﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

namespace Telerik.Windows.Documents.Common.FormatProviders
{
    /// <summary>
    /// Represents base binary format provider.
    /// </summary>
    public abstract class FormatProviderBase<T> : IFormatProvider<T>
    {
        /// <summary>
        /// Gets the supported extensions.
        /// </summary>
        /// <value>
        /// The supported extensions.
        /// </value>
        public abstract IEnumerable<string> SupportedExtensions
        {
            get;
        }

        /// <summary>
        /// Gets a value indicating whether format provider can import.
        /// </summary>
        /// <value>The value indicating whether can import.</value>
        public abstract bool CanImport { get; }

        /// <summary>
        /// Gets a value indicating whether format provider can export.
        /// </summary>
        /// <value>The value indicating whether can export.</value>
        public abstract bool CanExport { get; }

        /// <summary>
        /// Imports the specified input stream.
        /// </summary>
        /// <param name="input">The input stream.</param>
        /// <returns>The imported document.</returns>
        public T Import(Stream input)
        {
            if (!this.CanImport)
            {
                throw new InvalidOperationException("Import not supported.");
            }

            return this.ImportOverride(input);
        }

        /// <summary>
        /// Exports the specified document to the output stream.
        /// </summary>
        /// <param name="document">The document.</param>
        /// <param name="output">The output stream.</param>
        public void Export(T document, Stream output)
        {
            if (!this.CanExport)
            {
                throw new InvalidOperationException("Export not supported.");
            }

            this.ExportOverride(document, output);
        }

        /// <summary>
        /// Imports the specified input.
        /// </summary>
        /// <param name="input">The input.</param>
        /// <returns>The imported document.</returns>
        protected virtual T ImportOverride(Stream input)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Exports the specified document.
        /// </summary>
        /// <param name="document">The document.</param>
        /// <param name="output">The output.</param>
        protected virtual void ExportOverride(T document, Stream output)
        {
            throw new NotImplementedException();
        }
    }
}
