﻿using System;
using System.IO;
using System.Linq;
using System.Xml;
using Telerik.Windows.Documents.Common.FormatProviders.OpenXml.Utilities;
using Telerik.Windows.Documents.Common.Model.Data;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Parts;
using Telerik.Windows.Documents.Utilities;
#if NETSTANDARD
using Telerik.Zip;
#else
using Telerik.Windows.Zip;
#endif

namespace Telerik.Windows.Documents.FormatProviders.OpenXml.Export
{
    internal abstract class OpenXmlExporter<T, TContext>
        where T : OpenXmlPartsManager
        where TContext : IOpenXmlExportContext
    {
        private readonly DateTime zipPackageCreationDateTime;

        public OpenXmlExporter()
        {
            this.zipPackageCreationDateTime = DateTime.Now;
        }

        public void Export(Stream output, TContext context, OpenXmlExportSettings settings)
        {
            Guard.ThrowExceptionIfNull(output, "output");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(settings, "settings");

            using (ZipArchive archive = new ZipArchive(output, ZipArchiveMode.Create, true, null))
            {
                T partsManager = this.CreatePartsManager();

                this.InitializeParts(partsManager, context);

                foreach (IResource resource in context.Resources)
                {
                    string resourceName = this.CreateResourceName(resource);
                    partsManager.RegisterResource(resource.Extension);
#if WPF
                    try
                    {
                        using (Stream stream = new MemoryStream(resource.Data))
                        {
                            AddResourceToArchive(archive, resourceName, stream);
                        }
                    }
                    catch (System.Net.WebException)
                    {
                        continue;
                    }
#else
                    using (Stream stream = new MemoryStream(resource.Data))
                    {
                        AddResourceToArchive(archive, resourceName, stream);
                    }
#endif
                }

                for (int level = OpenXmlPartLevel.Level7; level >= OpenXmlPartLevel.Level1; level--)
                {
                    OpenXmlPartBase[] levelParts = partsManager.Parts.Where(p => p.Level == level).ToArray();
                    foreach (OpenXmlPartBase part in levelParts)
                    {
                        this.AddOpenPartToArchive(archive, ExportPart(part, context), part.Name);
                    }
                }
            }
        }

        protected static void AddResourceToArchive(ZipArchive archive, string name, Stream stream)
        {
            Guard.ThrowExceptionIfNull(archive, "package");
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");
            Guard.ThrowExceptionIfNull(stream, "stream");

            using (ZipArchiveEntry entry = archive.CreateEntry(PathHelper.NormalizePathForExport(name)))
            using (Stream entryStream = entry.Open())
            {
                stream.CopyTo(entryStream);
            }
        }

        protected static Stream ExportPart(OpenXmlPartBase part, TContext context)
        {
            Guard.ThrowExceptionIfNull(part, "part");

            using (MemoryStream memoryStream = new MemoryStream())
            {
                using (XmlWriter xmlWriter = XmlWriter.Create(memoryStream))
                {
                    IOpenXmlWriter writer = new OpenXmlPartWriter(xmlWriter);
                    part.Export(writer, context);
                    xmlWriter.Flush();

                    return new MemoryStream(memoryStream.GetBuffer(), 0, (int)memoryStream.Length);
                }
            }
        }

        protected abstract T CreatePartsManager();

        protected abstract string CreateResourceName(IResource resource);

        protected virtual void InitializeParts(T partsManager, TContext context)
        {
            Guard.ThrowExceptionIfNull(partsManager, "partsManager");
            Guard.ThrowExceptionIfNull(context, "context");

            partsManager.AddPart(new ContentTypesPart(partsManager));
        }

        protected void AddOpenPartToArchive(ZipArchive archive, Stream partStream, string partName)
        {
            Guard.ThrowExceptionIfNull(archive, "archive");

            DeflateSettings settings = new DeflateSettings();
            settings.CompressionLevel = CompressionLevel.Fastest;

            using (ZipArchiveEntry entry = archive.CreateEntry(PathHelper.NormalizePathForExport(partName), settings))
            using (Stream entryStream = entry.Open())
            {
                entry.LastWriteTime = this.zipPackageCreationDateTime;
                partStream.CopyTo(entryStream);
            }
        }
    }
}
