﻿using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Xml;
using Telerik.Windows.Documents.Common.FormatProviders.OpenXml.Utilities;
using Telerik.Windows.Documents.Common.Model.Data;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Import;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Parts;
using Telerik.Windows.Documents.Utilities;

#if NETSTANDARD
using Telerik.Zip;
#else
using Telerik.Windows.Zip;
#endif

namespace Telerik.Windows.Documents.FormatProviders.OpenXml
{
    internal abstract class OpenXmlImporter<T>
        where T : OpenXmlPartsManager
    {
        private const int ResourcesLevel = OpenXmlPartLevel.Level3;

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1822:MarkMembersAsStatic")]
        public void Import(Stream input, IOpenXmlImportContext context)
        {
            Guard.ThrowExceptionIfNull(input, "input");
            Guard.ThrowExceptionIfNull(context, "context");

            using (ZipArchive archive = new ZipArchive(input, ZipArchiveMode.Read, true, null))
            {
                Dictionary<string, ZipArchiveEntry> zipEntries = GetZipEntries(archive);

                T partsManager = this.CreatePartsManager();

                context.BeginImport();

                ContentTypesPart contentTypesPart = new ContentTypesPart(partsManager);
                partsManager.AddPart(contentTypesPart);
                ImportXlsxPartFromArchive(zipEntries[PathHelper.NormalizePathForExport(contentTypesPart.Name)], contentTypesPart, context);

                PrepareAllParts(partsManager, zipEntries, contentTypesPart);

                for (int level = OpenXmlPartLevel.Level2; level <= OpenXmlPartLevel.Level7; level++)
                {
                    foreach (OpenXmlPartBase part in partsManager.Parts.Where(p => p.Level == level))
                    {
                        ImportXlsxPartFromArchive(zipEntries[PathHelper.NormalizePathForExport(part.Name)], part, context);
                    }

                    if (level == ResourcesLevel)
                    {
                        ImportResources(context, zipEntries);
                    }
                }

                context.EndImport();

                foreach (KeyValuePair<string, ZipArchiveEntry> pair in zipEntries)
                {
                    pair.Value.Dispose();
                }
            }
        }

        protected abstract T CreatePartsManager();

        private static Dictionary<string, ZipArchiveEntry> GetZipEntries(ZipArchive archive)
        {
            Guard.ThrowExceptionIfNull(archive, "archive");

            Dictionary<string, ZipArchiveEntry> zipPackageEntries = new Dictionary<string, ZipArchiveEntry>();
            foreach (ZipArchiveEntry entry in archive.Entries)
            {
                zipPackageEntries[PathHelper.NormalizePath(entry.FullName)] = entry;
            }

            return zipPackageEntries;
        }

        private static void GetStreamFromZipPackage(ZipArchiveEntry entry, out Stream stream)
        {
            Guard.ThrowExceptionIfNull(entry, "entry");

            stream = new MemoryStream();

            using (Stream input = entry.Open())
            {
                input.CopyTo(stream);
            }

            stream.Seek(0, SeekOrigin.Begin);
        }

        private static void ImportXlsxPartFromArchive(ZipArchiveEntry zipEntry, OpenXmlPartBase part, IOpenXmlImportContext context)
        {
            Guard.ThrowExceptionIfNull(zipEntry, "zipEntry");
            Guard.ThrowExceptionIfNull(part, "part");

            Stream stream;
            GetStreamFromZipPackage(zipEntry, out stream);
            using (stream)
            {
                XmlReaderSettings settings = new XmlReaderSettings();
                settings.IgnoreComments = true;
                settings.IgnoreWhitespace = true;

                IOpenXmlReader reader = new OpenXmlPartReader(XmlReader.Create(stream, settings));
                part.Import(reader, context);
            }
        }

        private static void ImportResources(IOpenXmlImportContext context, Dictionary<string, ZipArchiveEntry> zipEntries)
        {
            foreach (KeyValuePair<string, ZipArchiveEntry> entry in zipEntries)
            {
                string resourceKey = PathHelper.NormalizePathFromImport(entry.Key);
                string extension = OpenXmlHelper.GetExtension(resourceKey);
                if (ResourcesFactory.CanCreateInstance(extension))
                {
                    Stream stream = null;
                    try
                    {
                        GetStreamFromZipPackage(entry.Value, out stream);
                        context.RegisterResource(resourceKey, ResourcesFactory.CreateInstance(stream, extension));
                    }
                    finally
                    {
                        if (stream != null)
                        {
                            stream.Close();
                        }
                    }
                }
            }
        }

        private static void PrepareAllParts(OpenXmlPartsManager partsManager, Dictionary<string, ZipArchiveEntry> zipEntries, ContentTypesPart contentTypesPart)
        {
            foreach (KeyValuePair<string, ZipArchiveEntry> entry in zipEntries)
            {
                string partName = PathHelper.NormalizePathFromImport(entry.Key);
                if (!partsManager.ContainsPart(partName))
                {
                    OpenXmlPartBase part = contentTypesPart.CreatePart(partName);
                    if (part != null)
                    {
                        partsManager.AddPart(part);
                    }
                }
            }
        }
    }
}
