﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Export;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Import;
using Telerik.Windows.Documents.Model.Drawing.Charts;

namespace Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements.Chart
{
    internal class PlotAreaElement : ChartElementBase
    {
        private const int MaxNumberOfAxes = 4;

        public const int PrimaryCategoryAxisIndex = 0;
        public const int PrimaryValueAxisIndex = 1;
        public const int SecondaryCategoryAxisIndex = 2;
        public const int SecondaryValueAxisIndex = 3;

        private readonly List<SeriesGroup> importedSeriesGroups;
        private readonly List<Axis> importedAxes;

        private DocumentChart chart;
        private int currentAxisCollectionIndex = 0;

        public PlotAreaElement(OpenXmlPartsManager partsManager)
            : base(partsManager)
        {
            this.importedSeriesGroups = new List<SeriesGroup>();
            this.importedAxes = new List<Axis>();
        }

        public override string ElementName
        {
            get
            {
                return OpenXmlElementNames.PlotAreaElementName;
            }
        }

        public override bool AlwaysExport
        {
            get
            {
                return true;
            }
        }

        public void CopyPropertiesFrom(DocumentChart chart)
        {
            this.chart = chart;
        }

        protected override IEnumerable<OpenXmlElementBase> EnumerateChildElements(IOpenXmlExportContext context)
        {
            foreach (SeriesGroup seriesGroup in this.chart.SeriesGroups)
            {
                switch (seriesGroup.SeriesType)
                {
                    case SeriesType.Bar:
                        {
                            BarChartElement barChartElement = this.CreateElement<BarChartElement>(OpenXmlElementNames.BarChartElementName);
                            barChartElement.Chart = this.chart;
                            barChartElement.SeriesGroup = seriesGroup as BarSeriesGroup;
                            barChartElement.CopyPropertiesFrom();
                            yield return barChartElement;
                            break;
                        }

                    case SeriesType.Line:
                        {
                            LineChartElement lineChartElement = this.CreateElement<LineChartElement>(OpenXmlElementNames.LineChartElementName);
                            lineChartElement.Chart = this.chart;
                            lineChartElement.SeriesGroup = seriesGroup as LineSeriesGroup;
                            lineChartElement.CopyPropertiesFrom();
                            yield return lineChartElement;
                            break;
                        }

                    case SeriesType.Area:
                        {
                            AreaChartElement areaChartElement = this.CreateElement<AreaChartElement>(OpenXmlElementNames.AreaChartElementName);
                            areaChartElement.Chart = this.chart;
                            areaChartElement.SeriesGroup = seriesGroup as AreaSeriesGroup;
                            areaChartElement.CopyPropertiesFrom();
                            yield return areaChartElement;
                            break;
                        }

                    case SeriesType.Pie:
                        {
                            DoughnutSeriesGroup doughnutGroup = seriesGroup as DoughnutSeriesGroup;
                            PieSeriesGroup pieGroup = seriesGroup as PieSeriesGroup;

                            if (doughnutGroup != null)
                            {
                                DoughnutChartElement doughnutChartElement = this.CreateElement<DoughnutChartElement>(OpenXmlElementNames.DoughnutChartElementName);
                                doughnutChartElement.Chart = this.chart;
                                doughnutChartElement.SeriesGroup = doughnutGroup;
                                doughnutChartElement.CopyPropertiesFrom();
                                yield return doughnutChartElement;
                            }
                            else if (pieGroup != null)
                            {
                                PieChartElement pieChartElement = this.CreateElement<PieChartElement>(OpenXmlElementNames.PieChartElementName);
                                pieChartElement.Chart = this.chart;
                                pieChartElement.SeriesGroup = pieGroup;
                                yield return pieChartElement;
                            }

                            break;
                        }

                    case SeriesType.Scatter:
                        {
                            ScatterChartElement scatterChartElement = this.CreateElement<ScatterChartElement>(OpenXmlElementNames.ScatterChartElementName);
                            scatterChartElement.Chart = this.chart;
                            scatterChartElement.CopyPropertiesFrom(seriesGroup as ScatterSeriesGroup);
                            yield return scatterChartElement;
                            break;
                        }

                    case SeriesType.Bubble:
                        {
                            BubbleChartElement scatterChartElement = this.CreateElement<BubbleChartElement>(OpenXmlElementNames.BubbleChartElementName);
                            scatterChartElement.Chart = this.chart;
                            scatterChartElement.CopyPropertiesFrom(seriesGroup as BubbleSeriesGroup);
                            yield return scatterChartElement;
                            break;
                        }

                    default:
                        {
                            throw new NotSupportedException();
                        }
                }
            }

            if (this.chart.PrimaryAxes != null)
            {
                yield return this.CreateElementForAxis(context, this.chart.PrimaryAxes.CategoryAxis, PrimaryCategoryAxisIndex, PrimaryValueAxisIndex, AxisPosition.Bottom);
                yield return this.CreateElementForAxis(context, this.chart.PrimaryAxes.ValueAxis, PrimaryValueAxisIndex, PrimaryCategoryAxisIndex, AxisPosition.Left);
            }

            if (this.chart.SecondaryAxes != null)
            {
                yield return this.CreateElementForAxis(context, this.chart.SecondaryAxes.CategoryAxis, SecondaryCategoryAxisIndex, SecondaryValueAxisIndex, AxisPosition.Bottom);
                yield return this.CreateElementForAxis(context, this.chart.SecondaryAxes.ValueAxis, SecondaryValueAxisIndex, SecondaryCategoryAxisIndex, AxisPosition.Left);
            }
        }

        private AxisElementBase CreateElementForAxis(IOpenXmlExportContext context, Axis axis, int axisIndex, int crossingAxisIndex, AxisPosition position)
        {
            switch (axis.AxisType)
            {
                case AxisType.Category:
                    {
                        CategoryAxisElement categoryAxisElement = this.CreateElement<CategoryAxisElement>(OpenXmlElementNames.CategoryAxisElementName);
                        categoryAxisElement.CopyPropertiesFrom(context, axis, axisIndex, crossingAxisIndex, position);
                        return categoryAxisElement;
                    }
                case AxisType.Value:
                    {
                        ValueAxisElement valueAxisElement = this.CreateElement<ValueAxisElement>(OpenXmlElementNames.ValueAxisElementName);
                        valueAxisElement.CopyPropertiesFrom(context, axis, axisIndex, crossingAxisIndex, position);
                        return valueAxisElement;
                    }
                case AxisType.Date:
                    {
                        DateAxisElement dateAxisElement = this.CreateElement<DateAxisElement>(OpenXmlElementNames.DateAxisElementName);
                        dateAxisElement.CopyPropertiesFrom(context, axis, axisIndex, crossingAxisIndex, position);
                        return dateAxisElement;
                    }
                default:
                    throw new NotSupportedException();
            }
        }

        public void CopyPropertiesTo(IOpenXmlImportContext context, DocumentChart chart)
        {
            context.PairSeriesGroupsWithAxes();

            foreach (SeriesGroup seriesGroup in this.importedSeriesGroups)
            {
                chart.SeriesGroups.Add(seriesGroup);
            }

            if (this.importedAxes.Count > 4 ||
                this.importedAxes.Count % 2 != 0)
            {
                throw new InvalidOperationException("The number of axes is invalid.");
            }

            if (this.importedAxes.Count == 2)
            {
                chart.PrimaryAxes = new AxisGroup();
            }
            else if (this.importedAxes.Count == 4)
            {
                chart.PrimaryAxes = new AxisGroup();
                chart.SecondaryAxes = new AxisGroup();
            }

            int axisCollectionIndex = 0;

            foreach (Axis axis in this.importedAxes)
            {
                switch (axisCollectionIndex)
                {
                    case PrimaryCategoryAxisIndex:
                        chart.PrimaryAxes.CategoryAxis = axis;
                        break;
                    case PrimaryValueAxisIndex:
                        chart.PrimaryAxes.ValueAxis = axis;
                        break;
                    case SecondaryCategoryAxisIndex:
                        chart.SecondaryAxes.CategoryAxis = axis;
                        break;
                    case SecondaryValueAxisIndex:
                        chart.SecondaryAxes.ValueAxis = axis;
                        break;
                    default:
                        throw new InvalidOperationException("The number of axes is invalid.");
                }

                axisCollectionIndex++;
            }
        }

        protected override void OnAfterReadChildElement(IOpenXmlImportContext context, OpenXmlElementBase childElement)
        {
            if (childElement is ChartGroupElementBase)
            {
                this.CreateSeriesGroup(context, childElement as ChartGroupElementBase);
            }
            else if (childElement is AxisElementBase)
            {
                this.CreateAxis(context, childElement as AxisElementBase);
            }
        }

        private void CreateSeriesGroup(IOpenXmlImportContext context, ChartGroupElementBase chartGroupElement)
        {
            SeriesGroup seriesGroup = null;

            switch (chartGroupElement.ElementName)
            {
                case OpenXmlElementNames.BarChartElementName:
                    seriesGroup = new BarSeriesGroup();
                    break;
                case OpenXmlElementNames.LineChartElementName:
                    seriesGroup = new LineSeriesGroup();
                    break;
                case OpenXmlElementNames.PieChartElementName:
                    seriesGroup = new PieSeriesGroup();
                    break;
                case OpenXmlElementNames.DoughnutChartElementName:
                    seriesGroup = new DoughnutSeriesGroup();
                    break;
                case OpenXmlElementNames.AreaChartElementName:
                    seriesGroup = new AreaSeriesGroup();
                    break;
                case OpenXmlElementNames.ScatterChartElementName:
                    seriesGroup = new ScatterSeriesGroup();
                    break;

                case OpenXmlElementNames.BubbleChartElementName:
                    seriesGroup = new BubbleSeriesGroup();
                    break;

                default:
                    throw new NotSupportedException();
            }

            chartGroupElement.CopyPropertiesTo(context, seriesGroup);
            this.importedSeriesGroups.Add(seriesGroup);
        }

        private void CreateAxis(IOpenXmlImportContext context, AxisElementBase axisElement)
        {
            Axis newAxis = null;

            switch (axisElement.ElementName)
            {
                case OpenXmlElementNames.CategoryAxisElementName:
                    newAxis = new CategoryAxis();
                    break;
                case OpenXmlElementNames.ValueAxisElementName:
                    newAxis = new ValueAxis();
                    break;
                case OpenXmlElementNames.DateAxisElementName:
                    newAxis = new DateAxis();
                    break;
                default:
                    throw new NotSupportedException();
            }

            AxisGroupName axisGroup;

            switch (this.currentAxisCollectionIndex)
            {
                case PrimaryCategoryAxisIndex:
                    axisGroup = AxisGroupName.Primary;
                    break;
                case PrimaryValueAxisIndex:
                    axisGroup = AxisGroupName.Primary;
                    break;
                case SecondaryCategoryAxisIndex:
                    axisGroup = AxisGroupName.Secondary;
                    break;
                case SecondaryValueAxisIndex:
                    axisGroup = AxisGroupName.Secondary;
                    break;
                default:
                    throw new InvalidOperationException("The number of axes is invalid.");
            }

            (axisElement as AxisElementBase).CopyPropertiesTo(context, newAxis, axisGroup);
            this.importedAxes.Add(newAxis);
            this.currentAxisCollectionIndex++;
        }

        protected override void ClearOverride()
        {
            this.importedAxes.Clear();
            this.importedSeriesGroups.Clear();
            this.currentAxisCollectionIndex = 0;
            this.chart = null;
        }
    }
}
