﻿using System;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Export;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Import;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements.Drawing;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements.Theme;
#if !SILVERLIGHT && !NETSTANDARD
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements.Chart;
using Telerik.Windows.Documents.Model.Drawing.Charts;
#endif
using Telerik.Windows.Documents.Model.Drawing.Shapes;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Model.Drawing.Theming;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements.Common
{
    internal class ShapePropertiesElement : OpenXmlElementBase
    {
        private readonly OpenXmlNamespace ns;

        private readonly OpenXmlChildElement<TransformElement> transform;
        private readonly OpenXmlChildElement<PresetGeometryElement> presetGeometry;
        private readonly OpenXmlChildElement<OutlineElement> outline;
        private readonly OpenXmlChildElement<SolidFillElement> solidFill;
        private readonly OpenXmlChildElement<NoFillElement> noFill;

        public ShapePropertiesElement(OpenXmlPartsManager partsManager, OpenXmlNamespace ns)
            : base(partsManager)
        {
            Guard.ThrowExceptionIfNull(ns, "ns");

            this.ns = ns;

            this.transform = this.RegisterChildElement<TransformElement>(OpenXmlElementNames.TransformElementName, OpenXmlUniquePoolIds.DrawingTransformElementUniquePoolId);
            this.presetGeometry = this.RegisterChildElement<PresetGeometryElement>(OpenXmlElementNames.PresetGeometryElementName);
            this.solidFill = this.RegisterChildElement<SolidFillElement>(OpenXmlElementNames.SolidFillElementName);
            this.noFill = this.RegisterChildElement<NoFillElement>(OpenXmlElementNames.NoFillElementName);

            // Outline should be after the fill, otherwise the fill will not be visible in Excel
            this.outline = this.RegisterChildElement<OutlineElement>(OpenXmlElementNames.OutlineElementName);
        }

        public override string ElementName
        {
            get
            {
                return OpenXmlElementNames.ShapePropertiesElementName;
            }
        }

        public override OpenXmlNamespace Namespace
        {
            get
            {
                return this.ns;
            }
        }

        public override bool AlwaysExport
        {
            get
            {
                return true;
            }
        }

        public TransformElement TransformElement
        {
            get
            {
                return this.transform.Element;
            }
        }

        public PresetGeometryElement PresetGeometryElement
        {
            get
            {
                return this.presetGeometry.Element;
            }
        }

        public OutlineElement OutlineElement
        {
            get
            {
                return this.outline.Element;
            }
        }

        public SolidFillElement SolidFillElement
        {
            get
            {
                return this.solidFill.Element;
            }
        }

        public NoFillElement NoFillElement
        {
            get
            {
                return this.noFill.Element;
            }
        }

        public void CopyPropertiesFrom(IOpenXmlExportContext context, ShapeBase shape)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(shape, "shape");

            this.CreateElement(this.transform);
            this.TransformElement.CopyPropertiesFrom(context, shape);

            this.CreateElement(this.presetGeometry);
            this.PresetGeometryElement.CopyPropertiesFrom(context, shape);

            this.ExportOutline(context, shape.Outline);
            this.ExportFill(context, shape.Fill);
        }

        public void CopyPropertiesTo(IOpenXmlImportContext context, ShapeBase shape)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(shape, "shape");

            if (this.TransformElement != null)
            {
                this.TransformElement.CopyPropertiesTo(context, shape);
                this.ReleaseElement(this.transform);
            }

            if (this.PresetGeometryElement != null)
            {
                this.PresetGeometryElement.CopyPropertiesTo(context, shape);
                this.ReleaseElement(this.presetGeometry);
            }

            this.ImportOutline(shape.Outline);

            shape.Fill = this.ImportFill();
        }

        private Fill ImportFill()
        {
            Fill fill = null;

            if (this.SolidFillElement != null)
            {
                fill = this.SolidFillElement.CreateSolidFill();
                this.ReleaseElement(this.solidFill);
            }

            if (this.NoFillElement != null)
            {
                fill = new NoFill();
                this.ReleaseElement(this.noFill);
            }

            return fill;
        }

#if !SILVERLIGHT && !NETSTANDARD
        public void CopyPropertiesFrom(IOpenXmlExportContext context, Axis axis)
        {
            Guard.ThrowExceptionIfNull(axis, "axis");
            Guard.ThrowExceptionIfNull(context, "context");

            this.ExportOutline(context, axis.Outline);
        }

        public void CopyPropertiesTo(IOpenXmlImportContext context, Axis axis)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(axis, "axis");

            this.ImportOutline(axis.Outline);
        }

        public void CopyPropertiesFrom(IOpenXmlExportContext context, ChartLine line)
        {
            Guard.ThrowExceptionIfNull(line, "line");
            Guard.ThrowExceptionIfNull(context, "context");

            this.ExportOutline(context, line.Outline);
        }

        public void CopyPropertiesTo(IOpenXmlImportContext context, ChartLine line)
        {
            Guard.ThrowExceptionIfNull(line, "line");
            Guard.ThrowExceptionIfNull(context, "context");

            this.ImportOutline(line.Outline);
        }

        public void CopyPropertiesFrom(IOpenXmlExportContext context, SeriesBase series)
        {
            Guard.ThrowExceptionIfNull(series, "series");
            Guard.ThrowExceptionIfNull(context, "context");

            this.ExportOutline(context, series.Outline);
        }

        public void CopyPropertiesTo(IOpenXmlImportContext context, SeriesInfo seriesInfo)
        {
            Guard.ThrowExceptionIfNull(seriesInfo, "seriesInfo");
            Guard.ThrowExceptionIfNull(context, "context");

            this.ImportOutline(seriesInfo.Outline);
        }

        public void CopyPropertiesFrom(IOpenXmlExportContext context, Marker marker)
        {
            Guard.ThrowExceptionIfNull(marker, "marker");
            Guard.ThrowExceptionIfNull(context, "context");

            this.ExportFill(context, marker.Fill);
            this.ExportOutline(context, marker.Outline);
        }

        public void CopyPropertiesTo(IOpenXmlImportContext context, Marker marker)
        {
            Guard.ThrowExceptionIfNull(marker, "marker");
            Guard.ThrowExceptionIfNull(context, "context");

            marker.Fill = this.ImportFill();
            this.ImportOutline(marker.Outline);
        }
#endif

        private void ImportOutline(Outline outline)
        {
            if (this.OutlineElement != null)
            {
                this.OutlineElement.CopyPropertiesTo(outline);
                this.ReleaseElement(this.outline);
            }
        }

        private void ExportOutline(IOpenXmlExportContext context, Outline outline)
        {
            if (outline != null &&
                (outline.Fill != null || outline.Width != null))
            {
                this.CreateElement(this.outline);
                this.OutlineElement.CopyPropertiesFrom(context, outline);
            }
        }

        private void ExportFill(IOpenXmlExportContext context, Fill fill)
        {
            if (fill != null)
            {
                switch (fill.ShapeFillType)
                {
                    case FillType.Solid:
                        this.CreateElement(this.solidFill);
                        this.SolidFillElement.CopyPropertiesFrom(context, fill as SolidFill);
                        break;
                    case FillType.NoFill:
                        this.CreateElement(this.noFill);
                        break;
                    default:
                        throw new NotSupportedException("This type of fill is not supported.");
                }
            }
        }
    }
}
