﻿using System.Collections.Generic;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Export;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Import;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements.Common;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Parts;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements.ContentTypes
{
    internal class TypesElement : OpenXmlPartRootElementBase
    {
        private readonly Dictionary<string, string> defaults;
        private readonly Dictionary<string, string> overrides;

        public TypesElement(OpenXmlPartsManager partsManager, OpenXmlPartBase part)
            : base(partsManager, part)
        {
            this.defaults = new Dictionary<string, string>();
            this.overrides = new Dictionary<string, string>();
        }

        public override OpenXmlNamespace Namespace
        {
            get
            {
                return OpenXmlNamespaces.ContentTypesNamespace;
            }
        }

        public override string ElementName
        {
            get
            {
                return OpenXmlElementNames.TypesElementName;
            }
        }

        public OpenXmlPartBase CreatePart(string partName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(partName, "partName");

            string contentType;
            if (!this.overrides.TryGetValue(partName, out contentType))
            {
                string extension = OpenXmlHelper.GetExtension(partName);
                if (!this.defaults.TryGetValue(extension, out contentType))
                {
                    return null;
                }
            }

            return this.PartsManager.CreatePart(contentType, partName);
        }

        public void RegisterDefault(string extension, string contentType)
        {
            Guard.ThrowExceptionIfNullOrEmpty(extension, "extension");
            Guard.ThrowExceptionIfNullOrEmpty(contentType, "contentType");

            this.defaults[extension] = contentType;
        }

        public void RegisterOverride(string partName, string contentType)
        {
            Guard.ThrowExceptionIfNullOrEmpty(partName, "partName");
            Guard.ThrowExceptionIfNullOrEmpty(contentType, "contentType");

            this.overrides[partName] = contentType;
        }

        protected override IEnumerable<OpenXmlElementBase> EnumerateChildElements(IOpenXmlExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            foreach (KeyValuePair<string, string> d in this.defaults)
            {
                DefaultElement defaultElement = this.CreateElement<DefaultElement>(OpenXmlElementNames.DefaultElementName);
                defaultElement.Extension = d.Key;
                defaultElement.ContentType = d.Value;

                yield return defaultElement;
            }

            foreach (KeyValuePair<string, string> o in this.overrides)
            {
                OverrideElement overrideElement = this.CreateElement<OverrideElement>(OpenXmlElementNames.OverrideElementName);
                overrideElement.ContentType = o.Value;
                overrideElement.PartName = o.Key;

                yield return overrideElement;
            }
        }

        protected override void OnAfterReadChildElement(IOpenXmlImportContext context, OpenXmlElementBase element)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(element, "element");

            switch (element.ElementName)
            {
                case OpenXmlElementNames.DefaultElementName:
                    DefaultElement defaultElement = (DefaultElement)element;
                    this.RegisterDefault(defaultElement.Extension, defaultElement.ContentType);

                    break;

                case OpenXmlElementNames.OverrideElementName:
                    OverrideElement overrideElement = (OverrideElement)element;
                    this.RegisterOverride(overrideElement.PartName, overrideElement.ContentType);

                    break;
            }
        }

        protected override void ClearOverride()
        {
            this.defaults.Clear();
            this.overrides.Clear();
        }
    }
}