﻿using Telerik.Windows.Documents.FormatProviders.OpenXml.Export;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Import;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements.Common;
using Telerik.Windows.Documents.Model.Drawing.Shapes;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements.Drawing
{
    internal class TransformElement : OpenXmlElementBase
    {
        private readonly OpenXmlNamespace ns;

        private readonly OpenXmlChildElement<SizeElement> extents;
        private readonly OpenXmlChildElement<OffsetElement> offset;

        private readonly BoolOpenXmlAttribute isHorizontallyFlipped;
        private readonly BoolOpenXmlAttribute isVerticallyFlipped;
        private readonly IntOpenXmlAttribute rotationAngle;

        public TransformElement(OpenXmlPartsManager partsManager, OpenXmlNamespace ns)
            : base(partsManager)
        {
            this.ns = ns;

            this.offset = this.RegisterChildElement<OffsetElement>(OpenXmlElementNames.OffsetElementName);
            this.extents = this.RegisterChildElement<SizeElement>(OpenXmlElementNames.ExtentsElementName);

            this.isHorizontallyFlipped = this.RegisterAttribute(new BoolOpenXmlAttribute("flipH"));
            this.isVerticallyFlipped = this.RegisterAttribute(new BoolOpenXmlAttribute("flipV"));
            this.rotationAngle = this.RegisterAttribute(new IntOpenXmlAttribute("rot"));
        }

        public override string ElementName
        {
            get
            {
                return OpenXmlElementNames.TransformElementName;
            }
        }

        public override OpenXmlNamespace Namespace
        {
            get
            {
                return this.ns;
            }
        }

        public bool IsHorizontallyFlipped
        {
            get
            {
                return this.isHorizontallyFlipped.Value;
            }
            set
            {
                this.isHorizontallyFlipped.Value = value;
            }
        }

        public bool IsVerticallyFlipped
        {
            get
            {
                return this.isVerticallyFlipped.Value;
            }
            set
            {
                this.isVerticallyFlipped.Value = value;
            }
        }

        public int RotationAngle
        {
            get
            {
                return this.rotationAngle.Value;
            }
            set
            {
                this.rotationAngle.Value = value;
            }
        }


        public SizeElement ExtentsElement
        {
            get
            {
                return this.extents.Element;
            }
        }

        public OffsetElement OffsetElement
        {
            get
            {
                return this.offset.Element;
            }
        }

        public void CopyPropertiesFrom(IOpenXmlExportContext context, ShapeBase shape)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(shape, "shape");

            if (!shape.SizeInternal.IsEmpty)
            {
                this.CreateElement(this.extents);
                this.ExtentsElement.ExtentWidth = (int)shape.Width;
                this.ExtentsElement.ExtentHeight = (int)shape.Height;
            }

            this.CreateElement(this.offset);
            this.OffsetElement.X = 0;
            this.OffsetElement.Y = 0;

            if (shape.IsHorizontallyFlipped)
            {
                this.IsHorizontallyFlipped = shape.IsHorizontallyFlipped;
            }

            if (shape.IsVerticallyFlipped)
            {
                this.IsVerticallyFlipped = shape.IsVerticallyFlipped;
            }

            if (shape.RotationAngle != 0)
            {
                this.RotationAngle = (int)((shape.RotationAngle) * 60000);
            }
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1822:MarkMembersAsStatic")]
        public void CopyPropertiesTo(IOpenXmlImportContext context, ShapeBase shape)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(shape, "shape");

            shape.RotationAngle = (double)this.RotationAngle / 60000;
            shape.IsHorizontallyFlipped = this.IsHorizontallyFlipped;
            shape.IsVerticallyFlipped = this.IsVerticallyFlipped;

            if (this.ExtentsElement != null && this.ExtentsElement.ExtentWidth != 0 && this.ExtentsElement.ExtentHeight != 0)
            {
                shape.Width = this.ExtentsElement.ExtentWidth;
                shape.Height = this.ExtentsElement.ExtentHeight;
            }
        }
    }
}
