﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Xml;
using Telerik.Windows.Documents.Core.Data;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Export;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Import;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Parts;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements
{
    internal abstract class OpenXmlElementBase
    {
        private readonly OpenXmlPartsManager partsManager;
        private readonly Dictionary<string, OpenXmlAttributeBase> attributes;
        private readonly QueueDictionary<string, IOpenXmlChildElement> childElements;

        public OpenXmlElementBase(OpenXmlPartsManager partsManager)
        {
            Guard.ThrowExceptionIfNull(partsManager, "partsManager");

            this.partsManager = partsManager;

            this.attributes = new Dictionary<string, OpenXmlAttributeBase>();
            this.childElements = new QueueDictionary<string, IOpenXmlChildElement>();
            this.InnerText = string.Empty;
        }

        public abstract string ElementName { get; }

        /// <summary>
        /// There is property with the same name in OpenXmlChildElement of T class. This property is introduced to fix a bug caused by the other property.
        /// When creating a new element the UniquePoolId is used which causes a check if there are elements with the same name in the object pool.
        /// When the element is released and is returned to the pool the ElementName propery is used. This way the create method never finds the already
        /// created objects and the pool for the ElementName objects becomes larger and larger and the objects are never leaving it. In order to remove this
        /// hack all UniquePoolId properties must be removed and the ElementsFactory should be refactored so instead of matcihng the element factory method not only by name, 
        /// to match them by name and namespace. Refactoring of the ElementsFactory might not be enough for removing the need for UniuqePoolId property.
        /// </summary>
        public string UniquePoolId { get; set; }

        // TODO: Think of removing this.
        public virtual bool AlwaysExport
        {
            get
            {
                return false;
            }
        }

        public virtual OpenXmlNamespace Namespace
        {
            get
            {
                return null;
            }
        }

        public virtual IEnumerable<OpenXmlNamespace> Namespaces
        {
            get
            {
                return Enumerable.Empty<OpenXmlNamespace>();
            }
        }

        public string InnerText
        {
            get;
            set;
        }

        public OpenXmlPartsManager PartsManager
        {
            get
            {
                return this.partsManager;
            }
        }

        public OpenXmlPartBase Part
        {
            get;
            set;
        }

        protected IEnumerable<IOpenXmlChildElement> ChildElements
        {
            get
            {
                return this.childElements.Values;
            }
        }

        public void Write(IOpenXmlWriter writer, IOpenXmlExportContext context)
        {
            this.OnBeforeWrite(context);
            if (this.ShouldExport(context))
            {
                this.WriteElementStart(writer);
                this.WriteAttributes(writer);
                this.WriteInnerText(writer);

                foreach (IOpenXmlChildElement childElement in this.ChildElements)
                {
                    if (childElement.HasElement)
                    {
                        childElement.GetElement().Write(writer, context);
                        this.ReleaseElement(childElement);
                    }
                }

                foreach (OpenXmlElementBase childElement in this.EnumerateChildElements(context))
                {
                    childElement.Write(writer, context);
                    this.ReleaseElement(childElement);
                }

                WriteElementEnd(writer);
            }
        }

        public void Read(IOpenXmlReader reader, IOpenXmlImportContext context)
        {
            ReadElementStart(reader);

            if (this.ShouldImport(context))
            {
                bool isEmptyElement = reader.IsEmptyElement();
                this.ReadAttributes(reader, context);
                this.OnAfterReadAttributes(context);
                reader.Read();

                if (!isEmptyElement)
                {
                    XmlNodeType nodeType = reader.GetNodeType();
                    if (nodeType == XmlNodeType.Text || nodeType == XmlNodeType.SignificantWhitespace)
                    {
                        this.ReadInnerText(reader);
                        reader.Read();
                    }
                    else
                    {
                        this.OnBeforeReadInnerElements(context);
                        this.ReadChildElements(reader, context);
                        this.OnAfterReadInnerElements(context);
                    }
                }

                this.OnAfterRead(context);
            }
            else
            {
                reader.SkipElement();
            }
        }

        public void Clear()
        {
            foreach (OpenXmlAttributeBase attribute in this.attributes.Values)
            {
                attribute.ResetValue();
            }

            this.InnerText = string.Empty;
            this.Part = null;

            this.ClearOverride();
        }

        protected virtual void ClearOverride()
        {
        }

        protected virtual bool ShouldExport(IOpenXmlExportContext context)
        {
            if (this.AlwaysExport)
            {
                return true;
            }

            if (!string.IsNullOrEmpty(this.InnerText))
            {
                return true;
            }

            if (this.attributes.Values.Any(a => a.ShouldExport()))
            {
                return true;
            }

            return this.ChildElements.Any(ce => ce.HasElement && ce.GetElement().ShouldExport(context));
        }

        protected virtual IEnumerable<OpenXmlElementBase> EnumerateChildElements(IOpenXmlExportContext context)
        {
            return Enumerable.Empty<OpenXmlElementBase>();
        }

        protected virtual void OnBeforeWrite(IOpenXmlExportContext context)
        {
        }

        protected virtual bool ShouldImport(IOpenXmlImportContext context)
        {
            return !context.IsImportSuspended;
        }

        protected virtual void OnAfterReadAttributes(IOpenXmlImportContext context)
        {
        }

        protected virtual void OnBeforeReadInnerElements(IOpenXmlImportContext context)
        {
        }

        protected virtual void OnBeforeReadChildElement(IOpenXmlImportContext context, OpenXmlElementBase chiledElement)
        {
        }

        protected virtual void OnAfterReadChildElement(IOpenXmlImportContext context, OpenXmlElementBase childElement)
        {
        }

        protected virtual void OnAfterReadInnerElements(IOpenXmlImportContext context)
        {
        }

        protected virtual void OnAfterRead(IOpenXmlImportContext context)
        {
        }

        protected OpenXmlAttribute<T> RegisterAttribute<T>(string name, OpenXmlNamespace ns, T defaultValue, bool isRequired = false)
        {
            OpenXmlAttribute<T> res = new OpenXmlAttribute<T>(name, ns, defaultValue, isRequired);
            this.RegisterAttribute(res);

            return res;
        }

        protected OpenXmlAttribute<T> RegisterAttribute<T>(string name, OpenXmlNamespace ns, bool isRequired = false)
        {
            OpenXmlAttribute<T> res = new OpenXmlAttribute<T>(name, ns, isRequired);
            this.RegisterAttribute(res);

            return res;
        }

        protected OpenXmlAttribute<T> RegisterAttribute<T>(string name, T defaultValue, bool isRequired = false)
        {
            return this.RegisterAttribute<T>(name, null, defaultValue, isRequired);
        }

        protected OpenXmlAttribute<T> RegisterAttribute<T>(string name, bool isRequired = false)
        {
            return this.RegisterAttribute<T>(name, null, default(T), isRequired);
        }

        protected OpenXmlCountAttribute RegisterCountAttribute()
        {
            OpenXmlCountAttribute countAttribute = new OpenXmlCountAttribute();
            this.RegisterAttribute(countAttribute);

            return countAttribute;
        }

        protected SpaceAttribute RegisterSpaceAttribute()
        {
            SpaceAttribute spaceAttribute = new SpaceAttribute(this);
            this.RegisterAttribute(spaceAttribute);

            return spaceAttribute;
        }

        protected T RegisterAttribute<T>(T attribute) where T : OpenXmlAttributeBase
        {
            Guard.ThrowExceptionIfNull(attribute, "attribute");

            string namespaceURI = null;
            if (attribute.Namespace != null)
            {
                namespaceURI = attribute.Namespace.Value;
            }

            string attributeKey = GetAttributeKey(attribute.Name, namespaceURI);
            this.attributes[attributeKey] = attribute;

            return attribute;
        }

        protected OpenXmlChildElement<T> RegisterChildElement<T>(string elementName)
            where T : OpenXmlElementBase
        {
            return this.RegisterChildElement<T>(elementName, elementName);
        }

        protected OpenXmlChildElement<T> RegisterChildElement<T>(string elementName, string uniqueElementName)
            where T : OpenXmlElementBase
        {
            Guard.ThrowExceptionIfNullOrEmpty(elementName, "elementName");
            Guard.ThrowExceptionIfNullOrEmpty(uniqueElementName, "uniqueElementName");

            OpenXmlChildElement<T> child = new OpenXmlChildElement<T>(elementName, uniqueElementName);
            this.RegisterChildElementInternal(child);

            return child;
        }

        protected OpenXmlChildElement<T> RegisterChildElement<T>(OpenXmlChildElement<T> childElement)
            where T : OpenXmlElementBase
        {
            Guard.ThrowExceptionIfNull(childElement, "childElement");

            this.RegisterChildElementInternal(childElement);

            return childElement;
        }

        protected virtual OpenXmlElementBase CreateElement(string elementName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(elementName, "elementName");

            OpenXmlElementBase element = this.partsManager.CreateElement(elementName, this.Part);

#if DEBUG
            if (element != null)
            {
                System.Diagnostics.Debug.Assert(element.IsEmpty(), "Element is not empty. Propably it was not released correctly.");
            }
#endif

            return element;
        }

        protected T CreateElement<T>(string elementName) where T : OpenXmlElementBase
        {
            return (T)this.CreateElement(elementName);
        }

        protected virtual void ReleaseElementOverride(OpenXmlElementBase element)
        {
            Guard.ThrowExceptionIfNull(element, "element");

            foreach (IOpenXmlChildElement childElement in element.ChildElements)
            {
                element.ReleaseElement(childElement);
            }

            this.partsManager.ReleaseElement(element);
        }

        // TODO: Think of moving to OpenXmlChildElement.
        protected void CreateElement(IOpenXmlChildElement childElement)
        {
            OpenXmlElementBase element = this.CreateElement(childElement.UniquePoolId);
            element.UniquePoolId = childElement.UniquePoolId;
            childElement.SetElement(element);
        }

        protected void ReleaseElement(IOpenXmlChildElement childElement)
        {
            if (childElement.HasElement)
            {
                this.ReleaseElement(childElement.GetElement());
                childElement.ClearElement();
            }
        }

        protected bool TryGetChildElement(string elementName, out IOpenXmlChildElement childElement)
        {
            return this.childElements.TryGetValue(elementName, out childElement);
        }

        private static void WriteElementEnd(IOpenXmlWriter writer)
        {
            writer.WriteElementEnd();
        }

        private static void ReadElementStart(IOpenXmlReader reader)
        {
            while (reader.GetNodeType() != XmlNodeType.Element)
            {
                if (!reader.Read())
                {
                    throw new InvalidOperationException("Element not found");
                }
            }
        }

        private void RegisterChildElementInternal(IOpenXmlChildElement child)
        {
            this.childElements[child.ElementName] = child;
        }

        private void ReleaseElement(OpenXmlElementBase element)
        {
            this.ReleaseElementOverride(element);
        }

        // TODO: Think of refactoring this method.
        private void ReadChildElements(IOpenXmlReader reader, IOpenXmlImportContext context)
        {
            string elementName;
            OpenXmlElementBase element;
            IOpenXmlChildElement childElement;
            while (!reader.IsElementOfType(XmlNodeType.EndElement))
            {
                switch (reader.GetNodeType())
                {
                    case XmlNodeType.Element:
                        if (reader.GetElementName(out elementName) && this.TryGetChildElement(elementName, out childElement))
                        {
                            this.CreateElement(childElement);
                            element = childElement.GetElement();

                            this.OnBeforeReadChildElement(context, element);
                            element.Read(reader, context);
                            this.OnAfterReadChildElement(context, element);
                        }
                        else if (reader.GetElementName(out elementName))
                        {
                            element = this.CreateElement(elementName);
                            if (element != null)
                            {
                                this.OnBeforeReadChildElement(context, element);
                                element.Read(reader, context);
                                this.OnAfterReadChildElement(context, element);

                                this.ReleaseElement(element);
                            }
                            else
                            {
                                reader.SkipElement();
                            }
                        }
                        else
                        {
                            reader.SkipElement();
                        }

                        break;

                    default:
                        reader.SkipElement();

                        break;
                }
            }

            reader.Read();
        }

        private void ReadAttributes(IOpenXmlReader reader, IOpenXmlImportContext context)
        {
            if (reader.HasAttributes() && reader.MoveToFirstAttribute())
            {
                OpenXmlAttributeBase openXmlAttribute;
                string attribute;
                string value;
                string namespaceURI;

                do
                {
                    if (!reader.GetAttributeNameAndValue(out attribute, out namespaceURI, out value))
                    {
                        throw new InvalidOperationException("Attribute not found.");
                    }

                    string attributeKey = GetAttributeKey(attribute, namespaceURI);
                    if (this.attributes.TryGetValue(attributeKey, out openXmlAttribute))
                    {
                        openXmlAttribute.SetStringValue(context, value);
                    }
                }
                while (reader.MoveToNextAttribute());
            }
        }

        private static string GetAttributeKey(string name, string namespaceURI)
        {
            if (string.IsNullOrEmpty(namespaceURI))
            {
                return name;
            }

            return string.Format("{0}:{1}", namespaceURI, name);
        }

        private void ReadInnerText(IOpenXmlReader reader)
        {
            string value;
            if (reader.GetElementValue(out value))
            {
                this.InnerText = value;
                reader.Read();
            }
            else
            {
                reader.SkipElement();
            }
        }

        private void WriteElementStart(IOpenXmlWriter writer)
        {
            writer.WriteElementStart(this);
            foreach (OpenXmlNamespace ns in this.Namespaces)
            {
                writer.WriteNamespace(ns);
            }
        }

        private void WriteAttributes(IOpenXmlWriter writer)
        {
            foreach (KeyValuePair<string, OpenXmlAttributeBase> attribute in this.attributes)
            {
                if (attribute.Value.ShouldExport())
                {
                    writer.WriteAttribute(attribute.Value);
                }
            }
        }

        private void WriteInnerText(IOpenXmlWriter writer)
        {
            writer.WriteValue(this.InnerText);
        }

#if DEBUG

        private bool IsEmpty()
        {
            foreach (IOpenXmlChildElement childElement in this.ChildElements)
            {
                if (childElement.HasElement)
                {
                    return false;
                }
            }

            foreach (OpenXmlAttributeBase attribute in this.attributes.Values)
            {
                if (attribute.HasValue)
                {
                    return false;
                }
            }

            return true;
        }

#endif
    }
}