﻿using System.Collections.Generic;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Parts;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.FormatProviders.OpenXml.Model
{
    internal abstract class OpenXmlElementsPoolBase<T> where T : OpenXmlPartsManager
    {
        private readonly T partsManager;
        private readonly Dictionary<string, Stack<OpenXmlElementBase>> store;
        private readonly object syncRoot;

        public OpenXmlElementsPoolBase(T partsManager)
        {
            Guard.ThrowExceptionIfNull(partsManager, "partsManager");

            this.partsManager = partsManager;
            this.store = new Dictionary<string, Stack<OpenXmlElementBase>>();
            this.syncRoot = new object();
        }

        public T PartsManager
        {
            get
            {
                return this.partsManager;
            }
        }

        public OpenXmlElementBase CreateElement(string elementName, OpenXmlPartBase part)
        {
            Guard.ThrowExceptionIfNullOrEmpty(elementName, "elementName");
            Guard.ThrowExceptionIfNull(part, "part");

            OpenXmlElementBase element;
            Stack<OpenXmlElementBase> elementsStack;
            if (!this.store.TryGetValue(elementName, out elementsStack) || elementsStack.Count == 0)
            {
                element = this.CreateInstance(elementName);
            }
            else
            {
                lock (this.syncRoot)
                {
                    if (elementsStack.Count == 0)
                    {
                        element = this.CreateInstance(elementName);
                    }
                    else
                    {
                        element = elementsStack.Pop();
                    }
                }
            }

            if (element != null)
            {
                element.Part = part;
            }

            return element;
        }

        public void ReleaseElement(OpenXmlElementBase element)
        {
            Guard.ThrowExceptionIfNull(element, "element");

            string elementName = element.UniquePoolId ?? element.ElementName;

            element.Clear();
            Stack<OpenXmlElementBase> elementsStack;
            if (!this.store.TryGetValue(elementName, out elementsStack))
            {
                lock (this.syncRoot)
                {
                    if (!this.store.TryGetValue(elementName, out elementsStack))
                    {
                        elementsStack = new Stack<OpenXmlElementBase>();
                        this.store[elementName] = elementsStack;
                    }
                }
            }

            this.store[elementName].Push(element);
        }

        protected virtual OpenXmlElementBase CreateInstance(string elementName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(elementName, "elementName");

            if (OpenXmlElementsFactory.CanCreateInstance(elementName))
            {
                return OpenXmlElementsFactory.CreateInstance(elementName, this.PartsManager);
            }

            return null;
        }
    }
}
