﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Telerik.Windows.Documents.Common.FormatProviders.OpenXml.Utilities;
using Telerik.Windows.Documents.Core.Data;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Parts;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.FormatProviders.OpenXml.Model
{
    internal abstract class OpenXmlPartsManager
    {
        private readonly HashSet<string> usedParts;
        private readonly StackCollection<OpenXmlPartBase> parts;

        private ContentTypesPart contentTypesPart;

        public OpenXmlPartsManager()
        {
            this.parts = new StackCollection<OpenXmlPartBase>();
            this.usedParts = new HashSet<string>();
        }

        public IEnumerable<OpenXmlPartBase> Parts
        {
            get
            {
                return this.parts;
            }
        }

        public bool ContainsPart(string partName)
        {
            return this.usedParts.Contains(partName);
        }

        public void AddPart(OpenXmlPartBase part)
        {
            Guard.ThrowExceptionIfNull(part, "part");

            this.parts.AddLast(part);
            this.MarkPartNameAsUsed(part.Name);
        }

        public void RegisterPart(OpenXmlPartBase part)
        {
            Guard.ThrowExceptionIfNull(part, "part");

            if (this.contentTypesPart == null)
            {
                this.contentTypesPart = this.GetRequiredElementByName<ContentTypesPart>(OpenXmlPartNames.ContentTypesPartName);
            }

            this.parts.AddLast(part);
            this.MarkPartNameAsUsed(part.Name);
            this.contentTypesPart.RegisterPart(part);
        }

        public void RegisterResource(string extension)
        {
            if (this.contentTypesPart == null)
            {
                this.contentTypesPart = this.GetRequiredElementByName<ContentTypesPart>(OpenXmlPartNames.ContentTypesPartName);
            }

            this.contentTypesPart.RegisterResource(extension);
        }

        public string CreateApplicationRelationship(string target, string type, string targetMode = null)
        {
            Guard.ThrowExceptionIfNullOrEmpty(target, "target");
            Guard.ThrowExceptionIfNullOrEmpty(type, "type");

            RelationshipsPart relationshipPart = this.GetRelationshipsPart(OpenXmlPartNames.ApplicationRelationshipsPartName);

            return relationshipPart.CreateRelationship(target, type, targetMode);
        }

        public string CreateRelationship(string targetPartName, string target, string type, string targetMode = null)
        {
            Guard.ThrowExceptionIfNullOrEmpty(targetPartName, "targetPartName");
            Guard.ThrowExceptionIfNullOrEmpty(target, "target");
            Guard.ThrowExceptionIfNullOrEmpty(type, "type");

            RelationshipsPart relationshipPart = this.GetRelationshipsPartForPart(targetPartName);

            return relationshipPart.CreateRelationship(target, type, targetMode);
        }

        public static string GetRelationshipRelativeTarget(string relationshipPartName, string target)
        {
            Guard.ThrowExceptionIfNull(relationshipPartName, "relationshipPartName");
            Guard.ThrowExceptionIfNullOrEmpty(target, "target");
            Guard.ThrowExceptionIfNullOrEmpty(target, "target");

            string relationPartDirectory = Path.GetDirectoryName(relationshipPartName);
            string fullPathToSubstract = relationPartDirectory.Substring(0, relationPartDirectory.Length - PathHelper.RelationDirectory.Length);
            fullPathToSubstract = fullPathToSubstract.Replace('\\', '/');
            string targetResult = target.Replace('\\', '/');

            string pathToSubstract = string.Empty;
            string substractedPath = string.Empty;
            while (pathToSubstract == substractedPath && !string.IsNullOrEmpty(fullPathToSubstract))
            {
                int firstIndex = GetDirectorySeparatorIndex(fullPathToSubstract, 0);
                if (firstIndex == -1 && targetResult.Length < firstIndex + 1)
                {
                    break;
                }

                pathToSubstract = fullPathToSubstract.Substring(0, firstIndex + 1);
                substractedPath = targetResult.Substring(0, firstIndex + 1);

                if (pathToSubstract == substractedPath)
                {
                    targetResult = targetResult.Substring(firstIndex + 1);
                    fullPathToSubstract = fullPathToSubstract.Substring(firstIndex + 1);
                }
                else
                {
                    break;
                }
            }

            int directoryCount = fullPathToSubstract.Where(ch => ch == '/').Count();
            if (directoryCount > 0)
            {
                for (int i = 0; i < directoryCount; i++)
                {
                    targetResult = "../" + targetResult;
                }
            }

            return targetResult;
        }

        public string GetRelationshipTarget(string targetPartName, string relationshipId)
        {
            Guard.ThrowExceptionIfNullOrEmpty(targetPartName, "targetPartName");
            Guard.ThrowExceptionIfNullOrEmpty(relationshipId, "relationshipId");

            RelationshipsPart relationshipPart = this.GetRelationshipsPartForPart(targetPartName);

            return relationshipPart.GetRelationshipTarget(relationshipId);
        }

        public string GetRelationshipId(string targetPartName, string target)
        {
            Guard.ThrowExceptionIfNullOrEmpty(targetPartName, "targetPartName");
            Guard.ThrowExceptionIfNullOrEmpty(target, "target");

            RelationshipsPart relationshipPart = this.GetRelationshipsPartForPart(targetPartName);

            return relationshipPart.GetRelationshipId(target);
        }

        public abstract OpenXmlElementBase CreateElement(string elementName, OpenXmlPartBase part);

        public abstract void ReleaseElement(OpenXmlElementBase element);

        public virtual OpenXmlPartBase CreatePart(string partType, string partName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(partType, "partType");
            Guard.ThrowExceptionIfNullOrEmpty(partName, "partName");

            if (OpenXmlPartsFactory.CanCreateInstance(partType))
            {
                return OpenXmlPartsFactory.CreateInstance(partType, this, partName);
            }

            return null;
        }

        public T GetPart<T>(string partName)
            where T : OpenXmlPartBase
        {
            return this.GetElementByName<T>(partName);
        }

        protected T GetElementByName<T>(string partName) where T : OpenXmlPartBase
        {
            return (T)this.parts.GetElementByName(partName);
        }

        protected T GetRequiredElementByName<T>(string partName)
            where T : OpenXmlPartBase
        {
            T result = this.GetElementByName<T>(partName);
            if (result == null)
            {
                throw new InvalidOperationException();
            }

            return result;
        }

        private RelationshipsPart GetRelationshipsPartForPart(string partName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(partName, "partName");

            string fileName = Path.GetFileName(partName);
            string directoryName = Path.GetDirectoryName(partName);
            string relationshipPartName = PathHelper.NormalizePath(Path.Combine(directoryName, PathHelper.RelationDirectory, fileName + ".rels"));

            return this.GetRelationshipsPart(relationshipPartName);
        }

        private RelationshipsPart GetRelationshipsPart(string name)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");

            if (!this.ContainsPart(name))
            {
                this.RegisterPart(new RelationshipsPart(this, name));
            }

            return this.GetElementByName<RelationshipsPart>(name);
        }

        private void MarkPartNameAsUsed(string partName)
        {
            this.usedParts.Add(partName);
        }

        private static int GetDirectorySeparatorIndex(string pathToSubstract, int startIndex)
        {
            int index = pathToSubstract.IndexOf(Path.DirectorySeparatorChar, startIndex);
            if (index == -1)
            {
                index = pathToSubstract.IndexOf(Path.AltDirectorySeparatorChar, startIndex);
            }

            return index;
        }
    }
}