﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Common.FormatProviders.XmlBased.Export;
using Telerik.Windows.Documents.Common.FormatProviders.XmlBased.Model.Attributes;
using Telerik.Windows.Documents.Common.FormatProviders.XmlBased.Model.Namespaces;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Common.FormatProviders.XmlBased.Model.Elements
{
    internal abstract class XmlElementBase<TAttribute>
        where TAttribute : XmlAttribute
    {
        private readonly Dictionary<string, TAttribute> attributes;

        public XmlElementBase()
        {
            this.attributes = new Dictionary<string, TAttribute>();
        }

        public abstract string Name
        {
            get;
        }

        public string InnerText
        {
            get;
            set;
        }

        public virtual XmlNamespace Namespace
        {
            get
            {
                return null;
            }
        }

        public virtual IEnumerable<XmlNamespace> Namespaces
        {
            get
            {
                return Enumerable.Empty<XmlNamespace>();
            }
        }

        public IEnumerable<TAttribute> Attributes
        {
            get
            {
                return this.attributes.Values;
            }
        }

        public void Clear()
        {
            foreach (XmlAttribute attribute in this.attributes.Values)
            {
                attribute.ResetValue();
            }

            this.ClearOverride();
        }

        protected static void WriteAttribute(IXmlWriter writer, TAttribute attribute)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(attribute, "attribute");

            if (attribute.Namespace == null)
            {
                writer.WriteAttribute(attribute.Name, attribute.GetValue());
            }
            else
            {
                writer.WriteAttribute(attribute.Namespace.LocalName, attribute.Name, attribute.GetValue());
            }
        }

        protected virtual void ClearOverride()
        {
        }

        protected void RegisterAttribute(TAttribute attribute)
        {
            Guard.ThrowExceptionIfNull(attribute, "attribute");
            if (attribute.Namespace != null)
            {
                this.attributes[string.Format("{0}:{1}", attribute.Namespace.LocalName, attribute.Name)] = attribute;
            }
            else
            {
                this.attributes[attribute.Name] = attribute;
            }
        }

        protected void TrySetAttributeValue(string name, string value)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");

            TAttribute attribute;
            if (this.attributes.TryGetValue(name, out attribute))
            {
                attribute.SetValue(value);
            }
        }

        protected void WriteElementStart(IXmlWriter writer)
        {
            if (this.Namespace != null)
            {
                writer.WriteElementStart(this.Namespace.LocalName, this.Name, this.Namespace.Value);
            }
            else
            {
                writer.WriteElementStart(this.Name);
            }

            foreach (XmlNamespace ns in this.Namespaces)
            {
                if (string.IsNullOrEmpty(ns.Prefix))
                {
                    writer.WriteAttribute(ns.LocalName, ns.Value);
                }
                else
                {
                    writer.WriteAttribute(ns.Prefix, ns.LocalName, ns.Value);
                }
            }
        }
    }
}
