﻿using System;
using Telerik.Windows.Documents.Model.Drawing.Shapes;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Model.Drawing.Theming;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Model.Drawing.Charts
{
    /// <summary>
    /// Represents one of the axes of the chart.
    /// </summary>
    public abstract class Axis
    {
        private double? min;
        private double? max;
        private bool isVisible;
        private string numberFormat;
        private readonly Outline outline;
        private readonly ChartLine majorGridlines;

        /// <summary>
        /// Initializes a new instance of the <see cref="Axis" /> class.
        /// </summary>
        protected Axis()
        {
            this.isVisible = true;
            this.outline = new Outline();
            this.outline.ShapeOutlineChanged += this.ShapeOutline_ShapeOutlineChanged;
            this.majorGridlines = new ChartLine();
            this.majorGridlines.ChartLineChanged += this.MajorGridlines_ChartLineChanged;
            this.numberFormat = string.Empty;
        }

        /// <summary>
        /// Gets the type of the axis.
        /// </summary>
        public abstract AxisType AxisType
        {
            get;
        }

        /// <summary>
        /// Gets or sets the smallest value of the axis. If the value is null, the smallest value will be determined automatically.
        /// For data axes, this value is an OLE Automation date value. For categorical axis, this number is the number of the category in the
        /// succession of the category collection.
        /// </summary>
        public double? Min
        {
            get
            {
                return this.min;
            }
            set
            {
                if (this.min != value)
                {
                    this.min = value;
                    this.OnAxisChanged();
                }
            }
        }

        /// <summary>
        /// Gets or sets the largest value of the axis. If the value is null, the largest value will be determined automatically.
        /// For data axes, this value is an OLE Automation date value. For categorical axis, this number is the number of the category in the
        /// succession of the category collection.
        /// </summary>
        public double? Max
        {
            get
            {
                return this.max;
            }
            set
            {
                if (this.max != value)
                {
                    this.max = value;
                    this.OnAxisChanged();
                }
            }
        }

        /// <summary>
        /// Gets or sets the value indicating whether the axis is visible.
        /// </summary>
        public bool IsVisible
        {
            get
            {
                return this.isVisible;
            }
            set
            {
                if (this.isVisible != value)
                {
                    this.isVisible = value;
                    this.OnAxisChanged();
                }
            }
        }

        /// <summary>
        /// Gets or sets the format string that is applied to the values of the axis.
        /// </summary>
        public string NumberFormat
        {
            get
            {
                return this.numberFormat;
            }
            set
            {
                if (this.numberFormat != value)
                {
                    Guard.ThrowExceptionIfNull(value, "value");

                    this.numberFormat = value;
                    this.OnAxisChanged();
                }
            }
        }

        /// <summary>
        /// Gets the outline of the axis.
        /// </summary>
        public Outline Outline
        {
            get
            {
                return this.outline;
            }
        }

        /// <summary>
        /// Gets the major gridlines of the axis.
        /// </summary>
        public ChartLine MajorGridlines
        {
            get
            {
                return this.majorGridlines;
            }
        }

        /// <summary>
        /// Creates a deep copy of the object.
        /// </summary>
        /// <returns>The deep copy of the object.</returns>
        public Axis Clone()
        {
            Axis newAxis = (Axis)Activator.CreateInstance(this.GetType());

            newAxis.Min = this.Min;
            newAxis.Max = this.Max;
            newAxis.IsVisible = this.IsVisible;
            newAxis.NumberFormat = this.NumberFormat;
            newAxis.Outline.Fill = this.Outline.Fill.Clone();
            newAxis.Outline.Width = this.Outline.Width;
            newAxis.MajorGridlines.Outline.Fill = this.MajorGridlines.Outline.Fill.Clone();
            newAxis.MajorGridlines.Outline.Width = this.MajorGridlines.Outline.Width;

            return newAxis;
        }

        private void ShapeOutline_ShapeOutlineChanged(object sender, EventArgs e)
        {
            this.OnAxisChanged();
        }

        private void MajorGridlines_ChartLineChanged(object sender, EventArgs e)
        {
            this.OnAxisChanged();
        }

        internal event EventHandler AxisChanged;

        internal void OnAxisChanged()
        {
            if (this.AxisChanged != null)
            {
                this.AxisChanged(this, EventArgs.Empty);
            }
        }
    }
}