﻿using System;

namespace Telerik.Windows.Documents.Model.Drawing.Charts
{
    /// <summary>
    /// Represents a chart.
    /// </summary>
    public class DocumentChart
    {
        private readonly SeriesGroupCollection seriesGroups;

        private AxisGroup primaryAxes;
        private AxisGroup secondaryAxes;
        private Legend legend;
        private Title title;

        /// <summary>
        /// Initializes a new instance of the <see cref="DocumentChart" /> class.
        /// </summary>
        public DocumentChart()
        {
            this.seriesGroups = new SeriesGroupCollection();
            this.seriesGroups.CollectionChanged += this.SeriesGroups_CollectionChanged;
            this.seriesGroups.SeriesGroupsChanged += this.SeriesGroupss_SeriesGroupChanged;
        }

        /// <summary>
        /// Represents a collection of the groups in which the series of the chart are organized.
        /// </summary>
        public SeriesGroupCollection SeriesGroups
        {
            get
            {
                return this.seriesGroups;
            }
        }

        /// <summary>
        /// Represents the primary group of axes of the chart.
        /// </summary>
        public AxisGroup PrimaryAxes
        {
            get
            {
                return this.primaryAxes;
            }
            set
            {
                if (this.primaryAxes != value)
                {
                    if (this.primaryAxes != null)
                    {
                        this.primaryAxes.CollectionChanged -= this.Axes_CollectionChanged;
                        this.primaryAxes.AxesChanged -= this.Axes_AxesChanged;
                    }

                    this.primaryAxes = value;

                    if (this.primaryAxes != null)
                    {
                        this.primaryAxes.CollectionChanged += this.Axes_CollectionChanged;
                        this.primaryAxes.AxesChanged += this.Axes_AxesChanged;
                    }
                }
            }
        }

        /// <summary>
        /// Represents the secondary group of axes of the chart. It is used when there is more than one group of series (combo chart).
        /// </summary>
        public AxisGroup SecondaryAxes
        {
            get
            {
                return this.secondaryAxes;
            }
            set
            {
                if (this.secondaryAxes != null)
                {
                    this.secondaryAxes.CollectionChanged -= this.Axes_CollectionChanged;
                    this.secondaryAxes.AxesChanged -= this.Axes_AxesChanged;
                }

                this.secondaryAxes = value;

                if (this.secondaryAxes != null)
                {
                    this.secondaryAxes.CollectionChanged += this.Axes_CollectionChanged;
                    this.secondaryAxes.AxesChanged += this.Axes_AxesChanged;
                }
            }
        }

        /// <summary>
        /// Represents the title of the chart.
        /// </summary>
        public Title Title
        {
            get
            {
                return this.title;
            }

            set
            {
                if (this.title != value)
                {
                    this.title = value;
                    this.OnChartChanged();
                }
            }
        }

        /// <summary>
        /// Gets or sets the chart legend of the chart.
        /// </summary>
        public Legend Legend
        {
            get
            {
                return this.legend;
            }
            set
            {
                if (this.legend != value)
                {
                    if (this.legend != null)
                    {
                        this.legend.LegendChanged -= this.Legend_LegendChanged;
                    }

                    this.legend = value;

                    if (this.legend != null)
                    {
                        this.legend.LegendChanged += this.Legend_LegendChanged;
                    }

                    this.OnChartChanged();
                }
            }
        }

        /// <summary>
        /// Creates a deep copy of the object.
        /// </summary>
        /// <returns>The deep copy of the object.</returns>
        public DocumentChart Clone()
        {
            DocumentChart newChart = new DocumentChart();

            foreach (SeriesGroup seriesGroup in this.SeriesGroups)
            {
                newChart.SeriesGroups.Add(seriesGroup.Clone());
            }

            newChart.PrimaryAxes = this.PrimaryAxes != null ? this.PrimaryAxes.Clone() : null;
            newChart.SecondaryAxes = this.SecondaryAxes != null ? this.SecondaryAxes.Clone() : null;
            newChart.Title = this.Title != null ? this.Title.Clone() : null;
            newChart.Legend = this.Legend != null ? this.Legend.Clone() : null;

            return newChart;
        }

        private void Axes_AxesChanged(object sender, EventArgs e)
        {
            this.OnChartChanged();
        }

        private void Axes_CollectionChanged(object sender, EventArgs e)
        {
            this.OnChartChanged();
        }

        private void SeriesGroupss_SeriesGroupChanged(object sender, EventArgs e)
        {
            this.OnChartChanged();
        }

        private void SeriesGroups_CollectionChanged(object sender, EventArgs e)
        {
            this.OnChartChanged();
        }

        private void Legend_LegendChanged(object sender, EventArgs e)
        {
            this.OnChartChanged();
        }

        internal event EventHandler ChartChanged;

        private void OnChartChanged()
        {
            if (this.ChartChanged != null)
            {
                this.ChartChanged(this, EventArgs.Empty);
            }
        }
    }
}
