﻿using System;
using System.Collections;
using System.Collections.Generic;

namespace Telerik.Windows.Documents.Model.Drawing.Charts
{
    /// <summary>
    /// Represents a base class for a collection of series.
    /// </summary>
    public abstract class SeriesCollection : IEnumerable<SeriesBase>
    {
        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.GetEnumerator();
        }

        /// <summary>
        /// Enumerates the series in the collection.
        /// </summary>
        /// <returns></returns>
        public IEnumerator<SeriesBase> GetEnumerator()
        {
            return this.GetEnumeratorOverride();
        }

        internal abstract IEnumerator<SeriesBase> GetEnumeratorOverride();

        /// <summary>
        /// Creates a new series, adds it to the collection and returns it.
        /// </summary>
        /// <returns>The new series object.</returns>
        public SeriesBase Add()
        {
            return this.AddOverride();
        }

        /// <summary>
        /// Creates a new series with data siutable for bubble series, adds it to the collection and returns it.
        /// </summary>
        /// <param name="xValuesData">The data for the X values of the series.</param>
        /// <param name="yValuesData">The data for the Y values of the series.</param>
        /// <param name="bubbleSizesData">The data for the bubble size values of the series.</param>
        /// <param name="title">The title of the series.</param>
        /// <returns>The new series object.</returns>
        public SeriesBase AddBubble(IChartData xValuesData, IChartData yValuesData, IChartData bubbleSizesData, Title title = null)
        {
            return this.AddOverride(xValuesData, yValuesData, bubbleSizesData, title);
        }

        /// <summary>
        /// Creates a new series with data siutable for scatter series, adds it to the collection and returns it.
        /// </summary>
        /// <param name="xValuesData">The data for the X values of the series.</param>
        /// <param name="yValuesData">The data for the Y values of the series.</param>
        /// <param name="title">The title of the series.</param>
        /// <returns>The new series object.</returns>
        public SeriesBase AddScatter(IChartData xValuesData, IChartData yValuesData, Title title = null)
        {
            return this.AddOverride(xValuesData, yValuesData, null, title);
        }

        /// <summary>
        /// Creates a new series using the specified data, adds it to the collection and returns it.
        /// </summary>
        /// <param name="categoriesData">The data for the categories of the series.</param>
        /// <param name="valuesData">The data for the values of the series.</param>
        /// <param name="title">The title of the series.</param>
        /// <returns>The new series object.</returns>
        public SeriesBase Add(IChartData categoriesData, IChartData valuesData, Title title = null)
        {
            return this.AddOverride(categoriesData, valuesData, null, title);
        }

        internal abstract SeriesBase AddOverride();

        internal abstract SeriesBase AddOverride(IChartData xValuesData, IChartData valuesData, IChartData bubbleSizesData, Title title = null);

        /// <summary>
        /// Adds a new series to the collection.
        /// </summary>
        /// <param name="series">The series to be added.</param>
        public void Add(SeriesBase series)
        {
            this.AddOverride(series);
        }

        internal abstract void AddOverride(SeriesBase series);

        /// <summary>
        /// Removes the specified series from the collection.
        /// </summary>
        /// <param name="series">The series to be removed.</param>
        public void Remove(SeriesBase series)
        {
            this.RemoveOverride(series);
        }

        internal abstract void RemoveOverride(SeriesBase series);

        internal abstract event EventHandler CollectionChanged;

        internal abstract event EventHandler SeriesChanged;
    }
}
