﻿using System;
using System.Collections.Generic;

namespace Telerik.Windows.Documents.Model.Drawing.Charts
{
    /// <summary>
    /// Represents a collection of series.
    /// </summary>
    /// <typeparam name="T">The type of series the collection holds.</typeparam>
    public class SeriesCollection<T> : SeriesCollection
        where T : SeriesBase

    {
        private readonly List<T> innerList;
        private readonly SeriesType seriesType;

        internal SeriesCollection(SeriesType seriesType)
        {
            this.innerList = new List<T>();
            this.seriesType = seriesType;
        }

        /// <summary>
        /// Enumerates the series in the collection as objects of a concrete series class.
        /// </summary>
        public new IEnumerator<T> GetEnumerator()
        {
            return this.GetEnumeratorOverride() as IEnumerator<T>;
        }

        internal override IEnumerator<SeriesBase> GetEnumeratorOverride()
        {
            return this.innerList.GetEnumerator();
        }

        /// <summary>
        /// Creates a new series, adds it to the collection and returns it.
        /// </summary>
        /// <returns>The new series object.</returns>
        public new T Add()
        {
            return this.AddOverride() as T;
        }

        internal override SeriesBase AddOverride()
        {
            SeriesBase newSeries = SeriesFactory.GetSeries(this.seriesType);
            T concreteNewSeries = newSeries as T;
            this.Add(concreteNewSeries);
            return concreteNewSeries;
        }

        /// <summary>
        /// Creates a new series with data siutable for bubble series, adds it to the collection and returns it.
        /// </summary>
        /// <param name="xValuesData">The data for the X values of the series.</param>
        /// <param name="yValuesData">The data for the Y values of the series.</param>
        /// <param name="bubbleSizesData">The data for the bubble size values of the series.</param>
        /// <param name="title">The title of the series.</param>
        /// <returns>The new series object.</returns>
        public new T AddBubble(IChartData xValuesData, IChartData yValuesData, IChartData bubbleSizesData, Title title = null)
        {
            return this.AddOverride(xValuesData, yValuesData, bubbleSizesData, title) as T;
        }

        /// <summary>
        /// Creates a new series with data siutable for scatter series, adds it to the collection and returns it.
        /// </summary>
        /// <param name="xValuesData">The data for the X values of the series.</param>
        /// <param name="yValuesData">The data for the Y values of the series.</param>
        /// <param name="title">The title of the series.</param>
        /// <returns>The new series object.</returns>
        public new T AddScatter(IChartData xValuesData, IChartData yValuesData, Title title = null)
        {
            return this.AddOverride(xValuesData, yValuesData, null, title) as T;
        }

        /// <summary>
        /// Creates a new series using the specified data, adds it to the collection and returns it.
        /// </summary>
        /// <param name="categoriesData">The data for the categories of the series.</param>
        /// <param name="valuesData">The data for the values of the series.</param>
        /// <param name="title">The title of the series.</param>
        /// <returns>The new series object.</returns>
        public new T Add(IChartData categoriesData, IChartData valuesData, Title title = null)
        {
            return this.AddOverride(categoriesData, valuesData, null, title) as T;
        }

        internal override SeriesBase AddOverride(IChartData xValuesData, IChartData valuesData, IChartData bubbleSizesData, Title title = null)
        {
            SeriesBase newSeries = SeriesFactory.GetSeries(this.seriesType, xValuesData, valuesData, bubbleSizesData, title);
            T concreteNewSeries = newSeries as T;
            this.Add(concreteNewSeries);
            return concreteNewSeries;
        }

        /// <summary>
        /// Adds a new series to the collection.
        /// </summary>
        /// <param name="series">The series to be added.</param>
        public void Add(T series)
        {
            this.AddOverride(series);
        }

        internal override void AddOverride(SeriesBase series)
        {
            T seriesToAdd = series as T;

            if (seriesToAdd == null)
            {
                throw new ArgumentException("The series is of type that is incompatible with the collection", "series");
            }

            seriesToAdd.SeriesChanged += this.Series_SeriesChanged;
            this.innerList.Add(seriesToAdd);
            this.OnCollectionChanged();
        }

        /// <summary>
        /// Removes the specified series from the collection.
        /// </summary>
        /// <param name="series">The series to be removed.</param>
        public void Remove(T series)
        {
            this.RemoveOverride(series);
        }

        internal override void RemoveOverride(SeriesBase series)
        {
            T seriesToRemove = series as T;

            if (seriesToRemove == null)
            {
                throw new InvalidOperationException();
            }

            if (this.innerList.Contains(seriesToRemove))
            {
                series.SeriesChanged -= this.Series_SeriesChanged;
                this.innerList.Remove(seriesToRemove);
                this.OnCollectionChanged();
            }
        }

        private void Series_SeriesChanged(object sender, EventArgs e)
        {
            this.OnSeriesChanged();
        }

        internal override event EventHandler CollectionChanged;

        private void OnCollectionChanged()
        {
            if (this.CollectionChanged != null)
            {
                this.CollectionChanged(this, EventArgs.Empty);
            }
        }

        internal override event EventHandler SeriesChanged;

        private void OnSeriesChanged()
        {
            if (this.SeriesChanged != null)
            {
                this.SeriesChanged(this, EventArgs.Empty);
            }
        }
    }
}