﻿using System;
#if WPF || NETSTANDARD
using System.Net;
#endif
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Core.Imaging;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Model.Data;

namespace Telerik.Windows.Documents.Model.Drawing.Shapes
{
    /// <summary>
    /// Represents image element.
    /// </summary>
    public sealed class Image : ShapeBase
    {
        private ImageSource imageSource;
        private bool preferRelativeToOriginalResize = true;

        /// <summary>
        /// Initializes a new instance of the <see cref="Image" /> class.
        /// </summary>
        internal Image()
        {
            this.LockAspectRatio = true;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Image" /> class.
        /// </summary>
        /// <param name="other">The other image.</param>
        internal Image(Image other)
            : this(other, true)
        {
        }

        private Image(Image other, bool copyImageSource)
            : base(other)
        {
            if (copyImageSource)
            {
                this.imageSource = other.ImageSource;
            }

            this.PreferRelativeToOriginalResize = other.PreferRelativeToOriginalResize;
        }

        /// <summary>
        /// Gets or sets the image source.
        /// </summary>
        /// <value>The image source.</value>
        public ImageSource ImageSource
        {
            get
            {
                return this.imageSource;
            }
            set
            {
                if (this.imageSource != value)
                {
                    ImageSource oldValue = this.imageSource;
                    this.imageSource = value;
                    this.OnImageSourceChanged(oldValue, this.imageSource);
                }
            }
        }

        /// <summary>
        /// Gets or sets the value indicating whether the scale resizing should be relative to the original or the current size of the image.
        /// </summary>
        /// <value>The value indicating whether the scale resizing should be relative to the original or the current size of the image.</value>
        public bool PreferRelativeToOriginalResize
        {
            get
            {
                return this.preferRelativeToOriginalResize;
            }
            set
            {
                if (this.preferRelativeToOriginalResize != value)
                {
                    this.preferRelativeToOriginalResize = value;
                }
            }
        }

#if WPF || NETSTANDARD
        /// <summary>
        /// Initializes the size.
        /// </summary>
        protected override void InitializeSize()
        {
            if (this.ImageSource != null)
            {
                this.Size = ImageSizeDecodersManager.GetSize(this.ImageSource.Data, this.ImageSource.Extension);
            }
        }
#endif

        internal event EventHandler<ResourceChangedEventArgs> ImageSourceChanged;

        internal Image Clone(bool cloneImageSource)
        {
            Image clonedImage = new Image(this, !cloneImageSource);

            if (cloneImageSource)
            {
                if (this.ImageSource != null)
                {
                    clonedImage.imageSource = this.ImageSource.Clone();
                }
            }

            return clonedImage;
        }

        private void OnImageSourceChanged(ImageSource oldValue, ImageSource newValue)
        {
            if (this.ImageSourceChanged != null)
            {
                this.ImageSourceChanged(this, new ResourceChangedEventArgs { OldValue = oldValue, NewValue = newValue });
            }
        }

        internal void RecalculateSize()
        {
            if (ImageSource == null)
            {
                return;
            }

            Size imageSize = default(Size);
            try
            {
                imageSize = ImageSizeDecodersManager.GetSize(this.ImageSource.Data, this.ImageSource.Extension);
            }
#if WPF || NETSTANDARD
            catch (WebException)
            {
                return;
            }
#endif
            catch (InvalidOperationException)
            {
                return;
            }

            if (imageSize == Size.Empty)
            {
                return;
            }           

            if (!this.HasHeight ^ !this.HasWidth)
            {
                var sourceRatio = imageSize.Height / imageSize.Width;
                double height = this.Height, width = this.Width;

                if (this.HasHeight)
                {
                    if (this.LockAspectRatio)
                    {
                        width = this.Height / sourceRatio;
                    }
                    else
                    {
                        width = imageSize.Width;
                    }                    
                }
                if (this.HasWidth)
                {
                    if (this.LockAspectRatio)
                    {
                        height = this.Width * sourceRatio;
                    }
                    else
                    {
                        height = imageSize.Height;
                    }
                }
                this.Size = new Size(width, height);
            }
            if (!this.HasHeight && !this.HasWidth)
            {
                this.Size = new Size(imageSize.Width, imageSize.Height);
            }
        }
    }
}
