﻿#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif

#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Model.Drawing.Theming;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Model.Drawing.Shapes
{
    /// <summary>
    /// Represents shape base element.
    /// </summary>
    public abstract class ShapeBase
    {
        internal static readonly Size DefaultSize = new Size(1, 1);

        private string name;
        private Size size = Size.Empty;
        private bool hasSize;
        private bool isVerticallyFlipped;
        private bool isHorizontallyFlipped;
        private double rotationAngle;
        private bool lockAspectRatio;
        private readonly Outline outline;
        private Fill fill;

        /// <summary>
        /// Initializes a new instance of the <see cref="ShapeBase" /> class.
        /// </summary>
        internal ShapeBase()
        {
            this.outline = new Outline();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ShapeBase" /> class by copying an Image instance.
        /// </summary>
        /// <param name="other">The other image.</param>
        protected ShapeBase(ShapeBase other)
            : this()
        {
            this.name = other.name;
            this.size = other.size;
            this.hasSize = other.hasSize;
            this.HasWidth = other.HasWidth;
            this.HasHeight = other.HasHeight;
            this.isVerticallyFlipped = other.isVerticallyFlipped;
            this.isHorizontallyFlipped = other.isHorizontallyFlipped;
            this.rotationAngle = other.rotationAngle;
            this.lockAspectRatio = other.LockAspectRatio;
            this.outline = other.outline.Clone();
            if (other.fill != null)
            {
                this.fill = other.fill.Clone();
            }
        }

        internal bool HasWidth { get; private set; }

        internal bool HasHeight { get; private set; }

        /// <summary>
        /// Gets or sets the name.
        /// </summary>
        /// <value>The name.</value>
        public string Name
        {
            get
            {
                return this.name;
            }
            set
            {
                if (this.name != value)
                {
                    this.name = value;
                }
            }
        }

        /// <summary>
        /// Gets or sets the width.
        /// </summary>
        /// <value>The width.</value>
        public double Width
        {
            get
            {
                return this.Size.Width;
            }
            set
            {
                Guard.ThrowExceptionIfNaN(value, "value");
                Guard.ThrowExceptionIfPositiveInfinity(value, "value");
                Guard.ThrowExceptionIfLessThan(0, value, "value");

                if (this.size.Width != value)
                {
                    this.EnsureSizeIsNotEmpy();
                    this.size.Width = value;
                    this.hasSize = true;
                    this.HasWidth = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets the height.
        /// </summary>
        /// <value>The height.</value>
        public double Height
        {
            get
            {
                return this.Size.Height;
            }
            set
            {
                Guard.ThrowExceptionIfNaN(value, "value");
                Guard.ThrowExceptionIfPositiveInfinity(value, "value");
                Guard.ThrowExceptionIfLessThan(0, value, "value");

                if (this.size.Height != value)
                {
                    this.EnsureSizeIsNotEmpy();
                    this.size.Height = value;
                    this.hasSize = true;
                    this.HasHeight = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets the size.
        /// </summary>
        /// <value>The size.</value>
        public Size Size
        {
            get
            {
                if (!this.hasSize)
                {
                    this.InitializeSize();
                }

                return this.size;
            }
            set
            {
                Guard.ThrowExceptionIfNaN(value.Width, "value.Width");
                Guard.ThrowExceptionIfNaN(value.Height, "value.Height");
                Guard.ThrowExceptionIfPositiveInfinity(value.Width, "value.Width");
                Guard.ThrowExceptionIfPositiveInfinity(value.Height, "value.Height");

                this.size = value;
                this.hasSize = true;
            }
        }

        /// <summary>
        /// Gets the size silently so the size's auto-initializing cannot be invoked.
        /// </summary>
        /// <value>The size.</value>
        internal Size SizeInternal
        {
            get
            {
                return this.size;
            }
        }

        /// <summary>
        /// Initializes the size.
        /// </summary>
        protected virtual void InitializeSize()
        {
            this.Size = ShapeBase.DefaultSize;
        }

        /// <summary>
        /// Gets or sets the value indicating if the shape is vertically flipped.
        /// </summary>
        /// <value>The value indicating if the shape is vertically flipped.</value>
        public bool IsVerticallyFlipped
        {
            get
            {
                return this.isVerticallyFlipped;
            }
            set
            {
                if (this.isVerticallyFlipped != value)
                {
                    this.isVerticallyFlipped = value;
                }
            }
        }

        /// <summary>
        /// Gets or sets the value indicating if the shape is horizontally flipped.
        /// </summary>
        /// <value>The value indicating if the shape is horizontally flipped.</value>
        public bool IsHorizontallyFlipped
        {
            get
            {
                return this.isHorizontallyFlipped;
            }
            set
            {
                if (this.isHorizontallyFlipped != value)
                {
                    this.isHorizontallyFlipped = value;
                }
            }
        }

        /// <summary>
        /// Gets or sets the rotation angle.
        /// </summary>
        /// <value>The rotation angle.</value>
        public double RotationAngle
        {
            get
            {
                return this.rotationAngle;
            }
            set
            {
                if (this.rotationAngle != value)
                {
                    this.rotationAngle = value;
                }
            }
        }

        /// <summary>
        /// Gets or sets the value indicating whether the aspect ratio between the width and height should remain constant.
        /// </summary>
        /// <value>The value indicating whether the aspect ratio between the width and height should remain constant.</value>
        public bool LockAspectRatio
        {
            get
            {
                return this.lockAspectRatio;
            }
            set
            {
                if (this.lockAspectRatio != value)
                {
                    this.lockAspectRatio = value;
                }
            }
        }

        /// <summary>
        /// Gets or sets the fill of the shape.
        /// </summary>
        public Fill Fill
        {
            get
            {
                return this.fill;
            }
            set
            {
                if (this.fill != value)
                {
                    this.fill = value;
                }
            }
        }

        /// <summary>
        /// Gets the outline of the shape.
        /// </summary>
        public Outline Outline
        {
            get
            {
                return this.outline;
            }
        }

        internal int Id { get; set; }

        /// <summary>
        /// Sets the width of the shape.
        /// </summary>
        /// <param name="respectLockAspectRatio">A value indicating whether the aspect ratio lock should be respected.</param>
        /// <param name="width">The new width.</param>
        public void SetWidth(bool respectLockAspectRatio, double width)
        {
            if (respectLockAspectRatio && this.hasSize)
            {
                this.ChangeHeightAccordingToAspectRatio(width);
            }

            this.Width = width;
        }

        /// <summary>
        /// Sets the height of the shape.
        /// </summary>
        /// <param name="respectLockAspectRatio">A value indicating whether the aspect ratio lock should be respected.</param>
        /// <param name="height">The new height.</param>
        public void SetHeight(bool respectLockAspectRatio, double height)
        {
            if (respectLockAspectRatio && this.hasSize)
            {
                this.ChangeWidthAccordingToAspectRatio(height);
            }

            this.Height = height;
        }

        private void ChangeHeightAccordingToAspectRatio(double newWidth)
        {
            if (this.LockAspectRatio && this.Width != 0)
            {
                double aspectRatio = this.Height / this.Width;
                this.size.Height = newWidth * aspectRatio;
            }
        }

        private void ChangeWidthAccordingToAspectRatio(double newHeight)
        {
            if (this.LockAspectRatio && this.Height != 0)
            {
                double aspectRatio = this.Width / this.Height;
                this.size.Width = newHeight * aspectRatio;
            }
        }

        private void EnsureSizeIsNotEmpy()
        {
            if (this.size.IsEmpty)
            {
                this.Size = ShapeBase.DefaultSize;
            }
        }
    }
}