﻿using System.ComponentModel;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Theming;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
#endif
#if NETSTANDARD
using Telerik.Documents.Media;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Utilities;

#if (WPF || SILVERLIGHT) && !NETCORE
namespace Telerik.Windows.Documents.Spreadsheet.Model
#else
namespace Telerik.Documents.Common.Model
#endif
{
    /// <summary>
    /// Represents a color which can be used in a theme.
    /// </summary>
    [TypeConverter(typeof(ThemableColorTypeConverter))]
    public class ThemableColor : IThemableObject<Color>
    {
        private readonly Color? localValue;
        private readonly ThemeColorType? themeColorType;
        private readonly ColorShadeType? colorShadeType;
        private readonly double? tintAndShade;
        private readonly bool isAutomatic;

        /// <summary>
        /// Initializes a new instance of the <see cref="ThemableColor" /> class.
        /// </summary>
        /// <param name="color">The color.</param>
        public ThemableColor(Color color)
            : this(color, false)
        {
            this.localValue = color;
        }


        /// <summary>
        /// Initializes a new instance of the <see cref="ThemableColor" /> class.
        /// </summary>
        /// <param name="color">The color.</param>
        /// <param name="isAutomatic">The is automatic.</param>
        public ThemableColor(Color color, bool isAutomatic)
        {
            this.isAutomatic = isAutomatic;
            this.localValue = color;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ThemableColor" /> class.
        /// </summary>
        /// <param name="themeColorType">The theme color type.</param>
        /// <param name="colorShadeType">The color shade type.</param>
        public ThemableColor(ThemeColorType themeColorType, ColorShadeType? colorShadeType = null)
        {
            this.themeColorType = themeColorType;
            this.colorShadeType = colorShadeType;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ThemableColor" /> class.
        /// </summary>
        /// <param name="themeColorType">Theme color type.</param>
        /// <param name="tintAndShade">The tint and shade.</param>
        public ThemableColor(ThemeColorType themeColorType, double tintAndShade)
            : this(themeColorType)
        {
            Guard.ThrowExceptionIfOutOfRange(-1, 1, tintAndShade, "tintAndShade");
            this.tintAndShade = tintAndShade;
        }

        /// <summary>
        /// Gets the local value of the color.
        /// </summary>
        public Color LocalValue
        {
            get
            {
                return this.localValue.Value;
            }
        }

        /// <summary>
        /// Gets the value indicating if the color is automatic. Automatic colors may be modified by a consumer as appropriate.
        /// </summary>
        /// <value>Value indicating if the color is automatic.</value>
        public bool IsAutomatic
        {
            get
            {
                return this.isAutomatic;
            }
        }

        /// <summary>
        /// Gets the theme color type.
        /// </summary>
        /// <value>The theme color type.</value>
        public ThemeColorType ThemeColorType
        {
            get
            {
                return this.themeColorType.Value;
            }
        }

        /// <summary>
        /// Gets the color shade type.
        /// </summary>
        /// <value>The color shade type.</value>
        public ColorShadeType? ColorShadeType
        {
            get
            {
                return this.colorShadeType;
            }
        }

        /// <summary>
        /// Gets the tint and shade.
        /// </summary>
        /// <value>The tint and shade.</value>
        public double? TintAndShade
        {
            get
            {
                return this.tintAndShade;
            }
        }

        /// <summary>
        /// Gets the value indicating if the instance is from a theme.
        /// </summary>
        /// <value>The value indicating if the instance is from a theme.</value>
        public bool IsFromTheme
        {
            get
            {
                return this.themeColorType.HasValue;
            }
        }

        /// <summary>
        /// Creates new themable color from Argb.
        /// </summary>
        /// <param name="alfa">The alfa.</param>
        /// <param name="red">The red.</param>
        /// <param name="green">The green.</param>
        /// <param name="blue">The blue.</param>
        /// <returns></returns>
        public static ThemableColor FromArgb(byte alfa, byte red, byte green, byte blue)
        {
            return new ThemableColor(Color.FromArgb(alfa, red, green, blue));
        }

        /// <summary>
        /// Compares two themable colors.
        /// </summary>
        /// <param name="first">The first themable color.</param>
        /// <param name="second">The second themable color.</param>
        /// <returns>If the two themable colors are equal.</returns>
        public static bool operator ==(ThemableColor first, ThemableColor second)
        {
            if ((object)first == null)
            {
                return (object)second == null;
            }

            return first.Equals(second);
        }

        /// <summary>
        /// Compares two themable colors.
        /// </summary>
        /// <param name="first">The first themable color.</param>
        /// <param name="second">The second themable color.</param>
        /// <returns>If the two themable colors are equal.</returns>
        public static bool operator !=(ThemableColor first, ThemableColor second)
        {
            return !(first == second);
        }

        /// <summary>
        /// Explicitly cast color to themable color.
        /// </summary>
        /// <param name="value">The color.</param>
        /// <returns>Themable color.</returns>
        public static explicit operator ThemableColor(Color value)
        {
            return new ThemableColor(value);
        }

        /// <summary>
        /// Gets the actual value.
        /// </summary>
        /// <param name="theme">The theme.</param>
        /// <returns>The actual value.</returns>
        public Color GetActualValue(DocumentTheme theme)
        {
            Guard.ThrowExceptionIfNull(theme, "theme");

            return this.GetActualValue(theme.ColorScheme);
        }

        /// <summary>
        /// Gets the actual value.
        /// </summary>
        /// <param name="colorScheme">The color scheme.</param>
        /// <returns>The actual color.</returns>
        public Color GetActualValue(ThemeColorScheme colorScheme)
        {
            Guard.ThrowExceptionIfNull(colorScheme, "colorScheme");

            double actualTintAndShade;
            Color color;
            if (this.IsFromTheme)
            {
                color = colorScheme[this.ThemeColorType].Color;

                if (this.tintAndShade.HasValue)
                {
                    actualTintAndShade = this.tintAndShade.Value;
                }
                else if (this.colorShadeType.HasValue)
                {
                    actualTintAndShade = colorScheme.GetTintAndShade(this.themeColorType.Value, this.colorShadeType.Value);
                }
                else
                {
                    actualTintAndShade = 0;
                }
            }
            else
            {
                color = this.localValue.Value;
                actualTintAndShade = 0;
            }

            return ColorsHelper.UpdateTint(color, actualTintAndShade);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            ThemableColor other = obj as ThemableColor;

            return other != null &&
                this.localValue == other.localValue &&
                this.isAutomatic == other.isAutomatic &&
                this.themeColorType == other.themeColorType &&
                this.tintAndShade == other.tintAndShade &&
                this.colorShadeType == other.colorShadeType;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return ObjectExtensions.CombineHashCodes(
                this.localValue.GetHashCode(),
                this.themeColorType.GetHashCode(),
                this.tintAndShade.GetHashCode(),
                this.colorShadeType.GetHashCode());
        }



        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
            if (this.IsFromTheme)
            {
                return this.ThemeColorType.ToString();
            }
            else
            {
                if (this.isAutomatic)
                {
                    return string.Format("Auto({0})", this.localValue.ToString());
                }
                else
                {
                    return this.localValue.ToString();
                }
            }
        }
    }
}
