﻿#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Theming;
#else
#endif
using Telerik.Windows.Documents.Utilities;

#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
#else
using System.Windows.Media;
#endif

#if (WPF || SILVERLIGHT) && !NETCORE
namespace Telerik.Windows.Documents.Spreadsheet.Model
#else
namespace Telerik.Documents.Common.Model
#endif
{
    /// <summary>
    /// Represents a font family which can be used in a theme.
    /// </summary>
    public class ThemableFontFamily : IThemableObject<FontFamily>
    {
        private readonly FontFamily localValue;
        private readonly ThemeFontType? themeFontType;

        /// <summary>
        /// Initializes a new instance of the <see cref="ThemableFontFamily" /> class.
        /// </summary>
        /// <param name="fontFamily">The font family.</param>
        public ThemableFontFamily(FontFamily fontFamily)
        {
            if (string.IsNullOrEmpty(fontFamily.Source))
            {
                this.themeFontType = ThemeFontType.Minor;
            }
            else
            {
                this.localValue = fontFamily;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ThemableFontFamily" /> class.
        /// </summary>
        /// <param name="familyName">Name of the family.</param>
        public ThemableFontFamily(string familyName)
            : this(new FontFamily(familyName))
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ThemableFontFamily" /> class.
        /// </summary>
        /// <param name="themeFontType">Type of the theme font.</param>
        public ThemableFontFamily(ThemeFontType themeFontType)
        {
            this.themeFontType = themeFontType;
        }

        /// <summary>
        /// Gets the local value.
        /// </summary>
        /// <value>The local value.</value>
        public FontFamily LocalValue
        {
            get
            {
                return this.localValue;
            }
        }

        /// <summary>
        /// Gets the theme font type.
        /// </summary>
        /// <value>The theme font type.</value>
        public ThemeFontType ThemeFontType
        {
            get
            {
                return this.themeFontType.Value;
            }
        }

        /// <summary>
        /// Gets the value indicating if the instance is from a theme.
        /// </summary>
        /// <value>The value indicating if the instance is from a theme.</value>
        public bool IsFromTheme
        {
            get
            {
                return this.localValue == null;
            }
        }

        /// <summary>
        /// Compares two themable font families.
        /// </summary>
        /// <param name="first">The first themable font family.</param>
        /// <param name="second">The second themable font family.</param>
        /// <returns>If the two themable font families are equal.</returns>
        public static bool operator ==(ThemableFontFamily first, ThemableFontFamily second)
        {
            if ((object)first == null)
            {
                return (object)second == null;
            }

            return first.Equals(second);
        }

        /// <summary>
        /// Compares two themable font families.
        /// </summary>
        /// <param name="first">The first themable font family.</param>
        /// <param name="second">The second themable font family.</param>
        /// <returns>If the two themable font families are not equal.</returns>
        public static bool operator !=(ThemableFontFamily first, ThemableFontFamily second)
        {
            return !(first == second);
        }

        /// <summary>
        /// Casts explicitly FontFamily object to themable font family.
        /// </summary>
        /// <param name="value">The font family.</param>
        /// <returns>Themable font family.</returns>
        public static explicit operator ThemableFontFamily(FontFamily value)
        {
            return new ThemableFontFamily(value);
        }

        /// <summary>
        /// Gets the actual value.
        /// </summary>
        /// <param name="theme">The theme.</param>
        /// <returns>The actual value.</returns>
        public FontFamily GetActualValue(DocumentTheme theme)
        {
            Guard.ThrowExceptionIfNull(theme, "theme");

            return this.GetActualValue(theme.FontScheme);
        }

        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current
        /// <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
            if (this.IsFromTheme)
            {
                return this.ThemeFontType.ToString();
            }
            else
            {
                return this.localValue.ToString();
            }
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            ThemableFontFamily other = obj as ThemableFontFamily;

            if (other == null)
            {
                return false;
            }

            return ObjectExtensions.EqualsOfT(this.localValue, other.localValue) &&
                this.themeFontType == other.themeFontType;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return ObjectExtensions.CombineHashCodes(
                this.localValue.GetHashCodeOrZero(),
                this.themeFontType.GetHashCodeOrZero());
        }

        private FontFamily GetActualValue(ThemeFontScheme themeFontScheme, FontLanguageType fontLanguageType = FontLanguageType.Latin)
        {
            Guard.ThrowExceptionIfNull(themeFontScheme, "themeFontScheme");

            FontFamily fontFamily = this.IsFromTheme ? themeFontScheme[this.ThemeFontType][fontLanguageType].FontFamily : this.localValue;

            return fontFamily;
        }
    }
}
