﻿using System.Collections;
using System.Collections.Generic;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Spreadsheet.Core;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Model;
#endif
#if NETSTANDARD
using Telerik.Documents.Media;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Utilities;

#if (WPF || SILVERLIGHT) && !NETCORE
namespace Telerik.Windows.Documents.Spreadsheet.Theming
#else
namespace Telerik.Documents.Common.Model
#endif
{
    /// <summary>
    /// Represents the color scheme of a theme.
    /// </summary>
    [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1710:IdentifiersShouldHaveCorrectSuffix")]
    public class ThemeColorScheme : NamedObjectBase, IEnumerable<ThemeColor>
    {
        private readonly Dictionary<ThemeColorType, ThemeColor> themeColorTypeToColor;

        private int? hashCode;

        /// <summary>
        /// Initializes a new instance of the <see cref="ThemeColorScheme" /> class.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="background1">The first background.</param>
        /// <param name="text1">The first text color.</param>
        /// <param name="background2">The second background.</param>
        /// <param name="text2">The second text color.</param>
        /// <param name="accent1">The first accent.</param>
        /// <param name="accent2">The second accent.</param>
        /// <param name="accent3">The third accent.</param>
        /// <param name="accent4">The fourth accent.</param>
        /// <param name="accent5">The fifth accent.</param>
        /// <param name="accent6">The sixth accent.</param>
        /// <param name="hyperlink">The hyperlink color.</param>
        /// <param name="followedHyperlink">The followed hyperlink color.</param>
        public ThemeColorScheme(string name,
            Color background1,
            Color text1,
            Color background2,
            Color text2,
            Color accent1,
            Color accent2,
            Color accent3,
            Color accent4,
            Color accent5,
            Color accent6,
            Color hyperlink,
            Color followedHyperlink)
            : base(name)
        {
            this.themeColorTypeToColor = new Dictionary<ThemeColorType, ThemeColor>();

            this.AddThemeColor(ThemeColorType.Background1, background1);
            this.AddThemeColor(ThemeColorType.Text1, text1);
            this.AddThemeColor(ThemeColorType.Background2, background2);
            this.AddThemeColor(ThemeColorType.Text2, text2);
            this.AddThemeColor(ThemeColorType.Accent1, accent1);
            this.AddThemeColor(ThemeColorType.Accent2, accent2);
            this.AddThemeColor(ThemeColorType.Accent3, accent3);
            this.AddThemeColor(ThemeColorType.Accent4, accent4);
            this.AddThemeColor(ThemeColorType.Accent5, accent5);
            this.AddThemeColor(ThemeColorType.Accent6, accent6);
            this.AddThemeColor(ThemeColorType.Hyperlink, hyperlink);
            this.AddThemeColor(ThemeColorType.FollowedHyperlink, followedHyperlink);
        }

        internal ThemeColorScheme(string name, Dictionary<ThemeColorType, ThemeColor> themeColorTypeToColor)
            : base(name)
        {
            Guard.ThrowExceptionIfNull(themeColorTypeToColor, "themeColorTypeToColor");

            this.themeColorTypeToColor = themeColorTypeToColor;
        }

        /// <summary>
        /// Gets the <see cref="ThemeColor" /> with the specified color type.
        /// </summary>
        /// <value>The theme color.</value>
        public ThemeColor this[ThemeColorType colorType]
        {
            get
            {
                return this.themeColorTypeToColor[colorType];
            }
        }

        /// <summary>
        /// Gets the tint and shade.
        /// </summary>
        /// <param name="themeColorType">Type of the theme color.</param>
        /// <param name="colorShadeType">Type of the color shade.</param>
        /// <returns>The tint and shade value.</returns>
        public double GetTintAndShade(ThemeColorType themeColorType, ColorShadeType colorShadeType)
        {
            Color mainColor = this[themeColorType].Color;

            return ColorsHelper.GetTintAndShadeForColorAndIndex(mainColor, colorShadeType);
        }

        /// <summary>
        /// Creates deep copy of this theme color scheme.
        /// </summary>
        /// <returns>The cloned theme color scheme.</returns>
        public ThemeColorScheme Clone()
        {
            Dictionary<ThemeColorType, ThemeColor> colors = new Dictionary<ThemeColorType, ThemeColor>();
            foreach (KeyValuePair<ThemeColorType, ThemeColor> pair in this.themeColorTypeToColor)
            {
                colors.Add(pair.Key, pair.Value.Clone());
            }

            return new ThemeColorScheme(this.Name, colors);
        }

        /// <inheritdoc />
        public override bool Equals(object obj)
        {
            ThemeColorScheme other = obj as ThemeColorScheme;
            if (other == null || this.Name != other.Name)
            {
                return false;
            }

            foreach (KeyValuePair<ThemeColorType, ThemeColor> pair in this.themeColorTypeToColor)
            {
                if (!ObjectExtensions.EqualsOfT(pair.Value, other[pair.Key]))
                {
                    return false;
                }
            }

            return true;
        }

        /// <inheritdoc />
        public override int GetHashCode()
        {
            if (!this.hashCode.HasValue)
            {
                this.hashCode = this.Name.GetHashCode();

                foreach (KeyValuePair<ThemeColorType, ThemeColor> pair in this.themeColorTypeToColor)
                {
                    this.hashCode = ObjectExtensions.CombineHashCodes(this.hashCode.Value, pair.Value.GetHashCode());
                }
            }

            return this.hashCode.Value;
        }

        /// <inheritdoc />
        public IEnumerator<ThemeColor> GetEnumerator()
        {
            return this.themeColorTypeToColor.Values.GetEnumerator();
        }

        /// <inheritdoc />
        IEnumerator IEnumerable.GetEnumerator()
        {
            return ((IEnumerable<ThemeColor>)this).GetEnumerator();
        }

        private void AddThemeColor(ThemeColorType themeColorType, Color color)
        {
            this.themeColorTypeToColor.Add(themeColorType, new ThemeColor(color, themeColorType));
        }
    }
}