﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Utilities;

#if (WPF || SILVERLIGHT) && !NETCORE
namespace Telerik.Windows.Documents.Spreadsheet.Theming
#else
namespace Telerik.Documents.Common.Model
#endif
{
    /// <summary>
    /// Represents the font scheme of a theme.
    /// </summary>
    public class ThemeFontScheme : NamedObjectBase
    {
        private readonly Dictionary<ThemeFontType, ThemeFonts> themeFontTypeToFont;

        private int? hashCode;

        /// <summary>
        /// Initializes a new instance of the <see cref="ThemeFontScheme" /> class.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="latinMajorFontName">Name of the latin headings font.</param>
        /// <param name="latinMinorFontName">Name of the latin body font.</param>
        /// <param name="eastAsianMajorFontName">Name of the east asian headings font.</param>
        /// <param name="eastAsianMinorFontName">Name of the east asian body font.</param>
        /// <param name="complexScriptMajorFontName">Name of the complex script headings font.</param>
        /// <param name="complexScriptMinorFontName">Name of the complex script body font.</param>
        public ThemeFontScheme(string name,
            string latinMajorFontName = null,
            string latinMinorFontName = null,
            string eastAsianMajorFontName = null,
            string eastAsianMinorFontName = null,
            string complexScriptMajorFontName = null,
            string complexScriptMinorFontName = null)
            : base(name)
        {
            this.themeFontTypeToFont = new Dictionary<ThemeFontType, ThemeFonts>();
            this.themeFontTypeToFont.Add(ThemeFontType.Major, new ThemeFonts(latinMajorFontName, eastAsianMajorFontName, complexScriptMajorFontName));
            this.themeFontTypeToFont.Add(ThemeFontType.Minor, new ThemeFonts(latinMinorFontName, eastAsianMinorFontName, complexScriptMinorFontName));
        }

        internal ThemeFontScheme(string name, Dictionary<ThemeFontType, ThemeFonts> themeFontTypeToFont)
            : base(name)
        {
            Guard.ThrowExceptionIfNull(themeFontTypeToFont, "themeFontTypeToFont");

            this.themeFontTypeToFont = themeFontTypeToFont;
        }

        /// <summary>
        /// Gets the <see cref="ThemeFonts" /> with the specified font type.
        /// </summary>
        /// <value>The theme fonts.</value>
        public ThemeFonts this[ThemeFontType fontType]
        {
            get
            {
                return this.themeFontTypeToFont[fontType];
            }
        }

        /// <summary>
        /// Creates deep copy of this theme font scheme.
        /// </summary>
        /// <returns>The cloned theme font scheme.</returns>
        public ThemeFontScheme Clone()
        {
            Dictionary<ThemeFontType, ThemeFonts> themeFontShemes = new Dictionary<ThemeFontType, ThemeFonts>();

            foreach (KeyValuePair<ThemeFontType, ThemeFonts> pair in this.themeFontTypeToFont)
            {
                themeFontShemes.Add(pair.Key, pair.Value.Clone());
            }

            return new ThemeFontScheme(this.Name, themeFontShemes);
        }

        /// <inheritdoc />
        public override bool Equals(object obj)
        {
            ThemeFontScheme other = obj as ThemeFontScheme;
            if (other == null || this.Name != other.Name)
            {
                return false;
            }

            foreach (KeyValuePair<ThemeFontType, ThemeFonts> pair in this.themeFontTypeToFont)
            {
                if (!ObjectExtensions.EqualsOfT(pair.Value, other[pair.Key]))
                {
                    return false;
                }
            }

            return true;
        }

        /// <inheritdoc />
        public override int GetHashCode()
        {
            if (!this.hashCode.HasValue)
            {
                this.hashCode = this.Name.GetHashCode();

                foreach (KeyValuePair<ThemeFontType, ThemeFonts> pair in this.themeFontTypeToFont)
                {
                    this.hashCode = ObjectExtensions.CombineHashCodes(this.hashCode.Value, pair.Value.GetHashCode());
                }
            }

            return this.hashCode.Value;
        }
    }
}