﻿using System.Collections.Generic;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Model;
#endif
using Telerik.Windows.Documents.Utilities;

#if (WPF || SILVERLIGHT) && !NETCORE
namespace Telerik.Windows.Documents.Spreadsheet.Theming
#else
namespace Telerik.Documents.Common.Model
#endif
{
    /// <summary>
    /// A collection of fonts for a theme, each corresponding to a language type.
    /// </summary>
    public class ThemeFonts
    {
        private readonly Dictionary<FontLanguageType, ThemeFont> fontLanguageTypeToFont;

        /// <summary>
        /// Initializes a new instance of the <see cref="ThemeFonts" /> class.
        /// </summary>
        /// <param name="latinFontName">Name of the latin font.</param>
        /// <param name="eastAsianFontName">Name of the east asian font.</param>
        /// <param name="complexScriptFontName">Name of the complex script font.</param>
        public ThemeFonts(string latinFontName = null, string eastAsianFontName = null, string complexScriptFontName = null)
        {
            this.fontLanguageTypeToFont = new Dictionary<FontLanguageType, ThemeFont>();

            this.AddThemeFont(latinFontName, FontLanguageType.Latin);
            this.AddThemeFont(eastAsianFontName, FontLanguageType.EastAsian);
            this.AddThemeFont(complexScriptFontName, FontLanguageType.ComplexScript);
        }

        internal ThemeFonts(Dictionary<FontLanguageType, ThemeFont> fontLanguageTypeToFont)
        {
            Guard.ThrowExceptionIfNull(fontLanguageTypeToFont, "fontLanguageTypeToFont");

            this.fontLanguageTypeToFont = fontLanguageTypeToFont;
        }

        /// <summary>
        /// Gets the <see cref="ThemeFont" /> with the specified font language type.
        /// </summary>
        /// <value>The theme font.</value>
        public ThemeFont this[FontLanguageType fontLanguageType]
        {
            get
            {
                return this.fontLanguageTypeToFont[fontLanguageType];
            }
        }

        /// <summary>
        /// Creates deep copy of this theme fonts.
        /// </summary>
        /// <returns>The cloned theme fonts.</returns>
        public ThemeFonts Clone()
        {
            Dictionary<FontLanguageType, ThemeFont> themeFonts = new Dictionary<FontLanguageType, ThemeFont>();

            foreach (KeyValuePair<FontLanguageType, ThemeFont> pair in this.fontLanguageTypeToFont)
            {
                if (pair.Value != null)
                {
                    themeFonts.Add(pair.Key, new ThemeFont(pair.Value.FontFamily, pair.Value.FontLanguageType));
                }
                else
                {
                    themeFonts.Add(pair.Key, null);
                }
            }

            return new ThemeFonts(themeFonts);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            ThemeFonts other = obj as ThemeFonts;
            foreach (KeyValuePair<FontLanguageType, ThemeFont> pair in this.fontLanguageTypeToFont)
            {
                if (!ObjectExtensions.EqualsOfT(pair.Value, other[pair.Key]))
                {
                    return false;
                }
            }

            return true;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            int hashCode = 0;
            foreach (KeyValuePair<FontLanguageType, ThemeFont> pair in this.fontLanguageTypeToFont)
            {
                hashCode = ObjectExtensions.CombineHashCodes(hashCode, pair.Value.GetHashCode());
            }

            return hashCode;
        }

        private void AddThemeFont(string fontName, FontLanguageType fontLanguageType)
        {
            ThemeFont themeFont = null;
            if (!string.IsNullOrEmpty(fontName))
            {
                themeFont = new ThemeFont(fontName, fontLanguageType);
            }

            this.fontLanguageTypeToFont.Add(fontLanguageType, themeFont);
        }
    }
}