﻿using System;
using System.Collections.Generic;

namespace Telerik.Windows.Documents.Common.Model.Protection
{
    /// <summary>
    /// Represents a hash algorithms provider.
    /// </summary>
    public static class HashingAlgorithmsProvider
    {
        private static readonly Dictionary<string, IProtectionAlgorithm> algorithms;
        private static bool isFipsEnforced;

        static HashingAlgorithmsProvider()
        {
            algorithms = new Dictionary<string, IProtectionAlgorithm>();
        }

        /// <summary>
        /// Enforces Federal Information Processing Standard (FIPS) Publication 140-2.
        /// </summary>
        /// <remarks>
        /// Calling this method will cause some hashing algorithms to be excluded from the supported set of algorithms.
        /// This method must be call before using any hashing algorithms.
        /// </remarks>
        public static void EnforceFips1402()
        {
            isFipsEnforced = true;
        }

        internal static IProtectionAlgorithm GetProtectionAlgorithm(string algorithmName)
        {
            Initialize();

            IProtectionAlgorithm algorithm;
            if (!algorithms.TryGetValue(algorithmName, out algorithm))
            {
                throw new InvalidOperationException("Hashing algorithm not supported");
            }

            return algorithm;
        }

        private static void Initialize()
        {
            if (algorithms.Count == 0)
            {
                if (!isFipsEnforced)
                {
                    algorithms.Add(RIPEMD160.Name, new RIPEMD160());
                    algorithms.Add(Whirlpool.Name, new Whirlpool());
                }

                algorithms.Add(SHA1.Name, new SHA1());
                algorithms.Add(SHA256.Name, new SHA256());
                algorithms.Add(SHA384.Name, new SHA384());
                algorithms.Add(SHA512.Name, new SHA512());
            }
        }
    }
}
