﻿using System;
using System.Text;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Common.Model.Protection
{
    /// <summary>
    /// Contains helper methods for generating protection salt and hash values.
    /// </summary>
    internal static class ProtectionHelper
    {
        private static readonly int saltLength = 16;
        private static readonly Random random = new Random();
        private static readonly char nConstant = 'N';
        private static readonly char kConstant = 'K';

        /// <summary>
        /// Generates base64 salt.
        /// </summary>
        /// <returns></returns>
        public static string GenerateSaltBase64()
        {
            byte[] buffer = new byte[saltLength];
            random.NextBytes(buffer);
            return Convert.ToBase64String(buffer);
        }

        /// <summary>
        /// Generates base64 hash.
        /// </summary>
        /// <param name="salt">The salt.</param>
        /// <param name="password">The password.</param>
        /// <param name="algorithmName">Name of the algorithm.</param>
        /// <param name="spinCount">The spin count.</param>
        /// <returns></returns>
        public static string GenerateHashBase64(string salt, string password, string algorithmName, int spinCount)
        {
            Guard.ThrowExceptionIfNullOrEmpty(salt, "salt");
            Guard.ThrowExceptionIfNullOrEmpty(password, "password");
            Guard.ThrowExceptionIfNullOrEmpty(algorithmName, "algorithmName");
            Guard.ThrowExceptionIfLessThan(1, spinCount, "spinCount");

            IProtectionAlgorithm algorithm = HashingAlgorithms.GetProtectionAlgorithm(algorithmName);

            byte[] legacyHashBytes = Encoding.Unicode.GetBytes(string.Format("{0:X8}", password));
            byte[] hash = Concatenate(Convert.FromBase64String(salt), legacyHashBytes);

            for (int i = 0; i < spinCount; i++)
            {
                hash = algorithm.ComputeHash(hash);
                hash = Concatenate(hash, i);
            }
            hash = algorithm.ComputeHash(hash);

            return Convert.ToBase64String(hash);
        }

        /// <summary>
        /// Generates the password using OOXML legacy scheme.
        /// </summary>
        /// <param name="password">The password.</param>
        /// <returns></returns>
        public static String GeneratePasswordLegacyScheme(String password)
        {
            Guard.ThrowExceptionIfNullOrEmpty(password, "password");

            Int32 pLength = password.Length;
            Int32 hash = 0;
            if (pLength == 0) return hash.ToString("X");

            for (Int32 i = pLength - 1; i >= 0; i--)
            {
                hash = hash >> 14 & 0x01 | hash << 1 & 0x7fff;
                hash ^= password[i];
            }
            hash = hash >> 14 & 0x01 | hash << 1 & 0x7fff;
            hash ^= 0x8000 | nConstant << 8 | kConstant;
            hash ^= pLength;
            return hash.ToString("X");
        }

        public static bool IsPasswordCorrect(string password, string hash, string salt, string algorithmName, int spinCount)
        {
            bool isPasswordCorrect = false;

            if (string.IsNullOrEmpty(hash) && string.IsNullOrEmpty(password))
            {
                isPasswordCorrect = true;
            }
            else if (!string.IsNullOrEmpty(hash) && !string.IsNullOrEmpty(password))
            {
                string passwordHash = ProtectionHelper.GenerateHashBase64(salt, password, algorithmName, spinCount);
                isPasswordCorrect = string.Equals(hash, passwordHash, StringComparison.Ordinal);
            }

            return isPasswordCorrect;
        }

        private static byte[] Concatenate(byte[] a, int b)
        {
            byte[] bytes = BitConverter.GetBytes(b);

            byte[] result = new byte[a.Length + bytes.Length];
            Array.Copy(a, result, a.Length);

            result[a.Length] = bytes[0];
            result[a.Length + 1] = bytes[1];
            result[a.Length + 2] = bytes[2];
            result[a.Length + 3] = bytes[3];

            return result;
        }

        private static byte[] Concatenate(byte[] a, byte[] b)
        {
            byte[] result = new byte[a.Length + b.Length];
            Array.Copy(a, result, a.Length);
            Array.Copy(b, 0, result, a.Length, b.Length);
            return result;
        }
    }
}
