﻿using System;
using System.Text;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Common.Model.Protection
{
    internal abstract class ProtectionHelperBase
    {
        private static readonly int saltLength = 16;
        private static readonly Random random = new Random();

        /// <summary>
        /// Generates base64 salt.
        /// </summary>
        /// <returns></returns>
        internal virtual string GenerateSaltBase64()
        {
            byte[] buffer = new byte[saltLength];
            random.NextBytes(buffer);
            return Convert.ToBase64String(buffer);
        }

        /// <summary>
        /// Generates base64 hash.
        /// </summary>
        /// <param name="salt">The salt.</param>
        /// <param name="password">The password.</param>
        /// <param name="algorithmName">Name of the algorithm.</param>
        /// <param name="spinCount">The spin count.</param>
        /// <returns></returns>
        internal virtual string GenerateHashBase64(string password, string salt, string algorithmName, int spinCount)
        {
            Guard.ThrowExceptionIfNullOrEmpty(password, "password");
            Guard.ThrowExceptionIfNullOrEmpty(salt, "salt");
            Guard.ThrowExceptionIfNullOrEmpty(algorithmName, "algorithmName");
            Guard.ThrowExceptionIfLessThan(1, spinCount, "spinCount");

            password = this.PreprocessPassword(password);

            byte[] legacyHashBytes = Encoding.Unicode.GetBytes(string.Format("{0:X8}", password));

            IProtectionAlgorithm algorithm = HashingAlgorithmsProvider.GetProtectionAlgorithm(algorithmName);

            byte[] hash = Concatenate(Convert.FromBase64String(salt), legacyHashBytes);

            for (int i = 0; i < spinCount; i++)
            {
                hash = algorithm.ComputeHash(hash);
                hash = Concatenate(hash, i);
            }
            hash = algorithm.ComputeHash(hash);

            return Convert.ToBase64String(hash);
        }

        protected virtual string PreprocessPassword(string password)
        {
            return password;
        }

        /// <summary>
        /// Determines whether [is password correct] [the specified password].
        /// </summary>
        /// <param name="password">The password.</param>
        /// <param name="hash">The hash.</param>
        /// <param name="salt">The salt.</param>
        /// <param name="algorithmName">Name of the algorithm.</param>
        /// <param name="spinCount">The spin count.</param>
        /// <returns></returns>
        internal virtual bool IsPasswordCorrect(string password, string hash, string salt, string algorithmName, int spinCount)
        {
            bool isPasswordCorrect = false;

            if (string.IsNullOrEmpty(hash) && string.IsNullOrEmpty(password))
            {
                isPasswordCorrect = true;
            }
            else if (!string.IsNullOrEmpty(hash) && !string.IsNullOrEmpty(password))
            {
                string passwordHash = this.GenerateHashBase64(password, salt, algorithmName, spinCount);

                isPasswordCorrect = string.Equals(hash, passwordHash, StringComparison.Ordinal);
            }

            return isPasswordCorrect;
        }

        private static byte[] Concatenate(byte[] a, int b)
        {
            byte[] bytes = BitConverter.GetBytes(b);

            byte[] result = new byte[a.Length + bytes.Length];
            Array.Copy(a, result, a.Length);

            result[a.Length] = bytes[0];
            result[a.Length + 1] = bytes[1];
            result[a.Length + 2] = bytes[2];
            result[a.Length + 3] = bytes[3];

            return result;
        }

        private static byte[] Concatenate(byte[] a, byte[] b)
        {
            byte[] result = new byte[a.Length + b.Length];
            Array.Copy(a, result, a.Length);
            Array.Copy(b, 0, result, a.Length, b.Length);
            return result;
        }
    }
}
