﻿using System;

/*
The Bouncy Castle Cryptographic C#® API

License:

The Bouncy Castle License
Copyright (c) 2000-2011 The Legion Of The Bouncy Castle (http://www.bouncycastle.org)
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sub license, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
 */

namespace Telerik.Windows.Documents.Common.Model.Protection
{
    /*
       Contributors:

       The following people have contributed to the C# Bouncy Castle Cryptography Package.
       Thanks, may your castles never deflate!

       Kaiser Yang <kaiseryang@yahoo.com> - initial port of the lightweight API and ASN.1 library. Finding BigInteger loop problem.
       Asier Murciego <a.murciego@captiva-sys.es> - Further patching to BigInteger library.
       Megan Woods <megan.woods@widestreet.com.au> - X509 certificate generation, RSA/DSA digest signature classes.
       David Del Vecchio <ddelvecc@virginia.edu> - patches to RSA Pkcs1 Signature generation OID issues, help with clarifications on DateTime and certificates.
       Nelson Fernandez <nelson-bc@kpanic.com.ar> - patches to allow compilation under mono.
       Paulo Soares <psoares@consiste.pt> - patches to X509CertificateParser, C# port of JZlib plus inflater/deflater streams, C# port of Apache BZip2 classes.
       Pawel Niewiadomski <11110000b@gmail.com> - patches for X509 and CMS, unit test for time classes.
       Jesper Johansen <jesper@hc.jay.net> - bug fix for DerT61String encodings.
       Adam Sternberg <agsternberg@gmail.com> - identified problem with generation of PGP public keyrings.
       Kirill Zhuklinets <zhuklinets_k@gaz-is.ru> - initial submission of bulk of Asn1.Esf classes (RFC 3126).
       Dr Andrew Gray <andrew.gray@rcrt.co.uk> - identified problem with BigInteger.ModPow for negative exponents.
       Mauricio Ulate <mulate@gmail.com> - identified problem with non-ASCII pass phrases in PGP.
       John Allberg <John.Allberg@teliasonera.com> - initial implementation of CryptoApiRandomGenerator.
       Mattias Öhrn <mattias.ohrn@gmail.com> - identified problem with Pkcs12Store.Save and provided fix.
       Jen Andre <jandre@gmail.com> - initial implementation of case-insensitive searches for PGP keyrings.
       #Cyrille37# <cyrille37@gmail.com> - identified problem with BigInteger.ModInverse for negative values.
       David Reis Jr <davidreis@yahoo.com> - bug fix for X509CrlStoreSelector handling of NextUpdate, fix handling of null parameters for DSA in key factories, initial port of Pkix namespace and supporting tests.
       Ivan Peev <ivan.peev@cozyroc.com> - bug fix for version string displayed in PGP armored output.
       Hector Ornelas Aciga <hector.ornelas@sat.gob.mx> - patch to add support for PKCS#5 Scheme 2 keys.
       Tom Van Holle <tvh@dsoft.be> - patch to add new class: Pkcs10CertificationRequestDelaySigned.
    */

    internal sealed class Whirlpool : IDigest, IProtectionAlgorithm
    {
        public const string Name = "WHIRLPOOL";

        private const int ByteLength = 64;

        private const int DigestLengthBytes = 512 / 8;
        private const int Rounds = 10;
        private const int ReductionPolynomial = 0x011d; // 2^8 + 2^4 + 2^3 + 2 + 1;

        private static readonly int[] sbox =
        {
            0x18, 0x23, 0xc6, 0xe8, 0x87, 0xb8, 0x01, 0x4f, 0x36, 0xa6, 0xd2, 0xf5, 0x79, 0x6f, 0x91, 0x52,
            0x60, 0xbc, 0x9b, 0x8e, 0xa3, 0x0c, 0x7b, 0x35, 0x1d, 0xe0, 0xd7, 0xc2, 0x2e, 0x4b, 0xfe, 0x57,
            0x15, 0x77, 0x37, 0xe5, 0x9f, 0xf0, 0x4a, 0xda, 0x58, 0xc9, 0x29, 0x0a, 0xb1, 0xa0, 0x6b, 0x85,
            0xbd, 0x5d, 0x10, 0xf4, 0xcb, 0x3e, 0x05, 0x67, 0xe4, 0x27, 0x41, 0x8b, 0xa7, 0x7d, 0x95, 0xd8,
            0xfb, 0xee, 0x7c, 0x66, 0xdd, 0x17, 0x47, 0x9e, 0xca, 0x2d, 0xbf, 0x07, 0xad, 0x5a, 0x83, 0x33,
            0x63, 0x02, 0xaa, 0x71, 0xc8, 0x19, 0x49, 0xd9, 0xf2, 0xe3, 0x5b, 0x88, 0x9a, 0x26, 0x32, 0xb0,
            0xe9, 0x0f, 0xd5, 0x80, 0xbe, 0xcd, 0x34, 0x48, 0xff, 0x7a, 0x90, 0x5f, 0x20, 0x68, 0x1a, 0xae,
            0xb4, 0x54, 0x93, 0x22, 0x64, 0xf1, 0x73, 0x12, 0x40, 0x08, 0xc3, 0xec, 0xdb, 0xa1, 0x8d, 0x3d,
            0x97, 0x00, 0xcf, 0x2b, 0x76, 0x82, 0xd6, 0x1b, 0xb5, 0xaf, 0x6a, 0x50, 0x45, 0xf3, 0x30, 0xef,
            0x3f, 0x55, 0xa2, 0xea, 0x65, 0xba, 0x2f, 0xc0, 0xde, 0x1c, 0xfd, 0x4d, 0x92, 0x75, 0x06, 0x8a,
            0xb2, 0xe6, 0x0e, 0x1f, 0x62, 0xd4, 0xa8, 0x96, 0xf9, 0xc5, 0x25, 0x59, 0x84, 0x72, 0x39, 0x4c,
            0x5e, 0x78, 0x38, 0x8c, 0xd1, 0xa5, 0xe2, 0x61, 0xb3, 0x21, 0x9c, 0x1e, 0x43, 0xc7, 0xfc, 0x04,
            0x51, 0x99, 0x6d, 0x0d, 0xfa, 0xdf, 0x7e, 0x24, 0x3b, 0xab, 0xce, 0x11, 0x8f, 0x4e, 0xb7, 0xeb,
            0x3c, 0x81, 0x94, 0xf7, 0xb9, 0x13, 0x2c, 0xd3, 0xe7, 0x6e, 0xc4, 0x03, 0x56, 0x44, 0x7f, 0xa9,
            0x2a, 0xbb, 0xc1, 0x53, 0xdc, 0x0b, 0x9d, 0x6c, 0x31, 0x74, 0xf6, 0x46, 0xac, 0x89, 0x14, 0xe1,
            0x16, 0x3a, 0x69, 0x09, 0x70, 0xb6, 0xd0, 0xed, 0xcc, 0x42, 0x98, 0xa4, 0x28, 0x5c, 0xf8, 0x86
        };

        private static readonly long[] c0 = new long[256];
        private static readonly long[] c1 = new long[256];
        private static readonly long[] c2 = new long[256];
        private static readonly long[] c3 = new long[256];
        private static readonly long[] c4 = new long[256];
        private static readonly long[] c5 = new long[256];
        private static readonly long[] c6 = new long[256];
        private static readonly long[] c7 = new long[256];

        private readonly long[] rc = new long[Rounds + 1];

        /*
            * increment() can be implemented in this way using 2 arrays or
            * by having some temporary variables that are used to set the
            * value provided by EIGHT[i] and carry within the loop.
            *
            * not having done any timing, this seems likely to be faster
            * at the slight expense of 32*(sizeof short) bytes
            */
        private static readonly short[] eight = new short[BitcountArraySize];

        static Whirlpool()
        {
            eight[BitcountArraySize - 1] = 8;

            for (int i = 0; i < 256; i++)
            {
                int v1 = sbox[i];
                int v2 = MaskWithReductionPolynomial(v1 << 1);
                int v4 = MaskWithReductionPolynomial(v2 << 1);
                int v5 = v4 ^ v1;
                int v8 = MaskWithReductionPolynomial(v4 << 1);
                int v9 = v8 ^ v1;

                c0[i] = PackIntoLong(v1, v1, v4, v1, v8, v5, v2, v9);
                c1[i] = PackIntoLong(v9, v1, v1, v4, v1, v8, v5, v2);
                c2[i] = PackIntoLong(v2, v9, v1, v1, v4, v1, v8, v5);
                c3[i] = PackIntoLong(v5, v2, v9, v1, v1, v4, v1, v8);
                c4[i] = PackIntoLong(v8, v5, v2, v9, v1, v1, v4, v1);
                c5[i] = PackIntoLong(v1, v8, v5, v2, v9, v1, v1, v4);
                c6[i] = PackIntoLong(v4, v1, v8, v5, v2, v9, v1, v1);
                c7[i] = PackIntoLong(v1, v4, v1, v8, v5, v2, v9, v1);
            }
        }

        public Whirlpool()
        {
            this.rc[0] = 0L;
            for (int r = 1; r <= Rounds; r++)
            {
                int i = 8 * (r - 1);
                this.rc[r] = (long)((ulong)c0[i] & 0xff00000000000000L) ^
                    (c1[i + 1] & 0x00ff000000000000L) ^
                    (c2[i + 2] & 0x0000ff0000000000L) ^
                    (c3[i + 3] & 0x000000ff00000000L) ^
                    (c4[i + 4] & 0x00000000ff000000L) ^
                    (c5[i + 5] & 0x0000000000ff0000L) ^
                    (c6[i + 6] & 0x000000000000ff00L) ^
                    (c7[i + 7] & 0x00000000000000ffL);
            }
        }

        private static long PackIntoLong(int b7, int b6, int b5, int b4, int b3, int b2, int b1, int b0)
        {
            return
                ((long)b7 << 56) ^
                ((long)b6 << 48) ^
                ((long)b5 << 40) ^
                ((long)b4 << 32) ^
                ((long)b3 << 24) ^
                ((long)b2 << 16) ^
                ((long)b1 << 8) ^
                b0;
        }

        /*
            * int's are used to prevent sign extension.  The values that are really being used are
            * actually just 0..255
            */
        private static int MaskWithReductionPolynomial(int input)
        {
            int rv = input;
            if (rv >= 0x100L) // high bit set
            {
                rv ^= ReductionPolynomial; // reduced by the polynomial
            }
            return rv;
        }

        // --------------------------------------------------------------------------------------//

        // -- buffer information --
        private const int BitcountArraySize = 32;
        private readonly byte[] buffer = new byte[64];
        private int bufferPosition;
        private readonly short[] bitCount = new short[BitcountArraySize];

        // -- internal hash state --
        private readonly long[] hash = new long[8];
        private readonly long[] k = new long[8]; // the round key
        private readonly long[] l = new long[8];
        private readonly long[] block = new long[8]; // mu (buffer)
        private readonly long[] state = new long[8]; // the current "cipher" state

        /**
            * Copy constructor. This will copy the state of the provided message
            * digest.
        */
        public Whirlpool(Whirlpool originalDigest)
        {
            Array.Copy(originalDigest.rc, 0, this.rc, 0, this.rc.Length);

            Array.Copy(originalDigest.buffer, 0, this.buffer, 0, this.buffer.Length);

            this.bufferPosition = originalDigest.bufferPosition;
            Array.Copy(originalDigest.bitCount, 0, this.bitCount, 0, this.bitCount.Length);

            // -- internal hash state --
            Array.Copy(originalDigest.hash, 0, this.hash, 0, this.hash.Length);
            Array.Copy(originalDigest.k, 0, this.k, 0, this.k.Length);
            Array.Copy(originalDigest.l, 0, this.l, 0, this.l.Length);
            Array.Copy(originalDigest.block, 0, this.block, 0, this.block.Length);
            Array.Copy(originalDigest.state, 0, this.state, 0, this.state.Length);
        }

        public string AlgorithmName
        {
            get { return Name; }
        }

        public int GetDigestSize()
        {
            return DigestLengthBytes;
        }

        public int DoFinal(byte[] output, int outOff)
        {
            // sets output[outOff] .. output[outOff+DIGEST_LENGTH_BYTES]
            this.Finish();

            for (int i = 0; i < 8; i++)
            {
                ConvertLongToByteArray(this.hash[i], output, outOff + (i * 8));
            }

            this.Reset();

            return this.GetDigestSize();
        }

        /**
            * Reset the chaining variables
        */
        public void Reset()
        {
            // set variables to null, blank, whatever
            this.bufferPosition = 0;
            Array.Clear(this.bitCount, 0, this.bitCount.Length);
            Array.Clear(this.buffer, 0, this.buffer.Length);
            Array.Clear(this.hash, 0, this.hash.Length);
            Array.Clear(this.k, 0, this.k.Length);
            Array.Clear(this.l, 0, this.l.Length);
            Array.Clear(this.block, 0, this.block.Length);
            Array.Clear(this.state, 0, this.state.Length);
        }

        // this takes a buffer of information and fills the block
        private void ProcessFilledBuffer()
        {
            // copies into the block...
            for (int i = 0; i < this.state.Length; i++)
            {
                this.block[i] = BytesToLongFromBuffer(this.buffer, i * 8);
            }
            this.ProcessBlock();
            this.bufferPosition = 0;
            Array.Clear(this.buffer, 0, this.buffer.Length);
        }

        private static long BytesToLongFromBuffer(byte[] buffer, int startPos)
        {
            long rv = (((buffer[startPos + 0] & 0xffL) << 56) |
                ((buffer[startPos + 1] & 0xffL) << 48) |
                ((buffer[startPos + 2] & 0xffL) << 40) |
                ((buffer[startPos + 3] & 0xffL) << 32) |
                ((buffer[startPos + 4] & 0xffL) << 24) |
                ((buffer[startPos + 5] & 0xffL) << 16) |
                ((buffer[startPos + 6] & 0xffL) << 8) |
                ((buffer[startPos + 7]) & 0xffL));

            return rv;
        }

        private static void ConvertLongToByteArray(long inputLong, byte[] outputArray, int offSet)
        {
            for (int i = 0; i < 8; i++)
            {
                outputArray[offSet + i] = (byte)((inputLong >> (56 - (i * 8))) & 0xff);
            }
        }

        private void ProcessBlock()
        {
            // buffer contents have been transferred to the _block[] array via
            // processFilledBuffer

            // compute and apply K^0
            for (int i = 0; i < 8; i++)
            {
                this.state[i] = this.block[i] ^ (this.k[i] = this.hash[i]);
            }

            // iterate over the rounds
            for (int round = 1; round <= Rounds; round++)
            {
                for (int i = 0; i < 8; i++)
                {
                    this.l[i] = 0;
                    this.l[i] ^= c0[(int)(this.k[(i - 0) & 7] >> 56) & 0xff];
                    this.l[i] ^= c1[(int)(this.k[(i - 1) & 7] >> 48) & 0xff];
                    this.l[i] ^= c2[(int)(this.k[(i - 2) & 7] >> 40) & 0xff];
                    this.l[i] ^= c3[(int)(this.k[(i - 3) & 7] >> 32) & 0xff];
                    this.l[i] ^= c4[(int)(this.k[(i - 4) & 7] >> 24) & 0xff];
                    this.l[i] ^= c5[(int)(this.k[(i - 5) & 7] >> 16) & 0xff];
                    this.l[i] ^= c6[(int)(this.k[(i - 6) & 7] >> 8) & 0xff];
                    this.l[i] ^= c7[(int)(this.k[(i - 7) & 7]) & 0xff];
                }

                Array.Copy(this.l, 0, this.k, 0, this.k.Length);

                this.k[0] ^= this.rc[round];

                // apply the round transformation
                for (int i = 0; i < 8; i++)
                {
                    this.l[i] = this.k[i];

                    this.l[i] ^= c0[(int)(this.state[(i - 0) & 7] >> 56) & 0xff];
                    this.l[i] ^= c1[(int)(this.state[(i - 1) & 7] >> 48) & 0xff];
                    this.l[i] ^= c2[(int)(this.state[(i - 2) & 7] >> 40) & 0xff];
                    this.l[i] ^= c3[(int)(this.state[(i - 3) & 7] >> 32) & 0xff];
                    this.l[i] ^= c4[(int)(this.state[(i - 4) & 7] >> 24) & 0xff];
                    this.l[i] ^= c5[(int)(this.state[(i - 5) & 7] >> 16) & 0xff];
                    this.l[i] ^= c6[(int)(this.state[(i - 6) & 7] >> 8) & 0xff];
                    this.l[i] ^= c7[(int)(this.state[(i - 7) & 7]) & 0xff];
                }

                // save the current state
                Array.Copy(this.l, 0, this.state, 0, this.state.Length);
            }

            // apply Miuaguchi-Preneel compression
            for (int i = 0; i < 8; i++)
            {
                this.hash[i] ^= this.state[i] ^ this.block[i];
            }
        }

        public void Update(byte input)
        {
            this.buffer[this.bufferPosition] = input;

            //Console.WriteLine("adding to buffer = "+_buffer[_bufferPos]);

            ++this.bufferPosition;

            if (this.bufferPosition == this.buffer.Length)
            {
                this.ProcessFilledBuffer();
            }

            this.Increment();
        }

        private void Increment()
        {
            int carry = 0;
            for (int i = this.bitCount.Length - 1; i >= 0; i--)
            {
                int sum = (this.bitCount[i] & 0xff) + eight[i] + carry;

                carry = sum >> 8;
                this.bitCount[i] = (short)(sum & 0xff);
            }
        }

        public void BlockUpdate(byte[] input, int inOff, int length)
        {
            while (length > 0)
            {
                this.Update(input[inOff]);
                ++inOff;
                --length;
            }
        }

        private void Finish()
        {
            /*
                * this makes a copy of the current bit length. at the expense of an
                * object creation of 32 bytes rather than providing a _stopCounting
                * boolean which was the alternative I could think of.
                */
            byte[] bitLength = this.CopyBitLength();

            this.buffer[this.bufferPosition++] |= 0x80;

            if (this.bufferPosition == this.buffer.Length)
            {
                this.ProcessFilledBuffer();
            }

            /*
                * Final block contains
                * [ ... data .... ][0][0][0][ length ]
                *
                * if [ length ] cannot fit.  Need to create a new block.
                */
            if (this.bufferPosition > 32)
            {
                while (this.bufferPosition != 0)
                {
                    this.Update(0);
                }
            }

            while (this.bufferPosition <= 32)
            {
                this.Update(0);
            }

            // copy the length information to the final 32 bytes of the
            // 64 byte block....
            Array.Copy(bitLength, 0, this.buffer, 32, bitLength.Length);

            this.ProcessFilledBuffer();
        }

        private byte[] CopyBitLength()
        {
            byte[] rv = new byte[BitcountArraySize];
            for (int i = 0; i < rv.Length; i++)
            {
                rv[i] = (byte)(this.bitCount[i] & 0xff);
            }
            return rv;
        }

        public int GetByteLength()
        {
            return ByteLength;
        }

        public byte[] ComputeHash(byte[] buffer)
        {
            byte[] result = new byte[this.GetDigestSize()];

            this.BlockUpdate(buffer, 0, buffer.Length);
            this.DoFinal(result, 0);

            return result;
        }
    }
}
