﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Data
{
    /// <summary>
    /// Provides based functionality for collections of document elements.
    /// </summary>
    /// <typeparam name="T">The type of the document elements which should be added to the collection.</typeparam>
    /// <typeparam name="TOwner">The type of the owner of the collection.</typeparam>
    public abstract class DocumentElementCollectionBase<T, TOwner> : IList<T>
    {
        private readonly List<T> store;

        internal DocumentElementCollectionBase(TOwner owner)
        {
            Guard.ThrowExceptionIfNull(owner, "owner");

            this.Owner = owner;
            this.store = new List<T>();
        }

        /// <summary>
        /// Gets the number of elements contained in the collection.
        /// </summary>
        /// <returns>The number of elements contained in the collection.</returns>
        public int Count
        {
            get
            {
                return this.store.Count;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the collection is read-only.
        /// </summary>
        /// <returns><c>true</c> if the collection is read-only; otherwise, <c>false</c>.</returns>
        bool ICollection<T>.IsReadOnly
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Gets the element holding the collection.
        /// </summary>
        /// <value>The owner element.</value>
        protected TOwner Owner { get; private set; }

        /// <summary>
        /// Gets or sets the element at the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <returns></returns>
        public T this[int index]
        {
            get
            {
                return this.store[index];
            }
            set
            {
                this.VerifyDocumentElementOnInsert(value);

                var oldElement = this.store[index];

                this.store[index] = value;

                this.OnAfterRemove(oldElement);

                this.SetParent(oldElement, default(TOwner));
                this.SetParent(value, this.Owner);
            }
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<T> GetEnumerator()
        {
            return this.store.GetEnumerator();
        }

        /// <summary>
        /// Determines the index of a specific element in the collection.
        /// </summary>
        /// <param name="item">The element to locate in the collection.</param>
        /// <returns>
        /// The index of <paramref name="item" /> if found in the list; otherwise, -1.
        /// </returns>
        public int IndexOf(T item)
        {
            return this.store.IndexOf(item);
        }

        /// <summary>
        /// Inserts an element to the collection at the specified index.
        /// </summary>
        /// <param name="index">The zero-based index at which <paramref name="item" /> should be inserted.</param>
        /// <param name="item">The element to insert into the collection.</param>
        public void Insert(int index, T item)
        {
            this.VerifyDocumentElementOnInsert(item);

            this.store.Insert(index, item);

            this.SetParent(item, this.Owner);
        }

        /// <summary>
        /// Inserts an elements to the collection at the specified index.
        /// </summary>
        /// <param name="index">The zero-based index at which <paramref name="items" /> should be inserted.</param>
        /// <param name="items">The elements to be inserted into the collection.</param>
        public void InsertRange(int index, IEnumerable<T> items)
        {
            foreach (var item in items)
            {
                this.VerifyDocumentElementOnInsert(item);
            }

            this.store.InsertRange(index, items);

            foreach (var item in items)
            {
                this.SetParent(item, this.Owner);
            }
        }


        /// <summary>
        /// Removes the element at the specified index.
        /// </summary>
        /// <param name="index">The zero-based index of the item to remove.</param>
        public void RemoveAt(int index)
        {
            T item = this.store[index];

            this.store.RemoveAt(index);

            this.OnAfterRemove(item);

            this.SetParent(item, default(TOwner));
        }

        /// <summary>
        /// Removes the range.
        /// </summary>
        /// <param name="index">The zero-based index of the item to remove.</param>
        /// <param name="count">The number of elements to remove.</param>
        public void RemoveRange(int index, int count)
        {
            var items = this.store.Skip(index).Take(count).ToArray();

            this.store.RemoveRange(index, count);

            foreach (var item in items)
            {
                this.OnAfterRemove(item);
                this.SetParent(item, default(TOwner));
            }
        }

        /// <summary>
        /// Called when the element is removed.
        /// </summary>
        /// <param name="item">The element.</param>
        protected virtual void OnAfterRemove(T item)
        {
        }

        /// <summary>
        /// Removes all items from the collection.
        /// </summary>
        public void Clear()
        {
            for (int i = this.store.Count - 1; i >= 0; i--)
            {
                this.RemoveAt(i);
            }
        }

        /// <summary>
        /// Determines whether the collection contains a specific element.
        /// </summary>
        /// <param name="item">The element to locate in the collection.</param>
        /// <returns>
        /// <c>true</c> if <paramref name="item" /> is found in the collection; otherwise, <c>false</c>.
        /// </returns>
        public bool Contains(T item)
        {
            return this.store.Contains(item);
        }

        /// <summary>
        /// Copies the elements of the collection to an <see cref="Array"/>, starting at a particular <see cref="Array"/> index.
        /// </summary>
        /// <param name="array">The array.</param>
        /// <param name="arrayIndex">Index of the array.</param>
        public void CopyTo(T[] array, int arrayIndex)
        {
            this.store.CopyTo(array, arrayIndex);
        }

        /// <summary>
        /// Removes the first occurrence of a specific object from the collection.
        /// </summary>
        /// <param name="item">The object to remove from the collection.</param>
        /// <returns>
        /// <c>true</c> if <paramref name="item" /> was successfully removed from the collection; otherwise, <c>false</c>. This method also returns false if <paramref name="item" /> is not found in the original <see cref="T:System.Collections.Generic.ICollection`1" />.
        /// </returns>
        public bool Remove(T item)
        {
            bool result = this.store.Remove(item);
            if (result)
            {
                this.OnAfterRemove(item);
                this.SetParent(item, default(TOwner));
            }

            return result;
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.store.GetEnumerator();
        }

        /// <summary>
        /// Adds an item to the collection.
        /// </summary>
        /// <param name="item">The element to add to the collection.</param>
        public void Add(T item)
        {
            this.VerifyDocumentElementOnInsert(item);

            this.store.Add(item);

            this.SetParent(item, this.Owner);
        }

        /// <summary>
        /// Verifies the validity of the document element before it is inserted in the collection. 
        /// </summary>
        /// <param name="item">The item.</param>
        protected abstract void VerifyDocumentElementOnInsert(T item);

        /// <summary>
        /// Sets the parent of the document element.
        /// </summary>
        /// <param name="item">The document element.</param>
        /// <param name="parent">The parent which should be set to the document element.</param>
        protected abstract void SetParent(T item, TOwner parent);
    }
}
