﻿using System;
#if NETSTANDARD
using Telerik.Documents.Primitives;
using Point = Telerik.Documents.Primitives.Point;
using MediaMatrix = Telerik.Documents.Primitives.Matrix;
#else
using System.Windows;
using Point = System.Windows.Point;
using MediaMatrix = System.Windows.Media.Matrix;
#endif

namespace Telerik.Windows.Documents.Core.Data
{
    internal struct Matrix
    {
        public Matrix(double m11, double m12, double m21, double m22, double offsetX, double offsetY)
            : this()
        {
            this.M11 = m11;
            this.M12 = m12;
            this.M21 = m21;
            this.M22 = m22;
            this.OffsetX = offsetX;
            this.OffsetY = offsetY;
        }
        public Matrix(MediaMatrix matrix)
            : this(matrix.M11, matrix.M12, matrix.M21, matrix.M22, matrix.OffsetX, matrix.OffsetY)
        {
        }

        public static Matrix Identity
        {
            get
            {
                return new Matrix(1, 0, 0, 1, 0, 0);
            }
        }

        public double Determinant
        {
            get
            {
                return (this.M11 * this.M22) - (this.M12 * this.M21);
            }
        }

        public double M11 { get; set; }

        public double M12 { get; set; }

        public double M21 { get; set; }

        public double M22 { get; set; }

        public double OffsetX { get; set; }

        public double OffsetY { get; set; }

        public static Matrix operator *(Matrix matrix1, Matrix matrix2)
        {
            return new Matrix((matrix1.M11 * matrix2.M11) + (matrix1.M12 * matrix2.M21), (matrix1.M11 * matrix2.M12) + (matrix1.M12 * matrix2.M22), (matrix1.M21 * matrix2.M11) + (matrix1.M22 * matrix2.M21), (matrix1.M21 * matrix2.M12) + (matrix1.M22 * matrix2.M22), ((matrix1.OffsetX * matrix2.M11) + (matrix1.OffsetY * matrix2.M21)) + matrix2.OffsetX, ((matrix1.OffsetX * matrix2.M12) + (matrix1.OffsetY * matrix2.M22)) + matrix2.OffsetY);
        }

        public static bool operator ==(Matrix a, Matrix b)
        {
            return a.M11 == b.M11 && a.M21 == b.M21 && a.M12 == b.M12 && a.M22 == b.M22 && a.OffsetX == b.OffsetX && a.OffsetY == b.OffsetY;
        }

        public static bool operator !=(Matrix a, Matrix b)
        {
            return !(a == b);
        }

        public bool IsIdentity()
        {
            return this == Matrix.Identity;
        }

        public Matrix Translate(double offsetX, double offsetY)
        {
            this.OffsetX += offsetX;
            this.OffsetY += offsetY;

            return this;
        }

        public Matrix Scale(double scaleX, double scaleY, double centerX = 0, double centerY = 0)
        {
            this = new Matrix(scaleX, 0, 0, scaleY, centerX, centerY) * this;

            return this;
        }

        public Matrix ScaleAppend(double scaleX, double scaleY, double centerX = 0, double centerY = 0)
        {
            this = this * new Matrix(scaleX, 0, 0, scaleY, centerX, centerY);

            return this;
        }

        public Matrix Rotate(double angle, double centerX = 0, double centerY = 0)
        {
            Matrix matrix = new Matrix();
            angle = Math.PI * angle / 180;
            double sin = Math.Sin(angle);
            double cos = Math.Cos(angle);
            double offsetX = (centerX * (1.0 - cos)) + (centerY * sin);
            double offsetY = (centerY * (1.0 - cos)) - (centerX * sin);
            matrix.SetMatrix(cos, sin, -sin, cos, offsetX, offsetY);
            this = matrix * this;

            return this;
        }

        public bool Equals(Matrix value)
        {
            return this.M11 == value.M11 &&
                   this.M12 == value.M12 &&
                   this.M21 == value.M21 &&
                   this.M22 == value.M22 &&
                   this.OffsetX == value.OffsetX &&
                   this.OffsetY == value.OffsetY;
        }

        public Point Transform(Point point)
        {
            double x = point.X;
            double y = point.Y;

            double rx = (x * this.M11) + (y * this.M21) + this.OffsetX;
            double ry = (x * this.M12) + (y * this.M22) + this.OffsetY;

            return new Point(rx, ry);
        }

        public Rect Transform(Rect rect)
        {
            Point p1 = new Point(rect.Top, rect.Left);
            Point p2 = new Point(rect.Bottom, rect.Right);

            return new Rect(this.Transform(p1), this.Transform(p2));
        }

        internal MediaMatrix ToMatrix()
        {
            return new MediaMatrix(this.M11, this.M12, this.M21, this.M22, this.OffsetX, this.OffsetY);
        }

        public override int GetHashCode()
        {
            unchecked
            {
                int result = 17;
                result = (result * 23) + this.M11.GetHashCode();
                result = (result * 23) + this.M12.GetHashCode();
                result = (result * 23) + this.M21.GetHashCode();
                result = (result * 23) + this.M22.GetHashCode();
                result = (result * 23) + this.OffsetX.GetHashCode();
                result = (result * 23) + this.OffsetY.GetHashCode();

                return result;
            }
        }

        public override bool Equals(object obj)
        {
            if (obj == null || !(obj is Matrix))
            {
                return false;
            }

            return this.Equals((Matrix)obj);
        }

        public override string ToString()
        {
            return string.Format("{0} {1} 0 | {2} {3} 0 | {4} {5} 1", this.M11, this.M12, this.M21, this.M22, this.OffsetX, this.OffsetY);
        }

        private void SetMatrix(double m11, double m12, double m21, double m22, double offsetX, double offsetY)
        {
            this.M11 = m11;
            this.M12 = m12;
            this.M21 = m21;
            this.M22 = m22;
            this.OffsetX = offsetX;
            this.OffsetY = offsetY;
        }
    }
}