﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace Telerik.Windows.Documents.Core.Data
{
    /// <summary>
    /// Represents Stack collection.
    /// </summary>
    public class StackCollection<T> : ICollection<T> where T : IStackCollectionElement
    {
        private readonly Dictionary<string, T> nameToElement;
        private readonly LinkedList<T> elements;

        /// <summary>
        /// Initializes a new instance of the StackCollection class.
        /// </summary>
        public StackCollection()
        {
            this.nameToElement = new Dictionary<string, T>();
            this.elements = new LinkedList<T>();
        }

        /// <summary>
        /// Gets the number of elements contained in the collection.
        /// </summary>
        /// <returns>The number of elements contained in the collection.
        /// </returns>
        /// <value></value>
        public int Count
        {
            get
            {
                return this.elements.Count;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the <see cref="T:System.Collections.Generic.ICollection`1" />
        /// is read-only.
        /// </summary>
        /// <returns>true if the <see cref="T:System.Collections.Generic.ICollection`1" />
        /// is read-only; otherwise, false.</returns>
        /// <value></value>
        public bool IsReadOnly
        {
            get
            {
                return ((ICollection<T>)this.elements).IsReadOnly;
            }
        }

        /// <summary>
        /// Gets the name of the element by.
        /// </summary>
        /// <param name="elementName">Name of the element.</param>
        /// <returns></returns>
        public T GetElementByName(string elementName)
        {
            T element;
            if (this.nameToElement.TryGetValue(elementName, out element))
            {
                return element;
            }

            return default(T);
        }

        /// <summary>
        /// Adds the specified item.
        /// </summary>
        /// <param name="item">The item.</param>
        public void Add(T item)
        {
            this.AddToDictionary(item);
            this.elements.AddLast(item);
        }

        /// <summary>
        /// Adds the before.
        /// </summary>
        /// <param name="presentedElementName">Name of the presented element.</param>
        /// <param name="layer">The layer.</param>
        /// <returns></returns>
        public bool AddBefore(string presentedElementName, T layer)
        {
            LinkedListNode<T> node = this.GetNodeByName(presentedElementName);
            if (node != null)
            {
                this.AddToDictionary(layer);
                this.elements.AddBefore(node, layer);
                return true;
            }

            return false;
        }

        /// <summary>
        /// Adds the after.
        /// </summary>
        /// <param name="presentedElementName">Name of the presented element.</param>
        /// <param name="element">The element.</param>
        /// <returns></returns>
        public bool AddAfter(string presentedElementName, T element)
        {
            LinkedListNode<T> node = this.GetNodeByName(presentedElementName);
            if (node != null)
            {
                this.AddToDictionary(element);
                this.elements.AddAfter(node, element);
                return true;
            }

            return false;
        }

        /// <summary>
        /// Adds the first.
        /// </summary>
        /// <param name="element">The element.</param>
        public void AddFirst(T element)
        {
            this.AddToDictionary(element);
            this.elements.AddFirst(element);
        }

        /// <summary>
        /// Adds the last.
        /// </summary>
        /// <param name="element">The element.</param>
        public void AddLast(T element)
        {
            this.AddToDictionary(element);
            this.elements.AddLast(element);
        }

        /// <summary>
        /// Determines whether [contains] [the specified item].
        /// </summary>
        /// <param name="item">The item.</param>
        /// <returns></returns>
        public bool Contains(T item)
        {
            return this.Contains(item.Name);
        }

        /// <summary>
        /// Determines whether [contains] [the specified element name].
        /// </summary>
        /// <param name="elementName">Name of the element.</param>
        /// <returns></returns>
        public bool Contains(string elementName)
        {
            return this.nameToElement.ContainsKey(elementName);
        }

        /// <summary>
        /// Removes the specified item.
        /// </summary>
        /// <param name="item">The item.</param>
        /// <returns></returns>
        public bool Remove(T item)
        {
            this.RemoveFromDictionary(item);
            return this.elements.Remove(item);
        }

        /// <summary>
        /// Removes the specified element name.
        /// </summary>
        /// <param name="elementName">Name of the element.</param>
        /// <returns></returns>
        public bool Remove(string elementName)
        {
            T layer;
            if (this.nameToElement.TryGetValue(elementName, out layer))
            {
                return this.Remove(layer);
            }

            return false;
        }

        /// <summary>
        /// Copies to.
        /// </summary>
        /// <param name="array">The array.</param>
        /// <param name="arrayIndex">Index of the array.</param>
        public void CopyTo(T[] array, int arrayIndex)
        {
            this.elements.CopyTo(array, arrayIndex);
        }

        /// <summary>
        /// Removes all items from the <see cref="T:System.Collections.Generic.ICollection`1" />.
        /// </summary>
        /// <exception cref="T:System.NotSupportedException">The <see cref="T:System.Collections.Generic.ICollection`1" />
        /// is read-only. </exception>
        public void Clear()
        {
            this.nameToElement.Clear();
            this.elements.Clear();
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<T> GetEnumerator()
        {
            return this.elements.GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return ((IEnumerable)this.elements).GetEnumerator();
        }

        private void AddToDictionary(T element)
        {
            if (this.nameToElement.ContainsKey(element.Name))
            {
                throw new InvalidOperationException("Element with name " + element.Name + " already exists.");
            }

            this.nameToElement.Add(element.Name, element);
        }

        private void RemoveFromDictionary(T element)
        {
            this.nameToElement.Remove(element.Name);
        }

        private LinkedListNode<T> GetNodeByName(string elementName)
        {
            LinkedListNode<T> node = this.elements.First;
            while (node != null && node.Value.Name != elementName)
            {
                node = node.Next;
            }

            return node;
        }
    }
}
