﻿using System;
using System.Linq;
using System.Windows;
#if !NETSTANDARD
using System.Windows.Media;
#else
using Telerik.Documents.Core.Fonts;
#endif

namespace Telerik.Windows.Documents.Core.Fonts
{
    /// <summary>
    /// Represents fonts properties class.
    /// </summary>
    public class FontProperties
    {
        private readonly FontFamily fontFamily;
        private readonly FontStyle fontStyle;
        private readonly FontWeight fontWeight;
        private readonly bool isMonoSpaced;

        /// <summary>
        /// Initializes a new instance of the <see cref="FontProperties" /> class.
        /// </summary>
        /// <param name="fontFamily">The font family.</param>
        /// <param name="fontStyle">The font style.</param>
        /// <param name="fontWeight">The font weight.</param>
        public FontProperties(FontFamily fontFamily, FontStyle fontStyle, FontWeight fontWeight)
        {
            this.fontFamily = fontFamily;
            this.fontStyle = fontStyle;
            this.fontWeight = fontWeight;
            this.isMonoSpaced = SystemFontsManager.IsMonospaced(fontFamily.Source);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="FontProperties" /> class.
        /// </summary>
        /// <param name="fontFamily">The font family.</param>
        public FontProperties(FontFamily fontFamily)
            : this(fontFamily, FontStyles.Normal, FontWeights.Normal)
        {
        }

        /// <summary>
        /// Gets font family name.
        /// </summary>
        /// <value>The name of the font family.</value>
        public string FontFamilyName
        {
            get
            {
                return this.FontFamily.Source;
            }
        }

        /// <summary>
        /// Gets the font weight.
        /// </summary>
        /// <value>The font weight.</value>
        public FontWeight FontWeight
        {
            get
            {
                return this.fontWeight;
            }
        }

        /// <summary>
        /// Gets the font style.
        /// </summary>
        /// <value>The font style.</value>
        public FontStyle FontStyle
        {
            get
            {
                return this.fontStyle;
            }
        }

        /// <summary>
        /// Gets the font family.
        /// </summary>
        /// <value>The font family.</value>
        public FontFamily FontFamily
        {
            get
            {
                return this.fontFamily;
            }
        }

        /// <summary>
        /// Gets if font is mono spaced.
        /// </summary>
        /// <value>The is mono spaced.</value>
        public bool IsMonoSpaced
        {
            get
            {
                return this.isMonoSpaced;
            }
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            FontProperties other = obj as FontProperties;
            if (other == null)
            {
                return false;
            }

#if !NETSTANDARD
            return this.FontFamilyName == other.FontFamilyName && this.FontStyle == other.FontStyle && this.FontWeight == other.FontWeight;
#else
            return this.FontFamilyName.Equals(other.FontFamilyName) && this.FontStyle.Equals(other.FontStyle) && this.FontWeight.Equals(other.FontWeight);
#endif
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            unchecked
            {
                int result = 17;
                result = result * 23 + this.fontFamily.GetHashCode();
                result = result * 23 + this.fontStyle.GetHashCode();
                result = result * 23 + this.fontWeight.GetHashCode();

                return result;
            }
        }
    }
}
