﻿/* Copyright 2012 Mozilla Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* Portions of /src/core/colorspace.js file as modified by Progress EAD - Copyright © 2016 Progress EAD. All rights reserved.
*/

using System;

namespace Telerik.Windows.Documents.Core.Imaging
{
    internal struct Color
    {
        private byte a;
        private byte r;
        private byte g;
        private byte b;

        public Color(byte a, byte r, byte g, byte b)
        {
            this.a = a;
            this.r = r;
            this.g = g;
            this.b = b;
        }

        public Color(byte r, byte g, byte b) 
            : this(255, r, g, b)
        {
        }

        public static Color Transparent
        {
            get
            {
                return Color.FromArgb(0, 0, 0, 0);
            }
        }

        public static Color Black
        {
            get
            {
                return Color.FromArgb(255, 0, 0, 0);
            }
        }

        public static Color White
        {
            get
            {
                return Color.FromArgb(255, 255, 255, 255);
            }
        }

        public byte A
        {
            get
            {
                return this.a;
            }
            set
            {
                this.a = value;
            }
        }

        public byte R
        {
            get
            {
                return this.r;
            }
            set
            {
                this.r = value;
            }
        }

        public byte G
        {
            get
            {
                return this.g;
            }
            set
            {
                this.g = value;
            }
        }

        public byte B
        {
            get
            {
                return this.b;
            }
            set
            {
                this.b = value;
            }
        }

        public static byte ConvertColorComponentToByte(double component)
        {
            if (component > 1)
            {
                return 255;
            }
            
            if (component < 0)
            {
                return 0;
            }
            
            return (byte)(component * 255);
        }

        public static Color FromColor(byte alpha, Color c)
        {
            return Color.FromArgb(alpha, c.r, c.g, c.b);
        }

        public static Color FromArgb(int alpha, int red, int green, int blue, int bits)
        {
            byte a = ConvertColorComponentToByte(alpha, bits);
            byte r = ConvertColorComponentToByte(red, bits);
            byte g = ConvertColorComponentToByte(green, bits);
            byte b = ConvertColorComponentToByte(blue, bits);
           
            return Color.FromArgb(a, r, g, b);
        }

        public static Color FromArgb(double alpha, double red, double green, double blue)
        {
            byte a = ConvertColorComponentToByte(alpha);
            byte r = ConvertColorComponentToByte(red);
            byte g = ConvertColorComponentToByte(green);
            byte b = ConvertColorComponentToByte(blue);
           
            return Color.FromArgb(a, r, g, b);
        }

        public static Color FromArgb(byte alpha, byte red, byte green, byte blue)
        {
            return new Color(alpha, red, green, blue);
        }

        public static Color FromGray(double gray)
        {
            return Color.FromGray(ConvertColorComponentToByte(gray));
        }

        public static Color FromGray(int gray, int bits)
        {
            return Color.FromGray(ConvertColorComponentToByte(gray, bits));
        }

        public static Color FromGray(byte gray)
        {
            return new Color(gray, gray, gray);
        }

        public static Color FromCmyk(int cyan, int magenta, int yellow, int black, int bits)
        {
            double max = (1 << bits) - 1;
            double c = cyan / max;
            double m = magenta / max;
            double y = yellow / max;
            double k = black / max;
           
            return Color.FromCmyk(c, m, y, k);
        }

        public static Color FromCmyk(byte cyan, byte magenta, byte yellow, byte black)
        {
            double c = cyan / 255.0;
            double m = magenta / 255.0;
            double y = yellow / 255.0;
            double k = black / 255.0;

            return Color.FromCmyk(c, m, y, k);
        }

        public static Color FromCmyk(double cyan, double magenta, double yellow, double black)
        {
            double c = Color.RestrictComponentToDoubleLimits(cyan);
            double m = Color.RestrictComponentToDoubleLimits(magenta);
            double y = Color.RestrictComponentToDoubleLimits(yellow);
            double k = Color.RestrictComponentToDoubleLimits(black);

            double red =
              (c * (-4.387332384609988 * c + 54.48615194189176 * m +
                    18.82290502165302 * y + 212.25662451639585 * k +
                    -285.2331026137004) +
               m * (1.7149763477362134 * m - 5.6096736904047315 * y +
                    -17.873870861415444 * k - 5.497006427196366) +
               y * (-2.5217340131683033 * y - 21.248923337353073 * k +
                    17.5119270841813) +
               k * (-21.86122147463605 * k - 189.48180835922747) + 255);
            double green =
              (c * (8.841041422036149 * c + 60.118027045597366 * m +
                    6.871425592049007 * y + 31.159100130055922 * k +
                    -79.2970844816548) +
               m * (-15.310361306967817 * m + 17.575251261109482 * y +
                    131.35250912493976 * k - 190.9453302588951) +
               y * (4.444339102852739 * y + 9.8632861493405 * k - 24.86741582555878) +
               k * (-20.737325471181034 * k - 187.80453709719578) + 255);
            double blue =
              (c * (0.8842522430003296 * c + 8.078677503112928 * m +
                    30.89978309703729 * y - 0.23883238689178934 * k +
                    -14.183576799673286) +
               m * (10.49593273432072 * m + 63.02378494754052 * y +
                    50.606957656360734 * k - 112.23884253719248) +
               y * (0.03296041114873217 * y + 115.60384449646641 * k +
                    -193.58209356861505) +
               k * (-22.33816807309886 * k - 180.12613974708367) + 255);

            byte r = Color.RestrictComponentToByteLimits(red);
            byte g = Color.RestrictComponentToByteLimits(green);
            byte b = Color.RestrictComponentToByteLimits(blue);

            return new Color(r, g, b);
        }

        public static double RestrictComponentToDoubleLimits(double component)
        {
            if (component > 1)
            {
                return 1;
            }
            else if (component < 0)
            {
                return 0;
            }
            else
            {
                return component;
            }
        }

        public static Color FromPixel(int pixel)
        {
            byte a, r, g, b;
            GetComponentsFromPixel(pixel, out a, out r, out g, out b);

            return FromArgb(a, r, g, b);
        }

        public static void GetComponentsFromPixel(int pixel, out byte a, out byte r, out byte g, out byte b)
        {
            b = (byte)(pixel & 0xFF);
            g = (byte)((pixel >> 8) & 0xFF);
            r = (byte)((pixel >> 16) & 0xFF);
            a = (byte)((pixel >> 24) & 0xFF);
        }

        public static int GetPixelFromComponents(byte r, byte g, byte b)
        {
            byte a = 255;
            int pixel = b;

            pixel = pixel | (g << 8);
            pixel = pixel | (r << 16);
            pixel = pixel | (a << 24);

            return pixel;
        }

        public static bool operator ==(Color a, Color b)
        {
            return a.Equals(b);
        }

        public static bool operator !=(Color a, Color b)
        {
            return !(a == b);
        }

        public byte GetGrayComponent()
        {
            return this.r;
        }

#if !NETSTANDARD
        public int ToPixel()
        {
#if WPF
            return this.a << 24 | this.r << 16 | this.g << 8 | this.b;
#elif SILVERLIGHT
            var a = this.A;
            var ai = a + 1;
            
            return (a << 24) | ((byte)((this.R * ai) >> 8) << 16) | ((byte)((this.G * ai) >> 8) << 8) | ((byte)((this.B * ai) >> 8));
#endif
        }
#endif

        public override int GetHashCode()
        {
            unchecked
            {
                int result = 17;
                result = (result * 23) + this.a.GetHashCode();
                result = (result * 23) + this.r.GetHashCode();
                result = (result * 23) + this.g.GetHashCode();
                result = (result * 23) + this.b.GetHashCode();
              
                return result;
            }
        }

        public override bool Equals(object obj)
        {
            if (obj == null || !(obj is Color))
            {
                return false;
            }
            return this.Equals((Color)obj);
        }

        public bool Equals(Color value)
        {
            return this.a == value.a &&
                   this.r == value.r &&
                   this.g == value.g &&
                   this.b == value.b;
        }

        private static byte ConvertColorComponentToByte(int component, int bits)
        {
            double max = (1 << bits) - 1;
            
            return ConvertColorComponentToByte(component / max);
        }

        private static byte RestrictComponentToByteLimits(double component)
        {
            if (component > 255)
            {
                return 255;
            }
            else if (component < 0)
            {
                return 0;
            }
            else
            {
                return (byte)component;
            }
        }

        public override string ToString()
        {
            return string.Format("{0:X2}{1:X2}{2:X2}{3:X2}", this.A, this.R, this.G, this.B);
        }
    }
}