﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Imaging
{
    internal abstract class ImageSizeDecoderBase : IImageSizeDecoder
    {
        public static readonly Size DefaultNotDecodedImageSize = Size.Empty;

        protected abstract IEnumerable<string> SupportedExtensions { get; }

        protected virtual int SignatureLength
        {
            get
            {
                return 0;
            }
        }

        protected virtual int BytesToSkipAfterSignature
        {
            get
            {
                return 0;
            }
        }

        public bool CanDecode(string imageTypeExtension)
        {
            Guard.ThrowExceptionIfNull(imageTypeExtension, "imageTypeExtension");

            return this.SupportedExtensions.Any(extension => string.Equals(extension, imageTypeExtension.Trim().Trim('.'), StringComparison.OrdinalIgnoreCase));
        }

        public Size DecodeSize(byte[] bytes)
        {
            if (this.IsValidImage(bytes))
            {
                return this.DecodeSizeInternal(bytes);
            }
            else
            {
                return DefaultNotDecodedImageSize;
            }
        }

        public Size DecodeSize(Stream stream)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");
            Guard.ThrowExceptionIfFalse(stream.CanSeek, "Cannot decode non-seekable stream.");

            using (MemoryStream ms = new MemoryStream())
            {
                stream.Seek(0, SeekOrigin.Begin);
                stream.CopyTo(ms);
                ms.Seek(0, SeekOrigin.Begin);

                if (this.IsValidImage(ms))
                {
                    return this.DecodeSizeInternal(ms);
                }
                else
                {
                    return DefaultNotDecodedImageSize;
                }
            }
        }

        protected abstract Size DecodeSizeInternal(Stream stream);

        protected abstract Size DecodeSizeInternal(byte[] bytes);

        protected virtual bool IsValidImage(Stream stream)
        {
            return true;
        }

        protected virtual bool IsValidImage(byte[] bytes)
        {
            return true;
        }

        protected int GetSizeStartIndexInBytes()
        {
            return this.SignatureLength + this.BytesToSkipAfterSignature;
        }

        protected static bool IsSignatureValid(byte[] bytes, byte[] signature)
        {
            if (bytes.Length >= signature.Length)
            {
                for (int i = 0; i < signature.Length; i++)
                {
                    if (bytes[i] != signature[i])
                    {
                        return false;
                    }
                }
            }

            return true;
        }

        protected static bool IsSignatureValid(Stream stream, byte[] signature)
        {
            stream.Seek(0, SeekOrigin.Begin);

            BinaryReader reader = new BinaryReader(stream);
            byte[] readBytes = reader.ReadBytes(signature.Length);

            return IsSignatureValid(readBytes, signature);
        }

        protected static int ReadLittleEndianInt32(BinaryReader binaryReader)
        {
            byte[] bytes = new byte[sizeof(int)];
            for (int i = 0; i < sizeof(int); i += 1)
            {
                bytes[sizeof(int) - 1 - i] = binaryReader.ReadByte();
            }
            return BitConverter.ToInt32(bytes, 0);
        }

        /// <summary>
        /// Reads the little endian 4 bytes from the given start index.
        /// </summary>
        /// <param name="imageBytes">The image bytes.</param>
        /// <param name="startIndex">The start index.</param>
        /// <returns></returns>
        protected static int ReadLittleEndianInt32(byte[] imageBytes, int startIndex)
        {
            byte[] bytes = new byte[sizeof(int)];
            for (int i = 0; i < sizeof(int); i += 1)
            {
                bytes[sizeof(int) - 1 - i] = imageBytes[i + startIndex];
            }
            return BitConverter.ToInt32(bytes, 0);
        }

        /// <summary>
        /// Reads 2 bytes from the given start index.
        /// </summary>
        /// <param name="imageBytes">The image bytes.</param>
        /// <param name="startIndex">The start index.</param>
        /// <returns></returns>
        protected static int ReadInt16(byte[] imageBytes, int startIndex)
        {
            byte[] bytes = new byte[sizeof(short)];
            for (int i = 0; i < sizeof(short); i++)
            {
                bytes[i] = imageBytes[i + startIndex];
            }
            return BitConverter.ToInt16(bytes, 0);
        }

        /// <summary>
        /// Reads 4 bytes from the given start index.
        /// </summary>
        /// <param name="imageBytes">The image bytes.</param>
        /// <param name="startIndex">The start index.</param>
        /// <returns></returns>
        protected static int ReadInt32(byte[] imageBytes, int startIndex)
        {
            byte[] bytes = new byte[sizeof(int)];
            for (int i = 0; i < sizeof(int); i++)
            {
                bytes[i] = imageBytes[i + startIndex];
            }
            return BitConverter.ToInt32(bytes, 0);
        }
    }
}