﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Markers;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Tables;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg.Decoder
{
    internal class JpegDecoder : JpegDecoderBase
    {       
        public JpegFrame Frame
        {
            get;
            set;
        }

        private bool ShouldScanByHorizontalBlocks
        {
            get
            {
                return this.ScanHeader.NumberOfComponents == 1 && this.FrameHeader.ImageComponents > 1;
            }
        }
        
        public JpegDecoder(byte[] data)
            : base(data)
        {
        }

        public JpegImage Decode()
        {
            if (!base.TryReadJpegStartOfImageMarker())
            {
                throw new InvalidOperationException("JPEG is not valid.");
            }

            JpegMarker marker;
            while ((marker = this.Reader.ReadNextJpegMarker()).MarkerType != JpegMarkerType.SOF)
            {
                marker.InterpretMarker(this);
            }

            this.DecodeFrame((StartOfFrameMarker)marker);

            return this.Rasterize(this.Frame);
        }

        private void DecodeFrame(StartOfFrameMarker startOfFrameMarker)
        {
            startOfFrameMarker.InterpretMarker(this);
            this.ComponentDecoder = startOfFrameMarker.ScanDecoder;

            this.Frame = new JpegFrame(this.FrameHeader.Width, this.FrameHeader.Height);
            JpegMarker marker;

            while ((marker = this.Reader.ReadNextJpegMarker()).MarkerType != JpegMarkerType.EOI)
            {
                if (marker.MarkerType == JpegMarkerType.SOS)
                {
                    this.DecodeScan((StartOfScanMarker)marker);
                }
                else if (marker.MarkerType == JpegMarkerType.SOF)
                {
                    throw new NotSupportedException("Jpeg images with more than one frame are not supported!");
                }
                else
                {
                    marker.InterpretMarker(this);
                }
            }
        }

        private void DecodeScan(StartOfScanMarker startOfScan)
        {
            startOfScan.InterpretMarker(this);

            if (this.Frame.Scan == null)
            {
                JpegComponent[] components = new JpegComponent[this.FrameHeader.ImageComponents];

                for (int i = 0; i < this.FrameHeader.ImageComponents; i++)
                {
                    components[i] = this.CreateJpegComponent(this.ScanHeader.Components[i]);
                }

                this.Frame.Scan = new JpegScan(components);
            }

            JpegScan scan = this.Frame.Scan;
            RestartMarker.RestartDecoder(this);
            RestartMarker.RestartScan(scan);

            this.ApplyHuffmanTablesToComponents();
            int expectedMcuCount = this.CalculateExpectedMcuCount();
            this.CalculateRestartInterval(expectedMcuCount);

            JpegMarker marker = null;
            int mcuIndex = 0;
            do
            {
                while (mcuIndex < expectedMcuCount)
                {
                    for (int n = 0; n < this.RestartInterval && mcuIndex < expectedMcuCount; n++)
                    {
                        this.DecodeBlock(scan, mcuIndex);
                        mcuIndex++;
                    }

                    this.TryPeekMarkerWhileScaning(out marker);

                    if (marker != null && marker.MarkerType == JpegMarkerType.RST)
                    {
                        marker.InterpretMarker(this);
                    }
                    else
                    {
                        RestartMarker.RestartDecoder(this);
                    }

                    RestartMarker.RestartScan(scan);
                }
            }
            while ((marker == null || marker.MarkerType == JpegMarkerType.RST) && mcuIndex < expectedMcuCount);
        }

        private int CalculateExpectedMcuCount()
        {
            if (this.ScanHeader.NumberOfComponents == 1)
            {
                JpegComponent jpegComponent;
                if (this.Frame.Scan.TryGetComponentById(this.ScanHeader.Components[0].ComponentId, out jpegComponent))
                {
                    return jpegComponent.HorizontalBlocksCount * jpegComponent.VerticalBlocksCount;
                }

                throw new InvalidOperationException("Frame scan does not have any components!");
            }
            else
            {
                return this.FrameHeader.BlocksPerLine * this.FrameHeader.BlocksPerColumn;
            }
        }

        private void CalculateRestartInterval(int expectedMcuCount)
        {
            if (!this.IsRestartIntervalDefined)
            {
                this.RestartInterval = expectedMcuCount;
            }

            Guard.ThrowExceptionIfLessThanOrEqual(0, this.RestartInterval, "RestartInterval");
        }

        private void ApplyHuffmanTablesToComponents()
        {
            for (int i = 0; i < this.ScanHeader.Components.Length; i++)
            {
                JpegScanComponent scanComponent = this.ScanHeader.Components[i];

                JpegComponent jpegComponent;
                if (this.Frame.Scan.TryGetComponentById(scanComponent.ComponentId, out jpegComponent))
                {
                    jpegComponent.ACTableId = scanComponent.ACTableId;
                    jpegComponent.DCTableId = scanComponent.DCTableId;
                }
            }
        }

        private bool TryPeekMarkerWhileScaning(out JpegMarker marker)
        {
            marker = null;

            byte first = this.Reader.Peek();
            byte second = this.Reader.Peek(1);

            bool isJpegMarker = JpegMarker.IsJpegMarker(first, second);
            
            if (isJpegMarker)
            {
                bool isEndingWithMarkerIdentifier;
                do
                {
                    isEndingWithMarkerIdentifier = (second == JpegMarker.MarkerIdentifier);

                    if (isEndingWithMarkerIdentifier)
                    {
                        this.Reader.Read();
                        second = this.Reader.Peek(1);
                    }
                } while (isEndingWithMarkerIdentifier);

                marker = JpegMarker.GetMarker(second);

                return true;
            }

            return false;
        }

        private void DecodeBlock(JpegScan scan, int mcuIndex)
        {
            foreach (JpegScanComponent scanComponent in this.ScanHeader.Components)
            {
                JpegComponent jpegComponent;
                if (scan.TryGetComponentById(scanComponent.ComponentId, out jpegComponent))
                {
                    bool isSampleFactoredComponent = (jpegComponent.HSampleFactor > 1 || jpegComponent.VSampleFactor > 1);

                    if(this.ShouldScanByHorizontalBlocks && isSampleFactoredComponent)
                    {
                        int blockIndex = mcuIndex;
                        int blockRow = blockIndex / jpegComponent.HorizontalBlocksCount;
                        int blockColumn = blockIndex % jpegComponent.HorizontalBlocksCount;
                        int mcuRow = blockRow / jpegComponent.VSampleFactor;
                        int mcuColumn = blockColumn / jpegComponent.HSampleFactor;

                        int actualRowIndex = blockRow % jpegComponent.VSampleFactor;
                        int actualColumnIndex = blockColumn % jpegComponent.HSampleFactor;
                        int actualMcuIndex = mcuRow * jpegComponent.McuPerRow + mcuColumn;

                        Block block = jpegComponent.GetBlock(actualMcuIndex, actualRowIndex, actualColumnIndex);
                        this.ComponentDecoder.DecodeBlock(this, jpegComponent, block);
                    }
                    else if (!isSampleFactoredComponent || this.FrameHeader.ImageComponents == 1)
                    {
                        Block block = jpegComponent.GetBlock(mcuIndex);
                        this.ComponentDecoder.DecodeBlock(this, jpegComponent, block);
                    }
                    else
                    {
                        for (int rowIndex = 0; rowIndex < jpegComponent.VSampleFactor; rowIndex++)
                        {
                            for (int columnIndex = 0; columnIndex < jpegComponent.HSampleFactor; columnIndex++)
                            {
                                Block block = jpegComponent.GetBlock(mcuIndex, rowIndex, columnIndex);
                                this.ComponentDecoder.DecodeBlock(this, jpegComponent, block);
                            }
                        }
                    }
                }
            }
        }

        private JpegImage Rasterize(JpegFrame frame)
        {
            byte[][,] raster = new byte[frame.Scan.Components.Length][,];
            Action[] actions = new Action[frame.Scan.Components.Length];

            for (int i = 0; i < frame.Scan.Components.Length; i++)
            {
                int index = i;
                Action action = new Action(() =>
                {
                    JpegComponent component = frame.Scan.Components[index];

                    QuantizationTable table = base.GetQuantisationTable(component.QuantizationTableId);
                    component.Dequantize(table);
                    component.IDCT();
                    // component.DequantizeAndInverse(table);

                    if (frame.Scan.Components.Length == 1)
                    {
                        raster[index] = component.RasterizeNonInterleaved(frame.Scan, frame.Width, frame.Height);
                    }
                    else
                    {
                        raster[index] = component.RasterizeInterleaved(frame.Scan, frame.Width, frame.Height); 
                    }
                });

                actions[i] = action;
            }

            TasksHelper.DoAsync(actions);

            return new JpegImage(frame.Width, frame.Height, this.ColorTransform, raster);
        }

        private JpegComponent CreateJpegComponent(JpegScanComponent scanComponent)
        {
            JpegFrameComponent frameComponent = this.FrameHeader.GetComponent(scanComponent.ComponentId);
            JpegComponent component = new JpegComponent(scanComponent.ComponentId, frameComponent.HSampleFactor, frameComponent.VSampleFactor, frameComponent.QuantizationTableId, this);
            component.InitializeDecoding(this);

            return component;
        }
    }
}
