﻿using System;
using System.IO;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.ColorSpaces;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Markers;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Tables;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg.Decoder
{
    internal abstract class JpegDecoderBase : IHuffmanTablesOwner
    {
        private readonly IJpegReader reader;
        private readonly QuantizationTable[] quantizationTables;
        private readonly HuffmanTable[,] huffmanTables;

        public JpegDecoderBase(byte[] data)
            : this()
        {
            Guard.ThrowExceptionIfNull(data, "data");

            this.reader = new JpegByteArrayReader(data);
        }

        public JpegDecoderBase(Stream stream)
            : this()
        {
            Guard.ThrowExceptionIfNull(stream, "stream");

            this.reader = new JpegStreamReader(stream);
        }

        private JpegDecoderBase()
        {
            this.quantizationTables = new QuantizationTable[Constants.TablesCount];
            this.huffmanTables = new HuffmanTable[2, Constants.TablesCount];
            this.IsRestartIntervalDefined = false;

            this.ColorTransform = JpegColorSpace.Undefined;
        }

        ~JpegDecoderBase()
        {
            this.Reader.FreeResources();
        }

        public IJpegReader Reader
        {
            get
            {
                return this.reader;
            }
        }

        public FrameHeader FrameHeader
        {
            get;
            set;
        }

        public ScanHeader ScanHeader
        {
            get;
            set;
        }

        public ScanDecoder ComponentDecoder
        {
            get;
            set;
        }

        // This property is related to the color transform in the AdobeMarker.
        // For more information see the following link: http://graphicdesign.stackexchange.com/questions/12894/cmyk-jpegs-extracted-from-pdf-appear-inverted/15906#15906
        public JpegColorSpace ColorTransform
        {
            get;
            set;
        }

        public int RestartInterval
        {
            get;
            set;
        }

        public bool IsRestartIntervalDefined
        {
            get;
            set;
        }

        public void AddQuantisationTable(QuantizationTable table)
        {
            Guard.ThrowExceptionIfNull(table, "table");

            this.quantizationTables[table.TableIndex] = table;
        }

        public QuantizationTable GetQuantisationTable(int quantizationTableId)
        {
            return this.quantizationTables[quantizationTableId];
        }

        public void AddHuffmanTable(HuffmanTable table)
        {
            Guard.ThrowExceptionIfNull(table, "table");

            this.huffmanTables[(int)table.TableClass, table.TableIndex] = table;
        }

        public HuffmanTable GetHuffmanTable(TableClass tableClass, int tableIndex)
        {
            return this.huffmanTables[(int)tableClass, tableIndex];
        }

        protected bool TryReadJpegStartOfImageMarker()
        {
            if (this.Reader.Length <= 2)
            {
                return false;
            }

            byte first = this.Reader.Read();

            if (first == JpegMarker.MarkerIdentifier)
            {
                byte second = this.Reader.Read();
                JpegMarker marker = JpegMarker.GetMarker(second);

                return marker.MarkerType == JpegMarkerType.SOI;
            }

            return false;
        }
    }
}
