﻿using System;
using System.Collections.Generic;
using System.IO;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.ColorSpaces;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Markers;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg.Decoder
{
    internal class JpegHeadersDecoder : JpegDecoderBase
    {
        private static readonly HashSet<JpegMarkerType> endMarkerTypes;
        private static readonly HashSet<JpegMarkerType> markerTypesToSkip;

        static JpegHeadersDecoder()
        {
            endMarkerTypes = new HashSet<JpegMarkerType>();
            endMarkerTypes.Add(JpegMarkerType.EOI);
            endMarkerTypes.Add(JpegMarkerType.SOS);

            markerTypesToSkip = new HashSet<JpegMarkerType>();
            markerTypesToSkip.Add(JpegMarkerType.COM);
            markerTypesToSkip.Add(JpegMarkerType.DHT);
            markerTypesToSkip.Add(JpegMarkerType.DQT);
        }

        public JpegHeadersDecoder(byte[] data)
            : base(data)
        {
        }

        public JpegHeadersDecoder(Stream stream)
             : base(stream)
        {
        }

        public bool TryDecodeJpegImage(out JpegImageInfo imageInfo)
        {
            if (!base.TryReadJpegStartOfImageMarker())
            {
                imageInfo = null;
                return false;
            }

            bool readNextMarker = true;

            while (readNextMarker)
            {
                JpegMarker marker = this.Reader.ReadNextJpegMarker();

                if (JpegHeadersDecoder.ShouldStopDecoding(marker))
                {
                    readNextMarker = false;
                }
                else if (!this.TrySkipMarker(marker))
                {
                    marker.InterpretMarker(this);
                }
            }

            imageInfo = this.CreateJpegImageInfo();

            return true;
        }

        private JpegImageInfo CreateJpegImageInfo()
        {
            Guard.ThrowExceptionIfNull(this.FrameHeader, "FrameHeader");

            JpegImageInfo imageInfo = new JpegImageInfo();

            imageInfo.Width = this.FrameHeader.Width;
            imageInfo.Height = this.FrameHeader.Height;
            imageInfo.ColorSpace = this.GetColorSpaceInfo();
            imageInfo.HasAdobeInvertedColors = this.HasAdobeInvertedColors(imageInfo.ColorSpace);

            return imageInfo;
        }

        private bool HasAdobeInvertedColors(JpegColorSpace imageColorSpace)
        {
            bool isCmyk = imageColorSpace == JpegColorSpace.Cmyk;
            bool hasInvertingAdobeMarker = this.ColorTransform != JpegColorSpace.Undefined;

            return isCmyk && hasInvertingAdobeMarker;
        }

        private JpegColorSpace GetColorSpaceInfo()
        {
            switch(this.FrameHeader.ImageComponents)
            {
                case 1:
                    return JpegColorSpace.Grayscale;
                case 3:
                    return JpegColorSpace.Rgb;
                case 4:
                    return JpegColorSpace.Cmyk;
                default:
                    throw new NotSupportedException("Not supported jpeg components count!");
            }
        }

        private bool TrySkipMarker(JpegMarker marker)
        {
            if (JpegHeadersDecoder.markerTypesToSkip.Contains(marker.MarkerType))
            {
                JpegMarker.ReadLengthAndSkipData(this);

                return true;
            }

            return false;
        }

        private static bool ShouldStopDecoding(JpegMarker marker)
        {
            return JpegHeadersDecoder.endMarkerTypes.Contains(marker.MarkerType);
        }
    }
}
