﻿using System;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Tables;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg.Decoder
{
    internal class ProgressiveDCTDecoder : ScanDecoder
    {
        private int eobrun;
        private int successiveACState;
        private int successiveACNextValue;
        private Action<JpegDecoder, JpegComponent, Block> decodeBlockAction;

        public override void ResetDecoder(JpegDecoderBase decoder)
        {
            this.eobrun = 0;
            this.successiveACState = 0;
            this.successiveACNextValue = 0;

            if (decoder.ScanHeader.SpectralSelectionStart == 0)
            {
                if (decoder.ScanHeader.BitPositionHigh == 0)
                {
                    this.decodeBlockAction = this.DecodeDCFirst;
                }
                else
                {
                    this.decodeBlockAction = this.DecodeDCSuccessive;
                }
            }
            else
            {
                if (decoder.ScanHeader.BitPositionHigh == 0)
                {
                    this.decodeBlockAction = this.DecodeACFirst;
                }
                else
                {
                    this.decodeBlockAction = this.DecodeACSuccessive;
                }
            }
        }

        public override void DecodeBlock(JpegDecoder decoder, JpegComponent component, Block block)
        {
            Guard.ThrowExceptionIfNull(this.decodeBlockAction, "decodeBlockAction");

            this.decodeBlockAction(decoder, component, block);
        }

        private void DecodeDCFirst(JpegDecoder decoder, JpegComponent component, Block block)
        {
            int t = component.DcTable.Decode(decoder.Reader);

            int diff = t == 0 ? 0 : (decoder.Reader.ReceiveAndExtend(t) << decoder.ScanHeader.BitPositionLow);
            block[0] = component.LastDCCoefficient + diff;
            component.UpdateLastDCCoefficient(block);
        }

        private void DecodeDCSuccessive(JpegDecoder decoder, JpegComponent component, Block block)
        {
            block[0] |= (decoder.Reader.ReadBit() << decoder.ScanHeader.BitPositionLow);
        }

        private void DecodeACFirst(JpegDecoder decoder, JpegComponent component, Block block)
        {

            if (this.eobrun > 0)
            {
                this.eobrun--;
                return;
            }

            int k = decoder.ScanHeader.SpectralSelectionStart;
            int e = decoder.ScanHeader.SpectralSelectionEnd;

            while (k <= e)
            {
                int rs = component.AcTable.Decode(decoder.Reader);
                int s = rs & 15;
                int r = rs >> 4;

                if (s == 0)
                {
                    if (r < 15)
                    {
                        this.eobrun = decoder.Reader.Receive(r) + (1 << r) - 1;
                        break;
                    }

                    k += 16;
                    continue;
                }
                
                k += r;

                int z = ZigZagScan.ZigZagMap[k];
                block[z] = decoder.Reader.ReceiveAndExtend(s) * (1 << decoder.ScanHeader.BitPositionLow);

                k++;
            }
        }

        private void DecodeACSuccessive(JpegDecoder decoder, JpegComponent component, Block block)
        {
            int k = decoder.ScanHeader.SpectralSelectionStart;
            int e = decoder.ScanHeader.SpectralSelectionEnd;
            int r = 0;
            
            while (k <= e)
            {
                int z = ZigZagScan.ZigZagMap[k];

                switch(this.successiveACState)
                {
                    case 0: // initial state
                        int rs = component.AcTable.Decode(decoder.Reader);
                        int s = rs & 15;
                        r = rs >> 4;

                        if (s == 0)
                        {
                            if (r < 15)
                            {
                                eobrun = decoder.Reader.Receive(r) + (1 << r);
                                this.successiveACState = 4;
                            }
                            else
                            {
                                r = 16;
                                this.successiveACState = 1;
                            }
                        }
                        else
                        {
                            if (s != 1)
                            {
                                throw new InvalidOperationException("invalid ACn encoding");
                            }

                            this.successiveACNextValue = decoder.Reader.ReceiveAndExtend(s);
                            this.successiveACState = r != 0 ? 2 : 3;
                        }
                        continue;
                    case 1: //skipping r zero items
                    case 2:
                        if (block[z] != 0)
                        {
                            block[z] += (decoder.Reader.ReadBit() << decoder.ScanHeader.BitPositionLow);
                        }
                        else
                        {
                            r--;
                            if (r == 0)
                            {
                                this.successiveACState = (successiveACState == 2 ? 3 : 0);
                            }
                        }
                        break;
                    case 3: // set value for a zero item
                        if (block[z] != 0)
                        {
                            block[z] += (decoder.Reader.ReadBit() << decoder.ScanHeader.BitPositionLow);
                        }
                        else
                        {
                            block[z] = (this.successiveACNextValue << decoder.ScanHeader.BitPositionLow);
                            this.successiveACState = 0;
                        }
                        break;
                    case 4: // eob
                        if (block[z] != 0)
                        {
                            block[z] += (decoder.Reader.ReadBit() << decoder.ScanHeader.BitPositionLow);
                        }
                        break;
                }

                k++;
            }

            if (this.successiveACState == 4)
            {
                this.eobrun--;

                if (this.eobrun == 0)
                {
                    this.successiveACState = 0;
                }
            }
        }
    }
}
