﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Tables;

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg
{
    internal static class DiscreteCosineTransform
    {
        #region Fields

        // IDCT coefficients.
        private static readonly double[,] coefficients;
        private static readonly double[,] transposedCoefficients;

        #endregion


        #region Constructors

        static DiscreteCosineTransform()
        {
            coefficients = BuildCoefficients();
            transposedCoefficients = BuildTransposedCoefitients(coefficients);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Implements Fast FDCT.
        /// </summary>
        /// <param name="input">Input block.</param>
        /// <returns>Output block.</returns>
        public static FloatBlock ForwardDCT(Block input)
        {
            FloatBlock output = new FloatBlock(input);

            float tmp0, tmp1, tmp2, tmp3, tmp4, tmp5, tmp6, tmp7;
            float tmp10, tmp11, tmp12, tmp13;
            float z1, z2, z3, z4, z5, z11, z13;

            for (int row = 0; row < Constants.BlockHeight; row++)
            {
                for (int column = 0; column < Constants.BlockWidth; column++)
                {
                    output[row, column] = output[row, column] - 128f;
                }
            }

            // Pass 1: process rows.
            for (int row = 0; row < 8; row++)
            {
                tmp0 = output[row, 0] + output[row, 7];
                tmp7 = output[row, 0] - output[row, 7];
                tmp1 = output[row, 1] + output[row, 6];
                tmp6 = output[row, 1] - output[row, 6];
                tmp2 = output[row, 2] + output[row, 5];
                tmp5 = output[row, 2] - output[row, 5];
                tmp3 = output[row, 3] + output[row, 4];
                tmp4 = output[row, 3] - output[row, 4];

                // Even part
                tmp10 = tmp0 + tmp3;
                tmp13 = tmp0 - tmp3;
                tmp11 = tmp1 + tmp2;
                tmp12 = tmp1 - tmp2;

                output[row, 0] = tmp10 + tmp11;
                output[row, 4] = tmp10 - tmp11;

                z1 = (tmp12 + tmp13) * (float)0.707106781;
                output[row, 2] = tmp13 + z1;
                output[row, 6] = tmp13 - z1;

                // Odd part
                tmp10 = tmp4 + tmp5;
                tmp11 = tmp5 + tmp6;
                tmp12 = tmp6 + tmp7;

                // The rotator is modified from fig 4-8 to avoid extra negations.
                z5 = (tmp10 - tmp12) * (float)0.382683433;
                z2 = ((float)0.541196100) * tmp10 + z5;
                z4 = ((float)1.306562965) * tmp12 + z5;
                z3 = tmp11 * ((float)0.707106781);

                z11 = tmp7 + z3;
                z13 = tmp7 - z3;

                output[row, 5] = z13 + z2;
                output[row, 3] = z13 - z2;
                output[row, 1] = z11 + z4;
                output[row, 7] = z11 - z4;
            }

            // Pass 2: process columns
            for (int column = 0; column < 8; column++)
            {
                tmp0 = output[0, column] + output[7, column];
                tmp7 = output[0, column] - output[7, column];
                tmp1 = output[1, column] + output[6, column];
                tmp6 = output[1, column] - output[6, column];
                tmp2 = output[2, column] + output[5, column];
                tmp5 = output[2, column] - output[5, column];
                tmp3 = output[3, column] + output[4, column];
                tmp4 = output[3, column] - output[4, column];

                // Even part
                tmp10 = tmp0 + tmp3;
                tmp13 = tmp0 - tmp3;
                tmp11 = tmp1 + tmp2;
                tmp12 = tmp1 - tmp2;

                output[0, column] = tmp10 + tmp11;
                output[4, column] = tmp10 - tmp11;

                z1 = (tmp12 + tmp13) * 0.707106781f;
                output[2, column] = tmp13 + z1;
                output[6, column] = tmp13 - z1;

                // Odd part
                tmp10 = tmp4 + tmp5;
                tmp11 = tmp5 + tmp6;
                tmp12 = tmp6 + tmp7;

                // The rotator is modified from fig 4-8 to avoid extra negations.
                z5 = (tmp10 - tmp12) * 0.382683433f;
                z2 = 0.541196100f * tmp10 + z5;
                z4 = 1.306562965f * tmp12 + z5;
                z3 = tmp11 * 0.707106781f;

                z11 = tmp7 + z3;
                z13 = tmp7 - z3;

                output[5, column] = z13 + z2;
                output[3, column] = z13 - z2;
                output[1, column] = z11 + z4;
                output[7, column] = z11 - z4;
            }

            return output;
        }

        public static Block InverseDCT(Block input)
        {
            Block output = new Block();
            double[] tmp = new double[Constants.BlockWidth * Constants.BlockHeight];

            double sum;
            int idx = 0;
            for (int i = 0; i < Constants.BlockWidth; i++)
            {
                for (int j = 0; j < Constants.BlockHeight; j++)
                {
                    sum = 0;

                    for (int k = 0; k < Constants.BlockWidth; k++)
                    {
                        sum += input[i * Constants.BlockHeight + k] * coefficients[k, j];
                    }

                    tmp[idx++] = sum;
                }
            }

            for (int i = 0; i < Constants.BlockWidth; i++)
            {
                for (int j = 0; j < Constants.BlockHeight; j++)
                {
                    sum = 128f;

                    for (int k = 0; k < Constants.BlockWidth; k++)
                    {
                        sum += transposedCoefficients[i, k] * tmp[k * Constants.BlockHeight + j];
                    }

                    output[i, j] = Round(sum);
                }
            }

            return output;
        }

        private static byte Round(double f)
        {
            if (f < 0)
            {
                return 0;
            }
            else if (f > 255)
            {
                return 255;
            }
            else
            {
                return (byte)(f + 0.5);
            }
        }

        private static double[,] BuildCoefficients()
        {
            double[,] c = new double[Constants.BlockWidth, Constants.BlockHeight];

            for (int i = 0; i < Constants.BlockWidth; i++)
            {
                for (int j = 0; j < Constants.BlockHeight; j++)
                {
                    c[i, j] = i == 0 ? 0.353553391d : 
                        (0.5 * Math.Cos(((2.0 * j + 1) * i * Math.PI) / 16.0));
                }
            }

            return c;
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1814:PreferJaggedArraysOverMultidimensional", MessageId = "0#"), System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1814:PreferJaggedArraysOverMultidimensional", MessageId = "Body"), System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1814:PreferJaggedArraysOverMultidimensional", MessageId = "Return")]
        private static double[,] BuildTransposedCoefitients(double[,] coefitients)
        {
            double[,] result = new double[Constants.BlockWidth, Constants.BlockHeight];

            for (int i = 0; i < Constants.BlockWidth; i++)
            {
                for (int j = 0; j < Constants.BlockHeight; j++)
                {
                    result[j, i] = coefitients[i, j];
                }
            }

            return result;
        }

        #endregion
    }
}
