﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Tables;

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg.Encoder
{
    internal class BaselineDCTEncoder : ScanEncoder
    {
        private const byte LuminanceTableIdx = 0;
        private const byte ChrominanceTableIdx = 1;

        public override void EncodeBlock(JpegEncoder encoder, JpegComponent component, Block block)
        {
            int k, i;

            // The DC portion.
            // The DC coefficients are coded differentially, using a one-dimensional predictor, 
            // PRED, which is the quantized DC value from the most recently coded 8 × 8 block from the same component. 
            int diff = block[0] - component.LastDCCoefficient;
            int temp2 = diff;

            if (diff < 0)
            {
                diff = -diff;
                temp2--;
            }

            int nbits = 0;
            while (diff != 0)
            {
                nbits++;
                diff >>= 1;
            }

            component.DcTable.Encode(encoder.Writer, nbits);
            if (nbits != 0)
            {
                encoder.Writer.WriteBits(nbits, temp2);
            }

            // The AC portion

            int coefficientPosition = 0; //  position of the coefficient in ZZ relative to the previous non-zero coefficient (i.e. the run-length of zero coefficients between non-zero coefficients). 

            for (k = 1; k < Constants.BlockSize; k++)
            {
                diff = block[k];
                if (diff == 0)
                {
                    coefficientPosition++;
                }
                else
                {
                    while (coefficientPosition > 15)
                    {
                        // Since the run length of zero coefficients may exceed 15, the value ’RRRRSSSS’ = X’F0’ 
                        // is defined to represent a run length of 15 zero coefficients followed by a coefficient of zero amplitude. 
                        // (This can be interpreted as a run length of 16 zero coefficients.)
                        component.AcTable.Encode(encoder.Writer, 0xF0);
                        coefficientPosition -= 16;
                    }
                    
                    temp2 = diff;
                    if (diff < 0)
                    {
                        diff = -diff;
                        temp2--;
                    }

                    nbits = 1;
                    while ((diff >>= 1) != 0)
                    {
                        nbits++;
                    }

                    i = (coefficientPosition << 4) + nbits;
                    component.AcTable.Encode(encoder.Writer, i);
                    encoder.Writer.WriteBits(nbits, temp2);

                    coefficientPosition = 0;
                }
            }

            if (coefficientPosition > 0)
            {
                component.AcTable.Encode(encoder.Writer, 0);
            }
        }

        /// <summary>
        /// Prepare Huffman tables.
        /// </summary>
        /// <param name="jpegEncoder">Encoder to add table to.</param>
        public override void PrepareHuffmanTables(JpegEncoder jpegEncoder)
        {
            HuffmanTable ydcTable = new HuffmanTable(TableClass.DCTable, 0, JpegEncoder.StandardDCLuminanceLengths, JpegEncoder.StandardDCLuminanceValues);
            jpegEncoder.AddHuffmanTable(ydcTable);

            HuffmanTable yacTable = new HuffmanTable(TableClass.ACTable, 0, JpegEncoder.StandardACLuminanceLengths, JpegEncoder.StandardACLuminanceValues);
            jpegEncoder.AddHuffmanTable(yacTable);

            HuffmanTable cbdcTable = new HuffmanTable(TableClass.DCTable, 1, JpegEncoder.StandardDCChromianceLengths, JpegEncoder.StandardDCChromianceValues);
            jpegEncoder.AddHuffmanTable(cbdcTable);

            HuffmanTable cbacTable = new HuffmanTable(TableClass.ACTable, 1, JpegEncoder.StandardACChromianceLengths, JpegEncoder.StandardACChromianceValues);
            jpegEncoder.AddHuffmanTable(cbacTable);
        }

        /// <summary>
        /// Prepare base quantization tables.
        /// </summary>
        /// <param name="encoder">Encoder to add table to.</param>
        public override void PrepareQuantizationTables(JpegEncoder encoder)
        {
            byte[] luminanceTable = BaselineDCTEncoder.ScaleQuantizationTable(
                encoder.Parameters.LuminanceTable,
                encoder.Parameters.QuantizingQuality);
            encoder.AddQuantizationTable(new QuantizationTable(LuminanceTableIdx, luminanceTable));

            byte[] chrominanceTable = BaselineDCTEncoder.ScaleQuantizationTable(
                encoder.Parameters.ChrominanceTable,
                encoder.Parameters.QuantizingQuality);
            encoder.AddQuantizationTable(new QuantizationTable(ChrominanceTableIdx, chrominanceTable));
        }

        public override void ResetEncoder(JpegEncoder encoder)
        {
            throw new NotImplementedException();
        }


        private static byte[] ScaleQuantizationTable(byte[] inTable, float qualityScale)
        {
            byte[] outTable = new byte[64];
            long temp;

            if (qualityScale <= 0)
            {
                qualityScale = 1;
            }
            else if (qualityScale > 100)
            {
                qualityScale = 100f;
            }
            else if (qualityScale < 50)
            {
                qualityScale = 5000f / qualityScale;
            }
            else
            {
                qualityScale = 200f - qualityScale * 2.0f;
            }

            for (int i = 0; i < 64; i++)
            {
                temp = (long)(inTable[i] * qualityScale / 100.0f);
                temp = Math.Max(temp, 1);
                temp = Math.Min(temp, 255);
                outTable[i] = (byte)temp;
            }
            return outTable;
        }
    }
}
