﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Markers;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Tables;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg.Encoder
{
    internal class JpegEncoder : IDisposable, IHuffmanTablesOwner
    {
        internal static readonly byte[] DefaultChromianceQuantizationTable = 
        {
            17,  18,  24,  47,  99,  99,  99,  99,
            18,  21,  26,  66,  99,  99,  99,  99,
            24,  26,  56,  99,  99,  99,  99,  99,
            47,  66,  99,  99,  99,  99,  99,  99,
            99,  99,  99,  99,  99,  99,  99,  99,
            99,  99,  99,  99,  99,  99,  99,  99,
            99,  99,  99,  99,  99,  99,  99,  99,
            99,  99,  99,  99,  99,  99,  99,  99
        };

        internal static readonly byte[] DefaultLuminanceQuantizationTable = 
        {
            16,  11,  10,  16,  24,  40,  51,  61,
            12,  12,  14,  19,  26,  58,  60,  55,
            14,  13,  16,  24,  40,  57,  69,  56,
            14,  17,  22,  29,  51,  87,  80,  62,
            18,  22,  37,  56,  68, 109, 103,  77,
            24,  35,  55,  64,  81, 104, 113,  92,
            49,  64,  78,  87, 103, 121, 120, 101,
            72,  92,  95,  98, 112, 100, 103,  99
        };

        internal static readonly byte[] StandardDCLuminanceLengths = { 0, 1, 5, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0 };
        internal static readonly byte[] StandardDCLuminanceValues = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };

        internal static readonly byte[] StandardDCChromianceLengths = { 0, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0 };
        internal static readonly byte[] StandardDCChromianceValues = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };


        internal static readonly byte[] StandardACLuminanceLengths = { 0, 2, 1, 3, 3, 2, 4, 3, 5, 5, 4, 4, 0, 0, 1, 0x7d };
        internal static readonly byte[] StandardACLuminanceValues = 
        {
            0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12,
            0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07,
            0x22, 0x71, 0x14, 0x32, 0x81, 0x91, 0xa1, 0x08,
            0x23, 0x42, 0xb1, 0xc1, 0x15, 0x52, 0xd1, 0xf0,
            0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0a, 0x16,
            0x17, 0x18, 0x19, 0x1a, 0x25, 0x26, 0x27, 0x28,
            0x29, 0x2a, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
            0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49,
            0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59,
            0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
            0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
            0x7a, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
            0x8a, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
            0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
            0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
            0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5,
            0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4,
            0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xe1, 0xe2,
            0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea,
            0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
            0xf9, 0xfa
        };

        internal static readonly byte[] StandardACChromianceLengths = { 0, 2, 1, 2, 4, 4, 3, 4, 7, 5, 4, 4, 0, 1, 2, 0x77 };
        internal static readonly byte[] StandardACChromianceValues =
        {
            0x00, 0x01, 0x02, 0x03, 0x11, 0x04, 0x05, 0x21,
            0x31, 0x06, 0x12, 0x41, 0x51, 0x07, 0x61, 0x71,
            0x13, 0x22, 0x32, 0x81, 0x08, 0x14, 0x42, 0x91,
            0xa1, 0xb1, 0xc1, 0x09, 0x23, 0x33, 0x52, 0xf0,
            0x15, 0x62, 0x72, 0xd1, 0x0a, 0x16, 0x24, 0x34,
            0xe1, 0x25, 0xf1, 0x17, 0x18, 0x19, 0x1a, 0x26,
            0x27, 0x28, 0x29, 0x2a, 0x35, 0x36, 0x37, 0x38,
            0x39, 0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48,
            0x49, 0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58,
            0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68,
            0x69, 0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78,
            0x79, 0x7a, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
            0x88, 0x89, 0x8a, 0x92, 0x93, 0x94, 0x95, 0x96,
            0x97, 0x98, 0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5,
            0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4,
            0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3,
            0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2,
            0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda,
            0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
            0xea, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
            0xf9, 0xfa
        };

        private static readonly byte[] ComponentId = { 1, 2, 3, 4 };
        private static readonly byte[] HSamplingFactor = { 1, 1, 1, 1 };
        private static readonly byte[] VSamplingFactor = { 1, 1, 1, 1 };
        private static readonly byte[] QTableNumber = { 0, 1, 1, 1 };
        private static readonly byte[] DCTableNumber = { 0, 1, 1, 1 };
        private static readonly byte[] ACTableNumber = { 0, 1, 1, 1 };

        private readonly JpegImage jpegImage;
        private readonly JpegWriter writer;
        private readonly JpegEncoderParameters encoderParameters;

        private QuantizationTable[] quantizationTables;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1814:PreferJaggedArraysOverMultidimensional", MessageId = "Member")]
        private HuffmanTable[,] huffmanTables;

        /// <summary>
        /// Initializes a new instance of the JpegEncoder class.
        /// </summary>
        /// <param name="jpegImage">JPEG image to be encoded.</param>
        /// <param name="encoderParameters">JPEG encoder parameters.</param>
        public JpegEncoder(JpegImage jpegImage, JpegEncoderParameters encoderParameters)
            : this(encoderParameters)
        {
            this.jpegImage = jpegImage;
        }

        private JpegEncoder(JpegEncoderParameters encoderParameters)
        {
            this.encoderParameters = encoderParameters;
            this.writer = new JpegWriter();
            this.InitializeTables();
        }

        public ScanEncoder ComponentEncoder
        {
            get;
            set;
        }

        public JpegFrame Frame
        {
            get;
            set;
        }

        public FrameHeader FrameHeader
        {
            get;
            set;
        }

        /// <summary>
        /// Gets number of lines (height). 
        /// </summary>
        public int Height
        {
            get
            {
                return this.jpegImage.Height;
            }
        }

        public IEnumerable<HuffmanTable> HuffmanTables
        {
            get
            {
                foreach (HuffmanTable table in this.huffmanTables)
                {
                    if (table != null)
                    {
                        yield return table;
                    }
                }
            }
        }

        /// <summary>
        /// Gets encoder parameters.
        /// </summary>
        public JpegEncoderParameters Parameters
        {
            get
            {
                return this.encoderParameters;
            }
        }

        public IEnumerable<QuantizationTable> QuantizationTables
        {
            get
            {
                foreach (QuantizationTable table in this.quantizationTables)
                {
                    if (table != null)
                    {
                        yield return table;
                    }
                }
            }
        }

        public ushort RestartInterval
        {
            get;
            set;
        }

        public ScanHeader ScanHeader
        {
            get;
            set;
        }

        /// <summary>
        /// Gets number of the samples per line (width).
        /// </summary>
        public int Width
        {
            get
            {
                return this.jpegImage.Width;
            }
        }

        public JpegWriter Writer
        {
            get
            {
                return this.writer;
            }
        }

        public void AddQuantizationTable(QuantizationTable table)
        {
            Guard.ThrowExceptionIfNull(table, "table");
            this.quantizationTables[table.TableIndex] = table;
        }

        public void AddHuffmanTable(HuffmanTable table)
        {
            Guard.ThrowExceptionIfNull(table, "table");
            this.huffmanTables[(int)table.TableClass, table.TableIndex] = table;
        }

        public void Dispose()
        {
            this.writer.Dispose();
        }

        public byte[] Encode()
        {
            // Write SOI marker.
            JpegMarker marker = JpegMarker.GetMarker(JpegMarkerType.SOI);
            marker.WriteMarker(this);

            marker = JpegMarker.GetMarker(JpegMarkerType.APP0);
            marker.WriteMarker(this);

            StartOfFrameMarker sofMarker = JpegMarker.GetSOFMarkerForEncoding(this.Parameters.EncodingType);
            this.ComponentEncoder = sofMarker.ScanEncoder;
            
            // Prepare quantization tables.
            sofMarker.ScanEncoder.PrepareQuantizationTables(this);

            // Write quantization tables.
            marker = JpegMarker.GetMarker(JpegMarkerType.DQT);
            marker.WriteMarker(this);

            // Write frame header.
            this.FrameHeader = new FrameHeader()
            {
                Height = this.Height,
                SamplePrecision = (byte)this.encoderParameters.SamplePrecision,                
                Width = this.Width
            };
            
            for (int i = 0; i < this.jpegImage.NumberOfComponents; i++)
            {
                JpegFrameComponent component = new JpegFrameComponent(
                    JpegEncoder.ComponentId[i],
                    JpegEncoder.HSamplingFactor[i],
                    JpegEncoder.VSamplingFactor[i],
                    JpegEncoder.QTableNumber[i]);
                this.FrameHeader.AddComponent(component);
            }

            sofMarker.WriteMarker(this);

            // Prepare Huffman tables
            sofMarker.ScanEncoder.PrepareHuffmanTables(this);

            // Write Huffman tables
            marker = JpegMarker.GetMarker(JpegMarkerType.DHT);
            marker.WriteMarker(this);

            // Write scan header.
            this.ScanHeader = new ScanHeader()
            {
                BitPositionHigh = 0,
                BitPositionLow = 0,
                NumberOfComponents = this.jpegImage.NumberOfComponents,
                SpectralSelectionStart = 0,
                SpectralSelectionEnd = 63
            };

            for (int i = 0; i < this.jpegImage.NumberOfComponents; i++)
            {
                JpegScanComponent component = new JpegScanComponent(
                    JpegEncoder.ComponentId[i],
                    JpegEncoder.DCTableNumber[i],
                    JpegEncoder.ACTableNumber[i]);
                this.ScanHeader.AddComponent(i, component);
            }

            marker = JpegMarker.GetMarker(JpegMarkerType.SOS);
            marker.WriteMarker(this);

            // Encode scan
            this.EncodeScan();

            // Write EOI marker
            marker = JpegMarker.GetMarker(JpegMarkerType.EOI);
            marker.WriteMarker(this);

            return this.writer.Data;
        }

        private void EncodeBlocks(JpegScan scan)
        {
            foreach(JpegComponent component in scan.Components)
            {
                // for DCT-based processes, set the DC prediction (PRED) to zero for all components in the scan 
                component.Restart();
            }

            for (int row = 0; row < scan.MinVerticalBlocksCount; row++)
            {
                int blockStartYPos = row * Constants.BlockHeight;

                for (int column = 0; column < scan.MinHorizontalBlocksCount; column++)
                {
                    int blockStartXPos = column * Constants.BlockWidth;

                    for (int i = 0; i < scan.Components.Length; i++)
                    {
                        JpegComponent component = scan.Components[i];
                        List<Block> blocks = component.PrepareBlocks(this.jpegImage.Raster, blockStartYPos, blockStartXPos);
                        List<FloatBlock> encodingBlocks = JpegComponent.FDCT(blocks);
                
                        QuantizationTable table = this.quantizationTables[component.QuantizationTableId];
                        List<Block> quantizedBlocks = JpegComponent.Quantize(encodingBlocks, table);
                        component.EncodeNextBlocks(this, quantizedBlocks);
                    }
                }
            }

            this.Writer.Flush();
        }

        private void EncodeScan()
        {
            JpegComponent[] components = new JpegComponent[this.FrameHeader.ImageComponents];

            for (int i = 0; i < this.FrameHeader.ImageComponents; i++)
            {
                components[i] = this.GetJpegComponent(this.ScanHeader.Components[i]);
            }

            JpegScan scan = new JpegScan(components);
            scan.InitializeEncoding(this);

            this.EncodeBlocks(scan);
        }

        private JpegComponent GetJpegComponent(JpegScanComponent scanComponent)
        {
            JpegFrameComponent frameComponent = this.FrameHeader.GetComponent(scanComponent.ComponentId);
            JpegComponent component = new JpegComponent(
                scanComponent.ComponentId, 
                frameComponent.HSampleFactor, 
                frameComponent.VSampleFactor, 
                frameComponent.QuantizationTableId,
                this);

            component.ACTableId = scanComponent.ACTableId;
            component.DCTableId = scanComponent.DCTableId;

            return component;
        }

        public HuffmanTable GetHuffmanTable(TableClass tableClass, int tableIndex)
        {
            return this.huffmanTables[(int)tableClass, tableIndex];
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1814:PreferJaggedArraysOverMultidimensional", MessageId = "Body")]
        private void InitializeTables()
        {
            this.quantizationTables = new QuantizationTable[Constants.TablesCount];
            this.huffmanTables = new HuffmanTable[2, Constants.TablesCount];
        }
    }
}
