﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg.Encoder
{
    /// <summary>
    /// Represents parameters of the JPEG encoder.
    /// </summary>
    public class JpegEncoderParameters : INotifyPropertyChanged
    {
        private byte[] chromianceTable = JpegEncoder.DefaultChromianceQuantizationTable;
        private byte[] luminanceTable = JpegEncoder.DefaultLuminanceQuantizationTable;
        private float quantizingQuality = 100.0f;
        private JpegEncodingType encodingType = JpegEncodingType.BaselineDct;
        private int samplePrecision = 8;

        /// <summary>
        /// Initializes a new instance of the JpegEncoderParameters class.
        /// </summary>
        public JpegEncoderParameters()
        {
            this.ChrominanceTable = new byte[JpegEncoder.DefaultChromianceQuantizationTable.Length];
            JpegEncoder.DefaultChromianceQuantizationTable.CopyTo(this.ChrominanceTable, 0);

            this.LuminanceTable = new byte[JpegEncoder.DefaultLuminanceQuantizationTable.Length];
            JpegEncoder.DefaultLuminanceQuantizationTable.CopyTo(this.LuminanceTable, 0);
        }

        /// <summary>
        /// Occurs when a property value changes.
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        /// <summary>
        /// Gets or sets a 64 byte array which corresponds to a JPEG Chrominance Quantization table.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1819:PropertiesShouldNotReturnArrays")]
        public byte[] ChrominanceTable
        {
            get 
            { 
                return this.chromianceTable; 
            }

            set 
            { 
                this.chromianceTable = value;
                this.OnPropertyChanged("ChrominanceTable");
            }
        }

        /// <summary>
        /// Gets or sets type of the JPEG encoder.
        /// </summary>
        public JpegEncodingType EncodingType
        {
            get
            {
                return this.encodingType;
            }

            set
            {
                this.encodingType = value;
                this.OnPropertyChanged("EncodingType");
            }
        }

        /// <summary>
        /// Gets or sets a 64 byte array which corresponds to a JPEG Luminance Quantization table.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1819:PropertiesShouldNotReturnArrays")]
        public byte[] LuminanceTable
        {
            get 
            { 
                return this.luminanceTable; 
            }

            set 
            { 
                this.luminanceTable = value;
                this.OnPropertyChanged("LuminanceTable");
            }
        }

        /// <summary>
        /// Gets or sets quantizing quality.
        /// </summary>
        public float QuantizingQuality
        {
            get
            {
                return this.quantizingQuality;
            }

            set
            {
                this.quantizingQuality = value;
                this.OnPropertyChanged("QuantizingQuality");
            }
        }

        /// <summary>
        /// Gets or sets the precision in bits for the samples of the components in the frame.
        /// </summary>
        /// <remarks>CCITT T.81, Annex B, section B.2.2 Frame header syntax (see P parameter in the Table B.2).</remarks>
        public int SamplePrecision
        {
            get
            {
                return this.samplePrecision;
            }

            set
            {
                this.samplePrecision = value;
                this.OnPropertyChanged("SamplePrecision");
            }
        }

        private void OnPropertyChanged(string propertyName)
        {
            if (this.PropertyChanged != null)
            {
                this.PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
            }
        }
    }
}
