﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Markers;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Tables;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Utils;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg.Encoder
{
    internal class JpegWriter : WriterBase
    {
        private readonly BitsWriter bitsWriter;
        private readonly BitsWriter bufferBitsWriter;

        /// <summary>
        /// Initializes a new instance of the JpegWriter class.
        /// </summary>
        public JpegWriter()
            : base()
        {
            this.bitsWriter = new BitsWriter();
            this.bufferBitsWriter = new BitsWriter();
        }

        public void Flush()
        {
            if (!this.bufferBitsWriter.IsFull && !this.bufferBitsWriter.IsEmpty)
            {
                this.bufferBitsWriter.WriteBits((byte)0, this.bufferBitsWriter.BitsLeft);
            }

            this.WriteJpegByte(this.bufferBitsWriter.Data);
            this.bufferBitsWriter.Clear();
        }

        /// <summary>
        /// Write 4 bits from the given byte value.
        /// </summary>
        /// <param name="value">Byte to get bits from.</param>
        public void Write4(byte value)
        {
            this.bitsWriter.WriteBits(value, 4);

            if (this.bitsWriter.IsFull)
            {
                base.Write(this.bitsWriter.Data);
                this.bitsWriter.Clear();
            }
        }

        /// <summary>
        /// Write byte,
        /// </summary>
        /// <param name="value">Byte to write.</param>
        public void Write8(byte value)
        {
            base.Write(value);
        }

        /// <summary>
        /// Write usignded short value.
        /// </summary>
        /// <param name="value">Value to write.</param>
        /// <returns></returns>
        public void Write16(ushort value)
        {
            byte[] bytes = BitConverter.GetBytes(value);
            base.WriteBE(bytes, 2);
        }

        /// <summary>
        /// Write JPEG information structures.
        /// </summary>
        /// <typeparam name="T">Table type.</typeparam>
        /// <param name="tables">Tables to write.</param>
        /// <param name="extraBytesCount">Numbe of the extra bytes whould be added to the length of the table list.</param>
        public void WriteJpegTables<T>(IEnumerable<T> tables, ushort extraBytesCount) where T : JpegTable
        {
            ushort totalLength = (ushort)(tables.Sum(t => t.Length) + extraBytesCount);
            this.Write16(totalLength);

            foreach (T table in tables)
            {
                table.Write(this);
            }
        }

        /// <summary>
        /// Writes a bits.
        /// </summary>
        /// <param name="n">Number of bits.</param>
        /// <param name="bits">Value to get bits from.</param>
        public void WriteBits(int n, int bits)
        {
            if (n > 0)
            {
                if (this.bufferBitsWriter.BitsLeft >= n)
                {
                    this.bufferBitsWriter.WriteBits((byte)bits, n);
                }
                else
                {
                    bits <<= sizeof(int) * 8 - n;

                    while (n > 0)
                    {
                        int writeCount = Math.Min(n, this.bufferBitsWriter.BitsLeft);
                        byte writeBits = (byte)(bits >> (sizeof(int) * 8 - writeCount));
                        this.bufferBitsWriter.WriteBits(writeBits, writeCount);
                        bits = bits << writeCount;
                        n -= writeCount;

                        if (this.bufferBitsWriter.IsFull)
                        {
                            this.WriteJpegByte(this.bufferBitsWriter.Data);
                            this.bufferBitsWriter.Clear();
                        }
                    }
                }
            }

            if (this.bufferBitsWriter.IsFull)
            {
                this.WriteJpegByte(this.bufferBitsWriter.Data);
                this.bufferBitsWriter.Clear();
            }
        }

        /// <summary>
        /// Writes a JPEG marker.
        /// </summary>
        /// <param name="marker">Mrker to write.</param>
        public void WriteJpegMarker(JpegMarker marker)
        {
            this.Write(JpegMarker.MarkerIdentifier);
            this.Write(marker.Code);
        }

        /// <summary>
        /// Restart buffer writer.
        /// </summary>
        public void Restart()
        {
            this.bufferBitsWriter.Clear();
        }

        private void WriteJpegByte(byte bits)
        {
            this.Write(bits);
            if (bits == JpegMarker.MarkerIdentifier)
            {
                this.Write(JpegMarker.EscapeFFByte);
            }
        }
    }
}
