﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.ColorSpaces;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Utilities;
#if !NETSTANDARD
using MediaNamespace = System.Windows.Media;
#else
using MediaNamespace = Telerik.Documents.Media;
#endif

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg
{
    internal class JpegImage
    {
        private readonly byte[][,] raster;
        private readonly int width;
        private readonly int height;
        private readonly int numberOfComponents;
        private readonly JpegColorSpace colorTransform;

        public JpegImage(int width, int height, JpegColorSpace colorTransform, byte[] byteRaster)
        {
            Guard.ThrowExceptionIfNull(byteRaster, "byteRaster");

            this.width = width;
            this.height = height;
            this.colorTransform = colorTransform;
            this.numberOfComponents = JpegImage.GetNumberOfComponents(colorTransform);
            this.raster = this.CreateEmptyRaster();
            
            Action<byte[], int, int> setRasterPixel = null;

            switch(colorTransform)
            {
                case JpegColorSpace.Rgb:
                    setRasterPixel = this.SetRgbPixel;
                    break;
                case JpegColorSpace.YCbCr:
                    setRasterPixel = this.SetYCbCrPixel;
                    break;
                case JpegColorSpace.Grayscale:
                    setRasterPixel = this.SetGrayscalePixel;
                    break;
                default:
                    throw new NotSupportedException(string.Format("Not supported colorspace: {0}", colorTransform));
            }

            for (int row = 0; row < height; row++)
            {
                for (int column = 0; column < width; column++)
                {
                    setRasterPixel(byteRaster, row, column);
                }
            }
        }

        internal JpegImage(int width, int height, JpegColorSpace colorTransform, byte[][,] raster)
        {
            Guard.ThrowExceptionIfNull(raster, "raster");

            this.width = width;
            this.height = height;
            this.colorTransform = colorTransform;
            this.raster = raster;
            this.numberOfComponents = this.raster.Length;
        }

        // This property is related to the color transform in the AdobeMarker.
        // For more information see the following link: http://graphicdesign.stackexchange.com/questions/12894/cmyk-jpegs-extracted-from-pdf-appear-inverted/15906#15906
        public JpegColorSpace ColorTransform
        {
            get
            {
                return this.colorTransform;
            }
        }

        public int Height
        {
            get
            {
                return this.height;
            }
        }

        public int NumberOfComponents
        {
            get
            {
                return this.numberOfComponents;
            }
        }

        public int Width
        {
            get
            {
                return this.width;
            }
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1814:PreferJaggedArraysOverMultidimensional", MessageId = "Member")]
        public byte[][,] Raster
        {
            get
            {
                return this.raster;
            }
        }

        public static int GetNumberOfComponents(JpegColorSpace colorSpace)
        {
            switch (colorSpace)
            {
                case JpegColorSpace.Rgb:
                case JpegColorSpace.YCbCr:
                    return 3;

                case JpegColorSpace.YCCK:
                case JpegColorSpace.Cmyk:
                    return 4;

                case JpegColorSpace.Grayscale:
                    return 1;

                case JpegColorSpace.Unknown:
                case JpegColorSpace.Undefined:
                default:
                    throw new NotSupportedException("Color space is not supported.");
            }
        }

        public int[] GetPixels(JpegColorSpace colorSpace)
        {
            switch (colorSpace)
            {
                case JpegColorSpace.Grayscale:
                    return this.GetGrayscalePixels();
                case JpegColorSpace.Rgb:
                    return this.GetRgbPixels();
                case JpegColorSpace.Cmyk:
                    return this.GetCmykPixels();
                default:
                    throw new NotSupportedException("Color space is not supported.");
            }
        }

        public byte[] GetData()
        {
            switch (this.ColorTransform)
            {
                case JpegColorSpace.YCbCr:
                    return this.GetRgbData();
                case JpegColorSpace.YCCK:
                    return this.GetCmykData();
                case JpegColorSpace.Unknown:
                    return this.GetDefaultData();
                case JpegColorSpace.Undefined:
                    switch (this.raster.Length)
                    {
                        case 1:
                            return this.GetDefaultData();
                        case 3:
                            return this.GetRgbData();
                        case 4: 
                            return this.GetCmykData();
                        default:
                            throw new NotSupportedException("Color space is not supported.");
                    }
                default:
                    throw new NotSupportedException("Color space is not supported.");
            }
        }

        private int[] GetGrayscalePixels()
        {
            int[] pixels = new int[this.width * this.height];
            int index = 0;
            for (int i = 0; i < this.height; i++)
            {
                for (int j = 0; j < this.width; j++)
                {
                    pixels[index++] = ColorsHelper.ToPixel(ColorsHelper.FromGray((byte)this.raster[0][i, j]));
                }
            }

            return pixels;
        }

        private int[] GetCmykPixels()
        {
            int[] pixels = new int[this.width * this.height];
            int index = 0;
            for (int i = 0; i < this.height; i++)
            {
                for (int j = 0; j < this.width; j++)
                {
                    byte y = this.raster[0][i, j];
                    byte cb = this.raster[1][i, j];
                    byte cr = this.raster[2][i, j];
                    byte k = this.raster[3][i, j];

                    YCbCrK.ToCmyk(ref y, ref cb, ref cr, ref k);
                    pixels[index++] = ColorsHelper.ToPixel(ColorsHelper.FromCmyk(y, cb, cr, k));
                }
            }

            return pixels;
        }

        private int[] GetRgbPixels()
        {
            int[] pixels = new int[this.width * this.height];
            int index = 0;
            for (int i = 0; i < this.height; i++)
            {
                for (int j = 0; j < this.width; j++)
                {
                    byte y = this.raster[0][i, j];
                    byte cb = this.raster[1][i, j];
                    byte cr = this.raster[2][i, j];

                    YCbCr.ToRgb(ref y, ref cb, ref cr);
                    pixels[index++] = ColorsHelper.ToPixel(MediaNamespace.Color.FromArgb(255, y, cb, cr));
                }
            }

            return pixels;
        }

        private byte[] GetDefaultData()
        {
            byte[] data = new byte[this.width * this.height * this.raster.Length];
            int index = 0;
            
            for (int i = 0; i < this.height; i++)
            {
                for (int j = 0; j < this.width; j++)
                {
                    for (int k = 0; k < this.raster.Length; k++)
                    {
                        data[index++] = this.raster[k][i, j];
                    }
                }
            }

            return data;
        }

        private byte[] GetCmykData()
        {
            byte[] data = new byte[this.width * this.height * 4];
            int index = 0;
            for (int i = 0; i < this.height; i++)
            {
                for (int j = 0; j < this.width; j++)
                {
                    byte y = this.raster[0][i, j];
                    byte cb = this.raster[1][i, j];
                    byte cr = this.raster[2][i, j];
                    byte k = this.raster[3][i, j];

                    YCbCrK.ToCmyk(ref y, ref cb, ref cr, ref k);
                    data[index++] = y;
                    data[index++] = cb;
                    data[index++] = cr;
                    data[index++] = k;
                }
            }

            return data;
        }

        private byte[] GetRgbData()
        {
            byte[] data = new byte[this.width * this.height * 3];
            int index = 0;
            for (int i = 0; i < this.height; i++)
            {
                for (int j = 0; j < this.width; j++)
                {
                    byte y = this.raster[0][i, j];
                    byte cb = this.raster[1][i, j];
                    byte cr = this.raster[2][i, j];

                    YCbCr.ToRgb(ref y, ref cb, ref cr);
                    data[index++] = y;
                    data[index++] = cb;
                    data[index++] = cr;
                }
            }

            return data;
        }

        private void SetRgbPixel(byte[] fromByteRaster, int row, int column)
        {
            byte r = this.GetPixelComponent(fromByteRaster, row, column, 0);
            byte g = this.GetPixelComponent(fromByteRaster, row, column, 1);
            byte b = this.GetPixelComponent(fromByteRaster, row, column, 2);

            YCbCr.FromRgb(ref r, ref g, ref b);

            this.raster[0][row, column] = r;
            this.raster[1][row, column] = g;
            this.raster[2][row, column] = b;
        }

        private void SetYCbCrPixel(byte[] fromByteRaster, int row, int column)
        {
            this.raster[0][row, column] = this.GetPixelComponent(fromByteRaster, row, column, 0);
            this.raster[1][row, column] = this.GetPixelComponent(fromByteRaster, row, column, 1);
            this.raster[2][row, column] = this.GetPixelComponent(fromByteRaster, row, column, 2);
        }

        private void SetGrayscalePixel(byte[] fromByteRaster, int row, int column)
        {
            this.raster[0][row, column] = this.GetPixelComponent(fromByteRaster, row, column, 0);
        }

        private byte GetPixelComponent(byte[] byteRaster, int row, int column, int componentId)
        {
            byte colorComponent = byteRaster[(row * this.width + column) * this.numberOfComponents + componentId];

            return colorComponent;
        }

        private byte[][,] CreateEmptyRaster()
        {
            byte[][,] raster = new byte[this.numberOfComponents][,];

            for (int componentId = 0; componentId < this.numberOfComponents; componentId++)
            {
                raster[componentId] = new byte[this.height, this.width];
            }

            return raster;
        }
    }
}
