﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Decoder;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Encoder;

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg.Markers
{
    /// <summary>
    /// Represents JFIF segment (APP0 marker).
    /// </summary>
    internal class JfifMarker : JpegMarker
    {
        /// <summary>
        /// JFIF segment format.
        /// </summary>
        internal const string JFIF_Identifier = "JFIF";         

        /// <summary>
        /// JFIF extension (JFXX) segment format. Currently is not supported.
        /// </summary>
        internal const string JFXX_Identifier = "JFXX";

        internal const int DEFAULT_LENGTH = 16; // = 16 for usual JPEG, no thumbnail
        private const byte DefaultVersionHi = 1;
        private const byte DefaultVersionLo = 1;
        private const byte DefaultDensityUnits = 0;
        private const ushort DefaultXDensity = 1;
        private const ushort DefaultYDensity = 1;
        private const byte DefaultThumbnailWidth = 0;
        private const byte DefaultThumbnailHeight = 0;

        private string identifier = JfifMarker.JFIF_Identifier;
        private ushort length = JfifMarker.DEFAULT_LENGTH; 	
        private byte versionHi = DefaultVersionHi; // 1
        private byte versionLo = DefaultVersionLo; // 1
        private byte densityUnits = DefaultDensityUnits; // 0 - No units, aspect ratio only specified.
        private ushort xdensity = DefaultXDensity;  // 1
        private ushort ydensity = DefaultYDensity;  // 1
        private byte thumbnWidth = DefaultThumbnailWidth; // 0
        private byte thumbnHeight = DefaultThumbnailHeight; // 0
        private byte[] thumbnailData = null;

        /// <summary>
        /// Gets or sets units for pixel density fields.
        /// </summary>
        /// <remarks>
        /// <list type="bullet">
        /// <item>0 - No units, aspect ratio only specified.</item>
        /// <item>1 - Pixels per inch.</item>
        /// <item>2 - Pixels per centimetre.</item>
        /// </list>
        /// </remarks>
        public byte DensityUnits
        {
            get
            {
                return this.densityUnits;
            }

            set
            {
                this.densityUnits = value;
            }
        }

        /// <summary>
        /// Gets or sets horizontal pixel density.
        /// </summary>
        public ushort DensityX
        {
            get
            {
                return this.xdensity;
            }

            set
            {
                this.xdensity = value;
            }
        }

        /// <summary>
        /// Gets or sets vertical pixel density.
        /// </summary>
        public ushort DensityY
        {
            get
            {
                return this.ydensity;
            }

            set
            {
                this.ydensity = value;
            }
        }

        /// <summary>
        /// Gets length of the marker segment.
        /// </summary>
        public override ushort Length
        {
            get
            {
                return this.length;
            }
        }

        /// <summary>
        /// Gets or sets identifier.
        /// </summary>
        public string Identifier
        {
            get
            {
                return this.identifier;
            }

            set
            {
                if (value != JfifMarker.JFIF_Identifier
                    && value != JfifMarker.JFXX_Identifier)
                {
                    throw new ArgumentOutOfRangeException("value");
                }

                this.identifier = value;
            }
        }

        public override JpegMarkerType MarkerType
        {
            get
            {
                return JpegMarkerType.APP0;
            }
        }

        /// <summary>
        /// Gets or sets embedded JFIF thumbnail data.
        /// </summary>
        public byte[] ThumbnailData
        {
            get
            {
                return this.thumbnailData;
            }

            set
            {
                this.thumbnailData = value;
            }
        }

        /// <summary>
        /// Gets or sets vertical size of embedded JFIF thumbnail in pixels.
        /// </summary>
        public byte ThumbnailHeight
        {
            get
            {
                return this.thumbnHeight;
            }

            set
            {
                this.thumbnHeight = value;
                this.UpdateMarkerLength();
            }
        }

        /// <summary>
        /// Gets or sets horizontal size of embedded JFIF thumbnail in pixels.
        /// </summary>
        public byte ThumbnailWidth
        {
            get
            {
                return this.thumbnWidth;
            }

            set
            {
                this.thumbnWidth = value;
                this.UpdateMarkerLength();
            }
        }

        public override void InterpretMarker(JpegDecoderBase decoder)
        {
            this.length = decoder.Reader.Read16();
            byte[] txt = new byte[5];
            decoder.Reader.Read(txt, 5);
#if WPF
            this.identifier = Encoding.ASCII.GetString(txt, 0, 4);
#else
            this.identifier = Encoding.UTF8.GetString(txt, 0, 4);
#endif

            if (this.identifier == JfifMarker.JFIF_Identifier)
            {
                this.versionHi = decoder.Reader.Read8();
                this.versionLo = decoder.Reader.Read8();
                this.densityUnits = decoder.Reader.Read8();
                this.xdensity = decoder.Reader.Read16();
                this.ydensity = decoder.Reader.Read16();
                this.thumbnWidth = decoder.Reader.Read8();
                this.thumbnHeight = decoder.Reader.Read8();

                if (this.thumbnWidth != 0 && this.thumbnHeight != 0)
                {
                    this.thumbnailData = new byte[3 * this.thumbnWidth * this.thumbnHeight];
                    decoder.Reader.Read(thumbnailData, thumbnailData.Length);
                }
            }
            else
            {
                decoder.Reader.Seek(this.length - 7, System.IO.SeekOrigin.Current);
                Debug.Assert(true, string.Format("{0} segment format is not supported", this.Identifier));
            }
        }

        public override void WriteMarker(JpegEncoder encoder)
        {
            base.WriteMarker(encoder);

            encoder.Writer.Write16(DEFAULT_LENGTH);
            string jfifId = JfifMarker.JFIF_Identifier;
#if WPF
            byte[] txt = Encoding.ASCII.GetBytes(jfifId);
#else
            byte[] txt = Encoding.UTF8.GetBytes(jfifId);
#endif
            encoder.Writer.Write(txt, txt.Length);
            encoder.Writer.Write8(0);

            encoder.Writer.Write8(DefaultVersionHi);
            encoder.Writer.Write8(DefaultVersionLo);
            encoder.Writer.Write8(DefaultDensityUnits);
            encoder.Writer.Write16(DefaultXDensity);
            encoder.Writer.Write16(DefaultYDensity);
            encoder.Writer.Write8(DefaultThumbnailWidth);
            encoder.Writer.Write8(DefaultThumbnailHeight);
        }

        private void UpdateMarkerLength()
        {
            if (this.identifier == JfifMarker.JFIF_Identifier)
            {
                this.length = (ushort)(JfifMarker.DEFAULT_LENGTH + 3 * this.thumbnWidth * this.thumbnHeight);
            }
        }
    }
}
