﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Decoder;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Encoder;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg.Markers
{
    internal abstract class JpegMarker
    {
        public const int MarkerIdentifier = 0xFF;
        public const int EscapeFFByte = 0x00;
        private static readonly Dictionary<byte, JpegMarker> markers;
        
        static JpegMarker()
        {
            markers = new Dictionary<byte, JpegMarker>();

            RegisterMarker(0xE0, new JfifMarker());
            RegisterMarker(0xEE, new AdobeMarker());
            RegisterMarker(0xD8, new StartOfImageMarker());
            RegisterMarker(0xC0, 0xC1, new StartOfFrameMarker(JpegEncodingType.BaselineDct));
            RegisterMarker(0xC9, new StartOfFrameMarker(JpegEncodingType.BaselineDct));
            RegisterMarker(0xC2, new StartOfFrameMarker(JpegEncodingType.ProgressiveDct));
            RegisterMarker(0xCA, new StartOfFrameMarker(JpegEncodingType.ProgressiveDct));
            RegisterMarker(0xC3, new StartOfFrameMarker(JpegEncodingType.NotSupported));
            RegisterMarker(0xC5, 0xC7, new StartOfFrameMarker(JpegEncodingType.NotSupported));
            RegisterMarker(0xCB, new StartOfFrameMarker(JpegEncodingType.NotSupported));
            RegisterMarker(0xCB, 0xCF, new StartOfFrameMarker(JpegEncodingType.NotSupported));
            RegisterMarker(0xC4, new DefineHuffmanTableMarker());
            RegisterMarker(0xDB, new DefineQuantizationTableMarker());
            RegisterMarker(0xDD, new DefineRestartIntervalMarker());
            RegisterMarker(0xDA, new StartOfScanMarker());
            RegisterMarker(0xD0, 0xD7, new RestartMarker());
            RegisterMarker(0xD9, new EndOfImageMarker());
            RegisterMarker(0xFE, new CommentMarker());
        }

        /// <summary>
        /// Gets length of the marker segment.
        /// </summary>
        public abstract ushort Length
        {
            get;
        }

        /// <summary>
        /// Gets marker type.
        /// </summary>
        public abstract JpegMarkerType MarkerType
        {
            get;
        }

        /// <summary>
        /// Gets marker code.
        /// </summary>
        public byte Code
        {
            get;
            private set;
        }

        public static JpegMarker GetMarker(byte code)
        {
            JpegMarker marker;
            if (markers.TryGetValue(code, out marker))
            {
                return marker;
            }

            return new NotSupportedMarker() { Code = code };
        }

        public static JpegMarker GetMarker(JpegMarkerType markerType)
        {
            JpegMarker marker = (from mrk in JpegMarker.markers.Values
                                 where mrk.MarkerType == markerType
                                 select mrk).FirstOrDefault();

            return marker;
        }

        public static StartOfFrameMarker GetSOFMarkerForEncoding(JpegEncodingType encoding)
        {
            StartOfFrameMarker marker = (from mrk in JpegMarker.markers.Values
                                         where mrk.MarkerType == JpegMarkerType.SOF && ((StartOfFrameMarker)mrk).EncodingType == encoding
                                         select mrk).FirstOrDefault() as StartOfFrameMarker;

            return marker;
        }

        public abstract void InterpretMarker(JpegDecoderBase decoder);

        public static bool IsJpegMarker(byte b1, byte b2)
        {
            return IsJpegMarker(b1) && b2 != EscapeFFByte;
        }

        public static bool IsJpegMarker(byte b)
        {
            return b == MarkerIdentifier;
        }

        public static ushort ReadLengthAndSkipData(JpegDecoderBase decoder)
        {
            Guard.ThrowExceptionIfNull(decoder, "decoder");

            ushort length = decoder.Reader.Read16();
            decoder.Reader.Seek(length - 2, System.IO.SeekOrigin.Current);

            return length;
        }

        public virtual void WriteMarker(JpegEncoder encoder)
        {
            encoder.Writer.WriteJpegMarker(this);
        }

        private static void RegisterMarker(byte fromCode, byte toCode, JpegMarker marker)
        {
            for (byte code = fromCode; code <= toCode; code++)
            {
                RegisterMarker(code, marker);
            }
        }

        private static void RegisterMarker(byte code, JpegMarker marker)
        {
            marker.Code = code;
            markers[code] = marker;
        }
    }
}
