﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Decoder;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Encoder;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg.Tables
{
    internal class FrameHeader : JpegTable
    {
        private ushort numberOfLines;
        private ushort samplesPerLine;
        private Dictionary<int, JpegFrameComponent> components;
        private byte imageComponents;
        private int? blocksPerLine;
        private int? blocksPerColumn;

        public FrameHeader()
        {
            this.MaxHSampleFactor = 1;
            this.MaxVSampleFactor = 1;
        }
        
        /// <summary>
        /// Gets number of image components in frame.
        /// </summary>
        /// <remarks>CCITT T.81, Annex B, section B.2.2 Frame header syntax (see Nf parameter in the Table B.2).</remarks>
        public int ImageComponents
        {
            get
            {
                return this.imageComponents;
            }
        }

        /// <summary>
        /// Gets number of lines (height). 
        /// </summary>
        public int Height
        {
            get
            {
                return this.numberOfLines;
            }

            internal set
            {
                this.numberOfLines = (ushort)value;
            }
        }

        /// <summary>
        /// Gets length of the frame header.
        /// </summary>
        /// <remarks>CCITT T.81, Annex B, section B.2.2 Frame header syntax (see Lf parameter in the Table B.2).</remarks>
        public override ushort Length
        {
            get 
            {
                return (ushort)(8 + 3 * this.imageComponents); 
            }
        }

        /// <summary>
        /// Gets the precision in bits for the samples of the components in the frame.
        /// </summary>
        /// <remarks>CCITT T.81, Annex B, section B.2.2 Frame header syntax (see P parameter in the Table B.2).</remarks>
        public byte SamplePrecision
        {
            get;
            internal set;
        }

        /// <summary>
        /// Gets number of the samples per line (width).
        /// </summary>
        public int Width
        {
            get
            {
                return this.samplesPerLine;
            }

            internal set
            {
                this.samplesPerLine = (ushort)value;
            }
        }

        public byte MaxHSampleFactor 
        {
            get;
            private set;
        }

        public byte MaxVSampleFactor 
        { 
            get; 
            private set; 
        }

        public int BlocksPerLine
        {
            get
            {
                if (!this.blocksPerLine.HasValue)
                {
                    this.blocksPerLine = (int)Math.Ceiling((double)this.Width / Constants.BlockWidth / this.MaxHSampleFactor);
                }
                
                return this.blocksPerLine.Value;
            }
        }

        public int BlocksPerColumn
        {
            get
            {
                if (!this.blocksPerColumn.HasValue)
                {
                    this.blocksPerColumn = (int)Math.Ceiling((double)this.Height / Constants.BlockHeight / this.MaxVSampleFactor);
                }

                return this.blocksPerColumn.Value;
            }
        }

        /// <summary>
        /// Add frame component (image color component).
        /// </summary>
        /// <param name="component">JPEG frame component to add.</param>
        public void AddComponent(JpegFrameComponent component)
        {
            if (this.components == null)
            {
                this.components = new Dictionary<int, JpegFrameComponent>();
            }
            this.components[component.ComponentId] = component;
        }

        public JpegFrameComponent GetComponent(int componentId)
        {
            return this.components[componentId];
        }

        /// <summary>
        /// Reads JPEG frame header.
        /// </summary>
        /// <param name="reader">JPEG reader.</param>
        public override void Read(IJpegReader reader)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            
            reader.Read16();
            this.SamplePrecision = reader.Read8();

            this.numberOfLines = reader.Read16();
            this.samplesPerLine = reader.Read16();
            this.imageComponents = reader.Read8();

            this.components = new Dictionary<int, JpegFrameComponent>(this.imageComponents);
            
            for (int i = 0; i < this.imageComponents; i++)
            {
                byte componentId = reader.Read8();
                byte hSampleFactor = reader.Read4();
                byte vSampleFactor = reader.Read4();
                byte qTableId = reader.Read8();

                this.MaxHSampleFactor = Math.Max(this.MaxHSampleFactor, hSampleFactor);
                this.MaxVSampleFactor = Math.Max(this.MaxVSampleFactor, vSampleFactor);

                this.components[componentId] = new JpegFrameComponent(componentId, hSampleFactor, vSampleFactor, qTableId);
            }
        }

        /// <summary>
        /// Writes JPEG frame header.
        /// </summary>
        /// <param name="writer">JPEG writer.</param>
        public override void Write(JpegWriter writer)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");

            this.imageComponents = (byte)this.components.Count;

            writer.Write16(this.Length);
            writer.Write8(this.SamplePrecision);
            writer.Write16(this.numberOfLines);
            writer.Write16(this.samplesPerLine);
            writer.Write8(this.imageComponents);

            foreach (JpegFrameComponent component in this.components.Values)
            {
                writer.Write8(component.ComponentId);
                writer.Write4(component.HSampleFactor);
                writer.Write4(component.VSampleFactor);
                writer.Write8(component.QuantizationTableId);
            }
        }
    }
}
