﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Decoder;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Encoder;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg.Tables
{
    internal class HuffmanTable : JpegTable
    {
        #region Constants

        public const int HuffmanTableClasses = 2;
        private const int HuffmanLengthsCount = 16;

        #endregion


        #region Fields

        private byte[] huffSize;
        private short[] huffCode;
        private byte[] bits;
        private short[] huffVals;
        private short[] minCode;
        private short[] maxCode;
        private short[] valPtr;
        private byte[] encodingSize;
        private short[] encodingCode;

        #endregion


        #region Constructors

        public HuffmanTable()
        {
        }

        public HuffmanTable(TableClass tableClass, byte tableIndex, byte[] lengths, byte[] values)
        {
            this.TableClass = tableClass;
            this.TableIndex = tableIndex;

            this.bits = new byte[lengths.Length];
            lengths.CopyTo(this.bits, 0);

            this.huffVals = new short[values.Length];
            values.CopyTo(this.huffVals, 0);

            this.GenerateSizeTable();
            this.GenerateCodeTable();
            this.GenerateEncoderTables();
        }

        #endregion


        #region Properties

        /// <summary>
        /// Gets a length of the Huffman table.
        /// </summary>
        /// <remarks>CCITT T.81, Annex B, section B.2.4.2 Huffman table-specification syntax (see Tc, Th, Li and Vi,j parameters in the Table B.5).</remarks>
        public override ushort Length
        {
            get
            {
                return (ushort)(HuffmanTable.HuffmanLengthsCount + 1 + this.bits.Sum(b => b));
            }
        }

        public byte TableIndex
        {
            get;
            private set;
        }

        public TableClass TableClass
        {
            get;
            private set;
        }


        #endregion


        #region Methods

        public HuffmanTable Clone()
        {
            HuffmanTable other = new HuffmanTable();
            other.huffSize = (byte[])this.huffSize.Clone();
            other.huffCode = (short[])this.huffCode.Clone();
            other.bits = (byte[])this.bits.Clone();
            other.huffVals = (short[])this.huffVals.Clone();
            other.minCode = (short[])this.minCode.Clone();
            other.maxCode = (short[])this.maxCode.Clone();
            other.valPtr = (short[])this.valPtr.Clone();

            return other;
        }

        public int Decode(IJpegReader reader)
        {
            int i = 0;
            int bits = reader.ReadBit();

            short code = (short)bits;

            while (code > this.maxCode[i])
            {
                i++;
                code <<= 1;
                bits = reader.ReadBit();

                int result = code | (short)bits;
                code = (short)result;
            }

            int value = this.huffVals[code + this.valPtr[i]];

            if (value < 0)
            {
                value = 256 + value;
            }

            return value;
        }

        public void Encode(JpegWriter writer, int nbits)
        {
            writer.WriteBits(this.encodingSize[nbits], this.encodingCode[nbits]);
        }

        private void GenerateSizeTable()
        {
            List<byte> huffSize = new List<byte>();
            short k = 0;
            for (int i = 0; i < this.bits.Length; i++)
            {
                for (int j = 0; j < this.bits[i]; j++)
                {
                    huffSize.Add((byte)(i + 1));
                    k++;
                }
            }

            huffSize.Add(0);

            this.huffSize = huffSize.ToArray();
        }

        private void GenerateCodeTable()
        {
            short k = 0;
            short code = 0;
            short si = this.huffSize[0];
            List<short> huffCode = new List<short>();
            int length = this.huffSize.Length - 1;

            for (short i = 0; i < length; i++)
            {
                while (this.huffSize[k] == si)
                {
                    huffCode.Add(code);
                    code++;
                    k++;
                }

                code <<= 1;
                si++;
            }

            this.huffCode = huffCode.ToArray();
        }

        private void GenerateEncoderTables()
        {
            this.encodingSize = new byte[256];
            this.encodingCode = new short[256];

            int length = this.huffSize.Length - 1;

            for (int i = 0; i < length; i++)
            {
                this.encodingCode[this.huffVals[i]] = this.huffCode[i];
                this.encodingSize[this.huffVals[i]] = this.huffSize[i];
            }
        }

        private void GenerateDecoderTables()
        {
            this.minCode = new short[HuffmanLengthsCount];
            this.maxCode = new short[HuffmanLengthsCount];
            this.valPtr = new short[HuffmanLengthsCount];

            for (int i = 0; i < this.minCode.Length; i++)
            {
                this.minCode[i] = -1;
            }

            for (int i = 0; i < this.maxCode.Length; i++)
            {
                this.maxCode[i] = -1;
            }

            short bitCount = 0;
            for (int i = 0; i < 16; i++)
            {
                if (bits[i] != 0)
                {
                    this.valPtr[i] = bitCount;
                }

                for (int j = 0; j < bits[i]; j++)
                {
                    if (this.huffCode[j + bitCount] < this.minCode[i] || this.minCode[i] == -1)
                    {
                        this.minCode[i] = this.huffCode[j + bitCount];
                    }

                    if (this.huffCode[j + bitCount] > this.maxCode[i])
                    {
                        this.maxCode[i] = this.huffCode[j + bitCount];
                    }
                }

                if (this.minCode[i] != -1)
                {
                    this.valPtr[i] = (short)(this.valPtr[i] - this.minCode[i]);
                }

                bitCount += this.bits[i];
            }
        }

        private void InitializeHuffmanTables()
        {
            this.GenerateSizeTable();
            this.GenerateCodeTable();
            this.GenerateDecoderTables();
        }

        #endregion


        #region Overrides

        /// <summary>
        /// Reads Huffman table.
        /// </summary>
        /// <param name="reader">JPEG reader.</param>
        public override void Read(IJpegReader reader)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");

            this.TableClass = (TableClass)reader.Read4();
            this.TableIndex = reader.Read4();

            this.bits = new byte[HuffmanTable.HuffmanLengthsCount];
            int count = 0;
            for (int i = 0; i < this.bits.Length; i++)
            {
                this.bits[i] = reader.Read8();
                count += this.bits[i];
            }

            this.huffVals = new short[count];
            for (int i = 0; i < this.huffVals.Length; i++)
            {
                this.huffVals[i] = reader.Read8();
            }

            this.InitializeHuffmanTables();
        }

        /// <summary>
        /// Writes Huffman table.
        /// </summary>
        /// <param name="writer">JPEG writer.</param>
        public override void Write(JpegWriter writer)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");

            writer.Write4((byte)this.TableClass);
            writer.Write4((byte)this.TableIndex);

            for (int i = 0; i < this.bits.Length; i++)
            {
                writer.Write8(this.bits[i]);
            }

            for (int i = 0; i < this.huffVals.Length; i++)
            {
                writer.Write8((byte)this.huffVals[i]);
            }
        }

        #endregion
    }
}
