﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Decoder;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Encoder;

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg.Tables
{
    internal class JpegComponent
    {
        #region Fields

        private readonly byte id;
        private readonly byte quantizationTableId;
        private readonly List<Block> blocks;
        private readonly byte hSampleFactor;
        private readonly byte vSampleFactor;
        private readonly IHuffmanTablesOwner huffmanTablesOwner;
        private int lastDCCoefficient;

        // See A.2.4 Completion of partial MCU.
        private int completedHeight;
        private int completedWidth;

        // private bool lastColumnIsDummy = false;
        // private bool lastRowIsDummy = false;

        private int verticalBlocksCount;
        private int horizontalBlocksCount;
        private int mcuPerRow;

        #endregion


        #region Constructors

        public JpegComponent(byte id, byte hSampleFactor, byte vSampleFactor, byte quantizationTableId, IHuffmanTablesOwner huffmanTablesOwner)
        {
            this.id = id;
            this.hSampleFactor = hSampleFactor;
            this.vSampleFactor = vSampleFactor;
            this.quantizationTableId = quantizationTableId;
            this.huffmanTablesOwner = huffmanTablesOwner;
            this.lastDCCoefficient = 0;

            this.blocks = new List<Block>();
        }

        #endregion


        #region Properties

        public int ACTableId
        {
            get;
            set;
        }

        public int DCTableId
        {
            get;
            set;
        }

        public HuffmanTable AcTable
        {
            get
            {
                return this.huffmanTablesOwner.GetHuffmanTable(TableClass.ACTable, this.ACTableId);
            }
        }

        public HuffmanTable DcTable
        {
            get
            {
                return this.huffmanTablesOwner.GetHuffmanTable(TableClass.DCTable, this.DCTableId); 
            }
        }

        public int HorizontalBlocksCount
        {
            get
            {
                return this.horizontalBlocksCount;
            }
        }

        public byte HSampleFactor
        {
            get
            {
                return this.hSampleFactor;
            }
        }

        public byte Id
        {
            get
            {
                return this.id;
            }
        }

        public int LastDCCoefficient
        {
            get
            {
                return this.lastDCCoefficient;
            }
        }

        public int QuantizationTableId
        {
            get
            {
                return this.quantizationTableId;
            }
        }

        public int VerticalBlocksCount
        {
            get
            {
                return this.verticalBlocksCount;
            }
        }

        public byte VSampleFactor
        {
            get
            {
                return this.vSampleFactor;
            }
        }

        public int McuPerRow
        {
            get
            {
                return this.mcuPerRow;
            }
        }

        #endregion


        #region Methods

        public static List<FloatBlock> FDCT(List<Block> blocksToProcess)
        {
            List<FloatBlock> encodingBlocks = new List<FloatBlock>();

            for (int i = 0; i < blocksToProcess.Count; i++)
            {
                FloatBlock transformedBlock = DiscreteCosineTransform.ForwardDCT(blocksToProcess[i]);
                encodingBlocks.Add(transformedBlock);
            }

            return encodingBlocks;
        }

        public static List<Block> Quantize(List<FloatBlock> encodingBlocks, QuantizationTable table)
        {
            List<Block> quantizedBlocks = new List<Block>();

            for (int i = 0; i < encodingBlocks.Count; i++)
            {
                FloatBlock floatBlock = encodingBlocks[i];

                for (int k = 0; k < Constants.BlockSize; k++)
                {
                    floatBlock[k] *= (float)table.Divisor[k];
                }

                quantizedBlocks.Add(ZigZagScan.ZigZag(floatBlock).ToBlock());
            }

            return quantizedBlocks;
        }

        public Block GetBlock(int mcuIndex, int rowIndex, int columnIndex)
        {
            int blockIndex = (mcuIndex * this.VSampleFactor * this.HSampleFactor) + (rowIndex * this.HSampleFactor) + columnIndex;

            return this.GetBlock(blockIndex);
        }

        public Block GetBlock(int index)
        {
            while (index >= this.blocks.Count)
            {
                this.blocks.Add(new Block());
            }

            return this.blocks[index];
        }

        public void UpdateLastDCCoefficient(Block block)
        {
            this.lastDCCoefficient = block[0];
        }
        
        public void DequantizeAndInverse(QuantizationTable quantizationTable)
        {
            for (int i = 0; i < this.blocks.Count; i++)
            {
                quantizationTable.DequantizeAndInverse(this.blocks[i]);
            }
        }
        
        public void Dequantize(QuantizationTable table)
        {
            foreach (Block block in this.blocks)
            {
                for (int k = 0; k < Constants.BlockSize; k++)
                {
                    int z = ZigZagScan.ZigZagMap[k];
                    block[z] *= table[k];
                }
            }
        }

        public void EncodeNextBlocks(JpegEncoder encoder, List<Block> quantizedBlocks)
        {
            int lastEncodedBlock = 0;

            for (int i = 0; i < this.VSampleFactor; i++)
            {
                for (int j = 0; j < this.HSampleFactor; j++)
                {
                    encoder.ComponentEncoder.EncodeBlock(encoder, this, quantizedBlocks[lastEncodedBlock]);
                    this.lastDCCoefficient = quantizedBlocks[lastEncodedBlock][0];
                    lastEncodedBlock++;
                }
            }
        }

        public void IDCT()
        {
            for (int i = 0; i < this.blocks.Count; i++)
            {
                this.blocks[i] = DiscreteCosineTransform.InverseDCT(this.blocks[i]);
            }
        }

        public void InitializeDecoding(JpegDecoder decoder)
        {
            this.horizontalBlocksCount = (int)Math.Ceiling(Math.Ceiling((double)decoder.FrameHeader.Width / Constants.BlockWidth)
                * this.HSampleFactor / decoder.FrameHeader.MaxHSampleFactor);

            this.verticalBlocksCount = (int)Math.Ceiling(Math.Ceiling((double)decoder.FrameHeader.Height / Constants.BlockHeight) *
                this.VSampleFactor / decoder.FrameHeader.MaxVSampleFactor);

            this.mcuPerRow = this.CalculateMcuPerRow();
            
            // TODO: This code should generate the initial blocks buffer size. However there is some issue when rasterizing with this buffer size. Currently the buffer size is constructed by using GetBlock(index) method.
            //int blocksPerLineForMcu = decoder.FrameHeader.BlocksPerLine * this.HSampleFactor;
            //int blocksPerColumnForMcu = decoder.FrameHeader.BlocksPerColumn * this.VSampleFactor;
            //int blocksBuffer = blocksPerColumnForMcu * (blocksPerLineForMcu + 1);
            //for (int i = 0; i < blocksBuffer; i++)
            //{
            //    this.blocks.Add(new Block());
            //}
        }

        public void InitializeEncoding(JpegScan scan, JpegEncoder encoder)
        {
            int hFactor = scan.MaxHSampleFactor / this.HSampleFactor;
            int vFactor = scan.MaxVSampleFactor / this.VSampleFactor;

            // See A.2.4 Completion of partial MCU.
            this.completedWidth = ((encoder.Width % Constants.BlockWidth != 0) ?
                ((int)Math.Ceiling((double)encoder.Width / (double)Constants.BlockWidth)) * Constants.BlockWidth :
                encoder.Width) *
                hFactor;
            // this.lastColumnIsDummy = this.completedWidth != (encoder.Width * hFactor);

            // results in a multiple of 8 for compWidthz
            // this will make the rest of the program fail for the unlikely
            // event that someone tries to compress an 16 x 16 pixel image
            // which would of course be worse than pointless

            this.horizontalBlocksCount = (int)Math.Ceiling((double)this.completedWidth / (double)Constants.BlockWidth);

            this.completedHeight = ((encoder.Height % Constants.BlockHeight != 0) ?
                ((int)Math.Ceiling((double)encoder.Height / (double)Constants.BlockHeight)) * Constants.BlockHeight :
                encoder.Height) *
                vFactor;
            // this.lastRowIsDummy = this.completedHeight != ((encoder.Height / scan.MaxVSampleFactor) * this.VSampleFactor);

            this.verticalBlocksCount = (int)Math.Ceiling((double)this.completedHeight / (double)Constants.BlockHeight);

            this.mcuPerRow = this.CalculateMcuPerRow();
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1814:PreferJaggedArraysOverMultidimensional", MessageId = "0#")]
        public List<Block> PrepareBlocks(byte[][,] raster, int blockStartYPos, int blockStartXPos)
        {
            List<Block> blocks = new List<Block>();

            for (int i = 0; i < this.VSampleFactor; i++)
            {
                for (int j = 0; j < this.HSampleFactor; j++)
                {
                    Block block = new Block(raster[this.Id - 1], blockStartXPos, blockStartYPos, j, i);
                    blocks.Add(block);
                }
            }

            return blocks;
        }

        public byte[,] RasterizeInterleaved(JpegScan scan, int width, int height)
        {
            byte[,] raster = new byte[height, width];
            int blockIndex = 0;
            int incrementBlock = 0;
            int x = 0;
            int y = 0;
            int lastBlockHeight = 0;

            int hFactor = scan.MaxHSampleFactor / this.HSampleFactor;
            int vFactor = scan.MaxVSampleFactor / this.VSampleFactor;

            int scaledBlockWidth = Constants.BlockWidth * hFactor;
            int scaledBlockHeight = Constants.BlockHeight * vFactor;

            while (blockIndex < this.blocks.Count)
            {
                int blockWidth = 0;
                int blockHeight = 0;

                if (x >= width)
                {
                    x = 0;
                    y += incrementBlock;
                }

                for (int i = 0; i < this.VSampleFactor; i++)
                {
                    blockWidth = 0;
                    for (int j = 0; j < this.HSampleFactor; j++)
                    {
                        Block block = this.blocks[blockIndex++];

                        RasterizeBlockInterleaved(raster, block, width, height, x, y, scaledBlockWidth, scaledBlockHeight, hFactor, vFactor);

                        blockWidth += scaledBlockWidth;
                        x += scaledBlockWidth;
                        blockHeight = scaledBlockHeight;
                    }

                    y += blockHeight;
                    x -= blockWidth;
                    lastBlockHeight += blockHeight;
                }

                y -= lastBlockHeight;
                incrementBlock = lastBlockHeight;
                lastBlockHeight = 0;
                x += blockWidth;
            }

            return raster;
        }

        public byte[,] RasterizeNonInterleaved(JpegScan scan, int width, int height)
        {
            byte[,] raster = new byte[height, width];
            int blockIndex = 0;
            int incrementBlock = 0;
            int x = 0;
            int y = 0;
            int lastBlockHeight = 0;

            int hFactor = scan.MaxHSampleFactor / this.HSampleFactor;
            int vFactor = scan.MaxVSampleFactor / this.VSampleFactor;

            int scaledBlockWidth = Constants.BlockWidth * hFactor;
            int scaledBlockHeight = Constants.BlockHeight * vFactor;

            while (blockIndex < this.blocks.Count)
            {
                int blockWidth = 0;
                int blockHeight = 0;

                if (x >= width)
                {
                    x = 0;
                    y += incrementBlock;
                }

                blockWidth = 0;

                if (blockIndex >= this.blocks.Count)
                {
                    continue;
                }

                Block block = this.blocks[blockIndex++];

                RasterizeBlockNonInterleaved(raster, block, width, height, x, y, scaledBlockWidth, scaledBlockHeight);

                blockWidth += scaledBlockWidth;
                x += scaledBlockWidth;
                blockHeight = scaledBlockHeight;


                y += blockHeight;
                x -= blockWidth;
                lastBlockHeight += blockHeight;


                y -= lastBlockHeight;
                incrementBlock = lastBlockHeight;
                lastBlockHeight = 0;
                x += blockWidth;
            }

            return raster;
        }

        public void Restart()
        {
            this.lastDCCoefficient = 0;
        }

        private int CalculateMcuPerRow()
        {
            return (int)Math.Ceiling((double)this.HorizontalBlocksCount / this.HSampleFactor);
        }

        private static void RasterizeBlockInterleaved(byte[,] raster, Block block, int width, int height, int x, int y, int blockWidth, int blockHeight, int hFactor, int vFactor)
        { 
            int yMax = blockHeight;
            if (y + yMax > height)
            {
                yMax = height - y;
            }

            int xMax = blockWidth;
            if (x + xMax > width)
            {
                xMax = width - x;
            }

            if (vFactor == 1 && hFactor == 1)
            {
                for (int i = 0; i < yMax; i++)
                {
                    for (int j = 0; j < xMax; j++)
                    {
                        raster[i + y, j + x] = (byte)block[i, j];
                    }
                }
            }
            else if (hFactor == 2 && vFactor == 2 && xMax == blockWidth && yMax == blockHeight)
            {
                for (int i = 0; i < Constants.BlockHeight; i++)
                {
                    int by = i * 2 + y;
                    for (int j = 0; j < Constants.BlockWidth; j++)
                    {
                        byte val = (byte)block[i, j];
                        int bx = j * 2 + x;
                        raster[by, bx] = val;
                        raster[by, bx + 1] = val;
                        raster[by + 1, bx] = val;
                        raster[by + 1, bx + 1] = val;
                    }
                }
            }
            else
            {
                for (int i = 0; i < yMax; i++)
                {
                    int by = i / vFactor;
                    for (int j = 0; j < xMax; j++)
                    {
                        int bx = j / hFactor;
                        raster[i + y, j + x] = (byte)block[by, bx];
                    }
                }
            }
        }

        private static void RasterizeBlockNonInterleaved(byte[,] raster, Block block, int width, int height, int x, int y, int blockWidth, int blockHeight)
        { 
            int yMax = blockHeight;
            if (y + yMax > height)
            {
                yMax = height - y;
            }

            int xMax = blockWidth;
            if (x + xMax > width)
            {
                xMax = width - x;
            }

            for (int i = 0; i < yMax; i++)
            {
                for (int j = 0; j < xMax; j++)
                {
                    raster[i + y, j + x] = (byte)block[i, j];
                }
            }
        }

        #endregion 
    }
}
