﻿using System;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Decoder;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Encoder;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg.Tables
{
    internal class QuantizationTable : JpegTable
    {
        private const int dctCos1 = 4017;   // cos(pi/16)
        private const int dctSin1 = 799;   // sin(pi/16)
        private const int dctCos3 = 3406;   // cos(3*pi/16)
        private const int dctSin3 = 2276;   // sin(3*pi/16)
        private const int dctCos6 = 1567;   // cos(6*pi/16)
        private const int dctSin6 = 3784;   // sin(6*pi/16)
        private const int dctSqrt2 = 5793;   // sqrt(2)
        private const int dctSqrt1d2 = 2896;  // sqrt(2) / 2
        private static readonly double[] CosineScaleFactor = { 1.0, 1.387039845, 1.306562965, 1.175875602, 1.0, 0.785694958, 0.541196100, 0.275899379 };
        
        #region Fields

        private ushort[] data;
        private double[] divisor;

        #endregion

        /// <summary>
        /// Initializes a new instance of the QuantizationTable class.
        /// </summary>
        public QuantizationTable()
        {
        }

        /// <summary>
        /// Initializes a new instance of the QuantizationTable class.
        /// </summary>
        /// <param name="tableIndex">Table index</param>
        /// <param name="byteData">Table data.</param>
        public QuantizationTable(byte tableIndex, byte[] byteData)
        {
            this.TableIndex = tableIndex;
            this.data = new ushort[byteData.Length];
            byteData.CopyTo(this.data, 0);

            this.divisor = new double[byteData.Length];

            int index = 0;
            for (int row = 0; row < Constants.BlockHeight; row++)
            {
                for (int column = 0; column < Constants.BlockWidth; column++)
                {
                    this.divisor[index] = (1.0d /
                        ((double)this.data[index] * QuantizationTable.CosineScaleFactor[row] * QuantizationTable.CosineScaleFactor[column] * Constants.BlockWidth));

                    index++;
                }
            }
        }


        #region Properties

        public double[] Divisor
        {
            get
            {
                return this.divisor;
            }
        }

        /// <summary>
        /// Gets a length of all quantization table parameters.
        /// </summary>
        /// <remarks>CCITT T.81, Annex B, section B.2.4.1 Quantization table-specification syntax (see Pq, Tq and Qr parameters in the Table B.4).</remarks>
        public override ushort Length
        {
            get
            {
                return (ushort)(Constants.BlockSize * (1 + this.Precision) + 1);
            }
        }

        /// <summary>
        /// Gets Quantization table element precision.
        /// </summary>
        /// <remarks>CCITT T.81, Annex B, section B.2.4.1 Quantization table-specification syntax (see Pq parameter in the Table B.4).
        /// Specifies the precision of the Qk values. Value 0 indicates 8-bit Qk values; value 1 indicates 16-bit Qk values. 
        /// Pq shall be zero for 8 bit sample precision P (see B.2.2).
        /// </remarks>
        public byte Precision
        {
            get;
            internal set;
        }

        public byte TableIndex
        {
            get;
            internal set;
        }

        #endregion


        #region Indexers

        public ushort this[int index]
        {
            get
            {
                return this.data[index];
            }
        }

        #endregion


        #region Overrides

        /// <summary>
        /// Reads all quantization table parameters.
        /// </summary>
        /// <param name="reader">JPEG reader.</param>
        public override void Read(IJpegReader reader)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");

            this.Precision = reader.Read4();
            this.TableIndex = reader.Read4();

            this.data = new ushort[Constants.BlockSize];

            for (int i = 0; i < this.data.Length; i++)
            {
                switch (this.Precision)
                {
                    case 0:
                        this.data[i] = reader.Read8();
                        break;

                    case 1:
                        this.data[i] = reader.Read16();
                        break;
                }
            }
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            int index = 0;
            for (int i = 0; i < Constants.BlockHeight; i++)
            {
                for (int j = 0; j < Constants.BlockWidth; j++)
                {
                    sb.AppendFormat("{0} ", this.data[index++]);
                }

                sb.AppendLine();
            }

            return sb.ToString();
        }

        /// <summary>
        /// Writes all quantization table parameters.
        /// </summary>
        /// <param name="writer"></param>
        public override void Write(JpegWriter writer)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");

            writer.Write4(this.Precision);
            writer.Write4(this.TableIndex);

            for (int i = 0; i < this.data.Length; i++)
            {
                switch (this.Precision)
                {
                    case 0:
                        writer.Write8((byte)this.data[ZigZagScan.ZigZagMap[i]]);
                        break;

                    case 1:
                        writer.Write16(this.data[ZigZagScan.ZigZagMap[i]]);
                        break;
                }
            }
        }

        #endregion


        #region Methods

        // A port of poppler's IDCT method which in turn is taken from:
        //   Christoph Loeffler, Adriaan Ligtenberg, George S. Moschytz,
        //   "Practical Fast 1-D DCT Algorithms with 11 Multiplications",
        //   IEEE Intl. Conf. on Acoustics, Speech & Signal Processing, 1989,
        //   988-991.
        public void DequantizeAndInverse(Block block)
        {
            int v0, v1, v2, v3, v4, v5, v6, v7, t;

            // dequant
            for (int i = 0; i < 64; i++)
            {
                block[i] = block[i] * this[i];
            }

            // inverse DCT on rows
            for (int i = 0; i < 8; i++)
            {
                var row = 8 * i;

                // check for all-zero AC coefficients
                if (block[1 + row] == 0 && block[2 + row] == 0 && block[3 + row] == 0 &&
                    block[4 + row] == 0 && block[5 + row] == 0 && block[6 + row] == 0 &&
                    block[7 + row] == 0)
                {
                    t = (dctSqrt2 * block[0 + row] + 512) >> 10;
                    block[0 + row] = t;
                    block[1 + row] = t;
                    block[2 + row] = t;
                    block[3 + row] = t;
                    block[4 + row] = t;
                    block[5 + row] = t;
                    block[6 + row] = t;
                    block[7 + row] = t;

                    continue;
                }

                // stage 4
                v0 = (dctSqrt2 * block[0 + row] + 128) >> 8;
                v1 = (dctSqrt2 * block[4 + row] + 128) >> 8;
                v2 = block[2 + row];
                v3 = block[6 + row];
                v4 = (dctSqrt1d2 * (block[1 + row] - block[7 + row]) + 128) >> 8;
                v7 = (dctSqrt1d2 * (block[1 + row] + block[7 + row]) + 128) >> 8;
                v5 = block[3 + row] << 4;
                v6 = block[5 + row] << 4;

                // stage 3
                t = (v0 - v1 + 1) >> 1;
                v0 = (v0 + v1 + 1) >> 1;
                v1 = t;
                t = (v2 * dctSin6 + v3 * dctCos6 + 128) >> 8;
                v2 = (v2 * dctCos6 - v3 * dctSin6 + 128) >> 8;
                v3 = t;
                t = (v4 - v6 + 1) >> 1;
                v4 = (v4 + v6 + 1) >> 1;
                v6 = t;
                t = (v7 + v5 + 1) >> 1;
                v5 = (v7 - v5 + 1) >> 1;
                v7 = t;

                // stage 2
                t = (v0 - v3 + 1) >> 1;
                v0 = (v0 + v3 + 1) >> 1;
                v3 = t;
                t = (v1 - v2 + 1) >> 1;
                v1 = (v1 + v2 + 1) >> 1;
                v2 = t;
                t = (v4 * dctSin3 + v7 * dctCos3 + 2048) >> 12;
                v4 = (v4 * dctCos3 - v7 * dctSin3 + 2048) >> 12;
                v7 = t;
                t = (v5 * dctSin1 + v6 * dctCos1 + 2048) >> 12;
                v5 = (v5 * dctCos1 - v6 * dctSin1 + 2048) >> 12;
                v6 = t;

                // stage 1
                block[0 + row] = v0 + v7;
                block[7 + row] = v0 - v7;
                block[1 + row] = v1 + v6;
                block[6 + row] = v1 - v6;
                block[2 + row] = v2 + v5;
                block[5 + row] = v2 - v5;
                block[3 + row] = v3 + v4;
                block[4 + row] = v3 - v4;
            }

            // inverse DCT on columns
            for (int i = 0; i < 8; i++)
            {
                var col = i;

                // check for all-zero AC coefficients
                if (block[1 * 8 + col] == 0 && block[2 * 8 + col] == 0 && block[3 * 8 + col] == 0 &&
                    block[4 * 8 + col] == 0 && block[5 * 8 + col] == 0 && block[6 * 8 + col] == 0 &&
                    block[7 * 8 + col] == 0)
                {
                    t = (dctSqrt2 * block[i + 0] + 8192) >> 14;
                    block[0 * 8 + col] = t;
                    block[1 * 8 + col] = t;
                    block[2 * 8 + col] = t;
                    block[3 * 8 + col] = t;
                    block[4 * 8 + col] = t;
                    block[5 * 8 + col] = t;
                    block[6 * 8 + col] = t;
                    block[7 * 8 + col] = t;
                    continue;
                }

                // stage 4
                v0 = (dctSqrt2 * block[0 * 8 + col] + 2048) >> 12;
                v1 = (dctSqrt2 * block[4 * 8 + col] + 2048) >> 12;
                v2 = block[2 * 8 + col];
                v3 = block[6 * 8 + col];
                v4 = (dctSqrt1d2 * (block[1 * 8 + col] - block[7 * 8 + col]) + 2048) >> 12;
                v7 = (dctSqrt1d2 * (block[1 * 8 + col] + block[7 * 8 + col]) + 2048) >> 12;
                v5 = block[3 * 8 + col];
                v6 = block[5 * 8 + col];

                // stage 3
                t = (v0 - v1 + 1) >> 1;
                v0 = (v0 + v1 + 1) >> 1;
                v1 = t;
                t = (v2 * dctSin6 + v3 * dctCos6 + 2048) >> 12;
                v2 = (v2 * dctCos6 - v3 * dctSin6 + 2048) >> 12;
                v3 = t;
                t = (v4 - v6 + 1) >> 1;
                v4 = (v4 + v6 + 1) >> 1;
                v6 = t;
                t = (v7 + v5 + 1) >> 1;
                v5 = (v7 - v5 + 1) >> 1;
                v7 = t;

                // stage 2
                t = (v0 - v3 + 1) >> 1;
                v0 = (v0 + v3 + 1) >> 1;
                v3 = t;
                t = (v1 - v2 + 1) >> 1;
                v1 = (v1 + v2 + 1) >> 1;
                v2 = t;
                t = (v4 * dctSin3 + v7 * dctCos3 + 2048) >> 12;
                v4 = (v4 * dctCos3 - v7 * dctSin3 + 2048) >> 12;
                v7 = t;
                t = (v5 * dctSin1 + v6 * dctCos1 + 2048) >> 12;
                v5 = (v5 * dctCos1 - v6 * dctSin1 + 2048) >> 12;
                v6 = t;

                // stage 1
                block[0 * 8 + col] = v0 + v7;
                block[7 * 8 + col] = v0 - v7;
                block[1 * 8 + col] = v1 + v6;
                block[6 * 8 + col] = v1 - v6;
                block[2 * 8 + col] = v2 + v5;
                block[5 * 8 + col] = v2 - v5;
                block[3 * 8 + col] = v3 + v4;
                block[4 * 8 + col] = v3 - v4;
            }

            // convert to 8-bit integers
            for (int i = 0; i < 64; i++)
            {
                var q = block[i];
                q = (q <= -2056) ? 0 : (q >= 2024) ? 255 : (q + 2056) >> 4;
                block[i] = q;
            }
        }

        #endregion
    }
}
