﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Decoder;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Encoder;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg.Tables
{
    internal class ScanHeader : JpegTable
    {
        #region Fields

        private JpegScanComponent[] components;

        #endregion


        #region Properties

        /// <summary>
        /// Gets Successive approximation bit position high.
        /// </summary>
        /// <remarks>CCITT T.81, Annex B, section B.2.3 Scan header syntax (see Ah parameter in the Table B.3).
        /// This parameter specifies the point transform used in the preceding scan (i.e. successive approximation bit position low 
        /// in the preceding scan) for the band of coefficients specified by Ss and Se. This parameter shall be set to zero for the 
        /// first scan of each band of coefficients. In the lossless mode of operations this parameter has no meaning. It shall be set to zero.
        /// </remarks>
        public byte BitPositionHigh
        {
            get;
            internal set;
        }

        /// <summary>
        /// Gets Successive approximation bit position low or point transform.
        /// </summary>
        /// <remarks>CCITT T.81, Annex B, section B.2.3 Scan header syntax (see Al parameter in the Table B.3).
        /// In the DCT modes of operation this parameter specifies the point transform, i.e. bit position low, used before coding the band 
        /// of coefficients specified by Ss and Se. This parameter shall be set to zero for the sequential DCT processes. In the lossless 
        /// mode of operations, this parameter specifies the point transform, Pt.
        /// </remarks>
        public byte BitPositionLow
        {
            get;
            internal set;
        }

        public JpegScanComponent[] Components
        {
            get
            {
                return this.components;
            }
        }

        /// <summary>
        /// Gets a scan header length.
        /// </summary>
        /// <remarks>CCITT T.81, Annex B, section B.2.3 Scan header syntax (see Ls parameter in the Table B.3).</remarks>
        public override ushort Length
        {
            get
            {
                return (ushort)(6 + 2 * this.NumberOfComponents);
            }
        }

        public int NumberOfComponents
        {
            get
            {
                return this.components.Length;
            }

            internal set
            {
                this.components = new JpegScanComponent[value];
            }
        }

        /// <summary>
        /// Gets End of spectral selection.
        /// </summary>
        /// <remarks>CCITT T.81, Annex B, section B.2.3 Scan header syntax (see Se parameter in the Table B.3).
        /// Specifies the last DCT coefficient in each block in zig-zag order which shall be coded in the scan. 
        /// This parameter shall be set to 63 for the sequential DCT processes. In the lossless mode of operations 
        /// this parameter has no meaning. It shall be set to zero.
        /// </remarks>
        public byte SpectralSelectionEnd
        {
            get;
            internal set;
        }

        /// <summary>
        /// Gets Start of spectral or predictor selection.
        /// </summary>
        /// <remarks>CCITT T.81, Annex B, section B.2.3 Scan header syntax (see Ss parameter in the Table B.3).
        /// In the DCT modes of operation, this parameter specifies the first DCT coefficient in each block in zig-zag order 
        /// which shall be coded in the scan. This parameter shall be set to zero for the sequential DCT processes. 
        /// In the lossless mode of operations this parameter is used to select the predictor.
        /// </remarks>
        public byte SpectralSelectionStart
        {
            get;
            internal set;
        }

        #endregion


        #region Overrides

        /// <summary>
        /// Add frame component (image color component).
        /// </summary>
        /// <param name="index">Index of the component.</param>
        /// <param name="component">JPEG frame component to add.</param>
        public void AddComponent(int index, JpegScanComponent component)
        {
            this.components[index] = component;
        }

        /// <summary>
        /// Reads JPEG scan header.
        /// </summary>
        /// <param name="reader">JPEG reader.</param>
        public override void Read(IJpegReader reader)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");

            reader.Read16();
            byte components = reader.Read8();

            this.components = new JpegScanComponent[components];

            for (int i = 0; i < this.components.Length; i++)
            {
                byte componentId = reader.Read8();
                byte dcTableId = reader.Read4();
                byte acTableId = reader.Read4();

                this.components[i] = new JpegScanComponent(componentId, dcTableId, acTableId);
            }

            this.SpectralSelectionStart = reader.Read8();
            this.SpectralSelectionEnd = reader.Read8();
            this.BitPositionHigh = reader.Read4();
            this.BitPositionLow = reader.Read4();
        }

        /// <summary>
        /// Writes JPEG scan header.
        /// </summary>
        /// <param name="writer">JPEG writer.</param>
        public override void Write(JpegWriter writer)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");

            writer.Write16(this.Length);
            writer.Write8((byte)this.NumberOfComponents);

            for (int i = 0; i < this.components.Length; i++)
            {
                JpegScanComponent component = this.components[i];
                writer.Write8(component.ComponentId);
                writer.Write4(component.DCTableId);
                writer.Write4(component.ACTableId);
            }

            writer.Write8(this.SpectralSelectionStart);
            writer.Write8(this.SpectralSelectionEnd);
            writer.Write4(this.BitPositionHigh);
            writer.Write4(this.BitPositionLow);
        }

        #endregion
    }
}
