﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Primitives;

namespace Telerik.Windows.Documents.Core.Imaging.Jpeg2000.Decoder
{
    internal class Jpeg2000HeadersDecoder
    {
        private const string SizeBoxType = "ihdr";
        private const string ColorSpaceBoxType = "colr";
        private const string Jpeg2000HeaderBoxType = "jp2h";
        private const string Jpeg2000ContentBoxType = "jp2c";
        private const short StartOfContentMarker = -177;
        private const short SizMarker = -175;
        private readonly byte[] data;
        private static readonly byte[] jpeg2000HeaderBytesStart;

        static Jpeg2000HeadersDecoder()
        {
            jpeg2000HeaderBytesStart = new byte[] { 0x00, 0x00, 0x00, 0x0C, 0x6A, 0x50, 0x20, 0x20, 0x0D, 0x0A, 0x87, 0x0A };
        }

        public Jpeg2000HeadersDecoder(byte[] data)
        {
            this.data = data;
        }

        private bool HasReadAllHeaders
        {
            get
            {
                return this.Size.HasValue && this.ColorSpace.HasValue && this.NumberOfComponents.HasValue;
            }
        }

        private bool HasReadAnyHeader
        {
            get
            {
                return this.Size.HasValue || this.ColorSpace.HasValue || this.NumberOfComponents.HasValue;
            }
        }

        private SizeI? Size
        {
            get;
            set;
        }

        private Jpeg2000ColorSpace? ColorSpace
        {
            get;
            set;
        }

        private int? NumberOfComponents
        {
            get;
            set;
        }

        public bool TryDecodeJpeg2000Image(out Jpeg2000ImageInfo imageInfo)
        {
            Jpeg2000Reader reader = new Jpeg2000Reader(this.data);

            if (!Jpeg2000HeadersDecoder.TryReadJpeg2000StartOfImageMarker(reader))
            {
                imageInfo = null;
                return false;
            }

            foreach (Box box in this.ReadBoxes(reader))
            {
                if (box.Type.Equals(Jpeg2000HeaderBoxType))
                {
                    this.ParseJpeg2000HeaderBox(box);
                }
                else if (box.Type.Equals(Jpeg2000ContentBoxType))
                {
                    this.ParseJpeg2000ContentBox(box);
                }
            }

            return this.TryCreateImageInfo(out imageInfo);
        }

        private IEnumerable<Box> ReadBoxes(Jpeg2000Reader reader)
        {
            while (!reader.EndOfFile && !this.HasReadAllHeaders)
            {
                Box box = reader.ReadBox();

                yield return box;
            }
        }

        private void ParseJpeg2000ContentBox(Box contentBox)
        {
            Jpeg2000Reader contentReader = new Jpeg2000Reader(contentBox.Content);

            short soc = contentReader.ReadBigEndianShort();
            short siz = contentReader.ReadBigEndianShort();

            bool isValidJpeg2000ContentStart = (soc == StartOfContentMarker) && (siz == SizMarker);

            if (isValidJpeg2000ContentStart)
            {
                this.ReadSizMarkerHeaderBytes(contentReader);
            }
        }

        private void ReadSizMarkerHeaderBytes(Jpeg2000Reader reader)
        {
            ushort sizMarkerSegmentLength = reader.ReadBigEndianUShort();
            ushort capabilityOfCodestream = reader.ReadBigEndianUShort();
            int imageWidth = reader.ReadBigEndianInt();
            int imageHeight = reader.ReadBigEndianInt();
            int imageOffsetX = reader.ReadBigEndianInt();
            int imageOffsetY = reader.ReadBigEndianInt();
            int tileWidth = reader.ReadBigEndianInt();
            int tileHeight = reader.ReadBigEndianInt();
            int tileOffsetX = reader.ReadBigEndianInt();
            int tileOffsetY = reader.ReadBigEndianInt();
            ushort numberOfComponents = reader.ReadBigEndianUShort();

            bool isValidSizSegment =
                sizMarkerSegmentLength > 0 &&
                capabilityOfCodestream <= 2 &&
                imageWidth >= 0 &&
                imageHeight >= 0 &&
                imageOffsetX >= 0 &&
                imageOffsetY >= 0 &&
                tileWidth >= 0 &&
                tileHeight >= 0 &&
                tileOffsetX >= 0 &&
                tileOffsetY >= 0 &&
                numberOfComponents > 0;
            
            if (isValidSizSegment)
            {
                this.Size = this.Size ?? new SizeI(imageWidth, imageHeight);
                this.NumberOfComponents = numberOfComponents;
            }
        }

        private void ParseJpeg2000HeaderBox(Box headerBox)
        {
            Jpeg2000Reader headerReader = new Jpeg2000Reader(headerBox.Content);

            foreach (Box box in this.ReadBoxes(headerReader))
            {
                if (box.Type.Equals(SizeBoxType))
                {
                    this.ParseSizeBox(box);
                }
                else if (box.Type.Equals(ColorSpaceBoxType))
                {
                    this.ParseColorSpaceBox(box);
                }
            }
        }

        private void ParseSizeBox(Box box)
        {
            Jpeg2000Reader sizeReader = new Jpeg2000Reader(box.Content);
            uint height = sizeReader.ReadBigEndianUInt();
            uint width = sizeReader.ReadBigEndianUInt();

            this.Size = new SizeI((int)width, (int)height);
        }

        private void ParseColorSpaceBox(Box box)
        {
            if (this.ColorSpace.HasValue)
            {
                // Skip unsupported complex colorspaces.
                return;
            }

            this.ColorSpace = Jpeg2000ColorSpace.Unknown;
            Jpeg2000Reader colorSpaceReader = new Jpeg2000Reader(box.Content);
            byte b = colorSpaceReader.Read();
            bool isEnumeratedColorSpaceMethod = (b == 1);
            bool isIccProfiledColorSpaceMethod = (b == 2);            

            if (isEnumeratedColorSpaceMethod)
            {
                colorSpaceReader.Read();
                colorSpaceReader.Read();
                uint colorSpace = colorSpaceReader.ReadBigEndianUInt();

                switch (colorSpace)
                {
                    case 12: // CMYK
                        this.ColorSpace = Jpeg2000ColorSpace.Cmyk;
                        break;
                    case 16: // sRgb
                    case 20: // esRGB
                    case 21: // ROMM-RGB
                        this.ColorSpace = Jpeg2000ColorSpace.Rgb;
                        break;
                    case 17: // Grayscale
                        this.ColorSpace = Jpeg2000ColorSpace.Grayscale;
                        break;
                    case 3: // YCbCr
                    case 4: // YCbCr
                    case 9: // PhotoYCC
                    case 11: // CMY
                    case 13: // YCCK
                    case 18: // sYCC
                    case 22: // YPbPr
                    case 23: // YPbPr
                    case 24: // e-sYCC
                        this.ColorSpace = Jpeg2000ColorSpace.Unknown;
                        break;
                    default:
                        this.ColorSpace = Jpeg2000ColorSpace.Unknown;
                        break;
                }
            }
            else if (isIccProfiledColorSpaceMethod)
            {
                this.ColorSpace = Jpeg2000ColorSpace.Unknown;
            }
        }

        private bool TryCreateImageInfo(out Jpeg2000ImageInfo imageInfo)
        {
            bool hasReadAnyHeader = this.HasReadAnyHeader;

            if (hasReadAnyHeader)
            {
                imageInfo = new Jpeg2000ImageInfo();

                if (this.Size.HasValue)
                {
                    imageInfo.Width = this.Size.Value.Width;
                    imageInfo.Height = this.Size.Value.Height;
                }

                if (this.ColorSpace.HasValue)
                {
                    imageInfo.ColorSpace = this.ColorSpace.Value;
                }

                if (this.NumberOfComponents.HasValue)
                {
                    imageInfo.NumberOfComponents = this.NumberOfComponents.Value;
                }
            }
            else
            {
                imageInfo = null;
            }

            return hasReadAnyHeader;
        }

        private static bool TryReadJpeg2000StartOfImageMarker(Jpeg2000Reader reader)
        {
            bool isValidHeader = reader.Length >= Jpeg2000HeadersDecoder.jpeg2000HeaderBytesStart.Length;

            if (isValidHeader)
            {
                foreach (byte headerByte in Jpeg2000HeadersDecoder.jpeg2000HeaderBytesStart)
                {
                    byte actualByte = reader.Read();

                    if (headerByte != actualByte)
                    {
                        isValidHeader = false;
                        break;
                    }
                }
            }

            return isValidHeader;
        }
    }
}
