﻿using System;
using System.Linq;
using System.Windows;
using System.Windows.Input;
using Telerik.Windows.Documents.Core.Input.PointerHandlers.Args;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Input.PointerHandlers
{
    /// <summary>
    /// Represents pointer handlers controller base class.
    /// </summary>
    public abstract class PointerHandlersControllerBase
    {
        private const string InvalidCaptureNextCallExceptionMessage = "Cannot call CaptureNext before calling RegisterLeftButtonDown or RegisterRightButtonDown!";

        private readonly PointerHandlersStack handlers;

        private PointerButtonDownArgs lastButtonDownArgs = null;
        private IPointerHandler capturedHandler = null;

        /// <summary>
        /// Initializes a new instance of the <see cref="PointerHandlersControllerBase" /> class.
        /// </summary>
        protected PointerHandlersControllerBase()
        {
            this.handlers = new PointerHandlersStack();
        }

        internal abstract Cursor DefaultCursor { get; }

        internal PointerHandlersStack Handlers
        {
            get
            {
                return this.handlers;
            }
        }

        internal IPointerHandler GetHandlerByName(string name)
        {
            return this.handlers.GetElementByName(name);
        }

        internal void Initialize()
        {
            foreach (IPointerHandler handler in this.handlers)
            {
                handler.Initialize();
            }
        }

        internal virtual Cursor GetCursor(Point position, SourceType source)
        {
            if (this.capturedHandler != null)
            {
                return this.capturedHandler.GetCursor(position, source);
            }

            IPointerHandler handler = this.GetFirstHandler(position, source);

            return handler == null ? this.DefaultCursor : handler.GetCursor(position, source);
        }

        internal virtual void RegisterRightButtonDown(PointerRightButtonDownArgs args)
        {
            foreach (IPointerHandler handler in this.handlers)
            {
                if (handler.ShouldHandle(args.Position, args.Source))
                {
                    handler.RegisterRightButtonDown(args);
                    if (args.IsHandled)
                    {
                        this.lastButtonDownArgs = args;
                        this.Capture(handler);

                        break;
                    }
                }
            }
        }

        internal virtual void RegisterRightButtonUp(PointerRightButtonUpArgs args)
        {
            if (this.capturedHandler != null)
            {
                this.capturedHandler.RegisterRightButtonUp(args);
                this.ReleaseCapture();
            }
        }

        internal virtual void RegisterLeftButtonDown(PointerLeftButtonDownArgs args)
        {
            foreach (IPointerHandler handler in this.handlers)
            {
                if (handler.ShouldHandle(args.Position, args.Source))
                {
                    handler.RegisterLeftButtonDown(args);
                    if (args.IsHandled)
                    {
                        this.lastButtonDownArgs = args;
                        this.Capture(handler);

                        break;
                    }
                }
            }
        }

        internal virtual void RegisterLeftButtonUp(PointerLeftButtonUpArgs args)
        {
            if (this.capturedHandler != null)
            {
                this.capturedHandler.RegisterLeftButtonUp(args);
                this.ReleaseCapture();
            }
        }

        internal virtual void RegisterMove(PointerMoveArgs args)
        {
            if (this.capturedHandler != null)
            {
                this.capturedHandler.RegisterMove(args);
            }
        }

        internal virtual void CaptureNext(IPointerHandler handlerToSkip)
        {
            if (this.lastButtonDownArgs == null)
            {
                throw new InvalidOperationException(PointerHandlersControllerBase.InvalidCaptureNextCallExceptionMessage);
            }

            bool shouldSkip = true;

            foreach (IPointerHandler handler in this.handlers)
            {
                if (handler == handlerToSkip)
                {
                    shouldSkip = false;

                    continue;
                }

                if (shouldSkip)
                {
                    continue;
                }
                else if (handler.ShouldHandle(this.lastButtonDownArgs.Position, this.lastButtonDownArgs.Source))
                {
                    bool isHandled = this.TryRegisterLastButtonDown(handler);
                    if (isHandled)
                    {
                        this.Capture(handler);
                        break;
                    }
                }
            }
        }

        internal virtual void RegisterLostCapture(PointerLostCaptureArgs args)
        {
            if (this.capturedHandler != null)
            {
                this.capturedHandler.RegisterLostCapture(args);
                this.capturedHandler = null;
            }
        }

        internal void Release()
        {
            foreach (IPointerHandler handler in this.handlers)
            {
                handler.Release();
            }
        }

        private void Capture(IPointerHandler handler)
        {
            Guard.ThrowExceptionIfNull(handler, "handler");
            this.capturedHandler = handler;
            this.capturedHandler.Capture(new PointerCaptureArgs());
        }

        private void ReleaseCapture()
        {
            if (this.capturedHandler != null)
            {
                this.capturedHandler.ReleaseCapture(new PointerReleaseCaptureArgs());
                this.capturedHandler = null;
            }
        }

        private IPointerHandler GetFirstHandler(Point position, SourceType source)
        {
            foreach (IPointerHandler handler in this.handlers)
            {
                if (handler.ShouldHandle(position, source))
                {
                    return handler;
                }
            }

            return null;
        }

        private bool TryRegisterLastButtonDown(IPointerHandler handler)
        {
            bool isHandled = false;

            if (this.lastButtonDownArgs.ButtonType == ButtonType.Left)
            {
                PointerLeftButtonDownArgs args = new PointerLeftButtonDownArgs()
                {
                    IsSimulated = true,
                    CtrlPressed = this.lastButtonDownArgs.CtrlPressed,
                    ShiftPressed = this.lastButtonDownArgs.ShiftPressed,
                    Source = this.lastButtonDownArgs.Source,
                    Position = this.lastButtonDownArgs.Position,
                    Clicks = this.lastButtonDownArgs.Clicks,
                };

                handler.RegisterLeftButtonDown(args);
                isHandled = args.IsHandled;
            }
            else if (this.lastButtonDownArgs.ButtonType == ButtonType.Right)
            {
                PointerRightButtonDownArgs args = new PointerRightButtonDownArgs()
                {
                    IsSimulated = true,
                    CtrlPressed = this.lastButtonDownArgs.CtrlPressed,
                    ShiftPressed = this.lastButtonDownArgs.ShiftPressed,
                    Source = this.lastButtonDownArgs.Source,
                    Position = this.lastButtonDownArgs.Position,
                    Clicks = this.lastButtonDownArgs.Clicks,
                };

                handler.RegisterRightButtonDown(args);
                isHandled = args.IsHandled;
            }

            return isHandled;
        }
    }
}
