﻿#if !NETSTANDARD
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows;
using Telerik.Windows.Documents.Core.Fonts;
using Telerik.Windows.Documents.Utilities;
#endif

namespace Telerik.Windows.Documents.Core.TextMeasurer
{
    /// <summary>
    /// Represents text measurer that can be used in multi-threaded applications.
    /// </summary>
    internal partial class RadTextMeasurer
#if !NETSTANDARD
        : ITextMeasurer
#endif
    {
        internal const int BaseFontSize = 100;

#if !NETSTANDARD
        internal static readonly Regex LigatureLanguageRegex = new Regex(IsLigatureLanguagePattern);

        private const string BaselineString = "’'`\"\\@#$%^&*()+=-_/|:].,!~abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890";
        private const string IsLigatureLanguagePattern = @"[\p{IsDevanagari}]";

        private static readonly Dictionary<string, Size> letterSizeCache;
        private static readonly Dictionary<string, Dictionary<string, double>> kerningPairToFontKerningValue;
        private static readonly Dictionary<string, double> fontBaseLineOffsetCache;

        static RadTextMeasurer()
        {
            letterSizeCache = new Dictionary<string, Size>();
            kerningPairToFontKerningValue = new Dictionary<string, Dictionary<string, double>>();
            fontBaseLineOffsetCache = new Dictionary<string, double>();
        }

        /// <summary>
        /// Measures the text.
        /// </summary>
        /// <param name="textProperties"></param>
        /// <param name="fontProperties">The font properties.</param>
        /// <returns></returns>
        public TextMeasurementInfo MeasureText(TextProperties textProperties, FontProperties fontProperties)
        {
            double width = 0;
            double ratio = textProperties.Size / BaseFontSize;

            if (textProperties.FlowDirection == FlowDirection.RightToLeft || LigatureLanguageRegex.IsMatch(textProperties.Text))
            {
                width = MeasureWholeTextWidth(textProperties, fontProperties);
            }
            else
            {
                width = MeasureTextWidth(textProperties, fontProperties) * ratio;
            }

            TextMeasurementInfo info = new TextMeasurementInfo();
            info.Size = new Size(width, GetLineHeight(fontProperties) * ratio);
            info.BaselineOffset = GetBaseLineOffset(fontProperties, textProperties.FlowDirection) * info.Size.Height;

            return info;
        }

        internal static double GetPairWidthKerningError(string pair, FontProperties fontProperties)
        {
            string font = fontProperties.FontFamily.Source;
            if (!SystemFontsManager.HasKerning(font))
            {
                return 0;
            }

            double kerning = 0;
            Dictionary<string, double> fontToKerningValue;

            lock (kerningPairToFontKerningValue)
            {
                if (!kerningPairToFontKerningValue.TryGetValue(pair, out fontToKerningValue))
                {
                    if (fontToKerningValue == null)
                    {
                        fontToKerningValue = new Dictionary<string, double>();

                        kerningPairToFontKerningValue[pair] = fontToKerningValue;
                    }
                }
            }

            StringBuilder builder = new StringBuilder();
            builder.Append(fontProperties.FontFamily.Source);
            builder.Append(fontProperties.FontWeight);
            builder.Append(fontProperties.FontStyle);

            string key = builder.ToString();
            lock (fontToKerningValue)
            {
                if (!fontToKerningValue.TryGetValue(key, out kerning))
                {
                    kerning = CalculateKerning(pair, fontProperties);
                    fontToKerningValue.Add(key, kerning);
                }
            }

            return kerning;
        }

        private static double MeasureTextWidth(TextProperties textProperties, FontProperties fontProperties)
        {
            string text = textProperties.Text;
            double width = 0;
            // TODO: Optimization for monospaced fonts.
            //if (fontProperties.IsMonoSpaced)
            //{
            //    Size symbolSize = new Size(0, 0);
            //    double tabWidth = GetSymbolSize('\t', fontProperties).Width;
            //    symbolSize = GetSymbolSize(DocumentsHelper.LineHeightMeasureSymbol, fontProperties);

            //    int tabsCount = GetTabsCount(text);
            //    width = (tabsCount * tabWidth) + ((text.Length - tabsCount) * symbolSize.Width);
            //}
            //else
            {
                Size symbolSize = new Size(0, 0);
                for (int i = 0; i < text.Length; i++)
                {
                    if (i + 1 < text.Length)
                    {
                        string pair = text[i].ToString() + text[i + 1];
                        double kerningError = GetPairWidthKerningError(pair, fontProperties);
                        width += kerningError;
                    }

                    symbolSize = GetSymbolSize(text[i], fontProperties);
                    width += symbolSize.Width;
                }
            }

            return width;
        }

        private static double GetLineHeight(FontProperties fontProperties)
        {
            Size res = GetSymbolSize(DocumentsHelper.LineHeightMeasureSymbol, fontProperties);

            return res.Height;
        }

        private static double CalculateKerning(string pair, FontProperties fontProperties)
        {
            double firstCharWidth = GetSymbolSize(pair[0], fontProperties).Width;
            double secontCharWidth = GetSymbolSize(pair[1], fontProperties).Width;
#if SILVERLIGHT
            string text = pair;
#elif WPF
            string text = DocumentsHelper.EscapeWhiteSpaces(pair);
#endif
            double pairWidth = MeasureTextSize(text, fontProperties).Width;
            double error = pairWidth - (firstCharWidth + secontCharWidth);

            return error;
        }

        private static Size GetSymbolSize(char symbol, FontProperties fontProperties)
        {
            Size symbolSize;
            StringBuilder builder = new StringBuilder();
            builder.Append(symbol);
            builder.Append(fontProperties.FontFamily.Source);
            builder.Append(fontProperties.FontWeight);
            builder.Append(fontProperties.FontStyle);
            string key = builder.ToString();

            lock (letterSizeCache)
            {
                if (!letterSizeCache.TryGetValue(key, out symbolSize))
                {
#if SILVERLIGHT
                    string text = symbol.ToString();
#elif WPF
                    string text = DocumentsHelper.EscapeWhiteSpaces(symbol.ToString());
#endif
                    symbolSize = MeasureTextSize(text, fontProperties);
                    letterSizeCache[key] = symbolSize;
                }
            }

            return symbolSize;
        }
#endif
    }
}