﻿using System;
using System.Collections.Generic;
#if WPF
using System.Drawing.Text;
#endif
using System.Globalization;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text.RegularExpressions;
#if !NETSTANDARD
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Fonts
{
    internal static class SystemFontsManager
    {
        private static readonly HashSet<string> fontsWithKerning; // = new HashSet<string> { "Arial", "Calibri", "Times New Roman" };
        private static readonly HashSet<string> monospacedFonts; // = new HashSet<string> { "Consolas", "Courier New", "Lucida Console" };
        private static readonly List<string> systemFonts;

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
        static SystemFontsManager()
        {
            fontsWithKerning = new HashSet<string>();
            monospacedFonts = new HashSet<string>();
            Dictionary<string, string> fileNameToFontName = new Dictionary<string, string>();
            string manifestResourceName = string.Empty;

#if NETSTANDARD
            manifestResourceName = "Telerik.Documents.Core.Core.TextMeasurer.FontsMetaInfo.Fonts.meta";
#else
            manifestResourceName = "Telerik.Windows.Documents.Core.TextMeasurer.FontsMetaInfo.Fonts.meta";
#endif
            using (var stream = Assembly.GetExecutingAssembly().GetManifestResourceStream(manifestResourceName))
            using (StreamReader reader = new StreamReader(stream))
            {
                // Lines format: Name;IsMonospaced;HasKerning;FileName1;FileName2;...;
                string line;
                while ((line = reader.ReadLine()) != null)
                {
                    // Skip comments.
                    if (line.StartsWith("//"))
                    {
                        continue;
                    }

                    var parts = line.Split(new char[] { ';' }, StringSplitOptions.RemoveEmptyEntries);
                    var fontName = parts[0];

                    if (parts[1] == "1")
                    {
                        monospacedFonts.Add(fontName);
                    }

                    if (parts[2] == "1")
                    {
                        fontsWithKerning.Add(fontName);
                    }

                    for (int i = 3; i < parts.Length; i++)
                    {
                        fileNameToFontName[parts[i].ToLower(CultureInfo.InvariantCulture)] = fontName;
                    }
                }
            }

            HashSet<string> systemFontsSet = new HashSet<string>();

#if !NETSTANDARD
#if SILVERLIGHT
            foreach (var typeface in System.Windows.Media.Fonts.SystemTypefaces)
            {
                GlyphTypeface glyphTypeface;
                try
                {
                    if (typeface.TryGetGlyphTypeface(out glyphTypeface))
                    {
                        string fileName = SanitizeFontFileName(Path.GetFileName(glyphTypeface.GetFontFileName()));
                        string fontName;

                        if (fileNameToFontName.TryGetValue(fileName, out fontName))
                        {
                            systemFontsSet.Add(fontName);
                        }
                    }
                }
                catch
                {
                }
            }
#else
            Action<FontFamily> addSystemFont = fontFamily =>
            {
                foreach (Typeface typeface in fontFamily.GetTypefaces())
                {
                    GlyphTypeface glyphTypeface;
                    try
                    {
                        if (typeface.TryGetGlyphTypeface(out glyphTypeface))
                        {
                            string fileName = SanitizeFontFileName(Path.GetFileName(glyphTypeface.GetFontFileName()));
                            string fontName;

                            if (fileNameToFontName.TryGetValue(fileName, out fontName))
                            {
                                systemFontsSet.Add(fontName);
                            }
                        }
                    }
                    catch
                    {
                    }
                }
            };

            var systemFontsList = System.Windows.Media.Fonts.SystemFontFamilies as IList<FontFamily>;

            if (systemFontsList != null)
            {
                for (int i = 0; i < systemFontsList.Count; i++)
                {
                    try
                    {
                        addSystemFont(systemFontsList[i]);
                    }
                    catch
                    {
                        continue;
                    }
                }
            }
            else
            {
                foreach (var family in System.Windows.Media.Fonts.SystemFontFamilies)
                {
                    addSystemFont(family);
                }
            }
#endif
#endif

#if WPF
            using (InstalledFontCollection fontsCollection = new InstalledFontCollection())
            {
                foreach (var font in fontsCollection.Families)
                {
                    if (!systemFontsSet.Contains(font.Name))
                    {
                        systemFontsSet.Add(font.Name);
                    }
                }
            }
#endif

            systemFonts = new List<string>(systemFontsSet);
            systemFonts.Sort();
        }

        public static IEnumerable<string> GetAvailableFonts()
        {
            return systemFonts;
        }

        public static bool HasKerning(string fontName)
        {
            return fontsWithKerning.Contains(fontName);
        }

        public static bool IsMonospaced(string fontFamily)
        {
            if (fontFamily == null)
            {
                return false;
            }

            var fontNames = fontFamily.Split(',');
            foreach (var fontName in fontNames)
            {
                if (!monospacedFonts.Contains(fontName.Trim()))
                {
                    return false;
                }
            }

            return true;
        }

        /// <summary>
        /// This method will trigger the initial load of system fonts.
        /// </summary>
        internal static void WarmUp()
        {
        }

        private static string SanitizeFontFileName(string fileName)
        {
            fileName = fileName.ToLower(CultureInfo.InvariantCulture);

            return Regex.Replace(Path.GetFileNameWithoutExtension(fileName), "_[0-9]+$", string.Empty) + Path.GetExtension(fileName);
        }
    }
}
