﻿using Telerik.Windows.Documents.Utilities;

namespace Telerik.Documents.Media
{
    /// <summary>
    /// Represents color object.
    /// </summary>
    public struct Color
    {
        /// <summary>
        /// Gets or sets the alpha component.
        /// </summary>
        /// <value>The alpha component.</value>
        public byte A { get; set; }

        /// <summary>
        /// Gets or sets the red component.
        /// </summary>
        /// <value>The red component.</value>
        public byte R { get; set; }

        /// <summary>
        /// Gets or sets the green component.
        /// </summary>
        /// <value>The green component.</value>
        public byte G { get; set; }

        /// <summary>
        /// Gets or sets the blue component.
        /// </summary>
        /// <value>The blue component.</value>
        public byte B { get; set; }

        /// <summary>
        /// Creates new color from Argb.
        /// </summary>
        /// <param name="alfa">The alfa.</param>
        /// <param name="red">The red.</param>
        /// <param name="green">The green.</param>
        /// <param name="blue">The blue.</param>
        /// <returns></returns>
        public static Color FromArgb(byte alfa, byte red, byte green, byte blue)
        {
            return new Color() { A = alfa, R = red, G = green, B = blue };
        }

        /// <summary>
        /// Creates new color from RGB.
        /// </summary>
        /// <param name="red">The red.</param>
        /// <param name="green">The green.</param>
        /// <param name="blue">The blue.</param>
        /// <returns></returns>
        public static Color FromRgb(byte red, byte green, byte blue)
        {
            return FromArgb(255, red, green, blue);
        }

        internal static Color FromUInt32(uint argb)
        {
            Color color = new Color();
            color.A = (byte)((argb & -16777216) >> 24);
            color.R = (byte)((argb & 16711680) >> 16);
            color.G = (byte)((argb & 65280) >> 8);
            color.B = (byte)(argb & 255);
            return color;
        }

        /// <summary>
        /// Compares two color instances.
        /// </summary>
        /// <param name="color1">First color.</param>
        /// <param name="color2">Second color.</param>
        /// <returns>If two colors are equal.</returns>
        public static bool operator ==(Color color1, Color color2)
        {
            return color1.Equals(color2);
        }

        /// <summary>
        /// Compares two color instances.
        /// </summary>
        /// <param name="color1">First color.</param>
        /// <param name="color2">Second color.</param>
        /// <returns>If two colors are not equal.</returns>
        public static bool operator !=(Color color1, Color color2)
        {
            return !(color1 == color2);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return ObjectExtensions.CombineHashCodes(this.A, this.R, this.G, this.B);
        }

        /// <summary>
        /// Determines whether the specified <see cref="Color"/> is equal to the current object.
        /// </summary>
        /// <param name="value">The color.</param>
        public bool Equals(Color value)
        {
            return this.A == value.A &&
                   this.R == value.R &&
                   this.G == value.G &&
                   this.B == value.B;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// <c>true</c>if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            if (obj != null && obj is Color)
            {
                return this.Equals((Color)obj);
            }

            return false;
        }
    }
}
