﻿using System;
using System.Collections.Generic;
using System.Globalization;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Theming;
#else
using Telerik.Documents.Common.Model;
#endif
#if NETSTANDARD
using Telerik.Documents.Media;
#else
using System.Windows.Media;
#endif

namespace Telerik.Windows.Documents.Media
{
    internal class ColorsHelper
    {
        private static readonly List<TintValues> tintValuesList = new List<TintValues>
        {
            new TintValues(0, 0, new double[] { 0.5, 0.35, 0.25, 0.15, 0.05 }),
            new TintValues(1, 19, new double[] { 0.9, 0.75, 0.5, 0.25, 0.1 }),
            new TintValues(20, 79, new double[] { 0.80, 0.60, 0.40, -0.25, -0.5 }),
            new TintValues(80, 99, new double[] { -0.10, -0.25, -0.50, -0.75, -0.9 }),
            new TintValues(100, 100, new double[] { -0.05, -0.15, -0.25, -0.35, -0.5 }),
        };

        public static Color HexStringToColor(string hexColor)
        {
            // Remove the # at the front.
            if (hexColor.StartsWith("#"))
            {
                hexColor = hexColor.Substring(1, hexColor.Length - 1);
            }

            byte a = 255;
            byte r = 255;
            byte g = 255;
            byte b = 255;

            int start = 0;

            // Handle ARGB strings (8 characters long).
            if (hexColor.Length == 8)
            {
                a = byte.Parse(hexColor.Substring(0, 2), NumberStyles.HexNumber);
                start = 2;
            }

            //Handle contracted RGB strings (3 characters long)
            if (hexColor.Length == 3)
            {
                hexColor = string.Format("{0}{0}{1}{1}{2}{2}", hexColor[0], hexColor[1], hexColor[2]);
            }

            // Convert RGB characters to bytes.
            r = byte.Parse(hexColor.Substring(start, 2), NumberStyles.HexNumber);
            g = byte.Parse(hexColor.Substring(start + 2, 2), NumberStyles.HexNumber);
            b = byte.Parse(hexColor.Substring(start + 4, 2), NumberStyles.HexNumber);

            return Color.FromArgb(a, r, g, b);
        }

#if !NETSTANDARD
        public static ColorShadeType? GetColorShadeTypeForColorAndTint(Color color, double tint)
        {
            HlsaColor hlsColor = HlsaColor.FromColor(color);
            double[] array = GetTintValuesForHlsaColor(hlsColor);

            double roundedTint = Math.Round(tint, 2);
            for (int i = 0; i < array.Length; i++)
            {
                double t = Math.Round(array[i], 2);
                if (AreEqual(t, roundedTint))
                {
                    return (ColorShadeType)i;
                }
            }

            return null;
        }
#endif

        public static double GetTintAndShadeForColorAndIndex(Color color, ColorShadeType colorShadeType)
        {
            HlsaColor hlsColor = HlsaColor.FromColor(color);
            double[] array = GetTintValuesForHlsaColor(hlsColor);

            return array[(int)colorShadeType];
        }

        public static Color UpdateTint(Color color, double tint)
        {
            if (tint != 0)
            {
                HlsaColor hlsColor = HlsaColor.FromColor(color);
                int colorLuminosity = (int)(hlsColor.Luminance * 240);
                if (tint <= 0f)
                {
                    double luminosity = colorLuminosity * -tint;

                    return ColorFromHlsa(hlsColor.Hue, (colorLuminosity - luminosity) / 240, hlsColor.Saturation);
                }
                else
                {
                    int num = 240 - colorLuminosity;

                    return ColorFromHlsa(hlsColor.Hue, (colorLuminosity + (num * tint)) / 240, hlsColor.Saturation);
                }
            }
            else
            {
                return color;
            }
        }

        public static Color ColorFromHlsa(double hue, double luminosity, double saturation)
        {
            return HlsaColor.ToColor(new HlsaColor(hue, luminosity, saturation, 1));
        }

        public static Color FromGray(byte gray)
        {
            return Color.FromArgb(255, gray, gray, gray);
        }

        public static Color FromCmyk(byte cyan, byte magenta, byte yellow, byte black)
        {
            byte r = (byte)(255 - Math.Min(255, cyan + black));
            byte g = (byte)(255 - Math.Min(255, magenta + black));
            byte b = (byte)(255 - Math.Min(255, yellow + black));

            return Color.FromArgb(255, r, g, b);
        }

#if !NETSTANDARD
        public static void GenerateThemeGradients(List<Color> themeColors,
            out List<Color> mainPaletteColors, out List<ThemableColor> mainPaletteThemableColors)
        {
            mainPaletteColors = new List<Color>();
            mainPaletteThemableColors = new List<ThemableColor>();

            for (int themeColorIndex = 0; themeColorIndex < themeColors.Count; themeColorIndex++)
            {
                Color color = themeColors[themeColorIndex];
                HlsaColor hlsColor = HlsaColor.FromColor(color);

                double[] themeColorTintValues = GetTintValuesForHlsaColor(hlsColor);
                for (int i = 0; i < themeColorTintValues.Length; i++)
                {
                    double tintAndShade = themeColorTintValues[i];
                    Color modifiedColor = UpdateTint(color, (float)tintAndShade);
                    mainPaletteColors.Add(modifiedColor);

                    mainPaletteThemableColors.Add(new ThemableColor((ThemeColorType)themeColorIndex, (ColorShadeType)i));
                }
            }
        }
#endif

        public static int ToPixel(Color color)
        {
#if WPF || NETSTANDARD
            return color.A << 24 | color.R << 16 | color.G << 8 | color.B;
#elif SILVERLIGHT
            var a = color.A;
            var ai = a + 1;

            return (a << 24) | ((byte)((color.R * ai) >> 8) << 16) | ((byte)((color.G * ai) >> 8) << 8) | ((byte)((color.B * ai) >> 8));
#endif
        }

        public static int GetBrightness(Color color)
        {
            return (int)Math.Sqrt(
               (color.R * color.R * .241) +
               (color.G * color.G * .691) +
               (color.B * color.B * .068));
        }

        public static void GetLuminanceModulationAndOffsetFromTint(double tint, out double lumModulation, out double lumOffset)
        {
            if (tint < 0)
            {
                lumModulation = (1 + tint);
                lumOffset = 0;
            }
            else
            {
                lumModulation = (1 - tint);
                lumOffset = tint;
            }
        }

        public static double GetTintFromLuminanceModulationAndOffset(double lumModulation, double lumOffset)
        {
            double tint;

            if (lumOffset != 0)
            {
                tint = lumOffset;
            }
            else
            {
                tint = lumModulation - 1;
            }

            return tint;
        }

        private static bool AreEqual(double f1, double f2)
        {
            return Math.Abs(f1 - f2) < 0.001;
        }

        private static double[] GetTintValuesForHlsaColor(HlsaColor color)
        {
            int luminosityInPercents = (int)(color.Luminance * 100);

            foreach (TintValues values in tintValuesList)
            {
                if (luminosityInPercents >= values.Start &&
                    luminosityInPercents <= values.End)
                {
                    return values.TintValuesArray;
                }
            }

            return null;
        }

        private class TintValues
        {
            private readonly int start;
            private readonly int end;
            private readonly double[] tintValues;

            public TintValues(int start, int end, double[] tintValues)
            {
                this.start = start;
                this.end = end;
                this.tintValues = tintValues;
            }

            public int Start
            {
                get
                {
                    return this.start;
                }
            }

            public int End
            {
                get
                {
                    return this.end;
                }
            }

            public double[] TintValuesArray
            {
                get
                {
                    return this.tintValues;
                }
            }
        }
    }
}
