﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Utilities;
#if !NETSTANDARD
using System.Windows.Media;
#else
using Telerik.Documents.Media;
#endif

namespace Telerik.Windows.Documents.Media
{
    internal class HlsaColor
    {
        private const double OneSixth = 1.0 / 6.0;
        private const double OneThird = 1.0 / 3.0;
        private const double TwoThird = 2.0 / 3.0;

        private readonly double alpha;
        private readonly double hue;
        private readonly double luminance;
        private readonly double saturation;

        public HlsaColor(double hue, double luminance, double saturation, double alpha)
        {
            this.hue = MathUtilities.Clamp(hue, 0.0, 360.0, 5);
            this.luminance = MathUtilities.Clamp(luminance, 0.0, 1.0, 5);
            this.saturation = MathUtilities.Clamp(saturation, 0.0, 1.0, 5);
            this.alpha = MathUtilities.Clamp(alpha, 0.0, 1.0, 5);
        }

        public double Alpha
        {
            get
            {
                return this.alpha;
            }
        }

        public double Hue
        {
            get
            {
                return this.hue;
            }
        }

        public double Luminance
        {
            get
            {
                return this.luminance;
            }
        }

        public double Saturation
        {
            get
            {
                return this.saturation;
            }
        }

        public static HlsaColor FromColor(Color color)
        {
            double r = color.R / 255.0;
            double g = color.G / 255.0;
            double b = color.B / 255.0;
            double a = color.A;
            double maxc = Math.Max(r, Math.Max(g, b));
            double minc = Math.Min(r, Math.Min(g, b));
            double l = (minc + maxc) / 2.0;
            double s = 0;
            double rc = 0, gc = 0, bc = 0;
            double h;
            
            if (minc == maxc)
            {
                return new HlsaColor(0.0, l, 0.0, a);
            }
         
            if (l <= 0.5)
            {
                s = (maxc - minc) / (maxc + minc);
            }
            else
            {
                s = (maxc - minc) / (2.0 - maxc - minc);
            }

            rc = (maxc - r) / (maxc - minc);
            gc = (maxc - g) / (maxc - minc);
            bc = (maxc - b) / (maxc - minc);
            
            if (r == maxc)
            {
                h = bc - gc;
            }
            else if (g == maxc)
            {
                h = 2.0 + rc - bc;
            }
            else
            {
                h = 4.0 + gc - rc;
            }
            
            h = h / 6.0;
            h = MathUtilities.CustomMod(h);

            return new HlsaColor(h * 360, l, s, a);
        }

        public static Color ToColor(HlsaColor color)
        {
            double h = color.Hue / 360;
            double l = color.Luminance;
            double s = color.Saturation;
            double a = color.Alpha;
            double m2 = 0, m1 = 0;

            if (s == 0.0)
            {
                return Color.FromArgb(
                    (byte)MathUtilities.Clamp(a * 255, 0.0, 255, 5),
                    (byte)MathUtilities.Clamp(l * 255, 0.0, 255, 5),
                    (byte)MathUtilities.Clamp(l * 255, 0.0, 255, 5),
                    (byte)MathUtilities.Clamp(l * 255, 0.0, 255, 5));
            }

            if (l <= 0.5)
            {
                m2 = l * (1.0 + s);
            }
            else
            {
                m2 = l + s - (l * s);
            }

            m1 = (2.0 * l) - m2;

            double r = ConvertValue(m1, m2, h + OneThird);
            double g = ConvertValue(m1, m2, h);
            double b = ConvertValue(m1, m2, h - OneThird);

            return Color.FromArgb(
                (byte)MathUtilities.Clamp(a * 255, 0.0, 255, 5),
                (byte)MathUtilities.Clamp(r * 255, 0.0, 255, 5),
                (byte)MathUtilities.Clamp(g * 255, 0.0, 255, 5),
                (byte)MathUtilities.Clamp(b * 255, 0.0, 255, 5));
        }

        public static bool operator ==(HlsaColor instance1, HlsaColor instance2)
        {
            if (object.ReferenceEquals(instance1, instance2))
            {
                return true;
            }

            if (instance1 as object == null || instance2 as object == null)
            {
                return false;
            }

            return AreEqual(instance1, instance2);
        }

        public static bool operator !=(HlsaColor instance1, HlsaColor instance2)
        {
            return !(instance1 == instance2);
        }

        public override bool Equals(object obj)
        {
            var other = obj as HlsaColor;

            if (obj == null || other == null)
            {
                return false;
            }

            return AreEqual(this, other);
        }

        public bool Equals(HlsaColor otherInstance)
        {
            return AreEqual(this, otherInstance);
        }

        public override int GetHashCode()
        {
            return ObjectExtensions.CombineHashCodes((int)this.Hue, (int)this.Luminance, (int)this.Saturation, (int)this.Alpha);
        }

        private static bool AreEqual(HlsaColor instance1, HlsaColor instance2)
        {
            if (instance1 == null || instance2 == null)
            {
                return false;
            }

            return
                instance1.Hue == instance2.Hue &&
                instance1.Luminance == instance2.Luminance &&
                instance1.Saturation == instance2.Saturation &&
                instance1.Alpha == instance2.Alpha;
        }

        private static double ConvertValue(double m1, double m2, double hue)
        {
            hue = MathUtilities.CustomMod(hue);

            if (hue < OneSixth)
            {
                return m1 + (((m2 - m1) * hue) * 6.0);
            }

            if (hue < 0.5)
            {
                return m2;
            }

            if (hue < TwoThird)
            {
                return m1 + (((m2 - m1) * (TwoThird - hue)) * 6.0);
            }

            return m1;
        }
    }
}
