﻿using System.IO;
using Telerik.Windows.Documents.Common.Model.Data;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Media
{
    /// <summary>
    /// Encapsulates data needed for creation of an image
    /// </summary>
    public class ImageSource : IResource
    {
        private const string NameFormat = "image{0}.{1}";

        private readonly byte[] data;
        private readonly string extension;

        /// <summary>
        /// Initializes a new instance of the <see cref="ImageSource" /> class.
        /// </summary>
        /// <param name="stream">The stream which represents the image.</param>
        /// <param name="extension">The format of the image.</param>
        public ImageSource(Stream stream, string extension)
            : this(extension)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");

            this.data = new byte[stream.Length];
            if (stream.CanSeek)
            {
                stream.Seek(0, SeekOrigin.Begin);
            }

            stream.Read(this.data, 0, this.data.Length);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ImageSource" /> class.
        /// </summary>
        /// <param name="data">Byte array containing representation of the image source.</param> 
        /// <param name="extension">The image file extension.</param>
        public ImageSource(byte[] data, string extension)
            : this(extension)
        {
            Guard.ThrowExceptionIfNull(data, "data");

            this.data = data;
        }

        internal ImageSource()
        {
        }

        private ImageSource(string extension)
        {
            Guard.ThrowExceptionIfNullOrEmpty(extension, "extension");

            this.extension = PathExtension.StripExtension(extension).ToLowerInvariant();
        }

        string IResource.Extension
        {
            get
            {
                return this.Extension;
            }
        }

        /// <summary>
        /// Gets the image file extension.
        /// </summary>
        /// <value>The image file extension.</value>
        public virtual string Extension
        {
            get
            {
                return this.extension;
            }
        }

        byte[] IResource.Data
        {
            get
            {
                return this.Data;
            }
        }

        /// <summary>
        /// Gets the byte array representation of the image source.
        /// </summary>
        /// <value>The byte array representation of the image source.</value>
        public virtual byte[] Data
        {
            get
            {
                return this.data;
            }
        }

        internal int Id
        {
            get
            {
                return ((IResource)this).Id;
            }
        }

        string IResource.Name
        {
            get
            {
                return string.Format(NameFormat, ((IResource)this).Id, this.Extension);
            }
        }

        int IResource.Id { get; set; }

        ResourceManager IResource.Owner { get; set; }

        internal virtual ImageSource Clone()
        {
            byte[] clonedData = (byte[])this.Data.Clone();
            ImageSource imageSource = new ImageSource(clonedData, this.Extension);

            return imageSource;
        }
    }
}