﻿using System;
using System.Linq;

namespace Telerik.Windows.Documents.Media
{
    /// <summary>
    /// Contains methods for converting DPI(Device Independent Pixels) to other unit types.
    /// </summary>
    public static class Unit
    {
        private const double DefaultDpi = 96.0;
        private const double CmToInch = 2.54;
        private const double PointToInch = 72.0;
        private const double PicaToInch = 6.0;
        private const double TwipToInch = 1440.0;
        private const double EmuToInch = 914400.0;

        /// <summary>
        /// Converts dips to points.
        /// </summary>
        /// <param name="value">Pixels.</param>
        /// <returns>Points.</returns>
        public static double DipToPoint(double value)
        {
            return (value * Unit.PointToInch) / Unit.DefaultDpi;
        }

        /// <summary>
        /// Converts dips to points.
        /// </summary>
        /// <param name="pixels">Pixels.</param>
        /// <returns>Points.</returns>
        public static int DipToPointI(double pixels)
        {
            return (int)Math.Round(Unit.DipToPoint(pixels));
        }

        /// <summary>
        /// Converts dips to picas.
        /// </summary>
        /// <param name="value">Pixels.</param>
        /// <returns>Picas.</returns>
        public static double DipToPica(double value)
        {
            return (value * Unit.PicaToInch) / Unit.DefaultDpi;
        }

        /// <summary>
        /// Converts dips to centimeters.
        /// </summary>
        /// <param name="value">Pixels.</param>
        /// <returns>Centimeters.</returns>
        public static double DipToCm(double value)
        {
            return (value * Unit.CmToInch) / Unit.DefaultDpi;
        }

        /// <summary>
        /// Converts dips to millimeters.
        /// </summary>
        /// <param name="value">Pixels.</param>
        /// <returns>Millimeters.</returns>
        public static double DipToMm(double value)
        {
            return DipToCm(value * 10);
        }

        /// <summary>
        /// Converts dips to inches.
        /// </summary>
        /// <param name="value">Pixels.</param>
        /// <returns>Inches.</returns>
        public static double DipToInch(double value)
        {
            return value / Unit.DefaultDpi;
        }

        /// <summary>
        /// Converts dips to twips.
        /// </summary>
        /// <param name="value">Pixels.</param>
        /// <returns>Twips.</returns>
        public static double DipToTwip(double value)
        {
            return (value * Unit.TwipToInch) / Unit.DefaultDpi;
        }

        /// <summary>
        /// Converts dips to EMUs.
        /// </summary>
        /// <param name="value">Pixels.</param>
        /// <returns>EMUs.</returns>
        public static double DipToEmu(double value)
        {
            return (value * Unit.EmuToInch) / Unit.DefaultDpi;
        }

        /// <summary>
        /// Converts dips to EMUs.
        /// </summary>
        /// <param name="value">Pixels.</param>
        /// <returns>EMUs.</returns>
        public static int DipToEmuI(double value)
        {
            return (int)Math.Round((value * Unit.EmuToInch) / Unit.DefaultDpi);
        }

        /// <summary>
        /// Converts dips to twips.
        /// </summary>
        /// <param name="value">Pixels.</param>
        /// <returns>Twips.</returns>
        public static int DipToTwipI(double value)
        {
            return (int)Math.Round(Unit.DipToTwip(value));
        }

        /// <summary>
        /// Converts dips to twips.
        /// </summary>
        /// <param name="value">Pixels.</param>
        /// <returns>Twips.</returns>
        public static float DipToTwipF(double value)
        {
            return (float)Unit.DipToTwip(value);
        }

        /// <summary>
        /// Converts twips to dips.
        /// </summary>
        /// <param name="value">Twips.</param>
        /// <returns>Dips.</returns>
        public static float TwipToDipF(double value)
        {
            return (float)Unit.TwipToDip(value);
        }

        /// <summary>
        /// Converts twips to dips.
        /// </summary>
        /// <param name="value">Twips.</param>
        /// <returns>Dips.</returns>
        public static int TwipToDipI(double value)
        {
            return (int)Unit.TwipToDip(value);
        }

        /// <summary>
        /// Converts DIPs to units.
        /// </summary>
        /// <param name="value">Pixels.</param>
        /// <param name="type">UnitType.</param>
        /// <returns>Units.</returns>
        public static double DipToUnit(double value, UnitType type)
        {
            if (type == UnitType.Cm)
            {
                return Unit.DipToCm(value);
            }
            else if (type == UnitType.Inch)
            {
                return Unit.DipToInch(value);
            }
            else if (type == UnitType.Mm)
            {
                return Unit.DipToMm(value);
            }
            else if (type == UnitType.Pica)
            {
                return Unit.DipToPica(value);
            }
            else if (type == UnitType.Point)
            {
                return Unit.DipToPoint(value);
            }
            else if (type == UnitType.Twip)
            {
                return Unit.DipToTwip(value);
            }
            else if (type == UnitType.Emu)
            {
                return Unit.DipToEmu(value);
            }
            else if (type == UnitType.Dip)
            {
                return value;
            }

            throw new ArgumentException("Unknown UnitType");
        }

        /// <summary>
        /// Converts to points dips.
        /// </summary>
        /// <param name="value">Points.</param>
        /// <returns>Pixels.</returns>
        public static double PointToDip(double value)
        {
            return (value * Unit.DefaultDpi) / Unit.PointToInch;
        }

        /// <summary>
        /// Converts to points dips.
        /// </summary>
        /// <param name="value">Picas.</param>
        /// <returns>Pixels.</returns>
        public static double PicaToDip(double value)
        {
            return (value * Unit.DefaultDpi) / Unit.PicaToInch;
        }

        /// <summary>
        /// Converts EMUs to dips.
        /// </summary>
        /// <param name="value">EMUs.</param>
        /// <returns>Pixels.</returns>
        public static double EmuToDip(double value)
        {
            return Math.Round((value * Unit.DefaultDpi) / Unit.EmuToInch, 2);
        }

        /// <summary>
        /// Converts centimeters to dips.
        /// </summary>
        /// <param name="value">Centimeters.</param>
        /// <returns>Pixels.</returns>
        public static double CmToDip(double value)
        {
            return (value * Unit.DefaultDpi) / Unit.CmToInch;
        }

        /// <summary>
        /// Converts millimeters to dips.
        /// </summary>
        /// <param name="value">Millimeters.</param>
        /// <returns>Pixels.</returns>
        public static double MmToDip(double value)
        {
            return CmToDip(value / 10);
        }

        /// <summary>
        /// Converts inches to dips.
        /// </summary>
        /// <param name="value">Inches.</param>
        /// <returns>Pixels.</returns>
        public static double InchToDip(double value)
        {
            return value * Unit.DefaultDpi;
        }

        /// <summary>
        /// Converts twips to dips.
        /// </summary>
        /// <param name="value">Twips.</param>
        /// <returns>Pixels.</returns>
        public static double TwipToDip(double value)
        {
            return (value * Unit.DefaultDpi) / Unit.TwipToInch;
        }

        /// <summary>
        /// Converts Units to dips.
        /// </summary>
        /// <param name="value">Units.</param>
        /// <param name="type">UnitType.</param>
        /// <returns>Pixels.</returns>
        public static double UnitToDip(double value, UnitType type)
        {
            if (type == UnitType.Cm)
            {
                return Unit.CmToDip(value);
            }
            else if (type == UnitType.Inch)
            {
                return Unit.InchToDip(value);
            }
            else if (type == UnitType.Mm)
            {
                return Unit.MmToDip(value);
            }
            else if (type == UnitType.Pica)
            {
                return Unit.PicaToDip(value);
            }
            else if (type == UnitType.Point)
            {
                return Unit.PointToDip(value);
            }
            else if (type == UnitType.Twip)
            {
                return Unit.TwipToDip(value);
            }
            else if (type == UnitType.Emu)
            {
                return Unit.EmuToDip(value);
            }
            else if (type == UnitType.Dip)
            {
                return value;
            }

            throw new ArgumentException("Unknown UnitType");
        }

        /// <summary>
        /// Converts pixels to units of measurement.
        /// </summary>
        /// <param name="basePixelSize">Base pixel size.</param>
        /// <param name="value">Pixels.</param>
        /// <returns>Ems.</returns>
        public static double PixelToEm(double basePixelSize, double value)
        {
            return value / basePixelSize;
        }

        /// <summary>
        /// Converts units of measurement  to pixels.
        /// </summary>
        /// <param name="basePixelSize">Base pixel size.</param>
        /// <param name="value">Ems.</param>
        /// <returns>Pixels.</returns>
        public static double EmToPixel(double basePixelSize, double value)
        {
            return basePixelSize * value;
        }

        /// <summary>
        /// Converts pixels to percents.
        /// </summary>
        /// <param name="basePixelSize">Base pixel size.</param>
        /// <param name="value">Pixels.</param>
        /// <returns>Percents.</returns>
        public static double PixelToPercent(double basePixelSize, double value)
        {
            return (value / basePixelSize) * 100;
        }

        /// <summary>
        /// Converts percents to pixels.
        /// </summary>
        /// <param name="basePixelSize">Base pixel size.</param>
        /// <param name="value">Percents.</param>
        /// <returns>Pixels.</returns>
        public static double PercentToPixel(double basePixelSize, double value)
        {
            return basePixelSize * (value / 100);
        }

        /// <summary>
        ///  Converts Units to pixel.
        /// </summary>
        /// <param name="basePixelSize">Base pixel size.</param>
        /// <param name="value">The value.</param>
        /// <param name="type">The type.</param>
        /// <returns></returns>
        public static double UnitToPixel(double basePixelSize, double value, UnitType type)
        {
            if (type == UnitType.Em)
            {
                return Unit.EmToPixel(basePixelSize, value);
            }
            else if (type == UnitType.Percent)
            {
                return Unit.PercentToPixel(basePixelSize, value);
            }
            else if (type == UnitType.Dip)
            {
                return value;
            }
            else
            {
                return Unit.UnitToDip(value, type);
            }
        }

        /// <summary>
        /// Converts Pixels the unit.
        /// </summary>
        /// <param name="basePixelSize">Base pixel size.</param>
        /// <param name="value">Pixel.</param>
        /// <param name="type">The type.</param>
        /// <returns></returns>
        public static double PixelToUnit(double basePixelSize, double value, UnitType type)
        {
            if (type == UnitType.Em)
            {
                return Unit.PixelToEm(basePixelSize, value);
            }
            else if (type == UnitType.Percent)
            {
                return Unit.PixelToPercent(basePixelSize, value);
            }
            else if (type == UnitType.Dip)
            {
                return value;
            }
            else
            {
                return Unit.DipToUnit(value, type);
            }
        }

        
        /// <summary>
        /// Determines whether unit type is relative.
        /// </summary>
        /// <param name="type">The unit type.</param>
        /// <returns></returns>
        public static bool IsRelativeUnitType(UnitType type)
        {
            return type == UnitType.Em || type == UnitType.Percent;
        }
    }
}