﻿using System;

namespace Telerik.Documents.Primitives
{
    /// <summary>
    /// Represents a Matrix structure.
    /// </summary>
    public struct Matrix
    {
        private double m11;
        private double m12;
        private double m21;
        private double m22;
        private double offsetX;
        private double offsetY;
        private bool isInitialized;
        private MatrixType matrixType;

        /// <summary>
        /// Initializes a new instance of the <see cref="Matrix"/> struct.
        /// m11, m12, 0
        /// m21, m22, 0 
        /// offsetX, offsetY, 1
        /// </summary>
        public Matrix(double m11, double m12, double m21, double m22, double offsetX, double offsetY)
            : this()
        {
            this.m11 = m11;
            this.m12 = m12;
            this.m21 = m21;
            this.m22 = m22;
            this.offsetX = offsetX;
            this.offsetY = offsetY;
            this.isInitialized = true;
        }

        /// <summary>
        /// Gets an identity matrix.
        /// </summary>
        public static Matrix Identity
        {
            get
            {
                return new Matrix(1, 0, 0, 1, 0, 0);
            }
        }

        /// <summary>
        /// Gets a value indicating whether this <see cref="Matrix"/> is equal to the identity matrix.
        /// </summary>
        public bool IsIdentity
        {
            get
            {
                return this == Matrix.Identity;
            }
        }

        /// <summary>
        /// Gets the determinant.
        /// </summary>
        public double Determinant
        {
            get
            {
                if (this.IsIdentity)
                {
                    return 1d;
                }

                return (this.M11 * this.M22) - (this.M12 * this.M21);
            }
        }

        /// <summary>
        /// Gets or sets the M11 value.
        /// </summary>
        public double M11
        {
            get
            {
                if (!this.isInitialized)
                {
                    return 1d;
                }

                return this.m11;
            }
            set
            {
                this.EnsureInitialized();
                this.m11 = value;
                this.UpdateType();
            }
        }

        /// <summary>
        /// Gets or sets the M12 value.
        /// </summary>
        public double M12
        {
            get
            {
                if (!this.isInitialized)
                {
                    return 0d;
                }

                return this.m12;
            }
            set
            {
                this.EnsureInitialized();
                this.m12 = value;
                this.UpdateType();
            }
        }

        /// <summary>
        /// Gets or sets the M21 value.
        /// </summary>
        public double M21
        {
            get
            {
                if (!this.isInitialized)
                {
                    return 0d;
                }

                return this.m21;
            }
            set
            {
                this.EnsureInitialized();
                this.m21 = value;
                this.UpdateType();
            }
        }

        /// <summary>
        /// Gets or sets the M22 value.
        /// </summary>
        public double M22
        {
            get
            {
                if (!this.isInitialized)
                {
                    return 1d;
                }

                return this.m22;
            }
            set
            {
                this.EnsureInitialized();
                this.m22 = value;
                this.UpdateType();
            }
        }

        /// <summary>
        /// Gets or sets the OffsetX value.
        /// </summary>
        public double OffsetX
        {
            get
            {
                if (!this.isInitialized)
                {
                    return 0d;
                }

                return this.offsetX;
            }
            set
            {
                this.EnsureInitialized();
                this.offsetX = value;
                this.UpdateType();
            }
        }

        /// <summary>
        /// Gets or sets the OffsetY value.
        /// </summary>
        public double OffsetY
        {
            get
            {
                if (!this.isInitialized)
                {
                    return 0d;
                }

                return this.offsetY;
            }
            set
            {
                this.EnsureInitialized();
                this.offsetY = value;
                this.UpdateType();
            }
        }

        /// <summary>
        /// Returns a new instance of <see cref="Matrix"/> that is the result of the multiplication if matrix1 and matrix2.
        /// </summary>
        /// <param name="matrix1">The left matrix to multiply.</param>
        /// <param name="matrix2">The right matrix to multiply.</param>
        /// <returns></returns>
        public static Matrix operator *(Matrix matrix1, Matrix matrix2)
        {
            Matrix result;

            if (matrix1.IsIdentity)
            {
                result = matrix2;
            }
            else if (matrix2.IsIdentity)
            {
                result = matrix1;
            }
            else
            {
                result = new Matrix((matrix1.M11 * matrix2.M11) + (matrix1.M12 * matrix2.M21),
                    (matrix1.M11 * matrix2.M12) + (matrix1.M12 * matrix2.M22),
                    (matrix1.M21 * matrix2.M11) + (matrix1.M22 * matrix2.M21),
                    (matrix1.M21 * matrix2.M12) + (matrix1.M22 * matrix2.M22),
                    ((matrix1.OffsetX * matrix2.M11) + (matrix1.OffsetY * matrix2.M21)) + matrix2.OffsetX,
                    ((matrix1.OffsetX * matrix2.M12) + (matrix1.OffsetY * matrix2.M22)) + matrix2.OffsetY);
            }

            return result;
        }

        /// <summary>
        /// Compares two <see cref="Matrix"/> instances for exact equality.
        /// </summary>
        /// <param name="a">The first <see cref="Matrix"/> to compare.</param>
        /// <param name="b">The second <see cref="Matrix"/> to compare.</param>
        /// <returns>True if the two <see cref="Matrix"/>s are exactly equal; otherwise false.</returns>
        public static bool operator ==(Matrix a, Matrix b)
        {
            return a.M11 == b.M11 && a.M21 == b.M21 && a.M12 == b.M12 && a.M22 == b.M22 && a.OffsetX == b.OffsetX && a.OffsetY == b.OffsetY;
        }

        /// <summary>
        /// Compares two <see cref="Matrix"/> instances for exact inequality.
        /// </summary>
        /// <param name="a">The first <see cref="Matrix"/> to compare.</param>
        /// <param name="b">The second <see cref="Matrix"/> to compare.</param>
        /// <returns>True if the two <see cref="Matrix"/>s are exactly unequal; otherwise false.</returns>
        public static bool operator !=(Matrix a, Matrix b)
        {
            return !(a == b);
        }

        /// <summary>
        /// Translates this <see cref="Matrix"/>.
        /// </summary>
        /// <param name='offsetX'>The offset in the x dimension.</param>
        /// <param name='offsetY'>The offset in the y dimension.</param>
        public Matrix Translate(double offsetX, double offsetY)
        {
            this.OffsetX += offsetX;
            this.OffsetY += offsetY;

            return this;
        }

        /// <summary>
        /// Prepends a scale around the origin to this <see cref="Matrix"/>.
        /// </summary>
        /// <param name='scaleX'>The scale factor in the x dimension.</param>
        /// <param name='scaleY'>The scale factor in the y dimension.</param>
        /// <param name='centerX'>The centerX of scale. Default 0.</param>
        /// <param name='centerY'>The centerY of scale. Default 0.</param>
        public Matrix ScalePrepend(double scaleX, double scaleY, double centerX = 0, double centerY = 0)
        {
            this = new Matrix(scaleX, 0, 0, scaleY, centerX - scaleX * centerX, centerY - scaleY * centerY) * this;

            return this;
        }

        /// <summary>
        /// Scales this <see cref="Matrix"/> around the origin.
        /// </summary>
        /// <param name='scaleX'>The scale factor in the x dimension.</param>
        /// <param name='scaleY'>The scale factor in the y dimension.</param>
        /// <param name='centerX'>The centerX of scale. Default 0.</param>
        /// <param name='centerY'>The centerY of scale. Default 0.</param>
        public Matrix Scale(double scaleX, double scaleY, double centerX = 0, double centerY = 0)
        {
            this = this * new Matrix(scaleX, 0, 0, scaleY, centerX - scaleX * centerX, centerY - scaleY * centerY);

            return this;
        }

        /// <summary>
        /// Rotates this <see cref="Matrix"/> at the given point.
        /// </summary>
        /// <param name='angle'>The angle to rotate.</param>
        /// <param name='centerX'>The centerX of rotation. Default 0.</param>
        /// <param name='centerY'>The centerY of rotation. Default 0.</param>
        public Matrix Rotate(double angle, double centerX = 0, double centerY = 0)
        {
            angle = Math.PI * angle / 180;

            double sin = Math.Sin(angle);
            double cos = Math.Cos(angle);
            double offsetX = (centerX * (1.0 - cos)) + (centerY * sin);
            double offsetY = (centerY * (1.0 - cos)) - (centerX * sin);

            Matrix matrix = new Matrix(cos, sin, -sin, cos, offsetX, offsetY);

            ////this = matrix * this;
            this = this * matrix;

            return this;
        }

        internal void RotateAt(double angle, double centerX = 0, double centerY = 0)
        {
            this.Rotate(angle, centerX, centerY);
        }

        /// <summary>
        /// Returns the result of transforming a given point by this <see cref="Matrix"/>.
        /// </summary>
        /// <param name="point"> The Point to transform.</param>
        internal Point Transform(Point point)
        {
            double x = point.X;
            double y = point.Y;

            if (this.IsIdentity)
            {
                return new Point(x, y);
            }

            double rx = (x * this.M11) + (y * this.M21) + this.OffsetX;
            double ry = (x * this.M12) + (y * this.M22) + this.OffsetY;

            return new Point(rx, ry);
        }

        internal void TransformRect(ref Rect rect)
        {
            if (rect.IsEmpty)
            {
                return;
            }

            if (this.IsIdentity)
            {
                return;
            }

            if ((this.matrixType & MatrixType.Scaling) != MatrixType.Identity)
            {
                rect.X *= this.m11;
                rect.Y *= this.m22;
                rect.Width *= this.m11;
                rect.Height *= this.m22;
                if (rect.Width < 0)
                {
                    rect.X += rect.Width;
                    rect.Width = -rect.Width;
                }
                if (rect.Height < 0)
                {
                    rect.Y += rect.Height;
                    rect.Height = -rect.Height;
                }
            }

            if ((this.matrixType & MatrixType.Translation) != MatrixType.Identity)
            {
                rect.X += this.offsetX;
                rect.Y += this.offsetY;
            }

            if (this.matrixType == MatrixType.Unknown)
            {
                Point point = this.Transform(rect.TopLeft);
                Point point1 = this.Transform(rect.TopRight);
                Point point2 = this.Transform(rect.BottomRight);
                Point point3 = this.Transform(rect.BottomLeft);
                rect.X = Math.Min(Math.Min(point.X, point1.X), Math.Min(point2.X, point3.X));
                rect.Y = Math.Min(Math.Min(point.Y, point1.Y), Math.Min(point2.Y, point3.Y));
                rect.Width = Math.Max(Math.Max(point.X, point1.X), Math.Max(point2.X, point3.X)) - rect.X;
                rect.Height = Math.Max(Math.Max(point.Y, point1.Y), Math.Max(point2.Y, point3.Y)) - rect.Y;
            }
        }

        /// <inheritdoc/>
        public override int GetHashCode()
        {
            if (this.IsIdentity)
            {
                return 0;
            }
            else
            {
                return this.M11.GetHashCode() ^
                       this.M12.GetHashCode() ^
                       this.M21.GetHashCode() ^
                       this.M22.GetHashCode() ^
                       this.OffsetX.GetHashCode() ^
                       this.OffsetY.GetHashCode();
            }
        }

        /// <inheritdoc/>
        public override bool Equals(object obj)
        {
            if (obj == null || !(obj is Matrix))
            {
                return false;
            }

            return this.Equals((Matrix)obj);
        }

        /// <inheritdoc/>
        public override string ToString()
        {
            return string.Format("{0} {1} 0 | {2} {3} 0 | {4} {5} 1", this.M11, this.M12, this.M21, this.M22, this.OffsetX, this.OffsetY);
        }

        private bool Equals(Matrix value)
        {
            return this.M11.Equals(value.M11) &&
                   this.M12.Equals(value.M12) &&
                   this.M21.Equals(value.M21) &&
                   this.M22.Equals(value.M22) &&
                   this.OffsetX.Equals(value.OffsetX) &&
                   this.OffsetY.Equals(value.OffsetY);
        }

        private void EnsureInitialized()
        {
            if (!this.isInitialized)
            {
                this.m11 = 1;
                this.m12 = 0;
                this.m21 = 0;
                this.m22 = 1;
                this.offsetX = 0;
                this.offsetY = 0;

                this.isInitialized = true;
            }
        }

        private void UpdateType()
        {
            this.matrixType = MatrixType.Identity;
            if (this.m21 != 0 || this.m12 != 0)
            {
                this.matrixType = MatrixType.Unknown;
                return;
            }

            if (this.m11 != 1 || this.m22 != 1)
            {
                this.matrixType = MatrixType.Scaling;
            }

            if (this.offsetX != 0 || this.offsetY != 0)
            {
                this.matrixType |= MatrixType.Translation;
            }

            if ((this.matrixType & (MatrixType.Translation | MatrixType.Scaling)) == MatrixType.Identity)
            {
                this.matrixType = MatrixType.Identity;
            }
        }
    }
}
