﻿using System;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Primitives
{
    /// <summary>
    /// Represents padding or margin information.
    /// </summary>
    public class Padding
    {
        private const double EmptyValue = 0;

        /// <summary>
        /// An empty padding.
        /// </summary>
        public static readonly Padding Empty = new Padding(EmptyValue);

        private readonly double top;
        private readonly double left;
        private readonly double right;
        private readonly double bottom;

        /// <summary>
        /// Initializes a new instance of the <see cref="Padding" /> class and sets all paddings to a given value.
        /// </summary>
        /// <param name="all">The value in device independent pixels (1/96 inch).</param>
        public Padding(double all)
        {
            this.top = all;
            this.left = all;
            this.right = all;
            this.bottom = all;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Padding" /> class.
        /// </summary>
        /// <param name="left">The left padding in device independent pixels (1/96 inch).</param>
        /// <param name="top">The top padding in device independent pixels (1/96 inch).</param>
        /// <param name="right">The right padding in device independent pixels (1/96 inch).</param>
        /// <param name="bottom">The bottom padding in device independent pixels (1/96 inch).</param>
        public Padding(double left, double top, double right, double bottom)
        {
            this.top = top;
            this.left = left;
            this.right = right;
            this.bottom = bottom;
        }

        /// <summary>
        /// Gets the top padding.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The top padding.</value>
        public double Top
        {
            get
            {
                return this.top;
            }
        }

        /// <summary>
        /// Gets the bottom padding.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The bottom padding.</value>
        public double Bottom
        {
            get
            {
                return this.bottom;
            }
        }

        /// <summary>
        /// Gets the left padding.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The left padding.</value>
        public double Left
        {
            get
            {
                return this.left;
            }
        }

        /// <summary>
        /// Gets the right padding.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The right padding.</value>
        public double Right
        {
            get
            {
                return this.right;
            }
        }

        internal bool IsEmpty
        {
            get
            {
                return this.Left == this.Right && this.Right == this.Top && this.top == this.Bottom && this.Bottom == EmptyValue;
            }
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// Returns true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            Padding other = obj as Padding;
            if (other == null)
            {
                return false;
            }

            return this.Left.Equals(other.Left) &&
                   this.Top.Equals(other.Top) &&
                   this.Right.Equals(other.Right) &&
                   this.Bottom.Equals(other.Bottom);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return ObjectExtensions.CombineHashCodes(
                this.Left.GetHashCode(),
                this.Top.GetHashCode(),
                this.Right.GetHashCode(),
                this.Bottom.GetHashCode());
        }

        /// <summary>
        /// Determines whether the specified paddings are equal.
        /// </summary>
        /// <returns>True if the paddings are equal.</returns>
        public static bool operator ==(Padding a, Padding b)
        {
            // If both are null, or both are same instance, return true.
            if (Object.ReferenceEquals(a, b))
            {
                return true;
            }

            // If one is null, but not both, return false.
            if (((object)a == null) || ((object)b == null))
            {
                return false;
            }

            return a.Equals(b);
        }

        /// <summary>
        /// Determines whether the specified paddings are different.
        /// </summary>
        /// <returns>True if the paddings are different.</returns>
        public static bool operator !=(Padding a, Padding b)
        {
            return !(a == b);
        }

        internal Padding Clone()
        {
            Padding padding = new Padding(this.Left, this.Top, this.Right, this.Bottom);

            return padding;
        }

        /// <inheritdoc />
        public override string ToString()
        {
            return string.Format("{0}, {1}, {2}, {3}", this.Left, this.Top, this.Right, this.Bottom);
        }
    }
}