﻿using System;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Documents.Primitives
{
    /// <summary>
    /// Rect structure representing a rectangle, defined by two double coordinates and a double width and double height.
    /// </summary>
    public struct Rect
    {
        private readonly static Rect emptyRect;

        private double x;
        private double y;
        private double width;
        private double height;

        static Rect()
        {
            emptyRect = new Rect();
            emptyRect.x = Double.PositiveInfinity;
            emptyRect.y = Double.PositiveInfinity;
            emptyRect.width = Double.NegativeInfinity;
            emptyRect.height = Double.NegativeInfinity;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Rect"/> structure.
        /// </summary>
        /// <param name="x">The X coordinate.</param>
        /// <param name="y">The Y coordinate.</param>
        /// <param name="width">The width. Should be non-negative.</param>
        /// <param name="height">The height. Should be non-negative.</param>
        public Rect(double x, double y, double width, double height)
        {
            Guard.ThrowExceptionIfLessThan(0, width, "width");
            Guard.ThrowExceptionIfLessThan(0, height, "height");

            this.x = x;
            this.y = y;
            this.width = width;
            this.height = height;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Rect"/> structure.
        /// </summary>
        /// <param name="location">The <see cref="Point"/> location, that determines the coordinates.</param>
        /// <param name="size">The <see cref="Size"/> size, that defines the width and height. It should not be Empty. </param>
        internal Rect(Point location, Size size) : this(location.X, location.Y, size.Width, size.Height)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Rect"/> structure at (0;0) coordinates.
        /// </summary>
        /// <param name="size">The <see cref="Size"/> size, that defines the width and height. It should not be Empty. </param>
        internal Rect(Size size)
            : this(new Point(), size)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Rect"/> structure. Sets the values of X, Y, Width and Height from the two points provided.
        /// </summary>
        internal Rect(Point point1,
                    Point point2)
        {
            this.x = Math.Min(point1.X, point2.X);
            this.y = Math.Min(point1.Y, point2.Y);

            // Max with 0 to prevent double -epsilon. 
            this.width = Math.Max(Math.Max(point1.X, point2.X) - x, 0);
            this.height = Math.Max(Math.Max(point1.Y, point2.Y) - y, 0);
        }

        /// <summary>
        /// Gets an empty <see cref="Rect"/>. It has negative-infinity Width and Height and positive-infinity X and Y.
        /// </summary>
        public static Rect Empty
        {
            get
            {
                return emptyRect;
            }
        }

        /// <summary>
        /// Returns true if this is the Empty <see cref="Rect"/>.
        /// </summary>
        public bool IsEmpty
        {
            get
            {
                return this.width < 0;
            }
        }

        /// <summary>
        /// Size - The Size representing the area of the Rectangle
        /// </summary>
        public Size Size
        {
            get
            {
                if (this.IsEmpty)
                {
                    return Size.Empty;
                }
                else
                {
                    return new Size(this.width, this.height);
                }
            }
            set
            {
                if (value.IsEmpty)
                {
                    this = emptyRect;
                }
                else
                {
                    if (this.IsEmpty)
                    {
                        throw new System.InvalidOperationException("Rect_CannotModifyEmptyRect");
                    }

                    this.width = value.Width;
                    this.height = value.Height;
                }
            }
        }

        /// <summary>
        /// Gets or sets the X coordinate. Cannot set X if this is an Empty <see cref="Rect"/>.
        /// </summary>
        public double X
        {
            get
            {
                return this.x;
            }
            set
            {
                Guard.ThrowExceptionIfTrue(this.IsEmpty, "Cannot modify empty rectangle.");

                this.x = value;
            }
        }

        /// <summary>
        /// Gets or sets the Y coordinate. Cannot set Y if this is an Empty <see cref="Rect"/>.
        /// </summary>
        public double Y
        {
            get
            {
                return this.y;
            }
            set
            {
                Guard.ThrowExceptionIfTrue(this.IsEmpty, "Cannot modify empty rectangle.");

                this.y = value;
            }
        }

        /// <summary>
        /// Gets or sets the Width. Cannot set Width if this is an Empty <see cref="Rect"/>. Should be non-negative. 
        /// </summary>
        public double Width
        {
            get
            {
                return this.width;
            }
            set
            {
                Guard.ThrowExceptionIfTrue(this.IsEmpty, "Cannot modify empty rectangle.");

                Guard.ThrowExceptionIfLessThan(0, value, "value");

                this.width = value;
            }
        }
        /// <summary>
        /// Gets or sets the Height. Cannot set Height if this is an Empty <see cref="Rect"/>. Should be non-negative. 
        /// </summary>
        public double Height
        {
            get
            {
                return this.height;
            }
            set
            {
                Guard.ThrowExceptionIfTrue(this.IsEmpty, "Cannot modify empty rectangle.");

                Guard.ThrowExceptionIfLessThan(0, value, "value");

                this.height = value;
            }
        }

        /// <summary>
        /// Gets the value of the X coordinate.
        /// </summary>
        public double Left
        {
            get
            {
                return this.X;
            }
        }

        /// <summary>
        /// Gets the value of the Y coordinate.
        /// </summary>
        public double Top
        {
            get
            {
                return this.Y;
            }
        }

        /// <summary>
        /// Gets the X + Width value. If this is an Empty <see cref="Rect"/> returns negative-infinity.
        /// </summary>
        public double Right
        {
            get
            {
                if (this.IsEmpty)
                {
                    return Double.NegativeInfinity;
                }

                return this.X + this.Width;
            }
        }

        /// <summary>
        /// Gets the Y + Height value. If this is an Empty <see cref="Rect"/> gets negative-infinity.
        /// </summary>
        public double Bottom
        {
            get
            {
                if (this.IsEmpty)
                {
                    return Double.NegativeInfinity;
                }

                return this.Y + this.Height;
            }
        }

        /// <summary>
        /// TopLeft Property - This is a read-only alias for the Point which is at X, Y
        /// If this is the empty rectangle, the value will be positive infinity, positive infinity.
        /// </summary>
        public Point TopLeft
        {
            get
            {
                return new Point(Left, Top);
            }
        }

        /// <summary>
        /// TopRight Property - This is a read-only alias for the Point which is at X + Width, Y
        /// If this is the empty rectangle, the value will be negative infinity, positive infinity.
        /// </summary>
        public Point TopRight
        {
            get
            {
                return new Point(Right, Top);
            }
        }

        /// <summary>
        /// BottomLeft Property - This is a read-only alias for the Point which is at X, Y + Height
        /// If this is the empty rectangle, the value will be positive infinity, negative infinity.
        /// </summary>
        public Point BottomLeft
        {
            get
            {
                return new Point(Left, Bottom);
            }
        }

        /// <summary>
        /// BottomRight Property - This is a read-only alias for the Point which is at X + Width, Y + Height
        /// If this is the empty rectangle, the value will be negative infinity, negative infinity.
        /// </summary>
        public Point BottomRight
        {
            get
            {
                return new Point(Right, Bottom);
            }
        }

        /// <summary>
        /// Compares two <see cref="Rect"/> instances for exact equality.
        /// </summary>
        /// <param name="firstRect">The first <see cref="Rect"/> to compare.</param>
        /// <param name="secondRect">The second <see cref="Rect"/> to compare.</param>
        /// <returns>True if the two <see cref="Rect"/>s are exactly equal; otherwise false.</returns>
        public static bool operator ==(Rect firstRect, Rect secondRect)
        {
            return firstRect.X == secondRect.X
                && firstRect.Y == secondRect.Y
                && firstRect.Width == secondRect.Width
                && firstRect.Height == secondRect.Height;
        }

        /// <summary>
        /// Compares two <see cref="Rect"/> instances for exact inequality.
        /// </summary>
        /// <param name="firstRect">The first <see cref="Rect"/> to compare.</param>
        /// <param name="secondRect">The second <see cref="Rect"/> to compare.</param>
        /// <returns>True if the two <see cref="Rect"/>s are exactly unequal; otherwise false.</returns>
        public static bool operator !=(Rect firstRect, Rect secondRect)
        {
            return !(firstRect == secondRect);
        }

        /// <inheritdoc/>
        public override int GetHashCode()
        {
            if (this.IsEmpty)
            {
                return 0;
            }
            else
            {
                return this.X.GetHashCode() ^ this.Y.GetHashCode() ^ this.Width.GetHashCode() ^ this.Height.GetHashCode();
            }
        }

        /// <inheritdoc/>
        public override bool Equals(object obj)
        {
            if (obj == null || !(obj is Rect))
            {
                return false;
            }

            return this.Equals((Rect)obj);
        }

        private bool Equals(Rect other)
        {
            if (this.IsEmpty)
            {
                return other.IsEmpty;
            }

            return this.X.Equals(other.X)
                && this.Y.Equals(other.Y)
                && this.Width.Equals(other.Width)
                && this.Height.Equals(other.Height);
        }

        /// <summary>
        /// Returns a value indicating whether a Point is inside the rectangle, including the edges.
        ///</summary>
        internal bool Contains(Point point)
        {
            return this.Contains(point.X, point.Y);
        }

        /// <summary>
        /// Returns a value indicating whether the x, y coordinates are inside the rectangle, including the edges.
        /// </summary>
        public bool Contains(double x, double y)
        {
            if (this.IsEmpty)
            {
                return false;
            }

            ////var t =  ((x >= this.x) && (x - width <= this.x) &&
            ////        (y >= this.y) && (y - height <= this.y));

            bool isInside = this.X <= x && this.X >= x - this.Width
                && this.Y <= y && this.Y >= y - this.Height;

            return isInside;
        }

        /////// <summary>
        /////// Contains - Returns true if the Rect non-Empty and is entirely contained within the
        /////// rectangle, inclusive of the edges.
        /////// Returns false otherwise
        /////// </summary>
        ////public bool Contains(Rect rect)
        ////{
        ////    if (IsEmpty || rect.IsEmpty)
        ////    {
        ////        return false;
        ////    }

        ////    return (x <= rect.x &&
        ////            y <= rect.y &&
        ////            x + width >= rect.x + rect.width &&
        ////            y + height >= rect.y + rect.height);
        ////}

        /// <summary>
        /// IntersectsWith - Returns true if the Rect intersects with this rectangle
        /// Returns false otherwise.
        /// Note that if one edge is coincident, this is considered an intersection.
        /// </summary>
        /// <returns>
        /// Returns true if the Rect intersects with this rectangle
        /// Returns false otherwise.
        /// or Height
        /// </returns>
        /// <param name="rect"> Rect </param>
        public bool IntersectsWith(Rect rect)
        {
            if (this.IsEmpty || rect.IsEmpty)
            {
                return false;
            }

            return (rect.Left <= this.Right) &&
                   (rect.Right >= this.Left) &&
                   (rect.Top <= this.Bottom) &&
                   (rect.Bottom >= this.Top);
        }

        /// <summary>
        /// Intersect - Update this rectangle to be the intersection of this and rect
        /// If either this or rect are Empty, the result is Empty as well.
        /// </summary>
        /// <param name="rect"> The rect to intersect with this </param>
        public void Intersect(Rect rect)
        {
            if (!this.IntersectsWith(rect))
            {
                this = Rect.Empty;
            }
            else
            {
                double left = Math.Max(this.Left, rect.Left);
                double top = Math.Max(this.Top, rect.Top);

                //  Max with 0 to prevent double weirdness from causing us to be (-epsilon..0)
                this.width = Math.Max(Math.Min(this.Right, rect.Right) - left, 0);
                this.height = Math.Max(Math.Min(this.Bottom, rect.Bottom) - top, 0);

                x = left;
                y = top;
            }
        }

        /////// <summary>
        /////// Intersect - Return the result of the intersection of rect1 and rect2.
        /////// If either this or rect are Empty, the result is Empty as well.
        /////// </summary>
        ////public static Rect Intersect(Rect rect1, Rect rect2)
        ////{
        ////    rect1.Intersect(rect2);
        ////    return rect1;
        ////}

        /// <summary>
        /// This <see cref="Rect"/> becomes a the result of union of this and otherRect.
        /// </summary>
        public void Union(Rect otherRect)
        {
            if (this.IsEmpty)
            {
                this = otherRect;
                return;
            }

            if (otherRect.IsEmpty)
            {
                return;
            }

            double newX = Math.Min(this.X, otherRect.X);
            double newY = Math.Min(this.Y, otherRect.Y);

            double maxRight = Math.Max(this.Right, otherRect.Right);
            double newWidth = Math.Max(maxRight - newX, 0);

            double maxBottom = Math.Max(this.Bottom, otherRect.Bottom);
            double newHeight = Math.Max(maxBottom - newY, 0);

            this.X = newX;
            this.Y = newY;
            this.Width = newWidth;
            this.Height = newHeight;
        }

        /////// <summary>
        /////// Union - Return the result of the union of rect1 and rect2.
        /////// </summary>
        ////public static Rect Union(Rect rect1, Rect rect2)
        ////{
        ////    rect1.Union(rect2);
        ////    return rect1;
        ////}

        /////// <summary>
        /////// Union - Update this rectangle to be the union of this and point.
        /////// </summary>
        ////public void Union(Point point)
        ////{
        ////    Union(new Rect(point, point));
        ////}

        /////// <summary>
        /////// Union - Return the result of the union of rect and point.
        /////// </summary>
        ////public static Rect Union(Rect rect, Point point)
        ////{
        ////    rect.Union(new Rect(point, point));
        ////    return rect;
        ////}


        /////// <summary>
        /////// Offset - translate the Location by the offset provided
        /////// If this is Empty, this method is illegal.
        /////// </summary>
        ////public void Offset(double offsetX, double offsetY)
        ////{
        ////    if (IsEmpty)
        ////    {
        ////        throw new System.InvalidOperationException("Rect_CannotCallMethod");
        ////    }

        ////    x += offsetX;
        ////    y += offsetY;
        ////}

        /////// <summary>
        /////// Offset - return the result of offsetting rect by the offset provided
        /////// If this is Empty, this method is illegal.
        /////// </summary>
        ////public static Rect Offset(Rect rect, double offsetX, double offsetY)
        ////{
        ////    rect.Offset(offsetX, offsetY);
        ////    return rect;
        ////}

        /////// <summary>
        /////// Inflate - inflate the bounds by the size provided, in all directions
        /////// If this is Empty, this method is illegal.
        /////// </summary>
        ////public void Inflate(Size size)
        ////{
        ////    Inflate(size.Width, size.Height);
        ////}

        /////// <summary>
        /////// Inflate - inflate the bounds by the size provided, in all directions.
        /////// If -width is > Width / 2 or -height is > Height / 2, this Rect becomes Empty
        /////// If this is Empty, this method is illegal.
        /////// </summary>
        ////public void Inflate(double width, double height)
        ////{
        ////    if (IsEmpty)
        ////    {
        ////        throw new System.InvalidOperationException("SRID.Rect_CannotCallMethod");
        ////    }

        ////    x -= width;
        ////    y -= height;

        ////    // Do two additions rather than multiplication by 2 to avoid spurious overflow
        ////    // That is: (A + 2 * B) != ((A + B) + B) if 2*B overflows.
        ////    // Note that multiplication by 2 might work in this case because A should start
        ////    // positive & be "clamped" to positive after, but consider A = Inf & B = -MAX.
        ////    this.width += width;
        ////    this.width += width;
        ////    this.height += height;
        ////    this.height += height;

        ////    // We catch the case of inflation by less than -width/2 or -height/2 here.  This also
        ////    // maintains the invariant that either the Rect is Empty or _width and _height are
        ////    // non-negative, even if the user parameters were NaN, though this isn't strictly maintained
        ////    // by other methods.
        ////    if (!(this.width >= 0 && this.height >= 0))
        ////    {
        ////        this = emptyRect;
        ////    }
        ////}

        /////// <summary>
        /////// Inflate - return the result of inflating rect by the size provided, in all directions
        /////// If this is Empty, this method is illegal.
        /////// </summary>
        ////public static Rect Inflate(Rect rect, Size size)
        ////{
        ////    rect.Inflate(size.Width, size.Height);
        ////    return rect;
        ////}

        /////// <summary>
        /////// Inflate - return the result of inflating rect by the size provided, in all directions
        /////// If this is Empty, this method is illegal.
        /////// </summary>
        ////public static Rect Inflate(Rect rect, double width, double height)
        ////{
        ////    rect.Inflate(width, height);
        ////    return rect;
        ////}

        /////// <summary>
        /////// Returns the bounds of the transformed rectangle.
        /////// The Empty Rect is not affected by this call.
        /////// </summary>
        /////// <returns>
        /////// The rect which results from the transformation.
        /////// </returns>
        /////// <param name="rect"> The Rect to transform. </param>
        /////// <param name="matrix"> The Matrix by which to transform. </param>
        ////public static Rect Transform(Rect rect, Matrix matrix)
        ////{
        ////    Matrix.TransformRect(ref rect, ref matrix);
        ////    return rect;
        ////}

        /////// <summary>
        /////// Updates rectangle to be the bounds of the original value transformed
        /////// by the matrix.
        /////// The Empty Rect is not affected by this call.        
        /////// </summary>
        /////// <param name="matrix"> Matrix </param>
        ////public void Transform(Matrix matrix)
        ////{
        ////    Matrix.TransformRect(ref this, ref matrix);
        ////}

        /////// <summary>
        /////// Scale the rectangle in the X and Y directions
        /////// </summary>
        /////// <param name="scaleX"> The scale in X </param>
        /////// <param name="scaleY"> The scale in Y </param>
        ////public void Scale(double scaleX, double scaleY)
        ////{
        ////    if (IsEmpty)
        ////    {
        ////        return;
        ////    }

        ////    x *= scaleX;
        ////    y *= scaleY;
        ////    width *= scaleX;
        ////    height *= scaleY;

        ////    // If the scale in the X dimension is negative, we need to normalize X and Width
        ////    if (scaleX < 0)
        ////    {
        ////        // Make X the left-most edge again
        ////        x += width;

        ////        // and make Width positive
        ////        width *= -1;
        ////    }

        ////    // Do the same for the Y dimension
        ////    if (scaleY < 0)
        ////    {
        ////        // Make Y the top-most edge again
        ////        y += height;

        ////        // and make Height positive
        ////        height *= -1;
        ////    }
        ////}
    }
}
