﻿using Telerik.Windows.Documents.Utilities;

namespace Telerik.Documents.Primitives
{
    /// <summary>
    /// Size structure defined by width and height.
    /// </summary>
    public struct Size
    {
        private static readonly Size emptySize;
        private double width;
        private double height;

        static Size()
        {
            emptySize = new Size();
            emptySize.width = double.NegativeInfinity;
            emptySize.height = double.NegativeInfinity;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Size"/> structure.
        /// </summary>
        /// <param name="width">The Width. Should be non-negative.</param>
        /// <param name="height">The Height. Should be non-negative.</param>
        public Size(double width, double height)
        {
            Guard.ThrowExceptionIfLessThan(0, width, "width");
            Guard.ThrowExceptionIfLessThan(0, height, "height");

            this.width = width;
            this.height = height;
        }

        /// <summary>
        /// Gets an empty <see cref="Size"/>. It has negative-infinity Width and Height.
        /// </summary>
        public static Size Empty
        {
            get
            {
                return emptySize;
            }
        }

        /// <summary>
        /// Returns true if this size is the Empty size.
        /// </summary>
        public bool IsEmpty
        {
            get
            {
                return this.width < 0;
            }
        }

        /// <summary>
        /// Gets or sets the width. Should be non-negative.
        /// </summary>
        public double Width
        {
            get
            {
                return this.width;
            }
            set
            {
                Guard.ThrowExceptionIfLessThan(0, value, "value");

                this.width = value;
            }
        }

        /// <summary>
        /// Gets or sets the height. Should be non-negative.
        /// </summary>
        public double Height
        {
            get
            {
                return this.height;
            }
            set
            {
                Guard.ThrowExceptionIfLessThan(0, value, "value");

                this.height = value;
            }
        }

        /// <summary>
        /// Compares two <see cref="Size"/> instances for exact equality.
        /// </summary>
        /// <param name="firstSize">The first <see cref="Size"/> to compare.</param>
        /// <param name="secondSize">The second <see cref="Size"/> to compare.</param>
        /// <returns>True if the two <see cref="Size"/>s are exactly equal; otherwise false.</returns>
        public static bool operator ==(Size firstSize, Size secondSize)
        {
            return ((firstSize.Width == secondSize.Width)
                && (firstSize.Height == secondSize.Height));
        }

        /// <summary>
        /// Compares two <see cref="Size"/> instances for exact inequality.
        /// </summary>
        /// <param name="firstSize">The first <see cref="Size"/> to compare.</param>
        /// <param name="secondSize">The second <see cref="Size"/> to compare.</param>
        /// <returns>True if the two <see cref="Size"/>s are exactly unequal; otherwise false.</returns>
        public static bool operator !=(Size firstSize, Size secondSize)
        {
            return !(firstSize == secondSize);
        }

        /// <inheritdoc/>
        public override int GetHashCode()
        {
            if (this.IsEmpty)
            {
                return 0;
            }

            return this.Width.GetHashCode() ^ this.Height.GetHashCode();
        }

        /// <inheritdoc/>
        public override bool Equals(object obj)
        {
            if (obj != null && obj is Size)
            {
                return this.Equals((Size)obj);
            }

            return false;
        }

        private bool Equals(Size other)
        {
            if (this.IsEmpty)
            {
                return other.IsEmpty;
            }

            return this.Width.Equals(other.Width) && this.Height.Equals(other.Height);
        }
    }
}
