﻿namespace Telerik.Documents.Primitives
{
    /// <summary>
    /// Describes the thickness of a rectangular border.
    /// </summary>
    public struct Thickness
    {
        private const double EmptyValue = 0;

        /// <summary>
        /// An empty thickness.
        /// </summary>
        public static readonly Thickness Empty = new Thickness(EmptyValue);

        /// <summary>
        /// Initializes a new instance of the <see cref="Thickness"/> struct.
        /// </summary>
        /// <param name="length">The uniform length for left, top, right and bottom sides.</param>
        public Thickness(double length)
        {
            this.Left = this.Top = this.Right = this.Bottom = length;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Thickness"/> struct.
        /// </summary>
        /// <param name="left">The thickness for the left side.</param>
        /// <param name="top">The thickness for the top side.</param>
        /// <param name="right">The thickness for the right side.</param>
        /// <param name="bottom">The thickness for the bottom side.</param>
        public Thickness(double left, double top, double right, double bottom)
        {
            this.Left = left;
            this.Top = top;
            this.Right = right;
            this.Bottom = bottom;
        }

        /// <summary>
        /// The thickness on the left.
        /// </summary>
        public double Left
        {
            get;
            set;
        }

        /// <summary>
        /// The thickness on the top.
        /// </summary>
        public double Top
        {
            get;
            set;
        }

        /// <summary>
        /// The thickness on the right.
        /// </summary>
        public double Right
        {
            get;
            set;
        }

        /// <summary>
        /// The thickness on the bottom.
        /// </summary>
        public double Bottom
        {
            get;
            set;
        }

        /// <summary>
        /// Returns the hash code for this instance.
        /// </summary>
        /// <returns>A 32-bit signed integer hash code.</returns>
        public override int GetHashCode()
        {
            return this.Left.GetHashCode() ^ this.Top.GetHashCode() ^ this.Right.GetHashCode() ^ this.Bottom.GetHashCode();
        }

        /// <summary>
        /// Returns a value indicating whether this instance and a specified Thickness object represent the same value.
        /// </summary>
        /// <param name="obj">A Thickness object to compare to this instance.</param>
        /// <returns> true if obj is equal to this instance; otherwise, false.</returns>
        public override bool Equals(object obj)
        {
            if (obj == null || !(obj is Thickness))
            {
                return false;
            }

            return this.Equals((Thickness)obj);
        }

        /// <summary>
        /// Compares the value of two Thickness structures for equality.
        /// </summary>
        /// <param name="a">The first structure to compare.</param>
        /// <param name="b">The other structure to compare.</param>
        /// <returns>true if the two instances of Thickness are equal; otherwise, false.</returns>
        public static bool operator ==(Thickness a, Thickness b)
        {
            if (a == null && b == null)
            {
                return true;
            }

            if (a == null || b == null)
            {
                return false;
            }

            return a.Equals(b);
        }

        /// <summary>
        /// Compares two Thickness structures for inequality.
        /// </summary>
        /// <param name="a">The first structure to compare</param>
        /// <param name="b">The other structure to compare.</param>
        /// <returns>true if the two instances of Thickness are not equal; otherwise, false.</returns>
        public static bool operator !=(Thickness a, Thickness b)
        {
            return !(a == b);
        }

        private bool Equals(Thickness other)
        {
            return this.Left.Equals(other.Left)
                && this.Top.Equals(other.Top)
                && this.Right.Equals(other.Right)
                && this.Bottom.Equals(other.Bottom);
        }

        internal Thickness Clone()
        {
            Thickness padding = new Thickness(this.Left, this.Top, this.Right, this.Bottom);

            return padding;
        }
    }
}