﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif

namespace Telerik.Windows.Documents.Utilities
{
    internal class MathUtilities
    {
        private static readonly int[] romanNumeralValues = new int[] { 1000, 900, 500, 400, 100, 90, 50, 40, 10, 9, 5, 4, 1 };
        private static readonly string[] romanNumerals = new string[] { "m", "cm", "d", "cd", "c", "xc", "l", "xl", "x", "ix", "v", "iv", "i" };

        /// <summary>
        /// Finds the minimum and maximum of a quadratic function a*x*x + b*x + c.
        /// </summary>
        /// <param name="interval">The interval.</param>
        /// <param name="a">The a coeficient.</param>
        /// <param name="b">The b coeficient.</param>
        /// <param name="c">The c coeficient.</param>
        /// <param name="d">The free coeficient.</param>
        /// <returns>The bounding interval.</returns>
        public static Interval FindLocalMinimumAndMaximum(Interval interval, double a, double b, double c, double d)
        {
            return FindLocalMinimumAndMaximum(FindLocalExtrema(interval, a, b, c, d));
        }

        /// <summary>
        /// Finds the minimum and maximum of a quadratic function a*x*x + b*x + c.
        /// </summary>
        /// <param name="interval">The interval.</param>
        /// <param name="a">The a coeficient.</param>
        /// <param name="b">The b coeficient.</param>
        /// <param name="c">The free coeficient.</param>
        /// <returns>The bounding interval.</returns>
        public static Interval FindLocalMinimumAndMaximum(Interval interval, double a, double b, double c)
        {
            return FindLocalMinimumAndMaximum(FindLocalExtrema(interval, a, b, c));
        }

        private static Interval FindLocalMinimumAndMaximum(IEnumerable<Point> localExtrema)
        {
            double minimum = double.MaxValue;
            double maximum = double.MinValue;

            foreach (Point localExtemum in localExtrema)
            {
                if (localExtemum.Y < minimum)
                {
                    minimum = localExtemum.Y;
                }

                if (localExtemum.Y > maximum)
                {
                    maximum = localExtemum.Y;
                }
            }

            return new Interval(minimum, maximum);
        }

        /// <summary>
        /// Finds the local extrema of a quadratic function a*x*x + b*x + c.
        /// </summary>
        /// <param name="interval">The interval.</param>
        /// <param name="a">The a coeficient.</param>
        /// <param name="b">The b coeficient.</param>
        /// <param name="c">The free coeficient.</param>
        /// <returns>The local extrema.</returns>
        public static IEnumerable<Point> FindLocalExtrema(Interval interval, double a, double b, double c)
        {
            List<double> pointsToCheck = new List<double>();
            pointsToCheck.Add(interval.Start);

            double zeroDerivative;
            if (TrySolveLinear(2 * a, b, out zeroDerivative))
            {
                if (interval.Start < zeroDerivative && zeroDerivative < interval.End)
                {
                    pointsToCheck.Add(zeroDerivative);
                }
            }

            pointsToCheck.Add(interval.End);

            foreach (double x in pointsToCheck)
            {
                Point localExtremum = new Point(x, a * x * x + b * x + c);

                yield return localExtremum;
            }
        }

        /// <summary>
        /// Finds the local extrema of a cubic function a*x*x*x + b*x*x + c*x + d.
        /// </summary>
        /// <param name="interval">The interval.</param>
        /// <param name="a">The a coeficient.</param>
        /// <param name="b">The b coeficient.</param>
        /// <param name="c">The c coeficient.</param>
        /// <param name="d">The free coeficient.</param>
        /// <returns>The local extrema.</returns>
        public static IEnumerable<Point> FindLocalExtrema(Interval interval, double a, double b, double c, double d)
        {
            List<double> pointsToCheck = new List<double>();
            pointsToCheck.Add(interval.Start);

            double[] zeroDerivatives;
            if (TrySolveQuadratic(3 * a, 2 * b, c, out zeroDerivatives))
            {
                foreach (double zeroDerivative in zeroDerivatives)
                {
                    if (interval.Start < zeroDerivative && zeroDerivative < interval.End)
                    {
                        pointsToCheck.Add(zeroDerivative);
                    }
                }
            }

            pointsToCheck.Add(interval.End);

            foreach (double x in pointsToCheck)
            {
                Point localExtremum = new Point(x, a * x * x * x + b * x * x + c * x + d);

                yield return localExtremum;
            }
        }

        /// <summary>
        /// Solves the quadratic equation a*x*x + b*x + c = 0.
        /// </summary>
        /// <param name="a">The coeficient before x*x.</param>
        /// <param name="b">The coeficient before x.</param>
        /// <param name="c">The free coeficient.</param>
        /// <param name="x">The x.</param>
        /// <returns>True if finite number of finite real solutions exist. Else returns false.</returns>
        public static bool TrySolveQuadratic(double a, double b, double c, out double[] x)
        {
            x = SolveQuadratic(a, b, c).Where((solution) =>
                {
                    return !double.IsNaN(solution) && !double.IsInfinity(solution);
                }).ToArray();

            return x.Length > 0;
        }

        /// <summary>
        /// Solves the quadratic equation a*x*x + b*x + c = 0.
        /// </summary>
        /// <param name="a">The coeficient before x*x.</param>
        /// <param name="b">The coeficient before x.</param>
        /// <param name="c">The free coeficient.</param>
        /// <returns>Returns the real solutions of the equation. Returns PositiveInfinity if every real number is solution.</returns>
        public static IEnumerable<double> SolveQuadratic(double a, double b, double c)
        {
            if (a.IsZero())
            {
                double linearSolve = SolveLinear(b, c);

                if (!double.IsNaN(linearSolve))
                {
                    yield return linearSolve;
                }
            }
            else
            {
                double d = b * b - 4 * a * c;

                if (d.IsZero())
                {
                    yield return -b / (2 * a);
                }
                else if (d > 0)
                {
                    double sqrtD = Math.Sqrt(d);

                    double x1 = (-b - sqrtD) / (2 * a);
                    double x2 = (-b + sqrtD) / (2 * a);

                    if(a > 0)
                    {
                        yield return x1;
                        yield return x2;
                    }
                    else
                    {
                        yield return x2;
                        yield return x1;
                    }
                }
            }
        }

        /// <summary>
        /// Solves the equation a*x + b = 0
        /// </summary>
        /// <param name="a">The coeficient before x.</param>
        /// <param name="b">The free coeficient.</param>
        /// <param name="x">The x.</param>
        /// <returns>True if finite number of finite real solution exists. Else returns false.</returns>
        public static bool TrySolveLinear(double a, double b, out double x)
        {
            x = SolveLinear(a, b);

            return !double.IsNaN(x) && !double.IsInfinity(x);
        }

        /// <summary>
        /// Solves the equation a*x + b = 0
        /// </summary>
        /// <param name="a">The coeficient before x.</param>
        /// <param name="b">The free coeficient.</param>
        /// <returns>The appropriate value of x. Returns NaN if no solution is available. Returns PositiveInfinity when every x is solution.</returns>
        public static double SolveLinear(double a, double b)
        {
            if (a.IsZero())
            {
                return b.IsZero() ? double.PositiveInfinity : double.NaN;
            }
            else
            {
                return -b / a;
            }
        }

        public static double Clamp(double value, double minimum, double maximum, int precision)
        {
            double clampedValue = 0;

            if (value > maximum)
            {
                clampedValue = maximum;
            }
            else
            {
                clampedValue = value;
            }

            if (clampedValue < minimum)
            {
                clampedValue = minimum;
            }

            return Math.Round(clampedValue, precision);
        }

        public static double CustomMod(double number)
        {
            if (number > 0)
            {
                return number - Math.Floor(number);
            }
            else if (number < 0)
            {
                double abs = Math.Abs(number);

                return 1 - (abs - Math.Floor(abs));
            }

            return 0;
        }

        public static string IntegerToRomanString(int number)
        {
            if (number < 0)
            {
                throw new ArgumentException("Value must be positive.", "number");
            }

            if (number == 0)
            {
                return "n";
            }

            if (number > 32767)
            {
                number++;
            }
            number %= 32767 + 1;

            StringBuilder result = new StringBuilder();

            for (int i = 0; i < 13; i++)
            {
                while (number >= romanNumeralValues[i])
                {
                    number -= romanNumeralValues[i];
                    result.Append(romanNumerals[i]);
                }
            }

            return result.ToString();
        }

        public static double GetDistance(Point p1, Point p2)
        {
            return Math.Sqrt(((p1.X - p2.X) * (p1.X - p2.X)) + ((p1.Y - p2.Y) * (p1.Y - p2.Y)));
        }
    }
}
