﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

namespace Telerik.Windows.Documents.Utilities
{
    internal abstract class ReaderBase
    {
        private readonly byte[] data;
        private readonly Stack<long> beginReadingPositions;

        private long position;

        public ReaderBase(byte[] data)
           : this()
        {
            this.data = data;
        }

        internal ReaderBase()
        {
            this.position = 0;
            this.beginReadingPositions = new Stack<long>();
        }

        public virtual bool EndOfFile
        {
            get
            {
                if (this.data == null)
                {
                    return true;
                }

                return this.Position >= this.Length;
            }
        }

        public virtual long Length
        {
            get
            {
                return this.data.Length;
            }
        }

        public byte[] Data
        {
            get
            {
                return this.data;
            }
        }

        public virtual void BeginReadingBlock()
        {
            this.beginReadingPositions.Push(this.position);
        }

        public virtual void EndReadingBlock()
        {
            if (this.beginReadingPositions.Count > 0)
            {
                this.position = this.beginReadingPositions.Pop();
            }
        }

        public virtual byte Read()
        {
            if (this.EndOfFile)
            {
                Guard.ThrowPositionOutOfRangeException();
            }

            return this.data[this.position++];
        }

        public virtual byte Peek(int skip = 0)
        {
            if (this.position + skip >= this.data.Length)
            {
                Guard.ThrowPositionOutOfRangeException();
            }

            return this.data[this.position + skip];
        }

        public virtual long Position
        {
            get
            {
                return this.position;
            }
        }

        public int Read(byte[] buffer, int count)
        {
            int read = 0;

            for (int i = 0; i < count; i++)
            {
                if (this.Position >= this.Length)
                {
                    return read;
                }

                buffer[i] = this.Read();
                read++;
            }

            return read;
        }

        public int ReadBE(byte[] buffer, int count)
        {
            int read = 0;

            for (int i = count - 1; i >= 0; i--)
            {
                if (this.Position >= this.Length)
                {
                    return read;
                }

                buffer[i] = this.Read();
                read++;
            }

            return read;
        }

        public virtual void Seek(long offset, SeekOrigin origin)
        {
            switch (origin)
            {
                case SeekOrigin.Begin:
                    this.position = offset;
                    break;
                case SeekOrigin.Current:
                    this.position += offset;
                    break;
                case SeekOrigin.End:
                    this.position = this.Length - offset;
                    break;
            }
        }
    }
}