﻿using System;
using System.ComponentModel;

namespace Telerik.Windows.Documents.Utilities
{
    internal class ResourceLoader : IDisposable
    {
        #region Classes

        class CancellationToken
        {
            public CancellationToken(object argument)
            {
                this.Argument = argument;
            }

            public object Argument { get; set; }

            public DoWorkEventArgs DoWorkArgs { get; set; }
        }

        #endregion

        #region Constructors

        public ResourceLoader()
        {
            this.isLoading = false;
        }

        #endregion

        #region Fields

        private bool isLoading;
        private CancellationToken currentThreadToken;
        private BackgroundWorker backgroundWorker;

        #endregion

        #region Events

        internal event EventHandler<DoWorkEventArgs> DoWork;
        internal event EventHandler<WorkCompletedEventArgs> WorkCompleted;

        private void OnWorkCompleted(WorkCompletedEventArgs args)
        {
            if (this.WorkCompleted != null)
            {
                this.WorkCompleted(this, args);
            }
        }

        private void OnDoWork(DoWorkEventArgs args)
        {
            if (this.DoWork != null)
            {
                this.DoWork(this, args);
            }
        }

        #endregion

        #region Properties

        public bool IsLoading
        {
            get
            {
                return this.isLoading;
            }
        }

        #endregion

        #region Methods

        private void AttachToBackgroundWorkerEvents(BackgroundWorker worker)
        {
            if (worker == null)
            {
                return;
            }
            worker.DoWork += this.Worker_DoWork;
            worker.RunWorkerCompleted += this.Worker_RunWorkerCompleted;
        }

        private void DetachToBackgroundWorkerEvents(BackgroundWorker worker)
        {
            if (worker == null)
            {
                return;
            }
            worker.DoWork -= this.Worker_DoWork;
            worker.RunWorkerCompleted -= this.Worker_RunWorkerCompleted;
        }

        internal void Load(object argument)
        {
            if (this.isLoading)
            {
                throw new InvalidOperationException("The Load method has been already called.");
            }

            this.isLoading = true;
            this.currentThreadToken = new CancellationToken(argument);
            this.DetachToBackgroundWorkerEvents(this.backgroundWorker);
            this.backgroundWorker = new BackgroundWorker();
            this.backgroundWorker.WorkerReportsProgress = false;
            this.backgroundWorker.WorkerSupportsCancellation = true;
            this.AttachToBackgroundWorkerEvents(this.backgroundWorker);

            this.backgroundWorker.RunWorkerAsync(this.currentThreadToken);
        }

        internal void StopLoading()
        {
            if (this.currentThreadToken == null)
            {
                return;
            }
            
            if (this.currentThreadToken.DoWorkArgs != null)
            {
                this.currentThreadToken.DoWorkArgs.Canceled = true;
            }
            
            if (this.backgroundWorker != null)
            {
                this.backgroundWorker.CancelAsync();
            }
            
            this.currentThreadToken = null;
            this.isLoading = false;
        }

        public void Dispose()
        {
#if WPF
            if (this.backgroundWorker != null)
            {
                this.backgroundWorker.Dispose();
            }
#endif
        }

        #endregion

        #region EventHandlers

        private void Worker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            CancellationToken token = (CancellationToken)e.Result;
            if (token != null && !token.DoWorkArgs.Canceled && !e.Cancelled && e.Error == null)
            {
                this.OnWorkCompleted(new WorkCompletedEventArgs { Result = token.DoWorkArgs.Result });
            }
         
            this.isLoading = false;
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
        private void Worker_DoWork(object sender, System.ComponentModel.DoWorkEventArgs e)
        {
            BackgroundWorker bw = (BackgroundWorker)sender;
            CancellationToken token = (CancellationToken)e.Argument;
            if (token == null)
            {
                return;
            }
            DoWorkEventArgs args = new DoWorkEventArgs();
            args.Argument = token.Argument;
            token.DoWorkArgs = args;
            if (e.Cancel || bw.CancellationPending)
            {
                return;
            }
            try
            {
                this.OnDoWork(args);
            }
            catch
            {
            }
            if (!args.Canceled)
            {
                e.Result = token;
            }
        }
        #endregion
    }
}
