﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

namespace Telerik.Windows.Documents.Utilities
{
    internal abstract class WriterBase : IDisposable
    {
        private readonly MemoryStream data;
        private readonly Stack<long> beginWritingPositions;

        /// <summary>
        /// Initializes a new instance of the WriterBase class.
        /// </summary>
        public WriterBase()
        {
            this.data = new MemoryStream();
            this.beginWritingPositions = new Stack<long>();
        }

        /// <summary>
        /// Gets writer data.
        /// </summary>
        public byte[] Data
        {
            get
            {
                return this.data.ToArray();
            }
        }

        public bool EndOfFile
        {
            get
            {
                if (this.data == null)
                {
                    return true;
                }

                return this.Position >= this.Length;
            }
        }

        public long Length
        {
            get
            {
                return this.data.Length;
            }
        }

        public virtual long Position
        {
            get
            {
                return this.data.Position;
            }
        }

        public virtual void BeginWritingBlock()
        {
            this.beginWritingPositions.Push(this.data.Position);
        }

        public void Dispose()
        {
            this.data.Dispose();
        }

        public virtual void EndWritingBlock()
        {
            if (this.beginWritingPositions.Count > 0)
            {
                this.data.Position = this.beginWritingPositions.Pop();
            }
        }

        /// <summary>
        /// Writes single byte.
        /// </summary>
        /// <param name="byteToWrite"></param>
        public virtual void Write(byte byteToWrite)
        {
            this.data.WriteByte(byteToWrite);
        }

        /// <summary>
        /// Writes a block of bytes.
        /// </summary>
        /// <param name="buffer">Buffer.</param>
        /// <param name="count">Bytes count.</param>
        public void Write(byte[] buffer, int count)
        {
            this.data.Write(buffer, 0, count);
        }

        /// <summary>
        /// Writes buffer in reverce direction.
        /// </summary>
        /// <param name="buffer">Buffer.</param>
        /// <param name="count">Bytes count.</param>
        /// <returns></returns>
        public void WriteBE(byte[] buffer, int count)
        {
            for (int i = count - 1; i >= 0; i--)
            {
                this.Write(buffer[i]);
            }
        }

        /// <summary>
        /// Sets the position.
        /// </summary>
        /// <param name="offset"></param>
        /// <param name="origin"></param>
        public virtual void Seek(long offset, SeekOrigin origin)
        {
            this.data.Seek(offset, origin);
        }
    }
}
