﻿using System;
using System.IO;
using System.Linq;
using System.Windows;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Core.Imaging;
using Telerik.Windows.Documents.Core.Tests.Utils;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Decoder;
using Telerik.Windows.Documents.Core.Imaging.Jpeg;
using Telerik.Windows.Documents.Core.Imaging.Jpeg2000.Decoder;
using Telerik.Windows.Documents.Core.Imaging.Jpeg2000;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Tests.Imaging
{
    [TestClass]
    public class ImageSizeDecodersTests
    {
        private static readonly Size commonFormatImageSize = new Size(329, 259);
        private static readonly Size iconImageSize = new Size(48,48);
        private static readonly byte[] emptyBytes = new byte[0];
        private static readonly byte[] invalidImageBytes = Enumerable.Repeat<byte>(2, 64).ToArray();

        [TestMethod]
        public void CanDecode_WhenImageExtensionContainsDot_ReturnsCorrectDecoder()
        {
            PngImageSizeDecoder pngDecoder = new PngImageSizeDecoder();
            Assert.IsTrue(pngDecoder.CanDecode(".png"));
        }

        [TestMethod]
        public void CanDecode_WhenImageExtensionContainsDotAndSpace_ReturnsCorrectDecoder()
        {
            PngImageSizeDecoder pngDecoder = new PngImageSizeDecoder();
            Assert.IsTrue(pngDecoder.CanDecode(" .png"));
        }

        [TestMethod]
        public void CanDecode_WhenImageExtensionDoesNotContainDot_ReturnsCorrectDecoder()
        {
            PngImageSizeDecoder pngDecoder = new PngImageSizeDecoder();
            Assert.IsTrue(pngDecoder.CanDecode("png"));
        }

        [TestMethod]
        public void CanDecode_WhenImageExtensionHaveTrailingSpaces_ReturnsCorrectDecoder()
        {
            PngImageSizeDecoder pngDecoder = new PngImageSizeDecoder();
            Assert.IsTrue(pngDecoder.CanDecode(" png "));
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void CanDecode_WhenImageExtensionIsNull_ThrowArgumentNullException()
        {
            PngImageSizeDecoder pngDecoder = new PngImageSizeDecoder();
            Assert.IsTrue(pngDecoder.CanDecode(null));
        }

        [TestMethod]
        public void TryDecodeEmptyImageByteArray_WhenJpegHeadersDecoder()
        {
            JpegHeadersDecoder decoder = new JpegHeadersDecoder(emptyBytes);
            JpegImageInfo imageInfo;
            Assert.IsFalse(decoder.TryDecodeJpegImage(out imageInfo));
        }

        [TestMethod]
        public void TryDecodeEmptyImageStream_WhenJpegHeadersDecoder()
        {
            JpegHeadersDecoder decoder = new JpegHeadersDecoder(new MemoryStream());
            JpegImageInfo imageInfo;
            Assert.IsFalse(decoder.TryDecodeJpegImage(out imageInfo));
        }

        [TestMethod]
        public void TryDecodeEmptyImage_WhenJpeg2000HeadersDecoder()
        {
            Jpeg2000HeadersDecoder decoder = new Jpeg2000HeadersDecoder(new byte[0]);
            Jpeg2000ImageInfo imageInfo;
            Assert.IsFalse(decoder.TryDecodeJpeg2000Image(out imageInfo));
        }

        [TestMethod]        
        [ExpectedException(typeof(InvalidOperationException))]
        public void TryDecodeEmptyImage_WhenJpegDecoder()
        {
            JpegDecoder decoder = new JpegDecoder(new byte[0]);
            decoder.Decode();
        }

        [TestMethod]
        public void PngImageSizeDecoder_WhenPngAndImageIsFromStream_DecodeImageSize()
        {
            AssertSizeIsDecodedFromStream(new PngImageSizeDecoder(), commonFormatImageSize, "png");
        }

        [TestMethod]
        public void JpegImageSizeDecoder_WhenJpgAndImageIsFromStream_DecodeImageSize()
        {
            AssertSizeIsDecodedFromStream(new JpegImageSizeDecoder(), commonFormatImageSize, "jpg");
        }

        [TestMethod]
        public void BitmapImageSizeDecoder_WhenBmpAndImageIsFromStream_DecodeImageSize()
        {
            AssertSizeIsDecodedFromStream(new BitmapImageSizeDecoder(), commonFormatImageSize, "bmp");
        }

        [TestMethod]
        public void GifImageSizeDecoder_WhenGifAndImageIsFromStream_DecodeImageSize()
        {
            AssertSizeIsDecodedFromStream(new GifImageSizeDecoder(), commonFormatImageSize, "gif");
        }

        [TestMethod]
        public void TifImageSizeDecoder_WhenTifAndImageIsFromStream_DecodeImageSize()
        {
            AssertSizeIsDecodedFromStream(new TifImageSizeDecoder(), commonFormatImageSize, "tif");
        }

        [TestMethod]
        public void MetafileImageSizeDecoder_WhenIcoAndImageIsFromStream_DecodeImageSize()
        {
            AssertSizeIsDecodedFromStream(new MetafileImageSizeDecoder(), iconImageSize, "ico");
        }

        [TestMethod]
        public void MetafileImageSizeDecoder_WhenEmfAndImageIsFromStream_DecodeImageSize()
        {
            AssertSizeIsDecodedFromStream(new MetafileImageSizeDecoder(), new Size(1135,337), "emf");
        }

        [TestMethod]
        public void MetafileImageSizeDecoder_WhenWmfAndImageIsFromStream_DecodeImageSize()
        {
            AssertSizeIsDecodedFromStream(new MetafileImageSizeDecoder(), new Size(3404, 1010), "wmf");
        }

        [TestMethod]
        public void PngImageSizeDecoder_WhenPngAndImageIsFromByteArray_DecodeImageSize()
        {
            AssertSizeIsDecodedFromByteArray(new PngImageSizeDecoder(), commonFormatImageSize, "png");
        }

        [TestMethod]
        public void JpegImageSizeDecoder_WhenJpgAndImageIsFromByteArray_DecodeImageSize()
        {
            AssertSizeIsDecodedFromByteArray(new JpegImageSizeDecoder(), commonFormatImageSize, "jpg");
        }

        [TestMethod]
        public void BitmapImageSizeDecoder_WhenBmpAndImageIsFromByteArray_DecodeImageSize()
        {
            AssertSizeIsDecodedFromByteArray(new BitmapImageSizeDecoder(), commonFormatImageSize, "bmp");
        }

        [TestMethod]
        public void GifImageSizeDecoder_WhenGifAndImageIsFromByteArray_DecodeImageSize()
        {
            AssertSizeIsDecodedFromByteArray(new GifImageSizeDecoder(), commonFormatImageSize, "gif");
        }

        [TestMethod]
        public void TifImageSizeDecoder_WhenTifAndImageIsFromByteArray_DecodeImageSize()
        {
            AssertSizeIsDecodedFromByteArray(new TifImageSizeDecoder(), commonFormatImageSize, "tif");
        }

        [TestMethod]
        public void MetafileImageSizeDecoder_WhenIcoAndImageIsFromByteArray_DecodeImageSize()
        {
            AssertSizeIsDecodedFromByteArray(new MetafileImageSizeDecoder(), iconImageSize, "ico");
        }

        [TestMethod]
        public void MetafileImageSizeDecoder_WhenEmfAndImageIsFromByteArray_DecodeImageSize()
        {
            AssertSizeIsDecodedFromByteArray(new MetafileImageSizeDecoder(), new Size(1135, 337), "emf");
        }

        [TestMethod]
        public void MetafileImageSizeDecoder_WhenWmfAndImageIsFromByteArray_DecodeImageSize()
        {
            AssertSizeIsDecodedFromByteArray(new MetafileImageSizeDecoder(), new Size(3404, 1010), "wmf");
        }

        [TestMethod]
        public void PngImageSizeDecoder_WhenByteArrayIsEmpty_DoesNotDecodeImageSizeAndReturnsSizeEmpty()
        {
            AssertSizeIsNotDecoded(new PngImageSizeDecoder(), emptyBytes);
        }

        [TestMethod]
        public void JpegImageSizeDecoder_WhenByteArrayIsEmpty_DoesNotDecodeImageSizeAndReturnsSizeEmpty()
        {
            AssertSizeIsNotDecoded(new JpegImageSizeDecoder(), emptyBytes);
        }

        [TestMethod]
        public void BitmapImageSizeDecoder_WhenByteArrayIsEmpty_DoesNotDecodeImageSizeAndReturnsSizeEmpty()
        {
            AssertSizeIsNotDecoded(new BitmapImageSizeDecoder(), emptyBytes);
        }

        [TestMethod]
        public void GifImageSizeDecoder_WhenByteArrayIsEmpty_DoesNotDecodeImageSizeAndReturnsSizeEmpty()
        {
            AssertSizeIsNotDecoded(new GifImageSizeDecoder(), emptyBytes);
        }

        [TestMethod]
        public void TifImageSizeDecoder_WhenByteArrayIsEmpty_DoesNotDecodeImageSizeAndReturnsSizeEmpty()
        {
            AssertSizeIsNotDecoded(new TifImageSizeDecoder(), emptyBytes);
        }

        [TestMethod]
        public void MetafileImageSizeDecoder_WhenByteArrayIsEmpty_DoesNotDecodeImageSizeAndReturnsSizeEmpty()
        {
            AssertSizeIsNotDecoded(new MetafileImageSizeDecoder(), emptyBytes);
        }

        [TestMethod]
        public void PngImageSizeDecoder_WhenByteArrayIsInvalidImageByteArray_DoesNotDecodeImageSizeAndReturnsSizeEmpty()
        {
            AssertSizeIsNotDecoded(new PngImageSizeDecoder(), invalidImageBytes);
        }

        [TestMethod]
        public void JpegImageSizeDecoder_WhenByteArrayIsInvalidImageByteArray_DoesNotDecodeImageSizeAndReturnsSizeEmpty()
        {
            AssertSizeIsNotDecoded(new JpegImageSizeDecoder(), invalidImageBytes);
        }

        [TestMethod]
        public void BitmapImageSizeDecoder_WhenByteArrayIsInvalidImageByteArray_DoesNotDecodeImageSizeAndReturnsSizeEmpty()
        {
            AssertSizeIsNotDecoded(new BitmapImageSizeDecoder(), invalidImageBytes);
        }

        [TestMethod]
        public void GifImageSizeDecoder_WhenByteArrayIsInvalidImageByteArray_DoesNotDecodeImageSizeAndReturnsSizeEmpty()
        {
            AssertSizeIsNotDecoded(new GifImageSizeDecoder(), invalidImageBytes);
        }

        [TestMethod]
        public void TifImageSizeDecoder_WhenByteArrayIsInvalidImageByteArray_DoesNotDecodeImageSizeAndReturnsSizeEmpty()
        {
            AssertSizeIsNotDecoded(new TifImageSizeDecoder(), invalidImageBytes);
        }

        [TestMethod]
        public void MetafileImageSizeDecoder_WhenByteArrayIsInvalidImageByteArray_DoesNotDecodeImageSizeAndReturnsSizeEmpty()
        {
            AssertSizeIsNotDecoded(new MetafileImageSizeDecoder(), invalidImageBytes);
        }

        [TestMethod]
        public void ImageSizeDecodersManager_WhenByteArrayIsNull_DoesNotDecodeImageSizeAndReturnsSizeEmpty()
        {
            Size resultSize = ImageSizeDecodersManager.GetSize(imageBytes: null, imageTypeExtension: "png");

            // Assert
            Assert.AreEqual(Size.Empty, resultSize, "Size is decoded, but it shouldn't!");
        }

        private static void AssertSizeIsDecodedFromStream(IImageSizeDecoder decoder, Size expectedSize, string imageExtension)
        {
            // Arrange
            Size size = new Size(double.PositiveInfinity, double.PositiveInfinity);

            // Act
            using (Stream stream = TestHelper.GetImageStream(imageExtension))
            {
                size = decoder.DecodeSize(stream);
            }

            // Assert
            Assert.AreEqual(expectedSize, size);
        }

        private static void AssertSizeIsDecodedFromByteArray(IImageSizeDecoder decoder, Size expectedSize, string imageExtension)
        {
            // Arrange
            Size size = new Size(double.PositiveInfinity, double.PositiveInfinity);

            // Act
            using (Stream stream = TestHelper.GetImageStream(imageExtension))
            {
                size = decoder.DecodeSize(stream.ReadAllBytes());
            }

            // Assert
            Assert.AreEqual(expectedSize, size);
        }

        private static void AssertSizeIsDecodedFromBytes(IImageSizeDecoder decoder, Size expectedSize, string imageExtension)
        {
            // Arrange
            Size size = new Size(double.PositiveInfinity, double.PositiveInfinity);

            // Act
            using (Stream stream = TestHelper.GetImageStream(imageExtension))
            {
                size = decoder.DecodeSize(stream.ReadAllBytes());
            }

            // Assert
            Assert.AreEqual(expectedSize, size);
        }

        private static void AssertSizeIsNotDecoded(IImageSizeDecoder decoder, byte[] bytes)
        {
            // Act
            Size resultSize = decoder.DecodeSize(bytes);

            // Assert
            Assert.AreEqual(Size.Empty, resultSize, "Size is decoded, but it shouldn't!");
        }
    }
}