﻿using System;
using System.Windows;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Core.Input;
using Telerik.Windows.Documents.Core.Input.PointerHandlers;
using System.Windows.Input;
using Telerik.Windows.Documents.Core.Input.PointerHandlers.Args;

namespace Telerik.Windows.Documents.Core.Tests.InputTests
{
    [TestClass]
    public class PointerHandlerControllerBaseTests
    {
        private class TestPointerHandler : IPointerHandler
        {
            private readonly Point topLeft;
            private readonly Point bottomRight;

            public string Name
            {
                get;
                private set;
            }

            public PointerHandlersControllerBase Controller
            {
                get;
                private set;
            }

            public int LeftButtonDownCounts { get; private set; }
            public int LeftButtonDownHandlesCount { get; private set; }
            public int LeftButtonUpCounts { get; private set; }
            public int RightButtonDownCounts { get; private set; }
            public int RightButtonDownHandlesCount { get; private set; }
            public int RightButtonUpCounts { get; private set; }
            public int CaptureCounts { get; set; }
            public int MoveCounts { get; private set; }

            public bool CanMove { get; set; }
            public bool CanGetLeftPress { get; set; }
            public bool CanGetRightPress { get; set; }

            public TestPointerHandler(Point topLeft, Point bottomRight, string name, PointerHandlersControllerBase controller)
            {
                this.topLeft = topLeft;
                this.bottomRight = bottomRight;
                this.Name = name;
                this.Controller = controller;

                this.CanMove = true;
                this.CanGetLeftPress = true;
                this.CanGetRightPress = true;

                this.Initialize();
            }

            public void ResetCounters()
            {
                this.LeftButtonDownCounts = 0;
                this.LeftButtonDownHandlesCount = 0;
                this.LeftButtonUpCounts = 0;
                this.RightButtonDownCounts = 0;
                this.RightButtonDownHandlesCount = 0;
                this.RightButtonUpCounts = 0;
                this.CaptureCounts = 0;
                this.MoveCounts = 0;
            }

            public void Initialize()
            {
                this.ResetCounters();
            }

            public void Release()
            {
            }

            public bool ShouldHandle(Point position, SourceType source)
            {
                return position.X >= this.topLeft.X && position.X <= this.bottomRight.X
                    && position.Y >= this.topLeft.Y && position.Y <= this.bottomRight.Y;
            }

            public Cursor GetCursor(Point position, SourceType source)
            {
                return this.ShouldHandle(position, source) ? Cursors.Hand : Cursors.None;
            }

            public void RegisterLeftButtonDown(PointerLeftButtonDownArgs args)
            {
                if (this.CanGetLeftPress)
                {
                    if (!args.IsSimulated || this.CanMove)
                    {
                        args.IsHandled = true;
                        this.LeftButtonDownHandlesCount++;
                    }
                }

                this.LeftButtonDownCounts++;
            }

            public void RegisterLeftButtonUp(PointerLeftButtonUpArgs args)
            {
                if (this.CanGetLeftPress)
                {
                    args.IsHandled = true;
                }

                this.LeftButtonUpCounts++;
            }

            public void RegisterRightButtonDown(PointerRightButtonDownArgs args)
            {
                if (this.CanGetRightPress)
                {
                    if (!args.IsSimulated || this.CanMove)
                    {
                        args.IsHandled = true;
                        this.RightButtonDownHandlesCount++;
                    }
                }

                this.RightButtonDownCounts++;
            }

            public void RegisterRightButtonUp(PointerRightButtonUpArgs args)
            {
                if (this.CanGetRightPress)
                {
                    args.IsHandled = true;
                }

                this.RightButtonUpCounts++;
            }

            public void RegisterMove(PointerMoveArgs args)
            {
                if (this.CanMove)
                {
                    args.IsHandled = true;
                }
                else
                {
                    this.Controller.CaptureNext(this);
                }

                this.MoveCounts++;
            }

            public void RegisterLostCapture(PointerLostCaptureArgs args)
            {
                args.IsHandled = true;
                this.CaptureCounts--;
            }

            public void Capture(PointerCaptureArgs args)
            {
                this.CaptureCounts++;
            }


            public void ReleaseCapture(PointerReleaseCaptureArgs args)
            {
                this.CaptureCounts--;
            }

            public override string ToString()
            {
                return String.Format("TestPointerHandler: {0}", this.Name);
            }
        }

        private class TestPointerHandlersController : PointerHandlersControllerBase
        {
            public TestPointerHandler FirstHandler { get; private set; }
            public TestPointerHandler SecondHandler { get; private set; }
            public TestPointerHandler ThirdHandler { get; private set; }
            public TestPointerHandler ForthHandler { get; private set; }

            public TestPointerHandlersController()
            {
                this.FirstHandler = new TestPointerHandler(new Point(0, 0), new Point(20, 20), "FirstHandler", this);
                this.SecondHandler = new TestPointerHandler(new Point(10, 10), new Point(30, 30), "SecondHandler", this);
                this.ThirdHandler = new TestPointerHandler(new Point(10, 10), new Point(20, 20), "ThirdHandler", this);
                this.ForthHandler = new TestPointerHandler(new Point(0, 0), new Point(30, 30), "ForthHandler", this);

                this.Handlers.AddLast(this.FirstHandler);
                this.Handlers.AddLast(this.SecondHandler);
                this.Handlers.AddLast(this.ThirdHandler);
                this.Handlers.AddLast(this.ForthHandler);

                this.Initialize();
            }

            public void ResetCounters()
            {
                foreach (IPointerHandler handler in this.Handlers)
                {
                    (handler as TestPointerHandler).ResetCounters();
                }
            }

            internal override Cursor DefaultCursor
            {
                get { return Cursors.Arrow; }
            }
        }

        [TestMethod]
        public void PointerDownTest()
        {
            TestPointerHandlersController controller = new TestPointerHandlersController();
            Point point = new Point(15, 15);

            controller.RegisterLeftButtonDown(new PointerLeftButtonDownArgs() { Position = point });
            Assert.AreEqual(1, controller.FirstHandler.LeftButtonDownCounts);
            Assert.AreEqual(1, controller.FirstHandler.LeftButtonDownHandlesCount);
            Assert.AreEqual(0, controller.SecondHandler.LeftButtonDownCounts);
            Assert.AreEqual(0, controller.ThirdHandler.LeftButtonDownCounts);
            Assert.AreEqual(0, controller.ForthHandler.LeftButtonDownCounts);

            controller.FirstHandler.CanGetLeftPress = false;
            controller.SecondHandler.CanGetLeftPress = false;

            controller.RegisterLeftButtonDown(new PointerLeftButtonDownArgs() { Position = point });
            Assert.AreEqual(2, controller.FirstHandler.LeftButtonDownCounts);
            Assert.AreEqual(1, controller.FirstHandler.LeftButtonDownHandlesCount);
            Assert.AreEqual(1, controller.SecondHandler.LeftButtonDownCounts);
            Assert.AreEqual(0, controller.SecondHandler.LeftButtonDownHandlesCount);
            Assert.AreEqual(1, controller.ThirdHandler.LeftButtonDownCounts);
            Assert.AreEqual(1, controller.ThirdHandler.LeftButtonDownHandlesCount);
            Assert.AreEqual(0, controller.ForthHandler.LeftButtonDownCounts);

            controller.ResetCounters();

            controller.RegisterRightButtonDown(new PointerRightButtonDownArgs() { Position = point });
            Assert.AreEqual(1, controller.FirstHandler.RightButtonDownCounts);
            Assert.AreEqual(1, controller.FirstHandler.RightButtonDownHandlesCount);
            Assert.AreEqual(0, controller.SecondHandler.RightButtonDownCounts);
            Assert.AreEqual(0, controller.ThirdHandler.RightButtonDownCounts);
            Assert.AreEqual(0, controller.ForthHandler.RightButtonDownCounts);

            controller.FirstHandler.CanGetRightPress = false;
            controller.SecondHandler.CanGetRightPress = false;

            controller.RegisterRightButtonDown(new PointerRightButtonDownArgs() { Position = point });
            Assert.AreEqual(2, controller.FirstHandler.RightButtonDownCounts);
            Assert.AreEqual(1, controller.FirstHandler.RightButtonDownHandlesCount);
            Assert.AreEqual(1, controller.SecondHandler.RightButtonDownCounts);
            Assert.AreEqual(0, controller.SecondHandler.RightButtonDownHandlesCount);
            Assert.AreEqual(1, controller.ThirdHandler.RightButtonDownCounts);
            Assert.AreEqual(1, controller.ThirdHandler.RightButtonDownHandlesCount);
            Assert.AreEqual(0, controller.ForthHandler.RightButtonDownCounts);
        }

        [TestMethod]
        public void PointerUpTest()
        {
            TestPointerHandlersController controller = new TestPointerHandlersController();
            Point point = new Point(15, 15);

            controller.RegisterLeftButtonUp(new PointerLeftButtonUpArgs() { Position = point });
            Assert.AreEqual(0, controller.FirstHandler.LeftButtonUpCounts);
            Assert.AreEqual(0, controller.SecondHandler.LeftButtonUpCounts);
            Assert.AreEqual(0, controller.ThirdHandler.LeftButtonUpCounts);
            Assert.AreEqual(0, controller.ForthHandler.LeftButtonUpCounts);


            controller.RegisterLeftButtonDown(new PointerLeftButtonDownArgs() { Position = point });
            controller.RegisterLeftButtonUp(new PointerLeftButtonUpArgs() { Position = point });
            Assert.AreEqual(1, controller.FirstHandler.LeftButtonUpCounts);
            Assert.AreEqual(0, controller.SecondHandler.LeftButtonUpCounts);
            Assert.AreEqual(0, controller.ThirdHandler.LeftButtonUpCounts);
            Assert.AreEqual(0, controller.ForthHandler.LeftButtonUpCounts);

            controller.RegisterLeftButtonUp(new PointerLeftButtonUpArgs() { Position = point });
            Assert.AreEqual(1, controller.FirstHandler.LeftButtonUpCounts);
            Assert.AreEqual(0, controller.SecondHandler.LeftButtonUpCounts);
            Assert.AreEqual(0, controller.ThirdHandler.LeftButtonUpCounts);
            Assert.AreEqual(0, controller.ForthHandler.LeftButtonUpCounts);


            controller.FirstHandler.CanGetLeftPress = false;
            controller.SecondHandler.CanGetLeftPress = false;

            controller.RegisterLeftButtonDown(new PointerLeftButtonDownArgs() { Position = point });
            controller.RegisterLeftButtonUp(new PointerLeftButtonUpArgs() { Position = point });
            Assert.AreEqual(1, controller.FirstHandler.LeftButtonUpCounts);
            Assert.AreEqual(0, controller.SecondHandler.LeftButtonUpCounts);
            Assert.AreEqual(1, controller.ThirdHandler.LeftButtonUpCounts);
            Assert.AreEqual(0, controller.ForthHandler.LeftButtonUpCounts);


            controller.ResetCounters();


            controller.RegisterRightButtonUp(new PointerRightButtonUpArgs() { Position = point });
            Assert.AreEqual(0, controller.FirstHandler.RightButtonUpCounts);
            Assert.AreEqual(0, controller.SecondHandler.RightButtonUpCounts);
            Assert.AreEqual(0, controller.ThirdHandler.RightButtonUpCounts);
            Assert.AreEqual(0, controller.ForthHandler.RightButtonUpCounts);


            controller.RegisterRightButtonDown(new PointerRightButtonDownArgs() { Position = point });
            controller.RegisterRightButtonUp(new PointerRightButtonUpArgs() { Position = point });
            Assert.AreEqual(1, controller.FirstHandler.RightButtonUpCounts);
            Assert.AreEqual(0, controller.SecondHandler.RightButtonUpCounts);
            Assert.AreEqual(0, controller.ThirdHandler.RightButtonUpCounts);
            Assert.AreEqual(0, controller.ForthHandler.RightButtonUpCounts);

            controller.RegisterRightButtonUp(new PointerRightButtonUpArgs() { Position = point });
            Assert.AreEqual(1, controller.FirstHandler.RightButtonUpCounts);
            Assert.AreEqual(0, controller.SecondHandler.RightButtonUpCounts);
            Assert.AreEqual(0, controller.ThirdHandler.RightButtonUpCounts);
            Assert.AreEqual(0, controller.ForthHandler.RightButtonUpCounts);


            controller.FirstHandler.CanGetRightPress = false;
            controller.SecondHandler.CanGetRightPress = false;

            controller.RegisterRightButtonDown(new PointerRightButtonDownArgs() { Position = point });
            controller.RegisterRightButtonUp(new PointerRightButtonUpArgs() { Position = point });
            Assert.AreEqual(1, controller.FirstHandler.RightButtonUpCounts);
            Assert.AreEqual(0, controller.SecondHandler.RightButtonUpCounts);
            Assert.AreEqual(1, controller.ThirdHandler.RightButtonUpCounts);
            Assert.AreEqual(0, controller.ForthHandler.RightButtonUpCounts);
        }

        [TestMethod]
        public void PointerMoveFromLeftPressTest()
        {
            TestPointerHandlersController controller = new TestPointerHandlersController();

            controller.RegisterLeftButtonDown(new PointerLeftButtonDownArgs() { Position = new Point(15, 15) });

            controller.RegisterMove(new PointerMoveArgs() { Position = new Point(16, 16) });
            controller.RegisterMove(new PointerMoveArgs() { Position = new Point(18, 18) });
            controller.RegisterMove(new PointerMoveArgs() { Position = new Point(24, 24) });

            controller.RegisterLeftButtonUp(new PointerLeftButtonUpArgs() { Position = new Point(25, 25) });
            controller.RegisterLeftButtonUp(new PointerLeftButtonUpArgs() { Position = new Point(25, 25) });

            Assert.AreEqual(1, controller.FirstHandler.LeftButtonDownCounts);
            Assert.AreEqual(1, controller.FirstHandler.LeftButtonDownHandlesCount);
            Assert.AreEqual(0, controller.SecondHandler.LeftButtonDownCounts);
            Assert.AreEqual(0, controller.ThirdHandler.LeftButtonDownCounts);
            Assert.AreEqual(0, controller.ForthHandler.LeftButtonDownCounts);

            Assert.AreEqual(3, controller.FirstHandler.MoveCounts);
            Assert.AreEqual(0, controller.SecondHandler.MoveCounts);
            Assert.AreEqual(0, controller.ThirdHandler.MoveCounts);
            Assert.AreEqual(0, controller.ForthHandler.MoveCounts);

            Assert.AreEqual(1, controller.FirstHandler.LeftButtonUpCounts);
            Assert.AreEqual(0, controller.SecondHandler.LeftButtonUpCounts);
            Assert.AreEqual(0, controller.ThirdHandler.LeftButtonUpCounts);
            Assert.AreEqual(0, controller.ForthHandler.LeftButtonUpCounts);
        }

        [TestMethod]
        public void PointerMoveLowerLayerFromLeftPressTest()
        {
            TestPointerHandlersController controller = new TestPointerHandlersController();

            controller.FirstHandler.CanMove = false;
            controller.SecondHandler.CanMove = false;

            controller.RegisterLeftButtonDown(new PointerLeftButtonDownArgs() { Position = new Point(15, 15) });

            controller.RegisterMove(new PointerMoveArgs() { Position = new Point(16, 16) });
            controller.RegisterMove(new PointerMoveArgs() { Position = new Point(18, 18) });
            controller.RegisterMove(new PointerMoveArgs() { Position = new Point(24, 24) });

            controller.RegisterLeftButtonUp(new PointerLeftButtonUpArgs() { Position = new Point(25, 25) });
            controller.RegisterLeftButtonUp(new PointerLeftButtonUpArgs() { Position = new Point(25, 25) });

            Assert.AreEqual(1, controller.FirstHandler.LeftButtonDownCounts);
            Assert.AreEqual(1, controller.FirstHandler.LeftButtonDownHandlesCount);
            Assert.AreEqual(1, controller.SecondHandler.LeftButtonDownCounts);
            Assert.AreEqual(0, controller.SecondHandler.LeftButtonDownHandlesCount);
            Assert.AreEqual(1, controller.ThirdHandler.LeftButtonDownCounts);
            Assert.AreEqual(1, controller.ThirdHandler.LeftButtonDownHandlesCount);
            Assert.AreEqual(0, controller.ForthHandler.LeftButtonDownCounts);
            Assert.AreEqual(0, controller.ForthHandler.LeftButtonDownHandlesCount);

            Assert.AreEqual(1, controller.FirstHandler.MoveCounts);
            Assert.AreEqual(0, controller.SecondHandler.MoveCounts);
            Assert.AreEqual(2, controller.ThirdHandler.MoveCounts);
            Assert.AreEqual(0, controller.ForthHandler.MoveCounts);

            Assert.AreEqual(0, controller.FirstHandler.LeftButtonUpCounts);
            Assert.AreEqual(0, controller.SecondHandler.LeftButtonUpCounts);
            Assert.AreEqual(1, controller.ThirdHandler.LeftButtonUpCounts);
            Assert.AreEqual(0, controller.ForthHandler.LeftButtonUpCounts);
        }

        [TestMethod]
        public void PointerMoveFromRightPressTest()
        {
            TestPointerHandlersController controller = new TestPointerHandlersController();

            controller.RegisterRightButtonDown(new PointerRightButtonDownArgs() { Position = new Point(15, 15) });

            controller.RegisterMove(new PointerMoveArgs() { Position = new Point(16, 16) });
            controller.RegisterMove(new PointerMoveArgs() { Position = new Point(18, 18) });
            controller.RegisterMove(new PointerMoveArgs() { Position = new Point(24, 24) });

            controller.RegisterRightButtonUp(new PointerRightButtonUpArgs() { Position = new Point(25, 25) });
            controller.RegisterRightButtonUp(new PointerRightButtonUpArgs() { Position = new Point(25, 25) });

            Assert.AreEqual(1, controller.FirstHandler.RightButtonDownCounts);
            Assert.AreEqual(1, controller.FirstHandler.RightButtonDownHandlesCount);
            Assert.AreEqual(0, controller.SecondHandler.RightButtonDownCounts);
            Assert.AreEqual(0, controller.ThirdHandler.RightButtonDownCounts);
            Assert.AreEqual(0, controller.ForthHandler.RightButtonDownCounts);

            Assert.AreEqual(3, controller.FirstHandler.MoveCounts);
            Assert.AreEqual(0, controller.SecondHandler.MoveCounts);
            Assert.AreEqual(0, controller.ThirdHandler.MoveCounts);
            Assert.AreEqual(0, controller.ForthHandler.MoveCounts);

            Assert.AreEqual(1, controller.FirstHandler.RightButtonUpCounts);
            Assert.AreEqual(0, controller.SecondHandler.RightButtonUpCounts);
            Assert.AreEqual(0, controller.ThirdHandler.RightButtonUpCounts);
            Assert.AreEqual(0, controller.ForthHandler.RightButtonUpCounts);
        }

        [TestMethod]
        public void PointerMoveLowerLayerFromRightPressTest()
        {
            TestPointerHandlersController controller = new TestPointerHandlersController();

            controller.FirstHandler.CanMove = false;
            controller.SecondHandler.CanMove = false;

            controller.RegisterRightButtonDown(new PointerRightButtonDownArgs() { Position = new Point(15, 15) });

            controller.RegisterMove(new PointerMoveArgs() { Position = new Point(16, 16) });
            controller.RegisterMove(new PointerMoveArgs() { Position = new Point(18, 18) });
            controller.RegisterMove(new PointerMoveArgs() { Position = new Point(24, 24) });

            controller.RegisterRightButtonUp(new PointerRightButtonUpArgs() { Position = new Point(25, 25) });
            controller.RegisterRightButtonUp(new PointerRightButtonUpArgs() { Position = new Point(25, 25) });

            Assert.AreEqual(1, controller.FirstHandler.RightButtonDownCounts);
            Assert.AreEqual(1, controller.FirstHandler.RightButtonDownHandlesCount);
            Assert.AreEqual(1, controller.SecondHandler.RightButtonDownCounts);
            Assert.AreEqual(0, controller.SecondHandler.RightButtonDownHandlesCount);
            Assert.AreEqual(1, controller.ThirdHandler.RightButtonDownCounts);
            Assert.AreEqual(1, controller.ThirdHandler.RightButtonDownHandlesCount);
            Assert.AreEqual(0, controller.ForthHandler.RightButtonDownCounts);
            Assert.AreEqual(0, controller.ForthHandler.RightButtonDownHandlesCount);

            Assert.AreEqual(1, controller.FirstHandler.MoveCounts);
            Assert.AreEqual(0, controller.SecondHandler.MoveCounts);
            Assert.AreEqual(2, controller.ThirdHandler.MoveCounts);
            Assert.AreEqual(0, controller.ForthHandler.MoveCounts);

            Assert.AreEqual(0, controller.FirstHandler.RightButtonUpCounts);
            Assert.AreEqual(0, controller.SecondHandler.RightButtonUpCounts);
            Assert.AreEqual(1, controller.ThirdHandler.RightButtonUpCounts);
            Assert.AreEqual(0, controller.ForthHandler.RightButtonUpCounts);
        }
    }
}
