﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Core.Fonts;
using Telerik.Windows.Documents.Core.Tests.Utils;
using Telerik.Windows.Documents.Core.TextMeasurer;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Tests.TextMeasurementTests
{
    [TestClass]
    public class TestRadTextMeasurer
    {
        private readonly RadTextMeasurer textMeasurer = new RadTextMeasurer();

        [TestMethod]
        public void TestMeasureWithUnknownFont()
        {
            FontProperties fontProperties = new FontProperties(new FontFamily("SomeFontThatIsNotPresentForSure"));
            try
            {
                TextMeasurementInfo result = this.textMeasurer.MeasureText(TestHelper.SampleString, fontProperties);
            }
            catch
            {
                Assert.Fail("Measure with unknown font problem.");
            }
        }
#if WPF
        [TestMethod]
        public void TestWarmUpDifferentFonts()
        {
            TestHelper.ExecuteActionWithDifferentFontProperties((fontProperties) =>
            {
                this.textMeasurer.MeasureText(TestHelper.SampleString, fontProperties);
            });
        }

        [TestMethod]
        public void TestCorrectnessDifferentFonts()
        {
            TestHelper.ExecuteActionWithDifferentFontProperties((fontProperties) =>
            {
                TextBlock block = TestHelper.CreateTextBlock(fontProperties);
                this.AssertAreEqual(block, new TextProperties("a a"), fontProperties);
            });

        }

#if DEBUG
        // Too unstable.
        ////[TestMethod]
#endif
        public void TestSpeedAndCorrectnessWithLotsOfWordsDifferentFonts()
        {
            TestHelper.ExecuteActionWithDifferentFontProperties((fontProperties) =>
            {
                TextBlock block = TestHelper.CreateTextBlock(fontProperties);

                SnippetPerformanceCounterRelease textBlockCounter = new SnippetPerformanceCounterRelease("");
                SnippetPerformanceCounterRelease textMeasurerCounter = new SnippetPerformanceCounterRelease("");
                TestHelper.ExecuteActionWithDifferentWords((word) =>
                {
                    this.AssertAreEqualAndMeasurePerformance(block, word, fontProperties, textBlockCounter, textMeasurerCounter);
                });

                textBlockCounter.End();
                textMeasurerCounter.End();

                Assert.IsTrue(textMeasurerCounter.TotalMilliseconds <= textBlockCounter.TotalMilliseconds);
            });
        }

        [TestMethod]
        public void TestCorrectnessWithLotsOfWordsDifferentFontsDifferentLanguages()
        {
            TestHelper.ExecuteActionWithDifferentFontProperties((fontProperties) =>
            {
                TextBlock block = TestHelper.CreateTextBlock(fontProperties);

                TestHelper.ExecuteActionWithDifferentWordsDifferentLanguages((word) =>
                {
                    this.AssertAreEqual(block, word, fontProperties);
                });
            });
        }
#endif
        [TestMethod]
        public void TestBaselineOffsetCorrectness()
        {
            TestHelper.ExecuteActionWithDifferentFontFamilies((fontFamily) =>
            {
                FontProperties fontProperties = new FontProperties(fontFamily, FontStyles.Normal, FontWeights.Bold);
                TextBlock block = TestHelper.CreateTextBlock(fontProperties);
                TestHelper.ExecuteActionWithDifferentWords((word) =>
                {
                    block.FontSize = word.Size;
                    block.Text = word.Text;

                    double baseLineOffset = block.GetBaselineOffset();
                    double radBaseLineOffset = this.textMeasurer.MeasureText(word, fontProperties).BaselineOffset;

                    Assert.IsTrue(TestHelper.AreClose(baseLineOffset, radBaseLineOffset));
                });

            });
        }
#if WPF
        [TestMethod]
        public void TestSpecialWordCorrectness()
        {
            TextProperties text = new TextProperties("average,average,average,average,average,average,average,", 12);
            TestHelper.ExecuteActionWithDifferentFontProperties((fontProperties) =>
            {
                TextBlock block = TestHelper.CreateTextBlock(fontProperties);
                this.AssertAreEqual(block, text, fontProperties);
            });
        }

        [TestMethod]
        public void TestKerningCorrectness()
        {
            TextProperties text = new TextProperties("T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,T,", 12);
            TestHelper.ExecuteActionWithDifferentFontProperties((fontProperties) =>
            {
                TextBlock block = TestHelper.CreateTextBlock(fontProperties);
                this.AssertAreEqual(block, text, fontProperties);
            });
        }
#endif
        //Bug:148378
        ////[TestMethod]
        public void TestLigaturesCorrectness()
        {
            TextProperties text = new TextProperties("tttttttttttttttttttttttttttttttttttttttttt", 14);
            FontProperties fontProperties = new FontProperties(new FontFamily("Calibri"));
            TextBlock block = TestHelper.CreateTextBlock(fontProperties);
            
            this.AssertAreEqual(block, text, fontProperties);
        }

#if WPF
        [TestMethod]
        public void TestTextMeasurementMultiThreaded()
        {
            TextProperties textProperties = new TextProperties("test", 12);

            Dictionary<FontProperties, Size> expectedValues = new Dictionary<FontProperties, Size>();
            TestHelper.ExecuteActionWithDifferentFontProperties((fontProperties) =>
            {
                TextBlock block = TestHelper.CreateTextBlock(fontProperties);
                block.FontSize = textProperties.Size;
                block.Text = textProperties.Text;
                expectedValues[fontProperties] = block.GetSize();
            });

            Action a = new Action(() =>
            {
                TestHelper.ExecuteActionWithDifferentFontProperties((fontProperties) =>
                {
                    Size textBlockSize = expectedValues[fontProperties];
                    Size textMeasurerSize = this.textMeasurer.MeasureText(textProperties, fontProperties).Size;

                    Assert.IsTrue(
                        TestHelper.AreClose(textBlockSize.Width, textMeasurerSize.Width) &&
                        TestHelper.AreClose(textBlockSize.Height, textMeasurerSize.Height));
                });
            });

            TasksHelper.DoAsync(a, a, a, a, a, a, a, a);
        }



        [TestMethod]
        public void TestTextMeasurementSingleThread()
        {
            TextProperties textProperties = new TextProperties("test", 12);

            Dictionary<FontProperties, Size> expectedValues = new Dictionary<FontProperties, Size>();
            TestHelper.ExecuteActionWithDifferentFontProperties((fontProperties) =>
            {
                TextBlock block = TestHelper.CreateTextBlock(fontProperties);
                block.FontSize = textProperties.Size;
                block.Text = textProperties.Text;
                expectedValues[fontProperties] = block.GetSize();
            });

            TestHelper.ExecuteActionWithDifferentFontProperties((fontProperties) =>
            {
                Size textBlockSize = expectedValues[fontProperties];
                Size textMeasurerSize = this.textMeasurer.MeasureText(textProperties, fontProperties).Size;

                Assert.IsTrue(
                    TestHelper.AreClose(textBlockSize.Width, textMeasurerSize.Width) &&
                    TestHelper.AreClose(textBlockSize.Height, textMeasurerSize.Height));
            });
        }

        [TestMethod]
        public void TestMeasureTextWrapping()
        {
            double wrappingWidth = 61;
            TextProperties text = new TextProperties("сад сад сдас дсадsadasdsadsadsadsadsad сад сад сад сад сад сад сад сад сад сад садсад асд сад сад ", 12);
            TestHelper.ExecuteActionWithDifferentFontProperties((fontProperties) =>
            {
                TextBlock block = TestHelper.CreateTextBlock(fontProperties);

                this.AssertAreEqualWrapping(block, text, fontProperties, wrappingWidth);
            });
        }
#endif
        [TestMethod]
        public void TestMeasureWhiteSpaces()
        {
            TextProperties textProperties = new TextProperties("\r\n\r\n");
            FontProperties fontProperties = new FontProperties(new FontFamily("Arial"));

            TextBlock textBlock = TestHelper.CreateTextBlock(fontProperties);

            textBlock.Text = textProperties.Text;
            textBlock.FontSize = textProperties.Size;
            Size textBlockSize = textBlock.GetSize();
            Size textMeasurerSize = this.textMeasurer.MeasureText(textProperties, fontProperties).Size;
            Assert.IsTrue(
                TestHelper.AreClose(textBlockSize.Width, textMeasurerSize.Width));
        }

        private void AssertAreEqual(TextBlock textBlock, TextProperties textProperties, FontProperties fontProperties)
        {
            textBlock.Text = textProperties.Text;
            textBlock.FontSize = textProperties.Size;
            Size textBlockSize = textBlock.GetSize();
            Size textMeasurerSize = this.textMeasurer.MeasureText(textProperties, fontProperties).Size;
            Assert.IsTrue(
                TestHelper.AreClose(textBlockSize.Width, textMeasurerSize.Width) &&
                TestHelper.AreClose(textBlockSize.Height, textMeasurerSize.Height));
        }

        private void AssertAreEqualWrapping(TextBlock textBlock, TextProperties textProperties, FontProperties fontProperties, double wrappingWidth)
        {
            textBlock.Text = textProperties.Text;
            textBlock.FontSize = textProperties.Size;
            textBlock.MaxWidth = wrappingWidth;
            textBlock.Width = wrappingWidth;
            textBlock.TextWrapping = TextWrapping.Wrap;

            Size blockSize = textBlock.GetSize();
            Size radBlockSize = RadTextMeasurer.MeasureTextWithWrapping(textProperties, fontProperties, wrappingWidth).Size;
            Assert.IsTrue(TestHelper.AreClose(blockSize.Height, radBlockSize.Height));
        }

        private void AssertAreEqualAndMeasurePerformance(TextBlock textBlock, TextProperties textProperties, FontProperties fontProperties, SnippetPerformanceCounterRelease textBlockCounter, SnippetPerformanceCounterRelease textMeasurerCounter)
        {
            textBlockCounter.Start();
            textBlock.FontSize = textProperties.Size;
            textBlock.Text = textProperties.Text;
            Size blockSize = textBlock.GetSize();
            textBlockCounter.Stop();

            textMeasurerCounter.Start();
            Size radBlockSize = this.textMeasurer.MeasureText(textProperties, fontProperties).Size;
            textMeasurerCounter.Stop();

            Assert.IsTrue(
            TestHelper.AreClose(blockSize.Width, radBlockSize.Width) &&
            TestHelper.AreClose(blockSize.Height, radBlockSize.Height));
        }
    }
}
