﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Linq;
using System.Windows;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Tests.Utils
{
    [TestClass]
    public class MathUtilitiesTests
    {
        [TestMethod]
        public void TrySolveLinearEquationValidEquationTest()
        {
            double x;
            bool isSolvable = MathUtilities.TrySolveLinear(3, 2, out x);

            Assert.IsTrue(isSolvable);

            double expected = -2.0 / 3;
            Assert.AreEqual(expected, x);
        }

        [TestMethod]
        public void TrySolveLinearEquationNoRealSolutionTest()
        {
            double x;
            bool isSolvable = MathUtilities.TrySolveLinear(0, 2, out x);

            Assert.IsFalse(isSolvable);
        }

        [TestMethod]
        public void TrySolveLinearEquationAllRealSolutionsTest()
        {
            double x;
            bool isSolvable = MathUtilities.TrySolveLinear(0, 0, out x);

            Assert.IsFalse(isSolvable);
        }

        [TestMethod]
        public void TrySolveQuadraticEquationTwoSolutionsTest()
        {
            double[] x;
            bool isSolvable = MathUtilities.TrySolveQuadratic(1, -2, -3, out x);

            Assert.IsTrue(isSolvable);
            TestHelper.AssertAreEqual(new double[] { -1, 3}, x);
        }

        [TestMethod]
        public void TrySolveQuadraticEquationTwoSolutionsNegativeFirstCoeficientTest()
        {
            double[] x;
            bool isSolvable = MathUtilities.TrySolveQuadratic(-2, 5, 3, out x);

            Assert.IsTrue(isSolvable);
            TestHelper.AssertAreEqual(new double[] { -1.0 / 2, 3 }, x);
        }

        [TestMethod]
        public void TrySolveQuadraticEquationOneSolutionTest()
        {
            double[] x;
            bool isSolvable = MathUtilities.TrySolveQuadratic(1, 6, 9, out x);

            Assert.IsTrue(isSolvable);
            TestHelper.AssertAreEqual(new double[] { -3 }, x);
        }

        [TestMethod]
        public void TrySolveQuadraticEquationEveryRealIsSolutionTest()
        {
            double[] x;
            bool isSolvable = MathUtilities.TrySolveQuadratic(0, 0, 0, out x);

            Assert.IsFalse(isSolvable);
        }

        [TestMethod]
        public void TrySolveQuadraticEquationNoSolutionTest()
        {
            double[] x;
            bool isSolvable = MathUtilities.TrySolveQuadratic(0, 0, 1, out x);

            Assert.IsFalse(isSolvable);
        }

        [TestMethod]
        public void TrySolveQuadraticEquationNoRealSolutionTest()
        {
            double[] x;
            bool isSolvable = MathUtilities.TrySolveQuadratic(1, 0, 1, out x);

            Assert.IsFalse(isSolvable);
        }

        [TestMethod]
        public void TryGetCubicLocalExtremaFourExtremaTest()
        {
            double a = -4, b = 15, c = 18, d = -6;
            Interval interval = new Interval(-0.75, 3.25);
            Func<double, Point> function = (x) => { return new Point(x, a * x * x * x + b * x * x + c * x + d); };

            Point[] extrema = MathUtilities.FindLocalExtrema(interval, a, b, c, d).ToArray();
            Point[] expected = new Point[] { function(interval.Start), function(-1.0 / 2), function(3), function(interval.End) };

            TestHelper.AssertAreEqual(expected, extrema);
        }

        [TestMethod]
        public void TryGetCubicLocalExtremaThreeExtremaItervalStartCoincideExtremumTest()
        {
            double a = -4, b = 15, c = 18, d = -6;
            Interval interval = new Interval(-0.5, 3.25);
            Func<double, Point> function = (x) => { return new Point(x, a * x * x * x + b * x * x + c * x + d); };

            Point[] extrema = MathUtilities.FindLocalExtrema(interval, a, b, c, d).ToArray();
            Point[] expected = new Point[] { function(interval.Start), function(3), function(interval.End) };

            TestHelper.AssertAreEqual(expected, extrema);
        }

        [TestMethod]
        public void TryGetCubicLocalExtremaThreeExtremaItervalEndCoincideExtremumTest()
        {
            double a = -4, b = 15, c = 18, d = -6;
            Interval interval = new Interval(-1, 3);
            Func<double, Point> function = (x) => { return new Point(x, a * x * x * x + b * x * x + c * x + d); };

            Point[] extrema = MathUtilities.FindLocalExtrema(interval, a, b, c, d).ToArray();
            Point[] expected = new Point[] { function(interval.Start), function(-1.0 / 2), function(interval.End) };

            TestHelper.AssertAreEqual(expected, extrema);
        }

        [TestMethod]
        public void TryGetCubicLocalExtremaTwoExtremaBothIntervalEndsCoincideExtremumTest()
        {
            double a = -4, b = 15, c = 18, d = -6;
            Interval interval = new Interval(0.5, 3);
            Func<double, Point> function = (x) => { return new Point(x, a * x * x * x + b * x * x + c * x + d); };

            Point[] extrema = MathUtilities.FindLocalExtrema(interval, a, b, c, d).ToArray();
            Point[] expected = new Point[] { function(interval.Start), function(interval.End) };

            TestHelper.AssertAreEqual(expected, extrema);
        }

        [TestMethod]
        public void TryGetCubicLocalExtremaThreeExtremaTest()
        {
            double a = -4, b = 15, c = 18, d = -6;
            Interval interval = new Interval(0, 4);
            Func<double, Point> function = (x) => { return new Point(x, a * x * x * x + b * x * x + c * x + d); };

            Point[] extrema = MathUtilities.FindLocalExtrema(interval, a, b, c, d).ToArray();
            Point[] expected = new Point[] { function(interval.Start), function(3), function(interval.End) };

            TestHelper.AssertAreEqual(expected, extrema);
        }

        [TestMethod]
        public void TryGetCubicLocalExtremaIntervalOutOfExtremaTest()
        {
            double a = -4, b = 15, c = 18, d = -6;
            Interval interval = new Interval(4, 5);
            Func<double, Point> function = (x) => { return new Point(x, a * x * x * x + b * x * x + c * x + d); };

            Point[] extrema = MathUtilities.FindLocalExtrema(interval, a, b, c, d).ToArray();
            Point[] expected = new Point[] { function(interval.Start), function(interval.End) };

            TestHelper.AssertAreEqual(expected, extrema);
        }

        [TestMethod]
        public void TryGetCubicLocalExtremaIntervalBetweenExtremaTest()
        {
            double a = -4, b = 15, c = 18, d = -6;
            Interval interval = new Interval(4, 5);
            Func<double, Point> function = (x) => { return new Point(x, a * x * x * x + b * x * x + c * x + d); };

            Point[] extrema = MathUtilities.FindLocalExtrema(interval, a, b, c, d).ToArray();
            Point[] expected = new Point[] { function(interval.Start), function(interval.End) };

            TestHelper.AssertAreEqual(expected, extrema);
        }

        [TestMethod]
        public void TryGetCubicLocalExtremaThreeExtremaAllExremaInIntervalTest()
        {
            double a = 1, b = 9, c = 27, d = -3;
            Interval interval = new Interval(-4, 1);
            Func<double, Point> function = (x) => { return new Point(x, a * x * x * x + b * x * x + c * x + d); };

            Point[] extrema = MathUtilities.FindLocalExtrema(interval, a, b, c, d).ToArray();
            Point[] expected = new Point[] { function(interval.Start), function(-3), function(interval.End) };

            TestHelper.AssertAreEqual(expected, extrema);
        }

        [TestMethod]
        public void TryGetQuadraticLocalExtremaThreeExtremaTest()
        {
            double a = 1, b = 4, c = 2;
            Interval interval = new Interval(-4, 1);
            Func<double, Point> function = (x) => { return new Point(x, a * x * x + b * x + c); };

            Point[] extrema = MathUtilities.FindLocalExtrema(interval, a, b, c).ToArray();
            Point[] expected = new Point[] { function(interval.Start), function(-2), function(interval.End) };

            TestHelper.AssertAreEqual(expected, extrema);
        }

        [TestMethod]
        public void TryGetQuadraticLocalExtremaTwoExtremaIntevalStartCoincideExtremumTest()
        {
            double a = 1, b = 4, c = 2;
            Interval interval = new Interval(-2, 1);
            Func<double, Point> function = (x) => { return new Point(x, a * x * x + b * x + c); };

            Point[] extrema = MathUtilities.FindLocalExtrema(interval, a, b, c).ToArray();
            Point[] expected = new Point[] { function(interval.Start), function(interval.End) };

            TestHelper.AssertAreEqual(expected, extrema);
        }

        [TestMethod]
        public void TryGetQuadraticLocalExtremaTwoExtremaIntevalEndCoincideExtremumTest()
        {
            double a = 1, b = 4, c = 2;
            Interval interval = new Interval(-3, -2);
            Func<double, Point> function = (x) => { return new Point(x, a * x * x + b * x + c); };

            Point[] extrema = MathUtilities.FindLocalExtrema(interval, a, b, c).ToArray();
            Point[] expected = new Point[] { function(interval.Start), function(interval.End) };

            TestHelper.AssertAreEqual(expected, extrema);
        }

        [TestMethod]
        public void TryGetQuadraticLocalExtremaTwoExtremaIntevalNotContainingExtremumTest()
        {
            double a = 1, b = 4, c = 2;
            Interval interval = new Interval(0, 1);
            Func<double, Point> function = (x) => { return new Point(x, a * x * x + b * x + c); };

            Point[] extrema = MathUtilities.FindLocalExtrema(interval, a, b, c).ToArray();
            Point[] expected = new Point[] { function(interval.Start), function(interval.End) };

            TestHelper.AssertAreEqual(expected, extrema);
        }

        [TestMethod]
        public void TryGetCubicMinimumMaximum()
        {
            Interval interval = new Interval(-2, 5);
            Interval minimumMaximum = MathUtilities.FindLocalMinimumAndMaximum(interval, -4, 15, 18, -6);
            Interval expectedMinimumMaximum = new Interval(-41, 75);

            Assert.AreEqual(expectedMinimumMaximum, minimumMaximum);
        }

        [TestMethod]
        public void TryGetQuadraticMinimumMaximum()
        {
            Interval interval = new Interval(-3, 5);
            Interval minimumMaximum = MathUtilities.FindLocalMinimumAndMaximum(interval, 1, 4, 1);
            Interval expectedMinimumMaximum = new Interval(-3, 46);

            Assert.AreEqual(expectedMinimumMaximum, minimumMaximum);
        }
    }
}
