using System;
using System.IO;

using Org.BouncyCastle.Utilities;

/*
The Bouncy Castle Cryptographic C# API

License:

The Bouncy Castle License
Copyright (c) 2000 - 2017 The Legion of the Bouncy Castle Inc. (http://www.bouncycastle.org)
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
DEALINGS IN THE SOFTWARE.
 */

namespace Org.BouncyCastle.Asn1
{
    /**
	 * A Der encoded set object
	 */
    internal class DerSet
		: Asn1Set
	{
		public static readonly DerSet Empty = new DerSet();

		public static DerSet FromVector(
			Asn1EncodableVector v)
		{
			return v.Count < 1 ? Empty : new DerSet(v);
		}

		internal static DerSet FromVector(
			Asn1EncodableVector	v,
			bool				needsSorting)
		{
			return v.Count < 1 ? Empty : new DerSet(v, needsSorting);
		}

		/**
		 * create an empty set
		 */
		public DerSet()
			: base(0)
		{
		}

		/**
		 * @param obj - a single object that makes up the set.
		 */
		public DerSet(
			Asn1Encodable obj)
			: base(1)
		{
			AddObject(obj);
		}

		public DerSet(
			params Asn1Encodable[] v)
			: base(v.Length)
		{
			foreach (Asn1Encodable o in v)
			{
				AddObject(o);
			}

			Sort();
		}

		/**
		 * @param v - a vector of objects making up the set.
		 */
		public DerSet(
			Asn1EncodableVector v)
			: this(v, true)
		{
		}

		internal DerSet(
			Asn1EncodableVector	v,
			bool				needsSorting)
			: base(v.Count)
		{
			foreach (Asn1Encodable o in v)
			{
				AddObject(o);
			}

			if (needsSorting)
			{
				Sort();
			}
		}

		/*
		 * A note on the implementation:
		 * <p>
		 * As Der requires the constructed, definite-length model to
		 * be used for structured types, this varies slightly from the
		 * ASN.1 descriptions given. Rather than just outputing Set,
		 * we also have to specify Constructed, and the objects length.
		 */
		internal override void Encode(
			DerOutputStream derOut)
		{
			// TODO Intermediate buffer could be avoided if we could calculate expected length
			MemoryStream bOut = new MemoryStream();
			DerOutputStream dOut = new DerOutputStream(bOut);

			foreach (Asn1Encodable obj in this)
			{
				dOut.WriteObject(obj);
			}

            Platform.Dispose(dOut);

            byte[] bytes = bOut.ToArray();

			derOut.WriteEncoded(Asn1Tags.Set | Asn1Tags.Constructed, bytes);
		}
	}
}
