﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.Serialization;
using System.Text;
using Telerik.Windows.Documents.Exceptions;

namespace Telerik.Windows.Documents.Fixed.Exceptions
{
    /// <summary>
    /// Represents not supported stream type exception.
    /// </summary>
#if WPF
    [Serializable]
#endif
    public class NotSupportedStreamTypeException : NotSupportedFeatureException
    {
        private const string DefaultMessage = "The stream type is not supported.";
        private const string CanReadMessage = "The input stream should support Read operation.";
        private const string CanSeekMessage = "The input stream should support Seek operation.";

        /// <summary>
        /// Initializes a new instance of the <see cref="NotSupportedStreamTypeException" /> class.
        /// </summary>
        public NotSupportedStreamTypeException() : base(DefaultMessage)
        {
        }
         
        /// <summary>
        /// Initializes a new instance of the <see cref="NotSupportedStreamTypeException" /> class.
        /// </summary>
        /// <param name="message">The message.</param>
        public NotSupportedStreamTypeException(string message) : base(message)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="NotSupportedStreamTypeException" /> class.
        /// </summary>
        /// <param name="supportsRead">The supports read.</param>
        /// <param name="supportsSeek">The supports seek.</param>
        public NotSupportedStreamTypeException(bool supportsRead, bool supportsSeek)
            : this(supportsRead, supportsSeek, BuildMessage(supportsRead, supportsSeek))
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="NotSupportedStreamTypeException" /> class.
        /// </summary>
        /// <param name="supportsRead">The supports read.</param>
        /// <param name="supportsSeek">The supports seek.</param>
        /// <param name="message">The message.</param>
        public NotSupportedStreamTypeException(bool supportsRead, bool supportsSeek, string message) : base(message)
        { 
            this.SupportsRead = supportsRead;
            this.SupportsSeek = supportsSeek;
        }
         
        /// <summary>
        /// Initializes a new instance of the <see cref="NotSupportedStreamTypeException" /> class.
        /// </summary>
        /// <param name="message">The message.</param>
        /// <param name="cause">The cause.</param>
        public NotSupportedStreamTypeException(string message, Exception cause) : base(message, cause)
        {
        }

#if WPF

        /// <summary>
        /// Initializes a new instance of the <see cref="NotSupportedStreamTypeException" /> class.
        /// </summary>
        /// <param name="info">The info.</param>
        /// <param name="context">The context.</param>
        protected NotSupportedStreamTypeException(SerializationInfo info, StreamingContext context) : base(info, context)
        { 
            this.SupportsSeek = info.GetBoolean("SupportsSeek");
            this.SupportsRead = info.GetBoolean("SupportsRead");
        }
#endif

        /// <summary>
        /// Gets or sets the supports seek.
        /// </summary>
        /// <value>The supports seek.</value>
        public bool SupportsSeek { private set; get; }

        /// <summary>
        /// Gets or sets the supports read.
        /// </summary>
        /// <value>The supports read.</value>
        public bool SupportsRead { private set; get; }

#if WPF 
        /// <summary>
        /// Provides COM objects with version-independent access to the <see cref="M:System.Exception.GetObjectData(System.Runtime.Serialization.SerializationInfo,System.Runtime.Serialization.StreamingContext)" />
        /// method
        /// </summary>
        /// <param name="info">The <see cref="T:System.Runtime.Serialization.SerializationInfo" />
        /// object that holds the serialized object data about the exception being thrown.</param>
        /// <param name="context">The <see cref="T:System.Runtime.Serialization.StreamingContext" />
        /// structure that contains contextual information about the source or destination.</param>
        [System.Security.SecurityCriticalAttribute()]
        public override void GetObjectData(SerializationInfo info, StreamingContext context)
        {
            info.AddValue("SupportsSeek", this.SupportsSeek);
            info.AddValue("SupportsRead", this.SupportsRead);

            base.GetObjectData(info, context);
        }
#endif

        private static string BuildMessage(bool supportsRead, bool supportsSeek)
        {
            StringBuilder builder = new StringBuilder(DefaultMessage);

            if (!supportsRead)
            {
                builder.AppendFormat(" {0}", CanReadMessage);
            }

            if (!supportsSeek)
            {
                builder.AppendFormat(" {0}", CanSeekMessage);
            }

            return builder.ToString();
        }
    }
}
