﻿using System;
using System.Collections.Generic;
using System.Linq;
#if !NETSTANDARD
using System.Windows;
#else
using Telerik.Documents.Primitives;
#endif
using Telerik.Windows.Documents.Core.Fonts.Glyphs;
using Telerik.Windows.Documents.Fixed.Model.Fonts;

namespace Telerik.Windows.Documents.Core.Fonts
{
    internal abstract class FontSource
    {
        private static readonly object lockObject = new object();

        private string fontFamily;
        private double? underlineThickness;
        private double? underlinePosition;
        private bool? isBold;
        private bool? isItalic;
        private double? ascent;
        private double? descent;
        private double? lineGap;
        private double? capHeight;
        private double? stemV;
        private double? italicAngle;
        private Rect? boundingBox;
        private byte[] data;
        private IEnumerable<FontFlag> flags;

        public string FontFamily
        {
            get
            {
                lock (lockObject)
                {
                    if (string.IsNullOrEmpty(this.fontFamily))
                    {
                        lock (lockObject)
                        {
                            this.fontFamily = this.GetFontFamily();
                        }
                    }

                    return this.fontFamily;
                }
            }
        }

        public double UnderlineThickness
        {
            get
            {
                lock (lockObject)
                {
                    if (!this.underlineThickness.HasValue)
                    {
                        lock (lockObject)
                        {
                            this.underlineThickness = this.GetUnderlineThickness();
                        }
                    }

                    return this.underlineThickness.Value;
                }
            }
        }

        public double UnderlinePosition
        {
            get
            {
                lock (lockObject)
                {
                    if (!this.underlinePosition.HasValue)
                    {
                        lock (lockObject)
                        {
                            this.underlinePosition = this.GetUnderlinePosition();

                        }
                    }

                    return this.underlinePosition.Value;
                }
            }
        }

        public bool IsBold
        {
            get
            {
                lock (lockObject)
                {
                    if (!this.isBold.HasValue)
                    {
                        lock (lockObject)
                        {
                            this.isBold = this.GetIsBold();
                        }
                    }

                    return this.isBold.Value;
                }
            }
        }

        public bool IsItalic
        {
            get
            {
                lock (lockObject)
                {
                    if (!this.isItalic.HasValue)
                    {
                        lock (lockObject)
                        {
                            this.isItalic = this.GetIsItalic();
                        }
                    }

                    return this.isItalic.Value;
                }
            }
        }

        public double Ascent
        {
            get
            {
                lock (lockObject)
                {
                    if (!this.ascent.HasValue)
                    {
                        lock (lockObject)
                        {
                            this.ascent = this.GetAscent();
                        }
                    }

                    return this.ascent.Value;
                }
            }
        }

        public double Descent
        {
            get
            {
                lock (lockObject)
                {
                    if (!this.descent.HasValue)
                    {
                        lock (lockObject)
                        {
                            this.descent = this.GetDescent();
                        }
                    }

                    return this.descent.Value;
                }
            }
        }

        public double LineGap
        {
            get
            {
                lock (lockObject)
                {
                    if (!this.lineGap.HasValue)
                    {
                        lock (lockObject)
                        {
                            this.lineGap = this.GetLineGap();
                        }
                    }

                    return this.lineGap.Value;
                }
            }
        }

        public double CapHeight
        {
            get
            {
                lock (lockObject)
                {
                    if (!this.capHeight.HasValue)
                    {
                        lock (lockObject)
                        {
                            this.capHeight = this.GetCapHeight();
                        }
                    }

                    return this.capHeight.Value;
                }
            }
        }

        public double StemV
        {
            get
            {
                lock (lockObject)
                {
                    if (!this.stemV.HasValue)
                    {
                        lock (lockObject)
                        {
                            this.stemV = this.GetStemV();
                        }
                    }

                    return this.stemV.Value;
                }
            }
        }

        public double ItalicAngle
        {
            get
            {
                lock (lockObject)
                {
                    if (!this.italicAngle.HasValue)
                    {
                        lock (lockObject)
                        {
                            this.italicAngle = this.GetItalicAngle();
                        }
                    }

                    return this.italicAngle.Value;
                }
            }
        }

        public Rect BoundingBox
        {
            get
            {
                lock (lockObject)
                {
                    if (!this.boundingBox.HasValue)
                    {
                        lock (lockObject)
                        {
                            this.boundingBox = this.GetBoundingBox();
                        }
                    }

                    return this.boundingBox.Value;
                }
            }
        }

        public byte[] Data
        {
            get
            {
                lock (lockObject)
                {
                    if (this.data == null)
                    {
                        lock (lockObject)
                        {
                            this.data = this.GetData();
                        }
                    }

                    return this.data;
                }
            }
        }

        public IEnumerable<FontFlag> Flags
        {
            get
            {
                lock (lockObject)
                {
                    if (this.flags == null)
                    {
                        lock (lockObject)
                        {
                            this.flags = this.GetFlags();
                        }
                    }

                    return this.flags;
                }
            }
        }

        public virtual IEnumerable<FontFlag> GetFlags()
        {
            if (this.IsItalic)
            {
                yield return FontFlag.Italic;
            }
        }

        public abstract bool GetIsItalic();

        public abstract bool GetIsBold();

        public abstract string GetFontFamily();

        public abstract double GetUnderlineThickness();

        public abstract double GetUnderlinePosition();

        public abstract double GetAscent();

        public abstract double GetDescent();

        public abstract double GetLineGap();

        public abstract double GetCapHeight();

        public abstract double GetStemV();

        public abstract double GetItalicAngle();

        public abstract Rect GetBoundingBox();

        public abstract byte[] GetData();

        public abstract bool TryGetCharCode(int unicode, out int charCode);

        public abstract bool TryGetGlyphId(int unicode, out ushort glyphId);

        public double GetAdvancedWidth(int charCode)
        {
            lock (lockObject)
            {
                double value = this.GetAdvancedWidthOverride(charCode);

                return value;
            }
        }

        public abstract double GetAdvancedWidthOverride(int charCode);

        public void GetAdvancedWidth(Glyph glyph)
        {
            lock (lockObject)
            {
                this.GetAdvancedWidthOverride(glyph);
            }
        }

        public abstract void GetAdvancedWidthOverride(Glyph glyph);

        public void InitializeGlyphOutlines(Glyph glyph, double fontSize)
        {
            lock (lockObject)
            {
                this.InitializeGlyphOutlinesOverride(glyph, fontSize);
            }
        }

        public abstract void InitializeGlyphOutlinesOverride(Glyph glyph, double fontSize);
    }
}
