﻿using System;
using System.Collections.Generic;
using System.IO;
using Telerik.Windows.Documents.Core.Fonts.OpenType;
using Telerik.Windows.Documents.Core.Fonts.Type1.Type1Format;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Fonts;
using Telerik.Windows.Documents.Fixed.Utilities;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Fonts
{
    internal abstract class FontsManagerBase
    {
        public const string Courier = "Courier";
        public const string CourierBold = "Courier-Bold";
        public const string CourierBoldOblique = "Courier-BoldOblique";
        public const string CourierOblique = "Courier-Oblique";
        public const string Helvetica = "Helvetica";
        public const string HelveticaBold = "Helvetica-Bold";
        public const string HelveticaBoldOblique = "Helvetica-BoldOblique";
        public const string HelveticaOblique = "Helvetica-Oblique";
        public const string TimesRoman = "Times-Roman";
        public const string TimesBold = "Times-Bold";
        public const string TimesBoldItalic = "Times-BoldItalic";
        public const string TimesItalic = "Times-Italic";
        public const string Symbol = "Symbol";
        public const string ZapfDingbats = "ZapfDingbats";

        private static readonly Dictionary<string, string> alternativeNames;
        private static readonly Dictionary<string, StandardFontDescriptor> standardFontDescriptors;
        private static Dictionary<string, byte[]> standardFontStreams;
        private readonly Dictionary<string, Type1FontSource> standardFontSources;

        static FontsManagerBase()
        {
            standardFontDescriptors = new Dictionary<string, StandardFontDescriptor>();
            alternativeNames = new Dictionary<string, string>();

            InitializeStandardFontStreams();
            InitializeStandardFontDescriptors();
            InitializeAlternativeNames();
        }

        public FontsManagerBase()
        {
            this.standardFontSources = new Dictionary<string, Type1FontSource>();

            this.InitializeStandardFontSources();
        }

        public abstract OpenTypeFontSource GetOpenTypeFontSource(FontProperties fontProperties);

        public static bool IsStandardFontName(string name)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");

            return standardFontDescriptors.ContainsKey(name) || alternativeNames.ContainsKey(name);
        }

        public static StandardFontDescriptor GetStandardFontDescriptor(string fontName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(fontName, "fontName");

            StandardFontDescriptor fd;
            string name = GetStandardFontName(fontName);
            if (!standardFontDescriptors.TryGetValue(name, out fd))
            {
                throw new ArgumentException("Font name is not a standard font.", fontName);
            }

            return fd;
        }

        public Type1FontSource GetStandardFontSource(string fontName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(fontName, "fontName");

            Type1FontSource fontSource;
            string name = GetStandardFontName(fontName);

            if (!this.standardFontSources.TryGetValue(name, out fontSource))
            {
                throw new ArgumentException("Font name is not a standard font.", fontName);
            }

            return fontSource;
        }

        public Type1FontSource GetType1FallbackFontSource(string fontName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(fontName, "fontName");

            string styles = fontName.ToLower();
            bool isBold = FontsHelper.IsBold(styles);
            bool isItalic = FontsHelper.IsItalic(styles);

            string fallback = Helvetica;

            if (isBold)
            {
                if (isItalic)
                {
                    fallback = HelveticaBoldOblique;
                }
                else
                {
                    fallback = HelveticaBold;
                }
            }
            else if (isItalic)
            {
                fallback = HelveticaOblique;
            }

            return this.GetStandardFontSource(fallback);
        }

        private static Stream GetApplicationResourceStream(string fontName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(fontName, "fontName");
#if NETSTANDARD
            string resourceName = string.Format(@"Telerik.Documents.Fixed.Resources.Fonts.{0}.pfb", fontName);
#else
            string resourceName = string.Format(@"Telerik.Windows.Documents.Fixed.Resources.Fonts.{0}.pfb", fontName);
#endif

            return ResourcesHelper.GetApplicationResourceStream(resourceName);
        }

        private static string GetStandardFontName(string fontName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(fontName, "fontName");

            string name = fontName;
            if (alternativeNames.ContainsKey(fontName))
            {
                name = alternativeNames[fontName];
            }

            return name;
        }

        private static void RegisterStandardFontStream(string fontName, Stream stream)
        {
            Guard.ThrowExceptionIfNullOrEmpty(fontName, "fontName");
            Guard.ThrowExceptionIfNull(stream, "stream");

            standardFontStreams[fontName] = Extensions.ReadAllBytes(stream);
        }

        private static void RegisterStandardFontDescriptor(string name, StandardFontDescriptor fontDescriptor)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");
            Guard.ThrowExceptionIfNull(fontDescriptor, "fontDescriptor");

            standardFontDescriptors[name] = fontDescriptor;
        }

        private static void InitializeStandardFontDescriptors()
        {
            StandardFontDescriptor courier = new StandardFontDescriptor(629, -157);
            StandardFontDescriptor helvetica = new StandardFontDescriptor(718, -207);
            StandardFontDescriptor times = new StandardFontDescriptor(683, -217);
            StandardFontDescriptor symbolic = new StandardFontDescriptor(1000, 0);

            RegisterStandardFontDescriptor(Courier, courier);
            RegisterStandardFontDescriptor(CourierBold, courier);
            RegisterStandardFontDescriptor(CourierBoldOblique, courier);
            RegisterStandardFontDescriptor(CourierOblique, courier);

            RegisterStandardFontDescriptor(Helvetica, helvetica);
            RegisterStandardFontDescriptor(HelveticaBold, helvetica);
            RegisterStandardFontDescriptor(HelveticaBoldOblique, helvetica);
            RegisterStandardFontDescriptor(HelveticaOblique, helvetica);

            RegisterStandardFontDescriptor(TimesRoman, times);
            RegisterStandardFontDescriptor(TimesBold, times);
            RegisterStandardFontDescriptor(TimesBoldItalic, times);
            RegisterStandardFontDescriptor(TimesItalic, times);

            RegisterStandardFontDescriptor(Symbol, symbolic);
            RegisterStandardFontDescriptor(ZapfDingbats, symbolic);
        }

        private static void InitializeAlternativeNames()
        {
            RegisterAlternativeName(Helvetica, "Arial", "Arial MT");
            RegisterAlternativeName(HelveticaBold, "Arial,Bold", "Arial MT,Bold");
            RegisterAlternativeName(HelveticaBoldOblique, "Arial,BoldItalic", "Arial MT,BoldItalic");
            RegisterAlternativeName(HelveticaOblique, "Arial,Italic", "Arial MT,Italic");

            RegisterAlternativeName(Courier, "Courier New");
            RegisterAlternativeName(CourierBold, "Courier New,Bold");
            RegisterAlternativeName(CourierBoldOblique, "Courier New,BoldItalic");
            RegisterAlternativeName(CourierOblique, "Courier New,Italic");

            RegisterAlternativeName(TimesRoman, "Times New Roman");
            RegisterAlternativeName(TimesBold, "Times New Roman,Bold");
            RegisterAlternativeName(TimesBoldItalic, "Times New Roman,BoldItalic");
            RegisterAlternativeName(TimesItalic, "Times New Roman,Italic");
        }

        private static void RegisterAlternativeName(string original, params string[] alternatives)
        {
            foreach (string str in alternatives)
            {
                alternativeNames[str] = original;
            }
        }

        private static void InitializeStandardFontStreams()
        {
            if (standardFontStreams == null)
            {
                standardFontStreams = new Dictionary<string, byte[]>();

                RegisterStandardFontStream(Courier, GetApplicationResourceStream(Courier));
                RegisterStandardFontStream(CourierBold, GetApplicationResourceStream(CourierBold));
                RegisterStandardFontStream(CourierBoldOblique, GetApplicationResourceStream(CourierBoldOblique));
                RegisterStandardFontStream(CourierOblique, GetApplicationResourceStream(CourierOblique));
                RegisterStandardFontStream(Helvetica, GetApplicationResourceStream(Helvetica));
                RegisterStandardFontStream(HelveticaBold, GetApplicationResourceStream(HelveticaBold));
                RegisterStandardFontStream(HelveticaBoldOblique, GetApplicationResourceStream(HelveticaBoldOblique));
                RegisterStandardFontStream(HelveticaOblique, GetApplicationResourceStream(HelveticaOblique));
                RegisterStandardFontStream(TimesRoman, GetApplicationResourceStream(TimesRoman));
                RegisterStandardFontStream(TimesBold, GetApplicationResourceStream(TimesBold));
                RegisterStandardFontStream(TimesBoldItalic, GetApplicationResourceStream(TimesBoldItalic));
                RegisterStandardFontStream(TimesItalic, GetApplicationResourceStream(TimesItalic));
                RegisterStandardFontStream(Symbol, GetApplicationResourceStream(Symbol));
                RegisterStandardFontStream(ZapfDingbats, GetApplicationResourceStream(ZapfDingbats));
            }
        }

        private static Type1FontSource CreateType1FontSource(byte[] data)
        {
            try
            {
                Type1FontSource fs = new Type1FontSource(data);

                return fs;
            }
            catch
            {
                throw new InvalidOperationException("Cannot create standard font.");
            }
        }

        private void InitializeStandardFontSources()
        {
            foreach (var keyValue in standardFontStreams)
            {
                this.RegisterStandardFontSource(keyValue.Key, CreateType1FontSource(keyValue.Value));
            }
        }

        private void RegisterStandardFontSource(string fontName, Type1FontSource fontSource)
        {
            Guard.ThrowExceptionIfNullOrEmpty(fontName, "fontName");
            Guard.ThrowExceptionIfNull(fontSource, "fontSource");

            this.standardFontSources[fontName] = fontSource;
        }
    }
}
