﻿using System.Collections.Generic;
using System.IO;
using System.Windows;
using System.Windows.Media;
using Telerik.Windows.Documents.Core.Fonts.OpenType;
using Telerik.Windows.Documents.Core.Fonts.OpenType.Utils;
using Telerik.Windows.Documents.Fixed.Exceptions;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Fonts
{
    internal class FontsManager : FontsManagerBase
    {
        private readonly Dictionary<FontProperties, OpenTypeFontSource> fonts;
        private readonly Dictionary<FallbackRecord, OpenTypeFontSource> fallbacks;

        public FontsManager() 
        {
            this.fonts = new Dictionary<FontProperties, OpenTypeFontSource>();
            this.fallbacks = new Dictionary<FallbackRecord, OpenTypeFontSource>();
        }

        public override OpenTypeFontSource GetOpenTypeFontSource(FontProperties fontProperties)
        {
            OpenTypeFontSource res;
            if (!this.fonts.TryGetValue(fontProperties, out res))
            {
                GlyphTypeface glyphTypeface = GetGlyphTypeface(fontProperties.FontFamilyName, fontProperties.FontStyle, fontProperties.FontWeight);
                res = this.GetFontSourceFromGlyphTypeface(fontProperties.FontFamilyName, glyphTypeface);
                this.fonts[fontProperties] = res;
            }

            return res;
        }

        public OpenTypeFontSource GetFallbackFontSource(FontProperties descr, string text)
        {
            FallbackRange range = FallbackRange.GetFallbackRange(text[0]);
            FallbackRecord fr = new FallbackRecord(range, descr);
            OpenTypeFontSource res;
            if (!this.fallbacks.TryGetValue(fr, out res))
            {
                foreach (string fallbackFontFamily in range.FallbackFontFamilies)
                {
                    GlyphTypeface glyphTypeface = GetGlyphTypeface(fallbackFontFamily, descr.FontStyle, descr.FontWeight);
                    res = this.GetFontSourceFromGlyphTypeface(fallbackFontFamily, glyphTypeface);
                    if (res != null)
                    {
                        break;
                    }
                }
                if (res == null)
                {
                    throw new NotSupportedFontFamilyException();
                }
                this.fallbacks[fr] = res;
            }
            return res;
        }

        private static GlyphTypeface GetGlyphTypeface(string familyName, FontStyle style, FontWeight weight)
        {
            Typeface tf = new Typeface(new FontFamily(familyName), style, weight, FontStretches.Normal);
            GlyphTypeface gtf;
            tf.TryGetGlyphTypeface(out gtf);
            return gtf;
        }

        private static IEnumerable<OpenTypeFontSourceBase> GetFontSources(Stream fontUri)
        {
            OpenTypeFontReader reader = new OpenTypeFontReader(Extensions.ReadAllBytes(fontUri));
            switch (OpenTypeFontSource.GetFontType(reader))
            {
                case OpenTypeFontSourceType.TrueType:

                    yield return ReadTrueTypeFont(reader);
                    break;

                case OpenTypeFontSourceType.TrueTypeCollection:
                    TrueTypeCollection ttc = ReadTrueTypeCollection(reader);
                    foreach (OpenTypeFontSourceBase fs in ttc.Fonts)
                    {
                        yield return fs;
                    }

                    break;

                default:
                    yield break;
            }
        }

        private static FontProperties CreateFontDescriptorFromFontSource(FontSource source)
        {
            FontStyle fs = source.IsItalic ? FontStyles.Italic : FontStyles.Normal;
            FontWeight fw = source.IsBold ? FontWeights.Bold : FontWeights.Normal;
            FontProperties fd = new FontProperties(new FontFamily(source.FontFamily), fs, fw);

            return fd;
        }

        private static OpenTypeFontSourceBase ReadTrueTypeFont(OpenTypeFontReader reader)
        {
            OpenTypeFontSource tt = new OpenTypeFontSource(reader);
            return tt;
        }

        private static TrueTypeCollection ReadTrueTypeCollection(OpenTypeFontReader reader)
        {
            TrueTypeCollection ttc = new TrueTypeCollection(reader);
            ttc.Initialize();
            return ttc;
        }

        private void RegisterFontSource(FontProperties descr, OpenTypeFontSource source)
        {
            this.fonts[descr] = source;
        }

        private OpenTypeFontSource GetFontSourceFromGlyphTypeface(string fontFamily, GlyphTypeface gtf)
        {
            OpenTypeFontSource res = null;
            if (gtf == null)
            {
                return null;
            }

            bool isBold = gtf.Weight == FontWeights.Bold && gtf.StyleSimulations != StyleSimulations.BoldItalicSimulation && gtf.StyleSimulations != StyleSimulations.BoldSimulation;
            bool isItalic = gtf.Style != FontStyles.Normal && gtf.StyleSimulations != StyleSimulations.BoldItalicSimulation && gtf.StyleSimulations != StyleSimulations.ItalicSimulation;
            foreach (OpenTypeFontSource fs in GetFontSources(gtf.GetFontStream()))
            {
                if (fs.FontFamily == fontFamily && isBold == fs.IsBold && isItalic == fs.IsItalic)
                {
                    res = fs;
                    continue;
                }

                this.RegisterFontSource(CreateFontDescriptorFromFontSource(fs), fs);
            }

            return res;
        }
    }
}
