﻿using System;
using System.Collections.Generic;
using System.Linq;
#if !NETSTANDARD
using System.Windows;
#else
using Telerik.Documents.Primitives;
#endif
using Telerik.Windows.Documents.Core.Fonts.OpenType.Tables;
using Telerik.Windows.Documents.Core.Fonts.OpenType.Tables.TrueTypeOutlines;
using Telerik.Windows.Documents.Core.Fonts.OpenType.Tables.TrueTypeTables;
using Telerik.Windows.Documents.Core.Fonts.OpenType.Tables.Utils;
using Telerik.Windows.Documents.Core.Fonts.Type1;
using Telerik.Windows.Documents.Core.Fonts.Type1.CFFFormat;
using Telerik.Windows.Documents.Core.Fonts.Utils;
using Telerik.Windows.Documents.Fixed.Model.Fonts;

namespace Telerik.Windows.Documents.Core.Fonts.OpenType
{
    internal partial class OpenTypeFontSource : OpenTypeFontSourceBase
    {
        private static object lockObj = new object();
        private readonly FontProperties fontProperties;

        private Dictionary<ushort, GlyphData> glyphsData;

        private Dictionary<uint, TableRecord> tables;
        private IndexToLocation loca;
        private MaxProfile maxp;
        private Name name;
        private Post post;

        public OpenTypeFontSource(FontProperties fontProperties, OpenTypeFontReader reader) : base(reader)
        {
            this.Offset = 0;
            this.fontProperties = fontProperties;
            this.Initialize();
        }

        public OpenTypeFontSource(OpenTypeFontReader reader) : this(null, reader)
        {
        }

        internal override Outlines GetOutlines()
        {
            return this.OffsetTable.HasOpenTypeOutlines ? Outlines.OpenType : Outlines.TrueType;
        }

        internal override CMap GetCMap()
        {
            var value = new CMap(this);
            this.ReadTableData(value);

            return value;
        }

        internal override HorizontalMetrics GetHMtx()
        {
            var value = new HorizontalMetrics(this);
            this.ReadTableData(value);

            return value;
        }

        internal override Kerning GetKern()
        {
            var value = new Kerning(this);
            this.ReadTableData(value);

            return value;
        }

        internal override GlyphSubstitution GetGSub()
        {
            var value = new GlyphSubstitution(this);
            this.ReadTableData(value);

            return value;
        }

        internal override Head GetHead()
        {
            var value = new Head(this);
            this.ReadTableData(value);

            return value;
        }

        internal override HorizontalHeader GetHHea()
        {
            var value = new HorizontalHeader(this);
            this.ReadTableData(value);

            return value;
        }

        internal Post Post
        {
            get
            {
                lock (lockObj)
                {
                    if (this.post == null)
                    {
                        lock (lockObj)
                        {
                            this.post = this.ReadTableData(Tags.POST_TABLE, Post.ReadPostTable);
                        }
                    }

                    return this.post;
                }
            }
        }

        internal Name Name
        {
            get
            {
                lock (lockObj)
                {
                    if (this.name == null)
                    {
                        lock (lockObj)
                        {
                            this.name = this.ReadTableData(Tags.NAME_TABLE, Name.ReadNameTable);
                        }
                    }
                    return this.name;
                }
            }
        }

        internal IndexToLocation Loca
        {
            get
            {
                lock (lockObj)
                {
                    if (this.loca == null)
                    {
                        lock (lockObj)
                        {
                            this.loca = new IndexToLocation(this);
                            this.ReadTableData(this.loca);
                        }
                    }
                    return this.loca;
                }
            }
        }

        internal MaxProfile MaxP
        {
            get
            {
                lock (lockObj)
                {
                    if (this.maxp == null)
                    {
                        lock (lockObj)
                        {
                            this.maxp = new MaxProfile(this);
                            this.ReadTableData(this.maxp);
                        }
                    }
                    return this.maxp;
                }
            }
        }

        internal Dictionary<uint, TableRecord> Tables
        {
            get
            {
                return this.tables;
            }
        }

        internal OffsetTable OffsetTable
        {
            get;
            private set;
        }

        internal override ushort GetGlyphCount()
        {
            return this.MaxP.NumGlyphs;
        }

        internal long Offset { get; set; }

        public override string GetFontFamily()
        {
            return this.Name.FontFamily ?? (this.fontProperties == null ? null : this.fontProperties.FontFamilyName);
        }

        public override IEnumerable<FontFlag> GetFlags()
        {
            foreach (FontFlag flag in base.GetFlags())
            {
                yield return flag;
            }

            yield return FontFlag.Symbolic;
        }

        public override bool GetIsBold()
        {
            return this.Head.IsBold;
        }

        public override bool GetIsItalic()
        {
            return this.Head.IsItalic;
        }

        public override double GetCapHeight()
        {
            return 0;
        }

        public override double GetStemV()
        {
            return 0;
        }

        public override double GetItalicAngle()
        {
            return this.Post.ItalicAngle;
        }

        public override double GetUnderlineThickness()
        {
            return this.Scaler.FUnitsToPixels(this.Post.UnderlineThickness, 1000);
        }

        public override double GetUnderlinePosition()
        {
            return this.Scaler.FUnitsToPixels(this.Post.UnderlinePosition, 1000);
        }

        public override Rect GetBoundingBox()
        {
            return new Rect(
                this.Scaler.FUnitsToPixels(this.Head.BBox.X, 1000),
                this.Scaler.FUnitsToPixels(this.Head.BBox.Y, 1000),
                this.Scaler.FUnitsToPixels(this.Head.BBox.Width, 1000),
                this.Scaler.FUnitsToPixels(this.Head.BBox.Height, 1000));
        }

        public override byte[] GetData()
        {
            return this.Reader.Data;
        }

        internal override CFFFontSource GetCFF()
        {
            if (!this.OffsetTable.HasOpenTypeOutlines)
            {
                return null;
            }

            var value = this.ReadCFFTable();
            return value;
        }

        public static OpenTypeFontSourceType GetFontType(OpenTypeFontReader reader)
        {
            OpenTypeFontSourceType res = OpenTypeFontSourceType.Unknown;
            reader.BeginReadingBlock();
            uint tag = reader.ReadULong();
            reader.EndReadingBlock();
            reader.BeginReadingBlock();
            double v = reader.ReadFixed();
            reader.EndReadingBlock();

            if (tag == Tags.TRUE_TYPE_COLLECTION)
            {
                res = OpenTypeFontSourceType.TrueTypeCollection;
            }
            else if (v == 1.0) // TODO: OpenType font with CCF outlines.
            {
                res = OpenTypeFontSourceType.TrueType;
            }
            return res;
        }

        public override bool TryGetCharCode(int unicode, out int charCode)
        {
            bool result = false;
            ushort glyphId;
            charCode = CMap.MISSING_GLYPH_ID;
            if (this.TryGetGlyphId(unicode, out glyphId))
            {
                ushort uCharCode = CMap.MISSING_GLYPH_ID;
                CMapTable table = this.CMap.GetCMapTable(3, 0);
                if (table != null)
                {
                    result = table.TryGetCharId(glyphId, out uCharCode);
                }

                table = this.CMap.GetCMapTable(1, 0);
                if (table != null)
                {
                    result = table.TryGetCharId(glyphId, out uCharCode);
                }

                charCode = uCharCode;

                return result;
            }

            return false;
        }

        public override double GetAdvancedWidthOverride(int glyphId)
        {
            return this.Scaler.GetAdvancedWidth((ushort)glyphId, 1000);
        }

        internal override GlyphData GetGlyphData(ushort glyphIndex)
        {
            GlyphData res;
            if (!this.glyphsData.TryGetValue(glyphIndex, out res))
            {
                long offset = this.Loca.GetOffset(glyphIndex);

                TableRecord glyfRecord;
                this.tables.TryGetValue(Tags.GLYF_TABLE, out glyfRecord);
                if (offset == -1 || glyfRecord == null || offset >= glyfRecord.Offset + glyfRecord.Length)
                {
                    return new GlyphData(this, glyphIndex);
                }

                offset += glyfRecord.Offset;
                this.Reader.BeginReadingBlock();
                this.Reader.Seek(offset, System.IO.SeekOrigin.Begin);
                res = GlyphData.ReadGlyf(this, glyphIndex);
                this.Reader.EndReadingBlock();
                this.glyphsData[glyphIndex] = res;
            }
            return res;
        }

        private void ReadTableData<T>(T table) where T : TrueTypeTableBase
        {
            TableRecord record;
            if (!this.tables.TryGetValue(table.Tag, out record))
            {
                return;
            }
            long offset = this.Offset + record.Offset;
            this.Reader.BeginReadingBlock();
            this.Reader.Seek(offset, System.IO.SeekOrigin.Begin);
            table.Read(this.Reader);
            table.Offset = offset;
            this.Reader.EndReadingBlock();
        }

        private T ReadTableData<T>(uint tag, ReadTableFormatDelegate<T> readTableDelegate) where T : TrueTypeTableBase
        {
            TableRecord record;
            if (!this.tables.TryGetValue(tag, out record))
            {
                return null;
            }
            this.Reader.BeginReadingBlock();
            long offset = this.Offset + record.Offset;
            this.Reader.Seek(offset, System.IO.SeekOrigin.Begin);
            T table = readTableDelegate(this, this.Reader);
            table.Offset = offset;
            this.Reader.EndReadingBlock();
            return table;
        }

        private int GetTableLength(uint tag)
        {
            TableRecord record = this.tables[tag];
            int len = (int)(this.Reader.Length - record.Offset);
            foreach (TableRecord r in this.tables.Values)
            {
                int l = (int)(r.Offset - record.Offset);
                if (l > 0 && l < len)
                {
                    len = l;
                }
            }
            return len;
        }

        private CFFFontSource ReadCFFTable()
        {
            int len = this.GetTableLength(Tags.CFF_TABLE);
            byte[] data = new byte[len];

            this.Reader.BeginReadingBlock();
            long offset = this.Offset + this.tables[Tags.CFF_TABLE].Offset;
            this.Reader.Seek(offset, System.IO.SeekOrigin.Begin);
            this.Reader.Read(data, len);
            this.Reader.EndReadingBlock();

            CFFFontFile file = new CFFFontFile(data);
            return file.FontSource;
        }

        private void ReadTableRecords()
        {
            this.tables = new Dictionary<uint, TableRecord>();
            for (int i = 0; i < this.OffsetTable.NumTables; i++)
            {
                TableRecord tr = new TableRecord();
                tr.Read(this.Reader);
                this.tables[tr.Tag] = tr;
            }
        }

        private void Initialize()
        {
            this.OffsetTable = new OffsetTable();
            this.OffsetTable.Read(this.Reader);

            this.ReadTableRecords();
            this.glyphsData = new Dictionary<ushort, GlyphData>();
        }
    }
}
