﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Core.Fonts.Glyphs;
using Telerik.Windows.Documents.Core.Fonts.OpenType.Tables;
using Telerik.Windows.Documents.Core.Fonts.OpenType.Tables.TrueTypeOutlines;
using Telerik.Windows.Documents.Core.Fonts.OpenType.Tables.TrueTypeTables;
using Telerik.Windows.Documents.Core.Fonts.Type1.CFFFormat;

namespace Telerik.Windows.Documents.Core.Fonts.OpenType
{
    internal abstract class OpenTypeFontSourceBase : FontSource
    {
        private static object lockObject = new object();

        private readonly OpenTypeGlyphScaler scaler;
        private readonly OpenTypeFontReader reader;

        private Outlines? outlines;
        private CFFFontSource cff;
        private CMap cMap;
        private Head head;
        private HorizontalHeader hHea;
        private HorizontalMetrics hMtx;
        private Kerning kern;
        private GlyphSubstitution gSub;
        private ushort? glyphCount;

        public OpenTypeFontSourceBase(OpenTypeFontReader reader)
            : base()
        {
            this.reader = reader;
            this.scaler = new OpenTypeGlyphScaler(this);
        }

        internal OpenTypeFontReader Reader
        {
            get
            {
                return this.reader;
            }
        }

        internal Outlines Outlines
        {
            get
            {
                lock (lockObject)
                {
                    if (!this.outlines.HasValue)
                    {
                        lock (lockObject)
                        {

                            this.outlines = this.GetOutlines();
                        }
                    }

                    return this.outlines.Value;
                }
            }
        }

        internal CFFFontSource CFF
        {
            get
            {
                lock (lockObject)
                {
                    if (this.cff == null)
                    {
                        lock (lockObject)
                        {

                            this.cff = this.GetCFF();
                        }
                    }

                    return this.cff;
                }
            }
        }

        internal CMap CMap
        {
            get
            {
                lock (lockObject)
                {
                    if (this.cMap == null)
                    {
                        lock (lockObject)
                        {

                            this.cMap = this.GetCMap();
                        }
                    }

                    return this.cMap;
                }
            }
        }

        internal Head Head
        {
            get
            {
                lock (lockObject)
                {
                    if (this.head == null)
                    {
                        lock (lockObject)
                        {

                            this.head = this.GetHead();
                        }
                    }

                    return this.head;
                }
            }
        }

        internal HorizontalHeader HHea
        {
            get
            {
                lock (lockObject)
                {
                    if (this.hHea == null)
                    {
                        lock (lockObject)
                        {

                            this.hHea = this.GetHHea();
                        }
                    }

                    return this.hHea;
                }
            }
        }

        internal HorizontalMetrics HMtx
        {
            get
            {
                lock (lockObject)
                {
                    if (this.hMtx == null)
                    {
                        lock (lockObject)
                        {

                            this.hMtx = this.GetHMtx();
                        }
                    }

                    return this.hMtx;
                }
            }
        }

        internal Kerning Kern
        {
            get
            {
                lock (lockObject)
                {
                    if (this.kern == null)
                    {
                        lock (lockObject)
                        {

                            this.kern = this.GetKern();
                        }
                    }

                    return this.kern;
                }
            }
        }

        internal GlyphSubstitution GSub
        {
            get
            {
                lock (lockObject)
                {
                    if (this.gSub == null)
                    {
                        lock (lockObject)
                        {

                            this.gSub = this.GetGSub();
                        }
                    }

                    return this.gSub;
                }
            }
        }

        internal ushort GlyphCount
        {
            get
            {
                lock (lockObject)
                {
                    if (!this.glyphCount.HasValue)
                    {
                        lock (lockObject)
                        {

                            this.glyphCount = this.GetGlyphCount();
                        }
                    }

                    return this.glyphCount.Value;
                }
            }
        }

        internal abstract Outlines GetOutlines();

        internal abstract CFFFontSource GetCFF();

        internal abstract CMap GetCMap();

        internal abstract Head GetHead();

        internal abstract HorizontalHeader GetHHea();

        internal abstract HorizontalMetrics GetHMtx();

        internal abstract Kerning GetKern();

        internal abstract GlyphSubstitution GetGSub();

        internal abstract ushort GetGlyphCount();

        internal ushort NumberOfHorizontalMetrics
        {
            get
            {
                return this.HHea.NumberOfHMetrics;
            }
        }

        internal OpenTypeGlyphScaler Scaler
        {
            get
            {
                return this.scaler;
            }
        }

        public override void GetAdvancedWidthOverride(Glyph glyph)
        {
            glyph.AdvancedWidth = this.scaler.GetAdvancedWidth(glyph.GlyphId, 1);
        }

        internal Script GetScript(uint tag)
        {
            return this.GSub.GetScript(tag);
        }

        internal Feature GetFeature(ushort index)
        {
            return this.GSub.GetFeature(index);
        }

        internal Lookup GetLookup(ushort index)
        {
            return this.GSub.GetLookup(index);
        }

        internal abstract GlyphData GetGlyphData(ushort glyphID);

        public override bool TryGetGlyphId(int unicode, out ushort glyphId)
        {
            return this.CMap.TryGetGlyphId(unicode, out glyphId);
        }

        public double GetLineHeight(double fontSize)
        {
            double height = Math.Abs(this.HHea.Ascender) + Math.Abs(this.HHea.Descender) + this.HHea.LineGap;
            return this.Scaler.FUnitsToPixels(height, fontSize);
        }

        public double GetBaselineOffset(double fontSize)
        {
            double height = Math.Abs(this.HHea.Ascender) + Math.Abs(this.HHea.LineGap);

            return this.Scaler.FUnitsToPixels(height, fontSize);
        }

        public override void InitializeGlyphOutlinesOverride(Glyph glyph, double fontSize)
        {
            this.scaler.GetScaleGlyphOutlines(glyph, fontSize);
        }

        public override double GetAscent()
        {
            return this.Scaler.FUnitsToPixels(this.HHea.Ascender, 1000);
        }

        public override double GetDescent()
        {
            return this.Scaler.FUnitsToPixels(this.HHea.Descender, 1000);
        }

        public override double GetLineGap()
        {
            return this.Scaler.FUnitsToPixels(this.HHea.LineGap, 1000);
        }

        internal void Write(FontWriter writer)
        {
            writer.WriteString(this.FontFamily);
            ushort fl = 0;
            if (this.IsBold)
            {
                fl |= 1;
            }
            if (this.IsItalic)
            {
                fl |= 2;
            }
            writer.WriteUShort(fl);
            writer.WriteUShort(this.GlyphCount);
            this.Head.Write(writer);
            this.CMap.Write(writer);
            this.HHea.Write(writer);
            this.HMtx.Write(writer);
            this.Kern.Write(writer);
            this.GSub.Write(writer);
        }
    }
}
