﻿using System;
using System.Collections.Generic;
using System.Linq;
#if !NETSTANDARD
using System.Windows;
#else
using Telerik.Documents.Primitives;
#endif
using Telerik.Windows.Documents.Core.Fonts.Glyphs;
using Telerik.Windows.Documents.Core.Fonts.OpenType.Tables.TrueTypeOutlines;
using Telerik.Windows.Documents.Core.Fonts.Utils;
using Telerik.Windows.Documents.Core.Shapes;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Fonts.OpenType
{
    internal class OpenTypeGlyphScaler
    {
        internal const double Dpi = 72;
        private const double Ppi = 72;

        private readonly OpenTypeFontSourceBase fontFile;

        public OpenTypeGlyphScaler(OpenTypeFontSourceBase fontFile)
        {
            Guard.ThrowExceptionIfNull(fontFile, "fontFile");

            this.fontFile = fontFile;
        }
        
        public double GetAdvancedWidth(ushort glyphId, double fontSize)
        {
            return this.FUnitsToPixels(this.fontFile.HMtx.GetAdvancedWidth(glyphId), fontSize);
        }

        public double FUnitsToPixels(int units, double fontSize)
        {
            return this.FUnitsToPixels((double)units, fontSize);
        }

        public double FUnitsToPixels(double units, double fontSize)
        {
            return units * Dpi * fontSize / (Ppi * this.fontFile.Head.UnitsPerEm);
        }

        public Point FUnitsOutlinePointToPixels(Point unitPoint, double fontSize)
        {
            double x = this.FUnitsToPixels(unitPoint.X, fontSize);
            double y = this.FUnitsToPixels(unitPoint.Y, fontSize);
            return new Point(x, -y);
        }

        public Point FUnitsPointToPixels(Point unitPoint, double fontSize)
        {
            double x = this.FUnitsToPixels(unitPoint.X, fontSize);
            double y = this.FUnitsToPixels(unitPoint.Y, fontSize);
            return new Point(x, y);
        }

        public void GetScaleGlyphOutlines(Glyph glyph, double fontSize)
        {
            Guard.ThrowExceptionIfNull(glyph, "glyph");

            switch (this.fontFile.Outlines)
            {
                case Outlines.OpenType:
                    this.CreateOpenTypeOutlines(glyph, fontSize);
                    break;
                case Outlines.TrueType:
                    this.CreateTrueTypeOutlines(glyph, fontSize);
                    break;
            }
        }

        private static Point GetMidPoint(Point a, Point b)
        {
            return new Point((a.X + b.X) / 2, (a.Y + b.Y) / 2);
        }

        private static LineSegment CreateLineSegment(Point point)
        {
            LineSegment s = new LineSegment();
            s.Point = point;
          
            return s;
        }

        private static QuadraticBezierSegment CreateBezierSegment(Point control, Point end)
        {
            QuadraticBezierSegment s = new QuadraticBezierSegment();
            s.Point1 = control;
            s.Point2 = end;
           
            return s;
        }

        private PathFigure CreatePathFigureFromContour(OutlinePoint[] points, double fontSize)
        {
            List<OutlinePoint> finalPoints = GetFinalGlyphOutlinePoints(points);

            PathFigure f = new PathFigure();

            int startIndex = finalPoints[0].IsOnCurve ? 0 : 1;
            f.StartPoint = this.FUnitsOutlinePointToPixels(finalPoints[startIndex].Point, fontSize);
            Func<int, int> getNextIndex = (index) => { return (index + 1) % finalPoints.Count; };
            int currentIndex = startIndex;

            do
            {
                int firstIndex = getNextIndex(currentIndex);
                OutlinePoint firstPoint = finalPoints[firstIndex];

                if (firstPoint.IsOnCurve)
                {
                    f.Segments.Add(CreateLineSegment(this.FUnitsOutlinePointToPixels(firstPoint.Point, fontSize)));
                    currentIndex = firstIndex;
                }
                else
                {
                    int secondIndex = getNextIndex(firstIndex);
                    OutlinePoint secondPoint = finalPoints[secondIndex];
                    f.Segments.Add(CreateBezierSegment(this.FUnitsOutlinePointToPixels(firstPoint.Point, fontSize), this.FUnitsOutlinePointToPixels(secondPoint.Point, fontSize)));
                    currentIndex = secondIndex;
                }
            }
            while (currentIndex != startIndex);
          
            f.IsClosed = true;
            f.IsFilled = true;
          
            return f;
        }

        private static List<OutlinePoint> GetFinalGlyphOutlinePoints(OutlinePoint[] points)
        {
            List<OutlinePoint> finalPoints = new List<OutlinePoint>();

            for (int i = 0; i < points.Length; i++)
            {
                OutlinePoint currentPoint = points[i];
                OutlinePoint nextPoint = points[(i + 1) % points.Length];
                finalPoints.Add(currentPoint);

                if (!currentPoint.IsOnCurve && !nextPoint.IsOnCurve)
                {
                    OutlinePoint pointOnCurve = new OutlinePoint(1);
                    pointOnCurve.Point = GetMidPoint(currentPoint.Point, nextPoint.Point);
                    finalPoints.Add(pointOnCurve);
                }
            }

            return finalPoints;
        }

        private void CreateOpenTypeOutlines(Glyph glyph, double fontSize)
        {
            Guard.ThrowExceptionIfNull(glyph, "glyph");

            this.fontFile.CFF.InitializeGlyphOutlines(glyph, fontSize);
        }
        
        private void CreateTrueTypeOutlines(Glyph glyph, double fontSize)
        {
            Guard.ThrowExceptionIfNull(glyph, "glyph");

            glyph.Outlines = new GlyphOutlinesCollection();
            GlyphData gd = this.fontFile.GetGlyphData(glyph.GlyphId);
            foreach (var contour in gd.Contours)
            {
                glyph.Outlines.Add(this.CreatePathFigureFromContour(contour, fontSize));
            }
        }
    }
}
