﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Core.Fonts.OpenType.Tables.Utils;

namespace Telerik.Windows.Documents.Core.Fonts.OpenType.Tables
{
    internal class CMap : TrueTypeTableBase
    {
        private static readonly object lockObject = new object();

        private const ushort UNICODE_PLATFORM_ID = 0;
        private const ushort WINDOWS_PLATFORM_ID = 3;
        private const ushort DEFAULT_SEMANTIC_ID = 0;
        private const ushort SYMBOL_ENCODING_ID = 0;
        private const ushort UNICODE_ENCODING_ID = 1;

        internal const ushort MISSING_GLYPH_ID = 0;

        private static readonly List<Tuple<ushort, ushort>> supportedPlatformIdToEncodingIdCMapTables;

        static CMap()
        {
            CMap.supportedPlatformIdToEncodingIdCMapTables = new List<Tuple<ushort, ushort>>();
            CMap.supportedPlatformIdToEncodingIdCMapTables.Add(new Tuple<ushort, ushort>(WINDOWS_PLATFORM_ID, UNICODE_ENCODING_ID));
            CMap.supportedPlatformIdToEncodingIdCMapTables.Add(new Tuple<ushort, ushort>(WINDOWS_PLATFORM_ID, SYMBOL_ENCODING_ID));
        }

        public CMap(OpenTypeFontSourceBase fontFile)
            : base(fontFile)
        {
        }

        private EncodingRecord[] encodings;
        private Dictionary<EncodingRecord, CMapTable> tables;

        private bool isInitialized;
        private CMapTable encoding;

        internal override uint Tag
        {
            get { return Tags.CMAP_TABLE; }
        }

        private void Initialize()
        {
            lock (lockObject)
            {
                if (this.isInitialized)
                {
                    return;
                }

                for (int i = 0; i < CMap.supportedPlatformIdToEncodingIdCMapTables.Count && this.encoding == null; i++)
                {
                    var platformIdToEncodingIdTuple = CMap.supportedPlatformIdToEncodingIdCMapTables[i];
                    this.encoding = this.GetCMapTable(platformIdToEncodingIdTuple.Item1, platformIdToEncodingIdTuple.Item2);
                }

                this.isInitialized = true;
            }
        }

        private CMapTable GetCMapTable(OpenTypeFontReader reader, EncodingRecord record)
        {
            CMapTable table;
            if (!this.tables.TryGetValue(record, out table))
            {
                reader.BeginReadingBlock();
                reader.Seek(this.Offset + record.Offset, System.IO.SeekOrigin.Begin);
                table = CMapTable.ReadCMapTable(reader);
                reader.EndReadingBlock();
                this.tables[record] = table;
            }
            return table;
        }

        public virtual CMapTable GetCMapTable(ushort platformId, ushort encodingId)
        {
            if (this.encodings == null)
            {
                return null;
            }
            EncodingRecord record = Enumerable.FirstOrDefault(this.encodings, e => e.PlatformId == platformId && e.EncodingId == encodingId);
            if (record == null)
            {
                return null;
            }

            return this.GetCMapTable(this.Reader, record);
        }

        public bool TryGetGlyphId(int unicode, out ushort glyphId)
        {
            glyphId = MISSING_GLYPH_ID;

            this.Initialize();

            //// TODO: Handle non-unicode symbols.
            if (this.encoding == null)
            {
                return false;
            }

            return this.encoding.TryGetGlyphId(unicode, out glyphId);
        }

        public override void Read(OpenTypeFontReader reader)
        {
            lock (lockObject)
            {
                reader.ReadUShort();
                ushort numTables = reader.ReadUShort();

                this.encodings = new EncodingRecord[numTables];
                this.tables = new Dictionary<EncodingRecord, CMapTable>(numTables);
                for (int i = 0; i < numTables; i++)
                {
                    EncodingRecord record = new EncodingRecord();
                    record.Read(reader);
                    this.encodings[i] = record;
                }
            }
        }

        internal override void Write(FontWriter writer)
        {
            this.Initialize();
            this.encoding.Write(writer);
        }

        internal override void Import(OpenTypeFontReader reader)
        {
            this.encoding = CMapTable.ImportCMapTable(reader);
            this.isInitialized = true;
        }
    }
}
