﻿using System;
using System.Linq;

namespace Telerik.Windows.Documents.Core.Fonts.OpenType.Tables
{
    internal class GlyphSubstitution : TrueTypeTableBase
    {
        #region Construcotrs

        public GlyphSubstitution(OpenTypeFontSourceBase fontFile)
            : base(fontFile)
        {
        }

        #endregion


        #region Fields

        private ushort scriptListOffset;
        private ushort featureListOffset;
        private ushort lookupListOffset;

        private ScriptList scriptList;
        private LookupList lookupList;
        private FeatureList featureList;

        #endregion


        #region Properties

        internal override uint Tag
        {
            get { return Tags.GSUB_TABLE; }
        }

        #endregion


        #region Methods

        private void ReadTable<T>(T table, ushort offset) where T : TableBase
        {
            if (offset == 0)
            {
                return;
            }
            long tableOffset = this.Offset + offset;
            table.Offset = tableOffset;
            this.Reader.BeginReadingBlock();
            this.Reader.Seek(tableOffset, System.IO.SeekOrigin.Begin);
            table.Read(this.Reader);
            this.Reader.EndReadingBlock();
        }

        private void ReadScriptList()
        {
            this.scriptList = new ScriptList(this.FontSource);
            this.ReadTable(this.scriptList, this.scriptListOffset);
        }

        private void ReadFeatureList()
        {
            this.featureList = new FeatureList(this.FontSource);
            this.ReadTable(this.featureList, this.featureListOffset);
        }

        private void ReadLookupList()
        {
            this.lookupList = new LookupList(this.FontSource);
            this.ReadTable(this.lookupList, this.lookupListOffset);
        }

        
        
        public Script GetScript(uint tag)
        {
            if (this.scriptList == null)
            {
                this.ReadScriptList();
            }
            return this.scriptList.GetScript(tag);
        }

        public Feature GetFeature(ushort index)
        {
            if (this.featureList == null)
            {
                this.ReadFeatureList();
            }
            return this.featureList.GetFeature(index);
        }

        public Lookup GetLookup(ushort index)
        {
            if (this.lookupList == null)
            {
                this.ReadLookupList();
            }
            return this.lookupList.GetLookup(index);
        }

        #endregion


        #region Overrides

        public override void Read(OpenTypeFontReader reader)
        {
            reader.ReadFixed();
            this.scriptListOffset = reader.ReadUShort();
            this.featureListOffset = reader.ReadUShort();
            this.lookupListOffset = reader.ReadUShort();
        }

        internal override void Write(FontWriter writer)
        {
            this.ReadScriptList();
            this.ReadFeatureList();
            this.ReadLookupList();
            if (this.scriptList == null || this.featureList == null || this.lookupList == null)
            {
                writer.Write(0);
                return;
            }
            writer.Write(1);
            this.scriptList.Write(writer);
            this.featureList.Write(writer);
            this.lookupList.Write(writer);
        }

        internal override void Import(OpenTypeFontReader reader)
        {
            byte isNull = reader.Read();
            if (isNull > 0)
            {
                this.scriptList = new ScriptList(this.FontSource);
                this.featureList = new FeatureList(this.FontSource);
                this.lookupList = new LookupList(this.FontSource);

                this.scriptList.Import(reader);
                this.featureList.Import(reader);
                this.lookupList.Import(reader);
            }
        }

        #endregion
    }
}
