﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Core.Fonts.Utils;

namespace Telerik.Windows.Documents.Core.Fonts.OpenType.Tables
{
    internal class LigatureSubstitution : SubTable
    {
        #region Constructors

        public LigatureSubstitution(OpenTypeFontSourceBase fontFile)
            : base(fontFile)
        {
        }

        #endregion


        #region Fields

        private ushort coverageOffset;
        private ushort[] ligatureSetOffsets;

        private LigatureSet[] ligatureSets;

        private Coverage coverage;

        #endregion


        #region Properties

        public Coverage Coverage
        {
            get
            {
                if (this.coverage == null)
                {
                    this.coverage = this.ReadCoverage(this.Reader, this.coverageOffset);
                }
                return this.coverage;
            }
        }

        public LigatureSet[] LigatureSets
        {
            get
            {
                if (this.ligatureSets == null)
                {
                    this.ligatureSets = new LigatureSet[this.ligatureSetOffsets.Length];
                    for (int i = 0; i < this.ligatureSets.Length; i++)
                    {
                        this.ligatureSets[i] = this.ReadLigatureSet(this.Reader, this.ligatureSetOffsets[i]);
                    }
                }
                return this.ligatureSets;
            }
        }

        #endregion


        #region Methods

        private LigatureSet ReadLigatureSet(OpenTypeFontReader reader, ushort offset)
        {
            reader.BeginReadingBlock();
            long ligatureSetOffset = this.Offset + offset;
            reader.Seek(ligatureSetOffset, System.IO.SeekOrigin.Begin);
            LigatureSet ls = new LigatureSet(this.FontSource);
            ls.Read(reader);
            ls.Offset = ligatureSetOffset;
            reader.EndReadingBlock();
            return ls;
        }

        #endregion


        #region Overrides

        public override GlyphsSequence Apply(GlyphsSequence glyphIDs)
        {
            GlyphsSequence res = new GlyphsSequence(glyphIDs.Count);
            for (int i = 0; i < glyphIDs.Count; i++)
            {
                int coverageIndex = this.Coverage.GetCoverageIndex(glyphIDs[i].GlyphId);
                if (coverageIndex < 0)
                {
                    res.Add(glyphIDs[i]);
                }
                else
                {
                    Ligature l = this.LigatureSets[coverageIndex].FindLigature(glyphIDs, i);
                    if (l != null)
                    {
                        res.Add(l.LigatureGlyphId);
                        i += l.Length;
                    }
                    else
                    {
                        res.Add(glyphIDs[i]);
                    }
                }
            }
            return res;
        }

        public override void Read(OpenTypeFontReader reader)
        {
            reader.ReadUShort();
            this.coverageOffset = reader.ReadUShort();
            ushort ligSetCount = reader.ReadUShort();
            this.ligatureSetOffsets = new ushort[ligSetCount];
            for (int i = 0; i < ligSetCount; i++)
            {
                this.ligatureSetOffsets[i] = reader.ReadUShort();
            }
        }

        internal override void Write(FontWriter writer)
        {
            this.Coverage.Write(writer);
            writer.WriteUShort((ushort)this.ligatureSetOffsets.Length);
            foreach (LigatureSet ls in this.LigatureSets)
            {
                ls.Write(writer);
            }
        }

        internal override void Import(OpenTypeFontReader reader)
        {
            this.coverage = Coverage.ImportCoverageTable(this.FontSource, reader);
            ushort ligSetCount = reader.ReadUShort();
            this.ligatureSets = new LigatureSet[ligSetCount];
            for (int i = 0; i < ligSetCount; i++)
            {
                LigatureSet ls = new LigatureSet(this.FontSource);
                ls.Import(reader);
                this.ligatureSets[i] = ls;
            }
        }

        #endregion
    }
}
