﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Core.PostScript.Data;
using Telerik.Windows.Documents.Fixed.Exceptions;

namespace Telerik.Windows.Documents.Core.Fonts.Type1.CFFFormat.CFFTables
{
    internal class Charset : CFFTable
    {
        #region Consts

        internal const ushort NotDefIndex = 0;

        #endregion


        #region Constructors

        public Charset(ICFFFontFile file, long offset, int count)
            : base(file, offset)
        {
            this.count = count;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Charset" /> class.
        /// </summary>
        /// <param name="file">Font file.</param>
        /// <param name="glyphs">Array of cids or character codes, depending of the font encoding.</param>
        public Charset(ICFFFontFile file, ushort[] glyphs)
            : base(file, -1)
        {
            this.Initialize(glyphs);
        }

        #endregion


        #region Fields

        private readonly int count;
        private Dictionary<string, ushort> nameToGidMapping;
        private Dictionary<ushort, ushort> cidToGidMapping;
        private string[] names;

        #endregion


        #region Indexers

        public ushort this[string name]
        {
            get
            {
                ushort res;
                if (this.nameToGidMapping.TryGetValue(name, out res))
                {
                    return res;
                }
                return NotDefIndex;
            }
        }

        public string this[ushort index]
        {
            get
            {
                return this.names[index];
            }
        }

        #endregion


        #region Methods

        public bool TryGetGlyphIdByCharacterName(string name, out ushort glyphId)
        {
            return this.nameToGidMapping.TryGetValue(name, out glyphId);
        }

        public bool TryGetGlyphIdByCharacterIdentifier(ushort cid, out ushort glyphId)
        {
            return this.cidToGidMapping.TryGetValue(cid, out glyphId);
        }

        private List<ushort> ReadFormat0(CFFFontReader reader)
        {
            int cnt = this.count - 1;
            List<ushort> glyphs = new List<ushort>(cnt);
            for (int i = 0; i < cnt; i++)
            {
                glyphs.Add(reader.ReadSID());
            }
            return glyphs;
        }

        private List<ushort> ReadFormat1(CFFFontReader reader)
        {
            int cnt = this.count - 1;
            List<ushort> glyphs = new List<ushort>(cnt);
            ushort first;
            byte left;
            while (glyphs.Count < cnt)
            {
                first = reader.ReadSID();
                left = reader.ReadCard8();
                glyphs.Add(first);
                for (int j = 0; j < left; j++)
                {
                    glyphs.Add((ushort)(first + j + 1));
                }
            }
            return glyphs;
        }

        private List<ushort> ReadFormat2(CFFFontReader reader)
        {
            int cnt = this.count - 1;
            List<ushort> glyphs = new List<ushort>(cnt);
            ushort first;
            ushort left;
            while (glyphs.Count < cnt)
            {
                first = reader.ReadSID();
                left = reader.ReadCard16();
                glyphs.Add(first);
                for (int j = 0; j < left; j++)
                {
                    glyphs.Add((ushort)(first + j + 1));
                }
            }
            return glyphs;
        }

        private void Initialize(ushort[] glyphs)
        {
            this.nameToGidMapping = new Dictionary<string, ushort>(glyphs.Length);
            this.names = new string[glyphs.Length];
            this.cidToGidMapping = new Dictionary<ushort, ushort>(glyphs.Length);

            for (ushort i = 0; i < glyphs.Length; i++)
            {
                ushort charEncoding = glyphs[i];
                string name = this.File.ReadString(charEncoding);
                ushort glyphId = (ushort)(i + 1);

                this.nameToGidMapping[name] = glyphId;
                this.names[i] = name;

                this.cidToGidMapping[charEncoding] = glyphId;
            }
        }

        #endregion


        #region Overrides

        public override void Read(CFFFontReader reader)
        {
            byte format = reader.ReadCard8();
            List<ushort> glyphs;
            switch (format)
            {
                case 0:
                    glyphs = this.ReadFormat0(reader);
                    break;
                case 1:
                    glyphs = this.ReadFormat1(reader);
                    break;
                case 2:
                    glyphs = this.ReadFormat2(reader);
                    break;
                default:
                    throw new NotSupportedCharsetFormatException(format);
            }

            this.Initialize(glyphs.ToArray());
        }

        #endregion
    }
}
