﻿using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Core.Data;
using Telerik.Windows.Documents.Core.Fonts.Glyphs;
using Telerik.Windows.Documents.Core.Fonts.Type1.CFFFormat.Data;
using Telerik.Windows.Documents.Core.Fonts.Type1.CharStrings;
using Telerik.Windows.Documents.Core.Fonts.Type1.Data;
using Telerik.Windows.Documents.Core.Fonts.Type1.Type1Format.Converters;
using Telerik.Windows.Documents.Core.Fonts.Type1.Utils;
using Telerik.Windows.Documents.Core.PostScript.Data;
using Telerik.Windows.Documents.Fixed.Model.Fonts.Encoding;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.Fonts.Type1.Type1Format.Dictionaries
{
    internal class Type1Font : PostScriptObject, IBuildCharHolder
    {
        private readonly BuildChar buildChar;
        private readonly Dictionary<string, GlyphData> glyphOutlines;
        private readonly Property<PostScriptArray> fontMatrix;
        private readonly Property<FontInfo> fontInfo;
        private readonly Property<PostScriptArray> encoding;
        private readonly Property<PostScriptDictionary> charStrings;
        private readonly Property<Private> priv;
        private readonly Property<PostScriptArray> fontBBox;

        public Type1Font()
        {
            this.buildChar = new BuildChar(this);
            this.glyphOutlines = new Dictionary<string, GlyphData>();

            this.fontMatrix = this.CreateProperty<PostScriptArray>(new PropertyDescriptor { Name = "FontMatrix" }, PostScriptArray.MatrixIdentity);
            this.fontInfo = this.CreateProperty<FontInfo>(new PropertyDescriptor { Name = "FontInfo" }, Type1Converters.PostScriptObjectConverter);
            this.encoding = this.CreateProperty<PostScriptArray>(new PropertyDescriptor { Name = "Encoding" }, Type1Converters.EncodingConverter, PredefinedEncoding.StandardEncoding.ToArray());
            this.charStrings = this.CreateProperty<PostScriptDictionary>(new PropertyDescriptor { Name = "CharStrings" });
            this.priv = this.CreateProperty<Private>(new PropertyDescriptor { Name = "Private" }, Type1Converters.PostScriptObjectConverter);
            this.fontBBox = this.CreateProperty<PostScriptArray>(new PropertyDescriptor { Name = "FontBBox" });
        }
        
        public PostScriptArray FontMatrix
        {
            get
            {
                return this.fontMatrix.GetValue();
            }
        }

        public PostScriptArray FontBBox
        {
            get
            {
                return this.fontBBox.GetValue();
            }
        }

        public FontInfo FontInfo
        {
            get
            {
                return this.fontInfo.GetValue();
            }
        }

        public PostScriptArray Encoding
        {
            get
            {
                return this.encoding.GetValue();
            }
        }

        public PostScriptDictionary CharStrings
        {
            get
            {
                return this.charStrings.GetValue();
            }
        }

        public Private Private
        {
            get
            {
                return this.priv.GetValue();
            }
        }

        public ushort GetAdvancedWidth(string glyphName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(glyphName, "glyphName");

            GlyphData data = this.GetGlyphData(glyphName);
            
            return data.AdvancedWidth;
        }

        public void GetGlyphOutlines(Glyph glyph, double fontSize)
        {
            Guard.ThrowExceptionIfNull(glyph, "glyph");

            GlyphData data = this.GetGlyphData(glyph.Name);
            GlyphOutlinesCollection outlines = data.Oultlines.Clone();
            Matrix m = this.FontMatrix.ToMatrix();
            m.ScaleAppend(fontSize, -fontSize);
            outlines.Transform(m);
            glyph.Outlines = outlines;
        }

        public GlyphData GetGlyphData(string name)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");

            GlyphData res;
            if (!this.glyphOutlines.TryGetValue(name, out res))
            {
                res = this.ReadGlyphData(name);
                this.glyphOutlines[name] = res;
            }

            return res;
        }

        public bool TryGetCharCode(ushort unicode, out ushort charId)
        {
            charId = 0;
            string name = AdobeGlyphList.GetName((char)unicode);
            if (this.Encoding.Contains(name))
            {
                charId = (ushort)this.Encoding.IndexOf(name);

                return true;
            }

            return false;
        }

        public byte[] GetSubr(int index)
        {
            return this.Private.GetSubr(index);
        }

        public byte[] GetGlobalSubr(int index)
        {
            throw new System.NotImplementedException();
        }

        internal string GetGlyphName(ushort cid)
        {
            if (this.Encoding == null)
            {
                return PostScriptConstants.NotDef;
            }

            string name = this.Encoding.GetElementAs<string>(cid);

            return name ?? PostScriptConstants.NotDef;
        }

        private GlyphData ReadGlyphData(string name)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");

            PostScriptString str = this.CharStrings.GetElementAs<PostScriptString>(name);
            this.buildChar.Execute(str.ToByteArray());
            return new GlyphData(buildChar.GlyphOutlines, (ushort?)buildChar.Width);

        } 
    }
}
