﻿using System;
using Telerik.Windows.Documents.Core.Imaging;
using Telerik.Windows.Documents.Fixed.Exceptions;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.PdfConverters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.Utils;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Filters;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Internal;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.ColorSpaces
{
    internal abstract class ColorSpaceOld : PdfObjectOld, IColorSpace
    {
        private const string DeviceGray = "DeviceGray";
        private const string DeviceGrayAbrv = "G";
        private const string DeviceRgb = "DeviceRGB";
        private const string DeviceRgbAbrv = "RGB";
        private const string DeviceCmyk = "DeviceCMYK";
        private const string DeviceCmykAbrv = "CMYK";
        private const string CalGray = "CalGray";
        private const string CalRgb = "CalRGB";
        private const string CalCmyk = "CalCMYK";
        private const string ICCBased = "ICCBased";
        private const string Indexed = "Indexed";
        private const string IndexedAbrv = "I";
        private const string Pattern = "Pattern";
        private const string Separation = "Separation";
        private const string DeviceN = "DeviceN";

        public ColorSpaceOld(PdfContentManager contentManager)
            : base(contentManager)
        {
        }

        public abstract int ComponentCount
        {
            get;
        }

        public virtual Brush DefaultBrush
        {
            get
            {
                return new SolidColorBrush(Color.Black);
            }
        }

        public abstract ColorSpace Type { get; }

        public static object[] CreateComponents(ColorSpaceOld cs, byte component)
        {
            object[] res = new object[cs.ComponentCount];
            for (int i = 0; i < cs.ComponentCount; i++)
            {
                res[i] = component;
            }

            return res;
        }

        public static bool IsColorSpace(string name)
        {
            name = GetColorSpaceName(name);

            switch (name)
            {
                case DeviceGray:
                case DeviceRgb:
                case DeviceCmyk:
                case CalCmyk:
                case CalGray:
                case CalRgb:
                case ICCBased:
                case Indexed:
                case Pattern:
                case Separation:
                    return true;
            }

            return false;
        }

        public static ColorSpaceOld CreateColorSpace(PdfContentManager contentManager, ColorSpace value)
        {
            switch (value)
            {
                case ColorSpace.RGB:
                    return new DeviceRgbOld(contentManager);

                case ColorSpace.Gray:
                    return new DeviceGrayOld(contentManager);

                default:
                    throw new NotSupportedException("Color space is not supported");
            }
        }

        public static ColorSpaceOld CreateColorSpace(PdfContentManager contentManager, string name)
        {
            name = GetColorSpaceName(name);
            switch (name)
            {
                case DeviceGray:
                    return new DeviceGrayOld(contentManager);
                case DeviceRgb:
                    return new DeviceRgbOld(contentManager);
                case DeviceCmyk:
                case CalCmyk:
                    return new DeviceCmykOld(contentManager);
                case CalGray:
                    return new CalGrayOld(contentManager);
                case CalRgb:
                    return new CalRgbOld(contentManager);
                case ICCBased:
                    return new ICCBasedOld(contentManager);
                case Indexed:
                    return new IndexedOld(contentManager);
                case Pattern:
                    return new PatternColorSpaceOld(contentManager);
                case Separation:
                    return new SeparationOld(contentManager);
                case DeviceN:
                    return new DeviceNOld(contentManager);
                default:
                    throw new NotSupportedColorSpaceException(name);
            }
        }

        public static ColorSpaceOld CreateColorSpace(PdfContentManager contentManager, PdfArrayOld array)
        {
            Guard.ThrowExceptionIfNull(contentManager, "contentManager");
            Guard.ThrowExceptionIfNull(array, "array");

            PdfNameOld name = array.GetElement<PdfNameOld>(0);

            ColorSpaceOld cs = CreateColorSpace(contentManager, name.Value);

            ICCBasedOld iccBased = cs as ICCBasedOld;
            if (iccBased != null)
            {
                iccBased.Profile = array.GetElement<ICCProfileOld>(1);

                return iccBased;
            }

            IndexedOld indexed = cs as IndexedOld;
            if (indexed != null)
            {
                indexed.Init(array);

                return indexed;
            }

            SeparationOld separation = cs as SeparationOld;
            if (separation != null)
            {
                separation.Init(array);

                return separation;
            }

            DeviceNOld deviceN = cs as DeviceNOld;
            if (deviceN != null)
            {
                deviceN.Init(array);

                return deviceN;
            }

            PatternColorSpaceOld pattern = cs as PatternColorSpaceOld;
            if (pattern != null && array.Count > 1)
            {
                pattern.UnderlyingColorSpace = array.GetElement<ColorSpaceOld>(1, Converters.ColorSpaceConverter);

                return pattern;
            }

            return cs;
        }

        public abstract Brush GetBrush(PdfResourceOld resources, object[] pars);

        public abstract Color GetColor(object[] pars);

        public abstract Color GetColor(byte[] bytes, int offset);

        public virtual void Clear()
        {
        }

        public override string ToString()
        {
            return this.GetType().Name;
        }

        public PixelContainer GetPixels(IImageDescriptor image, bool applyMask)
        {
            PixelContainer result = this.GetPixelsOverride(image, applyMask);
            this.Clear();

            return result;
        }

        public abstract double[] GetDefaultDecodeArray(int bitsPerComponent);

        protected abstract PixelContainer GetPixelsOverride(IImageDescriptor image, bool applyMask);

        protected static Color GetGrayColor(object[] pars)
        {
            Guard.ThrowExceptionIfNull(pars, "pars");

            double g;
            Helper.UnboxDouble(pars[pars.Length - 1], out g);

            return Color.FromGray(g);
        }

        protected static Color GetRgbColor(object[] pars)
        {
            Guard.ThrowExceptionIfNull(pars, "pars");

            int len = pars.Length;

            double r, g, b;
            Helper.UnboxDouble(pars[len - 3], out r);
            Helper.UnboxDouble(pars[len - 2], out g);
            Helper.UnboxDouble(pars[len - 1], out b);

            return Color.FromArgb(1, r, g, b);
        }

        protected static Color GetCmykColor(object[] pars)
        {
            Guard.ThrowExceptionIfNull(pars, "pars");

            int len = pars.Length;

            double c, m, y, k;
            Helper.UnboxDouble(pars[len - 4], out c);
            Helper.UnboxDouble(pars[len - 3], out m);
            Helper.UnboxDouble(pars[len - 2], out y);
            Helper.UnboxDouble(pars[len - 1], out k);

            return Color.FromCmyk(c, m, y, k);
        }

        protected static Color GetGrayColor(byte[] bytes, int offset)
        {
            Guard.ThrowExceptionIfNull(bytes, "bytes");

            return Color.FromGray(bytes[offset]);
        }

        protected static Color GetRgbColor(byte[] bytes, int offset)
        {
            Guard.ThrowExceptionIfNull(bytes, "bytes");

            return Color.FromArgb(255, bytes[offset], bytes[offset + 1], bytes[offset + 2]);
        }

        protected static Color GetCmykColor(byte[] bytes, int offset)
        {
            Guard.ThrowExceptionIfNull(bytes, "bytes");

            return Color.FromCmyk(bytes[offset], bytes[offset + 1], bytes[offset + 2], bytes[offset + 3]);
        }

        private static string GetColorSpaceName(string name)
        {
            switch (name)
            {
                case DeviceGrayAbrv:
                    return DeviceGray;
                case DeviceRgbAbrv:
                    return DeviceRgb;
                case DeviceCmykAbrv:
                    return DeviceCmyk;
                case IndexedAbrv:
                    return Indexed;
            }

            return name;
        }
    }
}
