﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Fixed.Exceptions;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.Enums;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.PdfConverters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.Text;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.Utils;
using Telerik.Windows.Documents.Fixed.Model.Internal;
using Telerik.Windows.Documents.Fixed.Model.Internal.Classes;
using FontSource = Telerik.Windows.Documents.Core.Fonts.FontSource;
using GlyphOutlinesCollection = Telerik.Windows.Documents.Core.Fonts.Glyphs.GlyphOutlinesCollection;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.Fonts
{
    internal abstract class FontBaseOld : PdfObjectOld
    {
        public const int FontSize = 100;

        private const string Type0 = "Type0";
        private const string Type1 = "Type1";
        private const string Type3 = "Type3";
        private const string TrueType = "TrueType";
        private const string MMType1 = "MMType1";

        private readonly InstantLoadProperty<CMapStream> unicode;
        private readonly Dictionary<CharCodeOld, GlyphOutlinesCollection> glyphOutlinesCache;

        private CMapOld mapping;

        public FontBaseOld(PdfContentManager contentManager)
            : base(contentManager)
        {
            this.unicode = this.CreateInstantLoadProperty<CMapStream>(new PdfPropertyDescriptor { Name = "ToUnicode" }, Converters.CMapStreamConverter);

            this.glyphOutlinesCache = new Dictionary<CharCodeOld, GlyphOutlinesCollection>();
        }

        public CMapStream ToUnicode
        {
            get
            {
                return this.unicode.GetValue();
            }
            set
            {
                this.unicode.SetValue(value);
            }
        }

        protected FontsManagerOld FontsManager
        {
            get
            {
                return this.ContentManager.FontsManager;
            }
        }

        protected CMapOld Mapping
        {
            get
            {
                if (this.mapping == null)
                {
                    this.mapping = this.FontsManager.GetFontToUnicode(this);
                }
                return this.mapping;
            }
        }

        protected abstract FontSource FontSource { get; }

        public static FontBaseOld CreateFont(PdfContentManager contentManager, PdfDictionaryOld dictionary)
        {
            FontBaseOld res;
            PdfNameOld subtype = dictionary.GetElement<PdfNameOld>(PdfNames.Subtype);
            switch (subtype.Value)
            {
                case Type0:
                    res = new Type0FontOld(contentManager);
                    break;
                case Type1:
                    res = new Type1FontOld(contentManager);
                    break;
                case MMType1:
                    res = new MMType1FontOld(contentManager);
                    break;
                case TrueType:
                    res = new TrueTypeFontOld(contentManager);
                    break;
                default:
                    throw new NotSupportedFontException(subtype.Value); 
            }
            
            res.Load(dictionary);
            
            return res;
        }

        public virtual double GetGlyphWidth(GlyphOld glyph)
        {
            if (this.FontSource != null)
            {
                this.FontSource.GetAdvancedWidth(glyph);
            }
            else
            {
                glyph.AdvancedWidth = 1;
            }

            return glyph.AdvancedWidth;
        }

        public virtual IEnumerable<GlyphOld> RenderGlyphs(PdfContext context, PdfStringOld str)
        {
            foreach (GlyphOld g in this.GetGlyphs(str))
            {
                this.RenderGlyph(context, g);
                yield return g;
            }
        }

        protected virtual IEnumerable<GlyphOld> GetGlyphs(PdfStringOld str)
        {
            if (str == null || str.Value == null)
            {
                yield break;
            }

            if (this.Mapping != null)
            {
                var text = this.Mapping.ToUnicode(str.Value);
                foreach (var p in text)
                {
                    yield return new GlyphOld { CharId = p.Item2, ToUnicode = p.Item1 };
                }
            }
            else
            {
                foreach (byte b in str.Value)
                {
                    yield return new GlyphOld { CharId = new CharCodeOld(b), ToUnicode = ((char)b).ToString() };
                }
            }
        }

        protected void GetGlyphOutlines(GlyphOld glyph)
        {
            if (this.FontSource != null)
            {
                GlyphOutlinesCollection outlines;
                if (this.glyphOutlinesCache.TryGetValue(glyph.CharId, out outlines))
                {
                    glyph.Outlines = outlines;
                }
                else
                {
                    this.FontSource.InitializeGlyphOutlines(glyph, FontSize);
                    this.glyphOutlinesCache[glyph.CharId] = glyph.Outlines;
                }
            }
        }

        protected virtual void RenderGlyph(PdfContext context, GlyphOld glyph)
        {
            glyph.Key = this.ContentManager.ResourceManager.GetFontKey(this);
            glyph.FontSize = context.TextState.FontSize;
            glyph.CharSpacing = context.TextState.CharSpacing;
            glyph.WordSpacing = glyph.IsSpace ? context.TextState.WordSpacing : 0;
            glyph.Rise = context.TextState.Rise;
            glyph.Fill = context.GraphicsState.GetBrushWithAlpha();
            glyph.Stroke = context.GraphicsState.GetStrokeBrushWithAlpha();
            glyph.HorizontalScaling = context.TextState.HorizontalScaling;
            RenderingMode mode = context.TextState.RenderingMode;
            glyph.IsFilled = mode == RenderingMode.Fill || mode == RenderingMode.FillClipping || mode == RenderingMode.FillStroke || mode == RenderingMode.FillStrokeClipping;
            glyph.IsStroked = mode == RenderingMode.FillStroke || mode == RenderingMode.FillStrokeClipping || mode == RenderingMode.Stroke || mode == RenderingMode.StrokeClipping;
            glyph.StrokeThickness = context.GraphicsState.LineWidth;
            glyph.TransformMatrix = context.GetTextRenderingMatrix();
            glyph.TextTranslation = context.TextState.TextTranslation;
            this.GetGlyphOutlines(glyph);
            this.GetGlyphToUnicode(glyph);
            glyph.Width = this.GetGlyphWidth(glyph);
            context.UpdateTextMatrix(glyph);
            glyph.FreezeGlyphInfo();
        }

        private void GetGlyphToUnicode(GlyphOld glyph)
        {
            string unicode;
            if (this.Mapping != null && this.Mapping.TryGetToUnicode(glyph.CharId, out unicode))
            {
                glyph.ToUnicode = unicode;
            }
            else
            {
                glyph.ToUnicode = ((char)glyph.CharId.IntValue).ToString();
            }
        }
    }
}
