﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.Exceptions;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.PdfConverters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.Text;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Filters;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.Filters
{
    /// <summary>
    /// Represents the filters manager.
    /// </summary>
    internal class FiltersManagerOld
    {
        private static readonly Dictionary<string, IPdfFilter> filters;

        static FiltersManagerOld()
        {
            filters = new Dictionary<string, IPdfFilter>();

            InitializeFilters();
        }

        /// <summary>
        /// Registers the specified PDF filter.
        /// </summary>
        /// <param name="filter">The filter.</param>
        public static void RegisterFilter(IPdfFilter filter)
        {
            filters[filter.Name] = filter;
        }

        internal static byte[] Decode(PdfDictionaryOld dictionary, PdfNameOld[] filters, byte[] data, DecodeParameters[] decodeParams)
        {
            Guard.ThrowExceptionIfNull(dictionary, "dictionary");
            Guard.ThrowExceptionIfNull(filters, "filters");
            Guard.ThrowExceptionIfNull(data, "data");

            FormatProviders.Pdf.Filters.PdfObject obj = new FormatProviders.Pdf.Filters.PdfObject(dictionary);
            DecodeParameters parms = decodeParams == null ? null : decodeParams[0];
            if (filters.Length == 0)
            {
                return data;
            }

            IPdfFilter filter = CreateFilter(filters[0].Value);
            int i = 1;
            while (true)
            {
                data = filter.Decode(obj, data, parms);
                if (i == filters.Length)
                {
                    return data;
                }

                parms = decodeParams == null ? null : decodeParams[i];
                filter = CreateFilter(filters[i].Value);
                i++;
            }
        }

        internal static PdfNameOld[] GetFilters(PdfContentManager contentManager, PdfDictionaryOld dictionary)
        {
            Guard.ThrowExceptionIfNull(contentManager, "contentManager");
            Guard.ThrowExceptionIfNull(dictionary, "dictionary");

            object obj;
            if (!dictionary.TryGetValue(PdfNames.Filter, out obj))
            {
                return new PdfNameOld[0];
            }

            PdfNameOld[] result = (PdfNameOld[])Converters.FiltersConverter.Convert(typeof(PdfNameOld[]), contentManager, obj);
            if (result == null)
            {
                return new PdfNameOld[0];
            }
            else
            {
                return result;
            }
        }

        internal static string[] GetFilterValues(PdfContentManager contentManager, PdfDictionaryOld dictionary)
        {
            PdfNameOld[] names = GetFilters(contentManager, dictionary);

            string[] result = null;

            if (names != null)
            {
                result = new string[names.Length];

                for (int i = 0; i < result.Length; i++)
                {
                    PdfNameOld name = names[i];
                    string filter = FormatProviders.Pdf.Model.PdfNames.GetNameFromAbbreviation(name.Value);
                    result[i] = filter;
                }
            }

            return result;
        }

        internal static DecodeParameters[] GetDecodeParameters(PdfContentManager contentManager, PdfDictionaryOld dictionary)
        {
            Guard.ThrowExceptionIfNull(contentManager, "contentManager");
            Guard.ThrowExceptionIfNull(dictionary, "dictionary");

            object obj;
            if (!dictionary.TryGetValue(PdfNames.DecodeParameters, out obj))
            {
                return null;
            }

            return (DecodeParameters[])Converters.DecodeParametersConverter.Convert(typeof(DecodeParameters[]), contentManager, obj);
        }

        private static void InitializeFilters()
        {
            FlateDecode fd = new FlateDecode();
            DctDecode  dct = new DctDecode ();
            ASCII85Decode ascii85 = new ASCII85Decode();
            ASCIIHexDecode asciihex = new ASCIIHexDecode();
            LZWDecode lzw = new LZWDecode();
            CCITTFaxDecode ccit = new CCITTFaxDecode();
            RunLengthDecode rl = new RunLengthDecode();
            JBIG2Decode jbig2 = new JBIG2Decode();

            filters[fd.Name] = fd;
            filters[dct.Name] = dct;
            filters[ascii85.Name] = ascii85;
            filters[asciihex.Name] = asciihex;
            filters[lzw.Name] = lzw;
            filters[ccit.Name] = ccit;
            filters[rl.Name] = rl;
            filters[jbig2.Name] = jbig2;
        }

        private static IPdfFilter CreateFilter(string filterName)
        {
            IPdfFilter filter;
            filterName = FormatProviders.Pdf.Model.PdfNames.GetNameFromAbbreviation(filterName);
            if (filters.TryGetValue(filterName, out filter))
            {
                return filter;
            }

            throw new NotSupportedFilterException(filterName); 
        }
    }
}
