﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Telerik.Windows.Documents.Core.Data;
using Telerik.Windows.Documents.Core.Shapes;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.Annot;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.Forms;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.Navigation;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.PdfConverters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader.Parsers;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.Collections;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Fixed.Model.Internal;
using Telerik.Windows.Documents.Fixed.Model.Internal.Classes;
using Telerik.Windows.Documents.Fixed.Model.Internal.Collections;
using Telerik.Windows.Documents.Fixed.Model.Navigation;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Utilities;
using RadFixedDocument = Telerik.Windows.Documents.Fixed.Model.RadFixedDocument;
using Stream = System.IO.Stream;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf
{
    /// <summary>
    /// Provides functionality for importing PDF documents.
    /// </summary>
    public partial class PdfFormatProvider
    {
        private readonly Dictionary<PageOld, RadFixedPageInternal> pagesMapping;
        private readonly Stream stream;
        private readonly FormatProviderSettings settings;
        private readonly bool isOldPdfViewer;

        private AcroFormCacheManager acroFormCacheManager;
        private DocumentChangesManager changesManager;
        private PdfContentManager contentManager;
        private PageOld[] pages;
        private RadFixedDocumentInternal document;

        /// <summary>
        /// Initializes a new instance of the <see cref="PdfFormatProvider" /> class.
        /// </summary>
        /// <param name="sourceStream">The source stream.</param>
        public PdfFormatProvider(Stream sourceStream)
            : this(sourceStream, FormatProviderSettings.ReadAllAtOnce)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="PdfFormatProvider" /> class. This overload should be used in order to import a document in RadPdfViewer.
        /// </summary>
        /// <param name="sourceStream">The stream.</param>
        /// <param name="settings">The settings.</param>
        public PdfFormatProvider(Stream sourceStream, FormatProviderSettings settings)
            : this(sourceStream, settings, true)
        {
        }

        internal PdfFormatProvider(Stream sourceStream, FormatProviderSettings settings, bool copyStreamForReadAllAtOnceMode)
            : this()
        {
            Guard.ThrowExceptionIfNull(sourceStream, "sourceStream");

            PdfImporter.ValidateInputStreamCanReadAndSeek(sourceStream);

            this.settings = settings;
            if (copyStreamForReadAllAtOnceMode && this.settings.ReadingMode == ReadingMode.AllAtOnce)
            {
                this.stream = new MemoryStream();

                sourceStream.Seek(0, SeekOrigin.Begin);
                sourceStream.CopyTo(this.stream);
            }
            else
            {
                this.stream = sourceStream;
            }

            this.stream.Seek(0, SeekOrigin.Begin);

            this.document = null;
            this.pagesMapping = new Dictionary<PageOld, RadFixedPageInternal>();

            this.isOldPdfViewer = true;
        }

        internal DocumentChangesManager ChangesManager
        {
            get
            {
                return this.changesManager;
            }
        }

        internal event EventHandler<OnDocumentExceptionEventArgs> OnException;

        /// <summary>
        /// Creates a <see cref="RadFixedDocument"/> that represents the PDF file.
        /// </summary>
        /// <returns>The fixed document.</returns>
        public RadFixedDocument Import()
        {
            if (!this.isOldPdfViewer)
            {
                throw new NotSupportedException("This method is only supported when initializing PdfFormatProvider through the PdfFormatProvider(Stream, FormatProviderSettings) constructor.");
            }
#if DEBUG
            PdfContentExtractor.ClearExtractedContent();
#endif
            this.acroFormCacheManager = new AcroFormCacheManager(this.ImportSettings);
            this.contentManager = new PdfContentManager(this, this.stream);

            RadFixedDocument document = this.acroFormCacheManager.FieldsAndWidgetsImportContext.Document;
            this.document = new RadFixedDocumentInternal(this, document);
            document.InternalDocument = this.document;
            PagesContentCache pagesContentCache = new PagesContentCache();
            ThreadSafePageContentManagerOld contentManager = new ThreadSafePageContentManagerOld(this);
            document.CacheManager = new LoadOnDemandPagesCacheManager(pagesContentCache, contentManager);

            this.changesManager = new DocumentChangesManager(document, this.stream);

            this.document.OnException += this.Document_OnException;
            this.LoadFormFields();
            this.RecalculateMissingWidgetAppearances();
            this.LoadPages();
            this.LoadBookmarks();
            this.LoadPageMode();
            this.document.OnException -= this.Document_OnException;

            document.SupportsSelection = true;
            document.TextDocument = this.document.TextDocument;

            return document;
        }

        internal ImageDataSource GetImageSource(Image image)
        {
            try
            {
                Guard.ThrowExceptionIfNull(image, "image");

                if (image.ImageSourceKey.StencilColor.HasValue)
                {
                    return this.contentManager.ResourceManager.GetImageSourceWithStencilColor(image.ImageSourceKey);
                }
                else
                {
                    return this.contentManager.ResourceManager.GetImageSource(image.ImageSourceKey);
                }
            }
            catch (Exception e)
            {
                this.OnDocumentException(e);

                return null;
            }
        }

        internal RadFixedPageInternal GetRadFixedPageFromPage(PageOld page)
        {
            Guard.ThrowExceptionIfNull(page, "page");

            return this.pagesMapping[page];
        }

        internal bool TryGetWidgetFromWidgetObject(WidgetOld source, out Widget widget)
        {
            return this.acroFormCacheManager.TryGetWidgetFromWidgetObject(source, out widget);
        }

        internal ContentCollection ParseOnlyTextRelatedContent(RadFixedPageInternal page)
        {
            ContentCollection collection = this.ParsePageContent(page, true);
            collection.Arrange(page.TransformMatrix);

            return collection;
        }

        internal ContentCollection ParsePageContent(RadFixedPageInternal page)
        {
            return this.ParsePageContent(page, false);
        }

        private ContentCollection ParsePageContent(RadFixedPageInternal page, bool skipNonTextRelatedContent)
        {
            ContentCollection collection = new ContentCollection();

            try
            {
                PageOld source = this.GetPageFromRadFixedPage(page);
                Container cropContainer = new Container();
                cropContainer.Clip = PathGeometry.CreateRectangle(page.CropBox);

                if (source.Contents != null)
                {
                    cropContainer.Content = this.GetPageContent(source, skipNonTextRelatedContent);
                }

                Matrix m = Matrix.Identity;
                m.Translate(-page.CropBox.X, -page.CropBox.Y);
                cropContainer.TransformMatrix = m;

                collection.AddChild(cropContainer);
            }
            catch (Exception e)
            {
                this.OnDocumentException(e);
            }

            return collection;
        }

        internal void LoadPageAnnotations(RadFixedPage page)
        {
            PageOld source = this.GetPageFromRadFixedPage(page.InternalRadFixedPage);
            this.CopyAnnotationsFromPageSource(source, page.Annotations);
            page.InternalRadFixedPage.Arrange();
        }

        internal Stream GetStream()
        {
            return this.stream;
        }

        internal void RecalculateWidgetAppearancesOld(string fieldName)
        {
            FormField field = this.document.PublicDocument.AcroForm.FormFields[fieldName];
            RadFixedDocument recalculatedDocument = this.BuildDocumentWithRecalculatedWidgetAppearances(Enumerable.Repeat<FormField>(field, 1));
            FormField regeneratedField = recalculatedDocument.AcroForm.FormFields[fieldName];
            Widget[] regeneratedWidgets = regeneratedField.Widgets.ToArray();
            int widgetIndex = 0;

            foreach (Widget widget in field.Widgets)
            {
                Widget regeneratedWidget = regeneratedWidgets[widgetIndex];
                widget.AppearanceOld = regeneratedWidget.AppearanceOld;
                widgetIndex++;
            }
        }

        private static bool ShouldRecalculateWidgetAppearance(Widget widget, bool shouldRecalculateAllAppearances)
        {
            return widget.AppearanceOld == null || (shouldRecalculateAllAppearances && widget.WidgetContentType == WidgetContentType.VariableContent);
        }

        private void Document_OnException(object sender, OnDocumentExceptionEventArgs e)
        {
            this.OnOnException(e);
        }

        private void OnOnException(OnDocumentExceptionEventArgs args)
        {
            if (this.OnException != null)
            {
                this.OnException(this, args);
            }
        }

        private void OnDocumentException(Exception e)
        {
            if (this.document != null)
            {
                OnDocumentExceptionEventArgs args = new OnDocumentExceptionEventArgs(e);
                if (args.Handle)
                {
                    this.document.OnOnException(args);
                    return;
                }
            }

            throw e;
        }

        private void CopyAnnotationsFromPageSource(PageOld source, AnnotationCollection annotations)
        {
            Guard.ThrowExceptionIfNull(source, "source");
            Guard.ThrowExceptionIfNull(annotations, "annotations");

            if (source.Annots != null)
            {
                for (int i = 0; i < source.Annots.Count; i++)
                {
                    AnnotationOld annot = source.Annots.GetElement<AnnotationOld>(i, Converters.AnnotConverter);
                    if (annot != null)
                    {
                        Annotation res = PdfElementToFixedElementTranslator.CreateAnnotation(annot, source, this);
                        if (res != null)
                        {
                            annotations.Add(res);
                        }
                    }
                }
            }
        }

        private ContentCollection GetPageContent(PageOld source, bool skipNonTextRelatedContent)
        {
            Guard.ThrowExceptionIfNull(source, "source");
            Guard.ThrowExceptionIfNull(source.Contents, "source.Contents");

            ContentCollection res = new ContentCollection();
            ContentStreamParser csp = new ContentStreamParser(this.contentManager, source.Contents.Data, source.Clip, source.Resources, source.Reference);
            csp.OnException += ContentStreamParser_OnException;
            IEnumerable<IContentElement> parseResult = skipNonTextRelatedContent ? csp.ParseOnlyTextRelatedContent() : csp.ParseContent();

            foreach (ContentElement ce in parseResult)
            {
                PdfElementToFixedElementTranslator.ToSLCoordinates(ce);
                res.Add(ce);
            }

            csp.OnException -= ContentStreamParser_OnException;

            return res;
        }

        private void ContentStreamParser_OnException(object sender, OnDocumentExceptionEventArgs e)
        {
            this.OnDocumentException(e.DocumentException);
        }

        private void LoadFormFields()
        {
            if (this.contentManager.DocumentCatalog.AcroForm != null && this.contentManager.DocumentCatalog.AcroForm.Fields != null)
            {
                try
                {
                    PdfImporter importer = new PdfImporter();
                    importer.Import(this.stream, this.acroFormCacheManager.FieldsAndWidgetsImportContext);

                    int fieldIndex = 0;
                    int widgetIndex = 0;

                    foreach (FormFieldNodeOld formFieldNode in this.contentManager.DocumentCatalog.AcroForm.Fields)
                    {
                        FormField field = this.acroFormCacheManager.FieldsAndWidgetsImportContext.GetField(fieldIndex);

                        if (field.FieldType == FormFieldType.Signature)
                        {
                            PdfElementToFixedElementTranslator.InitializeSignatureField((SignatureField)field, this.contentManager, formFieldNode);
                        }

                        this.contentManager.RegisterFormField(formFieldNode.Reference, field);

                        foreach (WidgetOld widgetObject in this.contentManager.GetChildWidgets(formFieldNode))
                        {
                            Widget widget = this.acroFormCacheManager.FieldsAndWidgetsImportContext.GetWidget(widgetIndex);
                            this.acroFormCacheManager.MapWidget(widgetObject, widget);
                            PdfElementToFixedElementTranslator.ApplyCommonAnnotationProperties(widgetObject, null, widget);

                            widgetIndex++;
                        }

                        fieldIndex++;
                    }
                }
                catch (Exception e)
                {
                    this.OnDocumentException(e);
                }
            }
        }

        private void RecalculateMissingWidgetAppearances()
        {
            bool shouldRecalculateAllAppearances = this.document.PublicDocument.AcroForm.ViewersShouldRecalculateWidgetAppearances;

            FormField[] fieldsToRegenerate = this.document.PublicDocument.AcroForm.FormFields.Where(field =>
            {
                return field.Widgets.Any(widget => ShouldRecalculateWidgetAppearance(widget, shouldRecalculateAllAppearances));
            }).ToArray();

            if (fieldsToRegenerate.Length > 0)
            {
                foreach (FormField field in fieldsToRegenerate)
                {
                    foreach (Widget widgetToRecalculate in field.Widgets.Where(widget => ShouldRecalculateWidgetAppearance(widget, shouldRecalculateAllAppearances)))
                    {
                        widgetToRecalculate.InvalidateAppearance();
                    }
                }

                RadFixedDocument recalculatedDocument = this.BuildDocumentWithRecalculatedWidgetAppearances(fieldsToRegenerate);

                foreach (FormField field in fieldsToRegenerate)
                {
                    FormField regeneratedField = recalculatedDocument.AcroForm.FormFields[field.Name];
                    Widget[] regeneratedWidgets = regeneratedField.Widgets.ToArray();
                    int widgetIndex = 0;

                    foreach (Widget widget in field.Widgets)
                    {
                        if (ShouldRecalculateWidgetAppearance(widget, shouldRecalculateAllAppearances))
                        {
                            Widget regeneratedWidget = regeneratedWidgets[widgetIndex];
                            widget.AppearanceOld = regeneratedWidget.AppearanceOld;
                        }

                        widgetIndex++;
                    }
                }

            }
        }

        private RadFixedDocument BuildDocumentWithRecalculatedWidgetAppearances(IEnumerable<FormField> fieldsToRegenerate)
        {
            RadFixedDocument initialDocument = new RadFixedDocument();

            foreach (FormField field in fieldsToRegenerate)
            {
                initialDocument.AcroForm.FormFields.Add(field);
            }

            using (MemoryStream pdfFile = new MemoryStream())
            {
                PdfExporter exporter = new PdfExporter();
                int startId = this.acroFormCacheManager.BiggestObjectId + 1;
                IRadFixedDocumentExportContext exportContext = new FieldsAndWidgetsPropertiesExportContext(startId, initialDocument, new PdfExportSettings());
                exporter.Export(exportContext, pdfFile);

                PdfFormatProvider pdfViewerProvider = new PdfFormatProvider(pdfFile);
                RadFixedDocument recalculatedDocument = pdfViewerProvider.Import();
                this.acroFormCacheManager.BiggestObjectId = pdfViewerProvider.acroFormCacheManager.BiggestObjectId;

                return recalculatedDocument;
            }
        }

        private void LoadPages()
        {
            this.document.Pages = new PagesCollection(this.GetPages(this.document));
        }

        private IEnumerable<RadFixedPageInternal> GetPages(RadFixedDocumentInternal doc)
        {
            Guard.ThrowExceptionIfNull(doc, "doc");

            RadFixedPageInternal page;
            int pageNo = 0;

            IEnumerable<PageOld> pages = this.contentManager.GetPages();
            if (pages == null)
            {
                return Enumerable.Empty<RadFixedPageInternal>();
            }

            int pageCount = Enumerable.Count(pages);
            RadFixedPageInternal[] fixedPages = new RadFixedPageInternal[pageCount];
            this.pages = new PageOld[pageCount];

            foreach (PageOld pageNode in pages)
            {
                pageNo++;
                page = PdfElementToFixedElementTranslator.CreateFixedPage(doc, pageNode);
                page.PublicPage.PageNumber = pageNo;
                this.pages[pageNo - 1] = pageNode;
                this.pagesMapping[pageNode] = page;
                fixedPages[pageNo - 1] = page;
            }

            return fixedPages;
        }

        private PageOld GetPageFromRadFixedPage(RadFixedPageInternal page)
        {
            Guard.ThrowExceptionIfNull(page, "page");

            return this.pages[page.PublicPage.PageNumber - 1];
        }

        private void LoadBookmarks()
        {
            if (this.contentManager.DocumentCatalog.Outlines != null)
            {
                OutlineItemOld outline = this.contentManager.DocumentCatalog.Outlines.First;

                while (outline != null)
                {
                    try
                    {
                        BookmarkItem bookmark = PdfElementToFixedElementTranslator.CreateBookmarkItemHierarchy(outline, this);
                        this.document.PublicDocument.Bookmarks.Add(bookmark);
                    }
                    catch (Exception e)
                    {
                        this.OnDocumentException(e);
                    }

                    outline = outline.Next;
                }
            }
        }

        private void LoadPageMode()
        {
            PdfNameOld pageModeName = this.contentManager.DocumentCatalog.PageMode;
            this.document.PublicDocument.PageMode = PdfElementToFixedElementTranslator.CreatePageMode(pageModeName);
        }
    }
}
