﻿using System.Collections.Generic;
using System.Diagnostics;
using Telerik.Windows.Documents.Core.PostScript.Data;
using Telerik.Windows.Documents.Utilities;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader.Readers;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.Text;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.Utils;
using Telerik.Windows.Documents.Fixed.Model.Internal.Classes;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader.Parsers
{
	internal class CMapStreamParser
	{
        private readonly CMapStream cmap;
        private readonly PdfContentManager contentManager;
        private readonly CMapStreamReader reader;

		private CMapOld mappings;

        public CMapStreamParser(PdfContentManager contentManager, CMapStream cmap)
		{
            Guard.ThrowExceptionIfNull(contentManager, "contentManager");
            Guard.ThrowExceptionIfNull(cmap, "cmap");

            this.contentManager = contentManager;
			this.cmap = cmap;

			this.reader = new CMapStreamReader(this.ContentManager.ReadData(this.cmap.Reference));
		}

		public CMapStreamReader Reader
		{
            get
            {
                return this.reader;
            }
		}

		public PdfContentManager ContentManager
		{
            get
            {
                return this.contentManager;
            }
		}

        public CMapOld ParseCMap()
        {
            this.mappings = new CMapOld();
            object[] pars;
            while (true)
            {
                pars = this.ReadObjectsToToken(Token.Operator);
                if (this.Reader.TokenType == Token.EndOfFile)
                {
                    break;
		}

                this.InvokeOperator(this.Reader.Result, pars);
            }

            if (this.cmap.UseCMap != null)
            {
                CMapStreamParser parser = new CMapStreamParser(this.ContentManager, this.cmap.UseCMap);
                this.mappings.UseCMap = parser.ParseCMap();
            }

            return this.mappings;
        }

		private object[] ReadObjectsToToken(Token endToken)
		{
			Stack<object> stack = new Stack<object>();

			Token tok;
			while ((tok = this.Reader.ReadToken()) != endToken)
			{
				switch (tok)
				{
					case Token.ArrayStart:
						stack.Push(this.ParseCurrentTokenToArray());
						break;
					case Token.Boolean:
						stack.Push(this.ParseCurrentTokenToBool());
						break;
					case Token.Real:
						stack.Push(this.ParseCurrentTokenToReal());
						break;
					case Token.DictionaryStart:
						stack.Push(this.ParseCurrentTokenToDictionary());
						break;
					case Token.String:
						stack.Push(this.ParseCurrentTokenToString());
						break;
					case Token.Integer:
						stack.Push(this.ParseCurrentTokenToInt());
						break;
					case Token.Name:
						stack.Push(this.ParseCurrentTokenToName());
						break;
					case Token.Null:
						stack.Push(null);
						break;
					case Token.EndOfFile:
						return new object[0];
					case Token.Unknown:
						break;
				}
			}

			if (stack.Count == 0)
			{
				return null;
			}

			int cnt = stack.Count - 1;
			object[] res = new object[cnt + 1];
			while (cnt >= 0)
			{
				res[cnt] = stack.Pop();
				cnt--;
			}
			
			return res;
		}

        private void BeginCodeSpaceRange(int count)
        {
            CharCodeOld begin, end;
            for (int i = 0; i < count; i++)
            {
                begin = this.ReadCharCode();
                end = this.ReadCharCode();
                this.mappings.AddCodeRange(begin, end);
            }
        }

		private void BeginBFChar(int count)
		{
            CharCodeOld cc;
            PdfStringOld unicode;
			for (int i = 0; i < count; i++)
			{
				cc = this.ReadCharCode();
                Token tok = this.Reader.ReadToken();
                if (tok == Token.String)
                {
                    unicode = this.ParseCurrentTokenToString();
                    this.mappings.AddCIDtoUnicodeMapping(cc, unicode.ToUnicodeString());
                }
                else if (tok == Token.Integer)
                {
                    this.mappings.AddCIDtoUnicodeMapping(cc, BytesHelperOld.GetUnicodeChar(this.ParseCurrentTokenToInt()).ToString());
                }
			}
			
            this.Reader.ReadToken();
		}

        private void BeginBFRange(int count)
        {
            CharCodeOld begin, end;
            PdfStringOld unicode;
            PdfArrayOld unicodes;
            Token tok;
            for (int i = 0; i < count; i++)
            {
                begin = this.ReadCharCode();
                end = this.ReadCharCode();
                tok = this.Reader.ReadToken();
                if (tok == Token.String)
                {
                    unicode = this.ParseCurrentTokenToString();
                    this.MapRange(begin, end, unicode);
                }
                else if (tok == Token.Integer)
                {
                    this.MapRange(begin, end, this.ParseCurrentTokenToInt());
                }
                else if (tok == Token.ArrayStart)
                {
                    unicodes = this.ParseCurrentTokenToArray();
                    this.MapRange(begin, end, unicodes);
                }
            }

            this.Reader.ReadToken();
        }

        private void BeginCIDChar(int count)
        {
            CharCodeOld cc;
            for (int i = 0; i < count; i++)
            {
                cc = this.ReadCharCode();
                Token tok = this.Reader.ReadToken();
                if (tok == Token.Integer)
                {
                    this.mappings.AddICCtoCIDMapping(cc, this.ParseCurrentTokenToInt());
                }
                else
                {
                    Debug.Assert(false, "Unsupported CID");
                }
            }
          
            this.Reader.ReadToken();
        }

        private void BeginCIDRange(int count)
        {
            CharCodeOld begin, end;
            Token tok;
            for (int i = 0; i < count; i++)
            {
                begin = this.ReadCharCode();
                end = this.ReadCharCode();
                tok = this.Reader.ReadToken();
                if (tok == Token.Integer)
                {
                    this.MapCIDRange(begin, end, this.ParseCurrentTokenToInt());
                }
                else
                {
                    Debug.Assert(false, "Not supported");
                }
            }

            this.Reader.ReadToken();
        }

        private void BeginNotDefChar(int count)
        {
            CharCodeOld cc;
            for (int i = 0; i < count; i++)
            {
                cc = this.ReadCharCode();
                Token tok = this.Reader.ReadToken();
                if (tok == Token.Integer)
                {
                    this.mappings.AddNotDefRange(cc, cc, this.ParseCurrentTokenToInt());
                }
                else
                {
                    Debug.Assert(false, "Unsupported CID");
                }
            }

            this.Reader.ReadToken(); // Reads endbfchar operator.
        }

        private void BeginNotDefRange(int count)
        {
            CharCodeOld cc, cc2;
            for (int i = 0; i < count; i++)
            {
                cc = this.ReadCharCode();
                cc2 = this.ReadCharCode();
                Token tok = this.Reader.ReadToken();
                if (tok == Token.Integer)
                {
                    this.mappings.AddNotDefRange(cc, cc2, this.ParseCurrentTokenToInt());
                }
                else
                {
                    Debug.Assert(false, "Unsupported CID");
                }
            }

            this.Reader.ReadToken(); // Reads endbfchar operator.
        }

        private void MapCIDRange(CharCodeOld begin, CharCodeOld end, int cid)
        {
            for (CharCodeOld i = begin; i < end; i++, cid++)
            {
                this.mappings.AddICCtoCIDMapping(i, cid);
            }

            this.mappings.AddICCtoCIDMapping(end, cid++);
        }

        private void MapRange(CharCodeOld begin, CharCodeOld end, int unicode)
        {
            for (CharCodeOld i = begin; i < end; i++, unicode++)
            {
                this.mappings.AddCIDtoUnicodeMapping(i, BytesHelperOld.GetUnicodeChar(unicode).ToString());
            }

            this.mappings.AddCIDtoUnicodeMapping(end, BytesHelperOld.GetUnicodeChar(unicode++).ToString());
        }

		private void MapRange(CharCodeOld begin, CharCodeOld end, PdfStringOld unicode)
		{
			for (CharCodeOld i = begin; i < end; i++)
			{
                this.mappings.AddCIDtoUnicodeMapping(i, unicode.ToUnicodeString());
                unicode.Increment();
			}

            this.mappings.AddCIDtoUnicodeMapping(end, unicode.ToUnicodeString());
		}

		private void MapRange(CharCodeOld begin, CharCodeOld end, PdfArrayOld unicodes)
		{
            int cnt = 0;
            PdfStringOld obj;

			for (CharCodeOld i = begin; i < end; i++)
			{
				obj = unicodes.GetElement<PdfStringOld>(cnt++);
                this.mappings.AddCIDtoUnicodeMapping(i, obj.ToUnicodeString());
			}

            obj = unicodes.GetElement<PdfStringOld>(cnt++);
            this.mappings.AddCIDtoUnicodeMapping(end, obj.ToUnicodeString());
		}

		private void InvokeOperator(string op, object[] pars)
		{
			int len = pars.Length;
			switch (op)
			{
                case CMapOperators.begincodespacerange:
                    this.BeginCodeSpaceRange((int)pars[len - 1]);
                    break;
				case CMapOperators.beginbfchar:
					this.BeginBFChar((int)pars[len - 1]);
					break;
				case CMapOperators.beginbfrange:
					this.BeginBFRange((int)pars[len - 1]);
					break;
                case CMapOperators.begincidchar:
                    this.BeginCIDChar((int)pars[len - 1]);
                    break;
                case CMapOperators.begincidrange:
                    this.BeginCIDRange((int)pars[len - 1]);
                    break;
                case CMapOperators.beginnotdefchar:
                    this.BeginNotDefChar((int)pars[len - 1]);
                    break;
                case CMapOperators.beginnotdefrange:
                    this.BeginNotDefRange((int)pars[len - 1]);
                    break;
			}
		}

		private bool ParseCurrentTokenToBool()
		{
			return this.Reader.Result == Keywords.True;
		}

		private int ParseCurrentTokenToInt()
		{
			int res;
			int.TryParse(this.Reader.Result, out res);
		
			return res;
		}

		private double ParseCurrentTokenToReal()
		{
			double r;
			double.TryParse(this.Reader.Result, System.Globalization.NumberStyles.Any, System.Globalization.CultureInfo.InvariantCulture.NumberFormat, out r);
			
			return r;
		}

		private PdfStringOld ParseCurrentTokenToString()
		{
            if (this.Reader.TokenType == Token.String)
            {
                return new PdfStringOld(this.ContentManager, this.Reader.BytesToken);
            }
            
            return null;
		}

		private PdfNameOld ParseCurrentTokenToName()
		{
			if (this.Reader.TokenType == Token.Name)
			{
				return new PdfNameOld(this.ContentManager, this.Reader.Result);
			}
			
			return null;
		}

		private PdfArrayOld ParseCurrentTokenToArray()
		{
			if (this.Reader.TokenType == Token.ArrayStart)
			{
				PdfArrayOld arr = new PdfArrayOld(this.ContentManager);
				object[] content = this.ReadObjectsToToken(Token.ArrayEnd);
                arr.Load(content);
				return arr;
			}
			
			return null;
		}

		private PdfDictionaryOld ParseCurrentTokenToDictionary()
		{
			if (this.Reader.TokenType == Token.DictionaryStart)
			{
				PdfDictionaryOld dict = new PdfDictionaryOld(this.ContentManager);
				object[] content = this.ReadObjectsToToken(Token.DictionaryEnd);
                dict.Load(content);
				return dict;
			}
			
			return null;
		}

        private CharCodeOld ReadCharCode()
        {
            this.Reader.ReadToken();
           
            return new CharCodeOld(this.Reader.BytesToken);
        }
	}
}
