﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Core.Imaging;
using Telerik.Windows.Documents.Core.PostScript.Data;
using Telerik.Windows.Documents.Core.Shapes;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Utilities;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.Enums;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.Fonts;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.XObjects;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.Filters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader.Readers;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.Text;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.Utils;
using Telerik.Windows.Documents.Fixed.Model.Internal;
using Telerik.Windows.Documents.Fixed.Model.Internal.Classes;
using ColorSpaceOld = Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.ColorSpaces.ColorSpaceOld;
using Telerik.Windows.Documents.Fixed.Exceptions;
using Rect = System.Windows.Rect;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.PdfConverters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.ColorSpaces;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader.Parsers
{
    internal class ContentStreamParser
    {
        private readonly PdfContentManager contentManager;
        private readonly ContentStreamReader reader;
        private readonly PdfContext context;
        private readonly PdfResourceOld resources;
        private readonly IndirectReferenceOld contentOwnerReference;
        private int inlineImageCount = 0;

        public ContentStreamParser(PdfContentManager contentManager, byte[] data, Rect clip, PdfResourceOld resources, Brush stencilBrush, IndirectReferenceOld contentOwnerReference)
            : this(contentManager, data, clip, resources, contentOwnerReference)
        {
            Guard.ThrowExceptionIfNull(stencilBrush, "stencilBrush");

            this.context.GraphicsState.Brush = stencilBrush;
        }

        public ContentStreamParser(PdfContentManager contentManager, byte[] data, Rect clip, PdfResourceOld resources, IndirectReferenceOld contentOwnerReference)
        {
            Guard.ThrowExceptionIfNull(contentManager, "contentManager");
            Guard.ThrowExceptionIfNull(data, "data");
#if DEBUG
            PdfContentExtractor.ExportContentStream(contentOwnerReference, data);
#endif   
            this.contentManager = contentManager;
            this.resources = resources;
            this.reader = new ContentStreamReader(data);
            this.context = new PdfContext(this.contentManager, clip);
            this.contentOwnerReference = contentOwnerReference;
        }

        public IEnumerable<ContentElement> ParseContent()
        {
            return this.ParseContent(false);
        }

        public IEnumerable<ContentElement> ParseOnlyTextRelatedContent()
        {
            return this.ParseContent(true);
        }

        private IEnumerable<ContentElement> ParseContent(bool skipNonTextRelatedContent)
        {
            List<ContentElement> contents = new List<ContentElement>();
            contents.Add(this.context.GraphicsState.ClippingContainer);

            object[] pars;
            while (true)
            {
                pars = this.ReadObjectsToToken(Token.Operator);
                if (this.reader.TokenType == Token.EndOfFile)
                {
                    break;
                }

                this.InvokeOperator(this.reader.Result, pars, skipNonTextRelatedContent);
            }

            return contents;
        }

        internal event EventHandler<OnDocumentExceptionEventArgs> OnException;

        private object[] ReadObjectsToToken(Token endToken)
        {
            return this.ReadObjectsToToken(endToken, false);
        }

        private object[] ReadObjectsToToken(Token endToken, bool replaceAbbreviations)
        {
            Stack<object> stack = new Stack<object>();

            Token tok;
            while ((tok = this.reader.ReadToken()) != endToken)
            {
                switch (tok)
                {
                    case Token.ArrayStart:
                        stack.Push(this.ParseCurrentTokenToArray(replaceAbbreviations));
                        break;
                    case Token.Boolean:
                        stack.Push(this.ParseCurrentTokenToBool());
                        break;
                    case Token.Real:
                        stack.Push(this.ParseCurrentTokenToReal());
                        break;
                    case Token.DictionaryStart:
                        stack.Push(this.ParseCurrentTokenToDictionary(replaceAbbreviations));
                        break;
                    case Token.String:
                        stack.Push(this.ParseCurrentTokenToString());
                        break;
                    case Token.Integer:
                        stack.Push(this.ParseCurrentTokenToInt());
                        break;
                    case Token.Name:
                        stack.Push(this.ParseCurrentTokenToName(replaceAbbreviations));
                        break;
                    case Token.Null:
                        stack.Push(null);
                        break;
                    case Token.EndOfFile:
                        return new object[0];
                    case Token.Unknown:
                        break;
                }
            }

            if (stack.Count == 0)
            {
                return new object[0];
            }

            int cnt = stack.Count - 1;
            object[] res = new object[cnt + 1];
            while (cnt >= 0)
            {
                res[cnt] = stack.Pop();
                cnt--;
            }
            return res;
        }

        private void OnOnException(OnDocumentExceptionEventArgs args)
        {
            if (this.OnException != null)
            {
                this.OnException(this, args);
            }
        }

        private FontBaseOld GetFont(PdfNameOld name)
        {
            Guard.ThrowExceptionIfNull(name, "name");

            return this.resources.GetFont(name);
        }

        private ShadingOld GetShanding(PdfNameOld name)
        {
            Guard.ThrowExceptionIfNull(name, "name");

            return this.resources.GetShading(name);
        }

        private XObject GetXObject(PdfNameOld name)
        {
            Guard.ThrowExceptionIfNull(name, "name");

            return this.resources.GetXObject(name);
        }

        private ColorSpaceOld GetColorSpace(PdfNameOld name)
        {
            Guard.ThrowExceptionIfNull(name, "name");

            if (ColorSpaceOld.IsColorSpace(name.Value))
            {
                return ColorSpaceOld.CreateColorSpace(this.contentManager, name.Value);
            }

            return this.resources.GetColorSpace(name);
        }

        private ExtGStateOld GetExtGState(PdfNameOld name)
        {
            Guard.ThrowExceptionIfNull(name, "name");

            return this.resources.GetExtGState(name);
        }

        private void AddContentElement(ContentElement element)
        {
            if (element != null)
            {
                this.context.GraphicsState.ClippingContainer.Content.AddChild(element);
            }
        }

        private void ReadInlineImage(bool skipImage)
        {
            int imageInlineId = ++inlineImageCount;

            ImageResourceKey resourceKey = this.CreateInlineImageResourceKey(this.contentOwnerReference, imageInlineId);

            if(skipImage)
            {
                this.SkipInlineImage(resourceKey);
            }
            else
            {
                Image image = this.ReadAndRegisterInlineImageSource(resourceKey);
                this.AddContentElement(image);
            }
        }

        private ImageResourceKey CreateInlineImageResourceKey(IndirectReferenceOld reference, int id)
        {
            Guard.ThrowExceptionIfNull(reference, "reference");

            ImageResourceKey key = new ImageResourceKey();
            key.Id = id;
            key.Reference = reference;
            key.Type = ResourceType.Local;

            return key;
        }

        private void SkipInlineImage(ImageResourceKey resourceKey)
        {
            InlineImageInfo imageInfo;
            if (this.context.ResourcesManager.TryGetInlineImageInfo(resourceKey, out imageInfo))
            {
                this.reader.Seek(imageInfo.EndOfImagePosition, System.IO.SeekOrigin.Begin);
            }
            else
            {
                byte[] data;
                PdfDictionaryOld dictionary;
                this.ReadRawInlineImageData(out dictionary, out data);
            }
        }

        private Image ReadAndRegisterInlineImageSource(ImageResourceKey resourceKey)
        {
            Guard.ThrowExceptionIfNull(resourceKey, "resourceKey");
            
            Image image = new Image();
            image.ImageSourceKey = resourceKey;
            image.TransformMatrix = this.context.GraphicsState.Ctm;

            InlineImageInfo imageInfo;
            if (this.context.ResourcesManager.TryGetInlineImageInfo(resourceKey, out imageInfo))
            {
                this.ApplyInlineImageStencilColor(resourceKey, imageInfo.XImage);
                this.reader.Seek(imageInfo.EndOfImagePosition, System.IO.SeekOrigin.Begin);
            }
            else
            {
                byte[] rawData;
                PdfDictionaryOld dictionary;
                this.ReadRawInlineImageData(out dictionary, out rawData);

                XImage xImage = new XImage(this.contentManager);
                xImage.Load(dictionary);
                this.ApplyInlineImageStencilColor(resourceKey, xImage);
                imageInfo = new InlineImageInfo(xImage, dictionary, rawData, this.reader.Position);
                this.contentManager.ResourceManager.RegisterInlineImage(imageInfo, resourceKey);
            }

            return image;
        }

        private void ReadRawInlineImageData(out PdfDictionaryOld imageDictionary, out byte[] rawImageData)
        {
            object[] res = this.ReadObjectsToToken(Token.Operator, true);
            imageDictionary = new PdfDictionaryOld(this.contentManager);
            imageDictionary.Load(res);

            if (imageDictionary.ContainsKey(PdfNames.ColorSpace))
            {
                PdfNameOld pdfName = imageDictionary.GetElement<PdfNameOld>(PdfNames.ColorSpace, Converters.PdfNameConverter);
                if (pdfName != null)
                {
                    imageDictionary[PdfNames.ColorSpace] = this.GetColorSpace(pdfName);
                }
            }

            this.reader.SkipWhiteSpaces();
            long position = this.reader.Position;
            StreamPart part = new StreamPart(this.reader.Stream, position, Keywords.EndOfInlineImage);
            rawImageData = this.reader.ReadRawData(position, part);
            this.reader.ReadToken();
        }

        private void ApplyInlineImageStencilColor(ImageResourceKey resourceKey, XImage xImage)
        {
            if ((xImage.ImageMask != null && xImage.ImageMask.Value))
            {
                resourceKey.StencilColor = this.context.StencilColor;
            }
        }

        private void InvokeGeneralGraphicsStateOperator(string op, object[] pars)
        {
            Guard.ThrowExceptionIfNullOrEmpty(op, "op");

            int len = pars != null ? pars.Length : 0;
            switch (op)
            {
                case GeneralGraphicsStateOperators.w:
                    double w;
                    Helper.UnboxDouble(pars[len - 1], out w);
                    this.context.SetLineWidth(w);
                    break;
                case GeneralGraphicsStateOperators.J:
                    this.context.SetLineCap((int) pars[len - 1]);
                    break;
                case GeneralGraphicsStateOperators.j:
                    this.context.SetLineJoin((int) pars[len - 1]);
                    break;
                case GeneralGraphicsStateOperators.M:
                    double m;
                    Helper.UnboxDouble(pars[len - 1], out m);
                    this.context.SetMitterLimit(m);
                    break;
                case GeneralGraphicsStateOperators.d:
                    double d;
                    Helper.UnboxDouble(pars[len - 1], out d);
                    this.context.SetDashPattern((PdfArrayOld) pars[len - 2], d);
                    break;
                case GeneralGraphicsStateOperators.ri:
                    this.context.SetRenderingIntent(pars[len - 1] as string);
                    break;
                case GeneralGraphicsStateOperators.i:
                    double i;
                    Helper.UnboxDouble(pars[len - 1], out i);
                    this.context.SetFlatness(i);
                    break;
                case GeneralGraphicsStateOperators.gs:
                    this.context.SetGraphicState(this.GetExtGState((PdfNameOld) pars[len - 1]));
                    break;
                case GeneralGraphicsStateOperators.q:
                    this.context.SaveGraphicState();
                    break;
                case GeneralGraphicsStateOperators.Q:
                    this.context.RestoreGraphicState();
                    break;
                case GeneralGraphicsStateOperators.cm:
                    double m11, m12, m21, m22, offsetX, offsetY;
                    Helper.UnboxDouble(pars[len - 6], out m11);
                    Helper.UnboxDouble(pars[len - 5], out m12);
                    Helper.UnboxDouble(pars[len - 4], out m21);
                    Helper.UnboxDouble(pars[len - 3], out m22);
                    Helper.UnboxDouble(pars[len - 2], out offsetX);
                    Helper.UnboxDouble(pars[len - 1], out offsetY);
                    this.context.SetMatrix(m11, m12, m21, m22, offsetX, offsetY);
                    break;
            }
        }

        private void InvokeColorOperator(string op, object[] pars)
        {
            Guard.ThrowExceptionIfNullOrEmpty(op, "op");

            int len = pars != null ? pars.Length : 0;
            double a, b, c, d;
            switch (op)
            {
                case ColorOperators.CS:
                    this.context.SetStrokeColorSpace(this.GetColorSpace(pars[len - 1] as PdfNameOld));
                    break;
                case ColorOperators.cs:
                    this.context.SetColorSpace(this.GetColorSpace(pars[len - 1] as PdfNameOld));
                    break;
                case ColorOperators.SCN:
                case ColorOperators.SC:
                    this.context.SetStrokeColor(this.context.GraphicsState.StrokeColorSpace.GetBrush(this.resources, pars));
                    break;
                case ColorOperators.sc:
                case ColorOperators.scn:
                    this.context.SetColor(this.context.GraphicsState.ColorSpace.GetBrush(this.resources, pars));
                    break;
                case ColorOperators.g:
                    Helper.UnboxDouble(pars[len - 1], out a);
                    this.context.SetGrayColor(a);
                    break;
                case ColorOperators.G:
                    Helper.UnboxDouble(pars[len - 1], out a);
                    this.context.SetGrayStrokeColor(a);
                    break;
                case ColorOperators.RG:
                    Helper.UnboxDouble(pars[len - 3], out a);
                    Helper.UnboxDouble(pars[len - 2], out b);
                    Helper.UnboxDouble(pars[len - 1], out c);
                    this.context.SetRgbStrokeColor(Color.FromArgb(1, a, b, c));
                    break;
                case ColorOperators.rg:
                    Helper.UnboxDouble(pars[len - 3], out a);
                    Helper.UnboxDouble(pars[len - 2], out b);
                    Helper.UnboxDouble(pars[len - 1], out c);
                    this.context.SetRgbColor(Color.FromArgb(1, a, b, c));
                    break;
                case ColorOperators.K:
                    Helper.UnboxDouble(pars[len - 4], out a);
                    Helper.UnboxDouble(pars[len - 3], out b);
                    Helper.UnboxDouble(pars[len - 2], out c);
                    Helper.UnboxDouble(pars[len - 1], out d);
                    this.context.SetCmykStrokeColor(Color.FromCmyk(a, b, c, d));
                    break;
                case ColorOperators.k:
                    Helper.UnboxDouble(pars[len - 4], out a);
                    Helper.UnboxDouble(pars[len - 3], out b);
                    Helper.UnboxDouble(pars[len - 2], out c);
                    Helper.UnboxDouble(pars[len - 1], out d);
                    this.context.SetCmykColor(Color.FromCmyk(a, b, c, d));
                    break;
            }
        }

        private void InvokeTextOperator(string op, object[] pars)
        {
            Guard.ThrowExceptionIfNullOrEmpty(op, "op");

            int len = pars != null ? pars.Length : 0;
            double a, b, c, d, e, f;
            switch (op)
            {
                case TextOperators.BT:
                    this.context.BeginText();
                    break;
                case TextOperators.ET:
                    this.context.EndText();
                    break;
                case TextOperators.Tw:
                    Helper.UnboxDouble(pars[len - 1], out a);
                    this.context.SetWordSpacing(a);
                    break;
                case TextOperators.Tz:
                    Helper.UnboxDouble(pars[len - 1], out a);
                    this.context.SetHorizontalScaling(a);
                    break;
                case TextOperators.TL:
                    Helper.UnboxDouble(pars[len - 1], out a);
                    this.context.SetTextLeading(a);
                    break;
                case TextOperators.Tf:
                    Helper.UnboxDouble(pars[len - 1], out a);
                    FontBaseOld font = this.GetFont(pars[len - 2] as PdfNameOld);
                    this.context.SetFont(font, a);
                    break;
                case TextOperators.Tr:
                    this.context.SetRenderingMode((RenderingMode) pars[len - 1]);
                    break;
                case TextOperators.Ts:
                    Helper.UnboxDouble(pars[len - 1], out a);
                    this.context.SetRise(a);
                    break;
                case TextOperators.Td:
                    Helper.UnboxDouble(pars[len - 2], out a);
                    Helper.UnboxDouble(pars[len - 1], out b);
                    this.context.MoveToNextLine(a, b);
                    break;
                case TextOperators.TD:
                    Helper.UnboxDouble(pars[len - 2], out a);
                    Helper.UnboxDouble(pars[len - 1], out b);
                    this.context.MoveToNextLineWithTextLeading(a, b);
                    break;
                case TextOperators.Tm:
                    Helper.UnboxDouble(pars[len - 6], out a);
                    Helper.UnboxDouble(pars[len - 5], out b);
                    Helper.UnboxDouble(pars[len - 4], out c);
                    Helper.UnboxDouble(pars[len - 3], out d);
                    Helper.UnboxDouble(pars[len - 2], out e);
                    Helper.UnboxDouble(pars[len - 1], out f);
                    this.context.SetTextMatrix(a, b, c, d, e, f);
                    break;
                case TextOperators.Tj:
                    this.AddContentElement(this.context.DrawText((PdfStringOld) pars[len - 1]));
                    break;
                case TextOperators.TStar:
                    this.context.MoveToNextLineWithCurrentTextLeading();
                    break;
                case TextOperators.Ap:
                    this.AddContentElement(this.context.MoveToNextLineAndDrawText((PdfStringOld) pars[len - 1]));
                    break;
                case TextOperators.Qu:
                    Helper.UnboxDouble(pars[len - 3], out a);
                    Helper.UnboxDouble(pars[len - 2], out b);
                    this.AddContentElement(this.context.MoveToNextLineAndDrawText((PdfStringOld) pars[len - 1], a, b));
                    break;
                case TextOperators.TJ:
                    this.AddContentElement(this.context.DrawText((PdfArrayOld) pars[len - 1]));
                    break;
                case TextOperators.Tc:
                    Helper.UnboxDouble(pars[len - 1], out a);
                    this.context.SetCharSpacing(a);
                    break;
            }
        }

        private void InvokePathOperator(string op, object[] pars)
        {
            Guard.ThrowExceptionIfNullOrEmpty(op, "op");

            int len = pars != null ? pars.Length : 0;
            double a, b, c, d, e, f;
            switch (op)
            {
                case PathOperators.m:
                    Helper.UnboxDouble(pars[len - 2], out a);
                    Helper.UnboxDouble(pars[len - 1], out b);
                    this.context.MoveTo(a, b);
                    break;
                case PathOperators.l:
                    Helper.UnboxDouble(pars[len - 2], out a);
                    Helper.UnboxDouble(pars[len - 1], out b);
                    this.context.LineTo(a, b);
                    break;
                case PathOperators.c:
                    Helper.UnboxDouble(pars[len - 6], out a);
                    Helper.UnboxDouble(pars[len - 5], out b);
                    Helper.UnboxDouble(pars[len - 4], out c);
                    Helper.UnboxDouble(pars[len - 3], out d);
                    Helper.UnboxDouble(pars[len - 2], out e);
                    Helper.UnboxDouble(pars[len - 1], out f);
                    this.context.CurveTo(a, b, c, d, e, f);
                    break;
                case PathOperators.v:
                    Helper.UnboxDouble(pars[len - 4], out a);
                    Helper.UnboxDouble(pars[len - 3], out b);
                    Helper.UnboxDouble(pars[len - 2], out c);
                    Helper.UnboxDouble(pars[len - 1], out d);
                    this.context.HCurveTo(a, b, c, d);
                    break;
                case PathOperators.y:
                    Helper.UnboxDouble(pars[len - 4], out a);
                    Helper.UnboxDouble(pars[len - 3], out b);
                    Helper.UnboxDouble(pars[len - 2], out c);
                    Helper.UnboxDouble(pars[len - 1], out d);
                    this.context.VCurveTo(a, b, c, d);
                    break;
                case PathOperators.h:
                    this.context.Close();
                    break;
                case PathOperators.re:
                    Helper.UnboxDouble(pars[len - 4], out a);
                    Helper.UnboxDouble(pars[len - 3], out b);
                    Helper.UnboxDouble(pars[len - 2], out c);
                    Helper.UnboxDouble(pars[len - 1], out d);
                    this.context.Rectangle(a, b, c, d);
                    break;
                case PathOperators.S:
                    this.AddContentElement(this.context.Stroke());
                    break;
                case PathOperators.s:
                    this.AddContentElement(this.context.CloseAndStroke());
                    break;
                case PathOperators.f:
                    this.AddContentElement(this.context.Fill());
                    break;
                case PathOperators.F:
                    this.AddContentElement(this.context.Fill());
                    break;
                case PathOperators.fStar:
                    this.AddContentElement(this.context.Fill(FillRule.EvenOdd));
                    break;
                case PathOperators.B:
                    this.AddContentElement(this.context.FillAndStroke());
                    break;
                case PathOperators.BStar:
                    this.AddContentElement(this.context.FillAndStroke(FillRule.EvenOdd));
                    break;
                case PathOperators.b:
                    this.AddContentElement(this.context.CloseFillAndStroke());
                    break;
                case PathOperators.bStar:
                    this.AddContentElement(this.context.CloseFillAndStroke(FillRule.EvenOdd));
                    break;
                case PathOperators.n:
                    this.context.EndPath();
                    break;
                case PathOperators.W:
                    this.context.SetClippingPath();
                    break;
                case PathOperators.WStar:
                    this.context.SetClippingPath(FillRule.EvenOdd);
                    break;
                case PathOperators.Sh:
                    this.AddContentElement(this.context.ApplyShading(this.GetShanding((PdfNameOld) pars[len - 1])));
                    break;
            }
        }

        private void InvokeXObjectOperator(string op, object[] pars, bool skipNonTextRelatedContent)
        {
            Guard.ThrowExceptionIfNullOrEmpty(op, "op");
            Guard.ThrowExceptionIfNull(pars, "pars");

            PdfNameOld name = pars[pars.Length - 1] as PdfNameOld;
            XObject obj = this.GetXObject(name);

            XImage image = obj as XImage;
            if (image != null)
            {
                if(!skipNonTextRelatedContent)
                {
                    this.AddContentElement(this.context.CreateXImage(image));
                }
            }
            else
            {
                XForm form = obj as XForm;
                if (form != null)
                {
                    this.AddContentElement(this.context.CreateXForm(this.resources, form, skipNonTextRelatedContent));
                }
                else
                {
                    throw new NotSupportedXObjectTypeException(obj.GetType().ToString());
                }
            }
        }

        private void InvokeImageOperator(string op, bool skipImage)
        {
            switch (op)
            {
                case ImageOperators.BI:
                    this.ReadInlineImage(skipImage);
                    break;
                case ImageOperators.EI:
                    break;
            }
        }

        private void InvokeOperator(string op, object[] pars, bool skipNonTextRelatedOperators)
        {
            try
            {
                if (TextOperators.IsOperator(op))
                {
                    this.InvokeTextOperator(op, pars);
                }
                else if (GeneralGraphicsStateOperators.IsOperator(op))
                {
                    this.InvokeGeneralGraphicsStateOperator(op, pars);
                }
                else if (XObjectOperators.IsOperator(op))
                {
                    this.InvokeXObjectOperator(op, pars, skipNonTextRelatedOperators);
                }
                else if (ImageOperators.IsOperator(op))
                {
                    this.InvokeImageOperator(op, skipNonTextRelatedOperators);
                }
                else if (!skipNonTextRelatedOperators)
                {
                    if (ColorOperators.IsOperator(op))
                    {
                        this.InvokeColorOperator(op, pars);
                    }
                    else if (PathOperators.IsOperator(op))
                    {
                        this.InvokePathOperator(op, pars);
                    }
                }
            }
            catch (Exception e)
            {
                this.OnOnException(new OnDocumentExceptionEventArgs(e));
            }
        }

        private bool ParseCurrentTokenToBool()
        {
            return this.reader.Result == Keywords.True;
        }

        private int ParseCurrentTokenToInt()
        {
            int res;
            int.TryParse(this.reader.Result, out res);

            return res;
        }

        private double ParseCurrentTokenToReal()
        {
            double r;
            double.TryParse(this.reader.Result, System.Globalization.NumberStyles.Any, System.Globalization.CultureInfo.InvariantCulture.NumberFormat, out r);

            return r;
        }

        private PdfStringOld ParseCurrentTokenToString()
        {
            if (this.reader.TokenType == Token.String)
            {
                return new PdfStringOld(this.contentManager, this.reader.BytesToken);
            }

            return null;
        }

        private PdfNameOld ParseCurrentTokenToName(bool replaceAbbreviation)
        {
            if (this.reader.TokenType == Token.Name)
            {
                if (!replaceAbbreviation)
                {
                    return new PdfNameOld(this.contentManager, this.reader.Result);
                }
                else
                {
                    return new PdfNameOld(this.contentManager, FormatProviders.Pdf.Model.PdfNames.GetNameFromAbbreviation(this.reader.Result));
                }
            }

            return null;
        }

        private PdfArrayOld ParseCurrentTokenToArray(bool replaceAbbreviations)
        {
            if (this.reader.TokenType == Token.ArrayStart)
            {
                PdfArrayOld arr = new PdfArrayOld(this.contentManager);
                object[] content = this.ReadObjectsToToken(Token.ArrayEnd, replaceAbbreviations);
                arr.Load(content);

                return arr;
            }

            return null;
        }

        private PdfDictionaryOld ParseCurrentTokenToDictionary(bool replaceAbbreviations)
        {
            if (this.reader.TokenType == Token.DictionaryStart)
            {
                PdfDictionaryOld dict = new PdfDictionaryOld(this.contentManager);
                object[] content = this.ReadObjectsToToken(Token.DictionaryEnd, replaceAbbreviations);
                dict.Load(content);

                return dict;
            }

            return null;
        }
    }
}
