﻿#if DEBUG
using System;
using System.Collections.Generic;
using System.IO;
using Telerik.Windows.Documents.Core.Imaging;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.XObjects;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Data;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Primitives;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader.Parsers
{
    public static class PdfContentExtractor
    {
        private static readonly HashSet<IndirectReferenceOld> contentStreamOwners;
        private static int jpegsCount;
        private static int missingReferenceCount;

        static PdfContentExtractor()
        {
            PdfContentExtractor.Directory = @"C:\Temp\";
            PdfContentExtractor.ShouldExtract = false;
            PdfContentExtractor.contentStreamOwners = new HashSet<IndirectReferenceOld>();
            ResetNumberingsAndCaches();
        }

        /// <summary>
        /// Gets or sets whether the content should be extracted. Default value is false.
        /// </summary>
        public static bool ShouldExtract
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets the parent directory of the extraction directory. Default value is C:\Temp\.
        /// </summary>
        public static string Directory
        {
            get;
            set;
        }

        private static bool CancelExtraction
        {
            get
            {
#if WPF
                return !ShouldExtract;
#else
                return true;
#endif
            }
        }

        private static string ExtractionDir
        {
            get
            {
                return System.IO.Path.Combine(Directory, "PDF Extracted Content");
            }
        }

        private static string ContentStreamsDir
        {
            get
            {
                return System.IO.Path.Combine(ExtractionDir, "ContentStreams");
            }
        }

        private static string JpegsDir
        {
            get
            {
                return System.IO.Path.Combine(ExtractionDir, "Jpeg Images");
            }
        }

        internal static string ImagesDir
        {
            get
            {
                return System.IO.Path.Combine(ExtractionDir, "Images");
            }
        }

        internal static string SMaskDir
        {
            get
            {
                return System.IO.Path.Combine(ExtractionDir, "SMasks");
            }
        }

        internal static string MasksDir
        {
            get
            {
                return System.IO.Path.Combine(ExtractionDir, "Masks");
            }
        }

        internal static string MaskResultImages
        {
            get
            {
                return System.IO.Path.Combine(ExtractionDir, "Mask Result Images");
            }
        }

        private static string[] ExtractionDirectories
        {
            get
            {
                return new string[] { Directory, ExtractionDir, ContentStreamsDir, JpegsDir, ImagesDir, SMaskDir, MasksDir, MaskResultImages };
            }
        }

        internal static void ExportContentStream(IndirectReferenceOld contentOwnerReference, byte[] data)
        {
            if (PdfContentExtractor.contentStreamOwners.Add(contentOwnerReference))
            {
                ExportBytes(data, ContentStreamsDir, string.Format(@"ContentStream {0}.txt", contentOwnerReference));
            }
        }

        internal static void ExportJpegImage(byte[] data, string fileNamePrefix)
        {
            ExportBytes(data, JpegsDir, string.Format(@"{0}{1}.jpg", fileNamePrefix, ++jpegsCount));
        }

        internal static void ExportToFile(IImageDescriptor image, string filePath, bool shouldApplyMask)
        {
            PixelContainer pixels = image.ColorSpace.GetPixels(image, shouldApplyMask);

            ExportToFile(new SizeI(image.Width, image.Height), filePath, (x, y) =>
            {
                return pixels.GetColorFromIndex(y * image.Width + x);
            });
        }

        internal static void ExportToFile(DecodedImageDataSource imageSource, string filePath)
        {
            ExportToFile(new SizeI(imageSource.Width, imageSource.Height), filePath, (x, y) =>
            {
                return imageSource.Data.GetColorFromIndex(y * imageSource.Width + x);
            });
        }

        internal static string GetFileName(XImage image, string directory, string fileStart)
        {
            return System.IO.Path.Combine(directory, string.Format("{0} {1}.png", fileStart,
                image.Reference == null ? string.Format("missingReference{0}", missingReferenceCount++) : image.Reference.ToString()));
        }

        internal static void ExportToFile(SizeI imageSize, string filePath, Func<int, int, Color> getPixel)
        {
#if WPF
            if (CancelExtraction) return;

            EnsureExtractionDirectories();

            int width = imageSize.Width;
            int height = imageSize.Height;

            var pixelFormat = System.Windows.Media.PixelFormats.Bgra32;
            var stride = width * 4; // bytes per row

            byte[] pixelData = new byte[height * stride];

            for (int y = 0; y < height; ++y)
            {
                for (int x = 0; x < width; ++x)
                {
                    Color color = getPixel(x, y);

                    var index = (y * stride) + (x * 4);

                    pixelData[index] = color.B;
                    pixelData[index + 1] = color.G;
                    pixelData[index + 2] = color.R;
                    pixelData[index + 3] = color.A;
                }
            }

            var bitmap = System.Windows.Media.Imaging.BitmapSource.Create(width, height, 96, 96, pixelFormat, null, pixelData, stride);

            using (Stream file = File.OpenWrite(filePath))
            {
                System.Windows.Media.Imaging.PngBitmapEncoder encoder = new System.Windows.Media.Imaging.PngBitmapEncoder();
                encoder.Frames.Add(System.Windows.Media.Imaging.BitmapFrame.Create(bitmap));
                encoder.Save(file);
            }
#endif
        }

        internal static void ClearExtractedContent()
        {
            if (CancelExtraction) return;

            try
            {
                ResetNumberingsAndCaches();

                if (System.IO.Directory.Exists(ExtractionDir))
                {
                    System.IO.Directory.Delete(ExtractionDir, true);
                }
            }
            catch
            {
                System.Windows.MessageBox.Show("Could not delete old extraction directory! Please close the directory from FileExplorer and try again!");
            }
            finally
            {
                EnsureExtractionDirectories();
#if WPF
                System.Diagnostics.Process.Start(ExtractionDir);
#endif
            }
        }

        private static void ResetNumberingsAndCaches()
        {
            PdfContentExtractor.jpegsCount = 0;
            PdfContentExtractor.missingReferenceCount = 0;
            PdfContentExtractor.contentStreamOwners.Clear();
        }

        private static void ExportBytes(byte[] data, string directory, string fileName)
        {
            if (CancelExtraction) return;

            EnsureExtractionDirectories();
            File.WriteAllBytes(System.IO.Path.Combine(directory, fileName), data);
        }

        private static void EnsureExtractionDirectories()
        {
            if (CancelExtraction) return;

            foreach (string directory in ExtractionDirectories)
            {
                if (!System.IO.Directory.Exists(directory))
                {
                    System.IO.Directory.CreateDirectory(directory);
                }
            }
        }
    }
}
#endif
