﻿using System;
using System.Collections.Generic;
using System.IO;
using Telerik.Windows.Documents.Core.PostScript.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Data;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader.Parsers
{
    internal class PdfParser : PdfParserBase
    {
        public PdfParser(PdfContentManager contentManager, Stream stream)
            : base(contentManager, stream)
        {
        }

        public bool TryReadIndirectObject(long offset, bool isEncrypted, out IndirectObjectOld indirectObject)
        {
            if (offset == -1)
            {
                indirectObject = null;
                return false;
            }

            indirectObject = new IndirectObjectOld();
            long lastPosition = this.Reader.Position;
            this.Reader.Seek(offset, SeekOrigin.Begin);
            indirectObject.ObjectNumber = this.ReadInt();
            indirectObject.GenerationNumber = this.ReadInt();
            this.Reader.ReadToken(); // Token must be IndirectObjectStart.
            object[] res = isEncrypted ? this.ReadObjectsToToken(indirectObject, Token.IndirectObjectEnd) : this.ReadObjectsToToken(Token.IndirectObjectEnd);
            indirectObject.Value = res[0];
            this.Reader.Seek(lastPosition, SeekOrigin.Begin);

            return true;
        }

        public TrailerOld InitializeCrossReferences()
        {
            this.Reader.SeekToStartXRef();
            this.Reader.GoToNextLine(); // Skip startxref keyword

            long offset = this.ReadLong();
            if (offset == -1)
            {
                throw new NotSupportedException("Startxref keyword cannot be found");
            }

            TrailerOld trailer = this.ReadCrossReferenceCollection(offset);

            return trailer;
        }

        public PdfDictionaryOld ParseCurrentTokenToDictionary(IndirectObjectOld containingObject)
        {
            if (this.Reader.TokenType == Token.DictionaryStart)
            {
                PdfDictionaryOld dict = new PdfDictionaryOld(this.ContentManager);
                object[] content = this.ReadObjectsToToken(containingObject, Token.DictionaryEnd);
                dict.Load(content);
                return dict;
            }
            return null;
        }

        protected override void ParseCurrentTokenOverride(Token token, Stack<object> stack)
        {
            switch (token)
            {
                case Token.StreamStart:
                    stack.Push(this.ParseCurrentTokenToStream(stack.Pop() as PdfDictionaryOld));
                    break;
            }
        }

        protected object[] ReadObjectsToToken(IndirectObjectOld containingObject, Token endToken)
        {
            Stack<object> stack = new Stack<object>();

            Token tok;
            while (!this.Reader.EndOfFile && (tok = this.Reader.ReadToken()) != endToken)
            {
                switch (tok)
                {
                    case Token.ArrayStart:
                        stack.Push(this.ParseCurrentTokenToArray(containingObject));
                        break;
                    case Token.Boolean:
                        stack.Push(this.ParseCurrentTokenToBool());
                        break;
                    case Token.Real:
                        stack.Push(this.ParseCurrentTokenToReal());
                        break;
                    case Token.DictionaryStart:
                        stack.Push(this.ParseCurrentTokenToDictionary(containingObject));
                        break;
                    case Token.String:
                        stack.Push(this.ParseCurrentTokenToString(containingObject));
                        break;
                    case Token.Integer:
                        stack.Push(this.ParseCurrentTokenToInt());
                        break;
                    case Token.Name:
                        stack.Push(this.ParseCurrentTokenToName());
                        break;
                    case Token.Null:
                        stack.Push(null);
                        break;
                    case Token.IndirectReference:
                        int generationNo = ((PdfIntOld)stack.Pop()).Value;
                        int objectNo = ((PdfIntOld)stack.Pop()).Value;
                        stack.Push(this.ParseCurrentTokenToIndirectReference(objectNo, generationNo));
                        break;
                    case Token.Unknown:
                        break;
                    case Token.StreamStart:
                        stack.Push(this.ParseCurrentTokenToStream(stack.Pop() as PdfDictionaryOld, containingObject));
                        break;
                }
            }

            int cnt = stack.Count - 1;
            object[] res = new object[cnt + 1];
            while (cnt >= 0)
            {
                res[cnt] = stack.Pop();
                cnt--;
            }

            return res;
        }

        private long ReadLong()
        {
            if (this.Reader.ReadToken() == Token.Integer)
            {
                long res;
                long.TryParse(this.Reader.Result, out res);
                return res;
            }
            return -1;
        }

        private PdfDictionaryOld ReadDictionary()
        {
            this.Reader.ReadToken();
            return this.ParseCurrentTokenToDictionary();
        }

        private PdfDataStream ParseCurrentTokenToStream(PdfDictionaryOld dict, IndirectObjectOld containingObject = null)
        {
            PdfDataStream stream = new PdfDataStream(this.Reader, dict, containingObject);
            stream.FixLengthAndPosition(this.Reader.Position);

            return stream;
        }

        private PdfStringOld ParseCurrentTokenToString(IndirectObjectOld containingObject)
        {
            if (this.Reader.TokenType == Token.String)
            {
                return new PdfStringOld(this.ContentManager, this.ContentManager.DecryptString(containingObject, this.Reader.BytesToken));
            }

            return null;
        }

        private PdfArrayOld ParseCurrentTokenToArray(IndirectObjectOld containingObject)
        {
            if (this.Reader.TokenType == Token.ArrayStart)
            {
                PdfArrayOld arr = new PdfArrayOld(this.ContentManager);
                object[] content = this.ReadObjectsToToken(containingObject, Token.ArrayEnd);
                arr.Load(content);
                return arr;
            }

            return null;
        }

        private CrossReferenceEntryOld ReadCrossReferenceEntry()
        {
            string line = this.Reader.ReadLine();
            while (string.IsNullOrEmpty(line))
            {
                line = this.Reader.ReadLine();
            }

            string[] res = line.Split(new char[] { ' ' });

            int offset, gn;
            int.TryParse(res[0], out offset);
            int.TryParse(res[1], out gn);

            return new CrossReferenceEntryOld { Field2 = offset, Field3 = gn, Type = res[2] == "f" ? CrossReferenceEntryTypeOld.Free : CrossReferenceEntryTypeOld.Used };
        }

        private TrailerOld InitTrailerAndCrossReferenceTable(long offset)
        {
            this.Reader.Seek(offset, SeekOrigin.Begin);

            Token tok;

            if ((tok = this.Reader.ReadToken()) == Token.XRef)
            {
                tok = this.ParseCrossReferenceTable();
            }

            if (tok == Token.Trailer)
            {
                PdfDictionaryOld dict = this.ReadDictionary(); // Read trailer dictionary;
                TrailerOld t = new TrailerOld(this.ContentManager);
                t.Load(dict);

                if (t.XRefStm != null)
                {
                    this.InitTrailerAndCrossReferenceStream(t.XRefStm.Value);
                }

                if (t.Prev != null)
                {
                    this.ReadCrossReferenceCollection(t.Prev.Value);
                }

                return t;
            }

            return null;
        }

        private Token ParseCrossReferenceTable()
        {
            Token tok;

            while ((tok = this.Reader.ReadToken()) == Token.Integer)
            {
                int first = this.ParseCurrentTokenToInt().Value, count = this.ReadInt(); // Each Subsection must start with two integers: first object number and object count.
                this.Reader.GoToNextLine();
                for (int i = 0; i < count; i++)
                {
                    this.ContentManager.CrossReferences.AddCrossReferenceEntry(first + i, this.ReadCrossReferenceEntry());
                }
            }

            return tok;
        }

        private TrailerOld InitTrailerAndCrossReferenceStream(long offset)
        {
            byte[] data;
            CrossReferenceStreamOld crossReferenceStream;

            this.ReadCrossReferenceStream(offset, out crossReferenceStream, out data);
            crossReferenceStream.Append(crossReferenceStream, data);
            TrailerOld trailer = new TrailerOld(this.ContentManager);
            trailer.CopyPropertiesFrom(crossReferenceStream);

            if (crossReferenceStream.Prev != null)
            {
                this.ReadCrossReferenceCollection(crossReferenceStream.Prev.Value);
            }

            return trailer;
        }

        private void ReadCrossReferenceStream(long offset, out CrossReferenceStreamOld stream, out byte[] data)
        {
            stream = null;
            data = null;
            IndirectObjectOld obj;

            if (this.TryReadIndirectObject(offset, false, out obj))
            {
                stream = new CrossReferenceStreamOld(this.ContentManager);
                PdfDataStream dataStream = (PdfDataStream)obj.Value;
                stream.Load(obj);
                data = dataStream.ReadData(this.ContentManager);
            }
        }

        private TrailerOld ReadCrossReferenceCollection(long offset)
        {
            this.Reader.Seek(offset, SeekOrigin.Begin);
            TrailerOld trailer = null;
            Token token = this.Reader.ReadToken();

            if (token == Token.XRef)
            {
                trailer = this.InitTrailerAndCrossReferenceTable(offset);
            }
            else
            {
                trailer = this.InitTrailerAndCrossReferenceStream(offset);
            }

            return trailer;
        }
    }
}
