﻿using System.Collections.Generic;
using System.IO;
using Telerik.Windows.Documents.Utilities;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.Encryption;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.Trees;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader.Parsers;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.Text;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.Annot;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.Forms;
#if WPF
using Telerik.Windows.Documents.Fixed.Model.DigitalSignatures;
#endif

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader
{
    internal class PdfContentManager
    {
#if SILVERLIGHT
        private readonly object lockObject;
#endif
        private readonly PdfFormatProvider provider;
        private readonly CrossReferenceCollection crossReferences;
        private readonly Dictionary<int, ObjectStreamOld> objectStreams;
        private readonly Dictionary<IndirectReferenceOld, object> indirectReferencesCache;
        private readonly Dictionary<IndirectReferenceOld, IndirectObjectOld> contentCache;
        private readonly Dictionary<IndirectReferenceOld, FormField> fieldsCache;
        private readonly Dictionary<FormFieldNodeOld, List<WidgetOld>> nodeToWidgets;
        private readonly List<PageOld> pages;
        private readonly PdfParser pdfParser;
        private readonly TrailerOld trailer;
        private readonly EncryptOld encrypt;
        private readonly PdfArrayOld id;

        public PdfContentManager(PdfFormatProvider provider, Stream stream) 
            : this()
        {
            Guard.ThrowExceptionIfNull(provider, "provider");
            Guard.ThrowExceptionIfNull(stream, "stream");

            this.provider = provider;
#if SILVERLIGHT
            this.lockObject = new object();
#endif
            this.pdfParser = new PdfParser(this, stream);

            this.trailer = this.pdfParser.InitializeCrossReferences();
            this.encrypt = this.trailer.Encrypt;
            this.id = this.trailer.ID;

            if (this.encrypt != null)
            {
                FormatProviders.Pdf.Model.Encryption.Encrypt.ValidateEncryptionAlgorithmValue((int)this.encrypt.Algorithm.Value);

                if (this.id != null)
                {
                    this.encrypt.Initialize(this.id, string.Empty);
                }
            }
        }

        /// <summary>
        /// Use for test purposes only!
        /// </summary>
        internal PdfContentManager()
        {
            this.crossReferences = new CrossReferenceCollection();
            this.objectStreams = new Dictionary<int, ObjectStreamOld>();
            this.contentCache = new Dictionary<IndirectReferenceOld, IndirectObjectOld>();
            this.fieldsCache = new Dictionary<IndirectReferenceOld, FormField>();
            this.nodeToWidgets = new Dictionary<FormFieldNodeOld, List<WidgetOld>>();
            this.indirectReferencesCache = new Dictionary<IndirectReferenceOld, object>();
            this.pages = new List<PageOld>();
            this.ResourceManager = new PdfResourceManager(this);
            this.FontsManager = new FontsManagerOld();
        }

        public DocumentCatalogOld DocumentCatalog
        {
            get
            {
                return this.trailer.Root;
            }
        }

        public CrossReferenceCollection CrossReferences
        {
            get
            {
                return this.crossReferences;
            }
        }

        public PdfResourceManager ResourceManager
        {
            get;
            private set;
        }

        public FontsManagerOld FontsManager
        {
            get;
            private set;
        }

        private EncryptOld Encrypt
        {
            get
            {
                return this.encrypt;
            }
        }

        public void RegisterIndirectReference(IndirectReferenceOld reference, object obj)
        {
            Guard.ThrowExceptionIfNull(reference, "reference");

            this.indirectReferencesCache[reference] = obj;
        }

        public bool TryGetPdfObject(IndirectReferenceOld reference, out object obj)
        {
            Guard.ThrowExceptionIfNull(reference, "reference");

            return this.indirectReferencesCache.TryGetValue(reference, out obj);
        }

        public void RegisterFormField(IndirectReferenceOld reference, FormField field)
        {
            Guard.ThrowExceptionIfNull(reference, "reference");

            this.fieldsCache[reference] = field;
        }

        public bool TryGetFormField(IndirectReferenceOld reference, out FormField field)
        {
            Guard.ThrowExceptionIfNull(reference, "reference");

            return this.fieldsCache.TryGetValue(reference, out field);
        }

        public IEnumerable<WidgetOld> GetChildWidgets(FormFieldNodeOld parent)
        {
            List<WidgetOld> childWidgets;
            if (this.nodeToWidgets.TryGetValue(parent, out childWidgets))
            {
                foreach (WidgetOld widget in childWidgets)
                {
                    yield return widget;
                }
            }
        }

        public void AddWidgetParent(WidgetOld widgetObject, FormFieldNodeOld parent)
        {
            widgetObject.FormField = parent;

            List<WidgetOld> childWidgets;
            if (!this.nodeToWidgets.TryGetValue(parent, out childWidgets))
            {
                childWidgets = new List<WidgetOld>();
                this.nodeToWidgets.Add(parent, childWidgets);
            }

            childWidgets.Add(widgetObject);
        }

        public byte[] DecryptStream(IndirectObjectOld containingObject, byte[] inputData)
        {
            Guard.ThrowExceptionIfNull(containingObject, "containingObject");
            Guard.ThrowExceptionIfNull(inputData, "inputData");

            if (this.Encrypt != null)
            {
                return this.Encrypt.DecryptStream(containingObject.ObjectNumber, containingObject.GenerationNumber, inputData);
            }
            else
            {
                return inputData;
            }
        }

        public byte[] DecryptString(IndirectObjectOld containingObject, byte[] inputData)
        {
            Guard.ThrowExceptionIfNull(containingObject, "containingObject");
            Guard.ThrowExceptionIfNull(inputData, "inputData");

            if (this.Encrypt != null)
            {
                return this.Encrypt.DecryptString(containingObject.ObjectNumber, containingObject.GenerationNumber, inputData);
            }
            else
            {
                return inputData;
            }
        }

        public byte[] ReadData(IndirectReferenceOld reference)
        {
            return this.ReadData(reference, true);
        }

        public byte[] ReadData(IndirectReferenceOld reference, bool shouldDecode)
        {
            Guard.ThrowExceptionIfNull(reference, "reference");

            lock (this.provider)
            {
                IndirectObjectOld obj;
                if (this.TryReadIndirectObject(reference, true, out obj))
                {
                    PdfDataStream dataStream = (PdfDataStream)obj.Value;
                    if (shouldDecode)
                    {
                        return dataStream.ReadData(this);
                    }
                    else
                    {
                        return dataStream.ReadDataWithoutDecoding(this);
                    }
                }

                return new byte[0];
            }
        }

        public bool TryGetIndirectObject(IndirectReferenceOld reference, out IndirectObjectOld indirectObject)
        {
            Guard.ThrowExceptionIfNull(reference, "reference");

            return this.TryGetIndirectObject(reference, true, out indirectObject);
        }

        public bool TryGetIndirectObject(IndirectReferenceOld reference, bool isEncrypted, out IndirectObjectOld indirectObject)
        {
            Guard.ThrowExceptionIfNull(reference, "reference");

            if (!this.contentCache.TryGetValue(reference, out indirectObject))
            {
                if (!this.TryReadIndirectObject(reference, isEncrypted, out indirectObject))
                {
                    return false;
                }

                PdfDataStream stream = indirectObject.Value as PdfDataStream;
                if (stream == null)
                {
                    this.contentCache[reference] = indirectObject;
                }
            }

            return true;
        }

        public virtual bool TryReadIndirectObject(IndirectReferenceOld fromRef, bool isEncrypted, out IndirectObjectOld indirectObject)
        {
            Guard.ThrowExceptionIfNull(fromRef, "fromRef");

            CrossReferenceEntryOld entry;
            if (!this.CrossReferences.TryGetCrossReferenceEntry(fromRef.ObjectNumber, out entry) || entry.Type == CrossReferenceEntryTypeOld.Free)
            {
                indirectObject = null;
                return false;
            }

            if (this.TryReadIndirectObject(entry, isEncrypted, out indirectObject))
            {
                indirectObject.GenerationNumber = fromRef.GenerationNumber;
                indirectObject.ObjectNumber = fromRef.ObjectNumber;

                return true;
            }

            return false;
        }

        public void LoadIndirectObject<T>(T obj) where T : PdfObjectOld
        {
            Guard.ThrowExceptionIfNull(obj, "obj");

            if (obj.IsLoaded || obj.Reference == null)
            {
                return;
            }

            IndirectObjectOld res;
            if (this.TryGetIndirectObject(obj.Reference, out res))
            {
                obj.Load(res);
            }
        }

        public bool TryGetSimpleType<T>(IndirectReferenceOld reference, out T simpleType)
            where T : class
        {
            Guard.ThrowExceptionIfNull(reference, "reference");

            simpleType = null;
            IndirectObjectOld obj;
            if (this.TryGetIndirectObject(reference, out obj))
            {
                simpleType = obj.Value as T;
            }

            return simpleType != null;
        }

        public IEnumerable<PageOld> GetPages()
        {
            Guard.ThrowExceptionIfNull(this.DocumentCatalog, "Catalog");
            Guard.ThrowExceptionIfNull(this.DocumentCatalog.Pages, "Pages");

            this.LoadPages(this.DocumentCatalog.Pages.Reference);

            return this.pages;
        }

        public bool TryReadIndirectObject(long offset, bool isEncrypted, out IndirectObjectOld indirectObject)
        {
            return this.pdfParser.TryReadIndirectObject(offset, isEncrypted, out indirectObject);
        }

#if WPF
        public ISourceStream GetSourceStream()
        {
            FileSourceStream sourceStream = new FileSourceStream(this.pdfParser.Reader.Stream);

            return sourceStream;
        }
#endif

        private void LoadPages(IndirectReferenceOld root)
        {
            this.LoadPages(root, null);
        }

        private void LoadPages(IndirectReferenceOld root, PageTreeNodeOld parent)
        {
            Guard.ThrowExceptionIfNull(root, "root");

            IndirectObjectOld obj;
            if (!this.TryGetIndirectObject(root, out obj))
            {
                return;
            }

            PdfNameOld type = obj.GetObjectType();
            if (type == null)
            {
                return;
            }

            switch (type.Value)
            {
                case "Pages":
                    PageTreeNodeOld node = new PageTreeNodeOld(this);
                    node.Load(obj);
                    node.Parent = parent;
                    foreach (IndirectReferenceOld ir in node.Children)
                    {
                        this.LoadPages(ir, node);
                    }

                    break;

                case "Page":
                    PageOld page = new PageOld(this);
                    page.Load(obj);
                    page.Parent = parent;
                    this.pages.Add(page);
                    break;
            }
        }

        private bool TryReadIndirectObject(CrossReferenceEntryOld entry, bool isEncrypted, out IndirectObjectOld indirectObject)
        {
            Guard.ThrowExceptionIfNull(entry, "entry");

#if SILVERLIGHT
            lock (this.lockObject)
#else
            lock (this.provider)
#endif
            {
                switch (entry.Type)
                {
                    case CrossReferenceEntryTypeOld.Used:

                        return this.TryReadIndirectObject(entry.Field2, isEncrypted, out indirectObject);

                    case CrossReferenceEntryTypeOld.Compressed:

                        ObjectStreamOld str;
                        if (this.TryGetObjectStream((int)entry.Field2, isEncrypted, out str))
                        {
                            indirectObject = str.ReadIndirectObject(entry.Field3);
                            return true;
                        }
                        else
                        {
                            indirectObject = null;
                            return false;
                        }

                    default:

                        indirectObject = null;
                        return false;
                }
            }
        }

        private bool TryGetObjectStream(int objNum, bool isEncrypted, out ObjectStreamOld objectStream)
        {
            if (!this.objectStreams.TryGetValue(objNum, out objectStream))
            {
                IndirectObjectOld obj;
                if (!this.TryReadIndirectObject(this.CrossReferences[objNum], isEncrypted, out obj))
                {
                    return false;
                }

                objectStream = new ObjectStreamOld(this);
                objectStream.Load(obj);
                this.objectStreams[objNum] = objectStream;
            }

            return true;
        }
    }
}