﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Core.Imaging;
using Telerik.Windows.Documents.Utilities;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.Fonts;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.XObjects;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader.Parsers;
using Telerik.Windows.Documents.Fixed.Model.Internal;
using Telerik.Windows.Documents.Fixed.Model.Internal.Classes;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.Filters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.Text;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Filters;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Fixed.Model.Extensions;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader
{
    internal class PdfResourceManager
    {
        private readonly Dictionary<ImageResourceKey, InlineImageInfo> localImages;
        private readonly Dictionary<ImageResourceKey, XImage> globalImages;
        private readonly Dictionary<FontBaseOld, ResourceKey> localFontKeys;
        private readonly Dictionary<XForm, IEnumerable<ContentElement>> xFormContents;
        private readonly PdfContentManager contentManager;

        public PdfResourceManager(PdfContentManager contentManager)
        {
            Guard.ThrowExceptionIfNull(contentManager, "contentManager");

            this.localImages = new Dictionary<ImageResourceKey, InlineImageInfo>();
            this.globalImages = new Dictionary<ImageResourceKey, XImage>();
            this.xFormContents = new Dictionary<XForm, IEnumerable<ContentElement>>();
            this.localFontKeys = new Dictionary<FontBaseOld, ResourceKey>();

            this.contentManager = contentManager;
        }

        public ImageDataSource GetImageSource(ImageResourceKey resourceKey)
        {
            Guard.ThrowExceptionIfNull(resourceKey, "resourceKey");

            switch (resourceKey.Type)
            {
                case ResourceType.Global:
                    return this.GetGlobalImageSource(resourceKey);

                case ResourceType.Local:
                    return this.GetInlineImageSource(resourceKey);

                default:
                    throw new NotSupportedException("Resource key type is not supported.");
            }
        }

        public ImageDataSource GetImageSourceWithStencilColor(ImageResourceKey resourceKey)
        {
            Guard.ThrowExceptionIfNull(resourceKey, "resourceKey");
            Guard.ThrowExceptionIfNull(resourceKey.StencilColor, "resourceKey.StencilColor");

            if (resourceKey.Type == ResourceType.Local)
            {
                return this.GetInlineImageSource(resourceKey);
            }
            else
            {
                return this.GetGlobalImageSourceWithStencilColor(resourceKey);
            }
        }

        public bool TryGetInlineImageInfo(ImageResourceKey resourceKey, out InlineImageInfo imageInfo)
        {
            return this.localImages.TryGetValue(resourceKey, out imageInfo);
        }

        public void RegisterInlineImage(InlineImageInfo imageInfo, ImageResourceKey resourceKey)
        {
            Guard.ThrowExceptionIfNull(imageInfo, "imageInfo");
            Guard.ThrowExceptionIfNull(resourceKey, "resourceKey");

            this.localImages[resourceKey] = imageInfo;
        }

        public ImageResourceKey RegisterXImage(XImage image, Color stencilColor)
        {
            Guard.ThrowExceptionIfNull(image, "image");
            Guard.ThrowExceptionIfNull(image.Reference, "reference");

            IndirectReferenceOld reference = image.Reference;
            ImageResourceKey key = new ImageResourceKey();
            key.Id = reference.ObjectNumber;
            key.Type = ResourceType.Global;

            if(image.ImageMask != null && image.ImageMask.Value)
            {
                key.StencilColor = stencilColor;
            }

            this.globalImages[key] = image;

            return key;
        }

        public ResourceKey GetFontKey(FontBaseOld font)
        {
            if (font.Reference != null)
            {
                return new ResourceKey() { Id = font.Reference.ObjectNumber, Type = ResourceType.Global };
            }
            else
            {
                ResourceKey fontKey;
                if (this.localFontKeys.TryGetValue(font, out fontKey))
                {
                    return fontKey;
                }
                else
                {
                    this.localFontKeys[font] = new ResourceKey() { Id = this.localFontKeys.Count, Type = ResourceType.Local };
                    return this.localFontKeys[font];
                }
            }
        }

        public IEnumerable<ContentElement> GetXFormContent(PdfResourceOld baseResources, XForm form, bool skipNonTextRelatedOffset)
        {
            Guard.ThrowExceptionIfNull(form, "form");

            form.Load();
            IEnumerable<ContentElement> res;
            if (!this.xFormContents.TryGetValue(form, out res))
            {
                byte[] data = this.contentManager.ReadData(form.Reference);
                PdfResourceOld resources = form.Resources == null ? baseResources : form.Resources;
                ContentStreamParser p = new ContentStreamParser(this.contentManager, data, form.Clip, resources, form.Reference);

                if (skipNonTextRelatedOffset)
                {
                    res = p.ParseOnlyTextRelatedContent();
                }
                else
                {
                    res = p.ParseContent();
                    this.xFormContents[form] = res;
                }
            }

            return res;
        }

        private ImageDataSource GetGlobalImageSource(ImageResourceKey resourceKey)
        {
            Guard.ThrowExceptionIfNull(resourceKey, "resourceKey");

            XImage xImage;
            if (!this.globalImages.TryGetValue(resourceKey, out xImage))
            {
                throw new NotSupportedException();
            }

            return xImage.CreateImageSource(xImage);
        }

        private ImageDataSource GetGlobalImageSourceWithStencilColor(ImageResourceKey resourceKey)
        {
            Guard.ThrowExceptionIfNull(resourceKey, "resourceKey");
            Guard.ThrowExceptionIfNull(resourceKey.StencilColor, "resourceKey.StencilColor");

            XImage xImage;
            if (!this.globalImages.TryGetValue(resourceKey, out xImage))
            {
                throw new NotSupportedException();
            }
            
            return xImage.CreateImageSourceWithStencilColor(xImage, resourceKey.StencilColor.Value);
        }

        private ImageDataSource GetInlineImageSource(ImageResourceKey resourceKey)
        {
            InlineImageInfo imageInfo = this.localImages[resourceKey];

            ImageDataSource imageSource = resourceKey.StencilColor.HasValue ?
                imageInfo.XImage.CreateImageSourceWithStencilColor(imageInfo, resourceKey.StencilColor.Value) :
                imageInfo.XImage.CreateImageSource(imageInfo);

            return imageSource;
        }
    }
}
