﻿using System.Collections.Generic;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Text;
using Telerik.Windows.Documents.Fixed.Utilities.Rendering;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export
{
    internal class PdfContentExportContext : IPdfContentExportContext
    {
        private readonly IPdfExportContext owner;
        private readonly IResourceHolder resourceHolder;
        private readonly IContentRootElement contentRoot;
        private readonly Matrix rootDipToPointTransform;
        private readonly ContentElementsGraph graph;
        private Marker previousContentMarker;

        public PdfContentExportContext(IPdfExportContext owner, IResourceHolder resourceHolder, IContentRootElement contentRoot)
        {
            Guard.ThrowExceptionIfNull(owner, "owner");
            Guard.ThrowExceptionIfNull(resourceHolder, "resourceHolder");
            Guard.ThrowExceptionIfNull(contentRoot, "contentRoot");

            this.owner = owner;
            this.resourceHolder = resourceHolder;
            this.contentRoot = contentRoot;
            this.rootDipToPointTransform = owner.GetDipToPdfPointTransformation(this.contentRoot);
            this.graph = new ContentElementsGraph(this.contentRoot);
        }

        public IPdfExportContext Owner
        {
            get
            {
                return this.owner;
            }
        }

        public IResourceHolder ResourceHolder
        {
            get
            {
                return this.resourceHolder;
            }
        }

        public IContentRootElement ContentRoot
        {
            get
            {
                return this.contentRoot;
            }
        }

        public Matrix RootDipToPointTransformation
        {
            get
            {
                return this.rootDipToPointTransform;
            }
        }

        public MarkedContentExportInfo GetCurrentContentMarkerExportInfo(Marker currentContentMarker)
        {
            // Currently handling nested makers is not supported, so we track only the current marker.
            MarkedContentExportInfo exportInfo = new MarkedContentExportInfo();

            bool areDifferentMarkers = currentContentMarker != this.previousContentMarker;
            exportInfo.ShouldExportMarkedContentStart = (areDifferentMarkers && currentContentMarker != null);
            exportInfo.ShouldExportMarkedContentEnd = (areDifferentMarkers && this.previousContentMarker != null);
            this.previousContentMarker = currentContentMarker;

            return exportInfo;
        }

        public ResourceEntry GetResource(FontBase font)
        {
            Guard.ThrowExceptionIfNull(font, "font");

            if (this.resourceHolder.Resources.Fonts == null)
            {
                this.resourceHolder.Resources.Fonts = new PdfDictionary();
            }

            ResourceEntry resource = this.Owner.GetResource(font);
            this.resourceHolder.Resources.Fonts[resource.ResourceKey] = resource.Resource.Reference;

            return resource;
        }

        public ResourceEntry GetResource(Fixed.Model.Resources.FormSource formSource)
        {
            Guard.ThrowExceptionIfNull(formSource, "formSource");

            ResourceEntry resource = this.Owner.GetResource(formSource);
            this.AddXObjectsResourceEntry(resource);

            return resource;
        }

        public ResourceEntry GetResource(Fixed.Model.Resources.ImageSource imageSource)
        {
            Guard.ThrowExceptionIfNull(imageSource, "imageSource");

            ResourceEntry resource = this.Owner.GetResource(imageSource);
            this.AddXObjectsResourceEntry(resource);

            return resource;
        }

        public ResourceEntry GetResource(PatternColor pattern)
        {
            Guard.ThrowExceptionIfNull(pattern, "pattern");

            if (this.resourceHolder.Resources.Patterns == null)
            {
                this.resourceHolder.Resources.Patterns = new PdfDictionary();
            }

            ResourceEntry resource = this.Owner.GetResource(pattern, this);
            this.resourceHolder.Resources.Patterns[resource.ResourceKey] = resource.Resource.Reference;

            return resource;
        }

        public ResourceEntry GetResource(ColorSpaceBase first, ColorSpaceBase second)
        {
            Guard.ThrowExceptionIfNull(first, "first");
            Guard.ThrowExceptionIfNull(second, "second");

            if (this.resourceHolder.Resources.ColorSpaces == null)
            {
                this.resourceHolder.Resources.ColorSpaces = new PdfDictionary();
            }

            ResourceEntry resource = this.Owner.GetResource(first, second);
            this.resourceHolder.Resources.ColorSpaces[resource.ResourceKey] = resource.Resource.Reference;

            return resource;
        }

        public ResourceEntry GetResource(ExtGState state)
        {
            Guard.ThrowExceptionIfNull(state, "state");

            if (this.resourceHolder.Resources.ExtGState == null)
            {
                this.resourceHolder.Resources.ExtGState = new PdfDictionary();
            }

            ResourceEntry resource = this.Owner.GetResource(state);
            this.resourceHolder.Resources.ExtGState[resource.ResourceKey] = resource.Resource.Reference;

            return resource;
        }

        public void SetUsedCharacters(FontBase font, TextCollection glyphs)
        {
            Guard.ThrowExceptionIfNull(font, "font");
            Guard.ThrowExceptionIfNull(glyphs, "glyphs");

            this.Owner.SetUsedCharacters(font, glyphs);
        }

        public IEnumerable<CharInfo> GetUsedCharacters(FontBase font)
        {
            Guard.ThrowExceptionIfNull(font, "font");

            return this.Owner.GetUsedCharacters(font);
        }

        public IEnumerable<ContentElementBase> GetClippingChildren(Fixed.Model.Clipping clipping)
        {
            return this.graph.GetClippingChildren(clipping);
        }

        public IEnumerable<ContentElementBase> GetRootElements()
        {
            return this.graph.GetRootElements();
        }

        private void AddXObjectsResourceEntry(ResourceEntry resource)
        {
            if (this.resourceHolder.Resources.XObjects == null)
            {
                this.resourceHolder.Resources.XObjects = new PdfDictionary();
            }

            this.resourceHolder.Resources.XObjects[resource.ResourceKey] = resource.Resource.Reference;
        }
    }
}
