﻿using System;
using System.Collections.Generic;
using System.Linq;
#if NETSTANDARD
using Matrix = Telerik.Documents.Primitives.Matrix;
#else
using Matrix = System.Windows.Media.Matrix;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Annotations;
#if WPF
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DigitalSignature;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Fonts;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Forms;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Objects;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Encryption;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Fixed.Model.Text;
using Telerik.Windows.Documents.Utilities;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.GraphicsState;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export
{
    internal class PdfExportContext : IPdfExportContext
    {
        private const string FontResourcePrefix = "F";
        private const string FormXObjectResourcePrefix = "X";
        private const string ImageResourcePrefix = "I";
        private const string PatternResourcePrefix = "P";
        private const string GraphicResourcePrefix = "G";
        private const string ColorResourcePrefix = "C";
        private const int ResourcesInitialIndex = 1;
        private static readonly double dipToPoint = Media.Unit.DipToPoint(1);
        private readonly CrossReferenceCollection crossReferenceCollection;
        private readonly Queue<IndirectObject> indirectObjects;
        private readonly InstanceIdCache<IPdfPrimitive, IndirectObject> registeredIndirectObjects;
        private readonly Dictionary<FontBase, ResourceEntry> fontResources;
        private readonly InstanceIdCache<FormSource, ResourceEntry> formResources;
        private readonly InstanceIdCache<ImageSource, ResourceEntry> imageResources;
        private readonly InstanceIdCache<PatternColor, ResourceEntry> patternResources;
        private readonly InstanceIdCache<PatternColor, ResourceEntry> uncoloredTilingResources;
        private readonly InstanceIdCache<ExtGState, ResourceEntry> extGStateResources;
        private readonly Dictionary<string, ResourceEntry> colorspaceResources;
        private readonly Dictionary<FontBase, HashSet<CharInfo>> usedCharacters;
        private readonly Dictionary<double, Matrix> contentRootHeightToDipToPdfPointTransformations;
        private readonly InstanceIdCache<RadFixedPage, Page> fixedPageToPage;
        private readonly InstanceIdCache<Widget, WidgetObject> widgetModelToWidgetObject;
        private readonly InstanceIdCache<FormField, FormFieldNode> fieldToNode;
        private readonly byte[] documentId;
        private readonly PdfExportSettings settings;
        private readonly RadFixedDocumentInfo documentInfo;

        private int fontResourcesCount;
        private int formXObjectResourcesCount;
        private int imageResourcesCount;
        private int patternResourcesCount;
        private int graphicResourcesCount;
        private int colorResourcesCount;
        private IndirectObject currentIndirectObject;
#if WPF
        private SignatureExportInfo signatureExportInfo;
#endif

        public PdfExportContext(RadFixedDocumentInfo documentInfo, PdfExportSettings settings)
        {
            Guard.ThrowExceptionIfNull(settings, "settings");

            this.settings = settings;
            this.documentInfo = documentInfo;
            this.documentId = CreateDocumentId();

            this.registeredIndirectObjects = new InstanceIdCache<IPdfPrimitive, IndirectObject>();

            this.fontResources = new Dictionary<FontBase, ResourceEntry>();
            this.formResources = new InstanceIdCache<FormSource, ResourceEntry>();
            this.imageResources = new InstanceIdCache<ImageSource, ResourceEntry>();
            this.patternResources = new InstanceIdCache<PatternColor, ResourceEntry>();
            this.uncoloredTilingResources = new InstanceIdCache<PatternColor, ResourceEntry>();
            this.extGStateResources = new InstanceIdCache<ExtGState, ResourceEntry>();
            this.colorspaceResources = new Dictionary<string, ResourceEntry>();

            this.fontResourcesCount = ResourcesInitialIndex;
            this.formXObjectResourcesCount = ResourcesInitialIndex;
            this.imageResourcesCount = ResourcesInitialIndex;
            this.patternResourcesCount = ResourcesInitialIndex;
            this.graphicResourcesCount = ResourcesInitialIndex;
            this.colorResourcesCount = ResourcesInitialIndex;
            this.CanWriteColors = true;

            this.indirectObjects = new Queue<IndirectObject>();
            this.usedCharacters = new Dictionary<FontBase, HashSet<CharInfo>>();

            this.crossReferenceCollection = new CrossReferenceCollection();
            this.crossReferenceCollection.AddCrossReferenceEntry(0, new CrossReferenceEntry(0, 65535, CrossReferenceEntryType.Free));
            this.contentRootHeightToDipToPdfPointTransformations = new Dictionary<double, Matrix>();
            this.fixedPageToPage = new InstanceIdCache<RadFixedPage, Page>();
            this.widgetModelToWidgetObject = new InstanceIdCache<Widget, WidgetObject>();
            this.fieldToNode = new InstanceIdCache<FormField, FormFieldNode>();
#if WPF
            this.signatureExportInfo = new SignatureExportInfo();
#endif
        }

        public bool CanWriteColors
        {
            get;
            set;
        }

        public Encrypt Encryption
        {
            get;
            set;
        }

        public PdfExportSettings Settings
        {
            get
            {
                return this.settings;
            }
        }

        public byte[] DocumentId
        {
            get
            {
                return this.documentId;
            }
        }

        public IEnumerable<FontBase> FontResources
        {
            get
            {
                return this.fontResources.Keys;
            }
        }

        public Queue<IndirectObject> IndirectObjectsQueue
        {
            get
            {
                return this.indirectObjects;
            }
        }

        public CrossReferenceCollection CrossReferenceCollection
        {
            get
            {
                return this.crossReferenceCollection;
            }
        }

        public RadFixedDocumentInfo DocumentInfo
        {
            get
            {
                return this.documentInfo;
            }
        }

        public IPdfContentExportContext AcroFormContentExportContext { get; set; }

#if WPF
        public SignatureExportInfo SignatureExportInfo
        {
            get
            {
                return this.signatureExportInfo;
            }
        }
#endif

        public IndirectObject CreateIndirectObject(IPdfPrimitive primitive)
        {
            Guard.ThrowExceptionIfNull(primitive, "primitive");

            return this.CreateIndirectObject(primitive, true);
        }

        public void BeginExportIndirectObject(IndirectObject indirectObject, long offset)
        {
            Guard.ThrowExceptionIfNull(indirectObject, "indirectObject");
            Guard.ThrowExceptionIfNotNull(this.currentIndirectObject, "currentIndirectObject");

            this.currentIndirectObject = indirectObject;
            CrossReferenceEntryType entryType = indirectObject.Content.Type == PdfElementType.Null ? CrossReferenceEntryType.Free : CrossReferenceEntryType.Used;
            this.crossReferenceCollection.UpdateCrossReferenceEntry(indirectObject.Reference.ObjectNumber, offset, 0, entryType);
        }

        public void EndExportIndirectObject()
        {
            this.currentIndirectObject.ReleaseContent();
            this.currentIndirectObject = null;
        }

        public FormFieldNode CreateFormFieldObject(FormField field, bool includeInIndirectObjectQueue)
        {
            FormFieldNode node;
            if (!this.fieldToNode.TryGetValue(field, out node))
            {
                node = new FormFieldNode();
                this.fieldToNode.Add(field, node);
            }

            IndirectObject fieldObject = this.CreateIndirectObject(node, false);

            if (includeInIndirectObjectQueue)
            {
                this.IndirectObjectsQueue.Enqueue(fieldObject);
            }

            return node;
        }

        public ResourceEntry GetResource(FontBase resource)
        {
            ResourceEntry resourceEntry;
            if (!this.fontResources.TryGetValue(resource, out resourceEntry))
            {
                FontObject font = PdfFontsFactory.CreateFont(resource.Type);

                resourceEntry = this.CreateResourceEntry(font, FontResourcePrefix);
                this.fontResources[resource] = resourceEntry;
            }

            return resourceEntry;
        }

        public ResourceEntry GetResource(FormSource resource)
        {
            ResourceEntry resourceEntry;
            if (!this.formResources.TryGetValue(resource, out resourceEntry))
            {
                FormXObject xForm = new FormXObject();
                xForm.CopyPropertiesFrom(this, resource);

                if (!resource.Matrix.IsIdentity)
                {
                    xForm.Matrix = resource.Matrix.ToPdfArray();
                }

                resourceEntry = this.CreateResourceEntry(xForm, FormXObjectResourcePrefix, true);
                this.formResources[resource] = resourceEntry;
            }

            return resourceEntry;
        }

        public ResourceEntry GetResource(ImageSource resource)
        {
            ResourceEntry resourceEntry;
            if (!this.imageResources.TryGetValue(resource, out resourceEntry))
            {
                ImageXObject xImage = new ImageXObject();
                xImage.CopyPropertiesFrom(this, resource);
                resourceEntry = this.CreateResourceEntry(xImage, ImageResourcePrefix, true);
                this.imageResources[resource] = resourceEntry;
            }

            return resourceEntry;
        }

        public ResourceEntry GetResource(ExtGState resource)
        {
            ResourceEntry resourceEntry;
            if (!this.extGStateResources.TryGetValue(resource, out resourceEntry))
            {
                ExtGStateObject extGState = new Model.Elements.GraphicsState.ExtGStateObject();
                extGState.CopyPropertiesFrom(this, resource);

                resourceEntry = this.CreateResourceEntry(extGState, GraphicResourcePrefix, true);
                this.extGStateResources[resource] = resourceEntry;
            }

            return resourceEntry;
        }

        public ResourceEntry GetResource(ColorSpaceBase first, ColorSpaceBase second)
        {
            Guard.ThrowExceptionIfNull(first, "first");
            Guard.ThrowExceptionIfNull(second, "second");

            string key = string.Format("{0}{1}", first.Name, second.Name);

            ResourceEntry resourceEntry;
            if (!this.colorspaceResources.TryGetValue(key, out resourceEntry))
            {
                PdfArray array = new PdfArray() { new PdfName(first.Name), new PdfName(second.Name) };

                resourceEntry = this.CreateResourceEntry(array, ColorResourcePrefix, true);
                this.colorspaceResources[key] = resourceEntry;
            }

            return resourceEntry;

        }

        public ResourceEntry GetResource(PatternColor resource, IPdfContentExportContext context)
        {
            Guard.ThrowExceptionIfNull(resource, "resource");
            Guard.ThrowExceptionIfNull(context, "context");

            PatternColor properKey = resource;
            InstanceIdCache<PatternColor, ResourceEntry> properResources = this.patternResources;
            if (resource.PatternType == PatternType.Tiling)
            {
                UncoloredTiling uncoloredTiling = resource as UncoloredTiling;
                if (uncoloredTiling != null)
                {
                    properKey = uncoloredTiling.Tiling;
                    properResources = this.uncoloredTilingResources;
                }
            }

            ResourceEntry resourceEntry;
            if (!properResources.TryGetValue(properKey, out resourceEntry))
            {
                Model.Elements.Patterns.PatternColorObject pattern = Model.Elements.Patterns.PatternColorObject.CreateInstance((int)resource.PatternType);
                pattern.CopyPropertiesFrom(context, resource);

                resourceEntry = this.CreateResourceEntry(pattern, PatternResourcePrefix, true);
                properResources[properKey] = resourceEntry;
            }

            return resourceEntry;
        }

        public byte[] EncryptStream(byte[] data)
        {
            Guard.ThrowExceptionIfNull(data, "data");

            if (this.Encryption == null || this.currentIndirectObject == null || this.Encryption == this.currentIndirectObject.Content)
            {
                return data;
            }

            return this.Encryption.EncryptStream(this.currentIndirectObject.Reference.ObjectNumber, this.currentIndirectObject.Reference.GenerationNumber, data);
        }

        public byte[] EncryptString(byte[] data)
        {
            Guard.ThrowExceptionIfNull(data, "data");

            if (this.Encryption == null || this.currentIndirectObject == null || this.Encryption == this.currentIndirectObject.Content)
            {
                return data;
            }

            return this.Encryption.EncryptString(this.currentIndirectObject.Reference.ObjectNumber, this.currentIndirectObject.Reference.GenerationNumber, data);
        }

        public IPdfContentExportContext CreateContentExportContext(IResourceHolder resourceHolder, IContentRootElement contentRoot)
        {
            Guard.ThrowExceptionIfNull(resourceHolder, "resourceHolder");
            Guard.ThrowExceptionIfNull(contentRoot, "contentRoot");

            return new PdfContentExportContext(this, resourceHolder, contentRoot);
        }

        public void SetUsedCharacters(FontBase font, TextCollection glyphs)
        {
            Guard.ThrowExceptionIfNull(font, "font");
            Guard.ThrowExceptionIfNull(glyphs, "glyphs");

            HashSet<CharInfo> usedCharacters;
            if (!this.usedCharacters.TryGetValue(font, out usedCharacters))
            {
                usedCharacters = new HashSet<CharInfo>();
                this.usedCharacters[font] = usedCharacters;
            }

            foreach (CharInfo glyph in glyphs.Characters)
            {
                usedCharacters.Add(glyph);
            }
        }

        public IEnumerable<CharInfo> GetUsedCharacters(FontBase font)
        {
            Guard.ThrowExceptionIfNull(font, "font");

            HashSet<CharInfo> usedCharacters;
            if (!this.usedCharacters.TryGetValue(font, out usedCharacters))
            {
                return Enumerable.Empty<CharInfo>();
            }

            return usedCharacters;
        }

        public Matrix GetDipToPdfPointTransformation(IContentRootElement contentRoot)
        {
            Matrix matrix;

            if (!this.contentRootHeightToDipToPdfPointTransformations.TryGetValue(contentRoot.Size.Height, out matrix))
            {
                matrix = CalculateDipToPdfPointTransformation(contentRoot);
                this.contentRootHeightToDipToPdfPointTransformations.Add(contentRoot.Size.Height, matrix);
            }

            return matrix;
        }

        public static Matrix CalculateDipToPdfPointTransformation(IContentRootElement contentRoot)
        {
            return new Matrix(dipToPoint, 0, 0, -dipToPoint, 0, Media.Unit.DipToPoint(contentRoot.Size.Height));
        }

        public void MapPages(RadFixedPage fixedPage, Page page)
        {
            Guard.ThrowExceptionIfNull(fixedPage, "fixedPage");
            Guard.ThrowExceptionIfNull(page, "page");

            this.fixedPageToPage[fixedPage] = page;
        }

        public bool TryGetPage(RadFixedPage fixedPage, out Page page)
        {
            Guard.ThrowExceptionIfNull(fixedPage, "fixedPage");

            return this.fixedPageToPage.TryGetValue(fixedPage, out page);
        }

        public void MapWidgets(Widget widgetModel, WidgetObject widgetObject)
        {
            Guard.ThrowExceptionIfNull(widgetModel, "widgetModel");
            Guard.ThrowExceptionIfNull(widgetObject, "widgetObject");

            this.widgetModelToWidgetObject[widgetModel] = widgetObject;
        }

        public bool TryGetWidget(Widget widgetModel, out WidgetObject widgetObject)
        {
            Guard.ThrowExceptionIfNull(widgetModel, "widgetModel");

            return this.widgetModelToWidgetObject.TryGetValue(widgetModel, out widgetObject);
        }

        protected IndirectObject CreateIndirectObject(IPdfPrimitive primitive, bool includeInIndirectObjectsQueue)
        {
            int objectNumber = this.GetNextObjectNumber();
            IndirectObject indirectObject = this.CreateIndirectObject(primitive, objectNumber, includeInIndirectObjectsQueue, true);

            return indirectObject;
        }

        protected IndirectObject CreateIndirectObject(IPdfPrimitive primitive, int objectNumber, bool includeInIndirectObjectsQueue, bool addToCrossReferences)
        {
            Guard.ThrowExceptionIfNull(primitive, "primitive");

            IndirectObject res;
            if (!this.registeredIndirectObjects.TryGetValue(primitive, out res))
            {
                IndirectReference indirectReference = new IndirectReference(objectNumber, 0);
                res = new IndirectObject(indirectReference, primitive);
                this.registeredIndirectObjects[primitive] = res;

                if (includeInIndirectObjectsQueue)
                {
                    this.IndirectObjectsQueue.Enqueue(res);
                }

                if (addToCrossReferences)
                {
                    this.crossReferenceCollection.AddCrossReferenceEntry(objectNumber, new CrossReferenceEntry(0, 0, CrossReferenceEntryType.Free));
                }
            }

            return res;
        }

        protected virtual int GetNextObjectNumber()
        {
            return this.CrossReferenceCollection.MaxObjectNumber + 1;
        }

        private ResourceEntry CreateResourceEntry(IPdfPrimitive primitive, string keyPrefix, bool includeInQueue = false)
        {
            IndirectObject indirectObject = this.CreateIndirectObject(primitive, includeInQueue);
            string key = this.GetNextResourceKey(keyPrefix);
            ResourceEntry resourceEntry = new ResourceEntry(key, indirectObject);

            return resourceEntry;
        }

        private string GetNextResourceKey(string keyPrefix)
        {
            switch (keyPrefix)
            {
                case FontResourcePrefix:
                    return string.Format("{0}{1}", keyPrefix, this.fontResourcesCount++);
                case FormXObjectResourcePrefix:
                    return string.Format("{0}{1}", keyPrefix, this.formXObjectResourcesCount++);
                case ImageResourcePrefix:
                    return string.Format("{0}{1}", keyPrefix, this.imageResourcesCount++);
                case PatternResourcePrefix:
                    return string.Format("{0}{1}", keyPrefix, this.patternResourcesCount++);
                case GraphicResourcePrefix:
                    return string.Format("{0}{1}", keyPrefix, this.graphicResourcesCount++);
                case ColorResourcePrefix:
                    return string.Format("{0}{1}", keyPrefix, this.colorResourcesCount++);
                default:
                    throw new ArgumentException("Invalid prefix argument: {0}", keyPrefix);
            }
        }

        private static byte[] CreateDocumentId()
        {
            DateTime today = DateTime.Now;
            string id = today.ToString("o", System.Globalization.CultureInfo.InvariantCulture);
            return MD5Core.GetHash(System.Text.Encoding.UTF8.GetBytes(id));
        }
    }
}
