﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Fonts;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Encryption;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.Model;
#if WPF
using Telerik.Windows.Documents.Fixed.Model.DigitalSignatures;
#endif
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export
{
    internal class PdfExporter
    {
        public void Export(IRadFixedDocumentExportContext context, Stream output)
        {
            Guard.ThrowExceptionIfNull(output, "output");
            Guard.ThrowExceptionIfNull(context, "context");

            PdfWriter writer = new PdfWriter(output);
#if WPF
            context.SignatureExportInfo.AddDocumentStartPosition((int)writer.Position);
#endif
            writer.WriteDocumentStart();

            if (context.Settings.ComplianceLevel == PdfComplianceLevel.None && context.Settings.IsEncrypted)
            {
                Encrypt encryption = CreateEncryptionBasedOnPdfSettings(context, context.Settings);
                context.CreateIndirectObject(encryption);
                context.Encryption = encryption;
            }

            DocumentCatalog catalog = new DocumentCatalog();
            catalog.CopyPropertiesFrom(context);

            PdfExporter.WriteCatalogAndRelatedProperties(writer, context, catalog);
            PdfExporter.WriteFontsFromContext(writer, context);

            long crossReferenceBeginingOffset = writer.Position;

            bool hasSignedSignatureField = PdfExporter.HasSignatureField(context);
            if (hasSignedSignatureField)
            {
                PdfExporter.WriteCrossReferenceStream(writer, context, catalog, null);
            }
            else
            {
                PdfExporter.WriteTrailerAndCrossReferenceTable(writer, context, catalog, null);
            }

            PdfExporter.WriteDocumentEnd(crossReferenceBeginingOffset, writer, context);

#if WPF
            context.SignatureExportInfo.AddDocumentEndPosition((int)writer.Position - 1);
            PdfExporter.UpdateSignatures(writer, context);
#endif
        }


        internal static void WriteCatalogAndRelatedProperties(PdfWriter writer, IPdfExportContext context, DocumentCatalog catalog)
        {
            context.CreateIndirectObject(catalog);
            PdfExporter.WritePendingIndirectObjects(context, writer);
        }

        internal static void WriteFontsFromContext(PdfWriter writer, IPdfExportContext context)
        {
            foreach (FontBase resource in context.FontResources)
            {
                ResourceEntry resourceEntry = context.GetResource(resource);

                FontObject font = resourceEntry.Resource.Content as FontObject;

                if (font != null)
                {
                    font.CopyPropertiesFrom(context, resource);
                }

                context.IndirectObjectsQueue.Enqueue(resourceEntry.Resource);
            }

            PdfExporter.WritePendingIndirectObjects(context, writer);
        }

        internal static void WritePendingIndirectObjects(IPdfExportContext context, PdfWriter writer)
        {
            while (context.IndirectObjectsQueue.Count > 0)
            {
                IndirectObject indirectObject = context.IndirectObjectsQueue.Dequeue();
                indirectObject.Write(writer, context);
            }
        }

        internal static void WriteTrailerAndCrossReferenceTable(PdfWriter writer, IPdfExportContext context, DocumentCatalog root, PdfInt previousTablePosition)
        {
            writer.WriteLine(PdfNames.CrossReferenceTableMarker);

            IEnumerable<int> orderedObjectNumbers = context.CrossReferenceCollection.Entries.OrderBy(item => item.Key).Select(item => item.Key);
            IEnumerable<Tuple<int, int>> sectionInfos = CalculateCrossReferenceSectionInfos(orderedObjectNumbers);

            foreach (Tuple<int, int> sectionInfo in sectionInfos)
            {
                int sectionStartId = sectionInfo.Item1;
                int sectionElementsCount = sectionInfo.Item2;

                writer.WriteLine("{0} {1}", sectionStartId, sectionElementsCount);
                int maxObjectNumber = sectionStartId + sectionElementsCount - 1;

                for(int objectNumber = sectionStartId; objectNumber <= maxObjectNumber; objectNumber++)
                {
                    CrossReferenceEntry entry = context.CrossReferenceCollection.GetCrossReferenceEntry(objectNumber);
                    entry.Write(writer, context);
                }
            }

            Trailer trailer = new Trailer();
            trailer.CopyPropertiesFrom(context);
            trailer.Root = root;
            trailer.Prev = previousTablePosition;
            trailer.Write(writer, context);
            writer.WriteLine();
        }

        private static IEnumerable<Tuple<int, int>> CalculateCrossReferenceSectionInfos(IEnumerable<int> orderedObjectNumbers)
        {
            int? previousId = null;
            int currentSectionCount = 0;

            foreach (int id in orderedObjectNumbers)
            {
                if (previousId.HasValue && (id != previousId + 1))
                {
                    yield return new Tuple<int, int>(previousId.Value - currentSectionCount + 1, currentSectionCount);
                    currentSectionCount = 1;
                }
                else
                {
                    currentSectionCount++;
                }

                previousId = id;
            }

            if (previousId.HasValue)
            {
                yield return new Tuple<int, int>(previousId.Value - currentSectionCount + 1, currentSectionCount);
            }
        }

        internal static void WriteCrossReferenceStream(PdfWriter writer, IPdfExportContext context, DocumentCatalog root, PdfInt previousCrossReferencePosition)
        {
            CrossReferenceStream crossReferenceStream = new CrossReferenceStream();
            IndirectObject indirectObject = context.CreateIndirectObject(crossReferenceStream);
            crossReferenceStream.CopyPropertiesFrom(context);
            crossReferenceStream.Root = root;
            crossReferenceStream.Prev = previousCrossReferencePosition;

            // we are writting the pending objects as the create indirect object will add the cross reference stream to a queue.
            PdfExporter.WritePendingIndirectObjects(context, writer);
        }

        internal static void WriteDocumentEnd(long offset, PdfWriter writer, IPdfExportContext context)
        {
            writer.WriteLine(PdfNames.StartXRef);
            writer.WriteLine("{0}", offset);
            writer.WriteDocumentEnd();
        }

        private static Encrypt CreateEncryptionBasedOnPdfSettings(IPdfExportContext context, PdfExportSettings settings)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(settings, "settings");

            Encrypt result = new StandardEncrypt();
            result.CopyPropertiesFrom(context, settings);

            return result;
        }

        private static bool HasSignatureField(IRadFixedDocumentExportContext context)
        {
            bool hasSignature = false;
#if WPF
            foreach (FormField item in context.Document.AcroForm.FormFields)
            {
                if (item.FieldType == FormFieldType.Signature)
                {
                    hasSignature = true;
                    break;
                }
            }
#endif
            return hasSignature;
        }

#if WPF
        internal static void UpdateSignatures(PdfWriter writer, IRadFixedDocumentExportContext context)
        {
            foreach (FormField item in context.Document.AcroForm.FormFields)
            {
                if (item.FieldType == FormFieldType.Signature)
                {
                    SignatureField signatureField = (SignatureField)item;
                    if (signatureField.Signature != null && signatureField.Signature.CanSign)
                    {
                        UpdateSignatureByteRange(signatureField.Name, writer, context);

                        UpdateSignatureContent(signatureField.Signature, writer, context);
                    }
                }
            }
        }

        internal static void UpdateSignatureByteRange(string singatureFieldName, PdfWriter writer, IPdfExportContext context)
        {
            writer.Seek(context.SignatureExportInfo.GetSignaturePositions(singatureFieldName).ByteRangeStartPosition, SeekOrigin.Begin);

            int[] byteRange = context.SignatureExportInfo.ComposeByteRangeArray(singatureFieldName);

            PdfArray array = new PdfArray();
            array.Add(new PdfInt(byteRange[0]));
            array.Add(new PdfInt(byteRange[1]));
            array.Add(new PdfInt(byteRange[2]));
            array.Add(new PdfInt(byteRange[3]));

            // override the inital byte range.
            array.Write(writer, context);

            // overriding the old array not needed symbols with blank space.
            int symbolsCount = (int)(context.SignatureExportInfo.GetSignaturePositions(singatureFieldName).ByteRangeEndPosition - writer.Position);
            for (int i = 0; i < symbolsCount; i++)
            {
                writer.WriteSeparator();
            }
        }

        internal static void UpdateSignatureContent(Signature signature, PdfWriter writer, IPdfExportContext context)
        {
            int[] byteRange = context.SignatureExportInfo.ComposeByteRangeArray(signature.Properties.FieldName);

            List<SourcePart> sourceParts = new List<SourcePart>();
            SourcePart sourcePart1 = new SourcePart(byteRange[0], byteRange[1]);
            SourcePart sourcePart2 = new SourcePart(byteRange[2], byteRange[3]);

            sourceParts.Add(sourcePart1);
            sourceParts.Add(sourcePart2);

            ByteRangeComposer byteRangeComposer = new ByteRangeComposer(new FileSourceStream(writer.OutputStream), sourceParts);
            byte[] byteArrayCompositionForHash = byteRangeComposer.Compose();

            byte[] contents = signature.CalculateContentsPackage(byteArrayCompositionForHash);

            writer.Seek(context.SignatureExportInfo.GetSignaturePositions(signature.Properties.FieldName).ContentStartPosition, SeekOrigin.Begin);

            PdfHexString pdfContents = new PdfHexString(contents);
            pdfContents.Write(writer, context);
        }
#endif
    }
}
