﻿using System;
using System.IO;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export
{
    internal class PdfWriter
    {
        private const byte SpaceToken = (byte)' ';
        private static readonly byte[] newLineToken = new byte[] {(byte)'\r', (byte)'\n' };

        private readonly Stream output;
        private readonly Encoding encoding;

        public PdfWriter(Stream output)
        {
            Guard.ThrowExceptionIfNull(output, "output");

            this.output = output;
            this.encoding = PdfEncoding.Encoding;
        }

        public Stream OutputStream
        {
            get
            {
                return this.output;
            }
        }

        public long Position
        {
            get
            {
                return this.output.Position;
            }
        }

        public long Seek(int offset, SeekOrigin seekOrigin)
        {
            long newPosition = this.output.Seek(offset, seekOrigin);

            return newPosition;
        }

        public void WriteSeparator()
        {
            this.WriteRaw(SpaceToken);
        }

        public void Write(string value)
        {
            byte[] bytes = this.encoding.GetBytes(value);

            this.WriteRaw(bytes);
        }

        public void Write(string value, params object[] args)
        {
            this.Write(string.Format(value, args));
        }

        public virtual void WriteLine()
        {
            this.WriteRaw(newLineToken);
        }

        public void WriteLine(string line)
        {
            this.Write(line);
            this.WriteLine();
        }

        public void WriteLine(string line, params object[] arg)
        {
            this.WriteLine(string.Format(line, arg));
        }

        public void Write(byte[] data)
        {
            this.WriteRaw(data);
        }

        public void WriteDocumentStart()
        {
            this.WriteLine(PdfNames.PdfDocumentHeader);
            this.WriteLine(PdfNames.BinaryMarker);
        }

        public void WriteDocumentEnd()
        {
            this.WriteLine(PdfNames.EndOfFileMarker);
        }

        public void WritePdfName(string name)
        {
            Guard.ThrowExceptionIfNull(name, "name");

            this.Write("{0}{1}", PdfNames.PdfNameIdentifier, name);
        }

        public void WritePdfObjectDescriptor(IPdfExportContext context, PdfObjectDescriptor descriptor)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(descriptor, "descriptor");

            if (!string.IsNullOrEmpty(descriptor.Type))
            {
                this.WritePdfName(PdfNames.Type);
                this.WriteSeparator();
                this.WritePdfName(descriptor.Type);
                this.WriteSeparator();
            }

            if (descriptor.SubType != null)
            {
                this.WritePdfName(descriptor.SubTypeProperty);
                this.WriteSeparator();
                descriptor.SubType.Write(this, context);
                this.WriteSeparator();
            }
        }

        public void WriteIndirectReference(IndirectReference reference)
        {
            this.Write("{0} {1} {2}", reference.ObjectNumber, reference.GenerationNumber, PdfNames.IndirectReferenceMarker);
        }

        private void WriteRaw(byte b)
        {
            this.output.WriteByte(b);
        }

        private void WriteRaw(byte[] data)
        {
            this.output.Write(data, 0, data.Length);
        }
    }
}
