﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Model.Fonts.Encoding;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export.Writers
{
    internal class CMapWriter
    {
        private const int MaxItemsInRange = 100;
        private const string CodeSpaceRange = "codespacerange";
        private const string BfChar = "bfchar";
        private const string BfRange = "bfrange";

        public void Write(PdfWriter writer, IPdfExportContext context, CMapEncoding encoding)
        {
            writer.WriteLine("/CIDInit /ProcSet findresource begin");
            writer.WriteLine("12 dict begin");
            writer.WriteLine("begincmap");
            WriteCidSystemInfo(writer, encoding.CidSystemInfo.Value);
            WriteNameDefinition(writer, "CMapName", encoding.CMapName.Value);
            WriteIntegerDefinition(writer, "CMapType", encoding.CMapType.Value);

            // CodeSpaceRanges should be imported and exported correctly!
            CharCodeRange[] codeSpaceRanges = new CharCodeRange[] { new CharCodeRange(new CharCode(0x0000, 2), new CharCode(0xFFFF, 2)) };
            IList<Tuple<CharCode, string>> bfchars = encoding.CharCodeToUnicodeMappings.ToList();

            // TODO: All string objects should be encrypted with the export context encryption!
            WriteItems(writer, CodeSpaceRange, codeSpaceRanges, (w, item) => { w.WriteLine("<{0}><{1}>", item.Start, item.End); });
            WriteItems(writer, BfChar, bfchars, (w, item) => { w.WriteLine("<{0}><{1}>", item.Item1, BytesHelper.ToHexString(System.Text.Encoding.BigEndianUnicode.GetBytes(item.Item2))); });

            WriteItems(writer, BfRange, encoding.UnicodeRanges, WriteRangeAction);

            writer.WriteLine(@"endcmap CMapName currentdict /CMap defineresource pop end end");
        }

        private static void WriteRangeAction(PdfWriter writer, CharCodeRange charCodeRange)
        {
            if (charCodeRange.IsSequential)
            {
                string unicodeString = BytesHelper.ToHexString(System.Text.Encoding.BigEndianUnicode.GetBytes(charCodeRange.Unicodes[0]));
                writer.WriteLine("<{0}><{1}> <{2}>", charCodeRange.Start, charCodeRange.End, unicodeString);
            }
            else
            {
                writer.WriteLine("<{0}><{1}> [{2}]", charCodeRange.Start, charCodeRange.End, charCodeRange.Unicodes
                    .Select(s => string.Format("<{0}>", BytesHelper.ToHexString(System.Text.Encoding.BigEndianUnicode.GetBytes(s))))
                    .Aggregate((a, b) => a + " " + b));
            }
        }

        private static void WriteNameDefinition(PdfWriter writer, string propertyName, string value)
        {
            WriteDefinition(writer, propertyName, string.Format("/{0}", value));
        }

        private static void WriteIntegerDefinition(PdfWriter writer, string propertyName, int value)
        {
            WriteDefinition(writer, propertyName, value.ToString());
        }

        private static void WriteCidSystemInfo(PdfWriter writer, CidSystemInfo info)
        {
            WriteDefinition(writer, "CIDSystemInfo", string.Format("<< /Registry ({0})/Ordering ({1})/Supplement {2}>>",
                info.Registry, info.Ordering, info.Supplement));
        }

        private static void WriteDefinition(PdfWriter writer, string propertyName, string value)
        {
            writer.WriteLine("/{0} {1} def", propertyName, value);
        }

        private static void WriteItems<T>(PdfWriter writer, string header, IList<T> ranges, Action<PdfWriter, T> writeItemAction)
        {
            int count = ranges.Count;
            int sections = GetRangesCount(count);
            for (int i = 0; i < sections; i++)
            {
                int sectionLength = Math.Min(count, MaxItemsInRange);
                int firstIndex = i * MaxItemsInRange;

                if (firstIndex + sectionLength >= count)
                {
                    sectionLength = count - firstIndex;
                }

                writer.WriteLine("{0} begin{1}", sectionLength, header);
                for (int j = 0; j < sectionLength; j++)
                {
                    T item = ranges[firstIndex + j];
                    writeItemAction(writer, item);
                }

                writer.WriteLine("end{0}", header);
            }
        }

        private static int GetRangesCount(int itemsCount)
        {
            int count = itemsCount / MaxItemsInRange;
            if (itemsCount % MaxItemsInRange > 0)
            {
                count++;
            }

            return count;
        }

    }
}
