﻿using System;
using System.Collections.Generic;
using System.IO;
#if !NETSTANDARD
using System.Windows.Media;
using System.Windows.Media.Imaging;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader.Parsers;
#endif
using Telerik.Windows.Documents.Core.Imaging.Jpeg;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Decoder;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Encoder;
using Telerik.Windows.Documents.Fixed.Model.Extensions;
using Telerik.Windows.Documents.Utilities;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.ColorSpaces;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Filters
{
    /// <summary>
    /// Represents DCT filter.
    /// </summary>
    public class DctDecode : IPdfFilter
    {
        /// <summary>
        /// Gets the name of the PDF filter.
        /// </summary>
        /// <value>The name.</value>
        public string Name
        {
            get
            {
                return PdfFilterNames.DCTDecode;
            }
        }

        /// <summary>
        /// Encodes the specified data.
        /// </summary>
        /// <param name="encodedObject">The encoded object.</param>
        /// <param name="inputData">The input data.</param>
        /// <returns>The encoded data.</returns>
        public byte[] Encode(PdfObject encodedObject, byte[] inputData)
        {
            Guard.ThrowExceptionIfNull(encodedObject, "encodedObject");
            Guard.ThrowExceptionIfNull(inputData, "inputData");

            JpegColorSpace colorSpace = GetJpegColorSpace(encodedObject.ColorSpace);

            return Encode(encodedObject.Width, encodedObject.Height, (float)encodedObject.ExportContext.Settings.ImageQuality, colorSpace, inputData);
        }

        internal static byte[] Encode(int width, int height, float imageQuality, JpegColorSpace colorSpace, byte[] inputData)
        {
            Guard.ThrowExceptionIfNull(inputData, "inputData");

            JpegImage image = new JpegImage(width, height, colorSpace, inputData);
            JpegEncoderParameters parameters = new JpegEncoderParameters();
            parameters.QuantizingQuality = imageQuality;
            JpegEncoder encoder = new JpegEncoder(image, parameters);

            return encoder.Encode();
        }

        /// <summary>
        /// Decodes the specified data.
        /// </summary>
        /// <param name="decodedObject">The decoded object.</param>
        /// <param name="inputData">The input data.</param>
        /// <param name="parms"></param>
        /// <returns></returns>
        public virtual byte[] Decode(PdfObject decodedObject, byte[] inputData, DecodeParameters parms)
        {
            Guard.ThrowExceptionIfNull(decodedObject, "decodedObject");
            Guard.ThrowExceptionIfNull(inputData, "inputData");
#if DEBUG && !NETSTANDARD
            PdfContentExtractor.ExportJpegImage(inputData, "DctDecode");
#endif

#if SILVERLIGHT || NETSTANDARD
            return DctDecode.DecodeWithJpegDecoder(inputData);
#else
            return DctDecode.DecodeWithBitmapImage(decodedObject, inputData);
#endif      
        }

        /// <summary>
        /// Decodes the data with Telerik's internal JpegDecoder class. 
        /// <para>This method was used in previous versions and you can achieve backwards compatibility by using it in the Decode method.</para>
        /// </summary>
        /// <param name="inputData">The input data.</param>
        /// <returns>The decoded data.</returns>
#if SILVERLIGHT
        private static byte[] DecodeWithJpegDecoder(byte[] inputData)
#else
        protected static byte[] DecodeWithJpegDecoder(byte[] inputData)
#endif
        {
            JpegDecoder decoder = new JpegDecoder(inputData);
            JpegImage image = decoder.Decode();

            return image.GetData();
        }

        private static JpegColorSpace GetJpegColorSpace(ColorSpace colorSpace)
        {
            switch (colorSpace)
            {
                case ColorSpace.RGB:
                    return JpegColorSpace.Rgb;
                case ColorSpace.Gray:
                    return JpegColorSpace.Grayscale;
                default:
                    throw new NotSupportedException(string.Format("Not supported colorspace: {0}", colorSpace));
            }
        }

#if !SILVERLIGHT && !NETSTANDARD
        private static byte[] DecodeWithBitmapImage(PdfObject decodedObject, byte[] inputData)
        {
            using (MemoryStream imageStream = new MemoryStream(inputData))
            {
                BitmapSource bitmap = DctDecode.GetBitmap(imageStream, decodedObject);
                ColorSpace colorSpace = decodedObject.ColorSpace;

                byte[] resultData;
                if (colorSpace == ColorSpace.Gray && DctDecode.TryGetGray8DataFromImage(bitmap, out resultData))
                {
                    return resultData;
                }
                else
                {
                    IEnumerable<int> pixels = colorSpace == ColorSpace.CMYK ? ImageSourceExtensions.GetCmykPixels(bitmap) : ImageSourceExtensions.GetPixels(bitmap);
                    int numberOfComponents = DctDecode.GetNumberOfComponents(colorSpace);
                    bool invertCmykColors = DctDecode.GetIfCmykColorsShouldBeInverted(colorSpace, inputData);
                    Func<int, IEnumerable<byte>> getPixelData = DctDecode.GetPixelToDataConverter(colorSpace, invertCmykColors);
                    int numberOfPixels = bitmap.PixelWidth * bitmap.PixelHeight;
                    resultData = new byte[numberOfPixels * numberOfComponents];

                    int dataIndex = 0;
                    foreach (int pixel in pixels)
                    {
                        foreach (byte component in getPixelData(pixel))
                        {
                            resultData[dataIndex] = component;
                            dataIndex++;
                        }
                    }

                    return resultData;
                }
            }
        }

        private static bool GetIfCmykColorsShouldBeInverted(ColorSpace colorSpace, byte[] inputData)
        {
            bool invertCmykColors = false;

            if (colorSpace == ColorSpace.CMYK)
            {
                JpegImageInfo imageInfo;
                JpegHeadersDecoder decoder = new JpegHeadersDecoder(inputData);
                invertCmykColors = decoder.TryDecodeJpegImage(out imageInfo) && imageInfo.HasAdobeInvertedColors;
            }

            return invertCmykColors;
        }

        private static BitmapSource GetBitmap(Stream imageStream, PdfObject decodedObject)
        {
            BitmapImage image = new BitmapImage();
            image.BeginInit();
            image.StreamSource = imageStream;
            image.CreateOptions = BitmapCreateOptions.PreservePixelFormat;
            image.DecodePixelWidth = decodedObject.Width;
            image.EndInit();

            return image;
        }

        private static bool TryGetGray8DataFromImage(BitmapSource source, out byte[] resultData)
        {
            if (source.Format == PixelFormats.Gray8)
            {
                resultData = new byte[source.PixelWidth * source.PixelHeight];
                checked
                {
                    source.CopyPixels(resultData, source.PixelWidth, 0);
                }

                return true;
            }

            resultData = null;
            return false;
        }

        private static int GetNumberOfComponents(ColorSpace colorSpace)
        {
            switch (colorSpace)
            {
                case ColorSpace.CMYK:
                    return 4;
                case ColorSpace.RGB:
                    return 3;
                case ColorSpace.Gray:
                    return 1;
                default:
                    throw new NotSupportedException(string.Format("Not supported colorspace: {0}", colorSpace));
            }
        }

        private static Func<int, IEnumerable<byte>> GetPixelToDataConverter(ColorSpace colorSpace, bool inverseCmykData)
        {
            switch (colorSpace)
            {
                case ColorSpace.CMYK:
                    if (inverseCmykData)
                    {
                        return DctDecode.GetInversedCmykDataFromPixel;
                    }
                    else
                    {
                        return DctDecode.GetCmykDataFromPixel;
                    }
                case ColorSpace.RGB:
                    return GetRgbDataFromPixel;
                case ColorSpace.Gray:
                    return GetGrayDataFromPixel;
                default:
                    throw new NotSupportedException(string.Format("Not supported colorspace: {0}", colorSpace));
            }
        }

        private static IEnumerable<byte> GetCmykDataFromPixel(int pixel)
        {
            yield return (byte)(pixel & 0xFF);
            yield return (byte)((pixel >> 8) & 0xFF);
            yield return (byte)((pixel >> 16) & 0xFF);
            yield return (byte)((pixel >> 24) & 0xFF);
        }

        private static IEnumerable<byte> GetInversedCmykDataFromPixel(int pixel)
        {
            foreach (byte component in DctDecode.GetCmykDataFromPixel(pixel))
            {
                yield return (byte)(255 - component);
            }
        }

        private static IEnumerable<byte> GetGrayDataFromPixel(int pixel)
        {
            yield return (byte)(pixel & 0xFF);
        }

        private static IEnumerable<byte> GetRgbDataFromPixel(int pixel)
        {
            yield return (byte)((pixel >> 16) & 0xFF);
            yield return (byte)((pixel >> 8) & 0xFF);
            yield return (byte)(pixel & 0xFF);
        }
#endif
    }
}
