﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Fixed.Exceptions;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Filters
{
    /// <summary>
    /// Represents the PDF filters manager.
    /// </summary>
    public static class FiltersManager
    {
        private static readonly Dictionary<string, IPdfFilter> filters;

        static FiltersManager()
        {
            filters = new Dictionary<string, IPdfFilter>();

            InitializeDefaultFilters();
        }

        /// <summary>
        /// Registers the specified PDF filter.
        /// </summary>
        /// <param name="filter">The filter.</param>
        public static void RegisterFilter(IPdfFilter filter)
        {
            filters[filter.Name] = filter;
#if !NETSTANDARD
            Old.Pdf.Filters.FiltersManagerOld.RegisterFilter(filter);
#endif
        }

        internal static byte[] Encode(PdfObject encodeObject, byte[] data, PdfArray filters)
        {
            Guard.ThrowExceptionIfNull(encodeObject, "encodeObject");
            Guard.ThrowExceptionIfNull(data, "data");
            Guard.ThrowExceptionIfNull(filters, "filters");

            foreach (PdfName name in filters)
            {
                IPdfFilter filter = GetFilter(name.Value);
                data = filter.Encode(encodeObject, data);
            }

            return data;
        }

        internal static byte[] Decode(PostScriptReader reader, IPdfImportContext context, PdfDictionary dictionary, PdfStreamBase stream, PdfArray filters)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");

            byte[] data = stream.ReadRawPdfData();

            if (filters != null)
            {
                DecodeParameters[] decodeParameters = stream.DecodeParameters;
                PdfObject obj = new PdfObject(reader, context, dictionary);
                string[] filtersArray = null;
                filtersArray = new string[filters.Count];

                for (int i = 0; i < filtersArray.Length; i++)
                {
                    PdfName filterName;
                    if (filters.TryGetElement<PdfName>(reader, context, i, out filterName))
                    {
                        filtersArray[i] = filterName.Value;
                    }
                }

                data = Decode(obj, data, filtersArray, decodeParameters);
            }

            return data;
        }

        internal static byte[] Decode(PdfObject obj, byte[] rawData, string[] filters, DecodeParameters[] decodeParameters)
        {
            byte[] data = rawData;

            if (filters != null)
            {
                for (int i = 0; i < filters.Length; i++)
                {
                    string filterName = filters[i];

                    if (filterName != null)
                    {
                        IPdfFilter filter = GetFilter(filterName);
                        data = filter.Decode(obj, data, decodeParameters == null ? null : decodeParameters[i]);
                    }
                }
            }

            return data;
        }

        private static IPdfFilter GetFilter(string name)
        {
            string filterName = PdfNames.GetNameFromAbbreviation(name);

            IPdfFilter filter;
            if (!filters.TryGetValue(filterName, out filter))
            {
                throw new NotSupportedFilterException(filterName);
            }

            return filter;
        }

        private static void InitializeDefaultFilters()
        {
            RegisterFilter(new FlateDecode());
            RegisterFilter(new DctDecode());
            RegisterFilter(new ASCII85Decode());
            RegisterFilter(new ASCIIHexDecode());
            RegisterFilter(new LZWDecode());
            RegisterFilter(new CCITTFaxDecode());
            RegisterFilter(new RunLengthDecode());
            RegisterFilter(new JBIG2Decode());
        }
    }
}
