﻿/* Copyright 2012 Mozilla Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* Portions of /src/parser.js file as modified by Progress EAD - Copyright © 2016 Progress EAD. All rights reserved.
*/

using System;
using System.Collections.Generic;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Filters
{
    internal partial class LzwDecoder
    {
        private readonly double Log2 = Math.Log(2);

        private LzwDecoderState lzwState;
        private int bitsCached = 0;
        private int cachedData = 0;
        private bool eof;

        private int position = 0;
        private byte[] data;

        public LzwDecoder(byte[] data, int earlyChange)
        {
            this.lzwState = new LzwDecoderState(earlyChange);
            this.data = data;
        }

        public byte[] GetBytes()
        {
            List<byte> bytes = new List<byte>();

            while (!this.eof)
            {
                bytes.AddRange(this.ReadBlock());
            }

            return bytes.ToArray();
        }

        private int? ReadBits(int n)
        {
            int bitsCached = this.bitsCached;
            int cachedData = this.cachedData;

            while (bitsCached < n)
            {
                byte? c = this.GetByte();
                if (c == null)
                {
                    this.eof = true;
                    return null;
                }

                cachedData = (cachedData << 8) | c.Value;
                bitsCached += 8;
            }

            this.bitsCached = (bitsCached -= n);
            this.cachedData = cachedData;

            return (int)((uint)cachedData >> bitsCached) & ((1 << n) - 1);
        }

        private byte? GetByte()
        {
            if (this.position >= this.data.Length)
            {
                return null;
            }

            return this.data[this.position++];
        }

        private List<byte> ReadBlock()
        {
            List<byte> readBytes = new List<byte>();

            int blockSize = 512;

            LzwDecoderState lzwState = this.lzwState;
            if (lzwState == null)
            {
                // eof was found
                return readBytes;
            }

            int earlyChange = lzwState.EarlyChange;
            int nextCode = lzwState.NextCode;
            byte[] dictionaryValues = lzwState.DictionaryValues;
            int[] dictionaryLengths = lzwState.DictionaryLengths;
            int[] dictionaryPrevCodes = lzwState.DictionaryPrevCodes;
            int codeLength = lzwState.CodeLength;
            int prevCode = lzwState.PrevCode;
            byte[] currentSequence = lzwState.CurrentSequence;
            int currentSequenceLength = lzwState.CurrentSequenceLength;

            for (int i = 0; i < blockSize; i++)
            {
                var code = this.ReadBits(codeLength);
                var hasPrev = currentSequenceLength > 0;

                if (code < 256)
                {
                    currentSequence[0] = (byte)code.Value;
                    currentSequenceLength = 1;
                }
                else if (code >= 258)
                {
                    if (code < nextCode)
                    {
                        currentSequenceLength = dictionaryLengths[code.Value];

                        for (int j = currentSequenceLength - 1, q = code.Value; j >= 0; j--)
                        {
                            currentSequence[j] = dictionaryValues[q];
                            q = dictionaryPrevCodes[q];
                        }
                    }
                    else
                    {
                        currentSequence[currentSequenceLength++] = currentSequence[0];
                    }
                }
                else if (code == 256)
                {
                    codeLength = 9;
                    nextCode = 258;
                    currentSequenceLength = 0;
                    continue;
                }
                else
                {
                    this.eof = true;
                    this.lzwState = null;
                    break;
                }

                if (hasPrev)
                {
                    dictionaryPrevCodes[nextCode] = prevCode;
                    dictionaryLengths[nextCode] = dictionaryLengths[prevCode] + 1;
                    dictionaryValues[nextCode] = currentSequence[0];
                    nextCode++;

                    if (((nextCode + earlyChange) & (nextCode + earlyChange - 1)) == 0)
                    {
                        codeLength = (int) Math.Min(Math.Log(nextCode + earlyChange) / Log2 + 1, 12) | 0;
                    }
                }
                prevCode = code.Value;

                for (int k = 0; k < currentSequenceLength; k++)
                {
                    readBytes.Add(currentSequence[k]);
                }
            }

            lzwState.NextCode = nextCode;
            lzwState.CodeLength = codeLength;
            lzwState.PrevCode = prevCode;
            lzwState.CurrentSequenceLength = currentSequenceLength;

            return readBytes;
        }
    }
}
