﻿#if !NETSTANDARD
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.ColorSpaces;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.ColorSpaces;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Filters
{
    /// <summary>
    /// Represents a PDF object.
    /// </summary>
    public class PdfObject
    {
        private const string BitsPerComponentName = "BitsPerComponent";
        private const string WidthName = "Width";
        private const string HeightName = "Height";

#if !NETSTANDARD
        private readonly Old.Pdf.DocumentModel.Data.PdfDictionaryOld oldDictionary;
#endif
        private readonly PdfDictionary dictionary;
        private readonly PostScriptReader reader;
        private readonly IPdfImportContext importContext;
        private readonly IPdfExportContext exportContext;
        private readonly ColorSpace colorSpace;
        private readonly int width;
        private readonly int height;

        /// <summary>
        /// This constructor is used for exporting stream objects.
        /// </summary>
        internal PdfObject(IPdfExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            this.exportContext = context;
        }

        /// <summary>
        /// This constructor is used for exporting image stream objects.
        /// </summary>
        internal PdfObject(IPdfExportContext context, int width, int height, ColorSpace colorSpace)
            : this(context)
        {
            this.width = width;
            this.height = height;
            this.colorSpace = colorSpace;
        }

        /// <summary>
        /// This constructor is used for decoding <see cref="Fixed.Model.Resources.ImageSource"/> objects.
        /// </summary>
        internal PdfObject(int width, int height, ColorSpace colorSpace)
        {
            this.width = width;
            this.height = height;
            this.colorSpace = colorSpace;
        }

        /// <summary>
        /// This constructor is used for decoding stream objects on RadPdfProcessing import.
        /// </summary>
        internal PdfObject(PostScriptReader reader, IPdfImportContext context, PdfDictionary dictionary)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(dictionary, "dictionary");

            this.reader = reader;
            this.importContext = context;
            this.dictionary = dictionary;
        }

#if !NETSTANDARD
        /// <summary>
        /// This constructor is used for decoding stream objects on RadPdfViewer import.
        /// </summary>
        internal PdfObject(Old.Pdf.DocumentModel.Data.PdfDictionaryOld dictionary)
        {
            Guard.ThrowExceptionIfNull(dictionary, "dictionary");

            this.oldDictionary = dictionary;
        }
#endif

        /// <summary>
        /// Gets the color space.
        /// </summary>
        /// <value>The color space.</value>
        public ColorSpace ColorSpace
        {
            get
            {
                return this.GetColorSpace();
            }
            set
            {
                this.SetColorSpace(value);
            }
        }

        /// <summary>
        /// Gets the bits per component.
        /// </summary>
        /// <value>The bits per component.</value>
        public int BitsPerComponent
        {
            get
            {
                int bpc;
                if (this.TryGetIntValue(BitsPerComponentName, out bpc))
                {
                    return bpc;
                }

                return 0;
            }
            internal set
            {
                this.SetIntValue(BitsPerComponentName, value);
            }
        }

        /// <summary>
        /// Gets the width.
        /// </summary>
        /// <value>The width.</value>
        public int Width
        {
            get
            {
                int width;
                if (!this.TryGetIntValue(WidthName, out width))
                {
                    width = this.width;
                }

                return width;
            }
            internal set
            {
                this.SetIntValue(WidthName, value);
            }
        }

        /// <summary>
        /// Gets the height.
        /// </summary>
        /// <value>The height.</value>
        public int Height
        {
            get
            {
                int height;
                if (!this.TryGetIntValue(HeightName, out height))
                {
                    height = this.height;
                }

                return height;
            }
            internal set
            {
                this.SetIntValue(HeightName, value);
            }
        }

        internal IPdfExportContext ExportContext
        {
            get
            {
                return this.exportContext;
            }
        }

#if !NETSTANDARD
        internal PdfContentManager ContentManager
        {
            get
            {
                if (this.oldDictionary != null)
                {
                    return this.oldDictionary.ContentManager;
                }

                return null;
            }
        }
#endif

        /// <summary>
        /// Returns <see cref="ColorSpaceObject"/> from this.dictionary when using PdfProcessing model;
        /// and ColorSpaceOld when using PdfViewer model.
        /// </summary>
        internal T GetColorSpaceInternal<T>()
            where T : class
        {
            if (this.dictionary != null)
            {
                ColorSpaceObject colorSpace;
                if (this.dictionary.TryGetElement<ColorSpaceObject>(this.reader, this.importContext, PdfNames.ColorSpace, out colorSpace))
                {
                    return colorSpace as T;
                }
            }
#if !NETSTANDARD
            if (this.oldDictionary != null)
            {
                if (this.oldDictionary.ContainsKey(PdfNames.ColorSpace))
                {
                    return this.oldDictionary.GetElement<ColorSpaceOld>(PdfNames.ColorSpace, Old.Pdf.DocumentModel.PdfConverters.Converters.ColorSpaceConverter) as T;
                }
            }
#endif
            return default(T);
        }

        /// <summary>
        /// Returns <see cref="ColorSpaceObject"/> from this.dictionary when using PdfProcessing;
        /// and <see cref="ColorSpaceObject"/> from ColorSpaceOld when using PdfViewer.
        /// </summary>
        internal ColorSpaceObject GetColorSpaceObject()
        {
            if (this.dictionary != null)
            {
                ColorSpaceObject colorSpace;
                if (this.dictionary.TryGetElement<ColorSpaceObject>(this.reader, this.importContext, PdfNames.ColorSpace, out colorSpace))
                {
                    return colorSpace;
                }
            }

#if !NETSTANDARD
            if (this.oldDictionary != null)
            {
                if (this.oldDictionary.ContainsKey(PdfNames.ColorSpace))
                {
                    var colorSpaceOld = this.oldDictionary.GetElement<ColorSpaceOld>(PdfNames.ColorSpace, Old.Pdf.DocumentModel.PdfConverters.Converters.ColorSpaceConverter);

                    return OldToProcessingTranslator.GetColorSpaceObjectFromColorSpaceOld(colorSpaceOld);
                }
            }
#endif
            return null;
        }

        private bool TryGetIntValue(string index, out int value)
        {
            value = 0;
            if (this.dictionary != null)
            {
                PdfInt result;
                this.dictionary.TryGetElement<PdfInt>(reader, importContext, index, out result);
                value = result.Value;

                return true;
            }

#if !NETSTANDARD
            if (this.oldDictionary != null)
            {
                return this.oldDictionary.TryGetInt(index, out value);
            }
#endif
            return false;
        }

        private void SetIntValue(string index, int value)
        {
            if (this.dictionary != null)
            {
                this.dictionary[index] = new PdfInt(value);
            }

#if !NETSTANDARD
            if (this.oldDictionary != null)
            {
                this.oldDictionary[index] = new Old.Pdf.DocumentModel.Data.PdfIntOld(this.oldDictionary.ContentManager, value);
            }
#endif
        }

        private ColorSpace GetColorSpace()
        {
            if (this.dictionary != null)
            {
                ColorSpaceObject colorSpace = this.GetColorSpaceInternal<Pdf.Model.Elements.ColorSpaces.ColorSpaceObject>();
                if (colorSpace != null)
                {
                    return colorSpace.Public;
                }
            }

#if !NETSTANDARD
            if (this.oldDictionary != null)
            {
                Old.Pdf.DocumentModel.Core.ColorSpaces.ColorSpaceOld colorSpace = this.GetColorSpaceInternal<Old.Pdf.DocumentModel.Core.ColorSpaces.ColorSpaceOld>();
                if (colorSpace != null)
                {
                    return colorSpace.Type;
                }
            }
#endif
            return this.colorSpace;
        }

        private void SetColorSpace(ColorSpace value)
        {
            if (this.dictionary != null)
            {
                this.dictionary[PdfNames.ColorSpace] = Model.Elements.ColorSpaces.ColorSpaceManager.CreateColorSpaceObject(value);
            }

#if !NETSTANDARD
            if (this.oldDictionary != null)
            {
                this.oldDictionary[PdfNames.ColorSpace] = Old.Pdf.DocumentModel.Core.ColorSpaces.ColorSpaceOld.CreateColorSpace(this.oldDictionary.ContentManager, value);
            }
#endif
        }
    }
}
