﻿using System;
using System.Collections.Generic;
using System.IO;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser.Keywords;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Annotations;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Destinations;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Forms;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Encryption;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import
{
    internal abstract class BaseImportContext : IPdfImportContext
    {
        private readonly Dictionary<FormFieldNode, FormField> nodeToField;
        private readonly Dictionary<FormFieldNode, List<WidgetObject>> nodeToWidgets;
        private readonly Dictionary<WidgetObject, Widget> widgetObjectToWidget;
        private readonly Dictionary<IndirectReference, IPdfPrimitive> indirectObjects;
        private readonly Dictionary<PageTreeNode, PdfDictionary> pagesToPdfDictionaries;
        private readonly Dictionary<string, NamedDestinationObject> nameToDestination;
        private readonly Stack<IndirectReference> indirectReferencesStack;
        private readonly CrossReferenceCollection crossReferences;
        private readonly BeginEndCounter innerStreamCounter;
        private readonly PdfImportSettings settings;
        private readonly List<Page> pages;
        private PostScriptReader reader;

        protected BaseImportContext(PdfImportSettings settings)
        {
            Guard.ThrowExceptionIfNull(settings, "settings");

            this.settings = settings;
            this.nodeToField = new Dictionary<FormFieldNode, FormField>();
            this.nodeToWidgets = new Dictionary<FormFieldNode, List<WidgetObject>>();
            this.widgetObjectToWidget = new Dictionary<WidgetObject, Widget>();
            this.crossReferences = new CrossReferenceCollection();
            this.indirectObjects = new Dictionary<IndirectReference, IPdfPrimitive>();
            this.pagesToPdfDictionaries = new Dictionary<PageTreeNode, PdfDictionary>();
            this.nameToDestination = new Dictionary<string, NamedDestinationObject>();
            this.indirectReferencesStack = new Stack<IndirectReference>();
            this.innerStreamCounter = new BeginEndCounter();
            this.pages = new List<Page>();
        }

        public CrossReferenceCollection CrossReferences
        {
            get
            {
                return this.crossReferences;
            }
        }

        public PdfImportSettings ImportSettings
        {
            get
            {
                return this.settings;
            }
        }

        public ReferenceProperty<DocumentCatalog> Root
        {
            get;
            set;
        }

        public byte[] DocumentId
        {
            get;
            set;
        }

        public Encrypt Encryption
        {
            get;
            set;
        }

        public List<Page> Pages
        {
            get
            {
                return this.pages;
            }
        }

        public IndirectReference CurrentIndirectReference
        {
            get
            {
                if (this.indirectReferencesStack.Count == 0)
                {
                    return null;
                }

                return this.indirectReferencesStack.Peek();
            }
        }

        public int StartXRefOffset { get; private set; }

        public bool StartsWithCrossReferenceTable { get; private set; }

        protected bool IsReadingStreamInnerContent
        {
            get
            {
                return this.innerStreamCounter.IsUpdateInProgress;
            }
        }

        public PostScriptReader Reader
        {
            get
            {
                return this.reader;
            }
        }

        public void RegisterIndirectObject(IndirectReference reference, IPdfPrimitive primitive)
        {
            Guard.ThrowExceptionIfNull(reference, "reference");
            Guard.ThrowExceptionIfNull(primitive, "primitive");

            this.indirectObjects[reference] = primitive;
        }

        public bool TryGetIndirectObject(IndirectReference reference, out IPdfPrimitive primitive)
        {
            Guard.ThrowExceptionIfNull(reference, "reference");

            return this.indirectObjects.TryGetValue(reference, out primitive);
        }

        public IndirectObject ReadIndirectObject(IndirectReference reference)
        {
            Guard.ThrowExceptionIfNull(reference, "reference");


            using (this.reader.Reader.BeginReadingBlock())
            {
                IndirectObject result;
                CrossReferenceEntry entry;
                this.indirectReferencesStack.Push(reference);

                if (this.CrossReferences.TryGetCrossReferenceEntry(reference, out entry))
                {
                    switch (entry.Type)
                    {
                        case CrossReferenceEntryType.Used:
                            result = this.ReadUsedIndirectObject(entry);
                            break;
                        case CrossReferenceEntryType.Compressed:
                            result = this.ReadCompressedIndirectObject(reference, entry);
                            break;
                        default:
                            result = this.ReadFreeIndirectObject(reference);
                            break;
                    }

                    if (result.Reference != reference)
                    {
                        result = this.ReadFreeIndirectObject(reference);
                    }
                }
                else
                {
                    result = this.ReadFreeIndirectObject(reference);
                }

                this.indirectReferencesStack.Pop();

                return result;
            }

        }

        public void MapWidgets(WidgetObject widgetObject, Widget widget)
        {
            this.widgetObjectToWidget[widgetObject] = widget;
        }

        public bool TryGetWidget(WidgetObject widgetObject, out Widget widget)
        {
            return this.widgetObjectToWidget.TryGetValue(widgetObject, out widget);
        }

        public void MapFields(FormFieldNode node, FormField field)
        {
            this.nodeToField[node] = field;
        }

        public bool TryGetField(FormFieldNode node, out FormField field)
        {
            return this.nodeToField.TryGetValue(node, out field);
        }

        public void AddWidgetParent(WidgetObject widgetObject, FormFieldNode parent)
        {
            widgetObject.FormField = parent;

            List<WidgetObject> childWidgets;
            if (!this.nodeToWidgets.TryGetValue(parent, out childWidgets))
            {
                childWidgets = new List<WidgetObject>();
                this.nodeToWidgets.Add(parent, childWidgets);
            }

            childWidgets.Add(widgetObject);
        }

        public IEnumerable<WidgetObject> GetChildWidgets(FormFieldNode parent)
        {
            List<WidgetObject> childWidgets;
            if (this.nodeToWidgets.TryGetValue(parent, out childWidgets))
            {
                foreach (WidgetObject widget in childWidgets)
                {
                    yield return widget;
                }
            }
        }

        public void BeginImport(Stream pdfFileStream)
        {
            Guard.ThrowExceptionIfNull(pdfFileStream, "pdfFileStream");
            Guard.ThrowExceptionIfNotNull(this.reader, "this.reader");

            PdfImporter.ValidateInputStreamCanReadAndSeek(pdfFileStream);

            if (!PdfImporter.SeekToStartXRef(pdfFileStream))
            {
                throw new NotSupportedException("StartXRef keyword cannot be found.");
            }

            this.reader = new PostScriptReader(pdfFileStream, new KeywordCollection());
            this.PrepareCrossReferencesForImport();

            this.BeginImportOverride();
        }

        public IDisposable BeginImportOfStreamInnerContent()
        {
            return this.innerStreamCounter.Begin();
        }

        public byte[] DecryptStream(IndirectReference reference, byte[] data)
        {
            Guard.ThrowExceptionIfNull(data, "data");

            if (!this.ShouldDecryptStream(reference))
            {
                return data;
            }

            return this.Encryption.DecryptStream(reference.ObjectNumber, reference.GenerationNumber, data);
        }

        public byte[] DecryptString(byte[] data)
        {
            Guard.ThrowExceptionIfNull(data, "data");

            if (!this.ShouldDecryptString())
            {
                return data;
            }

            return this.Encryption.DecryptString(this.CurrentIndirectReference.ObjectNumber, this.CurrentIndirectReference.GenerationNumber, data);
        }

        public bool TryGetNestedPrimitive<T>(IPdfPrimitive primitive, IEnumerable<string> dictionaryKeys, out T result)
            where T : class, IPdfPrimitive
        {
            IPdfPrimitive currentPrimitive = primitive;

            if (currentPrimitive.Type == PdfElementType.IndirectReference)
            {
                IndirectReference reference = (IndirectReference)currentPrimitive;
                IndirectObject indirectObject = this.ReadIndirectObject(reference);
                currentPrimitive = indirectObject.Content;
            }

            foreach (string key in dictionaryKeys)
            {
                if (currentPrimitive.Type == PdfElementType.Dictionary)
                {
                    PdfDictionary dictionary = (PdfDictionary)currentPrimitive;

                    if (!dictionary.ContainsKey(key))
                    {
                        result = null;
                        return false;
                    }

                    currentPrimitive = dictionary[key];

                    if (currentPrimitive.Type == PdfElementType.IndirectReference)
                    {
                        IndirectReference reference = (IndirectReference)currentPrimitive;
                        IndirectObject indirectObject = this.ReadIndirectObject(reference);
                        currentPrimitive = indirectObject.Content;
                    }
                }
                else
                {
                    result = null;
                    return false;
                }
            }

            result = currentPrimitive as T;
            return result != null;
        }

        public void MapPageToPdfDictionary(PageTreeNode page, PdfDictionary dictionary)
        {
            this.pagesToPdfDictionaries[page] = dictionary;
        }

        public void UnmapPdfDictionary(PageTreeNode page)
        {
            this.pagesToPdfDictionaries.Remove(page);
        }

        public bool TryGetPdfDictionary(PageTreeNode page, out PdfDictionary dictionary)
        {
            return this.pagesToPdfDictionaries.TryGetValue(page, out dictionary);
        }

        public void MapNameToNamedDestination(string name, NamedDestinationObject destination)
        {
            this.nameToDestination[name] = destination;
        }

        public bool TryGetNamedDestination(string name, out NamedDestinationObject destination)
        {
            return this.nameToDestination.TryGetValue(name, out destination);
        }

        protected abstract void BeginImportOverride();

        internal static int ReadStartXRefOffset(PostScriptReader reader, IPdfImportContext context)
        {
            reader.Reader.ReadLine(); // Skip startxref word
            PdfInt offset = reader.Read<PdfInt>(context, PdfElementType.PdfInt);

            if (offset == null)
            {
                throw new NotSupportedException("Offset to XRef cannot be found.");
            }

            return offset.Value;
        }

        private void PrepareCrossReferencesForImport()
        {
            this.StartXRefOffset = ReadStartXRefOffset(this.Reader, this);
            this.Reader.Reader.Seek(this.StartXRefOffset, SeekOrigin.Begin);
            this.StartsWithCrossReferenceTable = this.CrossReferences.PeekCrossReferenceTableMarker(this.Reader, this);
            this.CrossReferences.Read(this.Reader, this);

            if (this.Encryption != null)
            {
                Encrypt.ValidateEncryptionAlgorithmValue(this.Encryption.Algorithm.Value);

                if (!this.Encryption.AuthenticateUserPassword(this.DocumentId, this.settings.GetUserPassword()))
                {
                    throw new InvalidOperationException("Password is not correct.");
                }
            }
        }

        private bool ShouldDecryptStream(IndirectReference reference)
        {
            return !this.IsReadingStreamInnerContent && this.Encryption != null && reference != null;
        }

        private bool ShouldDecryptString()
        {
            return !this.IsReadingStreamInnerContent && this.Encryption != null && this.CurrentIndirectReference != null;
        }

        private IndirectObject ReadCompressedIndirectObject(IndirectReference reference, CrossReferenceEntry entry)
        {
            int objectStreamId = (int)entry.Field1;
            IndirectReference objectStreamReference = new IndirectReference(objectStreamId, 0);
            PdfObjectDescriptor descriptor = PdfObjectDescriptors.GetPdfObjectDescriptor<ObjectStream>();
            ObjectStream objectStream = (ObjectStream)descriptor.Converter.Convert(typeof(ObjectStream), this.reader, this, objectStreamReference);
            IPdfPrimitive result = objectStream.ReadObjectContent(entry, this);

            return new IndirectObject(reference, result);
        }

        private IndirectObject ReadUsedIndirectObject(CrossReferenceEntry entry)
        {
            long offset = entry.Field1;
            this.reader.Reader.Seek(offset, System.IO.SeekOrigin.Begin);
            IndirectObject result = this.reader.Read<IndirectObject>(this, PdfElementType.IndirectObject);

            return result;
        }


        private IndirectObject ReadFreeIndirectObject(IndirectReference reference)
        {
            IndirectObject result = new IndirectObject(reference, PdfNull.Instance);

            return result;
        }
    }
}
