﻿using System;
using System.Collections.Generic;
using System.IO;
#if NETSTANDARD
using Matrix = Telerik.Documents.Primitives.Matrix;
#else
using Matrix = System.Windows.Media.Matrix;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser.Keywords;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Operators;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Collections;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import
{
    internal class ContentStreamInterpreter : IOperandsInterpreter
    {
        private readonly Stack<GraphicsState> graphicsStateStack;
        private readonly Stack<TextState> textStateStack;
        private readonly OperandsCollection operands;
        private readonly IPdfContentImportContext context;
        private readonly PostScriptReader reader;

        public ContentStreamInterpreter(Stream stream, IPdfContentImportContext context)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");
            Guard.ThrowExceptionIfNull(context, "context");

            this.context = context;

            this.graphicsStateStack = new Stack<GraphicsState>();

            this.textStateStack = new Stack<TextState>();
            this.operands = new OperandsCollection();

            this.PathFigures = new PathFigureCollection();

            this.reader = new PostScriptReader(stream, new ContentStreamKeywordCollection());
        }

        public OperandsCollection Operands
        {
            get
            {
                return this.operands;
            }
        }

        public Matrix TextRenderingMatrix
        {
            get
            {
                var textMatrix = this.TextState.TextMatrix;
                var currentTransformationMatrix = this.GraphicState.Matrix;

                return textMatrix.MultiplyBy(currentTransformationMatrix);
            }
        }

        public GraphicsState GraphicState
        {
            get
            {
                return this.graphicsStateStack.Peek();
            }
        }

        public TextState TextState
        {
            get
            {
                return this.textStateStack.Peek();
            }
        }

        public PostScriptReader Reader
        {
            get
            {
                return this.reader;
            }
        }

        public PathFigure CurrentPathFigure { get; set; }

        public PathFigureCollection PathFigures { get; private set; }

        public void Execute()
        {
            IPdfPrimitive[] primitives = this.ReadContentPrimitives();

            for (int i = 0; i < primitives.Length; i++)
            {
                IPdfPrimitive primitive = primitives[i];
                if (primitive != null)
                {
                    if (primitive.Type == PdfElementType.Operator)
                    {
                        ContentStreamOperator oper = (ContentStreamOperator)primitive;

                        try
                        {
                            oper.Execute(this, this.context);
                        }
                        catch (NotImplementedException)
                        {
#if DEBUG
                            throw;
#else
                            this.Operands.Clear();
#endif
                        }
                    }
                    else
                    {
                        this.Operands.AddLast(primitive);
                    }
                }
            }
        }

        public void SaveGraphicState()
        {
            this.graphicsStateStack.Push(new GraphicsState(this.GraphicState));
            this.textStateStack.Push(new TextState(this.TextState));
        }

        public void RestoreGraphicState()
        {
            this.graphicsStateStack.Pop();
            this.textStateStack.Pop();
        }

        public GeometryBase CalculateGeometry()
        {
            var pathGeometry = new PathGeometry();
            foreach (var figure in this.PathFigures)
            {
                if (figure.Segments.Count != 0)
                {
                    pathGeometry.Figures.Add(figure);
                }
            }

            return pathGeometry;
        }

        public void ApplyTextProperties(TextPropertiesOwner textProperties)
        {
            if (this.TextState.Font == null)
            {
                textProperties.Font = FontsRepository.Helvetica;
            }
            else
            {
                textProperties.Font = this.context.Owner.GetFont(this.Reader, this.TextState.Font);
            }

            textProperties.CharacterSpacing = this.TextState.CharacterSpacing;
            textProperties.WordSpacing = this.TextState.WordSpacing;
            textProperties.TextRise = this.TextState.TextRise;
            textProperties.FontSize = this.TextState.FontSize;
            textProperties.HorizontalScaling = this.TextState.HorizontalScaling;
            textProperties.RenderingMode = this.TextState.RenderingMode;
            textProperties.TextMatrix = this.TextState.TextMatrix;
            this.ApplyGeometryProperties(textProperties.GeometryProperties);
        }

        public void ApplyGeometryProperties(GeometryPropertiesOwner geometryProperties)
        {
            if (geometryProperties.IsFilled)
            {
                geometryProperties.Fill = this.GraphicState.CalculateFillColor(this.Reader, this.context);
            }

            if (geometryProperties.IsStroked)
            {
                geometryProperties.Stroke = this.GraphicState.CalculateStrokeColor(this.Reader, this.context);
                geometryProperties.StrokeLineCap = this.GraphicState.StrokeLineCap;
                geometryProperties.StrokeLineJoin = this.GraphicState.StrokeLineJoin;
                geometryProperties.StrokeThickness = this.GraphicState.StrokeLineWidth;
                geometryProperties.StrokeDashArray = this.GraphicState.StrokeDashArray;
                geometryProperties.StrokeDashOffset = this.GraphicState.StrokeDashOffset;
                geometryProperties.MiterLimit = this.GraphicState.MiterLimit;
            }
        }

        public void ClearGeometry()
        {
            this.PathFigures.Clear();
            this.CurrentPathFigure = null;
        }

        public void EnsurePathFigure()
        {
            if (this.CurrentPathFigure == null)
            {
                this.CurrentPathFigure = this.PathFigures.AddPathFigure();
            }
        }

        private IPdfPrimitive[] ReadContentPrimitives()
        {
            using (this.context.Owner.BeginImportOfStreamInnerContent())
            {
                this.Reader.Reader.Seek(0, SeekOrigin.Begin);
                this.graphicsStateStack.Push(new GraphicsState());
                this.textStateStack.Push(new TextState());
                IPdfPrimitive[] primitives = this.Reader.Read(this.context.Owner);

                return primitives;
            }
        }
    }
}