﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import
{
    internal class LoadOnDemandPagesCacheManager : IPagesCacheManager
    {
        private readonly IPagesContentCache pagesContentCache;
        private readonly IPageContentManager pageContentManager;
        private readonly object lockObject;
        private readonly Dictionary<RadFixedPage, int> pageToLoadingRequestsCount;

        public LoadOnDemandPagesCacheManager(IPagesContentCache pagesContentCache, IPageContentManager contentManager)
        {
            this.pagesContentCache = pagesContentCache;
            this.pageContentManager = contentManager;

            this.lockObject = new object();
            this.pageToLoadingRequestsCount = new Dictionary<RadFixedPage, int>();
        }

        public void SavePage(RadFixedPage page)
        {
            this.pagesContentCache.CachePageContent(page);
        }

        public IDisposable BeginUsingPage(RadFixedPage page)
        {
            lock (this.lockObject)
            {
                int currentRequestsCount;
                if (!this.pageToLoadingRequestsCount.TryGetValue(page, out currentRequestsCount))
                {
                    currentRequestsCount = 0;
                }

                bool loadPageContent = this.pageContentManager.IsPageContentEmpty(page) && currentRequestsCount == 0;
                if (loadPageContent)
                {
                    try
                    {
                        this.LoadPage(page);
                    }
                    catch
                    {
                        this.pageContentManager.UnloadPageContent(page);
                        throw;
                    }
                }

                this.pageToLoadingRequestsCount[page] = currentRequestsCount + 1;
            }

            return new DisposableObject(() => this.EndUsingPage(page));
        }

        public void Clear()
        {
            this.pagesContentCache.Clear();
        }

        private void EndUsingPage(RadFixedPage page)
        {

            lock (this.lockObject)
            {
                this.pageToLoadingRequestsCount[page] -= 1;

                if (this.pageToLoadingRequestsCount[page] == 0)
                {
                    this.pageContentManager.UnloadPageContent(page);
                }
            }
        }

        private void LoadPage(RadFixedPage page)
        {
            bool isCached = this.pagesContentCache.TryLoadCachedPageContent(page);

            if (!isCached)
            {
                this.pageContentManager.LoadPageContent(page);
            }
        }
    }
}
