﻿using System;
using System.Collections.Generic;
using System.Windows;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Fonts;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Forms;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.GraphicsState;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Objects;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Resources;
#if !NETSTANDARD
using Matrix = System.Windows.Media.Matrix;
#else
using Telerik.Documents.Primitives;
using Matrix = Telerik.Documents.Primitives.Matrix;
#endif
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import
{
    internal class RadFixedDocumentImportContext : BaseImportContext, IRadFixedDocumentImportContext
    {
        private static readonly double PointToDip = Media.Unit.PointToDip(1);
        private readonly RadFixedDocument document;
        private readonly Dictionary<FontObject, FontBase> fonts;
        private readonly Dictionary<ImageXObject, ImageSource> imageSources;
        private readonly Dictionary<FormXObject, FormSource> formSources;
        private readonly Dictionary<ExtGStateObject, ExtGState> extGStates;
        private readonly Dictionary<Page, RadFixedPage> pagesToFixedPages;
        private readonly Dictionary<RadFixedPage, Page> fixedPagesToPages;
        private readonly Dictionary<Page, Matrix> pdfPointToDipTransformations;
        private IPdfContentImportContext acroFormContentImportContext;

        public RadFixedDocumentImportContext(PdfImportSettings settings)
            : base(settings)
        {
            this.document = new RadFixedDocument();
            this.fonts = new Dictionary<FontObject, FontBase>();
            this.imageSources = new Dictionary<ImageXObject, ImageSource>();
            this.formSources = new Dictionary<FormXObject, FormSource>();
            this.pagesToFixedPages = new Dictionary<Page, RadFixedPage>();
            this.fixedPagesToPages = new Dictionary<RadFixedPage, Page>();
            this.pdfPointToDipTransformations = new Dictionary<Page, Matrix>();
            this.extGStates = new Dictionary<ExtGStateObject, ExtGState>();
        }

        public RadFixedDocument Document
        {
            get
            {
                return this.document;
            }
        }
        
        public FontBase GetFont(PostScriptReader reader, FontObject font)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(font, "font");

            FontBase result;
            if (!this.fonts.TryGetValue(font, out result))
            {
                result = font.ToFont(reader, this);
                this.fonts[font] = result;
            }

            return result;
        }

        public FormSource GetFormSource(PostScriptReader reader, FormXObject form)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(form, "form");

            FormSource formSource;
            if (!this.formSources.TryGetValue(form, out formSource))
            {
                formSource = form.ToFormSource(reader, this);
                this.formSources[form] = formSource;
            }

            return formSource;
        }

        public ImageSource GetImageSource(PostScriptReader reader, ImageXObject image)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(image, "image");

            ImageSource imageSource;
            if (!this.imageSources.TryGetValue(image, out imageSource))
            {
                imageSource = image.ToImageSource();
                this.imageSources[image] = imageSource;
            }

            return imageSource;
        }

        public ExtGState GetExtGState(PostScriptReader reader, ExtGStateObject extGState)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(extGState, "extGState");

            ExtGState state;
            if (!this.extGStates.TryGetValue(extGState, out state))
            {
                state = extGState.ToExtGState();
                this.extGStates[extGState] = state;
            }

            return state;
        }

        public void MapPages(Page page, RadFixedPage fixedPage)
        {
            this.fixedPagesToPages[fixedPage] = page;
            this.pagesToFixedPages[page] = fixedPage;
            this.pdfPointToDipTransformations[page] = this.CalculatePdfPointToDipTransformation(page);
        }

        public RadFixedPage GetFixedPage(Page page)
        {
            return this.pagesToFixedPages[page];
        }

        public Page GetPage(RadFixedPage fixedPage)
        {
            return this.fixedPagesToPages[fixedPage];
        }

        public Matrix GetPdfPointToDipTransformation(Page page)
        {
            return this.pdfPointToDipTransformations[page];
        }

        public IPdfContentImportContext GetAcroFormContentImportContext()
        {
            if (this.acroFormContentImportContext == null)
            {
                AcroFormObject acroForm = this.Root.GetValue().AcroForm;

                if (acroForm != null)
                {
                    this.acroFormContentImportContext = new PdfContentImportContext(this, acroForm, new EmptyContentRoot());
                }
            }

            return this.acroFormContentImportContext;
        }

        protected override void BeginImportOverride()
        {
            DocumentCatalog catalog = this.Root.GetValue();
            catalog.CopyPropertiesTo(this.Reader, this);
        }

        private Matrix CalculatePdfPointToDipTransformation(Page page)
        {
            Rect rect = page.MediaBox.ToRect(this.Reader, this);

            return new Matrix(PointToDip, 0, 0, -PointToDip, 0, Media.Unit.PointToDip(rect.Height));
        }
    }
}
