﻿using System;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Converters
{
    internal class Converter : IConverter
    {
        public IPdfPrimitive Convert(Type type, PostScriptReader reader, IPdfImportContext context, IPdfPrimitive value)
        {
            Guard.ThrowExceptionIfNull(type, "type");

            if (value == null)
            {
                return null;
            }

            switch (value.Type)
            {
                case PdfElementType.IndirectReference:

                    IPdfPrimitive result;
                    IndirectReference reference = (IndirectReference)value;
                    if (!context.TryGetIndirectObject(reference, out result))
                    {
                        result = this.ConvertFromIndirectReference(type, reader, context, (IndirectReference)value);
                        context.RegisterIndirectObject(reference, result);
                    }

                    return result.Type == PdfElementType.Null ? null : result;

                case PdfElementType.IndirectObject:

                    return this.Convert(type, reader, context, ((IndirectObject)value).Content);

                case PdfElementType.Array:

                    return this.ConvertFromArray(type, reader, context, (PdfArray)value);

                case PdfElementType.Stream:

                    return this.ConvertFromStream(type, reader, context, (PdfStream)value);

                case PdfElementType.Dictionary:

                    return this.ConvertFromDictionary(type, reader, context, (PdfDictionary)value);

                case PdfElementType.String:

                    return this.ConvertFromString(type, reader, context, (PdfString)value);

                case PdfElementType.PdfName:

                    return this.ConvertFromName(type, reader, context, (PdfName)value);

                case PdfElementType.PdfInt:

                    return this.ConvertFromInt(type, reader, context, (PdfInt)value);

                case PdfElementType.PdfReal:

                    return this.ConvertFromReal(type, reader, context, (PdfReal)value);

                default:

                    return value;
            }
        }

        protected virtual IPdfPrimitive ConvertFromIndirectReference(Type type, PostScriptReader reader, IPdfImportContext context, IndirectReference reference)
        {
            Guard.ThrowExceptionIfNull(type, "type");
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(reference, "reference");

            IndirectObject indirectObject = context.ReadIndirectObject(reference);
            IPdfPrimitive primitive = indirectObject.Content;

            return this.Convert(type, reader, context, primitive);
        }

        protected virtual IPdfPrimitive ConvertFromDictionary(Type type, PostScriptReader reader, IPdfImportContext context, PdfDictionary dictionary)
        {
            Guard.ThrowExceptionIfNull(type, "type");
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(dictionary, "dictionary");

            PdfObject res = this.CreateInstance(type, reader, context, dictionary) as PdfObject;
            if (res != null)
            {
                res.Load(reader, context, dictionary);

                return res;
            }

            return dictionary;
        }

        protected virtual IPdfPrimitive ConvertFromStream(Type type, PostScriptReader reader, IPdfImportContext context, PdfStream stream)
        {
            Guard.ThrowExceptionIfNull(type, "type");
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(stream, "stream");

            PdfStreamObjectBase res = this.CreateInstance(type, reader, context, stream.Dictionary) as PdfStreamObjectBase;
            if (res != null)
            {
                res.Load(reader, context, stream);

                return res;
            }

            return stream;
        }

        protected virtual IPdfPrimitive ConvertFromName(Type type, PostScriptReader reader, IPdfImportContext context, PdfName name)
        {
            Guard.ThrowExceptionIfNull(type, "type");
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(name, "name");

            return name;
        }

        protected virtual IPdfPrimitive ConvertFromString(Type type, PostScriptReader reader, IPdfImportContext context, PdfString str)
        {
            Guard.ThrowExceptionIfNull(type, "type");
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(str, "str");

            return str;
        }

        protected virtual IPdfPrimitive ConvertFromArray(Type type, PostScriptReader reader, IPdfImportContext context, PdfArray array)
        {
            Guard.ThrowExceptionIfNull(type, "type");
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(array, "array");

            return array;
        }

        protected virtual IPdfPrimitive ConvertFromInt(Type type, PostScriptReader reader, IPdfImportContext context, PdfInt i)
        {
            Guard.ThrowExceptionIfNull(type, "type");
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(i, "i");

            return i;
        }

        protected virtual IPdfPrimitive ConvertFromReal(Type type, PostScriptReader reader, IPdfImportContext context, PdfReal r)
        {
            Guard.ThrowExceptionIfNull(type, "type");
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(r, "r");

            return r;
        }

        protected virtual object CreateInstance(Type type, PostScriptReader reader, IPdfImportContext context, PdfDictionary dictionary)
        {
            Guard.ThrowExceptionIfNull(type, "type");
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(dictionary, "dictionary");

            return Activator.CreateInstance(type);
        }
    }
}
