﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Filters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Converters
{
    internal static class DecodeParametersConverter
    {
        public static DecodeParameters[] Convert(PostScriptReader reader, IPdfImportContext context, IPdfPrimitive primitive)
        {
            object result = Strip(reader, context, primitive);
            
            Dictionary<string, object> dictionary = result as Dictionary<string, object>;
            if (dictionary != null)
            {
                return CreateDecodeParameters(new Dictionary<string, object>[] { dictionary});
            }

            object[] array = result as object[];
            if (array != null)
            {
                return CreateDecodeParameters(array);
            }

            return null;
        }

        public static PrimitiveWrapper CreateDecodeParms(Dictionary<string, object>[] decodeParameters)
        {
            if (decodeParameters == null)
            {
                return null;
            }

            PdfArray array = new PdfArray();

            foreach (Dictionary<string, object> dictionary in decodeParameters)
            {
                PdfDictionary pdfDictionary = new PdfDictionary();

                foreach (KeyValuePair<string, object> pair in dictionary)
                {
                    pdfDictionary[pair.Key] = CreatePrimitive(pair.Value);
                }

                array.Add(pdfDictionary);
            }

            return new PrimitiveWrapper(array);
        }

        private static DecodeParameters[] CreateDecodeParameters(object[] decodeParametersArray)
        {
            Guard.ThrowExceptionIfNull(decodeParametersArray, "decodeParametersArray");

            DecodeParameters[] result = new DecodeParameters[decodeParametersArray.Length];
            for (int i = 0; i < result.Length; i++)
            {
                Dictionary<string, object> dictionary = (Dictionary<string, object>)decodeParametersArray[i];
                result[i] = new DecodeParameters(dictionary);
            }

            return result;
        }

        private static DecodeParameters CreateDecodeParameters(PostScriptReader reader, IPdfImportContext context, PdfDictionary dictionary)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(dictionary, "dictionary");

            DecodeParameters res = new DecodeParameters();

            foreach (var key in dictionary.Keys)
            {
                res[key] = Strip(reader, context, dictionary[key]);
            }

            return res;
        }

        private static object Strip(PostScriptReader reader, IPdfImportContext context, IPdfPrimitive primitive)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(primitive, "primitive");

            IPdfSimpleType simpleType = primitive as IPdfSimpleType;
            if (simpleType != null)
            {
                return simpleType.Value;
            }

            switch (primitive.Type)
            {
                case PdfElementType.IndirectReference: 
                    return StripIndirectReference(reader, context, (IndirectReference)primitive);

                case PdfElementType.Dictionary:
                    return StripDictionary(reader, context, (PdfDictionary)primitive);

                case PdfElementType.Array:
                    return StripArray(reader, context, (PdfArray)primitive);
            }

            throw new NotSupportedException("Type is not supported.");
        }

        private static object StripIndirectReference(PostScriptReader reader, IPdfImportContext context, IndirectReference reference)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(reference, "reference");

            IPdfPrimitive result;
            if (!context.TryGetIndirectObject(reference, out result))
            {
                result = context.ReadIndirectObject(reference);
                context.RegisterIndirectObject(reference, result);
            }

            return Strip(reader, context, result);
        }

        private static object[] StripArray(PostScriptReader reader, IPdfImportContext context, PdfArray array)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(array, "array");

            object[] res = new object[array.Count];
            for (int i = 0; i < array.Count; i++)
            {
                res[i] = Strip(reader, context, array[i]);
            }

            return res;
        }

        private static Dictionary<string, object> StripDictionary(PostScriptReader reader, IPdfImportContext context, PdfDictionary dictionary)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(dictionary, "dictionary");

            Dictionary<string, object> res = new Dictionary<string, object>();
            foreach (var key in dictionary.Keys)
            {
                res[key] = Strip(reader, context, dictionary[key]);
            }

            return res;
        }

        private static IPdfPrimitive CreatePrimitive(object obj)
        {
            if (obj == null)
            {
                return PdfNull.Instance;
            }

            String text = obj as String;
            if (text != null)
            {
                return new PdfName(text);
            }

            if (obj is int)
            {
                return new PdfInt((int)obj);
            }

            if (obj is double)
            {
                return new PdfReal((double)obj);
            }

            if (obj is bool)
            {
                return new PdfBool((bool)obj);
            }

            Array array = obj as Array;
            if (array != null)
            {
                PdfArray pdfArray = new PdfArray();

                foreach (object element in array)
                {
                    IPdfPrimitive primitive = CreatePrimitive(element);
                    pdfArray.Add(primitive);
                }

                return pdfArray;
            }

            throw new NotSupportedException(string.Format("Cannot create PdfPrimitive from type: {0}", obj.GetType()));
        }
    }
}
