﻿using System;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Annotations;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Forms;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Converters
{
    internal class FieldsConverter : Converter
    {
        protected override IPdfPrimitive ConvertFromArray(Type type, PostScriptReader reader, IPdfImportContext context, PdfArray array)
        {
            FieldsConvertionContext convertionContext = new FieldsConvertionContext(reader, context, new FormFieldsTree(), array, null);
            bool isRootLevel = true;
            this.LoadFormFields(convertionContext, isRootLevel);

            return convertionContext.FieldsTree;
        }

        internal static bool IsWidget(PostScriptReader reader, IPdfImportContext context, PdfDictionary dictionary)
        {
            PdfArray kids;
            bool hasKids = dictionary.TryGetElement(reader, context, "Kids", out kids) && kids != null && kids.Count > 0;

            if (hasKids)
            {
                return false;
            }

            bool isWidget = false;
            PdfName subtype;

            if (dictionary.TryGetElement(reader, context, "Subtype", out subtype))
            {
                isWidget = subtype.Value == "Widget";
            }

            return isWidget;
        }

        internal static bool ContainsMergedFieldProperties(PdfDictionary dictionary)
        {
            return dictionary.ContainsKey("FT") || dictionary.ContainsKey("T");
        }

        private void LoadFormFields(FieldsConvertionContext context, bool isRootLevel = false)
        {
            bool addParentAsTerminalNode = false;

            for (int i = 0; i < context.Kids.Count; i++)
            {
                IPdfPrimitive child = context.Kids[i];

                if (child.Type == PdfElementType.IndirectReference)
                {
                    IndirectReference childReference = (IndirectReference)child;
                    IndirectObject indirectObject = context.ImportContext.ReadIndirectObject(childReference);

                    if (indirectObject.Content.Type == PdfElementType.Dictionary)
                    {
                        PdfDictionary childDictionary = (PdfDictionary)indirectObject.Content;

                        this.LoadFieldFromDictionary(context, isRootLevel, childReference, childDictionary, ref addParentAsTerminalNode);
                    }
                }
            }

            if (addParentAsTerminalNode)
            {
                context.FieldsTree.Add(context.Parent);
            }
        }

        private void LoadFieldFromDictionary(
            FieldsConvertionContext context,
            bool isRootLevel,
            IndirectReference indirectReference,
            PdfDictionary pdfDictionary,
            ref bool addParentAsTerminalNode)
        {
            // Root fields should not have parent.
            bool shouldSkipField = isRootLevel && pdfDictionary.ContainsKey("Parent");

            if (!shouldSkipField)
            {
                if (this.TryLoadWidget(context, indirectReference, pdfDictionary))
                {
                    bool isMergedWidget = FieldsConverter.ContainsMergedFieldProperties(pdfDictionary);
                    addParentAsTerminalNode = addParentAsTerminalNode || !isMergedWidget;
                }
                else
                {
                    this.LoadFormField(context, indirectReference, pdfDictionary);
                }
            }
        }

        private bool TryLoadWidget(
            FieldsConvertionContext context,
            IndirectReference fieldReference,
            PdfDictionary widgetDictionary)
        {
            bool isWidget = FieldsConverter.IsWidget(context.Reader, context.ImportContext, widgetDictionary);

            if (isWidget)
            {
                WidgetObject widget = new WidgetObject();
                widget.Load(context.Reader, context.ImportContext, widgetDictionary);
                context.ImportContext.RegisterIndirectObject(fieldReference, widget);

                bool isMergedWidget = FieldsConverter.ContainsMergedFieldProperties(widgetDictionary);

                if (isMergedWidget)
                {
                    FormFieldNode node = new FormFieldNode();
                    node.Load(context.Reader, context.ImportContext, widgetDictionary);
                    context.FieldsTree.Add(node);
                    context.ImportContext.AddWidgetParent(widget, node);
                }
                else
                {
                    context.ImportContext.AddWidgetParent(widget, context.Parent);
                }
            }

            return isWidget;
        }

        private void LoadFormField(FieldsConvertionContext context, IndirectReference fieldReference, PdfDictionary fieldDictionary)
        {
            FormFieldNode node = new FormFieldNode();
            node.Load(context.Reader, context.ImportContext, fieldDictionary);
            context.ImportContext.RegisterIndirectObject(fieldReference, node);

            if (node.Kids == null || node.Kids.Count == 0)
            {
                context.FieldsTree.Add(node);
            }
            else
            {
                FieldsConvertionContext childFieldsConversionContext =
                    new FieldsConvertionContext(context.Reader, context.ImportContext, context.FieldsTree, node.Kids, node);

                this.LoadFormFields(childFieldsConversionContext);
            }
        }

        private class FieldsConvertionContext
        {
            private readonly FormFieldsTree fields;
            private readonly PdfArray currentLevelKids;
            private readonly FormFieldNode currentLevelParent;
            private readonly PostScriptReader reader;
            private readonly IPdfImportContext importContext;

            public FieldsConvertionContext(PostScriptReader reader, IPdfImportContext importContext, FormFieldsTree fields, PdfArray currentLevelKids, FormFieldNode currentLevelParent)
            {
                this.fields = fields;
                this.currentLevelKids = currentLevelKids;
                this.currentLevelParent = currentLevelParent;
                this.reader = reader;
                this.importContext = importContext;
            }

            public PdfArray Kids
            {
                get
                {
                    return this.currentLevelKids;
                }
            }

            public FormFieldNode Parent
            {
                get
                {
                    return this.currentLevelParent;
                }
            }

            public FormFieldsTree FieldsTree
            {
                get
                {
                    return this.fields;
                }
            }

            public PostScriptReader Reader
            {
                get
                {
                    return this.reader;
                }
            }

            public IPdfImportContext ImportContext
            {
                get
                {
                    return this.importContext;
                }
            }
        }
    }
}
