﻿using System;
using System.Collections.Generic;
#if !NETSTANDARD
using System.Windows;
#else
using Telerik.Documents.Primitives;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Objects;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Annotations
{
    internal abstract class AnnotationObject : PdfObject
    {
        public const string RectName = "Rect";
        public const string AppearanceName = "AP";
        public const string AppearanceStateName = "AS";
        private readonly DirectProperty<PdfArray> rect;
        private readonly DirectProperty<PdfInt> flags;
        private readonly DirectProperty<PdfName> appearanceState;
        private readonly ReferenceProperty<BorderStyle> borderStyle;
        private readonly ReferenceProperty<Appearances> appearances;
        private readonly ReferenceProperty<Page> parentPage;

        public AnnotationObject()
        {
            this.rect = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor(RectName, true));
            this.parentPage = this.RegisterReferenceProperty<Page>(new PdfPropertyDescriptor("P", false, PdfPropertyRestrictions.MustBeIndirectReference));
            this.flags = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("F"), new PdfInt(0));
            this.appearanceState = this.RegisterDirectProperty<PdfName>(new PdfPropertyDescriptor(AppearanceStateName));
            this.borderStyle = this.RegisterReferenceProperty<BorderStyle>(new PdfPropertyDescriptor("BS"));
            this.appearances = this.RegisterReferenceProperty<Appearances>(new PdfPropertyDescriptor(AppearanceName));
        }

        public virtual bool IsSupported
        {
            get
            {
                return true;
            }
        }

        public PdfArray Rect 
        {
            get
            {
                return this.rect.GetValue();
            }
            set
            {
                this.rect.SetValue(value);
            }
        }

        public Page ParentPage
        {
            get
            {
                return this.parentPage.GetValue();
            }
            set
            {
                this.parentPage.SetValue(value);
            }
        }

        public PdfInt Flags
        {
            get
            {
                return this.flags.GetValue();
            }
            set
            {
                this.flags.SetValue(value);
            }
        }

        public BorderStyle BorderStyle
        {
            get
            {
                return this.borderStyle.GetValue();
            }
            set
            {
                this.borderStyle.SetValue(value);
            }
        }

        public Appearances Appearances
        {
            get
            {
                return this.appearances.GetValue();
            }
            set
            {
                this.appearances.SetValue(value);
            }
        }

        public PdfName AppearanceState
        {
            get
            {
                return this.appearanceState.GetValue();
            }
            set
            {
                this.appearanceState.SetValue(value);
            }
        }

        public Annotation ToAnnotation(PostScriptReader reader, IRadFixedDocumentImportContext context, double pageHeightInDip)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
                        
            Annotation annotation = this.ToAnnotationOverride(reader, context);

            if (annotation == null)
            {
                return null;
            }

            annotation.Rect = this.ReadRect(reader, context, pageHeightInDip);
            this.ImportFlagsToAnnotation(annotation);
            
            if (this.BorderStyle != null)
            {
                annotation.Border = new AnnotationBorder();
                this.BorderStyle.CopyPropertiesTo(annotation.Border);
            }

            if (this.Appearances != null)
            {
                if(this.AppearanceState == null)
                {
                    annotation.Appearances = this.CreateSingleStateAppearances(reader, context);
                }
                else
                {
                    MultiStateAppearances appearances = this.CreateMultiStateAppearances(reader, context);
                    appearances.CurrentState = this.AppearanceState.Value;
                    annotation.Appearances = appearances;
                }

                annotation.DoOnAppearancesImport();
            }

            return annotation;
        }

        public abstract Annotation ToAnnotationOverride(PostScriptReader reader, IRadFixedDocumentImportContext context);

        protected void CopyPropertiesFrom(IPdfExportContext context, Annotation annotation, RadFixedPage fixedPage)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(annotation, "annotation");
            Guard.ThrowExceptionIfNull(fixedPage, "fixedPage");

            this.Rect = ConvertRect(annotation.Rect, fixedPage.Size.Height).ToPointPdfArray();
            this.BorderStyle = new BorderStyle();
            this.BorderStyle.CopyPropertiesFrom(context, annotation.Border);
            this.ExportAnnotationFlagsFromAnnotation(annotation);

            if(annotation.Appearances != null)
            {
                switch(annotation.Appearances.AppearancesType)
                {
                    case AnnotationAppearancesType.SingleStateAppearances:
                        this.Appearances = this.CreateSingleStateAppearances(context, (SingleStateAppearances)annotation.Appearances);
                        break;
                    case AnnotationAppearancesType.MultiStateAppearances:
                        this.Appearances = this.CreateMultiStateAppearances(context, (MultiStateAppearances)annotation.Appearances);
                        break;
                    default:
                        throw new NotSupportedException(string.Format("Not supported appearances type: {0}", annotation.Appearances.AppearancesType));
                }
            }

            Page mappedPage;
            if (context.TryGetPage(fixedPage, out mappedPage))
            {
                this.ParentPage = mappedPage;
            }
        }

        internal static Rect ConvertRect(Rect rect, double height)
        {
            return new Rect(rect.X, height - rect.Y - rect.Height, rect.Width, rect.Height);
        }

        internal static void ImportFlagsToAnnotation(Annotation annotation, int flags)
        {
            FlagReader<AnnotationFlag> flagReader = new FlagReader<AnnotationFlag>(flags);

            annotation.IsDisplayedWhenNotSupported = !flagReader.IsSet(AnnotationFlag.Invisble);
            annotation.IsHidden = flagReader.IsSet(AnnotationFlag.Hidden);
            annotation.IsPrintable = flagReader.IsSet(AnnotationFlag.Print);
            annotation.IsZoomingWithPage = !flagReader.IsSet(AnnotationFlag.NoZoom);
            annotation.IsRotatedWithPage = !flagReader.IsSet(AnnotationFlag.NoRotate);
            annotation.IsVisibleInViewerUI = !flagReader.IsSet(AnnotationFlag.NoView);
            annotation.IsReadOnly = flagReader.IsSet(AnnotationFlag.ReadOnly);
            annotation.IsLockedByPositionAndSize = flagReader.IsSet(AnnotationFlag.Locked);
            annotation.IsTogglingVisibilityInViewerUI = flagReader.IsSet(AnnotationFlag.ToggleNoView);
            annotation.IsContentLocked = flagReader.IsSet(AnnotationFlag.LockedContents);
        }

        private Rect ReadRect(PostScriptReader reader, IRadFixedDocumentImportContext context, double pageHeightInDip)
        {
            Rect rect = this.Rect.ToDipRect(reader, context);
            Rect convertedRect = ConvertRect(rect, pageHeightInDip);

            return convertedRect;
        }

        private void ImportFlagsToAnnotation(Annotation annotation)
        {
            AnnotationObject.ImportFlagsToAnnotation(annotation, this.Flags.Value);
        }

        private void ExportAnnotationFlagsFromAnnotation(Annotation annotation)
        {
            FlagWriter<AnnotationFlag> flagWriter = new FlagWriter<AnnotationFlag>();

            flagWriter.SetFlagOnCondition(AnnotationFlag.Invisble, !annotation.IsDisplayedWhenNotSupported);
            flagWriter.SetFlagOnCondition(AnnotationFlag.Hidden, annotation.IsHidden);
            flagWriter.SetFlagOnCondition(AnnotationFlag.Print, annotation.IsPrintable);
            flagWriter.SetFlagOnCondition(AnnotationFlag.NoZoom, !annotation.IsZoomingWithPage);
            flagWriter.SetFlagOnCondition(AnnotationFlag.NoRotate, !annotation.IsRotatedWithPage);
            flagWriter.SetFlagOnCondition(AnnotationFlag.NoView, !annotation.IsVisibleInViewerUI);
            flagWriter.SetFlagOnCondition(AnnotationFlag.ReadOnly, annotation.IsReadOnly);
            flagWriter.SetFlagOnCondition(AnnotationFlag.Locked, annotation.IsLockedByPositionAndSize);
            flagWriter.SetFlagOnCondition(AnnotationFlag.ToggleNoView, annotation.IsTogglingVisibilityInViewerUI);
            flagWriter.SetFlagOnCondition(AnnotationFlag.LockedContents, annotation.IsContentLocked);

            this.Flags = new PdfInt(flagWriter.ResultFlags);
        }

        private SingleStateAppearances CreateSingleStateAppearances(PostScriptReader reader, IRadFixedDocumentImportContext context)
        {
            SingleStateAppearances appearances = null;

            if (this.Appearances.NormalAppearance != null && this.Appearances.NormalAppearance.SingleStateAppearance != null)
            {
                appearances = new SingleStateAppearances();
                appearances.NormalAppearance = context.GetFormSource(reader, this.Appearances.NormalAppearance.SingleStateAppearance);

                if (this.Appearances.DownAppearance != null && this.Appearances.DownAppearance.SingleStateAppearance != null)
                {
                    appearances.MouseDownAppearance = context.GetFormSource(reader, this.Appearances.DownAppearance.SingleStateAppearance);
                }

                if (this.Appearances.RolloverAppearance != null && this.Appearances.RolloverAppearance.SingleStateAppearance != null)
                {
                    appearances.MouseOverAppearance = context.GetFormSource(reader, this.Appearances.RolloverAppearance.SingleStateAppearance);
                }
            }

            return appearances;
        }

        private MultiStateAppearances CreateMultiStateAppearances(PostScriptReader reader, IRadFixedDocumentImportContext context)
        {
            MultiStateAppearances multiStateAppearances = new MultiStateAppearances();

            if (this.Appearances.NormalAppearance != null)
            {
                foreach (KeyValuePair<string, FormXObject> stateToAppearance in this.Appearances.NormalAppearance.StateAppearances)
                {
                    string stateName = stateToAppearance.Key;
                    FormXObject stateNormalAppearanceObject = stateToAppearance.Value;
                    SingleStateAppearances appearances = new SingleStateAppearances();
                    multiStateAppearances.AddAppearance(stateName, appearances);
                    appearances.NormalAppearance = context.GetFormSource(reader, stateNormalAppearanceObject);

                    FormXObject stateDownAppearanceObject;
                    if (this.Appearances.DownAppearance != null && this.Appearances.DownAppearance.TryGetStateAppearance(stateName, out stateDownAppearanceObject))
                    {
                        appearances.MouseDownAppearance = context.GetFormSource(reader, stateDownAppearanceObject);
                    }

                    FormXObject stateRolloverAppearanceObject;
                    if (this.Appearances.RolloverAppearance != null && this.Appearances.RolloverAppearance.TryGetStateAppearance(stateName, out stateRolloverAppearanceObject))
                    {
                        appearances.MouseOverAppearance = context.GetFormSource(reader, stateRolloverAppearanceObject);
                    }
                }
            }

            return multiStateAppearances;
        }

        private Appearances CreateMultiStateAppearances(IPdfExportContext context, MultiStateAppearances multiStateAppearances)
        {
            Dictionary<string, FormXObject> normalMapping = null;
            Dictionary<string, FormXObject> downMapping = null;
            Dictionary<string, FormXObject> rolloverMapping = null;

            this.AppearanceState = new PdfName(multiStateAppearances.CurrentState);

            foreach(KeyValuePair<string, SingleStateAppearances> stateAppearances in multiStateAppearances)
            {
                string stateName = stateAppearances.Key;
                SingleStateAppearances singleStateAppearances = stateAppearances.Value;

                if(singleStateAppearances.NormalAppearance != null)
                {
                    if(normalMapping == null)
                    {
                        normalMapping = new Dictionary<string, FormXObject>();
                    }

                    FormXObject normalXForm = GetAppearanceXForm(context, singleStateAppearances.NormalAppearance);
                    normalMapping.Add(stateName, normalXForm);

                    if (singleStateAppearances.MouseDownAppearance != null)
                    {
                        if (downMapping == null)
                        {
                            downMapping = new Dictionary<string, FormXObject>();
                        }

                        FormXObject downXForm = GetAppearanceXForm(context, singleStateAppearances.MouseDownAppearance);
                        downMapping.Add(stateName, downXForm);
                    }

                    if (singleStateAppearances.MouseOverAppearance != null)
                    {
                        if (rolloverMapping == null)
                        {
                            rolloverMapping = new Dictionary<string, FormXObject>();
                        }

                        FormXObject rolloverXForm = GetAppearanceXForm(context, singleStateAppearances.MouseOverAppearance);
                        rolloverMapping.Add(stateName, rolloverXForm);
                    }
                }
            }

            Appearances result = CreateMultiStateAppearances(normalMapping, downMapping, rolloverMapping);

            return result;
        }

        private Appearances CreateMultiStateAppearances(Dictionary<string, FormXObject> normalMapping, Dictionary<string, FormXObject> downMapping, Dictionary<string, FormXObject> rolloverMapping)
        {
            if (normalMapping == null)
            {
                return null;
            }
            else
            {
                return new Appearances()
                {
                    NormalAppearance = new Appearance(normalMapping),
                    DownAppearance = downMapping == null ? null : new Appearance(downMapping),
                    RolloverAppearance = rolloverMapping == null ? null : new Appearance(rolloverMapping),
                };
            }
        }

        private Appearances CreateSingleStateAppearances(IPdfExportContext context, SingleStateAppearances singleStateAppearances)
        {
            Appearances result = null;

            if (singleStateAppearances.NormalAppearance != null)
            {
                result = new Appearances();
                FormXObject normalXForm = GetAppearanceXForm(context, singleStateAppearances.NormalAppearance);
                result.NormalAppearance = new Appearance(normalXForm);

                if (singleStateAppearances.MouseDownAppearance != null)
                {
                    FormXObject downXForm = GetAppearanceXForm(context, singleStateAppearances.MouseDownAppearance);
                    result.DownAppearance = new Appearance(downXForm);
                }

                if (singleStateAppearances.MouseOverAppearance != null)
                {
                    FormXObject rolloverXForm = GetAppearanceXForm(context, singleStateAppearances.MouseOverAppearance);
                    result.RolloverAppearance = new Appearance(rolloverXForm);
                }
            }

            return result;
        }

        private static FormXObject GetAppearanceXForm(IPdfExportContext context, FormSource appearanceSource)
        {
            ResourceEntry entry = context.GetResource(appearanceSource);
            FormXObject xForm = (FormXObject)entry.Resource.Content;

            return xForm;
        }
    }
}
