﻿using System;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Forms;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Objects;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Resources;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Annotations
{
    internal class WidgetPropertiesExporter : VariableTextObjectPropertiesExporter<WidgetObject>
    {
        public WidgetPropertiesExporter(WidgetObject node, IPdfExportContext exportContext) 
            : base(node, exportContext)
        {
        }

        public void ExportWidgetProperties(Widget widget)
        {
            this.ExportVariableProperties(widget.TextProperties);
            this.ExportHighlightingMode(widget.HighlightingMode);
            this.ExportDynamicAppearanceCharacteristics(widget);
        }

        private void ExportHighlightingMode(HighlightingMode highlightingMode)
        {
            switch(highlightingMode)
            {
                case HighlightingMode.InvertBorderOfAnnotationRectangle:
                    this.Node.HighlightingMode = new PdfName(HighlightingModes.InvertBorderOfAnnotationRectangle);
                    break;
                case HighlightingMode.UseAnnotationDownAppearance:
                    this.Node.HighlightingMode = new PdfName(HighlightingModes.UseAnnotationPushAppearance);
                    break;
                case HighlightingMode.NoHighlighting:
                    this.Node.HighlightingMode = new PdfName(HighlightingModes.NoHighlighting);
                    break;
                default:
                    this.Node.HighlightingMode = new PdfName(HighlightingModes.InvertContentOfAnnotationRectangle);
                    break;
            }
        }

        private void ExportDynamicAppearanceCharacteristics(Widget widget)
        {
            this.Node.DynamicAppearanceCharacteristics = new AppearanceCharacteristics();

            switch (widget.WidgetContentType)
            {
                case WidgetContentType.VariableContent:
                case WidgetContentType.SignatureContent:
                    Widget<DynamicAppearanceCharacteristics> dynamicWidget = (Widget<DynamicAppearanceCharacteristics>)widget;
                    this.CopyDynamicAppearance(dynamicWidget.AppearanceCharacteristics);
                    break;
                case WidgetContentType.TwoStatesContent:
                    TwoStatesButtonWidget buttonWidget = (TwoStatesButtonWidget)widget;
                    this.CopyButtonDynamicAppearance(buttonWidget.AppearanceCharacteristics);
                    break;
                case WidgetContentType.PushButtonContent:
                    PushButtonWidget pushButtonWidget = (PushButtonWidget)widget;
                    this.CopyPushButtonDynamicAppearance(pushButtonWidget.AppearanceCharacteristics);
                    break;
            }
        }

        private void CopyDynamicAppearance(DynamicAppearanceCharacteristics appearance)
        {
            this.Node.DynamicAppearanceCharacteristics.Rotation = Page.GetRotate(appearance.Rotation);
            this.Node.DynamicAppearanceCharacteristics.BorderColor = WidgetPropertiesExporter.GetColorArray(appearance.BorderColor);
            this.Node.DynamicAppearanceCharacteristics.Background = WidgetPropertiesExporter.GetColorArray(appearance.Background);
        }

        private void CopyButtonDynamicAppearance(ButtonAppearanceCharacteristics appearance)
        {
            this.CopyDynamicAppearance(appearance);

            this.Node.DynamicAppearanceCharacteristics.ButtonNormalCaption = appearance.NormalCaption.ToPdfString();
        }

        private void CopyPushButtonDynamicAppearance(PushButtonAppearanceCharacteristics appearance)
        {
            this.CopyButtonDynamicAppearance(appearance);

            this.Node.DynamicAppearanceCharacteristics.DownCaption = appearance.MouseDownCaption.ToPdfString();
            this.Node.DynamicAppearanceCharacteristics.RolloverCaption = appearance.MouseOverCaption.ToPdfString();
            this.Node.DynamicAppearanceCharacteristics.NormalIcon = this.GetIcon(appearance.NormalIconSource);
            this.Node.DynamicAppearanceCharacteristics.DownIcon = this.GetIcon(appearance.MouseDownIconSource);
            this.Node.DynamicAppearanceCharacteristics.RolloverIcon = this.GetIcon(appearance.MouseOverIconSource);
            this.Node.DynamicAppearanceCharacteristics.IconFit = this.GetIconFit(appearance.IconFitOptions);
            this.Node.DynamicAppearanceCharacteristics.PushButtonTextPosition = new PdfInt((int)appearance.IconAndCaptionPosition);
        }

        private IconFit GetIconFit(IconFitOptions iconFitOptions)
        {
            IconFit iconFit = new IconFit();

            switch(iconFitOptions.ScaleCondition)
            {
                case IconScaleCondition.ScaleIfBiggerThanAnnotationRectangle:
                    iconFit.ScaleCondition = new PdfName(IconScaleConditions.ScaleIfBiggerThanAnnotationRectangle);
                    break;
                case IconScaleCondition.ScaleIfSmallerThanAnnotationRectangle:
                    iconFit.ScaleCondition = new PdfName(IconScaleConditions.ScaleIfSmallerThanAnnotationRectangle);
                    break;
                case IconScaleCondition.NeverScale:
                    iconFit.ScaleCondition = new PdfName(IconScaleConditions.NeverScale);
                    break;
                default:
                    iconFit.ScaleCondition = new PdfName(IconScaleConditions.AlwaysScale);
                    break;
            }

            iconFit.ScaleType = (iconFitOptions.ScalingType == IconScalingType.FitExactly) ? new PdfName(IconScalingTypes.FitExactly) : new PdfName(IconScalingTypes.Proportional);
            iconFit.TranslationAfterProportionalScaling = new PdfArray(new PdfReal(iconFitOptions.BlankSpaceFromTheLeftSide), new PdfReal(iconFitOptions.BlankSpaceFromTheBottomSide));
            iconFit.FitIgnoringBorderWidth = new PdfBool(iconFitOptions.IgnoreBorderWidth);

            return iconFit;
        }

        private FormXObject GetIcon(FormSource source)
        {
            if(source == null)
            {
                return null;
            }

            ResourceEntry entry = this.Context.GetResource(source);
            FormXObject xForm = (FormXObject)entry.Resource.Content;

            return xForm;
        }

        private static PdfArray GetColorArray(RgbColor color)
        {
            if(color == null)
            {
                return null;
            }

            PdfReal r = ColorSpaces.ColorObjectBase.ConvertFromByte(color.R);
            PdfReal g = ColorSpaces.ColorObjectBase.ConvertFromByte(color.G);
            PdfReal b = ColorSpaces.ColorObjectBase.ConvertFromByte(color.B);

            return new PdfArray(r, g, b);
        }
    }
}
