﻿using System;
using Telerik.Windows.Documents.Core.Imaging;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Forms;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Objects;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Resources;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Annotations
{
    internal class WidgetPropertiesImporter : VariableTextObjectPropertiesImporter<WidgetObject>
    {
        public WidgetPropertiesImporter(WidgetObject node, PostScriptReader reader, IRadFixedDocumentImportContext context) 
            : base(node, reader, context)
        {
        }

        public void ImportWidgetProperties(Widget widget)
        {
            widget.TextProperties = this.ReadTextProperties();
            widget.HighlightingMode = this.ReadHighlightingMode();

            if (this.Node.DynamicAppearanceCharacteristics != null)
            {
                switch (widget.WidgetContentType)
                {
                    case WidgetContentType.VariableContent:
                    case WidgetContentType.SignatureContent:
                        Widget<DynamicAppearanceCharacteristics> dynamicWidget = (Widget<DynamicAppearanceCharacteristics>)widget;
                        dynamicWidget.AppearanceCharacteristics = new DynamicAppearanceCharacteristics();
                        this.ReadDynamicAppearanceProperties(dynamicWidget.AppearanceCharacteristics);
                        break;
                    case WidgetContentType.TwoStatesContent:
                        TwoStatesButtonWidget twoStateButton = (TwoStatesButtonWidget)widget;
                        twoStateButton.AppearanceCharacteristics = new ButtonAppearanceCharacteristics();
                        this.ReadButtonDynamicAppearanceProperties(twoStateButton.AppearanceCharacteristics);
                        break;
                    case WidgetContentType.PushButtonContent:
                        PushButtonWidget pushButton = (PushButtonWidget)widget;
                        pushButton.AppearanceCharacteristics = new PushButtonAppearanceCharacteristics();
                        this.ReadPushButtonDynamicAppearanceProperties(pushButton.AppearanceCharacteristics);
                        break;
                    default:
                        throw new NotSupportedException(string.Format("Not supported widget type: {0}", widget.WidgetContentType));
                }
            }
        }

        private void ReadPushButtonDynamicAppearanceProperties(PushButtonAppearanceCharacteristics appearance)
        {
            this.ReadButtonDynamicAppearanceProperties(appearance);
            
            appearance.MouseDownCaption = GetCaption(this.Node.DynamicAppearanceCharacteristics.DownCaption);
            appearance.MouseOverCaption = GetCaption(this.Node.DynamicAppearanceCharacteristics.RolloverCaption);   
            appearance.NormalIconSource = this.GetIconSource(this.Node.DynamicAppearanceCharacteristics.NormalIcon);
            appearance.MouseDownIconSource = this.GetIconSource(this.Node.DynamicAppearanceCharacteristics.DownIcon);
            appearance.MouseOverIconSource = this.GetIconSource(this.Node.DynamicAppearanceCharacteristics.RolloverIcon);
            appearance.IconFitOptions.ScaleCondition = this.ReadScaleCondition();
            appearance.IconFitOptions.ScalingType = this.ReadScalingType();
            double[] array = this.Node.DynamicAppearanceCharacteristics.IconFit.TranslationAfterProportionalScaling.ToDoubleArray();
            appearance.IconFitOptions.BlankSpaceFromTheLeftSide = array[0];
            appearance.IconFitOptions.BlankSpaceFromTheBottomSide = array[1];
            appearance.IconFitOptions.IgnoreBorderWidth = this.Node.DynamicAppearanceCharacteristics.IconFit.FitIgnoringBorderWidth.Value;
            appearance.IconAndCaptionPosition = (IconAndCaptionPosition)this.Node.DynamicAppearanceCharacteristics.PushButtonTextPosition.Value;
        }

        private IconScalingType ReadScalingType()
        {
            string type = this.Node.DynamicAppearanceCharacteristics.IconFit.ScaleType.Value;

            return type.Equals(IconScalingTypes.FitExactly) ? IconScalingType.FitExactly : IconScalingType.Proportional;
        }

        private IconScaleCondition ReadScaleCondition()
        {
            string condition = this.Node.DynamicAppearanceCharacteristics.IconFit.ScaleCondition.Value;
            
            switch(condition)
            {
                case IconScaleConditions.ScaleIfBiggerThanAnnotationRectangle:
                    return IconScaleCondition.ScaleIfBiggerThanAnnotationRectangle;
                case IconScaleConditions.ScaleIfSmallerThanAnnotationRectangle:
                    return IconScaleCondition.ScaleIfSmallerThanAnnotationRectangle;
                case IconScaleConditions.NeverScale:
                    return IconScaleCondition.NeverScale;
                default:
                    return IconScaleCondition.AlwaysScale;
            }
        }

        private void ReadButtonDynamicAppearanceProperties(ButtonAppearanceCharacteristics appearance)
        {
            this.ReadDynamicAppearanceProperties(appearance);
            
            appearance.NormalCaption = GetCaption(this.Node.DynamicAppearanceCharacteristics.ButtonNormalCaption);
        }

        private void ReadDynamicAppearanceProperties(DynamicAppearanceCharacteristics appearance)
        {
            appearance.Rotation = Page.GetRotation(this.Node.DynamicAppearanceCharacteristics.Rotation);
            appearance.Background = GetColor(this.Node.DynamicAppearanceCharacteristics.Background);
            appearance.BorderColor = GetColor(this.Node.DynamicAppearanceCharacteristics.BorderColor);
        }

        private FormSource GetIconSource(FormXObject form)
        {
            if(form != null)
            {
                return form.ToFormSource(this.Reader, this.Context);
            }

            return null;
        }

        private static string GetCaption(PdfString caption)
        {
            if (caption != null)
            {
                return caption.ToString();
            }

            return null;
        }

        private static RgbColor GetColor(PdfArray colorArray)
        {
            if(colorArray == null)
            {
                return null;
            }

            double[] array = colorArray.ToDoubleArray();

            switch(colorArray.Count)
            {
                case 1:
                    byte gray = Color.ConvertColorComponentToByte(array[0]);

                    return new RgbColor(gray, gray, gray);
                case 3:
                    byte r = Color.ConvertColorComponentToByte(array[0]);
                    byte g = Color.ConvertColorComponentToByte(array[1]);
                    byte b = Color.ConvertColorComponentToByte(array[2]);

                    return new RgbColor(r, g, b);
                case 4:
                    double c = array[0];
                    double m = array[1];
                    double y = array[2];
                    double k = array[3];
                    Color color = Color.FromCmyk(c, m, y, k);

                    return new RgbColor(color.R, color.G, color.B);
                default:
                    return null;
            }
        }

        private HighlightingMode ReadHighlightingMode()
        {
            string modeValue = this.Node.HighlightingMode.Value;

            switch(modeValue)
            {
                case HighlightingModes.NoHighlighting:
                    return HighlightingMode.NoHighlighting;
                case HighlightingModes.InvertBorderOfAnnotationRectangle:
                    return HighlightingMode.InvertBorderOfAnnotationRectangle;
                case HighlightingModes.UseAnnotationPushAppearance:
                case HighlightingModes.UseAnnotationToggleAppearance:
                    return HighlightingMode.UseAnnotationDownAppearance;
                default:
                    return HighlightingMode.InvertContentOfAnnotationRectangle;
            }
        }
    }
}
