﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.Exceptions;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Filters;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.ColorSpaces
{
    internal static class ColorSpaceManager
    {
        private static readonly Dictionary<string, ColorSpace> stringToColoSpaceEnumMapper = new Dictionary<string, ColorSpace>();
        private static readonly Dictionary<ColorSpace, string> coloSpaceEnumToStringMapper = new Dictionary<ColorSpace, string>();
        private static readonly Dictionary<ColorSpace, Func<ColorSpaceObject>> creators = new Dictionary<ColorSpace, Func<ColorSpaceObject>>();

        static ColorSpaceManager()
        {
            RegisterMapping(ColorSpaceNames.CalGray, ColorSpace.CalGray, () => { return new CalGrayColorSpaceObject(); });
            RegisterMapping(ColorSpaceNames.CalRgb, ColorSpace.CalRgb, () => { return new CalRgbColorSpaceObject(); });
            RegisterMapping(ColorSpaceNames.DeviceCmyk, ColorSpace.CMYK, () => { return new DeviceCmykColorSpaceObject(); });
            RegisterMapping(ColorSpaceNames.DeviceN, ColorSpace.DeviceN, () => { return new DeviceNColorSpaceObject(); });
            RegisterMapping(ColorSpaceNames.DeviceGray, ColorSpace.Gray, () => { return new DeviceGrayColorSpaceObject(); });
            RegisterMapping(ColorSpaceNames.IccBased, ColorSpace.ICCBased, () => { return new IccBasedColorSpaceObject(); });
            RegisterMapping(ColorSpaceNames.Indexed, ColorSpace.Indexed, () => { return new IndexedColorSpaceObject(); });
            RegisterMapping(ColorSpaceNames.Lab, ColorSpace.Lab, () => { return new LabColorSpaceObject(); });
            RegisterMapping(ColorSpaceNames.Pattern, ColorSpace.Pattern, () => { return new PatternColorSpaceObject(); });
            RegisterMapping(ColorSpaceNames.DeviceRgb, ColorSpace.RGB, () => { return new DeviceRgbColorSpaceObject(); });
            RegisterMapping(ColorSpaceNames.Separation, ColorSpace.Separation, () => { return new SeparationColorSpaceObject(); });
        }

        public static bool IsColorSpaceName(string name)
        {
            return stringToColoSpaceEnumMapper.ContainsKey(name);
        }

        public static ColorSpaceObject CreateColorSpaceObject(string colorSpaceName)
        {
            ColorSpace? enumValue = ColorSpaceManager.GetColorSpaceEnumValue(colorSpaceName);
            if (enumValue.HasValue)
            {
                ColorSpaceObject colorSpaceObject = ColorSpaceManager.CreateColorSpaceObject(enumValue.Value);

                return colorSpaceObject;
            }

            throw new NotSupportedColorSpaceException(colorSpaceName);
        }

        public static ColorSpaceObject CreateColorSpaceObject(ColorSpace enumValue)
        {
            Func<ColorSpaceObject> func;
            if (creators.TryGetValue(enumValue, out func))
            {
                ColorSpaceObject colorSpaceObject = func();

                return colorSpaceObject;
            }

            throw new NotSupportedColorSpaceException("The provided color space is not supported.");
        }

        public static ColorSpaceObject CreateColorSpaceObject(ColorSpaceBase colorSpaceBase)
        {
            ColorSpace? enumValue = ColorSpaceManager.GetColorSpaceEnumValue(colorSpaceBase.Name);

            if (enumValue.HasValue)
            {
                ColorSpaceObject colorSpaceObject = ColorSpaceManager.CreateColorSpaceObject(enumValue.Value);

                if (colorSpaceObject != null)
                {
                    colorSpaceObject.CopyPropertiesFrom(colorSpaceBase);

                    return colorSpaceObject;
                }
            }

            throw new NotSupportedColorSpaceException(string.Format("The provided color space: {0} is not supported.", colorSpaceBase.Name));
        }

        private static ColorSpace? GetColorSpaceEnumValue(string colorSpaceName)
        {
            ColorSpace enumValue;
            if (stringToColoSpaceEnumMapper.TryGetValue(colorSpaceName, out enumValue))
            {
                return enumValue;
            }

            return null;
        }

        private static void RegisterMapping(string name, ColorSpace enumValue, Func<ColorSpaceObject> creator)
        {
            stringToColoSpaceEnumMapper.Add(name, enumValue);
            coloSpaceEnumToStringMapper.Add(enumValue, name);
            creators.Add(enumValue, creator);
        }
    }
}
