﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common
{
    internal abstract class PdfObject : PdfPrimitive
    {
        private readonly Dictionary<string, IPdfProperty> properties;

        public PdfObject()
        {
            this.properties = new Dictionary<string, IPdfProperty>();
        }

        public override PdfElementType Type
        {
            get
            {
                return PdfElementType.PdfObject;
            }
        }

        internal Dictionary<string, IPdfProperty> Properties
        {
            get
            {
                return this.properties;
            }
        }

        public override void Write(PdfWriter writer, IPdfExportContext context)
        {
            WritePdfPropertiesDictionary(this, writer, context);
        }
 
        internal static void WritePdfPropertiesDictionary(PdfObject instance, PdfWriter writer, IPdfExportContext context, bool shouldWriteObjectDescriptor = true)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(context, "context");

            if (instance.Properties.Count == 0)
            {
                return;
            }

            writer.Write(PdfNames.PdfDictionaryStart);

            if (shouldWriteObjectDescriptor)
            {
                PdfObjectDescriptor pdfObjectDescriptor = PdfObjectDescriptors.GetPdfObjectDescriptor(instance);
                writer.WritePdfObjectDescriptor(context, pdfObjectDescriptor);
            }

            foreach (IPdfProperty property in instance.Properties.Values)
            {
                if ((property.Descriptor.IsRequired && property.HasValue) || property.HasNonDefaultValue || property.Descriptor.AlwaysExport)
                {
                    writer.WritePdfName(property.Descriptor.Name);
                    writer.WriteSeparator(); 

                    WritePropertyValue(writer, context, property);

                    writer.WriteSeparator();
                }
            }

            writer.Write(PdfNames.PdfDictionaryEnd);
        }

        public virtual void Read(PostScriptReader reader, IPdfImportContext context)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            PdfDictionary dictionary = reader.Read<PdfDictionary>(context, PdfElementType.Dictionary);
            if (dictionary != null)
            {
                this.LoadFromDictionary(reader, context, dictionary);
            }
        }

        public void Load(PostScriptReader reader, IPdfImportContext context, IPdfPrimitive primitive)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(primitive, "primitive");

            if (primitive.Type == PdfElementType.Dictionary)
            {
                this.LoadFromDictionary(reader, context, (PdfDictionary)primitive);
            }
            else
            {
                this.LoadOverride(reader, context, primitive);
            }
        }

        protected virtual void LoadOverride(PostScriptReader reader, IPdfImportContext context, IPdfPrimitive primitive)
        {
        }

        protected static void WritePropertyValue(PdfWriter writer, IPdfExportContext context, IPdfProperty property)
        {
            IPdfPrimitive value = property.GetValue();
            switch (property.Descriptor.Restrictions)
            {
                case PdfPropertyRestrictions.MustBeIndirectReference:
                    WriteIndirectReference(writer, context, value);
                    break;
                case PdfPropertyRestrictions.ContainsOnlyIndirectReferences:
                    WriteIndirectReferencesCollection(writer, context, value);
                    break;
                default:
                    if (value.ExportAs == PdfElementType.PdfStreamObject)
                    {
                        WriteIndirectReference(writer, context, value);
                    }
                    else
                    {
                        value.Write(writer, context);
                    }
                    break;
            }
        }

        protected DirectProperty<T> RegisterDirectProperty<T>(PdfPropertyDescriptor descriptor)
            where T : class, IPdfPrimitive
        {
            DirectProperty<T> res = new DirectProperty<T>(descriptor);
            this.RegisterPdfProperty(res);

            return res;
        }

        protected DirectProperty<T> RegisterDirectProperty<T>(PdfPropertyDescriptor descriptor, T initialValue)
            where T : class, IPdfPrimitive
        {
            DirectProperty<T> res = new DirectProperty<T>(descriptor, initialValue);
            this.RegisterPdfProperty(res);

            return res;
        }

        protected ReferenceProperty<T> RegisterReferenceProperty<T>(PdfPropertyDescriptor descriptor)
            where T : class, IPdfPrimitive
        {
            ReferenceProperty<T> res = new ReferenceProperty<T>(descriptor);
            this.RegisterPdfProperty(res);

            return res;
        }

        protected ReferenceProperty<T> RegisterReferenceProperty<T>(PdfPropertyDescriptor descriptor, T initialValue)
            where T : class, IPdfPrimitive
        {
            ReferenceProperty<T> res = new ReferenceProperty<T>(descriptor, initialValue);
            this.RegisterPdfProperty(res);

            return res;
        }

        protected virtual void RegisterPdfProperty(IPdfProperty property)
        {
            this.RegisterPdfProperty(property.Descriptor.Name, property);
        }

        protected static void WriteIndirectReferencesCollection(PdfWriter writer, IPdfExportContext context, IPdfPrimitive primitive)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(context, "context");

            switch (primitive.Type)
            {
                case Types.PdfElementType.Array:
                    ((PdfArray)primitive).Write(writer, context, true);

                    break;
            }
        }

        private void RegisterPdfProperty(string name, IPdfProperty property)
        {
            this.properties[name] = property;
        }

        private void LoadFromDictionary(PostScriptReader reader, IPdfImportContext context, PdfDictionary dictionary)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(dictionary, "dictionary");

            IPdfProperty prop;
            foreach (KeyValuePair<string, IPdfPrimitive> pair in dictionary)
            {
                if (this.properties.TryGetValue(pair.Key, out prop))
                {
                    prop.SetValue(reader, context, pair.Value);
                }
            }
        }
    }
}
