﻿using System;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Filters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common
{
    internal abstract class PdfStreamObjectBase : PdfObject
    {
        private readonly DirectProperty<PdfInt> length;
        private readonly DirectProperty<PdfArray> filters;
        private readonly ReferenceProperty<PrimitiveWrapper> decodeParms;

        public PdfStreamObjectBase()
        {
            this.length = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("Length", true));
            this.filters = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor("Filter", true));
            this.decodeParms = this.RegisterReferenceProperty<PrimitiveWrapper>(new PdfPropertyDescriptor("DecodeParms"));
        }

        public override PdfElementType Type
        {
            get
            {
                return PdfElementType.PdfStreamObject;
            }
        }

        public PdfArray Filters
        {
            get
            {
                return this.filters.GetValue();
            }
            set
            {
                this.filters.SetValue(value);
            }
        }

        public PrimitiveWrapper DecodeParms
        {
            get
            {
                return this.decodeParms.GetValue();
            }
            set
            {
                this.decodeParms.SetValue(value);
            }
        }

        public PdfInt Length
        {
            get
            {
                return this.length.GetValue();
            }
            set
            {
                this.length.SetValue(value);
            }
        }

        public virtual bool ShouldEncryptData
        {
            get
            {
                return true;
            }
        }

        public override void Write(PdfWriter writer, IPdfExportContext context)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(context, "context");

            if (this.ExportAs == PdfElementType.PdfStreamObject)
            {
                byte[] data = this.GetData(context);
                this.Filters = this.GetExportFilters(context);

                if (this.Filters != null)
                {
                    data = FiltersManager.Encode(this.CreateEncodeObject(context), data, this.Filters);
                }

                if (this.ShouldEncryptData)
                {
                    data = context.EncryptStream(data);
                }

                this.WriteDictionaryAndEncodedData(writer, context, data);
            }
            else
            {
                base.Write(writer, context);
            }
        }

        protected void WriteDictionaryAndEncodedData(PdfWriter writer, IPdfExportContext context, byte[] encodedData)
        {
            this.Length = new PdfInt(encodedData.Length);

            base.Write(writer, context);
            PdfStreamObjectBase.WriteEncodedData(writer, context, encodedData);
        }

        protected override void LoadOverride(PostScriptReader reader, IPdfImportContext context, IPdfPrimitive primitive)
        {
            switch (primitive.Type)
            {
                case PdfElementType.Stream:
                    PdfStreamBase stream = (PdfStreamBase)primitive;
                    this.Load(reader, context, stream.Dictionary);
                    this.InterpretData(reader, context, stream);

                    break;

                default:
                    base.LoadOverride(reader, context, primitive);

                    break;
            }
        }

        protected virtual byte[] GetData(IPdfExportContext context)
        {
            throw new NotSupportedException();
        }

        protected virtual void InterpretData(PostScriptReader reader, IPdfImportContext context, PdfStreamBase stream)
        {
            throw new NotSupportedException();
        }

        protected virtual PdfArray GetExportFilters(IPdfExportContext context)
        {
            return new PdfArray(new PdfName(PdfFilterNames.FlateDecode));
        }

        protected virtual Filters.PdfObject CreateEncodeObject(IPdfExportContext context)
        {
            return new Filters.PdfObject(context);
        }

        internal static void WriteEncodedData(PdfWriter writer, IPdfExportContext context, byte[] encodedData)
        {
            writer.WriteLine();
            writer.WriteLine(PdfNames.PdfStreamStart);
            writer.Write(encodedData);
            writer.WriteLine();
            writer.Write(PdfNames.PdfStreamEnd);
        }
    }
}
