﻿#if !NETSTANDARD
using System.Windows;
using System.Windows.Media;
#else
using Telerik.Documents.Primitives;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Navigation;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Destinations
{
    internal class DestinationObject : PdfArrayObject, IDestinationObject
    {
        private readonly ReferenceProperty<Page> page;
        private readonly DirectProperty<PdfName> destinationType;
        private readonly DirectProperty<PdfReal> left;
        private readonly DirectProperty<PdfReal> top;
        private readonly DirectProperty<PdfReal> bottom;
        private readonly DirectProperty<PdfReal> right;
        private readonly DirectProperty<PdfReal> zoom;

        public DestinationObject()
        {
            this.page = this.RegisterReferenceProperty<Page>(new PdfPropertyDescriptor("Page", true, PdfPropertyRestrictions.MustBeIndirectReference));
            this.destinationType = this.RegisterDirectProperty<PdfName>(new PdfPropertyDescriptor("DestinationType", true));
            this.left = this.RegisterDirectProperty<PdfReal>(new PdfPropertyDescriptor("Left"));
            this.bottom = this.RegisterDirectProperty<PdfReal>(new PdfPropertyDescriptor("Bottom"));
            this.right = this.RegisterDirectProperty<PdfReal>(new PdfPropertyDescriptor("Right"));
            this.top = this.RegisterDirectProperty<PdfReal>(new PdfPropertyDescriptor("Top"));
            this.zoom = this.RegisterDirectProperty<PdfReal>(new PdfPropertyDescriptor("Zoom"));
        }

        public Page Page
        {
            get
            {
                return this.page.GetValue();
            }
            set
            {
                this.page.SetValue(value);
            }
        }

        public PdfName DestinationType
        {
            get
            {
                return this.destinationType.GetValue();
            }
            set
            {
                this.destinationType.SetValue(value);
            }
        }

        public PdfReal Left
        {
            get
            {
                return this.left.GetValue();
            }
            set
            {
                this.left.SetValue(value);
            }
        }

        public PdfReal Bottom
        {
            get
            {
                return this.bottom.GetValue();
            }
            set
            {
                this.bottom.SetValue(value);
            }
        }

        public PdfReal Right
        {
            get
            {
                return this.right.GetValue();
            }
            set
            {
                this.right.SetValue(value);
            }
        }

        public PdfReal Top
        {
            get
            {
                return this.top.GetValue();
            }
            set
            {
                this.top.SetValue(value);
            }
        }

        public PdfReal Zoom
        {
            get
            {
                return this.zoom.GetValue();
            }
            set
            {
                this.zoom.SetValue(value);
            }
        }

        bool IDestinationObject.IsNamed
        {
            get
            {
                return false;
            }
        }

        public void CopyPropertiesFrom(IPdfExportContext context, Destination destination)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(destination, "destination");

            Page mappedPage;
            RadFixedPage fixedPage = destination.Page;

            if (fixedPage != null && context.TryGetPage(fixedPage, out mappedPage))
            {
                this.Page = mappedPage;
            }

            Matrix matrix = fixedPage != null ? context.GetDipToPdfPointTransformation(fixedPage) : new Matrix();
            this.DestinationType = new PdfName(DestinationTypes.GetDestinationType(destination.DestinationType));

            switch (destination.DestinationType)
            {
                case Fixed.Model.Navigation.DestinationType.Location:
                    Location location = destination as Location;
                    this.Left = GetXInPdfPoint(matrix, location.Left);
                    this.Top = GetYInPdfPoint(matrix, location.Top);
                    this.Zoom = location.Zoom.HasValue ? location.Zoom.Value.ToPdfReal() : null;

                    this.IgnoreProperties(this.right, this.bottom);

                    break;

                case Fixed.Model.Navigation.DestinationType.PageHorizontalFit:
                    PageHorizontalFit pageHorizontalFit = destination as PageHorizontalFit;
                    this.Top = GetYInPdfPoint(matrix, pageHorizontalFit.Top);

                    this.IgnoreProperties(this.left, this.bottom, this.right, this.zoom);

                    break;

                case Fixed.Model.Navigation.DestinationType.PageVerticalFit:
                    PageVerticalFit pageVerticalFit = destination as PageVerticalFit;
                    this.Left = GetXInPdfPoint(matrix, pageVerticalFit.Left);

                    this.IgnoreProperties(this.top, this.bottom, this.right, this.zoom);

                    break;

                case Fixed.Model.Navigation.DestinationType.RectangleFit:
                    RectangleFit rectangleFit = destination as RectangleFit;
                    this.Left = GetXInPdfPoint(matrix, rectangleFit.Left);
                    this.Bottom = GetYInPdfPoint(matrix, rectangleFit.Bottom);
                    this.Right = GetXInPdfPoint(matrix, rectangleFit.Right);
                    this.Top = GetYInPdfPoint(matrix, rectangleFit.Top);

                    this.IgnoreProperties(this.zoom);

                    break;

                case Fixed.Model.Navigation.DestinationType.BoundingRectangleHorizontalFit:
                    BoundingRectangleHorizontalFit boundingRectangleHorizontalFit = destination as BoundingRectangleHorizontalFit;
                    this.Top = GetYInPdfPoint(matrix, boundingRectangleHorizontalFit.Top);

                    this.IgnoreProperties(this.left, this.right, this.bottom, this.zoom);

                    break;

                case Fixed.Model.Navigation.DestinationType.BoundingRectangleVerticalFit:
                    BoundingRectangleVerticalFit boundingRectangleVerticalFit = destination as BoundingRectangleVerticalFit;
                    this.Left = GetXInPdfPoint(matrix, boundingRectangleVerticalFit.Left);

                    this.IgnoreProperties(this.top, this.right, this.bottom, this.zoom);

                    break;

                case Fixed.Model.Navigation.DestinationType.PageFit:
                    {
                        if (this.Top == null)
                        {
                            this.IgnoreProperty(this.top);
                        }

                        if (this.Bottom == null)
                        {
                            this.IgnoreProperty(this.bottom);
                        }

                        if (this.Left == null)
                        {
                            this.IgnoreProperty(this.left);
                        }

                        if (this.Right == null)
                        {
                            this.IgnoreProperty(this.right);
                        }

                        if (this.Zoom == null)
                        {
                            this.IgnoreProperty(this.zoom);
                        }
                    }
                    break;
            }
        }

        public Destination ToDestination(PostScriptReader reader, IRadFixedDocumentImportContext context)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            DestinationType destinationType = DestinationTypes.GetDestinationType(this.DestinationType.Value);
            RadFixedPage destinationPage = this.Page != null ? context.GetFixedPage(this.Page) : null;
            Matrix matrix = this.Page != null ? context.GetPdfPointToDipTransformation(this.Page) : new Matrix();

            switch (destinationType)
            {
                case Fixed.Model.Navigation.DestinationType.Location:
                    return new Location()
                    {
                        Page = destinationPage,
                        Left = GetXInDip(matrix, this.Left),
                        Top = GetYInDip(matrix, this.Top),
                        Zoom = (this.Zoom != null && this.Zoom.Value > 0) ? ((double?)this.Zoom.Value) : null
                    };

                case Fixed.Model.Navigation.DestinationType.PageHorizontalFit:
                    return new PageHorizontalFit()
                    {
                        Page = destinationPage,
                        Top = GetYInDip(matrix, this.Top)
                    };

                case Fixed.Model.Navigation.DestinationType.PageVerticalFit:
                    return new PageVerticalFit()
                    {
                        Page = destinationPage,
                        Left = GetXInDip(matrix, this.Left)
                    };

                case Fixed.Model.Navigation.DestinationType.RectangleFit:
                    return new RectangleFit()
                    {
                        Page = destinationPage,
                        Left = GetXInDip(matrix, this.Left),
                        Bottom = GetYInDip(matrix, this.Bottom),
                        Right = GetXInDip(matrix, this.Right),
                        Top = GetYInDip(matrix, this.Top)
                    };

                case Fixed.Model.Navigation.DestinationType.BoundingRectangleHorizontalFit:
                    return new BoundingRectangleHorizontalFit()
                    {
                        Page = destinationPage,
                        Top = GetYInDip(matrix, this.Top)
                    };

                case Fixed.Model.Navigation.DestinationType.BoundingRectangleVerticalFit:
                    return new BoundingRectangleVerticalFit()
                    {
                        Page = destinationPage,
                        Left = GetXInDip(matrix, this.Left)
                    };

                default:
                    return Destination.CreateDestination(destinationType, destinationPage);
            }
        }

        private static double? GetXInDip(Matrix pointToDipMatrix, PdfReal xCoordinateInPdfPoint)
        {
            if (xCoordinateInPdfPoint != null)
            {
                return pointToDipMatrix.Transform(new Point(xCoordinateInPdfPoint.Value, 0)).X;
            }

            return null;
        }

        private static double? GetYInDip(Matrix pointToDipMatrix, PdfReal yCoordinateInPdfPoint)
        {
            if (yCoordinateInPdfPoint != null)
            {
                return pointToDipMatrix.Transform(new Point(0, yCoordinateInPdfPoint.Value)).Y;
            }

            return null;
        }

        private static PdfReal GetXInPdfPoint(Matrix dipToPointMatrix, double? xCoordinateInDip)
        {
            if (xCoordinateInDip.HasValue)
            {
                return dipToPointMatrix.Transform(new Point(xCoordinateInDip.Value, 0)).X.ToPdfReal();
            }

            return null;
        }

        private static PdfReal GetYInPdfPoint(Matrix dipToPointMatrix, double? yCoordinateInDip)
        {
            if (yCoordinateInDip.HasValue)
            {
                return dipToPointMatrix.Transform(new Point(0, yCoordinateInDip.Value)).Y.ToPdfReal();
            }

            return null;
        }
    }
}

