﻿using System;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Actions;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.Model.Actions;
using Telerik.Windows.Documents.Fixed.Model.Navigation;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Destinations
{
    internal class NamedDestinationObject : PdfObject, IDestinationObject
    {
        private IPdfPrimitive store;
        private DestinationObject destination;
        private GoToActionObject goToAction;

        public string Name { get; internal set; }

        bool IDestinationObject.IsNamed
        {
            get
            {
                return true;
            }
        }

        public DestinationObject Destination
        {
            get
            {
                this.ThrowExceptionIfNotLoaded();

                return this.destination;
            }
            internal set
            {
                this.destination = value;
            }
        }

        public GoToActionObject GoToAction
        {
            get
            {
                this.ThrowExceptionIfNotLoaded();

                return this.goToAction;
            }
            internal set
            {
                this.goToAction = value;
            }
        }

        public override void Write(PdfWriter writer, IPdfExportContext context)
        {
            PdfLiteralString name = new PdfLiteralString(this.Name, StringType.UTFEncodedTextString);
            name.Write(writer, context);
        }

        internal void SetDestinationPrimitive(string name, IPdfPrimitive primitive)
        {
            Guard.ThrowExceptionIfNull(name, "name");
            Guard.ThrowExceptionIfNull(primitive, "primitive");

            this.Name = name;
            this.store = primitive;
        }

        internal void CopyPropertiesFrom(IPdfExportContext context, NamedDestination namedDestination)
        {
            this.Name = namedDestination.Name;

            if (namedDestination.Destination != null)
            {
                this.Destination = new DestinationObject();
                this.Destination.CopyPropertiesFrom(context, namedDestination.Destination);
            }

            if (namedDestination.GoToAction != null)
            {
                this.GoToAction = new GoToActionObject();
                this.GoToAction.CopyPropertiesFrom(context, namedDestination.GoToAction);
            }
        }

        internal void CopyPropertiesTo(PostScriptReader reader, IRadFixedDocumentImportContext context, NamedDestination namedDestination)
        {
            this.LoadFromStore(reader, context);

            namedDestination.Name = this.Name;

            DestinationObject destinationObject = this.Destination;
            GoToActionObject goToActionObject = this.GoToAction;
            if (destinationObject != null)
            {
                Destination destination = destinationObject.ToDestination(reader, context);

                namedDestination.Destination = destination;
            }
            else if (goToActionObject != null)
            {
                GoToAction goToAction = (GoToAction)goToActionObject.ToAction(reader, context);
                namedDestination.GoToAction = goToAction;
            }
        }

        private void LoadFromStore(PostScriptReader reader, IRadFixedDocumentImportContext context)
        {
            IPdfPrimitive value = this.store;
            IndirectReference indirectReference = value as IndirectReference;
            if (indirectReference != null)
            {
                if (!context.TryGetIndirectObject(indirectReference, out value))
                {
                    IndirectObject indirectObject = context.ReadIndirectObject(indirectReference);
                    value = indirectObject.Content;
                    context.RegisterIndirectObject(indirectReference, value);
                }
            }

            GoToActionObject goToAction;
            DestinationObject destination;
            if (this.TryConvert(reader, context, value, out goToAction))
            {
                this.GoToAction = goToAction;
            }
            else if (this.TryConvert(reader, context, value, out destination))
            {
                this.Destination = destination;
            }
            else
            {
                throw new NotSupportedException("Not supported named destination");
            }
        }

        private bool TryConvert<T>(PostScriptReader reader, IRadFixedDocumentImportContext context,
            IPdfPrimitive value, out T result)
            where T : class, IPdfPrimitive
        {
            PdfObjectDescriptor goToActionDescriptor = PdfObjectDescriptors.GetPdfObjectDescriptor<T>();
            IPdfPrimitive converted = goToActionDescriptor.Converter.Convert(typeof(T), reader, context, value);
            result = converted as T;

            return result != null;
        }

        internal NamedDestination ToNamedDestination(IRadFixedDocumentImportContext context)
        {
            return context.Document.NamedDestinations[this.Name];
        }

        private void ThrowExceptionIfNotLoaded()
        {
            if (this.destination == null && this.goToAction == null && this.store != null)
            {
                throw new InvalidOperationException("Named destination not loaded");
            }
        }
    }
}
