﻿using System;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.Model.DigitalSignatures;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Utilities;
using System.Linq;
using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using System.Security.Cryptography.X509Certificates;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DigitalSignature
{
    internal class SignatureObject : PdfObject
    {
        private readonly DirectProperty<PdfName> filter;
        private readonly DirectProperty<PdfName> subFilter;
        private readonly DirectProperty<SignatureContents> contents;
        private readonly DirectProperty<PdfArray> cert;
        private readonly DirectProperty<SignatureByteRange> byteRange;
        private readonly ReferenceProperty<PdfArray> reference;
        private readonly DirectProperty<PdfArray> changes;
        private readonly DirectProperty<PdfString> name;
        private readonly DirectProperty<PdfString> location;
        private readonly DirectProperty<PdfString> reason;
        private readonly DirectProperty<PdfString> contactInfo;
        private readonly DirectProperty<PdfString> timeOfSigning;
        private readonly DirectProperty<PdfInt> signatureHandlerVersion;
        private readonly DirectProperty<PdfInt> signatureDictionaryVersion;
        private string signatureFieldName;

        public SignatureObject()
        {
            this.filter = this.RegisterDirectProperty<PdfName>(new PdfPropertyDescriptor("Filter"));
            this.subFilter = this.RegisterDirectProperty<PdfName>(new PdfPropertyDescriptor("SubFilter"));
            this.byteRange = this.RegisterDirectProperty<SignatureByteRange>(new PdfPropertyDescriptor("ByteRange"));
            this.contents = this.RegisterDirectProperty<SignatureContents>(new PdfPropertyDescriptor("Contents"));
            this.cert = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor("Cert"));

            this.reference = this.RegisterReferenceProperty<PdfArray>(new PdfPropertyDescriptor("Reference"));

            this.changes = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor("Changes"));
            this.name = this.RegisterDirectProperty<PdfString>(new PdfPropertyDescriptor("Name"));
            this.timeOfSigning = this.RegisterDirectProperty<PdfString>(new PdfPropertyDescriptor("M"));
            this.location = this.RegisterDirectProperty<PdfString>(new PdfPropertyDescriptor("Location"));
            this.reason = this.RegisterDirectProperty<PdfString>(new PdfPropertyDescriptor("Reason"));
            this.contactInfo = this.RegisterDirectProperty<PdfString>(new PdfPropertyDescriptor("ContactInfo"));
            this.signatureHandlerVersion = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("R"));
            this.signatureDictionaryVersion = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("V"));
        }

        public PdfName Filter
        {
            get
            {
                return this.filter.GetValue();
            }
            set
            {
                this.filter.SetValue(value);
            }
        }

        public PdfName SubFilter
        {
            get
            {
                return this.subFilter.GetValue();
            }
            set
            {
                this.subFilter.SetValue(value);
            }
        }

        public SignatureContents Contents
        {
            get
            {
                return this.contents.GetValue();
            }
            set
            {
                this.contents.SetValue(value);
            }
        }

        public PdfArray Cert
        {
            get
            {
                return this.cert.GetValue();
            }
            set
            {
                this.cert.SetValue(value);
            }
        }

        public SignatureByteRange ByteRange
        {
            get
            {
                return this.byteRange.GetValue();
            }
            set
            {
                this.byteRange.SetValue(value);
            }
        }

        public PdfArray Reference
        {
            get
            {
                return this.reference.GetValue();
            }
            set
            {
                this.reference.SetValue(value);
            }
        }

        public PdfArray Changes
        {
            get
            {
                return this.changes.GetValue();
            }
            set
            {
                this.changes.SetValue(value);
            }
        }

        public PdfString Name
        {
            get
            {
                return this.name.GetValue();
            }
            set
            {
                this.name.SetValue(value);
            }
        }

        public PdfString TimeOfSigning
        {
            get
            {
                return this.timeOfSigning.GetValue();
            }
            set
            {
                this.timeOfSigning.SetValue(value);
            }
        }

        public PdfString Location
        {
            get
            {
                return this.location.GetValue();
            }
            set
            {
                this.location.SetValue(value);
            }
        }

        public PdfString Reason
        {
            get
            {
                return this.reason.GetValue();
            }
            set
            {
                this.reason.SetValue(value);
            }
        }

        public PdfString ContactInfo
        {
            get
            {
                return this.contactInfo.GetValue();
            }
            set
            {
                this.contactInfo.SetValue(value);
            }
        }

        public PdfInt SignatureHandlerVersion
        {
            get
            {
                return this.signatureHandlerVersion.GetValue();
            }
            set
            {
                this.signatureHandlerVersion.SetValue(value);
            }
        }

        public PdfInt SignatureDictionaryVersion
        {
            get
            {
                return this.signatureDictionaryVersion.GetValue();
            }
            set
            {
                this.signatureDictionaryVersion.SetValue(value);
            }
        }

        internal void CopyPropertiesTo(Signature signature, PostScriptReader reader, IRadFixedDocumentImportContext context)
        {
            Guard.ThrowExceptionIfNull(signature, "signature");
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            SignatureDataPdfProperties properties = signature.Properties.PdfProperties;

            this.Filter.CopyToProperty(properties.Filter, (filter) => filter.Value);
            this.SubFilter.CopyToProperty(properties.SubFilter, (subFilter) => subFilter.Value);
            this.Contents.CopyToProperty(properties.Contents, (contents) => contents.Value);
            this.Cert.CopyToProperty(properties.Cert, (cert) => ToCertificateCollection(cert));
            this.ByteRange.CopyToProperty(properties.ByteRange, (byteRange) => ((PdfArray)byteRange.Primitive).ToIntArray());

            this.Reference.CopyToProperty(properties.Reference, (reference) => SignatureObject.ReferenceToModelElements(reference, reader, context));

            this.Changes.CopyToProperty(properties.Changes, (changes) => changes.ToIntArray());
            this.Name.CopyToProperty(properties.Name, (name) => name.ToString());

            if (this.TimeOfSigning != null)
            {
                DateTime? dateTime = SignatureDateCodec.Decode(this.TimeOfSigning.ToString());
                if (dateTime != null)
                {
                    properties.TimeOfSigning.Value = dateTime.Value;
                }
            }

            this.Location.CopyToProperty(properties.Location, (location) => location.ToString());
            this.Reason.CopyToProperty(properties.Reason, (reason) => reason.ToString());
            this.ContactInfo.CopyToProperty(properties.ContactInfo, (contactInfo) => contactInfo.ToString());
            this.SignatureHandlerVersion.CopyToProperty(properties.SignatureHandlerVersion, (signatureHandlerVersion) => signatureHandlerVersion.Value);
            this.SignatureDictionaryVersion.CopyToProperty(properties.SignatureDictionaryVersion, (signatureDictionaryVersion) => signatureDictionaryVersion.Value);

            FileSourceStream sourceStream = new FileSourceStream(reader.Reader.Stream);
            signature.Properties.SourceStream = sourceStream;
        }

        public void CopyPropertiesFrom(Signature signature)
        {
            Guard.ThrowExceptionIfNull(signature, "signature");
            Guard.ThrowExceptionIfNull(signature.Properties, "signature.Properties");

            this.signatureFieldName = signature.Properties.FieldName;

            SignatureDataPdfProperties properties = signature.Properties.PdfProperties;

            this.Filter = properties.Filter.ToPrimitive((filter) => filter.ToPdfName());
            this.SubFilter = properties.SubFilter.ToPrimitive((filter) => filter.ToPdfName());
            this.Contents = properties.Contents.ToPrimitive((contents) => new SignatureContents(contents));
            this.Cert = properties.Cert.ToPrimitive((cert) => FromCertificateCollection(cert));

            this.ByteRange = SignatureByteRange.ToSignatureByteRange(properties.ByteRange);

            this.Reference = SignatureObject.ToPdfArray(properties.Reference.Value);

            this.Changes = properties.Changes.ToPrimitive((changes) => changes.ToPdfArray());
            this.Name = properties.Name.ToPrimitive((name) => name.ToPdfLiteralString(StringType.ASCII));
            this.TimeOfSigning = properties.TimeOfSigning.ToPrimitive((timeOfSigning) => SignatureDateCodec.Encode(timeOfSigning).ToPdfLiteralString(StringType.ASCII));
            this.Location = properties.Location.ToPrimitive((location) => location.ToPdfLiteralString(StringType.ASCII));
            this.Reason = properties.Reason.ToPrimitive((reason) => reason.ToPdfLiteralString(StringType.ASCII));
            this.ContactInfo = properties.ContactInfo.ToPrimitive((contactInfo) => contactInfo.ToPdfLiteralString(StringType.ASCII));
            this.SignatureHandlerVersion = properties.SignatureHandlerVersion.ToPrimitive((signatureHandlerVersion) => signatureHandlerVersion.ToPdfInt());
            this.SignatureDictionaryVersion = properties.SignatureDictionaryVersion.ToPrimitive((signatureDictionaryVersion) => signatureDictionaryVersion.ToPdfInt());
        }

        private static List<SignatureReference> ReferenceToModelElements(PdfArray array, PostScriptReader reader, IRadFixedDocumentImportContext context)
        {
            Guard.ThrowExceptionIfNull(reader, "array");
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            List<SignatureReference> result = new List<SignatureReference>();
            foreach (var arrayItem in array)
            {
                SignatureReference signatureReference = new SignatureReference();

                SignatureReferenceObject signatureReferenceElement = new SignatureReferenceObject();
                signatureReferenceElement.Load(reader, context, arrayItem);
                signatureReferenceElement.CopyPropertiesTo(signatureReference, reader, context);

                result.Add(signatureReference);
            }

            return result;
        }

        private static PdfArray ToPdfArray(IEnumerable<SignatureReference> array)
        {
            if (array == null)
            {
                return null;
            }

            PdfArray result = new PdfArray();
            foreach (var value in array)
            {
                SignatureReferenceObject signatureReferenceElement = new SignatureReferenceObject();
                signatureReferenceElement.CopyPropertiesFrom(value);

                result.Add(signatureReferenceElement);
            }

            return result;
        }

        private X509Certificate2Collection ToCertificateCollection(PdfArray array)
        {
            X509Certificate2Collection certificateCollection = new X509Certificate2Collection();
            for (int i = 0; i < array.Count; i++)
            {
                byte[] rawData = ((PdfString)array[i]).Value;
                X509Certificate2 cert = new X509Certificate2(rawData);
                certificateCollection.Add(cert);
            }

            return certificateCollection;
        }

        private PdfArray FromCertificateCollection(X509Certificate2Collection certificateCollection)
        {
            PdfArray array = new PdfArray();

            for (int i = 0; i < certificateCollection.Count; i++)
            {
                PdfLiteralString pdfLiteralString = new PdfLiteralString(certificateCollection[i].RawData);
                array.Add(pdfLiteralString);
            }

            return array;
        }

        public override void Write(PdfWriter writer, IPdfExportContext context)
        {
            context.SignatureExportInfo.SetSignatureFieldName(this.signatureFieldName);

            base.Write(writer, context);

            context.SignatureExportInfo.ClearSignatureFieldName();
        }
    }
}
