﻿using System;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.DigitalSignatures;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DigitalSignature
{
    internal class SignatureReferenceObject : PdfObject
    {
        private readonly DirectProperty<PdfName> transformMethod;
        private readonly DirectProperty<PrimitiveWrapper> transformParams;
        private readonly ReferenceProperty<PrimitiveWrapper> data;
        private readonly DirectProperty<PdfArray> digestLocation;
        private readonly DirectProperty<PdfName> digestMethod;
        private readonly DirectProperty<PdfString> digestValue;

        public SignatureReferenceObject()
        {
            this.transformMethod = this.RegisterDirectProperty<PdfName>(new PdfPropertyDescriptor("TransformMethod", isRequired: true, alwaysExport: true));
            this.transformParams = this.RegisterDirectProperty<PrimitiveWrapper>(new PdfPropertyDescriptor("TransformParams"));
            this.data = this.RegisterReferenceProperty<PrimitiveWrapper>(new PdfPropertyDescriptor("Data"));
            this.digestMethod = this.RegisterDirectProperty<PdfName>(new PdfPropertyDescriptor("DigestMethod"));
            this.digestValue = this.RegisterDirectProperty<PdfString>(new PdfPropertyDescriptor("DigestValue"));
            this.digestLocation = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor("DigestLocation"));
        }

        public PdfName TransformMethod
        {
            get
            {
                return this.transformMethod.GetValue();
            }
            set
            {
                this.transformMethod.SetValue(value);
            }
        }

        public PrimitiveWrapper TransformParams
        {
            get
            {
                return this.transformParams.GetValue();
            }
            set
            {
                this.transformParams.SetValue(value);
            }
        }

        public PrimitiveWrapper Data
        {
            get
            {
                return this.data.GetValue();
            }
            set
            {
                this.data.SetValue(value);
            }
        }

        public PdfName DigestMethod
        {
            get
            {
                return this.digestMethod.GetValue();
            }
            set
            {
                this.digestMethod.SetValue(value);
            }
        }

        public PdfString DigestValue
        {
            get
            {
                return this.digestValue.GetValue();
            }
            set
            {
                this.digestValue.SetValue(value);
            }
        }

        public PdfArray DigestLocation
        {
            get
            {
                return this.digestLocation.GetValue();
            }
            set
            {
                this.digestLocation.SetValue(value);
            }
        }

        internal void CopyPropertiesFrom(SignatureReference signatureReference)
        {
            // TODO: implement export functionality
        }

        internal void CopyPropertiesTo(SignatureReference signatureReference, PostScriptReader reader, IRadFixedDocumentImportContext context)
        {
            Guard.ThrowExceptionIfNull(signatureReference, "signatureReference");
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            this.TransformMethod.CopyToProperty(signatureReference.TransformMethod, (transformMethod) => transformMethod.Value);

            if (signatureReference.TransformParameters != null)
            {
                signatureReference.TransformParameters.Value = ToTransformParameters(this.TransformMethod, this.TransformParams.Primitive, reader, context);
            }

            //this.data.CopyToProperty(signatureReference.Data, (data) => data);
            this.DigestMethod.CopyToProperty(signatureReference.DigestMethod, (digestMethod) => digestMethod.Value);
            this.DigestValue.CopyToProperty(signatureReference.DigestValue, (digestValue) => digestValue.Value);
            this.DigestLocation.CopyToProperty(signatureReference.DigestLocation, (digestLocation) => digestLocation.ToIntArray());
        }

        internal static TransformParameters ToTransformParameters(PdfName transformMethod, IPdfPrimitive primitive, PostScriptReader reader, IRadFixedDocumentImportContext context)
        {
            TransformParameters transformParameters = null;
            TransformParametersElementBase transformParametersElement = null;

            switch (transformMethod.Value)
            {
                case TransformMethods.DocMDP:
                    transformParametersElement = new DocMdpTransformParametersElement();
                    transformParameters = new DocMdpTransformParameters();
                    break;
                case TransformMethods.UR:
                    transformParametersElement = new UrTransformParametersElement();
                    transformParameters = new UrTransformParameters();
                    break;
                case TransformMethods.FieldMDP:
                    transformParametersElement = new FieldMdpTransformParametersElement();
                    transformParameters = new FieldMdpTransformParameters();
                    break;
                case TransformMethods.Identity:
                    transformParametersElement = null;
                    break;
                default:
                    throw new NotSupportedException(string.Format("Not supported transform method: {0}", transformMethod.Value));
            }

            if(transformParametersElement == null)
            {
                return null;
            }

            transformParametersElement.CopyPropertiesTo(transformParameters);

            return transformParameters;
        }
    }
}
