﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure
{
    internal class CrossReferenceStream : Trailer
    {
        private const int DefaultEntryType = 1;
        private readonly DirectProperty<PdfArray> index;
        private readonly DirectProperty<PdfArray> w;

        public CrossReferenceStream()
            : base()
        {
            this.index = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor("Index", false, PdfPropertyRestrictions.MustBeDirectObject));
            this.w = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor("W", true, PdfPropertyRestrictions.MustBeDirectObject));
        }

        public PdfArray Index
        {
            get
            {
                return this.index.GetValue();
            }
            set
            {
                this.index.SetValue(value);
            }
        }

        public PdfArray W
        {
            get
            {
                return this.w.GetValue();
            }
            set
            {
                this.w.SetValue(value);
            }
        }

        protected override void LoadOverride(PostScriptReader reader, IPdfImportContext context, IPdfPrimitive primitive)
        {
            switch (primitive.Type)
            {
                case PdfElementType.Stream:
                    PdfStream stream = (PdfStream)primitive;
                    this.Load(reader, context, stream.Dictionary);
                    byte[] data = stream.ReadDecodedPdfData();
                    this.LoadReferences(data, context);
                    break;
                default:
                    throw new InvalidOperationException("CrossReferenceStream should be loaded from stream object!");
            }
        }

        private void LoadReferences(byte[] data, IPdfImportContext context)
        {
            int[] index = this.CalculateIndexArrayForImport();
            int[] w = this.CalculateWArray();

            int sectionsCount = index.Length / 2;
            int position = 0;

            for (int section = 0; section < sectionsCount; section++)
            {
                int startObjectNumber = index[section * 2];
                int countOfObjectsInSection = index[section * 2 + 1];

                for (int i = 0; i < countOfObjectsInSection; i++)
                {
                    CrossReferenceEntry entry = new CrossReferenceEntry();

                    int type = w[0] > 0 ? CrossReferenceStream.GetInt(data, position, w[0]) : CrossReferenceStream.DefaultEntryType;
                    entry.Type = CrossReferenceStream.GetEntyType(type);
                    position += w[0];

                    entry.Field1 = CrossReferenceStream.GetInt(data, position, w[1]);
                    position += w[1];

                    entry.Field2 = CrossReferenceStream.GetInt(data, position, w[2]);
                    position += w[2];

                    int objectNumber = startObjectNumber + i;
                    context.CrossReferences.AddCrossReferenceEntry(objectNumber, entry);
                }
            }
        }

        public override void Write(PdfWriter writer, IPdfExportContext context)
        {
            KeyValuePair<int, CrossReferenceEntry>[] orderedCrossReferenceCollection = context.CrossReferenceCollection.Entries.OrderBy(item => item.Key).ToArray();

            this.Index = CrossReferenceStream.CalculateIndexArrayForExport(context, orderedCrossReferenceCollection);

            int[] w = new int[3];
            w[0] = 1;
            w[1] = CrossReferenceStream.CalculateMinBytesLength(writer.Position);
            w[2] = 1;

            this.W = w.ToPdfArray();

            PdfObject.WritePdfPropertiesDictionary(this, writer, context);

            int bytesNeeded = 0;
            for (int i = 0; i < w.Length; i++)
            {
                bytesNeeded += w[i];
            }

            int dataLength = orderedCrossReferenceCollection.Length * bytesNeeded;
            byte[] data = new byte[dataLength];

            int position = 0;
            foreach (var pair in orderedCrossReferenceCollection)
            {
                CrossReferenceEntry entry = pair.Value;

                CrossReferenceStream.SetInt(data, position, w[0], entry.Type == CrossReferenceEntryType.Used ? 1 : 0);
                position += w[0];

                CrossReferenceStream.SetInt(data, position, w[1], entry.Field1);
                position += w[1];

                CrossReferenceStream.SetInt(data, position, w[2], entry.Field2);
                position += w[2];
            }

            PdfStreamObjectBase.WriteEncodedData(writer, context, data);
        }

        private static PdfArray CalculateIndexArrayForExport(IPdfExportContext context, KeyValuePair<int, CrossReferenceEntry>[] orderedCrossReferenceCollection)
        {
            List<int> result = new List<int>();

            int? firstEntryInSectionKey = null;
            int prevEntryKey = 0;
            int entriesInSectionCount = 0;

            foreach (var pair in orderedCrossReferenceCollection)
            {
                if (!firstEntryInSectionKey.HasValue)
                {
                    firstEntryInSectionKey = pair.Key;
                    entriesInSectionCount = 1;
                }
                else if (prevEntryKey + 1 == pair.Key)
                {
                    entriesInSectionCount++;
                }
                else
                {
                    result.Add(firstEntryInSectionKey.Value);
                    result.Add(entriesInSectionCount);

                    firstEntryInSectionKey = pair.Key;
                    entriesInSectionCount = 1;
                }

                prevEntryKey = pair.Key;
            }

            result.Add(firstEntryInSectionKey.Value);
            result.Add(entriesInSectionCount);

            return result.ToPdfArray();
        }

        private static CrossReferenceEntryType GetEntyType(int type)
        {
            switch (type)
            {
                case 1:
                    return CrossReferenceEntryType.Used;
                case 2:
                    return CrossReferenceEntryType.Compressed;
                default:
                    return CrossReferenceEntryType.Free;
            }
        }

        private static int GetInt(byte[] data, int position, int length)
        {
            int number = 0;

            for (int i = 0; i < length; i++)
            {
                byte b = data[position + i];
                int shiftAmount = length - 1 - i;
                int shiftResult = (b << (shiftAmount * 8));
                number |= shiftResult;
            }

            return number;
        }

        private static int CalculateMinBytesLength(long number)
        {
            int bitsCount = 64;
            int maxNonZeroBit = 0;

            for (int i = 0; i < bitsCount; i++)
            {
                long result = number >> i;
                result &= 1;
                if (result == 1)
                {
                    maxNonZeroBit = i;
                }
            }

            int minBytesLength = (maxNonZeroBit / 8) + 1;

            return minBytesLength;
        }

        /// <summary>
        /// Sets a calculated byte at the specified position in the data byte array for the parameter number.
        /// This is reverse of the GetInt method.
        /// </summary>
        /// <param name="data">The data to be filled.</param>
        /// <param name="position">The position in the data array.</param>
        /// <param name="length">The length in bytes reserved for a specific field value from W.</param>
        /// <param name="number">The number aka field value.</param>
        private static void SetInt(byte[] data, int position, int length, long number)
        {
            for (int i = 0; i < length; i++)
            {
                int shiftAmount = length - 1 - i;
                long shiftResult = (number >> (shiftAmount * 8));
                byte b = (byte)shiftResult;
                data[position + i] = b;
            }
        }

        private int[] CalculateWArray()
        {
            return new int[] { ((PdfInt)this.W[0]).Value, ((PdfInt)this.W[1]).Value, ((PdfInt)this.W[2]).Value };
        }

        private int[] CalculateIndexArrayForImport()
        {
            if (this.Index == null)
            {
                return new int[] { 0, this.Size.Value };
            }
            else
            {
                int[] index = new int[this.Index.Count];

                for (int i = 0; i < index.Length; i++)
                {
                    index[i] = ((PdfInt)this.Index[i]).Value;
                }

                return index;
            }
        }
    }
}
