﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Destinations;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Forms;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Metadata;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.NamedTrees;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Outlines;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Collections;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Fixed.Model.Navigation;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure
{
    internal class DocumentCatalog : PdfObject
    {
        public const string PagesName = "Pages";
        public const string MetadataName = "Metadata";
        public const string OutputIntentsName = "OutputIntents";
        public const string AcroFormName = "AcroForm";
        public const string DestsName = "Dests";
        public const string NamesName = "Names";

        private const string OutlinesName = "Outlines";
        private const string PageModeName = "PageMode";
        private const string UseOutlines = "UseOutlines";
        private const string UseNone = "UseNone";

        private readonly ReferenceProperty<PageTreeNode> pages;
        private readonly ReferenceProperty<DocumentInfo> documentInfo;
        private readonly ReferenceProperty<PdfArray> outputIntents;
        private readonly ReferenceProperty<NamedDestinationsObject> dests;
        private readonly ReferenceProperty<AcroFormObject> acroForm;
        private readonly ReferenceProperty<OutlinesObject> outlines;
        private readonly ReferenceProperty<NamesObject> names;
        private readonly DirectProperty<PdfName> pageMode;

        public DocumentCatalog()
            : base()
        {
            this.pages = this.RegisterReferenceProperty<PageTreeNode>(new PdfPropertyDescriptor(PagesName, true, PdfPropertyRestrictions.MustBeIndirectReference));
            this.documentInfo = this.RegisterReferenceProperty<DocumentInfo>(new PdfPropertyDescriptor(MetadataName, false, PdfPropertyRestrictions.MustBeIndirectReference));
            this.outputIntents = this.RegisterReferenceProperty<PdfArray>(new PdfPropertyDescriptor(OutputIntentsName, false));
            this.dests = this.RegisterReferenceProperty<NamedDestinationsObject>(new PdfPropertyDescriptor(DestsName, false, PdfPropertyRestrictions.MustBeIndirectReference));
            this.names = this.RegisterReferenceProperty<NamesObject>(new PdfPropertyDescriptor(NamesName, false, PdfPropertyRestrictions.MustBeIndirectReference));
            this.acroForm = this.RegisterReferenceProperty<AcroFormObject>(new PdfPropertyDescriptor(AcroFormName, false, PdfPropertyRestrictions.MustBeIndirectReference));
            this.outlines = this.RegisterReferenceProperty<OutlinesObject>(new PdfPropertyDescriptor(OutlinesName, false, PdfPropertyRestrictions.MustBeIndirectReference));
            this.pageMode = this.RegisterDirectProperty<PdfName>(new PdfPropertyDescriptor(PageModeName), new PdfName(UseNone));
        }

        public PageTreeNode Pages
        {
            get
            {
                return this.pages.GetValue();
            }
            set
            {
                this.pages.SetValue(value);
            }
        }

        public DocumentInfo DocumentInfo
        {
            get
            {
                return this.documentInfo.GetValue();
            }
            set
            {
                this.documentInfo.SetValue(value);
            }
        }

        public PdfArray OutputIntents
        {
            get
            {
                return this.outputIntents.GetValue();
            }
            set
            {
                this.outputIntents.SetValue(value);
            }
        }

        public NamedDestinationsObject Dests
        {
            get
            {
                return this.dests.GetValue();
            }
            set
            {
                this.dests.SetValue(value);
            }
        }

        public NamesObject Names
        {
            get
            {
                return this.names.GetValue();
            }
            set
            {
                this.names.SetValue(value);
            }
        }

        public AcroFormObject AcroForm
        {
            get
            {
                return this.acroForm.GetValue();
            }
            set
            {
                this.acroForm.SetValue(value);
            }
        }

        internal OutlinesObject Outlines
        {
            get
            {
                return this.outlines.GetValue();
            }
            set
            {
                this.outlines.SetValue(value);
            }
        }

        internal PdfName PageMode
        {
            get
            {
                return this.pageMode.GetValue();
            }
            set
            {
                this.pageMode.SetValue(value);
            }
        }

        public void CopyPropertiesTo(PostScriptReader reader, IRadFixedDocumentImportContext context)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            this.CopyFormFieldsTo(reader, context);
            this.CopyPagePropertiesTo(reader, context);
            this.CopyOutlinesTo(reader, context);
            this.CopyPageModeTo(context);
        }

        public void CopyPropertiesFrom(IRadFixedDocumentExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            bool shouldExportAcroForm = context.Document.AcroForm.FormFields.Count > 0;

            if (shouldExportAcroForm)
            {
                this.CopyAcroFormProperties(context);
                context.PrepareWidgetAppearancesForExport();
            }

            this.CopyPagePropertiesFrom(context);

            if (shouldExportAcroForm)
            {
                this.CopyFieldPropertiesFrom(context);
            }

            this.CopyOutlinesPropertiesFrom(context);

            this.DocumentInfo = new DocumentInfo();
            this.DocumentInfo.CopyPropertiesFrom(context);

            this.CopyPageModeFrom(context);
        }

        private void CopyPageModeFrom(IRadFixedDocumentExportContext context)
        {
            string pageModeValue;

            switch (context.Document.PageMode)
            {
                case Fixed.Model.Common.PageMode.UseNone:
                    pageModeValue = UseNone;
                    break;
                case Fixed.Model.Common.PageMode.UseBookmarks:
                    pageModeValue = UseOutlines;
                    break;
                default:
                    pageModeValue = UseNone;
                    break;
            }

            this.PageMode = new PdfName(pageModeValue);
        }

        private void CopyAcroFormProperties(IRadFixedDocumentExportContext context)
        {
            this.AcroForm = new AcroFormObject();
            this.AcroForm.NeedAppearance = new PdfBool(context.Document.AcroForm.ViewersShouldRecalculateWidgetAppearances);
            this.AcroForm.Resources = new PdfResource();
            context.AcroFormContentExportContext = context.CreateContentExportContext(this.AcroForm, new EmptyContentRoot());
        }

        private void CopyFieldPropertiesFrom(IRadFixedDocumentExportContext context)
        {
            this.AcroForm.Fields = new FormFieldsTree();

            foreach (FormField field in context.Document.AcroForm.FormFields)
            {
                FormFieldNode node = context.CreateFormFieldObject(field, true);
                node.CopyFormField(field, context);

                this.AcroForm.Fields.Add(node);
#if WPF
                if (field.FieldType == FormFieldType.Signature && this.AcroForm.SigFlags == null)
                {
                    this.AcroForm.SetSignatureFlags();
                }
#endif
            }
        }

        private void CopyPagePropertiesFrom(IRadFixedDocumentExportContext context)
        {
            this.Pages = new PageTreeNode();
            if (context.Settings.ComplianceLevel != PdfComplianceLevel.None)
            {
                this.OutputIntents = new PdfArray(new OutputIntent());
            }

            PdfArray pagesCopies = new PdfArray();

            foreach (RadFixedPage fixedPage in context.Document.Pages)
            {
                Page page = new Page();
                page.Parent = this.Pages;
                context.MapPages(fixedPage, page);
            }

            foreach (RadFixedPage fixedPage in context.Document.Pages)
            {
                Page page;
                context.TryGetPage(fixedPage, out page);

                this.CopyRadFixedPageProperties(fixedPage, page, context);

                pagesCopies.Add(page);
            }

            this.CopyNamedDestinationsFrom(context);

            this.Pages.Children = pagesCopies;
            this.Pages.Count = new PdfInt();
            this.Pages.Count.Value = context.Document.Pages.Count;
        }

        private void CopyRadFixedPageProperties(RadFixedPage source, Page destination, IRadFixedDocumentExportContext context)
        {
            if (context.Document.CacheManager != null)
            {
                using (context.Document.CacheManager.BeginUsingPage(source))
                {
                    destination.CopyPropertiesFrom(context, source);
                }
            }
            else
            {
                destination.CopyPropertiesFrom(context, source);
            }
        }

        private void CopyOutlinesPropertiesFrom(IRadFixedDocumentExportContext context)
        {
            BookmarksCollection bookmarks = context.Document.Bookmarks;
            int bookmarksCount = bookmarks.Count;

            if (bookmarksCount == 0)
            {
                return;
            }

            this.Outlines = new OutlinesObject();
            this.Outlines.OutlineType = new PdfName(OutlinesName);
            this.Outlines.Count = new PdfInt(bookmarksCount);

            OutlineItemObject prevOutlineItemObject = null;

            for (int i = 0; i < bookmarksCount; i++)
            {
                BookmarkItem bookmarkItem = bookmarks[i];
                OutlineItemObject outlineItemObject = new OutlineItemObject();
                outlineItemObject.CopyPropertiesFrom(context, bookmarkItem);
                outlineItemObject.Parent = this.Outlines;

                if (prevOutlineItemObject != null)
                {
                    outlineItemObject.Prev = prevOutlineItemObject;
                    prevOutlineItemObject.Next = outlineItemObject;
                }

                if (i == 0)
                {
                    this.Outlines.First = outlineItemObject;
                }

                if (i == bookmarksCount - 1)
                {
                    this.Outlines.Last = outlineItemObject;
                }

                prevOutlineItemObject = outlineItemObject;
            }
        }

        private void CopyPageModeTo(IRadFixedDocumentImportContext context)
        {
            switch (this.PageMode.Value)
            {
                case UseOutlines:
                    context.Document.PageMode = Fixed.Model.Common.PageMode.UseBookmarks;
                    break;
                default:
                    context.Document.PageMode = Fixed.Model.Common.PageMode.UseNone;
                    break;
            }
        }

        private void CopyOutlinesTo(PostScriptReader reader, IRadFixedDocumentImportContext context)
        {
            if (this.Outlines != null)
            {
                OutlineItemObject outline = this.Outlines.First;

                while (outline != null)
                {
                    BookmarkItem bookmark = outline.ToBookmarkItem(reader, context);
                    context.Document.Bookmarks.Add(bookmark);

                    outline = outline.Next;
                }
            }
        }

        private void CopyFormFieldsTo(PostScriptReader reader, IRadFixedDocumentImportContext context)
        {
            if (this.AcroForm != null && this.AcroForm.Fields != null)
            {
                context.Document.AcroForm.ViewersShouldRecalculateWidgetAppearances = this.AcroForm.NeedAppearance.Value;

                foreach (FormFieldNode node in this.AcroForm.Fields)
                {
                    FormField field = node.ToFormField(reader, context, true);
                    context.Document.AcroForm.FormFields.Add(field);
                }
            }
        }

        private void CopyPagePropertiesTo(PostScriptReader reader, IRadFixedDocumentImportContext context)
        {
            List<Page> contextPages = context.Pages;
            IList<Page> pages = this.Pages.EnumeratePages(reader, context);

            switch (context.ImportSettings.ReadingMode)
            {
                case ReadingMode.AllAtOnce:
                    this.CopyAllPageProperties(reader, context, contextPages, pages);
                    break;
                case ReadingMode.OnDemand:
                    this.CopyLoadOnDemandPageProperties(reader, context, contextPages, pages);
                    break;
                default:
                    throw new InvalidOperationException("Unrecognized ReadingMode value - " + context.ImportSettings.ReadingMode);
            }
        }

        private void CopyNamedDestinationsTo(PostScriptReader reader, IRadFixedDocumentImportContext context)
        {
            List<NamedDestinationObject> store = null;
            if (this.Dests != null && this.Dests.Store != null)
            {
                store = this.Dests.Store;
            }
            else if (this.Names != null && this.Names.Dests != null)
            {
                if (this.Names.Dests.Destinations != null)
                {
                    store = this.Names.Dests.Destinations.Store;
                }
                else
                {
                    store = this.Names.Dests.EnumerateChildDestinations(reader, context).ToList();
                }
            }

            if (store == null)
            {
                return;
            }

            for (int i = 0; i < store.Count; i++)
            {
                NamedDestinationObject namedDestinationObject = store[i];
                context.MapNameToNamedDestination(namedDestinationObject.Name, namedDestinationObject);
            }

            for (int i = 0; i < store.Count; i++)
            {
                NamedDestinationObject namedDestinationObject = store[i];
                NamedDestination namedDestination = new NamedDestination();
                namedDestinationObject.CopyPropertiesTo(reader, context, namedDestination);
                context.Document.NamedDestinations.Add(namedDestination);
            }
        }

        private void CopyNamedDestinationsFrom(IRadFixedDocumentExportContext context)
        {
            this.Names = new NamesObject();
            this.Names.Dests = new DestinationsNamesTreeNode();
            NamedDestinationsArrayObject destinationsArray = new NamedDestinationsArrayObject();
            this.Names.Dests.Destinations = destinationsArray;

            foreach (NamedDestination destination in context.Document.NamedDestinations)
            {
                NamedDestinationObject namedDestinationObject = new NamedDestinationObject();
                namedDestinationObject.CopyPropertiesFrom(context, destination);
                destinationsArray.Store.Add(namedDestinationObject);
            }
        }

        private void CopyAllPageProperties(PostScriptReader reader, IRadFixedDocumentImportContext context, List<Page> contextPages, IList<Page> pages)
        {
            for (int i = 0; i < pages.Count; i++)
            {
                Page page = pages[i];
                RadFixedPage fixedPage = this.AddPage(context, contextPages, page, i);

                page.CopyPropertiesTo(reader, context, fixedPage);
                page.CopyPageContentTo(context, fixedPage);
            }

            this.CopyNamedDestinationsTo(reader, context);

            for (int i = 0; i < pages.Count; i++)
            {
                Page page = pages[i];
                RadFixedPage fixedPage = context.GetFixedPage(page);
                page.CopyPageAnnotationsTo(context, fixedPage);
            }
        }

        private void CopyLoadOnDemandPageProperties(PostScriptReader reader, IRadFixedDocumentImportContext context, List<Page> contextPages, IList<Page> pages)
        {
            for (int i = 0; i < pages.Count; i++)
            {
                Page page = pages[i];
                RadFixedPage fixedPage = this.AddPage(context, contextPages, page, i);

                page.CopyPropertiesTo(reader, context, fixedPage);
            }

            this.CopyNamedDestinationsTo(reader, context);
        }

        private RadFixedPage AddPage(IRadFixedDocumentImportContext context, List<Page> contextPages, Page page, int pageIndex)
        {
            RadFixedPage fixedPage = context.Document.Pages.AddPage();
            fixedPage.PageNumber = pageIndex + 1;
            context.MapPages(page, fixedPage);
            contextPages.Add(page);

            return fixedPage;
        }
    }
}
