﻿using System.Windows;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Annotations;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure
{
    internal class Page : PageTreeNode
    {
        public const string ContentsName = "Contents";
        public const string AnnotsName = "Annots";
        private const int RotationMultiplier = 90;
        private readonly ReferenceProperty<ContentStream> contentsProperty;
        private readonly DirectProperty<PdfArray> annotsProperty;

        public Page()
        {
            this.contentsProperty = this.RegisterReferenceProperty<ContentStream>(new PdfPropertyDescriptor(ContentsName));
            this.annotsProperty = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor(AnnotsName));
        }

        public override PageTreeNodeType PageTreeNodeType
        {
            get
            {
                return PageTreeNodeType.Page;
            }
        }

        public ContentStream Contents
        {
            get
            {
                return this.contentsProperty.GetValue();
            }
            set
            {
                this.contentsProperty.SetValue(value);
            }
        }

        public PdfArray Annots
        {
            get
            {
                return this.annotsProperty.GetValue();
            }
            set
            {
                this.annotsProperty.SetValue(value);
            }
        }

        public Rotation Rotation
        {
            get
            {
                return Page.GetRotation(this.Rotate);
            }
        }

        public static Rotation GetRotation(PdfInt rotate)
        {
            int value = rotate.Value;
            Rotation rotation = GetRotation(value);

            return rotation;
        }

        public static Rotation GetRotation(int value)
        {
            int rotationValue = ((value / RotationMultiplier) % 4) * RotationMultiplier;

            return (Rotation)rotationValue;
        }

        public void CopyPropertiesFrom(IPdfExportContext context, RadFixedPage fixedPage)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(fixedPage, "fixedPage");

            this.Resources = new PdfResource();
            this.MediaBox = fixedPage.ToBottomLeftCoordinateSystem(fixedPage.MediaBox);
            this.CropBox = fixedPage.ToBottomLeftCoordinateSystem(fixedPage.CropBox);
            this.Contents = new ContentStream();
            this.Contents.CopyPropertiesFrom(context, this, fixedPage);

            if (fixedPage.Annotations != null && fixedPage.Annotations.Count > 0)
            {
                this.Annots = new PdfArray();

                foreach (var annotation in fixedPage.Annotations)
                {
                    AnnotationObject annot = AnnotationFactory.CreatePdfAnnotation(context, annotation, fixedPage);
                    IndirectObject annotObject = context.CreateIndirectObject(annot);
                    this.Annots.Add(annotObject.Reference);
                }
            }

            this.Rotate = Page.GetRotate(fixedPage.Rotation);
        }

        public void CopyPropertiesTo(PostScriptReader reader, IRadFixedDocumentImportContext context, RadFixedPage fixedPage)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(fixedPage, "fixedPage");

            fixedPage.MediaBox = this.ConvertToTopLeftBox(this.MediaBox, reader, context);
            fixedPage.CropBox = this.ConvertToTopLeftBox(this.CropBox, reader, context);
            fixedPage.Rotation = this.Rotation;
        }

        internal void CopyPageContentTo(IRadFixedDocumentImportContext context, RadFixedPage fixedPage)
        {
            if (this.Contents != null)
            {
                this.Contents.CopyPropertiesTo(context, this, fixedPage);
            }
        }

        internal void CopyPageAnnotationsTo(IRadFixedDocumentImportContext context, RadFixedPage fixedPage)
        {
            if (this.Annots != null)
            {
                for (int i = 0; i < this.Annots.Count; i++)
                {
                    AnnotationObject pdfAnnotation;
                    this.Annots.TryGetElement(context.Reader, context, i, out pdfAnnotation);

                    if (!pdfAnnotation.IsSupported)
                    {
                        continue;
                    }

                    Annotation annotation = pdfAnnotation.ToAnnotation(context.Reader, context, fixedPage.Size.Height);

                    if (annotation != null)
                    {
                        fixedPage.Annotations.Add(annotation);
                    }
                }
            }
        }

        internal static PdfInt GetRotate(Rotation rotation)
        {
            return ((int)rotation).ToPdfInt();
        }

        internal Rect ConvertToTopLeftBox(PdfArray pdfArray, PostScriptReader reader, IPdfImportContext context)
        {
            double rootHeight = this.MediaBox.ToRect(reader, context).Height;
            Rect rect = pdfArray.ToRect(reader, context);
            Rect topLeftBox = rect.ToTopLeftCoordinateSystem(rootHeight);

            return topLeftBox;
        }
    }
}
