﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure
{
    internal class PageTreeNode : PdfObject, IResourceHolder, ITreeNode<PageTreeNode>
    {
        public const string PageName = "Page";
        public const string PageTreeNodeName = "Pages";
        public const string ResourcesName = "Resources";
        public const string MediaBoxName = "MediaBox";
        public const string CropBoxName = "CropBox";
        public const string RotateName = "Rotate";
        private readonly ReferenceProperty<PdfArray> children;
        private readonly ReferenceProperty<PageTreeNode> parent;
        private readonly DirectProperty<PdfInt> count;
        private readonly DirectProperty<PdfArray> mediaBox;
        private readonly DirectProperty<PdfArray> cropBox;
        private readonly DirectProperty<PdfResource> resources;
        private readonly DirectProperty<PdfInt> rotate;

        public PageTreeNode()
        {
            this.children = this.RegisterReferenceProperty<PdfArray>(new PdfPropertyDescriptor("Kids", false, PdfPropertyRestrictions.ContainsOnlyIndirectReferences));
            this.parent = this.RegisterReferenceProperty<PageTreeNode>(new PdfPropertyDescriptor("Parent", false, PdfPropertyRestrictions.MustBeIndirectReference));
            this.count = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("Count"));
            this.mediaBox = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor(MediaBoxName, true));
            this.cropBox = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor(CropBoxName));
            this.resources = this.RegisterDirectProperty<PdfResource>(new PdfPropertyDescriptor(ResourcesName));
            this.rotate = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor(RotateName));
        }

        public virtual PageTreeNodeType PageTreeNodeType
        {
            get
            {
                return PageTreeNodeType.PageTreeNode;
            }
        }

        public PdfArray Children
        {
            get
            {
                return this.children.GetValue();
            }
            set
            {
                this.children.SetValue(value);
            }
        }

        public PageTreeNode Parent
        {
            get
            {
                return this.parent.GetValue();
            }
            set
            {
                this.parent.SetValue(value);
            }
        }

        public PdfInt Count
        {
            get
            {
                return this.count.GetValue();
            }
            set
            {
                this.count.SetValue(value);
            }
        }

        public PdfArray MediaBox
        {
            get
            {
                return this.GetInheritableProperty((pageTreeNode) => pageTreeNode.mediaBox.GetValue());
            }
            set
            {
                this.mediaBox.SetValue(value);
            }
        }

        public PdfArray CropBox
        {
            get
            {
                PdfArray cropArray = this.GetInheritableProperty((pageTreeNode) => pageTreeNode.cropBox.GetValue());

                if (cropArray == null)
                {
                    cropArray = this.MediaBox;
                }

                return cropArray;
            }
            set
            {
                this.cropBox.SetValue(value);
            }
        }

        public PdfResource Resources
        {
            get
            {
                return this.GetInheritableProperty((pageTreeNode) => pageTreeNode.resources.GetValue());
            }
            set
            {
                this.resources.SetValue(value);
            }
        }

        public PdfInt Rotate
        {
            get
            {
                return this.GetInheritableProperty((pageTreeNode) => pageTreeNode.rotate.GetValue(), 0.ToPdfInt());
            }
            set
            {
                this.rotate.SetValue(value);
            }
        }

        PageTreeNode ITreeNode<PageTreeNode>.NodeValue
        {
            get
            {
                return this;
            }
        }

        ITreeNode<PageTreeNode> ITreeNode<PageTreeNode>.ParentNode
        {
            get
            {
                return this.Parent;
            }
        }

        public static PageTreeNode CreateInstance(PdfName typeName)
        {
            Guard.ThrowExceptionIfNull(typeName, "typeName");


            switch (typeName.Value)
            {
                case PageTreeNode.PageName:

                    return new Page();

                case PageTreeNode.PageTreeNodeName:

                    return new PageTreeNode();

                default:

                    throw new NotSupportedException("Type is not supported.");
            }
        }

        public IList<Page> EnumeratePages(PostScriptReader reader, IPdfImportContext context)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            List<Page> result = new List<Page>();
            for (int i = 0; i < this.Children.Count; i++)
            {
                PageTreeNode pageTreeNode;
                if (this.Children.TryGetElement(reader, context, i, out pageTreeNode))
                {
                    if (pageTreeNode.PageTreeNodeType == PageTreeNodeType.Page)
                    {
                        result.Add((Page)pageTreeNode);
                    }
                    else if (pageTreeNode.PageTreeNodeType == PageTreeNodeType.PageTreeNode)
                    {
                        IList<Page> pages = pageTreeNode.EnumeratePages(reader, context);
                        result.AddRange(pages);
                    }
                }
            }

            return result;
        }
    }
}
