﻿using System.Collections.Generic;
using System.Linq;
#if !NETSTANDARD
using System.Windows;
using System.Windows.Media;
#else
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Primitives;
#endif
using Telerik.Windows.Documents.Core.Fonts;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.CMaps;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Text;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Fonts
{
    internal abstract class FontObject : PdfObject
    {
        private readonly DirectProperty<PdfName> baseFont;
        private readonly ReferenceProperty<ToUnicodeCMap> toUnicode;

        public FontObject()
        {
            this.baseFont = this.RegisterDirectProperty<PdfName>(new PdfPropertyDescriptor("BaseFont", true));
            this.toUnicode = this.RegisterReferenceProperty<ToUnicodeCMap>(new PdfPropertyDescriptor("ToUnicode", true));
        }

        public PdfName BaseFont
        {
            get
            {
                return this.baseFont.GetValue();
            }
            set
            {
                this.baseFont.SetValue(value);
            }
        }

        public ToUnicodeCMap ToUnicode
        {
            get
            {
                return this.toUnicode.GetValue();
            }
            set
            {
                this.toUnicode.SetValue(value);
            }
        }

        public FontBase ToFont(PostScriptReader reader, IPdfImportContext context)
        {
            this.PrepareForImport(reader, context);
            FontBase font; 

            if (!this.TryCreateFontFromFontFile(out font) &&
                !this.TryCreateFontFromFontName(out font) &&
                !this.TryCreateFontFromFontFamily(out font))
            {
                font = this.CreateFontWithoutFontFile();
            }

            if (font != null)
            {
                this.CopyPropertiesToFont(font, reader, context);
            }

            return font;
        }

        public IEnumerable<CharInfo> GetCharacters(PdfString str)
        {
            Guard.ThrowExceptionIfNull(str, "str");

            return this.GetCharactersOverride(str);
        }

        public void CopyPropertiesFrom(IPdfExportContext context, FontBase font)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(font, "font");

            this.CopyPropertiesFromOverride(context, font);
        }

        protected abstract bool TryGetFontFamily(out FontFamily fontFamily);
        protected abstract bool TryCreateFontFromFontFile(out FontBase font);
        protected abstract FontBase CreateFontWithoutFontFile();
        protected abstract bool TryCreateFontFromFontProperties(FontProperties fontProperties, out FontBase font);

        protected virtual void CopyPropertiesFromOverride(IPdfExportContext context, FontBase font)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(font, "font");

            this.BaseFont = font.FontName.ToPdfName();
            this.ToUnicode = font.ToUnicode.ToPrimitive((toUnicode) => new ToUnicodeCMap(toUnicode), () => CreateToUnicodeCMapFromFont(context, font));
        }

        protected virtual IEnumerable<CharInfo> GetCharactersOverride(PdfString str)
        {
            Guard.ThrowExceptionIfNull(str, "str");

            foreach (byte b in str.Value)
            {
                CharCode code = new CharCode(b);

                yield return new CharInfo(GetToUnicode(code), code);
            }
        }

        protected virtual string GetToUnicode(CharCode code)
        {
            string unicode;
            if (this.ToUnicode != null && this.ToUnicode.Encoding.TryGetToUnicode(code, out unicode))
            {
                return unicode;
            }

            return ((char)code.Code).ToString();
        }

        protected virtual void PrepareForImport(PostScriptReader reader, IPdfImportContext context)
        {
        }

        protected virtual bool TryCreateFontFromFontName(out FontBase font)
        {
            string fontFamily;
            FontWeight fontWeight;
            FontStyle fontStyle;
            FontsHelper.GetFontFamily(this.BaseFont.Value, out fontFamily, out fontWeight, out fontStyle);
            FontProperties properties = new FontProperties(new FontFamily(fontFamily), fontStyle, fontWeight);

            return this.TryCreateFontFromFontProperties(properties, out font);
        }

        protected void CopyFontDescriptorPropertiesToFont(FontDescriptor descriptor, FontBase font)
        {
            if (descriptor != null)
            {
                descriptor.Ascent.CopyToProperty(font.Ascent, (ascent) => ascent.Value / FontBase.FontRatio);
                descriptor.Descent.CopyToProperty(font.Descent, (descent) => descent.Value / FontBase.FontRatio);
                descriptor.CapHeight.CopyToProperty(font.CapHeight, (capHeight) => capHeight.Value);
                descriptor.FontBBox.CopyToProperty(font.BoundingBox, (box) =>
                    {
                        double[] array = box.ToDoubleArray();
                        double x = array[0];
                        double y = array[1];
                        double width = array[2] - x;
                        double height = array[3] - y;

                        return new Rect(x, y, width, height);
                    });
                descriptor.ItalicAngle.CopyToProperty(font.ItalicAngle, (angle) => angle.Value);
                descriptor.StemV.CopyToProperty(font.StemV, (stemV) => stemV.Value);
                descriptor.FontFamily.CopyToProperty(font.FontFamily, (fontFamily) => fontFamily.ToString());
                descriptor.CharSet.CopyToProperty(font.CharSet, (charSet) => charSet.ToString());
                descriptor.FontStretch.CopyToProperty(font.FontStretch, (fontStretch) => fontStretch.Value);
                descriptor.FontWeight.CopyToProperty(font.FontWeight, (weight) => weight.Value);
                descriptor.Leading.CopyToProperty(font.Leading, (leading) => leading.Value);
                descriptor.XHeight.CopyToProperty(font.XHeight, (xHeight) => xHeight.Value);
                descriptor.StemH.CopyToProperty(font.StemH, (stemH) => stemH.Value);
                descriptor.AvgWidth.CopyToProperty(font.AvgWidth, (width) => width.Value);
                descriptor.MaxWidth.CopyToProperty(font.MaxWidth, (width) => width.Value);
                descriptor.MissingWidth.CopyToProperty(font.MissingWidth, (width) => width.Value);

                if(descriptor.Flags != null)
                {
                    this.CopyFlagsToFont(descriptor.Flags.Value, font);
                }           
            }
        }

        protected virtual void CopyPropertiesToFont(FontBase font, PostScriptReader reader, IPdfImportContext context)
        {
            this.ToUnicode.CopyToProperty(font.ToUnicode, (toUnicode) => toUnicode.Encoding);
        }

        private void CopyFlagsToFont(int flags, FontBase font)
        {
            FlagReader<FontFlag> reader = new FlagReader<FontFlag>(flags);

            font.IsFixedPitch = reader.IsSet(FontFlag.FixedPitch);
            font.IsSerif = reader.IsSet(FontFlag.Serif);
            font.IsSymbolic = reader.IsSet(FontFlag.Symbolic);
            font.IsScript = reader.IsSet(FontFlag.Script);
            font.IsNonSymbolic = reader.IsSet(FontFlag.Nonsymbolic);
            font.IsItalic = reader.IsSet(FontFlag.Italic);
            font.IsAllCap = reader.IsSet(FontFlag.AllCap);
            font.IsSmallCap = reader.IsSet(FontFlag.SmallCap);
            font.IsForcingBold = reader.IsSet(FontFlag.ForceBold);
        }

        private bool TryCreateFontFromFontFamily(out FontBase font)
        {
            FontFamily fontFamily;
            bool success = false;

            if (this.TryGetFontFamily(out fontFamily))
            {
                success = this.TryCreateFontFromFontProperties(new FontProperties(fontFamily), out font);
            }
            else
            {
                font = null;
            }

            return success;
        }

        private static ToUnicodeCMap CreateToUnicodeCMapFromFont(IPdfExportContext context, FontBase font)
        {
            ToUnicodeCMap toUnicodeCMap = null;
            IEnumerable<CharInfo> usedCharacters = context.GetUsedCharacters(font);

            SimpleFont simpleFont = font as SimpleFont;
            bool hasEncoding = simpleFont != null && simpleFont.Encoding.HasValue;

            // NOTE: This check is not by specification. However, it's added to fix issue 
            // with VaraiableContentWidget's appearance when rendered in Adobe Reader.
            bool shouldCreateToUnicodeMap = usedCharacters.Count() > 0 && !hasEncoding;

            if (shouldCreateToUnicodeMap)
            {
                toUnicodeCMap = new ToUnicodeCMap();
                toUnicodeCMap.CopyPropertiesFrom(context, usedCharacters);
            }

            return toUnicodeCMap;
        }
    }
}
