﻿using System;
using System.Linq;
#if !NETSTANDARD
using System.Windows.Media;
#else
using Telerik.Documents.Core.Fonts;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Fonts.Encoding;
using Telerik.Windows.Documents.Fixed.Model.Text;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Fonts
{
    internal abstract class SimpleFontObject : FontObject
    {
        private readonly DirectProperty<PdfInt> firstChar;
        private readonly DirectProperty<PdfInt> lastChar;
        private readonly DirectProperty<PdfArray> widths;
        private readonly ReferenceProperty<FontDescriptor> fontDescriptor;
        private readonly ReferenceProperty<SimpleFontEncoding> encoding;

        public SimpleFontObject()
        {
            this.fontDescriptor = this.RegisterReferenceProperty<FontDescriptor>(new PdfPropertyDescriptor("FontDescriptor", true, PdfPropertyRestrictions.MustBeIndirectReference));
            this.encoding = this.RegisterReferenceProperty<SimpleFontEncoding>(new PdfPropertyDescriptor("Encoding", false, PdfPropertyRestrictions.MustBeIndirectReference));
            this.firstChar = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("FirstChar", true));
            this.lastChar = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("LastChar", true));
            this.widths = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor("Widths", true));
        }

        public PdfInt FirstChar
        {
            get
            {
                return this.firstChar.GetValue();
            }
            set
            {
                this.firstChar.SetValue(value);
            }
        }

        public PdfInt LastChar
        {
            get
            {
                return this.lastChar.GetValue();
            }
            set
            {
                this.lastChar.SetValue(value);
            }
        }

        public PdfArray Widths
        {
            get
            {
                return this.widths.GetValue();
            }
            set
            {
                this.widths.SetValue(value);
            }
        }

        public SimpleFontEncoding Encoding
        {
            get
            {
                return this.encoding.GetValue();
            }
            set
            {
                this.encoding.SetValue(value);
            }
        }

        public FontDescriptor FontDescriptor
        {
            get
            {
                return this.fontDescriptor.GetValue();
            }
            set
            {
                this.fontDescriptor.SetValue(value);
            }
        }

        protected abstract bool TryCreateSimpleFontFromFontFile(out SimpleFont font);
        protected abstract SimpleFont CreateSimpleFontWithoutFontFile();

        protected override bool TryGetFontFamily(out FontFamily fontFamily)
        {
            fontFamily = null;

            if (this.FontDescriptor != null && this.FontDescriptor.FontFamily != null)
            {
                fontFamily = new FontFamily(this.FontDescriptor.FontFamily.ToString());
            }

            return fontFamily != null;
        }

        protected override bool TryCreateFontFromFontFile(out FontBase font)
        {
            SimpleFont simpleFont;
            if (this.TryCreateSimpleFontFromFontFile(out simpleFont))
            {
                font = simpleFont;
                return true;
            }

            font = null;
            return false;
        }

        protected override FontBase CreateFontWithoutFontFile()
        {
            SimpleFont font = this.CreateSimpleFontWithoutFontFile();

            return font;
        }

        protected override void CopyPropertiesFromOverride(IPdfExportContext context, FontBase font)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(font, "font");

            base.CopyPropertiesFromOverride(context, font);

            var simpleFont = (SimpleFont)font;

            if (font.Type != FontType.Standard || font.HasFontDescriptorProperties)
            {
                this.FontDescriptor = new FontDescriptor();
                this.FontDescriptor.CopyPropertiesFrom(context, font);
            }

            this.Encoding = simpleFont.Encoding.ToPrimitive((encoding) => new SimpleFontEncoding()
            {
                BaseEncoding = encoding.PredefinedEncoding.ToPrimitive((name) => new PdfName(name)),
                Differences = encoding.Differences.ToPrimitive((differencesRanges) => CreateDifferencesArray(differencesRanges))
            });

            if (!this.TryCopyWidthsFromFont(simpleFont))
            {
                this.CalculateUsedGlyphsWidths(context, font);
            }
        }

        protected override void CopyPropertiesToFont(FontBase font, PostScriptReader reader, IPdfImportContext context)
        {
            SimpleFont simpleFont = font as SimpleFont;

            if (simpleFont != null)
            {
                this.FirstChar.CopyToProperty(simpleFont.FirstChar, (firstChar) => firstChar.Value);
                this.LastChar.CopyToProperty(simpleFont.LastChar, (lastChar) => lastChar.Value);
                this.Widths.CopyToProperty(simpleFont.Widths, (widths) => widths.ToIntArray());
                this.Encoding.CopyToProperty(simpleFont.Encoding, (encoding) =>
                    {
                        SimpleEncoding simpleEncoding = new SimpleEncoding();
                        encoding.BaseEncoding.CopyToProperty(simpleEncoding.PredefinedEncoding, (name) => name.Value);
                        encoding.Differences.CopyToProperty(simpleEncoding.Differences, (array) => CreateDifferencesRanges(array));

                        return simpleEncoding;
                    });
            }

            base.CopyFontDescriptorPropertiesToFont(this.FontDescriptor, font);
            base.CopyPropertiesToFont(font, reader, context);
        }

        private static PdfArray CreateDifferencesArray(DifferencesRanges differencesRanges)
        {
            PdfArray differences = new PdfArray();

            foreach (DifferencesRange differencesRange in differencesRanges)
            {
                differences.Add(new PdfInt(differencesRange.StartCharCode));

                foreach (string characterName in differencesRange.Differences)
                {
                    differences.Add(new PdfName(characterName));
                }
            }

            return differences;
        }

        private static DifferencesRanges CreateDifferencesRanges(PdfArray array)
        {
            DifferencesRanges ranges = new DifferencesRanges();
            DifferencesRange currentRange = null;

            foreach (IPdfPrimitive primitive in array)
            {
                switch (primitive.Type)
                {
                    case PdfElementType.PdfInt:
                        PdfInt charCode = (PdfInt)primitive;
                        currentRange = ranges.AddDifferencesRange((byte)charCode.Value);
                        break;
                    case PdfElementType.PdfName:
                        PdfName name = (PdfName)primitive;
                        currentRange.AddNextDifference(name.Value);
                        break;
                    default:
                        throw new InvalidOperationException(string.Format("Unexpected element type: {0}", primitive.Type));
                }
            }

            return ranges;
        }

        private bool TryCopyWidthsFromFont(SimpleFont font)
        {
            bool shouldCopyImportedProperties = font.Widths.HasValue && font.FirstChar.HasValue && font.LastChar.HasValue;

            if (shouldCopyImportedProperties)
            {
                this.Widths = font.Widths.ToPrimitive((widths) => widths.ToPdfArray());
                this.FirstChar = font.FirstChar.Value.ToPdfInt();
                this.LastChar = font.LastChar.Value.ToPdfInt();
            }

            return shouldCopyImportedProperties;
        }

        private void CalculateUsedGlyphsWidths(IPdfExportContext context, FontBase font)
        {
            if (font.Type != FontType.Standard)
            {
                CharInfo[] usedCharacters = context.GetUsedCharacters(font).OrderBy(g => g.CharCode.Code).ToArray();

                if (usedCharacters.Length > 0)
                {
                    this.FirstChar = new PdfInt(usedCharacters[0].CharCode.Code);
                    this.LastChar = new PdfInt(usedCharacters[usedCharacters.Length - 1].CharCode.Code);
                    PdfArray widths = new PdfArray();
                    for (int charCode = this.FirstChar.Value; charCode <= this.LastChar.Value; charCode++)
                    {
                        widths.Add(new PdfInt((int)font.GetWidth(charCode)));
                    }

                    this.Widths = widths;
                }
            }
        }
    }
}
