﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Annotations;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
#if WPF
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DigitalSignature;
#endif

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Forms
{
    internal class FieldPropertiesExporter : VariableTextObjectPropertiesExporter<FormFieldNode>
    {
        public FieldPropertiesExporter(FormFieldNode node, IPdfExportContext exportContext) 
            : base(node, exportContext)
        {
        }

        private FlagWriter<FieldFlag> FlagWriter { get; set; }

        public void ExportFieldProperties(FormField field)
        {
            this.FlagWriter = new FlagWriter<FieldFlag>();

            switch (field.FieldType)
            {
                case FormFieldType.CheckBox:
                case FormFieldType.RadioButton:
                case FormFieldType.PushButton:
                    this.CopyFromButton(field);
                    break;
                case FormFieldType.ListBox:
                case FormFieldType.ComboBox:
                    this.CopyFromChoice((ChoiceField)field);
                    break;
                case FormFieldType.CombTextBox:
                case FormFieldType.TextBox:
                    this.CopyFromText((TextField)field);
                    break;
                case FormFieldType.Signature:
                    this.CopyFromSignature((SignatureField)field);
                    break;
                default:
                    throw new NotSupportedException(string.Format("Not supported field type: {0}", field.FieldType));
            }

            this.CopyCommonFieldProperties(field);
        }

        private void CopyCommonFieldProperties(FormField field)
        {
            this.FlagWriter.SetFlagOnCondition(FieldFlag.Required, field.IsRequired);
            this.FlagWriter.SetFlagOnCondition(FieldFlag.NoExport, field.ShouldBeSkipped);
            this.FlagWriter.SetFlagOnCondition(FieldFlag.ReadOnly, field.IsReadOnly);

            this.Node.PartialName = field.Name.ToPdfString();
            this.Node.UserInterfaceName = field.UserInterfaceName.ToPdfString();
            this.Node.MappingName = field.MappingName.ToPdfString();
            this.Node.FieldFlags = new PdfInt(this.FlagWriter.ResultFlags);
            this.ExportVariableProperties(field.TextProperties);

            this.CopyWidgets(field.Widgets);
        }

        private void CopyWidgets(IEnumerable<Widget> widgets)
        {
            foreach (Widget widgetModel in widgets)
            {
                WidgetObject widget;
                if (this.Context.TryGetWidget(widgetModel, out widget))
                {
                    if (this.Node.Kids == null)
                    {
                        this.Node.Kids = new PdfArray();
                    }

                    IndirectObject widgetObject = this.Context.CreateIndirectObject(widget);
                    this.Node.Kids.Add(widgetObject.Reference);
                }

                // Release all appearances after the widget export.
                widgetModel.Appearances = null;
            }
        }

        private void CopyFromButton(FormField buttonField)
        {
            this.Node.FieldType = new PdfName(FormFieldTypes.Button);

            switch (buttonField.FieldType)
            {
                case FormFieldType.CheckBox:
                    CheckBoxField checkBox = (CheckBoxField)buttonField;
                    this.CopyFromCheckBox(checkBox);
                    break;
                case FormFieldType.RadioButton:
                    RadioButtonField radio = (RadioButtonField)buttonField;
                    this.FlagWriter.SetFlag(FieldFlag.Radio);
                    this.FlagWriter.SetFlagOnCondition(FieldFlag.NoToggleToOff, !radio.AllowToggleOff);
                    this.FlagWriter.SetFlagOnCondition(FieldFlag.RadiosInUnison, radio.ShouldUpdateRadiosInUnison);
                    this.CopyFromRadioButton(radio);
                    break;
                case FormFieldType.PushButton:
                    this.FlagWriter.SetFlag(FieldFlag.PushButton);
                    break;
            }
        }

        private void CopyFromCheckBox(CheckBoxField checkBox)
        {
            PdfName fieldValue = checkBox.IsChecked ? new PdfName(CheckBoxField.DefaultExportValue) : new PdfName(TwoStatesButtonWidget.OffState);
            this.Node.FieldValue = new PrimitiveWrapper(fieldValue);

            PdfName defaultValue = checkBox.IsCheckedByDefault ? new PdfName(CheckBoxField.DefaultExportValue) : new PdfName(TwoStatesButtonWidget.OffState);
            this.Node.DefaultFieldValue = new PrimitiveWrapper(defaultValue);

            if(checkBox.ExportValue != CheckBoxField.DefaultExportValue)
            {
                PdfArray array = new PdfArray();
                
                foreach(Widget widget in this.GetPageExportedWidgets(checkBox))
                {
                    array.Add(checkBox.ExportValue.ToPdfString());
                }

                if(array.Count > 0)
                {
                    this.Node.Options = array;
                }
            }
        }

        private void CopyFromRadioButton(RadioButtonField radio)
        {
            PdfArray optionsArray = new PdfArray();
            bool shouldUpdateRadiosInUnison = radio.ShouldUpdateRadiosInUnison;

            foreach (RadioButtonWidget widget in this.GetPageExportedWidgets(radio))
            {
                optionsArray.Add(widget.Option.Value.ToPdfString());
                
                if (radio.Value != null && this.Node.FieldValue == null)
                {
                    string valueState;
                    RadioOption option = radio.Value;

                    if (this.TryGetOptionOnStateFromWidget(widget, option, shouldUpdateRadiosInUnison, out valueState))
                    {
                        this.Node.FieldValue = new PrimitiveWrapper(new PdfName(valueState));
                    }
                }

                if (radio.DefaultValue != null && this.Node.DefaultFieldValue == null)
                {
                    string valueState;
                    RadioOption option = radio.DefaultValue;

                    if (this.TryGetOptionOnStateFromWidget(widget, option, shouldUpdateRadiosInUnison, out valueState))
                    {
                        this.Node.DefaultFieldValue = new PrimitiveWrapper(new PdfName(valueState));
                    }
                }
            }

            if(optionsArray.Count > 0)
            {
                this.Node.Options = optionsArray;
            }
        }

        private bool TryGetOptionOnStateFromWidget(RadioButtonWidget widget, RadioOption option, bool shouldUpdateRadiosInUnison, out string stateName)
        {
            bool isWidgetWithSameValue = widget.Option.Equals(option, shouldUpdateRadiosInUnison);

            if (isWidgetWithSameValue)
            {
                MultiStateAppearances appearances = (MultiStateAppearances)widget.Appearances;
                stateName = appearances.Where(x => x.Key != TwoStatesButtonWidget.OffState).First().Key;

                return true;
            }

            stateName = null;

            return false;
        }

        private IEnumerable<Widget> GetPageExportedWidgets(FormField formField)
        {
            foreach (Widget widgetModel in formField.Widgets)
            {
                WidgetObject widget;
                if (this.Context.TryGetWidget(widgetModel, out widget))
                {
                    yield return widgetModel;
                }
            }
        }

        private void CopyFromChoice(ChoiceField choiceField)
        {
            this.Node.FieldType = new PdfName(FormFieldTypes.Choice);
            this.FlagWriter.SetFlagOnCondition(FieldFlag.CommitOnSelChange, choiceField.ShouldCommitOnSelectionChange);
            this.CopyFromChoiceOptionsCollection(choiceField.Options);

            switch (choiceField.FieldType)
            {
                case FormFieldType.ComboBox:
                    ComboBoxField comboBox = (ComboBoxField)choiceField;
                    this.FlagWriter.SetFlag(FieldFlag.Combo);
                    this.FlagWriter.SetFlagOnCondition(FieldFlag.Edit, comboBox.HasEditableTextBox);
                    this.FlagWriter.SetFlagOnCondition(FieldFlag.DoNotSpellCheck, !comboBox.ShouldSpellCheck);
                    this.CopyComboBoxValues(comboBox);
                    break;
                case FormFieldType.ListBox:
                    ListBoxField listBox = (ListBoxField)choiceField;
                    this.FlagWriter.SetFlagOnCondition(FieldFlag.MultiSelect, listBox.AllowMultiSelection);
                    this.Node.TopIndex = new PdfInt(listBox.TopIndex);
                    this.CopyListBoxValues(listBox);
                    break;
            }
        }

        private void CopyFromChoiceOptionsCollection(ChoiceOptionCollection options)
        {
            this.Node.Options = new PdfArray();

            foreach (ChoiceOption option in options)
            {
                IPdfPrimitive optionPrimitive;
                if (option.UserInterfaceValue == null)
                {
                    optionPrimitive = option.Value.ToPdfString();
                }
                else
                {
                    optionPrimitive = new PdfArray(option.Value.ToPdfString(), option.UserInterfaceValue.ToPdfString());
                }

                this.Node.Options.Add(optionPrimitive);
            }
        }

        private void CopyComboBoxValues(ComboBoxField comboBox)
        {
            if (comboBox.Value != null)
            {
                int optionIndex;
                bool hasFoundOptionIndex = comboBox.Options.TryGetOptionIndex(comboBox.Value, out optionIndex);

                if (comboBox.HasEditableTextBox || hasFoundOptionIndex)
                {
                    this.Node.FieldValue = new PrimitiveWrapper(comboBox.Value.Value.ToPdfString());
                }

                if (hasFoundOptionIndex)
                {
                    this.Node.SortedIndices = new PdfArray(new PdfInt(optionIndex));
                }
            }

            int defaultValueIndex;
            if (comboBox.DefaultValue != null && comboBox.Options.TryGetOptionIndex(comboBox.DefaultValue, out defaultValueIndex))
            {
                this.Node.DefaultFieldValue = new PrimitiveWrapper(comboBox.DefaultValue.Value.ToPdfString());
            }
        }

        private void CopyListBoxValues(ListBoxField listBox)
        {
            int[] selectedIndices = listBox.GetSelectedIndicesSorted();

            if (selectedIndices.Length > 0)
            {
                this.Node.FieldValue = new PrimitiveWrapper(this.GetSelectedValueArray(listBox.Options, selectedIndices));

                PdfArray sortedIndices = new PdfArray();

                foreach(int index in selectedIndices)
                {
                    sortedIndices.Add(new PdfInt(index));
                }

                this.Node.SortedIndices = sortedIndices;
            }

            int[] defaultSelectedIndices = listBox.GetDefaultSelectedIndicesSorted();

            if (defaultSelectedIndices.Length > 0)
            {
                this.Node.DefaultFieldValue = new PrimitiveWrapper(this.GetSelectedValueArray(listBox.Options, defaultSelectedIndices));
            }
        }

        private PdfArray GetSelectedValueArray(ChoiceOptionCollection options, int[] sortedSelectedIndices)
        {
            PdfArray array = new PdfArray();

            foreach(int index in sortedSelectedIndices)
            {
                ChoiceOption option = options[index];
                array.Add(option.Value.ToPdfString());
            }

            return array;
        }

        private void CopyFromText(TextField textField)
        {
            this.Node.FieldType = new PdfName(FormFieldTypes.Text);

            switch(textField.FieldType)
            {
                case FormFieldType.TextBox:
                    TextBoxField textBox = (TextBoxField)textField;
                    this.FlagWriter.SetFlagOnCondition(FieldFlag.Multiline, textBox.IsMultiline);
                    this.FlagWriter.SetFlagOnCondition(FieldFlag.Password, textBox.IsPassword);
                    this.FlagWriter.SetFlagOnCondition(FieldFlag.FileSelect, textBox.IsFileSelect);
                    this.FlagWriter.SetFlagOnCondition(FieldFlag.DoNotSpellCheck, !textBox.ShouldSpellCheck);
                    this.FlagWriter.SetFlagOnCondition(FieldFlag.DoNotScroll, !textBox.AllowScroll);

                    if(textBox.MaxLengthOfInputCharacters.HasValue)
                    {
                        this.Node.MaxLengthOfInputCharacters = new PdfInt(textBox.MaxLengthOfInputCharacters.Value);
                    }
                    break;
                case FormFieldType.CombTextBox:
                    CombTextBoxField combs = (CombTextBoxField)textField;
                    this.FlagWriter.SetFlag(FieldFlag.Comb);
                    this.Node.MaxLengthOfInputCharacters = new PdfInt(combs.MaxLengthOfInputCharacters);
                    break;
            }

            if(textField.Value != null)
            {
                this.Node.FieldValue = new PrimitiveWrapper(textField.Value.ToPdfString());
            }

            if(textField.DefaultValue != null)
            {
                this.Node.DefaultFieldValue = new PrimitiveWrapper(textField.DefaultValue.ToPdfString());
            }
        }

        private void CopyFromSignature(SignatureField signatureField)
        {
            this.Node.FieldType = new PdfName(FormFieldTypes.Signature);
#if WPF
            if (signatureField.Signature != null)
            {
                SignatureObject signatureObject = new SignatureObject();
                signatureObject.CopyPropertiesFrom(signatureField.Signature);
                IndirectObject indirectObject = this.Context.CreateIndirectObject(signatureObject);
                this.Node.FieldValue = new PrimitiveWrapper(indirectObject.Reference);
            }
#endif
        }
    }
}
