﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Annotations;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
#if WPF
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DigitalSignature;
using Telerik.Windows.Documents.Fixed.Model.DigitalSignatures;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Objects;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Forms
{
    internal class FieldPropertiesImporter : VariableTextObjectPropertiesImporter<FormFieldNode>
    {
        public FieldPropertiesImporter(FormFieldNode node, PostScriptReader reader, IRadFixedDocumentImportContext context)
            : base(node, reader, context)
        {
        }

        public void ImportFieldProperties(FormField field)
        {
            field.TextProperties = this.ReadTextProperties();

            switch (field.FieldType)
            {
                case FormFieldType.CheckBox:
                case FormFieldType.PushButton:
                case FormFieldType.RadioButton:
                    this.InitializeButtonFieldProperties(field);
                    break;
                case FormFieldType.ListBox:
                case FormFieldType.ComboBox:
                    this.InitializeChoiceFieldProperties((ChoiceField)field);
                    break;
                case FormFieldType.CombTextBox:
                case FormFieldType.TextBox:
                    this.InitializeTextFieldProperties((TextField)field);
                    break;
                case FormFieldType.Signature:
                    this.InitializeSignatureFieldProperties((SignatureField)field);
                    break;
                default:
                    throw new NotSupportedException(string.Format("Not supported field type: {0}", field.FieldType));
            }
        }

        private void InitializeButtonFieldProperties(FormField field)
        {
            switch (field.FieldType)
            {
                case FormFieldType.CheckBox:
                    CheckBoxField checkBoxField = (CheckBoxField)field;
                    this.InitializeCheckBoxFieldProperties(checkBoxField);
                    break;
                case FormFieldType.RadioButton:
                    RadioButtonField radioButtonField = (RadioButtonField)field;
                    this.InitializeRadioButtonFieldProperties(radioButtonField);
                    break;
            }
        }

        private void InitializeRadioButtonFieldProperties(RadioButtonField radioButtonField)
        {
            string fieldValue = this.Node.FieldValue.ToText();
            string defaultValue = this.Node.DefaultFieldValue.ToText();

            if (this.Node.Options != null)
            {
                this.ImportOptionsArray(radioButtonField);

                int index = 0;
                foreach (WidgetObject widgetObject in this.Context.GetChildWidgets(this.Node))
                {
                    RadioButtonWidget widget = (RadioButtonWidget)widgetObject.ToWidget(this.Reader, this.Context);
                    widget.Option = radioButtonField.Options[index];

                    string stateName;
                    if (!string.IsNullOrEmpty(fieldValue) && radioButtonField.Value == null && widgetObject.TryGetOnStateName(out stateName) && stateName.Equals(fieldValue))
                    {
                        radioButtonField.Value = widget.Option;
                    }

                    if (!string.IsNullOrEmpty(defaultValue) && radioButtonField.DefaultValue == null && widgetObject.TryGetOnStateName(out stateName) && stateName.Equals(defaultValue))
                    {
                        radioButtonField.DefaultValue = widget.Option;
                    }

                    index++;
                }
            }
            else
            {
                foreach(WidgetObject widgetObject in this.Context.GetChildWidgets(this.Node))
                {
                    string stateName;
                    if (widgetObject.TryGetOnStateName(out stateName))
                    {
                        RadioOption option = new RadioOption(stateName);
                        radioButtonField.Options.Add(option);
                        RadioButtonWidget widget = (RadioButtonWidget)widgetObject.ToWidget(this.Reader, this.Context);
                        widget.Option = option;

                        if (!string.IsNullOrEmpty(fieldValue) && radioButtonField.Value == null && stateName.Equals(fieldValue))
                        {
                            radioButtonField.Value = option;
                        }

                        if (!string.IsNullOrEmpty(defaultValue) && radioButtonField.DefaultValue == null && stateName.Equals(defaultValue))
                        {
                            radioButtonField.Value = option;
                        }
                    }
                }
            }
        }

        private void ImportOptionsArray(RadioButtonField radioButtonField)
        {
            foreach (IPdfPrimitive primitive in this.Node.Options)
            {
                PdfString value;
                if (primitive.Type == PdfElementType.String)
                {
                    value = (PdfString)primitive;
                }
                else
                {
                    PdfArray array = (PdfArray)primitive;
                    value = (PdfString)array[0];
                }

                radioButtonField.Options.Add(new RadioOption(value.ToString()));
            }
        }

        private void InitializeCheckBoxFieldProperties(CheckBoxField checkBoxField)
        {
            bool hasFoundExportValue = false;

            if (this.Node.Options != null)
            {
                checkBoxField.ExportValue = this.Node.Options[0].ToString();
                hasFoundExportValue = true;
            }

            string value = this.Node.FieldValue.ToText();

            if (value != null && value != TwoStatesButtonWidget.OffState)
            {
                checkBoxField.IsChecked = true;

                if(!hasFoundExportValue)
                {
                    checkBoxField.ExportValue = value;
                }
            }

            string defaultValue = this.Node.DefaultFieldValue.ToText();
            checkBoxField.IsCheckedByDefault = defaultValue != null && defaultValue != TwoStatesButtonWidget.OffState;

            if(!hasFoundExportValue)
            {
                foreach(WidgetObject widgetObject in this.Context.GetChildWidgets(this.Node))
                {
                    string stateName;
                    if(widgetObject.TryGetOnStateName(out stateName))
                    {
                        checkBoxField.ExportValue = stateName;
                        break;
                    }
                }
            }
        }

        private void InitializeChoiceFieldProperties(ChoiceField field)
        {
            InitializeChoiceOptionCollection(field.Options);            

            switch(field.FieldType)
            {
                case FormFieldType.ComboBox:
                    ComboBoxField comboBoxField = (ComboBoxField)field;
                    InitializeComboBoxFieldProperties(comboBoxField);
                    break;
                case FormFieldType.ListBox:
                    ListBoxField listBoxField = (ListBoxField)field;
                    InitializeListBoxFieldProperties(listBoxField);
                    break;
            }
        }

        private void InitializeComboBoxFieldProperties(ComboBoxField comboBoxField)
        {
            string value = this.Node.FieldValue.ToText();
            ChoiceOption indexValue = null;

            if (this.Node.SortedIndices != null)
            {
                int i = ((PdfInt)this.Node.SortedIndices[0]).Value;
                indexValue = comboBoxField.Options[i];
            }

            if (value != null)
            {
                int index;
                if(indexValue != null && indexValue.Value.Equals(value))
                {
                    comboBoxField.Value = indexValue;
                }
                else if (comboBoxField.Options.TryGetValueIndex(value, out index))
                {
                    comboBoxField.Value = comboBoxField.Options[index];
                }
                else
                {
                    comboBoxField.Value = new ChoiceOption(value);
                }                
            }
            else if (indexValue != null)
            {
                comboBoxField.Value = indexValue;
            }

            int defaultValueIndex;
            string defaultValue = this.Node.DefaultFieldValue.ToText();

            if(defaultValue != null && comboBoxField.Options.TryGetValueIndex(defaultValue, out defaultValueIndex))
            {
                comboBoxField.DefaultValue = comboBoxField.Options[defaultValueIndex];
            }
        }

        private void InitializeListBoxFieldProperties(ListBoxField listBoxField)
        {
            if(this.Node.SortedIndices != null)
            {
                listBoxField.Value = new ChoiceOption[this.Node.SortedIndices.Count];

                for (int i = 0; i < listBoxField.Value.Length; i++)
                {
                    int index = ((PdfInt)this.Node.SortedIndices[i]).Value;
                    listBoxField.Value[i] = listBoxField.Options[index];
                }
            }
            else
            {
                listBoxField.Value = this.InitializeListBoxValueFromValuesPrimitive(listBoxField.Options, this.Node.FieldValue);
            }

            if (this.Node.DefaultFieldValue != null)
            {
                listBoxField.DefaultValue = this.InitializeListBoxValueFromValuesPrimitive(listBoxField.Options, this.Node.DefaultFieldValue);
            }

            if (this.Node.TopIndex != null)
            {
                listBoxField.TopIndex = this.Node.TopIndex.Value;
            }
        }

        private ChoiceOption[] InitializeListBoxValueFromValuesPrimitive(ChoiceOptionCollection options, PrimitiveWrapper valuesPrimitive)
        {
            ChoiceOption[] choices = null;

            if (valuesPrimitive != null)
            {
                int index;
                string value;
                HashSet<int> indices = new HashSet<int>();

                switch (valuesPrimitive.Type)
                {
                    case PdfElementType.Array:
                        PdfArray array = (PdfArray)valuesPrimitive.Primitive;

                        foreach(IPdfPrimitive element in array)
                        {
                            value = element.ToString();
                            if (options.TryGetValueIndex(value, out index))
                            {
                                indices.Add(index);
                            }
                        }
                        break;
                    default:
                        value = valuesPrimitive.Primitive.ToString();
                        if (options.TryGetValueIndex(value, out index))
                        {
                            indices.Add(index);
                        }
                        break;
                }

                if(indices.Count > 0)
                {
                    int i = 0;
                    choices = new ChoiceOption[indices.Count];

                    foreach(int optionIndex in indices.OrderBy(x => x))
                    {
                        choices[i] = options[optionIndex];
                        i++;
                    }
                }
            }

            return choices;
        }

        private void InitializeChoiceOptionCollection(ChoiceOptionCollection optionCollection)
        {
            if (this.Node.Options != null)
            {
                foreach (IPdfPrimitive primitive in this.Node.Options)
                {
                    ChoiceOption option;

                    switch(primitive.Type)
                    {
                        case PdfElementType.Array:
                            PdfArray complexChoice = (PdfArray)primitive;
                            option = new ChoiceOption(complexChoice[0].ToString());

                            if (complexChoice.Count > 1)
                            {
                                option.UserInterfaceValue = complexChoice[1].ToString();
                            }
                            break;
                        default:
                            option = new ChoiceOption(primitive.ToString());
                            break;
                    }

                    optionCollection.Add(option);
                }
            }
        }

        private void InitializeTextFieldProperties(TextField field)
        {
            field.Value = this.Node.FieldValue.ToText();
            field.DefaultValue = this.Node.DefaultFieldValue.ToText();
        }

        private void InitializeSignatureFieldProperties(SignatureField signatureField)
        {
#if WPF
            if (this.Node.FieldValue != null)
            {
                signatureField.Signature = new Signature();

                SignatureObject signatureObject = new SignatureObject();
                signatureObject.Load(Reader, Context, this.Node.FieldValue.Primitive);
                signatureObject.CopyPropertiesTo(signatureField.Signature, Reader, Context);
            }
#endif
        }        
    }
}
