﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Annotations;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Forms
{
    internal class FormFieldNode : PdfObject, ITreeNode<FormFieldNode>, IVariableTextPropertiesObject
    {
        public const string TName = "T";
        public const string FieldValueName = "V";
        public const string KidsName = "Kids";
        public const string ParentName = "Parent";
        public const string OptionsName = "Opt";
        public const string SortedIndicesName = "I";

        private readonly DirectProperty<PdfName> fieldType;
        private readonly ReferenceProperty<FormFieldNode> parent;
        private readonly ReferenceProperty<PdfArray> kids;
        private readonly DirectProperty<PdfString> partialName;
        private readonly DirectProperty<PdfString> userInterfaceName;
        private readonly DirectProperty<PdfString> mappingName;
        private readonly DirectProperty<PdfInt> fieldFlags;
        private readonly DirectProperty<PdfInt> maxLengthOfInputCharacters;
        private readonly ReferenceProperty<PrimitiveWrapper> fieldValue;
        private readonly ReferenceProperty<PrimitiveWrapper> defaultFieldValue;
        private readonly ReferenceProperty<PdfDictionary> additionalActions;
        private readonly ReferenceProperty<PdfString> defaultAppearance;
        private readonly DirectProperty<PdfInt> quadding;
        private readonly ReferenceProperty<PdfArray> options;
        private readonly DirectProperty<PdfInt> topIndex;
        private readonly ReferenceProperty<PdfArray> sortedIndices;
        private readonly ReferenceProperty<SignatureFieldLock> signatureFieldLock;
        private readonly ReferenceProperty<SignatureFieldSeed> seedValue;

        public FormFieldNode()
        {
            this.fieldType = this.RegisterDirectProperty<PdfName>(new PdfPropertyDescriptor("FT", false, PdfPropertyRestrictions.MustBeDirectObject));
            this.parent = this.RegisterReferenceProperty<FormFieldNode>(new PdfPropertyDescriptor(ParentName, false, PdfPropertyRestrictions.MustBeIndirectReference));
            this.kids = this.RegisterReferenceProperty<PdfArray>(new PdfPropertyDescriptor(KidsName));
            this.partialName = this.RegisterDirectProperty<PdfString>(new PdfPropertyDescriptor(TName));
            this.userInterfaceName = this.RegisterDirectProperty<PdfString>(new PdfPropertyDescriptor("TU"));
            this.mappingName = this.RegisterDirectProperty<PdfString>(new PdfPropertyDescriptor("TM"));
            this.fieldFlags = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("Ff"));
            this.maxLengthOfInputCharacters = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("MaxLen"));
            this.fieldValue = this.RegisterReferenceProperty<PrimitiveWrapper>(new PdfPropertyDescriptor(FieldValueName, false));
            this.defaultFieldValue = this.RegisterReferenceProperty<PrimitiveWrapper>(new PdfPropertyDescriptor("DV"));
            this.additionalActions = this.RegisterReferenceProperty<PdfDictionary>(new PdfPropertyDescriptor("AA"));
            this.defaultAppearance = this.RegisterReferenceProperty<PdfString>(new PdfPropertyDescriptor("DA"));
            this.quadding = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("Q"));
            this.options = this.RegisterReferenceProperty<PdfArray>(new PdfPropertyDescriptor(OptionsName));
            this.topIndex = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("TI"), new PdfInt(0));
            this.sortedIndices = this.RegisterReferenceProperty<PdfArray>(new PdfPropertyDescriptor(SortedIndicesName));
            this.signatureFieldLock = this.RegisterReferenceProperty<SignatureFieldLock>(new PdfPropertyDescriptor("Lock", false, PdfPropertyRestrictions.MustBeIndirectReference));
            this.seedValue = this.RegisterReferenceProperty<SignatureFieldSeed>(new PdfPropertyDescriptor("SV", false, PdfPropertyRestrictions.MustBeIndirectReference));
        }

        public PdfName FieldType
        {
            get
            {
                return this.GetInheritableProperty((formFieldNode) => formFieldNode.fieldType.GetValue());
            }
            set
            {
                this.fieldType.SetValue(value);
            }
        }

        public FormFieldNode Parent
        {
            get
            {
                return this.parent.GetValue();
            }
            set
            {
                this.parent.SetValue(value);
            }
        }

        public PdfArray Kids
        {
            get
            {
                return this.kids.GetValue();
            }
            set
            {
                this.kids.SetValue(value);
            }
        }

        public PdfString PartialName
        {
            get
            {
                return this.partialName.GetValue();
            }
            set
            {
                this.partialName.SetValue(value);
            }
        }

        public PdfString UserInterfaceName
        {
            get
            {
                return this.userInterfaceName.GetValue();
            }
            set
            {
                this.userInterfaceName.SetValue(value);
            }
        }

        public PdfString MappingName
        {
            get
            {
                return this.mappingName.GetValue();
            }
            set
            {
                this.mappingName.SetValue(value);
            }
        }

        public PdfInt FieldFlags
        {
            get
            {
                return this.GetInheritableProperty((formFieldNode) => formFieldNode.fieldFlags.GetValue(), new PdfInt(0));
            }
            set
            {
                this.fieldFlags.SetValue(value);
            }
        }

        public PdfInt MaxLengthOfInputCharacters
        {
            get
            {
                return this.GetInheritableProperty((formFieldNode) => formFieldNode.maxLengthOfInputCharacters.GetValue());
            }
            set
            {
                this.maxLengthOfInputCharacters.SetValue(value);
            }
        }

        public PrimitiveWrapper FieldValue
        {
            get
            {
                return this.GetInheritableProperty((formFieldNode) => formFieldNode.fieldValue.GetValue());
            }
            set
            {
                this.fieldValue.SetValue(value);
            }
        }

        public PrimitiveWrapper DefaultFieldValue
        {
            get
            {
                return this.GetInheritableProperty((formFieldNode) => formFieldNode.defaultFieldValue.GetValue());
            }
            set
            {
                this.defaultFieldValue.SetValue(value);
            }
        }

        public PdfDictionary AdditionalActions
        {
            get
            {
                return this.additionalActions.GetValue();
            }
            set
            {
                this.additionalActions.SetValue(value);
            }
        }

        public PdfArray Options
        {
            get
            {
                return this.GetInheritableProperty((formFieldNode) => formFieldNode.options.GetValue());
            }
            set
            {
                this.options.SetValue(value);
            }
        }

        public PdfInt TopIndex
        {
            get
            {
                return this.topIndex.GetValue();
            }
            set
            {
                this.topIndex.SetValue(value);
            }
        }

        public PdfArray SortedIndices
        {
            get
            {
                return this.sortedIndices.GetValue();
            }
            set
            {
                this.sortedIndices.SetValue(value);
            }
        }

        public SignatureFieldLock SignаtureFieldLock
        {
            get
            {
                return this.signatureFieldLock.GetValue();
            }
            set
            {
                this.signatureFieldLock.SetValue(value);
            }
        }

        public SignatureFieldSeed SeedValue
        {
            get
            {
                return this.seedValue.GetValue();
            }
            set
            {
                this.seedValue.SetValue(value);
            }
        }

        public string FullName
        {
            get
            {
                string name = null;

                FormFieldNode node = this;

                while (node != null)
                {
                    if (node.PartialName != null)
                    {
                        if (name == null)
                        {
                            name = node.PartialName.ToString();
                        }
                        else
                        {
                            name = string.Format("{0}.{1}", node.PartialName.ToString(), name);
                        }
                    }

                    node = node.Parent;
                }

                return name;
            }
        }

        FormFieldNode ITreeNode<FormFieldNode>.NodeValue
        {
            get
            {
                return this;
            }
        }

        ITreeNode<FormFieldNode> ITreeNode<FormFieldNode>.ParentNode
        {
            get
            {
                return this.Parent;
            }
        }

        public PdfString GetDefaultAppearance(AcroFormObject form)
        {
            return this.GetInheritableProperty((formFieldNode) => formFieldNode.defaultAppearance.GetValue(), form.DefaultAppearance);
        }

        public void SetDefaultAppearance(PdfString value)
        {
            this.defaultAppearance.SetValue(value);
        }

        public PdfInt GetQuadding(AcroFormObject form)
        {
            return this.GetInheritableProperty((formFieldNode) => formFieldNode.quadding.GetValue(), form.Quadding);
        }

        public void SetQuadding(PdfInt value)
        {
            this.quadding.SetValue(value);
        }

        internal void CopyFormField(FormField field, IPdfExportContext context)
        {
            Guard.ThrowExceptionIfNull(field, "field");

            FieldPropertiesExporter fieldExporter = new FieldPropertiesExporter(this, context);
            fieldExporter.ExportFieldProperties(field);            
        }        

        internal FormField ToFormField(PostScriptReader reader, IRadFixedDocumentImportContext context, bool importFieldProperties)
        {
            FormField field;
            if (!context.TryGetField(this, out field))
            {
                FieldCreationContext fieldContext = new FieldCreationContext(context, this);
                field = FieldsFactory.CreateField(fieldContext);
                context.MapFields(this, field);

                if(importFieldProperties)
                {
                    FieldPropertiesImporter propertiesImporter = new FieldPropertiesImporter(this, reader, context);
                    propertiesImporter.ImportFieldProperties(field);
                }
            }

            return field;
        }
    }
}
