﻿using System;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Streaming;
#if !NETSTANDARD
using System.Windows;
#else
using Telerik.Documents.Primitives;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Objects;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Objects
{
    internal class FormXObject : XObjectBase
    {
        private readonly DirectProperty<PdfArray> boundingBox;
        private readonly DirectProperty<PdfDictionary> resources;
        private readonly DirectProperty<PdfArray> matrix;
        private readonly byte[] encodedContentData;
        private byte[] decodedContentData;

        public FormXObject(PdfStreamBase importedStream, PdfFileSource fileSource, PdfFileStreamExportContext exportContext)
            : this()
        {
            this.encodedContentData = importedStream.ReadRawPdfData();
            this.Load(fileSource.Context.Reader, fileSource.Context, importedStream.Dictionary);

            if (this.Filters != null)
            {
                this.Filters = ResourceRenamer.GetArrayWithRenamedResources(this.Filters, fileSource, exportContext);
            }

            if (this.DecodeParms != null)
            {
                switch (this.DecodeParms.Type)
                {
                    case PdfElementType.Dictionary:
                        PdfDictionary importedDictionary = (PdfDictionary)this.DecodeParms.Primitive;
                        this.DecodeParms = new PrimitiveWrapper(ResourceRenamer.GetDictionaryWithRenamedResources(importedDictionary, fileSource, exportContext));
                        break;
                    case PdfElementType.Array:
                        PdfArray importedArray = (PdfArray)this.DecodeParms.Primitive;
                        this.DecodeParms = new PrimitiveWrapper(ResourceRenamer.GetArrayWithRenamedResources(importedArray, fileSource, exportContext));
                        break;
                }
            }
        }

        public FormXObject(byte[] decodedContentData)
            : this()
        {
            this.decodedContentData = decodedContentData;
        }

        public FormXObject()
        {
            this.boundingBox = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor("BBox", true));
            this.resources = this.RegisterDirectProperty<PdfDictionary>(new PdfPropertyDescriptor("Resources"));
            this.matrix = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor("Matrix"));
        }

        public PdfArray BoundingBox
        {
            get
            {
                return this.boundingBox.GetValue();
            }
            set
            {
                this.boundingBox.SetValue(value);
            }
        }

        public PdfArray Matrix
        {
            get
            {
                return this.matrix.GetValue();
            }
            set
            {
                this.matrix.SetValue(value);
            }
        }

        public PdfDictionary Resources
        {
            get
            {
                return this.resources.GetValue();
            }
            set
            {
                this.resources.SetValue(value);
            }
        }

        public override XObjectType XObjectType
        {
            get
            {
                return XObjectType.Form;
            }
        }

        internal byte[] DecodedContentData
        {
            get
            {
                return this.decodedContentData;
            }
        }

        public override void Write(PdfWriter writer, IPdfExportContext context)
        {
            if (this.decodedContentData == null)
            {
                this.WriteDictionaryAndEncodedData(writer, context, this.encodedContentData);
            }
            else
            {
                base.Write(writer, context);
            }
        }

        public void CopyPropertiesFrom(IPdfExportContext context, FormSource formSource)
        {
            this.CopyPropertiesFrom(context, formSource, formSource.BoundingBox);
        }

        public void CopyPropertiesFrom(IPdfExportContext context, IContentRootElement contentRoot, Rect boundingBox)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(contentRoot, "contentRoot");

            ResourceHolder resourceHolder = new ResourceHolder();
            this.decodedContentData = ContentStream.BuildContentData(context, resourceHolder, contentRoot);
            this.BoundingBox = contentRoot.ToBottomLeftCoordinateSystem(boundingBox);
            this.CopyResourcesFrom(resourceHolder);
        }

        protected override byte[] GetData(IPdfExportContext context)
        {
            if (this.decodedContentData == null)
            {
                return base.GetData(context);
            }
            else
            {
                return this.decodedContentData;
            }
        }

        protected override void InterpretData(PostScriptReader reader, IPdfImportContext context, PdfStreamBase stream)
        {
            this.decodedContentData = stream.ReadDecodedPdfData();
        }

        internal FormSource ToFormSource(PostScriptReader reader, IRadFixedDocumentImportContext context)
        {
            FormSource formSource = new FormSource();
            Rect boundingBoxRect = this.BoundingBox.ToRect(reader, context);
            formSource.BoundingBox = boundingBoxRect.ToTopLeftCoordinateSystem(boundingBoxRect.Height);

            ResourceHolder resourceHolder = new ResourceHolder();

            if (this.Resources != null)
            {
                resourceHolder.Resources.Load(reader, context, this.Resources);
            }

            if (this.Matrix != null)
            {
                formSource.Matrix = this.Matrix.ToMatrix(reader, context);
            }

            ContentStream.ParseContentData(this.decodedContentData, context, resourceHolder, formSource);

            return formSource;
        }

        internal void CopyResourcesFrom(IResourceHolder resourceHolder)
        {
            this.CopyResourcesFrom(resourceHolder,
                (property) =>
                {
                    return property.GetValue();
                });
        }

        internal void CopyResourcesFrom(IResourceHolder resourceHolder, PdfFileSource fileSource, PdfFileStreamExportContext exportContext)
        {
            this.CopyResourcesFrom(resourceHolder,
                (property) =>
                {
                    IPdfPrimitive primitive = property.GetValue();
                    IPdfPrimitive renamedResourcesPrimitive = null;

                    switch (primitive.Type)
                    {
                        case PdfElementType.Dictionary:
                            PdfDictionary initialDictionary = (PdfDictionary)primitive;
                            renamedResourcesPrimitive = ResourceRenamer.GetDictionaryWithRenamedResources(initialDictionary, fileSource, exportContext);
                            break;
                        case PdfElementType.Array:
                            PdfArray initialArray = (PdfArray)primitive;
                            renamedResourcesPrimitive = ResourceRenamer.GetArrayWithRenamedResources(initialArray, fileSource, exportContext);
                            break;
                        default:
                            throw new InvalidOperationException(string.Format("Invalid resource primitive type: {0}", primitive.Type));
                    }

                    return renamedResourcesPrimitive;
                });
        }

        private void CopyResourcesFrom(IResourceHolder resourceHolder, Func<IPdfProperty, IPdfPrimitive> getValueFromProperty)
        {
            foreach (IPdfProperty property in resourceHolder.Resources.Properties.Values)
            {
                if (property.HasValue && property.HasNonDefaultValue)
                {
                    if (this.Resources == null)
                    {
                        this.Resources = new PdfDictionary();
                    }

                    this.Resources[property.Descriptor.Name] = getValueFromProperty(property);
                }
            }
        }
    }
}
