﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Utilities;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Converters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.ColorSpaces;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Objects
{
    internal class ImageXObject : XObjectBase
    {
        public const string ImageMaskPropertyName = "ImageMask";

        private readonly DirectProperty<PdfInt> width;
        private readonly DirectProperty<PdfInt> height;
        private readonly DirectProperty<ColorSpaceObject> colorSpace;
        private readonly DirectProperty<PdfInt> bitsPerComponent;
        private readonly DirectProperty<PdfArray> decode;
        private readonly DirectProperty<PdfBool> imageMask;
        private readonly ReferenceProperty<Mask> mask;
        private readonly ReferenceProperty<ImageXObject> sMask;        

        private PdfStreamBase stream;
        private byte[] encodedImageData;
        private PdfArray filters;

        private byte[] EncodedImageData
        {
            get
            {
                if (this.encodedImageData == null && this.stream != null)
                {
                    this.encodedImageData = this.stream.ReadRawPdfData();
                }

                return this.encodedImageData;
            }
            set
            {
                this.encodedImageData = value;
            }
        }

        public ImageXObject()
        {
            this.width = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("Width", true));
            this.height = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("Height", true));
            this.colorSpace = this.RegisterDirectProperty<ColorSpaceObject>(new PdfPropertyDescriptor("ColorSpace"));
            this.bitsPerComponent = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("BitsPerComponent"));
            this.decode = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor("Decode"));
            this.imageMask = this.RegisterDirectProperty<PdfBool>(new PdfPropertyDescriptor(ImageMaskPropertyName));
            this.mask = this.RegisterReferenceProperty<Mask>(new PdfPropertyDescriptor("Mask"));
            this.sMask = this.RegisterReferenceProperty<ImageXObject>(new PdfPropertyDescriptor("SMask"));
            this.filters = Fixed.Model.FixedDocumentDefaults.DefaultImageFilters.ToPdfArray();
        }

        public PdfInt Width
        {
            get
            {
                return this.width.GetValue();
            }
            set
            {
                this.width.SetValue(value);
            }
        }

        public PdfInt Height
        {
            get
            {
                return this.height.GetValue();
            }
            set
            {
                this.height.SetValue(value);
            }
        }

        public ColorSpaceObject ColorSpace
        {
            get
            {
                return this.colorSpace.GetValue();
            }
            set
            {
                this.colorSpace.SetValue(value);
            }
        }

        public PdfInt BitsPerComponent
        {
            get
            {
                return this.bitsPerComponent.GetValue();
            }
            set
            {
                this.bitsPerComponent.SetValue(value);
            }
        }

        public PdfArray Decode
        {
            get
            {
                return this.decode.GetValue();
            }
            set
            {
                this.decode.SetValue(value);
            }
        }

        public PdfBool ImageMask
        {
            get
            {
                return this.imageMask.GetValue();
            }
            set
            {
                this.imageMask.SetValue(value);
            }
        }

        public Mask Mask
        {
            get
            {
                return this.mask.GetValue();
            }
            set
            {
                this.mask.SetValue(value);
            }
        }

        public ImageXObject SMask
        {
            get
            {
                return this.sMask.GetValue();
            }
            set
            {
                this.sMask.SetValue(value);
            }
        }

        public override XObjectType XObjectType
        {
            get
            {
                return XObjectType.Image;
            }
        }

        public override void Write(PdfWriter writer, IPdfExportContext context)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(context, "context");

            if (this.ExportAs == PdfElementType.PdfStreamObject)
            {
                byte[] data = this.GetData(context);
                this.Filters = this.GetExportFilters(context);

                if (this.ShouldEncryptData)
                {
                    data = context.EncryptStream(data);
                }

                this.Length = new PdfInt(data.Length);

                WritePdfPropertiesDictionary(this, writer, context);

                writer.WriteLine();
                writer.WriteLine(PdfNames.PdfStreamStart);
                writer.Write(data);
                writer.WriteLine();
                writer.Write(PdfNames.PdfStreamEnd);
            }
            else
            {
                WritePdfPropertiesDictionary(this, writer, context);
            }
        }

        public void CopyPropertiesFrom(IPdfExportContext context, ImageSource imageSource)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(imageSource, "imageSource");

            this.Width = new PdfInt(imageSource.Width);
            this.Height = new PdfInt(imageSource.Height);
            this.EncodedImageData = imageSource.GetEncodedImageData(context.Settings.ImageQuality);
            this.filters = imageSource.Filters.ToPdfArray();

            this.ColorSpace = imageSource.ColorSpace.ToPrimitive((colorSpace) => ColorSpaceManager.CreateColorSpaceObject(colorSpace));
            this.ImageMask = imageSource.ImageMask.ToPrimitive((imageMask) => new PdfBool(imageMask));
            this.BitsPerComponent = imageSource.BitsPerComponent.ToPrimitive((bits) => new PdfInt(bits));
            this.Mask = imageSource.Mask.ToPrimitive((imageMask) => Mask.CalculateMask(imageMask, context));
            this.SMask = imageSource.SMask.ToPrimitive((sMask) => ImageXObject.CalculateSMask(sMask, context));

            this.DecodeParms = DecodeParametersConverter.CreateDecodeParms(imageSource.DecodeParameters);

            if (imageSource.DecodeArray != null)
            {
                this.Decode = imageSource.DecodeArray.ToPdfArray();
            }
        }

        internal ImageSource ToImageSource()
        {
            int bitsPerComponent = this.BitsPerComponent != null ? this.BitsPerComponent.Value : 0;
            var colorSpace = this.ColorSpace != null ? this.ColorSpace.ToColorSpace() : null;
            string[] filters = this.Filters != null ? this.Filters.ToStringArray() : new string[0];
            double[] decodeArray = this.Decode != null ? this.Decode.ToDoubleArray() : null;
            bool imageMask = this.ImageMask != null ? this.ImageMask.Value : false;
            ImageSource imageSource =
                new ImageSource(this.EncodedImageData, bitsPerComponent, this.Width.Value, this.Height.Value, colorSpace, filters, decodeArray, imageMask);
            
            imageSource.DecodeParameters = this.stream.DecodeParameters;

            this.Mask.CopyToProperty(imageSource.Mask, Mask.CreateImageMask);
            this.SMask.CopyToProperty(imageSource.SMask, (smask) => smask.ToImageSource());

            return imageSource;
        }

        protected override byte[] GetData(IPdfExportContext context)
        {
            byte[] data = this.EncodedImageData;
            this.EncodedImageData = null;

            return data;
        }

        protected override void InterpretData(PostScriptReader reader, IPdfImportContext context, PdfStreamBase stream)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");

            this.stream = stream;
        }

        protected override PdfArray GetExportFilters(IPdfExportContext context)
        {
            if (this.stream != null)
            {
                return (PdfArray)this.stream.Filters;
            }

            return this.filters;
        }

        protected override Filters.PdfObject CreateEncodeObject(IPdfExportContext context)
        {
            var colorSpace = this.ColorSpace != null ? this.ColorSpace.Public : Pdf.Filters.ColorSpace.Undefined;
            return new Filters.PdfObject(context, this.Width.Value, this.Height.Value, colorSpace);
        }

        private static Mask CalculateMask(ImageMask imageMask, IPdfExportContext context)
        {
            Mask mask;

            if (imageMask == null)
            {
                mask = null;
            }
            else if (imageMask.IsColorKeyMask)
            {
                mask = new Mask(imageMask.ColorMask.Value.ToPdfArray());
            }
            else
            {
                mask = new Mask();
                mask.CopyPropertiesFrom(context, imageMask.Image.Value);
                mask.ImageMask = new PdfBool(true);
            }

            return mask;
        }

        private static ImageXObject CalculateSMask(ImageSource sMask, IPdfExportContext context)
        {
            ImageXObject image;

            if (sMask == null)
            {
                image = null;
            }
            else
            {
                ResourceEntry entry = context.GetResource(sMask);
                image = (ImageXObject)entry.Resource.Content;
            }

            return image;
        }
    }
}