﻿using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Actions;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.ColorSpaces;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Destinations;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.Collections;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Navigation;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Outlines
{
    internal class OutlineItemObject : BaseOutlineObject
    {
        private readonly DirectProperty<PdfString> title;
        private readonly ReferenceProperty<BaseOutlineObject> parent;
        private readonly ReferenceProperty<OutlineItemObject> prev;
        private readonly ReferenceProperty<OutlineItemObject> next;
        private readonly ReferenceProperty<IDestinationObject> destination;
        private readonly ReferenceProperty<ActionObject> action;
        private readonly DirectProperty<PdfArray> color;
        private readonly DirectProperty<PdfInt> flag;

        public OutlineItemObject()
        {
            this.title = this.RegisterDirectProperty<PdfString>(new PdfPropertyDescriptor("Title", true));
            this.parent = this.RegisterReferenceProperty<BaseOutlineObject>(new PdfPropertyDescriptor("Parent", true, PdfPropertyRestrictions.MustBeIndirectReference));
            this.prev = this.RegisterReferenceProperty<OutlineItemObject>(new PdfPropertyDescriptor("Prev", false, PdfPropertyRestrictions.MustBeIndirectReference));
            this.next = this.RegisterReferenceProperty<OutlineItemObject>(new PdfPropertyDescriptor("Next", false, PdfPropertyRestrictions.MustBeIndirectReference));
            this.destination = this.RegisterReferenceProperty<IDestinationObject>(new PdfPropertyDescriptor("Dest"));
            this.action = this.RegisterReferenceProperty<ActionObject>(new PdfPropertyDescriptor("A"));
            this.color = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor("C"));
            this.flag = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("F"));
        }

        internal PdfString Title
        {
            get
            {
                return this.title.GetValue();
            }
            set
            {
                this.title.SetValue(value);
            }
        }

        internal BaseOutlineObject Parent
        {
            get
            {
                return this.parent.GetValue();
            }
            set
            {
                this.parent.SetValue(value);
            }
        }

        internal OutlineItemObject Prev
        {
            get
            {
                return this.prev.GetValue();
            }
            set
            {
                this.prev.SetValue(value);
            }
        }

        internal OutlineItemObject Next
        {
            get
            {
                return this.next.GetValue();
            }
            set
            {
                this.next.SetValue(value);
            }
        }

        internal IDestinationObject Destination
        {
            get
            {
                return this.destination.GetValue();
            }
            set
            {
                this.destination.SetValue(value);
            }
        }

        internal ActionObject Action
        {
            get
            {
                return this.action.GetValue();
            }
            set
            {
                this.action.SetValue(value);
            }
        }

        internal PdfArray Color
        {
            get
            {
                return this.color.GetValue();
            }
            set
            {
                this.color.SetValue(value);
            }
        }

        internal PdfInt Flag
        {
            get
            {
                return this.flag.GetValue();
            }
            set
            {
                this.flag.SetValue(value);
            }
        }

        internal BookmarkItem ToBookmarkItem(PostScriptReader reader, IRadFixedDocumentImportContext context)
        {
            string title = this.Title != null ? this.Title.ToString() : string.Empty;
            BookmarkItem bookmark = null;

            if (this.Destination != null)
            {
                if (this.Destination.IsNamed)
                {
                    NamedDestinationObject namedDestinationObject = (NamedDestinationObject)this.Destination;
                    NamedDestination namedDestination = namedDestinationObject.ToNamedDestination(context);
                    bookmark = new BookmarkItem(title, namedDestination);
                }
                else
                {
                    DestinationObject destinationObject = (DestinationObject)this.Destination;
                    Destination destination = destinationObject.ToDestination(reader, context);
                    bookmark = new BookmarkItem(title, destination);
                }
            }

            if (this.Action != null)
            {
                Fixed.Model.Actions.Action action = this.Action.ToAction(reader, context);
                bookmark = new BookmarkItem(title, action);
            }

            if (bookmark == null)
            {
                bookmark = new BookmarkItem(title);
            }

            bookmark.IsExpanded = this.Count != null && this.Count.Value > 0;
            bookmark.TextStyle = this.Flag != null ? (BookmarkItemStyles)this.Flag.Value : BookmarkItemStyles.Normal;

            this.PopulateChildBookmarkItems(bookmark, this.First, reader, context);

            if (this.Color != null)
            {
                RgbColor color = this.ToRgbColor();

                bookmark.TextColor = color;
            }

            return bookmark;
        }

        internal void CopyPropertiesFrom(IRadFixedDocumentExportContext context, BookmarkItem bookmarkItem)
        {
            this.Title = new PdfHexString(bookmarkItem.Title, StringType.UTFEncodedTextString);

            BookmarksCollection children = bookmarkItem.Children;
            int childrenCount = children.Count;
            this.Count = new PdfInt(bookmarkItem.IsExpanded ? childrenCount : -childrenCount);

            this.CopyChildBookmarksPropertiesFrom(context, children, childrenCount);

            if (bookmarkItem.Destination != null)
            {
                DestinationObject destinationObject = new DestinationObject();
                destinationObject.CopyPropertiesFrom(context, bookmarkItem.Destination);

                this.Destination = destinationObject;
            }

            if (bookmarkItem.NamedDestination != null)
            {
                NamedDestinationObject destinationObject = new NamedDestinationObject();
                destinationObject.CopyPropertiesFrom(context, bookmarkItem.NamedDestination);

                this.Destination = destinationObject;
            }

            if (bookmarkItem.Action != null)
            {
                this.Action = ActionObject.CreateAction(context, bookmarkItem.Action);
            }

            if (bookmarkItem.TextColor != null)
            {
                PdfReal redComponent = ColorObjectBase.ConvertFromByte(bookmarkItem.TextColor.R);
                PdfReal greenComponent = ColorObjectBase.ConvertFromByte(bookmarkItem.TextColor.G);
                PdfReal blueComponent = ColorObjectBase.ConvertFromByte(bookmarkItem.TextColor.B);

                this.Color = new PdfArray(redComponent, greenComponent, blueComponent);
            }

            this.Flag = new PdfInt((int)bookmarkItem.TextStyle);
        }

        private void PopulateChildBookmarkItems(
            BookmarkItem parentBookmark,
            OutlineItemObject childOutline,
            PostScriptReader reader,
            IRadFixedDocumentImportContext context)
        {
            while (childOutline != null)
            {
                BookmarkItem childBookmark = childOutline.ToBookmarkItem(reader, context);
                parentBookmark.Children.Add(childBookmark);

                childOutline = childOutline.Next;
            }
        }

        private RgbColor ToRgbColor()
        {
            PdfReal redComponent;

            if (!this.Color[0].TryConvertToPdfReal(out redComponent))
            {
                redComponent = new PdfReal(0);
            }

            PdfReal greenComponent;

            if (!this.Color[1].TryConvertToPdfReal(out greenComponent))
            {
                redComponent = new PdfReal(0);
            }

            PdfReal blueComponent;

            if (!this.Color[2].TryConvertToPdfReal(out blueComponent))
            {
                redComponent = new PdfReal(0);
            }

            RgbColor color = new RgbColor()
            {
                R = ColorObjectBase.ConvertToByte(redComponent),
                G = ColorObjectBase.ConvertToByte(greenComponent),
                B = ColorObjectBase.ConvertToByte(blueComponent)
            };

            return color;
        }

        private void CopyChildBookmarksPropertiesFrom(IRadFixedDocumentExportContext context, BookmarksCollection children, int childrenCount)
        {
            OutlineItemObject prevOutlineItemObject = null;

            for (int i = 0; i < childrenCount; i++)
            {
                BookmarkItem child = children[i];

                OutlineItemObject childOutlineItemObject = new OutlineItemObject();
                childOutlineItemObject.CopyPropertiesFrom(context, child);
                childOutlineItemObject.Parent = this;

                if (prevOutlineItemObject != null)
                {
                    childOutlineItemObject.Prev = prevOutlineItemObject;
                    prevOutlineItemObject.Next = childOutlineItemObject;
                }

                if (i == 0)
                {
                    this.First = childOutlineItemObject;
                }

                if (i == childrenCount - 1)
                {
                    this.Last = childOutlineItemObject;
                }

                prevOutlineItemObject = childOutlineItemObject;
            }
        }
    }
}
