﻿using System;
using System.Collections.Generic;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.ColorSpaces;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Functions;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Common.Functions;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Patterns
{
    internal abstract class GradientShading : Shading
    {
        private readonly DirectProperty<PdfArray> coordinates;
        private readonly DirectProperty<PdfArray> domain;
        private readonly ReferenceProperty<FunctionObject> function;
        private readonly DirectProperty<PdfArray> extend;

        public GradientShading()
        {
            this.coordinates = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor("Coords"));
            this.domain = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor("Domain"), new PdfArray(new PdfReal(0), new PdfReal(1)));
            this.function = this.RegisterReferenceProperty<FunctionObject>(new PdfPropertyDescriptor("Function"));
            this.extend = this.RegisterDirectProperty<PdfArray>(new PdfPropertyDescriptor("Extend"), new PdfArray(new PdfBool(false), new PdfBool(false)));
        }

        public PdfArray Coordinates
        {
            get
            {
                return this.coordinates.GetValue();
            }
            set
            {
                this.coordinates.SetValue(value);
            }
        }

        public PdfArray Domain
        {
            get
            {
                return this.domain.GetValue();
            }
            set
            {
                this.domain.SetValue(value);
            }
        }

        public FunctionObject Function
        {
            get
            {
                return this.function.GetValue();
            }
            set
            {
                this.function.SetValue(value);
            }
        }

        public PdfArray Extend
        {
            get
            {
                return this.extend.GetValue();
            }
            set
            {
                this.extend.SetValue(value);
            }
        }

        public override ColorBase ToColor(Import.Parser.PostScriptReader reader, Import.IPdfContentImportContext context, Matrix matrix)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
             
            Gradient gradient = this.ToColorOverride(reader, context, matrix);

            gradient.ExtendBefore = ((PdfBool)this.Extend[0]).Value;
            gradient.ExtendAfter = ((PdfBool)this.Extend[1]).Value;
             
            gradient.Background = this.GetBackgroundColor(reader, context);  

            return gradient; 
        }

        protected void InitializeGradientStops(Import.Parser.PostScriptReader reader, Import.IPdfContentImportContext context, Gradient gradient, Matrix matrix, double distance)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(gradient, "gradient");

            FunctionBase function = this.Function.ToFunction(reader, context.Owner);
            double t0 = this.Domain[0].ToReal();
            double t1 = this.Domain[1].ToReal();

            IEnumerable<double> tCoeficients = GradientStopsCalculator.GetGradientStopPositionCoeficients(function, matrix, t0, t1, distance, this.ColorSpace.Name);

            foreach (double t in tCoeficients)
            {
                double[] functionResult = function.Execute(new double[] { t });
                PdfArray pdfFunctionResult = functionResult.ToPdfArray();
                ColorObjectBase colorObject = this.ColorSpace.GetColor(context, reader, pdfFunctionResult);
                ColorBase color = colorObject.ToColor(reader, context);
                SimpleColor simpleColor = (SimpleColor)color;

                gradient.GradientStops.Add(new Fixed.Model.ColorSpaces.GradientStop(simpleColor, t));
            }
        }

        protected override void CopyPropertiesFromOverride(IPdfContentExportContext context, Gradient gradient)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(gradient, "gradient");

            this.Extend = new PdfArray();
            this.Extend.Add(gradient.ExtendBefore.ToPdfBool());
            this.Extend.Add(gradient.ExtendAfter.ToPdfBool());

            this.Function = CreateFunctionObject(context.Owner, gradient);
        }

        protected abstract Gradient ToColorOverride(Import.Parser.PostScriptReader reader, Import.IPdfContentImportContext context, Matrix matrix);
        
        private static FunctionObject CreateFunctionObject(IPdfExportContext exportContext, Gradient gradient)
        {
            FunctionBase function = CreateFunction(gradient);
            FunctionObject functionObject = FunctionObject.CreateFunction(new PdfInt((int)function.FunctionType));
            functionObject.CopyPropertiesFrom(exportContext, function);

            return functionObject;
        }

        private static FunctionBase CreateFunction(Gradient gradient)
        {
            if (gradient.GradientStops.Count > 1)
            {
                return CreateStitchingFunction(gradient);
            }
            else if (gradient.GradientStops.Count == 1)
            {
                var gradientStop = gradient.GradientStops.First();

                return CreateGradientFunction(0, 1, gradientStop.Color, gradientStop.Color);
            }
            else
            {
                return CreateGradientFunction(0, 1, RgbColors.Transparent, RgbColors.Transparent);
            }
        }

        private static FunctionBase CreateStitchingFunction(Gradient gradient)
        {
            Fixed.Model.ColorSpaces.GradientStop[] gradientStops = gradient.GradientStops.OrderBy(s => s.Offset).ToArray();
            double[] domain = new double[] { 0, 1 };
            List<double> bounds = new List<double>();
            List<double> encode = new List<double>();
            List<FunctionBase> functions = new List<FunctionBase>();

            int last = gradientStops.Length - 1;
            int beforeLast = last - 1;
            if (gradientStops[0].Offset > 0)
            {
                functions.Add(CreateGradientFunction(
                   0,
                   gradientStops[0].Offset,
                   gradientStops[0].Color,
                   gradientStops[0].Color));

                encode.AddRange(new double[] { 0, gradientStops[0].Offset });
                bounds.Add(gradientStops[0].Offset);
            }

            for (int i = 0; i < last; i++)
            {
                functions.Add(CreateGradientFunction(
                    gradientStops[i].Offset,
                    gradientStops[i + 1].Offset, 
                    gradientStops[i].Color, 
                    gradientStops[i + 1].Color));

                encode.AddRange(new double[] { gradientStops[i].Offset, gradientStops[i + 1].Offset });

                if (i < beforeLast)
                {
                    bounds.Add(gradientStops[i + 1].Offset);
                }
            }

            if (gradientStops[last].Offset < 1)
            {
                functions.Add(CreateGradientFunction(
                   gradientStops[last].Offset,
                   1,
                   gradientStops[last].Color,
                   gradientStops[last].Color));

                encode.AddRange(new double[] { gradientStops[last].Offset, 1 });
                bounds.Add(gradientStops[last].Offset);
            }

            StitchingFunction stitchingFunction = new StitchingFunction(functions.ToArray(), bounds.ToArray(), encode.ToArray(), domain, null);

            return stitchingFunction;
        }

        private static FunctionBase CreateGradientFunction(double startOffset, double endOffset, SimpleColor startColor, SimpleColor endColor)
        {
            Guard.ThrowExceptionIfNull(startColor, "startColor");
            Guard.ThrowExceptionIfNull(endColor, "endColor");

            double[] domain = new double[] { startOffset, endOffset };
            double[] range = new double[] { 0, 1, 0, 1, 0, 1 };
            int[] size = new int[] { 2 };
            int bitsPerSample = 8;

            int[][] samples = new int[2][];
            samples[0] = startColor.GetColorComponents();
            samples[1] = endColor.GetColorComponents();

            SampledFunction function = new SampledFunction(samples, size, bitsPerSample, domain, range, null, null);

            return function;
        }        
    }
}
