﻿using System;
using System.Collections.Generic;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Common.Functions;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Patterns
{
    internal static class GradientStopsCalculator
    {
        public static IEnumerable<double> GetGradientStopPositionCoeficients(FunctionBase function, Matrix matrix, double t0, double t1, double distance, string colorspaceName)
        {
            yield return t0;

            IEnumerable<double> tCoeficients;
            if (!(IsColorSpaceComplientToMinimalStopsInterpolation(colorspaceName) && TryGetMinimalNumberOfGradientStops(function, t0, t1, out tCoeficients)))
            {
                tCoeficients = CalculateEvenlyDistributedStopCoeficients(t0, t1, matrix, distance);
            }

            foreach(double coeficient in tCoeficients)
            {
                yield return coeficient;
            }

            yield return t1;
        }

        /// <summary>
        /// Calculates step appropriate for calculating neighbouring colors in a linear gradient.
        /// </summary>
        /// <param name="gradient">The linear gradient.</param>
        /// <returns>The step.</returns>
        public static double CalculateLinearGradientStopsMaximalStep(IGradient gradient)
        {
            double dist = MathUtilities.GetDistance(gradient.StartPoint, gradient.EndPoint);
            double step = CalculateGradientStopsMaximalStep(gradient.Position, dist);

            return step;
        }

        /// <summary>
        /// Calculates step appropriate for calculating neighbouring colors in a radial gradient.
        /// </summary>
        /// <param name="gradient">The radial gradient.</param>
        /// <returns>The step.</returns>
        public static double CalculateRadialGradientStopsMaximalStep(IRadialGradient gradient)
        {
            double dist = MathUtilities.GetDistance(gradient.StartPoint, gradient.EndPoint);
            double dR = Math.Abs(gradient.EndRadius - gradient.StartRadius);
            double step = CalculateGradientStopsMaximalStep(gradient.Position, dR + dist);

            return step;
        }

        private static double CalculateGradientStopsMaximalStep(Matrix matrix, double distance)
        {
            double d1 = Math.Sqrt(Math.Pow(matrix.M11 + matrix.M21, 2) + Math.Pow(matrix.M12 + matrix.M22, 2));
            double d2 = Math.Sqrt(Math.Pow(matrix.M11 - matrix.M21, 2) + Math.Pow(matrix.M12 - matrix.M22, 2));
            double transformedDistance = distance * Math.Max(d1, d2);
            double step = 1 / (transformedDistance * 3);

            return step;
        }

        private static bool IsColorSpaceComplientToMinimalStopsInterpolation(string colorSpaceName)
        {
            return colorSpaceName == ColorSpaceNames.DeviceRgb || colorSpaceName == ColorSpaceNames.DeviceGray;
        }

        private static bool TryGetMinimalNumberOfGradientStops(FunctionBase function, double t0, double t1, out IEnumerable<double> tCoeficients)
        {
            if (function.FunctionType == FunctionType.Sampled)
            {
                tCoeficients = GetGradientStopCoeficients((SampledFunction)function, t0, t1);
                return true;
            }
            else if (function.FunctionType == FunctionType.Stitching)
            {
                bool canGetMinimalNumberOfGradientStops = TryGetMinimalNumberOfGradientStops((StitchingFunction)function, t0, t1, out tCoeficients);
                return canGetMinimalNumberOfGradientStops;
            }
            else
            {
                tCoeficients = null;
                return false;
            }
        }

        private static bool TryGetMinimalNumberOfGradientStops(StitchingFunction function, double t0, double t1, out IEnumerable<double> tCoeficients)
        {
            int functionIndex = 0;
            double intervalStart = t0;
            List<double> aggregatedCoeficients = new List<double>();
            IEnumerable<double> innerCoeficients;
            tCoeficients = null;

            foreach (double boundValue in function.Bounds)
            {
                double intervalEnd = CalculateBoundsGradientStopCoeficient(function, boundValue, t0, t1);
                FunctionBase innerFunction = function.Functions[functionIndex];

                if (!TryGetMinimalNumberOfGradientStops(innerFunction, intervalStart, intervalEnd, out innerCoeficients))
                {
                    return false;
                }

                aggregatedCoeficients.AddRange(innerCoeficients);
                aggregatedCoeficients.Add(intervalEnd);

                functionIndex++;
                intervalStart = intervalEnd;
            }

            FunctionBase lastInnerFunction = function.Functions[functionIndex];

            if (!TryGetMinimalNumberOfGradientStops(lastInnerFunction, intervalStart, t1, out innerCoeficients))
            {
                return false;
            }

            aggregatedCoeficients.AddRange(innerCoeficients);

            tCoeficients = aggregatedCoeficients;
            return true;
        }

        private static double CalculateBoundsGradientStopCoeficient(StitchingFunction function, double boundValue, double t0, double t1)
        {
            double domainLength = function.Domain[1] - function.Domain[0];

            if (domainLength == 0)
            {
                return t0;
            }

            double leftRatio = (boundValue - function.Domain[0]) / domainLength;
            double rightRatio = (function.Domain[1] - boundValue) / domainLength;
            double gradientStopCoeficient = t0 * rightRatio + t1 * leftRatio;

            return gradientStopCoeficient;
        }

        private static IEnumerable<double> GetGradientStopCoeficients(SampledFunction function, double t0, double t1)
        {
            int numberOfSamples = function.Size[0];
            int numberOfGradientStops = numberOfSamples - 2;
            double step = (t1 - t0) / (numberOfGradientStops + 1);

            for (int i = 1; i <= numberOfGradientStops; i++)
            {
                double t = t0 + i * step;
                yield return t;
            }
        }

        private static IEnumerable<double> CalculateEvenlyDistributedStopCoeficients(double t0, double t1, Matrix matrix, double distance)
        {
            double step = CalculateGradientStopsMaximalStep(matrix, distance);

            if (t0 < t1)
            {
                for (double t = t0 + step; t < t1; t += step)
                {
                    yield return t;
                }
            }
            else
            {
                for (double t = t0 - step; t > t1; t -= step)
                {
                    yield return t;
                }
            }
        }
    }
}
