﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Fixed.Exceptions;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Encryption
{
    internal abstract class Encrypt : PdfObject
    {
        protected static readonly byte[] PasswordPadding =
        {
          0x28, 0xBF, 0x4E, 0x5E, 0x4E, 0x75, 0x8A, 0x41, 0x64, 0x00, 0x4E, 0x56, 0xFF, 0xFA, 0x01, 0x08,
          0x2E, 0x2E, 0x00, 0xB6, 0xD0, 0x68, 0x3E, 0x80, 0x2F, 0x0C, 0xA9, 0xFE, 0x64, 0x53, 0x69, 0x7A,
        };

        private readonly DirectProperty<PdfInt> algorithm;
        private readonly DirectProperty<PdfInt> length;
        private readonly DirectProperty<CryptFilters> cryptFilters;
        private readonly DirectProperty<PdfName> filter;
        private readonly DirectProperty<PdfName> streamsFilter;
        private readonly DirectProperty<PdfName> stringsFilter;

        public Encrypt()
        {
            this.filter = this.RegisterDirectProperty<PdfName>(new PdfPropertyDescriptor("Filter", true));
            this.algorithm = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("V"), new PdfInt(0));
            this.length = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("Length"), new PdfInt(40));
            this.cryptFilters = this.RegisterDirectProperty<CryptFilters>(new PdfPropertyDescriptor("CF"));
            this.streamsFilter = this.RegisterDirectProperty<PdfName>(new PdfPropertyDescriptor("StmF"), new PdfName(PdfNames.Identity));
            this.stringsFilter = this.RegisterDirectProperty<PdfName>(new PdfPropertyDescriptor("StrF"), new PdfName(PdfNames.Identity));
        }

        public PdfName Filter
        {
            get
            {
                return this.filter.GetValue();
            }
            set
            {
                this.filter.SetValue(value);
            }
        }

        public PdfName StreamsFilter
        {
            get
            {
                return this.streamsFilter.GetValue();
            }
            set
            {
                this.streamsFilter.SetValue(value);
            }
        }

        public PdfName StringsFilter
        {
            get
            {
                return this.stringsFilter.GetValue();
            }
            set
            {
                this.stringsFilter.SetValue(value);
            }
        }

        public PdfInt Algorithm
        {
            get
            {
                return this.algorithm.GetValue();
            }
            set
            {
                this.algorithm.SetValue(value);
            }
        }

        public PdfInt Length
        {
            get
            {
                return this.length.GetValue();
            }
            set
            {
                this.length.SetValue(value);
            }
        }

        public CryptFilters CryptFilters
        {
            get
            {
                return this.cryptFilters.GetValue();
            }
            set
            {
                this.cryptFilters.SetValue(value);
            }
        }

        public static byte[] PadOrTruncatePassword(string password)
        {
            byte[] passwordAsBytes = string.IsNullOrEmpty(password) ? new byte[0] : PdfEncoding.Encoding.GetBytes(password);

            byte[] padded = new byte[32];
            int numContributingPasswordBytes = passwordAsBytes.Length > padded.Length ? padded.Length : passwordAsBytes.Length;
            Array.Copy(passwordAsBytes, 0, padded, 0, numContributingPasswordBytes);
            if (passwordAsBytes.Length < padded.Length)
            {
                Array.Copy(PasswordPadding, 0, padded, passwordAsBytes.Length, padded.Length - passwordAsBytes.Length);
            }

            return padded;
        }

        public static Encrypt CreateInstance(PdfName type)
        {
            switch (type.Value)
            {
                case "Standard": return new StandardEncrypt();

                default:

                    throw new NotSupportedException("Encryption type is not supported.");
            }
        }

        public abstract bool AuthenticateUserPassword(byte[] id, string password);

        public abstract bool AuthenticateOwnerPassword(byte[] id, string password);

        public abstract byte[] EncryptStream(int objectNo, int generationNo, byte[] data);

        public abstract byte[] DecryptStream(int objectNo, int generationNo, byte[] data);

        public abstract byte[] EncryptString(int objectNo, int generationNo, byte[] data);

        public abstract byte[] DecryptString(int objectNo, int generationNo, byte[] data);

        public void CopyPropertiesFrom(IPdfExportContext context, PdfExportSettings settings)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(settings, "settings");

            this.CopyPropertiesFromOverride(context, settings);
        }

        protected virtual void CopyPropertiesFromOverride(IPdfExportContext context, PdfExportSettings settings)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(settings, "settings");
        }

        internal static void ValidateEncryptionAlgorithmValue(int algorithmValue)
        {
            if (algorithmValue < 0 || algorithmValue == 3 || algorithmValue > 4)
            {
                throw new NotSupportedEncryptionException(algorithmValue);
            }
        }
    }
}
