﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Security.Cryptography;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Encryption
{
    internal class StandardEncrypt : Encrypt, IStandardCryptFilterMethodProvider
    {
        private static readonly byte[] aesSAlTbytes = { 0x73, 0x41, 0x6c, 0x54 };
        private readonly DirectProperty<PdfInt> revision;
        private readonly DirectProperty<PdfString> owner;
        private readonly DirectProperty<PdfString> user;
        private readonly DirectProperty<PdfInt> permissions;

        private byte[] encryptionKey;

        public StandardEncrypt()
        {
            this.revision = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("R", true));
            this.owner = this.RegisterDirectProperty<PdfString>(new PdfPropertyDescriptor("O", true));
            this.user = this.RegisterDirectProperty<PdfString>(new PdfPropertyDescriptor("U", true));
            this.permissions = this.RegisterDirectProperty<PdfInt>(new PdfPropertyDescriptor("P", true));
        }

        public PdfInt Revision
        {
            get
            {
                return this.revision.GetValue();
            }
            set
            {
                this.revision.SetValue(value);
            }
        }

        public PdfString Owner
        {
            get
            {
                return this.owner.GetValue();
            }
            set
            {
                this.owner.SetValue(value);
            }
        }

        public PdfString User
        {
            get
            {
                return this.user.GetValue();
            }
            set
            {
                this.user.SetValue(value);
            }
        }

        public PdfInt Permissions
        {
            get
            {
                return this.permissions.GetValue();
            }
            set
            {
                this.permissions.SetValue(value);
            }
        }

        public override bool AuthenticateUserPassword(byte[] id, string password)
        {
            this.encryptionKey = CalculateEncryptionKey(password, id, this.Length.Value, (int)this.Revision.Value, this.Owner.Value, this.Permissions.Value, false);
            byte[] passwordBytes = CalculateUserPassword(this.encryptionKey, id);

            bool isCorrect = true;
            for (int i = 0; i < 16; i++)
            {
                if (this.User.Value[i] != passwordBytes[i])
                {
                    isCorrect = false;
                    break;
                }
            }

            return isCorrect;
        }

        public override bool AuthenticateOwnerPassword(byte[] id, string password)
        {
            throw new NotImplementedException();
        }

        public override byte[] EncryptStream(int objectNo, int generationNo, byte[] data)
        {
            Guard.ThrowExceptionIfNull(data, "data");

            string cryptFilter = this.Algorithm.Value == 4 ? this.StreamsFilter.Value : null;
            StandardEncryptionContext encryptionContext = GetEncryptionContext(this, objectNo, generationNo, data, this.encryptionKey, cryptFilter);
            byte[] res = Encrypt(encryptionContext);

            return res;
        }

        public override byte[] DecryptStream(int objectNo, int generationNo, byte[] data)
        {
            return this.EncryptStream(objectNo, generationNo, data);
        }

        public override byte[] EncryptString(int objectNo, int generationNo, byte[] data)
        {
            Guard.ThrowExceptionIfNull(data, "data");

            string cryptFilter = this.Algorithm.Value == 4 ? this.StringsFilter.Value : null;
            StandardEncryptionContext encryptionContext = GetEncryptionContext(this, objectNo, generationNo, data, this.encryptionKey, cryptFilter);
            byte[] res = Encrypt(encryptionContext);

            return res;
        }

        public override byte[] DecryptString(int objectNo, int generationNo, byte[] data)
        {
            return this.EncryptString(objectNo, generationNo, data);
        }

        protected override void CopyPropertiesFromOverride(IPdfExportContext context, PdfExportSettings settings)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(settings, "settings");
            
            int revision = FixedDocumentDefaults.StandardEncryptRevision;
            int length = FixedDocumentDefaults.StandardEncryptionKeyLength;
            int permissions = GetPermissionBits();

            this.Filter = new PdfName(PdfNames.StandardEncrypt);
            this.Revision = new PdfInt(revision);
            this.Algorithm = new PdfInt(2);
            this.Length = new PdfInt(length);
            this.Permissions = new PdfInt(permissions);

            byte[] owner = CalculateOwnerPassword(settings.OwnerPassword, settings.UserPassword, revision, length);
            this.encryptionKey = CalculateEncryptionKey(settings.UserPassword, context.DocumentId, length, revision, owner, permissions, false);
            byte[] user = CalculateUserPassword(this.encryptionKey, context.DocumentId);

            this.Owner = new PdfHexString(owner);
            this.User = new PdfHexString(user);

            base.CopyPropertiesFromOverride(context, settings);
        }
        
        internal static byte[] CalculateEncryptionKey(string password, byte[] id, int keyBitLength, int revision, byte[] oValue, int permissions, bool encryptMetadata)
        {
            List<byte> tmp = new List<byte>();

            byte[] passwordBytes = PadOrTruncatePassword(password);
            tmp.AddRange(passwordBytes);
            tmp.AddRange(oValue);

            byte[] bytes = BitConverter.GetBytes(permissions);
            tmp.AddRange(bytes);

            if (id != null)
            {
                tmp.AddRange(id);
            }

            if (revision >= 4 && encryptMetadata)
            {
                byte[] b = new byte[4];
                for (int i = 0; i < 4; ++i)
                {
                    b[i] = 0xFF;
                }

                tmp.AddRange(bytes);
            }

            byte[] hash = MD5Core.GetHash(tmp.ToArray());
            int keyLength = GetKeyLength(revision, keyBitLength);
            byte[] key = new byte[keyLength];
            Array.Copy(hash, key, key.Length);

            if (revision >= 3)
            {
                for (int i = 0; i < 50; ++i)
                {
                    hash = MD5Core.GetHash(key);
                    Array.Copy(hash, key, key.Length);
                }
            }
            
            return key;
        }

        internal static byte[] CalculateOwnerPassword(string ownerPassword, string userPassword, int revision, int keyBitLength)
        {
            byte[] ownerPasswordBytes = PadOrTruncatePassword(string.IsNullOrEmpty(ownerPassword) ? userPassword : ownerPassword);
            
            byte[] hash = MD5Core.GetHash(ownerPasswordBytes);
            if (revision >= 3)
            {
                for (int i = 0; i < 50; ++i)
                {
                    hash = MD5Core.GetHash(hash);
                }
            }

            int keyLength = GetKeyLength(revision, keyBitLength);
            byte[] key = new byte[keyLength];
            Array.Copy(hash, key, keyLength);

            RC4 rc4 = new RC4();
            rc4.PrepareKey(key);

            byte[] userPasswordBytes = PadOrTruncatePassword(userPassword);
            byte[] oValue = new byte[userPasswordBytes.Length];

            rc4.Encrypt(userPasswordBytes, oValue);
            if (revision >= 3)
            {
                byte[] newKey = new byte[key.Length];
                byte[] newOValue = new byte[oValue.Length];

                for (int i = 0; i < 19; ++i)
                {
                    CreateRC4Key(key, i + 1, newKey);

                    rc4 = new RC4();
                    rc4.PrepareKey(newKey);
                    rc4.Encrypt(oValue, newOValue);

                    Array.Copy(newOValue, oValue, newOValue.Length);
                }
            }

            return oValue;
        }

        internal static byte[] CalculateUserPassword(byte[] key, byte[] id)
        {
            List<byte> tmp = new List<byte>();
            tmp.AddRange(PasswordPadding);
            if (id != null)
            {
                tmp.AddRange(id);
            }
            
            byte[] hash = MD5Core.GetHash(tmp.ToArray());

            RC4 rc4 = new RC4();
            rc4.PrepareKey(key);

            byte[] uValue = new byte[hash.Length];
            rc4.Encrypt(hash, uValue);

            byte[] newKey = new byte[key.Length];
            byte[] newUValue = new byte[uValue.Length];

            for (int i = 0; i < 19; ++i)
            {
                CreateRC4Key(key, i + 1, newKey);

                rc4 = new RC4();
                rc4.PrepareKey(newKey);
                rc4.Encrypt(uValue, newUValue);

                Array.Copy(newUValue, uValue, newUValue.Length);
            }

            byte[] result = new byte[32];
            Array.Copy(uValue, result, uValue.Length);

            return result;
        }

        internal static byte[] Encrypt(StandardEncryptionContext context)
        {
            Guard.ThrowExceptionIfNull(context.InputData, "context.InputData");

            if (context.IsIdentityEncryption)
            {
                return context.InputData;
            }

            byte[] tmp = GetEmptyInitialArray(context);
            Array.Copy(context.EncryptionKey, tmp, context.EncryptionKey.Length);
            byte[] objNoBytes = new byte[3];
            BytesHelper.GetBytes(context.ObjectNumber, objNoBytes);
            byte[] genNoBytes = new byte[2];
            BytesHelper.GetBytes(context.GenerationNumber, genNoBytes);
            Array.Copy(objNoBytes, 0, tmp, context.EncryptionKey.Length, objNoBytes.Length);
            Array.Copy(genNoBytes, 0, tmp, context.EncryptionKey.Length + objNoBytes.Length, genNoBytes.Length);

            if (context.CryptFilterMethod == CryptFilterMethod.AESV2)
            {
                Array.Copy(aesSAlTbytes, 0, tmp, tmp.Length - aesSAlTbytes.Length, aesSAlTbytes.Length);
            }

            tmp = MD5Core.GetHash(tmp);
            int n = Math.Min(16, context.EncryptionKey.Length + 5);

            byte[] key = new byte[n];
            Array.Copy(tmp, key, n);

            byte[] result;

            if (context.CryptFilterMethod == CryptFilterMethod.AESV2)
            {
                result = CalculateAesResult(context.InputData, key);
            }
            else
            {
                result = CalculateRC4Result(context.InputData, key);
            }

            return result;
        }

        internal static StandardEncryptionContext GetEncryptionContext(IStandardCryptFilterMethodProvider stdCF, int objectNo, int generationNo, byte[] data, byte[] encryptionKey, string cryptFilter)
        {
            StandardEncryptionContext encryptionContext = new StandardEncryptionContext()
            {
                ObjectNumber = objectNo,
                GenerationNumber = generationNo,
                InputData = data,
                EncryptionKey = encryptionKey,
                IsIdentityEncryption = cryptFilter == null ? false : cryptFilter == PdfNames.Identity,
            };

            if (cryptFilter == null)
            {
                encryptionContext.CryptFilterMethod = CryptFilterMethod.V2;
            }
            else if (cryptFilter == PdfNames.StandardCryptFilter)
            {
                string cfm = stdCF.GetStandardCryptFilterMethod();

                CryptFilterMethod cryptFilterMethod;
                if (!Enum.TryParse(cfm, false, out cryptFilterMethod))
                {
                    throw new NotSupportedException(string.Format("Not supported crypt filter method: {0}", cfm));
                }

                encryptionContext.CryptFilterMethod = cryptFilterMethod;
            }
            else if (!encryptionContext.IsIdentityEncryption)
            {
                throw new NotSupportedException(string.Format("Only {0} and {1} crypt filter names are supported", PdfNames.Identity, PdfNames.StandardCryptFilter));
            }

            return encryptionContext;
        }

        private static byte[] CalculateAesResult(byte[] cipherTextCombined, byte[] key)
        {
#if SILVERLIGHT
            using (Aes aesAlg = new AesManaged())
#else
            using (Aes aesAlg = Aes.Create())
#endif
            {
                aesAlg.Key = key;

                byte[] IV = new byte[aesAlg.BlockSize / 8];
                byte[] cipherText = new byte[cipherTextCombined.Length - IV.Length];

                Array.Copy(cipherTextCombined, IV, IV.Length);
                Array.Copy(cipherTextCombined, IV.Length, cipherText, 0, cipherText.Length);

                aesAlg.IV = IV;
#if !SILVERLIGHT
                aesAlg.Mode = CipherMode.CBC;
#endif

                // Create a decrytor to perform the stream transform. 
                ICryptoTransform decryptor = aesAlg.CreateDecryptor(aesAlg.Key, aesAlg.IV);

                // Create the streams used for decryption.
                using (var msDecrypt = new MemoryStream(cipherText))
                {
                    using (CryptoStream csDecrypt = new CryptoStream(msDecrypt, decryptor, CryptoStreamMode.Read))
                    {
                        byte[] resFromCryptoStream = csDecrypt.ReadAllBytes();

                        return resFromCryptoStream;
                    }
                }
            }
        }

        private static byte[] CalculateRC4Result(byte[] inputData, byte[] key)
        {
            RC4 rc4 = new RC4();
            rc4.PrepareKey(key);

            byte[] res = new byte[inputData.Length];
            rc4.Encrypt(inputData, res);

            return res;
        }

        private static byte[] GetEmptyInitialArray(StandardEncryptionContext context)
        {
            switch (context.CryptFilterMethod)
            {
                case CryptFilterMethod.V2:
                    return new byte[context.EncryptionKey.Length + 5];
                case CryptFilterMethod.AESV2:
                    return new byte[context.EncryptionKey.Length + 9]; // 4 extra sAlT bytes. 
                default:
                    throw new NotSupportedException(string.Format("Not supported crypt filter method: {0}", context.CryptFilterMethod));
            }
        }

        private static int GetPermissionBits()
        {
            IEnumerable<int> reservedBitThatAreOnes = GetReservedBitThatAreOnes();
            FlagWriter<PermissionFlag> flagWriter = new FlagWriter<PermissionFlag>(reservedBitThatAreOnes);

            //// TODO: Use PermissionFlag enum to write specific permissions with the writer.

            return flagWriter.ResultFlags;
        }

        private static IEnumerable<int> GetReservedBitThatAreOnes()
        {
            yield return 7;
            yield return 8;

            for (int number = 13; number <= 32; number++)
            {
                yield return number;
            }
        }

        private static void CreateRC4Key(byte[] key, int step, byte[] result)
        {
            for (int i = 0; i < key.Length; i++)
            {
                result[i] = (byte)(key[i] ^ (byte)step);
            }
        }

        private static int GetKeyLength(int revision, int keyBitLength)
        {
            return revision == 2 ? 5 : (keyBitLength / 8);
        }

        string IStandardCryptFilterMethodProvider.GetStandardCryptFilterMethod()
        {
            Guard.ThrowExceptionIfNull(this.CryptFilters, "CryptFilters");
            Guard.ThrowExceptionIfNull(this.CryptFilters.StandardCryptFilter, "CryptFilters.StandardCryptFilter");
            string cfm = this.CryptFilters.StandardCryptFilter.CryptFilterMethod.Value;

            return cfm;
        }
    }
}
