﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types
{
    internal class CrossReferenceCollection
    {
        private readonly Dictionary<int, CrossReferenceEntry> crossReferenceEntries;

        public CrossReferenceCollection()
        {
            this.crossReferenceEntries = new Dictionary<int, CrossReferenceEntry>();
            this.MaxObjectNumber = int.MinValue;
        }

        public int Count
        {
            get
            {
                return this.crossReferenceEntries.Count;
            }
        }

        public int MaxObjectNumber { get; private set; }

        internal IEnumerable<KeyValuePair<int, CrossReferenceEntry>> Entries
        {
            get
            {
                return this.crossReferenceEntries;
            }
        }

        public void AddCrossReferenceEntry(int objectNumber, CrossReferenceEntry entry)
        {
            Guard.ThrowExceptionIfNull(entry, "entry");

            if (!this.crossReferenceEntries.ContainsKey(objectNumber))
            {
                this.crossReferenceEntries[objectNumber] = entry;
                this.MaxObjectNumber = Math.Max(this.MaxObjectNumber, objectNumber);
            }
        }

        public CrossReferenceEntry GetCrossReferenceEntry(int objectNumber)
        {
            return this.crossReferenceEntries[objectNumber];
        }

        public void UpdateCrossReferenceEntry(int objectNumber, long field1, int field2, CrossReferenceEntryType type)
        {
            CrossReferenceEntry entry = this.crossReferenceEntries[objectNumber];
            entry.Field1 = field1;
            entry.Field2 = field2;
            entry.Type = type;
        }

        public bool TryGetCrossReferenceEntry(IndirectReference reference, out CrossReferenceEntry entry)
        {
            return this.crossReferenceEntries.TryGetValue(reference.ObjectNumber, out entry);
        }

        public void Read(PostScriptReader reader, IPdfImportContext context)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            bool isCrossReferenceTable = this.PeekCrossReferenceTableMarker(reader, context);

            if (isCrossReferenceTable)
            {
                reader.Reader.ReadLine();
                this.ReadCrossReferenceTable(reader, context);
            }
            else
            {
                this.ReadCrossReferenceStream(reader, context);
            }
        }

        internal bool PeekCrossReferenceTableMarker(PostScriptReader reader, IPdfImportContext context)
        {
            using (reader.Reader.BeginReadingBlock())
            {
                string token = reader.Reader.ReadLine().Trim();
                bool hasFoundCrossReferenceTableMarker = token == PdfNames.CrossReferenceTableMarker;

                return hasFoundCrossReferenceTableMarker;
            }
        }

        private void ReadCrossReferenceTable(PostScriptReader reader, IPdfImportContext context)
        {
            bool readTableSection = true;
            string[] fields = reader.Reader.ReadSplitLine();
            int first = int.Parse(fields[0]);
            int count = int.Parse(fields[1]);

            while (readTableSection)
            {
                for (int i = 0; i < count; i++)
                {
                    CrossReferenceEntry entry = new CrossReferenceEntry();
                    entry.Read(reader);

                    this.AddCrossReferenceEntry(first + i, entry);
                }

                // This handles the case when the trailer object is written on a single row.
                using (reader.Reader.BeginReadingBlock())
                {
                    fields = reader.Reader.ReadSplitLine();
                    readTableSection = fields.Length >= 2 && int.TryParse(fields[0], out first) && int.TryParse(fields[1], out count);
                }

                if (readTableSection)
                {
                    reader.Reader.ReadLine();
                }
            }

            Trailer trailer = new Trailer();
            trailer.Read(reader, context);

            this.InterpretTrailer(trailer, reader, context);
        }

        private void ReadCrossReferenceStream(PostScriptReader reader, IPdfImportContext context)
        {
            IndirectObject indirectObject = reader.Read<IndirectObject>(context, PdfElementType.IndirectObject);
            CrossReferenceStream crossReferenceStream = new CrossReferenceStream();
            crossReferenceStream.Load(reader, context, indirectObject.Content);

            this.InterpretTrailer(crossReferenceStream, reader, context);
        }

        private void InterpretTrailer(Trailer trailer, PostScriptReader reader, IPdfImportContext context)
        {
            trailer.CopyPropertiesTo(reader, context);

            if (trailer.XRefStm != null)
            {
                reader.Reader.Seek(trailer.XRefStm.Value, SeekOrigin.Begin);
                this.Read(reader, context);
            }

            if (trailer.Prev != null)
            {
                reader.Reader.Seek(trailer.Prev.Value, SeekOrigin.Begin);
                this.Read(reader, context);
            }
        }
    }
}
