﻿using System;
using System.Text.RegularExpressions;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types
{
    internal class PdfLiteralString : PdfString
    {
        private readonly StringType stringType;

        public PdfLiteralString()
        {
            this.stringType = StringType.Byte;
        }

        public PdfLiteralString(byte[] initialValue)
            : base(initialValue)
        {
            bool isUnicode = PdfString.IsUnicode(this.Value);

            this.stringType = isUnicode ? StringType.UTFEncodedTextString : StringType.Byte;
        }

        public PdfLiteralString(string initialValue, StringType stringType)
            : base(initialValue, stringType)
        {
            this.stringType = stringType;
        }

        public StringType StringType
        {
            get
            {
                return this.stringType;
            }
        }

        public override StringObjectType StringObjectType
        {
            get { return StringObjectType.Literal; }
        }

        public override string ToString()
        {
            if (this.Value == null)
            {
                return string.Empty;
            }

            var text = GetString(this.Value, this.StringType);
            return text;
        }

        public string ToStringAs(StringType stringType)
        {
            var text = GetString(this.Value, stringType);
            return text;
        }

        public override void Write(Export.PdfWriter writer, Export.IPdfExportContext context)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(context, "context");

            var encryptedValue = context.EncryptString(this.Value);
            var text = GetString(encryptedValue, this.StringType);
            var escapedText = EscapeSpecialCharacters(text);

            writer.Write(string.Format("{0}{1}{2}", Characters.LiteralStringStart, escapedText, Characters.LiteralStringEnd));
        }

        private static string GetString(byte[] bytes, StringType stringType)
        {
            switch (stringType)
            {
                case StringType.Byte:
                    return PdfString.GetPdfEncodedTextString(bytes);

                case StringType.ASCII:
                    return System.Text.Encoding.UTF8.GetString(bytes, 0, bytes.Length);

                case StringType.UTFEncodedTextString:
                    return PdfString.GetUtfEncodedTextString(bytes);

                default:
                    throw new NotImplementedException(string.Format("{0} is not supported.", stringType.ToString()));

            }
        }

        private static string EscapeSpecialCharacters(string text)
        {
            // Escape all backslashes and brackets with a backslash.
            return Regex.Replace(text, @"\\|\)|\(", EscapeSpecialCharacterMatch);
        }

        private static string EscapeSpecialCharacterMatch(Match m)
        {
            switch (m.Value)
            {
                case @"\":
                    return @"\\";
                case @")":
                    return @"\)";
                case @"(":
                    return @"\(";
                default:
                    return m.Value;
            }
        }

    }
}
