﻿using System.IO;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Filters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser.Keywords;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Converters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types
{
    internal abstract class PdfStreamBase : PdfPrimitive
    {
        private readonly PdfDictionary dictionary;
        private readonly PdfInt length;
        private readonly long offset;
        private readonly IPdfImportContext context;
        private readonly Stream stream;
        private byte[] rawPdfData;
        private byte[] decodedPdfData;
        private readonly IndirectReference currentIndirectReference;
        private IPdfPrimitive filters;
        private readonly PostScriptReader reader;
        private DecodeParameters[] decodeParameters = null;

        public PdfStreamBase(IPdfImportContext context, Stream stream, PdfDictionary dictionary, PdfInt length, long offset)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(stream, "stream");
            Guard.ThrowExceptionIfNull(dictionary, "dictionary");

            this.context = context;
            this.currentIndirectReference = context.CurrentIndirectReference;

            this.stream = stream;
            this.dictionary = dictionary;
            this.length = length;
            this.offset = offset;
            this.reader = new PostScriptReader(this.stream, new KeywordCollection());
        }

        public IPdfPrimitive Filters
        {
            get
            {
                if (this.filters == null)
                {
                    dictionary.TryGetElement(this.reader, this.context, PdfNames.Filters, out this.filters);
                }

                return this.filters;
            }
        }

        public DecodeParameters[] DecodeParameters
        {
            get
            {
                if (this.decodeParameters == null)
                {
                    this.decodeParameters = PdfStreamBase.ReadDecodeParameters(this.reader, this.context, this.dictionary);
                }

                return this.decodeParameters;
            }
        }

        public override PdfElementType Type
        {
            get
            {
                return PdfElementType.Stream;
            }
        }

        public PdfDictionary Dictionary
        {
            get
            {
                return this.dictionary;
            }
        }

        protected IndirectReference CurrentIndirectReference
        {
            get
            {
                return this.currentIndirectReference;
            }
        }

        protected PostScriptReader Reader
        {
            get
            {
                return this.reader;
            }
        }

        protected IPdfImportContext Context
        {
            get
            {
                return this.context;
            }
        }

        internal PdfInt Length
        {
            get
            {
                return this.length;
            }
        }

        internal Stream Stream
        {
            get
            {
                return this.stream;
            }
        }

        internal long Offset
        {
            get
            {
                return this.offset;
            }
        }

        public byte[] ReadRawPdfData()
        {
            if (this.rawPdfData == null)
            {
                this.rawPdfData = this.ReadRawPdfDataOverride();
            }

            return this.rawPdfData;
        }

        public byte[] ReadDecodedPdfData()
        {
            if (this.decodedPdfData == null)
            {
                this.decodedPdfData = this.DecodePdfData();
            }

            return this.decodedPdfData;
        }

        protected abstract byte[] ReadRawPdfDataOverride();

        protected abstract int CalculateLength(Reader reader);

        public override void Write(PdfWriter writer, IPdfExportContext context)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(context, "context");

            this.dictionary.Write(writer, context);
            byte[] rawData = this.ReadRawPdfData();

            PdfStreamObjectBase.WriteEncodedData(writer, context, rawData);
        }

        internal byte[] CalculateStreamLengthAndReadData()
        {
            this.reader.Reader.Seek(this.offset, SeekOrigin.Begin);
            int calculateLength = this.CalculateLength(this.reader.Reader);
            this.reader.Reader.Seek(this.offset, SeekOrigin.Begin);
            byte[] data = this.reader.Reader.Read(calculateLength);

            return data;
        }

        private byte[] DecodePdfData()
        {
            PdfArray filters = this.Filters as PdfArray;

            if (filters == null)
            {
                PdfName filter = this.Filters as PdfName;
                if (filter != null)
                {
                    filters = new PdfArray(new PdfName[] { filter });
                }
                else
                {
                    filters = new PdfArray();
                }
            }

            byte[] data = FiltersManager.Decode(reader, context, dictionary, this, filters);

            return data;
        }

        private static DecodeParameters[] ReadDecodeParameters(PostScriptReader reader, IPdfImportContext context, PdfDictionary dictionary)
        {
            DecodeParameters[] decodeParameters = null;

            IPdfPrimitive parms;
            if (dictionary.TryGetElement(PdfNames.DecodeParameters, out parms))
            {
                decodeParameters = DecodeParametersConverter.Convert(reader, context, parms);
            }

            return decodeParameters;
        }
    }
}
