﻿using System.Collections.Generic;
using System.IO;
using Telerik.Windows.Documents.Common.FormatProviders;
#if !NETSTANDARD
using System;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader.Parsers;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Text;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf
{
    /// <summary>
    /// Represents PDF format provider.
    /// </summary>
    public partial class PdfFormatProvider : BinaryFormatProviderBase<RadFixedDocument>
    {
        private static readonly string[] supportedExtensions = new string[] { ".pdf" };

        /// <summary>
        /// Initializes a new instance of the <see cref="PdfFormatProvider" /> class. This overload should be used in order to import a document with the RadPdfProcessing library.
        /// </summary>
        public PdfFormatProvider()
        {
            this.ImportSettings = new PdfImportSettings();
            this.ExportSettings = new PdfExportSettings();
        }

        /// <summary>
        /// Gets the supported extensions.
        /// </summary>
        /// <value>The supported extensions.</value>
        public override IEnumerable<string> SupportedExtensions
        {
            get
            {
                return supportedExtensions;
            }
        }

        /// <summary>
        /// Gets or sets the import settings.
        /// </summary>
        /// <value>The import settings.</value>
        public PdfImportSettings ImportSettings { get; set; }

        /// <summary>
        /// Gets or sets the export settings.
        /// </summary>
        /// <value>The export settings.</value>
        public PdfExportSettings ExportSettings { get; set; }

        /// <summary>
        /// Gets a value indicating whether can import.
        /// </summary>
        /// <value>The value indicating whether can import.</value>
        public override bool CanImport
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Gets a value indicating whether can export.
        /// </summary>
        /// <value>The value indicating whether can export.</value>
        public override bool CanExport
        {
            get
            {
                return true;
            }
        }

        internal override bool ShouldDisposeMemoryStreamOnImport
        {
            get
            {
                return this.ImportSettings.ReadingMode == ReadingMode.AllAtOnce;
            }
        }

        /// <summary>
        /// Imports the specified input.
        /// </summary>
        /// <param name="input">The input.</param>
        /// <returns>The imported document.</returns>
        protected override RadFixedDocument ImportOverride(Stream input)
        {
            Guard.ThrowExceptionIfNull(input, "input");

#if !NETSTANDARD
            if (this.isOldPdfViewer)
            {
                throw new NotSupportedException("This method is only supported when initializing PdfFormatProvider through the PdfFormatProvider() constructor.");
            }

#if DEBUG
            PdfContentExtractor.ClearExtractedContent();
#endif
#endif
            PdfImporter importer = new PdfImporter();
            RadFixedDocumentImportContext context = new RadFixedDocumentImportContext(this.ImportSettings);
            importer.Import(input, context);

            RadFixedDocument document = context.Document;

            if (this.ImportSettings.ReadingMode == ReadingMode.OnDemand)
            {
                PagesContentCache pagesContentCache = new PagesContentCache();

                PageContentManager pageContentLoader = new PageContentManager(context);
                ThreadSafePageContentManager threadSafeContentManager = new ThreadSafePageContentManager(pageContentLoader);

                document.CacheManager = new LoadOnDemandPagesCacheManager(pagesContentCache, threadSafeContentManager);
            }
            else
            {
                document.CacheManager = new AllAtOncePagesCacheManager();
            }

            if (this.ImportSettings.IsPdfViewer)
            {
                this.ApplyPdfViewerRelatedProperties(input, document);
            }

            return document;
        }

        /// <summary>
        /// Exports the specified document.
        /// </summary>
        /// <param name="document"></param>
        /// <param name="output">The output.</param>
        protected override void ExportOverride(RadFixedDocument document, Stream output)
        {
            Guard.ThrowExceptionIfNull(document, "document");
            Guard.ThrowExceptionIfNull(output, "output");

#if !NETSTANDARD
            if (this.isOldPdfViewer)
            {
                throw new NotSupportedException("This method is only supported when initializing PdfFormatProvider through the PdfFormatProvider() constructor.");
            }
#endif

            using (new RadFixedDocumentLicenseCheck(document))
            {
                PdfExporter exporter = new PdfExporter();
                exporter.Export(new RadFixedDocumentExportContext(document, this.ExportSettings), output);
            }
        }

        private void ApplyPdfViewerRelatedProperties(Stream input, RadFixedDocument document)
        {
            TextDocument textDocument = new TextDocument(document);
            document.TextDocument = textDocument;
            document.SupportsSelection = true;
        }
    }
}
