﻿using System;
using System.IO;
using System.Windows;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Streaming.IncrementalUpdate.InteractiveForms;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Streaming.IncrementalUpdate
{
    internal sealed class PdfIncrementalStreamWriter : IDisposable
    {
        private readonly PdfIncrementalStreamExportContext context;
        private readonly PdfFileSource fileSource;
        private readonly bool leaveStreamOpen;
        private readonly DisposeValidator disposeValidator;

        public PdfIncrementalStreamWriter(PdfFileSource initialFile, Stream resultStream)
            : this(initialFile, resultStream, false)
        {
        }

        public PdfIncrementalStreamWriter(PdfFileSource initialFile, Stream resultStream, bool leaveStreamOpen)
        {
            this.fileSource = initialFile;
            this.leaveStreamOpen = leaveStreamOpen;
            this.disposeValidator = new DisposeValidator();
            this.context = new PdfIncrementalStreamExportContext(resultStream, initialFile.Context.CrossReferences.MaxObjectNumber);

            this.WriteFileBeginning();
        }

        public PagePropertiesEditor EditPageProperties(int pageIndex)
        {
            PdfPageSource page = this.fileSource.Pages[pageIndex];

            return new PagePropertiesEditor(this.context, page);
        }

        public CheckBoxPropertiesEditor EditCheckBoxProperties(string fieldName)
        {
            PdfFormFieldSource field = this.fileSource.FormFields[fieldName];
            CheckBoxPropertiesEditor propertiesEditor = new CheckBoxPropertiesEditor(this.context, field);

            return propertiesEditor;
        }

        public RadioButtonPropertiesEditor EditRadioButtonProperties(string fieldName)
        {
            PdfFormFieldSource field = this.fileSource.FormFields[fieldName];
            RadioButtonPropertiesEditor propertiesEditor = new RadioButtonPropertiesEditor(this.context, field);

            return propertiesEditor;
        }

        public ComboBoxPropertiesEditor EditComboBoxProperties(string fieldName)
        {
            PdfFormFieldSource field = this.fileSource.FormFields[fieldName];
            ComboBoxPropertiesEditor propertiesEditor = new ComboBoxPropertiesEditor(this.context, field);

            return propertiesEditor;
        }

        public ListBoxPropertiesEditor EditListBoxProperties(string fieldName)
        {
            PdfFormFieldSource field = this.fileSource.FormFields[fieldName];
            ListBoxPropertiesEditor propertiesEditor = new ListBoxPropertiesEditor(this.context, field);

            return propertiesEditor;
        }

        public TextFieldPropertiesEditor EditTextFieldProperties(string fieldName)
        {
            PdfFormFieldSource field = this.fileSource.FormFields[fieldName];
            TextFieldPropertiesEditor propertiesEditor = new TextFieldPropertiesEditor(this.context, field);

            return propertiesEditor;
        }

#if WPF
        public SignaturePropertiesEditor EditSignatureProperties(string fieldName)
        {
            PdfFormFieldSource field = this.fileSource.FormFields[fieldName];
            SignaturePropertiesEditor propertiesEditor = new SignaturePropertiesEditor(this.context, field);

            return propertiesEditor;
        }
#endif

        public void Dispose()
        {
            try
            {
                this.disposeValidator.Dispose();
                this.WriteFileEnding();
            }
            finally
            {
                if (!this.leaveStreamOpen)
                {
                    this.context.Writer.OutputStream.Dispose();
                }
            }
        }

        private void WriteFileBeginning()
        {
            Stream sourceStream = this.fileSource.Context.Reader.Reader.Stream;
            sourceStream.Seek(0, SeekOrigin.Begin);
#if WPF
            context.SignatureExportInfo.AddDocumentStartPosition((int)this.context.Writer.Position);
#endif
            sourceStream.CopyTo(this.context.Writer.OutputStream);
        }

        private void WriteFileEnding()
        {
            PdfExporter.WritePendingIndirectObjects(this.context, this.context.Writer);

            DocumentCatalog catalog = this.fileSource.Context.Root.GetValue();
            IndirectReference rootReference = this.fileSource.Context.Root.Reference;
            this.context.RegisterIndirectReference(catalog, rootReference.ObjectNumber, false);
            PdfExporter.WriteFontsFromContext(this.context.Writer, this.context);

            long crossReferenceBeginingOffset = this.context.Writer.Position;
            PdfInt previousOffset = new PdfInt(this.fileSource.Context.StartXRefOffset);
            bool previousStartsWithTable = this.fileSource.Context.StartsWithCrossReferenceTable;

            if (previousStartsWithTable)
            {
                PdfExporter.WriteTrailerAndCrossReferenceTable(this.context.Writer, this.context, catalog, previousOffset);
            }
            else
            {
                PdfExporter.WriteCrossReferenceStream(this.context.Writer, this.context, catalog, previousOffset);
            }

            PdfExporter.WriteDocumentEnd(crossReferenceBeginingOffset, this.context.Writer, this.context);
#if WPF
            this.context.SignatureExportInfo.AddDocumentEndPosition((int)this.context.Writer.Position - 1);

            this.UpdateSignatures(this.context.Writer, this.context);
#endif
        }
#if WPF
        private void UpdateSignatures(PdfWriter writer, PdfIncrementalStreamExportContext context)
        {
            if (context.SignatureToUpdate != null)
            {
                PdfExporter.UpdateSignatureByteRange(context.SignatureToUpdate.Properties.FieldName, writer, context);
                PdfExporter.UpdateSignatureContent(context.SignatureToUpdate, writer, context);
            }
        }
#endif
    }
}
