﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Streaming
{
    /// <summary>
    /// Represent source PDF file.
    /// </summary>
    public sealed class PdfFileSource : IDisposable
    {
        private readonly Stream fileStream;
        private readonly bool leaveStreamOpen;
        private PdfPageSource[] pages;
        private PdfFormFieldSourceCollection formFields;
        private readonly PdfSourceImportContext context;
        private readonly HashSet<IndirectReference> pageTreeReferences;
        private readonly DisposeValidator disposeValidator;

        /// <summary>
        /// Initializes a new instance of the <see cref="PdfFileSource" /> class.
        /// </summary>
        /// <param name="pdfFileStream">The pdf file stream.</param>
        public PdfFileSource(Stream pdfFileStream)
            : this(pdfFileStream, false)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="PdfFileSource" /> class.
        /// </summary>
        /// <param name="pdfFileStream">The pdf file stream.</param>
        /// <param name="leaveStreamOpen">Boolean value indicating whether or not to close the stream when <see cref="PdfFileSource" /> instance is disposed. The default value is false.</param>
        public PdfFileSource(Stream pdfFileStream, bool leaveStreamOpen)
            : this(pdfFileStream, new PdfImportSettings(), leaveStreamOpen)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="PdfFileSource" /> class.
        /// </summary>
        /// <param name="pdfFileStream">The PDF file <see cref="Stream"/> instance.</param>
        /// <param name="importSettings">The <see cref="PdfImportSettings" /> instance specifying settings for PDF file import.</param>
        /// <param name="leaveStreamOpen">Boolean value indicating whether or not to close the stream when <see cref="PdfFileSource" /> instance is disposed. The default value is false.</param>
        public PdfFileSource(Stream pdfFileStream, PdfImportSettings importSettings, bool leaveStreamOpen)
        {
            Guard.ThrowExceptionIfNull(pdfFileStream, "pdfFileStream");
            Guard.ThrowExceptionIfNull(importSettings, "importSettings");

            this.fileStream = pdfFileStream;
            this.leaveStreamOpen = leaveStreamOpen;
            this.context = new PdfSourceImportContext(importSettings);
            this.pageTreeReferences = new HashSet<IndirectReference>();
            this.disposeValidator = new DisposeValidator();
            this.Context.BeginImport(pdfFileStream);
        }

        /// <summary>
        /// Gets the pages.
        /// </summary>
        /// <value>The pages.</value>
        public PdfPageSource[] Pages
        {
            get
            {
                if(this.pages == null)
                {
                    this.pages = this.LoadPages();
                }

                return this.pages;
            }
        }

        internal PdfFormFieldSourceCollection FormFields
        {
            get
            {
                if(this.formFields == null)
                {
                    this.formFields = new PdfFormFieldSourceCollection(this);
                }

                return this.formFields;
            }
        }

        internal PdfSourceImportContext Context
        {
            get
            {
                return this.context;
            }
        }

        internal bool IsDisposed
        {
            get
            {
                return this.disposeValidator.IsDisposed;
            }
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing,
        /// or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            try
            {
                this.disposeValidator.Dispose();
                this.OnDisposed();
            }
            finally
            {
                if (!this.leaveStreamOpen)
                {
                    this.fileStream.Dispose();
                }
            }
        }

        internal event EventHandler Disposed;

        internal bool IsPageTreeReference(IndirectReference reference)
        {
            return this.pageTreeReferences.Contains(reference);
        }

        private PdfPageSource[] LoadPages()
        {
            DocumentCatalog catalog = this.Context.Root.GetValue();
            IEnumerable<PdfPageSource> pageSources = this.EnumeratePageSources(catalog.Pages);

            return pageSources.ToArray();
        }

        private IEnumerable<PdfPageSource> EnumeratePageSources(PageTreeNode pageNode)
        {
            for (int i = 0; i < pageNode.Children.Count; i++)
            {
                PageTreeNode pageTreeNode;
                if (pageNode.Children.TryGetElement(this.context.Reader, this.Context, i, out pageTreeNode))
                {
                    if (pageTreeNode.PageTreeNodeType == PageTreeNodeType.Page)
                    {
                        Page page = (Page)pageTreeNode;
                        IndirectReference reference = (IndirectReference)pageNode.Children[i];
                        this.pageTreeReferences.Add(reference);
                        this.Context.Pages.Add(page);
                        
                        yield return new PdfPageSource(this, reference, page);
                    }
                    else if (pageTreeNode.PageTreeNodeType == PageTreeNodeType.PageTreeNode)
                    {
                        foreach (PdfPageSource pageSource in this.EnumeratePageSources(pageTreeNode))
                        {
                            yield return pageSource;
                        }
                    }
                }
            }
        }        

        private void OnDisposed()
        {
            if (this.Disposed != null)
            {
                this.Disposed(this, EventArgs.Empty);
            }
        }
    }
}
