﻿using System;
using System.Collections.Generic;
using System.IO;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Metadata;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Streaming
{
    internal class PdfFileStreamExportContext : PdfExportContext, IResourceRenamingExportContext
    {
        private readonly PdfWriter writer;
        private readonly Dictionary<PdfFileSource, Dictionary<IndirectReference, IndirectReference>> sourceReferencesToContextReferencesMapping;
        private readonly Dictionary<PdfPageSource, IndirectReference> pageSourceToXFormReference;
        private readonly InstanceIdCache<RadFixedPage, IndirectReference> pageRootToXFormReference;
        private readonly PageTreeNode pagesRoot;
        private readonly IndirectObject pagesRootObject;
        private readonly RadFixedPage licenceMessageContentOwner;
        private bool hasFinalizedDocumentRoot;

        public PdfFileStreamExportContext(Stream stream, RadFixedDocumentInfo documentInfo, PdfExportSettings exportSettings) 
            : base(documentInfo, exportSettings)
        {
            this.pagesRoot = new PageTreeNode();
            this.pagesRoot.Children = new PdfArray();
            this.pagesRootObject = this.CreateIndirectObject(this.pagesRoot, false);
            this.writer = new PdfWriter(stream);
            this.sourceReferencesToContextReferencesMapping = new Dictionary<PdfFileSource, Dictionary<IndirectReference, IndirectReference>>();
            this.pageSourceToXFormReference = new Dictionary<PdfPageSource, IndirectReference>();
            this.pageRootToXFormReference = new InstanceIdCache<RadFixedPage, IndirectReference>();

            string message;
            if (RadFixedDocumentLicenseCheck.TryGetLicenseMessage(out message))
            {
                this.licenceMessageContentOwner = new RadFixedPage();
                RadFixedDocumentLicenseCheck.AddLicenceMessageContent(this.licenceMessageContentOwner, message);
            }

            this.hasFinalizedDocumentRoot = false;
        }

        public PdfWriter Writer
        {
            get
            {
                return this.writer;
            }
        }

        public RadFixedPage LicenceMessageContentOwner
        {
            get
            {
                return this.licenceMessageContentOwner;
            }
        }

        public IndirectReference AddNextPageReferenceToGetReference(Page page)
        {
            page.Parent = this.pagesRoot;
            this.pagesRoot.Children.Add(page);
            IndirectObject pageObject = this.CreateIndirectObject(page, false);

            return pageObject.Reference;
        }

        public DocumentCatalog GetFinalizedDocumentRoot()
        {
            Guard.ThrowExceptionIfTrue(this.hasFinalizedDocumentRoot, "hasFinalizedDocumentRoot");

            DocumentCatalog root = new DocumentCatalog();
            root.DocumentInfo = new DocumentInfo();
            root.DocumentInfo.CopyPropertiesFrom(this);
            this.pagesRoot.Count = new PdfInt(this.pagesRoot.Children.Count);
            root.Pages = this.pagesRoot;

            this.IndirectObjectsQueue.Enqueue(this.pagesRootObject);

            return root;
        }

        public void AddXFormReference(PdfPageSource page, IndirectReference reference)
        {
            this.pageSourceToXFormReference.Add(page, reference);
        }

        public bool TryGetXFormReference(PdfPageSource page, out IndirectReference xFormReference)
        {
            return this.pageSourceToXFormReference.TryGetValue(page, out xFormReference);
        }

        public void AddXFormReference(RadFixedPage pageRoot, IndirectReference reference)
        {
            this.pageRootToXFormReference.Add(pageRoot, reference);
        }

        public bool TryGetXFormReference(RadFixedPage pageRoot, out IndirectReference xFormReference)
        {
            return this.pageRootToXFormReference.TryGetValue(pageRoot, out xFormReference);
        }

        public void AddPageSourceToContextReferenceMapping(PdfPageSource pageSource, IndirectReference contextPageReference)
        {
            Dictionary<IndirectReference, IndirectReference> referencesMapping = this.GetReferencesMapping(pageSource.FileSource);

            referencesMapping[pageSource.PageReference] = contextPageReference;
        }

        public void AddAnnotationDictionariesMappings(PdfPageSource pageSource, IndirectReference sourceAnnotationReference, IndirectReference contextAnnotationReference)
        {
            Dictionary<IndirectReference, IndirectReference> referencesMapping = this.GetReferencesMapping(pageSource.FileSource);
            
            referencesMapping[sourceAnnotationReference] = contextAnnotationReference;
        }

        public void AddSourceToContextReferenceMapping(PdfFileSource source, IndirectReference sourceReference, IndirectReference contextReference)
        {
            Dictionary<IndirectReference, IndirectReference> referencesMapping = this.GetReferencesMapping(source);

            referencesMapping.Add(sourceReference, contextReference);
        }

        public bool TryGetContextIndirectReference(PdfFileSource source, IndirectReference sourceReference, out IndirectReference contextReference)
        {
            Dictionary<IndirectReference, IndirectReference> referencesMapping = this.GetReferencesMapping(source);

            return referencesMapping.TryGetValue(sourceReference, out contextReference);
        }

        private Dictionary<IndirectReference, IndirectReference> GetReferencesMapping(PdfFileSource source)
        {
            Guard.ThrowExceptionIfTrue(source.IsDisposed, "source.IsDisposed");

            Dictionary<IndirectReference, IndirectReference> sourceReferenceToContextReference;
            if (!this.sourceReferencesToContextReferencesMapping.TryGetValue(source, out sourceReferenceToContextReference))
            {
                sourceReferenceToContextReference = new Dictionary<IndirectReference, IndirectReference>();
                this.sourceReferencesToContextReferencesMapping.Add(source, sourceReferenceToContextReference);
                source.Disposed += this.PdfFileSourceDisposed;
            }

            return sourceReferenceToContextReference;
        }

        private void PdfFileSourceDisposed(object sender, EventArgs e)
        {
            PdfFileSource fileSource = (PdfFileSource)sender;
            fileSource.Disposed -= this.PdfFileSourceDisposed;
            this.ClearCache(fileSource);
        }

        private void ClearCache(PdfFileSource fileSource)
        {
            this.sourceReferencesToContextReferencesMapping.Remove(fileSource);
            
            foreach(PdfPageSource pageSource in fileSource.Pages)
            {
                this.pageSourceToXFormReference.Remove(pageSource);
            }
        }
    }
}
