﻿using System;
using System.Collections;
using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Converters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Forms;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Streaming
{
    internal class PdfFormFieldSourceCollection : IEnumerable<PdfFormFieldSource>
    {
        private readonly Dictionary<string, PdfFormFieldSource> fields;

        public PdfFormFieldSourceCollection(PdfFileSource fileSource)
        {
            this.fields = new Dictionary<string, PdfFormFieldSource>();

            foreach(PdfFormFieldSource fieldSource in EnumerateFieldSources(fileSource))
            {
                this.fields.Add(fieldSource.FieldName, fieldSource);
            }
        }        

        /// <summary>
        /// Gets PdfFormFieldSource instance by field name.
        /// </summary>
        /// <param name="fieldName">The field name.</param>
        /// <returns></returns>
        public PdfFormFieldSource this[string fieldName]
        {
            get
            {
                return this.fields[fieldName];
            }
        }

        /// <summary>
        /// Gets the count of the fields in the collection.
        /// </summary>
        public int Count
        {
            get
            {
                return this.fields.Count;
            }
        }

        public IEnumerator<PdfFormFieldSource> GetEnumerator()
        {
            return this.fields.Values.GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.fields.Values.GetEnumerator();
        }

        private static IEnumerable<PdfFormFieldSource> EnumerateFieldSources(PdfFileSource fileSource)
        {
            string[] propertyNames = { DocumentCatalog.AcroFormName, AcroFormObject.FieldsName };

            PdfArray fieldsArray;
            if (fileSource.Context.TryGetNestedPrimitive(fileSource.Context.Root.Reference, propertyNames, out fieldsArray))
            {
                foreach (PdfFormFieldSource fieldSource in EnumerateFieldSources(fileSource, fieldsArray, null, string.Empty))
                {
                    yield return fieldSource;
                }
            }
        }

        private static IEnumerable<PdfFormFieldSource> EnumerateFieldSources(PdfFileSource fileSource, PdfArray currentLevelNodes, IndirectObject parentObject, string parentName)
        {
            bool addParentAsTerminalNode = false;

            foreach (IPdfPrimitive item in currentLevelNodes)
            {
                if (item.Type == PdfElementType.IndirectReference)
                {
                    IndirectReference childReference = (IndirectReference)item;
                    IndirectObject childObject = fileSource.Context.ReadIndirectObject(childReference);

                    if(childObject.Content.Type == PdfElementType.Dictionary)
                    {
                        PdfDictionary dictionary = (PdfDictionary)childObject.Content;

                        if (FieldsConverter.IsWidget(fileSource.Context.Reader, fileSource.Context, dictionary))
                        {
                            if (FieldsConverter.ContainsMergedFieldProperties(dictionary))
                            {
                                string fieldName = GetFieldName(dictionary, parentName);
                                yield return new PdfFormFieldSource(fileSource, childObject, fieldName);
                            }
                            else
                            {
                                addParentAsTerminalNode = true;
                            }
                        }
                        else
                        {
                            string fieldName = GetFieldName(dictionary, parentName);

                            PdfArray kids;
                            if(dictionary.TryGetElement<PdfArray>(fileSource.Context.Reader, fileSource.Context, FormFieldNode.KidsName, out kids) && kids.Count > 0)
                            {
                                foreach (PdfFormFieldSource fieldSource in EnumerateFieldSources(fileSource, kids, childObject, fieldName))
                                {
                                    yield return fieldSource;
                                }
                            }
                            else
                            {
                                yield return new PdfFormFieldSource(fileSource, childObject, fieldName);
                            }
                        }
                    }
                }
            }

            if (addParentAsTerminalNode)
            {
                yield return new PdfFormFieldSource(fileSource, parentObject, parentName);
            }
        }

        private static string GetFieldName(PdfDictionary fieldNodeDictionary, string parentName)
        {
            if (fieldNodeDictionary.ContainsKey(FormFieldNode.TName))
            {
                PdfString partialName = fieldNodeDictionary[FormFieldNode.TName] as PdfString;

                if(partialName != null)
                {
                    string partialNameValue = partialName.ToString();

                    if(!string.IsNullOrEmpty(partialNameValue))
                    {
                        if(string.IsNullOrEmpty(parentName))
                        {
                            return partialNameValue;
                        }
                        else
                        {
                            return string.Format("{0}.{1}", parentName, partialNameValue);
                        }
                    }
                }
            }

            return parentName;
        }
    }
}
