﻿using System;
using System.Collections.Generic;
using System.IO;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Annotations;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Objects;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Operators;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Utilities.Rendering;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Streaming
{
    /// <summary>
    /// Represent a PDF page stream that writes page content directly to a given stream on Dispose method call.
    /// </summary>
    public sealed class PdfPageStreamWriter : IDisposable
    {
        private readonly PdfPageStreamWriterContext pageContext;
        private readonly PdfDictionary xObjectResources;
        private readonly Dictionary<IndirectReference, string> xFormResourceIds;
        private readonly PdfArray annotations;
        private readonly PdfWriter contentWriter;
        private readonly MemoryStream contentStream;
        private readonly PositionStack positionStack;
        private readonly DisposeValidator disposeValidator;
        private int currentResourceId;

        internal PdfPageStreamWriter(PdfPageStreamWriterContext pageContext)
        {
            this.pageContext = pageContext;
            this.xObjectResources = new PdfDictionary();
            this.xFormResourceIds = new Dictionary<IndirectReference, string>();
            this.annotations = new PdfArray();
            this.contentStream = new MemoryStream();
            this.contentWriter = new PdfWriter(this.contentStream);
            this.positionStack = new PositionStack(new SimplePosition());
            this.disposeValidator = new DisposeValidator();
            this.currentResourceId = 0;
        }

        /// <summary>
        /// Gets the size of the physical medium 
        /// on which the page is intended to be displayed or printed.
        /// </summary>
        public Size PageSize
        {
            get
            {
                return new Size(this.pageContext.MediaBox.Width, this.pageContext.MediaBox.Height);
            }
        }

        /// <summary>
        /// Gets the rectangle defining the boundaries of the physical medium 
        /// on which the page is intended to be displayed or printed.
        /// </summary>
        public Rect MediaBox
        {
            get
            {
                return this.pageContext.MediaBox;
            }
        }

        /// <summary>
        /// Gets the rectangle defining the boundaries of the visible region of the page. 
        /// When the page is displayed or printed, its contents are to be clipped (cropped) to this rectangle.
        /// </summary>
        public Rect CropBox
        {
            get
            {
                return this.pageContext.CropBox;
            }
        }

        /// <summary>
        /// Gets the page rotation.
        /// </summary>
        public Rotation PageRotation
        {
            get
            {
                return this.pageContext.PageRotation;
            }
        }

        /// <summary>
        /// Gets or sets the current content position.
        /// </summary>
        public IPosition ContentPosition
        {
            get
            {
                return this.positionStack.Position;
            }
            set
            {
                this.positionStack.Position = value;
            }
        }

        private PdfStreamWriterSettings Settings
        {
            get
            {
                return this.pageContext.Settings;
            }
        }

        private PdfFileStreamExportContext ExportContext
        {
            get
            {
                return this.pageContext.ExportContext;
            }
        }

        /// <summary>
        /// Writes the content of a <see cref="PdfPageSource"/>.
        /// </summary>
        /// <param name="pdfPageSource">The PDF page source.</param>
        public void WriteContent(PdfPageSource pdfPageSource)
        {
            Matrix? transformation = this.CalculateTransformationMatrix(pdfPageSource.Size);

            if (this.Settings.WriteAnnotations)
            {
                this.AppendAnnotations(pdfPageSource, transformation);
            }

            IndirectReference contentReference = this.GetXFormReference(pdfPageSource);
            this.AppendXFormContent(contentReference, transformation);
        }

        /// <summary>
        /// Writes the content of a <see cref="RadFixedPage"/>.
        /// </summary>
        /// <param name="page">The page instace.</param>
        public void WriteContent(RadFixedPage page)
        {
            Matrix? transformation = this.CalculateTransformationMatrix(page.Size);

            if (this.Settings.WriteAnnotations)
            {
                this.AppendAnnotations(page, transformation);
            }

            IndirectReference contentReference = this.GetXFormReference(page);
            this.AppendXFormContent(contentReference, transformation);
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing,
        /// or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            this.disposeValidator.Dispose();

            if (this.ExportContext.LicenceMessageContentOwner != null)
            {
                this.WriteLicenseMessage(this.ExportContext.LicenceMessageContentOwner);
            }

            this.WritePdfPage();
            this.contentStream.Dispose();
            this.pageContext.DoOnPageWritingEndedAction();
        }

        /// <summary>
        /// Saves current content position and returns IDisposable action that restores the position when disposed.
        /// </summary>
        /// <returns>IDisposable action that restores the position when disposed.</returns>
        public IDisposable SaveContentPosition()
        {
            return this.positionStack.SavePosition();
        }

        private static T GetNonReferencePrimitive<T>(IPdfImportContext importContext, IPdfPrimitive primitive)
            where T : class, IPdfPrimitive
        {
            T result = primitive as T;

            if (result == null)
            {
                IndirectReference reference = (IndirectReference)primitive;
                IndirectObject resultObject = importContext.ReadIndirectObject(reference);
                result = (T)resultObject.Content;
            }

            return result;
        }

        private IndirectReference GetXFormReference(PdfPageSource pageSource)
        {
            IndirectReference xFormReference;
            if (!this.ExportContext.TryGetXFormReference(pageSource, out xFormReference))
            {
                FormXObject xForm = this.CreateXFormWithContents(pageSource);
                Rect visibleContentBox = PageLayoutHelper.GetVisibleContentBox(pageSource);
                xForm.BoundingBox = visibleContentBox.ToBottomLeftCoordinateSystem(pageSource.Size.Height);

                if (pageSource.Page.Resources != null)
                {
                    xForm.CopyResourcesFrom(pageSource.Page, pageSource.FileSource, this.ExportContext);
                }

                xFormReference = this.ExportContext.CreateIndirectObject(xForm).Reference;
                this.ExportContext.AddXFormReference(pageSource, xFormReference);

                PdfExporter.WritePendingIndirectObjects(this.ExportContext, this.ExportContext.Writer);
            }

            return xFormReference;
        }

        private IndirectReference GetXFormReference(RadFixedPage pageRoot)
        {
            IndirectReference xFormReference;
            if (!this.ExportContext.TryGetXFormReference(pageRoot, out xFormReference))
            {
                FormXObject xForm = new FormXObject();
                Rect visiblePageRectangle = PageLayoutHelper.GetVisibleContentBox(pageRoot);
                xForm.CopyPropertiesFrom(this.ExportContext, pageRoot, visiblePageRectangle);
                xFormReference = this.ExportContext.CreateIndirectObject(xForm).Reference;

                this.ExportContext.AddXFormReference(pageRoot, xFormReference);

                PdfExporter.WritePendingIndirectObjects(this.ExportContext, this.ExportContext.Writer);
            }

            return xFormReference;
        }

        private void AppendAnnotations(PdfPageSource pdfPageSource, Matrix? transformation)
        {
            PdfSourceImportContext importContext = pdfPageSource.FileSource.Context;
            List<PdfDictionary> oldAnnotations;
            List<PdfDictionary> newAnnotationsEmptyDictionaries;
            this.RegisterOldAndNewAnnotationDictionariesMapping(pdfPageSource, out oldAnnotations, out newAnnotationsEmptyDictionaries);

            for (int i = 0; i < oldAnnotations.Count; i++)
            {
                PdfDictionary annotDict = oldAnnotations[i];
                PdfDictionary renamedDict = newAnnotationsEmptyDictionaries[i];
                ResourceRenamer.FillEmptyDictionaryWithRenamedResources(renamedDict, annotDict, pdfPageSource.FileSource, this.ExportContext);

                if (transformation.HasValue)
                {
                    IPdfPrimitive annotRectPrimitive = annotDict[AnnotationObject.RectName];
                    PdfArray rectArray = GetNonReferencePrimitive<PdfArray>(importContext, annotRectPrimitive);
                    Rect rect = rectArray.ToRect(importContext.Reader, importContext);
                    Rect transformedRect = transformation.Value.Transform(rect);
                    renamedDict[AnnotationObject.RectName] = transformedRect.ToPdfArray();
                }

                IndirectObject annotObject = this.ExportContext.CreateIndirectObject(renamedDict);
                this.annotations.Add(annotObject.Reference);
            }
        }

        private void RegisterOldAndNewAnnotationDictionariesMapping(PdfPageSource pdfPageSource, out List<PdfDictionary> oldAnnotations, out List<PdfDictionary> newAnnotationsEmptyDictionaries)
        {
            oldAnnotations = new List<PdfDictionary>();
            newAnnotationsEmptyDictionaries = new List<PdfDictionary>();

            PdfDictionary sourcePageProperties = pdfPageSource.PageDictionary;
            IPdfPrimitive annotsValue;
            if (sourcePageProperties.TryGetElement(Page.AnnotsName, out annotsValue))
            {
                PdfSourceImportContext importContext = pdfPageSource.FileSource.Context;
                PdfArray sourceArray = GetNonReferencePrimitive<PdfArray>(importContext, annotsValue);

                foreach (IPdfPrimitive annotPrimitive in sourceArray)
                {
                    if (annotPrimitive != null && annotPrimitive.Type != PdfElementType.Null)
                    {
                        PdfDictionary emptyNewDict = new PdfDictionary();
                        IndirectObject newAnnotationObject = this.ExportContext.CreateIndirectObject(emptyNewDict);

                        PdfDictionary annotDict;
                        if (annotPrimitive.Type == PdfElementType.Dictionary)
                        {
                            annotDict = (PdfDictionary)annotPrimitive;
                        }
                        else
                        {
                            IndirectReference reference = (IndirectReference)annotPrimitive;
                            IndirectObject resultObject = importContext.ReadIndirectObject(reference);
                            annotDict = (PdfDictionary)resultObject.Content;
                            this.ExportContext.AddAnnotationDictionariesMappings(pdfPageSource, reference, newAnnotationObject.Reference);
                        }

                        oldAnnotations.Add(annotDict);
                        newAnnotationsEmptyDictionaries.Add(emptyNewDict);
                    }
                }
            }
        }

        private void AppendAnnotations(RadFixedPage fixedPage, Matrix? transformation)
        {
            foreach (Fixed.Model.Annotations.Annotation annotation in fixedPage.Annotations)
            {
                AnnotationObject annot = AnnotationFactory.CreatePdfAnnotation(this.ExportContext, annotation, fixedPage);

                if (transformation.HasValue)
                {
                    PdfReal x1 = (PdfReal)annot.Rect[0];
                    PdfReal y1 = (PdfReal)annot.Rect[1];
                    PdfReal x2 = (PdfReal)annot.Rect[2];
                    PdfReal y2 = (PdfReal)annot.Rect[3];
                    Rect rect = PdfObjectsExtensions.CreateRectFromArrayValues(x1, y1, x2, y2);
                    Rect transformedRect = transformation.Value.Transform(rect);
                    annot.Rect = transformedRect.ToPdfArray();
                }

                IndirectObject annotObject = this.ExportContext.CreateIndirectObject(annot);
                this.annotations.Add(annotObject.Reference);
            }
        }

        private FormXObject CreateXFormWithContents(PdfPageSource pdfPageSource)
        {
            PdfDictionary sourceProperties = pdfPageSource.PageDictionary;

            FormXObject xForm;

            IPdfPrimitive contentsValue;
            if (sourceProperties.TryGetElement(Page.ContentsName, out contentsValue))
            {
                PdfStream importedStream;
                if (this.TryGetSingleContentStream(pdfPageSource.FileSource, contentsValue, out importedStream))
                {
                    xForm = new FormXObject(importedStream, pdfPageSource.FileSource, this.ExportContext);
                }
                else
                {
                    xForm = new FormXObject(pdfPageSource.Page.Contents.Data);
                }
            }
            else
            {
                xForm = new FormXObject(new byte[0]);
            }

            return xForm;
        }

        private bool TryGetSingleContentStream(PdfFileSource fileSource, IPdfPrimitive contentPrimitive, out PdfStream importedStream)
        {
            bool result = false;

            if (contentPrimitive.Type == PdfElementType.IndirectReference)
            {
                IndirectReference singleReference = (IndirectReference)contentPrimitive;
                IndirectObject contentObject = fileSource.Context.ReadIndirectObject(singleReference);

                if (contentObject.Content.Type == PdfElementType.Array)
                {
                    PdfArray array = (PdfArray)contentObject.Content;
                    result = this.TryGetSingleContentStreamFromDirectArray(fileSource, array, out importedStream);
                }
                else
                {
                    importedStream = (PdfStream)contentObject.Content;
                    result = true;
                }
            }
            else
            {
                PdfArray array = (PdfArray)contentPrimitive;
                result = this.TryGetSingleContentStreamFromDirectArray(fileSource, array, out importedStream);
            }

            return result;
        }

        private bool TryGetSingleContentStreamFromDirectArray(PdfFileSource fileSource, PdfArray contentArray, out PdfStream importedStream)
        {
            if (contentArray.Count == 1)
            {
                IndirectReference singleContentReference = (IndirectReference)contentArray[0];
                IndirectObject contentObject = fileSource.Context.ReadIndirectObject(singleContentReference);
                importedStream = (PdfStream)contentObject.Content;

                return true;
            }

            importedStream = null;
            return false;
        }

        private void AppendXFormContent(IndirectReference reference, Matrix? transformation)
        {
            if (transformation.HasValue)
            {
                this.contentWriter.WriteLine(ContentStreamOperatorNames.SaveGraphicsState);

                foreach (PdfReal operand in PdfPageStreamWriter.GetMatrixOperands(transformation.Value))
                {
                    operand.Write(this.contentWriter, this.ExportContext);
                    this.contentWriter.WriteSeparator();
                }

                this.contentWriter.Write(ContentStreamOperatorNames.ConcatMatrix);
                this.contentWriter.WriteLine();
            }

            string resourceId = this.GetXFormResourceId(reference);
            this.contentWriter.WritePdfName(resourceId);
            this.contentWriter.WriteSeparator();
            this.contentWriter.Write(ContentStreamOperatorNames.Paint);
            this.contentWriter.WriteLine();

            if (transformation.HasValue)
            {
                this.contentWriter.WriteLine(ContentStreamOperatorNames.RestorGraphicsState);
            }
        }

        private static IEnumerable<PdfReal> GetMatrixOperands(Matrix matrix)
        {
            yield return new PdfReal(matrix.M11);
            yield return new PdfReal(matrix.M12);
            yield return new PdfReal(matrix.M21);
            yield return new PdfReal(matrix.M22);
            yield return new PdfReal(matrix.OffsetX);
            yield return new PdfReal(matrix.OffsetY);
        }

        private Matrix? CalculateTransformationMatrix(Size formSize)
        {
            if (formSize.Equals(this.PageSize) && this.ContentPosition.Matrix.IsIdentity)
            {
                return null;
            }
            else
            {
                Matrix m = this.ContentPosition.Matrix;
                double transformOffsetX = Unit.DipToPoint(m.M21 * formSize.Height + m.OffsetX);
                double transformOffsetY = Unit.DipToPoint(this.PageSize.Height - m.M22 * formSize.Height - m.OffsetY);
                Matrix transform = new Matrix(m.M11, -m.M12, -m.M21, m.M22, transformOffsetX, transformOffsetY);

                //// The above transform matrix may also be calculated with the following code.
                //// However, the following code uses many Matrix calculations and is much slower compared to the above code.
                //// Matrix tranform = new Matrix()
                ////    .ScaleMatrix(RadFixedDocumentImportContext.PointToDip, -RadFixedDocumentImportContext.PointToDip)
                ////    .TranslateMatrix(0, formSize.Height)
                ////    .MultiplyBy(this.ContentWriterSettings.ContentPosition.Matrix)
                ////    .ScaleMatrix(BaseExportContext.dipToPoint, -BaseExportContext.dipToPoint)
                ////    .TranslateMatrix(0, pageHeight);

                return transform;
            }
        }

        private string GetXFormResourceId(IndirectReference reference)
        {
            string resourceId;
            if (!this.xFormResourceIds.TryGetValue(reference, out resourceId))
            {
                resourceId = string.Format("XForm{0}", this.currentResourceId++);
                this.xObjectResources[resourceId] = reference;
                this.xFormResourceIds.Add(reference, resourceId);
            }

            return resourceId;
        }

        private void WritePdfPage()
        {
            this.pageContext.PageObject.MediaBox = this.MediaBox.ToBottomLeftCoordinateSystem(this.PageSize.Height);
            this.pageContext.PageObject.CropBox = this.CropBox.ToBottomLeftCoordinateSystem(this.PageSize.Height);
            this.pageContext.PageObject.Rotate = Page.GetRotate(this.PageRotation);
            this.pageContext.PageObject.Contents = new ContentStream(this.contentStream.ToArray());
            this.pageContext.PageObject.Resources = new PdfResource();
            this.pageContext.PageObject.Resources.XObjects = this.xObjectResources;
            this.pageContext.PageObject.Annots = this.annotations.Count > 0 ? this.annotations : null;

            IndirectObject pageObject = this.ExportContext.CreateIndirectObject(this.pageContext.PageObject);
            this.ExportContext.IndirectObjectsQueue.Enqueue(pageObject);
            PdfExporter.WritePendingIndirectObjects(this.ExportContext, this.ExportContext.Writer);
        }

        private void WriteLicenseMessage(RadFixedPage licenceMessageContentOwner)
        {
            using (this.SaveContentPosition())
            {
                this.ContentPosition = new SimplePosition();
                this.WriteContent(licenceMessageContentOwner);
            }
        }
    }
}
