﻿using System;
using System.IO;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Streaming
{
    /// <summary>
    /// Represents PDF file stream that writes file content directly to a given stream.
    /// </summary>
    public sealed class PdfStreamWriter : IDisposable
    {
        private readonly Stream stream;
        private readonly bool leaveStreamOpen;
        private readonly PdfFileStreamExportContext context;
        private readonly PdfStreamWriterSettings settings;
        private readonly DisposeValidator disposeValidator;
        private bool isWritingPage;

        /// <summary>
        /// Initializes a new instance of the <see cref="PdfStreamWriter" /> class.
        /// </summary>
        /// <param name="stream">The stream.</param>
        public PdfStreamWriter(Stream stream)
            : this(stream, false)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="PdfStreamWriter" /> class.
        /// </summary>
        /// <param name="stream">The stream.</param>
        /// <param name="leaveStreamOpen">Boolean value indicating whether or not to close the stream after <see cref="PdfStreamWriter" /> instance is disposed.</param>
        public PdfStreamWriter(Stream stream, bool leaveStreamOpen)
        {
            this.stream = stream;
            this.leaveStreamOpen = leaveStreamOpen;
            PdfExportSettings exportSettings = new PdfExportSettings();
            RadFixedDocumentInfo documentInfo = new RadFixedDocumentInfo();
            this.context = new PdfFileStreamExportContext(stream, documentInfo, exportSettings);
            this.settings = new PdfStreamWriterSettings(exportSettings, documentInfo);
            this.disposeValidator = new DisposeValidator();

            this.BeginDocument();
        }

        /// <summary>
        /// The settings 
        /// </summary>
        public PdfStreamWriterSettings Settings
        {
            get
            {
                return this.settings;
            }
        }

        internal PdfFileStreamExportContext Context
        {
            get
            {
                return this.context;
            }
        }

        /// <summary>
        /// Writes the page.
        /// </summary>
        /// <param name="pageSource">The page source.</param>
        public void WritePage(PdfPageSource pageSource)
        {
            Page emptyPageObject = new Page();
            IndirectReference pageReference = this.context.AddNextPageReferenceToGetReference(emptyPageObject);
            this.context.AddPageSourceToContextReferenceMapping(pageSource, pageReference);

            using (PdfPageStreamWriter pageWriter = this.BeginPage(emptyPageObject, pageSource.MediaBox, pageSource.CropBox, pageSource.Rotation))
            {
                pageWriter.WriteContent(pageSource);
            }
        }

        /// <summary>
        /// Writes the page.
        /// </summary>
        /// <param name="page">The page.</param>
        public void WritePage(RadFixedPage page)
        {
            Page emptyPageObject = this.CreateNextPageEmptyPageObject();
            this.context.MapPages(page, emptyPageObject);

            using (PdfPageStreamWriter pageWriter = this.BeginPage(emptyPageObject, page.MediaBox, page.CropBox, page.Rotation))
            {
                pageWriter.WriteContent(page);
            }
        }

        /// <summary>
        /// Begins the page.
        /// </summary>
        /// <param name="pageSize">Size of the page.</param>
        /// <returns>The page stream.</returns>
        public PdfPageStreamWriter BeginPage(Size pageSize)
        {
            return this.BeginPage(pageSize, Rotation.Rotate0);
        }

        /// <summary>
        /// Begins the page.
        /// </summary>
        /// <param name="pageSize">Size of the page.</param>
        /// <param name="rotation">The rotation.</param>
        /// <returns>The page stream.</returns>
        public PdfPageStreamWriter BeginPage(Size pageSize, Rotation rotation)
        {
            Rect mediaBox = new Rect(0, 0, pageSize.Width, pageSize.Height);
            return this.BeginPage(mediaBox, mediaBox, rotation);
        }

        /// <summary>
        /// Begins the page.
        /// </summary>
        /// <param name="mediaBox">Defines the boundaries of the physical medium 
        /// on which the page is intended to be displayed or printed.</param>
        /// <param name="cropBox">Defines the boundaries of the visible region of the page. 
        /// When the page is displayed or printed, its contents are to be clipped (cropped) to this rectangle.</param>
        /// <param name="rotation">The rotation.</param>
        /// <returns>The page stream.</returns>
        public PdfPageStreamWriter BeginPage(Rect mediaBox, Rect cropBox, Rotation rotation)
        {
            Page emptyPageObject = this.CreateNextPageEmptyPageObject();
            return this.BeginPage(emptyPageObject, mediaBox, cropBox, rotation);
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing,
        /// or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            try
            {
                this.disposeValidator.Dispose();
                this.WriteAllPendingPdfObjects();
            }
            finally
            {
                if (!this.leaveStreamOpen)
                {
                    this.stream.Dispose();
                }
            }
        }

        internal void EndWritingPage()
        {
            Guard.ThrowExceptionIfFalse(this.isWritingPage, "isWritingPage");

            this.isWritingPage = false;
        }

        private void BeginDocument()
        {
            this.context.Writer.WriteDocumentStart();
        }

        private PdfPageStreamWriter BeginPage(Page emptyPageObject, Rect mediaBox, Rect cropBox, Rotation rotation)
        {
            if (this.isWritingPage)
            {
                throw new InvalidOperationException("Cannot begin a new page before the previous page is written.");
            }

            this.isWritingPage = true;

            PdfPageStreamWriterContext pageContext = new PdfPageStreamWriterContext()
            {
                PageObject = emptyPageObject,
                MediaBox = mediaBox,
                CropBox = cropBox,
                PageRotation = rotation,
                Settings = this.settings,
                ExportContext = this.context,
                DoOnPageWritingEndedAction = this.EndWritingPage
            };

            PdfPageStreamWriter page = new PdfPageStreamWriter(pageContext);

            return page;
        }

        private Page CreateNextPageEmptyPageObject()
        {
            Page emptyPageObject = new Page();
            this.context.AddNextPageReferenceToGetReference(emptyPageObject);

            return emptyPageObject;
        }

        private void WriteAllPendingPdfObjects()
        {
            DocumentCatalog root = this.context.GetFinalizedDocumentRoot();
            PdfExporter.WriteCatalogAndRelatedProperties(this.context.Writer, this.context, root);
            PdfExporter.WriteFontsFromContext(this.context.Writer, this.context);

            long crossReferenceBeginingOffset = this.context.Writer.Position;
            PdfExporter.WriteTrailerAndCrossReferenceTable(this.context.Writer, this.context, root, null);
            PdfExporter.WriteDocumentEnd(crossReferenceBeginingOffset, this.context.Writer, context);
        }
    }
}
