﻿using System;
using System.Collections.Generic;
using System.IO;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser.Keywords;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Streaming
{
    internal static class ResourceRenamer
    {
        public static void FillEmptyDictionaryWithRenamedResources(PdfDictionary emptyDictionary, PdfDictionary original, PdfFileSource fileSource, IResourceRenamingExportContext context)
        {
            RenamingContext renamingContext = new RenamingContext(fileSource, context);
            ResourceRenamer.FillEmptyDictionaryWithRenamedResources(emptyDictionary, original, renamingContext);
        }

        public static PdfDictionary GetDictionaryWithRenamedResources(PdfDictionary original, PdfFileSource fileSource, IResourceRenamingExportContext context)
        {
            RenamingContext renamingContext = new RenamingContext(fileSource, context);
            return ResourceRenamer.GetDictionaryWithRenamedResources(original, renamingContext);
        }

        public static PdfArray GetArrayWithRenamedResources(PdfArray original, PdfFileSource fileSource, IResourceRenamingExportContext context)
        {
            RenamingContext renamingContext = new RenamingContext(fileSource, context);
            return ResourceRenamer.GetArrayWithRenamedResources(original, renamingContext);
        }

        private static PdfDictionary GetDictionaryWithRenamedResources(PdfDictionary original, RenamingContext renamingContext)
        {
            PdfDictionary result = new PdfDictionary();
            ResourceRenamer.FillEmptyDictionaryWithRenamedResources(result, original, renamingContext);

            return result;
        }

        private static PdfArray GetArrayWithRenamedResources(PdfArray original, RenamingContext renamingContext)
        {
            PdfArray result = new PdfArray();
            ResourceRenamer.FillEmptyArrayWithRenamedResources(result, original, renamingContext);

            return result;
        }

        private static void FillEmptyDictionaryWithRenamedResources(PdfDictionary emptyDictionary, PdfDictionary original, RenamingContext renamingContext)
        {
            foreach (KeyValuePair<string, IPdfPrimitive> pair in original)
            {
                IPdfPrimitive primitiveWithClonedResources = ResourceRenamer.GetPrimitiveWithRenamedResources(pair.Value, renamingContext);
                emptyDictionary[pair.Key] = primitiveWithClonedResources;
            }
        }

        private static void FillEmptyArrayWithRenamedResources(PdfArray emptyArray, PdfArray original, RenamingContext renamingContext)
        {
            foreach (IPdfPrimitive primitive in original)
            {
                IPdfPrimitive primitiveWithClonedResources = ResourceRenamer.GetPrimitiveWithRenamedResources(primitive, renamingContext);
                emptyArray.Add(primitiveWithClonedResources);
            }
        }

        private static PdfStream CreateStreamWithEmptyDictionary(PdfStream original, RenamingContext renamingContext)
        {
            Stream stream = original.Stream;
            PdfInt length = original.Length;
            long offset = original.Offset;

            if (renamingContext.FileSource.Context.Encryption != null)
            {
                // We create new partial PDF file stream containing only the decrypted stream data and endstream keyword. 
                // If we omit endstream keyword InvalidOperationException is thrown while PdfStreamWriter is writing the data.
                byte[] decryptedData = original.ReadRawPdfData();     
                length = new PdfInt(decryptedData.Length);
                offset = 0;
                stream = new MemoryStream();
                stream.Write(decryptedData, 0, decryptedData.Length);
                stream.Write(StreamStartKeyword.EndStreamKeywordBytes, 0, PdfNames.PdfStreamEnd.Length);
            }

            return new PdfStream(renamingContext.FileSource.Context, stream, new PdfDictionary(), length, offset);
        }

        private static IPdfPrimitive GetPrimitiveWithRenamedResources(IPdfPrimitive original, RenamingContext renamingContext)
        {
            if(original == null)
            {
                return PdfNull.Instance;
            }
            else
            {
                IPdfPrimitive primitiveWithClonedResources;

                switch (original.Type)
                {
                    case PdfElementType.Dictionary:
                        PdfDictionary innerDictionary = (PdfDictionary)original;
                        primitiveWithClonedResources = ResourceRenamer.GetDictionaryWithRenamedResources(innerDictionary, renamingContext);
                        break;
                    case PdfElementType.Array:
                        PdfArray array = (PdfArray)original;
                        primitiveWithClonedResources = ResourceRenamer.GetArrayWithRenamedResources(array, renamingContext);
                        break;
                    case PdfElementType.IndirectReference:
                        IndirectReference reference = (IndirectReference)original;
                        primitiveWithClonedResources = ResourceRenamer.GetContextReference(reference, renamingContext);
                        break;
                    case PdfElementType.Stream:
                        PdfStream originalStream = (PdfStream)original;
                        PdfStream resultStream = ResourceRenamer.CreateStreamWithEmptyDictionary(originalStream, renamingContext);
                        ResourceRenamer.FillEmptyDictionaryWithRenamedResources(resultStream.Dictionary, originalStream.Dictionary, renamingContext);
                        primitiveWithClonedResources = resultStream;
                        break;
                    default:
                        primitiveWithClonedResources = original;
                        break;
                }

                return primitiveWithClonedResources;
            }
        }

        private static IPdfPrimitive GetContextReference(IndirectReference sourceReference, RenamingContext renamingContext)
        {
            IndirectReference contextReference;
            if (!renamingContext.ExportContext.TryGetContextIndirectReference(renamingContext.FileSource, sourceReference, out contextReference))
            {
                if (renamingContext.FileSource.IsPageTreeReference(sourceReference) || renamingContext.FileSource.Context.Root.Reference.Equals(sourceReference))
                {
                    return PdfNull.Instance;
                }
                else
                {
                    contextReference = ResourceRenamer.AddContextIndirectReference(sourceReference, renamingContext);
                }                            
            }

            return contextReference;
        }

        private static IndirectReference AddContextIndirectReference(IndirectReference sourceReference, RenamingContext renamingContext)
        {
            IndirectObject sourceObject = renamingContext.FileSource.Context.ReadIndirectObject(sourceReference);
            IPdfPrimitive sourceContent = sourceObject.Content ?? PdfNull.Instance;
            IndirectReference contextReference;

            switch (sourceContent.Type)
            {
                case PdfElementType.Dictionary:
                    PdfDictionary sourceDictionary = (PdfDictionary)sourceContent;
                    PdfDictionary contextDictionary = new PdfDictionary();
                    contextReference = ResourceRenamer.CreateContextReferenceMapping(contextDictionary, sourceReference, renamingContext);
                    ResourceRenamer.FillEmptyDictionaryWithRenamedResources(contextDictionary, sourceDictionary, renamingContext);
                    break;
                case PdfElementType.Array:
                    PdfArray sourceArray = (PdfArray)sourceContent;
                    PdfArray contextArray = new PdfArray();
                    contextReference = ResourceRenamer.CreateContextReferenceMapping(contextArray, sourceReference, renamingContext);
                    ResourceRenamer.FillEmptyArrayWithRenamedResources(contextArray, sourceArray, renamingContext);
                    break;
                case PdfElementType.Stream:
                    PdfStream sourceStream = (PdfStream)sourceContent;
                    PdfStream contextStream = ResourceRenamer.CreateStreamWithEmptyDictionary(sourceStream, renamingContext);
                    contextReference = ResourceRenamer.CreateContextReferenceMapping(contextStream, sourceReference, renamingContext);
                    ResourceRenamer.FillEmptyDictionaryWithRenamedResources(contextStream.Dictionary, sourceStream.Dictionary, renamingContext);
                    break;
                default:
                    IPdfPrimitive contextPrimitive = sourceContent;
                    contextReference = ResourceRenamer.CreateContextReferenceMapping(contextPrimitive, sourceReference, renamingContext);
                    break;
            }

            return contextReference;
        }

        private static IndirectReference CreateContextReferenceMapping(IPdfPrimitive contextPrimitive, IndirectReference sourceReference, RenamingContext renamingContext)
        {
            IndirectObject contextObject = renamingContext.ExportContext.CreateIndirectObject(contextPrimitive);
            IndirectReference contextReference = contextObject.Reference;
            renamingContext.ExportContext.AddSourceToContextReferenceMapping(renamingContext.FileSource, sourceReference, contextReference);

            return contextReference;
        }

        private class RenamingContext
        {
            public readonly PdfFileSource FileSource;
            public readonly IResourceRenamingExportContext ExportContext;

            public RenamingContext(PdfFileSource fileSource, IResourceRenamingExportContext exportContext)
            {
                this.FileSource = fileSource;
                this.ExportContext = exportContext;
            }
        }
    }
}
