﻿using System.Collections.Generic;
using System.IO;
using System.Linq;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Streaming;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Streaming.IncrementalUpdate;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Streaming.IncrementalUpdate.InteractiveForms;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities
{
    internal class DocumentChangesManager
    {
        private class PagePropertyChanges
        {
            public bool IsRotationChanged { get; set; }
            public bool IsSizeChanged { get; set; }
        }

        private readonly RadFixedDocument document;
        private readonly Stream inputStream;
        private readonly HashSet<string> modifiedFieldsNames;
        private readonly Dictionary<int, PagePropertyChanges> modifiedPages;

        internal DocumentChangesManager(RadFixedDocument document, Stream input)
        {
            this.document = document;
            this.inputStream = input;
            this.modifiedFieldsNames = new HashSet<string>();
            this.modifiedPages = new Dictionary<int, PagePropertyChanges>();
        }

        public bool HasChanges
        {
            get
            {
                return this.modifiedFieldsNames.Count > 0 || this.modifiedPages.Count > 0;
            }
        }

        public void MarkFieldAsModified(string fieldName)
        {
            this.modifiedFieldsNames.Add(fieldName);
        }

        public void MarkPageRotationAsModified(int pageIndex)
        {
            PagePropertyChanges changes = this.GetPagePropertyChanges(pageIndex);
            changes.IsRotationChanged = true;
        }

        public void MarkPageSizeAsModified(int pageIndex)
        {
            PagePropertyChanges changes = this.GetPagePropertyChanges(pageIndex);
            changes.IsSizeChanged = true;
        }

        public void SaveChanges(Stream outputStream)
        {
            using (PdfFileSource source = new PdfFileSource(this.inputStream, true))
            {
                using (PdfIncrementalStreamWriter writer = new PdfIncrementalStreamWriter(source, outputStream, true))
                {
                    foreach (string fieldName in this.modifiedFieldsNames)
                    {
                        FormField field = this.document.AcroForm.FormFields[fieldName];
                        this.WriteFieldValueChanges(writer, field);
                    }

                    foreach (int modifiedPageIndex in this.modifiedPages.Keys)
                    {
                        this.WritePagePropertiesChanges(writer, modifiedPageIndex);
                    }
                }
            }
        }

        private void WriteFieldValueChanges(PdfIncrementalStreamWriter writer, FormField field)
        {
            switch (field.FieldType)
            {
                case FormFieldType.CheckBox:
                    this.WriteCheckBoxChanges(writer, (CheckBoxField)field);
                    break;
                case FormFieldType.RadioButton:
                    this.WriteRadioButtonChanges(writer, (RadioButtonField)field);
                    break;
                case FormFieldType.ComboBox:
                    this.WriteComboBoxChanges(writer, (ComboBoxField)field);
                    break;
                case FormFieldType.ListBox:
                    this.WriteListBoxChanges(writer, (ListBoxField)field);
                    break;
                case FormFieldType.TextBox:
                case FormFieldType.CombTextBox:
                    this.WriteTextFieldChanges(writer, (TextField)field);
                    break;
            }
        }

        private void WriteCheckBoxChanges(PdfIncrementalStreamWriter writer, CheckBoxField field)
        {
            using (CheckBoxPropertiesEditor editor = writer.EditCheckBoxProperties(field.Name))
            {
                editor.SetIsChecked(field.IsChecked);
            }
        }

        private void WriteRadioButtonChanges(PdfIncrementalStreamWriter writer, RadioButtonField field)
        {
            using (RadioButtonPropertiesEditor editor = writer.EditRadioButtonProperties(field.Name))
            {
                if (field.Value == null)
                {
                    editor.ClearValue();
                }
                else
                {
                    int index = 0;

                    foreach (RadioButtonWidget widget in field.Widgets)
                    {
                        if (widget.IsSelected)
                        {
                            break;
                        }

                        index++;
                    }

                    if (index == field.Widgets.Count)
                    {
                        editor.ClearValue();
                    }
                    else
                    {
                        editor.SetValue(index);
                    }
                }
            }
        }

        private void WriteComboBoxChanges(PdfIncrementalStreamWriter writer, ComboBoxField field)
        {
            using (ComboBoxPropertiesEditor editor = writer.EditComboBoxProperties(field.Name))
            {
                int selectedIndex;
                if (field.Options.TryGetOptionIndex(field.Value, out selectedIndex))
                {
                    editor.SetValue(selectedIndex);
                }
                else
                {
                    string value = field.HasEditableTextBox ? (field.Value == null ? null : field.Value.Value) : null;
                    editor.SetValue(value);
                }

                SetAppearances(editor, field);
            }
        }

        private void WriteListBoxChanges(PdfIncrementalStreamWriter writer, ListBoxField field)
        {
            using (ListBoxPropertiesEditor editor = writer.EditListBoxProperties(field.Name))
            {
                int[] sortedSelectedIndices = field.GetSelectedIndicesSorted();
                editor.SetValue(sortedSelectedIndices);
                SetAppearances(editor, field);
            }
        }

        private void WriteTextFieldChanges(PdfIncrementalStreamWriter writer, TextField field)
        {
            using (TextFieldPropertiesEditor editor = writer.EditTextFieldProperties(field.Name))
            {
                editor.SetValue(field.Value);
                SetAppearances(editor, field);
            }
        }

        private static void SetAppearances(SingleStateAppearanceFieldPropertiesEditor editor, FormField<VariableContentWidget> field)
        {
            editor.SetAppearance(field.Widgets.Select(widget => widget.Content.NormalContentSource));
        }

        private void WritePagePropertiesChanges(PdfIncrementalStreamWriter writer, int pageIndex)
        {
            RadFixedPage page = this.document.Pages[pageIndex];
            PagePropertyChanges changes = this.GetPagePropertyChanges(pageIndex);

            using (PagePropertiesEditor editor = writer.EditPageProperties(pageIndex))
            {
                if (changes.IsRotationChanged)
                {
                    editor.SetRotation(page.Rotation);
                }

                if (changes.IsSizeChanged)
                {
                    editor.SetMediaBox(page.MediaBox);
                }
            }
        }

        private PagePropertyChanges GetPagePropertyChanges(int pageIndex)
        {
            PagePropertyChanges changes;
            if (!this.modifiedPages.TryGetValue(pageIndex, out changes))
            {
                changes = new PagePropertyChanges();
                this.modifiedPages.Add(pageIndex, changes);
            }

            return changes;
        }
    }
}
