﻿using System;
using System.Collections.Generic;
using System.Windows;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Fixed.Exceptions;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.ColorSpaces;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Common;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Patterns;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Objects;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities
{
    internal static class PdfObjectsExtensions
    {
        private static readonly double PointToDipConversion = Unit.PointToDip(1);
        private static readonly double DipToPointConversion = Unit.DipToPoint(1);

        public static P ToPrimitive<P, T>(this PdfProperty<T> pdfProperty, Func<T, P> convertToPrimitive, Func<P> getDefaultValue = null)
            where P : IPdfSharedObject
        {
            P primitive = default(P);

            if (pdfProperty.HasValue)
            {
                primitive = convertToPrimitive(pdfProperty.Value);
            }
            else if (getDefaultValue != null)
            {
                primitive = getDefaultValue();
            }

            return primitive;
        }

        public static void CopyToProperty<P, T>(this P primitive, PdfProperty<T> pdfProperty, Func<P, T> getValueFromPrimitive)
            where P : IPdfSharedObject
        {
            if (primitive != null)
            {
                pdfProperty.Value = getValueFromPrimitive(primitive);
            }
        }

        public static V GetInheritableProperty<T, V>(this ITreeNode<T> node, Func<T, V> getPropertyFromNodeValue, V defautValue = null)
            where V : class
        {
            ITreeNode<T> currentNode = node;
            V propertyValue = null;

            while (currentNode != null && propertyValue == null)
            {
                propertyValue = getPropertyFromNodeValue(currentNode.NodeValue);
                currentNode = currentNode.ParentNode;
            }

            return propertyValue ?? defautValue;
        }

        public static PdfArray ToPdfArray(this IEnumerable<int> array)
        {
            if (array == null)
            {
                return null;
            }

            PdfArray result = new PdfArray();
            foreach (var value in array)
            {
                result.Add(new PdfInt(value));
            }

            return result;
        }

        public static PdfArray ToPdfArray(this IEnumerable<double> array)
        {
            if (array == null)
            {
                return null;
            }

            PdfArray result = new PdfArray();
            foreach (var value in array)
            {
                result.Add(value.ToPdfReal());
            }

            return result;
        }

        public static string ToText(this PrimitiveWrapper textWrapper)
        {
            string text = null;

            if (textWrapper != null && (textWrapper.Type == PdfElementType.String || textWrapper.Type == PdfElementType.PdfName))
            {
                text = textWrapper.Primitive.ToString();
            }

            return text;
        }

        public static V[] ToArray<P, V>(this PdfArray array, PostScriptReader reader, IPdfImportContext context)
            where P : PdfSimpleType<V>
        {
            if (array == null)
            {
                return null;
            }

            V[] result = new V[array.Count];

            for (int i = 0; i < result.Length; i++)
            {
                P element;
                array.TryGetElement(reader, context, i, out element);

                if (element != null)
                {
                    result[i] = element.Value;
                }
            }

            return result;
        }

        public static double[] ToDoubleArray(this PdfArray array)
        {
            return array.ToArray((primitive) => primitive.ToReal());
        }

        public static int[] ToIntArray(this PdfArray array)
        {
            return array.ToArray((primitive) => (int)primitive.ToReal());
        }

        public static double ToReal(this IPdfSharedObject value)
        {
#if !NETSTANDARD
            double real;

            if (value.IsOldSchema)
            {
                Old.Pdf.Utils.Helper.UnboxDouble(value, out real);
            }
            else
            {
                real = ToReal(value as IPdfPrimitive);
            }

            return real;
#else
            return ToReal(value as IPdfPrimitive);
#endif
        }

        private static double ToReal(IPdfPrimitive value)
        {
            var pdfReal = value as PdfReal;
            if (pdfReal != null)
            {
                return pdfReal.Value;
            }

            var pdfInt = value as PdfInt;
            if (pdfInt != null)
            {
                return pdfInt.Value;
            }

            var pdfBool = value as PdfBool;
            if (pdfBool != null)
            {
                return pdfBool.Value ? 1d : 0d;
            }

            throw new ArgumentException(string.Format("The {0} type cannot be converted to a real numeric value.", value.GetType().Name));
        }

        public static string[] ToStringArray(this PdfArray array)
        {
            return array.ToArray((primitive) => primitive.ToString());
        }

        private static T[] ToArray<T>(this PdfArray array, Func<IPdfPrimitive, T> primitiveConverter)
        {
            T[] result = new T[array.Count];

            for (int i = 0; i < array.Count; i++)
            {
                result[i] = primitiveConverter(array[i]);
            }

            return result;
        }

        public static PdfArray ToPdfArray(this Rect rect)
        {
            PdfArray array = new PdfArray();
            array.Add(rect.X.ToPdfReal());
            array.Add(rect.Y.ToPdfReal());
            array.Add((rect.X + rect.Width).ToPdfReal());
            array.Add((rect.Y + rect.Height).ToPdfReal());

            return array;
        }

        public static PdfArray ToPointPdfArray(this Rect rect)
        {
            PdfArray array = new PdfArray();
            array.Add(Unit.DipToPoint(rect.X).ToPdfReal());
            array.Add(Unit.DipToPoint(rect.Y).ToPdfReal());
            array.Add(Unit.DipToPoint(rect.Right).ToPdfReal());
            array.Add(Unit.DipToPoint(rect.Bottom).ToPdfReal());

            return array;
        }

        public static Rect ToDipRect(this PdfArray array, PostScriptReader reader, IPdfImportContext context)
        {
            Guard.ThrowExceptionIfNull(array, "array");
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            PdfReal x1;
            PdfReal y1;
            PdfReal x2;
            PdfReal y2;

            array.TryGetElement(reader, context, 0, out x1);
            array.TryGetElement(reader, context, 1, out y1);
            array.TryGetElement(reader, context, 2, out x2);
            array.TryGetElement(reader, context, 3, out y2);

            return new Rect(new Point(Unit.PointToDip(x1.Value), Unit.PointToDip(y1.Value)), new Point(Unit.PointToDip(x2.Value), Unit.PointToDip(y2.Value)));
        }

        public static Matrix ToMatrix(this PdfArray array, PostScriptReader reader, IPdfImportContext context)
        {
            Guard.ThrowExceptionIfNull(array, "array");
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            PdfReal m11;
            PdfReal m12;
            PdfReal m21;
            PdfReal m22;
            PdfReal offsetX;
            PdfReal offsetY;

            array.TryGetElement(reader, context, 0, out m11);
            array.TryGetElement(reader, context, 1, out m12);
            array.TryGetElement(reader, context, 2, out m21);
            array.TryGetElement(reader, context, 3, out m22);
            array.TryGetElement(reader, context, 4, out offsetX);
            array.TryGetElement(reader, context, 5, out offsetY);

            return new Matrix()
            {
                M11 = m11.Value,
                M12 = m12.Value,
                M21 = m21.Value,
                M22 = m22.Value,
                OffsetX = offsetX.Value,
                OffsetY = offsetY.Value
            };
        }

        public static Rect ToRect(this PdfArray array, PostScriptReader reader, IPdfImportContext context)
        {
            Guard.ThrowExceptionIfNull(array, "array");
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            PdfReal x1;
            PdfReal y1;
            PdfReal x2;
            PdfReal y2;

            array.TryGetElement(reader, context, 0, out x1);
            array.TryGetElement(reader, context, 1, out y1);
            array.TryGetElement(reader, context, 2, out x2);
            array.TryGetElement(reader, context, 3, out y2);

            return CreateRectFromArrayValues(x1, y1, x2, y2);
        }

        public static Rect CreateRectFromArrayValues(PdfReal x1, PdfReal y1, PdfReal x2, PdfReal y2)
        {
            return new Rect(new Point(x1.Value, y1.Value), new Point(x2.Value, y2.Value));
        }

        public static PdfArray ToPdfArray(this Matrix matrix)
        {
            PdfArray array = new PdfArray();
            array.Add(matrix.M11.ToPdfReal());
            array.Add(matrix.M12.ToPdfReal());
            array.Add(matrix.M21.ToPdfReal());
            array.Add(matrix.M22.ToPdfReal());
            array.Add(matrix.OffsetX.ToPdfReal());
            array.Add(matrix.OffsetY.ToPdfReal());

            return array;
        }

        public static PdfArray ToPdfArray(this string[] array)
        {
            PdfArray pdfArray = new PdfArray();

            foreach (string text in array)
            {
                pdfArray.Add(text.ToPdfName());
            }

            return pdfArray;
        }

        public static PdfString ToPdfString(this string text)
        {
            if (text == null)
            {
                return null;
            }

            bool containsOnlyAsciiSymbols = true;

            foreach (char symbol in text)
            {
                if (symbol > 127)
                {
                    containsOnlyAsciiSymbols = false;
                    break;
                }
            }

            if (containsOnlyAsciiSymbols)
            {
                return text.ToPdfLiteralString(StringType.ASCII);
            }
            else
            {
                return text.ToPdfHexString();
            }
        }

        public static PdfHexString ToPdfHexString(this string text)
        {
            if (text == null)
            {
                return null;
            }

            byte[] bytes = PdfEncoding.Encoding.GetBytes(text);

            return new PdfHexString(bytes);
        }

        public static PdfHexString ToPdfHexString(this byte[] array)
        {
            PdfHexString pdfHexString = new PdfHexString(array);

            return pdfHexString;
        }

        public static PdfName ToPdfName(this string str)
        {
            return new PdfName(str);
        }

        public static PdfLiteralString ToPdfLiteralString(this string str, StringType type)
        {
            if (str == null)
            {
                return null;
            }

            return new PdfLiteralString(str, type);
        }

        public static PdfBool ToPdfBool(this bool value)
        {
            return new PdfBool(value);
        }

        public static PdfInt ToPdfInt(this int i)
        {
            return new PdfInt(i);
        }

        public static PdfReal ToPdfReal(this double? d)
        {
            if (d.HasValue)
            {
                return d.Value.ToPdfReal();
            }

            return null;
        }

        internal static bool TryConvertToPdfReal(this IPdfPrimitive value, out PdfReal pdfReal)
        {
            pdfReal = GetNumeric(value).ToPdfReal();

            return pdfReal != null;
        }

        internal static double? GetNumeric(this IPdfPrimitive value)
        {
            if (value != null)
            {
                PdfInt pdfInt = value as PdfInt;
                if (pdfInt != null)
                {
                    return pdfInt.Value;
                }

                PdfReal pdfReal = value as PdfReal;
                if (pdfReal != null)
                {
                    return pdfReal.Value;
                }
            }

            return null;
        }

        public static PdfReal ToPdfReal(this double d)
        {
            return new PdfReal(Math.Round(d, 8));
        }

        public static PdfReal ToPdfReal(this int i)
        {
            return new PdfReal(i);
        }

        public static ColorObjectBase ToColor(this ColorBase color, IPdfContentExportContext context)
        {
            switch (color.ColorSpace.Name)
            {
                case Fixed.Model.ColorSpaces.ColorSpaceNames.DeviceRgb:

                    return ((RgbColor)color).ToRgbColor(context);

                case Fixed.Model.ColorSpaces.ColorSpaceNames.Pattern:

                    return ((PatternColor)color).ToPatternColor(context);

                default:

                    throw new NotSupportedColorSpaceException(color.ColorSpace.Name);
            }
        }

        public static RgbColorObject ToRgbColor(this RgbColor color, IPdfContentExportContext context)
        {
            RgbColorObject result = new RgbColorObject();
            result.CopyPropertiesFrom(context, color);

            return result;
        }

        public static PatternColorObject ToPatternColor(this PatternColor color, IPdfContentExportContext context)
        {
            PatternColorObject result = PatternColorObject.CreateInstance((int)color.PatternType);
            result.CopyPropertiesFrom(context, color);

            return result;
        }

        public static Matrix GetScaledPosition(this Form form, bool applyTopLeftTransformation)
        {
            double widthScale = (form.FormSource != null && form.FormSource.Size.Width != 0) ? (form.Width / form.FormSource.Size.Width) : 1;
            double heightScale = (form.FormSource != null && form.FormSource.Size.Height != 0) ? (form.Height / form.FormSource.Size.Height) : 1;
            Matrix scaleMatrix = new Matrix(widthScale, 0, 0, heightScale, 0, 0);
            Matrix position;

            if (applyTopLeftTransformation)
            {
                Matrix topLeftMatrix = PdfObjectsExtensions.CalculateTopLeftCoordinateMatrix(form.Height);
                position = topLeftMatrix.MultiplyBy(form.Position.Matrix);
            }
            else
            {
                position = form.Position.Matrix;
            }

            scaleMatrix = scaleMatrix.MultiplyBy(position);

            return scaleMatrix;
        }

        public static Matrix ToTopLeftCoordinateSystem(this IContentRootElement root, Matrix matrix)
        {
            Guard.ThrowExceptionIfNull(root, "root");

            double rootHeight = PdfObjectsExtensions.GetRootHeight(root);
            Matrix topLeftTransformation = PdfObjectsExtensions.CalculateTopLeftCoordinateMatrix(rootHeight);
            Matrix result = matrix.MultiplyBy(topLeftTransformation);

            return result;
        }

        public static Rect ToTopLeftCoordinateSystem(this Rect rect, double rootHeightInPoints)
        {
            double rootHeightToDip = Unit.PointToDip(rootHeightInPoints);
            Matrix topLeftTransformation = PdfObjectsExtensions.CalculateTopLeftCoordinateMatrix(rootHeightToDip);

            Rect result = topLeftTransformation.Transform(rect);

            return result;
        }

        public static PdfArray ToBottomLeftCoordinateSystem(this IContentRootElement root, Rect rect)
        {
            Guard.ThrowExceptionIfNull(root, "root");

            double rootHeight = PdfObjectsExtensions.GetRootHeight(root);
            double rootHeightInPoint = Unit.DipToPoint(rootHeight);
            Matrix bottomLeftTransformation = PdfObjectsExtensions.CalculateBottomLeftCoordinateMatrix(rootHeightInPoint);
            Rect result = bottomLeftTransformation.Transform(rect);

            return result.ToPdfArray();
        }

        public static PdfArray ToBottomLeftCoordinateSystem(this Rect rect, double rootHeightInDip)
        {
            double rootHeightInPoint = Unit.DipToPoint(rootHeightInDip);
            Matrix bottomLeftTransformation = PdfObjectsExtensions.CalculateBottomLeftCoordinateMatrix(rootHeightInPoint);
            Rect result = bottomLeftTransformation.Transform(rect);

            return result.ToPdfArray();
        }

        private static Matrix CalculateTopLeftCoordinateMatrix(double rootHeightInDip)
        {
            Matrix topLeftTransformation = new Matrix(PointToDipConversion, 0, 0, -PointToDipConversion, 0, rootHeightInDip);

            return topLeftTransformation;
        }

        private static Matrix CalculateBottomLeftCoordinateMatrix(double rootHeightInPoint)
        {
            Matrix bottomLeftTransformation = new Matrix(DipToPointConversion, 0, 0, -DipToPointConversion, 0, rootHeightInPoint);

            return bottomLeftTransformation;
        }

        private static double GetRootHeight(IContentRootElement root)
        {
            // TODO: See if only RadFixedPage and FormSource should be threated specially. May be in case of Tiling patterns we should consider some other root elements.
            bool hasRootHeightTransformation = (root is RadFixedPage) || (root is FormSource);

            if (hasRootHeightTransformation)
            {
                return root.Size.Height;
            }
            else
            {
                return 0;
            }
        }
    }
}
