﻿/* Microsoft Permissive License (Ms-PL)
 * Derived from <a href="http://imagetools.codeplex.com" />
 * This license governs use of the accompanying software. If you use the software, you accept this license. If you do not accept the license, do not use the software.
 * 
 * 1. Definitions
 *  The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
 *  A "contribution" is the original software, or any additions or changes to the software.
 *  A "contributor" is any person that distributes its contribution under this license.
 *  "Licensed patents" are a contributor's patent claims that read directly on its contribution.
 *  
 * 2. Grant of Rights
 * (A) Copyright Grant- Subject to the terms of this license, including the license conditions and limitations in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative works of its contribution, and distribute its contribution or any derivative works that you create.
 * (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of its contribution in the software or derivative works of the contribution in the software.
 * 
 * 3. Conditions and Limitations
 * (A) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
 * (B) If you bring a patent claim against any contributor over patents that you claim are infringed by the software, your patent license from such contributor to the software ends automatically.
 * (C) If you distribute any portion of the software, you must retain all copyright, patent, trademark, and attribution notices that are present in the software.
 * (D) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do so under a license that complies with this license.
 * (E) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions. You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
 * */

using System;
using System.IO;
using System.Windows.Media.Imaging;
using Telerik.Windows.Zip;

namespace Telerik.Windows.Documents.Fixed.Imaging.Silverlight
{
    /// <summary>
    /// Encodes a collection of BitmapFrame objects to a PNG image stream.
    /// </summary>
    [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "Png")]
    internal class PngBitmapEncoder : BitmapEncoder, IDisposable
    {
        private const int MaxBlockSize = 0xFFFF;

        private readonly PngCodecInfo codecInfo;
        private Stream pngStream;
        private byte[] pixels;
        private WriteableBitmap writableImage;

        /// <summary>
        /// Initializes a new instance of the PngBitmapEncoder class.
        /// </summary>
        public PngBitmapEncoder()
        {
            this.codecInfo = new PngCodecInfo();
        }

        /// <summary>
        /// Gets information that describes this codec.
        /// </summary>
        public override BitmapCodecInfo CodecInfo
        {
            get
            {
                return this.codecInfo;
            }
        }

        /// <summary>
        /// Disposes the encoder.
        /// </summary>
        public sealed override void Dispose()
        {
            this.Dispose(true);

            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Encodes a bitmap image to a specified System.IO.Stream.
        /// </summary>
        /// <param name="stream">Stream.</param>
        public override void Save(Stream stream)
        {
            base.Save(stream);

            using (Stream encoderStream = this.Encode(this.Frames[0]))
            {
                encoderStream.Position = 0;
                encoderStream.WriteToStream(stream);
            }
        }

        /// <summary>
        /// Disposes the encoder.
        /// </summary>
        /// <param name="cleanUpBoth">Clean up.</param>
        protected virtual void Dispose(bool cleanUpBoth)
        {
            if (cleanUpBoth && this.pngStream != null)
                this.pngStream.Dispose();
        }

        private void CreateWritableData()
        {
            int width = this.writableImage.PixelWidth;
            int height = this.writableImage.PixelHeight;

            this.pixels = new byte[height * width * 4];

            int i = 0;
            foreach (int pixel in this.writableImage.Pixels)
            {
                byte[] bpixel = BitConverter.GetBytes(pixel);
                byte tmp = bpixel[0];
                bpixel[0] = bpixel[2];
                bpixel[2] = tmp;
                bpixel.CopyTo(this.pixels, i);
                i += 4;
            }
        }

        private void WriteHeader()
        {
            byte[] pngHeader = new byte[] { 0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A };
            this.pngStream.Write(pngHeader, 0, 8);

            PngHeader header = new PngHeader();
            header.Width = this.writableImage.PixelWidth;
            header.Height = this.writableImage.PixelHeight;
            header.ColorType = 6;
            header.BitDepth = 8;
            header.FilterMethod = 0;
            header.CompressionMethod = 0;
            header.InterlaceMethod = 0;

            this.WriteHeaderChunk(header);
        }

        private Stream Encode(WriteableBitmap image)
        {
            this.writableImage = image;

            this.pngStream = new MemoryStream();

            this.WriteHeader();
            this.CreateWritableData();

            byte[] fourByteData = new byte[4];

            byte[] size = BitConverter.GetBytes((uint)Math.Round(1 / 2.2 * 100000));
            fourByteData[0] = size[3];
            fourByteData[1] = size[2];
            fourByteData[2] = size[1];
            fourByteData[3] = size[0];

            this.WriteChunk(PngChunkTypes.Gamma, fourByteData);

            this.WriteDataChunks();

            this.WriteEndChunk();

            this.pngStream.Flush();
            return this.pngStream;
        }

        private void WriteHeaderChunk(PngHeader header)
        {
            byte[] chunkData = new byte[13], buffer = null;

            // Write the width integer.
            buffer = BitConverter.GetBytes(header.Width);
            Array.Reverse(buffer);

            Array.Copy(buffer, 0, chunkData, 0, 4);

            // Write the height integer.
            buffer = BitConverter.GetBytes(header.Height);
            Array.Reverse(buffer);

            Array.Copy(buffer, 0, chunkData, 4, 4);

            chunkData[8] = header.BitDepth;
            chunkData[9] = header.ColorType;
            chunkData[10] = header.CompressionMethod;
            chunkData[11] = header.FilterMethod;
            chunkData[12] = header.InterlaceMethod;

            this.WriteChunk(PngChunkTypes.Header, chunkData);
        }

        private void WriteChunk(string type, byte[] data)
        {
            this.WriteChunk(type, data, 0, data != null ? data.Length : 0);
        }

        private void WriteChunk(string type, byte[] data, int offset, int length)
        {
            // Write the length integer.
            byte[] lengthArray = BitConverter.GetBytes(length);
            Array.Reverse(lengthArray);

            this.pngStream.Write(lengthArray, 0, 4);

            byte[] typeArray = new byte[4];
            typeArray[0] = (byte)type[0];
            typeArray[1] = (byte)type[1];
            typeArray[2] = (byte)type[2];
            typeArray[3] = (byte)type[3];

            this.pngStream.Write(typeArray, 0, 4);

            if (data != null)
            {
                this.pngStream.Write(data, offset, length);
            }

            // Write the length integer.
            CRC32 crc = new CRC32();

            crc.UpdateValue(0, typeArray, 0, typeArray.Length);
            if (data != null)
            {
                crc.UpdateValue(0, data, offset, length);
            }

            byte[] crcArray = BitConverter.GetBytes((uint)crc.CrcValue);
            Array.Reverse(crcArray);

            this.pngStream.Write(crcArray, 0, 4);
        }

        private void WriteDataChunks()
        {
            byte[] data = new byte[(this.writableImage.PixelWidth * this.writableImage.PixelHeight * 4)
                + this.writableImage.PixelHeight];

            int rowLength = (this.writableImage.PixelWidth * 4) + 1;

            for (int y = 0; y < this.writableImage.PixelHeight; y++)
            {
                byte compression = 0;
                if (y > 0)
                {
                    compression = 2;
                }
                data[y * rowLength] = compression;

                for (int x = 0; x < this.writableImage.PixelWidth; x++)
                {
                    // Calculate the offset for the new array.
                    int dataOffset = (y * rowLength) + (x * 4) + 1;

                    // Calculate the offset for the original pixel array.
                    int pixelOffset = ((y * this.writableImage.PixelWidth) + x) * 4;

                    data[dataOffset + 0] = this.pixels[pixelOffset + 0];
                    data[dataOffset + 1] = this.pixels[pixelOffset + 1];
                    data[dataOffset + 2] = this.pixels[pixelOffset + 2];
                    data[dataOffset + 3] = this.pixels[pixelOffset + 3];

                    if (y > 0)
                    {
                        int lastOffset = (((y - 1) * this.writableImage.PixelWidth) + x) * 4;

                        data[dataOffset + 0] -= this.pixels[lastOffset + 0];
                        data[dataOffset + 1] -= this.pixels[lastOffset + 1];
                        data[dataOffset + 2] -= this.pixels[lastOffset + 2];
                        data[dataOffset + 3] -= this.pixels[lastOffset + 3];
                    }
                }
            }

            byte[] buffer = null;
            int bufferLength = 0;

            using (MemoryStream tempStream = new MemoryStream())
            {
                using (CompressedStream cStream = new CompressedStream(tempStream, StreamOperationMode.Write, new DeflateSettings()))
                {
                    // writting image data to zip output stream using 32k blocks
                    int length = 0x8000;
                    int offset = 0;
                    while (offset < data.Length)
                    {
                        length = data.Length - offset > length ? length : data.Length - offset;
                        cStream.Write(data, offset, length);

                        offset += length;
                    }

                    cStream.Close();

                    bufferLength = (int)tempStream.Length;
                    buffer = tempStream.GetBuffer();
                }
            }

            int numChunks = bufferLength / MaxBlockSize;

            if (bufferLength % MaxBlockSize != 0)
            {
                numChunks++;
            }

            for (int i = 0; i < numChunks; i++)
            {
                int length = bufferLength - (i * MaxBlockSize);

                if (length > MaxBlockSize)
                {
                    length = MaxBlockSize;
                }

                this.WriteChunk(PngChunkTypes.Data, buffer, i * MaxBlockSize, length);
            }
        }

        private void WriteEndChunk()
        {
            this.WriteChunk(PngChunkTypes.End, null);
        }

        /// <summary>
        /// Contains a list of possible chunk type identifier.
        /// </summary>
        private static class PngChunkTypes
        {
            /// <summary>
            /// The first chunk in a .png file. Can only exists once. Contains 
            /// common information like the width and the height of the image or
            /// the used compression method.
            /// </summary>
            public const string Header = "IHDR";

            /// <summary>
            /// The PLTE chunk contains from 1 to 256 palette entries, each a three byte
            /// series in the RGB format.
            /// </summary>
            public const string Palette = "PLTE";

            /// <summary>
            /// The IDAT chunk contains the actual image data. The image can contains more
            /// than one chunk of this type. All chunks together are the whole image.
            /// </summary>
            public const string Data = "IDAT";

            /// <summary>
            /// This chunk must appear last. It marks the end of the PNG DataStream. 
            /// The chunk's data field is empty. 
            /// </summary>
            public const string End = "IEND";

            /// <summary>
            /// This chunk specifies that the image uses simple transparency: 
            /// either alpha values associated with palette entries (for indexed-color images) 
            /// or a single transparent color (for grayscale and true-color images). 
            /// </summary>
            public const string PaletteAlpha = "tRNS";

            /// <summary>
            /// Textual information that the encoder wishes to record with the image can be stored in 
            /// "tEXt" chunks. Each "tEXt" chunk contains a keyword and a text string.
            /// </summary>
            public const string Text = "tEXt";

            /// <summary>
            /// This chunk specifies the relationship between the image samples and the desired 
            /// display output intensity.
            /// </summary>
            public const string Gamma = "gAMA";
        }

        /// <summary>
        /// Represents the .png header chunk.
        /// </summary>
        private class PngHeader
        {
            /// <summary>
            /// The dimension in x-direction of the image in pixels.
            /// </summary>
            public int Width { get; set; }

            /// <summary>
            /// The dimension in y-direction of the image in pixels.
            /// </summary>
            public int Height { get; set; }

            /// <summary>
            /// Bit depth is a single-byte integer giving the number of bits per sample 
            /// or per palette index (not per pixel). Valid values are 1, 2, 4, 8, and 16, 
            /// although not all values are allowed for all color types. 
            /// </summary>
            public byte BitDepth { get; set; }

            /// <summary>
            /// Color type is a integer that describes the interpretation of the 
            /// image data. Color type codes represent sums of the following values: 
            /// 1 (palette used), 2 (color used), and 4 (alpha channel used).
            /// </summary>
            public byte ColorType { get; set; }

            /// <summary>
            /// Indicates the method  used to compress the image data. At present, 
            /// only compression method 0 (deflate/inflate compression with a sliding 
            /// window of at most 32768 bytes) is defined.
            /// </summary>
            public byte CompressionMethod { get; set; }

            /// <summary>
            /// Indicates the preprocessing method applied to the image 
            /// data before compression. At present, only filter method 0 
            /// (adaptive filtering with five basic filter types) is defined.
            /// </summary>
            public byte FilterMethod { get; set; }

            /// <summary>
            /// Indicates the transmission order of the image data. 
            /// Two values are currently defined: 0 (no interlace) or 1 (Adam7 interlace).
            /// </summary>
            public byte InterlaceMethod { get; set; }
        }

        private class CRC32
        {
            private static readonly uint[] table = new uint[] 
            { 
                0, 0x77073096, 0xee0e612c, 0x990951ba, 0x76dc419, 0x706af48f, 0xe963a535, 0x9e6495a3, 0xedb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988, 0x9b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91, 
                0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de, 0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7, 0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec, 0x14015c4f, 0x63066cd9, 0xfa0f3d63, 0x8d080df5, 
                0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172, 0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b, 0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940, 0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59, 
                0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423, 0xcfba9599, 0xb8bda50f, 0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924, 0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d, 
                0x76dc4190, 0x1db7106, 0x98d220bc, 0xefd5102a, 0x71b18589, 0x6b6b51f, 0x9fbfe4a5, 0xe8b8d433, 0x7807c9a2, 0xf00f934, 0x9609a88e, 0xe10e9818, 0x7f6a0dbb, 0x86d3d2d, 0x91646c97, 0xe6635c01, 
                0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e, 0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457, 0x65b0d9c6, 0x12b7e950, 0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65, 
                0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2, 0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb, 0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0, 0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9, 
                0x5005713c, 0x270241aa, 0xbe0b1010, 0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f, 0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17, 0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad, 
                0xedb88320, 0x9abfb3b6, 0x3b6e20c, 0x74b1d29a, 0xead54739, 0x9dd277af, 0x4db2615, 0x73dc1683, 0xe3630b12, 0x94643b84, 0xd6d6a3e, 0x7a6a5aa8, 0xe40ecf0b, 0x9309ff9d, 0xa00ae27, 0x7d079eb1, 
                0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb, 0x196c3671, 0x6e6b06e7, 0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc, 0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5, 
                0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252, 0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b, 0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55, 0x316e8eef, 0x4669be79, 
                0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236, 0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f, 0xc5ba3bbe, 0xb2bd0b28, 0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d, 
                0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x26d930a, 0x9c0906a9, 0xeb0e363f, 0x72076785, 0x5005713, 0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0xcb61b38, 0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0xbdbdf21, 
                0x86d3d2d4, 0xf1d4e242, 0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777, 0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c, 0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45, 
                0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2, 0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db, 0xaed16a4a, 0xd9d65adc, 0x40df0b66, 0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9, 
                0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605, 0xcdd70693, 0x54de5729, 0x23d967bf, 0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94, 0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
            };

            private long crc;

            public long CrcValue
            {
                get
                {
                    return this.crc;
                }
                set
                {
                    this.crc = (uint)value;
                }
            }

            public long UpdateValue(long value, byte[] buffer, int index, int length)
            {
                if (value == 1 || buffer == null)
                {
                    this.crc = 0;
                }
                if (buffer != null)
                {
                    if (buffer == null)
                    {
                        throw new ArgumentNullException("buffer");
                    }

                    if (length < 0)
                    {
                        throw new ArgumentOutOfRangeException("length", "Length cannot be less than zero");
                    }

                    if (index < 0 || index + length > buffer.Length)
                    {
                        throw new ArgumentOutOfRangeException("index");
                    }
                    this.crc ^= 0xFFFFFFFF;
                    for (int i = 0; i < length; i++)
                    {
                        this.crc = table[(((int)this.crc) ^ buffer[index + i]) & 255] ^ (this.crc >> 8);
                    }
                    this.crc ^= 0xFFFFFFFF;
                }
                return this.crc;
            }
        }
    }
}
