﻿using System;
using System.Linq;

namespace Telerik.Windows.Documents.Fixed.Model.Annotations
{
    /// <summary>
    /// Represents different annotation types.
    /// </summary>
    public enum AnnotationType
    {
        /// <summary>
        /// Stands for the link annotation, which represents either a hypertext link to a destination elsewhere in the document or an action to be performed.
        /// </summary>
        Link,

        /// <summary>
        /// Stands for the widget annotation, which is used by interactive forms to represent the appearance of fields and to manage user interactions.
        /// </summary>
        Widget,

        /// <summary>
        /// Stands for the text annotation, a “sticky note” attached to a point in the PDF document.
        /// </summary>
        Text,

        /// <summary>
        /// Stands for the free text annotation, which displays text directly on the page.
        /// </summary>
        FreeText,

        /// <summary>
        /// Stands for the line annotation, which displays a single straight line on the page.
        /// </summary>
        Line,

        /// <summary>
        /// Stands for the square annotation, which displays a rectangle on the page.
        /// </summary>
        Square,

        /// <summary>
        /// Stands for the circle annotation, which displays an ellipse on the page.
        /// </summary>
        Circle,

        /// <summary>
        /// Stands for the polygon annotation, which displays a closed polygon on the page.
        /// </summary>
        Polygon,

        /// <summary>
        /// Stands for the polyline annotation which is similar to the polygon annotation,
        /// except that the first and last vertex of the polygon are not implicitly connected.
        /// </summary>
        PolyLine,

        /// <summary>
        /// Stands for the highlight annotation which appears as a highlight in the text of the document.
        /// </summary>
        Highlight,

        /// <summary>
        /// Stands for the underline annotation which appears as an underline in the text of the document.
        /// </summary>
        Underline,

        /// <summary>
        /// Stands for the squiggly annotation which appears as a jagged underline in the text of the document.
        /// </summary>
        Squiggly,

        /// <summary>
        /// Stands for the strikeout annotation which appears as a strikeout in the text of the document.
        /// </summary>
        StrikeOut,

        /// <summary>
        /// Stands for the stamp annotation which displays text or graphics intended to look as if they were stamped on the page with a rubber stamp.
        /// </summary>
        Stamp,

        /// <summary>
        /// Stands for the caret annotation which is a visual symbol that indicates the presence of text edits.
        /// </summary>
        Caret,

        /// <summary>
        /// Stands for the ink annotation which represents a freehand “scribble” composed of one or more disjoint paths.
        /// </summary>
        Ink,

        /// <summary>
        /// Stands for the popup annotation which displays text in a pop-up window for entry and editing.
        /// </summary>
        Popup,

        /// <summary>
        /// Stands for the file attachment annotation which contains a reference to a file.
        /// </summary>
        FileAttachment,

        /// <summary>
        /// Stands for the sound annotation which contains sound recorded from the computer’s microphone or imported from a file.
        /// </summary>
        Sound,

        /// <summary>
        /// Stands for the movie annotation which contains animated graphics and sound to be presented on the computer screen and through the speakers.
        /// </summary>
        Movie,

        /// <summary>
        /// Stands for the screen annotation which specifies a region of a page upon which media clips may be played.
        /// </summary>
        Screen,

        /// <summary>
        /// Stands for the printer mark annotation which represents a mark, added to a page to assist in identifying components of a multiple-plate job and maintaining consistent output during production.
        /// </summary>
        PrinterMark,

        /// <summary>
        /// Stands for the trapnet annotation which defines the trapping characteristics for a page of a PDF document.
        /// </summary>
        TrapNet,

        /// <summary>
        /// Stands for the watermark annotation which is used to represent graphics that are expected to be printed at a fixed size and position on a page, regardless of the dimensions of the printed page.
        /// </summary>
        Watermark,

        /// <summary>
        /// Stands for the 3D annotation which is the means by which 3D artwork is represented in a PDF document.
        /// </summary>
        ThreeD,

        /// <summary>
        /// Stands for the redact annotation which identifies content that is intended to be removed from the document.
        /// </summary>
        Redact
    }
}
