﻿using System;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Annotations
{
    /// <summary>
    /// Represents options specifying how to fit the button icon inside the annotation rectangle.
    /// </summary>
    public class IconFitOptions
    {
        private double leftSpacePercent;
        private double bottomSpacePercent;

        /// <summary>
        /// Initializes a new instance of <see cref="IconFitOptions"/> class.
        /// </summary>
        internal IconFitOptions()
        {
            this.ScaleCondition = IconScaleCondition.AlwaysScale;
            this.ScalingType = IconScalingType.Proportional;
            this.BlankSpaceFromTheLeftSide = 0.5;
            this.BlankSpaceFromTheBottomSide = 0.5;
            this.IgnoreBorderWidth = false;
        }

        internal IconFitOptions(IconFitOptions other)
        {
            this.ScaleCondition = other.ScaleCondition;
            this.ScalingType = other.ScalingType;
            this.BlankSpaceFromTheLeftSide = other.BlankSpaceFromTheLeftSide;
            this.BlankSpaceFromTheBottomSide = other.BlankSpaceFromTheBottomSide;
            this.IgnoreBorderWidth = other.IgnoreBorderWidth;
        }

        /// <summary>
        /// Gets or sets the conditions specifying when the icon should be scaled.
        /// </summary>
        public IconScaleCondition ScaleCondition { get; set; }

        /// <summary>
        /// Gets or sets the type of the scaling.
        /// </summary>
        public IconScalingType ScalingType { get; set; }

        /// <summary>
        /// Gets or sets the percent of the blank space from the left side compared to the whole horizontal blank space occured after a proportional scaling operation.
        /// The value must be in range between 0 and 1. The default value is 0.5 meaning that the left and right blank spaces are equal and the icon should be centered horizontally after the scaling.
        /// </summary>
        public double BlankSpaceFromTheLeftSide
        {
            get
            {
                return this.leftSpacePercent;
            }
            set
            {
                Guard.ThrowExceptionIfOutOfRange(0, 1, value, "value");

                this.leftSpacePercent = value;
            }
        }

        /// <summary>
        /// Gets or sets the percent of the blank space from the bottom side compared to the whole vertical blank space occured after a proportional scaling operation.
        /// The value must be in range between 0 and 1. The default value is 0.5 meaning that the bottom and top blank spaces are equal and the icon should be centered vertically after the scaling.
        /// </summary>
        public double BlankSpaceFromTheBottomSide
        {
            get
            {
                return this.bottomSpacePercent;
            }
            set
            {
                Guard.ThrowExceptionIfOutOfRange(0, 1, value, "value");

                this.bottomSpacePercent = value;
            }
        }

        /// <summary>
        /// Gets or sets boolean value indicating whether to ignore annotation border thickness when making the icon fit calculations.
        /// </summary>
        public bool IgnoreBorderWidth { get; set; }
    }
}
