﻿using System;
using Telerik.Windows.Documents.Fixed.Model.Editing;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Fixed.Model.Resources;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif

namespace Telerik.Windows.Documents.Fixed.Model.Annotations
{
    /// <summary>
    /// Represetns a push button widget.
    /// </summary>
    public class PushButtonWidget : Widget<PushButtonAppearanceCharacteristics>, IContentAnnotation
    {
        private readonly AnnotationContentSource content;

        internal PushButtonWidget(FormField field)
            : base(field)
        {
            this.content = new AnnotationContentSource();
        }

        /// <summary>
        /// Gets the widget content type.
        /// </summary>
        public sealed override WidgetContentType WidgetContentType
        {
            get
            {
                return WidgetContentType.PushButtonContent;
            }
        }

        /// <summary>
        /// Gets the source defining the visual content of the widget.
        /// This content is with bigger priority compared to Widget appearance characteristics and text properties and it is visualized by default when opening the exported document in some PDF viewer.
        /// </summary>
        public AnnotationContentSource Content
        {
            get
            {
                this.EnsureContentIsUpToDate();
                return this.content;
            }
        }

        internal override Rect ButtonContentBox
        {
            get
            {
                if (this.AppearanceCharacteristics.IconFitOptions.IgnoreBorderWidth)
                {
                    return this.Rect;
                }
                else
                {
                    return base.ButtonContentBox;
                }
            }
        }

        /// <summary>
        /// Recalculates the Widget <see cref="AnnotationContentSource"/> content based on Widget text properties and dynamic appearance characteristics.
        /// </summary>
        public sealed override void RecalculateContent()
        {
            this.Content.NormalContentSource = this.RecalculateAppearance(this.DrawNormalContent);
            this.Content.MouseDownContentSource = this.RecalculateAppearance(this.DrawMouseDownContent);
            this.Content.MouseOverContentSource = this.RecalculateAppearance(this.DrawMouseOverContent);
        }

        internal override void ConsumeImportedAppearances(AnnotationAppearances appearances)
        {
            if (appearances.AppearancesType == AnnotationAppearancesType.SingleStateAppearances)
            {
                SingleStateAppearances singeStateAppearances = (SingleStateAppearances)appearances;
                this.Content.Initialize(singeStateAppearances);
            }
        }

        internal void PrepareAppearancesForExport()
        {
            this.Appearances = new SingleStateAppearances(this.Content);
        }

        internal override Widget CreateClonedWidgetInstance(RadFixedDocumentCloneContext cloneContext)
        {
            FormField field = cloneContext.GetClonedField(this.Field);
            PushButtonWidget clonedWidget = new PushButtonWidget(field);
            clonedWidget.Content.Initialize(this.Content);
            clonedWidget.AppearanceCharacteristics = new PushButtonAppearanceCharacteristics(this.AppearanceCharacteristics);

            return clonedWidget;
        }
        private void DrawNormalContent(FixedContentEditor editor)
        {
            this.DrawContent(editor, this.AppearanceCharacteristics.NormalCaption, this.AppearanceCharacteristics.NormalIconSource);
        }

        private void DrawMouseDownContent(FixedContentEditor editor)
        {
            this.DrawContent(editor, this.AppearanceCharacteristics.MouseDownCaption, this.AppearanceCharacteristics.MouseDownIconSource);
        }

        private void DrawMouseOverContent(FixedContentEditor editor)
        {
            this.DrawContent(editor, this.AppearanceCharacteristics.MouseOverCaption, this.AppearanceCharacteristics.MouseOverIconSource);
        }

        private void DrawContent(FixedContentEditor editor, string caption, FormSource icon)
        {
            if (string.IsNullOrEmpty(caption))
            {
                if (icon != null && this.AppearanceCharacteristics.IconAndCaptionPosition != IconAndCaptionPosition.NoIconCaptionOnly)
                {
                    this.DrawIcon(editor, icon, this.ButtonContentBox);
                }
            }
            else
            {
                Block block = new Block();
                this.TextProperties.CopyTo(block);
                block.InsertText(caption);

                if (icon == null)
                {
                    if (this.AppearanceCharacteristics.IconAndCaptionPosition != IconAndCaptionPosition.NoCaptionIconOnly)
                    {
                        this.DrawCaptionCentered(editor, block);
                    }
                }
                else
                {
                    this.DrawCaptionAndIcon(editor, block, icon);
                }
            }
        }

        private void DrawIcon(FixedContentEditor editor, FormSource icon, Rect contentBox)
        {
            bool shouldScale = this.ShouldScale(icon, contentBox);

            if (shouldScale)
            {
                if (this.AppearanceCharacteristics.IconFitOptions.ScalingType == IconScalingType.FitExactly)
                {
                    editor.Position.Translate(contentBox.X, contentBox.Y);
                    editor.DrawForm(icon, contentBox.Width, contentBox.Height);
                }
                else
                {
                    double scaleX = contentBox.Width / icon.Size.Width;
                    double scaleY = contentBox.Height / icon.Size.Height;
                    double scale = Math.Min(scaleX, scaleY);
                    double width = scale * icon.Size.Width;
                    double height = scale * icon.Size.Height;
                    double deltaVertical = contentBox.Width - width;
                    double deltaHorizontal = contentBox.Height - height;
                    double x = contentBox.X + deltaHorizontal * this.AppearanceCharacteristics.IconFitOptions.BlankSpaceFromTheLeftSide;
                    double y = contentBox.Y + deltaVertical * (1 - this.AppearanceCharacteristics.IconFitOptions.BlankSpaceFromTheBottomSide);
                    editor.Position.Translate(x, y);
                    editor.DrawForm(icon, width, height);
                }
            }
            else
            {
                editor.Position.Translate(contentBox.X, contentBox.Y);
                editor.DrawForm(icon);
            }
        }

        private bool ShouldScale(FormSource icon, Rect contentBox)
        {
            switch (this.AppearanceCharacteristics.IconFitOptions.ScaleCondition)
            {
                case IconScaleCondition.NeverScale:
                    return false;
                case IconScaleCondition.ScaleIfBiggerThanAnnotationRectangle:
                    return IsBiggerIcon(icon, contentBox);
                case IconScaleCondition.ScaleIfSmallerThanAnnotationRectangle:
                    return !IsBiggerIcon(icon, contentBox);
                default:
                    return true;
            }
        }

        private static bool IsBiggerIcon(FormSource icon, Rect contentBox)
        {
            bool isBigger = icon.Size.Width > contentBox.Width || icon.Size.Height > contentBox.Height;

            return isBigger;
        }

        private static double GetCenteringOffset(double containerWidth, double elementWidth)
        {
            return (containerWidth - elementWidth) / 2;
        }

        private void DrawCaptionCentered(FixedContentEditor editor, Block block)
        {
            Size size = block.Measure();
            double offsetX = GetCenteringOffset(editor.Root.Size.Width, size.Width);
            double offsetY = GetCenteringOffset(editor.Root.Size.Height, size.Height);
            editor.Position.Translate(offsetX, offsetY);
            editor.DrawBlock(block);
        }

        private void DrawCaptionAndIcon(FixedContentEditor editor, Block block, FormSource icon)
        {
            switch (this.AppearanceCharacteristics.IconAndCaptionPosition)
            {
                case IconAndCaptionPosition.CaptionOverIcon:
                    this.DrawIcon(editor, icon, this.ButtonContentBox);
                    this.DrawCaptionCentered(editor, block);
                    break;
                case IconAndCaptionPosition.NoCaptionIconOnly:
                    this.DrawIcon(editor, icon, this.ButtonContentBox);
                    break;
                case IconAndCaptionPosition.NoIconCaptionOnly:
                    this.DrawCaptionCentered(editor, block);
                    break;
                case IconAndCaptionPosition.CaptionAboveIcon:
                    this.DrawCaptionAboveIcon(editor, block, icon);
                    break;
                case IconAndCaptionPosition.CaptionBelowIcon:
                    this.DrawCaptionBelowIcon(editor, block, icon);
                    break;
                case IconAndCaptionPosition.CaptionToLeftOfIcon:
                    this.DrawCaptionToTheLeftOfIcon(editor, block, icon);
                    break;
                case IconAndCaptionPosition.CaptionToRightOfIcon:
                    this.DrawCationToTheRightOfIcon(editor, block, icon);
                    break;
            }
        }

        private void DrawCationToTheRightOfIcon(FixedContentEditor editor, Block block, FormSource icon)
        {
            Size size = block.Measure();
            Rect box = this.ButtonContentBox;

            if (size.Width + FixedDocumentDefaults.ButtonPadding < box.Width)
            {
                double width = box.Width - size.Width - FixedDocumentDefaults.ButtonPadding;
                double offsetX = box.X + width + FixedDocumentDefaults.ButtonPadding;
                double offsetY = GetCenteringOffset(box.Height, size.Height);
                this.DrawIcon(editor, icon, new Rect(box.X, box.Y, width, box.Height));
                editor.Position.Translate(offsetX, offsetY);
                editor.DrawBlock(block);
            }
            else
            {
                this.DrawCaptionCentered(editor, block);
            }
        }

        private void DrawCaptionToTheLeftOfIcon(FixedContentEditor editor, Block block, FormSource icon)
        {
            Size size = block.Measure();
            Rect box = this.ButtonContentBox;

            if (size.Width + FixedDocumentDefaults.ButtonPadding < box.Width)
            {
                double width = box.Width - size.Width - FixedDocumentDefaults.ButtonPadding;
                double offsetY = GetCenteringOffset(box.Height, size.Height);
                this.DrawIcon(editor, icon, new Rect(box.X + size.Width + FixedDocumentDefaults.ButtonPadding, box.Y, width, box.Height));
                editor.Position.Translate(box.X, offsetY);
                editor.DrawBlock(block);
            }
            else
            {
                this.DrawCaptionCentered(editor, block);
            }
        }

        private void DrawCaptionBelowIcon(FixedContentEditor editor, Block block, FormSource icon)
        {
            Size size = block.Measure();
            Rect box = this.ButtonContentBox;

            if (size.Height + FixedDocumentDefaults.ButtonPadding < box.Height)
            {
                double height = box.Height - size.Height - FixedDocumentDefaults.ButtonPadding;
                double offsetX = GetCenteringOffset(box.Width, size.Width);
                double offsetY = box.X + height + FixedDocumentDefaults.ButtonPadding;
                this.DrawIcon(editor, icon, new Rect(box.X, box.Y, box.Width, height));
                editor.Position.Translate(offsetX, offsetY);
                editor.DrawBlock(block);
            }
            else
            {
                this.DrawCaptionCentered(editor, block);
            }
        }

        private void DrawCaptionAboveIcon(FixedContentEditor editor, Block block, FormSource icon)
        {
            Size size = block.Measure();
            Rect box = this.ButtonContentBox;

            if (size.Height + FixedDocumentDefaults.ButtonPadding < box.Height)
            {
                double height = box.Height - size.Height - FixedDocumentDefaults.ButtonPadding;
                double offsetX = GetCenteringOffset(box.Width, size.Width);
                this.DrawIcon(editor, icon, new Rect(box.X, box.Y + size.Height + FixedDocumentDefaults.ButtonPadding, box.Width, height));
                editor.Position.Translate(offsetX, box.Y);
                editor.DrawBlock(block);
            }
            else
            {
                this.DrawCaptionCentered(editor, block);
            }
        }
    }
}
