﻿using System;
using Telerik.Windows.Documents.Core.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Utilities;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif

namespace Telerik.Windows.Documents.Fixed.Model.Annotations
{
    /// <summary>
    /// Represents the radio button widget.
    /// </summary>
    public class RadioButtonWidget : TwoStatesButtonWidget
    {
        private static readonly double diagonalDirectionCoordinateValue;
        private readonly RadioButtonField field;

        static RadioButtonWidget()
        {
            diagonalDirectionCoordinateValue = 1 / Math.Sqrt(2);
        }

        internal RadioButtonWidget(FormField field)
            : base(field)
        {
            this.field = (RadioButtonField)field;
        }

        /// <summary>
        /// Gets the option that is represented by the widget.
        /// </summary>
        public RadioOption Option { get; internal set; }

        internal bool IsSelected
        {
            get
            {
                Guard.ThrowExceptionIfNull(this.Option, "Option");

                return this.Option.Equals(this.field.Value, this.field.ShouldUpdateRadiosInUnison);
            }
        }

        private bool IsRadioCircleTemplate
        {
            get
            {
                return this.TextProperties.Font.Name == FontsManagerBase.ZapfDingbats && this.AppearanceCharacteristics.NormalCaption == TwoStatesButtonTemplateCaptions.Circle;
            }
        }

        internal sealed override TwoStatesButtonWidget CreateClonedTwoStateButtonWidget(RadFixedDocumentCloneContext cloneContext)
        {
            FormField field = cloneContext.GetClonedField(this.Field);
            RadioButtonWidget clonedWidget = new RadioButtonWidget(field);
            clonedWidget.Option = cloneContext.GetClonedOption(this.Option);

            return clonedWidget;
        }

        internal override GeometryBase CalculateBackgroundGeometry(Size size)
        {
            if (this.IsRadioCircleTemplate)
            {
                double borderThickness = this.Border != null ? this.Border.Width : 0;
                PathGeometry geometry = CreateCircleGeometry(size, borderThickness / 2);

                return geometry;
            }
            else
            {
                return base.CalculateBackgroundGeometry(size);

            }
        }

        internal override void CalculateBorder3dEffectGeometries(Size size, out PathGeometry pathAbove, out PathGeometry pathBelow)
        {
            if (this.IsRadioCircleTemplate)
            {
                double borderThickness = this.Border != null ? this.Border.Width : 0;
                Point center = GetCircleCenter(size);
                double outerRadius = GetCircleRadius(size, borderThickness);
                double innerRadius = GetCircleRadius(size, this.BorderThicknessIncludingStyleThickness);

                Point outerLeft, outerRight, innerLeft, innerRight;
                GetDiametralPoints(center, outerRadius, out outerLeft, out outerRight);
                GetDiametralPoints(center, innerRadius, out innerLeft, out innerRight);

                pathAbove = new PathGeometry();
                PathFigure figureAbove = pathAbove.Figures.AddPathFigure();
                figureAbove.IsClosed = true;
                figureAbove.StartPoint = outerLeft;
                figureAbove.Segments.AddArcSegment(outerRight, outerRadius, outerRadius);
                figureAbove.Segments.AddLineSegment(innerRight);
                figureAbove.Segments.AddArcSegment(innerLeft, innerRadius, innerRadius).SweepDirection = SweepDirection.Counterclockwise;

                pathBelow = new PathGeometry();
                PathFigure figureBelow = pathBelow.Figures.AddPathFigure();
                figureBelow.IsClosed = true;
                figureBelow.StartPoint = outerRight;
                figureBelow.Segments.AddArcSegment(outerLeft, outerRadius, outerRadius);
                figureBelow.Segments.AddLineSegment(innerLeft);
                figureBelow.Segments.AddArcSegment(innerRight, innerRadius, innerRadius).SweepDirection = SweepDirection.Counterclockwise;
            }
            else
            {
                base.CalculateBorder3dEffectGeometries(size, out pathAbove, out pathBelow);
            }
        }

        private static void GetDiametralPoints(Point center, double radius, out Point leftPoint, out Point rightPoint)
        {
            double directionOffset = diagonalDirectionCoordinateValue * radius;
            leftPoint = new Point(center.X - directionOffset, center.Y + directionOffset);
            rightPoint = new Point(center.X + directionOffset, center.Y - directionOffset);
        }

        internal override GeometryBase CalculateContentClippingGeometry(Size size)
        {
            if(this.IsRadioCircleTemplate)
            {
                double borderThicknessWithStyle = this.BorderThicknessIncludingStyleThickness;
                PathGeometry geometry = CreateCircleGeometry(size, borderThicknessWithStyle / 2);

                return geometry;
            }
            else
            {
                return base.CalculateContentClippingGeometry(size);
            }
        }

        internal override double ScaleBlockSize(Size blockSize)
        {
            if (this.IsRadioCircleTemplate)
            {
                Rect contentBox = this.ButtonContentBox;
                double circleRadius = GetCircleRadius(new Size(contentBox.Width, contentBox.Height), 0);
                double scale = 1.3 * circleRadius / blockSize.Width;

                return scale;
            }
            else
            {
                return base.ScaleBlockSize(blockSize);
            }
        }

        private static PathGeometry CreateCircleGeometry(Size boundingBox, double radiusOffset)
        {
            Point center = GetCircleCenter(boundingBox);
            double radius = GetCircleRadius(boundingBox, radiusOffset);
            Point leftDiametralPoint = new Point(center.X - radius, center.Y);
            Point rightDiametralPoint = new Point(center.X + radius, center.Y);

            PathGeometry geometry = new PathGeometry();
            PathFigure figure = geometry.Figures.AddPathFigure();
            figure.IsClosed = true;
            figure.StartPoint = leftDiametralPoint;
            figure.Segments.AddArcSegment(rightDiametralPoint, radius, radius);
            figure.Segments.AddArcSegment(leftDiametralPoint, radius, radius);
            return geometry;
        }

        private static Point GetCircleCenter(Size boundingBox)
        {
            return new Point(boundingBox.Width / 2, boundingBox.Height / 2);
        }

        private static double GetCircleRadius(Size boundingBox, double radiusOffset)
        {
            return (Math.Min(boundingBox.Width, boundingBox.Height) / 2) - radiusOffset;
        }
    }
}
