﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Core.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Model.Editing;
using Telerik.Windows.Documents.Fixed.Model.Editing.Layout;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Fixed.Model.Text;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif

namespace Telerik.Windows.Documents.Fixed.Model.Annotations
{
    /// <summary>
    /// Represents widget that has ON and OFF states.
    /// </summary>
    public class TwoStatesButtonWidget : Widget<ButtonAppearanceCharacteristics>
    {
        internal const string OffState = "Off";
        private readonly AnnotationContentSource onStateContent;
        private readonly AnnotationContentSource offStateContent;

        internal TwoStatesButtonWidget(FormField field)
            : base(field)
        {
            this.onStateContent = new AnnotationContentSource();
            this.offStateContent = new AnnotationContentSource();
        }

        /// <summary>
        /// Gets the widget content type.
        /// </summary>
        public sealed override WidgetContentType WidgetContentType
        {
            get
            {
                return WidgetContentType.TwoStatesContent;
            }
        }

        /// <summary>
        /// Represents the content source displayed when the widget is in its ON state. 
        /// This content is with bigger priority compared to Widget appearance characteristics and text properties and it is visualized by default when opening the exported document in some PDF viewer.
        /// </summary>
        public AnnotationContentSource OnStateContent
        {
            get
            {
                this.EnsureContentIsUpToDate();
                return this.onStateContent;
            }
        }

        /// <summary>
        /// Represents the content source displayed when the widget is in its OFF state.
        /// This content is with bigger priority compared to Widget appearance characteristics and text properties and it is visualized by default when opening the exported document in some PDF viewer.
        /// </summary>
        public AnnotationContentSource OffStateContent
        {
            get
            {
                this.EnsureContentIsUpToDate();
                return this.offStateContent;
            }
        }

        /// <summary>
        /// Recalculates the Widget <see cref="AnnotationContentSource"/> content based on Widget text properties and dynamic appearance characteristics.
        /// </summary>
        public sealed override void RecalculateContent()
        {
            this.OnStateContent.NormalContentSource = this.RecalculateAppearance(this.DrawOnStateContent);
            this.OnStateContent.MouseDownContentSource = null;
            this.OnStateContent.MouseOverContentSource = null;
            this.OffStateContent.NormalContentSource = this.RecalculateAppearance(this.DrawOffStateContent);
            this.OffStateContent.MouseDownContentSource = null;
            this.OffStateContent.MouseOverContentSource = null;
        }

        internal override void ConsumeImportedAppearances(AnnotationAppearances appearances)
        {
            switch (appearances.AppearancesType)
            {
                case AnnotationAppearancesType.SingleStateAppearances:
                    SingleStateAppearances singleStateAppearances = (SingleStateAppearances)appearances;
                    this.OnStateContent.Initialize(singleStateAppearances);
                    this.OffStateContent.Initialize(singleStateAppearances);
                    break;
                case AnnotationAppearancesType.MultiStateAppearances:
                    MultiStateAppearances multiStateAppearances = (MultiStateAppearances)appearances;
                    foreach (KeyValuePair<string, SingleStateAppearances> stateNameToAppearances in multiStateAppearances)
                    {
                        if (stateNameToAppearances.Key.Equals(TwoStatesButtonWidget.OffState))
                        {
                            this.OffStateContent.Initialize(stateNameToAppearances.Value);
                        }
                        else
                        {
                            this.OnStateContent.Initialize(stateNameToAppearances.Value);
                        }
                    }
                    break;
                default:
                    throw new NotSupportedException(string.Format("Not supported appearances type: {0}", appearances.AppearancesType));
            }
        }

        internal void PrepareAppearancesForExport(string onStateName, bool isChecked)
        {
            MultiStateAppearances appearances = new MultiStateAppearances();
            appearances.CurrentState = isChecked ? onStateName : TwoStatesButtonWidget.OffState;
            appearances.AddAppearance(onStateName, new SingleStateAppearances(this.OnStateContent));
            appearances.AddAppearance(TwoStatesButtonWidget.OffState, new SingleStateAppearances(this.OffStateContent));
            this.Appearances = appearances;
        }

        internal sealed override Widget CreateClonedWidgetInstance(RadFixedDocumentCloneContext cloneContext)
        {
            TwoStatesButtonWidget clonedWidget = this.CreateClonedTwoStateButtonWidget(cloneContext);
            clonedWidget.OnStateContent.Initialize(this.OnStateContent);
            clonedWidget.OffStateContent.Initialize(this.OffStateContent);
            clonedWidget.AppearanceCharacteristics = new ButtonAppearanceCharacteristics(this.AppearanceCharacteristics);

            return clonedWidget;
        }

        internal virtual TwoStatesButtonWidget CreateClonedTwoStateButtonWidget(RadFixedDocumentCloneContext cloneContext)
        {
            FormField field = cloneContext.GetClonedField(this.Field);
            TwoStatesButtonWidget clonedWidget = new TwoStatesButtonWidget(field);

            return clonedWidget;
        }

        internal virtual double ScaleBlockSize(Size blockSize)
        {
            double scale = 1;

            if (this.TextProperties.FontSize == 0)
            {
                Size contentBox = new Size(this.ButtonContentBox.Width, this.ButtonContentBox.Height);
                double scaleX = contentBox.Width / blockSize.Width;
                double scaleY = contentBox.Height / blockSize.Height;
                scale = Math.Min(scaleX, scaleY);
            }

            return scale;
        }

        // TODO: This property is used to workaround ZapfDingbats issues with charcode mapping. We should instead fix this issues and remove the property.        
        private bool IsZapfDingbatsFont
        {
            get
            {
                return this.TextProperties.Font.Name == FontsManagerBase.ZapfDingbats;
            }
        }

        private void DrawOnStateContent(FixedContentEditor editor)
        {
            if (!string.IsNullOrEmpty(this.AppearanceCharacteristics.NormalCaption) && !this.ButtonContentBox.IsEmpty)
            {
                TextFragment text = new TextFragment(this.TextProperties.PropertiesOwner);

                if (this.IsZapfDingbatsFont)
                {
                    text.TextCollection.Characters = TwoStatesButtonWidget.GetZapfDingbatsGlyphs(this.AppearanceCharacteristics.NormalCaption);
                }
                else
                {
                    text.TextCollection.Text = this.AppearanceCharacteristics.NormalCaption;
                }

                if (this.TextProperties.FontSize == 0)
                {
                    text.FontSize = FixedDocumentDefaults.FontSize;
                }

                double width = TextFragmentLayoutElement.MeasureWidth(text);
                double height = TextFragmentLayoutElement.MeasureHeight(text);
                Size blockSize = new Size(width, height);
                double scale = this.ScaleBlockSize(blockSize);

                if (scale != 1)
                {
                    blockSize = new Size(scale * blockSize.Width, scale * blockSize.Height);
                    text.FontSize *= scale;
                }

                double offsetX = (editor.Root.Size.Width - blockSize.Width) / 2;
                double offsetY = (editor.Root.Size.Height - blockSize.Height) / 2;
                double deltaY = this.IsZapfDingbatsFont ? text.FontSize * 0.16 : 0;
                editor.Position.Translate(offsetX, offsetY + blockSize.Height - deltaY);
                editor.Draw(text);
            }
        }

        private static IEnumerable<CharInfo> GetZapfDingbatsGlyphs(string normalCaption)
        {
            foreach (char symbol in normalCaption)
            {
                int value = symbol;
                CharCode code = value <= byte.MaxValue ? new CharCode((byte)value) : new CharCode(value, 2);
                yield return new CharInfo(symbol.ToString(), code);
            }
        }

        private void DrawOffStateContent(FixedContentEditor editor)
        {
            // Do nothing.
        }
    }
}
