﻿using System;
using System.IO;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Fixed.Model.Editing;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;

namespace Telerik.Windows.Documents.Fixed.Model.Annotations
{
    /// <summary>
    /// Represents widget that has variable content generated dynamically while interacting with the widget in a PDF viewer.
    /// </summary>
    public class VariableContentWidget : Widget<DynamicAppearanceCharacteristics>, IContentAnnotation
    {
        private readonly AnnotationContentSource content;

        internal VariableContentWidget(FormField field)
            : base(field)
        {
            this.content = new AnnotationContentSource();
        }

        /// <summary>
        /// Gets the widget content type.
        /// </summary>
        public sealed override WidgetContentType WidgetContentType
        {
            get
            {
                return WidgetContentType.VariableContent;
            }
        }

        /// <summary>
        /// The content used to initially visualize the Widget appearance in a PDF viewer. 
        /// This content is with bigger priority compared to Widget appearance characteristics and text properties and it is visualized by default when opening the exported document in some PDF viewer.
        /// </summary>
        public AnnotationContentSource Content
        {
            get
            {
                this.EnsureContentIsUpToDate();
                return this.content;
            }
        }

        internal override string ContentMarkerName
        {
            get
            {
                return "Tx";
            }
        }

        /// <summary>
        /// Recalculates the Widget <see cref="AnnotationContentSource"/> content based on Widget text properties and dynamic appearance characteristics.
        /// </summary>
        public sealed override void RecalculateContent()
        {
            this.Content.NormalContentSource = this.RecalculateAppearance(this.DrawRectangularAppearanceContent);
            this.Content.MouseDownContentSource = null;
            this.Content.MouseOverContentSource = null;
        }

        internal sealed override void ConsumeImportedAppearances(AnnotationAppearances appearances)
        {
            if (appearances.AppearancesType == AnnotationAppearancesType.SingleStateAppearances)
            {
                SingleStateAppearances singeStateAppearances = (SingleStateAppearances)appearances;
                this.Content.Initialize(singeStateAppearances);
            }
        }

        internal void PrepareAppearancesForExport()
        {
            this.Appearances = new SingleStateAppearances(this.Content);
        }

        internal sealed override Widget CreateClonedWidgetInstance(RadFixedDocumentCloneContext cloneContext)
        {
            FormField field = cloneContext.GetClonedField(this.Field);
            VariableContentWidget clonedWidget = new VariableContentWidget(field);
            clonedWidget.Content.Initialize(this.Content);
            clonedWidget.AppearanceCharacteristics = new DynamicAppearanceCharacteristics(this.AppearanceCharacteristics);

            return clonedWidget;
        }

        private void DrawRectangularAppearanceContent(FixedContentEditor editor)
        {
            switch (this.Field.FieldType)
            {
                case FormFieldType.ComboBox:
                    this.DrawComboBoxContent(editor);
                    break;
                case FormFieldType.ListBox:
                    this.DrawListBoxContent(editor);
                    break;
                case FormFieldType.CombTextBox:
                    this.DrawCombTextBoxContent(editor);
                    break;
                case FormFieldType.TextBox:
                    this.DrawTextBoxContent(editor);
                    break;
            }
        }

        private void DrawComboBoxContent(FixedContentEditor editor)
        {
            ComboBoxField combo = (ComboBoxField)this.Field;

            if (combo.Value != null)
            {
                this.DrawTextBoxContent(editor, combo.Value.UserInterfaceValue ?? combo.Value.Value, false);
            }
        }

        private void DrawListBoxContent(FixedContentEditor editor)
        {
            ListBoxField listBox = (ListBoxField)this.Field;
            double borderThickness = this.BorderThicknessIncludingStyleThickness;
            double selectionWidth = editor.Root.Size.Width - 2 * borderThickness;
            double availableWidth = selectionWidth - 2 * FixedDocumentDefaults.ListBoxChoicePadding.Width;
            double offsetY = borderThickness + FixedDocumentDefaults.ListBoxChoicePadding.Height;

            for (int index = listBox.TopIndex; index < listBox.Options.Count; index++)
            {
                ChoiceOption option = listBox.Options[index];
                bool isSelected = listBox.Value != null && listBox.Value.Contains(option);
                RectangleGeometry selectionRectangle = null;

                if (isSelected)
                {
                    using (editor.SaveGraphicProperties())
                    {
                        editor.GraphicProperties.IsFilled = true;
                        editor.GraphicProperties.IsStroked = false;
                        editor.GraphicProperties.StrokeThickness = 0;
                        editor.GraphicProperties.FillColor = FixedDocumentDefaults.ListBoxSelectionColor;
                        editor.Position.Translate(borderThickness, borderThickness);
                        selectionRectangle = new RectangleGeometry();
                        editor.Position.Translate(borderThickness, offsetY);
                        editor.DrawPath(selectionRectangle);
                    }
                }

                string text = option.UserInterfaceValue ?? option.Value;
                editor.Position.Translate(borderThickness + FixedDocumentDefaults.ListBoxChoicePadding.Width, offsetY);
                Size blockSize = this.DrawTextBlockContent(editor, new Size(availableWidth, double.PositiveInfinity), text, true, false);
                double height = blockSize.Height + 2 * FixedDocumentDefaults.ListBoxChoicePadding.Height;
                offsetY += height;

                if (isSelected)
                {
                    selectionRectangle.Rect = new Rect(0, 0, selectionWidth, height);
                }
            }
        }

        private void DrawCombTextBoxContent(FixedContentEditor editor)
        {
            CombTextBoxField textBox = (CombTextBoxField)this.Field;

            if (textBox.Value != null && textBox.MaxLengthOfInputCharacters > 0)
            {
                double borderThickness = this.BorderThicknessIncludingStyleThickness;
                double singleCharacterBoxWidth = (editor.Root.Size.Width - 2 * borderThickness - 2 * FixedDocumentDefaults.TextBoxTextPadding) / textBox.MaxLengthOfInputCharacters;
                double availableHeight = editor.Root.Size.Height - 2 * borderThickness - 2 * FixedDocumentDefaults.TextBoxTextPadding;
                Size singleCharacterBoxSize = new Size(singleCharacterBoxWidth, availableHeight);
                double xOffset = borderThickness + FixedDocumentDefaults.TextBoxTextPadding;
                double yOffset = borderThickness + FixedDocumentDefaults.TextBoxTextPadding;

                for (int i = 0; i < textBox.MaxLengthOfInputCharacters && i < textBox.Value.Length; i++)
                {
                    string textToInsert = textBox.Value[i].ToString();
                    editor.Position.Translate(xOffset, yOffset);
                    this.DrawTextBlockContent(editor, singleCharacterBoxSize, textToInsert, false, false);

                    xOffset += singleCharacterBoxWidth;
                }
            }
        }

        private void DrawTextBoxContent(FixedContentEditor editor)
        {
            TextBoxField textBox = (TextBoxField)this.Field;

            if (textBox.Value != null)
            {
                string textToInsert = textBox.IsPassword ? new String('*', textBox.Value.Length) : textBox.Value;

                if (textBox.MaxLengthOfInputCharacters.HasValue && textToInsert.Length > textBox.MaxLengthOfInputCharacters.Value)
                {
                    textToInsert = textToInsert.Substring(0, textBox.MaxLengthOfInputCharacters.Value);
                }

                bool allowMultilineText = textBox.IsMultiline;
                DrawTextBoxContent(editor, textToInsert, allowMultilineText);
            }
        }

        private void DrawTextBoxContent(FixedContentEditor editor, string textToInsert, bool allowMultilineText)
        {
            double borderThickness = this.BorderThicknessIncludingStyleThickness;
            editor.Position.Translate(FixedDocumentDefaults.TextBoxTextPadding + borderThickness, FixedDocumentDefaults.TextBoxTextPadding + borderThickness);
            double availableWidth = editor.Root.Size.Width - 2 * borderThickness - 2 * FixedDocumentDefaults.TextBoxTextPadding;
            double availableHeight = allowMultilineText ? double.PositiveInfinity : editor.Root.Size.Height - 2 * borderThickness - 2 * FixedDocumentDefaults.TextBoxTextPadding;
            bool wrapText = allowMultilineText;
            this.DrawTextBlockContent(editor, new Size(availableWidth, availableHeight), textToInsert, allowMultilineText, wrapText);
        }

        private Size DrawTextBlockContent(FixedContentEditor positionedEditor, Size availableSize, string text, bool allowMultilineText, bool wrapText)
        {
            Block block = new Block();
            this.TextProperties.CopyTo(block);
            block.VerticalAlignment = double.IsPositiveInfinity(availableSize.Height) ? Editing.Flow.VerticalAlignment.Top : Editing.Flow.VerticalAlignment.Center;
            this.InsertText(block, text, allowMultilineText);

            if (wrapText)
            {
                positionedEditor.DrawBlock(block, availableSize);
            }
            else
            {
                Size size = block.Measure();

                if (size.Width > availableSize.Width)
                {
                    block.HorizontalAlignment = Editing.Flow.HorizontalAlignment.Left;
                    positionedEditor.DrawBlock(block);
                }
                else
                {
                    positionedEditor.DrawBlock(block, availableSize);
                }
            }

            return block.DesiredSize;
        }

        private void InsertText(Block block, string text, bool allowMultilineText)
        {
            if (allowMultilineText)
            {
                using (StringReader reader = new StringReader(text))
                {
                    string line = reader.ReadLine();

                    while (line != null)
                    {
                        block.InsertText(line);

                        line = reader.ReadLine();

                        if (line != null)
                        {
                            block.InsertLineBreak();
                        }
                    }
                }
            }
            else
            {
                block.InsertText(text);
            }
        }
    }
}
