﻿using System;
using Telerik.Windows.Documents.Fixed.Model.Editing;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Utilities;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif

namespace Telerik.Windows.Documents.Fixed.Model.Annotations
{
    /// <summary>
    /// Represents a form field widget.
    /// </summary>
    public abstract class Widget : Annotation
    {
        private readonly FormField field;
        private VariableTextProperties textProperties;
        private bool isAppearanceInvalidated;

        internal Widget(FormField field)
        {
            Guard.ThrowExceptionIfNull(field, "field");

            this.field = field;
            this.HighlightingMode = HighlightingMode.InvertContentOfAnnotationRectangle;
            this.TextProperties = new VariableTextProperties(field.TextProperties);
            this.IsPrintable = true;
        }

        /// <summary>
        /// Gets or sets the highlighting mode.
        /// </summary>
        public HighlightingMode HighlightingMode { get; set; }

        /// <summary>
        /// Represents the text properties used by the PDF viewer in order to build the widget appearance dynamically.
        /// </summary>
        public VariableTextProperties TextProperties
        {
            get
            {
                return this.textProperties;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "value");

                this.textProperties = value;
            }
        }

        /// <summary>
        /// Gets or sets the parent field.
        /// </summary>
        public FormField Field
        {
            get
            {
                return this.field;
            }
        }

        /// <summary>
        /// Gets the widget content type.
        /// </summary>
        public abstract WidgetContentType WidgetContentType { get; }

        /// <summary>
        /// Gets the type of the annotation.
        /// </summary>
        /// <value>The type of the annotation.</value>
        public override AnnotationType Type
        {
            get
            {
                return AnnotationType.Widget;
            }
        }

        internal virtual string ContentMarkerName
        {
            get
            {
                return null;
            }
        }

        /// <summary>
        /// Recalculates the Widget <see cref="AnnotationContentSource"/> content based on Widget text properties and dynamic appearance characteristics.
        /// </summary>
        public abstract void RecalculateContent();

        internal void InvalidateAppearance()
        {
            this.isAppearanceInvalidated = true;
        }

        internal sealed override void DoOnAppearancesImport()
        {
            AnnotationAppearances appearances = this.Appearances;
            this.Appearances = null;

            this.ConsumeImportedAppearances(appearances);
        }

        internal void EnsureContentIsUpToDate()
        {
            if (this.isAppearanceInvalidated)
            {
                this.isAppearanceInvalidated = false;
                this.RecalculateContent();
            }
        }

        internal sealed override Annotation CreateClonedInstance(RadFixedDocumentCloneContext cloneContext)
        {
            Widget widget = this.CreateClonedWidgetInstance(cloneContext);

            widget.TextProperties = new VariableTextProperties(this.TextProperties);
            widget.HighlightingMode = this.HighlightingMode;

            return widget;
        }

        internal abstract Widget CreateClonedWidgetInstance(RadFixedDocumentCloneContext cloneContext);

        internal abstract void ConsumeImportedAppearances(AnnotationAppearances appearances);
    }

    /// <summary>
    /// Represents widget annotation with <see cref="DynamicAppearanceCharacteristics"/>.
    /// </summary>
    public abstract class Widget<T> : Widget
        where T : DynamicAppearanceCharacteristics, new()
    {
        private T appearanceCharacteristics;

        internal Widget(FormField field)
            : base(field)
        {
            this.AppearanceCharacteristics = new T();
        }

        /// <summary>
        /// Gets or sets the appearance characteristics used by a PDF viewer to construct the widget appearance dynamically.
        /// </summary>
        public T AppearanceCharacteristics
        {
            get
            {
                return this.appearanceCharacteristics;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "value");

                this.appearanceCharacteristics = value;
            }
        }

        internal double BorderThicknessIncludingStyleThickness
        {
            get
            {
                if (this.HasBorderWidth)
                {
                    return IsSimulated3dEffectStyle(this.Border.Style) ? this.Border.Width * 2 : this.Border.Width;
                }

                return 0;
            }
        }

        internal virtual Rect ButtonContentBox
        {
            get
            {
                double offset = this.BorderThicknessIncludingStyleThickness + FixedDocumentDefaults.ButtonPadding;
                double width = this.Rect.Width - 2 * offset;
                double height = this.Rect.Height - 2 * offset;

                if (width > 0 && height > 0)
                {
                    return new Rect(offset, offset, width, height);
                }
                else
                {
                    return new Rect();
                }
            }
        }

        private bool HasBorderWidth
        {
            get
            {
                return this.Border != null && this.Border.Style != AnnotationBorderStyle.None && this.Border.Width > 0;
            }
        }

        private bool HasSimulatedBorder3dEffect
        {
            get
            {
                return this.HasBorderWidth && IsSimulated3dEffectStyle(this.Border.Style);
            }
        }

        internal FormSource RecalculateAppearance(Action<FixedContentEditor> drawAppearanceContent)
        {
            if (this.Rect.Width == 0 || this.Rect.Height == 0)
            {
                return new FormSource();
            }

            Size size = new Size(this.Rect.Width, this.Rect.Height);
            FormSource source = new FormSource() { Size = size };
            FixedContentEditor editor = new FixedContentEditor(source);

            using (editor.SaveGraphicProperties())
            {
                this.ApplyBackgroundProperties(editor);
                GeometryBase backgroundGeometry = this.CalculateBackgroundGeometry(size);
                editor.DrawPath(backgroundGeometry);
            }

            if (this.HasSimulatedBorder3dEffect)
            {
                this.DrawSimulatedBorder3dEffect(editor);
            }

            GeometryBase contentClipping = this.CalculateContentClippingGeometry(size);

            using (editor.PushClipping(contentClipping))
            {
                if (this.ContentMarkerName != null)
                {
                    editor.CurrentMarker = new Marker(this.ContentMarkerName);
                }

                drawAppearanceContent(editor);
                editor.CurrentMarker = null;
            }

            return source;
        }

        internal virtual GeometryBase CalculateContentClippingGeometry(Size size)
        {
            double offset = this.BorderThicknessIncludingStyleThickness;
            Rect rectangularClipping = new Rect(offset, offset, size.Width - 2 * offset, size.Height - 2 * offset);

            return new RectangleGeometry(rectangularClipping);
        }

        internal virtual GeometryBase CalculateBackgroundGeometry(Size size)
        {
            double borderThickness = this.Border != null ? this.Border.Width : 0;
            Rect backgroundRect = new Rect(borderThickness / 2, borderThickness / 2, size.Width - borderThickness, size.Height - borderThickness);

            return new RectangleGeometry(backgroundRect);
        }

        internal virtual void CalculateBorder3dEffectGeometries(Size size, out PathGeometry pathAbove, out PathGeometry pathBelow)
        {
            double thickness = this.Border.Width;

            double[] xCoords = { thickness, 2 * thickness, size.Width - 2 * thickness, size.Width - thickness };
            double[] yCoords = { thickness, 2 * thickness, size.Height - 2 * thickness, size.Height - thickness };

            Point outerTopLeft = new Point(xCoords[0], yCoords[0]);
            Point outerBottomLeft = new Point(xCoords[0], yCoords[3]);
            Point outerBottomRight = new Point(xCoords[3], yCoords[3]);
            Point outerTopRight = new Point(xCoords[3], yCoords[0]);

            Point innerTopLeft = new Point(xCoords[1], yCoords[1]);
            Point innerBottomLeft = new Point(xCoords[1], yCoords[2]);
            Point innerBottomRight = new Point(xCoords[2], yCoords[2]);
            Point innerTopRight = new Point(xCoords[2], yCoords[1]);

            pathAbove = new PathGeometry();
            PathFigure darkFigure = pathAbove.Figures.AddPathFigure();
            darkFigure.IsClosed = true;
            darkFigure.StartPoint = outerTopLeft;
            darkFigure.Segments.AddLineSegment(outerTopRight);
            darkFigure.Segments.AddLineSegment(innerTopRight);
            darkFigure.Segments.AddLineSegment(innerTopLeft);
            darkFigure.Segments.AddLineSegment(innerBottomLeft);
            darkFigure.Segments.AddLineSegment(outerBottomLeft);

            pathBelow = new PathGeometry();
            PathFigure lightFigure = pathBelow.Figures.AddPathFigure();
            lightFigure.IsClosed = true;
            lightFigure.StartPoint = outerBottomRight;
            lightFigure.Segments.AddLineSegment(outerBottomLeft);
            lightFigure.Segments.AddLineSegment(innerBottomLeft);
            lightFigure.Segments.AddLineSegment(innerBottomRight);
            lightFigure.Segments.AddLineSegment(innerTopRight);
            lightFigure.Segments.AddLineSegment(outerTopRight);
        }

        private static bool IsSimulated3dEffectStyle(AnnotationBorderStyle style)
        {
            return style == AnnotationBorderStyle.Inset || style == AnnotationBorderStyle.Beveled;
        }

        private void DrawSimulatedBorder3dEffect(FixedContentEditor editor)
        {
            PathGeometry pathAbove, pathBelow;
            this.CalculateBorder3dEffectGeometries(editor.Root.Size, out pathAbove, out pathBelow);

            using (editor.SaveGraphicProperties())
            {
                editor.GraphicProperties.IsFilled = true;
                editor.GraphicProperties.IsStroked = false;
                editor.GraphicProperties.StrokeThickness = 0;
                editor.GraphicProperties.FillColor = this.Border.Style == AnnotationBorderStyle.Inset ? FixedDocumentDefaults.Border3dEffectDarkColor : FixedDocumentDefaults.BeveledBorderLightColor;
                editor.DrawPath(pathAbove);
                editor.GraphicProperties.FillColor = this.Border.Style == AnnotationBorderStyle.Inset ? FixedDocumentDefaults.InsetBorderLightColor : FixedDocumentDefaults.Border3dEffectDarkColor;
                editor.DrawPath(pathBelow);
            }
        }

        private void ApplyBackgroundProperties(FixedContentEditor editor)
        {
            editor.GraphicProperties.IsFilled = this.AppearanceCharacteristics.Background != null;

            if (editor.GraphicProperties.IsFilled)
            {
                editor.GraphicProperties.FillColor = this.AppearanceCharacteristics.Background;
            }

            editor.GraphicProperties.IsStroked = this.HasBorderWidth && this.AppearanceCharacteristics.BorderColor != null;

            if (editor.GraphicProperties.IsStroked)
            {
                editor.GraphicProperties.StrokeColor = this.AppearanceCharacteristics.BorderColor;
                editor.GraphicProperties.StrokeThickness = this.Border.Width;

                if (this.Border.Style == AnnotationBorderStyle.Dashed)
                {
                    editor.GraphicProperties.StrokeDashArray = this.Border.DashArray;
                }
            }
        }
    }
}
