﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Navigation;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Collections
{
    /// <summary>
    /// Collection of all top-level <see cref="BookmarkItem"/> instances for the current level of the bookmarks hierarchy.
    /// </summary>
    public class BookmarksCollection : DocumentElementCollection<BookmarkItem, IFixedDocumentElement>, IFixedDocumentElement
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="BookmarksCollection" /> class.
        /// </summary>
        /// <param name="parent">The parent instance for this bookmarks collection. 
        /// This parent is applied on each <see cref="BookmarkItem"/> instance added 
        /// to the collection.</param>
        internal BookmarksCollection(IFixedDocumentElement parent) : base(parent)
        {
        }

        /// <summary>
        /// Gets the parent for the current <see cref="BookmarksCollection" /> instance.
        /// </summary>
        IFixedDocumentElement IFixedDocumentElement.Parent
        {
            get
            {
                return this.Owner;
            }
        }

        /// <summary>
        /// Iterates the bookmarks hierarchy in Depth First Search manner.
        /// The bookmarks are iterated in preorder traversal - root, left child, right child.
        /// </summary>
        internal IEnumerable<BookmarkItem> IterateBookmarksInDepthFirstSearch()
        {
            foreach (BookmarkItem bookmark in this)
            {
                yield return bookmark;

                foreach (BookmarkItem childBookmark in bookmark.Children.IterateBookmarksInDepthFirstSearch())
                {
                    yield return childBookmark;
                }
            }
        }

        /// <summary>
        /// Iterates the bookmarks hierarchy in Breadth First Search manner.
        /// </summary>
        internal IEnumerable<BookmarkItem> IterateBookmarksInBreadthFirstSearch()
        {
            Queue<BookmarkItem> bookmarksQueue = new Queue<BookmarkItem>(this);

            while (bookmarksQueue.Count > 0)
            {
                BookmarkItem bookmark = bookmarksQueue.Dequeue();

                foreach (BookmarkItem childBookmark in bookmark.Children)
                {
                    bookmarksQueue.Enqueue(childBookmark);
                }

                yield return bookmark;
            }
        }

        /// <summary>
        /// Verifies the validity of the <see cref="BookmarkItem"/> before it is inserted in the collection. 
        /// An item is assumed to be valid when has not already been added in this or other <see cref="BookmarksCollection" /> instance, 
        /// i.e. it's parent is <c>null</c>.
        /// </summary>
        /// <param name="item">The item to add.</param>
        /// <exception cref="ArgumentNullException">Thrown when the item is <c>null</c>.</exception>
        /// <exception cref="InvalidOperationException">Thrown when the item has already been added in this or other <see cref="BookmarksCollection" /> instance.</exception>
        protected override void VerifyDocumentElementOnInsert(BookmarkItem item)
        {
            Guard.ThrowExceptionIfNull(item, "item");

            if (item.Parent != null)
            {
                throw new InvalidOperationException("The element has already been associated with a collection.");
            }
        }
    }
}
