﻿using System;
using System.Collections;
using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.Model.Actions;
using Telerik.Windows.Documents.Fixed.Model.EventArguments;
using Telerik.Windows.Documents.Fixed.Model.Navigation;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Collections
{
    /// <summary>
    /// Represents collection of named destinations.
    /// </summary>
    public sealed class NamedDestinationsCollection : IEnumerable<NamedDestination>
    {
        private readonly object lockObject = new object();
        private readonly IDestinationContainer parent;
        private readonly Dictionary<string, NamedDestination> store;

        /// <summary>
        /// Initializes a new instance of the <see cref="NamedDestinationsCollection" /> class.
        /// </summary>
        /// <param name="parent">The parent.</param>
        public NamedDestinationsCollection(IDestinationContainer parent)
        {
            this.parent = parent;
            this.store = new Dictionary<string, NamedDestination>();
        }

        /// <summary>
        /// Gets the named destination associated with the specified name.
        /// </summary>
        /// <param name="name">The name of the named destination to get.</param>
        /// <returns>The named destination associated with the specified name.</returns>
        /// <exception cref="KeyNotFoundException">If the specified name is not found.</exception>
        public NamedDestination this[string name]
        {
            get
            {
                return this.store[name];
            }
        }

        /// <summary>
        /// Gets a collection containing the names in the collection.
        /// </summary>
        public IEnumerable<string> Names
        {
            get
            {
                return this.store.Keys;
            }
        }

        /// <summary>
        /// Gets the number of named destinations contained in the collection.
        /// </summary>
        public int Count
        {
            get
            {
                return this.store.Count;
            }
        }

        /// <summary>
        /// Adds a named destination with the specified name and value to the collection and returns it.
        /// </summary>
        /// <param name="name">The name of the named destination.</param>
        /// <param name="destination">The Destination for the named destination.</param>
        /// <returns>The created named destination.</returns>
        /// <exception cref="System.ArgumentNullException">name is null.</exception>
        /// <exception cref="System.ArgumentNullException">An element with the same name already exists in the collection.</exception>
        public NamedDestination Add(string name, Destination destination)
        {
            Guard.ThrowExceptionIfNull(name, "name");
            Guard.ThrowExceptionIfNull(destination, "destination");

            NamedDestination value = this.CreateDestination(name);
            value.Destination = destination;

            return value;
        }

        /// <summary>
        /// Adds a named destination with the specified name and value to the collection and returns it.
        /// </summary>
        /// <param name="name">The name of the named destination.</param>
        /// <param name="goToAction">The GoToAction for the named destination.</param>
        /// <returns>The created named destination.</returns>
        /// <exception cref="System.ArgumentNullException">name is null.</exception>
        /// <exception cref="System.ArgumentNullException">An element with the same name already exists in the collection.</exception>
        public NamedDestination Add(string name, GoToAction goToAction)
        {
            Guard.ThrowExceptionIfNull(name, "name");
            Guard.ThrowExceptionIfNull(goToAction, "goToAction");

            NamedDestination value = this.CreateDestination(name);
            value.GoToAction = goToAction;

            return value;
        }

        /// <summary>
        /// Removes all named destination from the collection.
        /// </summary>
        public void Clear()
        {
            foreach (NamedDestination destination in this.store.Values)
            {
                destination.NameChanged -= this.Destination_NameChanged;
            }

            this.store.Clear();
        }

        /// <summary>
        /// Determines whether the collection contains the specified name.
        /// </summary>
        /// <param name="name">The name to locate in the collection.</param>
        /// <returns>true if the collection contains an element with the specified name; otherwise, false</returns>
        /// <exception cref="System.ArgumentNullException">name is null</exception>
        public bool ContainsName(string name)
        {
            return this.store.ContainsKey(name);
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>A numerator for the collection</returns>
        public IEnumerator<NamedDestination> GetEnumerator()
        {
            return this.store.Values.GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.store.Values.GetEnumerator();
        }

        /// <summary>
        /// Removes the named destination with the specified name from the collection.
        /// </summary>
        /// <param name="name">The name of the named destination to remove</param>
        /// <returns>true if the named destination is successfully found and removed; otherwise, false. This
        /// method returns false if name is not found in the collection.</returns>
        /// <exception cref="System.ArgumentException">name is null.</exception>
        public bool Remove(string name)
        {
            Guard.ThrowExceptionIfNull(name, "name");

            NamedDestination destination;
            if (this.store.TryGetValue(name, out destination))
            {
                destination.NameChanged -= this.Destination_NameChanged;

                destination.SetParent(null);

                return this.store.Remove(name);
            }

            return false;
        }

        /// <summary>
        ///  Gets the named destiation associated with the specified name.
        /// </summary>
        /// <param name="name">The name of the destination to get.</param>
        /// <param name="value">When this method returns, contains the destination associated with the specified name
        /// if the name is found; otherwise, null. This parameter is passed uninitialized.</param>
        /// <returns>true if the collection contains a named destination with
        /// the specified name; otherwise, false.</returns>
        /// <exception cref="System.ArgumentException">name is null</exception>
        public bool TryGetValue(string name, out NamedDestination value)
        {
            return this.store.TryGetValue(name, out value);
        }

        /// <summary>
        /// Renames a named destination in the collection.
        /// </summary>
        /// <param name="oldName">The old name of the destination.</param>
        /// <param name="newName">The new name of the destination.</param>
        public void Rename(string oldName, string newName)
        {
            lock (this.lockObject)
            {
                NamedDestination destination = this[oldName];
                this.Remove(oldName);
                this.Add(newName, destination);
                destination.NameChanged -= this.Destination_NameChanged;
                destination.Name = newName;
                destination.NameChanged += this.Destination_NameChanged;
            }
        }

        internal void Add(NamedDestination namedDestination)
        {
            Guard.ThrowExceptionIfNull(namedDestination, "namedDestination");

            lock (this.lockObject)
            {
                string name = namedDestination.Name;
                NamedDestination existingDestination;
                if (this.store.TryGetValue(name, out existingDestination))
                {
                    if (existingDestination == namedDestination)
                    {
                        throw new InvalidOperationException("The same destination is already added.");
                    }

                    name = this.GenerateNewName(name);
                }

                namedDestination.Name = name;
                namedDestination.NameChanged += this.Destination_NameChanged;

                this.Add(name, namedDestination);
            }
        }

        private void Add(string name, NamedDestination value)
        {
            lock (this.lockObject)
            {
                this.VerifyDocumentElementOnInsert(value);

                value.SetParent(this.parent);
                this.store.Add(name, value);
            }
        }

        private string GenerateNewName(string name)
        {
            lock (this.lockObject)
            {
                int counter = 0;

                while (counter < int.MaxValue)
                {
                    string newName = string.Format("{0}.{1}", name, counter++);
                    if (!this.store.ContainsKey(newName))
                    {
                        return newName;
                    }
                }
            }

            throw new InvalidOperationException(string.Format("The named destinations with same name can not be more than {0}", int.MaxValue));
        }

        private NamedDestination CreateDestination(string name)
        {
            NamedDestination value = new NamedDestination();
            value.Name = name;
            value.NameChanged += this.Destination_NameChanged;
            this.Add(value.Name, value);

            return value;
        }

        private void Destination_NameChanged(object sender, NamedDestinationNameChangedEventArgs args)
        {
            this.Rename(args.OldValue, args.NewValue);
        }

        private void VerifyDocumentElementOnInsert(NamedDestination item)
        {
            Guard.ThrowExceptionIfNull(item, "item");

            if (item.Parent != null && item.Parent != this.parent)
            {
                throw new InvalidOperationException("The element has already been associated with a collection.");
            }
        }
    }
}
