﻿using Telerik.Windows.Documents.Utilities;
using Telerik.Windows.Documents.Fixed.Model.Common.Functions;
using System;
using Telerik.Windows.Documents.Core.Imaging;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.ColorSpaces
{
    internal class DeviceN : ColorSpaceBase, IMulticomponentColorSpace, IColorSpace
    {
        public override int ComponentCount
        {
            get
            {
                return this.Names.Length;
            }
        }

        public override string Name
        {
            get
            {
                return ColorSpaceNames.DeviceN;
            }
        }

        public string[] Names { get; set; }

        public ColorSpaceBase AlternateColorSpace { get; set; }

        public FunctionBase TintTransform { get; set; }

        public override PixelContainer GetPixels(IImageDescriptor image, bool applyMask)
        {
            return GetDeviceNPixels(image, this, applyMask);
        }

        public override double[] GetDefaultDecodeArray(int bitsPerComponent)
        {
            return GetDefaultDecodeArray(this);
        }

        internal static double[] GetDefaultDecodeArray(IMulticomponentColorSpace cs)
        {
            return GetDecodeArrayFromComponents(cs.ComponentCount);
        }

        internal static PixelContainer GetDeviceNPixels(IImageDescriptor image, IMulticomponentColorSpace cs, bool applyMask)
        {
            int componentsCount = cs.ComponentCount;
            int bitsPerComponent = image.BitsPerComponent;
            byte[] data = image.GetDecodedData();
            ImageBitReader reader = new ImageBitReader(data, bitsPerComponent, image.Width);
            PixelContainer pixels = PixelContainerFactory.GetDeviceNImagePixelContainer(image.Width, image.Height);
            double[] decode = GetDecodeArray(image);

            double[] yMin = new double[componentsCount];
            double[] yMax = new double[componentsCount];

            for (int i = 0; i < componentsCount; i++)
            {
                int minIndex = 2 * i;
                int maxIndex = minIndex + 1;
                yMin[i] = GetDecodeArrayComponent(decode, minIndex);
                yMax[i] = GetDecodeArrayComponent(decode, maxIndex);
            }

            double xMin = 0;
            double xMax = Math.Pow(2, bitsPerComponent) - 1;
            IMask mask = image.Mask;
            bool shouldMaskColorComponents = ShouldMaskColorComponents(mask, applyMask);
            Func<int[], Color> getPixel = (componentsArray) =>
            {
                double[] colorComponents = new double[componentsCount];

                for (int k = 0; k < componentsCount; k++)
                {
                    colorComponents[k] = FunctionBase.Interpolate(componentsArray[k], xMin, xMax, yMin[k], yMax[k]);
                }

                return cs.GetColor(colorComponents);
            };

            int[] components = new int[componentsCount];
            for (int i = 0; i < image.Height; i++)
            {
                for (int j = 0; j < image.Width; j++)
                {
                    for (int k = 0; k < componentsCount; k++)
                    {
                        components[k] = reader.Read();
                    }

                    pixels.Add(GetPixelFromComponents(components, getPixel, mask, shouldMaskColorComponents));
                }
            }

            return GetPixelArray(image, pixels, applyMask);
        }

        internal override ColorBase GetColor(double[] pars)
        {
            Guard.ThrowExceptionIfNull(pars, "pars");

            // TODO: try caching ColorBase result for specific pars to speed up this method. 
            double[] alternatePars = this.TintTransform.Execute(pars);

            return this.AlternateColorSpace.GetColor(alternatePars);
        }

        /// <summary>
        /// Called from <see cref="Indexed"/> to get color from lookup array.
        /// </summary>
        internal override ColorBase GetColorFromLookup(byte[] lookupData, int offset)
        {
            Guard.ThrowExceptionIfNull(lookupData, "lookupData");

            double[] colorants = new double[this.ComponentCount];
            for (int i = 0; i < this.ComponentCount; i++)
            {
                colorants[i] =  lookupData[offset + i];
            }

            return this.GetColor(colorants);
        }

        Color IMulticomponentColorSpace.GetColor(double[] components)
        {
            ColorBase color = this.GetColor(components);

            return color.ToColor();
        }
    }
}
