﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Fixed.Model.Data;

namespace Telerik.Windows.Documents.Fixed.Model.ColorSpaces
{
    /// <summary>
    /// Represents base class for gradient color.
    /// </summary>
    public abstract class Gradient : PatternColor, IGradient
    {
        private readonly GradientStopCollection gradientStops;

        /// <summary>
        /// Initializes a new instance of the <see cref="Gradient" /> class.
        /// </summary>
        public Gradient(Point startPoint, Point endPoint)
            : this(startPoint, endPoint, SimplePosition.Default)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Gradient" /> class.
        /// </summary>
        /// <param name="startPoint">The start point.</param>
        /// <param name="endPoint">The end point.</param>
        /// <param name="position">The position.</param>
        public Gradient(Point startPoint, Point endPoint, IPosition position)
        {
            this.StartPoint = startPoint;
            this.EndPoint = endPoint;
            this.Position = position;

            this.gradientStops = new GradientStopCollection();

            this.ExtendBefore = FixedDocumentDefaults.Extend;
            this.ExtendAfter = FixedDocumentDefaults.Extend;
        }

        /// <summary>
        /// Gets or sets the start point.
        /// </summary>
        /// <value>The start point.</value>
        public Point StartPoint { get; set; }

        /// <summary>
        /// Gets or sets the end point.
        /// </summary>
        /// <value>The end point.</value>
        public Point EndPoint { get; set; }

        /// <summary>
        /// Specifies if the gradient should extends before the start point.
        /// </summary>
        /// <value>The extend before.</value>
        public bool ExtendBefore { get; set; }

        /// <summary>
        /// Specifies if the gradient should extends after the end point.
        /// </summary>
        /// <value>True to extend.</value>
        public bool ExtendAfter { get; set; }

        /// <summary>
        /// Gets or sets the background color.
        /// </summary>
        /// <value>The background color.</value>
        public ColorBase Background { get; set; }

        /// <summary>
        /// Gets the gradient stops.
        /// </summary>
        /// <value>The gradient stops.</value>
        public GradientStopCollection GradientStops
        {
            get
            {
                return this.gradientStops;
            }
        }

        /// <summary>
        /// Gets or sets the position.
        /// </summary>
        /// <value>The position.</value>
        public override Data.IPosition Position { get; set; }

        internal override PatternType PatternType
        {
            get
            {
                return PatternType.Gradient;
            }
        }

        internal abstract GradientType GradientType
        {
            get;
        }

        IEnumerable<IGradientStop> IGradient.GradientStops
        {
            get
            {
                foreach (IGradientStop stop in this.GradientStops)
                {
                    yield return stop;
                }
            }
        }

        Rect? IGradient.BoundingBox
        {
            get
            {
                return null;
            }
        }

        Core.Imaging.Color IGradient.StartColor
        {
            get
            {
                GradientStop firstStop = this.GradientStops[0];

                return firstStop.Color.ToColor();
            }
        }

        Core.Imaging.Color IGradient.EndColor
        {
            get
            {
                GradientStop endStop = this.GradientStops[this.GradientStops.Count - 1];

                return endStop.Color.ToColor();
            }
        }

        Core.Imaging.Color? IGradient.Background
        {
            get
            {
                Core.Imaging.Color? color = null;

                if (this.Background != null)
                {
                    color = this.Background.ToColor();
                }

                return color;
            }
        }

        /// <summary>
        /// Equalses the specified other.
        /// </summary>
        /// <param name="other">The other.</param>
        /// <returns></returns>
        public override bool Equals(ColorBase other)
        {
            Gradient gradient = other as Gradient;
            if (gradient == null)
            {
                return false;
            }

            if (this.GradientStops.Count != gradient.GradientStops.Count)
            {
                return false;
            }

            for (int i = 0; i < this.GradientStops.Count; i++)
            {
                if (!this.GradientStops[i].Equals(other))
                {
                    return false;
                }
            }

            return this.Background == gradient.Background &&
                this.EndPoint == gradient.EndPoint &&
                this.ExtendAfter == gradient.ExtendAfter &&
                this.ExtendBefore == gradient.ExtendBefore &&
                this.StartPoint == gradient.StartPoint;
        }
    }
}
