﻿using Telerik.Windows.Documents.Utilities;
using Telerik.Windows.Documents.Core.Imaging;
using System;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.Common.Functions;

namespace Telerik.Windows.Documents.Fixed.Model.ColorSpaces
{
    internal class Indexed : ColorSpaceBase, IIndexedColorSpace
    {
        public override int ComponentCount
        {
            get { return 1; }
        }

        public override string Name
        {
            get
            {
                return ColorSpaceNames.Indexed;
            }
        }

        public ColorBase DefaultColor
        {
            get
            {
                return new RgbColor(0, 0, 0);
            }
        }

        public ColorSpaceBase Base { get; set; }

        public int HiVal { get; set; }

        public byte[] Lookup { get; set; }

        public Color GetColor(int index)
        {
            ColorBase color = this.GetColorAt(index);

            return color.ToColor();
        }

        public Color[] GetAllColors()
        {
            return GetAllColors(this);
        }

        public override double[] GetDefaultDecodeArray(int bitsPerComponent)
        {
            return GetIndexedDecodeArray(bitsPerComponent);
        }

        public override PixelContainer GetPixels(IImageDescriptor image, bool applyMask)
        {
            return GetIndexedPixels(image, this, applyMask);
        }

        internal static PixelContainer GetIndexedPixels(IImageDescriptor image, IIndexedColorSpace cs, bool applyMask)
        {
            int bitsPerComponent = image.BitsPerComponent;
            byte[] data = image.GetDecodedData();
            ImageBitReader reader = new ImageBitReader(data, bitsPerComponent, image.Width);
            bool hasMask = HasMask(image);
            PixelContainer pixels = PixelContainerFactory.GetIndexedImagePixelContainer(hasMask, image.Width, image.Height, cs.GetAllColors());

            double[] decode = GetDecodeArray(image);
            double yMin = GetDecodeArrayComponent(decode, 0);
            double yMax = GetDecodeArrayComponent(decode, 1);

            double xMin = 0;
            double xMax = Math.Pow(2, bitsPerComponent) - 1;

            IMask mask = image.Mask;
            bool shouldMaskColorComponents = ShouldMaskColorComponents(mask, applyMask);
            Func<int[], Color> getPixel = (componentsArray) => 
            {
                return cs.GetColor((int)FunctionBase.Interpolate(componentsArray[0], xMin, xMax, yMin, yMax));
            };

            int[] components = new int[1];
            for (int i = 0; i < image.Height; i++)
            {
                for (int j = 0; j < image.Width; j++)
                {
                    components[0] = reader.Read();

                    pixels.Add(GetPixelFromComponents(components, getPixel, mask, shouldMaskColorComponents));
                }
            }

            return GetPixelArray(image, pixels, applyMask);
        }

        internal static double[] GetIndexedDecodeArray(int bitsPerComponent)
        {
            double[] decode = new double[2];
            decode[1] = Math.Pow(2, bitsPerComponent) - 1;

            return decode;
        }

        internal static Color[] GetAllColors(IIndexedColorSpace cs)
        {
            Color[] result = new Color[cs.HiVal + 1];

            for (int i = 0; i <= cs.HiVal; i++)
            {
                Color color = cs.GetColor(i);
                result[i] = color;
            }

            return result;
        }

        internal override ColorBase GetColor(double[] pars)
        {
            Guard.ThrowExceptionIfNull(pars, "pars");

            int index = ((int)pars[0]);

            return this.GetColorAt(index);
        }

        /// <summary>
        /// Called from <see cref="Indexed"/> to get color from lookup array.
        /// </summary>
        internal override ColorBase GetColorFromLookup(byte[] lookupData, int offset)
        {
            throw new NotSupportedException("This method is not supported.");
        }

        private ColorBase GetColorAt(int index)
        {
            if (this.Base == null || this.Lookup == null)
            {
                return this.DefaultColor;
            }

            if (index < 0)
            {
                index = 0;
            }

            if (index > this.HiVal)
            {
                index = this.HiVal;
            }

            ColorSpaceBase colorSpaceBase = this.Base;

            IccBased iccBasedColorSpace = colorSpaceBase as IccBased;
            if (iccBasedColorSpace != null)
            {
                colorSpaceBase = iccBasedColorSpace.GetCachedAlternateColorSpace();
            }
            
            return colorSpaceBase.GetColorFromLookup(this.Lookup, index * colorSpaceBase.ComponentCount);
        }
    }
}