﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Core.Imaging;

namespace Telerik.Windows.Documents.Fixed.Model.ColorSpaces
{
    /// <summary>
    /// Represents RGB color.
    /// </summary>
    public class RgbColor : SimpleColor
    {
        private static readonly DeviceRgb DeviceRgbColorSpace = new DeviceRgb();
        internal const byte DefaultAlphaValue = 255;

        /// <summary>
        /// Initializes a new instance of the <see cref="RgbColor" /> class.
        /// </summary>
        public RgbColor()
        {
            this.A = DefaultAlphaValue;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="RgbColor" /> class.
        /// </summary>
        /// <param name="r">The red component.</param>
        /// <param name="g">The green component.</param>
        /// <param name="b">The blue component.</param>
        public RgbColor(byte r, byte g, byte b)
            : this(DefaultAlphaValue, r, g, b)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="RgbColor" /> class.
        /// </summary>
        /// <param name="a">The alpha component.</param>
        /// <param name="r">The red component.</param>
        /// <param name="g">The green component.</param>
        /// <param name="b">The blue component.</param>
        public RgbColor(byte a, byte r, byte g, byte b)
        {
            this.A = a;
            this.R = r;
            this.G = g;
            this.B = b;
        }

        internal RgbColor(RgbColor other)
            : this(other.A, other.R, other.G, other.B)
        {
        }

        internal RgbColor(Color color)
            : this(color.A, color.R, color.G, color.B)
        {
        }

        /// <summary>
        /// Gets or sets the alpha component.
        /// </summary>
        /// <value>The alpha component.</value>
        public byte A { get; set; }

        /// <summary>
        /// Gets or sets the red component.
        /// </summary>
        /// <value>The red component.</value>
        public byte R { get; set; }

        /// <summary>
        /// Gets or sets the green component.
        /// </summary>
        /// <value>The green component.</value>
        public byte G { get; set; }

        /// <summary>
        /// Gets or sets the blue component.
        /// </summary>
        /// <value>The blue component.</value>
        public byte B { get; set; }

        internal override ColorSpaceBase ColorSpace
        {
            get
            {
                return RgbColor.DeviceRgbColorSpace;
            }
        }

        internal override bool IsTransparent
        {
            get
            {
                return this.A == 0;
            }
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return ((int)this.A << 24) | ((int)this.R << 16) | ((int)this.G << 8) | this.B;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            if (obj == null)
            {
                return false;
            }

            return this.Equals(obj as ColorBase);
        }

        /// <summary>
        /// Equalses the specified other.
        /// </summary>
        /// <param name="other">The other.</param>
        /// <returns></returns>
        public override bool Equals(ColorBase other)
        {
            if (other == null)
            {
                return false;
            }

            RgbColor color = other as RgbColor;
            return this.R == color.R
                && this.G == color.G
                && this.B == color.B
                && this.A == color.A;
        }

        /// <summary>
        /// Returns the text representation of the color.
        /// </summary>
        /// <returns>Returns the text representation of the color.</returns>
        public override string ToString()
        {
            return string.Format("<A={0} R={1} G={2} B={3}>", this.A, this.R, this.G, this.B);
        }

        internal override int[] GetColorComponents()
        {
            return new int[] { this.R, this.G, this.B };
        }

        internal override Color ToColor()
        {
            return new Color(this.A, this.R, this.G, this.B);
        }
    }
}
