﻿using System;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Common.Functions
{
    internal abstract class FunctionBase
    {
        private double[] domain;
        private double[] range;

        protected FunctionBase(double[] domain, double[] range)
        {
            Guard.ThrowExceptionIfNull(domain, "domain");

            if (this.IsRangeRequired)
            {
                Guard.ThrowExceptionIfNull(range, "range");
            }

            this.domain = domain;
            this.range = range;
        }

        public abstract FunctionType FunctionType { get; }

        public double[] Domain
        {
            get
            {
                return this.domain;
            }
        }

        public double[] Range
        {
            get
            {
                return this.range;
            }
        }

        public int OutputValuesCount
        {
            get
            {
                return this.Range.Length / 2;
            }
        }

        protected abstract bool IsRangeRequired { get; }

        public static double Interpolate(double x, double xMin, double xMax, double yMin, double yMax)
        {
            return yMin + ((x - xMin) * ((yMax - yMin) / (xMax - xMin)));
        }

        public static double ClipValue(double value, double min, double max)
        {
            if (value < min)
            {
                return min;
            }

            if (value > max)
            {
                return max;
            }

            return value;
        }

        public double[] Execute(double[] inputValues)
        {
            Guard.ThrowExceptionIfNull(inputValues, "inputValues");

            double[] clippedInputValues = this.ClipInputValues(inputValues);
            double[] outputs = this.ExecuteOverride(clippedInputValues);
            double[] clippedOutputValues = this.ClipOutputValues(outputs);

            return clippedOutputValues;
        }

        public virtual byte[] GetFunctionData()
        {
            throw new InvalidCastException(string.Format("Function type {0} cannot provide data.", this.FunctionType));
        }

        protected abstract double[] ExecuteOverride(double[] clippedInputValues);

        private double[] ClipInputValues(double[] inputValues)
        {
            double[] result = new double[inputValues.Length];

            for (int i = 0; i < inputValues.Length; i++)
            {
                double min = this.Domain[2 * i];
                double max = this.Domain[(2 * i) + 1];

                result[i] = ClipValue(inputValues[i], min, max);
            }

            return result;
        }

        private double[] ClipOutputValues(double[] outputValues)
        {
            if (this.Range == null)
            {
                return outputValues; 
            }

            double[] result = new double[outputValues.Length];

            for (int i = 0; i < outputValues.Length; i++)
            {
                double min = this.Range[2 * i];
                double max = this.Range[(2 * i) + 1];
                result[i] = ClipValue(outputValues[i], min, max);
            }

            return result;
        }
    }
}
