﻿using System;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Common.Functions
{
    internal class SampledFunction : FunctionBase
    {
        private readonly int[] size;
        private readonly int bitsPerSample;
        private readonly double[] encode;
        private readonly double[] decode;
        private int[][] samples;
        private byte[] data;

        public SampledFunction(int[][] samples, int[] size, int bitsPerSample, double[] domain, double[] range, double[] encode, double[] decode)
            : base(domain, range)
        {
            Guard.ThrowExceptionIfNull(samples, "samples");
            GuardRequiredInitializationValues(size, bitsPerSample);

            this.samples = samples;
            this.size = size;
            this.bitsPerSample = bitsPerSample;
            this.encode = encode;
            this.decode = decode;
        }

        public SampledFunction(byte[] data, int[] size, int bitsPerSample, double[] domain, double[] range, double[] encode, double[] decode)
            : base(domain, range)
        {
            GuardRequiredInitializationValues(size, bitsPerSample);
            
            this.data = data ?? new byte[0];
            this.size = size;
            this.bitsPerSample = bitsPerSample;
            this.encode = encode;
            this.decode = decode;
        }

        public override FunctionType FunctionType
        {
            get
            {
                return FunctionType.Sampled;
            }
        }

        public double[] Encode
        {
            get
            {
                return this.encode;
            }
        }

        public double[] Decode
        {
            get
            {
                return this.decode;
            }
        }

        public int[] Size
        {
            get
            {
                return this.size;
            }
        }

        public int BitsPerSample
        {
            get
            {
                return this.bitsPerSample;
            }
        }

        protected override bool IsRangeRequired
        {
            get
            {
                return true;
            }
        }

        public override byte[] GetFunctionData()
        {
            this.EnsureData();

            return this.data;
        }

        protected override double[] ExecuteOverride(double[] clippedInputValues)
        {
            int[] flooredEncodedValues = new int[clippedInputValues.Length];
            double[] interpolationCoeficients = new double[clippedInputValues.Length];
            int nonZeroInterpolationCoeficientsCount = 0;

            for (int i = 0; i < clippedInputValues.Length; i++)
            {
                double minD = this.Domain[2 * i];
                double maxD = this.Domain[(2 * i) + 1];
                double minE = this.GetEncode(2 * i);
                double maxE = this.GetEncode((2 * i) + 1);
                int size = this.Size[i];

                double clippedInputValue = clippedInputValues[i];
                double encodedValue = Interpolate(clippedInputValue, minD, maxD, minE, maxE);
                double clippedEncodedValue = ClipValue(encodedValue, 0, size - 1);
                int flooredEncodedValue = (int)clippedEncodedValue;
                flooredEncodedValues[i] = flooredEncodedValue;
                double nonIntegerPart = clippedEncodedValue - flooredEncodedValue;

                if (IsNonZeroInterpolationCoeficient(nonIntegerPart))
                {
                    interpolationCoeficients[i] = nonIntegerPart;
                    nonZeroInterpolationCoeficientsCount++;
                }
            }

            double[] result = this.CalculateMultidimensionalLinearInterpolationResult(flooredEncodedValues, interpolationCoeficients, nonZeroInterpolationCoeficientsCount);

            return result;
        }

        private static bool IsNonZeroInterpolationCoeficient(double interpolationCoeficient)
        {
            return interpolationCoeficient != 0;
        }

        private static void GuardSupportedDimensionsCount(int nonZeroInterpolationCoeficientsCount)
        {
            if (nonZeroInterpolationCoeficientsCount > 31)
            {
                // It is higly unlikely to have function with 30 input arguments. Usually the arguments count are between 1 and 4, so this restriction for 31 arguments should not be met.
                throw new NotSupportedException(string.Format("Functions with input dimension bigger than 31 are not supported."));
            }
        }

        private static void GuardRequiredInitializationValues(int[] size, int bitsPerSample)
        {
            Guard.ThrowExceptionIfNull(size, "size");

            switch (bitsPerSample)
            {
                case 1:
                case 2:
                case 4:
                case 8:
                case 12:
                case 16:
                case 24:
                case 32:
                    break;
                default:
                    throw new ArgumentException(string.Format("{0} is not a valid bitsPerSample value!", bitsPerSample));
            }
        }

        private double GetEncode(int index)
        {
            double encode;

            if (this.Encode == null)
            {
                bool isEven = (index & 1) == 0;
                
                if (isEven)
                {
                    encode = 0;
                }
                else
                {
                    int sizeIndex = index / 2;
                    int size = this.Size[sizeIndex];
                    encode = size - 1;
                }
            }
            else
            {
                encode = this.Encode[index];
            }

            return encode;
        }

        private double GetDecode(int index)
        {
            double[] array = this.Decode ?? this.Range;
            double decode = array[index];

            return decode;
        }

        private double[] CalculateMultidimensionalLinearInterpolationResult(int[] flooredEncodedValues, double[] interpolationCoeficients, int nonZeroInterpolationCoeficientsCount)
        {
            GuardSupportedDimensionsCount(nonZeroInterpolationCoeficientsCount);

            this.EnsureSamples();
            double[] result = new double[this.OutputValuesCount];
            int cubeVerticesCount = 1 << nonZeroInterpolationCoeficientsCount;

            for (int cubeVertex = 0; cubeVertex < cubeVerticesCount; cubeVertex++)
            {
                int[] inputValues = new int[flooredEncodedValues.Length];
                double vertexCoeficient = 1;

                for (int inputIndex = 0, vertexDimensionsBits = 0; inputIndex < inputValues.Length; inputIndex++)
                {
                    int inputValue = flooredEncodedValues[inputIndex];
                    double interpolationCoeficient = interpolationCoeficients[inputIndex];

                    if (IsNonZeroInterpolationCoeficient(interpolationCoeficient))
                    {
                        bool isOnOppositeSide = ((cubeVertex >> vertexDimensionsBits) & 1) == 1;

                        if (isOnOppositeSide)
                        {
                            inputValue += 1;
                            vertexCoeficient *= interpolationCoeficient;
                        }
                        else
                        {
                            vertexCoeficient *= (1 - interpolationCoeficient);
                        }

                        vertexDimensionsBits++;
                    }

                    inputValues[inputIndex] = inputValue;
                }

                int index = this.GetSampleIndex(inputValues);                
                int[] outputValues = this.samples[index];

                for (int outputIndex = 0; outputIndex < outputValues.Length; outputIndex++)
                {
                    result[outputIndex] = result[outputIndex] + vertexCoeficient * outputValues[outputIndex];
                }
            }

            this.DecodeResult(result);

            return result;
        }

        private void EnsureSamples()
        {
            if (this.samples == null)
            {
                BitReader reader = new BitReader(this.data, this.BitsPerSample);

                int samplesCount = 1;
                for (int i = 0; i < this.Size.Length; i++)
                {
                    int s = this.Size[i];
                    samplesCount *= s;
                }

                this.samples = new int[samplesCount][];
                int outputValuesCount = this.OutputValuesCount;

                for (int i = 0; i < samplesCount; i++)
                {
                    this.samples[i] = new int[outputValuesCount];

                    for (int j = 0; j < outputValuesCount; j++)
                    {
                        this.samples[i][j] = reader.Read();
                    }
                }
            }
        }

        private void EnsureData()
        {
            if (this.data == null)
            {
                int samplesCount = this.samples.Length;
                int outputValuesCount = this.Range.Length / 2;
                int numberOfRecords = samplesCount * outputValuesCount;
                BitWriter bitWriter = new BitWriter(numberOfRecords, this.BitsPerSample);

                for (int i = 0; i < samplesCount; i++)
                {
                    int[] sample = this.samples[i];

                    for (int j = 0; j < outputValuesCount; j++)
                    {
                        int record = sample[j];
                        bitWriter.Write(record);
                    }
                }

                this.data = bitWriter.ResultBits;
            }
        }

        private void DecodeResult(double[] result)
        {
            int maxN = (1 << this.BitsPerSample) - 1;

            for (int i = 0; i < result.Length; i++)
            {
                double minD = this.GetDecode(2 * i);
                double maxD = this.GetDecode((2 * i) + 1);
                result[i] = FunctionBase.Interpolate(result[i], 0, maxN, minD, maxD);
            }
        }

        private int GetSampleIndex(int[] encodedValues)
        {
            int sum = encodedValues[0];
            int sizeFromPreviousDimensions = 1;

            for (int i = 1; i < encodedValues.Length; i++)
            {
                int s = this.Size[i - 1];
                sizeFromPreviousDimensions *= s;

                sum += sizeFromPreviousDimensions * encodedValues[i];
            }

            return sum;
        }
    }
}
