﻿using System;
using System.Linq;
using System.Windows;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Data
{
    /// <summary>
    /// Represents simple position.
    /// </summary>
    public class SimplePosition : IPosition
    {
        private readonly static Point defaultRotateCenter = new Point(0, 0);
        private readonly static Point defaultScaleCenter = new Point(0, 0);

        private double offsetX;
        private double offsetY;
        private double scaleX;
        private double scaleY;
        private double angle;
        private Point rotateCenter;
        private Point scaleCenter;

        /// <summary>
        /// Initializes a new instance of the <see cref="SimplePosition" /> class.
        /// </summary>
        public SimplePosition()
        {
            this.InitializeDefaults();
        }

        /// <summary>
        /// Gets the default.
        /// </summary>
        /// <value>The default.</value>
        public static SimplePosition Default
        {
            get
            {
                return new SimplePosition();
            }
        }

        /// <summary>
        /// Gets or sets the matrix.
        /// </summary>
        /// <value>The matrix.</value>
        public Matrix Matrix
        {
            get
            { 
                var matrix = new Matrix(); 
                matrix = matrix.ScaleMatrixAt(this.scaleX, this.scaleY, this.scaleCenter.X, this.scaleCenter.Y);
                matrix = matrix.RotateMatrixAt(this.angle, this.rotateCenter.X, this.rotateCenter.Y);
                matrix = matrix.TranslateMatrix(this.offsetX, this.offsetY);
                
                return matrix;
            }
        }

        /// <summary>
        /// Appends the specified scale vector.
        /// </summary>
        /// <param name="scaleX">The scale X.</param>
        /// <param name="scaleY">The scale Y.</param>
        public void Scale(double scaleX, double scaleY)
        {
            this.scaleX = scaleX;
            this.scaleY = scaleY;
            this.scaleCenter = defaultScaleCenter;
        }

        /// <summary>
        /// Applies a scale by the specified amount about the specified point.
        /// </summary>
        /// <param name="scaleX">The scale X.</param>
        /// <param name="scaleY">The scale Y.</param>
        /// <param name="centerX">The center X.</param>
        /// <param name="centerY">The center Y.</param>
        public void ScaleAt(double scaleX, double scaleY, double centerX, double centerY)
        {
            this.scaleX = scaleX;
            this.scaleY = scaleY;
            this.scaleCenter = new Point(centerX, centerY);
        }  

        /// <summary>
        /// Applies a rotation of the specified angle about the origin.
        /// </summary>
        /// <param name="angle">The angle.</param>
        public void Rotate(double angle)
        {
            this.angle = angle;
            this.rotateCenter = defaultRotateCenter;
        }

        /// <summary>
        /// Rotates this matrix about the specified point.
        /// </summary>
        /// <param name="angle">The angle.</param>
        /// <param name="centerX">The center X.</param>
        /// <param name="centerY">The center Y.</param>
        public void RotateAt(double angle, double centerX, double centerY)
        {
            this.angle = angle;
            this.rotateCenter = new Point(centerX, centerY);
        }

        /// <summary>
        /// Appends a translation of the specified offsets.
        /// </summary>
        /// <param name="offsetX">The offset X.</param>
        /// <param name="offsetY">The offset Y.</param>
        public void Translate(double offsetX, double offsetY)
        {
            this.offsetX = offsetX;
            this.offsetY = offsetY;
        }

        /// <summary>
        /// Clears this instance.
        /// </summary>
        public void Clear()
        {
            this.InitializeDefaults();
        }

        /// <summary>
        /// Clones this instance.
        /// </summary>
        /// <returns></returns>
        public IPosition Clone()
        {
            SimplePosition position = new SimplePosition();
            position.Translate(this.offsetX, this.offsetY);
            position.ScaleAt(this.scaleX, this.scaleY, this.scaleCenter.X, this.scaleCenter.Y);
            position.RotateAt(this.angle, this.rotateCenter.X, this.rotateCenter.Y);

            return position;
        }
         
        private void InitializeDefaults()
        {
            this.offsetX = 0;
            this.offsetY = 0;
            this.scaleX = 1;
            this.scaleY = 1;
            this.angle = 0;
            this.rotateCenter = defaultRotateCenter;
            this.scaleCenter = defaultScaleCenter;
        }
    }
}
