﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Cryptography;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.DigitalSignatures
{
    /// <summary>
    /// Defines a hash algorithm factory.
    /// </summary>
    public static class HashAlgorithmsManager
    {
        private static Dictionary<string, HashAlgorithm> store = new Dictionary<string, HashAlgorithm>();

        static HashAlgorithmsManager()
        {
            RegisterHashAlgorithm(SignatureOids.SHA1, new SHA1Managed());
            RegisterHashAlgorithm(SignatureOids.SHA256, new SHA256Managed());
            RegisterHashAlgorithm(SignatureOids.SHA384, new SHA384Managed());
            RegisterHashAlgorithm(SignatureOids.SHA512, new SHA512Managed());
#if !NETCORE && !NETSTANDARD
            RegisterHashAlgorithm(SignatureOids.MD5, new MD5Cng());
#endif
            RegisterHashAlgorithm(SignatureOids.MD5RSA, new MD5CryptoServiceProvider());

        }

        /// <summary>
        /// Creates an instance of a predefined hash algorithm class from an Oid.
        /// </summary>
        /// <param name="algorithmOid">The algorithm oid.</param>
        /// <returns></returns>
        internal static HashAlgorithm CreateHashAlgorithm(Oid algorithmOid)
        {
            Guard.ThrowExceptionIfNull(algorithmOid, "algorithmOid");

            HashAlgorithm result;
            if (!store.TryGetValue(algorithmOid.Value, out result))
            {
                Oid mainAlgorithmOid = SignatureOids.GetMainAlgorithmOid(algorithmOid);
                if (mainAlgorithmOid != null)
                {
                    store.TryGetValue(mainAlgorithmOid.Value, out result);
                }
            }

            return result;
        }

        /// <summary>
        /// Registers a hash algorithm.
        /// </summary>
        /// <param name="oid">The oid.</param>
        /// <param name="hashAlgorithm">The hash algorithm.</param>
        public static void RegisterHashAlgorithm(Oid oid, HashAlgorithm hashAlgorithm)
        {
            Guard.ThrowExceptionIfNull(oid, "Oid");
            Guard.ThrowExceptionIfNullOrEmpty(oid.Value, "oid.Value");
            Guard.ThrowExceptionIfNull(hashAlgorithm, "hashAlgorithm");

            store.Add(oid.Value, hashAlgorithm);
        }

        /// <summary>
        /// Determines whether the <see cref="HashAlgorithmsManager"/> contains a definition of a hash algorithm for the specificied Oid.
        /// </summary>
        /// <param name="oid">The oid.</param>
        /// <returns></returns>
        public static bool HasRegisteredHashAlgorithm(Oid oid)
        {
            Guard.ThrowExceptionIfNull(oid, "oid");

            bool hasRegisteredHashAlgorithm = store.ContainsKey(oid.Value);

            return hasRegisteredHashAlgorithm;
        }

        /// <summary>
        /// Unregisters a hash algorithm.
        /// </summary>
        /// <param name="oid">The oid.</param>
        public static void UnregisterHashAlgorithm(Oid oid)
        {
            if (HasRegisteredHashAlgorithm(oid))
            {
                store.Remove(oid.Value);
            }
        }
    }
}
