﻿using System;
using System.IO;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using Org.BouncyCastle.Asn1;

namespace Telerik.Windows.Documents.Fixed.Model.DigitalSignatures
{
    /// <summary>
    /// Defines a class used for handling the validation of digital signature in a PKCS#1 cryptographic message syntax.
    /// </summary>
    public class Pkcs1 : SignatureValidationHandlerBase
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="Pkcs1" /> class.
        /// </summary>
        public Pkcs1()
        {
        }

        /// <summary>
        /// When overridden in a derived class, validates the digital signature.
        /// </summary>
        /// <param name="dataProperties"></param>
        /// <param name="validationProperties"></param>
        /// <returns>
        /// An instance of the <see cref="SignatureValidationResult" /> class describing the result from the validation.
        /// </returns>
        protected override SignatureValidationResult ValidateOverride(SignatureDataProperties dataProperties, SignatureValidationProperties validationProperties)
        {
            SignatureValidationResultBuilder validationResultBuilder = new SignatureValidationResultBuilder();

            if (dataProperties.Certificates != null && dataProperties.Certificates.Count > 0)
            {
                validationResultBuilder.BuildCertificates(dataProperties.Certificates);

                X509Certificate2 cert = this.GetSigningCertificate(dataProperties.Certificates);
                RSACryptoServiceProvider rsa = (RSACryptoServiceProvider)cert.PublicKey.Key;

                Asn1InputStream asn1InputStream = new Asn1InputStream(new MemoryStream(dataProperties.Contents));
                Asn1OctetString asnOctetString = (Asn1OctetString)asn1InputStream.ReadObject();
                byte[] digest = asnOctetString.GetOctets();

                byte[] byteArrayCompositionForHash = this.GetByteArrayCompositionForHash();

                Oid hashAlgorithmOid = SignatureOids.GetMainAlgorithmOid(cert.SignatureAlgorithm);
                this.SetResultHashAlgorithm(validationResultBuilder, hashAlgorithmOid);

                bool isHashValid = rsa.VerifyData(byteArrayCompositionForHash, hashAlgorithmOid.Value, digest);

                validationResultBuilder.BuildIsDocumentModified(!isHashValid);
                this.ValidateCertificates(validationResultBuilder, dataProperties.Certificates);
            }
            else
            {
                throw new NotSupportedException("No certificates are specified in the signature \"Cert\" property.");
            }

            validationResultBuilder.BuildSignerInformation(dataProperties.Name);

            SignatureValidationResult validationResult = validationResultBuilder.GetResult();
            return validationResult;
        }
    }
}
