﻿using System;
using System.Linq;
using System.Security.Cryptography;
using System.Security.Cryptography.Pkcs;

namespace Telerik.Windows.Documents.Fixed.Model.DigitalSignatures
{
    /// <summary>
    /// Defines a class used for handling the validation of digital signature in a PKCS#7 cryptographic message syntax with detached signature.
    /// </summary>
    public class Pkcs7Detached : SignatureValidationHandlerBase
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="Pkcs7Detached" /> class.
        /// </summary>
        public Pkcs7Detached()
        {
        }

        /// <summary>
        /// When overridden in a derived class, validates the digital signature.
        /// </summary>
        /// <param name="dataProperties"></param>
        /// <param name="validationProperties"></param>
        /// <returns>
        /// An instance of the <see cref="SignatureValidationResult" /> class describing the result from the validation.
        /// </returns>
        protected override SignatureValidationResult ValidateOverride(SignatureDataProperties dataProperties, SignatureValidationProperties validationProperties)
        {
            SignatureValidationResultBuilder validationResultBuilder = new SignatureValidationResultBuilder();

            byte[] byteArrayCompositionForHash = this.GetByteArrayCompositionForHash();

            ContentInfo contentInfo = new ContentInfo(byteArrayCompositionForHash);

            SignedCms signedCms = new SignedCms(contentInfo, true);
            signedCms.Decode(dataProperties.Contents);

            this.ValidateCertificates(validationResultBuilder, signedCms.Certificates);
            validationResultBuilder.BuildCertificates(signedCms.Certificates);

            if (signedCms.SignerInfos.Count == 1)
            {
                SignerInfo signerInfo = signedCms.SignerInfos[0];

                CryptographicAttributeObject rawMessageDigest = null;
                foreach (var attribute in signerInfo.SignedAttributes)
                {
                    if(attribute.Oid.Value == SignatureOids.MESSAGE_DIGEST.Value)
                    {
                        rawMessageDigest = attribute;
                    }
                }

                if(rawMessageDigest != null)
                {
                    Pkcs9MessageDigest pkcs9MessageDigest = (Pkcs9MessageDigest)rawMessageDigest.Values[0];
                    byte[] messageDigest = pkcs9MessageDigest.MessageDigest;

                    bool isHashValid = this.ValidateHash(messageDigest, signerInfo.DigestAlgorithm);
                    validationResultBuilder.BuildIsDocumentModified(!isHashValid);
                }
                else
                {
                    try
                    {
                        signerInfo.CheckSignature(true);
                        validationResultBuilder.BuildIsDocumentModified(false);
                    }
                    catch
                    {
                        validationResultBuilder.BuildIsDocumentModified(true);
                    }
                }

                this.SetResultHashAlgorithm(validationResultBuilder, signerInfo.DigestAlgorithm);
            }
            else
            {
                throw new NotSupportedException("There is more then one signer specified for a single signature.");
            }

            validationResultBuilder.BuildSignerInformation(dataProperties.Name);

            SignatureValidationResult validationResult = validationResultBuilder.GetResult();
            return validationResult;
        }
    }
}
