﻿using System;
using System.Security.Cryptography;
using System.Security.Cryptography.Pkcs;
using System.Security.Cryptography.X509Certificates;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.DigitalSignatures
{
    /// <summary>
    /// Defines a class used for signing and verifying a digital signature.
    /// </summary>
    public class Signature
    {
        private X509Certificate2 certificate;

        /// <summary>
        /// Initializes a new instance of the <see cref="Signature" /> class.
        /// </summary>
        /// <param name="certificate">The certificate used for signing the signature.</param>
        public Signature(X509Certificate2 certificate) : this()
        {
            Guard.ThrowExceptionIfNull(certificate, "certificate");

            this.certificate = certificate;

            this.BuildInitalSignState();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Signature" /> class.
        /// </summary>
        internal Signature()
        {
            this.Properties = new SignatureDataProperties();
        }

        /// <summary>
        /// Gets or sets the signature data properties.
        /// </summary>
        /// <value>The signature data properties.</value>
        public SignatureDataProperties Properties { get; set; }

        /// <summary>
        /// Gets a value indicating whether the signature supports validation. Only imported signatures supports validation.
        /// </summary>
        /// <value><c>true</c> if the signature is imported; otherwise, <c>false</c>.</value>
        public bool SupportsValidation
        {
            get
            {
                return this.certificate == null;
            }
        }

        internal bool CanSign
        {
            get
            {
                return this.certificate != null;
            }
        }

        /// <summary>
        /// Tries to validate the signature. Return <c>true</c> on success; otherwise <c>false</c>.
        /// </summary>
        /// <param name="validationResult">When this method returns, contains the signature validation result.</param>
        /// <returns>Return <c>true</c> on success; otherwise <c>false</c>.</returns>
        public bool TryValidate(out SignatureValidationResult validationResult)
        {
            try
            {
                validationResult = this.Validate();

                return true;
            }
            catch
            {

            }

            validationResult = default(SignatureValidationResult);
            return false;
        }

        /// <summary>
        /// Tries to validate the signature. Return <c>true</c> on success; otherwise <c>false</c>.
        /// </summary>
        /// <param name="validationProperties">The validation properties.</param>
        /// <param name="validationResult">When this method returns, contains the signature validation result.</param>
        /// <returns>Return <c>true</c> on success; otherwise <c>false</c>.</returns>
        public bool TryValidate(SignatureValidationProperties validationProperties, out SignatureValidationResult validationResult)
        {
            Guard.ThrowExceptionIfNull(validationProperties, "validationProperties");

            try
            {
                validationResult = this.Validate(validationProperties);

                return true;
            }
            catch
            {

            }

            validationResult = default(SignatureValidationResult);
            return false;
        }

        /// <summary>
        /// Validates this digital signature instance.
        /// </summary>
        /// <returns>An instance of the <see cref="SignatureValidationResult"/> class.</returns>
        public SignatureValidationResult Validate()
        {
            SignatureValidationProperties validationProperties = new SignatureValidationProperties();

            SignatureValidationResult validationResult = this.Validate(validationProperties);

            return validationResult;
        }

        /// <summary>
        /// Validates this digital signature instance with specified validation properties.
        /// </summary>
        /// <param name="validationProperties">The validation properties.</param>
        /// <returns></returns>
        public SignatureValidationResult Validate(SignatureValidationProperties validationProperties)
        {
            Guard.ThrowExceptionIfNull(validationProperties, "validationProperties");

            if (!this.Properties.SourceStream.CanRead)
            {
                throw new InvalidOperationException("Document source stream does not support reading. Ensure that the stream is opened.");
            }

            if (!this.Properties.SourceStream.CanSeek)
            {
                throw new InvalidOperationException("Document source stream does not support seek operations. Ensure that the stream is opened.");
            }

            SignatureValidationResult validationResult;

            SignatureValidationHandlerBase validationHandler = SignatureValidationHandlersManager.GetHandler(this.Properties.SubFilter);
            if (validationHandler == null)
            {
                throw new NotSupportedException(string.Format("No signature validation handler was found for the subfilter: {0}", this.Properties.SubFilter));
            }
            else
            {
                validationResult = validationHandler.Validate(this.Properties, validationProperties);
            }

            if (validationResult == null)
            {
                throw new NotSupportedException("No signature validation result was constucted during the validation");
            }

            return validationResult;
        }

        internal void BuildInitalSignState()
        {
            this.Properties.SubFilter = SignatureSubFilterNames.ADBE_PKCS7_DETACHED;

            // inital contents.
            byte[] array = new byte[1] { 1 };
            byte[] package = this.CalculateContentsPackage(array);
            this.Properties.Contents = package;
            this.Properties.TimeOfSigning = DateTime.Now;

            // inital byte range
            int[] byteRange = new int[4];
            byteRange[0] = int.MaxValue;
            byteRange[1] = int.MaxValue;
            byteRange[2] = int.MaxValue;
            byteRange[3] = int.MaxValue;
            this.Properties.ByteRange = byteRange;
        }

        internal byte[] CalculateContentsPackage(byte[] byteArrayCompositionForHash)
        {
            ContentInfo contentInfo = new ContentInfo(byteArrayCompositionForHash);

            SignedCms signedCms = new SignedCms(contentInfo, true);

            byte[] package = null;
            try
            {
                package = this.GetContentPackage(signedCms, X509IncludeOption.WholeChain);
            }
            catch (CryptographicException)
            {
                package = this.GetContentPackage(signedCms, X509IncludeOption.EndCertOnly);
            }

            return package;
        }

        private byte[] GetContentPackage(SignedCms signedCms, X509IncludeOption includeOption)
        {
            byte[] package;
            CmsSigner signer = new CmsSigner(this.certificate);
            signer.IncludeOption = includeOption;
            signedCms.ComputeSignature(signer);
            package = signedCms.Encode();
            return package;
        }
    }
}
