﻿using System;
using System.Security.Cryptography.X509Certificates;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.DigitalSignatures
{
    /// <summary>
    ///  Defines a set of data properties represending a digital signature.
    /// </summary>
    public class SignatureDataProperties
    {
        private ISourceStream sourceStream;

        /// <summary>
        /// Initializes a new instance of the <see cref="SignatureDataProperties" /> class.
        /// </summary>
        public SignatureDataProperties()
        {
            this.PdfProperties = new SignatureDataPdfProperties();
        }

        /// <summary>
        /// Gets the name of the signature form field associated with this properties.
        /// </summary>
        /// <value>The name of the signature form field.</value>
        public string FieldName { get; internal set; }

        /// <summary>
        /// Gets the name of the preferred signature handler to use when validating this signature.
        /// </summary>
        /// <value>The filter.</value>
        public string Filter
        {
            get
            {
                return this.PdfProperties.Filter.Value;
            }
            internal set
            {
                this.PdfProperties.Filter.Value = value;
            }
        }

        /// <summary>
        /// Gets a name that describes the encoding of the signature value and key information in the signature dictionary.
        /// </summary>
        /// <value>The sub filter.</value>
        public string SubFilter
        {
            get
            {
                return this.PdfProperties.SubFilter.Value;
            }
            internal set
            {
                this.PdfProperties.SubFilter.Value = value;
            }
        }

        /// <summary>
        /// Gets the signature value.
        /// </summary>
        /// <value>The contents.</value>
        public byte[] Contents
        {
            get
            {
                return this.PdfProperties.Contents.Value;
            }
            internal set
            {
                this.PdfProperties.Contents.Value = value;
            }
        }

        /// <summary>
        /// Gets a collection of X.509 certificate chain used when signing and verifying signatures.
        /// </summary>
        /// <value>The certificates.</value>
        public X509Certificate2Collection Certificates
        {
            get
            {
                return this.PdfProperties.Cert.Value;
            }
            internal set
            {
                this.PdfProperties.Cert.Value = value;
            }
        }

        /// <summary>
        /// Gets an array of pairs of integers (starting byte offset, length in bytes) describing the exact byte range for the digest calculation.
        /// </summary>
        /// <value>The byte range.</value>
        public int[] ByteRange
        {
            get
            {
                return this.PdfProperties.ByteRange.Value;
            }
            internal set
            {
                this.PdfProperties.ByteRange.Value = value;
            }
        }

        /// <summary>
        /// Gets an array of three integers specifying changes to the document that have been made between the previous signature and this signature.
        /// </summary>
        /// <value>The changes.</value>
        public int[] Changes
        {
            get
            {
                return this.PdfProperties.Changes.Value;
            }
            internal set
            {
                this.PdfProperties.Changes.Value = value;
            }
        }

        /// <summary>
        /// Gets the name of the person or authority signing the document.
        /// </summary>
        /// <value>The name.</value>
        public string Name
        {
            get
            {
                return this.PdfProperties.Name.Value;
            }
            set
            {
                this.PdfProperties.Name.Value = value;
            }
        }

        /// <summary>
        /// Gets the CPU host name or physical location of the signing.
        /// </summary>
        /// <value>The location.</value>
        public string Location
        {
            get
            {
                return this.PdfProperties.Location.Value;
            }
            set
            {
                this.PdfProperties.Location.Value = value;
            }
        }

        /// <summary>
        /// Gets the reason for the signing.
        /// </summary>
        /// <value>The reason.</value>
        public string Reason
        {
            get
            {
                return this.PdfProperties.Reason.Value;
            }
            set
            {
                this.PdfProperties.Reason.Value = value;
            }
        }

        /// <summary>
        /// Gets contact information enabling a recipient to contact the signer to verify the signature.
        /// </summary>
        /// <value>The contact info.</value>
        public string ContactInfo
        {
            get
            {
                return this.PdfProperties.ContactInfo.Value;
            }
            set
            {
                this.PdfProperties.ContactInfo.Value = value;
            }
        }

        /// <summary>
        /// Gets the version of the signature handler used to create the signature.
        /// </summary>
        /// <value>The signature handler version.</value>
        public int SignatureHandlerVersion
        {
            get
            {
                return this.PdfProperties.SignatureHandlerVersion.Value;
            }
            internal set
            {
                this.PdfProperties.SignatureHandlerVersion.Value = value;
            }
        }

        /// <summary>
        /// Gets the version of the signature dictionary format.
        /// </summary>
        /// <value>The signature dictionary version.</value>
        public int SignatureDictionaryVersion
        {
            get
            {
                return this.PdfProperties.SignatureDictionaryVersion.Value;
            }
            internal set
            {
                this.PdfProperties.SignatureDictionaryVersion.Value = value;
            }
        }

        /// <summary>
        /// Gets or sets the time of signing.
        /// </summary>
        /// <value>The time of signing.</value>
        public DateTime TimeOfSigning
        {
            get
            {
                return this.PdfProperties.TimeOfSigning.Value;
            }
            set
            {
                this.PdfProperties.TimeOfSigning.Value = value;
            }
        }

        internal SignatureDataPdfProperties PdfProperties { get; private set; }

        internal ISourceStream SourceStream
        {
            get
            {
                return this.sourceStream;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "SourceStream");

                if (this.sourceStream != null)
                {
                    throw new ArgumentException("SourceStream is already defined.");
                }

                this.sourceStream = value;
            }
        }

        internal bool HasValidationRequiredProperties
        {
            get
            {
                if (!string.IsNullOrEmpty(this.SubFilter) && this.Contents != null && this.ByteRange != null)
                {
                    return true;
                }

                return false;
            }
        }
    }
}
