﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;

namespace Telerik.Windows.Documents.Fixed.Model.DigitalSignatures
{
    /// <summary>
    /// Defines a base class with the most basic need of functionality used for digital signature validaiton.
    /// </summary>
    public abstract class SignatureValidationHandlerBase
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="SignatureValidationHandlerBase" /> class.
        /// </summary>
        public SignatureValidationHandlerBase()
        {
        }

        /// <summary>
        /// Gets the signature data properties.
        /// </summary>
        /// <value>The signature data properties.</value>
        private SignatureDataProperties DataProperties { get; set; }

        /// <summary>
        /// Gets the signature validation properties.
        /// </summary>
        /// <value>The signature validation properties.</value>
        private SignatureValidationProperties ValidationProperties { get; set; }

        /// <summary>
        ///  When overridden in a derived class, validates the digital signature.
        /// </summary>
        /// <param name="dataProperties"></param>
        /// <param name="validationProperties"></param>
        /// <returns>An instance of the <see cref="SignatureValidationResult" /> class describing the result from the validation.</returns>
        protected abstract SignatureValidationResult ValidateOverride(SignatureDataProperties dataProperties, SignatureValidationProperties validationProperties);

        /// <summary>
        /// Validates a digital signature.
        /// </summary>
        /// <returns>An instance of the <see cref="SignatureValidationResult" /> class describing the result from the validation.</returns>
        internal SignatureValidationResult Validate(SignatureDataProperties dataProperties, SignatureValidationProperties validationProperties)
        {
            this.DataProperties = dataProperties;
            this.ValidationProperties = validationProperties;

            SignatureValidationResult validationResult = this.ValidateOverride(dataProperties, validationProperties);
            if (string.IsNullOrEmpty(validationResult.FieldName))
            {
                SignatureValidationResultBuilder validationResultBuilder = new SignatureValidationResultBuilder(validationResult);
                validationResultBuilder.BuildFieldName(this.DataProperties.FieldName);
            }

            return validationResult;
        }

        /// <summary>
        /// Computes a hash.
        /// </summary>
        /// <param name="hashAlgorithmOid">The hash algorithm oid.</param>
        internal byte[] ComputeHash(Oid hashAlgorithmOid)
        {
            HashAlgorithm hashAlgorithm = HashAlgorithmsManager.CreateHashAlgorithm(hashAlgorithmOid);

            byte[] computedHash = this.ComputeHash(hashAlgorithm);
            return computedHash;
        }

        /// <summary>
        /// Computes a hash.
        /// </summary>
        /// <param name="hashAlgorithm">The hash algorithm.</param>
        internal byte[] ComputeHash(HashAlgorithm hashAlgorithm)
        {
            byte[] byteArrayCompositionForHash = this.GetByteArrayCompositionForHash();

            byte[] computedHash = hashAlgorithm.ComputeHash(byteArrayCompositionForHash);

            return computedHash;
        }

        /// <summary>
        /// Gets the byte array over which the hash should be computed.
        /// </summary>
        protected byte[] GetByteArrayCompositionForHash()
        {
            IEnumerable<SourcePart> sourceParts = this.GetSourceParts();

            ByteRangeComposer byteRangeComposer = new ByteRangeComposer(this.DataProperties.SourceStream, sourceParts);
            byte[] byteArrayCompositionForHash = byteRangeComposer.Compose();
            return byteArrayCompositionForHash;
        }

        /// <summary>
        /// Gets a collection from a <see cref="SourcePart"/>.
        /// </summary>
        /// <returns></returns>
        internal IEnumerable<SourcePart> GetSourceParts()
        {
            int[] byteRanges = this.DataProperties.ByteRange;

            List<SourcePart> sourceParts = new List<SourcePart>(byteRanges.Length / 2);

            for (int i = 0; i < byteRanges.Length; i += 2)
            {
                int offset = byteRanges[i];
                int length = byteRanges[i + 1];
                SourcePart sourcePart = new SourcePart(offset, length);
                sourceParts.Add(sourcePart);
            }

            return sourceParts;
        }

        /// <summary>
        /// Validates the hash.
        /// </summary>
        /// <param name="originalHash">The original hash.</param>
        /// <param name="hashAlgorithmOid">The hash algorithm oid used for computing a comparing hash.</param>
        /// <returns></returns>
        internal bool ValidateHash(byte[] originalHash, Oid hashAlgorithmOid)
        {
            byte[] computedHash = this.ComputeHash(hashAlgorithmOid);
            bool isHashValid = this.ValidateHash(originalHash, computedHash);

            return isHashValid;
        }

        /// <summary>
        /// Validates the hash.
        /// </summary>
        /// <param name="originalHash">The original hash.</param>
        /// <param name="computedHash">The computed hash.</param>
        /// <returns></returns>
        internal bool ValidateHash(byte[] originalHash, byte[] computedHash)
        {
            for (int i = 0; i < originalHash.Length; i++)
            {
                if (originalHash[i] != computedHash[i])
                {
                    return false;
                }
            }

            return true;
        }
        
        /// <summary>
        ///  Validates the certificates and builds the corresponding data into the validation result.
        /// </summary>
        /// <param name="validationResultBuilder">The validation result builder.</param>
        /// <param name="certificateCollection">The certificate collection.</param>
        internal void ValidateCertificates(SignatureValidationResultBuilder validationResultBuilder, X509Certificate2Collection certificateCollection)
        {
            X509Certificate2 certificateToValidate = this.GetSigningCertificate(certificateCollection);

            if (certificateToValidate != null)
            {
                bool isCertificateFullyValid = this.ValidationProperties.Chain.Build(certificateToValidate);
                if (isCertificateFullyValid)
                {
                    validationResultBuilder.BuildIsCertificateValid(true);
                }
                else
                {
                    bool isCertificateValid = !this.ValidationProperties.Chain.ChainStatus
                       .Where(ch => this.ValidationProperties.ChainStatusFlags.HasFlag(ch.Status)).Any();

                    validationResultBuilder.BuildIsCertificateValid(isCertificateValid);
                    validationResultBuilder.BuildCertificatesChainElements(this.ValidationProperties.Chain.ChainElements);
                }
            }
        }

        /// <summary>
        /// Builds the hash algorithm into the the validation result argument.
        /// </summary>
        /// <param name="validationResultBuilder">The validation result builder.</param>
        /// <param name="digestAlgorithm">The digest algorithm.</param>
        internal void SetResultHashAlgorithm(SignatureValidationResultBuilder validationResultBuilder, Oid digestAlgorithm)
        {
            Oid hashAlgorithm = SignatureOids.GetMainAlgorithmOid(digestAlgorithm);

            validationResultBuilder.BuildHashAlgorithm(hashAlgorithm);
        }

        internal X509Certificate2 GetSigningCertificate(X509Certificate2Collection certificateCollection)
        {
            X509Certificate2 certificateToValidate = null;

            if (certificateCollection.Count == 1)
            {
                certificateToValidate = certificateCollection[0];
            }
            else
            {
                foreach (var cert in certificateCollection)
                {
                    bool isIssuer = IsIssuerCertificate(certificateCollection, cert);
                    if (isIssuer)
                    {
                        this.ValidationProperties.Chain.ChainPolicy.ExtraStore.Add(cert);
                    }
                    else
                    {
                        certificateToValidate = cert;
                    }
                }
            }

            return certificateToValidate;
        }

        private bool IsIssuerCertificate(X509Certificate2Collection certificateCollection, X509Certificate2 issuerCertificateCandidate)
        {
            foreach (var cert in certificateCollection)
            {
                if (cert.IssuerName.Name == issuerCertificateCandidate.SubjectName.Name)
                {
                    return true;
                }
            }

            return false;
        }
    }
}