﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.DigitalSignatures
{
    /// <summary>
    /// Defines a digital signature validation handlers manager.
    /// </summary>
    public static class SignatureValidationHandlersManager
    {
        private static Dictionary<string, SignatureValidationHandlerBase> store;

        static SignatureValidationHandlersManager()
        {
            store = new Dictionary<string, SignatureValidationHandlerBase>();

            RegisterHandler(SignatureSubFilterNames.ADBE_X509_RSA_SHA1, new Pkcs1());
            RegisterHandler(SignatureSubFilterNames.ADBE_PKCS7_SHA1, new Pkcs7());
            RegisterHandler(SignatureSubFilterNames.ADBE_PKCS7_DETACHED, new Pkcs7Detached());
        }

        /// <summary>
        /// Registers the signature validation handler.
        /// </summary>
        /// <param name="subFilter">The sub filter.</param>
        /// <param name="validationHandler">The instantiate func.</param>
        public static void RegisterHandler(string subFilter, SignatureValidationHandlerBase validationHandler)
        {
            Guard.ThrowExceptionIfNullOrEmpty(subFilter, "subFilter");
            Guard.ThrowExceptionIfNull(validationHandler, "validationHandler");

            store.Add(subFilter, validationHandler);
        }

        /// <summary>
        /// Unregisters the signature validation handler.
        /// </summary>
        /// <param name="subFilter">The sub filter.</param>
        public static void UnregisterHandler(string subFilter)
        {
            Guard.ThrowExceptionIfNullOrEmpty(subFilter, "subFilter");

            if (HasRegisteredHandler(subFilter))
            {
                store.Remove(subFilter);
            }
        }

        /// <summary>
        /// Determines whether the <see cref="SignatureValidationHandlersManager"/> contains a definition of a signature handler for the specified sub filter.
        /// </summary>
        /// <param name="subFilter">The sub filter.</param>
        /// <returns>
        /// <c>true</c> if <paramref name="subFilter" /> is found in the <see cref="SignatureValidationHandlersManager"/>; otherwise, <c>false</c>.
        /// </returns>
        public static bool HasRegisteredHandler(string subFilter)
        {
            Guard.ThrowExceptionIfNullOrEmpty(subFilter, "subFilter");

            bool hasRegisteredHandler = store.ContainsKey(subFilter);

            return hasRegisteredHandler;
        }

        /// <summary>
        /// Instantiates a signature validation handler.
        /// </summary>
        /// <param name="subFilter">The sub filter.</param>
        /// <returns></returns>
        internal static SignatureValidationHandlerBase GetHandler(string subFilter)
        {
            Guard.ThrowExceptionIfNullOrEmpty(subFilter, "subFilter");

            SignatureValidationHandlerBase validationHandler;
            if (store.TryGetValue(subFilter, out validationHandler))
            {
                return validationHandler;
            }

            return null;
        }
    }
}
