﻿using System;
using System.Linq;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;

namespace Telerik.Windows.Documents.Fixed.Model.DigitalSignatures
{
    /// <summary>
    /// Defines a class used to construct an instance of the <see cref="SignatureValidationResult"/> class.
    /// </summary>
    public class SignatureValidationResultBuilder
    {
        private readonly SignatureValidationResult validationResult;

        /// <summary>
        /// Initializes a new instance of the <see cref="SignatureValidationResultBuilder" /> class.
        /// </summary>
        public SignatureValidationResultBuilder() : this(new SignatureValidationResult())
        {
        }

        internal SignatureValidationResultBuilder(SignatureValidationResult validationResult)
        {
            this.validationResult = validationResult;
        }

        /// <summary>
        /// Builds the name of the signature form field.
        /// </summary>
        /// <param name="fieldName">Name of the field.</param>
        public void BuildFieldName(string fieldName)
        {
            this.validationResult.FieldName = fieldName;
        }

        /// <summary>
        /// Builds the is document modified.
        /// </summary>
        /// <param name="isDocumentModified">if set to <c>true</c> [is document modified].</param>
        public void BuildIsDocumentModified(bool isDocumentModified)
        {
            this.validationResult.IsDocumentModified = isDocumentModified;
        }

        /// <summary>
        /// Builds the is certificate valid.
        /// </summary>
        /// <param name="isCertificateValid">if set to <c>true</c> [is certificate valid].</param>
        public void BuildIsCertificateValid(bool isCertificateValid)
        {
            this.validationResult.IsCertificateValid = isCertificateValid;
        }

        /// <summary>
        /// Builds the certificates.
        /// </summary>
        /// <param name="certificates">The certificates.</param>
        public void BuildCertificates(X509Certificate2Collection certificates)
        {
            this.validationResult.Certificates = certificates;
        }

        /// <summary>
        /// Builds the certificates chain elements.
        /// </summary>
        /// <param name="certificatesChainElements">The certificates chain elements.</param>
        public void BuildCertificatesChainElements(X509ChainElementCollection certificatesChainElements)
        {
            this.validationResult.CertificatesChainElements = certificatesChainElements;
        }

        /// <summary>
        /// Builds the signer information.
        /// </summary>
        /// <param name="signerInformation">The signer information.</param>
        public void BuildSignerInformation(string signerInformation)
        {
            if (!string.IsNullOrEmpty(signerInformation))
            {
                this.validationResult.SignerInformation = signerInformation;
            }
        }

        /// <summary>
        /// Builds the hash algorithm.
        /// </summary>
        /// <param name="hashAlgorithm">The hash algorithm.</param>
        public void BuildHashAlgorithm(Oid hashAlgorithm)
        {
            this.validationResult.HashAlgorithm = hashAlgorithm;
        }

        /// <summary>
        /// Gets the signature validation result.
        /// </summary>
        /// <returns>An instance of the <see cref="SignatureValidationResult"/> class.</returns>
        public SignatureValidationResult GetResult()
        {
            return this.validationResult;
        }
    }
}
