﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Core.Data;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Editing.Collections;
using Telerik.Windows.Documents.Fixed.Model.Editing.Flow;
using Telerik.Windows.Documents.Fixed.Model.Editing.Layout;
using Telerik.Windows.Documents.Fixed.Model.Editing.Lists;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Fixed.Model.Navigation;
using Telerik.Windows.Documents.Fixed.Model.Objects;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Fixed.Model.Text;
using Telerik.Windows.Documents.Utilities;
using Stream = System.IO.Stream;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Primitives;
#else
using FontFamily = System.Windows.Media.FontFamily;
using FontStyle = System.Windows.FontStyle;
using FontWeight = System.Windows.FontWeight;
using Point = System.Windows.Point;
using Rect = System.Windows.Rect;
using Size = System.Windows.Size;
#endif

namespace Telerik.Windows.Documents.Fixed.Model.Editing
{
    /// <summary>
    /// Represents block of flowing content.
    /// </summary>
    public class Block : FixedContentEditorBase, IBlockElement
    {
        internal const double Epsilon = 0.1;
        private readonly List<LineInfo> lines;
        private readonly OrderedSet<LayoutElementBase> pendingElements;
        private readonly Stack<HyperlinkStartLayoutElement> hyperlinkStarts;

        private PositionContentElement bullet;
        private double indentAfterBullet;
        private bool isBulletInserted;
        private bool isBulletInvalidated;

        /// <summary>
        /// Initializes a new instance of the <see cref="Block" /> class.
        /// </summary>
        public Block()
        {
            this.lines = new List<LineInfo>();
            this.pendingElements = new OrderedSet<LayoutElementBase>();
            this.hyperlinkStarts = new Stack<HyperlinkStartLayoutElement>();
            this.ResetBulletProperties();

            this.DesiredSize = new Size(0, 0);
            this.ActualSize = new Size(0, 0);
            this.LineSpacing = FixedDocumentDefaults.BlockLineSpacing;
            this.LineSpacingType = FixedDocumentDefaults.LineSpacingType;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Block" /> class.
        /// </summary>
        /// <param name="other">The other block.</param>
        public Block(Block other)
            : this()
        {
            this.SpacingBefore = other.SpacingBefore;
            this.SpacingAfter = other.SpacingAfter;
            this.LineSpacing = other.LineSpacing;
            this.FirstLineIndent = other.FirstLineIndent;
            this.LeftIndent = other.LeftIndent;
            this.RightIndent = other.RightIndent;
            this.BackgroundColor = other.BackgroundColor;
            this.HorizontalAlignment = other.HorizontalAlignment;
            this.VerticalAlignment = other.VerticalAlignment;
            this.TabStops = other.TabStops;
            this.DefaultTabStopWidth = other.DefaultTabStopWidth;

            this.TextProperties.CopyFrom(other.TextProperties);
            this.GraphicProperties.CopyFrom(other.GraphicProperties);
        }

        /// <summary>
        /// Gets the desired size.
        /// </summary>
        /// <value>The desired size.</value>
        public Size DesiredSize { get; private set; }

        /// <summary>
        /// Gets the actual size of the block after it is drawn.
        /// </summary>
        /// <value>The actual size.</value>
        public Size ActualSize { get; private set; }

        /// <summary>
        /// Gets if there is any pending content after the last measure.
        /// </summary>
        /// <value>
        /// Boolean value indicating whether there is any pending content after the last measure.
        /// </value>
        public bool HasPendingContent
        {
            get
            {
                return this.PendingElements.Any();
            }
        }

        /// <summary>
        /// Gets or sets the bullet element.
        /// </summary>
        /// <value>The bullet element.</value>
        public PositionContentElement Bullet
        {
            get
            {
                return this.bullet;
            }
            set
            {
                if (this.bullet != value)
                {
                    this.bullet = value;
                    this.InvalidateBullet();
                }
            }
        }

        /// <summary>
        /// Gets or sets the indent after the bullet element.
        /// </summary>
        /// <value>The indent after the bullet element.</value>
        public double IndentAfterBullet
        {
            get
            {
                return this.indentAfterBullet;
            }
            set
            {
                if (this.indentAfterBullet != value)
                {
                    this.indentAfterBullet = value;
                    this.InvalidateBullet();
                }
            }
        }

        /// <summary>
        /// Gets or sets the spacing before.
        /// </summary>
        /// <value>The spacing before.</value>
        public double SpacingBefore { get; set; }

        /// <summary>
        /// Gets or sets the spacing after.
        /// </summary>
        /// <value>The spacing after.</value>
        public double SpacingAfter { get; set; }

        /// <summary>
        /// Gets or sets the line spacing.
        /// </summary>
        /// <value>The line spacing.</value>
        public double LineSpacing { get; set; }

        /// <summary>
        /// Gets or sets the value indicating how the spacing between lines is calculated.
        /// </summary>
        /// <value>The type line spacing rule.</value>
        public HeightType LineSpacingType { get; set; }

        /// <summary>
        /// Gets or sets the first line indent.
        /// </summary>
        /// <value>The first line indent.</value>
        public double FirstLineIndent { get; set; }

        /// <summary>
        /// Gets or sets the left indent.
        /// </summary>
        /// <value>The left indent.</value>
        public double LeftIndent { get; set; }

        /// <summary>
        /// Gets or sets the right indent.
        /// </summary>
        /// <value>The right indent.</value>
        public double RightIndent { get; set; }

        /// <summary>
        /// Gets or sets the background color.
        /// </summary>
        /// <value>The background color.</value>
        public ColorBase BackgroundColor { get; set; }

        /// <summary>
        /// Gets or sets the horizontal alignment.
        /// </summary>
        /// <value>The horizontal alignment.</value>
        public HorizontalAlignment HorizontalAlignment { get; set; }

        /// <summary>
        /// Gets or sets the vertical alignment.
        /// </summary>
        /// <value>The vertical alignment.</value>
        public VerticalAlignment VerticalAlignment { get; set; }

        /// <summary>
        /// Indicates if the block contains any elements.
        /// </summary>
        /// <value>True if the block does not contain any elements.</value>
        public bool IsEmpty
        {
            get
            {
                return !this.lines.Any() && !this.pendingElements.Any();
            }
        }

        /// <summary>
        /// Gets the pending elements.
        /// </summary>
        /// <value>The pending elements.</value>
        public IEnumerable<LayoutElementBase> PendingElements
        {
            get
            {
                return this.pendingElements;
            }
        }

        internal TabStopCollection TabStops { get; set; }

        internal double DefaultTabStopWidth { get; set; }

        internal bool HasBackground
        {
            get
            {
                return this.BackgroundColor != null && !this.BackgroundColor.IsTransparent;
            }
        }

        internal IEnumerable<LineInfo> Lines
        {
            get
            {
                foreach (LineInfo line in this.lines)
                {
                    yield return line;
                }
            }
        }

        /// <summary>
        /// Inserts the text.
        /// </summary>
        /// <param name="text">The text.</param>
        public void InsertText(string text)
        {
            Guard.ThrowExceptionIfNull(text, "text");

            if (!string.IsNullOrEmpty(text))
            {
                foreach (TextFragmentLayoutElement element in TextFragmentLayoutElement.CreateTextFragmentLayoutElements(this, text))
                {
                    this.Insert(element);
                }
            }
        }

        /// <summary>
        /// Inserts the text using the given font family.
        /// </summary>
        /// <param name="fontFamily">The font family.</param>
        /// <param name="text">The text.</param>
        public void InsertText(FontFamily fontFamily, string text)
        {
            Guard.ThrowExceptionIfNull(fontFamily, "fontFamily");
            Guard.ThrowExceptionIfNull(text, "text");

            this.TextProperties.TrySetFont(fontFamily);
            this.InsertText(text);
        }

        /// <summary>
        /// Inserts the text using the given font family, font style and font weight.
        /// </summary>
        /// <param name="fontFamily">The font family.</param>
        /// <param name="fontStyle">The font style.</param>
        /// <param name="fontWeight">The font weight.</param>
        /// <param name="text">The text.</param>
        public void InsertText(FontFamily fontFamily, FontStyle fontStyle, FontWeight fontWeight, string text)
        {
            Guard.ThrowExceptionIfNull(fontFamily, "fontFamily");
            Guard.ThrowExceptionIfNull(fontStyle, "fontStyle");
            Guard.ThrowExceptionIfNull(fontWeight, "fontWeight");
            Guard.ThrowExceptionIfNull(text, "text");

            this.TextProperties.TrySetFont(fontFamily, fontStyle, fontWeight);
            this.InsertText(text);
        }

        /// <summary>
        /// Inserts a line break.
        /// </summary>
        public void InsertLineBreak()
        {
            this.Insert(new LineBreakLayoutElement(this.TextProperties.Font, this.TextProperties.FontSize));
        }

        /// <summary>
        /// Inserts a form.
        /// </summary>
        /// <param name="source">The form source.</param>
        public void InsertForm(FormSource source)
        {
            this.AppendForm(source);
        }

        /// <summary>
        /// Inserts a form.
        /// </summary>
        /// <param name="source">The form source.</param>
        /// <param name="width">The width.</param>
        /// <param name="height">The height.</param>
        public void InsertForm(FormSource source, double width, double height)
        {
            this.AppendForm(source, width, height);
        }

        /// <summary>
        /// Inserts a form.
        /// </summary>
        /// <param name="source">The form source.</param>
        /// <param name="size">The size.</param>
        public void InsertForm(FormSource source, Size size)
        {
            this.AppendForm(source, size);
        }

        /// <summary>
        /// Inserts an image.
        /// </summary>
        /// <param name="stream">The stream that contains image.</param>
        public void InsertImage(Stream stream)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");

            this.ImageEditor.DrawImage(stream);
        }

        /// <summary>
        /// Inserts an image.
        /// </summary>
        /// <param name="stream">The stream that contains image.</param>
        /// <param name="width">The width.</param>
        /// <param name="height">The height.</param>
        public void InsertImage(Stream stream, double width, double height)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");

            this.ImageEditor.DrawImage(stream, width, height);
        }

        /// <summary>
        /// Inserts an image.
        /// </summary>
        /// <param name="stream">The stream that contains image.</param>
        /// <param name="size">The size.</param>
        public void InsertImage(Stream stream, Size size)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");

            this.InsertImage(stream, size.Width, size.Height);
        }

        /// <summary>
        /// Inserts an image.
        /// </summary>
        /// <param name="source">The image source.</param>
        public void InsertImage(ImageSource source)
        {
            Guard.ThrowExceptionIfNull(source, "source");

            this.ImageEditor.DrawImage(source);
        }

        /// <summary>
        /// Inserts an image.
        /// </summary>
        /// <param name="source">The image source.</param>
        /// <param name="size">The size.</param>
        public void InsertImage(ImageSource source, Size size)
        {
            Guard.ThrowExceptionIfNull(source, "source");

            this.InsertImage(source, size.Width, size.Height);
        }

        /// <summary>
        /// Inserts an image.
        /// </summary>
        /// <param name="source">The image source.</param>
        /// <param name="width">The width.</param>
        /// <param name="height">The height.</param>
        public void InsertImage(ImageSource source, double width, double height)
        {
            Guard.ThrowExceptionIfNull(source, "source");

            this.ImageEditor.DrawImage(source, width, height);
        }

        /// <summary>
        /// Inserts a line.
        /// </summary>
        /// <param name="point1">The point1.</param>
        /// <param name="point2">The point2.</param>
        public void InsertLine(Point point1, Point point2)
        {
            this.GraphicsEditor.DrawLine(point1, point2);
        }

        /// <summary>
        /// Inserts a rectangle.
        /// </summary>
        /// <param name="rectangle">The rectangle.</param>
        public void InsertRectangle(Rect rectangle)
        {
            this.GraphicsEditor.DrawRectangle(rectangle);
        }

        /// <summary>
        /// Inserts an ellipse.
        /// </summary>
        /// <param name="center">The center.</param>
        /// <param name="radiusX">The radius X.</param>
        /// <param name="radiusY">The radius Y.</param>
        public void InsertEllipse(Point center, double radiusX, double radiusY)
        {
            this.GraphicsEditor.DrawEllipse(center, radiusX, radiusY);
        }

        /// <summary>
        /// Inserts a circle.
        /// </summary>
        /// <param name="center">The center.</param>
        /// <param name="radius">The radius.</param>
        public void InsertCircle(Point center, double radius)
        {
            this.InsertEllipse(center, radius, radius);
        }

        /// <summary>
        /// Inserts a path.
        /// </summary>
        /// <param name="geometry">The geometry.</param>
        public void InsertPath(GeometryBase geometry)
        {
            this.GraphicsEditor.DrawPath(geometry);
        }

        /// <summary>
        /// Inserts hyperlink start.
        /// </summary>
        /// <param name="uri">The URI.</param>
        public void InsertHyperlinkStart(Uri uri)
        {
            Guard.ThrowExceptionIfNull(uri, "uri");

            Actions.UriAction uriAction = new Actions.UriAction(uri);
            ActionHyperlinkStartLayoutElement hyperlinkStart = new ActionHyperlinkStartLayoutElement(uriAction, this.TextProperties.Font, this.TextProperties.FontSize);
            this.InsertHyperlinkStart(hyperlinkStart);
        }

        /// <summary>
        /// Inserts hyperlink start.
        /// </summary>
        /// <param name="destination">The destination.</param>
        public void InsertHyperlinkStart(Destination destination)
        {
            Guard.ThrowExceptionIfNull(destination, "destination");

            DestinationHyperlinkStartLayoutElement hyperlinkStart = new DestinationHyperlinkStartLayoutElement(destination, this.TextProperties.Font, this.TextProperties.FontSize);
            this.InsertHyperlinkStart(hyperlinkStart);
        }

        /// <summary>
        /// Inserts hyperlink end.
        /// </summary>
        public void InsertHyperlinkEnd()
        {
            this.Insert(this.hyperlinkStarts.Pop().CreateEnd());
        }

        /// <summary>
        /// Inserts the specified element.
        /// </summary>
        /// <param name="element">The element.</param>
        public void Insert(PositionContentElement element)
        {
            Guard.ThrowExceptionIfNull(element, "element");

            this.Append(element);
        }

        /// <summary>
        /// Inserts the specified layout element.
        /// </summary>
        /// <param name="element">The element.</param>
        public void Insert(LayoutElementBase element)
        {
            Guard.ThrowExceptionIfNull(element, "element");

            LineInfo line = this.GetCurrentLine();
            line.Add(element);
        }

        /// <summary>
        /// Inserts the specified layout elements.
        /// </summary>
        /// <param name="elements">The elements.</param>
        public void InsertRange(IEnumerable<LayoutElementBase> elements)
        {
            Guard.ThrowExceptionIfNull(elements, "elements");

            if (elements.Any())
            {
                LineInfo line = this.GetCurrentLine();
                line.AddRange(elements);
            }
        }

        /// <summary>
        /// Measures the block in the infinite size.
        /// </summary>
        /// <returns>The block size.</returns>
        public Size Measure()
        {
            return this.Measure(new Size(double.PositiveInfinity, double.PositiveInfinity));
        }

        /// <summary>
        /// Measures the block element with specified available size.
        /// </summary>
        /// <param name="availableSize">The available size.</param>
        /// <returns>The result size.</returns>
        public Size Measure(Size availableSize)
        {
            if (this.TryCalculateDesiredSizeOnMinimalSizesScenarios(availableSize))
            {
                return this.DesiredSize;
            }

            Stack<LayoutElementBase> elements = this.PrepareElementsStackBeforeMeasure();
            List<LayoutElementBase> currentLineElements = new List<LayoutElementBase>();
            IEnumerable<LayoutElementBase> pendingElements;
            LayoutElementBase layoutElement;
            LineInfo line;
            LineMeasureContext lineMeasureContext;

            double contentWidth = 0;
            double height = this.SpacingBefore;
            double actualWidth = availableSize.Width - this.LeftIndent - this.RightIndent;
            double actualHeight = availableSize.Height;
            double currentLineWidth = 0;
            double offsetX = this.GetFirstLineLeftOffset();

            line = null;
            while (elements.Count > 0)
            {
                layoutElement = elements.Pop();
                lineMeasureContext = new LineMeasureContext(elements, actualWidth - (offsetX + currentLineWidth));
                if (!layoutElement.CanFit(actualWidth, offsetX + currentLineWidth))
                {
                    line = layoutElement.CompleteLine(actualWidth, offsetX + currentLineWidth, currentLineElements, out pendingElements);
                    currentLineElements.Clear();
                    currentLineWidth = 0;

                    foreach (LayoutElementBase pendingElement in pendingElements.Reverse())
                    {
                        elements.Push(pendingElement);
                    }

                    Block.CompleteLineEnding(line, elements);

                    if (this.ShouldMoveLineToPending(line, height, actualHeight))
                    {
                        break;
                    }
                    else
                    {
                        this.AddLineToMeasuring(line, ref contentWidth, ref height);

                        offsetX = this.GetLeftOffset();
                        line = null;

                        if (height > actualHeight)
                        {
                            break;
                        }
                    }
                }
                else
                {
                    currentLineElements.Add(layoutElement);
                    layoutElement.PrepareCalculatingWidthOnMeasure(lineMeasureContext);
                    currentLineWidth += layoutElement.Width;
                }
            }

            bool shouldAddFinalElements = line != null || currentLineElements.Count > 0;

            if (shouldAddFinalElements)
            {
                line = line ?? new LineInfo();
                line.AddRange(currentLineElements);

                if (this.ShouldMoveLineToPending(line, height, actualHeight))
                {
                    this.AddToPendingElements(line.Elements);
                }
                else
                {
                    this.AddLineToMeasuring(line, ref contentWidth, ref height);
                }
            }

            this.AddToPendingElements(elements);

            if (!this.pendingElements.Any())
            {
                height += this.SpacingAfter;
            }

            double significantRightIndent = Math.Max(this.RightIndent, 0);
            double width = contentWidth + this.LeftIndent + significantRightIndent;

            double normalizedWidth = Math.Max(width, 0);

            this.DesiredSize = new Size(normalizedWidth, height);

            return this.DesiredSize;
        }

        internal void InsertTab()
        {
            this.Insert(new TabLayoutElement(this.TabStops, this.DefaultTabStopWidth, this.TextProperties));
        }

        private IEnumerable<LayoutElementBase> GetAllElements()
        {
            foreach (LineInfo line in this.lines)
            {
                foreach (LayoutElementBase element in line.Elements)
                {
                    yield return element;
                }
            }

            foreach (LayoutElementBase element in this.pendingElements)
            {
                yield return element;
            }
        }

        private bool TryCalculateDesiredSizeOnMinimalSizesScenarios(Size availableSize)
        {
            if (availableSize == RadFixedDocumentEditor.MinimalWidthMeasureSize)
            {
                double maxWidth = 0;
                foreach (LayoutElementBase element in this.GetAllElements())
                {
                    maxWidth = Math.Max(element.GetMinWidth(), maxWidth);
                }

                // TODO: Think about the height.
                this.DesiredSize = new Size(maxWidth, availableSize.Height);

                return true;
            }
            else if (availableSize == RadFixedDocumentEditor.MinimalMeasureSize)
            {
                LayoutElementBase element = this.GetAllElements().FirstOrDefault();
                if (element != null)
                {
                    LineInfo tempElementParent = new LineInfo();
                    tempElementParent.Add(element);
                    double lineHeight = this.GetActualLineHeight(tempElementParent);

                    this.DesiredSize = new Size(element.GetMinMeasureSize().Width, lineHeight);
                }
                else
                {
                    this.DesiredSize = new Size(0, 0);
                }


                return true;
            }

            return false;
        }

        private static void CompleteLineEnding(LineInfo line, Stack<LayoutElementBase> pendingElements)
        {
            if (line.IsEndingWithLineBreak)
            {
                return;
            }

            while (pendingElements.Count > 0)
            {
                LayoutElementBase element = pendingElements.Peek();

                if (!element.CanAddToLineEnding)
                {
                    break;
                }

                pendingElements.Pop();

                LayoutElementBase firstPart, secondPart;
                if (element.TrySplitToAddToLineEnding(out firstPart, out secondPart))
                {
                    line.Add(firstPart);
                    pendingElements.Push(secondPart);
                }
                else
                {
                    line.Add(element);
                }
            }
        }

        private void AddLineToMeasuring(LineInfo line, ref double width, ref double height)
        {
            if (width < line.TrimmedWidth)
            {
                width = line.TrimmedWidth;
            }

            this.AddLine(line);
            double lineHeight = this.GetActualLineHeight(line);
            height += lineHeight;
        }

        private Stack<LayoutElementBase> PrepareElementsStackBeforeMeasure()
        {
            Stack<LayoutElementBase> elements = new Stack<LayoutElementBase>();

            for (int i = this.pendingElements.Count - 1; i >= 0; i--)
            {
                elements.Push(this.pendingElements[i]);
            }

            for (int i = this.lines.Count - 1; i >= 0; i--)
            {
                LineInfo line = this.lines[i];

                for (int j = line.Elements.Count() - 1; j >= 0; j--)
                {
                    elements.Push(line.Elements.ElementAt(j));
                }
            }

            this.InsertListBullet(elements);

            this.lines.Clear();
            this.pendingElements.Clear();

            return elements;
        }

        private bool ShouldMoveLineToPending(LineInfo line, double currentBlockHeight, double maxBlockHeight)
        {
            double lineHeight = this.GetActualLineHeight(line);

            return this.lines.Any() && currentBlockHeight + lineHeight > maxBlockHeight;
        }

        /// <summary>
        /// Draws the block element with the specified editor.
        /// </summary>
        /// <param name="editor">The editor.</param>
        /// <param name="boundingRect">The bounding rect.</param>
        public void Draw(FixedContentEditor editor, Rect boundingRect)
        {
            Guard.ThrowExceptionIfNull(editor, "editor");

            this.Measure(new Size(boundingRect.Width, boundingRect.Height));
            this.Arrange(editor, boundingRect);
        }

        /// <summary>
        /// Splits this instance.
        /// </summary>
        /// <returns>The remaining block.</returns>
        public Block Split()
        {
            Block block = new Block(this);
            block.InsertRange(this.PendingElements);
            block.FirstLineIndent = 0;
            block.SpacingBefore = 0;

            return block;
        }

        /// <summary>
        /// Splits this instance.
        /// </summary>
        /// <returns>The remaining block element.</returns>
        IBlockElement IBlockElement.Split()
        {
            return this.Split();
        }

        /// <summary>
        /// Sets the bullet from the corresponding list and list level.
        /// </summary>
        /// <param name="list">The list.</param>
        /// <param name="listLevel">The list level.</param>
        public void SetBullet(List list, int listLevel)
        {
            Guard.ThrowExceptionIfOutOfRange(FixedDocumentDefaults.FirstListLevelIndex, list.Levels.Count - 1, listLevel, "listLevel");

            list.Indexer.MoveToNextIndex(listLevel);
            ListLevel level = list.Levels[listLevel];

            this.Bullet = level.BulletNumberingFormat.GetBulletNumberingElement(list.Indexer);

            TextFragment textFragment = this.Bullet as TextFragment;

            if (textFragment != null)
            {
                using (this.SaveProperties())
                {
                    level.CharacterProperties.CopyTo(this);
                    this.CopyPropertiesTo(textFragment);
                }
            }

            this.FirstLineIndent = level.ParagraphProperties.FirstLineIndent;
            this.LeftIndent = level.ParagraphProperties.LeftIndent;
            this.IndentAfterBullet = level.IndentAfterBullet;
        }

        /// <summary>
        /// Clears the block children.
        /// </summary>
        public void Clear()
        {
            this.lines.Clear();
            this.hyperlinkStarts.Clear();
            this.pendingElements.Clear();
            this.ResetBulletProperties();
        }

        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current
        /// <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
            StringBuilder builder = new StringBuilder();

            foreach (LineInfo line in this.Lines)
            {
                builder.AppendLine(line.ToString());
            }

            return builder.ToString();
        }

        internal static void Draw(IEnumerable<LayoutElementBase> lineElements, DrawLayoutElementContext context)
        {
            double offsetX = context.OffsetX;

            foreach (LayoutElementBase layoutElement in lineElements)
            {
                layoutElement.Draw(new DrawLayoutElementContext(context.Editor, context.Block, context.Line, offsetX, context.OffsetY, context.BoundingRectangle));
                offsetX += layoutElement.Width;
            }
        }

        internal static Path CreateBackgroundPath(Rect rect, ColorBase fill)
        {
            Path path = new Path();
            path.Geometry = new RectangleGeometry(new Rect(rect.X, rect.Y, rect.Width, rect.Height));
            path.IsFilled = true;
            path.IsStroked = false;
            path.Fill = fill;

            return path;
        }

        internal override void Append(PositionContentElement element)
        {
            IEnumerable<LayoutElementBase> layoutElements;

            if (this.TryGetLayoutElements(element, out layoutElements))
            {
                foreach (LayoutElementBase layoutElement in layoutElements)
                {
                    this.Insert(layoutElement);
                }
            }
        }

        internal IEnumerable<LineInfo> GetLinesBetween(LineInfo first, LineInfo last)
        {
            Guard.ThrowExceptionIfNull(first, "first");
            Guard.ThrowExceptionIfNull(last, "last");

            if (first != last)
            {
                bool skip = true;
                foreach (LineInfo line in this.lines)
                {
                    if (line == last)
                    {
                        yield break;
                    }

                    if (!skip)
                    {
                        yield return line;
                    }

                    if (skip && first == line)
                    {
                        skip = false;
                    }
                }
            }
        }

        internal bool IsPending(LayoutElementBase element)
        {
            Guard.ThrowExceptionIfNull(element, "element");

            return this.pendingElements.Contains(element);
        }

        internal double GetActualLineHeight(LineInfo lineInfo)
        {
            Guard.ThrowExceptionIfNull(lineInfo, "lineInfo");

            double height = lineInfo.BaselineOffset + lineInfo.Descent;
            if (height == 0)
            {
                return 0;
            }

            switch (this.LineSpacingType)
            {
                case HeightType.AtLeast:
                    return Math.Max(height, this.LineSpacing);

                case HeightType.Exact:
                    return this.LineSpacing;

                case HeightType.Auto:
                    return height + (lineInfo.LineSpacingBaselineOffset + lineInfo.LineSpacingDescent) * (this.LineSpacing - 1);

                default:
                    return height;
            }
        }

        private void InsertListBullet(Stack<LayoutElementBase> elements)
        {
            if (this.isBulletInvalidated)
            {
                this.isBulletInvalidated = false;

                if (this.isBulletInserted)
                {
                    elements.Pop();
                    this.isBulletInserted = false;
                }

                if (this.Bullet != null || this.IndentAfterBullet != 0)
                {
                    LineInfo bulletElementsLine = new LineInfo();

                    if (this.Bullet != null)
                    {
                        IEnumerable<LayoutElementBase> bulletElements = null;

                        if (this.TryGetLayoutElements(this.Bullet, out bulletElements))
                        {
                            bulletElementsLine.AddRange(bulletElements);
                        }
                    }

                    double indentAfterBullet = this.CalculateAlignedIndentAfterBullet(bulletElementsLine);
                    elements.Push(new BulletLayoutElement(bulletElementsLine, indentAfterBullet, this.TextProperties));
                    this.isBulletInserted = true;
                }
            }
        }

        private double CalculateAlignedIndentAfterBullet(LineInfo bulletElementsLine)
        {
            if (this.FirstLineIndent < 0)
            {
                double hangingIndent = -this.FirstLineIndent;

                if (hangingIndent > bulletElementsLine.TrimmedWidth)
                {
                    return hangingIndent - bulletElementsLine.Width;
                }
            }

            return this.IndentAfterBullet;
        }

        private void ResetBulletProperties()
        {
            this.indentAfterBullet = 0;
            this.isBulletInserted = false;
            this.isBulletInvalidated = false;
            this.bullet = null;
        }

        private bool TryGetLayoutElements(PositionContentElement element, out IEnumerable<LayoutElementBase> layoutElements)
        {
            Guard.ThrowExceptionIfNull(element, "element");
            layoutElements = null;
            bool isSupportedElementType = true;

            switch (element.ElementType)
            {
                case FixedDocumentElementType.TextFragment:
                    TextFragment textFragment = (TextFragment)element;
                    using (this.SaveProperties())
                    {
                        this.CopyPropertiesFrom(textFragment);
                        layoutElements = TextFragmentLayoutElement.CreateTextFragmentLayoutElements(this, textFragment.Text).ToArray();
                    }
                    break;
                case FixedDocumentElementType.Image:
                    layoutElements = new LayoutElementBase[] { new ImageLayoutElement((Image)element, this.TextProperties) };
                    break;
                case FixedDocumentElementType.Path:
                    layoutElements = new LayoutElementBase[] { new PathLayoutElement((Path)element, this.TextProperties) };
                    break;
                case FixedDocumentElementType.Form:
                    layoutElements = new LayoutElementBase[] { new FormLayoutElement((Form)element, this.TextProperties) };
                    break;
                default:
                    isSupportedElementType = false;
                    break;
            }

            return isSupportedElementType;
        }

        private void CopyPropertiesFrom(TextFragment textFragment)
        {
            this.GraphicProperties.CopyFrom(textFragment.TextProperties.GeometryProperties);
            this.TextProperties.CopyFrom(textFragment.TextProperties);
        }

        private void CopyPropertiesTo(TextFragment textFragment)
        {
            this.GraphicProperties.CopyTo(textFragment.TextProperties.GeometryProperties);
            this.TextProperties.CopyTo(textFragment.TextProperties);
        }

        private void InvalidateBullet()
        {
            this.isBulletInvalidated = true;
        }

        private static double CalculateStartOffsetWhenCentering(double sizeToCenter, double sizeToFitIn)
        {
            return CalculateStartOffsetWhenPositioningAtOpositeSide(sizeToCenter, sizeToFitIn) / 2;
        }

        private static double CalculateStartOffsetWhenPositioningAtOpositeSide(double sizeToPosition, double sizeToFitIn)
        {
            return sizeToFitIn - sizeToPosition;
        }

        private void AddToPendingElements(IEnumerable<LayoutElementBase> elements)
        {
            foreach (LayoutElementBase element in elements)
            {
                if (element.ShouldSplitToBlocks(this))
                {
                    LayoutElementBase newElement;
                    IEnumerable<LayoutElementBase> includedElements;
                    IEnumerable<LayoutElementBase> excludedElements;
                    element.SplitToBlocks(this, out newElement, out includedElements, out excludedElements);

                    foreach (LayoutElementBase e in includedElements)
                    {
                        this.Insert(e);
                    }

                    this.pendingElements.InsertRange(0, excludedElements);
                    this.pendingElements.Add(newElement);
                }
                else
                {
                    this.pendingElements.Add(element);
                }
            }
        }

        private void Arrange(FixedContentEditor editor, Rect boundingRect)
        {
            double blockWidth = double.IsPositiveInfinity(boundingRect.Width) ? this.DesiredSize.Width : boundingRect.Width;
            double blockHeight = double.IsPositiveInfinity(boundingRect.Height) ? this.DesiredSize.Height : boundingRect.Height;
            double offsetY = boundingRect.Y + this.GetVerticalOffset(blockHeight, this.DesiredSize.Height);
            if (this.HasBackground)
            {
                Path background = CreateBackgroundPath(new Rect(boundingRect.X + this.LeftIndent, boundingRect.Y + this.SpacingBefore, blockWidth - this.RightIndent - this.LeftIndent, this.DesiredSize.Height - this.SpacingBefore - this.SpacingAfter), this.BackgroundColor);
                editor.Draw(background);
            }

            for (int i = 0; i < this.lines.Count; i++)
            {
                LineInfo line = this.lines[i];
                double offsetX = boundingRect.X + this.GetHorizontalOffset(blockWidth, line, i);
                line.BoundingRect = new Rect(offsetX, offsetY, line.Width, this.GetActualLineHeight(line));
                Block.Draw(line.Elements, new DrawLayoutElementContext(editor, this, line, offsetX, offsetY, boundingRect));

                offsetY += line.BoundingRect.Height;
            }

            this.ActualSize = new Size(blockWidth, blockHeight);
        }

        private void AddLine(LineInfo line)
        {
            Guard.ThrowExceptionIfNull(line, "line");

            this.lines.Add(line);
        }

        private LineInfo GetCurrentLine()
        {
            if (this.lines.Count == 0)
            {
                this.lines.Add(new LineInfo());
            }

            return this.lines.Last();
        }

        private double GetHorizontalOffset(double blockWidth, LineInfo line, int lineIndex)
        {
            Guard.ThrowExceptionIfNull(line, "line");

            double horizontalOffset = this.LeftIndent + (lineIndex == 0 ? this.GetFirstLineLeftOffset() : this.GetLeftOffset());
            switch (this.HorizontalAlignment)
            {
                case HorizontalAlignment.Center:
                    return horizontalOffset + CalculateStartOffsetWhenCentering(line.TrimmedWidth, blockWidth - horizontalOffset - this.RightIndent);

                case HorizontalAlignment.Right:
                    return CalculateStartOffsetWhenPositioningAtOpositeSide(line.TrimmedWidth, blockWidth) - this.RightIndent;

                case HorizontalAlignment.Left:
                    return horizontalOffset;

                default:
                    return 0;
            }
        }

        private double GetVerticalOffset(double availableHeight, double blockHeight)
        {
            double verticalOffset = this.SpacingBefore;
            switch (this.VerticalAlignment)
            {
                case Telerik.Windows.Documents.Fixed.Model.Editing.Flow.VerticalAlignment.Center:
                    verticalOffset += CalculateStartOffsetWhenCentering(blockHeight, availableHeight);
                    break;
                case Telerik.Windows.Documents.Fixed.Model.Editing.Flow.VerticalAlignment.Bottom:
                    verticalOffset += CalculateStartOffsetWhenPositioningAtOpositeSide(blockHeight, availableHeight);
                    break;
            }

            return verticalOffset;
        }

        private double GetFirstLineLeftOffset()
        {
            if (this.FirstLineIndent < 0)
            {
                return 0;
            }
            else
            {
                return this.FirstLineIndent;
            }
        }

        private double GetLeftOffset()
        {
            if (this.FirstLineIndent < 0)
            {
                return -this.FirstLineIndent;
            }
            else
            {
                return 0;
            }
        }

        private void InsertHyperlinkStart(HyperlinkStartLayoutElement hyperlinkStart)
        {
            this.hyperlinkStarts.Push(hyperlinkStart);
            this.Insert(hyperlinkStart);
        }

        internal void EndOpenHyperlink()
        {
            while (hyperlinkStarts.Count > 0)
            {
                this.InsertHyperlinkEnd();
            }
        }
    }
}
