﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Editing.Collections
{
    /// <summary>
    /// Defines a base collection class.
    /// </summary>
    public abstract class CollectionBase<T> : IEnumerable<T>
    {
        private readonly List<T> store;

        internal CollectionBase()
        {
            this.store = new List<T>();
        }

        /// <summary>
        /// Gets the collection items count.
        /// </summary>
        /// <value>The count.</value>
        public int Count
        {
            get
            {
                return this.store.Count;
            }
        }

        /// <summary>
        /// Gets the element at the specified index.
        /// </summary>
        /// <value>The element.</value>
        public T this[int index]
        {
            get
            {
                return this.store[index];
            }
            internal set
            {
                Guard.ThrowExceptionIfNull(value, "value");

                this.store[index] = value;
            }
        }

        /// <summary>
        /// Adds the specified item to the collection.
        /// </summary>
        /// <param name="item">The item.</param>
        public void Add(T item)
        {
            Guard.ThrowExceptionIfNull(item, "item");

            this.OnAddingElement(item);
            this.store.Add(item);
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<T> GetEnumerator()
        {
            return this.store.GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return this.GetEnumerator();
        }

        internal virtual void OnAddingElement(T element)
        {
        }

        internal void AddRange(IEnumerable<T> range)
        {
            Guard.ThrowExceptionIfNull(range, "range");

            this.store.AddRange(range);
        }

        internal void RemoveRange(int index, int count)
        {
            this.store.RemoveRange(index, count);
        }
    }
}
