﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Model.Editing.Flow;
using Telerik.Windows.Documents.Fixed.Model.Editing.Tables;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Utilities;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using Point = System.Windows.Point;
using Rect = System.Windows.Rect;
using Size = System.Windows.Size;
#endif
using Stream = System.IO.Stream;

namespace Telerik.Windows.Documents.Fixed.Model.Editing
{
    /// <summary>
    /// Represents content editor that will allow you to easily create fixed content.
    /// </summary>
    public class FixedContentEditor : FixedContentEditorBase
    {
        private readonly IContentRootElement root;
        private readonly PositionStack positionStack;
        private readonly Stack<Clipping> clippingStack;

        /// <summary>
        /// Initializes a new instance of the <see cref="FixedContentEditor" /> class.
        /// </summary>
        /// <param name="root">The root.</param>
        public FixedContentEditor(IContentRootElement root)
            : this(root, SimplePosition.Default)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="FixedContentEditor" /> class.
        /// </summary>
        /// <param name="root">The root.</param>
        /// <param name="initialPosition">The initial position.</param>
        public FixedContentEditor(IContentRootElement root, IPosition initialPosition)
        {
            this.root = root;

            this.positionStack = new PositionStack(initialPosition);
            this.clippingStack = new Stack<Clipping>();
        }

        /// <summary>
        /// Gets the root element for the fixed content.
        /// </summary>
        /// <value>The root element.</value>
        public IContentRootElement Root
        {
            get
            {
                return this.root;
            }
        }

        /// <summary>
        /// Gets or sets the current position.
        /// </summary>
        /// <value>The position.</value>
        public IPosition Position
        {
            get
            {
                return this.positionStack.Position;
            }
            set
            {
                this.positionStack.Position = value;
            }
        }

        /// <summary>
        /// Gets the current clipping object.
        /// </summary>
        /// <value>The clipping.</value>
        public Clipping Clipping
        {
            get
            {
                if (this.clippingStack.Count > 0)
                {
                    return this.clippingStack.Peek();
                }

                return null;
            }
        }

        internal Marker CurrentMarker { get; set; }

        internal bool SupportsAnnotations
        {
            get
            {
                return this.Root.SupportsAnnotations;
            }
        }

        /// <summary>
        /// Pushes the clipping.
        /// </summary>
        /// <param name="clip">The clip.</param>
        /// <returns></returns>
        public IDisposable PushClipping(GeometryBase clip)
        {
            Clipping clipping = new Clipping();
            clipping.Clip = clip;
            clipping.Clipping = this.Clipping;

            this.clippingStack.Push(clipping);

            return new DisposableObject(() => this.PopClipping());
        }

        /// <summary>
        /// Pushes the clipping.
        /// </summary>
        /// <param name="clip">The clip.</param>
        /// <returns></returns>
        public IDisposable PushClipping(Rect clip)
        {
            return this.PushClipping(new RectangleGeometry { Rect = clip });
        }

        /// <summary>
        /// Pops the clipping.
        /// </summary>
        /// <returns></returns>
        public Clipping PopClipping()
        {
            if (this.clippingStack.Count > 0)
            {
                return this.clippingStack.Pop();
            }

            return null;
        }

        /// <summary>
        /// Draws the text using current text state at the current position.
        /// </summary>
        /// <param name="text">The text.</param>
        public void DrawText(string text)
        {
            Guard.ThrowExceptionIfNull(text, "text");

            Block block = new Block();
            block.TextProperties.CopyFrom(this.TextProperties);
            block.GraphicProperties.CopyFrom(this.GraphicProperties);
            block.InsertText(text);

            this.DrawBlock(block);
        }

        /// <summary>
        /// Draws the text.
        /// </summary>
        /// <param name="text">The text.</param>
        /// <param name="size">The size.</param>
        public void DrawText(string text, Size size)
        {
            Guard.ThrowExceptionIfNull(text, "text");

            Block block = new Block();
            block.TextProperties.CopyFrom(this.TextProperties);
            block.GraphicProperties.CopyFrom(this.GraphicProperties);
            block.InsertText(text);

            this.DrawBlock(block, size);
        }

        /// <summary>
        /// Draws the image.
        /// </summary>
        /// <param name="stream">The stream.</param>
        public void DrawImage(Stream stream)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");

            this.ImageEditor.DrawImage(stream);
        }

        /// <summary>
        /// Draws the image.
        /// </summary>
        /// <param name="stream">The stream.</param>
        /// <param name="width">The width.</param>
        /// <param name="height">The height.</param>
        public void DrawImage(Stream stream, double width, double height)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");

            this.ImageEditor.DrawImage(stream, width, height);
        }

        /// <summary>
        /// Draws the image.
        /// </summary>
        /// <param name="stream">The stream.</param>
        /// <param name="size">The size.</param>
        public void DrawImage(Stream stream, Size size)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");

            this.DrawImage(stream, size.Width, size.Height);
        }

        /// <summary>
        /// Draws the image.
        /// </summary>
        /// <param name="source">The source.</param>
        public void DrawImage(ImageSource source)
        {
            Guard.ThrowExceptionIfNull(source, "source");

            this.ImageEditor.DrawImage(source);
        }

        /// <summary>
        /// Draws the image.
        /// </summary>
        /// <param name="source">The source.</param>
        /// <param name="size">The size.</param>
        public void DrawImage(ImageSource source, Size size)
        {
            Guard.ThrowExceptionIfNull(source, "source");

            this.DrawImage(source, size.Width, size.Height);
        }

        /// <summary>
        /// Draws the image.
        /// </summary>
        /// <param name="source">The source.</param>
        /// <param name="width">The width.</param>
        /// <param name="height">The height.</param>
        public void DrawImage(ImageSource source, double width, double height)
        {
            Guard.ThrowExceptionIfNull(source, "source");

            this.ImageEditor.DrawImage(source, width, height);
        }

        /// <summary>
        /// Draws the form.
        /// </summary>
        /// <param name="source">The source.</param>
        public void DrawForm(FormSource source)
        {
            this.AppendForm(source);
        }

        /// <summary>
        /// Draws a form.
        /// </summary>
        /// <param name="source">The form source.</param>
        /// <param name="width">The width.</param>
        /// <param name="height">The height.</param>
        public void DrawForm(FormSource source, double width, double height)
        {
            this.AppendForm(source, width, height);
        }

        /// <summary>
        /// Draws a form.
        /// </summary>
        /// <param name="source">The form source.</param>
        /// <param name="size">The size.</param>
        public void DrawForm(FormSource source, Size size)
        {
            this.AppendForm(source, size);
        }


        /// <summary>
        /// Creates new <see cref="Widget"/> representing a <see cref="FormField"/> and draws the widget with a specified <see cref="Annotation"/> size.
        /// This method will add widget only in cases when the <see cref="FixedContentEditor"/> Root supports annotations.
        /// </summary>
        /// <typeparam name="T">The widget type.</typeparam>
        /// <param name="parentField">The form field the should be represented by the widget.</param>
        /// <param name="annotationSize">The <see cref="Annotation"/> size.</param>
        public void DrawWidget<T>(FormField<T> parentField, Size annotationSize)
            where T : Widget
        {
            Guard.ThrowExceptionIfNull(parentField, "parentField");

            T widget = parentField.Widgets.AddWidget();
            this.DrawWidgetInternal(widget, annotationSize);
        }

        /// <summary>
        /// Creates new <see cref="RadioButtonWidget"/> representing a <see cref="RadioButtonField"/> and draws the widget with a specified <see cref="Annotation"/> size.
        /// This method will add widget only in cases when the <see cref="FixedContentEditor"/> Root supports annotations.
        /// </summary>
        /// <param name="parentField">The form field the should be represented by the widget.</param>
        /// <param name="option">The option that should be represented by the widget.</param>
        /// <param name="annotationSize">The <see cref="Annotation"/> size.</param>
        public void DrawWidget(RadioButtonField parentField, RadioOption option, Size annotationSize)
        {
            Guard.ThrowExceptionIfNull(parentField, "parentField");

            RadioButtonWidget widget = parentField.Widgets.AddWidget(option);
            this.DrawWidgetInternal(widget, annotationSize);
        }

        /// <summary>
        /// Draws a line.
        /// </summary>
        /// <param name="point1">The point1.</param>
        /// <param name="point2">The point2.</param>
        public void DrawLine(Point point1, Point point2)
        {
            this.GraphicsEditor.DrawLine(point1, point2);
        }

        /// <summary>
        /// Draws a rectangle.
        /// </summary>
        /// <param name="rectangle">The rectangle.</param>
        public void DrawRectangle(Rect rectangle)
        {
            this.GraphicsEditor.DrawRectangle(rectangle);
        }

        /// <summary>
        /// Draws a ellipse.
        /// </summary>
        /// <param name="center">The center.</param>
        /// <param name="radiusX">The radius X.</param>
        /// <param name="radiusY">The radius Y.</param>
        public void DrawEllipse(Point center, double radiusX, double radiusY)
        {
            this.GraphicsEditor.DrawEllipse(center, radiusX, radiusY);
        }

        /// <summary>
        /// Draws a circle.
        /// </summary>
        /// <param name="center">The center.</param>
        /// <param name="radius">The radius.</param>
        public void DrawCircle(Point center, double radius)
        {
            this.DrawEllipse(center, radius, radius);
        }

        /// <summary>
        /// Draws a path with given geometry.
        /// </summary>
        /// <param name="geometry">The geometry.</param>
        public void DrawPath(GeometryBase geometry)
        {
            this.GraphicsEditor.DrawPath(geometry);
        }

        /// <summary>
        /// Draws the table.
        /// </summary>
        /// <param name="table">The table.</param>
        public void DrawTable(Table table)
        {
            table.Draw(this, new Rect(0, 0, double.PositiveInfinity, double.PositiveInfinity));
        }

        /// <summary>
        /// Draws the table with some fit table width.
        /// </summary>
        /// <param name="table">The table.</param>
        /// <param name="width">The width to fit the table in.</param>
        public void DrawTable(Table table, double width)
        {
            table.Draw(this, new Rect(0, 0, width, double.PositiveInfinity));
        }

        /// <summary>
        /// Draws the table.
        /// </summary>
        /// <param name="table">The table.</param>
        /// <param name="size">The size to fit the table in. Only the rows that fit in the given hight are drawn. The rest of the table can be taken with the table.Split() method.</param>
        public void DrawTable(Table table, Size size)
        {
            table.Draw(this, new Rect(0, 0, size.Width, size.Height));
        }

        /// <summary>
        /// Draws the block.
        /// </summary>
        /// <param name="block">The block.</param>
        public void DrawBlock(IBlockElement block)
        {
            Guard.ThrowExceptionIfNull(block, "block");

            this.DrawBlock(block, new Size(double.PositiveInfinity, double.PositiveInfinity));
        }

        /// <summary>
        /// Draws the block in given size.
        /// </summary>
        /// <param name="block">The block.</param>
        /// <param name="size">The size.</param>
        public void DrawBlock(IBlockElement block, Size size)
        {
            Guard.ThrowExceptionIfNull(block, "block");

            block.Draw(this, new Rect(0, 0, size.Width, size.Height));
        }

        /// <summary>
        /// Draws the specified element.
        /// </summary>
        /// <param name="element">The element.</param>
        public void Draw(PositionContentElement element)
        {
            Guard.ThrowExceptionIfNull(element, "element");

            this.Append(element);
        }

        /// <summary>
        /// Saves the position.
        /// </summary>
        /// <returns></returns>
        public IDisposable SavePosition()
        {
            return this.positionStack.SavePosition();
        }

        /// <summary>
        /// Restores the position.
        /// </summary>
        public void RestorePosition()
        {
            this.positionStack.RestorePosition();
        }

        internal IDisposable PushTransformedClipping(Rect rect)
        {
            return this.PushClipping(this.Position.Matrix.TransformRectangle(rect));
        }

        internal void AddAnnotation(Annotation annotation)
        {
            Guard.ThrowExceptionIfNull(annotation, "annotation");

            if (this.SupportsAnnotations)
            {
                annotation.Rect = this.Position.Matrix.Transform(annotation.Rect);
                this.Root.Annotations.Add(annotation);
            }
        }

        internal override void Append(PositionContentElement element)
        {
            Guard.ThrowExceptionIfNull(element, "element");

            element.Clipping = this.Clipping;
            element.Marker = this.CurrentMarker;
            element.Position = new MatrixPosition(element.Position.Matrix.MultiplyBy(this.Position.Matrix));
            this.Root.Content.Add(element);
        }

        private void DrawWidgetInternal(Widget widget, Size annotationSize)
        {
            widget.Rect = new Rect(0, 0, annotationSize.Width, annotationSize.Height);
            widget.RecalculateContent();

            this.AddAnnotation(widget);
        }
    }
}
