﻿using System;
using System.Collections.Generic;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Model.Objects;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Editing
{
    /// <summary>
    /// Represents base class for fixed content editing.
    /// </summary>
    public abstract class FixedContentEditorBase
    {
        private readonly TextProperties textProperties;
        private readonly GraphicProperties graphicProperties;

        private ImageEditor imageEditor;
        private GraphicsEditor graphicsEditor;

        internal FixedContentEditorBase()
        {
            this.textProperties = new TextProperties();
            this.graphicProperties = new GraphicProperties();
        }

        /// <summary>
        /// Gets the graphic state that is responsible for the graphics properties in the fixed content.
        /// </summary>
        /// <value>The current graphic state.</value>
        public GraphicProperties GraphicProperties
        {
            get
            {
                return this.graphicProperties;
            }
        }

        /// <summary>
        /// Gets the text state that is responsible for the text properties in the fixed content.
        /// </summary>
        /// <value>The text state.</value>
        public TextProperties TextProperties
        {
            get
            {
                return this.textProperties;
            }
        }

        internal ImageEditor ImageEditor
        {
            get
            {
                if (this.imageEditor == null)
                {
                    this.imageEditor = new ImageEditor(this);
                }

                return this.imageEditor;
            }
        }

        internal GraphicsEditor GraphicsEditor
        {
            get
            {
                if (this.graphicsEditor == null)
                {
                    this.graphicsEditor = new GraphicsEditor(this);
                }

                return this.graphicsEditor;
            }
        }

        /// <summary>
        /// Saves the graphic properties.
        /// </summary>
        public IDisposable SaveGraphicProperties()
        {
            this.GraphicProperties.Save();

            return new DisposableObject(this.RestoreGraphicProperties);
        }

        /// <summary>
        /// Restores the graphic properties.
        /// </summary>
        public void RestoreGraphicProperties()
        {
            this.GraphicProperties.Restore();
        }

        /// <summary>
        /// Saves the text properties.
        /// </summary>
        public IDisposable SaveTextProperties()
        {
            this.TextProperties.Save();

            return new DisposableObject(this.RestoreTextProperties);
        }

        /// <summary>
        /// Restores the text properties.
        /// </summary>
        public void RestoreTextProperties()
        {
            this.TextProperties.Restore();
        }

        /// <summary>
        /// Saves text and graphic properties.
        /// </summary>
        public IDisposable SaveProperties()
        {
            this.textProperties.Save();
            this.graphicProperties.Save();

            return new DisposableObject(this.RestoreProperties);
        }

        /// <summary>
        /// Restores text and graphic properties.
        /// </summary>
        public void RestoreProperties()
        {
            this.textProperties.Restore();
            this.graphicProperties.Restore();
        }

        internal void AppendForm(FormSource source)
        {
            Guard.ThrowExceptionIfNull(source, "source");

            this.Append(new Form() { FormSource = source });
        }

        internal void AppendForm(FormSource source, double width, double height)
        {
            Guard.ThrowExceptionIfNull(source, "source");

            this.Append(new Form() { FormSource = source, Width = width, Height = height });
        }

        internal void AppendForm(FormSource source, Size size)
        {
            Guard.ThrowExceptionIfNull(source, "source");

            this.Append(new Form() { FormSource = source, Width = size.Width, Height = size.Height });
        }

        internal abstract void Append(PositionContentElement element);
    }
}
