﻿#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
#else
using System.Windows;
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Editing.Flow
{
    /// <summary>
    /// Defines a set of properties which will be applied when an inline is created.
    /// </summary>
    public class CharacterProperties
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="CharacterProperties" /> class.
        /// </summary>
        public CharacterProperties()
        {
            this.FontSize = FixedDocumentDefaults.FontSize;
            this.Font = FixedDocumentDefaults.Font;
            this.ForegroundColor = FixedDocumentDefaults.ForegroundColor;
            this.HighlightColor = FixedDocumentDefaults.HighlightColor;
            this.BaselineAlignment = FixedDocumentDefaults.BaselineAlignment;
            this.UnderlinePattern = FixedDocumentDefaults.UnderlinePattern;
            this.UnderlineColor = FixedDocumentDefaults.ForegroundColor;
        }

        /// <summary>
        /// Gets or sets the size of the font.
        /// </summary>
        /// <value>The size of the font.</value>
        public double FontSize { get; set; }

        /// <summary>
        /// Gets or sets the font.
        /// </summary>
        /// <value>The font.</value>
        public FontBase Font { get; set; }

        /// <summary>
        /// Gets or sets the color of the foreground.
        /// </summary>
        /// <value>The color of the foreground.</value>
        public ColorBase ForegroundColor { get; set; }

        /// <summary>
        /// Gets or sets the color of the highlight.
        /// </summary>
        /// <value>The color of the highlight.</value>
        public ColorBase HighlightColor { get; set; }

        /// <summary>
        /// Gets or sets the baseline alignment.
        /// </summary>
        /// <value>The baseline alignment.</value>
        public BaselineAlignment BaselineAlignment { get; set; }

        /// <summary>
        /// Gets or sets the underline pattern.
        /// </summary>
        /// <value>The underline pattern.</value>
        public UnderlinePattern UnderlinePattern { get; set; }

        /// <summary>
        /// Gets or sets the color of the underline.
        /// </summary>
        /// <value>The color of the underline.</value>
        public ColorBase UnderlineColor { get; set; }

        /// <summary>
        /// Tries to set the current font.
        /// </summary>
        /// <param name="fontFamily">The font family.</param>
        /// <returns>True if succeeded.</returns>
        public bool TrySetFont(FontFamily fontFamily)
        {
            return this.TrySetFont(fontFamily, FontStyles.Normal, FontWeights.Normal);
        }

        /// <summary>
        /// Tries to set the current font.
        /// </summary>
        /// <param name="fontFamily">The font family.</param>
        /// <param name="fontStyle">The font style.</param>
        /// <param name="fontWeight">The font weight.</param>
        /// <returns>True if succeeded.</returns>
        public bool TrySetFont(FontFamily fontFamily, FontStyle fontStyle, FontWeight fontWeight)
        {
            FontBase font;
            bool result = FontsRepository.TryCreateFont(fontFamily, fontStyle, fontWeight, out font);
            if (result)
            {
                this.Font = font;
            }

            return result;
        }

        /// <summary>
        /// Copies the properties from.
        /// </summary>
        /// <param name="fromProperties">From properties.</param>
        public void CopyPropertiesFrom(CharacterProperties fromProperties)
        {
            this.FontSize = fromProperties.FontSize;
            this.Font = fromProperties.Font;
            this.ForegroundColor = fromProperties.ForegroundColor;
            this.HighlightColor = fromProperties.HighlightColor;
            this.BaselineAlignment = fromProperties.BaselineAlignment;
            this.UnderlineColor = fromProperties.UnderlineColor;
            this.UnderlinePattern = fromProperties.UnderlinePattern;
        }

        internal void CopyTo(Block block)
        {
            Guard.ThrowExceptionIfNull(block, "block");

            block.TextProperties.FontSize = this.FontSize;
            block.TextProperties.Font = this.Font;
            block.TextProperties.HighlightColor = this.HighlightColor;
            block.TextProperties.BaselineAlignment = this.BaselineAlignment;
            block.TextProperties.UnderlineColor = this.UnderlineColor;
            block.TextProperties.UnderlinePattern = this.UnderlinePattern;
            block.GraphicProperties.FillColor = this.ForegroundColor;
        }
    }
}
