﻿using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Editing.Flow
{
    /// <summary>
    /// Defines a set of properties which will be applied when a paragraph is created.
    /// </summary>
    public class ParagraphProperties
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="ParagraphProperties" /> class.
        /// </summary>
        public ParagraphProperties()
        {
            this.BackgroundColor = FixedDocumentDefaults.BackgroundColor;
            this.HorizontalAlignment = FixedDocumentDefaults.HorizontalAlignment;
            this.SpacingBefore = FixedDocumentDefaults.SpacingBefore;
            this.SpacingAfter = FixedDocumentDefaults.SpacingAfter;
            this.LineSpacing = FixedDocumentDefaults.LineSpacing;
            this.LineSpacingType = FixedDocumentDefaults.LineSpacingType;
            this.LeftIndent = FixedDocumentDefaults.LeftIndent;
            this.RightIndent = FixedDocumentDefaults.RightIndent;
            this.FirstLineIndent = FixedDocumentDefaults.FirstLineIndent;
            this.ListLevel = FixedDocumentDefaults.FirstListLevelIndex;
            this.ListId = null;
        }

        /// <summary>
        /// Gets or sets the horizontal alignment.
        /// </summary>
        /// <value>The horizontal alignment.</value>
        public HorizontalAlignment HorizontalAlignment { get; set; }

        /// <summary>
        /// Gets or sets the spacing before.
        /// </summary>
        /// <value>The spacing before.</value>
        public double SpacingBefore { get; set; }

        /// <summary>
        /// Gets or sets the spacing after.
        /// </summary>
        /// <value>The spacing after.</value>
        public double SpacingAfter { get; set; }

        /// <summary>
        /// Gets or sets the line spacing.
        /// </summary>
        /// <value>The line spacing.</value>
        public double LineSpacing { get; set; }

        /// <summary>
        /// Gets or sets the value indicating how the spacing between lines is calculated.
        /// </summary>
        /// <value>The type line spacing rule.</value>
        public HeightType LineSpacingType { get; set; }

        /// <summary>
        /// Gets or sets the first line indent.
        /// </summary>
        /// <value>The first line indent.</value>
        public double FirstLineIndent { get; set; }

        /// <summary>
        /// Gets or sets the left indent.
        /// </summary>
        /// <value>The left indent.</value>
        public double LeftIndent { get; set; }

        /// <summary>
        /// Gets or sets the right indent.
        /// </summary>
        /// <value>The right indent.</value>
        public double RightIndent { get; set; }

        /// <summary>
        /// Gets or sets the color of the background.
        /// </summary>
        /// <value>The color of the background.</value>
        public ColorBase BackgroundColor { get; set; }

        /// <summary>
        /// Gets or sets the list id.
        /// </summary>
        /// <value>The list id.</value>
        public int? ListId { get; set; }

        /// <summary>
        /// Gets or sets the list level.
        /// </summary>
        /// <value>The list level.</value>
        public int ListLevel { get; set; }

        /// <summary>
        /// Copies the properties from.
        /// </summary>
        /// <param name="fromProperties">From properties.</param>
        public void CopyPropertiesFrom(ParagraphProperties fromProperties)
        {
            Guard.ThrowExceptionIfNull(fromProperties, "other");

            this.HorizontalAlignment = fromProperties.HorizontalAlignment;
            this.SpacingBefore = fromProperties.SpacingBefore;
            this.SpacingAfter = fromProperties.SpacingAfter;
            this.LineSpacing = fromProperties.LineSpacing;
            this.LineSpacingType = fromProperties.LineSpacingType;
            this.FirstLineIndent = fromProperties.FirstLineIndent;
            this.LeftIndent = fromProperties.LeftIndent;
            this.RightIndent = fromProperties.RightIndent;
            this.BackgroundColor = fromProperties.BackgroundColor;
        }

        internal void CopyTo(Block block)
        {
            Guard.ThrowExceptionIfNull(block, "block");

            block.HorizontalAlignment = this.HorizontalAlignment;
            block.SpacingBefore = this.SpacingBefore;
            block.SpacingAfter = this.SpacingAfter;
            block.LineSpacing = this.LineSpacing;
            block.LineSpacingType = this.LineSpacingType;
            block.FirstLineIndent = this.FirstLineIndent;
            block.LeftIndent = this.LeftIndent;
            block.RightIndent = this.RightIndent;
            block.BackgroundColor = this.BackgroundColor;
        }
    }
}
