﻿using System;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Editing.Flow
{
    internal class SectionInfo
    {
        private readonly RadFixedDocument document;
        private readonly SectionProperties properties;
        private readonly SpacingsCalculator spacingCalculator;

        private FixedContentEditor editor;

        private double actualWidth;
        private double actualHeight;
        private double offsetX;
        private double offsetY;

        public SectionInfo(RadFixedDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");

            this.document = document;
            this.properties = new SectionProperties();
            this.spacingCalculator = new SpacingsCalculator(this);
        }

        public SectionProperties Properties
        {
            get
            {
                return this.properties;
            }
        }

        public Size RemainingSize
        {
            get
            {
                return new Size(Math.Max(0, this.actualWidth), Math.Max(0, this.actualHeight - this.offsetY));
            }
        }

        public bool IsStarted
        {
            get
            {
                return this.editor != null;
            }
        }

        public void StartNewPage(SectionProperties properties)
        {
            Guard.ThrowExceptionIfNull(properties, "properties");

            this.properties.CopyPropertiesFrom(properties);
            this.StartNewPage();
        }

        public void StartNewPage()
        {
            RadFixedPage page = this.document.Pages.AddPage();
            page.Size = this.Properties.PageSize;
            page.Rotation = this.Properties.PageRotation;

            this.editor = new FixedContentEditor(page, SimplePosition.Default);

            this.UpdateAvailableSize();
        }

        public void UpdateAvailableSize()
        {
            this.actualWidth = this.properties.PageSize.Width - this.properties.PageMargins.Left - this.properties.PageMargins.Right;
            this.actualHeight = this.properties.PageSize.Height - this.properties.PageMargins.Bottom;
            this.offsetX = this.properties.PageMargins.Left;
            this.offsetY = this.properties.PageMargins.Top;
            this.spacingCalculator.Clear();
        }

        public void BeginExportContent()
        {
            this.UpdateAvailableSize();
            this.spacingCalculator.BeginExportContent();
        }

        public bool CanFit(IBlockElement blockElement)
        {
            Size minimalSize = blockElement.Measure(RadFixedDocumentEditor.MinimalMeasureSize);

            return minimalSize.Height <= this.RemainingSize.Height;
        }

        public void InsertBlockElement(IBlockElement blockElement)
        {
            Guard.ThrowExceptionIfNull(this.editor, "editor");
            Guard.ThrowExceptionIfNull(blockElement, "blockElement");

            Block block = blockElement as Block;
            double spacingBefore = 0;
            if (block != null)
            {
                spacingBefore = block.SpacingBefore;
                block.SpacingBefore = this.spacingCalculator.GetSpacingBefore(block);

                if (this.spacingCalculator.ShouldExpandBackground(block))
                {
                    Path background = this.spacingCalculator.GetExpandedBackground(this.offsetX, this.offsetY);
                    this.editor.Draw(background);
                }
            }

            blockElement.Draw(this.editor, new Rect(this.offsetX, this.offsetY, this.RemainingSize.Width, this.RemainingSize.Height));
            this.offsetY += blockElement.DesiredSize.Height;

            this.spacingCalculator.SetPreviousBlock(blockElement);
            if (block != null)
            {
                block.SpacingBefore = spacingBefore;
            }
        }

        public void InsertWatermark(TextWatermarkSettings settings)
        {
            Block block = new Block();
            block.TextProperties.Font = settings.Font;
            block.GraphicProperties.FillColor = settings.ForegroundColor;
            block.InsertText(settings.Text);
            block.Measure();

            this.DrawWatermark(settings, block);
        }

        public void InsertWatermark(ImageWatermarkSettings settings)
        {
            Block block = new Block();
            block.InsertImage(settings.ImageSource, new Size(settings.Width, settings.Height));

            this.DrawWatermark(settings, block, true);
        }

        public bool IsFirstPage()
        {
            return this.document.Pages.Count == 1;
        }

        private void DrawWatermark(WatermarkSettingsBase settings, Block block, bool hasCorrectSize = false)
        {
            using (this.editor.SavePosition())
            {
                double offsetX = (this.actualWidth - settings.Width) / 2 + this.properties.PageMargins.Left;
                double offsetY = (this.actualHeight - this.properties.PageMargins.Top - settings.Height) / 2 + this.properties.PageMargins.Top;

                this.editor.Position.Translate(Math.Max(0, offsetX), Math.Max(0, offsetY));
                this.editor.Position.RotateAt(settings.Angle, settings.Width / 2, settings.Height / 2);

                if (!hasCorrectSize)
                {
                    this.editor.Position.Scale(settings.Width / block.DesiredSize.Width, settings.Height / block.DesiredSize.Height);
                }

                this.editor.DrawBlock(block);
            }
        }
    }
}