﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Graphics;

namespace Telerik.Windows.Documents.Fixed.Model.Editing
{
    /// <summary>
    /// Represents graphic state.
    /// </summary>
    public class GraphicProperties : PropertiesBase<GraphicProperties>
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="GraphicProperties" /> class.
        /// </summary>
        public GraphicProperties()
        {
            this.FillColor = FixedDocumentDefaults.Color;
            this.StrokeColor = FixedDocumentDefaults.Color;
            this.StrokeThickness = FixedDocumentDefaults.StrokeThickness;
            this.IsFilled = true;
            this.IsStroked = true;
        }

        /// <summary>
        /// Gets or sets whether the drawn geometry is filled.
        /// </summary>
        /// <value>Value indicating whether the geometry is filled.</value>
        public bool IsFilled { get; set; }

        /// <summary>
        /// Gets or sets whether the drawn geometry is stroked.
        /// </summary>
        /// <value>Value indicating whether the geometry is stroked.</value>
        public bool IsStroked { get; set; }

        /// <summary>
        /// Gets or sets the fill color.
        /// </summary>
        /// <value>The color of the fill.</value>
        public ColorBase FillColor { get; set; }

        /// <summary>
        /// Gets or sets the stroke color.
        /// </summary>
        /// <value>The color of the stroke.</value>
        public ColorBase StrokeColor { get; set; }

        /// <summary>
        /// Gets or sets the stroke thickness.
        /// </summary>
        /// <value>The stroke thickness.</value>
        public double StrokeThickness { get; set; }

        /// <summary>
        /// Gets or sets the miter limit.
        /// </summary>
        /// <value>The miter limit.</value>
        public double? MiterLimit { get; set; }

        /// <summary>
        /// Gets or sets the stroke dash offset.
        /// </summary>
        /// <value>The stroke dash offset.</value>
        public double StrokeDashOffset { get; set; }

        /// <summary>
        /// Gets or sets the stroke dash array.
        /// </summary>
        /// <value>The stroke dash array.</value>
        public IEnumerable<double> StrokeDashArray { get; set; }

        /// <summary>
        /// Gets or sets the stroke line join.
        /// </summary>
        /// <value>The stroke line join.</value>
        public LineJoin StrokeLineJoin { get; set; }

        /// <summary>
        /// Gets or sets the stroke line cap.
        /// </summary>
        /// <value>The stroke line cap.</value>
        public LineCap StrokeLineCap { get; set; }

        /// <summary>
        /// Copies the graphic properties from fromProperties instance.
        /// </summary>
        /// <param name="graphicProperties">The graphic properties.</param>
        public override void CopyFrom(GraphicProperties graphicProperties)
        {
            this.FillColor = graphicProperties.FillColor;
            this.StrokeColor = graphicProperties.StrokeColor;
            this.StrokeThickness = graphicProperties.StrokeThickness;
            this.MiterLimit = graphicProperties.MiterLimit;
            this.StrokeDashOffset = graphicProperties.StrokeDashOffset;
            this.StrokeDashArray = graphicProperties.StrokeDashArray;
            this.StrokeLineJoin = graphicProperties.StrokeLineJoin;
            this.StrokeLineCap = graphicProperties.StrokeLineCap;
            this.IsFilled = graphicProperties.IsFilled;
            this.IsStroked = graphicProperties.IsStroked;
        }

        internal void CopyFrom(GeometryPropertiesOwner geometry)
        {
            this.FillColor = geometry.Fill;
            this.StrokeColor = geometry.Stroke;
            this.StrokeThickness = geometry.StrokeThickness;
            this.MiterLimit = geometry.MiterLimit;
            this.StrokeDashOffset = geometry.StrokeDashOffset;
            this.StrokeDashArray = geometry.StrokeDashArray;
            this.StrokeLineJoin = geometry.StrokeLineJoin;
            this.StrokeLineCap = geometry.StrokeLineCap;
            this.IsFilled = geometry.IsFilled;
            this.IsStroked = geometry.IsStroked;
        }

        internal void CopyTo(GeometryPropertiesOwner geometry)
        {
            geometry.Fill = this.FillColor;
            geometry.Stroke = this.StrokeColor;
            geometry.StrokeThickness = this.StrokeThickness;
            geometry.MiterLimit = this.MiterLimit;
            geometry.StrokeDashOffset = this.StrokeDashOffset;
            geometry.StrokeDashArray = this.StrokeDashArray;
            geometry.StrokeLineJoin = this.StrokeLineJoin;
            geometry.StrokeLineCap = this.StrokeLineCap;
            geometry.IsFilled = this.IsFilled;
            geometry.IsStroked = this.IsStroked;
        }

    }
}
