﻿using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Model.Editing.Flow;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Fixed.Model.Text;
using Telerik.Windows.Documents.Utilities;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
using Matrix = System.Windows.Media.Matrix;
#endif

namespace Telerik.Windows.Documents.Fixed.Model.Editing.Layout
{
    internal class ContentElementLayoutElementBase<T> : LayoutElementBase
        where T : PositionContentElement
    {
        private readonly RenderingMode renderingMode;
        private readonly ColorBase underlineColor;
        private readonly UnderlinePattern underlinePattern;
        private readonly ColorBase highlightColor;
        private readonly T element;

        internal ContentElementLayoutElementBase(T element,
            double width,
            double trimmedWidth,
            double baselineOffset,
            FontBase font,
            double fontSize,
            RenderingMode renderingMode,
            ColorBase underlineColor,
            UnderlinePattern underlinePattern,
            ColorBase highlightColor)
            : base(width, trimmedWidth, baselineOffset, font, fontSize)
        {
            Guard.ThrowExceptionIfNull(element, "element");

            this.element = element;
            this.underlineColor = underlineColor;
            this.underlinePattern = underlinePattern;
            this.renderingMode = renderingMode;
            this.highlightColor = highlightColor;
        }

        internal ContentElementLayoutElementBase(T element, double width, double trimmedWidth, double baselineOffset, ContentElementLayoutElementBase<T> other)
            : this(element,
                    width,
                    trimmedWidth,
                    baselineOffset,
                    other.Font,
                    other.FontSize,
                    other.RenderingMode,
                    other.UnderlineColor,
                    other.UnderlinePattern,
                    other.HighlightColor)
        {
        }

        public T Element
        {
            get
            {
                return this.element;
            }
        }

        internal RenderingMode RenderingMode
        {
            get
            {
                return this.renderingMode;
            }
        }

        internal ColorBase UnderlineColor
        {
            get
            {
                return this.underlineColor;
            }
        }

        internal UnderlinePattern UnderlinePattern
        {
            get
            {
                return this.underlinePattern;
            }
        }

        internal ColorBase HighlightColor
        {
            get
            {
                return this.highlightColor;
            }
        }

        internal override LineInfo CompleteLine(double maxLineWidth, double currentLineOffsetX, List<LayoutElementBase> elementsInCurrentLine, out IEnumerable<LayoutElementBase> pendingLayoutElements)
        {
            LineInfo line = new LineInfo();
            line.AddRange(elementsInCurrentLine);

            if (elementsInCurrentLine.Count == 0)
            {
                line.Add(this);
                pendingLayoutElements = Enumerable.Empty<LayoutElementBase>();
            }
            else
            {
                pendingLayoutElements = new LayoutElementBase[] { this };
            }

            return line;
        }

        internal override void Draw(DrawLayoutElementContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            if (this.HighlightColor != null)
            {
                Path highlights = this.DrawHighlights(context);
                Matrix highlightsMatrix = new Matrix();
                highlightsMatrix = highlightsMatrix.TranslateMatrix(context.OffsetX, context.OffsetY);
                highlights.Position = new MatrixPosition(highlightsMatrix);
                context.Editor.Draw(highlights);
            }

            Matrix elementMatrix = new Matrix();
            elementMatrix = elementMatrix.TranslateMatrix(context.OffsetX, context.OffsetY);
            elementMatrix = this.Transform(context, elementMatrix);
            this.element.Position = new MatrixPosition(elementMatrix);
            context.Editor.Draw(this.element);

            if (this.RenderingMode != Text.RenderingMode.None && this.UnderlinePattern != UnderlinePattern.None)
            {
                Path underline = this.DrawUnderline(context.Line);
                Matrix underlineMatrix = new Matrix();
                underlineMatrix = underlineMatrix.TranslateMatrix(context.OffsetX, context.OffsetY + context.Line.BaselineOffset + context.Line.UnderlinePosition + context.Line.UnderlineThickness / 2);
                underline.Position = new MatrixPosition(underlineMatrix);
                context.Editor.Draw(underline);
            }
        }

        internal override bool CanFit(double maxTotalWidth, double currentLineOffsetX)
        {
            return maxTotalWidth.IsGreaterThanOrEqualTo(currentLineOffsetX + this.TrimmedWidth, Block.Epsilon);
        }

        protected virtual Matrix Transform(DrawLayoutElementContext context, Matrix transform)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            return transform;
        }

        private Path DrawHighlights(DrawLayoutElementContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            double height;
            switch (context.Block.LineSpacingType)
            {
                case HeightType.Exact:
                case HeightType.AtLeast:
                    height = context.Block.GetActualLineHeight(context.Line);
                    break;

                default:
                    height = context.Line.BaselineOffset + context.Line.Descent;
                    break;
            }

            Rect rect = new Rect(0, 0, this.GetWidthInLine(context.Line), height);
            Path highlight = new Path();
            highlight.Geometry = new RectangleGeometry(rect);
            highlight.IsFilled = true;
            highlight.IsStroked = false;
            highlight.Fill = this.HighlightColor;

            return highlight;
        }

        private Path DrawUnderline(LineInfo line)
        {
            Guard.ThrowExceptionIfNull(line, "line");
            Guard.ThrowExceptionIfNull(this.element, "element");

            Point startPoint = new Point(0, 0);
            Point endPoint = new Point(this.GetWidthInLine(line), 0);

            Path underline = new Path();
            PathGeometry pathGeometry = new PathGeometry();
            PathFigure figure = pathGeometry.Figures.AddPathFigure();
            figure.IsClosed = false;
            figure.StartPoint = startPoint;
            figure.Segments.AddLineSegment(endPoint);
            underline.Geometry = pathGeometry;
            underline.Stroke = this.UnderlineColor;
            underline.IsFilled = false;
            underline.IsStroked = true;
            underline.StrokeThickness = line.UnderlineThickness;

            return underline;
        }
    }
}
