﻿using System;
using System.Collections.Generic;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Fixed.Model.Editing.Collections;
using Telerik.Windows.Documents.Fixed.Model.Editing.Flow;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Editing.Layout
{
    internal class TabLayoutElement : LayoutElementBase
    {
        private const int FontDistance = 1;

        private readonly TabStopCollection tabStopCollection;
        private readonly double defaultTabStopWidth;

        private TabStop tabStop;
        private double currentLineOffsetX;
        private double tabStopTrailingElementsWidth;
        private double maxTotalWidth;
        private double lastTabWidthCached;

        public TabLayoutElement(TabStopCollection tabStopCollection, double defaultTabStopWidth, TextProperties textProperties)
            : base(defaultTabStopWidth, 0, TextFragmentLayoutElement.GetBaselineOffset(textProperties.Font, textProperties.FontSize), textProperties.Font, textProperties.FontSize)
        {
            this.tabStopCollection = tabStopCollection;
            this.defaultTabStopWidth = defaultTabStopWidth;
        }

        internal override bool ReflectsOnTabStopWidth
        {
            get
            {
                return false;
            }
        }

        internal override LineInfo CompleteLine(double maxLineWidth, double currentLineOffsetX, List<LayoutElementBase> elementsInCurrentLine, out IEnumerable<LayoutElementBase> pendingLayoutElements)
        {
            LineInfo line = new LineInfo();
            line.AddRange(elementsInCurrentLine);

            pendingLayoutElements = new List<LayoutElementBase>()
                {
                    this
                };

            return line;
        }

        internal override double GetWidth()
        {
            double tabWidth;
            if (this.defaultTabStopWidth > 0)
            {
                tabWidth = this.defaultTabStopWidth - (this.currentLineOffsetX % this.defaultTabStopWidth);
            }
            else
            {
                tabWidth = 0;
            }

            if (this.tabStopCollection.Count > 0)
            {
                this.tabStop = this.tabStopCollection.Where(t => t.Position >= this.currentLineOffsetX).FirstOrDefault();

                if (this.tabStop != null)
                {
                    bool isSupported = this.tabStop.Type != TabStopType.Bar && this.tabStop.Type != TabStopType.Clear;

                    if (isSupported)
                    {
                        tabWidth = this.tabStop.Position - this.currentLineOffsetX;

                        switch (this.tabStop.Type)
                        {
                            case TabStopType.Decimal:
                            case TabStopType.Right:
                                tabWidth = Math.Max(0, tabWidth - this.tabStopTrailingElementsWidth);
                                break;
                            case TabStopType.Center:
                                tabWidth = Math.Max(0, tabWidth - this.tabStopTrailingElementsWidth / 2);
                                break;
                        }
                    }
                }
            }

            tabWidth = (Math.Min(tabWidth, this.maxTotalWidth));
            this.lastTabWidthCached = tabWidth;

            return tabWidth;
        }

        internal override void PrepareCalculatingWidthOnMeasure(LineMeasureContext context)
        {
            if (this.ShouldCalculateTrailingWidth())
            {
                this.tabStopTrailingElementsWidth = this.GetTrailingElementsWidth(context, this.tabStop.Type == TabStopType.Decimal);
            }
        }

        internal override void Draw(DrawLayoutElementContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            if (this.tabStop == null)
            {
                return;
            }

            switch (this.tabStop.Type)
            {
                case TabStopType.Left:
                case TabStopType.Center:
                case TabStopType.Right:
                case TabStopType.Decimal:
                    this.DrawTabStopLeader(context);
                    break;
                case TabStopType.Bar:
                    this.DrawBarTabStop(context);
                    break;
                case TabStopType.Clear:
                default:
                    break;
            }
        }

        internal override bool CanFit(double maxTotalWidth, double currentLineOffsetX)
        {
            this.maxTotalWidth = maxTotalWidth;
            this.currentLineOffsetX = currentLineOffsetX;
            return this.currentLineOffsetX + this.Width <= maxTotalWidth;
        }

        private bool ShouldCalculateTrailingWidth()
        {
            return this.tabStop != null &&
                (this.tabStop.Type == TabStopType.Center || this.tabStop.Type == TabStopType.Right || this.tabStop.Type == TabStopType.Decimal);
        }

        private double GetTrailingElementsWidth(LineMeasureContext context, bool shouldStopOnDecimal)
        {
            double accumulatedWidth = 0;
            foreach (LayoutElementBase layoutElement in context.PendingElements)
            {
                if (layoutElement.ReflectsOnTabStopWidth)
                {
                    if (shouldStopOnDecimal)
                    {
                        accumulatedWidth += layoutElement.GetElementWidthFromStartToDecimal();
                    }
                    else
                    {
                        accumulatedWidth += layoutElement.Width;
                    }
                }
                else
                {
                    break;
                }

                if (accumulatedWidth >= context.RemainingLineWidth)
                {
                    return context.RemainingLineWidth;
                }
            }

            return accumulatedWidth;
        }

        private void DrawBarTabStop(DrawLayoutElementContext context)
        {
            double offsetX = context.BoundingRectangle.X + this.tabStop.Position;
            double offsetY = context.OffsetY;

            using (context.Editor.SaveGraphicProperties())
            {
                context.Editor.GraphicProperties.StrokeColor = context.Block.GraphicProperties.FillColor;
                context.Editor.DrawLine(new Point(offsetX, offsetY), new Point(offsetX, offsetY + this.Height * 2));
            }
        }

        private void DrawTabStopLeader(DrawLayoutElementContext context)
        {
            if (this.tabStop.Leader == TabStopLeader.None ||
                this.lastTabWidthCached <= FontDistance)
            {
                return;
            }

            double offsetX = context.OffsetX + FontDistance;
            double offsetY = context.OffsetY + this.Height - FontDistance;

            Point startPoint = new Point();
            Point endPoint = new Point();

            using (context.Editor.SaveGraphicProperties())
            {
                context.Editor.GraphicProperties.StrokeColor = context.Block.GraphicProperties.FillColor;
                context.Editor.GraphicProperties.IsFilled = false;
                double lineMiddle = context.OffsetY + this.Height / 2 + context.Line.Descent;
                switch (this.tabStop.Leader)
                {
                    case TabStopLeader.Dot:
                        startPoint = new Point(offsetX, offsetY);
                        endPoint = new Point(this.lastTabWidthCached + offsetX - FontDistance, offsetY);
                        context.Editor.GraphicProperties.StrokeDashArray = new double[] { 0.5, 2 };
                        context.Editor.GraphicProperties.StrokeLineCap = LineCap.Round;
                        break;
                    case TabStopLeader.Hyphen:
                        startPoint = new Point(offsetX, lineMiddle);
                        endPoint = new Point(this.lastTabWidthCached + offsetX - FontDistance, lineMiddle);
                        context.Editor.GraphicProperties.StrokeDashArray = new double[] { 3, 2 };
                        break;
                    case TabStopLeader.Underscore:
                        startPoint = new Point(offsetX, offsetY);
                        endPoint = new Point(offsetX + this.lastTabWidthCached - FontDistance, offsetY);
                        context.Editor.GraphicProperties.StrokeDashOffset = 0;
                        break;
                    case TabStopLeader.MiddleDot:
                        startPoint = new Point(offsetX, lineMiddle);
                        endPoint = new Point(this.lastTabWidthCached + offsetX - FontDistance, lineMiddle);
                        context.Editor.GraphicProperties.StrokeDashArray = new double[] { 0.5, 2 };
                        context.Editor.GraphicProperties.StrokeLineCap = LineCap.Round;
                        break;
                }

                context.Editor.DrawLine(startPoint, endPoint);
            }
        }
    }
}