using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Text;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Editing.Layout
{
    internal class TextFragmentLayoutElement : ContentElementLayoutElementBase<TextFragment>
    {
        private const double SubscriptAndSuperscriptSizeRatio = 0.65;
        private const double SubscriptPositionRatio = -0.141;
        private const double SuperscriptPositionRatio = 0.35;

        public TextFragmentLayoutElement(TextFragment textFragment, TextProperties properties, bool useTextFragmentProperties)
            : base(textFragment,
                MeasureWidth(textFragment),
                MeasureWidth(textFragment, textFragment.Text.TrimEnd()),
                useTextFragmentProperties ? MeasureHeight(textFragment) : GetBaselineOffset(properties.Font, properties.FontSize),
                useTextFragmentProperties ? textFragment.Font : properties.Font,
                useTextFragmentProperties ? textFragment.FontSize : properties.FontSize,
                properties.RenderingMode,
                properties.UnderlineColor,
                properties.UnderlinePattern,
                properties.HighlightColor)
        {
        }

        public TextFragmentLayoutElement(TextFragment textFragment, TextFragmentLayoutElement other)
            : base(textFragment,
                MeasureWidth(textFragment),
                MeasureWidth(textFragment, textFragment.Text.TrimEnd()),
                MeasureHeight(textFragment),
                other)
        {
        }

        internal override bool CanAddToLineEnding
        {
            get
            {
                string text = this.Element.Text;
                return text.Length == 0 || char.IsWhiteSpace(text[0]);
            }
        }

        internal override double GetElementWidthFromStartToDecimal()
        {
            int lastDigitIndex;
            if (this.TryGetDecimalIndex(out lastDigitIndex))
            {
                string text = this.Element.Text.Substring(0, lastDigitIndex);
                return MeasureWidth(this.Element, text);
            }
            else
            {
                return base.GetElementWidthFromStartToDecimal();
            }
        }

        private bool TryGetDecimalIndex(out int lastDigitIndex)
        {
            lastDigitIndex = 0;
            bool isInNumber = false;
            foreach (char ch in this.Element.Text.ToCharArray())
            {
                if (char.IsDigit(ch))
                {
                    isInNumber = true;
                }
                else if (isInNumber)
                {
                    return true;
                }

                lastDigitIndex++;
            }

            return false;
        }

        internal override bool CanSplitToMergeWithNextLayoutElement
        {
            get
            {
                string text = this.Element.Text;
                int whiteSpaceIndex = GetIndexOfLastWhiteSpace(text);

                if (whiteSpaceIndex > -1)
                {
                    int splitIndex = whiteSpaceIndex + 1;

                    if (splitIndex < text.Length)
                    {
                        return true;
                    }
                }

                return false;
            }
        }

        internal override bool CanMergeWithNextLayoutElement
        {
            get
            {
                string text = this.Element.Text;
                return text.Length > 0 && !char.IsWhiteSpace(text[text.Length - 1]);
            }
        }

        public static IEnumerable<LayoutElementBase> CreateTextFragmentLayoutElements(Block block, string text)
        {
            Guard.ThrowExceptionIfNull(block, "block");
            Guard.ThrowExceptionIfNullOrEmpty(text, "text");

            StringBuilder builder = new StringBuilder();
            FontBase font = block.TextProperties.Font;
            FontBase fallbackFont = null;
            foreach (var c in text)
            {
                if (IsCharacterInFont(font, c))
                {
                    if (fallbackFont != null)
                    {
                        TextFragment textFragment;

                        using (block.SaveTextProperties())
                        {
                            block.TextProperties.Font = fallbackFont;
                            textFragment = CreateTextFragment(block, builder.ToString());
                        }

                        yield return new TextFragmentLayoutElement(textFragment, block.TextProperties, false);

                        fallbackFont = null;
                        builder.Clear();
                    }
                }
                else
                {
                    if (fallbackFont == null && builder.Length > 0)
                    {
                        TextFragment textFragment = CreateTextFragment(block, builder.ToString());
                        yield return new TextFragmentLayoutElement(textFragment, block.TextProperties, true);

                        builder.Clear();
                    }

                    // TODO: Remove second condition.
                    if (fallbackFont == null || !IsCharacterInFont(fallbackFont, c))
                    {
                        fallbackFont = FindFallback(font, c);
                    }
                }

                builder.Append(c);
            }

            if (builder.Length > 0)
            {
                if (fallbackFont == null)
                {
                    TextFragment textFragment = CreateTextFragment(block, builder.ToString());
                    yield return new TextFragmentLayoutElement(textFragment, block.TextProperties, true);
                }
                else
                {
                    TextFragment textFragment;
                    using (block.SaveTextProperties())
                    {
                        block.TextProperties.Font = fallbackFont;

                        textFragment = CreateTextFragment(block, builder.ToString());
                    }

                    yield return new TextFragmentLayoutElement(textFragment, block.TextProperties, false);
                }
            }
        }


        public override string ToString()
        {
            return this.Element.ToString();
        }

        protected override Matrix Transform(DrawLayoutElementContext context, Matrix transform)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            return base.Transform(context, transform)
                .TranslateMatrix(0, context.GetLineActualBaselineOffset());
        }

        internal override bool CanFit(double maxTotalWidth, double currentLineOffsetX)
        {
            return this.TrimmedWidth == 0 || base.CanFit(maxTotalWidth, currentLineOffsetX);
        }

        internal override void SplitToMergeWithNextLayoutElement(out LayoutElementBase firstPart, out LayoutElementBase secondPart)
        {
            if (this.CanSplitToMergeWithNextLayoutElement)
            {
                int whiteSpaceIndex = GetIndexOfLastWhiteSpace(this.Element.Text);
                int splitIndex = whiteSpaceIndex + 1;

                string firstText, secondText;
                SplitInTwo(this.Element.Text, splitIndex, out firstText, out secondText);

                firstPart = new TextFragmentLayoutElement(new TextFragment(this.Element, firstText), this);
                secondPart = new TextFragmentLayoutElement(new TextFragment(this.Element, secondText), this);
            }
            else
            {
                base.SplitToMergeWithNextLayoutElement(out firstPart, out secondPart);
            }
        }

        internal override bool TrySplitToAddToLineEnding(out LayoutElementBase firstPart, out LayoutElementBase secondPart)
        {
            string firstText, secondText;

            if (TrySplitAfterFirstWhitespace(this.Element.Text, 0, out firstText, out secondText))
            {
                firstPart = new TextFragmentLayoutElement(new TextFragment(this.Element, firstText), this);
                secondPart = new TextFragmentLayoutElement(new TextFragment(this.Element, secondText), this);

                return true;
            }

            return base.TrySplitToAddToLineEnding(out firstPart, out secondPart);
        }

        internal override LineInfo CompleteLine(double maxLineWidth, double currentLineOffsetX, List<LayoutElementBase> elementsInCurrentLine, out IEnumerable<LayoutElementBase> pendingLayoutElements)
        {
            Guard.ThrowExceptionIfLessThan(maxLineWidth, currentLineOffsetX + this.Width, "lineWidth");

            LineInfo line = null;

            if (!this.TryCompleteLineByFittingFirstPart(maxLineWidth, currentLineOffsetX, elementsInCurrentLine, out line, out pendingLayoutElements))
            {
                if (!this.TryCompleteLineBySplittingCurrentLineElements(elementsInCurrentLine, out line, out pendingLayoutElements))
                {
                    line = this.CompleteLineByForceSplittingThisElement(maxLineWidth, currentLineOffsetX, elementsInCurrentLine, out pendingLayoutElements);
                }
            }

            return line;
        }

        internal override double GetMinWidth()
        {
            // TODO: Cache the MinWidth. Invalidate text property since it can be changed.
            double maxCharWidth = 0;
            for (int charIndex = 0; charIndex < this.Element.Text.Length; charIndex++)
            {
                maxCharWidth = Math.Max(MeasureWidth(this.Element, this.Element.Text[charIndex].ToString()), maxCharWidth);
            }

            return maxCharWidth;
        }

        internal override Size GetMinMeasureSize()
        {
            if (string.IsNullOrEmpty(this.Element.Text))
            {
                return new Size(0, 0);
            }
            else
            {
                string firstChar = this.Element.Text[0].ToString();
                double width = string.IsNullOrWhiteSpace(firstChar) ? 0 : MeasureWidth(this.Element, firstChar);
                double height = MeasureHeight(this.Element);

                return new Size(width, height);
            }
        }

        private LineInfo CompleteLineByForceSplittingThisElement(double maxLineWidth, double currentLineOffsetX, List<LayoutElementBase> elementsInCurrentLine, out IEnumerable<LayoutElementBase> pendingLayoutElements)
        {
            LineInfo line = new LineInfo();
            line.AddRange(elementsInCurrentLine);

            string text = this.Element.Text;
            double firstWidth = MeasureWidth(this.Element, text.Substring(0, 1));
            int splitIndex = 1;

            for (int i = 1; i < text.Length; i++)
            {
                string textPart = text.Substring(0, i + 1);
                double width = MeasureWidth(this.Element, textPart);

                if (width + currentLineOffsetX > maxLineWidth)
                {
                    splitIndex = i;

                    if (!char.IsWhiteSpace(text[splitIndex]))
                    {
                        break;
                    }
                }
            }

            bool isFirstElement = !elementsInCurrentLine.Any();
            bool doesFirstLetterFit = firstWidth + currentLineOffsetX <= maxLineWidth;
            bool shouldSplit = splitIndex < text.Length && !char.IsWhiteSpace(text[splitIndex]);
            shouldSplit = (shouldSplit && doesFirstLetterFit) || (shouldSplit && isFirstElement);

            TextFragmentLayoutElement firstPart = null;

            if (shouldSplit)
            {
                string firstText, secondText;
                SplitInTwo(text, splitIndex, out firstText, out secondText);

                firstPart = new TextFragmentLayoutElement(new TextFragment(this.Element, firstText), this);
                pendingLayoutElements = new TextFragmentLayoutElement[] { new TextFragmentLayoutElement(new TextFragment(this.Element, secondText), this) };
            }
            else if (isFirstElement)
            {
                firstPart = this;
                pendingLayoutElements = Enumerable.Empty<TextFragmentLayoutElement>();
            }
            else
            {
                pendingLayoutElements = new TextFragmentLayoutElement[] { this };
            }

            if (firstPart != null)
            {
                line.Add(firstPart);
            }

            return line;
        }

        private bool TryCompleteLineBySplittingCurrentLineElements(List<LayoutElementBase> elementsInCurrentLine, out LineInfo line, out IEnumerable<LayoutElementBase> pendingLayoutElements)
        {
            int indexOfLastElementToPreserve = elementsInCurrentLine.Count - 1;

            while (indexOfLastElementToPreserve > -1)
            {
                LayoutElementBase elementToTryPreserve = elementsInCurrentLine[indexOfLastElementToPreserve];

                if (elementToTryPreserve.CanMergeWithNextLayoutElement && !elementToTryPreserve.CanSplitToMergeWithNextLayoutElement)
                {
                    indexOfLastElementToPreserve--;
                }
                else
                {
                    break;
                }
            }

            if (indexOfLastElementToPreserve > -1)
            {
                line = this.CompleteLineBySplittingCurrentLineElements(elementsInCurrentLine, indexOfLastElementToPreserve, out pendingLayoutElements);

                return true;
            }

            line = null;
            pendingLayoutElements = null;

            return false;
        }

        private LineInfo CompleteLineBySplittingCurrentLineElements(List<LayoutElementBase> elementsInCurrentLine, int indexOfLastElementToPreserve, out IEnumerable<LayoutElementBase> pendingLayoutElements)
        {
            LayoutElementBase firstSplitPart = null, secondSplitPart = null;
            LayoutElementBase element = elementsInCurrentLine[indexOfLastElementToPreserve];
            bool shouldSplitElement = element.CanSplitToMergeWithNextLayoutElement;

            if (shouldSplitElement)
            {
                element.SplitToMergeWithNextLayoutElement(out firstSplitPart, out secondSplitPart);
            }

            int indexOfFirstElementToRemove = shouldSplitElement ? indexOfLastElementToPreserve : indexOfLastElementToPreserve + 1;
            int countToRemove = elementsInCurrentLine.Count - indexOfFirstElementToRemove;

            List<LayoutElementBase> pendingElementsList = new List<LayoutElementBase>();

            if (secondSplitPart != null)
            {
                pendingElementsList.Add(secondSplitPart);
            }

            if (countToRemove > 0)
            {
                pendingElementsList.AddRange(elementsInCurrentLine.Skip(shouldSplitElement ? indexOfFirstElementToRemove + 1 : indexOfFirstElementToRemove).Take(countToRemove));
                elementsInCurrentLine.RemoveRange(indexOfFirstElementToRemove, countToRemove);
            }

            pendingElementsList.Add(this);

            if (firstSplitPart != null)
            {
                elementsInCurrentLine.Add(firstSplitPart);
            }

            LineInfo line = new LineInfo();
            line.AddRange(elementsInCurrentLine);
            pendingLayoutElements = pendingElementsList;

            return line;
        }

        private bool TryCompleteLineByFittingFirstPart(double maxLineWidth, double currentLineOffsetX, IEnumerable<LayoutElementBase> elementsInCurrentLine, out LineInfo line, out IEnumerable<LayoutElementBase> pendingElements)
        {
            string text = this.Element.Text;
            string firstText, secondText;
            bool doesFirstPartFit = false;
            line = null;
            pendingElements = null;
            TextFragmentLayoutElement firstPart = null;

            if (TrySplitAfterFirstWhitespace(text, 0, out firstText, out secondText))
            {
                firstPart = new TextFragmentLayoutElement(new TextFragment(this.Element, firstText), this);
                doesFirstPartFit = firstPart.CanFit(maxLineWidth, currentLineOffsetX);

                if (doesFirstPartFit)
                {
                    bool isFitting = true;
                    TextFragmentLayoutElement biggestFittingFragment = firstPart;
                    string remainingText = secondText;

                    while (isFitting)
                    {
                        int startIndex = firstText.Length;

                        if (TrySplitAfterFirstWhitespace(text, startIndex, out firstText, out secondText))
                        {
                            firstPart = new TextFragmentLayoutElement(new TextFragment(this.Element, firstText), this);
                            isFitting = firstPart.CanFit(maxLineWidth, currentLineOffsetX);

                            if (isFitting)
                            {
                                biggestFittingFragment = firstPart;
                                remainingText = secondText;
                            }
                        }
                        else
                        {
                            isFitting = false;
                        }
                    }

                    firstPart = biggestFittingFragment;
                    pendingElements = new LayoutElementBase[] { new TextFragmentLayoutElement(new TextFragment(this.Element, remainingText), this) };
                }
                else if (char.IsWhiteSpace(text[0]))
                {
                    pendingElements = new TextFragmentLayoutElement[] { new TextFragmentLayoutElement(new TextFragment(this.Element, secondText), this) };
                    doesFirstPartFit = true;
                }
            }
            else if (string.IsNullOrWhiteSpace(text))
            {
                firstPart = this;
                pendingElements = Enumerable.Empty<LayoutElementBase>();
                doesFirstPartFit = true;
            }

            if (doesFirstPartFit)
            {
                line = new LineInfo();
                line.AddRange(elementsInCurrentLine);
                line.Add(firstPart);
            }

            return doesFirstPartFit;
        }

        public static double GetBaselineOffset(FontBase font, double fontSize)
        {
            Guard.ThrowExceptionIfNull(font, "font");

            return fontSize * (Math.Abs(font.Ascent.Value) + Math.Abs(font.LineGap));
        }

        private static bool TrySplitAfterFirstWhitespace(string text, int startIndex, out string firstPart, out string secondPart)
        {
            firstPart = null;
            secondPart = null;

            int firstWhiteSpaceIndex = GetIndexOfFirstWhiteSpace(text, startIndex);
            int splitIndex = GetIndexOfFirstNonWhiteSpace(text, firstWhiteSpaceIndex + 1);
            bool shouldSplit = firstWhiteSpaceIndex >= 0 && splitIndex > 0 && splitIndex < text.Length;

            if (shouldSplit)
            {
                SplitInTwo(text, splitIndex, out firstPart, out secondPart);
            }

            return shouldSplit;
        }

        private static int GetIndexOfFirstWhiteSpace(string text, int startIndex)
        {
            for (int i = startIndex; i < text.Length; i++)
            {
                if (char.IsWhiteSpace(text[i]))
                {
                    return i;
                }
            }

            return -1;
        }

        private static int GetIndexOfLastWhiteSpace(string text)
        {
            for (int i = text.Length - 1; i >= 0; i--)
            {
                if (char.IsWhiteSpace(text[i]))
                {
                    return i;
                }
            }

            return -1;
        }

        private static int GetIndexOfFirstNonWhiteSpace(string text, int startIndex)
        {
            for (int i = startIndex; i < text.Length; i++)
            {
                if (!char.IsWhiteSpace(text[i]))
                {
                    return i;
                }
            }

            return -1;
        }

        private static TextFragment CreateTextFragment(Block block, string text)
        {
            Guard.ThrowExceptionIfNull(block, "block");
            Guard.ThrowExceptionIfNullOrEmpty(text, "text");

            TextFragment textFragment;
            textFragment = new TextFragment();
            textFragment.Text = text;
            block.GraphicProperties.CopyTo(textFragment.TextProperties.GeometryProperties);
            block.TextProperties.CopyTo(textFragment.TextProperties);

            if (block.TextProperties.BaselineAlignment != Flow.BaselineAlignment.Baseline)
            {
                double height = block.TextProperties.FontSize * block.TextProperties.Font.Height;
                textFragment.FontSize = block.TextProperties.FontSize * SubscriptAndSuperscriptSizeRatio;

                switch (block.TextProperties.BaselineAlignment)
                {
                    case Flow.BaselineAlignment.Subscript:
                        textFragment.TextRise = height * SubscriptPositionRatio;
                        break;
                    case Flow.BaselineAlignment.Superscript:
                        textFragment.TextRise = height * SuperscriptPositionRatio;
                        break;
                }
            }

            return textFragment;
        }

        internal static double MeasureWidth(TextFragment textFragment)
        {
            Guard.ThrowExceptionIfNull(textFragment, "textFragment");

            return MeasureWidth(textFragment.TextCollection, textFragment.FontSize, textFragment.WordSpacing, textFragment.CharacterSpacing, textFragment.HorizontalScaling);
        }

        internal static double MeasureHeight(TextFragment textFragment)
        {
            Guard.ThrowExceptionIfNull(textFragment, "textFragment");

            return GetBaselineOffset(textFragment.Font, textFragment.FontSize);
        }

        internal static double MeasureWidth(CharCode charCode, FontBase font, double fontSize, double? wordSpacing, double? characterSpacing, double? horizontalScaling)
        {
            double width = font.GetWidth(charCode.Code) / FontBase.FontRatio;
            double tx = (width * fontSize);

            if (wordSpacing.HasValue && charCode.Code == 32 && charCode.Size == 1)
            {
                tx += wordSpacing.Value;
            }

            if (characterSpacing.HasValue)
            {
                tx += characterSpacing.Value;
            }

            if (horizontalScaling.HasValue)
            {
                tx *= horizontalScaling.Value;
            }

            return tx;
        }

        private static double MeasureWidth(TextFragment textFragment, string text)
        {
            Guard.ThrowExceptionIfNull(textFragment, "textFragment");
            Guard.ThrowExceptionIfNull(text, "text");

            if (string.IsNullOrEmpty(text))
            {
                return 0;
            }

            return MeasureWidth(new TextCollection(textFragment.Font, text), textFragment.FontSize, textFragment.WordSpacing, textFragment.CharacterSpacing, textFragment.HorizontalScaling);
        }

        private static double MeasureWidth(TextCollection textCollection, double fontSize, double? wordSpacing, double? characterSpacing, double? horizontalScaling)
        {
            Guard.ThrowExceptionIfNull(textCollection, "textCollection");

            double result = 0;

            foreach (CharInfo glyph in textCollection.Characters)
            {
                double tx = MeasureWidth(glyph.CharCode, textCollection.Font, fontSize, wordSpacing, characterSpacing, horizontalScaling);

                result += tx;
            }

            return result;
        }

        private static bool IsCharacterInFont(FontBase font, char c)
        {
            ushort glyphId;

            return font.TryGetGlyphId(c, out glyphId);
        }

        private static FontBase FindFallback(FontBase original, char c)
        {
            FallbackRange fallbackRange = FallbackRange.GetFallbackRange(c);
            if (fallbackRange == null)
            {
                return FallbackRange.DefaultRange.GetFallbackFont(original, c);
            }

            return fallbackRange.GetFallbackFont(original, c);
        }

        private static void SplitInTwo(string text, int secondStartIndex, out string firstPart, out string secondPart)
        {
            firstPart = text.Substring(0, secondStartIndex);
            secondPart = text.Substring(secondStartIndex);
        }
    }
}