﻿using System;
using System.Text;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
#else
using System.Windows.Media;
#endif

namespace Telerik.Windows.Documents.Fixed.Model.Editing.Lists
{
    internal static class ListFactory
    {
        private static readonly string[] bullets;
        private static readonly FontFamily[] bulletsFontFamily;
        private static readonly Func<int, string>[] numberingStyles;

        static ListFactory()
        {
            bullets = new string[] 
            {
                "",
                "o",
                ""
            };

            bulletsFontFamily = new FontFamily[] 
            {
                new FontFamily("Symbol"),
                new FontFamily("Courier New"),
                new FontFamily("Wingdings")
            };

            numberingStyles = new Func<int, string>[]
            {
                BulletNumberingFormats.GetNumber,
                BulletNumberingFormats.GetLowerLetter,
                BulletNumberingFormats.GetLowerRomanNumber,
            };
        }

        public static void InitializeList(List list, ListTemplateType listTemplateType)
        {
            switch (listTemplateType)
            {
                case ListTemplateType.BulletDefault:
                    InitializeBulletDefaultList(list);
                    break;
                case ListTemplateType.NumberedDefault:
                    InitializeNumberedDefaultList(list);
                    break;
                case ListTemplateType.NumberedParentheses:
                    InitializeNumberedParenthesesList(list);
                    break;
                case ListTemplateType.NumberedHierarchical:
                    InitializeNumberedHierarchicalList(list);
                    break;
                default:
                    throw new NotSupportedException(string.Format("ListTemplateType {0} is not supported!", listTemplateType));
            }
        }

        public static void InitializeBulletDefaultList(List list)
        {
            InitializeDefaultIndentationList(list, (level, levelIndex) =>
            {
                level.BulletNumberingFormat = new TextBulletNumberingFormat((indexer) => { return bullets[levelIndex % bullets.Length]; });
                level.CharacterProperties.TrySetFont(bulletsFontFamily[levelIndex % bulletsFontFamily.Length]);
            });
        }

        public static void InitializeNumberedDefaultList(List list)
        {
            InitializeDefaultIndentationList(list, (level, levelIndex) =>
                {
                    level.BulletNumberingFormat = new TextBulletNumberingFormat((indexer) => 
                    {
                        return string.Format("{0}.", numberingStyles[levelIndex % numberingStyles.Length](indexer.GetCurrentIndex(levelIndex))); 
                    });
                });
        }

        public static void InitializeNumberedParenthesesList(List list)
        {
            InitializeDefaultIndentationList(list, (level, levelIndex) =>
            {
                level.BulletNumberingFormat = new TextBulletNumberingFormat((indexer) =>
                {
                    return string.Format("{0})", numberingStyles[levelIndex % numberingStyles.Length](indexer.GetCurrentIndex(levelIndex)));
                });
            });
        }

        public static void InitializeNumberedHierarchicalList(List list)
        {
            InitializeDefaultIndentationList(list, (level, levelIndex) =>
            {
                level.BulletNumberingFormat = new TextBulletNumberingFormat((indexer) =>
                {
                    StringBuilder builder = new StringBuilder();

                    for (int i = 0; i <= levelIndex; i++)
                    {
                        builder.AppendFormat("{0}.", BulletNumberingFormats.GetNumber(indexer.GetCurrentIndex(i)));
                    }

                    return builder.ToString();
                });
            });
        }

        private static void InitializeDefaultIndentationList(List list, Action<ListLevel, int> onListLevelCreated)
        {
            for (int level = 0; level < FixedDocumentDefaults.DefaultListLevelsCount; level++)
            {
                ListLevel listLevel = list.Levels.AddListLevel();

                onListLevelCreated(listLevel, level);

                double indentationStep = FixedDocumentDefaults.ListLevelIndentationStep;
                listLevel.ParagraphProperties.LeftIndent = indentationStep * 2 + (level * indentationStep);
                listLevel.ParagraphProperties.FirstLineIndent = -indentationStep;
            }
        }
    }
}
