﻿using Telerik.Windows.Documents.Fixed.Model.Editing.Flow;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Editing.Lists
{
    /// <summary>
    /// Represent a list level.
    /// </summary>
    public class ListLevel
    {
        private readonly CharacterProperties characterProperties;
        private readonly ParagraphProperties paragraphProperties;
        private IBulletNumberingFormat numberingStyleFormat;
        private double indentAfterBullet;
        private int startIndex;

        /// <summary>
        /// Initializes a new instance of the <see cref="ListLevel" /> class.
        /// </summary>
        public ListLevel()
            : this(Telerik.Windows.Documents.Fixed.Model.Editing.Lists.BulletNumberingFormats.EmptyNumbering)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ListLevel" /> class.
        /// </summary>
        /// <param name="numberingStyleFormat">The numbering style format.</param>
        public ListLevel(IBulletNumberingFormat numberingStyleFormat)
        {
            this.IsAddedToListLevelCollection = false;
            this.BulletNumberingFormat = numberingStyleFormat;
            this.characterProperties = new CharacterProperties();
            this.paragraphProperties = new ParagraphProperties();
            this.StartIndex = FixedDocumentDefaults.DefaultListLevelStartIndex;
            this.RestartAfterLevel = FixedDocumentDefaults.DefaultListLevelRestartAfterLevel;
            this.IndentAfterBullet = FixedDocumentDefaults.DefaultIndentAfterBullet;
        }

        /// <summary>
        /// Gets or sets the start index.
        /// </summary>
        /// <value>The start index.</value>
        public int StartIndex
        {
            get
            {
                return this.startIndex;
            }
            set
            {
                if (this.startIndex != value)
                {
                    this.startIndex = value;
                    this.RestartCurrentNumber();
                }
            }
        }

        /// <summary>
        /// Gets or sets the restart after level.
        /// </summary>
        /// <value>The restart after level.</value>
        public int RestartAfterLevel { get; set; }

        /// <summary>
        /// Gets the bullet numbering character properties.
        /// </summary>
        /// <value>The bullet numbering character properties.</value>
        public CharacterProperties CharacterProperties
        {
            get
            {
                return this.characterProperties;
            }
        }

        /// <summary>
        /// Gets the paragraph properties.
        /// </summary>
        /// <value>The paragraph properties.</value>
        public ParagraphProperties ParagraphProperties
        {
            get
            {
                return this.paragraphProperties;
            }
        }

        /// <summary>
        /// Gets or sets the indent after the bullet.
        /// </summary>
        /// <value>The indent after the bullet.</value>
        public double IndentAfterBullet
        {
            get
            {
                return this.indentAfterBullet;
            }
            set
            {
                Guard.ThrowExceptionIfLessThan(0, value, "value");
                this.indentAfterBullet = value;
            }
        }

        /// <summary>
        /// Gets or sets the bullet numbering format.
        /// </summary>
        /// <value>The bullet numbering format.</value>
        public IBulletNumberingFormat BulletNumberingFormat
        {
            get
            {
                return this.numberingStyleFormat;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "value");
                this.numberingStyleFormat = value;
            }
        }

        internal bool IsAddedToListLevelCollection { get; set; }

        internal int LastInsertedIndex { get; set; }

        internal void RestartCurrentNumber()
        {
            this.LastInsertedIndex = this.StartIndex - 1;
        }
    }
}
