﻿using System;
using Telerik.Windows.Documents.Fixed.Model.Editing.Collections;
using Telerik.Windows.Documents.Fixed.Model.Editing.Flow;
using Telerik.Windows.Documents.Fixed.Model.Editing.Tables;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Utilities;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Primitives;
#else
using FontFamily = System.Windows.Media.FontFamily;
using FontStyle = System.Windows.FontStyle;
using FontWeight = System.Windows.FontWeight;
using Size = System.Windows.Size;
#endif

namespace Telerik.Windows.Documents.Fixed.Model.Editing
{
    /// <summary>
    /// Represents editor that will allow you to create RadFixedDocument
    /// </summary>
    public class RadFixedDocumentEditor : IDisposable
    {
        internal static readonly Size MinimalMeasureSize = new Size(0, 0);
        internal static readonly Size MinimalWidthMeasureSize = new Size(0, double.PositiveInfinity);

        private readonly RadFixedDocument document;
        private readonly ListCollection lists;
        private readonly SectionInfo section;
        private readonly Block paragraph;

        private readonly SectionProperties sectionProperties;
        private readonly ParagraphProperties paragraphProperties;
        private readonly CharacterProperties characterProperties;
        private bool isParagraphStarted;

        /// <summary>
        /// Initializes a new instance of the <see cref="RadFixedDocumentEditor" /> class.
        /// </summary>
        /// <param name="document">The document.</param>
        public RadFixedDocumentEditor(RadFixedDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");

            this.document = document;

            this.section = new SectionInfo(this.document);
            this.paragraph = new Block();
            this.lists = new ListCollection();

            this.sectionProperties = new SectionProperties();
            this.paragraphProperties = new ParagraphProperties();
            this.characterProperties = new CharacterProperties();
            this.isParagraphStarted = false;
        }

        internal event EventHandler<RadFixedPageCreatedEventArgs> PageCreated;

        /// <summary>
        /// Gets the document.
        /// </summary>
        /// <value>The document.</value>
        public RadFixedDocument Document
        {
            get
            {
                return this.document;
            }
        }

        /// <summary>
        /// Gets the lists.
        /// </summary>
        /// <value>The lists.</value>
        public ListCollection Lists
        {
            get
            {
                return this.lists;
            }
        }

        /// <summary>
        /// Gets the section properties.
        /// </summary>
        /// <value>The section properties.</value>
        public SectionProperties SectionProperties
        {
            get
            {
                return this.sectionProperties;
            }
        }

        /// <summary>
        /// Gets the paragraph properties.
        /// </summary>
        /// <value>The paragraph properties.</value>
        public ParagraphProperties ParagraphProperties
        {
            get
            {
                return this.paragraphProperties;
            }
        }

        /// <summary>
        /// Gets the character properties.
        /// </summary>
        /// <value>The character properties.</value>
        public CharacterProperties CharacterProperties
        {
            get
            {
                return this.characterProperties;
            }
        }

        /// <summary>
        /// Inserts new paragraph.
        /// </summary>
        public void InsertParagraph()
        {
            this.AssureSectionExists();
            this.EndParagraph();
            this.StartParagraph();
        }

        /// <summary>
        /// Inserts the text.
        /// </summary>
        /// <param name="text">The text.</param>
        public void InsertRun(string text)
        {
            Guard.ThrowExceptionIfNullOrEmpty(text, "text");

            this.AssureParagraphExists();
            this.CharacterProperties.CopyTo(this.paragraph);
            this.paragraph.InsertText(text);
        }

        /// <summary>
        /// Inserts the text using the given font family.
        /// </summary>
        /// <param name="fontFamily">The font family.</param>
        /// <param name="text">The text.</param>
        public void InsertRun(FontFamily fontFamily, string text)
        {
            Guard.ThrowExceptionIfNull(fontFamily, "fontFamily");
            Guard.ThrowExceptionIfNullOrEmpty(text, "text");

            this.CharacterProperties.TrySetFont(fontFamily);
            this.InsertRun(text);
        }

        /// <summary>
        /// Inserts the text using the given font family, font style and font weight.
        /// </summary>
        /// <param name="fontFamily">The font family.</param>
        /// <param name="fontStyle">The font style.</param>
        /// <param name="fontWeight">The font weight.</param>
        /// <param name="text">The text.</param>
        public void InsertRun(FontFamily fontFamily, FontStyle fontStyle, FontWeight fontWeight, string text)
        {
            Guard.ThrowExceptionIfNull(fontFamily, "fontFamily");
            Guard.ThrowExceptionIfNull(fontStyle, "fontStyle");
            Guard.ThrowExceptionIfNull(fontWeight, "fontWeight");
            Guard.ThrowExceptionIfNullOrEmpty(text, "text");

            this.CharacterProperties.TrySetFont(fontFamily, fontStyle, fontWeight);
            this.InsertRun(text);
        }

        /// <summary>
        /// Inserts the text and moves to the next line.
        /// </summary>
        /// <param name="text">The text.</param>
        public void InsertLine(string text)
        {
            Guard.ThrowExceptionIfNullOrEmpty(text, "text");

            this.InsertRun(text);
            this.InsertLineBreak();
        }

        /// <summary>
        /// Inserts a line break.
        /// </summary>
        public void InsertLineBreak()
        {
            this.AssureParagraphExists();
            this.CharacterProperties.CopyTo(this.paragraph);
            this.paragraph.InsertLineBreak();
        }

        /// <summary>
        /// Inserts the image inline.
        /// </summary>
        /// <param name="imageSource">The image source.</param>
        /// <param name="size">The size.</param>
        public void InsertImageInline(ImageSource imageSource, Size size)
        {
            Guard.ThrowExceptionIfNull(imageSource, "imageSource");

            this.AssureParagraphExists();
            this.CharacterProperties.CopyTo(this.paragraph);
            this.paragraph.InsertImage(imageSource, size);
        }

        /// <summary>
        /// Inserts the image inline.
        /// </summary>
        /// <param name="imageSource">The image source.</param>
        public void InsertImageInline(ImageSource imageSource)
        {
            Guard.ThrowExceptionIfNull(imageSource, "imageSource");

            this.AssureParagraphExists();
            this.CharacterProperties.CopyTo(this.paragraph);
            this.paragraph.InsertImage(imageSource);
        }

        /// <summary>
        /// Inserts the form inline.
        /// </summary>
        /// <param name="formSource">The form source.</param>
        public void InsertFormInline(FormSource formSource)
        {
            Guard.ThrowExceptionIfNull(formSource, "formSource");

            this.AssureParagraphExists();
            this.CharacterProperties.CopyTo(this.paragraph);
            this.paragraph.InsertForm(formSource);
        }

        /// <summary>
        /// Inserts the form inline.
        /// </summary>
        /// <param name="formSource">The form source.</param>
        /// <param name="size">The form size.</param>
        public void InsertFormInline(FormSource formSource, Size size)
        {
            Guard.ThrowExceptionIfNull(formSource, "formSource");

            this.AssureParagraphExists();
            this.CharacterProperties.CopyTo(this.paragraph);
            this.paragraph.InsertForm(formSource, size);
        }

        /// <summary>
        /// Inserts new section.
        /// </summary>
        public void InsertSectionBreak()
        {
            this.StartNewPage(this.SectionProperties);
        }

        /// <summary>
        /// Inserts new page.
        /// </summary>
        public void InsertPageBreak()
        {
            this.AssureSectionExists();
            this.StartNewPage();
        }

        /// <summary>
        /// Inserts the table.
        /// </summary>
        /// <param name="table">The table.</param>
        public void InsertTable(Table table)
        {
            Guard.ThrowExceptionIfNull(table, "table");

            this.InsertBlock(table);
        }

        /// <summary>
        /// Inserts the block element.
        /// </summary>
        /// <param name="block">The block.</param>
        public void InsertBlock(IBlockElement block)
        {
            Guard.ThrowExceptionIfNull(block, "block");

            this.AssureSectionExists();
            this.EndParagraph();
            if (!this.section.CanFit(block))
            {
                this.StartNewPage();
            }

            this.section.InsertBlockElement(block);

            while (block.HasPendingContent)
            {
                block = block.Split();
                this.StartNewPage();
                this.section.InsertBlockElement(block);
            }
        }

        private void StartNewPage()
        {
            this.StartNewPage(this.section.Properties);
        }

        private void StartNewPage(SectionProperties sectionProperties)
        {
            this.EndParagraph();
            this.section.StartNewPage(sectionProperties);
            this.OnPageCreated(this.section);
            this.section.BeginExportContent();
        }

        /// <summary>
        /// Finishes the the pages and closes the document.
        /// </summary>
        public void Dispose()
        {
            this.EndParagraph();
        }

        private void AssureSectionExists()
        {
            if (!this.section.IsStarted)
            {
                this.InsertSectionBreak();
            }
        }

        private void AssureParagraphExists()
        {
            this.AssureSectionExists();
            this.StartParagraph();
        }

        private void StartParagraph()
        {
            if (!this.isParagraphStarted)
            {
                this.ParagraphProperties.CopyTo(this.paragraph);

                if (this.ParagraphProperties.ListId.HasValue)
                {
                    Telerik.Windows.Documents.Fixed.Model.Editing.Lists.List list;
                    if (!this.Lists.TryGetList(this.ParagraphProperties.ListId.Value, out list))
                    {
                        throw new ArgumentException(String.Format("There is no list with id {0}", this.ParagraphProperties.ListId.Value));
                    }

                    this.paragraph.SetBullet(list, this.ParagraphProperties.ListLevel);
                }

                this.isParagraphStarted = true;
            }
        }

        private void EndParagraph()
        {
            if (this.isParagraphStarted)
            {
                this.isParagraphStarted = false;
                this.InsertBlock(this.paragraph);
                this.paragraph.Clear();
            }
        }

        private void OnPageCreated(SectionInfo section)
        {
            if (this.PageCreated != null)
            {
                this.PageCreated(this, new RadFixedPageCreatedEventArgs(section));
            }
        }
    }
}
