﻿using System;
using System.Collections.Generic;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Editing.Collections;
using Telerik.Windows.Documents.Fixed.Model.Editing.Flow;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Editing.Tables
{
    /// <summary>
    /// Represent a table.
    /// </summary>
    public class Table : IBlockElement
    {
        private readonly int firstRowIndex;
        private readonly TableRowCollection rows;
        private readonly List<TableRow> pendingRows;

        private Dictionary<int, double> fixedColumnWidths;
        private Dictionary<int, double> minColumnWidths;
        private List<double> finalColumnWidths;
        private double[] columnLefts;
        private double borderSpacing;
        private TableBorders borders;
        private bool hasCells;

        /// <summary>
        /// Initializes a new instance of the <see cref="Table" /> class.
        /// </summary>
        public Table()
            : this(0)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Table" /> class.
        /// </summary>
        /// <param name="firstRowIndex">The index of the first row.</param>
        private Table(int firstRowIndex)
        {
            this.firstRowIndex = firstRowIndex;
            this.rows = new TableRowCollection();
            this.pendingRows = new List<TableRow>();
            this.fixedColumnWidths = new Dictionary<int, double>();
            this.finalColumnWidths = new List<double>();
            this.minColumnWidths = new Dictionary<int, double>();

            this.Borders = new TableBorders();
            this.Background = FixedDocumentDefaults.BackgroundColor;
            this.DefaultCellProperties = new CellProperties()
            {
                Padding = FixedDocumentDefaults.DefaultCellPadding,
                Background = FixedDocumentDefaults.BackgroundColor
            };

            this.LayoutType = FixedDocumentDefaults.DefaultLayoutType;
            this.BorderSpacing = FixedDocumentDefaults.DefaultBorderSpacing;
            this.BorderCollapse = FixedDocumentDefaults.DefaultBorderCollapse;
            this.IsLastRowSplit = false;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Table" /> class.
        /// </summary>
        /// <param name="other">The other.</param>
        /// <param name="firstRowIndex">The index of the first row.</param>
        private Table(Table other, int firstRowIndex)
            : this(firstRowIndex)
        {
            this.Borders = other.Borders;
            this.Margin = other.Margin;
            this.Background = other.Background;
            this.DefaultCellProperties = other.DefaultCellProperties;
            this.BorderSpacing = other.BorderSpacing;
            this.BorderCollapse = other.BorderCollapse;
            this.fixedColumnWidths = other.fixedColumnWidths;
            this.minColumnWidths = other.minColumnWidths;
            this.PreferredWidth = other.PreferredWidth;
        }

        /// <summary>
        /// Gets or sets the margin.
        /// </summary>
        /// <value>The margin.</value>
        public Thickness Margin { get; set; }

        /// <summary>
        /// Gets or sets the layout type of the table.
        /// </summary>
        /// <value>The layout type.</value>
        public TableLayoutType LayoutType { get; set; }

        /// <summary>
        /// Gets the borders.
        /// </summary>
        /// <value>The borders.</value>
        public TableBorders Borders
        {
            get
            {
                return this.borders;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "value");

                this.borders = value;
            }
        }

        /// <summary>
        /// Gets or sets the background of the table.
        /// </summary>
        /// <value>The background of the table.</value>
        public ColorBase Background { get; set; }

        /// <summary>
        /// Gets the default cell properties.
        /// </summary>
        /// <value>The default cell properties.</value>
        public CellProperties DefaultCellProperties { get; private set; }

        /// <summary>
        /// Gets the desired size.
        /// </summary>
        /// <value>The desired size.</value>
        public Size DesiredSize { get; private set; }

        /// <summary>
        /// Gets or sets the border spacing.
        /// </summary>
        /// <value>The border spacing.</value>
        public double BorderSpacing
        {
            get
            {
                return this.borderSpacing;
            }
            set
            {
                Guard.ThrowExceptionIfLessThan(0, value, "value");
                this.borderSpacing = value;
            }
        }

        /// <summary>
        /// Gets or sets borders collapse state.
        /// </summary>
        /// <value>Value indicating the border collapse state.</value>
        public BorderCollapse BorderCollapse { get; set; }

        /// <summary>
        /// Gets the rows.
        /// </summary>
        /// <value>The rows.</value>
        public TableRowCollection Rows
        {
            get
            {
                return this.rows;
            }
        }

        /// <summary>
        /// Gets if there is any pending content after the last measure.
        /// </summary>
        /// <value>
        /// Boolean value indicating whether there is any pending content after the last measure.
        /// </value>
        public bool HasPendingContent
        {
            get
            {
                return this.pendingRows.Any();
            }
        }

        internal double? PreferredWidth { get; set; }

        internal IEnumerable<TableCell> Cells
        {
            get
            {
                for (int i = 0; i < this.rows.Count; i++)
                {
                    IEnumerable<TableCell> ownedCells;
                    if (i == 0)
                    {
                        ownedCells = this.rows[i].AllIntersectingCells;
                    }
                    else
                    {
                        ownedCells = this.rows[i].Cells;
                    }

                    foreach (TableCell cell in ownedCells)
                    {
                        yield return cell;
                    }
                }
            }
        }

        private Rect BordersLayoutRectangle
        {
            get
            {
                Thickness thickness = this.Borders.Thickness;

                double left = -this.BorderSpacing / 2;
                double top = -this.BorderSpacing / 2;
                double width = this.DesiredSize.Width - this.Margin.Left - this.Margin.Right - (thickness.Left + thickness.Right) / 2;
                double height = this.DesiredSize.Height - this.Margin.Top - this.Margin.Bottom - (thickness.Top + thickness.Bottom) / 2;

                if (this.BorderCollapse == Tables.BorderCollapse.Separate)
                {
                    left -= thickness.Left / 2;
                    top -= thickness.Top / 2;
                }

                return new Rect(left, top, width, height);
            }
        }

        private double WidthMargins
        {
            get
            {
                Thickness thickness = this.Borders.Thickness;
                double widthMargins = this.Margin.Left + this.Margin.Right + this.BorderSpacing;

                if (this.BorderCollapse == Tables.BorderCollapse.Collapse)
                {
                    widthMargins += (thickness.Left + thickness.Right) / 2;
                }
                else
                {
                    widthMargins += thickness.Left + thickness.Right;
                }

                return widthMargins;
            }
        }

        private double HeightMargins
        {
            get
            {
                Thickness thickness = this.Borders.Thickness;
                double heightMargins = this.Margin.Top + this.Margin.Bottom + this.BorderSpacing;

                if (this.BorderCollapse == Tables.BorderCollapse.Collapse)
                {
                    heightMargins += (thickness.Top + thickness.Bottom) / 2;
                }
                else
                {
                    heightMargins += thickness.Top + thickness.Bottom;
                }

                return heightMargins;
            }
        }

        private bool IsSplitTable { get; set; }

        private bool IsLastRowSplit { get; set; }

        /// <summary>
        /// Measures this instance in infinity.
        /// </summary>
        /// <returns>The measured size.</returns>
        public Size Measure()
        {
            return this.Measure(new Size(double.PositiveInfinity, double.PositiveInfinity));
        }

        /// <summary>
        /// Measures the block element with specified available size.
        /// </summary>
        /// <param name="availableSize">The available size.</param>
        /// <returns>The result size.</returns>
        public Size Measure(Size availableSize)
        {
            this.OnBeforeMeasure();

            if (this.TryCalculateDesiredSizeOnMinimalSizesScenarios(availableSize))
            {
                return this.DesiredSize;
            }

            double widthMargins = this.WidthMargins;
            double heightMargins = this.HeightMargins;
            availableSize = new Size(Math.Max(0, availableSize.Width - widthMargins), Math.Max(0, availableSize.Height - heightMargins));

            if (!IsSplitTable)
            {
                this.MeasureInfiniteWidthLayout();

                double newAvailableWidth;
                if (this.ShouldMeasureFiniteWidthLayout(availableSize.Width, out newAvailableWidth))
                {
                    this.MeasureFiniteWidthLayout(newAvailableWidth);
                }

                this.ExpandAccordingToTablePreferredWidth();

                this.CalculateColumnLefts();
            }

            this.CalculateRowHeights();
            this.CalculateRowTopsAndSplit(availableSize.Height);

            foreach (TableCell cell in this.Cells)
            {
                this.CalculateCellLayoutRectangle(cell, this.rows.Count - 1);
            }

            TableRow lastRow = this.rows[this.rows.Count - 1];
            double tableHeight = lastRow.Top + lastRow.Height;
            double tableWidth = this.columnLefts[this.columnLefts.Length - 1] + this.finalColumnWidths[this.finalColumnWidths.Count - 1];
            this.DesiredSize = new Size(tableWidth + widthMargins, tableHeight + heightMargins);

            return this.DesiredSize;
        }

        private void ExpandAccordingToTablePreferredWidth()
        {
            if (this.PreferredWidth.HasValue)
            {
                double totalColumnWidth = this.GetTotalColumnWidth();

                if (totalColumnWidth < this.PreferredWidth)
                {
                    IEnumerable<int> toExpand = this.DetermineColumnsToExpand();

                    double difference = this.PreferredWidth.Value - totalColumnWidth;
                    double differencePercolumn = difference / toExpand.Count();

                    foreach(int columnIndex in toExpand)
                    {
                        double proposedNewWidth = this.finalColumnWidths[columnIndex] + differencePercolumn;
                        this.finalColumnWidths[columnIndex] = proposedNewWidth;
                    }
                }
            }
        }

        private IEnumerable<int> DetermineColumnsToExpand()
        {
            if (this.fixedColumnWidths.Count == this.finalColumnWidths.Count)
            {
                return this.fixedColumnWidths.Keys;
            }

            List<int> columns = new List<int>();

            for(int i = 0; i < this.finalColumnWidths.Count; i++)
            {
                if(!this.fixedColumnWidths.ContainsKey(i))
                {
                    columns.Add(i);
                }
            }

            return columns;
        }

        private bool ShouldMeasureFiniteWidthLayout(double availableWidth, out double newAvailableWidth)
        {
            bool isWidthFinite = this.PreferredWidth.HasValue|| !(double.IsInfinity(availableWidth) || double.IsNaN(availableWidth));

            newAvailableWidth = this.PreferredWidth.HasValue ? this.PreferredWidth.Value : availableWidth;

            if (isWidthFinite)
            {
                if (this.LayoutType == TableLayoutType.FixedWidth)
                {
                    return true;
                }
                else if (this.LayoutType == TableLayoutType.AutoFit)
                {
                    double totalColumnWidth = 0;

                    for (int i = 0; i < this.finalColumnWidths.Count; i++)
                    {
                        totalColumnWidth += this.finalColumnWidths[i];
                    }

                   return totalColumnWidth > newAvailableWidth;
                }
            }

            return false;
        }

        private double GetTotalColumnWidth()
        {
            double totalColumnWidth = 0;

            for (int i = 0; i < this.finalColumnWidths.Count; i++)
            {
                totalColumnWidth += this.finalColumnWidths[i];
            }

            return totalColumnWidth;
        }

        private bool TryCalculateDesiredSizeOnMinimalSizesScenarios(Size availableSize)
        {
            if (!this.hasCells)
            {
                this.DesiredSize = new Size(this.WidthMargins, this.HeightMargins);
                return true;
            }

            if (availableSize == RadFixedDocumentEditor.MinimalMeasureSize)
            {
                this.DesiredSize = this.MeasureMinimalSplitSize();
                return true;
            }

            return false;
        }

        /// <summary>
        /// Draws the block element with the specified editor.
        /// </summary>
        /// <param name="editor">The editor.</param>
        /// <param name="boundingRect">The bounding rect.</param>
        public void Draw(FixedContentEditor editor, Rect boundingRect)
        {
            Guard.ThrowExceptionIfNull(editor, "editor");

            this.Measure(new Size(boundingRect.Width, boundingRect.Height));

            using (editor.SavePosition())
            {
                using (editor.SaveProperties())
                {
                    Thickness thickness = this.Borders.Thickness;
                    double offsetX = boundingRect.X + this.Margin.Left + this.BorderSpacing / 2;
                    double offsetY = boundingRect.Y + this.Margin.Top + this.BorderSpacing / 2;

                    if (this.BorderCollapse == Tables.BorderCollapse.Collapse)
                    {
                        offsetX += thickness.Left / 2;
                        offsetY += thickness.Top / 2;
                    }
                    else
                    {
                        offsetX += thickness.Left;
                        offsetY += thickness.Top;
                    }

                    //// This is because the editor uses SimplePosition.
                    offsetX += editor.Position.Matrix.OffsetX;
                    offsetY += editor.Position.Matrix.OffsetY;
                    editor.Position.Translate(offsetX, offsetY);

                    Rect tableBordersRect = this.BordersLayoutRectangle;
                    this.DrawBackground(editor, this.Background, tableBordersRect, this.Borders.Thickness);

                    foreach (TableCell cell in this.Cells)
                    {
                        ColorBase background = cell.Background ?? this.DefaultCellProperties.Background;
                        this.DrawBackground(editor, background, cell.BordersLayoutRectangle, cell.BordersThickness);
                    }

                    this.Borders.Draw(editor, tableBordersRect);

                    foreach (TableCell cell in this.Cells)
                    {
                        this.DrawCellContent(editor, cell);
                        cell.ActualBorders.Draw(editor, cell.BordersLayoutRectangle);
                    }
                }
            }
        }

        /// <summary>
        /// Splits this instance.
        /// </summary>
        /// <returns>The remaining block element.</returns>
        public Table Split()
        {
            int firstRowIndex = this.firstRowIndex + (this.IsLastRowSplit ? this.rows.Count - 1 : this.rows.Count);

            Table table = new Table(this, firstRowIndex);
            table.rows.AddRange(this.pendingRows);
            table.hasCells = this.hasCells && table.rows.Count > 0;
            table.fixedColumnWidths = this.fixedColumnWidths;
            table.finalColumnWidths = this.finalColumnWidths;
            table.minColumnWidths = this.minColumnWidths;
            table.columnLefts = this.columnLefts;

            this.FixSharedMergedCells(table);

            table.IsSplitTable = true;

            return table;
        }

        /// <summary>
        /// Splits this instance.
        /// </summary>
        /// <returns>The remaining block element.</returns>
        IBlockElement IBlockElement.Split()
        {
            return this.Split();
        }

        private Size MeasureMinimalSplitSize()
        {
            IEnumerable<TableCell> firstRowCells;

            if (this.IsSplitTable)
            {
                firstRowCells = this.rows[0].AllIntersectingCells;
            }
            else
            {
                firstRowCells = this.rows[0].Cells;
            }

            double zeroHeight = 0;
            double minimalHeight = 0;

            foreach (TableCell cell in firstRowCells)
            {
                double fixedWidth = this.GetFixedWidth(cell);
                Size desiredSize = cell.Measure(new Size(fixedWidth, zeroHeight));

                minimalHeight = Math.Max(minimalHeight, desiredSize.Height);
            }

            double minimalWidth = this.GetTotalFixedWidth();
            minimalWidth += this.WidthMargins;
            minimalHeight += this.HeightMargins;

            return new Size(minimalWidth, minimalHeight);
        }

        private double GetTotalFixedWidth()
        {
            double fixedWidth = 0;

            foreach (KeyValuePair<int, double> columnToFixedWidth in this.fixedColumnWidths)
            {
                fixedWidth += columnToFixedWidth.Value;
            }

            return fixedWidth;
        }

        private double GetFixedWidth(TableCell cell)
        {
            double totalFixedWidth = 0;

            for (int column = cell.FromColumn; column <= cell.ToColumn; column++)
            {
                double fixedWidth;
                if (this.fixedColumnWidths.TryGetValue(column, out fixedWidth))
                {
                    totalFixedWidth += fixedWidth;
                }
            }

            return totalFixedWidth;
        }

        private int GetRelativeRowIndex(int absoluteRowIndex)
        {
            return absoluteRowIndex - this.firstRowIndex;
        }

        private int GetAbsoluteRowIndex(int relativeRowIndex)
        {
            return relativeRowIndex + this.firstRowIndex;
        }

        private void CalculateRowHeights()
        {
            foreach (TableCell cell in this.Cells)
            {
                int topIndex = Math.Max(0, this.GetRelativeRowIndex(cell.FromRow));
                int bottomIndex = this.GetRelativeRowIndex(cell.ToRow);
                double currentHeight = this.GetRowsHeight(topIndex, bottomIndex);

                this.CalculateCellLayoutRectangle(cell, this.rows.Count - 1);
                cell.Measure(new Size(cell.LayoutRectangle.Width, double.MaxValue));

                double cellContentHeight = cell.DesiredSize.Height;

                if (cellContentHeight > currentHeight)
                {
                    double deltaHeight = (cellContentHeight - currentHeight) / (bottomIndex - topIndex + 1);

                    for (int rowIndex = topIndex; rowIndex <= bottomIndex; rowIndex++)
                    {
                        this.rows[rowIndex].Height += deltaHeight;
                    }
                }
            }
        }

        private void OnBeforeMeasure()
        {
            IEnumerable<TableRow> pendingsToAdd = this.IsLastRowSplit ? this.pendingRows.Skip(1) : this.pendingRows;
            this.rows.AddRange(pendingsToAdd);
            this.pendingRows.Clear();

            if (this.IsSplitTable)
            {
                foreach (TableRow row in this.rows)
                {
                    row.Height = 0;
                    row.Top = 0;
                }
            }
            else
            {
                foreach (TableRow row in this.rows)
                {
                    row.Height = 0;
                    row.Top = 0;
                    row.AllIntersectingCells.Clear();
                }

                this.columnLefts = null;
                this.minColumnWidths.Clear();
                this.fixedColumnWidths.Clear();
                this.finalColumnWidths.Clear();

                this.PrepareCellsAndArrangeColumns();
            }
        }

        private void FixSharedMergedCells(Table splitTable)
        {
            if (this.rows.Count == 0 || splitTable.rows.Count == 0)
            {
                return;
            }

            if (this.IsLastRowSplit)
            {
                this.FixSharedMergedCellsWhenIsLastRowSplit(splitTable);
            }
            else
            {
                this.FixSharedMergedCellsWhenNoRowIsSplit(splitTable);
            }
        }

        private void FixSharedMergedCellsWhenNoRowIsSplit(Table splitTable)
        {
            int lastRelativeRowIndex = this.rows.Count - 1;
            TableRow lastRow = this.rows[lastRelativeRowIndex];
            int lastRowIndex = this.GetAbsoluteRowIndex(lastRelativeRowIndex);
            int maxAffectedRowIndex = -1;

            Dictionary<int, TableCell> fromColumnToSplitMergedCell = new Dictionary<int, TableCell>();

            foreach (TableCell cell in lastRow.AllIntersectingCells)
            {
                if (cell.ToRow > lastRowIndex)
                {
                    TableCell splitMergedCell = cell.Split();
                    fromColumnToSplitMergedCell.Add(splitMergedCell.FromColumn, splitMergedCell);
                    maxAffectedRowIndex = Math.Max(maxAffectedRowIndex, cell.ToRow);
                }
            }

            if (!fromColumnToSplitMergedCell.Any())
            {
                return;
            }

            int affectedRowsCount = maxAffectedRowIndex - lastRowIndex;

            for (int rowIndex = 0; rowIndex < affectedRowsCount; rowIndex++)
            {
                TableRow row = splitTable.rows[rowIndex].Copy();
                splitTable.Rows[rowIndex] = row;

                for (int i = 0; i < row.AllIntersectingCells.Count; i++)
                {
                    TableCell cell = row.AllIntersectingCells[i];

                    if (cell.FromRow <= lastRowIndex)
                    {
                        row.AllIntersectingCells[i] = fromColumnToSplitMergedCell[cell.FromColumn];
                    }
                }
            }
        }

        private void FixSharedMergedCellsWhenIsLastRowSplit(Table splitTable)
        {
            TableRow splitRow = splitTable.rows[0];
            int splitRowIndex = splitTable.firstRowIndex;

            Dictionary<int, TableCell> fromColumnToSplitMergedCell = new Dictionary<int, TableCell>();
            int maxAffectedRowIndex = -1;

            foreach (TableCell cell in splitRow.AllIntersectingCells)
            {
                if (cell.ToRow > splitRowIndex)
                {
                    fromColumnToSplitMergedCell.Add(cell.FromColumn, cell);
                    maxAffectedRowIndex = Math.Max(maxAffectedRowIndex, cell.ToRow);
                }
            }

            if (!fromColumnToSplitMergedCell.Any())
            {
                return;
            }

            int affectedRowsCount = maxAffectedRowIndex - splitRowIndex;

            for (int rowIndex = 1; rowIndex <= affectedRowsCount; rowIndex++)
            {
                TableRow row = splitTable.rows[rowIndex].Copy();
                splitTable.rows[rowIndex] = row;

                for (int i = 0; i < row.AllIntersectingCells.Count; i++)
                {
                    TableCell cell = row.AllIntersectingCells[i];

                    if (cell.FromRow <= splitRowIndex)
                    {
                        row.AllIntersectingCells[i] = fromColumnToSplitMergedCell[cell.FromColumn];
                    }
                }
            }
        }

        private void DrawCellContent(FixedContentEditor editor, TableCell cell)
        {
            Guard.ThrowExceptionIfNull(editor, "editor");
            Guard.ThrowExceptionIfNull(cell, "cell");

            cell.Draw(editor);
        }

        private void DrawBackground(FixedContentEditor editor, ColorBase background, Rect bordersMiddleLinesRectangle, Thickness bordersThickness)
        {
            if (background != null)
            {
                editor.GraphicProperties.IsFilled = true;
                editor.GraphicProperties.IsStroked = false;
                editor.GraphicProperties.FillColor = background;

                double left = bordersMiddleLinesRectangle.Left + bordersThickness.Left / 2;
                double top = bordersMiddleLinesRectangle.Top + bordersThickness.Top / 2;
                double width = Math.Max(0, bordersMiddleLinesRectangle.Width - (bordersThickness.Left + bordersThickness.Right) / 2);
                double height = Math.Max(0, bordersMiddleLinesRectangle.Height - (bordersThickness.Top + bordersThickness.Bottom) / 2);

                editor.DrawRectangle(new Rect(left, top, width, height));
            }
        }

        private void MeasureFiniteWidthLayout(double tableFitWidth)
        {
            double totalFloatingWidth = 0;
            double totalFixedWidth = 0;
            double totalFixedMinWidth = 0;

            for (int columnIndex = 0; columnIndex < this.finalColumnWidths.Count; columnIndex++)
            {
                double fixedWidth;

                if (this.fixedColumnWidths.TryGetValue(columnIndex, out fixedWidth))
                {
                    totalFixedWidth += fixedWidth;
                    totalFixedMinWidth += Math.Max(fixedWidth, this.minColumnWidths[columnIndex]);
                }
                else
                {
                    totalFloatingWidth += this.finalColumnWidths[columnIndex];
                }
            }

            if (totalFixedMinWidth > tableFitWidth)
            {
                double totalWidth = totalFloatingWidth + totalFixedWidth;

                for (int columnIndex = 0; columnIndex < this.finalColumnWidths.Count; columnIndex++)
                {
                    double fixedWidth;
                    double portion;
                    if (this.fixedColumnWidths.TryGetValue(columnIndex, out fixedWidth))
                    {
                        portion = fixedWidth / totalWidth;
                    }
                    else
                    {
                        portion = this.finalColumnWidths[columnIndex] / totalWidth;
                    }

                    this.finalColumnWidths[columnIndex] = portion * tableFitWidth;
                }
            }
            else
            {
                double availableFloatingWidth = tableFitWidth - totalFixedMinWidth;

                for (int columnIndex = 0; columnIndex < this.finalColumnWidths.Count; columnIndex++)
                {
                    if (!this.fixedColumnWidths.ContainsKey(columnIndex))
                    {
                        double portion = this.finalColumnWidths[columnIndex] / totalFloatingWidth;
                        this.finalColumnWidths[columnIndex] = portion * availableFloatingWidth;
                    }
                }
            }
        }

        private void MeasureInfiniteWidthLayout()
        {
            foreach (KeyValuePair<int, double> columnToFixedWidth in this.fixedColumnWidths)
            {
                this.finalColumnWidths[columnToFixedWidth.Key] = columnToFixedWidth.Value;
            }

            for (int column = 0; column < this.finalColumnWidths.Count; column++)
            {
                this.finalColumnWidths[column] = Math.Max(this.finalColumnWidths[column], this.minColumnWidths[column]);
            }

            foreach (TableCell cell in this.Cells)
            {
                this.FitCellWidth(cell);
            }
        }

        private void PrepareCellsAndArrangeColumns()
        {
            this.hasCells = false;

            for (int i = 0; i < this.rows.Count; i++)
            {
                TableRow row = this.rows[i];
                Stack<TableCell> mergedCells = new Stack<TableCell>(row.AllIntersectingCells.OrderByDescending((cell) => cell.FromColumn));
                row.AllIntersectingCells.Clear();
                int currentColumnIndex = 0;

                foreach (TableCell cell in row.Cells)
                {
                    this.hasCells = true;
                    this.SetDefaultProperties(cell);
                    currentColumnIndex = Table.MoveToNextFreeColumn(currentColumnIndex, row, mergedCells);

                    cell.FromRow = i;
                    cell.FromColumn = currentColumnIndex;
                    currentColumnIndex = cell.ToColumn + 1;

                    row.AllIntersectingCells.Add(cell);

                    this.MarkUsedCellRange(cell);

                    if (cell.ColumnSpan <= 1)
                    {
                        this.SetFixedColumnWidths(cell);
                    }
                }

                row.AllIntersectingCells.AddRange(mergedCells);
            }

            foreach (TableCell cell in this.Cells)
            {
                if (cell.ColumnSpan > 1)
                {
                    this.SetFixedColumnWidths(cell);
                }

                this.SetMinimumColumnWidths(cell);
            }
        }

        private void SetMinimumColumnWidths(TableCell cell)
        {
            double averageMinColumnWidths = cell.CalculateNonSplittableCellWidth() / cell.ColumnSpan;

            for (int column = cell.FromColumn; column <= cell.ToColumn; column++)
            {
                this.minColumnWidths[column] = this.minColumnWidths.ContainsKey(column) ?
                                               Math.Max(averageMinColumnWidths, this.minColumnWidths[column]) :
                                               averageMinColumnWidths;
            }
        }

        private void SetDefaultProperties(TableCell cell)
        {
            cell.DefaultCellProperties = this.DefaultCellProperties;
            cell.BorderSpacing = this.BorderSpacing;
            cell.BorderCollapse = this.BorderCollapse;
        }

        private static int MoveToNextFreeColumn(int currentColumn, TableRow row, Stack<TableCell> mergedCells)
        {
            TableCell next = mergedCells.Count > 0 ? mergedCells.Peek() : null;

            while (next != null && next.FromColumn <= currentColumn)
            {
                currentColumn = Math.Max(currentColumn, next.ToColumn + 1);
                row.AllIntersectingCells.Add(mergedCells.Pop());
                next = mergedCells.Count > 0 ? mergedCells.Peek() : null;
            }

            return currentColumn;
        }

        private void CalculateCellLayoutRectangle(TableCell cell, int maxBottomRowIndex)
        {
            int topRowIndex = Math.Max(0, this.GetRelativeRowIndex(cell.FromRow));
            int bottomRowIndex = Math.Min(maxBottomRowIndex, this.GetRelativeRowIndex(cell.ToRow));

            double left = this.columnLefts[cell.FromColumn];
            double top = this.rows[topRowIndex].Top;
            double width = this.columnLefts[cell.ToColumn] + this.finalColumnWidths[cell.ToColumn] - left;
            double height = this.rows[bottomRowIndex].Bottom - top;

            cell.LayoutRectangle = new Rect(left, top, width, height);
        }

        private void CalculateRowTopsAndSplit(double maxTableHeight)
        {
            int currentRowIndex = 0;
            TableRow currentRow = this.rows[currentRowIndex];
            currentRow.Top = 0;

            if (currentRow.Bottom <= maxTableHeight)
            {
                for (currentRowIndex = 1; currentRowIndex < this.rows.Count; currentRowIndex++)
                {
                    currentRow = this.rows[currentRowIndex];
                    currentRow.Top = this.rows[currentRowIndex - 1].Bottom;

                    if (currentRow.Bottom > maxTableHeight)
                    {
                        break;
                    }
                }
            }

            if (currentRowIndex < this.rows.Count)
            {
                this.SplitAndMoveToPending(currentRowIndex, maxTableHeight);
            }
        }

        private void SplitAndMoveToPending(int splitRowIndex, double maxTableHeight)
        {
            TableRow splitRow;

            if (this.TrySplitRow(splitRowIndex, maxTableHeight, out splitRow))
            {
                this.pendingRows.Add(splitRow);
                this.MoveRowsToPending(splitRowIndex + 1);
            }
            else if (splitRowIndex == 0)
            {
                this.MoveRowsToPending(splitRowIndex + 1);
            }
            else
            {
                this.MoveRowsToPending(splitRowIndex);
            }
        }

        private bool TrySplitRow(int rowIndex, double maxTableHeight, out TableRow splitRow)
        {
            Guard.ThrowExceptionIfLessThan(0, maxTableHeight, "maxTableHeight");

            splitRow = null;
            this.IsLastRowSplit = this.CheckIfRowCanBeSplit(rowIndex, maxTableHeight);

            if (this.IsLastRowSplit)
            {
                TableRow row = this.rows[rowIndex];
                row.Height = 0;
                splitRow = new TableRow();

                foreach (TableCell cell in row.AllIntersectingCells)
                {
                    double cellHeightInRow = cell.DesiredSize.Height - (row.Top - cell.LayoutRectangle.Top);
                    row.Height = Math.Max(row.Height, cellHeightInRow);

                    TableCell splitCell = cell.Split();
                    splitRow.AllIntersectingCells.Add(splitCell);
                }
            }
            else if (rowIndex > 0)
            {
                this.MeasureLastRowMergedCells(rowIndex - 1);
            }

            return this.IsLastRowSplit;
        }

        private void MeasureLastRowMergedCells(int lastRowIndex)
        {
            TableRow lastRow = this.rows[lastRowIndex];

            foreach (TableCell cell in lastRow.AllIntersectingCells)
            {
                if (cell.ToRow >= lastRowIndex + 1)
                {
                    this.CalculateCellLayoutRectangle(cell, lastRowIndex);
                    Rect layoutRect = cell.LayoutRectangle;
                    cell.Measure(new Size(layoutRect.Width, layoutRect.Height));
                }
            }
        }

        private bool CheckIfRowCanBeSplit(int splitRowIndex, double maxTableHeight)
        {
            bool canSplitRow = true;
            TableRow row = this.rows[splitRowIndex];

            foreach (TableCell cell in row.AllIntersectingCells)
            {
                this.CalculateCellLayoutRectangle(cell, splitRowIndex);
                double maxCellHeight = maxTableHeight - cell.LayoutRectangle.Top;
                Size desiredSize = cell.Measure(new Size(cell.LayoutRectangle.Width, maxCellHeight));

                if (desiredSize.Height > maxCellHeight)
                {
                    canSplitRow = false;
                }
            }

            return canSplitRow;
        }

        private void MoveRowsToPending(int firstRowIndex)
        {
            for (int i = firstRowIndex; i < this.rows.Count; i++)
            {
                this.pendingRows.Add(this.rows[i]);
            }

            int rowsToRemove = this.rows.Count - firstRowIndex;

            if (rowsToRemove > 0)
            {
                this.rows.RemoveRange(firstRowIndex, rowsToRemove);
            }
        }

        private void CalculateColumnLefts()
        {
            this.columnLefts = new double[this.finalColumnWidths.Count];

            for (int column = 1; column < this.finalColumnWidths.Count; column++)
            {
                this.columnLefts[column] = this.columnLefts[column - 1] + this.finalColumnWidths[column - 1];
            }
        }

        private void SetFixedColumnWidths(TableCell cell)
        {
            if (cell.PreferredWidth.HasValue)
            {
                double currentFixedWidth = 0;
                int floatingColumns = 0;

                for (int column = cell.FromColumn; column <= cell.ToColumn; column++)
                {
                    double currentWidth;
                    if (this.fixedColumnWidths.TryGetValue(column, out currentWidth))
                    {
                        currentFixedWidth += currentWidth;
                    }
                    else
                    {
                        floatingColumns++;
                    }
                }

                double remainingWidth = Math.Max(0, cell.PreferredWidth.Value - currentFixedWidth);

                if (remainingWidth > 0)
                {
                    if (floatingColumns > 0)
                    {
                        double averageWidth = remainingWidth / floatingColumns;

                        for (int column = cell.FromColumn; column <= cell.ToColumn; column++)
                        {
                            if (!this.fixedColumnWidths.ContainsKey(column))
                            {
                                this.fixedColumnWidths[column] = averageWidth;
                            }
                        }
                    }
                    else
                    {
                        double averageWidth = remainingWidth / cell.ColumnSpan;

                        for (int column = cell.FromColumn; column <= cell.ToColumn; column++)
                        {
                            this.fixedColumnWidths[column] += averageWidth;
                        }
                    }
                }
            }
        }

        private void FitCellWidth(TableCell cell)
        {
            double fixedWidth, floatingWidth;
            int floatingColumnsCount = this.GetFloatingColumnsCount(cell.FromColumn, cell.ToColumn, out fixedWidth, out floatingWidth);

            double fitWidth = floatingColumnsCount == 0 ? fixedWidth : double.PositiveInfinity;
            Size cellContentSize = cell.Measure(new Size(fitWidth, double.PositiveInfinity));

            if (floatingColumnsCount > 0)
            {
                double currentWidth = fixedWidth + floatingWidth;

                if (cellContentSize.Width > currentWidth)
                {
                    double deltaWidth = (cellContentSize.Width - currentWidth) / floatingColumnsCount;

                    for (int column = cell.FromColumn; column <= cell.ToColumn; column++)
                    {
                        if (!this.fixedColumnWidths.ContainsKey(column))
                        {
                            this.finalColumnWidths[column] += deltaWidth;
                        }
                    }
                }
            }
        }

        private double GetRowsHeight(int fromRelativeRowIndex, int toRelativeRowIndex)
        {
            double height = 0;

            for (int rowIndex = fromRelativeRowIndex; rowIndex <= toRelativeRowIndex; rowIndex++)
            {
                height += this.rows[rowIndex].Height;
            }

            return height;
        }

        private int GetFloatingColumnsCount(int fromColumnIndex, int toColumnIndex, out double fixedWidth, out double floatingWidth)
        {
            fixedWidth = 0;
            floatingWidth = 0;
            int floatingColumnsCount = 0;

            for (int columnIndex = fromColumnIndex; columnIndex <= toColumnIndex; columnIndex++)
            {
                double fixedColumnWidth;

                if (this.fixedColumnWidths.TryGetValue(columnIndex, out fixedColumnWidth))
                {
                    fixedWidth += fixedColumnWidth;
                }
                else
                {
                    floatingWidth += this.finalColumnWidths[columnIndex];
                    floatingColumnsCount++;
                }
            }

            return floatingColumnsCount;
        }

        private void MarkUsedCellRange(TableCell cell)
        {
            this.AddMissingRows(cell.ToRow);
            this.AddMissingColumns(cell.ToColumn);

            for (int rowIndex = cell.FromRow + 1; rowIndex <= cell.ToRow; rowIndex++)
            {
                rows[rowIndex].AllIntersectingCells.Add(cell);
            }
        }

        private void AddMissingColumns(int usedColumnIndex)
        {
            while (this.finalColumnWidths.Count <= usedColumnIndex)
            {
                this.finalColumnWidths.Add(0);
            }
        }

        private void AddMissingRows(int usedRowIndex)
        {
            while (this.rows.Count <= usedRowIndex)
            {
                this.Rows.AddTableRow();
            }
        }
    }
}
