﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Editing.Collections;
using Telerik.Windows.Documents.Fixed.Model.Editing.Flow;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Editing.Tables
{
    /// <summary>
    /// Represents a cell in the table.
    /// </summary>
    public class TableCell
    {
        private readonly BlockCollection blocks;
        private readonly List<IBlockElement> pendingBlocks;

        private int rowSpan;
        private int columnSpan;
        private TableCellBorders borders;

        internal TableCell()
        {
            this.Borders = new TableCellBorders();
            this.blocks = new BlockCollection();
            this.pendingBlocks = new List<IBlockElement>();
            this.rowSpan = 1;
            this.columnSpan = 1;
            this.IsLastBlockSplit = false;
        }

        private TableCell(TableCell other)
            : this()
        {
            this.Padding = other.Padding;
            this.Borders = other.Borders;
            this.Background = other.Background;
            this.DefaultCellProperties = other.DefaultCellProperties;
            this.BorderSpacing = other.BorderSpacing;
            this.BorderCollapse = other.BorderCollapse;
            this.FromRow = other.FromRow;
            this.FromColumn = other.FromColumn;
            this.RowSpan = other.RowSpan;
            this.ColumnSpan = other.ColumnSpan;
        }

        /// <summary>
        /// Gets or sets the row span.
        /// </summary>
        /// <value>The row span.</value>
        public int RowSpan
        {
            get
            {
                return this.rowSpan;
            }
            set
            {
                Guard.ThrowExceptionIfLessThan(1, value, "value");
                this.rowSpan = value;
            }
        }

        /// <summary>
        /// Gets or sets the column span.
        /// </summary>
        /// <value>The column span.</value>
        public int ColumnSpan
        {
            get
            {
                return this.columnSpan;
            }
            set
            {
                Guard.ThrowExceptionIfLessThan(1, value, "value");
                this.columnSpan = value;
            }
        }

        /// <summary>
        /// Gets or sets the preferred width.
        /// </summary>
        /// <value>The preferred width.</value>
        public double? PreferredWidth { get; set; }

        /// <summary>
        /// Gets the blocks for the cell.
        /// </summary>
        /// <value>The blocks.</value>
        public BlockCollection Blocks
        {
            get
            {
                return this.blocks;
            }
        }

        /// <summary>
        /// Gets or sets the padding.
        /// </summary>
        /// <value>The padding.</value>
        public Thickness? Padding { get; set; }

        /// <summary>
        /// Gets the borders.
        /// </summary>
        /// <value>The borders.</value>
        public TableCellBorders Borders
        {
            get
            {
                return this.borders;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "value");
                this.borders = value;
            }
        }

        /// <summary>
        /// Gets or sets the background.
        /// </summary>
        /// <value>The background.</value>
        public ColorBase Background { get; set; }

        /// <summary>
        /// Gets the size.
        /// </summary>
        /// <value>The size.</value>
        public Size Size
        {
            get
            {
                return new Size(this.LayoutRectangle.Width, this.LayoutRectangle.Height);
            }
        }

        /// <summary>
        /// Gets the desired size.
        /// </summary>
        /// <value>The value of the desired size.</value>
        internal Size DesiredSize
        {
            get
            {
                double width = this.DesiredContentSize.Width + this.WidthPaddings;
                double height = this.DesiredContentSize.Height + this.HeightPaddings;

                return new Size(width, height);
            }
        }

        internal Rect LayoutRectangle { get; set; }
        internal int FromRow { get; set; }
        internal int FromColumn { get; set; }
        internal CellProperties DefaultCellProperties { get; set; }
        internal double BorderSpacing { get; set; }
        internal BorderCollapse BorderCollapse { get; set; }

        internal int ToRow
        {
            get
            {
                return this.FromRow + this.RowSpan - 1;
            }
        }

        internal int ToColumn
        {
            get
            {
                return this.FromColumn + this.ColumnSpan - 1;
            }
        }

        internal bool HasPendingContent
        {
            get
            {
                return this.pendingBlocks.Count > 0;
            }
        }

        internal Rect BordersLayoutRectangle
        {
            get
            {
                Thickness thickness = this.BordersThickness;

                double left = this.LayoutRectangle.Left + this.BorderSpacing / 2;
                double top = this.LayoutRectangle.Top + this.BorderSpacing / 2;
                double width = this.LayoutRectangle.Width - this.BorderSpacing;
                double height = this.LayoutRectangle.Height - this.BorderSpacing;

                if (this.BorderCollapse == Tables.BorderCollapse.Separate)
                {
                    left += thickness.Left / 2;
                    top += thickness.Left / 2;
                    width -= (thickness.Left + thickness.Right) / 2;
                    height -= (thickness.Top + thickness.Bottom) / 2;
                }

                width = Math.Max(0, width);
                height = Math.Max(0, height);

                return new Rect(left, top, width, height);
            }
        }

        internal Thickness BordersThickness
        {
            get
            {
                return this.ActualBorders.Thickness;
            }
        }

        internal TableCellBorders ActualBorders
        {
            get
            {
                Border left = this.Borders.Left ?? this.DefaultCellProperties.Borders.Left;
                Border top = this.Borders.Top ?? this.DefaultCellProperties.Borders.Top;
                Border right = this.Borders.Right ?? this.DefaultCellProperties.Borders.Right;
                Border bottom = this.Borders.Bottom ?? this.DefaultCellProperties.Borders.Bottom;
                Border diagonalUp = this.Borders.DiagonalUp ?? this.DefaultCellProperties.Borders.DiagonalUp;
                Border diagonalDown = this.Borders.DiagonalDown ?? this.DefaultCellProperties.Borders.DiagonalDown;

                return new TableCellBorders(left, top, right, bottom, diagonalUp, diagonalDown);
            }
        }

        private Thickness ActualPadding
        {
            get
            {
                return this.Padding.HasValue ? this.Padding.Value : this.DefaultCellProperties.Padding;
            }
        }

        private double WidthPaddings
        {
            get
            {
                Thickness bordersThickness = this.BordersThickness;
                Thickness actualPadding = this.ActualPadding;
                double widthPaddings = actualPadding.Left + actualPadding.Right + this.BorderSpacing;

                if (this.BorderCollapse == Tables.BorderCollapse.Collapse)
                {
                    widthPaddings += (bordersThickness.Left + bordersThickness.Right) / 2;
                }
                else
                {
                    widthPaddings += bordersThickness.Left + bordersThickness.Right;
                }

                return widthPaddings;
            }
        }

        private double HeightPaddings
        {
            get
            {
                Thickness bordersThickness = this.BordersThickness;
                Thickness actualPadding = this.ActualPadding;
                double heightPaddings = actualPadding.Top + actualPadding.Bottom + this.BorderSpacing;

                if (this.BorderCollapse == Tables.BorderCollapse.Collapse)
                {
                    heightPaddings += (bordersThickness.Top + bordersThickness.Bottom) / 2;
                }
                else
                {
                    heightPaddings += bordersThickness.Top + bordersThickness.Bottom;
                }

                return heightPaddings;
            }
        }

        private Size DesiredContentSize
        {
            get;
            set;
        }

        private bool IsLastBlockSplit
        {
            get;
            set;
        }

        internal void Draw(FixedContentEditor editor)
        {
            Guard.ThrowExceptionIfNull(editor, "editor");

            Thickness bordersThickness = this.BordersThickness;
            Thickness actualPadding = this.ActualPadding;
            double availableWidth = Math.Max(0, this.LayoutRectangle.Width - this.WidthPaddings);
            double availableHeight = Math.Max(0, this.LayoutRectangle.Height - this.HeightPaddings);

            double offsetX = this.LayoutRectangle.X + actualPadding.Left + this.BorderSpacing / 2;
            double offsetY = this.LayoutRectangle.Y + actualPadding.Top + this.BorderSpacing / 2;
            double width = this.DesiredContentSize.Width.IsEqualTo(availableWidth) ?
                           Math.Max(this.DesiredContentSize.Width, availableWidth) :
                           availableWidth;
            double height = this.DesiredContentSize.Height.IsEqualTo(availableHeight) ?
                            Math.Max(this.DesiredContentSize.Height, availableHeight) :
                            availableHeight;

            if (this.BorderCollapse == Tables.BorderCollapse.Collapse)
            {
                offsetX += bordersThickness.Left / 2;
                offsetY += bordersThickness.Top / 2;
            }
            else
            {
                offsetX += bordersThickness.Left;
                offsetY += bordersThickness.Top;
            }

            using (editor.PushTransformedClipping(new Rect(offsetX, offsetY, width, height)))
            {
                foreach (IBlockElement blockElement in this.blocks)
                {
                    blockElement.Draw(editor, new Rect(offsetX, offsetY, width, height));
                    offsetY += blockElement.DesiredSize.Height;
                    height -= blockElement.DesiredSize.Height;
                }
            }
        }

        /// <summary>
        /// Measures this instance in infinity.
        /// </summary>
        /// <returns>The result size.</returns>
        internal Size Measure()
        {
            return this.Measure(new Size(double.PositiveInfinity, double.PositiveInfinity));
        }

        /// <summary>
        /// Measures the cell in the specified size.
        /// </summary>
        /// <param name="availableSize">The available.</param>
        /// <returns>The result size.</returns>
        internal Size Measure(Size availableSize)
        {
            double widthPaddings = this.WidthPaddings;
            double heightPaddings = this.HeightPaddings;
            availableSize = new Size(Math.Max(0, availableSize.Width - widthPaddings), Math.Max(0, availableSize.Height - heightPaddings));

            double width = 0;
            double height = 0;
            double remainingHeight = availableSize.Height;

            IEnumerable<IBlockElement> pendingsToAdd = this.IsLastBlockSplit ? this.pendingBlocks.Skip(1) : this.pendingBlocks;
            this.blocks.AddRange(pendingsToAdd);
            this.pendingBlocks.Clear();

            for (int i = 0; i < this.blocks.Count && remainingHeight >= 0; i++)
            {
                IBlockElement blockElement = this.blocks[i];
                Size remainingSize = new Size(availableSize.Width, remainingHeight);
                Size blockSize = blockElement.Measure(remainingSize);

                if (height + blockSize.Height > availableSize.Height || blockElement.HasPendingContent)
                {
                    bool isFirstBlock = i == 0;
                    this.IsLastBlockSplit = this.TrySplitAndMoveToPending(i, remainingSize);

                    bool shouldAddHeight = this.IsLastBlockSplit || isFirstBlock;

                    if (!shouldAddHeight)
                    {
                        break;
                    }
                }

                if (blockSize.Width > width)
                {
                    width = blockSize.Width;
                }

                height += blockSize.Height;
                remainingHeight = availableSize.Height - height;
            }

            this.DesiredContentSize = new Size(width, height);

            return this.DesiredSize;
        }

        internal double CalculateNonSplittableCellWidth()
        {
            double maxWidth = 0;

            foreach (IBlockElement block in this.blocks)
            {
                block.Measure(RadFixedDocumentEditor.MinimalWidthMeasureSize);
                maxWidth = Math.Max(maxWidth, block.DesiredSize.Width);
            }

            return maxWidth + this.WidthPaddings;
        }

        private bool TrySplitAndMoveToPending(int splitBlockIndex, Size availableSize)
        {
            IBlockElement blockElement = this.blocks[splitBlockIndex];
            bool isFirstBlock = splitBlockIndex == 0;
            bool isBlockSplit = false;

            if (blockElement.DesiredSize.Height > availableSize.Height && !isFirstBlock)
            {
                this.MoveBlocksToPending(splitBlockIndex);
            }
            else
            {
                if (blockElement.HasPendingContent)
                {
                    this.pendingBlocks.Add(blockElement.Split());
                    isBlockSplit = true;
                }

                this.MoveBlocksToPending(splitBlockIndex + 1);
            }

            return isBlockSplit;
        }

        private void MoveBlocksToPending(int firstBlockIndex)
        {
            for (int i = firstBlockIndex; i < this.blocks.Count; i++)
            {
                this.pendingBlocks.Add(this.blocks[i]);
            }

            this.blocks.RemoveRange(firstBlockIndex, this.blocks.Count - firstBlockIndex);
        }

        internal TableCell Split()
        {
            TableCell cell = new TableCell(this);
            cell.blocks.AddRange(this.pendingBlocks);

            return cell;
        }

        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current
        /// <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
            //return string.Format("Cell r:{0}, c:{1}, rSp:{2}, cSp:{3}, layoutRect:{4}, desiredSize:{5}, contentSize:{6}",
            //    this.FromRow, this.FromColumn, this.RowSpan, this.ColumnSpan, this.LayoutRectangle, this.DesiredSize, this.DesiredContentSize);

            StringBuilder builder = new StringBuilder();

            foreach (IBlockElement block in this.Blocks)
            {
                builder.AppendLine(String.Format("{0}, {1}, {2}", '{', block.ToString(), '}'));
            }

            return builder.ToString();
        }
    }
}
