﻿#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
#else
using System.Windows.Media;
using FontStyles = System.Windows.FontStyles;
using FontWeights = System.Windows.FontWeights;
using FontStyle = System.Windows.FontStyle;
using FontWeight = System.Windows.FontWeight;
#endif
using Telerik.Windows.Documents.Fixed.Model.Editing.Flow;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Text;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Editing
{
    /// <summary>
    /// Represents the text state properties.
    /// </summary>
    public class TextProperties : PropertiesBase<TextProperties>
    {
        private FontBase font;

        /// <summary>
        /// Initializes a new instance of the <see cref="TextProperties" /> class.
        /// </summary>
        public TextProperties()
        {
            this.RenderingMode = FixedDocumentDefaults.TextRenderingMode;
            this.FontSize = FixedDocumentDefaults.FontSize;
            this.Font = FixedDocumentDefaults.Font;
            this.UnderlinePattern = FixedDocumentDefaults.UnderlinePattern;
            this.UnderlineColor = FixedDocumentDefaults.Color;
            this.BaselineAlignment = FixedDocumentDefaults.BaselineAlignment;
            this.HighlightColor = FixedDocumentDefaults.HighlightColor;
        }

        /// <summary>
        /// Gets or sets the text decorations.
        /// </summary>
        /// <value>The text decorations.</value>
        public UnderlinePattern UnderlinePattern { get; set; }

        /// <summary>
        /// Gets or sets the color of the underline.
        /// </summary>
        /// <value>The color of the underline.</value>
        public ColorBase UnderlineColor { get; set; }

        /// <summary>
        /// Gets or sets the color of the highlight.
        /// </summary>
        /// <value>The color of the highlight.</value>
        public ColorBase HighlightColor { get; set; }

        /// <summary>
        /// Gets or sets the character spacing.
        /// </summary>
        /// <value>The character spacing.</value>
        public double? CharacterSpacing { get; set; }

        /// <summary>
        /// Gets or sets the word spacing.
        /// </summary>
        /// <value>The word spacing.</value>
        public double? WordSpacing { get; set; }

        /// <summary>
        /// Gets or sets the horizontal scaling.
        /// </summary>
        /// <value>The horizontal scaling.</value>
        public double? HorizontalScaling { get; set; }

        /// <summary>
        /// Gets or sets the size of the font.
        /// </summary>
        /// <value>The size of the font.</value>
        public double FontSize { get; set; }

        /// <summary>
        /// Gets or sets the rendering mode.
        /// </summary>
        /// <value>The rendering mode.</value>
        public RenderingMode RenderingMode { get; set; }

        /// <summary>
        /// Gets or sets the baseline alignment.
        /// </summary>
        /// <value>The baseline alignment.</value>
        public BaselineAlignment BaselineAlignment { get; set; }

        /// <summary>
        /// Gets or sets the current font.
        /// </summary>
        /// <value>The font.</value>
        public FontBase Font
        {
            get
            {
                return this.font;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "Font");

                this.font = value;
            }
        }

        /// <summary>
        /// Tries to set the current font.
        /// </summary>
        /// <param name="fontFamily">The font family.</param>
        /// <returns>True if succeeded.</returns>
        public bool TrySetFont(FontFamily fontFamily)
        {
            return this.TrySetFont(fontFamily, FontStyles.Normal, FontWeights.Normal);
        }

        /// <summary>
        /// Tries to set the current font.
        /// </summary>
        /// <param name="fontFamily">The font family.</param>
        /// <param name="fontStyle">The font style.</param>
        /// <param name="fontWeight">The font weight.</param>
        /// <returns>True if succeeded.</returns>
        public bool TrySetFont(FontFamily fontFamily, FontStyle fontStyle, FontWeight fontWeight)
        {
            FontBase font;
            bool result = FontsRepository.TryCreateFont(fontFamily, fontStyle, fontWeight, out font);
            if (result)
            {
                this.Font = font;
            }

            return result;
        }

        /// <summary>
        /// Copies the text properties from other instance.
        /// </summary>
        /// <param name="textProperties">The text properties.</param>
        public override void CopyFrom(TextProperties textProperties)
        {
            this.CharacterSpacing = textProperties.CharacterSpacing;
            this.WordSpacing = textProperties.WordSpacing;
            this.HorizontalScaling = textProperties.HorizontalScaling;
            this.FontSize = textProperties.FontSize;
            this.RenderingMode = textProperties.RenderingMode;
            this.Font = textProperties.Font;
            this.UnderlinePattern = textProperties.UnderlinePattern;
            this.UnderlineColor = textProperties.UnderlineColor;
            this.HighlightColor = textProperties.HighlightColor;
        }

        internal void CopyFrom(TextPropertiesOwner textProperties)
        {
            this.CharacterSpacing = textProperties.CharacterSpacing;
            this.WordSpacing = textProperties.WordSpacing;
            this.HorizontalScaling = textProperties.HorizontalScaling;
            this.FontSize = textProperties.FontSize;
            this.RenderingMode = textProperties.RenderingMode;
            this.Font = textProperties.Font;
        }

        internal void CopyTo(TextPropertiesOwner textProperties)
        {
            textProperties.CharacterSpacing = this.CharacterSpacing;
            textProperties.WordSpacing = this.WordSpacing;
            textProperties.HorizontalScaling = this.HorizontalScaling;
            textProperties.FontSize = this.FontSize;
            textProperties.RenderingMode = this.RenderingMode;
            textProperties.Font = this.Font;
        }
    }
}
