﻿using System;
using Telerik.Windows.Documents.Core.Imaging;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Filters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Primitives;

namespace Telerik.Windows.Documents.Fixed.Model.Extensions
{
    internal static class ImageExtensions
    {
        internal static ImageDataSource ToImageDataSource(this IImageDescriptor image)
        {
#if !SILVERLIGHT
            bool canBePassedToUIWithoutDecode = image.IsSimpleJpeg();

            if (canBePassedToUIWithoutDecode)
            {
                byte[] encodedData = image.GetEncodedData();

                return new JpegImageDataSource(image.Width, image.Height, encodedData);
            }
#endif
            return image.ToDecodedImageDataSource();
        }

        internal static DecodedImageDataSource ToDecodedImageDataSource(this IImageDescriptor image)
        {
            DecodedImageDataSource resultImage;

            if (image.SMask != null)
            {
                resultImage = CalculateAlphaMaskedPixels(image);
            }
            else
            {
                IMask mask = image.Mask;
                PixelContainer pixels = image.ColorSpace.GetPixels(image, true);
                SizeI imageSize = mask != null ? mask.GetMaskedImageSize(image) : new SizeI(image.Width, image.Height);

                resultImage = new DecodedImageDataSource(imageSize.Width, imageSize.Height, pixels);
            }

            return resultImage;
        }

        internal static DecodedImageDataSource ToDecodedImageDataSource(this IImageDescriptor image, Color stencilColor)
        {
            PixelContainer pixels = GetStencilMaskedPixels(image, stencilColor);
            DecodedImageDataSource source = new DecodedImageDataSource(image.Width, image.Height, pixels);

            return source;
        }

        internal static int GetScaledPixelIndex(this IImageDescriptor image, int scaledPixelX, int scaledPixelY, double scaleX, double scaleY)
        {
            int x = (int)(scaledPixelX * scaleX);
            int y = (int)(scaledPixelY * scaleY);
            int index = y * image.Width + x;

            return index;
        }

#if !SILVERLIGHT
        internal static bool IsSimpleJpeg(this IImageDescriptor image)
        {
            bool hasSingleJpegFilter = image.Filters != null && image.Filters.Length == 1 && image.Filters[0] == PdfFilterNames.DCTDecode;

            if (hasSingleJpegFilter && !ColorSpaceBase.HasMask(image))
            {
                bool hasInvertedColors = JpegHasInvertedColors(image);

                return !hasInvertedColors;
            }

            return false;
        }
#endif

        private static PixelContainer GetStencilMaskedPixels(IImageDescriptor image, Color stencilColor)
        {
            byte[] data = image.GetDecodedData();

            if (data == null)
            {
                return null;
            }

            int unmaskedColor = ImageMask.GetUnmaskedColor(image);
            int width = image.Width;
            int height = image.Height;

            ImageBitReader reader = new ImageBitReader(data, 1, width);
            Color transparent = Color.Transparent;
            PixelContainer pixels = PixelContainerFactory.GetStencilImagePixelContainer(width, height, stencilColor, transparent);

            for (int i = 0; i < height; i++)
            {
                for (int j = 0; j < width; j++)
                {
                    int bit = reader.Read();

                    if (bit == unmaskedColor)
                    {
                        pixels.Add(stencilColor);
                    }
                    else
                    {
                        pixels.Add(transparent);
                    }
                }
            }

            return pixels;
        }

        private static DecodedImageDataSource CalculateAlphaMaskedPixels(IImageDescriptor image)
        {
            IImageDescriptor sMask = image.SMask;

            int width = Math.Max(image.Width, sMask.Width);
            int height = Math.Max(image.Height, sMask.Height);

            double imageScaleX = image.Width / (double)width;
            double imageScaleY = image.Height / (double)height;
            double sMaskScaleX = sMask.Width / (double)width;
            double sMaskScaleY = sMask.Height / (double)height;
            DecodedImageDataSource result;

            // memory optimization for equal size of image and sMask.
            if (image.Width == sMask.Width && image.Height == sMask.Height)
            {
                result = GetImageSourceWhenSameImageSize(image, height, width, imageScaleX, imageScaleY);
            }
            else
            {
                result = GetImageSourceWhenDifferentImageSize(image, width, height, imageScaleX, imageScaleY, sMaskScaleX, sMaskScaleY);
            }

            return result;
        }

        private static DecodedImageDataSource GetImageSourceWhenSameImageSize(IImageDescriptor image, int height, int width, double imageScaleX, double imageScaleY)
        {
            PixelContainer imagePixels = image.ColorSpace.GetPixels(image, false);
            PixelContainer sMaskPixels = image.SMask.ColorSpace.GetPixels(image.SMask, false);

            for (int y = 0; y < height; y++)
            {
                for (int x = 0; x < width; x++)
                {
                    int imageIndex = image.GetScaledPixelIndex(x, y, imageScaleX, imageScaleY);

                    Color sMaskColor = sMaskPixels.GetColorFromIndex(imageIndex);
                    Color pixelColor = imagePixels.GetColorFromIndex(imageIndex);
                    pixelColor.A = sMaskColor.GetGrayComponent();

                    imagePixels.SetColorToIndex(pixelColor, imageIndex);
                }
            }

            return new DecodedImageDataSource(width, height, imagePixels);
        }

        private static DecodedImageDataSource GetImageSourceWhenDifferentImageSize(IImageDescriptor image, int width, int height, double imageScaleX, double imageScaleY, double sMaskScaleX, double sMaskScaleY)
        {
            PixelContainer imagePixels = image.ColorSpace.GetPixels(image, false);
            PixelContainer sMaskPixels = image.SMask.ColorSpace.GetPixels(image.SMask, false);
            PixelContainer resultPixels = PixelContainerFactory.GetMaskedImagePixelContainer(width, height);

            for (int y = 0; y < height; y++)
            {
                int rowStartIndex = y * width;

                for (int x = 0; x < width; x++)
                {
                    int imageIndex = image.GetScaledPixelIndex(x, y, imageScaleX, imageScaleY);
                    int maskIndex = image.SMask.GetScaledPixelIndex(x, y, sMaskScaleX, sMaskScaleY);

                    Color imagePixelColor = imagePixels.GetColorFromIndex(imageIndex);
                    Color sMaskColor = sMaskPixels.GetColorFromIndex(maskIndex);
                    Color pixelColor = new Color(sMaskColor.GetGrayComponent(), imagePixelColor.R, imagePixelColor.G, imagePixelColor.B);

                    int resultIndex = rowStartIndex + x;
                    resultPixels.SetColorToIndex(pixelColor, resultIndex);
                }
            }

            return new DecodedImageDataSource(width, height, resultPixels);
        }

#if !SILVERLIGHT

        private static bool JpegHasInvertedColors(IImageDescriptor image)
        {
            int numberOfComponents = image.ColorSpace.ComponentCount;
            bool isCmykColorSpace = numberOfComponents == 4;

            if (isCmykColorSpace)
            {
                return true;
            }

            if (image.Decode != null)
            {
                double[] defaultDecode = image.ColorSpace.GetDefaultDecodeArray(image.BitsPerComponent);
                int maxIndex = Math.Min(defaultDecode.Length, image.Decode.Length);

                for (int index = 0; index < maxIndex; index++)
                {
                    double value = image.Decode[index];
                    double expected = defaultDecode[index];

                    if (value != expected)
                    {
                        return true;
                    }
                }
            }

            return false;
        }
#endif
    }
}
