﻿using System;
using System.IO;
using System.Windows.Media.Imaging;
using Telerik.Windows.Documents.Fixed.Imaging.Silverlight;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Resources
{
    /// <summary>
    /// Extensions for creating image source from BitmapSource.
    /// </summary>
    public static class ImageSourceExtensions
    {
        /// <summary>
        /// Creates ImageSource form the BitmapSource
        /// </summary>
        /// <param name="bitmapSource">The bitmap source.</param>
        /// <returns>The ImageSource</returns>
        public static ImageSource ToImageSource(this BitmapSource bitmapSource)
        {
            Guard.ThrowExceptionIfNull(bitmapSource, "bitmapSource");

            return bitmapSource.ToImageSource(FixedDocumentDefaults.ImageQuality);
        }

        /// <summary>
        /// Toes the image source.
        /// </summary>
        /// <param name="bitmapSource">The bitmap source.</param>
        /// <param name="imageQuality">The image quality.</param>
        /// <returns></returns>
        public static ImageSource ToImageSource(this BitmapSource bitmapSource, FormatProviders.Pdf.Export.ImageQuality imageQuality)
        {
            Guard.ThrowExceptionIfNull(bitmapSource, "bitmapSource");

            return new ImageSource(bitmapSource, imageQuality);
        }

        internal static WriteableBitmap ToBitmapSource(this ImageDataSource dataSource)
        {
            WriteableBitmap result;

            if (dataSource.IsEmpty)
            {
                result = null;
            }
            else if (dataSource.ImageDataType == ImageDataType.Decoded)
            {
                DecodedImageDataSource decodedData = (DecodedImageDataSource)dataSource;
                result = GetBitmapSourceFromPixels(decodedData.Width, decodedData.Height, decodedData.Data.Array);
            }
            else
            {
                throw new NotSupportedException();
            }

            return result;
        }

        internal static BitmapSource CreateBitmapSource(Stream stream)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");

            stream.Seek(0, SeekOrigin.Begin);
            BitmapImage bitmapImage = new BitmapImage();
            bitmapImage.SetSource(stream);

            return bitmapImage;
        }

        internal static Stream CreateImageStream(BitmapSource bitmapSource)
        {
            PngBitmapEncoder encoder = new PngBitmapEncoder();
            encoder.Frames.Add(BitmapFrame.Create(bitmapSource));
            MemoryStream memory = new MemoryStream();
            encoder.Save(memory);

            return memory;
        }

        internal static int[] GetPixels(BitmapSource source)
        {
            Guard.ThrowExceptionIfNull(source, "source");

            WriteableBitmap writeableBitmap = source as WriteableBitmap;
            if (writeableBitmap == null)
            {
                writeableBitmap = new WriteableBitmap(source);
            }

            return writeableBitmap.Pixels;
        }

        private static WriteableBitmap GetBitmapSourceFromPixels(int width, int height, int[] pixels)
        {
            if (pixels.Length != width * height)
            {
                throw new ArgumentException("Pixel dimensions don't match");
            }

            WriteableBitmap writeableBitmap = new WriteableBitmap(width, height);
            Array.Copy(pixels, writeableBitmap.Pixels, pixels.Length);

            return writeableBitmap;
        }
    }
}
