﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Windows;
using System.Windows.Media.Imaging;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Filters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Primitives;
using Telerik.Windows.Documents.Utilities;
using Color = System.Windows.Media.Color;
using PixelFormats = System.Windows.Media.PixelFormats;

namespace Telerik.Windows.Documents.Fixed.Model.Extensions
{
    /// <summary>
    /// Extensions for creating image source from BitmapSource.
    /// </summary>
    public static class ImageSourceExtensions
    {
        /// <summary>
        /// Creates ImageSource form the BitmapSource
        /// </summary>
        /// <param name="bitmapSource">The bitmap source.</param>
        /// <returns>The ImageSource</returns>
        public static ImageSource ToImageSource(this BitmapSource bitmapSource)
        {
            Guard.ThrowExceptionIfNull(bitmapSource, "bitmapSource");

            return bitmapSource.ToImageSource(FixedDocumentDefaults.ImageQuality);
        }

        /// <summary>
        /// Creates ImageSource form the BitmapSource
        /// </summary>
        /// <param name="bitmapSource">The bitmap source.</param>
        /// <param name="imageQuality">The image quality.</param>
        /// <returns>The ImageSource</returns>
        public static ImageSource ToImageSource(this BitmapSource bitmapSource, FormatProviders.Pdf.Export.ImageQuality imageQuality)
        {
            Guard.ThrowExceptionIfNull(bitmapSource, "bitmapSource");

            return new ImageSource(bitmapSource, imageQuality);
        }

        internal static BitmapSource ToBitmapSource(this ImageDataSource dataSource, SizeI? modifiedSize = null)
        {
            BitmapSource result;

            if (dataSource.IsEmpty)
            {
                result = null;
            }
            else if (dataSource.ImageDataType == ImageDataType.Jpeg)
            {
                JpegImageDataSource jpegData = (JpegImageDataSource)dataSource;
                result = CreateImageSourceFromEncodedJpeg(jpegData, modifiedSize);
            }
            else if (dataSource.ImageDataType == ImageDataType.Decoded)
            {
                DecodedImageDataSource decodedData = (DecodedImageDataSource)dataSource;
                result = CreateImageSourceFromDecodedData(decodedData, modifiedSize);
            }
            else
            {
                throw new NotSupportedException();
            }

            return result;
        }

        internal static BitmapSource CreateBitmapSource(Stream stream)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");

            stream.Seek(0, SeekOrigin.Begin);

            BitmapImage bitmapImage = new BitmapImage();
            bitmapImage.BeginInit();
            bitmapImage.StreamSource = stream;
            bitmapImage.CreateOptions = BitmapCreateOptions.None;
            bitmapImage.EndInit();

            return bitmapImage;
        }

        internal static Stream CreateImageStream(BitmapSource bitmapSource)
        {
            PngBitmapEncoder encoder = new PngBitmapEncoder();
            encoder.Frames.Add(BitmapFrame.Create(bitmapSource));
            MemoryStream stream = new MemoryStream();
            encoder.Save(stream);

            return stream;        
        }

        internal static IEnumerable<int> GetCmykPixels(BitmapSource source)
        {
            int[] pixels = new int[source.PixelWidth * source.PixelHeight];
            BitmapSource cmykSource = source.Format == PixelFormats.Cmyk32 ? source : new FormatConvertedBitmap(source, PixelFormats.Cmyk32, null, 0);

            checked
            {
                return EnumerateBitmapPixels(cmykSource);
            }
        }

        internal static IEnumerable<int> GetPixels(BitmapSource source)
        {
            if (source.Format == PixelFormats.Bgr32 || source.Format == PixelFormats.Bgra32 || source.Format == PixelFormats.Pbgra32)
            {
                checked
                {
                    return EnumerateBitmapPixels(source);
                }
            }
            else if (source.Format == PixelFormats.Indexed8)
            {
                return EnumeratePixelsFromIndexedBitmap(source);
            }
            else
            {
                FormatConvertedBitmap converted = new FormatConvertedBitmap(source, PixelFormats.Bgra32, null, 0);
                return EnumerateBitmapPixels(converted);
            }
        }

        internal static BitmapSource GetBitmapSourceFromPixels(int width, int height, int[] pixels)
        {
            var dpiX = 96d;
            var dpiY = 96d;
            var pixelFormat = PixelFormats.Bgra32;
            var bytesPerPixel = (pixelFormat.BitsPerPixel + 7) / 8;
            var stride = bytesPerPixel * width;

            return BitmapSource.Create(width, height, dpiX, dpiY, pixelFormat, null, pixels, stride);
        }

        private static BitmapSource CreateImageSourceFromEncodedJpeg(JpegImageDataSource source, SizeI? modifiedSize = null)
        {
            BitmapImage bitmapImage = new BitmapImage();

            bitmapImage.BeginInit();

            bitmapImage.StreamSource = new MemoryStream(source.Data);

            if (modifiedSize.HasValue)
            {
                bitmapImage.DecodePixelWidth = modifiedSize.Value.Width;
                bitmapImage.DecodePixelHeight = modifiedSize.Value.Height;
            }

            bitmapImage.EndInit();
            bitmapImage.Freeze();

            return bitmapImage;
        }

        private static BitmapSource CreateImageSourceFromDecodedData(DecodedImageDataSource source, SizeI? modifiedSize = null)
        {
            BitmapSource initialSource;
            if (source.Data.DataType == ImageDecodedDataType.Monochrome)
            {
                MonochromePixelContainer monochromePixelContainer = (MonochromePixelContainer)source.Data;
                List<Color> colorList = monochromePixelContainer.Palette.Select(c => Color.FromArgb(c.A, c.R, c.G, c.B)).ToList();
                BitmapPalette palette = new BitmapPalette(colorList);

                initialSource = BitmapSource.Create(source.Width, source.Height, 96, 96, PixelFormats.Indexed1, palette, source.Data.Array, (source.Width + 7) / 8);
            }
            else if (source.Data.DataType == ImageDecodedDataType.Indexed)
            {
                IndexedPixelContainer indexedPixelContainer = (IndexedPixelContainer)source.Data;
                List<Color> colorList = indexedPixelContainer.Palette.Select(c => Color.FromArgb(c.A, c.R, c.G, c.B)).ToList();
                BitmapPalette palette = new BitmapPalette(colorList);

                initialSource = BitmapSource.Create(source.Width, source.Height, 96, 96, PixelFormats.Indexed8, palette, source.Data.Array, source.Width);
            }
            else if (source.Data.DataType == ImageDecodedDataType.Gray)
            {
                initialSource = BitmapSource.Create(source.Width, source.Height, 96, 96, PixelFormats.Gray8, null, source.Data.Array, source.Width);
            }
            else
            {
                initialSource = BitmapSource.Create(source.Width, source.Height, 96, 96, PixelFormats.Bgra32, null, source.Data.Array, source.Width * 4);
            }

            BitmapSource result = initialSource;

            if (modifiedSize.HasValue)
            {
                var drawingVisual = new System.Windows.Media.DrawingVisual();
                System.Windows.Media.DrawingContext drawingContext = drawingVisual.RenderOpen();
                drawingContext.DrawImage(result, new Rect(0, 0, modifiedSize.Value.Width, modifiedSize.Value.Height));
                drawingContext.Close();

                RenderTargetBitmap resizer = new RenderTargetBitmap(modifiedSize.Value.Width, modifiedSize.Value.Height, 96, 96, PixelFormats.Pbgra32);
                resizer.Render(drawingVisual);

                result = resizer;
            }

            result.Freeze();

            return result;
        }

        private static IEnumerable<int> EnumerateBitmapPixels(BitmapSource source)
        {
            // Avoiding the creation of a single large int array as optimization for large images.
            int[] row = new int[source.PixelWidth];
            for (int h = 0; h < source.PixelHeight; h++)
            {
                source.CopyPixels(new System.Windows.Int32Rect(0, h, source.PixelWidth, 1), row, source.PixelWidth * 4, 0);
                for (int w = 0; w < source.PixelWidth; w++)
                {
                    yield return row[w];
                }
            }
        }

        private static IEnumerable<int> EnumeratePixelsFromIndexedBitmap(BitmapSource source)
        {
            byte[] indices = new byte[source.PixelWidth * source.PixelHeight];
            source.CopyPixels(indices, source.PixelWidth, 0);
            for (int i = 0; i < indices.Length; ++i)
            {
                Color c = source.Palette.Colors[indices[i]];
                int pixel = (c.A << 24) | (c.R << 16) | (c.G << 8) | (c.B << 0);
                yield return pixel;
            }
        }
    }
}
