﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Core.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Fonts.Encoding;

namespace Telerik.Windows.Documents.Fixed.Model.Fonts
{
    internal abstract class CidFontBase : FontBase
    {
        private readonly PdfProperty<string> lang;
        private readonly PdfProperty<byte[]> cidSet;
        private readonly PdfProperty<CidWidths<int>> widths;
        private readonly PdfProperty<int> defaultWidth;
        private readonly PdfProperty<CidWidths<CidWidthVertical>> widthsVertical;
        private readonly PdfProperty<CidWidthVertical> defaultWidthsVertical;
        private readonly PdfProperty<CidToGidMap> cidToGidMapping;
        private readonly PdfProperty<CidSystemInfo> cidSystemInfo;
        private readonly PdfProperty<CMapEncoding> encoding;
        private Dictionary<int, double> charCodeToWidthCache;

        public CidFontBase(string name, FontSource fontSource)
            : base(name, fontSource)
        {
            this.lang = new PdfProperty<string>();
            this.cidSet = new PdfProperty<byte[]>();
            this.widths = new PdfProperty<CidWidths<int>>();
            this.defaultWidth = new PdfProperty<int>();
            this.widthsVertical = new PdfProperty<CidWidths<CidWidthVertical>>();
            this.defaultWidthsVertical = new PdfProperty<CidWidthVertical>();
            this.cidToGidMapping = new PdfProperty<CidToGidMap>();
            this.cidSystemInfo = new PdfProperty<CidSystemInfo>(() => Fonts.Encoding.CidSystemInfo.Default);
            this.encoding = new PdfProperty<CMapEncoding>();
        }

        internal override bool HasFontDescriptorProperties
        {
            get
            {
                return base.HasFontDescriptorProperties ||
                    this.Lang.HasValue ||
                    this.CidSet.HasValue;
            }
        }

        internal PdfProperty<string> Lang
        {
            get
            {
                return this.lang;
            }
        }

        internal PdfProperty<byte[]> CidSet
        {
            get
            {
                return this.cidSet;
            }
        }

        internal PdfProperty<CidWidths<int>> Widths
        {
            get
            {
                return this.widths;
            }
        }

        internal PdfProperty<int> DefaultWidth
        {
            get
            {
                return this.defaultWidth;
            }
        }

        internal PdfProperty<CidWidths<CidWidthVertical>> WidthsVertical
        {
            get
            {
                return this.widthsVertical;
            }
        }

        internal PdfProperty<CidWidthVertical> DefaultWidthsVertical
        {
            get
            {
                return this.defaultWidthsVertical;
            }
        }

        internal PdfProperty<CidToGidMap> CharIdToGlyphIdMapping
        {
            get
            {
                return this.cidToGidMapping;
            }
        }

        internal PdfProperty<CidSystemInfo> CidSystemInfo
        {
            get
            {
                return this.cidSystemInfo;
            }
        }

        internal PdfProperty<CMapEncoding> Encoding
        {
            get
            {
                return this.encoding;
            }
        }

        internal override FontSource ActualFontSource
        {
            get
            {
                return this.FontSource;
            }
        }

        internal override double GetWidth(int charCode)
        {
            double width;
            if (!this.TryGetWidthFromCache(charCode, out width))
            {
                if (this.DefaultWidth.HasValue)
                {
                    width = this.DefaultWidth.Value;
                }
                else
                {
                    width = base.GetWidth(charCode);
                }
            }

            return width;
        }

        private bool TryGetWidthFromCache(int charCode, out double width)
        {
            width = 0;
            bool hasFoundWidth = false;

            if (this.Widths.HasValue)
            {
                this.EnsureWidthsCacheIsLoaded();
                hasFoundWidth = this.charCodeToWidthCache.TryGetValue(charCode, out width);
            }

            return hasFoundWidth;
        }

        private void EnsureWidthsCacheIsLoaded()
        {
            if (this.charCodeToWidthCache == null)
            {
                this.charCodeToWidthCache = new Dictionary<int, double>();

                foreach (CidWidthsRange<int> range in this.Widths.Value)
                {
                    int charCode = range.StartCharCode;

                    foreach (int width in range.Widths)
                    {
                        this.charCodeToWidthCache[charCode] = width;
                        charCode++;
                    }
                }
            }
        }
    }
}
