﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Fonts.Encoding
{
    internal class CMapEncoding
    {
        private readonly Dictionary<CharCode, string> unicodeMapping;
        private readonly List<CharCodeRange> unicodeRanges;
        private readonly PdfProperty<CidSystemInfo> cidSystemInfo;
        private readonly PdfProperty<string> cMapName;
        private readonly PdfProperty<int> cMapType;

        public CMapEncoding()
        {
            this.unicodeMapping = new Dictionary<CharCode, string>();
            this.unicodeRanges = new List<CharCodeRange>();
            this.cidSystemInfo = new PdfProperty<CidSystemInfo>(() => Encoding.CidSystemInfo.Default);
            this.cMapName = new PdfProperty<string>(() => "Adobe-Identity-UCS");
            this.cMapType = new PdfProperty<int>(() => 2);
        }

        public PdfProperty<CidSystemInfo> CidSystemInfo
        {
            get
            {
                return this.cidSystemInfo;
            }
        }

        public PdfProperty<string> CMapName
        {
            get
            {
                return this.cMapName;
            }
        }

        public PdfProperty<int> CMapType
        {
            get
            {
                return this.cMapType;
            }
        }

        public IEnumerable<Tuple<CharCode, string>> CharCodeToUnicodeMappings
        {
            get
            {
                return unicodeMapping.Select(kv => new Tuple<CharCode, string>(kv.Key, kv.Value));
            }
        }

        internal List<CharCodeRange> UnicodeRanges
        {
            get
            {
                return this.unicodeRanges;
            }
        }

        public void AddUnicodeMapping(CharCode code, string unicode)
        {
            Guard.ThrowExceptionIfNull(code, "code");
            Guard.ThrowExceptionIfNull(unicode, "unicode");

            //// TODO: Handle different glyphs with equal CharCodes.

            if (!this.unicodeMapping.ContainsKey(code))
            {
                this.unicodeMapping.Add(code, unicode);
            }
        }

        public bool TryGetToUnicode(CharCode code, out string unicode)
        {
            Guard.ThrowExceptionIfNull(code, "code");

            return this.unicodeMapping.TryGetValue(code, out unicode) || this.TryGetValueFromRange(code, out unicode);
        }

        internal void AddUnicodeRange(CharCode start, CharCode end, List<string> unicodeStrings)
        {
            CharCodeRange charCodeRange = new CharCodeRange(start, end, unicodeStrings, isSequential: true);

            this.unicodeRanges.Add(charCodeRange);
        }

        internal void AddUnicodeRangeArray(CharCode start, CharCode end, List<string> unicodeStrings)
        {
            CharCodeRange charCodeRange = new CharCodeRange(start, end, unicodeStrings, isSequential: false);

            this.unicodeRanges.Add(charCodeRange);
        }

        private bool TryGetValueFromRange(CharCode charCode, out string unicode)
        {
            unicode = null;
            CharCodeRange rangeToFind = new CharCodeRange(charCode, charCode);

            CharCodeRange[] unicodeRangesArray = this.UnicodeRanges.ToArray();
            int index = Array.BinarySearch(unicodeRangesArray, rangeToFind);

            if (index < 0)
            {
                return false;
            }

            CharCodeRange matchedRange = unicodeRangesArray[index];
            List<string> unicodes = matchedRange.Unicodes;
            unicode = unicodes[charCode.Code - matchedRange.Start.Code];

            return unicode != null;
        }
    }
}
