﻿using System;
using System.Collections.Generic;
using System.Linq;
#if !NETSTANDARD
using System.Windows;
#else
using Telerik.Documents.Primitives;
#endif
using Telerik.Windows.Documents.Core.Fonts;
using Telerik.Windows.Documents.Core.Fonts.Glyphs;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Model.Fonts.Encoding;
using Telerik.Windows.Documents.Fixed.Model.Text;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Fonts
{
    /// <summary>
    /// Represents the base class for the font object.
    /// </summary>
    public abstract class FontBase : IFontDescriptor
    {
        internal const double FontRatio = 1000;
        internal const double GlyphOutlinesCacheFontSize = 100;

        private readonly string name;
        private readonly FontSource fontSource;
        private readonly PdfProperty<string> fontFamily;
        private readonly PdfProperty<string> fontStretch;
        private readonly PdfProperty<string> charSet;
        private readonly PdfProperty<double> fontWeight;
        private readonly PdfProperty<Rect> fontBBox;
        private readonly PdfProperty<double> italicAngle;
        private readonly PdfProperty<double> ascent;
        private readonly PdfProperty<double> descent;
        private readonly PdfProperty<double> leading;
        private readonly PdfProperty<double> capHeight;
        private readonly PdfProperty<double> xHeight;
        private readonly PdfProperty<double> stemV;
        private readonly PdfProperty<double> stemH;
        private readonly PdfProperty<double> avgWidth;
        private readonly PdfProperty<double> maxWidth;
        private readonly PdfProperty<double> missingWidth;
        private readonly PdfProperty<CMapEncoding> toUnicode;
        private readonly PdfProperty<FontFileInfo> fontFileInfo;
        private readonly Dictionary<CharCode, Glyph> glyphsCache;

        internal FontBase(string name, FontSource fontSource)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");
            Guard.ThrowExceptionIfNull(fontSource, "fontSource");

            this.name = name;
            this.fontSource = fontSource;
            this.glyphsCache = new Dictionary<CharCode, Glyph>();

            this.fontFamily = new PdfProperty<string>(this.GetFontSourceFontFamily);
            this.fontStretch = new PdfProperty<string>();
            this.charSet = new PdfProperty<string>();
            this.fontWeight = new PdfProperty<double>();
            this.fontBBox = new PdfProperty<Rect>(this.GetFontSourceBoundingBox);
            this.italicAngle = new PdfProperty<double>(this.GetFontSourceItalicAngle);
            this.ascent = new PdfProperty<double>(this.GetFontSourceAscent);
            this.descent = new PdfProperty<double>(this.GetFontSourceDescent);
            this.leading = new PdfProperty<double>();
            this.capHeight = new PdfProperty<double>(this.GetFontSourceCapHeight);
            this.xHeight = new PdfProperty<double>();
            this.stemV = new PdfProperty<double>(this.GetFontSourceStemV);
            this.stemH = new PdfProperty<double>();
            this.avgWidth = new PdfProperty<double>();
            this.maxWidth = new PdfProperty<double>();
            this.missingWidth = new PdfProperty<double>();
            this.toUnicode = new PdfProperty<CMapEncoding>();
            this.fontFileInfo = new PdfProperty<FontFileInfo>();

            this.InitializeFlagsFromFontSource();
        }

        /// <summary>
        /// Gets the name of the font.
        /// </summary>
        /// <value>The name.</value>
        public string Name
        {
            get
            {
                return this.name;
            }
        }

        internal virtual string FontName
        {
            get
            {
                return this.Name;
            }
        }

        internal abstract FontType Type { get; }

        internal double UnderlinePosition
        {
            get
            {
                return this.FontSource.UnderlinePosition / FontRatio;
            }
        }

        internal double UnderlineThickness
        {
            get
            {
                return this.FontSource.UnderlineThickness / FontRatio;
            }
        }

        internal double Height
        {
            get
            {
                return Math.Abs(this.Ascent.Value) + Math.Abs(this.Descent.Value);
            }
        }

        internal virtual double LineGap
        {
            get
            {
                return this.FontSource.LineGap / FontRatio;
            }
        }

        internal virtual bool IsBold
        {
            get
            {
                return this.FontSource.IsBold;
            }
        }

        internal virtual bool HasFontDescriptorProperties
        {
            get
            {
                return 
                this.Ascent.HasValue ||
                this.Descent.HasValue ||
                this.FontFamily.HasValue ||
                this.FontStretch.HasValue ||
                this.FontWeight.HasValue ||
                this.BoundingBox.HasValue ||
                this.ItalicAngle.HasValue ||
                this.Leading.HasValue ||
                this.CapHeight.HasValue ||
                this.XHeight.HasValue ||
                this.StemH.HasValue ||
                this.StemV.HasValue ||
                this.AvgWidth.HasValue ||
                this.MaxWidth.HasValue ||
                this.MissingWidth.HasValue ||
                this.FontFileInfo.HasValue ||
                this.CharSet.HasValue;
            }
        }

        internal PdfProperty<double> Ascent
        {
            get
            {
                return this.ascent;
            }
        }

        internal PdfProperty<double> Descent
        {
            get
            {
                return this.descent;
            }
        }

        internal PdfProperty<double> ItalicAngle
        {
            get
            {
                return this.italicAngle;
            }
        }

        internal PdfProperty<string> FontStretch
        {
            get
            {
                return this.fontStretch;
            }
        }

        internal PdfProperty<string> CharSet
        {
            get
            {
                return this.charSet;
            }
        }

        internal PdfProperty<double> FontWeight
        {
            get
            {
                return this.fontWeight;
            }
        }

        internal PdfProperty<double> CapHeight
        {
            get
            {
                return this.capHeight;
            }
        }

        internal PdfProperty<double> StemV
        {
            get
            {
                return this.stemV;
            }
        }

        internal PdfProperty<Rect> BoundingBox
        {
            get
            {
                return this.fontBBox;
            }
        }

        internal PdfProperty<CMapEncoding> ToUnicode
        {
            get
            {
                return this.toUnicode;
            }
        }

        internal PdfProperty<string> FontFamily
        {
            get
            {
                return this.fontFamily;
            }
        }

        internal PdfProperty<double> Leading
        {
            get
            {
                return this.leading;
            }
        }

        internal PdfProperty<double> XHeight
        {
            get
            {
                return this.xHeight;
            }
        }
        
        internal PdfProperty<double> StemH
        {
            get
            {
                return this.stemH;
            }
        }

        internal PdfProperty<double> AvgWidth
        {
            get
            {
                return this.avgWidth;
            }
        }

        internal PdfProperty<double> MaxWidth
        {
            get
            {
                return this.maxWidth;
            }
        }

        internal PdfProperty<double> MissingWidth
        {
            get
            {
                return this.missingWidth;
            }
        }

        internal PdfProperty<FontFileInfo> FontFileInfo
        {
            get
            {
                return this.fontFileInfo;
            }
        }

        internal bool IsFixedPitch { get; set; }

        internal bool IsSerif { get; set; }

        internal bool IsSymbolic { get; set; }

        internal bool IsScript { get; set; }

        internal bool IsNonSymbolic { get; set; }

        internal bool IsItalic { get; set; }

        internal bool IsAllCap { get; set; }

        internal bool IsSmallCap { get; set; }

        internal bool IsForcingBold { get; set; }

        internal FontSource FontSource
        {
            get
            {
                return this.fontSource;
            }
        }

        /// <summary>
        /// The <see cref="FontSource"/> or a fallback <see cref="Core.Fonts.FontSource"/> used when rendering the glyph outlines.
        /// </summary>
        internal abstract FontSource ActualFontSource { get; }

        bool IFontDescriptor.IsSymbolic
        {
            get
            {
                return this.IsSymbolic;
            }
        }

        bool IFontDescriptor.IsNonSymbolic
        {
            get
            {
                return this.IsNonSymbolic;
            }
        }

        internal bool TryGetGlyphId(int unicode, out ushort glyphId)
        {
            return this.FontSource.TryGetGlyphId(unicode, out glyphId);
        }

        internal virtual IEnumerable<CharInfo> CalculateCharacters(string text)
        {
            Guard.ThrowExceptionIfNull(text, "text");

            foreach (char ch in text)
            {
                int charId;
                if (this.FontSource.TryGetCharCode(ch, out charId))
                {
                    yield return new CharInfo(ch.ToString(), new CharCode((byte)charId));
                }
            }
        }

        internal abstract void InitializeCoreGlyph(Glyph glyph, CharCode charcode);
        
        internal GlyphOutlinesCollection GetCachedOutlines(CharCode charCode)
        {
            Glyph glyph = this.GetCachedGlyph(charCode);

            if (glyph.Outlines == null)
            {
                this.ActualFontSource.InitializeGlyphOutlines(glyph, GlyphOutlinesCacheFontSize);
            }

            return glyph.Outlines;
        }

        internal virtual double GetWidth(int charCode)
        {
            return this.FontSource.GetAdvancedWidth(charCode);
        }

        internal virtual byte[] ComputeSubset(IEnumerable<CharInfo> usedCharacters)
        {
            return this.FontSource.Data;
        }

        private double GetFontSourceStemV()
        {
            return this.FontSource.StemV;
        }

        private double GetFontSourceCapHeight()
        {
            return this.FontSource.CapHeight;
        }

        private double GetFontSourceDescent()
        {
            return this.FontSource.Descent / FontRatio;
        }

        private double GetFontSourceAscent()
        {
            return this.FontSource.Ascent / FontRatio;
        }

        private double GetFontSourceItalicAngle()
        {
            return this.FontSource.ItalicAngle;
        }

        private string GetFontSourceFontFamily()
        {
            return this.FontSource.FontFamily;
        }

        private Rect GetFontSourceBoundingBox()
        {
            double x = this.FontSource.BoundingBox.X;
            double y = this.Descent.Value * FontRatio;
            double width = this.FontSource.BoundingBox.Width;
            double height = (this.Ascent.Value + Math.Abs(this.Descent.Value)) * FontRatio;
            Rect boundingBox = new Rect(x, y, width, height);

            return boundingBox;
        }

        private Glyph GetCachedGlyph(CharCode charCode)
        {
            Glyph glyph;
            if (!this.glyphsCache.TryGetValue(charCode, out glyph))
            {
                glyph = new Glyph();
                this.InitializeCoreGlyph(glyph, charCode);
                this.glyphsCache.Add(charCode, glyph);
            }

            return glyph;
        }

        private void InitializeFlagsFromFontSource()
        {
            foreach (FontFlag flag in this.FontSource.Flags)
            {
                switch (flag)
                {
                    case FontFlag.FixedPitch:
                        this.IsFixedPitch = true;
                        break;
                    case FontFlag.Serif:
                        this.IsSerif = true;
                        break;
                    case FontFlag.Symbolic:
                        this.IsSymbolic = true;
                        break;
                    case FontFlag.Script:
                        this.IsScript = true;
                        break;
                    case FontFlag.Nonsymbolic:
                        this.IsNonSymbolic = true;
                        break;
                    case FontFlag.Italic:
                        this.IsItalic = true;
                        break;
                    case FontFlag.AllCap:
                        this.IsAllCap = true;
                        break;
                    case FontFlag.SmallCap:
                        this.IsSmallCap = true;
                        break;
                    case FontFlag.ForceBold:
                        this.IsForcingBold = true;
                        break;
                    default:
                        throw new NotSupportedException(string.Format("Not supported font flag: {0}", flag));
                }
            }
        }
    }
}
