﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Core.Fonts.OpenType;
using Telerik.Windows.Documents.Utilities;
using FontProperties = Telerik.Windows.Documents.Core.Fonts.FontProperties;

namespace Telerik.Windows.Documents.Fixed.Model.Fonts
{
    /// <summary>
    /// Represents the fonts repository. It can be used to create new font object.
    /// </summary>
    public static partial class FontsRepository
    {
        private const double HelveticaDescent = -211;
        private const double HelveticaAscent = 905;
        private const double HelveticaLineGap = 32;

        private const double TimesDescent = -216;
        private const double TimesAscent = 891;
        private const double TimesLineGap = 42;

        private const double CourierDescent = -300;
        private const double CourierAscent = 832;
        private const double CourierLineGap = 0;

        private const double SymbolDescent = -219;
        private const double SymbolAscent = 1005;
        private const double SymbolLineGap = 0;

        private const double ZapfDingbatsDescent = -220;
        private const double ZapfDingbatsAscent = 1000;
        private const double ZapfDingbatsLineGap = 0;

        private static readonly Dictionary<FontProperties, FontBase> fontPropertiesToFont;
        private static readonly Dictionary<string, Func<StandardFont>> standardFontCreators;

        private static readonly StandardFont helvetica;
        private static readonly StandardFont helveticaBold;
        private static readonly StandardFont helveticaOblique;
        private static readonly StandardFont helveticaBoldOblique;
        private static readonly StandardFont courier;
        private static readonly StandardFont courierBold;
        private static readonly StandardFont courierOblique;
        private static readonly StandardFont courierBoldOblique;
        private static readonly StandardFont timesRoman;
        private static readonly StandardFont timesBold;
        private static readonly StandardFont timesItalic;
        private static readonly StandardFont timesBoldItalic;
        private static readonly StandardFont symbol;
        private static readonly StandardFont zapfDingbats;

        static FontsRepository()
        {
            fontPropertiesToFont = new Dictionary<FontProperties, FontBase>();
            standardFontCreators = new Dictionary<string, Func<StandardFont>>();
            FontsManager = new Core.Fonts.FontsManager();
            RegisterStandartFontCreators();

            FontsRepository.helvetica = CreateStandardFontByName(Core.Fonts.FontsManager.Helvetica);
            FontsRepository.helveticaBold = CreateStandardFontByName(Core.Fonts.FontsManager.HelveticaBold);
            FontsRepository.helveticaOblique = CreateStandardFontByName(Core.Fonts.FontsManager.HelveticaOblique);
            FontsRepository.helveticaBoldOblique = CreateStandardFontByName(Core.Fonts.FontsManager.HelveticaBoldOblique);
            FontsRepository.courier = CreateStandardFontByName(Core.Fonts.FontsManager.Courier);
            FontsRepository.courierBold = CreateStandardFontByName(Core.Fonts.FontsManager.CourierBold);
            FontsRepository.courierOblique = CreateStandardFontByName(Core.Fonts.FontsManager.CourierOblique);
            FontsRepository.courierBoldOblique = CreateStandardFontByName(Core.Fonts.FontsManager.CourierBoldOblique);
            FontsRepository.timesRoman = CreateStandardFontByName(Core.Fonts.FontsManager.TimesRoman);
            FontsRepository.timesBold = CreateStandardFontByName(Core.Fonts.FontsManager.TimesBold);
            FontsRepository.timesItalic = CreateStandardFontByName(Core.Fonts.FontsManager.TimesItalic);
            FontsRepository.timesBoldItalic = CreateStandardFontByName(Core.Fonts.FontsManager.TimesBoldItalic);
            FontsRepository.symbol = CreateStandardFontByName(Core.Fonts.FontsManager.Symbol);
            FontsRepository.zapfDingbats = CreateStandardFontByName(Core.Fonts.FontsManager.ZapfDingbats);

#if NETSTANDARD
            FontsRepository.RegisterStandartFonts();
#endif
        }

        /// <summary>
        /// Gets the symbol font object.
        /// </summary>
        public static FontBase Symbol
        {
            get
            {
                return symbol;
            }
        }

        /// <summary>
        /// Gets the zapfdingbats font object.
        /// </summary>
        public static FontBase ZapfDingbats
        {
            get
            {
                return zapfDingbats;
            }
        }

        /// <summary>
        /// Gets the times toman font object.
        /// </summary>
        public static FontBase TimesRoman
        {
            get
            {
                return timesRoman;
            }
        }

        /// <summary>
        /// Gets the times bold font object.
        /// </summary>
        public static FontBase TimesBold
        {
            get
            {
                return timesBold;
            }
        }

        /// <summary>
        /// Gets the times bold italic font object.
        /// </summary>
        public static FontBase TimesBoldItalic
        {
            get
            {
                return timesBoldItalic;
            }
        }

        /// <summary>
        /// Gets the times italic font object.
        /// </summary>
        public static FontBase TimesItalic
        {
            get
            {
                return timesItalic;
            }
        }

        /// <summary>
        /// Gets the helvetica font object.
        /// </summary>
        /// <value>The helvetica.</value>
        public static FontBase Helvetica
        {
            get
            {
                return helvetica;
            }
        }

        /// <summary>
        /// Gets the helvetica bold font object.
        /// </summary>
        /// <value>The helvetica.</value>
        public static FontBase HelveticaBold
        {
            get
            {
                return helveticaBold;
            }
        }

        /// <summary>
        /// Gets the helvetica oblique font object.
        /// </summary>
        /// <value>The helvetica.</value>
        public static FontBase HelveticaOblique
        {
            get
            {
                return helveticaOblique;
            }
        }

        /// <summary>
        /// Gets the helvetica bold oblique font object.
        /// </summary>
        /// <value>The helvetica.</value>
        public static FontBase HelveticaBoldOblique
        {
            get
            {
                return helveticaBoldOblique;
            }
        }

        /// <summary>
        /// Gets the courier font object.
        /// </summary>
        public static FontBase Courier
        {
            get
            {
                return courier;
            }
        }

        /// <summary>
        /// Gets the courier bold font object.
        /// </summary>
        public static FontBase CourierBold
        {
            get
            {
                return courierBold;
            }
        }

        /// <summary>
        /// Gets the courier oblique font object.
        /// </summary>
        public static FontBase CourierOblique
        {
            get
            {
                return courierOblique;
            }
        }

        /// <summary>
        /// Gets the courier bold oblique font object.
        /// </summary>
        public static FontBase CourierBoldOblique
        {
            get
            {
                return courierBoldOblique;
            }
        }

        internal static Core.Fonts.FontsManager FontsManager
        {
            get;
            set;
        }

        /// <summary>
        /// Registers font.
        /// </summary>
        /// <param name="fontFamily">The font family.</param>
        /// <param name="fontStyle">The font style.</param>
        /// <param name="fontWeight">The font weight.</param>
        /// <param name="data">The font data.</param>
        public static void RegisterFont(FontFamily fontFamily, FontStyle fontStyle, FontWeight fontWeight, byte[] data)
        {
            Guard.ThrowExceptionIfNull(fontFamily, "fontFamily");
            Guard.ThrowExceptionIfNull(fontStyle, "fontStyle");
            Guard.ThrowExceptionIfNull(fontWeight, "fontWeight");
            Guard.ThrowExceptionIfNull(data, "data");

            FontProperties fontProperties = new FontProperties(fontFamily, fontStyle, fontWeight);
            RegisterFont(fontProperties, CreateFont(fontProperties, data));
        }

        internal static bool TryCreateStandardFont(string fontName, out StandardFont font)
        {
            Guard.ThrowExceptionIfNullOrEmpty(fontName, "fontName");

            Func<StandardFont> createFont;

            if (standardFontCreators.TryGetValue(fontName, out createFont))
            {
                font = createFont();
                return true;
            }
            else
            {
                font = null;
                return false;
            }
        }

        private static void RegisterStandartFontCreators()
        {
            RegisterStandardFontCreator(Core.Fonts.FontsManager.Helvetica, HelveticaAscent, HelveticaDescent, HelveticaLineGap, false, false);
            RegisterStandardFontCreator(Core.Fonts.FontsManager.HelveticaBold, HelveticaAscent, HelveticaDescent, HelveticaLineGap, true, false);
            RegisterStandardFontCreator(Core.Fonts.FontsManager.HelveticaOblique, HelveticaAscent, HelveticaDescent, HelveticaLineGap, false, true);
            RegisterStandardFontCreator(Core.Fonts.FontsManager.HelveticaBoldOblique, HelveticaAscent, HelveticaDescent, HelveticaLineGap, true, true);

            RegisterStandardFontCreator(Core.Fonts.FontsManager.Courier, CourierAscent, CourierDescent, CourierLineGap, false, false);
            RegisterStandardFontCreator(Core.Fonts.FontsManager.CourierBold, CourierAscent, CourierDescent, CourierLineGap, true, false);
            RegisterStandardFontCreator(Core.Fonts.FontsManager.CourierOblique, CourierAscent, CourierDescent, CourierLineGap, false, true);
            RegisterStandardFontCreator(Core.Fonts.FontsManager.CourierBoldOblique, CourierAscent, CourierDescent, CourierLineGap, true, true);

            RegisterStandardFontCreator(Core.Fonts.FontsManager.TimesRoman, TimesAscent, TimesDescent, TimesLineGap, false, false);
            RegisterStandardFontCreator(Core.Fonts.FontsManager.TimesBold, TimesAscent, TimesDescent, TimesLineGap, true, false);
            RegisterStandardFontCreator(Core.Fonts.FontsManager.TimesItalic, TimesAscent, TimesDescent, TimesLineGap, false, true);
            RegisterStandardFontCreator(Core.Fonts.FontsManager.TimesBoldItalic, TimesAscent, TimesDescent, TimesLineGap, true, true);

            RegisterStandardFontCreator(Core.Fonts.FontsManager.Symbol, SymbolAscent, SymbolDescent, SymbolLineGap, false, false);
            RegisterStandardFontCreator(Core.Fonts.FontsManager.ZapfDingbats, ZapfDingbatsAscent, ZapfDingbatsDescent, ZapfDingbatsLineGap, false, false);
        }

#if NETSTANDARD
        private static void RegisterStandartFonts()
        {

            RegisterFont(new FontProperties(new FontFamily(Core.Fonts.FontsManager.Helvetica)), FontsRepository.helvetica);

            RegisterFont(new FontProperties(new FontFamily(Core.Fonts.FontsManager.Helvetica), FontStyles.Normal, FontWeights.Bold), FontsRepository.helveticaBold);

            RegisterFont(new FontProperties(new FontFamily(Core.Fonts.FontsManager.Helvetica), FontStyles.Italic, FontWeights.Normal), FontsRepository.helveticaOblique);

            RegisterFont(new FontProperties(new FontFamily(Core.Fonts.FontsManager.Helvetica), FontStyles.Italic, FontWeights.Bold), FontsRepository.helveticaBoldOblique);

            RegisterFont(new FontProperties(new FontFamily(Core.Fonts.FontsManager.Courier), FontStyles.Normal, FontWeights.Normal), FontsRepository.courier);

            RegisterFont(new FontProperties(new FontFamily(Core.Fonts.FontsManager.Courier), FontStyles.Normal, FontWeights.Bold), FontsRepository.courierBold);

            RegisterFont(new FontProperties(new FontFamily(Core.Fonts.FontsManager.Courier), FontStyles.Italic, FontWeights.Normal), FontsRepository.courierOblique);

            RegisterFont(new FontProperties(new FontFamily(Core.Fonts.FontsManager.Courier), FontStyles.Italic, FontWeights.Bold), FontsRepository.courierBoldOblique);

            RegisterFont(new FontProperties(new FontFamily(Core.Fonts.FontsManager.TimesRoman), FontStyles.Normal, FontWeights.Normal), FontsRepository.timesRoman);

            RegisterFont(new FontProperties(new FontFamily(Core.Fonts.FontsManager.TimesRoman), FontStyles.Normal, FontWeights.Bold), FontsRepository.timesBold);

            RegisterFont(new FontProperties(new FontFamily(Core.Fonts.FontsManager.TimesRoman), FontStyles.Italic, FontWeights.Normal), FontsRepository.timesItalic);

            RegisterFont(new FontProperties(new FontFamily(Core.Fonts.FontsManager.TimesRoman), FontStyles.Italic, FontWeights.Bold), FontsRepository.timesBoldItalic);

            RegisterFont(new FontProperties(new FontFamily(Core.Fonts.FontsManager.Symbol), FontStyles.Normal, FontWeights.Normal), FontsRepository.symbol);

            RegisterFont(new FontProperties(new FontFamily(Core.Fonts.FontsManager.ZapfDingbats), FontStyles.Normal, FontWeights.Normal), FontsRepository.zapfDingbats);
        }
#endif

        private static StandardFont CreateStandardFont(string fontName, double ascent, double descent, double lineGap, bool isBold, bool isItalic)
        {
            return new StandardFont(fontName, ascent, descent, lineGap, isBold, isItalic, FontsManager.GetStandardFontSource(fontName));
        }

        private static void RegisterFont(FontProperties fontProperties, FontBase font)
        {
            Guard.ThrowExceptionIfNull(fontProperties, "fontProperties");
            Guard.ThrowExceptionIfNull(font, "font");

            fontPropertiesToFont[fontProperties] = font;
        }

        private static void RegisterStandardFontCreator(string fontName, double ascent, double descent, double lineGap, bool isBold, bool isItalic)
        {
            standardFontCreators[fontName] = () => CreateStandardFont(fontName, ascent, descent, lineGap, isBold, isItalic);
        }

        private static StandardFont CreateStandardFontByName(string fontName)
        {
            return standardFontCreators[fontName]();
        }

        private static FontBase CreateFont(FontProperties fontProperties, byte[] data)
        {
            FontBase font;
            OpenTypeFontReader reader = new OpenTypeFontReader(data);
            switch (OpenTypeFontSource.GetFontType(reader))
            {
                case Core.Fonts.OpenTypeFontSourceType.TrueType:

                    font = CreateFontFromOpenTypeFontSource(new OpenTypeFontSource(fontProperties, reader));

                    break;

                case Core.Fonts.OpenTypeFontSourceType.TrueTypeCollection:

                    font = CreateFontFromTrueTypeCollection(fontProperties, new TrueTypeCollection(reader));

                    break;

                default:

                    font = null;

                    break;
            }

            return font;
        }

        private static FontBase CreateFontFromTrueTypeCollection(FontProperties fontProperties, TrueTypeCollection trueTypeCollection)
        {
            trueTypeCollection.Initialize();
            foreach (OpenTypeFontSource fontSource in trueTypeCollection.Fonts)
            {
                if (fontSource.FontFamily == fontProperties.FontFamilyName)
                {
                    return CreateFontFromOpenTypeFontSource(fontSource);
                }
            }

            return null;
        }

        private static FontBase CreateFontFromOpenTypeFontSource(OpenTypeFontSource fontSource)
        {
            // TODO: Handle CFF fonts. 

            return new CidType2Font(CreateFontName(fontSource), fontSource);
        }

        private static string CreateFontName(OpenTypeFontSource fontSource)
        {
            string fontName = fontSource.FontFamily;
            if (fontSource.IsBold)
            {
                if (fontSource.IsItalic)
                {
                    fontName += ",BoldItalic";
                }
                else
                {
                    fontName += ",Bold";
                }
            }
            else if (fontSource.IsItalic)
            {
                fontName += ",Italic";
            }

            return fontName;
        }
    }
}
