﻿using System;
using Telerik.Windows.Documents.Core.Fonts;
using Telerik.Windows.Documents.Core.Fonts.Glyphs;
using Telerik.Windows.Documents.Core.Fonts.OpenType;
using Telerik.Windows.Documents.Core.Fonts.OpenType.Tables;
using Telerik.Windows.Documents.Core.PostScript.Data;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Model.Fonts.Encoding;
#if !NETSTANDARD
using System.Windows.Media;
#else
using Telerik.Documents.Core.Fonts;
#endif

namespace Telerik.Windows.Documents.Fixed.Model.Fonts
{
    internal class TrueTypeFont : SimpleFont
    {
        TrueTypeGlyphInitializationInfo initializationInfo;

        public TrueTypeFont(string name, FontSource fontSource)
            : base(name, fontSource)
        {
        }

        internal override FontType Type
        {
            get
            {
                return FontType.TrueType;
            }
        }

        internal override FontSource ActualFontSource
        {
            get
            {
                this.EnsureGlyphInitializationInfo();
                FontSource actualFontSource = this.initializationInfo.FontSource ?? (FontSource)EmptyFontSource.Instance;

                return actualFontSource;
            }
        }

        internal static TrueTypeGlyphInitializationInfo GetGlyphInitializationInfo(OpenTypeFontSource fontSource, IFontDescriptor fontDescriptor, ISimpleEncoding encoding)
        {
            TrueTypeGlyphInitializationInfo info = new TrueTypeGlyphInitializationInfo() { Platform = TrueTypeGlyphInitializationInfo.PlatformType.Unknown, Encoding = encoding, FontSource = fontSource };

            if (fontSource == null)
            {
                return info;
            }

            bool isMacOrWinNamedEncoding = encoding != null && encoding.IsNamedEncoding &&
                (encoding.BaseEncodingName == PredefinedSimpleEncoding.MacRomanEncodingName ||
                    encoding.BaseEncodingName == PredefinedSimpleEncoding.WinAnsiEncodingName);

            if ((fontDescriptor != null && fontDescriptor.IsNonSymbolic) || isMacOrWinNamedEncoding)
            {
                info.CMap = fontSource.CMap.GetCMapTable(3, 1);
                if (info.CMap != null)
                {
                    info.Platform = TrueTypeGlyphInitializationInfo.PlatformType.MicrosoftNonSymbolic;
                }
                else
                {
                    info.CMap = fontSource.CMap.GetCMapTable(1, 0);
                    if (info.CMap != null)
                    {
                        info.Platform = TrueTypeGlyphInitializationInfo.PlatformType.MacintoshNonSymbolic;
                    }
                }
            }
            else if ((fontDescriptor != null && fontDescriptor.IsSymbolic) || encoding == null)
            {
                info.CMap = fontSource.CMap.GetCMapTable(3, 0);
                if (info.CMap != null)
                {
                    byte firstByte;
                    if (TryCalculateFirstByte(info.CMap, out firstByte))
                    {
                        info.AppendingFirstByte = firstByte;
                        info.Platform = TrueTypeGlyphInitializationInfo.PlatformType.MicrosoftSymbolic;
                    }
                }
                else
                {
                    info.CMap = fontSource.CMap.GetCMapTable(1, 0);
                    if (info.CMap != null)
                    {
                        info.Platform = TrueTypeGlyphInitializationInfo.PlatformType.MacintoshSymbolic;
                    }
                }
            }

            return info;
        }

        internal static void InitializeCoreGlyph(Glyph glyphInfo, TrueTypeGlyphInitializationInfo initializationInfo, byte b)
        {
            switch (initializationInfo.Platform)
            {
                case TrueTypeGlyphInitializationInfo.PlatformType.MicrosoftNonSymbolic:
                    InitializeMicrosoftNonSymbolicGlyph(glyphInfo, initializationInfo, b);
                    break;
                case TrueTypeGlyphInitializationInfo.PlatformType.MacintoshNonSymbolic:
                    InitializeMacintoshNonSymbolicGlyph(glyphInfo, initializationInfo, b);
                    break;
                case TrueTypeGlyphInitializationInfo.PlatformType.MicrosoftSymbolic:
                    InitializeMicrosoftSymbolicGlyph(glyphInfo, initializationInfo, b);
                    break;
                case TrueTypeGlyphInitializationInfo.PlatformType.MacintoshSymbolic:
                    InitializeMacintoshSymbolicGlyph(glyphInfo, initializationInfo, b);
                    break;
            }
        }

        internal override void InitializeCoreGlyph(Glyph glyphInfo, CharCode charcode)
        {
            byte b = (byte)charcode.Code;
            this.EnsureGlyphInitializationInfo();
            InitializeCoreGlyph(glyphInfo, this.initializationInfo, b);
        }

        private static void InitializeMicrosoftNonSymbolicGlyph(Glyph g, TrueTypeGlyphInitializationInfo initializationInfo, byte b)
        {
            string name = initializationInfo.Encoding.GetName(b);
            g.Name = name;

            ushort glyphId;
            if (AdobeGlyphList.IsSupportedPdfName(name))
            {
                char unicode = AdobeGlyphList.GetUnicode(name);
                initializationInfo.CMap.TryGetGlyphId(unicode, out glyphId);
                g.GlyphId = glyphId;
            }
            else
            {                
                g.GlyphId = GetGlyphIdFromPostTable(initializationInfo.FontSource, g.Name);
            }
        }

        private static void InitializeMacintoshNonSymbolicGlyph(Glyph g, TrueTypeGlyphInitializationInfo initializationInfo, byte b)
        {
            string name = initializationInfo.Encoding.GetName(b);
            g.Name = name;

            if (name == PostScriptConstants.NotDef)
            {
                g.GlyphId = GetGlyphIdFromPostTable(initializationInfo.FontSource, name);
            }
            else
            {
                byte charId = PredefinedSimpleEncoding.StandardMacRomanEncoding.GetCharId(name);
                ushort glyphId;
                initializationInfo.CMap.TryGetGlyphId(charId, out glyphId);
                g.GlyphId = glyphId;
            }
        }

        private static void InitializeMicrosoftSymbolicGlyph(Glyph g, TrueTypeGlyphInitializationInfo initializationInfo, byte b)
        {
            ushort glyphId;
            ushort ch = AppendByte(initializationInfo.AppendingFirstByte.Value, b);
            initializationInfo.CMap.TryGetGlyphId(ch, out glyphId);
            g.GlyphId = glyphId;
        }

        private static void InitializeMacintoshSymbolicGlyph(Glyph g, TrueTypeGlyphInitializationInfo initializationInfo, byte b)
        {
            ushort glyphId;
            initializationInfo.CMap.TryGetGlyphId(b, out glyphId);
            g.GlyphId = glyphId;
        }

        private static ushort GetGlyphIdFromPostTable(OpenTypeFontSource fontSource, string name)
        {
            if (fontSource.Post == null)
            {
                return CMap.MISSING_GLYPH_ID;
            }

            return fontSource.Post.GetGlyphId(name);
        }

        private static bool TryCalculateFirstByte(CMapTable cMap, out byte firstByte)
        {
            bool success = false;
            firstByte = 0;

            try
            {
                ushort firstCode = cMap.FirstCode;
                CharCode charCode = new CharCode(firstCode);
                byte[] bytes = charCode.ToBytes();
                firstByte = bytes[0];

                success = true;
            }
            catch (NotSupportedException)
            {
            }

            return success;
        }

        private static ushort AppendByte(byte firstByte, byte b)
        {
            CharCode charCode = new CharCode(new byte[] { firstByte, b });
            ushort result = (ushort)charCode.Code;

            return result;
        }

        private void EnsureGlyphInitializationInfo()
        {
            if (this.initializationInfo == null)
            {
                OpenTypeFontSource fontSource = this.FontSource as OpenTypeFontSource;

                if (fontSource == null)
                {
                    FontProperties fallbackFontProperties = new FontProperties(new FontFamily(FixedDocumentDefaults.TrueTypeFallbackFontName));
                    fontSource = FontsRepository.FontsManager.GetOpenTypeFontSource(fallbackFontProperties);
                }

                this.initializationInfo = GetGlyphInitializationInfo(fontSource, this, this.Encoding.Value);
            }
        }
    }
}