﻿using System;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.InteractiveForms
{
    /// <summary>
    /// Represents a checkbox field.
    /// </summary>
    public class CheckBoxField : FormField<TwoStatesButtonWidget>
    {
        internal const string DefaultExportValue = "Yes";
        private string exportValue;

        /// <summary>
        /// Initializes a new instance of <see cref="CheckBoxField"/> class. 
        /// </summary>
        /// <param name="fieldName">The field name.</param>
        public CheckBoxField(string fieldName) 
            : base(fieldName)
        {
            this.ExportValue = DefaultExportValue;
        }

        /// <summary>
        /// Gets the field type.
        /// </summary>
        public sealed override FormFieldType FieldType
        {
            get
            {
                return FormFieldType.CheckBox;
            }
        }

        /// <summary>
        /// Gets or sets a boolean value indicating whether the field is checked or not.
        /// </summary>
        public bool IsChecked { get; set; }

        /// <summary>
        /// Gets or sets a boolean value indicating whether the check box should be checked when the form is reset to its default values.
        /// </summary>
        public bool IsCheckedByDefault { get; set; }
                
        /// <summary>
        /// Gets or sets the field export values.
        /// </summary>
        public string ExportValue
        {
            get
            {
                return this.exportValue;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "value");

                this.exportValue = value;
            }
        }

        internal sealed override VariableTextProperties GetDefaultTextProperties()
        {
            VariableTextProperties textProperties = new VariableTextProperties();
            textProperties.Font = FontsRepository.ZapfDingbats;
            textProperties.FontSize = 0;

            return textProperties;
        }

        internal sealed override TwoStatesButtonWidget CreateEmptyWidget()
        {
            return new TwoStatesButtonWidget(this);
        }

        internal sealed override void PrepareWidgetAppearancesForExport()
        {
            foreach (TwoStatesButtonWidget widget in this.Widgets)
            {
                widget.PrepareAppearancesForExport(CheckBoxField.DefaultExportValue, this.IsChecked);
            }
        }

        internal sealed override FormField GetClonedInstanceWithoutWidgetsOverride(RadFixedDocumentCloneContext radFixedDocumentCloneContext)
        {
            CheckBoxField field = new CheckBoxField(this.Name);

            field.IsChecked = this.IsChecked;
            field.IsCheckedByDefault = this.IsCheckedByDefault;
            field.ExportValue = this.ExportValue;

            return field;
        }

        internal sealed override void InitializeWidgetAppearanceProperties(TwoStatesButtonWidget widget)
        {
            widget.AppearanceCharacteristics.NormalCaption = TwoStatesButtonTemplateCaptions.Check;
            widget.Border = new AnnotationBorder(FixedDocumentDefaults.ButtonWidgetBorderThickness, AnnotationBorderStyle.Solid, null);
            widget.AppearanceCharacteristics.BorderColor = RgbColors.Black;
            widget.AppearanceCharacteristics.Background = RgbColors.White;
        }
    }
}
