﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.InteractiveForms
{
    /// <summary>
    /// Represents a collection of field options.
    /// </summary>
    /// <typeparam name="T"></typeparam>
    public class FieldOptionCollection<T> : IEnumerable<T>
        where T : FieldOptionBase
    {
        private readonly List<T> options;

        internal FieldOptionCollection()
        {
            this.options = new List<T>();
        }

        /// <summary>
        /// Gets or sets the element at the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <returns></returns>
        public T this[int index]
        {
            get
            {
                return this.options[index];
            }
            set
            {

                Guard.ThrowExceptionIfNull(value, "value");

                this.OnBeforeRemoveOption(this.options[index]);
                this.OnBeforeAddOption(value);
                this.options[index] = value;
            }
        }

        /// <summary>
        /// Gets the number of elements contained in the collection.
        /// </summary>
        /// <returns>The number of elements contained in the collection.</returns>
        public int Count
        {
            get
            {
                return this.options.Count;
            }
        }

        /// <summary>
        /// Adds option.
        /// </summary>
        /// <param name="option">The option.</param>
        public void Add(T option)
        {
            this.OnBeforeAddOption(option);
            this.options.Add(option);
        }

        /// <summary>
        /// Removes option at specified index.
        /// </summary>
        /// <param name="index">The option index.</param>
        public void RemoveAt(int index)
        {
            this.OnBeforeRemoveOption(this.options[index]);
            this.options.RemoveAt(index);
        }

        /// <summary>
        /// Removes all elements from the collection.
        /// </summary>
        public void Clear()
        {
            this.OnBeforeRemoveOptions(this.options);

            this.options.Clear();
        }

        /// <summary>
        /// Iterates the options in the collection.
        /// </summary>
        /// <returns>The options.</returns>
        public IEnumerator<T> GetEnumerator()
        {
            return this.options.GetEnumerator();
        }

        /// <summary>
        /// Iterates the options in the collection.
        /// </summary>
        /// <returns>The options.</returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.options.GetEnumerator();
        }

        internal bool TryGetOptionIndex(T option, out int index)
        {
            if (option == null)
            {
                index = -1;
            }
            else
            {
                index = this.options.IndexOf(option);
            }

            return index > -1;
        }

        internal bool TryGetValueIndex(string value, out int index)
        {
            index = 0;

            foreach (T option in this)
            {
                if (option.Value.Equals(value))
                {
                    return true;
                }

                index++;
            }

            return false;
        }

        internal virtual void OnBeforeAddOption(T option)
        {
            Guard.ThrowExceptionIfTrue(option.IsAdded, "option.IsAdded");

            option.IsAdded = true;
        }

        internal virtual void OnBeforeRemoveOptions(IEnumerable<T> options)
        {
            foreach(T option in options)
            {
                option.IsAdded = false;
            }
        }

        private void OnBeforeRemoveOption(T option)
        {
            this.OnBeforeRemoveOptions(Enumerable.Repeat(option, 1));
        }
    }
}
