﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.InteractiveForms
{
    /// <summary>
    /// Represents a form field.
    /// </summary>
    public abstract class FormField : IInstanceIdOwner
    {
        private VariableTextProperties textProperties;
        private readonly string name;
        private readonly int id;

        /// <summary>
        /// Initializes a new instance of <see cref="FormField"/>. 
        /// </summary>
        /// <param name="fieldName">The field name.</param>
        internal FormField(string fieldName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(fieldName, "fieldName");

            this.name = fieldName;
            this.id = InstanceIdGenerator.GetNextId();
        }

        /// <summary>
        /// Gets the field type.
        /// </summary>
        public abstract FormFieldType FieldType { get; }

        /// <summary>
        /// Gets the field name.
        /// </summary>
        public string Name
        {
            get
            {
                return this.name;
            }
        }

        /// <summary>
        /// Represents the text properties used by the PDF viewer in order to build the fields appearance dynamically.
        /// </summary>
        public VariableTextProperties TextProperties
        {
            get
            {
                if (this.textProperties == null)
                {
                    this.textProperties = this.GetDefaultTextProperties();
                }

                return this.textProperties;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "value");

                this.textProperties = value;
            }
        }

        /// <summary>
        /// Gets or sets an alternate field name to be used in place of the actual field name wherever the field must be identified in the user interface.
        /// For instance this may be used for error or status messages referring to the field. 
        /// This text is also useful when extracting the document’s contents in support of accessibility to users with disabilities or for other purposes.
        /// </summary>
        public string UserInterfaceName { get; set; }

        /// <summary>
        /// Gets or sets the mapping name to be used when exporting interactive form field data from the document.
        /// </summary>
        public string MappingName { get; set; }

        /// <summary>
        /// Gets or sets boolean value indicating whether the field is readonly.
        /// </summary>
        public bool IsReadOnly { get; set; }

        /// <summary>
        /// Gets or sets boolean value indicating whether the field must have value at the time it is exported by submitting the form.
        /// </summary>
        public bool IsRequired { get; set; }

        /// <summary>
        /// Gets or sets boolean value indicating whether the field should be skipped when submitting the form.
        /// </summary>
        public bool ShouldBeSkipped { get; set; }

        /// <summary>
        /// Gets the widgets.
        /// </summary>
        public IEnumerable<Widget> Widgets
        {
            get
            {
                return this.GetWidgets();
            }
        }

        internal abstract IEnumerable<Widget> GetWidgets();

        internal abstract Widget AddWidget();

        internal FormField GetClonedInstanceWithoutWidgets(RadFixedDocumentCloneContext cloneContext)
        {
            FormField cloning = this.GetClonedInstanceWithoutWidgetsOverride(cloneContext);
            cloning.UserInterfaceName = this.UserInterfaceName;
            cloning.MappingName = this.MappingName;
            cloning.IsReadOnly = this.IsReadOnly;
            cloning.IsRequired = this.IsRequired;
            cloning.ShouldBeSkipped = this.ShouldBeSkipped;
            cloning.TextProperties = new VariableTextProperties(this.TextProperties);

            return cloning;
        }

        internal virtual VariableTextProperties GetDefaultTextProperties()
        {
            return new VariableTextProperties();
        }

        internal abstract FormField GetClonedInstanceWithoutWidgetsOverride(RadFixedDocumentCloneContext cloneContext);

        internal abstract void AddClonedWidget(Widget clonedWidget);

        int IInstanceIdOwner.InstanceId
        {
            get
            {
                return this.id;
            }
        }

        internal void InvalidateWidgetAppearances()
        {
            foreach(Widget widget in this.Widgets)
            {
                widget.InvalidateAppearance();
            }
        }

        internal abstract void PrepareWidgetAppearancesForExport();
    }

    /// <summary>
    /// Represents FormField providing a collection of type specific widgets.
    /// </summary>
    /// <typeparam name="T">The widget type.</typeparam>
    public abstract class FormField<T> : FormField, IWidgetCreator<T>
        where T : Widget
    {
        private WidgetCollection<T> widgets;

        internal FormField(string fieldName)
            : base(fieldName)
        {
        }

        /// <summary>
        /// Gets the widgets.
        /// </summary>
        public new WidgetCollection<T> Widgets
        {
            get
            {
                if (this.widgets == null)
                {
                    this.widgets = this.CreateWidgetsCollection();
                }

                return this.widgets;
            }
        }

        internal sealed override IEnumerable<Widget> GetWidgets()
        {
            foreach (T widget in this.Widgets)
            {
                yield return widget;
            }
        }

        internal sealed override Widget AddWidget()
        {
            return this.Widgets.AddEmptyWidget();
        }

        internal sealed override void AddClonedWidget(Widget clonedWidget)
        {
            this.Widgets.AddClonedWidget(clonedWidget);
        }

        internal WidgetCollection<T> CreateWidgetsCollection()
        {
            return new WidgetCollection<T>(this, this.InitializeWidgetAppearanceProperties);
        }

        internal abstract void InitializeWidgetAppearanceProperties(T widget);

        internal abstract T CreateEmptyWidget();

        T IWidgetCreator<T>.CreateWidget()
        {
            return this.CreateEmptyWidget();
        }
    }
}
