﻿using System;
using System.Collections;
using System.Collections.Generic;

namespace Telerik.Windows.Documents.Fixed.Model.InteractiveForms
{
    /// <summary>
    /// Represents a collection of document fields.
    /// </summary>
    public class FormFieldCollection : IEnumerable<FormField>
    {
        private readonly Dictionary<string, FormField> fields;

        internal FormFieldCollection()
        {
            this.fields = new Dictionary<string, FormField>();
        }

        /// <summary>
        /// Gets FormField instance by field name.
        /// </summary>
        /// <param name="fieldName">The field name.</param>
        /// <returns></returns>
        public FormField this[string fieldName]
        {
            get
            {
                return this.fields[fieldName];
            }
        }

        /// <summary>
        /// Creates new <see cref="PushButtonField"/> and adds it to the collection.
        /// </summary>
        /// <param name="fieldName">The field name.</param>
        /// <returns>The result field.</returns>
        public PushButtonField AddPushButton(string fieldName)
        {
            PushButtonField field = new PushButtonField(fieldName);
            this.Add(field);

            return field;
        }

        /// <summary>
        /// Creates new <see cref="CheckBoxField"/> and adds it to the collection.
        /// </summary>
        /// <param name="fieldName">The field name.</param>
        /// <returns>The result field.</returns>
        public CheckBoxField AddCheckBox(string fieldName)
        {
            CheckBoxField field = new CheckBoxField(fieldName);
            this.Add(field);

            return field;
        }

        /// <summary>
        /// Creates new <see cref="RadioButtonField"/> and adds it to the collection.
        /// </summary>
        /// <param name="fieldName">The field name.</param>
        /// <returns>The result field.</returns>
        public RadioButtonField AddRadioButton(string fieldName)
        {
            RadioButtonField field = new RadioButtonField(fieldName);
            this.Add(field);

            return field;
        }

        /// <summary>
        /// Creates new <see cref="CombTextBoxField"/> and adds it to the collection.
        /// </summary>
        /// <param name="fieldName">The field name.</param>
        /// <returns>The result field.</returns>
        public CombTextBoxField AddCombTextBox(string fieldName)
        {
            CombTextBoxField field = new CombTextBoxField(fieldName);
            this.Add(field);

            return field;
        }

        /// <summary>
        /// Creates new <see cref="TextBoxField"/> and adds it to the collection.
        /// </summary>
        /// <param name="fieldName">The field name.</param>
        /// <returns>The result field.</returns>
        public TextBoxField AddTextBox(string fieldName)
        {
            TextBoxField field = new TextBoxField(fieldName);
            this.Add(field);

            return field;
        }

        /// <summary>
        /// Creates new <see cref="ComboBoxField"/> and adds it to the collection.
        /// </summary>
        /// <param name="fieldName">The field name.</param>
        /// <returns>The result field.</returns>
        public ComboBoxField AddComboBox(string fieldName)
        {
            ComboBoxField field = new ComboBoxField(fieldName);
            this.Add(field);

            return field;
        }

        /// <summary>
        /// Creates new <see cref="ListBoxField"/> and adds it to the collection.
        /// </summary>
        /// <param name="fieldName">The field name.</param>
        /// <returns>The result field.</returns>
        public ListBoxField AddListBox(string fieldName)
        {
            ListBoxField field = new ListBoxField(fieldName);
            this.Add(field);

            return field;
        }

        /// <summary>
        /// Creates new <see cref="SignatureField"/> and adds it to the collection.
        /// </summary>
        /// <param name="fieldName">The field name.</param>
        /// <returns>The result field.</returns>
        public SignatureField AddSignature(string fieldName)
        {
            SignatureField field = new SignatureField(fieldName);
            this.Add(field);

            return field;
        }

        /// <summary>
        /// Adds field to the collection.
        /// </summary>
        /// <param name="field">The field to be added.</param>
        public void Add(FormField field)
        {
            this.fields.Add(field.Name, field);
        }

        /// <summary>
        /// Removes field from the collection.
        /// </summary>
        /// <param name="field">The field to be removed.</param>
        public void Remove(FormField field)
        {
            this.fields.Remove(field.Name);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="fieldName"></param>
        /// <returns></returns>
        public bool Contains(string fieldName)
        {
            return this.fields.ContainsKey(fieldName);
        }

        /// <summary>
        /// Gets the count of the fields in the collection.
        /// </summary>
        public int Count
        {
            get
            {
                return this.fields.Count;
            }
        }

        /// <summary>
        /// Gets the enumerator for the elements in the collection.
        /// </summary>
        /// <returns>The enumerator.</returns>
        public IEnumerator<FormField> GetEnumerator()
        {
            return this.fields.Values.GetEnumerator();
        }

        /// <summary>
        /// Gets the enumerator for the elements in the collection.
        /// </summary>
        /// <returns></returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.fields.Values.GetEnumerator();
        }
    }
}
