﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace Telerik.Windows.Documents.Fixed.Model.InteractiveForms
{

    /// <summary>
    /// Represents a listbox field.
    /// </summary>
    public sealed class ListBoxField : ChoiceField
    {
        private ChoiceOption[] value;
        private int topIndex;

        /// <summary>
        /// Initializes a new instance of <see cref="ListBoxField"/> class. 
        /// </summary>
        /// <param name="fieldName">The field name.</param>
        public ListBoxField(string fieldName)
            : base(fieldName)
        {
        }

        /// <summary>
        /// Gets the field type.
        /// </summary>
        public sealed override FormFieldType FieldType
        {
            get
            {
                return FormFieldType.ListBox;
            }
        }

        /// <summary>
        /// Gets or sets boolean value indicating whether to allow multiple items selection.
        /// </summary>
        public bool AllowMultiSelection { get; set; }

        /// <summary>
        /// Gets or sets an array of selected values.
        /// </summary>
        public ChoiceOption[] Value
        {
            get
            {
                return this.value;
            }
            set
            {
                this.value = value;
                this.InvalidateWidgetAppearances();
            }
        }

        /// <summary>
        /// Gets or set the default selected values used when the form is reset to its default values.
        /// </summary>
        public ChoiceOption[] DefaultValue { get; set; }

        /// <summary>
        /// Gets or sets the index of the first item to be visible when displaying the ListBox.
        /// </summary>
        public int TopIndex
        {
            get
            {
                return this.topIndex;
            }
            set
            {
                if(this.topIndex != value)
                {
                    this.topIndex = value;
                    this.InvalidateWidgetAppearances();
                }
            }
        }

        internal override ChoiceField GetClonedChoiceField(RadFixedDocumentCloneContext cloneContext)
        {
            ListBoxField field = new ListBoxField(this.Name);
            field.AllowMultiSelection = this.AllowMultiSelection;
            field.TopIndex = this.TopIndex;
            field.Value = ListBoxField.GetClonedValue(this.Value, cloneContext);
            field.DefaultValue = ListBoxField.GetClonedValue(this.DefaultValue, cloneContext);

            return field;
        }

        internal int[] GetSelectedIndicesSorted()
        {
            return this.GetIndices(this.Value).OrderBy(i => i).ToArray();
        }

        internal int[] GetDefaultSelectedIndicesSorted()
        {
            return this.GetIndices(this.DefaultValue).OrderBy(i => i).ToArray();
        }

        private IEnumerable<int> GetIndices(ChoiceOption[] options)
        {
            if(options == null)
            {
                yield break;
            }

            foreach(ChoiceOption option in options)
            {
                int index;
                if(this.Options.TryGetOptionIndex(option, out index))
                {
                    yield return index;
                }
            }
        }

        private static ChoiceOption[] GetClonedValue(ChoiceOption[] value, RadFixedDocumentCloneContext cloneContext)
        {
            if(value == null)
            {
                return null;
            }

            ChoiceOption[] cloning = new ChoiceOption[value.Length];

            for(int i = 0; i < cloning.Length; i++)
            {
                cloning[i] = cloneContext.GetClonedOption(value[i]);
            }

            return cloning;
        }
    }
}
