﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.Fonts;

namespace Telerik.Windows.Documents.Fixed.Model.InteractiveForms
{
    /// <summary>
    /// Represents a field which usually provides group with several radio button options.
    /// </summary>
    public sealed class RadioButtonField : FormField, IWidgetCreator<RadioButtonWidget>
    {
        private readonly RadioOptionCollection options;
        private readonly RadioButtonWidgetCollection widgets;

        /// <summary>
        /// Initializes a new instance of <see cref="RadioButtonField"/> class. 
        /// </summary>
        /// <param name="fieldName">The field name.</param>
        public RadioButtonField(string fieldName)
            : base(fieldName)
        {
            this.options = new RadioOptionCollection();
            this.widgets = new RadioButtonWidgetCollection(this);
            this.ShouldUpdateRadiosInUnison = true;
        }

        /// <summary>
        /// Gets the field type.
        /// </summary>
        public override FormFieldType FieldType
        {
            get
            {
                return FormFieldType.RadioButton;
            }
        }

        /// <summary>
        /// Gets the collection of options to choose from.
        /// </summary>
        public RadioOptionCollection Options
        {
            get
            {
                return this.options;
            }
        }

        /// <summary>
        /// Gets the collection of widgets which are representing the radio button options.
        /// </summary>
        public new RadioButtonWidgetCollection Widgets
        {
            get
            {
                return this.widgets;
            }
        }

        /// <summary>
        /// Gets or sets the field value.
        /// </summary>
        public RadioOption Value { get; set; }

        /// <summary>
        /// Gets or sets the default field value used when the form is reset to its default values.
        /// </summary>
        public RadioOption DefaultValue { get; set; }

        /// <summary>
        /// Gets or sets boolean value indicating whether to allow deselecting all radio buttons.
        /// </summary>
        public bool AllowToggleOff { get; set; }

        /// <summary>
        /// Gets or sets boolean value indicating whether to turn on and off in unison radio buttons that have the same value.
        /// </summary>
        public bool ShouldUpdateRadiosInUnison { get; set; }

        internal sealed override VariableTextProperties GetDefaultTextProperties()
        {
            VariableTextProperties textProperties = new VariableTextProperties();
            textProperties.Font = FontsRepository.ZapfDingbats;

            return textProperties;
        }

        internal sealed override IEnumerable<Widget> GetWidgets()
        {
            foreach (RadioButtonWidget widget in this.Widgets)
            {
                yield return widget;
            }
        }

        internal sealed override Widget AddWidget()
        {
            return this.Widgets.AddEmptyWidget();
        }

        internal sealed override void AddClonedWidget(Widget clonedWidget)
        {
            this.Widgets.AddClonedWidget(clonedWidget);
        }

        internal sealed override void PrepareWidgetAppearancesForExport()
        {
            int stateIndex = 0;

            if (this.ShouldUpdateRadiosInUnison)
            {
                Dictionary<string, string> valueToStateName = new Dictionary<string, string>();

                foreach (RadioButtonWidget widget in this.Widgets)
                {
                    string onStateName;
                    if(!valueToStateName.TryGetValue(widget.Option.Value, out onStateName))
                    {
                        onStateName = stateIndex.ToString();
                        valueToStateName.Add(widget.Option.Value, onStateName);
                    }
                    
                    widget.PrepareAppearancesForExport(onStateName, widget.IsSelected);
                    stateIndex++;
                }
            }
            else
            {
                foreach (RadioButtonWidget widget in this.Widgets)
                {
                    string onStateName = stateIndex.ToString();
                    widget.PrepareAppearancesForExport(onStateName, widget.IsSelected);
                    stateIndex++;
                }
            }
        }

        internal sealed override FormField GetClonedInstanceWithoutWidgetsOverride(RadFixedDocumentCloneContext cloneContext)
        {
            RadioButtonField field = new RadioButtonField(this.Name);

            foreach(RadioOption option in this.options)
            {
                field.Options.Add(cloneContext.GetClonedOption(option));
            }

            field.Value = cloneContext.GetClonedOption(this.Value);
            field.DefaultValue = cloneContext.GetClonedOption(this.DefaultValue);
            field.AllowToggleOff = this.AllowToggleOff;
            field.ShouldUpdateRadiosInUnison = this.ShouldUpdateRadiosInUnison;

            return field;
        }

        RadioButtonWidget IWidgetCreator<RadioButtonWidget>.CreateWidget()
        {
            return new RadioButtonWidget(this);
        }
    }
}
