﻿using System;

namespace Telerik.Windows.Documents.Fixed.Model.InteractiveForms
{
    /// <summary>
    /// Represents a textbox field.
    /// </summary>
    public sealed class TextBoxField : TextField
    {
        private bool isMultiline;
        private bool isPassword;
        private int? maxLengthOfInputCharacters;

        /// <summary>
        /// Initializes a new instance of <see cref="TextBoxField"/> class. 
        /// </summary>
        /// <param name="fieldName">The field name.</param>
        public TextBoxField(string fieldName) 
            : base(fieldName)
        {
        }

        /// <summary>
        /// Gets the field type.
        /// </summary>
        public override FormFieldType FieldType
        {
            get
            {
                return FormFieldType.TextBox;
            }
        }

        /// <summary>
        /// Gets or sets boolean value indicating whether multiline text input is allowed.
        /// </summary>
        public bool IsMultiline
        {
            get
            {
                return this.isMultiline;
            }
            set
            {
                if(this.isMultiline != value)
                {
                    this.InvalidateWidgetAppearances();
                    this.isMultiline = value;
                }
            }
        }

        /// <summary>
        /// Gets or sets boolean value indicating whether the text input is a password.
        /// </summary>
        public bool IsPassword
        {
            get
            {
                return this.isPassword;
            }
            set
            {
                if (this.isPassword != value)
                {
                    this.InvalidateWidgetAppearances();
                    this.isPassword = value;
                }
            }
        }

        /// <summary>
        /// Gets or sets boolean value indicating whether the field represents pathname of a file whose contents are to be sumbitted as the value of the field.
        /// </summary>
        public bool IsFileSelect { get; set; }

        /// <summary>
        /// Gets or sets boolean value indicating whether the inputted text should be spell checked.
        /// </summary>
        public bool ShouldSpellCheck { get; set; }

        /// <summary>
        /// Gets or sets boolean value indicating whether to allow scrolling for bigger text content.
        /// If the scrolling is not allowed then the maximum text input is restricted to the Widget annotation rectangle.
        /// </summary>
        public bool AllowScroll { get; set; }

        /// <summary>
        /// Gets or sets the max length of the input characters. If null then the text input is not restricted from these value.
        /// </summary>
        public int? MaxLengthOfInputCharacters
        {
            get
            {
                return this.maxLengthOfInputCharacters;
            }
            set
            {
                if (this.maxLengthOfInputCharacters != value)
                {
                    this.InvalidateWidgetAppearances();
                    this.maxLengthOfInputCharacters = value;
                }
            }
        }

        internal sealed override TextField GetClonedTextField(RadFixedDocumentCloneContext cloneContext)
        {
            TextBoxField field = new TextBoxField(this.Name);
            field.IsMultiline = this.IsMultiline;
            field.IsPassword = this.IsPassword;
            field.IsFileSelect = this.IsFileSelect;
            field.ShouldSpellCheck = this.ShouldSpellCheck;
            field.AllowScroll = this.AllowScroll;
            field.MaxLengthOfInputCharacters = this.MaxLengthOfInputCharacters;

            return field;
        }
    }
}
