﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.Annotations.EventArgs;
using Telerik.Windows.Documents.Fixed.Model.Internal.Collections;
using Telerik.Windows.Documents.Fixed.Model.Navigation;
using Telerik.Windows.Documents.Fixed.Search;
using Telerik.Windows.Documents.Fixed.Text.Internal;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Internal
{
    /// <summary>
    /// Hosts a fixed-format document.
    /// </summary>
    internal class RadFixedDocumentInternal
    {
        private readonly PdfFormatProvider formatProvider;
        private PagesCollection pages;
        private TextDocumentOld textDocument;
        private TextSearch textSearch;

        /// <summary>
        /// Initializes a new instance of the <see cref="RadFixedDocument"/> class.
        /// </summary>
        internal RadFixedDocumentInternal(PdfFormatProvider formatProvider, RadFixedDocument publicDocument)
        {
            Guard.ThrowExceptionIfNull(formatProvider, "formatProvider");

            this.formatProvider = formatProvider;
            this.PublicDocument = publicDocument;
        }

        /// <summary>
        /// Occurs when exception is thrown.
        /// </summary>
        public event EventHandler<OnDocumentExceptionEventArgs> OnException;

        internal event EventHandler<AnnotationEventArgs> AnnotationClicked;

        public RadFixedDocument PublicDocument { get; private set; }

        /// <summary>
        /// Gets the collection of the document pages.
        /// </summary>
        /// <value>
        /// The pages collection.
        /// </value>
        public PagesCollection Pages
        {
            get
            {
                return this.pages;
            }
            internal set
            {
                if (this.pages != value)
                {
                    this.pages = value;
                }
            }
        }

        /// <summary>
        /// Gets all destinations in the document.
        /// </summary>
        /// <value>The destinations.</value>
        public IEnumerable<Destination> Destinations
        {
            get
            {
                List<Destination> res = new List<Destination>();
                foreach (RadFixedPageInternal page in this.Pages)
                {
                    foreach (Annotation a in page.Annotations)
                    {
                        Link l = a as Link;
                        if (l != null && l.Destination != null)
                        {
                            res.Add(l.Destination);
                        }
                    }
                }
                return res;
            }
        }

        /// <summary>
        /// Gets all annotations in the document.
        /// </summary>
        /// <value>The annotations.</value>
        public IEnumerable<Annotation> Annotations
        {
            get
            {
                List<Annotation> res = new List<Annotation>();
                foreach (RadFixedPageInternal page in this.Pages)
                {
                    res.AddRange(page.Annotations);
                }
                return res;
            }
        }

        internal PdfFormatProvider FormatProvider
        {
            get
            {
                return this.formatProvider;
            }
        }

        internal TextDocumentOld TextDocument
        {
            get
            {
                if (this.textDocument == null)
                {
                    this.textDocument = new TextDocumentOld(this);
                }

                return this.textDocument;
            }
        }

        internal TextSearch TextSearch
        {
            get
            {
                if (this.textSearch == null)
                {
                    this.textSearch = new TextSearch(this.TextDocument);
                }

                return this.textSearch;
            }
        }

        internal void OnAnnotationClicked(object sender, AnnotationEventArgs e)
        {
            if (this.AnnotationClicked != null)
            {
                this.AnnotationClicked(sender, e);
            }
        }

        internal void OnOnException(OnDocumentExceptionEventArgs args)
        {
            if (this.OnException != null)
            {
                this.OnException(this, args);
            }
        }
    }
}
