﻿using System.ComponentModel;
using Telerik.Windows.Documents.Fixed.Model.Actions;
using Telerik.Windows.Documents.Fixed.Model.Collections;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Navigation
{
    /// <summary>
    /// Represents a PDF bookmark (outline) item.
    /// </summary>
    public class BookmarkItem : FixedDocumentElementBase, IContextClonable<BookmarkItem>
    {
        private readonly BookmarksCollection children;

        private string title;
        private RgbColor textColor;

        /// <summary>
        /// Initializes a new instance of the <see cref="BookmarkItem" /> class.
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        [System.Obsolete("Use the BookmarkItem(string title, Destination destination) or BookmarkItem(string title, NamedDestination namedDestination), or BookmarkItem(string title, Action action) constructor instead.")]
        public BookmarkItem() : this(string.Empty) { }

        /// <summary>
        /// Initializes a new instance of the <see cref="BookmarkItem" /> class.
        /// </summary>
        /// <param name="title">The text to be displayed for this bookmark item.</param>
        /// <param name="destination">The destination to be displayed when this bookmark item is activated.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when the title or destination is <c>null</c>.</exception>
        public BookmarkItem(string title, Destination destination)
            : this(title)
        {
            Guard.ThrowExceptionIfNull(destination, "destination");

            this.Destination = destination;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="BookmarkItem" /> class.
        /// </summary>
        /// <param name="title">The text to be displayed for this bookmark item.</param>
        /// <param name="action">The action to be performed when this bookmark item is activated.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when the title or action is <c>null</c>.</exception>
        public BookmarkItem(string title, Action action)
            : this(title)
        {
            Guard.ThrowExceptionIfNull(action, "action");

            this.Action = action;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="BookmarkItem" /> class.
        /// </summary>
        /// <param name="title">The text to be displayed for this bookmark item.</param>
        /// <param name="namedDestination">The named destination to be displayed when this bookmark item is activated.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when the title or namedDestination is <c>null</c>.</exception>
        public BookmarkItem(string title, NamedDestination namedDestination)
            : this(title)
        {
            Guard.ThrowExceptionIfNull(namedDestination, "namedDestination");

            this.NamedDestination = namedDestination;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="BookmarkItem" /> class.
        /// </summary>
        /// <param name="title">The text to be displayed for this bookmark item.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when the title is <c>null</c>.</exception>
        internal BookmarkItem(string title)
        {
            Guard.ThrowExceptionIfNull(title, "title");

            this.Title = title;
            this.IsExpanded = false;
            this.TextStyle = BookmarkItemStyles.Normal;
            this.TextColor = RgbColors.Black;
            this.children = new BookmarksCollection(this);
        }

        /// <summary>
        /// Gets or sets whether the bookmark item is open or closed by default. For example, when 
        /// an item is open by default, its children are initially visible when document is loaded.
        /// </summary>
        public bool IsExpanded { get; set; }

        /// <summary>
        /// Gets the destination to be displayed when this bookmark item is activated.
        /// </summary>
        public Destination Destination { get; private set; }

        /// <summary>
        /// Gets the named destination to be displayed when this bookmark item is activated.
        /// </summary>
        public NamedDestination NamedDestination { get; private set; }

        /// <summary>
        /// Gets the action to be performed when this bookmark item is activated.
        /// </summary>
        public Action Action { get; private set; }

        /// <summary>
        /// Gets or sets the style characteristics for displaying the bookmark item's text.
        /// </summary>
        public BookmarkItemStyles TextStyle { get; set; }

        /// <summary>
        /// Gets or sets the color of the bookmark item's text in RGB color space.
        /// </summary>
        /// <exception cref="System.ArgumentNullException">Thrown when the value is <c>null</c>.</exception>
        public RgbColor TextColor
        {
            get
            {
                return this.textColor;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "TextColor");

                this.textColor = value;
            }
        }

        /// <summary>
        /// Gets or sets the text to be displayed for this bookmark item.
        /// </summary>
        /// <exception cref="System.ArgumentNullException">Thrown when the value is <c>null</c>.</exception>
        public string Title
        {
            get
            {
                return this.title;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "Title");

                this.title = value;
            }
        }

        /// <summary>
        /// Gets the immediate child elements for this bookmark item.
        /// </summary>
        public BookmarksCollection Children
        {
            get
            {
                return this.children;
            }
        }

        internal override FixedDocumentElementType ElementType
        {
            get
            {
                return FixedDocumentElementType.BookmarkItem;
            }
        }

        BookmarkItem IContextClonable<BookmarkItem>.Clone(RadFixedDocumentCloneContext cloneContext)
        {
            RgbColor textColor = this.TextColor;

            BookmarkItem clonedBookmark = new BookmarkItem(this.Title)
            {
                IsExpanded = this.IsExpanded,
                TextStyle = this.TextStyle,
                TextColor = new RgbColor(textColor.A, textColor.R, textColor.G, textColor.B)
            };

            if (this.Destination != null)
            {
                clonedBookmark.Destination = cloneContext.GetClonedDestination(this.Destination);
            }

            if (this.Action != null)
            {
                clonedBookmark.Action = this.Action.Clone(cloneContext);
            }

            if (this.NamedDestination != null)
            {
                clonedBookmark.NamedDestination = cloneContext.GetClonedNamedDestination(this.NamedDestination);
            }

            for (int i = 0; i < this.Children.Count; i++)
            {
                BookmarkItem child = this.Children[i];

                BookmarkItem clonedChild = ((IContextClonable<BookmarkItem>)child).Clone(cloneContext);

                clonedBookmark.Children.Add(clonedChild);
            }

            return clonedBookmark;
        }
    }
}
