﻿using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.Collections;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Navigation;
using Telerik.Windows.Documents.Fixed.Selection;
using Telerik.Windows.Documents.Fixed.Text;
#if !NETSTANDARD
using System;
using Telerik.Windows.Documents.Fixed.Model.Internal;
#endif
using Telerik.Windows.Documents.Utilities;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;

namespace Telerik.Windows.Documents.Fixed.Model
{
    /// <summary>
    /// Represents RadFixedDocument.
    /// </summary>
    public class RadFixedDocument : FixedDocumentElementBase, IFixedDocument, IDestinationContainer
    {
        private readonly AcroForm acroForm;
        private readonly PageCollection pages;
        private readonly NamedDestinationsCollection namedDestinations;
        private readonly RadFixedDocumentInfo documentInfo;
        private readonly BookmarksCollection bookmarks;

        private TextSelection selection;
        private TextPosition caretPosition;

#if !NETSTANDARD
        private RadFixedDocumentInternal internalDocument;
#endif

        /// <summary>
        /// Initializes a new instance of the <see cref="RadFixedDocument" /> class.
        /// </summary>
        public RadFixedDocument()
        {
            this.acroForm = new AcroForm();
            this.pages = new PageCollection(this);
            this.namedDestinations = new NamedDestinationsCollection(this);
            this.documentInfo = new RadFixedDocumentInfo();
            this.bookmarks = new BookmarksCollection(this);
            this.PageMode = PageMode.UseNone;
            this.SupportsSelection = false;
        }

        /// <summary>
        /// Gets the document info.
        /// </summary>
        /// <value>The document info.</value>
        public RadFixedDocumentInfo DocumentInfo
        {
            get
            {
                return this.documentInfo;
            }
        }

        /// <summary>
        /// Gets the pages collection.
        /// </summary>
        /// <value>The pages collection.</value>
        public PageCollection Pages
        {
            get
            {
                return this.pages;
            }
        }

        /// <summary>
        /// Gets the annotations collection.
        /// </summary>
        /// <value>The annotations.</value>
        public IEnumerable<Annotation> Annotations
        {
            get
            {
                foreach (RadFixedPage page in this.Pages)
                {
                    foreach (Annotation annotation in page.Annotations)
                    {
                        yield return annotation;
                    }
                }
            }
        }

        /// <summary>
        /// Gets the named destinations collection.
        /// </summary>
        /// <value>The destinations.</value>
        public NamedDestinationsCollection NamedDestinations
        {
            get
            {
                return this.namedDestinations;
            }
        }

        /// <summary>
        /// Represents the interactive form properties of the current document.
        /// </summary>
        public AcroForm AcroForm
        {
            get
            {
                return this.acroForm;
            }
        }

        /// <summary>
        /// Gets the document bookmarks (outlines) collection.
        /// </summary>
        public BookmarksCollection Bookmarks
        {
            get
            {
                return this.bookmarks;
            }
        }

        /// <summary>
        /// Gets or sets how the document should be displayed when opened.
        /// </summary>
        public PageMode PageMode { get; set; }

        /// <summary>
        /// Merges this document with the specified source document.
        /// </summary>
        /// <param name="source">The source document.</param>
        public void Merge(RadFixedDocument source)
        {
            Guard.ThrowExceptionIfNull(source, "source");

            RadFixedDocument clonedDocument = source.Clone();

            while (clonedDocument.NamedDestinations.Count > 0)
            {
                NamedDestination namedDestination = clonedDocument.NamedDestinations.First();
                clonedDocument.NamedDestinations.Remove(namedDestination.Name);
                this.NamedDestinations.Add(namedDestination);
            }

            while (clonedDocument.Pages.Count > 0)
            {
                RadFixedPage page = clonedDocument.Pages.First();
                clonedDocument.Pages.Remove(page);
                this.Pages.Add(page);
            }

            while (clonedDocument.Bookmarks.Count > 0)
            {
                BookmarkItem bookmark = clonedDocument.Bookmarks.First();
                clonedDocument.Bookmarks.Remove(bookmark);
                this.Bookmarks.Add(bookmark);
            }
        }

        /// <summary>
        /// Clones the document content.
        /// </summary>
        /// <returns>The cloned RadFixedDocument instance.</returns>
        public RadFixedDocument Clone()
        {
            RadFixedDocument clonedDocument = new RadFixedDocument();
            RadFixedDocumentCloneContext cloneContext = new RadFixedDocumentCloneContext();

            foreach (RadFixedPage page in this.Pages)
            {
                clonedDocument.Pages.Add(cloneContext.GetClonedPage(page));
            }

            foreach (NamedDestination namedDestination in this.NamedDestinations)
            {
                clonedDocument.namedDestinations.Add(cloneContext.GetClonedNamedDestination(namedDestination));
            }

            clonedDocument.AcroForm.ViewersShouldRecalculateWidgetAppearances = this.AcroForm.ViewersShouldRecalculateWidgetAppearances;

            foreach (FormField field in this.AcroForm.FormFields)
            {
                FormField clonedField = cloneContext.GetClonedField(field);
                clonedDocument.AcroForm.FormFields.Add(clonedField);

                foreach (Widget widget in field.Widgets)
                {
                    Widget clonedWidget = (Widget)cloneContext.GetClonedAnnotation(widget);
                    clonedField.AddClonedWidget(clonedWidget);
                }
            }

            for (int i = 0; i < this.Bookmarks.Count; i++)
            {
                BookmarkItem bookmark = this.Bookmarks[i];
                clonedDocument.Bookmarks.Add(cloneContext.GetClonedBookmark(bookmark));
            }

            clonedDocument.DocumentInfo.Author = this.DocumentInfo.Author;
            clonedDocument.DocumentInfo.Description = this.DocumentInfo.Description;
            clonedDocument.DocumentInfo.Title = this.DocumentInfo.Title;

            return clonedDocument;
        }

#if !NETSTANDARD
        // Legacy

        /// <summary>
        /// Occurs when exception is thrown. This property only works with RadPdfViewer.
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public event EventHandler<OnDocumentExceptionEventArgs> OnException;
#endif

        /// <summary>
        /// Gets the selection of the document.
        /// </summary>
        /// <value>The selection.</value>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public TextSelection Selection
        {
            get
            {
                if (this.selection == null)
                {
                    this.selection = new TextSelection(this.TextDocument);
                }

                return this.selection;
            }
        }

        /// <summary>
        /// Gets or sets the caret position in the document. This property only works with RadPdfViewer.
        /// </summary>
        /// <value>The caret position.</value>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public TextPosition CaretPosition
        {
            get
            {
                if (this.caretPosition == null)
                {
                    this.caretPosition = new TextPosition(this.TextDocument);
                }

                return this.caretPosition;
            }
        }

        internal TextDocument TextDocument { get; set; }

        internal bool SupportsSelection { get; set; }

        /// <summary>
        /// Gets or sets the Cache Manager responsible for page content caching operations.
        /// </summary>
        internal IPagesCacheManager CacheManager { get; set; }

#if !NETSTANDARD
        internal RadFixedDocumentInternal InternalDocument
        {
            get
            {
                return this.internalDocument;
            }
            set
            {
                if (this.internalDocument != value)
                {
                    if (this.internalDocument != null)
                    {
                        this.internalDocument.OnException -= this.InternalDocument_OnException;
                    }

                    this.internalDocument = value;

                    if (this.internalDocument != null)
                    {
                        this.internalDocument.OnException += this.InternalDocument_OnException;
                    }
                }
            }
        }

        private void InternalDocument_OnException(object sender, OnDocumentExceptionEventArgs e)
        {
            if (this.OnException != null)
            {
                this.OnException(this, e);
            }
        }

        // End Legacy
#endif

        internal override FixedDocumentElementType ElementType
        {
            get
            {
                return FixedDocumentElementType.RadFixedDocument;
            }
        }
    }
}
