﻿using System;
using System.ComponentModel;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
using Telerik.Windows.Documents.Fixed.Model.Internal;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
#endif
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.Collections;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Data;

namespace Telerik.Windows.Documents.Fixed.Model
{
    /// <summary>
    /// Represents fixed page.
    /// </summary>
    public class RadFixedPage : FixedDocumentElementBase, IContentRootElement, IContextClonable<RadFixedPage>, IInstanceIdOwner, IFixedPage
    {
#if SILVERLIGHT
        private readonly object lockObject;
#endif
        private readonly int id;
        private readonly ContentElementCollection content;

        private AnnotationCollection annotations;
        private Rect? cropBox;

        /// <summary>
        /// Initializes a new instance of the <see cref="RadFixedPage" /> class.
        /// </summary>
        public RadFixedPage()
            : this(null)
        {
        }

        internal RadFixedPage(RadFixedDocument parent)
            : base(parent)
        {
#if SILVERLIGHT
            this.lockObject = new object();
#endif
            this.id = InstanceIdGenerator.GetNextId();
            this.content = new ContentElementCollection(this);
            this.MediaBox = new Rect(0, 0, FixedDocumentDefaults.PageSize.Width, FixedDocumentDefaults.PageSize.Height);
            this.Rotation = Rotation.Rotate0;
        }

        /// <summary>
        /// Gets if the element supports annotations.
        /// </summary>
        /// <value>True if the element supports annotations.</value>
        public bool SupportsAnnotations
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Gets the content elements collection.
        /// </summary>
        /// <value>The content elements collection.</value>
        public ContentElementCollection Content
        {
            get
            {
                return this.content;
            }
        }

        /// <summary>
        /// Gets the annotations collection.
        /// </summary>
        /// <value>The annotations collection.</value>
        public AnnotationCollection Annotations
        {
            get
            {
                this.LoadAnnotations();

                return this.annotations;
            }
        }

        /// <summary>
        /// Gets or sets the size of the physical medium on which the page is intended to be displayed or printed.
        /// <para>For more control, use <see cref="MediaBox"/> and <see cref="CropBox"/> properties.</para>
        /// </summary>
        public Size Size
        {
            get
            {
#if SILVERLIGHT
                return new Size(this.MediaBox.Width, this.MediaBox.Height);
#else
                return this.MediaBox.Size;
#endif
            }
            set
            {
                this.MediaBox = new Rect(this.MediaBox.X, this.MediaBox.Y, value.Width, value.Height);
            }
        }

        /// <summary>
        /// Gets or sets the rectangle defining the boundaries of the physical medium 
        /// on which the page is intended to be displayed or printed.
        /// </summary>
        public Rect MediaBox { get; set; }

        /// <summary>
        /// Gets or sets the rectangle defining the boundaries of the visible region of the page. 
        /// When the page is displayed or printed, its contents are to be clipped (cropped) to this rectangle.
        /// </summary>
        public Rect CropBox
        {
            get
            {
                if (this.cropBox.HasValue)
                {
                    return this.cropBox.Value;
                }

                return this.MediaBox;
            }
            set
            {
                this.cropBox = value;
            }
        }

        /// <summary>
        /// Gets or sets the rotate angle.
        /// </summary>
        /// <value>The rotate angle.</value>
        public Rotation Rotation { get; set; }

        // Legacy

#if !NETSTANDARD
        /// <summary>
        /// Gets the fixed page's width. This property only works with RadPdfViewer.
        /// </summary>
        /// <value>The width.</value>
        [EditorBrowsable(EditorBrowsableState.Never)]
        [Obsolete("Use Telerik.Windows.Documents.Fixed.Utilities.PageLayoutHelper GetActualWidth method instead.")]
        public double ActualWidth
        {
            get
            {
                if (this.Rotation == Rotation.Rotate0 || this.Rotation == Rotation.Rotate180)
                {
                    return this.Size.Width;
                }
                else
                {
                    return this.Size.Height;
                }
            }
        }

        /// <summary>
        /// Gets the fixed page's height. This property only works with RadPdfViewer.
        /// </summary>
        /// <value>The height.</value>
        [EditorBrowsable(EditorBrowsableState.Never)]
        [Obsolete("Use Telerik.Windows.Documents.Fixed.Utilities.PageLayoutHelper GetActualHeight method instead.")]
        public double ActualHeight
        {
            get
            {
                if (this.Rotation == Rotation.Rotate0 || this.Rotation == Rotation.Rotate180)
                {
                    return this.Size.Height;
                }
                else
                {
                    return this.Size.Width;
                }
            }
        }

        /// <summary>
        /// Gets a value indicating if the page has content. This property only works with RadPdfViewer.
        /// </summary>
        /// <value>A value indicating if the page has content.</value>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public bool HasContent
        {
            get
            {
                return this.InternalRadFixedPage.HasContent;
            }
        }

        /// <summary>
        /// Gets the document that contains the page. This property only works with RadPdfViewer.
        /// </summary>
        /// <value>The document.</value>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public RadFixedDocument Document
        {
            get
            {
                return (RadFixedDocument)this.Parent;
            }
        }
#endif
        /// <summary>
        /// Gets the current page number. This property only works with RadPdfViewer.
        /// </summary>
        /// <value>The page no.</value>
        [EditorBrowsable(EditorBrowsableState.Never)]
        [Obsolete("Use RadFixedDocument.Pages.IndexOf(RadFixedPage) method instead.")]
        public int PageNo
        {
            get {
                return this.PageNumber;
            }
            set
            {
                this.PageNumber = value;
            }
        }

        internal int PageNumber { get; set; }

        internal bool HasLoadedAnnotations
        {
            get
            {
                return this.annotations != null;
            }
        }

#if !NETSTANDARD
        internal RadFixedPageInternal InternalRadFixedPage
        {
            get;
            set;
        }
#endif
        // End legacy

        internal override FixedDocumentElementType ElementType
        {
            get
            {
                return FixedDocumentElementType.RadFixedPage;
            }
        }

        int IInstanceIdOwner.InstanceId
        {
            get
            {
                return this.id;
            }
        }

        internal void LoadAnnotations()
        {
            if (!this.HasLoadedAnnotations)
            {
                this.annotations = new AnnotationCollection(this);

#if !NETSTANDARD
                if (this.InternalRadFixedPage != null)
                {
                    PdfFormatProvider formatProvider = this.InternalRadFixedPage.Document.FormatProvider;

#if SILVERLIGHT
                    lock (this.lockObject)
#else
                    lock (formatProvider)
#endif
                    {
                        formatProvider.LoadPageAnnotations(this);
                    }
                }
#endif
            }
        }

        RadFixedPage IContextClonable<RadFixedPage>.Clone(RadFixedDocumentCloneContext cloneContext)
        {
            RadFixedPage clonedPage = new RadFixedPage();

            clonedPage.Size = this.Size;
            clonedPage.Rotation = this.Rotation;

            foreach (PositionContentElement element in this.Content)
            {
                clonedPage.Content.Add(element.Clone(cloneContext));
            }

            foreach (Annotation annotation in this.Annotations)
            {
                clonedPage.Annotations.Add(cloneContext.GetClonedAnnotation(annotation));
            }

            return clonedPage;
        }
    }
}
