﻿using System;
using System.IO;
using Telerik.Windows.Documents.Core.Imaging.Jpeg;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.Decoder;
using Telerik.Windows.Documents.Core.Imaging.Jpeg2000;
using Telerik.Windows.Documents.Core.Imaging.Jpeg2000.Decoder;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Filters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Core.Imaging.Jpeg.ColorSpaces;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Model.Resources
{
    /// <summary>
    /// Represents encoded image data.
    /// </summary>
    public class EncodedImageData
    {
        /// <summary>
        /// Gets the data.
        /// </summary>
        /// <value>The data.</value>
        public byte[] Data { get; private set; }

        /// <summary>
        /// Gets the alpha channel data. This data should be encoded with same bitsPerComponent, width, height and filters as the image data.
        /// The colorspace of the alpha channel is not related to the colorspace of the image data and the alpha data colorspace is always DeviceGray.
        /// </summary>
        public byte[] AlphaChannel { get; private set; }

        /// <summary>
        /// Gets the bits per component.
        /// </summary>
        /// <value>The bits per component.</value>
        public int BitsPerComponent { get; private set; }

        /// <summary>
        /// Gets the width.
        /// </summary>
        /// <value>The width.</value>
        public int Width { get; private set; }

        /// <summary>
        /// Gets the height.
        /// </summary>
        /// <value>The height.</value>
        public int Height { get; private set; }

        /// <summary>
        /// Gets the color space.
        /// </summary>
        /// <value>The color space.</value>
        public string ColorSpace { get; private set; }

        /// <summary>
        /// Gets the filters.
        /// </summary>
        /// <value>The filters.</value>
        public string[] Filters { get; private set; }

        internal double[] DecodeArray { get; set; }

        /// <summary>
        /// Initializes a new instance of the <see cref="EncodedImageData" /> class.
        /// </summary>
        /// <param name="data">The data.</param>
        /// <param name="bitsPerComponent">The bits per component.</param>
        /// <param name="width">The width.</param>
        /// <param name="height">The height.</param>
        /// <param name="colorSpace">The color space.</param>
        /// <param name="filters">The filters.</param>
        public EncodedImageData(byte[] data, int bitsPerComponent, int width, int height, string colorSpace, string[] filters) 
            : this(data, null, bitsPerComponent, width, height, colorSpace, filters)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="EncodedImageData" /> class.
        /// </summary>
        /// <param name="data">The data.</param>
        /// <param name="alphaChannel">The alpha channel data. This data should be encoded with same bitsPerComponent, width, height and filters as the image data.
        /// The colorspace of the alpha channel is not related to the colorspace of the image data and the alpha data colorspace is always DeviceGray.</param>
        /// <param name="bitsPerComponent">The bits per component.</param>
        /// <param name="width">The width.</param>
        /// <param name="height">The height.</param>
        /// <param name="colorSpace">The color space.</param>
        /// <param name="filters">The filters.</param>
        public EncodedImageData(byte[] data, byte[] alphaChannel, int bitsPerComponent, int width, int height, string colorSpace, string[] filters)
        {
            Guard.ThrowExceptionIfNull(data, "data");

            this.ColorSpace = colorSpace;
            this.Filters = filters;
            this.Data = data;
            this.AlphaChannel = alphaChannel;
            this.BitsPerComponent = bitsPerComponent;
            this.Width = width;
            this.Height = height;
        }

        internal static bool TryCreateFromUnknownImageData(byte[] unknownImageData, out EncodedImageData encodedImageData)
        {
            encodedImageData = null;

            return TryCreateJpegEncodedImagaData(unknownImageData, out encodedImageData) || TryCreateJpeg2000EncodedImagaData(unknownImageData, out encodedImageData);
        }

        private static bool TryCreateJpegEncodedImagaData(byte[] data, out EncodedImageData encodedImageData)
        {
            string colorspace;
            JpegImageInfo jpegInfo;
            JpegHeadersDecoder jpegDecoder = new JpegHeadersDecoder(data);
            if (jpegDecoder.TryDecodeJpegImage(out jpegInfo) && TryGetJpegColorspace(jpegInfo.ColorSpace, out colorspace))
            {
                encodedImageData = new EncodedImageData(data, ImageSource.DefaultBitsPerComponent, jpegInfo.Width, jpegInfo.Height, colorspace, new string[] { PdfFilterNames.DCTDecode });
                encodedImageData.DecodeArray = jpegInfo.HasAdobeInvertedColors ? new double[] { 1, 0, 1, 0, 1, 0, 1, 0 } : null;

                return true;
            }

            encodedImageData = null;
            return false;
        }

        private static bool TryCreateJpeg2000EncodedImagaData(byte[] data, out EncodedImageData encodedImageData)
        {
            Jpeg2000ImageInfo jpxInfo;
            Jpeg2000HeadersDecoder jpxDecoder = new Jpeg2000HeadersDecoder(data);
            if (jpxDecoder.TryDecodeJpeg2000Image(out jpxInfo))
            {
                encodedImageData = new EncodedImageData(data, 0, jpxInfo.Width, jpxInfo.Height, null, new string[] { PdfFilterNames.JPXDecode });
                return true;
            }

            encodedImageData = null;
            return false;
        }

        private static bool TryGetJpegColorspace(JpegColorSpace jpegColorspace, out string colorspace)
        {
            colorspace = null;

            switch(jpegColorspace)
            {
                case JpegColorSpace.Rgb:
                    colorspace = ColorSpaceNames.DeviceRgb;
                    break;
                case JpegColorSpace.Grayscale:
                    colorspace = ColorSpaceNames.DeviceGray;
                    break;
                case JpegColorSpace.Cmyk:
                    colorspace = ColorSpaceNames.DeviceCmyk;
                    break;
            }

            return colorspace != null;
        }

    }
}