﻿using System;
using System.Windows;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Extensions;
using Telerik.Windows.Documents.Primitives;

namespace Telerik.Windows.Documents.Fixed.Model.Resources
{
    /// <summary>
    /// This class handles the logic for both ImageMask and ColorKeyMask depending on the constructor.
    /// </summary>
    internal class ImageMask : IMask
    {
        private readonly PdfProperty<ImageSource> image;
        private readonly PdfProperty<int[]> colorMask;

        public ImageMask(ImageSource image)
            : this()
        {
            this.Image.Value = image;
        }

        public ImageMask(int[] colorMaskArray)
            : this()
        {
            this.ColorMask.Value = colorMaskArray;
        }

        private ImageMask()
        {
            this.image = new PdfProperty<ImageSource>();
            this.colorMask = new PdfProperty<int[]>();
        }

        public PdfProperty<ImageSource> Image
        {
            get
            {
                return this.image;
            }
        }

        public PdfProperty<int[]> ColorMask
        {
            get
            {
                return this.colorMask;
            }
        }

        public bool IsColorKeyMask
        {
            get
            {
                return this.colorMask.HasValue;
            }
        }

        public byte[] Data
        {
            get
            {
                if (this.IsColorKeyMask)
                {
                    return null;
                }
                else
                {
                    IImageDescriptor image = this.Image.Value;
                    return image.GetDecodedData();
                }
            }
        }

        public SizeI GetMaskedImageSize(IImageDescriptor image)
        {
            if (this.IsColorKeyMask)
            {
                return new SizeI(image.Width, image.Height);
            }
            else
            {
                IImageDescriptor mask = this.Image.Value;
                return GetMaskedImageSize(mask, image);
            }
        }

        public PixelContainer MaskImage(IImageDescriptor image, PixelContainer pixels)
        {
            if (this.IsColorKeyMask)
            {
                return pixels;
            }
            else
            {
                return MaskImage(this.Image.Value, image, pixels);
            }
        }

        public bool ShouldMaskColorComponents(int[] components)
        {
            if (this.IsColorKeyMask)
            {
                int[] componentRanges = this.ColorMask.Value;

                for (int i = 0; i < componentRanges.Length; i += 2)
                {
                    int component = components[i / 2];

                    if (component < componentRanges[i] || componentRanges[i + 1] < component)
                    {
                        return false;
                    }
                }

                return true;
            }
            else
            {
                return false;
            }
        }

        internal static SizeI GetMaskedImageSize(IImageDescriptor mask, IImageDescriptor image)
        {
            int maskedWidth = Math.Max(image.Width, mask.Width);
            int maskedHeight = Math.Max(image.Height, mask.Height);

            return new SizeI(maskedWidth, maskedHeight);
        }

        internal static PixelContainer MaskImage(IImageDescriptor mask, IImageDescriptor image, PixelContainer pixels)
        {
            ImageMaskingContext context = new ImageMaskingContext();
            context.Image = image;
            context.MaskedSize = GetMaskedImageSize(mask, image);
            context.MaskScale = new Size(mask.Width / (double)context.MaskedSize.Width, mask.Height / (double)context.MaskedSize.Height);
            context.ImageScale = new Size(image.Width / (double)context.MaskedSize.Width, image.Height / (double)context.MaskedSize.Height);
            context.Pixels = pixels;
            context.UnmaskedColor = GetUnmaskedColor(mask);
            context.BitReader = new ImageBitReader(mask.GetDecodedData(), 1, mask.Width);

            // memory optimization for equal size of image and Mask.
            if (context.MaskedSize.Width == image.Width && context.MaskedSize.Height == image.Height)
            {
                return GetMaskImageWhenSameImageSize(context);
            }
            else
            {
                return GetMaskImageWhenDifferentImageSize(context);
            }
        }

        internal static int GetUnmaskedColor(IImageDescriptor image)
        {
            int color = 0;
            double[] decode = image.Decode;

            if (decode != null && decode.Length > 0)
            {
                color = (int)decode[0];
            }

            return color;
        }

        internal static bool IsMasked(ImageBitReader bitReader, int unmaskedColor, int row, int column)
        {
            int bit = bitReader.GetBitAt(row, column);

            return bit != unmaskedColor;
        }

        private static PixelContainer GetMaskImageWhenSameImageSize(ImageMaskingContext c)
        {
            for (int y = 0; y < c.MaskedSize.Height; y++)
            {
                int rowStartIndex = y * c.MaskedSize.Width;

                for (int x = 0; x < c.MaskedSize.Width; x++)
                {
                    int maskColumn = (int)(x * c.MaskScale.Width);
                    int maskRow = (int)(y * c.MaskScale.Height);
                    int maskedImageIndex = rowStartIndex + x;

                    if (IsMasked(c.BitReader, c.UnmaskedColor, maskRow, maskColumn))
                    {
                        c.Pixels.SetColorToIndex(Core.Imaging.Color.Transparent, maskedImageIndex);
                    }
                }
            }

            return c.Pixels;
        }

        private static PixelContainer GetMaskImageWhenDifferentImageSize(ImageMaskingContext c)
        {
            PixelContainer maskedPixels = PixelContainerFactory.GetMaskedImagePixelContainer(c.MaskedSize.Width, c.MaskedSize.Height);


            for (int y = 0; y < c.MaskedSize.Height; y++)
            {
                int rowStartIndex = y * c.MaskedSize.Width;

                for (int x = 0; x < c.MaskedSize.Width; x++)
                {
                    int maskColumn = (int)(x * c.MaskScale.Width);
                    int maskRow = (int)(y * c.MaskScale.Height);
                    int maskedImageIndex = rowStartIndex + x;

                    if (IsMasked(c.BitReader, c.UnmaskedColor, maskRow, maskColumn))
                    {
                        maskedPixels.SetColorToIndex(Core.Imaging.Color.Transparent, maskedImageIndex);
                    }
                    else
                    {
                        int imageIndex = c.Image.GetScaledPixelIndex(x, y, c.ImageScale.Width, c.ImageScale.Height);
                        Core.Imaging.Color color = c.Pixels.GetColorFromIndex(imageIndex);
                        maskedPixels.SetColorToIndex(color, maskedImageIndex);
                    }
                }
            }

            return maskedPixels;
        }

        private class ImageMaskingContext
        {
            public int UnmaskedColor;
            public SizeI MaskedSize;
            public Size ImageScale;
            public Size MaskScale;
            public PixelContainer Pixels;
            public ImageBitReader BitReader;
            public IImageDescriptor Image;
        }
    }
}
