﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
#if !NETSTANDARD
using System.Windows.Media.Imaging;
using Telerik.Windows.Documents.Core.Imaging;
using Telerik.Windows.Documents.Fixed.Model.Extensions;
#endif
using Telerik.Windows.Documents.Core.Imaging.Jpeg2000;
using Telerik.Windows.Documents.Core.Imaging.Jpeg2000.Decoder;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Filters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.ColorSpaces;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Utilities;


namespace Telerik.Windows.Documents.Fixed.Model.Resources
{
    /// <summary>
    /// ImageSource represents a single, constant set of pixels at a certain size.
    /// </summary>
    public partial class ImageSource : IInstanceIdOwner, IImageDescriptor
    {
        private const int OpaqueAlpha = 255;
        internal const int DefaultBitsPerComponent = 8;

        internal static readonly ColorSpaceBase DefaultColorSpace = new DeviceRgb();
        private readonly int id;
        private readonly byte[] unknownImageData;
        private readonly PdfProperty<ImageMask> mask;
        private readonly PdfProperty<ImageSource> sMask;
        private readonly PdfProperty<ColorSpaceBase> colorSpace;
        private readonly PdfProperty<int> bitsPerComponent;
        private readonly PdfProperty<double[]> decodeArray;
        private readonly PdfProperty<bool> imageMask;
        private double[] defaultDecodeArray;

        private int width;
        private int height;
        private string[] filters;
        private byte[] encodedImageData;
        private bool isInitialized;
        private ImageQuality imageQuality;

        /// <summary>
        /// Initializes a new instance of the <see cref="ImageSource" /> class.
        /// </summary>
        /// <param name="stream">The stream.</param>
        public ImageSource(Stream stream)
            : this()
        {
            this.unknownImageData = BytesHelper.GetBytes(stream);
            this.isInitialized = false;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ImageSource" /> class.
        /// </summary>
        /// <param name="stream">The stream.</param>
        /// <param name="imageQuality">The image quality.</param>
        public ImageSource(Stream stream, ImageQuality imageQuality)
            : this()
        {
            byte[] imageData = BytesHelper.GetBytes(stream);
            this.CalculateEncodedDataFromUnknownData(imageData, imageQuality);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ImageSource" /> class.
        /// </summary>
        /// <param name="imageSourceInfo">The image source info.</param>
        public ImageSource(EncodedImageData imageSourceInfo)
            : this()
        {
            Guard.ThrowExceptionIfNull(imageSourceInfo, "imageSourceInfo");

            this.InitializeEncodedData(imageSourceInfo);
        }

#if !NETSTANDARD
        /// <summary>
        /// Initializes a new instance of the <see cref="ImageSource" /> class.
        /// </summary>
        /// <param name="bitmapSource">The bitmap source.</param>
        public ImageSource(BitmapSource bitmapSource)
            : this(ImageSourceExtensions.CreateImageStream(bitmapSource))
        {
            this.InitializeImageInfo(bitmapSource);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ImageSource" /> class.
        /// </summary>
        /// <param name="bitmapSource">The bitmap source.</param>
        /// <param name="imageQuality">The image quality.</param>
        public ImageSource(BitmapSource bitmapSource, ImageQuality imageQuality)
            : this()
        {
            this.imageQuality = imageQuality;
            this.InitializeEncodedData(bitmapSource);
        }
#endif

        internal ImageSource(byte[] encodedImageData, int bitsPerComponent, int width, int height, ColorSpaceBase colorSpace, string[] filters, double[] decodeArray, bool imageMask)
            : this()
        {
            Guard.ThrowExceptionIfNull(encodedImageData, "encodedImageData");

            this.encodedImageData = encodedImageData;
            this.InitializeImageInfo(colorSpace, filters, width, height, bitsPerComponent, decodeArray, imageMask);
        }

        private ImageSource()
        {
            this.id = InstanceIdGenerator.GetNextId();
            this.defaultDecodeArray = null;
            this.mask = new PdfProperty<ImageMask>();
            this.sMask = new PdfProperty<ImageSource>();
            this.colorSpace = new PdfProperty<ColorSpaceBase>();
            this.bitsPerComponent = new PdfProperty<int>();
            this.decodeArray = new PdfProperty<double[]>(() => this.defaultDecodeArray);
            this.imageQuality = FixedDocumentDefaults.ImageQuality;
            this.imageMask = new PdfProperty<bool>();
        }

        /// <summary>
        /// Gets image height.
        /// </summary>
        public int Height
        {
            get
            {
                this.EnsureImageInfo();
                return this.height;
            }
        }

        /// <summary>
        /// Gets image width.
        /// </summary>
        public int Width
        {
            get
            {
                this.EnsureImageInfo();
                return this.width;
            }
        }

        /// <summary>
        /// Gets or sets the decode array which specifies a linear mapping of each component value to a number that would be appropriate as a component value in the color space of the image.
        /// </summary>
        /// <value>The decode array.</value>
        public double[] DecodeArray
        {
            get
            {
                return this.decodeArray.Value;
            }
            set
            {
                this.decodeArray.Value = value;
            }
        }

        internal DecodeParameters[] DecodeParameters
        {
            get;
            set;
        }

        internal string[] Filters
        {
            get
            {
                this.EnsureImageInfo();
                return this.filters;
            }
        }

        internal PdfProperty<int> BitsPerComponent
        {
            get
            {
                this.EnsureImageInfo();
                return this.bitsPerComponent;
            }
        }

        internal PdfProperty<ColorSpaceBase> ColorSpace
        {
            get
            {
                this.EnsureImageInfo();
                return this.colorSpace;
            }
        }

        internal PdfProperty<ImageMask> Mask
        {
            get
            {
                return this.mask;
            }
        }

        internal PdfProperty<ImageSource> SMask
        {
            get
            {
                return this.sMask;
            }
        }

        internal PdfProperty<bool> ImageMask
        {
            get
            {
                return this.imageMask;
            }
        }

        int IInstanceIdOwner.InstanceId
        {
            get
            {
                return this.id;
            }
        }

        int IImageDescriptor.BitsPerComponent
        {
            get
            {
                return this.BitsPerComponent.Value;
            }
        }

        double[] IImageDescriptor.Decode
        {
            get
            {
                return this.DecodeArray;
            }
        }

        IColorSpace IImageDescriptor.ColorSpace
        {
            get
            {
                if (!this.ColorSpace.HasValue && this.ImageMask.HasValue && this.ImageMask.Value)
                {
                    return new DeviceGray();
                }

                return this.ColorSpace.Value;
            }
        }

        IMask IImageDescriptor.Mask
        {
            get
            {
                return this.Mask.Value;
            }
        }

        IImageDescriptor IImageDescriptor.SMask
        {
            get
            {
                return this.SMask.Value;
            }
        }

        string[] IImageDescriptor.Filters
        {
            get
            {
                return this.Filters;
            }
        }

        byte[] IImageDescriptor.GetDecodedData()
        {
            this.EnsureImageInfo();

            ColorSpace colorSpaceValue = FormatProviders.Pdf.Filters.ColorSpace.Undefined;

            if (this.ColorSpace.HasValue)
            {
                ColorSpaceObject colorSpaceObject = ColorSpaceManager.CreateColorSpaceObject(this.ColorSpace.Value);
                colorSpaceValue = colorSpaceObject.Public;
            }

            PdfObject obj = new PdfObject(this.Width, this.Height, colorSpaceValue);
            byte[] data = FiltersManager.Decode(obj, this.encodedImageData, this.Filters, this.DecodeParameters);

            return data;
        }

        byte[] IImageDescriptor.GetEncodedData()
        {
            byte[] data = this.GetEncodedImageData(FixedDocumentDefaults.ImageQuality);

            return data;
        }

        /// <summary>
        /// Gets the encoded image data.
        /// </summary>
        /// <returns>The encoded image data.</returns>
        public EncodedImageData GetEncodedImageData()
        {
            byte[] data = this.GetEncodedImageData(FixedDocumentDefaults.ImageQuality);
            string colorSpace = this.ColorSpace.HasValue ? this.ColorSpace.Value.Name : null;

            return new EncodedImageData(data, this.BitsPerComponent.Value, this.Width, this.Height, colorSpace, this.Filters);
        }

#if !NETSTANDARD
        /// <summary>
        /// Creates <see cref="BitmapSource"/> from the image source instance.
        /// </summary>
        /// <returns>The <see cref="BitmapSource"/> instance.</returns>
        public BitmapSource GetBitmapSource()
        {
            ImageDataSource dataSource = this.ToImageDataSource();

            return dataSource.ToBitmapSource();
        }

        internal BitmapSource GetBitmapSource(ColorBase stencilColor)
        {
            Guard.ThrowExceptionIfNull(stencilColor, "stencilColor");
            Guard.ThrowExceptionIfFalse(this.ImageMask.Value, "ImageMask.Value");

            Color color = stencilColor.ToColor();
            ImageDataSource dataSource = this.ToDecodedImageDataSource(color);

            return dataSource.ToBitmapSource();
        }
#endif

        internal byte[] GetEncodedImageData(ImageQuality imageQuality)
        {
            if (this.unknownImageData != null)
            {
                this.EnsureEncodedDataFromUnknownImageData(imageQuality);
            }

            return this.encodedImageData;
        }

        private static PdfArray CreateFilters(string[] filters)
        {
            PdfArray array = new PdfArray();

            foreach (string filter in filters)
            {
                array.Add(new PdfName(filter));
            }

            return array;
        }

        private static bool CheckIfJpeg2000(byte[] imageData)
        {
            Jpeg2000ImageInfo imageInfo;
            Jpeg2000HeadersDecoder headersDecoder = new Jpeg2000HeadersDecoder(imageData);

            return headersDecoder.TryDecodeJpeg2000Image(out imageInfo);
        }

        private void EnsureEncodedDataFromUnknownImageData(ImageQuality imageQuality)
        {
            bool shouldRecalculateData = this.encodedImageData == null || this.imageQuality != imageQuality;

            if (shouldRecalculateData)
            {
                this.CalculateEncodedDataFromUnknownData(this.unknownImageData, imageQuality);
            }
        }

        private void InitializeEncodedData(EncodedImageData encodedData)
        {
            this.encodedImageData = encodedData.Data;
            this.InitializeImageInfo(encodedData);

            if (encodedData.AlphaChannel == null)
            {
                this.SMask.ClearValue();
            }
            else
            {
                EncodedImageData sMaskData = new EncodedImageData(encodedData.AlphaChannel, encodedData.BitsPerComponent, encodedData.Width, encodedData.Height, ColorSpaceNames.DeviceGray, encodedData.Filters);
                this.SMask.Value = new ImageSource(sMaskData);
            }
        }

#if !NETSTANDARD
        private void InitializeEncodedData(BitmapSource bitmapSource)
        {
            this.InitializeImageInfo(bitmapSource);
            byte[] data, alpha;
            GetData(bitmapSource, out data, out alpha);
            this.encodedImageData = FormatProviders.Pdf.Filters.DctDecode.Encode(this.Width, this.Height, (float)imageQuality, FixedDocumentDefaults.DefaultImageColorspace, data);

            if (alpha != null)
            {
                byte[] encodedAlpha = FormatProviders.Pdf.Filters.DctDecode.Encode(this.Width, this.Height, (float)imageQuality, FixedDocumentDefaults.DefaultSMaskColorspace, alpha);
                ColorSpaceBase colorSpace = new DeviceGray();
                this.SMask.Value = new ImageSource(encodedAlpha, DefaultBitsPerComponent, this.Width, this.Height, colorSpace, FixedDocumentDefaults.DefaultImageFilters, null, false);
            }
            else
            {
                this.SMask.ClearValue();
            }
        }

        private static void GetData(BitmapSource bitmapSource, out byte[] data, out byte[] alpha)
        {
            Guard.ThrowExceptionIfNull(bitmapSource, "bitmapSource");

            int numberOfPixels = bitmapSource.PixelHeight * bitmapSource.PixelWidth;
            data = new byte[numberOfPixels * 3];
            alpha = new byte[numberOfPixels];
            bool shouldExportAlpha = false;
            byte a, r, g, b;
            int pixelIndex = 0;
            foreach (int pixel in ImageSourceExtensions.GetPixels(bitmapSource))
            {
                Core.Imaging.Color.GetComponentsFromPixel(pixel, out a, out r, out g, out b);
                data[3 * pixelIndex] = r;
                data[3 * pixelIndex + 1] = g;
                data[3 * pixelIndex + 2] = b;
                alpha[pixelIndex] = a;
                pixelIndex++;

                if (a != OpaqueAlpha)
                {
                    shouldExportAlpha = true;
                }
            }

            if (!shouldExportAlpha)
            {
                alpha = null;
            }
        }
#endif

        private void EnsureImageInfo()
        {
            if (!this.isInitialized)
            {
                this.InitializeImageInfoFromUnknownData(this.unknownImageData, this.imageQuality);
            }
        }

        private void InitializeImageInfo(EncodedImageData encodedImageData)
        {
            this.InitializeImageInfo(CreateColorSpace(encodedImageData.ColorSpace), encodedImageData.Filters,
                encodedImageData.Width, encodedImageData.Height, encodedImageData.BitsPerComponent, encodedImageData.DecodeArray);
        }

#if !NETSTANDARD
        private void InitializeImageInfo(BitmapSource bitmapSource)
        {
            if (bitmapSource.PixelWidth == 0 || bitmapSource.PixelHeight == 0)
            {
                throw new NotSupportedException("Not supported image format. PixelWidth and PixelHeight must be positive numbers!");
            }

            this.InitializeImageInfo(DefaultColorSpace, FixedDocumentDefaults.DefaultImageFilters,
                bitmapSource.PixelWidth, bitmapSource.PixelHeight, DefaultBitsPerComponent, null);
        }
#endif

        private void InitializeImageInfo(ColorSpaceBase colorSpace, string[] filters, int width, int height, int bitsPerComponent, double[] decodeArray, bool imageMask = false)
        {
            this.filters = filters ?? new string[0];
            this.width = width;
            this.height = height;

            if (bitsPerComponent > 0)
            {
                this.bitsPerComponent.Value = bitsPerComponent;
            }
            else
            {
                this.bitsPerComponent.ClearValue();
            }

            if (colorSpace != null && !imageMask)
            {
                this.colorSpace.Value = colorSpace;
            }
            else
            {
                this.colorSpace.ClearValue();
            }

            this.defaultDecodeArray = decodeArray;
            this.imageMask.Value = imageMask;

            if (imageMask)
            {
                this.bitsPerComponent.Value = 1;
            }

            this.isInitialized = true;
        }

        private void InitializeImageInfoFromUnknownData(byte[] unknownData, ImageQuality imageQuality)
        {
            this.DoOnUnknownData(unknownData, imageQuality, this.InitializeImageInfo
#if NETSTANDARD
                );
#else
                , this.InitializeImageInfo);
#endif
        }

        private void CalculateEncodedDataFromUnknownData(byte[] unknownData, ImageQuality imageQuality)
        {
            this.DoOnUnknownData(unknownData, imageQuality, this.InitializeEncodedData
#if NETSTANDARD
                );
#else
                , this.InitializeEncodedData);
#endif
        }

        private void DoOnUnknownData(byte[] unknownData, ImageQuality imageQuality, Action<EncodedImageData> doOnEncodedData
#if NETSTANDARD
            )
#else
            , Action<BitmapSource> doOnBitmapSource)
#endif
        {
            this.imageQuality = imageQuality;

            EncodedImageData encodedData;
            if (imageQuality == ImageQuality.High && EncodedImageData.TryCreateFromUnknownImageData(unknownData, out encodedData))
            {
                doOnEncodedData(encodedData);
            }
            else if (ImageSource.CheckIfJpeg2000(unknownData))
            {
                throw new NotSupportedException("Cannot export Jpeg2000 image with ImageQuality different than High!");
            }
            else
            {
#if NETSTANDARD
                throw new NotSupportedException("Can not export other than Jpeg and Jpeg2000 and ImageQuality different than High");
#else
                BitmapSource source = ImageSourceExtensions.CreateBitmapSource(new MemoryStream(unknownData));
                doOnBitmapSource(source);
#endif
            }
        }

        private static ColorSpaceBase CreateColorSpace(string name)
        {
            // Colorspace is not required when using JPX filter.
            if (string.IsNullOrEmpty(name))
            {
                return null;
            }

            switch (name)
            {
                case ColorSpaceNames.DeviceRgb:
                    return new DeviceRgb();

                case ColorSpaceNames.DeviceCmyk:
                    return new DeviceCmyk();

                case ColorSpaceNames.DeviceGray:
                    return new DeviceGray();

                default:
                    throw new ArgumentException(string.Format("The color space with name {0} is not supported.", name));
            }
        }
    }
}