﻿using System;
using System.Collections.Generic;
#if !NETSTANDARD
using System.Windows.Media;
#else
using Telerik.Documents.Primitives;
#endif
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Fixed.Model.Text;

namespace Telerik.Windows.Documents.Fixed.Model
{
    internal class TextPropertiesOwner
    {
        private readonly TextCollection textCollection;
        private readonly GeometryPropertiesOwner geometryProperties;
        private RenderingMode renderingMode;

        public TextPropertiesOwner()
        {
            this.textCollection = new TextCollection();
            this.geometryProperties = new GeometryPropertiesOwner();

            this.RenderingMode = FixedDocumentDefaults.TextRenderingMode;
            this.FontSize = FixedDocumentDefaults.FontSize;
            this.Font = FixedDocumentDefaults.Font;
            this.TextMatrix = Matrix.Identity;
        }

        public TextPropertiesOwner(TextPropertiesOwner other)
        {
            this.textCollection = new TextCollection(other.TextCollection);
            this.geometryProperties = new GeometryPropertiesOwner(other.GeometryProperties);

            this.CharacterSpacing = other.CharacterSpacing;
            this.WordSpacing = other.WordSpacing;
            this.HorizontalScaling = other.HorizontalScaling;
            this.TextRise = other.TextRise;
            this.FontSize = other.FontSize;
            this.RenderingMode = other.RenderingMode;
            this.Font = other.Font;
            this.TextMatrix = other.TextMatrix;
        }

        /// <summary>
        /// Gets or sets the character spacing.
        /// </summary>
        /// <value>The character spacing.</value>
        public double? CharacterSpacing { get; set; }

        /// <summary>
        /// Gets or sets the word spacing.
        /// </summary>
        /// <value>The word spacing.</value>
        public double? WordSpacing { get; set; }

        /// <summary>
        /// Gets or sets the horizontal scaling.
        /// </summary>
        /// <value>The horizontal scaling.</value>
        public double? HorizontalScaling { get; set; }

        /// <summary>
        /// Gets or sets the text rise.
        /// </summary>
        /// <value>The text rise.</value>
        public double? TextRise { get; set; }

        /// <summary>
        /// Gets or sets the size of the font.
        /// </summary>
        /// <value>The size of the font.</value>
        public double FontSize { get; set; }

        /// <summary>
        /// Gets or sets the rendering mode.
        /// </summary>
        /// <value>The rendering mode.</value>
        public RenderingMode RenderingMode
        {
            get
            {
                return this.renderingMode;
            }
            set
            {
                this.renderingMode = value;
                this.UpdateIsFilled(value);
                this.UpdateIsStroked(value);
            }
        }

        /// <summary>
        /// Gets or sets the font.
        /// </summary>
        /// <value>The font.</value>
        public FontBase Font
        {
            get
            {
                return this.TextCollection.Font;
            }
            set
            {
                this.TextCollection.Font = value;
            }
        }

        /// <summary>
        /// Gets or sets the fill color.
        /// </summary>
        /// <value>The fill color.</value>
        public ColorBase Fill
        {
            get
            {
                return this.GeometryProperties.Fill;
            }
            set
            {
                this.GeometryProperties.Fill = value;
            }
        }

        /// <summary>
        /// Gets or sets the stroke color.
        /// </summary>
        /// <value>The stroke color.</value>
        public ColorBase Stroke
        {
            get
            {
                return this.GeometryProperties.Stroke;
            }
            set
            {
                this.GeometryProperties.Stroke = value;
            }
        }

        /// <summary>
        /// Gets or sets the width of the stroke line.
        /// </summary>
        /// <value>The width of the stroke line.</value>
        public double StrokeThickness
        {
            get
            {
                return this.GeometryProperties.StrokeThickness;
            }
            set
            {
                this.GeometryProperties.StrokeThickness = value;
            }
        }

        /// <summary>
        /// Gets or sets the stroke line cap.
        /// </summary>
        /// <value>The stroke line cap.</value>
        public LineCap StrokeLineCap
        {
            get
            {
                return this.GeometryProperties.StrokeLineCap;
            }
            set
            {
                this.GeometryProperties.StrokeLineCap = value;
            }
        }

        /// <summary>
        /// Gets or sets the stroke line join.
        /// </summary>
        /// <value>The stroke line join.</value>
        public LineJoin StrokeLineJoin
        {
            get
            {
                return this.GeometryProperties.StrokeLineJoin;
            }
            set
            {
                this.GeometryProperties.StrokeLineJoin = value;
            }
        }

        /// <summary>
        /// Gets or sets the stroke dash array.
        /// </summary>
        /// <value>The stroke dash array.</value>
        public IEnumerable<double> StrokeDashArray
        {
            get
            {
                return this.GeometryProperties.StrokeDashArray;
            }
            set
            {
                this.GeometryProperties.StrokeDashArray = value;
            }
        }

        /// <summary>
        /// Gets or sets the stroke dash offset.
        /// </summary>
        /// <value>The stroke dash offset.</value>
        public double StrokeDashOffset
        {
            get
            {
                return this.GeometryProperties.StrokeDashOffset;
            }
            set
            {
                this.GeometryProperties.StrokeDashOffset = value;
            }
        }

        /// <summary>
        /// Gets or sets the miter limit.
        /// </summary>
        /// <value>The miter limit.</value>
        public double? MiterLimit
        {
            get
            {
                return this.GeometryProperties.MiterLimit;
            }
            set
            {
                this.GeometryProperties.MiterLimit = value;
            }
        }

        internal double? TextTranslation { get; set; }

        internal Matrix TextMatrix
        {
            get;
            set;
        }

        internal GeometryPropertiesOwner GeometryProperties
        {
            get
            {
                return this.geometryProperties;
            }
        }

        internal TextCollection TextCollection
        {
            get
            {
                return this.textCollection;
            }
        }

        private void UpdateIsFilled(RenderingMode renderingMode)
        {
            this.GeometryProperties.IsFilled = renderingMode == RenderingMode.Fill ||
                    renderingMode == RenderingMode.FillAndAddToClippingPath ||
                    renderingMode == RenderingMode.FillAndStroke ||
                    renderingMode == RenderingMode.FillStrokeAndAddToClippingPath;
        }

        private void UpdateIsStroked(RenderingMode renderingMode)
        {
            this.GeometryProperties.IsStroked = renderingMode == RenderingMode.Stroke ||
                       renderingMode == RenderingMode.StrokeAndAddToClippingPath ||
                       renderingMode == RenderingMode.FillAndStroke ||
                       renderingMode == RenderingMode.FillStrokeAndAddToClippingPath;
        }
    }
}
