﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using Telerik.Windows.Documents.Core.Fonts.Type1.Type1Format.Dictionaries;
using Telerik.Windows.Documents.Core.PostScript.Data;
using Telerik.Windows.Documents.Core.PostScript.Operators;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.PostScript
{
    internal class Interpreter
    {
        private readonly Dictionary<string, Type1Font> fonts;
        private readonly PostScriptDictionary systemDict;

        private PostScriptArray rd;
        private PostScriptArray nd;
        private PostScriptArray np;

        public Interpreter()
        {
            this.fonts = new Dictionary<string, Type1Font>();
            this.systemDict = new PostScriptDictionary();

            this.InitializeSystemDict();
        }

        internal OperandsCollection Operands
        {
            get;
            set;
        }

        internal Stack<PostScriptDictionary> DictionaryStack
        {
            get;
            private set;
        }

        internal Stack<PostScriptArray> ArrayStack
        {
            get;
            private set;
        }

        internal PostScriptDictionary CurrentDictionary
        {
            get
            {
                return this.DictionaryStack.Peek();
            }
        }

        internal PostScriptArray CurrentArray
        {
            get
            {
                return this.ArrayStack.Peek();
            }
        }

        internal PostScriptReader Reader
        {
            get;
            private set;
        }

        internal PostScriptDictionary SystemDict
        {
            get
            {
                return this.systemDict;
            }
        }

        internal PostScriptArray RD
        {
            get
            {
                if (this.rd == null)
                {
                    object result;
                    if (this.UserDict.TryGetValue(Operator.RD, out result) ||
                        this.UserDict.TryGetValue(Operator.RDAlternate, out result))
                    {
                        this.rd = (PostScriptArray)result;
                    }
                }

                return this.rd;
            }
            set
            {
                this.rd = value;
            }
        }

        internal PostScriptArray ND
        {
            get
            {
                if (this.nd == null)
                {
                    object result;
                    if (this.UserDict.TryGetValue(Operator.ND, out result) ||
                        this.UserDict.TryGetValue(Operator.NDAlternate, out result))
                    {
                        this.nd = (PostScriptArray)result;
                    }
                }

                return this.nd;
            }
            set
            {
                this.nd = value;
            }
        }

        internal PostScriptArray NP
        {
            get
            {
                if (this.np == null)
                {
                    object result;
                    if (this.UserDict.TryGetValue(Operator.NP, out result) ||
                        this.UserDict.TryGetValue(Operator.NPAlternate, out result))
                    {
                        this.np = (PostScriptArray)result;
                    }
                }

                return this.np;
            }
            set
            {
                this.np = value;
            }
        }

        internal PostScriptDictionary UserDict
        {
            get;
            private set;
        }

        public Dictionary<string, Type1Font> Fonts
        {
            get
            {
                return this.fonts;
            }
        }

        public void Execute(byte[] data)
        {
            this.Operands = new OperandsCollection();
            this.DictionaryStack = new Stack<PostScriptDictionary>();
            this.ArrayStack = new Stack<PostScriptArray>();
            this.UserDict = new PostScriptDictionary();

            this.Reader = new PostScriptReader(data);
            while (!this.Reader.EndOfFile)
            {
                Token token = this.Reader.ReadToken();
                if (token == Token.Operator)
                {
                    Operator op = Operator.FindOperator(this.Reader.Result);
                    op.Execute(this);
                }
                else if (token != Token.Unknown)
                {
                    this.Operands.AddLast(ParseOperand(token, this.Reader));
                }
            }
        }

        internal void ExecuteProcedure(PostScriptArray proc)
        {
            Guard.ThrowExceptionIfNull(proc, "proc");

            foreach (object o in proc)
            {
                Operator op = o as Operator;
                if (op != null)
                {
                    op.Execute(this);
                }
                else
                {
                    this.Operands.AddLast(o);
                }
            }
        }

        private void InitializeSystemDict()
        {
            this.systemDict[Names.FontDirectory] = new PostScriptDictionary();
        }

        private static object ParseOperand(Token token, PostScriptReader reader)
        {
            switch (token)
            {
                case Token.Integer:
                    return ParseInt(reader.Result);
                case Token.Real:
                    return ParseReal(reader.Result);
                case Token.ArrayStart:
                    return ParseArray(reader);
                case Token.Boolean:
                    return ParseBool(reader.Result);
                case Token.DictionaryStart:
                    return ParseDictionary(reader);
                case Token.String:
                    return new PostScriptString(reader.Result);
                case Token.Name:
                    return reader.Result;
                case Token.Keyword:
                    return Keywords.GetValue(reader.Result);
                case Token.Operator:
                    return Operator.FindOperator(reader.Result);
            }
            return null;
        }

        private static object ParseBool(string b)
        {
            return b == Keywords.True;
        }

        private static int ParseInt(string str)
        {
            return int.Parse(str, CultureInfo.InvariantCulture.NumberFormat);
        }

        private static double ParseReal(string str)
        {
            return double.Parse(str, CultureInfo.InvariantCulture.NumberFormat);
        }

        private static PostScriptArray ParseArray(PostScriptReader reader)
        {
            Token tok;
            PostScriptArray arr = new PostScriptArray();
            while (!reader.EndOfFile && (tok = reader.ReadToken()) != Token.ArrayEnd)
            {
                object res = ParseOperand(tok, reader);

                if (tok == Token.Operator && res == null)
                {
                    continue;
                }

                arr.Add(res);
            }

            return arr;
        }

        private static PostScriptDictionary ParseDictionary(PostScriptReader reader)
        {
            Token tok;
            PostScriptDictionary dict = new PostScriptDictionary();
            while (!reader.EndOfFile && (tok = reader.ReadToken()) != Token.ArrayEnd)
            {
                string key;

                switch (tok)
                {
                    case Token.String:
                    case Token.Name:
                        key = reader.Result;
                        break;
                    default:
                        reader.ReadToken();
                        continue;
                }

                dict[key] = ParseOperand(reader.ReadToken(), reader);
            }

            return dict;
        }
    }
}
