﻿using System;
using System.IO;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser.Data
{
    internal class StreamPart : Stream
    {
        private readonly Stream baseStream;
        private readonly long offset, length;
        private readonly byte[] buffer;

        public StreamPart(Stream baseStream, long offset, string endToken)
        {
            Guard.ThrowExceptionIfLessThan(0, offset, "offset");

            this.baseStream = baseStream;
            this.offset = offset;
            this.buffer = new byte[Constants.BufferSize];
            this.length = this.CalculateStreamSize(endToken);
            this.baseStream.Seek(this.offset, SeekOrigin.Begin);
        }

        public override bool CanRead
        {
            get { return this.baseStream.CanRead; }
        }

        public override bool CanSeek
        {
            get { return this.baseStream.CanSeek; }
        }

        public override bool CanWrite
        {
            get { return this.baseStream.CanWrite; }
        }

        public override long Length
        {
            get { return this.length; }
        }

        public override long Position
        {
            get
            {
                return this.baseStream.Position - this.offset;
            }
            set
            {
                this.baseStream.Position = value + this.offset;
            }
        }

        public override void Flush()
        {
            this.baseStream.Flush();
        }

        public override int Read(byte[] buffer, int offset, int count)
        {
            int bytesToRead = (int)Math.Min(count, this.Length - this.Position);
            if (bytesToRead == 0)
            {
                return 0;
            }

            return this.baseStream.Read(buffer, offset, bytesToRead);
        }

        public override long Seek(long offset, SeekOrigin origin)
        {
            switch (origin)
            {
                case SeekOrigin.Begin:
                    Guard.ThrowExceptionIfOutOfRange(0, this.offset + this.length, offset, "offset");

                    return baseStream.Seek(this.offset + offset, SeekOrigin.Begin) - this.offset;

                default:
                    throw new NotSupportedException();
            }
        }

        public override void SetLength(long value)
        {
            throw new NotSupportedException();
        }

        public override void Write(byte[] buffer, int offset, int count)
        {
            throw new NotSupportedException();
        }

        private long CalculateStreamSize(string token)
        {
            this.baseStream.Seek(this.offset, SeekOrigin.Begin);
            StringBuilder tmp = new StringBuilder();
            StringBuilder tok = new StringBuilder(token);
            StringBuilder lineBreak = new StringBuilder();
            int len;
            long pos = this.offset;
            bool hasMach = false;
            while (pos < this.baseStream.Length)
            {
                len = this.baseStream.Read(this.buffer, 0, Constants.BufferSize);
                for (int i = 0; i < len; i++)
                {
                    if (hasMach && Characters.IsDelimiter(this.buffer[i]))
                    {
                        break;
                    }

                    tmp.Append((char)this.buffer[i]);
                    pos++;
                    if (tok.Length < tmp.Length)
                    {
                        char ch = tmp[0];
                        tmp.Remove(0, 1);
                        lineBreak.Append(ch);
                        if (lineBreak.Length > 2)
                        {
                            lineBreak.Remove(0, 1);
                        }
                    }

                    hasMach = tmp.Equals(tok);
                }

                if (hasMach)
                {
                    break;
                }
            }

            if (hasMach)
            {
                long length = pos - offset - tmp.Length;

                if (lineBreak.Length > 0)
                {
                    string lb = lineBreak.ToString();
                    if (lb.First() == Characters.CarriageReturn)
                    {
                        length--;
                    }

                    if (lb.Last() == Characters.LineFeed)
                    {
                        length--;
                    }
                }

                return length;
            }

            return -1;
        }
    }
}
