﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Utilities;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Operators;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Operators.InlineImage;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser.Keywords
{
    internal class ImageDataKeyword : Keyword
    {
        private static readonly byte[] endStreamKeywordBytes;

        static ImageDataKeyword()
        {
            endStreamKeywordBytes = Encoding.UTF8.GetBytes(ContentStreamOperatorNames.EndInlineImage);
        }

        public override string Name
        {
            get
            {
                return ContentStreamOperatorNames.ImageData;
            }
        }

        public override void Complete(PostScriptReader reader, IPdfImportContext context)
        {
            // Read PdfStream to "EI". Push PdfStream operand and EI operator.

            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            // if the ColorSpace is valid push the image, otherwise skip it.
            bool skipImage = false;

            long startOffset = reader.Reader.Stream.Position;
            PdfInt length = null;
            PdfDictionary dictionary = new PdfDictionary();

            IPdfPrimitive token = reader.PopToken();

            while (!(token is BeginInlineImage))
            {
                IPdfPrimitive value = token;

                IPdfPrimitive keyPrimitive = reader.PopToken();
                PdfName originalKey = (PdfName)keyPrimitive;
                PdfName key = new PdfName(PdfNames.GetNameFromAbbreviation(originalKey.Value));

                if (key.Value.Equals(PdfNames.ColorSpace))
                {
                    PdfName colorSpaceName = value as PdfName;
                    if (colorSpaceName == null)
                    {
                        // TODO: Support for IndexColorSpace
                        skipImage = true;
                    }
                    else
                    {
                        value = new PdfName(PdfNames.GetNameFromAbbreviation(colorSpaceName.Value));
                        dictionary[key.Value] = value;
                    }
                }
                else
                {
                    dictionary[key.Value] = value;
                }

                token = reader.PopToken();
            }

            reader.Reader.Seek(startOffset, SeekOrigin.Begin);
            int calculatedLength = CalculateStreamLength(reader.Reader);
            length = new PdfInt(calculatedLength);

            if (!skipImage)
            {
                reader.PushToken(new InlineImagePdfStream(context, reader.Reader.Stream, dictionary, length, startOffset));
                reader.PushToken(new EndInlineImage());
            }
        }

        public static int CalculateStreamLength(Reader reader)
        {
            long startOffset = reader.Position;
            long lastOffset = FindPositionBeforePdfStreamEndKeyword(reader);
            int calculatedLength = (int)(lastOffset - startOffset + 1);

            return calculatedLength;
        }

        private static long FindPositionBeforePdfStreamEndKeyword(Reader reader)
        {
            long keywordStartIndex = reader.IndexOf(endStreamKeywordBytes);

            return keywordStartIndex - 1;
        }
    }
}
