﻿using System;
using System.IO;
using System.Linq;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser.Keywords
{
    internal class StreamStartKeyword : Keyword
    {
        internal static readonly byte[] EndStreamKeywordBytes;

        static StreamStartKeyword()
        {
            EndStreamKeywordBytes = System.Text.Encoding.UTF8.GetBytes(PdfNames.PdfStreamEnd);
        }

        public override string Name
        {
            get
            {
                return PdfNames.PdfStreamStart;
            }
        }

        public override void Complete(PostScriptReader reader, IPdfImportContext context)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            //// This handles missing carriage return symbol after stream keyword.
            reader.Reader.Seek(-1, SeekOrigin.Current);

            reader.Reader.ReadLine();
            long startOffset = reader.Reader.Stream.Position;

            PdfInt length = null;
            PdfDictionary dictionary = (PdfDictionary)reader.PopToken();
            dictionary.TryGetElement(reader, context, PdfNames.Length, out length);
            bool isCorrectLegth = false;

            if (length != null)
            {
                reader.Reader.Seek(length.Value, SeekOrigin.Current);
                isCorrectLegth = TryFindPdfStreamEndOnNextRow(reader.Reader);
            }

            if (!isCorrectLegth)
            {
                reader.Reader.Seek(startOffset, SeekOrigin.Begin);
                int calculatedLength = CalculateStreamLength(reader.Reader);
                length = new PdfInt(calculatedLength);
            }
            
            reader.PushToken(new PdfStream(context, reader.Reader.Stream, dictionary, length, startOffset));
        }

        public static bool TryFindPdfStreamEndOnNextRow(Reader reader)
        {
            long initialPosition = reader.Position;
            long keywordIndex = reader.IndexOf(EndStreamKeywordBytes);

            if(keywordIndex > -1)
            {
                using (reader.BeginReadingBlock())
                {
                    reader.Seek(initialPosition, SeekOrigin.Begin);
                    int distance = (int)(keywordIndex - initialPosition);
                    byte[] bytesBetween = reader.Read(distance);

                    foreach (byte b in bytesBetween)
                    {
                        if (!(Characters.IsCarriageReturn(b) || Characters.IsLineFeed(b)))
                        {
                            return false;
                        }
                    }

                    return true;
                }                    
            }

            return false;
        }

        public static int CalculateStreamLength(Reader reader)
        {
            long startOffset = reader.Position;
            long lastOffset = FindPositionBeforePdfStreamEndKeyword(reader);
            int calculatedLength = (int)(lastOffset - startOffset + 1);

            return calculatedLength;
        }

        private static long FindPositionBeforePdfStreamEndKeyword(Reader reader)
        {
            long keywordStartIndex = reader.IndexOf(EndStreamKeywordBytes);

            if (keywordStartIndex > -1)
            {
                long positionBeforeEndStreamKeyword;
                long positionAfterEndStreamKeyword = reader.Position;
                reader.Seek(keywordStartIndex, SeekOrigin.Begin);
                byte currentByte = reader.Peek();

                while (!Characters.IsCarriageReturn(currentByte) && !Characters.IsLineFeed(currentByte))
                {
                    reader.Seek(-1, SeekOrigin.Current);
                    currentByte = reader.Peek();
                }

                if (Characters.IsLineFeed(currentByte))
                {
                    reader.Seek(-1, SeekOrigin.Current);
                    currentByte = reader.Peek();
                    positionBeforeEndStreamKeyword = Characters.IsCarriageReturn(currentByte) ? reader.Position - 1 : reader.Position;
                }
                else
                {
                    positionBeforeEndStreamKeyword = reader.Position - 1;
                }

                reader.Seek(positionAfterEndStreamKeyword, SeekOrigin.Begin);

                return positionBeforeEndStreamKeyword;
            }

            throw new InvalidOperationException("Cannot find endstream keyword!");
        }
    }
}
