﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Telerik.Windows.Documents.Data.StateMachine;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser.Keywords;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser.PrimitiveParsers;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser
{
    internal class PostScriptReader
    {
        private const string InitialState = "Initial";

        private readonly Reader reader;
        private readonly StateMachine<PostScriptReaderArgs> stateMachine;
        private readonly Stack<Stack<IPdfPrimitive>> collectionValuesStack;

        private readonly NumberParser numberParser;
        private readonly HexStringParser hexStringParser;
        private readonly LiteralStringParser literalStringParser;
        private readonly ArrayParser arrayParser;
        private readonly DictionaryParser dictionaryParser;
        private readonly NameParser nameParser;
        private readonly KeywordParser keywordParser;

        private PdfElementType? lastTokenType;
        private byte? sign;

        public PostScriptReader(Stream stream, KeywordCollection keywords)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");

            this.collectionValuesStack = new Stack<Stack<IPdfPrimitive>>();
            this.reader = new Reader(stream);

            this.numberParser = new NumberParser(this);
            this.hexStringParser = new HexStringParser(this);
            this.literalStringParser = new LiteralStringParser(this);
            this.arrayParser = new ArrayParser(this);
            this.dictionaryParser = new DictionaryParser(this);
            this.nameParser = new NameParser(this);
            this.keywordParser = new KeywordParser(this, keywords);

            this.stateMachine = this.BuildStateMachine();
        }

        public Reader Reader
        {
            get
            {
                return this.reader;
            }
        }

        public IPdfPrimitive[] Read(IPdfImportContext context)
        {
            return this.Read(context, this.Reader.Length);
        }

        public IPdfPrimitive[] Read(IPdfImportContext context, long endPosition)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            this.PushCollection();
            PostScriptReaderArgs readerArgs = new PostScriptReaderArgs(context, this.reader.Read());
            this.stateMachine.Start(this.stateMachine.States.GetStateByName(InitialState), readerArgs);

            while (this.Reader.Position < endPosition)
            {
                readerArgs = new PostScriptReaderArgs(context, this.reader.Read());
                this.stateMachine.GoToNextState(readerArgs);
            }

            this.stateMachine.Stop(new PostScriptReaderArgs(context, 0));

            Stack<IPdfPrimitive> poppedCollection = this.PopCollection();
            IPdfPrimitive[] result = poppedCollection.Reverse().ToArray();

            return result;
        }

        public T Read<T>(IPdfImportContext context, PdfElementType tokenType)
            where T : IPdfPrimitive
        {
            Guard.ThrowExceptionIfNull(context, "context");

            this.PushCollection();

            this.lastTokenType = null;
            bool hasReachedTokenType = false;
            int initialStackCount = this.collectionValuesStack.Count;
            PostScriptReaderArgs readerArgs = new PostScriptReaderArgs(context, this.reader.Read());
            this.stateMachine.Start(this.stateMachine.States.GetStateByName(InitialState), readerArgs);

            this.reader.CacheLength();

            while (!this.reader.EndOfFile && !hasReachedTokenType)
            {
                readerArgs = new PostScriptReaderArgs(context, this.reader.Read());
                this.stateMachine.GoToNextState(readerArgs);
                hasReachedTokenType = this.lastTokenType == tokenType && this.collectionValuesStack.Count == initialStackCount;

                if (this.reader.EndOfFile && !hasReachedTokenType)
                {
                    // Set initial state when end-of-file is reached with no token state transition
                    this.stateMachine.GoToState(this.stateMachine.States.GetStateByName(InitialState), readerArgs);
                }
            }

            this.reader.ClearCachedLength();

            Stack<IPdfPrimitive> collection = this.PopCollection();
            T token = (T)collection.Pop();

            return token;
        }

        public void ClearCollections()
        {
            this.collectionValuesStack.Clear();
        }

        public Stack<IPdfPrimitive> PeekCollection()
        {
            return this.collectionValuesStack.Peek();
        }

        public void PushCollection()
        {
            this.collectionValuesStack.Push(new Stack<IPdfPrimitive>());
        }

        public Stack<IPdfPrimitive> PopCollection()
        {
            return this.collectionValuesStack.Pop();
        }

        public void PushToken(IPdfPrimitive primitive)
        {
            if (primitive != null)
            {
                this.lastTokenType = primitive.Type;
            }

            this.PeekCollection().Push(primitive);
        }

        public IPdfPrimitive PopToken()
        {
            if (this.PeekCollection().Count > 0)
            {
                return this.PeekCollection().Pop();
            }

            return null;
        }

        private StateMachine<PostScriptReaderArgs> BuildStateMachine()
        {
            StateMachine<PostScriptReaderArgs> res = new StateMachine<PostScriptReaderArgs>();

            State<PostScriptReaderArgs> initial = new State<PostScriptReaderArgs>(InitialState);
            State<PostScriptReaderArgs> sign = new State<PostScriptReaderArgs>(
                "Sign",
                 onEnterAction: (b) =>
                    {
                        this.sign = b.Byte;
                    },
                onStayAction: (b) =>
                {
                    this.sign = b.Byte;
                },
                onLeaveAction: (b) =>
                    {
                        if (!Characters.IsNumberCharacter(b.Byte))
                        {
                            this.sign = null;
                        }
                    }

            );
            State<PostScriptReaderArgs> number =
                new State<PostScriptReaderArgs>(
                    "Number",
                    onEnterAction: (b) =>
                    {
                        if (this.sign.HasValue)
                        {
                            this.numberParser.Start((char)this.sign.Value);
                            this.sign = null;
                        }
                        this.numberParser.Append(b);
                    },
                    onStayAction: this.numberParser.Append,
                    onLeaveAction: (b) => { this.numberParser.Complete(); },
                    onStopAction: (b) => { this.numberParser.Complete(); });

            State<PostScriptReaderArgs> hexStringOrDictionaryStart = new State<PostScriptReaderArgs>("HexStringOrDictionaryStart");
            State<PostScriptReaderArgs> literalString = new State<PostScriptReaderArgs>("LiteralString", onStayAction: this.literalStringParser.Append, onLeaveAction: this.literalStringParser.Complete);
            State<PostScriptReaderArgs> arrayStart = new State<PostScriptReaderArgs>("ArrayStart", onEnterAction: (b) => { this.arrayParser.Start(); }, onStayAction: (b) => { this.arrayParser.Start(); });
            State<PostScriptReaderArgs> arrayEnd = new State<PostScriptReaderArgs>("ArrayEnd", onEnterAction: (b) => { this.arrayParser.End(); }, onStayAction: (b) => { this.arrayParser.End(); });
            State<PostScriptReaderArgs> dictionaryStart = new State<PostScriptReaderArgs>("DictionaryStart", onEnterAction: (b) => { this.dictionaryParser.Start(); });
            State<PostScriptReaderArgs> dictionaryEndStart = new State<PostScriptReaderArgs>("DictionaryEndMiddle");
            State<PostScriptReaderArgs> dictionaryEnd = new State<PostScriptReaderArgs>("DictionaryEnd", onEnterAction: (b) => { this.dictionaryParser.End(); });

            Action<PostScriptReaderArgs> nameStartLeaveAndStayAction = (b) =>
            {
                if (!Characters.IsValidNameCharacter(b.Byte))
                {
                    this.nameParser.Complete();
                }
            };

            State<PostScriptReaderArgs> nameStart = new State<PostScriptReaderArgs>("NameStart", onLeaveAction: nameStartLeaveAndStayAction, onStayAction: nameStartLeaveAndStayAction);

            State<PostScriptReaderArgs> name = new State<PostScriptReaderArgs>("Name", onEnterAction: this.nameParser.Append, onStayAction: this.nameParser.Append, onLeaveAction: (b) => { this.nameParser.Complete(); }, onStopAction: (b) => { this.nameParser.Complete(); });
            State<PostScriptReaderArgs> keyword = new State<PostScriptReaderArgs>("Keyword", onEnterAction: this.keywordParser.Append, onStayAction: this.keywordParser.Append, onLeaveAction: this.keywordParser.Complete, onStopAction: this.keywordParser.Complete);
            State<PostScriptReaderArgs> hexString = new State<PostScriptReaderArgs>("HexString", onEnterAction: this.hexStringParser.Append, onStayAction: this.hexStringParser.Append, onLeaveAction: this.hexStringParser.Complete);
            State<PostScriptReaderArgs> comment = new State<PostScriptReaderArgs>("Comment");
            State<PostScriptReaderArgs> emptyHexStringEnd = new State<PostScriptReaderArgs>("EmptyHexString", onEnterAction: this.hexStringParser.Complete);

            res.States.Add(initial);
            res.States.Add(sign);
            res.States.Add(number);
            res.States.Add(hexStringOrDictionaryStart);
            res.States.Add(literalString);
            res.States.Add(arrayStart);
            res.States.Add(arrayEnd);
            res.States.Add(dictionaryStart);
            res.States.Add(dictionaryEndStart);
            res.States.Add(dictionaryEnd);
            res.States.Add(nameStart);
            res.States.Add(name);
            res.States.Add(keyword);
            res.States.Add(hexString);
            res.States.Add(comment);
            res.States.Add(emptyHexStringEnd);

            Transition<PostScriptReaderArgs> initialToNumber = new Transition<PostScriptReaderArgs>(initial, number, (args) => { return Characters.IsNumberCharacter(args.Byte); });
            Transition<PostScriptReaderArgs> initialToSign = new Transition<PostScriptReaderArgs>(initial, sign, (args) => { return Characters.IsSign(args.Byte); });
            Transition<PostScriptReaderArgs> initialToHexStringOrDictionaryStart = new Transition<PostScriptReaderArgs>(initial, hexStringOrDictionaryStart, (args) => { return Characters.IsHexStringOrDictionaryStart(args.Byte); });
            Transition<PostScriptReaderArgs> initialToLiteralString = new Transition<PostScriptReaderArgs>(initial, literalString, (args) => { return Characters.IsLiteralStringStart(args.Byte); });
            Transition<PostScriptReaderArgs> initialToArrayStart = new Transition<PostScriptReaderArgs>(initial, arrayStart, (args) => { return Characters.IsArrayStart(args.Byte); });
            Transition<PostScriptReaderArgs> initialToArrayEnd = new Transition<PostScriptReaderArgs>(initial, arrayEnd, (args) => { return Characters.IsArrayEnd(args.Byte); });
            Transition<PostScriptReaderArgs> initialToDictionaryEndStart = new Transition<PostScriptReaderArgs>(initial, dictionaryEndStart, (args) => { return Characters.IsHexStringOrDictionaryEnd(args.Byte); });
            Transition<PostScriptReaderArgs> initialToNameStart = new Transition<PostScriptReaderArgs>(initial, nameStart, (args) => { return Characters.IsNameStart(args.Byte); });
            Transition<PostScriptReaderArgs> initialToKeyword = new Transition<PostScriptReaderArgs>(initial, keyword, (args) => { return Characters.IsKeywordStartCharacter(args.Byte); });
            Transition<PostScriptReaderArgs> initialToComment = new Transition<PostScriptReaderArgs>(initial, comment, (args) => { return Characters.IsCommentStart(args.Byte); });

            res.Transitions.AddTransition(initialToNumber);
            res.Transitions.AddTransition(initialToSign);
            res.Transitions.AddTransition(initialToHexStringOrDictionaryStart);
            res.Transitions.AddTransition(initialToLiteralString);
            res.Transitions.AddTransition(initialToArrayStart);
            res.Transitions.AddTransition(initialToArrayEnd);
            res.Transitions.AddTransition(initialToDictionaryEndStart);
            res.Transitions.AddTransition(initialToNameStart);
            res.Transitions.AddTransition(initialToKeyword);
            res.Transitions.AddTransition(initialToComment);

            Transition<PostScriptReaderArgs> signToInitial = new Transition<PostScriptReaderArgs>(sign, initial, (args) => { return !Characters.IsNumberCharacter(args.Byte) && !Characters.IsSign(args.Byte); });
            Transition<PostScriptReaderArgs> signToSign = new Transition<PostScriptReaderArgs>(sign, (args) => { return Characters.IsSign(args.Byte); });
            Transition<PostScriptReaderArgs> signToNumber = new Transition<PostScriptReaderArgs>(sign, number, (args) => { return Characters.IsNumberCharacter(args.Byte); });

            res.Transitions.AddTransition(signToNumber);
            res.Transitions.AddTransition(signToSign);
            res.Transitions.AddTransition(signToInitial);

            Transition<PostScriptReaderArgs> numberToHexStringOrDictionaryStart = new Transition<PostScriptReaderArgs>(number, hexStringOrDictionaryStart, (args) => { return Characters.IsHexStringOrDictionaryStart(args.Byte); });
            Transition<PostScriptReaderArgs> numberToLiteralString = new Transition<PostScriptReaderArgs>(number, literalString, (args) => { return Characters.IsLiteralStringStart(args.Byte); });
            Transition<PostScriptReaderArgs> numberToArrayStart = new Transition<PostScriptReaderArgs>(number, arrayStart, (args) => { return Characters.IsArrayStart(args.Byte); });
            Transition<PostScriptReaderArgs> numberToArrayEnd = new Transition<PostScriptReaderArgs>(number, arrayEnd, (args) => { return Characters.IsArrayEnd(args.Byte); });
            Transition<PostScriptReaderArgs> numberToDictionaryEndStart = new Transition<PostScriptReaderArgs>(number, dictionaryEndStart, (args) => { return Characters.IsHexStringOrDictionaryEnd(args.Byte); });
            Transition<PostScriptReaderArgs> numberToNameStart = new Transition<PostScriptReaderArgs>(number, nameStart, (args) => { return Characters.IsNameStart(args.Byte); });
            Transition<PostScriptReaderArgs> numberToNumber = new Transition<PostScriptReaderArgs>(number, this.numberParser.IsNumber);
            Transition<PostScriptReaderArgs> numberToComment = new Transition<PostScriptReaderArgs>(number, comment, (args) => { return Characters.IsCommentStart(args.Byte); });
            Transition<PostScriptReaderArgs> numberToInitial = new Transition<PostScriptReaderArgs>(number, initial, (args) => { return !this.numberParser.IsNumber(args); });

            res.Transitions.AddTransition(numberToHexStringOrDictionaryStart);
            res.Transitions.AddTransition(numberToLiteralString);
            res.Transitions.AddTransition(numberToArrayStart);
            res.Transitions.AddTransition(numberToArrayEnd);
            res.Transitions.AddTransition(numberToDictionaryEndStart);
            res.Transitions.AddTransition(numberToNameStart);
            res.Transitions.AddTransition(numberToNumber);
            res.Transitions.AddTransition(numberToComment);
            res.Transitions.AddTransition(numberToInitial);

            Transition<PostScriptReaderArgs> hexStringOrDictionaryStartToDictionaryStart = new Transition<PostScriptReaderArgs>(hexStringOrDictionaryStart, dictionaryStart, (args) => { return Characters.IsHexStringOrDictionaryStart(args.Byte); });
            Transition<PostScriptReaderArgs> hexStringOrDictionaryStartToEmptyHexString = new Transition<PostScriptReaderArgs>(hexStringOrDictionaryStart, emptyHexStringEnd, (args) => { return Characters.IsHexStringOrDictionaryEnd(args.Byte); });
            Transition<PostScriptReaderArgs> hexStringOrDictionaryStartToHexString = new Transition<PostScriptReaderArgs>(hexStringOrDictionaryStart, hexString, (args) => { return true; });

            res.Transitions.AddTransition(hexStringOrDictionaryStartToDictionaryStart);
            res.Transitions.AddTransition(hexStringOrDictionaryStartToEmptyHexString);
            res.Transitions.AddTransition(hexStringOrDictionaryStartToHexString);

            Transition<PostScriptReaderArgs> emptyHexStringEndToNumber = new Transition<PostScriptReaderArgs>(emptyHexStringEnd, number, (args) => { return Characters.IsNumberCharacter(args.Byte); });
            Transition<PostScriptReaderArgs> emptyHexStringEndToHexStringOrDictionaryStart = new Transition<PostScriptReaderArgs>(emptyHexStringEnd, hexStringOrDictionaryStart, (args) => { return Characters.IsHexStringOrDictionaryStart(args.Byte); });
            Transition<PostScriptReaderArgs> emptyHexStringEndToLiteralString = new Transition<PostScriptReaderArgs>(emptyHexStringEnd, literalString, (args) => { return Characters.IsLiteralStringStart(args.Byte); });
            Transition<PostScriptReaderArgs> emptyHexStringEndToArrayStart = new Transition<PostScriptReaderArgs>(emptyHexStringEnd, arrayStart, (args) => { return Characters.IsArrayStart(args.Byte); });
            Transition<PostScriptReaderArgs> emptyHexStringEndToArrayEnd = new Transition<PostScriptReaderArgs>(emptyHexStringEnd, arrayEnd, (args) => { return Characters.IsArrayEnd(args.Byte); });
            Transition<PostScriptReaderArgs> emptyHexStringEndToDictionaryEndStart = new Transition<PostScriptReaderArgs>(emptyHexStringEnd, dictionaryEndStart, (args) => { return Characters.IsHexStringOrDictionaryEnd(args.Byte); });
            Transition<PostScriptReaderArgs> emptyHexStringEndToNameStart = new Transition<PostScriptReaderArgs>(emptyHexStringEnd, nameStart, (args) => { return Characters.IsNameStart(args.Byte); });
            Transition<PostScriptReaderArgs> emptyHexStringEndToKeyword = new Transition<PostScriptReaderArgs>(emptyHexStringEnd, keyword, (args) => { return Characters.IsKeywordStartCharacter(args.Byte); });
            Transition<PostScriptReaderArgs> emptyHexStringEndToComment = new Transition<PostScriptReaderArgs>(emptyHexStringEnd, comment, (args) => { return Characters.IsCommentStart(args.Byte); });

            res.Transitions.AddTransition(emptyHexStringEndToNumber);
            res.Transitions.AddTransition(emptyHexStringEndToHexStringOrDictionaryStart);
            res.Transitions.AddTransition(emptyHexStringEndToLiteralString);
            res.Transitions.AddTransition(emptyHexStringEndToArrayStart);
            res.Transitions.AddTransition(emptyHexStringEndToArrayEnd);
            res.Transitions.AddTransition(emptyHexStringEndToDictionaryEndStart);
            res.Transitions.AddTransition(emptyHexStringEndToNameStart);
            res.Transitions.AddTransition(emptyHexStringEndToKeyword);
            res.Transitions.AddTransition(emptyHexStringEndToComment);

            Transition<PostScriptReaderArgs> literalStringToInitial = new Transition<PostScriptReaderArgs>(literalString, initial, this.literalStringParser.IsLiteralStringEnd);
            Transition<PostScriptReaderArgs> literalStringToLiteralString = new Transition<PostScriptReaderArgs>(literalString, (b) => { return !this.literalStringParser.IsLiteralStringEnd(b); });

            res.Transitions.AddTransition(literalStringToInitial);
            res.Transitions.AddTransition(literalStringToLiteralString);

            Transition<PostScriptReaderArgs> arrayStartToSign = new Transition<PostScriptReaderArgs>(arrayStart, sign, (args) => { return Characters.IsSign(args.Byte); });
            Transition<PostScriptReaderArgs> arrayStartToNumber = new Transition<PostScriptReaderArgs>(arrayStart, number, (args) => { return Characters.IsNumberCharacter(args.Byte); });
            Transition<PostScriptReaderArgs> arrayStartToHexStringOrDictionaryStart = new Transition<PostScriptReaderArgs>(arrayStart, hexStringOrDictionaryStart, (args) => { return Characters.IsHexStringOrDictionaryStart(args.Byte); });
            Transition<PostScriptReaderArgs> arrayStartToLiteralString = new Transition<PostScriptReaderArgs>(arrayStart, literalString, (args) => { return Characters.IsLiteralStringStart(args.Byte); });
            Transition<PostScriptReaderArgs> arrayStartToArrayStart = new Transition<PostScriptReaderArgs>(arrayStart, (args) => { return Characters.IsArrayStart(args.Byte); });
            Transition<PostScriptReaderArgs> arrayStartToArrayEnd = new Transition<PostScriptReaderArgs>(arrayStart, arrayEnd, (args) => { return Characters.IsArrayEnd(args.Byte); });
            Transition<PostScriptReaderArgs> arrayStartToNameStart = new Transition<PostScriptReaderArgs>(arrayStart, nameStart, (args) => { return Characters.IsNameStart(args.Byte); });
            Transition<PostScriptReaderArgs> arrayStartToKeyword = new Transition<PostScriptReaderArgs>(arrayStart, keyword, (args) => { return Characters.IsKeywordStartCharacter(args.Byte); });
            Transition<PostScriptReaderArgs> arrayStartToComment = new Transition<PostScriptReaderArgs>(arrayStart, comment, (args) => { return Characters.IsCommentStart(args.Byte); });

            res.Transitions.AddTransition(arrayStartToNumber);
            res.Transitions.AddTransition(arrayStartToSign);
            res.Transitions.AddTransition(arrayStartToHexStringOrDictionaryStart);
            res.Transitions.AddTransition(arrayStartToLiteralString);
            res.Transitions.AddTransition(arrayStartToArrayStart);
            res.Transitions.AddTransition(arrayStartToArrayEnd);
            res.Transitions.AddTransition(arrayStartToNameStart);
            res.Transitions.AddTransition(arrayStartToKeyword);
            res.Transitions.AddTransition(arrayStartToComment);

            Transition<PostScriptReaderArgs> arrayEndToNumber = new Transition<PostScriptReaderArgs>(arrayEnd, number, (args) => { return Characters.IsNumberCharacter(args.Byte); });
            Transition<PostScriptReaderArgs> arrayEndToHexStringOrDictionaryStart = new Transition<PostScriptReaderArgs>(arrayEnd, hexStringOrDictionaryStart, (args) => { return Characters.IsHexStringOrDictionaryStart(args.Byte); });
            Transition<PostScriptReaderArgs> arrayEndToLiteralString = new Transition<PostScriptReaderArgs>(arrayEnd, literalString, (args) => { return Characters.IsLiteralStringStart(args.Byte); });
            Transition<PostScriptReaderArgs> arrayEndToArrayStart = new Transition<PostScriptReaderArgs>(arrayEnd, arrayStart, (args) => { return Characters.IsArrayStart(args.Byte); });
            Transition<PostScriptReaderArgs> arrayEndToArrayEnd = new Transition<PostScriptReaderArgs>(arrayEnd, (args) => { return Characters.IsArrayEnd(args.Byte); });
            Transition<PostScriptReaderArgs> arrayEndToDictionaryEndStart = new Transition<PostScriptReaderArgs>(arrayEnd, dictionaryEndStart, (args) => { return Characters.IsHexStringOrDictionaryEnd(args.Byte); });
            Transition<PostScriptReaderArgs> arrayEndToNameStart = new Transition<PostScriptReaderArgs>(arrayEnd, nameStart, (args) => { return Characters.IsNameStart(args.Byte); });
            Transition<PostScriptReaderArgs> arrayEndToKeyword = new Transition<PostScriptReaderArgs>(arrayEnd, keyword, (args) => { return Characters.IsKeywordStartCharacter(args.Byte); });
            Transition<PostScriptReaderArgs> arrayEndToComment = new Transition<PostScriptReaderArgs>(arrayEnd, comment, (args) => { return Characters.IsCommentStart(args.Byte); });

            res.Transitions.AddTransition(arrayEndToNumber);
            res.Transitions.AddTransition(arrayEndToHexStringOrDictionaryStart);
            res.Transitions.AddTransition(arrayEndToLiteralString);
            res.Transitions.AddTransition(arrayEndToArrayStart);
            res.Transitions.AddTransition(arrayEndToArrayEnd);
            res.Transitions.AddTransition(arrayEndToDictionaryEndStart);
            res.Transitions.AddTransition(arrayEndToNameStart);
            res.Transitions.AddTransition(arrayEndToKeyword);
            res.Transitions.AddTransition(arrayEndToComment);

            Transition<PostScriptReaderArgs> dictionaryStartToNumber = new Transition<PostScriptReaderArgs>(dictionaryStart, number, (args) => { return Characters.IsNumberCharacter(args.Byte); });
            Transition<PostScriptReaderArgs> dictionaryStartToHexStringOrDictionaryStart = new Transition<PostScriptReaderArgs>(dictionaryStart, hexStringOrDictionaryStart, (args) => { return Characters.IsHexStringOrDictionaryStart(args.Byte); });
            Transition<PostScriptReaderArgs> dictionaryStartToLiteralString = new Transition<PostScriptReaderArgs>(dictionaryStart, literalString, (args) => { return Characters.IsLiteralStringStart(args.Byte); });
            Transition<PostScriptReaderArgs> dictionaryStartToArrayStart = new Transition<PostScriptReaderArgs>(dictionaryStart, arrayStart, (args) => { return Characters.IsArrayStart(args.Byte); });
            Transition<PostScriptReaderArgs> dictionaryStartToNameStart = new Transition<PostScriptReaderArgs>(dictionaryStart, nameStart, (args) => { return Characters.IsNameStart(args.Byte); });
            Transition<PostScriptReaderArgs> dictionaryStartToKeyword = new Transition<PostScriptReaderArgs>(dictionaryStart, keyword, (args) => { return Characters.IsKeywordStartCharacter(args.Byte); });
            Transition<PostScriptReaderArgs> dictionaryStartToComment = new Transition<PostScriptReaderArgs>(dictionaryStart, comment, (args) => { return Characters.IsCommentStart(args.Byte); });
            Transition<PostScriptReaderArgs> dictionaryStartToDictionaryEndStart = new Transition<PostScriptReaderArgs>(dictionaryStart, dictionaryEndStart, (args) => { return Characters.IsHexStringOrDictionaryEnd(args.Byte); });

            res.Transitions.AddTransition(dictionaryStartToNumber);
            res.Transitions.AddTransition(dictionaryStartToHexStringOrDictionaryStart);
            res.Transitions.AddTransition(dictionaryStartToLiteralString);
            res.Transitions.AddTransition(dictionaryStartToArrayStart);
            res.Transitions.AddTransition(dictionaryStartToNameStart);
            res.Transitions.AddTransition(dictionaryStartToKeyword);
            res.Transitions.AddTransition(dictionaryStartToComment);
            res.Transitions.AddTransition(dictionaryStartToDictionaryEndStart);

            Transition<PostScriptReaderArgs> dictionaryEndStartToDictionaryEnd = new Transition<PostScriptReaderArgs>(dictionaryEndStart, dictionaryEnd, (args) => { return Characters.IsHexStringOrDictionaryEnd(args.Byte); });

            res.Transitions.AddTransition(dictionaryEndStartToDictionaryEnd);

            Transition<PostScriptReaderArgs> dictionaryEndToNumber = new Transition<PostScriptReaderArgs>(dictionaryEnd, number, (args) => { return Characters.IsNumberCharacter(args.Byte); });
            Transition<PostScriptReaderArgs> dictionaryEndToHexStringOrDictionaryStart = new Transition<PostScriptReaderArgs>(dictionaryEnd, hexStringOrDictionaryStart, (args) => { return Characters.IsHexStringOrDictionaryStart(args.Byte); });
            Transition<PostScriptReaderArgs> dictionaryEndToLiteralString = new Transition<PostScriptReaderArgs>(dictionaryEnd, literalString, (args) => { return Characters.IsLiteralStringStart(args.Byte); });
            Transition<PostScriptReaderArgs> dictionaryEndToArrayStart = new Transition<PostScriptReaderArgs>(dictionaryEnd, arrayStart, (args) => { return Characters.IsArrayStart(args.Byte); });
            Transition<PostScriptReaderArgs> dictionaryEndToArrayEnd = new Transition<PostScriptReaderArgs>(dictionaryEnd, arrayEnd, (args) => { return Characters.IsArrayEnd(args.Byte); });
            Transition<PostScriptReaderArgs> dictionaryEndToDictionaryEndStart = new Transition<PostScriptReaderArgs>(dictionaryEnd, dictionaryEndStart, (args) => { return Characters.IsHexStringOrDictionaryEnd(args.Byte); });
            Transition<PostScriptReaderArgs> dictionaryEndToNameStart = new Transition<PostScriptReaderArgs>(dictionaryEnd, nameStart, (args) => { return Characters.IsNameStart(args.Byte); });
            Transition<PostScriptReaderArgs> dictionaryEndToKeyword = new Transition<PostScriptReaderArgs>(dictionaryEnd, keyword, (args) => { return Characters.IsKeywordStartCharacter(args.Byte); });
            Transition<PostScriptReaderArgs> dictionaryEndToComment = new Transition<PostScriptReaderArgs>(dictionaryEnd, comment, (args) => { return Characters.IsCommentStart(args.Byte); });

            res.Transitions.AddTransition(dictionaryEndToNumber);
            res.Transitions.AddTransition(dictionaryEndToHexStringOrDictionaryStart);
            res.Transitions.AddTransition(dictionaryEndToLiteralString);
            res.Transitions.AddTransition(dictionaryEndToArrayStart);
            res.Transitions.AddTransition(dictionaryEndToArrayEnd);
            res.Transitions.AddTransition(dictionaryEndToDictionaryEndStart);
            res.Transitions.AddTransition(dictionaryEndToNameStart);
            res.Transitions.AddTransition(dictionaryEndToKeyword);
            res.Transitions.AddTransition(dictionaryEndToComment);

            Transition<PostScriptReaderArgs> nameStartToName = new Transition<PostScriptReaderArgs>(nameStart, name, (args) => { return Characters.IsValidNameCharacter(args.Byte); });
            // Added NameStart to NameStart to handle documents with empty name values
            Transition<PostScriptReaderArgs> nameStartToNameStart = new Transition<PostScriptReaderArgs>(nameStart, nameStart, (args) => { return Characters.IsNameStart(args.Byte); });
            Transition<PostScriptReaderArgs> nameStartToArrayStart = new Transition<PostScriptReaderArgs>(nameStart, arrayStart, (args) => { return Characters.IsArrayStart(args.Byte); });
            Transition<PostScriptReaderArgs> nameStartToArrayEnd = new Transition<PostScriptReaderArgs>(nameStart, arrayEnd, (args) => { return Characters.IsArrayEnd(args.Byte); });
            Transition<PostScriptReaderArgs> nameStartToDictionaryEnd = new Transition<PostScriptReaderArgs>(nameStart, dictionaryEnd, (args) => { return Characters.IsHexStringOrDictionaryEnd(args.Byte); });
            Transition<PostScriptReaderArgs> nameStartToInitial = new Transition<PostScriptReaderArgs>(nameStart, initial, (args) => { return !Characters.IsValidNameCharacter(args.Byte); });

            res.Transitions.AddTransition(nameStartToName);
            res.Transitions.AddTransition(nameStartToNameStart);
            res.Transitions.AddTransition(nameStartToArrayStart);
            res.Transitions.AddTransition(nameStartToArrayEnd);
            res.Transitions.AddTransition(nameStartToDictionaryEnd);
            res.Transitions.AddTransition(nameStartToInitial);

            Transition<PostScriptReaderArgs> nameToHexStringOrDictionaryStart = new Transition<PostScriptReaderArgs>(name, hexStringOrDictionaryStart, (args) => { return Characters.IsHexStringOrDictionaryStart(args.Byte); });
            Transition<PostScriptReaderArgs> nameToLiteralString = new Transition<PostScriptReaderArgs>(name, literalString, (args) => { return Characters.IsLiteralStringStart(args.Byte); });
            Transition<PostScriptReaderArgs> nameToArrayStart = new Transition<PostScriptReaderArgs>(name, arrayStart, (args) => { return Characters.IsArrayStart(args.Byte); });
            Transition<PostScriptReaderArgs> nameToArrayEnd = new Transition<PostScriptReaderArgs>(name, arrayEnd, (args) => { return Characters.IsArrayEnd(args.Byte); });
            Transition<PostScriptReaderArgs> nameToDictionaryEndStart = new Transition<PostScriptReaderArgs>(name, dictionaryEndStart, (args) => { return Characters.IsHexStringOrDictionaryEnd(args.Byte); });
            Transition<PostScriptReaderArgs> nameToNameStart = new Transition<PostScriptReaderArgs>(name, nameStart, (args) => { return Characters.IsNameStart(args.Byte); });
            Transition<PostScriptReaderArgs> nameToComment = new Transition<PostScriptReaderArgs>(name, comment, (args) => { return Characters.IsCommentStart(args.Byte); });
            Transition<PostScriptReaderArgs> nameToName = new Transition<PostScriptReaderArgs>(name, (args) => { return Characters.IsValidNameCharacter(args.Byte); });
            Transition<PostScriptReaderArgs> nameToInitial = new Transition<PostScriptReaderArgs>(name, initial, (args) => { return !Characters.IsValidNameCharacter(args.Byte); });

            res.Transitions.AddTransition(nameToHexStringOrDictionaryStart);
            res.Transitions.AddTransition(nameToLiteralString);
            res.Transitions.AddTransition(nameToArrayStart);
            res.Transitions.AddTransition(nameToArrayEnd);
            res.Transitions.AddTransition(nameToDictionaryEndStart);
            res.Transitions.AddTransition(nameToNameStart);
            res.Transitions.AddTransition(nameToComment);
            res.Transitions.AddTransition(nameToName);
            res.Transitions.AddTransition(nameToInitial);

            Transition<PostScriptReaderArgs> keywordToHexStringOrDictionaryStart = new Transition<PostScriptReaderArgs>(keyword, hexStringOrDictionaryStart, (args) => { return Characters.IsHexStringOrDictionaryStart(args.Byte); });
            Transition<PostScriptReaderArgs> keywordToLiteralString = new Transition<PostScriptReaderArgs>(keyword, literalString, (args) => { return Characters.IsLiteralStringStart(args.Byte); });
            Transition<PostScriptReaderArgs> keywordToArrayStart = new Transition<PostScriptReaderArgs>(keyword, arrayStart, (args) => { return Characters.IsArrayStart(args.Byte); });
            Transition<PostScriptReaderArgs> keywordToArrayEnd = new Transition<PostScriptReaderArgs>(keyword, arrayEnd, (args) => { return Characters.IsArrayEnd(args.Byte); });
            Transition<PostScriptReaderArgs> keywordToDictionaryEndStart = new Transition<PostScriptReaderArgs>(keyword, dictionaryEndStart, (args) => { return Characters.IsHexStringOrDictionaryEnd(args.Byte); });
            Transition<PostScriptReaderArgs> keywordToNameStart = new Transition<PostScriptReaderArgs>(keyword, nameStart, (args) => { return Characters.IsNameStart(args.Byte); });
            Transition<PostScriptReaderArgs> keywordToKeyword = new Transition<PostScriptReaderArgs>(keyword, keyword, (args) => { return Characters.IsKeywordCharacter(args.Byte); });
            Transition<PostScriptReaderArgs> keywordToComment = new Transition<PostScriptReaderArgs>(keyword, comment, (args) => { return Characters.IsCommentStart(args.Byte); });
            Transition<PostScriptReaderArgs> keywordToInitial = new Transition<PostScriptReaderArgs>(keyword, initial, (args) => { return true; });

            res.Transitions.AddTransition(keywordToHexStringOrDictionaryStart);
            res.Transitions.AddTransition(keywordToLiteralString);
            res.Transitions.AddTransition(keywordToArrayStart);
            res.Transitions.AddTransition(keywordToArrayEnd);
            res.Transitions.AddTransition(keywordToDictionaryEndStart);
            res.Transitions.AddTransition(keywordToNameStart);
            res.Transitions.AddTransition(keywordToKeyword);
            res.Transitions.AddTransition(keywordToComment);
            res.Transitions.AddTransition(keywordToInitial);

            Transition<PostScriptReaderArgs> hexStringToInitial = new Transition<PostScriptReaderArgs>(hexString, initial, (args) => { return Characters.IsHexStringOrDictionaryEnd(args.Byte); });
            Transition<PostScriptReaderArgs> hexStringToHexString = new Transition<PostScriptReaderArgs>(hexString, (b) => { return true; });

            res.Transitions.AddTransition(hexStringToInitial);
            res.Transitions.AddTransition(hexStringToHexString);

            Transition<PostScriptReaderArgs> commentToInitial = new Transition<PostScriptReaderArgs>(comment, initial, (args) => { return Characters.IsLineFeed(args.Byte); });

            res.Transitions.AddTransition(commentToInitial);

            return res;
        }
    }
}
