﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser.PrimitiveParsers
{
    internal class LiteralStringParser
    {
        private const byte LeftBracket = (byte)'(';
        private const byte RightBracket = (byte)')';
        private const byte EscapingSlash = (byte)'\\';
        private const byte CarriageReturn = (byte)'\r';
        private const byte EndOfLineMarker = (byte)'\n';
        private const byte HorizontalTab = (byte)'\t';
        private const byte Backspace = (byte)'\b';
        private const byte FormFeed = (byte)'\f';
        private static readonly Dictionary<byte, byte> singleEscapedCharacterToByte;

        private readonly PostScriptReader parser;
        private readonly Func<byte, bool>[] escapedByteHandlers;
        private readonly Stack<byte> octalSequenceStack;
        private readonly List<byte> token;

        private bool hasReadCarriageReturn;
        private bool isInEscapeMode;
        private int openedParentheses;

        static LiteralStringParser()
        {
            singleEscapedCharacterToByte = new Dictionary<byte, byte>();
            singleEscapedCharacterToByte.Add((byte)'n', EndOfLineMarker);
            singleEscapedCharacterToByte.Add((byte)'r', CarriageReturn);
            singleEscapedCharacterToByte.Add((byte)'t', HorizontalTab);
            singleEscapedCharacterToByte.Add((byte)'b', Backspace);
            singleEscapedCharacterToByte.Add((byte)'f', FormFeed);
            singleEscapedCharacterToByte.Add(LeftBracket, LeftBracket);
            singleEscapedCharacterToByte.Add(RightBracket, RightBracket);
            singleEscapedCharacterToByte.Add(EscapingSlash, EscapingSlash);
        }

        public LiteralStringParser(PostScriptReader parser)
        {
            Guard.ThrowExceptionIfNull(parser, "parser");

            this.parser = parser;
            this.token = new List<byte>();
            this.octalSequenceStack = new Stack<byte>();

            this.escapedByteHandlers = new Func<byte, bool>[]
            {
                this.TryHandleEscapedOctalSequence,
                this.TryHandleEndOfLineSymbolsInEscapeMode,
                this.TryHandleSingleEscapedSymbol
            };
        }

        public void Append(PostScriptReaderArgs args)
        {
            byte character = args.Byte;

            if (this.isInEscapeMode)
            {
                this.HandleByteInEscapeMode(character);
            }
            else
            {
                this.HandleByteWhenNotEscaping(character);
            }
        }

        public bool IsLiteralStringEnd(PostScriptReaderArgs args)
        {
            if (args.Byte == RightBracket && this.openedParentheses == 0)
            {
                return !this.isInEscapeMode || this.octalSequenceStack.Count > 0 || this.hasReadCarriageReturn;
            }

            return false;
        }

        public void Complete(PostScriptReaderArgs args)
        {
            Guard.ThrowExceptionIfNull(args, "args");

            if (this.hasReadCarriageReturn && !this.isInEscapeMode)
            {
                this.token.Add(EndOfLineMarker);
            }
            else if (this.octalSequenceStack.Count > 0)
            {
                this.AddAndClearAccumulatedOctalNumber();
            }

            byte[] tokenArray = this.token.ToArray();
            byte[] decryptedToken = args.Context.DecryptString(tokenArray);
            PdfLiteralString literalString = new PdfLiteralString(decryptedToken);
            this.parser.PushToken(literalString);

            this.Reset();
        }

        private void HandleByteInEscapeMode(byte character)
        {
            bool isValidEscaping = false;

            for(int i = 0; i < this.escapedByteHandlers.Length; i ++)
            {
                Func<byte, bool> tryHandleByteInEscapedModeHandler = this.escapedByteHandlers[i];
                if (tryHandleByteInEscapedModeHandler(character))
                {
                    isValidEscaping = true;
                    break;
                }
            }

            if (!isValidEscaping)
            {
                this.StopEscapingIgnoringPreviousBackslash(character);
            }
        }

        private void HandleByteWhenNotEscaping(byte character)
        {
            this.EnsureAddingPreviouslyReadCariageReturn(character);

            if (character == EscapingSlash)
            {
                this.isInEscapeMode = true;
            }
            else
            {
                if (!this.hasReadCarriageReturn)
                {
                    this.AppendByteCountingBrackets(character);
                }
            }
        }

        private bool TryHandleEscapedOctalSequence(byte character)
        {
            bool shouldHandleOctalSequence = !this.hasReadCarriageReturn;

            if (shouldHandleOctalSequence)
            {
                bool isOctalCharacter = Characters.IsOctalChar(character);
                shouldHandleOctalSequence = isOctalCharacter || this.octalSequenceStack.Count > 0;

                if (isOctalCharacter)
                {
                    this.octalSequenceStack.Push(character);

                    if (this.octalSequenceStack.Count == 3)
                    {
                        this.AddAndClearAccumulatedOctalNumber();
                    }
                }
                else if (this.octalSequenceStack.Count > 0)
                {
                    this.AddAndClearAccumulatedOctalNumber();
                    this.HandleByteWhenNotEscaping(character);
                }
            }

            return shouldHandleOctalSequence;
        }

        private bool TryHandleEndOfLineSymbolsInEscapeMode(byte character)
        {
            bool hasHandledEndOfLineSymbol = character == EndOfLineMarker || character == CarriageReturn || this.hasReadCarriageReturn;

            if (character == EndOfLineMarker)
            {
                this.hasReadCarriageReturn = false;
                this.isInEscapeMode = false;
            }
            else if (character == CarriageReturn)
            {
                if (this.hasReadCarriageReturn)
                {
                    this.isInEscapeMode = false;
                }

                this.hasReadCarriageReturn = true;
            }
            else if(this.hasReadCarriageReturn)
            {
                this.hasReadCarriageReturn = false;
                this.isInEscapeMode = false;
                this.HandleByteWhenNotEscaping(character);
            }

            return hasHandledEndOfLineSymbol;
        }

        private bool TryHandleSingleEscapedSymbol(byte character)
        {
            byte symbol;
            if (LiteralStringParser.singleEscapedCharacterToByte.TryGetValue(character, out symbol))
            {
                this.isInEscapeMode = false;
                this.token.Add(symbol);

                return true;
            }

            return false;
        }

        private void AddAndClearAccumulatedOctalNumber()
        {
            this.isInEscapeMode = false;

            int multiplier = 1;
            int result = 0;

            while(this.octalSequenceStack.Count > 0)
            {
                byte number = this.octalSequenceStack.Pop();
                int numberValue = number - Characters.Zero;
                result += numberValue * multiplier;
                multiplier *= 8;
            }

            byte byteOctal = (byte)(result % 256);

            this.token.Add(byteOctal);
        }

        private void StopEscapingIgnoringPreviousBackslash(byte character)
        {
            this.isInEscapeMode = false;
            this.HandleByteWhenNotEscaping(character);
        }

        private void EnsureAddingPreviouslyReadCariageReturn(byte currentCharacter)
        {
            if (this.hasReadCarriageReturn && currentCharacter != EndOfLineMarker)
            {
                this.token.Add(EndOfLineMarker);
            }

            this.hasReadCarriageReturn = currentCharacter == CarriageReturn;
        }

        private void AppendByteCountingBrackets(byte character)
        {
            if (character == LeftBracket)
            {
                this.openedParentheses++;
            }
            else if (character == RightBracket)
            {
                this.openedParentheses--;
            }

            this.token.Add(character);
        }

        private void Reset()
        {
            this.token.Clear();
            this.isInEscapeMode = false;
            this.hasReadCarriageReturn = false;
        }
    }
}
