﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser
{
    internal class Reader
    {
        private readonly Stream input;
        private readonly Stack<long> beginReadingPositions;
        private readonly Stack<ManualResetEvent> waitHandles;
        private readonly StringBuilder stringBuilder;
        private readonly object syncRoot;

        private long cachedLength = -1;
        private int? lockedThread;

        public Reader(Stream input)
        {
            Guard.ThrowExceptionIfNull(input, "input");

            this.input = input;
            this.beginReadingPositions = new Stack<long>();
            this.waitHandles = new Stack<ManualResetEvent>();
            this.stringBuilder = new StringBuilder();
            this.syncRoot = new object();
        }

        internal Stream Stream
        {
            get
            {
                return this.input;
            }
        }

        public bool EndOfFile
        {
            get
            {
                return this.Position >= this.Length;
            }
        }

        public long Length
        {
            get
            {
                if (this.cachedLength != -1)
                {
                    return this.cachedLength;
                }
                else
                {
                    return this.input.Length;
                }
            }
        }

        public virtual long Position
        {
            get
            {
                return this.input.Position;
            }
        }

        public virtual IDisposable BeginReadingBlock()
        {
            if (!this.lockedThread.HasValue || Thread.CurrentThread.ManagedThreadId != this.lockedThread.Value)
            {
                lock (this.syncRoot)
                {
                    if (this.lockedThread.HasValue && Thread.CurrentThread.ManagedThreadId != this.lockedThread.Value)
                    {
                        ManualResetEvent waitHandle = new ManualResetEvent(false);
                        waitHandles.Push(waitHandle);

                        waitHandle.WaitOne();
                    }

                    this.lockedThread = Thread.CurrentThread.ManagedThreadId;
                }
            }

            this.beginReadingPositions.Push(this.Position);

            return new DisposableObject(this.EndReadingBlock);
        }

        public virtual void EndReadingBlock()
        {
            if (this.beginReadingPositions.Count > 0)
            {
                this.Seek(this.beginReadingPositions.Pop(), SeekOrigin.Begin);
            }

            if (this.beginReadingPositions.Count == 0)
            {
                this.lockedThread = null;

                if (this.waitHandles.Count > 0)
                {
                    this.waitHandles.Pop().Set();
                }
            }
        }

        public byte Peek()
        {
            byte b = this.Read();
            this.Seek(-1, SeekOrigin.Current);

            return b;
        }

        public virtual byte Read()
        {
            int res = this.input.ReadByte();
            if (res < 0)
            {
                throw new EndOfStreamException();
            }

            return (byte)res;
        }

        public byte[] Read(int count)
        {
            byte[] data = new byte[count];
            this.input.Read(data, 0, count);

            return data;
        }

        public byte[] ReadBE(int count)
        {
            byte[] buffer = this.Read(count);
            int half = buffer.Length / 2;
            byte tmp;
            for (int i = 0; i < half; i++)
            {
                tmp = buffer[i];
                buffer[i] = buffer[count - i];
                buffer[count - i] = tmp;
            }

            return buffer;
        }

        public string ReadLine()
        {
            this.CacheLength();

            this.stringBuilder.Clear();
            byte b;
            while (!this.EndOfFile && !Characters.IsLineFeed(b = this.Read()))
            {
                if (!Characters.IsCarriageReturn(b))
                {
                    this.stringBuilder.Append((char)b);
                }
                else if (!this.EndOfFile && !Characters.IsLineFeed(b = this.Peek()))
                {
                    break;
                }
            }

            this.ClearCachedLength();

            return this.stringBuilder.ToString();
        }

        public string[] ReadSplitLine()
        {
            char[] separators = Telerik.Windows.Documents.Core.PostScript.Data.PostScriptConstants.Separators;
            return this.ReadLine().Split(separators, StringSplitOptions.RemoveEmptyEntries);
        }

        public void Seek(long offset, SeekOrigin origin)
        {
            this.input.Seek(offset, origin);
        }

        public long IndexOf(byte[] byteSequence)
        {
            int byteSequenceIndexToCheck = 0;

            while (!this.EndOfFile && byteSequenceIndexToCheck != byteSequence.Length)
            {
                byte b = this.Read();

                if (b.Equals(byteSequence[byteSequenceIndexToCheck]))
                {
                    byteSequenceIndexToCheck++;
                }
                else if (byteSequenceIndexToCheck != 0)
                {
                    this.Seek(-byteSequenceIndexToCheck, SeekOrigin.Current);
                    byteSequenceIndexToCheck = 0;
                }
            }

            if (byteSequenceIndexToCheck == byteSequence.Length)
            {
                long startPosition = this.Position - byteSequenceIndexToCheck;

                return startPosition;
            }
            else
            {
                return -1;
            }
        }

        public void CacheLength()
        {
            if (input is MemoryStream || input is FileStream)
            {
                this.cachedLength = input.Length;
            }
        }

        public void ClearCachedLength()
        {
            this.cachedLength = -1;
        }
    }
}
