﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Core.PostScript.Data;
using Telerik.Windows.Documents.Core.PostScript.Encryption;
using Telerik.Windows.Documents.Core.PostScript.Operators;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Core.PostScript
{
    internal class PostScriptReader : ReaderBase, IPostScriptReader
    {
        #region Constructors

        public PostScriptReader(byte[] data)
            : base(data)
        {
            this.encryption = new EncryptionCollection();
            this.decryptedBuffer = new Queue<byte>();
        }

        #endregion


        #region Fields

        private readonly EncryptionCollection encryption;
        private readonly Queue<byte> decryptedBuffer;

        #endregion


        #region Properties

        public string Result { get;set; }

        public override long Position
        {
            get
            {
                return base.Position - this.decryptedBuffer.Count;
            }
        }

        #endregion


        #region Methods

        public Token ReadToken()
        {
            this.SkipUnusedCharacters();

            if (this.EndOfFile)
            {
                return Token.Unknown;
            }

            switch ((char)this.Peek())
            {
                case Characters.Name:
                    return this.ReadName();
                case Characters.ExecuteableArrayStart:
                case Characters.ArrayStart:
                    this.Read();
                    return Token.ArrayStart;
                case Characters.ArrayEnd:
                case Characters.ExecuteableArrayEnd:
                    this.Read();
                    return Token.ArrayEnd;
                case Characters.LiteralStringStart:
                    return this.ReadLiteralString();
                case Characters.HexadecimalStringEnd:
                    return this.ReadHexadecimalString();
            }

            if (Characters.IsValidNumberChar(this))
            {
                return this.ReadNumber();
            }

            if (Characters.IsLetter(this.Peek()) || this.Peek() == Characters.Minus || this.Peek() == Characters.VerticalBar)
            {
                return this.ReadOperatorOrKeyword();
            }

            this.Read();
            return Token.Unknown;
        }
        
        internal void SkipUnusedCharacters()
        {
            PostScriptReaderHelper.SkipUnusedCharacters(this);
        }

        internal void GoToNextLine()
        {
            PostScriptReaderHelper.GoToNextLine(this);
        }

        internal void SkipWhiteSpaces()
        {
            PostScriptReaderHelper.SkipWhiteSpaces(this);
        }
        
        private Token ReadOperatorOrKeyword()
        {
            this.Result = PostScriptReaderHelper.ReadKeyword(this);
            
            switch (this.Result)
            {
                case Keywords.True:
                case Keywords.False:
                    return Token.Boolean;
            }

            if (Keywords.IsKeyword(this.Result))
            {
                return Token.Keyword;
            }

            if (Operator.IsOperator(this.Result))
            {
                return Token.Operator;
            }

            return Token.Unknown;
        }

        private Token ReadName()
        {
            this.Result = PostScriptReaderHelper.ReadName(this);
            return Token.Name;
        }

        private Token ReadNumber()
        {
            this.Result = PostScriptReaderHelper.ReadNumber(this);
            return Enumerable.Contains(this.Result, Characters.DecimalPoint) ? Token.Real : Token.Integer;
        }

        private Token ReadHexadecimalString()
        {
            this.Result = PostScriptReaderHelper.GetString(PostScriptReaderHelper.ReadHexadecimalString(this));
            return Token.String;
        }

        private Token ReadLiteralString()
        {
            this.Result = PostScriptReaderHelper.GetString(PostScriptReaderHelper.ReadLiteralString(this));
            return Token.String;
        }

        #endregion


        #region Overrides

        public override void BeginReadingBlock()
        {
            if (this.encryption.HasEncryption)
            {
                Guard.ThrowNotSupportedException();
            }
            base.BeginReadingBlock();
        }

        public override void EndReadingBlock()
        {
            if (this.encryption.HasEncryption)
            {
                Guard.ThrowNotSupportedException();
            }
            base.EndReadingBlock();
        }

        public override void Seek(long offset, System.IO.SeekOrigin origin)
        {
            if (this.encryption.HasEncryption)
            {
                Guard.ThrowNotSupportedException();
            }

            base.Seek(offset, origin);
        }

        public override byte Read()
        {
            if (this.encryption.HasEncryption)
            {
                if (this.decryptedBuffer.Count > 0)
                {
                    return this.decryptedBuffer.Dequeue();
                }

                return this.encryption.Decrypt(base.Read());
            }

            byte res = base.Read();
            if (res == Characters.CR)
            {
                if (!this.EndOfFile && this.Peek() == Characters.LF)
                {
                    this.Read();
                }
                res = (byte)Characters.LF;
            }

            return res;
        }
       
        public override byte Peek(int skip = 0)
        {
            byte res = 0;
            if (this.encryption.HasEncryption)
            {
                if (skip < this.decryptedBuffer.Count)
                {
                    return Enumerable.ElementAt(this.decryptedBuffer, skip);
                }

                skip -= this.decryptedBuffer.Count;
                for (int i = 0; i <= skip; i++)
                {
                    res = this.encryption.Decrypt(base.Read());
                    this.decryptedBuffer.Enqueue(res);
                }

                return res;
            }

            res = base.Peek(skip);

            if (res == Characters.CR)
            {
                res = (byte)Characters.LF;
            }

            return res;
        }

        public void PushEncryption(EncryptionBase encrypt)
        {
            this.encryption.PushEncryption(encrypt);
        }

        public void PopEncryption()
        {
            this.encryption.PopEncryption();

            if (!this.encryption.HasEncryption)
            {
                this.decryptedBuffer.Clear();
            }
        }

        byte IPostScriptReader.Read(bool appendLineFeed)
        {
            return this.Read();
        }

        #endregion

        
    }
}
