﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Core.PostScript.Data;

namespace Telerik.Windows.Documents.Core.PostScript
{
    internal static class PostScriptReaderHelper
    {
        static PostScriptReaderHelper()
        {
            charValues = new int[128];
            for (int i = 0; i < 10; ++i)
            {
                charValues['0' + i] = i;
            }
            for (int i = 0; i < 6; ++i)
            {
                charValues['a' + i] = 0xA + i;
                charValues['A' + i] = 0xA + i;
            }
        }

        private static int[] charValues;

        public static void GoToNextLine(IPostScriptReader reader, bool appendLineFeed = true)
        {
            while (!reader.EndOfFile && reader.Read(appendLineFeed) != Characters.LF)
            {
            }
        }
        public static void SkipUnusedCharacters(IPostScriptReader reader)
        {
            SkipWhiteSpaces(reader);

            while (!reader.EndOfFile && reader.Peek() == Characters.Comment)
            {
                GoToNextLine(reader);
                SkipWhiteSpaces(reader);
            }
        }
        public static void SkipWhiteSpaces(IPostScriptReader reader)
        {
            while (!reader.EndOfFile && Characters.IsWhiteSpace(reader.Peek()))
            {
                reader.Read();
            }
        }
        public static string GetString(byte[] bytes)
        {
            return System.Text.Encoding.UTF8.GetString(bytes, 0, bytes.Length);
        }
        public static string ReadNumber(IPostScriptReader reader)
        {
            StringBuilder token = new StringBuilder();

            while (!reader.EndOfFile && Characters.IsValidNumberChar(reader))
            {
                token.Append((char)reader.Read());
            }
            return token.ToString();
        }
        public static string ReadName(IPostScriptReader reader)
        {
            StringBuilder token = new StringBuilder();
            reader.Read();
            while (!reader.EndOfFile && !Characters.IsDelimiter(reader.Peek()))
            {
                char nextChar = (char)reader.Read();
                if (nextChar == '#')
                {
                    string result;
                    if (TryReadHexChar(reader, out result))
                    {
                        token.Append(result);
                    }
                    else
                    {
                        token.Append(nextChar);
                        token.Append(result);
                    }
                }
                else
                {
                    token.Append(nextChar);
                }
            }
            return token.ToString();
        }
        public static string ReadKeyword(IPostScriptReader reader)
        {
            StringBuilder token = new StringBuilder();
            while (!reader.EndOfFile && !Characters.IsDelimiter(reader.Peek()) && !char.IsNumber((char)reader.Peek()))
            {
                token.Append((char)reader.Read());
            }

            return token.ToString();
        }
        public static byte[] ReadHexadecimalString(IPostScriptReader reader)
        {
            StringBuilder hex = new StringBuilder();
            reader.Read();

            while (!reader.EndOfFile && reader.Peek() != Characters.HexadecimalStringEnd)
            {
                if (Characters.IsHexChar(reader.Peek()))
                {
                    hex.Append((char)reader.Read());
                }
                else
                {
                    reader.Read();
                }
            }
            if (!reader.EndOfFile)
            {
                reader.Read();
            }
            if (hex.Length % 2 == 1)
            {
                hex.Append("0");
            }
            return GetBytesFromHexString(hex.ToString());
        }
        public static byte[] ReadLiteralString(IPostScriptReader reader)
        {
            List<byte> lit = new List<byte>();
            reader.Read();

            int opened = 1;
            while (opened > 0)
            {
                if (reader.Peek() == Characters.StringEscapeCharacter)
                {
                    if (PostScriptReaderHelper.IsValidEscape(reader.Peek(1)))
                    {
                        reader.Read();
                        if (reader.Peek() != Characters.CR && reader.Peek() != Characters.LF)
                        {
                            lit.Add((byte)PostScriptReaderHelper.GetSymbolFromEscapeSymbol(reader.Read()));
                        }
                        else
                        {
                            SkipWhiteSpaces(reader);
                        }
                        continue;
                    }
                    else if (Characters.IsOctalChar(reader.Peek(1)))
                    {
                        int br = 3;
                        reader.Read();
                        List<byte> bytes = new List<byte>();
                        
                        while (Characters.IsOctalChar(reader.Peek()) && br > 0)
                        {
                            bytes.Add((byte)(reader.Read() - '0'));
                            br--;
                        }
                        
                        while (bytes.Count < 3)
                        {
                            bytes.Insert(0, 0);
                        }
                        
                        int oct = 0;
                        int st = 1;
                        for (int i = 2; i >= 0; i--)
                        {
                            oct += bytes[i] * st;
                            st *= 8;
                        }

                        lit.Add((byte)oct);
                        
                        continue;
                    }
                    else
                    {
                        reader.Read();
                        continue;
                    }
                }
                else if (reader.Peek() == Characters.LiteralStringStart)
                {
                    opened++;
                }
                else if (reader.Peek() == Characters.LiteralStringEnd)
                {
                    opened--;
                    if (opened == 0)
                    {
                        break;
                    }
                }
                lit.Add(reader.Read());
            }
            reader.Read();
            return lit.ToArray();
        }
        public static byte[] GetBytesFromHexString(string hexString)
        {
            if (string.IsNullOrEmpty(hexString))
            {
                return new byte[0];
            }
            if (hexString.Length % 2 != 0)
            {
                hexString = "0" + hexString;
            }
            byte[] result = new byte[hexString.Length >> 1];
            for (int i = 0, l = hexString.Length; i < l; i += 2)
            {
                result[i >> 1] = (byte)((charValues[hexString[i]] << 4) | charValues[hexString[i + 1]]);
            }
            return result;
        }
        public static bool IsValidEscape(int b)
        {
            switch (b)
            {
                case '(':
                case ')':
                case '\\':
                case Characters.LF:
                case Characters.CR:
                case 'n':
                case 'r':
                case 't':
                case 'b':
                case 'f':
                    return true;
            }
            return false;
        }
        public static char GetSymbolFromEscapeSymbol(int symbol)
        {
            switch (symbol)
            {
                case '(':
                    return '(';
                case ')':
                    return ')';
                case '\\':
                    return '\\';
                case 'n':
                    return '\n';
                case 'r':
                    return '\r';
                case 't':
                    return '\t';
                case 'b':
                    return '\b';
                case 'f':
                    return '\f';
                default:
                    return '\0';
            }
        }

        private static bool TryReadHexChar(IPostScriptReader reader, out string result)
        {
            StringBuilder token = new StringBuilder();
            for (int i = 0; i < 2; i++)
            {
                char ch = (char)reader.Read();
                if (reader.EndOfFile || Characters.IsDelimiter(ch) || !Characters.IsHexChar(ch))
                {
                    result = token.ToString();
                    return false;
                }
                token.Append(ch);
            }

            result = ((char)GetBytesFromHexString(token.ToString())[0]).ToString();
            return true;
        }

    }
}
