﻿using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Fixed.Text;

namespace Telerik.Windows.Documents.Fixed.Search
{
    internal static class KMPTextSearch
    {
        private static int[] CalculatePrefixArray(string input)
        {
            int[] prefixArray = new int[input.Length];

            prefixArray[0] = 0;
            int k = 0;
            for (int i = 1; i < input.Length; i++)
            {
                while (k > 0 && (input[k] != input[i]))
                {
                    k = prefixArray[k - 1];
                }

                if (input[k] == input[i])
                {
                    k++;
                }

                prefixArray[i] = k;
            }

            return prefixArray;
        }
        private static IEnumerable<SearchResult> Find(TextPage page, string searchText, int startIndex, bool matchCase, bool wholeWordsOnly, bool singleMatch)
        {
            string input = page.ToString(TextSearchOptions.TextSearchLinesSeparator);
            startIndex += TextSearch.GetStartIndexOffset(startIndex, page, TextSearchOptions.TextSearchLinesSeparator);

            List<SearchResult> result = new List<SearchResult>();

            if (startIndex < 0 || startIndex >= input.Length)
            {
                return result;
            }

            if (!matchCase)
            {
                input = input.ToLower();
                searchText = searchText.ToLower();
            }

            int[] prefixes = CalculatePrefixArray(searchText);
            int k = 0;
            for (int i = startIndex; i < input.Length; i++)
            {
                while (k > 0 && searchText[k] != input[i])
                {
                    k = prefixes[k - 1];
                }

                if (searchText[k] == input[i])
                {
                    k++;
                }

                if (k == searchText.Length)
                {
                    bool matchFound = true;

                    if (wholeWordsOnly)
                    {
                        if (i < input.Length - 1 && !WordSeparators.IsWordSeparator(input[i + 1]))
                        {
                            matchFound = false;
                        }

                        if (i - searchText.Length >= 0 && !WordSeparators.IsWordSeparator(input[i - searchText.Length]))
                        {
                            matchFound = false;
                        }
                    }

                    if (matchFound)
                    {
                        int foundIndex = i - searchText.Length + 1;
                        foundIndex = foundIndex - TextSearch.GetFoundIndexOffset(foundIndex, page, TextSearchOptions.TextSearchLinesSeparator);

                        TextPosition start = new TextPosition(page, foundIndex);
                        TextPosition end = new TextPosition(page, foundIndex + searchText.Length);
                        result.Add(new SearchResult(start, end));
                        if (singleMatch)
                        {
                            return result;
                        }
                    }

                    k = prefixes[k - 1];
                }
            }

            return result;
        }

        public static IEnumerable<SearchResult> FindAll(TextDocument document, string searchText, bool matchCase, bool wholeWordsOnly)
        {
            List<SearchResult> res = new List<SearchResult>();
            for (int i = 0; i < document.PagesCount; i++)
            {
                res.AddRange(Find(document.GetTextPage(i), searchText, 0, matchCase, wholeWordsOnly, false));
            }
            return res;
        }

        public static SearchResult FindNext(TextDocument document, TextPosition startPosition, string searchText, bool matchCase, bool wholeWordsOnly)
        {
            IEnumerable<SearchResult> results;
            int pageIndex = document.IndexOf(startPosition.TextPage);
            int startIndex = startPosition.Index;
            for (int i = pageIndex; i < document.PagesCount; i++)
            {
                results = Find(document.GetTextPage(i), searchText, startIndex, matchCase, wholeWordsOnly, true);
                if (Enumerable.Count(results) > 0)
                {
                    return Enumerable.First(results);
                }
                startIndex = 0;
            }
            return SearchResult.NotFound;
        }

        public static SearchResult FindPrev(TextDocument document, TextPosition startPosition, string searchText, bool matchCase, bool wholeWordsOnly)
        {
            IEnumerable<SearchResult> results;
            int pageIndex = document.IndexOf(startPosition.TextPage);
            results = Find(document.GetTextPage(pageIndex), searchText, 0, matchCase, wholeWordsOnly, false);
            SearchResult res = Enumerable.LastOrDefault(results, r => r.Range.EndPosition <= startPosition);
            if (res != null)
            {
                return res;
            }

            for (int i = pageIndex - 1; i >= 0; i--)
            {
                results = Find(document.GetTextPage(i), searchText, 0, matchCase, false, false);
                if (Enumerable.Count(results) > 0)
                {
                    return Enumerable.Last(results);
                }
            }
            return SearchResult.NotFound;
        }
    }
}
