﻿using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using Telerik.Windows.Documents.Fixed.Text;

namespace Telerik.Windows.Documents.Fixed.Search
{
    internal static class RegexTextDocumentSearch
    {
        private static string CreateRegexWholeWord(string word)
        {
            string wordSeparatorRegEx = WordSeparators.GetRegex();

            word = "(" + word + ")";
            return wordSeparatorRegEx + word + wordSeparatorRegEx + "|" + //When the word is between two other words.
                   "^" + word + wordSeparatorRegEx + "|" + //When the word is the beginning of the text.
                   wordSeparatorRegEx + word + "$" + "|" + //When the word is the ending of the text.
                   "^" + word + "$";  //When the word is the beginning and the ending of the text.
        }

        private static int GetMatchGroupIndex(Match match)
        {
            for (int i = 1; i < match.Groups.Count; i++)
            {
                if (match.Groups[i].Success)
                {
                    return i;
                }
            }

            return 0;
        }

        private static IEnumerable<SearchResult> FindAll(TextPage page, string searchText, bool matchCase, bool wholeWordsOnly)
        {
            string inputText = page.ToString(TextSearchOptions.TextSearchLinesSeparator);

            RegexOptions options = matchCase ? RegexOptions.None : RegexOptions.IgnoreCase;

            if (wholeWordsOnly)
            {
                searchText = CreateRegexWholeWord(searchText);
            }

            MatchCollection matches = Regex.Matches(inputText, searchText, options);

            List<SearchResult> result = new List<SearchResult>();
            foreach (Match match in matches)
            {
                int matchGroupIndex = GetMatchGroupIndex(match);
                int matchIndex = match.Groups[matchGroupIndex].Index - TextSearch.GetFoundIndexOffset(match.Groups[matchGroupIndex].Index, page, TextSearchOptions.TextSearchLinesSeparator);

                TextPosition start = new TextPosition(page, matchIndex);
                TextPosition end = new TextPosition(page, matchIndex + match.Groups[matchGroupIndex].Length);

                result.Add(new SearchResult(start, end));
            }

            return result;
        }

        private static SearchResult FindSingle(TextPage page, string searchText, int startIndex, bool matchCase, bool wholeWordsOnly)
        {
            startIndex += TextSearch.GetStartIndexOffset(startIndex, page, TextSearchOptions.TextSearchLinesSeparator);
            string inputText = page.ToString(TextSearchOptions.TextSearchLinesSeparator).Substring(startIndex);

            RegexOptions options = matchCase ? RegexOptions.None : RegexOptions.IgnoreCase;

            if (wholeWordsOnly)
            {
                searchText = CreateRegexWholeWord(searchText);
            }

            Match match = Regex.Match(inputText, searchText, options);

            if (match.Success)
            {
                int matchGroupIndex = GetMatchGroupIndex(match);
                int matchIndex = match.Groups[matchGroupIndex].Index - TextSearch.GetFoundIndexOffset(startIndex + match.Groups[matchGroupIndex].Index, page, TextSearchOptions.TextSearchLinesSeparator);
                TextPosition end = new TextPosition(page, startIndex + matchIndex + match.Groups[matchGroupIndex].Length);
                TextPosition start = new TextPosition(page, startIndex + matchIndex);

                return new SearchResult(start, end);
            }

            return SearchResult.NotFound;
        }

        public static IEnumerable<SearchResult> FindAll(TextDocument document, string searchText, bool matchCase, bool wholeWordsOnly)
        {
            List<SearchResult> res = new List<SearchResult>();
            for (int i = 0; i < document.PagesCount; i++)
            {
                res.AddRange(FindAll(document.GetTextPage(i), searchText, matchCase, wholeWordsOnly));
            }

            return res;
        }

        public static SearchResult FindNext(TextDocument document, TextPosition startPosition, string searchText, bool matchCase, bool wholeWordsOnly)
        {
            int pageIndex = document.IndexOf(startPosition.TextPage);
            int startIndex = startPosition.Index;

            SearchResult res;
            for (int i = pageIndex; i < document.PagesCount; i++)
            {
                res = FindSingle(document.GetTextPage(i), searchText, startIndex, matchCase, wholeWordsOnly);
                if (res != SearchResult.NotFound)
                {
                    return res;
                }
                startIndex = 0;
            }

            return SearchResult.NotFound;
        }

        public static SearchResult FindPrev(TextDocument document, TextPosition startPosition, string searchText, bool matchCase, bool wholeWordsOnly)
        {
            IEnumerable<SearchResult> results;
            int pageIndex = document.IndexOf(startPosition.TextPage);
            results = FindAll(document.GetTextPage(pageIndex), searchText, matchCase, wholeWordsOnly);
            SearchResult res = Enumerable.LastOrDefault(results, r => r.Range.EndPosition <= startPosition);

            if (res != null)
            {
                return res;
            }

            for (int i = pageIndex - 1; i >= 0; i--)
            {
                results = FindAll(document.GetTextPage(i), searchText, matchCase, wholeWordsOnly);
                if (Enumerable.Count(results) > 0)
                {
                    return Enumerable.Last(results);
                }
            }

            return SearchResult.NotFound;
        }
    }
}
