﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.Text;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Search
{
    internal class TextSearch
    {
        private readonly TextDocument document;
        private readonly TextPosition startPosition;

        private TextSearchOptions options;
        private string searchText;

        public TextSearch(TextDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");

            this.document = document;
            this.startPosition = new TextPosition(this.document);
        }

        public SearchResult Find(string text, TextSearchOptions options)
        {
            Guard.ThrowExceptionIfNullOrEmpty(text, "text");
            Guard.ThrowExceptionIfNull(options, "options");

            this.InitFind(text, options, true);

            SearchResult res;
            if (options.UseRegularExpression)
            {
                res = RegexTextDocumentSearch.FindNext(this.document, this.startPosition, text, options.CaseSensitive, options.WholeWordsOnly);
            }
            else
            {
                res = KMPTextSearch.FindNext(this.document, this.startPosition, text, options.CaseSensitive, options.WholeWordsOnly);
            }

            if (res == SearchResult.NotFound)
            {
                this.startPosition.MoveToStartOfDocument();
            }
            else
            {
                this.startPosition.MoveToPosition(res.Range.StartPosition);
                this.startPosition.MoveToNextPosition();
            }

            return res;
        }

        public SearchResult FindPrevious(string text, TextSearchOptions options)
        {
            this.InitFind(text, options, false);

            SearchResult res;
            if (options.UseRegularExpression)
            {
                res = RegexTextDocumentSearch.FindPrev(this.document, this.startPosition, text, options.CaseSensitive, options.WholeWordsOnly);
            }
            else
            {
                res = KMPTextSearch.FindPrev(this.document, this.startPosition, text, options.CaseSensitive, options.WholeWordsOnly);
            }
            if (res == SearchResult.NotFound)
            {
                this.startPosition.MoveToEndOfDocument();
            }
            else
            {
                this.startPosition.MoveToPosition(res.Range.EndPosition);
                this.startPosition.MoveToPreviousPosition();
            }
            return res;
        }

        public IEnumerable<SearchResult> FindAll(string text, TextSearchOptions options)
        {
            if (options.UseRegularExpression)
            {
                return RegexTextDocumentSearch.FindAll(this.document, text, options.CaseSensitive, options.WholeWordsOnly);
            }
            else
            {
                return KMPTextSearch.FindAll(this.document, text, options.CaseSensitive, options.WholeWordsOnly);
            }
        }

        internal static int GetStartIndexOffset(int startIndex, TextPage page, string separator)
        {
            int offset = 0;
            foreach (Line line in page.Lines)
            {
                if (line.FirstIndex <= startIndex && startIndex < line.LastIndex)
                {
                    break;
                }

                if (line.LastWord != null && line.LastWord.EndsWithSpace)
                {
                    offset += separator.Length;
                }
                else
                {
                    offset += separator.Length - 1;
                }
            }

            return offset;
        }

        internal static int GetFoundIndexOffset(int foundIndex, TextPage page, string separator)
        {
            int offset = 0;
            foreach (Line line in page.Lines)
            {
                if (offset + line.LastIndex + separator.Length >= foundIndex)
                {
                    break;
                }

                if (line.LastWord.EndsWithSpace)
                {
                    offset += separator.Length;
                }
                else
                {
                    offset += separator.Length - 1; // the space at the end is missing but there is a position index, hence the -1
                }
            }

            return offset;
        }

        private void InitFind(string text, TextSearchOptions options, bool forward)
        {
            Guard.ThrowExceptionIfNull(text, "text");
            Guard.ThrowExceptionIfNull(options, "options");

            if (this.options != options)
            {
                this.options = options;
            }

            if (this.searchText != text)
            {
                this.searchText = text;
                if (forward)
                {
                    this.startPosition.MoveToStartOfDocument();
                }
                else
                {
                    this.startPosition.MoveToEndOfDocument();
                }
            }
        }
    }
}
