﻿using System.ComponentModel;

namespace Telerik.Windows.Documents.Fixed.Search
{
    /// <summary>
    /// Represent text search options.
    /// </summary>
    public class TextSearchOptions : INotifyPropertyChanged
    {
        #region Consts

        internal static readonly string TextSearchLinesSeparator = " ";

        #endregion


        #region StaticMembers

        /// <summary>
        /// Gets the default text search options.
        /// </summary>
        /// <value>The default.</value>
        public static TextSearchOptions Default
        {
            get
            {
                return new TextSearchOptions(false, false, false);
            }
        }

        #endregion


        #region Fields

        private bool useRegularExpression;
        private bool caseSensitive;
        private bool wholeWordsOnly;

        #endregion


        #region Events

        /// <summary>
        /// Occurs when a property value changes.
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        #endregion


        #region Constructors
        /// <summary>
        /// Initializes a new instance of the <see cref="TextSearchOptions" /> class.
        /// </summary>
        public TextSearchOptions()
            : this(false)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TextSearchOptions" /> class.
        /// </summary>
        /// <param name="caseSensitive">The case sensitive option.</param>
        public TextSearchOptions(bool caseSensitive)
            : this(caseSensitive, false)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TextSearchOptions" /> class.
        /// </summary>
        /// <param name="caseSensitive">The case sensitive.</param>
        /// <param name="useRegularExpression">The use regular expression.</param>
        public TextSearchOptions(bool caseSensitive, bool useRegularExpression)
            : this(caseSensitive, useRegularExpression, false)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TextSearchOptions" /> class.
        /// </summary>
        /// <param name="caseSensitive">The case sensitive.</param>
        /// <param name="useRegularExpression">The use regular expression.</param>
        /// <param name="wholeWordsOnly">The whole words only.</param>
        public TextSearchOptions(bool caseSensitive, bool useRegularExpression, bool wholeWordsOnly)
        {
            this.CaseSensitive = caseSensitive;
            this.UseRegularExpression = useRegularExpression;
            this.WholeWordsOnly = wholeWordsOnly;
        }

        #endregion


        #region Properties

        /// <summary>
        /// Gets or sets the use regular expression option.
        /// </summary>
        /// <value>The use regular expression.</value>
        public bool UseRegularExpression
        {
            get
            {
                return this.useRegularExpression;
            }
            set
            {
                if (value != this.useRegularExpression)
                {
                    this.useRegularExpression = value;
                    this.OnPropertyChanged("UseRegularExpression");
                }
            }
        }

        /// <summary>
        /// Gets or sets the case sensitive option.
        /// </summary>
        /// <value>The case sensitive.</value>
        public bool CaseSensitive
        {
            get
            {
                return this.caseSensitive;
            }
            set
            {
                if (value != this.caseSensitive)
                {
                    this.caseSensitive = value;
                    this.OnPropertyChanged("CaseSensitive");
                }
            }
        }

        /// <summary>
        /// Gets or sets the whole words only.
        /// </summary>
        /// <value>The whole words only.</value>
        public bool WholeWordsOnly
        {
            get
            {
                return this.wholeWordsOnly;
            }
            set
            {
                if (value != this.wholeWordsOnly)
                {
                    this.wholeWordsOnly = value;
                    this.OnPropertyChanged("WholeWordsOnly");
                }
            }
        }

        #endregion


        #region Methods

        private void OnPropertyChanged(string propertyName)
        {
            if (this.PropertyChanged != null)
            {
                this.PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
            }
        }

        #endregion

    }
}
