﻿using System;
using System.Threading;
using System.Threading.Tasks;
#if !NETSTANDARD
using System.ComponentModel;
using System.Windows;
using System.Windows.Media;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.Utils;
#else
using Telerik.Documents.Primitives;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
#endif
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Text;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Selection
{
    /// <summary>
    /// Represents text selection.
    /// </summary>
    public class TextSelection
    {
        private readonly TextDocument document;

        private bool isRunning;

        internal TextSelection(TextDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");
            this.document = document;

            this.StartPosition = null;
            this.EndPosition = null;
        }

        /// <summary>
        /// Occurs when selection is changing.
        /// </summary>
        public event EventHandler SelectionChanging;

        /// <summary>
        /// Occurs when selection is changed.
        /// </summary>
        public event EventHandler SelectionChanged;

        /// <summary>
        /// Gets the start position of the selection.
        /// </summary>
        /// <value>The start position.</value>
        public TextPosition StartPosition { get; private set; }

        /// <summary>
        /// Gets the end position of the selection.
        /// </summary>
        /// <value>The end position.</value>
        public TextPosition EndPosition { get; private set; }

        /// <summary>
        /// Checks if the selection is empty.
        /// </summary>
        /// <value>The is empty.</value>
        public bool IsEmpty
        {
            get
            {
                return this.StartPosition == this.EndPosition;
            }
        }

        /// <summary>
        /// Gets the selection geometry for given page.
        /// </summary>
        /// <param name="page">The page.</param>
        /// <returns></returns>
        public PathGeometry GetSelectionGeometry(RadFixedPage page)
        {
            TextPage textPage = this.document.GetTextPage(page.PageNumber - 1);

            return this.GetSelectionGeometry(textPage);
        }

        /// <summary>
        /// Sets the selection start.
        /// </summary>
        /// <param name="startPosition">The start position.</param>
        public void SetSelectionStart(TextPosition startPosition)
        {
            Guard.ThrowExceptionIfNull(startPosition, "startPosition");

            TextPosition endPosition = this.EndPosition == null ? startPosition : this.EndPosition;
            this.SetSelection(startPosition, endPosition);
        }

        /// <summary>
        /// Sets the selection end.
        /// </summary>
        /// <param name="endPosition">The end position.</param>
        public void SetSelectionEnd(TextPosition endPosition)
        {
            Guard.ThrowExceptionIfNull(endPosition, "endPosition");

            TextPosition startPosition = this.StartPosition == null ? endPosition : this.StartPosition;
            this.SetSelection(startPosition, endPosition);
        }

        /// <summary>
        /// Sets the selection.
        /// </summary>
        /// <param name="startPosition">The start position.</param>
        /// <param name="endPosition">The end position.</param>
        public void SetSelection(TextPosition startPosition, TextPosition endPosition)
        {
            Guard.ThrowExceptionIfNull(startPosition, "startPosition");
            Guard.ThrowExceptionIfNull(endPosition, "endPosition");

            if (startPosition.Document != endPosition.Document || endPosition.Document != this.document)
            {
                throw new ArgumentException("The start and the end position should be from the same document as the selection.");
            }

            this.OnSelectionChanging();

            this.StartPosition = new TextPosition(startPosition);
            this.EndPosition = new TextPosition(endPosition);

            this.OnSelectionChanged();
        }

        /// <summary>
        /// Selects the whole document.
        /// </summary>
        public void SelectAll()
        {
            TextPosition start = new TextPosition(this.document);
            start.MoveToStartOfDocument();

            TextPosition end = new TextPosition(this.document);
            end.MoveToEndOfDocument();

            this.SetSelection(start, end);
        }

        /// <summary>
        /// Clears the selection.
        /// </summary>
        public void Clear()
        {
            if (!this.IsEmpty)
            {
                this.OnSelectionChanging();

                this.StartPosition = null;
                this.EndPosition = null;

                this.OnSelectionChanged();
            }
        }

        /// <summary>
        /// Gets the selected text.
        /// </summary>
        /// <returns></returns>
        public string GetSelectedText()
        {
            if (this.IsEmpty)
            {
                return string.Empty;
            }

            return TextDocument.GetText(this.StartPosition, this.EndPosition, Line.LineSeparator);
        }

#if !NETSTANDARD
        /// <summary>
        /// Gets the selected text asynchronously. The callback is called on the Dispatcher.
        /// </summary>
        /// <param name="callback">The callback.</param>
        [EditorBrowsable(EditorBrowsableState.Never)]
        [Obsolete("Use GetSelectedTextAsync() method instead.")]
        public void GetSelectedTextAsync(Action<string> callback)
        {
            if (!this.isRunning)
            {
                this.isRunning = true;
                Thread th = new Thread(new ParameterizedThreadStart(this.GetSelectedText));
                th.Start(callback);
            }
        }
#endif

        /// <summary>
        /// Gets the selected text async.
        /// </summary>
        public Task<string> GetSelectedTextAsync()
        {
            if (!this.isRunning)
            {
                this.isRunning = true;
                Task<string> task = Task.Factory.StartNew<string>(this.GetSelectedText);

                return task.ContinueWith<string>(
                    new Func<Task<string>, string>((t) =>
                    {
                        this.isRunning = false;
                        return t.Result;
                    }));
            }
            else
            {
                return Task.Factory.StartNew<string>(() => string.Empty);
            }
        }

        /// <summary>
        /// Returns a string that represents the current object.
        /// </summary>
        /// <returns>A string that represents the current object.</returns>
        public override string ToString()
        {
            return this.GetSelectedText();
        }

        internal PathGeometry GetSelectionGeometry(TextPage page)
        {
            Guard.ThrowExceptionIfNull(page, "page");

            if (this.IsEmpty)
            {
                return new PathGeometry();
            }

            TextPosition startPosition;
            TextPosition endPosition;
            if (this.StartPosition < this.EndPosition)
            {
                startPosition = this.StartPosition;
                endPosition = this.EndPosition;
            }
            else
            {
                startPosition = this.EndPosition;
                endPosition = this.StartPosition;
            }

            TextPosition start = new TextPosition(page);
            TextPosition end = new TextPosition(page, page.LastIndex);
            if (startPosition > end || endPosition < start)
            {
                return new PathGeometry();
            }

            if (start < startPosition)
            {
                start = new TextPosition(startPosition);
            }

            if (end > endPosition)
            {
                end = endPosition;
            }

            startPosition = new TextPosition(start);
            endPosition = new TextPosition(end);

            PathGeometry res = new PathGeometry();
            res.FillRule = FillRule.Nonzero;

            Line firstLine = start.GetAssociatedLine();
            Line lastLine = end.GetAssociatedLine();

            Rect rect;
            if (firstLine != null && firstLine == lastLine)
            {
                rect = new Rect(new Point(startPosition.Location.X, firstLine.BoundingRect.Top), new Point(endPosition.Location.X, firstLine.BoundingRect.Bottom));
                res.Figures.Add(CreateSelectionRectangleFromBoundingRect(rect));

                return res;
            }

            Line line = null;
            do
            {
                if (line != null)
                {
                    res.Figures.Add(CreateSelectionRectangleFromBoundingRect(line.BoundingRect));
                }

                if (start.MoveToStartOfTheNextLine())
                {
                    line = start.GetAssociatedLine();
                }
                else
                {
                    return new PathGeometry();
                }
            }
            while (line != lastLine);

            if (firstLine != null)
            {
                rect = new Rect(new Point(startPosition.Location.X, firstLine.BoundingRect.Top), new Point(firstLine.BoundingRect.Right, firstLine.BoundingRect.Bottom));
                res.Figures.Add(CreateSelectionRectangleFromBoundingRect(rect));
            }

            if (lastLine != null)
            {
                rect = new Rect(new Point(lastLine.BoundingRect.Left, lastLine.BoundingRect.Top), new Point(endPosition.Location.X, lastLine.BoundingRect.Bottom));
                res.Figures.Add(CreateSelectionRectangleFromBoundingRect(rect));
            }

            return res;
        }

        internal bool ContainsPosition(TextPosition pos)
        {
            if (this.StartPosition == null || this.EndPosition == null)
            {
                return false;
            }

            return this.StartPosition < pos && pos < this.EndPosition;
        }

        private static PathFigure CreateSelectionRectangleFromBoundingRect(Rect boundingRect)
        {
            if (boundingRect.IsEmpty)
            {
                return new PathFigure();
            }

            PathFigure figure = new PathFigure();
            figure.IsClosed = true;
            figure.StartPoint = new Point(boundingRect.Left, boundingRect.Top);

#if !NETSTANDARD
            PolyLineSegment segment = new PolyLineSegment();
            segment.Points.Add(new Point(boundingRect.Right, boundingRect.Top));
            segment.Points.Add(new Point(boundingRect.Right, boundingRect.Bottom));
            segment.Points.Add(new Point(boundingRect.Left, boundingRect.Bottom));

            figure.Segments.Add(segment);
            figure.IsFilled = true;
#else
            figure.Segments.AddLineSegment(new Point(boundingRect.Right, boundingRect.Top));
            figure.Segments.AddLineSegment(new Point(boundingRect.Right, boundingRect.Bottom));
            figure.Segments.AddLineSegment(new Point(boundingRect.Left, boundingRect.Bottom));
#endif

            return figure;
        }

        private void OnSelectionChanging()
        {
            if (this.SelectionChanging != null)
            {
                this.SelectionChanging(this, EventArgs.Empty);
            }
        }

        private void OnSelectionChanged()
        {
            if (this.SelectionChanged != null)
            {
                this.SelectionChanged(this, EventArgs.Empty);
            }
        }

#if !NETSTANDARD
        private void GetSelectedText(object param)
        {
            Action<string> callback = param as Action<string>;
            if (callback == null)
            {
                return;
            }

            string selectedText = this.GetSelectedText();
            if (Helper.Dispatcher != null)
            {
                Helper.Dispatcher.BeginInvoke(new Action(() =>
                {
                    callback(selectedText);
                }));
            }
            else
            {
                callback(selectedText);
            }

            this.isRunning = false;
        }
#endif
    }
}
