﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
#if !NETSTANDARD
using System.Windows;
#else
using Telerik.Documents.Primitives;
#endif
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Text
{
    internal class Line
    {
        internal static readonly string WordsSeparator = " ";
        internal static readonly string LineSeparator = Environment.NewLine;
        internal static readonly double Delta = 0.00001;

        private readonly List<Word> words;
        private string value;

        public Line()
        {
            this.words = new List<Word>();
        }

        internal int WordsCount
        {
            get
            {
                return this.words.Count;
            }
        }

        internal int FirstIndex
        {
            get;
            private set;
        }

        internal int LastIndex
        {
            get;
            private set;
        }

        internal int Length
        {
            get;
            private set;
        }

        internal bool IsEmpty
        {
            get
            {
                return this.WordsCount == 0;
            }
        }

        internal Word FirstWord
        {
            get
            {
                return Enumerable.FirstOrDefault(this.words);
            }
        }

        internal Word LastWord
        {
            get
            {
                return Enumerable.LastOrDefault(this.words);
            }
        }

        internal Rect BoundingRect
        {
            get;
            private set;
        }

        public static int CompareWithCharacterIndex(Line line, int index)
        {
            if (line.FirstIndex <= index && index <= line.LastIndex)
            {
                return 0;
            }
            if (index > line.LastIndex)
            {
                return 1;
            }
            return -1;
        }

        private void CalculateValue()
        {
            StringBuilder sb = new StringBuilder();
            foreach (Word w in this.words)
            {
                sb.Append(w.ToString());
                if (w.EndsWithSpace)
                {
                    sb.Append(WordsSeparator);
                }
            }
            this.value = sb.ToString();
        }

        private void CalculateBoundingRect()
        {
            Rect rect;
            double minX, minY, maxX, maxY, height;
            minX = minY = double.MaxValue;
            maxX = maxY = height = double.MinValue;
            for (int i = 0; i < this.words.Count; i++)
            {
                {
                    rect = this.words[i].BoundingRect;
                    minX = Math.Min(rect.Left, minX);
                    minY = Math.Min(rect.Top, minY);
                    maxX = Math.Max(rect.Right, maxX);
                    maxY = Math.Max(rect.Bottom, maxY);
                    height = Math.Max(rect.Height, height);
                }
            }
            this.BoundingRect = new Rect(new Point(minX, minY), new Point(maxX, maxY));
        }

        internal Word GetWordFromIndex(int index)
        {
            return CollectionsExtensions.FindElement(this.words, index, Word.CompareWithCharacterIndex);
        }

        internal Word GetWordAt(int index)
        {
            return this.words[index];
        }

        internal void AddWord(Word word)
        {
            if (word.IsEmpty)
            {
                return;
            }
            this.words.Add(word);
        }

        internal void Finish()
        {
            if (this.IsEmpty)
            {
                return;
            }
            this.Length = Enumerable.Sum(this.words, w => w.Length);
            this.FirstIndex = this[0].FirstIndex;
            this.LastIndex = this[this.WordsCount - 1].LastIndex;
            this.CalculateBoundingRect();
            this.CalculateValue();
        }

        internal void FinishEmpty(Rect boundingRect)
        {
            if (!this.IsEmpty)
            {
                return;
            }
            this.Length = 0;
            this.FirstIndex = 0;
            this.LastIndex = 0;
            this.BoundingRect = boundingRect;
            this.value = "";
        }

        internal Word this[int index]
        {
            get
            {
                return this.words[index];
            }
        }

        internal int IndexOf(Word word)
        {
            return this.words.IndexOf(word);
        }

        internal IEnumerator<Word> GetEnumerator()
        {
            return this.words.GetEnumerator();
        }

        internal int GetIndexFromLocation(double location)
        {
            if (this.IsEmpty)
            {
                return 0;
            }
            Word w = CollectionsExtensions.FindElement(this.words, location, Word.CompareWithWordBoundingRect);
            return w.GetIndexFromLocation(location);
        }

        public override string ToString()
        {
            return this.value;
        }

        public static bool operator ==(Line left, Line right)
        {
            if (object.ReferenceEquals(left, null))
            {
                return object.ReferenceEquals(right, null);
            }

            return left.Equals(right);
        }

        public static bool operator !=(Line left, Line right)
        {
            if (object.ReferenceEquals(left, null))
            {
                return !object.ReferenceEquals(right, null);
            }

            return !left.Equals(right);
        }

        public override bool Equals(object obj)
        {
            Line other = obj as Line;
            if (other == null)
            {
                return false;
            }

            return
                this.FirstIndex.Equals(other.FirstIndex) &&
                this.LastIndex.Equals(other.LastIndex) &&
                this.value.Equals(other.value);
        }

        private bool WordsAreEqual(Line other)
        {
            for (int i = 0; i < this.words.Count; i++)
            {
                if (!this.words[i].Equals(other.words[i]))
                {
                    return false;
                }
            }

            return true;
        }

        public override int GetHashCode()
        {
            unchecked
            {
                int result = Telerik.Windows.Documents.Utilities.ObjectExtensions.CombineHashCodes(
                    this.FirstIndex.GetHashCode(),
                    this.LastIndex.GetHashCode());

                return result;
            }
        }
    }
}
