﻿using System.Collections.Generic;
using System.Text;
using Telerik.Windows.Documents.Fixed.Model;

namespace Telerik.Windows.Documents.Fixed.Text
{
    internal class TextDocument
    {
        private readonly RadFixedDocument fixedDocument;
        private readonly TextPage[] textPages;
        private readonly TextRecognizer textRecognizer;

        private readonly Dictionary<int, TextPage> pageIndexToTextPage;

        private int pagesCount;

        // This constructor will be used for tests only;
        internal TextDocument(int count)
        {
            this.textPages = new TextPage[count];
        }

        public TextDocument(RadFixedDocument document)
            : this()
        {
            this.fixedDocument = document;
            this.PagesCount = this.fixedDocument.Pages.Count;
        }

        protected TextDocument()
        {
            this.pageIndexToTextPage = new Dictionary<int, TextPage>();
            this.textRecognizer = new TextRecognizer(this);
        }

        internal int PagesCount
        {
            get
            {
                //Used only for test purposses
                if (this.textPages != null)
                {
                    return this.textPages.Length;
                }
                ////

                return this.pagesCount;
            }
            set
            {
                //TODO consider whether to make base class with abstract getter
                this.pagesCount = value;
            }
        }

        internal TextPage FirstPage
        {
            get
            {
                return this.GetTextPage(0);
            }
        }

        internal TextPage LastPage
        {
            get
            {
                return this.GetTextPage(this.PagesCount - 1);
            }
        }

        internal bool IsPdfProcessing
        {
            get
            {
                return this.fixedDocument != null;
            }
        }

        internal static string GetText(TextPosition startPosition, TextPosition endPosition, string lineSeparator)
        {
            TextPosition start;
            TextPosition end;

            if (startPosition < endPosition)
            {
                start = startPosition;
                end = endPosition;
            }
            else
            {
                start = endPosition;
                end = startPosition;
            }

            Line firstLine = start.GetAssociatedLine();
            Line lastLine = end.GetAssociatedLine();

            string str;
            if (firstLine == lastLine)
            {
                str = firstLine.ToString();
                return str.Substring(start.Index - firstLine.FirstIndex, end.Index - start.Index);
            }

            StringBuilder res = new StringBuilder();

            TextPosition textPosition = new TextPosition(start);
            str = firstLine.ToString();
            res.Append(str.Substring(start.Index - firstLine.FirstIndex));
            res.Append(lineSeparator);

            Line line = null;
            do
            {
                if (line != null)
                {
                    res.Append(line.ToString());
                    res.Append(lineSeparator);
                }

                bool moveToNextLine = textPosition.MoveToStartOfTheNextLine();
                if (moveToNextLine)
                {
                    line = textPosition.GetAssociatedLine();
                }
                else
                {
                    return string.Empty;
                }

            } while (!(line == lastLine && textPosition.Page.PageNumber == end.Page.PageNumber));

            str = lastLine.ToString();
            res.Append(str.Substring(0, end.Index - lastLine.FirstIndex));

            return res.ToString();
        }

        internal static char GetCharBeforePosition(TextPosition position)
        {
            if (position.Index == 0)
            {
                return ' ';
            }
            Line line = position.GetAssociatedLine();
            if (position.Index == line.FirstIndex)
            {
                return ' ';
            }
            return line.ToString()[position.Index - line.FirstIndex - 1];
        }

        internal static char GetCharAfterPosition(TextPosition position)
        {
            if (position.Index == position.TextPage.LastIndex)
            {
                return ' ';
            }
            Line line = position.GetAssociatedLine();
            if (position.Index == line.LastIndex)
            {
                return ' ';
            }
            return line.ToString()[position.Index - line.FirstIndex];
        }

        internal TextPage GetTextPage(int index)
        {
            if (index < 0 || index >= this.PagesCount)
            {
                return null;
            }

            TextPage result = null;

            // this code will be used for tests only;
            if (this.textPages != null)
            {
                result = this.textPages[index];
            }

            if (result == null)
            {
                result = this.CreateTextPage(index);

                // this code will be used for tests only;
                if (this.textPages != null)
                {
                    this.textPages[index] = result;
                }
            }

            return result;
        }

        // This methods will be used for tests only;
        internal void AddPage(int index, TextPage page)
        {
            this.textPages[index] = page;
        }

        internal int IndexOf(TextPage page)
        {
            int pageIndex = page.PageNumber - 1;

            return pageIndex;
        }

        internal TextPage GetNextPage(TextPage page)
        {
            int index = this.IndexOf(page);

            return this.GetTextPage(index + 1);
        }

        internal TextPage GetPrevPage(TextPage page)
        {
            int index = this.IndexOf(page);

            return this.GetTextPage(index - 1);
        }

        protected virtual TextPage CreateTextPage(int index)
        {
            TextPage textPage;

            if (!this.pageIndexToTextPage.TryGetValue(index, out textPage))
            {
                textPage = this.textRecognizer.CreateTextPage(this.fixedDocument.Pages[index]);

                if (textPage.Length > 0)
                {
                    this.pageIndexToTextPage.Add(index, textPage);
                }
            }

            return textPage;
        }

        public string ToString(string linesSeparator, string pagesSeparator)
        {
            StringBuilder sb = new StringBuilder();
            bool containsFormat = pagesSeparator.Contains("{0}");
            for (int i = 0; i < this.PagesCount; i++)
            {
                if (containsFormat)
                {
                    sb.Append(string.Format(pagesSeparator, i + 1));
                }
                else
                {
                    sb.Append(pagesSeparator);
                }
                sb.Append(linesSeparator);
                sb.Append(this.GetTextPage(i).ToString(linesSeparator));
            }
            return sb.ToString();
        }
    }
}
