﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
#if !NETSTANDARD
using System.Windows;
#else
using Telerik.Documents.Primitives;
#endif
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Text
{
    internal class TextPage
    {
        private const double Delta = 0.00001;
        private const double BoundingRectOffset = 15;

        private readonly TextDocument parent;
        private readonly RadFixedPage page;
        private readonly List<Line> lines;

        public TextPage(TextDocument parent, RadFixedPage fixedPage)
        {
            Guard.ThrowExceptionIfNull(parent, "parent");
            Guard.ThrowExceptionIfNull(fixedPage, "fixedPage");

            this.page = fixedPage;
            this.parent = parent;
            this.lines = new List<Line>();
        }

        public RadFixedPage Page
        {
            get
            {
                return this.page;
            }
        }

        public int PageNumber
        {
            get
            {
                return this.Page.PageNumber;
            }
        }

        public TextDocument Parent
        {
            get
            {
                return this.parent;
            }
        }

        public IEnumerable<Line> Lines
        {
            get
            {
                return this.lines;
            }
        }

        public int LinesCount
        {
            get
            {
                return this.lines.Count;
            }
        }

        public int Length
        {
            get;
            private set;
        }

        public int LastIndex
        {
            get;
            private set;
        }

        public Line LastLine
        {
            get
            {
                return Enumerable.LastOrDefault(this.lines);
            }
        }

        public static bool operator ==(TextPage left, TextPage right)
        {
            if (object.ReferenceEquals(left, null))
            {
                return object.ReferenceEquals(right, null);
            }

            if (object.ReferenceEquals(left, right))
            {
                return true;
            }

            return left.Equals(right);
        }

        public static bool operator !=(TextPage left, TextPage right)
        {
            if (object.ReferenceEquals(left, null))
            {
                return !object.ReferenceEquals(right, null);
            }

            return !left.Equals(right);
        }

        public static bool operator <(TextPage left, TextPage right)
        {
            Guard.ThrowExceptionIfNull(left, "left");
            Guard.ThrowExceptionIfNull(right, "right");

            if (left.Parent != right.Parent)
            {
                throw new ArgumentException("The text pages should be from the same TextDocument.");
            }

            return left.PageNumber < right.PageNumber;
        }

        public static bool operator >(TextPage left, TextPage right)
        {
            Guard.ThrowExceptionIfNull(left, "left");
            Guard.ThrowExceptionIfNull(right, "right");

            if (left.Parent != right.Parent)
            {
                throw new ArgumentException("The text pages should be from the same TextDocument.");
            }

            return left.PageNumber > right.PageNumber;
        }

        public override bool Equals(object obj)
        {
            TextPage other = obj as TextPage;
            if (other == null)
            {
                return false;
            }

            if (this.Parent != other.Parent)
            {
                throw new ArgumentException("The text pages should be from the same TextDocument.");
            }

            return this.PageNumber == other.PageNumber;
        }

        public override int GetHashCode()
        {
            unchecked
            {
                int result = ObjectExtensions.CombineHashCodes(
                    this.PageNumber.GetHashCode(), 
                    this.Page.GetHashCode());

                return result;
            }
        }

        public void Finish()
        {
            if (this.LinesCount > 0)
            {
                this.Length = Enumerable.Sum(this.Lines, l => l.Length);
                this.LastIndex = Enumerable.Last(this.lines).LastIndex;
            }
            else
            {
                Line empty = new Line();
                empty.FinishEmpty(new Rect(0, 0, 0, 0));
                this.lines.Add(empty);
            }
        }

        public void AddLine(Line line)
        {
            Guard.ThrowExceptionIfNull(line, "line");

            if (!line.IsEmpty)
            {
                this.lines.Add(line);
            }
        }

        public int IndexOfLine(Line line)
        {
            Guard.ThrowExceptionIfNull(line, "line");

            return this.lines.IndexOf(line);
        }

        public Line GetLineAt(int index)
        {
            return this.lines[index];
        }

        public Line GetLineFromIndex(int index)
        {
            return CollectionsExtensions.FindElement(this.lines, index, Line.CompareWithCharacterIndex);
        }

        public Word GetWordFromIndex(int index)
        {
            Line line = this.GetLineFromIndex(index);
            if (line == null)
            {
                return null;
            }

            return line.GetWordFromIndex(index);
        }

        public WordPosition GetWordPositionFromIndex(int index)
        {
            Word word = this.GetWordFromIndex(index);
            if (word == null)
            {
                return null;
            }

            return word.GetWordPositionFromIndex(index);
        }

        public Character GetCharacterFromIndex(int index)
        {
            Word word = this.GetWordFromIndex(index);
            if (word == null)
            {
                return null;
            }

            return word.GetCharacterFromIndex(index);
        }

        public Line GetNearestLine(Point location)
        {
            double minDistX, minDistY, distX, distY;
            minDistX = minDistY = double.MaxValue;
            Line res = null;
            Rect boundingRect;
            foreach (Line line in this.lines)
            {
                if (line.IsEmpty)
                {
                    continue;
                }

                boundingRect = line.BoundingRect;
                if (boundingRect.Contains(location))
                {
                    distX = CalcXDistanceToRect(boundingRect, location);
                    distY = CalcYDistanceToRect(boundingRect, location);

                    if (distY < minDistY)
                    {
                        minDistY = distY;
                        res = line;
                    }
                    else if (Math.Abs(minDistY - distY) < Delta && distX < minDistX)
                    {
                        minDistY = distY;
                        minDistX = distX;
                        res = line;
                    }
                }
            }

            if (res != null)
            {
                return res;
            }

            foreach (Line line in this.lines)
            {
                if (line.IsEmpty)
                {
                    continue;
                }

                boundingRect = new Rect(
                    line.BoundingRect.X - BoundingRectOffset,
                    line.BoundingRect.Y - BoundingRectOffset,
                    line.BoundingRect.Width + (2 * BoundingRectOffset),
                    line.BoundingRect.Height + (2 * BoundingRectOffset));

                if (boundingRect.Contains(location))
                {
                    distX = CalcXDistanceToCenter(boundingRect, location);
                    distY = CalcYDistanceToCenter(boundingRect, location);

                    if (distY < minDistY)
                    {
                        minDistY = distY;
                        res = line;
                    }
                    else if (Math.Abs(minDistY - distY) < Delta && distX < minDistX)
                    {
                        minDistY = distY;
                        minDistX = distX;
                        res = line;
                    }
                }
            }

            return res;
        }

        public string ToString(string linesSeparator)
        {
            StringBuilder sb = new StringBuilder();
            int last = this.lines.Count - 1;
            for (int i = 0; i <= last; i++)
            {
                sb.Append(this.lines[i].ToString());

                if (i != last)
                {
                    sb.Append(linesSeparator);
                }
            }

            return sb.ToString();
        }

        private static double CalcXDistanceToRect(Rect rect, Point p)
        {
            return Math.Min(Math.Abs(rect.Left - p.X), Math.Abs(rect.Right - p.X));
        }

        private static double CalcYDistanceToRect(Rect rect, Point p)
        {
            return Math.Min(Math.Abs(rect.Top - p.Y), Math.Abs(rect.Bottom - p.Y));
        }

        private static double CalcXDistanceToCenter(Rect rect, Point p)
        {
            return Math.Abs(AlgebraExtensions.Center(rect).X - p.X);
        }

        private static double CalcYDistanceToCenter(Rect rect, Point p)
        {
            return Math.Abs(AlgebraExtensions.Center(rect).Y - p.Y);
        }
    }
}
