﻿using System;
using System.Linq;
#if !NETSTANDARD
using Point = System.Windows.Point;
#else
using Point = Telerik.Documents.Primitives.Point;
#endif
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Text
{
    /// <summary>
    /// Represents text position.
    /// </summary>
    public class TextPosition
    {
        private int beginUpdateCount;
        private double lastTrackedLocation;

        /// <summary>
        /// Initializes a new instance of the <see cref="TextPosition" /> class.
        /// </summary>
        /// <param name="document">The document.</param>
        public TextPosition(RadFixedDocument document)
            : this(document.TextDocument)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TextPosition" /> class.
        /// </summary>
        /// <param name="position">The position.</param>
        public TextPosition(TextPosition position)
            : this(position.TextPage, position.Index)
        {
            this.lastTrackedLocation = position.lastTrackedLocation;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TextPosition" /> class.
        /// </summary>
        /// <param name="page">The page.</param>
        public TextPosition(RadFixedPage page)
            : this(TextPosition.GetTextPage(page))
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TextPosition" /> class.
        /// </summary>
        /// <param name="page">The page.</param>
        /// <param name="index">The index.</param>
        public TextPosition(RadFixedPage page, int index)
            : this(TextPosition.GetTextPage(page), index)
        {
        }

        internal TextPosition(TextPage page, int index)
        {
            Guard.ThrowExceptionIfNull(page, "page");

            if (index < 0 || page.LastIndex < index)
            {
                throw new ArgumentOutOfRangeException("index");
            }

            this.TextPage = page;
            this.Index = index;
        }

        internal TextPosition(TextDocument document)
            : this(document.FirstPage)
        {
        }

        internal TextPosition(TextPage page)
            : this(page, 0)
        {
        }

        /// <summary>
        /// Occurs when TextPosition is changing.
        /// </summary>
        public event EventHandler PositionChanging;

        /// <summary>
        /// Occurs when TextPosition is changed.
        /// </summary>
        public event EventHandler PositionChanged;

        /// <summary>
        /// Gets the index of the position.
        /// </summary>
        /// <value>The index.</value>
        public int Index { get; private set; }

        /// <summary>
        /// Gets the page that contains the position.
        /// </summary>
        /// <value>The page.</value>
        public RadFixedPage Page
        {
            get
            {
                return this.TextPage.Page;
            }
        }

        internal TextPage TextPage
        {
            get;
            private set;
        }

        internal TextDocument Document
        {
            get
            {
                return this.TextPage.Parent;
            }
        }

        internal Point Location
        {
            get
            {
                return this.GetLocation();
            }
        }

        /// <summary>
        /// Compares two text positions.
        /// </summary>
        /// <param name="left">Left text position.</param>
        /// <param name="right">Right text position.</param>
        /// <returns>Returns if the two positions are equal.</returns>
        public static bool operator ==(TextPosition left, TextPosition right)
        {
            if (object.ReferenceEquals(left, null))
            {
                return object.ReferenceEquals(right, null);
            }

            return left.Equals(right);
        }

        /// <summary>
        /// Compares two text positions.
        /// </summary>
        /// <param name="left">Left text position.</param>
        /// <param name="right">Right text position.</param>
        /// <returns>Returns if the two positions are not equal.</returns>
        public static bool operator !=(TextPosition left, TextPosition right)
        {
            if (object.ReferenceEquals(left, null))
            {
                return !object.ReferenceEquals(right, null);
            }

            return !left.Equals(right);
        }

        /// <summary>
        /// Compares two text positions.
        /// </summary>
        /// <param name="left">Left text position.</param>
        /// <param name="right">Right text position.</param>
        /// <returns>Returns if the left position is less than the right position.</returns>
        public static bool operator <(TextPosition left, TextPosition right)
        {
            Guard.ThrowExceptionIfNull(left, "left");
            Guard.ThrowExceptionIfNull(right, "right");

            if (left.Document != right.Document)
            {
                throw new ArgumentException("The two positions must be from the same TextDocument.");
            }

            if (left.TextPage == right.TextPage)
            {
                return left.Index < right.Index;
            }

            return left.TextPage < right.TextPage;
        }

        /// <summary>
        /// Compares two text positions.
        /// </summary>
        /// <param name="left">Left text position.</param>
        /// <param name="right">Right text position.</param>
        /// <returns>Returns if the left position is greater than the right position.</returns>
        public static bool operator >(TextPosition left, TextPosition right)
        {
            Guard.ThrowExceptionIfNull(left, "left");
            Guard.ThrowExceptionIfNull(right, "right");

            if (left.Document != right.Document)
            {
                throw new ArgumentException("The two positions must be from the same TextDocument.");
            }

            if (left.TextPage == right.TextPage)
            {
                return left.Index > right.Index;
            }

            return left.TextPage > right.TextPage;
        }

        /// <summary>
        /// Compares two text positions.
        /// </summary>
        /// <param name="left">Left text position.</param>
        /// <param name="right">Right text position.</param>
        /// <returns>Returns if the left position is less than or equals to the right position.</returns>
        public static bool operator <=(TextPosition left, TextPosition right)
        {
            Guard.ThrowExceptionIfNull(left, "left");
            Guard.ThrowExceptionIfNull(right, "right");

            if (left.Document != right.Document)
            {
                throw new ArgumentException("The two positions must be from the same TextDocument.");
            }

            return left < right || left == right;
        }

        /// <summary>
        /// Compares two text positions.
        /// </summary>
        /// <param name="left">Left text position.</param>
        /// <param name="right">Right text position.</param>
        /// <returns>Returns if the left position is greater than or equals to the right position.</returns>
        public static bool operator >=(TextPosition left, TextPosition right)
        {
            Guard.ThrowExceptionIfNull(left, "left");
            Guard.ThrowExceptionIfNull(right, "right");

            if (left.Document != right.Document)
            {
                throw new ArgumentException("The two positions must be from the same TextDocument.");
            }

            return left > right || left == right;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            TextPosition other = obj as TextPosition;
            if (other == null)
            {
                return false;
            }

            if (this.Document != other.Document)
            {
                throw new ArgumentException("The two positions must be from the same TextDocument.");
            }

            return this.Index == other.Index && this.TextPage == other.TextPage;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            int result = 17;
            result = (result * 23) + this.Index.GetHashCode();
            result = (result * 23) + ((TextPage != null) ? this.TextPage.GetHashCode() : 0);

            return result;
        }

        /// <summary>
        /// Moves to position.
        /// </summary>
        /// <param name="position">The position.</param>
        /// <returns></returns>
        public bool MoveToPosition(TextPosition position)
        {
            Guard.ThrowExceptionIfNull(position, "position");

            if (position.Document != this.Document)
            {
                throw new ArgumentException("Position should belong to the same document");
            }

            bool result = this.MoveToIndex(position.TextPage, position.Index);

            return result;
        }

        /// <summary>
        /// Moves to next position.
        /// </summary>
        /// <returns></returns>
        public bool MoveToNextPosition()
        {
            return this.ExecuteInsideBeginEndUpdate(() =>
            {
                int nextIndex = this.Index + 1;
                TextPage page = this.TextPage;
                if (nextIndex > this.TextPage.LastIndex)
                {
                    page = this.Document.GetNextPage(this.TextPage);
                    nextIndex = 0;
                }

                return this.MoveToIndex(page, nextIndex);
            });
        }

        /// <summary>
        /// Moves to previous position.
        /// </summary>
        /// <returns></returns>
        public bool MoveToPreviousPosition()
        {
            return this.ExecuteInsideBeginEndUpdate(() =>
            {
                int nextIndex = this.Index - 1;
                TextPage page = this.TextPage;
                if (nextIndex < 0)
                {
                    page = this.Document.GetPrevPage(this.TextPage);
                    if (page == null)
                    {
                        return false;
                    }

                    nextIndex = page.LastIndex;
                }

                return this.MoveToIndex(page, nextIndex);
            });
        }

        /// <summary>
        /// Moves to next word.
        /// </summary>
        /// <returns></returns>
        public bool MoveToNextWord()
        {
            return this.ExecuteInsideBeginEndUpdate(() =>
            {
                Line line = this.TextPage.GetLineFromIndex(this.Index);
                if (line == null)
                {
                    return false;
                }

                Word word = line.GetWordFromIndex(this.Index);
                if (word == null)
                {
                    return false;
                }

                int index = line.IndexOf(word);
                index++;
                Word res;
                Line nextLine;
                TextPage page = this.TextPage;
                if (index >= line.WordsCount)
                {
                    this.GetNextLine(line, out nextLine, out page);
                    if (nextLine == null)
                    {
                        return false;
                    }

                    res = nextLine.FirstWord;
                }
                else
                {
                    res = line[index];
                }

                return this.MoveToIndex(page, res.FirstIndex);
            });
        }

        /// <summary>
        /// Moves to previous word.
        /// </summary>
        /// <returns></returns>
        public bool MoveToPreviousWord()
        {
            return this.ExecuteInsideBeginEndUpdate(() =>
            {
                Line line = this.TextPage.GetLineFromIndex(this.Index);
                if (line == null)
                {
                    return false;
                }

                Word word = line.GetWordFromIndex(this.Index);
                if (word == null)
                {
                    return false;
                }

                if (word.FirstIndex < this.Index)
                {
                    return this.MoveToIndex(word.FirstIndex);
                }

                int index = line.IndexOf(word);
                index--;
                Word res;
                Line nextLine;
                TextPage page = this.TextPage;
                if (index < 0)
                {
                    this.GetPrevLine(line, out nextLine, out page);
                    if (nextLine == null)
                    {
                        return false;
                    }

                    res = nextLine.LastWord;
                }
                else
                {
                    res = line[index];
                }

                return this.MoveToIndex(page, res.FirstIndex);
            });
        }

        /// <summary>
        /// Moves to current word start.
        /// </summary>
        /// <returns></returns>
        public bool MoveToCurrentWordStart()
        {
            return this.ExecuteInsideBeginEndUpdate(() =>
            {
                Word word = this.TextPage.GetWordFromIndex(this.Index);
                if (word == null)
                {
                    return false;
                }

                return this.MoveToIndex(word.FirstIndex);
            });
        }

        /// <summary>
        /// Moves to current word end.
        /// </summary>
        /// <returns></returns>
        public bool MoveToCurrentWordEnd()
        {
            return this.ExecuteInsideBeginEndUpdate(() =>
            {
                Word word = this.TextPage.GetWordFromIndex(this.Index);
                if (word == null || word.LastIndex == this.Index)
                {
                    return false;
                }

                return this.MoveToIndex(word.LastIndex);
            });
        }

        /// <summary>
        /// Moves to line start.
        /// </summary>
        public bool MoveToLineStart()
        {
            Line line = this.GetAssociatedLine();
            if (line == null)
            {
                return false;
            }

            return this.MoveToIndex(line.FirstIndex);
        }

        /// <summary>
        /// Moves to line end.
        /// </summary>
        public bool MoveToLineEnd()
        {
            Line line = this.GetAssociatedLine();
            if (line == null)
            {
                return false;
            }

            return this.MoveToIndex(line.LastIndex);
        }

        /// <summary>
        /// Moves line up.
        /// </summary>
        /// <returns></returns>
        public bool MoveLineUp()
        {
            Line line = this.TextPage.GetLineFromIndex(this.Index);
            if (line == null)
            {
                return false;
            }

            Line prevLine;
            TextPage page;
            this.GetPrevLine(line, out prevLine, out page);
            if (page == null || prevLine == null)
            {
                return false;
            }

            double location = this.lastTrackedLocation;
            this.MoveToLineAtLocation(page, prevLine, location);
            this.lastTrackedLocation = location;

            return true;
        }

        /// <summary>
        /// Moves line down.
        /// </summary>
        /// <returns></returns>
        public bool MoveLineDown()
        {
            Line line = this.TextPage.GetLineFromIndex(this.Index);
            if (line == null)
            {
                return false;
            }

            TextPage page;
            Line nextLine;
            this.GetNextLine(line, out nextLine, out page);
            if (page == null || nextLine == null)
            {
                return false;
            }

            double location = this.lastTrackedLocation;
            this.MoveToLineAtLocation(page, nextLine, location);
            this.lastTrackedLocation = location;

            return true;
        }

        /// <summary>
        /// Moves to start of document.
        /// </summary>
        public bool MoveToStartOfDocument()
        {
            return this.MoveToIndex(this.Document.FirstPage, 0);
        }

        /// <summary>
        /// Moves to end of document.
        /// </summary>
        public bool MoveToEndOfDocument()
        {
            TextPage lastPage = this.Document.LastPage;

            return this.MoveToIndex(lastPage, lastPage.LastIndex);
        }

        /// <summary>
        /// Returns a string that represents the current object.
        /// </summary>
        /// <returns>A string that represents the current object.</returns>
        public override string ToString()
        {
            Word word = this.TextPage.GetWordFromIndex(this.Index);
            string text = word.ToString();

            return string.Format("{0}@{1}", text.Substring(0, this.Index - word.FirstIndex), text.Substring(this.Index - word.FirstIndex));
        }

        internal Line GetAssociatedLine()
        {
            return this.TextPage.GetLineFromIndex(this.Index);
        }

        internal Word GetAssociatedWord()
        {
            return this.TextPage.GetWordFromIndex(this.Index);
        }

        internal Character GetAssociatedCharacter()
        {
            return this.TextPage.GetCharacterFromIndex(this.Index);
        }

        internal bool MoveToStartOfTheNextLine()
        {
            TextPosition pos = new TextPosition(this);
            bool res = pos.MoveToLineEnd();
            res &= pos.MoveToNextPosition();
            if (!res)
            {
                return false;
            }

            return this.MoveToPosition(pos);
        }

        internal bool MoveToLocation(Point location)
        {
            return this.ExecuteInsideBeginEndUpdate(() =>
            {
                Line line = this.TextPage.GetNearestLine(location);
                if (line == null)
                {
                    return false;
                }

                int index = line.GetIndexFromLocation(location.X);

                return this.MoveToIndex(index);
            });
        }

        internal bool MoveToLocation(TextPage page, Point location)
        {
            return this.ExecuteInsideBeginEndUpdate(() =>
            {
                Line line = page.GetNearestLine(location);
                if (line == null)
                {
                    return false;
                }

                int index = line.GetIndexFromLocation(location.X);

                return this.MoveToIndex(page, index);
            });
        }

        internal bool MoveToIndex(int index)
        {
            return this.MoveToIndex(this.TextPage, index);
        }

        internal bool MoveToIndex(TextPage page, int index)
        {
            return this.ExecuteInsideBeginEndUpdate(() =>
            {
                if (page == null || index < 0 || index > page.LastIndex)
                {
                    return false;
                }

                this.TextPage = page;
                this.Index = index;

                return true;
            });
        }

        /// <summary>
        /// Called when TextPosition is changing.
        /// </summary>
        protected virtual void OnPositionChanging()
        {
            if (this.PositionChanging != null)
            {
                this.PositionChanging(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Called when TextPosition is changed.
        /// </summary>
        protected virtual void OnPositionChanged()
        {
            if (this.PositionChanged != null)
            {
                this.PositionChanged(this, EventArgs.Empty);
            }
        }

        private static TextPage GetTextPage(RadFixedPage page)
        {
            RadFixedDocument document = page.Parent as RadFixedDocument;
            Guard.ThrowExceptionIfNull(document, "RadFixedPage must be added to a RadFixedDocument");

            return document.TextDocument.GetTextPage(page.PageNumber - 1);
        }

        private Point GetLocation()
        {
            WordPosition pos = this.TextPage.GetWordPositionFromIndex(this.Index);
            if (pos == null)
            {
                return new Point();
            }

            return pos.Location;
        }

        private void MoveToLineAtLocation(TextPage page, Line line, double location)
        {
            Guard.ThrowExceptionIfNull(page, "page");
            Guard.ThrowExceptionIfNull(line, "line");

            int index = line.GetIndexFromLocation(location);
            this.MoveToIndex(page, index);
        }

        private void GetNextLine(Line line, out Line nextLine, out TextPage page)
        {
            Guard.ThrowExceptionIfNull(line, "line");

            int index = this.TextPage.IndexOfLine(line);
            index++;
            if (index < this.TextPage.LinesCount)
            {
                page = this.TextPage;
                nextLine = page.GetLineAt(index);

                return;
            }
            page = this.Document.GetNextPage(this.TextPage);
            nextLine = page == null ? null : Enumerable.FirstOrDefault(page.Lines);
        }

        private void GetPrevLine(Line line, out Line nextLine, out TextPage page)
        {
            Guard.ThrowExceptionIfNull(line, "line");

            int index = this.TextPage.IndexOfLine(line);
            index--;
            if (index >= 0)
            {
                page = this.TextPage;
                nextLine = page.GetLineAt(index);

                return;
            }

            page = this.Document.GetPrevPage(this.TextPage);
            nextLine = page == null ? null : Enumerable.Last(page.Lines);
        }

        private bool ExecuteInsideBeginEndUpdate(Func<bool> method)
        {
            this.BeginUpdate();
            bool result = method();
            this.EndUpdate(result);

            return result;
        }

        private void TrackLocation()
        {
            Word word = this.TextPage.GetWordFromIndex(this.Index);
            if (word != null)
            {
                WordPosition position = word.GetWordPositionFromIndex(this.Index);
                if (position != null)
                {
                    this.lastTrackedLocation = position.Location.X;
                }
            }
        }

        private void EndUpdate(bool shouldFirePositionChanged)
        {
            this.beginUpdateCount--;

            if (this.beginUpdateCount == 0)
            {
                this.TrackLocation();
                if (shouldFirePositionChanged)
                {
                    this.OnPositionChanged();
                }
            }
        }

        private void BeginUpdate()
        {
            if (this.beginUpdateCount == 0)
            {
                this.OnPositionChanging();
            }

            this.beginUpdateCount++;
        }
    }
}
