﻿using System;
#if !NETSTANDARD
using System.Windows;
using System.Windows.Media;
#else
using Telerik.Documents.Primitives;
#endif
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Utilities;
using Telerik.Windows.Documents.Fixed.Model.Text;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Operators.Text;
using Telerik.Windows.Documents.Fixed.Model.Editing.Layout;
using Telerik.Windows.Documents.Fixed.Utilities.Rendering;
using Telerik.Windows.Documents.Fixed.Model.Objects;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model.Fonts;

namespace Telerik.Windows.Documents.Fixed.Text
{
    internal class TextRecognizer
    {
        internal const double Delta = 1.0;

        private readonly object lockObject = new object();
        private readonly TextDocument parent;
        private TextPage currentTextPage;
        private Line currentLine;
        private Word currentWord;
        private Character currentCharacter;
        private int currentIndex;

        public TextRecognizer(TextDocument parent)
        {
            Guard.ThrowExceptionIfNull(parent, "parent");

            this.parent = parent;
        }

        public TextPage CreateTextPage(RadFixedPage fixedPage)
        {
            Guard.ThrowExceptionIfNull(fixedPage, "fixedPage");

            TextPage result = null;
            lock (lockObject)
            {
                this.StartNewTextPage(fixedPage);

                Matrix visiblePageContentTransformation = PageLayoutHelper.CalculateVisibibleContentTransformation(fixedPage);
                this.ProcessContentRoot(fixedPage, visiblePageContentTransformation);

                result = this.FinishCurrentTextPage();
            }

            return result;
        }

        protected void ProcessCharacter(Character character)
        {
            this.currentCharacter = character;
            if (this.IsLineBreak(character))
            {
                this.StartNewLine();
            }
            else if (this.IsWordEnd(character) || this.IsWhiteSpace(character))
            {
                this.StartNewWord();
            }

            this.AddToCurrentWord(character);
        }

        protected TextPage FinishCurrentTextPage()
        {
            this.FinishCurrentLine();
            this.currentTextPage.Finish();
            this.currentLine = null;
            this.currentWord = null;
            TextPage res = this.currentTextPage;
            this.currentTextPage = null;

            return res;
        }

        protected void StartNewTextPage(RadFixedPage fixedPage)
        {
            Guard.ThrowExceptionIfNull(fixedPage, "fixedPage");

            this.currentTextPage = new TextPage(this.parent, fixedPage);
            this.StartNewTextPageInternal();
        }

        protected virtual double GetScaledCharacterSpacing(Character character)
        {
            return character.CharSpacing * character.TextPosition.M11 * character.FontSize;
        }

        protected virtual double GetScaledTextTranslation(Character character)
        {
            return character.TextTranslation * character.TextPosition.M11 * character.FontSize;
        }

        private void ProcessContentRoot(IContentRootElement contentRoot, Matrix visibleContentTransformation)
        {
            foreach (ContentElementBase contentElement in contentRoot.Content)
            {
                if (contentElement.ElementType == FixedDocumentElementType.TextFragment)
                {
                    TextFragment textFragment = (TextFragment)contentElement;
                    this.ProcessTextFragment(textFragment, visibleContentTransformation);
                }
                else if (contentElement.ElementType == FixedDocumentElementType.Form)
                {
                    Form form = (Form)contentElement;
                    this.ProcessForm(form, visibleContentTransformation);
                }
                else if (contentElement is IContentRootElement)
                {
                    this.ProcessContentRoot(contentElement as IContentRootElement, visibleContentTransformation);
                }
            }
        }

        private void ProcessTextFragment(TextFragment textFragment, Matrix visibleContentTransformation)
        {
            Matrix currentGlyphDisplacementMatrix = new Matrix();

            Matrix positionMatrix = ShowText.FlipTextTransformation.MultiplyBy(textFragment.Position.Matrix);
            positionMatrix = positionMatrix.MultiplyBy(visibleContentTransformation);

            foreach (CharInfo glyph in textFragment.TextCollection.Characters)
            {
                if (!string.IsNullOrWhiteSpace(glyph.Unicode))
                {
                    Matrix glyphMatrix = currentGlyphDisplacementMatrix.MultiplyBy(positionMatrix);

                    // The glyph width should contain only the character whithout the character spacing.
                    double glyphWidth = TextFragmentLayoutElement.MeasureWidth(
                        charCode: glyph.CharCode,
                        font: textFragment.TextCollection.Font,
                        fontSize: textFragment.FontSize,
                        characterSpacing: null,
                        wordSpacing: null,
                        horizontalScaling: textFragment.HorizontalScaling);

                    Rect boundingRect = this.CalculateBoundingRect(glyph, glyphMatrix, glyphWidth, textFragment);
                    double charSpacing = textFragment.CharacterSpacing.HasValue ? textFragment.CharacterSpacing.Value : 0;
                    double textTranslation = textFragment.TextTranslation.HasValue ? textFragment.TextTranslation.Value : 0;

                    //TODO Cache the characters?
                    Character character = new Character(
                        boundingRect,
                        glyph.Unicode,
                        charSpacing,
                        textFragment.Position.Matrix,
                        textFragment.FontSize,
                        textTranslation);

                    this.ProcessCharacter(character);
                }

                double glyphDisplacement = TextFragmentLayoutElement.MeasureWidth(
                        charCode: glyph.CharCode,
                        font: textFragment.TextCollection.Font,
                        fontSize: textFragment.FontSize,
                        characterSpacing: textFragment.CharacterSpacing,
                        wordSpacing: textFragment.WordSpacing,
                        horizontalScaling: textFragment.HorizontalScaling);

                currentGlyphDisplacementMatrix = TranslateText.GetTranslatedTextMatrix(currentGlyphDisplacementMatrix, glyphDisplacement);
            }
        }

        private void ProcessForm(Form form, Matrix visibleContentTransformation)
        {
            Matrix formMatrix = form.GetScaledPosition(applyTopLeftTransformation: false);

            Matrix visibleFormContentTransformation = formMatrix.MultiplyBy(visibleContentTransformation);

            this.ProcessContentRoot(form.FormSource, visibleFormContentTransformation);
        }

        private Rect CalculateBoundingRect(CharInfo charInfo, Matrix glyphTransform, double glyphWidth, TextFragment textFragment)
        {
            FontBase font = textFragment.Font;

            double ascent = font.Ascent.Value;
            double descent = font.Descent.Value;
            double textRise = textFragment.TextRise ?? 0;

            double fontSize = textFragment.FontSize;
            double yPosition = (descent * fontSize) + textRise;

            double baseLineFactor = Math.Max(1, (ascent - (2 * descent)) / 1000);
            double glyphHeight = baseLineFactor * fontSize;

            Rect boundingRect = new Rect(new Point(0, 0), new Size(1, 1));
            boundingRect = new Matrix(glyphWidth, 0, 0, glyphHeight, 0, yPosition).Transform(boundingRect);
            boundingRect = glyphTransform.Transform(boundingRect);

            return boundingRect;
        }

        private void StartNewTextPageInternal()
        {
            this.StartNewLine();
            this.currentIndex = 0;
        }

        private bool ShouldInclude(string str)
        {
            return !Extensions.IsNullEmptyOrWhiteSpace(str);
        }

        private Character GetLastCharacter()
        {
            if (this.currentWord.IsEmpty)
            {
                if (this.currentLine.IsEmpty)
                {
                    Line line = this.currentTextPage.LastLine;
                    if (line == null)
                    {
                        return null;
                    }

                    return line.LastWord.LastCharacter;
                }

                return this.currentLine.LastWord.LastCharacter;
            }

            return this.currentWord.LastCharacter;
        }

        private bool IsWhiteSpace(Character character)
        {
            Guard.ThrowExceptionIfNull(character, "character");

            if (Extensions.IsNullEmptyOrWhiteSpace(character.ToUnicode))
            {
                return true;
            }

            Character lastChar = this.GetLastCharacter();
            if (lastChar == null)
            {
                return false;
            }

            double diff = character.BoundingRect.Left - lastChar.BoundingRect.Right;
            double scaledCharacterSpacing = Math.Max(0, this.GetScaledCharacterSpacing(character));
            double scaledTextTranslation = Math.Max(0, this.GetScaledTextTranslation(character));

            bool areDistantCharacters = diff - scaledCharacterSpacing - scaledTextTranslation > Delta;

            return areDistantCharacters || diff < -character.BoundingRect.Width / 2;
        }

        private bool IsPunctuation(string str)
        {
            if (string.IsNullOrEmpty(str) || str.Length > 1)
            {
                return false;
            }

            return char.IsPunctuation(str[0]);
        }

        private bool IsWordEnd(Character character)
        {
            Guard.ThrowExceptionIfNull(character, "character");

            Character lastChar = this.GetLastCharacter();
            if (lastChar == null)
            {
                return false;
            }

            return this.IsPunctuation(character.ToUnicode) ^ this.IsPunctuation(lastChar.ToUnicode);
        }

        private bool IsLineBreak(Character character)
        {
            Guard.ThrowExceptionIfNull(character, "character");

            Character lastChar = this.GetLastCharacter();
            if (lastChar == null)
            {
                return false;
            }

            double epsilon = 0.1;
            double lastTop = lastChar.BoundingRect.Top;
            double lastBottom = lastChar.BoundingRect.Bottom;
            double characterTop = character.BoundingRect.Top;
            double characterBottom = character.BoundingRect.Bottom;

            bool isLastContainingCharacter = lastTop.IsGreaterThanOrEqualTo(characterTop, epsilon) && lastBottom.IsLessThanOrEqualTo(characterBottom, epsilon);
            bool isCharacterContainingLast = characterTop.IsGreaterThanOrEqualTo(lastTop, epsilon) && characterBottom.IsLessThanOrEqualTo(lastBottom, epsilon);

            if (isLastContainingCharacter || isCharacterContainingLast)
            {
                return false;
            }

            return true;
        }

        private void AddToCurrentWord(Character character)
        {
            Guard.ThrowExceptionIfNull(character, "character");

            if (!this.ShouldInclude(character.ToUnicode))
            {
                return;
            }

            this.currentWord.AddPosition(this.currentIndex++);
            this.currentWord.AddCharacter(character);
        }

        private void StartNewWord()
        {
            if (this.currentWord != null)
            {
                if (this.currentWord.IsEmpty)
                {
                    return;
                }

                this.FinishCurrentWord();
            }

            this.currentWord = new Word();
        }

        private void FinishCurrentWord()
        {
            this.FinishCurrentWord(false);
        }

        private void FinishCurrentWord(bool forcePositionAtTheEnd)
        {
            if (!this.currentWord.IsEmpty)
            {
                if (forcePositionAtTheEnd)
                {
                    this.currentWord.AddPosition(this.currentIndex++);
                    this.currentWord.Finish(false);
                }
                else if (this.IsWhiteSpace(this.currentCharacter))
                {
                    this.currentWord.AddPosition(this.currentIndex++);
                    this.currentWord.Finish(true);
                }
                else
                {
                    this.currentWord.Finish(false);
                }

                this.currentLine.AddWord(this.currentWord);
                this.currentWord = null;
            }
        }

        private void StartNewLine()
        {
            if (this.currentLine != null)
            {
                this.FinishCurrentLine();
            }

            this.StartNewWord();
            this.currentLine = new Line();
        }

        private void FinishCurrentLine()
        {
            this.FinishCurrentWord(true);

            if (!this.currentLine.IsEmpty)
            {
                this.currentLine.Finish();
                this.currentTextPage.AddLine(this.currentLine);
                this.currentLine = null;
            }
        }
    }
}
