﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
#if !NETSTANDARD
using System.Windows;
#else
using Telerik.Documents.Primitives;
#endif
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Text
{
    internal class Word
    {
        internal const double DELTA = 0.00001;

        public static int CompareWithCharacterIndex(Word word, int index)
        {
            if (word.FirstIndex <= index && index <= word.LastIndex)
            {
                return 0;
            }
            if (index > word.LastIndex)
            {
                return 1;
            }
            return -1;
        }

        public static int CompareWithWordPosition(WordPosition position, double location)
        {
            if (position.Location.X == location)
            {
                return 0;
            }
            if (location > position.Location.X)
            {
                return 1;
            }
            return -1;
        }

        public static int CompareWithWordBoundingRect(Word word, double location)
        {
            if (location < word.BoundingRect.Left)
            {
                return -1;
            }
            if (word.BoundingRect.Right - location < DELTA && Enumerable.Last(word.positions).Location.X >= location)
            {
                return 0;
            }
            if (word.BoundingRect.Right < location)
            {
                return 1;
            }
            return 0;
        }

        public Word()
        {
            this.chars = new List<Character>();
            this.positions = new List<WordPosition>();
            this.children = new List<object>();
        }

        private readonly List<Character> chars;
        private readonly List<WordPosition> positions;
        private readonly List<object> children;

        internal int FirstIndex
        {
            get;
            private set;
        }

        internal int LastIndex
        {
            get;
            private set;
        }

        internal bool EndsWithSpace
        {
            get;
            private set;
        }

        internal int Length
        {
            get
            {
                return this.chars.Count;
            }
        }

        internal bool IsEmpty
        {
            get
            {
                return this.chars.Count == 0 && this.positions.Count == 0;
            }
        }

        internal Rect BoundingRect
        {
            get;
            private set;
        }

        internal Character LastCharacter
        {
            get
            {
                return Enumerable.LastOrDefault(this.chars);
            }
        }

        public int CharactersCount
        {
            get
            {
                return this.chars.Count;
            }
        }

        private double GetDistanceToWordPositionAtIndex(int index, double location)
        {
            if (this.positions.Count <= index || index < 0)
            {
                return -1;
            }
            return Math.Abs(this.positions[index].Location.X - location);
        }

        private Point GetStart(int index)
        {
            if (index >= this.children.Count || index < 0)
            {
                return new Point();
            }
            Point p;
            if (this.children[index] is Character)
            {
                Character gl = (Character)this.children[index];
                p = new Point(gl.BoundingRect.X, gl.BoundingRect.Y);
            }
            else
            {
                p = ((WordPosition)this.children[index]).Location;
            }
            return p;
        }

        private Point GetEnd(int index)
        {
            if (index > this.children.Count || index < 0)
            {
                return new Point();
            }
            Point p;
            if (this.children[index] is Character)
            {
                Character gl = (Character)this.children[index];
                p = new Point(gl.BoundingRect.Right, gl.BoundingRect.Y);
            }
            else
            {
                p = ((WordPosition)this.children[index]).Location;
            }
            return p;
        }

        private void CalculateLocationByIndex(int index)
        {
            if (index > this.children.Count || index < 0 || !(this.children[index] is WordPosition))
            {
                return;
            }
            Point left;
            Point right;

            if (index == 0)
            {
                left = right = this.GetStart(index + 1);
            }
            else if (index == this.children.Count - 1)
            {
                left = right = this.GetEnd(index - 1);
            }
            else
            {
                left = this.GetEnd(index - 1);
                right = this.GetStart(index + 1);
            }
            ((WordPosition)this.children[index]).Location = new Point(Math.Min(left.X, right.X) + Math.Abs(left.X - right.X) / 2, right.Y);
        }

        private void CalculateBoundingRect()
        {
            Rect rect;
            double minX, minY, maxX, maxY, height;
            minX = minY = double.MaxValue;
            maxX = maxY = height = double.MinValue;


            for (int i = 0; i < this.children.Count; i++)
            {
                if (this.children[i] is WordPosition)
                {
                    this.CalculateLocationByIndex(i);
                }
                else // this.children[i] is Character
                {
                    rect = ((Character)this.children[i]).BoundingRect;
                    minX = Math.Min(rect.Left, minX);
                    minY = Math.Min(rect.Top, minY);
                    maxX = Math.Max(rect.Right, maxX);
                    maxY = Math.Max(rect.Bottom, maxY);
                    height = Math.Max(rect.Height, height);
                }
            }
            this.BoundingRect = new Rect(new Point(minX, minY), new Point(maxX, maxY));
        }

        internal void AddCharacter(Character ch)
        {
            this.chars.Add(ch);
            this.children.Add(ch);
        }

        internal void AddPosition(int index)
        {
            WordPosition position = new WordPosition(index);
            this.positions.Add(position);
            this.children.Add(position);
        }

        internal WordPosition GetWordPositionFromIndex(int index)
        {
            int ind = index - this.FirstIndex;
            if (ind < 0 || ind >= this.positions.Count)
            {
                return null;
            }
            return this.positions[ind];
        }

        public Character GetCharacterAt(int index)
        {
            return this.chars[index];
        }

        internal Character GetCharacterFromIndex(int index)
        {
            WordPosition pos = this.GetWordPositionFromIndex(index);
            if (pos == null)
            {
                return null;
            }
            int ind = this.children.IndexOf(pos);
            if (ind < 0)
            {
                return null;
            }
            int direction = 1;
            if (ind == this.children.Count - 1)
            {
                direction = -1;
            }
            while (!(this.children[ind] is Character))
            {
                ind += direction;
            }
            return (Character)this.children[ind];
        }

        internal void Finish()
        {
            Finish(true);
        }

        internal void Finish(bool endsWithSpace)
        {
            if (this.IsEmpty)
            {
                return;
            }

            this.EndsWithSpace = endsWithSpace;
            this.FirstIndex = Enumerable.First(this.positions).Index;
            this.LastIndex = Enumerable.Last(this.positions).Index;
            this.CalculateBoundingRect();
        }

        internal int GetIndexFromLocation(double location)
        {
            WordPosition pos = CollectionsExtensions.FindElement(this.positions, location, CompareWithWordPosition);
            int index = this.positions.IndexOf(pos);
            
            double minDist = this.GetDistanceToWordPositionAtIndex(index, location);
            int res = index;

            double dist = this.GetDistanceToWordPositionAtIndex(index - 1, location);
            if (dist > 0 && dist < minDist)
            {
                minDist = dist;
                res = index - 1;
            }
            dist = this.GetDistanceToWordPositionAtIndex(index + 1, location);
            if (dist > 0 && dist < minDist)
            {
                minDist = dist;
                res = index + 1;
            }
            return this.positions[res].Index;
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            foreach (Character ch in this.chars)
            {
                if (ch.Char == 0)
                {
                    sb.Append(' ');
                    continue;
                }
                sb.Append(ch);
            }
            return sb.ToString();
        }
    }
}
