﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Utilities.Rendering
{
    internal class ContentElementsGraph
    {
        private readonly Dictionary<Clipping, OrderedHashSet<ContentElementBase>> elementsInClipping;
        private readonly OrderedHashSet<ContentElementBase> rootElements;

        public ContentElementsGraph(IContentRootElement contentElement)
        {
            Guard.ThrowExceptionIfNull(contentElement, "contentElement");

            this.elementsInClipping = new Dictionary<Clipping, OrderedHashSet<ContentElementBase>>();
            this.rootElements = new OrderedHashSet<ContentElementBase>();

            this.BuildGraph(contentElement);
        }

        public IEnumerable<ContentElementBase> GetRootElements()
        {
            return this.rootElements.Elements;
        }

        public IEnumerable<ContentElementBase> GetClippingChildren(Clipping clipping)
        {
            Guard.ThrowExceptionIfNull(clipping, "clipping");

            return this.elementsInClipping[clipping].Elements;
        }

        private void BuildGraph(IContentRootElement contentRootElement)
        {
            foreach (var contentElement in contentRootElement.Content)
            {
                if (contentElement.Clipping == null)
                {
                    this.rootElements.Add(contentElement);
                }
                else
                {
                    var clipping = contentElement.Clipping;
                    this.AddEdge(clipping, contentElement);
                    this.BuildClippingsEdges(contentElement.Clipping);
                }
            }
        }

        private void AddEdge(Clipping clipping, ContentElementBase contentElement)
        {
            OrderedHashSet<ContentElementBase> children;
            if (this.elementsInClipping.TryGetValue(clipping, out children))
            {
                children.Add(contentElement);
            }
            else
            {
                children = new OrderedHashSet<ContentElementBase>();
                children.Add(contentElement);
                this.elementsInClipping.Add(clipping, children);
            }
        }

        private void BuildClippingsEdges(Clipping clipping)
        {
            HashSet<Clipping> visitedClippings = new HashSet<Clipping>();
            visitedClippings.Add(clipping);

            Clipping parent = clipping.Clipping;
            while (parent != null)
            {
                if (visitedClippings.Contains(parent))
                {
                    throw new ArgumentException("Cycling connection between clippings cannot be created.");
                }

                this.AddEdge(parent, clipping);
                clipping = parent;
                parent = parent.Clipping;
                visitedClippings.Add(clipping);
            }

            this.rootElements.Add(clipping);
        }

        private class OrderedHashSet<T>
        {
            private readonly HashSet<T> elements;
            private readonly List<T> elementsList;

            public OrderedHashSet()
            {
                this.elements = new HashSet<T>();
                this.elementsList = new List<T>();
            }

            public IEnumerable<T> Elements
            {
                get
                {
                    foreach (T element in this.elementsList)
                    {
                        yield return element;
                    }
                }
            }

            public void Add(T element)
            {
                int countBefore = this.elements.Count;
                this.elements.Add(element);

                if (this.elements.Count > countBefore)
                {
                    this.elementsList.Add(element);
                }
            }
        }
    }
}
