﻿using System;
using System.Collections.Generic;
#if !NETSTANDARD
using System.Windows;
using System.Windows.Media;
#else
using Telerik.Documents.Primitives;
#endif
using Telerik.Windows.Documents.Core.Fonts;
using Telerik.Windows.Documents.Core.Fonts.Glyphs;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Operators.Text;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Editing.Layout;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Fixed.Model.Objects;
using Telerik.Windows.Documents.Fixed.Model.Text;
using Telerik.Windows.Documents.Utilities;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;

namespace Telerik.Windows.Documents.Fixed.Utilities.Rendering
{
    internal class ContentRootRenderingHelper
    {
        private readonly ContentElementsGraphCache contentElementsGraphCache;
        private readonly Stack<Rect> containersBoundsStack;

        internal ContentRootRenderingHelper()
        {
            this.contentElementsGraphCache = new ContentElementsGraphCache();
            this.containersBoundsStack = new Stack<Rect>();
        }

        internal void RenderContentElements(IFixedContentRenderer renderer, Rect viewport, IContentRootElement element, Rect elementBounds)
        {
            using (this.PushContainerBounds(elementBounds))
            {
                ContentElementsGraph contentElementsGraph = this.contentElementsGraphCache.GetGraph(element);

                foreach (var rootElement in contentElementsGraph.GetRootElements())
                {
                    this.RenderContentElementHierarchy(renderer, viewport, contentElementsGraph, rootElement);
                }
            }
        }

        internal IDisposable EnsureCleanBoundsCache()
        {
            return new DisposableObject(this.ClearBoundsCache);
        }

        private void RenderContentElementHierarchy(IFixedContentRenderer renderer, Rect viewport, ContentElementsGraph contentElementsGraph, ContentElementBase parent)
        {
            Clipping clipping = parent as Clipping;

            if (clipping != null)
            {
                this.RenderClippedElements(renderer, viewport, contentElementsGraph, clipping);
            }
            else
            {
                this.RenderContentElement(renderer, viewport, parent);
            }
        }

        private void RenderClippedElements(IFixedContentRenderer renderer, Rect viewport, ContentElementsGraph contentElementsGraph, Clipping clipping)
        {
            using (renderer.PushTransform(clipping.Transform))
            using (renderer.PushClipping(clipping.Clip))
            {
                IEnumerable<ContentElementBase> children = contentElementsGraph.GetClippingChildren(clipping);

                foreach (var child in children)
                {
                    using (renderer.PushTransform(clipping.Transform.InverseMatrix()))
                    {
                        this.RenderContentElementHierarchy(renderer, viewport, contentElementsGraph, child);
                    }
                }
            }
        }

        private void RenderContentElement(IFixedContentRenderer renderer, Rect viewport, ContentElementBase element)
        {
            if (this.IsInViewport(viewport, element))
            {
                switch (element.ElementType)
                {
                    case FixedDocumentElementType.TextFragment:
                        TextFragment textFragment = (TextFragment)element;
                        this.RenderTextFragment(renderer, textFragment);
                        break;
                    case FixedDocumentElementType.Path:
                        Path path = (Path)element;
                        this.RenderPath(renderer, path);
                        break;
                    case FixedDocumentElementType.Image:
                        Image image = (Image)element;
                        this.RenderImage(renderer, image);
                        break;
                    case FixedDocumentElementType.Form:
                        Form form = (Form)element;
                        this.RenderForm(renderer, form);
                        break;
                }
            }
        }

        private void RenderTextFragment(IFixedContentRenderer renderer, TextFragment textFragment)
        {
            Rect currentContainerBounds = this.PeekContainerBounds();

            if (textFragment.Font.ActualFontSource != EmptyFontSource.Instance)
            {
                double fontSizeScale = textFragment.FontSize / FontBase.GlyphOutlinesCacheFontSize;
                double horizontalScaling = textFragment.HorizontalScaling ?? 1;
                double textRise = textFragment.TextRise ?? 0;
                Matrix fontMatrix = new Matrix(fontSizeScale * horizontalScaling, 0, 0, fontSizeScale, 0, textRise) * ShowText.FlipTextTransformation;

                // The TextMatrix is inversed and applied again in the curent matrix to prevent a bug resulting in incorect glyph scale 
                // caused due to incorect multiply order of the TextMatrix and the current matrix of the TextFragment's poistion.
                Matrix positionMatrix = (ShowText.FlipTextTransformation * textFragment.TextMatrix).InverseMatrix() * textFragment.Position.Matrix;
                Matrix currentTextMatrix = textFragment.TextMatrix;

                GeometryPropertiesOwner properties = textFragment.TextProperties.GeometryProperties;

                using (renderer.PushTransform(positionMatrix))
                {
                    foreach (CharInfo charInfo in textFragment.TextCollection.Characters)
                    {
                        FontBase font = textFragment.TextCollection.Font;
                        GlyphOutlinesCollection outlines = font.GetCachedOutlines(charInfo.CharCode);

                        if (outlines != null)
                        {
                            Matrix glyphMatrix = fontMatrix * currentTextMatrix;
                            GlyphRenderingContext renderingContext = new GlyphRenderingContext(textFragment, outlines, glyphMatrix, currentContainerBounds);
                            renderer.RenderGlyph(renderingContext);
                        }

                        double width = TextFragmentLayoutElement.MeasureWidth(charInfo.CharCode, font, textFragment.FontSize, textFragment.WordSpacing, textFragment.CharacterSpacing, textFragment.HorizontalScaling);
                        currentTextMatrix = TranslateText.GetTranslatedTextMatrix(currentTextMatrix, width);
                    }
                }
            }
            else if (!string.IsNullOrEmpty(textFragment.Text))
            {
                using (renderer.PushTransform(textFragment.Position.Matrix))
                {
                    TextRenderingContext renderingContext = new TextRenderingContext(textFragment, currentContainerBounds);
                    renderer.RenderText(renderingContext);
                }
            }
        }

        private void RenderPath(IFixedContentRenderer renderer, Path path)
        {
            if (path.Geometry != null)
            {
                using (renderer.PushTransform(path.Position.Matrix))
                {
                    Rect currentContainerBounds = this.PeekContainerBounds();
                    GeometryRenderingContext renderingContext = new GeometryRenderingContext(path, currentContainerBounds);
                    renderer.RenderGeometry(renderingContext);
                }
            }
        }

        private void RenderImage(IFixedContentRenderer renderer, Image image)
        {
            if (image.ImageSource != null && image.Width > 0 && image.Height > 0)
            {
                using (renderer.PushTransform(image.Position.Matrix))
                {
                    Rect currentContainerBounds = this.PeekContainerBounds();
                    ImageRenderingContext renderingContext = new ImageRenderingContext(image, currentContainerBounds);
                    renderer.RenderImage(renderingContext);
                }
            }
        }

        private void RenderForm(IFixedContentRenderer renderer, Form form)
        {
            if (form.FormSource != null && form.Width > 0 && form.Height > 0)
            {
                Rect boundingBox = form.FormSource.BoundingBox;
                bool applyTopLeftTransformation = false;
                Matrix formPosition = form.GetScaledPosition(applyTopLeftTransformation);

                using (renderer.PushTransform(formPosition))
                using (renderer.PushTransform(form.FormSource.Matrix))
                using (renderer.PushClipping(new Model.Graphics.RectangleGeometry(boundingBox)))
                {
                    this.RenderContentElements(renderer, boundingBox, form.FormSource, boundingBox);
                }
            }
        }

        private bool IsInViewport(Rect viewport, ContentElementBase element)
        {
            // TODO: Calculate element bounding box and intersect with viewport...
            return true;
        }

        private IDisposable PushContainerBounds(Rect bounds)
        {
            this.containersBoundsStack.Push(bounds);

            return new DisposableObject(this.PopContainerBounds);
        }

        private void PopContainerBounds()
        {
            this.containersBoundsStack.Pop();
        }

        private Rect PeekContainerBounds()
        {
            return this.containersBoundsStack.Peek();
        }

        private void ClearBoundsCache()
        {
            this.containersBoundsStack.Clear();
        }
    }
}