﻿#if !NETSTANDARD
using System.Windows;
using System.Windows.Media;
#else
using Telerik.Documents.Primitives;
#endif
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Fixed.Utilities.Rendering.Annotations;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Utilities.Rendering
{
    /// <summary>
    /// Helper class dedicated to handle the logic for rendering the page content with a <see cref="IFixedContentRenderer"/>.
    /// </summary>
    internal class FixedPageContentRenderingHelper
    {
        private readonly RadFixedPage page;
        private readonly ContentRootRenderingHelper contentRootRenderingHelper;
        private readonly BaseAppearanceProvider appearanceProvider;

        /// <summary>
        /// Initializes a new instance of the <see cref="FixedPageContentRenderingHelper" /> class.
        /// </summary>
        /// <param name="page">The page instance to be rendered.</param>
        public FixedPageContentRenderingHelper(RadFixedPage page)
        {
            this.page = page;
            this.contentRootRenderingHelper = new ContentRootRenderingHelper();
            this.appearanceProvider = new WidgetAppearanceProvider(new DefaultAppearanceProvider(null));
        }

        /// <summary>
        /// Renders the page content elements.
        /// </summary>
        /// <param name="renderer">The actual UI renderer implementation.</param>
        public void RenderContentElements(IFixedContentRenderer renderer)
        {
            Rect viewport = PageLayoutHelper.CalculatePageBoundingRectangle(this.page);

            this.RenderContentElements(renderer, viewport);
        }

        /// <summary>
        /// Renders the page content elements located in the current viewport.
        /// </summary>
        /// <param name="renderer">The actual UI renderer implementation.</param>
        /// <param name="viewport">The current viewport.</param>
        public void RenderContentElements(IFixedContentRenderer renderer, Rect viewport)
        {
            Rect visibleContentBox = PageLayoutHelper.GetVisibleContentBox(this.page);
            Matrix visibleContentTransformation = PageLayoutHelper.CalculateVisibibleContentTransformation(this.page);

            using (this.contentRootRenderingHelper.EnsureCleanBoundsCache())
            using (renderer.PushTransform(visibleContentTransformation))
            using (renderer.PushClipping(new Model.Graphics.RectangleGeometry(visibleContentBox)))
            {
                this.contentRootRenderingHelper.RenderContentElements(renderer, viewport, this.page, visibleContentBox);
            }
        }

        /// <summary>
        /// Renders tiling content elements.
        /// </summary>
        /// <param name="renderer">The actual UI renderer implementation.</param>
        /// <param name="tiling">The tiling instance.</param>
        public void RenderTiling(IFixedContentRenderer renderer, TilingBase tiling)
        {
            using (renderer.PushClipping(new Model.Graphics.RectangleGeometry(tiling.BoundingBox)))
            {
                this.contentRootRenderingHelper.RenderContentElements(renderer, tiling.BoundingBox, tiling, tiling.BoundingBox);
            }
        }

        /// <summary>
        /// Renders the normal annotation appearance. This appearance is used when the annotation 
        /// is not interacting with the user. The appearance is also used for printing the annotation.
        /// </summary>
        /// <param name="renderer">The actual UI renderer implementation.</param>
        /// <param name="annotation">The annotation instance.</param>
        public void RenderNormalAnnotationAppearance(IFixedContentRenderer renderer, Annotation annotation)
        {
            FormSource normalAppearance;

            if (this.appearanceProvider.TryProvideAppearance(annotation, AnnotationAppearanceMode.Normal, out normalAppearance))
            {
                this.RenderAnnotationAppearance(renderer, annotation, normalAppearance);
            }
        }

        /// <summary>
        /// Renders the down annotation appearance. This appearance is used when the pointing device 
        /// of the viewer is pressed or held down within the annotation’s active area. When not present, 
        /// the annotation normal appearance.
        /// </summary>
        /// <param name="renderer">The actual UI renderer implementation.</param>
        /// <param name="annotation">The annotation instance.</param>
        public void RenderDownAnnotationAppearance(IFixedContentRenderer renderer, Annotation annotation)
        {
            FormSource downAppearance;

            if (this.appearanceProvider.TryProvideAppearance(annotation, AnnotationAppearanceMode.Down, out downAppearance))
            {
                this.RenderAnnotationAppearance(renderer, annotation, downAppearance);
            }
            else
            {
                this.RenderNormalAnnotationAppearance(renderer, annotation);
            }
        }

        private void RenderAnnotationAppearance(IFixedContentRenderer renderer, Annotation annotation, FormSource appearance)
        {
            double widthScale = (appearance.Size.Width != 0) ? (annotation.Rect.Width / appearance.Size.Width) : 1;
            double heightScale = (appearance.Size.Height != 0) ? (annotation.Rect.Height / appearance.Size.Height) : 1;
            Matrix scaleMatrix = new Matrix(widthScale, 0, 0, heightScale, annotation.Rect.X, annotation.Rect.Y);
            Matrix scaledAppearanceMatrix = appearance.Matrix.MultiplyBy(scaleMatrix);

            Matrix visibleContentTransformation = PageLayoutHelper.CalculateVisibibleContentTransformation(this.page);
            scaledAppearanceMatrix = scaledAppearanceMatrix.MultiplyBy(visibleContentTransformation);

            using (this.contentRootRenderingHelper.EnsureCleanBoundsCache())
            using (renderer.PushTransform(scaledAppearanceMatrix))
            using (renderer.PushClipping(new Model.Graphics.RectangleGeometry(appearance.BoundingBox)))
            {
                this.contentRootRenderingHelper.RenderContentElements(renderer, appearance.BoundingBox, appearance, appearance.BoundingBox);
            }
        }
    }
}
