﻿#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Fixed.Utilities.Rendering
{
    /// <summary>
    /// A helper class dedicated for layout related calculations of a fixed page.
    /// </summary>
    public static class PageLayoutHelper
    {
        /// <summary>
        /// Gets the rectangle of the visible region of the page, i.e. the intersection 
        /// between the MediaBox and the CropBox properties of the page.
        /// </summary>
        /// <returns>The visible page content rectangle.</returns>
        public static Rect GetVisibleContentBox(IFixedPage page)
        {
            Rect visibleContentBox = page.MediaBox;
            visibleContentBox.Intersect(page.CropBox);
            visibleContentBox = visibleContentBox != Rect.Empty ? visibleContentBox : new Rect(0, 0, 0, 0);

            return visibleContentBox;
        }

        /// <summary>
        /// Gets the width of the visible page content boundaries. This method considers the rotation of the page.
        /// </summary>
        /// <returns>The width of the rotated page content to be displayed.</returns>
        public static double GetActualWidth(IFixedPage page)
        {
            Rect visibleContentBox = GetVisibleContentBox(page);

            if (page.Rotation == Rotation.Rotate0 || page.Rotation == Rotation.Rotate180)
            {
                return visibleContentBox.Width;
            }
            else
            {
                return visibleContentBox.Height;
            }
        }

        /// <summary>
        /// Gets the height of the visible page content boundaries. This method considers the rotation of the page.
        /// </summary>
        /// <returns>The height of the rotated page content to be displayed.</returns>
        public static double GetActualHeight(IFixedPage page)
        {
            Rect visibleContentBox = GetVisibleContentBox(page);

            if (page.Rotation == Rotation.Rotate0 || page.Rotation == Rotation.Rotate180)
            {
                return visibleContentBox.Height;
            }
            else
            {
                return visibleContentBox.Width;
            }
        }

        internal static Rect CalculatePageBoundingRectangle(IFixedPage page)
        {
            Matrix pageTransform = CalculateRotatedPageTransformationInDip(page);
            Rect visibleContentBox = GetVisibleContentBox(page);
            Rect pointRect = new Rect(0, 0, visibleContentBox.Width, visibleContentBox.Height);
            Rect resultRect = pageTransform.Transform(pointRect);

            return resultRect;
        }

        internal static Matrix CalculateRotatedPageTransformationInDip(IFixedPage page)
        {
            Rect visibleContentBox = GetVisibleContentBox(page);

            double rotationAngle;
            Point rotationCenter;

            switch (page.Rotation)
            {
                case Rotation.Rotate90:
                    rotationAngle = 90;
                    rotationCenter = new Point(visibleContentBox.Height / 2, visibleContentBox.Height / 2);
                    break;
                case Rotation.Rotate180:
                    rotationAngle = 180;
                    rotationCenter = new Point(visibleContentBox.Width / 2, visibleContentBox.Height / 2);
                    break;
                case Rotation.Rotate270:
                    rotationAngle = 270;
                    rotationCenter = new Point(visibleContentBox.Width / 2, visibleContentBox.Width / 2);
                    break;
                default:
                    rotationAngle = 0;
                    rotationCenter = new Point();
                    break;
            }

            Matrix matrix = new Matrix();
#if SILVERLIGHT || NETSTANDARD
            matrix = matrix.RotateMatrixAt(rotationAngle, rotationCenter.X, rotationCenter.Y);
#else
            matrix.RotateAt(rotationAngle, rotationCenter.X, rotationCenter.Y);
#endif

            return matrix;
        }

        internal static Matrix CalculateVisibibleContentTransformation(IFixedPage page)
        {
            Rect visibleContentBox = GetVisibleContentBox(page);
            Matrix rotatedPageTransformation = CalculateRotatedPageTransformationInDip(page);
            // The visible content can have some offsets (X, Y). Since the rendered page starts from (0, 0) 
            // the page content needs to be translated (-X, -Y) to the beginning of the page.
            Matrix visibleContentTransformation = new Matrix(1, 0, 0, 1, -visibleContentBox.X, -visibleContentBox.Y).MultiplyBy(rotatedPageTransformation);

            return visibleContentTransformation;
        }

        internal static Matrix CalculatePointToDipPageTransformation(IFixedPage page)
        {
            Matrix matrix = new Matrix();

            double pointToDip = Unit.PointToDip(1);
            double scaleX = pointToDip, scaleY = -pointToDip, centerX = 0, centerY = page.MediaBox.Height / (1 + pointToDip);
            Matrix rotation = CalculateRotatedPageTransformationInDip(page);

#if SILVERLIGHT || NETSTANDARD
            matrix = matrix.ScaleMatrixAt(scaleX, scaleY, centerX, centerY);
            matrix = matrix.MultiplyBy(rotation);
#else
            matrix.ScaleAt(scaleX, scaleY, centerX, centerY);
            matrix.Append(rotation);
#endif

            return matrix;
        }
    }
}
