﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.DigitalSignatures;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;

namespace Telerik.Windows.Documents.Fixed.Tests.DigitalSignature
{
    [TestClass]
    public class DateCodecTests : SignatureTests
    {
        private static Dictionary<string, DateTime> dates;

        public DateCodecTests()
        {
            dates = new Dictionary<string, DateTime>();
            dates.Add("D:20111109150311Z", new DateTime(2011, 11, 09, 15, 3, 11));
            dates.Add("D:20140904111427+00'00'", new DateTime(2014, 09, 04, 11, 14, 27, DateTimeKind.Utc));
            dates.Add("D:20140904111427+02'00'", new DateTime(2014, 09, 04, 09, 14, 27, DateTimeKind.Utc));
            dates.Add("D:20140904111427+02'30'", new DateTime(2014, 09, 04, 08, 44, 27, DateTimeKind.Utc));
            dates.Add("D:20111202135540-05'00'", new DateTime(2011, 12, 02, 18, 55, 40, DateTimeKind.Utc));
            dates.Add("D:20130430100553-03'00'", new DateTime(2013, 04, 30, 13, 5, 53, DateTimeKind.Utc));
            dates.Add("D:20170419171740+03'00'", new DateTime(2017, 04, 19, 14, 17, 40, DateTimeKind.Utc));
            dates.Add("D:20170419171740+03'15'", new DateTime(2017, 04, 19, 14, 02, 40, DateTimeKind.Utc));
            dates.Add("D:20170419171740+03'35'", new DateTime(2017, 04, 19, 13, 42, 40, DateTimeKind.Utc));
            dates.Add("D:20170419171740+03'45'", new DateTime(2017, 04, 19, 13, 32, 40, DateTimeKind.Utc));
            dates.Add("D:20110211172549-05'00'", new DateTime(2011, 02, 11, 22, 25, 49, DateTimeKind.Utc));
            dates.Add("D:20120210135916-06'00'", new DateTime(2012, 02, 10, 19, 59, 16, DateTimeKind.Utc));
        }

        [TestMethod]
        public void TestDateParser_Decode()
        {
            foreach (var pair in dates)
            {
                DateTime? dateTime = SignatureDateCodec.Decode(pair.Key);
                Assert.IsNotNull(dateTime);

                DateTime universal = dateTime.Value.ToUniversalTime();
                Assert.AreEqual(pair.Value, universal, "The problematic date is: " + pair.Key);
            }
        }

        [TestMethod]
        public void TestDateParser_Encode()
        {
            foreach (var pair in dates)
            {
                if (pair.Key.Contains("Z"))
                {
                    continue;
                }

                DateTime localTime = pair.Value.ToLocalTime();

                string encodedLocalTime = SignatureDateCodec.Encode(localTime);

                DateTime? dateTime = SignatureDateCodec.Decode(encodedLocalTime);

                DateTime universal = dateTime.Value.ToUniversalTime();
                Assert.AreEqual(pair.Value, universal, "The problematic date is: " + pair.Key);
            }
        }

        [TestMethod]
        public void ImportDocument_VerifyTimeOfSigning()
        {
            //D: 201704 21 13 27 56 + 03'00'
            DateTime expected = new DateTime(2017, 04, 21, 10, 27, 56).ToLocalTime();

            Action<Signature, SignatureValidationResult> assertAction = (signature, validationResult) =>
            {
                Assert.IsTrue(expected.Equals(signature.Properties.TimeOfSigning));
            };

            RadFixedDocument document = this.Import("SignedWithSignedByCA.pdf");

            this.ValidateDocument(document, assertAction);
        }

        [TestMethod]
        public void ExportDocument_VerifyTimeOfSigning()
        {
            DateTime customSigningTime = new DateTime(2017, 04, 21, 10, 27, 56).ToUniversalTime();

            Action<Signature, SignatureValidationResult> assertAction = (signature, validationResult) =>
            {
                var expectedTime = customSigningTime.ToLocalTime();
                Assert.IsTrue(expectedTime.Equals(signature.Properties.TimeOfSigning));
            };

            RadFixedDocument document = Import("Unsigned.pdf");

            var field = document.AcroForm.FormFields.Where(f => f.FieldType == FormFieldType.Signature).First();
            SignatureField signatureField = (SignatureField)field;

            X509Certificate2 certificate = GetCertificate(SignatureTests.ChildCertificate);
            signatureField.Signature = new Signature(certificate);
            signatureField.Signature.Properties.Name = signatureField.Name + " Name";
            signatureField.Signature.Properties.TimeOfSigning = customSigningTime;

            this.ExportImportAndValidate(document, assertAction);
        }
    }
}
