﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Windows;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.DigitalSignatures;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Fixed.Tests.Utils;

namespace Telerik.Windows.Documents.Fixed.Tests.DigitalSignature
{
    public abstract class SignatureTests
    {
        public const string DocumentFolder = @"Documents\DigitalSignature\";
        public const string RootCertificate = "TempCA.pfx";
        public const string ChildCertificate = "SignedByCA.pfx";
        public const string NotTrustedRootCertificate = "NotTrustedRoot.pfx";
        public const string SignedByNotTrustedRootCertificate = "SignedByNotTrustedRoot.pfx";

        protected static X509Certificate2 GetCertificate(string certificateName)
        {
            X509Certificate2 certificate;

            using (Stream stream = TestUtils.GetApplicationResourceStream(@"DigitalSignature\Certificates\" + certificateName))
            {
                byte[] certRawData = new byte[stream.Length];
                stream.Read(certRawData, 0, (int) stream.Length);

                certificate = new X509Certificate2(certRawData, "test");
            }

            return certificate;
        }

        protected static Uri GetResourceUri(string resource)
        {
            AssemblyName assemblyName = new AssemblyName(typeof(SignatureTests).Assembly.FullName);
            string resourcePath = "/" + assemblyName.Name + ";component/" + resource;
            Uri resourceUri = new Uri(resourcePath, UriKind.Relative);

            return resourceUri;
        }

        protected void ValidateDocument(RadFixedDocument document)
        {
            Action<Signature, SignatureValidationResult> assertAction = (signature, validationResult) =>
            {
                Assert.IsFalse(validationResult.IsDocumentModified);
                Assert.IsTrue(validationResult.IsCertificateValid);
            };

            this.ValidateDocument(document, assertAction);
        }

        protected void ValidateDocument(RadFixedDocument document, Action<Signature, SignatureValidationResult> assertAction)
        {
            var signatureFields = document.AcroForm.FormFields.Where(f => f.FieldType == FormFieldType.Signature);
            foreach (SignatureField signatureField in signatureFields)
            {
                if (signatureField.Signature != null)
                {
                    SignatureValidationResult validationResult = signatureField.Signature.Validate();
                    assertAction(signatureField.Signature, validationResult);
                }
            }
        }

        protected void TryValidateDocument(RadFixedDocument document)
        {
            Action<SignatureValidationResult> assertAction = (validationResult) =>
            {
                Assert.IsFalse(validationResult.IsDocumentModified);
                Assert.IsTrue(validationResult.IsCertificateValid);
            };

            TryValidateDocument(document, assertAction);
        }

        protected void TryValidateDocument(RadFixedDocument document, Action<SignatureValidationResult> assertAction)
        {
            var signatureFields = document.AcroForm.FormFields.Where(f => f.FieldType == FormFieldType.Signature);
            foreach (SignatureField signatureField in signatureFields)
            {
                SignatureValidationResult validationResult;
                signatureField.Signature.TryValidate(out validationResult);

                assertAction(validationResult);
            }
        }

        protected void ExportImportAndValidate(RadFixedDocument document)
        {
            PdfFormatProvider provider = new PdfFormatProvider();

            using (Stream outputStream = new MemoryStream())
            {
                provider.Export(document, outputStream);
                var importedDocument = provider.Import(outputStream);

                this.ValidateDocument(importedDocument);
            }
        }

        protected void ExportImportAndValidate(RadFixedDocument document, Action<Signature, SignatureValidationResult> assertAction)
        {
            PdfFormatProvider provider = new PdfFormatProvider();

            using (Stream outputStream = new MemoryStream())
            {
                provider.Export(document, outputStream);
                var importedDocument = provider.Import(outputStream);

                this.ValidateDocument(importedDocument, assertAction);
            }
        }

        protected RadFixedDocument Import(string documentName)
        {
            RadFixedDocument document;

            MemoryStream memoryStream = new MemoryStream();

            using (Stream stream = TestUtils.GetApplicationResourceStream(DocumentFolder + documentName))
            {
                stream.CopyTo(memoryStream);
            }

            PdfFormatProvider fp = new PdfFormatProvider();
            document = fp.Import(memoryStream);
            return document;
        }

        protected RadFixedDocument Import(Stream stream)
        {
            PdfFormatProvider fp = new PdfFormatProvider();
            RadFixedDocument document = fp.Import(stream);

            return document;
        }
    }
}
