﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Security.Cryptography.X509Certificates;
using System.Text;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.DigitalSignatures;
using Telerik.Windows.Documents.Fixed.Model.Editing;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Fixed.Model.Resources;

namespace Telerik.Windows.Documents.Fixed.Tests.DigitalSignature
{
    [TestClass]
    public class SigningTests : SignatureTests
    {
        private static string SignatureName = "MySignature";

#if DEBUG
        [ClassInitialize]
        public static void Initialize(TestContext context)
        {
            X509Certificate2 certificate = GetCertificate(SignatureTests.RootCertificate);

            X509Store store = new X509Store(StoreName.Root, StoreLocation.CurrentUser);
            store.Open(OpenFlags.MaxAllowed);

            bool containsCertificate = store.Certificates.Contains(certificate);
            if (!containsCertificate)
            {
                store.Add(certificate);
            }
        }
#endif

        [TestMethod]
        public void CreateDocumentAndSign_WithWidgetAndFormSource()
        {
            RadFixedDocument document = new RadFixedDocument();
            var page = document.Pages.AddPage();

            var signatureField = new SignatureField(SignatureName);
            SignSignatureField(signatureField);

            FormSource formSource = GetFormSource(SignatureName);

            SignatureWidget widget = GetWidget(signatureField, formSource);
            page.Annotations.Add(widget);

            document.AcroForm.FormFields.Add(signatureField);

            this.ExportImportAndValidate(document);
        }

        [TestMethod]
        public void CreateDocumentAndSign_WithWidget()
        {
            RadFixedDocument document = new RadFixedDocument();
            var page = document.Pages.AddPage();

            var signatureField = new SignatureField(SignatureName);
            SignSignatureField(signatureField);

            SignatureWidget widget = GetWidget(signatureField);
            page.Annotations.Add(widget);

            document.AcroForm.FormFields.Add(signatureField);

            this.ExportImportAndValidate(document);
        }

        [TestMethod]
        public void CreateDocumentAndSign_WithEmptySignatureField()
        {
            RadFixedDocument document = new RadFixedDocument();
            var page = document.Pages.AddPage();

            var signatureField = new SignatureField(SignatureName);
            SignSignatureField(signatureField);

            document.AcroForm.FormFields.Add(signatureField);

            this.ExportImportAndValidate(document);
        }

        [TestMethod]
        public void ImportDocument_Sign()
        {
            RadFixedDocument document = Import("Unsigned.pdf");

            var field = document.AcroForm.FormFields.Where(f => f.FieldType == FormFieldType.Signature).First();
            SignatureField signatureField = (SignatureField)field;

            SignSignatureField(signatureField);
            this.ExportImportAndValidate(document);
        }

        [TestMethod]
        public void ImportDocument_AddSignatureFieldWithWidgetAndFormSorce_Sign()
        {
            RadFixedDocument document = Import("SimpleText.pdf");

            RadFixedPage page = document.Pages[0];

            var signatureField = new SignatureField(SignatureName);
            SignSignatureField(signatureField);

            FormSource formSource = GetFormSource(SignatureName);

            SignatureWidget widget = GetWidget(signatureField, formSource);
            page.Annotations.Add(widget);

            document.AcroForm.FormFields.Add(signatureField);

            this.ExportImportAndValidate(document);
        }

        [TestMethod]
        public void ImportDocument_AddSignatureFieldWithWidget_Sign()
        {
            RadFixedDocument document = Import("SimpleText.pdf");

            RadFixedPage page = document.Pages[0];

            var signatureField = new SignatureField(SignatureName);
            SignSignatureField(signatureField);

            SignatureWidget widget = GetWidget(signatureField);
            page.Annotations.Add(widget);

            document.AcroForm.FormFields.Add(signatureField);

            this.ExportImportAndValidate(document);
        }

        [TestMethod]
        public void ImportDocument_AddSignatureField_Sign()
        {
            RadFixedDocument document = Import("SimpleText.pdf");

            RadFixedPage page = document.Pages[0];

            var signatureField = new SignatureField(SignatureName);
            SignSignatureField(signatureField);

            document.AcroForm.FormFields.Add(signatureField);

            this.ExportImportAndValidate(document);
        }

        [TestMethod]
        public void CreateDocumentAndSign_WithSignedByNotTrustedRootCertificate()
        {
            RadFixedDocument document = new RadFixedDocument();
            var page = document.Pages.AddPage();

            var signatureField = new SignatureField(SignatureName);
            SignSignatureField(signatureField, SignatureTests.SignedByNotTrustedRootCertificate);

            FormSource formSource = GetFormSource(SignatureName);

            SignatureWidget widget = GetWidget(signatureField, formSource);
            page.Annotations.Add(widget);

            document.AcroForm.FormFields.Add(signatureField);

            Action<Signature, SignatureValidationResult> assertAction = (signature, validationResult) =>
            {
                Assert.IsFalse(validationResult.IsDocumentModified);
                Assert.IsFalse(validationResult.IsCertificateValid);
            };

            this.ExportImportAndValidate(document, assertAction);
        }

        private static void SignSignatureField(SignatureField signatureField)
        {
            SignSignatureField(signatureField, SignatureTests.ChildCertificate);
        }

        private static void SignSignatureField(SignatureField signatureField, string certificateName)
        {
            X509Certificate2 certificate = GetCertificate(certificateName);

            signatureField.Signature = new Signature(certificate);
            signatureField.Signature.Properties.Name = signatureField.Name + " Name";
        }

        private static SignatureWidget GetWidget(SignatureField signatureField, FormSource formSource = null)
        {
            SignatureWidget widget = signatureField.Widgets.AddWidget();
            widget.Rect = new Rect(new Point(0, 200), new Size(100, 100));
            widget.Border = new AnnotationBorder();
            widget.Border.Style = AnnotationBorderStyle.Solid;
            widget.Border.Width = 100;
            if (formSource != null)
            {
                widget.Content.NormalContentSource = formSource;
            }

            return widget;
        }

        private static FormSource GetFormSource(string signatureName)
        {
            var formSource = new FormSource();
            formSource.Size = new Size(120, 120);

            FixedContentEditor ed = new FixedContentEditor(formSource);
            ed.DrawCircle(new Point(50, 50), 20);
            ed.DrawText(signatureName);
            return formSource;
        }
    }
}
