﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Security.Cryptography.X509Certificates;
using System.Text;
#if NETSTANDARD
#else
using System.Windows;
using System.Windows.Resources;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.DigitalSignatures;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Fixed.Tests.Utils;

namespace Telerik.Windows.Documents.Fixed.Tests.DigitalSignature
{
    [TestClass]
    public class ValidationTests : SignatureTests
    {
#if DEBUG
        [ClassInitialize]
        public static void Initialize(TestContext context)
        {
            X509Certificate2 certificate = GetCertificate(SignatureTests.RootCertificate);

            X509Store store = new X509Store(StoreName.Root, StoreLocation.CurrentUser);
            store.Open(OpenFlags.MaxAllowed);

            bool containsCertificate = store.Certificates.Contains(certificate);
            if (!containsCertificate)
            {
                store.Add(certificate);
            }
        }
#endif

        [TestMethod]
        public void ValidateDocument_SignedByChildCertificate()
        {
            ImportAndValidate(@"Documents\DigitalSignature\SignedWithSignedByCA.pdf");
        }

        [TestMethod]
        public void ValidatedDocument_SignedByRootCertificate()
        {
            ImportAndValidate(@"Documents\DigitalSignature\SignedWithTempCA.pdf");
        }

        [TestMethod]
        public void ImportDocument_WithSignedAndUnsignedFields_Validate()
        {
            ImportAndValidate(@"Documents\DigitalSignature\SignedAndUnsignedFields.pdf");
        }

        [TestMethod]
        public void ImportDocument_ValidateWithCustomSignatureHandler()
        {
            CustomSignatureValidationHandler handler = new CustomSignatureValidationHandler();

            SignatureValidationHandlersManager.RegisterHandler("adbe.pkcs7.detached1", handler);

            string documentPath = @"Documents\DigitalSignature\CustomSignatureHandler.pdf";

            using (Stream stream = TestUtils.GetApplicationResourceStream(documentPath))
            {
                PdfFormatProvider fp = new PdfFormatProvider();
                RadFixedDocument document = fp.Import(stream);

                this.ValidateDocument(document);
            }
        }

        [TestMethod]
        public void ImportDocument_TryValidate()
        {
            this.ImportAndValidate(@"Documents\DigitalSignature\SignedWithSignedByCA.pdf", this.TryValidateDocument);
        }

        [TestMethod]
        public void ImportDocument_UseChainFlagsAndTryValidate()
        {
            RadFixedDocument document = this.Import("Invalid.pdf");

            IEnumerable<FormField> signatureFields = document.AcroForm.FormFields.Where(f => f.FieldType == FormFieldType.Signature);
            foreach (SignatureField signatureField in signatureFields)
            {
                SignatureValidationProperties validationProperties = new SignatureValidationProperties();
                SignatureValidationResult validationResult;
                if (!signatureField.Signature.TryValidate(validationProperties, out validationResult))
                {
                    throw new Exception("The validaiton of signature is not successful.");
                }

                Assert.IsFalse(validationResult.IsCertificateValid, "Invalid certificate before verification flags for field: " + signatureField.Name);

                validationProperties.Chain.ChainPolicy.VerificationFlags = X509VerificationFlags.AllFlags;

                if (!signatureField.Signature.TryValidate(validationProperties, out validationResult))
                {
                    throw new Exception("The validaiton of signature is not successful.");
                }

                Assert.IsTrue(validationResult.IsCertificateValid, "Invalid certificate after verification flags for field: " + signatureField.Name);
            }
        }

        [TestMethod]
        public void ImportDocument_ValidateInvalidSignature()
        {
            Action<Signature, SignatureValidationResult> assertAction = (signature, validationResult) =>
            {
                Assert.IsTrue(validationResult.IsDocumentModified);
                Assert.IsFalse(validationResult.IsCertificateValid);
            };

            RadFixedDocument document = this.Import("Invalid.pdf");
            this.ValidateDocument(document, assertAction);
        }

        [TestMethod]
        public void ImportDocument_ValidateUnknownSignature()
        {
            Action<Signature, SignatureValidationResult> assertAction = (signature, validationResult) =>
            {
                Assert.IsFalse(validationResult.IsDocumentModified, "IsDocumentModified");
                Assert.IsFalse(validationResult.IsCertificateValid, "IsCertificateValid");
            };

            RadFixedDocument document = this.Import("Unknown.pdf");
            this.ValidateDocument(document, assertAction);
        }

        private void ImportAndValidate(string documentPath)
        {
            this.ImportAndValidate(documentPath, this.ValidateDocument);
        }

        private void ImportAndValidate(string documentPath, Action<RadFixedDocument> validateAction)
        {
            RadFixedDocument document;

            using (Stream stream = TestUtils.GetApplicationResourceStream(documentPath))
            {
                PdfFormatProvider fp = new PdfFormatProvider();
                document = fp.Import(stream);

                validateAction(document);
            }
        }
    }
}
