﻿using System;
using System.Linq;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Editing;
using Telerik.Windows.Documents.Fixed.Model.Text;
using Telerik.Windows.Documents.Fixed.Tests.Utils;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
#if VISUAL
using Telerik.Windows.Documents.Flow.Tests.Utils;
#endif

namespace Telerik.Windows.Documents.Fixed.Tests.Editing
{
    [TestClass]
#if VISUAL
    public class MergeTests : DocumentFixedVisualBaseTest
#else
    public class MergeTests
#endif
    {
        [TestMethod]
        [TestCategory(PdfProcessingCategory.VisualTest)]
        public void Merge_ImportAndMergeShouldResultInSameTextCollectionGlyphs()
        {
            RadFixedDocument document = new PdfFormatProvider().Import(TestUtils.GetResourceBytes("Documents/Text/Telerik.pdf"));
            int expectedPageContentElementsCount = 3;

            Assert.AreEqual(1, document.Pages.Count);
            Assert.AreEqual(expectedPageContentElementsCount, document.Pages[0].Content.Count);

            CharInfo[][] importedGlyphs = new CharInfo[expectedPageContentElementsCount][];

            for (int i = 0; i < expectedPageContentElementsCount; i++)
            {
                importedGlyphs[i] = ((TextFragment)document.Pages[0].Content[i]).TextCollection.Characters.ToArray();
            }

            document.Merge(document);
            Assert.AreEqual(2, document.Pages.Count);

            for (int pageIndex = 0; pageIndex < document.Pages.Count; pageIndex++)
            {
                Assert.AreEqual(expectedPageContentElementsCount, document.Pages[pageIndex].Content.Count);

                for (int contentIndex = 0; contentIndex < expectedPageContentElementsCount; contentIndex++)
                {
                    CharInfo[] importedElementGlyphs = importedGlyphs[contentIndex];
                    CharInfo[] glyphsAfterMerge = ((TextFragment)document.Pages[pageIndex].Content[contentIndex]).TextCollection.Characters.ToArray();

                    Assert.AreEqual(importedElementGlyphs.Length, glyphsAfterMerge.Length);

                    for (int glyphIndex = 0; glyphIndex < importedElementGlyphs.Length; glyphIndex++)
                    {
                        CharInfo importedGlyph = importedElementGlyphs[glyphIndex];
                        CharInfo glyphAfterMerge = glyphsAfterMerge[glyphIndex];

                        Assert.AreEqual(importedGlyph, glyphAfterMerge);
                    }
                }
            }
#if VISUAL
            this.AssertVisual(document, UIInteractionHelper.PageDown(), 1);
#endif
        }

        [TestMethod]
        [TestCategory(PdfProcessingCategory.VisualTest)]
        public void Merge_WhenSourceAndTargetDocumentsContainsSinglePage_TargetDocumentContainsSourceDocumentPage()
        {
            // Arrange
            RadFixedDocument target = new RadFixedDocument();

            using (RadFixedDocumentEditor editor = new RadFixedDocumentEditor(target))
            {
                editor.InsertRun("Target");
            }

            RadFixedDocument source = new RadFixedDocument();
            using (RadFixedDocumentEditor editor = new RadFixedDocumentEditor(source))
            {
                editor.InsertRun("Source");
            }

            // Act
            target.Merge(source);

            // Assert
            Assert.AreEqual(2, target.Pages.Count);
            Assert.AreEqual(1, source.Pages.Count);

            StringBuilder builder = new StringBuilder();
            foreach (var text in target.Pages.Last().Content.Cast<TextFragment>())
            {
                builder.Append(text.Text);
            }
            Assert.AreEqual("Source", builder.ToString());
#if VISUAL
            this.AssertVisual(target, UIInteractionHelper.PageDown());
#endif

        }

        [TestMethod]
        [TestCategory(PdfProcessingCategory.VisualTest)]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Merge_WhenSourceDocumentIsNull_ThrowsArgumentNullException()
        {
            // Arrange
            RadFixedDocument target = new RadFixedDocument();

            using (RadFixedDocumentEditor editor = new RadFixedDocumentEditor(target))
            {
                editor.InsertRun("Target");
            }

            // Act
            target.Merge(null);
        }

        [TestMethod]
        [TestCategory(PdfProcessingCategory.VisualTest)]
        public void Merge_WhenSourceDocumentIsEmpty_TargetDocumentIsUnchanged()
        {
            // Arrange
            RadFixedDocument target = new RadFixedDocument();

            using (RadFixedDocumentEditor editor = new RadFixedDocumentEditor(target))
            {
                editor.InsertRun("Target");
            }

            RadFixedDocument source = new RadFixedDocument();

            // Act
            target.Merge(source);

            // Assert
            Assert.AreEqual(1, target.Pages.Count);

            StringBuilder builder = new StringBuilder();
            foreach (var text in target.Pages.First().Content.Cast<TextFragment>())
            {
                builder.Append(text.Text);
            }
            Assert.AreEqual("Target", builder.ToString());
#if VISUAL
            this.AssertVisual(target);
#endif
        }

        [TestMethod]
        [TestCategory(PdfProcessingCategory.VisualTest)]
        public void Merge_WhenTargetDocumentIsEmpty_TargetDocumentIsLikeSourceDocument()
        {
            // Arrange
            RadFixedDocument target = new RadFixedDocument();

            RadFixedDocument source = new RadFixedDocument();
            using (RadFixedDocumentEditor editor = new RadFixedDocumentEditor(source))
            {
                editor.InsertRun("Source");
            }

            // Act
            target.Merge(source);

            // Assert
            Assert.AreEqual(1, target.Pages.Count);

            StringBuilder builder = new StringBuilder();
            foreach (var text in target.Pages.First().Content.Cast<TextFragment>())
            {
                builder.Append(text.Text);
            }
            Assert.AreEqual("Source", builder.ToString());
#if VISUAL
            this.AssertVisual(target);
#endif
        }

#if VISUAL
        [TestMethod]
        [TestCategory(PdfProcessingCategory.VisualTest)]
        [Description("Team Pulse ID: 171309, TFS ID: 243197")]
        public void Merge_SourceDocumentContentIsNotClearedAfterMerge()
        {
            RadFixedDocument output = new RadFixedDocument();
            RadFixedDocument imported1 = this.LoadVisualTestDocument("example.pdf");
            RadFixedDocument imported2 = this.LoadVisualTestDocument("TestFile.pdf");
            RadFixedDocument imported3 = this.LoadVisualTestDocument("test.pdf");

            output.Merge(imported1);
            output.Merge(imported2);
            output.Merge(imported3);

            this.AssertVisual(output, UIInteractionHelper.PageDown(), 3);
            this.AssertVisual(imported1);
        }

        [TestMethod]
        [TestCategory(PdfProcessingCategory.VisualTest)]
        [Description("Team Pulse ID: 171309, TFS ID: 243197")]
        public void Merge_SourceDocumentFontPropertiesAreProperlyClonedAfterMerge()
        {
            RadFixedDocument output = new RadFixedDocument();
            RadFixedDocument imported = this.LoadVisualTestDocument("Acceleration_Program_Book.pdf");

            output.Merge(imported);            

            this.AssertVisual(output, UIInteractionHelper.PageDown(), 1);
        }
#endif
    }
}