﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Encryption;
using Telerik.Windows.Documents.Fixed.Model;

namespace Telerik.Windows.Documents.Fixed.Tests.Encryption
{
    [TestClass]
    public class StandardEncryptTests
    {
        private const int PermissionsDefaultValue = -3904;
        private static readonly byte[] DocumentId = { 0x54, 0x68, 0x75, 0x72, 0x73, 0x64, 0x61, 0x79, 0x2C, 0x20, 0x4A, 0x75, 0x6E, 0x65, 0x20, 0x39, 0x2C, 0x20, 0x32, 0x30, 0x31, 0x36, 0x34, 0x3A, 0x32, 0x35, 0x3A, 0x30, 0x39, 0x20, 0x50, 0x4D, };
        private static readonly byte[] ExpectedOwnerPasswordWhenEmptyPassword = { 0x36, 0x45, 0x1B, 0xD3, 0x9D, 0x75, 0x3B, 0x7C, 0x1D, 0x10, 0x92, 0x2C, 0x28, 0xE6, 0x66, 0x5A, 0xA4, 0xF3, 0x35, 0x3F, 0xB0, 0x34, 0x8B, 0x53, 0x68, 0x93, 0xE3, 0xB1, 0xDB, 0x5C, 0x57, 0x9B, };
        private static readonly byte[] ExpectedEncryptionKeyWhenEmptyPassword = { 0xA0, 0x08, 0x8D, 0xB4, 0xD9, 0x3A, 0xAC, 0xA3, 0x0E, 0xAB, 0x7A, 0xFE, 0xB5, 0x76, 0xBA, 0x7C, };
        private static readonly byte[] ExpectedOwnerPasswordWhenNonEmptyPassword = { 0x2A, 0xB8, 0xBE, 0xFC, 0xE9, 0x7F, 0xF8, 0x2B, 0x7A, 0x80, 0xE9, 0x5C, 0x0B, 0x90, 0xD2, 0x97, 0x03, 0x63, 0xA7, 0xC9, 0x04, 0xF8, 0x81, 0xF8, 0x23, 0x82, 0xF8, 0x8C, 0x9B, 0x63, 0x1C, 0xDA, };
        private static readonly byte[] ExpectedEncryptionKeyWhenNonEmptyPassword = { 0x50, 0x2A, 0x97, 0x08, 0x03, 0x19, 0x9A, 0x36, 0xDA, 0xA6, 0xF9, 0x12, 0xE3, 0x9D, 0xC5, 0x3E, };

        [TestMethod]
        public void TestStandardEncrypt_OwnerPasswordEncryptionWhenPasswordIsEmpty()
        {
            this.TestOwnerPasswordEncryption(CreateEmptyExportSettings(), ExpectedOwnerPasswordWhenEmptyPassword);
        }

        [TestMethod]
        public void TestStandardEncrypt_EncryptionKeyCalculationsWhenPasswordIsEmpty()
        {
            this.TestEncryptionKeyCalculations(CreateEmptyExportSettings(), ExpectedOwnerPasswordWhenEmptyPassword, ExpectedEncryptionKeyWhenEmptyPassword);
        }

        [TestMethod]
        public void TestStandardEncrypt_UserPasswordEncryptionWhenPasswordIsEmpty()
        {
            byte[] expectedUserPassword = { 0x88, 0x06, 0x8C, 0x15, 0xCE, 0x39, 0x95, 0x98, 0x99, 0xE9, 0x38, 0x0B, 0xB1, 0xAA, 0xDB, 0x5D, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
            this.TestUserPasswordEncryption(ExpectedEncryptionKeyWhenEmptyPassword, expectedUserPassword);
        }

        [TestMethod]
        public void TestStandardEncrypt_OwnerPasswordEncryptionWhenPasswordIsNotEmpty()
        {
            this.TestOwnerPasswordEncryption(CreateExportSettingsWithPassword(), ExpectedOwnerPasswordWhenNonEmptyPassword);
        }

        [TestMethod]
        public void TestStandardEncrypt_EncryptionKeyCalculationsWhenPasswordIsNotEmpty()
        {
            this.TestEncryptionKeyCalculations(CreateExportSettingsWithPassword(), ExpectedOwnerPasswordWhenNonEmptyPassword, ExpectedEncryptionKeyWhenNonEmptyPassword);
        }

        [TestMethod]
        public void TestStandardEncrypt_UserPasswordEncryptionWhenPasswordIsNotEmpty()
        {
            byte[] expectedUserPassword = { 0xAF, 0x52, 0x8C, 0x84, 0xB9, 0xA4, 0x15, 0x8F, 0xB3, 0x2C, 0x27, 0x97, 0xDA, 0x2D, 0x99, 0xD9, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
            this.TestUserPasswordEncryption(ExpectedEncryptionKeyWhenNonEmptyPassword, expectedUserPassword);
        }

        private void TestOwnerPasswordEncryption(PdfExportSettings settings, byte[] expectedOwnerPass)
        {
            int revision = FixedDocumentDefaults.StandardEncryptRevision;
            int keyLength = FixedDocumentDefaults.StandardEncryptionKeyLength;
            byte[] actualOwnerPass = StandardEncrypt.CalculateOwnerPassword(settings.OwnerPassword, settings.UserPassword, revision, keyLength);

            AssertByteArrays(actualOwnerPass, expectedOwnerPass);
        }

        private void TestEncryptionKeyCalculations(PdfExportSettings settings, byte[] ownerPasswordValue, byte[] expectedKey)
        {
            int revision = FixedDocumentDefaults.StandardEncryptRevision;
            int keyLength = FixedDocumentDefaults.StandardEncryptionKeyLength;
            byte[] actualKey = StandardEncrypt.CalculateEncryptionKey(settings.UserPassword, DocumentId, keyLength, revision, ownerPasswordValue, PermissionsDefaultValue, false);

            AssertByteArrays(actualKey, expectedKey);
        }

        private void TestUserPasswordEncryption(byte[] encryptionKey, byte[] expectedUserPassword)
        {
            byte[] actualUserPass = StandardEncrypt.CalculateUserPassword(encryptionKey, DocumentId);

            AssertByteArrays(actualUserPass, expectedUserPassword);
        }

        private static PdfExportSettings CreateEmptyExportSettings()
        {
            return new PdfExportSettings();
        }

        private static PdfExportSettings CreateExportSettingsWithPassword()
        {
            return new PdfExportSettings() { UserPassword = "test password" };
        }

        private static void AssertByteArrays(byte[] firstArray, byte[] secondArray)
        {
            if (firstArray == null)
            {
                Assert.IsTrue(secondArray == null, "One array is null and the other is not!");
            }

            Assert.AreEqual(firstArray.Length, secondArray.Length, "Arrays must have same elements count!");

            for (int i = 0; i < firstArray.Length; i++)
            {
                byte first = firstArray[i];
                byte second = secondArray[i];

                Assert.AreEqual(first, second, string.Format("Elements at index {0} are different!", i));
            }
        }
    }
}
