﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Collections.Generic;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Fixed.Model.Editing;
using Telerik.Windows.Documents.Fixed.Model.Editing.Collections;
using Telerik.Windows.Documents.Fixed.Model.Editing.Layout;
using Telerik.Windows.Documents.Fixed.Model.Editing.Tables;

namespace Telerik.Windows.Documents.Fixed.Tests.Flow
{
    [TestClass]
    public class TableTests
    {
        [TestMethod]
        public void Test_MeasureTableWithoutCells()
        {
            Size expectedSize = new Size();

            Table table = new Table();
            Size emptyTableSize = table.Measure();
            Assert.AreEqual(expectedSize, emptyTableSize);

            table.Rows.AddTableRow();
            Size emptyRowSize = table.Measure();
            Assert.AreEqual(expectedSize, emptyRowSize);
        }

        [TestMethod]
        public void Test_SingleRowMultipleCellsTableSplit()
        {
            Table table = new Table();
            table.Rows.Add(GenerateTableRow(new Size[][][]
                {
                    new Size[][] { new Size[] { new Size(10, 10), new Size(10, 20) }},
                    new Size[][] { new Size[] { new Size(10, 10), new Size(10, 15) }},
                    new Size[][] { new Size[] { new Size(20, 15) }},
                }));

            Table expectedSplit = new Table();
            expectedSplit.Rows.Add(GenerateTableRow(new Size[][][]
                {
                    new Size[][] { new Size[] { new Size(10, 20) }},
                    new Size[][] { new Size[] { new Size(10, 15) }},
                    new Size[][] { },
                }));

            SplitAndAssertTable(table, expectedSplit, new Size(45, 15));
        }

        [TestMethod]
        public void Test_SingleRowTableSplit()
        {
            Table table = new Table();
            table.Rows.Add(GenerateTableRow(new Size[][][]
                {
                    new Size[][] { new Size[] { new Size(10, 10), new Size(10, 20) }},
                }));

            Table expectedSplit = new Table();
            expectedSplit.Rows.Add(GenerateTableRow(new Size[][][]
                {
                    new Size[][] { new Size[] { new Size(10, 20) }},
                }));

            SplitAndAssertTable(table, expectedSplit, new Size(10, 10));
        }

        [TestMethod]
        public void Test_TableWithPreferredWidthWithMergedCellOnTop()
        {
            Table table = new Table();

            TableRow firstRow = table.Rows.AddTableRow();
            TableCell mergedCell = firstRow.Cells.AddTableCell();
            mergedCell.ColumnSpan = 3;
            mergedCell.PreferredWidth = 120;

            TableRow secondRow = table.Rows.AddTableRow();
            TableCell firstCell = secondRow.Cells.AddTableCell();
            firstCell.PreferredWidth = 30;
            TableCell secondCell = secondRow.Cells.AddTableCell();
            secondCell.PreferredWidth = 40;
            TableCell thirdCell = secondRow.Cells.AddTableCell();
            thirdCell.PreferredWidth = 50;

            table.Measure();

            double mergedWidth = mergedCell.LayoutRectangle.Width;
            Assert.AreEqual(mergedCell.PreferredWidth.Value, mergedWidth);

            double firstWidth = firstCell.LayoutRectangle.Width;
            Assert.AreEqual(firstCell.PreferredWidth.Value, firstWidth);

            double secondWidth = secondCell.LayoutRectangle.Width;
            Assert.AreEqual(secondCell.PreferredWidth.Value, secondWidth);

            double thirdWidth = thirdCell.LayoutRectangle.Width;
            Assert.AreEqual(thirdCell.PreferredWidth.Value, thirdWidth);
        }

        private static void SplitAndAssertTable(Table table, Table expectedSplit, Size measureSize)
        {
            table.Measure(measureSize);

            Assert.IsTrue(table.HasPendingContent);

            Table actualSplit = table.Split();
            expectedSplit.Measure();

            AssertTableCells(new List<TableCell>(expectedSplit.Cells), new List<TableCell>(actualSplit.Cells));
        }

        private static void AssertTableCells(List<TableCell> expectedCells, List<TableCell> actualCells)
        {
            Assert.AreEqual(expectedCells.Count, actualCells.Count);

            for (int cellIndex = 0; cellIndex < actualCells.Count; cellIndex++)
            {
                TableCell expectedCell = expectedCells[cellIndex];
                TableCell actualCell = actualCells[cellIndex];

                AssertCellBlocks(expectedCell.Blocks, actualCell.Blocks);
            }
        }

        private static void AssertCellBlocks(BlockCollection expectedBlocks, BlockCollection actualBlocks)
        {
            Assert.AreEqual(expectedBlocks.Count, actualBlocks.Count);

            for (int blockIndex = 0; blockIndex < actualBlocks.Count; blockIndex++)
            {
                List<LayoutElementBase> expectedElements = GetLayoutElements((Block)expectedBlocks[blockIndex]);
                List<LayoutElementBase> actualElements = GetLayoutElements((Block)actualBlocks[blockIndex]);

                Assert.AreEqual(expectedElements.Count, actualElements.Count);

                for (int elementIndex = 0; elementIndex < expectedElements.Count; elementIndex++)
                {
                    LayoutElementBase expectedElement = expectedElements[elementIndex];
                    LayoutElementBase actualElement = actualElements[elementIndex];

                    Assert.AreEqual(expectedElement.Width, actualElement.Width);
                    Assert.AreEqual(expectedElement.Height, actualElement.Height);
                }
            }
        }

        private static List<LayoutElementBase> GetLayoutElements(Block block)
        {
            List<LayoutElementBase> elements = new List<LayoutElementBase>();

            foreach (LineInfo line in block.Lines)
            {
                foreach (LayoutElementBase element in line.Elements)
                {
                    elements.Add(element);
                }
            }

            foreach (LayoutElementBase element in block.PendingElements)
            {
                elements.Add(element);
            }

            return elements;
        }

        private static TableRow GenerateTableRow(Size[][][] cellsBlocksContentBoxes)
        {
            TableRow row = new TableRow();

            if (cellsBlocksContentBoxes != null)
            {
                foreach (Size[][] cellBlocksContentBoxes in cellsBlocksContentBoxes)
                {
                    row.Cells.Add(GenerateTableCell(cellBlocksContentBoxes));
                }
            }

            return row;
        }

        private static TableCell GenerateTableCell(Size[][] cellBlocksContentBoxes)
        {
            TableCell cell = new TableCell();

            if (cellBlocksContentBoxes != null)
            {
                foreach (Size[] blockContentBoxes in cellBlocksContentBoxes)
                {
                    Block block = cell.Blocks.AddBlock();
                    block.GraphicProperties.IsStroked = false;

                    if (blockContentBoxes != null)
                    {
                        foreach (Size contentBox in blockContentBoxes)
                        {
                            block.InsertRectangle(new Rect(0, 0, contentBox.Width, contentBox.Height));
                        }
                    }
                }
            }

            return cell;
        }
    }
}
