﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.IO;
#if !NETSTANDARD
using System.Windows;
using System.Windows.Media;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Editing;
using Telerik.Windows.Documents.Fixed.Model.Text;
#endif
using Telerik.Windows.Documents.Core.Fonts;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser.Keywords;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Fonts;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Fonts.Cid;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Tests.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.Tests.Import;

namespace Telerik.Windows.Documents.Fixed.Tests.Fonts
{
    [TestClass]
    public class CidFontTests
    {
        private readonly TestPdfImportContext context;
        private readonly PostScriptReader reader;

        public CidFontTests()
        {
            this.context = new TestPdfImportContext();
            this.reader = new PostScriptReader(new MemoryStream(), new KeywordCollection());
        }

        [TestMethod]
        public void CidFontWidths_WhenIntAndArray_AreImportedCorrectly()
        {

            // Arrange
            // W[33[500]]
            CidFontType2Object expectedFont = GetExpectedFontWithWidths(new PdfArray(new PdfInt(33), new PdfArray(new PdfInt(500))));

            // Act
            CidType2Font actualFont = new CidType2Font("test", EmptyFontSource.Instance);
            expectedFont.CopyPropertiesToFont(actualFont, this.reader, this.context);

            // Assert
            this.AssertArraysAreEqual(expectedFont.Widths, this.ConvertCidWidthsToPdfArray(actualFont.Widths.Value));
        }

        [TestMethod]
        public void CidFontWidths_WhenIntAndIndirectArray_AreImportedCorrectly()
        {
            IndirectReference reference = new IndirectReference(1, 0);
            this.context.IndirectObjectsToRead[reference] = new IndirectObject(reference, new PdfArray(new PdfInt(500)));

            // Arrange
            // W[33[500]]
            CidFontType2Object expectedFont = GetExpectedFontWithWidths(new PdfArray(new PdfInt(33), reference));

            // Act
            CidType2Font actualFont = new CidType2Font("test", EmptyFontSource.Instance);
            expectedFont.CopyPropertiesToFont(actualFont, this.reader, this.context);

            // Assert
            this.AssertArraysAreEqual(expectedFont.Widths, this.ConvertCidWidthsToPdfArray(actualFont.Widths.Value));
        }

        [TestMethod]
        public void CidFontWidths_WhenHasMultipleIntegersAndArraysAndEndsWithArray_AreImportedCorrectly()
        {
            // Arrange
            // W[3[250]11 12 333 16[333]19 28 500 36[722 667]38[667 722 611 556]]
            CidFontType2Object expectedFont = GetExpectedFontWithWidths(new PdfArray(
                new PdfInt(3), new PdfArray(new PdfInt(250)),
                new PdfInt(11), new PdfInt(12), new PdfInt(333),
                new PdfInt(16), new PdfArray(new PdfInt(333)),
                new PdfInt(19), new PdfInt(28), new PdfInt(500),
                new PdfInt(36), new PdfArray(new PdfInt(722), new PdfInt(667)),
                new PdfInt(38), new PdfArray(new PdfInt(667), new PdfInt(722), new PdfInt(611), new PdfInt(556))));

            // Act
            CidType2Font actualFont = new CidType2Font("test", EmptyFontSource.Instance);
            expectedFont.CopyPropertiesToFont(actualFont, this.reader, this.context);

            // Assert
            this.AssertArraysAreEqual(expectedFont.Widths, this.ConvertCidWidthsToPdfArray(actualFont.Widths.Value));
        }

        [TestMethod]
        public void CidFontWidths_WhenHasMultipleIntegersAndArraysAndEndsWithIntegers_AreImportedCorrectly()
        {
            // Arrange
            // W[3[250]11 12 333 17[250]19 28 500]
            CidFontType2Object expectedFont = GetExpectedFontWithWidths(new PdfArray(
                new PdfInt(3), new PdfArray(new PdfInt(250)),
                new PdfInt(11), new PdfInt(12), new PdfInt(333),
                new PdfInt(17), new PdfArray(new PdfInt(250)),
                new PdfInt(19), new PdfInt(28), new PdfInt(500)));

            // Act
            CidType2Font actualFont = new CidType2Font("test", EmptyFontSource.Instance);
            expectedFont.CopyPropertiesToFont(actualFont, this.reader, this.context);

            // Assert
            this.AssertArraysAreEqual(expectedFont.Widths, this.ConvertCidWidthsToPdfArray(actualFont.Widths.Value));
        }

        [TestMethod]
        public void CidSet_Export_WhenFontHasNoUsedCharacters()
        {
            CidSet set = new CidSet();
            CidType2Font font = new CidType2Font("test", EmptyFontSource.Instance);
            TestPdfExportContext context = new TestPdfExportContext();
            set.CopyPropertiesFrom(context, font);

            Assert.IsNotNull(set.Data);
            Assert.AreEqual(0, set.Data.Length);
        }

#if !NETSTANDARD
        [TestMethod]
        public void AssertCharIdToGlyphIdMapping_IsIdentity_WhenExportDocumentWithCidFontAndImportItAgain()
        {
            using (MemoryStream stream = new MemoryStream())
            {
                PdfFormatProvider provider = new PdfFormatProvider();

                //Add a page Number
                Block block = new Block();

                block.TextProperties.TrySetFont(new FontFamily("Arial"));

                block.InsertText("123");

                RadFixedDocument document = new RadFixedDocument();
                FixedContentEditor editor = new FixedContentEditor(document.Pages.AddPage());
                editor.DrawBlock(block);

                provider.Export(document, stream);

                stream.Seek(0, SeekOrigin.Begin);

                document = provider.Import(stream);

                RadFixedPage firstPage = document.Pages[0];
                foreach (ContentElementBase content in firstPage.Content)
                {
                    TextFragment textFragment = content as TextFragment;
                    if (textFragment != null)
                    {
                        CidType2Font font = textFragment.Font as CidType2Font;

                        Assert.IsNotNull(font);

                        Assert.IsTrue(font.CharIdToGlyphIdMapping.HasValue);

                        Assert.IsTrue(font.CharIdToGlyphIdMapping.Value.IsIdentityMapping);
                    }
                }
            }
        }
#endif

        private void AssertArraysAreEqual(PdfArray expectedArray, PdfArray actualArray)
        {
            Assert.AreEqual(expectedArray.Count, actualArray.Count);

            for (int i = 0; i < expectedArray.Count; i++)
            {
                IPdfPrimitive expectedPrimitive = this.GetDirectPrimitive(expectedArray[i]);
                IPdfPrimitive actualPrimitive = this.GetDirectPrimitive(actualArray[i]);

                Assert.AreEqual(expectedPrimitive.Type, actualPrimitive.Type);

                if (expectedPrimitive.Type == PdfElementType.PdfInt)
                {
                    Assert.AreEqual(expectedPrimitive, actualPrimitive);
                }
                else if (expectedPrimitive.Type == PdfElementType.Array)
                {
                    this.AssertArraysAreEqual((PdfArray)expectedPrimitive, (PdfArray)actualPrimitive);
                }
            }
        }

        private IPdfPrimitive GetDirectPrimitive(IPdfPrimitive primitive)
        {
            if (primitive.Type == PdfElementType.IndirectReference)
            {
                IndirectReference reference = (IndirectReference)primitive;
                IPdfPrimitive directPrimitive = this.context.ReadIndirectObject(reference).Content;

                return directPrimitive;
            }

            return primitive;
        }

        private PdfArray ConvertCidWidthsToPdfArray(CidWidths<int> cidWidths)
        {
            PdfArray array = new PdfArray();

            foreach (CidWidthsRange<int> range in cidWidths)
            {
                array.Add(new PdfInt(range.StartCharCode));

                if (range.IsConstantWidthRange)
                {
                    array.Add(new PdfInt(range.EndCharCode));
                    array.Add(new PdfInt(range.ConstantWidth));
                }
                else
                {
                    PdfArray widthsArray = new PdfArray();

                    foreach (int width in range.Widths)
                    {
                        widthsArray.Add(new PdfInt(width));
                    }

                    array.Add(widthsArray);
                }
            }

            return array;
        }

        private static CidFontType2Object GetExpectedFontWithWidths(PdfArray widths)
        {
            CidFontType2Object expectedFont = new CidFontType2Object();
            expectedFont.FontDescriptor = new FontDescriptor();
            expectedFont.Widths = widths;

            return expectedFont;
        }
    }
}