﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.IO;
using System.Text;
using Telerik.Windows.Documents.Core.PostScript.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Data;
using PDFReader = Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader.Readers.PdfReader;
using Telerik.Windows.Documents.Fixed.Tests.Utils;

namespace Telerik.Windows.Documents.Fixed.Tests.FormatProviders.Old.Pdf.DocumentModel.Core
{
    [TestClass]
    public class ContentStreamOldTests
    {
        private const string FirstStreamElement = @"q
            /F1 21 Tf
            1 0 0 1 51.42 730.86 Tm
            0.384 0.392 0.388 rg
            0.384 0.392 0.388 RG
            Q";

        private const string SecondStreamElement = @"q
            /F1 21 Tf
            1 0 0 1 568.92 769.56 Tm
            0 g
            0 G
            Q";

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void AppendData_WhenStreamToAppendIsNull_ShouldThrow()
        {
            // Arrange
            ContentStreamOld contentStream = new ContentStreamOld(TestUtils.ContentManager);

            // Act && Assert
            contentStream.AppendData(null);
        }

        [TestMethod]
        public void AppendData_ShouldAppendDatatWithWhiteSpace()
        {
            // Arrange
            Stream firstStreamElement = new MemoryStream(Encoding.UTF8.GetBytes(FirstStreamElement + Keywords.StreamEnd + " "));
            PDFReader firstReader = new PDFReader(firstStreamElement);
            PdfDictionaryOld dictionary = new PdfDictionaryOld(TestUtils.ContentManager);
            PdfDataStream firstPdfData = new PdfDataStream(firstReader, dictionary);

            Stream secondStreamElement = new MemoryStream(Encoding.UTF8.GetBytes(SecondStreamElement + Keywords.StreamEnd + " "));
            PDFReader secondReader = new PDFReader(secondStreamElement);
            PdfDataStream secondPdfData = new PdfDataStream(secondReader, dictionary);

            ContentStreamOld contentStreamOld = new ContentStreamOld(TestUtils.ContentManager);

            string expected = FirstStreamElement + " " + SecondStreamElement;

            // Act
            contentStreamOld.AppendData(firstPdfData);
            contentStreamOld.AppendData(secondPdfData);

            // Assert
            string result = Encoding.UTF8.GetString(contentStreamOld.Data);
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void AppendData_WhenDataToAppendIsEmpty_ShouldSetEmptyArray()
        {
            // Arrange
            Stream streamElement = new MemoryStream(Encoding.UTF8.GetBytes(Keywords.StreamEnd + " "));
            PDFReader reader = new PDFReader(streamElement);
            PdfDictionaryOld dictionary = new PdfDictionaryOld(TestUtils.ContentManager);
            PdfDataStream pdfData = new PdfDataStream(reader, dictionary);

            ContentStreamOld contentStreamOld = new ContentStreamOld(TestUtils.ContentManager);

            // Act
            contentStreamOld.AppendData(pdfData);

            // Assert
            Assert.IsNotNull(contentStreamOld.Data);
            Assert.AreEqual(0, contentStreamOld.Data.Length);
        }

        [TestMethod]
        public void AppendData_ShouldSetIsLoadedProperty()
        {
            // Arrange
            Stream firstStreamElement = new MemoryStream(Encoding.UTF8.GetBytes(FirstStreamElement + Keywords.StreamEnd + " "));
            PDFReader firstReader = new PDFReader(firstStreamElement);
            PdfDictionaryOld dictionary = new PdfDictionaryOld(TestUtils.ContentManager);
            PdfDataStream firstPdfData = new PdfDataStream(firstReader, dictionary);

            ContentStreamOld contentStreamOld = new ContentStreamOld(TestUtils.ContentManager);

            // Act
            contentStreamOld.AppendData(firstPdfData);

            // Assert
            Assert.IsTrue(contentStreamOld.IsLoaded);
        }

        [TestMethod]
        public void AppendData_WhenFirstStreamElementEndsWithWhiteSpace_ShouldNotAppendhWhiteSpace()
        {
            // Arrange
            Stream firstStreamElement = new MemoryStream(Encoding.UTF8.GetBytes(FirstStreamElement + " " + Keywords.StreamEnd + " "));
            PDFReader firstReader = new PDFReader(firstStreamElement);
            PdfDictionaryOld dictionary = new PdfDictionaryOld(TestUtils.ContentManager);
            PdfDataStream firstPdfData = new PdfDataStream(firstReader, dictionary);

            Stream secondStreamElement = new MemoryStream(Encoding.UTF8.GetBytes(SecondStreamElement + Keywords.StreamEnd + " "));
            PDFReader secondReader = new PDFReader(secondStreamElement);
            PdfDataStream secondPdfData = new PdfDataStream(secondReader, dictionary);

            ContentStreamOld contentStreamOld = new ContentStreamOld(TestUtils.ContentManager);

            string expected = FirstStreamElement + " " + SecondStreamElement;

            // Act
            contentStreamOld.AppendData(firstPdfData);
            contentStreamOld.AppendData(secondPdfData);

            // Assert
            string result = Encoding.UTF8.GetString(contentStreamOld.Data);
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void AppendData_WhenSecondStreamElementStartsWithWhiteSpace_ShouldNotAppendWhiteSpace()
        {
            // Arrange
            Stream firstStreamElement = new MemoryStream(Encoding.UTF8.GetBytes(FirstStreamElement + Keywords.StreamEnd + " "));
            PDFReader firstReader = new PDFReader(firstStreamElement);
            PdfDictionaryOld dictionary = new PdfDictionaryOld(TestUtils.ContentManager);
            PdfDataStream firstPdfData = new PdfDataStream(firstReader, dictionary);

            Stream secondStreamElement = new MemoryStream(Encoding.UTF8.GetBytes(" " + SecondStreamElement + Keywords.StreamEnd + " "));
            PDFReader secondReader = new PDFReader(secondStreamElement);
            PdfDataStream secondPdfData = new PdfDataStream(secondReader, dictionary);

            ContentStreamOld contentStreamOld = new ContentStreamOld(TestUtils.ContentManager);

            string expected = FirstStreamElement + " " + SecondStreamElement;

            // Act
            contentStreamOld.AppendData(firstPdfData);
            contentStreamOld.AppendData(secondPdfData);

            // Assert
            string result = Encoding.UTF8.GetString(contentStreamOld.Data);
            Assert.AreEqual(expected, result);
        }
    }
}
