﻿using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.Actions;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Core.Navigation;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
using Telerik.Windows.Documents.Fixed.Model.Actions;
using Telerik.Windows.Documents.Fixed.Model.Navigation;
using Telerik.Windows.Documents.Fixed.Tests.Utils;

namespace Telerik.Windows.Documents.Fixed.Tests.FormatProviders.Old.Pdf
{
    [TestClass]
    public class PdfElementToFixedElementTranslatorTests
    {
        private PdfContentManager contentManager = TestUtils.ContentManager;

        private const string RootTitle = "Root";
        private const string FirstNodeTitle = "First Node";
        private const string SecondNodeTitle = "Second Node";
        private const string FirstLeafTitle = "First Leaf";
        private const string SecondLeafTitle = "Second Leaf";

        [TestMethod]
        public void CreateBookmarkItemHierarchy_WhenCountIsAboveZero_ShouldCreateNodesWithIsExpandedSetToTrue()
        {
            // Arrange
            OutlineItemOld rootOutline = new OutlineItemOld(this.contentManager);
            rootOutline.Count = new PdfIntOld(this.contentManager, int.MaxValue);

            OutlineItemOld firstOutlineNode = new OutlineItemOld(this.contentManager);
            firstOutlineNode.Count = new PdfIntOld(this.contentManager, 1);
            rootOutline.First = firstOutlineNode;

            PdfFormatProvider provider = new PdfFormatProvider();

            // Act
            BookmarkItem bookmark = PdfElementToFixedElementTranslator.CreateBookmarkItemHierarchy(rootOutline, provider);

            // Assert
            Assert.IsTrue(bookmark.IsExpanded);

            BookmarkItem childBookmark = bookmark.Children[0];
            Assert.IsTrue(childBookmark.IsExpanded);
        }

        [TestMethod]
        public void CreateBookmarkItemHierarchy_WhenCountIsEqualOrBelowZero_ShouldCreateNodesWithIsExpandedSetToFalse()
        {
            // Arrange
            OutlineItemOld rootOutline = new OutlineItemOld(this.contentManager);
            rootOutline.Count = new PdfIntOld(this.contentManager, 0);

            OutlineItemOld firstOutlineNode = new OutlineItemOld(this.contentManager);
            firstOutlineNode.Count = new PdfIntOld(this.contentManager, int.MinValue);
            rootOutline.First = firstOutlineNode;

            PdfFormatProvider provider = new PdfFormatProvider();

            // Act
            BookmarkItem bookmark = PdfElementToFixedElementTranslator.CreateBookmarkItemHierarchy(rootOutline, provider);

            // Assert
            Assert.IsFalse(bookmark.IsExpanded);

            BookmarkItem childBookmark = bookmark.Children[0];
            Assert.IsFalse(childBookmark.IsExpanded);
        }

        [TestMethod]
        public void CreateBookmarkItemHierarchy_WhenCountIsNull_ShouldCreateNodesWithIsExpandedSetToFalse()
        {
            // Arrange
            OutlineItemOld rootOutline = new OutlineItemOld(this.contentManager);

            OutlineItemOld firstOutlineNode = new OutlineItemOld(this.contentManager);
            rootOutline.First = firstOutlineNode;

            PdfFormatProvider provider = new PdfFormatProvider();

            // Act
            BookmarkItem bookmark = PdfElementToFixedElementTranslator.CreateBookmarkItemHierarchy(rootOutline, provider);

            // Assert
            Assert.IsFalse(bookmark.IsExpanded);

            BookmarkItem childBookmark = bookmark.Children[0];
            Assert.IsFalse(childBookmark.IsExpanded);
        }

        [TestMethod]
        public void CreateBookmarkItemHierarchy_WhenTitleIsNull_ShouldCreateNodesWithEmptyTitle()
        {
            // Arrange
            OutlineItemOld rootOutline = new OutlineItemOld(this.contentManager);

            OutlineItemOld firstOutlineNode = new OutlineItemOld(this.contentManager);
            rootOutline.First = firstOutlineNode;

            PdfFormatProvider provider = new PdfFormatProvider();

            // Act
            BookmarkItem bookmark = PdfElementToFixedElementTranslator.CreateBookmarkItemHierarchy(rootOutline, provider);

            // Assert
            Assert.AreEqual(string.Empty, bookmark.Title);

            BookmarkItem childBookmark = bookmark.Children[0];
            Assert.AreEqual(string.Empty, childBookmark.Title);
        }

        [TestMethod]
        public void CreateBookmarkItemHierarchy_WhenTitleIsNotNull_ShouldCreateNodesWithCorrectTitle()
        {
            // Arrange
            OutlineItemOld rootOutline = new OutlineItemOld(this.contentManager);
            rootOutline.Title = new PdfStringOld(this.contentManager, Encoding.ASCII.GetBytes(RootTitle));

            OutlineItemOld firstOutlineNode = new OutlineItemOld(this.contentManager);
            firstOutlineNode.Title = new PdfStringOld(this.contentManager, Encoding.ASCII.GetBytes(FirstNodeTitle));
            rootOutline.First = firstOutlineNode;

            PdfFormatProvider provider = new PdfFormatProvider();

            // Act
            BookmarkItem bookmark = PdfElementToFixedElementTranslator.CreateBookmarkItemHierarchy(rootOutline, provider);

            // Assert
            Assert.AreEqual(RootTitle, bookmark.Title);

            BookmarkItem childBookmark = bookmark.Children[0];
            Assert.AreEqual(FirstNodeTitle, childBookmark.Title);
        }

        [TestMethod]
        public void CreateBookmarkItemHierarchy_WhenActionIsNull_ShouldCreateNodesWithNullAction()
        {
            // Arrange
            OutlineItemOld rootOutline = new OutlineItemOld(this.contentManager);

            OutlineItemOld firstOutlineNode = new OutlineItemOld(this.contentManager);
            rootOutline.First = firstOutlineNode;

            PdfFormatProvider provider = new PdfFormatProvider();

            // Act
            BookmarkItem bookmark = PdfElementToFixedElementTranslator.CreateBookmarkItemHierarchy(rootOutline, provider);

            // Assert
            Assert.IsNull(bookmark.Action);

            BookmarkItem childBookmark = bookmark.Children[0];
            Assert.IsNull(childBookmark.Action);
        }

        [TestMethod]
        public void CreateBookmarkItemHierarchy_WhenActionIsNotNull_ShouldCreateNodeWithCorrectAction()
        {
            // Arrange
            OutlineItemOld otuline = new OutlineItemOld(this.contentManager);
            otuline.Title = new PdfStringOld(this.contentManager, Encoding.ASCII.GetBytes(RootTitle));
            GoToActionOld action = new GoToActionOld(this.contentManager);
            DestinationOld destination = new DestinationOld(contentManager);
            destination.Type = DestinationTypeOld.XYZ;
            action.Destination = destination;
            otuline.Action = action;

            PdfFormatProvider provider = new PdfFormatProvider();

            // Act
            BookmarkItem bookmark = PdfElementToFixedElementTranslator.CreateBookmarkItemHierarchy(otuline, provider);

            // Assert
            GoToAction goToAction = bookmark.Action as GoToAction;
            Assert.IsNotNull(goToAction);

            Location location = goToAction.Destination as Location;
            Assert.IsNotNull(location);

            Assert.AreEqual(location.DestinationType, DestinationType.Location);
        }

        [TestMethod]
        public void CreateBookmarkItemHierarchy_WhenDestinationIsNull_ShouldCreateNodesWithNullDestination()
        {
            // Arrange
            OutlineItemOld rootOutline = new OutlineItemOld(this.contentManager);

            OutlineItemOld firstOutlineNode = new OutlineItemOld(this.contentManager);
            rootOutline.First = firstOutlineNode;

            PdfFormatProvider provider = new PdfFormatProvider();

            // Act
            BookmarkItem bookmark = PdfElementToFixedElementTranslator.CreateBookmarkItemHierarchy(rootOutline, provider);

            // Assert
            Assert.IsNull(bookmark.Destination);

            BookmarkItem childBookmark = bookmark.Children[0];
            Assert.IsNull(childBookmark.Destination);
        }

        [TestMethod]
        public void CreateBookmarkItemHierarchy_WhenDestinationIsNotNull_ShouldCreateNodeWithCorrectDestination()
        {
            // Arrange
            OutlineItemOld otuline = new OutlineItemOld(this.contentManager);
            otuline.Title = new PdfStringOld(this.contentManager, Encoding.ASCII.GetBytes(RootTitle));
            DestinationOld destination = new DestinationOld(contentManager);
            destination.Type = DestinationTypeOld.XYZ;
            otuline.Destination = destination;

            PdfFormatProvider provider = new PdfFormatProvider();

            // Act
            BookmarkItem bookmark = PdfElementToFixedElementTranslator.CreateBookmarkItemHierarchy(otuline, provider);

            // Assert
            Location location = bookmark.Destination as Location;
            Assert.IsNotNull(location);

            Assert.AreEqual(location.DestinationType, DestinationType.Location);
        }

        [TestMethod]
        public void CreateBookmarkItemHierarchy_ShouldCreateValidBookmarksTree()
        {
            // Arrange
            OutlineItemOld outline = this.CreateOutlinesTree();
            PdfFormatProvider provider = new PdfFormatProvider();

            // Act
            BookmarkItem bookmark = PdfElementToFixedElementTranslator.CreateBookmarkItemHierarchy(outline, provider);

            // Assert
            this.AssertBookmarksTree(bookmark);
        }

        private OutlineItemOld CreateOutlinesTree()
        {
            OutlineItemOld root = new OutlineItemOld(this.contentManager);
            root.Title = new PdfStringOld(this.contentManager, Encoding.ASCII.GetBytes(RootTitle));

            OutlineItemOld firstNode = new OutlineItemOld(this.contentManager);
            firstNode.Title = new PdfStringOld(this.contentManager, Encoding.ASCII.GetBytes(FirstNodeTitle));
            root.First = firstNode;

            OutlineItemOld secondNode = new OutlineItemOld(this.contentManager);
            secondNode.Title = new PdfStringOld(this.contentManager, Encoding.ASCII.GetBytes(SecondNodeTitle));
            firstNode.Next = secondNode;

            OutlineItemOld firstLeaf = new OutlineItemOld(this.contentManager);
            firstLeaf.Title = new PdfStringOld(this.contentManager, Encoding.ASCII.GetBytes(FirstLeafTitle));
            firstNode.First = firstLeaf;

            OutlineItemOld secondLeaf = new OutlineItemOld(this.contentManager);
            secondLeaf.Title = new PdfStringOld(this.contentManager, Encoding.ASCII.GetBytes(SecondLeafTitle));
            secondNode.First = secondLeaf;

            return root;
        }

        private void AssertBookmarksTree(BookmarkItem root)
        {
            Assert.AreEqual(RootTitle, root.Title);
            Assert.AreEqual(2, root.Children.Count);

            BookmarkItem firstNode = root.Children[0];
            Assert.AreEqual(FirstNodeTitle, firstNode.Title);
            Assert.AreEqual(1, firstNode.Children.Count);

            BookmarkItem secondNode = root.Children[1];
            Assert.AreEqual(SecondNodeTitle, secondNode.Title);
            Assert.AreEqual(1, secondNode.Children.Count);

            BookmarkItem firstLeaf = firstNode.Children[0];
            Assert.AreEqual(FirstLeafTitle, firstLeaf.Title);
            Assert.AreEqual(0, firstLeaf.Children.Count);

            BookmarkItem secondLeaf = secondNode.Children[0];
            Assert.AreEqual(SecondLeafTitle, secondLeaf.Title);
            Assert.AreEqual(0, secondLeaf.Children.Count);
        }
    }
}
