﻿using System;
using System.IO;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.DocumentModel.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader.Parsers;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;

namespace Telerik.Windows.Documents.Fixed.Tests.FormatProviders.Old.Pdf.PdfReader.Parsers
{
    [TestClass]
    public class PdfParserTests
    {
        private const string XRefWithLineFeed = "xref\n0 3\n0000000000 65535 f\n\n0000019690 00000 n\n\n0000000015 00000 n\n\ntrailer\n <</Size 3 /Root 117 0 R /Info 115 0 R /ID[<0A5B8D456FA23DE6E112438E3AD4A7DF><0A5B8D456FA23DE6E112438E3AD4A7DF>] >> \nstartxref\n0\n%%EOF";

        private const string XRefWitManyLineFeeds = "xref\n0 3\n0000000000 65535 f\n\n\n0000019690 00000 n\n\n\n\n\n0000000015 00000 n\n\n\n\n\ntrailer\n <</Size 3 /Root 117 0 R /Info 115 0 R /ID[<0A5B8D456FA23DE6E112438E3AD4A7DF><0A5B8D456FA23DE6E112438E3AD4A7DF>] >> \nstartxref\n0\n%%EOF";

        private const string XRefWithEmtySpaces = "xref\n0 3\n0000000000 65535 f\n \n0000019690 00000 n\n \n0000000015 00000 n\n \ntrailer\n <</Size 3 /Root 117 0 R /Info 115 0 R /ID[<0A5B8D456FA23DE6E112438E3AD4A7DF><0A5B8D456FA23DE6E112438E3AD4A7DF>] >> \nstartxref\n0\n%%EOF";

        private const string XRefWithCariageReturn = "xref\r0 3\r0000000000 65535 f\r\r0000019690 00000 n\r\r0000000015 00000 n\r\rtrailer\r <</Size 3 /Root 117 0 R /Info 115 0 R /ID[<0A5B8D456FA23DE6E112438E3AD4A7DF><0A5B8D456FA23DE6E112438E3AD4A7DF>] >> \rstartxref\r0\r%%EOF";

        private const string XRefWithCariageReturnAndLineFeed = "xref\r\n0 3\r\n0000000000 65535 f\r\n\r\n0000019690 00000 n\r\n\r\n0000000015 00000 n\r\n\r\ntrailer\r\n <</Size 3 /Root 117 0 R /Info 115 0 R /ID[<0A5B8D456FA23DE6E112438E3AD4A7DF><0A5B8D456FA23DE6E112438E3AD4A7DF>] >> \r\nstartxref\r\n0\r\n%%EOF";

        private const string XRefWitManyCarriageReturn = "xref\r0 3\r0000000000 65535 f\r\r\r0000019690 00000 n\r\r\r\r\r0000000015 00000 n\r\r\r\r\rtrailer\r <</Size 3 /Root 117 0 R /Info 115 0 R /ID[<0A5B8D456FA23DE6E112438E3AD4A7DF><0A5B8D456FA23DE6E112438E3AD4A7DF>] >> \rstartxref\r0\r%%EOF";

        [TestMethod]
        public void InitializeCrossReferences_WhenReferenceEntriesAreSeparatedWithOneLineFeed_ShouldParseEntriesCorrectly()
        {
            // Arrange
            Stream stream = new MemoryStream(Encoding.UTF8.GetBytes(XRefWithLineFeed));
            PdfContentManager contentManager = new PdfContentManager(new PdfFormatProvider(), stream);
            PdfParser pdfParser = new PdfParser(contentManager, stream);

            // Act
            pdfParser.InitializeCrossReferences();

            // Assert
            this.AssertCrossReferenceEntries(contentManager);
        }

        [TestMethod]
        public void InitializeCrossReferences_WhenReferenceEntriesAreSeparatedWithManyLineFeeds_ShouldParseEntriesCorrectly()
        {
            // Arrange
            Stream stream = new MemoryStream(Encoding.UTF8.GetBytes(XRefWitManyLineFeeds));
            PdfContentManager contentManager = new PdfContentManager(new PdfFormatProvider(), stream);
            PdfParser pdfParser = new PdfParser(contentManager, stream);

            // Act
            pdfParser.InitializeCrossReferences();

            // Assert
            this.AssertCrossReferenceEntries(contentManager);
        }

        [TestMethod]
        [ExpectedException(typeof(IndexOutOfRangeException))]
        public void InitializeCrossReferences_WhenReferenceEntriesAreSeparatedWithEmptySpaces_ShouldThrow()
        {
            // Arrange
            Stream stream = new MemoryStream(Encoding.UTF8.GetBytes(XRefWithEmtySpaces));
            PdfContentManager contentManager = new PdfContentManager(new PdfFormatProvider(), stream);
            PdfParser pdfParser = new PdfParser(contentManager, stream);

            // Act && Assert
            pdfParser.InitializeCrossReferences();
        }

        [TestMethod]
        public void InitializeCrossReferences_WhenReferenceEntriesAreSeparatedWithCarrageReturn_ShouldParseEntriesCorrectly()
        {
            // Arrange
            Stream stream = new MemoryStream(Encoding.UTF8.GetBytes(XRefWithCariageReturn));
            PdfContentManager contentManager = new PdfContentManager(new PdfFormatProvider(), stream);
            PdfParser pdfParser = new PdfParser(contentManager, stream);

            // Act
            pdfParser.InitializeCrossReferences();

            // Assert
            this.AssertCrossReferenceEntries(contentManager);
        }

        [TestMethod]
        public void InitializeCrossReferences_WhenReferenceEntriesAreSeparatedWithCarrageReturnAndLineFeed_ShouldParseEntriesCorrectly()
        {
            // Arrange
            Stream stream = new MemoryStream(Encoding.UTF8.GetBytes(XRefWithCariageReturnAndLineFeed));
            PdfContentManager contentManager = new PdfContentManager(new PdfFormatProvider(), stream);
            PdfParser pdfParser = new PdfParser(contentManager, stream);

            // Act
            pdfParser.InitializeCrossReferences();

            // Assert
            this.AssertCrossReferenceEntries(contentManager);
        }

        [TestMethod]
        public void InitializeCrossReferences_WhenReferenceEntriesAreSeparatedWithManyCarriageReturn_ShouldParseEntriesCorrectly()
        {
            // Arrange
            Stream stream = new MemoryStream(Encoding.UTF8.GetBytes(XRefWitManyCarriageReturn));
            PdfContentManager contentManager = new PdfContentManager(new PdfFormatProvider(), stream);
            PdfParser pdfParser = new PdfParser(contentManager, stream);

            // Act
            pdfParser.InitializeCrossReferences();

            // Assert
            this.AssertCrossReferenceEntries(contentManager);
        }

        private void AssertCrossReferenceEntries(PdfContentManager contentManager)
        {
            CrossReferenceEntryOld firstEntry = contentManager.CrossReferences[0];
            Assert.AreEqual(0, firstEntry.Field2);
            Assert.AreEqual(65535, firstEntry.Field3);
            Assert.AreEqual(CrossReferenceEntryTypeOld.Free, firstEntry.Type);

            CrossReferenceEntryOld secondEntry = contentManager.CrossReferences[1];
            Assert.AreEqual(19690, secondEntry.Field2);
            Assert.AreEqual(0, secondEntry.Field3);
            Assert.AreEqual(CrossReferenceEntryTypeOld.Used, secondEntry.Type);

            CrossReferenceEntryOld thirdEntry = contentManager.CrossReferences[2];
            Assert.AreEqual(15, thirdEntry.Field2);
            Assert.AreEqual(0, thirdEntry.Field3);
            Assert.AreEqual(CrossReferenceEntryTypeOld.Used, thirdEntry.Type);

            CrossReferenceEntryOld fourthEntry;
            Assert.IsFalse(contentManager.CrossReferences.TryGetCrossReferenceEntry(3, out fourthEntry));
            Assert.IsNull(fourthEntry);
        }
    }
}
