﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.Model;

namespace Telerik.Windows.Documents.Fixed.Tests.FormatProviders.Pdf.Import
{
    [TestClass]
    public class LoadOnDemandPagesCacheManagerTests
    {
        private const string Text = "Text";

        [TestMethod]
        public void BeginUsingPage_ShouldLoadPageContent()
        {
            // Arrange
            TestPagesContentCache pagesContentCache = new TestPagesContentCache();
            TestPageContentManager pageContentManager = new TestPageContentManager();
            LoadOnDemandPagesCacheManager cacheManager = new LoadOnDemandPagesCacheManager(pagesContentCache, pageContentManager);
            RadFixedPage page = new RadFixedPage();

            // Act
            cacheManager.BeginUsingPage(page);

            // Assert
            int count;
            bool found = pageContentManager.PageToLoadedContentCount.TryGetValue(page, out count);
            Assert.IsTrue(found);
            Assert.AreEqual(1, count);

            found = pageContentManager.PageToUnloadedContentCount.TryGetValue(page, out count);
            Assert.IsFalse(found);
        }

        [TestMethod]
        public void BeginUsingPage_WhenDisposed_ShouldUnloadPageContent()
        {
            // Arrange
            TestPagesContentCache pagesContentCache = new TestPagesContentCache();
            TestPageContentManager pageContentManager = new TestPageContentManager();
            LoadOnDemandPagesCacheManager cacheManager = new LoadOnDemandPagesCacheManager(pagesContentCache, pageContentManager);
            RadFixedPage page = new RadFixedPage();

            // Act
            using (cacheManager.BeginUsingPage(page))
            {
            }

            // Assert
            int count;
            bool found = pageContentManager.PageToUnloadedContentCount.TryGetValue(page, out count);
            Assert.IsTrue(found);
            Assert.AreEqual(1, count);

            found = pageContentManager.PageToUnloadedContentCount.TryGetValue(page, out count);
            Assert.IsTrue(found);
            Assert.AreEqual(1, count);
        }

        [TestMethod]
        public void BeginUsingPage_WhenSeveralCalls_ShouldLoadAndUnloadPageContentCorrectNumberOfTimes()
        {
            // Arrange
            TestPagesContentCache pagesContentCache = new TestPagesContentCache();
            TestPageContentManager pageContentManager = new TestPageContentManager();
            LoadOnDemandPagesCacheManager cacheManager = new LoadOnDemandPagesCacheManager(pagesContentCache, pageContentManager);
            RadFixedPage page = new RadFixedPage();

            // Act
            using (cacheManager.BeginUsingPage(page))
            {
            }

            using (cacheManager.BeginUsingPage(page))
            {
            }

            using (cacheManager.BeginUsingPage(page))
            {
            }

            // Assert
            int count;
            bool found = pageContentManager.PageToLoadedContentCount.TryGetValue(page, out count);
            Assert.IsTrue(found);
            Assert.AreEqual(3, count);

            found = pageContentManager.PageToUnloadedContentCount.TryGetValue(page, out count);
            Assert.IsTrue(found);
            Assert.AreEqual(3, count);
        }

        [TestMethod]
        public void BeginUsingPage_WhenSeveralNestedCalls_ShouldLoadAndUnloadPageContentOnlyOnce()
        {
            // Arrange
            TestPagesContentCache pagesContentCache = new TestPagesContentCache();
            TestPageContentManager pageContentManager = new TestPageContentManager();
            LoadOnDemandPagesCacheManager cacheManager = new LoadOnDemandPagesCacheManager(pagesContentCache, pageContentManager);
            RadFixedPage page = new RadFixedPage();

            // Act
            using (cacheManager.BeginUsingPage(page))
            {
                using (cacheManager.BeginUsingPage(page))
                {
                    using (cacheManager.BeginUsingPage(page))
                    {
                    }
                }
            }

            // Assert
            int count;
            bool found = pageContentManager.PageToLoadedContentCount.TryGetValue(page, out count);
            Assert.IsTrue(found);
            Assert.AreEqual(1, count);

            found = pageContentManager.PageToUnloadedContentCount.TryGetValue(page, out count);
            Assert.IsTrue(found);
            Assert.AreEqual(1, count);
        }

        [TestMethod]
        public void BeginUsingPage_WhenPageIsNotEmpty_ShouldNotLoadPage()
        {
            // Arrange
            TestPagesContentCache pagesContentCache = new TestPagesContentCache();
            TestPageContentManager pageContentManager = new TestPageContentManager();
            pageContentManager.IsContentEmpty = false;
            LoadOnDemandPagesCacheManager cacheManager = new LoadOnDemandPagesCacheManager(pagesContentCache, pageContentManager);
            RadFixedPage page = new RadFixedPage();

            // Act
            cacheManager.BeginUsingPage(page);

            // Assert
            int count;
            bool found = pageContentManager.PageToLoadedContentCount.TryGetValue(page, out count);
            Assert.IsFalse(found);
        }

        [TestMethod]
        public void BeginUsingPage_WhenPageIsNotEmpty_ShouldUnloadPageContentOnDispose()
        {
            // Arrange
            TestPagesContentCache pagesContentCache = new TestPagesContentCache();
            TestPageContentManager pageContentManager = new TestPageContentManager();
            pageContentManager.IsContentEmpty = false;
            LoadOnDemandPagesCacheManager cacheManager = new LoadOnDemandPagesCacheManager(pagesContentCache, pageContentManager);
            RadFixedPage page = new RadFixedPage();

            // Act
            using (cacheManager.BeginUsingPage(page))
            {
            }

            // Assert
            int count;
            bool found = pageContentManager.PageToUnloadedContentCount.TryGetValue(page, out count);
            Assert.IsTrue(found);
            Assert.AreEqual(1, count);
        }

        [TestMethod]
        public void BeginUsingPage_ShouldTryLoadCachedPageContent()
        {
            // Arrange
            TestPagesContentCache pagesContentCache = new TestPagesContentCache();
            TestPageContentManager pageContentManager = new TestPageContentManager();
            LoadOnDemandPagesCacheManager cacheManager = new LoadOnDemandPagesCacheManager(pagesContentCache, pageContentManager);
            RadFixedPage page = new RadFixedPage();

            // Act
            cacheManager.BeginUsingPage(page);

            // Assert
            int count;
            bool found = pagesContentCache.PageToLoadedCachedPageCount.TryGetValue(page, out count);
            Assert.IsTrue(found);
            Assert.AreEqual(1, count);
        }

        [TestMethod]
        public void BeginUsingPage_WhenPageContentIsCached_ShouldNotLoadPageContent()
        {
            // Arrange
            TestPagesContentCache pagesContentCache = new TestPagesContentCache();
            pagesContentCache.IsCached = true;
            TestPageContentManager pageContentManager = new TestPageContentManager();
            LoadOnDemandPagesCacheManager cacheManager = new LoadOnDemandPagesCacheManager(pagesContentCache, pageContentManager);
            RadFixedPage page = new RadFixedPage();
            page.Content.AddTextFragment(Text);

            // Act
            cacheManager.BeginUsingPage(page);

            // Assert
            int count;
            bool found = pageContentManager.PageToLoadedContentCount.TryGetValue(page, out count);
            Assert.IsFalse(found);
        }

        [TestMethod]
        public void SavePage_ShouldSavePageContent()
        {
            // Arrange
            TestPagesContentCache pagesContentCache = new TestPagesContentCache();
            TestPageContentManager pageContentManager = new TestPageContentManager();
            LoadOnDemandPagesCacheManager cacheManager = new LoadOnDemandPagesCacheManager(pagesContentCache, pageContentManager);
            RadFixedPage page = new RadFixedPage();

            // Act
            cacheManager.SavePage(page);

            // Assert
            int count;
            bool found = pagesContentCache.PageToCachedPageCount.TryGetValue(page, out count);
            Assert.IsTrue(found);
            Assert.AreEqual(1, count);
        }

        [TestMethod]
        public void Clear_ShouldClearCachedPageContent()
        {
            // Arrange
            TestPagesContentCache pagesContentCache = new TestPagesContentCache();
            TestPageContentManager pageContentManager = new TestPageContentManager();
            LoadOnDemandPagesCacheManager cacheManager = new LoadOnDemandPagesCacheManager(pagesContentCache, pageContentManager);
            RadFixedPage page = new RadFixedPage();

            // Act
            cacheManager.Clear();

            // Assert
            Assert.AreEqual(1, pagesContentCache.ClearedCount);
        }

        [TestMethod]
        public void BeginUsingPage_WhenException_ShouldUnloadPageContent()
        {
            // Arrange
            TestPagesContentCache pagesContentCache = new TestPagesContentCache();
            TestPageContentManager pageContentManager = new TestPageContentManager(throwExceptionOnLoad: true);
            LoadOnDemandPagesCacheManager cacheManager = new LoadOnDemandPagesCacheManager(pagesContentCache, pageContentManager);
            RadFixedPage page = new RadFixedPage();

            // Act
            try
            {
                cacheManager.BeginUsingPage(page);
            }
            catch
            {
            }

            // Assert
            Assert.AreEqual(1, pageContentManager.PageToLoadedContentCount[page]);
            Assert.AreEqual(1, pageContentManager.PageToUnloadedContentCount[page]);
        }

        [TestMethod]
        [ExpectedException(typeof(NullReferenceException))]
        public void BeginUsingPage_WhenExceptionAndCalledAgain_ShouldThrow()
        {
            // Arrange
            TestPagesContentCache pagesContentCache = new TestPagesContentCache();
            TestPageContentManager pageContentManager = new TestPageContentManager(throwExceptionOnLoad: true);
            LoadOnDemandPagesCacheManager cacheManager = new LoadOnDemandPagesCacheManager(pagesContentCache, pageContentManager);
            RadFixedPage page = new RadFixedPage();

            // Act
            try
            {
                cacheManager.BeginUsingPage(page);
            }
            catch
            {
            }

            // Assert
            cacheManager.BeginUsingPage(page);
        }
    }
}
