﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser.Keywords;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Annotations;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Actions;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Fixed.Tests.Import;

namespace Telerik.Windows.Documents.Fixed.Tests.FormatProviders.Pdf.Import
{
    [TestClass]
    public class PageContentManagerTests
    {
        private const string PageContent = @"
1 0 0 rg
0 0 100 200 re
f";

        [TestMethod]
        public void LoadPageContent_ShouldLoadFixedPageFromPage()
        {
            // Arrange
            RadFixedPage fixedPage = new RadFixedPage();

            TestPdfImportContext importContext = new TestPdfImportContext();
            importContext.Reader = new PostScriptReader(new System.IO.MemoryStream(), new KeywordCollection());

            Page page = new Page();
            page.Contents = new ContentStream(System.Text.Encoding.ASCII.GetBytes(PageContent));

            LinkObject linkObject = new LinkObject();
            linkObject.Rect = new PdfArray(new PdfInt(0), new PdfInt(0), new PdfInt(100), new PdfInt(200));
            page.Annots = new PdfArray(linkObject);

            importContext.MapPages(page, fixedPage);
            PageContentManager pageContentManager = new PageContentManager(importContext);

            // Act
            pageContentManager.LoadPageContent(fixedPage);

            // Assert
            this.AssertLoadedPageContent(fixedPage);
        }

        [TestMethod]
        public void UnloadPageContent_WhenLoadOnDemand_ShouldClearContent()
        {
            // Arrange
            TestPdfImportContext importContext = new TestPdfImportContext();
            importContext.ImportSettings.ReadingMode = Fixed.FormatProviders.ReadingMode.OnDemand;
            PageContentManager pageContentManager = new PageContentManager(importContext);
            RadFixedPage page = new RadFixedPage();
            page.Content.AddTextFragment("Text");
            page.Annotations.AddLink(new GoToAction());

            // Act
            pageContentManager.UnloadPageContent(page);

            // Assert
            Assert.AreEqual(0, page.Content.Count);
            Assert.AreEqual(0, page.Annotations.Count);
        }

        [TestMethod]
        public void UnloadPageContent_WhenAllAtOnce_ShouldClearContent()
        {
            // Arrange
            TestPdfImportContext importContext = new TestPdfImportContext();
            importContext.ImportSettings.ReadingMode = Fixed.FormatProviders.ReadingMode.AllAtOnce;
            PageContentManager pageContentManager = new PageContentManager(importContext);
            RadFixedPage page = new RadFixedPage();
            page.Content.AddTextFragment("Text");
            page.Annotations.AddLink(new GoToAction());

            // Act
            pageContentManager.UnloadPageContent(page);

            // Assert
            Assert.AreEqual(1, page.Content.Count);
            Assert.AreEqual(1, page.Annotations.Count);
        }

        [TestMethod]
        public void IsPageContentEmpty_WhenContentAndAnnotationsAreEmpty_ShouldReturnTrue()
        {
            // Arrange
            TestPdfImportContext importContext = new TestPdfImportContext();
            importContext.ImportSettings.ReadingMode = Fixed.FormatProviders.ReadingMode.AllAtOnce;
            PageContentManager pageContentManager = new PageContentManager(importContext);
            RadFixedPage page = new RadFixedPage();

            // Act
            bool result = pageContentManager.IsPageContentEmpty(page);

            // Assert
            Assert.IsTrue(result);
        }

        [TestMethod]
        public void IsPageContentEmpty_WhenContentAndAnnotationsAreNotEmpty_ShouldReturnFalse()
        {
            // Arrange
            TestPdfImportContext importContext = new TestPdfImportContext();
            importContext.ImportSettings.ReadingMode = Fixed.FormatProviders.ReadingMode.AllAtOnce;
            PageContentManager pageContentManager = new PageContentManager(importContext);
            RadFixedPage page = new RadFixedPage();
            page.Content.AddTextFragment("Text");
            page.Annotations.AddLink(new GoToAction());

            // Act
            bool result = pageContentManager.IsPageContentEmpty(page);

            // Assert
            Assert.IsFalse(result);
        }

        [TestMethod]
        public void IsPageContentEmpty_WhenContentIsNotEmpty_ShouldReturnFalse()
        {
            // Arrange
            TestPdfImportContext importContext = new TestPdfImportContext();
            importContext.ImportSettings.ReadingMode = Fixed.FormatProviders.ReadingMode.AllAtOnce;
            PageContentManager pageContentManager = new PageContentManager(importContext);
            RadFixedPage page = new RadFixedPage();
            page.Content.AddTextFragment("Text");

            // Act
            bool result = pageContentManager.IsPageContentEmpty(page);

            // Assert
            Assert.IsFalse(result);
        }

        [TestMethod]
        public void IsPageContentEmpty_WhenAnnotationsAreNotEmpty_ShouldReturnFalse()
        {
            // Arrange
            TestPdfImportContext importContext = new TestPdfImportContext();
            importContext.ImportSettings.ReadingMode = Fixed.FormatProviders.ReadingMode.AllAtOnce;
            PageContentManager pageContentManager = new PageContentManager(importContext);
            RadFixedPage page = new RadFixedPage();
            page.Annotations.AddLink(new GoToAction());

            // Act
            bool result = pageContentManager.IsPageContentEmpty(page);

            // Assert
            Assert.IsFalse(result);
        }


        private void AssertLoadedPageContent(RadFixedPage fixedPage)
        {
            Assert.AreEqual(1, fixedPage.Content.Count);
            Path path = fixedPage.Content[0] as Path;
            Assert.IsNotNull(path);

            Assert.AreEqual(1, fixedPage.Annotations.Count);
            Link link = fixedPage.Annotations[0] as Link;
            Assert.IsNotNull(link);
        }
    }
}
