﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser.Keywords;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Converters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Annotations;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Forms;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.Tests.Import;

namespace Telerik.Windows.Documents.Fixed.Tests.FormatProviders.Pdf.Model.Converters
{
    [TestClass]
    public class FieldsConverterTests
    {
        [TestMethod]
        public void ConvertFromArray_WhenFieldIsAddedAtRootLevelAndHasParent_ShouldAddFieldOnlyOnce()
        {
            // Arrange
            TestPdfImportContext context = new TestPdfImportContext();

            PdfDictionary childDictionary = new PdfDictionary();
            childDictionary["T"] = new PdfLiteralString("ChildFieldName", StringType.ASCII);
            IndirectReference childReference = new IndirectReference(1, 0);
            context.IndirectObjectsToRead[childReference] = new IndirectObject(childReference, childDictionary);

            PdfDictionary parentDictionary = new PdfDictionary();
            parentDictionary["Kids"] = new PdfArray(childReference);
            childDictionary["Parent"] = parentDictionary;
            parentDictionary["T"] = new PdfLiteralString("ParentFieldName", StringType.ASCII);
            IndirectReference parentReference = new IndirectReference(2, 0);
            context.IndirectObjectsToRead[parentReference] = new IndirectObject(parentReference, parentDictionary);

            FieldsConverter fieldsConverter = new FieldsConverter();
            PostScriptReader reader = new PostScriptReader(new MemoryStream(), new KeywordCollection());
            PdfArray rootLevelFields = new PdfArray(parentReference, childReference);

            // Act
            FormFieldsTree formFieldsTree = fieldsConverter.Convert(typeof(FieldsConverterTests), reader, context, rootLevelFields) as FormFieldsTree;

            // Assert
            Assert.IsNotNull(formFieldsTree);
            List<FormFieldNode> fieldNodes = formFieldsTree.ToList();
            Assert.AreEqual(1, fieldNodes.Count);
            FormFieldNode fieldNode = fieldNodes[0];
            Assert.AreEqual("ParentFieldName.ChildFieldName", fieldNode.FullName);
        }

        [TestMethod]
        public void ConvertFromArray_WhenMergedFieldIsAddedAtRootLevelAndHasParent_ShouldAddFieldOnlyOnce()
        {
            // Arrange
            TestPdfImportContext context = new TestPdfImportContext();

            PdfDictionary childDictionary = new PdfDictionary();
            childDictionary["Subtype"] = new PdfName("Widget");
            childDictionary["T"] = new PdfLiteralString("ChildFieldName", StringType.ASCII);
            IndirectReference childReference = new IndirectReference(1, 0);
            context.IndirectObjectsToRead[childReference] = new IndirectObject(childReference, childDictionary);

            PdfDictionary parentDictionary = new PdfDictionary();
            parentDictionary["Kids"] = new PdfArray(childReference);
            childDictionary["Parent"] = parentDictionary;
            parentDictionary["T"] = new PdfLiteralString("ParentFieldName", StringType.ASCII);
            IndirectReference parentReference = new IndirectReference(2, 0);
            context.IndirectObjectsToRead[parentReference] = new IndirectObject(parentReference, parentDictionary);

            FieldsConverter fieldsConverter = new FieldsConverter();
            PostScriptReader reader = new PostScriptReader(new MemoryStream(), new KeywordCollection());
            PdfArray rootLevelFields = new PdfArray(parentReference, childReference);

            // Act
            FormFieldsTree formFieldsTree = fieldsConverter.Convert(typeof(FieldsConverterTests), reader, context, rootLevelFields) as FormFieldsTree;

            // Assert
            Assert.IsNotNull(formFieldsTree);
            List<FormFieldNode> fieldNodes = formFieldsTree.ToList();
            Assert.AreEqual(1, fieldNodes.Count);
            FormFieldNode fieldNode = fieldNodes[0];
            Assert.AreEqual("ParentFieldName.ChildFieldName", fieldNode.FullName);
        }

        [TestMethod]
        public void ConvertFromArray_WhenFieldWithKidsAndSubTypeWidget_ShouldImportAsField()
        {
            // Arrange
            TestPdfImportContext context = new TestPdfImportContext();

            PdfDictionary fieldDictionary = new PdfDictionary();
            fieldDictionary["Kids"] = new PdfArray(new PdfDictionary());
            fieldDictionary["T"] = new PdfLiteralString("FieldName", StringType.ASCII);
            fieldDictionary["Subtype"] = new PdfName("Widget");
            IndirectReference fieldReference = new IndirectReference(1, 0);
            context.IndirectObjectsToRead[fieldReference] = new IndirectObject(fieldReference, fieldDictionary);

            FieldsConverter fieldsConverter = new FieldsConverter();
            PostScriptReader reader = new PostScriptReader(new MemoryStream(), new KeywordCollection());
            PdfArray fields = new PdfArray(fieldReference);

            // Act
            FormFieldsTree formFieldsTree = fieldsConverter.Convert(typeof(FieldsConverterTests), reader, context, fields) as FormFieldsTree;

            // Assert
            Assert.AreEqual(1, context.RegisteredIndirectObjects.Count);
            IndirectObject indirectObject = context.RegisteredIndirectObjects[fieldReference];
            Assert.IsTrue(indirectObject.Content is FormFieldNode);

            FormFieldNode formFieldNode = indirectObject.Content as FormFieldNode;
            Assert.AreEqual("FieldName", formFieldNode.FullName);
        }

        [TestMethod]
        public void ConvertFromArray_WhenFieldWithEmptyKidsAndSubTypeWidget_ShouldImportAsWidget()
        {
            // Arrange
            TestPdfImportContext context = new TestPdfImportContext();

            PdfDictionary fieldDictionary = new PdfDictionary();
            fieldDictionary["Kids"] = new PdfArray();
            fieldDictionary["T"] = new PdfLiteralString("FieldName", StringType.ASCII);
            fieldDictionary["Subtype"] = new PdfName("Widget");
            IndirectReference fieldReference = new IndirectReference(1, 0);
            context.IndirectObjectsToRead[fieldReference] = new IndirectObject(fieldReference, fieldDictionary);

            FieldsConverter fieldsConverter = new FieldsConverter();
            PostScriptReader reader = new PostScriptReader(new MemoryStream(), new KeywordCollection());
            PdfArray fields = new PdfArray(fieldReference);

            // Act
            FormFieldsTree formFieldsTree = fieldsConverter.Convert(typeof(FieldsConverterTests), reader, context, fields) as FormFieldsTree;

            // Assert
            Assert.AreEqual(1, context.RegisteredIndirectObjects.Count);
            IndirectObject indirectObject = context.RegisteredIndirectObjects[fieldReference];
            Assert.IsTrue(indirectObject.Content is WidgetObject);
        }

        [TestMethod]
        public void ConvertFromArray_WhenMergedWidgetField_ShouldCreateFormFieldAndAddItToFieldsTree()
        {
            // Arrange
            TestPdfImportContext context = new TestPdfImportContext();

            PdfDictionary fieldDictionary = new PdfDictionary();
            string fieldName = "FieldName";
            fieldDictionary["T"] = new PdfLiteralString(fieldName, StringType.ASCII);
            fieldDictionary["Subtype"] = new PdfName("Widget");
            IndirectReference fieldReference = new IndirectReference(1, 0);
            context.IndirectObjectsToRead[fieldReference] = new IndirectObject(fieldReference, fieldDictionary);

            FieldsConverter fieldsConverter = new FieldsConverter();
            PostScriptReader reader = new PostScriptReader(new MemoryStream(), new KeywordCollection());
            PdfArray fields = new PdfArray(fieldReference);

            // Act
            FormFieldsTree formFieldsTree = fieldsConverter.Convert(typeof(FieldsConverterTests), reader, context, fields) as FormFieldsTree;

            // Assert
            Assert.AreEqual(1, formFieldsTree.Count());
            FormFieldNode formFieldNode = formFieldsTree.First();
            Assert.AreEqual(fieldName, formFieldNode.PartialName.ToString());
        }

        [TestMethod]
        public void ConvertFromArray_WhenMergedWidgetField_ShouldAddFormFieldAsWidgetParent()
        {
            // Arrange
            TestPdfImportContext context = new TestPdfImportContext();

            PdfDictionary fieldDictionary = new PdfDictionary();
            string fieldName = "FieldName";
            fieldDictionary["T"] = new PdfLiteralString(fieldName, StringType.ASCII);
            fieldDictionary["Subtype"] = new PdfName("Widget");
            IndirectReference fieldReference = new IndirectReference(1, 0);
            context.IndirectObjectsToRead[fieldReference] = new IndirectObject(fieldReference, fieldDictionary);

            FieldsConverter fieldsConverter = new FieldsConverter();
            PostScriptReader reader = new PostScriptReader(new MemoryStream(), new KeywordCollection());
            PdfArray fields = new PdfArray(fieldReference);

            // Act
            FormFieldsTree formFieldsTree = fieldsConverter.Convert(typeof(FieldsConverterTests), reader, context, fields) as FormFieldsTree;

            // Assert
            Assert.AreEqual(1, formFieldsTree.Count());
            FormFieldNode formFieldNode = formFieldsTree.First();

            Assert.IsTrue(context.ParentToWidgets.ContainsKey(formFieldNode));
            List<WidgetObject> childWidgets = context.ParentToWidgets[formFieldNode];
            Assert.AreEqual(1, childWidgets.Count);
        }

        [TestMethod]
        public void ConvertFromArray_WhenWidget_ShouldAddFormFieldAsWidgetParent()
        {
            // Arrange
            TestPdfImportContext context = new TestPdfImportContext();

            PdfDictionary fieldDictionary = new PdfDictionary();
            string fieldName = "ParentFieldName";
            fieldDictionary["T"] = new PdfLiteralString(fieldName, StringType.ASCII);
            IndirectReference fieldReference = new IndirectReference(1, 0);
            context.IndirectObjectsToRead[fieldReference] = new IndirectObject(fieldReference, fieldDictionary);

            PdfDictionary widgetDictionary = new PdfDictionary();
            widgetDictionary["Subtype"] = new PdfName("Widget");
            widgetDictionary["Parent"] = fieldDictionary;
            IndirectReference widgetReference = new IndirectReference(2, 0);
            context.IndirectObjectsToRead[widgetReference] = new IndirectObject(widgetReference, widgetDictionary);
            fieldDictionary["Kids"] = new PdfArray(widgetReference); 

            FieldsConverter fieldsConverter = new FieldsConverter();
            PostScriptReader reader = new PostScriptReader(new MemoryStream(), new KeywordCollection());
            PdfArray fields = new PdfArray(fieldReference);

            // Act
            FormFieldsTree formFieldsTree = fieldsConverter.Convert(typeof(FieldsConverterTests), reader, context, fields) as FormFieldsTree;

            // Assert
            Assert.AreEqual(1, formFieldsTree.Count());
            FormFieldNode formFieldNode = formFieldsTree.First();
            Assert.AreEqual(fieldName, formFieldNode.PartialName.ToString());

            Assert.AreEqual(1, context.ParentToWidgets.Count);

            List<WidgetObject> childWidgets = context.ParentToWidgets[formFieldNode];
            Assert.IsNotNull(childWidgets);
            Assert.AreEqual(1, childWidgets.Count);
        }
    }
}
