﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.IO;
using Telerik.Windows.Documents.Fixed.FormatProviders;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser.Keywords;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Converters;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.Tests.Import;

namespace Telerik.Windows.Documents.Fixed.Tests.FormatProviders.Pdf.Model.Converters
{
    [TestClass]
    public class PageTypeAwareConverterTests
    {
        [TestMethod]
        public void Convert_WhenLoadOnDemandWithEmptyPageProperties_ShouldCreatePageInstace()
        {
            // Arrange
            TestPdfImportContext context = new TestPdfImportContext();
            context.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            PdfDictionary pageDictionary = this.CreatePagePdfDictionary();

            IndirectReference pageReference = new IndirectReference(1, 0);
            context.IndirectObjectsToRead[pageReference] = new IndirectObject(pageReference, pageDictionary);

            PostScriptReader reader = new PostScriptReader(new MemoryStream(), new KeywordCollection());

            Page page = new Page();
            Func<PdfName, object> factoryFunction = (n) => { return page; };
            PageTypeAwareConverter converter = new PageTypeAwareConverter(factoryFunction);

            // Act
            Page actualPage = converter.Convert(typeof(Page), reader, context, pageReference) as Page;

            // Assert
            Assert.IsNotNull(actualPage);
            Assert.AreEqual(page, actualPage);

            this.AssertLoadOnDemandPagePropertiesAreNull(actualPage);
        }

        [TestMethod]
        public void Convert_WhenLoadOnDemandWithMediaBox_ShouldCreatePageInstaceWithMediaBox()
        {
            // Arrange
            TestPdfImportContext context = new TestPdfImportContext();
            context.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            PdfDictionary pageDictionary = this.CreatePagePdfDictionary();

            PdfArray mediaBoxArray = new PdfArray(new PdfInt[] { });
            pageDictionary["MediaBox"] = mediaBoxArray;
            Page parent = new Page();
            PdfArray parentMediaBoxArray = new PdfArray(new PdfInt[] { });
            parent.MediaBox = parentMediaBoxArray;
            pageDictionary["Parent"] = parent;

            IndirectReference pageReference = new IndirectReference(1, 0);
            context.IndirectObjectsToRead[pageReference] = new IndirectObject(pageReference, pageDictionary);

            PostScriptReader reader = new PostScriptReader(new MemoryStream(), new KeywordCollection());

            Page page = new Page();
            Func<PdfName, object> factoryFunction = (n) => { return page; };
            PageTypeAwareConverter converter = new PageTypeAwareConverter(factoryFunction);

            // Act
            Page actualPage = converter.Convert(typeof(Page), reader, context, pageReference) as Page;

            // Assert
            Assert.IsNotNull(actualPage);
            Assert.AreEqual(page, actualPage);
            Assert.IsNotNull(actualPage.MediaBox);
            Assert.AreEqual(mediaBoxArray, actualPage.MediaBox);

            this.AssertLoadOnDemandPagePropertiesAreNull(actualPage);
        }

        [TestMethod]
        public void Convert_WhenLoadOnDemandWithNullMediaBox_ShouldCreatePageInstaceWithInheritedMediaBox()
        {
            // Arrange
            TestPdfImportContext context = new TestPdfImportContext();
            context.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            PdfDictionary pageDictionary = this.CreatePagePdfDictionary();
            
            Page parent = new Page();
            PdfArray parentMediaBoxArray = new PdfArray(new PdfInt[] { });
            parent.MediaBox = parentMediaBoxArray;
            pageDictionary["Parent"] = parent;

            IndirectReference pageReference = new IndirectReference(1, 0);
            context.IndirectObjectsToRead[pageReference] = new IndirectObject(pageReference, pageDictionary);

            PostScriptReader reader = new PostScriptReader(new MemoryStream(), new KeywordCollection());

            Page page = new Page();
            Func<PdfName, object> factoryFunction = (n) => { return page; };
            PageTypeAwareConverter converter = new PageTypeAwareConverter(factoryFunction);

            // Act
            Page actualPage = converter.Convert(typeof(Page), reader, context, pageReference) as Page;

            // Assert
            Assert.IsNotNull(actualPage);
            Assert.AreEqual(page, actualPage);
            Assert.IsNotNull(actualPage.MediaBox);
            Assert.AreEqual(parentMediaBoxArray, actualPage.MediaBox);

            this.AssertLoadOnDemandPagePropertiesAreNull(actualPage);
        }

        [TestMethod]
        public void Convert_WhenLoadOnDemandWithCropBox_ShouldCreatePageInstaceWithCropBox()
        {
            // Arrange
            TestPdfImportContext context = new TestPdfImportContext();
            context.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            PdfDictionary pageDictionary = this.CreatePagePdfDictionary();

            PdfArray cropBoxArray = new PdfArray(new PdfInt[] { });
            pageDictionary["CropBox"] = cropBoxArray;
            Page parent = new Page();
            PdfArray parentCropboxArray = new PdfArray(new PdfInt[] { });
            parent.CropBox = parentCropboxArray;
            pageDictionary["Parent"] = parent;

            IndirectReference pageReference = new IndirectReference(1, 0);
            context.IndirectObjectsToRead[pageReference] = new IndirectObject(pageReference, pageDictionary);

            PostScriptReader reader = new PostScriptReader(new MemoryStream(), new KeywordCollection());

            Page page = new Page();
            Func<PdfName, object> factoryFunction = (n) => { return page; };
            PageTypeAwareConverter converter = new PageTypeAwareConverter(factoryFunction);

            // Act
            Page actualPage = converter.Convert(typeof(Page), reader, context, pageReference) as Page;

            // Assert
            Assert.IsNotNull(actualPage);
            Assert.AreEqual(page, actualPage);
            Assert.IsNotNull(actualPage.CropBox);
            Assert.AreEqual(cropBoxArray, actualPage.CropBox);

            this.AssertLoadOnDemandPagePropertiesAreNull(actualPage);
        }

        [TestMethod]
        public void Convert_WhenLoadOnDemandWithNullCropBox_ShouldCreatePageInstaceWithInheritedCropBox()
        {
            // Arrange
            TestPdfImportContext context = new TestPdfImportContext();
            context.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            PdfDictionary pageDictionary = this.CreatePagePdfDictionary();

            Page parent = new Page();
            PdfArray parentCropBoxArray = new PdfArray(new PdfInt[] { });
            parent.CropBox = parentCropBoxArray;
            pageDictionary["Parent"] = parent;

            IndirectReference pageReference = new IndirectReference(1, 0);
            context.IndirectObjectsToRead[pageReference] = new IndirectObject(pageReference, pageDictionary);

            PostScriptReader reader = new PostScriptReader(new MemoryStream(), new KeywordCollection());

            Page page = new Page();
            Func<PdfName, object> factoryFunction = (n) => { return page; };
            PageTypeAwareConverter converter = new PageTypeAwareConverter(factoryFunction);

            // Act
            Page actualPage = converter.Convert(typeof(Page), reader, context, pageReference) as Page;

            // Assert
            Assert.IsNotNull(actualPage);
            Assert.AreEqual(page, actualPage);
            Assert.IsNotNull(actualPage.CropBox);
            Assert.AreEqual(parentCropBoxArray, actualPage.CropBox);

            this.AssertLoadOnDemandPagePropertiesAreNull(actualPage);
        }

        [TestMethod]
        public void Convert_WhenLoadOnDemandWithRotate_ShouldCreatePageInstaceWithRotate()
        {
            // Arrange
            TestPdfImportContext context = new TestPdfImportContext();
            context.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            PdfDictionary pageDictionary = this.CreatePagePdfDictionary();
            
            PdfInt rotate = new PdfInt(90);
            pageDictionary["Rotate"] = rotate;
            Page parent = new Page();
            PdfInt parentRotate = new PdfInt(180);
            parent.Rotate = parentRotate;
            pageDictionary["Parent"] = parent;

            IndirectReference pageReference = new IndirectReference(1, 0);
            context.IndirectObjectsToRead[pageReference] = new IndirectObject(pageReference, pageDictionary);

            PostScriptReader reader = new PostScriptReader(new MemoryStream(), new KeywordCollection());

            Page page = new Page();
            Func<PdfName, object> factoryFunction = (n) => { return page; };
            PageTypeAwareConverter converter = new PageTypeAwareConverter(factoryFunction);

            // Act
            Page actualPage = converter.Convert(typeof(Page), reader, context, pageReference) as Page;

            // Assert
            Assert.IsNotNull(actualPage);
            Assert.AreEqual(page, actualPage);
            Assert.IsNotNull(actualPage.Rotate);
            Assert.AreEqual(rotate, actualPage.Rotate);

            this.AssertLoadOnDemandPagePropertiesAreNull(actualPage);
        }

        [TestMethod]
        public void Convert_WhenLoadOnDemandWithNullRotate_ShouldCreatePageInstaceWithInheritedRotate()
        {
            // Arrange
            TestPdfImportContext context = new TestPdfImportContext();
            context.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            PdfDictionary pageDictionary = this.CreatePagePdfDictionary();

            Page parent = new Page();
            PdfInt parentRotate = new PdfInt(180);
            parent.Rotate = parentRotate;
            pageDictionary["Parent"] = parent;

            IndirectReference pageReference = new IndirectReference(1, 0);
            context.IndirectObjectsToRead[pageReference] = new IndirectObject(pageReference, pageDictionary);

            PostScriptReader reader = new PostScriptReader(new MemoryStream(), new KeywordCollection());

            Page page = new Page();
            Func<PdfName, object> factoryFunction = (n) => { return page; };
            PageTypeAwareConverter converter = new PageTypeAwareConverter(factoryFunction);

            // Act
            Page actualPage = converter.Convert(typeof(Page), reader, context, pageReference) as Page;

            // Assert
            Assert.IsNotNull(actualPage);
            Assert.AreEqual(page, actualPage);
            Assert.IsNotNull(actualPage.Rotate);
            Assert.AreEqual(parentRotate, actualPage.Rotate);

            this.AssertLoadOnDemandPagePropertiesAreNull(actualPage);
        }

        private PdfDictionary CreatePagePdfDictionary()
        {
            PdfDictionary pageDictionary = new PdfDictionary();
            PdfName name = new PdfName("Page");
            pageDictionary["Type"] = name;

            pageDictionary["Contents"] = new ContentStream();
            pageDictionary["Annots"] = new PdfArray();
            pageDictionary["Resources"] = new PdfResource();

            return pageDictionary;
        }

        private void AssertLoadOnDemandPagePropertiesAreNull(Page page)
        {
            Assert.IsNull(page.Contents);
            Assert.IsNull(page.Annots);
            Assert.IsNull(page.Resources);
        }
    }
}
