﻿using System.IO;
using System.Linq;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser.Keywords;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Outlines;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Collections;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Navigation;
using Telerik.Windows.Documents.Fixed.Tests.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.Tests.Import;

namespace Telerik.Windows.Documents.Fixed.Tests.FormatProviders.Pdf.Model.Elements
{
    [TestClass]
    public class DocumentCatalogTests
    {
        private const string FirstChildTitle = "First Child";
        private const string SecondChildTitle = "Second Child";
        private const string ThirdChildTitle = "Third Child";

        [TestMethod]
        public void Constructor_ShouldSetPageModeToUseNone()
        {
            // Arrange & Act
            DocumentCatalog documentCatalog = new DocumentCatalog();

            // Assert
            Assert.AreEqual("UseNone", documentCatalog.PageMode.Value);
        }

        [TestMethod]
        public void CopyPropertiesTo_WhenOutlinesAreNull_ShouldDoNothing()
        {
            // Arrange
            PostScriptReader postScriptReader = new PostScriptReader(new MemoryStream(), new KeywordCollection());
            TestRadFixedDocumentImportContext context = new TestRadFixedDocumentImportContext();

            DocumentCatalog documentCatalog = new DocumentCatalog();
            documentCatalog.Outlines = null;

            PageTreeNode pageTreeNode = new PageTreeNode();
            pageTreeNode.Children = new PdfArray();
            documentCatalog.Pages = pageTreeNode;

            // Act
            documentCatalog.CopyPropertiesTo(postScriptReader, context);

            // Assert
            Assert.AreEqual(0, context.Document.Bookmarks.Count);
        }

        [TestMethod]
        public void CopyPropertiesTo_WhenOutlinesHasNoChildren_ShouldDoNothing()
        {
            // Arrange
            PostScriptReader postScriptReader = new PostScriptReader(new MemoryStream(), new KeywordCollection());
            TestRadFixedDocumentImportContext context = new TestRadFixedDocumentImportContext();

            DocumentCatalog documentCatalog = new DocumentCatalog();
            documentCatalog.Outlines = new OutlinesObject();

            PageTreeNode pageTreeNode = new PageTreeNode();
            pageTreeNode.Children = new PdfArray();
            documentCatalog.Pages = pageTreeNode;

            // Act
            documentCatalog.CopyPropertiesTo(postScriptReader, context);

            // Assert
            Assert.AreEqual(0, context.Document.Bookmarks.Count);
        }

        [TestMethod]
        public void CopyPropertiesTo_WhenOutlinesHasChildren_ShouldCreateBookmarksWithCorrectChildrenHierarchy()
        {
            // Arrange
            PostScriptReader postScriptReader = new PostScriptReader(new MemoryStream(), new KeywordCollection());
            TestRadFixedDocumentImportContext context = new TestRadFixedDocumentImportContext();

            DocumentCatalog documentCatalog = new DocumentCatalog();
            documentCatalog.Outlines = new OutlinesObject();

            OutlineItemObject firstChildOutlineObject = new OutlineItemObject();
            firstChildOutlineObject.Title = new PdfLiteralString(Encoding.UTF8.GetBytes(FirstChildTitle));
            documentCatalog.Outlines.First = firstChildOutlineObject;

            OutlineItemObject secondChildOutlineObject = new OutlineItemObject();
            firstChildOutlineObject.Next = secondChildOutlineObject;
            secondChildOutlineObject.Title = new PdfLiteralString(Encoding.UTF8.GetBytes(SecondChildTitle));

            OutlineItemObject thirdChildOutlineObject = new OutlineItemObject();
            firstChildOutlineObject.First = thirdChildOutlineObject;
            thirdChildOutlineObject.Title = new PdfLiteralString(Encoding.UTF8.GetBytes(ThirdChildTitle));

            PageTreeNode pageTreeNode = new PageTreeNode();
            pageTreeNode.Children = new PdfArray();
            documentCatalog.Pages = pageTreeNode;

            // Act
            documentCatalog.CopyPropertiesTo(postScriptReader, context);

            // Assert
            BookmarksCollection bookmarks = context.Document.Bookmarks;

            Assert.AreEqual(2, bookmarks.Count);

            BookmarkItem firstChild = bookmarks[0];
            Assert.AreEqual(FirstChildTitle, firstChild.Title);
            Assert.AreEqual(1, firstChild.Children.Count);

            BookmarkItem secondChild = bookmarks[1];
            Assert.AreEqual(SecondChildTitle, secondChild.Title);
            Assert.AreEqual(0, secondChild.Children.Count);

            BookmarkItem thirdChild = firstChild.Children[0];
            Assert.AreEqual(ThirdChildTitle, thirdChild.Title);
            Assert.AreEqual(0, thirdChild.Children.Count);
        }

        [TestMethod]
        public void CopyPropertiesTo_WhenPageModeIsNotSet_ShouldSetDocumentPageModeToUseNone()
        {
            // Arrange
            PostScriptReader postScriptReader = new PostScriptReader(new MemoryStream(), new KeywordCollection());
            TestRadFixedDocumentImportContext context = new TestRadFixedDocumentImportContext();

            DocumentCatalog documentCatalog = new DocumentCatalog();

            PageTreeNode pageTreeNode = new PageTreeNode();
            pageTreeNode.Children = new PdfArray();
            documentCatalog.Pages = pageTreeNode;

            // Act
            documentCatalog.CopyPropertiesTo(postScriptReader, context);

            // Assert
            Assert.AreEqual(PageMode.UseNone, context.Document.PageMode);
        }

        [TestMethod]
        public void CopyPropertiesTo_WhenHasPageMode_ShouldSetCorrectDocumentPageMode()
        {
            // Arrange
            PostScriptReader postScriptReader = new PostScriptReader(new MemoryStream(), new KeywordCollection());
            TestRadFixedDocumentImportContext context1 = new TestRadFixedDocumentImportContext();

            DocumentCatalog documentCatalog1 = new DocumentCatalog();
            documentCatalog1.PageMode = new PdfName("UseNone");

            PageTreeNode pageTreeNode = new PageTreeNode();
            pageTreeNode.Children = new PdfArray();
            documentCatalog1.Pages = pageTreeNode;

            TestRadFixedDocumentImportContext context2 = new TestRadFixedDocumentImportContext();

            DocumentCatalog documentCatalog2 = new DocumentCatalog();
            documentCatalog2.PageMode = new PdfName("UseOutlines");

            documentCatalog2.Pages = pageTreeNode;

            // Act
            documentCatalog1.CopyPropertiesTo(postScriptReader, context1);
            documentCatalog2.CopyPropertiesTo(postScriptReader, context2);

            // Assert
            Assert.AreEqual(PageMode.UseNone, context1.Document.PageMode);
            Assert.AreEqual(PageMode.UseBookmarks, context2.Document.PageMode);
        }

        [TestMethod]
        public void CopyPropertiesFrom_WhenHasNoBookmarks_ShouldDoNothing()
        {
            // Arrange
            TestRadFixedDocumentExportContext context = new TestRadFixedDocumentExportContext();
            context.Document = new RadFixedDocument();
            context.DocumentInfo = new RadFixedDocumentInfo();

            DocumentCatalog documentCatalog = new DocumentCatalog();

            // Act
            documentCatalog.CopyPropertiesFrom(context);

            // Assert
            Assert.IsNull(documentCatalog.Outlines);
        }

        [TestMethod]
        public void CopyPropertiesFrom_WhenHasBookmarks_ShouldSetOutlinesInstance()
        {
            // Arrange
            TestRadFixedDocumentExportContext context = new TestRadFixedDocumentExportContext();

            RadFixedDocument document = new RadFixedDocument();
            BookmarkItem firstChild = new BookmarkItem(FirstChildTitle);
            document.Bookmarks.Add(firstChild);

            BookmarkItem secondChild = new BookmarkItem(SecondChildTitle);
            document.Bookmarks.Add(secondChild);

            BookmarkItem thirdChild = new BookmarkItem(ThirdChildTitle);
            firstChild.Children.Add(thirdChild);

            context.Document = document;
            context.DocumentInfo = new RadFixedDocumentInfo();

            DocumentCatalog documentCatalog = new DocumentCatalog();

            // Act
            documentCatalog.CopyPropertiesFrom(context);

            // Assert
            Assert.IsNotNull(documentCatalog.Outlines);
        }

        [TestMethod]
        public void CopyPropertiesFrom_WhenHasBookmarks_ShouldCreateOutlinesWithCorrectCount()
        {
            // Arrange
            TestRadFixedDocumentExportContext context = new TestRadFixedDocumentExportContext();

            RadFixedDocument document = new RadFixedDocument();
            BookmarkItem firstChild = new BookmarkItem(FirstChildTitle);
            document.Bookmarks.Add(firstChild);

            BookmarkItem secondChild = new BookmarkItem(SecondChildTitle);
            document.Bookmarks.Add(secondChild);

            BookmarkItem thirdChild = new BookmarkItem(ThirdChildTitle);
            firstChild.Children.Add(thirdChild);

            context.Document = document;
            context.DocumentInfo = new RadFixedDocumentInfo();

            DocumentCatalog documentCatalog = new DocumentCatalog();

            // Act
            documentCatalog.CopyPropertiesFrom(context);

            // Assert
            Assert.AreEqual(2, documentCatalog.Outlines.Count.Value);
        }

        [TestMethod]
        public void CopyPropertiesFrom_WhenHasBookmarks_ShouldCreateOutlinesWithCorrectName()
        {
            // Arrange
            TestRadFixedDocumentExportContext context = new TestRadFixedDocumentExportContext();

            RadFixedDocument document = new RadFixedDocument();
            BookmarkItem firstChild = new BookmarkItem(FirstChildTitle);
            document.Bookmarks.Add(firstChild);

            BookmarkItem secondChild = new BookmarkItem(SecondChildTitle);
            document.Bookmarks.Add(secondChild);

            BookmarkItem thirdChild = new BookmarkItem(ThirdChildTitle);
            firstChild.Children.Add(thirdChild);

            context.Document = document;
            context.DocumentInfo = new RadFixedDocumentInfo();

            DocumentCatalog documentCatalog = new DocumentCatalog();

            // Act
            documentCatalog.CopyPropertiesFrom(context);

            // Assert
            Assert.AreEqual("Outlines", documentCatalog.Outlines.OutlineType.Value);
        }

        [TestMethod]
        public void CopyPropertiesFrom_WhenHasBookmarks_ShouldCreateOutlinesWithCorrectFirst()
        {
            // Arrange
            TestRadFixedDocumentExportContext context = new TestRadFixedDocumentExportContext();

            RadFixedDocument document = new RadFixedDocument();
            BookmarkItem firstChild = new BookmarkItem(FirstChildTitle);
            document.Bookmarks.Add(firstChild);

            BookmarkItem secondChild = new BookmarkItem(SecondChildTitle);
            document.Bookmarks.Add(secondChild);

            BookmarkItem thirdChild = new BookmarkItem(ThirdChildTitle);
            firstChild.Children.Add(thirdChild);

            context.Document = document;
            context.DocumentInfo = new RadFixedDocumentInfo();

            DocumentCatalog documentCatalog = new DocumentCatalog();

            // Act
            documentCatalog.CopyPropertiesFrom(context);

            // Assert
            byte[] expected = BytesHelper.MergeBytes(new byte[] { 255, 254 }, Encoding.BigEndianUnicode.GetBytes(FirstChildTitle));
            Assert.IsTrue(expected.SequenceEqual(documentCatalog.Outlines.First.Title.Value));
        }

        [TestMethod]
        public void CopyPropertiesFrom_WhenHasBookmarks_ShouldCreateOutlinesWithCorrectLast()
        {
            // Arrange
            TestRadFixedDocumentExportContext context = new TestRadFixedDocumentExportContext();

            RadFixedDocument document = new RadFixedDocument();
            BookmarkItem firstChild = new BookmarkItem(FirstChildTitle);
            document.Bookmarks.Add(firstChild);

            BookmarkItem secondChild = new BookmarkItem(SecondChildTitle);
            document.Bookmarks.Add(secondChild);

            BookmarkItem thirdChild = new BookmarkItem(ThirdChildTitle);
            firstChild.Children.Add(thirdChild);

            context.Document = document;
            context.DocumentInfo = new RadFixedDocumentInfo();

            DocumentCatalog documentCatalog = new DocumentCatalog();

            // Act
            documentCatalog.CopyPropertiesFrom(context);

            // Assert
            byte[] expected = BytesHelper.MergeBytes(new byte[] { 255, 254 }, Encoding.BigEndianUnicode.GetBytes(SecondChildTitle));
            Assert.IsTrue(expected.SequenceEqual(documentCatalog.Outlines.Last.Title.Value));
        }

        [TestMethod]
        public void CopyPropertiesFrom_WhenHasBookmarks_ShouldCreateOutlinesWithCorrectPrevAndNext()
        {
            // Arrange
            TestRadFixedDocumentExportContext context = new TestRadFixedDocumentExportContext();

            RadFixedDocument document = new RadFixedDocument();
            BookmarkItem firstChild = new BookmarkItem(FirstChildTitle);
            document.Bookmarks.Add(firstChild);

            BookmarkItem secondChild = new BookmarkItem(SecondChildTitle);
            document.Bookmarks.Add(secondChild);

            BookmarkItem thirdChild = new BookmarkItem(ThirdChildTitle);
            firstChild.Children.Add(thirdChild);

            context.Document = document;
            context.DocumentInfo = new RadFixedDocumentInfo();

            DocumentCatalog documentCatalog = new DocumentCatalog();

            // Act
            documentCatalog.CopyPropertiesFrom(context);

            // Assert
            OutlineItemObject firstOutlineItemObject = documentCatalog.Outlines.First;
            OutlineItemObject secondOutlineItemObject = documentCatalog.Outlines.Last;

            Assert.IsNull(firstOutlineItemObject.Prev);
            Assert.AreEqual(secondOutlineItemObject, firstOutlineItemObject.Next);

            Assert.AreEqual(firstOutlineItemObject, secondOutlineItemObject.Prev);
            Assert.IsNull(secondOutlineItemObject.Next);
        }

        [TestMethod]
        public void CopyPropertiesFrom_WhenHasBookmarks_ShouldCreateOutlinesWithCorrectParent()
        {
            // Arrange
            TestRadFixedDocumentExportContext context = new TestRadFixedDocumentExportContext();

            RadFixedDocument document = new RadFixedDocument();
            BookmarkItem firstChild = new BookmarkItem(FirstChildTitle);
            document.Bookmarks.Add(firstChild);

            BookmarkItem secondChild = new BookmarkItem(SecondChildTitle);
            document.Bookmarks.Add(secondChild);

            BookmarkItem thirdChild = new BookmarkItem(ThirdChildTitle);
            firstChild.Children.Add(thirdChild);

            context.Document = document;
            context.DocumentInfo = new RadFixedDocumentInfo();

            DocumentCatalog documentCatalog = new DocumentCatalog();

            // Act
            documentCatalog.CopyPropertiesFrom(context);

            // Assert
            OutlineItemObject firstOutlineItemObject = documentCatalog.Outlines.First;
            OutlineItemObject secondOutlineItemObject = documentCatalog.Outlines.Last;

            Assert.AreEqual(documentCatalog.Outlines, firstOutlineItemObject.Parent);
            Assert.AreEqual(documentCatalog.Outlines, secondOutlineItemObject.Parent);
        }

        [TestMethod]
        public void CopyPropertiesFrom_WhenHasBookmarks_ShouldCreateOutlinesWithCorrectChildren()
        {
            // Arrange
            TestRadFixedDocumentExportContext context = new TestRadFixedDocumentExportContext();

            RadFixedDocument document = new RadFixedDocument();
            BookmarkItem firstChild = new BookmarkItem(FirstChildTitle);
            document.Bookmarks.Add(firstChild);

            BookmarkItem secondChild = new BookmarkItem(SecondChildTitle);
            document.Bookmarks.Add(secondChild);

            BookmarkItem thirdChild = new BookmarkItem(ThirdChildTitle);
            firstChild.Children.Add(thirdChild);

            context.Document = document;
            context.DocumentInfo = new RadFixedDocumentInfo();

            DocumentCatalog documentCatalog = new DocumentCatalog();

            // Act
            documentCatalog.CopyPropertiesFrom(context);

            // Assert
            OutlineItemObject firstOutlineItemObject = documentCatalog.Outlines.First;
            OutlineItemObject secondOutlineItemObject = documentCatalog.Outlines.Last;

            Assert.AreEqual(-1, firstOutlineItemObject.Count.Value);
            Assert.IsNotNull(firstOutlineItemObject.First);

            byte[] expectedTitle = BytesHelper.MergeBytes(new byte[] { 255, 254 }, Encoding.BigEndianUnicode.GetBytes(ThirdChildTitle));
            Assert.IsTrue(expectedTitle.SequenceEqual(firstOutlineItemObject.First.Title.Value));

            Assert.AreEqual(0, secondOutlineItemObject.Count.Value);
            Assert.IsNull(secondOutlineItemObject.First);
        }

        [TestMethod]
        public void CopyPropertiesFrom_ShouldSetCorrectPageMode()
        {
            // Arrange
            TestRadFixedDocumentExportContext context1 = new TestRadFixedDocumentExportContext();
            context1.Document = new RadFixedDocument() { PageMode = PageMode.UseNone };
            context1.DocumentInfo = new RadFixedDocumentInfo();

            TestRadFixedDocumentExportContext context2 = new TestRadFixedDocumentExportContext();
            context2.Document = new RadFixedDocument() { PageMode = PageMode.UseBookmarks };
            context2.DocumentInfo = new RadFixedDocumentInfo();

            TestRadFixedDocumentExportContext context3 = new TestRadFixedDocumentExportContext();
            context3.Document = new RadFixedDocument() { PageMode = (PageMode)5 };
            context3.DocumentInfo = new RadFixedDocumentInfo();

            DocumentCatalog documentCatalog1 = new DocumentCatalog();
            DocumentCatalog documentCatalog2 = new DocumentCatalog();
            DocumentCatalog documentCatalog3 = new DocumentCatalog();

            // Act
            documentCatalog1.CopyPropertiesFrom(context1);
            documentCatalog2.CopyPropertiesFrom(context2);
            documentCatalog3.CopyPropertiesFrom(context3);

            // Assert
            Assert.AreEqual("UseNone", documentCatalog1.PageMode.Value);
            Assert.AreEqual("UseOutlines", documentCatalog2.PageMode.Value);
            Assert.AreEqual("UseNone", documentCatalog3.PageMode.Value);
        }
    }
}
