﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Text;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.DocumentStructure;

namespace Telerik.Windows.Documents.Fixed.Tests.FormatProviders.Pdf.Model.Elements.DocumentStructure
{
    [TestClass]
    public class ContentStreamTests
    {
        private const string FirstStreamElement = @"q
            /F1 21 Tf
            1 0 0 1 51.42 730.86 Tm
            0.384 0.392 0.388 rg
            0.384 0.392 0.388 RG
            Q";

        private const string SecondStreamElement = @"q
            /F1 21 Tf
            1 0 0 1 568.92 769.56 Tm
            0 g
            0 G
            Q";

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Concat_WhenStreamToAppendIsNull_ShouldThrow()
        {
            // Arrange
            ContentStream contentStream = new ContentStream();

            // Act && Assert
            contentStream.Concat(null);
        }

        [TestMethod]
        public void Concat_WhenLocalDataIsNull_ShouldNotThrow()
        {
            // Arrange
            ContentStream contentStream = new ContentStream();
            byte[] dataToAppend = Encoding.UTF8.GetBytes(SecondStreamElement);
            ContentStream contentStreamToAppend = new ContentStream(dataToAppend);

            // Act
            contentStream.Concat(contentStreamToAppend);

            // Assert
            string result = Encoding.UTF8.GetString(contentStream.Data);
            Assert.AreEqual(SecondStreamElement, result);
        }

        [TestMethod]
        public void Concat_WhenStreamDataToAppendIsNull_ShouldNotThrow()
        {
            // Arrange
            byte[] data = Encoding.UTF8.GetBytes(FirstStreamElement);
            ContentStream contentStream = new ContentStream(data);
            ContentStream contentStreamToAppend = new ContentStream();

            // Act
            contentStream.Concat(contentStreamToAppend);

            // Assert
            string result = Encoding.UTF8.GetString(contentStream.Data);
            Assert.AreEqual(FirstStreamElement, result);
        }

        [TestMethod]
        public void Concat_WhenAllDataIsNull_ShouldSetEmptyData()
        {
            // Arrange
            ContentStream contentStream = new ContentStream();
            ContentStream contentStreamToAppend = new ContentStream();

            // Act
            contentStream.Concat(contentStreamToAppend);

            // Assert
            Assert.IsNotNull(contentStream.Data);
            Assert.AreEqual(0, contentStream.Data.Length);
        }

        [TestMethod]
        public void Concat_ShouldConcatWithWhiteSpace()
        {
            // Arrange
            byte[] data = Encoding.UTF8.GetBytes(FirstStreamElement);
            ContentStream contentStream = new ContentStream(data);

            byte[] dataToAppend = Encoding.UTF8.GetBytes(SecondStreamElement);
            ContentStream contentStreamToAppend = new ContentStream(dataToAppend);

            string expected = FirstStreamElement + " " + SecondStreamElement;

            // Act
            contentStream.Concat(contentStreamToAppend);

            // Assert
            string result = Encoding.UTF8.GetString(contentStream.Data);
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void Concat_WhenLocalDataEndsWithWhiteSpace_ShouldNotAppendhWhiteSpace()
        {
            // Arrange
            byte[] data = Encoding.UTF8.GetBytes(FirstStreamElement + " ");
            ContentStream contentStream = new ContentStream(data);

            byte[] dataToAppend = Encoding.UTF8.GetBytes(SecondStreamElement);
            ContentStream contentStreamToAppend = new ContentStream(dataToAppend);

            string expected = FirstStreamElement + " " + SecondStreamElement;

            // Act
            contentStream.Concat(contentStreamToAppend);

            // Assert
            string result = Encoding.UTF8.GetString(contentStream.Data);
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void Concat_WhenDataToAppendStartsWithWhiteSpace_ShouldNotAppendWhiteSpace()
        {
            // Arrange
            byte[] data = Encoding.UTF8.GetBytes(FirstStreamElement);
            ContentStream contentStream = new ContentStream(data);

            byte[] dataToAppend = Encoding.UTF8.GetBytes(" " + SecondStreamElement);
            ContentStream contentStreamToAppend = new ContentStream(dataToAppend);

            string expected = FirstStreamElement + " " + SecondStreamElement;

            // Act
            contentStream.Concat(contentStreamToAppend);

            // Assert
            string result = Encoding.UTF8.GetString(contentStream.Data);
            Assert.AreEqual(expected, result);
        }
    }
}
