﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.IO;
using System.Text;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser.Keywords;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;

namespace Telerik.Windows.Documents.Fixed.Tests.FormatProviders.Pdf.Model.Types
{
    [TestClass]
    public class CrossReferenceEntryTests
    {
        private const string XRefWithLineFeed = "0000000000 65535 f\n\n0000019690 00000 n\n\n0000000015 00000 n";

        private const string XRefWithManyLineFeeds = "0000000000 65535 f\n\n\n0000019690 00000 n\n\n\n\n\n0000000015 00000 n";

        private const string XRefWithEmtySpaces = "0000000000 65535 f\n \n0000019690 00000 n\n     \n0000000015 00000 n";

        private const string XRefWithCariageReturn = "0000000000 65535 f\r\r0000019690 00000 n\r\r0000000015 00000 n";

        private const string XRefWithCariageReturnAndLineFeed = "0000000000 65535 f\r\n\r\n0000019690 00000 n\r\n\r\n0000000015 00000 n";

        private const string XRefWitManyCarriageReturn = "0000000000 65535 f\r\r\r0000019690 00000 n\r\r\r\r\r0000000015 00000 n";

        [TestMethod]
        public void Read_WhenReferenceEntriesAreSeparatedWithOneLineFeed_ShouldParseEntriesCorrectly()
        {
            // Arrange
            Stream stream = new MemoryStream(Encoding.UTF8.GetBytes(XRefWithLineFeed));
            PostScriptReader postScriptReader = new PostScriptReader(stream, new KeywordCollection());

            CrossReferenceEntry firstEntry = new CrossReferenceEntry();
            CrossReferenceEntry secondEntry = new CrossReferenceEntry();
            CrossReferenceEntry thirdEntry = new CrossReferenceEntry();

            // Act
            firstEntry.Read(postScriptReader);
            secondEntry.Read(postScriptReader);
            thirdEntry.Read(postScriptReader);

            // Assert
            this.AssertCrossReferenceEntries(firstEntry, secondEntry, thirdEntry);
        }

        [TestMethod]
        public void Read_WhenReferenceEntriesAreSeparatedWithManyLineFeeds_ShouldParseEntriesCorrectly()
        {
            // Arrange
            Stream stream = new MemoryStream(Encoding.UTF8.GetBytes(XRefWithManyLineFeeds));
            PostScriptReader postScriptReader = new PostScriptReader(stream, new KeywordCollection());

            CrossReferenceEntry firstEntry = new CrossReferenceEntry();
            CrossReferenceEntry secondEntry = new CrossReferenceEntry();
            CrossReferenceEntry thirdEntry = new CrossReferenceEntry();

            // Act
            firstEntry.Read(postScriptReader);
            secondEntry.Read(postScriptReader);
            thirdEntry.Read(postScriptReader);

            // Assert
            this.AssertCrossReferenceEntries(firstEntry, secondEntry, thirdEntry);
        }

        [TestMethod]
        public void Read_WhenReferenceEntriesAreSeparatedWithEmptySpaces_ShouldThrow()
        {
            // Arrange
            Stream stream = new MemoryStream(Encoding.UTF8.GetBytes(XRefWithEmtySpaces));
            PostScriptReader postScriptReader = new PostScriptReader(stream, new KeywordCollection());

            CrossReferenceEntry firstEntry = new CrossReferenceEntry();
            CrossReferenceEntry secondEntry = new CrossReferenceEntry();
            CrossReferenceEntry thirdEntry = new CrossReferenceEntry();

            // Act
            firstEntry.Read(postScriptReader);
            secondEntry.Read(postScriptReader);
            thirdEntry.Read(postScriptReader);

            // Assert
            this.AssertCrossReferenceEntries(firstEntry, secondEntry, thirdEntry);
        }

        [TestMethod]
        public void Read_WhenReferenceEntriesAreSeparatedWithCarrageReturn_ShouldParseEntriesCorrectly()
        {
            // Arrange
            Stream stream = new MemoryStream(Encoding.UTF8.GetBytes(XRefWithCariageReturn));
            PostScriptReader postScriptReader = new PostScriptReader(stream, new KeywordCollection());

            CrossReferenceEntry firstEntry = new CrossReferenceEntry();
            CrossReferenceEntry secondEntry = new CrossReferenceEntry();
            CrossReferenceEntry thirdEntry = new CrossReferenceEntry();

            // Act
            firstEntry.Read(postScriptReader);
            secondEntry.Read(postScriptReader);
            thirdEntry.Read(postScriptReader);

            // Assert
            this.AssertCrossReferenceEntries(firstEntry, secondEntry, thirdEntry);
        }

        [TestMethod]
        public void Read_WhenReferenceEntriesAreSeparatedWithCarrageReturnAndLineFeed_ShouldParseEntriesCorrectly()
        {
            // Arrange
            Stream stream = new MemoryStream(Encoding.UTF8.GetBytes(XRefWithCariageReturnAndLineFeed));
            PostScriptReader postScriptReader = new PostScriptReader(stream, new KeywordCollection());

            CrossReferenceEntry firstEntry = new CrossReferenceEntry();
            CrossReferenceEntry secondEntry = new CrossReferenceEntry();
            CrossReferenceEntry thirdEntry = new CrossReferenceEntry();

            // Act
            firstEntry.Read(postScriptReader);
            secondEntry.Read(postScriptReader);
            thirdEntry.Read(postScriptReader);

            // Assert
            this.AssertCrossReferenceEntries(firstEntry, secondEntry, thirdEntry);
        }

        [TestMethod]
        public void Read_WhenReferenceEntriesAreSeparatedWithManyCarriageReturn_ShouldParseEntriesCorrectly()
        {
            // Arrange
            Stream stream = new MemoryStream(Encoding.UTF8.GetBytes(XRefWitManyCarriageReturn));
            PostScriptReader postScriptReader = new PostScriptReader(stream, new KeywordCollection());

            CrossReferenceEntry firstEntry = new CrossReferenceEntry();
            CrossReferenceEntry secondEntry = new CrossReferenceEntry();
            CrossReferenceEntry thirdEntry = new CrossReferenceEntry();

            // Act
            firstEntry.Read(postScriptReader);
            secondEntry.Read(postScriptReader);
            thirdEntry.Read(postScriptReader);

            // Assert
            this.AssertCrossReferenceEntries(firstEntry, secondEntry, thirdEntry);
        }

        private void AssertCrossReferenceEntries(CrossReferenceEntry firstEntry, CrossReferenceEntry secondEntry, CrossReferenceEntry thirdEntry)
        {
            Assert.AreEqual(0, firstEntry.Field1);
            Assert.AreEqual(65535, firstEntry.Field2);
            Assert.AreEqual(CrossReferenceEntryType.Free, firstEntry.Type);

            Assert.AreEqual(19690, secondEntry.Field1);
            Assert.AreEqual(0, secondEntry.Field2);
            Assert.AreEqual(CrossReferenceEntryType.Used, secondEntry.Type);

            Assert.AreEqual(15, thirdEntry.Field1);
            Assert.AreEqual(0, thirdEntry.Field2);
            Assert.AreEqual(CrossReferenceEntryType.Used, thirdEntry.Type);
        }
    }
}
