﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.Model.Common.Functions;
using Telerik.Windows.Documents.Fixed.Tests.Utils;

namespace Telerik.Windows.Documents.Fixed.Tests.Functions
{
    [TestClass]
    public class SampledFunctionTests
    {
        public const double Epsilon = 1E-8;

        [TestMethod]
        public void SampledFunctionCalculationsOneInThreeOutWithTwoSamplesTest()
        {
            double[] domain = { 0, 1 };
            double[] encode = { 0, 1 };
            double[] decode = { 0, 1, 0, 1, 0, 1 };
            double[] range = decode;
            int[] size = { 2 };
            int[][] samples = { new int[] { 255, 0, 200 }, new int[] { 0, 255, 50 } };
            SampledFunction function = new SampledFunction(samples, size, 8, domain, range, encode, decode);

            int testSegmentsCount = 5;
            for (int testIndex = 0; testIndex <= testSegmentsCount; testIndex++)
            {
                double ratio = testIndex / testSegmentsCount;
                double[] expected = { 1 - ratio, ratio, (1 - ratio) * (200 / 255.0) + ratio * (50 / 255.0) };
                double[] value = function.Execute(new double[] { ratio });
                Assert.AreEqual(expected.Length, value.Length);

                for (int componentIndex = 0; componentIndex < expected.Length; componentIndex++)
                {
                    double expectedValue = expected[componentIndex];
                    double actualValue = value[componentIndex];
                    Assert.AreEqual(expectedValue, actualValue);
                }
            }
        }

        [TestMethod]
        public void SampledFunctionCalculationsOneInThreeOutWithFourSamplesTest()
        {
            double[] domain = { 0, 1 };
            double[] encode = { 0, 3 };
            double[] decode = { 0, 1, 0, 1, 0, 1 };
            double[] range = decode;
            int[] size = { 4 };
            int[][] samples = { new int[] { 200, 0, 0 }, new int[] { 100, 255, 0 }, new int[] { 0, 0, 150 }, new int[] { 50, 50, 50 } };
            SampledFunction function = new SampledFunction(samples, size, 8, domain, range, encode, decode);

            int testSegmentsCount = 9;
            for (int testIndex = 0; testIndex <= testSegmentsCount; testIndex++)
            {
                double ratio = (double)testIndex / testSegmentsCount;
                double scaledRatio = ratio * 3;
                int lowerIndex = (int)scaledRatio;
                int upperIndex = Math.Min(3, lowerIndex + 1);
                double localRatio = scaledRatio - lowerIndex;

                int[] firstSample = samples[lowerIndex];
                int[] secondSample = samples[upperIndex];
                double[] value = function.Execute(new double[] { ratio });
                Assert.AreEqual(3, value.Length);

                for (int componentIndex = 0; componentIndex < value.Length; componentIndex++)
                {
                    double expectedValue = ((1 - localRatio) * firstSample[componentIndex] + localRatio * secondSample[componentIndex]) / 255;
                    double actualValue = value[componentIndex];
                    Assert.IsTrue(TestUtils.EqualValues(expectedValue, actualValue, Epsilon));
                }
            }
        }

        [TestMethod]
        public void SampledFunctionCalculationsTwoInThreeOutTwoByTwoSamplesTests()
        {
            double[] domain = { 0, 1, 0, 1 };
            double[] encode = { 0, 1, 0, 1 };
            double[] decode = { 0, 1, 0, 1, 0, 1 };
            double[] range = decode;
            int[] size = { 2, 2 };
            int[][] samples = { new int[] { 200, 0, 0 }, new int[] { 100, 255, 0 }, new int[] { 0, 0, 150 }, new int[] { 50, 70, 80 } };
            SampledFunction function = new SampledFunction(samples, size, 8, domain, range, encode, decode);

            double[] zeroZeroValues = function.Execute(new double[] { 0, 0 });
            double[] oneZeroValues = function.Execute(new double[] { 1, 0 });
            double[] zeroOneValues = function.Execute(new double[] { 0, 1 });
            double[] oneOneValues = function.Execute(new double[] { 1, 1 });
            double[] complexScenarioValues = function.Execute(new double[] { 0.2, 0.6 });

            KeyValuePair<double[], double[]>[] actualToExpectedValues =
            {
                new KeyValuePair<double[], double[]>(zeroZeroValues, new double[] { 200 / 255d, 0, 0 }),
                new KeyValuePair<double[], double[]>(oneZeroValues, new double[] { 100 / 255d, 1, 0 }),
                new KeyValuePair<double[], double[]>(zeroOneValues, new double[] { 0, 0, 150 / 255d }),
                new KeyValuePair<double[], double[]>(oneOneValues, new double[] { 50 / 255d, 70 / 255d, 80 / 255d }),
                new KeyValuePair<double[], double[]>(complexScenarioValues, new double[]
                {
                    (0.8 * 0.4 * 200 + 0.2 * 0.4 * 100 + 0.2 * 0.6 * 50) / 255,
                    (0.2 * 0.4 * 255 + 0.2 * 0.6 * 70) / 255,
                    (0.8 * 0.6 * 150 + 0.2 * 0.6 * 80) / 255,
                }),
            };

            for (int scenarioIndex = 0; scenarioIndex < actualToExpectedValues.Length; scenarioIndex++)
            {
                double[] actualValues = actualToExpectedValues[scenarioIndex].Key;
                double[] expectedValues = actualToExpectedValues[scenarioIndex].Value;

                Assert.AreEqual(expectedValues.Length, actualValues.Length);

                for (int componentIndex = 0; componentIndex < expectedValues.Length; componentIndex++)
                {
                    double actual = actualValues[componentIndex];
                    double expected = expectedValues[componentIndex];
                    Assert.AreEqual(expected, actual);
                }
            }
        }

        [TestMethod]
        public void SampledFunctionCalculationsTwoInThreeOutThreeByFourSamplesTests()
        {
            double[] domain = { 0, 1, 0, 1 };
            double[] encode = { 0, 2, 0, 3 };
            double[] decode = { 0, 1, 0, 1, 0, 1 };
            double[] range = decode;
            int[] size = { 3, 4 };
            int[][] samples =
            {
                new int[] { 200, 100, 50 }, new int[] { 130, 255, 5 }, new int[] { 0, 25, 150 },
                new int[] { 20, 15, 70 }, new int[] { 110, 225, 55 }, new int[] { 40, 65, 250 },
                new int[] { 100, 100, 150 }, new int[] { 170, 55, 35 }, new int[] { 200, 195, 30 },
                new int[] { 130, 80, 40 }, new int[] { 30, 95, 10 }, new int[] { 0, 125, 170 },
            };
            SampledFunction function = new SampledFunction(samples, size, 8, domain, range, encode, decode);

            int testSegmentsCount = 9;
            for (int firstTestIndex = 0; firstTestIndex <= testSegmentsCount; firstTestIndex++)
            {
                for (int secondTestIndex = 0; secondTestIndex <= testSegmentsCount; secondTestIndex++)
                {
                    double firstRatio = (double)firstTestIndex / testSegmentsCount;
                    double firstScaledRatio = firstRatio * 2;
                    int firstLowerIndex = (int)firstScaledRatio;
                    int firstUpperIndex = Math.Min(2, firstLowerIndex + 1);
                    double firstLocalRatio = firstScaledRatio - firstLowerIndex;

                    double secondRatio = (double)secondTestIndex / testSegmentsCount;
                    double secondScaledRatio = secondRatio * 3;
                    int secondLowerIndex = (int)secondScaledRatio;
                    int secondUpperIndex = Math.Min(3, secondLowerIndex + 1);
                    double secondLocalRatio = secondScaledRatio - secondLowerIndex;

                    int[] zeroZeroSample = samples[firstLowerIndex + 3 * secondLowerIndex];
                    int[] oneZeroSample = samples[firstUpperIndex + 3 * secondLowerIndex];
                    int[] zeroOneSample = samples[firstLowerIndex + 3 * secondUpperIndex];
                    int[] oneOneSample = samples[firstUpperIndex + 3 * secondUpperIndex];
                    double[] value = function.Execute(new double[] { firstRatio, secondRatio });
                    Assert.AreEqual(3, value.Length);

                    for (int componentIndex = 0; componentIndex < value.Length; componentIndex++)
                    {
                        double expectedValue = (
                            (1 - firstLocalRatio) * (1 - secondLocalRatio) * zeroZeroSample[componentIndex] +
                            firstLocalRatio * (1 - secondLocalRatio) * oneZeroSample[componentIndex] +
                            (1 - firstLocalRatio) * secondLocalRatio * zeroOneSample[componentIndex] +
                            firstLocalRatio * secondLocalRatio * oneOneSample[componentIndex]) / 255;
                        double actualValue = value[componentIndex];
                        Assert.IsTrue(TestUtils.EqualValues(expectedValue, actualValue, Epsilon));
                    }
                }
            }
        }
    }
}
