﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Fonts;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.GraphicsState;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Utilities;
using Telerik.Windows.Documents.Fixed.Model.Collections;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
using Matrix = System.Windows.Media.Matrix;
#endif

namespace Telerik.Windows.Documents.Fixed.Tests.Import
{
    public abstract class TestContentStreamOperatorsBase
    {
        private readonly Dictionary<string, FontObject> fonts;
        private readonly Dictionary<string, ExtGStateObject> graphicStates;
        private static readonly double dipConversion;

        static TestContentStreamOperatorsBase()
        {
            dipConversion = Unit.PointToDip(1);
        }

        public TestContentStreamOperatorsBase()
        {
            this.fonts = new Dictionary<string, FontObject>();
            FontBase font = FontsRepository.Helvetica;
            FontObject type1Font = new Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Elements.Fonts.Type1FontObject();
            type1Font.BaseFont = new PdfName(font.Name);
            this.fonts["F1"] = type1Font;

            this.graphicStates = new Dictionary<string, ExtGStateObject>()
            {
                {
                    "State1", new ExtGStateObject()
                    {
                        AlphaConstant = new PdfReal(0.5),
                        StrokeAlphaConstant = new PdfReal(0.5),
                        AlphaSource = new PdfBool(false)
                    }
                },
                {
                    "StateWithAlphaWithoutAlphaSource",
                    new ExtGStateObject()
                    {
                        AlphaConstant = new PdfReal(0.2)
                    }
                }
            };
        }

        protected RadFixedPage GeneratePageContent(string contentStreamText)
        {
            RadFixedPage page = new RadFixedPage();
            TestPdfImportContext importContext = new TestPdfImportContext();
            TestPdfContentImportContext contentContext = new TestPdfContentImportContext(importContext, this.fonts, this.graphicStates, page);

            using (System.IO.MemoryStream stream = new System.IO.MemoryStream(System.Text.Encoding.UTF8.GetBytes(contentStreamText)))
            {
                ContentStreamInterpreter interpreter = new ContentStreamInterpreter(stream, contentContext);
                interpreter.Execute();
            }

            return page;
        }

        protected ExpectedPathGeometry GetExpectedRectangle(double x, double y, double width, double height, FillRule fillRule = FillRule.Nonzero)
        {
            return new ExpectedPathGeometry()
            {
                FillRule = fillRule,
                ExpectedFigures = new List<ExpectedFigure>()
                {
                    new ExpectedFigure()
                    {
                        StartPoint = new Point(x, y),
                        IsClosed = true,
                        ExpectedSegments = new List<ExpectedSegment>()
                        {
                            new ExpectedLineSegment() { Point = new Point(x + width, y) },
                            new ExpectedLineSegment() { Point = new Point(x + width, y + height) },
                            new ExpectedLineSegment() { Point = new Point(x, y + height) }
                        },
                    },
                }
            };
        }

        protected void AssertContentElementPosition(RadFixedPage page, PositionContentElement element, Matrix expectedPdfDocumentPosition)
        {
            this.AssertPosition(page, element.Position.Matrix, expectedPdfDocumentPosition);
        }

        protected void AssertPosition(RadFixedPage page, Matrix positionMatrix, Matrix expectedPdfDocumentPosition)
        {
            AssertPosition(page, positionMatrix, expectedPdfDocumentPosition, this.GetActualMatrix);
        }

        protected void AssertContentElementClipping(RadFixedPage page, ContentElementBase element, ExpectedClipping expectedClipping)
        {
            Clipping clipping = element.Clipping;

            if (expectedClipping == null)
            {
                Assert.IsNull(clipping);
            }
            else
            {
                Assert.IsNotNull(clipping);

                AssertGeometryPosition(page, clipping.Transform, expectedClipping.Transform);
                this.AssertGeometry(clipping.Clip, expectedClipping.Clip);
            }
        }

        protected void AssertGeometry(GeometryBase geometry, ExpectedGeometry expectedGeometry)
        {
            if (expectedGeometry is ExpectedPathGeometry)
            {
                Assert.IsInstanceOfType(geometry, typeof(PathGeometry));

                ExpectedPathGeometry expectedPathGeometry = expectedGeometry as ExpectedPathGeometry;
                PathGeometry pathGeometry = geometry as PathGeometry;

                this.AssertPathFigures(expectedPathGeometry, pathGeometry);
            }
            else
            {
                // Note that the expected geometry is always of type path. The rectangle geometry is
                // only exported, but upon import it is converted to path geometry. See page 227 of 
                // the PDF reference, explaining how the re geometry operator is converted to lines.
                Assert.Fail("Unexpected geometry type.");
            }
        }

        protected Matrix GetActualMatrix(RadFixedPage page, PositionContentElement element)
        {
            return this.GetActualMatrix(page, element.Position.Matrix);
        }

        protected static Matrix GetCoordinatesConversionMatrix(RadFixedPage page)
        {
            return new Matrix(dipConversion, 0, 0, -dipConversion, 0, page.Size.Height).InverseMatrix();
        }

        protected static Matrix GetGeometryActualMatrix(RadFixedPage page, Matrix geometryPositionMatrix)
        {
            Matrix conversionMatrix = GetCoordinatesConversionMatrix(page);
            Matrix actualMatrix = geometryPositionMatrix
                .MultiplyBy(conversionMatrix);

            return actualMatrix;
        }

        protected abstract Matrix GetActualMatrix(RadFixedPage page, Matrix positionMatrix);

        private void AssertPathFigures(ExpectedPathGeometry expectedPathGeometry, PathGeometry pathGeometry)
        {
            Assert.AreEqual(expectedPathGeometry.FillRule, pathGeometry.FillRule);

            List<ExpectedFigure> expectedFigures = expectedPathGeometry.ExpectedFigures;
            PathFigureCollection figures = pathGeometry.Figures;

            Assert.AreEqual(expectedFigures.Count, figures.Count);

            for (int i = 0; i < figures.Count; i++)
            {
                this.AssertFigureSegments(expectedFigures[i], figures[i]);

                Assert.AreEqual(expectedFigures[i].StartPoint, figures[i].StartPoint);
                Assert.AreEqual(expectedFigures[i].IsClosed, figures[i].IsClosed);
            }
        }

        private void AssertFigureSegments(ExpectedFigure expectedFigure, PathFigure pathFigure)
        {
            PathSegmentCollection segments = pathFigure.Segments;

            Assert.AreEqual(expectedFigure.ExpectedSegments.Count, segments.Count);

            for (int i = 0; i < segments.Count; i++)
            {
                ExpectedSegment expectedSegment = expectedFigure.ExpectedSegments[i];
                PathSegment segment = segments[i];

                if (expectedSegment is ExpectedLineSegment)
                {
                    Assert.IsInstanceOfType(segment, typeof(LineSegment));

                    ExpectedLineSegment expectedLineSegment = expectedSegment as ExpectedLineSegment;
                    LineSegment lineSegment = segment as LineSegment;

                    Assert.AreEqual(expectedLineSegment.Point, lineSegment.Point);
                }
                else if (expectedSegment is ExpectedBezierSegment)
                {
                    Assert.IsInstanceOfType(segment, typeof(BezierSegment));

                    ExpectedBezierSegment expectedBezierSegment = expectedSegment as ExpectedBezierSegment;
                    BezierSegment bezierSegment = segment as BezierSegment;

                    Assert.AreEqual(expectedBezierSegment.Point1, bezierSegment.Point1);
                    Assert.AreEqual(expectedBezierSegment.Point2, bezierSegment.Point2);
                    Assert.AreEqual(expectedBezierSegment.Point3, bezierSegment.Point3);
                }
                else
                {
                    Assert.Fail("Unexpected segment type.");
                }
            }
        }

        private static void AssertGeometryPosition(RadFixedPage page, Matrix positionMatrix, Matrix expectedPdfDocumentPosition)
        {
            AssertPosition(page, positionMatrix, expectedPdfDocumentPosition, GetGeometryActualMatrix);
        }

        private static void AssertPosition(RadFixedPage page, Matrix positionMatrix, Matrix expectedPdfDocumentPosition, Func<RadFixedPage, Matrix, Matrix> getActualMatrix)
        {
            Matrix actualMatrix = getActualMatrix(page, positionMatrix);

            string assertFailMessage = "Position does not meet expected value!";

            Assert.IsTrue(expectedPdfDocumentPosition.M11.IsEqualTo(actualMatrix.M11), assertFailMessage);
            Assert.IsTrue(expectedPdfDocumentPosition.M12.IsEqualTo(actualMatrix.M12), assertFailMessage);
            Assert.IsTrue(expectedPdfDocumentPosition.M21.IsEqualTo(actualMatrix.M21), assertFailMessage);
            Assert.IsTrue(expectedPdfDocumentPosition.M22.IsEqualTo(actualMatrix.M22), assertFailMessage);
            Assert.IsTrue(expectedPdfDocumentPosition.OffsetX.IsEqualTo(actualMatrix.OffsetX), assertFailMessage);
            Assert.IsTrue(expectedPdfDocumentPosition.OffsetY.IsEqualTo(actualMatrix.OffsetY), assertFailMessage);
        }
    }
}
