﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Utilities;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser.Keywords;

namespace Telerik.Windows.Documents.Fixed.Tests.Import
{
    [TestClass]
    public class TestPostScriptParser
    {
        [TestMethod]
        public void TestParseInt()
        {
            TestParser("+-1", -1);
            TestParser("++++++++++++1", 1);
            TestParser("+1", 1);
            TestParser("-1", -1);
            TestParser("+.1", .1);
            TestParser("-.1", -.1);
            TestParser("1.1", 1.1);
            TestParser("1", 1);
            TestParser("1 2      3      .22 12.3 22", 1, 2, 3, .22, 12.3, 22);
            TestParser("            1         ", 1);
            TestParser("  @#4  dsfsdf  @  dsf sdf    1   dsf . dsfsdf sdf dsf     ", 4, 1, 0.0);
        }

        [TestMethod]
        public void TestParseHexString()
        {
            TestParser("<>", BytesHelper.ToByteArray(""));
            TestParser("<4E6F762073686D6F7A206B6120706F702E>", BytesHelper.ToByteArray("4E6F762073686D6F7A206B6120706F702E"));
            TestParser("<4E6F76 2073\r\n686D6F7A206B     6120706    F702E>", BytesHelper.ToByteArray("4E6F762073686D6F7A206B6120706F702E"));
            TestParser("  dfsdf     <4E6zF76 2073\r\n686Dz6F7A206B     612z0706    F702zE>         sdfsd      ", BytesHelper.ToByteArray("4E6F762073686D6F7A206B6120706F702E"));
            TestParser("<901FA3>", BytesHelper.ToByteArray("901FA3"));
            TestParser("<901FA>", BytesHelper.ToByteArray("901FA0"));
            TestParser("<901FA><901FA>", BytesHelper.ToByteArray("901FA0"), BytesHelper.ToByteArray("901FA0"));
        }

        [TestMethod]
        public void TestParseLiteralString()
        {
            TestParser(@"()", System.Text.Encoding.UTF8.GetBytes(""));
            TestParser("(this is () valid string)", System.Text.Encoding.UTF8.GetBytes("this is () valid string"));
            TestParser("(alabala)", System.Text.Encoding.UTF8.GetBytes("alabala"));
            TestParser("   asdasd     (alabala)       sadasd      ", System.Text.Encoding.UTF8.GetBytes("alabala"));
            TestParser("(ala)(bala)", Encoding.UTF8.GetBytes("ala"), Encoding.UTF8.GetBytes("bala"));
            TestParser("(ala\r\nbala)", System.Text.Encoding.UTF8.GetBytes("ala\nbala"));
            TestParser("()", System.Text.Encoding.UTF8.GetBytes(""));
            TestParser("(special characters ( * ! & } ^ % and so on ) .)", System.Text.Encoding.UTF8.GetBytes("special characters ( * ! & } ^ % and so on ) ."));
            TestParser(@"(\0533)", System.Text.Encoding.UTF8.GetBytes("+3"));
            TestParser(@"(\053)", System.Text.Encoding.UTF8.GetBytes("+"));
            TestParser(@"(\53)", System.Text.Encoding.UTF8.GetBytes("+"));
            TestParser(@"(\))", System.Text.Encoding.UTF8.GetBytes(")"));
        }

        [TestMethod]
        public void TestParseArray()
        {
            TestParser("[]", CreateArrayAsParameter());
            TestParser("[[(t)]]", CreateArrayAsParameter(CreateArrayAsParameter(Encoding.UTF8.GetBytes("t"))));
            TestParser("[1 3]", CreateArrayAsParameter(1, 3));
            TestParser("[1]", CreateArrayAsParameter(1));
            TestParser("[1 (hello) 2]", CreateArrayAsParameter(1, Encoding.UTF8.GetBytes("hello"), 2));
            TestParser("[(hello)]", CreateArrayAsParameter(Encoding.UTF8.GetBytes("hello")));
            TestParser("[[1][2] dfdsfdsfdsff 44  [3]]", CreateArrayAsParameter(new object[] { 1 }, new object[] { 2 }, 44, new object[] { 3 }));
        }

        [TestMethod]
        public void TestParseArrayWithNegativeSigns()
        {
            TestParser("[-4500(hello)-2300(world)]", CreateArrayAsParameter(-4500, Encoding.UTF8.GetBytes("hello"), -2300, Encoding.UTF8.GetBytes("world")));
        }

        [TestMethod]
        public void TestParseName()
        {
            TestParser("/t 1", "t", 1);
            TestParser("/A#20B", "A B");
            TestParser("/test", "test");
            TestParser("asdasd/name1/name2", "name1", "name2");
        }

        [TestMethod]
        public void TestParseDictionary()
        {
            TestParser("<<>>", CreateDictionaryAsParameter());
            TestParser("[<</d<</t 1>>>>]", CreateArrayAsParameter(CreateDictionaryAsParameter("d", CreateDictionaryAsParameter("t", 1))));
            TestParser("<</d<</t 1>>>>", CreateDictionaryAsParameter("d", CreateDictionaryAsParameter("t", 1)));
            TestParser("<</t 1>>", CreateDictionaryAsParameter("t", 1));
            TestParser("<</t /t2>>", CreateDictionaryAsParameter("t", "t2"));
            TestParser("<</ 256>>", CreateDictionaryAsParameter("", 256));
        }

        [TestMethod]
        public void TestParseKeyword()
        {
            TestParser("true false null", true, false, null);
            TestParser("1 0 R", new IndirectReference(1, 0));
            TestParser("1 0 obj \r\n /Hello \r\n endobj", (values =>
            {
                IndirectObject indirectObj = (IndirectObject)values[0];
                AreEqual(new IndirectReference(1, 0), indirectObj.Reference);
                AreEqual("Hello", indirectObj.Content);
            }));
        }

        [TestMethod]
        public void TestParseComment()
        {
            TestParser("1%/ala 2 3 <<>> \n 2", 1, 2);
            TestParser("true%/ala 2 3 <<>> \r\n 2", true, 2);
        }

        [TestMethod]
        public void TestParsePdfStream()
        {
            TestPdfImportContext context = new TestPdfImportContext();
            context.RegisterIndirectObject(new IndirectReference(1, 0), new PdfInt(100));

            TestParser(context, "<</Length 1 0 R>>\r\nstream\r\n123\r\nendstream", new byte[] { (byte)'1', (byte)'2', (byte)'3' });
            TestParser("<<>>\r\nstream\r\n123\r\nendstream", new byte[] { (byte)'1', (byte)'2', (byte)'3' });
            TestParser("<</Length 1>>\r\nstream\r\n123\r\nendstream", new byte[] { (byte)'1', (byte)'2', (byte)'3' });
            TestParser("<</Length 3>>\r\nstream\r\n123\r\nendstream", new byte[] { (byte)'1', (byte)'2', (byte)'3' });
        }

        [TestMethod]
        public void TestParsePdfStream_NoNewLineBeforeEndStreamKeyword()
        {
            TestParser("<</Length 3>>\nstream\n567endstream", new byte[] { (byte)'5', (byte)'6', (byte)'7' });
        }

        [TestMethod]
        public void TestParsePdfStream_NoCarriageReturn()
        {
            TestParser("<</Length 3>>\nstream\n567\nendstream", new byte[] { (byte)'5', (byte)'6', (byte)'7' });
        }

        [TestMethod]
        public void TestReadDictionary()
        {
            string input = @"<<
/Type /Page
/Annots [6 0 R 7 0 R 8 0 R 9 0 R 10 0 R]
/Contents [12 0 R 13 0 R 14 0 R 15 0 R]
/MediaBox [0 0.02000 595.20 841.92]
/Parent 3 0 R
/Resources 
<<
/Font 11 0 R
/ProcSet [/PDF /Text /ImageC /ImageB]
>>
>>";

            IPdfImportContext context = new TestPdfImportContext();
            MemoryStream stream = new MemoryStream(System.Text.Encoding.UTF8.GetBytes(input));
            PostScriptReader reader = new PostScriptReader(stream, new KeywordCollection());

            PdfDictionary result = reader.Read<PdfDictionary>(context, PdfElementType.Dictionary);

            HashSet<string> expectedKeys =new HashSet<string>(new string[] { "Type", "Annots", "Contents", "MediaBox", "Parent", "Resources" });
            string[] actualKeys = result.Keys.ToArray();

            Assert.AreEqual(expectedKeys.Count, actualKeys.Length);

            for (int i = 0; i < actualKeys.Length; i++)
            {
                string actualKey = actualKeys[i];
                
                Assert.IsTrue(expectedKeys.Contains(actualKey));
            }
        }

        private static object[] CreateArrayAsParameter(params object[] parameters)
        {
            return new object[] { parameters };
        }

        private static Dictionary<string, object> CreateDictionaryAsParameter(params object[] parameters)
        {
            Dictionary<string, object> dict = new Dictionary<string, object>();
            for (int i = 0; i < parameters.Length; i += 2)
            {
                dict[(string)parameters[i]] = parameters[i + 1];
            }

            return dict;
        }

        private static void TestParser(string text, params object[] expectedValues)
        {
            TestParser(new TestPdfImportContext(), text, expectedValues);
        }

        private static void TestParser(TestPdfImportContext context, string text, params object[] expectedValues)
        {
            TestParser(context, text, (values =>
            {
                Assert.AreEqual(expectedValues.Length, values.Length);

                for (int i = 0; i < expectedValues.Length; i++)
                {
                    AreEqual(expectedValues[i], values[i]);
                }
            }));
        }

        private static void TestParser(string text, Action<IPdfPrimitive[]> areEqualAction)
        {
            TestParser(new TestPdfImportContext(), text, areEqualAction);
        }

        private static void TestParser(TestPdfImportContext context, string text, Action<IPdfPrimitive[]> areEqualAction)
        {
            using (var stream = new MemoryStream(System.Text.Encoding.UTF8.GetBytes(text)))
            {
                PostScriptReader parser = new PostScriptReader(stream, new KeywordCollection());

                IPdfPrimitive[] values = parser.Read(context);
                areEqualAction(values);
            }
        }

        private static void AreEqual(object expected, IPdfPrimitive actual)
        {
            if (actual == null)
            {
                Assert.AreEqual(expected, actual);
            }
            else
            {
                switch (actual.Type)
                {
                    case PdfElementType.PdfInt:
                        Assert.AreEqual(expected, ((PdfInt)actual).Value);

                        break;

                    case PdfElementType.PdfReal:
                        Assert.AreEqual(expected, ((PdfReal)actual).Value);

                        break;

                    case PdfElementType.PdfBool:
                        Assert.AreEqual(expected, ((PdfBool)actual).Value);

                        break;

                    case PdfElementType.String:
                        byte[] data = (byte[])expected;
                        PdfString str = (PdfString)actual;
                        Assert.AreEqual(data.Length, str.Value.Length);

                        for (int i = 0; i < data.Length; i++)
                        {
                            Assert.AreEqual(data[i], str.Value[i]);
                        }

                        break;

                    case PdfElementType.PdfName:
                        Assert.AreEqual(expected, ((PdfName)actual).Value);

                        break;

                    case PdfElementType.Array:
                        object[] objects = (object[])expected;
                        PdfArray arr = (PdfArray)actual;

                        Assert.AreEqual(objects.Length, arr.Count);

                        for (int i = 0; i < arr.Count; i++)
                        {
                            AreEqual(objects[i], arr[i]);
                        }

                        break;

                    case PdfElementType.Dictionary:
                        Dictionary<string, object> expectedDict = (Dictionary<string, object>)expected;
                        PdfDictionary actualDict = (PdfDictionary)actual;

                        foreach (string key in expectedDict.Keys)
                        {
                            AreEqual(expectedDict[key], actualDict[key]);
                        }

                        break;

                    case PdfElementType.Stream:
                        byte[] expectedData = (byte[])expected;
                        PdfStream stream = (PdfStream)actual;

                        Assert.AreEqual(expectedData.Length, stream.ReadDecodedPdfData().Length);
                        for (int i = 0; i < expectedData.Length; i++)
                        {
                            Assert.AreEqual(expectedData[i], stream.ReadDecodedPdfData()[i]);
                        }

                        break;

                    default:
                        Assert.AreEqual(expected, actual);

                        break;
                }
            }
        }
    }
}
