﻿using System;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import.Parser;
using System.Collections.Generic;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model;

namespace DocumentProcessingTests
{
    [TestClass]
    public class TestReader
    {
        private readonly byte[] endOfStreamKeywordBytes;

        public TestReader()
        {
            this.endOfStreamKeywordBytes = Encoding.UTF8.GetBytes(PdfNames.PdfStreamEnd);
        }

        [TestMethod]
        public void TestReadFromMultipleThreads()
        {
            Reader reader = TestReader.CreateReader("1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0");

            Task t1 = new Task(() =>
            {
                using (reader.BeginReadingBlock())
                {
                    reader.Seek(0, SeekOrigin.Begin);
                    char ch = (char)reader.Read();
                    Thread.Sleep(1000);

                    Assert.AreEqual('1', ch);
                }
            });

            Task t2 = new Task(() =>
            {
                using (reader.BeginReadingBlock())
                {
                    reader.Seek(66, SeekOrigin.Begin);
                    char ch = (char)reader.Read();
                    Thread.Sleep(1000);

                    Assert.AreEqual('4', ch);

                    using (reader.BeginReadingBlock())
                    {
                        reader.Seek(18, SeekOrigin.Begin);
                        ch = (char)reader.Read();
                        Thread.Sleep(1000);

                        Assert.AreEqual('0', ch);
                    }
                }
            });

            Task t3 = new Task(() =>
            {
                using (reader.BeginReadingBlock())
                {
                    reader.Seek(30, SeekOrigin.Begin);
                    char ch = (char)reader.Read();
                    Thread.Sleep(1000);

                    Assert.AreEqual('6', ch);

                    using (reader.BeginReadingBlock())
                    {
                        reader.Seek(56, SeekOrigin.Begin);
                        ch = (char)reader.Read();
                        Thread.Sleep(1000);

                        Assert.AreEqual('9', ch);
                    }
                }
            });

            t1.Start();
            t2.Start();
            t3.Start();

            Task.WaitAll(t1, t2, t3);
        }

        [TestMethod]
        public void TestReadLine()
        {
            Reader reader = TestReader.CreateReader("endstream\rendobj\n10 0 obj\r\n\r\n\n<<some text>>     \r \n%%EOF");

            string[] expectedLines = { "endstream", "endobj", "10 0 obj", "", "", "<<some text>>     ", " ", "%%EOF" };

            List<string> actualLines = new List<string>();

            while (!reader.EndOfFile)
            {
                actualLines.Add(reader.ReadLine());
            }

            Assert.AreEqual(expectedLines.Length, actualLines.Count);

            for (int i = 0; i < expectedLines.Length; i++)
            {
                string expected = expectedLines[i];
                string actual = actualLines[i];

                Assert.AreEqual(expected, actual);
            }
        }

        [TestMethod]
        public void TestReaderIndexOf_SimpleScenario()
        {
            Reader reader = TestReader.CreateReader("\r\nendstream\rendobj\n10 0 obj\r\n\r\n\n<<some text>>     \r \n%%EOF");
            long actual = reader.IndexOf(this.endOfStreamKeywordBytes);
            
            Assert.AreEqual(2, actual, "IndexOf Position!");
            Assert.AreEqual(11, reader.Position, "Reader Position!");
        }

        [TestMethod]
        public void TestReaderIndexOf_InBegining()
        {
            Reader reader = TestReader.CreateReader("endstream\rendobj\n10 0 obj\r\n\r\n\n<<some text>>     \r \n%%EOF");
            long actual = reader.IndexOf(this.endOfStreamKeywordBytes);
            
            Assert.AreEqual(0, actual, "IndexOf Position!");
            Assert.AreEqual(9, reader.Position, "Reader Position!");
        }

        [TestMethod]
        public void TestReaderIndexOf_InEnding()
        {
            Reader reader = TestReader.CreateReader("<<some text>>     \r \nendstream");
            long actual = reader.IndexOf(this.endOfStreamKeywordBytes);

            Assert.AreEqual(21, actual);
            Assert.IsTrue(reader.EndOfFile);
        }

        [TestMethod]
        public void TestReaderIndexOf_FirstOfMultipleMatches()
        {
            Reader reader = TestReader.CreateReader("<<some text>>     \r \nendstream  \r \nendstreamend");
            long actual = reader.IndexOf(this.endOfStreamKeywordBytes);

            Assert.AreEqual(21, actual, "IndexOf Position!");
            Assert.AreEqual(30, reader.Position, "Reader Position!");
        }

        [TestMethod]
        public void TestReaderIndexOf_FirstOfMultipleMatchesWhenPositionIsCoinsidingWithFirstStart()
        {
            Reader reader = TestReader.CreateReader("<<some text>>     \r \nendstream  \r \nendstreamend");
            reader.Seek(21, SeekOrigin.Begin);
            long actual = reader.IndexOf(this.endOfStreamKeywordBytes);

            Assert.AreEqual(21, actual, "IndexOf Position!");
            Assert.AreEqual(30, reader.Position, "Reader Position!");
        }

        [TestMethod]
        public void TestReaderIndexOf_SecondOfMultipleMatches()
        {
            Reader reader = TestReader.CreateReader("<<some text>>     \r \nendstream  \r \nendstreamendstreamendend");
            reader.Seek(22, SeekOrigin.Begin);
            long actual = reader.IndexOf(this.endOfStreamKeywordBytes);

            Assert.AreEqual(35, actual, "IndexOf Position!");
            Assert.AreEqual(44, reader.Position, "Reader Position!");
        }

        [TestMethod]
        public void TestReaderIndexOf_PartialMatchJustBeforeActualMatch()
        {
            Reader reader = TestReader.CreateReader("<<some text>>     \r \nendendstrendstreamendstrea  ");
            long actual = reader.IndexOf(this.endOfStreamKeywordBytes);

            Assert.AreEqual(30, actual, "IndexOf Position!");
            Assert.AreEqual(39, reader.Position, "Reader Position!");
        }

        [TestMethod]
        public void TestReaderIndexOf_PartialMatchJustBeforeActualMatchAndMatchAfterThat()
        {
            Reader reader = TestReader.CreateReader("<<some text>>     \r \nendendstrendstreamendstream  ");
            long actual = reader.IndexOf(this.endOfStreamKeywordBytes);

            Assert.AreEqual(30, actual, "IndexOf Position!");
            Assert.AreEqual(39, reader.Position, "Reader Position!");
        }

        [TestMethod]
        public void TestReaderIndexOf_PartialMatchWithMultipleCharactersCoinciding()
        {
            Reader reader = TestReader.CreateReader("<<some text>>123123125123125123  ");
            byte[] bytesToSearch = System.Text.Encoding.UTF8.GetBytes("123125");
            long actual = reader.IndexOf(bytesToSearch);

            Assert.AreEqual(16, actual, "IndexOf Position!");
            Assert.AreEqual(22, reader.Position, "Reader Position!");
        }

        [TestMethod]
        public void TestReaderIndexOf_CannotFindPosition()
        {
            Reader reader = TestReader.CreateReader("<<some text>>     \r \nendstrea\rm endstrea\nm endstrea\r\nm endstrea m ");
            reader.Seek(31, SeekOrigin.Begin);
            long actual = reader.IndexOf(this.endOfStreamKeywordBytes);

            Assert.AreEqual(-1, actual, "IndexOf Position!");
            Assert.IsTrue(reader.EndOfFile);
        }

        [TestMethod]
        public void TestReaderIndexOf_CannotFindPositionWhenAfterIt()
        {
            Reader reader = TestReader.CreateReader("<<some text>>     \r \nendendstrendstreamendstrea  ");
            reader.Seek(31, SeekOrigin.Begin);
            long actual = reader.IndexOf(this.endOfStreamKeywordBytes);

            Assert.AreEqual(-1, actual, "IndexOf Position!");
            Assert.IsTrue(reader.EndOfFile);
        }

        private static Reader CreateReader(string textToRead)
        {
            return new Reader(new MemoryStream(Encoding.UTF8.GetBytes(textToRead)));
        }
    }
}
