﻿using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.Tests.Utils;
using Telerik.Windows.Documents.Fixed.Text;

namespace Telerik.Windows.Documents.Fixed.Tests.Import.Text
{
    [TestClass]
    public class TestTextRecognizer
    {
        private const string LinesSeparator = " ";

        private const string ThreeLinesDocumentName = "threeLines";
        private const string ThreeLinesTextContent = "La<nl>lAlz<nl>l2";

        private const string TextWithTextArrayAndCharspacingDocumentName = "Text/TextWithTextArrayAndCharspacing";
        private const string TextWithTextArrayCharspacingAndTextTranslationDocumentName = "Text/TextWithTextArrayCharspacingAndTextTranslation";
        private const string TableOfContents = "Table of Contents";

        [TestMethod]
        public void ImportDocument_WhenMultipleLines_ShouldCreateCorrectLines()
        {
            // Arrange && Act
            TextDocument document = TestUtils.GetTextDocument(ThreeLinesDocumentName);

            // Assert
            this.AssertThreeLinesTextDocument(document);
        }

        [TestMethod]
        public void CreateDocument_WhenMultipleLines_ShouldCreateCorrectLines()
        {
            // Arrange && Act
            TextDocument document = TestUtils.CreateTextDocument(ThreeLinesTextContent);

            // Assert
            this.AssertThreeLinesTextDocument(document);
        }

        [TestMethod]
        public void ImportAndCreateDocument_WhenMultipleLines_ShouldProduceEqualResults()
        {
            TextDocument doc1 = TestUtils.CreateTextDocument(ThreeLinesTextContent);
            TextDocument doc2 = TestUtils.GetTextDocument(ThreeLinesDocumentName);

            for (int i = 0; i < doc1.PagesCount; i++)
            {
                TextPage page1 = doc1.GetTextPage(i);
                TextPage page2 = doc2.GetTextPage(i);

                Assert.AreEqual(page1.ToString(" "), page2.ToString(" "));

                for (int j = 0; j < page1.Lines.Count(); j++)
                {
                    Line line1 = page1.GetLineAt(j);
                    Line line2 = page2.GetLineAt(j);

                    Assert.AreEqual(line1.ToString(), line2.ToString());
                    Assert.AreEqual(line1.ToString(), line2.ToString());
                    Assert.AreEqual(line1.FirstIndex, line2.FirstIndex);
                    Assert.AreEqual(line1.LastIndex, line2.LastIndex);
                    Assert.AreEqual(line1.Length, line2.Length);

                    for (int k = 0; k < line1.WordsCount; k++)
                    {
                        Word word1 = line1.GetWordAt(k);
                        Word word2 = line2.GetWordAt(k);

                        Assert.AreEqual(word1.ToString(), word2.ToString());
                        Assert.AreEqual(word1.EndsWithSpace, word2.EndsWithSpace);
                        Assert.AreEqual(word1.FirstIndex, word2.FirstIndex);
                        Assert.AreEqual(word1.LastIndex, word2.LastIndex);
                        Assert.AreEqual(word1.Length, word2.Length);

                        for (int l = 0; l < word1.CharactersCount; l++)
                        {
                            Character char1 = word1.GetCharacterAt(l);
                            Character char2 = word2.GetCharacterAt(l);

                            Assert.AreEqual(char1.Char, char2.Char);
                        }
                    }
                }
            }
        }

        private void AssertThreeLinesTextDocument(TextDocument document)
        {
            TextPage textPage = document.GetTextPage(0);

            Assert.AreEqual(3, textPage.LinesCount);

            Line firstLine = textPage.GetLineAt(0);

            Assert.AreEqual("La", firstLine.ToString());
            Assert.AreEqual(0, firstLine.FirstIndex);
            Assert.AreEqual(2, firstLine.LastIndex);
            Assert.AreEqual(2, firstLine.Length);

            Line secondLine = textPage.GetLineAt(1);

            Assert.AreEqual("lAlz", secondLine.ToString());
            Assert.AreEqual(3, secondLine.FirstIndex);
            Assert.AreEqual(7, secondLine.LastIndex);
            Assert.AreEqual(4, secondLine.Length);

            Line thirdLine = textPage.GetLineAt(2);

            Assert.AreEqual("l2", thirdLine.ToString());
            Assert.AreEqual(8, thirdLine.FirstIndex);
            Assert.AreEqual(10, thirdLine.LastIndex);
            Assert.AreEqual(2, thirdLine.Length);
        }

        [TestMethod]
        public void ImportDocument_WhenTextWithCharacterSpacing_ShouldCreateCorrectTextPage()
        {
            // Arrange && Act
            TextDocument document = TestUtils.GetTextDocument(TextWithTextArrayAndCharspacingDocumentName);

            // Assert
            Assert.AreEqual(TableOfContents, document.FirstPage.ToString(LinesSeparator));
        }

        [TestMethod]
        public void ImportDocument_WhenTextWithTextTranslation_ShouldCreateCorrectTextPage()
        {
            // Arrange && Act
            TextDocument document = TestUtils.GetTextDocument(TextWithTextArrayCharspacingAndTextTranslationDocumentName);

            // Assert
            Assert.AreEqual(TableOfContents, document.FirstPage.ToString(LinesSeparator));
        }

        //[TestMethod]
        //public void TestSimpleWordRecognition()
        //{
        //    RadFixedDocument doc = TestUtils.CreateFixedDocument("Simple word recognition", 16, 0, 0);
        //    TextDocument text = new TextDocument(doc);
        //    TextPage page = text.GetTextPage(0);

        //    Assert.AreEqual(1, page.LinesCount);
        //    Assert.AreEqual(23, page.LastIndex);
        //    Assert.AreEqual(21, page.Length);

        //    Line line = page.Lines.First();
        //    Assert.AreEqual(3, line.WordsCount);
        //    Assert.AreEqual(0, line.FirstIndex);
        //    Assert.AreEqual(23, line.LastIndex);
        //    Assert.AreEqual(21, line.Length);

        //    Word word = line[0];
        //    Assert.AreEqual(6, word.Length);
        //    Assert.AreEqual("Simple", word.ToString());
        //    Assert.AreEqual(0, word.FirstIndex);
        //    Assert.AreEqual(6, word.LastIndex);

        //    word = line[1];
        //    Assert.AreEqual(4, word.Length);
        //    Assert.AreEqual("word", word.ToString());
        //    Assert.AreEqual(7, word.FirstIndex);
        //    Assert.AreEqual(11, word.LastIndex);

        //    word = line[2];
        //    Assert.AreEqual(11, word.Length);
        //    Assert.AreEqual("recognition", word.ToString());
        //    Assert.AreEqual(12, word.FirstIndex);
        //    Assert.AreEqual(23, word.LastIndex);
        //}

        //[TestMethod]
        //public void TestWordRecognitionWithWordBreaksAndPunctuation()
        //{
        //    RadFixedDocument doc = TestUtils.CreateFixedDocument("(test)()... this.             a-la\\ba/la", 16, 0, 0);
        //    TextDocument text = new TextDocument(doc);
        //    TextPage page = text.GetTextPage(0);

        //    Assert.AreEqual(1, page.LinesCount);
        //    Assert.AreEqual(28, page.LastIndex);
        //    Assert.AreEqual(26, page.Length);

        //    Line line = page.GetLineAt(0);
        //    Assert.AreEqual(12, line.WordsCount);

        //    Word word = line[0];
        //    Assert.AreEqual("(", word.ToString());
        //    Assert.AreEqual(0, word.FirstIndex);
        //    Assert.AreEqual(0, word.LastIndex);
        //    Assert.AreEqual(1, word.Length);

        //    word = line[1];
        //    Assert.AreEqual("test", word.ToString());
        //    Assert.AreEqual(1, word.FirstIndex);
        //    Assert.AreEqual(4, word.LastIndex);
        //    Assert.AreEqual(4, word.Length);

        //    word = line[2];
        //    Assert.AreEqual(")()...", word.ToString());
        //    Assert.AreEqual(5, word.FirstIndex);
        //    Assert.AreEqual(11, word.LastIndex);
        //    Assert.AreEqual(6, word.Length);

        //    word = line[3];
        //    Assert.AreEqual("this", word.ToString());
        //    Assert.AreEqual(12, word.FirstIndex);
        //    Assert.AreEqual(15, word.LastIndex);
        //    Assert.AreEqual(4, word.Length);

        //    word = line[4];
        //    Assert.AreEqual(".", word.ToString());
        //    Assert.AreEqual(16, word.FirstIndex);
        //    Assert.AreEqual(17, word.LastIndex);
        //    Assert.AreEqual(1, word.Length);

        //    word = line[5];
        //    Assert.AreEqual("a", word.ToString());
        //    Assert.AreEqual(18, word.FirstIndex);
        //    Assert.AreEqual(18, word.LastIndex);
        //    Assert.AreEqual(1, word.Length);

        //    word = line[6];
        //    Assert.AreEqual("-", word.ToString());
        //    Assert.AreEqual(19, word.FirstIndex);
        //    Assert.AreEqual(19, word.LastIndex);
        //    Assert.AreEqual(1, word.Length);

        //    word = line[7];
        //    Assert.AreEqual("la", word.ToString());
        //    Assert.AreEqual(20, word.FirstIndex);
        //    Assert.AreEqual(21, word.LastIndex);
        //    Assert.AreEqual(2, word.Length);

        //    word = line[8];
        //    Assert.AreEqual("\\", word.ToString());
        //    Assert.AreEqual(22, word.FirstIndex);
        //    Assert.AreEqual(22, word.LastIndex);
        //    Assert.AreEqual(1, word.Length);

        //    word = line[9];
        //    Assert.AreEqual("ba", word.ToString());
        //    Assert.AreEqual(23, word.FirstIndex);
        //    Assert.AreEqual(24, word.LastIndex);
        //    Assert.AreEqual(2, word.Length);

        //    word = line[10];
        //    Assert.AreEqual("/", word.ToString());
        //    Assert.AreEqual(25, word.FirstIndex);
        //    Assert.AreEqual(25, word.LastIndex);
        //    Assert.AreEqual(1, word.Length);

        //    word = line[11];
        //    Assert.AreEqual("la", word.ToString());
        //    Assert.AreEqual(26, word.FirstIndex);
        //    Assert.AreEqual(28, word.LastIndex);
        //    Assert.AreEqual(2, word.Length);
        //}

        //[TestMethod]
        //public void TestWordRecognitionDifferentFonts()
        //{
        //    RadFixedDocument doc = TestUtils.CreateFixedDocument("", 0, 0, 0);
        //    RadFixedPage page = doc.Pages[0];
        //    page.DrawString("Hello", Fonts.Helvetica, 71.25, 100, 12, 0, 0);
        //    page.DrawString("World", Fonts.HelveticaBold, 98.93000793457, 100, 12, 0, 0);
        //    TextDocument text = new TextDocument(doc);
        //    TextPage textPage = text.GetTextPage(0);

        //    Assert.AreEqual(1, textPage.LinesCount);

        //    Line line = textPage.GetLineAt(0);
        //    Assert.AreEqual(1, line.WordsCount);
        //    Assert.AreEqual("HelloWorld", line[0].ToString());
        //}

        //[TestMethod]
        //public void TestWordRecognitionDifferentFontSize()
        //{
        //    RadFixedDocument doc = TestUtils.CreateFixedDocument("", 0, 0, 0);
        //    RadFixedPage page = doc.Pages[0];
        //    page.Content.Clear();
        //    page.DrawString("as", Fonts.Helvetica, 71.25, 153, 12, 0, 0);
        //    page.DrawString("dasdsadsa", Fonts.Helvetica, 84.709, 100, 40, 0, 0);
        //    page.DrawString("sadasd", Fonts.Helvetica, 278, 153, 12, 0, 0);
        //    TextDocument text = new TextDocument(doc);
        //    TextPage textPage = text.GetTextPage(0);

        //    Assert.AreEqual(1, textPage.LinesCount);

        //    Line line = textPage.GetLineAt(0);
        //    Assert.AreEqual(1, line.WordsCount);
        //    Assert.AreEqual("asdasdsadsasadasd", line[0].ToString());
        //}

        //[TestMethod]
        //public void TestWordRecognitionTwoWords()
        //{
        //    RadFixedDocument doc = TestUtils.CreateFixedDocument("", 0, 0, 0);
        //    RadFixedPage page = doc.Pages[0];
        //    page.DrawString("Hello", Fonts.Helvetica, 71.25, 100, 12, 0, 0);
        //    page.DrawString("World", Fonts.Helvetica, 101.93000793457, 100, 12, 0, 0);
        //    TextDocument text = new TextDocument(doc);
        //    TextPage textPage = text.GetTextPage(0);
        //    Assert.AreEqual(1, textPage.LinesCount);

        //    Line line = textPage.GetLineAt(0);
        //    Assert.AreEqual(2, line.WordsCount);
        //    Assert.AreEqual("Hello", line[0].ToString());
        //    Assert.AreEqual("World", line[1].ToString());
        //}

        //[TestMethod]
        //public void TestLineRecognition()
        //{
        //    RadFixedDocument doc = TestUtils.CreateFixedDocument("Test<nl>New                        Lines<nl>Hello().<nl>World!", 16, 0, 0);
        //    TextDocument text = new TextDocument(doc);
        //    TextPage page = text.GetTextPage(0);

        //    Assert.AreEqual(4, page.LinesCount);

        //    Line line = page.GetLineAt(0);
        //    Assert.AreEqual(0, line.FirstIndex);
        //    Assert.AreEqual(4, line.LastIndex);
        //    Assert.AreEqual(1, line.WordsCount);
        //    Assert.AreEqual(4, line.Length);
        //    Assert.AreEqual("Test", line.ToString());

        //    line = page.GetLineAt(1);
        //    Assert.AreEqual(5, line.FirstIndex);
        //    Assert.AreEqual(14, line.LastIndex);
        //    Assert.AreEqual(2, line.WordsCount);
        //    Assert.AreEqual(8, line.Length);
        //    Assert.AreEqual("New Lines", line.ToString());

        //    line = page.GetLineAt(2);
        //    Assert.AreEqual(15, line.FirstIndex);
        //    Assert.AreEqual(23, line.LastIndex);
        //    Assert.AreEqual(2, line.WordsCount);
        //    Assert.AreEqual(8, line.Length);
        //    Assert.AreEqual("Hello().", line.ToString());

        //    line = page.GetLineAt(3);
        //    Assert.AreEqual(24, line.FirstIndex);
        //    Assert.AreEqual(30, line.LastIndex);
        //    Assert.AreEqual(2, line.WordsCount);
        //    Assert.AreEqual(6, line.Length);
        //    Assert.AreEqual("World!", line.ToString());
        //}

        //[TestMethod]
        //public void TestWordRecognitionWithCharSpacing()
        //{
        //    RadFixedDocument doc = TestUtils.CreateFixedDocument("ala bala hello world", 16, 10, 0);
        //    TextDocument text = new TextDocument(doc);
        //    TextPage page = text.GetTextPage(0);
        //    Assert.AreEqual(1, page.LinesCount);

        //    Line line = page.Lines.First();
        //    Assert.AreEqual(4, line.WordsCount);

        //    Word word = line[0];
        //    Assert.AreEqual("ala", word.ToString());

        //    word = line[1];
        //    Assert.AreEqual("bala", word.ToString());

        //    word = line[2];
        //    Assert.AreEqual("hello", word.ToString());

        //    word = line[3];
        //    Assert.AreEqual("world", word.ToString());
        //}

    }
}
