﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Actions;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Navigation;

namespace Telerik.Windows.Documents.Fixed.Tests.Integration.ImportExport.Bookmarks
{
    [TestClass]
    public class TestCreateExportBookmarks
    {
        [TestMethod]
        public void Bookmarks_WhenCreatedWithChildren_ShouldExportWithCorrectBookmarksHierarchy()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();

            RadFixedDocument document = new RadFixedDocument();

            string firstBookmarkTitle = "Bookmark 1";
            BookmarkItem firstBookmark = new BookmarkItem(firstBookmarkTitle);
            document.Bookmarks.Add(firstBookmark);

            string childBookmarkTitle = "Bookmark Child";
            firstBookmark.Children.Add(new BookmarkItem(childBookmarkTitle));
            firstBookmark.IsExpanded = true;

            string secondBookmarkTitle = "Bookmark 2";
            BookmarkItem secondBookmark = new BookmarkItem(secondBookmarkTitle);
            document.Bookmarks.Add(secondBookmark);

            // Act
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            Assert.AreEqual(2, document.Bookmarks.Count);

            BookmarkItem bookmark1 = document.Bookmarks[0];
            Assert.AreEqual(firstBookmarkTitle, bookmark1.Title);
            Assert.IsTrue(bookmark1.IsExpanded);
            Assert.AreEqual(1, bookmark1.Children.Count);

            BookmarkItem childBookmark = document.Bookmarks[0].Children[0];
            Assert.AreEqual(childBookmarkTitle, childBookmark.Title);
            Assert.IsFalse(childBookmark.IsExpanded);
            Assert.AreEqual(0, childBookmark.Children.Count);

            BookmarkItem bookmark2 = document.Bookmarks[1];
            Assert.AreEqual(secondBookmarkTitle, bookmark2.Title);
            Assert.IsFalse(bookmark2.IsExpanded);
            Assert.AreEqual(0, bookmark2.Children.Count);
        }

        [TestMethod]
        public void Bookmarks_WhenCreatedWithTextColor_ShouldExportWithCorrectTextColor()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();

            RadFixedDocument document = new RadFixedDocument();

            BookmarkItem bookmark = new BookmarkItem("Bookmark Title");
            bookmark.TextColor = new RgbColor(128, 0, 255);
            document.Bookmarks.Add(bookmark);

            // Act
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem result = document.Bookmarks[0];
            Assert.AreEqual(new RgbColor(128, 0, 255), result.TextColor);
        }

        [TestMethod]
        public void Bookmarks_WhenCreatedWithTextStyle_ShouldExportWithCorrectTextStyle()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();

            RadFixedDocument document = new RadFixedDocument();

            BookmarkItem bookmark = new BookmarkItem("Bookmark Title");
            bookmark.TextStyle = BookmarkItemStyles.Italic | BookmarkItemStyles.Bold;
            document.Bookmarks.Add(bookmark);

            // Act
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem result = document.Bookmarks[0];
            Assert.AreEqual(BookmarkItemStyles.Italic | BookmarkItemStyles.Bold, result.TextStyle);
        }

        [TestMethod]
        public void Bookmarks_WhenCreatedWithGoToActionWithLocationDestination_ShouldExportWithCorrectAction()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            RadFixedDocument document = new RadFixedDocument();

            Location destinationToAdd = new Location()
            {
                Top = 96,
                Left = 92,
                Zoom = 1,
                Page = document.Pages.AddPage()
            };

            GoToAction goToAction = new GoToAction(destinationToAdd);
            BookmarkItem bookmarkToAdd = new BookmarkItem("Bookmark Title", goToAction);

            document.Bookmarks.Add(bookmarkToAdd);

            // Act
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem bookmark = document.Bookmarks[0];

            Assert.IsNotNull(bookmark.Action);
            Assert.AreEqual(ActionType.GoTo, bookmark.Action.ActionType);

            GoToAction action = bookmark.Action as GoToAction;
            Assert.IsNotNull(action);

            Assert.IsNotNull(action.Destination);
            Assert.AreEqual(DestinationType.Location, action.Destination.DestinationType);

            Location destination = action.Destination as Location;
            Assert.IsNotNull(destination);

            Assert.AreEqual(document.Pages[0], destination.Page);
            Assert.AreEqual(96, destination.Top);
            Assert.AreEqual(92, destination.Left);
            Assert.AreEqual(1, destination.Zoom);
        }

        [TestMethod]
        public void Bookmarks_WhenCreatedWithNamedGoToActionWithLocationDestination_ShouldExportWithCorrectAction()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            RadFixedDocument document = new RadFixedDocument();

            Location destinationToAdd = new Location()
            {
                Top = 96,
                Left = 92,
                Zoom = 1,
                Page = document.Pages.AddPage()
            };

            GoToAction goToAction = new GoToAction(destinationToAdd);
            NamedDestination namedDestination = document.NamedDestinations.Add("Bookmark", goToAction);
            BookmarkItem bookmarkToAdd = new BookmarkItem("Bookmark Title", namedDestination);

            document.Bookmarks.Add(bookmarkToAdd);

            // Act
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem bookmark = document.Bookmarks[0];

            Assert.IsNull(bookmark.Action);
            Assert.IsNotNull(bookmark.NamedDestination);

            Assert.IsNull(bookmark.NamedDestination.Destination);

            GoToAction action = bookmark.NamedDestination.GoToAction;

            Assert.IsNotNull(action);

            Assert.AreEqual(ActionType.GoTo, action.ActionType);

            Assert.IsNotNull(action.Destination);
            Assert.AreEqual(DestinationType.Location, action.Destination.DestinationType);

            Location destination = action.Destination as Location;
            Assert.IsNotNull(destination);

            Assert.AreEqual(document.Pages[0], destination.Page);
            Assert.AreEqual(96, destination.Top);
            Assert.AreEqual(92, destination.Left);
            Assert.AreEqual(1, destination.Zoom);
        }

        [TestMethod]
        public void Bookmarks_WhenCreatedWithGoToActionWithNamedLocationDestination_ShouldExportWithCorrectAction()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            RadFixedDocument document = new RadFixedDocument();

            Location destinationToAdd = new Location()
            {
                Top = 96,
                Left = 92,
                Zoom = 1,
                Page = document.Pages.AddPage()
            };

            NamedDestination namedDestination = document.NamedDestinations.Add("Destination", destinationToAdd);

            GoToAction goToAction = new GoToAction(namedDestination);
            BookmarkItem bookmarkToAdd = new BookmarkItem("Bookmark Title", goToAction);

            document.Bookmarks.Add(bookmarkToAdd);

            // Act
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem bookmark = document.Bookmarks[0];

            Assert.IsNotNull(bookmark.Action);
            Assert.AreEqual(ActionType.GoTo, bookmark.Action.ActionType);

            GoToAction action = bookmark.Action as GoToAction;
            Assert.IsNotNull(action);

            Assert.IsNull(action.Destination);
            Assert.IsNotNull(action.NamedDestination);
            Assert.IsNotNull(action.NamedDestination.Destination);
            Assert.IsNull(action.NamedDestination.GoToAction);

            Location destination = action.NamedDestination.Destination as Location;
            Assert.IsNotNull(destination);

            Assert.AreEqual(document.Pages[0], destination.Page);
            Assert.AreEqual(96, destination.Top);
            Assert.AreEqual(92, destination.Left);
            Assert.AreEqual(1, destination.Zoom);
        }

        [TestMethod]
        public void Bookmarks_WhenCreatedWithNamedGoToActionWithNamedLocationDestination_ShouldExportWithCorrectAction()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            RadFixedDocument document = new RadFixedDocument();

            Location destinationToAdd = new Location()
            {
                Top = 96,
                Left = 92,
                Zoom = 1,
                Page = document.Pages.AddPage()
            };

            NamedDestination namedDestination = document.NamedDestinations.Add("Destination", destinationToAdd);
            GoToAction goToAction = new GoToAction(namedDestination);

            NamedDestination namedGoToAction = document.NamedDestinations.Add("Bookmark", goToAction);
            BookmarkItem bookmarkToAdd = new BookmarkItem("Bookmark Title", namedGoToAction);

            document.Bookmarks.Add(bookmarkToAdd);

            // Act
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem bookmark = document.Bookmarks[0];

            Assert.IsNull(bookmark.Action);
            Assert.IsNull(bookmark.Destination);
            Assert.IsNotNull(bookmark.NamedDestination);

            Assert.IsNull(bookmark.NamedDestination.Destination);

            GoToAction action = bookmark.NamedDestination.GoToAction;

            Assert.IsNotNull(action);

            Assert.IsNull(action.Destination);
            Assert.IsNotNull(action.NamedDestination);
            Assert.IsNotNull(action.NamedDestination.Destination);
            Assert.IsNull(action.NamedDestination.GoToAction);

            Location destination = action.NamedDestination.Destination as Location;
            Assert.IsNotNull(destination);

            Assert.AreEqual(document.Pages[0], destination.Page);
            Assert.AreEqual(96, destination.Top);
            Assert.AreEqual(92, destination.Left);
            Assert.AreEqual(1, destination.Zoom);
        }

        [TestMethod]
        public void Bookmarks_WhenCreatedWithGoToActionWithPageFitDestination_ShouldExportWithCorrectAction()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();

            RadFixedDocument document = new RadFixedDocument();

            PageFit destinationToAdd = new PageFit()
            {
                Page = document.Pages.AddPage()
            };

            GoToAction goToAction = new GoToAction(destinationToAdd);
            BookmarkItem bookmarkToAdd = new BookmarkItem("Bookmark Title", goToAction);

            document.Bookmarks.Add(bookmarkToAdd);

            // Act
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem bookmark = document.Bookmarks[0];

            Assert.IsNotNull(bookmark.Action);
            Assert.AreEqual(ActionType.GoTo, bookmark.Action.ActionType);

            GoToAction action = bookmark.Action as GoToAction;
            Assert.IsNotNull(action);

            Assert.IsNotNull(action.Destination);
            Assert.AreEqual(DestinationType.PageFit, action.Destination.DestinationType);

            PageFit destination = action.Destination as PageFit;
            Assert.IsNotNull(destination);

            Assert.AreEqual(document.Pages[0], destination.Page);
        }

        [TestMethod]
        public void Bookmarks_WhenCreatedWithGoToActionWithPageHorizontalFitDestination_ShouldExportWithCorrectAction()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();

            RadFixedDocument document = new RadFixedDocument();

            PageHorizontalFit destinationToAdd = new PageHorizontalFit()
            {
                Top = 964,
                Page = document.Pages.AddPage()
            };

            GoToAction goToAction = new GoToAction(destinationToAdd);
            BookmarkItem bookmarkToAdd = new BookmarkItem("Bookmark Title", goToAction);

            document.Bookmarks.Add(bookmarkToAdd);

            // Act
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem bookmark = document.Bookmarks[0];

            Assert.IsNotNull(bookmark.Action);
            Assert.AreEqual(ActionType.GoTo, bookmark.Action.ActionType);

            GoToAction action = bookmark.Action as GoToAction;
            Assert.IsNotNull(action);

            Assert.IsNotNull(action.Destination);
            Assert.AreEqual(DestinationType.PageHorizontalFit, action.Destination.DestinationType);

            PageHorizontalFit destination = action.Destination as PageHorizontalFit;
            Assert.IsNotNull(destination);

            Assert.AreEqual(document.Pages[0], destination.Page);
            Assert.AreEqual(964, destination.Top);
        }

        [TestMethod]
        public void Bookmarks_WhenCreatedWithGoToActionWithPageVerticalFitDestination_ShouldExportWithCorrectAction()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();

            RadFixedDocument document = new RadFixedDocument();


            PageVerticalFit destinationToAdd = new PageVerticalFit()
            {
                Left = 92,
                Page = document.Pages.AddPage()
            };

            GoToAction goToAction = new GoToAction(destinationToAdd);
            BookmarkItem bookmarkToAdd = new BookmarkItem("Bookmark Title", goToAction);

            document.Bookmarks.Add(bookmarkToAdd);

            // Act
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem bookmark = document.Bookmarks[0];

            Assert.IsNotNull(bookmark.Action);
            Assert.AreEqual(ActionType.GoTo, bookmark.Action.ActionType);

            GoToAction action = bookmark.Action as GoToAction;
            Assert.IsNotNull(action);

            Assert.IsNotNull(action.Destination);
            Assert.AreEqual(DestinationType.PageVerticalFit, action.Destination.DestinationType);

            PageVerticalFit destination = action.Destination as PageVerticalFit;
            Assert.IsNotNull(destination);

            Assert.AreEqual(document.Pages[0], destination.Page);
            Assert.AreEqual(92, destination.Left.Value);
        }

        [TestMethod]
        public void Bookmarks_WhenCreatedWithGoToActionWithBoundingRectangleHorizontalFitDestination_ShouldExportWithCorrectAction()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();

            RadFixedDocument document = new RadFixedDocument();

            BoundingRectangleHorizontalFit destinationToAdd = new BoundingRectangleHorizontalFit()
            {
                Top = 964,
                Page = document.Pages.AddPage()
            };

            GoToAction goToAction = new GoToAction(destinationToAdd);

            BookmarkItem bookmarkToAdd = new BookmarkItem("Bookmark Title", goToAction);

            document.Bookmarks.Add(bookmarkToAdd);

            // Act
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem bookmark = document.Bookmarks[0];

            Assert.IsNotNull(bookmark.Action);
            Assert.AreEqual(ActionType.GoTo, bookmark.Action.ActionType);

            GoToAction action = bookmark.Action as GoToAction;
            Assert.IsNotNull(action);

            Assert.IsNotNull(action.Destination);
            Assert.AreEqual(DestinationType.BoundingRectangleHorizontalFit, action.Destination.DestinationType);

            BoundingRectangleHorizontalFit destination = action.Destination as BoundingRectangleHorizontalFit;
            Assert.IsNotNull(destination);

            Assert.AreEqual(document.Pages[0], destination.Page);
            Assert.AreEqual(964, destination.Top);
        }

        [TestMethod]
        public void Bookmarks_WhenCreatedWithGoToActionWithBoundingRectangleVerticalFitDestination_ShouldExportWithCorrectAction()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();

            RadFixedDocument document = new RadFixedDocument();

            BoundingRectangleVerticalFit destinationToAdd = new BoundingRectangleVerticalFit()
            {
                Left = 960,
                Page = document.Pages.AddPage()
            };

            GoToAction goToAction = new GoToAction(destinationToAdd);
            BookmarkItem bookmarkToAdd = new BookmarkItem("Bookmark Title", goToAction);

            document.Bookmarks.Add(bookmarkToAdd);

            // Act
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem bookmark = document.Bookmarks[0];

            Assert.IsNotNull(bookmark.Action);
            Assert.AreEqual(ActionType.GoTo, bookmark.Action.ActionType);

            GoToAction action = bookmark.Action as GoToAction;
            Assert.IsNotNull(action);

            Assert.IsNotNull(action.Destination);
            Assert.AreEqual(DestinationType.BoundingRectangleVerticalFit, action.Destination.DestinationType);

            BoundingRectangleVerticalFit destination = action.Destination as BoundingRectangleVerticalFit;
            Assert.IsNotNull(destination);

            Assert.AreEqual(document.Pages[0], destination.Page);
            Assert.AreEqual(960, destination.Left);
        }
    }
}
