﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Fixed.Model.Text;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Tests.Utils;
using Telerik.Windows.Documents.Fixed.Model.Editing;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Fixed.Model.Objects;
using Rect = System.Windows.Rect;
using Point = System.Windows.Point;
using ScaleTransform = System.Windows.Media.ScaleTransform;
#endif
using Telerik.Windows.Documents.Core.Fonts;

namespace Telerik.Windows.Documents.Fixed.Tests.Integration.ImportExport.Clippings
{
    [TestClass]
#if VISUAL
    public class TestExportImportClippings : DocumentFixedVisualBaseTest
#else
    public class TestExportImportClippings : IntegrationTestClassBase
#endif
    {
        private static Clipping CreateRectangleGeometryClipping()
        {
            return new Clipping()
            {
                Clip = new RectangleGeometry()
                {
                    Rect = new Rect(50, 50, 50, 50)
                }
            };
        }

        private static Clipping CreatePathGeometryClipping(FillRule fillRule = FillRule.Nonzero)
        {
            PathGeometry pathGeometry = new PathGeometry();
            pathGeometry.FillRule = fillRule;

            PathFigure pathFigure = pathGeometry.Figures.AddPathFigure();
            pathFigure.StartPoint = new Point(4, 6);
            pathFigure.IsClosed = true;
            pathFigure.Segments.AddLineSegment().Point = new Point(10, 6);

            BezierSegment bezierSegment = pathFigure.Segments.AddBezierSegment();
            bezierSegment.Point1 = new Point(20, 20);
            bezierSegment.Point2 = new Point(25, 20);
            bezierSegment.Point3 = new Point(30, 30);

            return new Clipping()
            {
                Clip = pathGeometry
            };
        }

#if !NETSTANDARD
        private static Image CreateSampleImage(RadFixedPage page)
        {
            Image image = page.Content.AddImage();
            System.IO.Stream stream = TestUtils.GetApplicationResourceStream(string.Format("Images/{0}", "sample0.jpg"));
            ImageSource imageSource = new ImageSource(stream);
            image.ImageSource = imageSource;

            return image;
        }
#endif

        private static Path CreateSamplePath(RadFixedPage page)
        {
            Path path = page.Content.AddPath();
            PathGeometry pathGeometry = new PathGeometry();
            PathFigure pathFigure = pathGeometry.Figures.AddPathFigure();
            pathFigure.IsClosed = true;
            pathFigure.StartPoint = new Point(4, 4);
            pathFigure.Segments.AddLineSegment().Point = new Point(5, 10);
            pathFigure.Segments.AddLineSegment().Point = new Point(10, 10);
            BezierSegment bezierSegment = pathFigure.Segments.AddBezierSegment();
            bezierSegment.Point1 = new Point(11, 11);
            bezierSegment.Point2 = new Point(5, 11);
            bezierSegment.Point3 = new Point(3, 9);
            path.Geometry = pathGeometry;

            return path;
        }

        private static TextFragment CreateSampleTextFragment(RadFixedPage page)
        {
            StandardFont font;
            if (!FontsRepository.TryCreateStandardFont(FontsManager.Helvetica, out font))
            {
                Assert.Fail("Cannot create font.");
            }

            TextFragment textFragment = page.Content.AddTextFragment();
            textFragment.Position.Translate(30, 30);
            textFragment.Text = "Sample text";
            textFragment.Font = font;

            return textFragment;
        }

        private static RadFixedDocument CreateDocumentWithPage(out RadFixedPage page)
        {
            RadFixedDocument document = new RadFixedDocument();
            page = document.Pages.AddPage();

            return document;
        }

        private static Clipping CreateClippingWithTwoIntersectingTriangles(FillRule fillRule)
        {
            PathGeometry pathGeometry = new PathGeometry();
            pathGeometry.FillRule = fillRule;

            PathFigure firstTriangle = pathGeometry.Figures.AddPathFigure();
            firstTriangle.IsClosed = true;
            firstTriangle.StartPoint = new Point(100, 100);
            firstTriangle.Segments.AddLineSegment().Point = new Point(50, 0);
            firstTriangle.Segments.AddLineSegment().Point = new Point(0, 100);
            firstTriangle.Segments.AddLineSegment().Point = new Point(100, 100);

            PathFigure secondTriangle = pathGeometry.Figures.AddPathFigure();
            secondTriangle.IsClosed = true;
            secondTriangle.StartPoint = new Point(100, 50);
            secondTriangle.Segments.AddLineSegment().Point = new Point(0, 50);
            secondTriangle.Segments.AddLineSegment().Point = new Point(50, 150);
            secondTriangle.Segments.AddLineSegment().Point = new Point(100, 50);

            return new Clipping()
            {
                Clip = pathGeometry
            };
        }

        [TestMethod]
        [TestCategory(PdfProcessingCategory.VisualTest)]
        public void Test_Clipping_ExportImport_Path_With_ClipRectangleGeometry()
        {
            RadFixedPage page;
            RadFixedDocument document = CreateDocumentWithPage(out page);
            Path path = CreateSamplePath(page);

            path.Clipping = CreateRectangleGeometryClipping();

            ExportImportAndAssert(document);
        }

        [TestMethod]
        [TestCategory(PdfProcessingCategory.VisualTest)]
        public void Test_ClippingEvenOdd_ExportImport_Path_With_ClipRectangleGeometry()
        {
            RadFixedPage page;
            RadFixedDocument document = CreateDocumentWithPage(out page);
            Path path = CreateSamplePath(page);

            Clipping clipping = CreateRectangleGeometryClipping();
            path.Clipping = clipping;

            ExportImportAndAssert(document);
        }

        [TestMethod]
        [TestCategory(PdfProcessingCategory.VisualTest)]
        public void Test_Clipping_ExportImport_Path_With_ClipComplexGeometry()
        {
            RadFixedPage page;
            RadFixedDocument document = CreateDocumentWithPage(out page);

            Path path = page.Content.AddPath();
            path.Geometry = new RectangleGeometry(new Rect(0, 0, 100, 150));

            Clipping clipping = CreateClippingWithTwoIntersectingTriangles(FillRule.Nonzero);

            ExportImportAndAssert(document);
        }

        [TestMethod]
        [TestCategory(PdfProcessingCategory.VisualTest)]
        public void Test_ClippingEvenOdd_ExportImport_Path_With_ClipComplexGeometry()
        {
            RadFixedPage page;
            RadFixedDocument document = CreateDocumentWithPage(out page);

            Path path = page.Content.AddPath();
            path.Geometry = new RectangleGeometry(new Rect(0, 0, 100, 150));

            Clipping clipping = CreateClippingWithTwoIntersectingTriangles(FillRule.EvenOdd);
            path.Clipping = clipping;

            ExportImportAndAssert(document);
        }

        [TestMethod]
        [TestCategory(PdfProcessingCategory.VisualTest)]
        public void Test_Clipping_ExportImport_TextFragment_With_ClipPathGeometry()
        {
            RadFixedPage page;
            RadFixedDocument document = CreateDocumentWithPage(out page);
            TextFragment textFragment = CreateSampleTextFragment(page);

            textFragment.Clipping = CreatePathGeometryClipping();

            ExportImportAndAssert(document);
        }

        [TestMethod]
        [TestCategory(PdfProcessingCategory.VisualTest)]
        public void Test_ClippingEvenOdd_ExportImport_TextFragment_With_ClipPathGeometry()
        {
            RadFixedPage page;
            RadFixedDocument document = CreateDocumentWithPage(out page);
            TextFragment textFragment = CreateSampleTextFragment(page);

            Clipping clipping = CreatePathGeometryClipping(FillRule.EvenOdd);
            textFragment.Clipping = clipping;

            ExportImportAndAssert(document);
        }

#if !NETSTANDARD
        [TestMethod]
        [TestCategory(PdfProcessingCategory.VisualTest)]
        public void Test_Clipping_ExportImport_Image_NestedClippings()
        {
            RadFixedPage page;
            RadFixedDocument document = CreateDocumentWithPage(out page);

            Image image = CreateSampleImage(page);

            Clipping clipping = CreatePathGeometryClipping();
            clipping.Clipping = CreateRectangleGeometryClipping();
            clipping.Clipping.Clipping = CreatePathGeometryClipping();
            clipping.Clipping.Clipping.Clipping = CreateRectangleGeometryClipping();
            clipping.Clipping.Clipping.Clipping.Clipping = CreatePathGeometryClipping();

            image.Clipping = clipping;

            ExportImportAndAssert(document);
        }

        [TestMethod]
        [TestCategory(PdfProcessingCategory.VisualTest)]
        public void Test_Clipping_ExportImport_MultipleContentElements_NestedClippings()
        {
            RadFixedPage page;
            RadFixedDocument document = CreateDocumentWithPage(out page);

            TextFragment textFragment = CreateSampleTextFragment(page);
            Path path = CreateSamplePath(page);
            Image image = CreateSampleImage(page);

            Clipping clipping = CreatePathGeometryClipping();
            clipping.Clipping = CreateRectangleGeometryClipping();
            clipping.Clipping.Clipping = CreatePathGeometryClipping();

            textFragment.Clipping = clipping;
            image.Clipping = clipping;
            path.Clipping = clipping;

            ExportImportAndAssert(document);
        }
#endif

        [TestMethod]
        [TestCategory(PdfProcessingCategory.VisualTest)]
        public void Test_Clipping_FixedContentEditor_ElementWithoutClippingBeforeElementWithClipping()
        {
            RadFixedDocument document = new RadFixedDocument();

            FixedContentEditor editor = new FixedContentEditor(document.Pages.AddPage());
            Rect rect = new Rect(0, 0, 100, 20);
            editor.DrawRectangle(rect);
            using (editor.PushClipping(rect))
            {
                editor.GraphicProperties.FillColor = new RgbColor(255, 0, 0);
                editor.DrawText("text");
            }

            ExportImportAndAssert(document);
#if VISUAL
            this.AssertAreVisualyEqual(document);
#endif
        }

        [TestMethod]
        [TestCategory(PdfProcessingCategory.VisualTest)]
        public void Test_Clipping_FixedContentEditor_ElementWithAndWithoutClippingComplexOrder()
        {
            RadFixedDocument document = new RadFixedDocument();

            FixedContentEditor editor = new FixedContentEditor(document.Pages.AddPage());
            editor.GraphicProperties.StrokeThickness = 0;
            editor.GraphicProperties.IsFilled = true;
            editor.GraphicProperties.IsStroked = false;
#if !NETSTANDARD
            ScaleTransform transform = new ScaleTransform()
#else
            Telerik.Documents.Primitives.ScaleTransform transform = new Telerik.Documents.Primitives.ScaleTransform()
#endif
            {
                ScaleX = 400,
                ScaleY = 200
            };

            Func<Rect, Rect> getScaledRect = (rect) =>
                {
                    Point topLeft = transform.Transform(rect.TopLeft);
                    Point bottomRight = transform.Transform(rect.BottomRight);

                    return new Rect(topLeft, bottomRight);
                };

            Rect bounds = getScaledRect(new Rect(0, 0, 1, 1));

            editor.GraphicProperties.FillColor = RgbColors.Black;
            editor.DrawRectangle(getScaledRect(new Rect(0, 0, .6, 1)));

            using (editor.PushClipping(getScaledRect(new Rect(.2, .2, .8, .6))))
            {
                editor.GraphicProperties.FillColor = new RgbColor(255, 0, 0);
                editor.DrawRectangle(bounds);

                using (editor.PushClipping(getScaledRect(new Rect(.4, .4, .4, .2))))
                {
                    editor.GraphicProperties.FillColor = new RgbColor(0, 255, 0);
                    editor.DrawRectangle(bounds);
                }

                using (editor.PushClipping(getScaledRect(new Rect(0, .45, 1, .1))))
                {
                    editor.GraphicProperties.FillColor = new RgbColor(255, 255, 0);
                    editor.DrawRectangle(bounds);
                }
            }

            editor.GraphicProperties.FillColor = new RgbColor(0, 0, 255);
            editor.DrawRectangle(getScaledRect(new Rect(.45, .1, .1, .8)));
            ExportImportAndAssert(document);
#if VISUAL
            this.AssertAreVisualyEqual(document);
#endif
        }
    }
}
