﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Fixed.Tests.Utils;

namespace Telerik.Windows.Documents.Fixed.Tests.Integration.ImportExport.ColorSpaces
{
    [TestClass]
    public class TestGradient : IntegrationTestClassBase
    {
        [TestMethod]
        public void Test_LinearGradient_ExportImport_Fill_ExtendedBeforeAndAfter()
        {
            Path path;
            var document = this.CreatePageWithRect(out path);
            path.Fill = this.CreateSampleLinearGradient();

            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_RadialGradient_ExportImport_Fill_ExtendedBeforeAndAfter()
        {
            Path path;
            var document = this.CreatePageWithRect(out path);
            path.Fill = this.CreateSampleRadialGradient();

            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_LinearGradient_ExportImport_Stroke_ExtendBefore()
        {
            Path path;
            var document = this.CreatePageWithRect(out path);
            var gradient = this.CreateSampleLinearGradient();
            gradient.ExtendAfter = false;

            path.StrokeThickness = 20;
            path.Fill = new RgbColor() { A = 255, G = 255, B = 255 };
            path.Stroke = gradient;

            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_RadialGradient_ExportImport_Stroke_ExtendBefore()
        {
            Path path;
            var document = this.CreatePageWithRect(out path);
            var gradient = this.CreateSampleRadialGradient();
            gradient.ExtendAfter = false;

            path.StrokeThickness = 20;
            path.Fill = new RgbColor() { A = 255, G = 255, B = 255 };
            path.Stroke = gradient;

            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_LinearGradient_ExportImport_Stroke_WithoutExtend()
        {
            Path path;
            var document = this.CreatePageWithRect(out path);
            var gradient = this.CreateSampleLinearGradient();
            gradient.ExtendAfter = false;
            gradient.ExtendBefore = false;
            path.Stroke = gradient;

            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_RadialGradient_ExportImport_Stroke_WithoutExtend()
        {
            Path path;
            var document = this.CreatePageWithRect(out path);
            var gradient = this.CreateSampleRadialGradient();
            gradient.ExtendAfter = false;
            gradient.ExtendBefore = false;
            path.Stroke = gradient;

            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_LinearGradient_Import_MinimalGradientStopsCount()
        {
            PdfFormatProvider provider = new PdfFormatProvider();
            RadFixedDocument document = provider.Import(TestUtils.GetResourceBytes("Resources/ColorSpaces/LinearGradient.pdf"));
            this.AssertSampleDocumentWithLinearGradientWithTwoGradientStops(document);

            ExportImportAndAssert(document);
        }

        private void AssertSampleDocumentWithLinearGradientWithTwoGradientStops(RadFixedDocument document)
        {
            RadFixedPage page = document.Pages[0];

            Assert.AreEqual(13, page.Content.Count);
            Assert.IsFalse(page.Content.Where(c => (c as Path) == null).Any());

            Path gradientPath = (Path)page.Content[8];
            Assert.IsTrue(gradientPath.IsFilled);
            Assert.IsFalse(gradientPath.IsStroked);

            LinearGradient gradient = (LinearGradient)gradientPath.Fill;
            Assert.AreEqual(2, gradient.GradientStops.Count);

            GradientStop firstStop = gradient.GradientStops[0];
            GradientStop secondStop = gradient.GradientStops[1];

            Assert.AreEqual(new RgbColor(255, 0, 0), firstStop.Color);
            Assert.AreEqual(0, firstStop.Offset);
            Assert.AreEqual(new RgbColor(0, 255, 0), secondStop.Color);
            Assert.AreEqual(1, secondStop.Offset);
        }

        private RadFixedDocument CreatePageWithRect(out Path path)
        {
            var document = new RadFixedDocument();
            var page = document.Pages.AddPage();
            path = page.Content.AddPath();
            path.Geometry = new RectangleGeometry()
            {
                Rect = new Rect(10, 10, 350, 350)
            };

            return document;
        }

        private RadialGradient CreateSampleRadialGradient()
        {
            var startPoint = new Point()
            {
                X = 1,
                Y = 2.5
            };

            var endPoint = new Point()
            {
                X = 673,
                Y = 453
            };

            var gradient = new RadialGradient(startPoint, endPoint, 3.43, 8.4);
            gradient.GradientStops.Add(new GradientStop(new RgbColor(10, 11, 12), 0));
            gradient.GradientStops.Add(new GradientStop(new RgbColor(223, 224, 225), 1));

            return gradient;
        }

        private LinearGradient CreateSampleLinearGradient()
        {
            var startPoint = new Point()
            {
                X = 3,
                Y = 2.2
            };

            var endPoint = new Point()
            {
                X = 523,
                Y = 443
            };

            var gradient = new LinearGradient(startPoint, endPoint);
            gradient.GradientStops.Add(new GradientStop(new RgbColor(4, 5, 6), 0));
            gradient.GradientStops.Add(new GradientStop(new RgbColor(123, 124, 125), 1));

            return gradient;
        }
    }
}
