﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Model.Editing;
using Telerik.Windows.Documents.Fixed.Model.Editing.Flow;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Fixed.Model.Objects;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Fixed.Model.Text;
using Telerik.Windows.Documents.Fixed.Tests.Utils;
using Telerik.Windows.Documents.Utilities;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Primitives;
#else
using Point = System.Windows.Point;
using Rect = System.Windows.Rect;
using Size = System.Windows.Size;
using FontFamily = System.Windows.Media.FontFamily;
using Matrix = System.Windows.Media.Matrix;
#endif

namespace Telerik.Windows.Documents.Fixed.Tests.Integration.ImportExport.Editing
{
#if VISUAL
    public class FixedContentEditorTestBase : DocumentFixedVisualBaseTest
#else
    public class FixedContentEditorTestBase : IntegrationTestClassBase
#endif
    {
        private static readonly FontBase font;
        private static readonly double fontSize = Telerik.Windows.Documents.Media.Unit.PointToDip(11);

        static FixedContentEditorTestBase()
        {
            var fontFamily = new FontFamily("Calibri");
            // Calibri can not be used
//#if NETSTANDARD
//            using (var stream = TestUtils.GetApplicationResourceStream("CALIBRI.TTF"))
//            {
//                byte[] data = stream.ReadAllBytes();
//                FontsRepository.RegisterFont(fontFamily, FontStyles.Normal, FontWeights.Normal, data);
//            }
//#endif
            FontsRepository.TryCreateFont(fontFamily, out font);
        }

        protected void ExportAndAssert(RadFixedDocument expected, RadFixedDocument document)
        {
            using (System.IO.MemoryStream stream = new System.IO.MemoryStream())
            {
                AssertDocument(expected, ExportAndImport(stream, document));
            }
#if VISUAL
            this.AssertAreVisualyEqual(document);
#endif
        }

        protected void ExportAndAssert(RadFixedDocument expected, RadFixedDocument document, Action action, int repeatActionCount)
        {
            using (System.IO.MemoryStream stream = new System.IO.MemoryStream())
            {
                AssertDocument(expected, ExportAndImport(stream, document));
            }
#if VISUAL
            this.AssertAreVisualyEqual(document, action, repeatActionCount);
#endif
        }

        protected static RadFixedDocument CreateExpectedDocument(params IEnumerable<ContentElementBase>[] elements)
        {
            RadFixedDocument document = new RadFixedDocument();
            RadFixedPage page = document.Pages.AddPage();

            foreach (var positionElements in elements)
            {
                foreach (var element in positionElements)
                {
                    page.Content.Add(element);
                }
            }

            return document;
        }

        protected static FixedContentEditor CreateFixedContentEditor(RadFixedPage page)
        {
            FixedContentEditor editor = new FixedContentEditor(page);
            editor.TextProperties.FontSize = fontSize;
            editor.TextProperties.Font = font;
            editor.GraphicProperties.IsFilled = true;
            editor.GraphicProperties.IsStroked = true;
            editor.GraphicProperties.StrokeThickness = 10;
            editor.GraphicProperties.StrokeColor = new RgbColor(0, 0, 255);

            return editor;
        }

        protected static double GetLineHeight(Block block)
        {
            return (block.TextProperties.Font.Height + block.TextProperties.Font.LineGap) * block.TextProperties.FontSize;
        }

        protected static double GetLineHeight(Block block, double contentHeight, bool includeDescent = false)
        {
            return GetHeight(block, contentHeight, includeDescent);
        }

        protected static double GetHeight(Block block, double contentHeight, bool includeDescent = false)
        {
            double result = contentHeight;
            if (includeDescent || block.TextProperties.UnderlinePattern != UnderlinePattern.None)
            {
                result += Math.Abs(block.TextProperties.Font.Descent.Value) * block.TextProperties.FontSize;
            }

            return result;
        }

        protected static TextFragment CreateTextFragment(FixedContentEditor editor, Block block, string text, double offsetX, double offsetY, int lineIndex = 0)
        {
            TextFragment textFragment = new TextFragment();
            textFragment.FontSize = fontSize;
            textFragment.Text = text;
            textFragment.Font = font;
            Matrix matrix = new Matrix();
            matrix = matrix.TranslateMatrix(offsetX, offsetY + (Math.Abs(textFragment.Font.Ascent.Value + textFragment.Font.LineGap) * textFragment.FontSize));
            matrix = matrix.MultiplyBy(editor.Position.Matrix);
            textFragment.Position = new MatrixPosition(matrix);

            return textFragment;
        }

        protected static Image CreateImage(FixedContentEditor editor, Block block, ImageSource imageSource, double width, double height, double offsetX, double offsetY)
        {
            Image image = new Image();
            image.Width = width;
            image.Height = height;
            image.ImageSource = imageSource;
            Matrix matrix = new Matrix();
            matrix = matrix.MultiplyBy(image.Position.Matrix);
            matrix = matrix.TranslateMatrix(offsetX, offsetY);
            matrix = matrix.MultiplyBy(editor.Position.Matrix);
            image.Position = new MatrixPosition(matrix);

            return image;
        }

        protected static Path CreatePath(FixedContentEditor editor, Block block, GeometryBase geometry, double offsetX, double offsetY)
        {
            Path path = new Path();
            path.Geometry = geometry;
            Rect boundingRect = geometry.Bounds;
            Matrix matrix = new Matrix();
            double strokeThickness = block.GraphicProperties.IsStroked ? block.GraphicProperties.StrokeThickness / 2 : 0;
            matrix = matrix.TranslateMatrix(-boundingRect.X + offsetX + strokeThickness, -boundingRect.Y + offsetY + strokeThickness);
            matrix = matrix.MultiplyBy(editor.Position.Matrix);
            path.Position = new MatrixPosition(matrix);
            path.IsStroked = block.GraphicProperties.IsStroked;
            path.StrokeThickness = block.GraphicProperties.StrokeThickness;
            path.Stroke = block.GraphicProperties.StrokeColor;

            return path;
        }

        protected static PathGeometry CreateEllipse(Point center, double radiusX, double radiusY)
        {
            double controlPointRatio = (Math.Sqrt(2) - 1) * 4 / 3;
            var x0 = center.X - radiusX;
            var x1 = center.X - radiusX * controlPointRatio;
            var x2 = center.X;
            var x3 = center.X + radiusX * controlPointRatio;
            var x4 = center.X + radiusX;

            var y0 = center.Y - radiusY;
            var y1 = center.Y - radiusY * controlPointRatio;
            var y2 = center.Y;
            var y3 = center.Y + radiusY * controlPointRatio;
            var y4 = center.Y + radiusY;

            PathGeometry pathGeometry = new PathGeometry();
            PathFigure pathFigure = pathGeometry.Figures.AddPathFigure();
            pathFigure.StartPoint = new Point(x2, y0);
            pathFigure.IsClosed = true;

            pathFigure.Segments.AddBezierSegment(new Point(x3, y0), new Point(x4, y1), new Point(x4, y2));
            pathFigure.Segments.AddBezierSegment(new Point(x4, y3), new Point(x3, y4), new Point(x2, y4));
            pathFigure.Segments.AddBezierSegment(new Point(x1, y4), new Point(x0, y3), new Point(x0, y2));
            pathFigure.Segments.AddBezierSegment(new Point(x0, y1), new Point(x1, y0), new Point(x2, y0));

            return pathGeometry;
        }

        protected static PathGeometry CreateLine(Point point1, Point point2)
        {
            PathGeometry pathGeometry = new PathGeometry();
            PathFigure pathFigure = pathGeometry.Figures.AddPathFigure();
            pathFigure.StartPoint = point1;
            pathFigure.Segments.AddLineSegment(point2);

            return pathGeometry;
        }

        protected static ImageSource CreateImageSource()
        {
            return CreateImageSource("1.jpg");
        }

        protected static FormSource CreateFormSource()
        {
            FormSource formSource = Telerik.Windows.Documents.Fixed.Tests.Integration.ImportExport.Forms.TestExportImportForms.CreateFormSourceWithSimpleContentElements();

            return formSource;
        }

        protected static double CalculateLeftOffset(Block block, double blockWidth, double lineWidth, int lineIndex)
        {
            double leftIndent = block.LeftIndent;
            if (lineIndex == 0)
            {
                leftIndent += block.FirstLineIndent;
            }

            switch (block.HorizontalAlignment)
            {
                case HorizontalAlignment.Center:
                    return leftIndent + CalculateStartOffsetWhenCentering(lineWidth, blockWidth - leftIndent - block.RightIndent);

                case HorizontalAlignment.Right:
                    return CalculateStartOffsetWhenPositioningAtOpositeSide(lineWidth, blockWidth) - block.RightIndent;

                case HorizontalAlignment.Left:
                    return leftIndent;

                default:
                    return 0;
            }
        }

        protected static double CalculateTopOffset(Block block, double blockHeight, double availableSize)
        {
            double verticalOffset = block.SpacingBefore;
            switch (block.VerticalAlignment)
            {
                case Telerik.Windows.Documents.Fixed.Model.Editing.Flow.VerticalAlignment.Center:
                    verticalOffset += CalculateStartOffsetWhenCentering(blockHeight, availableSize);
                    break;
                case Telerik.Windows.Documents.Fixed.Model.Editing.Flow.VerticalAlignment.Bottom:
                    verticalOffset += CalculateStartOffsetWhenPositioningAtOpositeSide(blockHeight, availableSize);
                    break;
            }

            return verticalOffset;
        }

        protected static Path DrawHighlights(FixedContentEditor editor, Block block, Rect rect)
        {
            Matrix matrix = new Matrix();
            matrix = matrix.TranslateMatrix(rect.X, rect.Y);
            matrix = matrix.MultiplyBy(editor.Position.Matrix);

            Path highlight = new Path();
            highlight.Geometry = new RectangleGeometry(new Rect(0, 0, rect.Width, rect.Height));
            highlight.Position = new MatrixPosition(matrix);
            highlight.IsFilled = true;
            highlight.IsStroked = false;
            highlight.Fill = block.TextProperties.HighlightColor;

            return highlight;
        }

        protected static Path DrawUnderline(FixedContentEditor editor, Block block, Rect rect)
        {
            Matrix matrix = new Matrix();
            matrix = matrix.TranslateMatrix(0, rect.Height - (Math.Abs(block.TextProperties.Font.Descent.Value) * block.TextProperties.FontSize));
            matrix = matrix.TranslateMatrix(0, (Math.Abs(block.TextProperties.Font.UnderlinePosition) + block.TextProperties.Font.UnderlineThickness / 2) * block.TextProperties.FontSize);
            matrix = matrix.TranslateMatrix(rect.X, rect.Y);
            matrix = matrix.MultiplyBy(editor.Position.Matrix);

            Point startPoint = new Point(0, 0);
            Point endPoint = new Point(rect.Width, 0);

            Path underline = new Path();
            PathGeometry pathGeometry = new PathGeometry();
            PathFigure figure = pathGeometry.Figures.AddPathFigure();
            figure.IsClosed = false;
            figure.StartPoint = startPoint;
            figure.Segments.AddLineSegment(endPoint);
            underline.Geometry = pathGeometry;
            underline.Stroke = block.TextProperties.UnderlineColor;
            underline.Position = new MatrixPosition(matrix);
            underline.IsFilled = false;
            underline.IsStroked = true;
            underline.StrokeThickness = block.TextProperties.Font.UnderlineThickness * block.TextProperties.FontSize;

            return underline;
        }

        protected static IEnumerable<PositionContentElement> CreateElement(FixedContentEditor editor, Block block, PositionContentElement element, Rect rect)
        {
            if (block.TextProperties.HighlightColor != null)
            {
                yield return DrawHighlights(editor, block, rect);
            }

            yield return element;

            if (block.TextProperties.UnderlinePattern != UnderlinePattern.None)
            {
                yield return DrawUnderline(editor, block, rect);
            }
        }

        protected static Size MeasureText(FixedContentEditor editor, string text)
        {
            Block block = CreateBlock(editor);
            block.InsertText(text + DocumentsHelper.ZeroWidthSymbol);

            return block.Measure();
        }

        protected static Block CreateBlock(FixedContentEditor editor)
        {
            Block block = new Block();
            block.LineSpacing = 1;
            block.TextProperties.CopyFrom(editor.TextProperties);
            block.GraphicProperties.CopyFrom(editor.GraphicProperties);

            return block;
        }

        private static double CalculateStartOffsetWhenCentering(double sizeToCenter, double sizeToFitIn)
        {
            return CalculateStartOffsetWhenPositioningAtOpositeSide(sizeToCenter, sizeToFitIn) / 2;
        }

        private static double CalculateStartOffsetWhenPositioningAtOpositeSide(double sizeToPosition, double sizeToFitIn)
        {
            return sizeToFitIn - sizeToPosition;
        }
    }
}
