﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Core.Tests.Utils;
using System.IO;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
using Telerik.Windows.Documents.Fixed.Model.Objects;
using Telerik.Windows.Documents.Fixed.Model.Text;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Fixed.Tests.Utils;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Streaming;

namespace Telerik.Windows.Documents.Fixed.Tests.Integration.ImportExport.Encryption
{
    [TestClass]
    public class TestStandardEncryption : IntegrationTestClassBase
    {
        public const string EncryptedDocumentPath = "Documents/Encryption/EncryptedWithoutPassword.pdf";
        public const string PasswordEncryptedDocumentPath = "Documents/Encryption/EncryptedWithPass.pdf";
        public const string NotEncryptedDocumentPath = "Documents/threeLines.pdf";

        [TestMethod]
        public void Test_StandardEncryption_ImportExport_WithoutPassword()
        {
            PdfExportSettings exportSettings = new PdfExportSettings();
            exportSettings.IsEncrypted = true;
            CreateDocumentExportImportAndAssert(exportSettings: exportSettings);
        }

        [TestMethod]
        public void Test_StandardEncryption_ImportExport_InvalidPassword()
        {
            PdfExportSettings exportSettings = new PdfExportSettings();
            exportSettings.IsEncrypted = true;
            exportSettings.UserPassword = "alabala";

            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() => { CreateDocumentExportImportAndAssert(exportSettings: exportSettings); });
        }

        [TestMethod]
        public void Test_StandardEncryption_ImportExport_ValidPassword()
        {
            string password = "alabala";

            PdfExportSettings exportSettings = new PdfExportSettings();
            exportSettings.IsEncrypted = true;
            exportSettings.UserPassword = password;

            PdfImportSettings importSettings = new PdfImportSettings();
            importSettings.UserPasswordNeeded += (s, e) => { e.Password = password; };

            CreateDocumentExportImportAndAssert(importSettings, exportSettings);
        }

        [TestMethod]
        public void Test_StandardEncryption_PdfStreamWriter_InvalidPasswordScenario()
        {
            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
            {
                PdfImportSettings importSettings = new PdfImportSettings();
                importSettings.UserPasswordNeeded += (s, e) => { e.Password = "Pass"; };
                PdfFileSource passwordEncryptedDocument = new PdfFileSource(TestUtils.GetApplicationResourceStream(PasswordEncryptedDocumentPath), importSettings, false);
            });
        }

        [TestMethod]
        public void Test_StandardEncryption_PdfStreamWriter_MergeDifferentEncryptions()
        {
            MemoryStream resultPdf = new MemoryStream();

            using (PdfStreamWriter writer = new PdfStreamWriter(resultPdf, true))
            {
                using (PdfFileSource encryptedPdfSource = new PdfFileSource(TestUtils.GetApplicationResourceStream(EncryptedDocumentPath)))
                {
                    writer.WritePage(encryptedPdfSource.Pages[0]);
                }

                using (PdfFileSource notEncryptedDocument = new PdfFileSource(TestUtils.GetApplicationResourceStream(NotEncryptedDocumentPath)))
                {
                    writer.WritePage(notEncryptedDocument.Pages[0]);
                }

                PdfImportSettings importSettings = new PdfImportSettings();
                importSettings.UserPasswordNeeded += (s, e) => { e.Password = "pass"; };
                using (PdfFileSource passwordEncryptedDocument = new PdfFileSource(TestUtils.GetApplicationResourceStream(PasswordEncryptedDocumentPath), importSettings, false))
                {
                    writer.WritePage(passwordEncryptedDocument.Pages[0]);
                }
            }

            this.AssertDifferentEncryptionsAreMergedSuccessfully(resultPdf);
        }

        private void AssertDifferentEncryptionsAreMergedSuccessfully(MemoryStream resultPdf)
        {
            using (resultPdf)
            {
                PdfFormatProvider provider = new PdfFormatProvider();
                RadFixedDocument document = provider.Import(resultPdf);

                Assert.AreEqual(3, document.Pages.Count);

                RadFixedPage firstPage = document.Pages[0];
                Assert.AreEqual(1, firstPage.Content.Count);
                Form firstForm = (Form)firstPage.Content[0];
                Assert.IsNotNull(firstForm.FormSource);
                Assert.AreEqual(2, firstForm.FormSource.Content.Count);
                TextFragment firstText = (TextFragment)firstForm.FormSource.Content[0];
                Assert.AreEqual("Encrypted without password.", firstText.Text);

                RadFixedPage secondPage = document.Pages[1];
                Assert.AreEqual(1, secondPage.Content.Count);
                Form secondForm = (Form)secondPage.Content[0];
                Assert.IsNotNull(secondForm.FormSource);
                Assert.AreEqual(13, secondForm.FormSource.Content.Count);
                TextFragment secondText = (TextFragment)secondForm.FormSource.Content[1];
                Assert.AreEqual("L", secondText.Text);

                RadFixedPage thirdPage = document.Pages[2];
                Assert.AreEqual(1, thirdPage.Content.Count);
                Form thirdForm = (Form)thirdPage.Content[0];
                Assert.IsNotNull(thirdForm.FormSource);
                Assert.AreEqual(2, thirdForm.FormSource.Content.Count);
                TextFragment thirdText = (TextFragment)thirdForm.FormSource.Content[0];
                Assert.AreEqual("Encrypted with password", thirdText.Text);
                Image image = (Image)thirdForm.FormSource.Content[1];
                Assert.IsNotNull(image.ImageSource);
                Assert.AreEqual(161, image.ImageSource.Width);
                Assert.AreEqual(38, image.ImageSource.Height);
            }
        }

        private static void CreateDocumentExportImportAndAssert(PdfImportSettings importSettings = null, PdfExportSettings exportSettings = null)
        {
            RadFixedDocument document = CreateDocumentWithRectangleGeometry();
            ExportImportAndAssert(document, false, importSettings, exportSettings);
        }

        private static RadFixedDocument CreateDocumentWithRectangleGeometry()
        {
            RadFixedDocument document = new RadFixedDocument();
            CreatePageWithRectangleGeometries(document, new Rect(10, 10, 100, 100));

            return document;
        }

        protected static void CreatePageWithRectangleGeometries(RadFixedDocument doc, params Rect[] rectangles)
        {
            RadFixedPage page = doc.Pages.AddPage();
            foreach (Rect rect in rectangles)
            {
                Fixed.Model.Graphics.Path path = page.Content.AddPath();
                path.Geometry = new RectangleGeometry()
                {
                    Rect = rect
                };

                path.Fill = new RgbColor(255, 0, 0);
            }
        }
    }
}
