﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.IO;
#if NETSTANDARD
using Telerik.Documents.Primitives;
using Telerik.Documents.Core.Fonts;
#else
using System.Windows;
using System.Windows.Media;
using Telerik.Windows.Documents.Fixed.Tests.Utils;
#endif

#if !WPF && !SILVERLIGHT
using Padding = Telerik.Documents.Primitives.Thickness;
#else
using Telerik.Windows.Documents.Primitives;
#endif

using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Streaming;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Model.Editing;
using Telerik.Windows.Documents.Fixed.Model.Objects;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.FixedDocumentViewers.Tests.Integration;
using Telerik.Windows.Documents.Media;

namespace Telerik.Windows.Documents.Fixed.Tests.Integration.ImportExport.Forms
{
    [TestClass]
    public class TestExportImportForms : IntegrationTestClassBase
    {
        private const string XFormTestDocumentPath = "Documents/Forms/XForm scenarios.pdf";
        private static readonly Size FormSize = new Size(Unit.PointToDip(250), Unit.PointToDip(150));
        private static readonly double ShrinkCoef = 0.7;
        private static readonly double ExpandCoef = 1.6;
        private static readonly double[] GridHorizontalDimensions = { FormSize.Width * ShrinkCoef, FormSize.Height * ExpandCoef };
        private static readonly double[] GridVerticalDimensions = { FormSize.Width * ExpandCoef, FormSize.Height * ShrinkCoef };

        [TestMethod]
        public void Test_Forms_ExportImport_With_Transformations()
        {
            RadFixedDocument document = this.CreateDocumentWithFormsAndDifferentEditors();
            this.AssertExportImportAndAssertForms(document);
        }

#if !NETSTANDARD
        [TestMethod]
        public void Test_Forms_Import_ExportImport_MultipleScenarios()
        {
            RadFixedDocument importedDocument = Import(XFormTestDocumentPath);
            RadFixedDocument generatedDocument = this.CreateDocumentWithFormsAndDifferentEditors();
            AssertDocument(generatedDocument, importedDocument);
            this.AssertExportImportAndAssertForms(importedDocument);
        }
#endif

        [TestMethod]
        public void Test_Forms_ExportImport_With_BoundingBox()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();
            RadFixedPage page = document.Pages.AddPage();
            FormSource formSource = new FormSource();
            formSource.BoundingBox = new Rect(0, 0, 100, 200);
            page.Content.AddForm(formSource);

            // Act & Assert
            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_Forms_ExportImport_With_BoundingBoxPositiveOffsets()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();
            RadFixedPage page = document.Pages.AddPage();
            FormSource formSource = new FormSource();
            formSource.BoundingBox = new Rect(20, 30, 100, 200);
            page.Content.AddForm(formSource);

            // Act & Assert
            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_Forms_ExportImport_With_BoundingBoxNegativeOffsets()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();
            RadFixedPage page = document.Pages.AddPage();
            FormSource formSource = new FormSource();
            formSource.BoundingBox = new Rect(-300, -400, 100, 200);
            page.Content.AddForm(formSource);

            // Act & Assert
            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_Forms_PdfStreamWriter_ExportImport_With_CorrectBoundingBox()
        {
            using (MemoryStream stream = new MemoryStream())
            {
                // Arrange
                RadFixedDocument document = this.CreateDocumentWithIntersectingCropBoxAndMediaBox();

                // Act
                using (PdfStreamWriter writer = new PdfStreamWriter(stream, true))
                {
                    writer.WritePage(document.Pages[0]);
                    writer.WritePage(document.Pages[1]);
                    writer.WritePage(document.Pages[2]);
                    writer.WritePage(document.Pages[3]);
                }

                RadFixedDocument importedDocument = new PdfFormatProvider().Import(stream);

                // Assert
                Form firstForm = (Form)importedDocument.Pages[0].Content[0];
                Assert.AreEqual(new Rect(100, -500, 300, 400), firstForm.FormSource.BoundingBox);

                Form secondForm = (Form)importedDocument.Pages[1].Content[0];
                Assert.AreEqual(new Rect(0, 0, 300, 400), secondForm.FormSource.BoundingBox);

                Form thirdForm = (Form)importedDocument.Pages[2].Content[0];
                Assert.AreEqual(new Rect(500, -600, 300, 400), thirdForm.FormSource.BoundingBox);

                Form fourthForm = (Form)importedDocument.Pages[3].Content[0];
                Assert.AreEqual(new Rect(0, -400, 800, 400), fourthForm.FormSource.BoundingBox);
            }
        }

#if !NETSTANDARD
        [TestMethod]
        public void Test_Forms_PdfStreamWriter_Merge_ExportImport()
        {
            using (MemoryStream pdfStream = new MemoryStream())
            {
                using (PdfStreamWriter writer = new PdfStreamWriter(pdfStream, true))
                {
                    using (PdfFileSource fileSource = new PdfFileSource(TestUtils.GetApplicationResourceStream(XFormTestDocumentPath)))
                    {
                        RadFixedDocument generatedDocument = this.CreateDocumentWithFormsAndDifferentEditors();
                        Assert.AreEqual(1, fileSource.Pages.Length);
                        Assert.AreEqual(1, generatedDocument.Pages.Count);

                        writer.WritePage(fileSource.Pages[0]);
                        writer.WritePage(generatedDocument.Pages[0]);
                        writer.WritePage(fileSource.Pages[0]);
                        writer.WritePage(generatedDocument.Pages[0]);
                    }
                }

                RadFixedDocument streamWrittenDocument = new PdfFormatProvider().Import(pdfStream);
                FormSource pageForm = this.CreatePageFormSourceWithDifferentEditors();
                AssertTestTemplateRootInnerFormsHaveSameInstancesAndSize(pageForm);
                RadFixedDocument expectedGeneratedDocument = new RadFixedDocument();

                for (int i = 0; i < 4; i++)
                {
                    RadFixedPage page = expectedGeneratedDocument.Pages.AddPage();
                    page.MediaBox = new Rect(pageForm.Size);
                    page.CropBox = page.MediaBox;
                    page.Content.AddForm(pageForm);
                }

                AssertDocument(expectedGeneratedDocument, streamWrittenDocument);
                Form first = (Form)streamWrittenDocument.Pages[0].Content[0];
                Form second = (Form)streamWrittenDocument.Pages[1].Content[0];
                Form third = (Form)streamWrittenDocument.Pages[2].Content[0];
                Form forth = (Form)streamWrittenDocument.Pages[3].Content[0];
                Assert.IsTrue(first.FormSource == third.FormSource, "FormSources should be same instances!");
                Assert.IsTrue(second.FormSource == forth.FormSource, "FormSources should be same instances!");
                Assert.IsFalse(first.FormSource == second.FormSource, "FormSources should be different instances!");
                ExportImportAndAssert(streamWrittenDocument);
            }
        }
#endif

        internal static FormSource CreateFormSourceWithSimpleContentElements()
        {
            FormSource formSource = new FormSource();
            FixedContentEditor editor = new FixedContentEditor(formSource);
            formSource.Size = FormSize;

            using (editor.SavePosition())
            using (editor.SaveProperties())
            {
                editor.Position.Scale(Unit.PointToDip(1), -Unit.PointToDip(1));
                editor.Position.Translate(0, 200);

                editor.GraphicProperties.IsFilled = true;
                editor.GraphicProperties.IsStroked = true;
                editor.GraphicProperties.StrokeThickness = 5;
                editor.GraphicProperties.StrokeColor = new RgbColor(255, 0, 0);
                editor.GraphicProperties.FillColor = new RgbColor(0, 230, 0);
                editor.GraphicProperties.StrokeDashArray = new double[] { 7, 4 };
                editor.DrawRectangle(new Rect(new Size(250, 150)));
            }

            using (editor.SavePosition())
            using (editor.SaveProperties())
            {
                Matrix imageTransform = new Matrix(0.095152423788106, 0.0941529235382309, -0.0938666666666667, 0.0946666666666667, 213.6, 19.3333333333333);
                editor.Position = new MatrixPosition(imageTransform);
                Fixed.Model.Resources.ImageSource image = CreateImageSource("sample3.jpg");
                editor.DrawImage(image);
            }

            using (editor.SavePosition())
            using (editor.SaveProperties())
            {
                editor.Position.Scale(Unit.PointToDip(1), -Unit.PointToDip(1));
                editor.Position.Translate(0, 200);
                editor.GraphicProperties.StrokeThickness = 3;
                editor.GraphicProperties.IsFilled = true;
                editor.GraphicProperties.IsStroked = true;
                editor.DrawLine(new Point(), new Point(100, 100));
            }

            editor.Position.Scale(Unit.PointToDip(1), -Unit.PointToDip(1));
            editor.Position.Translate(133.33333, 66.66667);
            editor.GraphicProperties.IsFilled = true;
            editor.GraphicProperties.IsStroked = false;
            editor.GraphicProperties.FillColor = new RgbColor(255, 0, 0);
            editor.DrawRectangle(new Rect(new Size(25, 16)));

            editor.Position.Scale(Unit.PointToDip(1), Unit.PointToDip(1));
            editor.Position.Translate(133.33333, 49.409725555555553);
            editor.GraphicProperties.FillColor = RgbColors.Black;
            editor.TextProperties.TrySetFont(new FontFamily("Calibri"));
            editor.DrawText("Text");

            return formSource;
        }

        private RadFixedDocument CreateDocumentWithFormsAndDifferentEditors()
        {
            FormSource compositeForm = this.CreateFormSourceWithNestedForms();
            RadFixedDocument document = new RadFixedDocument();

            using (RadFixedDocumentEditor documentEditor = new RadFixedDocumentEditor(document))
            {
                documentEditor.SectionProperties.PageMargins = new Padding(0);
                documentEditor.InsertFormInline(compositeForm);
            }

            document.Pages[0].MediaBox = new Rect(compositeForm.Size);
            document.Pages[0].CropBox = document.Pages[0].MediaBox;
            FixedContentEditor pageEditor = new FixedContentEditor(document.Pages[0]);
            this.AddSimpleFormTwiceWithFixedContentEditor(pageEditor);

            return document;
        }

        private FormSource CreatePageFormSourceWithDifferentEditors()
        {
            FormSource compositeForm = this.CreateFormSourceWithNestedForms();
            FormSource pageForm = new FormSource();
            pageForm.Size = compositeForm.Size;

            pageForm.Content.AddForm(compositeForm);
            FixedContentEditor pageEditor = new FixedContentEditor(pageForm);
            this.AddSimpleFormTwiceWithFixedContentEditor(pageEditor);

            return pageForm;
        }

        private void AddSimpleFormTwiceWithFixedContentEditor(FixedContentEditor editor)
        {
            FormSource simpleForm = CreateFormSourceWithSimpleContentElements();
            Block bottomContainer = new Block();
            bottomContainer.InsertForm(simpleForm, new Size(GridHorizontalDimensions[0], GridVerticalDimensions[1]));
            bottomContainer.Draw(editor, new Rect(0, GridVerticalDimensions[0], GridHorizontalDimensions[0], GridVerticalDimensions[1]));

            editor.Position.Scale(ExpandCoef, ExpandCoef);
            editor.Position.Translate(editor.Root.Size.Width, 0);
            editor.Position.Rotate(90);
            editor.DrawForm(simpleForm);
        }

        private FormSource CreateFormSourceWithNestedForms()
        {
            double compositeWidth = GridHorizontalDimensions[0] + GridHorizontalDimensions[1];
            double compositeHeight = GridVerticalDimensions[0] + GridVerticalDimensions[1];
            FormSource formSource = new FormSource();
            formSource.Size = new Size(compositeWidth, compositeHeight);
            FixedContentEditor formSourceEditor = new FixedContentEditor(formSource);
            FormSource innerFormSource = CreateNestedFormSource();

            using (formSourceEditor.SavePosition())
            {
                formSourceEditor.Position.Scale(GridHorizontalDimensions[0], GridVerticalDimensions[0]);
                formSourceEditor.DrawForm(innerFormSource);
            }

            formSourceEditor.GraphicProperties.IsFilled = true;
            formSourceEditor.GraphicProperties.IsStroked = false;
            formSourceEditor.GraphicProperties.FillColor = new RgbColor(150, 255, 165, 0);
            formSourceEditor.DrawRectangle(new Rect(formSource.Size));

            formSourceEditor.Position.Scale(GridHorizontalDimensions[1], GridVerticalDimensions[1]);
            formSourceEditor.Position.Translate(GridHorizontalDimensions[0], GridVerticalDimensions[0]);
            formSourceEditor.DrawForm(innerFormSource);

            return formSource;
        }

        private static FormSource CreateNestedFormSource()
        {
            FormSource innerFormSource = new FormSource();
            innerFormSource.Size = new Size(1, 1);
            FixedContentEditor innerFormEditor = new FixedContentEditor(innerFormSource);
            innerFormEditor.GraphicProperties.StrokeThickness = 0.2;
            innerFormEditor.GraphicProperties.StrokeColor = new RgbColor(0, 0, 255);
            innerFormEditor.GraphicProperties.IsFilled = false;
            innerFormEditor.GraphicProperties.IsStroked = true;
            innerFormEditor.DrawCircle(new Point(0.5, 0.5), 0.4);

            return innerFormSource;
        }

        private void AssertExportImportAndAssertForms(RadFixedDocument expected)
        {
            this.AssertTestTemplateRootInnerFormsHaveSameInstancesAndSize(expected.Pages[0]);

            using (MemoryStream stream = new MemoryStream())
            {
                RadFixedDocument document = ExportAndImport(stream, expected);
                this.AssertTestTemplateRootInnerFormsHaveSameInstancesAndSize(expected.Pages[0]);
                AssertDocument(expected, document);
            }
        }

        private void AssertTestTemplateRootInnerFormsHaveSameInstancesAndSize(IContentRootElement testTemplateRoot)
        {
            Form first = (Form)testTemplateRoot.Content[0];
            Form second = (Form)testTemplateRoot.Content[1];
            Form third = (Form)testTemplateRoot.Content[2];
            Form firstInner = (Form)first.FormSource.Content[0];
            Form secondInner = (Form)first.FormSource.Content[2];

            Assert.IsNotNull(first.FormSource);
            Assert.IsNotNull(second.FormSource);
            Assert.IsNotNull(third.FormSource);
            Assert.IsNotNull(firstInner.FormSource);
            Assert.IsNotNull(secondInner.FormSource);

            AssertHelper.AssertAreEqual(first.FormSource.Size, testTemplateRoot.Size);
            AssertHelper.AssertAreEqual(first.FormSource.Size, new Size(553.333333333333, 673.333333333333));
            AssertHelper.AssertAreEqual(second.FormSource.Size, new Size(333.333333333333, 200));
            AssertHelper.AssertAreEqual(firstInner.FormSource.Size, new Size(1, 1));

            Assert.IsTrue(second.FormSource == third.FormSource, "FormSources should be same instances!");
            Assert.IsTrue(firstInner.FormSource == secondInner.FormSource, "Nested FormSources should be same instances!");
        }

        private RadFixedDocument CreateDocumentWithIntersectingCropBoxAndMediaBox()
        {
            RadFixedDocument document = new RadFixedDocument();

            RadFixedPage firstPage = document.Pages.AddPage();
            firstPage.MediaBox = new Rect(0, 0, 800, 1000);
            firstPage.CropBox = new Rect(100, 100, 300, 400);

            RadFixedPage secondPage = document.Pages.AddPage();
            secondPage.MediaBox = new Rect(0, 0, 800, 1000);
            secondPage.CropBox = new Rect(-100, 600, 400, 600);

            RadFixedPage thirdPage = document.Pages.AddPage();
            thirdPage.MediaBox = new Rect(0, 0, 800, 1000);
            thirdPage.CropBox = new Rect(500, -200, 500, 600);

            RadFixedPage fourthPage = document.Pages.AddPage();
            fourthPage.MediaBox = new Rect(0, 0, 800, 1000);
            fourthPage.CropBox = new Rect(-200, 200, 1200, 400);

            return document;
        }
    }
}
