﻿using System;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Graphics;

namespace Telerik.Windows.Documents.Fixed.Tests.Integration.ImportExport.Graphics
{
    [TestClass]
    public class TestExportImportPaths : GraphicsTestBase
    {
        private static LineSegment CreateSampleSegment()
        {
            return new LineSegment()
            {
                Point = new Point(32, 55)
            };
        }

        private static PathGeometry CreatePathGeometryWithTwoIntersectingTriangles(bool isClosed)
        {
            PathGeometry pathGeometry = new PathGeometry();

            PathFigure firstTriangle = pathGeometry.Figures.AddPathFigure();
            firstTriangle.IsClosed = isClosed;
            firstTriangle.StartPoint = new Point(100, 100);
            firstTriangle.Segments.AddLineSegment().Point = new Point(50, 0);
            firstTriangle.Segments.AddLineSegment().Point = new Point(0, 100);
            firstTriangle.Segments.AddLineSegment().Point = new Point(100, 100);

            PathFigure secondTriangle = pathGeometry.Figures.AddPathFigure();
            secondTriangle.IsClosed = isClosed;
            secondTriangle.StartPoint = new Point(100, 50);
            secondTriangle.Segments.AddLineSegment().Point = new Point(0, 50);
            secondTriangle.Segments.AddLineSegment().Point = new Point(50, 150);
            secondTriangle.Segments.AddLineSegment().Point = new Point(100, 50);

            return pathGeometry;
        }

        [TestMethod]
        public void Test_Path_ExportImport_WhenPathContainsSingleFilledPathFigure_WithSingleSegment()
        {
            Path path;
            RadFixedDocument document = CreateDocumentWithSegments(out path, CreateSampleSegment());
            path.Fill = new RgbColor(255, 0, 0);
            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_Path_ExportImport_WhenPathContainsSignlePathFigure_WithoutStrokeAndFill()
        {
            Path path;
            RadFixedDocument document = CreateDocumentWithSegments(out path, CreateSampleSegment());
            path.IsFilled = false;
            path.IsStroked = false;

            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_Path_ExportImport_FillAndStroke_EvenOdd()
        {
            Path path;
            RadFixedDocument document = CreateDocumentWithSegments(FillRule.EvenOdd, out path, CreateSampleSegment());
            path.IsFilled = true;
            path.IsStroked = true;
            
            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_Path_ExportImport_CloseFillAndStroke_EvenOdd()
        {
            // Arrange
            Path path;
            RadFixedDocument document = CreateDocumentWithPath(out path);
            path.IsFilled = true;
            path.IsStroked = true;

            bool isClosed = true;
            PathGeometry pathGeometry = CreatePathGeometry(new Point(1, 1), isClosed,
                    new LineSegment()
                    {
                        Point = new Point(32, 55)
                    });

            pathGeometry.FillRule = FillRule.EvenOdd;
            path.Geometry = pathGeometry;

            // Act & Assert
            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_Path_ExportImport_CloseFillAndStroke_Nonzero()
        {
            // Arrange
            Path path;
            RadFixedDocument document = CreateDocumentWithPath(out path);
            path.IsFilled = true;
            path.IsStroked = true;

            bool isClosed = true;
            PathGeometry pathGeometry = CreatePathGeometry(new Point(1, 1), isClosed,
                    new LineSegment()
                    {
                        Point = new Point(32, 55)
                    });

            pathGeometry.FillRule = FillRule.Nonzero;
            path.Geometry = pathGeometry;

            // Act & Assert
            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_ComplexPath_ExportImport_CloseFillAndStroke_EvenOdd()
        {
            // Arrange
            RadFixedPage page;
            RadFixedDocument document = CreateDocumentWithPage(out page);

            Path path = page.Content.AddPath();
            path.IsFilled = true;
            path.IsStroked = true;

            bool isClosed = true;
            PathGeometry pathGeometry = CreatePathGeometryWithTwoIntersectingTriangles(isClosed);

            pathGeometry.FillRule = FillRule.EvenOdd;
            path.Geometry = pathGeometry;

            // Act & Assert
            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_ComplexPath_ExportImport_CloseFillAndStroke_Nonzero()
        {
            // Arrange
            RadFixedPage page;
            RadFixedDocument document = CreateDocumentWithPage(out page);

            Path path = page.Content.AddPath();
            path.IsFilled = true;
            path.IsStroked = true;

            bool isClosed = true;
            PathGeometry pathGeometry = CreatePathGeometryWithTwoIntersectingTriangles(isClosed);

            pathGeometry.FillRule = FillRule.Nonzero;
            path.Geometry = pathGeometry;

            // Act & Assert
            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_Path_ExportImport_Fill_EvenOdd()
        {
            Path path;
            RadFixedDocument document = CreateDocumentWithSegments(FillRule.EvenOdd, out path, CreateSampleSegment());
            path.IsFilled = true;
            path.IsStroked = false;

            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_Path_ExportImport_WhenPathContainsSinglePathFigure_WithStroke_WithoutFill()
        {
            Path path;
            RadFixedDocument document = CreateDocumentWithSegments(out path, CreateSampleSegment());
            path.IsFilled = false;
            path.IsStroked = true;
            path.StrokeDashArray = new double[] { 48, 3, 1, 2, 3, 6, 0, 5, 4, 3 }; 
            path.StrokeDashOffset = 43;

            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_Path_ExportImport_WhenPathContainsSingleFilledPathFigure_WithMultipleSegments()
        {
            var lineSegment = new LineSegment()
            {
                Point = new Point(15, 15)
            };

            var lineSegment2 = new LineSegment()
            {
                Point = new Point(50, 3)
            };

            var bezierSegment = new BezierSegment()
            {
                Point1 = new Point(50, 5),
                Point2 = new Point(55, 10),
                Point3 = new Point(45, 15)
            };

            Path path;
            var document = CreateDocumentWithSegments(out path, lineSegment, lineSegment2, bezierSegment);
            path.Fill = new RgbColor(0, 255, 0);

            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_Path_ExportImport_WhenPathContainsMultiplePathFigures()
        {
            Path path;
            var document = CreateDocumentWithPath(out path);

            var pathGeometry = new PathGeometry();
            var figure1 = pathGeometry.Figures.AddPathFigure();
            figure1.StartPoint = new Point(4, 5);
            figure1.Segments.AddLineSegment().Point = new Point(25, 30);

            var figure2 = pathGeometry.Figures.AddPathFigure();
            figure2.StartPoint = new Point(30, 2);
            figure2.Segments.AddLineSegment().Point = new Point(43, 85);
            var bezierSegment = figure2.Segments.AddBezierSegment();
            bezierSegment.Point1 = new Point(2, 3);
            bezierSegment.Point2 = new Point(3, 4);
            bezierSegment.Point3 = new Point(4, 5);

            figure2.Segments.AddLineSegment().Point = new Point(343, 432);
            path.Geometry = pathGeometry;
            path.MiterLimit = 5.8556;

            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_Path_ExportImport_SingleRectangleGeometry()
        {
            RadFixedDocument document = GraphicsTestBase.CreateDocumentWithRectangleGeometries(new Rect(10, 5, 300, 400));

            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_Path_ExportImport_EmptyPathGeometry()
        {
            Path path;
            RadFixedDocument document = CreateDocumentWithPath(out path);
            path.Geometry = new PathGeometry();

            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_Path_ExportImport_EmptyRectangleGeometry()
        {
            Path path;
            RadFixedDocument document = CreateDocumentWithPath(out path);
            path.Geometry = new RectangleGeometry();
            path.MiterLimit = 0.32;

            ExportImportAndAssert(document);
        }

        [TestMethod]
        public void Test_Path_ExportImport_MultipleRectangleGeometries()
        {
            RadFixedPage page;
            RadFixedDocument document = CreateDocumentWithPage(out page);
            var path1 = page.Content.AddPath();
            path1.Geometry = new RectangleGeometry()
            {
                Rect = new Rect(4, 4, 23, 34)
            };

            path1.Fill = new RgbColor(10, 20, 30);
            path1.Stroke = new RgbColor(200, 210, 220);
            path1.StrokeLineCap = LineCap.Round;
            path1.StrokeLineJoin = LineJoin.Miter;
            path1.StrokeThickness = 43;

            var path2 = page.Content.AddPath();
            path2.Geometry = new RectangleGeometry()
            {
                Rect = new Rect(44, 66, 654, 342)
            };

            path2.Fill = new RgbColor(100, 120, 130);
            path2.Stroke = new RgbColor(203, 214, 225);
            path2.StrokeLineCap = LineCap.Square;
            path2.StrokeLineJoin = LineJoin.Round;
            path2.StrokeThickness = 3; 
            path2.StrokeDashArray = new double[] { 1.4, 5.6 };
            path2.StrokeDashOffset = 3.45;

            ExportImportAndAssert(document);
        }
         
        [TestMethod]
        public void Test_Path_ExportImport_MultiplePathsOnSamePage()
        {
            RadFixedPage page;
            RadFixedDocument document = CreateDocumentWithPage(out page);

            var path1 = page.Content.AddPath();
            path1.Geometry = new RectangleGeometry()
            {
                Rect = new Rect(23, 50, 943, 43)
            };

            path1.Fill = new RgbColor(23, 54, 49);
            path1.IsStroked = false;

            var path2 = page.Content.AddPath();
            path2.Geometry = CreatePathGeometry(
                    new Point(1, 1),
                    true,
                    new LineSegment() 
                    { 
                        Point = new Point(32, 55) 
                    },
                    new BezierSegment()
                    {
                        Point1 = new Point(44, 55),
                        Point2 = new Point(32, 11),
                        Point3 = new Point(12, 23)
                    },
                    new BezierSegment()
                    {
                        Point1 = new Point(54, 10),
                        Point2 = new Point(90, 99),
                        Point3 = new Point(1, 2)
                    },
                    new LineSegment()
                    {
                        Point = new Point(3, 2)
                    }
                );

            page.Content.AddPath().Geometry = new RectangleGeometry()
            {
                Rect = new Rect(54, 32, 21, 10)
            };

            ExportImportAndAssert(document);
        }
    }
}
