﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using Rect = System.Windows.Rect;
using Size = System.Windows.Size;
#endif
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Editing;
using Telerik.Windows.Documents.Fixed.Model.Editing.Flow;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;

namespace Telerik.Windows.Documents.Fixed.Tests.Integration.ImportExport.InteractiveForms
{
    public class TestAcroFormBase : IntegrationTestClassBase
    {
        protected virtual int ExpectedFieldsCount
        {
            get
            {
                return 8;
            }
        }

        public static void DrawPageWithWidgets(RadFixedPage page)
        {
            FixedContentEditor editor = new FixedContentEditor(page);
            using (editor.SaveGraphicProperties())
            {
                editor.GraphicProperties.IsFilled = true;
                editor.GraphicProperties.IsStroked = false;
                editor.GraphicProperties.StrokeThickness = 0;
                editor.GraphicProperties.FillColor = new RgbColor(209, 178, 234);
                editor.DrawRectangle(new Rect(50, 50, editor.Root.Size.Width - 100, editor.Root.Size.Height - 100));
            }

            editor.Position.Translate(100, 100);
            Size widgetDimensions = new Size(200, 30);

            foreach (FormField field in ((RadFixedDocument)page.Parent).AcroForm.FormFields)
            {
                switch (field.FieldType)
                {
                    case FormFieldType.CheckBox:
                        CheckBoxField check = (CheckBoxField)field;
                        DrawNextWidgetWithDescription(editor, "CheckBox", (e) => e.DrawWidget(check, widgetDimensions));
                        break;
                    case FormFieldType.ComboBox:
                        ComboBoxField combo = (ComboBoxField)field;
                        DrawNextWidgetWithDescription(editor, "ComboBox", (e) => e.DrawWidget(combo, widgetDimensions));
                        break;
                    case FormFieldType.CombTextBox:
                        CombTextBoxField comb = (CombTextBoxField)field;
                        DrawNextWidgetWithDescription(editor, "Comb TextBox", (e) => e.DrawWidget(comb, widgetDimensions));
                        break;
                    case FormFieldType.ListBox:
                        ListBoxField list = (ListBoxField)field;
                        DrawNextWidgetWithDescription(editor, "ListBox", (e) => e.DrawWidget(list, new Size(widgetDimensions.Width, widgetDimensions.Width)));
                        break;
                    case FormFieldType.PushButton:
                        PushButtonField push = (PushButtonField)field;
                        DrawNextWidgetWithDescription(editor, "Button", (e) => e.DrawWidget(push, widgetDimensions));
                        break;
                    case FormFieldType.RadioButton:
                        RadioButtonField radio = (RadioButtonField)field;
                        foreach (RadioOption option in radio.Options)
                        {
                            DrawNextWidgetWithDescription(editor, option.Value, (e) => e.DrawWidget(radio, option, widgetDimensions));
                        }
                        break;
                    case FormFieldType.Signature:
                        SignatureField signature = (SignatureField)field;
                        DrawNextWidgetWithDescription(editor, "Signature", (e) => e.DrawWidget(signature, widgetDimensions));
                        break;
                    case FormFieldType.TextBox:
                        TextBoxField textBox = (TextBoxField)field;
                        DrawNextWidgetWithDescription(editor, "TextBox", (e) => e.DrawWidget(textBox, widgetDimensions));
                        break;
                }
            }
        }

        public void AssertFieldsWidgetsExportAndImport(RadFixedDocument document)
        {
            this.AssertFields(document);
            HashSet<Widget> widgetsInFields = new HashSet<Widget>();

            foreach (FormField field in document.AcroForm.FormFields)
            {
                foreach (Widget widget in field.Widgets)
                {
                    bool isAdded = widgetsInFields.Add(widget);
                    Assert.IsTrue(isAdded);
                }
            }

            int count = 0;
            foreach (Annotation annotation in document.Annotations)
            {
                if (annotation.Type == AnnotationType.Widget)
                {
                    Widget widget = (Widget)annotation;
                    bool isContainedInField = widgetsInFields.Contains(widget);
                    Assert.IsTrue(isContainedInField);
                    count++;
                }
            }

            Assert.AreEqual(count, widgetsInFields.Count);

            ExportImportAndAssert(document);
        }

        protected virtual void AssertCheckBox(CheckBoxField field)
        {
            Assert.AreEqual("checkBox", field.Name);
            Assert.IsTrue(field.IsChecked);
        }

        protected virtual void AssertComboBox(ComboBoxField field)
        {
            Assert.AreEqual("combo", field.Name);

            for (int i = 0; i < field.Options.Count; i++)
            {
                ChoiceOption option = field.Options[i];
                string expectedValue = string.Format("Combo choice {0}", i + 1);
                Assert.AreEqual(expectedValue, option.Value);
                Assert.IsNull(option.UserInterfaceValue);
            }

            Assert.AreEqual(field.Value, field.Options[2]);
        }

        protected virtual void AssertCombTextBox(CombTextBoxField field)
        {
            Assert.AreEqual("comb", field.Name);
            Assert.AreEqual(10, field.MaxLengthOfInputCharacters);
            Assert.AreEqual("0123456789", field.Value);
        }

        protected virtual void AssertListBox(ListBoxField field)
        {
            Assert.AreEqual("list", field.Name);
            Assert.IsTrue(field.AllowMultiSelection);

            for (int i = 0; i < field.Options.Count; i++)
            {
                ChoiceOption option = field.Options[i];
                string expectedValue = string.Format("List choice {0}", i + 1);
                Assert.AreEqual(expectedValue, option.Value);
                Assert.IsNull(option.UserInterfaceValue);
            }

            Assert.IsNotNull(field.Value);
            Assert.AreEqual(2, field.Value.Length);
            Assert.AreEqual(field.Value[0], field.Options[0]);
            Assert.AreEqual(field.Value[1], field.Options[2]);
        }

        protected virtual void AssertPushButton(PushButtonField field)
        {
            Assert.AreEqual("push", field.Name);
        }

        protected virtual void AssertRadioButton(RadioButtonField field)
        {
            Assert.AreEqual("radio", field.Name);
            Assert.AreEqual(2, field.Options.Count);

            for (int i = 0; i < field.Options.Count; i++)
            {
                RadioOption option = field.Options[i];
                string expectedValue = string.Format("Radio option {0}", i + 1);
                Assert.AreEqual(expectedValue, option.Value);
            }

            Assert.AreEqual(field.Value, field.Options[1]);
        }

        protected virtual void AssertSignature(SignatureField field)
        {
            Assert.AreEqual("signiture", field.Name);
        }

        protected virtual void AssertTextBox(TextBoxField field)
        {
            Assert.AreEqual("textBox", field.Name);
            Assert.AreEqual("Sample text...", field.Value);
        }

        protected virtual void CreateFields(RadFixedDocument document)
        {
            CheckBoxField check = document.AcroForm.FormFields.AddCheckBox("checkBox");
            check.IsChecked = true;

            ComboBoxField combo = document.AcroForm.FormFields.AddComboBox("combo");
            combo.Options.Add(new ChoiceOption("Combo choice 1"));
            combo.Options.Add(new ChoiceOption("Combo choice 2"));
            combo.Options.Add(new ChoiceOption("Combo choice 3"));
            combo.Options.Add(new ChoiceOption("Combo choice 4"));
            combo.Options.Add(new ChoiceOption("Combo choice 5"));
            combo.Value = combo.Options[2];

            CombTextBoxField comb = document.AcroForm.FormFields.AddCombTextBox("comb");
            comb.MaxLengthOfInputCharacters = 10;
            comb.Value = "0123456789";

            ListBoxField list = document.AcroForm.FormFields.AddListBox("list");
            list.AllowMultiSelection = true;
            list.Options.Add(new ChoiceOption("List choice 1"));
            list.Options.Add(new ChoiceOption("List choice 2"));
            list.Options.Add(new ChoiceOption("List choice 3"));
            list.Options.Add(new ChoiceOption("List choice 4"));
            list.Options.Add(new ChoiceOption("List choice 5"));
            list.Options.Add(new ChoiceOption("List choice 6"));
            list.Options.Add(new ChoiceOption("List choice 7"));
            list.Value = new ChoiceOption[] { list.Options[0], list.Options[2] };

            PushButtonField push = document.AcroForm.FormFields.AddPushButton("push");

            RadioButtonField radio = document.AcroForm.FormFields.AddRadioButton("radio");
            radio.Options.Add(new RadioOption("Radio option 1"));
            radio.Options.Add(new RadioOption("Radio option 2"));
            radio.Value = radio.Options[1];

            SignatureField signature = document.AcroForm.FormFields.AddSignature("signiture");

            TextBoxField textBox = document.AcroForm.FormFields.AddTextBox("textBox");
            textBox.Value = "Sample text...";
        }

        private static void DrawNextWidgetWithDescription(FixedContentEditor editor, string description, Action<FixedContentEditor> drawWidgetWithEditor)
        {
            double padding = 20;
            drawWidgetWithEditor(editor);

            Size annotationSize = editor.Root.Annotations[editor.Root.Annotations.Count - 1].Rect.Size;
            double x = editor.Position.Matrix.OffsetX;
            double y = editor.Position.Matrix.OffsetY;

            Block block = new Block();
            block.TextProperties.FontSize = 20;
            block.VerticalAlignment = VerticalAlignment.Center;
            block.InsertText(description);
            editor.Position.Translate(x + annotationSize.Width + padding, y);
            editor.DrawBlock(block, new Size(editor.Root.Size.Width, annotationSize.Height));

            editor.Position.Translate(x, y + annotationSize.Height + padding);
        }

        private void AssertFields(RadFixedDocument document)
        {
            Assert.AreEqual(this.ExpectedFieldsCount, document.AcroForm.FormFields.Count);

            foreach (FormField field in document.AcroForm.FormFields)
            {
                switch (field.FieldType)
                {
                    case FormFieldType.CheckBox:
                        this.AssertCheckBox((CheckBoxField)field);
                        break;
                    case FormFieldType.ComboBox:
                        this.AssertComboBox((ComboBoxField)field);
                        break;
                    case FormFieldType.CombTextBox:
                        this.AssertCombTextBox((CombTextBoxField)field);
                        break;
                    case FormFieldType.ListBox:
                        this.AssertListBox((ListBoxField)field);
                        break;
                    case FormFieldType.PushButton:
                        this.AssertPushButton((PushButtonField)field);
                        break;
                    case FormFieldType.RadioButton:
                        this.AssertRadioButton((RadioButtonField)field);
                        break;
                    case FormFieldType.Signature:
                        this.AssertSignature((SignatureField)field);
                        break;
                    case FormFieldType.TextBox:
                        this.AssertTextBox((TextBoxField)field);
                        break;
                }
            }
        }
    }
}
