﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.FormatProviders;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Navigation;
using Telerik.Windows.Documents.Fixed.Tests.Utils;

namespace Telerik.Windows.Documents.Fixed.Tests.Integration.ImportExport.LoadOnDemand
{
    [TestClass]
    public class TestImportExportLoadOnDemandBookmarks
    {
        private const double Delta = 0.001;
        private static System.IO.Stream BookmarksFile;
        private static System.IO.Stream BookmarksWithStylesFile;
        private static System.IO.Stream BookmarksWithTextColorFile;

        [ClassInitialize]
        public static void Initialize(TestContext context)
        {
            TestImportExportLoadOnDemandBookmarks.BookmarksFile = TestUtils.GetApplicationResourceStream("Resources/Bookmarks/Bookmarks.pdf");
            TestImportExportLoadOnDemandBookmarks.BookmarksWithStylesFile = TestUtils.GetApplicationResourceStream("Resources/Bookmarks/Bookmarks-with-styles.pdf");
            TestImportExportLoadOnDemandBookmarks.BookmarksWithTextColorFile = TestUtils.GetApplicationResourceStream("Resources/Bookmarks/Bookmarks-with-TextColor.pdf");
        }

        [ClassCleanup]
        public static void Cleanup()
        {
            TestImportExportLoadOnDemandBookmarks.BookmarksFile = null;
            TestImportExportLoadOnDemandBookmarks.BookmarksWithStylesFile = null;
            TestImportExportLoadOnDemandBookmarks.BookmarksWithTextColorFile = null;
        }

        [TestMethod]
        public void Bookmarks_ShouldImportExportDocumentWithCorrectBookmarksHierarchy()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act
            RadFixedDocument document = provider.Import(TestImportExportLoadOnDemandBookmarks.BookmarksFile);
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            Assert.AreEqual(2, document.Bookmarks.Count);

            BookmarkItem bookmark1 = document.Bookmarks[0];
            Assert.AreEqual("Bookmark 1", bookmark1.Title);
            Assert.AreEqual(true, bookmark1.IsExpanded);
            Assert.AreEqual(2, bookmark1.Children.Count);

            BookmarkItem bookmark11 = bookmark1.Children[0];
            Assert.AreEqual("Bookmark 1.1", bookmark11.Title);
            Assert.AreEqual(false, bookmark11.IsExpanded);
            Assert.AreEqual(2, bookmark11.Children.Count);

            BookmarkItem bookmark111 = bookmark11.Children[0];
            Assert.AreEqual("Bookmark 1.1.1", bookmark111.Title);
            Assert.AreEqual(false, bookmark111.IsExpanded);
            Assert.AreEqual(0, bookmark111.Children.Count);

            BookmarkItem bookmark112 = bookmark11.Children[1];
            Assert.AreEqual("Bookmark 1.1.2", bookmark112.Title);
            Assert.AreEqual(false, bookmark112.IsExpanded);
            Assert.AreEqual(0, bookmark112.Children.Count);

            BookmarkItem bookmark12 = bookmark1.Children[1];
            Assert.AreEqual("Bookmark 1.2", bookmark12.Title);
            Assert.AreEqual(false, bookmark12.IsExpanded);
            Assert.AreEqual(0, bookmark12.Children.Count);

            BookmarkItem bookmark2 = document.Bookmarks[1];
            Assert.AreEqual("Bookmark 2", bookmark2.Title);
            Assert.AreEqual(false, bookmark2.IsExpanded);
            Assert.AreEqual(0, bookmark2.Children.Count);
        }

        [TestMethod]
        public void Bookmarks_WhenBookmarkWithLocationDestination_ShouldImportExportBookmarkWithCorrectDestination()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act
            RadFixedDocument document = provider.Import(TestImportExportLoadOnDemandBookmarks.BookmarksFile);
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem bookmark = document.Bookmarks[0];
            Assert.IsNotNull(bookmark.Destination);
            Assert.AreEqual(DestinationType.Location, bookmark.Destination.DestinationType);

            Location destination = bookmark.Destination as Location;
            Assert.IsNotNull(destination);

            Assert.AreEqual(document.Pages[0], destination.Page);
            Assert.AreEqual(96, destination.Top);
            Assert.AreEqual(92, destination.Left);
            Assert.AreEqual(1, destination.Zoom);
        }

        [TestMethod]
        public void Bookmarks_WhenBookmarkWithPageFitDestination_ShouldImportExportBookmarkWithCorrectDestination()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act
            RadFixedDocument document = provider.Import(TestImportExportLoadOnDemandBookmarks.BookmarksFile);
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem bookmark = document.Bookmarks[0].Children[0];
            Assert.IsNotNull(bookmark.Destination);
            Assert.AreEqual(DestinationType.PageFit, bookmark.Destination.DestinationType);

            PageFit destination = bookmark.Destination as PageFit;
            Assert.IsNotNull(destination);

            Assert.AreEqual(document.Pages[0], destination.Page);
        }

        [TestMethod]
        public void Bookmarks_WhenBookmarkWithPageHorizontalFitDestination_ShouldImportExportBookmarkWithCorrectDestination()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act
            RadFixedDocument document = provider.Import(TestImportExportLoadOnDemandBookmarks.BookmarksFile);
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem bookmark = document.Bookmarks[0].Children[0].Children[0];
            Assert.IsNotNull(bookmark.Destination);
            Assert.AreEqual(DestinationType.PageHorizontalFit, bookmark.Destination.DestinationType);

            PageHorizontalFit destination = bookmark.Destination as PageHorizontalFit;
            Assert.IsNotNull(destination);

            Assert.AreEqual(document.Pages[0], destination.Page);
            Assert.AreEqual(964, destination.Top);
        }

        [TestMethod]
        public void Bookmarks_WhenBookmarkWithPageVerticalFitDestination_ShouldImportExportBookmarkWithCorrectDestination()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act
            RadFixedDocument document = provider.Import(TestImportExportLoadOnDemandBookmarks.BookmarksFile);
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem bookmark = document.Bookmarks[0].Children[0].Children[1];
            Assert.IsNotNull(bookmark.Destination);
            Assert.AreEqual(DestinationType.PageVerticalFit, bookmark.Destination.DestinationType);

            PageVerticalFit destination = bookmark.Destination as PageVerticalFit;
            Assert.IsNotNull(destination);

            Assert.AreEqual(document.Pages[0], destination.Page);
            Assert.AreEqual(654.667, destination.Left.Value, Delta);
        }

        [TestMethod]
        public void Bookmarks_WhenBookmarkWithBoundingRectangleHorizontalFitDestination_ShouldImportExportBookmarkWithCorrectDestination()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act
            RadFixedDocument document = provider.Import(TestImportExportLoadOnDemandBookmarks.BookmarksFile);
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem bookmark = document.Bookmarks[0].Children[1];
            Assert.IsNotNull(bookmark.Destination);
            Assert.AreEqual(DestinationType.BoundingRectangleHorizontalFit, bookmark.Destination.DestinationType);

            BoundingRectangleHorizontalFit destination = bookmark.Destination as BoundingRectangleHorizontalFit;
            Assert.IsNotNull(destination);

            Assert.AreEqual(document.Pages[0], destination.Page);
            Assert.AreEqual(964, destination.Top);
        }

        [TestMethod]
        public void Bookmarks_WhenBookmarkWithBoundingRectangleVerticalFitDestination_ShouldImportExportBookmarkWithCorrectDestination()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act
            RadFixedDocument document = provider.Import(TestImportExportLoadOnDemandBookmarks.BookmarksFile);
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem bookmark = document.Bookmarks[1];
            Assert.IsNotNull(bookmark.Destination);
            Assert.AreEqual(DestinationType.BoundingRectangleVerticalFit, bookmark.Destination.DestinationType);

            BoundingRectangleVerticalFit destination = bookmark.Destination as BoundingRectangleVerticalFit;
            Assert.IsNotNull(destination);

            Assert.AreEqual(document.Pages[1], destination.Page);
            Assert.AreEqual(960, destination.Left);
        }

        [TestMethod]
        public void Bookmarks_WhenBookmarksWithStyles_ShouldImportExportBookmarkWithCorrectStyle()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act
            RadFixedDocument document = provider.Import(TestImportExportLoadOnDemandBookmarks.BookmarksWithStylesFile);
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem bookmark0 = document.Bookmarks[0];
            Assert.AreEqual(BookmarkItemStyles.Normal, bookmark0.TextStyle);

            BookmarkItem bookmark1 = document.Bookmarks[1];
            Assert.AreEqual(BookmarkItemStyles.Italic, bookmark1.TextStyle);

            BookmarkItem bookmark2 = document.Bookmarks[2];
            Assert.AreEqual(BookmarkItemStyles.Bold, bookmark2.TextStyle);

            BookmarkItem bookmark3 = document.Bookmarks[3];
            Assert.AreEqual(BookmarkItemStyles.Bold | BookmarkItemStyles.Italic, bookmark3.TextStyle);
        }

        [TestMethod]
        public void Bookmarks_WhenBookmarksWithTextColor_ShouldImportExportBookmarkWithCorrectTextColor()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act
            RadFixedDocument document = provider.Import(TestImportExportLoadOnDemandBookmarks.BookmarksWithTextColorFile);
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            BookmarkItem bookmark0 = document.Bookmarks[0];
            Assert.AreEqual(new RgbColor(255, 0, 0), bookmark0.TextColor);

            BookmarkItem bookmark1 = document.Bookmarks[1];
            Assert.AreEqual(new RgbColor(0, 255, 0), bookmark1.TextColor);

            BookmarkItem bookmark2 = document.Bookmarks[2];
            Assert.AreEqual(new RgbColor(0, 0, 255), bookmark2.TextColor);
        }

        [TestMethod]
        public void Bookmarks_WhenBookmarkIsAdded_ShouldExport()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;
            RadFixedDocument document = provider.Import(TestImportExportLoadOnDemandBookmarks.BookmarksFile);

            string title = "Added bookmark";
            BookmarkItem bookmarkToAdd = new BookmarkItem(title);

            // Act
            document.Bookmarks.Add(bookmarkToAdd);
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            // Assert
            Assert.AreEqual(3, document.Bookmarks.Count);

            BookmarkItem result = document.Bookmarks[2];
            Assert.AreEqual(title, result.Title);
        }
    }
}
