﻿#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.FormatProviders;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Fixed.Model.Text;
using Telerik.Windows.Documents.Fixed.Tests.Utils;
using Telerik.Windows.Documents.Fixed.Model.Objects;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Fixed.Model.Editing;
using Telerik.Windows.Documents.Fixed.Model.Actions;

namespace Telerik.Windows.Documents.Fixed.Tests.Integration.ImportExport.LoadOnDemand
{
    [TestClass]
    public class TestImportExportLoadOnDemandContent
    {
        private static System.IO.Stream SimpleContentFile;
        private static System.IO.Stream PageWithContentArrayFile;

        private const string LoremIpsum = "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Curabitur in erat mauris. Maecenas ultricies aliquet purus non imperdiet...";
        private const string SampleText = "Sample text";
        private const string FirstContentElement = "First content element";
        private const string SecondContentElement = "Second content element";

        [ClassInitialize]
        public static void Initialize(TestContext context)
        {
            TestImportExportLoadOnDemandContent.SimpleContentFile = TestUtils.GetApplicationResourceStream("Documents/SimpleContent.pdf");
            TestImportExportLoadOnDemandContent.PageWithContentArrayFile = TestUtils.GetApplicationResourceStream("Documents/PageWithContentArray.pdf");
        }

        [ClassCleanup]
        public static void Cleanup()
        {
            TestImportExportLoadOnDemandContent.SimpleContentFile = null;
            TestImportExportLoadOnDemandContent.PageWithContentArrayFile = null;
        }

        [TestMethod]
        public void LoadOnDemand_WhenPageLoaded_ShouldPopulatePageContent()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act && Assert
            RadFixedDocument document = provider.Import(SimpleContentFile);

            Assert.AreEqual(1, document.Pages.Count);

            RadFixedPage page = document.Pages[0];
            using (document.CacheManager.BeginUsingPage(page))
            {
                this.AssertImportedPageContent(page);
            }
        }

        [TestMethod]
        public void LoadOnDemand_WhenPageLoaded_ShouldPopulatePageAnnotations()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act && Assert
            RadFixedDocument document = provider.Import(SimpleContentFile);

            Assert.AreEqual(1, document.Pages.Count);

            RadFixedPage page = document.Pages[0];
            using (document.CacheManager.BeginUsingPage(page))
            {
                this.AssertImportedPageAnnotations(page);
            }
        }

        [TestMethod]
        public void LoadOnDemand_WhenPageNotLoaded_ShouldContainEmptyContent()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act && Assert
            RadFixedDocument document = provider.Import(SimpleContentFile);

            Assert.AreEqual(1, document.Pages.Count);

            RadFixedPage page = document.Pages[0];
            Assert.AreEqual(0, page.Content.Count);
        }

        [TestMethod]
        public void LoadOnDemand_WhenPageNotLoaded_ShouldContainEmptyAnnotations()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act && Assert
            RadFixedDocument document = provider.Import(SimpleContentFile);

            Assert.AreEqual(1, document.Pages.Count);

            RadFixedPage page = document.Pages[0];
            Assert.AreEqual(0, page.Annotations.Count);
        }

        [TestMethod]
        public void LoadOnDemand_WhenPageLoaded_ShouldClearContentAfterUnload()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act && Assert
            RadFixedDocument document = provider.Import(SimpleContentFile);

            Assert.AreEqual(1, document.Pages.Count);

            RadFixedPage page = document.Pages[0];
            using (document.CacheManager.BeginUsingPage(page))
            {
            }

            Assert.AreEqual(0, page.Content.Count);
        }

        [TestMethod]
        public void LoadOnDemand_WhenPageLoaded_ShouldClearAnnotationsAfterUnload()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act && Assert
            RadFixedDocument document = provider.Import(SimpleContentFile);

            Assert.AreEqual(1, document.Pages.Count);

            RadFixedPage page = document.Pages[0];
            using (document.CacheManager.BeginUsingPage(page))
            {
            }

            Assert.AreEqual(0, page.Annotations.Count);
        }

        [TestMethod]
        public void LoadOnDemand_WhenPageWithContentArrayLoadedUnloadedAndLoadedAgain_ShouldPopulatePageContent()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act && Assert
            RadFixedDocument document = provider.Import(PageWithContentArrayFile);

            Assert.AreEqual(1, document.Pages.Count);

            RadFixedPage page = document.Pages[0];
            document.CacheManager.BeginUsingPage(page).Dispose();

            using (document.CacheManager.BeginUsingPage(page))
            {
                Assert.AreEqual(2, page.Content.Count);

                TextFragment firstContent = page.Content[0] as TextFragment;
                Assert.IsNotNull(firstContent);
                Assert.AreEqual(FirstContentElement, firstContent.Text);

                TextFragment secondContent = page.Content[1] as TextFragment;
                Assert.IsNotNull(secondContent);
                Assert.AreEqual(SecondContentElement, secondContent.Text);
            }
        }

        [TestMethod]
        public void LoadOnDemand_WhenPageContentSaved_ShouldExportAddedContent()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act && Assert
            RadFixedDocument document = provider.Import(SimpleContentFile);

            Assert.AreEqual(1, document.Pages.Count);

            RadFixedPage page = document.Pages[0];
            using (document.CacheManager.BeginUsingPage(page))
            {
                TextFragment textFragment = new TextFragment(SampleText);
                textFragment.Position.Translate(300, 300);
                page.Content.Add(textFragment);

                document.CacheManager.SavePage(page);
            }

            this.ExportImportAndAssertContent(provider, ref document);
        }

        [TestMethod]
        public void LoadOnDemand_WhenPageAnnotationsSaved_ShouldExportAddedAnnotations()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act && Assert
            RadFixedDocument document = provider.Import(SimpleContentFile);

            Assert.AreEqual(1, document.Pages.Count);

            RadFixedPage page = document.Pages[0];
            using (document.CacheManager.BeginUsingPage(page))
            {
                FixedContentEditor editor = new FixedContentEditor(page);
                editor.Position.Translate(0, 100);
                CheckBoxField checkBox = document.AcroForm.FormFields.AddCheckBox("CheckBox");
                editor.DrawWidget(checkBox, new Size(50, 50));

                document.CacheManager.SavePage(page);
            }

            this.ExportImportAndAssertAnnotations(provider, ref document);
        }

        [TestMethod]
        public void LoadOnDemand_WhenPageNotSaved_ShouldNotExportAddedContent()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act && Assert
            RadFixedDocument document = provider.Import(SimpleContentFile);

            Assert.AreEqual(1, document.Pages.Count);

            RadFixedPage page = document.Pages[0];
            using (document.CacheManager.BeginUsingPage(page))
            {
                TextFragment textFragment = new TextFragment(SampleText);
                textFragment.Position.Translate(300, 300);
                page.Content.Add(textFragment);

                document.CacheManager.SavePage(page);

                TextFragment textFragment2 = new TextFragment(SampleText + "2");
                textFragment2.Position.Translate(300, 600);
                page.Content.Add(textFragment2);
            }

            this.ExportImportAndAssertContent(provider, ref document);
        }

        [TestMethod]
        public void LoadOnDemand_WhenPageNotSaved_ShouldNotExportAddedAnnotations()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act && Assert
            RadFixedDocument document = provider.Import(SimpleContentFile);

            Assert.AreEqual(1, document.Pages.Count);

            RadFixedPage page = document.Pages[0];
            using (document.CacheManager.BeginUsingPage(page))
            {
                FixedContentEditor editor = new FixedContentEditor(page);
                editor.Position.Translate(0, 100);
                CheckBoxField checkBox = document.AcroForm.FormFields.AddCheckBox("CheckBox");
                editor.DrawWidget(checkBox, new Size(50, 50));

                document.CacheManager.SavePage(page);

                editor.Position.Translate(0, 100);
                CheckBoxField checkBox2 = document.AcroForm.FormFields.AddCheckBox("CheckBox2");
                editor.DrawWidget(checkBox2, new Size(50, 50));
            }

            this.ExportImportAndAssertAnnotations(provider, ref document);
        }

        [TestMethod]
        public void LoadOnDemand_WhenPageContentModfied_ShouldNotLoadPage()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act && Assert
            RadFixedDocument document = provider.Import(SimpleContentFile);

            Assert.AreEqual(1, document.Pages.Count);

            RadFixedPage page = document.Pages[0];
            TextFragment textFragment = page.Content.AddTextFragment("Text");

            using (document.CacheManager.BeginUsingPage(page))
            {
                Assert.AreEqual(1, page.Content.Count);
                TextFragment loadedTextFragment = page.Content[0] as TextFragment;
                Assert.IsNotNull(loadedTextFragment);
                Assert.AreEqual(textFragment, loadedTextFragment);

                Assert.AreEqual(0, page.Annotations.Count);
            }
        }

        [TestMethod]
        public void LoadOnDemand_WhenPageAnnotationsModfied_ShouldNotLoadPage()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act && Assert
            RadFixedDocument document = provider.Import(SimpleContentFile);

            Assert.AreEqual(1, document.Pages.Count);

            RadFixedPage page = document.Pages[0];
            Link link = page.Annotations.AddLink(new GoToAction());

            using (document.CacheManager.BeginUsingPage(page))
            {
                Assert.AreEqual(0, page.Content.Count);

                Link loadedLink = page.Annotations[0] as Link;
                Assert.IsNotNull(loadedLink);
                Assert.AreEqual(link, loadedLink);
            }
        }

        [TestMethod]
        public void LoadOnDemand_WhenClear_ShouldNotExportAddedContent()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act && Assert
            RadFixedDocument document = provider.Import(SimpleContentFile);

            Assert.AreEqual(1, document.Pages.Count);

            RadFixedPage page = document.Pages[0];
            using (document.CacheManager.BeginUsingPage(page))
            {
                TextFragment textFragment = new TextFragment(SampleText);
                textFragment.Position.Translate(300, 300);
                page.Content.Add(textFragment);

                document.CacheManager.SavePage(page);
            }

            document.CacheManager.Clear();

            document = this.ExportImport(provider, document);

            Assert.AreEqual(1, document.Pages.Count);

            page = document.Pages[0];

            using (document.CacheManager.BeginUsingPage(page))
            {
                this.AssertImportedPageContent(page);
            }
        }

        [TestMethod]
        public void LoadOnDemand_WhenClear_ShouldNotExportAddedAnnotations()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider();
            provider.ImportSettings.ReadingMode = ReadingMode.OnDemand;

            // Act && Assert
            RadFixedDocument document = provider.Import(SimpleContentFile);

            Assert.AreEqual(1, document.Pages.Count);

            RadFixedPage page = document.Pages[0];
            using (document.CacheManager.BeginUsingPage(page))
            {
                FixedContentEditor editor = new FixedContentEditor(page);
                editor.Position.Translate(0, 100);
                CheckBoxField checkBox = document.AcroForm.FormFields.AddCheckBox("CheckBox");
                editor.DrawWidget(checkBox, new Size(50, 50));

                document.CacheManager.SavePage(page);
            }

            document.CacheManager.Clear();

            document = this.ExportImport(provider, document);

            Assert.AreEqual(1, document.Pages.Count);

            page = document.Pages[0];

            using (document.CacheManager.BeginUsingPage(page))
            {
                this.AssertImportedPageAnnotations(page);
            }
        }

        private void ExportImportAndAssertContent(PdfFormatProvider provider, ref RadFixedDocument document)
        {
            document = this.ExportImport(provider, document);
            Assert.AreEqual(1, document.Pages.Count);

            RadFixedPage page = document.Pages[0];
            using (document.CacheManager.BeginUsingPage(page))
            {
                this.AssertExportedPageContent(page);
                this.AssertImportedPageAnnotations(page);
            }
        }

        private void ExportImportAndAssertAnnotations(PdfFormatProvider provider, ref RadFixedDocument document)
        {
            document = this.ExportImport(provider, document);
            Assert.AreEqual(1, document.Pages.Count);

            RadFixedPage page = document.Pages[0];
            using (document.CacheManager.BeginUsingPage(page))
            {
                this.AssertExportedPageAnnotations(page);
                this.AssertImportedPageContent(page);
            }
        }

        private RadFixedDocument ExportImport(PdfFormatProvider provider, RadFixedDocument document)
        {
            byte[] data = provider.Export(document);
            document = provider.Import(data);

            return document;
        }

        private void AssertImportedPageContent(RadFixedPage page)
        {
            Assert.AreEqual(3, page.Content.Count);

            foreach (Fixed.Model.Common.ContentElementBase element in page.Content)
            {
                if (element.ElementType == Fixed.Model.Common.FixedDocumentElementType.TextFragment)
                {
                    TextFragment textFragment = element as TextFragment;
                    Assert.IsNotNull(textFragment);
                    Assert.AreEqual(LoremIpsum, textFragment.Text);
                }
                else if (element.ElementType == Fixed.Model.Common.FixedDocumentElementType.Path)
                {
                    Path path = element as Path;
                    Assert.IsNotNull(path);

                    RgbColor fill = path.Fill as RgbColor;
                    Assert.IsNotNull(fill);
                    Assert.AreEqual(255, fill.R);
                    Assert.AreEqual(0, fill.G);
                    Assert.AreEqual(0, fill.B);

                    RgbColor stroke = path.Stroke as RgbColor;
                    Assert.IsNotNull(stroke);
                    Assert.AreEqual(0, stroke.R);
                    Assert.AreEqual(0, stroke.G);
                    Assert.AreEqual(255, stroke.B);

                    PathGeometry geometry = path.Geometry as PathGeometry;
                    Assert.IsNotNull(geometry);
                    Assert.AreEqual(new Rect(0, 0, 200, 100), geometry.Bounds);

                }
                else if (element.ElementType == Fixed.Model.Common.FixedDocumentElementType.Image)
                {
                    Image image = element as Image;
                    Assert.IsNotNull(image);
                    Assert.AreEqual(300, image.Width);
                    Assert.AreEqual(372, image.Height);

                    ImageSource imageSource = image.ImageSource;
                    Assert.IsNotNull(imageSource);
                }
            }
        }

        private void AssertImportedPageAnnotations(RadFixedPage page)
        {
            Assert.AreEqual(2, page.Annotations.Count);

            VariableContentWidget variableContentWidget = page.Annotations[0] as VariableContentWidget;
            Assert.IsNotNull(variableContentWidget);
            TextBoxField textBox = variableContentWidget.Field as TextBoxField;
            Assert.IsNotNull(textBox);
            Assert.AreEqual("Text", textBox.Value);

            Widget pushButonWidget = page.Annotations[1] as PushButtonWidget;
            Assert.IsNotNull(pushButonWidget);
            PushButtonField pushButtonField = pushButonWidget.Field as PushButtonField;
            Assert.IsNotNull(pushButtonField);
        }

        private void AssertExportedPageContent(RadFixedPage page)
        {
            Assert.AreEqual(4, page.Content.Count);

            TextFragment textFragment = page.Content[3] as TextFragment;
            Assert.IsNotNull(textFragment);

            Assert.AreEqual(SampleText, textFragment.Text);
        }

        private void AssertExportedPageAnnotations(RadFixedPage page)
        {
            Assert.AreEqual(3, page.Annotations.Count);

            TwoStatesButtonWidget checkBoxWidget = page.Annotations[2] as TwoStatesButtonWidget;
            Assert.IsNotNull(checkBoxWidget);
        }
    }
}
