﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.IO;
using Telerik.Windows.Documents.Fixed.FormatProviders.Text;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Editing;

namespace Telerik.Windows.Documents.Fixed.Tests.Integration.ImportExport.Text
{
    [TestClass]
    public class TestTextFormatProvider : IntegrationTestClassBase
    {
        private const string TestDocumentFolder = @"Documents/Text/";

        [TestMethod]
        public void Export_WhenDocumentContainsLongText_TextIsSplittedAndContainsMultipleLines()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();
            using (RadFixedDocumentEditor editor = new RadFixedDocumentEditor(document))
            {
                editor.InsertRun("Very long long text. Very long long text.Very long long text. Very long long text.Very long long text. Very long long text. Very long long text.");
            }

            // Act
            string actualText = Export(document);

            // Assert
            string expectedText = "----------- Page1 ------------" +
                                  "\r\nVery long long text. Very long long text.Very long long text. Very long long text.Very long long text. Very" +
                                  "\r\nlong long text. Very long long text.";

            Assert.AreEqual(expectedText, actualText);
        }

        [TestMethod]
        public void Export_WhenDocumentContainsMultplePages_TextContainsAllPages()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();
            using (RadFixedDocumentEditor editor = new RadFixedDocumentEditor(document))
            {
                editor.InsertRun("Very long long text. Very long long text.Very long long text. Very long long text.Very long long text. Very long long text. Very long long text.");
                editor.InsertPageBreak();
                editor.InsertRun("Very long long text. Very long long text.Very long long text. Very long long text.Very long long text. Very long long text. Very long long text.");
            }

            // Act
            string actualText = Export(document);

            // Assert
            string expectedText = "----------- Page1 ------------\r\n" +
                "Very long long text. Very long long text.Very long long text. Very long long text.Very long long text. Very\r\n" +
                "long long text. Very long long text.----------- Page2 ------------\r\n" +
                "Very long long text. Very long long text.Very long long text. Very long long text.Very long long text. Very\r\n" +
                "long long text. Very long long text.";

            Assert.AreEqual(expectedText, actualText);
        }

        [TestMethod]
        public void Export_WhenDocumentIsImportedFromPdfFormatProviderAndContainsStringWrittenInParenthesis_TextIsCorrect()
        {
            // Arrange
            RadFixedDocument document = IntegrationTestClassBase.Import(TestDocumentFolder + "TextDocument.pdf");

            // Act
            string actualText = Export(document);

            // Assert
            string expectedText = "----------- Page1 ------------\r\n" +
                                  "Doc2.txt\r\n" +
                                  "This is document 2\r\n" +
                                  "EOD\r\n" +
                                  "Page 1";

            Assert.AreEqual(expectedText, actualText);
        }

        [TestMethod]
        public void Export_WhenTextIsDefinedInXFormObject_TextIsCorrect()
        {
            // Arrange
            RadFixedDocument document = IntegrationTestClassBase.Import(TestDocumentFolder + "TextInFormXObject.pdf");

            // Act
            string actualText = Export(document);

            // Assert
            string expectedText = "----------- Page1 ------------\r\n" +
                                  "Sample text";

            Assert.AreEqual(expectedText, actualText);
        }

        [TestMethod]
        public void Export_WhenToUnicodeCmap_TextIsCorrect()
        {
            // Arrange
            RadFixedDocument document = IntegrationTestClassBase.Import(TestDocumentFolder + "TextDocumentWithToUnicodeRanges.pdf");

            // Act
            string actualText = Export(document);

            // Assert
            string expectedText = "----------- Page1 ------------\r\n" +
                                  @"Test PdfFormatProvider
1. Generate Report and Export as PDF A3B with
Telerik Reporting
2. Import generated PDF with PDFFormatProvider
3. Export PDF with ComplianceLevel A3B
4. Open exported PDF and copy text in Editor
Report1";

            Assert.AreEqual(expectedText, actualText);
        }

        [TestMethod]
        public void ExportImport_WhenToUnicodeCmap_TextIsCorrect()
        {
            // Arrange
            RadFixedDocument document = IntegrationTestClassBase.Import(TestDocumentFolder + "TextDocumentWithToUnicodeRanges.pdf");
            document = IntegrationTestClassBase.ExportAndImport(new MemoryStream(), document);

            // Act
            string actualText = Export(document);

            // Assert
            string expectedText = "----------- Page1 ------------\r\n" +
                                  @"Test PdfFormatProvider
1. Generate Report and Export as PDF A3B with
Telerik Reporting
2. Import generated PDF with PDFFormatProvider
3. Export PDF with ComplianceLevel A3B
4. Open exported PDF and copy text in Editor
Report1";

            Assert.AreEqual(expectedText, actualText);
        }

        private static string Export(RadFixedDocument document)
        {
            return new TextFormatProvider().Export(document);
        }
    }
}
