﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Import;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Actions;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Fixed.Model.Navigation;
using Telerik.Windows.Documents.Fixed.Model.Objects;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Fixed.Model.Text;
using Telerik.Windows.Documents.Fixed.Tests.Utils;
using Telerik.Windows.Documents.FixedDocumentViewers.Tests.Integration;

namespace Telerik.Windows.Documents.Fixed.Tests.Integration
{
    public partial class IntegrationTestClassBase
    {
        private const string LocalTestFolder = @"..\..\..\..\..\..\TestApplications\DocumentProcessingTestApplication\bin\WPF\Output";

        protected static RadFixedDocument Import(string documentPath)
        {
            using (System.IO.Stream stream = TestUtils.GetApplicationResourceStream(documentPath))
            {
                return new PdfFormatProvider().Import(stream);
            }
        }

        protected static RadFixedDocument ExportAndImport(System.IO.Stream stream, RadFixedDocument expected, bool exportToFile = false, PdfImportSettings importSettings = null, PdfExportSettings exportSettings = null)
        {
            PdfFormatProvider provider = new PdfFormatProvider();
            if (importSettings != null)
            {
                provider.ImportSettings = importSettings;
            }

            if (exportSettings != null)
            {
                provider.ExportSettings = exportSettings;
            }

            if (exportToFile)
            {
                ExportToFile(expected);
            }

            provider.Export(expected, stream);

            return provider.Import(stream);
        }

        protected static void ExportImportAndAssert(RadFixedDocument expected, bool exportToFile = false, PdfImportSettings importSettings = null, PdfExportSettings exportSettings = null)
        {
            using (System.IO.MemoryStream stream = new System.IO.MemoryStream())
            {
                RadFixedDocument document = ExportAndImport(stream, expected, exportToFile, importSettings, exportSettings);
                AssertDocument(expected, document);
            }
        }

        protected static void AssertDocument(RadFixedDocument expected, RadFixedDocument document)
        {
            AssertHelper.AssertDocument(expected, document, false);
        }

        protected static void AssertPage(RadFixedPage expected, RadFixedPage page)
        {
            AssertHelper.AssertPage(expected, page, false);
        }

        internal static void AssertAnnotationInternal(Annotation expected, Annotation annotation)
        {
            AssertAnnotation(expected, annotation);
        }

        protected static void AssertAnnotation(Annotation expected, Annotation annotation)
        {
            Assert.IsNotNull(expected);
            Assert.IsNotNull(annotation);
            Assert.AreEqual(expected.Type, annotation.Type);
            AssertHelper.AssertAreEqual(expected.Rect, annotation.Rect);
            AssertAnnotationFlags(expected, annotation);

            switch (expected.Type)
            {
                case AnnotationType.Link:
                    AssertLink((Link)expected, (Link)annotation);
                    break;
                case AnnotationType.Widget:
                    AssertWidget((Widget)expected, (Widget)annotation);
                    break;
            }
        }

        internal static void AssertAnnotationFlags(Annotation expected, Annotation annotation)
        {
            Assert.AreEqual(expected.IsContentLocked, annotation.IsContentLocked);
            Assert.AreEqual(expected.IsDisplayedWhenNotSupported, annotation.IsDisplayedWhenNotSupported);
            Assert.AreEqual(expected.IsHidden, annotation.IsHidden);
            Assert.AreEqual(expected.IsLockedByPositionAndSize, annotation.IsLockedByPositionAndSize);
            Assert.AreEqual(expected.IsPrintable, annotation.IsPrintable);
            Assert.AreEqual(expected.IsReadOnly, annotation.IsReadOnly);
            Assert.AreEqual(expected.IsRotatedWithPage, annotation.IsRotatedWithPage);
            Assert.AreEqual(expected.IsTogglingVisibilityInViewerUI, annotation.IsTogglingVisibilityInViewerUI);
            Assert.AreEqual(expected.IsVisibleInViewerUI, annotation.IsVisibleInViewerUI);
            Assert.AreEqual(expected.IsZoomingWithPage, annotation.IsZoomingWithPage);
        }

        internal static void AssertLinkInternal(Link expected, Link link)
        {
            AssertLink(expected, link);
        }

        protected static void AssertLink(Link expected, Link link)
        {
            Assert.IsNotNull(expected);
            Assert.IsNotNull(link);

            if (expected.Action != null)
            {
                Assert.IsNotNull(link.Action);
                AssertAction(expected.Action, link.Action);
            }
            else if (expected.Destination != null)
            {
                Assert.IsNotNull(link.Destination);
                AssertDestination(expected.Destination, link.Destination);
            }
        }

        private static void AssertWidget(Widget expected, Widget widget)
        {
            Assert.IsNotNull(expected);
            Assert.IsNotNull(widget);
            Assert.IsNotNull(expected.Field);
            Assert.IsNotNull(widget.Field);
            Assert.AreEqual(expected.WidgetContentType, widget.WidgetContentType);
            Assert.AreEqual(expected.Field.Name, widget.Field.Name);
            Assert.AreEqual(expected.HighlightingMode, widget.HighlightingMode);
            AssertHelper.AssertTextProperties(expected.TextProperties, widget.TextProperties);

            switch (expected.WidgetContentType)
            {
                case WidgetContentType.VariableContent:
                    AssertVariableContentWidget((VariableContentWidget)expected, (VariableContentWidget)widget);
                    break;
                case WidgetContentType.SignatureContent:
                    AssertSignatureWidget((SignatureWidget)expected, (SignatureWidget)widget);
                    break;
                case WidgetContentType.TwoStatesContent:
                    AssertTwoStatesButtonWidget((TwoStatesButtonWidget)expected, (TwoStatesButtonWidget)widget);
                    break;
                case WidgetContentType.PushButtonContent:
                    AssertPushButtonWidget((PushButtonWidget)expected, (PushButtonWidget)widget);
                    break;
            }
        }

        private static void AssertVariableContentWidget(VariableContentWidget expected, VariableContentWidget widget)
        {
            AssertDynamicAppearance(expected.AppearanceCharacteristics, widget.AppearanceCharacteristics);
        }

        private static void AssertSignatureWidget(SignatureWidget expected, SignatureWidget widget)
        {
            AssertDynamicAppearance(expected.AppearanceCharacteristics, widget.AppearanceCharacteristics);
        }

        private static void AssertTwoStatesButtonWidget(TwoStatesButtonWidget expected, TwoStatesButtonWidget widget)
        {
            AssertButtonDynamicAppearance(expected.AppearanceCharacteristics, widget.AppearanceCharacteristics);
        }

        private static void AssertPushButtonWidget(PushButtonWidget expected, PushButtonWidget widget)
        {
            AssertPushButtonDynamicAppearance(expected.AppearanceCharacteristics, widget.AppearanceCharacteristics);
        }

        private static void AssertDynamicAppearance(DynamicAppearanceCharacteristics expected, DynamicAppearanceCharacteristics appearance)
        {
            Assert.IsNotNull(expected);
            Assert.IsNotNull(appearance);
            Assert.AreEqual(expected.Rotation, appearance.Rotation);
            Assert.AreEqual(expected.BorderColor, appearance.BorderColor);
            Assert.AreEqual(expected.Background, appearance.Background);
        }

        private static void AssertButtonDynamicAppearance(ButtonAppearanceCharacteristics expected, ButtonAppearanceCharacteristics appearance)
        {
            AssertDynamicAppearance(expected, appearance);
            Assert.AreEqual(expected.NormalCaption, appearance.NormalCaption);
        }

        private static void AssertPushButtonDynamicAppearance(PushButtonAppearanceCharacteristics expected, PushButtonAppearanceCharacteristics appearance)
        {
            AssertButtonDynamicAppearance(expected, appearance);
            Assert.AreEqual(expected.MouseOverCaption, appearance.MouseOverCaption);
            Assert.AreEqual(expected.MouseDownCaption, appearance.MouseDownCaption);
            AssertHelper.AssertFormSource(expected.NormalIconSource, appearance.NormalIconSource);
            AssertHelper.AssertFormSource(expected.MouseOverIconSource, appearance.MouseOverIconSource);
            AssertHelper.AssertFormSource(expected.MouseDownIconSource, appearance.MouseDownIconSource);
            Assert.AreEqual(expected.IconFitOptions.ScaleCondition, appearance.IconFitOptions.ScaleCondition);
            Assert.AreEqual(expected.IconFitOptions.ScalingType, appearance.IconFitOptions.ScalingType);
            Assert.AreEqual(expected.IconFitOptions.BlankSpaceFromTheLeftSide, appearance.IconFitOptions.BlankSpaceFromTheLeftSide);
            Assert.AreEqual(expected.IconFitOptions.BlankSpaceFromTheBottomSide, appearance.IconFitOptions.BlankSpaceFromTheBottomSide);
            Assert.AreEqual(expected.IconFitOptions.IgnoreBorderWidth, appearance.IconFitOptions.IgnoreBorderWidth);
            Assert.AreEqual(expected.IconAndCaptionPosition, appearance.IconAndCaptionPosition);
        }

        private static void AssertAnnotationContentSource(AnnotationContentSource expected, AnnotationContentSource content)
        {
            Assert.IsNotNull(expected);
            Assert.IsNotNull(content);
            AssertHelper.AssertFormSource(expected.NormalContentSource, content.NormalContentSource);
            AssertHelper.AssertFormSource(expected.MouseDownContentSource, content.MouseDownContentSource);
            AssertHelper.AssertFormSource(expected.MouseOverContentSource, content.MouseOverContentSource);
        }

        protected static void AssertAction(Action expected, Action action)
        {
            if (AssertHelper.AssertNotNull(expected, action))
            {
                Assert.IsNotNull(action);
                Assert.AreEqual(expected.ActionType, action.ActionType);

                switch (expected.ActionType)
                {
                    case ActionType.GoTo:
                        AssertGoToAction(expected as GoToAction, action as GoToAction);
                        break;

                    case ActionType.Uri:
                        AssertUriAction(expected as UriAction, action as UriAction);
                        break;
                }
            }
        }

        protected static void AssertGoToAction(GoToAction expected, GoToAction goToAction)
        {
            if (AssertHelper.AssertNotNull(expected, goToAction))
            {
                AssertDestination(expected.Destination, goToAction.Destination);
            }
        }

        protected static void AssertUriAction(UriAction expected, UriAction uriAction)
        {
            if (AssertHelper.AssertNotNull(expected, uriAction))
            {
                Assert.AreEqual(expected.IncludeMouseCoordinates, uriAction.IncludeMouseCoordinates);
                Assert.AreEqual(expected.Uri, uriAction.Uri);
            }
        }

        protected static void AssertDestination(Destination expected, Destination destination)
        {
            if (AssertHelper.AssertNotNull(expected, destination))
            {
                RadFixedDocument document = ((RadFixedDocument)expected.Page.Parent);
                int expectedIndex = document.Pages.IndexOf(expected.Page);

                RadFixedDocument actualDocument = ((RadFixedDocument)destination.Page.Parent);
                int destinationIndex = actualDocument.Pages.IndexOf(destination.Page);

                Assert.AreEqual(expectedIndex, destinationIndex);
                Assert.AreEqual(expected.DestinationType, destination.DestinationType);

                switch (expected.DestinationType)
                {
                    case DestinationType.Location:
                        AssertLocationDestination(expected as Location, destination as Location);
                        break;

                    case DestinationType.PageHorizontalFit:
                        AssertPageHorizontalFitDestination(expected as PageHorizontalFit, destination as PageHorizontalFit);
                        break;

                    case DestinationType.PageVerticalFit:
                        AssertPageVerticalFitDestination(expected as PageVerticalFit, destination as PageVerticalFit);
                        break;

                    case DestinationType.RectangleFit:
                        AssertRectangleFitDestination(expected as RectangleFit, destination as RectangleFit);
                        break;

                    case DestinationType.BoundingRectangleHorizontalFit:
                        AssertBoundingRectangleHorizontalFitDestination(expected as BoundingRectangleHorizontalFit, destination as BoundingRectangleHorizontalFit);
                        break;

                    case DestinationType.BoundingRectangleVerticalFit:
                        AssertBoundingRectangleVerticalFitDestination(expected as BoundingRectangleVerticalFit, destination as BoundingRectangleVerticalFit);
                        break;
                }
            }
        }

        private static void AssertBoundingRectangleVerticalFitDestination(BoundingRectangleVerticalFit expected, BoundingRectangleVerticalFit destination)
        {
            if (AssertHelper.AssertNotNull(expected, destination))
            {
                Assert.AreEqual(expected.Left, destination.Left);
            }
        }

        private static void AssertBoundingRectangleHorizontalFitDestination(BoundingRectangleHorizontalFit expected, BoundingRectangleHorizontalFit destination)
        {
            if (AssertHelper.AssertNotNull(expected, destination))
            {
                Assert.AreEqual(expected.Top, destination.Top);
            }
        }

        private static void AssertRectangleFitDestination(RectangleFit expected, RectangleFit destination)
        {
            if (AssertHelper.AssertNotNull(expected, destination))
            {
                Assert.AreEqual(expected.Top, destination.Top);
                Assert.AreEqual(expected.Bottom, destination.Bottom);
                Assert.AreEqual(expected.Left, destination.Left);
                Assert.AreEqual(expected.Right, destination.Right);
            }
        }

        private static void AssertPageVerticalFitDestination(PageVerticalFit expected, PageVerticalFit destination)
        {
            if (AssertHelper.AssertNotNull(expected, destination))
            {
                Assert.AreEqual(expected.Left, destination.Left);
            }
        }

        private static void AssertPageHorizontalFitDestination(PageHorizontalFit expected, PageHorizontalFit destination)
        {
            if (AssertHelper.AssertNotNull(expected, destination))
            {
                Assert.AreEqual(expected.Top, destination.Top);
            }
        }

        private static void AssertLocationDestination(Location expected, Location location)
        {
            if (AssertHelper.AssertNotNull(expected, location))
            {
                Assert.AreEqual(expected.Left, location.Left);
                Assert.AreEqual(expected.Top, location.Top);
                Assert.AreEqual(expected.Zoom, location.Zoom);
            }
        }

        protected static void AssertContentElement(ContentElementBase expected, ContentElementBase contentElement)
        {
            AssertHelper.AssertContentElement(expected, contentElement);
        }

        protected static void AssertImage(Image expected, Image image)
        {
            AssertHelper.AssertImage(expected, image);
        }

        protected static void AssertImageSource(ImageSource expected, ImageSource imageSource)
        {
            AssertHelper.AssertImageSource(expected, imageSource);
        }

        protected static void AssertPath(Path expected, Path path)
        {
            AssertHelper.AssertPath(expected, path);
        }

        protected static void AssertTextFragment(TextFragment expected, TextFragment textFragment)
        {
            AssertHelper.AssertTextFragment(expected, textFragment);
        }

        protected static void AssertFont(FontBase expected, FontBase font)
        {
            AssertHelper.AssertFont(expected, font);
        }

        protected static void AssertClipping(Clipping expected, Clipping clipping)
        {
            AssertHelper.AssertClipping(expected, clipping);
        }

        protected static void AssertPosition(IPosition expected, IPosition position)
        {
            AssertHelper.AssertPosition(expected, position);
        }

        protected static void AssertGeometry(GeometryBase expected, GeometryBase geometry)
        {
            AssertHelper.AssertGeometry(expected, geometry);
        }

        protected static void ExportToFile(RadFixedDocument document, string extension = "")
        {
            string fileName = System.DateTime.Now.ToString("yyyy.MM.dd hh-mm-ss") + extension + ".pdf";
            using (System.IO.Stream stream = System.IO.File.OpenWrite(System.IO.Path.Combine(LocalTestFolder, fileName)))
            {
                PdfFormatProvider provider = new PdfFormatProvider();
                provider.Export(document, stream);
            }
        }

        protected static ImageSource CreateImageSource(string name)
        {
            var stream = TestUtils.GetApplicationResourceStream(string.Format("Images/{0}", name));
            var imageSource = new ImageSource(stream);

            return imageSource;
        }
    }
}
