﻿using System.IO;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Collections;
using Telerik.Windows.Documents.Fixed.Model.Navigation;
using Telerik.Windows.Documents.Fixed.Tests.Utils;

namespace Telerik.Windows.Documents.Fixed.Tests.Integration.Old.Import.Navigation
{
    [TestClass]
    public class TestBookmarksImport
    {
        const string FirstBookmarkTitle = "Bookmark 1";
        const string FirstBookmarkChildTitle = "Bookmark 1.1";
        const string FirstBookmarkLeafTitle = "Bookmark 1.1.1";
        const string SecondBookmarkLeafTitle = "Bookmark 1.1.2";
        const string SecondBookmarkChildTitle = "Bookmark 1.2";
        const string SecondBookmarkTitle = "Bookmark 2";

        private static MemoryStream sourceStream;

        [ClassInitialize]
        public static void Initialize(TestContext context)
        {
            sourceStream = ResourceHelper.GetResourceMemoryStream("Documents/Navigation/Bookmarks.pdf");
        }

        [ClassCleanup]
        public static void Cleanup()
        {
            sourceStream.Dispose();
            sourceStream = null;
        }

        [TestMethod]
        public void Import_ShouldPopulateBookmarksCollection()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider(TestBookmarksImport.sourceStream);

            // Act
            RadFixedDocument document = provider.Import();

            // Assert
            this.AssertBookmarksHierarchy(document.Bookmarks);
        }

        [TestMethod]
        public void Import_ShouldCreateBookmarksWithCorrectParent()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider(TestBookmarksImport.sourceStream);

            // Act
            RadFixedDocument document = provider.Import();

            // Assert
            BookmarkItem firstBookmark = document.Bookmarks[0];
            Assert.AreEqual(document, firstBookmark.Parent);

            BookmarkItem firstBookmarkChild = firstBookmark.Children[0];
            Assert.AreEqual(firstBookmark, firstBookmarkChild.Parent);

            BookmarkItem firstBookmarkLeaf = firstBookmarkChild.Children[0];
            Assert.AreEqual(firstBookmarkChild, firstBookmarkLeaf.Parent);
        }

        [TestMethod]
        public void Import_WhenCountIsAboveZero_ShouldCreateBookmarksWithIsExpandedSetToTrue()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider(TestBookmarksImport.sourceStream);

            // Act
            RadFixedDocument document = provider.Import();

            // Assert
            BookmarkItem bookmark = document.Bookmarks[0];
            Assert.IsTrue(bookmark.IsExpanded);
        }

        [TestMethod]
        public void Import_WhenCountIsBelowZero_ShouldCreateBookmarksWithIsExpandedSetToFalse()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider(TestBookmarksImport.sourceStream);

            // Act
            RadFixedDocument document = provider.Import();

            // Assert
            BookmarkItem bookmark = document.Bookmarks[0].Children[0];
            Assert.IsFalse(bookmark.IsExpanded);
        }

        [TestMethod]
        public void Import_WhenCountIsNull_ShouldCreateBookmarksWithIsExpandedSetToFalse()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider(TestBookmarksImport.sourceStream);

            // Act
            RadFixedDocument document = provider.Import();

            // Assert
            BookmarkItem bookmark = document.Bookmarks[0].Children[1];
            Assert.IsFalse(bookmark.IsExpanded);
        }

        [TestMethod]
        public void Import_WhenLocationDestinationType_ShouldCreateBookmarksWithCorrectDestination()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider(TestBookmarksImport.sourceStream);

            // Act
            RadFixedDocument document = provider.Import();

            // Assert
            BookmarkItem bookmarkWithLocationDestination = document.Bookmarks[0];
            this.AssertBookmarkLocationDestination(bookmarkWithLocationDestination, document.Pages[0]);
        }

        [TestMethod]
        public void Import_WhenPageFitDestinationType_ShouldCreateBookmarksWithCorrectDestination()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider(TestBookmarksImport.sourceStream);

            // Act
            RadFixedDocument document = provider.Import();

            // Assert
            BookmarkItem bookmarkWithPageFitDestination = document.Bookmarks[0].Children[0];
            this.AssertBookmarkPageFitDestination(bookmarkWithPageFitDestination, document.Pages[0]);
        }

        [TestMethod]
        public void Import_WhenPageHorizontalFitDestinationType_ShouldCreateBookmarksWithCorrectDestination()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider(TestBookmarksImport.sourceStream);

            // Act
            RadFixedDocument document = provider.Import();

            // Assert
            BookmarkItem bookmarkWithPageHorizontalFitDestination = document.Bookmarks[0].Children[0].Children[0];
            this.AssertBookmarkPageHorizontalFitDestination(bookmarkWithPageHorizontalFitDestination, document.Pages[0]);
        }

        [TestMethod]
        public void Import_WhenPageVerticalFitDestinationType_ShouldCreateBookmarksWithCorrectDestination()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider(TestBookmarksImport.sourceStream);

            // Act
            RadFixedDocument document = provider.Import();

            // Assert
            BookmarkItem bookmarkWithPageVerticalFitDestination = document.Bookmarks[0].Children[0].Children[1];
            this.AssertBookmarkPageVerticalFitDestination(bookmarkWithPageVerticalFitDestination, document.Pages[0]);
        }

        [TestMethod]
        public void Import_WhenBoundingRectangleHorizontalFitDestinationType_ShouldCreateBookmarksWithCorrectDestination()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider(TestBookmarksImport.sourceStream);

            // Act
            RadFixedDocument document = provider.Import();

            // Assert
            BookmarkItem bookmarkWithBoundingRectangleHorizontalFitDestination = document.Bookmarks[0].Children[1];
            this.AssertBookmarkBoundingRectangleHorizontalFitDestination(bookmarkWithBoundingRectangleHorizontalFitDestination, document.Pages[0]);
        }

        [TestMethod]
        public void Import_WhenBoundingRectangleVerticalFitDestinationType_ShouldCreateBookmarksWithCorrectDestination()
        {
            // Arrange
            PdfFormatProvider provider = new PdfFormatProvider(TestBookmarksImport.sourceStream);

            // Act
            RadFixedDocument document = provider.Import();

            // Assert
            BookmarkItem bookmarkWithBoundingRectangleVerticalFitDestination = document.Bookmarks[1];
            this.AssertBookmarkBoundingRectangleVerticalFitDestination(bookmarkWithBoundingRectangleVerticalFitDestination, document.Pages[1]);
        }

        private void AssertBookmarksHierarchy(BookmarksCollection bookmarks)
        {
            Assert.AreEqual(2, bookmarks.Count);

            BookmarkItem firstBookmark = bookmarks[0];
            Assert.AreEqual(FirstBookmarkTitle, firstBookmark.Title);

            BookmarksCollection children = firstBookmark.Children;
            Assert.AreEqual(2, children.Count);

            BookmarkItem firstBookmarkChild = children[0];
            Assert.AreEqual(FirstBookmarkChildTitle, firstBookmarkChild.Title);

            BookmarksCollection leafs = firstBookmarkChild.Children;
            Assert.AreEqual(2, leafs.Count);

            BookmarkItem firstBookmarkLeaf = leafs[0];
            Assert.AreEqual(FirstBookmarkLeafTitle, firstBookmarkLeaf.Title);
            Assert.AreEqual(0, firstBookmarkLeaf.Children.Count);

            BookmarkItem secondBookmarkCLeaf = leafs[1];
            Assert.AreEqual(SecondBookmarkLeafTitle, secondBookmarkCLeaf.Title);
            Assert.AreEqual(0, secondBookmarkCLeaf.Children.Count);

            BookmarkItem secondBookmarkChild = children[1];
            Assert.AreEqual(SecondBookmarkChildTitle, secondBookmarkChild.Title);
            Assert.AreEqual(0, secondBookmarkChild.Children.Count);

            BookmarkItem secondBookmark = bookmarks[1];
            Assert.AreEqual(SecondBookmarkTitle, secondBookmark.Title);
            Assert.AreEqual(0, secondBookmark.Children.Count);
        }

        private void AssertBookmarkLocationDestination(BookmarkItem bookmark, RadFixedPage page)
        {
            Location location = bookmark.Destination as Location;

            Assert.IsNotNull(location);
            Assert.AreEqual(92, location.Left);
            Assert.AreEqual(96, location.Top);
            Assert.AreEqual(1, location.Zoom);
            Assert.AreEqual(page, location.Page);
        }

        private void AssertBookmarkPageFitDestination(BookmarkItem bookmark, RadFixedPage page)
        {
            PageFit pageFit = bookmark.Destination as PageFit;

            Assert.IsNotNull(pageFit);
            Assert.AreEqual(page, pageFit.Page);
        }

        private void AssertBookmarkPageHorizontalFitDestination(BookmarkItem bookmark, RadFixedPage page)
        {
            PageHorizontalFit pageHorizontalFit = bookmark.Destination as PageHorizontalFit;

            Assert.IsNotNull(pageHorizontalFit);
            Assert.AreEqual(964, pageHorizontalFit.Top);
            Assert.AreEqual(page, pageHorizontalFit.Page);
        }

        private void AssertBookmarkPageVerticalFitDestination(BookmarkItem bookmark, RadFixedPage page)
        {
            PageVerticalFit pageVerticalFit = bookmark.Destination as PageVerticalFit;

            Assert.IsNotNull(pageVerticalFit);
            Assert.AreEqual(654, (int)pageVerticalFit.Left);
            Assert.AreEqual(page, pageVerticalFit.Page);
        }

        private void AssertBookmarkBoundingRectangleHorizontalFitDestination(BookmarkItem bookmark, RadFixedPage page)
        {
            BoundingRectangleHorizontalFit boundingRectangleHorizontalFit = bookmark.Destination as BoundingRectangleHorizontalFit;

            Assert.IsNotNull(boundingRectangleHorizontalFit);
            Assert.AreEqual(964, (int)boundingRectangleHorizontalFit.Top);
            Assert.AreEqual(page, boundingRectangleHorizontalFit.Page);
        }

        private void AssertBookmarkBoundingRectangleVerticalFitDestination(BookmarkItem bookmark, RadFixedPage page)
        {
            BoundingRectangleVerticalFit boundingRectangleVerticalFit = bookmark.Destination as BoundingRectangleVerticalFit;

            Assert.IsNotNull(boundingRectangleVerticalFit);
            Assert.AreEqual(960, (int)boundingRectangleVerticalFit.Left);
            Assert.AreEqual(page, boundingRectangleVerticalFit.Page);
        }
    }
}
