﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Collections;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Navigation;

namespace Telerik.Windows.Documents.Fixed.Tests.Model.Collections
{
    [TestClass]
    public class BookmarksCollectionTests
    {
        private const string First = "First";
        private const string Second = "Second";
        private const string Third = "Third";
        private const string Fourth = "Fourth";
        private const string Fifth = "Fifth";

        private const string DfsResult = First + Second + Fourth + Fifth + Third;
        private const string BfsResult = First + Second + Third + Fourth + Fifth;

        [TestMethod]
        public void Constructor_ShouldSetCorrectParent()
        {
            // Arrange
            RadFixedDocument parent = new RadFixedDocument();

            // Act
            BookmarksCollection bookmarks = new BookmarksCollection(parent);

            // Assert
            Assert.AreEqual(parent, ((IFixedDocumentElement)bookmarks).Parent);
        }

        [TestMethod]
        [ExpectedException(typeof(InvalidOperationException))]
        public void Add_WhenItemHasBeenAddedToCurrentCollection_ShouldThrow()
        {
            // Arrange
            RadFixedDocument parent = new RadFixedDocument();
            BookmarksCollection bookmarks = new BookmarksCollection(parent);
            BookmarkItem bookmarkItem = new BookmarkItem(string.Empty);

            bookmarks.Add(bookmarkItem);

            // Act && Assert
            bookmarks.Add(bookmarkItem);
        }

        [TestMethod]
        [ExpectedException(typeof(InvalidOperationException))]
        public void Insert_WhenItemHasBeenAddedToCurrentCollection_ShouldThrow()
        {
            // Arrange
            RadFixedDocument parent = new RadFixedDocument();
            BookmarksCollection bookmarks = new BookmarksCollection(parent);
            BookmarkItem bookmarkItem = new BookmarkItem(string.Empty);

            bookmarks.Insert(0, bookmarkItem);

            // Act && Assert
            bookmarks.Insert(0, bookmarkItem);
        }

        [TestMethod]
        [ExpectedException(typeof(InvalidOperationException))]
        public void Add_WhenItemHasBeenAddedToOtherCollection_ShouldThrow()
        {
            // Arrange
            RadFixedDocument parent1 = new RadFixedDocument();
            BookmarksCollection bookmarks1 = new BookmarksCollection(parent1);

            RadFixedDocument parent2 = new RadFixedDocument();
            BookmarksCollection bookmarks2 = new BookmarksCollection(parent2);

            BookmarkItem bookmarkItem = new BookmarkItem(string.Empty);

            bookmarks1.Add(bookmarkItem);

            // Act && Assert
            bookmarks2.Add(bookmarkItem);
        }

        [TestMethod]
        [ExpectedException(typeof(InvalidOperationException))]
        public void Insert_WhenItemHasBeenAddedToOtherCollection_ShouldThrow()
        {
            // Arrange
            RadFixedDocument parent1 = new RadFixedDocument();
            BookmarksCollection bookmarks1 = new BookmarksCollection(parent1);

            RadFixedDocument parent2 = new RadFixedDocument();
            BookmarksCollection bookmarks2 = new BookmarksCollection(parent2);

            BookmarkItem bookmarkItem = new BookmarkItem(string.Empty);

            bookmarks1.Insert(0, bookmarkItem);

            // Act && Assert
            bookmarks2.Insert(0, bookmarkItem);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Add_WhenNull_ShouldThrow()
        {
            // Arrange
            RadFixedDocument parent = new RadFixedDocument();
            BookmarksCollection bookmarks = new BookmarksCollection(parent);

            // Act && Assert
            bookmarks.Add(null);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Insert_WhenNull_ShouldThrow()
        {
            // Arrange
            RadFixedDocument parent = new RadFixedDocument();
            BookmarksCollection bookmarks = new BookmarksCollection(parent);

            // Act && Assert
            bookmarks.Insert(0, null);
        }

        [TestMethod]
        public void IterateBookmarksInDepthFirstSearch_WhenNotEmpty_ShouldIterateElementsInCorrectOrder()
        {
            // Arrange
            RadFixedDocument parent = new RadFixedDocument();
            BookmarksCollection bookmarks = new BookmarksCollection(parent);
            this.PopulateBookmarksCollection(bookmarks);
            string path = string.Empty;

            // Act
            foreach (BookmarkItem bookmark in bookmarks.IterateBookmarksInDepthFirstSearch())
            {
                path += bookmark.Title;
            }

            // Assert
            Assert.AreEqual(DfsResult, path);
        }

        [TestMethod]
        public void IterateBookmarksInDepthFirstSearch_WhenEmpty_ShouldDoNothing()
        {
            // Arrange
            RadFixedDocument parent = new RadFixedDocument();
            BookmarksCollection bookmarks = new BookmarksCollection(parent);
            string path = string.Empty;

            // Act
            foreach (BookmarkItem bookmark in bookmarks.IterateBookmarksInDepthFirstSearch())
            {
                path += bookmark.Title;
            }

            // Assert
            Assert.AreEqual(string.Empty, path);
        }

        [TestMethod]
        public void IterateBookmarksInBreadthFirstSearch_WhenNotEmpty_ShouldIterateElementsInCorrectOrder()
        {
            // Arrange
            RadFixedDocument parent = new RadFixedDocument();
            BookmarksCollection bookmarks = new BookmarksCollection(parent);
            this.PopulateBookmarksCollection(bookmarks);
            string path = string.Empty;

            // Act
            foreach (BookmarkItem bookmark in bookmarks.IterateBookmarksInBreadthFirstSearch())
            {
                path += bookmark.Title;
            }

            // Assert
            Assert.AreEqual(BfsResult, path);
        }

        [TestMethod]
        public void IterateBookmarksInBreadthFirstSearch_WhenEmpty_ShouldDoNothing()
        {
            // Arrange
            RadFixedDocument parent = new RadFixedDocument();
            BookmarksCollection bookmarks = new BookmarksCollection(parent);
            string path = string.Empty;

            // Act
            foreach (BookmarkItem bookmark in bookmarks.IterateBookmarksInBreadthFirstSearch())
            {
                path += bookmark.Title;
            }

            // Assert
            Assert.AreEqual(string.Empty, path);
        }

        private void PopulateBookmarksCollection(BookmarksCollection bookmarks)
        {
            BookmarkItem first = new BookmarkItem(First);
            bookmarks.Add(first);

            BookmarkItem second = new BookmarkItem(Second);

            BookmarkItem third = new BookmarkItem(Third);

            first.Children.Add(second);
            first.Children.Add(third);

            BookmarkItem fourth = new BookmarkItem(Fourth);

            BookmarkItem fifth = new BookmarkItem(Fifth);

            second.Children.Add(fourth);
            second.Children.Add(fifth);
        }
    }
}
