﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Linq;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Actions;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Navigation;

namespace Telerik.Windows.Documents.Fixed.Tests.Model.Collections
{
    [TestClass]
    public class NamedDestinationsCollectionTests
    {
        [TestMethod]
        public void Document_HasNamedDestinationsCollectionInitialized()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();

            // Assert
            Assert.IsNotNull(document.NamedDestinations);
        }

        [TestMethod]
        public void NewNamedDestination_ShouldHaveCorrectParent()
        {
            // Arrange
            RadFixedDocument parent = new RadFixedDocument();

            // Act
            NamedDestination destination = parent.NamedDestinations.Add("TestDestination", new GoToAction());

            // Assert
            Assert.AreEqual(parent, ((IFixedDocumentElement)destination).Parent);
        }

        [TestMethod]
        [ExpectedException(typeof(InvalidOperationException))]
        public void NamedDestination_AddedToAnotherDocument_ShouldThrow()
        {
            // Arrange
            RadFixedDocument sourceDocument = new RadFixedDocument();
            RadFixedDocument targetDocument = new RadFixedDocument();

            // Act
            NamedDestination destination = sourceDocument.NamedDestinations.Add("TestDestination", new GoToAction());

            targetDocument.NamedDestinations.Add(destination);
        }

        [TestMethod]
        [ExpectedException(typeof(InvalidOperationException))]
        public void Add_WhenNamedDestinationInstanceHasBeenAddedToCurrentCollection_ShouldThrow()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();

            NamedDestination destination = document.NamedDestinations.Add("TestDestination", new GoToAction());

            document.NamedDestinations.Add(destination);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void Add_WhenItemHasBeenAddedToCurrentCollection_ShouldThrow()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();

            // Act
            document.NamedDestinations.Add("TestDestination", new GoToAction());
            document.NamedDestinations.Add("TestDestination", new GoToAction());
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void AddNamedActionAndNamedDestination_WithSameName_ShouldThrow()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();

            // Act
            document.NamedDestinations.Add("TestDestination", new GoToAction());
            RadFixedPage page = document.Pages.AddPage();
            Destination destination = Destination.CreateDestination(DestinationType.PageFit, page);
            document.NamedDestinations.Add("TestDestination", destination);
        }

        [TestMethod]
        public void RenameDestination_ShouldAffectCollection()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();

            NamedDestination destination = document.NamedDestinations.Add("Test Name", new GoToAction());

            // Act
            destination.Name = "New Name";

            // Assert
            Assert.IsFalse(document.NamedDestinations.ContainsName("Test Name"));
            Assert.IsTrue(document.NamedDestinations.ContainsName("New Name"));
        }

        [TestMethod]
        public void RenameDestination_UsingCollection_ShouldAffectDestination()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();

            NamedDestination destination = document.NamedDestinations.Add("Test Name", new GoToAction());

            // Act
            document.NamedDestinations.Rename("Test Name", "New Name");

            // Assert
            Assert.IsFalse(document.NamedDestinations.ContainsName("Test Name"));
            Assert.IsTrue(document.NamedDestinations.ContainsName("New Name"));
            Assert.AreEqual("New Name", destination.Name);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Add_WhenNullNameAndAction_ShouldThrow()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();

            // Act && Assert
            document.NamedDestinations.Add(null, new GoToAction());
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Add_WhenNullNameAndDestination_ShouldThrow()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();

            // Act && Assert
            RadFixedPage page = document.Pages.AddPage();
            Destination destination = Destination.CreateDestination(DestinationType.PageFit, page);
            document.NamedDestinations.Add(null, destination);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Add_WhenNullAction_ShouldThrow()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();

            // Act && Assert
            GoToAction goToAction = null;
            document.NamedDestinations.Add(null, goToAction);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Add_WhenNullDestination_ShouldThrow()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();

            // Act && Assert
            Destination destination = null;
            document.NamedDestinations.Add(null, destination);
        }

        [TestMethod]
        public void GetDestination_FromIndexer()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();
            NamedDestination expected = document.NamedDestinations.Add("Test", new GoToAction());

            // Act
            NamedDestination actual = document.NamedDestinations["Test"];

            // Assert
            Assert.AreEqual(expected, actual);
        }

        [TestMethod]
        public void GetDestinationNames_ReturnAllNames()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();
            document.NamedDestinations.Add("Test", new GoToAction());
            document.NamedDestinations.Add("Test1", new GoToAction());

            // Act & Assert
            Assert.AreEqual(2, document.NamedDestinations.Count);
            Assert.AreEqual("Test", document.NamedDestinations.Names.First());
            Assert.AreEqual("Test1", document.NamedDestinations.Names.Skip(1).First());
        }

        [TestMethod]
        public void Collection_Clear()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();
            document.NamedDestinations.Add("Test", new GoToAction());
            document.NamedDestinations.Add("Test1", new GoToAction());

            // Act
            document.NamedDestinations.Clear();

            //Assert
            Assert.AreEqual(0, document.NamedDestinations.Count);
        }

        [TestMethod]
        public void Collection_ContainsKey()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();
            document.NamedDestinations.Add("Test", new GoToAction());
            document.NamedDestinations.Add("Test1", new GoToAction());

            // Act & Assert
            Assert.IsTrue(document.NamedDestinations.ContainsName("Test"));
            Assert.IsTrue(document.NamedDestinations.ContainsName("Test1"));
            Assert.IsFalse(document.NamedDestinations.ContainsName("Test2"));
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Collection_ContainsNullKey_ShouldThrow()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();

            // Act & Assert
            Assert.IsTrue(document.NamedDestinations.ContainsName(null));
        }

        [TestMethod]
        public void Collection_RemoveExistingName()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();
            document.NamedDestinations.Add("Test", new GoToAction());
            document.NamedDestinations.Add("Test1", new GoToAction());

            // Act & Assert
            Assert.IsTrue(document.NamedDestinations.Remove("Test"));
            Assert.IsTrue(document.NamedDestinations.Remove("Test1"));
            Assert.IsFalse(document.NamedDestinations.Remove("Test2"));
        }

        [TestMethod]
        public void Collection_AddSameName_Renames()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();

            NamedDestination namedDestination = new NamedDestination();
            namedDestination.Name = "Test";
            namedDestination.GoToAction = new GoToAction();

            NamedDestination namedDestination1 = new NamedDestination();
            namedDestination1.Name = "Test";
            namedDestination1.GoToAction = new GoToAction();

            // ACT
            document.NamedDestinations.Add(namedDestination);
            document.NamedDestinations.Add(namedDestination1);

            // Assert
            Assert.AreEqual("Test", namedDestination.Name);
            Assert.AreEqual("Test.0", namedDestination1.Name);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Collection_RemoveNullName_ShouldThrow()
        {
            // Arrange
            RadFixedDocument document = new RadFixedDocument();

            // Act & Assert
            Assert.IsTrue(document.NamedDestinations.Remove(null));
        }
    }
}
