﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;

namespace Telerik.Windows.Documents.Fixed.Tests.Model.ColorSpaces
{
    [TestClass]
    public class IndexedTests
    {
        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void GetColor_WhenParsIsNull_ThrowsException()
        {
            Indexed indexedColorSpace = new Indexed();
            ColorBase result = indexedColorSpace.GetColor(null);
        }

        [TestMethod]
        public void GetColor_WhenParsIsNotNullAndNotEmpty_ReturnsCorrectRgbColor()
        {
            // Arrange
            double[] pars = new double[] { 0 };
            RgbColor expectedColorBase = new RgbColor(255, 255, 255);

            Indexed indexedColorSpace = new Indexed()
            {
                Base = new DeviceGray(),
                HiVal = 1,
                Lookup = new byte[] { 255 }
            };

            // Act
            ColorBase result = indexedColorSpace.GetColor(pars);

            RgbColor rgbColor = result as RgbColor;

            // Assert
            Assert.IsNotNull(rgbColor);

            // data is taken from the Base color space and from the lookup.
            Assert.AreEqual(expectedColorBase, rgbColor);
        }

        [TestMethod]
        public void GetColor_WhenBaseIsNull_ReturnsDefaultColor()
        {
            // Arrange
            double[] pars = new double[] { 0 };

            Indexed indexedColorSpace = new Indexed()
            {
                Base = null,
                HiVal = 1,
                Lookup = new byte[] { 255 }
            };

            ColorBase expectedColorBase = indexedColorSpace.DefaultColor;

            // Act
            ColorBase result = indexedColorSpace.GetColor(pars);

            // Assert
            Assert.IsNotNull(result);

            Assert.AreEqual(expectedColorBase, result);
        }


        [TestMethod]
        public void GetColor_WhenLookupIsNull_ReturnsDefaultColor()
        {
            // Arrange
            double[] pars = new double[] { 0 };

            Indexed indexedColorSpace = new Indexed()
            {
                Base = new DeviceGray(),
                HiVal = 1,
                Lookup = null
            };

            ColorBase expectedColorBase = indexedColorSpace.DefaultColor;

            // Act
            ColorBase result = indexedColorSpace.GetColor(pars);

            // Assert
            Assert.IsNotNull(result);

            Assert.AreEqual(expectedColorBase, result);
        }

        [TestMethod]
        public void GetColor_WhenParsFirstValueIsHigherThanHiVal_ReturnsCorrectRgbColor_UsingHiValAsIndex()
        {
            // Arrange
            double[] pars = new double[] { 3 };
            RgbColor expectedColorBase = new RgbColor(255, 255, 255);

            Indexed indexedColorSpace = new Indexed()
            {
                Base = new DeviceGray(),
                HiVal = 2,
                Lookup = new byte[] { 0, 0, 255, 128 }
            };

            // Act
            ColorBase result = indexedColorSpace.GetColor(pars);

            RgbColor rgbColor = result as RgbColor;

            // Assert
            Assert.IsNotNull(rgbColor);

            // data is taken from the Base color space and from the lookup.
            Assert.AreEqual(expectedColorBase, rgbColor);
        }

        [TestMethod]
        public void GetColor_WhenParsFirstValueIsLowerThan0_ReturnsCorrectRgbColor_Using0AsIndex()
        {
            // Arrange
            double[] pars = new double[] { -1 };
            RgbColor expectedColorBase = new RgbColor(255, 255, 255);

            Indexed indexedColorSpace = new Indexed()
            {
                Base = new DeviceGray(),
                HiVal = 2,
                Lookup = new byte[] { 255, 0, 0, 128 }
            };

            // Act
            ColorBase result = indexedColorSpace.GetColor(pars);

            RgbColor rgbColor = result as RgbColor;

            // Assert
            Assert.IsNotNull(rgbColor);

            // data is taken from the Base color space and from the lookup.
            Assert.AreEqual(expectedColorBase, rgbColor);
        }

        [TestMethod]
        public void GetColor_WhenBaseIsIccBased_ReturnsCorrectRgbColor_UsingBaseAlternateColor()
        {
            // Arrange
            double[] pars = new double[] { 1 };
            RgbColor expectedColorBase = new RgbColor(255, 255, 255);

            Indexed indexedColorSpace = new Indexed()
            {
                Base = new IccBased() { Alternate = new DeviceGray() },
                HiVal = 2,
                Lookup = new byte[] { 0, 255, 0, 128 }
            };

            // Act
            ColorBase result = indexedColorSpace.GetColor(pars);

            RgbColor rgbColor = result as RgbColor;

            // Assert
            Assert.IsNotNull(rgbColor);

            // data is taken from the Base color space and from the lookup.
            Assert.AreEqual(expectedColorBase, rgbColor);
        }

        [TestMethod]
        public void GetColor_WhenBaseIsIccBased_ReturnsCorrectRgbColor_UsingNInsteadOfAlternate()
        {
            // Arrange
            double[] pars = new double[] { 2 };
            RgbColor expectedColorBase = new RgbColor(128, 128, 128);

            Indexed indexedColorSpace = new Indexed()
            {
                Base = new IccBased() { N = 1 },
                HiVal = 3,
                Lookup = new byte[] { 0, 255, 128, 5 }
            };

            // Act
            ColorBase result = indexedColorSpace.GetColor(pars);

            RgbColor rgbColor = result as RgbColor;

            // Assert
            Assert.IsNotNull(rgbColor);

            // data is taken from the Base color space and from the lookup.
            Assert.AreEqual(expectedColorBase, rgbColor);
        }
    }
}
