﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Core.Fonts.Glyphs;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Model.Fonts;

namespace Telerik.Windows.Documents.Fixed.Tests.Model.Fonts
{
    [TestClass]
    public class FontBaseTests
    {
        [TestMethod]
        public void TestFontBase_GetCachedOutlines_UsesActualFontSourceInsteadOfInitialFontSource()
        {
            TestFontSource initialFontSource = new TestFontSource();
            TestFontSource actualFontSource = new TestFontSource();
            TestFont font = new TestFont("test", initialFontSource, actualFontSource);

            CharCode codeA = new CharCode((byte)'a');
            GlyphOutlinesCollection outlinesA = font.GetCachedOutlines(codeA);

            CharCode codeB = new CharCode((byte)'b');
            GlyphOutlinesCollection outlinesB = font.GetCachedOutlines(codeB);

            Assert.IsNotNull(outlinesA);
            Assert.IsNotNull(outlinesB);
            Assert.AreNotEqual(outlinesA, outlinesB);
            Assert.AreEqual(2, font.ActualFontSourceCounter);
            Assert.AreEqual(2, actualFontSource.TotalOutlinesCount);
            Assert.AreEqual(0, initialFontSource.TotalOutlinesCount);
        }

        [TestMethod]
        public void TestFontBase_GetCachedOutlines_UsesCachedGlyphOutlines()
        {
            TestFontSource initialFontSource = new TestFontSource();
            TestFontSource actualFontSource = new TestFontSource();
            TestFont font = new TestFont("test", initialFontSource, actualFontSource);

            CharCode codeA = new CharCode((byte)'a');
            CharCode codeB = new CharCode((byte)'b');

            GlyphOutlinesCollection outlines1 = font.GetCachedOutlines(codeA);
            GlyphOutlinesCollection outlines2 = font.GetCachedOutlines(codeB);
            GlyphOutlinesCollection outlines3 = font.GetCachedOutlines(codeA);
            GlyphOutlinesCollection outlines4 = font.GetCachedOutlines(codeB);
            GlyphOutlinesCollection outlines5 = font.GetCachedOutlines(codeB);

            Assert.IsNotNull(outlines1);
            Assert.IsNotNull(outlines2);
            Assert.AreNotEqual(outlines1, outlines2);
            Assert.AreEqual(outlines1, outlines3);
            Assert.AreEqual(outlines2, outlines4);
            Assert.AreEqual(outlines2, outlines5);

            Assert.AreEqual(2, font.ActualFontSourceCounter);
            Assert.AreEqual(2, actualFontSource.TotalOutlinesCount);
            Assert.AreEqual(0, initialFontSource.TotalOutlinesCount);

            int aCount = actualFontSource.GetOutlinesCount('a');
            int bCount = actualFontSource.GetOutlinesCount('b');
            Assert.AreEqual(1, aCount);
            Assert.AreEqual(1, bCount);
        }

        [TestMethod]
        public void TestFontBase_FontFlags_DefaultFontSourceHasNoFlags()
        {
            TestFontSource source = new TestFontSource();
            TestFont font = new TestFont("test", source, source);

            this.AssertFontFlags(font);
        }

        [TestMethod]
        public void TestFontBase_FontFlags_DefaultItalicFontSourceHasItalicFlag()
        {
            TestFontSource source = new TestFontSource(true);
            TestFont font = new TestFont("test", source, source);

            this.AssertFontFlags(font, FontFlag.Italic);
        }

        [TestMethod]
        public void TestFontBase_FontFlags_OpenTypeFontSourceHasSymbolicFlag()
        {
            TestOpenTypeFontSource source = new TestOpenTypeFontSource();
            TestFont font = new TestFont("test", source, source);

            this.AssertFontFlags(font, FontFlag.Symbolic);
        }

        [TestMethod]
        public void TestFontBase_FontFlags_ItalicOpenTypeFontSourceHasBothItalicAndSymbolicFlags()
        {
            TestOpenTypeFontSource source = new TestOpenTypeFontSource(true);
            TestFont font = new TestFont("test", source, source);

            this.AssertFontFlags(font, FontFlag.Symbolic, FontFlag.Italic);
        }

        private void AssertFontFlags(FontBase font, params FontFlag[] flags)
        {
            HashSet<FontFlag> flagSet = new HashSet<FontFlag>(flags);

            foreach (FontFlag flag in Enum.GetValues(typeof(FontFlag)))
            {
                bool isContained = flagSet.Contains(flag);

                switch (flag)
                {
                    case FontFlag.FixedPitch:
                        Assert.AreEqual(isContained, font.IsFixedPitch);
                        break;
                    case FontFlag.Serif:
                        Assert.AreEqual(isContained, font.IsSerif);
                        break;
                    case FontFlag.Symbolic:
                        Assert.AreEqual(isContained, font.IsSymbolic);
                        break;
                    case FontFlag.Script:
                        Assert.AreEqual(isContained, font.IsScript);
                        break;
                    case FontFlag.Nonsymbolic:
                        Assert.AreEqual(isContained, font.IsNonSymbolic);
                        break;
                    case FontFlag.Italic:
                        Assert.AreEqual(isContained, font.IsItalic);
                        break;
                    case FontFlag.AllCap:
                        Assert.AreEqual(isContained, font.IsAllCap);
                        break;
                    case FontFlag.SmallCap:
                        Assert.AreEqual(isContained, font.IsSmallCap);
                        break;
                    case FontFlag.ForceBold:
                        Assert.AreEqual(isContained, font.IsForcingBold);
                        break;
                    default:
                        throw new NotSupportedException(string.Format("Flag {0} is not supported!", flag));
                }
            }
        }
    }
}
