﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
#else
using System.Windows.Media;
using Telerik.Windows.Documents.Core.Fonts.OpenType;
using Telerik.Windows.Documents.Fixed.Model;
#endif
using Telerik.Windows.Documents.Core.Fonts;
using Telerik.Windows.Documents.Core.Fonts.Glyphs;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Fonts.Encoding;

namespace Telerik.Windows.Documents.Fixed.Tests.Model.Fonts
{

    [TestClass]
    public class TrueTypeFontTests
    {
        [TestMethod]
        public void TestTrueTypeFont_ActualFontSourceRemainsTheSameWhenOpenTypeFontSource()
        {
            TestOpenTypeFontSource fontSource = new TestOpenTypeFontSource();
            TrueTypeFont font = new TrueTypeFont("test", fontSource);
            FontSource actualFontSource = font.ActualFontSource;

            Assert.AreEqual(fontSource, actualFontSource);
        }

#if !NETSTANDARD
        [TestMethod]
        public void TestTrueTypeFont_ActualFontSourceFallbacksToArialWhenEmptyOriginalSource()
        {
            FontProperties properties = new FontProperties(new FontFamily("test"));
            TrueTypeFont font = new TrueTypeFont("test", EmptyFontSource.Instance);
            FontSource actualFontSource = font.ActualFontSource;
            FontProperties fallbackProperties = new FontProperties(new FontFamily(FixedDocumentDefaults.TrueTypeFallbackFontName));
            OpenTypeFontSource expectedFallbackFont = FontsRepository.FontsManager.GetOpenTypeFontSource(fallbackProperties);

            Assert.AreEqual(FixedDocumentDefaults.TrueTypeFallbackFontName, actualFontSource.FontFamily);
            Assert.AreEqual(expectedFallbackFont, actualFontSource);
        }
#endif

        [TestMethod]
        public void TestTrueTypeFont_GetGlyphInitializationInfoReturnsUnknownPlatformTypeWhenNullFontSource()
        {
            TrueTypeGlyphInitializationInfo info = TrueTypeFont.GetGlyphInitializationInfo(null, null, null);

            Assert.IsNotNull(info);
            Assert.AreEqual(TrueTypeGlyphInitializationInfo.PlatformType.Unknown, info.Platform);
            Assert.IsNull(info.AppendingFirstByte);
        }

        [TestMethod]
        public void TestTrueTypeFont_GetGlyphInitializationInfoReturnUnknownPlatformWhenNoCMapTable()
        {
            TestOpenTypeFontSource fontSource = new TestOpenTypeFontSource();
            TestFont font = new TestFont("test", fontSource, fontSource);
            SimpleEncoding encoding = new SimpleEncoding();
            TrueTypeGlyphInitializationInfo info = TrueTypeFont.GetGlyphInitializationInfo(fontSource, font, encoding);

            Assert.IsNotNull(info);
            Assert.AreEqual(TrueTypeGlyphInitializationInfo.PlatformType.Unknown, info.Platform);
            Assert.AreEqual(fontSource, info.FontSource);
            Assert.AreEqual(encoding, info.Encoding);
            Assert.IsNull(info.AppendingFirstByte);
        }

        [TestMethod]
        public void TestTrueTypeFont_GetGlyphInitializationInfoReturnUnknownPlatformWhenNoSuitableFontDescriptorAndEncoding()
        {
            TestOpenTypeFontSource fontSource = new TestOpenTypeFontSource();
            TestFont font = new TestFont("test", fontSource, fontSource);
            font.IsSymbolic = false;
            font.IsNonSymbolic = false;
            SimpleEncoding encoding = new SimpleEncoding();
            encoding.PredefinedEncoding.Value = PredefinedSimpleEncoding.StandardEncodingName;
            TestCMap map = (TestCMap)fontSource.CMap;
            TestCMapTable table31 = new TestCMapTable(31);
            TestCMapTable table10 = new TestCMapTable(10);
            TestCMapTable table30 = new TestCMapTable(30);
            map.AddCMapTable(3, 1, table31);
            map.AddCMapTable(1, 0, table10);
            map.AddCMapTable(3, 0, table30);
            TrueTypeGlyphInitializationInfo info = TrueTypeFont.GetGlyphInitializationInfo(fontSource, font, encoding);

            Assert.IsNotNull(info);
            Assert.AreEqual(TrueTypeGlyphInitializationInfo.PlatformType.Unknown, info.Platform);
            Assert.AreEqual(fontSource, info.FontSource);
            Assert.AreEqual(encoding, info.Encoding);
            Assert.IsNull(info.CMap);
            Assert.IsNull(info.AppendingFirstByte);
        }

        [TestMethod]
        public void TestTrueTypeFont_GetGlyphInitializationInfoIsMicrosoftNonSymbolicBecauseOfFontDescriptor()
        {
            TestOpenTypeFontSource fontSource = new TestOpenTypeFontSource();
            TestFont font = new TestFont("test", fontSource, fontSource);
            font.IsNonSymbolic = true;
            TestCMap map = (TestCMap)fontSource.CMap;
            TestCMapTable table31 = new TestCMapTable(31);
            TestCMapTable table10 = new TestCMapTable(10);
            map.AddCMapTable(3, 1, table31);
            map.AddCMapTable(1, 0, table10);
            TrueTypeGlyphInitializationInfo info = TrueTypeFont.GetGlyphInitializationInfo(fontSource, font, null);

            Assert.IsNotNull(info);
            Assert.AreEqual(TrueTypeGlyphInitializationInfo.PlatformType.MicrosoftNonSymbolic, info.Platform);
            Assert.AreEqual(fontSource, info.FontSource);
            Assert.IsNull(info.Encoding);
            Assert.IsNull(info.AppendingFirstByte);
            Assert.AreEqual(table31, info.CMap);
        }

        [TestMethod]
        public void TestTrueTypeFont_GetGlyphInitializationInfoIsMacintoshNonSymbolicBecauseOfEncoding()
        {
            TestOpenTypeFontSource fontSource = new TestOpenTypeFontSource();
            TestFont font = new TestFont("test", fontSource, fontSource);
            font.IsNonSymbolic = false;
            SimpleEncoding encoding = new SimpleEncoding();
            encoding.PredefinedEncoding.Value = PredefinedSimpleEncoding.MacRomanEncodingName;
            TestCMap map = (TestCMap)fontSource.CMap;
            TestCMapTable table = new TestCMapTable(2);
            map.AddCMapTable(1, 0, table);
            TrueTypeGlyphInitializationInfo info = TrueTypeFont.GetGlyphInitializationInfo(fontSource, font, encoding);

            Assert.IsNotNull(info);
            Assert.AreEqual(TrueTypeGlyphInitializationInfo.PlatformType.MacintoshNonSymbolic, info.Platform);
            Assert.AreEqual(fontSource, info.FontSource);
            Assert.AreEqual(encoding, info.Encoding);
            Assert.AreEqual(table, info.CMap);
            Assert.IsNull(info.AppendingFirstByte);
        }

        [TestMethod]
        public void TestTrueTypeFont_GetGlyphInitializationInfoIsMicrosoftSymbolicBecauseOfFontDescriptor()
        {
            TestOpenTypeFontSource fontSource = new TestOpenTypeFontSource();
            TestFont font = new TestFont("test", fontSource, fontSource);
            font.IsSymbolic = true;
            SimpleEncoding encoding = new SimpleEncoding();
            encoding.PredefinedEncoding.Value = PredefinedSimpleEncoding.PdfDocEncodingName;
            TestCMap map = (TestCMap)fontSource.CMap;
            TestCMapTable table31 = new TestCMapTable(31);
            TestCMapTable table10 = new TestCMapTable(10);
            TestCMapTable table30 = new TestCMapTable(1001);
            map.AddCMapTable(3, 1, table31);
            map.AddCMapTable(1, 0, table10);
            map.AddCMapTable(3, 0, table30);
            TrueTypeGlyphInitializationInfo info = TrueTypeFont.GetGlyphInitializationInfo(fontSource, font, encoding);

            Assert.IsNotNull(info);
            Assert.AreEqual(TrueTypeGlyphInitializationInfo.PlatformType.MicrosoftSymbolic, info.Platform);
            Assert.AreEqual(fontSource, info.FontSource);
            Assert.AreEqual(encoding, info.Encoding);
            Assert.AreEqual((byte)3, info.AppendingFirstByte);
            Assert.AreEqual(table30, info.CMap);
        }

        [TestMethod]
        public void TestTrueTypeFont_GetGlyphInitializationInfoIsMacintoshSymbolicBecauseOfEncoding()
        {
            TestOpenTypeFontSource fontSource = new TestOpenTypeFontSource();
            TestFont font = new TestFont("test", fontSource, fontSource);
            font.IsNonSymbolic = false;
            font.IsSymbolic = false;
            TestCMap map = (TestCMap)fontSource.CMap;
            TestCMapTable table = new TestCMapTable(2);
            map.AddCMapTable(1, 0, table);
            TrueTypeGlyphInitializationInfo info = TrueTypeFont.GetGlyphInitializationInfo(fontSource, font, null);

            Assert.IsNotNull(info);
            Assert.AreEqual(TrueTypeGlyphInitializationInfo.PlatformType.MacintoshSymbolic, info.Platform);
            Assert.AreEqual(fontSource, info.FontSource);
            Assert.IsNull(info.Encoding);
            Assert.IsNull(info.AppendingFirstByte);
            Assert.AreEqual(table, info.CMap);
        }

        [TestMethod]
        public void TestTrueTypeFont_GetGlyphInitializationInfoIsMicrosoftSymbolicBecauseOfEncodingDifferences()
        {
            TestOpenTypeFontSource fontSource = new TestOpenTypeFontSource();
            TestFont font = new TestFont("test", fontSource, fontSource);
            SimpleEncoding encoding = new SimpleEncoding();
            encoding.PredefinedEncoding.Value = PredefinedSimpleEncoding.WinAnsiEncodingName;
            encoding.Differences.Value = new DifferencesRanges();
            TestCMap map = (TestCMap)fontSource.CMap;
            TestCMapTable table31 = new TestCMapTable(31);
            TestCMapTable table10 = new TestCMapTable(10);
            TestCMapTable table30 = new TestCMapTable(1001);
            map.AddCMapTable(3, 1, table31);
            map.AddCMapTable(1, 0, table10);
            map.AddCMapTable(3, 0, table30);
            TrueTypeGlyphInitializationInfo info = TrueTypeFont.GetGlyphInitializationInfo(fontSource, font, encoding);

            Assert.IsNotNull(info);
            Assert.AreEqual(TrueTypeGlyphInitializationInfo.PlatformType.MicrosoftSymbolic, info.Platform);
            Assert.AreEqual(fontSource, info.FontSource);
            Assert.AreEqual(encoding, info.Encoding);
            Assert.AreEqual((byte)3, info.AppendingFirstByte);
            Assert.AreEqual(table30, info.CMap);
        }

        [TestMethod]
        public void TestTrueTypeFont_GetGlyphInitializationInfoIsMacintoshSymbolicBecauseOfEncodingDifferences()
        {
            TestOpenTypeFontSource fontSource = new TestOpenTypeFontSource();
            TestFont font = new TestFont("test", fontSource, fontSource);
            font.IsNonSymbolic = false;
            font.IsSymbolic = true;
            SimpleEncoding encoding = new SimpleEncoding();
            encoding.PredefinedEncoding.Value = PredefinedSimpleEncoding.MacRomanEncodingName;
            encoding.Differences.Value = new DifferencesRanges();
            TestCMap map = (TestCMap)fontSource.CMap;
            TestCMapTable table = new TestCMapTable(2);
            map.AddCMapTable(1, 0, table);
            TrueTypeGlyphInitializationInfo info = TrueTypeFont.GetGlyphInitializationInfo(fontSource, font, encoding);

            Assert.IsNotNull(info);
            Assert.AreEqual(TrueTypeGlyphInitializationInfo.PlatformType.MacintoshSymbolic, info.Platform);
            Assert.AreEqual(fontSource, info.FontSource);
            Assert.AreEqual(encoding, info.Encoding);
            Assert.IsNull(info.AppendingFirstByte);
            Assert.AreEqual(table, info.CMap);
        }

        [TestMethod]
        public void TestTrueTypeFont_InitializeCoreGlyph_MicrosoftNonSymbolic()
        {
            TestOpenTypeFontSource fontSource = new TestOpenTypeFontSource();
            TrueTypeFont font = new TrueTypeFont("test", fontSource);
            font.Encoding.Value = new SimpleEncoding();
            font.Encoding.Value.PredefinedEncoding.Value = PredefinedSimpleEncoding.WinAnsiEncodingName;

            int unicode = '\u0201';
            ushort expectedGlyphId = 3456;
            TestCMapTable table = new TestCMapTable(10);
            table.AddGlyphIdMapping(unicode, expectedGlyphId);
            TestCMap map = (TestCMap)fontSource.CMap;
            map.AddCMapTable(3, 1, table);

            Glyph g = new Glyph();
            font.InitializeCoreGlyph(g, new CharCode(135));

            Assert.AreEqual("adblgrave", g.Name);
            Assert.AreEqual(expectedGlyphId, g.GlyphId);
        }

        [TestMethod]
        public void TestTrueTypeFont_InitializeCoreGlyph_MacintoshNonSymbolic()
        {
            TestOpenTypeFontSource fontSource = new TestOpenTypeFontSource();
            TrueTypeFont font = new TrueTypeFont("test", fontSource);
            font.Encoding.Value = new SimpleEncoding();
            font.Encoding.Value.PredefinedEncoding.Value = PredefinedSimpleEncoding.WinAnsiEncodingName;

            ushort expectedGlyphId = 4567;
            TestCMapTable table = new TestCMapTable(15);
            table.AddGlyphIdMapping(224, expectedGlyphId);
            TestCMap map = (TestCMap)fontSource.CMap;
            map.AddCMapTable(1, 0, table);

            Glyph g = new Glyph();
            font.InitializeCoreGlyph(g, new CharCode(135));

            Assert.AreEqual("adblgrave", g.Name);
            Assert.AreEqual(expectedGlyphId, g.GlyphId);
        }

        [TestMethod]
        public void TestTrueTypeFont_InitializeCoreGlyph_MicrosoftSymbolic()
        {
            TestOpenTypeFontSource fontSource = new TestOpenTypeFontSource();
            TrueTypeFont font = new TrueTypeFont("test", fontSource);
            font.Encoding.Value = new SimpleEncoding();
            font.Encoding.Value.PredefinedEncoding.Value = PredefinedSimpleEncoding.StandardEncodingName;
            font.IsSymbolic = true;

            ushort expectedGlyphId = 12345;
            TestCMapTable table = new TestCMapTable(1001);
            table.AddGlyphIdMapping(903, expectedGlyphId);
            TestCMap map = (TestCMap)fontSource.CMap;
            map.AddCMapTable(3, 0, table);

            Glyph g = new Glyph();
            font.InitializeCoreGlyph(g, new CharCode(135));

            Assert.IsNull(g.Name);
            Assert.AreEqual(expectedGlyphId, g.GlyphId);
        }

        [TestMethod]
        public void TestTrueTypeFont_InitializeCoreGlyph_MacintoshSymbolic()
        {
            TestOpenTypeFontSource fontSource = new TestOpenTypeFontSource();
            TrueTypeFont font = new TrueTypeFont("test", fontSource);

            ushort expectedGlyphId = 54321;
            TestCMapTable table = new TestCMapTable(5);
            table.AddGlyphIdMapping(135, expectedGlyphId);
            TestCMap map = (TestCMap)fontSource.CMap;
            map.AddCMapTable(1, 0, table);

            Glyph g = new Glyph();
            font.InitializeCoreGlyph(g, new CharCode(135));

            Assert.IsNull(g.Name);
            Assert.AreEqual(expectedGlyphId, g.GlyphId);
        }
    }
}
