﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Actions;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Common;
using Telerik.Windows.Documents.Fixed.Model.Navigation;

namespace Telerik.Windows.Documents.Fixed.Tests.Model.Navigation
{
    [TestClass]
    public class BookmarkItemTests
    {
        [TestMethod]
        public void Constructor_ShouldSetNullParent()
        {
            // Act
            BookmarkItem bookmark = new BookmarkItem(string.Empty);

            // Assert
            Assert.IsNull(bookmark.Parent);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Constructor_WhenTitleIsNull_ShouldThrow()
        {
            // Act && Assert
            BookmarkItem bookmark = new BookmarkItem(null);
        }

        [TestMethod]
        public void Constructor_WhenTitleIsNotNull_ShouldSetCorrectTitle()
        {
            // Arrange
            string title = "Title";

            // Act
            BookmarkItem bookmark = new BookmarkItem(title);

            // Assert
            Assert.AreEqual(title, bookmark.Title);
        }

        [TestMethod]
        public void Constructor_ShouldSetIsExpandedToFalse()
        {
            // Act
            BookmarkItem bookmark = new BookmarkItem(string.Empty);

            // Assert
            Assert.AreEqual(false, bookmark.IsExpanded);
        }

        [TestMethod]
        public void Constructor_ShouldSetTextStyleToNormal()
        {
            // Act
            BookmarkItem bookmark = new BookmarkItem(string.Empty);

            // Assert
            Assert.AreEqual(BookmarkItemStyles.Normal, bookmark.TextStyle);
        }

        [TestMethod]
        public void Constructor_ShouldSetTextColorToBlack()
        {
            // Act
            BookmarkItem bookmark = new BookmarkItem(string.Empty);

            // Assert
            Assert.AreEqual(RgbColors.Black, bookmark.TextColor);
        }

        [TestMethod]
        public void Constructor_ShouldSetEmptyChildrenCollection()
        {
            // Act
            BookmarkItem bookmark = new BookmarkItem(string.Empty);

            // Assert
            Assert.IsNotNull(bookmark.Children);
            Assert.AreEqual(0, bookmark.Children.Count);
        }

        [TestMethod]
        public void Constructor_ShouldSetChildrenCollectionWithParentSetToTheBookmarkItemInstance()
        {
            // Act
            BookmarkItem bookmark = new BookmarkItem(string.Empty);

            // Assert
            Assert.IsNotNull(bookmark.Children);
            Assert.AreEqual(bookmark, ((IFixedDocumentElement)bookmark.Children).Parent);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Title_WhenNull_ShouldThrow()
        {
            // Arrange
            BookmarkItem bookmark = new BookmarkItem(string.Empty);

            // Act && Assert
            bookmark.Title = null;
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TextColor_WhenNull_ShouldThrow()
        {
            // Arrange
            BookmarkItem bookmark = new BookmarkItem(string.Empty);

            // Act && Assert
            bookmark.TextColor = null;
        }

        [TestMethod]
        public void ElementType_ShouldReturnCorrectType()
        {
            // Arrange
            BookmarkItem bookmark = new BookmarkItem(string.Empty);

            // Act
            FixedDocumentElementType elementType = bookmark.ElementType;

            // Assert
            Assert.AreEqual(FixedDocumentElementType.BookmarkItem, elementType);
        }

        [TestMethod]
        public void Clone_WhenHasChildren_ShouldCloneWithCorrectBookmarksHierarchy()
        {
            // Arrange
            RadFixedDocumentCloneContext cloneContext = new RadFixedDocumentCloneContext();

            string bookmarkTitle = "Bookmark Title";
            BookmarkItem bookmark = new BookmarkItem(bookmarkTitle);
            bookmark.IsExpanded = true;

            string childBookmarkTitle = "Child Bookmark";
            bookmark.Children.Add(new BookmarkItem(childBookmarkTitle));

            // Act
            BookmarkItem clonedBookmark = ((IContextClonable<BookmarkItem>)bookmark).Clone(cloneContext);

            // Assert
            Assert.AreEqual(bookmarkTitle, clonedBookmark.Title);
            Assert.IsTrue(clonedBookmark.IsExpanded);
            Assert.AreEqual(1, clonedBookmark.Children.Count);

            BookmarkItem clonedChildBookmark = clonedBookmark.Children[0];
            Assert.AreEqual(childBookmarkTitle, clonedChildBookmark.Title);
            Assert.IsFalse(clonedChildBookmark.IsExpanded);
            Assert.AreEqual(0, clonedChildBookmark.Children.Count);
        }

        [TestMethod]
        public void Clone_WhenHasTextStyle_ShouldCloneWithCorrectTextStyle()
        {
            // Arrange
            RadFixedDocumentCloneContext cloneContext = new RadFixedDocumentCloneContext();

            BookmarkItem bookmark = new BookmarkItem(string.Empty);
            bookmark.TextStyle = BookmarkItemStyles.Bold | BookmarkItemStyles.Italic;

            // Act
            BookmarkItem clonedBookmark = ((IContextClonable<BookmarkItem>)bookmark).Clone(cloneContext);

            // Assert
            Assert.AreEqual(BookmarkItemStyles.Bold | BookmarkItemStyles.Italic, clonedBookmark.TextStyle);
        }

        [TestMethod]
        public void Clone_WhenHasTextColor_ShouldCloneWithCorrectTextColor()
        {
            // Arrange
            RadFixedDocumentCloneContext cloneContext = new RadFixedDocumentCloneContext();

            BookmarkItem bookmark = new BookmarkItem(string.Empty);
            bookmark.TextColor = new RgbColor(255, 0, 128);

            // Act
            BookmarkItem clonedBookmark = ((IContextClonable<BookmarkItem>)bookmark).Clone(cloneContext);

            // Assert
            Assert.AreEqual(new RgbColor(255, 0, 128), clonedBookmark.TextColor);
        }

        [TestMethod]
        public void Clone_WhenHasDestination_ShouldCloneWithCorrectDestination()
        {
            // Arrange
            RadFixedDocumentCloneContext cloneContext = new RadFixedDocumentCloneContext();


            Location location = new Location()
            {
                Left = 20,
                Top = 50,
                Zoom = 30
            };

            BookmarkItem bookmark = new BookmarkItem(string.Empty, location);

            // Act
            BookmarkItem clonedBookmark = ((IContextClonable<BookmarkItem>)bookmark).Clone(cloneContext);

            // Assert
            Assert.IsNotNull(clonedBookmark.Destination);
            Assert.AreEqual(DestinationType.Location, clonedBookmark.Destination.DestinationType);

            Location destination = clonedBookmark.Destination as Location;
            Assert.IsNotNull(destination);

            Assert.AreEqual(20, destination.Left);
            Assert.AreEqual(50, destination.Top);
            Assert.AreEqual(30, destination.Zoom);
        }

        [TestMethod]
        public void Clone_WhenHasAction_ShouldCloneWithCorrectAction()
        {
            // Arrange
            RadFixedDocumentCloneContext cloneContext = new RadFixedDocumentCloneContext();

            Location location = new Location()
            {
                Left = 20,
                Top = 50,
                Zoom = 30
            };

            GoToAction goToAction = new GoToAction(location);
            BookmarkItem bookmark = new BookmarkItem(string.Empty, goToAction);

            // Act
            BookmarkItem clonedBookmark = ((IContextClonable<BookmarkItem>)bookmark).Clone(cloneContext);

            // Assert
            Assert.IsNotNull(clonedBookmark.Action);
            Assert.AreEqual(ActionType.GoTo, clonedBookmark.Action.ActionType);

            GoToAction action = clonedBookmark.Action as GoToAction;
            Assert.IsNotNull(action);

            Assert.IsNotNull(action.Destination);
            Assert.AreEqual(DestinationType.Location, action.Destination.DestinationType);

            Location destination = action.Destination as Location;
            Assert.IsNotNull(destination);

            Assert.AreEqual(20, destination.Left);
            Assert.AreEqual(50, destination.Top);
            Assert.AreEqual(30, destination.Zoom);
        }
    }
}
