﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Linq;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Actions;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Navigation;

namespace Telerik.Windows.Documents.Fixed.Tests.Model
{
    [TestClass]
    public class RadFixedDocumentTests
    {
        [TestMethod]
        public void Merge_WhenDocumentsWithBookmarks_ShouldMergeBookmarksWithCorrectBookmarksHierarchy()
        {
            // Arrange
            RadFixedDocument mergedDocument = new RadFixedDocument();
            string firstBookmarkTitle = "First Bookmark";
            mergedDocument.Bookmarks.Add(new BookmarkItem(firstBookmarkTitle));

            RadFixedDocument sourceDocument = new RadFixedDocument();
            string secondBookmarkTitle = "Second Bookmark";
            BookmarkItem secondBookmark = new BookmarkItem(secondBookmarkTitle);
            secondBookmark.IsExpanded = true;
            sourceDocument.Bookmarks.Add(secondBookmark);

            string childBookmarkTitle = "Child Bookmark";
            secondBookmark.Children.Add(new BookmarkItem(childBookmarkTitle));

            // Act
            mergedDocument.Merge(sourceDocument);

            // Assert
            Assert.AreEqual(2, mergedDocument.Bookmarks.Count);

            BookmarkItem bookmark1 = mergedDocument.Bookmarks[0];
            Assert.AreEqual(firstBookmarkTitle, bookmark1.Title);
            Assert.IsFalse(bookmark1.IsExpanded);
            Assert.AreEqual(0, bookmark1.Children.Count);

            BookmarkItem bookmark2 = mergedDocument.Bookmarks[1];
            Assert.AreEqual(secondBookmarkTitle, bookmark2.Title);
            Assert.IsTrue(bookmark2.IsExpanded);
            Assert.AreEqual(1, bookmark2.Children.Count);

            BookmarkItem childBookmark = bookmark2.Children[0];
            Assert.AreEqual(childBookmarkTitle, childBookmark.Title);
            Assert.IsFalse(childBookmark.IsExpanded);
            Assert.AreEqual(0, childBookmark.Children.Count);
        }

        [TestMethod]
        public void Merge_WhenDocumentsWithBookmarks_ShouldMergeBookmarksWithCorrectTextStyle()
        {
            // Arrange
            RadFixedDocument mergedDocument = new RadFixedDocument();
            mergedDocument.Bookmarks.Add(new BookmarkItem(string.Empty));

            RadFixedDocument sourceDocument = new RadFixedDocument();
            BookmarkItem secondBookmark = new BookmarkItem(string.Empty);
            secondBookmark.TextStyle = BookmarkItemStyles.Bold | BookmarkItemStyles.Italic;
            sourceDocument.Bookmarks.Add(secondBookmark);

            // Act
            mergedDocument.Merge(sourceDocument);

            // Assert
            BookmarkItem mergedBookmark = mergedDocument.Bookmarks[1];
            Assert.AreEqual(BookmarkItemStyles.Bold | BookmarkItemStyles.Italic, mergedBookmark.TextStyle);
        }

        [TestMethod]
        public void Merge_WhenDocumentsWithBookmarks_ShouldMergeBookmarksWithCorrectTextColor()
        {
            // Arrange
            RadFixedDocument mergedDocument = new RadFixedDocument();
            mergedDocument.Bookmarks.Add(new BookmarkItem(string.Empty));

            RadFixedDocument sourceDocument = new RadFixedDocument();
            BookmarkItem secondBookmark = new BookmarkItem(string.Empty);
            secondBookmark.TextColor = new RgbColor(255, 0, 128);
            sourceDocument.Bookmarks.Add(secondBookmark);

            // Act
            mergedDocument.Merge(sourceDocument);

            // Assert
            BookmarkItem mergedBookmark = mergedDocument.Bookmarks[1];
            Assert.AreEqual(new RgbColor(255, 0, 128), mergedBookmark.TextColor);
        }

        [TestMethod]
        public void Merge_WhenDocumentsWithBookmarks_ShouldMergeBookmarksWithCorrectDestination()
        {
            // Arrange
            RadFixedDocument mergedDocument = new RadFixedDocument();
            mergedDocument.Bookmarks.Add(new BookmarkItem(string.Empty));

            RadFixedDocument sourceDocument = new RadFixedDocument();

            Location location = new Location()
            {
                Left = 20,
                Top = 50,
                Zoom = 30
            };

            BookmarkItem secondBookmark = new BookmarkItem(string.Empty, location);
            sourceDocument.Bookmarks.Add(secondBookmark);

            // Act
            mergedDocument.Merge(sourceDocument);

            // Assert
            BookmarkItem mergedBookmark = mergedDocument.Bookmarks[1];
            Assert.IsNotNull(mergedBookmark.Destination);
            Assert.AreEqual(DestinationType.Location, mergedBookmark.Destination.DestinationType);

            Location destination = mergedBookmark.Destination as Location;
            Assert.IsNotNull(destination);

            Assert.AreEqual(20, destination.Left);
            Assert.AreEqual(50, destination.Top);
            Assert.AreEqual(30, destination.Zoom);
        }

        [TestMethod]
        public void Merge_WhenDocumentsWithBookmarks_ShouldMergeBookmarksWithCorrectAction()
        {
            // Arrange
            RadFixedDocument mergedDocument = new RadFixedDocument();
            mergedDocument.Bookmarks.Add(new BookmarkItem(string.Empty));

            RadFixedDocument sourceDocument = new RadFixedDocument();

            Location location = new Location()
            {
                Left = 20,
                Top = 50,
                Zoom = 30
            };

            GoToAction goToAction = new GoToAction(location);
            BookmarkItem secondBookmark = new BookmarkItem(string.Empty, goToAction);

            sourceDocument.Bookmarks.Add(secondBookmark);

            // Act
            mergedDocument.Merge(sourceDocument);

            // Assert
            BookmarkItem mergedBookmark = mergedDocument.Bookmarks[1];
            Assert.IsNotNull(mergedBookmark.Action);
            Assert.AreEqual(ActionType.GoTo, mergedBookmark.Action.ActionType);

            GoToAction action = mergedBookmark.Action as GoToAction;
            Assert.IsNotNull(action);

            Assert.IsNotNull(action.Destination);
            Assert.AreEqual(DestinationType.Location, action.Destination.DestinationType);

            Location destination = action.Destination as Location;
            Assert.IsNotNull(destination);

            Assert.AreEqual(20, destination.Left);
            Assert.AreEqual(50, destination.Top);
            Assert.AreEqual(30, destination.Zoom);
        }

        [TestMethod]
        public void Merge_WhenDocumentsWithNamedDestinations_ShouldRenameTheNamedDestinationsWithSameNames()
        {
            // Arrange
            RadFixedDocument mergedDocument = new RadFixedDocument();
            Location mergedLocation = new Location()
            {
                Left = 40,
                Top = 30,
                Zoom = 10
            };

            NamedDestination mergedNamedLocation = mergedDocument.NamedDestinations.Add("Location", mergedLocation);

            GoToAction mergedGoToAction = new GoToAction(mergedNamedLocation);

            NamedDestination mergedNamedAction = mergedDocument.NamedDestinations.Add("GoToAction", mergedGoToAction);

            BookmarkItem firstBookmark = new BookmarkItem("Bookmark", mergedNamedAction);
            mergedDocument.Bookmarks.Add(firstBookmark);


            RadFixedDocument sourceDocument = new RadFixedDocument();

            Location location = new Location()
            {
                Left = 20,
                Top = 50,
                Zoom = 30
            };

            NamedDestination namedLocation = sourceDocument.NamedDestinations.Add("Location", location);

            GoToAction goToAction = new GoToAction(namedLocation);

            NamedDestination namedAction = sourceDocument.NamedDestinations.Add("GoToAction", goToAction);

            BookmarkItem secondBookmark = new BookmarkItem("Bookmark", namedAction);

            sourceDocument.Bookmarks.Add(secondBookmark);

            // Act
            mergedDocument.Merge(sourceDocument);

            // Assert

            Assert.AreEqual(4, mergedDocument.NamedDestinations.Count);

            Assert.AreEqual("Location", mergedDocument.NamedDestinations.First().Name);
            Assert.AreEqual("GoToAction", mergedDocument.NamedDestinations.Skip(1).First().Name);
            Assert.AreEqual("Location.0", mergedDocument.NamedDestinations.Skip(2).First().Name);
            Assert.AreEqual("GoToAction.0", mergedDocument.NamedDestinations.Skip(3).First().Name);

            BookmarkItem mergedBookmark = mergedDocument.Bookmarks[1];
            Assert.IsNull(mergedBookmark.Action);
            Assert.IsNotNull(mergedBookmark.NamedDestination);

            GoToAction action = mergedBookmark.NamedDestination.GoToAction;
            Assert.IsNotNull(action);

            Assert.IsNull(action.Destination);
            Assert.IsNotNull(action.NamedDestination);
            Assert.AreEqual(DestinationType.Location, action.NamedDestination.Destination.DestinationType);

            Location destination = action.NamedDestination.Destination as Location;
            Assert.IsNotNull(destination);

            Assert.AreEqual(20, destination.Left);
            Assert.AreEqual(50, destination.Top);
            Assert.AreEqual(30, destination.Zoom);
        }

        [TestMethod]
        public void Merge_WhenDocumentsWithLinksWithNamedDestinations_ShouldContainLinksWithNamedDestinations()
        {
            // Arrange
            RadFixedDocument mergedDocument = new RadFixedDocument();
            Location mergedLocation = new Location()
            {
                Left = 40,
                Top = 30,
                Zoom = 10
            };

            NamedDestination mergedNamedLocation = mergedDocument.NamedDestinations.Add("Location", mergedLocation);

            GoToAction mergedGoToAction = new GoToAction(mergedNamedLocation);

            NamedDestination mergedNamedAction = mergedDocument.NamedDestinations.Add("GoToAction", mergedGoToAction);

            RadFixedPage page = mergedDocument.Pages.AddPage();
            page.Annotations.AddLink(mergedNamedAction);


            RadFixedDocument sourceDocument = new RadFixedDocument();

            Location location = new Location()
            {
                Left = 40,
                Top = 30,
                Zoom = 10
            };

            NamedDestination namedLocation = sourceDocument.NamedDestinations.Add("Location", location);

            GoToAction goToAction = new GoToAction(namedLocation);

            NamedDestination namedAction = sourceDocument.NamedDestinations.Add("GoToAction", goToAction);

            var sourcePage = sourceDocument.Pages.AddPage();

            sourcePage.Annotations.AddLink(namedAction);

            // Act
            mergedDocument.Merge(sourceDocument);

            // Assert

            Assert.AreEqual(4, mergedDocument.NamedDestinations.Count);

            Assert.AreEqual("Location", mergedDocument.NamedDestinations.First().Name);
            Assert.AreEqual("GoToAction", mergedDocument.NamedDestinations.Skip(1).First().Name);
            Assert.AreEqual("Location.0", mergedDocument.NamedDestinations.Skip(2).First().Name);
            Assert.AreEqual("GoToAction.0", mergedDocument.NamedDestinations.Skip(3).First().Name);

            Assert.AreEqual(2, mergedDocument.Annotations.Count());

            Link link = mergedDocument.Annotations.First() as Link;
            Assert.IsNotNull(link);
            Assert.IsNotNull(link.NamedDestination);
            Assert.AreEqual("GoToAction", link.NamedDestination.Name);

            GoToAction action = link.NamedDestination.GoToAction;
            Assert.IsNotNull(action);

            Assert.IsNull(action.Destination);
            Assert.IsNotNull(action.NamedDestination);
            Assert.AreEqual("Location", action.NamedDestination.Name);
            Assert.AreEqual(DestinationType.Location, action.NamedDestination.Destination.DestinationType);

            Location destination = action.NamedDestination.Destination as Location;
            Assert.IsNotNull(destination);

            Assert.AreEqual(40, destination.Left);
            Assert.AreEqual(30, destination.Top);
            Assert.AreEqual(10, destination.Zoom);

            link = mergedDocument.Annotations.Skip(1).First() as Link;
            Assert.IsNotNull(link);
            Assert.IsNotNull(link.NamedDestination);
            Assert.AreEqual("GoToAction.0", link.NamedDestination.Name);

            action = link.NamedDestination.GoToAction;
            Assert.IsNotNull(action);

            Assert.IsNull(action.Destination);
            Assert.IsNotNull(action.NamedDestination);
            Assert.AreEqual("Location.0", action.NamedDestination.Name);
            Assert.AreEqual(DestinationType.Location, action.NamedDestination.Destination.DestinationType);

            destination = action.NamedDestination.Destination as Location;
            Assert.IsNotNull(destination);

            Assert.AreEqual(40, destination.Left);
            Assert.AreEqual(30, destination.Top);
            Assert.AreEqual(10, destination.Zoom);
        }
    }
}
