﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Core.PostScript.Data;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.PdfReader.Readers;
using System.IO;
using Telerik.Windows.Documents.Fixed.FormatProviders.Old.Pdf.Text;

namespace Telerik.Windows.Documents.Fixed.Tests.PdfReaderTests
{
    [TestClass]
    public class TestPdfReader
    {
        private Stream CreateStreamFromString(string text)
        {
            MemoryStream ms = new MemoryStream(System.Text.Encoding.UTF8.GetBytes(text));
            return ms;
        }

        private string GetString(byte[] bytes)
        {
            return System.Text.Encoding.UTF8.GetString(bytes, 0, bytes.Length);
        }

        private string GetBigEndianString(byte[] bytes)
        {
            return System.Text.Encoding.BigEndianUnicode.GetString(bytes, 0, bytes.Length);
        }

        [TestMethod]
        public void TestReadBool()
        {
            string text = "true false";
            using (Stream s = this.CreateStreamFromString(text))
            {
                PdfReader reader = new PdfReader(s);
                Assert.AreEqual(Token.Boolean, reader.ReadToken());
                Assert.AreEqual("true", reader.Result);
                Assert.AreEqual(Token.Boolean, reader.ReadToken());
                Assert.AreEqual("false", reader.Result);
            }
        }

        [TestMethod]
        public void TestReadInt()
        {
            string text = "123 43445 +17 -98 0 0 0 0 0 0 34";
            using (Stream s = this.CreateStreamFromString(text))
            {
                PdfReader reader = new PdfReader(s);
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("123", reader.Result);
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("43445", reader.Result);
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("+17", reader.Result);
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("-98", reader.Result);
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("0", reader.Result);
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("0", reader.Result);
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("0", reader.Result);
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("0", reader.Result);
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("0", reader.Result);
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("0", reader.Result);
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("34", reader.Result);
            }
        }

        [TestMethod]
        public void TestReadReal()
        {
            string text = "34.5 -3.62 +123.6 4. -.002 0.0";
            using (Stream s = this.CreateStreamFromString(text))
            {
                PdfReader reader = new PdfReader(s);
                Assert.AreEqual(Token.Real, reader.ReadToken());
                Assert.AreEqual("34.5", reader.Result);
                Assert.AreEqual(Token.Real, reader.ReadToken());
                Assert.AreEqual("-3.62", reader.Result);
                Assert.AreEqual(Token.Real, reader.ReadToken());
                Assert.AreEqual("+123.6", reader.Result);
                Assert.AreEqual(Token.Real, reader.ReadToken());
                Assert.AreEqual("4.", reader.Result);
                Assert.AreEqual(Token.Real, reader.ReadToken());
                Assert.AreEqual("-.002", reader.Result);
                Assert.AreEqual(Token.Real, reader.ReadToken());
                Assert.AreEqual("0.0", reader.Result);
            }
        }

        [TestMethod]
        public void TestReadLiteralString()
        {
            string text = 
@"(This is a string)
(Strings may contain newlines\r\nand such.)
(Strings may contain balanced parentheses () and special characters (*!&}^% and so on).)
()";

            using (Stream s = this.CreateStreamFromString(text))
            {
                PdfReader reader = new PdfReader(s);
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual("This is a string", this.GetString(reader.BytesToken));
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual("Strings may contain newlines\r\nand such.", this.GetString(reader.BytesToken));
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual("Strings may contain balanced parentheses () and special characters (*!&}^% and so on).", this.GetString(reader.BytesToken));
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual("", this.GetString(reader.BytesToken));
            }
        }

        [TestMethod]
        public void TestBigEndianLiteralString()
        {
            string text =
"(\\000D\\000a\\000t\\000e\\000 \\000\\(\\000e\\000.\\000g\\000.\\000 \\0002\\0001\\000/\\0003\\000/\\0002\\0000\\0000\\0006\\000\\))";

            using (Stream s = this.CreateStreamFromString(text))
            {
                PdfReader reader = new PdfReader(s);
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual("Date (e.g. 21/3/2006)", this.GetBigEndianString(reader.BytesToken));
            }
        }

        [TestMethod]
        public void TestLiteralStringExcapeSequences()
        {
            string text = @"(\n\r\t\b\f\(\)\\)
(These \
two strings \
are the same.)
(A\053B)
(\053)
(\53)
";
            using (Stream s = this.CreateStreamFromString(text))
            {
                PdfReader reader = new PdfReader(s);
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual("\n\r\t\b\f()\\", this.GetString(reader.BytesToken));
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual("These two strings are the same.", this.GetString(reader.BytesToken));
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual("A+B", this.GetString(reader.BytesToken));
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual("+", this.GetString(reader.BytesToken));
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual("+", this.GetString(reader.BytesToken));
            }
        }

        [TestMethod]
        public void TestReadHexadecimalString()
        {
            string text = "<901FA3><901FA>";
            using (Stream s = this.CreateStreamFromString(text))
            {
                PdfReader reader = new PdfReader(s);
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual(0x90, reader.BytesToken[0]);
                Assert.AreEqual(0x1F, reader.BytesToken[1]);
                Assert.AreEqual(0xA3, reader.BytesToken[2]);
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual(0x90, reader.BytesToken[0]);
                Assert.AreEqual(0x1F, reader.BytesToken[1]);
                Assert.AreEqual(0xA0, reader.BytesToken[2]);
            }
        }

        [TestMethod]
        public void TestReadPdfName()
        {
            string text = @"
/A#20B
/Name1
/ASomewhatLongerName
/A;Name_With-Various***Characters?
/1.2
/$$
/@pattern
/.notdef
";
            using (Stream s = this.CreateStreamFromString(text))
            {
                PdfReader reader = new PdfReader(s);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("A B", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("Name1", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("ASomewhatLongerName", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("A;Name_With-Various***Characters?", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("1.2", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("$$", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("@pattern", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual(".notdef", reader.Result);
            }
        }

        public void TestPdfNameEscapeSequences()
        {
            string text = @"/paired#28#29parentheses/Adobe#20Green/A#42";
            using (Stream s = this.CreateStreamFromString(text))
            {
                PdfReader reader = new PdfReader(s);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("paired()parentheses", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("Adobe Green", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("AB", reader.Result);
            }
        }

        [TestMethod]
        public void TestReadPdfArray()
        {
            string text = @"[549 3.14 false (Ralph) /SomeName]";
            using (Stream s = this.CreateStreamFromString(text))
            {
                PdfReader reader = new PdfReader(s);
                Assert.AreEqual(Token.ArrayStart, reader.ReadToken());
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("549", reader.Result);
                Assert.AreEqual(Token.Real, reader.ReadToken());
                Assert.AreEqual("3.14", reader.Result);
                Assert.AreEqual(Token.Boolean, reader.ReadToken());
                Assert.AreEqual("false", reader.Result);
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual("Ralph", this.GetString(reader.BytesToken));
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("SomeName", reader.Result);
                Assert.AreEqual(Token.ArrayEnd, reader.ReadToken());
            }
        }

        [TestMethod]
        public void TestReadPdfDictionary()
        {
            string text = @"
<< 
/Type /Example
/Subtype /DictionaryExample
/Version 0.01
/IntegerItem 12
/StringItem (a string)
/SubDictionary 
<< 
/Item /Item
>>
>>
";
            using (Stream s = this.CreateStreamFromString(text))
            {
                PdfReader reader = new PdfReader(s);
                Assert.AreEqual(Token.DictionaryStart, reader.ReadToken());
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("Type", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("Example", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("Subtype", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("DictionaryExample", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("Version", reader.Result);
                Assert.AreEqual(Token.Real, reader.ReadToken());
                Assert.AreEqual("0.01", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("IntegerItem", reader.Result);
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("12", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("StringItem", reader.Result);
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual("a string", this.GetString(reader.BytesToken));
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("SubDictionary", reader.Result);
                Assert.AreEqual(Token.DictionaryStart, reader.ReadToken());
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("Item", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("Item", reader.Result);
                Assert.AreEqual(Token.DictionaryEnd, reader.ReadToken());
                Assert.AreEqual(Token.DictionaryEnd, reader.ReadToken());
            }
        }

        [TestMethod]
        public void TestReadPdfDataStream()
        {
            string text = @"
<< 
/Type /Example
/Subtype /DictionaryExample
>>
stream
endstream
";
            using (Stream s = this.CreateStreamFromString(text))
            {
                PdfReader reader = new PdfReader(s);
                Assert.AreEqual(Token.DictionaryStart, reader.ReadToken());
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("Type", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("Example", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("Subtype", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("DictionaryExample", reader.Result);
                Assert.AreEqual(Token.DictionaryEnd, reader.ReadToken());
                Assert.AreEqual(Token.StreamStart, reader.ReadToken());
                Assert.AreEqual(Token.StreamEnd, reader.ReadToken());
            }
        }

        [TestMethod]
        public void TestReadNullObject()
        {
            string text = "null";
            using (Stream s = this.CreateStreamFromString(text))
            {
                PdfReader reader = new PdfReader(s);
                Assert.AreEqual(Token.Null, reader.ReadToken());
                Assert.AreEqual("null", reader.Result);
            }
        }

        [TestMethod]
        public void TestReadIndirectObject()
        {
            string text = @"
12 0 obj
1234
endobj
";
            using (Stream s = this.CreateStreamFromString(text))
            {
                PdfReader reader = new PdfReader(s);
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("12", reader.Result);
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("0", reader.Result);
                Assert.AreEqual(Token.IndirectObjectStart, reader.ReadToken());
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("1234", reader.Result);
                Assert.AreEqual(Token.IndirectObjectEnd, reader.ReadToken());
            }
        }

        [TestMethod]
        public void TestReadInJunkAndComments()
        {
            string text =
"%some junk here\r\n" +
"%more comments blablabla \r\n" +
"0 %int\r\n" + 
"\r       1 %one more int\r\n" +
"\n  2 %and another int\r\n" +
"(Some text)    \r\r\r\n (And more)\r\n\r\n" +
"/Name1/Name2(string)3(string)null";
            using (Stream s = this.CreateStreamFromString(text))
            {
                PdfReader reader = new PdfReader(s);
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("0", reader.Result);
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("1", reader.Result);
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("2", reader.Result);
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual("Some text", this.GetString(reader.BytesToken));
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual("And more", this.GetString(reader.BytesToken));
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("Name1", reader.Result);
                Assert.AreEqual(Token.Name, reader.ReadToken());
                Assert.AreEqual("Name2", reader.Result);
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual("string", this.GetString(reader.BytesToken));
                Assert.AreEqual(Token.Integer, reader.ReadToken());
                Assert.AreEqual("3", reader.Result);
                Assert.AreEqual(Token.String, reader.ReadToken());
                Assert.AreEqual("string", this.GetString(reader.BytesToken));
                Assert.AreEqual(Token.Null, reader.ReadToken());
                Assert.AreEqual("null", reader.Result);
            }
        }
    }
}
