﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.FormatProviders.Pdf.Model.Types;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.PostScript.Parser.Operators.Utils;

namespace Telerik.Windows.Documents.Fixed.Tests.PostScript.Parser.Operators.Utils
{
    [TestClass]
    public class ToUnicodeRangeBuilderTests
    {
        [TestMethod]
        public void BuildToUnicodeMapping_WhenHasNoOperands_ShouldReturnEmptyMapping()
        {
            // Arrange
            TestCMapEncodingInterpreter interpreter = new TestCMapEncodingInterpreter();
            interpreter.Operands = new Fixed.FormatProviders.Pdf.Import.OperandsCollection();

            // Act
            IDictionary<CharCode, string> toUnicodeMapping = ToUnicodeRangeBuilder.BuildToUnicodeMapping(interpreter);

            // Assert
            Assert.AreEqual(0, toUnicodeMapping.Count);
        }

        [TestMethod]
        [ExpectedException(typeof(NotSupportedException))]
        public void BuildToUnicodeMapping_WhenRangeOperandIsNotPdfStringOrArray_ShouldThrow()
        {
            // Arrange
            TestCMapEncodingInterpreter interpreter = new TestCMapEncodingInterpreter();
            Fixed.FormatProviders.Pdf.Import.OperandsCollection operands = new Fixed.FormatProviders.Pdf.Import.OperandsCollection();
            operands.AddLast(new PdfHexString("a", StringType.ASCII));
            operands.AddLast(new PdfHexString("b", StringType.ASCII));
            operands.AddLast(new PdfInt(-1));
            interpreter.Operands = operands;

            // Act && Assert
            IDictionary<CharCode, string> toUnicodeMapping = ToUnicodeRangeBuilder.BuildToUnicodeMapping(interpreter);
        }

        [TestMethod]
        public void BuildToUnicodeMapping_WhenRangeOperandIsString_ShouldBuildCorrenctMapping()
        {
            // Arrange
            TestCMapEncodingInterpreter interpreter = new TestCMapEncodingInterpreter();
            Fixed.FormatProviders.Pdf.Import.OperandsCollection operands = new Fixed.FormatProviders.Pdf.Import.OperandsCollection();
            operands.AddLast(new PdfHexString("a", StringType.ASCII));
            operands.AddLast(new PdfHexString("b", StringType.ASCII));
            operands.AddLast(new PdfHexString("C", StringType.UTFEncodedTextString));
            interpreter.Operands = operands;

            // Act
            IDictionary<CharCode, string> toUnicodeMapping = ToUnicodeRangeBuilder.BuildToUnicodeMapping(interpreter);

            // Assert
            Assert.AreEqual(2, toUnicodeMapping.Count);

            string firstUnicode;
            Assert.IsTrue(toUnicodeMapping.TryGetValue(new CharCode(97, 1), out firstUnicode));
            Assert.AreEqual("C", firstUnicode);

            string secondUnicode;
            Assert.IsTrue(toUnicodeMapping.TryGetValue(new CharCode(98, 1), out secondUnicode));
            Assert.AreEqual("D", secondUnicode);
        }

        [TestMethod]
        public void BuildToUnicodeMapping_WhenManyStringRanges_ShouldBuildCorrenctMapping()
        {
            // Arrange
            TestCMapEncodingInterpreter interpreter = new TestCMapEncodingInterpreter();
            Fixed.FormatProviders.Pdf.Import.OperandsCollection operands = new Fixed.FormatProviders.Pdf.Import.OperandsCollection();
            operands.AddLast(new PdfHexString("a", StringType.ASCII));
            operands.AddLast(new PdfHexString("b", StringType.ASCII));
            operands.AddLast(new PdfHexString("C", StringType.UTFEncodedTextString));

            operands.AddLast(new PdfHexString("d", StringType.ASCII));
            operands.AddLast(new PdfHexString("e", StringType.ASCII));
            operands.AddLast(new PdfHexString("M", StringType.UTFEncodedTextString));

            interpreter.Operands = operands;

            // Act
            IDictionary<CharCode, string> toUnicodeMapping = ToUnicodeRangeBuilder.BuildToUnicodeMapping(interpreter);

            // Assert
            Assert.AreEqual(4, toUnicodeMapping.Count);

            string firstUnicode;
            Assert.IsTrue(toUnicodeMapping.TryGetValue(new CharCode(97, 1), out firstUnicode));
            Assert.AreEqual("C", firstUnicode);

            string secondUnicode;
            Assert.IsTrue(toUnicodeMapping.TryGetValue(new CharCode(98, 1), out secondUnicode));
            Assert.AreEqual("D", secondUnicode);

            string thirdUnicode;
            Assert.IsTrue(toUnicodeMapping.TryGetValue(new CharCode(100, 1), out thirdUnicode));
            Assert.AreEqual("M", thirdUnicode);

            string fourthUnicode;
            Assert.IsTrue(toUnicodeMapping.TryGetValue(new CharCode(101, 1), out fourthUnicode));
            Assert.AreEqual("N", fourthUnicode);
        }

        [TestMethod]
        public void BuildToUnicodeMapping_WhenRangeOperandIsArray_ShouldBuildCorrenctMapping()
        {
            // Arrange
            TestCMapEncodingInterpreter interpreter = new TestCMapEncodingInterpreter();
            Fixed.FormatProviders.Pdf.Import.OperandsCollection operands = new Fixed.FormatProviders.Pdf.Import.OperandsCollection();

            operands.AddLast(new PdfHexString("a", StringType.ASCII));
            operands.AddLast(new PdfHexString("b", StringType.ASCII));
            PdfArray array = new PdfArray(new PdfHexString("E", StringType.UTFEncodedTextString), new PdfHexString("X", StringType.UTFEncodedTextString));
            operands.AddLast(array);

            interpreter.Operands = operands;

            // Act
            IDictionary<CharCode, string> toUnicodeMapping = ToUnicodeRangeBuilder.BuildToUnicodeMapping(interpreter);

            // Assert
            Assert.AreEqual(2, toUnicodeMapping.Count);

            string firstUnicode;
            Assert.IsTrue(toUnicodeMapping.TryGetValue(new CharCode(97, 1), out firstUnicode));
            Assert.AreEqual("E", firstUnicode);

            string secondUnicode;
            Assert.IsTrue(toUnicodeMapping.TryGetValue(new CharCode(98, 1), out secondUnicode));
            Assert.AreEqual("X", secondUnicode);
        }

        [TestMethod]
        public void BuildToUnicodeMapping_WhenManyArrayRanges_ShouldBuildCorrenctMapping()
        {
            // Arrange
            TestCMapEncodingInterpreter interpreter = new TestCMapEncodingInterpreter();
            Fixed.FormatProviders.Pdf.Import.OperandsCollection operands = new Fixed.FormatProviders.Pdf.Import.OperandsCollection();

            operands.AddLast(new PdfHexString("a", StringType.ASCII));
            operands.AddLast(new PdfHexString("b", StringType.ASCII));
            PdfArray array1 = new PdfArray(new PdfHexString("E", StringType.UTFEncodedTextString), new PdfHexString("X", StringType.UTFEncodedTextString));
            operands.AddLast(array1);

            operands.AddLast(new PdfHexString("d", StringType.ASCII));
            operands.AddLast(new PdfHexString("e", StringType.ASCII));
            PdfArray array2 = new PdfArray(new PdfHexString("W", StringType.UTFEncodedTextString), new PdfHexString("Z", StringType.UTFEncodedTextString));
            operands.AddLast(array2);

            interpreter.Operands = operands;

            // Act
            IDictionary<CharCode, string> toUnicodeMapping = ToUnicodeRangeBuilder.BuildToUnicodeMapping(interpreter);

            // Assert
            Assert.AreEqual(4, toUnicodeMapping.Count);

            string firstUnicode;
            Assert.IsTrue(toUnicodeMapping.TryGetValue(new CharCode(97, 1), out firstUnicode));
            Assert.AreEqual("E", firstUnicode);

            string secondUnicode;
            Assert.IsTrue(toUnicodeMapping.TryGetValue(new CharCode(98, 1), out secondUnicode));
            Assert.AreEqual("X", secondUnicode);

            string thirdUnicode;
            Assert.IsTrue(toUnicodeMapping.TryGetValue(new CharCode(100, 1), out thirdUnicode));
            Assert.AreEqual("W", thirdUnicode);

            string fourthUnicode;
            Assert.IsTrue(toUnicodeMapping.TryGetValue(new CharCode(101, 1), out fourthUnicode));
            Assert.AreEqual("Z", fourthUnicode);
        }

        [TestMethod]
        public void BuildToUnicodeMapping_ShouldBuildMappingInCorrectOrder()
        {
            // Arrange
            TestCMapEncodingInterpreter interpreter = new TestCMapEncodingInterpreter();
            Fixed.FormatProviders.Pdf.Import.OperandsCollection operands = new Fixed.FormatProviders.Pdf.Import.OperandsCollection();

            operands.AddLast(new PdfHexString("c", StringType.ASCII));
            operands.AddLast(new PdfHexString("d", StringType.ASCII));
            operands.AddLast(new PdfHexString("C", StringType.UTFEncodedTextString));
            interpreter.Operands = operands;

            operands.AddLast(new PdfHexString("a", StringType.ASCII));
            operands.AddLast(new PdfHexString("b", StringType.ASCII));
            PdfArray array = new PdfArray(new PdfHexString("b", StringType.UTFEncodedTextString), new PdfHexString("a", StringType.UTFEncodedTextString));
            operands.AddLast(array);

            // Act
            IDictionary<CharCode, string> toUnicodeMapping = ToUnicodeRangeBuilder.BuildToUnicodeMapping(interpreter);

            // Assert
            Assert.AreEqual(4, toUnicodeMapping.Count);

            Assert.AreEqual(97, toUnicodeMapping.Skip(0).First().Key.Code);
            Assert.AreEqual("b", toUnicodeMapping.Skip(0).First().Value);

            Assert.AreEqual(98, toUnicodeMapping.Skip(1).First().Key.Code);
            Assert.AreEqual("a", toUnicodeMapping.Skip(1).First().Value);

            Assert.AreEqual(99, toUnicodeMapping.Skip(2).First().Key.Code);
            Assert.AreEqual("C", toUnicodeMapping.Skip(2).First().Value);

            Assert.AreEqual(100, toUnicodeMapping.Skip(3).First().Key.Code);
            Assert.AreEqual("D", toUnicodeMapping.Skip(3).First().Value);
        }
    }
}
