﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Model.Fonts.Encoding;
using Telerik.Windows.Documents.Fixed.PostScript.Parser.Operators.Utils;

namespace Telerik.Windows.Documents.Fixed.Tests.PostScript.Parser.Operators.Utils
{
    [TestClass]
    public class ToUnicodeRangeParserTests
    {
        [TestMethod]
        public void ParseToUnicodeMapping_WhenMappingIsEmpty_ShouldDoNothing()
        {
            // Arrange
            TestCMapEncodingInterpreter interpreter = new TestCMapEncodingInterpreter();
            interpreter.Operands = new Fixed.FormatProviders.Pdf.Import.OperandsCollection();
            interpreter.Encoding = new CMapEncoding();
            SortedDictionary<CharCode, string> toUnicodeMapping = new SortedDictionary<CharCode, string>();

            // Act
            ToUnicodeRangeParser.ParseToUnicodeMapping(interpreter, toUnicodeMapping);

            // Assert
            Assert.AreEqual(0, interpreter.Encoding.CharCodeToUnicodeMappings.Count());
            Assert.AreEqual(0, interpreter.Encoding.UnicodeRanges.Count);
        }

        [TestMethod]
        public void ParseToUnicodeMapping_WhenHasRangeWithSequentialUnicodes_ShouldAddCorrectly()
        {
            // Arrange
            TestCMapEncodingInterpreter interpreter = new TestCMapEncodingInterpreter();
            interpreter.Operands = new Fixed.FormatProviders.Pdf.Import.OperandsCollection();
            interpreter.Encoding = new CMapEncoding();

            SortedDictionary<CharCode, string> toUnicodeMapping = new SortedDictionary<CharCode, string>();
            toUnicodeMapping.Add(new CharCode(97), "c");
            toUnicodeMapping.Add(new CharCode(98), "d");

            // Act
            ToUnicodeRangeParser.ParseToUnicodeMapping(interpreter, toUnicodeMapping);

            // Assert
            Assert.AreEqual(0, interpreter.Encoding.CharCodeToUnicodeMappings.Count());
            Assert.AreEqual(1, interpreter.Encoding.UnicodeRanges.Count);

            CharCodeRange charCodeRange = interpreter.Encoding.UnicodeRanges.Skip(0).First();
            
            Assert.IsTrue(charCodeRange.IsSequential);
            Assert.AreEqual(97, charCodeRange.Start.Code);
            Assert.AreEqual(98, charCodeRange.End.Code);

            List<string> unicodes = charCodeRange.Unicodes;
            Assert.AreEqual(2, unicodes.Count);
            Assert.AreEqual("c", unicodes[0]);
            Assert.AreEqual("d", unicodes[1]);
        }

        [TestMethod]
        public void ParseToUnicodeMapping_WhenHasMultipleRangesWithSequentialUnicodes_ShouldAddCorrectly()
        {
            // Arrange
            TestCMapEncodingInterpreter interpreter = new TestCMapEncodingInterpreter();
            interpreter.Operands = new Fixed.FormatProviders.Pdf.Import.OperandsCollection();
            interpreter.Encoding = new CMapEncoding();

            SortedDictionary<CharCode, string> toUnicodeMapping = new SortedDictionary<CharCode, string>();
            toUnicodeMapping.Add(new CharCode(97), "c");
            toUnicodeMapping.Add(new CharCode(98), "d");
            toUnicodeMapping.Add(new CharCode(100), "x");
            toUnicodeMapping.Add(new CharCode(101), "y");

            // Act
            ToUnicodeRangeParser.ParseToUnicodeMapping(interpreter, toUnicodeMapping);

            // Assert
            Assert.AreEqual(0, interpreter.Encoding.CharCodeToUnicodeMappings.Count());
            Assert.AreEqual(2, interpreter.Encoding.UnicodeRanges.Count);

            CharCodeRange firstCharCodeRange = interpreter.Encoding.UnicodeRanges.Skip(0).First();

            Assert.IsTrue(firstCharCodeRange.IsSequential);
            Assert.AreEqual(97, firstCharCodeRange.Start.Code);
            Assert.AreEqual(98, firstCharCodeRange.End.Code);

            List<string> unicodes = firstCharCodeRange.Unicodes;
            Assert.AreEqual(2, unicodes.Count);
            Assert.AreEqual("c", unicodes[0]);
            Assert.AreEqual("d", unicodes[1]);

            CharCodeRange secondCharCodeRange = interpreter.Encoding.UnicodeRanges.Skip(1).First();
            
            Assert.IsTrue(secondCharCodeRange.IsSequential);
            Assert.AreEqual(100, secondCharCodeRange.Start.Code);
            Assert.AreEqual(101, secondCharCodeRange.End.Code);

            unicodes = secondCharCodeRange.Unicodes;
            Assert.AreEqual(2, unicodes.Count);
            Assert.AreEqual("x", unicodes[0]);
            Assert.AreEqual("y", unicodes[1]);
        }

        [TestMethod]
        public void ParseToUnicodeMapping_WhenHasRangeWithNonSequentialUnicodes_ShouldAddCorrectly()
        {
            // Arrange
            TestCMapEncodingInterpreter interpreter = new TestCMapEncodingInterpreter();
            interpreter.Operands = new Fixed.FormatProviders.Pdf.Import.OperandsCollection();
            interpreter.Encoding = new CMapEncoding();

            SortedDictionary<CharCode, string> toUnicodeMapping = new SortedDictionary<CharCode, string>();
            toUnicodeMapping.Add(new CharCode(97), "x");
            toUnicodeMapping.Add(new CharCode(98), "z");

            // Act
            ToUnicodeRangeParser.ParseToUnicodeMapping(interpreter, toUnicodeMapping);

            // Assert
            Assert.AreEqual(0, interpreter.Encoding.CharCodeToUnicodeMappings.Count());
            Assert.AreEqual(1, interpreter.Encoding.UnicodeRanges.Count);

            CharCodeRange charCodeRange = interpreter.Encoding.UnicodeRanges.Skip(0).First();

            Assert.IsFalse(charCodeRange.IsSequential);
            Assert.AreEqual(97, charCodeRange.Start.Code);
            Assert.AreEqual(98, charCodeRange.End.Code);

            List<string> unicodes = charCodeRange.Unicodes;
            Assert.AreEqual(2, unicodes.Count);
            Assert.AreEqual("x", unicodes[0]);
            Assert.AreEqual("z", unicodes[1]);
        }

        [TestMethod]
        public void ParseToUnicodeMapping_WhenHasMultipleRangesWithNonSequentialUnicodes_ShouldAddCorrectly()
        {
            // Arrange
            TestCMapEncodingInterpreter interpreter = new TestCMapEncodingInterpreter();
            interpreter.Operands = new Fixed.FormatProviders.Pdf.Import.OperandsCollection();
            interpreter.Encoding = new CMapEncoding();

            SortedDictionary<CharCode, string> toUnicodeMapping = new SortedDictionary<CharCode, string>();
            toUnicodeMapping.Add(new CharCode(97), "x");
            toUnicodeMapping.Add(new CharCode(98), "z");
            toUnicodeMapping.Add(new CharCode(100), "X");
            toUnicodeMapping.Add(new CharCode(101), "Z");

            // Act
            ToUnicodeRangeParser.ParseToUnicodeMapping(interpreter, toUnicodeMapping);

            // Assert
            Assert.AreEqual(0, interpreter.Encoding.CharCodeToUnicodeMappings.Count());
            Assert.AreEqual(2, interpreter.Encoding.UnicodeRanges.Count);

            CharCodeRange firstCharCodeRange = interpreter.Encoding.UnicodeRanges.Skip(0).First();

            Assert.IsFalse(firstCharCodeRange.IsSequential);
            Assert.AreEqual(97, firstCharCodeRange.Start.Code);
            Assert.AreEqual(98, firstCharCodeRange.End.Code);

            List<string> unicodes = firstCharCodeRange.Unicodes;
            Assert.AreEqual(2, unicodes.Count);
            Assert.AreEqual("x", unicodes[0]);
            Assert.AreEqual("z", unicodes[1]);

            CharCodeRange secondCharCodeRange = interpreter.Encoding.UnicodeRanges.Skip(1).First();

            Assert.IsFalse(secondCharCodeRange.IsSequential);
            Assert.AreEqual(100, secondCharCodeRange.Start.Code);
            Assert.AreEqual(101, secondCharCodeRange.End.Code);

            unicodes = secondCharCodeRange.Unicodes;
            Assert.AreEqual(2, unicodes.Count);
            Assert.AreEqual("X", unicodes[0]);
            Assert.AreEqual("Z", unicodes[1]);
        }

        [TestMethod]
        public void ParseToUnicodeMapping_WhenHasMultipleRangesWithSequentialAndNonSequentialUnicodes_ShouldAddCorrectly()
        {
            // Arrange
            TestCMapEncodingInterpreter interpreter = new TestCMapEncodingInterpreter();
            interpreter.Operands = new Fixed.FormatProviders.Pdf.Import.OperandsCollection();
            interpreter.Encoding = new CMapEncoding();

            SortedDictionary<CharCode, string> toUnicodeMapping = new SortedDictionary<CharCode, string>();
            toUnicodeMapping.Add(new CharCode(97), "c");
            toUnicodeMapping.Add(new CharCode(98), "d");
            toUnicodeMapping.Add(new CharCode(100), "X");
            toUnicodeMapping.Add(new CharCode(101), "Z");

            // Act
            ToUnicodeRangeParser.ParseToUnicodeMapping(interpreter, toUnicodeMapping);

            // Assert
            Assert.AreEqual(0, interpreter.Encoding.CharCodeToUnicodeMappings.Count());
            Assert.AreEqual(2, interpreter.Encoding.UnicodeRanges.Count);

            CharCodeRange firstCharCodeRange = interpreter.Encoding.UnicodeRanges.Skip(0).First();

            Assert.IsTrue(firstCharCodeRange.IsSequential);
            Assert.AreEqual(97, firstCharCodeRange.Start.Code);
            Assert.AreEqual(98, firstCharCodeRange.End.Code);

            List<string> unicodes = firstCharCodeRange.Unicodes;
            Assert.AreEqual(2, unicodes.Count);
            Assert.AreEqual("c", unicodes[0]);
            Assert.AreEqual("d", unicodes[1]);

            CharCodeRange secondCharCodeRange = interpreter.Encoding.UnicodeRanges.Skip(1).First();

            Assert.IsFalse(secondCharCodeRange.IsSequential);
            Assert.AreEqual(100, secondCharCodeRange.Start.Code);
            Assert.AreEqual(101, secondCharCodeRange.End.Code);

            unicodes = secondCharCodeRange.Unicodes;
            Assert.AreEqual(2, unicodes.Count);
            Assert.AreEqual("X", unicodes[0]);
            Assert.AreEqual("Z", unicodes[1]);
        }

        [TestMethod]
        public void ParseToUnicodeMapping_WhenNonSequentialCharcodes_ShouldAddCharcodesMappingWithoutRange()
        {
            // Arrange
            TestCMapEncodingInterpreter interpreter = new TestCMapEncodingInterpreter();
            interpreter.Operands = new Fixed.FormatProviders.Pdf.Import.OperandsCollection();
            interpreter.Encoding = new CMapEncoding();

            SortedDictionary<CharCode, string> toUnicodeMapping = new SortedDictionary<CharCode, string>();
            toUnicodeMapping.Add(new CharCode(97), "c");
            toUnicodeMapping.Add(new CharCode(100), "d");

            // Act
            ToUnicodeRangeParser.ParseToUnicodeMapping(interpreter, toUnicodeMapping);

            // Assert
            Assert.AreEqual(2, interpreter.Encoding.CharCodeToUnicodeMappings.Count());
            Assert.AreEqual(0, interpreter.Encoding.UnicodeRanges.Count);

            Tuple<CharCode, string> firstPair = interpreter.Encoding.CharCodeToUnicodeMappings.Skip(0).First();
            CharCode charCode = firstPair.Item1;
            Assert.AreEqual(97, charCode.Code);
            string unicode = firstPair.Item2;
            Assert.AreEqual("c", unicode);

            Tuple<CharCode, string> secondPair = interpreter.Encoding.CharCodeToUnicodeMappings.Skip(1).First();
            charCode = secondPair.Item1;
            Assert.AreEqual(100, charCode.Code);
            unicode = secondPair.Item2;
            Assert.AreEqual("d", unicode);
        }

        [TestMethod]
        public void ParseToUnicodeMapping_WhenManyRanges_ShouldAddCorrectly()
        {
            // Arrange
            TestCMapEncodingInterpreter interpreter = new TestCMapEncodingInterpreter();
            interpreter.Operands = new Fixed.FormatProviders.Pdf.Import.OperandsCollection();
            interpreter.Encoding = new CMapEncoding();

            SortedDictionary<CharCode, string> toUnicodeMapping = new SortedDictionary<CharCode, string>();
            toUnicodeMapping.Add(new CharCode(97), "e");
            toUnicodeMapping.Add(new CharCode(98), "y");
            toUnicodeMapping.Add(new CharCode(100), "z");
            toUnicodeMapping.Add(new CharCode(101), "a");
            toUnicodeMapping.Add(new CharCode(102), "b");
            toUnicodeMapping.Add(new CharCode(103), "c");
            toUnicodeMapping.Add(new CharCode(104), "d");
            toUnicodeMapping.Add(new CharCode(105), "x");
            toUnicodeMapping.Add(new CharCode(200), "W");

            // Act
            ToUnicodeRangeParser.ParseToUnicodeMapping(interpreter, toUnicodeMapping);

            // Assert
            Assert.AreEqual(4, interpreter.Encoding.UnicodeRanges.Count);

            CharCodeRange firstCharCodeRange = interpreter.Encoding.UnicodeRanges.Skip(0).First();

            Assert.IsFalse(firstCharCodeRange.IsSequential);
            Assert.AreEqual(97, firstCharCodeRange.Start.Code);
            Assert.AreEqual(98, firstCharCodeRange.End.Code);

            List<string> unicodes = firstCharCodeRange.Unicodes;
            Assert.AreEqual(2, unicodes.Count);
            Assert.AreEqual("e", unicodes[0]);
            Assert.AreEqual("y", unicodes[1]);

            CharCodeRange secondCharCodeRange = interpreter.Encoding.UnicodeRanges.Skip(1).First();

            Assert.IsFalse(secondCharCodeRange.IsSequential);
            Assert.AreEqual(100, secondCharCodeRange.Start.Code);
            Assert.AreEqual(101, secondCharCodeRange.End.Code);

            unicodes = secondCharCodeRange.Unicodes;
            Assert.AreEqual(2, unicodes.Count);
            Assert.AreEqual("z", unicodes[0]);
            Assert.AreEqual("a", unicodes[1]);

            CharCodeRange thirdCharCodeRange = interpreter.Encoding.UnicodeRanges.Skip(2).First();

            Assert.IsTrue(thirdCharCodeRange.IsSequential);
            Assert.AreEqual(102, thirdCharCodeRange.Start.Code);
            Assert.AreEqual(104, thirdCharCodeRange.End.Code);

            unicodes = thirdCharCodeRange.Unicodes;
            Assert.AreEqual(3, unicodes.Count);
            Assert.AreEqual("b", unicodes[0]);
            Assert.AreEqual("c", unicodes[1]);
            Assert.AreEqual("d", unicodes[2]);

            CharCodeRange fourthCharCodeRange = interpreter.Encoding.UnicodeRanges.Skip(3).First();

            Assert.IsFalse(fourthCharCodeRange.IsSequential);
            Assert.AreEqual(105, fourthCharCodeRange.Start.Code);
            Assert.AreEqual(105, fourthCharCodeRange.End.Code);

            unicodes = fourthCharCodeRange.Unicodes;
            Assert.AreEqual(1, unicodes.Count);
            Assert.AreEqual("x", unicodes[0]);

            Assert.AreEqual(1, interpreter.Encoding.CharCodeToUnicodeMappings.Count());
            Assert.AreEqual(200, interpreter.Encoding.CharCodeToUnicodeMappings.First().Item1.Code);
            Assert.AreEqual("W", interpreter.Encoding.CharCodeToUnicodeMappings.First().Item2);
        }
    }
}
