﻿#if !NETSTANDARD
using System.Windows;
#else
using Telerik.Documents.Primitives;
#endif
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Data;
using Telerik.Windows.Documents.Fixed.Model.Fonts;
using Telerik.Windows.Documents.Fixed.Model.Objects;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Fixed.Model.Text;
using Telerik.Windows.Documents.Fixed.Text;

namespace Telerik.Windows.Documents.Fixed.Tests.Text
{
    [TestClass]
    public class TextRecognizerTests
    {
        private const string LinesSeparator = " ";
        private const double Delta = 0.00001;

        [TestMethod]
        public void CreateTextPage_WhenTextIsDefinedInFormXObject_ShouldCreateCorrectTextPage()
        {
            // Arrange
            FormSource formSource = new FormSource();
            string text = "Sample text";
            formSource.Content.AddTextFragment(text);

            RadFixedPage page = new RadFixedPage();
            page.Content.AddForm(formSource);

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Assert.AreEqual(text, textPage.ToString(LinesSeparator));
        }

        [TestMethod]
        public void CreateTextPage_WhenTextIsDefinedInFormXObjectWithPositionMatrix_ShouldCrateTextPageWithCorrectCharacterBoundingRectangles()
        {
            // Arrange
            FormSource formSource = new FormSource();
            string text = "Sample text";
            TextFragment textFragment = formSource.Content.AddTextFragment(text);
            textFragment.Font = FontsRepository.Helvetica;
            textFragment.FontSize = 14;
            formSource.Size = new Size(100, 20);

            RadFixedPage page = new RadFixedPage();
            Form form = page.Content.AddForm(formSource);
            form.Position.Translate(100, 100);

            Rect firstCharacterBoundingRect = new Rect(100, 88.954, 9.338, 14);
            Rect lastCharacterBoundingRect = new Rect(170.028, 88.954, 3.892, 14);

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Line line = textPage.GetLineAt(0);
            Character firstCharacter = line.FirstWord.GetCharacterAt(0);

            Assert.AreEqual('S', firstCharacter.Char);
            this.AssertRectangle(firstCharacterBoundingRect, firstCharacter.BoundingRect);

            Character lastCharacter = line.LastWord.GetCharacterAt(3);

            Assert.AreEqual('t', lastCharacter.Char);
            this.AssertRectangle(lastCharacterBoundingRect, lastCharacter.BoundingRect);
        }

        [TestMethod]
        public void CreateTextPage_WhenTextIsDefinedInFormXObjectWithScaledPositionMatrix_ShouldCrateTextPageWithCorrectCharacterBoundingRectangles()
        {
            // Arrange
            FormSource formSource = new FormSource();
            string text = "Sample text";
            TextFragment textFragment = formSource.Content.AddTextFragment(text);
            textFragment.Font = FontsRepository.Helvetica;
            textFragment.FontSize = 14;
            formSource.Size = new Size(100, 20);

            RadFixedPage page = new RadFixedPage();
            Form form = page.Content.AddForm(formSource);
            form.Position.Translate(100, 100);
            form.Width = 200;
            form.Height = 40;

            Rect firstCharacterBoundingRect = new Rect(100, 77.908, 18.676, 28);
            Rect lastCharacterBoundingRect = new Rect(240.056, 77.908, 7.784, 28);

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Line line = textPage.GetLineAt(0);
            Character firstCharacter = line.FirstWord.GetCharacterAt(0);

            Assert.AreEqual('S', firstCharacter.Char);
            this.AssertRectangle(firstCharacterBoundingRect, firstCharacter.BoundingRect);

            Character lastCharacter = line.LastWord.GetCharacterAt(3);

            Assert.AreEqual('t', lastCharacter.Char);
            this.AssertRectangle(lastCharacterBoundingRect, lastCharacter.BoundingRect);
        }

        [TestMethod]
        public void CreateTextPage_WhenTextIsDefinedInFormXObjectWithPositionMatrixAndPageRotated90_ShouldCrateTextPageWithCorrectCharacterBoundingRectangles()
        {
            // Arrange
            FormSource formSource = new FormSource();
            string text = "Sample text";
            TextFragment textFragment = formSource.Content.AddTextFragment(text);
            textFragment.Font = FontsRepository.Helvetica;
            textFragment.FontSize = 14;
            formSource.Size = new Size(100, 20);

            RadFixedPage page = new RadFixedPage();
            page.Rotation = Rotation.Rotate90;
            Form form = page.Content.AddForm(formSource);
            form.Position.Translate(100, 100);

            Rect firstCharacterBoundingRect = new Rect(953.046, 100, 14, 9.338);
            Rect lastCharacterBoundingRect = new Rect(953.046, 170.028, 14, 3.892);

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Line firstLine = textPage.GetLineAt(0);
            Character firstCharacter = firstLine.FirstWord.GetCharacterAt(0);

            Assert.AreEqual('S', firstCharacter.Char);
            this.AssertRectangle(firstCharacterBoundingRect, firstCharacter.BoundingRect);

            Line lastLine = textPage.LastLine;
            Character lastCharacter = lastLine.FirstWord.GetCharacterAt(0);

            Assert.AreEqual('t', lastCharacter.Char);
            this.AssertRectangle(lastCharacterBoundingRect, lastCharacter.BoundingRect);
        }

        [TestMethod]
        public void CreateTextPage_WhenTextWithPositionMatrix_ShouldCrateTextPageWithCorrectCharacterBoundingRectangles()
        {
            // Arrange
            RadFixedPage page = new RadFixedPage();
            string text = "Sample text";
            TextFragment textFragment = page.Content.AddTextFragment(text);
            textFragment.Font = FontsRepository.Helvetica;
            textFragment.FontSize = 14;
            textFragment.Position.Translate(100, 100);

            Rect firstCharacterBoundingRect = new Rect(100, 88.954, 9.338, 14);
            Rect lastCharacterBoundingRect = new Rect(170.028, 88.954, 3.892, 14);

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Line line = textPage.GetLineAt(0);
            Character firstCharacter = line.FirstWord.GetCharacterAt(0);

            Assert.AreEqual('S', firstCharacter.Char);
            this.AssertRectangle(firstCharacterBoundingRect, firstCharacter.BoundingRect);

            Character lastCharacter = line.LastWord.GetCharacterAt(3);

            Assert.AreEqual('t', lastCharacter.Char);
            this.AssertRectangle(lastCharacterBoundingRect, lastCharacter.BoundingRect);
        }

        [TestMethod]
        public void CreateTextPage_WhenTextWithScaledPositionMatrix_ShouldCrateTextPageWithCorrectCharacterBoundingRectangles()
        {
            // Arrange
            RadFixedPage page = new RadFixedPage();
            string text = "Sample text";
            TextFragment textFragment = page.Content.AddTextFragment(text);
            textFragment.Font = FontsRepository.Helvetica;
            textFragment.FontSize = 1;
            textFragment.Position.Scale(14, 14);
            textFragment.Position.Translate(100, 100);

            Rect firstCharacterBoundingRect = new Rect(100, 88.954, 9.338, 14);
            Rect lastCharacterBoundingRect = new Rect(170.028, 88.954, 3.892, 14);

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Line line = textPage.GetLineAt(0);
            Character firstCharacter = line.FirstWord.GetCharacterAt(0);

            Assert.AreEqual('S', firstCharacter.Char);
            this.AssertRectangle(firstCharacterBoundingRect, firstCharacter.BoundingRect);

            Character lastCharacter = line.LastWord.GetCharacterAt(3);

            Assert.AreEqual('t', lastCharacter.Char);
            this.AssertRectangle(lastCharacterBoundingRect, lastCharacter.BoundingRect);
        }

        [TestMethod]
        public void CreateTextPage_WhenTextWithPositionMatrixAndPageRotated90_ShouldCrateTextPageWithCorrectCharacterBoundingRectangles()
        {
            // Arrange
            RadFixedPage page = new RadFixedPage();
            page.Rotation = Rotation.Rotate90;

            string text = "Sample text";
            TextFragment textFragment = page.Content.AddTextFragment(text);
            textFragment.Font = FontsRepository.Helvetica;
            textFragment.FontSize = 14;
            textFragment.Position.Translate(100, 100);

            Rect firstCharacterBoundingRect = new Rect(953.046, 100, 14, 9.338);
            Rect lastCharacterBoundingRect = new Rect(953.046, 170.028, 14, 3.892);

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Line firstLine = textPage.GetLineAt(0);
            Character firstCharacter = firstLine.FirstWord.GetCharacterAt(0);

            Assert.AreEqual('S', firstCharacter.Char);
            this.AssertRectangle(firstCharacterBoundingRect, firstCharacter.BoundingRect);

            Line lastLine = textPage.LastLine;
            Character lastCharacter = lastLine.FirstWord.GetCharacterAt(0);

            Assert.AreEqual('t', lastCharacter.Char);
            this.AssertRectangle(lastCharacterBoundingRect, lastCharacter.BoundingRect);
        }

        [TestMethod]
        public void CreateTextPage_WhenTextWithPositionMatrixAndTextRise_ShouldCrateTextPageWithCorrectCharacterBoundingRectangles()
        {
            // Arrange
            RadFixedPage page = new RadFixedPage();
            string text = "Sample text";
            TextFragment textFragment = page.Content.AddTextFragment(text);
            textFragment.Font = FontsRepository.Helvetica;
            textFragment.FontSize = 14;
            textFragment.TextRise = 5;
            textFragment.Position.Translate(100, 100);

            Rect firstCharacterBoundingRect = new Rect(100, 83.954, 9.338, 14);
            Rect lastCharacterBoundingRect = new Rect(170.028, 83.954, 3.892, 14);

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Line line = textPage.GetLineAt(0);
            Character firstCharacter = line.FirstWord.GetCharacterAt(0);

            Assert.AreEqual('S', firstCharacter.Char);
            this.AssertRectangle(firstCharacterBoundingRect, firstCharacter.BoundingRect);

            Character lastCharacter = line.LastWord.GetCharacterAt(3);

            Assert.AreEqual('t', lastCharacter.Char);
            this.AssertRectangle(lastCharacterBoundingRect, lastCharacter.BoundingRect);
        }

        [TestMethod]
        public void CreateTextPage_WhenTextWithPositionMatrixAndPositiveCharacterSpacing_ShouldCreateCorrectTextPage()
        {
            // Arrange
            RadFixedPage page = new RadFixedPage();
            string text = "Sample text";

            TextFragment textFragment = page.Content.AddTextFragment(text);
            textFragment.Font = FontsRepository.Helvetica;
            textFragment.FontSize = 14;
            textFragment.CharacterSpacing = 0.1;
            textFragment.Position.Translate(100, 100);

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Assert.AreEqual(text, textPage.ToString(LinesSeparator));
        }

        [TestMethod]
        public void CreateTextPage_WhenTextWithPositionMatrixAndPositiveCharacterSpacing_ShouldCrateTextPageWithCorrectCharacterBoundingRectangles()
        {
            // Arrange
            RadFixedPage page = new RadFixedPage();
            string text = "Sample text";
            TextFragment textFragment = page.Content.AddTextFragment(text);
            textFragment.Font = FontsRepository.Helvetica;
            textFragment.FontSize = 14;
            textFragment.CharacterSpacing = 0.1;
            textFragment.Position.Translate(100, 100);

            Rect firstCharacterBoundingRect = new Rect(100, 88.954, 9.338, 14);
            Rect lastCharacterBoundingRect = new Rect(171.028, 88.954, 3.892, 14);

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Line line = textPage.GetLineAt(0);
            Character firstCharacter = line.FirstWord.GetCharacterAt(0);

            Assert.AreEqual('S', firstCharacter.Char);
            this.AssertRectangle(firstCharacterBoundingRect, firstCharacter.BoundingRect);

            Character lastCharacter = line.LastWord.GetCharacterAt(3);

            Assert.AreEqual('t', lastCharacter.Char);
            this.AssertRectangle(lastCharacterBoundingRect, lastCharacter.BoundingRect);
        }

        [TestMethod]
        public void CreateTextPage_WhenTextWithScaledPositionMatrixAndPositiveCharacterSpacing_ShouldCreateCorrectTextPage()
        {
            // Arrange
            RadFixedPage page = new RadFixedPage();
            string text = "Sample text";

            TextFragment textFragment = page.Content.AddTextFragment(text);
            textFragment.Font = FontsRepository.Helvetica;
            textFragment.FontSize = 1;
            textFragment.CharacterSpacing = 0.1;
            textFragment.Position.Scale(14, 14);
            textFragment.Position.Translate(100, 100);

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Assert.AreEqual(text, textPage.ToString(LinesSeparator));
        }

        [TestMethod]
        public void CreateTextPage_WhenTextWithScaledPositionMatrixAndPositiveCharacterSpacing_ShouldCrateTextPageWithCorrectCharacterBoundingRectangles()
        {
            // Arrange
            RadFixedPage page = new RadFixedPage();
            string text = "Sample text";
            TextFragment textFragment = page.Content.AddTextFragment(text);
            textFragment.Font = FontsRepository.Helvetica;
            textFragment.FontSize = 1;
            textFragment.CharacterSpacing = 0.1;
            textFragment.Position.Scale(14, 14);
            textFragment.Position.Translate(100, 100);

            Rect firstCharacterBoundingRect = new Rect(100, 88.954, 9.338, 14);
            Rect lastCharacterBoundingRect = new Rect(184.028, 88.954, 3.892, 14);

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Line line = textPage.GetLineAt(0);
            Character firstCharacter = line.FirstWord.GetCharacterAt(0);

            Assert.AreEqual('S', firstCharacter.Char);
            this.AssertRectangle(firstCharacterBoundingRect, firstCharacter.BoundingRect);

            Character lastCharacter = line.LastWord.GetCharacterAt(3);

            Assert.AreEqual('t', lastCharacter.Char);
            this.AssertRectangle(lastCharacterBoundingRect, lastCharacter.BoundingRect);
        }

        [TestMethod]
        public void CreateTextPage_WhenTextWithPositionMatrixAndNegativeCharacterSpacing_ShouldCrateTextPageWithCorrectCharacterBoundingRectangles()
        {
            // Arrange
            RadFixedPage page = new RadFixedPage();
            string text = "Sample text";
            TextFragment textFragment = page.Content.AddTextFragment(text);
            textFragment.Font = FontsRepository.Helvetica;
            textFragment.FontSize = 14;
            textFragment.CharacterSpacing = -0.1;
            textFragment.Position.Translate(100, 100);

            Rect firstCharacterBoundingRect = new Rect(100, 88.954, 9.338, 14);
            Rect lastCharacterBoundingRect = new Rect(169.028, 88.954, 3.892, 14);

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Line line = textPage.GetLineAt(0);
            Character firstCharacter = line.FirstWord.GetCharacterAt(0);

            Assert.AreEqual('S', firstCharacter.Char);
            this.AssertRectangle(firstCharacterBoundingRect, firstCharacter.BoundingRect);

            Character lastCharacter = line.LastWord.GetCharacterAt(3);

            Assert.AreEqual('t', lastCharacter.Char);
            this.AssertRectangle(lastCharacterBoundingRect, lastCharacter.BoundingRect);
        }

        [TestMethod]
        public void CreateTextPage_WhenTextWithPositionMatrixAndHorizontalScaling_ShouldCrateTextPageWithCorrectCharacterBoundingRectangles()
        {
            // Arrange
            RadFixedPage page = new RadFixedPage();
            string text = "Sample text";
            TextFragment textFragment = page.Content.AddTextFragment(text);
            textFragment.Font = FontsRepository.Helvetica;
            textFragment.FontSize = 14;
            textFragment.HorizontalScaling = 2;
            textFragment.Position.Translate(100, 100);

            Rect firstCharacterBoundingRect = new Rect(100, 88.954, 18.676, 14);
            Rect lastCharacterBoundingRect = new Rect(240.056, 88.954, 7.784, 14);

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Line line = textPage.GetLineAt(0);
            Character firstCharacter = line.FirstWord.GetCharacterAt(0);

            Assert.AreEqual('S', firstCharacter.Char);
            this.AssertRectangle(firstCharacterBoundingRect, firstCharacter.BoundingRect);

            Character lastCharacter = line.LastWord.GetCharacterAt(3);

            Assert.AreEqual('t', lastCharacter.Char);
            this.AssertRectangle(lastCharacterBoundingRect, lastCharacter.BoundingRect);
        }

        [TestMethod]
        public void CreateTextPage_WhenTextWithPositionMatrixAndWordSpacing_ShouldCrateTextPageWithCorrectCharacterBoundingRectangles()
        {
            // Arrange
            RadFixedPage page = new RadFixedPage();
            string text = "Sample text";
            TextFragment textFragment = page.Content.AddTextFragment(text);
            textFragment.Font = FontsRepository.Helvetica;
            textFragment.FontSize = 14;
            textFragment.WordSpacing = 20;
            textFragment.Position.Translate(100, 100);

            Rect firstCharacterBoundingRect = new Rect(100, 88.954, 9.338, 14);
            Rect lastCharacterBoundingRect = new Rect(190.028, 88.954, 3.892, 14);

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Line line = textPage.GetLineAt(0);
            Character firstCharacter = line.FirstWord.GetCharacterAt(0);

            Assert.AreEqual('S', firstCharacter.Char);
            this.AssertRectangle(firstCharacterBoundingRect, firstCharacter.BoundingRect);

            Character lastCharacter = line.LastWord.GetCharacterAt(3);

            Assert.AreEqual('t', lastCharacter.Char);
            this.AssertRectangle(lastCharacterBoundingRect, lastCharacter.BoundingRect);
        }

        [TestMethod]
        public void CreateTextPage_WhenTextWithManySpaces_ShouldCreateTextPageWithOneSpace()
        {
            // Arrange
            RadFixedPage page = new RadFixedPage();
            string textWithManySpaces = "Sample              text";
            string textWithOneSpace = "Sample text";

            TextFragment textFragment = page.Content.AddTextFragment(textWithManySpaces);
            textFragment.Font = FontsRepository.Helvetica;
            textFragment.FontSize = 14;
            textFragment.Position.Translate(100, 100);

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Assert.AreEqual(textWithOneSpace, textPage.ToString(LinesSeparator));
        }

        [TestMethod]
        public void CreateTextPage_WhenTextWithManySpaces_ShouldCrateTextPageWithCorrectCharacterBoundingRectangles()
        {
            // Arrange
            RadFixedPage page = new RadFixedPage();
            string text = "Sample              text";

            TextFragment textFragment = page.Content.AddTextFragment(text);
            textFragment.Font = FontsRepository.Helvetica;
            textFragment.FontSize = 14;
            textFragment.Position.Translate(100, 100);

            Rect firstCharacterBoundingRect = new Rect(100, 88.954, 9.338, 14);
            Rect lastCharacterBoundingRect = new Rect(220.624, 88.954, 3.892, 14);

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Line line = textPage.GetLineAt(0);
            Character firstCharacter = line.FirstWord.GetCharacterAt(0);

            Assert.AreEqual('S', firstCharacter.Char);
            this.AssertRectangle(firstCharacterBoundingRect, firstCharacter.BoundingRect);

            Character lastCharacter = line.LastWord.GetCharacterAt(3);

            Assert.AreEqual('t', lastCharacter.Char);
            this.AssertRectangle(lastCharacterBoundingRect, lastCharacter.BoundingRect);
        }

        [TestMethod]
        public void CreateTextPage_WhenTextWithNegativeFontSize_ShouldNotThrow()
        {
            // Arrange
            RadFixedPage page = new RadFixedPage();
            TextFragment textFragment = page.Content.AddTextFragment("");
            textFragment.FontSize = -14;

            TextDocument textDocument = new TextDocument(new RadFixedDocument());
            TextRecognizer textRecognizer = new TextRecognizer(textDocument);

            // Act
            TextPage textPage = textRecognizer.CreateTextPage(page);

            // Assert
            Assert.IsNotNull(textPage);
        }

        private void AssertRectangle(Rect expected, Rect actual)
        {
            Assert.AreEqual(expected.X, actual.X, Delta);
            Assert.AreEqual(expected.Y, actual.Y, Delta);
            Assert.AreEqual(expected.Width, actual.Width, Delta);
            Assert.AreEqual(expected.Height, actual.Height, Delta);
        }
    }
}
