﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.Model.Annotations;
using Telerik.Windows.Documents.Fixed.Model.InteractiveForms;
using Telerik.Windows.Documents.Fixed.Model.Resources;
using Telerik.Windows.Documents.Fixed.Utilities.Rendering.Annotations;

namespace Telerik.Windows.Documents.Fixed.Tests.Utilities.Rendering.Annotations
{
    [TestClass]
    public class WidgetAppearanceProviderTests
    {
        private const string Off = "Off";
        private const string On = "On";

        [TestMethod]
        public void TryProvideAppearance_WhenAnnoationIsNull_ShouldReturnFalse()
        {
            // Arrange
            WidgetAppearanceProvider widgetAppearanceProvider = new WidgetAppearanceProvider(null);
            FormSource appearance;

            // Act
            bool found = widgetAppearanceProvider.TryProvideAppearance(null, AnnotationAppearanceMode.Normal, out appearance);

            // Assert
            Assert.IsFalse(found);
            Assert.IsNull(appearance);
        }

        [TestMethod]
        public void TryProvideAppearance_WhenAnnoationIsNotWidget_ShouldReturnFalse()
        {
            // Arrange
            WidgetAppearanceProvider widgetAppearanceProvider = new WidgetAppearanceProvider(null);
            Link annotation = new Link();
            FormSource appearance;

            // Act
            bool found = widgetAppearanceProvider.TryProvideAppearance(annotation, AnnotationAppearanceMode.Normal, out appearance);

            // Assert
            Assert.IsFalse(found);
            Assert.IsNull(appearance);
        }

        [TestMethod]
        public void TryProvideAppearance_WhenRadioButtonAndAppearancesAreNull_ShouldReturnFalse()
        {
            // Arrange
            WidgetAppearanceProvider widgetAppearanceProvider = new WidgetAppearanceProvider(null);
            RadioButtonWidget annotation = new RadioButtonWidget(new RadioButtonField("FieldName"));
            annotation.Option = new RadioOption("1");
            FormSource appearance;

            // Act
            bool found = widgetAppearanceProvider.TryProvideAppearance(annotation, AnnotationAppearanceMode.Normal, out appearance);

            // Assert
            Assert.IsFalse(found);
            Assert.IsNull(appearance);
        }

        [TestMethod]
        public void TryProvideAppearance_WhenCheckBoxAndAppearancesAreNull_ShouldReturnFalse()
        {
            // Arrange
            WidgetAppearanceProvider widgetAppearanceProvider = new WidgetAppearanceProvider(null);
            TwoStatesButtonWidget annotation = new TwoStatesButtonWidget(new CheckBoxField("FieldName"));
            FormSource appearance;

            // Act
            bool found = widgetAppearanceProvider.TryProvideAppearance(annotation, AnnotationAppearanceMode.Normal, out appearance);

            // Assert
            Assert.IsFalse(found);
            Assert.IsNull(appearance);
        }

        [TestMethod]
        public void TryProvideAppearance_WhenVariableContentWidgetAndAppearancesAreNull_ShouldReturnFalse()
        {
            // Arrange
            WidgetAppearanceProvider widgetAppearanceProvider = new WidgetAppearanceProvider(null);
            VariableContentWidget annotation = new VariableContentWidget(new TextBoxField("FieldName"));
            FormSource appearance;

            // Act
            bool found = widgetAppearanceProvider.TryProvideAppearance(annotation, AnnotationAppearanceMode.Normal, out appearance);

            // Assert
            Assert.IsFalse(found);
            Assert.IsNull(appearance);
        }

        [TestMethod]
        public void TryProvideAppearance_WhenRadioButtonWithOnStateAppearance_ShouldProvideAppearance()
        {
            // Arrange
            WidgetAppearanceProvider widgetAppearanceProvider = new WidgetAppearanceProvider(null);
            RadioButtonField field = new RadioButtonField("FieldName");
            RadioOption radioOption = new RadioOption("1");
            field.Value = radioOption;
            RadioButtonWidget annotation = new RadioButtonWidget(field);
            SingleStateAppearances appearances = new SingleStateAppearances();
            FormSource expectedAppearance = new FormSource();
            appearances.NormalAppearance = expectedAppearance;
            MultiStateAppearances multiStateAppearances = new MultiStateAppearances();
            multiStateAppearances.AddAppearance(On, appearances);
            annotation.ConsumeImportedAppearances(multiStateAppearances);
            annotation.Option = radioOption;
            FormSource appearance;

            // Act
            bool found = widgetAppearanceProvider.TryProvideAppearance(annotation, AnnotationAppearanceMode.Normal, out appearance);

            // Assert
            Assert.IsTrue(found);
            Assert.AreEqual(expectedAppearance, appearance);
        }

        [TestMethod]
        public void TryProvideAppearance_WhenRadioButtonWithOffStateAppearance_ShouldProvideAppearance()
        {
            // Arrange
            WidgetAppearanceProvider widgetAppearanceProvider = new WidgetAppearanceProvider(null);
            RadioButtonField field = new RadioButtonField("FieldName");
            RadioOption radioOption = new RadioOption("1");
            field.Value = null;
            RadioButtonWidget annotation = new RadioButtonWidget(field);
            SingleStateAppearances appearances = new SingleStateAppearances();
            FormSource expectedAppearance = new FormSource();
            appearances.NormalAppearance = expectedAppearance;
            MultiStateAppearances multiStateAppearances = new MultiStateAppearances();
            multiStateAppearances.AddAppearance(Off, appearances);
            annotation.ConsumeImportedAppearances(multiStateAppearances);
            annotation.Option = radioOption;
            FormSource appearance;

            // Act
            bool found = widgetAppearanceProvider.TryProvideAppearance(annotation, AnnotationAppearanceMode.Normal, out appearance);

            // Assert
            Assert.IsTrue(found);
            Assert.AreEqual(expectedAppearance, appearance);
        }

        [TestMethod]
        public void TryProvideAppearance_WhenCheckBoxWithOnStateAppearance_ShouldProvideAppearance()
        {
            // Arrange
            WidgetAppearanceProvider widgetAppearanceProvider = new WidgetAppearanceProvider(null);
            CheckBoxField field = new CheckBoxField("FieldName");
            field.IsChecked = true;
            TwoStatesButtonWidget annotation = new TwoStatesButtonWidget(field);
            SingleStateAppearances appearances = new SingleStateAppearances();
            FormSource expectedAppearance = new FormSource();
            appearances.NormalAppearance = expectedAppearance;
            MultiStateAppearances multiStateAppearances = new MultiStateAppearances();
            multiStateAppearances.AddAppearance(On, appearances);
            annotation.ConsumeImportedAppearances(multiStateAppearances);
            FormSource appearance;

            // Act
            bool found = widgetAppearanceProvider.TryProvideAppearance(annotation, AnnotationAppearanceMode.Normal, out appearance);

            // Assert
            Assert.IsTrue(found);
            Assert.AreEqual(expectedAppearance, appearance);
        }

        [TestMethod]
        public void TryProvideAppearance_WhenCheckBoxWithOffStateAppearance_ShouldProvideAppearance()
        {
            // Arrange
            WidgetAppearanceProvider widgetAppearanceProvider = new WidgetAppearanceProvider(null);
            CheckBoxField field = new CheckBoxField("FieldName");
            field.IsChecked = false;
            TwoStatesButtonWidget annotation = new TwoStatesButtonWidget(field);
            SingleStateAppearances appearances = new SingleStateAppearances();
            FormSource expectedAppearance = new FormSource();
            appearances.NormalAppearance = expectedAppearance;
            MultiStateAppearances multiStateAppearances = new MultiStateAppearances();
            multiStateAppearances.AddAppearance(Off, appearances);
            annotation.ConsumeImportedAppearances(multiStateAppearances);
            FormSource appearance;

            // Act
            bool found = widgetAppearanceProvider.TryProvideAppearance(annotation, AnnotationAppearanceMode.Normal, out appearance);

            // Assert
            Assert.IsTrue(found);
            Assert.AreEqual(expectedAppearance, appearance);
        }
    }
}
