﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.Model.Data;
#if NETSTANDARD
using Telerik.Documents.Primitives;
using Telerik.Windows.Documents.Utilities;
#else
using System.Windows;
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Fixed.Tests.Model.Common;
using Telerik.Windows.Documents.Fixed.Utilities.Rendering;

namespace Telerik.Windows.Documents.Fixed.Tests.Utilities.Rendering
{
    [TestClass]
    public class PageLayoutHelperTests
    {
        [TestMethod]
        public void GetVisibleContentBox_WhenCropBoxIsEmpty_ShouldReturnRectangleWithZeroBounds()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(20, 50, 100, 200);
            fixedPage.CropBox = Rect.Empty;

            Rect expected = new Rect(0, 0, 0, 0);

            // Act
            Rect result = PageLayoutHelper.GetVisibleContentBox(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void GetVisibleContentBox_WhenCropBoxIsEmptyAndMediaBoxHasNegativeOffsets_ShouldReturnRectangleWithZeroBounds()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(-1000, -2000, 800, 1000);
            fixedPage.CropBox = Rect.Empty;

            Rect expected = new Rect(0, 0, 0, 0);

            // Act
            Rect result = PageLayoutHelper.GetVisibleContentBox(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void GetVisibleContentBox_WhenMediaBoxIsEmpty_ShouldReturnRectangleWithZeroBounds()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = Rect.Empty;
            fixedPage.CropBox = new Rect(20, 50, 200, 100);

            Rect expected = new Rect(0, 0, 0, 0);

            // Act
            Rect result = PageLayoutHelper.GetVisibleContentBox(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void GetVisibleContentBox_WhenBothAreEmpty_ShouldReturnRectangleWithZeroBounds()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = Rect.Empty;
            fixedPage.CropBox = Rect.Empty;

            Rect expected = new Rect(0, 0, 0, 0);

            Rect result = PageLayoutHelper.GetVisibleContentBox(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void GetVisibleContentBox_WhenCropBoxFitsInMediaBox_ShouldReturnCorrectIntersection()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(100, 100, 600, 800);

            Rect expected = new Rect(100, 100, 600, 800);

            // Act
            Rect result = PageLayoutHelper.GetVisibleContentBox(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void GetVisibleContentBox_WhenCropBoxIntersectsMediaBoxFromLeftToRight_ShouldReturnCorrectIntersection()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(-100, 100, 1000, 800);

            Rect expected = new Rect(0, 100, 800, 800);

            // Act
            Rect result = PageLayoutHelper.GetVisibleContentBox(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void GetVisibleContentBox_WhenCropBoxIntersectsMediaBoxInBottomLeft_ShouldReturnCorrectIntersection()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(-200, 600, 600, 800);

            Rect expected = new Rect(0, 600, 400, 400);

            // Act
            Rect result = PageLayoutHelper.GetVisibleContentBox(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void GetVisibleContentBox_WhenCropBoxIntersectsMediaBoxInTopRight_ShouldReturnCorrectIntersection()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(400, -200, 600, 600);

            Rect expected = new Rect(400, 0, 400, 400);

            // Act
            Rect result = PageLayoutHelper.GetVisibleContentBox(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void GetVisibleContentBox_WhenCropBoxAndMediaBoxHaveNegativeOffsets_ShouldReturnCorrectIntersection()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(-1000, -2000, 800, 1000);
            fixedPage.CropBox = new Rect(-1200, -2200, 600, 600);

            Rect expected = new Rect(-1000, -2000, 400, 400);

            // Act
            Rect result = PageLayoutHelper.GetVisibleContentBox(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void GetVisibleContentBox_WhenCropBoxAndMediaBoxAreTouchingBySide_ShouldReturnRectangleWithZeroWidthAndHeight()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(800, 1000, 600, 600);

            Rect expected = new Rect(800, 1000, 0, 0);

            // Act
            Rect result = PageLayoutHelper.GetVisibleContentBox(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void GetVisibleContentBox_WhenCropBoxAndMediaBoxDoNotIntersect_ShouldReturnRectangleWithZeroBounds()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(1000, 1200, 600, 600);

            Rect expected = new Rect(0, 0, 0, 0);

            // Act
            Rect result = PageLayoutHelper.GetVisibleContentBox(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void GetActualWidth_WhenRotated0_ShouldReturnCorrectWidth()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(100, 100, 200, 400);
            fixedPage.Rotation = Rotation.Rotate0;

            // Act
            double result = PageLayoutHelper.GetActualWidth(fixedPage);

            // Assert
            Assert.AreEqual(200, result);
        }

        [TestMethod]
        public void GetActualWidth_WhenRotated90_ShouldReturnCorrectWidth()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(-400, 600, 600, 600);
            fixedPage.Rotation = Rotation.Rotate90;

            // Act
            double result = PageLayoutHelper.GetActualWidth(fixedPage);

            // Assert
            Assert.AreEqual(400, result);
        }

        [TestMethod]
        public void GetActualWidth_WhenRotated180_ShouldReturnCorrectWidth()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(600, -200, 600, 600);
            fixedPage.Rotation = Rotation.Rotate180;

            // Act
            double result = PageLayoutHelper.GetActualWidth(fixedPage);

            // Assert
            Assert.AreEqual(200, result);
        }

        [TestMethod]
        public void GetActualWidth_WhenRotated270_ShouldReturnCorrectWidth()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(-1000, -2000, 800, 1000);
            fixedPage.CropBox = new Rect(-1000, -2000, 200, 400);
            fixedPage.Rotation = Rotation.Rotate270;

            // Act
            double result = PageLayoutHelper.GetActualWidth(fixedPage);

            // Assert
            Assert.AreEqual(400, result);
        }

        [TestMethod]
        public void GetActualHeight_WhenRotated0_ShouldReturnCorrectHeight()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(100, 100, 200, 400);
            fixedPage.Rotation = Rotation.Rotate0;

            // Act
            double result = PageLayoutHelper.GetActualHeight(fixedPage);

            // Assert
            Assert.AreEqual(400, result);
        }

        [TestMethod]
        public void GetActualHeight_WhenRotated90_ShouldReturnCorrectHeight()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(-400, 600, 600, 600);
            fixedPage.Rotation = Rotation.Rotate90;

            // Act
            double result = PageLayoutHelper.GetActualHeight(fixedPage);

            // Assert
            Assert.AreEqual(200, result);
        }

        [TestMethod]
        public void GetActualHeight_WhenRotated180_ShouldReturnCorrectHeight()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(600, -200, 600, 600);
            fixedPage.Rotation = Rotation.Rotate180;

            // Act
            double result = PageLayoutHelper.GetActualHeight(fixedPage);

            // Assert
            Assert.AreEqual(400, result);
        }

        [TestMethod]
        public void GetActualHeight_WhenRotated270_ShouldReturnCorrectHeight()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(-1000, -2000, 800, 1000);
            fixedPage.CropBox = new Rect(-1000, -2000, 200, 400);
            fixedPage.Rotation = Rotation.Rotate270;

            // Act
            double result = PageLayoutHelper.GetActualHeight(fixedPage);

            // Assert
            Assert.AreEqual(200, result);
        }

        [TestMethod]
        public void CalculateRotatedPageTransformationInDip_WhenRotated0_ShouReturnCorrectTransformation()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(100, 100, 200, 400);
            fixedPage.Rotation = Rotation.Rotate0;

            Matrix expected = new Matrix();

            // Act
            Matrix result = PageLayoutHelper.CalculateRotatedPageTransformationInDip(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void CalculateRotatedPageTransformationInDip_WhenRotated90_ShouReturnCorrectTransformation()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(-400, 600, 600, 600);
            fixedPage.Rotation = Rotation.Rotate90;

            Matrix expected = new Matrix();
#if NETSTANDARD
            expected = expected.RotateMatrixAt(90, 200, 200);
#else
            expected.RotateAt(90, 200, 200);
#endif

            // Act
            Matrix result = PageLayoutHelper.CalculateRotatedPageTransformationInDip(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void CalculateRotatedPageTransformationInDip_WhenRotated180_ShouReturnCorrectTransformation()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(600, -200, 600, 600);
            fixedPage.Rotation = Rotation.Rotate180;

            Matrix expected = new Matrix();
#if NETSTANDARD
            expected = expected.RotateMatrixAt(180, 100, 200);
#else
            expected.RotateAt(180, 100, 200);
#endif

            // Act
            Matrix result = PageLayoutHelper.CalculateRotatedPageTransformationInDip(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void CalculateRotatedPageTransformationInDip_WhenRotated270_ShouReturnCorrectTransformation()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(-1000, -2000, 800, 1000);
            fixedPage.CropBox = new Rect(-1000, -2000, 200, 400);
            fixedPage.Rotation = Rotation.Rotate270;

            Matrix expected = new Matrix();
#if NETSTANDARD
            expected = expected.RotateMatrixAt(270, 100, 100);
#else
            expected.RotateAt(270, 100, 100);
#endif

            // Act
            Matrix result = PageLayoutHelper.CalculateRotatedPageTransformationInDip(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void CalculateRotatedPageTransformationInDip_WhenCropBoxIsEmpty_ShouldReturnEmptyRectangle()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(-1000, -2000, 800, 1000);
            fixedPage.CropBox = Rect.Empty;
            fixedPage.Rotation = Rotation.Rotate270;

            Matrix expected = new Matrix();
#if NETSTANDARD
            expected = expected.RotateMatrixAt(270, 0, 0);
#else
            expected.RotateAt(270, 0, 0);
#endif

            // Act
            Matrix result = PageLayoutHelper.CalculateRotatedPageTransformationInDip(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void CalculatePageBoundingRectangle_WhenRotated0_ShouReturnCorrectRectangle()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(100, 100, 200, 400);
            fixedPage.Rotation = Rotation.Rotate0;

            Rect expected = new Rect(0, 0, 200, 400);

            // Act
            Rect result = PageLayoutHelper.CalculatePageBoundingRectangle(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void CalculatePageBoundingRectangle_WhenRotated90_ShouReturnCorrectRectangle()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(-400, 600, 600, 600);
            fixedPage.Rotation = Rotation.Rotate90;

            Matrix matrix = new Matrix();
#if NETSTANDARD
            matrix = matrix.RotateMatrixAt(90, 200, 200);
#else
            matrix.RotateAt(90, 200, 200);
#endif

            Rect expected = this.Transform(matrix, new Rect(0, 0, 200, 400));

            // Act
            Rect result = PageLayoutHelper.CalculatePageBoundingRectangle(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void CalculatePageBoundingRectangle_WhenRotated180_ShouReturnCorrectRectangle()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(0, 0, 800, 1000);
            fixedPage.CropBox = new Rect(600, -200, 600, 600);
            fixedPage.Rotation = Rotation.Rotate180;

            Matrix matrix = new Matrix();
#if NETSTANDARD
            matrix = matrix.RotateMatrixAt(180, 100, 200);
#else
            matrix.RotateAt(180, 100, 200);
#endif

            Rect expected = this.Transform(matrix, new Rect(0, 0, 200, 400));

            // Act
            Rect result = PageLayoutHelper.CalculatePageBoundingRectangle(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        public void CalculatePageBoundingRectangle_WhenRotated270_ShouReturnCorrectRectangle()
        {
            // Arrange
            TestFixedPage fixedPage = new TestFixedPage();
            fixedPage.MediaBox = new Rect(-1000, -2000, 800, 1000);
            fixedPage.CropBox = new Rect(-1000, -2000, 200, 400);
            fixedPage.Rotation = Rotation.Rotate270;

            Matrix matrix = new Matrix();
#if NETSTANDARD
            matrix = matrix.RotateMatrixAt(270, 100, 100);
#else
            matrix.RotateAt(270, 100, 100);
#endif

            Rect expected = this.Transform(matrix, new Rect(0, 0, 200, 400));

            // Act
            Rect result = PageLayoutHelper.CalculatePageBoundingRectangle(fixedPage);

            // Assert
            Assert.AreEqual(expected, result);
        }

        private Rect Transform(Matrix matrix, Rect rect)
        {
            return new Rect(matrix.Transform(new Point(rect.Left, rect.Top)), matrix.Transform(new Point(rect.Right, rect.Bottom)));
        }
    }
}
