﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Common.Tables
{
    internal class TableCellDataGrid
    {
        public TableCellDataGrid()
        {
            this.Data = new List<TableRowData>();
        }

        public bool IsInitializing { get; set; }

        public List<TableRowData> Data { get; private set; }

        public TableCellData this[int y, int x]
        {
            get
            {
                return this.Data[y].TableCellDatas[x];
            }
            set
            {
                this.Data[y].TableCellDatas[x] = value;
            }
        }

        public static TableCellDataGrid CreateFromTable(Table table)
        {
            TableCellDataGrid tableCellDataGrid = new TableCellDataGrid();

            tableCellDataGrid.IsInitializing = true;

            int rowIndex = 0;
            foreach (TableRow row in table.Rows)
            {
                int cellIndex = 0;
                foreach (TableCell cell in row.Cells)
                {
                    while (tableCellDataGrid.EnsureIndexExistsAndGetTableCellData(rowIndex, cellIndex, row).Type != TableCellType.Empty)
                    {
                        ++cellIndex;
                    }
                    tableCellDataGrid.EnsureIndexExistsAndSetTableCellData(rowIndex, cellIndex, row, new TableCellData(TableCellType.Cell, cell));

                    for (int y = rowIndex + 1; y < rowIndex + cell.RowSpan; ++y)
                    {
                        TableCell emptyClone = CreateVerticallyMergedEmptyCell(cell, y);
                        TableRow correspondingRow = table.Rows[y];
                        tableCellDataGrid.EnsureIndexExistsAndSetTableCellData(y, cellIndex, correspondingRow, new TableCellData(TableCellType.VerticallyMerged, emptyClone));
                    }

                    for (int y = rowIndex; y < rowIndex + cell.RowSpan; ++y)
                    {
                        for (int x = cellIndex + 1; x < cellIndex + cell.ColumnSpan; ++x)
                        {
                            TableRow correspondingRow = table.Rows[y];
                            tableCellDataGrid.EnsureIndexExistsAndSetTableCellData(y, x, correspondingRow, new TableCellData(TableCellType.HorizontalyMerged, cell));
                        }
                    }

                    ++cellIndex;
                }
                ++rowIndex;
            }

            tableCellDataGrid.IsInitializing = false;

            return tableCellDataGrid;
        }

        public static Table FillTableFromTableCellsData(Table tableToFill, List<TableRowData> tableRowDatasList)
        {
            Guard.ThrowExceptionIfNull(tableRowDatasList, "tableCellDatasList");

            TableCellDataGrid tableCellDataGrid = new TableCellDataGrid();
            tableCellDataGrid.IsInitializing = true;

            int currentRowIndex = 0;
            foreach (TableRowData rowData in tableRowDatasList)
            {
                int currentColumnIndex = 0;

                TableRow currentRow = rowData.Row;
                tableToFill.Rows.Add(currentRow);

                foreach (TableCellData tableCellData in rowData.TableCellDatas)
                {
                    bool skipCell = false;

                    while (tableCellDataGrid.EnsureIndexExistsAndGetTableCellData(currentRowIndex, currentColumnIndex, currentRow).Type != TableCellType.Empty)
                    {
                        currentColumnIndex++;
                    }

                    int columnSpan = tableCellData.Cell.ColumnSpan;

                    if (tableCellData.Type == TableCellType.VerticallyMerged && currentRowIndex > 0 && tableCellDataGrid.EnsureIndexExistsAndGetTableCellData(currentRowIndex - 1, currentColumnIndex, currentRow).Type != TableCellType.Empty)
                    {
                        TableCellData tableCellAbove = tableCellDataGrid.EnsureIndexExistsAndGetTableCellData(currentRowIndex - 1, currentColumnIndex, currentRow);
                        TableCellDataGrid.AddCellsToGrid(tableCellDataGrid, currentRow, ref currentColumnIndex, currentRowIndex, columnSpan, tableCellAbove);

                        int maxRowSpanForCurrentCell = TableCellDataGrid.GetMaxRowSpanForCell(tableCellDataGrid, currentRow, currentColumnIndex, tableCellAbove, tableRowDatasList.Count);
                        if (tableCellAbove.Cell.RowSpan < maxRowSpanForCurrentCell)
                        {
                            tableCellAbove.Cell.RowSpan++;
                        }
                        skipCell = true;
                    }
                    else if (columnSpan > 1)
                    {
                        TableCellDataGrid.AddCellsToGrid(tableCellDataGrid, currentRow, ref currentColumnIndex, currentRowIndex, columnSpan, tableCellData);
                    }
                    else if (tableCellData.Type == TableCellType.HorizontalyMerged && currentColumnIndex > 0 && tableCellDataGrid.EnsureIndexExistsAndGetTableCellData(currentRowIndex, currentColumnIndex - 1, currentRow).Type != TableCellType.Empty)
                    {
                        TableCellData cellBefore = tableCellDataGrid.EnsureIndexExistsAndGetTableCellData(currentRowIndex, currentColumnIndex - 1, currentRow);
                        TableCellDataGrid.AddCellsToGrid(tableCellDataGrid, currentRow, ref currentColumnIndex, currentRowIndex, columnSpan, cellBefore);
                        cellBefore.Cell.ColumnSpan++;
                        skipCell = true;
                    }
                    else
                    {
                        TableCellDataGrid.AddCellsToGrid(tableCellDataGrid, currentRow, ref currentColumnIndex, currentRowIndex, columnSpan, tableCellData);
                    }

                    if (!skipCell)
                    {
                        currentRow.Cells.Add(tableCellData.Cell);
                    }
                }

                currentRowIndex++;
            }

            return tableToFill;
        }

        private static TableCell CreateVerticallyMergedEmptyCell(TableCell cell, int gridRowIndex)
        {
            TableCell emptyClone = cell.Clone();
            emptyClone.Blocks.Clear();
            emptyClone.Blocks.AddParagraph();
            emptyClone.GridColumnIndex = cell.GridColumnIndex;
            emptyClone.GridRowIndex = gridRowIndex;
            emptyClone.RowSpan = 1;

            return emptyClone;
        }

        private static int GetMaxRowSpanForCell(TableCellDataGrid tableCellDataGrid, TableRow currentRow, int currentColumnIndex, TableCellData tableCellData, int rowsCount)
        {
            bool indexFound = false;
            int startIndex = 0;

            for (int i = 0; i < rowsCount; i++)
            {
                if (tableCellDataGrid.EnsureIndexExistsAndGetTableCellData(i, currentColumnIndex - tableCellData.Cell.ColumnSpan, currentRow) == tableCellData)
                {
                    indexFound = true;
                    startIndex = i;
                    break;
                }
            }

            Guard.ThrowExceptionIfFalse(indexFound, "The start index of the cell could not miss.");

            int maxRowSpanForCell = rowsCount - startIndex;

            return maxRowSpanForCell;
        }

        private static void AddCellsToGrid(TableCellDataGrid tableCellDataGrid, TableRow currentRow, ref int currentColumnIndex, int currentRowIndex, int gridSpan, TableCellData tableCellData)
        {
            for (int limit = currentColumnIndex + gridSpan; currentColumnIndex < limit; currentColumnIndex++)
            {
                tableCellDataGrid.EnsureIndexExistsAndSetTableCellData(currentRowIndex, currentColumnIndex, currentRow, tableCellData);
            }
        }

        private TableCellData EnsureIndexExistsAndGetTableCellData(int y, int x, TableRow row)
        {
            this.EnsureIndexExists(y, x, row);
            return this.Data[y].TableCellDatas[x];
        }

        private void EnsureIndexExistsAndSetTableCellData(int y, int x, TableRow row, TableCellData value)
        {
            this.EnsureIndexExists(y, x, row);
            this.Data[y].TableCellDatas[x] = value;
        }

        private void EnsureIndexExists(int y, int x, TableRow row)
        {
            while (y >= this.Data.Count)
            {
                this.Data.Add(new TableRowData(row, new List<TableCellData>()));
            }

            while (x >= this.Data[y].TableCellDatas.Count)
            {
                this.Data[y].TableCellDatas.Add(new TableCellData());
            }
        }
    }
}