﻿using System.Collections.Generic;
using System.IO;
using Telerik.Windows.Documents.Common.FormatProviders;
using Telerik.Windows.Documents.Flow.FormatProviders.Common;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Import;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Validation;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Docx
{
    /// <summary>
    /// Represents Docx format provider.
    /// </summary>
    public class DocxFormatProvider : BinaryFormatProviderBase<RadFlowDocument>
    {
        private static readonly IEnumerable<string> supportedExtensions = new string[] { ".docx" };

        /// <summary>
        /// Initializes a new instance of the <see cref="DocxFormatProvider" /> class.
        /// </summary>
        public DocxFormatProvider()
        {
            this.ExportSettings = new DocxExportSettings();
        }

        /// <summary>
        /// Gets the supported extensions.
        /// </summary>
        /// <value>The supported extensions.</value>
        public override IEnumerable<string> SupportedExtensions
        {
            get
            {
                return DocxFormatProvider.supportedExtensions;
            }
        }

        /// <summary>
        /// Gets a value indicating whether can import.
        /// </summary>
        /// <value>The value indicating whether can import.</value>
        public override bool CanImport
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Gets a value indicating whether can export.
        /// </summary>
        /// <value>The value indicating whether can export.</value>
        public override bool CanExport
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Gets or sets the export settings.
        /// </summary>
        /// <value>The export settings.</value>
        public DocxExportSettings ExportSettings { get; set; }

        /// <summary>
        /// Imports the specified input.
        /// </summary>
        /// <param name="input">The input.</param>
        /// <returns>The imported document.</returns>
        protected override RadFlowDocument ImportOverride(Stream input)
        {
            Guard.ThrowExceptionIfNull(input, "input");

            DocxImporter importer = new DocxImporter();
            RadFlowDocumentImportContext context = new RadFlowDocumentImportContext();
            importer.Import(input, context);

            return context.Document;
        }

        /// <summary>
        /// Exports the specified document.
        /// </summary>
        /// <param name="document"></param>
        /// <param name="output">The output.</param>
        protected override void ExportOverride(RadFlowDocument document, Stream output)
        {
            Guard.ThrowExceptionIfNull(document, "document");
            Guard.ThrowExceptionIfNull(output, "output");

            this.ValidateAndRepairDocument(document);

            using (new RadFlowDocumentLicenseCheck(document))
            {
                DocxExporter exporter = new DocxExporter();
                RadFlowDocumentExportContext context = new RadFlowDocumentExportContext(document, this.ExportSettings);
                exporter.Export(output, context, this.ExportSettings);
            }
        }

        private void ValidateAndRepairDocument(RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");

            switch (this.ExportSettings.InvalidDocumentAction)
            {
                case InvalidDocumentAction.Repair:
                    DocxValidator.Repair(document);
                    break;

                case InvalidDocumentAction.ThrowException:
                    ValidationResult result = DocxValidator.Validate(document);
                    if (result.ResultType == ValidationResultType.Error)
                    {
                        throw new InvalidDocumentException(result.ValidationErrors);
                    }
                    break;

                default:
                    break;
            }
        }
    }
}
