﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Common.Model.Data;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Contexts;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Lists;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Fields;
#if !SILVERLIGHT && !NETSTANDARD
using System;
using Telerik.Windows.Documents.Model.Drawing.Shapes;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Parts;
#endif
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Theming;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Docx.Export
{
    internal class RadFlowDocumentExportContext : IDocxExportContext
    {
        private readonly RadFlowDocument document;
        private readonly Dictionary<IResource, string> resourceToRelationshipIdDictionary;
        private readonly Dictionary<HeaderFooterBase, int> headerFooterToPartNumberDictionary;
        private readonly Dictionary<HeaderFooterBase, string> headerFooterToRelationshipIdDictionary;

        private readonly FieldContext fieldContext;
        private readonly BookmarkContext bookmarkContext;
        private readonly PermissionContext permissionContext;
        private readonly CommentContext commentContext;
        private readonly ListsExportContext listsExportContext;

        public RadFlowDocumentExportContext(RadFlowDocument document, DocxExportSettings settings)
        {
            Guard.ThrowExceptionIfNull(document, "document");
            Guard.ThrowExceptionIfNull(settings, "settings");

            this.document = document;
            this.ExportSettings = settings;
            this.headerFooterToPartNumberDictionary = new Dictionary<HeaderFooterBase, int>();
            this.headerFooterToRelationshipIdDictionary = new Dictionary<HeaderFooterBase, string>();
            this.resourceToRelationshipIdDictionary = new Dictionary<IResource, string>();

            this.fieldContext = new FieldContext(false);
            this.bookmarkContext = new BookmarkContext(this.document);
            this.permissionContext = new PermissionContext(this.document);
            this.commentContext = new CommentContext(this.document);
            this.listsExportContext = new ListsExportContext(this.document);

            this.InitializeHeaders();
            this.InitializeFooters();
        }

        public DocumentTheme Theme
        {
            get
            {
                return this.Document.Theme;
            }
        }

        public RadFlowDocument Document
        {
            get
            {
                return this.document;
            }
        }

        public DocxExportSettings ExportSettings { get; private set; }

        public FieldContext FieldContext
        {
            get
            {
                return this.fieldContext;
            }
        }

        public BookmarkContext BookmarkContext
        {
            get
            {
                return this.bookmarkContext;
            }
        }

        public PermissionContext PermissionContext
        {
            get
            {
                return this.permissionContext;
            }
        }

        public CommentContext CommentContext
        {
            get
            {
                return this.commentContext;
            }
        }

        public ListsExportContext ListsExportContext
        {
            get
            {
                return this.listsExportContext;
            }
        }

        public ResourceManager Resources
        {
            get
            {
                return this.document.Resources;
            }
        }

        public IEnumerable<Section> GetSections()
        {
            return this.document.Sections;
        }

        public IEnumerable<Header> GetHeaders()
        {
            foreach (Section section in this.GetSections())
            {
                if (section.Headers.Even != null)
                {
                    yield return section.Headers.Even;
                }

                if (section.Headers.Default != null)
                {
                    yield return section.Headers.Default;
                }

                if (section.Headers.First != null)
                {
                    yield return section.Headers.First;
                }
            }
        }

        public IEnumerable<Footer> GetFooters()
        {
            foreach (Section section in this.GetSections())
            {
                if (section.Footers.Even != null)
                {
                    yield return section.Footers.Even;
                }

                if (section.Footers.Default != null)
                {
                    yield return section.Footers.Default;
                }

                if (section.Footers.First != null)
                {
                    yield return section.Footers.First;
                }
            }
        }

        public int GetHeaderFooterPartNumberByHeaderFooter(HeaderFooterBase headerFooter)
        {
            Guard.ThrowExceptionIfNull(headerFooter, "headerFooter");

            return this.headerFooterToPartNumberDictionary[headerFooter];
        }

        public string GetRelationshipIdByHeaderFooter(HeaderFooterBase headerFooter)
        {
            Guard.ThrowExceptionIfNull(headerFooter, "headerFooter");

            return this.headerFooterToRelationshipIdDictionary[headerFooter];
        }

        public void RegisterHeaderFooter(string relationshipId, HeaderFooterBase headerFooter)
        {
            Guard.ThrowExceptionIfNull(relationshipId, "relationshipId");
            Guard.ThrowExceptionIfNull(headerFooter, "headerFooter");

            this.headerFooterToRelationshipIdDictionary.Add(headerFooter, relationshipId);
        }

        public string GetRelationshipIdByResource(IResource resource)
        {
            Guard.ThrowExceptionIfNull(resource, "resource");

            return this.resourceToRelationshipIdDictionary[resource];
        }

        public void RegisterResource(string relationshipId, IResource resource)
        {
            Guard.ThrowExceptionIfNullOrEmpty(relationshipId, "relationshipId");
            Guard.ThrowExceptionIfNull(resource, "resource");

            this.resourceToRelationshipIdDictionary[resource] = relationshipId;
        }

#if !SILVERLIGHT && !NETSTANDARD
        public ChartPart GetChartPartForChart(ChartShape chart)
        {
            // Chart functionality is currently not implemented.
            throw new NotImplementedException();
        }

        public ChartShape GetChartForChartPart(ChartPart chartPart)
        {
            // Chart functionality is currently not implemented.
            throw new NotImplementedException();
        }
#endif

        private void InitializeHeaders()
        {
            int headerNumber = 1;
            foreach (Header header in this.GetHeaders())
            {
                this.headerFooterToPartNumberDictionary.Add(header, headerNumber++);
            }
        }

        private void InitializeFooters()
        {
            int footerNumber = 1;
            foreach (Footer footer in this.GetFooters())
            {
                this.headerFooterToPartNumberDictionary.Add(footer, footerNumber++);
            }
        }
    }
}
