﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.FormatProviders.Common;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Contexts;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Import;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Protection;
using Telerik.Windows.Documents.Flow.Model.Watermarks;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Document
{
    internal class ParagraphElement : DocumentElementBase
    {
        private readonly OpenXmlChildElement<ParagraphPropertiesElement> paragraphPropertiesChildElement;
        private Paragraph paragraph;
        private Section section;
        private Watermark watermark;

        public ParagraphElement(DocxPartsManager partsManager)
            : base(partsManager)
        {
            this.paragraphPropertiesChildElement = this.RegisterChildElement<ParagraphPropertiesElement>(DocxElementNames.ParagraphProperties);
        }

        public override string ElementName
        {
            get
            {
                return DocxElementNames.Paragraph;
            }
        }

        public Paragraph Paragraph
        {
            get
            {
                return this.paragraph;
            }
        }

        public Watermark Watermark
        {
            get
            {
                return this.watermark;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "value");
                this.watermark = value;
            }
        }

        private ParagraphPropertiesElement ParagraphPropertiesElement
        {
            get
            {
                return this.paragraphPropertiesChildElement.Element;
            }
        }

        public virtual void SetAssociatedFlowModelElement(Paragraph paragraph, Section section = null)
        {
            Guard.ThrowExceptionIfNull(paragraph, "paragraph");

            this.paragraph = paragraph;
            this.section = section;
        }

        protected override bool ShouldImport(IDocxImportContext context)
        {
            return true;
        }

        protected override bool ShouldExport(IDocxExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            return true;
        }

        protected override void OnBeforeWrite(IDocxExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            if (this.paragraph.Properties.HasLocalValues() || this.paragraph.IsLastInSectionButNotLastInDocument(this.section))
            {
                this.CreateElement(this.paragraphPropertiesChildElement);
                this.ParagraphPropertiesElement.SetAssociatedFlowModelElement(this.paragraph, this.section);
            }
        }

        protected override IEnumerable<OpenXmlElementBase> EnumerateChildElements(IDocxExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            foreach (InlineBase inline in this.paragraph.Inlines)
            {
                switch (inline.Type)
                {
                    case DocumentElementType.BookmarkRangeStart:
                        BookmarkRangeStart bookmarkRangeStart = (BookmarkRangeStart)inline;
                        BookmarkStartElement bookmarkStartElement = this.CreateElement<BookmarkStartElement>(DocxElementNames.BookmarkStart);
                        Bookmark bookmark = bookmarkRangeStart.Bookmark;
                        bookmarkStartElement.CopyPropertiesFrom(bookmark);
                        context.BookmarkContext.RegisterAnnotationToExport(bookmark, bookmarkStartElement.Id);
                        yield return bookmarkStartElement;

                        break;

                    case DocumentElementType.BookmarkRangeEnd:
                        BookmarkRangeEnd bookmarkRangeEnd = (BookmarkRangeEnd)inline;
                        BookmarkEndElement bookmarkEndElement = this.CreateElement<BookmarkEndElement>(DocxElementNames.BookmarkEnd);
                        bookmarkEndElement.Id = context.BookmarkContext.PopIdByRegisteredAnnotation(bookmarkRangeEnd.Bookmark);
                        yield return bookmarkEndElement;

                        break;

                    case DocumentElementType.CommentRangeStart:
                        CommentRangeStart commentRangeStart = (CommentRangeStart)inline;
                        CommentRangeStartElement commentStartElement = this.CreateElement<CommentRangeStartElement>(DocxElementNames.CommentRangeStart);
                        commentStartElement.Id = context.CommentContext.GetIdByRegisteredAnnotation(commentRangeStart.Comment);
                        yield return commentStartElement;

                        break;

                    case DocumentElementType.CommentRangeEnd:
                        CommentRangeEnd commentRangeEnd = (CommentRangeEnd)inline;
                        CommentRangeEndElement commentEndElement = this.CreateElement<CommentRangeEndElement>(DocxElementNames.CommentRangeEnd);
                        commentEndElement.Id = context.CommentContext.GetIdByRegisteredAnnotation(commentRangeEnd.Comment);
                        yield return commentEndElement;

                        RunElement commentReferenceRun = this.CreateElement<RunElement>(DocxElementNames.Run);
                        commentReferenceRun.SetAssociatedFlowModelElement(inline);
                        yield return commentReferenceRun;

                        break;

                    case DocumentElementType.PermissionRangeStart:
                        PermissionRangeStart permissionRangeStart = (PermissionRangeStart)inline;
                        PermissionStartElement permissionStartElement = this.CreateElement<PermissionStartElement>(DocxElementNames.PermissionStart);
                        PermissionRange permission = permissionRangeStart.Permission;
                        permissionStartElement.CopyPropertiesFrom(permission);
                        context.PermissionContext.RegisterAnnotationToExport(permission, permissionStartElement.Id);
                        yield return permissionStartElement;

                        break;

                    case DocumentElementType.PermissionRangeEnd:
                        PermissionRangeEnd permissionRangeEnd = (PermissionRangeEnd)inline;
                        PermissionEndElement permissionEndElement = this.CreateElement<PermissionEndElement>(DocxElementNames.PermissionEnd);
                        permissionEndElement.Id = context.PermissionContext.PopIdByRegisteredAnnotation(permissionRangeEnd.Permission);
                        yield return permissionEndElement;

                        break;

                    default:
                        RunElement runElement = this.CreateElement<RunElement>(DocxElementNames.Run);
                        runElement.SetAssociatedFlowModelElement(inline);

                        if (this.watermark != null)
                        {
                            runElement.Watermark = this.watermark;
                        }

                        yield return runElement;
                        break;
                }
            }
        }

        protected override void OnAfterReadAttributes(IDocxImportContext context)
        {
            this.ProcessHangingAnnotations(context);
        }

        protected override void OnBeforeReadChildElement(IDocxImportContext context, OpenXmlElementBase childElement)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(childElement, "element");

            switch (childElement.ElementName)
            {
                case DocxElementNames.Run:
                    ((RunElement)childElement).SetAssociatedFlowModelElementParent(this.paragraph);
                    break;

                case DocxElementNames.ParagraphProperties:
                    this.ParagraphPropertiesElement.SetAssociatedFlowModelElement(this.paragraph);
                    break;

                case DocxElementNames.Hyperlink:
                case DocxElementNames.SimpleField:
                    ((ParagraphContentElementBase)childElement).SetAssociatedFlowModelElementParent(this.paragraph);
                    break;
            }
        }

        protected override void OnAfterReadChildElement(IDocxImportContext context, OpenXmlElementBase childElement)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(childElement, "element");

            switch (childElement.ElementName)
            {
                case DocxElementNames.BookmarkStart:
                    BookmarkStartElement bookmarkStartElement = (BookmarkStartElement)childElement;
                    Bookmark bookmark = BookmarkContext.CreateBookmark(context.Document, bookmarkStartElement.Name, bookmarkStartElement.ColFirst, bookmarkStartElement.ColLast);
                    this.paragraph.Inlines.Add(bookmark.BookmarkRangeStart);
                    context.BookmarkContext.RegisterAnnotationToImport(bookmarkStartElement.Id, bookmark);

                    break;

                case DocxElementNames.BookmarkEnd:
                    BookmarkRangeEnd bookmarkRangeEnd = context.BookmarkContext.PopRegisteredAnnotationById(((BookmarkEndElement)childElement).Id).BookmarkRangeEnd;
                    this.paragraph.Inlines.Add(bookmarkRangeEnd);

                    break;

                case DocxElementNames.CommentRangeStart:
                    CommentRangeStartElement commentStartElement = (CommentRangeStartElement)childElement;
                    Comment startComment = context.CommentContext.GetImportedCommentById(commentStartElement.Id);
                    if (startComment == null)
                    {
                        startComment = CommentContext.CreateComment(context.Document);
                        context.CommentContext.AddImportedComment(commentStartElement.Id, startComment);
                    }
                    else if (startComment.CommentRangeEnd.Paragraph != null)
                    {
                        CommentRangeStart rangeStart = startComment.CommentRangeStart;
                        rangeStart.Paragraph.Inlines.Remove(rangeStart);
                    }
                    this.paragraph.Inlines.Add(startComment.CommentRangeStart);

                    break;

                case DocxElementNames.CommentRangeEnd:
                    CommentRangeEndElement commentEndElement = (CommentRangeEndElement)childElement;
                    Comment endComment = context.CommentContext.GetImportedCommentById(commentEndElement.Id);
                    if (endComment == null)
                    {
                        endComment = CommentContext.CreateComment(context.Document);
                        context.CommentContext.AddImportedComment(commentEndElement.Id, endComment);
                    }
                    else if (endComment.CommentRangeEnd.Paragraph != null)
                    {
                        CommentRangeEnd rangeEnd = endComment.CommentRangeEnd;
                        rangeEnd.Paragraph.Inlines.Remove(rangeEnd);
                    }
                    this.paragraph.Inlines.Add(endComment.CommentRangeEnd);

                    break;

                case DocxElementNames.PermissionStart:
                    PermissionStartElement permissionStartElement = (PermissionStartElement)childElement;
                    PermissionRange permission = PermissionContext.CreatePermission(context.Document, permissionStartElement);
                    if (permission != null)
                    {
                        this.paragraph.Inlines.Add(permission.Start);
                        context.PermissionContext.RegisterAnnotationToImport(permissionStartElement.Id, permission);
                    }

                    break;

                case DocxElementNames.PermissionEnd:
                    PermissionRangeEnd permissionRangeEnd = context.PermissionContext.PopRegisteredAnnotationById(((PermissionEndElement)childElement).Id).End;
                    if (permissionRangeEnd != null)
                    {
                        this.paragraph.Inlines.Add(permissionRangeEnd);
                    }

                    break;

                case DocxElementNames.Run:
                    RunElement runElement = (RunElement)childElement;
                    if (runElement.Watermark != null)
                    {
                        this.watermark = runElement.Watermark;
                    }

                    break;
            }
        }

        protected override void ClearOverride()
        {
            base.ClearOverride();

            this.section = null;
            this.paragraph = null;
            this.watermark = null;
        }

        private void ProcessHangingAnnotations(IDocxImportContext context)
        {
            this.ProcessHangingBookmarks(context);

            this.ProcessHangingPermissions(context);
        }

        private void ProcessHangingBookmarks(IDocxImportContext context)
        {
            foreach (BookmarkRangeStart rangeStart in context.BookmarkContext.GetHangingBookmarkStarts())
            {
                this.paragraph.Inlines.Add(rangeStart);
            }

            context.BookmarkContext.ClearHangingBookmarkStarts();

            foreach (int rangeEndId in context.BookmarkContext.GetHangingBookmarkEndIds())
            {
                this.paragraph.Inlines.Add(context.BookmarkContext.PopRegisteredAnnotationById(rangeEndId).BookmarkRangeEnd);
            }

            context.BookmarkContext.ClearHangingBookmarkEnds();
        }

        // TODO: refactor this to use one method for both bookmark and permission context.
        private void ProcessHangingPermissions(IDocxImportContext context)
        {
            foreach (PermissionRangeStart rangeStart in context.PermissionContext.GetHangingPermissionStarts())
            {
                this.paragraph.Inlines.Add(rangeStart);
            }

            context.PermissionContext.ClearHangingPermissionStarts();

            foreach (int rangeEndId in context.PermissionContext.GetHangingPermissionEndIds())
            {
                this.paragraph.Inlines.Add(context.PermissionContext.PopRegisteredAnnotationById(rangeEndId).End);
            }

            context.PermissionContext.ClearHangingPermissionEnds();
        }
    }
}