﻿using Telerik.Windows.Documents.Flow.FormatProviders.Common;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Import;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Styles;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Styles.ParagraphPropertiesElements;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Document
{
    internal class ParagraphPropertiesElement : DocumentElementBase
    {
        private readonly OpenXmlChildElement<SectionPropertiesElement> sectionPropertiesChildElement;

        private readonly OpenXmlChildElement<RunPropertiesElement> runPropertiesChildElement;
        private readonly OpenXmlChildElement<NumberingPropertiesElement> numberingPropertiesChildElement;

        private readonly OpenXmlChildElement<FlowDirectionElement> flowDirectionElement;
        private readonly OpenXmlChildElement<AlignmentElement> textAlignmentElement;
        private readonly OpenXmlChildElement<KeepOnOnePageElement> keepOnOnePageElement;
        private readonly OpenXmlChildElement<KeepWithNextParagraphElement> keepWithNextParagraphElement;
        private readonly OpenXmlChildElement<ApplyEastAsianLineBreakingRulesElement> applyEastAsianLineBreakingRulesElement;
        private readonly OpenXmlChildElement<PageBreakBeforeElement> pageBreakBefore;
        private readonly OpenXmlChildElement<OutlineLevelElement> outlineLevel;
        private readonly OpenXmlChildElement<ContextualSpacingElement> contextualSpacing;
        private readonly OpenXmlChildElement<MirrorIndentsElement> mirrorIndents;
        private readonly OpenXmlChildElement<ShadingElement> shadingChildElement;
        private readonly OpenXmlChildElement<IndentationElement> indentation;
        private readonly OpenXmlChildElement<TabStopsElement> tabStopsElement;
        private readonly OpenXmlChildElement<AllowOverflowPunctuationElement> allowOverflowPunctuation;
        private readonly OpenXmlChildElement<SpacingElement> spacing;
        private readonly OpenXmlChildElement<ParagraphBordersElement> borders;
        private readonly OpenXmlChildElement<StyleIdElement> styleIdElement;

        private Paragraph paragraph;
        private ParagraphProperties paragraphProperties;
        private Section section;

        public ParagraphPropertiesElement(DocxPartsManager partsManager)
            : base(partsManager)
        {
            this.styleIdElement = this.RegisterChildElement<StyleIdElement>(DocxElementNames.ParagraphStyleId);
            this.keepWithNextParagraphElement = this.RegisterChildElement<KeepWithNextParagraphElement>(DocxElementNames.KeepWithNextParagraph);
            this.keepOnOnePageElement = this.RegisterChildElement<KeepOnOnePageElement>(DocxElementNames.KeepOnOnePage);
            this.pageBreakBefore = this.RegisterChildElement<PageBreakBeforeElement>(DocxElementNames.PageBreakBefore);
            this.numberingPropertiesChildElement = this.RegisterChildElement<NumberingPropertiesElement>(DocxElementNames.NumberingProperties);
            this.borders = this.RegisterChildElement<ParagraphBordersElement>(DocxElementNames.ParagraphBorders);
            this.shadingChildElement = this.RegisterChildElement<ShadingElement>(DocxElementNames.Shading);
            this.tabStopsElement = this.RegisterChildElement<TabStopsElement>(DocxElementNames.TabStops);
            this.applyEastAsianLineBreakingRulesElement = this.RegisterChildElement<ApplyEastAsianLineBreakingRulesElement>(DocxElementNames.ApplyEastAsianLineBreakingRules);
            this.allowOverflowPunctuation = this.RegisterChildElement<AllowOverflowPunctuationElement>(DocxElementNames.AllowOverflowPunctuation);
            this.flowDirectionElement = this.RegisterChildElement<FlowDirectionElement>(DocxElementNames.FlowDirection);
            this.spacing = this.RegisterChildElement<SpacingElement>(DocxElementNames.Spacing);
            this.indentation = this.RegisterChildElement<IndentationElement>(DocxElementNames.Indentation);
            this.contextualSpacing = this.RegisterChildElement<ContextualSpacingElement>(DocxElementNames.ContextualSpacing);
            this.mirrorIndents = this.RegisterChildElement<MirrorIndentsElement>(DocxElementNames.MirrorIndents);
            this.textAlignmentElement = this.RegisterChildElement<AlignmentElement>(DocxElementNames.Alignment);
            this.outlineLevel = this.RegisterChildElement<OutlineLevelElement>(DocxElementNames.OutlineLevel);

            this.runPropertiesChildElement = this.RegisterChildElement<RunPropertiesElement>(DocxElementNames.RunProperties);
            this.sectionPropertiesChildElement = this.RegisterChildElement<SectionPropertiesElement>(DocxElementNames.SectionProperties);
        }

        public override string ElementName
        {
            get
            {
                return DocxElementNames.ParagraphProperties;
            }
        }

        public ParagraphProperties ParagraphProperties
        {
            get
            {
                return this.paragraphProperties;
            }
        }

        private SectionPropertiesElement SectionPropertiesElement
        {
            get
            {
                return this.sectionPropertiesChildElement.Element;
            }
        }

        private RunPropertiesElement RunPropertiesElement
        {
            get
            {
                return this.runPropertiesChildElement.Element;
            }
        }

        private NumberingPropertiesElement NumberingPropertiesElement
        {
            get
            {
                return this.numberingPropertiesChildElement.Element;
            }
        }

        private ParagraphBordersElement ParagraphBordersElement
        {
            get
            {
                return this.borders.Element;
            }
        }

        private TabStopsElement TabStopsElement
        {
            get
            {
                return this.tabStopsElement.Element;
            }
        }

        public void SetAssociatedFlowModelElement(Paragraph paragraph, Section section = null)
        {
            Guard.ThrowExceptionIfNull(paragraph, "paragraph");

            this.paragraph = paragraph;
            this.paragraphProperties = paragraph.Properties;
            this.section = section;
        }

        public void SetAssociatedFlowModelElement(ParagraphProperties paragraphProperties)
        {
            Guard.ThrowExceptionIfNull(paragraphProperties, "paragraphProperties");

            this.paragraphProperties = paragraphProperties;
        }

        protected override bool ShouldExport(IDocxExportContext context)
        {
            return true;
        }

        protected override void OnAfterRead(IDocxImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            if (this.flowDirectionElement.Element != null)
            {
                this.ParagraphProperties.FlowDirection.LocalValue = this.flowDirectionElement.Element.Value ? FlowDirection.RightToLeft : FlowDirection.LeftToRight;
                this.ReleaseElement(this.flowDirectionElement);
            }

            if (this.textAlignmentElement.Element != null)
            {
                this.ParagraphProperties.TextAlignment.LocalValue = this.textAlignmentElement.Element.Value;
                this.ReleaseElement(this.textAlignmentElement);
            }

            if (this.spacing.Element != null)
            {
                this.spacing.Element.ReadAttributes(this.ParagraphProperties);
                this.ReleaseElement(this.spacing);
            }

            if (this.keepOnOnePageElement.Element != null)
            {
                this.ParagraphProperties.KeepOnOnePage.LocalValue = this.keepOnOnePageElement.Element.Value;
                this.ReleaseElement(this.keepOnOnePageElement);
            }

            if (this.keepWithNextParagraphElement.Element != null)
            {
                this.ParagraphProperties.KeepWithNextParagraph.LocalValue = this.keepWithNextParagraphElement.Element.Value;
                this.ReleaseElement(this.keepWithNextParagraphElement);
            }

            if (this.applyEastAsianLineBreakingRulesElement.Element != null)
            {
                this.ParagraphProperties.ApplyEastAsianLineBreakingRules.LocalValue = this.applyEastAsianLineBreakingRulesElement.Element.Value;
                this.ReleaseElement(this.applyEastAsianLineBreakingRulesElement);
            }

            if (this.pageBreakBefore.Element != null)
            {
                this.ParagraphProperties.PageBreakBefore.LocalValue = this.pageBreakBefore.Element.Value;
                this.ReleaseElement(this.pageBreakBefore);
            }

            if (this.outlineLevel.Element != null)
            {
                this.ParagraphProperties.OutlineLevel.LocalValue = this.outlineLevel.Element.Value;
                this.ReleaseElement(this.outlineLevel);
            }

            if (this.contextualSpacing.Element != null)
            {
                this.ParagraphProperties.ContextualSpacing.LocalValue = this.contextualSpacing.Element.Value;
                this.ReleaseElement(this.contextualSpacing);
            }

            if (this.mirrorIndents.Element != null)
            {
                this.ParagraphProperties.MirrorIndents.LocalValue = this.mirrorIndents.Element.Value;
                this.ReleaseElement(this.mirrorIndents);
            }

            if (this.TabStopsElement != null)
            {
                this.ReleaseElement(this.tabStopsElement);
            }

            if (this.shadingChildElement.Element != null)
            {
                this.shadingChildElement.Element.ReadAttributes(this.ParagraphProperties);
                this.ReleaseElement(this.shadingChildElement);
            }

            if (this.indentation.Element != null)
            {
                this.indentation.Element.ReadAttributes(this.ParagraphProperties);
                this.ReleaseElement(this.indentation);
            }

            if (this.allowOverflowPunctuation.Element != null)
            {
                this.ParagraphProperties.AllowOverflowPunctuation.LocalValue = this.allowOverflowPunctuation.Element.Value;
                this.ReleaseElement(this.allowOverflowPunctuation);
            }

            if (this.numberingPropertiesChildElement.Element != null)
            {
                this.ReleaseElement(this.numberingPropertiesChildElement);
            }

            if (this.ParagraphBordersElement != null)
            {
                this.ReleaseElement(this.borders);
            }

            if (this.styleIdElement.Element != null)
            {
                this.ParagraphProperties.StyleId = this.styleIdElement.Element.Value;
                this.ReleaseElement(this.styleIdElement);
            }
        }

        protected override void OnBeforeWrite(IDocxExportContext context)
        {
            this.CreateSectionPropertiesElement();

            if (this.ParagraphProperties.FlowDirection.HasLocalValue)
            {
                this.CreateElement(this.flowDirectionElement);
                FlowDirection flowDirectionValue = this.ParagraphProperties.FlowDirection.LocalValue.Value;
                this.flowDirectionElement.Element.Value = flowDirectionValue == FlowDirection.RightToLeft ? true : false;
            }

            if (this.ParagraphProperties.TextAlignment.HasLocalValue)
            {
                this.CreateElement(this.textAlignmentElement);
                this.textAlignmentElement.Element.Value = this.ParagraphProperties.TextAlignment.LocalValue.Value;
            }

            if (this.ParagraphProperties.SpacingAfter.HasLocalValue ||
                this.ParagraphProperties.SpacingBefore.HasLocalValue ||
                this.ParagraphProperties.AutomaticSpacingAfter.HasLocalValue ||
                this.ParagraphProperties.AutomaticSpacingBefore.HasLocalValue ||
                this.ParagraphProperties.LineSpacing.HasLocalValue ||
                this.ParagraphProperties.LineSpacingType.HasLocalValue)
            {
                this.CreateElement(this.spacing);
                this.spacing.Element.FillAttributes(this.ParagraphProperties);
            }

            if (this.ParagraphProperties.KeepOnOnePage.HasLocalValue)
            {
                this.CreateElement(this.keepOnOnePageElement);
                this.keepOnOnePageElement.Element.Value = this.ParagraphProperties.KeepOnOnePage.LocalValue.Value;
            }

            if (this.ParagraphProperties.KeepWithNextParagraph.HasLocalValue)
            {
                this.CreateElement(this.keepWithNextParagraphElement);
                this.keepWithNextParagraphElement.Element.Value = this.ParagraphProperties.KeepWithNextParagraph.LocalValue.Value;
            }

            if (this.ParagraphProperties.ApplyEastAsianLineBreakingRules.HasLocalValue)
            {
                this.CreateElement(this.applyEastAsianLineBreakingRulesElement);
                this.applyEastAsianLineBreakingRulesElement.Element.Value = this.ParagraphProperties.ApplyEastAsianLineBreakingRules.LocalValue.Value;
            }

            if (this.ParagraphProperties.PageBreakBefore.HasLocalValue)
            {
                this.CreateElement(this.pageBreakBefore);
                this.pageBreakBefore.Element.Value = this.ParagraphProperties.PageBreakBefore.LocalValue.Value;
            }

            if (this.ParagraphProperties.OutlineLevel.HasLocalValue)
            {
                this.CreateElement(this.outlineLevel);
                this.outlineLevel.Element.Value = this.ParagraphProperties.OutlineLevel.LocalValue.Value;
            }

            if (this.ParagraphProperties.ContextualSpacing.HasLocalValue)
            {
                this.CreateElement(this.contextualSpacing);
                this.contextualSpacing.Element.Value = this.ParagraphProperties.ContextualSpacing.LocalValue.Value;
            }

            if (this.ParagraphProperties.MirrorIndents.HasLocalValue)
            {
                this.CreateElement(this.mirrorIndents);
                this.mirrorIndents.Element.Value = this.ParagraphProperties.MirrorIndents.LocalValue.Value;
            }

            if (this.ParagraphProperties.ShadingPattern.HasLocalValue ||
                this.ParagraphProperties.ShadingPatternColor.HasLocalValue ||
                this.ParagraphProperties.BackgroundColor.HasLocalValue)
            {
                this.CreateElement(this.shadingChildElement);
                this.shadingChildElement.Element.FillAttributes(this.ParagraphProperties);
            }

            if (this.ParagraphProperties.FirstLineIndent.HasLocalValue || this.ParagraphProperties.HangingIndent.HasLocalValue ||
                this.ParagraphProperties.LeftIndent.HasLocalValue || this.ParagraphProperties.RightIndent.HasLocalValue)
            {
                this.CreateElement(this.indentation);
                this.indentation.Element.FillAttributes(this.ParagraphProperties);
            }

            if (this.ParagraphProperties.TabStops.HasLocalValue)
            {
                this.CreateElement(this.tabStopsElement);
                this.TabStopsElement.SetAssociatedFlowModelElement(this.ParagraphProperties);
            }

            if (!this.ParagraphProperties.AllowOverflowPunctuation.LocalValue.Value.Equals(DocumentDefaultStyleSettings.AllowOverflowPunctuation))
            {
                this.CreateElement(this.allowOverflowPunctuation);
                this.allowOverflowPunctuation.Element.Value = this.ParagraphProperties.AllowOverflowPunctuation.LocalValue.Value;
            }

            if (this.ParagraphProperties.ListId.HasLocalValue || this.ParagraphProperties.ListLevel.HasLocalValue)
            {
                this.CreateElement(this.numberingPropertiesChildElement);
                this.NumberingPropertiesElement.SetAssociatedFlowModelElement(this.ParagraphProperties);
            }

            if (this.ParagraphProperties.Borders.HasLocalValue)
            {
                this.CreateElement(this.borders);
                this.ParagraphBordersElement.SetAssociatedFlowModelElement(this.ParagraphProperties);
            }

            if (!string.IsNullOrEmpty(this.ParagraphProperties.StyleId))
            {
                this.CreateElement(this.styleIdElement);
                this.styleIdElement.Element.Value = this.ParagraphProperties.StyleId;
            }

            if (this.ParagraphProperties.ParagraphMarkerProperties != null &&
                this.ParagraphProperties.OwnerDocumentElement.GetType() != typeof(RadFlowDocument) &&
                this.ParagraphProperties.ParagraphMarkerProperties.HasLocalValues())
            {
                this.CreateElement(this.runPropertiesChildElement);
                this.RunPropertiesElement.SetAssociatedFlowModelElement(this.ParagraphProperties.ParagraphMarkerProperties);
            }
        }

        protected override void OnBeforeReadChildElement(IDocxImportContext context, OpenXmlElementBase element)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            switch (element.ElementName)
            {
                case DocxElementNames.SectionProperties:
                    this.SectionPropertiesElement.SetAssociatedFlowModelElement(context.GetCurrentSection());
                    break;
                case DocxElementNames.RunProperties:
                    this.RunPropertiesElement.SetAssociatedFlowModelElement(this.ParagraphProperties.ParagraphMarkerProperties);
                    break;
                case DocxElementNames.NumberingProperties:
                    this.NumberingPropertiesElement.SetAssociatedFlowModelElement(this.ParagraphProperties);
                    break;
                case DocxElementNames.ParagraphBorders:
                    this.ParagraphBordersElement.SetAssociatedFlowModelElement(this.ParagraphProperties);
                    break;
                case DocxElementNames.TabStops:
                    this.TabStopsElement.SetAssociatedFlowModelElement(this.ParagraphProperties);
                    break;
            }
        }

        protected override void ClearOverride()
        {
            base.ClearOverride();

            this.paragraph = null;
            this.paragraphProperties = null;
            this.section = null;
        }

        private void CreateSectionPropertiesElement()
        {
            if (this.paragraph != null &&
                this.paragraph.IsLastInSectionButNotLastInDocument(this.section))
            {
                this.CreateElement(this.sectionPropertiesChildElement);

                this.SectionPropertiesElement.SetAssociatedFlowModelElement(this.section);
            }
        }
    }
}
