﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Contexts;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Import;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Fields;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Flow.Model.Watermarks;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Document
{
    internal class RunElement : DocumentElementBase
    {
        private readonly OpenXmlChildElement<RunPropertiesElement> runPropertiesChildElement;

        private InlineBase inline;
        private Run run;
        private Paragraph paragraph;
        private Watermark watermark;

        public RunElement(DocxPartsManager partsManager)
            : base(partsManager)
        {
            this.runPropertiesChildElement = this.RegisterChildElement<RunPropertiesElement>(DocxElementNames.RunProperties);
        }

        public override string ElementName
        {
            get
            {
                return DocxElementNames.Run;
            }
        }

        public override bool AlwaysExport
        {
            get
            {
                return true;
            }
        }

        public RunPropertiesElement RunPropertiesElement
        {
            get
            {
                return this.runPropertiesChildElement.Element;
            }
        }

        public Watermark Watermark
        {
            get
            {
                return this.watermark;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "value");
                this.watermark = value;
            }
        }

        public void SetAssociatedFlowModelElement(InlineBase inline)
        {
            Guard.ThrowExceptionIfNull(inline, "inline");

            this.inline = inline;
        }

        public void SetAssociatedFlowModelElementParent(Paragraph paragraph)
        {
            Guard.ThrowExceptionIfNull(paragraph, "paragraph");

            this.paragraph = paragraph;
        }

        protected override void ClearOverride()
        {
            this.paragraph = null;
            this.inline = null;
            this.run = null;
            this.watermark = null;
        }

        protected override bool ShouldImport(IDocxImportContext context)
        {
            // The Paragraph will be null if the Run is a direct child of a StructuredDocumentContent.
            return base.ShouldImport(context) || this.paragraph != null;
        }

        protected override void OnAfterRead(IDocxImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            if (this.RunPropertiesElement != null)
            {
                if (this.run != null)
                {
                    this.RunPropertiesElement.CopyPropertiesTo(this.run.Properties);
                }

                this.ReleaseElement(this.runPropertiesChildElement);
            }
        }

        protected override void OnBeforeWrite(IDocxExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            switch (this.inline.Type)
            {
                case DocumentElementType.Run:

                    Run run = (Run)this.inline;
                    if (run.Properties.HasLocalValues())
                    {
                        this.CreateElement(this.runPropertiesChildElement);
                        this.RunPropertiesElement.SetAssociatedFlowModelElement(run.Properties);
                    }

                    break;
            }
        }

        protected override IEnumerable<OpenXmlElementBase> EnumerateChildElements(IDocxExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            Run run = this.inline as Run;
            if (run != null && run.Text != null)
            {
                string[] texts = run.Text.Split('\t');
                bool isFirst = true;
                foreach (string txt in texts)
                {
                    if (!isFirst)
                    {
                        yield return this.CreateElement<TabElement>(DocxElementNames.Tab);
                    }

                    if (!string.IsNullOrEmpty(txt))
                    {
                        DocumentElementBase textElement;
                        if (context.FieldContext.IsInInstruction)
                        {
                            textElement = this.CreateElement<InstructionTextElement>(DocxElementNames.InstructionText);
                        }
                        else
                        {
                            textElement = this.CreateElement<TextElement>(DocxElementNames.Text);
                        }
                        textElement.InnerText = txt;

                        yield return textElement;
                    }

                    isFirst = false;
                }
            }

            ShapeInlineBase shape = this.inline as ShapeInlineBase;
            if (shape != null)
            {
                DrawingElement drawingElement = this.CreateElement<DrawingElement>(DocxElementNames.Drawing);
                drawingElement.CopyPropertiesFrom(context, shape);

                yield return drawingElement;
            }

            ShapeAnchorBase anchor = this.inline as ShapeAnchorBase;
            if (anchor != null)
            {
                DrawingElement drawingElement = this.CreateElement<DrawingElement>(DocxElementNames.Drawing);
                drawingElement.CopyPropertiesFrom(context, anchor);

                yield return drawingElement;
            }

            FieldCharacter fieldChar = this.inline as FieldCharacter;
            if (fieldChar != null)
            {
                context.FieldContext.OnFieldCharacter(fieldChar);

                FieldCharacterElement fieldCharElement = this.CreateElement<FieldCharacterElement>(DocxElementNames.FieldCharacter);
                fieldCharElement.FieldCharacterType = fieldChar.FieldCharacterType;
                if (fieldChar.FieldCharacterType == FieldCharacterType.Start && fieldChar.FieldInfo != null)
                {
                    fieldCharElement.IsLocked = fieldChar.FieldInfo.IsLocked;
                    fieldCharElement.IsDirty = fieldChar.FieldInfo.IsDirty;
                }

                yield return fieldCharElement;
            }

            Break br = this.inline as Break;
            if (br != null)
            {
                BreakElement breakElement = this.CreateElement<BreakElement>(DocxElementNames.Break);
                breakElement.CopyPropertiesFrom(context, br);

                yield return breakElement;
            }

            CommentRangeEnd commentEnd = this.inline as CommentRangeEnd;
            if (commentEnd != null)
            {
                CommentReferenceElement refElement = this.CreateElement<CommentReferenceElement>(DocxElementNames.CommentReference);
                refElement.Id = context.CommentContext.GetIdByRegisteredAnnotation(commentEnd.Comment);

                yield return refElement;
            }

            if (this.watermark != null)
            {
                VmlContainerElement pictElement = this.CreateElement<VmlContainerElement>(DocxElementNames.VmlContainer);
                pictElement.Watermark = this.watermark;

                yield return pictElement;
            }
        }

        protected override void OnAfterReadChildElement(IDocxImportContext context, OpenXmlElementBase element)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(element, "element");

            switch (element.ElementName)
            {
                case DocxElementNames.Text:
                case DocxElementNames.InstructionText:

                    if (this.run == null)
                    {
                        this.run = new Run(context.Document);
                        this.run.Text = string.Empty;
                        this.paragraph.Inlines.Add(this.run);
                    }

                    this.run.Text += element.InnerText;

                    break;

                case DocxElementNames.Tab:

                    if (this.run == null)
                    {
                        this.run = new Run(context.Document);
                        this.paragraph.Inlines.Add(this.run);
                    }

                    this.run.Text += "\t";

                    break;

                case DocxElementNames.Drawing:

                    InlineBase resultInline = ((DrawingElement)element).CreateInline(context);
                    if (resultInline != null)
                    {
                        this.paragraph.Inlines.Add(resultInline);
                    }

                    break;

                case DocxElementNames.FieldCharacter:
                    FieldCharacterElement fieldCharEl = (FieldCharacterElement)element;
                    FieldCharacter fieldChar = new FieldCharacter(context.Document, fieldCharEl.FieldCharacterType);

                    this.paragraph.Inlines.Add(fieldChar);
                    context.FieldContext.OnFieldCharacter(fieldChar);
                    if (fieldChar.FieldCharacterType == FieldCharacterType.Start)
                    {
                        context.FieldContext.SetIsDirty(fieldCharEl.IsDirty);
                        context.FieldContext.SetIsLocked(fieldCharEl.IsLocked);
                    }

                    break;

                case DocxElementNames.Break:
                    BreakElement breakElement = (BreakElement)element;
                    Break br = new Break(context.Document);
                    br.BreakType = breakElement.Type;
                    br.TextWrappingRestartLocation = breakElement.TextWrappingRestartLocation;
                    this.paragraph.Inlines.Add(br);

                    break;

                case DocxElementNames.CommentReference:
                    CommentReferenceElement commentReferenceElement = (CommentReferenceElement)element;
                    Comment comment = context.CommentContext.GetImportedCommentById(commentReferenceElement.Id);
                    if (comment != null)
                    {
                        if (comment.CommentRangeEnd.Paragraph == null)
                        {
                            this.paragraph.Inlines.Add(comment.CommentRangeEnd);
                        }
                    }
                    else
                    {
                        comment = CommentContext.CreateComment(context.Document);
                        this.paragraph.Inlines.Add(comment.CommentRangeStart);
                        this.paragraph.Inlines.Add(comment.CommentRangeEnd);
                        context.CommentContext.AddImportedComment(commentReferenceElement.Id, comment);
                    }
                    context.CommentContext.AddIdToCommentsPartImport(commentReferenceElement.Id);
                    break;

                case DocxElementNames.VmlContainer:
                    VmlContainerElement vmlContainerElement = (VmlContainerElement)element;
                    if (vmlContainerElement.Watermark != null)
                    {
                        this.watermark = vmlContainerElement.Watermark;
                    }
                    else if (vmlContainerElement.Image != null)
                    {
                        this.paragraph.Inlines.Add(vmlContainerElement.Image);
                    }
                    break;
            }
        }
    }
}