﻿using System.Collections.Generic;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Primitives;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Import;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Common;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements;
using Telerik.Windows.Documents.Model;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Document
{
    internal class SectionPropertiesElement : DocxElementBase
    {
        private readonly OpenXmlChildElement<TitlePageElement> titlePageChildElement;
        private readonly OpenXmlChildElement<SectionTypeElement> sectionTypeChildElement;
        private readonly OpenXmlChildElement<PageMarginsElement> pageMarginsChildElement;
        private readonly OpenXmlChildElement<PageSizeElement> pageSizeChildElement;
        private readonly OpenXmlChildElement<VerticalAlignmentElement> verticalAlignmentElement;
        private readonly OpenXmlChildElement<PageNumberingSettingsElement> pageNumberingSettingsElement;

        private Section section;

        public SectionPropertiesElement(DocxPartsManager partsManager)
            : base(partsManager)
        {
            this.sectionTypeChildElement = this.RegisterChildElement<SectionTypeElement>(DocxElementNames.SectionType);
            this.pageSizeChildElement = this.RegisterChildElement<PageSizeElement>(DocxElementNames.PageSize);
            this.pageMarginsChildElement = this.RegisterChildElement<PageMarginsElement>(DocxElementNames.PageMargins);
            this.pageNumberingSettingsElement = this.RegisterChildElement<PageNumberingSettingsElement>(DocxElementNames.PageNumberingSettings);
            this.verticalAlignmentElement = this.RegisterChildElement<VerticalAlignmentElement>(DocxElementNames.VerticalAlignment);
            this.titlePageChildElement = this.RegisterChildElement<TitlePageElement>(DocxElementNames.TitlePage);
        }

        public override string ElementName
        {
            get
            {
                return DocxElementNames.SectionProperties;
            }
        }

        private TitlePageElement TitlePageElement
        {
            get
            {
                return this.titlePageChildElement.Element;
            }
        }

        private SectionTypeElement SectionTypeElement
        {
            get
            {
                return this.sectionTypeChildElement.Element;
            }
        }

        private PageMarginsElement PageMarginsElement
        {
            get
            {
                return this.pageMarginsChildElement.Element;
            }
        }

        private PageSizeElement PageSizeElement
        {
            get
            {
                return this.pageSizeChildElement.Element;
            }
        }

        private VerticalAlignmentElement VerticalAlignmentElement
        {
            get
            {
                return this.verticalAlignmentElement.Element;
            }
        }

        private PageNumberingSettingsElement PageNumberingSettingsElement
        {
            get
            {
                return this.pageNumberingSettingsElement.Element;
            }
        }

        public void SetAssociatedFlowModelElement(Section section)
        {
            Guard.ThrowExceptionIfNull(section, "section");

            this.section = section;
        }

        protected override bool ShouldExport(IDocxExportContext context)
        {
            return true;
        }

        protected override void OnAfterReadChildElement(IDocxImportContext context, OpenXmlElementBase childElement)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(childElement, "childElement");

            switch (childElement.ElementName)
            {
                case DocxElementNames.HeaderReference:
                    HeaderReferenceElement headerReferenceElement = (HeaderReferenceElement)childElement;
                    Header header = this.section.Headers.Add(headerReferenceElement.HeaderFooterType);
                    context.RegisterHeader(header, headerReferenceElement.RelationshipId);
                    break;

                case DocxElementNames.FooterReference:
                    FooterReferenceElement footerReferenceElement = (FooterReferenceElement)childElement;
                    Footer footer = this.section.Footers.Add(footerReferenceElement.HeaderFooterType);
                    context.RegisterFooter(footer, footerReferenceElement.RelationshipId);
                    break;

                default:
                    break;
            }
        }

        protected override void OnAfterRead(IDocxImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            if (this.TitlePageElement != null)
            {
                this.section.Properties.HasDifferentFirstPageHeaderFooter.LocalValue = this.TitlePageElement.Value;
                this.ReleaseElement(this.titlePageChildElement);
            }

            if (this.SectionTypeElement != null)
            {
                this.section.Properties.SectionType.LocalValue = this.SectionTypeElement.SectionType;
                this.ReleaseElement(this.sectionTypeChildElement);
            }

            if (this.PageMarginsElement != null)
            {
                Padding padding = new Padding(this.PageMarginsElement.Left, this.PageMarginsElement.Top, this.PageMarginsElement.Right, this.PageMarginsElement.Bottom);
                this.section.Properties.PageMargins.LocalValue = padding;
                this.section.Properties.HeaderTopMargin.LocalValue = this.PageMarginsElement.Header;
                this.section.Properties.FooterBottomMargin.LocalValue = this.PageMarginsElement.Footer;

                this.ReleaseElement(this.pageMarginsChildElement);
            }

            if (this.PageSizeElement != null)
            {
                Size size = new Size(this.PageSizeElement.Width, this.PageSizeElement.Height);
                this.section.Properties.PageSize.LocalValue = size;
                this.section.Properties.PageOrientation.LocalValue = this.PageSizeElement.PageOrientation;

                this.ReleaseElement(this.pageSizeChildElement);
            }

            if (this.VerticalAlignmentElement != null)
            {
                this.section.Properties.VerticalAlignment.LocalValue = this.VerticalAlignmentElement.Value;
                this.ReleaseElement(this.verticalAlignmentElement);
            }

            if (this.PageNumberingSettingsElement != null)
            {
                this.PageNumberingSettingsElement.CopyPropertiesTo(this.section.PageNumberingSettings);

                this.ReleaseElement(this.pageNumberingSettingsElement);
            }

            context.CloseCurrentSection();
        }

        protected override void OnBeforeWrite(IDocxExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            bool localHasDifferentFirstPageHeaderFooter = this.section.Properties.HasDifferentFirstPageHeaderFooter.LocalValue.Value;
            if (!localHasDifferentFirstPageHeaderFooter.Equals(DocumentDefaultStyleSettings.HasDifferentFirstPageHeaderFooter))
            {
                this.CreateElement(this.titlePageChildElement);
                this.TitlePageElement.Value = localHasDifferentFirstPageHeaderFooter;
            }

            SectionType localSectionTypeValue = this.section.Properties.SectionType.LocalValue.Value;
            if (!localSectionTypeValue.Equals(DocumentDefaultStyleSettings.SectionType))
            {
                this.CreateElement(this.sectionTypeChildElement);
                this.SectionTypeElement.SectionType = localSectionTypeValue;
            }

            Padding localPageMargins = this.section.Properties.PageMargins.LocalValue;
            double localHeaderTopMargin = this.section.Properties.HeaderTopMargin.LocalValue.Value;
            double localFooterBottomMargin = this.section.Properties.FooterBottomMargin.LocalValue.Value;
            if (localPageMargins != DocumentDefaultStyleSettings.PageMargins ||
                localHeaderTopMargin != DocumentDefaultStyleSettings.SectionHeaderTopMargin ||
                localFooterBottomMargin != DocumentDefaultStyleSettings.SectionFooterBottomMargin)
            {
                this.CreateElement(this.pageMarginsChildElement);
                this.PageMarginsElement.Top = localPageMargins.Top;
                this.PageMarginsElement.Right = localPageMargins.Right;
                this.PageMarginsElement.Bottom = localPageMargins.Bottom;
                this.PageMarginsElement.Left = localPageMargins.Left;
                this.PageMarginsElement.Gutter = 0; // Not supported in the model.
                this.PageMarginsElement.Header = localHeaderTopMargin;
                this.PageMarginsElement.Footer = localFooterBottomMargin;
            }

            Size localPageSize = this.section.Properties.PageSize.LocalValue.Value;
            PageOrientation localPageOrientation = this.section.Properties.PageOrientation.LocalValue.Value;
            if (localPageSize != DocumentDefaultStyleSettings.PageSize ||
                localPageOrientation != DocumentDefaultStyleSettings.PageOrientation)
            {
                this.CreateElement(this.pageSizeChildElement);
                this.PageSizeElement.Width = localPageSize.Width;
                this.PageSizeElement.Height = localPageSize.Height;
                this.PageSizeElement.PageOrientation = localPageOrientation;
            }

            Flow.Model.Styles.VerticalAlignment localVerticalAlignment = this.section.Properties.VerticalAlignment.LocalValue;
            if (localVerticalAlignment != DocumentDefaultStyleSettings.VerticalAlignment)
            {
                this.CreateElement(this.verticalAlignmentElement);
                this.VerticalAlignmentElement.Value = localVerticalAlignment;
            }

            PageNumberingSettings localPageNumberingSettings = this.section.PageNumberingSettings;
            if (localPageNumberingSettings.ChapterSeparatorCharacter.HasValue ||
                localPageNumberingSettings.ChapterHeadingStyleIndex.HasValue ||
                localPageNumberingSettings.PageNumberFormat.HasValue ||
                localPageNumberingSettings.StartingPageNumber.HasValue)
            {
                this.CreateElement(this.pageNumberingSettingsElement);
                this.PageNumberingSettingsElement.CopyPropertiesFrom(localPageNumberingSettings);
            }
        }

        protected override IEnumerable<OpenXmlElementBase> EnumerateChildElements(IDocxExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            return this.CreateHeaderFooterReferenceElements(context);
        }

        protected override void ClearOverride()
        {
            base.ClearOverride();

            this.section = null;
        }

        private IEnumerable<HeaderFooterReferenceElementBase> CreateHeaderFooterReferenceElements(IDocxExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            HeaderFooterBase headerFooter;

            headerFooter = this.section.Headers.Even;
            if (headerFooter != null)
            {
                HeaderReferenceElement headerReference = this.CreateHeaderFooterReferenceElement<HeaderReferenceElement>(context, DocxElementNames.HeaderReference, headerFooter, HeaderFooterType.Even);
                yield return headerReference;
            }

            headerFooter = this.section.Headers.Default;
            if (headerFooter != null)
            {
                HeaderReferenceElement headerReference = this.CreateHeaderFooterReferenceElement<HeaderReferenceElement>(context, DocxElementNames.HeaderReference, headerFooter, HeaderFooterType.Default);
                yield return headerReference;
            }

            headerFooter = this.section.Headers.First;
            if (headerFooter != null)
            {
                HeaderReferenceElement headerReference = this.CreateHeaderFooterReferenceElement<HeaderReferenceElement>(context, DocxElementNames.HeaderReference, headerFooter, HeaderFooterType.First);
                yield return headerReference;
            }

            headerFooter = this.section.Footers.Even;
            if (headerFooter != null)
            {
                FooterReferenceElement footerReference = this.CreateHeaderFooterReferenceElement<FooterReferenceElement>(context, DocxElementNames.FooterReference, headerFooter, HeaderFooterType.Even);
                yield return footerReference;
            }

            headerFooter = this.section.Footers.Default;
            if (headerFooter != null)
            {
                FooterReferenceElement footerReference = this.CreateHeaderFooterReferenceElement<FooterReferenceElement>(context, DocxElementNames.FooterReference, headerFooter, HeaderFooterType.Default);
                yield return footerReference;
            }

            headerFooter = this.section.Footers.First;
            if (headerFooter != null)
            {
                FooterReferenceElement footerReference = this.CreateHeaderFooterReferenceElement<FooterReferenceElement>(context, DocxElementNames.FooterReference, headerFooter, HeaderFooterType.First);
                yield return footerReference;
            }
        }

        private T CreateHeaderFooterReferenceElement<T>(IDocxExportContext context, string elementName, HeaderFooterBase headerFooter, HeaderFooterType headerFooterType) where T : HeaderFooterReferenceElementBase
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(elementName, "elementName");
            Guard.ThrowExceptionIfNull(headerFooter, "headerFooter");

            T headerFooterReferenceElement = this.CreateElement<T>(elementName);
            headerFooterReferenceElement.HeaderFooterType = headerFooterType;
            headerFooterReferenceElement.RelationshipId = context.GetRelationshipIdByHeaderFooter(headerFooter);

            return headerFooterReferenceElement;
        }
    }
}
