﻿using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Flow.FormatProviders.Common.Tables;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Import;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Utilities;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Document
{
    internal class TableElement : DocumentElementBase
    {
        private readonly OpenXmlChildElement<TablePropertiesElement> tablePropertiesElementChildElement;
        private Table table;

        private List<TableRowData> tableRowDatas;

        public TableElement(DocxPartsManager partsManager)
            : base(partsManager)
        {
            this.tablePropertiesElementChildElement = this.RegisterChildElement<TablePropertiesElement>(DocxElementNames.TableProperties);
        }

        public override string ElementName
        {
            get
            {
                return DocxElementNames.Table;
            }
        }

        private TablePropertiesElement TablePropertiesElement
        {
            get
            {
                return this.tablePropertiesElementChildElement.Element;
            }
        }

        public void SetAssociatedFlowModelElement(Table table)
        {
            Guard.ThrowExceptionIfNull(table, "table");

            this.table = table;
            this.tableRowDatas = new List<TableRowData>();
        }

        protected override bool ShouldExport(IDocxExportContext context)
        {
            return this.table.Rows.Count > 0 || base.ShouldExport(context);
        }

        protected override IEnumerable<OpenXmlElementBase> EnumerateChildElements(IDocxExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            TableCellDataGrid tableCellDataGrid = TableCellDataGrid.CreateFromTable(this.table);
            IEnumerable<TableRowElement> tableRowElements = this.table.Rows.Zip(tableCellDataGrid.Data, this.CreateTableRowElement);

            return tableRowElements;
        }

        protected override void OnBeforeReadChildElement(IDocxImportContext context, OpenXmlElementBase childElement)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            switch (childElement.ElementName)
            {
                case DocxElementNames.TableRow:
                    ((TableRowElement)childElement).SetAssociatedFlowModelElement(new TableRow(this.table.Document), new List<TableCellData>());
                    break;
                case DocxElementNames.TableProperties:
                    this.TablePropertiesElement.SetAssociatedFlowModelElement(this.table.Properties);
                    break;
            }
        }

        protected override void OnAfterReadChildElement(IDocxImportContext context, OpenXmlElementBase childElement)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(childElement, "childElement");

            DocxHelper.AddHangingAnnotation(context, childElement);

            switch (childElement.ElementName)
            {
                case DocxElementNames.TableRow:
                    TableRowElement tableRowElement = (TableRowElement)childElement;
                    this.tableRowDatas.Add(new TableRowData(tableRowElement.TableRow, tableRowElement.TableCellDatas));
                    break;
            }
        }

        protected override void OnAfterRead(IDocxImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            TableCellDataGrid.FillTableFromTableCellsData(this.table, this.tableRowDatas);

            if (this.tablePropertiesElementChildElement != null)
            {
                this.ReleaseElement(this.tablePropertiesElementChildElement);
            }
        }

        protected override void OnBeforeWrite(IDocxExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            if (this.table.Properties.HasLocalValues())
            {
                this.CreateElement(this.tablePropertiesElementChildElement);
                this.TablePropertiesElement.SetAssociatedFlowModelElement(this.table.Properties);
            }
        }

        protected override void ClearOverride()
        {
            base.ClearOverride();

            this.table = null;
        }

        private TableRowElement CreateTableRowElement(TableRow tableRow, TableRowData tableRowData)
        {
            TableRowElement tableRowElement = this.CreateElement<TableRowElement>(DocxElementNames.TableRow);
            tableRowElement.SetAssociatedFlowModelElement(tableRow, tableRowData.TableCellDatas);

            return tableRowElement;
        }
    }
}