﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Import;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Common;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Lists.ListsInfo;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Lists
{
    internal class AbstractListElement : DocxElementBase
    {
        private readonly IntOpenXmlAttribute abstractListIdAttribute;
        private readonly OpenXmlChildElement<NumStyleLinkElement> numStyleLinkElement;
        private readonly OpenXmlChildElement<StyleLinkElement> styleLinkElement;
        private readonly OpenXmlChildElement<MultiLevelTypeElement> multiLevelTypeElement;
        private AbstractListInfo abstractListInfo;

        public AbstractListElement(DocxPartsManager partsManager)
            : base(partsManager)
        {
            this.abstractListIdAttribute = this.RegisterAttribute(new IntOpenXmlAttribute("abstractNumId", OpenXmlNamespaces.WordprocessingMLNamespace, true));

            this.multiLevelTypeElement = this.RegisterChildElement<MultiLevelTypeElement>(DocxElementNames.MultiLevelType);
            this.numStyleLinkElement = this.RegisterChildElement<NumStyleLinkElement>(DocxElementNames.NumStyleLink);
            this.styleLinkElement = this.RegisterChildElement<StyleLinkElement>(DocxElementNames.StyleLink);
        }

        public override string ElementName
        {
            get
            {
                return DocxElementNames.AbstractList;
            }
        }

        public void SetAssociatedElementInfo(AbstractListInfo abstractListInfo)
        {
            Guard.ThrowExceptionIfNull(abstractListInfo, "abstractListInfo");

            this.abstractListInfo = abstractListInfo;
        }

        protected override void OnBeforeWrite(IDocxExportContext context)
        {
            if (this.abstractListInfo.IsAbstractListForStyle)
            {
                this.abstractListIdAttribute.Value = context.ListsExportContext.GetDocxListIds(this.abstractListInfo.Id).AbstractListIdForStyle;

                this.CreateElement(this.numStyleLinkElement);
                this.numStyleLinkElement.Element.Value = this.abstractListInfo.NumStyleLink;
            }
            else
            {
                this.abstractListIdAttribute.Value = context.ListsExportContext.GetDocxListIds(this.abstractListInfo.Id).AbstractListId;

                if (!string.IsNullOrEmpty(this.abstractListInfo.StyleLink))
                {
                    this.CreateElement(this.styleLinkElement);
                    this.styleLinkElement.Element.Value = this.abstractListInfo.StyleLink;
                }
            }

            this.CreateElement(this.multiLevelTypeElement);
            this.multiLevelTypeElement.Element.Value = this.abstractListInfo.MultilevelType.Value;
        }

        protected override void OnAfterRead(IDocxImportContext context)
        {
            this.abstractListInfo.Id = this.abstractListIdAttribute.Value;

            if (this.numStyleLinkElement.Element != null)
            {
                this.abstractListInfo.NumStyleLink = this.numStyleLinkElement.Element.Value;
                this.ReleaseElement(this.numStyleLinkElement);
            }

            if (this.styleLinkElement.Element != null)
            {
                this.abstractListInfo.StyleLink = this.styleLinkElement.Element.Value;
                this.ReleaseElement(this.styleLinkElement);
            }

            if (this.multiLevelTypeElement.Element != null)
            {
                this.abstractListInfo.MultilevelType = this.multiLevelTypeElement.Element.Value;
                this.ReleaseElement(this.multiLevelTypeElement);
            }

            context.ListsImportContext.AbstractListInfoCollection.Add(this.abstractListInfo.Id, this.abstractListInfo);
        }

        protected override IEnumerable<OpenXmlElementBase> EnumerateChildElements(IDocxExportContext context)
        {
            if (!this.abstractListInfo.IsAbstractListForStyle)
            {
                for (int i = 0; i < this.abstractListInfo.Levels.Count; i++)
                {
                    ListLevelElement listLevelElement = this.CreateElement<ListLevelElement>(DocxElementNames.ListLevel);
                    listLevelElement.SetAssociatedElementInfo(this.abstractListInfo.Levels[i]);
                    listLevelElement.SetListLevelId(i);

                    yield return listLevelElement;
                }
            }
        }

        protected override void OnBeforeReadChildElement(IDocxImportContext context, OpenXmlElementBase childElement)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(childElement, "element");

            switch (childElement.ElementName)
            {
                case DocxElementNames.ListLevel:
                    ListLevelInfo listLevelInfo = new ListLevelInfo();
                    this.abstractListInfo.Levels.Add(listLevelInfo);
                    ((ListLevelElement)childElement).SetAssociatedElementInfo(listLevelInfo);
                    break;
            }
        }
    }
}
