﻿using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Import;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Common;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Document;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Lists.ListsInfo;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Styles;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Lists
{
    internal class ListLevelElement : DocxElementBase
    {
        private readonly IntOpenXmlAttribute listLevelIdAttribute;
        private readonly OpenXmlChildElement<StartIndexElement> startIndexElement;
        private readonly OpenXmlChildElement<NumberingStyleElement> numberingStyleElement;
        private readonly OpenXmlChildElement<NumberTextFormatElement> numberTextFormatElement;
        private readonly OpenXmlChildElement<RestartAfterLevelElement> restartAfterLevelElement;
        private readonly OpenXmlChildElement<AlignmentElement> alignmentElement;
        private readonly OpenXmlChildElement<StyleIdElement> styleIdElement;
        private readonly OpenXmlChildElement<IsLegalElement> isLegalElement;
        private readonly OpenXmlChildElement<ParagraphPropertiesElement> paragraphPropertiesChildElement;
        private readonly OpenXmlChildElement<RunPropertiesElement> runPropertiesChildElement;
        private ListLevelInfo listLevelInfo;
        private int listLevelId;

        public ListLevelElement(DocxPartsManager partsManager)
            : base(partsManager)
        {
            this.listLevelIdAttribute = this.RegisterAttribute(new IntOpenXmlAttribute("ilvl", OpenXmlNamespaces.WordprocessingMLNamespace, true));

            this.startIndexElement = this.RegisterChildElement<StartIndexElement>(DocxElementNames.StartIndex);
            this.numberingStyleElement = this.RegisterChildElement<NumberingStyleElement>(DocxElementNames.NumberingStyle);
            this.restartAfterLevelElement = this.RegisterChildElement<RestartAfterLevelElement>(DocxElementNames.RestartAfterLevel);
            this.styleIdElement = this.RegisterChildElement<StyleIdElement>(DocxElementNames.ParagraphStyleId);
            this.isLegalElement = this.RegisterChildElement<IsLegalElement>(DocxElementNames.IsLegal);
            this.numberTextFormatElement = this.RegisterChildElement<NumberTextFormatElement>(DocxElementNames.NumberTextFormat);
            this.alignmentElement = this.RegisterChildElement<AlignmentElement>(DocxElementNames.ListLevelAlignment);
            this.paragraphPropertiesChildElement = this.RegisterChildElement<ParagraphPropertiesElement>(DocxElementNames.ParagraphProperties);
            this.runPropertiesChildElement = this.RegisterChildElement<RunPropertiesElement>(DocxElementNames.RunProperties);
        }

        public override string ElementName
        {
            get
            {
                return DocxElementNames.ListLevel;
            }
        }

        public void SetAssociatedElementInfo(ListLevelInfo listLevelInfo)
        {
            Guard.ThrowExceptionIfNull(listLevelInfo, "listLevelInfo");

            this.listLevelInfo = listLevelInfo;
        }

        public void SetListLevelId(int listLevelId)
        {
            this.listLevelId = listLevelId;
        }

        protected override void OnBeforeWrite(IDocxExportContext context)
        {
            this.listLevelIdAttribute.Value = this.listLevelId;

            this.CreateElement(this.startIndexElement);
            this.startIndexElement.Element.Value = this.listLevelInfo.StartIndex.Value;

            this.CreateElement(this.numberingStyleElement);
            this.numberingStyleElement.Element.Value = this.listLevelInfo.NumberingStyle.Value;

            this.CreateElement(this.numberTextFormatElement);
            this.numberTextFormatElement.Element.Value = this.listLevelInfo.NumberTextFormat;

            if (this.listLevelInfo.RestartAfterLevel.HasValue && this.listLevelInfo.RestartAfterLevel.Value != DocumentDefaultStyleSettings.RestartAfterLevel)
            {
                this.CreateElement(this.restartAfterLevelElement);

                // restart after level is one based index in OpenXml documentation and we have to compensate.
                this.restartAfterLevelElement.Element.Value = this.listLevelInfo.RestartAfterLevel.Value + 1;
            }

            if (this.listLevelInfo.Alignment.HasValue && this.listLevelInfo.Alignment.Value != Alignment.Left)
            {
                this.CreateElement(this.alignmentElement);
                this.alignmentElement.Element.Value = this.listLevelInfo.Alignment.Value;
            }

            if (!string.IsNullOrEmpty(this.listLevelInfo.StyleId))
            {
                this.CreateElement(this.styleIdElement);
                this.styleIdElement.Element.Value = this.listLevelInfo.StyleId;
            }

            if (this.listLevelInfo.IsLegal.HasValue)
            {
                this.CreateElement(this.isLegalElement);
                this.isLegalElement.Element.Value = this.listLevelInfo.IsLegal.Value;
            }

            if (this.listLevelInfo.ParagraphProperties.HasLocalValues())
            {
                this.CreateElement(this.paragraphPropertiesChildElement);
                this.paragraphPropertiesChildElement.Element.SetAssociatedFlowModelElement(this.listLevelInfo.ParagraphProperties);
            }

            if (this.listLevelInfo.CharacterProperties.HasLocalValues())
            {
                this.CreateElement(this.runPropertiesChildElement);
                this.runPropertiesChildElement.Element.SetAssociatedFlowModelElement(this.listLevelInfo.CharacterProperties);
            }
        }

        protected override void OnAfterRead(IDocxImportContext context)
        {
            this.listLevelInfo.ListLevelId = this.listLevelIdAttribute.Value;

            if (this.startIndexElement.Element != null)
            {
                this.listLevelInfo.StartIndex = this.startIndexElement.Element.Value;
                this.ReleaseElement(this.startIndexElement);
            }

            if (this.numberingStyleElement.Element != null)
            {
                this.listLevelInfo.NumberingStyle = this.numberingStyleElement.Element.Value;
                this.ReleaseElement(this.numberingStyleElement);
            }

            if (this.numberTextFormatElement.Element != null)
            {
                this.listLevelInfo.NumberTextFormat = this.numberTextFormatElement.Element.Value;
                this.ReleaseElement(this.numberTextFormatElement);
            }

            if (this.restartAfterLevelElement.Element != null)
            {
                // restart after level is one based index in OpenXml documentation and we have to compensate.
                this.listLevelInfo.RestartAfterLevel = this.restartAfterLevelElement.Element.Value - 1;
                this.ReleaseElement(this.restartAfterLevelElement);
            }

            if (this.alignmentElement.Element != null)
            {
                this.listLevelInfo.Alignment = this.alignmentElement.Element.Value;
                this.ReleaseElement(this.alignmentElement);
            }

            if (this.styleIdElement.Element != null)
            {
                this.listLevelInfo.StyleId = this.styleIdElement.Element.Value;
                this.ReleaseElement(this.styleIdElement);
            }

            if (this.isLegalElement.Element != null)
            {
                this.listLevelInfo.IsLegal = this.isLegalElement.Element.Value;
                this.ReleaseElement(this.isLegalElement);
            }
        }

        protected override void OnBeforeReadChildElement(IDocxImportContext context, OpenXmlElementBase childElement)
        {
            switch (childElement.ElementName)
            {
                case DocxElementNames.ParagraphProperties:
                    (childElement as ParagraphPropertiesElement).SetAssociatedFlowModelElement(this.listLevelInfo.ParagraphProperties);
                    break;

                case DocxElementNames.RunProperties:
                    (childElement as RunPropertiesElement).SetAssociatedFlowModelElement(this.listLevelInfo.CharacterProperties);
                    break;
            }
        }
    }
}
