﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Import;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Common;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Lists.ListsInfo;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Parts;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Lists
{
    internal class ListsElement : DocxPartRootElementBase
    {
        public ListsElement(DocxPartsManager partsManager, OpenXmlPartBase part)
            : base(partsManager, part)
        {
        }

        public override string ElementName
        {
            get
            {
                return DocxElementNames.Lists;
            }
        }

        protected override bool ShouldExport(IDocxExportContext context)
        {
            return context.Document.Lists.Count > 0;
        }

        protected override void OnAfterRead(IDocxImportContext context)
        {
            foreach (NumberingListInfo numberingListInfo in context.ListsImportContext.NumberingListInfoCollection.Values)
            {
                AbstractListInfo abstractListInfo = context.ListsImportContext.AbstractListInfoCollection[numberingListInfo.AbstractListId];
                if (string.IsNullOrEmpty(abstractListInfo.NumStyleLink))
                {
                    // common list.
                    if (numberingListInfo.ListLevelOverrideCollection.Count == 0)
                    {
                        List list = abstractListInfo.GetList();
                        context.Document.Lists.Add(list);

                        // getting correct abstract list id. As the document may contain overridern lists linked to the same abstract list.
                        int nextAbstractListInfoId = abstractListInfo.Id;
                        context.ListsImportContext.AddAbstractListToListMapping(abstractListInfo.Id, list.Id, out nextAbstractListInfoId);
                        numberingListInfo.AbstractListId = nextAbstractListInfoId;
                    }
                    else
                    {
                        // list override.
                        AbstractListInfo abstractListInfoOverride = new AbstractListInfo(abstractListInfo);
                        for (int i = 0; i < numberingListInfo.ListLevelOverrideCollection.Count; i++)
                        {
                            ListLevelInfo listLevelInfo = numberingListInfo.ListLevelOverrideCollection[i];
                            abstractListInfoOverride.Levels[listLevelInfo.ListLevelId].CopyPropertiesFrom(listLevelInfo);
                        }

                        List list = abstractListInfoOverride.GetList();
                        context.Document.Lists.Add(list);

                        int nextAbstractListInfoId = abstractListInfo.Id;
                        context.ListsImportContext.AddAbstractListToListMapping(abstractListInfoOverride.Id, list.Id, out nextAbstractListInfoId);
                        numberingListInfo.AbstractListId = nextAbstractListInfoId;
                    }
                }
            }
        }

        protected override IEnumerable<OpenXmlElementBase> EnumerateChildElements(IDocxExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            // export abstract lists first.
            foreach (List list in context.Document.Lists)
            {
                bool complexList = !string.IsNullOrEmpty(list.StyleId);

                AbstractListElement abstractListElement = this.CreateElement<AbstractListElement>(DocxElementNames.AbstractList);
                abstractListElement.SetAssociatedElementInfo(new AbstractListInfo(list, false));
                yield return abstractListElement;

                if (complexList)
                {
                    AbstractListElement styleRelatedAbstractListElement = this.CreateElement<AbstractListElement>(DocxElementNames.AbstractList);
                    styleRelatedAbstractListElement.SetAssociatedElementInfo(new AbstractListInfo(list, true));
                    yield return styleRelatedAbstractListElement;
                }
            }

            foreach (List list in context.Document.Lists)
            {
                bool complexList = !string.IsNullOrEmpty(list.StyleId);

                // numbered list to abstract list for paragraph.
                NumberingListInfo numberingListInfo = new NumberingListInfo();
                numberingListInfo.Id = context.ListsExportContext.GetDocxListIds(list.Id).NumberedListIdForParagraph;
                numberingListInfo.AbstractListId = context.ListsExportContext.GetDocxListIds(list.Id).AbstractListId;

                NumberedListElement numberedListElement = this.CreateElement<NumberedListElement>(DocxElementNames.NumberedList);
                numberedListElement.SetAssociatedElementInfo(numberingListInfo);

                yield return numberedListElement;

                if (complexList)
                {
                    // numbered list to abstract list for style
                    NumberingListInfo styleNumberingListInfo = new NumberingListInfo();
                    styleNumberingListInfo.Id = context.ListsExportContext.GetDocxListIds(list.Id).NumberedListIdForStyle.Value;
                    styleNumberingListInfo.AbstractListId = context.ListsExportContext.GetDocxListIds(list.Id).AbstractListIdForStyle;

                    NumberedListElement styleNumberedListElement = this.CreateElement<NumberedListElement>(DocxElementNames.NumberedList);
                    styleNumberedListElement.SetAssociatedElementInfo(styleNumberingListInfo);

                    yield return styleNumberedListElement;
                }
            }
        }

        protected override void OnBeforeReadChildElement(IDocxImportContext context, OpenXmlElementBase childElement)
        {
            switch (childElement.ElementName)
            {
                case DocxElementNames.NumberedList:
                    (childElement as NumberedListElement).SetAssociatedElementInfo(new NumberingListInfo());
                    break;
                case DocxElementNames.AbstractList:
                    (childElement as AbstractListElement).SetAssociatedElementInfo(new AbstractListInfo());
                    break;
            }
        }
    }
}
