﻿using Telerik.Windows.Documents.Flow.FormatProviders.Common;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Import;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Common;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Document;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Styles.StylePropertiesElements;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Types;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Styles
{
    internal class StyleElement : DocxElementBase
    {
        private readonly OpenXmlAttribute<string> styleIdAttribute;
        private readonly MappedOpenXmlAttribute<StyleType> typeAttribute;
        private readonly BoolOpenXmlAttribute defaultAttribute;
        private readonly BoolOpenXmlAttribute customAttribute;

        private readonly OpenXmlChildElement<StyleNameElement> nameElement;
        private readonly OpenXmlChildElement<IsPrimaryElement> isPrimaryElement;
        private readonly OpenXmlChildElement<NextStyleIdElement> nextStyleIdElement;
        private readonly OpenXmlChildElement<LinkedStyleIdElement> linkedStyleIdElement;
        private readonly OpenXmlChildElement<BasedOnStyleIdElement> basedOnStyleIdElement;
        private readonly OpenXmlChildElement<UIPriorityElement> elementUIPriority;

        private readonly OpenXmlChildElement<RunPropertiesElement> runPropertiesChildElement;
        private readonly OpenXmlChildElement<ParagraphPropertiesElement> paragraphPropertiesChildElement;
        private readonly OpenXmlChildElement<TablePropertiesElement> tablePropertiesChildElement;
        private readonly OpenXmlChildElement<TableRowPropertiesElement> tableRowPropertiesChildElement;
        private readonly OpenXmlChildElement<TableCellPropertiesElement> tableCellPropertiesChildElement;

        private Style style;

        public StyleElement(DocxPartsManager partsManager)
            : base(partsManager)
        {
            this.styleIdAttribute = this.RegisterAttribute(new OpenXmlAttribute<string>("styleId", OpenXmlNamespaces.WordprocessingMLNamespace, true));
            this.typeAttribute = this.RegisterAttribute(new MappedOpenXmlAttribute<StyleType>("type", OpenXmlNamespaces.WordprocessingMLNamespace, TypeMappers.StyleTypeMapper, true));
            this.defaultAttribute = this.RegisterAttribute(new BoolOpenXmlAttribute("default", OpenXmlNamespaces.WordprocessingMLNamespace, true));
            this.customAttribute = this.RegisterAttribute(new BoolOpenXmlAttribute("customStyle", OpenXmlNamespaces.WordprocessingMLNamespace, true));

            this.nameElement = this.RegisterChildElement<StyleNameElement>(DocxElementNames.StyleName);
            this.basedOnStyleIdElement = this.RegisterChildElement<BasedOnStyleIdElement>(DocxElementNames.BasedOnStyleId);
            this.nextStyleIdElement = this.RegisterChildElement<NextStyleIdElement>(DocxElementNames.NextStyleId);
            this.linkedStyleIdElement = this.RegisterChildElement<LinkedStyleIdElement>(DocxElementNames.LinkedStyleId);
            this.elementUIPriority = this.RegisterChildElement<UIPriorityElement>(DocxElementNames.UIPriority);
            this.isPrimaryElement = this.RegisterChildElement<IsPrimaryElement>(DocxElementNames.IsPrimary);

            this.paragraphPropertiesChildElement = this.RegisterChildElement<ParagraphPropertiesElement>(DocxElementNames.ParagraphProperties);
            this.runPropertiesChildElement = this.RegisterChildElement<RunPropertiesElement>(DocxElementNames.RunProperties);
            this.tablePropertiesChildElement = this.RegisterChildElement<TablePropertiesElement>(DocxElementNames.TableProperties);
            this.tableRowPropertiesChildElement = this.RegisterChildElement<TableRowPropertiesElement>(DocxElementNames.TableRowProperties);
            this.tableCellPropertiesChildElement = this.RegisterChildElement<TableCellPropertiesElement>(DocxElementNames.TableCellProperties);
        }

        public override string ElementName
        {
            get
            {
                return DocxElementNames.Style;
            }
        }

        public Style Style
        {
            get
            {
                return this.style;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "style");

                this.style = value;
            }
        }

        protected override void OnAfterRead(IDocxImportContext context)
        {
            if (!this.styleIdAttribute.HasValue)
            {
                this.styleIdAttribute.Value = UndefinedStyleIdGenerator.GetNext();
            }

            if (!this.typeAttribute.HasValue)
            {
                this.typeAttribute.Value = StyleType.Paragraph;
            }

            this.Style = new Style(this.styleIdAttribute.Value, this.typeAttribute.Value);

            if (this.defaultAttribute.HasValue)
            {
                this.Style.IsDefault = this.defaultAttribute.Value;
            }

            if (this.customAttribute.HasValue)
            {
                this.Style.IsCustom = this.customAttribute.Value;
            }

            if (this.nameElement.Element != null)
            {
                this.Style.Name = this.nameElement.Element.Value;
            }

            if (this.isPrimaryElement.Element != null)
            {
                this.Style.IsPrimary = this.isPrimaryElement.Element.Value;
            }

            if (this.basedOnStyleIdElement.Element != null)
            {
                this.Style.BasedOnStyleId = this.basedOnStyleIdElement.Element.Value;
            }

            if (this.nextStyleIdElement.Element != null)
            {
                this.Style.NextStyleId = this.nextStyleIdElement.Element.Value;
            }

            if (this.linkedStyleIdElement.Element != null)
            {
                this.Style.LinkedStyleId = this.linkedStyleIdElement.Element.Value;
            }

            if (this.elementUIPriority.Element != null)
            {
                this.Style.UIPriority = this.elementUIPriority.Element.Value;
            }

            if (this.runPropertiesChildElement.Element != null)
            {
                if (this.Style.CharacterProperties != null)
                {
                    this.runPropertiesChildElement.Element.CopyPropertiesTo(this.Style.CharacterProperties);
                }
            }

            if (this.paragraphPropertiesChildElement.Element != null)
            {
                if (this.Style.ParagraphProperties != null)
                {
                    this.Style.ParagraphProperties.CopyPropertiesFrom(this.paragraphPropertiesChildElement.Element.ParagraphProperties);
                }
            }

            if (this.tablePropertiesChildElement.Element != null)
            {
                if (this.Style.TableProperties != null)
                {
                    this.Style.TableProperties.CopyPropertiesFrom(this.tablePropertiesChildElement.Element.TableProperties);
                }
            }

            if (this.tableRowPropertiesChildElement.Element != null)
            {
                if (this.Style.TableRowProperties != null)
                {
                    this.Style.TableRowProperties.CopyPropertiesFrom(this.tableRowPropertiesChildElement.Element.TableRowProperties);
                }
            }

            if (this.tableCellPropertiesChildElement.Element != null)
            {
                if (this.Style.TableCellProperties != null)
                {
                    this.Style.TableCellProperties.CopyPropertiesFrom(this.tableCellPropertiesChildElement.Element.TableCellProperties);
                }
            }

            this.ReleaseElements();
        }

        protected override void OnBeforeWrite(IDocxExportContext context)
        {
            this.styleIdAttribute.Value = this.Style.Id;

            this.typeAttribute.Value = this.Style.StyleType;

            if (this.Style.IsDefault)
            {
                this.defaultAttribute.Value = this.Style.IsDefault;
            }

            if (this.Style.IsCustom)
            {
                this.customAttribute.Value = this.Style.IsCustom;
            }

            // the style name will return the id if no name is defined.
            this.CreateElement(this.nameElement);
            this.nameElement.Element.Value = this.style.Name;

            if (this.Style.IsPrimary)
            {
                this.CreateElement(this.isPrimaryElement);
                this.isPrimaryElement.Element.Value = this.Style.IsPrimary;
            }

            if (!string.IsNullOrEmpty(this.Style.BasedOnStyleId))
            {
                this.CreateElement(this.basedOnStyleIdElement);
                this.basedOnStyleIdElement.Element.Value = this.Style.BasedOnStyleId;
            }

            if (!string.IsNullOrEmpty(this.Style.NextStyleId))
            {
                this.CreateElement(this.nextStyleIdElement);
                this.nextStyleIdElement.Element.Value = this.Style.NextStyleId;
            }

            if (!string.IsNullOrEmpty(this.Style.LinkedStyleId))
            {
                this.CreateElement(this.linkedStyleIdElement);
                this.linkedStyleIdElement.Element.Value = this.Style.LinkedStyleId;
            }

            if (this.Style.UIPriority != DocumentDefaultStyleSettings.UIPriority)
            {
                this.CreateElement(this.elementUIPriority);
                this.elementUIPriority.Element.Value = this.Style.UIPriority;
            }

            if (this.Style.CharacterProperties.HasLocalValues())
            {
                this.CreateElement(this.runPropertiesChildElement);
                this.runPropertiesChildElement.Element.SetAssociatedFlowModelElement(this.Style.CharacterProperties);
            }

            if (this.Style.ParagraphProperties != null && this.Style.ParagraphProperties.HasLocalValues())
            {
                this.CreateElement(this.paragraphPropertiesChildElement);
                this.paragraphPropertiesChildElement.Element.SetAssociatedFlowModelElement(this.Style.ParagraphProperties);
            }

            if (this.Style.TableProperties != null && this.Style.TableProperties.HasLocalValues())
            {
                this.CreateElement(this.tablePropertiesChildElement);
                this.tablePropertiesChildElement.Element.SetAssociatedFlowModelElement(this.Style.TableProperties);
            }

            if (this.Style.TableRowProperties != null && this.Style.TableRowProperties.HasLocalValues())
            {
                this.CreateElement(this.tableRowPropertiesChildElement);
                this.tableRowPropertiesChildElement.Element.SetAssociatedFlowModelElement(this.Style.TableRowProperties);
            }

            if (this.Style.TableCellProperties != null && this.Style.TableCellProperties.HasLocalValues())
            {
                this.CreateElement(this.tableCellPropertiesChildElement);
                this.tableCellPropertiesChildElement.Element.SetAssociatedFlowModelElement(this.Style.TableCellProperties);
            }
        }

        protected override void OnBeforeReadChildElement(IDocxImportContext context, OpenXmlElementBase childElement)
        {
            switch (childElement.ElementName)
            {
                case DocxElementNames.ParagraphProperties:
                    (childElement as ParagraphPropertiesElement).SetAssociatedFlowModelElement(new Style("paragraphStyle", StyleType.Table).ParagraphProperties);
                    break;

                case DocxElementNames.TableProperties:
                    (childElement as TablePropertiesElement).SetAssociatedFlowModelElement(new Style("tableStyle", StyleType.Table).TableProperties);
                    break;
                case DocxElementNames.TableRowProperties:
                    (childElement as TableRowPropertiesElement).SetAssociatedFlowModelElement(new Style("tableRowStyle", StyleType.Table).TableRowProperties);
                    break;
                case DocxElementNames.TableCellProperties:
                    (childElement as TableCellPropertiesElement).SetAssociatedFlowModelElement(new Style("tableCellStyle", StyleType.Table).TableCellProperties);
                    break;
            }
        }

        private void ReleaseElements()
        {
            if (this.nameElement.Element != null)
            {
                this.ReleaseElement(this.nameElement);
            }

            if (this.isPrimaryElement.Element != null)
            {
                this.ReleaseElement(this.isPrimaryElement);
            }

            if (this.basedOnStyleIdElement.Element != null)
            {
                this.ReleaseElement(this.basedOnStyleIdElement);
            }

            if (this.nextStyleIdElement.Element != null)
            {
                this.ReleaseElement(this.nextStyleIdElement);
            }

            if (this.linkedStyleIdElement.Element != null)
            {
                this.ReleaseElement(this.linkedStyleIdElement);
            }

            if (this.elementUIPriority.Element != null)
            {
                this.ReleaseElement(this.elementUIPriority);
            }

            if (this.runPropertiesChildElement.Element != null)
            {
                this.ReleaseElement(this.runPropertiesChildElement);
            }

            if (this.paragraphPropertiesChildElement.Element != null)
            {
                this.ReleaseElement(this.paragraphPropertiesChildElement);
            }

            if (this.tablePropertiesChildElement.Element != null)
            {
                this.ReleaseElement(this.tablePropertiesChildElement);
            }

            if (this.tableRowPropertiesChildElement.Element != null)
            {
                this.ReleaseElement(this.tableRowPropertiesChildElement);
            }

            if (this.tableCellPropertiesChildElement.Element != null)
            {
                this.ReleaseElement(this.tableCellPropertiesChildElement);
            }
        }
    }
}
