﻿using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Import;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Document;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements;
using Telerik.Windows.Documents.Primitives;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Styles
{
    internal class TableCellPaddingElement : DocumentElementBase
    {
        private readonly string elementName;
        private readonly OpenXmlChildElement<TableWidthElement> left;
        private readonly OpenXmlChildElement<TableWidthElement> top;
        private readonly OpenXmlChildElement<TableWidthElement> right;
        private readonly OpenXmlChildElement<TableWidthElement> bottom;

        private IPropertiesWithPadding properties;

        public TableCellPaddingElement(DocxPartsManager partsManager, string elementName)
            : base(partsManager)
        {
            this.elementName = elementName;
            this.top = this.RegisterChildElement<TableWidthElement>(DocxElementNames.Top);
            this.left = this.RegisterChildElement<TableWidthElement>(DocxElementNames.Left);
            this.bottom = this.RegisterChildElement<TableWidthElement>(DocxElementNames.Bottom);
            this.right = this.RegisterChildElement<TableWidthElement>(DocxElementNames.Right);
        }

        public override string ElementName
        {
            get
            {
                return this.elementName;
            }
        }

        public void SetAssociatedFlowModelElement(IPropertiesWithPadding properties)
        {
            Guard.ThrowExceptionIfNull(properties, "properties");

            this.properties = properties;
        }

        /// <summary>
        /// Creates the element. 
        /// We override this method in order to handle the situation where child element have the same name as the child element of another docx element.
        /// example: both table borders and table cell padding has child elements with name left, top, etc.
        /// </summary>
        /// <param name="elementName">Name of the element.</param>
        /// <returns></returns>
        protected override OpenXmlElementBase CreateElement(string elementName)
        {
            switch (elementName)
            {
                case DocxElementNames.Left:
                case DocxElementNames.Top:
                case DocxElementNames.Right:
                case DocxElementNames.Bottom:
                    TableWidthElement element = new TableWidthElement(this.PartsManager, elementName);
                    element.Part = this.Part;

                    return element;

                default:

                    return base.CreateElement(elementName);
            }
        }

        protected override void ReleaseElementOverride(OpenXmlElementBase element)
        {
            switch (element.ElementName)
            {
                case DocxElementNames.Left:
                case DocxElementNames.Top:
                case DocxElementNames.Right:
                case DocxElementNames.Bottom:
                    break;

                default:
                    base.ReleaseElementOverride(element);
                    break;
            }
        }

        protected override void OnAfterRead(IDocxImportContext context)
        {
            bool hasLocalBorderValue = false;
            double left = 0;
            double top = 0;
            double right = 0;
            double bottom = 0;

            if (this.left.Element != null)
            {
                hasLocalBorderValue = true;
                left = this.left.Element.ToDouble();
                this.ReleaseElement(this.left);
            }

            if (this.top.Element != null)
            {
                hasLocalBorderValue = true;
                top = this.top.Element.ToDouble();
                this.ReleaseElement(this.top);
            }

            if (this.right.Element != null)
            {
                hasLocalBorderValue = true;
                right = this.right.Element.ToDouble();
                this.ReleaseElement(this.right);
            }

            if (this.bottom.Element != null)
            {
                hasLocalBorderValue = true;
                bottom = this.bottom.Element.ToDouble();
                this.ReleaseElement(this.bottom);
            }

            if (hasLocalBorderValue)
            {
                this.properties.Padding.LocalValue = new Padding(left, top, right, bottom);
            }
        }

        protected override void OnBeforeWrite(IDocxExportContext context)
        {
            if (this.properties.Padding.HasLocalValue)
            {
                Padding padding = this.properties.Padding.LocalValue;

                this.CreateElement(this.left);
                this.left.Element.Value = new TableWidthUnit(TableWidthUnitType.Fixed, padding.Left);

                this.CreateElement(this.top);
                this.top.Element.Value = new TableWidthUnit(TableWidthUnitType.Fixed, padding.Top);

                this.CreateElement(this.right);
                this.right.Element.Value = new TableWidthUnit(TableWidthUnitType.Fixed, padding.Right);

                this.CreateElement(this.bottom);
                this.bottom.Element.Value = new TableWidthUnit(TableWidthUnitType.Fixed, padding.Bottom);
            }
        }
    }
}
