﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Media;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Import;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx.Utilities;
using Telerik.Windows.Documents.Flow.Model.Watermarks;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Types;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Model.Drawing.Shapes;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Docx.Model.Elements.Vector
{
    internal class ShapeElement : VectorElementBase
    {
        private const string ImageWatermarkId = "WordPictureWatermark";
        private const string TextWatermarkId = "PowerPlusWaterMarkObject";
        private const string ImageWatermarkType = "#_x0000_t75";
        private const string TextWatermarkType = "#_x0000_t136";
        private const string StyleString = "position:absolute;margin-left:0;margin-top:0;width:{0}pt;height:{1}pt;rotation:{2};z-index:-251658752;mso-position-horizontal:center;mso-position-horizontal-relative:margin;mso-position-vertical:center;mso-position-vertical-relative:margin";
        private const string TextpathStyleString = @"font-family:&quot;{0}&quot;";
        private const int UnitTypeLength = 2;

        private readonly Random randomGenerator;

        private readonly OpenXmlAttribute<string> id;
        private readonly OpenXmlAttribute<string> type;
        private readonly OpenXmlAttribute<string> style;
        private readonly ConvertedOpenXmlAttribute<Color> fillColor;
        private readonly BoolOpenXmlAttribute stroked;

        private Media.ImageSource imageSource;
        private Watermark watermark;
        private TextWatermarkSettings textWatermarkSettings;
        private ImageWatermarkSettings imageWatermarkSettings;

        public ShapeElement(DocxPartsManager partsManager)
            : base(partsManager)
        {
            this.id = this.RegisterAttribute(new OpenXmlAttribute<string>("id"));
            this.type = this.RegisterAttribute(new OpenXmlAttribute<string>("type"));
            this.style = this.RegisterAttribute(new OpenXmlAttribute<string>("style"));
            this.fillColor = this.RegisterAttribute(new ConvertedOpenXmlAttribute<Color>("fillcolor", Converters.NameOrHashedHexBinary3Converter));
            this.stroked = this.RegisterAttribute(new BoolOpenXmlAttribute("stroked"));

            this.randomGenerator = new Random();
        }

        public override string ElementName
        {
            get
            {
                return DocxElementNames.Shape;
            }
        }

        public string Id
        {
            get
            {
                return this.id.Value;
            }
            set
            {
                this.id.Value = value;
            }
        }

        public string Type
        {
            get
            {
                return this.type.Value;
            }
            set
            {
                this.type.Value = value;
            }
        }

        public string Style
        {
            get
            {
                return this.style.Value;
            }
            set
            {
                this.style.Value = value;
            }
        }

        public Color FillColor
        {
            get
            {
                return this.fillColor.Value;
            }
            set
            {
                this.fillColor.Value = value;
            }
        }

        public bool Stroked
        {
            get
            {
                return this.stroked.Value;
            }
            set
            {
                this.stroked.Value = value;
            }
        }

        public ShapeTypes ShapeType
        {
            get
            {
                if (this.Id.Contains(TextWatermarkId) || this.Id.Contains(ImageWatermarkId))
                {
                    return ShapeTypes.Watermark;
                }
                else if (this.imageSource != null)
                {
                    return ShapeTypes.Image;
                }
                else
                {
                    return ShapeTypes.Unknown;
                }
            }
        }

        public void SetAssociatedWatermarkElement(Watermark watermark)
        {
            Guard.ThrowExceptionIfNull(watermark, "watermark");

            this.watermark = watermark;
        }

        public Image CreateImage()
        {
            ShapeStyleInfo shapeStyleInfo = this.ReadShapeStyle();

            Image image = new Image()
            {
                ImageSource = this.imageSource,
                Width = shapeStyleInfo.Width,
                Height = shapeStyleInfo.Height,
                RotationAngle = shapeStyleInfo.RotationAngle
            };

            return image;
        }

        public Watermark CreateWatermark()
        {
            if (this.textWatermarkSettings == null && this.imageWatermarkSettings == null)
            {
                if (this.Id.Contains(TextWatermarkId))
                {
                    this.textWatermarkSettings = new TextWatermarkSettings();
                }
                else if (this.Id.Contains(ImageWatermarkId))
                {
                    this.imageWatermarkSettings = new ImageWatermarkSettings();
                }
                else
                {
                    Debug.Assert(false, "Unknown watermark settings!");
                    return null;
                }
            }

            ShapeStyleInfo shapeStyleInfo;
            if (this.textWatermarkSettings != null)
            {
                shapeStyleInfo = this.ReadShapeStyle();
                this.textWatermarkSettings.ForegroundColor = this.FillColor;
                this.textWatermarkSettings.Width = shapeStyleInfo.Width;
                this.textWatermarkSettings.Height = shapeStyleInfo.Height;
                this.textWatermarkSettings.Angle = shapeStyleInfo.RotationAngle;

                return new Watermark(this.textWatermarkSettings);
            }
            else if (this.imageWatermarkSettings != null)
            {
                shapeStyleInfo = this.ReadShapeStyle();
                this.imageWatermarkSettings.Width = shapeStyleInfo.Width;
                this.imageWatermarkSettings.Height = shapeStyleInfo.Height;
                this.imageWatermarkSettings.Angle = shapeStyleInfo.RotationAngle;
                this.imageWatermarkSettings.ImageSource = this.imageSource;

                return new Watermark(this.imageWatermarkSettings);
            }

            return null;
        }

        protected override void OnBeforeWrite(IDocxExportContext context)
        {
            if (this.watermark != null)
            {
                double widthInPoints;
                double heightInPoints;

                switch (this.watermark.WatermarkType)
                {
                    case WatermarkType.Image:

                        widthInPoints = Unit.DipToPoint(this.watermark.ImageSettings.Width);
                        heightInPoints = Unit.DipToPoint(this.watermark.ImageSettings.Height);

                        this.Id = ImageWatermarkId + this.randomGenerator.Next();
                        this.Type = ImageWatermarkType;
                        this.Style = string.Format(StyleString, widthInPoints, heightInPoints, this.watermark.ImageSettings.Angle);
                        break;

                    case WatermarkType.Text:

                        widthInPoints = Unit.DipToPoint(this.watermark.TextSettings.Width);
                        heightInPoints = Unit.DipToPoint(this.watermark.TextSettings.Height);

                        this.Id = TextWatermarkId + this.randomGenerator.Next();
                        this.Type = TextWatermarkType;
                        this.Style = string.Format(StyleString, widthInPoints, heightInPoints, this.watermark.TextSettings.Angle);
                        this.FillColor = this.watermark.TextSettings.ForegroundColor;
                        this.Stroked = false;
                        break;

                    default:
                        throw new NotSupportedException(this.watermark.WatermarkType + " is unsupported watermark type.");
                }
            }
        }

        protected override IEnumerable<OpenXmlElementBase> EnumerateChildElements(IDocxExportContext context)
        {
            if (this.watermark != null)
            {
                switch (this.watermark.WatermarkType)
                {
                    case WatermarkType.Image:
                        ImageDataElement imageDataElement = this.CreateElement<ImageDataElement>(DocxElementNames.ImageData);
                        if (this.watermark.ImageSettings.ImageSource != null)
                        {
                            imageDataElement.Id = this.CreateRelationshipIdFromImageSource(this.watermark.ImageSettings.ImageSource);
                        }
                        yield return imageDataElement;
                        break;
                    case WatermarkType.Text:
                        TextpathElement textpathElement = this.CreateElement<TextpathElement>(DocxElementNames.Textpath);
                        textpathElement.StringAttribute = this.watermark.TextSettings.Text;
                        textpathElement.Style = string.Format(TextpathStyleString, this.watermark.TextSettings.FontFamily);
                        yield return textpathElement;

                        FillElement fillElement = this.CreateElement<FillElement>(DocxElementNames.Fill);
                        fillElement.Opacity = this.watermark.TextSettings.Opacity;
                        yield return fillElement;

                        break;
                    default:
                        throw new NotSupportedException(this.watermark.WatermarkType + " is unsupported watermark type.");
                }
            }
        }

        protected override void OnAfterReadChildElement(IDocxImportContext context, OpenXmlElementBase childElement)
        {
            string elementName = childElement.ElementName;

            switch (elementName)
            {
                case DocxElementNames.Textpath:
                    if (this.textWatermarkSettings == null)
                    {
                        this.textWatermarkSettings = new TextWatermarkSettings();
                    }

                    TextpathElement textpathElement = childElement as TextpathElement;

                    FontFamily fontFamily = GetFontFamilyFromTextPathElementStyleString(textpathElement.Style);
                    if (fontFamily != null)
                    {
                        this.textWatermarkSettings.FontFamily = fontFamily;
                    }

                    if (!string.IsNullOrEmpty(textpathElement.StringAttribute))
                    {
                        this.textWatermarkSettings.Text = textpathElement.StringAttribute;
                    }

                    break;
                case DocxElementNames.Fill:
                    if (this.textWatermarkSettings == null)
                    {
                        this.textWatermarkSettings = new TextWatermarkSettings();
                    }

                    FillElement fillElement = childElement as FillElement;
                    this.textWatermarkSettings.Opacity = fillElement.Opacity;

                    break;
                case DocxElementNames.ImageData:
                    ImageDataElement imageDataElement = childElement as ImageDataElement;
                    try
                    {
                        this.imageSource = this.CreateImageSourceFromRelationshipId(imageDataElement.Id, context);
                    }
                    catch (KeyNotFoundException)
                    {
                        this.imageSource = null;
                    }                    
                    break;
            }
        }

        protected override void ClearOverride()
        {
            this.imageSource = null;
            this.watermark = null;
            this.textWatermarkSettings = null;
            this.imageWatermarkSettings = null;
        }

        private static bool TryParseSize(string input, out double size)
        {
            string valueString = input;
            string unitString = string.Empty;

            if (input.Length > UnitTypeLength)
            {
                unitString = input.Substring(input.Length - UnitTypeLength);
            }

            if (!string.IsNullOrEmpty(unitString))
            {
                valueString = input.Remove(input.Length - UnitTypeLength);
            }

            bool isParsed = false;
            if (double.TryParse(valueString, NumberStyles.Any, CultureInfo.InvariantCulture, out size))
            {
                switch (unitString)
                {
                    case "in":
                        size = Unit.InchToDip(size);
                        isParsed = true;
                        break;
                    case "pt":
                        size = Unit.PointToDip(size);
                        isParsed = true;
                        break;
                }
            }

            return isParsed;
        }

        private static FontFamily GetFontFamilyFromTextPathElementStyleString(string styleString)
        {
            int index = styleString.IndexOf("font-family:");
            if (index >= 0)
            {
                string fontFamilyString = styleString.Substring(index + "font-family:".Length);
                string[] splitFontFamilyString = fontFamilyString.Split(new string[] { "\"", "&quot;" }, StringSplitOptions.RemoveEmptyEntries);

                if (splitFontFamilyString.Length > 0)
                {
                    return new FontFamily(splitFontFamilyString[0]);
                }
            }

            return null;
        }

        private string CreateRelationshipIdFromImageSource(Telerik.Windows.Documents.Media.ImageSource resource)
        {
            string resourceName = DocxHelper.CreateResourceName(resource);

            return this.PartsManager.CreateRelationship(this.Part.Name, resourceName, OpenXmlRelationshipTypes.GetRelationshipTypeByExtension(resource.Extension));
        }

        private Telerik.Windows.Documents.Media.ImageSource CreateImageSourceFromRelationshipId(string relationshipId, IDocxImportContext context)
        {
            string target = this.PartsManager.GetRelationshipTarget(this.Part.Name, relationshipId);
            string resourceKey = this.Part.GetResourceName(target);

            return (Telerik.Windows.Documents.Media.ImageSource)context.GetResourceByResourceKey(resourceKey);
        }

        private ShapeStyleInfo ReadShapeStyle()
        {
            string[] styles = this.Style.Split(';');

            double width = 0;
            double height = 0;
            double rotationAngle = 0;

            string widthAttributeName = "width";
            string heightAttributeName = "height";
            string rotationAttributeName = "rotation";
            string attributeValueSeparator = ":";

            foreach (string style in styles)
            {
                if (style.StartsWith(widthAttributeName))
                {
                    string widthString = style.Substring(widthAttributeName.Length + attributeValueSeparator.Length);
                    if (!ShapeElement.TryParseSize(widthString, out width))
                    {
                        width = 0;
                    }
                }
                else if (style.StartsWith(heightAttributeName))
                {
                    string heightString = style.Substring(heightAttributeName.Length + attributeValueSeparator.Length);
                    if (!ShapeElement.TryParseSize(heightString, out height))
                    {
                        height = 0;
                    }
                }
                else if (style.StartsWith(rotationAttributeName))
                {
                    string angleString = style.Substring(rotationAttributeName.Length + attributeValueSeparator.Length);
                    if (!double.TryParse(angleString, out rotationAngle))
                    {
                        rotationAngle = 0;
                    }
                }
            }

            return new ShapeStyleInfo(width, height, rotationAngle);
        }
    }
}
