﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model;
using Telerik.Windows.Documents.Model.Drawing.Shapes;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Docx.Validation
{
    /// <summary>
    /// Represents Docx validator class.
    /// </summary>
    public static class DocxValidator
    {
        private static readonly Dictionary<DocumentElementType, List<IValidationRule>> validationRules;
        private static readonly Dictionary<IValidationRule, IRepairRule> repairRules;
        private static readonly List<KeyValuePair<DocumentElementBase, IValidationRule>> pendingRepairElements;

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Maintainability", "CA1506:AvoidExcessiveClassCoupling"), System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Maintainability", "CA1502:AvoidExcessiveComplexity")]
        static DocxValidator()
        {
            validationRules = new Dictionary<DocumentElementType, List<IValidationRule>>();
            repairRules = new Dictionary<IValidationRule, IRepairRule>();
            pendingRepairElements = new List<KeyValuePair<DocumentElementBase, IValidationRule>>();

            RegisterRules<CommentRangeStart>(DocumentElementType.CommentRangeStart,
                cr =>
                {
                    return cr.Comment.CommentRangeEnd.Paragraph != null;
                },
                cr =>
                {
                    Paragraph paragraph = cr.Comment.CommentRangeStart.Paragraph;
                    int index = paragraph.Inlines.IndexOf(cr.Comment.CommentRangeStart);
                    paragraph.Inlines[index] = cr.Comment.CommentRangeEnd;
                },
                "CommentRangeStart element cannot exists without corresponding end.",
                true);

            RegisterRules<Table>(DocumentElementType.Table,
                t => t.Children.Any(),
                t => t.Rows.AddTableRow().Cells.AddTableCell().Blocks.AddParagraph(),
                "Table element cannot be empty.");

            RegisterRules<TableRow>(DocumentElementType.TableRow,
                tr => tr.Children.Any(),
                tr => tr.Cells.AddTableCell().Blocks.AddParagraph(),
                "Table row element cannot be empty.");

            RegisterRules<TableCell>(DocumentElementType.TableCell,
                tc =>
                {
                    DocumentElementBase element = tc.Children.LastOrDefault();

                    return element != null && element.Type == DocumentElementType.Paragraph;
                },
                tc => tc.Blocks.AddParagraph(),
                "Table cell element cannot be empty.");

            RegisterRules<Section>(DocumentElementType.Section,
                s =>
                {
                    return s == s.Document.Sections.Last() || s.Children.Any();
                },
                s =>
                {
                    s.Blocks.AddParagraph();
                },
                "Only last section can be empty.");

            RegisterRules<Section>(DocumentElementType.Section,
                s =>
                {
                    return s == s.Document.Sections.Last() || (s.Blocks.Count != 0 && s.Blocks.Last() is Paragraph);
                },
                s =>
                {
                    s.Blocks.AddParagraph();
                },
                "Last element in section should be paragraph to preserve the section properties.");

            RegisterRules<ImageInline>(DocumentElementType.ImageInline,
                i =>
                {
                    return i.Image.ImageSource == null || OpenXmlHelper.IsSupportedImageSourceExtension(i.Image.ImageSource.Extension);
                },
                i =>
                {
                    i.Image.ImageSource = null;
                },
                "Image source format is not supported.");

            RegisterRules<ImageInline>(DocumentElementType.ImageInline,
                i => (i.Image.HasWidth || i.Image.Size.Width != 1) && (i.Image.HasHeight || i.Image.Size.Height != 1),
                i => i.Image.RecalculateSize(),
                "ImageSource size is not calculated.");

            RegisterRules<ImageInline>(DocumentElementType.ImageInline,
                i => !i.Image.Size.IsEmpty,
                i => i.Image.Size = ShapeBase.DefaultSize,
                "Image size cannot be Size.Empty.");

            RegisterRules<FloatingImage>(DocumentElementType.FloatingImage,
               fi => !fi.Image.Size.IsEmpty,
               fi => fi.Image.Size = ShapeBase.DefaultSize,
               "Image size cannot be Size.Empty.");
        }

        /// <summary>
        /// Validates if the specified document is compliant with Docx format validation rules.
        /// </summary>
        /// <param name="document">The document.</param>
        /// <returns>Returns validation result that contains result type and validation errors.</returns>
        public static ValidationResult Validate(RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");

            IEnumerable<ValidationError> errors = GetValidationErrors(document);

            if (errors.Any())
            {
                return new ValidationResult(errors);
            }
            else
            {
                return new ValidationResult();
            }
        }

        /// <summary>
        /// Repairs the specified document to be compliant with Docx format.
        /// </summary>
        /// <param name="document">The document.</param>
        public static void Repair(RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");

            foreach (DocumentElementBase child in EnumerateDocumentElements(document))
            {
                DocxValidator.RepairDocumentElement(child);
            }

            foreach (KeyValuePair<DocumentElementBase, IValidationRule> pendingRepairChild in pendingRepairElements)
            {
                DocxValidator.TryRepairDocumentElement(pendingRepairChild.Key, pendingRepairChild.Value);
            }
        }

        private static IEnumerable<ValidationError> GetValidationErrors(RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");

            foreach (DocumentElementBase child in EnumerateDocumentElements(document))
            {
                foreach (ValidationError error in GetValidationErrorsForDocumentElement(child))
                {
                    yield return error;
                }
            }
        }

        private static IEnumerable<ValidationError> GetValidationErrorsForDocumentElement(DocumentElementBase documentElement)
        {
            Guard.ThrowExceptionIfNull(documentElement, "documentElement");

            List<IValidationRule> rules;
            if (validationRules.TryGetValue(documentElement.Type, out rules))
            {
                foreach (IValidationRule rule in rules)
                {
                    if (!rule.IsCompliant(documentElement))
                    {
                        yield return new ValidationError(documentElement, rule.ErrorMessage);
                    }
                }
            }
        }

        private static void RepairDocumentElement(DocumentElementBase documentElement)
        {
            Guard.ThrowExceptionIfNull(documentElement, "documentElement");

            List<IValidationRule> rules;
            if (validationRules.TryGetValue(documentElement.Type, out rules))
            {
                foreach (IValidationRule rule in rules)
                {
                    if (rule.IsDocumentChanging)
                    {
                        pendingRepairElements.Add(new KeyValuePair<DocumentElementBase, IValidationRule>(documentElement, rule));
                    }
                    else
                    {
                        TryRepairDocumentElement(documentElement, rule);
                    }
                }
            }
        }

        private static void TryRepairDocumentElement(DocumentElementBase documentElement, IValidationRule rule)
        {
            IRepairRule repairRule;
            if (!rule.IsCompliant(documentElement) && repairRules.TryGetValue(rule, out repairRule))
            {
                repairRule.Repair(documentElement);
            }
        }

        private static IEnumerable<DocumentElementBase> EnumerateDocumentElements(DocumentElementBase root)
        {
            Guard.ThrowExceptionIfNull(root, "root");

            yield return root;

            foreach (DocumentElementBase child in root.Children)
            {
                foreach (DocumentElementBase innerChild in EnumerateDocumentElements(child))
                {
                    yield return innerChild;
                }
            }
        }

        private static void RegisterRules<T>(DocumentElementType elementType, Func<T, bool> predicate, Action<T> repairAction, string errorMessage, bool isDocumentChanging = false)
            where T : DocumentElementBase
        {
            Guard.ThrowExceptionIfNull(predicate, "predicate");
            Guard.ThrowExceptionIfNull(repairAction, "repairAction");
            Guard.ThrowExceptionIfNullOrEmpty(errorMessage, "errorMessage");

            List<IValidationRule> rules;
            if (!validationRules.TryGetValue(elementType, out rules))
            {
                rules = new List<IValidationRule>();
                validationRules[elementType] = rules;
            }

            IValidationRule rule = new ValidationRule<T>(predicate, errorMessage, isDocumentChanging);
            rules.Add(rule);

            repairRules[rule] = new RepairRule<T>(repairAction);
        }
    }
}
