﻿using System.Collections.Generic;
using System.Globalization;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Attributes;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Styles;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.BorderEvaluation;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html.Export
{
    internal class HtmlExportContext : IHtmlExportContext
    {
        private readonly RadFlowDocument document;
        private readonly HtmlExportSettings settings;
        private readonly HtmlStyleRepository htmlStyleRepository;
        private readonly ImageRepository imageRepository;
        private readonly Dictionary<Table, TableBorderGrid> tableBorderGrids;

        public HtmlExportContext(RadFlowDocument document, HtmlExportSettings settings)
        {
            Guard.ThrowExceptionIfNull(document, "document");
            Guard.ThrowExceptionIfNull(settings, "settings");

            this.document = document;
            this.settings = settings;

            this.htmlStyleRepository = new HtmlStyleRepository();
            this.imageRepository = new ImageRepository();
            this.tableBorderGrids = new Dictionary<Table, TableBorderGrid>();
        }

        public CultureInfo Culture
        {
            get
            {
                return CultureInfo.InvariantCulture;
            }
        }

        public RadFlowDocument Document
        {
            get
            {
                return this.document;
            }
        }

        public HtmlExportSettings Settings
        {
            get
            {
                return this.settings;
            }
        }

        public HtmlStyleRepository HtmlStyleRepository
        {
            get
            {
                return this.htmlStyleRepository;
            }
        }

        public ImageRepository ImageRepository
        {
            get
            {
                return this.imageRepository;
            }
        }

        public void BeginExport()
        {
            foreach (Style style in this.Document.StyleRepository.GetSortedTopologicallyStyles())
            {
                this.RegisterStyleForExport(style);
            }
        }

        public void ClearParagraphClasses(ClassNamesCollection classes)
        {
            Guard.ThrowExceptionIfNull(classes, "classes");

            if (classes.Count == 2 &&
                classes.PeekFirst() == StyleNamesConverter.PrefixedNormalStyleId &&
                classes.PeekLast() == StyleNamesConverter.PrefixedNormalWebStyleId)
            {
                classes.Clear();
            }
        }

        public void EndExport()
        {
        }

        public void BeginExportTable(Table table)
        {
            TableBorderGrid borderGrid = new TableBorderGrid(table);
            if (!table.HasCellSpacing)
            {
                borderGrid.AssureValid();
            }

            this.tableBorderGrids.Add(table, borderGrid);
        }

        public void EndExportTable(Table table)
        {
            this.tableBorderGrids.Remove(table);
        }

        public TableBorderGrid GetBorderGrid(Table table)
        {
            return this.tableBorderGrids.GetValueOrNull(table);
        }

        private void RegisterStyleForExport(Style style)
        {
            // Register style as element and class selector if it is among the registered styles (h1-6, a, etc).
            string styleName;
            if (StyleToElementSelectorMappings.TryGetElementSelector(style.Id, out styleName))
            {
                this.RegisterStyleForExport(style, SelectorType.Elements, styleName);
            }

            styleName = StyleNamesConverter.ConvertStyleNameOnExport(style.Id);
            if (style.StyleType == StyleType.Table)
            {
                this.RegisterTableStyle(styleName, style);
            }
            else
            {
                this.RegisterStyleForExport(style, SelectorType.Style, styleName);
            }
        }

        private void RegisterStyleForExport(Style style, SelectorType selectorType, string styleName)
        {
            Selector selector = new Selector(styleName, selectorType);

            // Normal (Web) style is exported without its parent Normal style, so it should include all his properties.
            if (style.Id == BuiltInStyleNames.NormalWebStyleId &&
                style.BasedOnStyleId == BuiltInStyleNames.NormalStyleId)
            {
                selector.CopyFrom(this, this.Document.StyleRepository.GetStyle(BuiltInStyleNames.NormalStyleId));
            }

            selector.CopyFrom(this, style);
            this.HtmlStyleRepository.RegisterStyle(styleName, selector);
        }

        private Selector RegisterTableStyle(string styleName, Style style)
        {
            Selector selector = new Selector(styleName, SelectorType.StyleForElements, HtmlElementNames.Table);
            selector.CopyFrom(this, style.CharacterProperties);
            selector.CopyFrom(this, style.ParagraphProperties);
            selector.CopyFrom(this, style.TableProperties);

            this.HtmlStyleRepository.RegisterStyle(styleName, selector);

            if (style.TableProperties.Borders.HasLocalValue || style.TableCellProperties.VerticalAlignment.LocalValue != TableCell.VerticalAlignmentPropertyDefinition.DefaultValue)
            {
                selector = new Selector(styleName, SelectorType.StyleForElements, HtmlElementNames.TableCell);

                // export the borders in table cell style only if the outside and inner borders are equal.
                if (style.TableProperties.Borders.LocalValue != null)
                {
                    TableBorders borders = new TableBorders(style.TableProperties.Borders.LocalValue.Left);
                    if (style.TableProperties.Borders.LocalValue.Equals(borders))
                    {
                        selector.CopyFrom(this, style.TableProperties.GetStyleProperty(Table.TableBordersPropertyDefinition), style.TableProperties);
                    }
                }

                selector.CopyFrom(this, style.TableCellProperties);
                this.HtmlStyleRepository.RegisterStyle(styleName, selector);
            }
            return selector;
        }
    }
}
