﻿using System;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html
{
    /// <summary>
    /// Contains settings for export with <see cref="HtmlFormatProvider"/>.
    /// </summary>
    public class HtmlExportSettings
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="HtmlExportSettings" /> class.
        /// </summary>
        public HtmlExportSettings()
        {
            this.StylesExportMode = StylesExportMode.Embedded;
            this.ImagesExportMode = ImagesExportMode.Embedded;
            this.IndentDocument = false;
            this.BordersMinimalThickness = 0;
        }

        /// <summary>
        /// Occurs when exporting styles to external source.
        /// </summary>
        public event EventHandler<ExternalStylesExportingEventArgs> ExternalStylesExporting;

        /// <summary>
        /// Occurs when an image is exporting.
        /// </summary>
        public event EventHandler<ImageExportingEventArgs> ImageExporting;

        /// <summary>
        /// Gets or sets the path to the folder that will contain the external image files.
        /// </summary>
        public string ImagesFolderPath { get; set; }

        /// <summary>
        /// Gets or sets the base path that will be set as value to the 'src' attribute of the 'image' elements.
        /// </summary>
        public string ImagesSourceBasePath { get; set; }

        /// <summary>
        /// Gets or sets the path to the file that will contain the external styles.
        /// </summary>
        public string StylesFilePath { get; set; }

        /// <summary>
        /// Gets or sets the value that will be set as 'href' attribute of the 'link' element pointing to the file containing the external styles.
        /// </summary>
        public string StylesSourcePath { get; set; }

        /// <summary>
        /// Gets or sets the images export mode.
        /// </summary>
        /// <value>
        /// The images export mode.
        /// </value>
        public ImagesExportMode ImagesExportMode { get; set; }

        /// <summary>
        /// Gets or sets the export mode for the styles in <see cref="Flow.Model.RadFlowDocument.StyleRepository"/>. The default value is <see cref="T:StylesExportMode.Embedded"/>.
        /// </summary>
        /// <value>The styles export mode.</value>
        public StylesExportMode StylesExportMode { get; set; }

        /// <summary>
        /// Gets or sets the document export level.
        /// </summary>
        /// <value>The document export level.</value>
        public DocumentExportLevel DocumentExportLevel { get; set; }

        /// <summary>
        /// Gets or sets a value indicating whether the document should be indented. The default value is <c>false</c>.
        /// </summary>
        /// <value>If the document should be indented.</value>
        public bool IndentDocument { get; set; }

        /// <summary>
        /// Gets or sets the minimal thickness of a border.
        /// </summary>
        /// <value>
        /// The borders minimal thickness.
        /// </value>
        public double BordersMinimalThickness { get; set; }

        internal void OnExternalStylesExporting(ExternalStylesExportingEventArgs e)
        {
            Guard.ThrowExceptionIfNull(e, "e");

            if (this.ExternalStylesExporting != null)
            {
                this.ExternalStylesExporting(this, e);
            }
        }

        internal void OnImageExporting(ImageExportingEventArgs e)
        {
            Guard.ThrowExceptionIfNull(e, "e");

            if (this.ImageExporting != null)
            {
                this.ImageExporting(this, e);
            }
        }
    }
}
