﻿using System.Collections.Generic;
using System.IO;
using Telerik.Windows.Documents.Common.FormatProviders;
using Telerik.Windows.Documents.Flow.FormatProviders.Common;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Import;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html
{
    /// <summary>
    /// Represents HTML format provider.
    /// </summary>
    public class HtmlFormatProvider : TextBasedFormatProviderBase<RadFlowDocument>
    {
        private static readonly IEnumerable<string> supportedExtensions = new string[] { ".html", ".htm" };

        /// <summary>
        /// Initializes a new instance of the <see cref="HtmlFormatProvider"/> class.
        /// </summary>
        public HtmlFormatProvider()
        {
            this.ImportSettings = new HtmlImportSettings();
            this.ExportSettings = new HtmlExportSettings();
        }

        /// <summary>
        /// Gets the supported extensions.
        /// </summary>
        /// <value>
        /// The supported extensions.
        /// </value>
        public override IEnumerable<string> SupportedExtensions
        {
            get
            {
                return supportedExtensions;
            }
        }

        /// <summary>
        /// Gets a value indicating whether format provider can import.
        /// </summary>
        /// <value>The value indicating whether can import.</value>
        public override bool CanImport
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Gets a value indicating whether format provider can export.
        /// </summary>
        /// <value>The value indicating whether can export.</value>
        public override bool CanExport
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Gets or sets the import settings.
        /// </summary>
        /// <value>The import settings.</value>
        public HtmlImportSettings ImportSettings { get; set; }

        /// <summary>
        /// Gets or sets the export settings.
        /// </summary>
        /// <value>The export settings.</value>
        public HtmlExportSettings ExportSettings { get; set; }

        /// <summary>
        /// Exports the specified document.
        /// </summary>
        /// <param name="document">The document.</param>
        /// <param name="output">The output.</param>
        protected override void ExportOverride(RadFlowDocument document, Stream output)
        {
            Guard.ThrowExceptionIfNull(document, "document");
            Guard.ThrowExceptionIfNull(output, "output");

            using (HtmlWriter writer = new HtmlWriter(output, this.ExportSettings))
            {
                using (new RadFlowDocumentLicenseCheck(document))
                {
                    HtmlExporter exporter = new HtmlExporter();
                    HtmlExportContext context = new HtmlExportContext(document, this.ExportSettings);
                    exporter.Export(writer, context);
                }
            }
        }

        /// <summary>
        /// Imports the specified input.
        /// </summary>
        /// <param name="input">The input.</param>
        /// <returns>The imported document.</returns>
        protected override RadFlowDocument ImportOverride(Stream input)
        {
            Guard.ThrowExceptionIfNull(input, "input");

            HtmlImporter importer = new HtmlImporter();

            return importer.Import(input, new HtmlImportContext(this.ImportSettings));
        }
    }
}
