﻿using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html
{
    internal class HtmlTextProcessor
    {
        public const char Space = ' ';

        private const string NonBreakingSpaceHtml = "&nbsp;";
        private const string NonBreakingTabHtml = "&nbsp; &nbsp; ";
        private const char NonBreakingSpace = (char)160;
        private const char Tab = '\t';

        private static readonly string doubleSpace = string.Format("{0}{0}", Space);
        private static readonly string doubleSpaceHtml = string.Format("{0}{1}", Space, NonBreakingSpaceHtml);
        private static readonly Regex whiteSpacesPattern = new Regex(@"[ \f\n\r\t\v]+", RegexOptions.Compiled);
        private static readonly HtmlTextProcessor textProcessor;

        private readonly Dictionary<int, string> htmlEscape;

        static HtmlTextProcessor()
        {
            textProcessor = new HtmlTextProcessor();
        }

        private HtmlTextProcessor()
        {
            this.htmlEscape = new Dictionary<int, string>();

            // CO Controls and Basic Latin
            this.RegisterHtmlEscape(34, "&quot;");         // " quotation mark
            this.RegisterHtmlEscape(38, "&amp;");          // & ampersand
            this.RegisterHtmlEscape(60, "&lt;");           // < less then
            this.RegisterHtmlEscape(62, "&gt;");           // > greater then

            // ISO 8859-1 Symbol Entities
            this.RegisterHtmlEscape(153, "&trade;");       // ™ trademark
            this.RegisterHtmlEscape(8482, "&trade;");      // ™ trademark
            this.RegisterHtmlEscape(NonBreakingSpace, NonBreakingSpaceHtml);            // non-breaking space
            this.RegisterHtmlEscape(Tab, NonBreakingTabHtml);            // tab
            this.RegisterHtmlEscape(161, "&iexcl;");       // ¡ inverted exclamation mark
            this.RegisterHtmlEscape(162, "&cent;");        // ¢ cent
            this.RegisterHtmlEscape(163, "&pound;");       // £ British Pound Sterling
            this.RegisterHtmlEscape(8364, "&euro;");       // € Euro
            this.RegisterHtmlEscape(164, "&curren;");      // ¤ currency
            this.RegisterHtmlEscape(165, "&yen;");         // ¥ Japanese Yen
            this.RegisterHtmlEscape(166, "&brvbar;");      // ¦ broken vertical bar
            this.RegisterHtmlEscape(167, "&sect;");        // § section
            this.RegisterHtmlEscape(168, "&uml;");         // ¨ spacing diaeresis
            this.RegisterHtmlEscape(169, "&copy;");        // © copyright
            this.RegisterHtmlEscape(170, "&ordf;");        // ª feminine ordinal indicator
            this.RegisterHtmlEscape(171, "&laquo;");       // « left guillemets (used as quotation marks in some languages)
            this.RegisterHtmlEscape(172, "&not;");         // ¬ negation
            this.RegisterHtmlEscape(173, "&shy;");         // ¬ negation
            this.RegisterHtmlEscape(174, "&reg;");         // ® registered
            this.RegisterHtmlEscape(175, "&macr;");        // ¯ spacing macron
            this.RegisterHtmlEscape(176, "&deg;");         // ° degree - e.g. 45°
            this.RegisterHtmlEscape(177, "&plusmn;");      // ± plus/minus
            this.RegisterHtmlEscape(178, "&sup2;");        // ² superscript 2
            this.RegisterHtmlEscape(179, "&sup3;");        // ³ superscript 3
            this.RegisterHtmlEscape(180, "&acute;");       // ´ spacing acute
            this.RegisterHtmlEscape(181, "&micro;");       // µ micro
            this.RegisterHtmlEscape(182, "&para;");        // ¶ paragraph
            this.RegisterHtmlEscape(183, "&middot;");      // · middle dot
            this.RegisterHtmlEscape(184, "&cedil;");       // ¸ spacing cedilla
            this.RegisterHtmlEscape(185, "&sup1;");        // ¹ superscript 1
            this.RegisterHtmlEscape(186, "&ordm;");        // º masculine ordinal indicator
            this.RegisterHtmlEscape(187, "&raquo;");       // » right guillemets (used as quotation marks in some languages)
            this.RegisterHtmlEscape(188, "&frac14;");      // ¼ fraction 1/4
            this.RegisterHtmlEscape(189, "&frac12;");      // ½ fraction 1/2
            this.RegisterHtmlEscape(190, "&frac34;");      // ¾ fraction 3/4
            this.RegisterHtmlEscape(191, "&iquest;");      // ¿ inverted question mark
            this.RegisterHtmlEscape(215, "&times;");       // × multiplication
            this.RegisterHtmlEscape(247, "&divide;");      // ÷ divide

            // ISO 8859-1 Character Entities
            this.RegisterHtmlEscape(192, "&Agrave;");      // À capital a, grave accent
            this.RegisterHtmlEscape(193, "&Aacute;");      // Á capital a, acute accent
            this.RegisterHtmlEscape(194, "&Acirc;");       // Â capital a, circumflex accent
            this.RegisterHtmlEscape(195, "&Atilde;");      // Ã capital a, tilde
            this.RegisterHtmlEscape(196, "&Auml;");        // Ä capital a, umlaut mark
            this.RegisterHtmlEscape(197, "&Aring;");       // Å capital a, ring
            this.RegisterHtmlEscape(198, "&AElig;");       // Æ capital ae
            this.RegisterHtmlEscape(199, "&Ccedil;");      // Ç capital c, cedilla
            this.RegisterHtmlEscape(200, "&Egrave;");      // È capital e, grave accent
            this.RegisterHtmlEscape(201, "&Eacute;");      // É capital e, acute accent
            this.RegisterHtmlEscape(202, "&Ecirc;");       // Ê capital e, circumflex accent
            this.RegisterHtmlEscape(203, "&Euml;");        // Ë capital e, umlaut mark
            this.RegisterHtmlEscape(204, "&Igrave;");      // Ì capital i, grave accent
            this.RegisterHtmlEscape(205, "&Iacute;");      // Í capital i, acute accent
            this.RegisterHtmlEscape(206, "&Icirc;");       // Î capital i, circumflex accent
            this.RegisterHtmlEscape(207, "&Iuml;");        // Ï capital i, umlaut mark
            this.RegisterHtmlEscape(208, "&ETH;");         // Ð capital eth, Icelandic
            this.RegisterHtmlEscape(209, "&Ntilde;");      // Ñ capital n, tilde
            this.RegisterHtmlEscape(210, "&Ograve;");      // Ò capital o, grave accent
            this.RegisterHtmlEscape(211, "&Oacute;");      // Ó capital o, acute accent
            this.RegisterHtmlEscape(212, "&Ocirc;");       // Ô capital o, circumflex accent
            this.RegisterHtmlEscape(213, "&Otilde;");      // Õ capital o, tilde
            this.RegisterHtmlEscape(214, "&Ouml;");        // Ö capital o, umlaut mark
            this.RegisterHtmlEscape(216, "&Oslash;");      // Ø capital o, slash
            this.RegisterHtmlEscape(217, "&Ugrave;");      // Ù capital u, grave accent
            this.RegisterHtmlEscape(218, "&Uacute;");      // Ú capital u, acute accent
            this.RegisterHtmlEscape(219, "&Ucirc;");       // Û capital u, circumflex accent
            this.RegisterHtmlEscape(220, "&Uuml;");        // Ü capital u, umlaut mark
            this.RegisterHtmlEscape(221, "&Yacute;");      // Ý capital y, acute accent
            this.RegisterHtmlEscape(222, "&THORN;");       // Þ capital THORN, Icelandic
            this.RegisterHtmlEscape(223, "&szlig;");       // ß small sharp s, German
            this.RegisterHtmlEscape(224, "&agrave;");      // à small a, grave accent
            this.RegisterHtmlEscape(225, "&aacute;");      // á small a, acute accent
            this.RegisterHtmlEscape(226, "&acirc;");       // â small a, circumflex accent
            this.RegisterHtmlEscape(227, "&atilde;");      // ã small a, tilde
            this.RegisterHtmlEscape(228, "&auml;");        // ä small a, umlaut mark
            this.RegisterHtmlEscape(229, "&aring;");       // å small a, ring
            this.RegisterHtmlEscape(230, "&aelig;");       // æ small ae
            this.RegisterHtmlEscape(231, "&ccedil;");      // ç small c, cedilla
            this.RegisterHtmlEscape(232, "&egrave;");      // è small e, grave accent
            this.RegisterHtmlEscape(233, "&eacute;");      // é small e, acute accent
            this.RegisterHtmlEscape(234, "&ecirc;");       // ê small e, circumflex accent
            this.RegisterHtmlEscape(235, "&euml;");        // ë small e, umlaut mark
            this.RegisterHtmlEscape(236, "&igrave;");      // ì small i, grave accent
            this.RegisterHtmlEscape(237, "&iacute;");      // í small i, acute accent
            this.RegisterHtmlEscape(238, "&icirc;");       // î small i, circumflex accent
            this.RegisterHtmlEscape(239, "&iuml;");        // ï small i, umlaut mark
            this.RegisterHtmlEscape(240, "&eth;");         // ð small eth, Icelandic
            this.RegisterHtmlEscape(241, "&ntilde;");      // ñ small n, tilde
            this.RegisterHtmlEscape(242, "&ograve;");      // ò small o, grave accent
            this.RegisterHtmlEscape(243, "&oacute;");      // ó small o, acute accent
            this.RegisterHtmlEscape(244, "&ocirc;");       // ô small o, circumflex accent
            this.RegisterHtmlEscape(245, "&otilde;");      // õ small o, tilde
            this.RegisterHtmlEscape(246, "&ouml;");        // ö small o, umlaut mark
            this.RegisterHtmlEscape(248, "&oslash;");      // ø small o, slash
            this.RegisterHtmlEscape(249, "&ugrave;");      // ù small u, grave accent
            this.RegisterHtmlEscape(250, "&uacute;");      // ú small u, acute accent
            this.RegisterHtmlEscape(251, "&ucirc;");       // û small u, circumflex accent
            this.RegisterHtmlEscape(252, "&uuml;");        // ü small u, umlaut mark
            this.RegisterHtmlEscape(253, "&yacute;");      // ý small y, acute accent
            this.RegisterHtmlEscape(254, "&thorn;");       // þ small thorn, Icelandic
            this.RegisterHtmlEscape(255, "&yuml;");        // ÿ small y, umlaut mark

            // Latin Extended-A
            this.RegisterHtmlEscape(338, "&OElig;");       // Œ latin capital ligature
            this.RegisterHtmlEscape(339, "&oelig;");       // œ latin small ligature
            this.RegisterHtmlEscape(352, "&Scaron;");      // Š latin capital letter S with caron
            this.RegisterHtmlEscape(353, "&scaron;");      // š latin small letter s with caron
            this.RegisterHtmlEscape(376, "&Yuml;");        // Ÿ latin capital letter Y with diaeresis

            // Spacing Modifier Letters
            this.RegisterHtmlEscape(710, "&circ;");        // ˆ modifier letter circumflex accent
            this.RegisterHtmlEscape(732, "&tilde;");       // ˜ small tilde

            // General Punctuation
            this.RegisterHtmlEscape(8211, "&ndash;");      // – en dash
            this.RegisterHtmlEscape(8212, "&mdash;");      // — em dash, used e.g. to make a break in the flow of a sentence, a long line
            this.RegisterHtmlEscape(8216, "&lsquo;");      // ‘ left single curly quote
            this.RegisterHtmlEscape(8217, "&rsquo;");      // ’ right single curly quote
            this.RegisterHtmlEscape(8218, "&sbquo;");      // ‚ single low-9 quotation mark
            this.RegisterHtmlEscape(8220, "&ldquo;");      // “ left double curly quote
            this.RegisterHtmlEscape(8221, "&rdquo;");      // ” right double curly quote

            this.RegisterHtmlEscape(8222, "&bdquo;");      // „ double low-9 quotation mark
            this.RegisterHtmlEscape(8224, "&dagger;");     // † dagger
            this.RegisterHtmlEscape(8225, "&Dagger;");     // ‡ double dagger
            this.RegisterHtmlEscape(8240, "&permil;");     // ‰ per mille sign
            this.RegisterHtmlEscape(8249, "&lsaquo;");     // ‹ single left-pointing angle quotation mark
            this.RegisterHtmlEscape(8250, "&rsaquo;");     // › single right-pointing angle quotation mark
            this.RegisterHtmlEscape(402, "&fnof;");        // ƒ latin small f with hook

            this.RegisterHtmlEscape(8226, "&bull;");       // • bullet = black small circle
            this.RegisterHtmlEscape(8230, "&hellip;");     // … horizontal ellipsis = three dot leader
            this.RegisterHtmlEscape(8242, "&prime;");      // ′ prime = minutes = feet
            this.RegisterHtmlEscape(8243, "&Prime;");      // ″ double prime = seconds = inches
            this.RegisterHtmlEscape(8254, "&oline;");      // ‾ overline = spacing overscore
            this.RegisterHtmlEscape(8260, "&frasl;");      // ⁄ fraction slash

            // Mathematical Operators
            this.RegisterHtmlEscape(8704, "&forall;");     // ∀ for all
            this.RegisterHtmlEscape(8706, "&part;");       // ∂ partial differential
            this.RegisterHtmlEscape(8707, "&exist;");      // ∃ there exists
            this.RegisterHtmlEscape(8709, "&empty;");      // ∅ empty set = null set = diameter
            this.RegisterHtmlEscape(8711, "&nabla;");      // ∇ nabla = backward difference
            this.RegisterHtmlEscape(8712, "&isin;");       // ∈ element of
            this.RegisterHtmlEscape(8713, "&notin;");      // ∉ not an element of
            this.RegisterHtmlEscape(8715, "&ni;");         // ∋ contains as member
            this.RegisterHtmlEscape(8719, "&prod;");       // ∏ n-ary product = product sign
            this.RegisterHtmlEscape(8721, "&sum;");        // ∑ n-ary sumation
            this.RegisterHtmlEscape(8722, "&minus;");      // − minus sign, counterpart for the plus sign from keyboard
            this.RegisterHtmlEscape(8727, "&lowast;");     // ∗ asterisk operator
            this.RegisterHtmlEscape(8730, "&radic;");      // √ square root
            this.RegisterHtmlEscape(8733, "&prop;");       // ∝ proportional to
            this.RegisterHtmlEscape(8734, "&infin;");      // ∞ infinity
            this.RegisterHtmlEscape(8736, "&ang;");        // ∠ angle
            this.RegisterHtmlEscape(8743, "&and;");        // ∧ logical and = wedge
            this.RegisterHtmlEscape(8744, "&or;");         // ∨ logical or = vee
            this.RegisterHtmlEscape(8745, "&cap;");        // ∩ intersection = cap
            this.RegisterHtmlEscape(8746, "&cup;");        // ∪ union = cup
            this.RegisterHtmlEscape(8747, "&int;");        // ∫ integral
            this.RegisterHtmlEscape(8756, "&there4;");     // ∴ therefore
            this.RegisterHtmlEscape(8764, "&sim;");        // ∼ tilde operator = varies with = similar to
            this.RegisterHtmlEscape(8773, "&cong;");       // ≅ approximately equal to
            this.RegisterHtmlEscape(8776, "&asymp;");      // ≈ almost equal to = asymptotic to
            this.RegisterHtmlEscape(8800, "&ne;");         // ≠ not equal to
            this.RegisterHtmlEscape(8801, "&equiv;");      // ≡ identical to
            this.RegisterHtmlEscape(8804, "&le;");         // ≤ less-than or equal to
            this.RegisterHtmlEscape(8805, "&ge;");         // ≥ greater-than or equal to
            this.RegisterHtmlEscape(8834, "&sub;");        // ⊂ subset of
            this.RegisterHtmlEscape(8835, "&sup;");        // ⊃ superset of
            this.RegisterHtmlEscape(8836, "&nsub;");       // ⊄ not a subset of
            this.RegisterHtmlEscape(8838, "&sube;");       // ⊆ subset of or equal to
            this.RegisterHtmlEscape(8839, "&supe;");       // ⊇ superset of or equal to
            this.RegisterHtmlEscape(8853, "&oplus;");      // ⊕ circled plus = direct sum
            this.RegisterHtmlEscape(8855, "&otimes;");     // ⊗ circled times = vector product
            this.RegisterHtmlEscape(8869, "&perp;");       // ⊥ up tack = orthogonal to = perpendicular
            this.RegisterHtmlEscape(8901, "&sdot;");       // ⋅ dot operator

            // Greek Letters & Symbols Character Entity Reference
            this.RegisterHtmlEscape(913, "&Alpha;");       // Α greek capital letter alpha
            this.RegisterHtmlEscape(914, "&Beta;");        // Β greek capital letter beta
            this.RegisterHtmlEscape(915, "&Gamma;");       // Γ greek capital letter gamma
            this.RegisterHtmlEscape(916, "&Delta;");       // Δ greek capital letter delta
            this.RegisterHtmlEscape(917, "&Epsilon;");     // Ε greek capital letter epsilon
            this.RegisterHtmlEscape(918, "&Zeta;");        // Ζ greek capital letter zeta
            this.RegisterHtmlEscape(919, "&Eta;");         // Η greek capital letter eta
            this.RegisterHtmlEscape(920, "&Theta;");       // Θ greek capital letter theta
            this.RegisterHtmlEscape(921, "&Iota;");        // Ι greek capital letter iota
            this.RegisterHtmlEscape(922, "&Kappa;");       // Κ greek capital letter kappa
            this.RegisterHtmlEscape(923, "&Lambda;");      // Λ greek capital letter lambda
            this.RegisterHtmlEscape(924, "&Mu;");          // Μ greek capital letter mu
            this.RegisterHtmlEscape(925, "&Nu;");          // Ν greek capital letter nu
            this.RegisterHtmlEscape(926, "&Xi;");          // Ξ greek capital letter xi
            this.RegisterHtmlEscape(927, "&Omicron;");     // Ο greek capital letter omicron
            this.RegisterHtmlEscape(928, "&Pi;");          // Π greek capital letter pi
            this.RegisterHtmlEscape(929, "&Rho;");         // Ρ greek capital letter rho
            this.RegisterHtmlEscape(931, "&Sigma;");       // Σ greek capital letter sigma
            this.RegisterHtmlEscape(932, "&Tau;");         // Τ greek capital letter tau
            this.RegisterHtmlEscape(933, "&Upsilon;");     // Υ greek capital letter upsilon
            this.RegisterHtmlEscape(934, "&Phi;");         // Φ greek capital letter phi
            this.RegisterHtmlEscape(935, "&Chi;");         // Χ greek capital letter chi
            this.RegisterHtmlEscape(936, "&Psi;");         // Ψ greek capital letter psi
            this.RegisterHtmlEscape(937, "&Omega;");       // Ω greek capital letter omega
            this.RegisterHtmlEscape(945, "&alpha;");       // α greek small letter alpha
            this.RegisterHtmlEscape(946, "&beta;");        // β greek small letter beta
            this.RegisterHtmlEscape(947, "&gamma;");       // γ greek small letter gamma
            this.RegisterHtmlEscape(948, "&delta;");       // δ greek small letter delta
            this.RegisterHtmlEscape(949, "&epsilon;");     // ε greek small letter epsilon
            this.RegisterHtmlEscape(950, "&zeta;");        // ζ greek small letter zeta
            this.RegisterHtmlEscape(951, "&eta;");         // η greek small letter eta
            this.RegisterHtmlEscape(952, "&theta;");       // θ greek small letter theta
            this.RegisterHtmlEscape(953, "&iota;");        // ι greek small letter iota
            this.RegisterHtmlEscape(954, "&kappa;");       // κ greek small letter kappa
            this.RegisterHtmlEscape(955, "&lambda;");      // λ greek small letter lambda
            this.RegisterHtmlEscape(956, "&mu;");          // μ greek small letter mu
            this.RegisterHtmlEscape(957, "&nu;");          // ν greek small letter nu
            this.RegisterHtmlEscape(958, "&xi;");          // ξ greek small letter xi
            this.RegisterHtmlEscape(959, "&omicron;");     // ο greek small letter omicron
            this.RegisterHtmlEscape(960, "&pi;");          // π greek small letter pi
            this.RegisterHtmlEscape(961, "&rho;");         // ρ greek small letter rho
            this.RegisterHtmlEscape(962, "&sigmaf;");      // ς greek small letter final sigma
            this.RegisterHtmlEscape(963, "&sigma;");       // σ greek small letter sigma
            this.RegisterHtmlEscape(964, "&tau;");         // τ greek small letter tau
            this.RegisterHtmlEscape(965, "&upsilon;");     // υ greek small letter upsilon
            this.RegisterHtmlEscape(966, "&phi;");         // φ greek small letter phi
            this.RegisterHtmlEscape(967, "&chi;");         // χ greek small letter chi
            this.RegisterHtmlEscape(968, "&psi;");         // ψ greek small letter psi
            this.RegisterHtmlEscape(969, "&omega;");       // ω greek small letter omega
            this.RegisterHtmlEscape(977, "&thetasym;");    // ϑ greek small letter theta symbol
            this.RegisterHtmlEscape(978, "&upsih;");       // ϒ greek upsilon with hook symbol
            this.RegisterHtmlEscape(982, "&piv;");         // ϖ greek pi symbol
        }

        public static HtmlTextProcessor TextProcessor
        {
            get
            {
                return textProcessor;
            }
        }

        public static string RemoveMultipleWhiteSpaces(string text)
        {
            return whiteSpacesPattern.Replace(text, Space.ToString());
        }

        public static string TrimStart(string text)
        {
            return text.TrimStart(Space);
        }

        public static string TrimEnd(string text)
        {
            return text.TrimEnd(Space);
        }

        public static string Normalize(string text)
        {
            return text.Replace(NonBreakingSpace, Space);
        }

        public static string StripText(string currentText, string lastText)
        {
            if (HtmlTextProcessor.IsSpace(currentText.First()))
            {
                currentText = HtmlTextProcessor.TrimStart(currentText);
                if (!string.IsNullOrEmpty(lastText) && !HtmlTextProcessor.IsSpace(lastText.Last()))
                {
                    currentText = HtmlTextProcessor.Space + currentText;
                }
            }

            if (!string.IsNullOrEmpty(currentText) && HtmlTextProcessor.IsSpace(currentText.Last()))
            {
                currentText = HtmlTextProcessor.TrimEnd(currentText) + HtmlTextProcessor.Space;
            }

            return currentText;
        }

        public static bool IsSpace(char ch)
        {
            return ch == Space;
        }

        public static string PreserveWhiteSpaces(string text)
        {
            return text.Replace(Tab.ToString(), NonBreakingTabHtml)
                       .Replace(doubleSpace, doubleSpaceHtml);
        }

        public string EscapeHtml(string text)
        {
            StringBuilder result = new StringBuilder();
            string escape;
            for (int i = 0; i < text.Length; i++)
            {
                if (this.htmlEscape.TryGetValue(text[i], out escape))
                {
                    result.Append(escape);
                }
                else if (text[i] < 32)
                {
                    continue;
                }
                else
                {
                    result.Append(text[i]);
                }
            }

            return result.ToString();
        }

        private void RegisterHtmlEscape(int ch, string escape)
        {
            Guard.ThrowExceptionIfNullOrEmpty(escape, "escape");

            this.htmlEscape.Add(ch, escape);
        }
    }
}
