﻿// Copyright (c) 2012 James Treworgy
// 
// LICENSE (MIT License)
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy 
// of this software and associated documentation files (the "Software"), to deal 
// in the Software without restriction, including without limitation the rights to 
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
// of the Software, and to permit persons to whom the Software is furnished to do 
// so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all 
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Dynamic;
using CsQuery.Utility;
using CsQuery.ExtensionMethods;
using CsQuery.ExtensionMethods.Internal;
using CsQuery.Engine;
using CsQuery.Implementation;

namespace CsQuery
{
    internal partial class CQ
    {

        /// <summary>
        /// Store arbitrary data associated with the specified element, and render it as JSON on the
        /// element in a format that can be read by the jQuery "Data()" methods.
        /// </summary>
        ///
        /// <param name="key">
        /// The name of the key to associate with this data object.
        /// </param>
        /// <param name="data">
        /// An string to be associated with the key.
        /// </param>
        ///
        /// <returns>
        /// The current CQ object.
        /// </returns>
        ///
        /// <url>
        /// http://api.jquery.com/data/#data1
        /// </url>

        public CQ Data(string key, string data)
        {
            foreach (IDomElement e in Elements)
            {
                e.SetAttribute("data-" + key, data);
            }
            return this;
        }

        /// <summary>
        /// Remove all data- attributes from the element.
        /// </summary>
        ///
        /// <returns>
        /// The current CQ object.
        /// </returns>
        ///
        /// <url>
        /// http://api.jquery.com/removeData/
        /// </url>

        public CQ RemoveData()
        {
            return RemoveData((string)null);
        }

        /// <summary>
        /// Remove a previously-stored piece of data identified by a key.
        /// </summary>
        ///
        /// <param name="key">
        /// A string naming the piece of data to delete, or pieces of data if the string has multiple
        /// values separated by spaces.
        /// </param>
        ///
        /// <returns>
        /// THe current CQ object.
        /// </returns>
        ///
        /// <url>
        /// http://api.jquery.com/removeData/
        /// </url>

        public CQ RemoveData(string key)
        {
            foreach (IDomElement el in Elements)
            {
                List<string> toRemove = new List<string>();
                foreach (var kvp in el.Attributes)
                {
                    bool match = String.IsNullOrEmpty(key) ?
                        kvp.Key.StartsWith("data-") :
                        kvp.Key == "data-" + key;
                    if (match)
                    {
                        toRemove.Add(kvp.Key);
                    }
                }
                foreach (string attr in toRemove)
                {
                    el.RemoveAttribute(attr);
                }
            }
            return this;
        }


        /// <summary>
        /// Remove all data from an element.
        /// </summary>
        ///
        /// <param name="keys">
        /// An array or space-separated string naming the pieces of data to delete.
        /// </param>
        ///
        /// <returns>
        /// The current CQ object.
        /// </returns>
        ///
        /// <url>
        /// http://api.jquery.com/removeData/
        /// </url>

        public CQ RemoveData(IEnumerable<string> keys)
        {
            foreach (var key in keys)
            {
                RemoveData(key);

            }
            return this;
        }




        /// <summary>
        /// Returns data as a string, with no attempt to parse it from JSON. This is the equivalent of
        /// using the Attr("data-{key}") method.
        /// </summary>
        ///
        /// <param name="key">
        /// The key identifying the data.
        /// </param>
        ///
        /// <returns>
        /// A string.
        /// </returns>

        public string DataRaw(string key)
        {
            return First().Attr("data-" + key);
        }

        /// <summary>
        /// Determine whether an element has any jQuery data associated with it.
        /// </summary>
        ///
        /// <returns>
        /// true if there is any data, false if not.
        /// </returns>
        ///
        /// <url>
        /// http://api.jquery.com/jQuery.hasData/
        /// </url>

        public bool HasData()
        {
            foreach (IDomElement el in Elements)
            {
                foreach (var kvp in el.Attributes)
                {
                    if (kvp.Key.StartsWith("data-"))
                    {
                        return true;
                    }
                }
            }
            return false;
        }

        
    }
}
