﻿// Copyright (c) 2012 James Treworgy
// 
// LICENSE (MIT License)
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy 
// of this software and associated documentation files (the "Software"), to deal 
// in the Software without restriction, including without limitation the rights to 
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
// of the Software, and to permit persons to whom the Software is furnished to do 
// so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all 
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CsQuery.Utility;
using CsQuery.ExtensionMethods;
using CsQuery.ExtensionMethods.Internal;
using CsQuery.Engine;
using CsQuery.Implementation;

namespace CsQuery
{
    internal partial class CQ
    {

        /// <summary>
        /// Reduce the set of matched elements to those that match the selector or pass the function's
        /// test.
        /// </summary>
        ///
        /// <param name="selector">
        /// A string containing a selector expression to match the current set of elements against.
        /// </param>
        ///
        /// <returns>
        /// A new CQ object.
        /// </returns>
        ///
        /// <url>
        /// http://api.jquery.com/filter/
        /// </url>

        public CQ Filter(string selector)
        {
            return NewInstance(FilterElements(SelectionSet, selector));

        }

        /// <summary>
        /// Reduce the set of matched elements to those that matching the element passed by parameter.
        /// </summary>
        ///
        /// <param name="element">
        /// The element to match.
        /// </param>
        ///
        /// <returns>
        /// A new CQ object.
        /// </returns>
        ///
        /// <url>
        /// http://api.jquery.com/filter/
        /// </url>

        public CQ Filter(IDomObject element)
        {
            return Filter(Objects.Enumerate(element));
        }

        /// <summary>
        /// Reduce the set of matched elements to those matching any of the elements in a sequence passed
        /// by parameter.
        /// </summary>
        ///
        /// <param name="elements">
        /// The elements to match.
        /// </param>
        ///
        /// <returns>
        /// A new CQ object.
        /// </returns>
        ///
        /// <url>
        /// http://api.jquery.com/filter/
        /// </url>

        public CQ Filter(IEnumerable<IDomObject> elements)
        {
            CQ filtered = NewInstance(this);
            filtered.SelectionSet.IntersectWith(elements);
            return filtered;
        }

        /// <summary>
        /// Reduce the set of matched elements to those that match the selector or pass the function's
        /// test.
        /// </summary>
        ///
        /// <remarks>
        /// This method doesn't offer anything that can't easily be accomplished with a LINQ "where"
        /// query but is included for completeness.
        /// </remarks>
        ///
        /// <param name="function">
        /// A function used as a test for each element in the set.
        /// </param>
        ///
        /// <returns>
        /// A new CQ object.
        /// </returns>
        ///
        /// <url>
        /// http://api.jquery.com/filter/
        /// </url>

        public CQ Filter(Func<IDomObject, bool> function)
        {
            CQ output = NewCqInDomain();

            List<IDomObject> filteredList = new List<IDomObject>();
            foreach (IDomObject obj in SelectionSet)
            {
                if (function(obj))
                {
                    filteredList.Add(obj);
                }
            }
            return output.SetSelection(filteredList, Order);
        }

        /// <summary>
        /// Reduce the set of matched elements to those that match the selector or pass the function's
        /// test.
        /// </summary>
        ///
        /// <remarks>
        /// This method doesn't offer anything that can't easily be accomplished with a LINQ "where"
        /// query but is included for completeness.
        /// </remarks>
        ///
        /// <param name="function">
        /// A function used as a test for each element in the set.
        /// </param>
        ///
        /// <returns>
        /// A new CQ object.
        /// </returns>
        ///
        /// <url>
        /// http://api.jquery.com/filter/
        /// </url>

        public CQ Filter(Func<IDomObject, int, bool> function)
        {
            CQ output = NewCqInDomain();
            List<IDomObject> filteredList = new List<IDomObject>();

            int index = 0;
            foreach (IDomObject obj in SelectionSet)
            {
                if (function(obj, index++))
                {
                    filteredList.Add(obj);
                }
            }

            return output.SetSelection(filteredList, Order);
        }
        

    }
}
