﻿// Copyright (c) 2012 James Treworgy
// 
// LICENSE (MIT License)
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy 
// of this software and associated documentation files (the "Software"), to deal 
// in the Software without restriction, including without limitation the rights to 
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
// of the Software, and to permit persons to whom the Software is furnished to do 
// so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all 
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CsQuery.Utility;
using CsQuery.ExtensionMethods;
using CsQuery.ExtensionMethods.Internal;
using CsQuery.Engine;
using CsQuery.Implementation;

namespace CsQuery
{
    internal partial class CQ
    {
        
        /// <summary>
        /// Remove all selected elements from the Document.
        /// </summary>
        ///
        /// <param name="selector">
        /// A selector expression that filters the set of matched elements to be removed.
        /// </param>
        ///
        /// <returns>
        /// The current CQ object.
        /// </returns>
        ///
        /// <url>
        /// http://api.jquery.com/remove/
        /// </url>

        public CQ Remove(string selector = null)
        {
            var list = !String.IsNullOrEmpty(selector) ?
                Filter(selector).SelectionSet :
                SelectionSet;

            // We need to copy first because selection can change
            List<IDomObject> removeList = list.ToList();

            List<bool> disconnected = list.Select(item => item.IsDisconnected ||
                item.Document != Document).ToList();

            
            for (int index = 0; index < removeList.Count; index++)
            {
                var el = removeList[index];
                if (disconnected[index])
                {
                    list.Remove(el);
                }
                if (el.ParentNode != null)
                {
                    el.Remove();
                }
            }
            return this;
        }

        /// <summary>
        /// This is synonymous with Remove in CsQuery, since there's nothing associated with an element
        /// that is not rendered. It is included for compatibility.
        /// </summary>
        ///
        /// <remarks>
        /// CsQuery does not maintain data such as initial visibility state when using Show/Hide, or an
        /// internal data structure when using Data methods. There is no data associated with an element
        /// that is not represented entirely through the markup that it will render. In the future, it's
        /// possible we may add such functionality for certain features, so it may be desirable to use
        /// Detach instead of Remove in those situations. This ensures forward compatibility.
        /// </remarks>
        ///
        /// <param name="selector">
        /// A selector expression that filters the set of matched elements to be removed.
        /// </param>
        ///
        /// <returns>
        /// A new CQ object.
        /// </returns>

        public CQ Detach(string selector = null)
        {
            return Remove(selector);
        }
        

    }
}
