﻿// Copyright (c) 2012 James Treworgy
// 
// LICENSE (MIT License)
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy 
// of this software and associated documentation files (the "Software"), to deal 
// in the Software without restriction, including without limitation the rights to 
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
// of the Software, and to permit persons to whom the Software is furnished to do 
// so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all 
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CsQuery.Utility;
using CsQuery.ExtensionMethods;
using CsQuery.ExtensionMethods.Internal;
using CsQuery.Engine;
using CsQuery.Implementation;

namespace CsQuery
{
    internal partial class CQ
    {
        /// <summary>
        /// Set the CSS width of each element in the set of matched elements.
        /// </summary>
        ///
        /// <remarks>
        /// There is no Width() method in CsQuery because this is a value calculated by the browser.
        /// While we can set the CSS of an element, it would be futile to try to return a useful value.
        /// If you want to inspect the current CSS width for an element, please use Css() methods
        /// instead. This ensures there is no confusion about the use of Width() in CsQuery.
        /// </remarks>
        ///
        /// <param name="value">
        /// An integer representing the number of pixels.
        /// </param>
        ///
        /// <returns>
        /// The current CQ object.
        /// </returns>
        ///
        /// <url>
        /// http://api.jquery.com/width/#width2
        /// </url>

        public CQ Width(int value)
        {
            return Width(value.ToString() + "px");
        }

        /// <summary>
        /// Set the CSS width of each element in the set of matched elements.
        /// </summary>
        ///
        /// <remarks>
        /// There are no Height() or Width() methods  in CsQuery because these are value calculated by
        /// the browser that depend on the page layout, as well as things like the browser window size
        /// which don't even exist in CsQuery. While we can set the CSS of an element, it would be futile
        /// to try to return a useful value. If you want to inspect the current CSS width for an element,
        /// please use Css() methods instead. This ensures there is no confusion about the use of Width()
        /// and Height()
        /// in CsQuery.
        /// </remarks>
        ///
        /// <param name="value">
        /// An integer along with a unit of measure appended (as a string), e.g. "100px".
        /// </param>
        ///
        /// <returns>
        /// The current CQ object.
        /// </returns>
        ///
        /// <url>
        /// http://api.jquery.com/width/#width2
        /// </url>

        public CQ Width(string value)
        {
            return Css("width", value);
        }

        /// <summary>
        /// Set the CSS width of each element in the set of matched elements.
        /// </summary>
        ///
        /// <remarks>
        /// There are no Height() or Width() methods  in CsQuery because these are value calculated by
        /// the browser that depend on the page layout, as well as things like the browser window size
        /// which don't even exist in CsQuery. While we can set the CSS of an element, it would be futile
        /// to try to return a useful value. If you want to inspect the current CSS width for an element,
        /// please use Css() methods instead. This ensures there is no confusion about the use of Width()
        /// and Height()
        /// in CsQuery.
        /// </remarks>
        ///
        /// <param name="value">
        /// An integer representing the number of pixels.
        /// </param>
        ///
        /// <returns>
        /// The current CQ object.
        /// </returns>
        ///
        /// <url>
        /// http://api.jquery.com/height/#height2
        /// </url>

        public CQ Height(int value)
        {
            return Height(value.ToString() + "px");
        }

        /// <summary>
        /// Set the CSS height of each element in the set of matched elements.
        /// </summary>
        ///
        /// <remarks>
        /// There are no Height() or Width() methods  in CsQuery because these are value calculated by
        /// the browser that depend on the page layout, as well as things like the browser window size
        /// which don't even exist in CsQuery. While we can set the CSS of an element, it would be futile
        /// to try to return a useful value. If you want to inspect the current CSS width for an element,
        /// please use Css() methods instead. This ensures there is no confusion about the use of Width()
        /// and Height()
        /// in CsQuery.
        /// </remarks>
        ///
        /// <param name="value">
        /// An integer along with a unit of measure appended (as a string), e.g. "100px".
        /// </param>
        ///
        /// <returns>
        /// The current CQ object.
        /// </returns>
        ///
        /// <url>
        /// http://api.jquery.com/height/#height2
        /// </url>

        public CQ Height(string value)
        {
            return Css("height", value);
        }
        

    }
}
