﻿// Copyright (c) 2012 James Treworgy
// 
// LICENSE (MIT License)
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy 
// of this software and associated documentation files (the "Software"), to deal 
// in the Software without restriction, including without limitation the rights to 
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
// of the Software, and to permit persons to whom the Software is furnished to do 
// so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all 
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CsQuery.Implementation
{
    /// <summary>
    /// A CSS style sheet.
    /// </summary>
    ///
    /// <url>
    /// http://www.w3.org/TR/DOM-Level-2-Style/css.html#CSS-CSSStyleSheet
    /// </url>

    internal class CSSStyleSheet : ICSSStyleSheet
    {
        #region constructors

        /// <summary>
        /// Constructor.
        /// </summary>
        ///
        /// <param name="ownerNode">
        /// The node that owns this item.
        /// </param>

        public CSSStyleSheet(IDomElement ownerNode)
        {
            OwnerNode = ownerNode;
        }

        #endregion

        #region private properties

        private IList<ICSSRule> _Rules;

        #endregion

        #region public properties

        /// <summary>
        /// Indicates whether the style sheet is applied to the document.
        /// </summary>

        public bool Disabled
        {
            get;set;
        }

        /// <summary>
        /// If the style sheet is a linked style sheet, the value of its attribute is its location. For
        /// inline style sheets, the value of this attribute is null.
        /// </summary>

        public string Href
        {
            get
            {
                return OwnerNode == null ? 
                    null : 
                    OwnerNode["href"];
            }
            set
            {
                if (OwnerNode == null)
                {
                    throw new InvalidOperationException("This CSSStyleSheet is not bound to an element node.");
                }
                OwnerNode["href"] = value;
            }
        }

        /// <summary>
        /// The node that associates this style sheet with the document. For HTML, this may be the
        /// corresponding LINK or STYLE element.
        /// </summary>
        ///
        /// <value>
        /// The owner node.
        /// </value>

        public IDomElement OwnerNode
        {
            get;
            protected set;
        }

        /// <summary>
        /// This specifies the style sheet language for this style sheet. This will always be "text/css".
        /// </summary>
        ///
        /// <value>
        /// The type.
        /// </value>

        public string Type
        {
            get { return "text/css"; }
        }

        /// <summary>
        /// Gets the CSS rules for this style sheet.
        /// </summary>
        ///
        /// <value>
        /// The CSS rules.
        /// </value>

        public IList<ICSSRule> CssRules
        {
            get {
                if (_Rules == null)
                {
                    _Rules = new List<ICSSRule>();
                }
                return _Rules;
            }
        }

        #endregion
    }
}
