﻿// Copyright (c) 2012 James Treworgy
// 
// LICENSE (MIT License)
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy 
// of this software and associated documentation files (the "Software"), to deal 
// in the Software without restriction, including without limitation the rights to 
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
// of the Software, and to permit persons to whom the Software is furnished to do 
// so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all 
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CsQuery.Implementation
{
    /// <summary>
    /// A CDATA node
    /// </summary>

    internal class DomCData : DomObject<DomCData>, IDomCData
    {

        /// <summary>
        /// Default constructor.
        /// </summary>

        public DomCData()
            : base()
        {
            _NonAttributeData = "";
        }

        /// <summary>
        /// Constructor that populates the node with the passed value.
        /// </summary>
        ///
        /// <param name="value">
        /// The contents of the CDATA node
        /// </param>

        public DomCData(string value)
            : base()
        {
            NodeValue = value;
        }

        private string _NonAttributeData;

        /// <summary>
        /// Gets or sets the node value. For CDATA nodes, this is the content.
        /// </summary>

        public override string NodeValue
        {
            get
            {
                return NonAttributeData;
            }
            set
            {
                NonAttributeData = value;
            }
        }

        /// <summary>
        /// Gets the type of the node. For CDATA nodes, this is NodeType.CDATA_SECTION_NODE.
        /// </summary>

        public override NodeType NodeType
        {
            get { return NodeType.CDATA_SECTION_NODE; }
        }


        #region IDomSpecialElement Members

        /// <summary>
        /// Gets or sets the non-attribute data in the tag. For CDATA nodes, this is the same as the
        /// content of the node..
        /// </summary>

        public string NonAttributeData
        {
            get
            {
                return _NonAttributeData;
            }
            set
            {
                _NonAttributeData = value ?? "";
            }
        }

        /// <summary>
        /// Gets a value indicating whether HTML is allowed as a child of this element. For CDATA nodes,
        /// this is always false.
        /// </summary>

        public override bool InnerHtmlAllowed
        {
            get { return false; }
        }

        /// <summary>
        /// Gets a value indicating whether this object has children. For CDATA nodes, this is always
        /// false.
        /// </summary>

        public override bool HasChildren
        {
            get { return false; }
        }

        /// <summary>
        /// Gets or sets the text of the CDATA element.
        /// </summary>

        public string Text
        {
            get
            {
                return NonAttributeData;
            }
            set
            {
                NonAttributeData = value;
            }
        }

        /// <summary>
        /// Makes a deep copy of this object.
        /// </summary>
        ///
        /// <returns>
        /// A copy of this object.
        /// </returns>

        public override DomCData Clone()
        {
            DomCData clone = new DomCData();
            clone.NonAttributeData = NonAttributeData;
            return clone;
        }

        IDomNode IDomNode.Clone()
        {
            return Clone();
        }
        object ICloneable.Clone()
        {
            return Clone();
        }
        #endregion

      
    }
    
}
