﻿// Copyright (c) 2012 James Treworgy
// 
// LICENSE (MIT License)
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy 
// of this software and associated documentation files (the "Software"), to deal 
// in the Software without restriction, including without limitation the rights to 
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
// of the Software, and to permit persons to whom the Software is furnished to do 
// so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all 
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System.Linq;
using CsQuery.HtmlParser;

namespace CsQuery.Implementation
{
    /// <summary>
    /// An implementation of properties shared by all form submittable elements.
    /// </summary>
    /// 
    /// <url>
    /// http://www.w3.org/html/wg/drafts/html/master/forms.html#category-submit
    /// </url>
    internal class FormSubmittableElement : FormReassociateableElement, IFormSubmittableElement
    {
        /// <summary>
        /// Constructor to specify the element's token ID.
        /// </summary>
        /// <param name="tokenId">The token ID of the element.</param>
        protected FormSubmittableElement(ushort tokenId)
            : base(tokenId)
        {
        }

        /// <summary>
        /// A form submittable element is disabled if it has the disabled attribute,
        /// or it is in a disabled fieldset and not in the legend.
        /// </summary>
        /// 
        /// <url>
        /// http://www.w3.org/html/wg/drafts/html/master/forms.html#attr-fe-disabled
        /// </url>
        public override bool Disabled
        {
            get
            {
                if (base.Disabled)
                {
                    return true;
                }

                IDomContainer fieldset = GetAncestors().FirstOrDefault(c => c.NodeName == "FIELDSET");
                if (fieldset == null || !fieldset.Disabled)
                {
                    return false;
                }

                IDomElement firstLegend = fieldset.GetDescendentElements().FirstOrDefault(o => o.NodeName == "LEGEND");
                return firstLegend == null || !GetAncestors().Contains(firstLegend);
            }
            set
            {
                SetProp(HtmlData.attrDISABLED, value);
            }
        }
    }
}
