﻿// Copyright (c) 2012 James Treworgy
// 
// LICENSE (MIT License)
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy 
// of this software and associated documentation files (the "Software"), to deal 
// in the Software without restriction, including without limitation the rights to 
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
// of the Software, and to permit persons to whom the Software is furnished to do 
// so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all 
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CsQuery.Engine
{
    /// <summary>
    /// A case-insensitive character equality comparer.
    /// </summary>

    internal class CharacterEqualityComparer : EqualityComparer<char>
    {
        /// <summary>
        /// Factory that creates an EqualityComparer&lt;char&gt; based on the case sensitivity
        /// requirements.
        /// </summary>
        ///
        /// <param name="isCaseSensitive">
        /// true if the EqualityComparer should be case sensitive.
        /// </param>
        ///
        /// <returns>
        /// An EqualityComparer&lt;char&gt;
        /// </returns>

        public static CharacterEqualityComparer Create(bool isCaseSensitive) {
            return isCaseSensitive ? 
                new CaseSensitiveCharacterEqualityComparer():
                new CharacterEqualityComparer();
        }

        /// <summary>
        /// Tests if two char objects are considered equal.
        /// </summary>
        ///
        /// <param name="x">
        /// Character to be compared.
        /// </param>
        /// <param name="y">
        /// Character to be compared.
        /// </param>
        ///
        /// <returns>
        /// true if the objects are considered equal, false if they are not.
        /// </returns>

        public override bool Equals(char x, char y)
        {
            return x.Equals(y);
                
        }

        /// <summary>
        /// Calculates the hash code for this object.
        /// </summary>
        ///
        /// <param name="obj">
        /// The object.
        /// </param>
        ///
        /// <returns>
        /// The hash code for this object.
        /// </returns>

        public override int GetHashCode(char obj)
        {
            return obj.GetHashCode();
               
        }
    }

    /// <summary>
    /// A case-sensitive character equality comparer.
    /// </summary>

    internal class CaseSensitiveCharacterEqualityComparer : CharacterEqualityComparer
    {
        /// <summary>
        /// Tests if two char objects are considered equal.
        /// </summary>
        ///
        /// <param name="x">
        /// Character to be compared.
        /// </param>
        /// <param name="y">
        /// Character to be compared.
        /// </param>
        ///
        /// <returns>
        /// true if the objects are considered equal, false if they are not.
        /// </returns>

        public override bool Equals(char x, char y)
        {
            return Char.ToLower(x).Equals(Char.ToLower(y));
        }

        /// <summary>
        /// Calculates the hash code for this object.
        /// </summary>
        ///
        /// <param name="obj">
        /// The object.
        /// </param>
        ///
        /// <returns>
        /// The hash code for this object.
        /// </returns>

        public override int  GetHashCode(char obj)
        {
            return Char.ToLower(obj).GetHashCode();
        }
    }
}
