﻿// Copyright (c) 2012 James Treworgy
// 
// LICENSE (MIT License)
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy 
// of this software and associated documentation files (the "Software"), to deal 
// in the Software without restriction, including without limitation the rights to 
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
// of the Software, and to permit persons to whom the Software is furnished to do 
// so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all 
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CsQuery.Engine;

namespace CsQuery
{
    /// <summary>
    /// The default DomIndexProvider.
    /// </summary>

    internal static class DomIndexProviders
    {
        /// <summary>
        /// Static constructor.
        /// </summary>

        static DomIndexProviders()
        {
            _RangedDomIndexProvider = new RangedDomIndexProvider();
            _SimpleDomIndexProvider = new SimpleDomIndexProvider();
            _NoDomIndexProvider = new NoDomIndexProvider();
        }

        private static IDomIndexProvider _RangedDomIndexProvider;
        private static IDomIndexProvider _SimpleDomIndexProvider;
        private static IDomIndexProvider _NoDomIndexProvider;
        
        /// <summary>
        /// Return a SimpleDomIndex provider
        /// </summary>
        ///
        /// <returns>
        /// The DomIndex instance
        /// </returns>

        public static IDomIndexProvider Simple
        {
            get
            {
                return _SimpleDomIndexProvider;
            }
        }

        /// <summary>
        /// Returns a RangedDomIndex provider
        /// </summary>

        public static IDomIndexProvider Ranged
        {
            get
            {
                return _RangedDomIndexProvider;
            }
        }

        /// <summary>
        /// Returns a NoDomIndex provider
        /// </summary>

        public static IDomIndexProvider None
        {
            get
            {
                return _NoDomIndexProvider;
            }
        }
    }

    /// <summary>
    ///  DomIndexProvider returning a SimpleDomIndex
    /// </summary>

    internal class SimpleDomIndexProvider : IDomIndexProvider
    {
        /// <summary>
        /// Return an instance of a DomIndex class.
        /// </summary>
        ///
        /// <returns>
        /// The DomIndex instance
        /// </returns>

        public IDomIndex GetDomIndex()
        {
            return new DomIndexSimple();
        }


    }

    /// <summary>
    /// DomIndexProvider returning a RangedDomIndex
    /// </summary>

    internal class RangedDomIndexProvider : IDomIndexProvider
    {
        /// <summary>
        /// Return an instance of a DomIndex class.
        /// </summary>
        ///
        /// <returns>
        /// The DomIndex instance
        /// </returns>

        public IDomIndex GetDomIndex()
        {
            return new DomIndexRanged();
        }
    }

    /// <summary>
    /// DomIndexProvider returning a RangedDomIndex
    /// </summary>

    internal class NoDomIndexProvider : IDomIndexProvider
    {
        /// <summary>
        /// Return an instance of a DomIndex class.
        /// </summary>
        ///
        /// <returns>
        /// The DomIndex instance
        /// </returns>

        public IDomIndex GetDomIndex()
        {
            return new DomIndexNone();
        }
    }
}
