﻿// Copyright (c) 2012 James Treworgy
// 
// LICENSE (MIT License)
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy 
// of this software and associated documentation files (the "Software"), to deal 
// in the Software without restriction, including without limitation the rights to 
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
// of the Software, and to permit persons to whom the Software is furnished to do 
// so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all 
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CsQuery.EquationParser
{
    /// <summary>
    /// Interface for a strongly-typed equation.
    /// </summary>
    ///
    /// <typeparam name="T">
    /// The type of value returned by the equation.
    /// </typeparam>

    internal interface IEquation<T> : IOperand<T>, IEquation where T : IConvertible
    {
        /// <summary>
        /// Execute the equation and return the result
        /// </summary>
        ///
        /// <param name="values">
        /// The values of the variables for this equation, in the order the variables were created.
        /// </param>
        ///
        /// <returns>
        /// The value.
        /// </returns>

        new T GetValue(params IConvertible[] values);

        /// <summary>
        /// Execute the equation; if an error occurs, return false.
        /// </summary>
        ///
        /// <param name="result">
        /// [out] The result.
        /// </param>
        ///
        /// <returns>
        /// true if it succeeds, false if it fails.
        /// </returns>

        bool TryGetValue(out T result);

        /// <summary>
        /// Execute the equation; if an error occurs, return false.
        /// </summary>
        ///
        /// <param name="result">
        /// [out] The result.
        /// </param>
        /// <param name="values">
        /// The values of the variables for this equation, in the order the variables were created.
        /// </param>
        ///
        /// <returns>
        /// true if it succeeds, false if it fails.
        /// </returns>

        bool TryGetValue(out T result, params IConvertible[] values);

        /// <summary>
        /// Makes a deep copy of this object.
        /// </summary>
        ///
        /// <returns>
        /// A copy of this object.
        /// </returns>

        new IEquation<T> Clone();

        /// <summary>
        /// Clone the equation, changing the output type.
        /// </summary>
        ///
        /// <typeparam name="U">
        /// Generic type parameter.
        /// </typeparam>
        ///
        /// <returns>
        /// A clone of the equation that returns type U.
        /// </returns>

        IEquation<U> CloneAs<U>() where U : IConvertible;
    }
}
