﻿// Copyright (c) 2012 James Treworgy
// 
// LICENSE (MIT License)
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy 
// of this software and associated documentation files (the "Software"), to deal 
// in the Software without restriction, including without limitation the rights to 
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
// of the Software, and to permit persons to whom the Software is furnished to do 
// so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all 
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;

namespace CsQuery.Implementation
{
    /// <summary>
    /// Creates a single virtual stream out of multiple input streams.
    /// </summary>

    internal class CombinedStream : BaseStream
    {
        bool _valid;
        readonly IEnumerator<Stream> _streams;

        /// <summary>
        /// Creates a single virtual stream out of multiple input streams.
        /// </summary>
        ///
        /// <param name="streams">
        /// The streams.
        /// </param>

        public CombinedStream(params Stream[] streams) : 
            this((IEnumerable<Stream>)streams) 
        { }

        /// <summary>
        /// Creates a single virtual stream out of multiple input streams.
        /// </summary>
        ///
        /// <param name="streams">
        /// The streams.
        /// </param>

        public CombinedStream(IEnumerable<Stream> streams)
        {
            _streams = streams.GetEnumerator();
            _valid = _streams.MoveNext();
        }

        /// <summary>
        /// Gets a value indicating whether we can read.
        /// </summary>

        public override bool CanRead { 
            get {
                return _valid && _streams.Current.CanRead; 
            } 
        }
        /// <summary> Reads from the next stream available </summary>
        public override int Read(byte[] buffer, int offset, int count)
        {
            if (count == 0) return 0;
            while (_valid)
            {
                int len = _streams.Current.Read(buffer, offset, count);
                if (len > 0)
                    return len;

                //_streams.Current.Dispose();
                _valid = _streams.MoveNext();
            }

            return 0;
        }
        ///// <summary> Disposes of all remaining streams. </summary>
        //protected override void Dispose(bool disposing)
        //{
        //    while (disposing && _valid)
        //    {
        //        _streams.Current.Dispose();
        //        _valid = _streams.MoveNext();
        //    }

        //    base.Dispose(disposing);
        //}
    }

    /// <summary>
    /// Base stream implementation
    /// </summary>

    internal abstract class BaseStream : Stream
    {
        /// <summary>
        /// When overridden in a derived class, gets a value indicating whether the current stream supports reading.
        /// </summary>
        public override bool CanRead { get { return false; } }

        /// <summary>
        /// When overridden in a derived class, gets a value indicating whether the current stream supports seeking.
        /// </summary>
        public override bool CanSeek { get { return false; } }

        /// <summary>
        /// When overridden in a derived class, gets a value indicating whether the current stream supports writing.
        /// </summary>
        public override bool CanWrite { get { return false; } }

        /// <summary>
        /// When overridden in a derived class, gets the length in bytes of the stream.
        /// </summary>
        public override long Length { get { throw new NotSupportedException(); } }

        /// <summary>
        /// When overridden in a derived class, gets or sets the position within the current stream.
        /// </summary>
        public override long Position
        {
            get { throw new NotSupportedException(); }
            set { throw new NotSupportedException(); }
        }

        /// <summary>
        /// When overridden in a derived class, clears all buffers for this stream and causes any buffered data to be written to the underlying device.
        /// </summary>
        public override void Flush()
        { }

        /// <summary>
        /// When overridden in a derived class, sets the position within the current stream.
        /// </summary>
        public override long Seek(long offset, SeekOrigin origin)
        {
            throw new NotSupportedException();
        }

        /// <summary>
        /// When overridden in a derived class, sets the length of the current stream.
        /// </summary>
        public override void SetLength(long value)
        {
            throw new NotSupportedException();
        }

        /// <summary>
        /// When overridden in a derived class, reads a sequence of bytes from the current stream and advances the position within the stream by the number of bytes read.
        /// </summary>
        public override int Read(byte[] buffer, int offset, int count)
        {
            throw new NotSupportedException();
        }

        /// <summary>
        /// When overridden in a derived class, writes a sequence of bytes to the current stream and advances the current position within this stream by the number of bytes written.
        /// </summary>
        public override void Write(byte[] buffer, int offset, int count)
        {
            throw new NotSupportedException();
        }

        /// <summary>
        /// Reads a byte from the stream and advances the position within the stream by one byte, or returns -1 if at the end of the stream.
        /// </summary>
        public override int ReadByte()
        {
            byte[] bytes = new byte[1];
            return Read(bytes, 0, 1) == 1 ? bytes[0] : -1;
        }

        /// <summary>
        /// Writes a byte to the current position in the stream and advances the position within the stream by one byte.
        /// </summary>
        public override void WriteByte(byte value)
        {
            Write(new byte[] { value }, 0, 1);
        }
    }
}