﻿// Copyright (c) 2012 James Treworgy
// 
// LICENSE (MIT License)
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy 
// of this software and associated documentation files (the "Software"), to deal 
// in the Software without restriction, including without limitation the rights to 
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
// of the Software, and to permit persons to whom the Software is furnished to do 
// so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all 
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CsQuery.Promises;

namespace CsQuery
{
    /// <summary>
    /// A static provider for methods that produce Promise-related objects
    /// </summary>

    internal static class When
    {
        /// <summary>
        /// Gets or sets a value indicating whether objects in the Promises library should be created in
        /// debug mode. This affects Deferred.FailOnResolutionExceptions.
        /// </summary>

        public static bool Debug
        {
            get;
            set;
        }

        /// <summary>
        /// Returns a new Deferred object, an object containing a promise and resolver methods.
        /// </summary>
        ///
        /// <returns>
        /// A new Deferred object.
        /// </returns>

        public static Deferred Deferred()
        {
            return new Deferred();
        }

        /// <summary>
        /// Returns a new Deferred object, an object containing a promise and resolver methods.
        /// </summary>
        ///
        /// <typeparam name="T">
        /// Generic type parameter determining the type of parameter that will be passed to the resolvers
        /// </typeparam>
        ///
        /// <returns>
        /// A new Deferred object
        /// </returns>

        public static Deferred<T> Deferred<T>()
        {
            return new Deferred<T>();
        }

        /// <summary>
        /// Returns a new promise that resolves when all of the promises passed by parameter have resolved
        /// </summary>
        ///
        /// <param name="promises">
        /// One or more IPromise objects
        /// </param>
        ///
        /// <returns>
        /// A new IPromise object
        /// </returns>

        public static IPromise All(params IPromise[] promises)
        {
            return new WhenAll(promises);

        }

        /// <summary>
        /// Returns a new promise that resolves when all of the promises passed by parameter have
        /// resolved, or when the time has elapsed
        /// </summary>
        ///
        /// <param name="timeoutMilliseconds">
        /// The timeout in milliseconds.
        /// </param>
        /// <param name="promises">
        /// One or more IPromise objects.
        /// </param>
        ///
        /// <returns>
        /// A new IPromise object.
        /// </returns>

        public static IPromise All(int timeoutMilliseconds, params IPromise[] promises)
        {
            return new WhenAll(timeoutMilliseconds,promises);
            
        }

        /// <summary>
        /// Return a promise that fails after the specified time. This is like Timer, but fails rather
        /// than succeeds after the time has elapsed.
        /// </summary>
        ///
        /// <param name="timeoutMilliseconds">
        /// The timeout in milliseconds.
        /// </param>
        ///
        /// <returns>
        /// A promise.
        /// </returns>

        public static IPromise Timeout(int timeoutMilliseconds)
        {
            return new Timeout(timeoutMilliseconds);
        }

        /// <summary>
        /// Return a promise that resolves successfully after the specified time. 
        /// </summary>
        ///
        /// <param name="timerMilliseconds">
        /// The timeout in milliseconds.
        /// </param>
        ///
        /// <returns>
        /// A promise.
        /// </returns>

        public static IPromise Timer(int timerMilliseconds)
        {
            return new Timeout(timerMilliseconds, true);
        }

        
    }
}
