﻿// Copyright (c) 2012 James Treworgy
// 
// LICENSE (MIT License)
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy 
// of this software and associated documentation files (the "Software"), to deal 
// in the Software without restriction, including without limitation the rights to 
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
// of the Software, and to permit persons to whom the Software is furnished to do 
// so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all 
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CsQuery.StringScanner
{
    /// <summary>
    /// An interface for pattern matching.
    /// 
    /// Something implementing this interface will be used as follows:
    /// 
    /// First, Initialize is called, passing in the source and the starting index where scanning
    /// should begin.
    /// 
    /// The Validate function then scans the string, and returns true if a valid match is found, and
    /// false if not.
    /// 
    /// The Result property should be populated by the function with the matching string, and the
    /// EndIndex property should be populated with the last position scanned (one after the last
    /// valid character that was returned). If no valid string was matched, EndIndex should equal the
    /// original StartIndex.
    /// </summary>

    internal interface IExpectPattern
    {
        /// <summary>
        /// Initializes the pattern
        /// </summary>
        ///
        /// <param name="startIndex">
        /// The start index.
        /// </param>
        /// <param name="source">
        /// Source for the.
        /// </param>

        void Initialize(int startIndex, char[] source);

        /// <summary>
        /// Validate the string and try to match something.
        /// </summary>
        ///
        /// <returns>
        /// true if a matching string was found, false if not.
        /// </returns>

        bool Validate();

        /// <summary>
        /// When a valid string was found, the string.
        /// </summary>
        ///
        /// <value>
        /// A string.
        /// </value>

        string Result { get; }

        /// <summary>
        /// Gets zero-based index of the ending postion. This is one position after the last matching
        /// character.
        /// </summary>
        ///
        /// <value>
        /// The end index.
        /// </value>

        int EndIndex { get; }
    }
}
