﻿// Copyright (c) 2012 James Treworgy
// 
// LICENSE (MIT License)
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy 
// of this software and associated documentation files (the "Software"), to deal 
// in the Software without restriction, including without limitation the rights to 
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
// of the Software, and to permit persons to whom the Software is furnished to do 
// so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all 
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Linq;
using System.Text;
using System.Reflection;
using System.Reflection.Emit;

namespace CsQuery.Utility
{
    /// <summary>
    /// Replacement for Activator.CreateInstance that caches the constructor fucnction, providing a
    /// significant performance improvement over Activator.CreateInstance. Calling with a value type
    /// will be deferred to Activator.CreateInstance.
    /// </summary>
    /// 
    ///
    /// <url>
    /// http://mironabramson.com/blog/post/2008/08/Fast-version-of-the-ActivatorCreateInstance-method-using-IL.aspx
    /// </url>

    internal static class FastActivator
    {
      
        private static ConcurrentDictionary<Type,CreateObject> creatorCache = new ConcurrentDictionary<Type,CreateObject>();

        private readonly static Type coType = typeof(CreateObject);
        private delegate object CreateObject();

        /// <summary>
        /// Create a new instance of type T
        /// </summary>
        ///
        /// <typeparam name="T">
        /// The type of object to create
        /// </typeparam>
        ///
        /// <returns>
        /// A new instance of type T
        /// </returns>

        public static T CreateInstance<T>() where T : class
        {
            return (T)CreateInstance(typeof(T));
        }

        /// <summary>
        /// Creates an instance.
        /// </summary>
        ///
        /// <param name="type">
        /// The Type to process.
        /// </param>
        ///
        /// <returns>
        /// The new instance.
        /// </returns>

        public static object CreateInstance(Type type)
        {
            CreateObject c;
            if (!type.IsClass)
            {
                return Activator.CreateInstance(type);
                
            }

            if (!creatorCache.TryGetValue(type, out c))
            {
                DynamicMethod dynMethod = new DynamicMethod("DM$OBJ_FACTORY_" + type.Name, type, null, type);
                ILGenerator ilGen = dynMethod.GetILGenerator();

                ilGen.Emit(OpCodes.Newobj, type.GetConstructor(Type.EmptyTypes));
                ilGen.Emit(OpCodes.Ret);
                c = (CreateObject)dynMethod.CreateDelegate(coType);
                creatorCache[type] = c;

            }
            return c();

        }

    }

}
