﻿// Copyright (c) 2013 Milkshake Software (http://milkshakesoftware.com)
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using CsQuery;
using PreMailer.Net.Sources;
using System;
using System.Collections.Generic;
using System.Linq;

namespace PreMailer.Net
{
    internal class PreMailer
	{
		private readonly CQ document;
        private readonly SortedList<string, StyleClass> styles;
        private readonly bool useCqSelection;

        private readonly CssParser cssParser;
        private readonly CssSelectorParser cssSelectorParser;
		private readonly List<string> warnings;

        private PreMailer(CQ cq, SortedList<string, StyleClass> styles, bool useCqSelection)
		{
			this.document = cq;
            this.styles = styles;
            this.useCqSelection = useCqSelection;
            this.warnings = new List<string>();

			this.cssParser = new CssParser();
			this.cssSelectorParser = new CssSelectorParser();
		}

        public static InlineResult MoveCssInline(CQ cq, SortedList<string, StyleClass> styles, bool useCqSelection)
        {
            PreMailer preMailer = new PreMailer(cq, styles, useCqSelection);

            return preMailer.Process();
        }

		private InlineResult Process()
		{
			// Gather all of the CSS that we can work with.
            CQ cssSourceNodes = GetStyleElements();
            RemoveStyleElements(cssSourceNodes);

            SortedList<string, StyleClass> validSelectors = CleanUnsupportedSelectors(this.styles);
            Dictionary<IDomObject, List<StyleClass>> elementsWithStyles = FindElementsWithStyles(validSelectors);
            Dictionary<IDomObject, StyleClass> mergedStyles = MergeStyleClasses(elementsWithStyles);
            this.ApplyStyles(mergedStyles);

			return new InlineResult(warnings);
		}

		/// <summary>
		/// Returns a collection of CQ nodes that can be used to source CSS content.<para/>
		/// Currently, only 'style' tags are supported.
		/// </summary>
		private CQ GetStyleElements()
		{
			// TODO: Add Source to Read CSS from LINK tags etc.
			// All we need to do here is update the selector in 'document.Find(...)' and then add
			// something that implements ICssSource to handle that type of link..
			// e.g. new LinkTagCssSource(node, baseUrl: "...");
            var elements = document.Find("style");
			return elements;
		}


		private void RemoveStyleElements(CQ cssSourceNodes)
		{
            foreach (IDomObject node in cssSourceNodes)
            {
				node.Remove();
			}
		}

		private SortedList<string, StyleClass> CleanUnsupportedSelectors(SortedList<string, StyleClass> selectors)
		{
			var result = new SortedList<string, StyleClass>();
			var failedSelectors = new List<StyleClass>();

			foreach (var selector in selectors)
			{
				if (!selector.Key.Contains(":"))
					result.Add(selector.Key, selector.Value);
				else
					failedSelectors.Add(selector.Value);
			}

			if (!failedSelectors.Any())
				return selectors;

			foreach (var failedSelector in failedSelectors)
			{
				warnings.Add(String.Format(
						"PreMailer.Net is unable to process the pseudo class/element '{0}' due to a limitation in CsQuery.",
						failedSelector.Name));
			}

			return result;
		}

		private Dictionary<IDomObject, List<StyleClass>> FindElementsWithStyles(
				SortedList<string, StyleClass> stylesToApply)
		{
			var result = new Dictionary<IDomObject, List<StyleClass>>();

			foreach (var style in stylesToApply)
			{
				var elementsForSelector = document[style.Value.Name];

				foreach (var el in elementsForSelector)
				{
                    if (this.useCqSelection && !document.Is(el))
                    {
                        continue;
                    }

					var existing = result.ContainsKey(el) ? result[el] : new List<StyleClass>();
					existing.Add(style.Value);
					result[el] = existing;
				}
			}

			return result;
		}

		private Dictionary<IDomObject, List<StyleClass>> SortBySpecificity(
				Dictionary<IDomObject, List<StyleClass>> styles)
		{
			var result = new Dictionary<IDomObject, List<StyleClass>>();

			foreach (var style in styles)
			{
				if (style.Key.Attributes != null)
				{
					var sortedStyles = style.Value.OrderBy(x => cssSelectorParser.GetSelectorSpecificity(x.Name)).ToList();

					if (String.IsNullOrWhiteSpace(style.Key.Attributes["style"]))
					{
						style.Key.SetAttribute("style", String.Empty);
					}
					else // Ensure that existing inline styles always win.
					{
						sortedStyles.Add(cssParser.ParseStyleClass("inline", style.Key.Attributes["style"]));
					}

					result[style.Key] = sortedStyles;
				}
			}

			return result;
		}

		private Dictionary<IDomObject, StyleClass> MergeStyleClasses(
				Dictionary<IDomObject, List<StyleClass>> styles)
		{
			var result = new Dictionary<IDomObject, StyleClass>();
			var stylesBySpecificity = SortBySpecificity(styles);

			foreach (var elemStyle in stylesBySpecificity)
			{
				// CSS Classes are assumed to be sorted by specifity now, so we can just merge these up.
				var merged = new StyleClass();
				foreach (var style in elemStyle.Value)
				{
					merged.Merge(style, true);
				}

				result[elemStyle.Key] = merged;
			}

			return result;
		}

		private void ApplyStyles(Dictionary<IDomObject, StyleClass> elementStyles)
		{
			foreach (var elemStyle in elementStyles)
			{
				var el = elemStyle.Key;
				el.SetAttribute("style", elemStyle.Value.ToString());
			}
		}
	}
}