﻿using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html.Import
{
    internal class RadFlowDocumentImportEditor
    {
        private readonly RadFlowDocument document;
        private readonly HtmlImportSettings importSettings;
        private readonly Stack<BlockContainerBase> blockContainersStack;
        private readonly Stack<Table> tablesStack;
        private readonly Stack<Dictionary<RowType, LinkedList<TableRow>>> tableRowTypesStack;
        private readonly Stack<TableRow> tableRowStack;

        private Paragraph currentParagraph;
        private bool hasCurrentHyperlink;

        public RadFlowDocumentImportEditor(RadFlowDocument document, HtmlImportSettings htmlImportSettings)
        {
            Guard.ThrowExceptionIfNull(document, "document");

            this.document = document;
            this.importSettings = htmlImportSettings;

            this.blockContainersStack = new Stack<BlockContainerBase>();
            this.tablesStack = new Stack<Table>();
            this.tableRowStack = new Stack<TableRow>();
            this.tableRowTypesStack = new Stack<Dictionary<RowType, LinkedList<TableRow>>>();

            this.PushBlockContainer(this.document.Sections.AddSection());
        }

        public RadFlowDocument Document
        {
            get
            {
                return this.document;
            }
        }

        public Paragraph CurrentParagraph
        {
            get
            {
                return this.currentParagraph;
            }
        }

        private BlockContainerBase CurrentBlockContainer
        {
            get
            {
                if (this.blockContainersStack.Count > 0)
                {
                    return this.blockContainersStack.Peek();
                }

                return null;
            }
        }

        private Table CurrentTable
        {
            get
            {
                if (this.tablesStack.Count > 0)
                {
                    return this.tablesStack.Peek();
                }

                return null;
            }
        }

        private TableRow CurrentTableRow
        {
            get
            {
                if (this.tableRowStack.Count > 0)
                {
                    return this.tableRowStack.Peek();
                }

                return null;
            }
        }

        private Dictionary<RowType, LinkedList<TableRow>> CurrentTableRows
        {
            get
            {
                if (this.tableRowTypesStack.Count > 0)
                {
                    return this.tableRowTypesStack.Peek();
                }

                return null;
            }
        }

        public void BeginHyperlink()
        {
            this.hasCurrentHyperlink = true;
        }

        public void EndHyperlink()
        {
            this.hasCurrentHyperlink = false;
        }

        public void BeginParagraph(Paragraph paragraph)
        {
            Guard.ThrowExceptionIfNull(paragraph, "paragraph");

            this.EndParagraph();
            this.currentParagraph = paragraph;
            this.CurrentBlockContainer.Blocks.Add(paragraph);
        }

        public void EndParagraph()
        {
            if (this.currentParagraph != null)
            {
                Run last = this.currentParagraph.Inlines.LastOrDefault() as Run;
                if (last != null)
                {
                    this.FinishRun(last);
                }

                if (this.currentParagraph.Inlines.Count == 0 &&
                    this.currentParagraph.ListId == DocumentDefaultStyleSettings.ListId)
                {
                    ((BlockContainerBase)this.currentParagraph.Parent).Blocks.Remove(this.currentParagraph);
                }

                ClearNormalStyleId(this.currentParagraph);

                this.currentParagraph = null;
            }
        }

        public void PushTable(Table table)
        {
            this.tablesStack.Push(table);
            this.tableRowTypesStack.Push(CreateTableRowsDictionary());

            this.CurrentBlockContainer.Blocks.Add(table);
        }

        public void PopTable()
        {
            if (this.CurrentTable != null)
            {
                Table currentTable = this.tablesStack.Pop();
                Dictionary<RowType, LinkedList<TableRow>> tableRows = this.tableRowTypesStack.Pop();

                InsertTableRows(currentTable, tableRows[RowType.Header]);
                InsertTableRows(currentTable, tableRows[RowType.Body]);
                InsertTableRows(currentTable, tableRows[RowType.Footer]);

                ClearTableNormalStyleId(currentTable);
            }
        }

        public void PushTableRow(TableRow row, RowType rowType = RowType.Body)
        {
            Guard.ThrowExceptionIfNull(row, "row");

            if (this.CurrentTable != null)
            {
                this.CurrentTableRows[rowType].AddLast(row);
                this.tableRowStack.Push(row);
            }
        }

        public void PopTableRow()
        {
            if (this.CurrentTableRow != null)
            {
                this.tableRowStack.Pop();
            }
        }

        public void InsertTableCell(TableCell cell, CellType cellType)
        {
            Guard.ThrowExceptionIfNull(cell, "cell");
            Guard.ThrowExceptionIfNull(cellType, "cellType");

            if (this.CurrentTableRow != null)
            {
                this.CurrentTableRow.Cells.Add(cell);
                this.PushBlockContainer(cell);
            }
        }

        public void InsertRun(Run run)
        {
            Guard.ThrowExceptionIfNull(run, "run");

            this.AssureParagraphExists();
            run.Text = HtmlTextProcessor.RemoveMultipleWhiteSpaces(run.Text);
            if (!string.IsNullOrEmpty(run.Text))
            {
                this.PrepareRunToInsert(run);

                if (!string.IsNullOrEmpty(run.Text))
                {
                    this.currentParagraph.Inlines.Add(run);
                }
            }
        }

        public void InsertInline(InlineBase inline)
        {
            Guard.ThrowExceptionIfNull(inline, "inline");

            this.AssureParagraphExists();
            Run lastRun = this.currentParagraph.Inlines.LastOrDefault() as Run;
            if (lastRun != null && inline.Type != DocumentElementType.Run && inline.Type != DocumentElementType.FieldCharacter)
            {
                this.FinishRun(lastRun);
            }

            this.currentParagraph.Inlines.Add(inline);
        }

        public void PushBlockContainer(BlockContainerBase blockContainer)
        {
            this.blockContainersStack.Push(blockContainer);

            this.EndParagraph();
        }

        public void PopBlockContainer()
        {
            this.EndParagraph();
            this.blockContainersStack.Pop();
        }

        public bool HasCurrentParagraph()
        {
            return this.currentParagraph != null;
        }

        public void FinishDocument()
        {
            this.EndParagraph();
            this.PopTable();
        }

        private static void ClearNormalStyleId(Paragraph paragraph)
        {
            Guard.ThrowExceptionIfNull(paragraph, "paragraph");

            if (paragraph.StyleId == StyleNamesConverter.PrefixedNormalStyleId)
            {
                paragraph.StyleId = null;
            }
        }

        private static void ClearTableNormalStyleId(Table table)
        {
            Guard.ThrowExceptionIfNull(table, "table");

            if (table.StyleId == StyleNamesConverter.PrefixedNormalTableStyleId)
            {
                table.StyleId = null;
            }
        }

        private static Dictionary<RowType, LinkedList<TableRow>> CreateTableRowsDictionary()
        {
            Dictionary<RowType, LinkedList<TableRow>> result = new Dictionary<RowType, LinkedList<TableRow>>();
            result[RowType.Header] = new LinkedList<TableRow>();
            result[RowType.Body] = new LinkedList<TableRow>();
            result[RowType.Footer] = new LinkedList<TableRow>();

            return result;
        }

        private static void InsertTableRows(Table table, LinkedList<TableRow> tableRows)
        {
            foreach (TableRow row in tableRows)
            {
                if (row.Cells.Count > 0)
                {
                    table.Rows.Add(row);
                }
            }
        }

        private void PrepareRunToInsert(Run run)
        {
            Run last = this.currentParagraph.Inlines.LastOrDefault() as Run;
            string lastRunText = last != null ? last.Text : string.Empty;

            run.Text = HtmlTextProcessor.StripText(run.Text, lastRunText);

            if (this.hasCurrentHyperlink && string.IsNullOrEmpty(run.StyleId))
            {
                run.StyleId = BuiltInStyleNames.HyperlinkStyleId;
            }
        }

        private void AssureParagraphExists()
        {
            if (!this.HasCurrentParagraph())
            {
                this.BeginParagraph(new Paragraph(this.Document));
            }
        }

        private void FinishRun(Run run)
        {
            Guard.ThrowExceptionIfNull(run, "run");

            run.Text = HtmlTextProcessor.TrimEnd(run.Text);

            if (this.importSettings.ReplaceNonBreakingSpaces)
            {
                run.Text = HtmlTextProcessor.Normalize(run.Text);
            }

            if (string.IsNullOrEmpty(run.Text))
            {
                this.currentParagraph.Inlines.Remove(run);
            }
        }
    }
}
