﻿using CsQuery;
using PreMailer.Net;
using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Styles;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html.Import
{
    internal class StyleEvaluator
    {
        private readonly CQ cq;
        private readonly SortedList<string, StyleClass> styles;
        private readonly StyleNamesInfo stylesNamesInfo;

        private readonly CssSelectorParser cssSelectorParser;

        private StyleClass defaultSpanStyle;

        public StyleEvaluator(CQ cq, SortedList<string, StyleClass> styles, StyleNamesInfo stylesNamesInfo)
        {
            Guard.ThrowExceptionIfNull(cq, "cq");
            Guard.ThrowExceptionIfNull(stylesNamesInfo, "stylesInfo");

            this.cq = cq;
            this.styles = styles;
            this.stylesNamesInfo = stylesNamesInfo;

            this.cssSelectorParser = new CssSelectorParser();
        }

        public HtmlStyleRepository Evaluate()
        {
            HtmlStyleRepository htmlStyleRepository = new HtmlStyleRepository();

            this.CreateDefaultSelectors(htmlStyleRepository);
            this.AddSelectorsForStyles(htmlStyleRepository);
            this.AddSelectorsForElementSelectors(htmlStyleRepository);
            this.AddEmptySelectorsForAllNonDefinedStyles(htmlStyleRepository);
            this.InlineStyles();

            return htmlStyleRepository;
        }

        private static void CopyAttributes(Selector selector, StyleClass style)
        {
            foreach (KeyValuePair<string, string> attribute in style.Attributes)
            {
                selector.RegisterProperty(attribute.Key, attribute.Value);
            }
        }

        private static string GetSelectorName(string styleName)
        {
            if (styleName.StartsWith("."))
            {
                return styleName.Substring(1);
            }

            return styleName;
        }

        private void AddSelectorsForElementSelectors(HtmlStyleRepository htmlStyleRepository)
        {
            foreach (string selectorString in StyleToElementSelectorMappings.GetSelectors())
            {
                bool any = false;
                string styleId = null;
                string prefixedStyleId = null;

                if (StyleToElementSelectorMappings.TryGetStyleId(selectorString, out styleId))
                {
                    prefixedStyleId = StyleNamesConverter.ConvertStyleNameOnExport(styleId);
                    any = this.ExplicitlySetClassName(selectorString, prefixedStyleId);
                }

                if (any && styleId != null)
                {
                    StyleClass style = this.styles[selectorString];
                    this.styles.Remove(selectorString);

                    if (!htmlStyleRepository.ContainsStyle(prefixedStyleId))
                    {
                        Selector selector = new Selector(prefixedStyleId, SelectorType.Style);
                        CopyAttributes(selector, style);
                        htmlStyleRepository.RegisterStyle(prefixedStyleId, selector);
                    }
                }
            }
        }

        private bool ExplicitlySetClassName(string selector, string className)
        {
            bool any = false;

            foreach (IDomObject element in this.cq[selector])
            {
                any = true;

                element.AddClass(className);
            }

            return any;
        }

        private void CreateDefaultSelectors(HtmlStyleRepository htmlStyleRepository)
        {
            htmlStyleRepository.DefaultParagraphStyle = this.CreateDefaultElementSelector(HtmlElementNames.Paragraph);
            htmlStyleRepository.DefaultCharacterStyle = this.CreateDefaultElementSelector(HtmlElementNames.Span, this.SaveDefaultSpanStyleClass);
        }

        private void AddSelectorsForStyles(HtmlStyleRepository htmlStyleRepository)
        {
            foreach (KeyValuePair<string, StyleClass> style in this.styles.ToList())
            {
                if (this.ShouldCreateStyle(style.Key))
                {
                    string selectorName = GetSelectorName(style.Key);

                    // If a <table> element has an associated style, we should inline that style later.
                    // Note: If a paragrah and table elements have the same style associated, we will inline the style properties for the paragrah as well.
                    if (!this.stylesNamesInfo.TableStyles.Contains(selectorName))
                    {
                        this.styles.Remove(style.Key);
                    }

                    Selector selector = new Selector(GetSelectorName(selectorName), SelectorType.Style);
                    CopyAttributes(selector, style.Value);
                    htmlStyleRepository.RegisterStyle(selector.Name, selector);
                }
            }
        }

        private void AddEmptySelectorsForAllNonDefinedStyles(HtmlStyleRepository htmlStyleRepository)
        {
            // Get all paragraphs and spans with 'class' attribute
            CQ paragraphsWithClassAttribute = this.cq.Select("p[class]");
            CQ spansWithClassAttribute = this.cq.Select("span[class]");

            IEnumerable<string> allUsedClassNames = paragraphsWithClassAttribute.Concat(spansWithClassAttribute)
                .Where(element => element.HasClasses)
                .Select(element => element.Classes.First())
                .Distinct();

            foreach (string className in allUsedClassNames)
            {
                if (!htmlStyleRepository.ContainsStyle(className))
                {
                    // This is class name which is not defined in the CSS.
                    Selector emptySelector = new Selector(className, SelectorType.Style);
                    htmlStyleRepository.RegisterStyle(className, emptySelector);
                }
            }
        }

        private void InlineStyles()
        {
            this.InlineDefaultSpanStyle();

            PreMailer.Net.PreMailer.MoveCssInline(this.cq, this.styles, false);
        }

        private void InlineDefaultSpanStyle()
        {
            if (this.defaultSpanStyle != null)
            {
                SortedList<string, StyleClass> defaultCharacterStyleList = new SortedList<string, StyleClass>();
                defaultCharacterStyleList.Add(HtmlElementNames.Span, this.defaultSpanStyle);

                // Select all spans without class or with empty class attribute.
                CQ spansWithoutClasses = this.cq.Select("span").Not(@"[class!=""""]");

                PreMailer.Net.PreMailer.MoveCssInline(spansWithoutClasses, defaultCharacterStyleList, true);
            }
        }

        private void SaveDefaultSpanStyleClass(StyleClass styleClass)
        {
            this.defaultSpanStyle = styleClass;
        }

        private Selector CreateDefaultElementSelector(string selectorName, Action<StyleClass> saveStyleCallback = null)
        {
            Selector selector = new Selector(selectorName, SelectorType.Elements);

            StyleClass style;
            if (this.styles.TryGetValue(selectorName, out style))
            {
                this.styles.Remove(selectorName);

                CopyAttributes(selector, style);

                if (saveStyleCallback != null)
                {
                    saveStyleCallback(style);
                }
            }

            return selector;
        }

        private bool ShouldCreateStyle(string selector)
        {
            CssSpecificity specificity = this.cssSelectorParser.CalculateSpecificity(selector);

            // All styles with selector containing only one class.
            bool isClassSelector =
                specificity.Classes == 1 &&
                specificity.ClassesAttributesPseudoElements == 1 &&
                specificity.Elements == 0 &&
                specificity.Ids == 0;

            return isClassSelector;
        }
    }
}
