﻿using System.Linq;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Styles;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html.Import
{
    internal class StylesImporter
    {
        private const string ConflictingStyleIdSuffix = "1";

        private readonly RadFlowDocument document;
        private readonly IHtmlImportContext context;

        private Style normalWebStyle;
        private Style defaultCharacterStyle;
        private StylesImporter(IHtmlImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            this.context = context;
            this.document = context.Document;
        }

        public static void Import(IHtmlImportContext context)
        {
            StylesImporter stylesImporter = new StylesImporter(context);

            stylesImporter.CreateTemplateStyles();
            stylesImporter.Import();
        }

        private void Import()
        {
            foreach (Selector selector in this.context.HtmlStyleRepository.GetStyles())
            {
                bool isStyleImported = false;

                if (this.ShouldImportParagraphStyle(selector))
                {
                    this.ImportParagraphStyle(selector);
                    isStyleImported = true;
                }

                if (this.ShouldImportCharacterStyle(selector))
                {
                    this.ImportCharacterStyle(selector);
                    isStyleImported = true;
                }

                System.Diagnostics.Debug.Assert(isStyleImported || !selector.HasProperties, "The non-empty selector named'" + selector.Name + "' was not imported.");
            }

            if (this.context.ShouldImportNormalWeb)
            {
                this.document.StyleRepository.Add(this.normalWebStyle);
            }
        }

        private void CreateTemplateStyles()
        {
            this.normalWebStyle = this.CreateNormalWebStyle();
            this.defaultCharacterStyle = this.CreateDefaultCharacterStyle();
        }

        private void ImportCharacterStyle(Selector selector)
        {
            Style characterStyle = this.CreateCharacterStyle(selector, this.defaultCharacterStyle);
            this.document.StyleRepository.Add(characterStyle);

            if (selector.Type == SelectorType.Style && characterStyle.Id != selector.Name)
            {
                // Style is renamed, change StyleId of all elements.
                this.document.EnumerateChildrenOfType<Run>()
                    .Where(run => run.StyleId == selector.Name)
                    .ForEach(run => run.StyleId = characterStyle.Id);
            }
        }

        private void ImportParagraphStyle(Selector selector)
        {
            Style paragraphStyle = this.CreateParagraphStyle(selector, this.normalWebStyle);
            this.document.StyleRepository.Add(paragraphStyle);

            if (selector.Type == SelectorType.Style && paragraphStyle.Id != selector.Name)
            {
                // Style is renamed, change StyleId of all elements.
                this.document.EnumerateChildrenOfType<Paragraph>()
                    .Where(p => p.StyleId == selector.Name)
                    .ForEach(p => p.StyleId = paragraphStyle.Id);
            }
        }

        private bool ShouldImportCharacterStyle(Selector selector)
        {
            return this.context.StyleNamesInfo.CharacterStyles.Contains(selector.Name);
        }

        private bool ShouldImportParagraphStyle(Selector selector)
        {
            if (StyleNamesConverter.IsTelerikExportedDefaultBuiltInStyle(selector.Name))
            {
                return selector.HasProperties;
            }
            else
            {
                return this.context.StyleNamesInfo.ParagraphStyles.Contains(selector.Name) || selector.HasProperties;
            }
        }

        private Style CreateDefaultCharacterStyle()
        {
            Style defaultCharacterStyle = new Style("temporaryStyle", StyleType.Character);
            this.context.HtmlStyleRepository.DefaultCharacterStyle.CopyTo(this.context, defaultCharacterStyle.CharacterProperties);

            return defaultCharacterStyle;
        }

        private Style CreateNormalWebStyle()
        {
            Style normalWebStyle = BuiltInStyles.GetStyle(BuiltInStyleNames.NormalWebStyleId);

            normalWebStyle.CharacterProperties.FontFamily.ClearValue();
            normalWebStyle.CharacterProperties.FontSize.ClearValue();

            normalWebStyle.ParagraphProperties.SpacingBefore.LocalValue = 6.66;
            normalWebStyle.ParagraphProperties.AutomaticSpacingBefore.LocalValue = true;
            normalWebStyle.ParagraphProperties.SpacingAfter.LocalValue = 6.66;
            normalWebStyle.ParagraphProperties.AutomaticSpacingAfter.LocalValue = true;

            double? spacingAfterDefaultValue = normalWebStyle.ParagraphProperties.SpacingAfter.LocalValue;
            double? spacingBeforeDefaultValue = normalWebStyle.ParagraphProperties.SpacingBefore.LocalValue;

            normalWebStyle.ParagraphProperties.SpacingAfter.ClearValue();
            normalWebStyle.ParagraphProperties.SpacingBefore.ClearValue();

            this.context.HtmlStyleRepository.DefaultParagraphStyle.CopyTo(this.context, normalWebStyle.ParagraphProperties);
            this.context.HtmlStyleRepository.DefaultParagraphStyle.CopyTo(this.context, normalWebStyle.CharacterProperties);

            if (normalWebStyle.ParagraphProperties.SpacingAfter.HasLocalValue)
            {
                normalWebStyle.ParagraphProperties.AutomaticSpacingAfter.LocalValue = false;
            }
            else
            {
                normalWebStyle.ParagraphProperties.SpacingAfter.LocalValue = spacingAfterDefaultValue;
            }

            if (normalWebStyle.ParagraphProperties.SpacingBefore.HasLocalValue)
            {
                normalWebStyle.ParagraphProperties.AutomaticSpacingBefore.LocalValue = false;
            }
            else
            {
                normalWebStyle.ParagraphProperties.SpacingBefore.LocalValue = spacingBeforeDefaultValue;
            }

            return normalWebStyle;
        }

        private Style CreateParagraphStyle(Selector selector, Style templateStyle)
        {
            Style paragraphStyle = this.CreateStyle(selector.Name, StyleType.Paragraph, templateStyle);

            double? spacingAfterDefaultValue = paragraphStyle.ParagraphProperties.SpacingAfter.LocalValue;
            double? spacingBeforeDefaultValue = paragraphStyle.ParagraphProperties.SpacingAfter.LocalValue;

            paragraphStyle.ParagraphProperties.SpacingAfter.ClearValue();
            paragraphStyle.ParagraphProperties.SpacingBefore.ClearValue();

            selector.CopyTo(this.context, paragraphStyle);

            if (paragraphStyle.ParagraphProperties.SpacingAfter.HasLocalValue)
            {
                paragraphStyle.ParagraphProperties.AutomaticSpacingAfter.LocalValue = false;
            }
            else
            {
                paragraphStyle.ParagraphProperties.SpacingAfter.LocalValue = spacingAfterDefaultValue;
            }

            if (paragraphStyle.ParagraphProperties.SpacingBefore.HasLocalValue)
            {
                paragraphStyle.ParagraphProperties.AutomaticSpacingBefore.LocalValue = false;
            }
            else
            {
                paragraphStyle.ParagraphProperties.SpacingBefore.LocalValue = spacingBeforeDefaultValue;
            }

            return paragraphStyle;
        }

        private Style CreateCharacterStyle(Selector selector, Style templateStyle)
        {
            Style characterStyle = this.CreateStyle(selector.Name, StyleType.Character, templateStyle);
            selector.CopyTo(this.context, characterStyle);

            return characterStyle;
        }

        private Style CreateStyle(string suggestedStyleId, StyleType styleType, Style templateStyle)
        {
            string styleId = this.ComputeStyleId(suggestedStyleId);

            Style style;
            if (BuiltInStyles.IsBuiltInStyle(styleId))
            {
                style = BuiltInStyles.GetStyle(styleId);
            }
            else
            {
                style = new Style(styleId, styleType);
                style.CopyStylePropertiesFrom(templateStyle);
                style.BasedOnStyleId = templateStyle.BasedOnStyleId;
            }

            return style;
        }

        private string ComputeStyleId(string styleId)
        {
            string uniqueStyleId;
            if (StyleNamesConverter.TryGetTelerikExportedBuiltInStyle(styleId, out uniqueStyleId))
            {
                return uniqueStyleId;
            }

            uniqueStyleId = styleId;
            while (this.document.StyleRepository.Contains(uniqueStyleId))
            {
                uniqueStyleId += StylesImporter.ConflictingStyleIdSuffix;
            }

            return uniqueStyleId;
        }
    }
}
