﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Common.FormatProviders.XmlBased.Model.Elements;
using Telerik.Windows.Documents.Common.FormatProviders.XmlBased.Model.Namespaces;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Import;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Attributes;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements.Text;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Styles;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements
{
    internal abstract class HtmlElementBase : XmlElementBase<HtmlAttribute>
    {
        private readonly HtmlContentManager contentManager;

        private readonly StyleAttribute styleAttribute;
        private readonly ClassAttribute classAttribute;
        private readonly HtmlStyleProperties inheritedProperties;
        private readonly HtmlBubblingProperties bubblingProperties;

        public HtmlElementBase(HtmlContentManager contentManager)
        {
            Guard.ThrowExceptionIfNull(contentManager, "contentManager");

            this.contentManager = contentManager;

            this.styleAttribute = new StyleAttribute();
            this.RegisterAttribute(this.styleAttribute);

            this.classAttribute = new ClassAttribute();
            this.RegisterAttribute(this.classAttribute);

            this.inheritedProperties = new HtmlStyleProperties();
            this.bubblingProperties = new HtmlBubblingProperties();
        }

        public ClassNamesCollection Classes
        {
            get
            {
                return this.classAttribute.Value;
            }
        }

        public StyleAttribute Style
        {
            get
            {
                return this.styleAttribute;
            }
        }

        public HtmlStyleProperties InheritedProperties
        {
            get
            {
                return this.inheritedProperties;
            }
        }

        public HtmlStyleProperties LocalProperties
        {
            get
            {
                return this.Style.Value;
            }
        }

        public HtmlBubblingProperties BubblingProperties
        {
            get
            {
                return this.bubblingProperties;
            }
        }

        protected HtmlContentManager ContentManager
        {
            get
            {
                return this.contentManager;
            }
        }

        protected virtual bool CanHaveStyle
        {
            get
            {
                return false;
            }
        }

        protected virtual bool ForceFullEndElement
        {
            get
            {
                return false;
            }
        }

        protected virtual bool PreserveWhiteSpaces
        {
            get
            {
                return false;
            }
        }

        public void Read(IHtmlReader reader, IHtmlImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            this.OnBeforeRead(reader, context);

            reader.BeginReadElement();

            while (reader.MoveToNextAttribute())
            {
                string name = reader.GetAttributeName();
                this.TrySetAttributeValue(name, reader.GetAttributeValue());
            }

            this.EvaluateLocalProperties(context);
            this.OnAfterReadAttributes(reader, context);
            this.EvaluatePropertiesAfterReadAttributes(context);

            string text = reader.GetInnerText();
            if (!string.IsNullOrEmpty(text))
            {
                this.OnReadInnerText(reader, context, text);
            }

            if (reader.HasChildElements())
            {
                do
                {
                    HtmlElementType elementType = reader.GetCurrentChildElementType();
                    HtmlElementBase innerElement;

                    if (elementType == HtmlElementType.Text)
                    {
                        innerElement = this.CreateElement(HtmlElementNames.Text);
                    }
                    else
                    {
                        string elementName = reader.GetCurrentChildElementName();
                        innerElement = this.CreateElement(elementName);
                    }

                    if (innerElement != null)
                    {
                        bool copyOnlyInheritableProperties = ShouldCopyOnlyInheritableProperties(reader.IsInsideSpanElement(), innerElement.GetType());

                        innerElement.CopyInheritablePropertiesFrom(this.LocalProperties, copyOnlyInheritableProperties);
                        this.OnBeforeReadChildElement(reader, context, innerElement);
                        innerElement.Read(reader, context);
                        this.OnAfterReadChildElement(reader, context, innerElement);
                        this.CopyBubblingPropertiesFrom(context, innerElement.BubblingProperties);
                        this.ReleaseElement(innerElement);
                    }
                }
                while (reader.MoveToNextChildElement());
            }

            reader.EndReadElement();

            this.OnAfterRead(reader, context);
        }

        public virtual void Write(IHtmlWriter writer, IHtmlExportContext context)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(context, "context");

            this.OnBeforeWrite(writer, context);

            this.WriteElementStart(writer);
            this.WriteAttributes(writer, context);

            this.WriteContent(writer, context);

            writer.WriteElementEnd(this.ForceFullEndElement);

            this.OnAfterWrite(writer, context);
        }

        protected virtual void WriteContent(IHtmlWriter writer, IHtmlExportContext context)
        {
            if (!string.IsNullOrEmpty(this.InnerText))
            {
                writer.WriteValue(this.InnerText, this.PreserveWhiteSpaces);
            }

            foreach (HtmlElementBase innerElement in this.OnEnumerateInnerElements(writer, context))
            {
                innerElement.Write(writer, context);
                this.ReleaseElement(innerElement);
            }
        }

        protected HtmlAttribute<T> RegisterAttribute<T>(string name, XmlNamespace ns, T defaultValue, bool isRequired = false)
        {
            HtmlAttribute<T> res = new HtmlAttribute<T>(name, ns, defaultValue, isRequired);
            this.RegisterAttribute(res);

            return res;
        }

        protected HtmlAttribute<T> RegisterAttribute<T>(string name, XmlNamespace ns, bool isRequired = false)
        {
            HtmlAttribute<T> res = new HtmlAttribute<T>(name, ns, isRequired);
            this.RegisterAttribute(res);

            return res;
        }

        protected HtmlAttribute<T> RegisterAttribute<T>(string name, T defaultValue, bool isRequired = false)
        {
            return this.RegisterAttribute<T>(name, null, defaultValue, isRequired);
        }

        protected HtmlAttribute<T> RegisterAttribute<T>(string name, bool isRequired = false)
        {
            return this.RegisterAttribute<T>(name, null, default(T), isRequired);
        }

        protected virtual void OnBeforeRead(IHtmlReader reader, IHtmlImportContext context)
        {
        }

        protected virtual void OnAfterReadAttributes(IHtmlReader reader, IHtmlImportContext context)
        {
        }

        protected virtual void EvaluatePropertiesAfterReadAttributes(IHtmlImportContext context)
        {
        }

        protected virtual void OnBeforeReadChildElement(IHtmlReader reader, IHtmlImportContext context, HtmlElementBase innerElement)
        {
        }

        protected virtual void OnAfterReadChildElement(IHtmlReader reader, IHtmlImportContext context, HtmlElementBase innerElement)
        {
        }

        protected virtual void OnAfterRead(IHtmlReader reader, IHtmlImportContext context)
        {
        }

        protected virtual void OnReadInnerText(IHtmlReader reader, IHtmlImportContext context, string text)
        {
        }

        protected virtual void OnBeforeWrite(IHtmlWriter writer, IHtmlExportContext context)
        {
        }

        protected virtual IEnumerable<HtmlElementBase> OnEnumerateInnerElements(IHtmlWriter writer, IHtmlExportContext context)
        {
            return Enumerable.Empty<HtmlElementBase>();
        }

        protected virtual void OnAfterWrite(IHtmlWriter writer, IHtmlExportContext context)
        {
        }

        protected void CopyLocalPropertiesFrom(IHtmlExportContext context, IElementWithProperties element)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(element, "element");

            if (element.Properties != null && element.Properties.HasLocalValues())
            {
                this.LocalProperties.CopyFrom(context, element.Properties);
            }
        }

        protected void CopyStyleFrom(IHtmlExportContext context, IElementWithStyle element)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(element, "element");

            string styleId;
            if (!string.IsNullOrEmpty(element.StyleId))
            {
                this.CopyStylePropertiesFrom(context, element.StyleId);
            }
            else if (this.GetDefaultStyleId(context, out styleId))
            {
                this.CopyStylePropertiesFrom(context, styleId);
            }
        }

        protected virtual bool GetDefaultStyleId(IHtmlExportContext context, out string styleId)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            styleId = null;

            return false;
        }

        protected void CopyStylePropertiesFrom(IHtmlExportContext context, string styleId)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNullOrEmpty(styleId, "styleId");

            switch (context.Settings.StylesExportMode)
            {
                case StylesExportMode.Embedded:
                case StylesExportMode.External:

                    this.SetClassNamesFromStyle(context, styleId);

                    break;

                case StylesExportMode.Inline:

                    this.CopyStylePropertiesAsLocalProperties(context, styleId);

                    break;
            }
        }

        protected void RemoveLocalPropertiesAlreadyInAppliedStyles(IHtmlExportContext context, string elementType)
        {
            if (this.Classes.Count > 0)
            {
                foreach (HtmlStyleProperty property in this.LocalProperties.ToList())
                {
                    foreach (string className in this.Classes.ToEnumerable())
                    {
                        Selector selector = context.HtmlStyleRepository.GetStyle(className, elementType);
                        if (selector != null)
                        {
                            HtmlStyleProperty propertyInSelector;
                            if (selector.TryGetProperty(property.Name, out propertyInSelector))
                            {
                                if (property.UnparsedValue.Equals(propertyInSelector.UnparsedValue))
                                {
                                    this.LocalProperties.RemoveProperty(property.Name);
                                }
                            }
                        }
                    }
                }
            }
        }

        protected void CopyLocalPropertiesTo(IHtmlImportContext context, IElementWithProperties element)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(element, "element");

            this.CopyLocalPropertiesTo(context, element.Properties);
        }

        protected void CopyLocalPropertiesTo(IHtmlImportContext context, DocumentElementPropertiesBase properties)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(properties, "properties");

            if (this.LocalProperties.HasProperties)
            {
                this.LocalProperties.CopyTo(context, properties, this.BubblingProperties);
            }
        }

        protected virtual void RuleOutBubblingProperties(IHtmlImportContext context, IElementWithProperties element)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(element, "element");

            this.RuleOutBubblingProperties(context, element.Properties);
        }

        protected virtual void RuleOutBubblingProperties(IHtmlImportContext context, DocumentElementPropertiesBase properties)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(properties, "properties");

            this.BubblingProperties.RuleOutProperties(context, properties);
        }

        protected void ApplyStyle(IHtmlImportContext context, IElementWithStyle element)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(element, "element");

            if (this.Classes.Count > 0)
            {
                element.StyleId = this.Classes.GetLast();
            }

            this.CopyStyleToOverride(context, element);
        }

        protected virtual void CopyStyleToOverride(IHtmlImportContext context, IElementWithStyle element)
        {
        }

        protected virtual HtmlElementBase CreateElement(string elementName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(elementName, "elementName");

            return this.contentManager.CreateElement(elementName);
        }

        protected T CreateElement<T>(string elementName)
            where T : HtmlElementBase
        {
            Guard.ThrowExceptionIfNullOrEmpty(elementName, "elementName");

            return (T)this.CreateElement(elementName);
        }

        protected void ReleaseElement(HtmlElementBase element)
        {
            Guard.ThrowExceptionIfNull(element, "element");

            this.contentManager.ReleaseElement(element);
        }

        protected override void ClearOverride()
        {
            base.ClearOverride();

            this.InheritedProperties.ClearProperties();
            this.BubblingProperties.Clear();
        }

        private static bool ShouldCopyOnlyInheritableProperties(bool isInSpanElement, Type elementType)
        {
            // All local properties of a span element should be inheritted to its children nevertheles they are inheritable or not.
            bool copyOnlyInheritableProperties = !(isInSpanElement &&
                                                   (elementType == typeof(TextElement) || elementType == typeof(TextBasedElement)));
            return copyOnlyInheritableProperties;
        }

        private void WriteAttributes(IHtmlWriter writer, IHtmlExportContext context)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(context, "context");

            foreach (HtmlAttribute attribute in this.Attributes)
            {
                if (attribute.ShouldExport(writer, context))
                {
                    WriteAttribute(writer, attribute);
                }
            }
        }

        private void CopyInheritablePropertiesFrom(HtmlStyleProperties parentProperties, bool onlyInheritable = true)
        {
            Guard.ThrowExceptionIfNull(parentProperties, "parentProperties");

            this.InheritedProperties.CopyInheritablePropertiesFrom(parentProperties, onlyInheritable);
        }

        private void EvaluateLocalProperties(IHtmlImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            HtmlStyleProperties properties = new HtmlStyleProperties();
            this.ApplyInheritedProperties(context, properties);
            this.ApplyStyleProperties(context, properties);
            this.ApplyLocalProperties(context, properties);

            this.Style.Value = properties;
        }

        private void ApplyInheritedProperties(IHtmlImportContext context, HtmlStyleProperties properties)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(properties, "properties");

            properties.CopyFrom(this.InheritedProperties);
        }

        private void ApplyStyleProperties(IHtmlImportContext context, HtmlStyleProperties properties)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(properties, "properties");

            if (this.Classes.Count > 0)
            {
                string style = null;
                if (this.CanHaveStyle)
                {
                    style = this.Classes.GetFirst();

                    Selector selector;
                    if (context.HtmlStyleRepository.TryGetStyle(style, out selector))
                    {
                        foreach (HtmlStyleProperty property in selector)
                        {
                            HtmlStylePropertyDescriptor descriptor;
                            if (property.Values.HasRelativeValues && HtmlStylePropertyDescriptors.TryGetPropertyDescriptor(property.Name, out descriptor) && descriptor.ApplyAsLocalValueIfRelative)
                            {
                                properties.RegisterProperty(property.Name, property.UnparsedValue);
                            }
                            else
                            {
                                properties.RemoveProperty(property.Name);
                            }
                        }
                    }
                }

                while (this.Classes.Count > 0)
                {
                    Selector selector;
                    if (context.HtmlStyleRepository.TryGetStyle(this.Classes.GetFirst(), out selector))
                    {
                        properties.CopyFrom(selector);
                    }
                }

                if (this.CanHaveStyle && !string.IsNullOrEmpty(style))
                {
                    this.Classes.AddFirst(style);
                }
            }
        }

        private void ApplyLocalProperties(IHtmlImportContext context, HtmlStyleProperties properties)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(properties, "properties");

            properties.CopyFrom(this.LocalProperties);
        }

        private void SetClassNamesFromStyle(IHtmlExportContext context, string styleId)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNullOrEmpty(styleId, "styleId");

            string convertedStyleId = StyleNamesConverter.ConvertStyleNameOnExport(styleId);
            if (context.HtmlStyleRepository.ContainsStyle(convertedStyleId))
            {
                this.Classes.AddFirst(convertedStyleId);

                Style style = context.Document.StyleRepository.GetStyle(styleId);
                if (style != null)
                {
                    while (!string.IsNullOrEmpty(style.BasedOnStyleId))
                    {
                        style = context.Document.StyleRepository.GetStyle(style.BasedOnStyleId);
                        if (style != null)
                        {
                            convertedStyleId = StyleNamesConverter.ConvertStyleNameOnExport(style.Id);
                            if (context.HtmlStyleRepository.ContainsStyle(convertedStyleId))
                            {
                                this.Classes.AddFirst(convertedStyleId);
                            }
                        }
                    }
                }
            }
        }

        private void CopyStylePropertiesAsLocalProperties(IHtmlExportContext context, string styleId)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNullOrEmpty(styleId, "styleId");

            Style style = context.Document.StyleRepository.GetStyle(styleId);
            if (style != null)
            {
                string localStyleId = StyleNamesConverter.ConvertStyleNameOnExport(style.Id);
                this.LocalProperties.CopyFrom(context.HtmlStyleRepository.GetStyle(localStyleId));

                while (!string.IsNullOrEmpty(style.BasedOnStyleId))
                {
                    style = context.Document.StyleRepository.GetStyle(style.BasedOnStyleId);
                    if (style != null)
                    {
                        localStyleId = StyleNamesConverter.ConvertStyleNameOnExport(style.Id);
                        this.LocalProperties.CopyFrom(context.HtmlStyleRepository.GetStyle(localStyleId));
                    }
                }
            }
        }

        private void CopyBubblingPropertiesFrom(IHtmlImportContext context, HtmlBubblingProperties bubblingProperties)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(bubblingProperties, "bubblingProperties");

            this.BubblingProperties.CopyFrom(bubblingProperties);
        }
    }
}
