﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Import;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Attributes;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements.Text;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Fields;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements
{
    internal class HyperlinkElement : BodyElementBase, IPhrasingElement
    {
        private readonly List<HtmlElementBase> innerElements;

        private readonly HtmlAttribute<string> uri;
        private readonly HtmlAttribute<string> target;
        private readonly HtmlAttribute<string> tooltip;

        private FieldInfo hyperlinkField;

        public HyperlinkElement(HtmlContentManager contentManager)
            : base(contentManager)
        {
            this.innerElements = new List<HtmlElementBase>();

            this.uri = this.RegisterAttribute<string>("href");
            this.target = this.RegisterAttribute<string>("target");
            this.tooltip = this.RegisterAttribute<string>("title");
        }

        public override string Name
        {
            get
            {
                return HtmlElementNames.Hyperlink;
            }
        }

        public Hyperlink Hyperlink
        {
            get;
            private set;
        }

        protected override bool CanHaveStyle
        {
            get
            {
                return true;
            }
        }

        public void CopyPropertiesFrom(IHtmlExportContext context, Hyperlink hyperlink)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(hyperlink, "hyperlink");

            this.Hyperlink = hyperlink;

            this.uri.Value = this.Hyperlink.Uri;
            if (!string.IsNullOrEmpty(this.Hyperlink.ToolTip))
            {
                this.tooltip.Value = this.Hyperlink.ToolTip;
            }
        }

        public void AppendInnerElement(HtmlElementBase innerElement)
        {
            Guard.ThrowExceptionIfNull(innerElement, "innerElement");

            this.innerElements.Add(innerElement);
        }

        protected override void OnAfterReadAttributes(IHtmlReader reader, IHtmlImportContext context)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            Hyperlink hyperlink = new Hyperlink(context.Document);

            if (this.uri.HasValue)
            {
                hyperlink.Uri = this.uri.Value;
            }

            if (this.target.HasValue)
            {
                // TODO: how to import this attribute.
            }

            if (this.tooltip.HasValue)
            {
                hyperlink.ToolTip = this.tooltip.Value;
            }

            this.hyperlinkField = new FieldInfo(context.Document, hyperlink);

            context.InsertInline(this.hyperlinkField.Start);
            context.InsertInline(new Run(context.Document) { Text = hyperlink.CreateHyperlinkCode() });
            context.InsertInline(this.hyperlinkField.Separator);

            context.BeginHyperlink();
        }

        protected override void OnAfterRead(IHtmlReader reader, IHtmlImportContext context)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            context.EndHyperlink();
            context.InsertInline(this.hyperlinkField.End);
        }

        protected override IEnumerable<HtmlElementBase> OnEnumerateInnerElements(IHtmlWriter writer, IHtmlExportContext context)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(context, "context");

            return this.innerElements;
        }

        protected override void ClearOverride()
        {
            base.ClearOverride();

            this.hyperlinkField = null;
            this.Hyperlink = null;
            this.innerElements.Clear();
        }
    }
}
