﻿using System;
using System.IO;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Import;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Attributes;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements.Text;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Model.Drawing.Shapes;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements
{
    internal class ImageElement : BodyElementBase, IPhrasingElement
    {
        private const string DataPrefix = "data:image/";
        private const string Base64Prefix = "base64,";
        private const string Delimiter = ";";

        private static readonly string base64SourceFormat = DataPrefix + "{0}" + Delimiter + Base64Prefix + "{1}";

        private readonly HtmlAttribute<string> source;
        private readonly HtmlAttribute<string> alternativeText;
        private readonly HtmlAttribute<string> title;
        private readonly HtmlAttribute<double> width;
        private readonly HtmlAttribute<double> height;

        private Image image;

        public ImageElement(HtmlContentManager contentManager)
            : base(contentManager)
        {
            this.source = this.RegisterAttribute<string>("src", true);
            this.alternativeText = this.RegisterAttribute<string>("alt");
            this.title = this.RegisterAttribute<string>("title");
            this.width = this.RegisterAttribute<double>("width");
            this.height = this.RegisterAttribute<double>("height");
        }

        public override string Name
        {
            get
            {
                return HtmlElementNames.Image;
            }
        }

        public string Source
        {
            get
            {
                return this.source.Value;
            }
            set
            {
                this.source.Value = value;
            }
        }

        public void SetAsociatedFlowElement(Image image)
        {
            Guard.ThrowExceptionIfNull(image, "image");

            this.image = image;
        }

        protected override void OnAfterReadAttributes(IHtmlReader reader, IHtmlImportContext context)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            ImageInline imageInline = new ImageInline(context.Document);
            if (this.Source != null)
            {
                // U+0020 SPACE, U+0009 CHARACTER TABULATION (tab), U+000A LINE FEED (LF), U+000C FORM FEED (FF), U+000D CARRIAGE RETURN (CR).
                string source = this.Source.Trim('\u0020', '\u0009', '\u000A', '\u000C', '\u000D');

                if (!string.IsNullOrEmpty(source))
                {
                    SetImageSource(context, imageInline, source);
                }
            }

            if (this.width.HasValue)
            {
                imageInline.Image.Width = this.width.Value;
            }

            if (this.height.HasValue)
            {
                imageInline.Image.Height = this.height.Value;
            }

            context.InsertInline(imageInline);
        }

        protected override void OnBeforeWrite(IHtmlWriter writer, IHtmlExportContext context)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(context, "context");

            ImageExportingEventArgs args = new ImageExportingEventArgs(this.image);
            context.Settings.OnImageExporting(args);

            if (!string.IsNullOrEmpty(args.AlternativeText))
            {
                this.alternativeText.Value = args.AlternativeText;
            }

            if (args.ExportSize && this.image.SizeInternal != Size.Empty)
            {
                this.width.Value = this.image.Width;
                this.height.Value = this.image.Height;
            }

            if (!string.IsNullOrEmpty(args.Title))
            {
                this.title.Value = args.Title;
            }

            this.ExportImageSource(context, args);
        }

        protected override void ClearOverride()
        {
            base.ClearOverride();

            this.image = null;
        }

        private static void SetImageSource(IHtmlImportContext context, ImageInline imageInline, string source)
        {
            if (source.StartsWith(DataPrefix))
            {
                imageInline.Image.ImageSource = CreateImageSourceFromBase64(source);
            }
            else
            {
                byte[] data = context.Settings.GetDataFromUri(source);
                imageInline.Image.ImageSource = new UriImageSource(new Uri(source, UriKind.RelativeOrAbsolute), data);
            }
        }

        private static ImageSource CreateImageSourceFromBase64(string source)
        {
            Guard.ThrowExceptionIfNullOrEmpty(source, "source");

            try
            {
                source = source.Replace(DataPrefix, string.Empty);
                source = source.Replace(Base64Prefix, string.Empty);
                string[] args = source.Split(new string[] { Delimiter }, StringSplitOptions.RemoveEmptyEntries);

                return new ImageSource(Convert.FromBase64String(args[1]), args[0]);
            }
            catch
            {
                throw new InvalidOperationException("The source is not in the required format");
            }
        }

        private static string BuildBase64Source(ImageSource imageSource)
        {
            Guard.ThrowExceptionIfNull(imageSource, "imageSource");

            return string.Format(base64SourceFormat, imageSource.Extension, Convert.ToBase64String(imageSource.Data));
        }

        private void ExportImageSource(IHtmlExportContext context, ImageExportingEventArgs args)
        {
            switch (context.Settings.ImagesExportMode)
            {
                case ImagesExportMode.Embedded:
                    {
                        ImageSource imageSource = this.image.ImageSource;
                        UriImageSource webImageSource = imageSource as UriImageSource;
                        if (webImageSource != null)
                        {
                            this.source.Value = webImageSource.Uri.OriginalString;
                        }
                        else
                        {
                            this.source.Value = BuildBase64Source(imageSource);
                        }
                    }
                    break;

                case ImagesExportMode.External:
                    {
                        if (args.Handled)
                        {
                            Guard.ThrowExceptionIfNull(args.Source, "Source");
                            this.source.Value = args.Source;
                        }
                        else
                        {
                            UriImageSource webImageSource = this.image.ImageSource as UriImageSource;
                            if (webImageSource != null)
                            {
                                this.source.Value = webImageSource.Uri.OriginalString;
                            }
                            else
                            {
                                this.ExportImageExternal(context);
                            }
                        }
                    }
                    break;

                default:
                    throw new NotSupportedException(string.Format("Unsupported {0}.", typeof(ImagesExportMode).Name));
            }
        }

        private void ExportImageExternal(IHtmlExportContext context)
        {
            string imageName;
            int imageSourceId = this.image.ImageSource.Id;
            if (!context.ImageRepository.TryGetImageName(imageSourceId, out imageName))
            {
                if (context.Settings.ImagesFolderPath == null)
                {
                    throw new NotSupportedException("When exporting image to external file, HtmlExportSettings.ImagesFolderPath should be set, or HtmlExportSettings.ExternalStylesExport should be marked as handled.");
                }

                imageName = context.ImageRepository.RegisterImage(imageSourceId);
                byte[] imageData = this.image.ImageSource.Data;

                File.WriteAllBytes(Path.Combine(context.Settings.ImagesFolderPath, imageName + "." + this.image.ImageSource.Extension), imageData);
            }

            this.source.Value = Path.Combine(context.Settings.ImagesSourceBasePath, imageName + "." + this.image.ImageSource.Extension);
        }
    }
}
