﻿using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Import;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Attributes;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Converters;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements.Lists
{
    internal abstract class ListElement : BlockContainerElementBase
    {
        private readonly HtmlAttribute<int> startIndexAttribute;

        private readonly List<HtmlElementBase> innerElements;
        private ListElement nextListLevelElement;

        public ListElement(HtmlContentManager contentManager)
            : base(contentManager)
        {
            this.innerElements = new List<HtmlElementBase>();
            this.SetInitialIds();

            this.startIndexAttribute = this.RegisterAttribute<int>("start");
        }

        internal int ListId { get; set; }

        internal int ListLevelIndex { get; set; }

        internal bool ShouldRestartLevel { get; set; }

        public void AppendInnerElement(IHtmlExportContext context, Paragraph paragraph)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(paragraph, "paragraph");
            Guard.ThrowExceptionIfLessThan<int>(0, paragraph.ListId, "paragraph.ListId");

            int paragraphListLevelIndex = (paragraph.ListLevel == DocumentDefaultStyleSettings.ListLevel) ? 0 : paragraph.ListLevel;
            if (this.ListLevelIndex == paragraphListLevelIndex)
            {
                this.ClearNextListLevelElement();

                List list = context.Document.Lists.GetList(this.ListId);
                this.ContentManager.ExportListManager.UpdateListIndexes(list, this.ListLevelIndex, this.ShouldRestartLevel);

                int levelBulletSerialIndex = this.ContentManager.ExportListManager.ListIndexes.TotalListItemsPerLevel[this.ListLevelIndex].BulletSerialIndex;

                ListLevelElement listLevelElement = this.CreateElement<ListLevelElement>(HtmlElementNames.ListLevel);
                listLevelElement.SetAssociatedFlowElement(paragraph);
                listLevelElement.SetValue(levelBulletSerialIndex);

                if (this.ShouldRestartLevel)
                {
                    this.ShouldRestartLevel = false;
                }

                this.innerElements.Add(listLevelElement);
            }
            else
            {
                ListElement listElementForListLevelIndex = this.GetListElement(paragraphListLevelIndex);
                if (listElementForListLevelIndex != null)
                {
                    if (listElementForListLevelIndex.innerElements.Count > 0)
                    {
                        HtmlElementBase lastInnerElement = listElementForListLevelIndex.innerElements.Last();
                        if (lastInnerElement is ListElement)
                        {
                            listElementForListLevelIndex.ClearNextListLevelElement();
                        }
                    }
                    listElementForListLevelIndex.AppendInnerElement(context, paragraph);
                }
                else
                {
                    if (this.nextListLevelElement != null)
                    {
                        this.nextListLevelElement.AppendInnerElement(context, paragraph);
                    }
                    else
                    {
                        this.AppendNewListLevel(context, paragraph);
                    }
                }
            }
        }

        public void AppendInnerElement(HtmlElementBase innerElement)
        {
            Guard.ThrowExceptionIfNull(innerElement, "innerElement");

            this.innerElements.Add(innerElement);
        }

        internal void ClearNextListLevelElement()
        {
            if (this.nextListLevelElement != null)
            {
                this.nextListLevelElement.ClearNextListLevelElement();
                this.nextListLevelElement = null;
            }
        }

        protected override void OnBeforeRead(IHtmlReader reader, IHtmlImportContext context)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            base.OnBeforeRead(reader, context);

            context.EndParagraph();
        }

        protected override void OnBeforeWrite(IHtmlWriter writer, IHtmlExportContext context)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(context, "context");

            List list = context.Document.Lists.GetList(this.ListId);
            ListLevel listLevel = list.Levels[this.ListLevelIndex];

            this.ContentManager.ExportListManager.BeginList();

            string convertedResult = string.Empty;
            string bulletType = "disc";

            if (listLevel.NumberingStyle == NumberingStyle.Bullet)
            {
                BulletListInfo bulletListInfo = this.ContentManager.ExportListManager.GetBulletListInfoBySymbol(listLevel.NumberTextFormat);
                if (bulletListInfo != null)
                {
                    bulletType = bulletListInfo.HtmlName;
                }
            }
            else if (HtmlConverters.ListNumberingStyleTypeConverter.ConvertBack(context, listLevel.NumberingStyle, null, out convertedResult))
            {
                bulletType = convertedResult;
            }

            this.LocalProperties.RegisterProperty("list-style-type", bulletType);
        }

        protected override IEnumerable<HtmlElementBase> OnEnumerateInnerElements(IHtmlWriter writer, IHtmlExportContext context)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(context, "context");

            foreach (HtmlElementBase innerElement in this.innerElements)
            {
                if (!(innerElement is ListElement))
                {
                    ((ListLevelElement)innerElement).PrepareToExport(context, this.ListLevelIndex);
                }

                yield return innerElement;
            }
        }

        protected override void OnAfterWrite(IHtmlWriter writer, IHtmlExportContext context)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(context, "context");

            base.OnAfterWrite(writer, context);

            this.ContentManager.ExportListManager.EndList();
        }

        protected override void OnAfterReadAttributes(IHtmlReader reader, IHtmlImportContext context)
        {
            ListLevel listLevel = this.ContentManager.ImportListManager.BeginLevel(context.Document);
            if (this.startIndexAttribute.HasValue)
            {
                listLevel.StartIndex = this.startIndexAttribute.Value;
            }

            this.ApplyAttributesOnListLevelAfterRead(context, listLevel);
        }

        protected abstract void ApplyAttributesOnListLevelAfterRead(IHtmlImportContext context, ListLevel listLevel);

        protected override void OnAfterRead(IHtmlReader reader, IHtmlImportContext context)
        {
            this.ContentManager.ImportListManager.EndLevel();
        }

        protected override void ClearOverride()
        {
            base.ClearOverride();

            if (this.nextListLevelElement != null)
            {
                this.nextListLevelElement.ClearOverride();
                this.nextListLevelElement = null;
            }

            this.SetInitialIds();
            this.innerElements.Clear();
        }

        private void SetInitialIds()
        {
            this.ListId = DocumentDefaultStyleSettings.ListId;
            this.ListLevelIndex = DocumentDefaultStyleSettings.ListId;
        }

        private ListElement GetListElement(int listLevelIndex)
        {
            if (this.ListLevelIndex == listLevelIndex)
            {
                return this;
            }
            else
            {
                if (this.nextListLevelElement != null)
                {
                    ListElement listElement = this.nextListLevelElement.GetListElement(listLevelIndex);
                    return listElement;
                }
                else
                {
                    return null;
                }
            }
        }

        private void AppendNewListLevel(IHtmlExportContext context, Paragraph paragraph)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(paragraph, "paragraph");

            int newListLevelIndex = this.ListLevelIndex + 1;

            ListElement newListElement = this.ContentManager.CreateListElement(context, paragraph, newListLevelIndex);
            List list = context.Document.Lists.GetList(newListElement.ListId);

            newListElement.ShouldRestartLevel = this.ContentManager.ExportListManager.ShouldRestartLevel(list, newListLevelIndex);
            newListElement.AppendInnerElement(context, paragraph);

            this.nextListLevelElement = newListElement;
            this.AppendInnerElement(newListElement);
        }
    }
}
