﻿using System.Linq;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Import;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Attributes;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Converters;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Styles;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;
using Telerik.Windows.Documents.Utilities;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements.Lists
{
    internal class ListLevelElement : ParagraphElementBase
    {
        private readonly HtmlAttribute<int> value;

        public ListLevelElement(HtmlContentManager contentManager)
            : base(contentManager)
        {
            this.value = this.RegisterAttribute<int>("value");
        }

        public override string Name
        {
            get
            {
                return HtmlElementNames.ListLevel;
            }
        }

        private HtmlImportListManager ImportListManager
        {
            get
            {
                return this.ContentManager.ImportListManager;
            }
        }

        private HtmlExportListManager ExportListManager
        {
            get
            {
                return this.ContentManager.ExportListManager;
            }
        }

        internal void SetValue(int value)
        {
            this.value.Value = value;
        }

        internal void PrepareToExport(IHtmlExportContext context, int listLevelIndex)
        {
            List list = context.Document.Lists.GetList(this.Paragraph.ListId);
            ListLevel listLevel = list.Levels[listLevelIndex];

            CharacterProperties paragraphMarkerProperties = this.Paragraph.Properties.ParagraphMarkerProperties;
            foreach (IStyleProperty property in paragraphMarkerProperties.StyleProperties)
            {
                IStylePropertyDefinition propertyDefinition = property.PropertyDefinition;
                IStyleProperty localProperty = listLevel.CharacterProperties.GetStyleProperty(propertyDefinition);

                if (localProperty.HasLocalValue)
                {
                    this.LocalProperties.CopyFrom(context, localProperty, listLevel.CharacterProperties);
                    this.ExportListManager.ExportedCharacterProperties.Add(localProperty);
                }
                else if (property.HasLocalValue)
                {
                    this.LocalProperties.CopyFrom(context, property, paragraphMarkerProperties);
                    this.ExportListManager.ExportedCharacterProperties.Add(property);
                }
            }
        }

        protected override void OnAfterWrite(IHtmlWriter writer, IHtmlExportContext context)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(context, "context");

            base.OnAfterWrite(writer, context);

            this.ExportListManager.ExportedCharacterProperties.Clear();
        }

        protected override void OnAfterReadAttributes(IHtmlReader reader, IHtmlImportContext context)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            base.OnAfterReadAttributes(reader, context);

            if (this.ImportListManager.List == null)
            {
                this.ImportListManager.BeginLevel(context.Document);
            }

            this.ImportListManager.ListIndexes.AddLevelIndex(this.ImportListManager.LevelIndex.Value);
            this.ImportListManager.LevelBulletSerialIndex += 1;
            this.ImportListManager.ResertAllSubsequentLevels();

            if (this.ImportListManager.LevelBulletSerialIndex == 1)
            {
                this.ApplyAttributesToListLevel(context);

                this.ImportListManager.LevelBulletSerialIndex = this.ImportListManager.GetLevel().StartIndex;
            }
            else
            {
                // split the list in two if the bulletSerialIndex is not the same as the value attribute.
                if (this.value.HasValue)
                {
                    int bulletSerialIndex = this.ImportListManager.LevelBulletSerialIndex;
                    if (bulletSerialIndex != this.value.Value)
                    {
                        // clear the list
                        this.ImportListManager.LevelIndex = 0;
                        this.ImportListManager.EndLevel();

                        // start new list
                        this.ImportListManager.BeginLevel(context.Document);

                        this.ApplyAttributesToListLevel(context);
                    }
                }
            }
        }

        protected override void OnBeforeRead(IHtmlReader reader, IHtmlImportContext context)
        {
            base.OnBeforeRead(reader, context);

            context.ListImportContext.BeginListLevel();
        }

        protected override void OnBeforeReadChildElement(IHtmlReader reader, IHtmlImportContext context, HtmlElementBase innerElement)
        {
            base.OnBeforeReadChildElement(reader, context, innerElement);

            if (this.Paragraph == null)
            {
                this.SetAssociatedFlowElement(this.EnsureParagraph(context));
                this.ApplyStyle(context, context.CurrentParagraph);
            }

            this.Paragraph.ListId = this.ImportListManager.List.Id;
            this.Paragraph.ListLevel = this.ImportListManager.LevelIndex.Value;

            context.ListImportContext.CurrentListLevelContentImporter.BeginChildRead();
        }

        protected override void OnAfterReadChildElement(IHtmlReader reader, IHtmlImportContext context, HtmlElementBase innerElement)
        {
            base.OnAfterReadChildElement(reader, context, innerElement);

            this.ApplyListLevelIndentationOnCurrentParagraphIfNecessary(context);

            context.ListImportContext.CurrentListLevelContentImporter.EndChildRead(this.Paragraph.Inlines.Count == 0);
        }

        protected override void OnAfterRead(IHtmlReader reader, IHtmlImportContext context)
        {
            base.OnAfterRead(reader, context);

            context.EndParagraph();
            context.ListImportContext.EndListLevel();
        }

        private void ApplyAttributesToListLevel(IHtmlImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            ListLevel listLevel = this.ImportListManager.GetLevel();
            if (this.value.HasValue)
            {
                listLevel.StartIndex = this.value.Value;
            }

            HtmlStyleProperty listStyleType = this.LocalProperties.Where(i => i.Name == HtmlStylePropertyDescriptors.ListStyleTypeDescriptor.Name).FirstOrDefault();
            if (listStyleType != null)
            {
                NumberingStyle numberingStyle;
                if (HtmlConverters.ListNumberingStyleTypeConverter.Convert(context, listStyleType, null, out numberingStyle))
                {
                    if (numberingStyle != NumberingStyle.None)
                    {
                        bool unorderedListWithOrderedStyleType = listLevel.NumberingStyle == NumberingStyle.Bullet &&
                                                                 numberingStyle != NumberingStyle.Bullet;

                        listLevel.NumberingStyle = numberingStyle;

                        if (numberingStyle == NumberingStyle.Bullet)
                        {
                            BulletListInfo bulletListInfo = this.ContentManager.ImportListManager.GetBulletListInfoByHtmlName(listStyleType.UnparsedValue);
                            if (bulletListInfo != null)
                            {
                                listLevel.NumberTextFormat = bulletListInfo.BulletSymbol;
                                listLevel.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily(bulletListInfo.SymbolFontFamily);
                            }
                        }
                        else if (unorderedListWithOrderedStyleType)
                        {
                            this.ContentManager.ImportListManager.ClearListLevel();
                            listLevel.NumberingStyle = numberingStyle;
                        }
                    }
                }
            }
        }
    }
}
