﻿using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Import;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements
{
    internal abstract class ParagraphElementBase : BodyElementBase
    {
        private const string EmptyParagraphExportCharacter = " ";

        private Paragraph paragraph;

        public ParagraphElementBase(HtmlContentManager contentManager)
            : base(contentManager)
        {
        }

        protected Paragraph Paragraph
        {
            get
            {
                return this.paragraph;
            }
        }

        protected override bool CanHaveStyle
        {
            get
            {
                return true;
            }
        }

        public void SetAssociatedFlowElement(Paragraph paragraph)
        {
            Guard.ThrowExceptionIfNull(paragraph, "paragraph");

            this.paragraph = paragraph;
        }

        protected void ApplyListLevelIndentationOnCurrentParagraphIfNecessary(IHtmlImportContext context)
        {
            if (context.HasCurrentParagraph() &&
                context.ListImportContext.IsInListLevel &&
                context.ListImportContext.CurrentListLevelContentImporter.ShouldApplyIndentationOnParagraph)
            {
                Paragraph currentParagraph = context.CurrentParagraph;
                this.CopyStyleToOverride(context, currentParagraph);

                if (currentParagraph.ListId == DocumentDefaultStyleSettings.ListId &&
                    currentParagraph.ListLevel == DocumentDefaultStyleSettings.ListLevel)
                {
                    ListLevel currentListLevel = this.ContentManager.ImportListManager.GetLevel();
                    currentParagraph.Indentation.LeftIndent = currentListLevel.ParagraphProperties.LeftIndent.GetActualValue().Value;
                }
            }
        }

        protected override void CopyStyleToOverride(IHtmlImportContext context, IElementWithStyle element)
        {
            if (element is Paragraph && string.IsNullOrEmpty(element.StyleId))
            {
                element.StyleId = BuiltInStyleNames.NormalWebStyleId;
                context.ShouldImportNormalWeb = true;
            }
        }

        protected override bool GetDefaultStyleId(IHtmlExportContext context, out string styleId)
        {
            styleId = context.Document.StyleRepository.GetDefaultStyle(StyleType.Paragraph).Id;

            return true;
        }

        protected override void OnBeforeWrite(IHtmlWriter writer, IHtmlExportContext context)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(context, "context");

            base.OnBeforeWrite(writer, context);

            this.CopyStyleFrom(context, this.Paragraph);
            this.CopyLocalPropertiesFrom(context, this.Paragraph);

            context.ClearParagraphClasses(this.Classes);
        }

        protected override void WriteContent(IHtmlWriter writer, IHtmlExportContext context)
        {
            base.WriteContent(writer, context);

            if (this.ShouldExportNonBreakingSpace())
            {
                writer.WriteValue(EmptyParagraphExportCharacter, true);
            }
        }

        protected override IEnumerable<HtmlElementBase> OnEnumerateInnerElements(IHtmlWriter writer, IHtmlExportContext context)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(context, "context");

            foreach (HtmlElementBase element in this.ContentManager.ExportParagraphInlines(context, this.Paragraph))
            {
                yield return element;
            }
        }

        protected override void ClearOverride()
        {
            base.ClearOverride();

            this.paragraph = null;
        }

        private bool ShouldExportNonBreakingSpace()
        {
            // When a paragraph is empty or contains only whitespace runs, a non-breaking space character should be exported.
            List<Run> paragraphRuns = this.paragraph.Inlines.OfType<Run>().ToList();

            bool shouldExportNonBreakingSpace = this.paragraph.Inlines.Count == 0 ||
                                                (paragraphRuns.Count == this.paragraph.Inlines.Count &&
                                                paragraphRuns.All(run => string.IsNullOrEmpty(HtmlTextProcessor.TrimEnd(run.Text))));

            return shouldExportNonBreakingSpace;
        }
    }
}