﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Import;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Attributes;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Converters;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Styles;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Primitives;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements.Tables
{
    internal class TableElement : BodyElementBase
    {
        private readonly HtmlAttribute<string> align;
        private readonly HtmlAttribute<string> cellPadding;

        private Table table;

        public TableElement(HtmlContentManager contentManager)
            : base(contentManager)
        {
            this.align = this.RegisterAttribute<string>("align");
            this.cellPadding = this.RegisterAttribute<string>("cellpadding");

            this.RegisterAttribute(new StyleValueAttribute("width", this.Style, null, false, LegacyConverters.NumericAttributeToPixelUnitTypeConverter));
            this.RegisterAttribute(new StyleValueAttribute("bgcolor", "background-color", this.Style));
            this.RegisterAttribute(new StyleValueAttribute("bordercolor", "border-color", this.Style));
            this.RegisterAttribute(new StyleValueAttribute("border", this.Style, null, false, LegacyConverters.NumericAttributeToPixelUnitTypeConverter));
            this.RegisterAttribute(new StyleValueAttribute("cellspacing", "border-spacing", this.Style, null, false, LegacyConverters.NumericAttributeToPixelUnitTypeConverter));
        }

        public override string Name
        {
            get
            {
                return HtmlElementNames.Table;
            }
        }

        protected override bool CanHaveStyle
        {
            get
            {
                return true;
            }
        }

        public void SetAssociatedFlowElement(Table table)
        {
            Guard.ThrowExceptionIfNull(table, "table");

            this.table = table;
        }

        protected override void OnBeforeWrite(IHtmlWriter writer, IHtmlExportContext context)
        {
            base.OnBeforeWrite(writer, context);

            context.BeginExportTable(this.table);

            this.CopyStyleFrom(context, this.table);
            this.CopyLocalPropertiesFrom(context, this.table);
            this.RemoveLocalPropertiesAlreadyInAppliedStyles(context, HtmlElementNames.Table);

            if (this.table.Alignment != DocumentDefaultStyleSettings.TableAlignment)
            {
                string alignmentAsString;
                if (HtmlValueMappers.AlignmentValueMapper.TryGetFromValue(this.table.Alignment, out alignmentAsString))
                {
                    this.align.Value = alignmentAsString;
                }
            }
        }

        protected override void OnAfterWrite(IHtmlWriter writer, IHtmlExportContext context)
        {
            base.OnAfterWrite(writer, context);

            context.EndExportTable(this.table);
        }

        protected override void OnAfterReadAttributes(IHtmlReader reader, IHtmlImportContext context)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            Table table = new Table(context.Document);

            this.ApplyStyle(context, table);
            this.CopyLocalPropertiesTo(context, table);

            if (this.align.HasValue)
            {
                Alignment alignment;
                if (HtmlValueMappers.AlignmentValueMapper.TryGetToValue(this.align.Value, out alignment))
                {
                    table.Alignment = alignment;
                }
            }

            if (this.cellPadding.HasValue)
            {
                if (!string.IsNullOrEmpty(this.cellPadding.Value))
                {
                    Padding cellPaddingValue;
                    HtmlStyleProperty htmlStyleProperty = new HtmlStyleProperty("cellpadding", this.cellPadding.Value);
                    if (HtmlConverters.PaddingConverter.Convert(context, htmlStyleProperty, table.Properties, out cellPaddingValue))
                    {
                        table.TableCellPadding = cellPaddingValue;
                    }
                }
            }

            context.PushTable(table);
        }

        protected override void OnAfterRead(IHtmlReader reader, IHtmlImportContext context)
        {
            Guard.ThrowExceptionIfNull(reader, "reader");
            Guard.ThrowExceptionIfNull(context, "context");

            context.PopTable();
        }

        protected override IEnumerable<HtmlElementBase> OnEnumerateInnerElements(IHtmlWriter writer, IHtmlExportContext context)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            Guard.ThrowExceptionIfNull(context, "context");

            foreach (TableRow row in this.table.Rows)
            {
                TableRowElement rowElement = this.CreateElement<TableRowElement>(HtmlElementNames.TableRow);
                rowElement.SetAssociatedFlowElement(row);

                yield return rowElement;
            }
        }

        protected override void ClearOverride()
        {
            base.ClearOverride();

            this.table = null;
        }

        protected override bool GetDefaultStyleId(IHtmlExportContext context, out string styleId)
        {
            styleId = context.Document.StyleRepository.GetDefaultStyle(StyleType.Table).Id;

            return true;
        }
    }
}
