﻿using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements.Lists;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements.Tables;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements.Text;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Fields;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Model.Drawing.Shapes;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html.Model
{
    internal class HtmlContentManager
    {
        private readonly HtmlImportListManager importListManager;
        private readonly HtmlExportListManager exportListManager;
        private readonly HtmlElementsPool elementsPool;
        private readonly Stack<HyperlinkElement> hyperlinkElementsStack;
        private ListElement currentExportingList;

        private bool skipUntilSeparateCharacter;

        public HtmlContentManager()
        {
            this.importListManager = new HtmlImportListManager();
            this.exportListManager = new HtmlExportListManager();

            this.elementsPool = new HtmlElementsPool(this);
            this.hyperlinkElementsStack = new Stack<HyperlinkElement>();

            this.skipUntilSeparateCharacter = false;
        }

        internal HtmlImportListManager ImportListManager
        {
            get
            {
                return this.importListManager;
            }
        }

        internal HtmlExportListManager ExportListManager
        {
            get
            {
                return this.exportListManager;
            }
        }

        internal ListElement CurrentExportingList
        {
            get
            {
                return this.currentExportingList;
            }
            set
            {
                this.currentExportingList = value;
            }
        }

        private HyperlinkElement HyperlinkElement
        {
            get
            {
                if (this.hyperlinkElementsStack.Count > 0)
                {
                    return this.hyperlinkElementsStack.Peek();
                }

                return null;
            }
        }

        public HtmlElementBase CreateElement(string name, bool fallback = false)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");

            HtmlElementBase result = this.elementsPool.CreateElement(name);
            if (result == null && fallback)
            {
                result = this.elementsPool.CreateElement(HtmlElementNames.Default);
            }

            return result;
        }

        public T CreateElement<T>(string name)
            where T : HtmlElementBase
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");

            return (T)this.CreateElement(name);
        }

        public void ReleaseElement(HtmlElementBase element)
        {
            Guard.ThrowExceptionIfNull(element, "element");

            element.Clear();
            this.elementsPool.ReleaseElement(element.Name, element);
        }

        internal ListElement CreateListElement(IHtmlExportContext context, Paragraph paragraph, int? forListLevelIndex = null)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(paragraph, "paragraph");

            List list = context.Document.Lists.GetList(paragraph.ListId);

            int listLevel = 0;
            if (!forListLevelIndex.HasValue)
            {
                listLevel = (paragraph.ListLevel == DocumentDefaultStyleSettings.ListLevel) ? 0 : paragraph.ListLevel;
            }
            else
            {
                listLevel = forListLevelIndex.Value;
            }

            ListElement localListElement = null;
            if (list.Levels[listLevel].NumberingStyle == NumberingStyle.Bullet)
            {
                localListElement = this.CreateElement<UnorderedListElement>(HtmlElementNames.UnorderedList);
            }
            else
            {
                localListElement = this.CreateElement<OrderedListElement>(HtmlElementNames.OrderedList);
            }

            localListElement.ListId = paragraph.ListId;
            localListElement.ListLevelIndex = listLevel;

            return localListElement;
        }

        internal IEnumerable<HtmlElementBase> ExportParagraphInlines(IHtmlExportContext context, Paragraph paragraph)
        {
            foreach (InlineBase inline in paragraph.Inlines)
            {
                foreach (HtmlElementBase element in this.ExportInline(context, inline))
                {
                    yield return element;
                }
            }

            if (this.HyperlinkElement != null)
            {
                Hyperlink hyperlink = this.HyperlinkElement.Hyperlink;

                yield return this.hyperlinkElementsStack.Pop();

                HyperlinkElement hyperlinkElement = this.CreateElement<HyperlinkElement>(HtmlElementNames.Hyperlink);
                hyperlinkElement.CopyPropertiesFrom(context, hyperlink);

                this.hyperlinkElementsStack.Push(hyperlinkElement);
            }
        }

        internal IEnumerable<HtmlElementBase> ExportBlockContainer(IHtmlExportContext context, BlockContainerBase blockContainer)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(blockContainer, "blockContainerBase");

            foreach (BlockBase block in blockContainer.Blocks)
            {
                foreach (HtmlElementBase element in this.ExportBlocks(context, block))
                {
                    yield return element;
                }
            }

            if (this.CurrentExportingList != null)
            {
                this.CurrentExportingList.ClearNextListLevelElement();

                yield return this.CurrentExportingList;

                this.CurrentExportingList = null;
            }
        }

        internal IEnumerable<HtmlElementBase> ExportBlocks(IHtmlExportContext context, BlockBase block)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(block, "block");

            switch (block.Type)
            {
                case DocumentElementType.Paragraph:
                    return this.ExportParagraph(context, (Paragraph)block);

                case DocumentElementType.Table:
                    return this.ExportTable(context, (Table)block);
            }

            return Enumerable.Empty<HtmlElementBase>();
        }

        private IEnumerable<HtmlElementBase> ExportInline(IHtmlExportContext context, InlineBase inline)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(inline, "inline");

            if (this.skipUntilSeparateCharacter)
            {
                this.skipUntilSeparateCharacter = (inline.Type == DocumentElementType.FieldCharacter) &&
                    ((FieldCharacter)inline).FieldCharacterType == FieldCharacterType.Separator;
            }
            else
            {
                switch (inline.Type)
                {
                    case DocumentElementType.Run: return this.ExportInline(this.CreateTextBasedElement(context, (Run)inline));

                    case DocumentElementType.FloatingImage: return this.ExportInline(this.CreateImageElement(context, ((FloatingImage)inline).Image));

                    case DocumentElementType.ImageInline: return this.ExportInline(this.CreateImageElement(context, ((ImageInline)inline).Image));

                    case DocumentElementType.FieldCharacter: return this.ExportFieldCharacter(context, (FieldCharacter)inline);

                    case DocumentElementType.Break: return this.ExportBreak(context, (Break)inline);
                }
            }

            return Enumerable.Empty<HtmlElementBase>();
        }

        private IEnumerable<HtmlElementBase> ExportInline(HtmlElementBase element)
        {
            if (this.HyperlinkElement != null)
            {
                this.HyperlinkElement.AppendInnerElement(element);
            }
            else
            {
                yield return element;
            }
        }

        private IEnumerable<HtmlElementBase> ExportBreak(IHtmlExportContext context, Break b)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(context, "context");

            switch (b.BreakType)
            {
                case BreakType.LineBreak: return this.ExportInline(this.CreateBreakElement(context, b));

                default: return Enumerable.Empty<HtmlElementBase>();
            }
        }

        private IEnumerable<HtmlElementBase> ExportFieldCharacter(IHtmlExportContext context, FieldCharacter fieldCharacter)
        {
            Hyperlink hyperlink = fieldCharacter.FieldInfo.Field as Hyperlink;
            if (fieldCharacter.FieldCharacterType == FieldCharacterType.Start)
            {
                this.skipUntilSeparateCharacter = true;

                if (hyperlink != null)
                {
                    HyperlinkElement hyperlinkElement = this.CreateElement<HyperlinkElement>(HtmlElementNames.Hyperlink);
                    hyperlinkElement.CopyPropertiesFrom(context, hyperlink);
                    if (this.HyperlinkElement != null)
                    {
                        this.HyperlinkElement.AppendInnerElement(hyperlinkElement);
                    }

                    this.hyperlinkElementsStack.Push(hyperlinkElement);
                }
            }
            else if (fieldCharacter.FieldCharacterType == FieldCharacterType.End && hyperlink != null)
            {
                yield return this.hyperlinkElementsStack.Pop();
            }
        }

        private TextBasedElement CreateTextBasedElement(IHtmlExportContext context, Run run)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(run, "run");

            TextBasedElement textElement = this.CreateElement<TextBasedElement>(HtmlElementNames.Span);
            textElement.SetAssociatedFlowElement(run);

            return textElement;
        }

        private ImageElement CreateImageElement(IHtmlExportContext context, Image image)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(image, "image");

            image.RecalculateSize();
            ImageElement imageElement = this.CreateElement<ImageElement>(HtmlElementNames.Image);
            imageElement.SetAsociatedFlowElement(image);

            return imageElement;
        }

        private LineBreakElement CreateBreakElement(IHtmlExportContext context, Break b)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(b, "b");

            LineBreakElement lineBreak = this.CreateElement<LineBreakElement>(HtmlElementNames.LineBreak);

            return lineBreak;
        }

        private IEnumerable<HtmlElementBase> ExportTable(IHtmlExportContext context, Table table)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(table, "table");

            TableElement tableElement = this.CreateElement<TableElement>(HtmlElementNames.Table);
            tableElement.SetAssociatedFlowElement(table);

            yield return tableElement;
        }

        private IEnumerable<HtmlElementBase> ExportParagraph(IHtmlExportContext context, Paragraph paragraph)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(paragraph, "paragraph");

            // return the current list if the paragraph is not associated with the it.
            if (this.CurrentExportingList != null && this.CurrentExportingList.ListId != paragraph.ListId)
            {
                this.CurrentExportingList.ClearNextListLevelElement();

                yield return this.CurrentExportingList;

                this.CurrentExportingList = null;
            }

            if (paragraph.ListId == DocumentDefaultStyleSettings.ListId)
            {
                ParagraphElement paragraphElement = this.CreateElement<ParagraphElement>(HtmlElementNames.Paragraph);
                paragraphElement.SetAssociatedFlowElement(paragraph);

                yield return paragraphElement;
            }
            else if (this.CurrentExportingList != null && this.CurrentExportingList.ListId == paragraph.ListId)
            {
                this.CurrentExportingList.AppendInnerElement(context, paragraph);
            }
            else
            {
                ListElement localListElement = this.CreateListElement(context, paragraph, 0);
                ListIndexes newListIndexes = new ListIndexes(localListElement.ListId);

                // transfer the list indexes if the list is the same as the previous one. Needed for continue list feature.
                if (this.ExportListManager.ListIndexes != null && localListElement.ListId == this.ExportListManager.ListIndexes.ListId)
                {
                    newListIndexes.TransferListIndexes(this.ExportListManager.ListIndexes);
                }

                this.ExportListManager.ListIndexes = newListIndexes;

                List list = context.Document.Lists.GetList(localListElement.ListId);
                localListElement.ShouldRestartLevel = this.ExportListManager.ShouldRestartLevel(list, 0);
                localListElement.AppendInnerElement(context, paragraph);
                this.CurrentExportingList = localListElement;
            }
        }
    }
}
