﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements.Lists;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements.Tables;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Elements.Text;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html.Model
{
    internal static class HtmlElementsFactory
    {
        private static readonly Dictionary<string, Func<HtmlContentManager, HtmlElementBase>> factoryMethods;

        static HtmlElementsFactory()
        {
            factoryMethods = new Dictionary<string, Func<HtmlContentManager, HtmlElementBase>>();

            // Default
            RegisterFactoryMethod(HtmlElementNames.Default, (cm) => { return new DefaultElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.Text, (cm) => { return new TextElement(cm); });

            // Html
            RegisterFactoryMethod(HtmlElementNames.Html, (cm) => { return new HtmlElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.Head, (cm) => { return new HeadElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.Body, (cm) => { return new BodyElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.Style, (cm) => { return new StyleElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.Link, (cm) => { return new LinkElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.Title, (cm) => { return new TitleElement(cm); });

            // Paragraph
            RegisterFactoryMethod(HtmlElementNames.Paragraph, (cm) => { return new ParagraphElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.Span, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Span); });
            RegisterFactoryMethod(HtmlElementNames.Bold, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Bold); });
            RegisterFactoryMethod(HtmlElementNames.Strong, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Strong); });
            RegisterFactoryMethod(HtmlElementNames.Strike, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Strike); });
            RegisterFactoryMethod(HtmlElementNames.Striketrough, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Striketrough); });
            RegisterFactoryMethod(HtmlElementNames.Emphasis, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Emphasis); });
            RegisterFactoryMethod(HtmlElementNames.Italic, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Italic); });
            RegisterFactoryMethod(HtmlElementNames.Definition, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Definition); });
            RegisterFactoryMethod(HtmlElementNames.Variable, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Variable); });
            RegisterFactoryMethod(HtmlElementNames.Cite, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Cite); });
            RegisterFactoryMethod(HtmlElementNames.Underline, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Underline); });
            RegisterFactoryMethod(HtmlElementNames.Superscript, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Superscript); });
            RegisterFactoryMethod(HtmlElementNames.Subscript, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Subscript); });
            RegisterFactoryMethod(HtmlElementNames.Code, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Code); });
            RegisterFactoryMethod(HtmlElementNames.Keyboard, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Keyboard); });
            RegisterFactoryMethod(HtmlElementNames.Sample, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Sample); });
            RegisterFactoryMethod(HtmlElementNames.Preformatted, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Preformatted); });
            RegisterFactoryMethod(HtmlElementNames.Teletype, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Teletype); });
            RegisterFactoryMethod(HtmlElementNames.Inserted, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Inserted); });
            RegisterFactoryMethod(HtmlElementNames.Deleted, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Deleted); });
            RegisterFactoryMethod(HtmlElementNames.Label, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Label); });
            RegisterFactoryMethod(HtmlElementNames.Quote, (cm) => { return new TextBasedElement(cm, HtmlElementNames.Quote); });
            RegisterFactoryMethod(HtmlElementNames.Center, (cm) => { return new CenterElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.LineBreak, (cm) => { return new LineBreakElement(cm); });

            // Lists
            RegisterFactoryMethod(HtmlElementNames.OrderedList, (cm) => { return new OrderedListElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.UnorderedList, (cm) => { return new UnorderedListElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.ListLevel, (cm) => { return new ListLevelElement(cm); });

            // Headings
            RegisterFactoryMethod(HtmlElementNames.Heading1, (cm) => { return new ParagraphElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.Heading2, (cm) => { return new ParagraphElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.Heading3, (cm) => { return new ParagraphElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.Heading4, (cm) => { return new ParagraphElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.Heading5, (cm) => { return new ParagraphElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.Heading6, (cm) => { return new ParagraphElement(cm); });

            // Table
            RegisterFactoryMethod(HtmlElementNames.Table, (cm) => { return new TableElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.TableRow, (cm) => { return new TableRowElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.TableCell, (cm) => { return new TableCellElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.TableHeaderCell, (cm) => { return new TableHeaderCellElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.TableHeader, (cm) => { return new TableHeaderElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.TableBody, (cm) => { return new TableBodyElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.TableFooter, (cm) => { return new TableFooterElement(cm); });

            // Other
            RegisterFactoryMethod(HtmlElementNames.Hyperlink, (cm) => { return new HyperlinkElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.Image, (cm) => { return new ImageElement(cm); });
            RegisterFactoryMethod(HtmlElementNames.Div, (cm) => { return new BlockElement(cm, HtmlElementNames.Div); });
        }

        public static bool TryCreateInstance(HtmlContentManager contentManager, string elementName, out HtmlElementBase element)
        {
            Guard.ThrowExceptionIfNull(contentManager, "contentManager");
            Guard.ThrowExceptionIfNullOrEmpty(elementName, "elementName");

            Func<HtmlContentManager, HtmlElementBase> method;
            element = null;
            if (factoryMethods.TryGetValue(elementName, out method))
            {
                element = method(contentManager);

                return true;
            }
            else
            {
                return false;
            }
        }

        private static void RegisterFactoryMethod(string elementName, Func<HtmlContentManager, HtmlElementBase> factoryMethod)
        {
            Guard.ThrowExceptionIfNullOrEmpty(elementName, "elementName");
            Guard.ThrowExceptionIfNull(factoryMethod, "factoryMethod");

            factoryMethods[elementName] = factoryMethod;
        }
    }
}
