﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Converters;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Styles
{
    internal static class HtmlPropertyMappers
    {
        private static readonly Dictionary<string, List<HtmlPropertyMapperElement>> htmlPropertiesToMapperElement = new Dictionary<string, List<HtmlPropertyMapperElement>>();
        private static readonly Dictionary<IStylePropertyDefinition, List<HtmlPropertyMapperElement>> stylePropertyDefinitionsToMapperElement = new Dictionary<IStylePropertyDefinition, List<HtmlPropertyMapperElement>>();

        static HtmlPropertyMappers()
        {
            InitializeMappers();
        }

        internal static IEnumerable<HtmlPropertyMapperElement> TryGetPropertyMappingElements(HtmlStyleProperty property)
        {
            return HtmlPropertyMappers.TryGetPropertyMappingElements(property.Name);
        }

        internal static IEnumerable<HtmlPropertyMapperElement> TryGetPropertyMappingElements(string propertyName)
        {
            if (htmlPropertiesToMapperElement.ContainsKey(propertyName))
            {
                foreach (HtmlPropertyMapperElement mapperElement in htmlPropertiesToMapperElement[propertyName])
                {
                    yield return mapperElement;
                }
            }

            yield break;
        }

        internal static IEnumerable<HtmlPropertyMapperElement> TryGetPropertyMappingElements(IStylePropertyDefinition stylePropertyDefinition)
        {
            if (stylePropertyDefinitionsToMapperElement.ContainsKey(stylePropertyDefinition))
            {
                foreach (HtmlPropertyMapperElement mapperElement in stylePropertyDefinitionsToMapperElement[stylePropertyDefinition])
                {
                    yield return mapperElement;
                }
            }

            yield break;
        }

        private static void AddMapperElement(HtmlPropertyMapperElement mapperElement)
        {
            if (mapperElement.IsDescriptorToDefinitionElement)
            {
                if (!htmlPropertiesToMapperElement.ContainsKey(mapperElement.HtmlPropertyDescriptor.Name))
                {
                    htmlPropertiesToMapperElement[mapperElement.HtmlPropertyDescriptor.Name] = new List<HtmlPropertyMapperElement>();
                }

                htmlPropertiesToMapperElement[mapperElement.HtmlPropertyDescriptor.Name].Add(mapperElement);
            }

            if (mapperElement.IsDefinitionToDescriptorElement)
            {
                IStylePropertyDefinition stylePropertyDefinition = mapperElement.PropertyDefinition;

                if (!stylePropertyDefinitionsToMapperElement.ContainsKey(stylePropertyDefinition))
                {
                    stylePropertyDefinitionsToMapperElement[stylePropertyDefinition] = new List<HtmlPropertyMapperElement>();
                }

                stylePropertyDefinitionsToMapperElement[stylePropertyDefinition].Add(mapperElement);
            }
        }

        private static void InitializeMappers()
        {
            InitFontStyleMapper();
            InitFontWeighteMapper();
            InitForegroundColorMapper();
            InitStrikethroughMapper();
            InitUnderlineMapper();
            InitFontSizeMapper();
            InitFontFamilyMapper();
            InitTextAlignMapper();
            InitBackgroundColorMapper();
            InitHighlightColorMapper();
            InitBaselineAlignmentMapper();

            InitMarginMappers();
            InitFlowDirectionMapper();
            InitFirstLineIndentMapper();
            InitLineSpacingMapper();

            InitTableAndTableCellWidthMappers();
            InitCellSpacingMapper();
            InitTableLayoutMapper();
            InitTableBordersColorMapper();
            InitTableBordersMapper();
            InitTableCellBordersMapper();
            InitTableCellVerticalAlignmentMapper();

            InitRowHeightMapper();

            InitPaddingMappers();
        }

        private static void InitTableBordersMapper()
        {
            HtmlPropertyMapperElement element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BorderPropertyDescriptor,
                Table.TableBordersPropertyDefinition,
                HtmlConverters.TableBordersConverter,
                true,
                true);
            AddMapperElement(element);

            element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BorderLeftPropertyDescriptor,
                Table.TableBordersPropertyDefinition,
                HtmlConverters.TableBordersConverter,
                true,
                true);
            AddMapperElement(element);

            element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BorderTopPropertyDescriptor,
               Table.TableBordersPropertyDefinition,
               HtmlConverters.TableBordersConverter,
               true,
               true);
            AddMapperElement(element);

            element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BorderRightPropertyDescriptor,
                Table.TableBordersPropertyDefinition,
                HtmlConverters.TableBordersConverter,
                true,
                true);
            AddMapperElement(element);

            element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BorderBottomPropertyDescriptor,
                Table.TableBordersPropertyDefinition,
                HtmlConverters.TableBordersConverter,
                true,
                true);
            AddMapperElement(element);
        }

        private static void InitTableBordersColorMapper()
        {
            HtmlPropertyMapperElement element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BorderColorPropertyDescriptor,
                Table.TableBordersPropertyDefinition,
                HtmlConverters.TableBordersColorConverter,
                true,
                true);
            AddMapperElement(element);
        }

        private static void InitTableCellBordersMapper()
        {
            HtmlPropertyMapperElement element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BorderPropertyDescriptor,
                TableCell.BordersPropertyDefinition,
                HtmlConverters.TableCellBordersConverter,
                true,
                true);
            AddMapperElement(element);

            element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BorderLeftPropertyDescriptor,
                TableCell.BordersPropertyDefinition,
                HtmlConverters.TableCellBordersConverter,
                true,
                true);
            AddMapperElement(element);

            element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BorderTopPropertyDescriptor,
               TableCell.BordersPropertyDefinition,
               HtmlConverters.TableCellBordersConverter,
               true,
               true);
            AddMapperElement(element);

            element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BorderRightPropertyDescriptor,
                TableCell.BordersPropertyDefinition,
                HtmlConverters.TableCellBordersConverter,
                true,
                true);
            AddMapperElement(element);

            element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BorderBottomPropertyDescriptor,
                TableCell.BordersPropertyDefinition,
                HtmlConverters.TableCellBordersConverter,
                true,
                true);
            AddMapperElement(element);
        }

        private static void InitTableCellVerticalAlignmentMapper()
        {
            HtmlPropertyMapperElement vertAlignElement = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.VerticalAlignmentPropertyDescriptor,
                TableCell.VerticalAlignmentPropertyDefinition,
                HtmlConverters.VerticalAlignmentConverter,
                true,
                true);
            AddMapperElement(vertAlignElement);
        }

        private static void InitPaddingMappers()
        {
            HtmlPropertyMapperElement padding = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.PaddingPropertyDescriptor,
                TableCell.PaddingPropertyDefinition,
                HtmlConverters.PaddingShortHandConverter,
                true,
                true);
            AddMapperElement(padding);

            HtmlPropertyMapperElement paddingLeft = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.PaddingLeftPropertyDescriptor,
                TableCell.PaddingPropertyDefinition,
                HtmlConverters.PaddingConverter,
                true,
                false);
            AddMapperElement(paddingLeft);

            HtmlPropertyMapperElement paddingRight = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.PaddingRightPropertyDescriptor,
                TableCell.PaddingPropertyDefinition,
                HtmlConverters.PaddingConverter,
                true,
                false);
            AddMapperElement(paddingRight);

            HtmlPropertyMapperElement paddingTop = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.PaddingTopPropertyDescriptor,
                TableCell.PaddingPropertyDefinition,
                HtmlConverters.PaddingConverter,
                true,
                false);
            AddMapperElement(paddingTop);

            HtmlPropertyMapperElement paddingBottom = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.PaddingBottomPropertyDescriptor,
                TableCell.PaddingPropertyDefinition,
                HtmlConverters.PaddingConverter,
                true,
                false);
            AddMapperElement(paddingBottom);
        }

        private static void InitCellSpacingMapper()
        {
            HtmlPropertyMapperElement borderSpacingElement = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BorderSpacingPropertyDescriptor,
                           Table.TableCellSpacingPropertyDefinition,
                           HtmlConverters.CellSpacingConverter,
                           true,
                           true);
            AddMapperElement(borderSpacingElement);

            HtmlPropertyMapperElement borderColapseElement = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BorderCollapsePropertyDescriptor,
                Table.TableCellSpacingPropertyDefinition,
                HtmlConverters.BorderCollapseConverter,
                false,
                true);
            AddMapperElement(borderColapseElement);
        }

        private static void InitTableAndTableCellWidthMappers()
        {
            HtmlPropertyMapperElement tableWidthElement = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.WidthPropertyDescriptor,
                Table.PreferredWidthPropertyDefinition,
                HtmlConverters.PreferredWidthConverter,
                true,
                true);
            AddMapperElement(tableWidthElement);

            HtmlPropertyMapperElement cellWidthElement = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.WidthPropertyDescriptor,
                TableCell.PreferredWidthPropertyDefinition,
                HtmlConverters.PreferredWidthConverter,
                true,
                true);
            AddMapperElement(cellWidthElement);
        }

        private static void InitTableLayoutMapper()
        {
            HtmlPropertyMapperElement element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.TableLayoutPropertyDescriptor,
                Table.LayoutTypePropertyDefinition,
                HtmlConverters.TableLayoutTypeConverter,
                true,
                true);
            AddMapperElement(element);
        }

        private static void InitRowHeightMapper()
        {
            HtmlPropertyMapperElement element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.RowHeightPropertyDescriptor,
                TableRow.HeightPropertyDefinition,
                HtmlConverters.RowHeightConverter,
                true,
                true);
            AddMapperElement(element);
        }

        private static void InitLineSpacingMapper()
        {
            HtmlPropertyMapperElement element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.LineHeightPropertyDescriptor,
                Paragraph.LineSpacingPropertyDefinition,
                HtmlConverters.LineSpacingConverter,
                true,
                false);
            AddMapperElement(element);

            element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.LineHeightPropertyDescriptor,
                Paragraph.LineSpacingTypePropertyDefinition,
                HtmlConverters.LineSpacingTypeConverter,
                true,
                true);
            AddMapperElement(element);
        }

        private static void InitFirstLineIndentMapper()
        {
            HtmlPropertyMapperElement firstLineIndentMapper = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.TextIndentPropertyDescriptor,
                Paragraph.FirstLineIndentPropertyDefinition,
                HtmlConverters.FirstLineIndentConverter,
                true,
                true);
            AddMapperElement(firstLineIndentMapper);

            HtmlPropertyMapperElement hangingIndentMapper = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.TextIndentPropertyDescriptor,
                Paragraph.HangingIndentPropertyDefinition,
                HtmlConverters.HangingIndentConverter,
                true,
                true);
            AddMapperElement(hangingIndentMapper);
        }

        private static void InitFlowDirectionMapper()
        {
            HtmlPropertyMapperElement element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.DirectionPropertyDescriptor,
                Paragraph.FlowDirectionPropertyDefinition,
                HtmlConverters.FlowDirectionConverter,
                true,
                true);
            AddMapperElement(element);

            element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.DirectionPropertyDescriptor,
              Run.FlowDirectionPropertyDefinition,
              HtmlConverters.FlowDirectionConverter,
              true,
              true);
            AddMapperElement(element);
        }

        private static void InitMarginMappers()
        {
            // Paragraphs
            AddMapperElement(new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.MarginLeftPropertyDescriptor,
                Paragraph.LeftIndentPropertyDefinition,
                HtmlConverters.DoubleConverter,
                true,
                true));

            AddMapperElement(new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.MarginRightPropertyDescriptor,
                 Paragraph.RightIndentPropertyDefinition,
                 HtmlConverters.DoubleConverter,
                 true,
                 true));

            AddMapperElement(new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.MarginTopPropertyDescriptor,
                 Paragraph.SpacingBeforePropertyDefinition,
                 HtmlConverters.SpacingBeforeConverter,
                 true,
                 true));

            AddMapperElement(new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.MarginBottomPropertyDescriptor,
                 Paragraph.SpacingAfterPropertyDefinition,
                 HtmlConverters.SpacingAfterConverter,
                 true,
                 true));

            AddMapperElement(new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.MarginTopPropertyDescriptor,
                 Paragraph.AutomaticSpacingBeforePropertyDefinition,
                 HtmlConverters.AutomaticSpacingBeforeConverter,
                 true,
                 false));

            AddMapperElement(new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.MarginBottomPropertyDescriptor,
                 Paragraph.AutomaticSpacingAfterPropertyDefinition,
                 HtmlConverters.AutomaticSpacingAfterConverter,
                 true,
                 false));

            // Table
            AddMapperElement(new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.MarginLeftPropertyDescriptor,
                Table.IndentPropertyDefinition,
                HtmlConverters.DoubleConverter,
                true,
                true));
        }

        private static void InitHighlightColorMapper()
        {
            HtmlPropertyMapperElement element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BackgroundColorPropertyDescriptor,
                 Run.HighlightColorPropertyDefinition,
                 HtmlConverters.ColorMediaConverter,
                 false,
                 true);
            AddMapperElement(element);
        }

        private static void InitBackgroundColorMapper()
        {
            HtmlPropertyMapperElement element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BackgroundColorPropertyDescriptor,
                Run.BackgroundColorPropertyDefinition,
                HtmlConverters.ThemableColorConverter,
                true,
                true);
            AddMapperElement(element);

            element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BackgroundColorPropertyDescriptor,
                Paragraph.BackgroundColorPropertyDefinition,
                HtmlConverters.ThemableColorConverter,
                true,
                true);
            AddMapperElement(element);

            element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BackgroundColorPropertyDescriptor,
                TableCell.BackgroundColorPropertyDefinition,
                HtmlConverters.ThemableColorConverter,
                true,
                true);
            AddMapperElement(element);

            element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.BackgroundColorPropertyDescriptor,
                Table.BackgroundColorPropertyDefinition,
                HtmlConverters.ThemableColorConverter,
                true,
                true);
            AddMapperElement(element);
        }

        private static void InitTextAlignMapper()
        {
            HtmlPropertyMapperElement element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.TextAlignPropertyDescriptor,
                Paragraph.TextAlignmentPropertyDefinition,
                HtmlConverters.AlignmentConverter,
                true,
                true);

            AddMapperElement(element);
        }

        private static void InitFontFamilyMapper()
        {
            HtmlPropertyMapperElement element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.FontFamilyPropertyDescriptor,
                Run.FontFamilyPropertyDefinition,
                HtmlConverters.ThemableFontFamilyConverter,
                true,
                true);

            AddMapperElement(element);
        }

        private static void InitFontSizeMapper()
        {
            HtmlPropertyMapperElement element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.FontSizePropertyDescriptor,
                Run.FontSizePropertyDefinition,
                HtmlConverters.FontSizeConverter,
                true,
                true);

            AddMapperElement(element);
        }

        private static void InitUnderlineMapper()
        {
            HtmlPropertyMapperElement patternElement = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.TextDecorationPropertyDescriptor,
                Run.UnderlinePatternPropertyDefinition,
                HtmlConverters.UnderlinePatternConverter,
                true,
                true);
            AddMapperElement(patternElement);
        }

        private static void InitStrikethroughMapper()
        {
            HtmlPropertyMapperElement element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.TextDecorationPropertyDescriptor,
                Run.StrikethroughPropertyDefinition,
                HtmlConverters.StrikethroughConverter,
                true,
                true);

            AddMapperElement(element);
        }

        private static void InitForegroundColorMapper()
        {
            HtmlPropertyMapperElement element = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.ColorPropertyDescriptor,
                Run.ForegroundColorPropertyDefinition,
                HtmlConverters.ThemableColorConverter,
                true,
                true);

            AddMapperElement(element);
        }

        private static void InitFontWeighteMapper()
        {
            HtmlPropertyMapperElement mapperElement = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.FontWeightPropertyDescriptor,
                Run.FontWeightPropertyDefinition,
                HtmlConverters.FontWeightConverter,
                true,
                true);

            AddMapperElement(mapperElement);
        }

        private static void InitFontStyleMapper()
        {
            HtmlPropertyMapperElement mapperElement = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.FontStylePropertyDescriptor,
                Run.FontStylePropertyDefinition,
                HtmlConverters.FontStyleConverter,
                true,
                true);

            AddMapperElement(mapperElement);
        }

        private static void InitBaselineAlignmentMapper()
        {
            HtmlPropertyMapperElement mapperElement = new HtmlPropertyMapperElement(HtmlStylePropertyDescriptors.VerticalAlignmentPropertyDescriptor,
                Run.BaselineAlignmentPropertyDefinition,
                HtmlConverters.BaselineAlignmentConverter,
                true,
                true);

            AddMapperElement(mapperElement);
        }
    }
}
