﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Export;
using Telerik.Windows.Documents.Flow.FormatProviders.Html.Import;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Html.Model.Styles
{
    internal class HtmlStyleProperties : IEnumerable<HtmlStyleProperty>
    {
        private readonly Dictionary<string, HtmlStyleProperty> htmlStyleProperties;

        public HtmlStyleProperties()
        {
            this.htmlStyleProperties = new Dictionary<string, HtmlStyleProperty>();
        }

        public bool HasProperties
        {
            get
            {
                return this.htmlStyleProperties.Count > 0;
            }
        }

        public void RegisterProperty(string name, string value)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");
            Guard.ThrowExceptionIfNullOrEmpty(value, "value");

            HtmlStyleProperty property = new HtmlStyleProperty(name, value);
            this.RegisterPropertyInternal(property);
        }

        public bool TryGetProperty(string name, out HtmlStyleProperty property)
        {
            if (this.htmlStyleProperties.ContainsKey(name))
            {
                property = this.htmlStyleProperties[name];
                return true;
            }

            property = null;
            return false;
        }

        public void RemoveProperty(string name)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");

            this.htmlStyleProperties.Remove(name);
        }

        public void Write(CssWriter writer)
        {
            foreach (HtmlStyleProperty property in this.htmlStyleProperties.Values)
            {
                writer.WriteStyleProperty(property.Name, property.UnparsedValue);
            }
        }

        public void CopyFrom(Selector style)
        {
            Guard.ThrowExceptionIfNull(style, "style");

            this.CopyPropertiesInternal(style, false);
        }

        public void CopyFrom(HtmlStyleProperties properties)
        {
            Guard.ThrowExceptionIfNull(properties, "properties");

            this.CopyPropertiesInternal(properties, false);
        }

        public void CopyInheritablePropertiesFrom(HtmlStyleProperties properties, bool onlyInheritable = true)
        {
            Guard.ThrowExceptionIfNull(properties, "properties");

            this.CopyPropertiesInternal(properties, onlyInheritable);
        }

        public void CopyFrom(IHtmlExportContext context, DocumentElementPropertiesBase properties)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            if (properties != null)
            {
                foreach (IStyleProperty property in properties.StyleProperties)
                {
                    this.CopyFrom(context, property, properties);
                }
            }
        }

        public void CopyFrom(IHtmlExportContext context, IStyleProperty property, DocumentElementPropertiesBase properties, bool shouldCalculateExportOfActualValue = true)
        {
            foreach (HtmlPropertyMapperElement mapperElement in HtmlPropertyMappers.TryGetPropertyMappingElements(property.PropertyDefinition))
            {
                bool shouldExportActualValue = HtmlStyleProperties.ShouldExportActualValue(property, mapperElement.HtmlPropertyDescriptor);
                if (shouldExportActualValue || property.HasLocalValue || !shouldCalculateExportOfActualValue)
                {
                    object value = (shouldExportActualValue || !shouldCalculateExportOfActualValue) ? property.GetActualValueAsObject() : property.GetLocalValueAsObject();
                    if (value != null)
                    {
                        string result;
                        if (mapperElement.Converter.ConvertBack(context, value, mapperElement.HtmlPropertyDescriptor, properties, out result))
                        {
                            if (!string.IsNullOrEmpty(result))
                            {
                                this.RegisterProperty(mapperElement.HtmlPropertyDescriptor.Name, result);
                            }
                        }
                    }
                }
            }
        }

        public void CopyTo(IHtmlImportContext context, DocumentElementPropertiesBase properties)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            this.CopyTo(context, properties, null);
        }

        public void CopyTo(IHtmlImportContext context, DocumentElementPropertiesBase properties, HtmlBubblingProperties bubblingProperties)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            if (properties != null)
            {
                foreach (KeyValuePair<string, HtmlStyleProperty> property in this.htmlStyleProperties)
                {
                    if (!string.IsNullOrEmpty(property.Value.UnparsedValue))
                    {
                        foreach (HtmlPropertyMapperElement mapperElement in HtmlPropertyMappers.TryGetPropertyMappingElements(property.Value))
                        {
                            IStylePropertyDefinition propertyDefinition = mapperElement.PropertyDefinition;
                            if (propertyDefinition != null)
                            {
                                IStyleProperty styleProperty = properties.GetStyleProperty(propertyDefinition);
                                if (styleProperty != null || mapperElement.HtmlPropertyDescriptor.IsBubblingInheritable)
                                {
                                    object result = null;
                                    if (mapperElement.Converter.Convert(context, property.Value, properties, out result))
                                    {
                                        if (result != null)
                                        {
                                            if (styleProperty != null)
                                            {
                                                styleProperty.SetValueAsObject(result);
                                            }
                                            else if (bubblingProperties != null && mapperElement.HtmlPropertyDescriptor.IsBubblingInheritable && properties.OwnerDocumentElement != null)
                                            {
                                                bubblingProperties.RegisterProperty(properties, property.Value, result);
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        public void ClearProperties()
        {
            this.htmlStyleProperties.Clear();
        }

        public IEnumerator<HtmlStyleProperty> GetEnumerator()
        {
            return this.htmlStyleProperties.Values.GetEnumerator();
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return this.GetEnumerator();
        }

        protected void CopyFromIncludingNonHtmlDefaultActualValues(IHtmlExportContext context, DocumentElementPropertiesBase properties)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            if (properties != null)
            {
                foreach (IStyleProperty property in properties.StyleProperties)
                {
                    bool shouldCopy = false;
                    bool shouldGetActualValue = false;

                    if (property.HasLocalValue)
                    {
                        shouldCopy = true;
                        shouldGetActualValue = false;
                    }
                    else if (StylePropertyHelper.IsStylePropertyWithNonDefaultHtmlActualValue(property.PropertyDefinition))
                    {
                        shouldCopy = true;
                        shouldGetActualValue = true;
                    }

                    if (shouldCopy)
                    {
                        foreach (HtmlPropertyMapperElement mapperElement in HtmlPropertyMappers.TryGetPropertyMappingElements(property.PropertyDefinition))
                        {
                            object value = shouldGetActualValue ? property.GetActualValueAsObject() : property.GetLocalValueAsObject();
                            if (value != null)
                            {
                                string result;
                                if (mapperElement.Converter.ConvertBack(context, value, mapperElement.HtmlPropertyDescriptor, properties, out result))
                                {
                                    if (!string.IsNullOrEmpty(result))
                                    {
                                        this.RegisterProperty(mapperElement.HtmlPropertyDescriptor.Name, result);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        private static bool ShouldExportActualValue(IStyleProperty styleProperty, HtmlStylePropertyDescriptor htmlDescriptor)
        {
            // These properties should be exported even if there are no local values.
            if (styleProperty.PropertyDefinition == TableCell.BordersPropertyDefinition ||
                styleProperty.PropertyDefinition == TableCell.PaddingPropertyDefinition)
            {
                return true;
            }

            // Borders-collapse should always be exported.
            if (htmlDescriptor == HtmlStylePropertyDescriptors.BorderCollapsePropertyDescriptor)
            {
                return true;
            }

            return false;
        }

        private void RegisterPropertyInternal(HtmlStyleProperty property)
        {
            Guard.ThrowExceptionIfNull(property, "property");

            if (this.htmlStyleProperties.ContainsKey(property.Name))
            {
                HtmlStyleProperty currentHtmlStyleProperty = new HtmlStyleProperty(property, this.htmlStyleProperties[property.Name]);
                this.htmlStyleProperties[property.Name] = currentHtmlStyleProperty;
            }
            else
            {
                this.htmlStyleProperties[property.Name] = property;
            }
        }

        private void CopyPropertiesInternal(HtmlStyleProperties properties, bool onlyInheritable)
        {
            foreach (HtmlStyleProperty property in properties.htmlStyleProperties.Values)
            {
                HtmlStylePropertyDescriptor propertyDescriptor;
                if (HtmlStylePropertyDescriptors.TryGetPropertyDescriptor(property.Name, out propertyDescriptor))
                {
                    bool shouldCopy = true;
                    if (onlyInheritable)
                    {
                        shouldCopy = propertyDescriptor.IsInheritable;
                    }

                    if (shouldCopy)
                    {
                        this.RegisterPropertyInternal(property);
                    }
                }
            }
        }
    }
}
