﻿using CsQuery.ExtensionMethods;
using System.Linq;
using Telerik.Windows.Documents.Fixed.Model.Editing;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Pdf.Export
{
    internal class ParagraphSpacingCalculator
    {
        private const double LargeSpacingInPoints = 14;
        private const double SmallSpacingInPoints = 0;

        private static readonly double largeSpacing = Unit.PointToDip(LargeSpacingInPoints);
        private static readonly double smallSpacing = Unit.PointToDip(SmallSpacingInPoints);

        private bool hasSpacingBefore;

        public ParagraphSpacingCalculator()
        {
            this.hasSpacingBefore = true;
        }

        public void SetBlockSpacingBeforeAndAfter(Block block, Paragraph paragraph, Paragraph nextParagraph)
        {
            Guard.ThrowExceptionIfNull(block, "block");
            Guard.ThrowExceptionIfNull(paragraph, "paragraph");

            if (nextParagraph != null && paragraph.ContextualSpacing && HaveSameStyle(paragraph, nextParagraph))
            {
                block.SpacingAfter = 0;
                this.hasSpacingBefore = false;
            }
            else
            {
                block.SpacingAfter = GetSpacingAfter(paragraph);
                this.hasSpacingBefore = true;
            }

            if (this.hasSpacingBefore)
            {
                block.SpacingBefore = GetSpacingBefore(paragraph);
            }
            else
            {
                block.SpacingBefore = 0;
            }
        }

        private static double GetSpacingBefore(Paragraph paragraph)
        {
            if (paragraph.Spacing.AutomaticSpacingBefore)
            {
                return GetAutomaticSpacingBefore(paragraph);
            }

            return paragraph.Spacing.SpacingBefore;
        }

        private static double GetSpacingAfter(Paragraph paragraph)
        {
            if (paragraph.Spacing.AutomaticSpacingAfter)
            {
                return GetAutomaticSpacingAfter(paragraph);
            }

            return paragraph.Spacing.SpacingAfter;
        }

        private static double GetAutomaticSpacingBefore(Paragraph paragraph)
        {
            if (!IsFirstParagraphIsBlock(paragraph) && IsFirstParagraphInList(paragraph))
            {
                return largeSpacing;
            }
            else
            {
                return smallSpacing;
            }
        }

        private static double GetAutomaticSpacingAfter(Paragraph paragraph)
        {
            if (IsInList(paragraph))
            {
                if (IsLastParagraphInList(paragraph))
                {
                    return largeSpacing;
                }

                return smallSpacing;
            }

            if (IsLastParagraphInTableCell(paragraph))
            {
                return smallSpacing;
            }

            return largeSpacing;
        }

        private static bool IsFirstParagraphIsBlock(Paragraph paragraph)
        {
            return paragraph.Parent.Children.First() == paragraph;
        }

        private static bool IsFirstParagraphInList(Paragraph paragraph)
        {
            if (!IsInList(paragraph))
            {
                return false;
            }

            if (IsFirstParagraphIsBlock(paragraph))
            {
                return true;
            }

            Paragraph previousParagraph = GetPreviousParagraph(paragraph);

            return previousParagraph == null ||
                !IsInList(previousParagraph) ||
                paragraph.ListId != previousParagraph.ListId;
        }

        private static bool IsLastParagraphIsBlock(Paragraph paragraph)
        {
            return paragraph.Parent.Children.Last() == paragraph;
        }

        private static bool IsLastParagraphInList(Paragraph paragraph)
        {
            if (!IsInList(paragraph))
            {
                return false;
            }

            if (IsLastParagraphIsBlock(paragraph))
            {
                return true;
            }

            Paragraph nextParagraph = GetNextParagraph(paragraph);
            if (nextParagraph == null)
            {
                return true;
            }

            return nextParagraph == null ||
                !IsInList(nextParagraph) ||
                paragraph.ListId != nextParagraph.ListId;
        }

        private static bool IsLastParagraphInTableCell(Paragraph paragraph)
        {
            TableCell parentTableCell = paragraph.Parent as TableCell;
            if (parentTableCell != null)
            {
                if (parentTableCell.Blocks.Last() == paragraph)
                {
                    return true;
                }
            }

            return false;
        }

        private static bool IsInList(Paragraph paragraph)
        {
            RadFlowDocument document = paragraph.Document;

            return document.Lists.GetList(paragraph.ListId) != null;
        }

        private static Paragraph GetPreviousParagraph(Paragraph paragraph)
        {
            int index = paragraph.Parent.Children.IndexOf(paragraph);
            if (index > 0)
            {
                return paragraph.Parent.Children.ElementAt(index - 1) as Paragraph;
            }

            return null;
        }

        private static Paragraph GetNextParagraph(Paragraph paragraph)
        {
            int index = paragraph.Parent.Children.IndexOf(paragraph);
            if (index < paragraph.Parent.Children.Count() - 1)
            {
                return paragraph.Parent.Children.ElementAt(index + 1) as Paragraph;
            }

            return null;
        }

        private static bool HaveSameStyle(Paragraph p1, Paragraph p2)
        {
            return p1.StyleId == p2.StyleId;
        }
    }
}
