﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Editing;
using Telerik.Windows.Documents.Fixed.Model.Editing.Flow;
using Telerik.Windows.Documents.Fixed.Model.Editing.Layout;
using Telerik.Windows.Documents.Fixed.Model.Editing.Lists;
using Telerik.Windows.Documents.Flow.FormatProviders.Pdf.Utils;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Fields;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Flow.Model.Watermarks;
using Telerik.Windows.Documents.Primitives;
using Telerik.Windows.Documents.Utilities;
using Ex = Telerik.Windows.Documents.Flow.FormatProviders.Pdf.Utils.Extensions;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Pdf.Export
{
    [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Maintainability", "CA1506:AvoidExcessiveClassCoupling")]
    internal class PdfExporter
    {
        private const char Tab = '\t';

        private readonly Dictionary<int, int> flowToFixedListId;
        private readonly PdfExportContext context;
        private readonly Stack<HyperlinkInfo> hyperlinksStack;
        private readonly ParagraphSpacingCalculator paragraphSpacingCalculator;
        private readonly HeadersFootersInfo headerFootersInfo;

        private RadFixedDocument document;
        private Section currentSection;
        private int currentSectionPageNo;
        private bool skipUntilNextFieldCharacter;

        public PdfExporter(PdfExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            this.context = context;
            this.flowToFixedListId = new Dictionary<int, int>();
            this.hyperlinksStack = new Stack<HyperlinkInfo>();
            this.paragraphSpacingCalculator = new ParagraphSpacingCalculator();
            this.headerFootersInfo = new HeadersFootersInfo();
        }

        public RadFixedDocumentEditor Editor
        {
            get;
            private set;
        }

        private HyperlinkInfo CurrentHyperlink
        {
            get
            {
                if (this.hyperlinksStack.Count > 0)
                {
                    return this.hyperlinksStack.Peek();
                }

                return null;
            }
        }

        public RadFixedDocument Export()
        {
            this.document = new RadFixedDocument();

            using (this.Editor = new RadFixedDocumentEditor(this.document))
            {
                this.Editor.PageCreated += this.Editor_PageCreated;
                this.ExportDocument(this.context.Document, this.Editor);
                this.Editor.PageCreated -= this.Editor_PageCreated;
            }

            return this.document;
        }

        private static bool TryCreateHyperlinkInfo(Hyperlink hyperlink, out HyperlinkInfo hyperlinkInfo)
        {
            Uri uri;
            hyperlinkInfo = null;
            if (Uri.TryCreate(hyperlink.Uri, UriKind.Absolute, out uri))
            {
                hyperlinkInfo = new HyperlinkInfo(uri);

                return true;
            }

            return false;
        }

        private static void ExportBreak(RadFixedDocumentEditor editor, Break b, Block block)
        {
            switch (b.BreakType)
            {
                case BreakType.LineBreak:
                    block.InsertLineBreak();
                    break;

                case BreakType.PageBreak:
                    if (editor != null)
                    {
                        editor.InsertBlock(block);
                        editor.InsertPageBreak();
                        block.Clear();
                    }

                    break;
            }
        }

        private static void OverrideParagraphProperties(ListLevel level, Block block)
        {
            if (block.FirstLineIndent != 0)
            {
                level.ParagraphProperties.FirstLineIndent = block.FirstLineIndent;
            }

            if (block.LeftIndent != 0)
            {
                level.ParagraphProperties.LeftIndent = block.LeftIndent;

                bool shouldFixWordsProcessingLeftIndent = block.FirstLineIndent >= 0 && level.ParagraphProperties.FirstLineIndent < 0;

                if (shouldFixWordsProcessingLeftIndent)
                {
                    level.ParagraphProperties.LeftIndent += level.ParagraphProperties.FirstLineIndent;
                }
            }
        }

        private static IEnumerable<string> SplitRunOnTabs(string runText)
        {
            StringBuilder builder = new StringBuilder();
            for (int i = 0; i < runText.Length; i++)
            {
                if (runText[i] == Tab)
                {
                    yield return builder.ToString();
                    yield return Tab.ToString();

                    builder.Clear();
                }
                else
                {
                    builder.Append(runText[i]);
                }
            }

            yield return builder.ToString();
            builder.Clear();
        }

        private static void ExportWatermark(SectionInfo sectionInfo, WatermarkCollection watermarks)
        {
            foreach (Watermark watermark in watermarks)
            {
                switch (watermark.WatermarkType)
                {
                    case WatermarkType.Image:
                        Fixed.Model.Editing.Flow.ImageWatermarkSettings imageSettings = new Fixed.Model.Editing.Flow.ImageWatermarkSettings();
                        imageSettings.CopyPropertiesFrom(watermark.ImageSettings);

                        sectionInfo.InsertWatermark(imageSettings);
                        break;
                    case WatermarkType.Text:
                        Fixed.Model.Editing.Flow.TextWatermarkSettings textSettings = new Fixed.Model.Editing.Flow.TextWatermarkSettings();
                        textSettings.CopyPropertiesFrom(watermark.TextSettings);

                        sectionInfo.InsertWatermark(textSettings);
                        break;
                }
            }
        }

        private void ExportDocument(RadFlowDocument document, RadFixedDocumentEditor editor)
        {
            Guard.ThrowExceptionIfNull(document, "document");
            Guard.ThrowExceptionIfNull(editor, "editor");

            foreach (Model.Lists.List list in document.Lists)
            {
                this.CreateList(list);
            }

            foreach (Section section in document.Sections)
            {
                this.currentSection = section;
                this.currentSectionPageNo = this.document.Pages.Count + 1;
                this.ExportSection(section, editor);
            }
        }

        private void CreateList(Telerik.Windows.Documents.Flow.Model.Lists.List flowList)
        {
            List list = this.Editor.Lists.AddList();
            this.flowToFixedListId[flowList.Id] = list.Id;

            foreach (Model.Lists.ListLevel flowLevel in flowList.Levels)
            {
                list.Levels.Add(this.CreateListLevel(flowLevel));
            }
        }

        private ListLevel CreateListLevel(Model.Lists.ListLevel flowLevel)
        {
            ListLevel listLevel = new ListLevel();
            listLevel.CharacterProperties.CopyPropertiesFrom(this.context, flowLevel.CharacterProperties);
            listLevel.RestartAfterLevel = flowLevel.RestartAfterLevel;
            listLevel.StartIndex = flowLevel.StartIndex;

            double firstLineIndent = Ex.GetFirstLineIndent(flowLevel.ParagraphProperties);
            listLevel.ParagraphProperties.FirstLineIndent = firstLineIndent;
            listLevel.ParagraphProperties.LeftIndent = Ex.CalculateActualLeftIndent(flowLevel.ParagraphProperties.LeftIndent.GetActualValue(), firstLineIndent);

            listLevel.BulletNumberingFormat = this.context.Settings.ExtensibilityManager.ListHelper.CreateBulletFormat(flowLevel);

            //// TODO: This property should be set depending on tab stops in RadWordsProcessing
            //// listLevel.IndentAfterBulletNumbering = ? 

            return listLevel;
        }

        private void ExportSection(Section section, RadFixedDocumentEditor editor)
        {
            Guard.ThrowExceptionIfNull(section, "section");
            Guard.ThrowExceptionIfNull(editor, "editor");

            editor.CopyPropertiesFrom(section.Properties);
            editor.InsertSectionBreak();

            foreach (IBlockElement blockElement in this.ExportBlockContainer(section, editor))
            {
                editor.InsertBlock(blockElement);
            }
        }

        private IEnumerable<IBlockElement> ExportBlockContainer(BlockContainerBase blockContainer, RadFixedDocumentEditor editor = null)
        {
            Guard.ThrowExceptionIfNull(blockContainer, "blockContainer");

            int last = blockContainer.Blocks.Count - 1;
            for (int i = 0; i < blockContainer.Blocks.Count; i++)
            {
                IBlockElement blockElement = this.ExportBlock(editor, blockContainer.Blocks[i], i == last ? null : blockContainer.Blocks[i + 1], i == 0);
                if (blockElement != null)
                {
                    yield return blockElement;
                }
            }
        }

        private IBlockElement ExportBlock(RadFixedDocumentEditor editor, BlockBase block, BlockBase nextBlock, bool isFirstBlockInContainer)
        {
            Guard.ThrowExceptionIfNull(block, "block");

            IBlockElement blockElement = null;
            switch (block.Type)
            {
                case DocumentElementType.Paragraph:
                    blockElement = this.ExportParagraph(editor, (Paragraph)block, nextBlock as Paragraph, isFirstBlockInContainer);
                    break;

                case DocumentElementType.Table:
                    blockElement = this.ExportTable((Table)block);
                    break;
            }

            return blockElement;
        }

        private Fixed.Model.Editing.Tables.Table ExportTable(Table table)
        {
            Fixed.Model.Editing.Tables.Table fixedTable = new Fixed.Model.Editing.Tables.Table();
            fixedTable.Borders = this.ExportTableBorders(table.Borders);
            fixedTable.Background = table.Properties.BackgroundColor.ToColor(this.context);

            Border defaultInnerHorizontalBorders = this.ExportBorder(table.Borders.InsideHorizontal);
            Border defaultInnerVerticalBorders = this.ExportBorder(table.Borders.InsideVertical);

            int lastRow = table.Rows.Count - 1;
            for (int i = 0; i < table.Rows.Count; i++)
            {
                TableRow row = table.Rows[i];
                Fixed.Model.Editing.Tables.TableRow tableRow = fixedTable.Rows.AddTableRow();
                Border innerHorizontalBorders = i < lastRow ? defaultInnerHorizontalBorders : null;
                for (int j = 0; j < row.Cells.Count; j++)
                {
                    int lastCell = row.Cells.Count - 1;
                    TableCell cell = row.Cells[j];
                    Border innerVerticalBorders = j < lastCell ? defaultInnerVerticalBorders : null;
                    this.ExportTableCell(cell, tableRow, innerHorizontalBorders, innerVerticalBorders);
                }
            }

            double availableContentWidth = this.Editor.GetAvailableContentWidth();
            fixedTable.CopyPropertiesFrom(this.context, table.Properties, availableContentWidth);

            return fixedTable;
        }

        private TableBorders ExportTableBorders(Model.Styles.TableBorders tableBorders)
        {
            return new TableBorders(this.ExportBorder(tableBorders.Left),
                this.ExportBorder(tableBorders.Top),
                this.ExportBorder(tableBorders.Right),
                this.ExportBorder(tableBorders.Bottom));
        }

        private void ExportTableCell(TableCell cell, Fixed.Model.Editing.Tables.TableRow tableRow, Border innerHorizontalBorder, Border innerVerticalBorder)
        {
            Fixed.Model.Editing.Tables.TableCell tableCell = tableRow.Cells.AddTableCell();
            tableCell.Background = cell.Properties.BackgroundColor.ToColor(this.context);
            tableCell.Borders = this.ExportTableCellBorders(cell.Borders, innerHorizontalBorder, innerVerticalBorder);
            tableCell.ColumnSpan = cell.ColumnSpan;
            tableCell.RowSpan = cell.RowSpan;
            tableCell.Padding = cell.Padding.ToThickness();
            tableCell.PreferredWidth = cell.PreferredWidth.ToPreferedWidth(this.Editor.GetAvailableContentWidth());

            foreach (IBlockElement block in this.ExportBlockContainer(cell))
            {
                tableCell.Blocks.Add(block);
            }

            if (!tableCell.Blocks.Any())
            {
                Block block = this.CreateDefaultPropertiesBlock();
                block.Insert(new SpaceLayoutElement(block.TextProperties));
                tableCell.Blocks.Add(block);
            }
        }

        private Block CreateDefaultPropertiesBlock()
        {
            Block block = new Block();
            block.CopyPropertiesFrom(this.context, this.context.DefaultValues.CharacterProperties);
            block.CopyPropertiesFrom(this.context, this.context.DefaultValues.ParagraphProperties);

            return block;
        }

        private Fixed.Model.Editing.Tables.TableCellBorders ExportTableCellBorders(Model.Styles.TableCellBorders tableCellBorders, Border innerHorizontalBorder, Border innerVerticalBorder)
        {
            return new Fixed.Model.Editing.Tables.TableCellBorders(this.ExportBorder(tableCellBorders.Left),
                this.ExportBorder(tableCellBorders.Top),
                this.ExportBorder(tableCellBorders.Right, innerVerticalBorder),
                this.ExportBorder(tableCellBorders.Bottom, innerHorizontalBorder),
                this.ExportBorder(tableCellBorders.DiagonalUp),
                this.ExportBorder(tableCellBorders.DiagonalDown));
        }

        private Border ExportBorder(Model.Styles.Border border)
        {
            return new Border(border.Thickness, border.Style.ToBorderStyle(), border.Color.ToColor(this.context));
        }

        private Border ExportBorder(Model.Styles.Border border, Border inheritedBorder)
        {
            if (border.Style == Model.Styles.BorderStyle.Inherit && inheritedBorder != null)
            {
                return inheritedBorder;
            }

            return new Border(border.Thickness, border.Style.ToBorderStyle(), border.Color.ToColor(this.context));
        }

        private Block ExportParagraph(RadFixedDocumentEditor editor, Paragraph paragraph, Paragraph nextParagraph, bool isFirstBlockInContainer)
        {
            Guard.ThrowExceptionIfNull(paragraph, "paragraph");

            if (!isFirstBlockInContainer && paragraph.PageBreakBefore)
            {
                editor.InsertPageBreak();
            }

            Block block = new Block();
            block.CopyPropertiesFrom(this.context, paragraph.Properties);
            this.paragraphSpacingCalculator.SetBlockSpacingBeforeAndAfter(block, paragraph, nextParagraph);

            foreach (HyperlinkInfo hyperlinkInfo in this.hyperlinksStack.Reverse())
            {
                this.ExportHyperlinkStart(hyperlinkInfo, block);
            }

            if (paragraph.ListId > -1)
            {
                int listId = this.GetListId(paragraph.ListId);
                int listLevel = paragraph.ListLevel > -1 ? paragraph.ListLevel : FixedDocumentDefaults.FirstListLevelIndex;
                List list = this.Editor.Lists[listId];
                ListLevel level = list.Levels[listLevel];

                using (level.ParagraphProperties.Save())
                {
                    OverrideParagraphProperties(level, block);

                    block.SetBullet(list, listLevel);
                    this.ExportInlines(editor, paragraph, block);
                }
            }
            else
            {
                this.ExportInlines(editor, paragraph, block);
            }

            return block;
        }

        private void ExportInlines(RadFixedDocumentEditor editor, Paragraph paragraph, Block block)
        {
            Guard.ThrowExceptionIfNull(paragraph, "paragraph");
            Guard.ThrowExceptionIfNull(block, "block");

            if (paragraph.Inlines.Any())
            {
                foreach (InlineBase inline in paragraph.Inlines)
                {
                    this.ExportInline(editor, inline, block);
                }

                block.EndOpenHyperlink();
            }
            else
            {
                block.Insert(new SpaceLayoutElement(block.TextProperties));
            }
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1800:DoNotCastUnnecessarily")]
        private void ExportInline(RadFixedDocumentEditor editor, InlineBase inline, Block block)
        {
            Guard.ThrowExceptionIfNull(inline, "inline");
            Guard.ThrowExceptionIfNull(block, "block");

            if (this.skipUntilNextFieldCharacter)
            {
                this.skipUntilNextFieldCharacter = inline.Type == DocumentElementType.FieldCharacter;
                if (inline.Type == DocumentElementType.FieldCharacter)
                {
                    this.ExportFieldCharacter((FieldCharacter)inline, block);
                }
            }
            else
            {
                switch (inline.Type)
                {
                    case DocumentElementType.Run:
                        this.ExportRun((Run)inline, block);
                        break;

                    case DocumentElementType.ImageInline:
                        this.ExportImageInline((ImageInline)inline, block);
                        break;

                    case DocumentElementType.Break:
                        ExportBreak(editor, (Break)inline, block);
                        break;

                    case DocumentElementType.FieldCharacter:
                        this.ExportFieldCharacter((FieldCharacter)inline, block);
                        break;
                }
            }
        }

        private void ExportFieldCharacter(FieldCharacter fieldCharacter, Block block)
        {
            Hyperlink hyperlink = fieldCharacter.FieldInfo.Field as Hyperlink;
            if (fieldCharacter.FieldCharacterType == FieldCharacterType.Start)
            {
                this.skipUntilNextFieldCharacter = true;
                if (hyperlink != null)
                {
                    HyperlinkInfo hyperlinkInfo;
                    if (TryCreateHyperlinkInfo(hyperlink, out hyperlinkInfo))
                    {
                        this.ExportHyperlinkStart(hyperlinkInfo, block);
                    }
                }
            }
            else if (fieldCharacter.FieldCharacterType == FieldCharacterType.End && hyperlink != null)
            {
                this.ExportHyperlinkEnd(block);
            }
            else if (fieldCharacter.FieldCharacterType == FieldCharacterType.Separator)
            {
                switch (fieldCharacter.FieldInfo.GetCode().Trim().ToLowerInvariant())
                {
                    case "page":
                        block.InsertText(this.GetCurrentPageNumber());
                        this.skipUntilNextFieldCharacter = true;
                        break;
                }
            }
        }

        private string GetCurrentPageNumber()
        {
            int pageNumber = this.document.Pages.Count;
            if (this.currentSection.PageNumberingSettings.StartingPageNumber.HasValue)
            {
                pageNumber += this.currentSection.PageNumberingSettings.StartingPageNumber.Value - 1;
            }

            if (this.currentSection.PageNumberingSettings.PageNumberFormat.HasValue)
            {
                return this.context.Settings.ExtensibilityManager.ListHelper.GetNumberText(
                    this.currentSection.PageNumberingSettings.PageNumberFormat.Value,
                    pageNumber);
            }
            else
            {
                return pageNumber.ToString();
            }
        }

        private void ExportHyperlinkEnd(Block block)
        {
            if (this.CurrentHyperlink != null && this.CurrentHyperlink.Uri != null)
            {
                block.InsertHyperlinkEnd();
                this.hyperlinksStack.Pop();
            }
        }

        private void ExportHyperlinkStart(HyperlinkInfo hyperlinkInfo, Block block)
        {
            if (hyperlinkInfo.Uri != null)
            {
                block.InsertHyperlinkStart(hyperlinkInfo.Uri);
            }

            if (this.hyperlinksStack.Count == 0 || this.hyperlinksStack.Peek() != hyperlinkInfo)
            {
                this.hyperlinksStack.Push(hyperlinkInfo);
            }
        }

        private void ExportImageInline(ImageInline imageInline, Block block)
        {
            Guard.ThrowExceptionIfNull(imageInline, "imageInline");
            Guard.ThrowExceptionIfNull(block, "editor");

            Fixed.Model.Resources.ImageSource imageSource;
            try
            {
                imageSource = this.context.GetImageSource(imageInline.Image.ImageSource);
            }
            catch (System.Net.WebException)
            {
                return;
            }

            imageInline.Image.RecalculateSize();
            block.InsertImage(imageSource, imageInline.Image.Size);
        }

        private void ExportRun(Run run, Block block)
        {
            Guard.ThrowExceptionIfNull(run, "run");
            Guard.ThrowExceptionIfNull(block, "editor");

            block.CopyPropertiesFrom(this.context, run.Properties);

            foreach (string runText in SplitRunOnTabs(run.Text))
            {
                if (runText == Tab.ToString())
                {
                    block.InsertTab();
                }
                else
                {
                    block.InsertText(runText);
                }
            }
        }

        private int GetListId(int flowListId)
        {
            return this.flowToFixedListId[flowListId];
        }

        private double ExportHeader(SectionInfo sectionInfo, Header header)
        {
            double headerHeight = 0;
            sectionInfo.Properties.PageMargins =
#if WPF || SILVERLIGHT
                new Padding(
#else
                new Thickness(
#endif
                    this.currentSection.PageMargins.Left,
                    this.currentSection.HeaderTopMargin,
                    this.currentSection.PageMargins.Right,
                    0);

            sectionInfo.UpdateAvailableSize();
            foreach (IBlockElement block in this.ExportBlockContainer(header))
            {
                sectionInfo.InsertBlockElement(block);
                headerHeight += block.DesiredSize.Height;
            }

            return headerHeight + this.currentSection.HeaderTopMargin;
        }

        private double ExportFooter(SectionInfo sectionInfo, Footer footer)
        {
            Size footerMeasureSize =
                new Size(this.currentSection.PageSize.Width - this.currentSection.PageMargins.Left - this.currentSection.PageMargins.Right,
                double.PositiveInfinity);

            double footerHeight = 0;
            List<IBlockElement> blocks = new List<IBlockElement>();
            foreach (IBlockElement block in this.ExportBlockContainer(footer))
            {
                blocks.Add(block);
                block.Measure(footerMeasureSize);
                footerHeight += block.DesiredSize.Height;
            }

            sectionInfo.Properties.PageMargins =
#if WPF || SILVERLIGHT
                new Padding(
#else
                new Thickness(
#endif
                    this.currentSection.PageMargins.Left,
                    this.currentSection.PageSize.Height - footerHeight - this.currentSection.FooterBottomMargin,
                    this.currentSection.PageMargins.Right,
                    this.currentSection.FooterBottomMargin);

            sectionInfo.UpdateAvailableSize();
            foreach (IBlockElement blockElement in blocks)
            {
                sectionInfo.InsertBlockElement(blockElement);
            }

            return footerHeight + this.currentSection.FooterBottomMargin;
        }

        private HeaderFooterType GetCurrentHeaderFooterType(int pageNo)
        {
            if (this.IsHeaderFooterFirst(pageNo))
            {
                return HeaderFooterType.First;
            }
            else if (this.IsHeaderFooterEven(pageNo))
            {
                return HeaderFooterType.Even;
            }
            else
            {
                return HeaderFooterType.Default;
            }
        }

        private bool IsHeaderFooterEven(int pageNo)
        {
            return this.context.Document.HasDifferentEvenOddPageHeadersFooters && (pageNo % 2) == 0;
        }

        private bool IsHeaderFooterFirst(int pageNo)
        {
            return this.currentSection.HasDifferentFirstPageHeaderFooter && pageNo == this.currentSectionPageNo;
        }

        private void Editor_PageCreated(object sender, RadFixedPageCreatedEventArgs e)
        {
            int pageCount = this.document.Pages.Count;

            this.FillCurrentHeaderFooters();
            HeaderFooterType headerFooterType = this.GetCurrentHeaderFooterType(pageCount);

            double headerHeight = 0;
            Header header = this.headerFootersInfo.GetLastUsedHeader(headerFooterType);
            if (header != null)
            {
                headerHeight = this.ExportHeader(e.Section, header);
            }

            double footerHeight = 0;
            Footer footer = this.headerFootersInfo.GetLastUsedFooter(headerFooterType);
            if (footer != null)
            {
                footerHeight = this.ExportFooter(e.Section, footer);
            }

            e.Section.Properties.PageMargins =
#if WPF || SILVERLIGHT
                new Padding(
#else
                new Thickness(
#endif
                    this.currentSection.PageMargins.Left,
                    Math.Max(headerHeight, this.currentSection.PageMargins.Top),
                    this.currentSection.PageMargins.Right,
                    Math.Max(footerHeight, this.currentSection.PageMargins.Bottom));

            e.Section.UpdateAvailableSize();

            if (header != null)
            {
                ExportWatermark(e.Section, header.Watermarks);
            }
        }

        private void FillCurrentHeaderFooters()
        {
            Headers sectionHeaders = this.currentSection.Headers;
            if (sectionHeaders.Default != null)
            {
                this.headerFootersInfo.SetLastUsedHeader(HeaderFooterType.Default, sectionHeaders.Default);
            }
            if (sectionHeaders.Even != null)
            {
                this.headerFootersInfo.SetLastUsedHeader(HeaderFooterType.Even, sectionHeaders.Even);
            }
            if (sectionHeaders.First != null)
            {
                this.headerFootersInfo.SetLastUsedHeader(HeaderFooterType.First, sectionHeaders.First);
            }

            Footers sectionFooters = this.currentSection.Footers;
            if (sectionFooters.Default != null)
            {
                this.headerFootersInfo.SetLastUsedFooter(HeaderFooterType.Default, sectionFooters.Default);
            }
            if (sectionFooters.Even != null)
            {
                this.headerFootersInfo.SetLastUsedFooter(HeaderFooterType.Even, sectionFooters.Even);
            }
            if (sectionFooters.First != null)
            {
                this.headerFootersInfo.SetLastUsedFooter(HeaderFooterType.First, sectionFooters.First);
            }
        }
    }
}