﻿using System;
using System.IO;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Media;
using Telerik.Documents.Primitives;
#else
using System.Windows;
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;
using Telerik.Windows.Documents.Fixed.Model.Editing;
using Telerik.Windows.Documents.Fixed.Model.Extensions;
using Telerik.Windows.Documents.Flow.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Model.Watermarks;
using Telerik.Windows.Documents.Utilities;
using Telerik.Windows.Documents.Primitives;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Flow.FormatProviders.Pdf.Utils
{
    internal static class Extensions
    {
        public static void CopyPropertiesFrom(this RadFixedDocumentEditor editor, SectionProperties properties)
        {
            Guard.ThrowExceptionIfNull(editor, "editor");
            Guard.ThrowExceptionIfNull(properties, "properties");

#if WPF || SILVERLIGHT
            editor.SectionProperties.PageMargins = properties.PageMargins.GetActualValue();
#else
            editor.SectionProperties.PageMargins = properties.PageMargins.GetActualValue().ToThickness();
#endif
            editor.SectionProperties.PageSize = GetPageSize(properties);
        }

        public static void CopyPropertiesFrom(this Fixed.Model.Editing.Block block, PdfExportContext context, ParagraphProperties properties)
        {
            Guard.ThrowExceptionIfNull(block, "block");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(properties, "properties");

            block.BackgroundColor = properties.BackgroundColor.ToColor(context);
            block.HorizontalAlignment = GetHorizontalAlignment(properties);
            block.FirstLineIndent = GetFirstLineIndent(properties);
            block.LeftIndent = CalculateActualLeftIndent(properties.LeftIndent.GetActualValue(), block.FirstLineIndent);
            block.RightIndent = GetValue(properties.RightIndent.GetActualValue());
            block.LineSpacing = GetValue(properties.LineSpacing.GetActualValue());
            block.LineSpacingType = GetHeightType(properties.LineSpacingType.GetActualValue());
            block.TabStops = GetTabStops(properties.TabStops.GetActualValue(), block);
            block.DefaultTabStopWidth = context.Document.DefaultTabStopWidth;
        }

        public static double CalculateActualLeftIndent(double? wordsProcessingLeftIndent, double firstLineIndent)
        {
            double leftIndent = GetValue(wordsProcessingLeftIndent);

            if (firstLineIndent < 0)
            {
                return leftIndent + firstLineIndent;
            }

            return leftIndent;
        }

        public static double GetFirstLineIndent(ParagraphProperties properties)
        {
            Guard.ThrowExceptionIfNull(properties, "properties");

            double? firstLineIndent = properties.HangingIndent.GetActualValue();

            if (firstLineIndent.HasValue && firstLineIndent.Value != 0)
            {
                return -firstLineIndent.Value;
            }
            else
            {
                firstLineIndent = GetValue(properties.FirstLineIndent.GetActualValue());

                return firstLineIndent.Value;
            }
        }

        public static void CopyPropertiesFrom(this Fixed.Model.Editing.Tables.Table fixedProperties, PdfExportContext context, TableProperties properties, double availableContentWidth)
        {
            Guard.ThrowExceptionIfNull(fixedProperties, "fixedProperties");
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(properties, "properties");

            fixedProperties.LayoutType = properties.GetLayoutType();

            TableWidthUnit actualValue = properties.PreferredWidth.GetActualValue();
            fixedProperties.PreferredWidth = actualValue.ToPreferedWidth(availableContentWidth);
        }

        public static void CopyPropertiesFrom(this Fixed.Model.Editing.Block block, PdfExportContext context, CharacterProperties properties)
        {
            Guard.ThrowExceptionIfNull(block, "block");
            Guard.ThrowExceptionIfNull(properties, "properties");

            FontFamily fontFamily = properties.FontFamily.GetActualValue().GetActualValue(context.Document.Theme);
            FontStyle? fontStyle = properties.FontStyle.GetActualValue();
            FontWeight? fontWeight = properties.FontWeight.GetActualValue();

            FontStyle actualFontStyle = fontStyle.HasValue ? fontStyle.Value : FontStyles.Normal;
            FontWeight actualFontWeight = fontWeight.HasValue ? fontWeight.Value : FontWeights.Normal;
            block.TextProperties.TrySetFont(fontFamily, actualFontStyle, actualFontWeight);
            block.TextProperties.FontSize = properties.FontSize.GetActualValue().Value;
            block.TextProperties.BaselineAlignment = GetBaselineAlignment(properties);
            block.GraphicProperties.FillColor = properties.ForegroundColor.ToColor(context);
            block.TextProperties.HighlightColor = properties.HighlightColor.ToColor(context);
            block.TextProperties.UnderlineColor = properties.UnderlineColor.ToColor(context, properties.ForegroundColor.GetActualValue());
            block.TextProperties.UnderlinePattern = GetUnderlinePattern(properties);
        }

        public static void CopyPropertiesFrom(this Fixed.Model.Editing.Flow.CharacterProperties fixedProperties, PdfExportContext context, CharacterProperties properties)
        {
            Guard.ThrowExceptionIfNull(fixedProperties, "fixedProperties");
            Guard.ThrowExceptionIfNull(properties, "properties");

            FontFamily fontFamily = properties.FontFamily.GetActualValue().GetActualValue(context.Document.Theme);
            FontStyle? fontStyle = properties.FontStyle.GetActualValue();
            FontWeight? fontWeight = properties.FontWeight.GetActualValue();

            FontStyle actualFontStyle = fontStyle.HasValue ? fontStyle.Value : FontStyles.Normal;
            FontWeight actualFontWeight = fontWeight.HasValue ? fontWeight.Value : FontWeights.Normal;
            fixedProperties.TrySetFont(fontFamily, actualFontStyle, actualFontWeight);
            fixedProperties.FontSize = properties.FontSize.GetActualValue().Value;
            fixedProperties.BaselineAlignment = GetBaselineAlignment(properties);
            fixedProperties.ForegroundColor = properties.ForegroundColor.ToColor(context);
            fixedProperties.HighlightColor = properties.HighlightColor.ToColor(context);
            fixedProperties.UnderlineColor = properties.UnderlineColor.ToColor(context);
            fixedProperties.UnderlinePattern = GetUnderlinePattern(properties);
        }

        public static void CopyPropertiesFrom(this Fixed.Model.Editing.Flow.TextWatermarkSettings fixedSettings, TextWatermarkSettings textWatermarkSettings)
        {
            Guard.ThrowExceptionIfNull(fixedSettings, "fixedSettings");
            Guard.ThrowExceptionIfNull(textWatermarkSettings, "textWatermarkSettings");

            RgbColor foreground = (RgbColor)textWatermarkSettings.ForegroundColor.ToColor();
            foreground.A = (byte)(foreground.A * textWatermarkSettings.Opacity);
            fixedSettings.ForegroundColor = foreground;

            fixedSettings.TrySetFont(textWatermarkSettings.FontFamily);
            fixedSettings.Angle = textWatermarkSettings.Angle;
            fixedSettings.Height = textWatermarkSettings.Height;
            fixedSettings.Width = textWatermarkSettings.Width;
            fixedSettings.Text = textWatermarkSettings.Text;
        }

        public static void CopyPropertiesFrom(this Fixed.Model.Editing.Flow.ImageWatermarkSettings fixedSettings, ImageWatermarkSettings imageWatermarkSettings)
        {
            Guard.ThrowExceptionIfNull(fixedSettings, "fixedSettings");
            Guard.ThrowExceptionIfNull(imageWatermarkSettings, "imageWatermarkSettings");

            using (MemoryStream imageData = new MemoryStream(imageWatermarkSettings.ImageSource.Data))
            {
                fixedSettings.ImageSource = new Fixed.Model.Resources.ImageSource(imageData, Fixed.FormatProviders.Pdf.Export.ImageQuality.High);
            }
            fixedSettings.Width = imageWatermarkSettings.Width;
            fixedSettings.Height = imageWatermarkSettings.Height;
            fixedSettings.Angle = imageWatermarkSettings.Angle;
        }

        public static double GetAvailableContentWidth(this RadFixedDocumentEditor editor)
        {
#if WPF || SILVERLIGHT
            Padding margins = editor.SectionProperties.PageMargins;
#else
            Padding margins = editor.SectionProperties.PageMargins.ToPadding();
#endif
            double availableWidth = editor.SectionProperties.PageSize.Width - margins.Left - margins.Right;

            return Math.Max(0, availableWidth);
        }

        public static double? ToPreferedWidth(this Model.Styles.TableWidthUnit tableWidthUnit, double hundredPercentWidth)
        {
            double? value = null;

            switch (tableWidthUnit.Type)
            {
                case TableWidthUnitType.Fixed:
                    value = tableWidthUnit.Value;
                    break;
                case TableWidthUnitType.Percent:
                    value = tableWidthUnit.Value * hundredPercentWidth / 100;
                    break;
            }

            if (value.HasValue && (double.IsNaN(value.Value) || double.IsInfinity(value.Value)))
            {
                value = null;
            }

            return value;
        }

        public static Fixed.Model.Editing.BorderStyle ToBorderStyle(this Model.Styles.BorderStyle borderStyle)
        {
            switch (borderStyle)
            {
                case Model.Styles.BorderStyle.Single:
                    return Fixed.Model.Editing.BorderStyle.Single;

                case Model.Styles.BorderStyle.None:
                    return Fixed.Model.Editing.BorderStyle.None;

                default:
                    return Fixed.Model.Editing.BorderStyle.None;
            }
        }

        public static Fixed.Model.ColorSpaces.ColorBase ToColor(this ThemableColor color, PdfExportContext context)
        {
            Guard.ThrowExceptionIfNull(color, "color");

            return color.GetActualValue(context.Document.Theme).ToColor();
        }

        public static Fixed.Model.ColorSpaces.ColorBase ToColor(this Model.Styles.Core.IStyleProperty<Color?> property, PdfExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(property, "property");

            Color? themableColor = property.GetActualValue();
            if (themableColor.HasValue)
            {
                return themableColor.Value.ToColor();
            }

            return null;
        }

        public static Fixed.Model.ColorSpaces.ColorBase ToColor(this Model.Styles.Core.IStyleProperty<ThemableColor> property, PdfExportContext context)
        {
            Guard.ThrowExceptionIfNull(property, "property");
            Guard.ThrowExceptionIfNull(context, "context");

            return property.GetActualValue().ToColor(context);
        }

        public static Fixed.Model.ColorSpaces.ColorBase ToColor(this Model.Styles.Core.IStyleProperty<ThemableColor> property, PdfExportContext context, ThemableColor alternative)
        {
            Guard.ThrowExceptionIfNull(property, "property");
            Guard.ThrowExceptionIfNull(context, "context");

            ThemableColor color = property.GetActualValue();
            if (color.IsAutomatic && alternative != null)
            {
                color = alternative;
            }

            return color.ToColor(context);
        }

        public static Thickness ToThickness(this Padding padding)
        {
            return new Thickness(padding.Left, padding.Top, padding.Right, padding.Bottom);
        }

        public static Padding ToPadding(this Thickness thickness)
        {
            return new Padding(thickness.Left, thickness.Top, thickness.Right, thickness.Bottom);
        }

        private static double GetValue(double? value)
        {
            if (value.HasValue)
            {
                return value.Value;
            }

            return 0;
        }

        private static Fixed.Model.Editing.Flow.UnderlinePattern GetUnderlinePattern(CharacterProperties properties)
        {
            Guard.ThrowExceptionIfNull(properties, "properties");

            UnderlinePattern? pattern = properties.UnderlinePattern.GetActualValue();
            if (pattern.HasValue)
            {
                switch (pattern.Value)
                {
                    case UnderlinePattern.Single:
                        return Fixed.Model.Editing.Flow.UnderlinePattern.Single;
                }
            }

            return Fixed.Model.Editing.Flow.UnderlinePattern.None;
        }

        private static Fixed.Model.Editing.Flow.BaselineAlignment GetBaselineAlignment(CharacterProperties properties)
        {
            Guard.ThrowExceptionIfNull(properties, "properties");

            Model.Styles.BaselineAlignment? baselineAlignment = properties.BaselineAlignment.GetActualValue();
            if (baselineAlignment.HasValue)
            {
                switch (baselineAlignment.Value)
                {
                    case Model.Styles.BaselineAlignment.Baseline:
                        return Fixed.Model.Editing.Flow.BaselineAlignment.Baseline;

                    case Model.Styles.BaselineAlignment.Subscript:
                        return Fixed.Model.Editing.Flow.BaselineAlignment.Subscript;

                    case Model.Styles.BaselineAlignment.Superscript:
                        return Fixed.Model.Editing.Flow.BaselineAlignment.Superscript;
                }
            }

            return Fixed.Model.Editing.Flow.BaselineAlignment.Baseline;
        }

        private static Size GetPageSize(SectionProperties properties)
        {
            Guard.ThrowExceptionIfNull(properties, "properties");

            Size size = properties.PageSize.GetActualValue().Value;

            return size;
        }

        private static Fixed.Model.Editing.Flow.HorizontalAlignment GetHorizontalAlignment(ParagraphProperties properties)
        {
            Guard.ThrowExceptionIfNull(properties, "properties");

            Alignment? alignment = properties.TextAlignment.GetActualValue();
            if (alignment.HasValue)
            {
                switch (alignment.Value)
                {
                    case Alignment.Center:
                        return Fixed.Model.Editing.Flow.HorizontalAlignment.Center;

                    case Alignment.Right:
                        return Fixed.Model.Editing.Flow.HorizontalAlignment.Right;

                    case Alignment.Left:
                        return Fixed.Model.Editing.Flow.HorizontalAlignment.Left;
                }
            }

            return Fixed.Model.Editing.Flow.HorizontalAlignment.Left;
        }

        private static Fixed.Model.Editing.HeightType GetHeightType(Model.Styles.HeightType? value)
        {
            if (value.HasValue)
            {
                switch (value.Value)
                {
                    case Model.Styles.HeightType.AtLeast:
                        return Fixed.Model.Editing.HeightType.AtLeast;

                    case Model.Styles.HeightType.Auto:
                        return Fixed.Model.Editing.HeightType.Auto;

                    case Model.Styles.HeightType.Exact:
                        return Fixed.Model.Editing.HeightType.Exact;
                }
            }

            return Fixed.Model.Editing.HeightType.Auto;
        }

        private static Fixed.Model.Editing.Flow.TableLayoutType GetLayoutType(this TableProperties properties)
        {
            Guard.ThrowExceptionIfNull(properties, "properties");

            TableLayoutType? layoutType = properties.LayoutType.GetActualValue();

            if (layoutType.HasValue)
            {
                switch (layoutType.Value)
                {
                    case TableLayoutType.AutoFit:
                        return Fixed.Model.Editing.Flow.TableLayoutType.AutoFit;

                    case TableLayoutType.FixedWidth:
                        return Fixed.Model.Editing.Flow.TableLayoutType.FixedWidth;
                }
            }

            return Fixed.Model.Editing.Flow.TableLayoutType.FixedWidth;
        }

        private static Telerik.Windows.Documents.Fixed.Model.Editing.Collections.TabStopCollection GetTabStops(TabStopCollection tabStopCollection, Block parentBlock)
        {
            Fixed.Model.Editing.Collections.TabStopCollection tabStops = new Telerik.Windows.Documents.Fixed.Model.Editing.Collections.TabStopCollection();
            foreach (TabStop tabStop in tabStopCollection)
            {
                tabStops.Add(GetTabStop(tabStop, parentBlock));
            }

            return tabStops;
        }

        private static Fixed.Model.Editing.Flow.TabStop GetTabStop(TabStop tabStop, Block parentBlock)
        {
            Fixed.Model.Editing.Flow.TabStop fixedTabStop = new Fixed.Model.Editing.Flow.TabStop(tabStop.Position);

            switch (tabStop.Type)
            {
                case TabStopType.Left:
                    fixedTabStop.Type = Fixed.Model.Editing.Flow.TabStopType.Left;
                    break;
                case TabStopType.Center:
                    fixedTabStop.Type = Fixed.Model.Editing.Flow.TabStopType.Center;
                    break;
                case TabStopType.Right:
                    fixedTabStop.Type = Fixed.Model.Editing.Flow.TabStopType.Right;
                    fixedTabStop.Position -= parentBlock.LeftIndent;
                    break;
                case TabStopType.Decimal:
                    fixedTabStop.Type = Fixed.Model.Editing.Flow.TabStopType.Decimal;
                    break;
                case TabStopType.Bar:
                    fixedTabStop.Type = Fixed.Model.Editing.Flow.TabStopType.Bar;
                    break;
                case TabStopType.Clear:
                    fixedTabStop.Type = Fixed.Model.Editing.Flow.TabStopType.Clear;
                    break;
                default:
                    throw new ArgumentException(string.Format("Unknown tab stop type: {0}", tabStop.Type));
            }

            switch (tabStop.Leader)
            {
                case TabStopLeader.None:
                    fixedTabStop.Leader = Fixed.Model.Editing.Flow.TabStopLeader.None;
                    break;
                case TabStopLeader.Dot:
                    fixedTabStop.Leader = Fixed.Model.Editing.Flow.TabStopLeader.Dot;
                    break;
                case TabStopLeader.Hyphen:
                    fixedTabStop.Leader = Fixed.Model.Editing.Flow.TabStopLeader.Hyphen;
                    break;
                case TabStopLeader.Underscore:
                    fixedTabStop.Leader = Fixed.Model.Editing.Flow.TabStopLeader.Underscore;
                    break;
                case TabStopLeader.MiddleDot:
                    fixedTabStop.Leader = Fixed.Model.Editing.Flow.TabStopLeader.MiddleDot;
                    break;
                default:
                    throw new ArgumentException(string.Format("Unknown tab stop leader: {0}", tabStop.Leader));
            }

            return fixedTabStop;
        }
    }
}
