﻿using System;
using System.Collections.Generic;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Media;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Utilities;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Flow.Model.Styles;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Export
{
    internal class HeaderExporter
    {
        // English (United States)
        private static readonly int DefaultLanguage = 1033;
        private static readonly int DefaultUnicodeSkipCount = 1;
        private static readonly int DefaultFontID = 0;

        private readonly RtfWriter writer;
        private readonly ExportContext context;

        public HeaderExporter(ExportContext context, RtfWriter writer)
        {
            this.context = context;
            this.writer = writer;
        }

        public void ExportHeader()
        {
            this.writer.WriteTag(RtfTags.TagEncodingAnsi);
            this.writer.WriteTag(RtfTags.TagEncodingAnsiCodePage, CharSetHelper.AnsiCodePage);
            this.writer.WriteTag(RtfTags.TagUnicodeSkipCount, DefaultUnicodeSkipCount);
            this.writer.WriteTag(RtfTags.TagDefaultFont, DefaultFontID);
            this.writer.WriteTag(RtfTags.TagDefaultLanguage, DefaultLanguage);

            this.WriteFontTable();
            this.WriteColorTable();
            this.WriteDefaultStyles();
            this.WriteStyleTable();

            ////// Info
            this.WriteListTable();
            this.WriteListOverridesTable();
            this.writer.DebugWriteNewLine();
        }

        private void WriteFontTable()
        {
            this.writer.DebugWriteNewLine();
            using (this.writer.WriteGroup(RtfTags.TagFontTable))
            {
                foreach (FontFamily font in this.context.FontTable.Keys)
                {
                    int fontId = this.context.FontTable[font];
                    using (this.writer.WriteGroup(RtfTags.TagFont, fontId))
                    {
                        this.writer.WriteText(font.Source);
                        this.writer.WriteText(";");
                    }
                }
            }
        }

        private void WriteColorTable()
        {
            this.writer.DebugWriteNewLine();
            using (this.writer.WriteGroup(RtfTags.TagColorTable))
            {
                for (int i = 0; i < this.context.ColorTable.Count; i++)
                {
                    Color color = this.context.ColorTable.Single(pair => pair.Value == i).Key;

                    this.writer.WriteTag(RtfTags.TagColorRed, color.R);
                    this.writer.WriteTag(RtfTags.TagColorGreen, color.G);
                    this.writer.WriteTag(RtfTags.TagColorBlue, color.B);
                    this.writer.WriteText(";");
                }
            }
        }

        private void WriteDefaultStyles()
        {
            this.writer.DebugWriteNewLine();
            using (this.writer.WriteGroup(RtfTags.TagDefaultCharacterProperties, true))
            {
                InlineExporter.ExportCharacterProperties(
                    this.context.Document.DefaultStyle.CharacterProperties, this.context, this.writer);
            }
            this.writer.DebugWriteNewLine();
            using (this.writer.WriteGroup(RtfTags.TagDefaultParagraphProperties, true))
            {
                ParagraphExporter.ExportParagraphProperties(
                    this.context.Document.DefaultStyle.ParagraphProperties, this.context, this.writer, true, false);
            }
        }

        private void WriteStyleTable()
        {
            using (this.writer.WriteGroup(RtfTags.TagStyleSheet))
            {
                foreach (Style style in this.context.Document.StyleRepository.GetSortedTopologicallyStyles())
                {
                    switch (style.StyleType)
                    {
                        case StyleType.Character:
                            this.ExportSpanStyle(style);
                            break;
                        case StyleType.Paragraph:
                            this.ExportParagraphStyle(style);
                            break;
                        case StyleType.Table:
                            this.ExportTableStyle(style);
                            break;
                        case StyleType.Numbering:
                        default:
                            // Do nothing
                            break;
                    }
                }
            }
        }

        private void WriteListTable()
        {
            if (this.context.Document.Lists == null || this.context.Document.Lists.Count == 0)
            {
                return;
            }
            using (this.writer.WriteGroup(RtfTags.TagListTable, true))
            {
                foreach (List list in this.context.Document.Lists)
                {
                    this.writer.DebugWriteNewLine();
                    this.WriteList(list);
                }
            }
        }
        private void WriteList(List list, bool shouldAssignDifferentId = false)
        {
            bool isCurrentListStyleNumberingDefinition = false;
            using (this.writer.WriteGroup(RtfTags.TagList))
            {
                this.writer.WriteTag(RtfTags.TagListTemplateId, -1); // undefined

                switch (list.MultilevelType)
                {
                    case MultilevelType.SingleLevel:
                        this.writer.WriteTag(RtfTags.TagListSingleLevel);
                        this.WriteListLevelInfo(list.Levels[0]);
                        break;

                    case MultilevelType.Multilevel:
                        this.WriteListLevelsInfo(list);
                        break;

                    case MultilevelType.HybridMultilevel:
                        this.writer.WriteTag(RtfTags.TagListIsHybrid);
                        this.WriteListLevelsInfo(list);
                        break;

                    default:
                        throw new NotSupportedException("Unsupported list mulitilevel type.");
                }

                using (this.writer.WriteGroup(RtfTags.TagListName))
                {
                    this.writer.WriteText(";");
                }

                // Here we should generate different id - it we have exported its Numbering style list before that.
                int id = list.Id;
                if (shouldAssignDifferentId)
                {
                    id = list.GetHashCode();
                    this.context.StyleNumberingListIdsToActualListIds.Add(list.Id, id);
                }
                this.writer.WriteTag(RtfTags.TagListId, id);

                if (!string.IsNullOrEmpty(list.StyleId))
                {
                    Style style = this.context.Document.StyleRepository.GetStyle(list.StyleId);
                    if (style.ParagraphProperties.ListId.HasLocalValue)
                    {
                        if (!this.context.ListNumberingStyleNamesToListIds.ContainsKey(list.StyleId))
                        {
                            this.context.ListNumberingStyleNamesToListIds.Add(list.StyleId, list.Id);
                            using (this.writer.WriteGroup(RtfTags.TagListStyleName))
                            {
                                this.writer.WriteText(list.StyleId);
                                this.writer.WriteText(";");
                            }
                            isCurrentListStyleNumberingDefinition = true;
                        }
                        else
                        {
                            this.writer.WriteTag(RtfTags.TagListStyleId, this.context.ListNumberingStyleNamesToListIds[list.StyleId]);
                        }
                    }
                }
            }

            if (isCurrentListStyleNumberingDefinition)
            {
                this.WriteList(list, true);
            }
        }

        private void WriteListLevelsInfo(List list)
        {
            for (int i = 0; i < list.Levels.Count; i++)
            {
                this.WriteListLevelInfo(list.Levels[i]);
            }
        }

        private void WriteListLevelInfo(ListLevel listLevel)
        {
            using (this.writer.WriteGroup(RtfTags.TagListLevel))
            {
                int rtfListNumberingFormat = (int)listLevel.NumberingStyle;

                this.writer.WriteTag(RtfTags.TagListLevelNumberingFormat, rtfListNumberingFormat);
                this.writer.WriteTag(RtfTags.TagListLevelNumberingFormatAlt, rtfListNumberingFormat);

                this.writer.WriteTag(RtfTags.TagListLevelJustification, RtfHelper.ListLevelAlignmentMapper.GetFromValue(listLevel.Alignment));
                this.writer.WriteTag(RtfTags.TagListLevelJustificationAlt, RtfHelper.ListLevelAlignmentMapper.GetFromValue(listLevel.Alignment));

                // By the RTF specification (2007) \leveltext control word is mandatory, but in OOXML is not, 
                // and thus if a convert is performed from DOCX to RTF, an exception can be thrown.
                string listLevelTextFormat = listLevel.NumberTextFormat;
                if (string.IsNullOrEmpty(listLevelTextFormat))
                {
                    using (this.writer.WriteGroup(RtfTags.TagListLevelText))
                    {
                        this.writer.WriteHex(0);
                    }
                }
                else
                {
                    if (listLevel.NumberingStyle != NumberingStyle.Bullet)
                    {
                        listLevelTextFormat = RtfHelper.ConvertNumberedListFormatToRtfLevelText(listLevelTextFormat);
                    }
                    this.WriteListLevelTextAndNumberingInfo(listLevelTextFormat);
                }

                if (listLevel.IsLegal)
                {
                    this.writer.WriteTag(RtfTags.TagListLevelIsLegal, 1);
                }
                InlineExporter.ExportCharacterProperties(listLevel.CharacterProperties, this.context, this.writer);

                this.writer.WriteTag(RtfTags.TagListLevelFollowChar, 0);
                this.writer.WriteTag(RtfTags.TagListLevelStartIndex, listLevel.StartIndex);
                if (listLevel.RestartAfterLevel != DocumentDefaultStyleSettings.RestartAfterLevel)
                {
                    this.writer.WriteTag(RtfTags.TagListLevelRestartAfterLevel);
                }
                if (!string.IsNullOrEmpty(listLevel.StyleId))
                {
                    int styleId = this.context.StyleTable[listLevel.StyleId];
                    this.writer.WriteTag(RtfTags.TagStyleParagraph, styleId);
                }
                ParagraphExporter.ExportParagraphProperties(listLevel.ParagraphProperties, this.context, this.writer, false);
            }

            this.writer.DebugWriteNewLine();
        }

        private void WriteListLevelTextAndNumberingInfo(string textTemplate)
        {
            string templateWithMarkers = textTemplate.Replace('@', 'a');
            templateWithMarkers = string.Format(templateWithMarkers, '@', '@', '@', '@', '@', '@', '@', '@', '@');
            string templateWithNumbers = string.Format(textTemplate, '0', '1', '2', '3', '4', '5', '6', '7', '8');
            List<int> parameterIndexes = new List<int>();
            using (this.writer.WriteGroup(RtfTags.TagListLevelText))
            {
                int length = templateWithMarkers.Length;

                // write length first
                this.writer.WriteHex(length);
                for (int i = 0; i < length; i++)
                {
                    if (templateWithMarkers[i] == '@')
                    {
                        parameterIndexes.Add(i);
                        int number = int.Parse(templateWithNumbers.Substring(i, 1));
                        this.writer.WriteHex(number);
                    }
                    else
                    {
                        this.writer.WriteChar(templateWithNumbers[i]);
                    }
                }
                this.writer.WriteChar(';');
            }
            using (this.writer.WriteGroup(RtfTags.TagListLevelNumbers))
            {
                foreach (int paramIndex in parameterIndexes)
                {
                    this.writer.WriteHex(paramIndex + 1);
                }
                this.writer.WriteChar(';');
            }
        }
        private void WriteListOverridesTable()
        {
            if (this.context.Document.Lists == null || this.context.Document.Lists.Count == 0)
            {
                return;
            }
            using (this.writer.WriteGroup(RtfTags.TagListOverrideTable, true))
            {
                foreach (List list in this.context.Document.Lists)
                {
                    this.WriteListOverride(list.Id);
                }

                foreach (int listId in this.context.StyleNumberingListIdsToActualListIds.Values)
                {
                    this.WriteListOverride(listId);
                }
            }
        }
        private void WriteListOverride(int listId)
        {
            using (this.writer.WriteGroup(RtfTags.TagListOverride))
            {
                this.writer.WriteTag(RtfTags.TagListId, listId);
                this.writer.WriteTag(RtfTags.TagListOverrideCount, 0);
                this.writer.WriteTag(RtfTags.TagListOverrideId, listId);
            }
        }

        private void ExportSpanStyle(Style style)
        {
            this.writer.DebugWriteNewLine();
            using (this.writer.WriteGroup(RtfTags.TagStyleCharacter, this.context.StyleTable[style.Id], true))
            {
                this.writer.WriteTag(RtfTags.TagStyleAdditive);
                this.ExportStyleMetaProperties(style);
                InlineExporter.ExportCharacterProperties(style.CharacterProperties, this.context, this.writer);
                this.writer.WriteText(style.Name + ";");
            }
        }
        private void ExportParagraphStyle(Style style)
        {
            this.writer.DebugWriteNewLine();
            using (this.writer.WriteGroup(RtfTags.TagStyleParagraph, this.context.StyleTable[style.Id]))
            {
                this.ExportStyleMetaProperties(style);
                InlineExporter.ExportCharacterProperties(style.CharacterProperties, this.context, this.writer);
                ParagraphExporter.ExportParagraphProperties(style.ParagraphProperties, this.context, this.writer, true, true, true);
                this.writer.WriteText(style.Name + ";");
            }
        }
        private void ExportTableStyle(Style style)
        {
            this.writer.DebugWriteNewLine();
            using (this.writer.WriteGroup(RtfTags.TagStyleTable, this.context.StyleTable[style.Id], true))
            {
                this.writer.WriteTag(RtfTags.TagStyleRowDefaults);
                this.ExportStyleMetaProperties(style);
                TableExporter.ExportTableStyle(style, this.context, this.writer);
                InlineExporter.ExportCharacterProperties(style.CharacterProperties, this.context, this.writer);
                ParagraphExporter.ExportParagraphProperties(style.ParagraphProperties, this.context, this.writer, true, true);
                this.writer.WriteText(style.Name + ";");
            }
        }

        private void ExportStyleMetaProperties(Style style)
        {
            int basedOnStyleId;
            if (!string.IsNullOrEmpty(style.BasedOnStyleId) && this.context.StyleTable.TryGetValue(style.BasedOnStyleId, out basedOnStyleId))
            {
                this.writer.WriteTag(RtfTags.TagStyleBasedOn, basedOnStyleId);
            }

            int linkedOnStyleId;
            if (!string.IsNullOrEmpty(style.LinkedStyleId) && this.context.StyleTable.TryGetValue(style.LinkedStyleId, out linkedOnStyleId))
            {
                this.writer.WriteTag(RtfTags.TagStyleLinked, linkedOnStyleId);
            }

            int nextStyleId;
            if (!string.IsNullOrEmpty(style.NextStyleId) && this.context.StyleTable.TryGetValue(style.NextStyleId, out nextStyleId))
            {
                this.writer.WriteTag(RtfTags.TagStyleNextParagraphStyle, nextStyleId);
            }

            if (style.IsPrimary)
            {
                this.writer.WriteTag(RtfTags.TagStyleIsPrimary);
            }

            if (style.UIPriority != DocumentDefaultStyleSettings.UIPriority)
            {
                this.writer.WriteTag(RtfTags.TagStyleUIPriority, style.UIPriority);
            }
        }
    }
}