﻿using System;
using System.Diagnostics;
using System.Globalization;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Utilities;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Model.Drawing.Shapes;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Export
{
    internal class ImageExporter
    {
        private readonly RtfDocumentExporter rtfExporter;
        private readonly RtfWriter writer;

        public ImageExporter(RtfDocumentExporter rtfExporter)
        {
            this.rtfExporter = rtfExporter;
            this.writer = this.rtfExporter.Writer;
        }

        public void ExportImage(Image image)
        {
            using (this.writer.WriteGroup(RtfTags.TagPictureWrapper, true))
            {
                this.ExportImageInternal(image, true);
            }
            if (this.rtfExporter.Settings.ExportImagesInCompatibilityMode)
            {
                using (this.writer.WriteGroup(RtfTags.TagPictureWrapperAlternative, false))
                {
                    this.ExportImageInternal(image, true);
                }
            }
        }

        public void ExportFloatingImage(FloatingImage floatingImage)
        {
            using (this.writer.WriteGroup(RtfTags.TagShape, false))
            {
                using (this.writer.WriteGroup(RtfTags.TagShapeInstructions, true))
                {
                    this.ExportShapeInstructions(floatingImage);
                }
                using (this.writer.WriteGroup(RtfTags.TagShapeResultDestination, false))
                {
                    this.writer.WriteTag(RtfTags.TagParagraphEnd);
                    this.writer.WriteTag(RtfTags.TagParagraphDefaults);
                    this.ExportImage(floatingImage.Image);
                    this.writer.WriteTag(RtfTags.TagParagraphEnd);
                }
            }
        }

        public void WriteImageDataProperty(Image image)
        {
            using (this.writer.WriteGroup(RtfTags.TagShapePopertyDestenation, false))
            {
                using (this.writer.WriteGroup(RtfTags.TagShapePopertyNameDestenation, false))
                {
                    this.writer.WriteText(RtfTags.ShapePictureDataProperty);
                }
                using (this.writer.WriteGroup(RtfTags.TagShapePopertyValueDestenation, false))
                {
                    this.ExportImageInternal(image, false);
                }
            }
        }

        public void WriteShapeProperty(string propertyName, int value)
        {
            this.WriteShapeProperty(propertyName, value.ToString());
        }

        public void WriteShapeProperty(string propertyName, string value)
        {
            this.writer.DebugWriteNewLine();
            using (this.writer.WriteGroup(RtfTags.TagShapePopertyDestenation, false))
            {
                using (this.writer.WriteGroup(RtfTags.TagShapePopertyNameDestenation, false))
                {
                    this.writer.WriteText(propertyName);
                }
                using (this.writer.WriteGroup(RtfTags.TagShapePopertyValueDestenation, false))
                {
                    this.writer.WriteText(value);
                }
            }
        }

        private void ExportImageInternal(Image imageInline, bool exportRotation)
        {
            using (this.writer.WriteGroup(RtfTags.TagPicture))
            {
                imageInline.RecalculateSize();
                Size imageSize = RtfHelper.EnsureShapeSize(imageInline.Size);
                using (this.writer.WriteGroup(RtfTags.TagInlinePictureShapeProperties, true))
                {
                    if (exportRotation && imageInline.RotationAngle != 0)
                    {
                        this.WriteShapeProperty(RtfTags.ShapeRotationProperty, RtfHelper.AngleToRtfValue(imageInline.RotationAngle));
                        imageSize = RtfHelper.CalculateImageSizeForRotateAngle(imageSize, imageInline.RotationAngle);
                    }
                    if (imageInline.IsHorizontallyFlipped)
                    {
                        this.WriteShapeProperty(RtfTags.ShapeHorizontalFlipProperty, 1);
                    }
                    if (imageInline.IsVerticallyFlipped)
                    {
                        this.WriteShapeProperty(RtfTags.ShapeVerticalFlipProperty, 1);
                    }
                }

                this.writer.WriteTag(RtfTags.TagPictureWidth, (int)Math.Round(imageSize.Width));
                this.writer.WriteTag(RtfTags.TagPictureHeight, (int)Math.Round(imageSize.Height));
                this.writer.WriteTag(RtfTags.TagPictureWidthGoal, Unit.DipToTwipI(imageSize.Width));
                this.writer.WriteTag(RtfTags.TagPictureHeightGoal, Unit.DipToTwipI(imageSize.Height));

                byte[] imageBytes = imageInline.ImageSource != null && imageInline.ImageSource.Data != null ? (byte[])imageInline.ImageSource.Data.Clone() : new byte[0];
                string extension = imageInline.ImageSource != null && imageInline.ImageSource.Extension != null ? imageInline.ImageSource.Extension : string.Empty;

                switch (extension.ToLower(CultureInfo.InvariantCulture).TrimStart('.'))
                {
                    case "png":
                        this.writer.WriteTag(RtfTags.TagPictureFormatPng);
                        break;
                    case "jpg":
                    case "jpeg":
                        this.writer.WriteTag(RtfTags.TagPictureFormatJpg);
                        break;
                    case "emf":
                        this.writer.WriteTag(RtfTags.TagPictureFormatEmf);
                        break;
                    case "pict":
                        this.writer.WriteTag(RtfTags.TagPictureFormatQuickDraw);
                        break;
                    case "wmf":
                        this.writer.WriteTag(RtfTags.TagPictureFormatWinMetafile, 8);
                        break;
                    case "bmp":
                        // Export bmp as pngblip otherwise MS Word refuses to open them
                        this.writer.WriteTag(RtfTags.TagPictureFormatPng);
                        break;
                    default:
                        // Probably some unsupported format - try exporting it as device-dependent bitmap 
                        Debug.Assert(false, string.Format("Unsupported image format: {0}", extension));
                        this.writer.WriteTag(RtfTags.TagPictureFormatWinBmp, 0);
                        break;
                }

                this.writer.WriteHex(imageBytes);
            }
        }

        private void ExportShapeInstructions(FloatingImage imageBlock)
        {
            Size imagesize = RtfHelper.EnsureShapeSize(imageBlock.Image.Size);
            imagesize = RtfHelper.CalculateImageSizeForRotateAngle(imagesize, imageBlock.Image.RotationAngle);
            this.writer.WriteTag(RtfTags.TagShapeLeft, Unit.DipToTwipI(imageBlock.HorizontalPosition.Offset));
            this.writer.WriteTag(RtfTags.TagShapeTop, Unit.DipToTwipI(imageBlock.VerticalPosition.Offset));
            this.writer.WriteTag(RtfTags.TagShapeRight, Unit.DipToTwipI(imageBlock.HorizontalPosition.Offset + imagesize.Width));
            this.writer.WriteTag(RtfTags.TagShapeBottom, Unit.DipToTwipI(imageBlock.VerticalPosition.Offset + imagesize.Height));
            this.writer.WriteTag(RtfTags.TagShapeIsInHeader, 0);

            switch (imageBlock.HorizontalPosition.RelativeFrom)
            {
                case HorizontalRelativeFrom.Column:
                    this.writer.WriteTag(RtfTags.TagShapeHorizontalRelativeToColumn);
                    break;
                case HorizontalRelativeFrom.Page:
                    this.writer.WriteTag(RtfTags.TagShapeHorizontalRelativeToPage);
                    break;
                case HorizontalRelativeFrom.Margin:
                case HorizontalRelativeFrom.LeftMargin:
                case HorizontalRelativeFrom.RightMargin:
                default:
                    this.writer.WriteTag(RtfTags.TagShapeHorizontalRelativeToMargin);
                    break;
            }
            this.writer.WriteTag(RtfTags.TagShapeHorizontalRelativeToIgnore);

            switch (imageBlock.VerticalPosition.RelativeFrom)
            {
                case VerticalRelativeFrom.Paragraph:
                    this.writer.WriteTag(RtfTags.TagShapeVerticalRelativeToParagraph);
                    break;
                case VerticalRelativeFrom.Page:
                    this.writer.WriteTag(RtfTags.TagShapeVerticalRelativeToPage);
                    break;
                case VerticalRelativeFrom.Margin:
                case VerticalRelativeFrom.TopMargin:
                case VerticalRelativeFrom.BottomMargin:
                default:
                    this.writer.WriteTag(RtfTags.TagShapeVerticalRelativeToMargin);
                    break;
            }
            this.writer.WriteTag(RtfTags.TagShapeVerticalRelativeToIgnore);

            this.writer.WriteTag(RtfTags.TagShapeWrappingType, RtfHelper.WrappingStyleMapper.GetFromValue(imageBlock.Wrapping.WrappingType));
            this.writer.WriteTag(RtfTags.TagShapeTextWrap, RtfHelper.TextWrapMapper.GetFromValue(imageBlock.Wrapping.TextWrap));
            this.writer.WriteTag(RtfTags.TagShapeIsBelowText, imageBlock.IsBehindDocument ? 1 : 0);
            this.writer.WriteTag(RtfTags.TagShapeZOrder, imageBlock.ZIndex);
            this.WriteShapeProperty(RtfTags.ShapeTypeProperty, RtfHelper.ImageShapeType);
            if (imageBlock.Image.RotationAngle != 0)
            {
                this.WriteShapeProperty(RtfTags.ShapeRotationProperty, RtfHelper.AngleToRtfValue(imageBlock.Image.RotationAngle));
            }
            this.WriteImageDataProperty(imageBlock.Image);
            this.WriteShapeProperty("pictureBiLevel", 0);
            this.WriteShapeProperty("fFilled", 0);

            if (imageBlock.HorizontalPosition.ValueType == PositionValueType.Alignment)
            {
                this.WriteShapeProperty(RtfTags.ShapeHorizontalPositionProperty, RtfHelper.RelativeHorizontalAlignmentMapper.GetFromValue(imageBlock.HorizontalPosition.Alignment));
            }
            this.WriteShapeProperty(RtfTags.ShapeHorizontalRelativeToPositionProperty, RtfHelper.HorizontalRelativeFromMapper.GetFromValue(imageBlock.HorizontalPosition.RelativeFrom));
            if (imageBlock.VerticalPosition.ValueType == PositionValueType.Alignment)
            {
                this.WriteShapeProperty(RtfTags.ShapeVerticalPositionProperty, RtfHelper.RelativeVerticalAlignmentMapper.GetFromValue(imageBlock.VerticalPosition.Alignment));
            }
            this.WriteShapeProperty(RtfTags.ShapeVerticalRelativeToPositionProperty, RtfHelper.VerticalRelativeFromMapper.GetFromValue(imageBlock.VerticalPosition.RelativeFrom));
            this.WriteShapeProperty(RtfTags.ShapeAllowOverlapProperty, imageBlock.AllowOverlap ? 1 : 0);
            this.WriteShapeProperty(RtfTags.ShapeBehindDocumentProperty, imageBlock.IsBehindDocument ? 1 : 0);
            this.WriteShapeProperty(RtfTags.ShapeLayoutInCellProperty, imageBlock.LayoutInCell ? 1 : 0);

            this.WriteShapeProperty(RtfTags.ShapeZIndexProperty, imageBlock.ZIndex);

            this.WriteShapeProperty(RtfTags.ShapeMarginLeftProperty, Unit.DipToEmuI(imageBlock.Margin.Left));
            this.WriteShapeProperty(RtfTags.ShapeMarginTopProperty, Unit.DipToEmuI(imageBlock.Margin.Top));
            this.WriteShapeProperty(RtfTags.ShapeMarginRightProperty, Unit.DipToEmuI(imageBlock.Margin.Right));
            this.WriteShapeProperty(RtfTags.ShapeMarginBottomProperty, Unit.DipToEmuI(imageBlock.Margin.Bottom));
        }
    }
}