﻿using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Utilities;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Fields;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Media;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Theming;
#else
using Telerik.Documents.Common.Model;
#endif
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Media;
#else
using System.Windows.Media;
using FontWeight = System.Windows.FontWeight;
using FontWeights = System.Windows.FontWeights;
using FontStyle = System.Windows.FontStyle;
using FontStyles = System.Windows.FontStyles;
#endif

namespace Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Export
{
    internal class InlineExporter
    {
        private readonly RtfDocumentExporter documentExporter;
        private readonly RtfWriter writer;

        public InlineExporter(RtfDocumentExporter exporter)
        {
            this.documentExporter = exporter;
            this.writer = exporter.Writer;
        }

        private ExportContext Context
        {
            get
            {
                return this.documentExporter.Context;
            }
        }

        public static void ExportCharacterProperties(CharacterProperties spanProps, ExportContext context, RtfWriter writer)
        {
            ExportFlowDirection(spanProps.FlowDirection.GetActualValue(), writer);

            DocumentTheme docTheme = spanProps.Document.Theme;

            ExportFontFamily(spanProps.FontFamily.GetActualValue(), context, docTheme, writer);

            ExportFontSize(spanProps.FontSize.GetActualValue(), writer);

            ExportFontStyle(spanProps.FontStyle.GetActualValue(), writer);

            ExportFontWeight(spanProps.FontWeight.GetActualValue(), writer);

            ExportStrikethrough(spanProps.Strikethrough.GetActualValue(), writer);

            ExportBaselineAlignment(spanProps.BaselineAlignment.GetActualValue(), writer);

            ExportColors(spanProps, docTheme, context, writer);

            // UnderlineDecoration
            writer.WriteTag(RtfHelper.UndelineDecorationToRtfTag(spanProps.UnderlinePattern.GetActualValue().Value));
        }

        public void ExportInline(InlineBase inline)
        {
            switch (inline.Type)
            {
                case DocumentElementType.Run:
                    this.ExportSpan((Run)inline);
                    break;
                case DocumentElementType.FieldCharacter:
                    this.ExportFieldCharacter((FieldCharacter)inline);
                    break;
                case DocumentElementType.ImageInline:
                    this.documentExporter.ImageExporter.ExportImage(((ImageInline)inline).Image);
                    break;
                case DocumentElementType.FloatingImage:
                    this.documentExporter.ImageExporter.ExportFloatingImage((FloatingImage)inline);
                    break;
                case DocumentElementType.BookmarkRangeStart:
                    this.ExportBookmarkRangeStart((BookmarkRangeStart)inline);
                    break;
                case DocumentElementType.BookmarkRangeEnd:
                    this.ExportBookmarkRangeEnd((BookmarkRangeEnd)inline);
                    break;
                case DocumentElementType.CommentRangeStart:
                    this.ExportCommentRangeStart((CommentRangeStart)inline);
                    break;
                case DocumentElementType.CommentRangeEnd:
                    this.ExportCommentRangeEnd((CommentRangeEnd)inline);
                    break;
                case DocumentElementType.Break:
                    this.ExportBreak((Break)inline);
                    break;
                default:
                    break;
            }
        }

        private static void ExportColors(CharacterProperties spanProps, DocumentTheme docTheme, ExportContext context, RtfWriter writer)
        {
            // ForegroundColor
            Color foregroundColor = spanProps.ForegroundColor.GetActualValue().GetActualValue(docTheme);
            if (foregroundColor != Run.ForegroundColorPropertyDefinition.DefaultValue.LocalValue && !RtfHelper.IsTransparentColor(foregroundColor))
            {
                writer.WriteTag(RtfTags.TagSpanForecolor, context.ColorTable[foregroundColor]);
            }

            // HighlightColor
            Color highlightColor = spanProps.HighlightColor.GetActualValue().Value;
            if (highlightColor != Run.HighlightColorPropertyDefinition.DefaultValue.Value && !RtfHelper.IsTransparentColor(highlightColor))
            {
                writer.WriteTag(RtfTags.TagSpanHighlight, context.ColorTable[highlightColor]);
                writer.WriteTag(RtfTags.TagSpanHighlightWord, context.ColorTable[highlightColor]);
            }

            // Shading
            Color shadingBack = spanProps.BackgroundColor.GetActualValue().GetActualValue(docTheme);
            Color shadingFore = spanProps.ShadingPatternColor.GetActualValue().GetActualValue(docTheme);
            if (!RtfHelper.IsTransparentColor(shadingBack) || !RtfHelper.IsTransparentColor(shadingFore))
            {
                if (!RtfHelper.IsTransparentColor(shadingBack))
                {
                    writer.WriteTag(RtfTags.TagSpanShadingBackground, context.ColorTable[shadingBack]);
                }

                if (!RtfHelper.IsTransparentColor(shadingFore))
                {
                    writer.WriteTag(RtfTags.TagSpanShadingForeground, context.ColorTable[shadingFore]);
                }

                writer.WriteTag(RtfTags.TagSpanShadingPercent, RtfHelper.ShadingPatternToRtfTag(spanProps.ShadingPattern.GetActualValue().Value));
            }

            // UnderlineColor
            ThemableColor themableUnderlineColor = spanProps.UnderlineColor.GetActualValue();
            Color actualUnderlineColor = themableUnderlineColor.GetActualValue(docTheme);
            if (!themableUnderlineColor.IsAutomatic && !RtfHelper.IsTransparentColor(actualUnderlineColor))
            {
                writer.WriteTag(RtfTags.TagUnderlineColor, context.ColorTable[actualUnderlineColor]);
            }
        }

        private static void ExportBaselineAlignment(BaselineAlignment? baseline, RtfWriter writer)
        {
            if (baseline != BaselineAlignment.Baseline)
            {
                if (baseline == BaselineAlignment.Subscript)
                {
                    writer.WriteTag(RtfTags.TagSubscript);
                }
                if (baseline == BaselineAlignment.Superscript)
                {
                    writer.WriteTag(RtfTags.TagSuperscript);
                }
            }
        }

        private static void ExportStrikethrough(bool? strikethrough, RtfWriter writer)
        {
            if (strikethrough.HasValue)
            {
                if (strikethrough.Value)
                {
                    writer.WriteTag(RtfTags.TagStrikethrough);
                }
                else
                {
                    writer.WriteTag(RtfTags.TagStrikethrough, 0);
                }
            }
        }

        private static void ExportFontWeight(FontWeight? fontWeight, RtfWriter writer)
        {
            if (fontWeight != FontWeights.Normal)
            {
                writer.WriteTag(RtfTags.TagBold);
            }
            else
            {
                writer.WriteTag(RtfTags.TagBold, 0);
            }
        }

        private static void ExportFontStyle(FontStyle? fontStyle, RtfWriter writer)
        {
            // FontStyle
            if (fontStyle != FontStyles.Normal)
            {
                writer.WriteTag(RtfTags.TagItalic);
            }
            else
            {
                writer.WriteTag(RtfTags.TagItalic, 0);
            }
        }

        private static void ExportFontSize(double? fontSize, RtfWriter writer)
        {
            if (fontSize.HasValue)
            {
                writer.WriteTag(RtfTags.TagFontSize, Unit.DipToPointI(fontSize.Value * 2));
            }
        }

        private static void ExportFontFamily(ThemableFontFamily actualFont, ExportContext context, DocumentTheme docTheme, RtfWriter writer)
        {
            if (actualFont != null)
            {
                int fontId = context.FontTable[actualFont.GetActualValue(docTheme)];
                if (fontId != 0)
                {
                    writer.WriteTag(RtfTags.TagFont, fontId);
                }
            }
        }

        private static void ExportFlowDirection(FlowDirection? flowDirection, RtfWriter writer)
        {
            // FlowDirection
            if (flowDirection == FlowDirection.LeftToRight)
            {
                writer.WriteTag(RtfTags.TagSpanLtr);
            }
            else
            {
                writer.WriteTag(RtfTags.TagSpanRtl);
            }
        }

        private void ExportFieldCharacter(FieldCharacter fldChar)
        {
            switch (fldChar.FieldCharacterType)
            {
                case FieldCharacterType.Start:
                    this.writer.WriteGroupStart(RtfTags.TagField, false);

                    if (fldChar.FieldInfo.IsLocked)
                    {
                        this.writer.WriteTag(RtfTags.TagFieldIsLocked);
                    }
                    if (fldChar.FieldInfo.IsDirty)
                    {
                        this.writer.WriteTag(RtfTags.TagFieldIsDirty);
                    }

                    this.writer.WriteGroup(RtfTags.TagFieldInstructions, true);
                    break;

                case FieldCharacterType.Separator:
                    this.writer.WriteGroupEnd(); // Instruction group
                    this.writer.WriteGroupStart(RtfTags.TagFieldResult, false);
                    break;

                case FieldCharacterType.End:

                    this.writer.WriteGroupEnd(); // Result or Instruction group
                    this.writer.WriteGroupEnd(); // Field group
                    break;
            }
        }

        private void ExportSpan(Run span)
        {
            string text = span.Text;
            if (string.IsNullOrEmpty(text))
            {
                return;
            }

            using (this.writer.WriteGroup())
            {
                this.ExportSpanProperties(span);

                this.writer.WriteText(text);
            }
        }

        private void ExportSpanProperties(Run span)
        {
            int styleId;
            if (!string.IsNullOrEmpty(span.Properties.StyleId) && this.Context.StyleTable.TryGetValue(span.Properties.StyleId, out styleId))
            {
                this.writer.WriteTag(RtfTags.TagStyleCharacter, styleId);
            }

            ExportCharacterProperties(span.Properties, this.Context, this.writer);
        }

        private void ExportBookmarkRangeStart(BookmarkRangeStart bookmarkRangeStart)
        {
            using (this.writer.WriteGroup(RtfTags.TagBookmarkStart, true))
            {
                this.writer.WriteText(bookmarkRangeStart.Bookmark.Name);
            }
        }

        private void ExportBookmarkRangeEnd(BookmarkRangeEnd bookmarkRangeEnd)
        {
            using (this.writer.WriteGroup(RtfTags.TagBookmarkEnd, true))
            {
                this.writer.WriteText(bookmarkRangeEnd.Bookmark.Name);
            }
        }

        private void ExportBreak(Break br)
        {
            switch (br.BreakType)
            {
                case BreakType.LineBreak:
                    using (this.writer.WriteGroup())
                    {
                        this.writer.WriteTag(RtfTags.TagLineBreakTextWrapping, RtfHelper.LineBreakTextWrappingMapper.GetFromValue(br.TextWrappingRestartLocation));
                        this.writer.WriteTag(RtfTags.TagLineBreak);
                        break;
                    }

                case BreakType.PageBreak:
                    this.writer.WriteTag(RtfTags.TagPageBreak);
                    break;

                case BreakType.ColumnBreak:
                    this.writer.WriteTag(RtfTags.TagColumnBreak);
                    break;
            }
        }

        private void ExportCommentRangeStart(CommentRangeStart commentRangeStart)
        {
            int id = this.Context.GetIdForComment(commentRangeStart.Comment);

            using (this.writer.WriteGroup(RtfTags.CommentStart, true))
            {
                this.writer.WriteText(id.ToString());
            }
        }

        private void ExportCommentRangeEnd(CommentRangeEnd commentRangeEnd)
        {
            int id = this.Context.GetIdForComment(commentRangeEnd.Comment);

            using (this.writer.WriteGroup(RtfTags.CommentEnd, true))
            {
                this.writer.WriteText(id.ToString());
            }

            this.documentExporter.CommentExporter.ExportComment(commentRangeEnd);
        }
    }
}