﻿using System;
using System.Collections.Generic;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Media;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Utilities;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;
using Telerik.Windows.Documents.Media;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Export
{
    internal class ParagraphExporter
    {
        private readonly RtfDocumentExporter documentExporter;
        private readonly RtfWriter writer;

        public ParagraphExporter(RtfDocumentExporter exporter)
        {
            this.documentExporter = exporter;
            this.writer = exporter.Writer;
        }

        private ExportContext Context
        {
            get
            {
                return this.documentExporter.Context;
            }
        }

        public static void ExportParagraphProperties(ParagraphProperties paragraphProperties, ExportContext context, RtfWriter writer, bool exportTabStops, bool exportIndentation = true, bool exportListId = false)
        {
            ExportParagraphSpacing(paragraphProperties, writer);

            Alignment textAlignment = paragraphProperties.TextAlignment.GetActualValue().Value;
            bool isRtl = paragraphProperties.FlowDirection.GetActualValue().Value == FlowDirection.RightToLeft;

            if (textAlignment != Alignment.Left || isRtl)
            {
                if (textAlignment == Alignment.LowKashida)
                {
                    writer.WriteTag(RtfTags.TagAlignKashida, 0);
                }
                else if (textAlignment == Alignment.MediumKashida)
                {
                    writer.WriteTag(RtfTags.TagAlignKashida, 10);
                }
                else if (textAlignment == Alignment.HighKashida)
                {
                    writer.WriteTag(RtfTags.TagAlignKashida, 20);
                }
                else
                {
                    writer.WriteTag(RtfHelper.AlignmentToRtfTag(textAlignment, isRtl));
                }
            }

            ParagraphBorders borders = paragraphProperties.Borders.GetActualValue();
            if (borders != null && borders != DocumentDefaultStyleSettings.ParagraphBorders)
            {
                BorderExporter.ExportParagraphBorders(borders, context, writer);
            }

            if (exportIndentation)
            {
                ParagraphExporter.ExportParagraphIndents(paragraphProperties, isRtl, writer);
            }

            var docTheme = paragraphProperties.Document.Theme;

            Color shadingBack = paragraphProperties.BackgroundColor.GetActualValue().GetActualValue(docTheme);
            Color shadingFore = paragraphProperties.ShadingPatternColor.GetActualValue().GetActualValue(docTheme);
            if (!RtfHelper.IsTransparentColor(shadingBack) || !RtfHelper.IsTransparentColor(shadingFore))
            {
                if (!RtfHelper.IsTransparentColor(shadingBack))
                {
                    writer.WriteTag(RtfTags.TagParagraphShadingBackground, context.ColorTable[shadingBack]);
                }

                if (!RtfHelper.IsTransparentColor(shadingFore))
                {
                    writer.WriteTag(RtfTags.TagParagraphShadingForeground, context.ColorTable[shadingFore]);
                }

                writer.WriteTag(RtfTags.TagParagraphShadingPercent, RtfHelper.ShadingPatternToRtfTag(paragraphProperties.ShadingPattern.GetActualValue().Value));
            }

            if (exportListId && paragraphProperties.ListId.HasLocalValue)
            {
                writer.WriteTag(RtfTags.TagListOverrideId, paragraphProperties.ListId.LocalValue.Value);
            }

            if (paragraphProperties.ListLevel.HasLocalValue && paragraphProperties.ListLevel.LocalValue != -1)
            {
                writer.WriteTag(RtfTags.TagParagraphListLevel, paragraphProperties.ListLevel.LocalValue.Value);
            }

            if (exportTabStops)
            {
                ExportTabStops(paragraphProperties.TabStops.GetActualValue(), writer);
            }
        }

        public void ExportParagraph(Paragraph paragraph)
        {
            this.ExportParagraphProperties(paragraph);

            foreach (InlineBase inline in paragraph.Inlines)
            {
                this.documentExporter.InlineExporter.ExportInline(inline);
            }

            // Don't export \par control for last paragraph in cell.
            TableCell parentCell = paragraph.Parent as TableCell;
            if (parentCell == null || (paragraph != parentCell.Blocks.Last()))
            {
                // We copy the character properties to the ParagraphMarkerProperties in order to sync them on export.
                if (!string.IsNullOrEmpty(paragraph.StyleId))
                {
                    GetParagraphMarkerPropertiesFromParagraphStyle(paragraph);
                }

                using (this.writer.WriteGroup())
                {
                    InlineExporter.ExportCharacterProperties(paragraph.Properties.ParagraphMarkerProperties, this.Context, this.writer);
                    this.writer.WriteTag(RtfTags.TagParagraphEnd);
                }
                this.writer.DebugWriteNewLine();
            }
        }

        private static void GetParagraphMarkerPropertiesFromParagraphStyle(Paragraph paragraph)
        {
            Style style = paragraph.Document.StyleRepository.GetStyle(paragraph.StyleId);
            if (style != null)
            {
                foreach (IStyleProperty property in style.CharacterProperties.StyleProperties)
                {
                    IStyleProperty paragraphMarkerProperty = paragraph.Properties.ParagraphMarkerProperties.GetStyleProperty(property.PropertyDefinition);
                    if (property.HasLocalValue && !paragraphMarkerProperty.HasLocalValue)
                    {
                        paragraphMarkerProperty.SetValueAsObject(property.GetLocalValueAsObject());
                    }
                }
            }
        }

        private static void ExportParagraphIndents(ParagraphProperties paragraphProperties, bool isRtl, RtfWriter writer)
        {
            double leftIndent = paragraphProperties.LeftIndent.GetActualValue().Value;
            double rightIndent = paragraphProperties.RightIndent.GetActualValue().Value;
            double firstLineIndent = paragraphProperties.FirstLineIndent.GetActualValue().Value;
            double hangingIndent = paragraphProperties.HangingIndent.GetActualValue().Value;

            // Swap indents for RTL
            if (isRtl)
            {
                double tmp = leftIndent;
                leftIndent = rightIndent;
                rightIndent = tmp;
            }

            if (hangingIndent > 0)
            {
                firstLineIndent = -hangingIndent;
            }

            int leftIndentTwip = Unit.DipToTwipI(leftIndent);
            int rightIndentTwip = Unit.DipToTwipI(rightIndent);

            if (firstLineIndent != 0)
            {
                writer.WriteTag(RtfTags.TagParagraphFirstLineIndent, Unit.DipToTwipI(firstLineIndent));
            }

            if (leftIndent != 0 || isRtl)
            {
                writer.WriteTag(RtfTags.TagParagraphLeftIndent, leftIndentTwip);
                writer.WriteTag(RtfTags.TagParagraphLeftIndentBiDirectional, leftIndentTwip);
            }

            if (rightIndent != 0 || isRtl)
            {
                writer.WriteTag(RtfTags.TagParagraphRightIndent, rightIndentTwip);
                writer.WriteTag(RtfTags.TagParagraphRightIndentBiDirectional, rightIndentTwip);
            }
        }

        private static void ExportTabStops(IEnumerable<TabStop> tabStops, RtfWriter writer)
        {
            foreach (TabStop tabStop in tabStops)
            {
                string leader = RtfHelper.TabStopLeaderMapper.GetFromValue(tabStop.Leader);
                if (!string.IsNullOrEmpty(leader))
                {
                    writer.WriteTag(leader);
                }

                if (tabStop.Type != TabStopType.Bar)
                {
                    // The default is start (left) and it shall not be exported.
                    string type = RtfHelper.TabStopTypeMapper.GetFromValue(tabStop.Type);
                    if (!string.IsNullOrEmpty(type))
                    {
                        writer.WriteTag(type);
                    }

                    writer.WriteTag(RtfTags.TagTabStopPosition, Unit.DipToTwipI(tabStop.Position));
                }
                else
                {
                    writer.WriteTag(RtfTags.TagTabStopBarPosition, Unit.DipToTwipI(tabStop.Position));
                }
            }
        }

        private static void ExportParagraphSpacing(ParagraphProperties paragraphProperties, RtfWriter writer)
        {
            // LineSpacing
            HeightType lineSpacingType = paragraphProperties.LineSpacingType.GetActualValue().Value;
            double lineSpacingValue = paragraphProperties.LineSpacing.GetActualValue().Value;
            switch (lineSpacingType)
            {
                case HeightType.Auto:
                    writer.WriteTag(RtfTags.TagParagraphLineSpacing, (int)Math.Round(lineSpacingValue * RtfHelper.DefaultLineSpacing));
                    writer.WriteTag(RtfTags.TagParagraphLineSpacingType, 1);
                    break;
                case HeightType.AtLeast:
                    writer.WriteTag(RtfTags.TagParagraphLineSpacing, Unit.DipToTwipI(lineSpacingValue));
                    writer.WriteTag(RtfTags.TagParagraphLineSpacingType, 0);
                    break;
                case HeightType.Exact:
                    writer.WriteTag(RtfTags.TagParagraphLineSpacing, -Unit.DipToTwipI(lineSpacingValue));
                    writer.WriteTag(RtfTags.TagParagraphLineSpacingType, 0);
                    break;
            }

            // 0 is the default value in RTF
            double spacingBefore = paragraphProperties.SpacingBefore.GetActualValue().Value;
            if (spacingBefore != 0.0)
            {
                writer.WriteTag(RtfTags.TagParagraphSpacingBefore, Unit.DipToTwipI(spacingBefore));
            }

            if (paragraphProperties.AutomaticSpacingBefore.GetActualValue().Value)
            {
                writer.WriteTag(RtfTags.TagParagraphAutomaticSpacingBefore, 1);
            }

            // 0 is the default value in RTF
            double spacingAfter = paragraphProperties.SpacingAfter.GetActualValue().Value;
            if (spacingAfter != 0.0)
            {
                writer.WriteTag(RtfTags.TagParagraphSpacingAfter, Unit.DipToTwipI(spacingAfter));
            }

            if (paragraphProperties.AutomaticSpacingAfter.GetActualValue().Value)
            {
                writer.WriteTag(RtfTags.TagParagraphAutomaticSpacingAfter, 1);
            }
        }

        // TODO: Remove this method when it is no longer needed 
        private static int GetParagraphNestingLevel(Paragraph paragraph)
        {
            int level = RtfHelper.GetNestingLevel(paragraph);
            return (level - 2) / 3;
        }

        private void ExportParagraphProperties(Paragraph paragraph)
        {
            this.writer.WriteTag(RtfTags.TagParagraphDefaults);

            if (this.Context.TableStylesStack.Peek().HasValue)
            {
                this.writer.WriteTag(RtfTags.TagStyleTableForParagraph, this.Context.TableStylesStack.Peek().Value);
            }

            int styleId;
            if (!string.IsNullOrEmpty(paragraph.StyleId) && this.Context.StyleTable.TryGetValue(paragraph.StyleId, out styleId))
            {
                this.writer.WriteTag(RtfTags.TagStyleParagraph, styleId);
            }

            if (paragraph.Properties.FlowDirection.GetActualValue().Value == FlowDirection.LeftToRight)
            {
                this.writer.WriteTag(RtfTags.TagParagraphLtr);
            }
            else
            {
                this.writer.WriteTag(RtfTags.TagParagraphRtl);
            }

            bool isInTable = (paragraph.Parent != null) && (paragraph.Parent is TableCell);
            if (isInTable)
            {
                this.writer.WriteTag(RtfTags.TagParagraphInTable);

                int nestingLevel = GetParagraphNestingLevel(paragraph);
                this.writer.WriteTag(RtfTags.TagParagraphNestingLevel, nestingLevel);
            }

            ExportTabStops(paragraph.TabStops, this.writer);

            bool exportIndentation = true;

            if (paragraph.ListId != -1)
            {
                int id = paragraph.ListId;

                // If a paragraph is pointing to List which is a numbering style definition.
                if (this.Context.StyleNumberingListIdsToActualListIds.ContainsKey(paragraph.ListId))
                {
                    id = this.Context.StyleNumberingListIdsToActualListIds[paragraph.ListId];
                }
                this.writer.WriteTag(RtfTags.TagListOverrideId, id);
                if (paragraph.ListLevel != -1)
                {
                    this.writer.WriteTag(RtfTags.TagParagraphListLevel, paragraph.ListLevel);
                }
                List list = paragraph.Document.Lists.GetList(paragraph.ListId);

                ListLevel listLevel;
                if (list.MultilevelType == MultilevelType.SingleLevel || paragraph.ListLevel == -1)
                {
                    listLevel = list.Levels[0];
                }
                else
                {
                    listLevel = list.Levels[paragraph.ListLevel];
                }

                string listLevelTextFormat = listLevel.NumberTextFormat;
                if (string.IsNullOrEmpty(listLevelTextFormat))
                {
                    // If a listlevel hasn't a number text format, the text should be set to '\tab'.
                    listLevelTextFormat = "\t";
                }

                using (this.writer.WriteGroup(RtfTags.TagListText))
                {
                    if (listLevel.NumberingStyle != NumberingStyle.Bullet)
                    {
                        listLevelTextFormat = RtfHelper.ConvertNumberedListFormatToRtfLevelText(listLevelTextFormat);
                    }
                    this.writer.WriteText(listLevelTextFormat);
                }
                ExportParagraphIndents(listLevel.ParagraphProperties, false, this.writer);

                exportIndentation = false;
            }

            ExportParagraphProperties(paragraph.Properties, this.Context, this.writer, false, exportIndentation);
        }
    }
}