﻿using System.Collections.Generic;
using System.IO;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Primitives;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Utilities;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Watermarks;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Model;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Export
{
    internal class RtfDocumentExporter
    {
        private readonly RadFlowDocument document;

        public RtfDocumentExporter(RadFlowDocument document, Stream output, RtfExportSettings settings)
        {
            Guard.ThrowExceptionIfNull(document, "document");
            Guard.ThrowExceptionIfNull(settings, "settings");

            this.Settings = settings;
            this.document = document;
            this.Writer = new RtfWriter(output);

            this.CommentExporter = new CommentExporter(this);
            this.ImageExporter = new ImageExporter(this);
            this.ParagraphExporter = new ParagraphExporter(this);
            this.InlineExporter = new InlineExporter(this);
        }

        public ExportContext Context { get; private set; }

        public RtfWriter Writer { get; private set; }

        public RtfExportSettings Settings { get; private set; }

        public ImageExporter ImageExporter { get; private set; }

        public CommentExporter CommentExporter { get; private set; }

        public ParagraphExporter ParagraphExporter { get; private set; }

        public InlineExporter InlineExporter { get; private set; }

        public void Export()
        {
            this.Context = new ExportContext(this.document);
            this.Context.InitializeContext();

            this.ExportDocument();
        }

        public void ExportBlock(BlockBase block)
        {
            if (block is Paragraph)
            {
                this.ParagraphExporter.ExportParagraph((Paragraph)block);
            }
            else if (block is Table)
            {
                this.ExportTable((Table)block);
            }
        }

        private void ExportDocument()
        {
            using (this.Writer.WriteGroup(RtfTags.TagRtf))
            {
                HeaderExporter headerExporter = new HeaderExporter(this.Context, this.Writer);
                headerExporter.ExportHeader();
                this.ExportDocumentProperties(this.document);

                foreach (Section section in this.document.Sections)
                {
                    this.ExportSection(section);
                }
            }

            this.Writer.Flush();
        }

        private void ExportDocumentProperties(RadFlowDocument doc)
        {
            this.Writer.WriteTag(RtfTags.TagNoUICompatibilityTag);
            this.Writer.WriteTag(RtfTags.TagLayoutMode, RtfHelper.DocumentViewTypeMapper.GetFromValue(doc.ViewType));
            if (doc.HasDifferentEvenOddPageHeadersFooters)
            {
                this.Writer.WriteTag(RtfTags.TagDifferentOddEvenHeaderFooter);
            }

            this.Writer.WriteTag(RtfTags.TagDocumentTabDefaultWidth, Unit.DipToTwipI(doc.DefaultTabStopWidth));
        }

        private void ExportSection(Section section)
        {
            bool isFirst = section == section.Parent.Children.First();
            if (!isFirst)
            {
                this.Writer.WriteTag(RtfTags.TagSectionEnd);
            }

            this.ExportSectionProperties(section);

            if (isFirst)
            {
                this.ExportDocumentVariables();
            }

            foreach (BlockBase block in section.Blocks)
            {
                this.ExportBlock(block);
            }
        }

        private void ExportSectionProperties(Section section)
        {
            this.Writer.WriteTag(RtfTags.TagSectionDefaults);
            if (section != section.Document.Sections.First())
            {
                string tag = RtfHelper.SectionTypeMapper.GetFromValue(section.SectionType);
                this.Writer.WriteTag(tag);
            }

#pragma warning disable 618
            if (section.PageOrientation == PageOrientation.Landscape ||
                section.PageOrientation == PageOrientation.Rotate270)
#pragma warning restore 618
            {
                this.Writer.WriteTag(RtfTags.TagSectionLandscapeMode);
            }

            Size pageSize = section.PageSize;
            this.Writer.WriteTag(RtfTags.TagSectionWidth, Unit.DipToTwipI(pageSize.Width));
            this.Writer.WriteTag(RtfTags.TagSectionHeight, Unit.DipToTwipI(pageSize.Height));

            Padding pageMargins = section.PageMargins;
            this.Writer.WriteTag(RtfTags.TagSectionMarginLeft, Unit.DipToTwipI(pageMargins.Left));
            this.Writer.WriteTag(RtfTags.TagSectionMarginRight, Unit.DipToTwipI(pageMargins.Right));
            this.Writer.WriteTag(RtfTags.TagSectionMarginTop, Unit.DipToTwipI(pageMargins.Top));
            this.Writer.WriteTag(RtfTags.TagSectionMarginBottom, Unit.DipToTwipI(pageMargins.Bottom));

            if (section.HasDifferentFirstPageHeaderFooter)
            {
                this.Writer.WriteTag(RtfTags.TagDifferentFirstHeaderFooter);
            }

            switch (section.VerticalAlignment)
            {
                case Model.Styles.VerticalAlignment.Top:
                    this.Writer.WriteTag(RtfTags.TagSectionVerticalAlignmentTop);
                    break;
                case Model.Styles.VerticalAlignment.Bottom:
                    this.Writer.WriteTag(RtfTags.TagSectionVerticalAlignmentBottom);
                    break;
                case Model.Styles.VerticalAlignment.Center:
                    this.Writer.WriteTag(RtfTags.TagSectionVerticalAlignmentCenter);
                    break;
                case Model.Styles.VerticalAlignment.Justified:
                    this.Writer.WriteTag(RtfTags.TagSectionVerticalAlignmentJustified);
                    break;
            }

            PageNumberingSettings pageNumberingSettings = section.PageNumberingSettings;
            if (pageNumberingSettings.StartingPageNumber.HasValue)
            {
                this.Writer.WriteTag(RtfTags.TagSectionRestartPageNumberingFromStartingPageNumber);
                this.Writer.WriteTag(RtfTags.TagSectionStartingPageNumber, section.PageNumberingSettings.StartingPageNumber.Value);
            }

            if (pageNumberingSettings.PageNumberFormat.HasValue)
            {
                string numberFormatTag;
                if (RtfHelper.NumberingStyleMapper.TryGetFromValue(pageNumberingSettings.PageNumberFormat.Value, out numberFormatTag))
                {
                    this.Writer.WriteTag(numberFormatTag);
                }
            }

            if (pageNumberingSettings.ChapterSeparatorCharacter.HasValue)
            {
                string chapterSeparatorTag;
                if (RtfHelper.ChapterSeparatorTypeMapper.TryGetFromValue(pageNumberingSettings.ChapterSeparatorCharacter.Value, out chapterSeparatorTag))
                {
                    this.Writer.WriteTag(chapterSeparatorTag);
                }
            }

            this.Writer.WriteTag(RtfTags.TagHeaderTopMargin, Unit.DipToTwipI(section.HeaderTopMargin));
            this.Writer.WriteTag(RtfTags.TagFooterBottomMargin, Unit.DipToTwipI(section.FooterBottomMargin));

            this.ExportHeaderFooterGroup(RtfTags.TagHeaderFirstPage, section.Headers.First);
            this.ExportHeaderFooterGroup(RtfTags.TagHeaderRightPages, section.Headers.Default);
            this.ExportHeaderFooterGroup(RtfTags.TagHeaderLeftPages, section.Headers.Even);

            this.ExportHeaderFooterGroup(RtfTags.TagFooterFirstPage, section.Footers.First);
            this.ExportHeaderFooterGroup(RtfTags.TagFooterRightPages, section.Footers.Default);
            this.ExportHeaderFooterGroup(RtfTags.TagFooterLeftPages, section.Footers.Even);
        }

        private void ExportHeaderFooterGroup(string tagName, HeaderFooterBase headerFooterBase)
        {
            if (headerFooterBase != null)
            {
                using (this.Writer.WriteGroup(tagName))
                {
                    Header header = headerFooterBase as Header;
                    if (header != null)
                    {
                        foreach (Watermark watermark in header.Watermarks)
                        {
                            WatermarkExporter exporter = new WatermarkExporter(this, watermark);
                            exporter.ExportWatermark();
                        }
                    }

                    foreach (BlockBase block in headerFooterBase.Blocks)
                    {
                        this.ExportBlock(block);
                    }
                }
            }
        }

        private void ExportDocumentVariables()
        {
            if (this.document.DocumentVariables.Count == 0)
            {
                return;
            }

            foreach (KeyValuePair<string, string> item in this.document.DocumentVariables)
            {
                using (this.Writer.WriteGroup(RtfTags.TagDocumentVariable, true))
                {
                    using (this.Writer.WriteGroup())
                    {
                        this.Writer.WriteText(item.Key);
                    }
                    using (this.Writer.WriteGroup())
                    {
                        this.Writer.WriteText(item.Value);
                    }
                }
            }
        }

        private void ExportTable(Table table)
        {
            TableExporter tableExporter = new TableExporter(this, table);
            tableExporter.ExportTable();
        }
    }
}