﻿using System;
using System.Diagnostics;
using System.IO;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Exceptions;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Export
{
    internal class RtfWriter : IDisposable
    {
        private static readonly string HexChars = "0123456789abcdef";
        private static readonly string UnicodeSubstituteChar = "?";

        private readonly Stream writer;
        private readonly StreamWriter textWriter;
        private int groupLevel;

        private bool shouldAddSpace;

        public RtfWriter(Stream o)
        {
            this.writer = o;
            this.textWriter = new StreamWriter(this.writer);
        }

        public void WriteTag(string tagName)
        {
            Debug.Assert(!string.IsNullOrEmpty(tagName), "Tag is empty");

            this.InnerWrite("\\");
            this.InnerWrite(tagName);
            this.shouldAddSpace = true;
        }

        public void WriteTag(string tagName, int parameter)
        {
            this.WriteTag(tagName);
            this.InnerWrite(parameter.ToString());
            this.shouldAddSpace = true;
        }

        public void WriteGroupStart()
        {
            this.InnerWrite("{");
            this.groupLevel++;
            this.shouldAddSpace = false;
        }

        public void WriteGroupStart(string tagName, bool isExtensionGroup)
        {
            this.WriteGroupStart();

            if (isExtensionGroup)
            {
                this.InnerWrite("\\*");
            }

            this.WriteTag(tagName);
        }

        public void WriteGroupStart(string tagName, int parameter, bool isExtensionGroup)
        {
            this.WriteGroupStart();

            if (isExtensionGroup)
            {
                this.InnerWrite("\\*");
            }

            this.WriteTag(tagName, parameter);
        }

        public RtfGroupContext WriteGroup()
        {
            this.WriteGroupStart();

            return new RtfGroupContext(this);
        }

        public RtfGroupContext WriteGroup(string tagName, bool isExtensionGroup = false)
        {
            this.WriteGroupStart(tagName, isExtensionGroup);
            return new RtfGroupContext(this);
        }

        public RtfGroupContext WriteGroup(string tagName, int parameter, bool isExtensionGroup = false)
        {
            this.WriteGroupStart(tagName, parameter, isExtensionGroup);
            return new RtfGroupContext(this);
        }

        public void WriteGroupEnd()
        {
            this.groupLevel--;

            if (this.groupLevel < 0)
            {
                throw new RtfException("Group level reached negative value");
            }
            this.InnerWrite("}");
            this.shouldAddSpace = false;
        }

        public void WriteText(string text)
        {
            if (text == null || text.Length == 0)
            {
                return;
            }

            if (this.shouldAddSpace)
            {
                this.InnerWriteChar(' ');
                this.shouldAddSpace = false;
            }

            for (int i = 0; i < text.Length; i++)
            {
                char c = text[i];
                this.WriteChar(c);
            }
        }

        public void WriteChar(char c)
        {
            ////string cStr = c.ToString();
            if (c >= 0 && c <= 127)
            {
                switch (c)
                {
                    case '\t':
                        this.WriteTag("tab");
                        this.InnerWriteChar(' ');
                        break;
                    case '\n':
                        this.WriteTag("tab");
                        this.InnerWriteChar(' ');
                        break;
                    case '\\':
                    case '{':
                    case '}':
                        this.InnerWriteChar('\\');
                        this.InnerWriteChar(c);
                        break;
                    default:
                        this.InnerWriteChar(c);
                        break;
                }
            }
            else
            {
                this.WriteUnicodeChar(c);
            }
        }

        public void WriteHex(int number)
        {
            this.InnerWrite(string.Format(@"\'{0:x2}", number));
        }

        public void WriteHex(byte[] bytes)
        {
            if (this.shouldAddSpace)
            {
                this.InnerWriteChar(' ');
                this.shouldAddSpace = false;
            }

            for (int i = 0, l = bytes.Length; i < l; ++i)
            {
                this.textWriter.Write(HexChars[bytes[i] >> 4]);
                this.textWriter.Write(HexChars[bytes[i] & 15]);
            }
        }

        [System.Diagnostics.Conditional("DEBUG")]
        public void DebugWriteNewLine()
        {
            this.InnerWrite(Environment.NewLine);
        }

        public void Flush()
        {
            this.textWriter.Flush();
            this.writer.Flush();
        }

        public void Dispose()
        {
            this.textWriter.Dispose();
            this.writer.Dispose();
        }

        private void InnerWrite(string text)
        {
            this.textWriter.Write(text);
        }

        private void InnerWriteChar(char c)
        {
            this.textWriter.Write(c);
        }

        private void WriteUnicodeChar(char c)
        {
            short code = (short)c;
            this.textWriter.Write(string.Format("\\u{0}{1}", code, UnicodeSubstituteChar));
        }
    }
}