﻿using System;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Utilities;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Flow.Model.Watermarks;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Model.Drawing.Shapes;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Export
{
    internal class WatermarkExporter
    {
        private readonly RtfDocumentExporter rtfExporter;
        private readonly Watermark watermark;
        private readonly RtfWriter writer;
        private readonly ImageExporter imageExporter;
        private readonly Random rand = new Random();

        public WatermarkExporter(RtfDocumentExporter rtfDocumentExporter, Watermark watermark)
        {
            Guard.ThrowExceptionIfNull(watermark, "watermark");

            this.rtfExporter = rtfDocumentExporter;
            this.watermark = watermark;
            this.writer = this.rtfExporter.Writer;
            this.imageExporter = this.rtfExporter.ImageExporter;
        }

        public void ExportWatermark()
        {
            switch (this.watermark.WatermarkType)
            {
                case WatermarkType.Text:
                    this.ExportTextWatermark(this.watermark.TextSettings);
                    break;
                case WatermarkType.Image:
                    this.ExportImageWatermark(this.watermark.ImageSettings);
                    break;
                default:
                    throw new NotSupportedException("Watermark type is not supported.");
            }
        }

        private void ExportTextWatermark(TextWatermarkSettings textSettings)
        {
            using (this.writer.WriteGroup(RtfTags.TagShape, false))
            {
                using (this.writer.WriteGroup(RtfTags.TagShapeInstructions, true))
                {
                    this.ExportTextShapeInstructions(textSettings);
                }
            }
        }

        private void ExportImageWatermark(ImageWatermarkSettings imageSettings)
        {
            if (imageSettings.ImageSource == null)
            {
                return;
            }

            using (this.writer.WriteGroup(RtfTags.TagShape, false))
            {
                Image image = new Image();
                image.Size = new Size(imageSettings.Width, imageSettings.Height);
                image.ImageSource = imageSettings.ImageSource;
                image.RotationAngle = imageSettings.Angle;

                using (this.writer.WriteGroup(RtfTags.TagShapeInstructions, true))
                {
                    this.ExportImageShapeInstructions(image);
                }

                using (this.writer.WriteGroup(RtfTags.TagShapeResultDestination, false))
                {
                    this.writer.WriteTag(RtfTags.TagParagraphEnd);
                    this.writer.WriteTag(RtfTags.TagParagraphDefaults);

                    this.imageExporter.ExportImage(image);
                    this.writer.WriteTag(RtfTags.TagParagraphEnd);
                }
            }
        }

        private void ExportImageShapeInstructions(Image image)
        {
            this.writer.DebugWriteNewLine();
            Size imagesize = RtfHelper.CalculateImageSizeForRotateAngle(image.Size, image.RotationAngle);

            this.writer.WriteTag(RtfTags.TagShapeLeft, 0);
            this.writer.WriteTag(RtfTags.TagShapeTop, 0);
            this.writer.WriteTag(RtfTags.TagShapeRight, Unit.DipToTwipI(imagesize.Width));
            this.writer.WriteTag(RtfTags.TagShapeBottom, Unit.DipToTwipI(imagesize.Height));

            this.writer.WriteTag(RtfTags.TagShapeIsInHeader, 1);

            this.writer.WriteTag(RtfTags.TagShapeHorizontalRelativeToColumn);
            this.writer.WriteTag(RtfTags.TagShapeHorizontalRelativeToIgnore);

            this.writer.WriteTag(RtfTags.TagShapeVerticalRelativeToParagraph);
            this.writer.WriteTag(RtfTags.TagShapeVerticalRelativeToIgnore);

            this.writer.WriteTag(RtfTags.TagShapeTextWrap, RtfHelper.TextWrapMapper.GetFromValue(TextWrap.BothSides));

            this.writer.WriteTag(RtfTags.TagShapeIsBelowText, 1);
            this.writer.WriteTag(RtfTags.TagShapeZOrder, 0);

            // Exporting the image behind the text.
            this.writer.WriteTag(RtfTags.TagShapeWrappingType, 3);

            this.imageExporter.WriteShapeProperty(RtfTags.ShapeTypeProperty, RtfHelper.ImageShapeType);

            if (image.RotationAngle != 0)
            {
                this.imageExporter.WriteShapeProperty(RtfTags.ShapeRotationProperty, RtfHelper.AngleToRtfValue(image.RotationAngle));
            }
            this.imageExporter.WriteImageDataProperty(image);

            this.imageExporter.WriteShapeProperty(RtfTags.ShapeHorizontalPositionProperty, RtfHelper.RelativeHorizontalAlignmentMapper.GetFromValue(RelativeHorizontalAlignment.Center));
            this.imageExporter.WriteShapeProperty(RtfTags.ShapeHorizontalRelativeToPositionProperty, RtfHelper.HorizontalRelativeFromMapper.GetFromValue(HorizontalRelativeFrom.Margin));

            this.imageExporter.WriteShapeProperty(RtfTags.ShapeVerticalPositionProperty, RtfHelper.RelativeVerticalAlignmentMapper.GetFromValue(RelativeVerticalAlignment.Center));
            this.imageExporter.WriteShapeProperty(RtfTags.ShapeVerticalRelativeToPositionProperty, RtfHelper.VerticalRelativeFromMapper.GetFromValue(VerticalRelativeFrom.Margin));

            this.imageExporter.WriteShapeProperty(RtfTags.ShapeBehindDocumentProperty, 1);
            this.imageExporter.WriteShapeProperty(RtfTags.ShapeLayoutInCellProperty, 0);

            this.imageExporter.WriteShapeProperty(RtfTags.ShapeNameProperty, RtfTags.ImageWatermarkShapeName + this.rand.Next(short.MaxValue));
        }

        private void ExportTextShapeInstructions(TextWatermarkSettings textSettings)
        {
            this.writer.DebugWriteNewLine();

            Size textSize = new Size(textSettings.Width, textSettings.Height);
            textSize = RtfHelper.CalculateImageSizeForRotateAngle(textSize, textSettings.Angle);

            this.writer.WriteTag(RtfTags.TagShapeLeft, 0);
            this.writer.WriteTag(RtfTags.TagShapeTop, 0);
            this.writer.WriteTag(RtfTags.TagShapeRight, Unit.DipToTwipI(textSize.Width));
            this.writer.WriteTag(RtfTags.TagShapeBottom, Unit.DipToTwipI(textSize.Height));

            this.writer.WriteTag(RtfTags.TagShapeIsInHeader, 1);

            this.writer.WriteTag(RtfTags.TagShapeHorizontalRelativeToMargin);
            this.writer.WriteTag(RtfTags.TagShapeHorizontalRelativeToIgnore);

            this.writer.WriteTag(RtfTags.TagShapeVerticalRelativeToMargin);
            this.writer.WriteTag(RtfTags.TagShapeVerticalRelativeToIgnore);

            this.writer.WriteTag(RtfTags.TagShapeTextWrap, RtfHelper.TextWrapMapper.GetFromValue(TextWrap.BothSides));

            this.writer.WriteTag(RtfTags.TagShapeIsBelowText, 1);
            this.writer.WriteTag(RtfTags.TagShapeZOrder, 0);

            // Exporting the image behind the text.
            this.writer.WriteTag(RtfTags.TagShapeWrappingType, 3);

            this.imageExporter.WriteShapeProperty(RtfTags.ShapeTypeProperty, RtfHelper.TextShapeType);

            if (textSettings.Angle != 0)
            {
                this.imageExporter.WriteShapeProperty(RtfTags.ShapeRotationProperty, RtfHelper.AngleToRtfValue(textSettings.Angle));
            }
            this.imageExporter.WriteShapeProperty(RtfTags.ShapeTextProperty, textSettings.Text);
            this.imageExporter.WriteShapeProperty(RtfTags.ShapeTextFontProperty, textSettings.FontFamily.Source);

            this.imageExporter.WriteShapeProperty(RtfTags.ShapeFillColorProperty, RtfHelper.ColorToInt(textSettings.ForegroundColor));
            int opacity = (int)RtfHelper.DoubleToRtfFixedPointSize(textSettings.Opacity);
            this.imageExporter.WriteShapeProperty(RtfTags.ShapeFillOpacityProperty, opacity);

            this.imageExporter.WriteShapeProperty(RtfTags.ShapeHorizontalPositionProperty, RtfHelper.RelativeHorizontalAlignmentMapper.GetFromValue(RelativeHorizontalAlignment.Center));
            this.imageExporter.WriteShapeProperty(RtfTags.ShapeHorizontalRelativeToPositionProperty, RtfHelper.HorizontalRelativeFromMapper.GetFromValue(HorizontalRelativeFrom.Margin));

            this.imageExporter.WriteShapeProperty(RtfTags.ShapeVerticalPositionProperty, RtfHelper.RelativeVerticalAlignmentMapper.GetFromValue(RelativeVerticalAlignment.Center));
            this.imageExporter.WriteShapeProperty(RtfTags.ShapeVerticalRelativeToPositionProperty, RtfHelper.VerticalRelativeFromMapper.GetFromValue(VerticalRelativeFrom.Margin));

            this.imageExporter.WriteShapeProperty(RtfTags.ShapeBehindDocumentProperty, 1);
            this.imageExporter.WriteShapeProperty(RtfTags.ShapeLayoutInCellProperty, 0);
            this.imageExporter.WriteShapeProperty(RtfTags.ShapeLineProperty, 0);

            this.imageExporter.WriteShapeProperty(RtfTags.ShapeNameProperty, RtfTags.TextWatermarkShapeName + this.rand.Next(short.MaxValue));
        }
    }
}
