using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Text;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Exceptions;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.Model;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.Parser
{
    internal sealed class RtfParser : RtfParserBase, IDisposable
    {
        private const int EndOfFile = -1;
        private readonly Stack<int> unicodeSkipCountStack = new Stack<int>();
        private readonly Stack<int> codePageStack = new Stack<int>();
        private readonly Dictionary<string, int> fontToCodePageMapping = new Dictionary<string, int>();
        private readonly MemoryStream hexDecodingBuffer = new MemoryStream();

        private StringBuilder curText;
        private int unicodeSkipCount;
        private int level;
        private int tagCountAtLastGroupStart;
        private int tagCount;
        private int fontTableStartLevel;
        private string targetFont;
        private bool expectingThemeFont;

        public RtfParser()
        {
        }

        public RtfParser(params RtfParserListenerBase[] listeners) : base(listeners)
        {
        }

        public void Dispose()
        {
            this.hexDecodingBuffer.Dispose();
        }

        protected override void DoParse(RtfSource rtfTextSource)
        {
            this.NotifyParseBegin();
            try
            {
                this.ParseRtf(rtfTextSource.Reader);
                this.NotifyParseSuccess();
            }
            catch (RtfException e)
            {
                this.NotifyParseFail(e);
                throw;
            }
            finally
            {
                this.NotifyParseEnd();
            }
        }

        private static bool IsASCIILetter(int character)
        {
            return (character >= 'a' && character <= 'z') || (character >= 'A' && character <= 'Z');
        }

        private static bool IsHexDigit(int character)
        {
            return (character >= '0' && character <= '9') ||
                   (character >= 'a' && character <= 'f') ||
                   (character >= 'A' && character <= 'F');
        }

        private static bool IsDigit(int character)
        {
            return character >= '0' && character <= '9';
        }

        private static int ReadOneByte(RtfReader reader)
        {
            int byteValue = reader.Read();
            if (byteValue == -1)
            {
                throw new RtfUnicodeEncodingException("UnexpectedEndOfFile");
            }
            return byteValue;
        }

        private static int PeekNextChar(RtfReader reader, bool mandatory)
        {
            int character = reader.Peek();
            if (mandatory && character == -1)
            {
                throw new RtfMultiByteEncodingException("EndOfFileInvalidCharacter");
            }
            return character;
        }

        private static void UpdateEncoding(RtfReader reader, RtfTag tag)
        {
            switch (tag.Name)
            {
                case RtfTags.TagEncodingAnsi:
                    UpdateEncoding(reader, CharSetHelper.AnsiCodePage);
                    break;
                case RtfTags.TagEncodingMac:
                    UpdateEncoding(reader, 10000);
                    break;
                case RtfTags.TagEncodingPc:
                    UpdateEncoding(reader, 437);
                    break;
                case RtfTags.TagEncodingPca:
                    UpdateEncoding(reader, 850);
                    break;
                case RtfTags.TagEncodingAnsiCodePage:
                    UpdateEncoding(reader, tag.ValueAsNumber);
                    break;
            }
        }

        private static void UpdateEncoding(RtfReader reader, int codePage)
        {
            reader.CodePageDecoder.CurrentCodePage = codePage;
        }

        private static char ReadOneChar(RtfReader reader)
        {
            return (char)reader.Read();
        }

        private void ParseRtf(RtfReader reader)
        {
            this.curText = new StringBuilder();

            this.unicodeSkipCountStack.Clear();
            this.codePageStack.Clear();
            this.unicodeSkipCount = 1;
            this.level = 0;
            this.tagCountAtLastGroupStart = 0;
            this.tagCount = 0;
            this.fontTableStartLevel = -1;
            this.targetFont = null;
            this.expectingThemeFont = false;
            this.fontToCodePageMapping.Clear();
            this.hexDecodingBuffer.SetLength(0);
            int groupCount = 0;
            int nextChar = PeekNextChar(reader, false);
            bool backslashAlreadyConsumed = false;
            while (nextChar != RtfParser.EndOfFile)
            {
                int peekChar = 0;
                bool peekCharValid = false;
                switch (nextChar)
                {
                    case '\\':
                        if (!backslashAlreadyConsumed)
                        {
                            reader.Read(); // must still consume the 'peek'ed char
                        }
                        int secondChar = PeekNextChar(reader, true);
                        switch (secondChar)
                        {
                            case '\\':
                            case '{':
                            case '}':
                                this.curText.Append(ReadOneChar(reader)); // must still consume the 'peek'ed char
                                break;
                            case '\n':
                            case '\r':
                                reader.Read(); // must still consume the 'peek'ed char
                                // must be treated as a 'par' tag if preceded by a backslash
                                // (see RTF spec page 144)
                                this.HandleTag(reader, new RtfTag(RtfTags.TagParagraphEnd));
                                break;
                            case '\'':
                                reader.Read(); // must still consume the 'peek'ed char
                                char hex1 = (char)ReadOneByte(reader);
                                char hex2 = (char)ReadOneByte(reader);
                                if (!IsHexDigit(hex1))
                                {
                                    throw new RtfHexEncodingException(string.Format("InvalidFirstHexDigit: {0}", hex1));
                                }
                                if (!IsHexDigit(hex2))
                                {
                                    throw new RtfHexEncodingException(string.Format("InvalidFirstHexDigit: {0}", hex2));
                                }
                                int decodedByte = int.Parse(string.Format("{0}{1}{2}", string.Empty, hex1, hex2), NumberStyles.HexNumber);
                                this.hexDecodingBuffer.WriteByte((byte)decodedByte);
                                peekChar = PeekNextChar(reader, false);
                                peekCharValid = true;
                                bool mustFlushHexContent = true;
                                if (peekChar == '\\')
                                {
                                    reader.Read();
                                    backslashAlreadyConsumed = true;
                                    int continuationChar = PeekNextChar(reader, false);
                                    if (continuationChar == '\'')
                                    {
                                        mustFlushHexContent = false;
                                    }
                                }
                                if (mustFlushHexContent)
                                {
                                    // we may _NOT_ handle hex content in a character-by-character way as
                                    // this results in invalid text for japanese/chinese content ...
                                    // -> we wait until the following content is non-hex and then flush the
                                    //    pending data. ugly but necessary with our decoding model.
                                    this.DecodeCurrentHexBuffer(reader);
                                }
                                break;
                            case '|':
                            case '~':
                            case '-':
                            case '_':
                            case ':':
                            case '*':
                                this.HandleTag(reader, new RtfTag(string.Format("{0}{1}", string.Empty, ReadOneChar(reader)))); // must still consume the 'peek'ed char
                                break;
                            default:
                                this.ParseTag(reader);
                                break;
                        }
                        break;
                    case '\n':
                    case '\r':
                        reader.Read(); // must still consume the 'peek'ed char
                        break;
                    case '\t':
                        reader.Read(); // must still consume the 'peek'ed char
                        // should be treated as a 'tab' tag (see RTF spec page 144)
                        this.HandleTag(reader, new RtfTag(RtfTags.TagTabulator));
                        break;
                    case '{':
                        reader.Read(); // must still consume the 'peek'ed char
                        this.FlushText();
                        this.NotifyGroupBegin();
                        this.tagCountAtLastGroupStart = this.tagCount;
                        this.unicodeSkipCountStack.Push(this.unicodeSkipCount);

                        this.codePageStack.Push(reader.CodePageDecoder.CurrentCodePage);
                        this.level++;
                        break;
                    case '}':
                        reader.Read(); // must still consume the 'peek'ed char
                        this.FlushText();
                        if (this.level > 0)
                        {
                            this.unicodeSkipCount = this.unicodeSkipCountStack.Pop();
                            if (this.fontTableStartLevel == this.level)
                            {
                                this.fontTableStartLevel = -1;
                                this.targetFont = null;
                                this.expectingThemeFont = false;
                            }
                            UpdateEncoding(reader, this.codePageStack.Pop());
                            this.level--;
                            this.NotifyGroupEnd();
                            groupCount++;
                        }
                        else
                        {
                            throw new RtfBraceNestingException("ToManyBraces");
                        }
                        break;
                    default:
                        this.curText.Append(ReadOneChar(reader)); // must still consume the 'peek'ed char
                        break;
                }
                if (this.level == 0 && this.IgnoreContentAfterRootGroup)
                {
                    break;
                }
                if (peekCharValid)
                {
                    nextChar = peekChar;
                }
                else
                {
                    nextChar = PeekNextChar(reader, false);
                    backslashAlreadyConsumed = false;
                }
            }
            reader.Close();

            if (this.level > 0)
            {
                throw new RtfBraceNestingException("ToFewBraces");
            }
            if (groupCount == 0)
            {
                throw new RtfEmptyDocumentException("NoRtfContent");
            }
            this.curText = null;
        }

        private void ParseTag(RtfReader reader)
        {
            StringBuilder tagName = new StringBuilder();
            StringBuilder tagValue = null;
            bool readingName = true;
            bool delimReached = false;

            int nextChar = PeekNextChar(reader, true);
            while (!delimReached)
            {
                if (readingName && IsASCIILetter(nextChar))
                {
                    tagName.Append(ReadOneChar(reader)); // must still consume the 'peek'ed char
                }
                else if (IsDigit(nextChar) || (nextChar == '-' && tagValue == null))
                {
                    readingName = false;
                    if (tagValue == null)
                    {
                        tagValue = new StringBuilder();
                    }
                    tagValue.Append(ReadOneChar(reader)); // must still consume the 'peek'ed char
                }
                else
                {
                    delimReached = true;
                    RtfTag newTag;
                    if (tagValue != null && tagValue.Length > 0)
                    {
                        newTag = new RtfTag(tagName.ToString(), tagValue.ToString());
                    }
                    else
                    {
                        newTag = new RtfTag(tagName.ToString());
                    }
                    bool skippedContent = this.HandleTag(reader, newTag);
                    if (nextChar == ' ' && !skippedContent)
                    {
                        reader.Read(); // must still consume the 'peek'ed char
                    }
                }
                if (!delimReached)
                {
                    nextChar = PeekNextChar(reader, true);
                }
            }
        }

        private bool HandleTag(RtfReader reader, RtfTag tag)
        {
            if (this.level == 0)
            {
                throw new RtfStructureException(string.Format("TagOnRootLevel: {0}", tag.ToString()));
            }

            if (this.tagCount < 4)
            {
                // This only handles the initial encoding tag in the header section
                UpdateEncoding(reader, tag);
            }

            string tagName = tag.Name;

            // Enable the font name detection in case the last tag was introducing a theme font
            bool detectFontName = this.expectingThemeFont;
            if (this.tagCountAtLastGroupStart == this.tagCount)
            {
                // Dirst tag in a group
                switch (tagName)
                {
                    case RtfTags.TagThemeFontLoMajor:
                    case RtfTags.TagThemeFontHiMajor:
                    case RtfTags.TagThemeFontDbMajor:
                    case RtfTags.TagThemeFontBiMajor:
                    case RtfTags.TagThemeFontLoMinor:
                    case RtfTags.TagThemeFontHiMinor:
                    case RtfTags.TagThemeFontDbMinor:
                    case RtfTags.TagThemeFontBiMinor:
                        // These introduce a new font, but the actual font tag will be
                        // The second tag in the group, so we must remember this condition ...
                        this.expectingThemeFont = true;
                        break;
                }

                // Always enable the font name detection also for the first tag in a group
                detectFontName = true;
            }
            if (detectFontName)
            {
                // first tag in a group:
                switch (tagName)
                {
                    case RtfTags.TagFont:
                        if (this.fontTableStartLevel > 0)
                        {
                            // in the font-table definition:
                            // -> remember the target font for charset mapping
                            this.targetFont = tag.FullName;
                            this.expectingThemeFont = false; // reset that state now
                        }
                        break;
                    case RtfTags.TagFontTable:
                        // -> remember we're in the font-table definition
                        this.fontTableStartLevel = this.level;
                        break;
                }
            }
            if (this.targetFont != null)
            {
                // within a font-tables font definition: perform charset mapping
                if (RtfTags.TagFontCharset.Equals(tagName))
                {
                    int charSet = tag.ValueAsNumber;
                    int codePage = CharSetHelper.GetCodePage(charSet);
                    this.fontToCodePageMapping[this.targetFont] = codePage;
                    UpdateEncoding(reader, codePage);
                }
            }
            if (this.fontToCodePageMapping.Count > 0 && RtfTags.TagFont.Equals(tagName))
            {
                if (this.fontToCodePageMapping.ContainsKey(tag.FullName))
                {
                    UpdateEncoding(reader, this.fontToCodePageMapping[tag.FullName]);
                }
            }

            bool skippedContent = false;
            switch (tagName)
            {
                case RtfTags.TagUnicodeCode:
                    this.curText.Append((char)tag.ValueAsNumber);

                    // skip over the indicated number of 'alternative representation' text
                    for (int i = 0; i < this.unicodeSkipCount; i++)
                    {
                        int nextChar = PeekNextChar(reader, true);
                        switch (nextChar)
                        {
                            case ' ':
                            case '\r':
                            case '\n':
                                reader.Read(); // consume peeked char
                                skippedContent = true;
                                if (i == 0)
                                {
                                    // the first whitespace after the tag
                                    // -> only a delimiter, doesn't count for skipping ...
                                    i--;
                                }
                                break;
                            case '\\':
                                reader.Read(); // consume peeked char
                                skippedContent = true;
                                int secondChar = ReadOneByte(reader); // mandatory
                                switch (secondChar)
                                {
                                    case '\'':
                                        // ok, this is a hex-encoded 'byte' -> need to consume both
                                        // hex digits too
                                        ReadOneByte(reader); // high nibble
                                        ReadOneByte(reader); // low nibble
                                        break;
                                }
                                break;
                            case '{':
                            case '}':
                                // don't consume peeked char and abort skipping
                                i = this.unicodeSkipCount;
                                break;
                            default:
                                reader.Read(); // consume peeked char
                                skippedContent = true;
                                break;
                        }
                    }
                    break;
                case RtfTags.TagUnicodeSkipCount:
                    int newSkipCount = tag.ValueAsNumber;
                    if (newSkipCount < 0 || newSkipCount > 10)
                    {
                        throw new RtfUnicodeEncodingException(string.Format("InvalidUnicodeSkipCount: {0}", tag.ToString()));
                    }
                    this.unicodeSkipCount = newSkipCount;
                    break;
                case RtfTags.TagBinary:
                    if (PeekNextChar(reader, true) == ' ')
                    {
                        reader.Read(); // consume peeked char
                    }

                    if (tag.ValueAsNumber > 0)
                    {
                        RtfBinary binary = new RtfBinary(reader.ReadBinary(tag.ValueAsNumber));
                        this.NotifyBinaryFound(binary);
                    }

                    break;
                default:
                    this.FlushText();
                    this.NotifyTagFound(tag);
                    break;
            }

            this.tagCount++;

            return skippedContent;
        }

        private void DecodeCurrentHexBuffer(RtfReader reader)
        {
            this.hexDecodingBuffer.Seek(0, SeekOrigin.Begin);

            int b;
            while ((b = this.hexDecodingBuffer.ReadByte()) != -1)
            {
                this.curText.Append(reader.CodePageDecoder.Convert(b));
            }

            this.hexDecodingBuffer.SetLength(0);
        }

        private void FlushText()
        {
            if (this.curText.Length > 0)
            {
                if (this.level == 0)
                {
                    throw new RtfStructureException(string.Format("TextOnRootLevel: {0}", this.curText.ToString()));
                }
                this.NotifyTextFound(new RtfText(this.curText.ToString()));
                this.curText.Remove(0, this.curText.Length);
            }
        }
    }
}