﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.Model;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.RtfTypes;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.RtfTypes.Fonts;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.RtfTypes.Images;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.RtfTypes.Lists;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.RtfTypes.Styles;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Fields;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Model.Watermarks;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import
{
    internal class RtfImportContext
    {
        private readonly Stack<RtfStyle> styleStack;

        private readonly Stack<RtfStructureStackItem> structureStack;

        private readonly BlockContainerBase partBlockContainer;

        private Section currentSection;

        private Paragraph currentParagraph;

        public RtfImportContext()
        {
            this.Document = new RadFlowDocument();

            this.FontTable = new RtfFontTable();
            this.ColorTable = new RtfColorTable();
            this.DocumentInfo = new RtfDocumentInfo();
            this.ImageBuilder = new RtfImageBuilder(this);
            this.ShapeBuilder = new RtfShapeBuilder(this);
            this.ListTable = new RtfListTable();
            this.ListOverrideTable = new RtfListOverrideTable();
            this.Comments = new Dictionary<string, Comment>();
            this.Bookmarks = new Dictionary<string, Bookmark>();

            this.StylesTable = new RtfStylesTable();
            this.styleStack = new Stack<RtfStyle>();
            this.styleStack.Push(new RtfStyle(this.Document));

            this.CurrentSection = this.Document.Sections.AddSection();
            this.structureStack = new Stack<RtfStructureStackItem>();
            this.FieldContext = new FieldContext(true);
        }

        public RtfImportContext(RtfImportContext other, BlockContainerBase partBlockContainer)
            : this(other)
        {
            Guard.ThrowExceptionIfNull(partBlockContainer, "partBlockContainer");
            this.partBlockContainer = partBlockContainer;
        }

        public RtfImportContext(RtfImportContext other)
        {
            this.Document = other.Document;

            this.FontTable = other.FontTable;
            this.ColorTable = other.ColorTable;
            this.StylesTable = other.StylesTable;
            this.ImageBuilder = other.ImageBuilder;
            this.ShapeBuilder = other.ShapeBuilder;
            this.ListTable = other.ListTable;
            this.ListOverrideTable = other.ListOverrideTable;
            this.Comments = new Dictionary<string, Comment>();
            this.Bookmarks = new Dictionary<string, Bookmark>();

            this.styleStack = new Stack<RtfStyle>();
            this.styleStack.Push(new RtfStyle(other.Document));

            this.structureStack = new Stack<RtfStructureStackItem>();
            this.FieldContext = new FieldContext(true);
        }

        ////private List<FieldRangeStart> fieldsToUpdateAfterImport;

        public RtfImageBuilder ImageBuilder { get; private set; }

        public RtfShapeBuilder ShapeBuilder { get; private set; }

        public RtfStylesTable StylesTable { get; set; }

        public FieldContext FieldContext { get; private set; }

        public RtfColorTable ColorTable { get; set; }

        public RtfFontTable FontTable { get; set; }

        public RtfListTable ListTable { get; set; }

        public RtfListOverrideTable ListOverrideTable { get; set; }

        public RtfDocumentInfo DocumentInfo { get; set; }

        public string DefaultFontId { get; set; }

        public FontFamily DefaultFont
        {
            get
            {
                FontFamily font = null;

                if (this.FontTable != null &&
                    !string.IsNullOrEmpty(this.DefaultFontId) &&
                    this.FontTable.ContainsFontWithId(this.DefaultFontId))
                {
                    font = this.FontTable.GetFontFamily(this.DefaultFontId);
                }

                return font;
            }
        }

        public string Generator { get; set; }

        public RadFlowDocument Document { get; set; }

        public Section CurrentSection
        {
            get
            {
                return this.currentSection;
            }
            set
            {
                if (this.IsImportingPartBlockContainer)
                {
                    throw new InvalidOperationException("Cannot set current section while importing child document part.");
                }

                this.currentSection = value;
            }
        }

        public BlockContainerBase CurrentBaseBlockConatiner
        {
            get
            {
                if (this.IsImportingPartBlockContainer)
                {
                    return this.partBlockContainer;
                }
                else
                {
                    return this.CurrentSection;
                }
            }
        }

        public bool IsImportingPartBlockContainer
        {
            get
            {
                return this.partBlockContainer != null;
            }
        }

        public RtfStyle CurrentStyle
        {
            get
            {
                return this.styleStack.Peek();
            }
        }

        public TableRowStyle CurrentRowStyle
        {
            get
            {
                return this.CurrentStyle.RowStyle;
            }
        }

        public CommentImporter CurrentCommentHandler { get; set; }

        public Dictionary<string, Bookmark> Bookmarks { get; set; }
        public Dictionary<string, Comment> Comments { get; set; }

        public void HandleSectionEnd()
        {
            if (this.IsImportingPartBlockContainer)
            {
                Debug.Assert(false, "Section end inside header/footer.");
            }
            else
            {
                this.FlushSection();
                this.CurrentSection = this.Document.Sections.AddSection();
                ////this.CurrentSection.LinkHeaderToPrevious(HeaderFooterType.Default);
                ////this.CurrentSection.LinkHeaderToPrevious(HeaderFooterType.Even);
                ////this.CurrentSection.LinkHeaderToPrevious(HeaderFooterType.First);
                ////this.CurrentSection.LinkFooterToPrevious(HeaderFooterType.Default);
                ////this.CurrentSection.LinkFooterToPrevious(HeaderFooterType.Even);
                ////this.CurrentSection.LinkFooterToPrevious(HeaderFooterType.First);
            }
        }

        public void HandleParagraphEnd()
        {
            this.EnsureCurrentParagraph();
            this.FlushCurrentParagraph();
        }

        public void HandleRowEnd()
        {
            Debug.Assert(this.currentParagraph == null, "Current paragraph should be null here!");

            // Put all cells on the top of the stack in single row.
            List<TableCell> cells = new List<TableCell>();

            int expectedNestingLevel = (this.structureStack.Count != 0) ? this.structureStack.Peek().NestingLevel : 0;
            while (this.structureStack.Count != 0 &&
                this.structureStack.Peek().Element is TableCell &&
                this.structureStack.Peek().NestingLevel == expectedNestingLevel)
            {
                RtfStructureStackItem structureItem = this.structureStack.Pop();
                Debug.Assert(expectedNestingLevel == structureItem.NestingLevel, "Unexpected nesting level");
                cells.Add(structureItem.Element as TableCell);
            }

            cells.Reverse();

            Debug.Assert(cells.Count == this.CurrentRowStyle.CellStyles.Count, "Cell styles don't match the number of cells");
            int rowNestingLevel = this.CurrentStyle.CurrentNestingLevel - 1;
            if (rowNestingLevel < 0)
            {
                Debug.Assert(false, "Row nesting level should not be negative");

                // Recover anyway
                rowNestingLevel = 0;
            }

            RtfStructureRowStackItem rowStackItem = new RtfStructureRowStackItem(
                new TableRow(this.Document),
                rowNestingLevel,
                this.CurrentRowStyle);
            rowStackItem.CellsToBeAdded = cells;

            // Clear current row style
            this.structureStack.Push(rowStackItem);

            // Reset current border if needed
            if (this.CurrentStyle.CurrentBorderedElementType == BorderedElementType.TableCell ||
                this.CurrentStyle.CurrentBorderedElementType == BorderedElementType.TableRow)
            {
                this.CurrentStyle.CurrentBorderedElementType = BorderedElementType.Paragraph;
                this.CurrentStyle.CurrentBorder = null;
            }
        }

        public void HandleCellEnd()
        {
            bool isEmtyCell = this.structureStack.Count == 0 || !this.structureStack.Peek().IsRowOrParagraph;

            Debug.Assert(isEmtyCell || this.CurrentStyle.IsInTable, "Cell end while not in table!");

            this.FlushCurrentParagraph();

            if (!this.CurrentStyle.IsInTable)
            {
                return;
            }

            // Put all paragraphs on the top of the stack in single cell.
            List<BlockBase> blocks = new List<BlockBase>();

            int expectedNestingLevel = this.CurrentStyle.CurrentNestingLevel;
            if (isEmtyCell)
            {
                if (this.structureStack.Count == 0)
                {
                    expectedNestingLevel = 1;
                }
                else
                {
                    expectedNestingLevel = this.structureStack.Peek().NestingLevel;
                }
            }

            List<RtfStructureRowStackItem> rows = new List<RtfStructureRowStackItem>();

            while (this.structureStack.Count > 0 &&
                this.structureStack.Peek().IsRowOrParagraph &&
                this.structureStack.Peek().NestingLevel == expectedNestingLevel)
            {
                RtfStructureStackItem topItem = this.structureStack.Pop();

                if (topItem.Element is Paragraph)
                {
                    // Flush nested row into table.
                    if (rows.Count != 0)
                    {
                        blocks.Add(this.FlushTableRowsIntoTable(rows));
                        rows.Clear();
                    }

                    blocks.Add((Paragraph)topItem.Element);
                }
                else
                {
                    RtfStructureRowStackItem rowStackItem = (RtfStructureRowStackItem)topItem;
                    rows.Add(rowStackItem);
                }
            }

            // Flush rows collection.
            if (rows.Count != 0)
            {
                blocks.Add(this.FlushTableRowsIntoTable(rows));
                rows.Clear();
            }

            blocks.Reverse();

            TableCell cell = new TableCell(this.Document);
            foreach (BlockBase block in blocks)
            {
                cell.Blocks.Add(block);
            }
            this.structureStack.Push(new RtfStructureStackItem(cell, expectedNestingLevel));

            // Reset current border if needed.
            if (this.CurrentStyle.CurrentBorderedElementType == BorderedElementType.TableCell)
            {
                this.CurrentStyle.CurrentBorderedElementType = BorderedElementType.Paragraph;
                this.CurrentStyle.CurrentBorder = null;
            }
        }

        public void FlushSection()
        {
            this.FlushCurrentParagraph();
            this.FlushStructureStack();
            if (this.CurrentSection != null)
            {
                this.CurrentStyle.ApplySectionStyle(this.CurrentSection);
            }
        }

        public void EnterGroup(RtfGroup group)
        {
            this.styleStack.Push(new RtfStyle(this.CurrentStyle));

            if (group.Destination == RtfTags.TagField)
            {
                FieldCharacter fieldChar = new FieldCharacter(this.Document, FieldCharacterType.Start);
                this.AddInline(fieldChar);
                this.FieldContext.OnFieldCharacter(fieldChar);
            }
        }

        public void LeaveGroup(RtfGroup group)
        {
            if (group.Destination == RtfTags.TagFieldInstructions)
            {
                FieldCharacter fieldChar = new FieldCharacter(this.Document, FieldCharacterType.Separator);
                this.AddInline(fieldChar);
                this.FieldContext.OnFieldCharacter(fieldChar);
            }
            else if (group.Destination == RtfTags.TagField)
            {
                FieldCharacter fieldChar = new FieldCharacter(this.Document, FieldCharacterType.End);
                this.AddInline(fieldChar);
                this.FieldContext.OnFieldCharacter(fieldChar);
            }

            this.styleStack.Pop();
        }

        public void AddText(string text)
        {
            this.EnsureCurrentParagraph();

            // Skip content that is hidden
            if (!this.CurrentStyle.IsHidden)
            {
                Run run = this.currentParagraph.Inlines.AddRun();
                run.Text = text;
                this.CurrentStyle.ApplyCharacterStyle(run);
            }
        }

        public void AddInline(InlineBase inline)
        {
            this.EnsureCurrentParagraph();

            // Skip content that is hidden
            if (!this.CurrentStyle.IsHidden)
            {
                this.currentParagraph.Inlines.Add(inline);
            }
        }

        public void AddWatermark(Watermark watermark)
        {
            ((Header)this.CurrentBaseBlockConatiner).Watermarks.Add(watermark);
        }

        public void ClearUnclosedBookmarkRanges()
        {
            foreach (Bookmark bookmark in this.Bookmarks.Values)
            {
                if (bookmark.BookmarkRangeEnd.Paragraph == null)
                {
                    // Bookmark end was never added so remove the bookmark start
                    bookmark.BookmarkRangeStart.Paragraph.Inlines.Remove(bookmark.BookmarkRangeStart);
                }
            }
        }

        public Comment GetComment(string commentId)
        {
            Comment comment;
            if (!this.Comments.ContainsKey(commentId))
            {
                comment = this.Document.Comments.AddComment();
                this.Comments[commentId] = comment;
            }
            else
            {
                comment = this.Comments[commentId];
            }

            return comment;
        }

        private static void SetCellsPreferredWidth(List<int> cellXs, Table table)
        {
            int previous = 0;
            TableRow row = table.Rows.Where(ro => ro.Cells.Count == cellXs.Count).FirstOrDefault();

            // Row is null if there is cell style, but it relates to no cell. 
            Debug.Assert(row != null, "Cell styles do not match the count of cells.");

            if (row != null)
            {
                for (int i = 0; i < cellXs.Count; i++)
                {
                    int width = cellXs[i] - previous;
                    row.Cells[i].PreferredWidth = new TableWidthUnit(Unit.TwipToDipF(width));
                    previous = cellXs[i];
                }
            }
        }

        private void EnsureCurrentParagraph()
        {
            if (this.currentParagraph == null)
            {
                this.currentParagraph = new Paragraph(this.Document);
            }
        }

        private Table FlushTableRowsIntoTable(List<RtfStructureRowStackItem> rowStackItems)
        {
            rowStackItems.Reverse();

            ////HashSet<int> cellXhash = new HashSet<int>();
            ////foreach (var rowItem in rowStackItems)
            ////{
            ////    foreach (var cellStyle in rowItem.RowStyle.CellStyles)
            ////    {
            ////        cellXhash.Add(cellStyle.CellXBoundry);
            ////    }
            ////}

            ////List<int> cellXs = new List<int>(cellXhash.Count);
            ////foreach (int cellx in cellXhash.OrderBy(x => x))
            ////{
            ////    cellXs.Add(cellx);
            ////}
            List<int> cellXs = rowStackItems
                .SelectMany(rowItem => rowItem.RowStyle.CellStyles)
                .Select(cellStyle => cellStyle.CellXBoundry)
                .Distinct()
                .OrderBy(x => x)
                .ToList();

            int rowsCount = rowStackItems.Count;
            int colsCount = cellXs.Count;

            TableCell[,] cellGrid = new TableCell[colsCount, rowsCount];

            for (int currentRowIndex = 0; currentRowIndex < rowStackItems.Count; currentRowIndex++)
            {
                RtfStructureRowStackItem rowStackItem = rowStackItems[currentRowIndex];
                TableRow row = rowStackItem.Row;
                List<TableCellStyle> cellStyles = rowStackItem.RowStyle.CellStyles;
                List<TableCell> cells = rowStackItem.CellsToBeAdded;
                rowStackItem.RowStyle.ApplyRowStyle(row);

                int currentColumnIndex = 0;
                for (int i = 0; i < cells.Count; i++)
                {
                    if (cellStyles.Count <= i)
                    {
                        // There is no style for this cell. This is a bug, but we should add the cell(with no style) anyway.
                        row.Cells.Add(cells[i]);
                        continue;
                    }

                    // IsInHorizontalRange is ignored because of Excel exporting it even when cells are not merged
                    ////if (cellStyles[i].IsInHorizontalRange)
                    ////{
                    ////    //skip
                    ////    continue;
                    ////}

                    int cellLastColumn = cellXs.IndexOf(cellStyles[i].CellXBoundry) + 1;
                    int colSpan = cellLastColumn - currentColumnIndex;

                    if (colSpan == 0)
                    {
                        // This is an empty cell - skip it
                        continue;
                    }

                    if (cellStyles[i].IsInVerticalRange && currentRowIndex > 0 && cellGrid[currentColumnIndex, currentRowIndex - 1] != null)
                    {
                        TableCell cellAbove = cellGrid[currentColumnIndex, currentRowIndex - 1];
                        for (int limit = currentColumnIndex + colSpan; currentColumnIndex < limit; ++currentColumnIndex)
                        {
                            cellGrid[currentColumnIndex, currentRowIndex] = cellAbove;
                        }
                        cellAbove.RowSpan++;
                    }
                    else
                    {
                        for (int limit = currentColumnIndex + colSpan; currentColumnIndex < limit; ++currentColumnIndex)
                        {
                            cellGrid[currentColumnIndex, currentRowIndex] = cells[i];
                        }

                        // Add the cell;
                        cells[i].ColumnSpan = colSpan;
                        cellStyles[i].ApplyCellStyle(cells[i]);
                        row.Cells.Add(cells[i]);
                    }
                }
            }

            Table table = new Table(this.Document);

            // Get table properties form the fist row of the table
            rowStackItems.First().RowStyle.ApplyTableStyle(table);

            for (int currentRowIndex = 0; currentRowIndex < rowStackItems.Count; currentRowIndex++)
            {
                RtfStructureRowStackItem rowStackItem = rowStackItems[currentRowIndex];
                if (rowStackItem.Row.Cells.Any())
                {
                    table.Rows.Add(rowStackItem.Row);
                }
                else
                {
                    Debug.Assert(false, "Row has no cells");

                    // Row has no cells - this means this is row consisted only of row-spanned cells
                    // We don't support this case so we will skip this row but we will have to reduce the row span of all the cells that intersect with it.
                    HashSet<TableCell> rowSpannedCells = new HashSet<TableCell>();
                    for (int i = 0; i < colsCount; i++)
                    {
                        rowSpannedCells.Add(cellGrid[i, currentRowIndex]);
                    }
                    foreach (TableCell cell in rowSpannedCells)
                    {
                        if (cell.RowSpan > 1)
                        {
                            cell.RowSpan--;
                        }
                    }
                }
            }

            if (table.Properties.LayoutType.LocalValue == TableLayoutType.FixedWidth)
            {
                SetCellsPreferredWidth(cellXs, table);
            }

            return table;
        }

        private void FlushCurrentParagraph()
        {
            if (this.currentParagraph != null)
            {
                this.CurrentStyle.ApplyParagraphStyle(this.currentParagraph);

                if (this.CurrentStyle.IsInTable)
                {
                    Debug.Assert(this.CurrentStyle.CurrentNestingLevel != 0, "Paragraph is marked as InTable but nesting level is 0");
                    int nestingLevel = Math.Max(1, this.CurrentStyle.CurrentNestingLevel);
                    this.structureStack.Push(new RtfStructureStackItem(this.currentParagraph, nestingLevel));
                }
                else
                {
                    this.FlushStructureStack();
                    this.CurrentBaseBlockConatiner.Blocks.Add(this.currentParagraph);
                }
            }
            this.currentParagraph = null;
        }

        private void FlushStructureStack()
        {
            if (this.structureStack.Count != 0)
            {
                List<RtfStructureRowStackItem> rows = new List<RtfStructureRowStackItem>();

                while (this.structureStack.Count != 0)
                {
                    RtfStructureStackItem item = this.structureStack.Pop();
                    RtfStructureRowStackItem rowStackItem = item as RtfStructureRowStackItem;

                    Debug.Assert(rowStackItem != null, "Unexpected structure item");
                    if (rowStackItem != null)
                    {
                        Debug.Assert(rowStackItem.NestingLevel == 0, "Unexpected nesting level");

                        rows.Add(rowStackItem);
                    }
                }

                Table table = this.FlushTableRowsIntoTable(rows);

                this.CurrentBaseBlockConatiner.Blocks.Add(table);
            }
        }
    }
}