using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.Model;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.RtfTypes.Styles;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import
{
    internal class RtfStylesTable : RtfElementIteratorBase
    {
        private static readonly string TableNormalDisplayName = "Normal Table";

        private bool isInitialized;

        private RtfImportContext context;

        private Dictionary<int, RtfStyleDefinitionInfo> dictStyles;

        public void ReadTable(RtfGroup group, RtfImportContext context)
        {
            if (this.isInitialized)
            {
                // If the style table is already defined - ignore subsequent definitions.
                Debug.Assert(false, "StylesTable already initialized");
                return;
            }

            this.context = context;
            this.dictStyles = new Dictionary<int, RtfStyleDefinitionInfo>();
            this.VisitGroupChildren(group, false);

            this.InitializeRelatedStyles();

            this.AddStylesToDocumentRepositiory();

            this.FillStylePropertiesFormInfos();

            this.isInitialized = true;
        }

        public void ImportDefaultSpanStyle(RtfGroup group, RtfImportContext currentContext)
        {
            RtfImportContext localContext = new RtfImportContext(currentContext);
            RtfStyleImporter importer = new RtfStyleImporter(localContext, StyleType.Character);
            RtfStyleDefinitionInfo styleInfo = importer.ImportStyleGroup(group);

            if (currentContext.DefaultFont != null && !styleInfo.ImportedSpanProperties.Properties.FontFamily.HasLocalValue)
            {
                styleInfo.ImportedSpanProperties.FontFamily = new ThemableFontFamily(currentContext.DefaultFont);
            }

            CopyPropertiesIfNotSame(styleInfo.ImportedSpanProperties.Properties, currentContext.Document.DefaultStyle.CharacterProperties);
        }

        public void ImportDefaultParagraphStyle(RtfGroup group, RtfImportContext currentContext)
        {
            RtfImportContext localContext = new RtfImportContext(currentContext);
            RtfStyleImporter importer = new RtfStyleImporter(localContext, StyleType.Paragraph);
            RtfStyleDefinitionInfo styleInfo = importer.ImportStyleGroup(group);

            CopyParagraphPropertiesIfNotSame(styleInfo.ImportedParagraphProperties.Properties, currentContext.Document.DefaultStyle.ParagraphProperties);
        }

        public Style GetStyleById(int styleId)
        {
            RtfStyleDefinitionInfo result;
            if (this.dictStyles.TryGetValue(styleId, out result))
            {
                return result.CurrentStyle;
            }
            return null;
        }

        protected override void DoVisitGroup(RtfGroup group)
        {
            switch (group.Destination)
            {
                case RtfTags.TagStyleParagraph:
                    this.ImportStyle(group, StyleType.Paragraph);
                    break;
                case RtfTags.TagStyleCharacter:
                    this.ImportStyle(group, StyleType.Character);
                    break;
                case RtfTags.TagStyleTable:
                    this.ImportStyle(group, StyleType.Table);
                    break;
                case RtfTags.TagStyleSection:
                    // No support for section style
                    break;
                default:
                    // The default paragraph style
                    this.ImportStyle(group, StyleType.Paragraph, true);

                    break;
            }
        }

        private static void CopyParagraphPropertiesIfNotSame(ParagraphProperties fromProperties, ParagraphProperties toProperties)
        {
            CopyPropertiesIfNotSame(fromProperties, toProperties);

            // TODO: Tab Stops
            ////toProperties.TabStops = new TabStopCollection(fromProperties.TabStops);
        }

        private static void CopyPropertiesIfNotSame(DocumentElementPropertiesBase fromProperties, DocumentElementPropertiesBase toProperties)
        {
            if (fromProperties == null || fromProperties.GetType() != toProperties.GetType())
            {
                return;
            }

            foreach (IStyleProperty fromProperty in fromProperties.StyleProperties)
            {
                if (fromProperty.HasLocalValue)
                {
                    IStyleProperty toProperty = toProperties.GetStyleProperty(fromProperty.PropertyDefinition);
                    object fromValue = fromProperty.GetActualValueAsObject();
                    object toValue = toProperty.GetActualValueAsObject();

                    if (fromValue != null && !fromValue.Equals(toValue))
                    {
                        toProperty.SetValueAsObject(fromValue);
                    }
                }
            }
        }

        private static void FillStylePropertiesFromInfo(RtfStyleDefinitionInfo styleInfo)
        {
            RtfStylesTable.CopyPropertiesIfNotSame(styleInfo.ImportedSpanProperties.Properties, styleInfo.CurrentStyle.CharacterProperties);

            if (styleInfo.CurrentStyle.StyleType == StyleType.Paragraph || styleInfo.CurrentStyle.StyleType == StyleType.Table)
            {
                RtfStylesTable.CopyParagraphPropertiesIfNotSame(styleInfo.ImportedParagraphProperties.Properties, styleInfo.CurrentStyle.ParagraphProperties);
            }

            if (styleInfo.CurrentStyle.StyleType == StyleType.Table)
            {
                RtfStylesTable.CopyPropertiesIfNotSame(styleInfo.ImportedTableProperties.Properties, styleInfo.CurrentStyle.TableProperties);
                RtfStylesTable.CopyPropertiesIfNotSame(styleInfo.ImportedTableRowProperties.Properties, styleInfo.CurrentStyle.TableRowProperties);
                RtfStylesTable.CopyPropertiesIfNotSame(styleInfo.ImportedTableCellProperties.Properties, styleInfo.CurrentStyle.TableCellProperties);

                // There are some LocalProperties that should not be set imported in styles, 
                // but are because RTF default value differs form the model default value - we should reset them
                styleInfo.CurrentStyle.TableProperties.LayoutType.LocalValue = DocumentDefaultStyleSettings.TableLayoutType;
            }
        }

        private void InitializeRelatedStyles()
        {
            foreach (RtfStyleDefinitionInfo style in this.dictStyles.Values)
            {
                if (style.BasedOnStyleId.HasValue)
                {
                    RtfStyleDefinitionInfo basedOnStyle;
                    if (this.dictStyles.TryGetValue(style.BasedOnStyleId.Value, out basedOnStyle))
                    {
                        style.CurrentStyle.BasedOnStyleId = basedOnStyle.CurrentStyle.Id;
                    }
                }

                if (style.LinkedStyleId.HasValue)
                {
                    RtfStyleDefinitionInfo linkedStyle;
                    if (this.dictStyles.TryGetValue(style.LinkedStyleId.Value, out linkedStyle))
                    {
                        style.CurrentStyle.LinkedStyleId = linkedStyle.CurrentStyle.Id;
                    }
                }

                if (style.NextStyleId.HasValue)
                {
                    RtfStyleDefinitionInfo nextStyle;
                    if (this.dictStyles.TryGetValue(style.NextStyleId.Value, out nextStyle))
                    {
                        style.CurrentStyle.NextStyleId = nextStyle.CurrentStyle.Id;
                    }
                }
            }
        }

        private void AddStylesToDocumentRepositiory()
        {
            foreach (RtfStyleDefinitionInfo style in this.dictStyles.Values)
            {
                Style existingStyle = this.context.Document.StyleRepository.GetStyle(style.CurrentStyle.Id);
                if (existingStyle != null)
                {
                    style.CurrentStyle.IsDefault = existingStyle.IsDefault;
                    style.CurrentStyle.IsCustom = existingStyle.IsCustom;
                }

                this.context.Document.StyleRepository.Add(style.CurrentStyle);
            }
        }

        private void FillStylePropertiesFormInfos()
        {
            foreach (Style style in this.context.Document.StyleRepository.GetSortedTopologicallyStyles())
            {
                RtfStyleDefinitionInfo styleInfo = this.dictStyles.Values.FirstOrDefault(info => info.CurrentStyle == style);
                if (styleInfo != null)
                {
                    RtfStylesTable.FillStylePropertiesFromInfo(styleInfo);
                }
            }
        }

        private void ImportStyle(RtfGroup group, StyleType type, bool isDefault = false)
        {
            RtfImportContext localContext = new RtfImportContext(this.context);
            RtfStyleImporter importer = new RtfStyleImporter(localContext, type);

            RtfStyleDefinitionInfo styleInfo = importer.ImportStyleGroup(group);

            if (isDefault)
            {
                styleInfo.StyleId = 0;
            }
            else
            {
                Debug.Assert(styleInfo.StyleId.HasValue, "Style has no ID");
            }

            if (styleInfo.CurrentStyle.Id == RtfStylesTable.TableNormalDisplayName)
            {
                styleInfo.CurrentStyle.Id = BuiltInStyleNames.NormalTableStyleId;
            }

            // Import only the first style definitions
            // there may be multiple style definitions for tables, but we don't support them yet
            if (!this.dictStyles.ContainsKey(styleInfo.StyleId.Value))
            {
                this.dictStyles[styleInfo.StyleId.Value] = styleInfo;
            }
        }
    }
}