﻿using System.Diagnostics;
using System.Text;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.Model;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Utilities;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.RtfTypes.Fonts
{
    internal sealed class RtfFont : RtfElementIteratorBase
    {
        private readonly StringBuilder fontNameBuffer = new StringBuilder();

        private int codePage;

        public string Id { get; private set; }

        public string Name { get; private set; }

        public string NameAlt { get; private set; }

        public RtfFontFamily FontFamily { get; private set; }

        public RtfFontPitch Pitch { get; private set; }

        public int CharSet { get; private set; }

        public int CodePage
        {
            get
            {
                // If a codepage is specified, it overrides the charset setting.
                if (this.codePage == 0)
                {
                    // Unspecified codepage - use the one derived from the charset.
                    return CharSetHelper.GetCodePage(this.CharSet);
                }
                return this.codePage;
            }
        }

        public override bool Equals(object obj)
        {
            if (obj == this)
            {
                return true;
            }

            RtfFont other = obj as RtfFont;

            return other != null &&
                   this.Id.Equals(other.Id) &&
                   this.FontFamily == other.FontFamily &&
                   this.Pitch == other.Pitch &&
                   this.CharSet == other.CharSet &&
                   this.codePage == other.codePage &&
                   this.Name.Equals(other.Name);
        }

        public override int GetHashCode()
        {
            return HashTool.AddHashCode(this.GetType().GetHashCode(), this.ComputeHashCode());
        }

        public override string ToString()
        {
            return string.Format("{0}:{1}", this.Id, this.Name);
        }

        public void FillFont(RtfGroup group)
        {
            this.ResetState();

            this.VisitGroupChildren(group, false);

            this.Name = this.GetFontName();
        }

        protected override void DoVisitGroup(RtfGroup group)
        {
            switch (group.Destination)
            {
                case RtfTags.TagFont:
                case RtfTags.TagThemeFontLoMajor:
                case RtfTags.TagThemeFontHiMajor:
                case RtfTags.TagThemeFontDbMajor:
                case RtfTags.TagThemeFontBiMajor:
                case RtfTags.TagThemeFontLoMinor:
                case RtfTags.TagThemeFontHiMinor:
                case RtfTags.TagThemeFontDbMinor:
                case RtfTags.TagThemeFontBiMinor:
                    this.VisitGroupChildren(group, false);
                    break;
                case RtfTags.TagFontAlt:
                    this.GetFontAlt(group);
                    break;
            }
        }

        protected override void DoVisitTag(RtfTag tag)
        {
            switch (tag.Name)
            {
                case RtfTags.TagThemeFontLoMajor:
                case RtfTags.TagThemeFontHiMajor:
                case RtfTags.TagThemeFontDbMajor:
                case RtfTags.TagThemeFontBiMajor:
                case RtfTags.TagThemeFontLoMinor:
                case RtfTags.TagThemeFontHiMinor:
                case RtfTags.TagThemeFontDbMinor:
                case RtfTags.TagThemeFontBiMinor:

                    // Skip and ignore for the moment.
                    break;
                case RtfTags.TagFont:
                    this.Id = tag.FullName;
                    break;
                case RtfTags.TagFontKindNil:
                    this.FontFamily = RtfFontFamily.Nil;
                    break;
                case RtfTags.TagFontKindRoman:
                    this.FontFamily = RtfFontFamily.Roman;
                    break;
                case RtfTags.TagFontKindSwiss:
                    this.FontFamily = RtfFontFamily.Swiss;
                    break;
                case RtfTags.TagFontKindModern:
                    this.FontFamily = RtfFontFamily.Modern;
                    break;
                case RtfTags.TagFontKindScript:
                    this.FontFamily = RtfFontFamily.Script;
                    break;
                case RtfTags.TagFontKindDecor:
                    this.FontFamily = RtfFontFamily.Decor;
                    break;
                case RtfTags.TagFontKindTech:
                    this.FontFamily = RtfFontFamily.Tech;
                    break;
                case RtfTags.TagFontKindBidi:
                    this.FontFamily = RtfFontFamily.Bidi;
                    break;
                case RtfTags.TagFontCharset:
                    this.CharSet = tag.ValueAsNumber;
                    break;
                case RtfTags.TagCodePage:
                    this.codePage = tag.ValueAsNumber;
                    break;
                case RtfTags.TagFontPitch:
                    switch (tag.ValueAsNumber)
                    {
                        case 0:
                            this.Pitch = RtfFontPitch.Default;
                            break;
                        case 1:
                            this.Pitch = RtfFontPitch.Fixed;
                            break;
                        case 2:
                            this.Pitch = RtfFontPitch.Variable;
                            break;
                    }
                    break;
            }
        }

        protected override void DoVisitText(RtfText text)
        {
            this.fontNameBuffer.Append(text.Text);
        }

        private int ComputeHashCode()
        {
            int hash = this.Id.GetHashCode();
            hash = HashTool.AddHashCode(hash, this.FontFamily);
            hash = HashTool.AddHashCode(hash, this.Pitch);
            hash = HashTool.AddHashCode(hash, this.CharSet);
            hash = HashTool.AddHashCode(hash, this.codePage);
            hash = HashTool.AddHashCode(hash, this.Name);
            return hash;
        }

        private string GetFontName()
        {
            string fontName = null;
            int len = this.fontNameBuffer.Length;
            if (len > 0)
            {
                if (this.fontNameBuffer[len - 1] == ';')
                {
                    fontName = this.fontNameBuffer.ToString().Substring(0, len - 1).Trim();
                }
                else
                {
                    // Fix for rtf files generated from the Telerik Reporting which don't have ";" are the end of font records
                    Debug.Assert(false, "No \";\" at the end of font record.");
                    fontName = this.fontNameBuffer.ToString().Trim();
                }

                if (fontName.Length == 0)
                {
                    fontName = null;
                }
            }

            return fontName;
        }

        private void ResetState()
        {
            this.CharSet = 0;
            this.codePage = 0;
            this.FontFamily = RtfFontFamily.Nil;
            this.Pitch = RtfFontPitch.Default;
            this.fontNameBuffer.Length = 0;
        }

        private void GetFontAlt(RtfGroup group)
        {
            this.NameAlt = RtfHelper.GetGroupText(group);
        }
    }
}