﻿using System;
using System.IO;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.Model;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Utilities;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Media;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.RtfTypes.Images
{
    internal class RtfImageBuilder : RtfElementIteratorBase
    {
        private readonly RtfImportContext context;
        private bool isInInlinePictureShapeProperties;

        private RtfImage currentImage;

        public RtfImageBuilder(RtfImportContext context)
        {
            this.context = context;
        }

        public ImageInline ReadImage(RtfGroup group)
        {
            Util.EnsureGroupDestination(group, RtfTags.TagPicture);
            this.currentImage = new RtfImage();

            this.VisitGroupChildren(group, false);

            ImageInline result = this.CreateImageInline(this.context.Document);

            this.currentImage = null;

            return result;
        }

        protected override void DoVisitTag(RtfTag tag)
        {
            switch (tag.Name)
            {
                case RtfTags.TagIsWordArtShape:
                    this.currentImage.IsWordArtShape = true;
                    break;
                case RtfTags.TagPictureFormatPng:
                    this.currentImage.Extension = "png";
                    break;
                case RtfTags.TagPictureFormatJpg:
                    this.currentImage.Extension = "jpg";
                    break;
                case RtfTags.TagPictureFormatWinDib:
                case RtfTags.TagPictureFormatWinBmp:
                    this.currentImage.Extension = "bmp";
                    break;
                case RtfTags.TagPictureFormatEmf:
                    this.currentImage.Extension = "emf";
                    break;
                case RtfTags.TagPictureFormatQuickDraw:
                    this.currentImage.Extension = "pict";
                    break;
                case RtfTags.TagPictureFormatWinMetafile:
                    this.currentImage.Extension = "wmf";
                    break;
                case RtfTags.TagPictureWidth:
                    this.currentImage.Width = tag.ValueAsNumber;
                    break;
                case RtfTags.TagPictureHeight:
                    this.currentImage.Height = tag.ValueAsNumber;
                    break;
                case RtfTags.TagPictureWidthGoal:
                    this.currentImage.DesiredWidth = Unit.TwipToDip(tag.ValueAsNumber);
                    break;
                case RtfTags.TagPictureHeightGoal:
                    this.currentImage.DesiredHeight = Unit.TwipToDip(tag.ValueAsNumber);
                    break;
                case RtfTags.TagPictureScaleX:
                    this.currentImage.ScaleX = tag.ValueAsNumber / 100.0;
                    break;
                case RtfTags.TagPictureScaleY:
                    this.currentImage.ScaleY = tag.ValueAsNumber / 100.0;
                    break;
                case RtfTags.TagPictureWrapper:
                case RtfTags.TagPictureWrapperAlternative:
                case RtfTags.TagPictureFormatOs2Metafile:
                case RtfTags.TagPicture:
                default:
                    // Not supported
                    break;
            }
        }

        protected override void DoVisitText(RtfText text)
        {
            this.currentImage.Data = text.Text;
        }

        protected override void DoVisitBinary(RtfBinary bin)
        {
            this.currentImage.BinaryData = bin.Data;
        }

        protected override void DoVisitGroup(RtfGroup group)
        {
            switch (group.Destination)
            {
                case RtfTags.TagInlinePictureShapeProperties:
                    if (!this.isInInlinePictureShapeProperties)
                    {
                        this.isInInlinePictureShapeProperties = true;
                        this.VisitGroupChildren(group, false);
                        this.isInInlinePictureShapeProperties = false;
                    }
                    break;
                case RtfTags.TagShapePopertyDestenation:
                    if (this.isInInlinePictureShapeProperties)
                    {
                        this.currentImage.ReadInstructionGroup(group);
                    }
                    break;
                case RtfTags.TagShapeResultDestination:
                    // Not used
                    break;
                default:
                    // Not recognized
                    break;
            }
        }

        private ImageInline CreateImageInline(RadFlowDocument doc)
        {
            if (this.currentImage == null)
            {
                return null;
            }

            if (this.currentImage.IsWordArtShape)
            {
                // This is WordArt image and should not be added to document
                return null;
            }

            double width;
            double height;

            if (this.currentImage.Extension == "bmp" ||
                this.currentImage.Extension == "pict")
            {
                width = Math.Max(this.currentImage.Width * this.currentImage.ScaleX, 1);
                height = Math.Max(this.currentImage.Height * this.currentImage.ScaleY, 1);
            }
            else
            {
                width = Math.Max(this.currentImage.DesiredWidth * this.currentImage.ScaleX, 1);
                height = Math.Max(this.currentImage.DesiredHeight * this.currentImage.ScaleY, 1);
            }

            ImageInline result = new ImageInline(doc);

            MemoryStream stream;
            if (this.currentImage.BinaryData != null)
            {
                stream = new MemoryStream(this.currentImage.BinaryData);
            }
            else
            {
                stream = new MemoryStream(RtfHelper.GetBytesFromHexString(this.currentImage.Data));
            }

            using (stream)
            {
                result.Image.ImageSource = new ImageSource(stream, this.currentImage.Extension);
            }

            result.Image.IsHorizontallyFlipped = this.currentImage.GetPropertyIntValue(RtfTags.ShapeHorizontalFlipProperty, 0) == 1;
            result.Image.IsVerticallyFlipped = this.currentImage.GetPropertyIntValue(RtfTags.ShapeVerticalFlipProperty, 0) == 1;

            double angle = RtfHelper.RtfValueToAngle(this.currentImage.GetPropertyIntValue(RtfTags.ShapeRotationProperty, 0));
            Size size = new Size(width, height);
            size = RtfHelper.CalculateImageSizeForRotateAngle(size, angle);

            result.Image.RotationAngle = angle;
            result.Image.Size = size;
            return result;
        }
    }
}