﻿using System;
using System.Diagnostics;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Primitives;
#else
using System.Windows;
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Primitives;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.Model;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Utilities;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Flow.Model.Watermarks;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Model.Drawing.Shapes;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.RtfTypes.Images
{
    internal class RtfShapeBuilder : RtfElementIteratorBase
    {
        private readonly RtfImportContext context;
        private bool isInInstructions;
        private RtfShape currentShape;

        public RtfShapeBuilder(RtfImportContext context)
        {
            this.context = context;
        }

        private RtfImageBuilder ImageBuilder
        {
            get
            {
                return this.context.ImageBuilder;
            }
        }

        public void ReadShape(RtfGroup group, out InlineBase inline, out Watermark watermark)
        {
            Util.EnsureGroupDestination(group, RtfTags.TagShape);
            this.currentShape = new RtfShape();

            this.VisitGroupChildren(group, false);

            if (this.IsWatermark())
            {
                watermark = this.CreateWatermark();
                inline = null;
            }
            else
            {
                watermark = null;
                inline = this.CreateInline();
            }

            this.Clear();
        }

        protected override void DoVisitTag(RtfTag tag)
        {
            switch (tag.Name)
            {
                case RtfTags.TagShapeTop:
                    this.currentShape.Top = Unit.TwipToDip(tag.ValueAsNumber);
                    break;
                case RtfTags.TagShapeLeft:
                    this.currentShape.Left = Unit.TwipToDip(tag.ValueAsNumber);
                    break;
                case RtfTags.TagShapeBottom:
                    this.currentShape.Bottom = Unit.TwipToDip(tag.ValueAsNumber);
                    break;
                case RtfTags.TagShapeRight:
                    this.currentShape.Right = Unit.TwipToDip(tag.ValueAsNumber);
                    break;
                case RtfTags.TagShapeZOrder:
                    this.currentShape.ZIndex = tag.ValueAsNumber;
                    break;
                case RtfTags.TagShapeIsInHeader:
                    this.currentShape.IsInHeader = !(tag.HasValue && tag.ValueAsNumber == 1);
                    break;
                case RtfTags.TagShapeHorizontalRelativeToPage:
                    this.currentShape.HorizontalRelativeFrom = HorizontalRelativeFrom.Page;
                    break;
                case RtfTags.TagShapeHorizontalRelativeToMargin:
                    this.currentShape.HorizontalRelativeFrom = HorizontalRelativeFrom.Margin;
                    break;
                case RtfTags.TagShapeHorizontalRelativeToColumn:
                    this.currentShape.HorizontalRelativeFrom = HorizontalRelativeFrom.Column;
                    break;
                case RtfTags.TagShapeHorizontalRelativeToIgnore:
                    this.currentShape.IgnoreHorizontalRelativeFrom = true;
                    break;
                case RtfTags.TagShapeVerticalRelativeToPage:
                    this.currentShape.VerticalRelativeFrom = VerticalRelativeFrom.Page;
                    break;
                case RtfTags.TagShapeVerticalRelativeToMargin:
                    this.currentShape.VerticalRelativeFrom = VerticalRelativeFrom.Margin;
                    break;
                case RtfTags.TagShapeVerticalRelativeToParagraph:
                    this.currentShape.VerticalRelativeFrom = VerticalRelativeFrom.Paragraph;
                    break;
                case RtfTags.TagShapeVerticalRelativeToIgnore:
                    this.currentShape.IgnoreVerticalRelativeFrom = true;
                    break;
                case RtfTags.TagShapeWrappingType:
                    this.currentShape.Wrapping.WrappingType = RtfHelper.WrappingStyleMapper.GetToValue(tag.ValueAsNumber);
                    break;
                case RtfTags.TagShapeTextWrap:
                    this.currentShape.Wrapping.TextWrap = RtfHelper.TextWrapMapper.GetToValue(tag.ValueAsNumber);
                    break;
                case RtfTags.TagShapeIsBelowText:
                    this.currentShape.IsBelowText = tag.HasValue && tag.ValueAsNumber == 1;
                    break;
                default:
                    // Not supported
                    break;
            }
        }

        protected override void DoVisitGroup(RtfGroup group)
        {
            switch (group.Destination)
            {
                case RtfTags.TagShapeInstructions:
                    if (!this.isInInstructions)
                    {
                        this.isInInstructions = true;
                        this.VisitGroupChildren(group, false);
                        this.isInInstructions = false;
                    }
                    break;
                case RtfTags.TagShapePopertyDestenation:
                    if (this.isInInstructions)
                    {
                        this.currentShape.ReadInstructionGroup(group);
                    }
                    break;
                case RtfTags.TagShapeResultDestination:
                    // Not used
                    break;
                default:
                    // Not recognized
                    break;
            }
        }

        private void Clear()
        {
            this.currentShape = null;
            this.isInInstructions = false;
        }

        private InlineBase CreateInline()
        {
            ImageInline imageInline = null;
            RtfGroup pictureGroup = null;
            this.currentShape.Properties.TryGetValue(RtfTags.ShapePictureDataProperty, out pictureGroup);

            Debug.Assert(pictureGroup != null, "No picture data found");
            if (pictureGroup != null)
            {
                imageInline = this.ImageBuilder.ReadImage(pictureGroup.Elements[1] as RtfGroup);
                if (imageInline == null)
                {
                    return null;
                }

                imageInline.Image.RotationAngle = RtfHelper.RtfValueToAngle(this.currentShape.GetPropertyIntValue(RtfTags.ShapeRotationProperty, 0));
                Size imageSize = new Size(this.currentShape.Right - this.currentShape.Left, this.currentShape.Bottom - this.currentShape.Top);
                imageInline.Image.Size = RtfHelper.CalculateImageSizeForRotateAngle(imageSize, imageInline.Image.RotationAngle);
            }
            else
            {
                return null;
            }

            FloatingImage result = new FloatingImage(this.context.Document, imageInline.Image);
            result.AllowOverlap = this.currentShape.GetPropertyBoolValue(RtfTags.ShapeAllowOverlapProperty, true);
            result.IsBehindDocument = this.currentShape.IsBelowText;
            result.Wrapping = this.currentShape.Wrapping;
            result.LayoutInCell = this.currentShape.GetPropertyBoolValue(RtfTags.ShapeLayoutInCellProperty, false);

            result.Margin = new Padding(
                Unit.EmuToDip(this.currentShape.GetPropertyIntValue(RtfTags.ShapeMarginLeftProperty, RtfHelper.DefaultFloatingBlockHorizontalMargin)),
                Unit.EmuToDip(this.currentShape.GetPropertyIntValue(RtfTags.ShapeMarginTopProperty, 0)),
                Unit.EmuToDip(this.currentShape.GetPropertyIntValue(RtfTags.ShapeMarginRightProperty, RtfHelper.DefaultFloatingBlockHorizontalMargin)),
                Unit.EmuToDip(this.currentShape.GetPropertyIntValue(RtfTags.ShapeMarginBottomProperty, 0)));

            result.HorizontalPosition = this.GetHorizontalPosition();
            result.VerticalPosition = this.GetVerticalPosition();
            result.ZIndex = this.currentShape.GetPropertyIntValue(RtfTags.ShapeZIndexProperty, this.currentShape.ZIndex);
            if (this.currentShape.GetPropertyIntValue(RtfTags.ShapePseudoInlineProperty, 0) == 1)
            {
                return imageInline;
            }
            else
            {
                return result;
            }
        }

        private VerticalPosition GetVerticalPosition()
        {
            VerticalPosition verticalPos = new VerticalPosition();
            verticalPos.RelativeFrom = RtfHelper.RtfValueToVerticalRelativeFrom(
                this.currentShape.GetPropertyIntValue(RtfTags.ShapeVerticalRelativeToPositionProperty, -1),
                this.currentShape.VerticalRelativeFrom);

            verticalPos.Offset = this.currentShape.Top;

            int align = this.currentShape.GetPropertyIntValue(RtfTags.ShapeVerticalPositionProperty, 0);
            if (align == 0)
            {
                verticalPos.ValueType = PositionValueType.Offset;
            }
            else
            {
                verticalPos.ValueType = PositionValueType.Alignment;
                verticalPos.Alignment = RtfHelper.RelativeVerticalAlignmentMapper.GetToValue(align);
            }

            return verticalPos;
        }

        private HorizontalPosition GetHorizontalPosition()
        {
            HorizontalPosition horizontalPos = new HorizontalPosition();
            horizontalPos.RelativeFrom = RtfHelper.RtfValueToHorizontalRelativeFrom(
                this.currentShape.GetPropertyIntValue(RtfTags.ShapeHorizontalRelativeToPositionProperty, -1),
                this.currentShape.HorizontalRelativeFrom);

            horizontalPos.Offset = this.currentShape.Left;

            int align = this.currentShape.GetPropertyIntValue(RtfTags.ShapeHorizontalPositionProperty, 0);
            if (align == 0)
            {
                horizontalPos.ValueType = PositionValueType.Offset;
            }
            else
            {
                horizontalPos.ValueType = PositionValueType.Alignment;
                horizontalPos.Alignment = RtfHelper.RelativeHorizontalAlignmentMapper.GetToValue(align);
            }

            return horizontalPos;
        }

        private Watermark CreateWatermark()
        {
            string name = this.currentShape.GetPropertyStringValue(RtfTags.ShapeNameProperty);
            if (name.Contains(RtfTags.TextWatermarkShapeName))
            {
                return this.CreateTextWatermark();
            }
            else if (name.Contains(RtfTags.ImageWatermarkShapeName))
            {
                return this.CreateImageWatermark();
            }
            Debug.Assert(false, "Cannot create watermark");
            return null;
        }

        private Watermark CreateImageWatermark()
        {
            InlineBase inline = this.CreateInline();
            FloatingImage floatingImage = inline as FloatingImage;
            ImageInline imageInline = inline as ImageInline;
            Image image = floatingImage != null ? floatingImage.Image : imageInline.Image;
            if (image != null)
            {
                ImageWatermarkSettings imageSettings = new ImageWatermarkSettings();
                imageSettings.Width = image.Width;
                imageSettings.Height = image.Height;
                imageSettings.ImageSource = image.ImageSource;
                imageSettings.Angle = image.RotationAngle;

                return new Watermark(imageSettings);
            }

            throw new InvalidOperationException("Cannot construct watermark");
        }

        private Watermark CreateTextWatermark()
        {
            TextWatermarkSettings textSettings = new TextWatermarkSettings();
            string fontFamily = this.currentShape.GetPropertyStringValue(RtfTags.ShapeTextFontProperty);
            if (!string.IsNullOrEmpty(fontFamily))
            {
                textSettings.FontFamily = new FontFamily(fontFamily);
            }
            textSettings.ForegroundColor = RtfHelper.IntToColor(this.currentShape.GetPropertyIntValue(RtfTags.ShapeFillColorProperty, int.MaxValue));
            string text = this.currentShape.GetPropertyStringValue(RtfTags.ShapeTextProperty);
            if (!string.IsNullOrEmpty(text))
            {
                textSettings.Text = text;
            }
            textSettings.Angle = RtfHelper.RtfValueToAngle(this.currentShape.GetPropertyIntValue(RtfTags.ShapeRotationProperty, 0));
            textSettings.Opacity = RtfHelper.RtfFixedPointSizeToDouble(this.currentShape.GetPropertyIntValue(RtfTags.ShapeFillOpacityProperty, 65536));
            textSettings.Width = this.currentShape.Right - this.currentShape.Left;
            textSettings.Height = this.currentShape.Bottom - this.currentShape.Top;

            return new Watermark(textSettings);
        }

        private bool IsWatermark()
        {
            string name = this.currentShape.GetPropertyStringValue(RtfTags.ShapeNameProperty);
            return name.Contains(RtfTags.TextWatermarkShapeName) || name.Contains(RtfTags.ImageWatermarkShapeName);
        }
    }
}