﻿using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.Model;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.TagHandlers;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Utilities;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Flow.Model.Styles;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.RtfTypes.Lists
{
    internal class ListLevelInfo : RtfElementIteratorBase
    {
        private static readonly Dictionary<string, ControlTagHandler> tagHandlers;

        private string rtfLevelText;
        private string rtfLevelNumbersInfo;
        private bool rtfNumberingFormatLocked;

        private bool isLevelTextValid;
        private string levelText;
        private RtfImportContext localContext;

        static ListLevelInfo()
        {
            ListLevelInfo.tagHandlers = new Dictionary<string, ControlTagHandler>();
            SpanStyleHandlers.InitializeSpanStyleHandlers(ListLevelInfo.tagHandlers);
            ParagraphStyleHandlers.InitializeParagraphStyleHandlers(ListLevelInfo.tagHandlers);
        }

        public ListLevelInfo()
        {
        }

        public int LevelStartIndex { get; private set; }
        public string LevelTemplateId { get; private set; }
        public NumberingStyle NumberingStyle { get; private set; }
        public Alignment Alignment { get; private set; }
        public bool IsLegal { get; private set; }
        public bool RestartAfterLevel { get; private set; }
        public int? StyleId { get; set; }

        public string LevelText
        {
            get
            {
                if (!this.isLevelTextValid && !string.IsNullOrEmpty(this.rtfLevelText))
                {
                    this.ConstructText();
                }
                return this.levelText;
            }
        }

        public void ReadListLevelInfo(RtfGroup group, RtfImportContext context)
        {
            Util.EnsureGroupDestination(group, RtfTags.TagListLevel);

            this.localContext = new RtfImportContext(context);

            this.VisitGroupChildren(group, false);
        }

        public ListLevel CreateListLevel(RtfImportContext context, int level)
        {
            ListLevel result = new ListLevel(context.CurrentStyle.CurrentList)
            {
                StartIndex = this.LevelStartIndex,
                NumberingStyle = this.NumberingStyle,
                Alignment = this.Alignment,
                IsLegal = this.IsLegal,
            };

            if (!string.IsNullOrEmpty(this.LevelText))
            {
                result.NumberTextFormat = this.LevelText;
            }

            if (this.RestartAfterLevel)
            {
                // In RTF RestartAfterLevel is used as a bool one (not like the docx). If it is set, its value is 0 (the first ListLevel).
                result.RestartAfterLevel = 0;
            }

            if (this.StyleId != null)
            {
                Style style = context.StylesTable.GetStyleById((int)this.StyleId);
                if (style != null && style.StyleType == StyleType.Paragraph)
                {
                    result.StyleId = style.Id;
                    if (!style.ParagraphProperties.ListLevel.HasLocalValue)
                    {
                        style.ParagraphProperties.ListLevel.LocalValue = level;
                    }
                    if (!style.ParagraphProperties.ListId.HasLocalValue)
                    {
                        style.ParagraphProperties.ListId.LocalValue = context.CurrentStyle.CurrentList.Id;
                    }
                    result.CharacterProperties.CopyPropertiesFrom(style.CharacterProperties);
                    result.ParagraphProperties.CopyPropertiesFrom(style.ParagraphProperties);
                }
            }

            result.CharacterProperties.CopyPropertiesFrom(this.localContext.CurrentStyle.CharacterStyle.Properties);
            result.ParagraphProperties.CopyPropertiesFrom(this.localContext.CurrentStyle.ParagraphStyle.Properties);

            return result;
        }

        protected override void DoVisitTag(RtfTag tag)
        {
            switch (tag.Name)
            {
                case RtfTags.TagListLevelNumberingFormat:
                    if (!this.rtfNumberingFormatLocked)
                    {
                        this.NumberingStyle = (NumberingStyle)tag.ValueAsNumber;
                    }
                    break;

                case RtfTags.TagListLevelNumberingFormatAlt:
                    this.rtfNumberingFormatLocked = true;
                    this.NumberingStyle = (NumberingStyle)tag.ValueAsNumber;
                    break;

                case RtfTags.TagListLevelStartIndex:
                    this.LevelStartIndex = tag.ValueAsNumber;
                    break;

                case RtfTags.TagListLevelJustificationAlt:
                case RtfTags.TagListLevelJustification:
                    this.Alignment = RtfHelper.ListLevelAlignmentMapper.GetToValue(tag.ValueAsNumber);
                    break;

                case RtfTags.TagStyleParagraph:
                    this.StyleId = tag.ValueAsNumber;
                    break;

                case RtfTags.TagListLevelIsLegal:
                    this.IsLegal = tag.ValueAsNumber == 0 ? false : true;
                    break;

                case RtfTags.TagListLevelRestartAfterLevel:
                    this.RestartAfterLevel = true;
                    break;
            }

            // Handle span properties with TagHandlers.
            ControlTagHandler handler;
            ListLevelInfo.tagHandlers.TryGetValue(tag.Name, out handler);
            if (handler != null)
            {
                handler(tag, this.localContext);
            }
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1800:DoNotCastUnnecessarily")]
        protected override void DoVisitGroup(RtfGroup group)
        {
            switch (group.Destination)
            {
                case RtfTags.TagListLevelText:
                    RtfTag templateId = group.Elements.FirstOrDefault(el => el is RtfTag && ((RtfTag)el).Name == RtfTags.TagListLevelTemplateId) as RtfTag;
                    this.LevelTemplateId = templateId != null ? templateId.ValueAsText : null;

                    string text = RtfHelper.GetGroupText(group);

                    Debug.Assert(!string.IsNullOrEmpty(text), "Could not find level text");
                    this.rtfLevelText = text.TrimEnd(';');
                    this.isLevelTextValid = false;

                    break;

                case RtfTags.TagListLevelNumbers:
                    Debug.Assert(group.Elements.Count == 2, "ListLevelNumbers group should have 2 elements");
                    if (group.Elements.Count == 2)
                    {
                        RtfText levelNumbersData = group.Elements[1] as RtfText;
                        Debug.Assert(levelNumbersData != null, "Unexpected element type");
                        if (levelNumbersData != null)
                        {
                            this.rtfLevelNumbersInfo = levelNumbersData.Text.TrimEnd(';');
                        }
                        this.isLevelTextValid = false;
                    }
                    break;
            }
        }

        private void ConstructText()
        {
            List<int> placeholderIndexes = new List<int>();
            if (!string.IsNullOrEmpty(this.rtfLevelNumbersInfo))
            {
                for (int i = 0; i < this.rtfLevelNumbersInfo.Length; i++)
                {
                    placeholderIndexes.Add(this.rtfLevelNumbersInfo[i]);
                }
            }

            placeholderIndexes.Sort();
            placeholderIndexes.Reverse();

            Debug.Assert(!string.IsNullOrEmpty(this.rtfLevelText), "LevelText is empty");
            Debug.Assert(this.rtfLevelText.Length == this.rtfLevelText[0] + 1, "LevelText is invalid");

            StringBuilder sb = new StringBuilder(this.rtfLevelText);

            foreach (int index in placeholderIndexes)
            {
                int param = this.rtfLevelText[index];
                sb.Remove(index, 1);
                sb.Insert(index, "%" + (param + 1).ToString());
            }

            this.isLevelTextValid = true;
            this.levelText = sb.ToString().Substring(1);
        }
    }
}