﻿using System.Collections.Generic;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Linq;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Exceptions;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.Model;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Utilities;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Flow.Model.Watermarks;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Rtf.Import.TagHandlers
{
    internal delegate bool ControlGroupHandler(RtfGroup group, RtfImportContext context);

    internal static class GroupTagHandlers
    {
        public static void InitializeGroupHandlers(Dictionary<string, ControlGroupHandler> groupHandlers)
        {
            ////// Header Groups
            groupHandlers[RtfTags.TagFontTable] = GroupTagHandlers.FontTableHandler;
            groupHandlers[RtfTags.TagColorTable] = GroupTagHandlers.ColorTableHandler;
            groupHandlers[RtfTags.TagGenerator] = GroupTagHandlers.GeneratorHandler;

            groupHandlers[RtfTags.TagDefaultCharacterProperties] = GroupTagHandlers.DefaultCharacterPropertiesHandler;
            groupHandlers[RtfTags.TagDefaultParagraphProperties] = GroupTagHandlers.DefaultParagraphGroupHandler;
            groupHandlers[RtfTags.TagStyleSheet] = GroupTagHandlers.StylesTableHandler;

            // Lists
            groupHandlers[RtfTags.TagListTable] = GroupTagHandlers.ListTableHandler;
            groupHandlers[RtfTags.TagListOverrideTable] = GroupTagHandlers.ListOverrideTableHandler;

            // Ignored groups
            groupHandlers[RtfTags.TagThemeData] = GroupTagHandlers.IgnoreGroupHandler;
            groupHandlers[RtfTags.TagColorSchemeMapping] = GroupTagHandlers.IgnoreGroupHandler;
            groupHandlers[RtfTags.TagFileTable] = GroupTagHandlers.IgnoreGroupHandler;
            groupHandlers[RtfTags.TagParagraphGroupProperties] = GroupTagHandlers.IgnoreGroupHandler;
            groupHandlers[RtfTags.TagRevisionMarksTable] = GroupTagHandlers.IgnoreGroupHandler;
            groupHandlers[RtfTags.TagRevisionSaveIdTable] = GroupTagHandlers.IgnoreGroupHandler;
            groupHandlers[RtfTags.TagUserProtectionInfoTable] = GroupTagHandlers.IgnoreGroupHandler;
            groupHandlers[RtfTags.TagParagraphNumberText] = GroupTagHandlers.IgnoreGroupHandler;

            //////Document Groups
            groupHandlers[RtfTags.TagInfo] = GroupTagHandlers.DocumentInfoGroupHandler;
            groupHandlers[RtfTags.TagPicture] = GroupTagHandlers.ImageGroupHandler;
            groupHandlers[RtfTags.TagShape] = GroupTagHandlers.ShapeGroupHandler;
            groupHandlers[RtfTags.TagNoNestedTablesDestenation] = GroupTagHandlers.IgnoreGroupHandler;
            groupHandlers[RtfTags.TagPictureWrapperAlternative] = GroupTagHandlers.IgnoreGroupHandler;
            groupHandlers[RtfTags.TagPictureWrapper] = GroupTagHandlers.VisitGroupHandler;
            groupHandlers[RtfTags.TagNestedTablePropertiesDestenation] = GroupTagHandlers.VisitGroupHandler;
            groupHandlers[RtfTags.TagListText] = GroupTagHandlers.IgnoreGroupHandler;
            groupHandlers[RtfTags.TagBookmarkStart] = GroupTagHandlers.BookmarkStartHandler;
            groupHandlers[RtfTags.TagBookmarkEnd] = GroupTagHandlers.BookmarkEndHandler;
            groupHandlers[RtfTags.TagField] = GroupTagHandlers.VisitGroupHandler;
            groupHandlers[RtfTags.TagFieldInstructions] = GroupTagHandlers.VisitGroupHandler;
            groupHandlers[RtfTags.TagFieldResult] = GroupTagHandlers.VisitGroupHandler;

            groupHandlers[RtfTags.CommentStart] = GroupTagHandlers.CommentStartHandler;
            groupHandlers[RtfTags.CommentEnd] = GroupTagHandlers.CommentEndHandler;
            groupHandlers[RtfTags.CommentAuthor] = GroupTagHandlers.CommentAuthorHandler;
            groupHandlers[RtfTags.CommentDefinition] = GroupTagHandlers.CommentDefinitionHandler;
            groupHandlers[RtfTags.CommentID] = GroupTagHandlers.CommentIdHandler;
            groupHandlers[RtfTags.CommentDate] = GroupTagHandlers.CommentDateHandler;
            groupHandlers[RtfTags.TagDocumentVariable] = GroupTagHandlers.DocumentVariableHandler;

            // Headers/Footers
            groupHandlers[RtfTags.TagHeaderAllPages] = GroupTagHandlers.HeaderFooterHandler;
            groupHandlers[RtfTags.TagHeaderFirstPage] = GroupTagHandlers.HeaderFooterHandler;
            groupHandlers[RtfTags.TagHeaderLeftPages] = GroupTagHandlers.HeaderFooterHandler;
            groupHandlers[RtfTags.TagHeaderRightPages] = GroupTagHandlers.HeaderFooterHandler;
            groupHandlers[RtfTags.TagFooterAllPages] = GroupTagHandlers.HeaderFooterHandler;
            groupHandlers[RtfTags.TagFooterFirstPage] = GroupTagHandlers.HeaderFooterHandler;
            groupHandlers[RtfTags.TagFooterLeftPages] = GroupTagHandlers.HeaderFooterHandler;
            groupHandlers[RtfTags.TagFooterRightPages] = GroupTagHandlers.HeaderFooterHandler;

            //////Ignore footnotes
            groupHandlers[RtfTags.TagFootnote] = GroupTagHandlers.IgnoreGroupHandler;
        }

        [SuppressMessage("Microsoft.Usage", "CA1801:ReviewUnusedParameters", MessageId = "context"), System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Usage", "CA1801:ReviewUnusedParameters", MessageId = "group")]
        private static bool IgnoreGroupHandler(RtfGroup group, RtfImportContext context)
        {
            return false;
        }

        [SuppressMessage("Microsoft.Usage", "CA1801:ReviewUnusedParameters", MessageId = "context"), System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Usage", "CA1801:ReviewUnusedParameters", MessageId = "group")]
        private static bool VisitGroupHandler(RtfGroup group, RtfImportContext context)
        {
            return true;
        }

        private static bool FontTableHandler(RtfGroup group, RtfImportContext context)
        {
            Util.EnsureGroupDestination(group, RtfTags.TagFontTable);
            context.FontTable.FillFontTable(group);
            return false;
        }

        private static bool ColorTableHandler(RtfGroup group, RtfImportContext context)
        {
            Util.EnsureGroupDestination(group, RtfTags.TagColorTable);
            context.ColorTable.FillColorTable(group);

            return false;
        }

        private static bool GeneratorHandler(RtfGroup group, RtfImportContext context)
        {
            Util.EnsureGroupDestination(group, RtfTags.TagGenerator);

            if (group.Elements.Count == 3 && group.Elements[2] is RtfText)
            {
                context.Generator = ((RtfText)group.Elements[2]).Text;
            }

            return false;
        }

        private static bool DocumentInfoGroupHandler(RtfGroup group, RtfImportContext context)
        {
            Util.EnsureGroupDestination(group, RtfTags.TagInfo);

            context.DocumentInfo.FillDocumentInfo(group);

            return false;
        }

        private static bool HeaderFooterHandler(RtfGroup group, RtfImportContext context)
        {
            HeaderFooterBase headerFooter;
            switch (group.Destination)
            {
                case RtfTags.TagHeaderAllPages:
                    headerFooter = context.CurrentSection.Headers.Add(HeaderFooterType.Default);
                    break;

                case RtfTags.TagHeaderRightPages:
                    headerFooter = context.CurrentSection.Headers.Add(HeaderFooterType.Default);
                    break;

                case RtfTags.TagHeaderFirstPage:
                    headerFooter = context.CurrentSection.Headers.Add(HeaderFooterType.First);
                    break;

                case RtfTags.TagHeaderLeftPages:
                    headerFooter = context.CurrentSection.Headers.Add(HeaderFooterType.Even);
                    break;

                case RtfTags.TagFooterAllPages:
                    headerFooter = context.CurrentSection.Footers.Add(HeaderFooterType.Default);
                    break;

                case RtfTags.TagFooterRightPages:
                    headerFooter = context.CurrentSection.Footers.Add(HeaderFooterType.Default);
                    break;

                case RtfTags.TagFooterFirstPage:
                    headerFooter = context.CurrentSection.Footers.Add(HeaderFooterType.First);
                    break;

                case RtfTags.TagFooterLeftPages:
                    headerFooter = context.CurrentSection.Footers.Add(HeaderFooterType.Even);
                    break;

                default:
                    throw new RtfUnexpectedElementException("Unexpected header/footer group.", group.Destination);
            }

            RtfImportContext importContext = new RtfImportContext(context, headerFooter);
            RtfDocumentImporter importer = new RtfDocumentImporter(importContext);
            importer.ImportRoot(group);

            return false;
        }

        private static bool DefaultCharacterPropertiesHandler(RtfGroup group, RtfImportContext context)
        {
            Util.EnsureGroupDestination(group, RtfTags.TagDefaultCharacterProperties);
            context.StylesTable.ImportDefaultSpanStyle(group, context);
            return false;
        }

        private static bool DefaultParagraphGroupHandler(RtfGroup group, RtfImportContext context)
        {
            Util.EnsureGroupDestination(group, RtfTags.TagDefaultParagraphProperties);
            context.StylesTable.ImportDefaultParagraphStyle(group, context);
            return false;
        }

        private static bool StylesTableHandler(RtfGroup group, RtfImportContext context)
        {
            Util.EnsureGroupDestination(group, RtfTags.TagStyleSheet);
            context.StylesTable.ReadTable(group, context);
            return false;
        }

        private static bool ImageGroupHandler(RtfGroup group, RtfImportContext context)
        {
            Util.EnsureGroupDestination(group, RtfTags.TagPicture);

            ImageInline image = context.ImageBuilder.ReadImage(group);
            if (image != null)
            {
                context.AddInline(image);
            }

            return false;
        }

        private static bool ShapeGroupHandler(RtfGroup group, RtfImportContext context)
        {
            InlineBase inline = null;
            Watermark watermark = null;
            context.ShapeBuilder.ReadShape(group, out inline, out watermark);
            if (inline != null)
            {
                context.AddInline(inline);
            }

            if (watermark != null)
            {
                context.AddWatermark(watermark);
            }

            return false;
        }

        private static bool BookmarkStartHandler(RtfGroup group, RtfImportContext context)
        {
            RtfText bmkNameElement = group.Elements.FirstOrDefault(e => e is RtfText) as RtfText;
            if (bmkNameElement != null && !context.Bookmarks.ContainsKey(bmkNameElement.Text))
            {
                Bookmark bookmark = new Bookmark(context.Document, bmkNameElement.Text);
                context.AddInline(bookmark.BookmarkRangeStart);
                context.Bookmarks[bmkNameElement.Text] = bookmark;
            }

            return false;
        }

        private static bool BookmarkEndHandler(RtfGroup group, RtfImportContext context)
        {
            RtfText bmkNameElement = group.Elements.FirstOrDefault(e => e is RtfText) as RtfText;
            if (bmkNameElement != null)
            {
                Bookmark bookmark = null;
                if (context.Bookmarks.TryGetValue(bmkNameElement.Text, out bookmark))
                {
                    context.AddInline(bookmark.BookmarkRangeEnd);
                }
            }

            return false;
        }

        private static bool CommentStartHandler(RtfGroup group, RtfImportContext context)
        {
            RtfText text = group.Elements.FirstOrDefault(e => e is RtfText) as RtfText;
            if (text != null)
            {
                Comment comment = context.GetComment(text.Text);
                CommentRangeStart commentStart = comment.CommentRangeStart;

                if (commentStart.Paragraph != null)
                {
                    commentStart.Paragraph.Inlines.Remove(commentStart);
                }

                context.AddInline(comment.CommentRangeStart);
            }

            return false;
        }

        private static bool CommentEndHandler(RtfGroup group, RtfImportContext context)
        {
            RtfText text = group.Elements.FirstOrDefault(e => e is RtfText) as RtfText;
            if (text != null)
            {
                Comment comment = context.GetComment(text.Text);
                CommentRangeEnd commentEnd = comment.CommentRangeEnd;

                if (commentEnd.Paragraph != null)
                {
                    commentEnd.Paragraph.Inlines.Remove(commentEnd);
                }

                context.AddInline(comment.CommentRangeEnd);
            }

            return false;
        }

        private static bool CommentIdHandler(RtfGroup group, RtfImportContext context)
        {
            Debug.Assert(context.CurrentCommentHandler == null, "context.CurrentCommentHandler should be null");

            context.CurrentCommentHandler = new CommentImporter(context);

            context.CurrentCommentHandler.ImportCommentId(group);

            return false;
        }

        private static bool CommentAuthorHandler(RtfGroup group, RtfImportContext context)
        {
            Debug.Assert(context.CurrentCommentHandler != null, "context.CurrentCommentHandler should not be null");

            context.CurrentCommentHandler.ImportAuthor(group);

            return false;
        }

        private static bool CommentDateHandler(RtfGroup group, RtfImportContext context)
        {
            Debug.Assert(context.CurrentCommentHandler != null, "context.CurrentCommentHandler should not be null");

            context.CurrentCommentHandler.ImportDate(group);

            return false;
        }

        private static bool CommentDefinitionHandler(RtfGroup group, RtfImportContext context)
        {
            if (context.CurrentCommentHandler == null)
            {
                // When comment Initials are not present.
                context.CurrentCommentHandler = new CommentImporter(context);
            }

            CommentImporter commentImporter = context.CurrentCommentHandler;
            commentImporter.ImportCommentDefinition(group);

            context.CurrentCommentHandler = null;

            return false;
        }

        private static bool ListTableHandler(RtfGroup group, RtfImportContext context)
        {
            Util.EnsureGroupDestination(group, RtfTags.TagListTable);

            context.ListTable.ReadTable(group, context);

            return false;
        }

        private static bool ListOverrideTableHandler(RtfGroup group, RtfImportContext context)
        {
            Util.EnsureGroupDestination(group, RtfTags.TagListOverrideTable);

            context.ListOverrideTable.ReadTable(group, context);

            return false;
        }

        private static bool DocumentVariableHandler(RtfGroup group, RtfImportContext context)
        {
            RtfGroup[] groups = group.Elements.Where(e => e is RtfGroup).Select(e => (RtfGroup)e).ToArray();

            if (groups.Length >= 2)
            {
                string name = RtfHelper.GetGroupText(groups[0]);
                string value = RtfHelper.GetGroupText(groups[1]);
                if (!string.IsNullOrEmpty(name))
                {
                    context.Document.DocumentVariables[name] = value;
                }
            }

            return false;
        }
    }
}