﻿using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model
{
    /// <summary>
    /// Represents a bookmark.
    /// </summary>
    public class Bookmark
    {
        internal static readonly string HiddenPrefix = "_";

        private readonly RadFlowDocument document;
        private readonly BookmarkRangeStart bookmarkRangeStart;
        private readonly BookmarkRangeEnd bookmarkRangeEnd;
        private readonly string name;
        private readonly int? fromColumn;
        private readonly int? toColumn;

        /// <summary>
        /// Initializes a new instance of the <see cref="Bookmark" /> class.
        /// </summary>
        /// <param name="document">The document.</param>
        /// <param name="name">The name.</param>
        public Bookmark(RadFlowDocument document, string name)
            : this(document, name, null, null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Bookmark" /> class.
        /// </summary>
        /// <param name="document">The document.</param>
        /// <param name="name">The name.</param>
        /// <param name="fromColumn">From column.</param>
        /// <param name="toColumn">To column.</param>
        public Bookmark(RadFlowDocument document, string name, int? fromColumn, int? toColumn)
        {
            Guard.ThrowExceptionIfNull(document, "document");
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");
            Guard.ThrowExceptionIfContainsWhitespace(name, "name");

            this.document = document;

            this.bookmarkRangeStart = new BookmarkRangeStart(document, this);
            this.bookmarkRangeEnd = new BookmarkRangeEnd(document, this);

            this.name = name;
            this.fromColumn = fromColumn;
            this.toColumn = toColumn;
        }

        /// <summary>
        /// Gets the associated document.
        /// </summary>
        /// <value>The document.</value>
        public RadFlowDocument Document
        {
            get
            {
                return this.document;
            }
        }

        /// <summary>
        /// Gets the bookmark range start.
        /// </summary>
        /// <value>The bookmark range start.</value>
        public BookmarkRangeStart BookmarkRangeStart
        {
            get
            {
                return this.bookmarkRangeStart;
            }
        }

        /// <summary>
        /// Gets the bookmark range end.
        /// </summary>
        /// <value>The bookmark range end.</value>
        public BookmarkRangeEnd BookmarkRangeEnd
        {
            get
            {
                return this.bookmarkRangeEnd;
            }
        }

        /// <summary>
        /// Gets the name of the bookmark.
        /// </summary>
        /// <value>The name.</value>
        public string Name
        {
            get
            {
                return this.name;
            }
        }

        /// <summary>
        /// Gets the index of the first column in the current row which is part of the bookmark.
        /// </summary>
        /// <value>From column.</value>
        public int? FromColumn
        {
            get
            {
                return this.fromColumn;
            }
        }

        /// <summary>
        /// Gets the index of the last column in the current row which is part of the bookmark.
        /// </summary>
        /// <value>To column.</value>
        public int? ToColumn
        {
            get
            {
                return this.toColumn;
            }
        }

        internal static bool IsValidBookmarkName(string bookmarkName, bool allowHidden)
        {
            if (string.IsNullOrEmpty(bookmarkName))
            {
                return false;
            }

            if (allowHidden && bookmarkName.StartsWith(Bookmark.HiddenPrefix))
            {
                bookmarkName = bookmarkName.Substring(Bookmark.HiddenPrefix.Length);
            }

            if (!char.IsLetter(bookmarkName[0]))
            {
                return false;
            }

            foreach (char c in bookmarkName)
            {
                if (!char.IsLetterOrDigit(c))
                {
                    return false;
                }
            }

            return true;
        }
    }
}