﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.Model.BorderEvaluation.GridItems;
using Telerik.Windows.Documents.Flow.Model.Styles;

namespace Telerik.Windows.Documents.Flow.Model.BorderEvaluation
{
    internal class TableBorderGrid
    {
        private readonly TableBordersCalculator tableBorderCalculator;

        private List<TableRowHash> rowHashList;
        private int tableBordersHash;
        private bool hasCellCpacingHash;
        private FlowDirection tableFlowDirection;
        private bool isValid;

        public TableBorderGrid(Table table)
        {
            this.Table = table;
            this.isValid = false;
            this.tableBorderCalculator = new TableBordersCalculator(this.Table);
        }

        public Table Table { get; private set; }

        private TableBordersCalculator TableBorderCalculator
        {
            get
            {
                return this.tableBorderCalculator;
            }
        }

        private VerticalBorderGridItem[,] VerticalBorders
        {
            get
            {
                return this.TableBorderCalculator.VerticalBorders;
            }
        }

        private HorizontalBorderGridItem[,] HorisontalBorders
        {
            get
            {
                return this.TableBorderCalculator.HorisontalBorders;
            }
        }

        private CrossBorderGridItem[,] CrossBorders
        {
            get
            {
                return this.TableBorderCalculator.CrossBorders;
            }
        }

        public VerticalBorderGridItem GetLeftVerticalBorder(int rowIndex, int columnIndex)
        {
            return this.VerticalBorders[rowIndex, columnIndex];
        }

        public HorizontalBorderGridItem GetTopHorizontalBorder(int rowIndex, int columnIndex)
        {
            return this.HorisontalBorders[rowIndex, columnIndex];
        }

        public CrossBorderGridItem GetTopLeftCrossBorder(int rowIndex, int columnIndex)
        {
            return this.CrossBorders[rowIndex, columnIndex];
        }

        public bool HasChanges()
        {
            if (this.tableFlowDirection != this.Table.FlowDirection)
            {
                return true;
            }

            if (this.hasCellCpacingHash != this.Table.HasCellSpacing)
            {
                return true;
            }

            if (this.tableBordersHash != this.Table.Borders.GetHashCode())
            {
                return true;
            }

            if (this.rowHashList == null)
            {
                return true;
            }

            using (IEnumerator<TableRow> rowIterator = this.Table.Rows.GetEnumerator())
            using (IEnumerator<TableRowHash> rowHashIterator = this.rowHashList.GetEnumerator())
            {
                bool rowIteratorMoved = rowIterator.MoveNext();
                bool rowHashIteratorMoved = rowHashIterator.MoveNext();
                while (rowIteratorMoved && rowHashIteratorMoved)
                {
                    if (rowIterator.Current != rowHashIterator.Current.Row)
                    {
                        return true;
                    }

                    using (IEnumerator<TableCell> cellIterator = rowIterator.Current.Cells.GetEnumerator())
                    using (IEnumerator<TableCellHash> cellHashIterator = rowHashIterator.Current.CellHashList.GetEnumerator())
                    {
                        bool cellIteratorMoved = cellIterator.MoveNext();
                        bool cellHashIteratorMoved = cellHashIterator.MoveNext();
                        while (cellIteratorMoved && cellHashIteratorMoved)
                        {
                            if (cellIterator.Current.RowSpan != cellHashIterator.Current.RowSpan ||
                                cellIterator.Current.ColumnSpan != cellHashIterator.Current.ColumnSpan ||
                                cellIterator.Current.Borders.GetHashCode() != cellHashIterator.Current.BordersHash)
                            {
                                return true;
                            }

                            cellIteratorMoved = cellIterator.MoveNext();
                            cellHashIteratorMoved = cellHashIterator.MoveNext();
                        }

                        if (cellIteratorMoved || cellHashIteratorMoved)
                        {
                            //one of the lists is longer
                            return true;
                        }
                    }

                    rowIteratorMoved = rowIterator.MoveNext();
                    rowHashIteratorMoved = rowHashIterator.MoveNext();
                }

                if (rowIteratorMoved || rowHashIteratorMoved)
                {
                    //one of the lists is longer
                    return true;
                }
            }

            return false;
        }

        public void CheckForChanges()
        {
            if (this.HasChanges())
            {
                this.Invalidate();
            }
        }

        public void Invalidate()
        {
            this.isValid = false;
        }

        internal void AssureValid()
        {
            if (!this.isValid)
            {
                this.RecalculateBordersGrid();
            }
        }

        private void RecalculateBordersGrid()
        {
            this.ReCalculateHash();

            this.TableBorderCalculator.ClearCollections();

            if (!this.Table.HasCellSpacing)
            {
                this.TableBorderCalculator.RecalculateBorders();
            }

            this.isValid = true;
        }

        private void ReCalculateHash()
        {
            this.hasCellCpacingHash = this.Table.HasCellSpacing;
            this.tableFlowDirection = this.Table.FlowDirection;

            this.tableBordersHash = this.Table.Borders.GetHashCode();

            this.rowHashList = new List<TableRowHash>();

            foreach (TableRow row in this.Table.Rows)
            {
                TableRowHash rowHash = new TableRowHash
                {
                    Row = row,
                    CellHashList = new List<TableCellHash>()
                };

                foreach (TableCell cell in row.Cells)
                {
                    rowHash.CellHashList.Add(new TableCellHash
                    {
                        BordersHash = cell.Borders.GetHashCode(),
                        RowSpan = cell.RowSpan,
                        ColumnSpan = cell.ColumnSpan
                    });
                }

                this.rowHashList.Add(rowHash);
            }
        }
    }
}