﻿using Telerik.Windows.Documents.Flow.Model.BorderEvaluation.GridItems;
using Telerik.Windows.Documents.Flow.Model.Styles;

namespace Telerik.Windows.Documents.Flow.Model.BorderEvaluation
{
    internal class TableBordersCalculator
    {
        private readonly Table table;

        internal TableBordersCalculator(Table table)
        {
            this.table = table;
        }

        internal VerticalBorderGridItem[,] VerticalBorders { get; private set; }

        internal HorizontalBorderGridItem[,] HorisontalBorders { get; private set; }

        internal CrossBorderGridItem[,] CrossBorders { get; private set; }

        internal Table Table
        {
            get
            {
                return this.table;
            }
        }

        internal virtual void RecalculateBorders()
        {
            this.InitializeNoCellCpacing();

            this.CalculateHorizontalAndVerticalBorders();

            this.CalculateHorizontalAndVerticalInheritedBorders();

            this.MergeContinuesBorders();

            this.CalculateCrossBorders();
        }

        internal void ClearCollections()
        {
            this.VerticalBorders = null;
            this.HorisontalBorders = null;
            this.CrossBorders = null;
        }

        private void InitializeNoCellCpacing()
        {
            int gridRowsCount = this.Table.Rows.Count;
            int gridColumnsCount = this.Table.GridColumnsCount;
            this.VerticalBorders = new VerticalBorderGridItem[gridRowsCount, gridColumnsCount + 1];
            for (int i = 0; i < this.Table.Rows.Count; i++)
            {
                for (int j = 0; j < gridColumnsCount + 1; j++)
                {
                    this.VerticalBorders[i, j] = new VerticalBorderGridItem();
                    this.VerticalBorders[i, j].ColumnIndex = j;
                }
            }

            this.HorisontalBorders = new HorizontalBorderGridItem[gridRowsCount + 1, gridColumnsCount];
            for (int i = 0; i < gridRowsCount + 1; i++)
            {
                for (int j = 0; j < gridColumnsCount; j++)
                {
                    this.HorisontalBorders[i, j] = new HorizontalBorderGridItem();
                    this.HorisontalBorders[i, j].RowIndex = i;
                }
            }

            this.CrossBorders = new CrossBorderGridItem[gridRowsCount + 1, gridColumnsCount + 1];
            for (int i = 0; i < gridRowsCount + 1; i++)
            {
                for (int j = 0; j < gridColumnsCount + 1; j++)
                {
                    this.CrossBorders[i, j] = new CrossBorderGridItem();
                    this.CrossBorders[i, j].RowIndex = i;
                    this.CrossBorders[i, j].ColumnIndex = j;
                }
            }
        }

        private void CalculateHorizontalAndVerticalBorders()
        {
            foreach (TableRow row in this.Table.Rows)
            {
                foreach (TableCell cell in row.Cells)
                {
                    // RowSpan
                    for (int rowIndex = cell.GridRowIndex; rowIndex < cell.GridRowIndex + cell.RowSpan; rowIndex++)
                    {
                        // left
                        this.VerticalBorders[rowIndex, cell.GridColumnIndex].IsInsideCell = false;
                        if (this.VerticalBorders[rowIndex, cell.GridColumnIndex].Border.CompareTo(cell.Borders.Left) < 0)
                        {
                            this.VerticalBorders[rowIndex, cell.GridColumnIndex].Border = cell.Borders.Left;
                        }

                        // right  
                        this.VerticalBorders[rowIndex, cell.GridColumnIndex + cell.ColumnSpan].IsInsideCell = false;
                        if (this.VerticalBorders[rowIndex, cell.GridColumnIndex + cell.ColumnSpan].Border.CompareTo(cell.Borders.Right) < 0)
                        {
                            this.VerticalBorders[rowIndex, cell.GridColumnIndex + cell.ColumnSpan].Border = cell.Borders.Right;
                        }
                    }

                    // ColumnSpan
                    for (int colIndex = cell.GridColumnIndex; colIndex < cell.GridColumnIndex + cell.ColumnSpan; colIndex++)
                    {
                        // top
                        this.HorisontalBorders[cell.GridRowIndex, colIndex].IsInsideCell = false;
                        if (this.HorisontalBorders[cell.GridRowIndex, colIndex].Border.CompareTo(cell.Borders.Top) < 0)
                        {
                            this.HorisontalBorders[cell.GridRowIndex, colIndex].Border = cell.Borders.Top;
                        }

                        // bottom
                        this.HorisontalBorders[cell.GridRowIndex + cell.RowSpan, colIndex].IsInsideCell = false;
                        if (this.HorisontalBorders[cell.GridRowIndex + cell.RowSpan, colIndex].Border.CompareTo(cell.Borders.Bottom) < 0)
                        {
                            this.HorisontalBorders[cell.GridRowIndex + cell.RowSpan, colIndex].Border = cell.Borders.Bottom;
                        }
                    }
                }
            }
        }

        private void CalculateHorizontalAndVerticalInheritedBorders()
        {
            TableBorders tableBorders = this.Table.Borders;
            Border noneBorder = new Border(BorderStyle.Inherit);
            int gridRowsCount = this.Table.GridRowsCount;
            int gridColumnsCount = this.Table.GridColumnsCount;

            for (int row = 0; row < gridRowsCount; row++)
            {
                // left borders
                if (this.VerticalBorders[row, 0].Border.Style == BorderStyle.Inherit)
                {
                    this.VerticalBorders[row, 0].Border = tableBorders.Left;
                }

                // right borders
                if (this.VerticalBorders[row, gridColumnsCount].Border.Style == BorderStyle.Inherit)
                {
                    this.VerticalBorders[row, gridColumnsCount].Border = tableBorders.Right;
                }

                // inner vertical borders
                for (int col = 1; col < gridColumnsCount; col++)
                {
                    if (this.VerticalBorders[row, col].IsInsideCell)
                    {
                        this.VerticalBorders[row, col].Border = noneBorder;
                    }
                    else if (this.VerticalBorders[row, col].Border.Style == BorderStyle.Inherit)
                    {
                        this.VerticalBorders[row, col].Border = tableBorders.InsideVertical;
                    }
                }
            }

            for (int col = 0; col < gridColumnsCount; col++)
            {
                // top borders
                if (this.HorisontalBorders[0, col].Border.Style == BorderStyle.Inherit)
                {
                    this.HorisontalBorders[0, col].Border = tableBorders.Top;
                }

                // right borders
                if (this.HorisontalBorders[gridRowsCount, col].Border.Style == BorderStyle.Inherit)
                {
                    this.HorisontalBorders[gridRowsCount, col].Border = tableBorders.Bottom;
                }

                // inner horizontal border
                for (int row = 1; row < gridRowsCount; row++)
                {
                    if (this.HorisontalBorders[row, col].IsInsideCell)
                    {
                        this.HorisontalBorders[row, col].Border = noneBorder;
                    }
                    else if (this.HorisontalBorders[row, col].Border.Style == BorderStyle.Inherit)
                    {
                        this.HorisontalBorders[row, col].Border = tableBorders.InsideHorizontal;
                    }
                }
            }
        }

        private void MergeContinuesBorders()
        {
            // Horizontal
            for (int row = 0; row < this.Table.GridRowsCount + 1; row++)
            {
                if (this.Table.GridColumnsCount > 0)
                {
                    HorizontalBorderGridItem current = this.HorisontalBorders[row, 0];
                    current.ColumnIndex = 0;
                    current.ColumnSpan = 1;

                    for (int col = 1; col < this.Table.GridColumnsCount; col++)
                    {
                        if (this.HorisontalBorders[row, col].Border == current.Border)
                        {
                            this.HorisontalBorders[row, col] = current;
                            current.ColumnSpan++;
                        }
                        else
                        {
                            current = this.HorisontalBorders[row, col];
                            current.ColumnIndex = col;
                            current.ColumnSpan = 1;
                        }
                    }
                }
            }

            // Vertical
            for (int col = 0; col < this.Table.GridColumnsCount + 1; col++)
            {
                if (this.Table.GridRowsCount > 0)
                {
                    VerticalBorderGridItem current = this.VerticalBorders[0, col];
                    current.RowIndex = 0;
                    current.RowSpan = 1;

                    for (int row = 1; row < this.Table.GridRowsCount; row++)
                    {
                        if (this.VerticalBorders[row, col].Border == current.Border)
                        {
                            this.VerticalBorders[row, col] = current;
                            current.RowSpan++;
                        }
                        else
                        {
                            current = this.VerticalBorders[row, col];
                            current.RowIndex = row;
                            current.RowSpan = 1;
                        }
                    }
                }
            }
        }

        private void CalculateCrossBorders()
        {
            for (int r = 0; r < this.Table.GridRowsCount + 1; r++)
            {
                for (int c = 0; c < this.Table.GridColumnsCount + 1; c++)
                {
                    Border top = (r > 0) ? this.VerticalBorders[r - 1, c].Border : null;
                    Border bot = (r < this.Table.GridRowsCount) ? this.VerticalBorders[r, c].Border : null;
                    Border verticalMax = Border.Max(top, bot);

                    Border left = (c > 0) ? this.HorisontalBorders[r, c - 1].Border : null;
                    Border right = (c < this.Table.GridColumnsCount) ? this.HorisontalBorders[r, c].Border : null;
                    Border horizontalMax = Border.Max(left, right);

                    Border max = Border.Max(verticalMax, horizontalMax);

                    if (horizontalMax != null)
                    {
                        this.CrossBorders[r, c].VerticalSize = horizontalMax.Thickness;
                    }

                    if (verticalMax != null)
                    {
                        this.CrossBorders[r, c].HorizontalSize = verticalMax.Thickness;
                    }

                    this.CrossBorders[r, c].Border = max;

                    CrossBorderDirection direction = CrossBorderDirection.None;
                    if (max == left)
                    {
                        if (this.Table.FlowDirection == FlowDirection.LeftToRight)
                        {
                            direction |= CrossBorderDirection.Left;
                        }
                        else
                        {
                            direction |= CrossBorderDirection.Right;
                        }
                    }
                    if (max == right)
                    {
                        if (this.Table.FlowDirection == FlowDirection.LeftToRight)
                        {
                            direction |= CrossBorderDirection.Right;
                        }
                        else
                        {
                            direction |= CrossBorderDirection.Left;
                        }
                    }
                    if (max == top)
                    {
                        direction |= CrossBorderDirection.Top;
                    }
                    if (max == bot)
                    {
                        direction |= CrossBorderDirection.Bottom;
                    }
                    this.CrossBorders[r, c].CrossBorderDirection = direction;
                }
            }
        }
    }
}