﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.Model.Styles;

namespace Telerik.Windows.Documents.Flow.Model.Cloning
{
    internal class StyleRepositoryMerger
    {
        private const string StyleNameMergeSuffix = "_1";

        private readonly StyleRepository targetStyleRepository;
        private readonly StyleRepository sourceStyleRepository;
        private readonly StyleMergeEqualityComparer styleComparer;
        private List<Style> addedStyles;

        public StyleRepositoryMerger(StyleRepository target, StyleRepository source)
        {
            this.targetStyleRepository = target;
            this.sourceStyleRepository = source;

            this.styleComparer = new StyleMergeEqualityComparer();
        }

        private enum AddStyleAction
        {
            None,
            Add,
            RenameAndAdd
        }

        public void Merge(CloneContext cloneContext)
        {
            this.addedStyles = new List<Style>();

            foreach (Style sourceStyle in this.sourceStyleRepository.Styles)
            {
                AddStyleAction addStyleAction = this.ComputeAddStyleAction(sourceStyle, cloneContext.MergeOptions.ConflictingStylesResolutionMode);
                Style clonedStyle = null;
                switch (addStyleAction)
                {
                    case AddStyleAction.Add:
                        clonedStyle = sourceStyle.Clone();
                        this.AddStyle(clonedStyle);
                        break;

                    case AddStyleAction.RenameAndAdd:
                        clonedStyle = sourceStyle.Clone();
                        clonedStyle.IsCustom = true;
                        this.RenameStyle(clonedStyle, cloneContext);
                        this.AddStyle(clonedStyle);
                        break;

                    case AddStyleAction.None:
                    default:
                        break;
                }

                if (clonedStyle != null && (clonedStyle.StyleType == StyleType.Numbering || clonedStyle.StyleType == StyleType.Paragraph))
                {
                    StyleRepositoryMerger.AssignOldListIdToNewStyleId(clonedStyle, cloneContext);
                }
            }

            this.UpdateStyleLinks(cloneContext);
        }

        private static void AssignOldListIdToNewStyleId(Style style, CloneContext cloneContext)
        {
            if (style.ParagraphProperties.ListId.HasLocalValue)
            {
                if (cloneContext.OldListsToStyles.ContainsKey(style.ParagraphProperties.ListId.LocalValue.Value))
                {
                    cloneContext.OldListsToStyles[style.ParagraphProperties.ListId.LocalValue.Value].Add(style.Id);
                }
                else
                {
                    List<string> styleIds = new List<string>();
                    styleIds.Add(style.Id);
                    cloneContext.OldListsToStyles.Add(style.ParagraphProperties.ListId.LocalValue.Value, styleIds);
                }
            }
        }

        private void UpdateStyleLinks(CloneContext cloneContext)
        {
            foreach (Style addedStyle in this.addedStyles)
            {
                if (addedStyle.BasedOnStyleId != null && cloneContext.RenamedStyles.ContainsKey(addedStyle.BasedOnStyleId))
                {
                    addedStyle.BasedOnStyleId = cloneContext.RenamedStyles[addedStyle.BasedOnStyleId];
                }

                if (addedStyle.LinkedStyleId != null && cloneContext.RenamedStyles.ContainsKey(addedStyle.LinkedStyleId))
                {
                    addedStyle.LinkedStyleId = cloneContext.RenamedStyles[addedStyle.LinkedStyleId];
                }

                if (addedStyle.NextStyleId != null && cloneContext.RenamedStyles.ContainsKey(addedStyle.NextStyleId))
                {
                    addedStyle.NextStyleId = cloneContext.RenamedStyles[addedStyle.NextStyleId];
                }
            }
        }

        private void AddStyle(Style style)
        {
            if (style.IsDefault)
            {
                style.IsDefault = false;
            }

            this.targetStyleRepository.Add(style);
            this.addedStyles.Add(style);
        }

        private void RenameStyle(Style style, CloneContext cloneContext)
        {
            string oldId = style.Id;
            string newId = this.ChooseRenamedStyledId(oldId);

            style.Id = newId;

            cloneContext.RenamedStyles.Add(oldId, newId);
        }

        private string ChooseRenamedStyledId(string oldId)
        {
            string newId;

            do
            {
                newId = oldId + StyleNameMergeSuffix;
                oldId = newId;
            }
            while (this.targetStyleRepository.Contains(newId));

            return newId;
        }

        private AddStyleAction ComputeAddStyleAction(Style style, ConflictingStylesResolutionMode conflictStyleResolutionMode)
        {
            if (!this.targetStyleRepository.Contains(style.Id))
            {
                return AddStyleAction.Add;
            }

            Style conflictedStyle = this.targetStyleRepository.GetStyle(style.Id);
            switch (conflictStyleResolutionMode)
            {
                case ConflictingStylesResolutionMode.UseTargetStyle:
                    if (conflictedStyle.StyleType == style.StyleType)
                    {
                        return AddStyleAction.None;
                    }
                    else
                    {
                        return AddStyleAction.RenameAndAdd;
                    }
                case ConflictingStylesResolutionMode.RenameSourceStyle:
                    if (this.styleComparer.Equals(conflictedStyle, style))
                    {
                        return AddStyleAction.None;
                    }
                    else
                    {
                        return AddStyleAction.RenameAndAdd;
                    }
                default:
                    throw new ArgumentException("Unsupported ConflictStyleResolutionMode.");
            }
        }
    }
}
