﻿using System;
using Telerik.Windows.Documents.Core.Data;
using Telerik.Windows.Documents.Flow.Model.Cloning;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Collections
{
    /// <summary>
    /// Defines a collection of document elements.
    /// </summary>
    /// <typeparam name="T">The type of the document elements in the collection.</typeparam>
    /// <typeparam name="TOwner">The type of the owner.</typeparam>
    public class DocumentElementCollection<T, TOwner> : DocumentElementCollectionBase<T, TOwner>
        where T : DocumentElementBase
        where TOwner : DocumentElementBase
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="DocumentElementCollection{T, TOwner}"/> class.
        /// </summary>
        /// <param name="owner">The owner.</param>
        internal DocumentElementCollection(TOwner owner) : base(owner)
        {
        }

        internal void AddClonedChildrenFrom(DocumentElementCollection<T, TOwner> fromCollection, CloneContext cloneContext)
        {
            Guard.ThrowExceptionIfNull(cloneContext, "cloneContext");
            Guard.ThrowExceptionIfNull(fromCollection, "fromCollection");

            foreach (T documentElement in fromCollection)
            {
                T clonedElement = (T)documentElement.CloneCore(cloneContext);
                if (clonedElement != null)
                {
                    this.Add(clonedElement);

                    documentElement.OnAfterCloneCore(cloneContext, clonedElement);
                }
            }
        }

        /// <summary>
        /// Sets the parent of the document element.
        /// </summary>
        /// <param name="item">The document element.</param>
        /// <param name="parent">The parent which should be set to the document element.</param>
        protected override void SetParent(T item, TOwner parent)
        {
            Guard.ThrowExceptionIfNull(item, "item");

            item.SetParent(parent);
        }

        /// <summary>
        /// Verifies the validity of the document element before it is inserted in the collection. 
        /// </summary>
        /// <param name="item">The item.</param>
        protected override void VerifyDocumentElementOnInsert(T item)
        {
            Guard.ThrowExceptionIfNull(item, "item");

            if (item.Document != this.Owner.Document)
            {
                throw new ArgumentException("The document element is associated with another document.", "item");
            }

            if (item.Parent != null)
            {
                throw new ArgumentException("The document element is already associated with a parent.", "item");
            }
        }
    }
}