﻿using System;
using System.Collections;
using System.Collections.Generic;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Collections
{
    /// <summary>
    /// Represents document variables collection.
    /// </summary>
    public class DocumentVariableCollection : IEnumerable<KeyValuePair<string, string>>
    {
        private readonly Dictionary<string, string> documentVariables;

        internal DocumentVariableCollection()
        {
            StringComparer comparer = StringComparer.OrdinalIgnoreCase;
            this.documentVariables = new Dictionary<string, string>(comparer);
        }

        /// <summary>
        /// Gets the count of document variables.
        /// </summary>
        /// <value>The count.</value>
        public int Count
        {
            get
            {
                return this.documentVariables.Count;
            }
        }

        /// <summary>
        /// Gets or sets the <see cref="string" /> with the specified name.
        /// </summary>
        /// <value>The document variable name</value>
        public string this[string name]
        {
            get
            {
                string value;
                if (!this.documentVariables.TryGetValue(name, out value))
                {
                    throw new KeyNotFoundException("The collection does not contains the specified variable name.");
                }

                return value;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "value");

                this.documentVariables[name] = value;
            }
        }

        /// <summary>
        /// Adds the specified document variable.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="value">The value.</param>
        public void Add(string name, string value)
        {
            this.documentVariables[name] = value;
        }

        /// <summary>
        /// Removes the document variable by a specified name.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <returns></returns>
        public bool Remove(string name)
        {
            return this.documentVariables.Remove(name);
        }

        /// <summary>
        /// Determines whether the collection contains a document variable with a specified name.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <returns></returns>
        public bool Contains(string name)
        {
            return this.documentVariables.ContainsKey(name);
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<KeyValuePair<string, string>> GetEnumerator()
        {
            return this.documentVariables.GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.GetEnumerator();
        }

        internal bool TryGetValue(string name, out string value)
        {
            return this.documentVariables.TryGetValue(name, out value);
        }

        internal void AddClonedDocumentVariablesFrom(DocumentVariableCollection fromDocumentVariableCollection)
        {
            foreach (KeyValuePair<string, string> documentVariable in fromDocumentVariableCollection)
            {
                this.documentVariables.Add(documentVariable.Key, documentVariable.Value);
            }
        }
    }
}