﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.Model.Cloning;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model
{
    /// <summary>
    /// Provides base class for flow document elements.
    /// </summary>
    public abstract partial class DocumentElementBase
    {
        private static readonly Type ParagraphType = typeof(Paragraph);
        private static readonly Type InlineType = typeof(InlineBase);

        private readonly RadFlowDocument document;

        private DocumentElementBase parent;

        internal DocumentElementBase()
        {
            if (this.GetType() != typeof(RadFlowDocument))
            {
                throw new InvalidOperationException("This constructor should be used for RadFlowDocument only.");
            }
        }

        internal DocumentElementBase(RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");

            this.document = document;
        }

        /// <summary>
        /// Gets the document to which this document element is associated.
        /// </summary>
        /// <value>
        /// The document.
        /// </value>
        public virtual RadFlowDocument Document
        {
            get
            {
                return this.document;
            }
        }

        /// <summary>
        /// Gets the parent of the element.
        /// </summary>
        /// <value>The parent.</value>
        public DocumentElementBase Parent
        {
            get
            {
                return this.parent;
            }
        }

        internal abstract IEnumerable<DocumentElementBase> Children { get; }

        internal virtual IEnumerable<DocumentElementBase> ContentChildren
        {
            get
            {
                return this.Children;
            }
        }

        internal abstract DocumentElementType Type { get; }

        /// <summary>
        /// Enumerates all child elements of given type.
        /// </summary>
        /// <typeparam name="T">The type.</typeparam>
        /// <returns></returns>
        public IEnumerable<T> EnumerateChildrenOfType<T>()
            where T : DocumentElementBase
        {
            return this.EnumerateChildrenOfTypeInternal<T>(this.Children);
        }

        /// <summary>
        /// Enumerates all direct child elements of a given type.
        /// <para>For example Headers/Footers are indirect children of a <see cref="Section"/> element and 
        /// Comments are indirect children of a <see cref="RadFlowDocument"/> element.</para>
        /// </summary>
        /// <typeparam name="T">The type of the T.</typeparam>
        /// <returns></returns>
        internal IEnumerable<T> EnumerateContentChildrenOfType<T>()
            where T : DocumentElementBase
        {
            return this.EnumerateChildrenOfTypeInternal<T>(this.ContentChildren);
        }

        /// <summary>
        /// Sets the parent.
        /// </summary>
        /// <param name="newParent">The new parent.</param>
        internal void SetParent(DocumentElementBase newParent)
        {
            if (this.Parent == newParent)
            {
                return;
            }

            DocumentElementBase oldParent = this.Parent;
            this.parent = newParent;
            if (oldParent != null)
            {
                oldParent.OnChildRemoved(this);
            }

            if (newParent != null)
            {
                newParent.OnChildAdded(this);
            }
        }

        /// <summary>
        /// Enumerates all parent elements of given type.
        /// </summary>
        /// <typeparam name="T">The type.</typeparam>
        /// <returns></returns>
        internal IEnumerable<T> EnumerateParentsOfType<T>() where T : DocumentElementBase
        {
            DocumentElementBase parent = this;
            while (parent != null)
            {
                if (parent is T)
                {
                    yield return (T)parent;
                }

                parent = parent.Parent;
            }
        }

        internal abstract DocumentElementBase CloneCore(CloneContext cloneContext);

        /// <summary>
        /// This methods can be override in derived class and if override should be invoked after CloneCore call.  
        /// </summary>
        /// <param name="cloneContext"></param>
        /// <param name="clonedElement"></param>
        internal virtual void OnAfterCloneCore(CloneContext cloneContext, DocumentElementBase clonedElement)
        {
        }

        /// <summary>
        /// Called when child element is added.
        /// </summary>
        /// <param name="child">The child element.</param>
        protected virtual void OnChildAdded(DocumentElementBase child)
        {
        }

        /// <summary>
        /// Called when child element is removed.
        /// </summary>
        /// <param name="child">The child element.</param>
        protected virtual void OnChildRemoved(DocumentElementBase child)
        {
        }

        private static IEnumerable<T> EnumerateChildrenOrSelfOfType<T>(DocumentElementBase element) where T : DocumentElementBase
        {
            if (element is T)
            {
                yield return (T)element;
            }

            if (CanHaveChildOfType(element.GetType(), typeof(T)))
            {
                foreach (DocumentElementBase child in element.Children)
                {
                    foreach (T subChild in EnumerateChildrenOrSelfOfType<T>(child))
                    {
                        yield return subChild;
                    }
                }
            }
        }

        private static bool CanHaveChildOfType(Type parentType, Type childType)
        {
            if (ParagraphType.IsAssignableFrom(parentType))
            {
                return InlineType.IsAssignableFrom(childType) || childType.IsAssignableFrom(InlineType);
            }

            return true;
        }

        private IEnumerable<T> EnumerateChildrenOfTypeInternal<T>(IEnumerable<DocumentElementBase> children)
            where T : DocumentElementBase
        {
            if (CanHaveChildOfType(this.GetType(), typeof(T)))
            {
                foreach (DocumentElementBase child in children)
                {
                    foreach (T subChild in EnumerateChildrenOrSelfOfType<T>(child))
                    {
                        yield return subChild;
                    }
                }
            }
        }
    }
}
