﻿using Telerik.Windows.Documents.Flow.Model.Cloning;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model
{
    /// <summary>
    /// Defines utility class used for importing document elements from a source <see cref="Telerik.Windows.Documents.Flow.Model.RadFlowDocument" /> into
    /// target <see cref="Telerik.Windows.Documents.Flow.Model.RadFlowDocument" />.
    /// </summary>
    public class DocumentElementImporter
    {
        private readonly RadFlowDocument targetDocument;
        private readonly RadFlowDocument sourceDocument;
        private readonly ConflictingStylesResolutionMode conflictingStylesResolutionMode;

        private bool isInitialized;
        private CloneContext cloneContext;

        /// <summary>
        /// Initializes a new instance of the <see cref="DocumentElementImporter" /> class.
        /// </summary>
        /// <param name="targetDocument">The target document.</param>
        /// <param name="sourceDocument">The source document.</param>
        /// <param name="conflictingStylesResolutionMode">The conflicting styles resolution mode.</param>
        public DocumentElementImporter(RadFlowDocument targetDocument, RadFlowDocument sourceDocument, ConflictingStylesResolutionMode conflictingStylesResolutionMode)
        {
            Guard.ThrowExceptionIfNull(targetDocument, "targetDocument");
            Guard.ThrowExceptionIfNull(sourceDocument, "sourceDocument");

            this.targetDocument = targetDocument;
            this.sourceDocument = sourceDocument;
            this.conflictingStylesResolutionMode = conflictingStylesResolutionMode;
        }

        /// <summary>
        /// Imports the specified document element from the source document. 
        /// The target document is adjusted in order the import to be correct.
        /// </summary>
        /// <typeparam name="T">The type.</typeparam>
        /// <param name="sourceElement">The document element from the source document which shall be imported to the target document.</param>
        /// <returns>The element to import.</returns>
        public T Import<T>(T sourceElement)
            where T : DocumentElementBase
        {
            Guard.ThrowExceptionIfNull(sourceElement, "sourceElement");
            Guard.ThrowExceptionIfFalse(sourceElement.Document == this.sourceDocument, "The element does not belong to the source document.");

            if (!this.isInitialized)
            {
                this.Initialize();
            }

            T result = (T)sourceElement.CloneCore(this.cloneContext);

            if (result != null)
            {
                sourceElement.OnAfterCloneCore(this.cloneContext, result);
            }

            return result;
        }

        private void Initialize()
        {
            MergeOptions mergeOptions = new MergeOptions()
            {
                ConflictingStylesResolutionMode = this.conflictingStylesResolutionMode
            };

            this.cloneContext = new CloneContext(this.targetDocument)
            {
                MergeOptions = mergeOptions
            };

            this.targetDocument.MergeWithoutChildren(this.sourceDocument, this.cloneContext);

            this.isInitialized = true;
        }
    }
}