using System.Linq;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Editing
{
    internal class EditorPosition
    {
        private readonly RadFlowDocument document;

        public EditorPosition(RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");

            this.document = document;
        }

        public Paragraph CurrentParagraph { get; private set; }

        public Table CurrentTable { get; private set; }

        public int InlineIndex { get; set; }

        public bool IsAtTableEnd
        {
            get
            {
                return this.CurrentTable != null;
            }
        }

        public bool IsInitialized
        {
            get
            {
                return this.CurrentBlock != null;
            }
        }

        public BlockBase CurrentBlock
        {
            get
            {
                if (this.IsAtTableEnd)
                {
                    return this.CurrentTable;
                }
                else
                {
                    return this.CurrentParagraph;
                }
            }
        }

        public RadFlowDocument Document
        {
            get
            {
                return this.document;
            }
        }

        public bool IsAtParagraphEnd()
        {
            return this.CurrentParagraph != null &&
                   this.CurrentParagraph.Inlines.Count == this.InlineIndex;
        }

        public bool IsInTableCell()
        {
            return this.CurrentParagraph != null &&
                   this.CurrentParagraph.EnumerateParentsOfType<TableCell>().FirstOrDefault() != null;
        }

        public void MoveToParagraphStart(Paragraph paragraph)
        {
            Guard.ThrowExceptionIfNull(paragraph, "paragraph");
            Guard.ThrowExceptionIfFalse(paragraph.Document == this.document, "document");

            this.MoveToParagraphInternal(paragraph, 0);
        }

        public void MoveToInlineEnd(InlineBase inline)
        {
            Guard.ThrowExceptionIfNull(inline, "inline");
            Guard.ThrowExceptionIfFalse(inline.Document == this.document, "document");

            this.MoveToParagraphInternal(inline.Paragraph, inline.Paragraph.Inlines.IndexOf(inline) + 1);
        }

        public void MoveToInlineStart(InlineBase inline)
        {
            Guard.ThrowExceptionIfNull(inline, "inline");
            Guard.ThrowExceptionIfFalse(inline.Document == this.document, "document");

            this.MoveToParagraphInternal(inline.Paragraph, inline.Paragraph.Inlines.IndexOf(inline));
        }

        public void MoveAfterTable(Table table)
        {
            Guard.ThrowExceptionIfNull(table, "table");
            Guard.ThrowExceptionIfFalse(table.Document == this.document, "document");

            this.CurrentTable = table;
            this.CurrentParagraph = null;
            this.InlineIndex = 0;
        }

        private void MoveToParagraphInternal(Paragraph paragraph, int index)
        {
            this.CurrentParagraph = paragraph;
            this.InlineIndex = index;
            this.CurrentTable = null;
        }
    }
}