﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Flow.Model.Cloning;
using Telerik.Windows.Documents.Flow.Model.Collections;
using Telerik.Windows.Documents.Flow.Model.Styles;

namespace Telerik.Windows.Documents.Flow.Model.Editing
{
    public partial class RadFlowDocumentEditor
    {
        private class InsertDocumentEditor
        {
            private readonly RadFlowDocumentEditor editor;

            private Paragraph insertTargetParagraphFormatting;
            private Section insertTargetSection;
            private bool isInsertPositionInParagraph;
            private bool hasInsertedMultipleParagraphs;
            private int clonedSectionIndex;
            private int initialInsertSectionIndex;

            public InsertDocumentEditor(RadFlowDocumentEditor editor)
            {
                this.editor = editor;
            }

            private ParagraphProperties FormattingOfInsertTargetParagraph
            {
                get
                {
                    return this.insertTargetParagraphFormatting.Properties;
                }
            }

            public void InsertDocument(RadFlowDocument sourceDocument, InsertDocumentOptions importOptions)
            {
                bool isTargetDocumentEmptyInTheBeginning = IsDocumentEmpty(this.editor.Document);

                this.editor.AssurePositionInitialized();

                InsertDocumentContext insertContext = new InsertDocumentContext(this.editor.Document, sourceDocument, importOptions ?? new InsertDocumentOptions());

                this.BeginInsertDocument(sourceDocument, insertContext);
                this.InsertDocumentInternal(sourceDocument, insertContext, isTargetDocumentEmptyInTheBeginning);
                this.EndInsertDocument();
            }
            
            private static void SwapSectionPropertiesAndHeaderFooters(Section first, Section second, CloneContext cloneContext)
            {
                Section tempSwappingSection = first.ClonePropertiesAndHeadersFooters(cloneContext);

                first.Properties.CopyPropertiesFrom(second.Properties);
                first.Headers.CloneHeadersFootersFrom(second.Headers, cloneContext);
                first.Footers.CloneHeadersFootersFrom(second.Footers, cloneContext);

                second.Properties.CopyPropertiesFrom(tempSwappingSection.Properties);
                second.Headers.CloneHeadersFootersFrom(tempSwappingSection.Headers, cloneContext);
                second.Footers.CloneHeadersFootersFrom(tempSwappingSection.Footers, cloneContext);
            }

            private static bool IsDocumentEmpty(RadFlowDocument document)
            {
                return document.Sections.Count == 0;
            }

            private static bool ShouldInsertParagraph(Paragraph paragraph, InsertDocumentContext insertContext, BlockCollection blocks, int insertedBlocksCount)
            {
                bool isParagraphDifferentThanTheLastParagraphMarker = paragraph != insertContext.LastParagraphMarker;
                if (isParagraphDifferentThanTheLastParagraphMarker)
                {
                    bool isNextBlockTable = blocks.Count > insertedBlocksCount + 1 && 
                                            blocks[insertedBlocksCount + 1].Type == DocumentElementType.Table;
                    return blocks.Count == 1 || !isNextBlockTable;
                }
                else
                {
                    return false;
                }
            }

            private static DocumentElementBase CloneElement(DocumentElementBase block, InsertDocumentContext insertContext)
            {
                DocumentElementBase clonedBlock = block.CloneCore(insertContext.CloneContext);
                if (clonedBlock != null)
                {
                    clonedBlock.OnAfterCloneCore(insertContext.CloneContext, clonedBlock);
                }

                return clonedBlock;
            }

            private void BeginInsertDocument(RadFlowDocument sourceDocument, InsertDocumentContext insertContext)
            {
                this.insertTargetParagraphFormatting = new Paragraph(this.editor.Document);
                this.clonedSectionIndex = 0;

                this.insertTargetSection = this.editor.position.CurrentBlock.EnumerateParentsOfType<Section>().FirstOrDefault();

                if (!insertContext.InsertDocumentOptions.InsertLastParagraphMarker)
                {
                    insertContext.LastParagraphMarker = sourceDocument.Sections
                                                                      .Last()
                                                                      .EnumerateContentChildrenOfType<Paragraph>()
                                                                      .LastOrDefault();
                }

                if (this.editor.position.CurrentParagraph != null)
                {
                    this.isInsertPositionInParagraph = true;
                    this.FormattingOfInsertTargetParagraph.CopyPropertiesFrom(this.editor.position.CurrentParagraph.Properties);
                }

                this.initialInsertSectionIndex = this.editor.Document.Sections.IndexOf(this.insertTargetSection) + 1;
            }

            private void InsertDocumentInternal(RadFlowDocument sourceDocument, InsertDocumentContext insertContext, bool isTargetDocumentEmpty)
            {
                Section tempSwappingSection = this.insertTargetSection;
                foreach (Section section in sourceDocument.Sections)
                {
                    this.hasInsertedMultipleParagraphs |= this.InsertBlocksCloned(section.Blocks, insertContext);

                    if (isTargetDocumentEmpty || (this.clonedSectionIndex != sourceDocument.Sections.Count - 1))
                    {
                        Section clonedSection = section.ClonePropertiesAndHeadersFooters(insertContext.CloneContext);
                        SwapSectionPropertiesAndHeaderFooters(tempSwappingSection, clonedSection, insertContext.CloneContext);
                        tempSwappingSection = clonedSection;

                        if ((this.clonedSectionIndex > 0 || sourceDocument.Sections.Count > 1) &&
                            this.clonedSectionIndex != sourceDocument.Sections.Count - 1)
                        {
                            int insertIndex = this.initialInsertSectionIndex + this.clonedSectionIndex;
                            if (insertIndex >= this.editor.Document.Sections.Count)
                            {
                                this.editor.Document.Sections.Add(clonedSection);
                            }
                            else
                            {
                                this.editor.Document.Sections.Insert(insertIndex, clonedSection);
                            }

                            this.editor.SetPositionAfterSection(clonedSection);
                        }
                    }

                    this.clonedSectionIndex++;
                }
            }

            private void EndInsertDocument()
            {
                if (this.hasInsertedMultipleParagraphs && this.isInsertPositionInParagraph)
                {
                    this.OverrideCurrentParagraphProperties(this.FormattingOfInsertTargetParagraph);
                }
            }

            private bool InsertBlocksCloned(BlockCollection blocks, InsertDocumentContext insertContext)
            {
                bool hasInsertedParagraph = false;
                int insertedBlocksCounter = 0;
                foreach (BlockBase block in blocks)
                {
                    switch (block.Type)
                    {
                        case DocumentElementType.Paragraph:
                            this.editor.PreparePositionForInsertInline();
                            Paragraph paragraph = (Paragraph)block;
                            foreach (InlineBase inline in paragraph.Inlines)
                            {
                                DocumentElementBase clonedInline = CloneElement(inline, insertContext);
                                this.editor.InsertInlineInternal((InlineBase)clonedInline);
                            }

                            Paragraph clonedParagraph = paragraph.CloneWithoutChildren(insertContext.CloneContext);
                            this.OverrideCurrentParagraphProperties(clonedParagraph.Properties);

                            if (ShouldInsertParagraph(paragraph, insertContext, blocks, insertedBlocksCounter))
                            {
                                this.editor.InsertParagraphInternal(clonedParagraph, false);
                                hasInsertedParagraph = true;
                            }
                            insertedBlocksCounter++;
                            break;
                        case DocumentElementType.Table:
                            DocumentElementBase clonedBlock = CloneElement(block, insertContext);
                            this.editor.InsertTableInternal((Table)clonedBlock, false);
                            insertedBlocksCounter++;
                            break;
                    }
                }

                return hasInsertedParagraph;
            }
                        
            private void OverrideCurrentParagraphProperties(ParagraphProperties properties)
            {
                Paragraph currentParagraph = this.editor.position.CurrentParagraph;
                if (currentParagraph != null)
                {
                    currentParagraph.Properties.ClearLocalValues();
                    currentParagraph.Properties.CopyPropertiesFrom(properties);
                }
            }
        }
    }
}