﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Core.Imaging;
using Telerik.Windows.Documents.Flow.Model.Fields;
using Telerik.Windows.Documents.Flow.Model.Protection;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Model.Watermarks;
using Telerik.Windows.Documents.Flow.TextSearch;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Editing
{
    /// <summary>
    /// Defines utility class used for editing <see cref="Telerik.Windows.Documents.Flow.Model.RadFlowDocument">RadFlowDocuments</see>
    /// </summary>
    public partial class RadFlowDocumentEditor
    {
        private readonly RadFlowDocument document;
        private readonly EditorPosition position;

        private readonly Run characterFormatting;
        private readonly Paragraph paragraphFormatting;
        private readonly Table tableFormatting;

        /// <summary>
        /// Initializes a new instance of the <see cref="RadFlowDocumentEditor" /> class.
        /// </summary>
        /// <param name="document">The document.</param>
        public RadFlowDocumentEditor(RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");

            this.document = document;
            this.position = new EditorPosition(this.document);

            this.characterFormatting = new Run(this.document);
            this.paragraphFormatting = new Paragraph(this.document);
            this.tableFormatting = new Table(this.document);
        }

        /// <summary>
        /// Gets the character formatting that will be used when creating new run elements.
        /// </summary>
        /// <value>The character formatting.</value>
        public CharacterProperties CharacterFormatting
        {
            get
            {
                return this.characterFormatting.Properties;
            }
        }

        /// <summary>
        /// Gets the paragraph formatting that will be used when creating new paragraphs.
        /// </summary>
        /// <value>The paragraph formatting.</value>
        public ParagraphProperties ParagraphFormatting
        {
            get
            {
                return this.paragraphFormatting.Properties;
            }
        }

        /// <summary>
        /// Gets the table formatting that will be used when creating new tables.
        /// </summary>
        /// <value>The table formatting.</value>
        public TableProperties TableFormatting
        {
            get
            {
                return this.tableFormatting.Properties;
            }
        }

        /// <summary>
        /// Gets the document to which this editor is associated.
        /// </summary>
        /// <value>
        /// The document.
        /// </value>
        public RadFlowDocument Document
        {
            get
            {
                return this.document;
            }
        }

        /// <summary>
        /// Inserts text.
        /// </summary>
        /// <param name="text">The text.</param>
        /// <returns>The inserted last inserted Run element.</returns>
        public Run InsertText(string text)
        {
            this.PreparePositionForInsertInline();

            Run run = null;
            string[] lines = TextHelper.GetLines(text);
            for (int i = 0; i < lines.Length; i++)
            {
                if (i > 0)
                {
                    this.InsertParagraph();
                }

                string line = lines[i];
                if (!string.IsNullOrEmpty(line))
                {
                    string[] lineParts = line.Split(new string[] { SpecialSymbols.VerticalTabSymbol }, StringSplitOptions.None);

                    for (int j = 0; j < lineParts.Length; j++)
                    {
                        if (j > 0)
                        {
                            this.InsertBreak(BreakType.LineBreak);
                        }

                        string linePart = lineParts[j];
                        if (!string.IsNullOrEmpty(linePart))
                        {
                            run = new Run(this.Document);
                            this.ApplyCharacterFormatting(run);
                            run.Text = linePart;

                            this.InsertInlineInternal(run);
                        }
                    }
                }
            }
            return run;
        }

        /// <summary>
        /// Inserts text and adds new paragraph after it.
        /// </summary>
        /// <param name="text">The text.</param>
        /// <returns>The inserted last inserted Run element.</returns>
        public Run InsertLine(string text)
        {
            Run run = this.InsertText(text);
            this.InsertParagraph();
            return run;
        }

        /// <summary>
        /// Inserts the hyperlink.
        /// </summary>
        /// <param name="text">The text.</param>
        /// <param name="uri">The URI.</param>
        /// <param name="isAnchor">True for hyperlinks pointing to a bookmark.</param>
        /// <param name="toolTip">The tool tip.</param>
        /// <returns></returns>
        public Hyperlink InsertHyperlink(string text, string uri, bool isAnchor, string toolTip = null)
        {
            this.PreparePositionForInsertInline();

            Hyperlink hyperlink = new Hyperlink(this.Document);
            hyperlink.Uri = uri;
            hyperlink.IsAnchor = isAnchor;
            hyperlink.ToolTip = toolTip;

            FieldInfo hyperlinkField = new FieldInfo(this.Document, hyperlink);

            this.InsertInlineInternal(hyperlinkField.Start);
            this.InsertText(hyperlink.CreateHyperlinkCode());
            this.InsertInlineInternal(hyperlinkField.Separator);

            // Insert with style if needed.
            if (string.IsNullOrEmpty(this.characterFormatting.StyleId))
            {
                if (!this.Document.StyleRepository.Contains(BuiltInStyleNames.HyperlinkStyleId))
                {
                    this.Document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.HyperlinkStyleId);
                }

                this.characterFormatting.StyleId = BuiltInStyleNames.HyperlinkStyleId;
                this.InsertText(text);
                this.characterFormatting.StyleId = string.Empty;
            }
            else
            {
                this.InsertText(text);
            }
            this.InsertInlineInternal(hyperlinkField.End);

            return hyperlink;
        }

        /// <summary>
        /// Inserts new field.
        /// </summary>
        /// <param name="code">The code fragment.</param>
        /// <param name="result">The result fragment.</param>
        /// <returns></returns>
        public FieldInfo InsertField(string code, string result)
        {
            this.PreparePositionForInsertInline();

            FieldInfo field = new FieldInfo(this.Document);

            this.InsertInlineInternal(field.Start);

            if (!string.IsNullOrEmpty(code))
            {
                this.InsertText(code);
            }

            this.InsertInlineInternal(field.Separator);

            if (!string.IsNullOrEmpty(result))
            {
                this.InsertText(result);
            }

            this.InsertInlineInternal(field.End);

            field.LoadFieldFromCode();
            return field;
        }

        /// <summary>
        /// Inserts a new section.
        /// </summary>
        /// <returns>The inserted Section element.</returns>
        public Section InsertSection()
        {
            Section section = new Section(this.Document);
            this.InsertSectionInternal(section);

            return section;
        }

        /// <summary>
        /// Inserts a new paragraph.
        /// </summary>
        /// <returns>The inserted Paragraph element.</returns>
        public Paragraph InsertParagraph()
        {
            Paragraph paragraph = new Paragraph(this.Document);
            this.InsertParagraphInternal(paragraph);

            return paragraph;
        }

        /// <summary>
        /// Inserts an inline image with its original size.
        /// </summary>
        /// <param name="stream">The stream.</param>
        /// <param name="extension">The extension.</param>
        /// <returns>The inserted inline image.</returns>
        public ImageInline InsertImageInline(Stream stream, string extension)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");
            Guard.ThrowExceptionIfNullOrEmpty(extension, "extension");

            return this.InsertImageInline(new ImageSource(stream, extension), ImageSizeDecodersManager.GetSize(stream, extension));
        }

        /// <summary>
        /// Inserts an inline image.
        /// </summary>
        /// <param name="source">The image source.</param>
        /// <param name="size">The size.</param>
        /// <returns>The inserted image inline.</returns>
        public ImageInline InsertImageInline(ImageSource source, Size size)
        {
            Guard.ThrowExceptionIfNull(source, "source");

            this.PreparePositionForInsertInline();

            ImageInline image = new ImageInline(this.Document);
            image.Image.ImageSource = source;
            image.Image.Size = size;
            this.InsertInlineInternal(image);

            return image;
        }

        /// <summary>
        /// Inserts the image inline.
        /// </summary>
        /// <param name="stream">The image stream.</param>
        /// <param name="extension">The image extension.</param>
        /// <param name="size">The size.</param>
        /// <returns>The inserted image inline.</returns>
        public ImageInline InsertImageInline(Stream stream, string extension, Size size)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");
            Guard.ThrowExceptionIfNullOrEmpty(extension, "extension");

            return this.InsertImageInline(new ImageSource(stream, extension), size);
        }

        /// <summary>
        /// Inserts a floating image with its original size.
        /// </summary>
        /// <param name="stream">The stream.</param>
        /// <param name="extension">The extension.</param>
        /// <returns>The inserted floating image.</returns>
        public FloatingImage InsertFloatingImage(Stream stream, string extension)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");
            Guard.ThrowExceptionIfNullOrEmpty(extension, "extension");

            return this.InsertFloatingImage(new ImageSource(stream, extension), ImageSizeDecodersManager.GetSize(stream, extension));
        }

        /// <summary>
        /// Inserts a floating image.
        /// </summary>
        /// <param name="stream">The image stream.</param>
        /// <param name="extension">The image extension.</param>
        /// <param name="size">The size.</param>
        /// <returns>The floating image inline.</returns>
        public FloatingImage InsertFloatingImage(Stream stream, string extension, Size size)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");
            Guard.ThrowExceptionIfNullOrEmpty(extension, "extension");

            return this.InsertFloatingImage(new ImageSource(stream, extension), size);
        }

        /// <summary>
        /// Inserts a floating image.
        /// </summary>
        /// <param name="source">The image source.</param>
        /// <param name="size">The size.</param>
        /// <returns>The floating image inline.</returns>
        public FloatingImage InsertFloatingImage(ImageSource source, Size size)
        {
            Guard.ThrowExceptionIfNull(source, "source");

            this.PreparePositionForInsertInline();

            FloatingImage image = new FloatingImage(this.Document);
            image.Image.ImageSource = source;
            image.Image.Size = size;
            this.InsertInlineInternal(image);

            return image;
        }

        /// <summary>
        /// Inserts an inline.
        /// </summary>
        /// <param name="inline">The inline.</param>
        /// <returns>The inline.</returns>
        public InlineBase InsertInline(InlineBase inline)
        {
            Guard.ThrowExceptionIfNull(inline, "inline");

            this.PreparePositionForInsertInline();

            this.InsertInlineInternal(inline);

            return inline;
        }

        /// <summary>
        /// Inserts a new table and positions the editor after the table.
        /// </summary>
        /// <returns> The inserted table.</returns>
        public Table InsertTable()
        {
            return this.InsertTable(0, 0);
        }

        /// <summary>
        /// Inserts a new table and positions the editor after the table.
        /// </summary>
        /// <param name="rows">The number of rows in the table.</param>
        /// <param name="columns">The number of columns in the table.</param>
        /// <returns>The inserted table.</returns>
        public Table InsertTable(int rows, int columns)
        {
            Guard.ThrowExceptionIfOutOfRange(0, int.MaxValue, rows, "rows");
            Guard.ThrowExceptionIfOutOfRange(0, int.MaxValue, columns, "columns");

            Table table = new Table(this.Document, rows, columns);
            this.InsertTableInternal(table);

            return table;
        }

        /// <summary>
        /// Inserts the break.
        /// </summary>
        /// <param name="type">The type of the break.</param>
        /// <returns>The inserted break.</returns>
        public Break InsertBreak(BreakType type)
        {
            Break br = new Break(this.Document);
            br.BreakType = type;

            if (type == BreakType.LineBreak || !this.position.IsInTableCell())
            {
                this.InsertInline(br);

                if (type == BreakType.PageBreak)
                {
                    this.InsertParagraph();
                }
            }
            else
            {
                // We should pass the parent Section element.
                Section currentSection = this.position.CurrentBlock.BlockContainer.EnumerateParentsOfType<Section>().FirstOrDefault();
                TableCell currentPositionCell = this.position.CurrentBlock.Parent.EnumerateParentsOfType<TableCell>().Last();

                this.InsertBreakRespectiveTable(currentPositionCell, currentSection, p => p.Inlines.Add(br));
            }

            return br;
        }

        /// <summary>
        /// Inserts the comment.
        /// </summary>
        /// <param name="text">The text of the comment.</param>
        /// <returns>The created comment.</returns>
        public Comment InsertComment(string text)
        {
            Comment comment = this.Document.Comments.AddComment();
            comment.Blocks.AddParagraph().Inlines.AddRun(text);

            this.InsertInline(comment.CommentRangeStart);
            this.InsertInline(comment.CommentRangeEnd);

            return comment;
        }

        /// <summary>
        /// Inserts the comment.
        /// </summary>
        /// <param name="text">The text.</param>
        /// <param name="inlineStart">The inline start. The start of the comment will be inserted before this inline.</param>
        /// <param name="inlineEnd">The inline end. The end of the comment will be inserted after this inline.</param>
        /// <returns></returns>
        public Comment InsertComment(string text, InlineBase inlineStart, InlineBase inlineEnd)
        {
            Guard.ThrowExceptionIfNull(inlineStart, "inlineStart");
            Guard.ThrowExceptionIfNull(inlineEnd, "inlineEnd");

            Comment comment = this.Document.Comments.AddComment();
            comment.Blocks.AddParagraph().Inlines.AddRun(text);

            return this.InsertComment(comment, inlineStart, inlineEnd);
        }

        /// <summary>
        /// Inserts the comment.
        /// </summary>
        /// <param name="comment">The comment.</param>
        /// <returns>The comment.</returns>
        public Comment InsertComment(Comment comment)
        {
            Guard.ThrowExceptionIfNull(comment, "comment");
            Guard.ThrowExceptionIfFalse(comment.Document == this.document, "Comment is from another document.");

            this.InsertInline(comment.CommentRangeStart);
            this.InsertInline(comment.CommentRangeEnd);

            return comment;
        }

        /// <summary>
        /// Inserts the comment.
        /// </summary>
        /// <param name="comment">The comment.</param>
        /// <param name="inlineStart">The inline start. The start of the comment will be inserted before this inline.</param>
        /// <param name="inlineEnd">The inline end. The end of the comment will be inserted after this inline.</param>
        /// <returns>The comment.</returns>
        public Comment InsertComment(Comment comment, InlineBase inlineStart, InlineBase inlineEnd)
        {
            Guard.ThrowExceptionIfNull(comment, "comment");
            Guard.ThrowExceptionIfNull(inlineStart, "inlineStart");
            Guard.ThrowExceptionIfNull(inlineEnd, "inlineEnd");
            Guard.ThrowExceptionIfFalse(comment.Document == this.document, "Comment is from another document.");

            this.MoveToInlineStart(inlineStart);
            this.InsertInline(comment.CommentRangeStart);
            this.MoveToInlineEnd(inlineEnd);
            this.InsertInline(comment.CommentRangeEnd);

            return comment;
        }

        /// <summary>
        /// Inserts the bookmark with specified name.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <returns>The created bookmark.</returns>
        public Bookmark InsertBookmark(string name)
        {
            // TODO: Write validation for bookmark name with Bookmark.IsValidBookmarkName()
            if (this.Document.GetBookmarkByName(name) != null)
            {
                throw new InvalidOperationException(string.Format("A Bookmark with name '{0}' already exists.", name));
            }

            Bookmark bookmark = new Bookmark(this.Document, name);

            this.InsertInline(bookmark.BookmarkRangeStart);
            this.InsertInline(bookmark.BookmarkRangeEnd);

            return bookmark;
        }

        /// <summary>
        /// Inserts the bookmark with specified name.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="inlineStart">The inline start. The start of the bookmark will be inserted before this inline.</param>
        /// <param name="inlineEnd">The inline end. The end of the bookmark will be inserted after this inline.</param>
        /// <returns>The created bookmark.</returns>
        public Bookmark InsertBookmark(string name, InlineBase inlineStart, InlineBase inlineEnd)
        {
            Guard.ThrowExceptionIfNull(inlineStart, "inlineStart");
            Guard.ThrowExceptionIfNull(inlineEnd, "inlineEnd");
            if (this.Document.EnumerateChildrenOfType<BookmarkRangeStart>().Any(b => b.Bookmark.Name == name))
            {
                throw new InvalidOperationException(string.Format("A Bookmark with name '{0}' already exists.", name));
            }

            Bookmark bookmark = new Bookmark(this.Document, name);

            this.MoveToInlineStart(inlineStart);
            this.InsertInline(bookmark.BookmarkRangeStart);

            this.MoveToInlineEnd(inlineEnd);
            this.InsertInline(bookmark.BookmarkRangeEnd);

            return bookmark;
        }

        /// <summary>
        /// Inserts the source document at the current position in the target document.
        /// </summary>
        /// <param name="sourceDocument">The source document.</param>
        public void InsertDocument(RadFlowDocument sourceDocument)
        {
            this.InsertDocument(sourceDocument, null);
        }

        /// <summary>
        /// Inserts the source document at the current position in the target document.
        /// </summary>
        /// <param name="sourceDocument">The source document.</param>
        /// <param name="insertOptions">The insert options.</param>
        public void InsertDocument(RadFlowDocument sourceDocument, InsertDocumentOptions insertOptions)
        {
            Guard.ThrowExceptionIfNull(sourceDocument, "sourceDocument");

            if (sourceDocument.Sections.Count == 0)
            {
                throw new ArgumentException("Source document does not contain any sections.");
            }

            if (this.position.IsInTableCell() && sourceDocument.Sections.Count > 1)
            {
                throw new InvalidOperationException("The source document cannot be inserted at the current position.");
            }

            InsertDocumentEditor insertDocumentEditor = new InsertDocumentEditor(this);
            insertDocumentEditor.InsertDocument(sourceDocument, insertOptions);
        }

        /// <summary>
        /// Deletes a bookmark by name.
        /// </summary>
        /// <param name="name">The name.</param>
        public void DeleteBookmark(string name)
        {
            BookmarkRangeStart start = this.Document.EnumerateChildrenOfType<BookmarkRangeStart>()
                                     .Where(b => b.Bookmark.Name == name)
                                     .FirstOrDefault();
            if (start == null)
            {
                throw new InvalidOperationException(string.Format("A Bookmark with name '{0}' does not exist.", name));
            }

            this.DeleteBookmark(start.Bookmark);
        }

        /// <summary>
        /// Deletes the bookmark.
        /// </summary>
        /// <param name="bookmark">The bookmark.</param>
        public void DeleteBookmark(Bookmark bookmark)
        {
            Guard.ThrowExceptionIfFalse(bookmark.Document == this.document, "Bookmark is from another document.");

            BookmarkRangeStart start = bookmark.BookmarkRangeStart;
            BookmarkRangeEnd end = bookmark.BookmarkRangeEnd;

            if (start.Paragraph == null || end.Paragraph == null)
            {
                throw new InvalidOperationException("Bookmark is not part of the document.");
            }

            start.Paragraph.Inlines.Remove(start);
            end.Paragraph.Inlines.Remove(end);
        }

        /// <summary>
        /// Inserts the permission range surrounding inline start and inline end.
        /// </summary>
        /// <param name="permissionRangeCredentials">The permission range credentials.</param>
        /// <param name="inlineStart">The inline start.</param>
        /// <param name="inlineEnd">The inline end.</param>
        /// <returns></returns>
        public PermissionRange InsertPermissionRange(PermissionRangeCredentials permissionRangeCredentials, InlineBase inlineStart, InlineBase inlineEnd)
        {
            Guard.ThrowExceptionIfNull(permissionRangeCredentials, "permissionRangeCredentials");
            Guard.ThrowExceptionIfNull(inlineStart, "inlineStart");
            Guard.ThrowExceptionIfNull(inlineEnd, "inlineEnd");

            PermissionRange permission = new PermissionRange(this.Document, permissionRangeCredentials);

            this.MoveToInlineStart(inlineStart);
            this.InsertInline(permission.Start);

            this.MoveToInlineEnd(inlineEnd);
            this.InsertInline(permission.End);

            return permission;
        }

        /// <summary>
        /// Inserts the permission range surrounding table cell.
        /// </summary>
        /// <param name="permissionRangeCredentials">The permission range credentials.</param>
        /// <param name="cell">The cell.</param>
        /// <returns></returns>
        public PermissionRange InsertPermissionRange(PermissionRangeCredentials permissionRangeCredentials, TableCell cell)
        {
            Guard.ThrowExceptionIfNull(permissionRangeCredentials, "permissionRangeCredentials");
            Guard.ThrowExceptionIfNull(cell, "cell");
            Guard.ThrowExceptionIfNull(cell.Row, "cell.Row");
            Guard.ThrowExceptionIfNull(cell.Row.Table, "cell.Row.Table");

            int fromColumn = cell.GridColumnIndex;
            int toColumn = cell.GridColumnIndex + cell.ColumnSpan - 1;
            PermissionRange permission = new PermissionRange(this.Document, permissionRangeCredentials, fromColumn, toColumn);

            Paragraph startParagraph = cell.EnumerateChildrenOfType<Paragraph>().FirstOrDefault();
            if (startParagraph == null)
            {
                startParagraph = cell.Blocks.AddParagraph();
            }

            this.position.MoveToParagraphStart(startParagraph);
            this.InsertInline(permission.Start);

            Paragraph nextParagraph = RadFlowDocumentEditor.GetParagraphFromNextBlockElement(cell);
            if (nextParagraph != null)
            {
                this.position.MoveToParagraphStart(nextParagraph);
            }

            this.InsertInline(permission.End);

            return permission;
        }

        /// <summary>
        /// Inserts the permission range surrounding table row.
        /// </summary>
        /// <param name="permissionRangeCredentials">The permission range credentials.</param>
        /// <param name="row">The row.</param>
        /// <returns></returns>
        public PermissionRange InsertPermissionRange(PermissionRangeCredentials permissionRangeCredentials, TableRow row)
        {
            Guard.ThrowExceptionIfNull(permissionRangeCredentials, "permissionRangeCredentials");
            Guard.ThrowExceptionIfNull(row, "row");
            Guard.ThrowExceptionIfLessThanOrEqual(0, row.Cells.Count, "row.Cells.Count");
            Guard.ThrowExceptionIfNull(row.Table, "row.Table");

            TableCell firstCell = row.Cells.First();
            TableCell lastCell = row.Cells.Last();

            int fromColumn = firstCell.GridColumnIndex;
            int toColumn = lastCell.GridColumnIndex + lastCell.ColumnSpan - 1;
            PermissionRange permission = new PermissionRange(this.Document, permissionRangeCredentials, fromColumn, toColumn);

            Paragraph startParagraph = firstCell.EnumerateChildrenOfType<Paragraph>().FirstOrDefault();
            if (startParagraph == null)
            {
                startParagraph = firstCell.Blocks.AddParagraph();
            }

            this.position.MoveToParagraphStart(startParagraph);
            this.InsertInline(permission.Start);

            Paragraph nextParagraph = RadFlowDocumentEditor.GetParagraphFromNextBlockElement(lastCell);
            if (nextParagraph != null)
            {
                this.position.MoveToParagraphStart(nextParagraph);
            }

            this.InsertInline(permission.End);

            return permission;
        }

        /// <summary>
        /// Deletes the permission range.
        /// </summary>
        /// <param name="permission">The permission.</param>
        public void DeletePermissionRange(PermissionRange permission)
        {
            Guard.ThrowExceptionIfFalse(permission.Document == this.document, "Permission is from another document.");

            PermissionRangeStart start = permission.Start;
            PermissionRangeEnd end = permission.End;

            if (start.Paragraph == null || end.Paragraph == null)
            {
                throw new InvalidOperationException("Bookmark is not part of the document.");
            }

            start.Paragraph.Inlines.Remove(start);
            end.Paragraph.Inlines.Remove(end);
        }

        /// <summary>
        /// Protects the editor document from changes with specified password in read only protection mode.
        /// </summary>
        /// <param name="password">The password.</param>
        public void Protect(string password)
        {
            this.Protect(password, ProtectionMode.ReadOnly);
        }

        /// <summary>
        /// Protects the editor document from changes with specified password in desired protection mode.
        /// </summary>
        /// <param name="password">The password.</param>
        /// <param name="protectionMode">The protection mode.</param>
        public void Protect(string password, ProtectionMode protectionMode)
        {
            if (this.Document.ProtectionSettings.Enforced)
            {
                throw new InvalidOperationException("Cannot protect a document that is already protected.");
            }

            if (!string.IsNullOrEmpty(password))
            {
                if (string.IsNullOrEmpty(this.Document.ProtectionSettings.Salt))
                {
                    this.Document.ProtectionSettings.Salt = FlowProtectionHelper.Instance.GenerateSaltBase64();
                }

                string hash = FlowProtectionHelper.Instance.GenerateHashBase64(
                    password, this.Document.ProtectionSettings.Salt, this.Document.ProtectionSettings.AlgorithmName, this.Document.ProtectionSettings.SpinCount);

                this.Document.ProtectionSettings.Hash = hash;
            }

            this.Document.ProtectionSettings.Enforced = true;
            this.Document.ProtectionSettings.ProtectionMode = protectionMode;
        }

        /// <summary>
        /// Tries to unprotect the editor document with specified password.
        /// </summary>
        /// <param name="password">The password.</param>
        /// <returns><c>True</c> if the password is correct; otherwise, <c>False</c>.</returns>
        public bool Unprotect(string password)
        {
            bool isPasswordCorrect = FlowProtectionHelper.Instance.IsPasswordCorrect(
                password,
                this.Document.ProtectionSettings.Hash,
                this.Document.ProtectionSettings.Salt,
                this.Document.ProtectionSettings.AlgorithmName,
                this.Document.ProtectionSettings.SpinCount);

            if (isPasswordCorrect)
            {
                this.Unprotect();
            }

            return isPasswordCorrect;
        }

        /// <summary>
        /// Unprotects the editor document.
        /// </summary>
        public void Unprotect()
        {
            this.Document.ProtectionSettings.Enforced = false;
            this.Document.ProtectionSettings.ProtectionMode = ProtectionMode.NoProtection;
            this.Document.ProtectionSettings.Hash = string.Empty;
            this.Document.ProtectionSettings.Salt = string.Empty;
        }

        /// <summary>
        /// Replaces all occurrences of a specified text in a <see cref="Telerik.Windows.Documents.Flow.Model.RadFlowDocument"/>.
        /// </summary>
        /// <param name="oldText">The searched text.</param>
        /// <param name="newText">The replace text.</param>
        public void ReplaceText(string oldText, string newText)
        {
            this.ReplaceText(oldText, newText, true, false);
        }

        /// <summary>
        /// Replaces all occurrences of a specified text in a <see cref="Telerik.Windows.Documents.Flow.Model.RadFlowDocument"/>.
        /// </summary>
        /// <param name="oldText">The searched text.</param>
        /// <param name="newText">The replace text.</param>
        /// <param name="matchCase">Indicates whether the casing should be matched.</param>
        /// <param name="matchWholeWord">Indicates whether only whole words should be matched.</param>
        public void ReplaceText(string oldText, string newText, bool matchCase, bool matchWholeWord)
        {
            Guard.ThrowExceptionIfNullOrEmpty(oldText, "oldText");

            this.ReplaceText(FindTextService.BuildRegex(oldText, matchCase, matchWholeWord), newText);
        }

        /// <summary>
        /// Replaces all occurrences of a matched text by the specified regex in a <see cref="Telerik.Windows.Documents.Flow.Model.RadFlowDocument"/>.
        /// </summary>
        /// <param name="regex">The regex.</param>
        /// <param name="newText">The replace text.</param>
        public void ReplaceText(Regex regex, string newText)
        {
            Guard.ThrowExceptionIfNull(regex, "regex");

            IEnumerable<RunTextMatch> matches = FindReplaceManager.Find(regex, this.Document);
            FindReplaceManager.Replace(matches, regex, newText, new ReplaceTextHandler());
        }

        /// <summary>
        /// Replaces character properties of all occurrences of the searched text in a <see cref="Telerik.Windows.Documents.Flow.Model.RadFlowDocument"/>.
        /// </summary>
        /// <param name="searchedText">The searched text.</param>
        /// <param name="replacePropertiesAction">The action which replaces character properties.</param>
        public void ReplaceStyling(string searchedText, Action<CharacterProperties> replacePropertiesAction)
        {
            this.ReplaceStyling(searchedText, true, false, replacePropertiesAction);
        }

        /// <summary>
        /// Replaces character properties of all occurrences of the searched text in a <see cref="Telerik.Windows.Documents.Flow.Model.RadFlowDocument"/>.
        /// </summary>
        /// <param name="searchedText">The searched text.</param>
        /// <param name="matchCase">Indicates whether the casing should be matched.</param>
        /// <param name="matchWholeWord">Indicates whether only whole words should be matched.</param>
        /// <param name="replacePropertiesAction">The action which replaces character properties.</param>
        public void ReplaceStyling(string searchedText, bool matchCase, bool matchWholeWord, Action<CharacterProperties> replacePropertiesAction)
        {
            Guard.ThrowExceptionIfNullOrEmpty(searchedText, "searchedText");

            this.ReplaceStyling(FindTextService.BuildRegex(searchedText, matchCase, matchWholeWord), replacePropertiesAction);
        }

        /// <summary>
        /// Replaces character properties of all occurrences of a 
        /// <para>matched text by the specified regex in a <see cref="Telerik.Windows.Documents.Flow.Model.RadFlowDocument"/>.</para> 
        /// </summary>
        /// <param name="regex">The regex.</param>
        /// <param name="replacePropertiesAction">The action which replaces character properties.</param>
        public void ReplaceStyling(Regex regex, Action<CharacterProperties> replacePropertiesAction)
        {
            Guard.ThrowExceptionIfNull(replacePropertiesAction, "replacePropertiesAction");

            IEnumerable<RunTextMatch> matches = FindReplaceManager.Find(regex, this.Document);
            FindReplaceManager.Replace(matches, regex, string.Empty, new ReplacePropertiesHandler(replacePropertiesAction));
        }

        /// <summary>
        /// Sets the watermark to the specified header.
        /// </summary>
        /// <param name="watermark">The watermark.</param>
        /// <param name="header">The header.</param>
        public void SetWatermark(Watermark watermark, Header header)
        {
            Guard.ThrowExceptionIfNull(watermark, "watermark");
            Guard.ThrowExceptionIfNull(header, "header");
            Guard.ThrowExceptionIfNotNull(watermark.Document, "Document");

            header.Watermarks.Clear();
            header.Watermarks.Add(watermark);
        }

        /// <summary>
        /// Sets the watermark to the header of the specified type associated with the specified section.
        /// </summary>
        /// <param name="watermark">The watermark.</param>
        /// <param name="section">The section.</param>
        /// <param name="headerType">Type of the header.</param>
        public void SetWatermark(Watermark watermark, Section section, HeaderFooterType headerType)
        {
            Guard.ThrowExceptionIfNull(watermark, "watermark");
            Guard.ThrowExceptionIfNull(section, "section");

            Header header;
            switch (headerType)
            {
                case HeaderFooterType.Default:
                    if (section.Headers.Default == null)
                    {
                        section.Headers.Add(HeaderFooterType.Default);
                    }
                    header = section.Headers.Default;
                    break;

                case HeaderFooterType.Even:
                    if (section.Headers.Even == null)
                    {
                        section.Headers.Add(HeaderFooterType.Even);
                    }
                    header = section.Headers.Even;
                    break;

                case HeaderFooterType.First:
                    if (section.Headers.First == null)
                    {
                        section.Headers.Add(HeaderFooterType.First);
                    }
                    header = section.Headers.First;
                    break;

                default:
                    throw new NotSupportedException("Unsupported header type.");
            }

            this.SetWatermark(watermark, header);
        }

        /// <summary>
        /// Moves the editor to the start of a paragraph.
        /// </summary>
        /// <param name="paragraph">The paragraph.</param>
        public void MoveToParagraphStart(Paragraph paragraph)
        {
            Guard.ThrowExceptionIfFalse(paragraph.Document == this.document, "document");

            this.position.MoveToParagraphStart(paragraph);
        }

        /// <summary>
        /// Moves the editor after an inline.
        /// </summary>
        /// <param name="inline">The inline.</param>
        public void MoveToInlineEnd(InlineBase inline)
        {
            Guard.ThrowExceptionIfFalse(inline.Document == this.document, "document");

            this.position.MoveToInlineEnd(inline);
        }

        /// <summary>
        /// Moves the editor before an inline.
        /// </summary>
        /// <param name="inline">The inline.</param>
        public void MoveToInlineStart(InlineBase inline)
        {
            Guard.ThrowExceptionIfFalse(inline.Document == this.document, "document");

            this.position.MoveToInlineStart(inline);
        }

        /// <summary>
        /// Moves the editor to a table end.
        /// </summary>
        /// <param name="table">The table.</param>
        public void MoveToTableEnd(Table table)
        {
            Guard.ThrowExceptionIfFalse(table.Document == this.document, "document");

            this.position.MoveAfterTable(table);
        }

        private static Table ClonePartOfTable(Table tableToClone, int fromRowIndex, int toRowIndex)
        {
            Table clonedTable = new Table(tableToClone.Document);
            clonedTable.Properties.CopyPropertiesFrom(tableToClone.Properties);
            for (int currentRowIndex = fromRowIndex; currentRowIndex < toRowIndex; currentRowIndex++)
            {
                clonedTable.Rows.Add(tableToClone.Rows[currentRowIndex].Clone());
            }

            return clonedTable;
        }

        private static Paragraph GetParagraphFromNextBlockElement(TableCell currentCell)
        {
            TableRow currentRow = currentCell.Row;
            Table currentTable = currentRow.Table;

            Paragraph nextParagraph = null;
            TableCell nextCell = null;

            for (int i = 1; i < currentRow.Cells.Count; i++)
            {
                if (currentRow.Cells[i - 1] == currentCell)
                {
                    nextCell = currentRow.Cells[i];
                    break;
                }
            }

            if (nextCell == null)
            {
                TableRow nextRow = null;
                for (int i = 1; i < currentTable.Rows.Count; i++)
                {
                    if (currentTable.Rows[i - 1] == currentRow)
                    {
                        nextRow = currentTable.Rows[i];
                        break;
                    }
                }

                if (nextRow != null)
                {
                    nextCell = nextRow.Cells.FirstOrDefault();
                    if (nextCell == null)
                    {
                        nextCell = nextRow.Cells.AddTableCell();
                    }
                }
            }

            if (nextCell != null)
            {
                nextParagraph = nextCell.EnumerateChildrenOfType<Paragraph>().FirstOrDefault();
                if (nextParagraph == null)
                {
                    nextParagraph = nextCell.Blocks.AddParagraph();
                }
            }

            if (nextParagraph == null)
            {
                BlockContainerBase parent = currentTable.Parent as BlockContainerBase;
                if (parent != null)
                {
                    for (int i = 1; i < parent.Blocks.Count; i++)
                    {
                        if (parent.Blocks[i - 1] == currentTable)
                        {
                            nextParagraph = parent.Blocks[i] as Paragraph;
                            break;
                        }
                    }
                }
            }

            return nextParagraph;
        }

        private void InsertInlineInternal(InlineBase run)
        {
            Guard.ThrowExceptionIfFalse(this.position.IsInitialized, "Editor position is not initialized.");
            Guard.ThrowExceptionIfFalse(this.position.CurrentParagraph != null, "Editor position is not correct.");

            this.position.CurrentParagraph.Inlines.Insert(this.position.InlineIndex, run);
            this.position.InlineIndex++;
        }

        private void InsertTableInternal(Table table, bool applyCurrentFormatting = true)
        {
            BlockContainerBase parentContainer;
            int insertIndex = 0;
            if (!this.position.IsInitialized)
            {
                parentContainer = this.AssureSection();
            }
            else
            {
                parentContainer = this.position.CurrentBlock.BlockContainer;
                if (this.position.CurrentParagraph != null && !this.position.IsAtParagraphEnd())
                {
                    // We are in the middle of the paragraph - we should split it.
                    Paragraph paragraphAfter = this.InsertParagraph();
                    insertIndex = parentContainer.Blocks.IndexOf(paragraphAfter);
                }
                else
                {
                    // Insert after the current block.
                    insertIndex = parentContainer.Blocks.IndexOf(this.position.CurrentBlock) + 1;
                }
            }

            if (applyCurrentFormatting)
            {
                this.ApplyTableFormatting(table);
            }

            parentContainer.Blocks.Insert(insertIndex, table);
            this.position.MoveAfterTable(table);
        }

        private void InsertParagraphInternal(Paragraph paragraph, bool applyCurrentFormatting = true)
        {
            BlockContainerBase parentContainer = null;
            if (!this.position.IsInitialized)
            {
                parentContainer = this.AssureSection();
            }

            if (applyCurrentFormatting)
            {
                this.ApplyParagraphFormatting(paragraph);
            }

            // Insert the new paragraph 
            if (parentContainer != null)
            {
                parentContainer.Blocks.Insert(0, paragraph);
                this.position.MoveToParagraphStart(paragraph);
            }
            else
            {
                // Insert the new paragraph after the current block
                BlockBase block = this.position.CurrentBlock;
                parentContainer = block.BlockContainer;
                int insertIndex = parentContainer.Blocks.IndexOf(block) + 1;
                parentContainer.Blocks.Insert(insertIndex, paragraph);
            }

            Paragraph currentParagraph = this.position.CurrentParagraph;
            if (currentParagraph != null)
            {
                // Move all inlines for the previous paragraph that are after the current position index
                int splitIndex = this.position.InlineIndex;
                while (currentParagraph.Inlines.Count > splitIndex)
                {
                    InlineBase inlineToMove = currentParagraph.Inlines.Last();
                    currentParagraph.Inlines.RemoveAt(currentParagraph.Inlines.Count - 1);
                    paragraph.Inlines.Insert(0, inlineToMove);
                }
            }

            // Move position to the new paragraph
            this.position.MoveToParagraphStart(paragraph);
        }

        private void InsertSectionInternal(Section section)
        {
            this.Document.Sections.Add(section);

            this.SetPositionAfterSection(section);
        }

        private void SetPositionAfterSection(Section section)
        {
            if (this.position.IsInitialized)
            {
                if (!this.position.IsInTableCell())
                {
                    if (this.position.IsAtTableEnd)
                    {
                        // Insert a paragraph after current table in which the section properties will be stored.
                        this.InsertParagraph();
                    }

                    this.position.CurrentParagraph.SetParent(section);
                    this.InsertParagraph();
                }
                else
                {
                    // We should add the section break before the parent cell.
                    TableCell currentPositionCell = this.position.CurrentBlock.Parent.EnumerateParentsOfType<TableCell>().Last();

                    Action<Paragraph> setParentSectionAction = new Action<Paragraph>(p =>
                    {
                        this.position.MoveToParagraphStart(p);
                        this.position.CurrentParagraph.SetParent(section);
                    });

                    this.InsertBreakRespectiveTable(currentPositionCell, section, setParentSectionAction);
                }
            }
        }

        private void PreparePositionForInsertInline()
        {
            this.AssurePositionInitialized();

            if (this.position.IsAtTableEnd)
            {
                // If the position is on table - find or create a paragraph after the table to insert in.
                Table table = this.position.CurrentTable;
                BlockContainerBase blockContainer = table.BlockContainer;
                int expectedParagraphIndex = blockContainer.Blocks.IndexOf(table) + 1;

                Paragraph paragraph;
                if (expectedParagraphIndex < blockContainer.Blocks.Count &&
                    blockContainer.Blocks[expectedParagraphIndex] is Paragraph)
                {
                    // Found suitable paragraph.
                    paragraph = (Paragraph)blockContainer.Blocks[expectedParagraphIndex];
                }
                else
                {
                    // No suitable paragraph -> insert one.
                    paragraph = this.InsertParagraph();
                }

                this.position.MoveToParagraphStart(paragraph);
            }
        }

        private void AssurePositionInitialized()
        {
            if (!this.position.IsInitialized)
            {
                // Find suitable paragraphs at the start of the document
                Section section = this.AssureSection();
                Paragraph paragraph = section.Blocks.FirstOrDefault() as Paragraph;
                if (paragraph == null)
                {
                    paragraph = new Paragraph(this.Document);
                    this.ApplyParagraphFormatting(paragraph);
                    section.Blocks.Insert(0, paragraph);
                }

                this.position.MoveToParagraphStart(paragraph);
            }
        }

        private Section AssureSection()
        {
            Section section = this.Document.Sections.FirstOrDefault();
            if (section == null)
            {
                section = this.Document.Sections.AddSection();
            }

            return section;
        }

        private void ApplyCharacterFormatting(Run run)
        {
            run.Properties.CopyPropertiesFrom(this.characterFormatting.Properties);
        }

        private void ApplyParagraphFormatting(Paragraph paragraph)
        {
            paragraph.Properties.CopyPropertiesFrom(this.paragraphFormatting.Properties);
        }

        private void ApplyTableFormatting(Table table)
        {
            table.Properties.CopyPropertiesFrom(this.tableFormatting.Properties);
        }

        private void InsertBreakRespectiveTable(TableCell currentPositionCell, BlockContainerBase parentContainer, Action<Paragraph> insertAction)
        {
            Table table = currentPositionCell.Table;
            if (table == null || parentContainer == null)
            {
                return;
            }

            BlockContainerBase tableContainer = table.BlockContainer;
            Paragraph paragraphBetweenTables = new Paragraph(this.Document);

            if (currentPositionCell.GridRowIndex != 0)
            {
                Table aboveTable = ClonePartOfTable(table, 0, currentPositionCell.GridRowIndex);
                Table belowTable = ClonePartOfTable(table, currentPositionCell.GridRowIndex, table.GridRowsCount);

                tableContainer.Blocks.Insert(tableContainer.Blocks.IndexOf(table) + 1, aboveTable);
                tableContainer.Blocks.Remove(table);
                tableContainer.Blocks.Add(paragraphBetweenTables);

                insertAction(paragraphBetweenTables);

                parentContainer.Blocks.Add(belowTable);
                this.position.MoveAfterTable(belowTable);
            }
            else
            {
                // Inserted break is in cell from the first table row, so we should insert the section break before table.
                tableContainer.Blocks.Insert(tableContainer.Blocks.IndexOf(table) - 1, paragraphBetweenTables);
                tableContainer.Blocks.Remove(table);

                insertAction(paragraphBetweenTables);

                parentContainer.Blocks.Add(table);
                this.position.MoveAfterTable(table);
            }
        }

        // TODO: Think of adding InsertImage with Image
        // TODO: Think of adding MoveToCellStart method
    }
}
