using System;
using Telerik.Windows.Documents.Flow.Model.Fields.Expressions;
using Telerik.Windows.Documents.Flow.Model.Fields.FieldCode;

namespace Telerik.Windows.Documents.Flow.Model.Fields
{
    /// <summary>
    /// A base class for fields with comparison argument.
    /// </summary>
    public abstract class ComparisonFieldBase : Field
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="ComparisonFieldBase" /> class.
        /// </summary>
        /// <param name="document">The document.</param>
        protected ComparisonFieldBase(RadFlowDocument document) : base(document)
        {
        }

        internal override bool ExpectComparisonArgument
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Evaluates a comparison argument.
        /// </summary>
        /// <param name="comparison">The comparison.</param>
        /// <param name="document">The document.</param>
        /// <returns></returns>
        internal static ComparisonFieldResult Compare(FieldComparison comparison, RadFlowDocument document)
        {
            // Validation
            if (!ComparisonFieldBase.IsComaprisonOperatorvalid(comparison.Operator))
            {
                return new ComparisonFieldResult(string.Format("Invalid comparison operator: {0}", comparison.Operator), true);
            }

            if (string.IsNullOrEmpty(comparison.LeftArgument))
            {
                return new ComparisonFieldResult(string.Format("Missing left argument{0}", comparison.Operator), true);
            }

            if (string.IsNullOrEmpty(comparison.RightArgument))
            {
                return new ComparisonFieldResult(string.Format("Missing right argument{0}", comparison.Operator), true);
            }

            bool result = false;
            if (comparison.IsLeftArgumentQuoted)
            {
                result = ComparisonFieldBase.CompareArgumentsAsString(comparison);
            }
            else
            {
                result = ComparisonFieldBase.CompareArgumentsAsExpressions(comparison, document);
            }

            ComparisonFieldResult comparisonResult = new ComparisonFieldResult(result ? "1" : "0");
            comparisonResult.CompareValue = result;

            return comparisonResult;
        }

        private static bool CompareArgumentsAsExpressions(FieldComparison comparison, RadFlowDocument document)
        {
            Expression leftExp = ExpressionParser.Parse(comparison.LeftArgument, document);
            Expression rightExp = ExpressionParser.Parse(comparison.RightArgument, document);

            double left = leftExp.GetResult().Value;
            double right = rightExp.GetResult().Value;
            double res = left - right;

            bool result = CompareCore(comparison.Operator, res);
            return result;
        }

        private static bool CompareArgumentsAsString(FieldComparison comparison)
        {
            // TODO: Add wild-cards logic here.
            int res = string.Compare(comparison.LeftArgument, comparison.RightArgument, StringComparison.Ordinal);

            return ComparisonFieldBase.CompareCore(comparison.Operator, res);
        }

        private static bool CompareCore(string op, double res)
        {
            bool result = false;
            switch (op)
            {
                case ExpressionConstants.Equal:
                    result = res == 0;
                    break;
                case ExpressionConstants.NotEqual:
                    result = res != 0;
                    break;
                case ExpressionConstants.GreaterThan:
                    result = res > 0;
                    break;
                case ExpressionConstants.GreaterThanOrEqualTo:
                    result = res >= 0;
                    break;
                case ExpressionConstants.LessThan:
                    result = res < 0;
                    break;
                case ExpressionConstants.LessThanOrEqualTo:
                    result = res <= 0;
                    break;
            }
            return result;
        }

        private static bool IsComaprisonOperatorvalid(string op)
        {
            return string.Equals(op, ExpressionConstants.Equal, System.StringComparison.Ordinal) ||
                   string.Equals(op, ExpressionConstants.NotEqual, System.StringComparison.Ordinal) ||
                   string.Equals(op, ExpressionConstants.GreaterThan, System.StringComparison.Ordinal) ||
                   string.Equals(op, ExpressionConstants.GreaterThanOrEqualTo, System.StringComparison.Ordinal) ||
                   string.Equals(op, ExpressionConstants.LessThan, System.StringComparison.Ordinal) ||
                   string.Equals(op, ExpressionConstants.LessThanOrEqualTo, System.StringComparison.Ordinal);
        }
    }
}