﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.Model.Fields.Expressions.Functions;
using Telerik.Windows.Documents.Flow.Model.Fields.Expressions.Operators;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Fields.Expressions
{
    internal static class ExpressionParser
    {
        private static readonly TokenHandler[] TokenHandlers;

        static ExpressionParser()
        {
            ExpressionParser.TokenHandlers = new TokenHandler[]
            {
                ExpressionParser.ParseLiteral,
                ExpressionParser.ParseBookmark,
                ExpressionParser.ParseUnaryOperator,
                ExpressionParser.ParseBinaryOperator,
                ExpressionParser.ParseFunctionStart,
                ExpressionParser.ParseFunction,
            };
        }

        private delegate bool TokenHandler(ExpressionToken token,
            Stack<Expression> expressionStack,
            RadFlowDocument document);

        public static Expression Parse(string input, RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNullOrEmpty(input, "input");

            Expression result;

            try
            {
                Stack<Expression> expressionStack = new Stack<Expression>();

                ExpressionTokenizer tokenizer = new ExpressionTokenizer(input);
                IEnumerable<ExpressionToken> rpnTokens = ExpressionParser.InfixToReversedPolishNotation(tokenizer);

                foreach (ExpressionToken token in rpnTokens)
                {
                    ExpressionParser.ParseToken(token, expressionStack, document);
                }

                if (expressionStack.Count != 1)
                {
                    throw new ExpressionException("Spreadsheet_ErrorExpressions_GeneralErrorMessage");
                }

                result = expressionStack.Pop();
            }
            catch (ExpressionException expressionEx)
            {
                result = new ErrorExpression(expressionEx);
            }

            return result;
        }

        private static IEnumerable<ExpressionToken> InfixToReversedPolishNotation(ExpressionTokenizer tokenizer)
        {
            Stack<ExpressionToken> stack = new Stack<ExpressionToken>();

            ExpressionToken token = tokenizer.Read();
            while (token != null)
            {
                if (token.TokenType == ExpressionTokenType.Function)
                {
                    yield return ExpressionToken.FunctionStart;
                    stack.Push(token);
                }
                else if (token.TokenType == ExpressionTokenType.ListSeparator)
                {
                    foreach (ExpressionToken item in PopUntilTokenOfType(stack, ExpressionTokenType.LeftParenthesis))
                    {
                        yield return item;
                    }
                }
                else if (OperatorInfos.IsOperator(token.TokenType))
                {
                    foreach (ExpressionToken expressionToken in GetOperandsFromStack(stack, token))
                    {
                        yield return expressionToken;
                    }

                    stack.Push(token);
                }
                else if (token.TokenType == ExpressionTokenType.LeftParenthesis)
                {
                    stack.Push(token);
                }
                else if (token.TokenType == ExpressionTokenType.RightParenthesis)
                {
                    foreach (ExpressionToken item in PopUntilTokenOfType(stack, ExpressionTokenType.LeftParenthesis))
                    {
                        yield return item;
                    }
                    stack.Pop();

                    if (stack.Count != 0 && stack.Peek().TokenType == ExpressionTokenType.Function)
                    {
                        yield return stack.Pop();
                    }
                }
                else
                {
                    if (token.TokenType != ExpressionTokenType.Space)
                    {
                        yield return token;
                    }
                }

                token = tokenizer.Read();
            }

            while (stack.Count != 0)
            {
                yield return stack.Pop();
            }
        }

        private static IEnumerable<ExpressionToken> PopUntilTokenOfType(Stack<ExpressionToken> stack, ExpressionTokenType expectedTokenType)
        {
            while (stack.Count > 0 && stack.Peek().TokenType != expectedTokenType)
            {
                yield return stack.Pop();
            }

            if (stack.Count == 0)
            {
                throw new ExpressionException(string.Format("Could not find token of type: {0}", expectedTokenType));
            }
            ExpressionParser.EnsureTokenOfType(stack.Peek(), expectedTokenType);
        }

        private static void ParseToken(ExpressionToken token, Stack<Expression> expressionStack, RadFlowDocument document)
        {
            bool handled = false;

            foreach (TokenHandler tokenHandler in TokenHandlers)
            {
                if (tokenHandler(token, expressionStack, document))
                {
                    handled = true;
                    break;
                }
            }

            if (!handled)
            {
                throw new ExpressionException("Could not handle token: " + token.TokenType);
            }
        }

        private static bool ParseLiteral(ExpressionToken token, Stack<Expression> expressionStack, RadFlowDocument document)
        {
            Expression value = null;

            switch (token.TokenType)
            {
                case ExpressionTokenType.Number:
                    value = new NumberExpression(token.NumberValue);
                    break;

                case ExpressionTokenType.True:
                    value = BooleanExpression.True;
                    break;

                case ExpressionTokenType.False:
                    value = BooleanExpression.False;
                    break;

                default:
                    return false;
            }

            expressionStack.Push(value);
            return true;
        }

        private static bool ParseUnaryOperator(ExpressionToken token, Stack<Expression> expressionStack, RadFlowDocument document)
        {
            Expression[] operands;
            Expression expression;
            switch (token.TokenType)
            {
                case ExpressionTokenType.UnaryMinus:
                    operands = ExpressionParser.PopNOperands(expressionStack, 1);
                    expression = new UnaryMinusExpression(operands[0]);
                    break;

                case ExpressionTokenType.Percent:
                    operands = ExpressionParser.PopNOperands(expressionStack, 1);
                    expression = new PercentExpression(operands[0]);
                    break;

                default:
                    return false;
            }

            expressionStack.Push(expression);

            return true;
        }

        private static bool ParseBinaryOperator(ExpressionToken token, Stack<Expression> expressionStack, RadFlowDocument document)
        {
            Func<Expression[]> getOpreands = () =>
            {
                return ExpressionParser.PopNOperands(expressionStack, 2);
            };

            BinaryOperatorExpression expression;
            if (BinaryOperatorExpressionFactory.TryCreateExpression(token.TokenType, getOpreands, out expression))
            {
                expressionStack.Push(expression);
                return true;
            }
            else
            {
                return false;
            }
        }

        private static bool ParseFunctionStart(ExpressionToken token, Stack<Expression> expressionStack, RadFlowDocument document)
        {
            if (token.TokenType == ExpressionTokenType.FunctionStart)
            {
                expressionStack.Push(new FunctionStartExpression());
                return true;
            }

            return false;
        }

        private static bool ParseFunction(ExpressionToken token, Stack<Expression> expressionStack, RadFlowDocument document)
        {
            if (token.TokenType != ExpressionTokenType.Function)
            {
                return false;
            }

            Expression[] arguments = ExpressionParser.ReadFunctionArguments(expressionStack);

            FunctionExpression functionExpression = FunctionExpressionFactory.CreateFunctionExpression(token.Value, arguments);
            expressionStack.Push(functionExpression);

            return true;
        }

        private static bool ParseBookmark(ExpressionToken token, Stack<Expression> expressionStack, RadFlowDocument document)
        {
            if (token.TokenType == ExpressionTokenType.Bookmark)
            {
                BookmarkExpression bookmarkExpression = new BookmarkExpression(token.Value, document);
                expressionStack.Push(bookmarkExpression);
                return true;
            }

            return false;
        }

        private static Expression[] ReadFunctionArguments(Stack<Expression> expressionStack)
        {
            Stack<Expression> result = new Stack<Expression>();
            while (expressionStack.Count > 0 && !(expressionStack.Peek() is FunctionStartExpression))
            {
                result.Push(expressionStack.Pop());
            }

            if (expressionStack.Count == 0 || !(expressionStack.Peek() is FunctionStartExpression))
            {
                throw new ExpressionException("Invalid function.");
            }

            expressionStack.Pop();

            return result.ToArray();
        }

        private static void EnsureTokenOfType(ExpressionToken token, ExpressionTokenType tokenType)
        {
            if (token.TokenType != tokenType)
            {
                throw new ExpressionException(string.Format("Expected token not found: {0}", tokenType));
            }
        }

        private static Expression[] PopNOperands(Stack<Expression> expressionStack, int n)
        {
            Stack<Expression> result = new Stack<Expression>();
            while (expressionStack.Count > 0 && n > 0)
            {
                result.Push(expressionStack.Pop());
                n--;
            }

            if (n != 0)
            {
                throw new ExpressionException("Not enough operands.");
            }

            return result.ToArray();
        }

        private static IEnumerable<ExpressionToken> GetOperandsFromStack(Stack<ExpressionToken> stack, ExpressionToken token)
        {
            while (stack.Count > 0 && OperatorInfos.IsOperator(stack.Peek().TokenType))
            {
                OperatorAssociativity tokenAssociativity = OperatorInfos.GetAssociativity(token.TokenType);
                int precedenceCompare = OperatorInfos.ComparePrecedence(token.TokenType, stack.Peek().TokenType);
                if ((tokenAssociativity == OperatorAssociativity.Left && precedenceCompare >= 0) ||
                    (tokenAssociativity == OperatorAssociativity.Right && precedenceCompare > 0))
                {
                    yield return stack.Pop();
                }
                else
                {
                    break;
                }
            }
        }
    }
}
