﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Text;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Fields.Expressions
{
    internal class ExpressionTokenizer
    {
        private static readonly Dictionary<string, ExpressionToken> StringToExpressionToken;
        private static readonly Dictionary<string, ExpressionToken> StringToBinaryOperatorExpressionToken;
        private static readonly Dictionary<string, ExpressionToken> StringToComparisonOperatorExpressionToken;

        private readonly TokenStringReader tokenReader;
        private readonly List<Func<ExpressionToken>> tokenHandlers;
        private readonly List<Func<string, ExpressionToken>> textTokenHandlers;

        private ExpressionToken previousNonSpaceToken;

        static ExpressionTokenizer()
        {
            StringToExpressionToken = new Dictionary<string, ExpressionToken>();
            StringToExpressionToken.Add(ExpressionToken.LeftParenthesis.Value, ExpressionToken.LeftParenthesis);
            StringToExpressionToken.Add(ExpressionToken.RightParenthesis.Value, ExpressionToken.RightParenthesis);

            StringToBinaryOperatorExpressionToken = new Dictionary<string, ExpressionToken>();
            StringToBinaryOperatorExpressionToken.Add(ExpressionToken.Plus.Value, ExpressionToken.Plus);
            StringToBinaryOperatorExpressionToken.Add(ExpressionToken.Minus.Value, ExpressionToken.Minus);
            StringToBinaryOperatorExpressionToken.Add(ExpressionToken.Multiply.Value, ExpressionToken.Multiply);
            StringToBinaryOperatorExpressionToken.Add(ExpressionToken.Divide.Value, ExpressionToken.Divide);
            StringToBinaryOperatorExpressionToken.Add(ExpressionToken.Power.Value, ExpressionToken.Power);

            StringToComparisonOperatorExpressionToken = new Dictionary<string, ExpressionToken>();
            StringToComparisonOperatorExpressionToken.Add(ExpressionToken.Equal.Value, ExpressionToken.Equal);
            StringToComparisonOperatorExpressionToken.Add(ExpressionToken.NotEqual.Value, ExpressionToken.NotEqual);
            StringToComparisonOperatorExpressionToken.Add(ExpressionToken.GreaterThan.Value, ExpressionToken.GreaterThan);
            StringToComparisonOperatorExpressionToken.Add(ExpressionToken.GreaterThanOrEqualTo.Value, ExpressionToken.GreaterThanOrEqualTo);
            StringToComparisonOperatorExpressionToken.Add(ExpressionToken.LessThan.Value, ExpressionToken.LessThan);
            StringToComparisonOperatorExpressionToken.Add(ExpressionToken.LessThanOrEqualTo.Value, ExpressionToken.LessThanOrEqualTo);
        }

        public ExpressionTokenizer(string input)
        {
            Guard.ThrowExceptionIfNullOrEmpty(input, "input");

            this.tokenReader = new TokenStringReader(input);

            this.tokenHandlers = new List<Func<ExpressionToken>>();
            this.textTokenHandlers = new List<Func<string, ExpressionToken>>();

            this.InitTokenHandlers();
        }

        public ExpressionToken Read()
        {
            ExpressionToken resultToken = null;

            resultToken = this.ReadInternal();

            if (resultToken != null && resultToken.TokenType != ExpressionTokenType.Space)
            {
                this.previousNonSpaceToken = resultToken;
            }

            return resultToken;
        }

        private static bool IsLiteral(string input, params string[] literals)
        {
            string inputToUpper = input.ToUpper(CultureInfo.InvariantCulture);
            for (int i = 0; i < literals.Length; i++)
            {
                if (inputToUpper == literals[i])
                {
                    return true;
                }
            }

            return false;
        }

        private static bool IsValidFunctionCellReferenceOrVariableChar(char nextChar)
        {
            if (ExpressionTokenizer.IsCharEqualTo(nextChar, ExpressionConstants.LeftParenthesis))
            {
                return false;
            }

            if (ExpressionTokenizer.IsValidFunctionChar(nextChar) || ExpressionTokenizer.IsValidNonEscapedCellReferenceChar(nextChar))
            {
                return true;
            }

            return ExpressionTokenizer.IsValidBookmarkChar(nextChar);
        }

        private static bool IsValidBookmarkChar(char nextChar)
        {
            return !ExpressionTokenizer.IsComparisonOperatorFirstChar(nextChar) &&
                   !ExpressionTokenizer.StringToBinaryOperatorExpressionToken.ContainsKey(nextChar.ToString()) &&
                   !ExpressionTokenizer.StringToExpressionToken.ContainsKey(nextChar.ToString()) &&
                   !ExpressionTokenizer.IsCharEqualTo(nextChar,
                       ExpressionConstants.TextQualifier,
                       ExpressionConstants.Percent,
                       ExpressionConstants.Space,
                       ExpressionConstants.NumberDecimalSeparator,
                       ExpressionConstants.SingleQuote,
                       "#",
                       "!",
                       "|",
                       "{",
                       "}",
                       "[",
                       "]",
                       "`",
                       "~",
                       ";",
                       "@",
                       ",");
        }

        private static bool IsValidFunctionChar(char nextChar)
        {
            return char.IsLetter(nextChar);
        }

        private static bool IsValidNonEscapedCellReferenceChar(char nextChar)
        {
            return char.IsLetterOrDigit(nextChar) ||
                   ExpressionTokenizer.IsCharEqualTo(nextChar, "_", "#", "!", "$");
        }

        private static bool IsComparisonOperatorFirstChar(char firstChar)
        {
            return ExpressionTokenizer.IsCharEqualTo(firstChar,
                ExpressionConstants.LessThan,
                ExpressionConstants.GreaterThan,
                ExpressionConstants.Equal);
        }

        private static bool IsComparisonOperatorSecondChar(char firstChar, char secondChar)
        {
            return (firstChar != ExpressionConstants.Equal[0] && secondChar == ExpressionConstants.Equal[0]) ||
                   (firstChar == ExpressionConstants.LessThan[0] && secondChar == ExpressionConstants.GreaterThan[0]);
        }

        private static bool TryParseNumber(string input, out double number)
        {
            bool parsed = double.TryParse(input, NumberStyles.Number, CultureInfo.InvariantCulture, out number);
            return parsed;
        }

        private static bool IsDecimalSeparator(char ch)
        {
            return ch.ToString() == ExpressionConstants.NumberDecimalSeparator;
        }

        private static bool IsCharEqualTo(char ch, params string[] values)
        {
            string character = ch.ToString();
            for (int i = 0; i < values.Length; i++)
            {
                if (character == values[i])
                {
                    return true;
                }
            }

            return false;
        }

        private void InitTokenHandlers()
        {
            this.tokenHandlers.Add(() => { return this.ReadSpaceToken(); });
            this.tokenHandlers.Add(() => { return this.ReadNumberToken(); });
            this.tokenHandlers.Add(() => { return this.ReadUnaryPlusMinusToken(); });
            this.tokenHandlers.Add(() => { return this.ReadPercentOperatorToken(); });
            this.tokenHandlers.Add(() => { return this.ReadBinaryOperatorToken(); });
            this.tokenHandlers.Add(() => { return this.ReadParenthesisToken(); });
            this.tokenHandlers.Add(() => { return this.ReadListSeparatorToken(); });
            this.tokenHandlers.Add(() => { return this.ReadFunctionOrBookmarkToken(); });

            this.textTokenHandlers.Add((value) => { return this.ReadLiteralToken(value); });
            this.textTokenHandlers.Add((value) => { return this.ReadFunctionToken(value); });
            this.textTokenHandlers.Add((value) => { return this.ReadBookmarkToken(value); });
        }

        private ExpressionToken ReadInternal()
        {
            ExpressionToken result = null;

            if (this.tokenReader.EndOfFile)
            {
                return null;
            }

            foreach (Func<ExpressionToken> tokenHandler in this.tokenHandlers)
            {
                result = tokenHandler();
                if (result != null)
                {
                    break;
                }
            }

            if (!this.tokenReader.EndOfFile)
            {
                if (result == null)
                {
                    throw new ExpressionException("Unexpected symbol.");
                }
            }

            return result;
        }

        private ExpressionToken ReadSpaceToken()
        {
            ExpressionToken result = null;

            char next = (char)this.tokenReader.Peek();
            if (ExpressionTokenizer.IsCharEqualTo(next, ExpressionConstants.Space))
            {
                StringBuilder builder = new StringBuilder();

                while (!this.tokenReader.EndOfFile && ExpressionTokenizer.IsCharEqualTo(next, ExpressionConstants.Space))
                {
                    builder.Append(next);
                    this.tokenReader.Read();
                    next = (char)this.tokenReader.Peek();
                }

                result = new ExpressionToken(ExpressionTokenType.Space, builder.ToString());
            }

            return result;
        }

        private ExpressionToken ReadNumberToken()
        {
            ExpressionToken result = null;

            double numberResult;
            if (this.TryReadNumber(out numberResult))
            {
                string tokenStringValue = numberResult.ToString(CultureInfo.InvariantCulture);
                result = new ExpressionToken(numberResult, tokenStringValue);
            }

            return result;
        }

        private ExpressionToken ReadUnaryPlusMinusToken()
        {
            ExpressionToken result = null;
            char nextChar = (char)this.tokenReader.Peek();

            if (this.IsValidUnaryPlusMinusOperator())
            {
                // The only unary operator is "-"
                if (nextChar.ToString() == ExpressionToken.UnaryMinus.Value)
                {
                    result = ExpressionToken.UnaryMinus;
                    this.tokenReader.Read();

                    if (this.tokenReader.EndOfFile)
                    {
                        throw new ExpressionException("Unexpected end of expression.");
                    }
                }
            }

            return result;
        }

        private ExpressionToken ReadPercentOperatorToken()
        {
            ExpressionToken result = null;
            char nextChar = (char)this.tokenReader.Peek();

            if (this.IsValidPercentOperator())
            {
                if (nextChar.ToString() == ExpressionToken.Percent.Value)
                {
                    result = ExpressionToken.Percent;
                    this.tokenReader.Read();
                }
            }

            return result;
        }

        private ExpressionToken ReadBinaryOperatorToken()
        {
            ExpressionToken result = null;

            char nextChar = (char)this.tokenReader.Peek();

            if (ExpressionTokenizer.StringToBinaryOperatorExpressionToken.TryGetValue(nextChar.ToString(), out result))
            {
                this.tokenReader.Read();
            }
            else if (ExpressionTokenizer.IsComparisonOperatorFirstChar(nextChar) && !this.IsValidUnaryPlusMinusOperator())
            {
                StringBuilder builder = new StringBuilder();
                nextChar = this.ReadAppendAndGetNextChar(nextChar, builder);

                if (ExpressionTokenizer.IsComparisonOperatorSecondChar(builder[0], nextChar))
                {
                    builder.Append(nextChar);
                    this.tokenReader.Read();
                }

                ExpressionTokenizer.StringToComparisonOperatorExpressionToken.TryGetValue(builder.ToString(), out result);

                if (result == null)
                {
                    throw new ExpressionException("Unexpected binary operator token: " + builder.ToString());
                }
            }

            if (result != null && !this.IsValidBinaryOperator())
            {
                throw new ExpressionException("Invalid binary operator usage.");
            }

            return result;
        }

        private ExpressionToken ReadParenthesisToken()
        {
            ExpressionToken result = null;

            char nextChar = (char)this.tokenReader.Peek();
            if (ExpressionTokenizer.StringToExpressionToken.TryGetValue(nextChar.ToString(), out result))
            {
                this.tokenReader.Read();
            }

            return result;
        }

        private ExpressionToken ReadListSeparatorToken()
        {
            ExpressionToken result = null;
            char nextChar = (char)this.tokenReader.Peek();

            if (ExpressionTokenizer.IsCharEqualTo(nextChar, ExpressionConstants.ListSeparator))
            {
                result = new ExpressionToken(ExpressionTokenType.ListSeparator, ExpressionConstants.ListSeparator);
                this.tokenReader.Read();
            }

            return result;
        }

        private ExpressionToken ReadFunctionOrBookmarkToken()
        {
            ExpressionToken result = null;
            char nextChar = (char)this.tokenReader.Peek();

            if (ExpressionTokenizer.IsValidFunctionChar(nextChar) ||
                ExpressionTokenizer.IsValidNonEscapedCellReferenceChar(nextChar) ||
                ExpressionTokenizer.IsValidBookmarkChar(nextChar))
            {
                StringBuilder builder = new StringBuilder();
                bool isCharValid = true;

                while (!this.tokenReader.EndOfFile && isCharValid)
                {
                    nextChar = this.ReadAppendAndGetNextChar(nextChar, builder);

                    isCharValid = ExpressionTokenizer.IsValidFunctionCellReferenceOrVariableChar(nextChar);
                }

                string value = builder.ToString();
                Guard.ThrowExceptionIfNullOrEmpty(value, "value");

                foreach (Func<string, ExpressionToken> textTokenHandler in this.textTokenHandlers)
                {
                    result = textTokenHandler(value);
                    if (result != null)
                    {
                        break;
                    }
                }

                if (result == null)
                {
                    throw new ExpressionException("Could not read function or bookmark token.");
                }
            }

            return result;
        }

        private ExpressionToken ReadFunctionToken(string value)
        {
            ExpressionToken result = null;
            char nextChar = (char)this.tokenReader.Peek();

            // This is needed to handle cases when there are whitespaces between the function and left paranthaesis
            // The space token will lost - but we dpn't needed it anyway. 
            if (ExpressionTokenizer.IsCharEqualTo(nextChar, ExpressionConstants.Space))
            {
                this.ReadSpaceToken();
            }

            nextChar = (char)this.tokenReader.Peek();
            if (ExpressionTokenizer.IsCharEqualTo(nextChar, ExpressionConstants.LeftParenthesis))
            {
                result = new ExpressionToken(ExpressionTokenType.Function, value);
            }

            return result;
        }

        private ExpressionToken ReadLiteralToken(string value)
        {
            if (ExpressionTokenizer.IsLiteral(value, ExpressionConstants.TrueLiteral))
            {
                return ExpressionToken.True;
            }
            else if (ExpressionTokenizer.IsLiteral(value, ExpressionConstants.FalseLiteral))
            {
                return ExpressionToken.False;
            }

            return null;
        }

        private ExpressionToken ReadBookmarkToken(string value)
        {
            ExpressionToken result = null;
            if (Bookmark.IsValidBookmarkName(value, true))
            {
                result = new ExpressionToken(ExpressionTokenType.Bookmark, value);
            }
            return result;
        }

        private bool IsValidUnaryPlusMinusOperator()
        {
            if (this.previousNonSpaceToken == null)
            {
                return true;
            }

            return this.IsValidUnaryOperatorPrecedingToken();
        }

        private bool IsValidUnaryOperatorPrecedingToken()
        {
            return this.previousNonSpaceToken.TokenType != ExpressionTokenType.Number &&
                   this.previousNonSpaceToken.TokenType != ExpressionTokenType.True &&
                   this.previousNonSpaceToken.TokenType != ExpressionTokenType.False &&
                   this.previousNonSpaceToken.TokenType != ExpressionTokenType.RightParenthesis &&
                   this.previousNonSpaceToken.TokenType != ExpressionTokenType.Percent &&
                   this.previousNonSpaceToken.TokenType != ExpressionTokenType.Bookmark;
        }

        private bool IsValidPercentOperator()
        {
            return this.previousNonSpaceToken != null &&
                   this.previousNonSpaceToken.TokenType != ExpressionTokenType.Bookmark &&
                   !this.IsValidUnaryOperatorPrecedingToken();
        }

        private bool IsValidBinaryOperator()
        {
            if (this.tokenReader.EndOfFile || this.previousNonSpaceToken == null)
            {
                return false;
            }

            char nextChar = (char)this.tokenReader.Peek();
            return !ExpressionTokenizer.IsCharEqualTo(nextChar,
                ExpressionConstants.RightParenthesis,
                ExpressionConstants.Multiplication,
                ExpressionConstants.Division,
                ExpressionConstants.Power);
        }

        private bool TryReadNumber(out double numberResult)
        {
            bool result = false;
            numberResult = 0;
            char nextChar = (char)this.tokenReader.Peek();

            if (char.IsDigit(nextChar) || ExpressionTokenizer.IsDecimalSeparator(nextChar))
            {
                StringBuilder builder = new StringBuilder();

                while (char.IsDigit(nextChar) || ExpressionTokenizer.IsDecimalSeparator(nextChar))
                {
                    nextChar = this.ReadAppendAndGetNextChar(nextChar, builder);
                }

                if (!ExpressionTokenizer.TryParseNumber(builder.ToString(), out numberResult))
                {
                    throw new ExpressionException("Could not parse number: " + builder.ToString());
                }

                result = true;
            }

            return result;
        }

        private char ReadAppendAndGetNextChar(char nextChar, StringBuilder builder)
        {
            this.tokenReader.Read();
            builder.Append(nextChar);
            return (char)this.tokenReader.Peek();
        }
    }
}