using System;
using System.Globalization;
using Telerik.Windows.Documents.Flow.Model.Fields.FieldCode;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Fields
{
    /// <summary>
    /// A base class for fields that can produce result fragment.
    /// </summary>
    public abstract class Field
    {
        private static readonly string DateFormattingSwitch = "@";
        private static readonly string NumericFormattingSwitch = "#";
        private static readonly string GeneralFormattingSwitch = "*";

        /// <summary>
        /// Initializes a new instance of the <see cref="Field" /> class.
        /// </summary>
        /// <param name="document">The document.</param>
        protected Field(RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");

            this.Document = document;
        }

        /// <summary>
        /// Gets the date time formatting.
        /// </summary>
        /// <value>The date time formatting.</value>
        public string DateTimeFormatting { get; private set; }

        /// <summary>
        /// Gets the numeric formatting.
        /// </summary>
        /// <value>The numeric formatting.</value>
        public string NumericFormatting { get; private set; }

        /// <summary>
        /// Gets the general formatting.
        /// </summary>
        /// <value>The general formatting.</value>
        public string GeneralFormatting { get; private set; }

        /// <summary>
        /// Gets the document to which this field is associated.
        /// </summary>
        /// <value>
        /// The document.
        /// </value>
        public RadFlowDocument Document { get; private set; }

        internal virtual bool ExpectComparisonArgument
        {
            get
            {
                return false;
            }
        }

        internal virtual bool PreserveResultOnUpdate
        {
            get
            {
                return false;
            }
        }

        internal void LoadParameters(FieldParameters parameters)
        {
            this.DateTimeFormatting = parameters.GetSwitchArgument(Field.DateFormattingSwitch);
            this.NumericFormatting = parameters.GetSwitchArgument(Field.NumericFormattingSwitch);
            this.GeneralFormatting = parameters.GetSwitchArgument(Field.GeneralFormattingSwitch);

            // TODO: set generic formatting
            this.LoadParametersOverride(parameters);
        }

        internal bool IsSwitchWithArgument(string sw)
        {
            if (sw.Length == 1 &&
                (sw[0] == FieldParserConstants.GeneralFormattingSwitch ||
                 sw[0] == FieldParserConstants.NumericFormattingSwitch ||
                 sw[0] == FieldParserConstants.DateFormattingSwitch))
            {
                return true;
            }

            return this.IsSwitchWithArgumentOverride(sw);
        }

        internal virtual void LoadParametersOverride(FieldParameters parameters)
        {
        }

        /// <summary>
        /// Gets the field result.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns></returns>
        internal abstract FieldResult GetFieldResult(FieldUpdateContext context);

        /// <summary>
        /// Updates field result fragment.
        /// </summary>
        /// <param name="context">The context.</param>
        internal bool UpdateResult(FieldUpdateContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            FieldResult fieldResult = this.GetFieldResult(context);

            if (fieldResult != null)
            {
                UpdateResult(context, fieldResult);
                return true;
            }
            else
            {
                return false;
            }
        }

        /// <summary>
        /// Encodes a parameter.
        /// </summary>
        /// <param name="parameter">The parameter.</param>
        /// <returns></returns>
        protected static string EncodeParameter(string parameter)
        {
            return TextHelper.AddSlashes(parameter);
        }

        /// <summary>
        /// Gets the formatted date.
        /// </summary>
        /// <param name="date">The date.</param>
        /// <returns></returns>
        protected string GetFormattedDate(DateTime date)
        {
            string result = string.Empty;
            string dateFormatting;

            if (!string.IsNullOrEmpty(this.DateTimeFormatting))
            {
                dateFormatting = this.DateTimeFormatting.Replace("am/pm", "tt").Replace("AM/PM", "tt");
            }
            else
            {
                dateFormatting = DateTimeFormatInfo.CurrentInfo.ShortDatePattern;
            }

            result = date.ToString(dateFormatting, CultureInfo.InvariantCulture);
            return result;
        }

        /// <summary>
        /// Determines whether a switch can have an argument.
        /// </summary>
        /// <param name="switchKey">The switch key.</param>
        /// <returns>True if the switch can have argument.</returns>
        protected virtual bool IsSwitchWithArgumentOverride(string switchKey)
        {
            return false;
        }

        private static void UpdateResult(FieldUpdateContext context, FieldResult fieldResult)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(fieldResult, "fieldResult");

            FieldCharacter endInline = context.FieldInfo.Separator ?? context.FieldInfo.End;
            CharacterProperties resultCharacterFormatting = InlineRangeEditor.GetCharacterFormattingFromFirstRunInRange(context.FieldInfo.Start, endInline);

            context.Editor.CharacterFormatting.CopyPropertiesFrom(resultCharacterFormatting);
            context.Editor.InsertText(fieldResult.Result);
        }
    }
}