﻿using System;
using System.Text;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Fields.FieldCode
{
    internal static class FieldCodeParser
    {
        public static FieldCodeParseResult ParseField(FieldInfo fieldInfo)
        {
            Guard.ThrowExceptionIfNull(fieldInfo, "field");
            if (fieldInfo.Start == null || fieldInfo.End == null)
            {
                throw new InvalidOperationException("There are no field characters attached to field");
            }

            FieldCharacter codeEnd = null;
            if (fieldInfo.Separator != null && fieldInfo.Separator.Parent != null)
            {
                codeEnd = fieldInfo.Separator;
            }
            else
            {
                codeEnd = fieldInfo.End;
            }

            FieldTokenizer tokenizer = new FieldTokenizer(fieldInfo.Start, codeEnd);
            Field field = FieldCodeParser.CreateField(tokenizer, fieldInfo.Document);
            FieldParameters fieldParams = FieldCodeParser.ReadFieldArguments(field, tokenizer);

            return new FieldCodeParseResult(field, fieldParams);
        }

        private static Field CreateField(FieldTokenizer tokenizer, RadFlowDocument document)
        {
            FieldToken typeToken = tokenizer.Read();
            if (typeToken == null ||
                !(typeToken.TokenType == FieldTokenType.Argument || typeToken.TokenType == FieldTokenType.ExpressionStart))
            {
                ParseErrorField errorField = new ParseErrorField(document, "Unknown token type parsed.");
                return errorField;
            }

            Field field = null;
            if (typeToken.TokenType == FieldTokenType.ExpressionStart)
            {
                field = new ExpressionField(document);
            }
            else
            {
                field = FieldFactory.CreateField(typeToken.Value, document);
            }

            return field;
        }

        private static FieldParameters ReadFieldArguments(Field field, FieldTokenizer tokenizer)
        {
            FieldParameters parameters = new FieldParameters();
            if (field is ParseErrorField)
            {
                return parameters;
            }

            // Read expression
            if (field is ExpressionField)
            {
                parameters.Expression = FieldCodeParser.ReadExpression(tokenizer);
            }

            // Read Comparison if the field expects one
            if (field.ExpectComparisonArgument)
            {
                parameters.Comparison = FieldCodeParser.ReadComparison(tokenizer);
            }

            // Read rest of the arguments.
            FieldToken nextToken = tokenizer.Read();
            FieldSwitch switchExpectingArgument = null;
            while (nextToken != null && nextToken.TokenType != FieldTokenType.Error)
            {
                if (nextToken.TokenType == FieldTokenType.Switch)
                {
                    switchExpectingArgument = null;

                    FieldSwitch fieldSwitch = new FieldSwitch(nextToken.Value);
                    parameters.AddSwitch(fieldSwitch);
                    if (field.IsSwitchWithArgument(fieldSwitch.SwitchValue))
                    {
                        switchExpectingArgument = fieldSwitch;
                    }
                }
                else if (nextToken.TokenType == FieldTokenType.Argument)
                {
                    if (switchExpectingArgument != null)
                    {
                        switchExpectingArgument.Argument = new FieldArgument(nextToken.Value);
                        switchExpectingArgument = null;
                    }
                    else
                    {
                        parameters.AddArgument(new FieldArgument(nextToken.Value));
                    }
                }
                nextToken = tokenizer.Read();
            }

            return parameters;
        }

        private static string ReadExpression(FieldTokenizer tokenizer)
        {
            StringBuilder sb = new StringBuilder();
            FieldToken nextToken = tokenizer.Read();
            while (nextToken != null &&
                   nextToken.TokenType != FieldTokenType.Error &&
                   IsTokenPartOfExpression(nextToken, true))
            {
                sb.Append(nextToken.Value);
                nextToken = tokenizer.Read();
            }

            return sb.ToString();
        }

        private static bool IsTokenPartOfExpression(FieldToken token, bool acceptWhiteSpaces)
        {
            if (token.TokenType == FieldTokenType.Argument)
            {
                return true;
            }

            if (acceptWhiteSpaces && token.TokenType == FieldTokenType.WhiteSpace)
            {
                return true;
            }

            return false;
        }

        private static FieldComparison ReadComparison(FieldTokenizer tokenizer)
        {
            FieldToken leftToken = tokenizer.Read();
            FieldToken operatorToken = tokenizer.Read();
            FieldToken rightToken = tokenizer.Read();

            string left = string.Empty;
            bool isLeftQuoted = false;
            string op = string.Empty;
            string right = string.Empty;
            bool isRightQuoted = false;

            if (leftToken != null && leftToken.TokenType == FieldTokenType.Argument)
            {
                left = leftToken.Value;
                isLeftQuoted = leftToken.IsQuoted;
            }

            if (operatorToken != null && operatorToken.TokenType == FieldTokenType.Argument)
            {
                op = operatorToken.Value;
            }

            if (rightToken != null && rightToken.TokenType == FieldTokenType.Argument)
            {
                right = rightToken.Value;
                isRightQuoted = rightToken.IsQuoted;
            }

            return new FieldComparison(left, isLeftQuoted, op, right, isRightQuoted);
        }
    }
}