using System.Collections.Generic;
using System.Linq;

namespace Telerik.Windows.Documents.Flow.Model.Fields.FieldCode
{
    /// <summary>
    /// Represents a parsed collection of field arguments and switches.
    /// </summary>
    public class FieldParameters
    {
        private readonly Dictionary<string, FieldSwitch> switches;
        private readonly List<FieldArgument> fieldArguments;

        /// <summary>
        /// Initializes a new instance of the <see cref="FieldParameters" /> class.
        /// </summary>
        public FieldParameters()
        {
            this.switches = new Dictionary<string, FieldSwitch>();
            this.fieldArguments = new List<FieldArgument>();
        }

        /// <summary>
        /// Gets the first field argument.
        /// </summary>
        /// <value>The first field argument.</value>
        public FieldArgument FirstArgument
        {
            get
            {
                return this.fieldArguments.FirstOrDefault();
            }
        }

        /// <summary>
        /// Gets the second field argument.
        /// </summary>
        /// <value>The second field argument.</value>
        public FieldArgument SecondArgument
        {
            get
            {
                if (this.fieldArguments.Count < 2)
                {
                    return null;
                }
                return this.fieldArguments[1];
            }
        }

        /// <summary>
        /// Gets the expression string.
        /// </summary>
        /// <value>The expression.</value>
        public string Expression { get; internal set; }

        /// <summary>
        /// Gets the field comparison element.
        /// </summary>
        /// <value>The comparison.</value>
        public FieldComparison Comparison { get; internal set; }

        /// <summary>
        /// Tries to get a switch with a given key.
        /// </summary>
        /// <param name="switchKey">The switch key.</param>
        /// <param name="result">The result.</param>
        /// <returns></returns>
        public bool TryGetSwitch(string switchKey, out FieldSwitch result)
        {
            return this.switches.TryGetValue(switchKey, out result);
        }

        /// <summary>
        /// Gets the switch argument.
        /// </summary>
        /// <param name="switchKey">The switch key.</param>
        /// <returns>The switch argument value if there is such switch, otherwise - null.</returns>
        public string GetSwitchArgument(string switchKey)
        {
            string result = null;

            FieldSwitch requestedSwitch = null;
            if (this.TryGetSwitch(switchKey.ToString(), out requestedSwitch))
            {
                result = requestedSwitch.Argument != null ? requestedSwitch.Argument.Value : string.Empty;
            }

            return result;
        }

        /// <summary>
        /// Determines whether a switch is defined.
        /// </summary>
        /// <param name="switchKey">The switch key.</param>
        /// <returns>True if the switch is defined.</returns>
        public bool IsSwitchDefined(string switchKey)
        {
            return this.switches.ContainsKey(switchKey.ToString());
        }

        internal void AddSwitch(FieldSwitch fieldSwitch)
        {
            this.switches[fieldSwitch.SwitchValue] = fieldSwitch;
        }

        internal void AddArgument(FieldArgument value)
        {
            this.fieldArguments.Add(value);
        }
    }
}