using System;
using System.Collections.Generic;
using System.Text;

namespace Telerik.Windows.Documents.Flow.Model.Fields.FieldCode
{
    internal class FieldTokenizer
    {
        private readonly TokenStringReader tokenReader;
        private readonly List<Func<FieldToken>> tokenHandlers;
        private bool isFirstNonWhitespaceToken = true;

        public FieldTokenizer(FieldCharacter start, FieldCharacter end)
        {
            this.tokenReader = new TokenStringReader(start, end);

            this.tokenHandlers = new List<Func<FieldToken>>();
            this.tokenHandlers.Add(() => { return this.ReadSwitchToken(); });
            this.tokenHandlers.Add(() => { return this.ReadArgumentToken(); });
            this.tokenHandlers.Add(() => { return this.ReadWhiteSpaceToken(); });
            this.tokenHandlers.Add(() => { return this.ReadExpressionStart(); });
        }

        public FieldToken Read(bool skipWhitesapce = true)
        {
            FieldToken result = this.ReadInternal();
            if (skipWhitesapce)
            {
                while (result != null && result.TokenType == FieldTokenType.WhiteSpace)
                {
                    result = this.ReadInternal();
                }
            }
            return result;
        }

        private static bool IsValidSwitchChar(char nextChar)
        {
            return (nextChar >= 'a' && nextChar <= 'z') ||
                   IsCharEqualTo(nextChar,
                       FieldParserConstants.DateFormattingSwitch,
                       FieldParserConstants.NumericFormattingSwitch,
                       FieldParserConstants.GeneralFormattingSwitch,
                       FieldParserConstants.PreventUpdateSwitch);
        }

        private static bool IsCharEqualTo(char character, params char[] values)
        {
            for (int i = 0; i < values.Length; i++)
            {
                if (character == values[i])
                {
                    return true;
                }
            }

            return false;
        }

        private FieldToken ReadInternal()
        {
            FieldToken result = null;

            if (this.tokenReader.EndOfFile)
            {
                return null;
            }

            foreach (Func<FieldToken> tokenHandler in this.tokenHandlers)
            {
                result = tokenHandler();
                if (result != null)
                {
                    break;
                }
            }

            if (this.isFirstNonWhitespaceToken &&
                result != null &&
                result.TokenType != FieldTokenType.WhiteSpace)
            {
                this.isFirstNonWhitespaceToken = false;
            }

            if (!this.tokenReader.EndOfFile)
            {
                if (result == null)
                {
                    result = new FieldToken(FieldTokenType.Error, string.Empty);
                    this.tokenReader.MoveToEnd();
                }
            }

            return result;
        }

        private FieldToken ReadSwitchToken()
        {
            FieldToken result = null;

            char nextChar = (char)this.tokenReader.Peek();
            if (FieldTokenizer.IsCharEqualTo(nextChar, FieldParserConstants.Backslash))
            {
                this.tokenReader.Read();

                nextChar = (char)this.tokenReader.Peek();
                StringBuilder sb = new StringBuilder();
                while (FieldTokenizer.IsValidSwitchChar(nextChar))
                {
                    this.tokenReader.Read();
                    sb.Append(nextChar);
                    nextChar = (char)this.tokenReader.Peek();
                }

                result = new FieldToken(FieldTokenType.Switch, sb.ToString());
            }

            return result;
        }

        private FieldToken ReadArgumentToken()
        {
            FieldToken result = null;

            char nextChar = (char)this.tokenReader.Peek();

            if (FieldTokenizer.IsCharEqualTo(nextChar, FieldParserConstants.Quote))
            {
                this.tokenReader.Read();

                string argument = this.ReadArgumentEscapedSequence();

                result = new FieldToken(FieldTokenType.Argument, argument, true);
            }
            else if (this.IsValidNonEscapedArgumentChar(nextChar))
            {
                string argument = this.ReadArgumentNonEscapedSequence();

                result = new FieldToken(FieldTokenType.Argument, argument);
            }

            return result;
        }

        private FieldToken ReadWhiteSpaceToken()
        {
            FieldToken result = null;
            {
                char nextChar = (char)this.tokenReader.Peek();

                if (char.IsWhiteSpace(nextChar))
                {
                    StringBuilder sb = new StringBuilder();

                    while (!this.tokenReader.EndOfFile &&
                           char.IsWhiteSpace(nextChar))
                    {
                        sb.Append(nextChar);
                        this.tokenReader.Read();
                        nextChar = (char)this.tokenReader.Peek();
                    }

                    result = new FieldToken(FieldTokenType.WhiteSpace, sb.ToString());
                }
            }

            return result;
        }

        private FieldToken ReadExpressionStart()
        {
            FieldToken result = null;
            if (this.isFirstNonWhitespaceToken)
            {
                char nextChar = (char)this.tokenReader.Peek();

                if (IsCharEqualTo(nextChar, FieldParserConstants.ExpressionStart))
                {
                    this.tokenReader.Read();
                    result = new FieldToken(FieldTokenType.ExpressionStart, nextChar.ToString());
                }
            }

            return result;
        }

        private string ReadArgumentNonEscapedSequence()
        {
            StringBuilder sb = new StringBuilder();
            char nextChar = (char)this.tokenReader.Peek();

            while (!this.tokenReader.EndOfFile &&
                   this.IsValidNonEscapedArgumentChar(nextChar))
            {
                sb.Append(nextChar);
                this.tokenReader.Read();
                nextChar = (char)this.tokenReader.Peek();
            }
            return sb.ToString();
        }

        private string ReadArgumentEscapedSequence()
        {
            StringBuilder sb = new StringBuilder();

            bool previousIsBackslash = false;
            while (!this.tokenReader.EndOfFile)
            {
                char nextChar = (char)this.tokenReader.Read();

                if (nextChar == FieldParserConstants.Quote && !previousIsBackslash)
                {
                    break;
                }
                else if (nextChar == FieldParserConstants.Backslash)
                {
                    if (previousIsBackslash)
                    {
                        sb.Append(nextChar);
                        previousIsBackslash = false;
                    }
                    else
                    {
                        previousIsBackslash = true;
                    }
                }
                else
                {
                    sb.Append(nextChar);
                    previousIsBackslash = false;
                }
            }
            return sb.ToString();
        }

        private bool IsValidNonEscapedArgumentChar(char nextChar)
        {
            if (this.isFirstNonWhitespaceToken && IsCharEqualTo(nextChar, FieldParserConstants.ExpressionStart))
            {
                return false;
            }

            return !char.IsWhiteSpace(nextChar) &&
                   !IsCharEqualTo(nextChar, FieldParserConstants.Quote);
        }
    }
}