﻿using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Fields
{
    internal class FieldContext
    {
        private readonly List<FieldCharacterInfo> currentFieldCharacters;
        private readonly bool isImporting;

        public FieldContext(bool isImporting)
        {
            this.isImporting = isImporting;
            this.currentFieldCharacters = new List<FieldCharacterInfo>();
        }

        public bool IsInInstruction
        {
            get
            {
                if (!this.currentFieldCharacters.Any())
                {
                    return false;
                }

                bool isLastCharStart = this.currentFieldCharacters.Last().Character.FieldCharacterType == FieldCharacterType.Start;

                return isLastCharStart;
            }
        }

        protected IEnumerable<FieldCharacter> CurrentFieldCharacters
        {
            get
            {
                return this.currentFieldCharacters.Select(info => info.Character);
            }
        }

        public void OnFieldCharacter(FieldCharacter fieldChar)
        {
            FieldCharacterInfo charInfo = new FieldCharacterInfo(fieldChar);
            switch (fieldChar.FieldCharacterType)
            {
                case FieldCharacterType.Start:
                    this.currentFieldCharacters.Add(charInfo);
                    break;
                case FieldCharacterType.Separator:
                    if (this.IsInInstruction)
                    {
                        this.currentFieldCharacters.Add(charInfo);
                    }
                    else
                    {
                        this.OnInvalidSeparatorFieldCharacter(fieldChar);
                    }
                    break;
                case FieldCharacterType.End:
                    if (this.currentFieldCharacters.Any())
                    {
                        this.OnFieldEnd(fieldChar);
                    }
                    else
                    {
                        this.OnInvalidEndFieldCharacter(fieldChar);
                    }
                    break;
            }
        }

        public void SetIsLocked(bool isLocked)
        {
            FieldCharacterInfo lastInfo = this.GetCurrentFieldStartInfo();
            if (lastInfo == null)
            {
                Debug.Fail("Cannot find field start to apply IsLocked.");
                return;
            }

            lastInfo.IsLocked = isLocked;
        }

        public void SetIsDirty(bool isDirty)
        {
            FieldCharacterInfo lastInfo = this.GetCurrentFieldStartInfo();
            if (lastInfo == null)
            {
                Debug.Fail("Cannot find field start to apply IsDirty.");
                return;
            }

            lastInfo.IsDirty = isDirty;
        }

        protected virtual void OnInvalidSeparatorFieldCharacter(FieldCharacter fieldCharacter)
        {
            Debug.Fail("No currently opened fields.");
        }

        protected virtual void OnInvalidEndFieldCharacter(FieldCharacter fieldCharacter)
        {
            Debug.Fail("Cannot consume separate char if not in instruction fragment.");
        }

        private FieldCharacterInfo GetCurrentFieldStartInfo()
        {
            if (!this.currentFieldCharacters.Any())
            {
                return null;
            }

            FieldCharacterInfo result = this.currentFieldCharacters.Last();
            if (result.Character.FieldCharacterType != FieldCharacterType.Start)
            {
                return null;
            }
            return result;
        }

        private void OnFieldEnd(FieldCharacter end)
        {
            FieldCharacterInfo start;
            FieldCharacterInfo separate = this.Pop();

            if (separate.Character.FieldCharacterType == FieldCharacterType.Separator)
            {
                start = this.Pop();
            }
            else
            {
                start = separate;
                separate = null;
            }

            Debug.Assert(start.Character.FieldCharacterType == FieldCharacterType.Start, "Could not find start character");

            if (this.isImporting)
            {
                FieldInfo field = new FieldInfo(start.Character, (separate != null) ? separate.Character : null, end);
                field.IsLocked = start.IsLocked;
                field.IsDirty = start.IsDirty;
                field.LoadFieldFromCode();
            }
        }

        private FieldCharacterInfo Pop()
        {
            FieldCharacterInfo f = this.currentFieldCharacters[this.currentFieldCharacters.Count - 1];
            this.currentFieldCharacters.RemoveAt(this.currentFieldCharacters.Count - 1);
            return f;
        }

        private class FieldCharacterInfo
        {
            public FieldCharacterInfo(FieldCharacter character)
            {
                Guard.ThrowExceptionIfNull(character, "character");
                this.Character = character;
            }

            public FieldCharacter Character { get; private set; }

            public bool IsLocked { get; set; }

            public bool IsDirty { get; set; }
        }
    }
}