﻿using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Flow.Model.Fields.FieldCode;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Fields
{
    /// <summary>
    /// An class that holds information of a filed that is inserted in the document.
    /// </summary>
    public class FieldInfo
    {
        private readonly RadFlowDocument document;
        private Field field;

        /// <summary>
        /// Initializes a new instance of the <see cref="FieldInfo" /> class.
        /// </summary>
        /// <param name="document">The document.</param>
        public FieldInfo(RadFlowDocument document)
            : this(document, null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="FieldInfo" /> class.
        /// </summary>
        /// <param name="document">The document.</param>
        /// <param name="field">The field that should be associated to the field info.</param>
        internal FieldInfo(RadFlowDocument document, Field field)
        {
            Guard.ThrowExceptionIfNull(document, "document");

            this.document = document;
            this.Field = field;
            this.AttachFieldCharacters(
                new FieldCharacter(this.Document, FieldCharacterType.Start),
                new FieldCharacter(this.Document, FieldCharacterType.Separator),
                new FieldCharacter(this.Document, FieldCharacterType.End));
        }

        internal FieldInfo(FieldCharacter start, FieldCharacter separate, FieldCharacter end)
        {
            Guard.ThrowExceptionIfNull(start, "start");
            Guard.ThrowExceptionIfNull(end, "end");

            this.document = start.Document;
            if (separate == null)
            {
                separate = new FieldCharacter(this.Document, FieldCharacterType.Separator);
            }
            this.AttachFieldCharacters(start, separate, end);
        }

        /// <summary>
        /// Gets the start field character.
        /// </summary>
        /// <value>The start field character.</value>
        public FieldCharacter Start { get; private set; }

        /// <summary>
        /// Gets the separator field character.
        /// </summary>
        /// <value>The separator field character.</value>
        public FieldCharacter Separator { get; private set; }

        /// <summary>
        /// Gets the end field character.
        /// </summary>
        /// <value>The end field character.</value>
        public FieldCharacter End { get; private set; }

        /// <summary>
        /// Gets the current field.
        /// </summary>
        /// <value>The current field.</value>
        public Field Field
        {
            get
            {
                return this.field;
            }
            internal set
            {
                this.field = value;
            }
        }

        /// <summary>
        /// Gets or sets the value indicating that the field is locked. Locked fields are not updated.
        /// </summary>
        /// <value>The IsLocked value.</value>
        public bool IsLocked { get; set; }

        /// <summary>
        /// Gets or sets the value indicating if the contents of the field should be updated before they are displayed
        /// if this functionality is supported by the next processing application.
        /// </summary>
        /// <value>The IsDirty value.</value>
        public bool IsDirty { get; set; }

        /// <summary>
        /// Gets the document to which this field is associated.
        /// </summary>
        /// <value>
        /// The document.
        /// </value>
        public RadFlowDocument Document
        {
            get
            {
                return this.document;
            }
        }

        /// <summary>
        /// Gets the Separator if it is inserted in the document, otherwise gets the End FieldCharacter.
        /// </summary>
        internal FieldCharacter CodeEnd
        {
            get
            {
                if (this.Separator.Paragraph == null)
                {
                    return this.End;
                }

                return this.Separator;
            }
        }

        /// <summary>
        /// Updates the field. This method will also update all fields inside the code fragment of the current field.
        /// </summary>
        public void UpdateField()
        {
            FieldUpdateContext context = new FieldUpdateContext(this);
            this.UpdateFieldInternal(context);
        }

        /// <summary>
        /// Gets the current code as string. 
        /// Note that if there are nested fields - their result fragment will used to generate the value returned by this method.
        /// </summary>
        /// <returns>Current code as string.</returns>
        public string GetCode()
        {
            if (this.Start.Parent == null || this.End.Parent == null)
            {
                return string.Empty;
            }

            string result = InlineRangeEditor.GetTextInRange(this.Start, this.CodeEnd, true, false);
            return result;
        }

        /// <summary>
        /// Gets current result as string. 
        /// Note that if there are nested fields - their result fragment will be used to generate the value returned by this method.
        /// </summary>
        /// <returns>Current result as string.</returns>
        public string GetResult()
        {
            if (this.Separator == null
                || this.Separator.Parent == null
                || this.End.Parent == null)
            {
                return string.Empty;
            }

            string result = InlineRangeEditor.GetTextInRange(this.Separator, this.End, true, true);
            return result;
        }

        internal void UpdateFieldInternal(FieldUpdateContext context)
        {
            IEnumerable<FieldCharacter> nestedFieldCharacters = InlineRangeEditor.EnumerateInlinesInRange(this.Start, this.CodeEnd)
                                                                                 .Where(inline => inline.Type == DocumentElementType.FieldCharacter)
                                                                                 .Cast<FieldCharacter>();

            List<FieldCharacter> fieldsToUpdate = FieldUpdateScheduler.GetOrderedListOfFieldToUpdate(nestedFieldCharacters);

            // Update this field last
            fieldsToUpdate.Add(this.Start);

            foreach (FieldCharacter fieldStart in fieldsToUpdate)
            {
                fieldStart.FieldInfo.UpdateFieldCore(context);
            }
        }

        //// TODO: Add set code and result methods

        internal FieldCodeParseResult LoadFieldFromCode()
        {
            // Parse field code.
            FieldCodeParseResult parseResult = FieldCodeParser.ParseField(this);

            this.Field = parseResult.Field;
            this.Field.LoadParameters(parseResult.FieldParameters);

            return parseResult;
        }

        internal void UpdateFieldCore(FieldUpdateContext context)
        {
            // Do not update locked field.
            if (this.IsLocked)
            {
                return;
            }

            FieldCodeParseResult parseResult = this.LoadFieldFromCode();
            context.Parameters = parseResult.FieldParameters;
            IEnumerable<InlineBase> oldResultFragment = null;

            // Delete result if needed.
            if (this.Separator.Paragraph == null)
            {
                context.Editor.MoveToInlineStart(this.End);
                context.Editor.InsertInline(this.Separator);
            }
            else
            {
                if (!this.Field.PreserveResultOnUpdate)
                {
                    // We are caching the current result, so if UpdateResult() returns false later, we restore the deleted inlines range from the InlineCollection.
                    oldResultFragment = InlineRangeEditor.EnumerateFieldInlinesInRange(this.Separator, this.End).ToList();
                    InlineRangeEditor.DeleteContent(this.Separator, this.End);
                }
                context.Editor.MoveToInlineEnd(this.Separator);
            }

            // Give control to Field to insert result.
            bool isResultUpdated = this.Field.UpdateResult(context);
            if (!isResultUpdated && oldResultFragment != null)
            {
                InlineRangeEditor.InsertInlinesInRange(this.Separator, oldResultFragment);
            }
        }

        /// <summary>
        /// Attaches the field to a set of field characters.
        /// </summary>
        /// <param name="start">The start.</param>
        /// <param name="separate">The separate.</param>
        /// <param name="end">The end.</param>
        private void AttachFieldCharacters(FieldCharacter start, FieldCharacter separate, FieldCharacter end)
        {
            Guard.ThrowExceptionIfNull(start, "start");
            Guard.ThrowExceptionIfNull(separate, "separate");
            Guard.ThrowExceptionIfNull(end, "end");

            Guard.ThrowExceptionIfFalse(start.FieldCharacterType == FieldCharacterType.Start, "Unexpected field character type");
            Guard.ThrowExceptionIfFalse(separate.FieldCharacterType == FieldCharacterType.Separator, "Unexpected field character type");
            Guard.ThrowExceptionIfFalse(end.FieldCharacterType == FieldCharacterType.End, "Unexpected field character type");

            Guard.ThrowExceptionIfFalse(start.FieldInfo == null, "Start field character already associated to another FieldInfo.");
            Guard.ThrowExceptionIfFalse(separate.FieldInfo == null, "Start field character already associated to another FieldInfo.");
            Guard.ThrowExceptionIfFalse(end.FieldInfo == null, "End field character already associated to another FieldInfo.");

            Guard.ThrowExceptionIfFalse(start.Document == separate.Document, "Start and Separator characters belong to different documents.");
            Guard.ThrowExceptionIfFalse(start.Document == end.Document, "Start and End characters belong to different documents.");

            this.Start = start;
            this.Start.FieldInfo = this;

            this.Separator = separate;
            this.Separator.FieldInfo = this;

            this.End = end;
            this.End.FieldInfo = this;
        }
    }
}