﻿using System.Text;
using Telerik.Windows.Documents.Flow.Model.Fields.FieldCode;

namespace Telerik.Windows.Documents.Flow.Model.Fields
{
    /// <summary>
    /// A hyperlink field.
    /// </summary>
    public class Hyperlink : Field
    {
        private static readonly string ToolTipSwitch = "o";
        private static readonly string TargetSwitch = "t";
        private static readonly string BookmarkSwitch = "l";

        /// <summary>
        /// Initializes a new instance of the <see cref="Hyperlink" /> class.
        /// </summary>
        /// <param name="document">The document.</param>
        public Hyperlink(RadFlowDocument document)
            : base(document)
        {
        }

        /// <summary>
        /// Gets the URI.
        /// </summary>
        /// <value>The URI.</value>
        public string Uri { get; internal set; }

        /// <summary>
        /// Gets a value indicating if the hyperlink points to a bookmark.
        /// </summary>
        /// <value>The is anchor.</value>
        public bool IsAnchor { get; internal set; }

        /// <summary>
        /// Gets the tool tip.
        /// </summary>
        /// <value>The tool tip.</value>
        public string ToolTip { get; internal set; }

        internal override bool PreserveResultOnUpdate
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Creates the hyperlink code.
        /// </summary>
        /// <returns></returns>
        internal string CreateHyperlinkCode()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(" HYPERLINK ");
            if (this.IsAnchor)
            {
                sb.Append("\\").Append(Hyperlink.BookmarkSwitch).Append(" ");
            }

            sb.Append("\"");
            sb.Append(Field.EncodeParameter(this.Uri));
            sb.Append("\"");

            if (!string.IsNullOrEmpty(this.ToolTip))
            {
                sb.Append(" \\").Append(Hyperlink.ToolTipSwitch).Append(" \"");
                sb.Append(Field.EncodeParameter(this.ToolTip));
                sb.Append("\"");
            }

            sb.Append(" ");
            return sb.ToString();
        }

        internal override void LoadParametersOverride(FieldParameters parameters)
        {
            FieldSwitch bmkSwitch = null;
            if (parameters.TryGetSwitch(Hyperlink.BookmarkSwitch, out bmkSwitch))
            {
                this.Uri = bmkSwitch.Argument != null ? bmkSwitch.Argument.Value : string.Empty;
                this.IsAnchor = true;
            }
            else
            {
                this.Uri = parameters.FirstArgument != null ? parameters.FirstArgument.Value : string.Empty;
                this.IsAnchor = false;
            }

            this.ToolTip = parameters.GetSwitchArgument(Hyperlink.ToolTipSwitch);
        }

        internal override FieldResult GetFieldResult(FieldUpdateContext context)
        {
            FieldResult fieldResult = null;

            // Update the content only if the field is empty.
            if (string.IsNullOrEmpty(context.FieldInfo.GetResult()))
            {
                fieldResult = new FieldResult(this.Uri);
            }

            return fieldResult;
        }

        /// <summary>
        /// Determines whether a switch can have an argument.
        /// </summary>
        /// <param name="switchKey">The switch key.</param>
        /// <returns>True if the switch can have argument.</returns>
        protected override bool IsSwitchWithArgumentOverride(string switchKey)
        {
            return switchKey == Hyperlink.BookmarkSwitch ||
                   switchKey == Hyperlink.TargetSwitch ||
                   switchKey == Hyperlink.ToolTipSwitch;
        }
    }
}