﻿using Telerik.Windows.Documents.Flow.Model.Fields.FieldCode;
using Telerik.Windows.Documents.Flow.Model.Fields.MailMerge;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Fields
{
    /// <summary>
    /// A merge field.
    /// </summary>
    public class MergeField : Field
    {
        internal static readonly string Type = "MERGEFIELD";

        private static readonly string PrefixSwitch = "b";
        private static readonly string SuffixSwitch = "f";
        private static readonly string MappedSwitch = "m";
        private static readonly string VerticalFormattingSwitch = "v";

        /// <summary>
        /// Initializes a new instance of the <see cref="MergeField" /> class.
        /// </summary>
        /// <param name="document">The document.</param>
        public MergeField(RadFlowDocument document) : base(document)
        {
        }

        /// <summary>
        /// Gets the prefix.
        /// </summary>
        /// <value>The prefix.</value>
        public string Prefix { get; internal set; }

        /// <summary>
        /// Gets the suffix.
        /// </summary>
        /// <value>The suffix.</value>
        public string Suffix { get; internal set; }

        internal bool IsMapped { get; set; }
        internal bool EnablesVerticalFormatting { get; set; }

        internal override void LoadParametersOverride(FieldParameters parameters)
        {
            this.Prefix = parameters.GetSwitchArgument(MergeField.PrefixSwitch);
            this.Suffix = parameters.GetSwitchArgument(MergeField.SuffixSwitch);

            this.IsMapped = parameters.IsSwitchDefined(MergeField.MappedSwitch);
            this.EnablesVerticalFormatting = parameters.IsSwitchDefined(MergeField.VerticalFormattingSwitch);
        }

        internal override FieldResult GetFieldResult(FieldUpdateContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            MergeFieldUpdateContext mergeFieldUpdateContext = context as MergeFieldUpdateContext;

            string result;
            if (mergeFieldUpdateContext == null || mergeFieldUpdateContext.CurrentRecord == null)
            {
                result = this.GetDefaultResult(context);
            }
            else
            {
                result = this.GetUpdatedResult(mergeFieldUpdateContext);
            }

            if (!string.IsNullOrEmpty(result))
            {
                result = this.Prefix + result + this.Suffix;
            }

            return new FieldResult(result ?? string.Empty);
        }

        /// <summary>
        /// Determines whether a switch can have an argument.
        /// </summary>
        /// <param name="switchKey">The switch key.</param>
        /// <returns>True if the switch can have argument.</returns>
        protected override bool IsSwitchWithArgumentOverride(string switchKey)
        {
            return
                switchKey == MergeField.PrefixSwitch ||
                switchKey == MergeField.SuffixSwitch ||
                switchKey == MergeField.MappedSwitch ||
                switchKey == MergeField.VerticalFormattingSwitch;
        }

        private string GetDefaultResult(FieldUpdateContext context)
        {
            if (context.Parameters.FirstArgument == null ||
                string.IsNullOrEmpty(context.Parameters.FirstArgument.Value))
            {
                return null;
            }

            string argument = context.Parameters.FirstArgument.Value.ToString();
            return "«" + argument + "»";
        }

        private string GetUpdatedResult(MergeFieldUpdateContext mergeFieldUpdateContext)
        {
            string propertyName = mergeFieldUpdateContext.Parameters.FirstArgument.Value;
            if (string.IsNullOrEmpty(propertyName))
            {
                return this.GetDefaultResult(mergeFieldUpdateContext);
            }

            // TODO: implement the mapped field. One property name to another name.
            object value = PropertyValueExtractor.GetValue(mergeFieldUpdateContext.CurrentRecord, propertyName);
            if (value == null)
            {
                return string.Empty;
            }

            return value.ToString();
        }
    }
}
