﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.Model.Cloning;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model
{
    /// <summary>
    /// Combines all headers or footers that can be available in a <see cref="Section"/>.
    /// </summary>
    public abstract class HeadersFootersBase<T>
        where T : HeaderFooterBase
    {
        private const string UnsupportedHeaderFooterTypeExceptionMessage = "Unsupported header/footer type.";

        private readonly RadFlowDocument document;
        private readonly Section section;

        private T @default;
        private T even;
        private T first;

        internal HeadersFootersBase(RadFlowDocument document, Section section)
        {
            this.document = document;
            this.section = section;
        }

        /// <summary>
        /// Gets the header or footer which should be applied to the first page.
        /// <para>
        /// Visualization of this header or footer depends additionally on <see cref="RadFlowDocument.HasDifferentEvenOddPageHeadersFooters"/> property.
        /// </para>
        /// </summary>
        /// <value>The first header or footer.</value>
        public T First
        {
            get
            {
                return this.first;
            }
        }

        /// <summary>
        /// Gets the header or footer which should be applied to the even pages.
        /// <para>
        /// Visualization of this header or footer depends additionally on <see cref="T:Section.HasDifferentFirstPageHeaderFooter"/> property.
        /// </para>
        /// </summary>
        /// <value>The even header or footer.</value>
        public T Even
        {
            get
            {
                return this.even;
            }
        }

        /// <summary>
        /// Gets the default header or footer which should be applied to the pages.
        /// </summary>
        /// <value>The default header or footer.</value>
        public T Default
        {
            get
            {
                return this.@default;
            }
        }

        /// <summary>
        /// Gets the document to which the current style belongs to.
        /// </summary>
        /// <value>The document.</value>
        public RadFlowDocument Document
        {
            get
            {
                return this.document;
            }
        }

        /// <summary>
        /// Gets the owner section.
        /// </summary>
        /// <value>The owner section.</value>
        public Section Section
        {
            get
            {
                return this.section;
            }
        }

        /// <summary>
        /// Adds default <see cref="Header"/> or <see cref="Footer"/>.
        /// </summary>
        /// <returns>The added <see cref="Header"/> or <see cref="Footer"/>.</returns>
        public T Add()
        {
            return this.Add(HeaderFooterType.Default);
        }

        /// <summary>
        /// Adds <see cref="Header"/> or <see cref="Footer"/> according to specified <see cref="HeaderFooterType"/>. 
        /// <para>
        /// Visualization of non-default headers or footers depends additionally on 
        /// <see cref="T:Section.HasDifferentFirstPageHeaderFooter"/> and <see cref="RadFlowDocument.HasDifferentEvenOddPageHeadersFooters"/> properties.
        /// </para>
        /// </summary>
        /// <param name="headerFooterType">Type of the header or footer.</param>
        /// <returns>The added <see cref="Header"/> or <see cref="Footer"/>.</returns>
        public T Add(HeaderFooterType headerFooterType)
        {
            switch (headerFooterType)
            {
                case HeaderFooterType.Default:
                    this.@default = this.CreateHeaderFooterInstanceInternal();
                    return this.Default;

                case HeaderFooterType.Even:
                    this.even = this.CreateHeaderFooterInstanceInternal();
                    return this.Even;

                case HeaderFooterType.First:
                    this.first = this.CreateHeaderFooterInstanceInternal();
                    return this.First;

                default:
                    throw new ArgumentException(UnsupportedHeaderFooterTypeExceptionMessage);
            }
        }

        /// <summary>
        /// Removes the <see cref="Header"/> or <see cref="Footer"/> with specified <see cref="HeaderFooterType"/>.
        /// </summary>
        /// <param name="headerFooterType">Type of the header footer.</param>
        public void Remove(HeaderFooterType headerFooterType)
        {
            switch (headerFooterType)
            {
                case HeaderFooterType.Default:
                    this.@default = null;
                    break;

                case HeaderFooterType.Even:
                    this.even = null;
                    break;

                case HeaderFooterType.First:
                    this.first = null;
                    break;

                default:
                    throw new ArgumentException(UnsupportedHeaderFooterTypeExceptionMessage);
            }
        }

        internal IEnumerable<T> GetInstances()
        {
            if (this.Default != null)
            {
                yield return this.Default;
            }
            if (this.First != null)
            {
                yield return this.First;
            }
            if (this.Even != null)
            {
                yield return this.Even;
            }
        }

        internal void CloneHeadersFootersFrom(HeadersFootersBase<T> headersFooters, CloneContext cloneContext)
        {
            Guard.ThrowExceptionIfNull(headersFooters, "headersFooters");
            Guard.ThrowExceptionIfNull(cloneContext, "cloneContext");

            if (headersFooters.First != null)
            {
                this.first = (T)headersFooters.First.CloneCore(cloneContext);
                if (this.first != null)
                {
                    headersFooters.First.OnAfterCloneCore(cloneContext, this.first);
                }
            }

            if (headersFooters.Even != null)
            {
                this.even = (T)headersFooters.Even.CloneCore(cloneContext);
                if (this.even != null)
                {
                    headersFooters.Even.OnAfterCloneCore(cloneContext, this.even);
                }
            }

            if (headersFooters.Default != null)
            {
                this.@default = (T)headersFooters.Default.CloneCore(cloneContext);
                if (this.@default != null)
                {
                    headersFooters.Default.OnAfterCloneCore(cloneContext, this.@default);
                }
            }
        }

        /// <summary>
        /// Creates an instance of <see cref="Header"/> or <see cref="Footer"/> class.
        /// </summary>
        /// <returns>The new <see cref="Header"/> or <see cref="Footer"/> instance.</returns>
        protected abstract T CreateHeaderFooterInstance();

        private T CreateHeaderFooterInstanceInternal()
        {
            T headerFooter = this.CreateHeaderFooterInstance();

            if (headerFooter == null)
            {
                throw new InvalidOperationException("Header/footer is null.");
            }

            if (headerFooter.Document != this.Document)
            {
                throw new InvalidOperationException("Header/footer has invalid associated document.");
            }

            if (headerFooter.Parent != this.Section)
            {
                throw new InvalidOperationException("Header/footer has invalid parent.");
            }

            return headerFooter;
        }
    }
}