using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Flow.Model.Fields;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model
{
    // TODO: Think of a more suitable name
    internal static class InlineRangeEditor
    {
        internal static IEnumerable<InlineBase> EnumerateInlinesInRange(InlineBase start, InlineBase end, bool respectParagraphs = false)
        {
            InlineRangeEditor.ValidateInlineRange(start, end);

            Paragraph currentParagraph = start.Paragraph;
            int curIndex = currentParagraph.Inlines.IndexOf(start) + 1;

            Paragraph endPar = end.Paragraph;
            int endIndex = endPar.Inlines.IndexOf(end);

            while (currentParagraph != null && currentParagraph != endPar)
            {
                while (curIndex < currentParagraph.Inlines.Count)
                {
                    yield return currentParagraph.Inlines[curIndex];
                    curIndex++;
                }
                curIndex = 0;
                int nextParIndex = currentParagraph.BlockContainer.Blocks.IndexOf(currentParagraph) + 1;
                currentParagraph = currentParagraph.BlockContainer.Blocks.Count > nextParIndex ? currentParagraph.BlockContainer.Blocks[nextParIndex] as Paragraph : null;

                if (respectParagraphs && currentParagraph != null)
                {
                    EndOfParagraphMarker br = new EndOfParagraphMarker(start.Document);

                    yield return br;
                }
            }

            while (curIndex < endIndex)
            {
                yield return currentParagraph.Inlines[curIndex];
                curIndex++;
            }
        }

        internal static IEnumerable<InlineBase> EnumerateFieldInlinesInRange(InlineBase start, InlineBase end, bool respectParagraphs = false)
        {
            FieldCharacter expectedFieldCharacter = null;
            foreach (InlineBase inline in InlineRangeEditor.EnumerateInlinesInRange(start, end, respectParagraphs))
            {
                if (expectedFieldCharacter != null)
                {
                    if (expectedFieldCharacter == inline)
                    {
                        expectedFieldCharacter = null;
                    }
                }
                else
                {
                    yield return inline;
                    FieldCharacter fieldChar = inline as FieldCharacter;
                    if (fieldChar != null && fieldChar.FieldCharacterType == FieldCharacterType.Start)
                    {
                        // We have reached start of code fragment - skip it.
                        expectedFieldCharacter = fieldChar.FieldInfo.Separator ?? fieldChar.FieldInfo.End;
                    }
                }
            }
        }

        internal static CharacterProperties GetCharacterFormattingFromFirstRunInRange(InlineBase start, InlineBase end)
        {
            foreach (InlineBase inline in EnumerateInlinesInRange(start, end))
            {
                Run run = inline as Run;
                if (run != null)
                {
                    return run.Properties;
                }
            }

            return null;
        }

        internal static string GetTextInRange(InlineBase start, InlineBase end, bool getContentFormFieldResultsOnly, bool respectParagraphs)
        {
            IEnumerable<InlineBase> inlines;
            if (getContentFormFieldResultsOnly)
            {
                inlines = EnumerateFieldInlinesInRange(start, end, respectParagraphs);
            }
            else
            {
                inlines = EnumerateInlinesInRange(start, end, respectParagraphs);
            }

            StringBuilder sb = new StringBuilder();
            foreach (InlineBase inline in inlines)
            {
                Run run = inline as Run;
                if (run != null)
                {
                    sb.Append(run.Text);
                    continue;
                }

                Break br = inline as Break;
                if (br != null && br.BreakType == BreakType.LineBreak)
                {
                    sb.Append(SpecialSymbols.VerticalTabSymbol);
                    continue;
                }

                EndOfParagraphMarker endOfParagraphMarker = inline as EndOfParagraphMarker;
                if (endOfParagraphMarker != null)
                {
                    sb.Append(Environment.NewLine);
                    continue;
                }
            }
            return sb.ToString();
        }

        internal static void InsertInlinesInRange(InlineBase start, IEnumerable<InlineBase> inlines)
        {
            Guard.ThrowExceptionIfNull(inlines, "inlines");

            Paragraph paragraph = start.Paragraph;
            int startIndex = paragraph.Inlines.IndexOf(start) + 1;

            paragraph.Inlines.InsertRange(startIndex, inlines);
        }

        internal static void DeleteContent(InlineBase start, InlineBase end)
        {
            InlineRangeEditor.ValidateInlineRange(start, end);
            if (start.Paragraph == end.Paragraph)
            {
                DeleteContentSameParagraph(start, end);
            }
            else
            {
                DeleteContentDifferentParagraphs(start, end);
            }
        }

        private static void DeleteContentSameParagraph(InlineBase start, InlineBase end)
        {
            Guard.ThrowExceptionIfFalse(start.Paragraph == end.Paragraph, "Start and end should belong to the same paragraph.");

            Paragraph paragraph = start.Paragraph;
            int startIndex = paragraph.Inlines.IndexOf(start) + 1;
            int endIndex = paragraph.Inlines.IndexOf(end);

            paragraph.Inlines.RemoveRange(startIndex, endIndex - startIndex);
        }

        private static void DeleteContentDifferentParagraphs(InlineBase start, InlineBase end)
        {
            Paragraph startPar = start.Paragraph;
            Paragraph endPar = end.Paragraph;

            // Delete blocks in the middle
            BlockContainerBase blockContainer = startPar.BlockContainer;
            int startParIndex = blockContainer.Blocks.IndexOf(startPar);
            int endParIndex = blockContainer.Blocks.IndexOf(endPar);
            blockContainer.Blocks.RemoveRange(startParIndex + 1, endParIndex - startParIndex - 1);

            // Delete inlines at the end of start paragraph.
            int startIndex = startPar.Inlines.IndexOf(start) + 1;
            startPar.Inlines.RemoveRange(startIndex, startPar.Inlines.Count - startIndex);

            // Delete paragraph at the beginning of end paragraph.
            int endIndex = endPar.Inlines.IndexOf(end);
            endPar.Inlines.RemoveRange(0, endIndex);

            // Insert the content of start paragraph in the end paragraph.
            List<InlineBase> inlines = startPar.Inlines.ToList();
            startPar.Inlines.RemoveRange(0, inlines.Count);
            endPar.Inlines.InsertRange(0, inlines);

            // Remove start paragraph.
            blockContainer.Blocks.RemoveAt(startParIndex);
        }

        private static void ValidateInlineRange(InlineBase start, InlineBase end)
        {
            Guard.ThrowExceptionIfNull(start, "start");
            Guard.ThrowExceptionIfNull(end, "end");
            Guard.ThrowExceptionIfFalse(start.Document == end.Document, "Start and end inlines should belong to the same document.");

            Paragraph startPar = start.Paragraph;
            Paragraph endPar = end.Paragraph;
            if (startPar == endPar)
            {
                if (!IsBefore(start, end))
                {
                    throw new InvalidOperationException("Start inline should be before end inline.");
                }
            }
            else
            {
                if (startPar.Parent != endPar.Parent)
                {
                    throw new InvalidOperationException("Start and end inlines should belong to paragraph in one block container.");
                }

                if (!IsBefore(startPar, endPar))
                {
                    throw new InvalidOperationException("Start inline should be before end inline.");
                }
            }
        }

        private static bool IsBefore(DocumentElementBase start, DocumentElementBase end)
        {
            Guard.ThrowExceptionIfFalse(start.Parent == end.Parent, "Start and End should have same parent.");

            foreach (DocumentElementBase el in start.Parent.Children)
            {
                if (el == start)
                {
                    return true;
                }
                if (el == end)
                {
                    return false;
                }
            }

            // Something is really wrong here.
            throw new InvalidOperationException("Child element not contained in parent.");
        }
    }
}