﻿using System;
using Telerik.Windows.Documents.Flow.Model.Styles;

namespace Telerik.Windows.Documents.Flow.Model.Lists
{
    /// <summary>
    /// Establishes functionality to share list properties between different paragraphs.
    /// </summary>
    public sealed class List
    {
        private readonly ListLevelCollection levels;
        private MultilevelType multilevelType;
        private int id;
        private string styleId;

        /// <summary>
        /// Initializes a new instance of the <see cref="List"/> class.
        /// <para>For predefined lists use <see cref="ListCollection.Add(ListTemplateType)" /> method.</para>
        /// </summary>
        public List()
        {
            this.levels = new ListLevelCollection(this);
            this.multilevelType = MultilevelType.HybridMultilevel;
            this.id = -1;
        }

        /// <summary>
        /// Gets the ID of the list.
        /// </summary>
        /// <value>The ID.</value>
        public int Id
        {
            get
            {
                return this.id;
            }
            internal set
            {
                this.id = value;
            }
        }

        /// <summary>
        /// Gets or sets the name of a numbering style associated with this list.
        /// </summary>
        /// <value>The name of the style.</value>
        public string StyleId
        {
            get
            {
                return this.styleId;
            }
            set
            {
                this.CheckIsValidStyle(value);

                this.styleId = value;

                if (this.Id > -1 && this.Document != null)
                {
                    Style style = this.Document.StyleRepository.GetStyle(this.styleId);
                    if (style != null)
                    {
                        // Guard.ThrowExceptionIfTrue(style.ParagraphProperties.ListId.HasLocalValue &&
                        // style.ParagraphProperties.ListId.LocalValue != this.Id,
                        // "Style is already refering another list.");
                        style.ParagraphProperties.ListId.LocalValue = this.Id;
                    }
                }
            }
        }

        /// <summary>
        /// Gets the levels.
        /// </summary>
        /// <value>The levels.</value>
        public ListLevelCollection Levels
        {
            get
            {
                return this.levels;
            }
        }

        /// <summary>
        /// Gets the document to which this list is associated.
        /// </summary>
        /// <value>The document.</value>
        public RadFlowDocument Document { get; internal set; }

        /// <summary>
        /// Gets or sets the type of the multilevel.
        /// </summary>
        /// <value>
        /// The type of the multilevel.
        /// </value>
        public MultilevelType MultilevelType
        {
            get
            {
                return this.multilevelType;
            }
            set
            {
                this.multilevelType = value;
            }
        }

        /// <summary>
        /// Gets or sets the name (alias) of the list.
        /// </summary>
        /// <value>The name.</value>
        internal string Name { get; set; }

        /// <summary>
        /// Creates deep copy of this list.
        /// </summary>
        /// <returns>The cloned list.</returns>
        public List Clone()
        {
            List list = new List();

            list.MultilevelType = this.MultilevelType;
            list.Id = this.Id;
            list.StyleId = this.StyleId;
            list.Name = this.Name;

            for (int i = 0; i < this.Levels.Count; i++)
            {
                list.Levels[i] = this.Levels[i].Clone(list);
            }

            return list;
        }

        private void CheckIsValidStyle(string value)
        {
            if (string.IsNullOrEmpty(value) || this.Document == null)
            {
                return;
            }

            Style style = this.Document.StyleRepository.GetStyle(value);
            if (style == null)
            {
                return;
            }

            if (style.StyleType != StyleType.Numbering)
            {
                throw new ArgumentException(string.Format("Style of type {0} can not be assigned to list.", style.StyleType));
            }
        }
    }
}
