﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Flow.Model.Cloning;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Lists
{
    /// <summary>
    /// Represents dynamic collection of lists.
    /// </summary>
    public sealed class ListCollection : IEnumerable<List>
    {
        private readonly List<List> listCollection = new List<List>();
        private readonly RadFlowDocument document;
        private readonly IdGenerator listStyleIdGenerator = new IdGenerator();

        internal ListCollection(RadFlowDocument parent)
        {
            this.document = parent;
        }

        /// <summary>
        /// Gets the number of lists contained in the collection.
        /// </summary>
        /// <returns>The number of lists contained in the collection.</returns>
        public int Count
        {
            get
            {
                return this.listCollection.Count;
            }
        }

        internal RadFlowDocument Document
        {
            get
            {
                return this.document;
            }
        }

        /// <summary>
        /// Gets the list.
        /// </summary>
        /// <param name="id">The id.</param>
        /// <returns></returns>
        public List GetList(int id)
        {
            return this.listCollection.Where(l => l.Id == id).FirstOrDefault();
        }

        /// <summary>
        /// Adds list according to the list template parameter.
        /// </summary>
        /// <param name="listTemplate">The list template.</param>
        /// <returns>The added list.</returns>
        public List Add(ListTemplateType listTemplate)
        {
            List list = ListFactory.GetList(listTemplate);
            this.Add(list);

            return list;
        }

        /// <summary>
        /// Adds an list to the collection.
        /// </summary>
        /// <param name="list">The list to add to the collection.</param>
        public void Add(List list)
        {
            if (list.Document != null)
            {
                throw new InvalidOperationException("The list is already added to the list collection of another document.");
            }

            list.Document = this.Document;
            list.Id = this.listStyleIdGenerator.GetNext();
            this.listCollection.Add(list);
        }

        /// <summary>
        /// Removes the specified list by list id.
        /// </summary>
        /// <param name="listId">The list id.</param>
        /// <returns>
        /// <c>true</c> if <paramref name="listId" /> was successfully removed from the collection; otherwise, <c>false</c>. This method also returns false if list with <paramref name="listId" /> is not found in the collection.
        /// </returns>
        public bool Remove(int listId)
        {
            List list = this.GetList(listId);

            return this.Remove(list);
        }

        /// <summary>
        /// Removes the first occurrence of a specific object from the collection.
        /// </summary>
        /// <param name="list">The list to remove from the collection.</param>
        /// <returns>
        /// <c>true</c> if <paramref name="list" /> was successfully removed from the collection; otherwise, <c>false</c>. This method also returns false if <paramref name="list" /> is not found in the collection.
        /// </returns>
        public bool Remove(List list)
        {
            bool result = this.listCollection.Remove(list);
            if (result)
            {
                foreach (Paragraph paragraph in list.Document.EnumerateChildrenOfType<Paragraph>())
                {
                    if (paragraph.ListId == list.Id)
                    {
                        paragraph.Properties.ListId.ClearValue();
                        paragraph.Properties.ListLevel.ClearValue();
                    }
                }

                foreach (Style style in list.Document.StyleRepository.Styles)
                {
                    if (style.ParagraphProperties.ListId.LocalValue == list.Id)
                    {
                        style.ParagraphProperties.ListId.ClearValue();
                        style.ParagraphProperties.ListLevel.ClearValue();
                    }
                }

                list.Document = default(RadFlowDocument);
            }

            return result;
        }

        /// <summary>
        /// Determines whether the collection contains a specific item.
        /// </summary>
        /// <param name="list">The item to locate in the collection.</param>
        /// <returns>
        /// <c>true</c> if <paramref name="list" /> is found in the collection; otherwise, <c>false</c>.
        /// </returns>
        public bool Contains(List list)
        {
            return this.listCollection.Contains(list);
        }

        /// <summary>
        /// Removes all lists from the collection.
        /// </summary>
        public void Clear()
        {
            foreach (List list in this.listCollection.ToList())
            {
                this.Remove(list);
            }
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<List> GetEnumerator()
        {
            return this.listCollection.GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.listCollection.GetEnumerator();
        }

        internal void AddClonedListsFrom(ListCollection fromListCollection)
        {
            Guard.ThrowExceptionIfNull(fromListCollection, "fromListCollection");

            foreach (List list in fromListCollection)
            {
                this.Add(list.Clone());
            }
        }

        internal void Merge(ListCollection withListCollection, CloneContext cloneContext)
        {
            Guard.ThrowExceptionIfNull(withListCollection, "withListCollection");
            Guard.ThrowExceptionIfNull(cloneContext, "cloneContext");

            ListCollectionMerger merger = new ListCollectionMerger(this, withListCollection);
            merger.Merge(cloneContext);

            // assigning new list ids to the related styles.
            foreach (KeyValuePair<int, int> pair in cloneContext.ReinitializedLists)
            {
                int oldListId = pair.Key;
                if (cloneContext.OldListsToStyles.ContainsKey(oldListId))
                {
                    foreach (string styleId in cloneContext.OldListsToStyles[oldListId])
                    {
                        Style style = this.Document.StyleRepository.GetStyle(styleId);
                        int newListId = cloneContext.ReinitializedLists[oldListId];
                        style.ParagraphProperties.ListId.LocalValue = newListId;
                    }
                }
            }
        }
    }
}
