﻿using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Lists
{
    /// <summary>
    /// Establishes functionality of list level.
    /// </summary>
    public sealed class ListLevel
    {
        private readonly CharacterProperties characterProperties;
        private readonly ParagraphProperties paragraphProperties;

        private List ownerList;
        private int startIndex = DocumentDefaultStyleSettings.StartIndex;
        private int restartAfterLevel = DocumentDefaultStyleSettings.RestartAfterLevel;
        private NumberingStyle numberingStyle = NumberingStyle.Bullet;
        private bool isLegal;
        private string numberTextFormat;
        private Alignment levelAlignment;
        private string styleId;

        internal ListLevel(List ownerList)
        {
            this.OwnerList = ownerList;
            this.characterProperties = new CharacterProperties(this, false);
            this.paragraphProperties = new ParagraphProperties(this, false);
        }

        /// <summary>
        /// Gets or sets the starting number of this level.
        /// </summary>
        /// <exception cref="T:System.ArgumentOutOfRangeException"><paramref name="value"/> is less than 0.</exception>
        /// <value>The start index.</value>
        public int StartIndex
        {
            get
            {
                return this.startIndex;
            }
            set
            {
                Guard.ThrowExceptionIfLessThan(0, value, "value");

                this.startIndex = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating the list level that must appear before this list level restarts to its start index. The value must be lower of an earlier level than this level. 
        /// The default value -1, means the start index will be restarted after the previous level occur.
        /// </summary>
        /// <exception cref="T:System.ArgumentOutOfRangeException"><paramref name="value"/> is less then -1 and greater then 8.</exception>
        /// <value>The restart after level.</value>
        public int RestartAfterLevel
        {
            get
            {
                return this.restartAfterLevel;
            }
            set
            {
                Guard.ThrowExceptionIfLessThan(-1, value, "value");
                Guard.ThrowExceptionIfGreaterThan(8, value, "value");

                this.restartAfterLevel = value;
            }
        }

        /// <summary>
        /// Gets or sets the number text format.
        /// </summary>
        /// <value>The number text format.</value>
        public string NumberTextFormat
        {
            get { return this.numberTextFormat; }
            set { this.numberTextFormat = value; }
        }

        /// <summary>
        /// Gets or sets the numbering style.
        /// </summary>
        /// <value>The numbering style.</value>
        public NumberingStyle NumberingStyle
        {
            get
            {
                return this.numberingStyle;
            }
            set
            {
                this.numberingStyle = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating if all inherited number formats should be displayed as decimal format.
        /// </summary>
        /// <value>The is legal.</value>
        public bool IsLegal
        {
            get { return this.isLegal; }
            set { this.isLegal = value; }
        }

        /// <summary>
        /// Gets or sets the name of a paragraph style associated with this list level.
        /// </summary>
        /// <value>The name of the style.</value>
        public string StyleId
        {
            get
            {
                return this.styleId;
            }
            set
            {
                this.styleId = value;

                if (this.Document != null)
                {
                    Style style = null;
                    if (BuiltInStyles.IsBuiltInStyle(this.styleId))
                    {
                        style = this.Document.StyleRepository.AddBuiltInStyle(this.styleId);
                    }
                    else
                    {
                        style = this.Document.StyleRepository.GetStyle(this.styleId);
                    }

                    Guard.ThrowExceptionIfTrue(style.StyleType != StyleType.Paragraph, "Only paragraph styles can be related to list levels.");

                    int listLevelId = this.GetListLevelId();
                    if (style != null && style.ParagraphProperties != null)
                    {
                        if (!style.ParagraphProperties.ListId.HasLocalValue &&
                            !style.ParagraphProperties.ListLevel.HasLocalValue &&
                            this.OwnerList.Id != Paragraph.ListIdPropertyDefinition.DefaultValue &&
                            listLevelId != Paragraph.ListLevelPropertyDefinition.DefaultValue)
                        {
                            style.ParagraphProperties.ListLevel.LocalValue = listLevelId;
                            style.ParagraphProperties.ListId.LocalValue = this.OwnerList.Id;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Gets or sets the alignment.
        /// </summary>
        /// <value>The alignment.</value>
        public Alignment Alignment
        {
            get { return this.levelAlignment; }
            set { this.levelAlignment = value; }
        }

        /// <summary>
        /// Gets the character properties.
        /// </summary>
        /// <value>The character properties.</value>
        public CharacterProperties CharacterProperties
        {
            get { return this.characterProperties; }
        }

        /// <summary>
        /// Gets the paragraph properties.
        /// </summary>
        /// <value>The paragraph properties.</value>
        public ParagraphProperties ParagraphProperties
        {
            get { return this.paragraphProperties; }
        }

        /// <summary>
        /// Gets the document to which this list level is associated.
        /// </summary>
        /// <value>The document.</value>
        public RadFlowDocument Document
        {
            get
            {
                if (this.OwnerList == null)
                {
                    return null;
                }

                return this.OwnerList.Document;
            }
        }

        internal List OwnerList
        {
            get
            {
                return this.ownerList;
            }
            set
            {
                this.ownerList = value;
            }
        }

        /// <summary>
        /// Creates deep copy of this list level, associated to <paramref name="ownerList"/>.
        /// </summary>
        /// <param name="ownerList">The owner list.</param>
        public ListLevel Clone(List ownerList)
        {
            Guard.ThrowExceptionIfNull(ownerList, "ownerList");

            ListLevel listLevel = new ListLevel(ownerList);

            listLevel.StartIndex = this.StartIndex;
            listLevel.RestartAfterLevel = this.RestartAfterLevel;
            listLevel.NumberTextFormat = this.NumberTextFormat;
            listLevel.NumberingStyle = this.NumberingStyle;
            listLevel.IsLegal = this.IsLegal;
            listLevel.StyleId = this.StyleId;
            listLevel.Alignment = this.Alignment;
            listLevel.CharacterProperties.CopyPropertiesFrom(this.CharacterProperties);
            listLevel.ParagraphProperties.CopyPropertiesFrom(this.ParagraphProperties);

            return listLevel;
        }

        private int GetListLevelId()
        {
            int listLevelId = -1;
            if (this.OwnerList == null)
            {
                return listLevelId;
            }

            for (int i = 0; i < this.OwnerList.Levels.Count; i++)
            {
                if (this.OwnerList.Levels[i] == this)
                {
                    listLevelId = i;
                    break;
                }
            }
            return listLevelId;
        }
    }
}
