﻿using System;
using System.Collections;
using System.Collections.Generic;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Lists
{
    /// <summary>
    /// Represents collection of list levels.
    /// </summary>
    public sealed class ListLevelCollection : IEnumerable<ListLevel>
    {
        private const int LevelsCount = 9;

        private readonly List<ListLevel> listLevelCollection;
        private readonly List ownerList;

        internal ListLevelCollection(List ownerList)
        {
            this.ownerList = ownerList;
            this.listLevelCollection = new List<ListLevel>(LevelsCount);
            this.InitializeListLevels(LevelsCount);
        }

        /// <summary>
        /// Gets the count.
        /// </summary>
        /// <value>The count.</value>
        public int Count
        {
            get
            {
                return this.listLevelCollection.Count;
            }
        }

        internal List OwnerList
        {
            get
            {
                return this.ownerList;
            }
        }

        /// <summary>
        /// Gets or sets the list at the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <exception cref="T:System.ArgumentOutOfRangeException"><paramref name="index"/> is less then 0 and greater then 8.</exception>
        /// <returns></returns>
        public ListLevel this[int index]
        {
            get
            {
                Guard.ThrowExceptionIfLessThan(0, index, "index");
                Guard.ThrowExceptionIfGreaterThan(8, index, "index");

                return this.listLevelCollection[index];
            }
            set
            {
                Guard.ThrowExceptionIfLessThan(0, index, "index");
                Guard.ThrowExceptionIfGreaterThan(8, index, "index");

                if (value.OwnerList != null && value.OwnerList != this.OwnerList)
                {
                    throw new InvalidOperationException("List level is already added to another list.");
                }

                ListLevel oldListLevel = this.listLevelCollection[index];

                this.listLevelCollection[index] = value;
                this.listLevelCollection[index].OwnerList = this.OwnerList;

                oldListLevel.OwnerList = null;
            }
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<ListLevel> GetEnumerator()
        {
            return this.listLevelCollection.GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.listLevelCollection.GetEnumerator();
        }

        private void InitializeListLevels(int collectionCount)
        {
            for (int i = 0; i < collectionCount; i++)
            {
                this.listLevelCollection.Add(new ListLevel(this.OwnerList));
            }
        }
    }
}
