﻿using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model
{
    /// <summary>
    /// Represents page numbering settings for all page numbers in the content of the current section.
    /// </summary>
    public class PageNumberingSettings
    {
        private readonly Section section;

        internal PageNumberingSettings(Section section)
        {
            this.section = section;
        }

        /// <summary>
        /// Gets or sets the chapter separator character that shall appear between the chapter heading style index and the page number.
        /// </summary>
        /// <value>The chapter separator character.</value>
        public ChapterSeparatorType? ChapterSeparatorCharacter
        {
            get
            {
                return this.section.Properties.ChapterSeparatorCharacter.GetActualValue();
            }
            set
            {
                this.section.Properties.ChapterSeparatorCharacter.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the index of the chapter heading style.
        /// </summary>
        /// <value>The index of the chapter heading style.</value>
        public int? ChapterHeadingStyleIndex
        {
            get
            {
                return this.section.Properties.ChapterHeadingStyleIndex.GetActualValue();
            }
            set
            {
                if (value.HasValue)
                {
                    Guard.ThrowExceptionIfOutOfRange(1, 9, value.Value, "value");
                }
                this.section.Properties.ChapterHeadingStyleIndex.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the page number format for the page numbering in the current section.
        /// </summary>
        /// <value>The page number format.</value>
        public NumberingStyle? PageNumberFormat
        {
            get
            {
                return this.section.Properties.PageNumberFormat.GetActualValue();
            }
            set
            {
                this.section.Properties.PageNumberFormat.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the starting page number.
        /// <para>This number appears on the first page of the section.</para>
        /// </summary>
        /// <value>The starting page number.</value>
        public int? StartingPageNumber
        {
            get
            {
                return this.section.Properties.StartingPageNumber.GetActualValue();
            }
            set
            {
                if (value.HasValue)
                {
                    Guard.ThrowExceptionIfLessThan(0, value.Value, "value");
                }
                this.section.Properties.StartingPageNumber.LocalValue = value;
            }
        }
    }
}