﻿using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Flow.Model.Cloning;
using Telerik.Windows.Documents.Flow.Model.Collections;
using Telerik.Windows.Documents.Flow.Model.Fields;
using Telerik.Windows.Documents.Flow.Model.Protection;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model
{
    /// <summary>
    /// A block-level flow content element used to group content into a paragraph.
    /// </summary>
    [DebuggerDisplay("\\{Paragraph\\}: {GetChildrenText()}")]
    public sealed partial class Paragraph : BlockBase, IElementWithStyle, IElementWithProperties
    {
        private readonly InlineCollection inlines;
        private readonly ParagraphProperties properties;

        /// <summary>
        /// Initializes a new instance of the <see cref="Paragraph"/> class.
        /// </summary>
        /// <param name="document">The document in which the element will be added.</param>
        public Paragraph(RadFlowDocument document)
            : base(document)
        {
            this.inlines = new InlineCollection(this);
            this.properties = new ParagraphProperties(this);
            this.Spacing = new ParagraphSpacing(this.Properties);
            this.Shading = new Shading(this.Properties);
            this.Indentation = new ParagraphIndentation(this.Properties);
        }

        /// <summary>
        /// Gets the inlines in this paragraphs.
        /// </summary>
        /// <value>The inlines.</value>
        public InlineCollection Inlines
        {
            get
            {
                return this.inlines;
            }
        }

        /// <summary>
        /// Gets or sets the style id.
        /// </summary>
        /// <value>The style id.</value>
        public string StyleId
        {
            get
            {
                return this.Properties.StyleId;
            }
            set
            {
                this.Properties.StyleId = value;
            }
        }

        /// <summary>
        /// Gets the properties.
        /// </summary>
        /// <value>The properties.</value>
        DocumentElementPropertiesBase IElementWithProperties.Properties
        {
            get
            {
                return this.properties;
            }
        }

        /// <summary>
        /// Gets the styling properties of this element.
        /// </summary>
        /// <value>The properties.</value>
        public ParagraphProperties Properties
        {
            get
            {
                return this.properties;
            }
        }

        /// <summary>
        /// Gets or sets the flow direction.
        /// </summary>
        /// <returns>The default value is <c>LeftToRight</c></returns>
        public FlowDirection FlowDirection
        {
            get
            {
                return this.Properties.FlowDirection.GetActualValue().Value;
            }
            set
            {
                this.Properties.FlowDirection.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets the paragraph text alignment.
        /// </summary>
        /// <returns>The default value is <c>Left</c></returns>
        public Alignment TextAlignment
        {
            get
            {
                return this.Properties.TextAlignment.GetActualValue().Value;
            }
            set
            {
                this.Properties.TextAlignment.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets the paragraph spacing.
        /// </summary>
        /// <value>The paragraph spacing.</value>
        public ParagraphSpacing Spacing { get; private set; }

        /// <summary>
        /// Gets or sets a value indicating if this paragraph should be rendered on one page when the document is shown in page view mode.
        /// </summary>
        /// <returns>The default value is <c>false</c></returns>
        public bool KeepOnOnePage
        {
            get
            {
                return this.Properties.KeepOnOnePage.GetActualValue().Value;
            }
            set
            {
                this.Properties.KeepOnOnePage.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating if this paragraph should be rendered at least partly on the one page with 
        /// the following paragraph when this is possible and when the document is shown in page view mode.
        /// </summary>
        /// <returns>The default value is <c>false</c></returns>
        public bool KeepWithNextParagraph
        {
            get
            {
                return this.Properties.KeepWithNextParagraph.GetActualValue().Value;
            }
            set
            {
                this.Properties.KeepWithNextParagraph.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating the outline level. The outline level defines the level of this paragraph in TOC field.
        /// </summary>
        /// <returns>The default value is <c>Level9</c> - no level.</returns>
        public OutlineLevel OutlineLevel
        {
            get
            {
                return this.Properties.OutlineLevel.GetActualValue().Value;
            }
            set
            {
                this.Properties.OutlineLevel.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether East-Asian line breaking rules are applied to this paragraph.
        /// </summary>
        /// <returns>The default value is <c>true</c>.</returns>
        public bool ApplyEastAsianLineBreakingRules
        {
            get
            {
                return this.Properties.ApplyEastAsianLineBreakingRules.GetActualValue().Value;
            }
            set
            {
                this.Properties.ApplyEastAsianLineBreakingRules.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating if this paragraph should be rendered on new page when the document is shown in page view mode.
        /// </summary>
        /// <returns>The default value is <c>false</c></returns>
        public bool PageBreakBefore
        {
            get
            {
                return this.Properties.PageBreakBefore.GetActualValue().Value;
            }
            set
            {
                this.Properties.PageBreakBefore.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the borders of this paragraph.
        /// </summary>
        /// <value>The borders.</value>
        public ParagraphBorders Borders
        {
            get
            {
                return this.Properties.Borders.GetActualValue();
            }
            set
            {
                this.Properties.Borders.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether spacing before/after are ignored 
        /// when preceding/following paragraph has same paragraph style.
        /// </summary>
        /// <returns>The default value is <c>false</c>.</returns>
        public bool ContextualSpacing
        {
            get
            {
                return this.Properties.ContextualSpacing.GetActualValue().Value;
            }
            set
            {
                this.Properties.ContextualSpacing.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether left and right indents should be swapped on odd pages.
        /// </summary>
        /// <returns>The default value is <c>false</c>.</returns>
        public bool MirrorIndents
        {
            get
            {
                return this.Properties.MirrorIndents.GetActualValue().Value;
            }
            set
            {
                this.Properties.MirrorIndents.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the tab stops of the paragraph.
        /// </summary>
        /// <value>The tab stops.</value>
        public TabStopCollection TabStops
        {
            get
            {
                return this.Properties.TabStops.GetActualValue();
            }
            set
            {
                this.Properties.TabStops.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets the shading which shall be applied to the extents of the paragraph.
        /// </summary>
        /// <value>The shading.</value>
        public Shading Shading { get; private set; }

        /// <summary>
        /// Gets the indentation of the paragraph.
        /// </summary>
        /// <value>The indentation.</value>
        public ParagraphIndentation Indentation { get; private set; }

        /// <summary>
        /// Gets or sets a value indicating whether the last punctuation character on a line can overflow in paragraph margin/indent.
        /// </summary>
        /// <returns>The default is <c>true</c>.</returns>
        public bool AllowOverflowPunctuation
        {
            get
            {
                return this.Properties.AllowOverflowPunctuation.GetActualValue().Value;
            }
            set
            {
                this.Properties.AllowOverflowPunctuation.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the list id.
        /// </summary>
        /// <value>The list id.</value>
        public int ListId
        {
            get
            {
                return this.Properties.ListId.GetActualValue().Value;
            }
            set
            {
                this.Properties.ListId.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets a value indicating that the paragraph is referencing list style.
        /// </summary>
        /// <returns>The default value is <c>-1</c> - list style is not referenced.</returns>
        public int ListLevel
        {
            get
            {
                return this.Properties.ListLevel.GetActualValue().Value;
            }
            set
            {
                this.Properties.ListLevel.LocalValue = value;
            }
        }

        internal override IEnumerable<DocumentElementBase> Children
        {
            get
            {
                return this.Inlines;
            }
        }

        /// <summary>
        /// Gets the type of the element.
        /// </summary>
        /// <value>The type.</value>
        internal override DocumentElementType Type
        {
            get
            {
                return DocumentElementType.Paragraph;
            }
        }

        /// <summary>
        /// Creates deep copy of this document element, associated to the current document.
        /// </summary>
        /// <returns>The cloned element.</returns>
        public Paragraph Clone()
        {
            return this.CloneInternal(null);
        }

        /// <summary>
        /// Creates deep copy of this document element, associated to <paramref name="document"/>.
        /// </summary>
        /// <param name="document">The document to which the cloned element should be associated.</param>
        /// <returns>The cloned element.</returns>
        public Paragraph Clone(RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");
            return this.CloneInternal(document);
        }

        internal override DocumentElementBase CloneCore(CloneContext cloneContext)
        {
            Guard.ThrowExceptionIfNull(cloneContext, "cloneContext");

            Paragraph paragraph = this.CloneWithoutChildren(cloneContext);

            paragraph.Inlines.AddClonedChildrenFrom(this.Inlines, cloneContext);

            if (cloneContext.CurrentSection == null)
            {
                paragraph.RemoveUnpairedAnnotationMarkers(cloneContext);
            }

            return paragraph;
        }

        internal Paragraph CloneWithoutChildren(CloneContext cloneContext)
        {
            Guard.ThrowExceptionIfNull(cloneContext, "cloneContext");

            Paragraph paragraph = new Paragraph(cloneContext.Document);
            paragraph.Properties.CopyPropertiesFrom(this.Properties);

            if (cloneContext.RenamedStyles.ContainsKey(paragraph.StyleId))
            {
                paragraph.StyleId = cloneContext.RenamedStyles[paragraph.StyleId];
            }

            if (cloneContext.ReinitializedLists.ContainsKey(paragraph.ListId))
            {
                paragraph.ListId = cloneContext.ReinitializedLists[paragraph.ListId];
            }

            return paragraph;
        }

        private Paragraph CloneInternal(RadFlowDocument document)
        {
            CloneContext cloneContext = new CloneContext(document ?? this.Document);
            cloneContext.FieldContext.DisableStrictMode();

            Paragraph clonedParagraph = (Paragraph)this.CloneCore(cloneContext);

            return clonedParagraph;
        }

        private void RemoveUnpairedAnnotationMarkers(CloneContext cloneContext)
        {
            this.RemoveUnpairedFieldCharacters(cloneContext);
            this.RemoveUnpairedBookmarkRanges(cloneContext);
            this.RemoveUnpairedCommentRanges(cloneContext);
            this.RemoveUnpairedPermissionRanges(cloneContext);
        }

        private void RemoveUnpairedCommentRanges(CloneContext cloneContext)
        {
            foreach (CommentRangeStart start in cloneContext.CommentContext.GetHangingAnnotationStarts())
            {
                this.Inlines.Remove(start);
            }

            foreach (CommentRangeEnd end in cloneContext.CommentContext.GetHangingAnnotationEnds())
            {
                this.Inlines.Remove(end);
            }
        }

        private void RemoveUnpairedBookmarkRanges(CloneContext cloneContext)
        {
            foreach (BookmarkRangeStart start in cloneContext.BookmarkContext.GetHangingAnnotationStarts())
            {
                this.Inlines.Remove(start);
            }

            foreach (BookmarkRangeEnd end in cloneContext.BookmarkContext.GetHangingAnnotationEnds())
            {
                this.Inlines.Remove(end);
            }
        }

        private void RemoveUnpairedFieldCharacters(CloneContext cloneContext)
        {
            foreach (FieldCharacter fieldCharacter in cloneContext.FieldContext.GetUnpairedFieldCharacters())
            {
                this.Inlines.Remove(fieldCharacter);
            }
        }

        private void RemoveUnpairedPermissionRanges(CloneContext cloneContext)
        {
            foreach (PermissionRangeStart start in cloneContext.PermissionContext.GetHangingAnnotationStarts())
            {
                this.Inlines.Remove(start);
            }

            foreach (PermissionRangeEnd end in cloneContext.PermissionContext.GetHangingAnnotationEnds())
            {
                this.Inlines.Remove(end);
            }
        }

        private string GetChildrenText()
        {
            StringBuilder stringBuilder = new StringBuilder();

            foreach (Run run in this.Inlines.Where(i => i is Run).Select(i => i as Run))
            {
                stringBuilder.Append(run.Text);
            }

            return stringBuilder.ToString();
        }
    }
}
