﻿using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Protection
{
    /// <summary>
    /// Represents permission range for document protection.
    /// </summary>
    public class PermissionRange
    {
        private readonly RadFlowDocument document;
        private readonly PermissionRangeStart permissionRangeStart;
        private readonly PermissionRangeEnd permissionRangeEnd;
        private readonly PermissionRangeCredentials permissionCredentials;
        private readonly int? fromColumn;
        private readonly int? toColumn;

        /// <summary>
        /// Initializes a new instance of the <see cref="PermissionRange" /> class.
        /// </summary>
        /// <param name="document">The document.</param>
        /// <param name="permissionRangeCredentials">The permission range credentials.</param>
        public PermissionRange(RadFlowDocument document, PermissionRangeCredentials permissionRangeCredentials)
            : this(document, permissionRangeCredentials, null, null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="PermissionRange" /> class.
        /// </summary>
        /// <param name="document">The document.</param>
        /// <param name="permissionRangeCredentials">The permission range credentials.</param>
        /// <param name="fromColumn">From column.</param>
        /// <param name="toColumn">To column.</param>
        public PermissionRange(RadFlowDocument document, PermissionRangeCredentials permissionRangeCredentials, int? fromColumn, int? toColumn)
        {
            Guard.ThrowExceptionIfNull(document, "document");
            Guard.ThrowExceptionIfNull(permissionRangeCredentials, "permissionRangeCredentials");

            this.document = document;
            this.permissionCredentials = permissionRangeCredentials;
            this.fromColumn = fromColumn;
            this.toColumn = toColumn;

            this.permissionRangeStart = new PermissionRangeStart(this.Document, this);
            this.permissionRangeEnd = new PermissionRangeEnd(this.Document, this);
        }

        /// <summary>
        /// Gets the associated document.
        /// </summary>
        /// <value>The document.</value>
        public RadFlowDocument Document
        {
            get
            {
                return this.document;
            }
        }

        /// <summary>
        /// Gets the permission range credentials.
        /// </summary>
        /// <value>The permission range credentials.</value>
        public PermissionRangeCredentials Credentials
        {
            get
            {
                return this.permissionCredentials;
            }
        }

        /// <summary>
        /// Gets the permission range start.
        /// </summary>
        /// <value>The permission range start.</value>
        public PermissionRangeStart Start
        {
            get
            {
                return this.permissionRangeStart;
            }
        }

        /// <summary>
        /// Gets the permission range end.
        /// </summary>
        /// <value>The permission range end.</value>
        public PermissionRangeEnd End
        {
            get
            {
                return this.permissionRangeEnd;
            }
        }

        /// <summary>
        /// Gets the index of the first column in the current row which is part of the permission.
        /// </summary>
        /// <value>From column.</value>
        public int? FromColumn
        {
            get
            {
                return this.fromColumn;
            }
        }

        /// <summary>
        /// Gets the index of the last column in the current row which is part of the permission.
        /// </summary>
        /// <value>To column.</value>
        public int? ToColumn
        {
            get
            {
                return this.toColumn;
            }
        }
    }
}
