﻿using System.Collections;
using System.Collections.Generic;
using System.Linq;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Theming;
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Common.Model.Data;
using Telerik.Windows.Documents.Flow.Model.Collections;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Utilities;
using Telerik.Windows.Documents.Flow.Model.Protection;
using Telerik.Windows.Documents.Flow.Model.Cloning;
using Telerik.Windows.Documents.Flow.Model.Fields;

namespace Telerik.Windows.Documents.Flow.Model
{
    /// <summary>
    /// Hosts flow document content.
    /// </summary>
    public sealed partial class RadFlowDocument : DocumentElementBase, IElementWithProperties
    {
        private readonly StyleRepository styleRepository;
        private readonly SectionCollection sections;
        private readonly ResourceManager resources;
        private readonly DocumentProperties properties;
        private readonly ListCollection lists;
        private readonly CommentCollection comments;
        private readonly DocumentVariableCollection documentVariables;

        private DocumentTheme theme = PredefinedThemeSchemes.DefaultTheme;

        /// <summary>
        /// Initializes a new instance of the <see cref="RadFlowDocument" /> class.
        /// </summary>
        public RadFlowDocument()
            : this(true)
        {
        }

        internal RadFlowDocument(bool setInitialDocumentDefaultValues)
        {
            this.DefaultStyle = new DocumentDefaultStyle(this);
            this.sections = new SectionCollection(this);
            this.styleRepository = new StyleRepository(this);
            this.resources = new ResourceManager();
            this.properties = new DocumentProperties(this);
            this.lists = new ListCollection(this);
            this.comments = new CommentCollection(this);
            this.documentVariables = new DocumentVariableCollection();
            this.ProtectionSettings = new ProtectionSettings();

            if (setInitialDocumentDefaultValues)
            {
                this.EnsureInitialDocumentDefaultValues();
            }

            this.RegisterDefaultStyles();
        }

        /// <summary>
        /// Gets the default styling of the document.
        /// </summary>
        /// <value>The default style.</value>
        public DocumentDefaultStyle DefaultStyle { get; private set; }

        /// <summary>
        /// Gets the properties.
        /// </summary>
        /// <value>
        /// The properties.
        /// </value>
        DocumentElementPropertiesBase IElementWithProperties.Properties
        {
            get
            {
                return this.properties;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether pages in this document will have different headers and footers for even and odd pages.
        /// </summary>
        /// <value>
        /// <c>true</c> if pages in this document will have different headers and footers for even and odd pages; otherwise, <c>false</c>.
        /// </value>
        public bool HasDifferentEvenOddPageHeadersFooters
        {
            get
            {
                return this.Properties.HasDifferentEvenOddPageHeadersFooters.GetActualValue().Value;
            }
            set
            {
                this.Properties.HasDifferentEvenOddPageHeadersFooters.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating the view type of the document.
        /// </summary>
        /// <value>
        /// The type of the view.
        /// </value>
        public DocumentViewType ViewType
        {
            get
            {
                return this.Properties.ViewType.GetActualValue().Value;
            }
            set
            {
                this.Properties.ViewType.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the default width of the tab stop.
        /// </summary>
        /// <value>The default width of the tab stop.</value>
        public double DefaultTabStopWidth
        {
            get
            {
                return this.Properties.DefaultTabStopWidth.GetActualValue().Value;
            }
            set
            {
                this.Properties.DefaultTabStopWidth.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets the document to which this document element is associated.
        /// </summary>
        /// <value>
        /// The document.
        /// </value>
        public override RadFlowDocument Document
        {
            get
            {
                return this;
            }
        }

        /// <summary>
        /// Gets the styling properties of this document.
        /// </summary>
        /// <value>
        /// The properties.
        /// </value>
        public DocumentProperties Properties
        {
            get
            {
                return this.properties;
            }
        }

        /// <summary>
        /// Gets the sections of the flow document.
        /// </summary>
        /// <value>The sections.</value>
        public SectionCollection Sections
        {
            get
            {
                return this.sections;
            }
        }

        /// <summary>
        /// Gets the style repository.
        /// </summary>
        /// <value>The style repository.</value>
        public StyleRepository StyleRepository
        {
            get
            {
                return this.styleRepository;
            }
        }

        /// <summary>
        /// Gets the lists associated with this document.
        /// </summary>
        /// <value>The lists.</value>
        public ListCollection Lists
        {
            get
            {
                return this.lists;
            }
        }

        /// <summary>
        /// Gets or sets the theme.
        /// </summary>
        /// <value>The theme.</value>
        public DocumentTheme Theme
        {
            get
            {
                return this.theme;
            }
            set
            {
                this.theme = value;
            }
        }

        /// <summary>
        /// Gets the comments of the document.
        /// </summary>
        /// <value>The comments.</value>
        public CommentCollection Comments
        {
            get
            {
                return this.comments;
            }
        }

        /// <summary>
        /// Gets the the document variables.
        /// </summary>
        /// <value>The document variables.</value>
        public DocumentVariableCollection DocumentVariables
        {
            get
            {
                return this.documentVariables;
            }
        }

        /// <summary>
        /// Gets or sets the protection settings.
        /// </summary>
        /// <value>The protection settings.</value>
        public ProtectionSettings ProtectionSettings { get; private set; }

        internal override IEnumerable<DocumentElementBase> Children
        {
            get
            {
                return this.Sections.Concat<DocumentElementBase>(this.Comments);
            }
        }

        internal override IEnumerable<DocumentElementBase> ContentChildren
        {
            get
            {
                return this.Sections;
            }
        }

        internal override DocumentElementType Type
        {
            get
            {
                return DocumentElementType.Document;
            }
        }

        internal ResourceManager Resources
        {
            get
            {
                return this.resources;
            }
        }

        /// <summary>
        /// Creates deep copy of this document.
        /// </summary>
        /// <returns>The cloned document.</returns>
        public RadFlowDocument Clone()
        {
            return (RadFlowDocument)this.CloneCore(new CloneContext(new RadFlowDocument(false)));
        }

        /// <summary>
        /// Merges this document with the specified source document.
        /// </summary>
        /// <param name="sourceDocument">The source document.</param>
        public void Merge(RadFlowDocument sourceDocument)
        {
            this.Merge(sourceDocument, null);
        }

        /// <summary>
        /// Merges this document with the specified source document.
        /// </summary>
        /// <param name="sourceDocument">The source document.</param>
        /// <param name="mergeOptions">The merge options.</param>
        public void Merge(RadFlowDocument sourceDocument, MergeOptions mergeOptions)
        {
            Guard.ThrowExceptionIfNull(sourceDocument, "sourceDocument");

            CloneContext cloneContext = new CloneContext(this)
            {
                MergeOptions = mergeOptions ?? new MergeOptions()
            };

            this.MergeWithoutChildren(sourceDocument, cloneContext);
            this.Sections.AddClonedChildrenFrom(sourceDocument.Sections, cloneContext);
        }

        /// <summary>
        /// Updates all supported fields in the document.
        /// </summary>
        public void UpdateFields()
        {
            IEnumerable<FieldCharacter> allFields = this.EnumerateChildrenOfType<FieldCharacter>();
            List<FieldCharacter> fieldsToUpdate = FieldUpdateScheduler.GetOrderedListOfFieldToUpdate(allFields);

            foreach (FieldCharacter fieldStart in fieldsToUpdate)
            {
                FieldUpdateContext context = new FieldUpdateContext(fieldStart.FieldInfo);
                fieldStart.FieldInfo.UpdateFieldCore(context);
            }
        }

        /// <summary>
        /// Performs mail merge on the document with the specified <paramref name="collection"/> of records.
        /// </summary>
        /// <param name="collection">The collection of records.</param>
        /// <returns>The merged document.</returns>
        public RadFlowDocument MailMerge(IEnumerable collection)
        {
            return MailMergeProcessor.Execute(this, collection);
        }

        internal void MergeWithoutChildren(RadFlowDocument sourceDocument, CloneContext cloneContext)
        {
            this.StyleRepository.Merge(sourceDocument.StyleRepository, cloneContext);

            // lists should be merged after styles
            this.Lists.Merge(sourceDocument.Lists, cloneContext);
        }

        // TODO: This method should be optimized.
        internal Bookmark GetBookmarkByName(string name)
        {
            BookmarkRangeStart bookmarkRangeStart = this.EnumerateChildrenOfType<BookmarkRangeStart>()
                .FirstOrDefault(b => b.Bookmark.Name == name);

            return bookmarkRangeStart != null ? bookmarkRangeStart.Bookmark : null;
        }

        internal override DocumentElementBase CloneCore(CloneContext cloneContext)
        {
            Guard.ThrowExceptionIfNull(cloneContext, "cloneContext");

            RadFlowDocument document = cloneContext.Document;

            document.DefaultStyle.CharacterProperties.CopyPropertiesFrom(this.DefaultStyle.CharacterProperties);
            document.DefaultStyle.ParagraphProperties.CopyPropertiesFrom(this.DefaultStyle.ParagraphProperties);
            document.Properties.CopyPropertiesFrom(this.Properties);
            document.Sections.AddClonedChildrenFrom(this.Sections, cloneContext);
            document.StyleRepository.AddClonedStylesFrom(this.StyleRepository);
            document.Lists.AddClonedListsFrom(this.Lists);
            document.DocumentVariables.AddClonedDocumentVariablesFrom(this.DocumentVariables);
            document.ProtectionSettings = this.ProtectionSettings.Clone();
            document.Theme = this.Theme.Clone();

            return document;
        }

        // setting the default propety values with which every MS Word document is starting.
        private void EnsureInitialDocumentDefaultValues()
        {
            this.DefaultStyle.CharacterProperties.FontSize.LocalValue = Unit.PointToDip(11);
            this.DefaultStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily(ThemeFontType.Minor);
            this.DefaultStyle.ParagraphProperties.SpacingAfter.LocalValue = 12;
            this.DefaultStyle.ParagraphProperties.LineSpacing.LocalValue = 1.15;
            this.DefaultStyle.ParagraphProperties.LineSpacingType.LocalValue = HeightType.Auto;
        }

        private void RegisterDefaultStyles()
        {
            this.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.NormalStyleId);
            this.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.NormalTableStyleId);
        }
    }
}
