﻿#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Media;
#else
using System.Windows;
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Flow.Model.Cloning;
using Telerik.Windows.Documents.Flow.Model.Styles;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Utilities;
using BaselineAlignment = Telerik.Windows.Documents.Flow.Model.Styles.BaselineAlignment;
using FlowDirection = Telerik.Windows.Documents.Flow.Model.Styles.FlowDirection;

namespace Telerik.Windows.Documents.Flow.Model
{
    /// <summary>
    /// An inline-level flow content element intended to contain a run of formatted or unformatted text.
    /// </summary>
    public sealed partial class Run : InlineBase, IElementWithStyle, IElementWithProperties
    {
        private readonly CharacterProperties properties;

        private string text;

        /// <summary>
        /// Initializes a new instance of the <see cref="Run"/> class.
        /// </summary>
        /// <param name="document">The document in which the element will be added.</param>
        public Run(RadFlowDocument document)
            : base(document)
        {
            this.text = string.Empty;

            this.properties = new CharacterProperties(this);
            this.Shading = new Shading(this.properties);
            this.Underline = new Underline(this.properties);
        }

        /// <summary>
        /// Gets the properties.
        /// </summary>
        /// <value>The properties.</value>
        DocumentElementPropertiesBase IElementWithProperties.Properties
        {
            get
            {
                return this.properties;
            }
        }

        /// <summary>
        /// Gets the styling properties of this element.
        /// </summary>
        /// <value>The properties.</value>
        public CharacterProperties Properties
        {
            get
            {
                return this.properties;
            }
        }

        /// <summary>
        /// Gets or sets the text.
        /// </summary>
        /// <value>The text.</value>
        public string Text
        {
            get
            {
                return this.text;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "value");

                this.text = TextHelper.RemoveNewLines(value);
            }
        }

        /// <summary>
        /// Gets or sets the style id.
        /// </summary>
        /// <value>The style id.</value>
        public string StyleId
        {
            get
            {
                return this.Properties.StyleId;
            }
            set
            {
                this.Properties.StyleId = value;
            }
        }

        /// <summary>
        /// Gets or sets the font family.
        /// </summary>
        /// <value>The font family.</value>
        public ThemableFontFamily FontFamily
        {
            get
            {
                return this.properties.FontFamily.GetActualValue();
            }
            set
            {
                this.properties.FontFamily.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the size of the font.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The size of the font.</value>
        public double FontSize
        {
            get
            {
                return this.Properties.FontSize.GetActualValue().Value;
            }
            set
            {
                this.Properties.FontSize.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets the shading.
        /// </summary>
        /// <value>The shading.</value>
        public Shading Shading { get; private set; }

        /// <summary>
        /// Gets or sets the font style.
        /// </summary>
        /// <returns>The default value is <c>Normal</c>.</returns>
        public FontStyle FontStyle
        {
            get
            {
                return this.Properties.FontStyle.GetActualValue().Value;
            }
            set
            {
                this.Properties.FontStyle.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the font weight.
        /// </summary>
        /// <returns>The default value is <c>Normal</c>.</returns>
        public FontWeight FontWeight
        {
            get
            {
                return this.Properties.FontWeight.GetActualValue().Value;
            }
            set
            {
                this.Properties.FontWeight.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the color of the foreground.
        /// </summary>
        /// <returns>The default value is <c>black</c>.</returns>
        public ThemableColor ForegroundColor
        {
            get
            {
                return this.Properties.ForegroundColor.GetActualValue();
            }
            set
            {
                this.Properties.ForegroundColor.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the color of the highlight.
        /// </summary>
        /// <returns>The default value is <c>Transparent</c>.</returns>
        public Color HighlightColor
        {
            get
            {
                return this.Properties.HighlightColor.GetActualValue().Value;
            }
            set
            {
                this.Properties.HighlightColor.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the baseline alignment.
        /// </summary>
        /// <returns>The default value is <c>Baseline</c>.</returns>
        public BaselineAlignment BaselineAlignment
        {
            get
            {
                return this.Properties.BaselineAlignment.GetActualValue().Value;
            }
            set
            {
                this.Properties.BaselineAlignment.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the strikethrough.
        /// </summary>
        /// <returns>The default value is <c>false</c>.</returns>
        public bool Strikethrough
        {
            get
            {
                return this.Properties.Strikethrough.GetActualValue().Value;
            }
            set
            {
                this.Properties.Strikethrough.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets the underline.
        /// </summary>
        /// <value>The underline.</value>
        public Underline Underline { get; private set; }

        /// <summary>
        /// Gets or sets the flow direction.
        /// </summary>
        /// <returns>The default value is <c>LeftToRight</c></returns>
        public FlowDirection FlowDirection
        {
            get
            {
                return this.Properties.FlowDirection.GetActualValue().Value;
            }
            set
            {
                this.Properties.FlowDirection.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets the type of the element.
        /// </summary>
        /// <value>The type.</value>
        internal override DocumentElementType Type
        {
            get
            {
                return DocumentElementType.Run;
            }
        }

        /// <summary>
        /// Creates deep copy of this document element, associated to the current document.
        /// </summary>
        /// <returns>The cloned element.</returns>
        public Run Clone()
        {
            return this.CloneInternal(null);
        }

        /// <summary>
        /// Creates deep copy of this document element, associated to <paramref name="document"/>.
        /// </summary>
        /// <param name="document">The document to which the cloned element should be associated.</param>
        /// <returns>The cloned element.</returns>
        public Run Clone(RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");
            return this.CloneInternal(document);
        }

        /// <summary>
        /// Returns a string that represents the current object.
        /// </summary>
        /// <returns>A string that represents the current object.</returns>
        public override string ToString()
        {
            return this.Text;
        }

        internal override DocumentElementBase CloneCore(CloneContext cloneContext)
        {
            Guard.ThrowExceptionIfNull(cloneContext, "cloneContext");

            Run run = new Run(cloneContext.Document);

            run.text = this.text;
            run.Properties.CopyPropertiesFrom(this.Properties);

            if (cloneContext.RenamedStyles.ContainsKey(run.StyleId))
            {
                run.StyleId = cloneContext.RenamedStyles[run.StyleId];
            }

            return run;
        }

#if !NETSTANDARD
        /// <summary>
        /// This API supports the internal infrastructure and is not intended to be used directly from your code.
        /// </summary>
        internal override string GetChildrenDebuggerDisplayText()
        {
            return DebugVisualizer.EscapeXmlContent(this.Text);
        }
#endif

        private Run CloneInternal(RadFlowDocument document)
        {
            return (Run)this.CloneCore(new CloneContext(document ?? this.Document));
        }
    }
}
