﻿using System;
using System.Collections.Generic;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Primitives;
using Telerik.Windows.Documents.Flow.Model.Cloning;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Model;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model
{
    /// <summary>
    /// A flow content element used for grouping of <see cref="BlockBase"/> elements.
    /// </summary>
    /// <remarks>
    /// In addition to the collection of blocks, section can contain one or more instances 
    /// of <see cref="Header"/> and <see cref="Footer"/> through the <see cref="Section.Headers"/> and <see cref="Section.Footers"/> properties.
    /// </remarks>
    public sealed partial class Section : BlockContainerBase, IElementWithProperties
    {
        private readonly SectionProperties properties;
        private readonly Headers headers;
        private readonly Footers footers;
        private readonly PageNumberingSettings pageNumberingSettings;

        /// <summary>
        /// Initializes a new instance of the <see cref="Section"/> class.
        /// </summary>
        /// <param name="document">The document in which the element will be added.</param>
        public Section(RadFlowDocument document)
            : base(document)
        {
            this.properties = new SectionProperties(this);
            this.headers = new Headers(document, this);
            this.footers = new Footers(document, this);
            this.pageNumberingSettings = new PageNumberingSettings(this);
        }

        /// <summary>
        /// Gets all headers associated to this section.
        /// </summary>
        /// <value>The headers.</value>
        public Headers Headers
        {
            get
            {
                return this.headers;
            }
        }

        /// <summary>
        /// Gets all footers associated to this section.
        /// </summary>
        /// <value>The footers.</value>
        public Footers Footers
        {
            get
            {
                return this.footers;
            }
        }

        /// <summary>
        /// Gets the properties.
        /// </summary>
        /// <value>The properties.</value>
        DocumentElementPropertiesBase IElementWithProperties.Properties
        {
            get
            {
                return this.properties;
            }
        }

        /// <summary>
        /// Gets the styling properties of this element.
        /// </summary>
        /// <value>The properties.</value>
        public SectionProperties Properties
        {
            get
            {
                return this.properties;
            }
        }

        /// <summary>
        /// Gets the page numbering settings.
        /// </summary>
        /// <value>The page numbering settings.</value>
        public PageNumberingSettings PageNumberingSettings
        {
            get
            {
                return this.pageNumberingSettings;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating if the section has different header/footer for its first page.
        /// </summary>
        /// <returns>The default value is <c>false</c>.</returns>
        public bool HasDifferentFirstPageHeaderFooter
        {
            get
            {
                return this.Properties.HasDifferentFirstPageHeaderFooter.GetActualValue().Value;
            }
            set
            {
                this.Properties.HasDifferentFirstPageHeaderFooter.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the page orientation. Specifies the actual paper size to use when printing the file.
        /// </summary>
        /// <value>The page orientation.</value>
        public PageOrientation PageOrientation
        {
            get
            {
                return this.Properties.PageOrientation.GetActualValue().Value;
            }
            set
            {
                this.Properties.PageOrientation.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the size of the page.
        /// The width and height are in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The size of the page.</value>
        /// <remarks>
        /// For setting common page sizes <see cref="PaperTypeConverter.ToSize"/> method can be used.
        /// </remarks>
        public Size PageSize
        {
            get
            {
                return this.Properties.PageSize.GetActualValue().Value;
            }
            set
            {
                this.Properties.PageSize.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the page margins.
        /// </summary>
        /// <value>The page margins.</value>
        public Padding PageMargins
        {
            get
            {
                return this.Properties.PageMargins.GetActualValue();
            }
            set
            {
                this.Properties.PageMargins.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the top margin of the header.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The header top margin.</value>
        public double HeaderTopMargin
        {
            get
            {
                return this.Properties.HeaderTopMargin.GetActualValue().Value;
            }
            set
            {
                this.Properties.HeaderTopMargin.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the bottom margin of the footer.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The footer bottom margin.</value>
        public double FooterBottomMargin
        {
            get
            {
                return this.Properties.FooterBottomMargin.GetActualValue().Value;
            }
            set
            {
                this.Properties.FooterBottomMargin.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the type of the section.
        /// </summary>
        /// <value>The type of the section.</value>
        public SectionType SectionType
        {
            get
            {
                return this.Properties.SectionType.GetActualValue().Value;
            }
            set
            {
                this.Properties.SectionType.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the vertical alignment.
        /// </summary>
        /// <value>The vertical alignment.</value>
        public Styles.VerticalAlignment VerticalAlignment
        {
            get
            {
                return this.Properties.VerticalAlignment.GetActualValue();
            }
            set
            {
                this.Properties.VerticalAlignment.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets the type of the element.
        /// </summary>
        /// <value>The type.</value>
        internal override DocumentElementType Type
        {
            get
            {
                return DocumentElementType.Section;
            }
        }

        internal override IEnumerable<DocumentElementBase> Children
        {
            get
            {
                return base.Children.Concat<DocumentElementBase>(this.Headers.GetInstances())
                                    .Concat<DocumentElementBase>(this.Footers.GetInstances());
            }
        }

        internal override IEnumerable<DocumentElementBase> ContentChildren
        {
            get
            {
                return base.Children;
            }
        }

        /// <summary>
        /// Rotates the page orientation and changes the size and margins of the related section.
        /// </summary>
        /// <param name="orientation">The page orientation.</param>
        public void Rotate(PageOrientation orientation)
        {
            if (!(orientation == PageOrientation.Portrait || orientation == PageOrientation.Landscape))
            {
                throw new ArgumentException("The page orientation is not supported.", "orientation");
            }

            if (this.PageOrientation == orientation)
            {
                return;
            }

            Padding newPadding = this.PageMargins;
            switch (orientation)
            {
                case PageOrientation.Landscape:
                    newPadding = new Padding(newPadding.Bottom, newPadding.Left, newPadding.Top, newPadding.Right);
                    break;
                case PageOrientation.Portrait:
                    newPadding = new Padding(newPadding.Top, newPadding.Right, newPadding.Bottom, newPadding.Left);
                    break;
            }

            this.PageMargins = newPadding;
            this.PageSize = new Size(this.PageSize.Height, this.PageSize.Width);
            this.PageOrientation = orientation;
        }

        /// <summary>
        /// Creates deep copy of this document element, associated to the current document.
        /// </summary>
        /// <returns>The cloned element.</returns>
        public Section Clone()
        {
            return this.CloneInternal();
        }

        /// <summary>
        /// Creates deep copy of this document element, associated to <paramref name="document"/>.
        /// </summary>
        /// <param name="document">The document to which the cloned element should be associated.</param>
        /// <returns>The cloned element.</returns>
        public Section Clone(RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");
            return this.CloneInternal(document);
        }

        internal override DocumentElementBase CloneCore(CloneContext cloneContext)
        {
            Section section = this.ClonePropertiesAndHeadersFooters(cloneContext);
            section.Blocks.AddClonedChildrenFrom(this.Blocks, cloneContext);

            return section;
        }

        internal Section ClonePropertiesAndHeadersFooters(CloneContext cloneContext)
        {
            Guard.ThrowExceptionIfNull(cloneContext, "cloneContext");

            Section section = new Section(cloneContext.Document);
            cloneContext.CurrentSection = section;

            section.Headers.CloneHeadersFootersFrom(this.Headers, cloneContext);
            section.Footers.CloneHeadersFootersFrom(this.Footers, cloneContext);

            section.Properties.CopyPropertiesFrom(this.Properties);

            return section;
        }

        private Section CloneInternal(RadFlowDocument document = null)
        {
            return (Section)this.CloneCore(new CloneContext(document ?? this.Document));
        }
    }
}