﻿using Telerik.Windows.Documents.Flow.Model.Cloning;
using Telerik.Windows.Documents.Model.Data;
using Telerik.Windows.Documents.Model.Drawing.Shapes;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Shapes
{
    /// <summary>
    /// Represents inline anchor that contains floating image.
    /// </summary>
    public sealed class FloatingImage : ShapeAnchorBase
    {
        private readonly Image image;

        /// <summary>
        /// Initializes a new instance of the <see cref="FloatingImage" /> class.
        /// </summary>
        /// <param name="document">The document.</param>
        public FloatingImage(RadFlowDocument document)
            : this(document, new Image())
        {
        }

        internal FloatingImage(RadFlowDocument document, Image image)
            : base(document)
        {
            this.image = image;
            this.image.ImageSourceChanged += this.Image_ImageSourceChanged;

            if (this.image.ImageSource != null)
            {
                this.Document.Resources.RegisterResource(this.image.ImageSource);
            }
        }

        /// <summary>
        /// Gets the image.
        /// </summary>
        /// <value>The image.</value>
        public Image Image
        {
            get
            {
                return this.image;
            }
        }

        /// <summary>
        /// Gets the type of the element.
        /// </summary>
        /// <value>The type.</value>
        internal override DocumentElementType Type
        {
            get
            {
                return DocumentElementType.FloatingImage;
            }
        }

        internal override ShapeBase Shape
        {
            get
            {
                return this.image;
            }
        }

        /// <summary>
        /// Creates deep copy of this document element, associated to the current document.
        /// </summary>
        /// <returns>The cloned element.</returns>
        public FloatingImage Clone()
        {
            return this.CloneInternal(null);
        }

        /// <summary>
        /// Creates deep copy of this document element, associated to <paramref name="document"/>.
        /// </summary>
        /// <param name="document">The document to which the cloned element should be associated.</param>
        /// <returns>The cloned element.</returns>
        public FloatingImage Clone(RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");
            return this.CloneInternal(document);
        }

        internal override DocumentElementBase CloneCore(CloneContext cloneContext)
        {
            Guard.ThrowExceptionIfNull(cloneContext, "cloneContext");

            Image clonedImage = this.Image.Clone(this.Document != cloneContext.Document);

            FloatingImage floatingImage = new FloatingImage(cloneContext.Document, clonedImage);
            floatingImage.ClonePropertiesFrom(this);

            return floatingImage;
        }

        private FloatingImage CloneInternal(RadFlowDocument document = null)
        {
            return (FloatingImage)this.CloneCore(new CloneContext(document ?? this.Document));
        }

        private void Image_ImageSourceChanged(object sender, ResourceChangedEventArgs e)
        {
            if (e.OldValue != null)
            {
                this.Document.Resources.ReleaseResource(e.OldValue);
            }

            if (e.NewValue != null)
            {
                this.Document.Resources.RegisterResource(e.NewValue);
            }
        }
    }
}
