﻿using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Model.Drawing.Shapes;
using Telerik.Windows.Documents.Utilities;
using Telerik.Windows.Documents.Primitives;

namespace Telerik.Windows.Documents.Flow.Model.Shapes
{
    /// <summary>
    /// Represents base inline holder for floating shape element.
    /// </summary>
    public abstract class ShapeAnchorBase : InlineBase
    {
        private ShapeWrapping wrapping;
        private Padding margin;
        private HorizontalPosition horizontalPosition;
        private VerticalPosition verticalPosition;

        internal ShapeAnchorBase(RadFlowDocument document)
            : base(document)
        {
            this.HorizontalPosition = new HorizontalPosition();
            this.VerticalPosition = new VerticalPosition();
            this.Wrapping = new ShapeWrapping();
            this.Margin = DocumentDefaultStyleSettings.FloatingImageMargin;
            this.LayoutInCell = true;
            this.AllowOverlap = true;
        }

        /// <summary>
        /// Specifies whether the shape shall be allowed to overlap the contents of the other shape object.
        /// </summary>
        /// <value>The allow overlap value.</value>
        public bool AllowOverlap { get; set; }

        /// <summary>
        /// Gets or sets value indicating if the anchor can be modified at runtime.
        /// </summary>
        /// <value>The value indicating if the anchor can be modified at runtime..</value>
        public bool IsLocked { get; set; }

        /// <summary>
        /// Gets or sets the shape wrapping type.
        /// </summary>
        /// <value>The wrapping.</value>
        public ShapeWrapping Wrapping
        {
            get
            {
                return this.wrapping;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "wrapping");

                this.wrapping = value;
            }
        }

        /// <summary>
        /// Gets or sets value indicating if the layout of the shape should be calculated
        /// relative to the cell that is holding the shape.
        /// </summary>
        /// <value>True if the layout of the shape should be calculated relative to the cell that is holding the shape.</value>
        public bool LayoutInCell { get; set; }

        /// <summary>
        /// Gets or sets the value indicating if the shape should be displayed behind document content.
        /// </summary>
        /// <value>The is behind document.</value>
        public bool IsBehindDocument { get; set; }

        /// <summary>
        /// Gets or sets the Z index of the shape.
        /// </summary>
        /// <value> The Z index of the shape.</value>
        public int ZIndex { get; set; }

        /// <summary>
        /// Gets or sets the margin.
        /// </summary>
        /// <value>The margin.</value>
        public Padding Margin
        {
            get
            {
                return this.margin;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "margin");

                this.margin = value;
            }
        }

        /// <summary>
        /// Gets or sets the horizontal position.
        /// </summary>
        /// <value>The horizontal position.</value>
        public HorizontalPosition HorizontalPosition
        {
            get
            {
                return this.horizontalPosition;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "horizontalPosition");

                this.horizontalPosition = value;
            }
        }

        /// <summary>
        /// Gets or sets the vertical position.
        /// </summary>
        /// <value>The vertical position.</value>
        public VerticalPosition VerticalPosition
        {
            get
            {
                return this.verticalPosition;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "verticalPosition");

                this.verticalPosition = value;
            }
        }

        internal abstract ShapeBase Shape { get; }

        /// <summary>
        /// Creates deep copy of the element properties of the <paramref name="fromShapeAnchor"/> and sets them to this instance.
        /// </summary>
        /// <param name="fromShapeAnchor">The shape anchor to clone property elements from.</param>
        protected void ClonePropertiesFrom(ShapeAnchorBase fromShapeAnchor)
        {
            this.AllowOverlap = fromShapeAnchor.AllowOverlap;
            this.IsLocked = fromShapeAnchor.IsLocked;
            this.Wrapping = fromShapeAnchor.Wrapping.Clone();
            this.LayoutInCell = fromShapeAnchor.LayoutInCell;
            this.IsBehindDocument = fromShapeAnchor.IsBehindDocument;
            this.ZIndex = fromShapeAnchor.ZIndex;
            this.Margin = fromShapeAnchor.Margin.Clone();
            this.HorizontalPosition = fromShapeAnchor.HorizontalPosition.Clone();
            this.VerticalPosition = fromShapeAnchor.VerticalPosition.Clone();
        }
    }
}