using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Shapes
{
    /// <summary>
    /// Specifies vertical shape positioning.
    /// </summary>
    public class VerticalPosition
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="VerticalPosition" /> class.
        /// </summary>
        public VerticalPosition()
            : this(VerticalRelativeFrom.Paragraph, 0d)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="VerticalPosition" /> class.
        /// </summary>
        /// <param name="relativeFrom">The relative from.</param>
        /// <param name="offset">The offset.</param>
        public VerticalPosition(VerticalRelativeFrom relativeFrom, double offset)
        {
            this.RelativeFrom = relativeFrom;
            this.ValueType = PositionValueType.Offset;
            this.Offset = offset;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="VerticalPosition" /> class.
        /// </summary>
        /// <param name="relativeFrom">The relative from.</param>
        /// <param name="alignment">The alignment.</param>
        public VerticalPosition(VerticalRelativeFrom relativeFrom, RelativeVerticalAlignment alignment)
        {
            this.RelativeFrom = relativeFrom;
            this.ValueType = PositionValueType.Alignment;
            this.Alignment = alignment;
        }

        /// <summary>
        /// Gets or sets the type of the positioning.
        /// </summary>
        /// <value>The type of the positioning.</value>
        public PositionValueType ValueType { get; set; }

        /// <summary>
        /// Gets or sets the relative horizontal positioning base.
        /// </summary>
        /// <value>The vertical relative from value.</value>
        public VerticalRelativeFrom RelativeFrom { get; set; }

        /// <summary>
        /// Gets or sets the alignment to be used if the position type is alignment.
        /// </summary>
        /// <value>The alignment.</value>
        public RelativeVerticalAlignment Alignment { get; set; }

        /// <summary>
        /// Gets or sets the offset to be used if the position type is offset.
        /// </summary>
        /// <value>The offset.</value>
        public double Offset { get; set; }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            VerticalPosition other = obj as VerticalPosition;
            if (other == null)
            {
                return false;
            }

            return this.ValueType == other.ValueType
                && this.RelativeFrom == other.RelativeFrom
                && this.Alignment == other.Alignment
                && this.Offset == other.Offset;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return ObjectExtensions.CombineHashCodes(
                this.ValueType.GetHashCode(),
                this.RelativeFrom.GetHashCode(),
                this.Alignment.GetHashCode(),
                this.Offset.GetHashCode());
        }

        internal VerticalPosition Clone()
        {
            VerticalPosition verticalPosition = new VerticalPosition()
            {
                Alignment = this.Alignment,
                Offset = this.Offset,
                RelativeFrom = this.RelativeFrom,
                ValueType = this.ValueType
            };

            return verticalPosition;
        }
    }
}