﻿using System;
using System.Collections.Generic;

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Establishes functionality to work with built in styles.
    /// </summary>
    public static class BuiltInStyles
    {
        private static readonly Dictionary<string, BuiltInStyleMetaData> styleGenerationMetadatas = new Dictionary<string, BuiltInStyleMetaData>();

        static BuiltInStyles()
        {
            BuiltInStyles.InitializeCreationOfStyles();
        }

        /// <summary>
        /// Gets built-in style by ID.
        /// </summary>
        /// <param name="styleId">The style ID.</param>
        /// <exception cref="T:System.ArgumentException"><paramref name="styleId"/> is not an ID of a built-in style.</exception>
        /// <returns>The built-in style.</returns>
        public static Style GetStyle(string styleId)
        {
            if (!BuiltInStyles.IsBuiltInStyle(styleId))
            {
                throw new ArgumentException("The style with ID '{0}' is not built-in style.", styleId);
            }

            BuiltInStyleMetaData metaData = styleGenerationMetadatas.GetValueOrNull(styleId);
            Style style = metaData.GetStyle();

            return style;
        }

        /// <summary>
        /// Determines whether a style ID is ID of a built-in style.
        /// </summary>
        /// <param name="styleId">The style ID.</param>
        /// <returns></returns>
        public static bool IsBuiltInStyle(string styleId)
        {
            return BuiltInStyles.styleGenerationMetadatas.ContainsKey(styleId);
        }

        /// <summary>
        /// Gets all primary styles.
        /// </summary>
        /// <returns></returns>
        public static IEnumerable<Style> GetAllPrimaryStyles()
        {
            LinkedList<Style> linkedList = new LinkedList<Style>();
            foreach (KeyValuePair<string, BuiltInStyleMetaData> item in BuiltInStyles.styleGenerationMetadatas)
            {
                BuiltInStyleMetaData metaData = item.Value;
                if (metaData.IsPrimary)
                {
                    linkedList.AddLast(metaData.GetStyle());
                }
            }
            return linkedList;
        }

        /// <summary>
        /// Gets all styles.
        /// </summary>
        /// <param name="type">The type.</param>
        /// <returns></returns>
        public static IEnumerable<Style> GetAllStyles(StyleType type)
        {
            LinkedList<Style> linkedList = new LinkedList<Style>();
            foreach (KeyValuePair<string, BuiltInStyleMetaData> item in BuiltInStyles.styleGenerationMetadatas)
            {
                BuiltInStyleMetaData metaData = item.Value;
                if (metaData.StyleType == type)
                {
                    linkedList.AddLast(metaData.GetStyle());
                }
            }
            return linkedList;
        }

        /// <summary>
        /// Gets all styles.
        /// </summary>
        /// <returns></returns>
        public static IEnumerable<Style> GetAllStyles()
        {
            LinkedList<Style> linkedList = new LinkedList<Style>();
            foreach (KeyValuePair<string, BuiltInStyleMetaData> item in BuiltInStyles.styleGenerationMetadatas)
            {
                BuiltInStyleMetaData metaData = item.Value;
                linkedList.AddLast(metaData.GetStyle());
            }
            return linkedList;
        }

        /// <summary>
        /// Gets the metadata for built-in style.
        /// </summary>
        /// <param name="styleId">The style id.</param>
        /// <returns></returns>
        internal static BuiltInStyleMetaData GetStyleMetadata(string styleId)
        {
            if (!BuiltInStyles.IsBuiltInStyle(styleId))
            {
                throw new ArgumentException("The style with ID '{0}' is not built-in style.", styleId);
            }

            BuiltInStyleMetaData metaData = styleGenerationMetadatas.GetValueOrNull(styleId);
            return metaData;
        }

        private static void InitializeCreationOfStyles()
        {
            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.NormalStyleName,
                new BuiltInStyleMetaData()
                {
                    StyleType = StyleType.Paragraph,
                    IsPrimary = true,
                    UIPriority = 0,
                    CreateStyleMethod = BuiltInStylesGenerator.NormalStyle
                });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.NormalWebStyleId,
                new BuiltInStyleMetaData()
                {
                    StyleType = StyleType.Paragraph,
                    IsPrimary = false,
                    UIPriority = 99,
                    CreateStyleMethod = BuiltInStylesGenerator.NormalWebStyle
                });

            BuiltInStyles.InitializeHeadingStyles();
            BuiltInStyles.InitializeTableStyles();
            BuiltInStyles.InitializeTocStyles();
            BuiltInStyles.InitializeFootnotesAndEndnotes();

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.CaptionStyleId,
               new BuiltInStyleMetaData()
               {
                   StyleType = StyleType.Paragraph,
                   IsPrimary = true,
                   UIPriority = 35,
                   CreateStyleMethod = BuiltInStylesGenerator.CaptionStyle
               });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.HyperlinkStyleName,
                new BuiltInStyleMetaData()
                {
                    StyleType = StyleType.Character,
                    IsPrimary = true,
                    UIPriority = 99,
                    CreateStyleMethod = BuiltInStylesGenerator.HyperlinkStyle
                });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.TofStyleId,
                new BuiltInStyleMetaData()
                {
                    StyleType = Styles.StyleType.Paragraph,
                    IsPrimary = false,
                    UIPriority = 99,
                    CreateStyleMethod = BuiltInStylesGenerator.TofStyle
                });
        }

        private static void InitializeTableStyles()
        {
            BuiltInStyles.AddTableNormalTableStyle();

            BuiltInStyles.AddPlainTableStyles();
        }

        private static void AddTableNormalTableStyle()
        {
            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.NormalTableStyleId,
                new BuiltInStyleMetaData()
                {
                    StyleType = StyleType.Table,
                    IsPrimary = false,
                    UIPriority = 59,
                    CreateStyleMethod = BuiltInStylesGenerator.TableNormalStyle
                });
        }

        private static void AddPlainTableStyles()
        {
            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.TableGridStyleId,
                   new BuiltInStyleMetaData()
                   {
                       StyleType = StyleType.Table,
                       IsPrimary = false,
                       UIPriority = 59,
                       CreateStyleMethod = BuiltInStylesGenerator.TableGridStyle
                   });
        }

        private static void InitializeHeadingStyles()
        {
            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetHeadingStyleIdByIndex(1),
                new BuiltInStyleMetaData()
                {
                    StyleType = StyleType.Paragraph,
                    IsPrimary = true,
                    UIPriority = 9,
                    CreateStyleMethod = BuiltInStylesGenerator.Heading1Style
                });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetHeadingStyleIdByIndex(2),
                new BuiltInStyleMetaData()
                {
                    StyleType = StyleType.Paragraph,
                    IsPrimary = true,
                    UIPriority = 9,
                    CreateStyleMethod = BuiltInStylesGenerator.Heading2Style
                });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetHeadingStyleIdByIndex(3),
                new BuiltInStyleMetaData()
                {
                    StyleType = StyleType.Paragraph,
                    IsPrimary = true,
                    UIPriority = 9,
                    CreateStyleMethod = BuiltInStylesGenerator.Heading3Style
                });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetHeadingStyleIdByIndex(4),
                new BuiltInStyleMetaData()
                {
                    StyleType = StyleType.Paragraph,
                    IsPrimary = true,
                    UIPriority = 9,
                    CreateStyleMethod = BuiltInStylesGenerator.Heading4Style
                });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetHeadingStyleIdByIndex(5),
                new BuiltInStyleMetaData()
                {
                    StyleType = StyleType.Paragraph,
                    IsPrimary = true,
                    UIPriority = 9,
                    CreateStyleMethod = BuiltInStylesGenerator.Heading5Style
                });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetHeadingStyleIdByIndex(6),
                new BuiltInStyleMetaData()
                {
                    StyleType = StyleType.Paragraph,
                    IsPrimary = true,
                    UIPriority = 9,
                    CreateStyleMethod = BuiltInStylesGenerator.Heading6Style
                });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetHeadingStyleIdByIndex(7),
                new BuiltInStyleMetaData()
                {
                    StyleType = StyleType.Paragraph,
                    IsPrimary = true,
                    UIPriority = 9,
                    CreateStyleMethod = BuiltInStylesGenerator.Heading7Style
                });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetHeadingStyleIdByIndex(8),
                new BuiltInStyleMetaData()
                {
                    StyleType = StyleType.Paragraph,
                    IsPrimary = true,
                    UIPriority = 9,
                    CreateStyleMethod = BuiltInStylesGenerator.Heading8Style
                });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetHeadingStyleIdByIndex(9),
                new BuiltInStyleMetaData()
                {
                    StyleType = StyleType.Paragraph,
                    IsPrimary = true,
                    UIPriority = 9,
                    CreateStyleMethod = BuiltInStylesGenerator.Heading9Style
                });
        }

        private static void InitializeTocStyles()
        {
            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetTocStyleIdByIndex(1),
               new BuiltInStyleMetaData()
               {
                   StyleType = StyleType.Paragraph,
                   IsPrimary = false,
                   UIPriority = 39,
                   CreateStyleMethod = BuiltInStylesGenerator.GetToc1Style
               });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetTocStyleIdByIndex(2),
               new BuiltInStyleMetaData()
               {
                   StyleType = StyleType.Paragraph,
                   IsPrimary = false,
                   UIPriority = 39,
                   CreateStyleMethod = BuiltInStylesGenerator.GetToc2Style
               });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetTocStyleIdByIndex(3),
               new BuiltInStyleMetaData()
               {
                   StyleType = StyleType.Paragraph,
                   IsPrimary = false,
                   UIPriority = 39,
                   CreateStyleMethod = BuiltInStylesGenerator.GetToc3Style
               });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetTocStyleIdByIndex(4),
               new BuiltInStyleMetaData()
               {
                   StyleType = StyleType.Paragraph,
                   IsPrimary = false,
                   UIPriority = 39,
                   CreateStyleMethod = BuiltInStylesGenerator.GetToc4Style
               });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetTocStyleIdByIndex(5),
               new BuiltInStyleMetaData()
               {
                   StyleType = StyleType.Paragraph,
                   IsPrimary = false,
                   UIPriority = 39,
                   CreateStyleMethod = BuiltInStylesGenerator.GetToc5Style
               });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetTocStyleIdByIndex(6),
               new BuiltInStyleMetaData()
               {
                   StyleType = StyleType.Paragraph,
                   IsPrimary = false,
                   UIPriority = 39,
                   CreateStyleMethod = BuiltInStylesGenerator.GetToc6Style
               });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetTocStyleIdByIndex(7),
               new BuiltInStyleMetaData()
               {
                   StyleType = StyleType.Paragraph,
                   IsPrimary = false,
                   UIPriority = 39,
                   CreateStyleMethod = BuiltInStylesGenerator.GetToc7Style
               });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetTocStyleIdByIndex(8),
               new BuiltInStyleMetaData()
               {
                   StyleType = StyleType.Paragraph,
                   IsPrimary = false,
                   UIPriority = 39,
                   CreateStyleMethod = BuiltInStylesGenerator.GetToc8Style
               });

            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.GetTocStyleIdByIndex(9),
               new BuiltInStyleMetaData()
               {
                   StyleType = StyleType.Paragraph,
                   IsPrimary = false,
                   UIPriority = 39,
                   CreateStyleMethod = BuiltInStylesGenerator.GetToc9Style
               });
        }

        private static void InitializeFootnotesAndEndnotes()
        {
            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.FootnoteReferenceStyleId,
                new BuiltInStyleMetaData()
                {
                    StyleType = StyleType.Character,
                    IsPrimary = false,
                    UIPriority = 99,
                    CreateStyleMethod = BuiltInStylesGenerator.FootnoteReferenceStyle
                });
            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.FootnoteTextStyleId,
                new BuiltInStyleMetaData()
                {
                    StyleType = StyleType.Paragraph,
                    IsPrimary = false,
                    UIPriority = 99,
                    CreateStyleMethod = BuiltInStylesGenerator.FootnoteTextStyle
                });
            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.EndnoteReferenceStyleId,
                new BuiltInStyleMetaData()
                {
                    StyleType = StyleType.Character,
                    IsPrimary = false,
                    UIPriority = 99,
                    CreateStyleMethod = BuiltInStylesGenerator.EndnoteReferenceStyle
                });
            BuiltInStyles.styleGenerationMetadatas.Add(BuiltInStyleNames.EndnoteTextStyleId,
                new BuiltInStyleMetaData()
                {
                    StyleType = StyleType.Paragraph,
                    IsPrimary = false,
                    UIPriority = 99,
                    CreateStyleMethod = BuiltInStylesGenerator.EndnoteTextStyle
                });
        }
    }
}
