﻿using Telerik.Windows.Documents.Flow.Model.Lists;

namespace Telerik.Windows.Documents.Flow.Model.Styles.Core
{
    internal class StyleProperty<TValue> : StylePropertyBase<TValue>
    {
        public StyleProperty(StylePropertyDefinition<TValue> propertyDefinition, DocumentElementPropertiesBase propertyContainer)
            : base(propertyDefinition, propertyContainer)
        {
        }

        public override TValue GetActualValue()
        {
            if (this.HasLocalValue)
            {
                return this.LocalValue;
            }

            return this.GetActualValueInternal();
        }

        internal TValue GetDocumentDefaultValue(RadFlowDocument rootDocument)
        {
            TValue result = default(TValue);
            if (rootDocument != null)
            {
                IStyleProperty documentDefaultStyleProperty = null;

                if (this.PropertyContainer is CharacterProperties)
                {
                    documentDefaultStyleProperty = rootDocument.DefaultStyle.CharacterProperties.GetStyleProperty(this.PropertyDefinition);
                }
                else if (this.PropertyContainer is ParagraphProperties)
                {
                    documentDefaultStyleProperty = rootDocument.DefaultStyle.ParagraphProperties.GetStyleProperty(this.PropertyDefinition);
                }

                if (documentDefaultStyleProperty != null)
                {
                    object documentDefaultValue = documentDefaultStyleProperty.GetLocalValueAsObject();
                    if (documentDefaultValue != null)
                    {
                        result = (TValue)documentDefaultValue;
                    }
                }
            }

            return result;
        }

        protected TValue GetActualValueInternal()
        {
            if (!this.PropertyContainer.SuppressStylePropertyEvaluation && this.PropertyContainer.Document != null)
            {
                if (this.PropertyContainer.OwnerDocumentElement != null)
                {
                    return this.GetValueFromDocumentElement();
                }
                else if (this.PropertyContainer.OwnerStyle != null)
                {
                    TValue styleValue = this.GetValueFromStyle();
                    if (styleValue != null)
                    {
                        return styleValue;
                    }
                    return this.DefaultValue;
                }
                else if (this.PropertyContainer.OwnerListLevel != null)
                {
                    return this.GetValueFromListLevel();
                }
            }

            return this.DefaultValue;
        }

        private TValue GetValueFromDocumentElement()
        {
            DocumentElementBase ownerElement = this.PropertyContainer.OwnerDocumentElement;
            TValue elementValue = this.GetElementValue(ownerElement);
            if (elementValue != null)
            {
                return elementValue;
            }

            TValue defaultStyleValue = this.GetValueFromDefaultStyle(ownerElement.Document);
            if (defaultStyleValue != null)
            {
                return defaultStyleValue;
            }

            return this.DefaultValue;
        }

        private TValue GetValueFromStyle()
        {
            string linkedStyleId = this.GetProperLinkedStyleId();
            if (!string.IsNullOrEmpty(linkedStyleId))
            {
                Style linkedStyle = this.PropertyContainer.OwnerStyle.Document.StyleRepository.GetStyle(linkedStyleId);
                if (linkedStyle != null)
                {
                    object linkedPropertyValue = linkedStyle.GetPropertyValue(this.PropertyDefinition);
                    if (linkedPropertyValue != null)
                    {
                        return (TValue)linkedPropertyValue;
                    }
                }
            }

            if (!string.IsNullOrEmpty(this.PropertyContainer.OwnerStyle.BasedOnStyleId))
            {
                Style basedOnStyle = this.PropertyContainer.OwnerStyle.Document.StyleRepository.GetStyle(this.PropertyContainer.OwnerStyle.BasedOnStyleId);
                if (basedOnStyle != null)
                {
                    object basedOnValue = basedOnStyle.GetPropertyValue(this.PropertyDefinition);
                    if (basedOnValue != null)
                    {
                        return (TValue)basedOnValue;
                    }
                }
            }

            // Linked style is based on paragraph style. Evaluating the character part.
            if (this.PropertyContainer.OwnerStyle.StyleType == StyleType.Character &&
                !string.IsNullOrEmpty(this.PropertyContainer.OwnerStyle.LinkedStyleId))
            {
                Style linkedStyle = this.PropertyContainer.OwnerStyle.Document.StyleRepository.GetStyle(this.PropertyContainer.OwnerStyle.LinkedStyleId);
                if (linkedStyle != null && linkedStyle.StyleType == StyleType.Paragraph && !string.IsNullOrEmpty(linkedStyle.BasedOnStyleId))
                {
                    Style basedOnStyle = this.PropertyContainer.OwnerStyle.Document.StyleRepository.GetStyle(linkedStyle.BasedOnStyleId);
                    if (basedOnStyle != null)
                    {
                        object basedOnValue = basedOnStyle.GetPropertyValue(this.PropertyDefinition);
                        if (basedOnValue != null)
                        {
                            return (TValue)basedOnValue;
                        }
                    }
                }
            }

            // Styles with type different from table (paragraph, character, etc) and not based on any other style 
            // do not fall back to the properties of the default style for this type.
            // example: if style "MyStyle" is not based on style "Normal" then it will not inherit any property from "Normal".
            if (this.PropertyContainer.OwnerStyle.StyleType == StyleType.Table && !this.PropertyContainer.OwnerStyle.IsDefault)
            {
                Style defaultStyle = this.PropertyContainer.OwnerStyle.Document.StyleRepository.GetDefaultStyle(this.PropertyContainer.OwnerStyle.StyleType);
                if (defaultStyle != null)
                {
                    object defaultStyleValue = defaultStyle.GetPropertyValue(this.PropertyDefinition);
                    if (defaultStyleValue != null)
                    {
                        return (TValue)defaultStyleValue;
                    }
                }
            }
            else
            {
                TValue documentDefaultValue = this.GetDocumentDefaultValue(this.PropertyContainer.OwnerStyle.Document);
                if (documentDefaultValue != null)
                {
                    return documentDefaultValue;
                }
            }

            return default(TValue);
        }

        private string GetProperLinkedStyleId()
        {
            if (this.PropertyContainer.OwnerStyle == null)
            {
                return null;
            }

            if (string.IsNullOrEmpty(this.PropertyContainer.OwnerStyle.LinkedStyleId))
            {
                return null;
            }

            if (this.PropertyContainer is CharacterProperties)
            {
                if (this.PropertyContainer.OwnerStyle.StyleType == StyleType.Character)
                {
                    return null;
                }

                return this.PropertyContainer.OwnerStyle.LinkedStyleId;
            }

            if (this.PropertyContainer is ParagraphProperties)
            {
                if (this.PropertyContainer.OwnerStyle.StyleType == StyleType.Paragraph)
                {
                    return null;
                }

                return this.PropertyContainer.OwnerStyle.LinkedStyleId;
            }

            return null;
        }

        private TValue GetElementValue(DocumentElementBase element)
        {
            TValue result = default(TValue);

            if (element == null)
            {
                return result;
            }

            RadFlowDocument rootDocument = element as RadFlowDocument;
            if (rootDocument != null)
            {
                result = this.GetValueFromDefaultStyle(rootDocument);
                return result;
            }

            IElementWithStyle parentWithStyle = element as IElementWithStyle;
            if (parentWithStyle != null)
            {
                if (string.IsNullOrEmpty(parentWithStyle.StyleId) && element.Type == DocumentElementType.Paragraph)
                {
                    Style style = element.Document.StyleRepository.GetDefaultStyle(StyleType.Paragraph);
                    if (style != null)
                    {
                        IStyleProperty styleProperty = style.GetStyleProperty(this.PropertyDefinition);
                        if (styleProperty != null && styleProperty.HasLocalValue)
                        {
                            object value = styleProperty.GetLocalValueAsObject();
                            if (value != null)
                            {
                                return (TValue)value;
                            }
                        }
                    }
                }
                else if (!string.IsNullOrEmpty(parentWithStyle.StyleId))
                {
                    Style style = element.Document.StyleRepository.GetStyle(parentWithStyle.StyleId);
                    if (style != null)
                    {
                        object value = style.GetPropertyValue(this.PropertyDefinition);
                        if (value != null)
                        {
                            return (TValue)value;
                        }
                    }
                }
            }

            // when the parent paragraph has assigned list level with assigned style.
            if (element.Type == DocumentElementType.Paragraph && this.PropertyDefinition != Paragraph.ListIdPropertyDefinition && this.PropertyDefinition != Paragraph.ListLevelPropertyDefinition)
            {
                Paragraph parentParagraph = (Paragraph)element;
                if (parentParagraph.ListId > Paragraph.ListIdPropertyDefinition.DefaultValue &&
                    parentParagraph.ListLevel > Paragraph.ListLevelPropertyDefinition.DefaultValue)
                {
                    List list = parentParagraph.Document.Lists.GetList(parentParagraph.ListId);
                    if (list != null && !string.IsNullOrEmpty(list.Levels[parentParagraph.ListLevel].StyleId))
                    {
                        Style style = element.Document.StyleRepository.GetStyle(list.Levels[parentParagraph.ListLevel].StyleId);
                        if (style != null)
                        {
                            object value = style.GetPropertyValue(this.PropertyDefinition);
                            if (value != null)
                            {
                                return (TValue)value;
                            }
                        }
                    }
                }
            }

            if (result == null)
            {
                DocumentElementBase parent;
                if (element is Table)
                {
                    parent = element.Document;
                }
                else
                {
                    parent = element.Parent;
                }

                result = this.GetElementValue(parent);
            }

            return result;
        }

        private TValue GetValueFromDefaultStyle(RadFlowDocument document, StyleType? assignableType = null)
        {
            TValue returnValue = default(TValue);
            if (!assignableType.HasValue)
            {
                // the default character style is not respected in MS Word.
                if (this.PropertyContainer.OwnerDocumentElement is Paragraph || this.PropertyContainer.OwnerDocumentElement is Run)
                {
                    assignableType = StyleType.Paragraph;
                }
                else if (this.PropertyContainer.OwnerDocumentElement is Table)
                {
                    assignableType = StyleType.Table;
                }
            }

            if (assignableType.HasValue)
            {
                Style defaultStyle = document.StyleRepository.GetDefaultStyle(assignableType.Value);
                if (defaultStyle != null)
                {
                    object defaultStyleValue = defaultStyle.GetPropertyValue(this.PropertyDefinition);
                    if (defaultStyleValue != null)
                    {
                        returnValue = (TValue)defaultStyleValue;
                    }
                }
            }
            return returnValue;
        }

        private TValue GetValueFromListLevel()
        {
            if (!string.IsNullOrEmpty(this.PropertyContainer.OwnerListLevel.StyleId))
            {
                Style style = this.PropertyContainer.OwnerListLevel.Document.StyleRepository.GetStyle(this.PropertyContainer.OwnerListLevel.StyleId);
                if (style != null && style.StyleType == StyleType.Paragraph)
                {
                    object propertyValue = style.GetPropertyValue(this.PropertyDefinition);
                    if (propertyValue != null)
                    {
                        return (TValue)propertyValue;
                    }
                }
            }

            TValue defaultStyleValue = this.GetValueFromDefaultStyle(this.PropertyContainer.OwnerListLevel.Document, StyleType.Paragraph);
            if (defaultStyleValue != null)
            {
                return defaultStyleValue;
            }

            return this.DefaultValue;
        }
    }
}