﻿using System;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif
#if NETSTANDARD
using Telerik.Documents.Media;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Describes the settings which are used to determine how the border will appear in the document. This class is immutable.
    /// </summary>
    [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1036:OverrideMethodsOnComparableTypes")]
    public class Border : IComparable<Border>
    {
        internal static readonly Border DefaultBorder = new Border(0, BorderStyle.Inherit, new ThemableColor(Colors.Black));
        internal static readonly Border EmptyBorder = new Border(0, BorderStyle.None, new ThemableColor(Colors.Black));

        private readonly ThemableColor color;
        private readonly bool shadow;
        private readonly bool frame;
        private readonly double spacing;
        private readonly BorderStyle style;
        private readonly double thickness;

        /// <summary>
        /// Initializes a new instance of the <see cref="Border" /> class.
        /// </summary>
        /// <param name="border">The border.</param>
        public Border(Border border)
        {
            this.thickness = border.Thickness;
            this.style = border.Style;
            this.color = border.Color;
            this.shadow = border.Shadow;
            this.frame = border.Frame;
            this.spacing = border.Spacing;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Border" /> class.
        /// </summary>
        /// <param name="style">The style.</param>
        public Border(BorderStyle style)
            : this(0, style, new ThemableColor(Colors.Transparent), false, false, 0)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Border" /> class.
        /// </summary>
        /// <param name="thickness">The thickness.</param>
        /// <param name="style">The style.</param>
        /// <param name="color">The color.</param>
        public Border(double thickness, BorderStyle style, ThemableColor color)
            : this(thickness, style, color, false, false, 0)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Border" /> class.
        /// </summary>
        /// <param name="thickness">The thickness.</param>
        /// <param name="style">The style.</param>
        /// <param name="color">The color.</param>
        /// <param name="shadow">The shadow.</param>
        /// <param name="frame">The frame.</param>
        /// <param name="spacing">The spacing.</param>
        public Border(double thickness, BorderStyle style, ThemableColor color, bool shadow, bool frame, double spacing)
        {
            this.thickness = thickness;
            this.style = style;
            this.shadow = shadow;
            this.frame = frame;
            this.spacing = spacing;

            if (color == null && (this.Style == BorderStyle.None || this.Style == BorderStyle.Inherit))
            {
                // It is allowed to pass null as color , only if the style is None or Inherit
                color = DefaultBorder.Color;
            }
            this.color = color;
        }

        /// <summary>
        /// Gets the color of the border.
        /// </summary>
        /// <value>The color.</value>
        public ThemableColor Color
        {
            get { return this.color; }
        }

        /// <summary>
        /// Gets a value indicating whether the border should have shadow effect.
        /// </summary>
        /// <value>The shadow.</value>
        public bool Shadow
        {
            get { return this.shadow; }
        }

        /// <summary>
        /// Gets a value indicating whether the border should have frame effect.
        /// </summary>
        /// <value>The frame.</value>
        public bool Frame
        {
            get { return this.frame; }
        }

        /// <summary>
        /// Gets the spacing of the border.
        /// </summary>
        /// <value>The spacing.</value>
        public double Spacing
        {
            get { return this.spacing; }
        }

        /// <summary>
        /// Gets the style of the border.
        /// </summary>
        /// <value>The style.</value>
        public BorderStyle Style
        {
            get { return this.style; }
        }

        /// <summary>
        /// Gets the thickness of the border.
        /// </summary>
        /// <value>The thickness.</value>
        public double Thickness
        {
            get { return this.thickness; }
        }

        /// <summary>
        /// Determines whether the specified borders are equal.
        /// </summary>
        /// <returns>True if the borders are equal.</returns>
        public static bool operator ==(Border first, Border second)
        {
            if (object.ReferenceEquals(first, second))
            {
                return true;
            }

            if (object.ReferenceEquals(first, null) ||
                object.ReferenceEquals(second, null))
            {
                return false;
            }

            return first.Equals(second);
        }

        /// <summary>
        /// Determines whether the specified borders are different.
        /// </summary>
        /// <returns>True if the borders are different.</returns>
        public static bool operator !=(Border first, Border second)
        {
            return !(first == second);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// Returns true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            Border other = obj as Border;
            if (other == null)
            {
                return false;
            }

            // Thickness
            int thicknessComp = this.Thickness.CompareTo(other.Thickness);
            if (thicknessComp != 0)
            {
                return false;
            }

            // Style
            int styleComp = (int)this.Style - (int)other.Style;
            if (styleComp != 0)
            {
                return false;
            }

            // Color
            if (this.Color == null && other.Color == null)
            {
                return true;
            }
            else if ((this.Color != null && other.Color == null) || (this.Color == null && other.Color != null))
            {
                return false;
            }

            if (!this.Color.Equals(other.Color))
            {
                return false;
            }

            // Spacing
            if (this.Spacing != other.Spacing)
            {
                return false;
            }

            // Shadow
            if (this.Shadow != other.Shadow)
            {
                return false;
            }

            // Shadow
            if (this.Frame != other.Frame)
            {
                return false;
            }

            return true;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return ObjectExtensions.CombineHashCodes(this.Style.GetHashCode(),
                this.Color.GetHashCode(),
                this.Thickness.GetHashCode());
        }

        /// <summary>
        /// Compares a borders to another.
        /// </summary>
        /// <param name="other">The other.</param>
        /// <returns></returns>
        public int CompareTo(Border other)
        {
            if (other == null)
            {
                return 1;
            }

            // Thickness compare
            int thicknessComp = this.Thickness.CompareTo(other.Thickness);
            if (thicknessComp != 0)
            {
                return thicknessComp;
            }

            // Style compare
            int styleComp = (int)this.Style - (int)other.Style;
            if (styleComp != 0)
            {
                return styleComp;
            }

            // TODO: Make this work for theme colors
            if (!this.Color.IsFromTheme && !other.Color.IsFromTheme)
            {
                Color thisColor = this.Color.LocalValue;
                Color otherColor = other.Color.LocalValue;

                // Color first compare
                int colComp = (thisColor.R + thisColor.B + (2 * thisColor.G)) -
                              (otherColor.R + otherColor.B + (2 * otherColor.G));

                if (colComp != 0)
                {
                    return colComp;
                }

                // Color second compare
                colComp = (thisColor.B + (2 * thisColor.G)) -
                          (otherColor.B + (2 * otherColor.G));
                if (colComp != 0)
                {
                    return colComp;
                }

                // Color third compare
                colComp = thisColor.G - otherColor.G;
                if (colComp != 0)
                {
                    return colComp;
                }
            }

            return 0;
        }

        internal static Border Max(Border x, Border y)
        {
            if (x == null)
            {
                return y;
            }

            if (y == null)
            {
                return x;
            }

            return x.CompareTo(y) > 0 ? x : y;
        }
    }
}