﻿using System;
using System.Collections.Generic;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Media;
#else
using System.Windows;
using System.Windows.Media;
#endif

#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Defines a set of properties which can be applied to a <see cref="Run"/> element.
    /// </summary>
    public sealed class CharacterProperties : DocumentElementPropertiesBase, IPropertiesWithShading
    {
        private static Func<CharacterProperties, IStyleProperty>[] stylePropertyGetters;

        private StyleProperty<double?> fontSizeStyleProperty;
        private StyleProperty<ThemableFontFamily> fontFamilyStyleProperty;
        private StyleProperty<FontStyle?> fontStyleStyleProperty;
        private StyleProperty<FontWeight?> fontWeightStyleProperty;
        private StyleProperty<ThemableColor> foregroundColorStyleProperty;
        private StyleProperty<Color?> highlightColorStyleProperty;
        private StyleProperty<BaselineAlignment?> baselineAlignmentStyleProperty;
        private StyleProperty<bool?> strikethroughStyleProperty;
        private StyleProperty<ThemableColor> backgroundColorStyleProperty;
        private StyleProperty<ThemableColor> shadingPatternColorStyleProperty;
        private StyleProperty<ShadingPattern?> shadingPatternStyleProperty;
        private StyleProperty<ThemableColor> underlineColorStyleProperty;
        private StyleProperty<UnderlinePattern?> underlinePatternStyleProperty;
        private StyleProperty<FlowDirection?> flowDirectionStyleProperty;

        /// <summary>
        /// Initializes a new instance of the <see cref="CharacterProperties" /> class.
        /// </summary>
        /// <param name="style">The style.</param>
        internal CharacterProperties(Style style)
            : base(style)
        {
            this.InitProperties();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CharacterProperties" /> class.
        /// </summary>
        /// <param name="run">The owner run.</param>
        internal CharacterProperties(Run run)
            : base(run, false)
        {
            this.InitProperties();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CharacterProperties" /> class.
        /// </summary>
        internal CharacterProperties(DocumentElementBase ownerElement, bool suppressStylePropertyEvaluation)
            : base(ownerElement, suppressStylePropertyEvaluation)
        {
            this.InitProperties();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CharacterProperties" /> class.
        /// </summary>
        internal CharacterProperties(ListLevel listLevel, bool suppressStylePropertyEvaluation)
            : base(listLevel, suppressStylePropertyEvaluation)
        {
            this.InitProperties();
        }

        /// <summary>
        /// Gets style property used to get or set the font family.
        /// </summary>
        /// <value>The font family.</value>
        public IStyleProperty<ThemableFontFamily> FontFamily
        {
            get
            {
                return this.fontFamilyStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the size of the font. 
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The size of the font.</value>
        public IStyleProperty<double?> FontSize
        {
            get
            {
                return this.fontSizeStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the font style.
        /// </summary>
        /// <returns>The default value is <c>Normal</c>.</returns>
        public IStyleProperty<FontStyle?> FontStyle
        {
            get
            {
                return this.fontStyleStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the font weight.
        /// </summary>
        /// <returns>The default value is <c>Normal</c>.</returns>
        public IStyleProperty<FontWeight?> FontWeight
        {
            get
            {
                return this.fontWeightStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the foreground color.
        /// </summary>
        /// <returns>The default value is <c>black</c>.</returns>
        public IStyleProperty<ThemableColor> ForegroundColor
        {
            get
            {
                return this.foregroundColorStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the color of the highlight.
        /// </summary>
        /// <returns>The default value is <c>Transparent</c>.</returns>
        public IStyleProperty<Color?> HighlightColor
        {
            get
            {
                return this.highlightColorStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the baseline alignment.
        /// </summary>
        /// <returns>The default value is <c>Baseline</c>.</returns>
        public IStyleProperty<BaselineAlignment?> BaselineAlignment
        {
            get
            {
                return this.baselineAlignmentStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the strikethrough.
        /// </summary>
        /// <returns>The default value is <c>false</c>.</returns>
        public IStyleProperty<bool?> Strikethrough
        {
            get
            {
                return this.strikethroughStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the color of the background.
        /// </summary>
        /// <value>The color of the background.</value>
        public IStyleProperty<ThemableColor> BackgroundColor
        {
            get
            {
                return this.backgroundColorStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating the color for any foreground pattern used in this shading.
        /// </summary>
        /// <value>The color of the pattern.</value>
        public IStyleProperty<ThemableColor> ShadingPatternColor
        {
            get
            {
                return this.shadingPatternColorStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating the pattern which shall be used to lay the pattern color over the background color for this shading.
        /// </summary>
        /// <returns>The default value is <c>Clear</c></returns>
        public IStyleProperty<ShadingPattern?> ShadingPattern
        {
            get
            {
                return this.shadingPatternStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the color for the underlining.
        /// </summary>
        /// <value>The color.</value>
        public IStyleProperty<ThemableColor> UnderlineColor
        {
            get
            {
                return this.underlineColorStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating the pattern which shall be used to create the underlining applied beneath the text.
        /// </summary>
        /// <returns>The default value is <c>None</c></returns>
        public IStyleProperty<UnderlinePattern?> UnderlinePattern
        {
            get
            {
                return this.underlinePatternStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the flow direction.
        /// </summary>
        /// <returns>The default value is <c>LeftToRight</c></returns>
        public IStyleProperty<FlowDirection?> FlowDirection
        {
            get
            {
                return this.flowDirectionStyleProperty;
            }
        }

        /// <summary>
        /// Gets the document.
        /// </summary>
        /// <value>The document.</value>
        RadFlowDocument IPropertiesWithShading.Document
        {
            get
            {
                return this.Document;
            }
        }

        /// <summary>
        /// Enumerates the style properties.
        /// </summary>
        /// <returns>The style properties.</returns>
        protected override IEnumerable<IStyleProperty> EnumerateStyleProperties()
        {
            yield return this.FontFamily;
            yield return this.FontSize;
            yield return this.FontStyle;
            yield return this.FontWeight;
            yield return this.ForegroundColor;
            yield return this.HighlightColor;
            yield return this.BaselineAlignment;
            yield return this.Strikethrough;
            yield return this.BackgroundColor;
            yield return this.ShadingPatternColor;
            yield return this.ShadingPattern;
            yield return this.UnderlineColor;
            yield return this.UnderlinePattern;
            yield return this.FlowDirection;
        }

        /// <summary>
        /// Gets specific style property or null.
        /// </summary>
        /// <param name="propertyDefinition">The property definition.</param>
        /// <returns></returns>
        protected override IStyleProperty GetStylePropertyOverride(IStylePropertyDefinition propertyDefinition)
        {
            if (propertyDefinition != DocumentElementPropertiesBase.StyleIdPropertyDefinition &&
                propertyDefinition.StylePropertyType != StylePropertyType.Character)
            {
                return null;
            }

            return stylePropertyGetters[propertyDefinition.GlobalPropertyIndex](this);
        }

        private static void InitStylePropertyGetters()
        {
            if (stylePropertyGetters != null)
            {
                return;
            }

            stylePropertyGetters = new Func<CharacterProperties, IStyleProperty>[15];
            stylePropertyGetters[DocumentElementPropertiesBase.StyleIdPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.StyleIdProperty; };
            stylePropertyGetters[Run.FontSizePropertyDefinition.GlobalPropertyIndex] = prop => { return prop.fontSizeStyleProperty; };
            stylePropertyGetters[Run.FontFamilyPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.fontFamilyStyleProperty; };
            stylePropertyGetters[Run.FontStylePropertyDefinition.GlobalPropertyIndex] = prop => { return prop.fontStyleStyleProperty; };
            stylePropertyGetters[Run.FontWeightPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.fontWeightStyleProperty; };
            stylePropertyGetters[Run.ForegroundColorPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.foregroundColorStyleProperty; };
            stylePropertyGetters[Run.HighlightColorPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.highlightColorStyleProperty; };
            stylePropertyGetters[Run.BaselineAlignmentPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.baselineAlignmentStyleProperty; };
            stylePropertyGetters[Run.StrikethroughPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.strikethroughStyleProperty; };
            stylePropertyGetters[Run.BackgroundColorPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.backgroundColorStyleProperty; };
            stylePropertyGetters[Run.ShadingPatternColorPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.shadingPatternColorStyleProperty; };
            stylePropertyGetters[Run.ShadingPatternPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.shadingPatternStyleProperty; };
            stylePropertyGetters[Run.UnderlineColorPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.underlineColorStyleProperty; };
            stylePropertyGetters[Run.UnderlinePatternPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.underlinePatternStyleProperty; };
            stylePropertyGetters[Run.FlowDirectionPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.flowDirectionStyleProperty; };
        }

        /// <summary>
        /// Initialize the properties.
        /// </summary>
        private void InitProperties()
        {
            this.fontSizeStyleProperty = new StyleProperty<double?>(Run.FontSizePropertyDefinition, this);
            this.fontFamilyStyleProperty = new StyleProperty<ThemableFontFamily>(Run.FontFamilyPropertyDefinition, this);
            this.fontStyleStyleProperty = new StyleProperty<FontStyle?>(Run.FontStylePropertyDefinition, this);
            this.fontWeightStyleProperty = new StyleProperty<FontWeight?>(Run.FontWeightPropertyDefinition, this);
            this.foregroundColorStyleProperty = new StyleProperty<ThemableColor>(Run.ForegroundColorPropertyDefinition, this);
            this.highlightColorStyleProperty = new StyleProperty<Color?>(Run.HighlightColorPropertyDefinition, this);
            this.baselineAlignmentStyleProperty = new StyleProperty<BaselineAlignment?>(Run.BaselineAlignmentPropertyDefinition, this);
            this.strikethroughStyleProperty = new StyleProperty<bool?>(Run.StrikethroughPropertyDefinition, this);
            this.backgroundColorStyleProperty = new StyleProperty<ThemableColor>(Run.BackgroundColorPropertyDefinition, this);
            this.shadingPatternColorStyleProperty = new StyleProperty<ThemableColor>(Run.ShadingPatternColorPropertyDefinition, this);
            this.shadingPatternStyleProperty = new StyleProperty<ShadingPattern?>(Run.ShadingPatternPropertyDefinition, this);
            this.underlineColorStyleProperty = new StyleProperty<ThemableColor>(Run.UnderlineColorPropertyDefinition, this);
            this.underlinePatternStyleProperty = new StyleProperty<UnderlinePattern?>(Run.UnderlinePatternPropertyDefinition, this);
            this.flowDirectionStyleProperty = new StyleProperty<FlowDirection?>(Run.FlowDirectionPropertyDefinition, this);

            InitStylePropertyGetters();
        }
    }
}