﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Defines common set of properties which can be applied to document element.
    /// </summary>
    public abstract class DocumentElementPropertiesBase : IElementWithStyle
    {
        /// <summary>
        /// Gets the style id property definition.
        /// </summary>
        protected static readonly StylePropertyDefinition<string> StyleIdPropertyDefinition =
            new StylePropertyDefinition<string>("StyleID", string.Empty, StylePropertyType.DocumentElement);

        private readonly bool suppressStylePropertyEvaluation;
        private Style ownerStyle;
        private DocumentElementBase ownerDocumentElement;
        private ListLevel ownerListLevel;
        private LocalProperty<string> styleIdProperty;

        /// <summary>
        /// Initializes a new instance of the <see cref="DocumentElementPropertiesBase" /> class.
        /// </summary>
        /// <param name="style">The style.</param>
        internal DocumentElementPropertiesBase(Style style)
        {
            this.SetOwnerStyle(style);
            this.InitProperties();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="DocumentElementPropertiesBase" /> class.
        /// </summary>
        /// <param name="documentElement">The document element.</param>
        /// <param name="suppressStylePropertyEvaluation">The suppress style property evaluation.</param>
        internal DocumentElementPropertiesBase(DocumentElementBase documentElement, bool suppressStylePropertyEvaluation)
        {
            this.suppressStylePropertyEvaluation = suppressStylePropertyEvaluation;
            this.SetOwnerDocumentElement(documentElement);
            this.InitProperties();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="DocumentElementPropertiesBase" /> class.
        /// </summary>
        /// <param name="listLevel">The list level.</param>
        /// <param name="suppressStylePropertyEvaluation">The suppress style property evaluation.</param>
        internal DocumentElementPropertiesBase(ListLevel listLevel, bool suppressStylePropertyEvaluation)
        {
            this.suppressStylePropertyEvaluation = suppressStylePropertyEvaluation;
            this.SetOwnerListLevel(listLevel);
            this.InitProperties();
        }

        /// <summary>
        /// Gets the style properties.
        /// </summary>
        /// <value>The style properties.</value>
        public IEnumerable<IStyleProperty> StyleProperties
        {
            get
            {
                foreach (IStyleProperty styleProperty in this.EnumerateStyleProperties())
                {
                    yield return styleProperty;
                }
                yield return this.styleIdProperty;
            }
        }

        /// <summary>
        /// Gets or sets the name of the style.
        /// </summary>
        /// <value>The name of the style.</value>
        public string StyleId
        {
            get { return this.styleIdProperty.GetActualValue(); }
            set { this.styleIdProperty.LocalValue = value; }
        }

        internal DocumentElementBase OwnerDocumentElement
        {
            get
            {
                return this.ownerDocumentElement;
            }
        }

        /// <summary>
        /// Gets value indicating whether the style property evaluation 
        /// should use styles and parents to calculate the property value.
        /// </summary>
        /// <value>True if only local or default values are used.</value>
        internal bool SuppressStylePropertyEvaluation
        {
            get
            {
                return this.suppressStylePropertyEvaluation;
            }
        }

        internal Style OwnerStyle
        {
            get
            {
                return this.ownerStyle;
            }
        }

        internal ListLevel OwnerListLevel
        {
            get
            {
                return this.ownerListLevel;
            }
        }

        internal RadFlowDocument Document
        {
            get
            {
                if (this.OwnerDocumentElement != null)
                {
                    return this.OwnerDocumentElement.Document;
                }

                if (this.OwnerStyle != null)
                {
                    return this.OwnerStyle.Document;
                }

                if (this.OwnerListLevel != null)
                {
                    return this.OwnerListLevel.Document;
                }

                return null;
            }
        }

        /// <summary>
        /// Gets the style id property.
        /// </summary>
        /// <value>The style id property.</value>
        protected IStyleProperty StyleIdProperty
        {
            get
            {
                return this.styleIdProperty;
            }
        }

        /// <summary>
        /// Gets specific style property or null.
        /// </summary>
        /// <param name="propertyDefinition">The property definition.</param>
        /// <returns></returns>
        public IStyleProperty GetStyleProperty(IStylePropertyDefinition propertyDefinition)
        {
            return this.GetStylePropertyOverride(propertyDefinition);
        }

        /// <summary>
        /// Gets specific style property or null.
        /// </summary>
        /// <param name="propertyName">Name of the property.</param>
        /// <returns></returns>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public IStyleProperty GetStyleProperty(string propertyName)
        {
            return this.StyleProperties.FirstOrDefault(p => p.PropertyDefinition.PropertyName == propertyName);
        }

        /// <summary>
        /// Copies the properties from.
        /// </summary>
        /// <param name="fromProperties">From properties.</param>
        public virtual void CopyPropertiesFrom(DocumentElementPropertiesBase fromProperties)
        {
            if (fromProperties == null || fromProperties.GetType() != this.GetType())
            {
                return;
            }

            foreach (IStyleProperty property in fromProperties.StyleProperties)
            {
                if (property.HasLocalValue)
                {
                    IStyleProperty toProperty = this.GetStyleProperty(property.PropertyDefinition);
                    toProperty.SetValueAsObject(property.GetLocalValueAsObject());
                }
            }
        }

        /// <summary>
        /// Clears all local property values.
        /// </summary>
        public void ClearLocalValues()
        {
            foreach (IStyleProperty property in this.StyleProperties)
            {
                if (property.HasLocalValue)
                {
                    property.ClearValue();
                }
            }
        }

        /// <summary>
        /// Determines whether there are local property values.
        /// </summary>
        /// <returns>True if there are local property values.</returns>
        public bool HasLocalValues()
        {
            return this.StyleProperties.Any(p => p.HasLocalValue);
        }

        /// <summary>
        /// Enumerates the style properties.
        /// </summary>
        /// <returns>The style properties.</returns>
        protected abstract IEnumerable<IStyleProperty> EnumerateStyleProperties();

        /// <summary>
        /// Gets specific style property or null.
        /// </summary>
        /// <param name="propertyDefinition">The property definition.</param>
        /// <returns></returns>
        protected abstract IStyleProperty GetStylePropertyOverride(IStylePropertyDefinition propertyDefinition);

        private void InitProperties()
        {
            this.styleIdProperty = new LocalProperty<string>(DocumentElementPropertiesBase.StyleIdPropertyDefinition, this);
        }

        private void SetOwnerStyle(Style style)
        {
            if (this.ownerStyle != null || this.ownerDocumentElement != null || this.ownerListLevel != null)
            {
                throw new InvalidOperationException("Owner cannot be set more than once.");
            }

            Guard.ThrowExceptionIfNull(style, "style");
            this.ownerStyle = style;
        }

        private void SetOwnerDocumentElement(DocumentElementBase documentElement)
        {
            if (this.ownerStyle != null || this.ownerDocumentElement != null || this.ownerListLevel != null)
            {
                throw new InvalidOperationException("Owner cannot be set more than once.");
            }

            Guard.ThrowExceptionIfNull(documentElement, "document element");
            this.ownerDocumentElement = documentElement;
        }

        private void SetOwnerListLevel(ListLevel listLevel)
        {
            if (this.ownerStyle != null || this.ownerDocumentElement != null || this.ownerListLevel != null)
            {
                throw new InvalidOperationException("Owner cannot be set more than once.");
            }

            Guard.ThrowExceptionIfNull(listLevel, "list level");
            this.ownerListLevel = listLevel;
        }
    }
}
