﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Holds settings that can be applied to <see cref="RadFlowDocument"/>.
    /// </summary>
    public sealed class DocumentProperties : DocumentElementPropertiesBase
    {
        private static Func<DocumentProperties, IStyleProperty>[] stylePropertyGetters;

        private LocalProperty<bool?> hasDifferentEvenOddPageHeadersFootersProperty;
        private LocalProperty<DocumentViewType?> viewTypeProperty;
        private LocalProperty<double?> defaultTabStopWidthProperty;

        internal DocumentProperties(RadFlowDocument document)
            : base(document, false)
        {
            this.InitProperties();
        }

        /// <summary>
        /// Gets the style property used to get or set a value indicating whether pages in this document will have different headers and footers for even and odd pages.
        /// </summary>
        public IStyleProperty<bool?> HasDifferentEvenOddPageHeadersFooters
        {
            get
            {
                return this.hasDifferentEvenOddPageHeadersFootersProperty;
            }
        }

        /// <summary>
        /// Gets the style property used to get or set a value indicating the view type of the document.
        /// </summary>
        public IStyleProperty<DocumentViewType?> ViewType
        {
            get
            {
                return this.viewTypeProperty;
            }
        }

        /// <summary>
        /// Gets the style property used to determine the default width value which shall be used when a tab stop should be generated.
        /// <para>The value is in device independent pixels (1/96 inch).</para>
        /// </summary>
        /// <value>The default width of the tab stop.</value>
        public IStyleProperty<double?> DefaultTabStopWidth
        {
            get
            {
                return this.defaultTabStopWidthProperty;
            }
        }

        /// <summary>
        /// Enumerates the style properties.
        /// </summary>
        /// <returns>The style properties.</returns>
        protected override IEnumerable<IStyleProperty> EnumerateStyleProperties()
        {
            yield return this.HasDifferentEvenOddPageHeadersFooters;
            yield return this.ViewType;
            yield return this.DefaultTabStopWidth;
        }

        /// <summary>
        /// Gets specific style property or null.
        /// </summary>
        /// <param name="propertyDefinition">The property definition.</param>
        /// <returns></returns>
        protected override IStyleProperty GetStylePropertyOverride(IStylePropertyDefinition propertyDefinition)
        {
            if (propertyDefinition != DocumentElementPropertiesBase.StyleIdPropertyDefinition &&
                propertyDefinition.StylePropertyType != StylePropertyType.Document)
            {
                return null;
            }

            return stylePropertyGetters[propertyDefinition.GlobalPropertyIndex](this);
        }

        private static void InitStylePropertyGetters()
        {
            if (stylePropertyGetters != null)
            {
                return;
            }

            stylePropertyGetters = new Func<DocumentProperties, IStyleProperty>[4];
            stylePropertyGetters[DocumentElementPropertiesBase.StyleIdPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.StyleIdProperty; };
            stylePropertyGetters[RadFlowDocument.HasDifferentEvenOddPageHeadersFootersPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.hasDifferentEvenOddPageHeadersFootersProperty; };
            stylePropertyGetters[RadFlowDocument.ViewTypePropertyDefinition.GlobalPropertyIndex] = prop => { return prop.viewTypeProperty; };
            stylePropertyGetters[RadFlowDocument.DefaultTabStopWidthPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.defaultTabStopWidthProperty; };
        }

        private void InitProperties()
        {
            this.hasDifferentEvenOddPageHeadersFootersProperty = new LocalProperty<bool?>(RadFlowDocument.HasDifferentEvenOddPageHeadersFootersPropertyDefinition, this);
            this.viewTypeProperty = new LocalProperty<DocumentViewType?>(RadFlowDocument.ViewTypePropertyDefinition, this);
            this.defaultTabStopWidthProperty = new LocalProperty<double?>(RadFlowDocument.DefaultTabStopWidthPropertyDefinition, this);

            InitStylePropertyGetters();
        }
    }
}