﻿#if (WPF || SILVERLIGHT) && !NETCORE
using System.Windows.Media;
using Telerik.Windows.Documents.Spreadsheet.Model;
#endif
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Defines set of borders which can be applied to paragraph element. This class is immutable.
    /// </summary>
    public class ParagraphBorders
    {
        private readonly Border top;
        private readonly Border bottom;
        private readonly Border left;
        private readonly Border right;
        private readonly Border between;

        /// <summary>
        /// Initializes a new instance of the <see cref="ParagraphBorders" /> class.
        /// </summary>
        public ParagraphBorders()
            : this(Border.DefaultBorder)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ParagraphBorders" /> class.
        /// </summary>
        /// <param name="all">All.</param>
        public ParagraphBorders(Border all)
        {
            Guard.ThrowExceptionIfNull(all, "all");

            this.left = all;
            this.top = all;
            this.right = all;
            this.bottom = all;
            this.between = all;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ParagraphBorders" /> class.
        /// </summary>
        /// <param name="leftBorder">The left border.</param>
        /// <param name="topBorder">The top border.</param>
        /// <param name="rightBorder">The right border.</param>
        /// <param name="bottomBorder">The bottom border.</param>
        public ParagraphBorders(Border leftBorder, Border topBorder, Border rightBorder, Border bottomBorder)
            : this(leftBorder, topBorder, rightBorder, bottomBorder, null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ParagraphBorders" /> class.
        /// </summary>
        /// <param name="leftBorder">The left border.</param>
        /// <param name="topBorder">The top border.</param>
        /// <param name="rightBorder">The right border.</param>
        /// <param name="bottomBorder">The bottom border.</param>
        /// <param name="between">The between.</param>
        public ParagraphBorders(Border leftBorder, Border topBorder, Border rightBorder, Border bottomBorder, Border between)
        {
            this.left = leftBorder ?? Border.DefaultBorder;
            this.top = topBorder ?? Border.DefaultBorder;
            this.right = rightBorder ?? Border.DefaultBorder;
            this.bottom = bottomBorder ?? Border.DefaultBorder;
            this.between = between ?? Border.DefaultBorder;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ParagraphBorders" /> class. Creates copy of the the source paragraph borders and replaces only the specified borders.
        /// </summary>
        /// <param name="source">The source.</param>
        /// <param name="leftBorder">The left border.</param>
        /// <param name="topBorder">The top border.</param>
        /// <param name="rightBorder">The right border.</param>
        /// <param name="bottomBorder">The bottom border.</param>
        /// <param name="between">The between.</param>
        public ParagraphBorders(ParagraphBorders source, Border leftBorder = null, Border topBorder = null, Border rightBorder = null, Border bottomBorder = null, Border between = null)
        {
            this.left = leftBorder ?? source.Left;
            this.top = topBorder ?? source.Top;
            this.right = rightBorder ?? source.Right;
            this.bottom = bottomBorder ?? source.Bottom;
            this.between = between ?? source.Between;
        }

        /// <summary>
        /// Gets the top border.
        /// </summary>
        /// <value>The top border.</value>
        public Border Top
        {
            get
            {
                return this.top;
            }
        }

        /// <summary>
        /// Gets the bottom border.
        /// </summary>
        /// <value>The bottom border.</value>
        public Border Bottom
        {
            get
            {
                return this.bottom;
            }
        }

        /// <summary>
        /// Gets the left border.
        /// </summary>
        /// <value>The left border.</value>
        public Border Left
        {
            get
            {
                return this.left;
            }
        }

        /// <summary>
        /// Gets the right border.
        /// </summary>
        /// <value>The right border.</value>
        public Border Right
        {
            get
            {
                return this.right;
            }
        }

        /// <summary>
        /// Gets the border, which should be applied between paragraphs, which have the same set of paragraph borders.
        /// </summary>
        /// <value>The between border.</value>
        public Border Between
        {
            get
            {
                return this.between;
            }
        }

        /// <summary>
        /// Determines whether the specified paragraph borders are equal.
        /// </summary>
        /// <returns>True if the paragraph borders are equal.</returns>
        public static bool operator ==(ParagraphBorders a, ParagraphBorders b)
        {
            // If both are null, or both are same instance, return true.
            if (object.ReferenceEquals(a, b))
            {
                return true;
            }

            // If one is null, but not both, return false.
            if (((object)a == null) || ((object)b == null))
            {
                return false;
            }

            return a.Equals(b);
        }

        /// <summary>
        /// Determines whether the specified paragraph borders are different.
        /// </summary>
        /// <returns>True if the paragraph borders are different.</returns>
        public static bool operator !=(ParagraphBorders a, ParagraphBorders b)
        {
            return !(a == b);
        }

        /// <summary>
        /// Sets the left.
        /// </summary>
        /// <param name="left">The left.</param>
        /// <returns>New instance of paragraph borders.</returns>
        public ParagraphBorders SetLeft(Border left)
        {
            return new ParagraphBorders(left, this.Top, this.Right, this.Bottom, this.Between);
        }

        /// <summary>
        /// Sets the top.
        /// </summary>
        /// <param name="top">The top.</param>
        /// <returns>New instance of paragraph borders.</returns>
        public ParagraphBorders SetTop(Border top)
        {
            return new ParagraphBorders(this.Left, top, this.Right, this.Bottom, this.Between);
        }

        /// <summary>
        /// Sets the right.
        /// </summary>
        /// <param name="right">The right.</param>
        /// <returns>New instance of paragraph borders.</returns>
        public ParagraphBorders SetRight(Border right)
        {
            return new ParagraphBorders(this.Left, this.Top, right, this.Bottom, this.Between);
        }

        /// <summary>
        /// Sets the bottom.
        /// </summary>
        /// <param name="bottom">The bottom.</param>
        /// <returns>New instance of paragraph borders.</returns>
        public ParagraphBorders SetBottom(Border bottom)
        {
            return new ParagraphBorders(this.Left, this.Top, this.Right, bottom, this.Between);
        }

        /// <summary>
        /// Sets the between.
        /// </summary>
        /// <param name="between">The between.</param>
        /// <returns>New instance of paragraph borders.</returns>
        public ParagraphBorders SetBetween(Border between)
        {
            return new ParagraphBorders(this.Left, this.Top, this.Right, this.Bottom, between);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// Returns true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            ParagraphBorders other = obj as ParagraphBorders;
            if (other == null)
            {
                return false;
            }

            return this.Left.Equals(other.Left) &&
                this.Top.Equals(other.Top) &&
                this.Right.Equals(other.Right) &&
                this.Bottom.Equals(other.Bottom) &&
                this.Between.Equals(other.Between);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return ObjectExtensions.CombineHashCodes(this.Left.GetHashCode(),
                this.Top.GetHashCode(),
                this.Right.GetHashCode(),
                this.Bottom.GetHashCode(),
                this.Between.GetHashCode());
        }
    }
}
