﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;
using Telerik.Windows.Documents.Flow.Model.Lists;

#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Defines a set of properties which can be applied to a <see cref="Paragraph"/> element.
    /// </summary>
    public sealed class ParagraphProperties : DocumentElementPropertiesBase, IPropertiesWithShading
    {
        private static Func<ParagraphProperties, IStyleProperty>[] stylePropertyGetters;

        private LocalProperty<bool?> allowOverflowPunctuationStyleProperty;
        private StyleProperty<FlowDirection?> flowDirectionStyleProperty;
        private StyleProperty<double?> lineSpacingStyleProperty;
        private LineSpacingTypeStyleProperty lineSpacingTypeStyleProperty;
        private StyleProperty<bool?> keepOnOnePageStyleProperty;
        private StyleProperty<bool?> keepWithNextParagraphStyleProperty;
        private StyleProperty<int?> listIdStyleProperty;
        private StyleProperty<int?> listLevelStyleProperty;
        private StyleProperty<OutlineLevel?> outlineLevelStyleProperty;
        private StyleProperty<bool?> applyEastAsianLineBreakingRulesStyleProperty;
        private StyleProperty<bool?> pageBreakBeforeStyleProperty;
        private StyleProperty<bool?> contextualSpacingStyleProperty;
        private StyleProperty<bool?> mirrorIndentsStyleProperty;
        private StyleProperty<Alignment?> textAlignmentStyleProperty;
        private StyleProperty<bool?> automaticSpacingBeforeStyleProperty;
        private StyleProperty<bool?> automaticSpacingAfterStyleProperty;
        private StyleProperty<double?> spacingAfterStyleProperty;
        private StyleProperty<double?> spacingBeforeStyleProperty;
        private StyleProperty<ParagraphBorders> bordersStyleProperty;
        private StyleProperty<ThemableColor> backgroundColorStyleProperty;
        private StyleProperty<ThemableColor> shadingPatternColorStyleProperty;
        private StyleProperty<ShadingPattern?> shadingPatternStyleProperty;
        private StyleProperty<double?> firstLineIndentStyleProperty;
        private StyleProperty<double?> hangingIndentStyleProperty;
        private StyleProperty<double?> leftIndentStyleProperty;
        private StyleProperty<double?> rightIndentStyleProperty;
        private TabStopCollectionStyleProperty tabStopsStyleProperty;

        /// <summary>
        /// Initializes a new instance of the <see cref="ParagraphProperties" /> class.
        /// </summary>
        /// <param name="style">The style.</param>
        internal ParagraphProperties(Style style)
            : base(style)
        {
            this.InitProperties();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ParagraphProperties" /> class.
        /// </summary>
        /// <param name="paragraph">The owner paragraph.</param>
        internal ParagraphProperties(Paragraph paragraph)
            : base(paragraph, false)
        {
            this.InitProperties();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ParagraphProperties" /> class.
        /// </summary>
        internal ParagraphProperties(RadFlowDocument document, bool suppressStylePropertyEvaluation)
            : base(document, suppressStylePropertyEvaluation)
        {
            this.InitProperties();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ParagraphProperties" /> class.
        /// </summary>
        internal ParagraphProperties(ListLevel listLevel, bool suppressStylePropertyEvaluation)
            : base(listLevel, suppressStylePropertyEvaluation)
        {
            this.InitProperties();
        }

        /// <summary>
        /// Gets style property used to get or set the flow direction.
        /// </summary>
        /// <returns>The default value is <c>LeftToRight</c></returns>
        public IStyleProperty<FlowDirection?> FlowDirection
        {
            get
            {
                return this.flowDirectionStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the paragraph text alignment.
        /// </summary>
        /// <returns>The default value is <c>Left</c></returns>
        public IStyleProperty<Alignment?> TextAlignment
        {
            get
            {
                return this.textAlignmentStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the spacing that should be added above the first line in a paragraph.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The spacing before value.</value>
        public IStyleProperty<double?> SpacingBefore
        {
            get
            {
                return this.spacingBeforeStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating whether a consumer shall automatically determine the spacing 
        /// before the paragraph based on its contents the automatic spacing after.
        /// </summary>
        /// <value>The automatic spacing after.</value>
        public IStyleProperty<bool?> AutomaticSpacingBefore
        {
            get
            {
                return this.automaticSpacingBeforeStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the spacing that should be added after the last line in a paragraph.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The spacing after value.</value>
        public IStyleProperty<double?> SpacingAfter
        {
            get
            {
                return this.spacingAfterStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating whether a consumer shall automatically determine the spacing 
        /// after the paragraph based on its contents the automatic spacing after.
        /// </summary>
        /// <value>The automatic spacing after.</value>
        public IStyleProperty<bool?> AutomaticSpacingAfter
        {
            get
            {
                return this.automaticSpacingAfterStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the value indicating the amount of vertical spacing between lines of text within a paragraph.
        /// <para>If <see cref="LineSpacingType" /> is Auto then the line height value times the value of the property, 
        /// otherwise the value is in device independent pixels (1/96 inch).</para>
        /// </summary>
        /// <value>The line spacing value.</value>
        public IStyleProperty<double?> LineSpacing
        {
            get
            {
                return this.lineSpacingStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the value indicating how the spacing between lines is calculated.
        /// </summary>
        /// <value>The type line spacing rule.</value>
        public IStyleProperty<HeightType?> LineSpacingType
        {
            get
            {
                return this.lineSpacingTypeStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating if this paragraph should be rendered on one page when the document is shown in page view mode. 
        /// </summary>
        /// <returns>The default value is <c>false</c></returns>
        public IStyleProperty<bool?> KeepOnOnePage
        {
            get
            {
                return this.keepOnOnePageStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating if this paragraph should be rendered at least partly on the one page with 
        /// the following paragraph when this is possible and when the document is shown in page view mode. 
        /// </summary>
        /// <returns>The default value is <c>false</c></returns>
        public IStyleProperty<bool?> KeepWithNextParagraph
        {
            get
            {
                return this.keepWithNextParagraphStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating the outline level. The outline level defines the level of this paragraph in TOC field. 
        /// </summary>
        /// <returns>The default value is <c>Level9</c> - no level.</returns>
        public IStyleProperty<OutlineLevel?> OutlineLevel
        {
            get
            {
                return this.outlineLevelStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating whether East-Asian line breaking rules are applied to this paragraph. 
        /// </summary>
        /// <returns>The default value is <c>true</c>.</returns>
        public IStyleProperty<bool?> ApplyEastAsianLineBreakingRules
        {
            get
            {
                return this.applyEastAsianLineBreakingRulesStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating if this paragraph should be rendered on new page when the document is shown in page view mode. 
        /// </summary>
        /// <returns>The default value is <c>false</c></returns>
        public IStyleProperty<bool?> PageBreakBefore
        {
            get
            {
                return this.pageBreakBeforeStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the paragraph borders.
        /// </summary>
        /// <value>The paragraph borders.</value>
        public IStyleProperty<ParagraphBorders> Borders
        {
            get
            {
                return this.bordersStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating whether spacing before/after are ignored 
        /// when preceding/following paragraph has same paragraph style.
        /// </summary>
        /// <returns>The default value is <c>false</c>.</returns>
        public IStyleProperty<bool?> ContextualSpacing
        {
            get
            {
                return this.contextualSpacingStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating whether left and right indents should be swapped on odd pages. 
        /// </summary>
        /// <returns>The default value is <c>false</c>.</returns>
        public IStyleProperty<bool?> MirrorIndents
        {
            get
            {
                return this.mirrorIndentsStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the color of the background.
        /// </summary>
        /// <value>The color of the background.</value>
        public IStyleProperty<ThemableColor> BackgroundColor
        {
            get
            {
                return this.backgroundColorStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating the color for any foreground pattern used in this shading.
        /// </summary>
        /// <value>The color of the pattern.</value>
        public IStyleProperty<ThemableColor> ShadingPatternColor
        {
            get
            {
                return this.shadingPatternColorStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating the pattern which shall be used to lay the pattern color over the background color for this shading.
        /// </summary>
        /// <returns>The default value is <c>Clear</c></returns>
        public IStyleProperty<ShadingPattern?> ShadingPattern
        {
            get
            {
                return this.shadingPatternStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating the additional indentation which shall be applied to the first line of the paragraph.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The first line indent.</value>
        public IStyleProperty<double?> FirstLineIndent
        {
            get
            {
                return this.firstLineIndentStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value which shall be removed from the indentation of the first line of the paragraph, 
        /// by moving the indentation on the first line back towards the beginning of the direction of text flow.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The hanging indent.</value>
        public IStyleProperty<double?> HangingIndent
        {
            get
            {
                return this.hangingIndentStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating the indentation which shall be applied to the left side of the whole paragraph.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The left indent.</value>
        public IStyleProperty<double?> LeftIndent
        {
            get
            {
                return this.leftIndentStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating the indentation which shall be applied to the right side of the whole paragraph.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The left indent.</value>
        public IStyleProperty<double?> RightIndent
        {
            get
            {
                return this.rightIndentStyleProperty;
            }
        }

        /// <summary>
        /// Gets local style property used to get or set a value indicating whether the last punctuation character on a line can overflow in paragraph margin/indent. 
        /// This property cannot be derived from a style.
        /// </summary>
        /// <returns>The default is <c>true</c>.</returns>
        public IStyleProperty<bool?> AllowOverflowPunctuation
        {
            get
            {
                return this.allowOverflowPunctuationStyleProperty;
            }
        }

        /// <summary>
        /// Gets the style property used to get or set the tab stops in the paragraph.
        /// </summary>
        /// <value>The tab stops.</value>
        public IStyleProperty<TabStopCollection> TabStops
        {
            get
            {
                return this.tabStopsStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the formatting properties of the glyph used to represent this paragraph.
        /// </summary>
        /// <value>The paragraph marker properties.</value>
        public CharacterProperties ParagraphMarkerProperties { get; private set; }

        /// <summary>
        /// Gets the document.
        /// </summary>
        /// <value>The document.</value>
        RadFlowDocument IPropertiesWithShading.Document
        {
            get
            {
                return this.Document;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating that the paragraph is referencing a list.
        /// </summary>
        /// <returns>The default value is <c>-1</c> - list style is not referenced.</returns>
        public IStyleProperty<int?> ListId
        {
            get
            {
                return this.listIdStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating that the paragraph is referencing list level.
        /// </summary>
        /// <returns>The default value is <c>-1</c> - list level is not referenced.</returns>
        public IStyleProperty<int?> ListLevel
        {
            get
            {
                return this.listLevelStyleProperty;
            }
        }

        /// <summary>
        /// Copy properties from specified source.
        /// </summary>
        /// <param name="fromProperties">The source from which to copy the properties.</param>
        public override void CopyPropertiesFrom(DocumentElementPropertiesBase fromProperties)
        {
            base.CopyPropertiesFrom(fromProperties);

            if (fromProperties == null || this.ParagraphMarkerProperties == null || fromProperties.GetType() != this.GetType())
            {
                return;
            }

            this.ParagraphMarkerProperties.CopyPropertiesFrom(((ParagraphProperties)fromProperties).ParagraphMarkerProperties);
        }

        /// <summary>
        /// Enumerates the style properties.
        /// </summary>
        /// <returns>The style properties.</returns>
        protected override IEnumerable<IStyleProperty> EnumerateStyleProperties()
        {
            yield return this.FlowDirection;
            yield return this.TextAlignment;
            yield return this.SpacingBefore;
            yield return this.AutomaticSpacingBefore;
            yield return this.SpacingAfter;
            yield return this.AutomaticSpacingAfter;
            yield return this.LineSpacing;
            yield return this.LineSpacingType;
            yield return this.KeepOnOnePage;
            yield return this.KeepWithNextParagraph;
            yield return this.OutlineLevel;
            yield return this.ApplyEastAsianLineBreakingRules;
            yield return this.PageBreakBefore;
            yield return this.Borders;
            yield return this.ContextualSpacing;
            yield return this.MirrorIndents;
            yield return this.BackgroundColor;
            yield return this.ShadingPatternColor;
            yield return this.ShadingPattern;
            yield return this.FirstLineIndent;
            yield return this.HangingIndent;
            yield return this.LeftIndent;
            yield return this.RightIndent;
            yield return this.AllowOverflowPunctuation;
            yield return this.TabStops;
            yield return this.ListLevel;
            yield return this.ListId;
        }

        /// <summary>
        /// Gets specific style property or null.
        /// </summary>
        /// <param name="propertyDefinition">The property definition.</param>
        /// <returns></returns>
        protected override IStyleProperty GetStylePropertyOverride(IStylePropertyDefinition propertyDefinition)
        {
            if (propertyDefinition != DocumentElementPropertiesBase.StyleIdPropertyDefinition &&
                propertyDefinition.StylePropertyType != StylePropertyType.Paragraph)
            {
                return null;
            }

            return stylePropertyGetters[propertyDefinition.GlobalPropertyIndex](this);
        }

        private static void InitStylePropertyGetters()
        {
            if (stylePropertyGetters != null)
            {
                return;
            }

            stylePropertyGetters = new Func<ParagraphProperties, IStyleProperty>[28];
            stylePropertyGetters[DocumentElementPropertiesBase.StyleIdPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.StyleIdProperty; };
            stylePropertyGetters[Paragraph.AllowOverflowPunctuationPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.allowOverflowPunctuationStyleProperty; };
            stylePropertyGetters[Paragraph.TextAlignmentPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.textAlignmentStyleProperty; };
            stylePropertyGetters[Paragraph.FlowDirectionPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.flowDirectionStyleProperty; };
            stylePropertyGetters[Paragraph.LineSpacingPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.lineSpacingStyleProperty; };
            stylePropertyGetters[Paragraph.LineSpacingTypePropertyDefinition.GlobalPropertyIndex] = prop => { return prop.lineSpacingTypeStyleProperty; };
            stylePropertyGetters[Paragraph.KeepOnOnePagePropertyDefinition.GlobalPropertyIndex] = prop => { return prop.keepOnOnePageStyleProperty; };
            stylePropertyGetters[Paragraph.KeepWithNextParagraphPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.keepWithNextParagraphStyleProperty; };
            stylePropertyGetters[Paragraph.ListIdPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.listIdStyleProperty; };
            stylePropertyGetters[Paragraph.ListLevelPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.listLevelStyleProperty; };
            stylePropertyGetters[Paragraph.OutlineLevelPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.outlineLevelStyleProperty; };
            stylePropertyGetters[Paragraph.ApplyEastAsianLineBreakingRulesPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.applyEastAsianLineBreakingRulesStyleProperty; };
            stylePropertyGetters[Paragraph.PageBreakBeforePropertyDefinition.GlobalPropertyIndex] = prop => { return prop.pageBreakBeforeStyleProperty; };
            stylePropertyGetters[Paragraph.ContextualSpacingPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.contextualSpacingStyleProperty; };
            stylePropertyGetters[Paragraph.MirrorIndentsPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.mirrorIndentsStyleProperty; };
            stylePropertyGetters[Paragraph.SpacingAfterPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.spacingAfterStyleProperty; };
            stylePropertyGetters[Paragraph.SpacingBeforePropertyDefinition.GlobalPropertyIndex] = prop => { return prop.spacingBeforeStyleProperty; };
            stylePropertyGetters[Paragraph.AutomaticSpacingBeforePropertyDefinition.GlobalPropertyIndex] = prop => { return prop.automaticSpacingBeforeStyleProperty; };
            stylePropertyGetters[Paragraph.AutomaticSpacingAfterPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.automaticSpacingAfterStyleProperty; };
            stylePropertyGetters[Paragraph.BordersPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.bordersStyleProperty; };
            stylePropertyGetters[Paragraph.BackgroundColorPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.backgroundColorStyleProperty; };
            stylePropertyGetters[Paragraph.ShadingPatternColorPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.shadingPatternColorStyleProperty; };
            stylePropertyGetters[Paragraph.ShadingPatternPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.shadingPatternStyleProperty; };
            stylePropertyGetters[Paragraph.FirstLineIndentPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.firstLineIndentStyleProperty; };
            stylePropertyGetters[Paragraph.HangingIndentPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.hangingIndentStyleProperty; };
            stylePropertyGetters[Paragraph.LeftIndentPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.leftIndentStyleProperty; };
            stylePropertyGetters[Paragraph.RightIndentPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.rightIndentStyleProperty; };
            stylePropertyGetters[Paragraph.TabStopsPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.tabStopsStyleProperty; };
        }

        /// <summary>
        /// Initialize the properties.
        /// </summary>
        private void InitProperties()
        {
            this.allowOverflowPunctuationStyleProperty = new LocalProperty<bool?>(Paragraph.AllowOverflowPunctuationPropertyDefinition, this);
            this.textAlignmentStyleProperty = new StyleProperty<Alignment?>(Paragraph.TextAlignmentPropertyDefinition, this);
            this.flowDirectionStyleProperty = new StyleProperty<FlowDirection?>(Paragraph.FlowDirectionPropertyDefinition, this);
            this.lineSpacingStyleProperty = new StyleProperty<double?>(Paragraph.LineSpacingPropertyDefinition, this);
            this.lineSpacingTypeStyleProperty = new LineSpacingTypeStyleProperty(Paragraph.LineSpacingTypePropertyDefinition, this);
            this.keepOnOnePageStyleProperty = new StyleProperty<bool?>(Paragraph.KeepOnOnePagePropertyDefinition, this);
            this.keepWithNextParagraphStyleProperty = new StyleProperty<bool?>(Paragraph.KeepWithNextParagraphPropertyDefinition, this);
            this.listIdStyleProperty = new StyleProperty<int?>(Paragraph.ListIdPropertyDefinition, this);
            this.listLevelStyleProperty = new StyleProperty<int?>(Paragraph.ListLevelPropertyDefinition, this);
            this.outlineLevelStyleProperty = new StyleProperty<OutlineLevel?>(Paragraph.OutlineLevelPropertyDefinition, this);
            this.applyEastAsianLineBreakingRulesStyleProperty = new StyleProperty<bool?>(Paragraph.ApplyEastAsianLineBreakingRulesPropertyDefinition, this);
            this.pageBreakBeforeStyleProperty = new StyleProperty<bool?>(Paragraph.PageBreakBeforePropertyDefinition, this);
            this.contextualSpacingStyleProperty = new StyleProperty<bool?>(Paragraph.ContextualSpacingPropertyDefinition, this);
            this.mirrorIndentsStyleProperty = new StyleProperty<bool?>(Paragraph.MirrorIndentsPropertyDefinition, this);
            this.spacingAfterStyleProperty = new StyleProperty<double?>(Paragraph.SpacingAfterPropertyDefinition, this);
            this.spacingBeforeStyleProperty = new StyleProperty<double?>(Paragraph.SpacingBeforePropertyDefinition, this);
            this.automaticSpacingBeforeStyleProperty = new StyleProperty<bool?>(Paragraph.AutomaticSpacingBeforePropertyDefinition, this);
            this.automaticSpacingAfterStyleProperty = new StyleProperty<bool?>(Paragraph.AutomaticSpacingAfterPropertyDefinition, this);
            this.bordersStyleProperty = new StyleProperty<ParagraphBorders>(Paragraph.BordersPropertyDefinition, this);

            this.backgroundColorStyleProperty = new StyleProperty<ThemableColor>(Paragraph.BackgroundColorPropertyDefinition, this);
            this.shadingPatternColorStyleProperty = new StyleProperty<ThemableColor>(Paragraph.ShadingPatternColorPropertyDefinition, this);
            this.shadingPatternStyleProperty = new StyleProperty<ShadingPattern?>(Paragraph.ShadingPatternPropertyDefinition, this);

            this.firstLineIndentStyleProperty = new StyleProperty<double?>(Paragraph.FirstLineIndentPropertyDefinition, this);
            this.hangingIndentStyleProperty = new StyleProperty<double?>(Paragraph.HangingIndentPropertyDefinition, this);
            this.leftIndentStyleProperty = new StyleProperty<double?>(Paragraph.LeftIndentPropertyDefinition, this);
            this.rightIndentStyleProperty = new StyleProperty<double?>(Paragraph.RightIndentPropertyDefinition, this);

            this.tabStopsStyleProperty = new TabStopCollectionStyleProperty(Paragraph.TabStopsPropertyDefinition, this);

            // ParagraphMarkerProperties should only be initialized when owner element is paragraph
            if (this.OwnerDocumentElement != null)
            {
                this.ParagraphMarkerProperties = new CharacterProperties(this.OwnerDocumentElement, true);
            }

            InitStylePropertyGetters();
        }
    }
}