﻿namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Specifies the inter-line and inter-paragraph spacing which shall be applied to the contents of a paragraph.
    /// </summary>
    public class ParagraphSpacing
    {
        private readonly ParagraphProperties paragraphProperties;

        /// <summary>
        /// Initializes a new instance of the <see cref="ParagraphSpacing" /> class.
        /// </summary>
        /// <param name="paragraphProperties">The paragraph properties.</param>
        public ParagraphSpacing(ParagraphProperties paragraphProperties)
        {
            this.paragraphProperties = paragraphProperties;
        }

        /// <summary>
        /// Gets or sets the spacing that should be added above the first line in a paragraph.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The spacing before value.</value>
        public double SpacingBefore
        {
            get
            {
                return this.paragraphProperties.SpacingBefore.GetActualValue().Value;
            }
            set
            {
                this.paragraphProperties.SpacingBefore.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets value indicating whether a consumer shall automatically determine the spacing 
        /// before the paragraph based on its contents the automatic spacing after.
        /// </summary>
        /// <value>The automatic spacing after.</value>
        public bool AutomaticSpacingBefore
        {
            get
            {
                return this.paragraphProperties.AutomaticSpacingBefore.GetActualValue().Value;
            }
            set
            {
                this.paragraphProperties.AutomaticSpacingBefore.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the spacing that should be added after the last line in a paragraph.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The spacing after value.</value>
        public double SpacingAfter
        {
            get
            {
                return this.paragraphProperties.SpacingAfter.GetActualValue().Value;
            }
            set
            {
                this.paragraphProperties.SpacingAfter.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets value indicating whether a consumer shall automatically determine the spacing 
        /// after the paragraph based on its contents the automatic spacing after.
        /// </summary>
        /// <value>The automatic spacing after.</value>
        public bool AutomaticSpacingAfter
        {
            get
            {
                return this.paragraphProperties.AutomaticSpacingAfter.GetActualValue().Value;
            }
            set
            {
                this.paragraphProperties.AutomaticSpacingAfter.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the value indicating the amount of vertical spacing between lines of text within a paragraph.
        /// <para>If <see cref="LineSpacingType" /> is Auto then the line height value times the value of the property, 
        /// otherwise the value is in device independent pixels (1/96 inch).</para>
        /// </summary>
        /// <value>The line spacing value.</value>
        public double LineSpacing
        {
            get
            {
                return this.paragraphProperties.LineSpacing.GetActualValue().Value;
            }
            set
            {
                this.paragraphProperties.LineSpacing.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the value indicating how the spacing between lines is calculated.
        /// </summary>
        /// <value>The type line spacing rule.</value>
        public HeightType LineSpacingType
        {
            get
            {
                return this.paragraphProperties.LineSpacingType.GetActualValue().Value;
            }
            set
            {
                this.paragraphProperties.LineSpacingType.LocalValue = value;
            }
        }
    }
}
