﻿using System;
using System.Collections.Generic;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Telerik.Windows.Documents.Primitives;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;
using Telerik.Windows.Documents.Model;

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Defines a set of properties which can be applied to a <see cref="Section"/> element.
    /// </summary>
    public sealed class SectionProperties : DocumentElementPropertiesBase
    {
        private static Func<SectionProperties, IStyleProperty>[] stylePropertyGetters;

        private LocalProperty<bool?> hasDifferentFirstPageHeaderFooterProperty;
        private LocalProperty<PageOrientation?> pageOrientationProperty;
        private LocalProperty<Size?> pageSizeProperty;
        private LocalProperty<Padding> pageMarginsProperty;
        private LocalProperty<double?> headerTopMarginProperty;
        private LocalProperty<double?> footerBottomMarginProperty;
        private LocalProperty<SectionType?> sectionTypeProperty;
        private LocalProperty<VerticalAlignment> verticalAlignmentProperty;
        private LocalProperty<ChapterSeparatorType?> chapterSeparatorCharacterProperty;
        private LocalProperty<int?> chapterHeadingStyleIndexProperty;
        private LocalProperty<NumberingStyle?> pageNumberFormatProperty;
        private LocalProperty<int?> startingPageNumberProperty;

        /// <summary>
        /// Initializes a new instance of the <see cref="SectionProperties" /> class.
        /// </summary>
        /// <param name="section">The owner section.</param>
        internal SectionProperties(Section section)
            : base(section, false)
        {
            this.InitProperties();
        }

        /// <summary>
        /// Gets the has different first page header footer.
        /// </summary>
        /// <value>The has different first page header footer.</value>
        public IStyleProperty<bool?> HasDifferentFirstPageHeaderFooter
        {
            get
            {
                return this.hasDifferentFirstPageHeaderFooterProperty;
            }
        }

        /// <summary>
        /// Gets the page orientation property.
        /// </summary>
        /// <value>The page orientation.</value>
        public IStyleProperty<PageOrientation?> PageOrientation
        {
            get
            {
                return this.pageOrientationProperty;
            }
        }

        /// <summary>
        /// Gets the page size property.
        /// The width and height are in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The page size property.</value>
        public IStyleProperty<Size?> PageSize
        {
            get
            {
                return this.pageSizeProperty;
            }
        }

        /// <summary>
        /// Gets the page margins property.
        /// </summary>
        /// <value>The page margins property.</value>
        public IStyleProperty<Padding> PageMargins
        {
            get
            {
                return this.pageMarginsProperty;
            }
        }

        /// <summary>
        /// Gets the header top margin property.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The header top margin property.</value>
        public IStyleProperty<double?> HeaderTopMargin
        {
            get
            {
                return this.headerTopMarginProperty;
            }
        }

        /// <summary>
        /// Gets the footer bottom margin property.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The footer bottom margin property.</value>
        public IStyleProperty<double?> FooterBottomMargin
        {
            get
            {
                return this.footerBottomMarginProperty;
            }
        }

        /// <summary>
        /// Gets the section type property.
        /// </summary>
        /// <value>The section type property.</value>
        public IStyleProperty<SectionType?> SectionType
        {
            get
            {
                return this.sectionTypeProperty;
            }
        }

        /// <summary>
        /// Gets the vertical alignment property.
        /// </summary>
        /// <value>The vertical alignment property.</value>
        public IStyleProperty<VerticalAlignment> VerticalAlignment
        {
            get
            {
                return this.verticalAlignmentProperty;
            }
        }

        /// <summary>
        /// Gets the chapter separator character that shall appear between the chapter heading style index and the page number.
        /// </summary>
        /// <value>The chapter separator character.</value>
        public IStyleProperty<ChapterSeparatorType?> ChapterSeparatorCharacter
        {
            get
            {
                return this.chapterSeparatorCharacterProperty;
            }
        }

        /// <summary>
        /// Gets the index of the chapter heading style.
        /// </summary>
        /// <value>The index of the chapter heading style.</value>
        public IStyleProperty<int?> ChapterHeadingStyleIndex
        {
            get
            {
                return this.chapterHeadingStyleIndexProperty;
            }
        }

        /// <summary>
        /// Gets the page number format for the page numbering in the current section.
        /// </summary>
        /// <value>The page number format.</value>
        public IStyleProperty<NumberingStyle?> PageNumberFormat
        {
            get
            {
                return this.pageNumberFormatProperty;
            }
        }

        /// <summary>
        /// Gets the starting page number.
        /// <para>This number appears on the first page of the section.</para>
        /// </summary>
        /// <value>The starting page number.</value>
        public IStyleProperty<int?> StartingPageNumber
        {
            get
            {
                return this.startingPageNumberProperty;
            }
        }

        /// <summary>
        /// Enumerates the style properties.
        /// </summary>
        /// <returns>The style properties.</returns>
        protected override IEnumerable<IStyleProperty> EnumerateStyleProperties()
        {
            yield return this.HasDifferentFirstPageHeaderFooter;
            yield return this.PageOrientation;
            yield return this.PageSize;
            yield return this.PageMargins;
            yield return this.VerticalAlignment;
            yield return this.HeaderTopMargin;
            yield return this.FooterBottomMargin;
            yield return this.SectionType;
            yield return this.ChapterSeparatorCharacter;
            yield return this.ChapterHeadingStyleIndex;
            yield return this.PageNumberFormat;
            yield return this.StartingPageNumber;
        }

        /// <summary>
        /// Gets specific style property or null.
        /// </summary>
        /// <param name="propertyDefinition">The property definition.</param>
        /// <returns></returns>
        protected override IStyleProperty GetStylePropertyOverride(IStylePropertyDefinition propertyDefinition)
        {
            if (propertyDefinition != DocumentElementPropertiesBase.StyleIdPropertyDefinition &&
                propertyDefinition.StylePropertyType != StylePropertyType.Section)
            {
                return null;
            }

            return stylePropertyGetters[propertyDefinition.GlobalPropertyIndex](this);
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Maintainability", "CA1502:AvoidExcessiveComplexity")]
        private static void InitStylePropertyGetters()
        {
            if (stylePropertyGetters != null)
            {
                return;
            }

            stylePropertyGetters = new Func<SectionProperties, IStyleProperty>[13];
            stylePropertyGetters[DocumentElementPropertiesBase.StyleIdPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.StyleIdProperty; };
            stylePropertyGetters[Section.HasDifferentFirstPageHeaderFooterPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.hasDifferentFirstPageHeaderFooterProperty; };
            stylePropertyGetters[Section.PageOrientationPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.pageOrientationProperty; };
            stylePropertyGetters[Section.PageSizePropertyDefinition.GlobalPropertyIndex] = prop => { return prop.pageSizeProperty; };
            stylePropertyGetters[Section.PageMarginsPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.pageMarginsProperty; };
            stylePropertyGetters[Section.VerticalAlignmentPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.verticalAlignmentProperty; };
            stylePropertyGetters[Section.HeaderTopMarginPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.headerTopMarginProperty; };
            stylePropertyGetters[Section.FooterBottomMarginPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.footerBottomMarginProperty; };
            stylePropertyGetters[Section.SectionTypePropertyDefinition.GlobalPropertyIndex] = prop => { return prop.sectionTypeProperty; };
            stylePropertyGetters[Section.ChapterSeparatorCharacterPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.chapterSeparatorCharacterProperty; };
            stylePropertyGetters[Section.ChapterHeadingStyleIndexPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.chapterHeadingStyleIndexProperty; };
            stylePropertyGetters[Section.PageNumberFormatPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.pageNumberFormatProperty; };
            stylePropertyGetters[Section.StartingPageNumberPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.startingPageNumberProperty; };
        }

        private void InitProperties()
        {
            this.hasDifferentFirstPageHeaderFooterProperty = new LocalProperty<bool?>(Section.HasDifferentFirstPageHeaderFooterPropertyDefinition, this);
            this.pageOrientationProperty = new LocalProperty<PageOrientation?>(Section.PageOrientationPropertyDefinition, this);
            this.pageSizeProperty = new LocalProperty<Size?>(Section.PageSizePropertyDefinition, this);
            this.pageMarginsProperty = new LocalProperty<Padding>(Section.PageMarginsPropertyDefinition, this);
            this.verticalAlignmentProperty = new LocalProperty<VerticalAlignment>(Section.VerticalAlignmentPropertyDefinition, this);
            this.headerTopMarginProperty = new LocalProperty<double?>(Section.HeaderTopMarginPropertyDefinition, this);
            this.footerBottomMarginProperty = new LocalProperty<double?>(Section.FooterBottomMarginPropertyDefinition, this);
            this.sectionTypeProperty = new LocalProperty<SectionType?>(Section.SectionTypePropertyDefinition, this);
            this.chapterSeparatorCharacterProperty = new LocalProperty<ChapterSeparatorType?>(Section.ChapterSeparatorCharacterPropertyDefinition, this);
            this.chapterHeadingStyleIndexProperty = new LocalProperty<int?>(Section.ChapterHeadingStyleIndexPropertyDefinition, this);
            this.pageNumberFormatProperty = new LocalProperty<NumberingStyle?>(Section.PageNumberFormatPropertyDefinition, this);
            this.startingPageNumberProperty = new LocalProperty<int?>(Section.StartingPageNumberPropertyDefinition, this);

            InitStylePropertyGetters();
        }
    }
}