﻿using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Defines a tab stop element. This class is immutable.
    /// </summary>
    public class TabStop
    {
        private readonly double position;
        private readonly TabStopType type;
        private readonly TabStopLeader leader;

        /// <summary>
        /// Initializes a new instance of the <see cref="TabStop" /> class.
        /// </summary>
        /// <param name="position">The position in device independent pixels (1/96 inch).</param>
        public TabStop(double position)
            : this(position, TabStopType.Left, TabStopLeader.None)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TabStop" /> class.
        /// </summary>
        /// <param name="position">The position in device independent pixels (1/96 inch).</param>
        /// <param name="type">The tab stop type.</param>
        public TabStop(double position, TabStopType type)
            : this(position, type, TabStopLeader.None)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TabStop" /> class.
        /// </summary>
        /// <param name="position">The position in device independent pixels (1/96 inch).</param>
        /// <param name="type">The tab stop type.</param>
        /// <param name="leader">The tab stop leader.</param>
        public TabStop(double position, TabStopType type, TabStopLeader leader)
        {
            this.position = position;
            this.type = type;
            this.leader = leader;
        }

        /// <summary>
        /// Gets or sets the position of the tab stop.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The position in points.</value>
        public double Position
        {
            get
            {
                return this.position;
            }
        }

        /// <summary>
        /// Gets or sets the tab stop type.
        /// </summary>
        /// <value>The tab stop type.</value>
        public TabStopType Type
        {
            get
            {
                return this.type;
            }
        }

        /// <summary>
        /// Gets or sets the tab stop leader.
        /// </summary>
        /// <value>The tab stop leader.</value>
        public TabStopLeader Leader
        {
            get
            {
                return this.leader;
            }
        }

        /// <summary>
        /// Determines whether the specified tab stops are equal.
        /// </summary>
        /// <returns>True if the tab stops are equal.</returns>
        public static bool operator ==(TabStop tabStop, TabStop otherTabStop)
        {
            if (object.ReferenceEquals(tabStop, otherTabStop))
            {
                return true;
            }

            if ((object)tabStop == null || ((object)otherTabStop == null))
            {
                return false;
            }

            return tabStop.Equals(otherTabStop);
        }

        /// <summary>
        /// Determines whether the specified tab stops are different.
        /// </summary>
        /// <returns>True if the tab stops are different.</returns>
        public static bool operator !=(TabStop tabStop, TabStop otherTabStop)
        {
            return !(tabStop == otherTabStop);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The object to compare with the current object.</param>
        /// <returns>
        /// Returns <c>true</c> if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, <c>false</c>.
        /// </returns>
        public override bool Equals(object obj)
        {
            TabStop other = obj as TabStop;
            if (other == null)
            {
                return false;
            }

            return this.Leader == other.Leader &&
                   this.Position == other.Position &&
                   this.Type == other.Type;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return ObjectExtensions.CombineHashCodes(this.Leader.GetHashCode(), this.Position.GetHashCode(), this.Type.GetHashCode());
        }
    }
}