﻿using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Represents a collection of <see cref="TabStop"/> elements. This class is immutable.
    /// </summary>
    public class TabStopCollection : IEnumerable<TabStop>
    {
        private readonly List<TabStop> tabStops;

        /// <summary>
        /// Initializes a new instance of the <see cref="TabStopCollection" /> class.
        /// </summary>
        public TabStopCollection()
        {
            this.tabStops = new List<TabStop>();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TabStopCollection" /> class with collection of given <see cref="TabStop"/> elements.
        /// </summary>
        /// <param name="tabStops">The tab stops.</param>
        public TabStopCollection(IEnumerable<TabStop> tabStops)
        {
            Guard.ThrowExceptionIfNull(tabStops, "tabStops");

            this.tabStops = new List<TabStop>(tabStops);
        }

        /// <summary>
        /// Gets the count of tab stop elements in the collection.
        /// </summary>
        /// <value>The count.</value>
        public int Count
        {
            get
            {
                return this.tabStops.Count;
            }
        }

        /// <summary>
        /// Determines whether the specified tab stop collections are equal.
        /// </summary>
        /// <returns>True if the tab stop collections are equal.</returns>
        public static bool operator ==(TabStopCollection tabStops, TabStopCollection otherTabStops)
        {
            if (object.ReferenceEquals(tabStops, otherTabStops))
            {
                return true;
            }

            if ((object)tabStops == null || ((object)otherTabStops == null))
            {
                return false;
            }

            return tabStops.Equals(otherTabStops);
        }

        /// <summary>
        /// Determines whether the specified tab stop collections are different.
        /// </summary>
        /// <returns>True if the tab stop collections are different.</returns>
        public static bool operator !=(TabStopCollection tabStops, TabStopCollection otherTabStops)
        {
            return !(tabStops == otherTabStops);
        }

        /// <summary>
        /// Returns a new instance of <see cref="TabStopCollection"/> class in which a specified tab stop is inserted.
        /// </summary>
        /// <param name="tabStop">The tab stop.</param>
        /// <returns>The tab stop collection containing the added tab stop.</returns>
        public TabStopCollection Insert(TabStop tabStop)
        {
            Guard.ThrowExceptionIfNull(tabStop, "tabStop");

            List<TabStop> newTabStopCollection = this.tabStops.ToList();
            newTabStopCollection.Add(tabStop);

            return new TabStopCollection(newTabStopCollection);
        }

        /// <summary>
        /// Returns a new instance of <see cref="TabStopCollection"/> class in which a specified tab stop is removed.
        /// </summary>
        /// <param name="tabStop">The tab stop.</param>
        /// <returns>The tab stop collection from which the specified tab stop is removed.</returns>                
        public TabStopCollection Remove(TabStop tabStop)
        {
            Guard.ThrowExceptionIfNull(tabStop, "tabStop");

            IList<TabStop> newTabStopCollection = this.tabStops.ToList();
            newTabStopCollection.Remove(tabStop);

            return new TabStopCollection(newTabStopCollection);
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<TabStop> GetEnumerator()
        {
            return this.tabStops.GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.GetEnumerator();
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The object to compare with the current object.</param>
        /// <returns>
        /// Returns <c>true</c> if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, <c>false</c>.
        /// </returns>
        public override bool Equals(object obj)
        {
            TabStopCollection other = obj as TabStopCollection;
            if ((object)other == null || this.Count != other.Count)
            {
                return false;
            }

            foreach (TabStop tabStop in this.tabStops)
            {
                if (!other.Contains(tabStop))
                {
                    return false;
                }
            }

            return true;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return ObjectExtensions.CombineHashCodes(this.tabStops.GetHashCode(), base.GetHashCode());
        }

        internal TabStopCollection Merge(IEnumerable<TabStop> tabStopsToMerge)
        {
            Guard.ThrowExceptionIfNull(tabStopsToMerge, "tabStopsToMerge");

            List<TabStop> mergedTabStops = new List<TabStop>(this.tabStops);
            mergedTabStops.AddRange(tabStopsToMerge.Where(ts2 => this.tabStops.All(ts1 => !ts1.Equals(ts2))));

            return new TabStopCollection(mergedTabStops);
        }
    }
}