﻿using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Defines set of borders which can be applied to table cell element. This class is immutable.
    /// </summary>
    public class TableCellBorders
    {
        private readonly Border top;
        private readonly Border bottom;
        private readonly Border left;
        private readonly Border right;
        private readonly Border insideHorizontal;
        private readonly Border insideVertical;
        private readonly Border diagonalDown;
        private readonly Border diagonalUp;

        /// <summary>
        /// Initializes a new instance of the <see cref="TableCellBorders" /> class.
        /// </summary>
        public TableCellBorders()
            : this(Border.DefaultBorder)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TableCellBorders" /> class. Setting only the outer borders.
        /// </summary>
        /// <param name="all">All.</param>
        public TableCellBorders(Border all)
            : this(all, all, all, all)
        {
            Guard.ThrowExceptionIfNull(all, "all");
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TableCellBorders" /> class.
        /// </summary>
        /// <param name="leftBorder">The left border.</param>
        /// <param name="topBorder">The top border.</param>
        /// <param name="rightBorder">The right border.</param>
        /// <param name="bottomBorder">The bottom border.</param>
        public TableCellBorders(Border leftBorder, Border topBorder, Border rightBorder, Border bottomBorder)
            : this(leftBorder, topBorder, rightBorder, bottomBorder, null, null, null, null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TableCellBorders" /> class.
        /// </summary>
        /// <param name="leftBorder">The left border.</param>
        /// <param name="topBorder">The top border.</param>
        /// <param name="rightBorder">The right border.</param>
        /// <param name="bottomBorder">The bottom border.</param>
        /// <param name="insideHorizontalBorder">The inside horizontal border.</param>
        /// <param name="insideVerticalBorder">The inside vertical border.</param>
        /// <param name="diagonalDownBorder">The diagonal down border.</param>
        /// <param name="diagonalUpBorder">The diagonal up border.</param>
        public TableCellBorders(Border leftBorder, Border topBorder, Border rightBorder, Border bottomBorder, Border insideHorizontalBorder, Border insideVerticalBorder, Border diagonalDownBorder, Border diagonalUpBorder)
        {
            this.left = leftBorder ?? Border.DefaultBorder;
            this.top = topBorder ?? Border.DefaultBorder;
            this.right = rightBorder ?? Border.DefaultBorder;
            this.bottom = bottomBorder ?? Border.DefaultBorder;
            this.insideHorizontal = insideHorizontalBorder ?? Border.DefaultBorder;
            this.insideVertical = insideVerticalBorder ?? Border.DefaultBorder;
            this.diagonalDown = diagonalDownBorder ?? Border.EmptyBorder;
            this.diagonalUp = diagonalUpBorder ?? Border.EmptyBorder;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TableCellBorders" /> class. Creates copy of the the source table cell borders and replaces only the specified borders.
        /// </summary>
        /// <param name="source">The source.</param>
        /// <param name="leftBorder">The left border.</param>
        /// <param name="topBorder">The top border.</param>
        /// <param name="rightBorder">The right border.</param>
        /// <param name="bottomBorder">The bottom border.</param>
        /// <param name="insideHorizontalBorder">The inside horizontal border.</param>
        /// <param name="insideVerticalBorder">The inside vertical border.</param>
        /// <param name="diagonalDownBorder">The diagonal down border.</param>
        /// <param name="diagonalUpBorder">The diagonal up border.</param>
        public TableCellBorders(TableCellBorders source, Border leftBorder = null, Border topBorder = null, Border rightBorder = null, Border bottomBorder = null, Border insideHorizontalBorder = null, Border insideVerticalBorder = null, Border diagonalDownBorder = null, Border diagonalUpBorder = null)
        {
            this.left = leftBorder ?? source.Left;
            this.top = topBorder ?? source.Top;
            this.right = rightBorder ?? source.Right;
            this.bottom = bottomBorder ?? source.Bottom;
            this.insideHorizontal = insideHorizontalBorder ?? source.InsideHorizontal;
            this.insideVertical = insideVerticalBorder ?? source.InsideVertical;
            this.diagonalDown = diagonalDownBorder ?? source.DiagonalDown;
            this.diagonalUp = diagonalUpBorder ?? source.DiagonalUp;
        }

        /// <summary>
        /// Gets the top border.
        /// </summary>
        /// <value>The top border.</value>
        public Border Top
        {
            get
            {
                return this.top;
            }
        }

        /// <summary>
        /// Gets the bottom border.
        /// </summary>
        /// <value>The bottom border.</value>
        public Border Bottom
        {
            get
            {
                return this.bottom;
            }
        }

        /// <summary>
        /// Gets the left border.
        /// </summary>
        /// <value>The left border.</value>
        public Border Left
        {
            get
            {
                return this.left;
            }
        }

        /// <summary>
        /// Gets the right border.
        /// </summary>
        /// <value>The right border.</value>
        public Border Right
        {
            get
            {
                return this.right;
            }
        }

        /// <summary>
        /// Gets the inner horizontal border.
        /// </summary>
        /// <value>The inner horizontal border.</value>
        public Border InsideHorizontal
        {
            get
            {
                return this.insideHorizontal;
            }
        }

        /// <summary>
        /// Gets the inner vertical border.
        /// </summary>
        /// <value>The inner vertical border.</value>
        public Border InsideVertical
        {
            get
            {
                return this.insideVertical;
            }
        }

        /// <summary>
        /// Gets the top left to bottom right diagonal border.
        /// </summary>
        /// <value>The diagonal down.</value>
        public Border DiagonalDown
        {
            get
            {
                return this.diagonalDown;
            }
        }

        /// <summary>
        /// Gets the top right to bottom left diagonal border.
        /// </summary>
        /// <value>The diagonal up.</value>
        public Border DiagonalUp
        {
            get
            {
                return this.diagonalUp;
            }
        }

        /// <summary>
        /// Determines whether the specified table cell borders are equal.
        /// </summary>
        /// <returns>True if the table cell borders are equal.</returns>
        public static bool operator ==(TableCellBorders a, TableCellBorders b)
        {
            // If both are null, or both are same instance, return true.
            if (object.ReferenceEquals(a, b))
            {
                return true;
            }

            // If one is null, but not both, return false.
            if (((object)a == null) || ((object)b == null))
            {
                return false;
            }

            return a.Equals(b);
        }

        /// <summary>
        /// Determines whether the specified table cell borders are different.
        /// </summary>
        /// <returns>True if the table cell borders are different.</returns>
        public static bool operator !=(TableCellBorders a, TableCellBorders b)
        {
            return !(a == b);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// Returns true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            TableCellBorders other = obj as TableCellBorders;
            if (other == null)
            {
                return false;
            }

            return this.Left.Equals(other.Left) &&
                this.Top.Equals(other.Top) &&
                this.Right.Equals(other.Right) &&
                this.Bottom.Equals(other.Bottom) &&
                this.InsideHorizontal.Equals(other.InsideHorizontal) &&
                this.InsideVertical.Equals(other.InsideVertical) &&
                this.DiagonalDown.Equals(other.DiagonalDown) &&
                this.DiagonalUp.Equals(other.DiagonalUp);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return ObjectExtensions.CombineHashCodes(this.Left.GetHashCode(),
                this.Top.GetHashCode(),
                this.Right.GetHashCode(),
                this.Bottom.GetHashCode(),
                this.InsideHorizontal.GetHashCode(),
                this.InsideVertical.GetHashCode(),
                this.DiagonalDown.GetHashCode(),
                this.DiagonalUp.GetHashCode());
        }
    }
}
