﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;
using Telerik.Windows.Documents.Primitives;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Defines a set of properties which can be applied to a <see cref="TableCell"/> element.
    /// </summary>
    public sealed class TableCellProperties : DocumentElementPropertiesBase, IPropertiesWithShading, IPropertiesWithPadding
    {
        private static Func<TableCellProperties, IStyleProperty>[] stylePropertyGetters;

        private StyleProperty<ThemableColor> backgroundColorStyleProperty;
        private StyleProperty<ThemableColor> shadingPatternColorStyleProperty;
        private StyleProperty<ShadingPattern?> shadingPatternStyleProperty;
        private StyleProperty<TableCellBorders> bordersStyleProperty;
        private StyleProperty<Padding> paddingStyleProperty;
        private LocalProperty<int?> columnSpanStyleProperty;
        private LocalProperty<int?> rowSpanStyleProperty;
        private LocalProperty<bool?> ignoreCellMarkerInRowHeightCalculationStyleProperty;
        private LocalProperty<bool?> canWrapContentStyleProperty;
        private LocalProperty<TableWidthUnit> preferredWidthStyleProperty;
        private LocalProperty<VerticalAlignment?> verticalAlignmentStyleProperty;
        private LocalProperty<TextDirection> textDirectionStyleProperty;

        /// <summary>
        /// Initializes a new instance of the <see cref="TableCellProperties" /> class.
        /// </summary>
        /// <param name="style">The style.</param>
        internal TableCellProperties(Style style)
            : base(style)
        {
            this.InitProperties();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TableCellProperties" /> class.
        /// </summary>
        /// <param name="cell">The owner cell.</param>
        internal TableCellProperties(TableCell cell)
            : base(cell, false)
        {
            this.InitProperties();
        }

        /// <summary>
        /// Gets style property used to get or set the borders.
        /// </summary>
        /// <value>The borders.</value>
        public IStyleProperty<TableCellBorders> Borders
        {
            get
            {
                return this.bordersStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the color of the background.
        /// </summary>
        /// <value>The color of the background.</value>
        public IStyleProperty<ThemableColor> BackgroundColor
        {
            get
            {
                return this.backgroundColorStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating the color for any foreground pattern used in this shading.
        /// </summary>
        /// <value>The color of the pattern.</value>
        public IStyleProperty<ThemableColor> ShadingPatternColor
        {
            get
            {
                return this.shadingPatternColorStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating the pattern which shall be used to lay the pattern color over the background color for this shading.
        /// </summary>
        /// <returns>The default value is <c>Clear</c></returns>
        public IStyleProperty<ShadingPattern?> ShadingPattern
        {
            get
            {
                return this.shadingPatternStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the padding.
        /// </summary>
        /// <value>The table cell padding.</value>
        public IStyleProperty<Padding> Padding
        {
            get
            {
                return this.paddingStyleProperty;
            }
        }

        /// <summary>
        /// Gets local style property used to get or set the column span. 
        /// This property cannot be derived from a style.
        /// </summary>
        /// <returns>The default value is <c>1</c>.</returns>
        public IStyleProperty<int?> ColumnSpan
        {
            get
            {
                return this.columnSpanStyleProperty;
            }
        }

        /// <summary>
        /// Gets local style property used to get or set the row span. 
        /// This property cannot be derived from a style.
        /// </summary>
        /// <returns>The default value is <c>1</c>.</returns>        
        public IStyleProperty<int?> RowSpan
        {
            get
            {
                return this.rowSpanStyleProperty;
            }
        }

        /// <summary>
        /// Gets local style property used to get or set a value indicating whether cell marker will be ignored when row height is calculated. 
        /// This property cannot be derived from a style.
        /// </summary>
        /// <returns>The default value is <c>false</c>.</returns>
        public IStyleProperty<bool?> IgnoreCellMarkerInRowHeightCalculation
        {
            get
            {
                return this.ignoreCellMarkerInRowHeightCalculationStyleProperty;
            }
        }

        /// <summary>
        /// Gets local style property used to get or set a value indicating whether the content can be wrapped during the table layout. 
        /// This property cannot be derived from a style.
        /// </summary>
        /// <returns>The default value is <c>true</c>.</returns>
        public IStyleProperty<bool?> CanWrapContent
        {
            get
            {
                return this.canWrapContentStyleProperty;
            }
        }

        /// <summary>
        /// Gets local style property used to get or set the preferred width. 
        /// This property cannot be derived from a style.
        /// </summary>
        /// <value>The preferred width.</value>
        public IStyleProperty<TableWidthUnit> PreferredWidth
        {
            get
            {
                return this.preferredWidthStyleProperty;
            }
        }

        /// <summary>
        /// Gets local style property used to get or set the vertical alignment. 
        /// This property cannot be derived from a style.
        /// </summary>
        /// <returns>The default value is <c>Top</c>.</returns>
        public IStyleProperty<VerticalAlignment?> VerticalAlignment
        {
            get
            {
                return this.verticalAlignmentStyleProperty;
            }
        }

        /// <summary>
        /// Gets local style property used to get or set the text direction. 
        /// This property cannot be derived from a style.
        /// </summary>
        /// <value>The text direction.</value>
        public IStyleProperty<TextDirection> TextDirection
        {
            get
            {
                return this.textDirectionStyleProperty;
            }
        }

        /// <summary>
        /// Gets the document.
        /// </summary>
        /// <value>The document.</value>
        RadFlowDocument IPropertiesWithShading.Document
        {
            get
            {
                return this.Document;
            }
        }

        /// <summary>
        /// Enumerates the style properties.
        /// </summary>
        /// <returns>The style properties.</returns>
        protected override IEnumerable<IStyleProperty> EnumerateStyleProperties()
        {
            yield return this.Borders;
            yield return this.BackgroundColor;
            yield return this.ShadingPatternColor;
            yield return this.ShadingPattern;
            yield return this.Padding;
            yield return this.ColumnSpan;
            yield return this.RowSpan;
            yield return this.IgnoreCellMarkerInRowHeightCalculation;
            yield return this.CanWrapContent;
            yield return this.PreferredWidth;
            yield return this.VerticalAlignment;
            yield return this.TextDirection;
        }

        /// <summary>
        /// Gets specific style property or null.
        /// </summary>
        /// <param name="propertyDefinition">The property definition.</param>
        /// <returns></returns>
        protected override IStyleProperty GetStylePropertyOverride(IStylePropertyDefinition propertyDefinition)
        {
            if (propertyDefinition != DocumentElementPropertiesBase.StyleIdPropertyDefinition &&
                propertyDefinition.StylePropertyType != StylePropertyType.TableCell)
            {
                return null;
            }

            return stylePropertyGetters[propertyDefinition.GlobalPropertyIndex](this);
        }

        private static void InitStylePropertyGetters()
        {
            if (stylePropertyGetters != null)
            {
                return;
            }

            stylePropertyGetters = new Func<TableCellProperties, IStyleProperty>[13];
            stylePropertyGetters[DocumentElementPropertiesBase.StyleIdPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.StyleIdProperty; };
            stylePropertyGetters[TableCell.BackgroundColorPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.backgroundColorStyleProperty; };
            stylePropertyGetters[TableCell.ShadingPatternColorPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.shadingPatternColorStyleProperty; };
            stylePropertyGetters[TableCell.ShadingPatternPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.shadingPatternStyleProperty; };
            stylePropertyGetters[TableCell.BordersPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.bordersStyleProperty; };
            stylePropertyGetters[TableCell.PaddingPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.paddingStyleProperty; };
            stylePropertyGetters[TableCell.ColumnSpanPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.columnSpanStyleProperty; };
            stylePropertyGetters[TableCell.RowSpanPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.rowSpanStyleProperty; };
            stylePropertyGetters[TableCell.IgnoreCellMarkerInRowHeightCalculationPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.ignoreCellMarkerInRowHeightCalculationStyleProperty; };
            stylePropertyGetters[TableCell.CanWrapContentPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.canWrapContentStyleProperty; };
            stylePropertyGetters[TableCell.PreferredWidthPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.preferredWidthStyleProperty; };
            stylePropertyGetters[TableCell.VerticalAlignmentPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.verticalAlignmentStyleProperty; };
            stylePropertyGetters[TableCell.TextDirectionPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.textDirectionStyleProperty; };
        }

        /// <summary>
        /// Initialize the properties.
        /// </summary>
        private void InitProperties()
        {
            this.backgroundColorStyleProperty = new StyleProperty<ThemableColor>(TableCell.BackgroundColorPropertyDefinition, this);
            this.shadingPatternColorStyleProperty = new StyleProperty<ThemableColor>(TableCell.ShadingPatternColorPropertyDefinition, this);
            this.shadingPatternStyleProperty = new StyleProperty<ShadingPattern?>(TableCell.ShadingPatternPropertyDefinition, this);
            this.bordersStyleProperty = new StyleProperty<TableCellBorders>(TableCell.BordersPropertyDefinition, this);
            this.paddingStyleProperty = new StyleProperty<Padding>(TableCell.PaddingPropertyDefinition, this);

            this.columnSpanStyleProperty = new LocalProperty<int?>(TableCell.ColumnSpanPropertyDefinition, this);
            this.rowSpanStyleProperty = new LocalProperty<int?>(TableCell.RowSpanPropertyDefinition, this);
            this.ignoreCellMarkerInRowHeightCalculationStyleProperty =
                new LocalProperty<bool?>(TableCell.IgnoreCellMarkerInRowHeightCalculationPropertyDefinition, this);
            this.canWrapContentStyleProperty = new LocalProperty<bool?>(TableCell.CanWrapContentPropertyDefinition, this);
            this.preferredWidthStyleProperty = new LocalProperty<TableWidthUnit>(TableCell.PreferredWidthPropertyDefinition, this);
            this.verticalAlignmentStyleProperty = new LocalProperty<VerticalAlignment?>(TableCell.VerticalAlignmentPropertyDefinition, this);
            this.textDirectionStyleProperty = new LocalProperty<TextDirection>(TableCell.TextDirectionPropertyDefinition, this);

            InitStylePropertyGetters();
        }
    }
}
