﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;
using Telerik.Windows.Documents.Primitives;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Defines a set of properties which can be applied to a <see cref="Table"/> element.
    /// </summary>
    public sealed class TableProperties : DocumentElementPropertiesBase, IPropertiesWithShading, IPropertiesWithPadding
    {
        private static Func<TableProperties, IStyleProperty>[] stylePropertyGetters;

        private StyleProperty<Alignment?> alignmentStyleProperty;
        private StyleProperty<Padding> tableCellPaddingStyleProperty;
        private StyleProperty<double?> tableCellSpacingStyleProperty;
        private StyleProperty<double?> indentStyleProperty;
        private StyleProperty<int?> rowBandingStyleProperty;
        private StyleProperty<int?> columnBandingStyleProperty;
        private StyleProperty<TableBorders> tableBordersStyleProperty;
        private StyleProperty<ThemableColor> backgroundColorStyleProperty;
        private StyleProperty<ThemableColor> shadingPatternColorStyleProperty;
        private StyleProperty<ShadingPattern?> shadingPatternStyleProperty;
        private LocalProperty<FlowDirection?> flowDirectionStyleProperty;
        private LocalProperty<TableWidthUnit> preferredWidthStyleProperty;
        private LocalProperty<TableLooks?> looksStyleProperty;
        private LocalProperty<TableLayoutType?> layoutTypeStyleProperty;
        private LocalProperty<bool?> overlapStyleProperty;

        // TODO Implement.
        ////private LocalProperty<FloatingPosition> floatingPositionStyleProperty;

        /// <summary>
        /// Initializes a new instance of the <see cref="TableProperties" /> class.
        /// </summary>
        /// <param name="style">The style.</param>
        internal TableProperties(Style style)
            : base(style)
        {
            this.InitProperties();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TableProperties" /> class.
        /// </summary>
        /// <param name="table">The owner table.</param>
        internal TableProperties(Table table)
            : base(table, false)
        {
            this.InitProperties();
        }

        /// <summary>
        /// Gets style property used to get or set the alignment of the table.
        /// </summary>
        /// <returns>The default value is <c>Left</c>.</returns>
        public IStyleProperty<Alignment?> Alignment
        {
            get
            {
                return this.alignmentStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the color of the background.
        /// </summary>
        /// <value>The color of the background.</value>
        public IStyleProperty<ThemableColor> BackgroundColor
        {
            get
            {
                return this.backgroundColorStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating the color for any foreground pattern used in this shading.
        /// </summary>
        /// <value>The color of the pattern.</value>
        public IStyleProperty<ThemableColor> ShadingPatternColor
        {
            get
            {
                return this.shadingPatternColorStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating the pattern which shall be used to lay the pattern color over the background color for this shading.
        /// </summary>
        /// <returns>The default value is <c>Clear</c></returns>
        public IStyleProperty<ShadingPattern?> ShadingPattern
        {
            get
            {
                return this.shadingPatternStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the top border.
        /// </summary>
        /// <value>The top border.</value>
        public IStyleProperty<TableBorders> Borders
        {
            get
            {
                return this.tableBordersStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set default padding of the cells inside the table.
        /// </summary>
        /// <value>The table cell padding.</value>
        public IStyleProperty<Padding> TableCellPadding
        {
            get
            {
                return this.tableCellPaddingStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating the spacing between adjacent cells and the edges of the table.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The default value is <c>0</c>.</value>
        public IStyleProperty<double?> TableCellSpacing
        {
            get
            {
                return this.tableCellSpacingStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set a value which shall be added before the leading edge of the table.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The default value is <c>0</c>.</value>
        public IStyleProperty<double?> Indent
        {
            get
            {
                return this.indentStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the number of banded rows.
        /// </summary>
        /// <returns>The default value is <c>0</c>.</returns>
        public IStyleProperty<int?> RowBanding
        {
            get
            {
                return this.rowBandingStyleProperty;
            }
        }

        /// <summary>
        /// Gets style property used to get or set the number of banded columns.
        /// </summary>
        /// <returns>The default value is <c>0</c>.</returns>
        public IStyleProperty<int?> ColumnBanding
        {
            get
            {
                return this.columnBandingStyleProperty;
            }
        }

        /// <summary>
        /// Gets local style property used to get or set the flow direction of cells inside the table. 
        /// This property cannot be derived from a style.
        /// </summary>
        /// <returns>The default value is <c>LeftToRight</c>.</returns>
        public IStyleProperty<FlowDirection?> FlowDirection
        {
            get
            {
                return this.flowDirectionStyleProperty;
            }
        }

        // TODO Implement.
        /////// <summary>
        /////// Gets style property used to get or set the floating position of the table.
        /////// </summary>
        /////// <value>The floating position.</value>
        ////public IStyleProperty<FloatingPosition> FloatingPosition
        ////{
        ////    get
        ////    {
        ////        return this.floatingPositionStyleProperty;
        ////    }
        ////}

        /// <summary>
        /// Gets local style property used to get or set the preferred width. This property cannot be derived from a style.
        /// </summary>
        /// <value>The preferred width.</value>
        public IStyleProperty<TableWidthUnit> PreferredWidth
        {
            get
            {
                return this.preferredWidthStyleProperty;
            }
        }

        /// <summary>
        /// Gets local style property used to get or set the value indicating which components of the conditional style should be applied if such exists. 
        /// This property cannot be derived from a style.
        /// </summary>
        /// <returns>The default value is <c>BandedRows | BandedColumns</c>.</returns>
        public IStyleProperty<TableLooks?> Looks
        {
            get
            {
                return this.looksStyleProperty;
            }
        }

        /// <summary>
        /// Gets local style property used to get or set a value indicating which algorithm shall be used to lay out the content of the table. 
        /// This property cannot be derived from a style.
        /// </summary>
        /// <returns>The default value is <c>AutoFit</c>.</returns>
        public IStyleProperty<TableLayoutType?> LayoutType
        {
            get
            {
                return this.layoutTypeStyleProperty;
            }
        }

        /// <summary>
        /// Gets local style property used to get or set a value indicating whether this floating table shall allow other floating tables to overlap its extents. 
        /// This property cannot be derived from a style.
        /// </summary>
        /// <returns>The default value is <c>true</c>.</returns>
        public IStyleProperty<bool?> Overlap
        {
            get
            {
                return this.overlapStyleProperty;
            }
        }

        /// <summary>
        /// Gets the document.
        /// </summary>
        /// <value>The document.</value>
        RadFlowDocument IPropertiesWithShading.Document
        {
            get
            {
                return this.Document;
            }
        }

        /// <summary>
        /// Gets style property used to get or set default padding of the cells inside the table.
        /// </summary>
        /// <value>The table cell padding.</value>
        IStyleProperty<Padding> IPropertiesWithPadding.Padding
        {
            get
            {
                return this.TableCellPadding;
            }
        }

        /// <summary>
        /// Enumerates the style properties.
        /// </summary>
        /// <returns>The style properties.</returns>
        protected override IEnumerable<IStyleProperty> EnumerateStyleProperties()
        {
            yield return this.Alignment;
            yield return this.BackgroundColor;
            yield return this.ShadingPatternColor;
            yield return this.ShadingPattern;
            yield return this.Borders;
            yield return this.TableCellPadding;
            yield return this.TableCellSpacing;
            yield return this.Indent;
            yield return this.RowBanding;
            yield return this.ColumnBanding;
            yield return this.FlowDirection;
            yield return this.PreferredWidth;
            yield return this.Looks;
            yield return this.LayoutType;
            yield return this.Overlap;
        }

        /// <summary>
        /// Gets specific style property or null.
        /// </summary>
        /// <param name="propertyDefinition">The property definition.</param>
        /// <returns></returns>
        protected override IStyleProperty GetStylePropertyOverride(IStylePropertyDefinition propertyDefinition)
        {
            if (propertyDefinition != DocumentElementPropertiesBase.StyleIdPropertyDefinition &&
               propertyDefinition.StylePropertyType != StylePropertyType.Table)
            {
                return null;
            }

            return stylePropertyGetters[propertyDefinition.GlobalPropertyIndex](this);
        }

        private static void InitStylePropertyGetters()
        {
            if (stylePropertyGetters != null)
            {
                return;
            }

            stylePropertyGetters = new Func<TableProperties, IStyleProperty>[16];
            stylePropertyGetters[DocumentElementPropertiesBase.StyleIdPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.StyleIdProperty; };
            stylePropertyGetters[Table.AlignmentPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.alignmentStyleProperty; };
            stylePropertyGetters[Table.TableCellPaddingPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.tableCellPaddingStyleProperty; };
            stylePropertyGetters[Table.TableCellSpacingPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.tableCellSpacingStyleProperty; };
            stylePropertyGetters[Table.IndentPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.indentStyleProperty; };
            stylePropertyGetters[Table.RowBandingPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.rowBandingStyleProperty; };
            stylePropertyGetters[Table.ColumnBandingPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.columnBandingStyleProperty; };
            stylePropertyGetters[Table.TableBordersPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.tableBordersStyleProperty; };
            stylePropertyGetters[Table.BackgroundColorPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.backgroundColorStyleProperty; };
            stylePropertyGetters[Table.ShadingPatternColorPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.shadingPatternColorStyleProperty; };
            stylePropertyGetters[Table.ShadingPatternPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.shadingPatternStyleProperty; };
            stylePropertyGetters[Table.FlowDirectionPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.flowDirectionStyleProperty; };
            stylePropertyGetters[Table.PreferredWidthPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.preferredWidthStyleProperty; };
            stylePropertyGetters[Table.LooksPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.looksStyleProperty; };
            stylePropertyGetters[Table.LayoutTypePropertyDefinition.GlobalPropertyIndex] = prop => { return prop.layoutTypeStyleProperty; };
            stylePropertyGetters[Table.OverlapPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.overlapStyleProperty; };
        }

        /// <summary>
        /// Initialize the properties.
        /// </summary>
        private void InitProperties()
        {
            this.alignmentStyleProperty = new StyleProperty<Alignment?>(Table.AlignmentPropertyDefinition, this);
            this.tableCellPaddingStyleProperty = new StyleProperty<Padding>(Table.TableCellPaddingPropertyDefinition, this);
            this.tableCellSpacingStyleProperty = new StyleProperty<double?>(Table.TableCellSpacingPropertyDefinition, this);
            this.indentStyleProperty = new StyleProperty<double?>(Table.IndentPropertyDefinition, this);
            this.rowBandingStyleProperty = new StyleProperty<int?>(Table.RowBandingPropertyDefinition, this);
            this.columnBandingStyleProperty = new StyleProperty<int?>(Table.ColumnBandingPropertyDefinition, this);
            this.tableBordersStyleProperty = new StyleProperty<TableBorders>(Table.TableBordersPropertyDefinition, this);
            this.backgroundColorStyleProperty = new StyleProperty<ThemableColor>(Table.BackgroundColorPropertyDefinition, this);
            this.shadingPatternColorStyleProperty = new StyleProperty<ThemableColor>(Table.ShadingPatternColorPropertyDefinition, this);
            this.shadingPatternStyleProperty = new StyleProperty<ShadingPattern?>(Table.ShadingPatternPropertyDefinition, this);
            this.flowDirectionStyleProperty = new LocalProperty<FlowDirection?>(Table.FlowDirectionPropertyDefinition, this);
            this.preferredWidthStyleProperty = new LocalProperty<TableWidthUnit>(Table.PreferredWidthPropertyDefinition, this);
            this.looksStyleProperty = new LocalProperty<TableLooks?>(Table.LooksPropertyDefinition, this);
            this.layoutTypeStyleProperty = new LocalProperty<TableLayoutType?>(Table.LayoutTypePropertyDefinition, this);
            this.overlapStyleProperty = new LocalProperty<bool?>(Table.OverlapPropertyDefinition, this);

            // TODO Implement.
            ////this.floatingPositionStyleProperty = new LocalProperty<FloatingPosition>(Table.FloatingPositionPropertyDefinition, this);

            InitStylePropertyGetters();
        }
    }
}