﻿using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Specifies the height of a <see cref="TableRow"/>. The height may be absolute or relative, depending on the property values.
    /// </summary>
    public class TableRowHeight
    {
        private readonly double value;
        private readonly HeightType type;

        /// <summary>
        /// Initializes a new instance of the <see cref="TableRowHeight" /> class. Useful for creating <see cref="TableRowHeight"/> with <see cref="HeightType.Auto"/>.
        /// </summary>
        /// <param name="type">The type.</param>
        public TableRowHeight(HeightType type)
            : this(type, 0)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TableRowHeight" /> class.
        /// </summary>
        /// <param name="type">The type of the height.</param>
        /// <param name="value">The row height in device independent pixels (1/96 inch).</param>
        public TableRowHeight(HeightType type, double value)
        {
            this.type = type;
            this.value = value;
        }

        /// <summary>
        /// Gets or sets the value.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The value.</value>
        public double Value
        {
            get
            {
                return this.value;
            }
        }

        /// <summary>
        /// Gets or sets the type of the height.
        /// </summary>
        /// <value>The type of the height.</value>
        public HeightType Type
        {
            get
            {
                return this.type;
            }
        }

        /// <summary>
        /// Determines whether the specified table row height are equal.
        /// </summary>
        /// <returns>True if the table row heights are equal.</returns>
        public static bool operator ==(TableRowHeight a, TableRowHeight b)
        {
            // If both are null, or both are same instance, return true.
            if (object.ReferenceEquals(a, b))
            {
                return true;
            }

            // If one is null, but not both, return false.
            if (((object)a == null) || ((object)b == null))
            {
                return false;
            }

            return a.Equals(b);
        }

        /// <summary>
        /// Determines whether the specified table row height are different.
        /// </summary>
        /// <returns>True if the table row heights are different.</returns>
        public static bool operator !=(TableRowHeight a, TableRowHeight b)
        {
            return !(a == b);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// Returns true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            TableRowHeight other = obj as TableRowHeight;
            if (other == null)
            {
                return false;
            }

            return this.Value.Equals(other.Value) && this.Type.Equals(other.Type);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return ObjectExtensions.CombineHashCodes(this.Value.GetHashCode(), this.Type.GetHashCode());
        }

        /// <inheritdoc />
        public override string ToString()
        {
            if (this.Type == HeightType.Auto)
            {
                return this.Type.ToString();
            }
            else
            {
                return string.Format("{0}, {1}", this.Type.ToString(), this.Value.ToString());
            }
        }
    }
}
