﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Defines a set of properties which can be applied to a <see cref="TableRow"/> element.
    /// </summary>
    public sealed class TableRowProperties : DocumentElementPropertiesBase
    {
        private static Func<TableRowProperties, IStyleProperty>[] stylePropertyGetters;

        private StyleProperty<double?> tableCellSpacingStyleProperty;
        private LocalProperty<bool?> repeatOnEveryPageStyleProperty;
        private LocalProperty<TableRowHeight> heightStyleProperty;
        private LocalProperty<bool?> canSplitStyleProperty;

        /// <summary>
        /// Initializes a new instance of the <see cref="TableRowProperties" /> class.
        /// </summary>
        /// <param name="style">The style.</param>
        internal TableRowProperties(Style style)
            : base(style)
        {
            this.InitProperties();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TableRowProperties" /> class.
        /// </summary>
        /// <param name="tableRow">The table row.</param>
        internal TableRowProperties(TableRow tableRow)
            : base(tableRow, false)
        {
            this.InitProperties();
        }

        /// <summary>
        /// Gets style property used to get or set a value indicating the spacing between adjacent cells and the edges of the table.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        public IStyleProperty<double?> TableCellSpacing
        {
            get
            {
                return this.tableCellSpacingStyleProperty;
            }
        }

        /// <summary>
        /// Gets local style property used to get or set a value indicating whether this row should be repeated on every new page. 
        /// This property cannot be derived from a style.
        /// </summary>
        /// <returns>The default value is <c>false</c>.</returns>        
        public IStyleProperty<bool?> RepeatOnEveryPage
        {
            get
            {
                return this.repeatOnEveryPageStyleProperty;
            }
        }

        /// <summary>
        /// Gets local style property used to get or set a value indicating whether the content of the row can be split across multiple pages. 
        /// This property cannot be derived from a style.
        /// </summary>
        /// <returns>The default value is <c>true</c>.</returns>
        public IStyleProperty<bool?> CanSplit
        {
            get
            {
                return this.canSplitStyleProperty;
            }
        }

        /// <summary>
        /// Gets local style property used to get or set the height. 
        /// This property cannot be derived from a style.
        /// </summary>
        /// <value>The height.</value>
        public IStyleProperty<TableRowHeight> Height
        {
            get
            {
                return this.heightStyleProperty;
            }
        }

        /// <summary>
        /// Enumerates the style properties.
        /// </summary>
        /// <returns>The style properties.</returns>
        protected override IEnumerable<IStyleProperty> EnumerateStyleProperties()
        {
            yield return this.TableCellSpacing;
            yield return this.RepeatOnEveryPage;
            yield return this.CanSplit;
            yield return this.Height;
        }

        /// <summary>
        /// Gets specific style property or null.
        /// </summary>
        /// <param name="propertyDefinition">The property definition.</param>
        /// <returns></returns>
        protected override IStyleProperty GetStylePropertyOverride(IStylePropertyDefinition propertyDefinition)
        {
            if (propertyDefinition != DocumentElementPropertiesBase.StyleIdPropertyDefinition &&
                propertyDefinition.StylePropertyType != StylePropertyType.TableRow)
            {
                return null;
            }

            return stylePropertyGetters[propertyDefinition.GlobalPropertyIndex](this);
        }

        private static void InitStylePropertyGetters()
        {
            if (stylePropertyGetters != null)
            {
                return;
            }

            stylePropertyGetters = new Func<TableRowProperties, IStyleProperty>[5];
            stylePropertyGetters[DocumentElementPropertiesBase.StyleIdPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.StyleIdProperty; };
            stylePropertyGetters[TableRow.TableCellSpacingPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.tableCellSpacingStyleProperty; };
            stylePropertyGetters[TableRow.RepeatOnEveryPagePropertyDefinition.GlobalPropertyIndex] = prop => { return prop.repeatOnEveryPageStyleProperty; };
            stylePropertyGetters[TableRow.HeightPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.heightStyleProperty; };
            stylePropertyGetters[TableRow.CanSplitPropertyDefinition.GlobalPropertyIndex] = prop => { return prop.canSplitStyleProperty; };
        }

        /// <summary>
        /// Initialize the properties.
        /// </summary>
        private void InitProperties()
        {
            this.tableCellSpacingStyleProperty = new StyleProperty<double?>(TableRow.TableCellSpacingPropertyDefinition, this);
            this.repeatOnEveryPageStyleProperty = new LocalProperty<bool?>(TableRow.RepeatOnEveryPagePropertyDefinition, this);
            this.heightStyleProperty = new LocalProperty<TableRowHeight>(TableRow.HeightPropertyDefinition, this);
            this.canSplitStyleProperty = new LocalProperty<bool?>(TableRow.CanSplitPropertyDefinition, this);

            InitStylePropertyGetters();
        }
    }
}