﻿using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Defines preferred width of a table or table cell.
    /// </summary>
    public class TableWidthUnit
    {
        private readonly TableWidthUnitType type;
        private readonly double value;

        /// <summary>
        /// Initializes a new instance of the <see cref="TableWidthUnit" /> class.
        /// </summary>
        /// <param name="type">The type.</param>
        public TableWidthUnit(TableWidthUnitType type)
            : this(type, double.NaN)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TableWidthUnit" /> class.
        /// </summary>
        /// <param name="value">The value in device independent pixels (1/96 inch).</param>
        public TableWidthUnit(double value)
            : this(TableWidthUnitType.Fixed, value)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TableWidthUnit" /> class.
        /// </summary>
        /// <param name="type">The type.</param>
        /// <param name="value">The value in device independent pixels (1/96 inch).</param>
        public TableWidthUnit(TableWidthUnitType type, double value)
        {
            this.type = type;
            this.value = value;
        }

        /// <summary>
        /// Gets or sets the unit type type.
        /// </summary>
        /// <value>The unit type.</value>
        public TableWidthUnitType Type
        {
            get
            {
                return this.type;
            }
        }

        /// <summary>
        /// Gets or sets the unit value value.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The unit value.</value>
        public double Value
        {
            get
            {
                return this.value;
            }
        }

        /// <summary>
        /// Determines whether the specified table width units are equal.
        /// </summary>
        /// <returns>True if the table width units are equal.</returns>
        public static bool operator ==(TableWidthUnit a, TableWidthUnit b)
        {
            // If both are null, or both are same instance, return true.
            if (object.ReferenceEquals(a, b))
            {
                return true;
            }

            // If one is null, but not both, return false.
            if (((object)a == null) || ((object)b == null))
            {
                return false;
            }

            return a.Equals(b);
        }

        /// <summary>
        /// Determines whether the specified table width units are different.
        /// </summary>
        /// <returns>True if the table width units are different.</returns>
        public static bool operator !=(TableWidthUnit a, TableWidthUnit b)
        {
            return !(a == b);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// Returns true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            TableWidthUnit other = obj as TableWidthUnit;
            if (other == null)
            {
                return false;
            }

            return this.Type.Equals(other.Type) && this.Value.Equals(other.Value);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return ObjectExtensions.CombineHashCodes(this.Type.GetHashCode(), this.Value.GetHashCode());
        }

        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current
        /// <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
            if (this.Type == TableWidthUnitType.Auto)
            {
                return this.Type.ToString();
            }
            else
            {
                return string.Format("{0}, {1}", this.Type.ToString(), this.Value.ToString());
            }
        }
    }
}