﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Establishes functionality to share style properties between different document elements.
    /// </summary>
    [DebuggerDisplay("Id = {Id}, Type = {StyleType}")]
    public class Style
    {
        private static readonly Dictionary<StylePropertyType, Func<Style, DocumentElementPropertiesBase>> StylePropertyTypeToPropertiesContainerEvaluator;
        private readonly StyleType styleType;
        private string id;
        private string name;
        private string basedOnStyleId;
        private string linkedStyleId;
        private bool isCustom;
        private bool isDefault;
        private CharacterProperties characterProperties;
        private ParagraphProperties paragraphProperties;
        private TableProperties tableProperties;
        private TableRowProperties tableRowProperties;
        private TableCellProperties tableCellProperties;

        static Style()
        {
            StylePropertyTypeToPropertiesContainerEvaluator = new Dictionary<StylePropertyType, Func<Style, DocumentElementPropertiesBase>>()
            {
                { StylePropertyType.Character, s => s.CharacterProperties },
                { StylePropertyType.Paragraph, s => s.ParagraphProperties },
                { StylePropertyType.Table, s => s.TableProperties },
                { StylePropertyType.TableRow, s => s.TableRowProperties },
                { StylePropertyType.TableCell, s => s.TableCellProperties },
                { StylePropertyType.DocumentElement, s => null },
            };
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Style" /> class with corresponding style ID and style type.
        /// </summary>
        /// <param name="id">The ID.</param>
        /// <param name="type">The type.</param>
        public Style(string id, StyleType type)
        {
            this.Id = id;
            this.styleType = type;
            this.UIPriority = DocumentDefaultStyleSettings.UIPriority;
        }

        /// <summary>
        /// Gets or sets the ID of the style.
        /// </summary>
        /// <value>The ID.</value>
        public string Id
        {
            get
            {
                return this.id;
            }
            set
            {
                if (this.Document != null)
                {
                    throw new InvalidOperationException("Cannot change the id of a style that is already added to a document.");
                }
                Guard.ThrowExceptionIfNullOrEmpty(value, "Id");

                this.id = value;
            }
        }

        /// <summary>
        /// Gets or sets the name which will represent the style in any UI. If this property is not set then the id of the style is returned.
        /// </summary>
        /// <value>The name.</value>
        public string Name
        {
            get
            {
                if (string.IsNullOrEmpty(this.name))
                {
                    return this.Id;
                }
                return this.name;
            }
            set
            {
                this.name = value;
            }
        }

        /// <summary>
        /// Gets the document to which the current style belongs to.
        /// </summary>
        /// <value>The document.</value>
        public RadFlowDocument Document { get; internal set; }

        /// <summary>
        /// Gets the type of the style.
        /// </summary>
        /// <value>The type of the style.</value>
        public StyleType StyleType
        {
            get
            {
                return this.styleType;
            }
        }

        /// <summary>
        /// Gets or sets the name of the based on style.
        /// </summary>
        /// <value>The name of the based on.</value>
        public string BasedOnStyleId
        {
            get
            {
                return this.basedOnStyleId;
            }
            set
            {
                this.CheckIsValidBasedOnStyle(value);

                this.basedOnStyleId = value;
            }
        }

        /// <summary>
        /// Gets or sets the next style id.
        /// </summary>
        /// <value>The next style id.</value>
        public string NextStyleId { get; set; }

        /// <summary>
        /// Gets or sets the linked style ID.
        /// </summary>
        /// <value>The linked style ID.</value>
        public string LinkedStyleId
        {
            get
            {
                return this.linkedStyleId;
            }
            set
            {
                // TODO: Add validation if the the style is added to a document.
                this.linkedStyleId = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating if the current style is default.
        /// </summary>
        /// <value>The is default.</value>
        public bool IsDefault
        {
            get
            {
                return this.isDefault;
            }
            set
            {
                if (this.Document != null)
                {
                    throw new InvalidOperationException("Cannot change the IsDefault property of a style that is already added to a document.");
                }
                this.isDefault = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating if the current style is custom. 
        /// The default value is <code>false</code>.
        /// </summary>
        public bool IsCustom
        {
            get
            {
                return this.isCustom;
            }
            set
            {
                this.isCustom = value;
            }
        }

        /// <summary>
        /// Gets or sets value indicating whether the style should be visible in any styles gallery.
        /// </summary>
        public bool IsPrimary { get; set; }

        /// <summary>
        /// Gets or sets the UI priority of the style.
        /// </summary>
        /// <value>The UI priority.</value>
        public int UIPriority { get; set; }

        /// <summary>
        /// Gets the character properties.
        /// </summary>
        /// <value>The character properties.</value>
        public CharacterProperties CharacterProperties
        {
            get
            {
                if (this.characterProperties == null)
                {
                    this.characterProperties = new CharacterProperties(this);
                }
                return this.characterProperties;
            }
        }

        /// <summary>
        /// Gets the paragraph properties.
        /// </summary>
        /// <value>The paragraph properties.</value>
        public ParagraphProperties ParagraphProperties
        {
            get
            {
                if (this.paragraphProperties == null &&
                    (this.StyleType != StyleType.Character || this.StyleType != StyleType.Numbering))
                {
                    this.paragraphProperties = new ParagraphProperties(this);
                }

                return this.paragraphProperties;
            }
        }

        /// <summary>
        /// Gets the table properties.
        /// </summary>
        /// <value>The table properties.</value>
        public TableProperties TableProperties
        {
            get
            {
                if (this.tableProperties == null && this.StyleType == StyleType.Table)
                {
                    this.tableProperties = new TableProperties(this);
                }
                return this.tableProperties;
            }
        }

        /// <summary>
        /// Gets the table row properties.
        /// </summary>
        /// <value>The table row properties.</value>
        public TableRowProperties TableRowProperties
        {
            get
            {
                if (this.tableRowProperties == null && this.StyleType == StyleType.Table)
                {
                    this.tableRowProperties = new TableRowProperties(this);
                }
                return this.tableRowProperties;
            }
        }

        /// <summary>
        /// Gets the table cell properties.
        /// </summary>
        /// <value>The table cell properties.</value>
        public TableCellProperties TableCellProperties
        {
            get
            {
                if (this.tableCellProperties == null && this.StyleType == StyleType.Table)
                {
                    this.tableCellProperties = new TableCellProperties(this);
                }
                return this.tableCellProperties;
            }
        }

        /// <summary>
        /// Gets the property value.
        /// </summary>
        /// <param name="stylePropertyDefinition">The style property definition.</param>
        /// <returns></returns>
        public object GetPropertyValue(IStylePropertyDefinition stylePropertyDefinition)
        {
            IStyleProperty styleProperty = this.GetStyleProperty(stylePropertyDefinition);
            if (styleProperty != null)
            {
                return styleProperty.GetActualValueAsObject();
            }

            return null;
        }

        /// <summary>
        /// Creates deep copy of this style.
        /// </summary>
        /// <returns>The cloned style.</returns>
        public Style Clone()
        {
            Style style = new Style(this.Id, this.StyleType);

            style.Name = this.Name;
            style.BasedOnStyleId = this.BasedOnStyleId;
            style.NextStyleId = this.NextStyleId;
            style.LinkedStyleId = this.LinkedStyleId;
            style.IsDefault = this.IsDefault;
            style.IsCustom = this.IsCustom;
            style.IsPrimary = this.IsPrimary;
            style.UIPriority = this.UIPriority;

            style.CopyStylePropertiesFrom(this);

            return style;
        }

        internal IStyleProperty GetStyleProperty(IStylePropertyDefinition stylePropertyDefinition)
        {
            Func<Style, DocumentElementPropertiesBase> evaluator = StylePropertyTypeToPropertiesContainerEvaluator.GetValueOrNull(stylePropertyDefinition.StylePropertyType);
            if (evaluator != null)
            {
                DocumentElementPropertiesBase propertiesContainer = evaluator(this);
                if (propertiesContainer != null)
                {
                    IStyleProperty styleProperty = propertiesContainer.GetStyleProperty(stylePropertyDefinition);
                    return styleProperty;
                }
            }

            return null;
        }

        internal void CopyStylePropertiesFrom(Style fromStyle)
        {
            if (this.CharacterProperties != null)
            {
                this.CharacterProperties.CopyPropertiesFrom(fromStyle.CharacterProperties);
            }

            if (this.ParagraphProperties != null)
            {
                this.ParagraphProperties.CopyPropertiesFrom(fromStyle.ParagraphProperties);
            }

            if (this.TableProperties != null)
            {
                this.TableProperties.CopyPropertiesFrom(fromStyle.TableProperties);
            }

            if (this.TableRowProperties != null)
            {
                this.TableRowProperties.CopyPropertiesFrom(fromStyle.TableRowProperties);
            }

            if (this.TableCellProperties != null)
            {
                this.TableCellProperties.CopyPropertiesFrom(fromStyle.TableCellProperties);
            }
        }

        private void CheckIsValidBasedOnStyle(string value)
        {
            if (string.IsNullOrEmpty(value) || this.Document == null)
            {
                return;
            }

            Style basedOnStyle = this.Document.StyleRepository.GetStyle(value);
            if (basedOnStyle == null)
            {
                return;
            }

            if (this.StyleType != basedOnStyle.StyleType)
            {
                throw new ArgumentException(string.Format("Style of type {0} can not be based on style of type {1}.", this.StyleType, basedOnStyle.StyleType));
            }

            if (this.HasCycles(basedOnStyle))
            {
                throw new InvalidOperationException("Cyclic style inheritance is not allowed.");
            }
        }

        private bool HasCycles(Style newBasedOn)
        {
            if (newBasedOn == null)
            {
                return false;
            }

            Style curr = newBasedOn;
            while (curr != null)
            {
                if (curr == this)
                {
                    return true;
                }

                string currId = curr.BasedOnStyleId;
                if (string.IsNullOrEmpty(currId))
                {
                    return false;
                }

                curr = this.Document.StyleRepository.GetStyle(currId);
            }

            return false;
        }
    }
}
