﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Flow.Model.Cloning;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Styles
{
    /// <summary>
    /// Represents dynamic style repository.
    /// </summary>
    public class StyleRepository
    {
        private readonly Dictionary<StyleType, Style> defaultStyles = new Dictionary<StyleType, Style>();
        private readonly Dictionary<string, Style> stylesDictionary = new Dictionary<string, Style>();

        /// <summary>
        /// Initializes a new instance of the <see cref="StyleRepository" /> class.
        /// </summary>
        /// <param name="document">The document.</param>
        internal StyleRepository(RadFlowDocument document)
        {
            this.Document = document;
        }

        /// <summary>
        /// Gets the document to which this style collection belongs to.
        /// </summary>
        /// <value>The document.</value>
        public RadFlowDocument Document { get; private set; }

        /// <summary>
        /// Gets an enumeration of the styles added to the repository.
        /// </summary>
        /// <value>The styles.</value>
        public IEnumerable<Style> Styles
        {
            get
            {
                return this.stylesDictionary.Values;
            }
        }

        /// <summary>
        /// Adds the specified style.
        /// </summary>
        /// <param name="style">The style.</param>
        /// <exception cref="T:System.ArgumentNullException"><paramref name="style"/> is part of another style repository.</exception>
        public void Add(Style style)
        {
            this.EnsureStyleRegistered(style);
        }

        /// <summary>
        /// Removes the specified style.
        /// </summary>
        /// <param name="style">The style.</param>
        public void Remove(Style style)
        {
            Guard.ThrowExceptionIfNull(style, "style");
            if (style.Document != this.Document)
            {
                throw new InvalidOperationException("Style should be from the same document repository.");
            }

            this.Remove(style.Id);
        }

        /// <summary>
        /// Removes the specified style by style id.
        /// </summary>
        /// <param name="styleId">The style id.</param>
        public void Remove(string styleId)
        {
            Style style = this.stylesDictionary.GetValueOrNull(styleId);
            if (style == null)
            {
                return;
            }

            this.stylesDictionary.Remove(styleId);
            style.Document = null;

            if (style.IsDefault)
            {
                Style currentDefaultStyle = this.defaultStyles.GetValueOrNull(style.StyleType);
                if (style == currentDefaultStyle)
                {
                    this.defaultStyles.Remove(style.StyleType);
                }
            }
        }

        /// <summary>
        /// Gets a style by id.
        /// </summary>
        /// <param name="styleId">The style id.</param>
        /// <returns>The style with the given id or null if such is not present in the repository.</returns>
        public Style GetStyle(string styleId)
        {
            return this.stylesDictionary.GetValueOrNull(styleId);
        }

        /// <summary>
        /// Gets the default style for a given type.
        /// </summary>
        /// <param name="type">The type.</param>
        /// <returns>The default style or null if such is not present in the repository.</returns>
        public Style GetDefaultStyle(StyleType type)
        {
            return this.defaultStyles.GetValueOrNull(type);
        }

        /// <summary>
        /// Determines whether there is a style with a given id in the repository.
        /// </summary>
        /// <param name="styleId">The style id.</param>
        /// <returns>true if <paramref name="styleId"/> is found in the <see cref="StyleRepository"/>; otherwise, false.</returns>
        public bool Contains(string styleId)
        {
            return this.stylesDictionary.ContainsKey(styleId);
        }

        /// <summary>
        /// Removes all styles from the repository.
        /// </summary>
        public void Clear()
        {
            foreach (string styleName in this.stylesDictionary.Keys.ToList())
            {
                this.Remove(styleName);
            }
        }

        /// <summary>
        /// Adds built-in style. Use IDs from <see cref="BuiltInStyleNames" /> class.
        /// </summary>
        /// <param name="styleId">The ID of the built-in style.</param>
        /// <exception cref="T:System.ArgumentException"><paramref name="styleId"/> is not an ID of a built-in style.</exception>
        /// <returns>The built-in style, added to the repository.</returns>
        public Style AddBuiltInStyle(string styleId)
        {
            if (!BuiltInStyles.IsBuiltInStyle(styleId))
            {
                throw new ArgumentException("The style with ID '{0}' is not built-in style.", styleId);
            }

            Style style = this.GetStyle(styleId);
            if (style != null)
            {
                return style;
            }

            style = BuiltInStyles.GetStyle(styleId);
            if (style != null)
            {
                this.Add(style);
            }

            return style;
        }

        internal void EnsureStyleRegistered(Style style)
        {
            if (style.Document != null && style.Document != this.Document)
            {
                throw new InvalidOperationException("Style from another document added to repository.");
            }

            Style registeredStyle = this.stylesDictionary.GetValueOrNull(style.Id);
            if (registeredStyle == null)
            {
                this.AddInternal(style);
            }
            else if (registeredStyle != style)
            {
                // some times MS Word replace non custom style with custom one.
                this.Remove(registeredStyle.Id);
                this.AddInternal(style);
            }
        }

        internal IEnumerable<Style> GetSortedTopologicallyStyles()
        {
            HashSet<Style> visited = new HashSet<Style>();
            List<Style> list = new List<Style>();

            foreach (Style style in this.Styles)
            {
                TopologicalSortVisit(style, this, visited, list);
            }

            return list;
        }

        internal void AddClonedStylesFrom(StyleRepository fromStyleCollection)
        {
            Guard.ThrowExceptionIfNull(fromStyleCollection, "fromStyleCollection");

            foreach (Style style in fromStyleCollection.Styles)
            {
                this.Add(style.Clone());
            }
        }

        internal void Merge(StyleRepository withStyleRepository, CloneContext cloneContext)
        {
            Guard.ThrowExceptionIfNull(withStyleRepository, "withStyleRepository");
            Guard.ThrowExceptionIfNull(cloneContext, "cloneContext");

            StyleRepositoryMerger merger = new StyleRepositoryMerger(this, withStyleRepository);
            merger.Merge(cloneContext);
        }

        private static void TopologicalSortVisit(Style style, StyleRepository styleCollection, HashSet<Style> visited, List<Style> result)
        {
            if (!visited.Contains(style))
            {
                visited.Add(style);

                if (!string.IsNullOrEmpty(style.BasedOnStyleId))
                {
                    Style basedOnStyle = styleCollection.GetStyle(style.BasedOnStyleId);
                    if (basedOnStyle != null)
                    {
                        TopologicalSortVisit(basedOnStyle, styleCollection, visited, result);
                    }
                }

                if (style.StyleType == StyleType.Character && !string.IsNullOrEmpty(style.LinkedStyleId))
                {
                    Style linkedStyle = styleCollection.GetStyle(style.LinkedStyleId);
                    if (linkedStyle != null)
                    {
                        TopologicalSortVisit(linkedStyle, styleCollection, visited, result);
                    }
                }

                result.Add(style);
            }
        }

        private void AddInternal(Style style)
        {
            if (this.stylesDictionary.ContainsKey(style.Id))
            {
                throw new InvalidOperationException(string.Format("Style with name {0} is already registered", style.Name));
            }

            this.stylesDictionary.Add(style.Id, style);

            if (style.IsDefault)
            {
                this.defaultStyles[style.StyleType] = style;
            }

            style.Document = this.Document;
        }
    }
}
