﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Core.Data;
using Telerik.Windows.Documents.Flow.Model.Cloning;
using Telerik.Windows.Documents.Flow.Model.Collections;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Primitives;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model
{
    /// <summary>
    /// Block-level flow content element that provides a grid-based organization.
    /// </summary>
    public sealed partial class Table : BlockBase, IElementWithStyle, IElementWithProperties
    {
        private readonly TableRowCollection rows;
        private readonly TableProperties properties;
        private bool isTableGridValid;
        private int gridColumnsCount;
        private int gridRowsCount;

        /// <summary>
        /// Initializes a new instance of the <see cref="Table"/> class.
        /// </summary>
        /// <param name="document">The document in which the element will be added.</param>
        public Table(RadFlowDocument document)
            : this(document, 0, 0)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Table" /> class with given number of rows and columns.
        /// </summary>
        /// <param name="document">The document.</param>
        /// <param name="rows">The number of rows.</param>
        /// <param name="columns">The number of columns.</param>
        public Table(RadFlowDocument document, int rows, int columns)
            : base(document)
        {
            Guard.ThrowExceptionIfOutOfRange(0, int.MaxValue, rows, "rows");
            Guard.ThrowExceptionIfOutOfRange(0, int.MaxValue, columns, "columns");

            this.rows = new TableRowCollection(this);
            this.properties = new TableProperties(this);
            this.Shading = new Shading(this.Properties);

            for (int r = 0; r < rows; r++)
            {
                TableRow row = this.Rows.AddTableRow();
                for (int c = 0; c < columns; c++)
                {
                    row.Cells.AddTableCell();
                }
            }
            this.AssureTableGrid();
        }

        /// <summary>
        /// Gets the properties.
        /// </summary>
        /// <value>The properties.</value>
        DocumentElementPropertiesBase IElementWithProperties.Properties
        {
            get
            {
                return this.properties;
            }
        }

        /// <summary>
        /// Gets the styling properties of this element.
        /// </summary>
        /// <value>The properties.</value>
        public TableProperties Properties
        {
            get
            {
                return this.properties;
            }
        }

        /// <summary>
        /// Gets the rows of the table.
        /// </summary>
        /// <value>The rows.</value>
        public TableRowCollection Rows
        {
            get
            {
                return this.rows;
            }
        }

        /// <summary>
        /// Gets or sets the style id.
        /// </summary>
        /// <value>The style id.</value>
        public string StyleId
        {
            get
            {
                return this.Properties.StyleId;
            }
            set
            {
                this.Properties.StyleId = value;
            }
        }

        /// <summary>
        /// Gets or sets the alignment of the table.
        /// </summary>
        /// <returns>The default value is <c>Left</c>.</returns>
        public Alignment Alignment
        {
            get
            {
                return this.Properties.Alignment.GetActualValue().Value;
            }
            set
            {
                this.Properties.Alignment.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the borders of the table.
        /// </summary>
        /// <value>The borders.</value>
        public TableBorders Borders
        {
            get
            {
                return this.Properties.Borders.GetActualValue();
            }
            set
            {
                this.Properties.Borders.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets the shading which shall be applied to the extents of the table.
        /// </summary>
        /// <value>The shading.</value>
        public Shading Shading { get; private set; }

        /// <summary>
        /// Gets the number of columns in the table grid.
        /// </summary>
        /// <value>The grid columns count.</value>
        public int GridColumnsCount
        {
            get
            {
                this.AssureTableGrid();
                return this.gridColumnsCount;
            }
        }

        /// <summary>
        /// Gets the number of rows in the table grid.
        /// </summary>
        /// <value>The grid rows count.</value>
        public int GridRowsCount
        {
            get
            {
                this.AssureTableGrid();
                return this.gridRowsCount;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating the spacing between adjacent cells and the edges of the table.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The table cell spacing.</value>
        public double TableCellSpacing
        {
            get
            {
                return this.Properties.TableCellSpacing.GetActualValue().Value;
            }
            set
            {
                this.Properties.TableCellSpacing.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets a value indicating whether there is cell spacing in the table.
        /// </summary>
        /// <value>True if there is cell spacing.</value>
        public bool HasCellSpacing
        {
            get
            {
                return this.TableCellSpacing > 0;
            }
        }

        /// <summary>
        /// Gets or sets default padding of the cells inside the table.
        /// </summary>
        /// <value>The table cell padding.</value>
        public Padding TableCellPadding
        {
            get
            {
                return this.Properties.TableCellPadding.GetActualValue();
            }
            set
            {
                this.Properties.TableCellPadding.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets a value which shall be added before the leading edge of the table.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The default value is <c>0</c>.</value>
        public double Indent
        {
            get
            {
                return this.Properties.Indent.GetActualValue().Value;
            }
            set
            {
                this.Properties.Indent.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets the flow direction of cells inside the table.
        /// </summary>
        /// <returns>The default value is <c>LeftToRight</c>.</returns>
        public FlowDirection FlowDirection
        {
            get
            {
                return this.Properties.FlowDirection.GetActualValue().Value;
            }
            set
            {
                this.Properties.FlowDirection.LocalValue = value;
            }
        }

        // TODO Implement.
        /////// <summary>
        /////// Gets or sets the floating position of the table.
        /////// </summary>
        /////// <value>The floating position.</value>
        ////public FloatingPosition FloatingPosition
        ////{
        ////    get
        ////    {
        ////        return this.Properties.FloatingPosition.GetActualValue();
        ////    }
        ////    set
        ////    {
        ////        this.Properties.FloatingPosition.SetValueAsObject(value);
        ////    }
        ////}

        /// <summary>
        /// Gets or sets the width of the preferred.
        /// </summary>
        /// <value>The width of the preferred.</value>
        public TableWidthUnit PreferredWidth
        {
            get
            {
                return this.Properties.PreferredWidth.GetActualValue();
            }
            set
            {
                this.Properties.PreferredWidth.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the value indicating which components of the conditional style should be applied if such exists.
        /// </summary>
        /// <returns>The default value is <c>BandedRows | BandedColumns</c>.</returns>
        public TableLooks Looks
        {
            get
            {
                return this.Properties.Looks.GetActualValue().Value;
            }
            set
            {
                this.Properties.Looks.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating which algorithm shall be used to lay out the content of the table.
        /// </summary>
        /// <returns>The default value is <c>AutoFit</c>.</returns>
        public TableLayoutType LayoutType
        {
            get
            {
                return this.Properties.LayoutType.GetActualValue().Value;
            }
            set
            {
                this.Properties.LayoutType.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether this floating table shall allow other floating tables to overlap its extents.
        /// </summary>
        /// <returns>The default value is <c>true</c>.</returns>
        public bool Overlap
        {
            get
            {
                return this.Properties.Overlap.GetActualValue().Value;
            }
            set
            {
                this.Properties.Overlap.LocalValue = value;
            }
        }

        internal override IEnumerable<DocumentElementBase> Children
        {
            get
            {
                return this.rows;
            }
        }

        internal override DocumentElementType Type
        {
            get
            {
                return DocumentElementType.Table;
            }
        }

        /// <summary>
        /// Creates deep copy of this document element, associated to the current document.
        /// </summary>
        /// <returns>The cloned element.</returns>
        public Table Clone()
        {
            return this.CloneInternal(null);
        }

        /// <summary>
        /// Creates deep copy of this document element, associated to <paramref name="document"/>.
        /// </summary>
        /// <param name="document">The document to which the cloned element should be associated.</param>
        /// <returns>The cloned element.</returns>
        public Table Clone(RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");
            return this.CloneInternal(document);
        }

        internal override DocumentElementBase CloneCore(CloneContext cloneContext)
        {
            Guard.ThrowExceptionIfNull(cloneContext, "cloneContext");

            Table table = new Table(cloneContext.Document);
            table.Properties.CopyPropertiesFrom(this.Properties);

            if (cloneContext.RenamedStyles.ContainsKey(table.StyleId))
            {
                table.StyleId = cloneContext.RenamedStyles[table.StyleId];
            }

            table.Rows.AddClonedChildrenFrom(this.Rows, cloneContext);

            return table;
        }

        /// <summary>
        /// Invalidates the table grid.
        /// </summary>
        internal void InvalidateTableGrid()
        {
            this.isTableGridValid = false;
        }

        /// <summary>
        /// Assures the table grid is valid.
        /// </summary>
        internal void AssureTableGrid()
        {
            if (!this.isTableGridValid)
            {
                this.RecalculateGridIndexes();
                this.isTableGridValid = true;
            }
        }

        /// <summary>
        /// Called when child element is added.
        /// </summary>
        /// <param name="child">The child element.</param>
        protected override void OnChildAdded(DocumentElementBase child)
        {
            base.OnChildAdded(child);
            this.InvalidateTableGrid();
        }

        /// <summary>
        /// Called when child element is removed.
        /// </summary>
        /// <param name="child">The child element.</param>
        protected override void OnChildRemoved(DocumentElementBase child)
        {
            base.OnChildRemoved(child);
            this.InvalidateTableGrid();
        }

        private void RecalculateGridIndexes()
        {
            ExpandoList<ExpandoList<bool>> visited = new ExpandoList<ExpandoList<bool>>();

            int cellIndex = 0;
            int rowIndex = 0;
            int maxColumnIndex = 0;
            int maxRowIndex = 0;
            foreach (TableRow row in this.Rows)
            {
                cellIndex = 0;
                foreach (TableCell cell in row.Cells)
                {
                    if (visited[rowIndex] == null)
                    {
                        visited[rowIndex] = new ExpandoList<bool>();
                    }

                    // Find the first empty gap in the grid.
                    while (visited[rowIndex][cellIndex])
                    {
                        cellIndex++;
                    }

                    // Mark the spread of the current cell.
                    for (int i = 0; i < cell.ColumnSpan; i++)
                    {
                        for (int j = 0; j < cell.RowSpan; j++)
                        {
                            if (visited[rowIndex + j] == null)
                            {
                                visited[rowIndex + j] = new ExpandoList<bool>();
                            }

                            visited[rowIndex + j][cellIndex + i] = true;
                        }
                    }

                    cell.GridRowIndex = rowIndex;
                    cell.GridColumnIndex = cellIndex;
                    maxColumnIndex = Math.Max(maxColumnIndex, cellIndex + cell.ColumnSpan - 1);
                    maxRowIndex = Math.Max(maxRowIndex, rowIndex + cell.RowSpan - 1);

                    cellIndex++;
                }

                row.GridRowIndex = rowIndex;
                rowIndex++;
            }

            this.gridColumnsCount = maxColumnIndex + 1;
            this.gridRowsCount = maxRowIndex + 1;
        }

        private Table CloneInternal(RadFlowDocument document = null)
        {
            return (Table)this.CloneCore(new CloneContext(document ?? this.Document));
        }
    }
}
