﻿using Telerik.Windows.Documents.Flow.Model.Cloning;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Primitives;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model
{
    /// <summary>
    /// A flow document element that defines a cell of content within a <see cref="Table"/>.
    /// </summary>
    public sealed partial class TableCell : BlockContainerBase, IElementWithProperties
    {
        private readonly TableCellProperties properties;
        private int gridRowIndex;
        private int gridColumnIndex;

        /// <summary>
        /// Initializes a new instance of the <see cref="TableCell"/> class.
        /// </summary>
        /// <param name="document">The document in which the element will be added.</param>
        public TableCell(RadFlowDocument document)
            : base(document)
        {
            this.properties = new TableCellProperties(this);
            this.Shading = new Shading(this.properties);
        }

        /// <summary>
        /// Gets the properties.
        /// </summary>
        /// <value>The properties.</value>
        DocumentElementPropertiesBase IElementWithProperties.Properties
        {
            get
            {
                return this.properties;
            }
        }

        /// <summary>
        /// Gets the styling properties of this element.
        /// </summary>
        /// <value>The properties.</value>
        public TableCellProperties Properties
        {
            get
            {
                return this.properties;
            }
        }

        /// <summary>
        /// Gets the parent row of the cell.
        /// </summary>
        /// <value>The row.</value>
        public TableRow Row
        {
            get
            {
                return (TableRow)this.Parent;
            }
        }

        /// <summary>
        /// Gets the parent table of the cell.
        /// </summary>
        /// <value>The table.</value>
        public Table Table
        {
            get
            {
                if (this.Row == null)
                {
                    return null;
                }

                return this.Row.Table;
            }
        }

        /// <summary>
        /// Gets or sets the borders.
        /// </summary>
        /// <value>The borders.</value>
        public TableCellBorders Borders
        {
            get
            {
                return this.Properties.Borders.GetActualValue();
            }
            set
            {
                this.Properties.Borders.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the shading which shall be applied to the extents of the cell.
        /// </summary>
        /// <value>The shading.</value>
        public Shading Shading { get; private set; }

        /// <summary>
        /// Gets or sets the padding.
        /// </summary>
        /// <value>The table cell padding.</value>
        public Padding Padding
        {
            get
            {
                return this.Properties.Padding.GetActualValue();
            }
            set
            {
                this.Properties.Padding.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the column span.
        /// </summary>
        /// <returns>The default value is <c>1</c>.</returns>
        public int ColumnSpan
        {
            get
            {
                return this.Properties.ColumnSpan.GetActualValue().Value;
            }
            set
            {
                this.InvalidateTableGrid();
                this.Properties.ColumnSpan.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the row span.
        /// </summary>
        /// <returns>The default value is <c>1</c>.</returns>        
        public int RowSpan
        {
            get
            {
                return this.Properties.RowSpan.GetActualValue().Value;
            }
            set
            {
                this.InvalidateTableGrid();
                this.Properties.RowSpan.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether cell marker will be ignored when row height is calculated.
        /// </summary>
        /// <returns>The default value is <c>false</c>.</returns>
        public bool IgnoreCellMarkerInRowHeightCalculation
        {
            get
            {
                return this.Properties.IgnoreCellMarkerInRowHeightCalculation.GetActualValue().Value;
            }
            set
            {
                this.Properties.IgnoreCellMarkerInRowHeightCalculation.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the content can be wrapped during the table layout.
        /// </summary>
        /// <returns>The default value is <c>true</c>.</returns>
        public bool CanWrapContent
        {
            get
            {
                return this.Properties.CanWrapContent.GetActualValue().Value;
            }
            set
            {
                this.Properties.CanWrapContent.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the preferred width.
        /// </summary>
        /// <value>The preferred width.</value>
        public TableWidthUnit PreferredWidth
        {
            get
            {
                return this.Properties.PreferredWidth.GetActualValue();
            }
            set
            {
                this.Properties.PreferredWidth.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the vertical alignment.
        /// </summary>
        /// <returns>The default value is <c>Top</c>.</returns>
        public VerticalAlignment VerticalAlignment
        {
            get
            {
                return this.Properties.VerticalAlignment.GetActualValue().Value;
            }
            set
            {
                this.Properties.VerticalAlignment.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the text direction.
        /// </summary>
        /// <value>The text direction.</value>
        public TextDirection TextDirection
        {
            get
            {
                return this.Properties.TextDirection.GetActualValue();
            }
            set
            {
                this.Properties.TextDirection.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets the column index of the cell in the table grid.
        /// </summary>
        /// <value>The column index of the cell in the table grid.</value>
        public int GridColumnIndex
        {
            get
            {
                this.AssureTableGrid();

                return this.gridColumnIndex;
            }
            internal set
            {
                this.gridColumnIndex = value;
            }
        }

        /// <summary>
        /// Gets the row index of the cell in the table grid.
        /// </summary>
        /// <value>The row index of the cell in the table grid.</value>
        public int GridRowIndex
        {
            get
            {
                this.AssureTableGrid();

                return this.gridRowIndex;
            }
            internal set
            {
                this.gridRowIndex = value;
            }
        }

        /// <summary>
        /// Gets the type of the element.
        /// </summary>
        /// <value>The type.</value>
        internal override DocumentElementType Type
        {
            get
            {
                return DocumentElementType.TableCell;
            }
        }

        /// <summary>
        /// Creates deep copy of this document element, associated to the current document.
        /// </summary>
        /// <returns>The cloned element.</returns>
        public TableCell Clone()
        {
            return this.CloneInternal(null);
        }

        /// <summary>
        /// Creates deep copy of this document element, associated to <paramref name="document"/>.
        /// </summary>
        /// <param name="document">The document to which the cloned element should be associated.</param>
        /// <returns>The cloned element.</returns>
        public TableCell Clone(RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");
            return this.CloneInternal(document);
        }

        internal override DocumentElementBase CloneCore(CloneContext cloneContext)
        {
            Guard.ThrowExceptionIfNull(cloneContext, "cloneContext");

            TableCell tableCell = new TableCell(cloneContext.Document);
            tableCell.Properties.CopyPropertiesFrom(this.Properties);
            tableCell.Blocks.AddClonedChildrenFrom(this.Blocks, cloneContext);

            return tableCell;
        }

        private void AssureTableGrid()
        {
            if (this.Table != null)
            {
                this.Table.AssureTableGrid();
            }
        }

        /// <summary>
        /// Invalidates the table grid.
        /// </summary>
        private void InvalidateTableGrid()
        {
            if (this.Row != null)
            {
                this.Row.InvalidateTableGrid();
            }
        }

        private TableCell CloneInternal(RadFlowDocument document = null)
        {
            return (TableCell)this.CloneCore(new CloneContext(document ?? this.Document));
        }
    }
}