﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Flow.Model.Cloning;
using Telerik.Windows.Documents.Flow.Model.Collections;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model
{
    /// <summary>
    /// A flow content element that defines a row within a <see cref="Table"/>.
    /// </summary>
    public sealed partial class TableRow : DocumentElementBase, IElementWithProperties
    {
        private readonly TableRowProperties properties;
        private readonly TableCellCollection cells;

        private int gridRowIndex;

        /// <summary>
        /// Initializes a new instance of the <see cref="TableRow"/> class.
        /// </summary>
        /// <param name="document">The document in which the element will be added.</param>
        public TableRow(RadFlowDocument document)
            : base(document)
        {
            this.cells = new TableCellCollection(this);
            this.properties = new TableRowProperties(this);
        }

        /// <summary>
        /// Gets the properties.
        /// </summary>
        /// <value>The properties.</value>
        DocumentElementPropertiesBase IElementWithProperties.Properties
        {
            get
            {
                return this.properties;
            }
        }

        /// <summary>
        /// Gets the styling properties of this element.
        /// </summary>
        /// <value>The properties.</value>
        public TableRowProperties Properties
        {
            get
            {
                return this.properties;
            }
        }

        /// <summary>
        /// Gets the cells in this row.
        /// </summary>
        /// <value>The cells.</value>
        public TableCellCollection Cells
        {
            get
            {
                return this.cells;
            }
        }

        /// <summary>
        /// Gets the parent table of the row.
        /// </summary>
        /// <value>The table.</value>
        public Table Table
        {
            get
            {
                return (Table)this.Parent;
            }
        }

        /// <summary>
        /// Gets the index of the row in the table grid.
        /// </summary>
        /// <value>The index of the row in the table grid.</value>
        public int GridRowIndex
        {
            get
            {
                if (this.Table != null)
                {
                    this.Table.AssureTableGrid();
                }

                return this.gridRowIndex;
            }
            internal set
            {
                this.gridRowIndex = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating the spacing between adjacent cells and the edges of the table.
        /// The value is in device independent pixels (1/96 inch).
        /// </summary>
        /// <value>The default value is <c>0</c>.</value>
        public double TableCellSpacing
        {
            get
            {
                return this.Properties.TableCellSpacing.GetActualValue().Value;
            }
            set
            {
                this.Properties.TableCellSpacing.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether this row should be repeated on every new page.
        /// </summary>
        /// <value>The default value is <c>false</c>.</value>
        public bool RepeatOnEveryPage
        {
            get
            {
                return this.Properties.RepeatOnEveryPage.GetActualValue().Value;
            }
            set
            {
                this.Properties.RepeatOnEveryPage.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the content of the row can be split across multiple pages.
        /// </summary>
        /// <returns>The default value is <c>true</c>.</returns>
        public bool CanSplit
        {
            get
            {
                return this.Properties.CanSplit.GetActualValue().Value;
            }
            set
            {
                this.Properties.CanSplit.LocalValue = value;
            }
        }

        /// <summary>
        /// Gets or sets the height.
        /// </summary>
        /// <value>The height.</value>
        public TableRowHeight Height
        {
            get
            {
                return this.Properties.Height.GetActualValue();
            }
            set
            {
                this.Properties.Height.LocalValue = value;
            }
        }

        internal override IEnumerable<DocumentElementBase> Children
        {
            get
            {
                return this.Cells;
            }
        }

        internal override DocumentElementType Type
        {
            get
            {
                return DocumentElementType.TableRow;
            }
        }

        /// <summary>
        /// Creates deep copy of this document element, associated to the current document.
        /// </summary>
        /// <returns>The cloned element.</returns>
        public TableRow Clone()
        {
            return this.CloneInternal(null);
        }

        /// <summary>
        /// Creates deep copy of this document element, associated to <paramref name="document"/>.
        /// </summary>
        /// <param name="document">The document to which the cloned element should be associated.</param>
        /// <returns>The cloned element.</returns>
        public TableRow Clone(RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");
            return this.CloneInternal(document);
        }

        internal override DocumentElementBase CloneCore(CloneContext cloneContext)
        {
            Guard.ThrowExceptionIfNull(cloneContext, "cloneContext");

            TableRow tableRow = new TableRow(cloneContext.Document);
            tableRow.Properties.CopyPropertiesFrom(this.Properties);
            tableRow.Cells.AddClonedChildrenFrom(this.Cells, cloneContext);

            return tableRow;
        }

        /// <summary>
        /// Invalidates the table grid.
        /// </summary>
        internal void InvalidateTableGrid()
        {
            if (this.Table != null)
            {
                this.Table.InvalidateTableGrid();
            }
        }

        /// <summary>
        /// Called when child element is the added.
        /// </summary>
        /// <param name="child">The child element.</param>
        protected override void OnChildAdded(DocumentElementBase child)
        {
            base.OnChildAdded(child);
            this.InvalidateTableGrid();
        }

        /// <summary>
        /// Called when child element is removed.
        /// </summary>
        /// <param name="child">The child element.</param>
        protected override void OnChildRemoved(DocumentElementBase child)
        {
            base.OnChildRemoved(child);
            this.InvalidateTableGrid();
        }

        private TableRow CloneInternal(RadFlowDocument document = null)
        {
            return (TableRow)this.CloneCore(new CloneContext(document ?? this.Document));
        }
    }
}
