﻿using System;
using System.IO;
using System.Net;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Media
{
    /// <summary>
    /// Encapsulates data needed for creation of an image taken from Uri source.
    /// </summary>
    public sealed class UriImageSource : ImageSource
    {
        private readonly Uri uri;
        private byte[] data;
        private string extension;

        /// <summary>
        /// Initializes a new instance of the <see cref="UriImageSource" /> class.
        /// </summary>
        /// <param name="uri">The URI of the image.</param>
        public UriImageSource(Uri uri)
            : this(uri, string.Empty)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="UriImageSource" /> class.
        /// </summary>
        /// <param name="uri">The URI of the image.</param>
        /// <param name="extension">The extension of the image.</param>
        public UriImageSource(Uri uri, string extension)
        {
            Guard.ThrowExceptionIfNull(uri, "uri");

            this.uri = uri;
            this.extension = string.IsNullOrEmpty(extension) ? string.Empty : PathExtension.StripExtension(extension);
        }

        internal UriImageSource(Uri uri, byte[] data)
        {
            Guard.ThrowExceptionIfNull(uri, "uri");

            this.uri = uri;
            this.data = data;
            this.extension = string.Empty;
        }

        /// <summary>
        /// Gets the URI of the image.
        /// </summary>
        /// <value>The URI of the image.</value>
        public Uri Uri
        {
            get
            {
                return this.uri;
            }
        }

        /// <summary>
        /// Gets the data.
        /// </summary>
        /// <value>The data.</value>
        public override byte[] Data
        {
            get
            {
                if (this.data == null)
                {
                    this.GetImageData();
                }

                return this.data;
            }
        }

        /// <summary>
        /// Gets the extension.
        /// </summary>
        /// <value>The extension.</value>
        public override string Extension
        {
            get
            {
                this.UpdateExtensionFromUri();

                if (string.IsNullOrEmpty(this.extension))
                {
                    this.GetImageData();
                }

                return this.extension;
            }
        }

        internal override ImageSource Clone()
        {
            UriImageSource clone = new UriImageSource(this.uri, this.data);
            clone.extension = this.extension;

            return clone;
        }

        private static string StripExtensionFromResponseHeader(string contentType)
        {
            int index = contentType.IndexOf('/');
            if (index == -1 || index == contentType.Length - 1)
            {
                throw new NotSupportedException(string.Format("The Content-Type: {0}", contentType));
            }

            string result = contentType.Substring(index + 1);
            return result;
        }

        private void GetImageData()
        {
            using (WebClient client = new WebClient())
            {
                this.data = client.DownloadData(this.Uri);

                this.UpdateExtensionFromUri();
                this.UpdateExtensionFromResponseHeaders(client.ResponseHeaders);
            }
        }

        private void UpdateExtensionFromResponseHeaders(WebHeaderCollection responseHeaders)
        {
            if (string.IsNullOrEmpty(this.extension))
            {
                for (int i = 0; i < responseHeaders.Count; i++)
                {
                    if (responseHeaders.Keys[i].ToString().ToLowerInvariant() == "Content-Type".ToLowerInvariant())
                    {
                        this.extension = StripExtensionFromResponseHeader(responseHeaders[i]);
                        break;
                    }
                }
            }
        }

        private void UpdateExtensionFromUri()
        {
            if (string.IsNullOrEmpty(this.extension))
            {
                string extension = Path.GetExtension(this.Uri.OriginalString);
                if (!string.IsNullOrEmpty(extension))
                {
                    this.extension = PathExtension.StripExtension(extension);
                }
            }
        }
    }
}