﻿using System;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Watermarks
{
    /// <summary>
    /// Represents watermark which can be applied to pages in a <see cref="Section"/>.
    /// </summary>
    public class Watermark
    {
        private readonly ImageWatermarkSettings imageSettings;
        private readonly TextWatermarkSettings textSettings;

        private RadFlowDocument document;

        /// <summary>
        /// Initializes a new instance of the <see cref="Watermark" /> class which represents image watermark.
        /// </summary>
        /// <param name="imageWatermarkSettings">The image watermark settings.</param>
        public Watermark(ImageWatermarkSettings imageWatermarkSettings)
        {
            Guard.ThrowExceptionIfNull(imageWatermarkSettings, "imageWatermarkSettings");

            this.imageSettings = imageWatermarkSettings;
            this.imageSettings.Watermark = this;
            this.WatermarkType = WatermarkType.Image;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Watermark" /> class which represents text watermark.
        /// </summary>
        /// <param name="textWatermarkSettings">The text watermark settings.</param>
        public Watermark(TextWatermarkSettings textWatermarkSettings)
        {
            Guard.ThrowExceptionIfNull(textWatermarkSettings, "textWatermarkSettings");

            this.textSettings = textWatermarkSettings;
            this.textSettings.Watermark = this;
            this.WatermarkType = WatermarkType.Text;
        }

        /// <summary>
        /// Gets the type of the watermark.
        /// </summary>
        /// <value>The type.</value>
        public WatermarkType WatermarkType { get; private set; }

        /// <summary>
        /// Gets the image watermark settings. This property is initialized only if the watermark is image watermark.
        /// </summary>
        /// <value>The image settings.</value>
        public ImageWatermarkSettings ImageSettings
        {
            get
            {
                return this.imageSettings;
            }
        }

        /// <summary>
        /// Gets the text watermark settings. This property is initialized only if the watermark is text watermark.
        /// </summary>
        /// <value>The text settings.</value>
        public TextWatermarkSettings TextSettings
        {
            get
            {
                return this.textSettings;
            }
        }

        /// <summary>
        /// Gets the document associated with the watermark.
        /// </summary>
        /// <value>The document.</value>
        public RadFlowDocument Document
        {
            get
            {
                return this.document;
            }
            internal set
            {
                if (value != null)
                {
                    Guard.ThrowExceptionIfNotNull(this.document, "document");
                }

                if (this.document != null && this.ImageSettings != null)
                {
                    this.ImageSettings.ReleaseImageSourceFromDocument();
                }

                this.document = value;

                if (this.document != null && this.ImageSettings != null)
                {
                    this.ImageSettings.RegisterImageSourceInDocument();
                }
            }
        }

        /// <summary>
        /// Creates a deep copy of this element, not associated with this document.
        /// </summary>
        /// <returns>A clone of the watermark.</returns>
        public Watermark Clone()
        {
            if (this.WatermarkType == WatermarkType.Image)
            {
                ImageWatermarkSettings clonedSettings = this.ImageSettings.Clone();
                return new Watermark(clonedSettings);
            }
            else if (this.WatermarkType == WatermarkType.Text)
            {
                TextWatermarkSettings clonedSettings = this.TextSettings.Clone();
                return new Watermark(clonedSettings);
            }
            else
            {
                throw new NotSupportedException("Unsupported watermark.");
            }
        }
    }
}
