﻿using System.Collections;
using System.Collections.Generic;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.Model.Watermarks
{
    /// <summary>
    /// Represents a collection of watermark elements.
    /// </summary>
    public class WatermarkCollection : IEnumerable<Watermark>
    {
        private readonly RadFlowDocument document;
        private readonly List<Watermark> innerList;

        /// <summary>
        /// Initializes a new instance of the <see cref="WatermarkCollection" /> class.
        /// </summary>
        /// <param name="document">The document.</param>
        public WatermarkCollection(RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");

            this.document = document;
            this.innerList = new List<Watermark>();
        }

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Flow.Model.Watermarks.Watermark" /> at the specified index.
        /// </summary>
        /// <value>The watermark.</value>
        public Watermark this[int index]
        {
            get
            {
                return this.innerList[index];
            }
        }

        /// <summary>
        /// Adds the specified watermark to the collection.
        /// </summary>
        /// <param name="watermark">The watermark.</param>
        public void Add(Watermark watermark)
        {
            Guard.ThrowExceptionIfNull(watermark, "watermark");
            Guard.ThrowExceptionIfNotNull(watermark.Document, "watermark.Document");

            watermark.Document = this.document;
            this.innerList.Add(watermark);
        }

        /// <summary>
        /// Removes the specified watermark from the collection.
        /// </summary>
        /// <param name="watermark">The watermark.</param>
        /// <returns></returns>
        public bool Remove(Watermark watermark)
        {
            Guard.ThrowExceptionIfNull(watermark, "watermark");

            watermark.Document = null;
            return this.innerList.Remove(watermark);
        }

        /// <summary>
        /// Clears this collection.
        /// </summary>
        public void Clear()
        {
            foreach (Watermark watermark in this.innerList)
            {
                watermark.Document = null;
            }

            this.innerList.Clear();
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        public IEnumerator GetEnumerator()
        {
            return this.innerList.GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        IEnumerator<Watermark> IEnumerable<Watermark>.GetEnumerator()
        {
            return this.innerList.GetEnumerator();
        }

        internal void AddClonedChildrenFrom(WatermarkCollection fromCollection)
        {
            Guard.ThrowExceptionIfNull(fromCollection, "fromCollection");

            foreach (Watermark watermark in fromCollection)
            {
                Watermark clonedElement = watermark.Clone();
                if (clonedElement != null)
                {
                    this.Add(clonedElement);
                }
            }
        }
    }
}
