﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace Telerik.Windows.Documents.Flow.TextSearch.Core
{
    /// <summary>
    /// Represents a collection of intervals mapped to objects.
    /// </summary>
    /// <typeparam name="TKey">The type of the T key.</typeparam>
    /// <typeparam name="V">The type of the V.</typeparam>
    internal class IntervalMap<TKey, V> : IEnumerable
        where TKey : IComparable<TKey>
    {
        private readonly Dictionary<Interval<TKey>, V> intervalToValueDictionary;

        public IntervalMap()
        {
            this.intervalToValueDictionary = new Dictionary<Interval<TKey>, V>();
        }

        internal Interval<TKey> LastIntervalCached { get; private set; }

        /// <summary>
        /// Adds the specified interval mapped to the value.
        /// </summary>
        /// <param name="intervalStart">The interval start.</param>
        /// <param name="intervalEnd">The interval end.</param>
        /// <param name="mappedValue">The mapped value.</param>
        public void Add(TKey intervalStart, TKey intervalEnd, V mappedValue)
        {
            Interval<TKey> interval = new Interval<TKey>(intervalStart, intervalEnd);

            this.intervalToValueDictionary.Add(interval, mappedValue);
        }

        /// <summary>
        /// Gets the value from an interval point.
        /// </summary>
        /// <param name="intervalPoint">The interval point.</param>
        /// <returns>The value which is mapped to the interval from the interval point.</returns>
        public V GetValueFromIntervalPoint(TKey intervalPoint)
        {
            Interval<TKey> currentInterval;

            int collectionIndex = 0;
            int leftBoundIndex = 0;
            int rightBoundIndex = this.intervalToValueDictionary.Keys.Count - 1;
            while (leftBoundIndex <= rightBoundIndex)
            {
                collectionIndex = (leftBoundIndex + rightBoundIndex) / 2;
                currentInterval = this.intervalToValueDictionary.ElementAt(collectionIndex).Key;

                if (currentInterval.Contains(intervalPoint))
                {
                    this.LastIntervalCached = currentInterval;
                    return this.intervalToValueDictionary[currentInterval];
                }
                else if (currentInterval.CompareTo(intervalPoint) == 1)
                {
                    rightBoundIndex = collectionIndex - 1;
                }
                else
                {
                    leftBoundIndex = collectionIndex + 1;
                }
            }

            return default(V);
        }

        public IEnumerator GetEnumerator()
        {
            return this.intervalToValueDictionary.GetEnumerator();
        }

        public void Clear()
        {
            this.intervalToValueDictionary.Clear();
        }
    }
}