﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;
using Telerik.Windows.Documents.Flow.TextSearch.Core;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.TextSearch
{
    /// <summary>
    /// Represents a handler which replace character properties of <see cref="Telerik.Windows.Documents.Flow.Model.Run" /> elements with replace text.
    /// </summary>
    internal class ReplacePropertiesHandler : ReplaceHandlerBase
    {
        private readonly Action<CharacterProperties> replacePropertiesAction;
        private readonly HashSet<Run> runsToRemove;
        private readonly IntervalMap<int, Run> runsToMerge;
        private readonly Queue<string> replaceTexts;

        private Run lastMatchedRun;
        private Paragraph lastRunParent;
        private int paragraphTextLength = 0;
        private string lastReplaceText = string.Empty;
        private int lastRunIndexInParagraph;

        /// <summary>
        /// Initializes a new instance of the <see cref="ReplacePropertiesHandler" /> class.
        /// </summary>
        /// <param name="replacePropertiesAction">The replace character properties action.</param>
        public ReplacePropertiesHandler(Action<CharacterProperties> replacePropertiesAction)
        {
            Guard.ThrowExceptionIfNull(replacePropertiesAction, "replacePropertiesAction");

            this.replacePropertiesAction = replacePropertiesAction;

            this.runsToRemove = new HashSet<Run>();
            this.runsToMerge = new IntervalMap<int, Run>();
            this.replaceTexts = new Queue<string>();

            this.lastMatchedRun = null;
        }

        private bool HasReplaceTextsQueued
        {
            get
            {
                return this.replaceTexts.Count != 0;
            }
        }

        /// <summary>
        /// Replaces matched text in a run.
        /// </summary>
        /// <param name="runTextMatch">The match info in a text of a run.</param>
        protected internal override void ReplaceMatch(RunTextMatch runTextMatch)
        {
            this.lastReplaceText = runTextMatch.FullMatchText;
            this.replaceTexts.Enqueue(runTextMatch.FullMatchText);

            ReplaceTextHandler.ReplaceTextInternal(runTextMatch, runTextMatch.FullMatchText);

            if (this.lastMatchedRun != runTextMatch.Run)
            {
                this.ReplaceProperties(this.lastMatchedRun);
            }

            this.lastMatchedRun = runTextMatch.Run;
        }

        /// <summary>
        /// The method is called after the replacement of all found matches.
        /// </summary>
        protected override void EndReplaceCore()
        {
            base.EndReplaceCore();

            if (this.lastMatchedRun != this.runsToRemove.LastOrDefault())
            {
                this.ReplaceProperties(this.lastMatchedRun);
            }

            this.MergeProcessedRuns();
            this.Clear();
        }

        private static List<Run> SplitRun(Regex regex, Run originalRun)
        {
            List<string> oldRunsText = SplitString(regex, originalRun.Text);

            if (string.IsNullOrEmpty(oldRunsText.FirstOrDefault()))
            {
                oldRunsText.RemoveAt(0);
            }

            return GetRuns(oldRunsText, originalRun);
        }

        private static List<string> SplitString(Regex regex, string input)
        {
            List<string> stringsSplitted = new List<string>();

            // We check if the regex matches the whole string.
            Match match = regex.Match(input, 0);
            if (!match.Success)
            {
                stringsSplitted.Add(input);
                return stringsSplitted;
            }

            int currentStringIndex = 0;
            do
            {
                stringsSplitted.Add(input.Substring(currentStringIndex, match.Index - currentStringIndex));
                currentStringIndex = match.Index + match.Length;

                match = match.NextMatch();
            }
            while (match.Success);

            stringsSplitted.Add(input.Substring(currentStringIndex, input.Length - currentStringIndex));
            return stringsSplitted;
        }

        private static List<Run> GetRuns(IEnumerable<string> oldRunTextSplitted, Run originalRun)
        {
            List<Run> runs = new List<Run>();
            foreach (string textPart in oldRunTextSplitted)
            {
                runs.Add(CreateRunCopy(originalRun, textPart));
            }

            return runs;
        }

        private static Run CreateRunCopy(Run runTemplate, string runText)
        {
            Run run = new Run(runTemplate.Document);
            run.Text = runText;
            run.Properties.CopyPropertiesFrom(runTemplate.Properties);

            return run;
        }

        private static bool ArePropertiesEqual(CharacterProperties characterProperties, CharacterProperties otherCharacterProperties)
        {
            foreach (IStyleProperty property in characterProperties.StyleProperties)
            {
                IStyleProperty otherProperty = otherCharacterProperties.GetStyleProperty(property.PropertyDefinition);
                object value = property.GetActualValueAsObject();
                object otherValue = otherProperty.GetActualValueAsObject();

                if (value == null && otherValue == null)
                {
                    continue;
                }
                else if (value != null && !value.Equals(otherValue))
                {
                    return false;
                }
            }

            return true;
        }

        private void ReplaceProperties(Run originalRun)
        {
            if (originalRun == null)
            {
                return;
            }

            Paragraph parent = originalRun.Paragraph;
            if (this.lastRunParent != parent)
            {
                this.paragraphTextLength = 0;
            }

            int currentRunIndexInParagraph = parent.Inlines.IndexOf(originalRun);

            // We have a full match and we should change the properties of the whole run => we shouldn't split it.
            if (originalRun.Text == this.replaceTexts.Peek())
            {
                this.replacePropertiesAction(originalRun.Properties);
                this.replaceTexts.Dequeue();

                // This is an optimization instead of iterating and accumulating the text length of the runs between this and the previous.
                // If they are not consecutive, we just do not put them for merging.
                if (currentRunIndexInParagraph - this.lastRunIndexInParagraph <= 1)
                {
                    this.runsToMerge.Add(this.paragraphTextLength,
                                     this.paragraphTextLength + originalRun.Text.Length,
                                     originalRun);
                }

                this.paragraphTextLength += originalRun.Text.Length;
            }
            else
            {
                int startingIndexOfNewRunInOldText = originalRun.Text.IndexOf(this.replaceTexts.Peek());

                int insertedRunsTextLength = 0;
                Run runWithReplacedProperties;

                List<Run> splittedRuns = SplitRun(this.Regex, originalRun);
                foreach (Run run in splittedRuns)
                {
                    // In case there are multiple occurrences of the replace text in the middle of the original run.
                    bool isReplacedRunOnly = string.IsNullOrEmpty(run.Text);

                    if (insertedRunsTextLength >= startingIndexOfNewRunInOldText || isReplacedRunOnly)
                    {
                        // This will return false if there was a match of the replaceText in old run text.
                        // E.g. runText = "alo", searchedText = "a", replaceText = "o"
                        string currentReplaceText = this.HasReplaceTextsQueued ? this.replaceTexts.Dequeue() : this.lastReplaceText;

                        runWithReplacedProperties = this.InsertRunWithReplacedProperties(originalRun, currentReplaceText, currentRunIndexInParagraph);
                        this.runsToMerge.Add(this.paragraphTextLength,
                                             this.paragraphTextLength + runWithReplacedProperties.Text.Length,
                                             runWithReplacedProperties);

                        this.paragraphTextLength += runWithReplacedProperties.Text.Length;
                        insertedRunsTextLength += runWithReplacedProperties.Text.Length;
                        currentRunIndexInParagraph++;
                    }

                    if (!isReplacedRunOnly)
                    {
                        parent.Inlines.Insert(currentRunIndexInParagraph, run);

                        this.paragraphTextLength += run.Text.Length;
                        insertedRunsTextLength += run.Text.Length;
                        currentRunIndexInParagraph++;
                    }
                }

                this.runsToRemove.Add(originalRun);
            }

            this.lastRunParent = parent;
            this.lastRunIndexInParagraph = currentRunIndexInParagraph;
        }

        private Run InsertRunWithReplacedProperties(Run runTemplate, string replaceText, int insertIndexInParagraph)
        {
            Run runWithReplacedProperties = CreateRunCopy(runTemplate, replaceText);
            this.replacePropertiesAction(runWithReplacedProperties.Properties);

            runTemplate.Paragraph.Inlines.Insert(insertIndexInParagraph, runWithReplacedProperties);

            return runWithReplacedProperties;
        }

        private void MergeProcessedRuns()
        {
            MergedRunInfo mergeRunInfo = new MergedRunInfo();

            KeyValuePair<Interval<int>, Run> previousIntervalRunPair = new KeyValuePair<Interval<int>, Run>();
            foreach (KeyValuePair<Interval<int>, Run> intervalToRun in this.runsToMerge)
            {
                // System.Diagnostics.Debug.WriteLine(string.Format("[{0}-{1}]-{2}", intervalToRun.Key.Start, intervalToRun.Key.End, intervalToRun.Value.Text));
                // We have consecutive runs with same properties that we should merge.
                if (previousIntervalRunPair.Value != null && previousIntervalRunPair.Key.End == intervalToRun.Key.Start &&
                    (mergeRunInfo.Run == null || ArePropertiesEqual(mergeRunInfo.Run.Properties, intervalToRun.Value.Properties)))
                {
                    mergeRunInfo.Run = previousIntervalRunPair.Value;
                    if (!mergeRunInfo.IsInsertIndexInitialized)
                    {
                        mergeRunInfo.ParagraphInsertIndex = mergeRunInfo.Run.Paragraph.Inlines.IndexOf(mergeRunInfo.Run);
                    }

                    mergeRunInfo.AppendText(mergeRunInfo.Run.Text);
                    this.runsToRemove.Add(previousIntervalRunPair.Value);
                }
                else if (mergeRunInfo.CanInsertMergedRun)
                {
                    mergeRunInfo.AppendText(previousIntervalRunPair.Value.Text);
                    this.InsertRunWithReplacedProperties(mergeRunInfo.Run, mergeRunInfo.Text, mergeRunInfo.ParagraphInsertIndex);

                    this.runsToRemove.Add(previousIntervalRunPair.Value);

                    mergeRunInfo = new MergedRunInfo();
                }

                previousIntervalRunPair = intervalToRun;
            }

            // We hit the case the last run should be merged.
            if (mergeRunInfo.CanInsertMergedRun)
            {
                mergeRunInfo.AppendText(previousIntervalRunPair.Value.Text);
                this.InsertRunWithReplacedProperties(mergeRunInfo.Run, mergeRunInfo.Text, mergeRunInfo.ParagraphInsertIndex);

                this.runsToRemove.Add(previousIntervalRunPair.Value);
            }
        }

        private void Clear()
        {
            foreach (Run runToRemove in this.runsToRemove)
            {
                runToRemove.Paragraph.Inlines.Remove(runToRemove);
            }

            this.runsToRemove.Clear();
            this.runsToMerge.Clear();
            this.replaceTexts.Clear();

            this.lastMatchedRun = null;
            this.paragraphTextLength = 0;
            this.lastRunIndexInParagraph = 0;
            this.lastReplaceText = string.Empty;
        }
    }
}