﻿using System;
using System.Collections.Generic;
using System.Text;
using Telerik.Windows.Documents.Flow.Model;

namespace Telerik.Windows.Documents.Flow.TextSearch
{
    /// <summary>
    /// Represents a handler to replace text.
    /// </summary>
    internal class ReplaceTextHandler : ReplaceHandlerBase
    {
        /// <summary>
        /// Gets or sets the replace text.
        /// </summary>
        /// <value>The replace text.</value>
        protected string ReplaceText { get; set; }

        internal static void ReplaceTextInternal(RunTextMatch runTextMatch, string replaceText)
        {
            Run run = runTextMatch.Run;
            string matchedText = run.Text.Substring(runTextMatch.StartIndex, runTextMatch.Length);
            string searchedText = runTextMatch.FullMatchText;

            // Starting run, so we replace the starting match with the whole replace text.
            run.Text = run.Text.Remove(runTextMatch.StartIndex, runTextMatch.Length);
            run.Text = run.Text.Insert(runTextMatch.StartIndex, replaceText);

            // We don't have a full match in the currently matched run, thus we should process runs next to it.
            if (matchedText != searchedText)
            {
                List<Run> matchedRunsToRemove = new List<Run>();
                StringBuilder builder = new StringBuilder();

                builder.Append(matchedText);

                Paragraph parent = run.Paragraph;
                string remainingText = searchedText.Substring(matchedText.Length, searchedText.Length - runTextMatch.Length);
                int remainingTextIndex;
                int nextRunIndex = parent.Inlines.IndexOf(run) + 1;
                for (int runIndex = nextRunIndex; runIndex < parent.Inlines.Count; runIndex++)
                {
                    Run nextRun = (Run)parent.Inlines[runIndex];
                    remainingTextIndex = 0;

                    while (remainingTextIndex < Math.Min(nextRun.Text.Length, remainingText.Length) &&
                           nextRun.Text[remainingTextIndex] == remainingText[remainingTextIndex])
                    {
                        builder.Append(remainingText[remainingTextIndex]);
                        remainingTextIndex++;
                    }

                    if (nextRun.Text.Length == remainingTextIndex)
                    {
                        // Middle run or ending run with fully matching text, so we need to remove it.
                        matchedRunsToRemove.Add(nextRun);
                    }
                    else if (nextRun.Text.Length > remainingTextIndex)
                    {
                        // Ending run with partially matching text, so we need to keep it and remove the matched text from it.
                        nextRun.Text = nextRun.Text.Substring(remainingTextIndex, nextRun.Text.Length - remainingTextIndex);
                    }

                    if (string.Equals(builder.ToString(), searchedText))
                    {
                        break;
                    }
                    else
                    {
                        remainingText = remainingText.Remove(0, remainingTextIndex);
                    }
                }

                foreach (Run runToRemove in matchedRunsToRemove)
                {
                    runToRemove.Paragraph.Inlines.Remove(runToRemove);
                }

                matchedRunsToRemove.Clear();
                builder.Clear();
            }
        }

        /// <summary>
        /// Replaces matched text in a run.
        /// </summary>
        /// <param name="runTextMatch">The match info in a text of a run.</param>
        protected internal override void ReplaceMatch(RunTextMatch runTextMatch)
        {
            ReplaceTextInternal(runTextMatch, this.ReplaceText);
        }

        /// <summary>
        /// Called before the replacement of all matches in a <see cref="Telerik.Windows.Documents.Flow.Model.RadFlowDocument" /> begins.
        /// </summary>
        /// <param name="replaceText">The replace text.</param>
        protected override void BeginReplaceCore(string replaceText)
        {
            base.BeginReplaceCore(replaceText);

            this.ReplaceText = replaceText;
        }
    }
}