﻿using System.Collections.Generic;
using System.Text.RegularExpressions;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Flow.TextSearch
{
    /// <summary>
    /// Represents a service for finding a text in a <see cref="Telerik.Windows.Documents.Flow.Model.RadFlowDocument"/>.
    /// </summary>
    internal class FindTextService
    {
        private const string WholeWordRegexPattern = @"\b";

        private readonly DocumentTextInfo documentTextInfo;

        public FindTextService(RadFlowDocument document)
        {
            Guard.ThrowExceptionIfNull(document, "document");

            this.documentTextInfo = new DocumentTextInfo(document);
        }

        public IEnumerable<RunTextMatch> Find(Regex regex)
        {
            Guard.ThrowExceptionIfNull(regex, "regex");

            int startIndex = 0;
            int initialRunTextLength = 0;
            int actualRunTextLength = 0;
            foreach (Match match in regex.Matches(this.documentTextInfo.Text))
            {
                Run matchedRun;
                if (this.documentTextInfo.TryGetRun(match.Index, out matchedRun))
                {
                    initialRunTextLength = this.documentTextInfo.GetLastRunInterval().End - this.documentTextInfo.GetLastRunInterval().Start;
                    actualRunTextLength = matchedRun.Text.Length;

                    // We subtract the original run index from the match index (because they are pointing to the initial document text) and 
                    // we add the accumulated text length difference from the replacements.
                    startIndex = match.Index - this.documentTextInfo.GetLastRunInterval().Start + (actualRunTextLength - initialRunTextLength);

                    int matchIndex = RunMatchHelper.GetMatchedIndexInRun(matchedRun, match.Value, startIndex);
                    int matchLength = RunMatchHelper.GetMatchedTextLengthInRun(matchedRun, match.Value, matchIndex);

                    yield return new RunTextMatch(matchedRun, matchIndex, matchLength, match.Value);
                }
            }
        }

        internal static Regex BuildRegex(string pattern, bool matchCase, bool matchWholeWord)
        {
            pattern = Regex.Escape(pattern);

            RegexOptions options = RegexOptions.None;

            if (!matchCase)
            {
                options = RegexOptions.IgnoreCase;
            }

            if (matchWholeWord)
            {
                pattern = string.Concat(WholeWordRegexPattern, pattern, WholeWordRegexPattern);
            }

            return new Regex(pattern, options);
        }
    }
}