﻿using System;
using System.IO;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.Tests.Integration;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Model.Protection;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
using Telerik.Windows.Documents.Flow.Tests.Utils;

namespace Telerik.Windows.Documents.Flow.Tests.Integration.Editing
{
    [TestClass]
    public class Protection : DocumentFlowVisualBaseTest
    {
        public static string TestFilesFolder = "VisualTestsDocuments/";
        readonly string currentUser = System.Environment.UserDomainName + "\\" + System.Environment.UserName;
        readonly string invalidUser = "invalidUser";
        readonly string password = "123";
        readonly string invalidPassword = "444";

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.DocumentProtection)]
        public void ProtectedDocument_CouldBeEditedByValidUserWithGrantedPermissions()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            CreateDocumentWithProtection(editor, currentUser);

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION });
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.DocumentProtection)]
        public void ProtectedDocument_CouldNotBeEditedByInvalidUserWithGrantedPermissions()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            CreateDocumentWithProtection(editor, invalidUser);

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION });
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.DocumentProtection)]
        public void ProtectedDocument_CouldNotBeEditedWithoutPermissionsGranted()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.InsertText("One paragraph");
            editor.InsertParagraph();
            editor.InsertText("Another paragraph with test text");
            editor.InsertParagraph();
            editor.InsertText("End");
            editor.InsertParagraph();
            editor.InsertText("The last paragraph");

            editor.Protect(password);
            editor.Document.ProtectionSettings.ProtectionMode = ProtectionMode.AllowComments;

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION });
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.DocumentProtection)]
        public void ProtectedDocument_ProtectedWithAlgoritmDifferentThanTheDefault_IsProtected()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.Document.ProtectionSettings.AlgorithmName = ProtectionAlgorithmNames.SHA1;
            editor.Protect(password, ProtectionMode.AllowComments);

            var start = editor.InsertText("A:");
            editor.InsertText("Middle");
            var end = editor.InsertText(":B");

            editor.InsertParagraph();
            editor.InsertText("Comment");

            PermissionRangeCredentials cred = new PermissionRangeCredentials(currentUser);
            editor.InsertPermissionRange(cred, start, end);

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION });
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.DocumentProtection)]
        public void ImportProtectedDocumentWithNoProtection_PermissionRangeCouldBeAdded()
        {
            RadFlowDocument document = TestHelper.GetDocumentFromFile(TestFilesFolder + "DocumentWithoutProtection.docx", new DocxFormatProvider());
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            var tables = document.EnumerateChildrenOfType<Table>();
            var row1 = tables.FirstOrDefault().Rows.FirstOrDefault();
            var row2 = tables.LastOrDefault().Rows.Last();

            PermissionRangeCredentials cred = new PermissionRangeCredentials(currentUser);
            PermissionRangeCredentials cred2 = new PermissionRangeCredentials(invalidUser);

            editor.InsertPermissionRange(cred, row1);
            editor.InsertPermissionRange(cred, row2);
            editor.InsertPermissionRange(cred2, row1);
            editor.Protect(password);

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION });
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.DocumentProtection)]
        public void ImportProtectedDocument_AddPermissionRangeAndExport()
        {
            RadFlowDocument document = TestHelper.GetDocumentFromFile(TestFilesFolder + "DocumentWithProtection.docx", new DocxFormatProvider());
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            var start = editor.InsertText("Edit protected ");
            var end = editor.InsertText("document");
            editor.InsertParagraph();

            PermissionRangeCredentials cred = new PermissionRangeCredentials(currentUser);
            editor.InsertPermissionRange(cred, start, end);

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION });
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.DocumentProtection)]
        public void ImportProtectedDocument_UnprotectAndExport()
        {
            RadFlowDocument document = TestHelper.GetDocumentFromFile(TestFilesFolder + "DocumentWithProtection.docx", new DocxFormatProvider());
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.Unprotect();

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION });
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.DocumentProtection)]
        public void RemoveProtectionAndThenSetItThroughEnforcedProperty()
        {
            RadFlowDocument document = TestHelper.GetDocumentFromFile(TestFilesFolder + "DocumentWithProtection.docx", new DocxFormatProvider());
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.Unprotect();
            document.ProtectionSettings.Enforced = true;
            document.ProtectionSettings.ProtectionMode = ProtectionMode.ReadOnly;

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION });
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.DocumentProtection)]
        public void ImportUnprotectedDocument_ProtectAndExport()
        {
            RadFlowDocument document = TestHelper.GetDocumentFromFile(TestFilesFolder + "DynamicFontsIssues.docx", new DocxFormatProvider());
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.Protect(password);

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION });
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.DocumentProtection)]
        public void DocumentProtectedWithSha1_IsUnprotectedUsingCorrectPassword()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            document.ProtectionSettings.AlgorithmName = ProtectionAlgorithmNames.SHA1;

            CreateDocumentWithProtection(editor, currentUser);
            Assert.IsTrue(document.ProtectionSettings.Enforced);

            bool isUnprotected = editor.Unprotect(password);
            Assert.IsTrue(isUnprotected);

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION });
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.DocumentProtection)]
        public void DocumentProtectedWithSha256_IsUnprotectedUsingCorrectPassword()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            document.ProtectionSettings.AlgorithmName = ProtectionAlgorithmNames.SHA256;

            CreateDocumentWithProtection(editor, currentUser);
            Assert.IsTrue(document.ProtectionSettings.Enforced);

            bool isUnprotected = editor.Unprotect(password);
            Assert.IsTrue(isUnprotected);

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION });
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.DocumentProtection)]
        public void DocumentProtectedWithSha384_IsUnprotectedUsingCorrectPassword()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            document.ProtectionSettings.AlgorithmName = ProtectionAlgorithmNames.SHA384;

            CreateDocumentWithProtection(editor, currentUser);
            Assert.IsTrue(document.ProtectionSettings.Enforced);

            bool isUnprotected = editor.Unprotect(password);
            Assert.IsTrue(isUnprotected);

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION });
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.DocumentProtection)]
        public void DocumentProtectedWithSha512_IsUnprotectedUsingCorrectPassword()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            document.ProtectionSettings.AlgorithmName = ProtectionAlgorithmNames.SHA512;

            CreateDocumentWithProtection(editor, currentUser);
            Assert.IsTrue(document.ProtectionSettings.Enforced);

            bool isUnprotected = editor.Unprotect(password);
            Assert.IsTrue(isUnprotected);

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION });
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.DocumentProtection)]
        public void DocumentProtectedWithSha512_IsNotUnprotectedUsingIncorrectPassword()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            document.ProtectionSettings.AlgorithmName = ProtectionAlgorithmNames.SHA512;

            CreateDocumentWithProtection(editor, currentUser);
            Assert.IsTrue(document.ProtectionSettings.Enforced);

            bool isUnprotected = editor.Unprotect(invalidPassword);
            Assert.IsFalse(isUnprotected);

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION });
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.DocumentProtection)]
        public void UnprotecDocumenttWithWrongPassword_DocumentIsNotUnprotectedTest()
        {
            RadFlowDocument document = TestHelper.GetDocumentFromFile(TestFilesFolder + "DocumentWithProtection.docx", new DocxFormatProvider());
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.Unprotect(invalidPassword);

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION });
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.DocumentProtection)]
        public void UprotectWithPasswordDocumentCreatedInMsWord()
        {
            RadFlowDocument document = TestHelper.GetDocumentFromFile(TestFilesFolder + "DocumentProtectedInMsWord.docx", new DocxFormatProvider());
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.Unprotect(password);

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION });
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.DocumentProtection)]
        public void UnprotectWithPasswordDocumentCreatedInRichTextBox()
        {
            RadFlowDocument document = TestHelper.GetDocumentFromFile(TestFilesFolder + "DocumentProtectedWithDpl.docx", new DocxFormatProvider());

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            editor.Unprotect(password);

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION });
        }

        private void CreateDocumentWithProtection(RadFlowDocumentEditor editor, string user)
        {
            var start = editor.InsertText("One paragraph");
            editor.InsertParagraph();
            editor.InsertText("Another paragraph with test text");
            editor.InsertParagraph();
            var end = editor.InsertText("End");
            editor.InsertParagraph();
            editor.InsertText("The last paragraph");

            PermissionRangeCredentials credentials = new PermissionRangeCredentials(user);

            editor.InsertPermissionRange(credentials, start, end);
            editor.Protect(password, ProtectionMode.AllowComments);
        }
    }
}
