﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Fixed.Tests.Integration;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
using Telerik.Windows.Documents.Flow.Tests.Utils;
using Telerik.Windows.Documents.Media;

namespace Telerik.Windows.Documents.Flow.Tests.Integration.Editing
{
    [TestClass]
    public class TabStops : DocumentFlowVisualBaseTest
    {
        readonly Dictionary<string, string> acronyms = new Dictionary<string, string>();
        public static string TestFilesFolder = "VisualTestsDocuments/";

        [TestInitialize]
        public void TestInitialize()
        {
            acronyms.Add("2moro", "Tomorrow");
            acronyms.Add("2nite", "Tonight");
            acronyms.Add("BRB", "Be Right Back");
            acronyms.Add("BTW", "By The Way -or- Bring The Wheelchair");
            acronyms.Add("B4N", "Bye For Now");
            acronyms.Add("BCNU", "Be Seeing You");
            acronyms.Add("BFF", "Best Friends Forever -or- Best Friend's Funeral");
            acronyms.Add("CYA", "Cover Your Ass -or- See Ya");
            acronyms.Add("DBEYR", "Don't Believe Everything You Read");
            acronyms.Add("FUD", "Fear, Uncertainty, and Disinformation");
            acronyms.Add("FWIW", "For What It's Worth -or- Forgot Where I Was");
        }

        // Default Tab set
        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.TapStops)]
        public void DefaultTabStopTest()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            document.DefaultTabStopWidth = Unit.InchToDip(2);

            AddTextAndTabsInDocument(editor);

            this.AssertVisual(document);
        }

        // Left alignment, Dot leader - different format providers
        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.TapStops)]
        public void LeftTabStopTypeAndDotTabStopLeaderTest()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            InsertDocumentContentWithTabStop(editor, 3, TabStopType.Left, TabStopLeader.Dot);

            this.AssertVisual(document);
        }

        // Right alignment, Hyphen leader - different format providers
        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.TapStops)]
        public void RightTabStopTypeAndHypehnTabStopLeaderTest_Docx()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            InsertDocumentContentWithTabStop(editor, 5, TabStopType.Right, TabStopLeader.Hyphen);

            this.AssertVisual(document);
        }

        // Center alignment, No Leader - different format providers
        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.TapStops)]
        public void CenterTabStopTypeAndNoTabStopLeaderTest()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            InsertDocumentContentWithTabStop(editor, 4, TabStopType.Center, TabStopLeader.None);

            this.AssertVisual(document);
        }

        // Decimal alignment, Underscore leader
        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.TapStops)]
        public void DecimalAlignmentTabStopTypeAndUnderscoreLeader()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            AddDocumentContentWithDecimalTabStops(editor);

            this.AssertVisual(document);
        }

        // Bar alignment
        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.TapStops)]
        public void BarTabStopTypeTest()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            InsertDocumentContentWithTabStop(editor, 4, TabStopType.Bar, TabStopLeader.None, 8);

            this.AssertVisual(document);
        }

        // Left TabStop combined with Paragraph Left Indent
        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.TapStops)]
        public void LeftTabStopCombinedWithParagraphLeftIndent()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.ParagraphFormatting.LeftIndent.LocalValue = Unit.InchToDip(1);
            this.InsertDocumentContentWithTabStop(editor, 3, TabStopType.Left, TabStopLeader.Hyphen);

            this.AssertVisual(document);
        }

        // Bug ID: 160762
        // Right TabStop combined with Paragraph Left Indent
        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.TapStops)]
        public void RightTabStopCombinedWithParagraphLeftIndent()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.ParagraphFormatting.LeftIndent.LocalValue = Unit.InchToDip(1);
            this.InsertDocumentContentWithTabStop(editor, 3, TabStopType.Right, TabStopLeader.Hyphen);

            this.AssertVisual(document);
        }

        // Define two styles with different tab stops
        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.TapStops)]
        public void TwoDifferentTabStopsDefinedInTwoDifferentStyles()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            Style style1 = new Style("first", StyleType.Paragraph);
            style1.ParagraphProperties.TabStops.LocalValue = new TabStopCollection().Insert(new TabStop(Unit.InchToDip(2), TabStopType.Left, TabStopLeader.Underscore));
            document.StyleRepository.Add(style1);

            Style style2 = new Style("second", StyleType.Paragraph);
            style2.ParagraphProperties.TabStops.LocalValue = new TabStopCollection().Insert(new TabStop(Unit.InchToDip(4), TabStopType.Left, TabStopLeader.Hyphen));
            document.StyleRepository.Add(style2);

            editor.InsertParagraph().StyleId = style1.Id;
            editor.InsertText("Tab stop is: ");
            editor.InsertText("\t");
            editor.InsertText("2\"");

            editor.InsertParagraph();

            editor.InsertParagraph().StyleId = style2.Id;
            editor.InsertText("Tab stop is: ");
            editor.InsertText("\t");
            editor.InsertText("4\"");

            this.AssertVisual(document);
        }

        // Import a document created in MS Word with three tab stops
        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.TapStops)]
        public void ImportExportDocumentWithThreeTabStops()
        {
            RadFlowDocument document = TestHelper.GetDocumentFromFile(TestFilesFolder + "ThreeTabStopsInADocument.docx", new DocxFormatProvider());

            this.AssertVisual(document);
        }

        // TabStops in Header
        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.TapStops)]
        public void TabStopsAreRespectedInHeader()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.InsertText("TabStops used in headers");
            var header = document.Sections.First().Headers.Add(HeaderFooterType.Default);

            foreach (var item in acronyms)
            {
                var testText = header.Blocks.AddParagraph().Inlines.AddRun(item.Key + "\t" + item.Value);
                testText.Paragraph.TabStops = new TabStopCollection().Insert(new TabStop(Unit.InchToDip(2), TabStopType.Left, TabStopLeader.MiddleDot));
            }

            this.AssertVisual(document);
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.TapStops)]
        public void WhenTheTextIsLongerThanTheDefinedTabStopLenghtRightTabStopIsNotRespectedTest()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            Style style1 = new Style("first", StyleType.Paragraph);
            style1.ParagraphProperties.TabStops.LocalValue = new TabStopCollection().Insert(new TabStop(Unit.InchToDip(2), TabStopType.Right, TabStopLeader.Underscore));
            document.StyleRepository.Add(style1);

            editor.InsertParagraph().StyleId = style1.Id;
            editor.InsertText("Tab stop is: ");
            editor.InsertText("\t");
            editor.InsertText("2\". This text should not respect the defined TabStop and should flow to the right");

            editor.InsertParagraph();

            this.AssertVisual(document);
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.TapStops)]
        public void SetThreeTabStopsToAParagraphThroughDocument()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            TabStopCollection tabStops = new TabStopCollection()
                .Insert(new TabStop(Unit.InchToDip(2), TabStopType.Left, TabStopLeader.Dot))
                .Insert(new TabStop(Unit.InchToDip(4), TabStopType.Left, TabStopLeader.Dot))
                .Insert(new TabStop(Unit.InchToDip(6), TabStopType.Left, TabStopLeader.Dot));

            Paragraph p = document.Sections.AddSection().Blocks.AddParagraph();
            p.TabStops = tabStops;
            p.Inlines.AddRun("First \t Second \t Third \t");

            this.AssertVisual(document);
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.TapStops)]
        public void RemoveAllTabStopsFromFirstPararaphOfImportedDocument()
        {
            RadFlowDocument document = TestHelper.GetDocumentFromFile(TestFilesFolder + "ThreeTabStopsInADocument.docx", new DocxFormatProvider());
            Paragraph p = document.EnumerateChildrenOfType<Paragraph>().First();

            for (int i = 0; i < 3; i++)
            {
                TabStop tabStop = p.TabStops.FirstOrDefault();
                var tabStops = p.TabStops.Remove(tabStop);
                p.TabStops = tabStops;
            }

            this.AssertVisual(document);
        }

        private void InsertDocumentContentWithTabStop(RadFlowDocumentEditor editor, double position, TabStopType type, TabStopLeader leader, int tabPresses = 1)
        {
            TabStop tabStop = new TabStop(Unit.InchToDip(position), type, leader);
            editor.ParagraphFormatting.TabStops.LocalValue = new TabStopCollection().Insert(tabStop);

            foreach (var item in acronyms)
            {
                editor.InsertText(item.Key);
                for (int i = 0; i < tabPresses; i++)
                {
                    editor.InsertText("\t");
                }
                editor.InsertText(item.Value);
                editor.InsertParagraph();
            }
        }

        private void AddTextAndTabsInDocument(RadFlowDocumentEditor editor)
        {
            editor.InsertText("one");
            editor.InsertText("\t");
            editor.InsertText("two");
            editor.InsertText("\t");
            editor.InsertText("three");
            editor.InsertText("\t");
            editor.InsertText("four");
        }

        private static void AddDocumentContentWithDecimalTabStops(RadFlowDocumentEditor editor)
        {
            string[] decimalValues = new string[] { "3.14", "4567.98", "-123.565", "0.0", "1111.0", "3.2m" };

            editor.ParagraphFormatting.TabStops.LocalValue = new TabStopCollection().Insert(new TabStop(Unit.InchToDip(3), TabStopType.Decimal, TabStopLeader.Underscore));

            foreach (var item in decimalValues)
            {
                editor.InsertText("Length: " + item.Length.ToString());
                editor.InsertText("\t");
                editor.InsertText(item);
                editor.InsertParagraph();
            }
        }
    }
}
