﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Linq;
using Telerik.Windows.Documents.Flow.FormatProviders.Html;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
using Telerik.Windows.Documents.Media;

namespace Telerik.Windows.Documents.Flow.Tests.Integration.HtmlImportExport
{
    [TestClass]
    public class UriImageSourceTests
    {
        private const string ImageUri = "http://www.telerik.com/sfimages/default-source/logos/telerik-logo-reversed.png";
        private const string InvalidImageUri = "http://www.telerik.com/sfimages/default-source/logos/telerik-logo-reversedbg.png";
        private const string TestFilesFolder = "Html/Images/";

        public RadFlowDocument ImportHtml()
        {
            HtmlFormatProvider provider = new HtmlFormatProvider();
            return provider.Import(string.Format("<p>hello world</p><img src=\"{0}\">", ImageUri));
        }

        [TestMethod]
        public void ExportImageWithUri_WhenImageSourceIsUri_AndExportModeIsExternal()
        {
            HtmlFormatProvider provider = new HtmlFormatProvider();
            provider.ExportSettings.ImagesExportMode = ImagesExportMode.External;

            RadFlowDocument document = this.ImportHtml();
            string html = provider.Export(document);
            Assert.IsTrue(html.Contains(ImageUri));
            Assert.IsFalse(html.Contains("<img src=\"data:image/png;base64,"));
        }

        [TestMethod]
        public void ExportImageWithUri_WhenImageSourceIsUri_AndExportModeIsEmbedded()
        {
            HtmlFormatProvider provider = new HtmlFormatProvider();
            provider.ExportSettings.ImagesExportMode = ImagesExportMode.Embedded;

            RadFlowDocument document = this.ImportHtml();
            string html = provider.Export(document);
            Assert.IsTrue(html.Contains(ImageUri));
            Assert.IsFalse(html.Contains("<img src=\"data:image/png;base64,"));
        }

        [TestMethod]
        public void ExportImageWithUri_WhenAnImageSourceIsCreatedFromUri()
        {
            HtmlFormatProvider provider = new HtmlFormatProvider();

            RadFlowDocument document = this.ImportHtml();
            ImageInline imageInline = document.EnumerateChildrenOfType<ImageInline>().First();

            UriImageSource newSource = new UriImageSource(new Uri(ImageUri));
            imageInline.Image.ImageSource = newSource;

            string html = provider.Export(document);
            Assert.IsTrue(html.Contains(ImageUri));
            Assert.IsFalse(html.Contains("<img src=\"data:image/png;base64,"));
        }

        [TestMethod]
        public void ExportImageWithUri_WhenAnImageSourceIsCreatedFromUriAndExtension()
        {
            HtmlFormatProvider provider = new HtmlFormatProvider();

            RadFlowDocument document = this.ImportHtml();
            ImageInline imageInline = document.EnumerateChildrenOfType<ImageInline>().First();

            UriImageSource newSource = new UriImageSource(new Uri(ImageUri), ".jpg");
            imageInline.Image.ImageSource = newSource;

            string html = provider.Export(document);
            Assert.IsTrue(html.Contains(ImageUri));
            Assert.IsFalse(html.Contains("<img src=\"data:image/png;base64,"));
        }

        [TestMethod]
        public void ImportExportHtmlFileWithInvalidImageUri()
        {
            HtmlFormatProvider provider = new HtmlFormatProvider();
            RadFlowDocument document = TestHelper.GetDocumentFromFile(string.Format("{0}HtmlWithInvalidImageUri.html", TestFilesFolder), provider);

            string html = provider.Export(document);
            Assert.IsTrue(html.Contains(InvalidImageUri));
            Assert.IsFalse(html.Contains("<img src=\"data:image/png;base64,"));
        }

        [TestMethod]
        public void ImportExportWithUri_ImageIsDownloadedWhenItsDataIsAccessed()
        {
            HtmlFormatProvider provider = new HtmlFormatProvider();

            RadFlowDocument document = TestHelper.GetDocumentFromFile(string.Format("{0}HtmlWithImages.html", TestFilesFolder), provider);
            ImageInline imageInline = document.EnumerateChildrenOfType<ImageInline>().First();
            UriImageSource uriImageSourceFromDocument = imageInline.Image.ImageSource as UriImageSource;

            ImageSource newImageSource = new ImageSource(uriImageSourceFromDocument.Data, uriImageSourceFromDocument.Extension);
            imageInline.Image.ImageSource = newImageSource;

            string html = provider.Export(document);
            Assert.IsFalse(html.Contains(ImageUri));
            Assert.IsTrue(html.Contains("<img src=\"data:image/png;base64,"));
        }

        [TestMethod]
        public void ImportImageWithUri_WhenTheUriIsRelative()
        {
            string imagePath = "../../missingImage.png";

            HtmlFormatProvider provider = new HtmlFormatProvider();

            RadFlowDocument document = provider.Import("<p>hello world</p><img src=\"" + imagePath + "\">");
            string html = provider.Export(document);
            Assert.IsTrue(html.Contains("<img src=\"" + imagePath + "\""));
            Assert.IsFalse(html.Contains("<img src=\"data:image/png;base64,"));
        }
    }
}