﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Flow.Model.Styles;

namespace Telerik.Windows.Documents.Flow.Tests.Integration.HtmlImportExport.Lists
{
    [TestClass]
    public class StartIndexTests : ListTestsBase
    {
        [TestMethod]
        public void SimpleList_IsImportedWithValueTwoAsStartIndexForTheFirstLevel()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(expected);

            List list = this.GetList(expected);
            list.Levels[0].StartIndex = 2;

            Paragraph p = editor.InsertText("first item").Paragraph;
            p.ListId = list.Id;
            p.ListLevel = 0;
            p.StyleId = BuiltInStyleNames.NormalWebStyleId;

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body><ol start=\"2\"><li>first item</li></ol></body></html>");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void SimpleList_IsImportedWithValueThreeAsStartIndexForTheFirstLevel()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(expected);

            List list = this.GetList(expected);
            list.Levels[0].StartIndex = 3;

            Paragraph p = editor.InsertText("first item").Paragraph;
            p.ListId = list.Id;
            p.ListLevel = 0;
            p.StyleId = BuiltInStyleNames.NormalWebStyleId;

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body><ol start=\"2\"><li value=\"3\">first item</li></ol></body></html>");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ListLevelWithDifferentStartIndexThenDefault_IsCorrectlyImported()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(expected);

            int startIndex = 7;

            List list = this.GetList(expected);
            list.Levels[0].StartIndex = startIndex;

            Paragraph p = editor.InsertText("0-1").Paragraph;
            p.ListId = list.Id;
            p.ListLevel = 0;
            p.StyleId = BuiltInStyleNames.NormalWebStyleId;

            this.AddParagraphsToDocument(editor, list, 0, 2);

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<ol start=\"" + startIndex + "\"><li>0-1</li><li>0-2</li></ol>");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ListLevelWitOverriddenStartIndex_IsCorrectlyImported()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(expected);

            int startIndex = 3;

            List list = this.GetList(expected);
            list.Levels[0].StartIndex = startIndex;

            Paragraph p = editor.InsertText("0-1").Paragraph;
            p.ListId = list.Id;
            p.ListLevel = 0;
            p.StyleId = BuiltInStyleNames.NormalWebStyleId;

            this.AddParagraphsToDocument(editor, list, 0, 2);

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<ol start=\"7\"><li value=\"" + startIndex + "\">0-1</li><li>0-2</li></ol>");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ListWithUncommonValueForTheLastTwoListLevelElements_ShouldBeImportedAsTwoLists()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(expected);

            List list1 = this.GetList(expected);
            list1.Levels[0].StartIndex = 3;

            Paragraph p = editor.InsertText("0-3").Paragraph;
            p.ListId = list1.Id;
            p.ListLevel = 0;
            p.StyleId = BuiltInStyleNames.NormalWebStyleId;

            this.AddParagraphsToDocument(editor, list1, 0, 4);
            this.AddParagraphsToDocument(editor, list1, 0, 5);

            List list2 = this.GetList(expected);
            list2.Levels[0].StartIndex = 10;

            this.AddParagraphsToDocument(editor, list2, 0, 10);
            this.AddParagraphsToDocument(editor, list2, 0, 11);

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<ol><li value=\"3\">0-3</li><li>0-4</li><li>0-5</li><li value=\"10\">0-10</li><li>0-11</li></ol>");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ListWithListStyleTypeNoneForTheFirstListLevelItemAndUncommonValueForTheLastTwoListLevelElements_ShouldBeImportedAsTwoLists()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(expected);

            List list1 = this.GetList(expected);
            list1.Levels[0].StartIndex = 3;

            Paragraph p = editor.InsertText("0-3").Paragraph;
            p.ListId = list1.Id;
            p.ListLevel = 0;
            p.StyleId = BuiltInStyleNames.NormalWebStyleId;

            this.AddParagraphsToDocument(editor, list1, 0, 4);
            this.AddParagraphsToDocument(editor, list1, 0, 5);

            List list2 = this.GetList(expected);
            list2.Levels[0].StartIndex = 10;

            this.AddParagraphsToDocument(editor, list2, 0, 10);
            this.AddParagraphsToDocument(editor, list2, 0, 11);

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<ol><li value=\"3\" style=\"list-style-type: none;\">0-3</li><li>0-4</li><li>0-5</li><li value=\"10\">0-10</li><li>0-11</li></ol>");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ListWithValueForEachListLevel_IsImportedWithValueThreeAsStartIndexForTheFirstLevel()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(expected);

            List list = this.GetList(expected);
            list.Levels[0].StartIndex = 3;

            Paragraph p = editor.InsertText("0-3").Paragraph;
            p.ListId = list.Id;
            p.ListLevel = 0;
            p.StyleId = BuiltInStyleNames.NormalWebStyleId;

            this.AddParagraphsToDocument(editor, list, 0, 4);
            this.AddParagraphsToDocument(editor, list, 0, 5);

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body><ol><li value=\"3\">0-3</li><li value=\"4\">0-4</li><li value=\"5\">0-5</li></ol></body></html>");

            //Assert
            this.AssertDocuments(expected, imported);
        }
    }
}
