﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;
using Telerik.Windows.Documents.Media;
#if WPF || SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Flow.Tests.Integration.HtmlImportExport.Paragraphs
{
    [TestClass]
    public abstract class CodeStyledTagsTestsBase : ParagraphTestsBase
    {
        public CodeStyledTagsTestsBase()
        {
            base.AsserterInfo.RegisterPropertiesToCheck(new IStylePropertyDefinition[]
            {
               Run.FontFamilyPropertyDefinition,
               Run.FontSizePropertyDefinition
            });
        }

        protected abstract string Tag { get; }

        [TestMethod]
        public void Test_CodeStyledTag_AsChild_Import()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();
            Run run = expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("test");
            run.FontFamily = new ThemableFontFamily("Courier New");
            run.FontSize = Unit.PointToDip(10);

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body><p><" + this.Tag + ">test</" + this.Tag + "></p></body></html>");

            //Assert
            this.AssertDocumentStructureAndProperties(expected, imported);
        }

        [TestMethod]
        public void Test_CodeStyledTag_AsParentElement_Import()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();
            Section section = expected.Sections.AddSection();

            Paragraph p = section.Blocks.AddParagraph();
            p.Properties.ParagraphMarkerProperties.FontSize.LocalValue = Unit.PointToDip(10);
            p.Properties.ParagraphMarkerProperties.FontFamily.LocalValue = new ThemableFontFamily("Courier New");
            Run run = p.Inlines.AddRun("first");
            run.FontFamily = new ThemableFontFamily("Courier New");
            run.FontSize = Unit.PointToDip(10);

            p = section.Blocks.AddParagraph();
            p.Properties.ParagraphMarkerProperties.FontSize.LocalValue = Unit.PointToDip(10);
            p.Properties.ParagraphMarkerProperties.FontFamily.LocalValue = new ThemableFontFamily("Courier New");
            run = p.Inlines.AddRun("second");
            run.FontFamily = new ThemableFontFamily("Courier New");
            run.FontSize = Unit.PointToDip(10);

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body><" + this.Tag + "><p>first</p><p>second</p></" + this.Tag + "></body></html>");

            //Assert
            this.AssertDocumentStructureAndProperties(expected, imported);
        }

        [TestMethod]
        public void Test_CodeStyledTag_AsParentElementSurroundedByText_Import()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();
            Section section = expected.Sections.AddSection();
            Paragraph p = section.Blocks.AddParagraph();

            Run run = p.Inlines.AddRun("before");
            run = p.Inlines.AddRun("after");
            run.FontFamily = new ThemableFontFamily("Courier New");
            run.FontSize = Unit.PointToDip(10);

            p = section.Blocks.AddParagraph();
            p.Properties.ParagraphMarkerProperties.FontSize.LocalValue = Unit.PointToDip(10);
            p.Properties.ParagraphMarkerProperties.FontFamily.LocalValue = new ThemableFontFamily("Courier New");
            run = p.Inlines.AddRun("first");
            run.FontFamily = new ThemableFontFamily("Courier New");
            run.FontSize = Unit.PointToDip(10);

            p = section.Blocks.AddParagraph();
            p.Properties.ParagraphMarkerProperties.FontSize.LocalValue = Unit.PointToDip(10);
            p.Properties.ParagraphMarkerProperties.FontFamily.LocalValue = new ThemableFontFamily("Courier New");
            run = p.Inlines.AddRun("second");
            run.FontFamily = new ThemableFontFamily("Courier New");
            run.FontSize = Unit.PointToDip(10);

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body>before<" + this.Tag + ">after<p>first</p><p>second</p></" + this.Tag + "></body></html>");

            //Assert
            this.AssertDocumentStructureAndProperties(expected, imported);
        }

        [TestMethod]
        public void Test_CodeStyledTag_AsChildElementInTableCell_Import()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();
            Table table = expected.Sections.AddSection().Blocks.AddTable();
            TableRow row = table.Rows.AddTableRow();
            TableCell cell = row.Cells.AddTableCell();

            Paragraph p = cell.Blocks.AddParagraph();

            Run run = p.Inlines.AddRun("before");
            run = p.Inlines.AddRun("after");
            run.FontFamily = new ThemableFontFamily("Courier New");
            run.FontSize = Unit.PointToDip(10);

            p = cell.Blocks.AddParagraph();
            p.Properties.ParagraphMarkerProperties.FontSize.LocalValue = Unit.PointToDip(10);
            p.Properties.ParagraphMarkerProperties.FontFamily.LocalValue = new ThemableFontFamily("Courier New");
            run = p.Inlines.AddRun("first");
            run.FontFamily = new ThemableFontFamily("Courier New");
            run.FontSize = Unit.PointToDip(10);

            p = cell.Blocks.AddParagraph();
            p.Properties.ParagraphMarkerProperties.FontSize.LocalValue = Unit.PointToDip(10);
            p.Properties.ParagraphMarkerProperties.FontFamily.LocalValue = new ThemableFontFamily("Courier New");
            run = p.Inlines.AddRun("second");
            run.FontFamily = new ThemableFontFamily("Courier New");
            run.FontSize = Unit.PointToDip(10);

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body><table><tr><td>before<" + this.Tag + ">after<p>first</p><p>second</p></" + this.Tag + "></td></tr></table></body></html>");

            //Assert
            this.AssertDocumentStructureAndProperties(expected, imported);
        }
    }

    [TestClass]
    public class CodeTagTests : CodeStyledTagsTestsBase
    {
        protected override string Tag
        {
            get
            {
                return "code";
            }
        }
    }

    [TestClass]
    public class KbdTagTests : CodeStyledTagsTestsBase
    {
        protected override string Tag
        {
            get
            {
                return "kbd";
            }
        }
    }

    [TestClass]
    public class SampTagTests : CodeStyledTagsTestsBase
    {
        protected override string Tag
        {
            get
            {
                return "samp";
            }
        }
    }

    [TestClass]
    public class TeletypeTagTests : CodeStyledTagsTestsBase
    {
        protected override string Tag
        {
            get
            {
                return "tt";
            }
        }
    }
}
