﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.FormatProviders.Html;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Tests.Utilities;

namespace Telerik.Windows.Documents.Flow.Tests.Integration.HtmlImportExport.Paragraphs
{
    [TestClass]
    public class NbspTests : ParagraphTestsBase
    {
        private const char NonBreakingSpaceChar = (char)160;

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.Section)]
        public void Test_Nbsp_Import()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("ala " + NonBreakingSpaceChar + " bala");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body><p>ala &nbsp; bala</p></body></html>");

            //Assert
            this.AssertDocumentsStructureOnly(expected, imported);

#if VISUAL
            this.AssertDocumentVisual(imported);
#endif
        }

#if VISUAL
        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.Section)]
        public void Test_Nbsp_ImportEmptyParagraphs()
        {
            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body><p>Parag&nbsp;&nbsp;&nbsp;raph 1</p><p>&nbsp;</p><p>&nbsp;</p><p>Paragraph 2</p></body>");
            this.AssertDocumentVisual(imported);
        }
#endif

        [TestMethod]
        public void Test_NbspBetweenSpans_Import()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("ala");
            paragraph.Inlines.AddRun(" " + NonBreakingSpaceChar + " "); // whitespace, nbsp, whitespace
            paragraph.Inlines.AddRun("bala");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body><p><span>ala</span> &nbsp; <span>bala</span></p></body></html>");

            //Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }

        [TestMethod]
        public void Test_Nbsp_ExportImport()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("ala " + NonBreakingSpaceChar + " bala");

            // Act
            RadFlowDocument imported = this.ExportImportDocument(expected, new HtmlFormatProvider());

            // Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.Section)]
        public void Nbsp_WhenAreMultipleInSameParagraph_Import()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("Parag" + NonBreakingSpaceChar
                                                                                        + NonBreakingSpaceChar
                                                                                        + NonBreakingSpaceChar
                                                                                        + NonBreakingSpaceChar
                                                                                        + NonBreakingSpaceChar + "raph");
            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body><p>Parag&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;raph</p></body></html>"); // 5 &nbsp; symbols.

            // Assert
            this.AssertDocumentsStructureOnly(expected, imported);
            this.AssertDocumentVisual(imported);
        }

        [TestMethod]
        public void Nbsp_WhenAreMultipleInSameParagraph_AndNormalizeStringsIsTrue_Import()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("Parag     raph");

            string html = "<html><body><p>Parag&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;raph</p></body></html>";

            // Act
            HtmlFormatProvider provider = new HtmlFormatProvider();
            provider.ImportSettings.ReplaceNonBreakingSpaces = true;
            RadFlowDocument imported = provider.Import(html); // 5 &nbsp; symbols should be replaced with whitespaces.

            // Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }

        [TestMethod]
        public void Nbsp_WhenAreMultipleInSameParagraph_ExportImport()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("Parag " + NonBreakingSpaceChar + " " + NonBreakingSpaceChar + " raph"); // 5 spaces.

            // Act
            RadFlowDocument imported = this.ExportImportDocument(expected, new HtmlFormatProvider());

            // Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }

        [TestMethod]
        public void Paragraph_WhenHasOnlyNbspSymbol_Import()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Section section = expected.Sections.AddSection();

            section.Blocks.AddParagraph().Inlines.AddRun("par1");
            section.Blocks.AddParagraph().Inlines.AddRun(NonBreakingSpaceChar.ToString());
            section.Blocks.AddParagraph().Inlines.AddRun("par3");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body><p>par1</p><p>&nbsp</p><p>par3</p></body></html>");

            // Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }

        [TestMethod]
        public void Paragraph_WhenIsEmpty_ExportImport()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Section section = expected.Sections.AddSection();

            section.Blocks.AddParagraph().Inlines.AddRun("par1");
            section.Blocks.AddParagraph().Inlines.AddRun(NonBreakingSpaceChar.ToString());
            section.Blocks.AddParagraph().Inlines.AddRun("par3");

            // Act
            RadFlowDocument imported = this.ExportImportDocument(expected, new HtmlFormatProvider());

            // Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }

        protected override void AssertDocumentVisual(RadFlowDocument document)
        {
#if VISUAL
            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION, PDFFILEEXTENSION, HTMLFILEEXTENSION }, new System.Drawing.Rectangle(400, 200, 400, 400));
#endif
        }
    }
}
