﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.FormatProviders.Html;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;

namespace Telerik.Windows.Documents.Flow.Tests.Integration.HtmlImportExport.Paragraphs
{
    [TestClass]
    public class SpanTests : ParagraphTestsBase
    {
        [TestMethod]
        public void Test_Span_Import()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("test");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body><span>test</span></body></html>");

            //Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }

        [TestMethod]
        public void Test_Span_ExportImport()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForExportImport(false);
            expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("test");

            // Act
            RadFlowDocument actual = this.ExportImportDocument(expected, new HtmlFormatProvider());

            //Assert
            this.AssertDocuments(expected, actual, true);
        }

        [TestMethod]
        public void Test_Underline_ExportImport()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForExportImport(false);
            Run run = expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("test");
            run.Properties.UnderlinePattern.LocalValue = UnderlinePattern.Single;

            // Act
            RadFlowDocument actual = this.ExportImportDocument(expected, new HtmlFormatProvider());

            //Assert
            this.AssertDocuments(expected, actual, true);
        }

        [TestMethod]
        public void Test_TwoSpanOnSameLine_Import()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("span1");
            paragraph.Inlines.AddRun("span2");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString(
@"<html>
    <body>
        <span>span1</span><span>span2</span>
    </body>
</html>");

            //Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }

        [TestMethod]
        public void Test_TwoSpanOnSameLineSpaceBetween_Import()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("span1");
            paragraph.Inlines.AddRun(" ");
            paragraph.Inlines.AddRun("span2");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString(
@"<html>
    <body>
        <span>span1</span> <span>span2</span>
    </body>
</html>");

            //Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }

        [TestMethod]
        public void Test_TwoSpansOnDifferentLines_Import()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("span1");
            paragraph.Inlines.AddRun(" ");
            paragraph.Inlines.AddRun("span2");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString(
@"<html>
    <body>
        <span>span1</span>
        <span>span2</span>
    </body>
</html>");

            //Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }

        [TestMethod]
        public void Test_SpanThatEndsWithMultipleWhiteSpaces_Import()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("test");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body><span>test                        </span></body></html>");

            //Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }

        [TestMethod]
        public void Test_SpanThatContainsMultipleWhiteSpaces_Import()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("test test ");
            paragraph.Inlines.AddRun("test");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body><span>test                        test </span>    <span>    test</span></body></html>");

            //Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }

        [TestMethod]
        public void Test_SpanSurroundedByText_Import()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("before");
            paragraph.Inlines.AddRun(" test ");
            paragraph.Inlines.AddRun("after");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body>before<span>    test    </span>after</body></html>");

            //Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }

        [TestMethod]
        public void Test_SpanWhiteSpaceBeforeAndAfter_Import()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("test");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body>     <span>test</span>    </body></html>");

            //Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }

        [TestMethod]
        public void Test_SpanWhiteSpaceAndTextAfter_Import()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("test");
            paragraph.Inlines.AddRun(" after");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString(
@"<html><body><span>test</span>    
            after</body></html>");

            //Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }

        [TestMethod]
        public void Test_SpanAndImage_Import()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("span1");
            paragraph.Inlines.AddImageInline();
            paragraph.Inlines.AddRun("span2");
            paragraph.Inlines.AddRun(" span3");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString(
@"<html>
    <body>
	<span>span1</span>      
		<img />                 
		       <span>span2</span>
			   span3

</body>
</html>");

            //Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }

        [TestMethod]
        public void Test_TextWithWhiteSpacesBefore_Import()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("test");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString(
@"<html>
    <body>
	           test</body>
</html>");

            //Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }

        [TestMethod]
        public void Test_SpansThatEndAndBeginWithWhiteSpaces_Import()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("span1 ");
            paragraph.Inlines.AddRun("span2");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString(
@"<html>
	<body>
		<span style>span1  </span>    <span>  span2</span>
	</body>
</html>");

            //Assert
            this.AssertDocumentsStructureOnly(expected, imported);
        }
    }
}
