﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Linq;
using Telerik.Windows.Documents.Flow.FormatProviders.Html;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Media;
#if WPF || SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Media;
#else
using FontWeights = System.Windows.FontWeights;
using System.Windows.Media;
#endif

namespace Telerik.Windows.Documents.Flow.Tests.Integration.HtmlImportExport.Styles
{
    [TestClass]
    public class StylesImportTests : StyleTestsBase
    {
        [TestMethod]
        public void StyleElement_InBody_Import()
        {
            RadFlowDocument expected = CreateDocument();
            Section section = expected.Sections.AddSection();
            Paragraph paragraph1 = section.Blocks.AddParagraph();
            Style c = CreateWebBasedParagraphStyle("Class");
            c.CharacterProperties.ForegroundColor.LocalValue = new ThemableColor(Colors.Blue);
            expected.StyleRepository.Add(c);

            Style c2 = CreateWebBasedParagraphStyle("Class2");
            c2.CharacterProperties.ForegroundColor.LocalValue = new ThemableColor(Colors.Red);
            expected.StyleRepository.Add(c2);

            Style c3 = CreateWebBasedParagraphStyle("Class3");
            c3.CharacterProperties.ForegroundColor.LocalValue = new ThemableColor(Colors.Green);
            expected.StyleRepository.Add(c3);

            paragraph1.StyleId = c.Id;
            paragraph1.Inlines.AddRun("test");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><head><style>.Class2 { color: red; }</style><style>.Class3 { color: green; }</style></head><body><p class='Class'>test</p><style>.Class { color: blue; }</style></body></html>");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void LinkElement_InBody_Import()
        {
            RadFlowDocument expected = CreateDocument();
            Section section = expected.Sections.AddSection();
            Paragraph paragraph1 = section.Blocks.AddParagraph();
            Style c = CreateWebBasedParagraphStyle("Class");
            c.CharacterProperties.ForegroundColor.LocalValue = new ThemableColor(Colors.Blue);
            expected.StyleRepository.Add(c);

            paragraph1.StyleId = c.Id;
            paragraph1.Inlines.AddRun("test");

            HtmlFormatProvider provider = new HtmlFormatProvider();
            provider.ImportSettings.LoadFromUri += (s, e) => { e.SetData(System.Text.Encoding.UTF8.GetBytes(".Class { color: blue; }")); };

            // Act
            RadFlowDocument imported = provider.Import("<html><head></head><body><p class='Class'>test</p><link type='text/css' rel='stylesheet' href='dfsdf'></body></html>");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void LinkElement_WhenRelationIsNotStyleSheet_Import()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();
            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph.StyleId = BuiltInStyleNames.NormalWebStyleId;
            paragraph.Inlines.AddRun("test");

            // Act
            HtmlFormatProvider provider = new HtmlFormatProvider();
            RadFlowDocument imported = provider.Import("<html><head><link rel='themeData' href='test'></head><body><p>test</p></body></html>");

            // Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void LinkElement_External_WhenTypeIsProvided_Import()
        {
            // Arrange
            int counter = 0;
            HtmlFormatProvider provider = new HtmlFormatProvider();
            provider.ImportSettings.LoadFromUri += (s, e) => 
            {
                e.SetData(System.Text.Encoding.UTF8.GetBytes(".Class { color: blue; }"));
                counter++;
            };

            // Act
            RadFlowDocument imported = provider.Import("<html><head><link rel='stylesheet' type='text/css' href='test'/></head><body><p class='Class'>test</p></body></html>");

            // Assert
            Assert.AreEqual(1, counter);
        }

        [TestMethod]
        public void LinkElement_External_WhenTypeIsNotProvided_Import()
        {
            // Arrange
            int counter = 0;
            HtmlFormatProvider provider = new HtmlFormatProvider();
            provider.ImportSettings.LoadFromUri += (s, e) =>
            {
                e.SetData(System.Text.Encoding.UTF8.GetBytes(".Class { color: blue; }"));
                counter++;
            };

            // Act
            RadFlowDocument imported = provider.Import("<html><head><link rel='stylesheet' href='test'/></head><body><p class='Class'>test</p></body></html>");

            // Assert
            Assert.AreEqual(1, counter);
        }

        [TestMethod]
        public void LinkElement_External_WhenTypeIsInvalid_Import()
        {
            // Arrange
            int counter = 0;
            HtmlFormatProvider provider = new HtmlFormatProvider();
            provider.ImportSettings.LoadFromUri += (s, e) =>
            {
                e.SetData(System.Text.Encoding.UTF8.GetBytes(".Class { color: blue; }"));
                counter++;
            };

            // Act
            RadFlowDocument imported = provider.Import("<html><head><link rel='stylesheet' type='invalid' href='test'></head><body><p class='Class'>test</p></body></html>");

            // Assert
            Assert.AreEqual(0, counter);
        }

        [TestMethod]
        public void Paragraph_InheritedPropertiesAreCleared()
        {
            RadFlowDocument expected = this.GetExpectedDocumentForImport(true);
            Section section = expected.Sections.AddSection();

            Paragraph paragraph1 = section.Blocks.AddParagraph();
            paragraph1.StyleId = BuiltInStyleNames.NormalWebStyleId;
            paragraph1.Properties.LineSpacing.LocalValue = 1.75;
            paragraph1.Properties.LineSpacingType.LocalValue = HeightType.Auto;
            paragraph1.Inlines.AddRun("p1");

            Paragraph paragraph2 = section.Blocks.AddParagraph();
            paragraph2.StyleId = BuiltInStyleNames.NormalWebStyleId;
            paragraph2.Inlines.AddRun("p2");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><body><div style='line-height: 175%'><p>p1</p></div><p>p2</p></body></html>");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void RelativeStyleProperties_AreConvertedToLocalPropertiesAndInheritsValues()
        {
            RadFlowDocument expected = CreateDocument();
            Style class1 = CreateWebBasedParagraphStyle("Class1");
            class1.CharacterProperties.FontSize.LocalValue = 6.666666666666666667;
            expected.StyleRepository.Add(class1);

            Section section = expected.Sections.AddSection();

            Paragraph paragraph1 = section.Blocks.AddParagraph();
            paragraph1.StyleId = class1.Id;
            paragraph1.Properties.ParagraphMarkerProperties.FontSize.LocalValue = 75;
            paragraph1.Inlines.AddRun("P1").FontSize = 75;

            Paragraph paragraph2 = section.Blocks.AddParagraph();
            paragraph2.StyleId = class1.Id;
            paragraph2.Properties.ParagraphMarkerProperties.FontSize.LocalValue = 50;
            paragraph2.Inlines.AddRun("P2").FontSize = 50;

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><head><style>.Class1 { font-size: 50%; } </style></head><body><div style='font-size: 150px'><p class='Class1'>P1</p></div><div style='font-size: 100px'><p class='Class1'>P2</p></div></body></html>");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void InheritedProperties_DoesNotOverridesStyleProperties()
        {
            RadFlowDocument expected = CreateDocument();
            Style class1 = CreateWebBasedParagraphStyle("Class1");
            class1.CharacterProperties.ForegroundColor.LocalValue = new ThemableColor(Colors.Red);
            expected.StyleRepository.Add(class1);

            Section section = expected.Sections.AddSection();

            Paragraph paragraph1 = section.Blocks.AddParagraph();
            paragraph1.StyleId = class1.Id;
            paragraph1.Inlines.AddRun("test");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><style> body { color: blue;} .Class1 { color: red;} </style><body><p class='Class1'>test</p></body></html>");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ClassPriorityNoStyle_IsFirstClassApplied()
        {
            RadFlowDocument expected = CreateDocument();
            Style class1 = CreateWebBasedParagraphStyle("Class1");
            class1.CharacterProperties.ForegroundColor.LocalValue = new ThemableColor(Colors.Red);
            class1.CharacterProperties.FontSize.LocalValue = 50;
            expected.StyleRepository.Add(class1);

            Style class2 = CreateWebBasedParagraphStyle("Class2");
            class2.CharacterProperties.ForegroundColor.LocalValue = new ThemableColor(Colors.Blue);
            expected.StyleRepository.Add(class2);

            Section section = expected.Sections.AddSection();

            Paragraph paragraph1 = section.Blocks.AddParagraph();
            paragraph1.Properties.ParagraphMarkerProperties.ForegroundColor.LocalValue = new ThemableColor(Colors.Blue);
            paragraph1.Properties.ParagraphMarkerProperties.FontSize.LocalValue = 50;
            Run run = paragraph1.Inlines.AddRun("test");
            run.ForegroundColor = new ThemableColor(Colors.Blue);
            run.FontSize = 50;

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><style>.Class1 { color: red; font-size: 50px; } .Class2 { color: blue; }</style><body><div class='Class1 Class2'>test</div></body></html>");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void DivWithClassName_IsStyleImportedAsLocalProperties()
        {
            // Arrange
            ThemableFontFamily fontFamily = new ThemableFontFamily("Comic Sans MS");
            RadFlowDocument expected = CreateDocument();
            Style style = CreateWebBasedParagraphStyle("myStyle");
            style.CharacterProperties.FontFamily.LocalValue = fontFamily;
            expected.StyleRepository.Add(style);

            Section section = expected.Sections.AddSection();

            Paragraph paragraph1 = section.Blocks.AddParagraph();
            paragraph1.Properties.ParagraphMarkerProperties.FontFamily.LocalValue = fontFamily;
            Run run = paragraph1.Inlines.AddRun("test");
            run.FontFamily = fontFamily;

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<html><style>.myStyle { font-family: 'Comic Sans MS';}</style><body><div class='myStyle'>test</div></body></html>");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ParagraphWithEmptyClassName_IsImportedWithNormalWebStyle()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();

            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph.StyleId = BuiltInStyleNames.NormalWebStyleId;
            paragraph.Inlines.AddRun("text");

            // Act 
            RadFlowDocument imported = this.ImportHtml("ParagraphWithEmptyClassName.html");

            // Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ParagraphWithoutClassName_IsImportedWithNormalWebStyle()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport(true);

            Section section = expected.Sections.AddSection();

            Paragraph paragraph1 = section.Blocks.AddParagraph();
            paragraph1.StyleId = BuiltInStyleNames.NormalWebStyleId;
            paragraph1.Inlines.AddRun("text");

            // Act
            RadFlowDocument imported = this.ImportHtml("ParagraphWithoutClassName.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void SpanWithoutClassName_NoStylesAreImported()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Section section = expected.Sections.AddSection();

            Paragraph paragraph1 = section.Blocks.AddParagraph();
            paragraph1.Inlines.AddRun("text");

            // Act
            RadFlowDocument imported = this.ImportHtml("SpanWithoutClassName.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ClassNamesUsedOnSpansWithoutStyles_AreImportedAsCharacterStyles()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Style class1Style = new Style("Class1", StyleType.Character);
            expected.StyleRepository.Add(class1Style);

            Style class2Style = new Style("Class2", StyleType.Character);
            expected.StyleRepository.Add(class2Style);

            Section section = expected.Sections.AddSection();

            Paragraph paragraph = section.Blocks.AddParagraph();
            Run run1 = paragraph.Inlines.AddRun("text1");
            run1.StyleId = class1Style.Id;
            Run run2 = paragraph.Inlines.AddRun("text2");
            run2.StyleId = class2Style.Id;

            // Act
            RadFlowDocument imported = this.ImportHtml("ClassNamesUsedOnSpansWithoutStyles.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ClassNamesUsedOnParagraphsWithoutStyles_AreImportedAsParagraphStyles()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Style class1Style = CreateWebBasedParagraphStyle("Class1");
            class1Style.BasedOnStyleId = BuiltInStyleNames.NormalStyleId;
            expected.StyleRepository.Add(class1Style);

            Style class2Style = CreateWebBasedParagraphStyle("Class2");
            class2Style.BasedOnStyleId = BuiltInStyleNames.NormalStyleId;
            expected.StyleRepository.Add(class2Style);

            Section section = expected.Sections.AddSection();

            Paragraph paragraph1 = section.Blocks.AddParagraph();
            paragraph1.StyleId = class1Style.Id;
            paragraph1.Inlines.AddRun("text1");

            Paragraph paragraph2 = section.Blocks.AddParagraph();
            paragraph2.StyleId = class2Style.Id;
            paragraph2.Inlines.AddRun("text2");

            // Act
            RadFlowDocument imported = this.ImportHtml("ClassNamesUsedOnParagraphsWithoutStyles.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ClassNamesUsedOnParagraphsAndSpansWithoutStyles_AreImportedAsParagraphAndChracterStyle()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Style class1ParagraphStyle = new Style("Class1", StyleType.Paragraph);
            class1ParagraphStyle.BasedOnStyleId = BuiltInStyleNames.NormalStyleId;
            class1ParagraphStyle.ParagraphProperties.AutomaticSpacingAfter.LocalValue = true;
            class1ParagraphStyle.ParagraphProperties.AutomaticSpacingBefore.LocalValue = true;
            class1ParagraphStyle.ParagraphProperties.SpacingAfter.LocalValue = 6.66;
            class1ParagraphStyle.ParagraphProperties.SpacingBefore.LocalValue = 6.66;
            expected.StyleRepository.Add(class1ParagraphStyle);

            Style class11CharacterStyle = new Style("Class11", StyleType.Character);
            expected.StyleRepository.Add(class11CharacterStyle);

            Section section = expected.Sections.AddSection();

            Paragraph paragraph1 = section.Blocks.AddParagraph();
            paragraph1.StyleId = class1ParagraphStyle.Id;
            paragraph1.Inlines.AddRun("text1");

            Paragraph paragraph2 = section.Blocks.AddParagraph();
            Run run = paragraph2.Inlines.AddRun("text2");
            run.StyleId = class11CharacterStyle.Id;

            // Act
            RadFlowDocument imported = this.ImportHtml("ClassNamesUsedOnParagraphsAndSpansWithoutStyles.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void MultipleClassNamesUsedParagraphWithoutStyles_OnlyLastClassNameIsImported()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Style class2Style = CreateWebBasedParagraphStyle("Class1");
            class2Style.BasedOnStyleId = BuiltInStyleNames.NormalStyleId;
            expected.StyleRepository.Add(class2Style);

            Section section = expected.Sections.AddSection();

            Paragraph paragraph1 = section.Blocks.AddParagraph();
            paragraph1.StyleId = class2Style.Id;
            paragraph1.Inlines.AddRun("text");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString("<p class=\"Class1 Class2\"><font>text</font></p>");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void CharacterPropertyStyleAppliedOnSpan_IsImportedAsParagraphAndCharacterStyle()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Style class1ParagraphStyle = StyleTestsBase.CreateWebBasedParagraphStyle("Class1");
            class1ParagraphStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            expected.StyleRepository.Add(class1ParagraphStyle);

            Style class11CharacterStyle = new Style("Class11", StyleType.Character);
            class11CharacterStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            expected.StyleRepository.Add(class11CharacterStyle);

            Run run = expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("text");
            run.StyleId = class11CharacterStyle.Id;

            // Act
            RadFlowDocument imported = this.ImportHtml("CharacterPropertyStyleAppliedOnSpan.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void CharacterPropertyStyleAppliedOnParagraph_IsImportedAsParagraphStyle()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Style class1ParagraphStyle = StyleTestsBase.CreateWebBasedParagraphStyle("Class1");
            class1ParagraphStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            expected.StyleRepository.Add(class1ParagraphStyle);

            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph.StyleId = class1ParagraphStyle.Id;
            paragraph.Inlines.AddRun("text");

            // Act
            RadFlowDocument imported = this.ImportHtml("CharacterPropertyStyleAppliedOnParagraph.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void CharacterPropertyStyleAppliedOnParagraphAndSpan_IsImportedAsParagraphAndCharacterStyle()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Style class1ParagraphStyle = StyleTestsBase.CreateWebBasedParagraphStyle("Class1");
            class1ParagraphStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            expected.StyleRepository.Add(class1ParagraphStyle);

            Style class11CharacterStyle = new Style("Class11", StyleType.Character);
            class11CharacterStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            expected.StyleRepository.Add(class11CharacterStyle);

            Section section = expected.Sections.AddSection();
            Paragraph paragraph1 = section.Blocks.AddParagraph();
            paragraph1.StyleId = class1ParagraphStyle.Id;
            paragraph1.Inlines.AddRun("text1");

            Run run = section.Blocks.AddParagraph().Inlines.AddRun("text2");
            run.StyleId = class11CharacterStyle.Id;

            // Act
            RadFlowDocument imported = this.ImportHtml("CharacterPropertyStyleAppliedOnParagraphAndSpan.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void CharacterPropertyStyleNotApplied_IsImportedAsParagraphStyle()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Style class1ParagraphStyle = StyleTestsBase.CreateWebBasedParagraphStyle("Class1");
            class1ParagraphStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            expected.StyleRepository.Add(class1ParagraphStyle);

            expected.Sections.AddSection();

            // Act
            RadFlowDocument imported = this.ImportHtml("CharacterPropertyStyleNotApplied.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ParagraphPropertyStyleAppliedOnParagraph_IsImportedAsParagraphStyle()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Style class1ParagraphStyle = StyleTestsBase.CreateWebBasedParagraphStyle("Class1");
            class1ParagraphStyle.ParagraphProperties.TextAlignment.LocalValue = Alignment.Center;
            expected.StyleRepository.Add(class1ParagraphStyle);

            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph.StyleId = class1ParagraphStyle.Id;
            paragraph.Inlines.AddRun("text");

            // Act
            RadFlowDocument imported = this.ImportHtml("ParagraphPropertyStyleAppliedOnParagraph.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ParagraphPropertyStyleAppliedOnSpan_IsImportedAsParagraphAndCharacterStyle()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Style class1ParagraphStyle = StyleTestsBase.CreateWebBasedParagraphStyle("Class1");
            class1ParagraphStyle.ParagraphProperties.TextAlignment.LocalValue = Alignment.Center;
            expected.StyleRepository.Add(class1ParagraphStyle);

            Style class11CharacterStyle = new Style("Class11", StyleType.Character);
            expected.StyleRepository.Add(class11CharacterStyle);

            Run run = expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("text");
            run.StyleId = class11CharacterStyle.Id;

            // Act
            RadFlowDocument imported = this.ImportHtml("ParagraphPropertyStyleAppliedOnSpan.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ParagraphPropertyStyleAppliedOnParagraphAndSpan_IsImportedAsParagraphAndCharacterStyle()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Style class1ParagraphStyle = StyleTestsBase.CreateWebBasedParagraphStyle("Class1");
            class1ParagraphStyle.ParagraphProperties.TextAlignment.LocalValue = Alignment.Center;
            expected.StyleRepository.Add(class1ParagraphStyle);

            Style class11CharacterStyle = new Style("Class11", StyleType.Character);
            expected.StyleRepository.Add(class11CharacterStyle);

            Section section = expected.Sections.AddSection();
            Paragraph paragraph1 = section.Blocks.AddParagraph();
            paragraph1.StyleId = class1ParagraphStyle.Id;
            paragraph1.Inlines.AddRun("text1");

            Run run = section.Blocks.AddParagraph().Inlines.AddRun("text2");
            run.StyleId = class11CharacterStyle.Id;

            // Act
            RadFlowDocument imported = this.ImportHtml("ParagraphPropertyStyleAppliedOnParagraphAndSpan.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ParagraphPropertyStyleNotApplied_IsImportedAsParagraphStyle()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Style class1ParagraphStyle = StyleTestsBase.CreateWebBasedParagraphStyle("Class1");
            class1ParagraphStyle.ParagraphProperties.TextAlignment.LocalValue = Alignment.Center;
            expected.StyleRepository.Add(class1ParagraphStyle);

            expected.Sections.AddSection();

            // Act
            RadFlowDocument imported = this.ImportHtml("ParagraphPropertyStyleNotApplied.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ParagraphAndCharacterPropertyStyleAppliedOnParagraph_IsImportedAsParagraphStyle()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Style class1ParagraphStyle = StyleTestsBase.CreateWebBasedParagraphStyle("Class1");
            class1ParagraphStyle.ParagraphProperties.TextAlignment.LocalValue = Alignment.Center;
            class1ParagraphStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            expected.StyleRepository.Add(class1ParagraphStyle);

            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph.StyleId = class1ParagraphStyle.Id;
            paragraph.Inlines.AddRun("text");

            // Act
            RadFlowDocument imported = this.ImportHtml("ParagraphAndCharacterPropertyStyleAppliedOnParagraph.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ParagraphAndCharacterPropertyStyleNotApplied_IsImportedAsParagraphStyle()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Style class1ParagraphStyle = StyleTestsBase.CreateWebBasedParagraphStyle("Class1");
            class1ParagraphStyle.ParagraphProperties.TextAlignment.LocalValue = Alignment.Center;
            class1ParagraphStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            expected.StyleRepository.Add(class1ParagraphStyle);

            expected.Sections.AddSection();

            // Act
            RadFlowDocument imported = this.ImportHtml("ParagraphAndCharacterPropertyStyleNotApplied.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ParagraphAndCharacterPropertyStyleAppliedOnSpan_IsImportedAsParagraphAndCharacterStyle()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Style class1ParagraphStyle = StyleTestsBase.CreateWebBasedParagraphStyle("Class1");
            class1ParagraphStyle.ParagraphProperties.TextAlignment.LocalValue = Alignment.Center;
            class1ParagraphStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            expected.StyleRepository.Add(class1ParagraphStyle);

            Style class11CharacterStyle = new Style("Class11", StyleType.Character);
            class11CharacterStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            expected.StyleRepository.Add(class11CharacterStyle);

            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            Run run = paragraph.Inlines.AddRun("text");
            run.StyleId = class11CharacterStyle.Id;

            // Act
            RadFlowDocument imported = this.ImportHtml("ParagraphAndCharacterPropertyStyleAppliedOnSpan.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ParagraphAndCharacterPropertyStyleAppliedOnParagraphAndSpan_IsImportedAsParagraphAndCharacterStyle()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Style class1ParagraphStyle = StyleTestsBase.CreateWebBasedParagraphStyle("Class1");
            class1ParagraphStyle.ParagraphProperties.TextAlignment.LocalValue = Alignment.Center;
            class1ParagraphStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            expected.StyleRepository.Add(class1ParagraphStyle);

            Style class11CharacterStyle = new Style("Class11", StyleType.Character);
            class11CharacterStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            expected.StyleRepository.Add(class11CharacterStyle);

            Section section = expected.Sections.AddSection();
            Paragraph paragraph1 = section.Blocks.AddParagraph();
            paragraph1.StyleId = class1ParagraphStyle.Id;
            paragraph1.Inlines.AddRun("text1");

            Run run = section.Blocks.AddParagraph().Inlines.AddRun("text2");
            run.StyleId = class11CharacterStyle.Id;

            // Act
            RadFlowDocument imported = this.ImportHtml("ParagraphAndCharacterPropertyStyleAppliedOnParagraphAndSpan.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void DefaultParagraphStyle_IsImportedAsNormalWebStyle()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport(true);

            Style normalWebStyle = expected.StyleRepository.GetStyle(BuiltInStyleNames.NormalWebStyleId);
            normalWebStyle.ParagraphProperties.TextAlignment.LocalValue = Alignment.Center;
            normalWebStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");

            Section section = expected.Sections.AddSection();

            Paragraph paragraph1 = section.Blocks.AddParagraph();
            paragraph1.StyleId = normalWebStyle.Id;
            paragraph1.Inlines.AddRun("text");

            // Act
            RadFlowDocument imported = this.ImportHtml("DefaultParagraphStyle.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void DefaultParagraphStyleWithMarginBottom_IsImportedAsNormalWebWithSpacingAfter()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport(true);

            Style normalWebStyle = expected.StyleRepository.GetStyle(BuiltInStyleNames.NormalWebStyleId);
            normalWebStyle.ParagraphProperties.AutomaticSpacingAfter.LocalValue = false;
            normalWebStyle.ParagraphProperties.SpacingAfter.LocalValue = 40;

            Paragraph paragraph1 = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph1.StyleId = normalWebStyle.Id;
            paragraph1.Inlines.AddRun("text");

            // Act
            RadFlowDocument imported = this.ImportHtml("DefaultParagraphStyleWithMarginBottom.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void DefaultParagraphStyleWithMarginTop_IsImportedAsNormalWebWithSpacingBefore()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport(true);

            Style normalWebStyle = expected.StyleRepository.GetStyle(BuiltInStyleNames.NormalWebStyleId);
            normalWebStyle.ParagraphProperties.AutomaticSpacingBefore.LocalValue = false;
            normalWebStyle.ParagraphProperties.SpacingBefore.LocalValue = 40;

            Paragraph paragraph1 = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph1.StyleId = normalWebStyle.Id;
            paragraph1.Inlines.AddRun("text");

            // Act
            RadFlowDocument imported = this.ImportHtml("DefaultParagraphStyleWithMarginTop.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void DefaultParagraphStyleAndParagraphPropertiesSetOnParagraph_PropertiesAreSetAsLocalValues()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport(true);

            Style normalWebStyle = expected.StyleRepository.GetStyle(BuiltInStyleNames.NormalWebStyleId);
            normalWebStyle.ParagraphProperties.TextAlignment.LocalValue = Alignment.Center;

            Section section = expected.Sections.AddSection();

            Paragraph paragraph1 = section.Blocks.AddParagraph();
            paragraph1.StyleId = normalWebStyle.Id;
            paragraph1.TextAlignment = Alignment.Right;
            paragraph1.Inlines.AddRun("text");

            // Act
            RadFlowDocument imported = this.ImportHtml("DefaultParagraphStyleAndParagraphPropertiesSetOnParagraph.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void DefaultParagraphStyleAndCharacterPropertiesSetOnParagraph_PropertiesAreSetAsLocalValuesOnRunAndParagraphSymbol()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport(true);

            Style normalWebStyle = expected.StyleRepository.GetStyle(BuiltInStyleNames.NormalWebStyleId);
            normalWebStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");

            Section section = expected.Sections.AddSection();

            Paragraph paragraph1 = section.Blocks.AddParagraph();
            paragraph1.StyleId = normalWebStyle.Id;
            paragraph1.Properties.ParagraphMarkerProperties.FontFamily.LocalValue = new ThemableFontFamily("Courier New");
            Run run1 = paragraph1.Inlines.AddRun("text");
            run1.FontFamily = new ThemableFontFamily("Courier New");

            // Act
            RadFlowDocument imported = this.ImportHtml("DefaultParagraphStyleAndCharacterPropertiesSetOnParagraph.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void DefaultSpanStyle_IsAppliedAsLocalValues()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Section section = expected.Sections.AddSection();

            Run run = section.Blocks.AddParagraph().Inlines.AddRun("text");
            run.FontFamily = new ThemableFontFamily("Comic Sans MS");

            // Act
            RadFlowDocument imported = this.ImportHtml("DefaultSpanStyle.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void DefaultSpanStyleAndPropertiesSetOnSpan_PropertiesAreSetAsLocalValues()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Section section = expected.Sections.AddSection();

            Run run = section.Blocks.AddParagraph().Inlines.AddRun("text");
            run.FontFamily = new ThemableFontFamily("Courier New");

            // Act
            RadFlowDocument imported = this.ImportHtml("DefaultSpanStyleAndPropertiesSetOnSpan.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void DefaultSpanStyleAndClassSetOnSpan_DefaultSpanStylePropertiesAreAddedToTheCharacterStyle()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Style class1Style = new Style("Class1", StyleType.Character);
            class1Style.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            expected.StyleRepository.Add(class1Style);

            //expected.StyleRepository.

            Section section = expected.Sections.AddSection();

            Run run = section.Blocks.AddParagraph().Inlines.AddRun("text");
            run.StyleId = class1Style.Id;

            // Act
            RadFlowDocument imported = this.ImportHtml("DefaultSpanStyleAndClassSetOnSpan.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void LocalParagraphPropertiesSetOnParagraph_PropertiesAreSetAsLocalValues()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport(true);

            Style normalWebStyle = expected.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.NormalWebStyleId);

            Paragraph paragraph1 = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph1.StyleId = normalWebStyle.Id;
            paragraph1.TextAlignment = Alignment.Center;
            paragraph1.Inlines.AddRun("text");

            // Act
            RadFlowDocument imported = this.ImportHtml("LocalParagraphPropertiesSetOnParagraph.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void LocalCharacterPropertiesSetOnParagraph_ProperiesAreSetAsLocalValuesOnRunAndParagraphSymbol()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport(true);

            Paragraph paragraph1 = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph1.StyleId = BuiltInStyleNames.NormalWebStyleId;
            paragraph1.Properties.ParagraphMarkerProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            Run run1 = paragraph1.Inlines.AddRun("text");
            run1.FontFamily = new ThemableFontFamily("Comic Sans MS");

            // Act
            RadFlowDocument imported = this.ImportHtml("LocalCharacterPropertiesSetOnParagraph.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void LocalCharacterPropertiesSetOnSpan_ProperiesAreSetAsLocalValuesOnRun()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Paragraph paragraph1 = expected.Sections.AddSection().Blocks.AddParagraph();
            Run run1 = paragraph1.Inlines.AddRun("text");
            run1.FontFamily = new ThemableFontFamily("Comic Sans MS");

            // Act
            RadFlowDocument imported = this.ImportHtml("LocalCharacterPropertiesSetOnSpan.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        // TODO uncomment when importing of default hyperlink style is ready.
        //[TestMethod]
        public void LocalCharacterPropertiesSetOnHyperlink_ProperiesAreSetAsLocalValues()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(expected);
            editor.InsertHyperlink("text", "www.telerik.com", false);
            Run run1 = expected.EnumerateChildrenOfType<Run>().First();
            run1.FontFamily = new ThemableFontFamily("Comic Sans MS");

            // Act
            RadFlowDocument imported = this.ImportHtml("LocalCharacterPropertiesSetOnHyperlink.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void LocalPropertiesSetOnDivContainingText_PropertiesAreSetAsLocalValuesOnParagraphAndRun()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Paragraph paragraph1 = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph1.TextAlignment = Alignment.Center;
            paragraph1.Properties.ParagraphMarkerProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            Run run1 = paragraph1.Inlines.AddRun("text");
            run1.FontFamily = new ThemableFontFamily("Comic Sans MS");

            // Act
            RadFlowDocument imported = this.ImportHtml("LocalPropertiesSetOnDivContainingText.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void LocalPropertiesSetOnDivContainingSpan_PropertiesAreSetAsLocalValuesOnParagraphAndRun()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Paragraph paragraph1 = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph1.TextAlignment = Alignment.Center;
            paragraph1.Properties.ParagraphMarkerProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            Run run1 = paragraph1.Inlines.AddRun("text");
            run1.FontFamily = new ThemableFontFamily("Comic Sans MS");

            // Act
            RadFlowDocument imported = this.ImportHtml("LocalPropertiesSetOnDivContainingSpan.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void LocalPropertiesSetOnDivContainingParagraph_PropertiesAreSetAsLocalValuesOnParagraphAndRun()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport(true);

            Paragraph paragraph1 = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph1.StyleId = BuiltInStyleNames.NormalWebStyleId;
            paragraph1.TextAlignment = Alignment.Center;
            paragraph1.Properties.ParagraphMarkerProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            Run run1 = paragraph1.Inlines.AddRun("text");
            run1.FontFamily = new ThemableFontFamily("Comic Sans MS");

            // Act
            RadFlowDocument imported = this.ImportHtml("LocalPropertiesSetOnDivContainingParagraph.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void IdSelectorStyleAppliedOnParagraph_IsSetAsLocalValues()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport(true);
            Style normalWebStyle = expected.StyleRepository.GetStyle(BuiltInStyleNames.NormalWebStyleId);

            Paragraph paragraph1 = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph1.StyleId = normalWebStyle.Id;
            paragraph1.TextAlignment = Alignment.Center;
            paragraph1.Inlines.AddRun("text");

            // Act
            RadFlowDocument imported = this.ImportHtml("IdSelectorStyleAppliedOnParagraph.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void IdSelectorStyleAppliedOnSpan_IsSetAsLocalValues()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Run run1 = expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("text");
            run1.FontFamily = new ThemableFontFamily("Comic Sans MS");

            // Act
            RadFlowDocument imported = this.ImportHtml("IdSelectorStyleAppliedOnSpan.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void IdSelectorStyleAppliedOnDivContainingSpan_IsAppliedAsLocalProperties()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            Paragraph paragraph1 = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph1.TextAlignment = Alignment.Center;
            paragraph1.Properties.ParagraphMarkerProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");

            Run run1 = paragraph1.Inlines.AddRun("text");
            run1.FontFamily = new ThemableFontFamily("Comic Sans MS");

            // Act
            RadFlowDocument imported = this.ImportHtml("IdSelectorStyleAppliedOnDivContainingSpan.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void StyleWithBuiltInName_IsImportedWithChangedNamed()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();
            Style paragraphStyle = StyleTestsBase.CreateWebBasedParagraphStyle("Normal1");
            paragraphStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");

            expected.StyleRepository.Add(paragraphStyle);

            Style characterStyle = new Style("Normal11", StyleType.Character);
            characterStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");
            expected.StyleRepository.Add(characterStyle);

            Paragraph paragraph1 = expected.Sections.AddSection().Blocks.AddParagraph();
            Run run = paragraph1.Inlines.AddRun("text");
            run.StyleId = characterStyle.Id;

            // Act
            RadFlowDocument imported = this.ImportHtml("StyleWithBuiltInName.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ParagraphStyleWithTelerikPrefixAndBuiltInName_IsImportedAsBuiltInStyle()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();
            Style style = expected.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.CaptionStyleId);
            Paragraph paragraph1 = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph1.StyleId = style.Id;
            paragraph1.Inlines.AddRun("text");

            // Act
            RadFlowDocument imported = this.ImportHtml("ParagraphStyleWithTelerikPrefixAndBuiltInName.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void CharacterStyleWithTelerikPrefixAndBuiltInName_IsImportedAsBuiltInStyle()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();
            Style style = expected.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.FootnoteReferenceStyleId);
            Run run = expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("text");
            run.StyleId = style.Id;

            // Act
            RadFlowDocument imported = this.ImportHtml("CharacterStyleWithTelerikPrefixAndBuiltInName.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ParagraphWithTelerikNormalStyle_IsImportedWithoutStyleId()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();
            expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("text");

            // Act
            RadFlowDocument imported = this.ImportHtml("ParagraphWithTelerikNormalStyle.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void ParagraphWithTelerikNormalStyle_WhenStyleIsModified_IsImportedAndNormalStyleIsModified()
        {
            // Arrange
            RadFlowDocument expected = CreateDocument();

            expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("text");
            Style normalStyle = expected.StyleRepository.GetStyle(BuiltInStyleNames.NormalStyleId);
            normalStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily(new FontFamily("Verdana"));
            normalStyle.CharacterProperties.FontSize.LocalValue = 30;

            // Act
            RadFlowDocument imported = this.ImportHtml("ParagraphWithModifiedTelerikNormalStyle.html");

            //Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void TableWithTelerikTableNormalStyle_IsImportedWithoutStyleId()
        {
            // Arrange 
            RadFlowDocument expected = this.GetExpectedDocumentForImport();

            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddTable().Rows.AddTableRow().Cells.AddTableCell().Blocks.AddParagraph();
            paragraph.StyleId = BuiltInStyleNames.NormalWebStyleId;
            paragraph.Inlines.AddRun("text");

            // Act
            RadFlowDocument imported = this.ImportHtml("TableWithTelerikTableNormalStyle.html");

            // Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void TableWithModifiedTelerikTableNormalStyle_TablePropertiesAreInlinedAndTableNormalStyleIsModified()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();

            Table table = expected.Sections.AddSection().Blocks.AddTable();
            Border border = new Border(2, BorderStyle.Single, new ThemableColor(Colors.Green));
            table.Borders = new TableBorders(border, border, border, border);

            Style tableNormalStyle = expected.StyleRepository.GetStyle(BuiltInStyleNames.NormalTableStyleId);
            tableNormalStyle.TableProperties.Borders.LocalValue = new TableBorders(border, border, border, border);
            tableNormalStyle.TableCellProperties.Borders.LocalValue = new TableCellBorders(border, border, border, border);

            TableRow row = table.Rows.AddTableRow();
            TableCell cell = row.Cells.AddTableCell();
            Paragraph paragraph = cell.Blocks.AddParagraph();
            paragraph.StyleId = BuiltInStyleNames.NormalWebStyleId;
            paragraph.Inlines.AddRun("text");

            // Act
            RadFlowDocument imported = this.ImportHtml("TableWithModifiedTelerikTableNormalStyle.html");

            // Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void TableWithClassName_TablePropertiesAreInlined()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();

            Style style = CreateWebBasedParagraphStyle("tablestyle");
            expected.StyleRepository.Add(style);

            Table table = expected.Sections.AddSection().Blocks.AddTable();
            Border border = new Border(2, BorderStyle.Single, new ThemableColor(Colors.Green));
            table.Borders = new TableBorders(border, border, border, border);
            table.StyleId = style.Id;

            TableRow row = table.Rows.AddTableRow();
            TableCell cell = row.Cells.AddTableCell();
            Paragraph paragraph = cell.Blocks.AddParagraph();
            paragraph.StyleId = BuiltInStyleNames.NormalWebStyleId;
            paragraph.Inlines.AddRun("test");

            // Act 
            RadFlowDocument imported = this.ImportHtml("TableWithClassContainingTableProperties.html");

            // Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void TableWithEmptyClassAttribute_IsImportedWithTableNormal()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();

            Table table = expected.Sections.AddSection().Blocks.AddTable();

            Paragraph paragraph = table.Rows.AddTableRow().Cells.AddTableCell().Blocks.AddParagraph();
            paragraph.StyleId = BuiltInStyleNames.NormalWebStyleId;
            paragraph.Inlines.AddRun("test");

            // Act
            RadFlowDocument imported = this.ImportHtml("TableWithEmptyClassName.html");

            // Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void SpanWithDifferentFormatting_WhenHasBackgroundAndPartOfInnerTextIsBold_BackgroundIsAppliedOnBoldedText()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport(false);
            Section section = expected.Sections.AddSection();
            Paragraph paragraph = section.Blocks.AddParagraph();

            Run run1 = new Run(expected);
            run1.Text = "outside";
            run1.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Red);
            run1.Properties.ForegroundColor.LocalValue = new ThemableColor(Colors.Green);
            run1.Properties.FontSize.LocalValue = 24;
            paragraph.Inlines.Add(run1);

            Run run2 = new Run(expected);
            run2.Text = "inside text";
            run2.Properties.CopyPropertiesFrom(run1.Properties);
            run2.Properties.FontWeight.LocalValue = FontWeights.Bold;
            paragraph.Inlines.Add(run2);

            Run run3 = new Run(expected);
            run3.Text = "outside";
            run3.Properties.CopyPropertiesFrom(run1.Properties);
            paragraph.Inlines.Add(run3);

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString(@"<html>
<body>
<span style=""font-size: 24px; color: #008000; background-color: #FF0000;"">outside<b>inside text</b>outside</span>
</body>
</html>");

            // Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void SpanWithDifferentFormatting_WhenHasParagraphElementInside_NonInheritablePropertiesAreNotInherittedFromParagraphElement()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();
            Section section = expected.Sections.AddSection();

            Paragraph firstParagraph = section.Blocks.AddParagraph();
            Run firstParagraphRun1 = new Run(expected);
            firstParagraphRun1.Text = "outside";
            firstParagraphRun1.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Red);
            firstParagraphRun1.Properties.ForegroundColor.LocalValue = new ThemableColor(Colors.Green);
            firstParagraphRun1.Properties.FontSize.LocalValue = 24;
            firstParagraph.Inlines.Add(firstParagraphRun1);

            Run firstParagraphRun2 = new Run(expected);
            firstParagraphRun2.Text = "inside text";
            firstParagraphRun2.Properties.CopyPropertiesFrom(firstParagraphRun1.Properties);
            firstParagraphRun2.Properties.FontWeight.LocalValue = FontWeights.Bold;
            firstParagraph.Inlines.Add(firstParagraphRun2);

            Paragraph secondParagraph = section.Blocks.AddParagraph();
            secondParagraph.StyleId = BuiltInStyleNames.NormalWebStyleId;
            secondParagraph.Properties.ParagraphMarkerProperties.CopyPropertiesFrom(firstParagraphRun2.Properties);
            secondParagraph.Properties.ParagraphMarkerProperties.BackgroundColor.ClearValue();
            Run secondParagraphRun = new Run(expected);
            secondParagraphRun.Text = "paragraph";
            secondParagraphRun.Properties.CopyPropertiesFrom(firstParagraphRun2.Properties);
            secondParagraphRun.Properties.BackgroundColor.ClearValue();
            secondParagraph.Inlines.Add(secondParagraphRun);

            Paragraph thirdParagraph = section.Blocks.AddParagraph();
            thirdParagraph.Properties.ParagraphMarkerProperties.CopyPropertiesFrom(firstParagraphRun1.Properties);
            thirdParagraph.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Red);
            Run thirdParagraphRun = new Run(expected);
            thirdParagraphRun.Text = "outside";
            thirdParagraphRun.Properties.CopyPropertiesFrom(firstParagraphRun1.Properties);
            thirdParagraph.Inlines.Add(thirdParagraphRun);

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString(@"<html>
<body>
<span style=""font-size: 24px; color: #008000; background-color: #FF0000;"">outside<b>inside text<p>paragraph</p></b>outside</span>
</body>
</html>");

            // Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void Span_WhenInnerTextPartsHaveDifferentFormatting_StylePropertiesAreInherittedCorrectly()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport(false);
            Section section = expected.Sections.AddSection();
            Paragraph paragraph = section.Blocks.AddParagraph();

            Run run1 = new Run(expected);
            run1.Text = "normal";
            paragraph.Inlines.Add(run1);

            Run run2 = new Run(expected);
            run2.Text = "bold";
            run2.Properties.FontWeight.LocalValue = FontWeights.Bold;
            paragraph.Inlines.Add(run2);

            Run run3 = new Run(expected);
            run3.Text = "color and bold";
            run3.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Blue);
            run3.Properties.FontWeight.LocalValue = FontWeights.Bold;
            paragraph.Inlines.Add(run3);

            Run run4 = new Run(expected);
            run4.Text = "normal";
            paragraph.Inlines.Add(run4);

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString(@"<html>
<body>
<span>normal<b>bold</b><font style=""background-color:#0000FF""><b>color and bold</b></font>normal</span>
</body>
</html>");

            // Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void SpansAroundTableElement_WhenTextInsideFirstSpanHasDifferentStyle_StylePropertiesAreInherittedOnlyInTheFirstSpan()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport();
            Section section = expected.Sections.AddSection();
            Paragraph paragraph = section.Blocks.AddParagraph();

            Run run1 = new Run(expected);
            run1.Text = "outside";
            run1.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Red);
            run1.Properties.ForegroundColor.LocalValue = new ThemableColor(Colors.Green);
            run1.Properties.FontSize.LocalValue = 24;
            paragraph.Inlines.Add(run1);

            Run run2 = new Run(expected);
            run2.Text = "inside text";
            run2.Properties.CopyPropertiesFrom(run1.Properties);
            run2.Properties.FontWeight.LocalValue = FontWeights.Bold;
            paragraph.Inlines.Add(run2);

            Run run3 = new Run(expected);
            run3.Text = "outside";
            run3.Properties.CopyPropertiesFrom(run1.Properties);
            paragraph.Inlines.Add(run3);

            Table table = section.Blocks.AddTable();
            Paragraph tableParagraph = table.Rows.AddTableRow().Cells.AddTableCell().Blocks.AddParagraph();
            tableParagraph.StyleId = BuiltInStyleNames.NormalWebStyleId;
            tableParagraph.Inlines.AddRun("test");

            // Act
            RadFlowDocument imported = this.ImportHtmlFromString(@"<html>
<body>
<span style=""font-size: 24px; color: #008000; background-color: #FF0000;"">outside<b>inside text</b>outside</span>
<table><tr><td><p><span>test</span></p></td></tr></table>
</body>
</html>");

            // Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void StyleProperty_WhenValueTypeIsInch_Import()
        {
            // Arrange & Act
            RadFlowDocument imported = this.ImportHtmlFromString(@"<html>
<head>
    <style>
        .titleheading {
            margin-left: 3in;
        }
    </style>
</head>
<body>
    <p class=""titleheading"">indent</p>
</body>
</html>");

            Paragraph paragraph = imported.EnumerateChildrenOfType<Paragraph>().FirstOrDefault();
            Assert.AreEqual(Unit.InchToDip(3), paragraph.Indentation.LeftIndent);
        }

        [TestMethod]
        public void StyleProperty_WhenValueTypeIsCm_Import()
        {
            // Arrange & Act
            RadFlowDocument imported = this.ImportHtmlFromString(@"<html>
<head>
    <style>
        .titleheading {
            margin-left: 3cm;
        }
    </style>
</head>
<body>
    <p class=""titleheading"">indent</p>
</body>
</html>");

            Paragraph paragraph = imported.EnumerateChildrenOfType<Paragraph>().FirstOrDefault();
            Assert.AreEqual(Unit.CmToDip(3), paragraph.Indentation.LeftIndent);
        }

        [TestMethod]
        public void StyleProperty_WhenValueTypeIsMm_Import()
        {
            // Arrange & Act
            RadFlowDocument imported = this.ImportHtmlFromString(@"<html>
<head>
    <style>
        .titleheading {
            margin-left: 20mm;
        }
    </style>
</head>
<body>
    <p class=""titleheading"">indent</p>
</body>
</html>");

            Paragraph paragraph = imported.EnumerateChildrenOfType<Paragraph>().FirstOrDefault();
            Assert.AreEqual(Unit.MmToDip(20), paragraph.Indentation.LeftIndent);
        }

        [TestMethod]
        public void StyleSelectorWithNameContainingDashes_IsImported()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport(false);

            Style style = CreateWebBasedParagraphStyle("telerik-reTable-1");
            expected.StyleRepository.Add(style);

            Table table = expected.Sections.AddSection().Blocks.AddTable();
            Border border = new Border(1, BorderStyle.Single, new ThemableColor(Colors.Red));
            table.Borders = new TableBorders(border, border, border, border);
            table.StyleId = style.Id;

            TableRow row = table.Rows.AddTableRow();
            TableCell cell = row.Cells.AddTableCell();
            Paragraph paragraph = cell.Blocks.AddParagraph();
            paragraph.Inlines.AddRun("test");

            // Act 
            RadFlowDocument imported = this.ImportHtml("TableWithClassNameContainingDashes.html");

            // Assert
            this.AssertDocuments(expected, imported);
        }

        //TODO: see story 156320 Extend the support for CSS properties applied to table and table cell borders
        //At the moment the test fails because the border-style and border-thickness are not supported
        //[TestMethod]
        public void TableWithClass_IndividualLongHandPropertiesAreImported()
        {
            // Arrange
            RadFlowDocument expected = this.GetExpectedDocumentForImport(false);

            Style style = CreateWebBasedParagraphStyle("telerikTable");
            expected.StyleRepository.Add(style);

            Table table = expected.Sections.AddSection().Blocks.AddTable();
            Border border = new Border(1, BorderStyle.Single, new ThemableColor(Colors.Red));
            table.Borders = new TableBorders(border, border, border, border);
            table.StyleId = style.Id;

            TableRow row = table.Rows.AddTableRow();
            TableCell cell = row.Cells.AddTableCell();
            Paragraph paragraph = cell.Blocks.AddParagraph();
            paragraph.Inlines.AddRun("test");

            // Act 
            RadFlowDocument imported = this.ImportHtml("TableWithClassWithIndividualProperties.html");

            // Assert
            this.AssertDocuments(expected, imported);
        }

        [TestMethod]
        public void HtmlImport_WhenPseudoClassStyleIsPresent_ShouldNotGenerateStyleInDocument()
        {
            // Arrange
            string importHtml = @"<HTML><style> .test:custom { color: red; } </style></HTML>";
            HtmlFormatProvider provider = new HtmlFormatProvider();

            // Act
            RadFlowDocument document = provider.Import(importHtml);

            // Assert
            Assert.IsFalse(document.StyleRepository.Contains("test"));
        }

        [TestMethod]
        public void HtmlImport_WhenPseudoElementStyleIsPresent_ShouldNotGenerateStyleInDocument()
        {
            // Arrange
            string importHtml = @"<HTML><style> .test::custom { color: red; } </style></HTML>";
            HtmlFormatProvider provider = new HtmlFormatProvider();

            // Act
            RadFlowDocument document = provider.Import(importHtml);

            // Assert
            Assert.IsFalse(document.StyleRepository.Contains("test"));
        }
    }
}