﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Tests.Utilities;

namespace Telerik.Windows.Documents.Flow.Tests.Integration.HtmlImportExport.Tables
{
    [TestClass]
    public class TableHeadBodyAndFootTests : HtmlTestBase
    {
        protected override string FolderName
        {
            get
            {
                return "Tables";
            }
        }

        private static readonly string startingHtml = @"
<html>
    <head><style></style></head>
<body>
<table>";
        private static readonly string endingHtml = @"
</table>
</body>
<html>";

        private static readonly string tableHead = @"
    <thead>
        <tr>
            <td><p>header</p></td>
        </tr>
    </thead>";

        private static readonly string tableBody = @"
    <tbody>
        <tr>
            <td><p>body</p></td>
        </tr>
    </tbody>";

        private static readonly string tableFoot = @"
    <tfoot>
        <tr>
            <td><p>footer</p></td>
        </tr>
    </tfoot>";

        [TestMethod]
        public void Table_WithHeadBodyAndFoot_Import_Html()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Table table = expected.Sections.AddSection().Blocks.AddTable();

            TableRow row = table.Rows.AddTableRow();
            TableCell cellHeader = row.Cells.AddTableCell();
            cellHeader.Blocks.AddParagraph().Inlines.AddRun("header");

            row = table.Rows.AddTableRow();
            TableCell cellBody = row.Cells.AddTableCell();
            cellBody.Blocks.AddParagraph().Inlines.AddRun("body");

            row = table.Rows.AddTableRow();
            TableCell cellFooter = row.Cells.AddTableCell();
            cellFooter.Blocks.AddParagraph().Inlines.AddRun("footer");

            // Act
            string html = startingHtml + tableHead + tableBody + tableFoot + endingHtml;
            RadFlowDocument imported = this.ImportHtmlFromString(html);

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(expected, imported);
        }

        [TestMethod]
        public void Table_WithHeadFootAndBody_Import_Html()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Table table = expected.Sections.AddSection().Blocks.AddTable();

            TableRow row = table.Rows.AddTableRow();
            TableCell cellHeader = row.Cells.AddTableCell();
            cellHeader.Blocks.AddParagraph().Inlines.AddRun("header");

            row = table.Rows.AddTableRow();
            TableCell cellBody = row.Cells.AddTableCell();
            cellBody.Blocks.AddParagraph().Inlines.AddRun("body");

            row = table.Rows.AddTableRow();
            TableCell cellFooter = row.Cells.AddTableCell();
            cellFooter.Blocks.AddParagraph().Inlines.AddRun("footer");

            // Act
            string html = startingHtml + tableHead + tableFoot + tableBody + endingHtml;
            RadFlowDocument imported = this.ImportHtmlFromString(html);

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(expected, imported);
        }

        [TestMethod]
        public void Table_WithFootHeadAndBody_Import_Html()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Table table = expected.Sections.AddSection().Blocks.AddTable();

            TableRow row = table.Rows.AddTableRow();
            TableCell cellHeader = row.Cells.AddTableCell();
            cellHeader.Blocks.AddParagraph().Inlines.AddRun("header");

            row = table.Rows.AddTableRow();
            TableCell cellBody = row.Cells.AddTableCell();
            cellBody.Blocks.AddParagraph().Inlines.AddRun("body");

            row = table.Rows.AddTableRow();
            TableCell cellFooter = row.Cells.AddTableCell();
            cellFooter.Blocks.AddParagraph().Inlines.AddRun("footer");

            // Act
            string html = startingHtml + tableFoot + tableHead + tableBody + endingHtml;
            RadFlowDocument imported = this.ImportHtmlFromString(html);

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(expected, imported);
        }

        [TestMethod]
        public void Table_WithFootBodyAndHead_Import_Html()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Table table = expected.Sections.AddSection().Blocks.AddTable();

            TableRow row = table.Rows.AddTableRow();
            TableCell cellHeader = row.Cells.AddTableCell();
            cellHeader.Blocks.AddParagraph().Inlines.AddRun("header");

            row = table.Rows.AddTableRow();
            TableCell cellBody = row.Cells.AddTableCell();
            cellBody.Blocks.AddParagraph().Inlines.AddRun("body");

            row = table.Rows.AddTableRow();
            TableCell cellFooter = row.Cells.AddTableCell();
            cellFooter.Blocks.AddParagraph().Inlines.AddRun("footer");

            // Act
            string html = startingHtml + tableFoot + tableBody + tableHead + endingHtml;
            RadFlowDocument imported = this.ImportHtmlFromString(html);

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(expected, imported);
        }

        [TestMethod]
        public void Table_WithBodyFootAndHead_Import_Html()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Table table = expected.Sections.AddSection().Blocks.AddTable();

            TableRow row = table.Rows.AddTableRow();
            TableCell cellHeader = row.Cells.AddTableCell();
            cellHeader.Blocks.AddParagraph().Inlines.AddRun("header");

            row = table.Rows.AddTableRow();
            TableCell cellBody = row.Cells.AddTableCell();
            cellBody.Blocks.AddParagraph().Inlines.AddRun("body");

            row = table.Rows.AddTableRow();
            TableCell cellFooter = row.Cells.AddTableCell();
            cellFooter.Blocks.AddParagraph().Inlines.AddRun("footer");

            // Act
            string html = startingHtml + tableBody + tableFoot + tableHead + endingHtml;
            RadFlowDocument imported = this.ImportHtmlFromString(html);

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(expected, imported);
        }

        [TestMethod]
        public void Table_WithBodyHeadAndFoot_Import_Html()
        {
            // Arrange
            RadFlowDocument expected = new RadFlowDocument();
            Table table = expected.Sections.AddSection().Blocks.AddTable();

            TableRow row = table.Rows.AddTableRow();
            TableCell cellHeader = row.Cells.AddTableCell();
            cellHeader.Blocks.AddParagraph().Inlines.AddRun("header");

            row = table.Rows.AddTableRow();
            TableCell cellBody = row.Cells.AddTableCell();
            cellBody.Blocks.AddParagraph().Inlines.AddRun("body");

            row = table.Rows.AddTableRow();
            TableCell cellFooter = row.Cells.AddTableCell();
            cellFooter.Blocks.AddParagraph().Inlines.AddRun("footer");

            // Act
            string html = startingHtml + tableBody + tableHead + tableFoot + endingHtml;
            RadFlowDocument imported = this.ImportHtmlFromString(html);

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(expected, imported);
        }
    }
}
