﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Linq;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
using Telerik.Windows.Documents.Flow.Tests.Utilities.XmlValidation;

namespace Telerik.Windows.Documents.Flow.Tests.Integration.ImportExport.Annotations
{
    [TestClass]
    public class BookmarksTests : PropertiesTestsBase
    {
        public const string TestFilesFolder = "Annotations/Bookmarks/";

        [TestMethod]
        public void Bookmark_BetweenInlines_ExportImport_Rtf()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("before ");
            editor.InsertBookmark("test");
            editor.InsertText(" after");

            //Act
            RadFlowDocument testDocument = this.ExportImportDocument(editor.Document, new RtfFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, testDocument);
        }

        [TestMethod]
        public void Bookmark_BetweenInlines_ExportImport_Docx()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("before ");
            editor.InsertBookmark("test");
            editor.InsertText(" after");

            //Act
            RadFlowDocument testDocument = this.ExportImportDocument(editor.Document, new DocxFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, testDocument);
        }

        [TestMethod]
        public void Bookmark_AroundInline_ExportImport_Rtf()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("before ");
            Run bookaredRun = editor.InsertText(" inside ");
            editor.InsertBookmark("test", bookaredRun, bookaredRun);
            editor.InsertText(" after");

            //Act
            RadFlowDocument testDocument = this.ExportImportDocument(editor.Document, new RtfFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, testDocument);
        }

        [TestMethod]
        public void Bookmark_AroundInline_ExportImport_Docx()
        {
            // Arrange 
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("before");
            Run run = editor.InsertText("inside");
            editor.InsertBookmark("test", run, run);
            editor.InsertText("after");

            // Act
            RadFlowDocument testDocument = this.ExportImportDocument(editor.Document, new DocxFormatProvider());

            // Assert
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, testDocument);
        }

        [TestMethod]
        public void Bookmark_AroundInline_Import_Rtf()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("before ");
            Run bookaredRun = editor.InsertText("inside");
            editor.InsertBookmark("test", bookaredRun, bookaredRun);
            editor.InsertText(" after");

            //Act
            RadFlowDocument testDocument = TestHelper.GetDocumentFromFile(string.Format("{0}BookmarkAroundInline.rtf", TestFilesFolder), new RtfFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, testDocument);
        }

        [TestMethod]
        public void Bookmark_AroundInline_Import_Docx()
        {
            // Arrange 
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("before");
            editor.InsertText(" ");
            Run run = editor.InsertText("inside");
            editor.InsertBookmark("test", run, run);
            editor.InsertText(" after");

            // Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}BookmarkAroundInline.docx", TestFilesFolder), new DocxFormatProvider());

            // Assert
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, importedDocument);
        }

        [TestMethod]
        public void Bookmark_AroundInline_Validate_Xml()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("before");
            Run run = editor.InsertText("inside");
            editor.InsertBookmark("test", run, run);
            editor.InsertText("after");

            XmlValidator.ValidateOpenXmlDocument(editor.Document);
        }

        [TestMethod]
        public void Bookmark_Import_InavlidBookmarkStart_Rtf()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("before");
            editor.InsertText("after");

            string rtf = @"{\rtf\ansi\ansicpg1252\uc1\deff0\deflang1033
{\fonttbl{\f0 Calibri;}}
{\colortbl\red0\green0\blue0;}
\pard before{\*\bkmkstart test}after\par}";


            //Act
            RadFlowDocument testDocument = new RtfFormatProvider().Import(rtf);

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, testDocument);
        }

        [TestMethod]
        public void Bookmark_Import_InavlidBookmarkEnd_Rtf()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("before");
            editor.InsertText("after");

            string rtf = @"{\rtf\ansi\ansicpg1252\uc1\deff0\deflang1033
{\fonttbl{\f0 Calibri;}}
{\colortbl\red0\green0\blue0;}
\pard before{\*\bkmkend test}after\par}";


            //Act
            RadFlowDocument testDocument = new RtfFormatProvider().Import(rtf);

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, testDocument);
        }

        [TestMethod]
        public void Bookmark_InsertedOnTwoParagraphs_Import_Rtf()
        {
            // Arrange 
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("first ");
            Run run1 = editor.InsertText("second");
            editor.InsertParagraph();
            Run run2 = editor.InsertText("third");
            editor.InsertBookmark("test", run1, run2);
            editor.InsertText(" fourth");

            // Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}BookmarkOnTwoParagraphs.rtf", TestFilesFolder), new RtfFormatProvider());

            // Assert
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, importedDocument);
        }

        [TestMethod]
        public void Bookmark_InsertedOnTwoParagraphs_Import_Docx()
        {
            // Arrange 
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("first");
            editor.InsertText(" ");
            Run run1 = editor.InsertText("second");
            editor.InsertParagraph();
            Run run2 = editor.InsertText("third");
            editor.InsertBookmark("test", run1, run2);
            editor.InsertText(" fourth");

            // Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}BookmarkOnTwoParagraphs.docx", TestFilesFolder), new DocxFormatProvider());

            // Assert
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, importedDocument);
        }

        [TestMethod]
        public void Bookmark_InsertedOnTwoParagraphs_ExportImport_Rtf()
        {
            // Arrange 
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("first ");
            Run run1 = editor.InsertText("second");
            editor.InsertParagraph();
            Run run2 = editor.InsertText("third");
            editor.InsertBookmark("test", run1, run2);
            editor.InsertText(" fourth");

            // Act
            RadFlowDocument testDocument = this.ExportImportDocument(editor.Document, new RtfFormatProvider());

            // Assert
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, testDocument);
        }

        [TestMethod]
        public void Bookmark_InsertedOnTwoParagraphs_ExportImport_Docx()
        {
            // Arrange 
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("first");
            editor.InsertText(" ");
            Run run1 = editor.InsertText("second");
            editor.InsertParagraph();
            Run run2 = editor.InsertText("third");
            editor.InsertBookmark("test", run1, run2);
            editor.InsertText(" fourth");

            // Act
            RadFlowDocument testDocument = this.ExportImportDocument(editor.Document, new DocxFormatProvider());

            // Assert
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, testDocument);
        }

        [TestMethod]
        public void Bookmark_InsertedOnTwoParagraphs_Validate_Xml()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("first");
            editor.InsertText(" ");
            Run run1 = editor.InsertText("second");
            editor.InsertParagraph();
            Run run2 = editor.InsertText("third");
            editor.InsertBookmark("test", run1, run2);
            editor.InsertText(" fourth");

            XmlValidator.ValidateOpenXmlDocument(editor.Document);
        }

        [TestMethod]
        public void Bookmark_InsertedInTable_OnTwoRowsAndTwoCols_Import_Docx()
        {
            // Arrange 
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}BookmarksInTableForTwoColumnsAndTwoRows.docx", TestFilesFolder), new DocxFormatProvider());

            // Act
            Bookmark bookmark = importedDocument.EnumerateChildrenOfType<BookmarkRangeStart>().FirstOrDefault().Bookmark;

            // Assert
            Assert.AreEqual("test", bookmark.Name);
            Assert.AreEqual(0, bookmark.FromColumn);
            Assert.AreEqual(1, bookmark.ToColumn);
        }

        [TestMethod]
        public void MultipleBookmarks_InComplexDocument_Import_Rtf()
        {
            // Arrange
            RadFlowDocument expectedDocument = this.CreateBookmarksImportExportComplexTestDocument();

            // Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}BookmarksComplex.rtf", TestFilesFolder), new RtfFormatProvider());

            // Assert
            StructureAsserter.AssertDocumentsAreEqual(expectedDocument, importedDocument);
        }

        [TestMethod]
        public void MultipleBookmarks_InComplexDocument_Import_Docx()
        {
            // Arrange
            RadFlowDocument expectedDocument = this.CreateBookmarksImportExportComplexTestDocument();

            // Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}BookmarksComplex.docx", TestFilesFolder), new DocxFormatProvider());

            // Assert
            StructureAsserter.AssertDocumentsAreEqual(expectedDocument, importedDocument);
        }

        [TestMethod]
        public void MultipleBookmarks_InComplexDocument_ExportImport_Rtf()
        {
            // Arrange
            RadFlowDocument expectedDocument = this.CreateBookmarksImportExportComplexTestDocument();

            // Act
            RadFlowDocument testDocument = this.ExportImportDocument(expectedDocument, new RtfFormatProvider());

            // Assert
            StructureAsserter.AssertDocumentsAreEqual(expectedDocument, testDocument);
        }

        [TestMethod]
        public void MultipleBookmarks_InComplexDocument_ExportImport_Docx()
        {
            // Arrange
            RadFlowDocument expectedDocument = this.CreateBookmarksImportExportComplexTestDocument();

            // Act
            RadFlowDocument testDocument = this.ExportImportDocument(expectedDocument, new DocxFormatProvider());

            // Assert
            StructureAsserter.AssertDocumentsAreEqual(expectedDocument, testDocument);
        }

        [TestMethod]
        public void MultipleBookmarks_WhenAreBetweenTableCellAndTableRows_Import_Docx()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            Table table = editor.InsertTable(2, 3);
            foreach (TableCell cell in table.EnumerateChildrenOfType<TableCell>())
            {
                cell.Blocks.AddParagraph();
            }

            TableCell cell1Row0 = table.Rows[0].Cells[1];
            Paragraph paragraphInCell1Row0 = cell1Row0.Blocks.First() as Paragraph;

            editor.MoveToParagraphStart(paragraphInCell1Row0);
            editor.InsertBookmark("1");

            Bookmark bookmark2 = new Bookmark(document, "2");
            editor.InsertInline(bookmark2.BookmarkRangeStart);

            TableCell cell2Row0 = table.Rows[0].Cells[2];
            Paragraph paragraphInCell2Row0 = cell2Row0.Blocks.First() as Paragraph;

            editor.MoveToParagraphStart(paragraphInCell2Row0);
            editor.InsertInline(bookmark2.BookmarkRangeEnd);

            TableCell cell0Row1 = table.Rows[1].Cells[0];
            Paragraph paragraphInCell0Row1 = cell0Row1.Blocks.First() as Paragraph;
            editor.MoveToParagraphStart(paragraphInCell0Row1);

            Bookmark bookmark3 = new Bookmark(document, "3");
            editor.InsertInline(bookmark3.BookmarkRangeStart);
            editor.InsertInline(bookmark3.BookmarkRangeEnd);

            Bookmark bookmark4 = new Bookmark(document, "4");
            editor.InsertInline(bookmark4.BookmarkRangeStart);

            editor.MoveToTableEnd(table);
            editor.InsertParagraph();
            editor.InsertInline(bookmark4.BookmarkRangeEnd);
            editor.InsertParagraph();

            // Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}BookmarksOutsideOfParagraph.docx", TestFilesFolder), new DocxFormatProvider());

            // Assert
            StructureAsserter.AssertDocumentsAreEqual(document, importedDocument);
        }

        [TestMethod]
        public void MultipleBookmarks_WhenTheirStartOrEndAreBetweenParagraphs_Import_Docx()
        {
            // Arrange 
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("This is sample text");
            editor.InsertParagraph();
            Run run1 = editor.InsertText("Another");
            editor.InsertBookmark("bmk", run1, run1);
            editor.InsertText(" sample ");
            Bookmark bmk2 = new Bookmark(editor.Document, "bmk2");
            editor.InsertInline(bmk2.BookmarkRangeStart);
            editor.InsertText("text");
            editor.InsertParagraph();
            editor.InsertInline(bmk2.BookmarkRangeEnd);
            editor.InsertText("Last line");

            // Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}BookmarkBetweenParagraphs.docx", TestFilesFolder), new DocxFormatProvider());

            // Assert
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, importedDocument);
        }

        [TestMethod]
        public void OneSection_Validate_Xml()
        {
            RadFlowDocument document = this.CreateBookmarksImportExportComplexTestDocument();
            XmlValidator.ValidateOpenXmlDocument(document);
        }

        private RadFlowDocument CreateBookmarksImportExportComplexTestDocument()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("This is sentence one. This is sentence ");
            Run run1 = editor.InsertText("two.");
            editor.InsertParagraph();
            Run run2 = editor.InsertText("This is");
            editor.InsertBookmark("bookmarkOn2Paragraphs", run1, run2);
            editor.InsertText(" sentence three on ");
            Run run3 = editor.InsertText("separate p");
            editor.InsertBookmark("bookmarkOn1Paragraph", run3, run3);
            editor.InsertText("aragraph.");
            editor.InsertParagraph();
            editor.InsertBookmark("emptyBookmark");

            return editor.Document;
        }
    }
}