﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Common.FormatProviders;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Tests.Utilities;

namespace Telerik.Windows.Documents.Flow.Tests.Integration.ImportExport.Fields
{
    [TestClass]
    public class MergeFieldTests : TestsBase
    {
        public const string TestFilesFolder = "Fields/MergeFields/";

        public static string StartingBracket = "«";
        public static string EndingBracket = "»";

        private void MergeField_WithoutSwitches_ExportImport(IFormatProvider<RadFlowDocument> provider)
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.InsertField("MERGEFIELD FirstName", "");
            editor.InsertText(" ");

            // Act
            RadFlowDocument imported = this.ExportImportDocument(document, provider);
            List<Person> persons = Person.GetList(1);
            RadFlowDocument mergedDocument = imported.MailMerge(persons);

            // Assert
            List<Run> runs = mergedDocument.EnumerateChildrenOfType<Run>().ToList();
            string text = runs[0].Text;
            Assert.AreEqual(text, persons[0].FirstName);
        }

        [TestMethod]
        public void MergeField_WithoutSwitches_ExportImportDocx_IsFieldMergedCorrectly()
        {
            this.MergeField_WithoutSwitches_ExportImport(new DocxFormatProvider());
        }

        [TestMethod]
        public void MergeField_WithoutSwitches_ExportImportRtf_IsFieldMergedCorrectly()
        {
            this.MergeField_WithoutSwitches_ExportImport(new RtfFormatProvider());
        }

        private void MergeField_WithPrefix_ExportImport(IFormatProvider<RadFlowDocument> provider)
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            string prefix = "Mr.";
            editor.InsertField("MERGEFIELD FirstName \\b \"" + prefix + "\"", "");
            editor.InsertText(" ");
            editor.InsertField("MERGEFIELD \\b \"" + prefix + "\" FirstName", "");

            // Act
            RadFlowDocument imported = this.ExportImportDocument(document, provider);
            List<Person> persons = Person.GetList(1);
            RadFlowDocument mergedDocument = imported.MailMerge(persons);

            // Assert
            List<Run> runs = mergedDocument.EnumerateChildrenOfType<Run>().ToList();
            string text = runs[0].Text;
            Assert.AreEqual(text, prefix + persons[0].FirstName);

            text = runs[2].Text;
            Assert.AreEqual(text, prefix + persons[0].FirstName);
        }

        [TestMethod]
        public void MergeField_WithPrefix_ExportImportToDocx_IsFieldMergedCorrectly()
        {
            this.MergeField_WithPrefix_ExportImport(new DocxFormatProvider());
        }

        [TestMethod]
        public void MergeField_WithPrefix_ExportImportToRtf_IsFieldMergedCorrectly()
        {
            this.MergeField_WithPrefix_ExportImport(new RtfFormatProvider());
        }

        private void MergeField_WithSuffix_ExportImport(IFormatProvider<RadFlowDocument> provider)
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            string suffix = ".";
            editor.InsertField("MERGEFIELD FirstName \\f \"" + suffix + "\"", "");
            editor.InsertText(" ");
            editor.InsertField("MERGEFIELD \\f \"" + suffix + "\" FirstName", "");

            // Act
            RadFlowDocument imported = this.ExportImportDocument(document, provider);
            List<Person> persons = Person.GetList(1);
            RadFlowDocument mergedDocument = imported.MailMerge(persons);

            // Assert
            List<Run> runs = mergedDocument.EnumerateChildrenOfType<Run>().ToList();
            string text = runs[0].Text;
            Assert.AreEqual(text, persons[0].FirstName + suffix);

            text = runs[2].Text;
            Assert.AreEqual(text, persons[0].FirstName + suffix);
        }

        [TestMethod]
        public void MergeField_WithSuffix_ExportImportDocx_IsFieldMergedCorrectly()
        {
            this.MergeField_WithSuffix_ExportImport(new DocxFormatProvider());
        }

        [TestMethod]
        public void MergeField_WithSuffix_ExportImportRtf_IsFieldMergedCorrectly()
        {
            this.MergeField_WithSuffix_ExportImport(new RtfFormatProvider());
        }

        private void MergeField_WithPrefixAndSuffix_ExportImport(IFormatProvider<RadFlowDocument> provider)
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            string prefix = "Mr.";
            string suffix = ".";
            editor.InsertField("MERGEFIELD FirstName \\b \"" + prefix + "\" \\f \"" + suffix + "\"", "");
            editor.InsertText(" ");
            editor.InsertField("MERGEFIELD \\b \"" + prefix + "\" \\f \"" + suffix + "\" FirstName", "");

            // Act
            RadFlowDocument imported = this.ExportImportDocument(document, provider);
            List<Person> persons = Person.GetList(1);
            RadFlowDocument mergedDocument = imported.MailMerge(persons);

            // Assert
            List<Run> runs = mergedDocument.EnumerateChildrenOfType<Run>().ToList();
            string text = runs[0].Text;
            Assert.AreEqual(text, prefix + persons[0].FirstName + suffix);

            text = runs[2].Text;
            Assert.AreEqual(text, prefix + persons[0].FirstName + suffix);
        }

        [TestMethod]
        public void MergeField_WithPrefixAndSuffix_ExportImportDocx_IsFieldMergedCorrectly()
        {
            this.MergeField_WithPrefixAndSuffix_ExportImport(new DocxFormatProvider());
        }

        [TestMethod]
        public void MergeField_WithPrefixAndSuffix_ExportImportRtf_IsFieldMergedCorrectly()
        {
            this.MergeField_WithPrefixAndSuffix_ExportImport(new RtfFormatProvider());
        }

        private void MergeField_WithSuffixAndPrefix_ExportImport(IFormatProvider<RadFlowDocument> provider)
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            string prefix = "Mr.";
            string suffix = ".";
            editor.InsertField("MERGEFIELD FirstName \\f \"" + suffix + "\" \\b \"" + prefix + "\"", "");
            editor.InsertText(" ");
            editor.InsertField("MERGEFIELD \\f \"" + suffix + "\" \\b \"" + prefix + "\" FirstName", "");

            // Act
            RadFlowDocument imported = this.ExportImportDocument(document, provider);
            List<Person> persons = Person.GetList(1);
            RadFlowDocument mergedDocument = imported.MailMerge(persons);

            // Assert
            List<Run> runs = mergedDocument.EnumerateChildrenOfType<Run>().ToList();
            string text = runs[0].Text;
            Assert.AreEqual(text, prefix + persons[0].FirstName + suffix);

            text = runs[2].Text;
            Assert.AreEqual(text, prefix + persons[0].FirstName + suffix);
        }

        [TestMethod]
        public void MergeField_WithSuffixAndPrefix_ExportImportDocx_IsFieldMergedCorrectly()
        {
            this.MergeField_WithSuffixAndPrefix_ExportImport(new DocxFormatProvider());
        }

        [TestMethod]
        public void MergeField_WithSuffixAndPrefix_ExportImportRtf_IsFieldMergedCorrectly()
        {
            this.MergeField_WithSuffixAndPrefix_ExportImport(new RtfFormatProvider());
        }

        // TODO: tests for vertical switch.

        [TestMethod]
        public void ImportAndMergeFieldFromDocx_WithoutSwitches_IsFieldMergedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.InsertField(" MERGEFIELD FirstName ", StartingBracket + "FirstName" + EndingBracket);

            // Act and Assert
            RadFlowDocument imported = TestHelper.GetDocumentFromFile(string.Format("{0}MergeFieldNoSwitches.docx", TestFilesFolder), new DocxFormatProvider());
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, imported);

            // Act
            List<Person> persons = Person.GetList(1);
            RadFlowDocument mergedDocument = imported.MailMerge(persons);

            // Assert
            List<Run> runs = mergedDocument.EnumerateChildrenOfType<Run>().ToList();
            string text = runs[0].Text;
            Assert.AreEqual(text, persons[0].FirstName);
        }

        [TestMethod]
        public void ImportAndMergeFieldFromRtf_WithoutSwitches_IsFieldMergedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.InsertField(" MERGEFIELD FirstName ", StartingBracket + "FirstName" + EndingBracket);

            // Act and Assert
            RadFlowDocument imported = TestHelper.GetDocumentFromFile(string.Format("{0}MergeFieldNoSwitches.rtf", TestFilesFolder), new RtfFormatProvider());
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, imported);

            // Act
            List<Person> persons = Person.GetList(1);
            RadFlowDocument mergedDocument = imported.MailMerge(persons);

            // Assert
            List<Run> runs = mergedDocument.EnumerateChildrenOfType<Run>().ToList();
            string text = runs[0].Text;
            Assert.AreEqual(text, persons[0].FirstName);
        }

        [TestMethod]
        public void ImportAndMergeFieldFromDocx_WithSwitches_IsFieldMergedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.InsertField(" MERGEFIELD  FirstName \\b Mr. \\f . ", "Mr.«FirstName».");
            editor.InsertParagraph();
            editor.InsertField(" MERGEFIELD \\b Mr. \\f . FirstName ", "Mr.«FirstName».");
            editor.InsertParagraph();
            editor.InsertField(" MERGEFIELD \\b Mr.  FirstName \\f . ", "Mr.«FirstName».");
            editor.InsertParagraph();
            editor.InsertField(" MERGEFIELD \\f . FirstName \\b Mr. ", "Mr.«FirstName».");

            // Act and Assert
            RadFlowDocument imported = TestHelper.GetDocumentFromFile(string.Format("{0}MergeFieldWithSwitches.docx", TestFilesFolder), new DocxFormatProvider());
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, imported);

            // Act
            List<Person> persons = Person.GetList(1);
            RadFlowDocument mergedDocument = imported.MailMerge(persons);

            // Assert
            List<Run> runs = mergedDocument.EnumerateChildrenOfType<Run>().ToList();

            foreach (Run run in runs)
            {
                Assert.AreEqual(run.Text, "Mr." + persons[0].FirstName + ".");
            }
        }

        [TestMethod]
        public void ImportAndMergeFieldFromRtf_WithSwitches_IsFieldMergedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.InsertField(" MERGEFIELD  FirstName \\b Mr. \\f . ", "Mr.«FirstName».");
            editor.InsertParagraph();
            editor.InsertField(" MERGEFIELD \\b Mr. \\f . FirstName ", "Mr.«FirstName».");
            editor.InsertParagraph();
            editor.InsertField(" MERGEFIELD \\b Mr.  FirstName \\f . ", "Mr.«FirstName».");
            editor.InsertParagraph();
            editor.InsertField(" MERGEFIELD \\f . FirstName \\b Mr. ", "Mr.«FirstName».");

            // Act and Assert
            RadFlowDocument imported = TestHelper.GetDocumentFromFile(string.Format("{0}MergeFieldWithSwitches.rtf", TestFilesFolder), new RtfFormatProvider());
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, imported);

            // Act
            List<Person> persons = Person.GetList(1);
            RadFlowDocument mergedDocument = imported.MailMerge(persons);

            // Assert
            List<Run> runs = mergedDocument.EnumerateChildrenOfType<Run>().ToList();

            foreach (Run run in runs)
            {
                Assert.AreEqual(run.Text, "Mr." + persons[0].FirstName + ".");
            }
        }
    }

    public class Person
    {
        public string FirstName { get; set; }
        public string FamilyName { get; set; }
        public string Company { get; set; }

        public static List<Person> GetList(int personCound)
        {
            List<Person> list = new List<Person>(personCound);

            for (int i = 0; i < personCound; i++)
            {
                Person person = new Person() { FirstName = "Nancy" + i, FamilyName = "Davolio" + i, Company = "Company" + i };
                list.Add(person);
            }

            return list;
        }
    }
}
