﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Linq;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
using Telerik.Windows.Documents.Flow.Tests.Utilities.XmlValidation;

namespace Telerik.Windows.Documents.Flow.Tests.Integration.ImportExport.Lists
{
    [TestClass]
    public class ListWithHeadingsTests : ListsTestBase
    {
        [TestMethod]
        public void List_WhenHeadingsAreAppliedOverListLevels_ExportImport_Docx()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            List list = this.GetList(document);
            for (int i = 0; i < 9; i++)
            {
                document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1));
            }

            Section section = document.Sections.AddSection();

            Paragraph p = section.Blocks.AddParagraph();
            p.ListId = list.Id;
            p.StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(1);

            RadFlowDocument importedDocument = this.StyleExportImportTemplate(document, new DocxFormatProvider());
            List importedList = importedDocument.Lists.GetList(0);
            Paragraph importedParagraph = importedDocument.EnumerateChildrenOfType<Paragraph>().FirstOrDefault();

            this.AssertListProperties(list, importedList);
            this.AssertHeadingStyles(importedDocument, list);

            Assert.AreEqual(importedParagraph.StyleId, p.StyleId);
            Assert.AreEqual(importedParagraph.ListId, importedList.Id);
        }

        [TestMethod]
        public void List_WhenHeadingsAreAppliedOverListLevels_ExportImport_Rtf()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            List list = this.GetList(document);
            for (int i = 0; i < 9; i++)
            {
                document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1));
            }

            Section section = document.Sections.AddSection();

            Paragraph p = section.Blocks.AddParagraph();
            p.ListId = list.Id;
            p.StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(1);

            RadFlowDocument importedDocument = this.StyleExportImportTemplate(document, new RtfFormatProvider());
            List importedList = importedDocument.Lists.GetList(0);
            Paragraph importedParagraph = importedDocument.EnumerateChildrenOfType<Paragraph>().FirstOrDefault();

            this.AssertListProperties(list, importedList);
            this.AssertHeadingStyles(importedDocument, list);

            Assert.AreEqual(importedParagraph.StyleId, p.StyleId);
            Assert.AreEqual(importedParagraph.ListId, importedList.Id);
        }

        [TestMethod]
        public void List_WhenHeadingsAreAppliedOverListLevels_Validate_Xml()
        {
            RadFlowDocument document = new RadFlowDocument();
            List list = this.GetList(document);
            for (int i = 0; i < 9; i++)
            {
                document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1));
            }

            Section section = document.Sections.AddSection();

            Paragraph p = section.Blocks.AddParagraph();
            p.ListId = list.Id;
            p.StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(1);

            XmlValidator.ValidateOpenXmlDocument(document);
        }

        [TestMethod]
        public void TwoLists_WhenHeadingsAreAppliedOverListLevels_ExportImport_Docx()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            List list = this.GetList(document);
            List listTwo = this.GetList(document);

            for (int i = 0; i < 9; i++)
            {
                document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1));
            }

            Section section = document.Sections.AddSection();

            Paragraph p = section.Blocks.AddParagraph();
            p.ListId = list.Id;
            p.StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(1);

            p = section.Blocks.AddParagraph();
            p.ListId = listTwo.Id;
            p.StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(1);

            RadFlowDocument importedDocument = this.StyleExportImportTemplate(document, new DocxFormatProvider());
            List importedList = importedDocument.Lists.GetList(0);
            List importedListTwo = importedDocument.Lists.GetList(1);
            Paragraph importedParagraph = importedDocument.EnumerateChildrenOfType<Paragraph>().FirstOrDefault();
            Paragraph importedParagraphTwo = importedDocument.EnumerateChildrenOfType<Paragraph>().LastOrDefault();

            this.AssertListProperties(list, importedList);
            this.AssertListProperties(listTwo, importedListTwo);
            this.AssertHeadingStyles(importedDocument, list);

            Assert.AreEqual(importedParagraph.StyleId, p.StyleId);
            Assert.AreEqual(importedParagraph.ListId, importedList.Id);

            Assert.AreEqual(importedParagraphTwo.StyleId, p.StyleId);
            Assert.AreEqual(importedParagraphTwo.ListId, importedListTwo.Id);
        }

        [TestMethod]
        public void TwoLists_WhenHeadingsAreAppliedOverListLevels_ExportImport_Rtf()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            List list = this.GetList(document);
            List listTwo = this.GetList(document);

            for (int i = 0; i < 9; i++)
            {
                document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1));
            }

            Section section = document.Sections.AddSection();

            Paragraph p = section.Blocks.AddParagraph();
            p.ListId = list.Id;
            p.StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(1);

            p = section.Blocks.AddParagraph();
            p.ListId = listTwo.Id;
            p.StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(1);

            RadFlowDocument importedDocument = this.StyleExportImportTemplate(document, new RtfFormatProvider());
            List importedList = importedDocument.Lists.GetList(0);
            List importedListTwo = importedDocument.Lists.GetList(1);
            Paragraph importedParagraph = importedDocument.EnumerateChildrenOfType<Paragraph>().FirstOrDefault();
            Paragraph importedParagraphTwo = importedDocument.EnumerateChildrenOfType<Paragraph>().LastOrDefault();

            this.AssertListProperties(list, importedList);
            this.AssertListProperties(listTwo, importedListTwo);
            this.AssertHeadingStyles(importedDocument, list);

            Assert.AreEqual(importedParagraph.StyleId, p.StyleId);
            Assert.AreEqual(importedParagraph.ListId, importedList.Id);

            Assert.AreEqual(importedParagraphTwo.StyleId, p.StyleId);
            Assert.AreEqual(importedParagraphTwo.ListId, importedListTwo.Id);
        }

        [TestMethod]
        public void TwoLists_WhenHeadingsAreAppliedOverListLevels_Validate_Xml()
        {
            RadFlowDocument document = new RadFlowDocument();
            List list = this.GetList(document);
            List listTwo = this.GetList(document);

            for (int i = 0; i < 9; i++)
            {
                document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1));
            }

            Section section = document.Sections.AddSection();

            Paragraph p = section.Blocks.AddParagraph();
            p.ListId = list.Id;
            p.StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(1);

            p = section.Blocks.AddParagraph();
            p.ListId = listTwo.Id;
            p.StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(1);

            XmlValidator.ValidateOpenXmlDocument(document);
        }

        [TestMethod]
        public void List_WhenHeadingsAreAppliedOverListLevels_Import_Docx()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            List list = this.GetList(document);
            list.MultilevelType = MultilevelType.Multilevel;


            // Act and Assert
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}ListWithHeadings.docx", TestFilesFolder), new DocxFormatProvider());
            List importedList = importedDocument.Lists.GetList(0);

            this.AssertListProperties(list, importedList);
            this.AssertHeadingStyles(importedDocument, list);
        }

        [TestMethod]
        public void List_WhenHeadingsAreAppliedOverListLevels_Import_Rtf()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            List list = this.GetList(document);
            list.MultilevelType = MultilevelType.Multilevel;

            // Act and Assert
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}ListWithHeadings.rtf", TestFilesFolder), new RtfFormatProvider());
            List importedList = importedDocument.Lists.GetList(0);

            this.ClearRedundantForRtfPropertyValues(importedList);
            this.ClearRedundantForRtfPropertyValues(list);

            this.AssertListProperties(list, importedList);
            this.AssertHeadingStyles(importedDocument, list);
        }

        [TestMethod]
        public void TwoLists_WhenHeadingsAreAppliedOverListLevels_Import_Docx()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            List list = this.GetList(document);
            list.MultilevelType = MultilevelType.Multilevel;
            List listTwo = this.GetList(document);
            listTwo.MultilevelType = MultilevelType.Multilevel;

            // Act and Assert
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}TwoListsWithHeadings.docx", TestFilesFolder), new DocxFormatProvider());
            List importedList = importedDocument.Lists.GetList(0);
            List importedListTwo = importedDocument.Lists.GetList(1);

            this.AssertListProperties(list, importedList);
            this.AssertListProperties(listTwo, importedListTwo);

            this.AssertHeadingStyles(importedDocument, list);

            System.Collections.Generic.List<Paragraph> paragraphs = importedDocument.EnumerateChildrenOfType<Paragraph>().ToList();
            for (int i = 0; i < 9; i++)
            {
                Assert.AreEqual(paragraphs[i].Properties.ListId.LocalValue, null);
            }

            for (int i = 10; i < 19; i++)
            {
                Assert.AreEqual(paragraphs[i].Properties.ListId.LocalValue, listTwo.Id);
                Assert.AreEqual(paragraphs[i].Properties.ListLevel.LocalValue, i - 10);
            }
        }

        [TestMethod]
        public void TwoLists_WhenHeadingsAreAppliedOverListLevels_Import_Rtf()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            List list = this.GetList(document);
            list.MultilevelType = MultilevelType.Multilevel;
            List listTwo = this.GetList(document);
            listTwo.MultilevelType = MultilevelType.Multilevel;

            // Act and Assert
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}TwoListsWithHeadings.rtf", TestFilesFolder), new RtfFormatProvider());
            List importedList = importedDocument.Lists.GetList(0);
            List importedListTwo = importedDocument.Lists.GetList(1);

            this.ClearRedundantForRtfPropertyValues(list);
            this.ClearRedundantForRtfPropertyValues(listTwo);
            this.ClearRedundantForRtfPropertyValues(importedList);
            this.ClearRedundantForRtfPropertyValues(importedListTwo);

            this.AssertListProperties(list, importedList);
            this.AssertListProperties(listTwo, importedListTwo);

            this.AssertHeadingStyles(importedDocument, list);
        }

        [TestMethod]
        public void List_WithStylesOnListLevels_AndOverridenSpacingAndIndentation_Rtf()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Style style = new Style("MyNormalStyle", StyleType.Paragraph);
            style.ParagraphProperties.LeftIndent.LocalValue = 110.4;
            style.ParagraphProperties.RightIndent.LocalValue = 86.4;
            style.ParagraphProperties.HangingIndent.LocalValue = 86.4;
            style.ParagraphProperties.LineSpacingType.LocalValue = HeightType.AtLeast;
            style.ParagraphProperties.LineSpacing.LocalValue = 40;
            document.StyleRepository.Add(style);

            List list = document.Lists.Add(ListTemplateType.NumberedDefault);
            foreach (ListLevel listLevel in list.Levels)
            {
                listLevel.StyleId = style.Id;
                listLevel.ParagraphProperties.LineSpacing.LocalValue = 60;
                listLevel.ParagraphProperties.SpacingBefore.LocalValue = 50;
                listLevel.ParagraphProperties.SpacingAfter.LocalValue = 40;
                listLevel.ParagraphProperties.LeftIndent.LocalValue = 200;
                listLevel.ParagraphProperties.RightIndent.LocalValue = 210;
                listLevel.ParagraphProperties.HangingIndent.LocalValue = 220;
            }
            // Act 
            RadFlowDocument importedDocument = this.StyleExportImportTemplate(document, new RtfFormatProvider());
            List importedList = importedDocument.Lists.GetList(0);

            // Assert
            this.AssertListProperties(list, importedList);
        }

        private List GetList(RadFlowDocument document)
        {
            List list = document.Lists.Add(ListTemplateType.NumberedDefault);
            for (int i = 0; i < list.Levels.Count; i++)
            {
                string textFormat = string.Empty;
                if (i > 0)
                {
                    textFormat = list.Levels[i - 1].NumberTextFormat + ".";
                }

                list.Levels[i].NumberTextFormat = textFormat + "%" + (i + 1);
                list.Levels[i].NumberingStyle = NumberingStyle.Decimal;
                list.Levels[i].Alignment = Alignment.Left;
                list.Levels[i].StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1);
                list.Levels[i].ParagraphProperties.LeftIndent.LocalValue = Math.Round((19.2 + ((i + 1) * 9.6)), 1, MidpointRounding.AwayFromZero);
                list.Levels[i].ParagraphProperties.HangingIndent.LocalValue = Math.Round(19.2 + ((i + 1) * 9.6), 1, MidpointRounding.AwayFromZero);
            }

            return list;
        }

        private void ClearRedundantForRtfPropertyValues(List list)
        {
            // We clear the following property values, because they have different values in RTF compared to our model 
            // and AssertListProperties() method will fail with them.
            foreach (ListLevel listLevel in list.Levels)
            {
                listLevel.ParagraphProperties.SpacingBefore.ClearValue();
                listLevel.ParagraphProperties.LineSpacing.ClearValue();
                listLevel.ParagraphProperties.LeftIndent.ClearValue();
                listLevel.ParagraphProperties.HangingIndent.ClearValue();
            }
        }

        private void AssertHeadingStyles(RadFlowDocument importedDocument, List list)
        {
            for (int i = 0; i < 9; i++)
            {
                Style style = importedDocument.StyleRepository.GetStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1));
                Assert.IsNotNull(style);
                Assert.AreEqual(style.ParagraphProperties.ListId.LocalValue, list.Id);
                if (i == 0)
                {
                    Assert.IsTrue(style.ParagraphProperties.ListLevel.LocalValue == null || style.ParagraphProperties.ListLevel.LocalValue == i);
                }
                else
                {
                    Assert.AreEqual(style.ParagraphProperties.ListLevel.LocalValue, i);
                }
            }
        }
    }
}