﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Model.Protection;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
using Telerik.Windows.Documents.Flow.Tests.Utilities.XmlValidation;

namespace Telerik.Windows.Documents.Flow.Tests.Integration.ImportExport.Protection
{
    [TestClass]
    public class DocumentProtectionTests : ProtectionTestsBase
    {
        private RadFlowDocumentEditor editor;

        [TestInitialize]
        public void TestInitialize()
        {
            this.editor = new RadFlowDocumentEditor(new RadFlowDocument());
        }

        [TestCleanup]
        public void TestCleanup()
        {
            this.editor = null;
        }

        [TestMethod]
        public void ProtectDocument_WithNoProtectionMode_ExportImport_Docx()
        {
            // Arrange
            this.editor.InsertText("test");
            this.editor.Protect("123", ProtectionMode.NoProtection);

            // Act
            RadFlowDocument importedDocument = this.ExportImportDocument(this.editor.Document, new DocxFormatProvider());

            //Assert
            StructureAsserter.AsserProtectionSettingsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void ProtectDocument_WithReadOnlyMode_ExportImport_Docx()
        {
            // Arrange
            this.editor.InsertText("test");
            this.editor.Protect("123", ProtectionMode.ReadOnly);

            // Act
            RadFlowDocument importedDocument = this.ExportImportDocument(this.editor.Document, new DocxFormatProvider());

            //Assert
            StructureAsserter.AsserProtectionSettingsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void ProtectDocument_WithAllowCommentsMode_ExportImport_Docx()
        {
            // Arrange
            this.editor.InsertText("test");
            this.editor.Protect("123", ProtectionMode.AllowComments);

            // Act
            RadFlowDocument importedDocument = this.ExportImportDocument(this.editor.Document, new DocxFormatProvider());

            //Assert
            StructureAsserter.AsserProtectionSettingsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void ProtectDocument_WithNoProtectionMode_Import_Docx()
        {
            // Arrange
            this.editor.InsertText("test");
            this.editor.Protect("123", ProtectionMode.NoProtection);

            // Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}ProtectedDocument_NoProtectionMode.docx", TestFilesFolder), new DocxFormatProvider());

            //Assert
            StructureAsserter.AsserProtectionSettingsAreEqual(this.editor.Document, importedDocument, true);
        }

        [TestMethod]
        public void ProtectDocument_WithReadOnlyMode_Import_Docx()
        {
            // Arrange
            this.editor.InsertText("test");
            this.editor.Protect("123", ProtectionMode.ReadOnly);

            // Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}ProtectedDocument_ReadOnlyMode.docx", TestFilesFolder), new DocxFormatProvider());

            //Assert
            StructureAsserter.AsserProtectionSettingsAreEqual(this.editor.Document, importedDocument, true);
        }

        [TestMethod]
        public void ProtectDocument_WithAllowCommentsMode_Import_Docx()
        {
            // Arrange
            this.editor.InsertText("test");
            this.editor.Protect("123", ProtectionMode.AllowComments);

            // Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}ProtectedDocument_AllowCommentsMode.docx", TestFilesFolder), new DocxFormatProvider());

            //Assert
            StructureAsserter.AsserProtectionSettingsAreEqual(this.editor.Document, importedDocument, true);
        }

        [TestMethod]
        public void ProtectDocument_WithAlgorithmSha1_ExportImport_Docx()
        {
            // Arrange
            this.editor.InsertText("test");
            this.editor.Document.ProtectionSettings.AlgorithmName = ProtectionAlgorithmNames.SHA1;
            this.editor.Protect("123");

            // Act
            RadFlowDocument importedDocument = this.ExportImportDocument(this.editor.Document, new DocxFormatProvider());

            //Assert
            StructureAsserter.AsserProtectionSettingsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void ProtectDocument_WithAlgorithmSha256_ExportImport_Docx()
        {
            // Arrange
            this.editor.InsertText("test");
            this.editor.Document.ProtectionSettings.AlgorithmName = ProtectionAlgorithmNames.SHA256;
            this.editor.Protect("123");

            // Act
            RadFlowDocument importedDocument = this.ExportImportDocument(this.editor.Document, new DocxFormatProvider());

            //Assert
            StructureAsserter.AsserProtectionSettingsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void ProtectDocument_WithAlgorithmSha384_ExportImport_Docx()
        {
            // Arrange
            this.editor.InsertText("test");
            this.editor.Document.ProtectionSettings.AlgorithmName = ProtectionAlgorithmNames.SHA384;
            this.editor.Protect("123");

            // Act
            RadFlowDocument importedDocument = this.ExportImportDocument(this.editor.Document, new DocxFormatProvider());

            //Assert
            StructureAsserter.AsserProtectionSettingsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void ProtectDocument_WithAlgorithmSha512_ExportImport_Docx()
        {
            // Arrange
            this.editor.InsertText("test");
            this.editor.Document.ProtectionSettings.AlgorithmName = ProtectionAlgorithmNames.SHA512;
            this.editor.Protect("123");

            // Act
            RadFlowDocument importedDocument = this.ExportImportDocument(this.editor.Document, new DocxFormatProvider());

            //Assert
            StructureAsserter.AsserProtectionSettingsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void UnProtectDocument_WithCorrectPassword_ExportImport_Docx()
        {
            // Arrange
            string password = "123";
            this.editor.InsertText("test");
            this.editor.Protect(password);

            // Act
            RadFlowDocument importedDocument = this.ExportImportDocument(this.editor.Document, new DocxFormatProvider());

            //Assert
            this.editor = new RadFlowDocumentEditor(importedDocument);
            bool isSuccessful = this.editor.Unprotect(password);

            // Assert
            Assert.IsTrue(isSuccessful);
        }

        [TestMethod]
        public void UnProtectDocument_WithIncorrectPassword_ExportImport_Docx()
        {
            // Arrange
            string password = "123";
            this.editor.InsertText("test");
            this.editor.Protect(password);

            // Act
            RadFlowDocument importedDocument = this.ExportImportDocument(this.editor.Document, new DocxFormatProvider());

            //Assert
            this.editor = new RadFlowDocumentEditor(importedDocument);
            bool isSuccessful = this.editor.Unprotect(password + "incorrect");

            // Assert
            Assert.IsFalse(isSuccessful);
        }

        [TestMethod]
        public void UnProtectDocument_WithAlgorithmSha512_Import_Docx()
        {
            // Arrange
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}ProtectedDocument_ReadOnlyMode.docx", TestFilesFolder), new DocxFormatProvider());

            // Act
            this.editor = new RadFlowDocumentEditor(importedDocument);
            bool isSuccessful = this.editor.Unprotect("123");

            // Assert
            Assert.AreEqual(ProtectionAlgorithmNames.SHA512, importedDocument.ProtectionSettings.AlgorithmName);
            Assert.IsTrue(isSuccessful);
        }

        [TestMethod]
        public void ProtectDocument_Validate_Xml()
        {
            // Arrange
            this.editor.InsertText("test");
            this.editor.Protect("123");

            XmlValidator.ValidateOpenXmlDocument(this.editor.Document);
        }
    }
}
