﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Linq;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Model.Protection;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
using Telerik.Windows.Documents.Flow.Tests.Utilities.XmlValidation;

namespace Telerik.Windows.Documents.Flow.Tests.Integration.ImportExport.Protection
{
    [TestClass]
    public class PermissionTests : ProtectionTestsBase
    {
        private RadFlowDocumentEditor editor;
        private PermissionRangeCredentials credentials;

        [TestInitialize]
        public void TestInitialize()
        {
            this.editor = new RadFlowDocumentEditor(new RadFlowDocument());
            this.credentials = new PermissionRangeCredentials("test");
        }

        [TestCleanup]
        public void TestCleanup()
        {
            this.editor = null;
            this.credentials = null;
        }

        [TestMethod]
        public void InsertPermission_AroundInlines_ExportImport_Docx()
        {
            // Arrange
            Run inlineStart = this.editor.InsertText("first");
            this.editor.InsertParagraph();
            Run inlineEnd = this.editor.InsertText("second");

            // Act
            this.editor.InsertPermissionRange(this.credentials, inlineStart, inlineEnd);

            RadFlowDocument importedDocument = this.ExportImportDocument(this.editor.Document, new DocxFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void InsertPermission_AroundInlines_Import_Docx()
        {
            // Arrange
            this.editor.InsertText("before ");
            Run inline = this.editor.InsertText("inside");
            this.editor.InsertText(" after.");

            this.editor.InsertParagraph();

            this.editor.InsertPermissionRange(this.credentials, inline, inline);

            //Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}PermissionAroundInline.docx", TestFilesFolder), new DocxFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void InsertPermission_AroundInlines_Validate_Xml()
        {
            // Arrange
            Run inlineStart = this.editor.InsertText("first");
            this.editor.InsertParagraph();
            Run inlineEnd = this.editor.InsertText("second");

            // Act
            this.editor.InsertPermissionRange(this.credentials, inlineStart, inlineEnd);

            XmlValidator.ValidateOpenXmlDocument(this.editor.Document);
        }

        [TestMethod]
        public void InsertPermission_AroundCell_ExportImport_Docx()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 2);
            TableRow row = table.Rows.First();

            // Act
            this.editor.InsertPermissionRange(this.credentials, row.Cells.First());

            RadFlowDocument importedDocument = this.ExportImportDocument(this.editor.Document, new DocxFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void InsertPermission_AroundCell_Validate_Xml()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 2);
            TableRow row = table.Rows.First();

            // Act
            this.editor.InsertPermissionRange(this.credentials, row.Cells.First());

            XmlValidator.ValidateOpenXmlDocument(this.editor.Document);
        }

        [TestMethod]
        public void InsertPermission_AroundCell_StartAndEndAreInsideParagraphs_Import_Docx()
        {
            // Arrange
            Table table = this.editor.InsertTable(1, 2);
            this.editor.InsertPermissionRange(this.credentials, table.Rows.First().Cells.First());

            this.editor.MoveToTableEnd(table);
            this.editor.InsertParagraph();

            //Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}PermissionAroundCell.docx", TestFilesFolder), new DocxFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void InsertPermission_AroundCell_StartIsInsideParagraph_EndIsAfterTableCell_Import_Docx()
        {
            // Arrange
            Table table = this.editor.InsertTable(1, 2);
            this.editor.MoveToTableEnd(table);
            this.editor.InsertParagraph();

            PermissionRange permission = new PermissionRange(this.editor.Document, this.credentials);

            Paragraph pragraph = table.Rows.First().Cells.First().Blocks.AddParagraph();
            pragraph.Inlines.Add(permission.Start);

            pragraph = table.Rows.First().Cells.Last().Blocks.AddParagraph();
            pragraph.Inlines.Add(permission.End);

            //Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}PermissionAroundCell_StartIsInsideParagraph_EndIsAfterTableCell.docx", TestFilesFolder), new DocxFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void InsertPermission_AroundCell_StartIsBeforeTableCell_EndIsAfterTableCell_Import_Docx()
        {
            // Arrange
            Table table = this.editor.InsertTable(1, 2);
            this.editor.MoveToTableEnd(table);
            this.editor.InsertParagraph();

            PermissionRange permission = new PermissionRange(this.editor.Document, this.credentials);

            Paragraph pragraph = table.Rows.First().Cells.First().Blocks.AddParagraph();
            pragraph.Inlines.Add(permission.Start);

            pragraph = table.Rows.First().Cells.Last().Blocks.AddParagraph();
            pragraph.Inlines.Add(permission.End);

            //Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}PermissionAroundCell_StartIsBeforeTableCell_EndIsAfterTableCell.docx", TestFilesFolder), new DocxFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void InsertPermission_AroundCell_StartIsInsideParagraph_EndIsAfterTableRow_Import_Docx()
        {
            // Arrange
            Table table = this.editor.InsertTable(1, 2);
            this.editor.MoveToTableEnd(table);
            Paragraph endParagraph = this.editor.InsertParagraph();

            PermissionRange permission = new PermissionRange(this.editor.Document, this.credentials);

            foreach (TableCell cell in this.editor.Document.EnumerateChildrenOfType<TableCell>())
            {
                cell.Blocks.AddParagraph();
            }

            Paragraph pragraph = table.Rows.First().Cells.First().Blocks.First() as Paragraph;
            pragraph.Inlines.Add(permission.Start);

            endParagraph.Inlines.Add(permission.End);

            //Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}PermissionAroundCell_StartIsInsideParagraph_EndIsAfterTableRow.docx", TestFilesFolder), new DocxFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void InsertPermission_AroundCell_WhenTheFirstCellIsColumnSpanned_ExportImport_Docx()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 1);
            TableCell firstCell = table.Rows.First().Cells.First();
            firstCell.ColumnSpan = 2;

            table.Rows.Last().Cells.AddTableCell();

            // Act
            this.editor.InsertPermissionRange(this.credentials, firstCell);

            RadFlowDocument importedDocument = this.ExportImportDocument(this.editor.Document, new DocxFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void InsertPermission_AroundCell_WhenTheFirstCellIsRowSpanned_ExportImport_Docx()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 1);
            TableCell firstCell = table.Rows.First().Cells.First();
            firstCell.RowSpan = 2;

            table.Rows.First().Cells.AddTableCell();

            // Act
            this.editor.InsertPermissionRange(this.credentials, firstCell);

            RadFlowDocument importedDocument = this.ExportImportDocument(this.editor.Document, new DocxFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void InsertPermission_AroundRow_ExportImport_Docx()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 2);
            TableRow firstRow = table.Rows.First();

            // Act
            this.editor.InsertPermissionRange(this.credentials, firstRow);

            RadFlowDocument importedDocument = this.ExportImportDocument(this.editor.Document, new DocxFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void InsertPermission_AroundRow_Validate_Xml()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 2);
            TableRow firstRow = table.Rows.First();

            // Act
            this.editor.InsertPermissionRange(this.credentials, firstRow);

            XmlValidator.ValidateOpenXmlDocument(this.editor.Document);
        }

        [TestMethod]
        public void InsertPermission_AroundRow_StartAndEndAreInsideParagraphs_Import_Docx()
        {
            // Arrange
            Table table = this.editor.InsertTable(1, 2);
            foreach (TableCell cell in table.EnumerateChildrenOfType<TableCell>())
            {
                cell.Blocks.AddParagraph();
            }

            TableRow firstRow = table.Rows.First();
            this.editor.MoveToTableEnd(table);
            this.editor.InsertParagraph();

            this.editor.InsertPermissionRange(this.credentials, firstRow);

            // Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}PermissionAroundRow.docx", TestFilesFolder), new DocxFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void InsertPermission_AroundRow_StartIsInsideParagraph_EndIsAfterTableRow_Import_Docx()
        {
            // Arrange
            Table table = this.editor.InsertTable(1, 2);
            foreach (TableCell cell in table.EnumerateChildrenOfType<TableCell>())
            {
                cell.Blocks.AddParagraph();
            }

            TableRow firstRow = table.Rows.First();
            this.editor.MoveToTableEnd(table);
            Paragraph endParagraph = this.editor.InsertParagraph();

            PermissionRange permission = new PermissionRange(this.editor.Document, this.credentials);

            Paragraph pragraph = firstRow.Cells.First().Blocks.First() as Paragraph;
            pragraph.Inlines.Add(permission.Start);

            endParagraph.Inlines.Add(permission.End);

            // Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}PermissionAroundRow_StartIsInsideParagraph_EndIsAfterTableRow.docx", TestFilesFolder), new DocxFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void InsertPermission_AroundRow_StartIsBeforeTableRow_EndIsAfterTableRow_Import_Docx()
        {
            // Arrange
            Table table = this.editor.InsertTable(1, 2);
            foreach (TableCell cell in table.EnumerateChildrenOfType<TableCell>())
            {
                cell.Blocks.AddParagraph();
            }

            TableRow firstRow = table.Rows.First();
            this.editor.MoveToTableEnd(table);
            Paragraph endParagraph = this.editor.InsertParagraph();

            PermissionRange permission = new PermissionRange(this.editor.Document, this.credentials);

            Paragraph pragraph = firstRow.Cells.First().Blocks.First() as Paragraph;
            pragraph.Inlines.Add(permission.Start);

            endParagraph.Inlines.Add(permission.End);

            // Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}PermissionAroundRow_StartIsBeforeTableRow_EndIsAfterTableRow.docx", TestFilesFolder), new DocxFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void InsertPermission_AroundRow_WhenTheFirstCellIsColumnSpanned_ExportImport_Docx()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 1);
            TableRow firstRow = table.Rows.First();
            TableCell firstCell = firstRow.Cells.First();
            firstCell.ColumnSpan = 2;

            table.Rows.Last().Cells.AddTableCell();

            // Act
            this.editor.InsertPermissionRange(this.credentials, firstCell);

            RadFlowDocument importedDocument = this.ExportImportDocument(this.editor.Document, new DocxFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(this.editor.Document, importedDocument);
        }

        [TestMethod]
        public void InsertPermission_AroundRow_WhenTheFirstCellIsColumnSpanned_Import_Docx()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 1);
            TableRow firstRow = table.Rows.First();
            TableCell firstCell = firstRow.Cells.First();
            firstCell.ColumnSpan = 2;

            table.Rows.Last().Cells.AddTableCell();

            foreach (TableCell cell in table.EnumerateChildrenOfType<TableCell>())
            {
                cell.Blocks.AddParagraph();
            }

            this.editor.InsertPermissionRange(this.credentials, firstCell);
            this.editor.MoveToTableEnd(table);
            this.editor.InsertParagraph();

            // Act
            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(string.Format("{0}PermissionAroundRow_WhenTheFirstCellIsColumnSpanned.docx", TestFilesFolder), new DocxFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(this.editor.Document, importedDocument);
        }

        // TODO: all of the exported cells for rows spanned cell will export containing paragraphs and content.
        //[TestMethod]
        public void InsertPermission_AroundRow_WhenTheFirstCellIsRowSpanned_ExportImport_Docx()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 1);
            TableRow firstRow = table.Rows.First();
            TableCell firstCell = firstRow.Cells.First();
            firstCell.RowSpan = 2;

            firstRow.Cells.AddTableCell();

            // Act
            this.editor.InsertPermissionRange(this.credentials, firstRow);

            RadFlowDocument importedDocument = this.ExportImportDocument(this.editor.Document, new DocxFormatProvider());

            //Assert
            StructureAsserter.AssertDocumentsAreEqual(this.editor.Document, importedDocument);
        }

        //[TestMethod]
        public void InsertPermission_AroundRow_WhenTheFirstCellIsRowSpanned_Import_Docx()
        {
            Assert.IsTrue(false);
        }
    }
}
