﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Linq;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Tests.Utilities;

namespace Telerik.Windows.Documents.Flow.Tests.Integration.ImportExport.Tables
{
    internal class TableHelper
    {
        private const string CellContentTemplate = "Cell [{0}, {1}]";

        public static void FillTable(Table table, int rows, int columns)
        {
            for (int row = 0; row < rows; row++)
            {
                TableRow tableRow = table.Rows.AddTableRow();
                for (int col = 0; col < columns; col++)
                {
                    tableRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun().Text = string.Format(CellContentTemplate, row, col);
                }
            }
        }

        public static void AssertTable(Table table, int rows, int columns)
        {
            Assert.AreEqual(rows, table.Rows.Count);
            for (int row = 0; row < table.Rows.Count; row++)
            {
                TableRow tableRow = table.Rows[row];
                Assert.AreEqual(columns, table.Rows.Count);
                for (int col = 0; col < table.Rows.Count; col++)
                {
                    TableCell cell = tableRow.Cells[col];
                    Assert.AreEqual(1, cell.Blocks.Count);
                    DocumentAsserter.AssertParagraphWithText(cell.Blocks[0], string.Format(CellContentTemplate, row, col));
                }
            }
        }

        public static void FillNestedTable(Table table)
        {
            int rows = 3;
            int columns = 3;
            int nestedRowIndex = 1;
            int nestedColIndex = 1;


            for (int row = 0; row < rows; row++)
            {
                TableRow tableRow = table.Rows.AddTableRow();
                for (int col = 0; col < columns; col++)
                {
                    TableCell cell = tableRow.Cells.AddTableCell();
                    cell.Blocks.AddParagraph().Inlines.AddRun().Text = string.Format(CellContentTemplate, row, col);

                    if (row == nestedRowIndex && col == nestedColIndex)
                    {
                        Table nestedTable = cell.Blocks.AddTable();
                        TableHelper.FillTable(nestedTable, 2, 2);
                    }
                }
            }
        }

        public static void AssertNestedTable(Table table, bool containsAdditionalParagraphAfterTheNestedTable)
        {
            int rows = 3;
            int columns = 3;
            int nestedRowIndex = 1;
            int nestedColIndex = 1;

            Assert.AreEqual(rows, table.Rows.Count);
            for (int row = 0; row < table.Rows.Count; row++)
            {
                TableRow tableRow = table.Rows[row];
                Assert.AreEqual(columns, table.Rows.Count);
                for (int col = 0; col < table.Rows.Count; col++)
                {
                    TableCell cell = tableRow.Cells[col];
                    DocumentAsserter.AssertParagraphWithText(cell.Blocks[0], string.Format(CellContentTemplate, row, col));
                    if (row == nestedRowIndex && col == nestedColIndex)
                    {
                        if (containsAdditionalParagraphAfterTheNestedTable)
                        {
                            Assert.AreEqual(3, cell.Blocks.Count);
                        }
                        else
                        {
                            Assert.AreEqual(2, cell.Blocks.Count);
                        }

                        Assert.IsInstanceOfType(cell.Blocks[1], typeof(Table));

                        TableHelper.AssertTable((Table)cell.Blocks[1], 2, 2);

                        if (containsAdditionalParagraphAfterTheNestedTable)
                        {
                            Assert.IsInstanceOfType(cell.Blocks[2], typeof(Paragraph));
                            Assert.IsFalse(((Paragraph)cell.Blocks[2]).Inlines.Any());
                        }
                    }
                    else
                    {
                        Assert.AreEqual(1, cell.Blocks.Count);
                    }
                }
            }
        }

        public static void FillTableWithHorizontalMerge(Table table)
        {
            TableRow row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 1), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 2), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 3), 1, 1);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 1), 2, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 3), 1, 1);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 1), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 2), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 3), 1, 1);
        }

        public static void FillTableWithTwoConsecutiveHorizontallyMergedGroups(Table table)
        {
            TableRow row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 1), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 2), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 3), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 4), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 5), 1, 1);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 1), 2, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 3), 2, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 5), 1, 1);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 1), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 2), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 3), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 4), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 5), 1, 1);
        }

        public static void FillTableWithVerticalMerge(Table table)
        {
            TableRow row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 1), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 2), 1, 1);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 1), 1, 2);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 2), 1, 1);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 2), 1, 1);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 3, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 3, 1), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 3, 2), 1, 1);
        }

        public static void FillTableWithVerticallyMergedFirstColumn(Table table)
        {
            TableRow row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 0), 1, 3);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 1), 1, 2);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 2), 1, 1);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 1), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 2), 1, 1);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 3, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 3, 1), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 3, 2), 1, 1);
        }

        public static void FillTableWithTwoConsecutiveVerticallyMergedGroups(Table table)
        {
            TableRow row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 1), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 2), 1, 1);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 1), 1, 2);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 2), 1, 1);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 2), 1, 1);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 3, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 3, 1), 1, 2);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 3, 2), 1, 1);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 4, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 4, 2), 1, 1);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 5, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 5, 1), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 5, 2), 1, 1);
        }

        public static void FillTableWithVerticalAndHorizontalMerge(Table table)
        {
            TableRow row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 1), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 2), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 0, 3), 1, 1);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 1), 2, 2);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 1, 3), 1, 1);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 2, 3), 1, 1);

            row = table.Rows.AddTableRow();
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 3, 0), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 3, 1), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 3, 2), 1, 1);
            TableHelper.AddCell(row, string.Format(CellContentTemplate, 3, 3), 1, 1);
        }

        private static void AddCell(TableRow tableRow, string content, int colspan, int rowspan)
        {
            TableCell cell = tableRow.Cells.AddTableCell();
            cell.Blocks.AddParagraph().Inlines.AddRun().Text = content;
            cell.ColumnSpan = colspan;
            cell.RowSpan = rowspan;
        }
    }
}
