﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.IO;
using Telerik.Windows.Documents.Common.FormatProviders;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
using Telerik.Windows.Documents.Flow.Tests.Utilities.XmlValidation;

namespace Telerik.Windows.Documents.Flow.Tests.Integration.ImportExport.Tables
{
    [TestClass]
    public class VerticallyMergedCellsTests
    {
        private const string TestFileNameWithoutExtension = "Tables/VerticallyMergedCells";

        [TestMethod]
        public void VerticallyMergedCells_Import_Rtf()
        {
            RadFlowDocument doc = TestHelper.GetDocumentFromFile(TestFileNameWithoutExtension + ".rtf", new RtfFormatProvider());

            //Assert
            this.AssertTableWithVerticalMerge(doc);
        }

        [TestMethod]
        public void VerticallyMergedCells_Import_Docx()
        {
            RadFlowDocument doc = TestHelper.GetDocumentFromFile(TestFileNameWithoutExtension + ".docx", new DocxFormatProvider());

            //Assert
            this.AssertTableWithVerticalMerge(doc);
        }

        [TestMethod]
        public void VerticallyMergedCells_ExportImport_Rtf()
        {
            this.VerticallyMergedCellsExportImportTemplate(new RtfFormatProvider());
        }

        [TestMethod]
        public void VerticallyMergedCells_ExportImport_Docx()
        {
            this.VerticallyMergedCellsExportImportTemplate(new DocxFormatProvider());
        }

        [TestMethod]
        public void OneSection_Validate_Xml()
        {
            RadFlowDocument document = new RadFlowDocument();
            Table table = document.Sections.AddSection()
                       .Blocks.AddTable();

            TableHelper.FillTableWithVerticalMerge(table);
            document.Sections[0].Blocks.AddParagraph();

            XmlValidator.ValidateOpenXmlDocument(document);
        }

        private void VerticallyMergedCellsExportImportTemplate(IFormatProvider<RadFlowDocument> provider)
        {
            //Arrange
            RadFlowDocument originalDoc = this.CreateDocumentWithTableWithVerticalMerge();

            //Act
            RadFlowDocument testDocument = null;
            using (Stream stream = new MemoryStream())
            {
                provider.Export(originalDoc, stream);
                stream.Seek(0, SeekOrigin.Begin);

                // Uncomment to debug.
                ////StreamReader r = new StreamReader(stream);
                ////var docString = r.ReadToEnd();
                ////stream.Seek(0, SeekOrigin.Begin);

                testDocument = provider.Import(stream);
            }

            //Assert
            this.AssertTableWithVerticalMerge(testDocument);
        }

        private RadFlowDocument CreateDocumentWithTableWithVerticalMerge()
        {
            RadFlowDocument document = new RadFlowDocument();
            Table table = document.Sections.AddSection()
                       .Blocks.AddTable();

            TableHelper.FillTableWithVerticalMerge(table);
            document.Sections[0].Blocks.AddParagraph();

            return document;
        }

        private void AssertTableWithVerticalMerge(RadFlowDocument doc)
        {
            //Assert
            Assert.AreEqual(1, doc.Sections.Count);
            Assert.AreEqual(2, doc.Sections[0].Blocks.Count);
            DocumentAsserter.AssertEmptyParagraph(doc.Sections[0].Blocks[1]);

            Table table = doc.Sections[0].Blocks[0] as Table;
            Table expected = new RadFlowDocument()
                .Sections.AddSection()
                .Blocks.AddTable();
            TableHelper.FillTableWithVerticalMerge(expected);

            StructureAsserter.AssertTablesAreEqual(expected, table);
        }

        [TestMethod]
        public void VerticallyMergedCells_Export_AssertOnlyOneOfMergedCellsHasContent_Docx()
        {
            //Arrange
            RadFlowDocument document = this.CreateDocumentWithTableWithVerticalMerge();

            //Act
            string mainPartXml = XmlValidationHelper.GetMainDocumentXmlFromDocument(document);

            //Assert
            int contentCount = 0;

            int currentIndex = 0;
            while (currentIndex >= 0)
            {
                currentIndex = mainPartXml.IndexOf("Cell [1, 1]", currentIndex + 1);
                if (currentIndex >= 0)
                {
                    contentCount++;
                }
            }

            Assert.AreEqual(1, contentCount);
        }

        [TestMethod]
        public void VerticallyMergedCellsFirstColumn_ExportImport_Rtf()
        {
            //Arrange
            RadFlowDocument originalDoc = this.CreateDocumentWithTableWithMergedFirstColumn();

            //Act
            RadFlowDocument testDocument = null;
            RtfFormatProvider provider = new RtfFormatProvider();
            using (Stream stream = new MemoryStream())
            {
                provider.Export(originalDoc, stream);
                stream.Seek(0, SeekOrigin.Begin);
                testDocument = provider.Import(stream);
            }

            //Assert
            this.AssertTableWithMergedFirstColumn(testDocument);
        }

        private RadFlowDocument CreateDocumentWithTableWithMergedFirstColumn()
        {
            RadFlowDocument document = new RadFlowDocument();
            Table table = document.Sections.AddSection()
                       .Blocks.AddTable();

            TableHelper.FillTableWithVerticallyMergedFirstColumn(table);
            document.Sections[0].Blocks.AddParagraph();

            return document;
        }

        private void AssertTableWithMergedFirstColumn(RadFlowDocument doc)
        {
            //Assert
            Assert.AreEqual(1, doc.Sections.Count);
            Assert.AreEqual(2, doc.Sections[0].Blocks.Count);
            DocumentAsserter.AssertEmptyParagraph(doc.Sections[0].Blocks[1]);

            Table table = doc.Sections[0].Blocks[0] as Table;
            Table expected = new RadFlowDocument()
                .Sections.AddSection()
                .Blocks.AddTable();
            TableHelper.FillTableWithVerticallyMergedFirstColumn(expected);

            StructureAsserter.AssertTablesAreEqual(expected, table);
        }
    }
}
