﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Linq;
using Telerik.Windows.Documents.Flow.FormatProviders.Docx;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Watermarks;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
using Telerik.Windows.Documents.Media;

namespace Telerik.Windows.Documents.Flow.Tests.Integration.ImportExport.Watermarks
{
    [TestClass]
    public class ImageWatermarkTests : WatermarkTestsBase
    {
        // Create watermark programmatically, export and import.

        [TestMethod]
        public void ImageWatermark_DefaultScenario_ExportImport_Docx()
        {
            RadFlowDocument originalDocument = new RadFlowDocument();
            originalDocument.Sections.AddSection();

            ImageWatermarkTests.SetImageWatermarkToHeader(originalDocument.Sections.First().Headers.Add(), 100, 50, 10, new ImageSource(TestHelper.GetImageStream("test.png"), "png"));

            RadFlowDocument importedDocument = this.ExportImportDocument(originalDocument, new DocxFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(originalDocument, importedDocument);
        }

        [TestMethod]
        public void ImageWatermark_DefaultScenario_ExportImport_Rtf()
        {
            RadFlowDocument originalDocument = new RadFlowDocument();
            originalDocument.Sections.AddSection();

            ImageWatermarkTests.SetImageWatermarkToHeader(originalDocument.Sections.First().Headers.Add(), 100, 50, 10, new ImageSource(TestHelper.GetImageStream("test.png"), "png"));

            RadFlowDocument importedDocument = this.ExportImportDocument(originalDocument, new RtfFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(originalDocument, importedDocument);
        }

        [TestMethod]
        public void ImageWatermark_NullImageSource_ExportImport_Docx()
        {
            RadFlowDocument originalDocument = new RadFlowDocument();
            originalDocument.Sections.AddSection();

            ImageWatermarkTests.SetImageWatermarkToHeader(originalDocument.Sections.First().Headers.Add(), 100, 50, 10, null);

            RadFlowDocument importedDocument = this.ExportImportDocument(originalDocument, new DocxFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(originalDocument, importedDocument);
        }

        [TestMethod]
        public void ImageWatermark_NullImageSource_ExportImport_Rtf()
        {
            RadFlowDocument originalDocument = new RadFlowDocument();
            originalDocument.Sections.AddSection();

            ImageWatermarkTests.SetImageWatermarkToHeader(originalDocument.Sections.First().Headers.Add(), 100, 50, 10, null);

            RadFlowDocument importedDocument = this.ExportImportDocument(originalDocument, new RtfFormatProvider());

            //When the image sourcee is null in RTF the watermark is not exported at all.
            RadFlowDocument expectedDocument = new RadFlowDocument();
            expectedDocument.Sections.AddSection();
            expectedDocument.Sections.First().Headers.Add();

            StructureAsserter.AssertDocumentsAreEqual(expectedDocument, importedDocument);
        }

        [TestMethod]
        public void ImageWatermark_WatermarkWithEmptySettings_ExportImport_Docx()
        {
            RadFlowDocument originalDocument = new RadFlowDocument();
            originalDocument.Sections.AddSection();

            Header header = originalDocument.Sections.First().Headers.Add();
            Watermark imageWatermark = new Watermark(new ImageWatermarkSettings());
            header.Watermarks.Add(imageWatermark);

            RadFlowDocument importedDocument = this.ExportImportDocument(originalDocument, new DocxFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(originalDocument, importedDocument);
        }

        [TestMethod]
        public void ImageWatermark_WatermarkWithEmptySettings_ExportImport_Rtf()
        {
            RadFlowDocument originalDocument = new RadFlowDocument();
            originalDocument.Sections.AddSection();

            Header header = originalDocument.Sections.First().Headers.Add();
            Watermark imageWatermark = new Watermark(new ImageWatermarkSettings());
            header.Watermarks.Add(imageWatermark);

            RadFlowDocument importedDocument = this.ExportImportDocument(originalDocument, new RtfFormatProvider());

            //When the image sourcee is null in RTF the watermark is not exported at all.
            RadFlowDocument expectedDocument = new RadFlowDocument();
            expectedDocument.Sections.AddSection();
            expectedDocument.Sections.First().Headers.Add();

            StructureAsserter.AssertDocumentsAreEqual(expectedDocument, importedDocument);
        }

        // Import documents with watermark.

        [TestMethod]
        public void ImageWatermark_WordGenerated_Import_Docx()
        {
            string fileName = "Watermarks/WordImageWatermark.docx";

            RadFlowDocument expectedDocument = new RadFlowDocument();
            expectedDocument.Sections.AddSection().Blocks.AddParagraph();

            Header defaultHeader = expectedDocument.Sections.First().Headers.Add();
            Header evenHeader = expectedDocument.Sections.First().Headers.Add(HeaderFooterType.Even);
            Header firstHeader = expectedDocument.Sections.First().Headers.Add(HeaderFooterType.First);

            double width = 90;
            double height = 30;
            double angle = 40;
            ImageSource imageSource = new ImageSource(TestHelper.GetImageStream("test.png"), "png");

            ImageWatermarkTests.SetImageWatermarkToHeader(defaultHeader, width, height, angle, imageSource);
            ImageWatermarkTests.SetImageWatermarkToHeader(evenHeader, width, height, angle, imageSource);
            ImageWatermarkTests.SetImageWatermarkToHeader(firstHeader, width, height, angle, imageSource);

            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(fileName, new DocxFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(expectedDocument, importedDocument);
        }

        [TestMethod]
        public void ImageWatermark_RichTextBoxGenerated_Import_Docx()
        {
            string fileName = "Watermarks/RTBImageWatermark.docx";

            RadFlowDocument expectedDocument = new RadFlowDocument();
            expectedDocument.Sections.AddSection().Blocks.AddParagraph();
            Header header = expectedDocument.Sections.First().Headers.Add();

            SetImageWatermarkToHeader(header, 90, 30, 0, new ImageSource(TestHelper.GetImageStream("test.png"), "png"));

            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(fileName, new DocxFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(expectedDocument, importedDocument);
        }

        [TestMethod]
        public void ImageWatermark_WordGenerated_Import_Rtf()
        {
            string fileName = "Watermarks/WordImageWatermark.rtf";

            RadFlowDocument expectedDocument = new RadFlowDocument();
            expectedDocument.Sections.AddSection().Blocks.AddParagraph();

            Header defaultHeader = expectedDocument.Sections.First().Headers.Add();
            defaultHeader.Blocks.AddParagraph();
            Header evenHeader = expectedDocument.Sections.First().Headers.Add(HeaderFooterType.Even);
            evenHeader.Blocks.AddParagraph();
            Header firstHeader = expectedDocument.Sections.First().Headers.Add(HeaderFooterType.First);
            firstHeader.Blocks.AddParagraph();

            double width = 90;
            double height = 30;
            double angle = 40;
            ImageSource imageSource = new ImageSource(TestHelper.GetImageStream("test.png"), "png");

            ImageWatermarkTests.SetImageWatermarkToHeader(defaultHeader, width, height, angle, imageSource);
            ImageWatermarkTests.SetImageWatermarkToHeader(evenHeader, width, height, angle, imageSource);
            ImageWatermarkTests.SetImageWatermarkToHeader(firstHeader, width, height, angle, imageSource);

            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(fileName, new RtfFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(expectedDocument, importedDocument);
        }

    }
}
