﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Media;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Flow.FormatProviders.Docx;
using Telerik.Windows.Documents.Flow.FormatProviders.Rtf;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Watermarks;
using Telerik.Windows.Documents.Flow.Tests.Utilities;

namespace Telerik.Windows.Documents.Flow.Tests.Integration.ImportExport.Watermarks
{
    [TestClass]
    public class TextWatermarkTests : WatermarkTestsBase
    {
        // Create watermark programmatically, export and import.

        [TestMethod]
        public void TextWatermark_DefaultScenario_ExportImport_Docx()
        {
            RadFlowDocument originalDocument = new RadFlowDocument();
            originalDocument.Sections.AddSection();

            TextWatermarkTests.SetTextWatermarkToHeader(originalDocument.Sections.First().Headers.Add(), 100, 50, 10, 0.4, Colors.Red, "Confidential", new FontFamily("Arial"));

            RadFlowDocument importedDocument = this.ExportImportDocument(originalDocument, new DocxFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(originalDocument, importedDocument);
        }

        [TestMethod]
        public void TextWatermark_DefaultScenario_ExportImport_Rtf()
        {
            RadFlowDocument originalDocument = new RadFlowDocument();
            originalDocument.Sections.AddSection();

            TextWatermarkTests.SetTextWatermarkToHeader(originalDocument.Sections.First().Headers.Add(), 100, 50, 10, 0.4, Colors.Red, "Confidential", new FontFamily("Arial"));

            RadFlowDocument importedDocument = this.ExportImportDocument(originalDocument, new RtfFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(originalDocument, importedDocument);
        }

        [TestMethod]
        public void TextWatermark_WatermarkWithEmptySettings_ExportImport_Docx()
        {
            RadFlowDocument originalDocument = new RadFlowDocument();
            originalDocument.Sections.AddSection();

            Header header = originalDocument.Sections.First().Headers.Add();
            Watermark textWatermark = new Watermark(new TextWatermarkSettings());
            header.Watermarks.Add(textWatermark);

            RadFlowDocument importedDocument = this.ExportImportDocument(originalDocument, new DocxFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(originalDocument, importedDocument);
        }

        [TestMethod]
        public void TextWatermark_WatermarkWithEmptySettings_ExportImport_Rtf()
        {
            RadFlowDocument originalDocument = new RadFlowDocument();
            originalDocument.Sections.AddSection();

            Header header = originalDocument.Sections.First().Headers.Add();
            Watermark textWatermark = new Watermark(new TextWatermarkSettings() { ForegroundColor = Color.FromArgb(255, 0, 0, 0) });
            header.Watermarks.Add(textWatermark);

            RadFlowDocument importedDocument = this.ExportImportDocument(originalDocument, new RtfFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(originalDocument, importedDocument);
        }

        // Import documents with watermark.

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.Watermark)]
        public void TextWatermark_RichTextBoxGenerated_Import_Docx()
        {
            string fileName = "Watermarks/RTBTextWatermark.docx";

            RadFlowDocument expectedDocument = new RadFlowDocument();
            expectedDocument.Sections.AddSection().Blocks.AddParagraph();
            Header header = expectedDocument.Sections.First().Headers.Add();

            TextWatermarkTests.SetTextWatermarkToHeader(header, 480, 96, 315, 0.25, Colors.Red, "CONFIDENTIAL", new FontFamily("Calibri"));

            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(fileName, new DocxFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(expectedDocument, importedDocument);

#if VISUAL
            AssertVisual(importedDocument, VisualTestsFormats);
#endif
        }

        [TestMethod]
        public void TextWatermark_WordGenerated_Import_Docx()
        {
            string fileName = "Watermarks/WordTextWatermark.docx";

            RadFlowDocument expectedDocument = CreateFlowDocumentWithFootersAndSectionWithParagraph();

            Header defaultHeader = expectedDocument.Sections.First().Headers.Add();
            Header evenHeader = expectedDocument.Sections.First().Headers.Add(HeaderFooterType.Even);
            Header firstHeader = expectedDocument.Sections.First().Headers.Add(HeaderFooterType.First);

            double width = 782.06666666666661;
            double height = 97.73333333333332;
            double angle = 315;
            double opacity = 0.5;
            Color color = Colors.Red;
            string text = "CONFIDENTIAL";
            FontFamily fontFamily = new FontFamily("Verdana");

            TextWatermarkTests.SetTextWatermarkToHeader(defaultHeader, width, height, angle, opacity, color, text, fontFamily);
            TextWatermarkTests.SetTextWatermarkToHeader(evenHeader, width, height, angle, opacity, color, text, fontFamily);
            TextWatermarkTests.SetTextWatermarkToHeader(firstHeader, width, height, angle, opacity, color, text, fontFamily);

            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(fileName, new DocxFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(expectedDocument, importedDocument);
        }

        [TestMethod]
        public void TextWatermark_WordGenerated_Import_Rtf()
        {
            string fileName = "Watermarks/WordTextWatermark.rtf";

            RadFlowDocument expectedDocument = CreateFlowDocumentWithFootersAndSectionWithParagraph();

            Header defaultHeader = expectedDocument.Sections.First().Headers.Add();
            defaultHeader.Blocks.AddParagraph();
            Header evenHeader = expectedDocument.Sections.First().Headers.Add(HeaderFooterType.Even);
            evenHeader.Blocks.AddParagraph();
            Header firstHeader = expectedDocument.Sections.First().Headers.Add(HeaderFooterType.First);
            firstHeader.Blocks.AddParagraph();

            double width = 782.06666666666672;
            double height = 97.733333333333334;
            double angle = 315;
            double opacity = 0.5;
            Color color = Colors.Red;
            string text = "CONFIDENTIAL";
            FontFamily fontFamily = new FontFamily("Verdana");

            TextWatermarkTests.SetTextWatermarkToHeader(defaultHeader, width, height, angle, opacity, color, text, fontFamily);
            TextWatermarkTests.SetTextWatermarkToHeader(evenHeader, width, height, angle, opacity, color, text, fontFamily);
            TextWatermarkTests.SetTextWatermarkToHeader(firstHeader, width, height, angle, opacity, color, text, fontFamily);

            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(fileName, new RtfFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(expectedDocument, importedDocument);
        }

        [TestMethod]
        public void TextWatermark_WordGenerated_InvalidFont_Import_Docx()
        {
            // Word is able to display the watermark with a different font.
            string fileName = "Watermarks/WordTextWatermarkInvalidFontFamily.docx";

            RadFlowDocument expectedDocument = CreateFlowDocumentWithFootersAndSectionWithParagraph();

            Header defaultHeader = expectedDocument.Sections.First().Headers.Add();
            Header evenHeader = expectedDocument.Sections.First().Headers.Add(HeaderFooterType.Even);
            Header firstHeader = expectedDocument.Sections.First().Headers.Add(HeaderFooterType.First);

            double width = 754.06666666666661;
            double height = 125.66666666666667;
            double angle = 315;
            double opacity = 1;
            Color color = Colors.Red;
            string text = "CONFIDENTIAL";
            FontFamily fontFamily = new FontFamily("Invalid");

            TextWatermarkTests.SetTextWatermarkToHeader(defaultHeader, width, height, angle, opacity, color, text, fontFamily);
            TextWatermarkTests.SetTextWatermarkToHeader(evenHeader, width, height, angle, opacity, color, text, fontFamily);
            TextWatermarkTests.SetTextWatermarkToHeader(firstHeader, width, height, angle, opacity, color, text, fontFamily);

            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(fileName, new DocxFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(expectedDocument, importedDocument);
        }

        [TestMethod]
        public void TextWatermark_WordGenerated_InvalidFont_Import_Rtf()
        {
            // Word is able to display the watermark with a different font.
            string fileName = "Watermarks/WordTextWatermarkInvalidFontFamily.rtf";

            RadFlowDocument expectedDocument = CreateFlowDocumentWithFootersAndSectionWithParagraph();

            Header defaultHeader = expectedDocument.Sections.First().Headers.Add();
            defaultHeader.Blocks.AddParagraph();
            Header evenHeader = expectedDocument.Sections.First().Headers.Add(HeaderFooterType.Even);
            evenHeader.Blocks.AddParagraph();
            Header firstHeader = expectedDocument.Sections.First().Headers.Add(HeaderFooterType.First);
            firstHeader.Blocks.AddParagraph();

            double width = 754.06666666666661;
            double height = 125.66666666666667;
            double angle = 315;
            double opacity = 1;
            Color color = Colors.Red;
            string text = "CONFIDENTIAL";
            FontFamily fontFamily = new FontFamily("Invalid");

            TextWatermarkTests.SetTextWatermarkToHeader(defaultHeader, width, height, angle, opacity, color, text, fontFamily);
            TextWatermarkTests.SetTextWatermarkToHeader(evenHeader, width, height, angle, opacity, color, text, fontFamily);
            TextWatermarkTests.SetTextWatermarkToHeader(firstHeader, width, height, angle, opacity, color, text, fontFamily);

            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(fileName, new RtfFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(expectedDocument, importedDocument);
        }

        [TestMethod]
        public void TextWatermark_WordGenerated_InvalidColor_Import_Docx()
        {
            // For the ST_ColorType any string for the color will result in a valid document. Word displays watermarks with invalid colors as white/transparent.
            string fileName = "Watermarks/WordTextWatermarkInvalidColor.docx";

            RadFlowDocument expectedDocument = CreateFlowDocumentWithFootersAndSectionWithParagraph();

            Header defaultHeader = expectedDocument.Sections.First().Headers.Add();
            Header evenHeader = expectedDocument.Sections.First().Headers.Add(HeaderFooterType.Even);
            Header firstHeader = expectedDocument.Sections.First().Headers.Add(HeaderFooterType.First);

            double width = 782.06666666666661;
            double height = 97.73333333333332;
            double angle = 315;
            double opacity = 1;
            Color color = new Color();
            string text = "CONFIDENTIAL";
            FontFamily fontFamily = new FontFamily("Verdana");

            TextWatermarkTests.SetTextWatermarkToHeader(defaultHeader, width, height, angle, opacity, color, text, fontFamily);
            TextWatermarkTests.SetTextWatermarkToHeader(evenHeader, width, height, angle, opacity, color, text, fontFamily);
            TextWatermarkTests.SetTextWatermarkToHeader(firstHeader, width, height, angle, opacity, color, text, fontFamily);

            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(fileName, new DocxFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(expectedDocument, importedDocument);
        }

        [TestMethod]
        public void TextWatermark_WordGenerated_ThreeDigitColor_Import_Docx()
        {
            // The text watermark forecolor is #f30 in the model.
            string fileName = "Watermarks/WordTextWatermarkThreeDigicColor.docx";

            RadFlowDocument expectedDocument = new RadFlowDocument();
            expectedDocument.Sections.AddSection().Blocks.AddParagraph();

            Header defaultHeader = expectedDocument.Sections.First().Headers.Add();

            double width = 549.86666666666656;
            double height = 329.93333333333328;
            double angle = 315;
            double opacity = 0.5;
            Color color = Color.FromRgb(255, 51, 0);
            string text = "DRAFT";

            TextWatermarkTests.SetTextWatermarkToHeader(defaultHeader, width, height, angle, opacity, color, text, new FontFamily("Calibri"));

            RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(fileName, new DocxFormatProvider());

            StructureAsserter.AssertDocumentsAreEqual(expectedDocument, importedDocument);
        }

        //TODO: Implement the ability to parse opacity presented in "64881f" format.
        // See bug 159589 Unable to import documents with watermark with opacity in format "64881f"
        //[TestMethod]
        //public void TextWatermark_WordGenerated_SmallOpacity_Import_Docx()
        //{
        //    // Setting the transparency to 1% in Word leads to opacity displayed as 64881f in the model.
        //    string fileName = "Watermarks/WordTextWatermarkSmallTransparency.docx";

        //    RadFlowDocument expectedDocument = this.CreateFlowDocumentWithFootersAndSectionWithParagraph();

        //    Header defaultHeader = expectedDocument.Sections.First().Headers.Add();
        //    Header evenHeader = expectedDocument.Sections.First().Headers.Add(HeaderFooterType.Even);
        //    Header firstHeader = expectedDocument.Sections.First().Headers.Add(HeaderFooterType.First);

        //    double width = 586.55;
        //    double height = 73.3;
        //    double angle = 315;
        //    double opacity = 0.99;
        //    Color color = Colors.Red;
        //    string text = "CONFIDENTIAL";
        //    FontFamily fontFamily = new FontFamily("Verdana");

        //    TextWatermarkTests.SetTextWatermarkToHeader(defaultHeader, width, height, angle, opacity, color, text, fontFamily);
        //    TextWatermarkTests.SetTextWatermarkToHeader(evenHeader, width, height, angle, opacity, color, text, fontFamily);
        //    TextWatermarkTests.SetTextWatermarkToHeader(firstHeader, width, height, angle, opacity, color, text, fontFamily);

        //    RadFlowDocument importedDocument = TestHelper.GetDocumentFromFile(fileName, new DocxFormatProvider());

        //    StructureAsserter.AssertDocumentsAreEqual(expectedDocument, importedDocument);
        //}
    }
}
