﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using Telerik.Windows.Documents.Flow.FormatProviders.Txt;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Tests.Integration.ImportExport.Tables;
using Telerik.Windows.Documents.Flow.Tests.Utilities;

namespace Telerik.Windows.Documents.Flow.Tests.Integration
{
    [TestClass]
    public class TxtFormatProviderTests
    {
        private TxtFormatProvider txtProvider;

        [TestInitialize]
        public void Init()
        {
            this.txtProvider = new TxtFormatProvider();
        }

        [TestMethod]
        public void TxtFormatProvider_Export_EmptyDocument()
        {
            RadFlowDocument doc = new RadFlowDocument();

            Assert.AreEqual(string.Empty, this.txtProvider.Export(doc));
        }

        [TestMethod]
        public void TxtFormatProvider_Export_SingleParagraph()
        {
            RadFlowDocument doc = new RadFlowDocument();
            string testStr = "This is a test paragraph.";
            doc.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun(testStr);

            Assert.AreEqual(testStr, this.txtProvider.Export(doc));
        }

        [TestMethod]
        public void TxtFormatProvider_Export_TextWithFormatting()
        {
            RadFlowDocument doc = new RadFlowDocument();
            Paragraph p = doc.Sections.AddSection().Blocks.AddParagraph();
            p.Inlines.AddRun("First ");
            p.Inlines.AddRun("bold ").FontSize = 50;
            p.Inlines.AddRun("last.");

            Assert.AreEqual("First bold last.", this.txtProvider.Export(doc));
        }

        [TestMethod]
        public void TxtFormatProvider_Export_MultipleParagraphs()
        {
            string testStr1 = "This is a first paragraph.";
            string testStr2 = "This is a second paragraph.";

            RadFlowDocument doc = new RadFlowDocument();
            Section s = doc.Sections.AddSection();
            s.Blocks.AddParagraph().Inlines.AddRun(testStr1);
            s.Blocks.AddParagraph().Inlines.AddRun(testStr2);

            Assert.AreEqual(testStr1 + Environment.NewLine + testStr2, this.txtProvider.Export(doc));
        }

        [TestMethod]
        public void TxtFormatProvider_Export_ParagraphsAndTables()
        {
            string expected =
                "Begin" + Environment.NewLine +
                "Cell [0, 0]" + Environment.NewLine +
                "Cell [0, 1]" + Environment.NewLine +
                "Cell [1, 0]" + Environment.NewLine +
                "Cell [1, 1]" + Environment.NewLine +
                "End";
            RadFlowDocument doc = new RadFlowDocument();
            Section s = doc.Sections.AddSection();
            s.Blocks.AddParagraph().Inlines.AddRun("Begin");
            Table table = s.Blocks.AddTable();
            TableHelper.FillTable(table, 2, 2);
            s.Blocks.AddParagraph().Inlines.AddRun("End");

            Assert.AreEqual(expected, this.txtProvider.Export(doc));
        }


        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TxtFormatProvider_Import_NullString()
        {
            RadFlowDocument expected = new RadFlowDocument();
            expected.Sections.AddSection().Blocks.AddParagraph();

            RadFlowDocument imported = this.txtProvider.Import(null);
        }

        [TestMethod]
        public void TxtFormatProvider_Import_EmptyString()
        {
            RadFlowDocument expected = new RadFlowDocument();
            expected.Sections.AddSection().Blocks.AddParagraph();

            RadFlowDocument imported = this.txtProvider.Import(string.Empty);

            StructureAsserter.AssertDocumentsAreEqual(expected, imported);
        }

        [TestMethod]
        public void TxtFormatProvider_Import_SingleParagraphs()
        {
            RadFlowDocument expected = new RadFlowDocument();
            expected.Sections.AddSection()
                .Blocks.AddParagraph()
                .Inlines.AddRun("This is a test paragraph.");

            RadFlowDocument imported = this.txtProvider.Import("This is a test paragraph.");

            StructureAsserter.AssertDocumentsAreEqual(expected, imported);
        }

        [TestMethod]
        public void TxtFormatProvider_Import_MultipleParagraphs()
        {
            string testStr1 = "This is a first paragraph.";
            string testStr2 = "This is a second paragraph.";

            RadFlowDocument expected = new RadFlowDocument();
            Section s = expected.Sections.AddSection();
            s.Blocks.AddParagraph().Inlines.AddRun(testStr1);
            s.Blocks.AddParagraph().Inlines.AddRun(testStr2);

            RadFlowDocument imported = this.txtProvider.Import(testStr1 + Environment.NewLine + testStr2);

            StructureAsserter.AssertDocumentsAreEqual(expected, imported);
        }

        [TestMethod]
        public void TxtFormatProvider_Import_InconsistentLineEndings()
        {
            string testStr = "first" + "\r\n" + "second" + "\n" + "third" + "\r" + "last";

            RadFlowDocument expected = new RadFlowDocument();
            Section s = expected.Sections.AddSection();
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "first";
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "second";
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "third";
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "last";

            RadFlowDocument imported = this.txtProvider.Import(testStr);

            StructureAsserter.AssertDocumentsAreEqual(expected, imported);
        }
    }
}
