﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
using Telerik.Windows.Documents.Flow.Model.Protection;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Media;
#else
using System.Windows;
using System.Windows.Media;
using Style = Telerik.Windows.Documents.Flow.Model.Styles.Style;
#endif
#if WPF || SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Theming;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Flow.FormatProviders.Docx;

namespace Telerik.Windows.Documents.Flow.Tests.Model
{
    [TestClass]
    public class CloneDocumentTests
    {
        public const string TestFilesFolder = "Fields/";

        [TestMethod]
        public void Clone_WhenInvoked_CreatesNewDocument()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            // Act
            RadFlowDocument clonedDocument = document.Clone();

            // Assert
            Assert.AreNotSame(document, clonedDocument);
        }

        [TestMethod]
        public void Clone_WhenPropertiesAreSet_CloningHasSameProperties()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            document.HasDifferentEvenOddPageHeadersFooters = true;
            document.ViewType = DocumentViewType.WebLayout;

            // Act
            RadFlowDocument clonedDocument = document.Clone();

            // Assert
            StylePropertyAsserter.AssertPropertiesAreEqual(clonedDocument.Properties, clonedDocument.Properties);
        }

        [TestMethod]
        public void Clone_WhenHasChildSections_CloningHasSameStructure()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            document.Sections.AddSection();
            document.Sections.AddSection();
            document.Sections.AddSection();

            // Act
            RadFlowDocument clonedDocument = document.Clone();

            // Assert
            StructureAsserter.AssertDocumentsAreEqual(document, clonedDocument);
        }

        [TestMethod]
        public void Clone_WhenHasCharacterStyleAdded_CloningHasSameStyleRepository()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Style testStyle = new Style("Test", StyleType.Character);
            testStyle.CharacterProperties.FontSize.LocalValue = 50;
            document.StyleRepository.Add(testStyle);

            // Act
            RadFlowDocument clonedDocument = document.Clone();

            // Assert
            StylePropertyAsserter.AssertStylesAreEqual(document, clonedDocument);
        }

        [TestMethod]
        public void Clone_WhenHasParagraphStyleAdded_CloningHasSameStyleRepository()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Style testStyle = new Style("Test", StyleType.Paragraph);
            testStyle.ParagraphProperties.FirstLineIndent.LocalValue = 50;
            document.StyleRepository.Add(testStyle);

            // Act
            RadFlowDocument clonedDocument = document.Clone();

            // Assert
            StylePropertyAsserter.AssertStylesAreEqual(document, clonedDocument);
        }

        [TestMethod]
        public void Clone_WhenHasTableStyleAdded_CloningHasSameStyleRepository()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Style testStyle = new Style("Test", StyleType.Table);

            testStyle.TableProperties.Indent.LocalValue = 50;
            testStyle.TableRowProperties.Height.LocalValue = new TableRowHeight(HeightType.Exact, 50);
            testStyle.TableCellProperties.PreferredWidth.LocalValue = new TableWidthUnit(50);

            document.StyleRepository.Add(testStyle);

            // Act
            RadFlowDocument clonedDocument = document.Clone();

            // Assert
            StylePropertyAsserter.AssertStylesAreEqual(document, clonedDocument);
        }

        [TestMethod]
        public void Clone_WhenHasStyleWithPropertiesSetAdded_CloningHasSameStyleRepository()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Style testStyle = new Style("Test", StyleType.Character);
            testStyle.BasedOnStyleId = "basedOnId";
            testStyle.IsCustom = true;
            testStyle.IsDefault = true;
            testStyle.IsPrimary = true;
            testStyle.LinkedStyleId = "linkedStyleId";
            testStyle.Name = "name";
            testStyle.NextStyleId = "nextStyleId";
            testStyle.UIPriority = 5;

            document.StyleRepository.Add(testStyle);

            // Act
            RadFlowDocument clonedDocument = document.Clone();

            // Assert
            StylePropertyAsserter.AssertStylesAreEqual(document, clonedDocument);
        }

        [TestMethod]
        public void Clone_WhenHasTemplateListAdded_CloningHasSameListsCollection()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            document.Lists.Add(ListTemplateType.NumberedDefault);

            // Act
            RadFlowDocument clonedDocument = document.Clone();

            // Assert
            ListPropertyAsserter.AssertListsAreEqual(document, clonedDocument);
        }

        [TestMethod]
        public void Clone_WhenHasListWithRelatedStyleAdded_CloningHasSameListsCollection()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            List list = document.Lists.Add(ListTemplateType.NumberedDefault);

            Style style = new Style("style", StyleType.Numbering);
            document.StyleRepository.Add(style);

            list.StyleId = style.Id;

            // Act
            RadFlowDocument clonedDocument = document.Clone();

            // Assert
            ListPropertyAsserter.AssertListsAreEqual(document, clonedDocument);
        }

        [TestMethod]
        public void Clone_WhenHasListWithLevelRelatedToHeadingsAdded_CloningHasSameListsCollection()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1));
            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(2));
            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(3));
            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(4));
            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(5));
            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(6));
            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(7));
            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(8));
            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(9));

            List list = document.Lists.Add(ListTemplateType.BulletDefault);
            for (int i = 0; i < list.Levels.Count; i++)
            {
                list.Levels[i].StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1);
            }

            // Act
            RadFlowDocument clonedDocument = document.Clone();

            // Assert
            ListPropertyAsserter.AssertListsAreEqual(document, clonedDocument);
        }

        [TestMethod]
        public void Clone_WhenHasListRelatedToStyleAndWithLevelRelatedToHeadingsAdded_CloningHasSameListsCollection()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1));
            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(2));
            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(3));
            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(4));
            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(5));
            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(6));
            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(7));
            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(8));
            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(9));

            List list = document.Lists.Add(ListTemplateType.BulletDefault);
            for (int i = 0; i < list.Levels.Count; i++)
            {
                list.Levels[i].StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1);
            }

            Style style = new Style("style", StyleType.Numbering);
            document.StyleRepository.Add(style);

            list.StyleId = style.Id;

            // Act
            RadFlowDocument clonedDocument = document.Clone();

            // Assert
            ListPropertyAsserter.AssertListsAreEqual(document, clonedDocument);
        }

        [TestMethod]
        public void Clone_WithDocumentProtection()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            document.ProtectionSettings.ProtectionMode = Flow.Model.Protection.ProtectionMode.AllowComments;
            document.ProtectionSettings.AlgorithmName = ProtectionAlgorithmNames.SHA1;

            // Act
            RadFlowDocument clonedDocument = document.Clone();

            // Assert
            Assert.AreNotSame(document, clonedDocument);
        }

        [TestMethod]
        public void Clone_WhenDocumentHasDocumentVariables_CloningHasSameDocumentVariablesCollection()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            document.DocumentVariables.Add("test", "value");

            // Act
            RadFlowDocument clonedDocument = document.Clone();

            // Assert
            Assert.AreEqual(1, clonedDocument.DocumentVariables.Count);
            string docVarValue = clonedDocument.DocumentVariables["test"];
            Assert.AreEqual("value", docVarValue);
        }

        [TestMethod]
        public void Clone_DocumentDefaultStyleProperties_CloningHasSameDefaultStylePropeties()
        {
            RadFlowDocument document = new RadFlowDocument();

            document.DefaultStyle.CharacterProperties.FontSize.LocalValue = 30;
            document.DefaultStyle.CharacterProperties.FontWeight.LocalValue = FontWeights.Bold;
            document.DefaultStyle.CharacterProperties.FontStyle.LocalValue = FontStyles.Italic;

            document.DefaultStyle.ParagraphProperties.TextAlignment.LocalValue = Alignment.Right;
            document.DefaultStyle.ParagraphProperties.BackgroundColor.LocalValue = new ThemableColor(Colors.Blue);
            document.DefaultStyle.ParagraphProperties.AutomaticSpacingBefore.LocalValue = true;

            RadFlowDocument clonedDocument = document.Clone();

            StylePropertyAsserter.AssertDefaultStylePropertiesAreEqual(document.DefaultStyle, clonedDocument.DefaultStyle);
        }

        [TestMethod]
        public void Clone_DocumentDefaultStyleProperties_CloningHasSameInitialValues()
        {
            RadFlowDocument document = new RadFlowDocument();

            document.DefaultStyle.CharacterProperties.FontSize.ClearValue();

            document.DefaultStyle.ParagraphProperties.LineSpacing.ClearValue();

            RadFlowDocument clonedDocument = document.Clone();

            StylePropertyAsserter.AssertDefaultStylePropertiesAreEqual(document.DefaultStyle, clonedDocument.DefaultStyle);
        }

        [TestMethod]
        public void Clone_DocumentTheme_CloningHasSameTheme()
        {
            RadFlowDocument document = new RadFlowDocument();

            document.Theme = this.GetTestDocumentTheme();

            RadFlowDocument clonedDocument = document.Clone();

            Assert.AreEqual(document.Theme, clonedDocument.Theme);
        }

        [TestMethod]
        public void Clone_DocumentContainingFieldsWithoutSeparator_CloneDoesNotThrowNullReferenceException()
        {
            RadFlowDocument imported = TestHelper.GetDocumentFromFile(
                string.Format("{0}CodeFields/SimpleFieldWithoutSeparator.docx", TestFilesFolder), new DocxFormatProvider());

            RadFlowDocument cloned = null;

            try
            {
                cloned = imported.Clone();
            }
            catch (NullReferenceException)
            {
                Assert.Fail("Should not throw NullReferenceException!");
            }
        }

        private DocumentTheme GetTestDocumentTheme()
        {
            Color[] colors = new Color[]
            {
                Colors.Chocolate,
                Colors.Blue,
                Colors.Black,
                Colors.Purple,
                Colors.Plum,
                Colors.Magenta,
                Colors.Lime,
                Colors.LightSkyBlue,
                Colors.Gold,
                Colors.Gray,
                Colors.Gainsboro,
                Colors.Firebrick
            };

            string[] fontFamilies = new[]
            {
                "Cambria",
                "Segoe UI"
            };

            string themeName = "TestTheme";

            ThemeColorScheme colorScheme = new ThemeColorScheme(themeName,
                  colors[0],
                  colors[1],
                  colors[2],
                  colors[3],
                  colors[4],
                  colors[5],
                  colors[6],
                  colors[7],
                  colors[8],
                  colors[9],
                  colors[10],
                  colors[11]);

            ThemeFontScheme fontScheme = new ThemeFontScheme(themeName, fontFamilies[0], fontFamilies[1]);

            return new DocumentTheme(themeName, colorScheme, fontScheme);
        }
    }
}
