﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Linq;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Tests.Utilities;

namespace Telerik.Windows.Documents.Flow.Tests.Model
{
    [TestClass]
    public class DocumentElementImporterTests
    {
        [TestMethod]
        public void Import_WhenElementIsRun_ReturnsClonedRunAndRunCanBeAddedToTargetDocument()
        {
            // Arrange
            RadFlowDocument sourceDocument = new RadFlowDocument();
            Run run = sourceDocument.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("test");

            RadFlowDocument targetDocument = new RadFlowDocument();
            DocumentElementImporter importer = new DocumentElementImporter(targetDocument, sourceDocument, ConflictingStylesResolutionMode.UseTargetStyle);

            // Act
            Run importedRun = importer.Import<Run>(run);

            // Assert
            Assert.AreEqual(importedRun.Document, targetDocument);
        }

        [TestMethod]
        public void Import_WhenElementIsParagraphAndContainsPartOfAnnotations_ReturnsCorrectParagraphWithoutUnpairedAnnotations()
        {
            // Arrange
            RadFlowDocument sourceDocument = new RadFlowDocument();
            Section section = sourceDocument.Sections.AddSection();
            Paragraph p1 = section.Blocks.AddParagraph();
            p1.Inlines.AddRun("start");

            Bookmark bmk = new Bookmark(sourceDocument, "bmk");
            p1.Inlines.Add(bmk.BookmarkRangeStart);

            Paragraph p2 = section.Blocks.AddParagraph();
            p2.Inlines.AddRun("end");
            p2.Inlines.Add(bmk.BookmarkRangeEnd);

            DocumentElementImporter importer = new DocumentElementImporter(new RadFlowDocument(), sourceDocument, ConflictingStylesResolutionMode.UseTargetStyle);

            // Act
            Paragraph importedParagraph = importer.Import<Paragraph>(p1);

            // Assert
            Assert.AreEqual(1, importedParagraph.Inlines.Count);
            Assert.AreEqual("start", (importedParagraph.Inlines[0] as Run).Text);
        }

        [TestMethod]
        public void Import_WhenElementIsSectionAndContainsAnnotationsSpaningTwoParagraphs_ReturnsCorrectSectionAndParagraphsContainTheirAnnotations()
        {
            // Arrange
            RadFlowDocument sourceDocument = new RadFlowDocument();
            Section section = sourceDocument.Sections.AddSection();
            Paragraph p1 = section.Blocks.AddParagraph();
            p1.Inlines.AddRun("start");

            Bookmark bmk = new Bookmark(sourceDocument, "bmk");
            p1.Inlines.Add(bmk.BookmarkRangeStart);

            Paragraph p2 = section.Blocks.AddParagraph();
            p2.Inlines.AddRun("end");
            p2.Inlines.Add(bmk.BookmarkRangeEnd);

            DocumentElementImporter importer = new DocumentElementImporter(new RadFlowDocument(), sourceDocument, ConflictingStylesResolutionMode.UseTargetStyle);

            // Act
            Section importedSection = importer.Import<Section>(section);

            // Assert
            System.Collections.Generic.List<Paragraph> paragraphs = importedSection.EnumerateChildrenOfType<Paragraph>().ToList();
            Assert.AreEqual(2, paragraphs.Count);
            Assert.AreEqual(1, importedSection.EnumerateChildrenOfType<BookmarkRangeStart>().Count());
            Assert.AreEqual(1, importedSection.EnumerateChildrenOfType<BookmarkRangeEnd>().Count());

            BookmarkRangeStart importedBmkStart = paragraphs[0].EnumerateChildrenOfType<BookmarkRangeStart>().FirstOrDefault();
            Assert.IsNotNull(importedBmkStart);
            Assert.AreEqual("bmk", importedBmkStart.Bookmark.Name);
            Assert.AreEqual(importedBmkStart.Bookmark.BookmarkRangeEnd, paragraphs[1].EnumerateChildrenOfType<BookmarkRangeEnd>().FirstOrDefault());
        }

        [TestMethod]
        public void Import_WhenTargetAndSourceDocumentsContainSameStylesAndParagraphElementIsWithSuchStyle_ReturnsParagraphWithRenamedStyle()
        {
            // Arrange
            string styleName = "MyStyle";

            RadFlowDocument sourceDocument = new RadFlowDocument();
            Section section = sourceDocument.Sections.AddSection();
            Paragraph p1 = section.Blocks.AddParagraph();
            p1.Inlines.AddRun("start");

            Style sourceStyle = new Style(styleName, StyleType.Paragraph);
            sourceStyle.CharacterProperties.FontSize.LocalValue = 30;
            sourceDocument.StyleRepository.Add(sourceStyle);

            p1.StyleId = sourceStyle.Id;

            RadFlowDocument targetDocument = new RadFlowDocument();
            Style targetStyle = new Style(styleName, StyleType.Paragraph);
            targetDocument.StyleRepository.Add(targetStyle);

            DocumentElementImporter importer = new DocumentElementImporter(targetDocument, sourceDocument, ConflictingStylesResolutionMode.RenameSourceStyle);

            // Act
            Paragraph importedParagraph = importer.Import<Paragraph>(p1);

            // Assert
            Style renamedStyle = targetDocument.StyleRepository.GetStyle(styleName + "_1");
            Assert.IsNotNull(renamedStyle);
            StylePropertyAsserter.AssertPropertiesAreEqual(sourceStyle.CharacterProperties, renamedStyle.CharacterProperties);
            Assert.AreEqual(renamedStyle.Id, importedParagraph.StyleId);
        }

        [TestMethod]
        public void Import_WhenElementIsNotChildOfSourceDocument_ThrowsInvalidOperationException()
        {
            // Arrange
            DocumentElementImporter importer = new DocumentElementImporter(new RadFlowDocument(), new RadFlowDocument(), ConflictingStylesResolutionMode.UseTargetStyle);

            // Act && Assert
            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
            {
                importer.Import<Run>(new Run(new RadFlowDocument()));
            });
        }

        [TestMethod]
        public void Import_WhenImportMethodIsInvokedMultipleConsecutiveTimesAndElementIsParagraphWithStyleFromSourceDocument_ReturnsParagraphWithCorrectlyOneTimeRenamedStyle()
        {
            // Arrange
            string styleName = "MyStyle";

            RadFlowDocument sourceDocument = new RadFlowDocument();
            Section section = sourceDocument.Sections.AddSection();
            Paragraph p1 = section.Blocks.AddParagraph();
            p1.Inlines.AddRun("start");

            Style sourceStyle = new Style(styleName, StyleType.Paragraph);
            sourceStyle.CharacterProperties.FontSize.LocalValue = 30;
            sourceDocument.StyleRepository.Add(sourceStyle);

            p1.StyleId = sourceStyle.Id;

            RadFlowDocument targetDocument = new RadFlowDocument();
            Style targetStyle = new Style(styleName, StyleType.Paragraph);
            targetDocument.StyleRepository.Add(targetStyle);

            DocumentElementImporter importer = new DocumentElementImporter(targetDocument, sourceDocument, ConflictingStylesResolutionMode.RenameSourceStyle);

            // Act
            Paragraph importedParagraph1 = importer.Import<Paragraph>(p1);
            Paragraph importedParagraph2 = importer.Import<Paragraph>(p1);

            // Assert
            Style renamedStyle = targetDocument.StyleRepository.GetStyle(styleName + "_1");
            Assert.IsNotNull(renamedStyle);
            StylePropertyAsserter.AssertPropertiesAreEqual(sourceStyle.CharacterProperties, renamedStyle.CharacterProperties);

            Assert.AreEqual(renamedStyle.Id, importedParagraph1.StyleId);
            Assert.AreEqual(renamedStyle.Id, importedParagraph2.StyleId);
        }

        [TestMethod]
        public void Import_WhenElementIsNull_ThrowsArgumentNullException()
        {
            // Arrange
            DocumentElementImporter importer = new DocumentElementImporter(new RadFlowDocument(), new RadFlowDocument(), ConflictingStylesResolutionMode.UseTargetStyle);

            // Act && Assert
            TestHelper.AssertExceptionIsThrown<ArgumentNullException>(() =>
            {
                importer.Import<Run>(null);
            });
        }
    }
}