﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Tests.Utilities;

namespace Telerik.Windows.Documents.Flow.Tests.Model.Editing
{
    [TestClass]
    public class RadFlowDocumentEditorBookmarksTests
    {
        private RadFlowDocumentEditor editor;

        [TestInitialize]
        public void TestInitialize()
        {
            RadFlowDocument document = new RadFlowDocument();
            this.editor = new RadFlowDocumentEditor(document);
        }

        [TestCleanup]
        public void TestCleanup()
        {
            this.editor = null;
        }

        [TestMethod]
        public void InsertBookmark_InEmptyDocument_WhenOnlyNameIsGiven_IsInserted()
        {
            // Arrange

            // Act
            string bookmarkName = "singleBookmark";
            this.editor.InsertBookmark(bookmarkName);

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            Bookmark bmk = new Bookmark(expected, bookmarkName);
            paragraph.Inlines.Add(bmk.BookmarkRangeStart);
            paragraph.Inlines.Add(bmk.BookmarkRangeEnd);

            StructureAsserter.AssertDocumentsAreEqual(expected, this.editor.Document);
        }

        [TestMethod]
        public void InsertBookmark_BetweenInlines_WhenOnlyNameIsGiven_IsInserted()
        {
            // Arrange
            this.editor.InsertText("run1");
            Run run2 = this.editor.InsertText("run2");
            this.editor.MoveToInlineStart(run2);

            // Act
            string bookmarkName = "singleBookmark";
            this.editor.InsertBookmark(bookmarkName);

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("run1");
            Bookmark bmk = new Bookmark(expected, bookmarkName);
            paragraph.Inlines.Add(bmk.BookmarkRangeStart);
            paragraph.Inlines.Add(bmk.BookmarkRangeEnd);
            paragraph.Inlines.AddRun("run2");

            StructureAsserter.AssertDocumentsAreEqual(expected, this.editor.Document);
        }

        [TestMethod]
        public void InsertBookmark_TwiceWithSameName_WhenOnlyNameIsGiven_ThrowsInvalidOperationException()
        {
            // Arrange

            // Act & Assert
            string bookmarkName = "singleBookmark";
            this.editor.InsertBookmark(bookmarkName);
            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
            {
                this.editor.InsertBookmark(bookmarkName);
            });
        }

        [TestMethod]
        public void InsertBookmark_WithInlineStartFormAnotherDocument_ThrowsInvalidOperationException()
        {
            // Arrange
            Run run = this.editor.InsertText("our run");
            RadFlowDocumentEditor anotherEditor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run runFromAnotherDocument = anotherEditor.InsertText("run");

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
            {
                this.editor.InsertBookmark("test", runFromAnotherDocument, run);
            });
        }

        [TestMethod]
        public void InsertBookmark_WithInlineEndFormAnotherDocument_ThrowsInvalidOperationException()
        {
            // Arrange
            Run run = this.editor.InsertText("our run");
            RadFlowDocumentEditor anotherEditor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run runFromAnotherDocument = anotherEditor.InsertText("run");

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
            {
                this.editor.InsertBookmark("test", run, runFromAnotherDocument);
            });
        }

        [TestMethod]
        public void InsertBookmark_WithNullInlineStart_ThrowsArgumentNullException()
        {
            // Arrange
            Run run = this.editor.InsertText("run");

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentNullException>(() =>
            {
                this.editor.InsertBookmark("test", null, run);
            });
        }

        [TestMethod]
        public void InsertBookmark_WithNullInlineEnd_ThrowsArgumentNullException()
        {
            // Arrange
            Run run = this.editor.InsertText("run");

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentNullException>(() =>
            {
                this.editor.InsertBookmark("test", run, null);
            });
        }

        [TestMethod]
        public void InsertBookmark_WhenBookmarkWithSameNameExistsInHeaderFooter_ThrowsInvalidOperationException()
        {
            // Arrange
            Header header = this.editor.Document.Sections.AddSection().Headers.Add();
            Paragraph paragraph = header.Blocks.AddParagraph();
            Bookmark bmk = new Bookmark(this.editor.Document, "bmk");
            paragraph.Inlines.Add(bmk.BookmarkRangeStart);
            paragraph.Inlines.Add(bmk.BookmarkRangeEnd);

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
            {
                this.editor.InsertBookmark("bmk");
            });
        }

        [TestMethod]
        public void InsertBookmark_ForSingleInline_IsInserted()
        {
            // Arrange

            // Act
            string bookmarkName = "singleBookmark";
            Run run = this.editor.InsertText("run");
            this.editor.InsertBookmark(bookmarkName, run, run);

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            Bookmark bmk = new Bookmark(expected, bookmarkName);
            paragraph.Inlines.Add(bmk.BookmarkRangeStart);
            paragraph.Inlines.AddRun("run");
            paragraph.Inlines.Add(bmk.BookmarkRangeEnd);

            StructureAsserter.AssertDocumentsAreEqual(expected, this.editor.Document);
        }

        [TestMethod]
        public void InsertBookmark_InMultipleParagraphs_IsInserted()
        {
            // Arrange
            this.editor.InsertText("first");
            Run start = this.editor.InsertText("second");
            this.editor.InsertParagraph();
            Run end = this.editor.InsertText("third");
            this.editor.InsertText("fourth");

            // Act
            string bookmarkName = "singleBookmark";
            this.editor.InsertBookmark(bookmarkName, start, end);

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Bookmark bmk = new Bookmark(expected, bookmarkName);

            Section section = expected.Sections.AddSection();
            Paragraph paragraph = section.Blocks.AddParagraph();
            paragraph.Inlines.AddRun("first");
            paragraph.Inlines.Add(bmk.BookmarkRangeStart);
            paragraph.Inlines.AddRun("second");

            paragraph = section.Blocks.AddParagraph();
            paragraph.Inlines.AddRun("third");
            paragraph.Inlines.Add(bmk.BookmarkRangeEnd);
            paragraph.Inlines.AddRun("fourth");

            StructureAsserter.AssertDocumentsAreEqual(expected, this.editor.Document);
        }

        [TestMethod]
        public void DeleteBookmark_ByName_DeletesBookmark()
        {
            // Arrange
            string bookmarkName = "singleBookmark";
            this.editor.InsertBookmark(bookmarkName);

            // Act
            this.editor.DeleteBookmark(bookmarkName);

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            expected.Sections.AddSection().Blocks.AddParagraph();
            StructureAsserter.AssertDocumentsAreEqual(expected, this.editor.Document);
        }

        [TestMethod]
        public void DeleteBookmark_ByObject_DeletesBookmark()
        {
            // Arrange
            string bookmarkName = "singleBookmark";
            Bookmark bmk = this.editor.InsertBookmark(bookmarkName);

            // Act
            this.editor.DeleteBookmark(bmk);

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            expected.Sections.AddSection().Blocks.AddParagraph();
            StructureAsserter.AssertDocumentsAreEqual(expected, this.editor.Document);
        }

        [TestMethod]
        public void DeleteBookmark_ByName_ForNonExistingBoookmark_ThrowsException()
        {
            // Arrange
            this.editor.InsertText("TestDocument");

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
                this.editor.DeleteBookmark("test"));
        }

        [TestMethod]
        public void DeleteBookmark_ByObject_ForNotInsertedBoookmark_ThrowsException()
        {
            // Arrange
            this.editor.InsertText("TestDocument");
            Bookmark bmk = new Bookmark(this.editor.Document, "test");


            // Act & Assert
            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
                this.editor.DeleteBookmark(bmk));
        }

        [TestMethod]
        public void DeleteBookmark_ByObject_ForDeletedBoookmark_ThrowsException()
        {
            // Arrange
            this.editor.InsertText("TestDocument");
            Bookmark bmk = this.editor.InsertBookmark("test");
            this.editor.DeleteBookmark(bmk);

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
                this.editor.DeleteBookmark(bmk));
        }

        [TestMethod]
        public void DeleteBookmark_ByObject_ForBookmarkFromAnotherDocument_ThrowsException()
        {
            // Arrange
            this.editor.InsertText("TestDocument");
            Bookmark bmk = new Bookmark(new RadFlowDocument(), "test");

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
                this.editor.DeleteBookmark(bmk));
        }
    }
}