﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Linq;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;

namespace Telerik.Windows.Documents.Flow.Tests.Model.Editing
{
    [TestClass]
    public class RadFlowDocumentEditorBreaksTests
    {
        [TestMethod]
        public void PageBreak_WhenIsInserted_ParagraphIsInsertedAfterIt()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            Break insertedBreak = editor.InsertBreak(BreakType.PageBreak);

            // Assert
            System.Collections.Generic.IEnumerable<Paragraph> paragraphs = editor.Document.EnumerateChildrenOfType<Paragraph>();

            Assert.AreEqual(2, paragraphs.Count());
            Assert.IsTrue(paragraphs.First().Inlines.Contains(insertedBreak));
            Assert.AreEqual(0, paragraphs.Last().Inlines.Count);
        }

        [TestMethod]
        public void PageBreak_WhenIsInsertedAfterText_TextAndBreakAreInSameParagraph()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("test");

            // Act
            editor.InsertBreak(BreakType.PageBreak);

            // Assert
            System.Collections.Generic.IEnumerable<Paragraph> paragraphs = editor.Document.EnumerateChildrenOfType<Paragraph>();

            Assert.AreEqual(2, paragraphs.Count());
            Assert.IsTrue(paragraphs.First().Inlines.Any(r => r is Run));
            Assert.IsTrue(paragraphs.First().Inlines.Any(br => br is Break));
            Assert.AreEqual(0, paragraphs.Last().Inlines.Count);
        }

        [TestMethod]
        public void LineBreak_WhenIsInsertedAfterText_AndTextIsInsertedAfterIt_AllInlinesAreInSameParagraph()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("test");

            // Act
            editor.InsertBreak(BreakType.LineBreak);
            editor.InsertText("teext 2");

            // Assert
            System.Collections.Generic.IEnumerable<Paragraph> paragraphs = editor.Document.EnumerateChildrenOfType<Paragraph>();

            Assert.AreEqual(1, paragraphs.Count());
            Assert.IsTrue(paragraphs.First().Inlines.Any(br => br is Break));
        }

        [TestMethod]
        public void PageBreak_WhenIsInserted_TextWrappingRestartLocationProperty_IsNextLine()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            editor.InsertBreak(BreakType.PageBreak);

            // Assert
            Break br = editor.Document.EnumerateChildrenOfType<Break>().First();

            Assert.AreEqual(BreakType.PageBreak, br.BreakType);
            Assert.AreEqual(TextWrappingRestartLocation.NextLine, br.TextWrappingRestartLocation);
        }

        [TestMethod]
        public void ColumnBreak_WhenIsInserted_TextWrappingRestartLocationProperty_IsNextLine()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            editor.InsertBreak(BreakType.ColumnBreak);

            // Assert
            Break br = editor.Document.EnumerateChildrenOfType<Break>().First();

            Assert.AreEqual(BreakType.ColumnBreak, br.BreakType);
            Assert.AreEqual(TextWrappingRestartLocation.NextLine, br.TextWrappingRestartLocation);
        }

        [TestMethod]
        public void LineBreak_WhenIsInserted_TextWrappingRestartLocationProperty_IsNextLine()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            editor.InsertBreak(BreakType.LineBreak);

            // Assert
            Break br = editor.Document.EnumerateChildrenOfType<Break>().First();

            Assert.AreEqual(BreakType.LineBreak, br.BreakType);
            Assert.AreEqual(TextWrappingRestartLocation.NextLine, br.TextWrappingRestartLocation);
        }

        [TestMethod]
        public void LineBreak_WhenIsInTable_AndTextIsInsertedAfterIt_AllInlinesAreInSameParagraph()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Table table = editor.InsertTable(3, 3);
            TableCell cell1 = table.Rows.First().Cells.First();
            Paragraph cell1Paragraph = cell1.Blocks.AddParagraph();
            editor.MoveToParagraphStart(cell1Paragraph);
            editor.InsertText("test");

            // Act
            editor.InsertBreak(BreakType.LineBreak);
            editor.InsertText("teext 2");

            // Assert
            Assert.AreEqual(3, cell1Paragraph.Inlines.Count);
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Table>().Count()); // We Assert the table is not splitted.
        }
    }
}