﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Linq;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Tests.Utilities;

namespace Telerik.Windows.Documents.Flow.Tests.Model.Editing
{
    [TestClass]
    public class RadFlowDocumentEditorCommentsTests
    {
        private RadFlowDocumentEditor editor;

        [TestInitialize]
        public void TestInitialize()
        {
            this.editor = new RadFlowDocumentEditor(new RadFlowDocument());
        }

        [TestCleanup]
        public void TestCleanup()
        {
            this.editor = null;
        }

        [TestMethod]
        public void InsertComment_WithNullInlineStart_ThrowsException()
        {
            // Arrange
            Run run = this.editor.InsertText("run");

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentNullException>(() =>
            {
                this.editor.InsertComment("text", null, run);
            });
        }

        [TestMethod]
        public void InsertComment_WithNullInlineEnd_ThrowsException()
        {
            // Arrange
            Run run = this.editor.InsertText("run");

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentNullException>(() =>
            {
                this.editor.InsertComment("text", run, null);
            });
        }

        [TestMethod]
        public void InsertComment_WithInlinesFromOtherDocument_ThrowsException()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Run run = document.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun();

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
            {
                this.editor.InsertComment("text", run, run);
            });
        }

        [TestMethod]
        public void Comment_InEmptyDocument_WhenOnlyAuthorAndBodyAreGiven_IsInserted()
        {
            // Act
            string text = "text";
            this.editor.InsertComment(text);

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            Comment comment = expected.Comments.AddComment();
            comment.Blocks.AddParagraph().Inlines.AddRun(text);
            paragraph.Inlines.Add(comment.CommentRangeStart);
            paragraph.Inlines.Add(comment.CommentRangeEnd);

            Assert.IsTrue(this.editor.Document.EnumerateChildrenOfType<CommentRangeStart>().Any());
            Assert.IsTrue(this.editor.Document.EnumerateChildrenOfType<CommentRangeEnd>().Any());

            StructureAsserter.AssertDocumentsAreEqual(expected, this.editor.Document);
        }

        [TestMethod]
        public void InsertComment_WhenIsAroundInline_CommentStartAndEndAreInserted()
        {
            // Arrange
            string runText = "single inline";
            Run run = this.editor.InsertText(runText);

            // Act
            string commentText = "John";
            this.editor.InsertComment(commentText, run, run);

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph paragraph = expected.Sections.AddSection().Blocks.AddParagraph();
            Comment comment = expected.Comments.AddComment();
            comment.Blocks.AddParagraph().Inlines.AddRun(commentText);
            paragraph.Inlines.Add(comment.CommentRangeStart);
            paragraph.Inlines.AddRun(runText);
            paragraph.Inlines.Add(comment.CommentRangeEnd);

            StructureAsserter.AssertDocumentsAreEqual(expected, this.editor.Document);
        }
    }
}