﻿using System;
using System.Collections.Generic;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Model.Fields;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;

namespace Telerik.Windows.Documents.Flow.Tests.Model.Editing
{
    [TestClass]
    public class RadFlowDocumentEditorStructureTests
    {
        [TestMethod]
        public void InsertText_InEmptyDocument()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            Run run = editor.InsertText("test");

            Assert.IsNotNull(run);

            RadFlowDocument expected = new RadFlowDocument();
            expected.Sections.AddSection()
                .Blocks.AddParagraph()
                .Inlines.AddRun().Text = "test";

            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertText_Twice_InEmptyDocument()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("test");
            editor.InsertText("1234");

            RadFlowDocument expected = new RadFlowDocument();
            Paragraph par = expected.Sections.AddSection()
                .Blocks.AddParagraph();
            par.Inlines.AddRun().Text = "test";
            par.Inlines.AddRun().Text = "1234";

            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertLine_InsertsParagraphAfterTheText()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertLine("1234");

            RadFlowDocument expected = new RadFlowDocument();
            Section s = expected.Sections.AddSection();
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "1234";
            s.Blocks.AddParagraph();

            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertNewParagraph_InEmptyDocument()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertParagraph();
            editor.InsertText("1234");

            RadFlowDocument expected = new RadFlowDocument();
            Section s = expected.Sections.AddSection();
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "1234";

            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertParagraph_InDocumentWithEmptySection_InsertOnlyOneParagraph()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            document.Sections.AddSection();

            // Act
            editor.InsertParagraph();

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Paragraph>().Count());
        }

        [TestMethod]
        public void InsertParagraph_InEmptyDocument_InsertOnlyOneParagraph()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            editor.InsertParagraph();

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Paragraph>().Count());
        }

        [TestMethod]
        public void AddTwoParagraphs_InEmptyDocument()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("test");
            editor.InsertParagraph();
            editor.InsertText("1234");

            RadFlowDocument expected = new RadFlowDocument();
            Section s = expected.Sections.AddSection();
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "test";
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "1234";

            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertParagraph_InTheMiddleOfParagraph_ShouldSplitTheParagraph()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run firstRun = editor.InsertText("test");
            editor.InsertText("1234");
            editor.MoveToInlineEnd(firstRun);

            // Act
            editor.InsertParagraph();

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Section s = expected.Sections.AddSection();
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "test";
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "1234";
            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertTable_InTheMiddleOfParagraph_ShouldSplitTheParagraph()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run firstRun = editor.InsertText("before");
            editor.InsertText("after");

            // Act
            editor.MoveToInlineEnd(firstRun);
            editor.InsertTable(3, 4);

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Section s = expected.Sections.AddSection();
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "before";
            s.Blocks.Add(new Table(expected.Document, 3, 4));
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "after";
            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertTable_InTheEndOfParagraph_ShouldNotSplitTheParagraph()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("before");

            // Act
            editor.InsertTable(3, 4);

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Section s = expected.Sections.AddSection();
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "before";
            s.Blocks.Add(new Table(expected.Document, 3, 4));
            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertTwoTable_ShouldNotAddParagraphsBetweenThem()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("before");

            // Act
            editor.InsertTable(3, 4);
            editor.InsertTable(2, 2);

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Section s = expected.Sections.AddSection();
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "before";
            s.Blocks.Add(new Table(expected.Document, 3, 4));
            s.Blocks.Add(new Table(expected.Document, 2, 2));
            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertParagraph_WhenOnTable_ShouldInsertEmptyParagraphAfterTheTable()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("before");
            editor.InsertTable(3, 4);

            //Act 
            editor.InsertParagraph();

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Section s = expected.Sections.AddSection();
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "before";
            s.Blocks.Add(new Table(expected.Document, 3, 4));
            s.Blocks.AddParagraph();
            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertText_WhenOnTable_AndThereIsNoParagraphAfter_ShouldCreateNewParagraph()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("before");
            editor.InsertTable(3, 4);

            //Act 
            editor.InsertText("after");

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Section s = expected.Sections.AddSection();
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "before";
            s.Blocks.Add(new Table(expected.Document, 3, 4));
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "after";
            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertText_WhenOnTable_AndThereIsParagraphAfter_ShouldInsertInTheParagraph()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("before");
            Table table = editor.InsertTable(3, 4);
            editor.InsertText("after");
            editor.MoveToTableEnd(table);

            //Act 
            editor.InsertText("NEW");

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Section s = expected.Sections.AddSection();
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "before";
            s.Blocks.Add(new Table(expected.Document, 3, 4));
            Paragraph par = s.Blocks.AddParagraph();
            par.Inlines.AddRun().Text = "NEW";
            par.Inlines.AddRun().Text = "after";
            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertParagraph_WhenPostionIsBetweenTables_ShouldInsertParagraphBetweenTables()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Table table = editor.InsertTable(3, 4);
            editor.InsertTable(2, 2);
            editor.MoveToTableEnd(table);

            // Act
            editor.InsertParagraph();

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Section s = expected.Sections.AddSection();
            s.Blocks.Add(new Table(expected.Document, 3, 4));
            s.Blocks.AddParagraph();
            s.Blocks.Add(new Table(expected.Document, 2, 2));
            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertText_WhenPostionIsBetweenTables_ShouldInsertTextBetweenTables()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Table table = editor.InsertTable(3, 4);
            editor.InsertTable(2, 2);
            editor.MoveToTableEnd(table);

            // Act
            editor.InsertText("test");

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Section s = expected.Sections.AddSection();
            s.Blocks.Add(new Table(expected.Document, 3, 4));
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "test";
            s.Blocks.Add(new Table(expected.Document, 2, 2));
            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertTable_InEmptyDocument_InsertsTableAtTheStartOfTheDocument()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            editor.InsertTable(3, 4);

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Section s = expected.Sections.AddSection();
            s.Blocks.Add(new Table(expected.Document, 3, 4));
            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertText_WithNewLines_ShouldInsertMultipleParagraphs()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            string test = "first" + "\r\n" + "second" + "\n" + "third" + "\r" + "last";
            editor.InsertText(test);

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Section s = expected.Sections.AddSection();
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "first";
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "second";
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "third";
            s.Blocks.AddParagraph().Inlines.AddRun().Text = "last";
            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertHyperlink_InsertsHyperlinkField()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            editor.InsertHyperlink("telerik", "telerik.com", false, "Telerik Site");

            // Assert
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph par = expected.Sections.AddSection().Blocks.AddParagraph();
            FieldInfo field = new FieldInfo(expected);

            par.Inlines.Add(field.Start);
            par.Inlines.AddRun().Text = " HYPERLINK \"telerik.com\" \\o \"Telerik Site\" ";
            par.Inlines.Add(field.Separator);
            par.Inlines.AddRun().Text = "telerik";
            par.Inlines.Add(field.End);
            field.UpdateField();

            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertImage_WithImageSource_InsertsImage()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            ImageInline result = editor.InsertImageInline(new ImageSource(TestHelper.GetImageStream("test.png"), "png"), new Size(60, 20));

            // Assert
            Assert.IsNotNull(result);
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph p = expected.Sections.AddSection().Blocks.AddParagraph();
            ImageInline image = TestHelper.CreatePngImage(expected);
            image.Image.Size = new Size(60, 20);
            p.Inlines.Add(image);

            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertImage_WithStream_InsertsImage()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            ImageInline result = editor.InsertImageInline(TestHelper.GetImageStream("test.png"), "png", new Size(60, 20));

            // Assert
            Assert.IsNotNull(result);
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph p = expected.Sections.AddSection().Blocks.AddParagraph();
            ImageInline image = TestHelper.CreatePngImage(expected);
            image.Image.Size = new Size(60, 20);
            p.Inlines.Add(image);

            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertInline_InsertsTheInline()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            Run run = new Run(editor.Document) { Text = "test" };
            InlineBase result = editor.InsertInline(run);

            // Assert
            Assert.AreSame(run, result);
            RadFlowDocument expected = new RadFlowDocument();
            expected.Sections.AddSection()
                .Blocks.AddParagraph()
                .Inlines.AddRun("test");

            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertFloatingImage_WithImageSource_InsertsFloatingImage()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            FloatingImage result = editor.InsertFloatingImage(new ImageSource(TestHelper.GetImageStream("test.png"), "png"), new Size(60, 20));

            // Assert
            Assert.IsNotNull(result);
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph p = expected.Sections.AddSection().Blocks.AddParagraph();
            FloatingImage image = TestHelper.CreatePngFloatingImage(expected);
            image.Image.Size = new Size(60, 20);
            p.Inlines.Add(image);

            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertFloatingImage_WithStream_InsertsFloatingImage()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            FloatingImage result = editor.InsertFloatingImage(TestHelper.GetImageStream("test.png"), "png", new Size(60, 20));

            // Assert
            Assert.IsNotNull(result);
            RadFlowDocument expected = new RadFlowDocument();
            Paragraph p = expected.Sections.AddSection().Blocks.AddParagraph();
            FloatingImage image = TestHelper.CreatePngFloatingImage(expected);
            image.Image.Size = new Size(60, 20);
            p.Inlines.Add(image);

            StructureAsserter.AssertDocumentsAreEqual(expected, editor.Document);
        }

        [TestMethod]
        public void InsertInline_WithNull_Fails()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentNullException>(() => editor.InsertInline(null));
        }

        [TestMethod]
        public void InsertFloatingImage_WithNullImageSource_Fails()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentNullException>(() => editor.InsertFloatingImage(null, new Size()));
        }

        [TestMethod]
        public void InsertFloatingImage_WithNullStream_Fails()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentNullException>(() => editor.InsertFloatingImage(null, "png", new Size()));
        }

        [TestMethod]
        public void InsertImage_WithNullImageSource_Fails()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentNullException>(() => editor.InsertImageInline(null, new Size()));
        }

        [TestMethod]
        public void InsertImage_WithNullStream_Fails()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentNullException>(() => editor.InsertImageInline(null, "png", new Size()));
        }

        [TestMethod]
        public void InsertSection_WhenDocumentIsEmpty_InsertOnlySection()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            editor.InsertSection();

            // Assert
            Assert.AreEqual(1, editor.Document.Sections.Count);
            Assert.IsFalse(editor.Document.EnumerateChildrenOfType<Paragraph>().Any());
        }

        [TestMethod]
        public void InsertSection_WhenDocumentHasTextContent_InsertsSectionAndParagraph()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("test");

            // Act
            editor.InsertSection();

            // Assert
            Assert.AreEqual(2, editor.Document.Sections.Count);
            Assert.AreEqual(2, editor.Document.EnumerateChildrenOfType<Paragraph>().Count());
        }

        [TestMethod]
        public void InsertText_WhenDocumentHasTextAndSectionIsInserted_TextIsInsertedInAddedSection()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run oldSectionRun = editor.InsertText("test");
            Section oldSection = (Section)oldSectionRun.Paragraph.BlockContainer;
            Section newSection = editor.InsertSection();

            // Act
            Run newSectionRun = editor.InsertText("new section");

            //Assert
            Assert.AreEqual(newSection, newSectionRun.Paragraph.BlockContainer);
            Assert.AreNotEqual(oldSection, newSection);
        }

        [TestMethod]
        public void InsertSection_WhenIsInTableWith3RowsAnd3Columns_AndCurrentParagraphIsInTableCellFromFirstRow_TableIsMovedToNewSection()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run oldSectionRun = editor.InsertText("test");
            Section oldSection = (Section)oldSectionRun.Paragraph.BlockContainer;
            editor.InsertParagraph();

            Table table = editor.InsertTable(3, 3);
            TableCell cell1 = table.Rows.First().Cells.First();
            Paragraph cell1Paragraph = cell1.Blocks.AddParagraph();
            editor.MoveToParagraphStart(cell1Paragraph);
            editor.InsertText("cell");

            // Act
            editor.MoveToParagraphStart(cell1Paragraph);
            Section newSection = editor.InsertSection();

            // Assert
            Assert.AreEqual(newSection, table.BlockContainer);
            Assert.IsTrue(newSection.EnumerateChildrenOfType<Table>().Any());
            Assert.IsFalse(oldSection.EnumerateChildrenOfType<Table>().Any());
        }

        [TestMethod]
        public void InsertSection_WhenIsInTableWith3RowsAnd3Columns_AndCurrentParagraphIsInTableCellFromMiddleRow_TableIsSplitted()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run oldSectionRun = editor.InsertText("test");
            Section oldSection = (Section)oldSectionRun.Paragraph.BlockContainer;
            editor.InsertParagraph();

            Table table = editor.InsertTable(3, 3);
            TableCell cell1 = table.Rows.First().Cells.First();
            Paragraph cell1Paragraph = cell1.Blocks.AddParagraph();
            editor.MoveToParagraphStart(cell1Paragraph);
            Run cell1run = editor.InsertText("cell1");

            TableCell cell2 = table.Rows[1].Cells.First();
            Paragraph cell2Paragraph = cell2.Blocks.AddParagraph();
            editor.MoveToParagraphStart(cell2Paragraph);
            Run cell2run = editor.InsertText("cell2");

            // Act
            editor.MoveToParagraphStart(cell2Paragraph);
            Section newSection = editor.InsertSection();

            Table aboveTable = oldSection.EnumerateChildrenOfType<Table>().FirstOrDefault();
            Table belowTable = newSection.EnumerateChildrenOfType<Table>().FirstOrDefault();

            // Assert
            Assert.AreEqual(1, aboveTable.GridRowsCount);
            Assert.AreEqual(2, belowTable.GridRowsCount);

            Run splittedCellAboveRun = ((Paragraph)aboveTable.Rows.First().Cells.First().Blocks.First()).Inlines.First() as Run;
            Run splittedCellBelowRun = ((Paragraph)belowTable.Rows.First().Cells.First().Blocks.First()).Inlines.First() as Run;
            Assert.AreEqual(cell1run.Text, splittedCellAboveRun.Text);
            Assert.AreEqual(cell2run.Text, splittedCellBelowRun.Text);
        }

        [TestMethod]
        public void InsertSection_WhenIsInTableWith3RowsAnd3Columns_AndCurrentParagraphIsInTableCellFromLastRow_TableIsSplitted()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run oldSectionRun = editor.InsertText("test");
            Section oldSection = (Section)oldSectionRun.Paragraph.BlockContainer;
            editor.InsertParagraph();

            Table table = editor.InsertTable(3, 3);
            TableCell cell1 = table.Rows.First().Cells.First();
            Paragraph cell1Paragraph = cell1.Blocks.AddParagraph();
            editor.MoveToParagraphStart(cell1Paragraph);
            Run cell1run = editor.InsertText("cell1");

            TableCell cell2 = table.Rows[1].Cells.First();
            Paragraph cell2Paragraph = cell2.Blocks.AddParagraph();
            editor.MoveToParagraphStart(cell2Paragraph);
            Run cell2run = editor.InsertText("cell2");


            TableCell cell3 = table.Rows[2].Cells.First();
            Paragraph cell3Paragraph = cell3.Blocks.AddParagraph();
            editor.MoveToParagraphStart(cell3Paragraph);
            Run cell3run = editor.InsertText("cell3");

            // Act
            editor.MoveToParagraphStart(cell3Paragraph);
            Section newSection = editor.InsertSection();

            Table aboveTable = oldSection.EnumerateChildrenOfType<Table>().FirstOrDefault();
            Table belowTable = newSection.EnumerateChildrenOfType<Table>().FirstOrDefault();

            // Assert
            Assert.AreEqual(2, aboveTable.GridRowsCount);
            Assert.AreEqual(1, belowTable.GridRowsCount);

            Run splittedCellAboveRun1 = ((Paragraph)aboveTable.Rows.First().Cells.First().Blocks.First()).Inlines.First() as Run;
            Run splittedCellAboveRun2 = ((Paragraph)aboveTable.Rows.Last().Cells.First().Blocks.First()).Inlines.First() as Run;
            Run splittedCellBelowRun = ((Paragraph)belowTable.Rows.First().Cells.First().Blocks.First()).Inlines.First() as Run;
            Assert.AreEqual(cell1run.Text, splittedCellAboveRun1.Text);
            Assert.AreEqual(cell2run.Text, splittedCellAboveRun2.Text);
            Assert.AreEqual(cell3run.Text, splittedCellBelowRun.Text);
        }

        [TestMethod]
        public void InsertSection_WhenIsInNestedTable_OuterTableIsSplittedFromMostOuterTableCellRow()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run oldSectionRun = editor.InsertText("test");
            Section oldSection = (Section)oldSectionRun.Paragraph.BlockContainer;
            editor.InsertParagraph();

            Table table = editor.InsertTable(3, 3);
            TableCell outerCellRow2 = table.Rows[1].Cells.First();
            Paragraph outerCellRow2Paragraph = outerCellRow2.Blocks.AddParagraph();
            editor.MoveToParagraphStart(outerCellRow2Paragraph);
            Run outerCellRow2Run = editor.InsertText("cell2");

            TableCell outerCellRow2NestedTableContainer = table.Rows[1].Cells[1];
            Paragraph outerCellRow2NestedTableContainerParagraph = outerCellRow2NestedTableContainer.Blocks.AddParagraph();
            editor.MoveToParagraphStart(outerCellRow2NestedTableContainerParagraph);

            Table nestedTable = editor.InsertTable(5, 5);
            Paragraph nestedTableCellParagraph = nestedTable.Rows[3].Cells[1].Blocks.AddParagraph();

            // Act
            editor.MoveToParagraphStart(nestedTableCellParagraph);
            Section newSection = editor.InsertSection();

            // Assert
            Table aboveTable = oldSection.EnumerateChildrenOfType<Table>().FirstOrDefault();
            Table belowTable = newSection.EnumerateChildrenOfType<Table>().Where(t => t.GridRowsCount != 5).FirstOrDefault();
            Assert.AreEqual(1, aboveTable.GridRowsCount);
            Assert.AreEqual(2, belowTable.GridRowsCount);

            Run splittedCellBelowRun = ((Paragraph)belowTable.Rows[0].Cells[0].Blocks.First()).Inlines.First() as Run;
            Assert.AreEqual(outerCellRow2Run.Text, splittedCellBelowRun.Text);

            Table newNestedTable = belowTable.EnumerateChildrenOfType<Table>().FirstOrDefault();
            Assert.AreEqual(5, newNestedTable.GridRowsCount);
        }

        [TestMethod]
        public void InsertPageBreak_WhenIsInTableWith3RowsAnd3Columns_AndCurrentParagraphIsInTableCellFromMiddleRow_TableIsSpltted()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("test");
            editor.InsertParagraph();

            Table table = editor.InsertTable(3, 3);
            TableCell cell1 = table.Rows.First().Cells.First();
            Paragraph cell1Paragraph = cell1.Blocks.AddParagraph();
            editor.MoveToParagraphStart(cell1Paragraph);
            Run cell1run = editor.InsertText("cell1");

            TableCell cell2 = table.Rows[1].Cells.First();
            Paragraph cell2Paragraph = cell2.Blocks.AddParagraph();
            editor.MoveToParagraphStart(cell2Paragraph);
            Run cell2run = editor.InsertText("cell2");

            // Act
            editor.MoveToParagraphStart(cell2Paragraph);
            editor.InsertBreak(BreakType.PageBreak);

            Table aboveTable = editor.Document.EnumerateChildrenOfType<Table>().Where(t => t.GridRowsCount == 1).FirstOrDefault();
            Table belowTable = editor.Document.EnumerateChildrenOfType<Table>().Where(t => t.GridRowsCount == 2).FirstOrDefault();

            // Assert
            Assert.AreEqual(2, editor.Document.EnumerateChildrenOfType<Table>().Count());

            Run splittedCellAboveRun = ((Paragraph)aboveTable.Rows.First().Cells.First().Blocks.First()).Inlines.First() as Run;
            Run splittedCellBelowRun = ((Paragraph)belowTable.Rows.First().Cells.First().Blocks.First()).Inlines.First() as Run;
            Assert.AreEqual(cell1run.Text, splittedCellAboveRun.Text);
            Assert.AreEqual(cell2run.Text, splittedCellBelowRun.Text);
        }

        [TestMethod]
        public void InsertPageBreak_WhenIsInTableWith3RowsAnd3Columns_AndCurrentParagraphIsInTableCellFromFirstRow_BreakIsInsertedBeforeTable()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("test");
            editor.InsertParagraph();

            Table table = editor.InsertTable(3, 3);
            TableCell cell1 = table.Rows.First().Cells.First();
            Paragraph cell1Paragraph = cell1.Blocks.AddParagraph();
            editor.MoveToParagraphStart(cell1Paragraph);
            editor.InsertText("cell");

            // Act
            editor.MoveToParagraphStart(cell1Paragraph);
            Break br = editor.InsertBreak(BreakType.PageBreak);

            // Assert
            Section section = editor.Document.Sections.First();
            int breakParagrahIndex = section.Blocks.IndexOf(br.Paragraph);
            int tableIndex = section.Blocks.IndexOf(table);
            Assert.IsTrue(breakParagrahIndex < tableIndex);
        }

        [TestMethod]
        public void InsertField_InsertsFieldCorrectly()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            editor.InsertField("FieldCodePartHere", "FieldResultPartHere");

            // Assert
            AssertFieldInserted(editor.Document, "FieldCodePartHere", "FieldResultPartHere");
        }

        [TestMethod]
        public void InsertField_WithEmtpyResult_InsertsFieldCorrectly()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            editor.InsertField("FieldCodePartHere", string.Empty);

            // Assert
            AssertFieldInserted(editor.Document, "FieldCodePartHere", null);
        }

        [TestMethod]
        public void InsertField_WithNullResult_InsertsFieldCorrectly()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            editor.InsertField("FieldCodePartHere", null);

            // Assert
            AssertFieldInserted(editor.Document, "FieldCodePartHere", null);
        }

        [TestMethod]
        public void InsertField_WithEmtpyCode_InsertsFieldCorrectly()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            editor.InsertField(string.Empty, "FieldResultPartHere");

            // Assert
            AssertFieldInserted(editor.Document, null, "FieldResultPartHere");
        }

        [TestMethod]
        public void InsertField_WithNullCode_InsertsFieldCorrectly()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            editor.InsertField(null, "FieldResultPartHere");

            // Assert
            AssertFieldInserted(editor.Document, null, "FieldResultPartHere");
        }

        [TestMethod]
        public void InsertDocument_WhenTargetDocumentIsEmpty_InsertsTheSourceDocumentAndResultSectionIsWithSourceSectionProperties()
        {
            // Arrange
            string runText = "source document run";
            RadFlowDocument sourceDocument = GetSimpleSourceDocumentToInsert(runText);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            editor.InsertDocument(sourceDocument);
            RadFlowDocument targetDocument = editor.Document;

            // Assert
            Assert.AreEqual(1, targetDocument.Sections.Count);
            Assert.AreEqual(PageOrientation.Landscape, targetDocument.Sections[0].PageOrientation);
            Assert.AreEqual(2, targetDocument.EnumerateChildrenOfType<Paragraph>().Count());
            Assert.AreEqual(runText, targetDocument.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
        }

        [TestMethod]
        public void InsertDocument_WhenTargetDocumentIsEmptyAndSourceDocumentContainsMultipleSections_InsertsTheSourceDocumentAndResultSectionsAreWithSourceSectionProperties()
        {
            // Arrange
            RadFlowDocument sourceDocument = new RadFlowDocument();
            sourceDocument.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("source1");
            sourceDocument.Sections[0].Rotate(PageOrientation.Landscape);
            sourceDocument.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("source2");
            sourceDocument.Sections[1].PageSize = PaperTypeConverter.ToSize(PaperTypes.A5);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            editor.InsertDocument(sourceDocument);
            RadFlowDocument targetDocument = editor.Document;

            // Assert
            Assert.AreEqual(2, targetDocument.Sections.Count);
            Assert.AreEqual(PageOrientation.Landscape, targetDocument.Sections[0].PageOrientation);
            Assert.AreEqual(PaperTypeConverter.ToSize(PaperTypes.A5), targetDocument.Sections[1].PageSize);
            Assert.AreEqual("source1", targetDocument.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
            Assert.AreEqual("source2", targetDocument.EnumerateChildrenOfType<Run>().LastOrDefault().Text);
        }

        [TestMethod]
        public void InsertDocument_WhenTargetDocumentHasSingleSectionAndSourceDocumentContainsSingleSection_InsertsTheSourceDocumentAndResultSectionIsWithTargetSectionProperties()
        {
            // Arrange
            string runText = "source document run";
            RadFlowDocument sourceDocument = GetSimpleSourceDocumentToInsert(runText);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("target document");

            // Act
            editor.InsertDocument(sourceDocument);
            RadFlowDocument targetDocument = editor.Document;

            // Assert
            Assert.AreEqual(1, targetDocument.Sections.Count);
            Assert.AreEqual(PageOrientation.Portrait, targetDocument.Sections[0].PageOrientation);
            Assert.AreEqual(2, targetDocument.EnumerateChildrenOfType<Paragraph>().Count());
            Assert.AreEqual("target document", targetDocument.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
            Assert.AreEqual(runText, targetDocument.EnumerateChildrenOfType<Run>().LastOrDefault().Text);
        }

        [TestMethod]
        public void InsertDocument_WhenTargetDocumentHasSingleSectionAndSourceDocumentHasMultipleSections_ResultDocumentHasTargetSectionPropertiesForLastSectionAndSourceDocumentLastSectionPropertiesAreIgnored()
        {
            // Arrange
            RadFlowDocumentEditor targetEditor = new RadFlowDocumentEditor(new RadFlowDocument());
            var run1 = targetEditor.InsertText("target document insert position:");
            targetEditor.InsertText("target document after insert position:");
            targetEditor.MoveToInlineEnd(run1);
            targetEditor.Document.Sections[0].Headers.Add().Blocks.AddParagraph().Inlines.AddRun("original");

            RadFlowDocumentEditor sourceEditor = new RadFlowDocumentEditor(new RadFlowDocument());
            sourceEditor.InsertText("first source section A5");
            var section = sourceEditor.Document.Sections[0];
            section.Headers.Add().Blocks.AddParagraph().Inlines.AddRun("first source section header");
            section.PageSize = PaperTypeConverter.ToSize(PaperTypes.A5);

            sourceEditor.InsertSection();
            sourceEditor.InsertText("second source section Landscape");
            section = sourceEditor.Document.Sections[1];
            section.Rotate(Telerik.Windows.Documents.Model.PageOrientation.Landscape);
            section.Headers.Add().Blocks.AddParagraph().Inlines.AddRun("second source section header");

            // Act
            targetEditor.InsertDocument(sourceEditor.Document, new InsertDocumentOptions() { InsertLastParagraphMarker = false });
            RadFlowDocument targetDocument = targetEditor.Document;

            // Assert
            var sections = targetDocument.Sections;
            Assert.AreEqual(2, sections.Count);

            // Section 1
            Assert.AreEqual(PageOrientation.Portrait, sections[0].PageOrientation);
            Assert.AreEqual(PaperTypeConverter.ToSize(PaperTypes.A5), sections[0].PageSize);
            Assert.AreEqual(3, sections[0].EnumerateChildrenOfType<Paragraph>().Count()); // One is in the header and the other stores the section props.

            var section1runs = (sections[0].Blocks.FirstOrDefault() as Paragraph).Inlines.Cast<Run>().ToList();
            Assert.AreEqual("target document insert position:", section1runs[0].Text);
            Assert.AreEqual("first source section A5", section1runs[1].Text);
            Assert.AreEqual("first source section header", sections[0].Headers.Default.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);

            // Section 2
            Assert.AreEqual(PageOrientation.Portrait, sections[1].PageOrientation);
            Assert.AreEqual(PaperTypeConverter.ToSize(PaperTypes.Letter), sections[1].PageSize);
            Assert.AreEqual(2, sections[1].EnumerateChildrenOfType<Paragraph>().Count()); // One is in the header.

            var section2runs = (sections[1].Blocks.FirstOrDefault() as Paragraph).Inlines.Cast<Run>().ToList();
            Assert.AreEqual("second source section Landscape", section2runs[0].Text);
            Assert.AreEqual("target document after insert position:", section2runs[1].Text);
            Assert.AreEqual("original", sections[1].Headers.Default.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);

        }

        [TestMethod]
        public void InsertDocument_WhenTargetDocumentHasMultipleSectionsAndSourceDocumentHasMultipleSections_ResultDocumentHasTargetSectionPropertiesForLastSectionAndSourceDocumentLastSectionPropertiesAreIgnored()
        {
            // Arrange
            RadFlowDocumentEditor targetEditor = new RadFlowDocumentEditor(new RadFlowDocument());
            var firstTargetSection = targetEditor.Document.Sections.AddSection();
            targetEditor.MoveToInlineEnd(firstTargetSection.Blocks.AddParagraph().Inlines.AddRun("section1"));
            firstTargetSection.Rotate(Telerik.Windows.Documents.Model.PageOrientation.Landscape);
            targetEditor.Document.Sections[0].Headers.Add().Blocks.AddParagraph().Inlines.AddRun("original");

            targetEditor.InsertSection();

            var run1 = targetEditor.InsertText("target document insert position:");
            targetEditor.InsertText("target document after insert position:");
            targetEditor.MoveToInlineEnd(run1);

            RadFlowDocumentEditor sourceEditor = new RadFlowDocumentEditor(new RadFlowDocument());
            sourceEditor.InsertText("first source section A5");
            var section = sourceEditor.Document.Sections[0];
            section.Headers.Add().Blocks.AddParagraph().Inlines.AddRun("first source section header");
            section.PageSize = PaperTypeConverter.ToSize(PaperTypes.A5);

            sourceEditor.InsertSection();
            sourceEditor.InsertText("second source section Landscape");
            section = sourceEditor.Document.Sections[1];
            section.Rotate(Telerik.Windows.Documents.Model.PageOrientation.Landscape);
            section.Headers.Add().Blocks.AddParagraph().Inlines.AddRun("second source section header");

            // Act
            targetEditor.InsertDocument(sourceEditor.Document, new InsertDocumentOptions() { InsertLastParagraphMarker = false });
            RadFlowDocument targetDocument = targetEditor.Document;

            // Assert
            var sections = targetDocument.Sections;
            Assert.AreEqual(3, sections.Count);

            // Section 1
            Assert.AreEqual(PageOrientation.Landscape, sections[0].PageOrientation);
            Assert.AreEqual(2, sections[0].EnumerateChildrenOfType<Paragraph>().Count()); // One is in the header.

            var section1runs = (sections[0].Blocks.FirstOrDefault() as Paragraph).Inlines.Cast<Run>().ToList();
            Assert.AreEqual("section1", section1runs[0].Text);
            Assert.AreEqual("original", sections[0].Headers.Default.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);

            // Section 2
            Assert.AreEqual(PageOrientation.Portrait, sections[1].PageOrientation);
            Assert.AreEqual(PaperTypeConverter.ToSize(PaperTypes.A5), sections[1].PageSize);
            Assert.AreEqual(3, sections[1].EnumerateChildrenOfType<Paragraph>().Count()); // One is in the header.

            var section2runs = (sections[1].Blocks.FirstOrDefault() as Paragraph).Inlines.Cast<Run>().ToList();
            Assert.AreEqual("target document insert position:", section2runs[0].Text);
            Assert.AreEqual("first source section A5", section2runs[1].Text);
            Assert.AreEqual("first source section header", sections[1].Headers.Default.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);

            // Section 3
            Assert.AreEqual(PageOrientation.Portrait, sections[2].PageOrientation);
            Assert.AreEqual(PaperTypeConverter.ToSize(PaperTypes.Letter), sections[2].PageSize);
            Assert.AreEqual(2, sections[2].EnumerateChildrenOfType<Paragraph>().Count()); // One is in the header.

            var section3runs = (sections[2].Blocks.FirstOrDefault() as Paragraph).Inlines.Cast<Run>().ToList();
            Assert.AreEqual("second source section Landscape", section3runs[0].Text);
            Assert.AreEqual("target document after insert position:", section3runs[1].Text);
            Assert.AreEqual("first source section header", sections[2].Headers.Default.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);

        }

        [TestMethod]
        public void InsertDocument_WhenInsertLastParagraphMarkerIsFalse_InsertsTheSourceDocumentWithoutTheLastParagraphOfTheSourceDocument()
        {
            // Arrange
            string runText = "source document run";
            RadFlowDocument sourceDocument = GetSimpleSourceDocumentToInsert(runText);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("target document");

            // Act
            editor.InsertDocument(sourceDocument, new InsertDocumentOptions() { InsertLastParagraphMarker = false });
            RadFlowDocument targetDocument = editor.Document;

            // Assert
            Assert.AreEqual(1, targetDocument.Sections.Count);
            Assert.AreEqual(PageOrientation.Portrait, targetDocument.Sections[0].PageOrientation);
            Assert.AreEqual(1, targetDocument.EnumerateChildrenOfType<Paragraph>().Count());
            Assert.AreEqual("target document", targetDocument.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
            Assert.AreEqual(runText, targetDocument.EnumerateChildrenOfType<Run>().LastOrDefault().Text);
        }

        [TestMethod]
        public void InsertDocument_WhenInsertLastParagraphMarkerIsTrueAndSourceDocumentHasMultipleSections_InsertsTheLastParagraphFromTheLastSection()
        {
            // Arrange
            string runText = "source document run";
            RadFlowDocument sourceDocument = GetSimpleSourceDocumentToInsert(runText);

            Paragraph lastSourceParagraph = sourceDocument.Sections.AddSection().Blocks.AddParagraph();
            lastSourceParagraph.Properties.LineSpacing.LocalValue = 100;
            lastSourceParagraph.Inlines.AddRun("last paragraph");

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("target document");

            // Act
            editor.InsertDocument(sourceDocument, new InsertDocumentOptions() { InsertLastParagraphMarker = true });
            RadFlowDocument targetDocument = editor.Document;

            // Assert
            Assert.AreEqual(2, targetDocument.Sections.Count);
            Assert.AreEqual(4, targetDocument.EnumerateChildrenOfType<Paragraph>().Count());

            var runs = targetDocument.EnumerateChildrenOfType<Run>().ToList();
            Assert.AreEqual(3, runs.Count);
            Assert.AreEqual("last paragraph", runs[2].Text);
            StylePropertyAsserter.AssertPropertiesAreEqual(lastSourceParagraph.Properties, runs[2].Paragraph.Properties);
        }

        [TestMethod]
        public void InsertDocument_WhenWithoutInsertDocumentOptionsAndTargetParagraphAndSourceParagraphHaveDifferentStyles_TheInsertTargetParagraphHaveTheStyleOfTheSourceParagraphAndNextParagraphHasTargetParagraphStyle()
        {
            // Arrange
            string runText = "source document run";
            RadFlowDocument sourceDocument = GetSimpleSourceDocumentToInsert(runText);
            sourceDocument.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(2));
            sourceDocument.EnumerateChildrenOfType<Paragraph>().FirstOrDefault().StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(2);

            RadFlowDocumentEditor targetEditor = new RadFlowDocumentEditor(new RadFlowDocument());
            targetEditor.Document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1));
            targetEditor.InsertText("target document").Paragraph.StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(1);

            // Act
            targetEditor.InsertDocument(sourceDocument);
            RadFlowDocument targetDocument = targetEditor.Document;

            // Assert
            var paragraphs = targetDocument.EnumerateChildrenOfType<Paragraph>().ToList();

            Assert.AreEqual(2, paragraphs.Count);
            Assert.AreEqual(BuiltInStyleNames.GetHeadingStyleIdByIndex(2), paragraphs[0].StyleId);
            Assert.AreEqual(BuiltInStyleNames.GetHeadingStyleIdByIndex(1), paragraphs[1].StyleId);
        }

        [TestMethod]
        public void InsertDocument_WhenEditorPositionIsBetweenTwoRunsAndInsertLastParagraphMarkerIsFalse_InsertsTheSourceDocumentBetweenRunsWithoutTheLastParagraphOfTheSourceDocument()
        {
            // Arrange
            string runText = "source document run";
            RadFlowDocument sourceDocument = GetSimpleSourceDocumentToInsert(runText);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            var run = editor.InsertText("target document");
            editor.InsertText("target document second run");
            editor.MoveToInlineEnd(run);

            // Act
            editor.InsertDocument(sourceDocument, new InsertDocumentOptions() { InsertLastParagraphMarker = false });
            RadFlowDocument targetDocument = editor.Document;

            // Assert
            Assert.AreEqual(1, targetDocument.Sections.Count);
            Assert.AreEqual(PageOrientation.Portrait, targetDocument.Sections[0].PageOrientation);
            Assert.AreEqual(1, targetDocument.EnumerateChildrenOfType<Paragraph>().Count());

            var runs = targetDocument.EnumerateChildrenOfType<Run>().ToList();
            Assert.AreEqual("target document", runs[0].Text);
            Assert.AreEqual(runText, runs[1].Text);
            Assert.AreEqual("target document second run", runs[2].Text);
        }

        [TestMethod]
        public void InsertDocument_WhenEditorPositionIsBetweenTwoRunsAndInsertLastParagraphMarkerIsFalseAndTargetParagraphHasStyle_InsertsTheSourceParagraphsWithTheirStyleAndWithoutTheLastParagraphOfTheSourceDocumentAndLastParagraphHasTargetParagraphStyle()
        {
            // Arrange
            RadFlowDocument sourceDocument = new RadFlowDocument();
            sourceDocument.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("source p1");
            sourceDocument.Sections[0].Blocks.AddParagraph().Inlines.AddRun("source p2");

            RadFlowDocumentEditor targetEditor = new RadFlowDocumentEditor(new RadFlowDocument());
            var run = targetEditor.InsertText("target document");
            targetEditor.InsertText("target document second run");
            targetEditor.MoveToInlineEnd(run);
            run.Paragraph.StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(1);
            targetEditor.Document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1));

            // Act
            targetEditor.InsertDocument(sourceDocument, new InsertDocumentOptions() { InsertLastParagraphMarker = false });
            RadFlowDocument targetDocument = targetEditor.Document;

            // Assert
            var paragraphs = targetDocument.EnumerateChildrenOfType<Paragraph>().ToList();
            Assert.AreEqual(2, paragraphs.Count);

            var runs = targetDocument.EnumerateChildrenOfType<Run>().ToList();
            Assert.AreEqual("target document", runs[0].Text);
            Assert.AreEqual("source p1", runs[1].Text);
            Assert.AreEqual("source p2", runs[2].Text);
            Assert.AreEqual("target document second run", runs[3].Text);

            Assert.AreEqual(runs[0].Paragraph, runs[1].Paragraph);
            Assert.AreEqual(runs[2].Paragraph, runs[3].Paragraph);
            Assert.AreEqual(string.Empty, runs[0].Paragraph.StyleId);
            Assert.AreEqual(BuiltInStyleNames.GetHeadingStyleIdByIndex(1), runs[3].Paragraph.StyleId);
        }

        [TestMethod]
        public void InsertDocument_WhenEditorPositionIsInTableAndSourceDocumentHasSingleSection_SourceDocumentBlocksAreInsertedInTheTableCell()
        {
            // Arrange
            string runText = "source document run";
            RadFlowDocument sourceDocument = GetSimpleSourceDocumentToInsert(runText);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run run = editor.InsertTable(3, 3).Rows[1].Cells[0].Blocks.AddParagraph().Inlines.AddRun("cell (1,0)");
            editor.MoveToInlineEnd(run);

            // Act
            editor.InsertDocument(sourceDocument);
            RadFlowDocument targetDocument = editor.Document;

            // Assert
            Assert.AreEqual(1, targetDocument.Sections.Count);
            Assert.AreEqual(PageOrientation.Portrait, targetDocument.Sections[0].PageOrientation);

            var tables = targetDocument.EnumerateChildrenOfType<Table>().ToList();
            Assert.AreEqual(1, tables.Count);

            var cellParagraphs = tables[0].Rows[1].Cells[0].EnumerateChildrenOfType<Paragraph>().ToList();
            Assert.AreEqual(2, cellParagraphs.Count);
            Assert.AreEqual("cell (1,0)", (cellParagraphs[0].Inlines[0] as Run).Text);
            Assert.AreEqual(runText, (cellParagraphs[0].Inlines[1] as Run).Text);
            Assert.AreEqual(0, cellParagraphs[1].Inlines.Count); // The last trailing paragraph. It shall be empty.
        }

        [TestMethod]
        public void InsertDocument_WhenTargetAndSourceDocumentsHaveSameNameStylesAndConflictingModeIsRename_TheResultDocumentContainsTwoStyles()
        {
            // Arrange
            string styleName = "MyStyle";

            RadFlowDocumentEditor targetEditor = new RadFlowDocumentEditor(new RadFlowDocument());
            var targetStyle = new Telerik.Windows.Documents.Flow.Model.Styles.Style(styleName, StyleType.Paragraph);
            targetEditor.Document.StyleRepository.Add(targetStyle);
            targetEditor.InsertText("p1").Paragraph.StyleId = styleName;

            RadFlowDocument sourceDocument = new RadFlowDocument();
            var sourceStyle = new Telerik.Windows.Documents.Flow.Model.Styles.Style(styleName, StyleType.Paragraph);
            sourceStyle.CharacterProperties.FontSize.LocalValue = 30;
            sourceDocument.StyleRepository.Add(sourceStyle);
            sourceDocument.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("p2").StyleId = styleName;

            // Act
            targetEditor.InsertDocument(sourceDocument, new InsertDocumentOptions()
            {
                ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle
            });

            // Assert
            var renamedSourceStyle = targetEditor.Document.StyleRepository.GetStyle(styleName + "_1"); // StyleRepositoryMerger style rename suffix const.
            Assert.IsNotNull(renamedSourceStyle);
            Assert.AreEqual(30, renamedSourceStyle.CharacterProperties.FontSize.LocalValue.Value);

            var nonRenamedTargetStyle = targetEditor.Document.StyleRepository.GetStyle(styleName);
            Assert.IsNotNull(nonRenamedTargetStyle);
            Assert.AreEqual(null, nonRenamedTargetStyle.CharacterProperties.FontSize.LocalValue);
        }

        [TestMethod]
        public void InsertDocument_WhenWithoutInsertDocumentOptions_TheParagraphWhereTheDocumentIsInsertedHasFormattingSameAsTheFirstSourceParagraph()
        {
            // Arrange
            RadFlowDocumentEditor targetEditor = new RadFlowDocumentEditor(new RadFlowDocument());
            var targetStyle = new Telerik.Windows.Documents.Flow.Model.Styles.Style("Style1", StyleType.Paragraph);
            targetEditor.Document.StyleRepository.Add(targetStyle);
            targetEditor.InsertText("p1").Paragraph.StyleId = "Style1";

            RadFlowDocument sourceDocument = new RadFlowDocument();
            var sourceStyle = new Telerik.Windows.Documents.Flow.Model.Styles.Style("Style2", StyleType.Paragraph);
            sourceStyle.CharacterProperties.FontSize.LocalValue = 30;
            sourceDocument.StyleRepository.Add(sourceStyle);
            sourceDocument.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("p2").Paragraph.StyleId = "Style2";

            // Act
            targetEditor.InsertDocument(sourceDocument);

            // Assert
            Paragraph targetParagraph = targetEditor.Document.EnumerateChildrenOfType<Run>().Where(r => r.Text == "p1").FirstOrDefault().Paragraph;
            Paragraph sourceParagraph = targetEditor.Document.EnumerateChildrenOfType<Run>().Where(r => r.Text == "p2").FirstOrDefault().Paragraph;

            Assert.AreEqual(targetParagraph, sourceParagraph);
            var style = targetEditor.Document.StyleRepository.GetStyle(targetParagraph.StyleId);

            Assert.IsNotNull(style);
            Assert.AreEqual("Style2", style.Id);
            Assert.AreEqual(30, style.CharacterProperties.FontSize.LocalValue.Value);
        }

        [TestMethod]
        public void InsertDocument_WhenTargetDocumentHasSingleParagraphAndSourceDocumentHasTwoParagraphsAndInsertLastParagraphMarkerIsFalse_FirstParagraphInResultDocumentHasStyleSameAsSourceFirstParagraphAndSecondParagraphHasStyleAsTargetParagraph()
        {
            // Arrange
            RadFlowDocument sourceDocument = new RadFlowDocument();
            sourceDocument.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(2));
            sourceDocument.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("source p1").Paragraph.StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(2);
            sourceDocument.Sections[0].Blocks.AddParagraph().Inlines.AddRun("source p2");

            RadFlowDocumentEditor targetEditor = new RadFlowDocumentEditor(new RadFlowDocument());
            targetEditor.Document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1));
            targetEditor.InsertText("target document").Paragraph.StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(1);

            // Act
            targetEditor.InsertDocument(sourceDocument, new InsertDocumentOptions()
            {
                InsertLastParagraphMarker = false
            });

            var paragraphs = targetEditor.Document.EnumerateChildrenOfType<Paragraph>().ToList();

            // Assert
            Assert.AreEqual(2, paragraphs.Count);
            Assert.AreEqual(BuiltInStyleNames.GetHeadingStyleIdByIndex(2), paragraphs[0].StyleId);
            Assert.AreEqual(BuiltInStyleNames.GetHeadingStyleIdByIndex(1), paragraphs[1].StyleId);
        }

        [TestMethod]
        public void InsertDocument_WhenTargetDocumentHasSingleParagraphAndSourceDocumentHasTwoParagraphsAndInsertLastParagraphMarkerIsTrue_FirstTwoParagraphsHaveStylesAsSourceParagraphsAndLastParagraphHasStyleAsTargetParagraph()
        {
            // Arrange
            RadFlowDocument sourceDocument = new RadFlowDocument();
            sourceDocument.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(2));
            sourceDocument.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("source p1").Paragraph.StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(2);
            sourceDocument.Sections[0].Blocks.AddParagraph().Inlines.AddRun("source p2");

            RadFlowDocumentEditor targetEditor = new RadFlowDocumentEditor(new RadFlowDocument());
            targetEditor.Document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1));
            targetEditor.InsertText("target document").Paragraph.StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(1);

            // Act
            targetEditor.InsertDocument(sourceDocument, new InsertDocumentOptions()
            {
                InsertLastParagraphMarker = true
            });

            var paragraphs = targetEditor.Document.EnumerateChildrenOfType<Paragraph>().ToList();

            // Assert
            Assert.AreEqual(3, paragraphs.Count);
            Assert.AreEqual(BuiltInStyleNames.GetHeadingStyleIdByIndex(2), paragraphs[0].StyleId);
            Assert.AreEqual(string.Empty, paragraphs[1].StyleId);
            Assert.AreEqual(BuiltInStyleNames.GetHeadingStyleIdByIndex(1), paragraphs[2].StyleId);
        }

        [TestMethod]
        public void InsertDocument_WhenSourceDocumentIsEmpty_ThrowsArgumentException()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act && Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(() =>
            {
                editor.InsertDocument(new RadFlowDocument());
            });
        }

        [TestMethod]
        public void InsertDocument_WhenEditorPositionIsInTableAndSourceDocumentHasMultipleSections_ThrowsInvalidOperationException()
        {
            // Arrange
            RadFlowDocumentEditor sourceEditor = new RadFlowDocumentEditor(new RadFlowDocument());
            sourceEditor.InsertText("first source section A5");
            var section = sourceEditor.Document.Sections[0];
            section.Headers.Add().Blocks.AddParagraph().Inlines.AddRun("first source section header");
            section.PageSize = PaperTypeConverter.ToSize(PaperTypes.A5);

            sourceEditor.InsertSection();
            sourceEditor.InsertText("second source section Landscape");
            section = sourceEditor.Document.Sections[1];
            section.Rotate(Telerik.Windows.Documents.Model.PageOrientation.Landscape);
            section.Headers.Add().Blocks.AddParagraph().Inlines.AddRun("second source section header");

            RadFlowDocumentEditor targetEditor = new RadFlowDocumentEditor(new RadFlowDocument());
            Table table = targetEditor.InsertTable(3, 3);
            table.Rows[0].Cells[0].Blocks.AddParagraph().Inlines.AddRun("cell (0,0)");
            Run run = table.Rows[1].Cells[0].Blocks.AddParagraph().Inlines.AddRun("cell (1,0)");
            targetEditor.MoveToInlineEnd(run);

            // Act && Assert
            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
            {
                targetEditor.InsertDocument(sourceEditor.Document);
            });
        }

        [TestMethod]
        public void InsertDocument_WhenFirstParagraphFromSourceDocumentHasLocalPropertyWhichSecondParagraphHasnt_InTheResultDocumentSecondParagraphShouldntInheritFirstParagraphProperties()
        {
            // Arrange
            RadFlowDocument sourceDocument = new RadFlowDocument();
            Section section = sourceDocument.Sections.AddSection();
            Paragraph p1 = section.Blocks.AddParagraph();
            p1.Inlines.AddRun("p1");
            p1.Properties.ListId.LocalValue = 0;
            p1.Properties.ListLevel.LocalValue = 0;
            p1.Properties.LineSpacing.LocalValue = 10;

            Paragraph p2 = section.Blocks.AddParagraph();
            p2.Inlines.AddRun("p2");

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            editor.InsertDocument(sourceDocument);
            var runs = editor.Document.EnumerateChildrenOfType<Run>().ToList();

            // Assert
            StylePropertyAsserter.AssertPropertiesAreEqual(p1.Properties, runs[0].Paragraph.Properties);
            StylePropertyAsserter.AssertPropertiesAreEqual(p2.Properties, runs[1].Paragraph.Properties);
        }

        [TestMethod]
        public void InsertDocument_WhenFirstParagraphFromSourceDocumentHasPropertiesFromStyleWhichSecondParagraphHasnt_InTheResultDocumentSecondParagraphShouldntInheritFirstParagraphProperties()
        {
            // Arrange
            RadFlowDocument sourceDocument = new RadFlowDocument();
            var style = new Telerik.Windows.Documents.Flow.Model.Styles.Style("style", StyleType.Paragraph);
            style.ParagraphProperties.ListId.LocalValue = 0;
            style.ParagraphProperties.ListLevel.LocalValue = 0;
            style.ParagraphProperties.LineSpacing.LocalValue = 10;
            sourceDocument.StyleRepository.Add(style);

            Section section = sourceDocument.Sections.AddSection();
            Paragraph p1 = section.Blocks.AddParagraph();
            p1.Inlines.AddRun("p1");
            p1.StyleId = style.Id;

            Paragraph p2 = section.Blocks.AddParagraph();
            p2.Inlines.AddRun("p2");

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            editor.InsertDocument(sourceDocument);
            var runs = editor.Document.EnumerateChildrenOfType<Run>().ToList();

            // Assert
            Paragraph importedParagraph1 = runs[0].Paragraph;
            Assert.AreEqual(0, importedParagraph1.ListId);
            Assert.AreEqual(0, importedParagraph1.ListLevel);
            StylePropertyAsserter.AssertPropertiesAreEqual(p1.Properties, importedParagraph1.Properties);

            Paragraph importedParagraph2 = runs[1].Paragraph;
            Assert.AreEqual(-1, importedParagraph2.ListId);
            Assert.AreEqual(-1, importedParagraph2.ListLevel);
            StylePropertyAsserter.AssertPropertiesAreEqual(p2.Properties, importedParagraph2.Properties);
        }

        [TestMethod]
        public void InsertDocument_WhenTargetDocumentIsEmptyAndSourceDocumentHasSingleParagraphWithList_InTheResultDocumentSecondParagraphShouldHaveDefaultParagraphProperties()
        {
            // Arrange
            RadFlowDocument sourceDocument = new RadFlowDocument();
            Section section = sourceDocument.Sections.AddSection();
            Paragraph p = section.Blocks.AddParagraph();
            p.Inlines.AddRun("p1");
            p.Properties.ListId.LocalValue = 0;
            p.Properties.ListLevel.LocalValue = 0;
            p.Properties.LineSpacing.LocalValue = 10;

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            // Act
            editor.InsertDocument(sourceDocument);

            Paragraph[] paragraphs = editor.Document.EnumerateChildrenOfType<Paragraph>().ToArray();
            ParagraphProperties defaultParagraphProperties = new Paragraph(editor.Document).Properties;

            // Assert
            StylePropertyAsserter.AssertPropertiesAreEqual(p.Properties, paragraphs[0].Properties);
            StylePropertyAsserter.AssertPropertiesAreEqual(defaultParagraphProperties, paragraphs[1].Properties);
        }

        [TestMethod]
        public void InsertDocument_WhenTargetDocumentIsEmptyAndContainsParagraphsAndTables_InTheResultDocumentIsNotAddedAnExtraParagraphBeforeTable()
        {
            // Arrange
            RadFlowDocument sourceDocument = new RadFlowDocument();
            sourceDocument.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("test");
            sourceDocument.Sections[0].Blocks.Add(new Table(sourceDocument, 3, 3));

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            
            // Act
            editor.InsertDocument(sourceDocument);

            // Assert
            BlockBase[] blocks = editor.Document.EnumerateChildrenOfType<BlockBase>().ToArray();
            
            Assert.AreEqual(2, blocks.Length);
            Assert.AreEqual(typeof(Paragraph), blocks[0].GetType());
            Assert.AreEqual(typeof(Table), blocks[1].GetType());
        }

        private static RadFlowDocument GetSimpleSourceDocumentToInsert(string runText)
        {
            RadFlowDocument document = new RadFlowDocument();
            document.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun(runText);
            document.Sections[0].Rotate(PageOrientation.Landscape);
            return document;
        }

        private void AssertFieldInserted(RadFlowDocument doc, string code, string result)
        {
            List<InlineBase> inlines = doc.EnumerateChildrenOfType<InlineBase>().ToList();
            int index = 0;

            Assert.AreEqual(FieldCharacterType.Start, ((FieldCharacter)inlines[index++]).FieldCharacterType);
            if (code != null)
            {
                Assert.AreEqual(code, ((Run)inlines[index++]).Text);
            }
            Assert.AreEqual(FieldCharacterType.Separator, ((FieldCharacter)inlines[index++]).FieldCharacterType);
            if (result != null)
            {
                Assert.AreEqual(result, ((Run)inlines[index++]).Text);
            }
            Assert.AreEqual(FieldCharacterType.End, ((FieldCharacter)inlines[index++]).FieldCharacterType);

            Assert.AreEqual(inlines.Count, index);
        }
    }
}
