﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Collections.Generic;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Media;
#else
using System.Windows;
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Model.Fields;
#if WPF || SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Flow.Tests.Model.Fields
{
    [TestClass]
    public class MailMergeFieldTests
    {
        public static string StartingBracket = "«";
        public static string EndingBracket = "»";

        private List<FieldCharacter> GetFieldCharacters(RadFlowDocument document)
        {
            List<FieldCharacter> fieldCharacters = document.EnumerateChildrenOfType<FieldCharacter>().Where(f => f.FieldCharacterType == FieldCharacterType.Start).ToList();
            return fieldCharacters;
        }

        [TestMethod]
        public void InsertAndMergeField_IsFieldMergedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.InsertField("MERGEFIELD FirstName", "");
            editor.InsertText(" ");
            editor.InsertField("MERGEFIELD FamilyName", "");
            editor.InsertText(" ");
            editor.InsertField("MERGEFIELD Company", "");

            // Act
            List<Person> persons = Person.GetList(10); // change person count for debugging.
            RadFlowDocument mergedDocument = document.MailMerge(persons);

            // Assert
            List<Section> sections = mergedDocument.EnumerateChildrenOfType<Section>().ToList();
            Assert.AreEqual(persons.Count, sections.Count);

            for (int i = 0; i < persons.Count; i++)
            {
                Section section = sections[i];
                List<Run> runs = section.EnumerateChildrenOfType<Run>().ToList();

                string text = runs[0].Text;
                Assert.AreEqual(text, persons[i].FirstName);

                text = runs[2].Text;
                Assert.AreEqual(text, persons[i].FamilyName);

                text = runs[4].Text;
                Assert.AreEqual(text, persons[i].Company);
            }
        }

        [TestMethod]
        public void InsertAndMergeField_WithNoMergeData_IsFieldMergedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.InsertField("MERGEFIELD FirstName", "");
            editor.InsertText(" ");
            editor.InsertField("MERGEFIELD FamilyName", "");
            editor.InsertText(" ");
            editor.InsertField("MERGEFIELD Company", "");

            // Act
            document.UpdateFields();

            // Assert
            List<FieldCharacter> fieldCharacters = this.GetFieldCharacters(document);

            string text = fieldCharacters[0].FieldInfo.GetResult();
            Assert.AreEqual(text, StartingBracket + "FirstName" + EndingBracket);

            text = fieldCharacters[1].FieldInfo.GetResult();
            Assert.AreEqual(text, StartingBracket + "FamilyName" + EndingBracket);

            text = fieldCharacters[2].FieldInfo.GetResult();
            Assert.AreEqual(text, StartingBracket + "Company" + EndingBracket);
        }

        [TestMethod]
        public void InsertAndMergeField_WithPrefixSwitch_IsFieldMergedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            string prefix = "Mr.";
            editor.InsertField("MERGEFIELD FirstName \\b \"" + prefix + "\"", "");
            editor.InsertText(" ");
            editor.InsertField("MERGEFIELD \\b \"" + prefix + "\" FirstName", "");

            // Act
            document.UpdateFields();

            // Assert
            List<FieldCharacter> fieldCharacters = this.GetFieldCharacters(document);

            string text = fieldCharacters[0].FieldInfo.GetResult();
            Assert.AreEqual(text, prefix + StartingBracket + "FirstName" + EndingBracket);

            text = fieldCharacters[1].FieldInfo.GetResult();
            Assert.AreEqual(text, prefix + StartingBracket + "FirstName" + EndingBracket);
        }

        [TestMethod]
        public void InsertAndMergeField_WithSuffixSwitch_IsFieldMergedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            string suffix = ".";
            editor.InsertField("MERGEFIELD FirstName \\f \"" + suffix + "\"", "");
            editor.InsertText(" ");
            editor.InsertField("MERGEFIELD \\f \"" + suffix + "\" FirstName", "");

            // Act
            document.UpdateFields();

            // Assert
            List<FieldCharacter> fieldCharacters = this.GetFieldCharacters(document);

            string text = fieldCharacters[0].FieldInfo.GetResult();
            Assert.AreEqual(text, StartingBracket + "FirstName" + EndingBracket + suffix);

            text = fieldCharacters[1].FieldInfo.GetResult();
            Assert.AreEqual(text, StartingBracket + "FirstName" + EndingBracket + suffix);
        }

        [TestMethod]
        public void InsertMergeField_UpdateTheFieldInfo_IsFieldMergedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.InsertField("MERGEFIELD FirstName", "");
            editor.InsertText(" ");
            editor.InsertField("MERGEFIELD FamilyName", "");
            editor.InsertText(" ");
            editor.InsertField("MERGEFIELD Company", "");

            // Act
            List<FieldCharacter> fieldCharacters = document.EnumerateChildrenOfType<FieldCharacter>().Where(f => f.FieldCharacterType == FieldCharacterType.Start).ToList();

            for (int i = 0; i < fieldCharacters.Count; i++)
            {
                fieldCharacters[i].FieldInfo.UpdateField();
            }

            // Assert
            string text = fieldCharacters[0].FieldInfo.GetResult();
            Assert.AreEqual(text, StartingBracket + "FirstName" + EndingBracket);

            text = fieldCharacters[1].FieldInfo.GetResult();
            Assert.AreEqual(text, StartingBracket + "FamilyName" + EndingBracket);

            text = fieldCharacters[2].FieldInfo.GetResult();
            Assert.AreEqual(text, StartingBracket + "Company" + EndingBracket);
        }

        [TestMethod]
        public void Merge_WhenFieldHasFormatting_IsFormattingPreservedAfterMerge()
        {
            //Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            editor.CharacterFormatting.FontSize.LocalValue = 9;
            editor.InsertField("MERGEFIELD FirstName", "");

            //Act
            List<Person> persons = Person.GetList(1);
            RadFlowDocument mergedDocument = document.MailMerge(persons);

            //Assert
            Paragraph paragraph = mergedDocument.Sections.First().Blocks.First() as Paragraph;
            Run run = paragraph.Inlines.First() as Run;

            Assert.IsNotNull(run);
            Assert.AreEqual(persons.First().FirstName, run.Text);
            Assert.AreEqual(9, run.Properties.FontSize.LocalValue);
        }

        [TestMethod]
        public void Merge_WhenFieldHasMultipleFormattedRuns_IsFormattingAsInFirstRun()
        {
            //Arrange
            RadFlowDocument document = new RadFlowDocument();
            Paragraph paragraph = document.Sections.AddSection().Blocks.AddParagraph();

            FieldInfo field = new FieldInfo(document);

            ThemableColor expectedForegroundColor = new ThemableColor(Colors.Blue);
            double expectedFontSize = 9;

            paragraph.Inlines.Add(field.Start);
            Run firstRun = paragraph.Inlines.AddRun("");
            firstRun.Properties.FontSize.LocalValue = expectedFontSize;
            firstRun.Properties.ForegroundColor.LocalValue = expectedForegroundColor;
            Run secondRun = paragraph.Inlines.AddRun("MERGEFIELD");
            secondRun.Properties.FontWeight.LocalValue = FontWeights.Bold;
            Run thirdhRun = paragraph.Inlines.AddRun(" FamilyName");
            thirdhRun.Properties.FontFamily.LocalValue = new ThemableFontFamily(new FontFamily("Times New Roman"));
            thirdhRun.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Yellow);
            paragraph.Inlines.Add(field.Separator);
            Run fourthRun = paragraph.Inlines.AddRun("");
            fourthRun.Properties.ForegroundColor.LocalValue = new ThemableColor(Colors.Red);
            paragraph.Inlines.Add(field.End);

            //Act
            List<Person> persons = Person.GetList(1);
            RadFlowDocument mergedDocument = document.MailMerge(persons);

            //Assert
            Paragraph mergedDocumentParagraph = mergedDocument.Sections.First().Blocks.First() as Paragraph;
            Run run = mergedDocumentParagraph.Inlines.First() as Run;

            Assert.IsNotNull(run);
            Assert.AreEqual(persons.First().FamilyName, run.Text);
            Assert.AreEqual(expectedFontSize, run.Properties.FontSize.LocalValue);
            Assert.AreEqual(expectedForegroundColor, run.Properties.ForegroundColor.LocalValue);
            Assert.IsNull(run.Properties.FontWeight.LocalValue);
            Assert.IsNull(run.Properties.FontFamily.LocalValue);
            Assert.IsNull(run.Properties.BackgroundColor.LocalValue);
        }

        [TestMethod]
        public void MailMerge_SourceContainingNewLines()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.InsertField("MERGEFIELD Company ", "«Company»");
            editor.InsertField("MERGEFIELD Company ", "");

            List<Person> source = Person.GetList(1);
            source[0].Company = "Phone \r\n Company";

            RadFlowDocument clonedDocument = document.MailMerge(source);

            Assert.AreEqual(3, clonedDocument.EnumerateChildrenOfType<Paragraph>().Count());

            Assert.AreEqual(4, clonedDocument.EnumerateChildrenOfType<InlineBase>().Count());
        }

        [TestMethod]
        public void MailMerge_SourceContainingVerticalTabs()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.InsertField("MERGEFIELD Company ", "«Company»");
            editor.InsertField("MERGEFIELD Company ", "");

            List<Person> source = Person.GetList(1);
            source[0].Company = "Phone \v Company";

            RadFlowDocument clonedDocument = document.MailMerge(source);

            Assert.AreEqual(1, clonedDocument.EnumerateChildrenOfType<Paragraph>().Count());

            Assert.AreEqual(6, clonedDocument.EnumerateChildrenOfType<InlineBase>().Count());
        }

        [TestMethod]
        public void MailMerge_SourceContainingNewLinesAndParagraphHasStyleAppliedToParagraph_TheResultParagraphShouldHaveStyle()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            ThemableColor paragraphBackgroundColor = ThemableColor.FromArgb(255, 255, 0, 0);

            FieldInfo fi = editor.InsertField("MERGEFIELD Company ", "«Company»");
            editor.InsertField("MERGEFIELD Company ", "");

            Paragraph p = fi.Start.Paragraph;
            p.Properties.BackgroundColor.LocalValue = paragraphBackgroundColor;

            List<Person> source = Person.GetList(1);
            source[0].Company = "Phone \r\n Company";

            RadFlowDocument clonedDocument = document.MailMerge(source);

            List<Paragraph> paragraphs = clonedDocument.EnumerateChildrenOfType<Paragraph>().ToList();


            Assert.AreEqual(3, paragraphs.Count);
            Assert.AreEqual(4, clonedDocument.EnumerateChildrenOfType<InlineBase>().Count());


            foreach (Paragraph paragraph in paragraphs)
            {
                Assert.AreEqual(paragraphBackgroundColor, paragraph.Properties.BackgroundColor.LocalValue);
            }
        }

        [TestMethod]
        public void MailMerge_SourceContainingNewLinesAndParagraphHasStyleAppliedToEditor_TheResultParagraphShouldHaveStyle()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            ThemableColor paragraphBackgroundColor = ThemableColor.FromArgb(255, 255, 0, 0);

            editor.ParagraphFormatting.BackgroundColor.LocalValue = paragraphBackgroundColor;
            FieldInfo fi = editor.InsertField("MERGEFIELD Company ", "«Company»");
            editor.InsertField("MERGEFIELD Company ", "");

            List<Person> source = Person.GetList(1);
            source[0].Company = "Phone \r\n Company";

            RadFlowDocument clonedDocument = document.MailMerge(source);

            List<Paragraph> paragraphs = clonedDocument.EnumerateChildrenOfType<Paragraph>().ToList();


            Assert.AreEqual(3, paragraphs.Count);
            Assert.AreEqual(4, clonedDocument.EnumerateChildrenOfType<InlineBase>().Count());


            foreach (Paragraph paragraph in paragraphs)
            {
                Assert.AreEqual(paragraphBackgroundColor, paragraph.Properties.BackgroundColor.LocalValue);
            }
        }

        [TestMethod]
        public void MailMerge_SourceContainingNewLinesAndFieldSpanningInTwoParagraphs_TheResultParagraphShouldHaveStyle()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            ThemableColor paragraphBackgroundColor = ThemableColor.FromArgb(255, 255, 0, 0);
            ThemableColor paragraphBackgroundColor1 = ThemableColor.FromArgb(255, 0, 255, 0);

            Paragraph p1 = editor.InsertParagraph();
            p1.Properties.BackgroundColor.LocalValue = paragraphBackgroundColor;
            Paragraph p2 = editor.InsertParagraph();
            p2.Properties.BackgroundColor.LocalValue = paragraphBackgroundColor1;

            FieldInfo field = new FieldInfo(document);
            p1.Inlines.Add(field.Start);
            p1.Inlines.AddRun("MERGEFIELD Company ");
            p1.Inlines.Add(field.Separator);
            p1.Inlines.AddRun("«Company»");
            p2.Inlines.Add(field.End);

            List<Person> source = Person.GetList(1);
            source[0].Company = "Phone \r\n Company \r\n Test";

            RadFlowDocument clonedDocument = document.MailMerge(source);

            List<Paragraph> paragraphs = clonedDocument.EnumerateChildrenOfType<Paragraph>().ToList();


            Assert.AreEqual(3, paragraphs.Count);
            Assert.AreEqual(3, clonedDocument.EnumerateChildrenOfType<InlineBase>().Count());


            Assert.AreEqual(paragraphBackgroundColor, paragraphs[0].Properties.BackgroundColor.LocalValue);
            Assert.AreEqual(paragraphBackgroundColor1, paragraphs[1].Properties.BackgroundColor.LocalValue);
            Assert.AreEqual(paragraphBackgroundColor1, paragraphs[2].Properties.BackgroundColor.LocalValue);
        }

        [TestMethod]
        public void MailMerge_SourceContainsNullString_FieldIsRemoved()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.InsertField("MERGEFIELD FirstName ", "«FirstName»");

            List<Person> source = new List<Person>()
            {
                new Person() { FirstName = null }
            };

            // Act
            RadFlowDocument mergedDocument = document.MailMerge(source);

            // Assert
            Assert.AreEqual(1, mergedDocument.Sections.Count);
            Assert.AreEqual(1, mergedDocument.Sections.First().Blocks.Count);
            Assert.AreEqual(0, ((Paragraph)mergedDocument.Sections.First().Blocks.First()).Inlines.Count);
        }

        [TestMethod]
        public void MailMerge_SourceContainsEmptyString_FieldIsRemoved()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.InsertField("MERGEFIELD FirstName ", "«FirstName»");

            List<Person> source = new List<Person>()
            {
                new Person() { FirstName = string.Empty }
            };

            // Act
            RadFlowDocument mergedDocument = document.MailMerge(source);

            // Assert
            Assert.AreEqual(1, mergedDocument.Sections.Count);
            Assert.AreEqual(1, mergedDocument.Sections.First().Blocks.Count);
            Assert.AreEqual(0, ((Paragraph)mergedDocument.Sections.First().Blocks.First()).Inlines.Count);
        }

        [TestMethod]
        public void MailMerge_SourceDoesNotContainSuchProperty_FieldIsRemoved()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            editor.InsertField("MERGEFIELD NonExistent ", "«NonExistent»");

            List<Person> source = new List<Person>()
            {
                new Person()
            };

            // Act
            RadFlowDocument mergedDocument = document.MailMerge(source);

            // Assert
            Assert.AreEqual(1, mergedDocument.Sections.Count);
            Assert.AreEqual(1, mergedDocument.Sections.First().Blocks.Count);
            Assert.AreEqual(0, ((Paragraph)mergedDocument.Sections.First().Blocks.First()).Inlines.Count);
        }
    }

    public class Person
    {
        public string FirstName { get; set; }
        public string FamilyName { get; set; }
        public string Company { get; set; }

        public static List<Person> GetList(int personCount)
        {
            List<Person> list = new List<Person>(personCount);

            for (int i = 0; i < personCount; i++)
            {
                Person person = new Person() { FirstName = "Nancy" + i, FamilyName = "Davolio" + i, Company = "Company" + i };
                list.Add(person);
            }

            return list;
        }
    }
}
