﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Flow.Model.Styles;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
#else
using System.Windows.Media;
#endif
#if WPF || SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Flow.Tests.Utilities;

namespace Telerik.Windows.Documents.Flow.Tests.Model.Lists
{
    [TestClass]
    public class ListsTests
    {
        [TestMethod]
        public void AddList_ToDocument()
        {
            RadFlowDocument document = new RadFlowDocument();
            List list = document.Lists.Add(ListTemplateType.NumberedDefault);

            Assert.IsNotNull(list);
            Assert.IsNotNull(list.Document);
            Assert.IsTrue(list.Id > -1);

            for (int i = 0; i < 9; i++)
            {
                Assert.IsNotNull(list.Levels[i]);
                Assert.IsNotNull(list.Levels[i].Document);
            }
        }

        [TestMethod]
        public void AddBuiltInList_ToDocument()
        {
            RadFlowDocument document = new RadFlowDocument();

            List list = document.Lists.Add(ListTemplateType.NumberedDefault);

            Assert.IsNotNull(list);
            Assert.IsNotNull(list.Document);
            Assert.IsTrue(list.Id > -1);

            for (int i = 0; i < 9; i++)
            {
                Assert.IsNotNull(list.Levels[i]);
                Assert.IsNotNull(list.Levels[i].Document);
            }
        }

        [TestMethod]
        public void RemoveList_FromDocument()
        {
            RadFlowDocument document = new RadFlowDocument();

            Assert.IsTrue(document.Lists.Count == 0);

            List list = document.Lists.Add(ListTemplateType.NumberedDefault);
            Assert.IsTrue(document.Lists.Count == 1);

            document.Lists.Remove(list);

            Assert.IsTrue(document.Lists.Count == 0);
        }

        [TestMethod]
        public void RemoveList_WithAssignedParagraphs_FromDocument()
        {
            RadFlowDocument document = new RadFlowDocument();

            Assert.IsTrue(document.Lists.Count == 0);

            List list = document.Lists.Add(ListTemplateType.NumberedDefault);
            Assert.IsTrue(document.Lists.Count == 1);

            Section section = document.Sections.AddSection();
            Paragraph p1 = section.Blocks.AddParagraph();
            p1.ListId = list.Id;
            p1.ListLevel = 0;

            Paragraph p2 = section.Blocks.AddParagraph();
            p2.ListId = list.Id;
            p2.ListLevel = 4;

            Paragraph p3 = section.Blocks.AddParagraph();
            p3.ListId = list.Id;
            p3.ListLevel = 2;

            document.Lists.Remove(list);

            Assert.IsTrue(document.Lists.Count == 0);
            Assert.IsTrue(p1.ListId == Paragraph.ListIdPropertyDefinition.DefaultValue);
            Assert.IsTrue(p1.ListLevel == Paragraph.ListLevelPropertyDefinition.DefaultValue);

            Assert.IsTrue(p2.ListId == Paragraph.ListIdPropertyDefinition.DefaultValue);
            Assert.IsTrue(p2.ListLevel == Paragraph.ListLevelPropertyDefinition.DefaultValue);

            Assert.IsTrue(p3.ListId == Paragraph.ListIdPropertyDefinition.DefaultValue);
            Assert.IsTrue(p3.ListLevel == Paragraph.ListLevelPropertyDefinition.DefaultValue);
        }

        [TestMethod]
        public void RemoveList_WithAssignedStyles_FromDocument()
        {
            RadFlowDocument document = new RadFlowDocument();

            Assert.IsTrue(document.Lists.Count == 0);

            List list = document.Lists.Add(ListTemplateType.NumberedDefault);
            Assert.IsTrue(document.Lists.Count == 1);

            Style style1 = new Style("1", StyleType.Paragraph);
            style1.ParagraphProperties.ListId.LocalValue = list.Id;
            style1.ParagraphProperties.ListLevel.LocalValue = 0;

            Style style2 = new Style("2", StyleType.Paragraph);
            style2.ParagraphProperties.ListId.LocalValue = list.Id;
            style2.ParagraphProperties.ListLevel.LocalValue = 1;

            Style style3 = new Style("3", StyleType.Paragraph);
            style3.ParagraphProperties.ListId.LocalValue = list.Id;
            style3.ParagraphProperties.ListLevel.LocalValue = 2;

            document.StyleRepository.Add(style1);
            document.StyleRepository.Add(style2);
            document.StyleRepository.Add(style3);

            document.Lists.Remove(list);

            Assert.IsTrue(document.Lists.Count == 0);

            Assert.IsNull(style1.ParagraphProperties.ListId.LocalValue);
            Assert.IsNull(style1.ParagraphProperties.ListLevel.LocalValue);
            Assert.IsTrue(style1.ParagraphProperties.ListId.GetActualValue() == Paragraph.ListIdPropertyDefinition.DefaultValue);
            Assert.IsTrue(style1.ParagraphProperties.ListLevel.GetActualValue() == Paragraph.ListLevelPropertyDefinition.DefaultValue);

            Assert.IsNull(style2.ParagraphProperties.ListId.LocalValue);
            Assert.IsNull(style2.ParagraphProperties.ListLevel.LocalValue);
            Assert.IsTrue(style2.ParagraphProperties.ListId.GetActualValue() == Paragraph.ListIdPropertyDefinition.DefaultValue);
            Assert.IsTrue(style2.ParagraphProperties.ListLevel.GetActualValue() == Paragraph.ListLevelPropertyDefinition.DefaultValue);

            Assert.IsNull(style3.ParagraphProperties.ListId.LocalValue);
            Assert.IsNull(style3.ParagraphProperties.ListLevel.LocalValue);
            Assert.IsTrue(style3.ParagraphProperties.ListId.GetActualValue() == Paragraph.ListIdPropertyDefinition.DefaultValue);
            Assert.IsTrue(style3.ParagraphProperties.ListLevel.GetActualValue() == Paragraph.ListLevelPropertyDefinition.DefaultValue);
        }

        [TestMethod]
        public void AssignNumberingStyle_ToList()
        {
            RadFlowDocument document = new RadFlowDocument();

            Assert.IsTrue(document.Lists.Count == 0);

            List list = document.Lists.Add(ListTemplateType.NumberedDefault);
            Assert.IsTrue(document.Lists.Count == 1);

            Style style = new Style("style", StyleType.Numbering);

            document.StyleRepository.Add(style);

            list.StyleId = style.Id;

            Assert.AreEqual(list.StyleId, style.Id);
            Assert.AreEqual(list.Id, style.ParagraphProperties.ListId.LocalValue.Value);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void AssignParagraphStyle_ToList()
        {
            RadFlowDocument document = new RadFlowDocument();

            Assert.IsTrue(document.Lists.Count == 0);

            List list = document.Lists.Add(ListTemplateType.NumberedDefault);
            Assert.IsTrue(document.Lists.Count == 1);

            Style style = new Style("style", StyleType.Paragraph);

            document.StyleRepository.Add(style);

            list.StyleId = style.Id;
        }

        [TestMethod]
        public void Clone_OnListWithModifiedProperties_CloningHasSameProperties()
        {
            // Arrange
            List list = new List();
            list.Levels[1].Alignment = Alignment.Center;
            list.Levels[1].CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily(new FontFamily("Comic Sans MS"));
            list.Levels[1].IsLegal = true;
            list.Levels[1].NumberingStyle = NumberingStyle.BahtText;
            list.Levels[1].NumberTextFormat = "__%1)__";
            list.Levels[1].ParagraphProperties.SpacingBefore.LocalValue = 7;
            list.Levels[1].RestartAfterLevel = 2;
            list.Levels[1].StartIndex = 3;
            list.Levels[1].StyleId = "test";
            list.Levels[3].Alignment = Alignment.Center;
            list.Levels[3].CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily(new FontFamily("Verdana"));
            list.Levels[3].IsLegal = true;
            list.Levels[3].NumberingStyle = NumberingStyle.BahtText;
            list.Levels[3].NumberTextFormat = "__%1)__";
            list.Levels[3].ParagraphProperties.SpacingBefore.LocalValue = 8;
            list.Levels[3].RestartAfterLevel = 3;
            list.Levels[3].StartIndex = 4;
            list.Levels[3].StyleId = "test1";
            list.MultilevelType = MultilevelType.Multilevel;
            list.StyleId = "test2";

            // Act
            List clone = list.Clone();

            // Assert
            StylePropertyAsserter.AssertListsAreEqual(list, clone);

            // Verify (partially) this is deep copy.
            Assert.AreNotSame(list.Levels, clone.Levels);
            Assert.AreNotSame(list.Levels[1], clone.Levels[1]);
            Assert.AreNotSame(list.Levels[1].CharacterProperties, clone.Levels[1].CharacterProperties);
            Assert.AreNotSame(list.Levels[1].ParagraphProperties, clone.Levels[1].ParagraphProperties);
        }
    }
}
