﻿using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
using Telerik.Windows.Documents.Flow.Model.Watermarks;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Media;
#else
using System.Windows.Media;
#endif

namespace Telerik.Windows.Documents.Flow.Tests.Model
{
    [TestClass]
    public class MergeDocumentsTests
    {
        [TestMethod]
        public void Merge_WithoutParameters_AddsSectionsFromTheSourceDocumentAfterTheSectionsOfTheTargetDocument()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            document1.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("first");

            RadFlowDocument document2 = new RadFlowDocument();
            document2.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("second");

            // Act
            document1.Merge(document2);

            // Assert
            List<Section> sections = document1.EnumerateChildrenOfType<Section>().ToList();
            Assert.AreEqual(2, document1.Sections.Count);
            StructureAsserter.AssertSectionsAreEqual(document1.Sections[0], sections[0]);
            StructureAsserter.AssertSectionsAreEqual(document2.Sections[0], sections[1]);
        }

        [TestMethod]
        public void Merge_UseTargetStyleAndSourceHasEqualStyle_SourceStyleIsNotAdded()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Style style1 = new Style("Style", StyleType.Character);
            style1.CharacterProperties.FontSize.LocalValue = 30;
            document1.StyleRepository.Add(style1);

            int stylesCountBefore = document1.StyleRepository.Styles.Count();

            RadFlowDocument document2 = new RadFlowDocument();
            Style style2 = new Style("Style", StyleType.Character);
            style2.CharacterProperties.FontSize.LocalValue = 30;
            document2.StyleRepository.Add(style2);

            MergeOptions mergeOptions = new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.UseTargetStyle };

            // Act
            document1.Merge(document2, mergeOptions);

            // Assert
            Assert.AreEqual(stylesCountBefore, document1.StyleRepository.Styles.Count());
            Assert.IsTrue(document1.StyleRepository.Contains("Style"));
            Assert.AreEqual(30, document1.StyleRepository.GetStyle("Style").CharacterProperties.FontSize.LocalValue.Value);
        }

        [TestMethod]
        public void Merge_UseTargetStyleAndSourceHasStyleWithSameIdAndDifferentProperties_SourceStyleIsNotAdded()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Style style1 = new Style("Style", StyleType.Character);
            style1.CharacterProperties.FontSize.LocalValue = 20;
            document1.StyleRepository.Add(style1);

            int stylesCountBefore = document1.StyleRepository.Styles.Count();

            RadFlowDocument document2 = new RadFlowDocument();
            Style style2 = new Style("Style", StyleType.Character);
            style2.CharacterProperties.FontSize.LocalValue = 30;
            document2.StyleRepository.Add(style2);

            MergeOptions mergeOptions = new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.UseTargetStyle };

            // Act
            document1.Merge(document2, mergeOptions);

            // Assert
            Assert.AreEqual(stylesCountBefore, document1.StyleRepository.Styles.Count());
            Assert.IsTrue(document1.StyleRepository.Contains("Style"));
            Assert.AreEqual(20, document1.StyleRepository.GetStyle("Style").CharacterProperties.FontSize.LocalValue.Value);
        }

        [TestMethod]
        public void Merge_UseTargetStyleAndSourceHasStyleWithSameIdButDifferentType_SourceStyleIsAdded()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Style style1 = new Style("Style", StyleType.Paragraph);
            style1.CharacterProperties.FontSize.LocalValue = 30;
            document1.StyleRepository.Add(style1);

            int stylesCountBefore = document1.StyleRepository.Styles.Count();

            RadFlowDocument document2 = new RadFlowDocument();
            Style style2 = new Style("Style", StyleType.Character);
            style2.CharacterProperties.FontSize.LocalValue = 30;
            document2.StyleRepository.Add(style2);

            MergeOptions mergeOptions = new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.UseTargetStyle };

            // Act
            document1.Merge(document2, mergeOptions);

            // Assert
            Assert.AreEqual(stylesCountBefore + 1, document1.StyleRepository.Styles.Count());

            Assert.IsTrue(document1.StyleRepository.Contains("Style"));
            Assert.AreEqual(StyleType.Paragraph, document1.StyleRepository.GetStyle("Style").StyleType);

            Assert.IsTrue(document1.StyleRepository.Contains("Style_1"));
            Assert.AreEqual(StyleType.Character, document1.StyleRepository.GetStyle("Style_1").StyleType);
        }

        [TestMethod]
        public void Merge_UseTargetStyleAndSourceHasDifferentStyle_SourceStyleIsAdded()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Style style1 = new Style("Style1", StyleType.Character);
            style1.CharacterProperties.FontSize.LocalValue = 30;
            document1.StyleRepository.Add(style1);

            int stylesCountBefore = document1.StyleRepository.Styles.Count();

            RadFlowDocument document2 = new RadFlowDocument();
            Style style2 = new Style("Style2", StyleType.Character);
            style2.CharacterProperties.FontSize.LocalValue = 30;
            document2.StyleRepository.Add(style2);

            MergeOptions mergeOptions = new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.UseTargetStyle };

            // Act
            document1.Merge(document2, mergeOptions);

            // Assert
            Assert.AreEqual(stylesCountBefore + 1, document1.StyleRepository.Styles.Count());
            Assert.IsTrue(document1.StyleRepository.Contains("Style1"));
            Assert.IsTrue(document1.StyleRepository.Contains("Style2"));
        }

        [TestMethod]
        public void Merge_RenameSourceStyleAndSourceHasEqualStyle_SourceStyleIsNotAdded()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Style style1 = new Style("Style", StyleType.Character);
            style1.CharacterProperties.FontSize.LocalValue = 30;
            document1.StyleRepository.Add(style1);

            int stylesCountBefore = document1.StyleRepository.Styles.Count();

            RadFlowDocument document2 = new RadFlowDocument();
            Style style2 = new Style("Style", StyleType.Character);
            style2.CharacterProperties.FontSize.LocalValue = 30;
            document2.StyleRepository.Add(style2);

            MergeOptions mergeOptions = new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle };

            // Act
            document1.Merge(document2, mergeOptions);

            // Assert
            Assert.AreEqual(stylesCountBefore, document1.StyleRepository.Styles.Count());
            Assert.IsTrue(document1.StyleRepository.Contains("Style"));
            Assert.AreEqual(30, document1.StyleRepository.GetStyle("Style").CharacterProperties.FontSize.LocalValue.Value);
        }

        [TestMethod]
        public void Merge_RenameSourceStyleAndSourceHasStyleWithSameIdAndDifferentProperties_SourceStyleIsAddedAndRenamed()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Style style1 = new Style("Style", StyleType.Character);
            style1.CharacterProperties.FontSize.LocalValue = 20;
            document1.StyleRepository.Add(style1);

            RadFlowDocument document2 = new RadFlowDocument();
            Style style2 = new Style("Style", StyleType.Character);
            style2.CharacterProperties.FontSize.LocalValue = 30;
            document2.StyleRepository.Add(style2);

            MergeOptions mergeOptions = new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle };

            // Act
            document1.Merge(document2, mergeOptions);

            // Assert
            Assert.IsTrue(document1.StyleRepository.Contains("Style"));
            Assert.AreEqual(20, document1.StyleRepository.GetStyle("Style").CharacterProperties.FontSize.LocalValue.Value);

            Assert.IsTrue(document1.StyleRepository.Contains("Style_1"));
            Assert.AreEqual(30, document1.StyleRepository.GetStyle("Style_1").CharacterProperties.FontSize.LocalValue.Value);
        }

        [TestMethod]
        public void Merge_RenameSourceStyleAndSourceHasStyleWithSameIdAndDifferentPropertiesAndStyleWithRenamedId_SourceStyleIsAddedAndRenamedToAvoidCollision()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Style style = new Style("Style", StyleType.Character);
            style.CharacterProperties.FontSize.LocalValue = 20;
            document1.StyleRepository.Add(style);

            Style style1 = new Style("Style_1", StyleType.Character);
            document1.StyleRepository.Add(style1);

            RadFlowDocument document2 = new RadFlowDocument();
            Style style2 = new Style("Style", StyleType.Character);
            style2.CharacterProperties.FontSize.LocalValue = 30;
            document2.StyleRepository.Add(style2);

            MergeOptions mergeOptions = new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle };

            // Act
            document1.Merge(document2, mergeOptions);

            // Assert
            Assert.IsTrue(document1.StyleRepository.Contains("Style"));
            Assert.AreEqual(20, document1.StyleRepository.GetStyle("Style").CharacterProperties.FontSize.LocalValue.Value);

            Assert.IsTrue(document1.StyleRepository.Contains("Style_1"));

            Assert.IsTrue(document1.StyleRepository.Contains("Style_1_1"));
            Assert.AreEqual(30, document1.StyleRepository.GetStyle("Style_1_1").CharacterProperties.FontSize.LocalValue.Value);
        }

        [TestMethod]
        public void Merge_RenameSourceStyleAndSourceHasDifferentStyle_SourceStyleIsAdded()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Style style1 = new Style("Style1", StyleType.Character);
            style1.CharacterProperties.FontSize.LocalValue = 30;
            document1.StyleRepository.Add(style1);

            int stylesCountBefore = document1.StyleRepository.Styles.Count();

            RadFlowDocument document2 = new RadFlowDocument();
            Style style2 = new Style("Style2", StyleType.Character);
            style2.CharacterProperties.FontSize.LocalValue = 30;
            document2.StyleRepository.Add(style2);

            MergeOptions mergeOptions = new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle };

            // Act
            document1.Merge(document2, mergeOptions);

            // Assert
            Assert.AreEqual(stylesCountBefore + 1, document1.StyleRepository.Styles.Count());
            Assert.IsTrue(document1.StyleRepository.Contains("Style1"));
            Assert.IsTrue(document1.StyleRepository.Contains("Style2"));
        }

        [TestMethod]
        public void Merge_WhenRenamedStylesIsAdded_StyleIdsOfParagraphsAreUpdated()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Style style1 = new Style("Style", StyleType.Paragraph);
            style1.ParagraphProperties.SpacingAfter.LocalValue = 20;
            document1.StyleRepository.Add(style1);
            document1.Sections.AddSection().Blocks.AddParagraph().StyleId = style1.Id;

            RadFlowDocument document2 = new RadFlowDocument();
            Style style2 = new Style("Style", StyleType.Paragraph);
            style2.ParagraphProperties.SpacingAfter.LocalValue = 30;
            document2.StyleRepository.Add(style2);
            document2.Sections.AddSection().Blocks.AddParagraph().StyleId = style2.Id;

            MergeOptions mergeOptions = new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle };

            // Act
            document1.Merge(document2, mergeOptions);

            // Assert
            List<Paragraph> mergedParagraphs = document1.EnumerateChildrenOfType<Paragraph>().ToList();
            Assert.AreEqual("Style", mergedParagraphs[0].StyleId);
            Assert.AreEqual("Style_1", mergedParagraphs[1].StyleId);
        }

        [TestMethod]
        public void Merge_WhenRenamedStylesIsAdded_StyleIdsOfTablesAreUpdated()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Style style1 = new Style("Style", StyleType.Table);
            style1.TableProperties.Indent.LocalValue = 20;
            document1.StyleRepository.Add(style1);
            document1.Sections.AddSection().Blocks.AddTable().StyleId = style1.Id;

            RadFlowDocument document2 = new RadFlowDocument();
            Style style2 = new Style("Style", StyleType.Table);
            style2.TableProperties.Indent.LocalValue = 30;
            document2.StyleRepository.Add(style2);
            document2.Sections.AddSection().Blocks.AddTable().StyleId = style2.Id;

            MergeOptions mergeOptions = new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle };

            // Act
            document1.Merge(document2, mergeOptions);

            // Assert
            List<Table> mergedTables = document1.EnumerateChildrenOfType<Table>().ToList();
            Assert.AreEqual("Style", mergedTables[0].StyleId);
            Assert.AreEqual("Style_1", mergedTables[1].StyleId);
        }

        [TestMethod]
        public void Merge_WhenRenamedStyleIsAdded_StyleIdsOfRunsAreUpdated()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Style style1 = new Style("Style", StyleType.Character);
            style1.CharacterProperties.FontSize.LocalValue = 20;
            document1.StyleRepository.Add(style1);
            document1.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("test").StyleId = style1.Id;

            RadFlowDocument document2 = new RadFlowDocument();
            Style style2 = new Style("Style", StyleType.Character);
            style2.CharacterProperties.FontSize.LocalValue = 30;
            document2.StyleRepository.Add(style2);
            document2.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("test").StyleId = style2.Id;

            MergeOptions mergeOptions = new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle };

            // Act
            document1.Merge(document2, mergeOptions);

            // Assert
            List<Run> mergedRuns = document1.EnumerateChildrenOfType<Run>().ToList();
            Assert.AreEqual("Style", mergedRuns[0].StyleId);
            Assert.AreEqual("Style_1", mergedRuns[1].StyleId);
        }

        [TestMethod]
        public void Merge_WhenRenamedStyleIsAdded_LinkedStyleIdsOfStyleLinkedToRenamedOneIsUpdated()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Style style1 = new Style("Style", StyleType.Character);
            style1.CharacterProperties.FontSize.LocalValue = 20;
            document1.StyleRepository.Add(style1);

            RadFlowDocument document2 = new RadFlowDocument();
            Style style2 = new Style("Style", StyleType.Character);
            style2.CharacterProperties.FontSize.LocalValue = 30;
            document2.StyleRepository.Add(style2);

            Style linkedStyle = new Style("Linked", StyleType.Paragraph);
            linkedStyle.LinkedStyleId = "Style";
            document2.StyleRepository.Add(linkedStyle);

            MergeOptions mergeOptions = new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle };

            // Act
            document1.Merge(document2, mergeOptions);

            // Assert
            Assert.AreEqual("Style_1", document1.StyleRepository.GetStyle("Linked").LinkedStyleId);
        }

        [TestMethod]
        public void Merge_WhenRenamedStyleIsAdded_BasedOnStyleIdsOfStylesBasedOnRenamedOneIsUpdated()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Style style1 = new Style("Style", StyleType.Character);
            style1.CharacterProperties.FontSize.LocalValue = 20;
            document1.StyleRepository.Add(style1);

            RadFlowDocument document2 = new RadFlowDocument();
            Style style2 = new Style("Style", StyleType.Character);
            style2.CharacterProperties.FontSize.LocalValue = 30;
            document2.StyleRepository.Add(style2);

            Style inheritedStyle = new Style("Inherited", StyleType.Character);
            inheritedStyle.BasedOnStyleId = "Style";
            document2.StyleRepository.Add(inheritedStyle);

            MergeOptions mergeOptions = new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle };

            // Act
            document1.Merge(document2, mergeOptions);

            // Assert
            Assert.AreEqual("Style_1", document1.StyleRepository.GetStyle("Inherited").BasedOnStyleId);
        }

        [TestMethod]
        public void Merge_WhenRenamedStyleIsAdded_NextStyleIdsOfConnectedStyleIsUpdated()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Style style1 = new Style("Style", StyleType.Character);
            style1.CharacterProperties.FontSize.LocalValue = 20;
            document1.StyleRepository.Add(style1);

            RadFlowDocument document2 = new RadFlowDocument();
            Style style2 = new Style("Style", StyleType.Character);
            style2.CharacterProperties.FontSize.LocalValue = 30;
            document2.StyleRepository.Add(style2);

            Style previousStyle = new Style("Previous", StyleType.Character);
            previousStyle.NextStyleId = "Style";
            document2.StyleRepository.Add(previousStyle);

            MergeOptions mergeOptions = new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle };

            // Act
            document1.Merge(document2, mergeOptions);

            // Assert
            Assert.AreEqual("Style_1", document1.StyleRepository.GetStyle("Previous").NextStyleId);
        }

        [TestMethod]
        public void Merge_WhenDefaultStyleIsAdded_IsDefaultOfSourceStyleIsSetToFalse()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Style style1 = new Style("Style1", StyleType.Character);
            style1.IsDefault = true;
            style1.CharacterProperties.FontSize.LocalValue = 30;
            document1.StyleRepository.Add(style1);

            RadFlowDocument document2 = new RadFlowDocument();
            Style style2 = new Style("Style2", StyleType.Character);
            style2.IsDefault = true;
            style2.CharacterProperties.FontSize.LocalValue = 30;
            document2.StyleRepository.Add(style2);

            MergeOptions mergeOptions = new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle };

            // Act
            document1.Merge(document2, mergeOptions);

            // Assert
            Assert.AreEqual("Style1", document1.StyleRepository.GetDefaultStyle(StyleType.Character).Id);
            Assert.IsFalse(document1.StyleRepository.GetStyle("Style2").IsDefault);
        }

        [TestMethod]
        public void Merge_WhenDifferentSimpleListsAreAdded()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            List list1 = document1.Lists.Add(ListTemplateType.NumberedDefault);

            Paragraph p1 = document1.Sections.AddSection().Blocks.AddParagraph();
            p1.ListId = list1.Id;
            p1.ListLevel = 2;

            RadFlowDocument document2 = new RadFlowDocument();
            List list2 = document1.Lists.Add(ListTemplateType.NumberedDefault);
            list2.Levels[0].StartIndex = 4;

            Paragraph p2 = document2.Sections.AddSection().Blocks.AddParagraph();
            p2.ListId = list2.Id;
            p2.ListLevel = 3;

            // Act
            document1.Merge(document2);

            // Assert
            Assert.AreEqual(document1.Lists.Count, 2);

            List<Paragraph> paragraphs = document1.EnumerateChildrenOfType<Paragraph>().ToList();

            List<List> lists = document1.Lists.ToList();

            Assert.AreEqual(paragraphs[0].ListId, lists[0].Id);
            Assert.AreEqual(paragraphs[0].ListLevel, 2);


            Assert.AreEqual(paragraphs[1].ListId, lists[1].Id);
            Assert.AreEqual(paragraphs[1].ListLevel, 3);
        }

        [TestMethod]
        public void Merge_WhenMultipleParagraphStylesAreAddedAndRelatedToExistingList_ListIdsInStylesAreUpdated()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            document1.Lists.Add(ListTemplateType.NumberedDefault);

            RadFlowDocument document2 = new RadFlowDocument();
            document2.Lists.Add(ListTemplateType.BulletDefault);
            List list2 = document2.Lists.Add(ListTemplateType.NumberedDefault);

            Style style1 = new Style("style1", StyleType.Paragraph);
            style1.ParagraphProperties.ListId.LocalValue = list2.Id;
            document2.StyleRepository.Add(style1);

            Style style2 = new Style("style2", StyleType.Paragraph);
            style2.ParagraphProperties.ListId.LocalValue = list2.Id;
            document2.StyleRepository.Add(style2);

            // Act
            document1.Merge(document2);

            // Assert
            Assert.AreEqual(document1.Lists.Count, 3);

            Style newStyle1 = document1.StyleRepository.GetStyle(style1.Id);
            Style newStyle2 = document1.StyleRepository.GetStyle(style2.Id);

            Assert.AreEqual(newStyle1.ParagraphProperties.ListId.LocalValue.Value, document1.Lists.LastOrDefault().Id);
            Assert.AreEqual(newStyle2.ParagraphProperties.ListId.LocalValue.Value, document1.Lists.LastOrDefault().Id);
        }

        [TestMethod]
        public void Merge_WhenMultipleExistingParagraphStylesAreAddedAndRelatedToList_ListIdsInStylesAreUpdated()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            List list1 = document1.Lists.Add(ListTemplateType.NumberedDefault);

            Style style1Existing = new Style("style1", StyleType.Paragraph);
            style1Existing.ParagraphProperties.ListId.LocalValue = list1.Id;
            document1.StyleRepository.Add(style1Existing);

            Style style2Existing = new Style("style2", StyleType.Paragraph);
            style2Existing.ParagraphProperties.ListId.LocalValue = list1.Id;
            document1.StyleRepository.Add(style2Existing);

            RadFlowDocument document2 = new RadFlowDocument();
            document2.Lists.Add(ListTemplateType.BulletDefault);
            List list2 = document2.Lists.Add(ListTemplateType.NumberedDefault);
            list2.Levels[0].StartIndex = 4;

            Style style1 = new Style("style1", StyleType.Paragraph);
            style1.ParagraphProperties.ListId.LocalValue = list2.Id;
            document2.StyleRepository.Add(style1);

            Style style2 = new Style("style2", StyleType.Paragraph);
            style2.ParagraphProperties.ListId.LocalValue = list2.Id;
            document2.StyleRepository.Add(style2);

            // Act
            document1.Merge(
                document2,
                new MergeOptions() { ConflictingStylesResolutionMode = Flow.Model.ConflictingStylesResolutionMode.RenameSourceStyle });

            // Assert
            Assert.AreEqual(document1.Lists.Count, 3);

            Style newStyle1 = document1.StyleRepository.GetStyle(style1.Id + "_1");
            Style newStyle2 = document1.StyleRepository.GetStyle(style2.Id + "_1");

            List mergedList = document1.Lists.LastOrDefault();
            Assert.AreEqual(mergedList.Levels[0].StartIndex, list2.Levels[0].StartIndex);

            Assert.AreEqual(newStyle1.ParagraphProperties.ListId.LocalValue.Value, mergedList.Id);
            Assert.AreEqual(newStyle2.ParagraphProperties.ListId.LocalValue.Value, mergedList.Id);
        }

        [TestMethod]
        public void Merge_WhenNumberingStyleIsAdded_StyleIdInListIsUpdated_ListIdInStyleIsUpdated()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            document1.Lists.Add(ListTemplateType.BulletDefault);

            RadFlowDocument document2 = new RadFlowDocument();
            List list1 = document2.Lists.Add(ListTemplateType.NumberedDefault);

            Style style = new Style("style1", StyleType.Numbering);
            style.ParagraphProperties.ListId.LocalValue = list1.Id;
            document2.StyleRepository.Add(style);

            list1.StyleId = style.Id;
            style.ParagraphProperties.ListId.LocalValue = list1.Id;

            // Act
            document1.Merge(document2);

            // Assert
            Assert.AreEqual(document1.Lists.Count, 2);

            List mergedList = document1.Lists.LastOrDefault();
            Assert.IsNotNull(mergedList);

            Style mergedStyle = document1.StyleRepository.GetStyle(style.Id);
            Assert.IsNotNull(mergedStyle);

            Assert.AreEqual(mergedList.StyleId, mergedStyle.Id);
            Assert.AreEqual(mergedList.Id, mergedStyle.ParagraphProperties.ListId.LocalValue.Value);
        }

        [TestMethod]
        public void Merge_WhenExistingNumberingStyleIsAdded_StyleIdInListIsUpdated_ListIdInStyleIsUpdated()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            List listExisting = document1.Lists.Add(ListTemplateType.BulletDefault);

            Style styleExisting = new Style("style1", StyleType.Numbering);
            document1.StyleRepository.Add(styleExisting);

            listExisting.StyleId = styleExisting.Id;
            styleExisting.ParagraphProperties.ListId.LocalValue = listExisting.Id;

            RadFlowDocument document2 = new RadFlowDocument();
            List list1 = document2.Lists.Add(ListTemplateType.NumberedDefault);

            Style style = new Style("style1", StyleType.Numbering);
            style.ParagraphProperties.ListId.LocalValue = list1.Id;
            document2.StyleRepository.Add(style);

            list1.StyleId = style.Id;
            style.ParagraphProperties.ListId.LocalValue = list1.Id;

            // Act
            document1.Merge(document2, new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle });

            // Assert
            Assert.AreEqual(document1.Lists.Count, 2);

            List mergedList = document1.Lists.LastOrDefault();
            Assert.IsNotNull(mergedList);

            Style mergedStyle = document1.StyleRepository.GetStyle(style.Id + "_1");
            Assert.IsNotNull(mergedStyle);

            Assert.AreEqual(mergedList.StyleId, mergedStyle.Id);
            Assert.AreEqual(mergedList.Id, mergedStyle.ParagraphProperties.ListId.LocalValue.Value);
        }

        [TestMethod]
        public void Merge_WhenExistingListIsAddedWithLevelRelatedToExistingHeadings_UpdatesOfAllIds()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();

            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1));
            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(2));
            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(3));
            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(4));
            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(5));
            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(6));
            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(7));
            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(8));
            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(9));

            List listExisting = document1.Lists.Add(ListTemplateType.BulletDefault);
            for (int i = 0; i < listExisting.Levels.Count; i++)
            {
                listExisting.Levels[i].StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1);
            }

            RadFlowDocument document2 = new RadFlowDocument();

            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1));
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(2));
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(3));
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(4));
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(5));
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(6));
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(7));
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(8));
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(9));

            List list1 = document2.Lists.Add(ListTemplateType.NumberedDefault);
            for (int i = 0; i < listExisting.Levels.Count; i++)
            {
                list1.Levels[i].StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1);
            }

            // Act
            document1.Merge(document2, new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle });

            // Assert
            Assert.AreEqual(document1.Lists.Count, 2);

            List firstList = document1.Lists.FirstOrDefault();
            for (int i = 0; i < firstList.Levels.Count; i++)
            {
                Assert.IsNull(document1.StyleRepository.GetStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1) + "_1"));

                Style style = document1.StyleRepository.GetStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1));

                Assert.AreEqual(firstList.Levels[i].StyleId, style.Id);
                Assert.AreEqual(style.ParagraphProperties.ListId.LocalValue, firstList.Id);
                Assert.AreEqual(style.ParagraphProperties.ListLevel.LocalValue, i);
            }

            List secondList = document1.Lists.LastOrDefault();
            for (int i = 0; i < secondList.Levels.Count; i++)
            {
                Style style = document1.StyleRepository.GetStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1));

                Assert.AreEqual(secondList.Levels[i].StyleId, style.Id);
                Assert.AreNotEqual(style.ParagraphProperties.ListId.LocalValue, secondList.Id);
                Assert.AreEqual(style.ParagraphProperties.ListLevel.LocalValue, i);
            }
        }

        [TestMethod]
        public void Merge_WhenExistingRelatedToStyleListIsAddedWithLevelRelatedToExistingHeadings_UpdatesOfAllIds()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();

            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1));
            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(2));
            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(3));
            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(4));
            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(5));
            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(6));
            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(7));
            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(8));
            document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(9));

            List listExisting = document1.Lists.Add(ListTemplateType.BulletDefault);
            Style styleExisting = new Style("styleExisting", StyleType.Numbering);
            document1.StyleRepository.Add(styleExisting);

            listExisting.StyleId = styleExisting.Id;
            styleExisting.ParagraphProperties.ListId.LocalValue = listExisting.Id;

            for (int i = 0; i < listExisting.Levels.Count; i++)
            {
                listExisting.Levels[i].StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1);
            }

            RadFlowDocument document2 = new RadFlowDocument();

            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1));
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(2));
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(3));
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(4));
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(5));
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(6));
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(7));
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(8));
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(9));

            List list1 = document2.Lists.Add(ListTemplateType.NumberedDefault);
            Style style1 = new Style("style1", StyleType.Numbering);
            document2.StyleRepository.Add(style1);

            list1.StyleId = style1.Id;
            style1.ParagraphProperties.ListId.LocalValue = list1.Id;

            for (int i = 0; i < listExisting.Levels.Count; i++)
            {
                list1.Levels[i].StyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1);
            }

            // Act
            document1.Merge(document2, new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle });

            // Assert
            Assert.AreEqual(document1.Lists.Count, 2);

            List firstList = document1.Lists.FirstOrDefault();
            for (int i = 0; i < firstList.Levels.Count; i++)
            {
                Assert.IsNull(document1.StyleRepository.GetStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1) + "_1"));

                Style style = document1.StyleRepository.GetStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1));

                Assert.AreEqual(firstList.Levels[i].StyleId, style.Id);
                Assert.AreEqual(style.ParagraphProperties.ListId.LocalValue, firstList.Id);
                Assert.AreEqual(style.ParagraphProperties.ListLevel.LocalValue, i);
            }

            List secondList = document1.Lists.LastOrDefault();
            for (int i = 0; i < secondList.Levels.Count; i++)
            {
                Style style = document1.StyleRepository.GetStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(i + 1));

                Assert.AreEqual(secondList.Levels[i].StyleId, style.Id);
                Assert.AreNotEqual(style.ParagraphProperties.ListId.LocalValue, secondList.Id);
                Assert.AreEqual(style.ParagraphProperties.ListLevel.LocalValue, i);
            }
        }

        [TestMethod]
        public void Merge_WhenExistingStyleIsRelatedToListAndOnTheSourceIsNot_AddBothOfStyles()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            List list = document1.Lists.Add(ListTemplateType.NumberedDefault);

            Style heading1 = document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1));

            heading1.ParagraphProperties.ListId.LocalValue = list.Id;

            RadFlowDocument document2 = new RadFlowDocument();
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1));

            // Act
            document1.Merge(document2, new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle });

            // Assert
            Assert.AreEqual(document1.Lists.Count, 1);

            Assert.IsNotNull(document1.StyleRepository.GetStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1)));
            Assert.IsNotNull(document1.StyleRepository.GetStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1) + "_1"));
        }

        [TestMethod]
        public void Merge_WhenExistingStyleIsRelatedToListLevelAndOnTheSourceIsNot_AddBothOfStyles()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            List list = document1.Lists.Add(ListTemplateType.NumberedDefault);

            Style heading1 = document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1));
            heading1.ParagraphProperties.ListId.LocalValue = list.Id;
            heading1.ParagraphProperties.ListLevel.LocalValue = 1;

            RadFlowDocument document2 = new RadFlowDocument();
            document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1));

            // Act
            document1.Merge(document2, new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle });

            // Assert
            Assert.AreEqual(document1.Lists.Count, 1);

            Assert.IsNotNull(document1.StyleRepository.GetStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1)));
            Assert.IsNotNull(document1.StyleRepository.GetStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1) + "_1"));
        }

        [TestMethod]
        public void Merge_WhenExistingStylesAreRelatedToListLevels_AddBothOfStyles()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            List existingList = document1.Lists.Add(ListTemplateType.NumberedDefault);

            Style existingHeading1 = document1.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1));
            existingHeading1.ParagraphProperties.ListId.LocalValue = existingList.Id;
            existingHeading1.ParagraphProperties.ListLevel.LocalValue = 1;

            RadFlowDocument document2 = new RadFlowDocument();
            List list2 = document2.Lists.Add(ListTemplateType.NumberedDefault);

            Style heading1 = document2.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1));
            heading1.ParagraphProperties.ListId.LocalValue = list2.Id;
            heading1.ParagraphProperties.ListLevel.LocalValue = 3;

            // Act
            document1.Merge(document2, new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle });

            // Assert
            Assert.AreEqual(document1.Lists.Count, 2);

            Assert.IsNotNull(document1.StyleRepository.GetStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1)));
            Assert.IsNotNull(document1.StyleRepository.GetStyle(BuiltInStyleNames.GetHeadingStyleIdByIndex(1) + "_1"));
        }

        [TestMethod]
        public void Merge_WhenBuiltInStyleIsRenamed_RenamedStyleHasSetIsCustomTrue()
        {
            // Arrange
            string headingStyleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(1);

            RadFlowDocument targetDocument = new RadFlowDocument();
            targetDocument.StyleRepository.AddBuiltInStyle(headingStyleId);

            RadFlowDocument sourceDocument = new RadFlowDocument();
            sourceDocument.StyleRepository.AddBuiltInStyle(headingStyleId);
            Style headingStyleInSourceDocument = sourceDocument.StyleRepository.GetStyle(headingStyleId);
            headingStyleInSourceDocument.CharacterProperties.FontSize.LocalValue = 30;

            MergeOptions mergeOptions = new MergeOptions() { ConflictingStylesResolutionMode = ConflictingStylesResolutionMode.RenameSourceStyle };

            // Act
            targetDocument.Merge(sourceDocument, mergeOptions);

            // Assert
            Assert.IsTrue(targetDocument.StyleRepository.Contains(headingStyleId));
            Assert.IsFalse(targetDocument.StyleRepository.GetStyle(headingStyleId).IsCustom);

            string renamedHeadingStyleId = headingStyleId + "_1";
            Assert.IsTrue(targetDocument.StyleRepository.Contains(renamedHeadingStyleId));
            Assert.IsTrue(targetDocument.StyleRepository.GetStyle(renamedHeadingStyleId).IsCustom);
        }

        [TestMethod]
        public void Merge_WhenSourceCollectionIsUsedWatermarks_TheyShouldBeTransferedToTargetCollection()
        {
            RadFlowDocument targetDocument = new RadFlowDocument();

            RadFlowDocument sourceDocument = new RadFlowDocument();

            sourceDocument.Sections.AddSection();

            TextWatermarkSettings settings = new TextWatermarkSettings()
            {
                Angle = 12,
                Width = 200,
                Height = 300,
                Opacity = 0.4,
                FontFamily = new FontFamily("Verdana"),
                ForegroundColor = Colors.Red,
                Text = "DRAFT"
            };

            Watermark textWatermark = new Watermark(settings);

            Header header = sourceDocument.Sections.First().Headers.Add(HeaderFooterType.Default);
            header.Watermarks.Add(textWatermark);

            targetDocument.Merge(sourceDocument);

            Assert.AreEqual(1, targetDocument.Sections.First().Headers.Default.Watermarks.Count());
        }
    }
}
