﻿using System;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Model.Fields;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
#if NETSTANDARD
using Telerik.Documents.Media;
#else
using System.Windows.Media;
#endif
#if WPF || SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Flow.Tests.Model
{
    [TestClass]
    public class ParagraphTests
    {
        [TestMethod]
        public void CloneWithParameter_WhenInvokedWithNull_ThrowsArgumentNullException()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Paragraph paragraph = new Paragraph(document1);

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentNullException>(() => paragraph.Clone(null));
        }

        [TestMethod]
        public void Clone_WithDocument_ClonesTheParagraphForTheDocument()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Paragraph paragraph = new Paragraph(document1);

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            Paragraph clonedParagraph = paragraph.Clone(document2);

            // Assert
            Assert.AreNotSame(paragraph, clonedParagraph);
            Assert.AreSame(document2, clonedParagraph.Document);
        }

        [TestMethod]
        public void Clone_WithoutDocument_ClonesTheParagraphForCurrentDocument()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Paragraph paragraph = new Paragraph(document);

            // Act
            Paragraph clonedParagraph = paragraph.Clone();

            // Assert
            Assert.AreNotSame(paragraph, clonedParagraph);
            Assert.AreSame(document, clonedParagraph.Document);
        }

        [TestMethod]
        public void Clone_WhenContainsChildFieldCharacters_ClonesTheParagraphAndAllFields()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();

            Paragraph paragraph = new Paragraph(document1);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document1);

            editor.MoveToParagraphStart(paragraph);
            editor.InsertText("Before ");
            FieldInfo field = editor.InsertField("AUTHOR \\* MERGEFORMAT", "Author Name");
            field.IsDirty = true;
            field.IsLocked = true;
            editor.InsertText(" after");

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            Paragraph clonedParagraph = paragraph.Clone(document2);

            // Assert
            StructureAsserter.AssertParagraphsAreEqual(paragraph, clonedParagraph);

            FieldInfo clonedField = clonedParagraph.EnumerateChildrenOfType<FieldCharacter>().First().FieldInfo;
            Assert.IsTrue(clonedParagraph.EnumerateChildrenOfType<FieldCharacter>().All(fieldCharacter => fieldCharacter.FieldInfo == clonedField), "Field characters reference incorrect field.");
        }

        [TestMethod]
        public void Clone_WhenPropertiesAreSet_CloningHasSameProperties()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();

            Paragraph paragraph = new Paragraph(document1);
            paragraph.AllowOverflowPunctuation = false;
            paragraph.ApplyEastAsianLineBreakingRules = false;

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            Paragraph clonedParagraph = paragraph.Clone(document2);

            // Assert
            StylePropertyAsserter.AssertPropertiesAreEqual(paragraph.Properties, clonedParagraph.Properties);
        }

        [TestMethod]
        public void Clone_WhenHasFieldStartedInPreviousParagraph_RemovesUnpairedFieldCharacters()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            document.Sections.AddSection().Blocks.AddParagraph();

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            editor.InsertHyperlink("John \r\nDoe", "www.telerik.com", false);

            Paragraph secondParagraph = (Paragraph)document.Sections[0].Blocks[1];

            // Act
            Paragraph clonedParagraph = secondParagraph.Clone();

            // Assert
            Assert.IsFalse(clonedParagraph.EnumerateChildrenOfType<FieldCharacter>().Any());
        }

        [TestMethod]
        public void Clone_WhenHasFieldEndingInFollowingParagraph_RemovesUnpairedFieldCharacters()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            document.Sections.AddSection().Blocks.AddParagraph();

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            editor.InsertHyperlink("John \r\nDoe", "www.telerik.com", false);

            Paragraph firstParagraph = (Paragraph)document.Sections[0].Blocks[0];

            // Act
            Paragraph clonedParagraph = firstParagraph.Clone();

            // Assert
            Assert.IsFalse(clonedParagraph.EnumerateChildrenOfType<FieldCharacter>().Any());
        }

        [TestMethod]
        public void Clone_WhenContainsBookmark_ClonesTheParagraphAndBookmarkStartAndEnd()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();

            Paragraph paragraph = new Paragraph(document1);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document1);

            editor.MoveToParagraphStart(paragraph);
            Run run = editor.InsertText("run");
            editor.InsertBookmark("test", run, run);

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            Paragraph clonedParagraph = paragraph.Clone(document2);

            // Assert
            StructureAsserter.AssertParagraphsAreEqual(paragraph, clonedParagraph);

            Bookmark bookmarkFromStart = clonedParagraph.EnumerateChildrenOfType<BookmarkRangeStart>().FirstOrDefault().Bookmark;
            Bookmark bookmarkFromEnd = clonedParagraph.EnumerateChildrenOfType<BookmarkRangeEnd>().FirstOrDefault().Bookmark;

            Assert.AreEqual(bookmarkFromStart, bookmarkFromEnd);
            Assert.AreEqual("test", bookmarkFromStart.Name);
        }

        [TestMethod]
        public void Clone_WhenContainsOnlyBookmarkStart_ClonesTheParagraph_ButBookmarkStartIsStripped()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Paragraph firstParagraph = editor.InsertParagraph();
            Run run1 = editor.InsertText("run1");
            editor.InsertParagraph();
            Run run2 = editor.InsertText("run2");
            editor.InsertBookmark("test", run1, run2);

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            Paragraph clonedParagraph = firstParagraph.Clone(document2);

            // Assert
            Assert.IsFalse(clonedParagraph.EnumerateChildrenOfType<AnnotationRangeStartBase>().Any());
        }

        [TestMethod]
        public void Clone_WhenContainsOnlyBookmarkEnd_ClonesTheParagraph_ButBookmarkEndIsStripped()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertParagraph();
            Run run1 = editor.InsertText("run1");
            Paragraph secondParagraph = editor.InsertParagraph();
            Run run2 = editor.InsertText("run2");
            editor.InsertBookmark("test", run1, run2);

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            Paragraph clonedParagraph = secondParagraph.Clone(document2);

            // Assert
            Assert.IsFalse(clonedParagraph.EnumerateChildrenOfType<AnnotationRangeEndBase>().Any());
        }

        [TestMethod]
        public void Clone_WhenContainsNestedBookmarks_ClonesTheParagraphAndBookmarks()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            Run outsideRunLeft = editor.InsertText("ne");
            Run middleRunLeft = editor.InsertText("s");
            Run innerRun = editor.InsertText("t");
            Run middleRunRight = editor.InsertText("e");
            Run outsideRunRight = editor.InsertText("d");

            editor.InsertBookmark("inner", innerRun, innerRun);
            editor.InsertBookmark("middle", middleRunLeft, middleRunRight);
            editor.InsertBookmark("outside", outsideRunLeft, outsideRunRight);

            // Act 
            Paragraph clonedParagraph = editor.Document.EnumerateChildrenOfType<Paragraph>().First().Clone();

            // Assert
            StructureAsserter.AssertDocumentsAreEqual(editor.Document, clonedParagraph.Document);
        }

        [TestMethod]
        public void Clone_WhenContainsCommentAroundInline_ClonesTheParagraphAndComment()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Paragraph paragraph = new Paragraph(document1);
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document1);

            editor.MoveToParagraphStart(paragraph);
            Run run = editor.InsertText("run");
            editor.InsertComment("test", run, run);

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            Paragraph clonedParagraph = paragraph.Clone(document2);

            // Assert
            Assert.IsTrue(clonedParagraph.EnumerateChildrenOfType<CommentRangeStart>().Any());
            Assert.IsTrue(clonedParagraph.EnumerateChildrenOfType<CommentRangeEnd>().Any());

            StructureAsserter.AssertParagraphsAreEqual(paragraph, clonedParagraph);
        }

        [TestMethod]
        public void Clone_WhenContainsComment_ClonesTheParagraphAndComment()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Paragraph paragraph = new Paragraph(document1);
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document1);

            editor.MoveToParagraphStart(paragraph);
            editor.InsertComment("test");

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            Paragraph clonedParagraph = paragraph.Clone(document2);

            // Assert
            Assert.IsTrue(clonedParagraph.EnumerateChildrenOfType<CommentRangeStart>().Any());
            Assert.IsTrue(clonedParagraph.EnumerateChildrenOfType<CommentRangeEnd>().Any());

            StructureAsserter.AssertParagraphsAreEqual(paragraph, clonedParagraph);
        }

        [TestMethod]
        public void Clone_WhenContainsCommentEnd_ClonesTheParagraph_ButCommentEndIsStripped()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertParagraph();
            Run run1 = editor.InsertText("run1");
            Paragraph secondParagraph = editor.InsertParagraph();
            Run run2 = editor.InsertText("run2");
            editor.InsertComment("test", run1, run2);

            // Act
            Paragraph clonedParagraph = secondParagraph.Clone(new RadFlowDocument());

            // Assert
            Assert.IsTrue(secondParagraph.EnumerateChildrenOfType<CommentRangeEnd>().Any());
            Assert.IsFalse(clonedParagraph.EnumerateChildrenOfType<CommentRangeStart>().Any());
            Assert.IsFalse(clonedParagraph.EnumerateChildrenOfType<CommentRangeEnd>().Any());
        }

        [TestMethod]
        public void Clone_WhenContainsOnlyCommentStart_ClonesTheParagraph_ButCommentStartIsStripped()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            Paragraph firstParagraph = editor.InsertParagraph();
            Run run1 = editor.InsertText("run1");
            editor.InsertParagraph();
            Run run2 = editor.InsertText("run2");
            editor.InsertComment("test", run1, run2);

            // Act
            Paragraph clonedParagraph = firstParagraph.Clone(new RadFlowDocument());

            // Assert
            Assert.IsTrue(firstParagraph.EnumerateChildrenOfType<CommentRangeStart>().Any());
            Assert.IsFalse(clonedParagraph.EnumerateChildrenOfType<CommentRangeStart>().Any());
            Assert.IsFalse(clonedParagraph.EnumerateChildrenOfType<CommentRangeEnd>().Any());
        }

        [TestMethod]
        public void Clone_WhenBreakIsInParagraph_ClonesBreakAndItsProperties()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            document.Sections.AddSection().Blocks.AddParagraph();

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Break br = editor.InsertBreak(BreakType.PageBreak);
            br.TextWrappingRestartLocation = TextWrappingRestartLocation.NextTextRegionUnblockedOnLeft;

            Paragraph secondParagraph = (Paragraph)document.Sections[0].Blocks[0];

            // Act
            Paragraph clonedParagraph = secondParagraph.Clone();
            Break clonedBreak = clonedParagraph.EnumerateChildrenOfType<Break>().First();

            // Assert
            Assert.AreEqual(br.BreakType, clonedBreak.BreakType);
            Assert.AreEqual(br.TextWrappingRestartLocation, clonedBreak.TextWrappingRestartLocation);
        }

        [TestMethod]
        public void Clone_WhenContainsTabStops_ClonesTheParagraphAndTabStops()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Paragraph paragraph = new Paragraph(document);

            TabStopCollection tabStops = new TabStopCollection().Insert(new TabStop(100, TabStopType.Bar, TabStopLeader.Underscore))
                                                                .Insert(new TabStop(200, TabStopType.Center))
                                                                .Insert(new TabStop(200, TabStopType.Center));

            paragraph.TabStops = tabStops;

            // Act
            Paragraph clonedParagraph = paragraph.Clone();

            // Assert
            Assert.AreEqual(3, clonedParagraph.TabStops.Count);
            StructureAsserter.AssertParagraphsAreEqual(paragraph, clonedParagraph);
        }

        [TestMethod]
        public void Clone_ParagraphParagraphMarkerProperties_CloningHasSameParagraphMarkerProperties()
        {
            RadFlowDocument document = new RadFlowDocument();

            Paragraph paragraph = new Paragraph(document);
            paragraph.Properties.ParagraphMarkerProperties.ForegroundColor.LocalValue = new ThemableColor(Colors.Blue);

            Paragraph cloned = paragraph.Clone();

            StylePropertyAsserter.AssertPropertiesAreEqual(paragraph.Properties.ParagraphMarkerProperties, cloned.Properties.ParagraphMarkerProperties);
        }

        [TestMethod]
        public void CopyProperties_ForParagraphMarkerProperties_CopyHasTheSameParagraphMarkerProperties()
        {
            RadFlowDocument document = new RadFlowDocument();

            Paragraph paragraph = new Paragraph(document);
            paragraph.Properties.ParagraphMarkerProperties.ForegroundColor.LocalValue = new ThemableColor(Colors.Blue);

            Paragraph newParagraph = new Paragraph(document);

            newParagraph.Properties.CopyPropertiesFrom(paragraph.Properties);

            StylePropertyAsserter.AssertPropertiesAreEqual(paragraph.Properties, newParagraph.Properties);
        }
    }
}