﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Model.Protection;

namespace Telerik.Windows.Documents.Flow.Tests.Model.Protection
{
    [TestClass]
    public class DocumentProtectionTests : ProtectionBaseTests
    {
        private RadFlowDocumentEditor editor;

        [TestInitialize]
        public void TestInitialize()
        {
            this.editor = new RadFlowDocumentEditor(new RadFlowDocument());
        }

        [TestCleanup]
        public void TestCleanup()
        {
            this.editor = null;
        }

        [TestMethod]
        public void ProtectDocument_WithDefaultMode()
        {
            // Act
            this.editor.Protect("password");

            // Assert
            this.AssertDocumentIsProtected(this.editor.Document.ProtectionSettings, ProtectionMode.ReadOnly, ProtectionAlgorithmNames.SHA512);
        }

        [TestMethod]
        public void ProtectDocument_WithReadOnlyMode()
        {
            // Act
            ProtectionMode protectionMode = ProtectionMode.ReadOnly;
            this.editor.Protect("password", protectionMode);

            // Assert
            this.AssertDocumentIsProtected(this.editor.Document.ProtectionSettings, protectionMode, ProtectionAlgorithmNames.SHA512);
        }

        [TestMethod]
        public void ProtectDocument_WithAllowCommentMode()
        {
            // Act
            ProtectionMode protectionMode = ProtectionMode.AllowComments;
            this.editor.Protect("password", protectionMode);

            // Assert
            this.AssertDocumentIsProtected(this.editor.Document.ProtectionSettings, protectionMode, ProtectionAlgorithmNames.SHA512);
        }

        [TestMethod]
        public void Unprotectection()
        {
            // Act
            this.editor.Protect("password");
            this.editor.Unprotect();

            // Assert
            Assert.IsFalse(this.editor.Document.ProtectionSettings.Enforced);

            Assert.IsTrue(this.editor.Document.ProtectionSettings.ProtectionMode == ProtectionMode.NoProtection);
            Assert.IsTrue(string.IsNullOrEmpty(this.editor.Document.ProtectionSettings.Hash));
            Assert.IsTrue(string.IsNullOrEmpty(this.editor.Document.ProtectionSettings.Salt));
        }

        [TestMethod]
        public void Unprotection_WithCorrectPassword()
        {
            // Assert
            string password = "password";
            this.editor.Protect(password);

            // Act
            bool isSuccessful = this.editor.Unprotect(password);

            // Assert
            Assert.IsTrue(isSuccessful);
        }

        [TestMethod]
        public void Unprotection_WithIncorrectPassword()
        {
            // Assert
            string password = "password";
            this.editor.Protect(password);

            // Act
            bool isSuccessful = this.editor.Unprotect(password + "incorrect");

            // Assert
            Assert.IsFalse(isSuccessful);
        }

        private void TestProtectionAlgorithms(string algorithmName)
        {
            // Arrange
            ProtectionMode protectionMode = ProtectionMode.ReadOnly;

            // Act
            this.editor.Document.ProtectionSettings.AlgorithmName = algorithmName;
            this.editor.Protect("password", protectionMode);

            // Assert
            this.AssertDocumentIsProtected(this.editor.Document.ProtectionSettings, protectionMode, algorithmName);
        }

        [TestMethod]
        public void ProtectDocument_WithAlgorithmSha1()
        {
            this.TestProtectionAlgorithms(ProtectionAlgorithmNames.SHA1);
        }

        [TestMethod]
        public void ProtectDocument_WithAlgorithmSha256()
        {
            this.TestProtectionAlgorithms(ProtectionAlgorithmNames.SHA256);
        }

        [TestMethod]
        public void ProtectDocument_WithAlgorithmSha512()
        {
            this.TestProtectionAlgorithms(ProtectionAlgorithmNames.SHA512);
        }

        [TestMethod]
        public void ProtectDocument_WithAlgorithmSha384()
        {
            this.TestProtectionAlgorithms(ProtectionAlgorithmNames.SHA384);
        }
    }
}
