﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Linq;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Model.Protection;
using Telerik.Windows.Documents.Flow.Tests.Utilities;

namespace Telerik.Windows.Documents.Flow.Tests.Model.Protection
{
    [TestClass]
    public class PermissionTests : ProtectionBaseTests
    {
        private RadFlowDocumentEditor editor;
        private PermissionRange permission;
        private PermissionRangeCredentials credentials;

        [TestInitialize]
        public void TestInitialize()
        {
            this.editor = new RadFlowDocumentEditor(new RadFlowDocument());
            this.credentials = new PermissionRangeCredentials("test");
            this.permission = new PermissionRange(this.editor.Document, this.credentials);
        }

        [TestCleanup]
        public void TestCleanup()
        {
            this.editor = null;
            this.credentials = null;
            this.permission = null;
        }

        [TestMethod]
        public void CreatePermissionCredentials_IndividualNameWithWhitespace_ThrowsException()
        {
            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(() => new PermissionRangeCredentials(" test"));

            TestHelper.AssertExceptionIsThrown<ArgumentException>(() => new PermissionRangeCredentials("test "));

            TestHelper.AssertExceptionIsThrown<ArgumentException>(() => new PermissionRangeCredentials("te st"));
        }

        [TestMethod]
        public void CreatePermissionCredentials_WithEmptyIndividualName_ThrowsException()
        {
            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(() => new PermissionRangeCredentials(string.Empty));
        }

        [TestMethod]
        public void CreatePermissionCredentials_WithNullEditingGroup_ThrowsException()
        {
            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(() => new PermissionRangeCredentials(null));
        }

        [TestMethod]
        public void CreatePermission_WithNullDocument_ThrowsException()
        {
            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(() => new PermissionRange(null, new PermissionRangeCredentials("test")));
        }

        [TestMethod]
        public void CreatePermission_WithNullPermissionCredentials_ThrowsException()
        {
            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(() => new PermissionRange(new RadFlowDocument(), null));
        }

        [TestMethod]
        public void CreatePermission_WithEditor_ExpectEditingGroupToBeNull()
        {
            PermissionRangeCredentials permissionCredentials = new PermissionRangeCredentials("test");

            Assert.IsNotNull(permissionCredentials.Editor);
            Assert.IsNull(permissionCredentials.EditingGroup);
        }

        [TestMethod]
        public void CreatePermission_WithEditingGroup_ExpectEditorToBeNull()
        {
            PermissionRangeCredentials permissionCredentials = new PermissionRangeCredentials(EditingGroup.Everyone);

            Assert.IsNotNull(permissionCredentials.EditingGroup);
            Assert.IsNull(permissionCredentials.Editor);
        }

        [TestMethod]
        public void DeletePermission()
        {
            // Arrange
            this.editor.InsertInline(this.permission.Start);
            this.editor.InsertText("no edit");
            this.editor.InsertInline(this.permission.End);

            // Assert
            PermissionRangeStart permissionRangeStart = this.editor.Document.EnumerateChildrenOfType<PermissionRangeStart>().FirstOrDefault();
            PermissionRangeEnd permissionRangeEnd = this.editor.Document.EnumerateChildrenOfType<PermissionRangeEnd>().First();

            Assert.IsNotNull(permissionRangeStart);
            Assert.IsNotNull(permissionRangeEnd);

            // Act
            this.editor.DeletePermissionRange(this.permission);

            // Assert
            permissionRangeStart = this.editor.Document.EnumerateChildrenOfType<PermissionRangeStart>().FirstOrDefault();
            permissionRangeEnd = this.editor.Document.EnumerateChildrenOfType<PermissionRangeEnd>().FirstOrDefault();

            Assert.IsNull(permissionRangeStart);
            Assert.IsNull(permissionRangeEnd);
        }

        [TestMethod]
        public void InsertPermission_UsingModel()
        {
            // Arrange & Act
            this.editor.InsertInline(this.permission.Start);
            this.editor.InsertText("no edit");
            this.editor.InsertInline(this.permission.End);

            // Assert
            Paragraph paragraph = this.editor.Document.Sections.First().Blocks.First() as Paragraph;
            Assert.IsNotNull(paragraph);

            InlineBase[] inlines = paragraph.Inlines.ToArray();
            Assert.AreEqual(3, inlines.Length);

            Assert.IsInstanceOfType(inlines[0], typeof(PermissionRangeStart));
            Assert.IsInstanceOfType(inlines[1], typeof(Run));
            Assert.IsInstanceOfType(inlines[2], typeof(PermissionRangeEnd));
        }

        [TestMethod]
        public void InsertPermission_AroundInlines()
        {
            // Arrange
            Run inlineStart = this.editor.InsertText("first");
            this.editor.InsertParagraph();
            Run inlineEnd = this.editor.InsertText("second");

            // Act
            this.editor.InsertPermissionRange(this.credentials, inlineStart, inlineEnd);

            // Assert
            InlineBase[] inlines = this.editor.Document.EnumerateChildrenOfType<InlineBase>().ToArray();

            Assert.IsInstanceOfType(inlines[0], typeof(PermissionRangeStart));
            Assert.IsInstanceOfType(inlines[3], typeof(PermissionRangeEnd));
        }

        [TestMethod]
        public void InsertPermission_AroundEmptyCell_AndEmptyNextCell()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 2);
            TableRow row = table.Rows.First();

            // Act
            this.editor.InsertPermissionRange(this.credentials, row.Cells.First());

            // Assert
            Paragraph paragraph = row.Cells.First().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeStart));

            paragraph = row.Cells.Last().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeEnd));
        }

        [TestMethod]
        public void InsertPermission_AroundCell_WithContent_AndNextCell_WithContent()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 2);
            TableRow row = table.Rows.First();
            TableCell cell = row.Cells.First();
            Paragraph startParagraph = cell.Blocks.AddParagraph();

            this.editor.MoveToParagraphStart(startParagraph);
            this.editor.InsertText("inner text");

            Paragraph nextParagraph = row.Cells.Last().Blocks.AddParagraph();
            this.editor.MoveToParagraphStart(nextParagraph);
            this.editor.InsertText("outer text");

            // Act
            this.editor.InsertPermissionRange(this.credentials, row.Cells.First());

            // Assert
            nextParagraph = row.Cells.First().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(nextParagraph.Inlines.First(), typeof(PermissionRangeStart));

            nextParagraph = row.Cells.Last().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(nextParagraph.Inlines.First(), typeof(PermissionRangeEnd));
        }

        [TestMethod]
        public void InsertPermission_AroundEmptyCell_AndNextCell_WithContent()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 2);
            TableRow row = table.Rows.First();

            Paragraph nextParagraph = row.Cells.Last().Blocks.AddParagraph();
            this.editor.MoveToParagraphStart(nextParagraph);
            this.editor.InsertText("outer text");

            // Act
            this.editor.InsertPermissionRange(this.credentials, row.Cells.First());

            // Assert
            nextParagraph = row.Cells.First().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(nextParagraph.Inlines.First(), typeof(PermissionRangeStart));

            nextParagraph = row.Cells.Last().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(nextParagraph.Inlines.First(), typeof(PermissionRangeEnd));
        }

        [TestMethod]
        public void InsertPermission_AroundCell_WithContent_AndEmptyNextCell()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 2);
            TableRow row = table.Rows.First();
            TableCell cell = row.Cells.First();
            Paragraph startParagraph = cell.Blocks.AddParagraph();

            this.editor.MoveToParagraphStart(startParagraph);
            this.editor.InsertText("inner text");

            // Act
            this.editor.InsertPermissionRange(this.credentials, row.Cells.First());

            // Assert
            Paragraph nextParagraph = row.Cells.First().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(nextParagraph.Inlines.First(), typeof(PermissionRangeStart));

            nextParagraph = row.Cells.Last().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(nextParagraph.Inlines.First(), typeof(PermissionRangeEnd));
        }

        [TestMethod]
        public void InsertPermission_AroundLastEmptyCell_WhenTheCellIsTheLastElementInDocument()
        {
            // Arrange
            Table table = this.editor.InsertTable(1, 1);
            TableRow row = table.Rows.First();

            // Act
            this.editor.InsertPermissionRange(this.credentials, row.Cells.First());

            // Assert
            Paragraph paragraph = row.Cells.First().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeStart));

            paragraph = row.Cells.Last().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.Last(), typeof(PermissionRangeEnd));
        }

        [TestMethod]
        public void InsertPermission_AroundEmptyRow_AndEmptyNextRow()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 2);
            TableRow firstRow = table.Rows.First();
            TableRow nextRow = table.Rows.Last();

            // Act
            this.editor.InsertPermissionRange(this.credentials, firstRow);

            // Assert
            Paragraph paragraph = firstRow.Cells.First().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeStart));

            paragraph = nextRow.Cells.First().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeEnd));
        }

        [TestMethod]
        public void InsertPermission_AroundRow_WithContent_AndEmptyNextRow()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 2);
            TableRow firstRow = table.Rows.First();
            TableRow nextRow = table.Rows.Last();

            TableCell cell = firstRow.Cells.AddTableCell();
            Paragraph startParagraph = cell.Blocks.AddParagraph();

            this.editor.MoveToParagraphStart(startParagraph);
            this.editor.InsertText("inner text");

            // Act
            this.editor.InsertPermissionRange(this.credentials, firstRow);

            // Assert
            Paragraph paragraph = firstRow.Cells.First().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeStart));

            paragraph = nextRow.Cells.First().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeEnd));
        }

        [TestMethod]
        public void InsertPermission_AroundRow_WithContent_AndNextRow_WithContent()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 2);
            TableRow firstRow = table.Rows.First();
            TableRow nextRow = table.Rows.Last();

            TableCell cell = firstRow.Cells.AddTableCell();
            Paragraph startParagraph = cell.Blocks.AddParagraph();

            this.editor.MoveToParagraphStart(startParagraph);
            this.editor.InsertText("inner text");

            Paragraph nextParagraph = nextRow.Cells.Last().Blocks.AddParagraph();
            this.editor.MoveToParagraphStart(nextParagraph);
            this.editor.InsertText("outer text");

            // Act
            this.editor.InsertPermissionRange(this.credentials, firstRow);

            // Assert
            Paragraph paragraph = firstRow.Cells.First().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeStart));

            paragraph = nextRow.Cells.First().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeEnd));
        }

        [TestMethod]
        public void InsertPermission_AroundEmptyRow_AndNextRow_WithContent()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 2);
            TableRow firstRow = table.Rows.First();
            TableRow nextRow = table.Rows.Last();

            Paragraph nextParagraph = nextRow.Cells.Last().Blocks.AddParagraph();
            this.editor.MoveToParagraphStart(nextParagraph);
            this.editor.InsertText("outer text");

            // Act
            this.editor.InsertPermissionRange(this.credentials, firstRow);

            // Assert
            Paragraph paragraph = firstRow.Cells.First().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeStart));

            paragraph = nextRow.Cells.First().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeEnd));

            Assert.AreEqual(1, this.editor.Document.EnumerateChildrenOfType<PermissionRangeStart>().Count());
            Assert.AreEqual(1, this.editor.Document.EnumerateChildrenOfType<PermissionRangeEnd>().Count());
        }

        [TestMethod]
        public void InsertPermission_AroundEmptyRow_WhenTheRowIsTheLastElementInDocument()
        {
            // Arrange
            Table table = this.editor.InsertTable(1, 1);
            TableRow firstRow = table.Rows.First();

            // Act
            this.editor.InsertPermissionRange(this.credentials, firstRow);

            // Assert

            Assert.AreEqual(1, this.editor.Document.EnumerateChildrenOfType<PermissionRangeStart>().Count());
            Assert.AreEqual(1, this.editor.Document.EnumerateChildrenOfType<PermissionRangeEnd>().Count());

            Paragraph paragraph = firstRow.Cells.First().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeStart));

            paragraph = firstRow.Cells.Last().Blocks.Last() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.Last(), typeof(PermissionRangeEnd));
        }

        [TestMethod]
        public void InsertPermission_AroundCell_WhenTheCellIsColumnSpanned()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 1);
            TableCell firstCell = table.Rows.First().Cells.First();
            firstCell.ColumnSpan = 2;

            table.Rows.Last().Cells.AddTableCell();

            Assert.AreEqual(1, table.Rows.First().Cells.Count);
            Assert.AreEqual(2, table.Rows.Last().Cells.Count);

            // Act
            this.editor.InsertPermissionRange(this.credentials, firstCell);

            // Assert
            Assert.AreEqual(1, table.Rows.First().Cells.Count);
            Assert.AreEqual(2, table.Rows.Last().Cells.Count);

            Assert.AreEqual(1, this.editor.Document.EnumerateChildrenOfType<PermissionRangeStart>().Count());
            Assert.AreEqual(1, this.editor.Document.EnumerateChildrenOfType<PermissionRangeEnd>().Count());

            Paragraph paragraph = firstCell.Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeStart));

            paragraph = table.Rows.Last().Cells.First().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeEnd));
        }

        [TestMethod]
        public void InsertPermission_AroundCell_WhenTheCellIsRowSpanned()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 1);
            TableCell firstCell = table.Rows.First().Cells.First();
            firstCell.RowSpan = 2;

            table.Rows.First().Cells.AddTableCell();

            Assert.AreEqual(2, table.Rows.First().Cells.Count);
            Assert.AreEqual(1, table.Rows.Last().Cells.Count);

            // Act
            this.editor.InsertPermissionRange(this.credentials, firstCell);

            // Assert
            Assert.AreEqual(2, table.Rows.First().Cells.Count);
            Assert.AreEqual(1, table.Rows.Last().Cells.Count);

            Assert.AreEqual(1, this.editor.Document.EnumerateChildrenOfType<PermissionRangeStart>().Count());
            Assert.AreEqual(1, this.editor.Document.EnumerateChildrenOfType<PermissionRangeEnd>().Count());

            Paragraph paragraph = firstCell.Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeStart));

            paragraph = table.Rows.First().Cells.Last().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeEnd));
        }

        [TestMethod]
        public void InsertPermission_AroundRow_WhenTheFirstCellIsColumnSpanned()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 1);
            TableRow firstRow = table.Rows.First();
            TableCell firstCell = firstRow.Cells.First();
            firstCell.ColumnSpan = 2;

            table.Rows.Last().Cells.AddTableCell();

            Assert.AreEqual(1, firstRow.Cells.Count);
            Assert.AreEqual(2, table.Rows.Last().Cells.Count);

            // Act
            this.editor.InsertPermissionRange(this.credentials, firstCell);

            // Assert
            Assert.AreEqual(1, firstRow.Cells.Count);
            Assert.AreEqual(2, table.Rows.Last().Cells.Count);

            Assert.AreEqual(1, this.editor.Document.EnumerateChildrenOfType<PermissionRangeStart>().Count());
            Assert.AreEqual(1, this.editor.Document.EnumerateChildrenOfType<PermissionRangeEnd>().Count());

            Paragraph paragraph = firstCell.Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeStart));

            paragraph = table.Rows.Last().Cells.First().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeEnd));
        }

        [TestMethod]
        public void InsertPermission_AroundRow_WhenTheFirstCellIsRowSpanned()
        {
            // Arrange
            Table table = this.editor.InsertTable(2, 1);
            TableRow firstRow = table.Rows.First();
            TableCell firstCell = firstRow.Cells.First();
            firstCell.RowSpan = 2;

            firstRow.Cells.AddTableCell();

            Assert.AreEqual(2, firstRow.Cells.Count);
            Assert.AreEqual(1, table.Rows.Last().Cells.Count);

            // Act
            this.editor.InsertPermissionRange(this.credentials, firstRow);

            // Assert
            Assert.AreEqual(2, firstRow.Cells.Count);
            Assert.AreEqual(1, table.Rows.Last().Cells.Count);

            Assert.AreEqual(1, this.editor.Document.EnumerateChildrenOfType<PermissionRangeStart>().Count());
            Assert.AreEqual(1, this.editor.Document.EnumerateChildrenOfType<PermissionRangeEnd>().Count());

            Paragraph paragraph = firstCell.Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeStart));

            paragraph = table.Rows.Last().Cells.First().Blocks.First() as Paragraph;
            Assert.IsInstanceOfType(paragraph.Inlines.First(), typeof(PermissionRangeEnd));
        }
    }
}
