﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.Model;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Media;
#else
using System.Windows;
using System.Windows.Media;
#endif
#if WPF || SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Tests.Utilities;

namespace Telerik.Windows.Documents.Flow.Tests.Model
{
    [TestClass]
    public class RunTests
    {
        [TestMethod]
        public void Constructor_TextValueIsEmptyString()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            // Act
            Run run = new Run(document);

            // Assert
            Assert.AreEqual(string.Empty, run.Text);
        }

        [TestMethod]
        public void SetText_WithNull_ThrowsArgumentNullException()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            // Act
            Run run = new Run(document);

            try
            {
                run.Text = null;
            }
            catch (ArgumentNullException)
            {
                return;
            }

            Assert.Fail("ArgumentNullException is expected.");
        }

        [TestMethod]
        public void SetText_WithNewLine_ClearsTheNewLines()
        {
            // Arrange
            Run run = new Run(new RadFlowDocument());

            // Act 
            string test = "First" + Environment.NewLine + "Second";
            run.Text = test;

            // Assert
            Assert.AreEqual("FirstSecond", run.Text);
        }

        [TestMethod]
        public void SetText_WithNewLinesAndCarrigeReturns_ClearsTheNewLines()
        {
            // Arrange
            Run run = new Run(new RadFlowDocument());

            // Act 
            string test = "First" + "\r\n" + "Second" + "\n" + "Third" + "\r" + "Last";
            run.Text = test;

            // Assert
            Assert.AreEqual("FirstSecondThirdLast", run.Text);
        }

        [TestMethod]
        public void CloneWithParameter_WhenInvokedWithNull_ThrowsArgumentNullException()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Run run = new Run(document1);

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentNullException>(() => run.Clone(null));
        }

        [TestMethod]
        public void Clone_WithDocument_ClonesTheRunForTheDocument()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Run run = new Run(document);

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            Run clonedRun = run.Clone(document2);

            // Assert
            Assert.AreNotSame(run, clonedRun);
            Assert.AreSame(document2, clonedRun.Document);
        }

        [TestMethod]
        public void Clone_WithoutDocument_ClonesTheRunForTheCurrentDocument()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Run run = new Run(document);

            // Act
            Run clonedRun = run.Clone(document);

            // Assert
            Assert.AreNotSame(run, clonedRun);
            Assert.AreSame(document, clonedRun.Document);
        }

        [TestMethod]
        public void Clone_WhenTextIsSet_ClonedRunHasSameText()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Run run = new Run(document);
            run.Text = "some text";

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            Run clonedRun = run.Clone(document2);

            // Assert
            Assert.AreEqual(run.Text, clonedRun.Text);
        }

        [TestMethod]
        public void Clone_WhenPropertiesAreSet_CloningHasSameProperties()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Run run = new Run(document);

            run.BaselineAlignment = Flow.Model.Styles.BaselineAlignment.Superscript;
            run.FontFamily = new ThemableFontFamily("Arial");
            run.FontSize = 50;
            run.FontStyle = FontStyles.Italic;
            run.FontWeight = FontWeights.Bold;
            run.ForegroundColor = new ThemableColor(Colors.Violet);
            run.HighlightColor = Colors.Yellow;
            run.Strikethrough = true;
            run.StyleId = "Heading1";
            run.Text = "some text";
            run.Underline.Color = new ThemableColor(Colors.Green);
            run.Underline.Pattern = UnderlinePattern.DashLongHeavy;
            run.Shading.BackgroundColor = new ThemableColor(Colors.Red);
            run.Shading.PatternColor = new ThemableColor(Colors.Blue);
            run.Shading.Pattern = ShadingPattern.Percent37;

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            Run clonedRun = run.Clone(document2);

            // Assert
            StylePropertyAsserter.AssertPropertiesAreEqual(run.Properties, clonedRun.Properties);
        }
    }
}
