﻿using System;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
using Telerik.Windows.Documents.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Model.Fields;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Primitives;
#if NETSTANDARD
using Telerik.Documents.Primitives;
#else
using System.Windows;
#endif
#if VISUAL
using System.Windows.Media;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Flow.Tests.Utils;
#endif

namespace Telerik.Windows.Documents.Flow.Tests.Model
{
#if VISUAL
    [TestClass]
    public class SectionTests : Telerik.Windows.Documents.Fixed.Tests.Integration.DocumentFlowVisualBaseTest
    {
#else
    [TestClass]
    public class SectionTests
    {
        public const string VISUALTESTCATEGORY = "VisualTests";
#endif
        [TestMethod]
        public void CloneWithParameter_WhenInvokedWithNull_ThrowsArgumentNullException()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            Section section = new Section(document1);

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentNullException>(() => section.Clone(null));
        }

        [TestMethod]
        public void Clone_WithDocument_ClonesTheSectionForTheDocument()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();

            Section section = new Section(document1);
            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            Section clonedSection = section.Clone(document2);

            // Assert
            Assert.AreNotSame(section, clonedSection);
            Assert.AreSame(document2, clonedSection.Document);
        }

        [TestMethod]
        public void Clone_WithoutDocument_ClonesTheSectionForTheCurrentDocument()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Section section = new Section(document);

            // Act
            Section clonedSection = section.Clone();

            // Assert
            Assert.AreNotSame(section, clonedSection);
            Assert.AreSame(document, clonedSection.Document);
        }

        [TestMethod]
        public void Clone_WhenPropertiesAreSet_CloningHasSameProperties()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();

            Section section = new Section(document1);
            section.PageOrientation = PageOrientation.Landscape;
            section.SectionType = SectionType.Continuous;

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            Section clonedSection = section.Clone(document2);

            // Assert
            StylePropertyAsserter.AssertPropertiesAreEqual(section.Properties, clonedSection.Properties);
        }

        [TestMethod]
        public void Clone_WhenHasChildParagraphs_CloningHasSameStructure()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();

            Section section = new Section(document1);
            section.Blocks.AddParagraph().Inlines.AddRun("some text");
            section.Blocks.AddParagraph().Inlines.AddRun("some text");

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            Section clonedSection = section.Clone(document2);

            // Assert
            StructureAsserter.AssertSectionsAreEqual(section, clonedSection);
        }

        [TestMethod]
        public void Clone_WhenHasHeadersAndFooters_CloningHasClonedHeadersFootersWithSameStructure()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();

            Section section = new Section(document1);
            section.Headers.Add(HeaderFooterType.First).Blocks.AddParagraph().Inlines.AddRun("first header");
            section.Headers.Add(HeaderFooterType.Even).Blocks.AddParagraph().Inlines.AddRun("even header");
            section.Headers.Add(HeaderFooterType.Default).Blocks.AddParagraph().Inlines.AddRun("odd header");
            section.Footers.Add(HeaderFooterType.First).Blocks.AddParagraph().Inlines.AddRun("first footer");
            section.Footers.Add(HeaderFooterType.Even).Blocks.AddParagraph().Inlines.AddRun("even footer");
            section.Footers.Add(HeaderFooterType.Default).Blocks.AddParagraph().Inlines.AddRun("odd footer");

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            Section clonedSection = section.Clone(document2);

            // Assert
            StructureAsserter.AssertSectionsAreEqual(section, clonedSection);
            Assert.AreSame(clonedSection, clonedSection.Headers.First.Parent);
            Assert.AreSame(clonedSection, clonedSection.Headers.Even.Parent);
            Assert.AreSame(clonedSection, clonedSection.Headers.Default.Parent);
            Assert.AreSame(clonedSection, clonedSection.Footers.First.Parent);
            Assert.AreSame(clonedSection, clonedSection.Footers.Even.Parent);
            Assert.AreSame(clonedSection, clonedSection.Footers.Default.Parent);
        }

        [TestMethod]
        public void Clone_WhenHasFieldsSpanningTwoParagraphs_CloningHasSameStructure()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Section section = document.Sections.AddSection();
            section.Blocks.AddParagraph();

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            editor.InsertHyperlink("John \r\nDoe", "www.telerik.com", false);

            // Act
            Section clonedSection = section.Clone();

            // Assert
            StructureAsserter.AssertSectionsAreEqual(section, clonedSection);

            FieldInfo clonedField = clonedSection.EnumerateChildrenOfType<FieldCharacter>().First().FieldInfo;
            Assert.IsTrue(clonedSection.EnumerateChildrenOfType<FieldCharacter>().All(fieldCharacter => fieldCharacter.FieldInfo == clonedField), "Field characters reference incorrect field.");
        }

        [TestMethod]
        public void Clone_WhenHasBookmarkSpanningTwoParagraphs_CloningHasSameStructure()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            Run run1 = editor.InsertText("run1");
            editor.InsertParagraph();
            Run run2 = editor.InsertText("run2");
            editor.InsertBookmark("test", run1, run2);

            Section section = editor.Document.Sections.First();

            // Act
            Section clonedSection = section.Clone();

            // Assert
            StructureAsserter.AssertSectionsAreEqual(section, clonedSection);
        }

        [TestMethod]
        public void Clone_WhenHasPageNumberingSettings_CloningSectionHasSamePageNumberingSettings()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Section expected = new Section(document);
            expected.PageNumberingSettings.ChapterHeadingStyleIndex = 1;
            expected.PageNumberingSettings.ChapterSeparatorCharacter = ChapterSeparatorType.EmDash;
            expected.PageNumberingSettings.PageNumberFormat = NumberingStyle.Decimal;
            expected.PageNumberingSettings.StartingPageNumber = 10;

            // Act
            Section copiedSection = expected.Clone();

            // Assert
            StructureAsserter.AssertSectionsAreEqual(expected, copiedSection);
        }

        [TestMethod]
        public void CopyPropertiesFrom_WhenSectionElementHasPageNumberingSettings_CopiesPageNumberingSettings()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Section expected = new Section(document);
            expected.PageNumberingSettings.ChapterHeadingStyleIndex = 1;
            expected.PageNumberingSettings.ChapterSeparatorCharacter = ChapterSeparatorType.EmDash;
            expected.PageNumberingSettings.PageNumberFormat = NumberingStyle.Decimal;
            expected.PageNumberingSettings.StartingPageNumber = 10;

            // Act
            Section copiedSection = new Section(document);
            copiedSection.Properties.CopyPropertiesFrom(expected.Properties);

            // Assert
            StructureAsserter.AssertSectionsAreEqual(expected, copiedSection);
        }

        [TestMethod]
        public void PageNumberingSettings_WhenTryToSetChapterHeadingStyleIndexLessThan1_ArgumentOutOfRangeExceptionIsThrown()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Section section = new Section(document);

            //Act
            try
            {
                section.PageNumberingSettings.ChapterHeadingStyleIndex = 0;
                Assert.Fail("Should throw ArgumentOutOfRangeException.");
            }
            catch (ArgumentOutOfRangeException)
            {
            }

            // Assert
            Assert.IsNull(section.PageNumberingSettings.ChapterHeadingStyleIndex);
        }

        [TestMethod]
        public void PageNumberingSettings_WhenTryToSetChapterHeadingStyleIndexMoreThan9_ArgumentOutOfRangeExceptionIsThrown()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Section section = new Section(document);

            //Act
            try
            {
                section.PageNumberingSettings.ChapterHeadingStyleIndex = 10;
                Assert.Fail("Should throw ArgumentOutOfRangeException.");
            }
            catch (ArgumentOutOfRangeException)
            {
            }

            // Assert
            Assert.IsNull(section.PageNumberingSettings.ChapterHeadingStyleIndex);
        }

        [TestMethod]
        public void PageNumberingSettings_WhenTryToSetStartingPageNumberLessThan0_ArgumentOutOfRangeExceptionIsThrown()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Section section = new Section(document);

            //Act
            try
            {
                section.PageNumberingSettings.StartingPageNumber = -1;
                Assert.Fail("Should throw ArgumentOutOfRangeException.");
            }
            catch (ArgumentOutOfRangeException)
            {
            }

            // Assert
            Assert.IsNull(section.PageNumberingSettings.StartingPageNumber);
        }

        [TestMethod]
        public void Rotate_WhenPageOrientationParameterIsNotLandscapeOrPortrait_ArgumentExceptionIsThrown()
        {
            // Arrange
            Section section = new Section(new RadFlowDocument());
            PageOrientation defaultPageOrientation = section.PageOrientation;

            // Act
            try
            {
#pragma warning disable 618
                section.Rotate(PageOrientation.Rotate180);
#pragma warning restore 618
                Assert.Fail("Should throw ArgumentException.");
            }
            catch (ArgumentException)
            {
            }

            // Assert
            Assert.AreEqual(defaultPageOrientation, section.PageOrientation);
        }

        [TestMethod]
        public void RotateToLandscape_WhenPageOrientationIsPortrait_SectionPageOrientationMarginsAndSizeAreChanged()
        {
            // Arrange
            Section section = new Section(new RadFlowDocument());

            Padding pageMargins = new Padding(10, 20, 30, 40);
            section.PageMargins = pageMargins;
            Size pageSize = new Size(100, 50);
            section.PageSize = pageSize;

            // Act
            section.Rotate(PageOrientation.Landscape);

            // Assert
            Assert.AreEqual(PageOrientation.Landscape, section.PageOrientation);

            Assert.AreEqual(pageSize.Height, section.PageSize.Width);
            Assert.AreEqual(pageSize.Width, section.PageSize.Height);

            Assert.AreEqual(pageMargins.Left, section.PageMargins.Top);
            Assert.AreEqual(pageMargins.Top, section.PageMargins.Right);
            Assert.AreEqual(pageMargins.Right, section.PageMargins.Bottom);
            Assert.AreEqual(pageMargins.Bottom, section.PageMargins.Left);
        }

        [TestMethod]
        public void RotateToPortrait_WhenPageOrientationIsLandscape_SectionPageOrientationMarginsAndSizeAreChanged()
        {
            // Arrange
            Section section = new Section(new RadFlowDocument());
            section.PageOrientation = PageOrientation.Landscape;

            Padding pageMargins = new Padding(10, 20, 30, 40);
            section.PageMargins = pageMargins;
            Size pageSize = new Size(100, 50);
            section.PageSize = pageSize;

            // Act
            section.Rotate(PageOrientation.Portrait);

            // Assert
            Assert.AreEqual(PageOrientation.Portrait, section.PageOrientation);

            Assert.AreEqual(pageSize.Height, section.PageSize.Width);
            Assert.AreEqual(pageSize.Width, section.PageSize.Height);

            Assert.AreEqual(pageMargins.Left, section.PageMargins.Bottom);
            Assert.AreEqual(pageMargins.Top, section.PageMargins.Left);
            Assert.AreEqual(pageMargins.Right, section.PageMargins.Top);
            Assert.AreEqual(pageMargins.Bottom, section.PageMargins.Right);
        }

        [TestMethod]
        public void RotateToLandscape_WhenPageOrientationIsLandscape_DoesntChangeAnything()
        {
            // Arrange
            Section section = new Section(new RadFlowDocument());
            section.PageOrientation = PageOrientation.Landscape;

            Padding pageMargins = new Padding(10, 20, 30, 40);
            section.PageMargins = pageMargins;
            Size pageSize = new Size(100, 50);
            section.PageSize = pageSize;

            // Act
            section.Rotate(PageOrientation.Landscape);

            // Assert
            Assert.AreEqual(PageOrientation.Landscape, section.PageOrientation);
            Assert.AreEqual(pageSize, section.PageSize);
            Assert.AreEqual(pageMargins, section.PageMargins);
        }

        [TestMethod]
        public void RotateToPortrait_WhenPageOrientationIsPortrait_DoesntChangeAnything()
        {
            // Arrange
            Section section = new Section(new RadFlowDocument());
            section.PageOrientation = PageOrientation.Portrait;

            Padding pageMargins = new Padding(10, 20, 30, 40);
            section.PageMargins = pageMargins;
            Size pageSize = new Size(100, 50);
            section.PageSize = pageSize;

            // Act
            section.Rotate(PageOrientation.Portrait);

            // Assert
            Assert.AreEqual(PageOrientation.Portrait, section.PageOrientation);
            Assert.AreEqual(pageSize, section.PageSize);
            Assert.AreEqual(pageMargins, section.PageMargins);
        }

        [TestMethod]
        public void Headers_WhenEnumeratingChildrenOfSection_AreEnumerated()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Section section = document.Sections.AddSection();
            section.Headers.Add(HeaderFooterType.First);

            // Act
            int headerCount = section.EnumerateChildrenOfType<Header>().ToList().Count;

            // Assert
            Assert.AreEqual(1, headerCount);
        }

        [TestMethod]
        public void Footers_WhenEnumeratingChildrenOfSection_AreEnumerated()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Section section = document.Sections.AddSection();
            section.Footers.Add(HeaderFooterType.First);

            // Act
            int footerCount = section.EnumerateChildrenOfType<Footer>().ToList().Count;

            // Assert
            Assert.AreEqual(1, footerCount);
        }

#if VISUAL
        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.Section)]
        public void Rotate_WhenTwoSectionsExisting_RotatesOnlyOneSection()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Section firstSection = editor.InsertSection();
            Section secondSection = editor.InsertSection();

            Padding pageMargins = new Padding(10, 20, 30, 40);
            firstSection.PageMargins = pageMargins;
            secondSection.PageMargins = pageMargins;
            Size pageSize = new Size(100, 50);
            firstSection.PageSize = pageSize;
            secondSection.PageSize = pageSize;

            // Act
            firstSection.Rotate(PageOrientation.Landscape);

            // Assert
            Assert.AreEqual(PageOrientation.Portrait, secondSection.PageOrientation);
            Assert.AreEqual(pageSize, secondSection.PageSize);
            Assert.AreEqual(pageMargins, secondSection.PageMargins);

            Assert.AreEqual(PageOrientation.Landscape, firstSection.PageOrientation);
            Assert.AreEqual(pageSize.Height, firstSection.PageSize.Width);
            Assert.AreEqual(pageSize.Width, firstSection.PageSize.Height);
            Assert.AreEqual(pageMargins.Left, firstSection.PageMargins.Top);
            Assert.AreEqual(pageMargins.Top, firstSection.PageMargins.Right);
            Assert.AreEqual(pageMargins.Right, firstSection.PageMargins.Bottom);
            Assert.AreEqual(pageMargins.Bottom, firstSection.PageMargins.Left);

            this.AssertVisual(editor.Document);
        }

        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.Section)]
        public void Rotate_WhenPageOrientationParameterIsLandscape_SectionPageOrientationIsChangedToLandscape()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Section section = document.Sections.AddSection();
            PageOrientation defaultPageOrientation = section.PageOrientation;

            // Act
            section.Rotate(PageOrientation.Landscape);

            // Assert
            Assert.AreEqual(PageOrientation.Portrait, defaultPageOrientation);
            Assert.AreEqual(PageOrientation.Landscape, section.PageOrientation);

            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION, PDFFILEEXTENSION, RTFFILEEXTENSION });
        }

        // Team Pulse bug ID: 141135, TFS bug ID: 225400
        [TestMethod]
        [TestCategory(WordsProcessingCategory.VisualTest)]
        [TestCategory(WordsProcessingCategory.Section)]
        public void SectionBreakIsNotOmittedAfterExport()
        {
            RadFlowDocument document = new RadFlowDocument();

            // Create a table
            Border border = new Border(2, BorderStyle.Single, new ThemableColor(Colors.Black));
            Table table = new Table(document, 3, 3);
            table.Borders = new TableBorders(border);
            // Insert first section in the document
            Section section1 = document.Sections.AddSection();
            section1.Blocks.Add(table);
            section1.SectionType = Documents.Flow.Model.Styles.SectionType.NextPage;

            // Insert second section in the document
            Section section2 = document.Sections.AddSection();
            section2.Blocks.AddParagraph().Inlines.AddRun("Next page");

            // The generated document should have two pages
            this.AssertVisual(document, new string[] { DOCXFILEEXTENSION, PDFFILEEXTENSION, RTFFILEEXTENSION }, UIInteractionHelper.PageDown(), 1, null);
        }
#endif
    }
}