﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Primitives;

namespace Telerik.Windows.Documents.Flow.Tests.Model.Shapes
{
    [TestClass]
    public class FloatingImageTests
    {
        [TestMethod]
        public void CloneWithParameter_WhenInvokedWithNull_ThrowsArgumentNullException()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            FloatingImage floatingImage = new FloatingImage(document1);

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentNullException>(() => floatingImage.Clone(null));
        }

        [TestMethod]
        public void Clone_WithDocument_ClonesTheImageForTheDocument()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            FloatingImage floatingImage = new FloatingImage(document1);

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            FloatingImage clonedFloatingImage = floatingImage.Clone(document2);

            // Assert
            Assert.AreNotSame(floatingImage, clonedFloatingImage);
            Assert.AreSame(document2, clonedFloatingImage.Document);
        }

        [TestMethod]
        public void Clone_WithoutDocument_ClonesTheImageForTheCurrentDocumentAndWithSameImageSource()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            FloatingImage floatingImage = new FloatingImage(document);
            floatingImage.Image.ImageSource = new ImageSource(new byte[] { 1, 2, 3 }, "jpg");

            // Act
            FloatingImage clonedFloatingImage = floatingImage.Clone();

            // Assert
            Assert.AreNotSame(floatingImage, clonedFloatingImage);
            Assert.AreSame(document, clonedFloatingImage.Document);
            Assert.AreSame(floatingImage.Image.ImageSource, clonedFloatingImage.Image.ImageSource);
        }

        [TestMethod]
        public void Clone_WithDocument_ClonesTheImageForTheDocumentAndTheImageSource()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            FloatingImage floatingImage = new FloatingImage(document1);
            floatingImage.Image.ImageSource = new ImageSource(new byte[] { 1, 2, 3 }, "jpg");

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            FloatingImage clonedFloatingImage = floatingImage.Clone(document2);

            // Assert
            Assert.AreNotSame(floatingImage, clonedFloatingImage);
            Assert.AreSame(document2, clonedFloatingImage.Document);
            Assert.AreNotSame(floatingImage.Image.ImageSource, clonedFloatingImage.Image.ImageSource);
            CollectionAssert.AreEqual(new byte[] { 1, 2, 3 }, clonedFloatingImage.Image.ImageSource.Data);
            Assert.AreEqual("jpg", clonedFloatingImage.Image.ImageSource.Extension);
        }

        [TestMethod]
        public void Clone_WhenImageSourceIsNull_ClonesTheImageWithNullImageSource()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            FloatingImage floatingImage = new FloatingImage(document1);

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            FloatingImage clonedFloatingImage = floatingImage.Clone(document2);

            // Assert
            Assert.IsNull(floatingImage.Image.ImageSource);
            Assert.IsNull(clonedFloatingImage.Image.ImageSource);
        }

        [TestMethod]
        public void Clone_WhenPropertiesAreSet_CloningHasSameProperties()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            FloatingImage floatingImage = new FloatingImage(document1);
            floatingImage.AllowOverlap = false;
            floatingImage.HorizontalPosition.Offset = 17;
            floatingImage.Image.Height = 13;
            floatingImage.Image.Width = 15;
            floatingImage.IsBehindDocument = true;
            floatingImage.IsLocked = true;
            floatingImage.LayoutInCell = false;
            floatingImage.Margin = new Padding(5);
            floatingImage.VerticalPosition.Offset = 23;
            floatingImage.Wrapping.TextWrap = TextWrap.LeftOnly;
            floatingImage.ZIndex = 53;

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            FloatingImage clonedFloatingImage = floatingImage.Clone(document2);

            // Assert
            Assert.AreEqual(false, clonedFloatingImage.AllowOverlap);
            Assert.AreNotSame(floatingImage.HorizontalPosition, clonedFloatingImage.HorizontalPosition);
            Assert.AreEqual(17, clonedFloatingImage.HorizontalPosition.Offset);
            Assert.AreNotSame(floatingImage.Image, clonedFloatingImage.Image);
            Assert.AreEqual(13, clonedFloatingImage.Image.Height);
            Assert.AreEqual(15, clonedFloatingImage.Image.Width);
            Assert.AreEqual(true, clonedFloatingImage.IsBehindDocument);
            Assert.AreEqual(true, clonedFloatingImage.IsLocked);
            Assert.AreEqual(false, clonedFloatingImage.LayoutInCell);
            Assert.AreEqual(new Padding(5), clonedFloatingImage.Margin);
            Assert.AreNotSame(floatingImage.VerticalPosition, clonedFloatingImage.VerticalPosition);
            Assert.AreEqual(23, clonedFloatingImage.VerticalPosition.Offset);
            Assert.AreNotSame(floatingImage.Wrapping, clonedFloatingImage.Wrapping);
            Assert.AreEqual(TextWrap.LeftOnly, clonedFloatingImage.Wrapping.TextWrap);
            Assert.AreEqual(53, clonedFloatingImage.ZIndex);
        }
    }
}
