﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
using Telerik.Windows.Documents.Media;

namespace Telerik.Windows.Documents.Flow.Tests.Model.Shapes
{
    [TestClass]
    public class ImageInlineTests
    {
        [TestMethod]
        public void CloneWithParameter_WhenInvokedWithNull_ThrowsArgumentNullException()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            ImageInline imageInline = new ImageInline(document1);

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentNullException>(() => imageInline.Clone(null));
        }

        [TestMethod]
        public void Clone_WithDocument_ClonesTheImageForTheDocument()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            ImageInline imageInline = new ImageInline(document1);

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            ImageInline clonedImageInline = imageInline.Clone(document2);

            // Assert
            Assert.AreNotSame(imageInline, clonedImageInline);
            Assert.AreSame(document2, clonedImageInline.Document);
        }

        [TestMethod]
        public void Clone_WithoutDocumentWithoutImageSource_ClonesTheImageForTheDocument()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            ImageInline imageInline = new ImageInline(document1);

            // Act
            ImageInline clonedImageInline = imageInline.Clone(document1);

            // Assert
            Assert.AreNotSame(imageInline, clonedImageInline);
            Assert.AreSame(document1, clonedImageInline.Document);
        }

        [TestMethod]
        public void Clone_WithoutDocument_ClonesTheImageForTheCurrentDocumentAndWithSameImageSource()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            ImageInline imageInline = new ImageInline(document);
            imageInline.Image.ImageSource = new ImageSource(new byte[] { 1, 2, 3 }, "jpg");

            // Act
            ImageInline clonedImageInline = imageInline.Clone();

            // Assert
            Assert.AreNotSame(imageInline, clonedImageInline);
            Assert.AreSame(document, clonedImageInline.Document);
            Assert.AreSame(imageInline.Image.ImageSource, clonedImageInline.Image.ImageSource);
        }

        [TestMethod]
        public void Clone_WithDocument_ClonesTheImageForTheDocumentAndTheImageSource()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            ImageInline imageInline = new ImageInline(document1);
            imageInline.Image.ImageSource = new ImageSource(new byte[] { 1, 2, 3 }, "jpg");

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            ImageInline clonedImageInline = imageInline.Clone(document2);

            // Assert
            Assert.AreNotSame(imageInline, clonedImageInline);
            Assert.AreSame(document2, clonedImageInline.Document);
            Assert.AreNotSame(imageInline.Image.ImageSource, clonedImageInline.Image.ImageSource);
            CollectionAssert.AreEqual(new byte[] { 1, 2, 3 }, clonedImageInline.Image.ImageSource.Data);
            Assert.AreEqual("jpg", clonedImageInline.Image.ImageSource.Extension);
        }

        [TestMethod]
        public void Clone_WhenImageSourceIsNull_ClonesTheImageWithNullImageSource()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            ImageInline imageInline = new ImageInline(document1);

            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            ImageInline clonedImageInline = imageInline.Clone(document2);

            // Assert
            Assert.IsNull(imageInline.Image.ImageSource);
            Assert.IsNull(clonedImageInline.Image.ImageSource);
        }

        [TestMethod]
        public void SetWidth_WhenValueIsNegative_ThrowsArgumentException()
        {
            // Arrange
            ImageInline image = new ImageInline(new RadFlowDocument());

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(new Action(() =>
            {
                image.Image.SetWidth(false, -1);
            }));
        }

        [TestMethod]
        public void SetHeight_WhenValueIsNegative_ThrowsArgumentException()
        {
            // Arrange
            ImageInline image = new ImageInline(new RadFlowDocument());

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(new Action(() =>
            {
                image.Image.SetHeight(false, -1);
            }));
        }

        [TestMethod]
        public void SetWidth_WhenValueIsNegativeInfinity_ThrowsArgumentException()
        {
            // Arrange
            ImageInline image = new ImageInline(new RadFlowDocument());

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(new Action(() =>
            {
                image.Image.SetWidth(false, double.NegativeInfinity);
            }));
        }

        [TestMethod]
        public void SetHeight_WhenValueIsNegativeInfinity_ThrowsArgumentException()
        {
            // Arrange
            ImageInline image = new ImageInline(new RadFlowDocument());

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(new Action(() =>
            {
                image.Image.SetHeight(false, double.NegativeInfinity);
            }));
        }

        [TestMethod]
        public void SetWidth_WhenValueIsNaN_ThrowsArgumentException()
        {
            // Arrange
            ImageInline image = new ImageInline(new RadFlowDocument());

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(new Action(() =>
            {
                image.Image.SetWidth(false, double.NaN);
            }));
        }

        [TestMethod]
        public void SetHeight_WhenValueIsNaN_ThrowsArgumentException()
        {
            // Arrange
            ImageInline image = new ImageInline(new RadFlowDocument());

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(new Action(() =>
            {
                image.Image.SetHeight(false, double.NaN);
            }));
        }

        [TestMethod]
        public void SetWidth_WhenValueIsPositiveInfinity_ThrowsArgumentException()
        {
            // Arrange
            ImageInline image = new ImageInline(new RadFlowDocument());

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(new Action(() =>
            {
                image.Image.SetWidth(false, double.PositiveInfinity);
            }));
        }

        [TestMethod]
        public void SetHeight_WhenValueIsPositiveInfinity_ThrowsArgumentException()
        {
            // Arrange
            ImageInline image = new ImageInline(new RadFlowDocument());

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(new Action(() =>
            {
                image.Image.SetHeight(false, double.PositiveInfinity);
            }));
        }

        [TestMethod]
        public void InsertInlineImage_WhenWidthIsNaN_ThrowsArgumentException()
        {
            // Arrange
            ImageInline inlineImage = new ImageInline(new RadFlowDocument());

            Action action = new Action(() =>
            {
                inlineImage.Image.Width = double.NaN;
            });

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(action);
        }

        [TestMethod]
        public void InsertInlineImage_WhenHeightIsNaN_ThrowsArgumentException()
        {
            // Arrange
            ImageInline inlineImage = new ImageInline(new RadFlowDocument());

            Action action = new Action(() =>
            {
                inlineImage.Image.Height = double.NaN;
            });

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(action);
        }

        [TestMethod]
        public void InsertInlineImage_WhenWidthIsPositiveInfinity_ThrowsArgumentException()
        {
            // Arrange
            ImageInline inlineImage = new ImageInline(new RadFlowDocument());

            Action action = new Action(() =>
            {
                inlineImage.Image.Width = double.PositiveInfinity;
            });

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(action);
        }

        [TestMethod]
        public void InsertInlineImage_WhenHeightIsPositiveInfinity_ThrowsArgumentException()
        {
            // Arrange
            ImageInline inlineImage = new ImageInline(new RadFlowDocument());

            Action action = new Action(() =>
            {
                inlineImage.Image.Height = double.PositiveInfinity;
            });

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(action);
        }

        [TestMethod]
        public void InsertInlineImage_WhenWidthIsNegativeInfinity_ThrowsArgumentException()
        {
            // Arrange
            ImageInline inlineImage = new ImageInline(new RadFlowDocument());

            Action action = new Action(() =>
            {
                inlineImage.Image.Width = double.NegativeInfinity;
            });

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(action);
        }

        [TestMethod]
        public void InsertInlineImage_WhenHeightIsNegativeInfinity_ThrowsArgumentException()
        {
            // Arrange
            ImageInline inlineImage = new ImageInline(new RadFlowDocument());

            Action action = new Action(() =>
            {
                inlineImage.Image.Height = double.NegativeInfinity;
            });

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(action);
        }

        [TestMethod]
        public void InsertInlineImage_WhenWidthIsLessThan0_ThrowsArgumentException()
        {
            // Arrange
            ImageInline inlineImage = new ImageInline(new RadFlowDocument());

            Action action = new Action(() =>
            {
                inlineImage.Image.Width = -1;
            });

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(action);
        }

        [TestMethod]
        public void InsertInlineImage_WhenHeightIsLessThan0_ThrowsArgumentException()
        {
            // Arrange
            ImageInline inlineImage = new ImageInline(new RadFlowDocument());

            Action action = new Action(() =>
            {
                inlineImage.Image.Height = -1;
            });

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<ArgumentException>(action);
        }
    }
}