﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;

namespace Telerik.Windows.Documents.Flow.Tests.Model.Styles
{
    [TestClass]
    public class LineSpacingTypeStylePropertyTests
    {
        [TestMethod]
        public void DefaultStyleProperties_WhenLineSpacingIsNotDefinedAndLineSpacingTypeIs_LineSpacingTypeIsWithDefaultValue()
        {
            // Arrange & Act
            RadFlowDocument document = new RadFlowDocument();
            document.DefaultStyle.ParagraphProperties.LineSpacing.ClearValue();
            document.DefaultStyle.ParagraphProperties.LineSpacingType.LocalValue = HeightType.Exact;

            //Assert
            Assert.AreEqual(HeightType.Auto, document.DefaultStyle.ParagraphProperties.LineSpacingType.GetActualValue().Value);
        }

        [TestMethod]
        public void DefaultStyleProperties_WhenLineSpacingAndLineSpacingTypeAreDefined_LineSpacingTypeIsWithDefaultValue()
        {
            // Arrange & Act
            RadFlowDocument document = new RadFlowDocument();
            document.DefaultStyle.ParagraphProperties.LineSpacing.LocalValue = 1.15;
            document.DefaultStyle.ParagraphProperties.LineSpacingType.LocalValue = HeightType.Exact;

            //Assert
            Assert.AreEqual(1.15, document.DefaultStyle.ParagraphProperties.LineSpacing.GetActualValue().Value);
            Assert.AreEqual(HeightType.Exact, document.DefaultStyle.ParagraphProperties.LineSpacingType.GetActualValue().Value);
        }

        [TestMethod]
        public void LineSpacingType_WhenHasLocalValueForParagraphAndLineSpacingHasLocalValue_IsWithItsLocalValue()
        {
            // Arrange & Act
            Paragraph paragraph = CreateParagraph();

            paragraph.Properties.LineSpacing.LocalValue = 10;
            paragraph.Properties.LineSpacingType.LocalValue = HeightType.AtLeast;

            // Assert
            Assert.AreEqual(10, paragraph.Spacing.LineSpacing);
            Assert.AreEqual(HeightType.AtLeast, paragraph.Spacing.LineSpacingType);
        }

        [TestMethod]
        public void LineSpacingType_WhenHasNotLocalValueForParagraphAndLineSpacingHasLocalValue_HasValueAuto()
        {
            // Arrange & Act
            Paragraph paragraph = CreateParagraph();

            paragraph.Properties.LineSpacing.LocalValue = 10;

            // Assert
            Assert.AreEqual(10, paragraph.Spacing.LineSpacing);
            Assert.AreEqual(HeightType.Auto, paragraph.Spacing.LineSpacingType);
        }

        [TestMethod]
        public void LineSpacingType_WhenHasLocalValueForParagraphAndLineSpacingHasNotLocalValue_HasValueAuto()
        {
            // Arrange & Act
            Paragraph paragraph = CreateParagraph();

            paragraph.Properties.LineSpacingType.LocalValue = HeightType.AtLeast;

            // Assert
            Assert.AreEqual(HeightType.Auto, paragraph.Spacing.LineSpacingType);
        }

        [TestMethod]
        public void LineSpacingType_WhenHasNotLocalValueForParagraphAndLineSpacingHasNotLocalValue_HasValueAuto()
        {
            // Arrange & Act
            Paragraph paragraph = CreateParagraph();

            // Assert
            Assert.AreEqual(HeightType.Auto, paragraph.Spacing.LineSpacingType);
        }

        [TestMethod]
        public void LineSpacingType_WhenElementHasLocalValueForLineSpacingAndTypeAndHasStyleApplied_ValueIsLocal()
        {
            // Arrange & Act
            Style style = new Style("test", StyleType.Paragraph);
            style.ParagraphProperties.LineSpacing.LocalValue = 2;
            style.ParagraphProperties.LineSpacingType.LocalValue = HeightType.AtLeast;

            Paragraph paragraph = CreateParagraph();
            paragraph.Properties.LineSpacing.LocalValue = 15;
            paragraph.Properties.LineSpacingType.LocalValue = HeightType.Exact;

            paragraph.StyleId = style.Id;
            paragraph.Document.StyleRepository.Add(style);

            // Assert
            Assert.AreEqual(15, paragraph.Spacing.LineSpacing);
            Assert.AreEqual(HeightType.Exact, paragraph.Spacing.LineSpacingType);
        }

        [TestMethod]
        public void LineSpacingType_WhenElementHasLocalValueForLineSpacingTypeButLineSpacingIsDefinedInStyle_ValueIsFromStyle()
        {
            // Arrange & Act
            Style style = new Style("test", StyleType.Paragraph);
            style.ParagraphProperties.LineSpacing.LocalValue = 2;
            style.ParagraphProperties.LineSpacingType.LocalValue = HeightType.AtLeast;

            Paragraph paragraph = CreateParagraph();
            paragraph.Properties.LineSpacingType.LocalValue = HeightType.Exact;

            paragraph.StyleId = style.Id;
            paragraph.Document.StyleRepository.Add(style);

            // Assert
            Assert.AreEqual(2, paragraph.Spacing.LineSpacing);
            Assert.AreEqual(HeightType.AtLeast, paragraph.Spacing.LineSpacingType);
        }

        [TestMethod]
        public void LineSpacingType_WhenElementHasLocalValueForLineSpacingTypeButLineSpacingIsDefinedInBasedOnStyle_ValueIsFromBasedOnStyle()
        {
            // Arrange & Act
            Style baseStyle = new Style("base", StyleType.Paragraph);
            baseStyle.ParagraphProperties.LineSpacing.LocalValue = 2;

            Style style = new Style("test", StyleType.Paragraph);
            style.ParagraphProperties.LineSpacingType.LocalValue = HeightType.AtLeast;
            style.BasedOnStyleId = baseStyle.Id;

            Paragraph paragraph = CreateParagraph();
            paragraph.Properties.LineSpacingType.LocalValue = HeightType.Exact;

            paragraph.StyleId = style.Id;
            paragraph.Document.StyleRepository.Add(style);
            paragraph.Document.StyleRepository.Add(baseStyle);

            // Assert
            Assert.AreEqual(2, paragraph.Spacing.LineSpacing);
            Assert.AreEqual(HeightType.Auto, paragraph.Spacing.LineSpacingType);
        }

        [TestMethod]
        public void LineSpacingType_WhenElementHasLocalValueForLineSpacingTypeInBasedOnStyleButLineSpacingIsDefinedIOnStyle_ValueIsFromStyle()
        {
            // Arrange & Act
            Style baseStyle = new Style("base", StyleType.Paragraph);
            baseStyle.ParagraphProperties.LineSpacingType.LocalValue = HeightType.AtLeast;

            Style style = new Style("test", StyleType.Paragraph);
            style.ParagraphProperties.LineSpacing.LocalValue = 2;
            style.BasedOnStyleId = baseStyle.Id;

            Paragraph paragraph = CreateParagraph();
            paragraph.Properties.LineSpacingType.LocalValue = HeightType.Exact;

            paragraph.StyleId = style.Id;
            paragraph.Document.StyleRepository.Add(style);
            paragraph.Document.StyleRepository.Add(baseStyle);

            // Assert
            Assert.AreEqual(2, paragraph.Spacing.LineSpacing);
            Assert.AreEqual(HeightType.Auto, paragraph.Spacing.LineSpacingType);
        }

        [TestMethod]
        public void LineSpacingType_WhenParagraphElementHasStyleIdWhichRefersToNotAddedStyle_IsEvaluatedProperly()
        {
            // Arrange & Act
            Style style = new Style("test", StyleType.Paragraph);
            style.ParagraphProperties.LineSpacing.LocalValue = 20;
            style.ParagraphProperties.LineSpacingType.LocalValue = HeightType.Exact;

            Paragraph paragraph = CreateParagraph();
            paragraph.StyleId = style.Id;

            // Assert
            Assert.IsFalse(paragraph.Properties.LineSpacing.HasLocalValue);
            Assert.AreEqual(1.15, paragraph.Properties.LineSpacing.GetActualValueAsObject());
            Assert.IsFalse(paragraph.Properties.LineSpacingType.HasLocalValue);
            Assert.AreEqual(HeightType.Auto, paragraph.Properties.LineSpacingType.GetActualValueAsObject());
        }

        [TestMethod]
        public void LineSpacingType_WhenIsSetOnStyleWithoutDocument_HasLocalValue()
        {
            // Arrange & Act
            Style style = new Style("test", StyleType.Paragraph);
            style.ParagraphProperties.LineSpacingType.LocalValue = HeightType.Exact;

            // Assert
            Assert.IsTrue(style.ParagraphProperties.LineSpacingType.HasLocalValue);
            Assert.AreEqual(HeightType.Exact, style.ParagraphProperties.LineSpacingType.GetActualValueAsObject());
        }

        private static Paragraph CreateParagraph()
        {
            return new RadFlowDocument().Sections.AddSection().Blocks.AddParagraph();
        }
    }
}