﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Linq;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Tests.Utilities;

namespace Telerik.Windows.Documents.Flow.Tests.Model.Styles
{
    [TestClass]
    public class StyleRepositoryTests
    {
        [TestMethod]
        public void Add_WhenStyleIsInAnotherDocument_ThrowsInvalidOperationException()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            RadFlowDocument document2 = new RadFlowDocument();

            try
            {
                // Act
                document1.StyleRepository.Add(document2.StyleRepository.GetStyle(BuiltInStyleNames.NormalStyleId));

                // Assert
                Assert.Fail("Should throw exception.");
            }
            catch (InvalidOperationException)
            {
            }
        }

        [TestMethod]
        public void Add_WhenStyleIsNotAssociatedWithDocument_AddsTheStyle()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Style style = new Style("test", StyleType.Paragraph);

            // Act
            document.StyleRepository.Add(style);

            // Assert
            Assert.IsNotNull(document.StyleRepository.GetStyle(style.Id));
        }

        [TestMethod]
        public void ContainsStyleTest()
        {
            RadFlowDocument document = new RadFlowDocument();
            string styleId = "MyCustomStyle";

            Assert.IsFalse(document.StyleRepository.Contains("myCustomStyleId"));

            Style style = new Style(styleId, StyleType.Paragraph);
            document.StyleRepository.Add(style);

            Assert.IsTrue(document.StyleRepository.Contains(styleId));
        }

        [TestMethod]
        public void AddBuiltInStylesTest()
        {
            RadFlowDocument document = new RadFlowDocument();
            string styleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(1);

            Assert.IsNull(document.StyleRepository.GetStyle(styleId));

            document.StyleRepository.AddBuiltInStyle(styleId);

            Assert.IsNotNull(document.StyleRepository.GetStyle(styleId));
        }

        [TestMethod]
        public void RemoveStyle_RemovesStyleAndSetsDocumentToNull()
        {
            RadFlowDocument document = new RadFlowDocument();
            string styleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(1);

            Style heading1 = document.StyleRepository.AddBuiltInStyle(styleId);

            Assert.IsNotNull(heading1);

            document.StyleRepository.Remove(heading1);

            Assert.IsNull(document.StyleRepository.GetStyle(styleId));
            Assert.IsNull(heading1.Document);
        }

        [TestMethod]
        public void Remove_StyleFromAnotherDocument_ThrowsException()
        {
            // Arrange
            RadFlowDocument document1 = new RadFlowDocument();
            RadFlowDocument document2 = new RadFlowDocument();

            // Act
            Style styleFromOtherDocument = document2.StyleRepository.GetStyle(BuiltInStyleNames.NormalStyleId);

            // Act & Assert
            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
            {
                document1.StyleRepository.Remove(styleFromOtherDocument);
            });
        }

        [TestMethod]
        public void ClearStylesTest()
        {
            RadFlowDocument document = new RadFlowDocument();
            string styleId = BuiltInStyleNames.GetHeadingStyleIdByIndex(1);

            document.StyleRepository.AddBuiltInStyle(styleId);

            Assert.IsTrue(document.StyleRepository.Styles.Count() > 0);

            document.StyleRepository.Clear();

            Assert.IsTrue(document.StyleRepository.Styles.Count() == 0);
            Assert.IsNull(document.StyleRepository.GetDefaultStyle(StyleType.Character));
            Assert.IsNull(document.StyleRepository.GetDefaultStyle(StyleType.Paragraph));
            Assert.IsNull(document.StyleRepository.GetDefaultStyle(StyleType.Table));
            Assert.IsNull(document.StyleRepository.GetDefaultStyle(StyleType.Numbering));
        }

        [TestMethod]
        public void GetStyleTest()
        {
            RadFlowDocument document = new RadFlowDocument();
            string styleId = "MyCustomStyle";

            Assert.IsFalse(document.StyleRepository.Contains("myCustomStyleId"));

            Style styleToAdd = new Style(styleId, StyleType.Paragraph);
            document.StyleRepository.Add(styleToAdd);

            Style style = document.StyleRepository.GetStyle(styleId);

            Assert.IsNotNull(style);
            Assert.IsTrue(style.Id == styleToAdd.Id);
        }

        [TestMethod]
        public void ChangeStyleId_WhenStyleIsAdded_ThrowsException()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Style style = new Style("TestStyle", StyleType.Paragraph);
            document.StyleRepository.Add(style);

            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
            {
                style.Id = "Changed";
            });
        }

        [TestMethod]
        public void ChangeIsDefault_WhenStyleIsAdded_ThrowsException()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Style style = new Style("TestStyle", StyleType.Paragraph);
            document.StyleRepository.Add(style);

            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
            {
                style.IsDefault = true;
            });
        }
    }
}
