﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Linq;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Model.Styles;

namespace Telerik.Windows.Documents.Flow.Tests.Model.Styles
{
    [TestClass]
    public class TabStopsTests
    {
        [TestMethod]
        public void AddTabStopToTabStopsCollection_WhenSameTabStopIsAddedMultipleTimes_TabStopIsAddedMultipleTimes()
        {
            // Arrange
            TabStop tabStop = new TabStop(2.5, TabStopType.Left, TabStopLeader.Underscore);

            // Act
            TabStopCollection tabStops = new TabStopCollection().Insert(tabStop)
                                                                .Insert(tabStop)
                                                                .Insert(tabStop);

            // Assert
            Assert.AreEqual(3, tabStops.Count);
            foreach (TabStop ts in tabStops)
            {
                Assert.AreEqual(ts, tabStop);
            }
        }

        [TestMethod]
        public void RemoveTabStopFromTabStopsCollection_WhenSameTabStopIsAddedMultipleTimes_RemovesOnlySingleTabStopFromCollection()
        {
            // Arrange
            TabStop tabStop = new TabStop(2.5, TabStopType.Left, TabStopLeader.Underscore);

            TabStopCollection tabStops = new TabStopCollection().Insert(tabStop)
                                                                .Insert(tabStop)
                                                                .Insert(tabStop);
            // Act
            tabStops = tabStops.Remove(tabStop);

            // Assert
            Assert.AreEqual(2, tabStops.Count);
            foreach (TabStop ts in tabStops)
            {
                Assert.AreEqual(ts, tabStop);
            }
        }

        [TestMethod]
        public void AddTabStop_DoNotAddTheTabStopToTheSameCollectionButToTheNewlyReturned()
        {
            // Arrange
            TabStopCollection tabStops = new TabStopCollection();

            // Act
            TabStopCollection returnedCollection = tabStops.Insert(new TabStop(5, TabStopType.Bar, TabStopLeader.Hyphen));

            // Assert
            Assert.AreEqual(0, tabStops.Count);
            Assert.AreEqual(1, returnedCollection.Count);
        }

        [TestMethod]
        public void TabStopsCollectionActualValue_WhenParagraphHasTabStopsFromStyleAndLocalValueIsNull_IsFromStyle()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Style style = new Style("test", StyleType.Paragraph);
            style.ParagraphProperties.TabStops.LocalValue = new TabStopCollection().Insert(new TabStop(100, TabStopType.Left, TabStopLeader.Dot));
            document.StyleRepository.Add(style);

            Paragraph paragraph = new Paragraph(document);
            paragraph.StyleId = style.Id;

            // Act
            TabStopCollection tabStops = paragraph.TabStops;

            // Assert
            Assert.AreEqual(1, tabStops.Count);
            Assert.AreEqual(new TabStop(100, TabStopType.Left, TabStopLeader.Dot), tabStops.FirstOrDefault());
        }

        [TestMethod]
        public void TabStopsCollectionActualValue_WhenParagraphHasTabStopsFromLocalValueAndStyle_IsCombinedFromLocalAndStyle()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Style style = new Style("test", StyleType.Paragraph);
            style.ParagraphProperties.TabStops.LocalValue = new TabStopCollection().Insert(new TabStop(100, TabStopType.Left, TabStopLeader.Dot));
            document.StyleRepository.Add(style);

            Paragraph paragraph = new Paragraph(document);
            paragraph.StyleId = style.Id;
            paragraph.Properties.TabStops.LocalValue = new TabStopCollection().Insert(new TabStop(200, TabStopType.Decimal, TabStopLeader.Dot));

            // Act
            TabStopCollection tabStops = paragraph.TabStops;

            // Assert
            Assert.AreEqual(2, tabStops.Count);
            Assert.AreEqual(new TabStop(100, TabStopType.Left, TabStopLeader.Dot), tabStops.FirstOrDefault());
            Assert.AreEqual(new TabStop(200, TabStopType.Decimal, TabStopLeader.Dot), tabStops.LastOrDefault());
        }

        [TestMethod]
        public void TabStopsCollectionActualValue_WhenParagraphHasTabStopsFromLocalValueAndStyleAndBasedStyle_IsCombinedAll()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Style baseStyle = new Style("base", StyleType.Paragraph);
            baseStyle.ParagraphProperties.TabStops.LocalValue = new TabStopCollection().Insert(new TabStop(100, TabStopType.Right));
            document.StyleRepository.Add(baseStyle);

            Style style = new Style("test", StyleType.Paragraph);
            style.BasedOnStyleId = baseStyle.Id;
            style.ParagraphProperties.TabStops.LocalValue = new TabStopCollection().Insert(new TabStop(100, TabStopType.Left, TabStopLeader.Dot));
            document.StyleRepository.Add(style);

            Paragraph paragraph = new Paragraph(document);
            paragraph.StyleId = style.Id;
            paragraph.Properties.TabStops.LocalValue = new TabStopCollection().Insert(new TabStop(200, TabStopType.Decimal, TabStopLeader.Dot));

            // Act
            TabStopCollection tabStops = paragraph.TabStops;

            // Assert
            Assert.AreEqual(3, tabStops.Count);
            Assert.AreEqual(new TabStop(100, TabStopType.Right), tabStops.FirstOrDefault());
            Assert.AreEqual(new TabStop(100, TabStopType.Left, TabStopLeader.Dot), tabStops.ElementAt(1));
            Assert.AreEqual(new TabStop(200, TabStopType.Decimal, TabStopLeader.Dot), tabStops.LastOrDefault());
        }

        [TestMethod]
        public void TabStopCollection_WhenHasEqualElementsOnDifferentPositions_EqualsReturnsTrue()
        {
            // Arrange
            TabStop stop1 = new TabStop(100);
            TabStop stop2 = new TabStop(100, TabStopType.Right);
            TabStop stop3 = new TabStop(100, TabStopType.Right, TabStopLeader.MiddleDot);

            // Act
            TabStopCollection collection1 = new TabStopCollection().Insert(stop1)
                                                                   .Insert(stop2)
                                                                   .Insert(stop3);

            TabStopCollection collection2 = new TabStopCollection().Insert(stop3)
                                                                   .Insert(stop1)
                                                                   .Insert(stop2);

            // Assert
            Assert.AreEqual(collection1, collection2);
        }

        [TestMethod]
        public void TabStopCollectionInheritance_WhenParagraphTabStopCollectionIsModified_DefaultTabStopCollectionIsNotModified()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            TabStop defaultTabStop = new TabStop(4, TabStopType.Center);
            document.DefaultStyle.ParagraphProperties.TabStops.LocalValue = new TabStopCollection().Insert(defaultTabStop);

            Paragraph paragraph = editor.InsertParagraph();
            TabStopCollection actualTabStopCollection = paragraph.TabStops;

            Assert.AreEqual(document.DefaultStyle.ParagraphProperties.TabStops.LocalValue, actualTabStopCollection);

            // Act
            TabStop newTabStop = new TabStop(10, TabStopType.Left);
            paragraph.TabStops = new TabStopCollection().Insert(newTabStop);

            // Assert
            Assert.AreNotEqual(document.DefaultStyle.ParagraphProperties.TabStops.LocalValue, paragraph.TabStops);
        }

        [TestMethod]
        public void ModifyDefaultTabStopCollection_NothingIsChanged()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Assert.AreEqual(0, document.DefaultStyle.ParagraphProperties.TabStops.GetActualValue().Count);

            // Act
            Paragraph.TabStopsPropertyDefinition.DefaultValue.Insert(new TabStop(3));

            // Assert
            Assert.AreEqual(0, document.DefaultStyle.ParagraphProperties.TabStops.GetActualValue().Count);
        }

        [TestMethod]
        public void TabStopInheritance_WhenParagraphTabStopIsModified_DefatulTabStopIsNotModified()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            TabStop defaultTabStop = new TabStop(4, TabStopType.Center);
            document.DefaultStyle.ParagraphProperties.TabStops.LocalValue = new TabStopCollection().Insert(defaultTabStop);

            Paragraph paragraph = editor.InsertParagraph();
            TabStop actualTabStop = paragraph.TabStops.First();

            Assert.AreEqual(defaultTabStop, actualTabStop);

            TabStop newTabStop = new TabStop(10, TabStopType.Left);
            paragraph.TabStops = new TabStopCollection().Insert(newTabStop);

            TabStop actualDefaultTabStop = document.DefaultStyle.ParagraphProperties.TabStops.LocalValue.First();

            Assert.AreEqual(defaultTabStop, actualDefaultTabStop);
        }
    }
}