﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
using Telerik.Windows.Documents.Media;

namespace Telerik.Windows.Documents.Flow.Tests
{
    [TestClass]
    public class TestSimpleStyleProperties
    {
        [TestMethod]
        public void TestInitialValuesInCharacterProperties()
        {
            RadFlowDocument document = new RadFlowDocument();
            Section section = document.Sections.AddSection();
            Paragraph paragraph = section.Blocks.AddParagraph();
            Run run = paragraph.Inlines.AddRun();

            Assert.IsFalse(run.Properties.FontSize.HasLocalValue);
            Assert.AreEqual(run.FontSize, Unit.PointToDip(11));

            Assert.IsFalse(run.Properties.FontFamily.HasLocalValue);

            Assert.IsTrue(run.Properties.FontSize.LocalValue.HasValue == false);
            Assert.IsTrue(run.Properties.FontFamily.LocalValue == null);
        }

        [TestMethod]
        public void TestInitialValuesInParagraphProperties()
        {
            RadFlowDocument document = new RadFlowDocument();
            Section section = document.Sections.AddSection();
            Paragraph paragraph = section.Blocks.AddParagraph();

            Assert.IsFalse(paragraph.Properties.TextAlignment.HasLocalValue);
            Assert.IsTrue(paragraph.Properties.TextAlignment.LocalValue == null);
        }

        [TestMethod]
        public void TestCharacterPropertiesCopyPropertiesFrom()
        {
            RadFlowDocument document = new RadFlowDocument();
            Section section = document.Sections.AddSection();
            Paragraph paragraph = section.Blocks.AddParagraph();
            Run firstRun = paragraph.Inlines.AddRun();
            firstRun.FontSize = 44;
            firstRun.StyleId = "hello";

            Run secondRun = paragraph.Inlines.AddRun();

            secondRun.Properties.CopyPropertiesFrom(firstRun.Properties);

            StylePropertyAsserter.AssertPropertiesAreEqual(firstRun.Properties, secondRun.Properties);
        }

        [TestMethod]
        public void TestParagraphPropertiesCopyPropertiesFrom()
        {
            RadFlowDocument document = new RadFlowDocument();
            Section section = document.Sections.AddSection();
            Paragraph firstParagraph = section.Blocks.AddParagraph();
            Paragraph secondParagraph = section.Blocks.AddParagraph();

            firstParagraph.Spacing.SpacingAfter = 14;
            firstParagraph.StyleId = "hello";

            secondParagraph.Properties.CopyPropertiesFrom(firstParagraph.Properties);

            StylePropertyAsserter.AssertPropertiesAreEqual(firstParagraph.Properties, secondParagraph.Properties);
        }

        [TestMethod]
        public void TestTablePropertiesCopyPropertiesFrom()
        {
            RadFlowDocument document = new RadFlowDocument();
            Table firstTable = document.Sections.AddSection().Blocks.AddTable();
            Table secondTable = document.Sections.AddSection().Blocks.AddTable();

            firstTable.Alignment = Alignment.Center;
            firstTable.StyleId = "hello";

            secondTable.Properties.CopyPropertiesFrom(firstTable.Properties);

            StylePropertyAsserter.AssertPropertiesAreEqual(firstTable.Properties, secondTable.Properties);
        }

        [TestMethod]
        public void TestTableRowPropertiesCopyPropertiesFrom()
        {
            RadFlowDocument document = new RadFlowDocument();
            TableRow firstRow = document.Sections.AddSection().Blocks.AddTable().Rows.AddTableRow();
            TableRow secondRow = document.Sections.AddSection().Blocks.AddTable().Rows.AddTableRow();

            firstRow.RepeatOnEveryPage = true;

            secondRow.Properties.CopyPropertiesFrom(firstRow.Properties);

            StylePropertyAsserter.AssertPropertiesAreEqual(firstRow.Properties, secondRow.Properties);
        }

        [TestMethod]
        public void TestTableCellPropertiesCopyPropertiesFrom()
        {
            RadFlowDocument document = new RadFlowDocument();
            TableCell firstCell = document.Sections.AddSection().Blocks.AddTable().Rows.AddTableRow().Cells.AddTableCell();
            TableCell secondCell = document.Sections.AddSection().Blocks.AddTable().Rows.AddTableRow().Cells.AddTableCell();

            firstCell.VerticalAlignment = VerticalAlignment.Center;

            secondCell.Properties.CopyPropertiesFrom(firstCell.Properties);

            StylePropertyAsserter.AssertPropertiesAreEqual(firstCell.Properties, secondCell.Properties);
        }

        [TestMethod]
        public void TestCopyProperties_FromCharacterPropertiesToParagraphProperties()
        {
            RadFlowDocument document = new RadFlowDocument();
            Section section = document.Sections.AddSection();
            Paragraph paragraph = section.Blocks.AddParagraph();
            Run run = paragraph.Inlines.AddRun("text");

            paragraph.Properties.CopyPropertiesFrom(run.Properties);
        }

        [TestMethod]
        public void TestEnumerableProperties()
        {
            RadFlowDocument document = new RadFlowDocument();
            Section section = document.Sections.AddSection();
            Paragraph paragraph = section.Blocks.AddParagraph();

            bool hasIterated = false;
            foreach (IStyleProperty property in paragraph.Properties.StyleProperties)
            {
                hasIterated = true;
            }

            Assert.IsTrue(hasIterated);
        }

        [TestMethod]
        public void TestGetStylePropertyByStylePropertyDefinition_Of_DocumentProperties()
        {
            RadFlowDocument document = new RadFlowDocument();

            IStyleProperty styleProperty = document.Properties.GetStyleProperty(Paragraph.AllowOverflowPunctuationPropertyDefinition);
            Assert.IsNull(styleProperty);

            styleProperty = document.Properties.GetStyleProperty(RadFlowDocument.HasDifferentEvenOddPageHeadersFootersPropertyDefinition);
            Assert.IsNotNull(styleProperty);
        }

        [TestMethod]
        public void TestGetStylePropertyByStylePropertyDefinition_Of_SectionProperties()
        {
            RadFlowDocument document = new RadFlowDocument();
            Section section = document.Sections.AddSection();

            IStyleProperty styleProperty = section.Properties.GetStyleProperty(Paragraph.AllowOverflowPunctuationPropertyDefinition);
            Assert.IsNull(styleProperty);

            styleProperty = section.Properties.GetStyleProperty(Section.FooterBottomMarginPropertyDefinition);
            Assert.IsNotNull(styleProperty);
        }

        [TestMethod]
        public void TestGetStylePropertyByStylePropertyDefinition_Of_CharacterProperties()
        {
            RadFlowDocument document = new RadFlowDocument();
            Section section = document.Sections.AddSection();
            Paragraph paragraph = section.Blocks.AddParagraph();
            Run run = paragraph.Inlines.AddRun("text");

            IStyleProperty styleProperty = run.Properties.GetStyleProperty(Paragraph.AllowOverflowPunctuationPropertyDefinition);
            Assert.IsNull(styleProperty);

            styleProperty = run.Properties.GetStyleProperty(Run.BackgroundColorPropertyDefinition);
            Assert.IsNotNull(styleProperty);
        }

        [TestMethod]
        public void TestGetStylePropertyByStylePropertyDefinition_Of_ParagraphProperties()
        {
            RadFlowDocument document = new RadFlowDocument();
            Section section = document.Sections.AddSection();
            Paragraph paragraph = section.Blocks.AddParagraph();
            Run run = paragraph.Inlines.AddRun("text");

            IStyleProperty styleProperty = paragraph.Properties.GetStyleProperty(Run.BackgroundColorPropertyDefinition);
            Assert.IsNull(styleProperty);

            styleProperty = paragraph.Properties.GetStyleProperty(Paragraph.AllowOverflowPunctuationPropertyDefinition);
            Assert.IsNotNull(styleProperty);
        }

        [TestMethod]
        public void TestGetStylePropertyByStylePropertyDefinition_Of_TableProperties()
        {
            RadFlowDocument document = new RadFlowDocument();
            Section section = document.Sections.AddSection();
            Table table = section.Blocks.AddTable();

            IStyleProperty styleProperty = table.Properties.GetStyleProperty(Run.BackgroundColorPropertyDefinition);
            Assert.IsNull(styleProperty);

            styleProperty = table.Properties.GetStyleProperty(Table.ColumnBandingPropertyDefinition);
            Assert.IsNotNull(styleProperty);
        }

        [TestMethod]
        public void TestGetStylePropertyByStylePropertyDefinition_Of_TableRowProperties()
        {
            RadFlowDocument document = new RadFlowDocument();
            Section section = document.Sections.AddSection();
            Table table = section.Blocks.AddTable();
            TableRow row = table.Rows.AddTableRow();

            IStyleProperty styleProperty = row.Properties.GetStyleProperty(Run.BackgroundColorPropertyDefinition);
            Assert.IsNull(styleProperty);

            styleProperty = row.Properties.GetStyleProperty(TableRow.CanSplitPropertyDefinition);
            Assert.IsNotNull(styleProperty);
        }

        [TestMethod]
        public void TestGetStylePropertyByStylePropertyDefinition_Of_TableCellProperties()
        {
            RadFlowDocument document = new RadFlowDocument();
            Section section = document.Sections.AddSection();
            Table table = section.Blocks.AddTable();
            TableRow row = table.Rows.AddTableRow();
            TableCell cell = row.Cells.AddTableCell();

            IStyleProperty styleProperty = cell.Properties.GetStyleProperty(Run.BackgroundColorPropertyDefinition);
            Assert.IsNull(styleProperty);

            styleProperty = cell.Properties.GetStyleProperty(TableCell.CanWrapContentPropertyDefinition);
            Assert.IsNotNull(styleProperty);
        }
    }
}
