﻿using System;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Media;
#else
using System.Windows;
using System.Windows.Media;
#endif
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
#if WPF || SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif
using Style = Telerik.Windows.Documents.Flow.Model.Styles.Style;
using Telerik.Windows.Documents.Flow.Model.Editing;

namespace Telerik.Windows.Documents.Flow.Tests.Model.Styles
{
    [TestClass]
    public class TestStyleInheritance
    {
        [TestMethod]
        [ExpectedException(typeof(InvalidOperationException))]
        public void TestCyclicInheritanceDetection()
        {
            RadFlowDocument document = new RadFlowDocument();

            Style style1 = new Style("style1", StyleType.Character);
            Style style2 = new Style("style2", StyleType.Character);
            Style style3 = new Style("style3", StyleType.Character);

            document.StyleRepository.Add(style1);
            document.StyleRepository.Add(style2);
            document.StyleRepository.Add(style3);

            style2.BasedOnStyleId = style1.Id;
            style3.BasedOnStyleId = style2.Id;
            style1.BasedOnStyleId = style3.Id;
            Assert.IsTrue(false, "A cyclic exception should be thrown!");
        }

        [TestMethod]
        public void TestSetValue()
        {
            FontWeight localValue = FontWeights.Bold;

            Style style = new Style("id", StyleType.Character);
            Assert.IsFalse(style.CharacterProperties.FontWeight.HasLocalValue);

            style.CharacterProperties.FontWeight.LocalValue = FontWeights.Bold;
            Assert.IsTrue(style.CharacterProperties.FontWeight.HasLocalValue);
            StylePropertyAsserter.ConfirmValues(style.CharacterProperties.FontWeight.LocalValue, localValue);
        }

        [TestMethod]
        public void TestClearValue()
        {
            Style style = new Style("id", StyleType.Character);
            Assert.IsFalse(style.CharacterProperties.FontWeight.HasLocalValue);

            style.CharacterProperties.FontWeight.LocalValue = FontWeights.Bold;
            Assert.IsTrue(style.CharacterProperties.FontWeight.HasLocalValue);

            style.CharacterProperties.FontWeight.ClearValue();
            Assert.IsFalse(style.CharacterProperties.FontWeight.HasLocalValue);
        }

        [TestMethod]
        public void TestBasicBasedOnInheritance()
        {
            RadFlowDocument document = new RadFlowDocument();

            Style characterStyle = new Style("characterStyle", StyleType.Character);
            Style characterBasedOnStyle = new Style("characterBasedOnStyle", StyleType.Character);

            characterBasedOnStyle.CharacterProperties.FontWeight.LocalValue = FontWeights.Bold;
            characterStyle.BasedOnStyleId = characterBasedOnStyle.Id;

            document.StyleRepository.Add(characterStyle);
            document.StyleRepository.Add(characterBasedOnStyle);

            StylePropertyAsserter.AssertPropertyValue(characterStyle, Run.FontWeightPropertyDefinition, FontWeights.Bold);
        }

        [TestMethod]
        public void TestBasicBasedOnInheritanceWithLocalValueOnTheChildStyle()
        {
            FontWeight expectedValue = FontWeights.Normal;
            RadFlowDocument document = new RadFlowDocument();

            Style characterStyle = new Style("characterStyle", StyleType.Character);
            Style characterBasedOnStyle = new Style("characterBasedOnStyle", StyleType.Character);

            characterStyle.CharacterProperties.FontWeight.LocalValue = expectedValue;
            characterBasedOnStyle.CharacterProperties.FontWeight.LocalValue = FontWeights.Bold;
            characterStyle.BasedOnStyleId = characterBasedOnStyle.Id;

            document.StyleRepository.Add(characterStyle);
            document.StyleRepository.Add(characterBasedOnStyle);

            StylePropertyAsserter.AssertPropertyValue(characterStyle, Run.FontWeightPropertyDefinition, expectedValue);
        }

        [TestMethod]
        public void TestMultipleBasedOnInheritanceWithLocalValueOnTheChildStyle()
        {
            FontWeight expectedValue = FontWeights.Bold;
            RadFlowDocument document = new RadFlowDocument();

            Style childStyle = new Style("childStyle", StyleType.Character);
            Style parentStyle = new Style("parentStyle", StyleType.Character);
            Style grandParentStyle = new Style("grandParentStyle", StyleType.Character);
            Style grandGrandParentStyle = new Style("grandGrandParentStyle", StyleType.Character);

            grandGrandParentStyle.CharacterProperties.FontWeight.LocalValue = expectedValue;
            childStyle.BasedOnStyleId = parentStyle.Id;
            parentStyle.BasedOnStyleId = grandParentStyle.Id;
            grandParentStyle.BasedOnStyleId = grandGrandParentStyle.Id;

            document.StyleRepository.Add(childStyle);
            document.StyleRepository.Add(parentStyle);
            document.StyleRepository.Add(grandParentStyle);
            document.StyleRepository.Add(grandGrandParentStyle);

            StylePropertyAsserter.AssertPropertyValue(childStyle, Run.FontWeightPropertyDefinition, expectedValue);
        }

        [TestMethod]
        public void TestInheritanceFromDocumentDefaultValuesToStyle()
        {
            ThemableColor expectedValue = new ThemableColor(Colors.Red);
            RadFlowDocument document = new RadFlowDocument();
            document.DefaultStyle.CharacterProperties.ForegroundColor.LocalValue = new ThemableColor(Colors.Red);

            Style paragraphStyle = new Style("ParagraphStyle", StyleType.Paragraph);
            Style paragraphMidleStyle = new Style("ParagraphMidleStyle", StyleType.Paragraph);
            Style paragraphBasedOnStyle = new Style("ParagraphBasedOnStyle", StyleType.Paragraph);

            paragraphStyle.BasedOnStyleId = paragraphMidleStyle.Id;
            paragraphMidleStyle.BasedOnStyleId = paragraphBasedOnStyle.Id;

            document.StyleRepository.Add(paragraphStyle);
            document.StyleRepository.Add(paragraphMidleStyle);
            document.StyleRepository.Add(paragraphBasedOnStyle);

            StylePropertyAsserter.AssertPropertyValue(paragraphStyle, Run.ForegroundColorPropertyDefinition, expectedValue);
        }

        [TestMethod]
        public void TestInheritanceFromDocumentDefaultValuesToParagraphWithStyle()
        {
            Alignment expectedValue = Alignment.Right;
            RadFlowDocument document = new RadFlowDocument();
            document.DefaultStyle.ParagraphProperties.TextAlignment.LocalValue = Alignment.Right;

            Style paragraphStyle = new Style("ParagraphStyle", StyleType.Paragraph);
            Style paragraphMidleStyle = new Style("ParagraphMidleStyle", StyleType.Paragraph);
            Style paragraphBasedOnStyle = new Style("ParagraphBasedOnStyle", StyleType.Paragraph);

            paragraphStyle.BasedOnStyleId = paragraphMidleStyle.Id;
            paragraphMidleStyle.BasedOnStyleId = paragraphBasedOnStyle.Id;

            document.StyleRepository.Add(paragraphStyle);
            document.StyleRepository.Add(paragraphMidleStyle);
            document.StyleRepository.Add(paragraphBasedOnStyle);


            Paragraph paragraph = document.Sections.AddSection().Blocks.AddParagraph();
            paragraph.StyleId = paragraphStyle.Id;

            StylePropertyAsserter.AssertPropertyValue(paragraph, Paragraph.TextAlignmentPropertyDefinition, expectedValue);
        }

        [TestMethod]
        public void TestInheritanceFromDocumentDefaultValuesToSpanWithStyle()
        {
            RadFlowDocument document = new RadFlowDocument();
            document.DefaultStyle.CharacterProperties.ForegroundColor.LocalValue = new ThemableColor(Colors.Red);

            Style paragraphStyle = new Style("ParagraphStyle", StyleType.Paragraph);
            Style paragraphMidleStyle = new Style("ParagraphMidleStyle", StyleType.Paragraph);
            Style paragraphBasedOnStyle = new Style("ParagraphBasedOnStyle", StyleType.Paragraph);

            paragraphStyle.BasedOnStyleId = paragraphMidleStyle.Id;
            paragraphMidleStyle.BasedOnStyleId = paragraphBasedOnStyle.Id;

            document.StyleRepository.Add(paragraphStyle);
            document.StyleRepository.Add(paragraphMidleStyle);
            document.StyleRepository.Add(paragraphBasedOnStyle);

            Paragraph paragraph = document.Sections.AddSection().Blocks.AddParagraph();
            paragraph.StyleId = paragraphStyle.Id;
            Run run = paragraph.Inlines.AddRun();

            StylePropertyAsserter.AssertPropertyValue(run, Run.ForegroundColorPropertyDefinition,
                document.DefaultStyle.CharacterProperties.ForegroundColor.GetActualValue());
        }

        [TestMethod]
        public void TestElementsWithStyleHasNoLocalValue()
        {
            RadFlowDocument document = new RadFlowDocument();

            StylePropertyAsserter.HasNoLocalProperties(new Run(document).Properties.StyleProperties);
            StylePropertyAsserter.HasNoLocalProperties(new Paragraph(document).Properties.StyleProperties);
            StylePropertyAsserter.HasNoLocalProperties(new Table(document).Properties.StyleProperties);
            StylePropertyAsserter.HasNoLocalProperties(new TableRow(document).Properties.StyleProperties);
            StylePropertyAsserter.HasNoLocalProperties(new TableCell(document).Properties.StyleProperties);
        }

        [TestMethod]
        public void TestInheritanceOfParagraphStyle()
        {
            Alignment expectedValue = Alignment.Right;
            RadFlowDocument document = new RadFlowDocument();

            Style paragraphStyle = new Style("ParagraphStyle", StyleType.Paragraph);
            Style paragraphBasedOnStyle = new Style("ParagraphBasedOnStyle", StyleType.Paragraph);

            paragraphBasedOnStyle.ParagraphProperties.TextAlignment.LocalValue = expectedValue;
            paragraphStyle.BasedOnStyleId = paragraphBasedOnStyle.Id;

            document.StyleRepository.Add(paragraphStyle);
            document.StyleRepository.Add(paragraphBasedOnStyle);

            StylePropertyAsserter.AssertPropertyValue(paragraphStyle, Paragraph.TextAlignmentPropertyDefinition, expectedValue);
        }

        [TestMethod]
        public void TestInheritanceOfTableStyle()
        {
            Alignment expectedValue = Alignment.Right;
            RadFlowDocument document = new RadFlowDocument();

            Style tableStyle = new Style("tableStyle", StyleType.Table);
            Style tableBasedOnStyle = new Style("tableBasedOnStyle", StyleType.Table);

            tableBasedOnStyle.ParagraphProperties.TextAlignment.LocalValue = expectedValue;
            tableStyle.BasedOnStyleId = tableBasedOnStyle.Id;

            document.StyleRepository.Add(tableStyle);
            document.StyleRepository.Add(tableBasedOnStyle);

            StylePropertyAsserter.AssertPropertyValue(tableStyle, Paragraph.TextAlignmentPropertyDefinition, expectedValue);
        }

        [TestMethod]
        public void TestCopyPropertiesForCharacterProperties()
        {
            RadFlowDocument document = new RadFlowDocument();
            Run run = document.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun();
            run.Text = "hello world";

            run.FontFamily = new ThemableFontFamily("Calibri");
            run.BaselineAlignment = Flow.Model.Styles.BaselineAlignment.Subscript;
            run.FontSize = 48;
            run.FontStyle = FontStyles.Italic;
            run.FontWeight = FontWeights.Bold;
            run.ForegroundColor = new ThemableColor(Colors.Red);
            run.HighlightColor = Colors.Blue;
            run.Strikethrough = true;
            run.Underline.Color = new ThemableColor(Colors.Purple);
            run.Underline.Pattern = UnderlinePattern.Dash;
            run.Shading.BackgroundColor = new ThemableColor(Colors.Pink);
            run.Shading.Pattern = ShadingPattern.Percent10;
            run.Shading.PatternColor = new ThemableColor(Colors.Cyan);

            Run newRun = document.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun();
            newRun.Properties.CopyPropertiesFrom(run.Properties);

            StylePropertyAsserter.VerifyCopyProperties(run.Properties, newRun.Properties);
        }

        [TestMethod]
        public void TestCopyPropertiesForParagraphProperties()
        {
            RadFlowDocument document = new RadFlowDocument();
            Paragraph paragraph = document.Sections.AddSection().Blocks.AddParagraph();

            paragraph.FlowDirection = Flow.Model.Styles.FlowDirection.RightToLeft;
            paragraph.AllowOverflowPunctuation = true;
            paragraph.ApplyEastAsianLineBreakingRules = true;
            paragraph.Borders = new ParagraphBorders(new Border(1, BorderStyle.Single, new ThemableColor(Colors.Red)));
            paragraph.ContextualSpacing = true;
            paragraph.Indentation.FirstLineIndent = 12;
            paragraph.Indentation.HangingIndent = 12;
            paragraph.Indentation.LeftIndent = 8;
            paragraph.Indentation.RightIndent = 8;
            paragraph.KeepOnOnePage = true;
            paragraph.KeepWithNextParagraph = true;
            // TODO: Uncomment when list styles are implemented.            
            //paragraph.ListLevel = 2;
            paragraph.MirrorIndents = true;
            paragraph.OutlineLevel = OutlineLevel.Level3;
            paragraph.PageBreakBefore = true;
            paragraph.Shading.BackgroundColor = new ThemableColor(Colors.Pink);
            paragraph.Shading.Pattern = ShadingPattern.Percent10;
            paragraph.Shading.PatternColor = new ThemableColor(Colors.Cyan);
            paragraph.Spacing.SpacingAfter = 10;
            paragraph.Spacing.SpacingBefore = 10;
            paragraph.Spacing.LineSpacing = 1.5;
            paragraph.Spacing.LineSpacingType = HeightType.Exact;
            paragraph.Spacing.AutomaticSpacingAfter = false;
            paragraph.Spacing.AutomaticSpacingBefore = false;

            Paragraph newParagraph = document.Sections.AddSection().Blocks.AddParagraph();
            newParagraph.Properties.CopyPropertiesFrom(paragraph.Properties);

            StylePropertyAsserter.VerifyCopyProperties(paragraph.Properties, newParagraph.Properties);
        }

        [TestMethod]
        public void TestCopyPropertiesForTableProperties()
        {
            RadFlowDocument document = new RadFlowDocument();
            Table table = document.Sections.AddSection().Blocks.AddTable();

            table.Alignment = Alignment.Center;
            table.Borders = new TableBorders(new Border(1, BorderStyle.Single, new ThemableColor(Colors.Red)));
            // TODO Implement.
            //table.FloatingPosition = new FloatingPosition();
            table.FlowDirection = Flow.Model.Styles.FlowDirection.RightToLeft;
            table.Indent = 12;
            table.LayoutType = TableLayoutType.FixedWidth;
            table.PreferredWidth = new TableWidthUnit(TableWidthUnitType.Fixed, 680);
            table.Overlap = true;
            table.Looks = TableLooks.FirstColumn | TableLooks.FirstRow;
            table.Shading.BackgroundColor = new ThemableColor(Colors.Pink);
            table.Shading.Pattern = ShadingPattern.Percent10;
            table.Shading.PatternColor = new ThemableColor(Colors.Cyan);

            Table newTable = document.Sections.AddSection().Blocks.AddTable();

            newTable.Properties.CopyPropertiesFrom(table.Properties);
            StylePropertyAsserter.VerifyCopyProperties(table.Properties, newTable.Properties);
        }

        [TestMethod]
        public void TestCopyPropertiesForTableRowProperties()
        {
            RadFlowDocument document = new RadFlowDocument();
            TableRow tableRow = document.Sections.AddSection().Blocks.AddTable().Rows.AddTableRow();

            tableRow.CanSplit = true;
            tableRow.Height = new TableRowHeight(HeightType.Exact, 40);
            tableRow.RepeatOnEveryPage = true;
            tableRow.TableCellSpacing = 30;

            TableRow newTableRow = document.Sections.AddSection().Blocks.AddTable().Rows.AddTableRow();
            newTableRow.Properties.CopyPropertiesFrom(tableRow.Properties);

            StylePropertyAsserter.VerifyCopyProperties(tableRow.Properties, newTableRow.Properties);
        }

        [TestMethod]
        public void TestCopyPropertiesForTableCellProperties()
        {
            RadFlowDocument document = new RadFlowDocument();
            TableCell tableCell = document.Sections.AddSection().Blocks.AddTable().Rows.AddTableRow().Cells.AddTableCell();

            tableCell.Borders = new TableCellBorders(new Border(1, BorderStyle.Single, new ThemableColor(Colors.Red)));
            tableCell.CanWrapContent = true;
            tableCell.ColumnSpan = 2;
            tableCell.IgnoreCellMarkerInRowHeightCalculation = true;
            tableCell.Padding = new Primitives.Padding(4);
            tableCell.PreferredWidth = new TableWidthUnit(TableWidthUnitType.Fixed, 50);
            tableCell.RowSpan = 2;
            tableCell.Shading.BackgroundColor = new ThemableColor(Colors.Pink);
            tableCell.Shading.Pattern = ShadingPattern.Percent10;
            tableCell.Shading.PatternColor = new ThemableColor(Colors.Cyan);
            tableCell.VerticalAlignment = Flow.Model.Styles.VerticalAlignment.Justified;

            TableCell newTableCell = document.Sections.AddSection().Blocks.AddTable().Rows.AddTableRow().Cells.AddTableCell();
            newTableCell.Properties.CopyPropertiesFrom(tableCell.Properties);
            StylePropertyAsserter.VerifyCopyProperties(tableCell.Properties, newTableCell.Properties);
        }

        [TestMethod]
        public void TestPropertyDefaultValueIsActualValueWhenBasedOnStyleIsNotInStyleRepository()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Style style = new Style("style", StyleType.Table);
            document.StyleRepository.Add(style);
            style.BasedOnStyleId = "basedOnStyle";

            // Act
            TableBorders actualValue = style.TableProperties.Borders.GetActualValue();

            // Assert
            TableBorders expectedValue = Table.TableBordersPropertyDefinition.DefaultValue;
            Assert.AreEqual(expectedValue, actualValue);
        }

        [TestMethod]
        public void TestPropertyDefaultValueIsActualValueWhenStyleIsNotInStyleRepository()
        {
            // Arrange
            Style style = new Style("style", StyleType.Table);

            // Act
            TableBorders actualValue = style.TableProperties.Borders.GetActualValue();

            // Assert
            TableBorders expectedValue = Table.TableBordersPropertyDefinition.DefaultValue;
            Assert.AreEqual(expectedValue, actualValue);
        }

        [TestMethod]
        public void TestFontSizeEvaluationOfTextWhen_RunIsInATable_TheAssociatedTableStyleHasCustomValue_NormalStyleHasDefaultValue()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            Table table = editor.InsertTable(2, 2);

            Paragraph paragraph = table.Rows.First().Cells.First().Blocks.AddParagraph();
            Run run = paragraph.Inlines.AddRun("hello");

            Style tableGridStyle = document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.TableGridStyleId);
            tableGridStyle.CharacterProperties.FontSize.LocalValue = 48;

            table.StyleId = tableGridStyle.Id;

            Assert.AreEqual(tableGridStyle.CharacterProperties.FontSize.LocalValue, run.FontSize);
        }

        [TestMethod]
        public void TestFontSizeEvaluationOfTextWhen_RunIsInATable_TheAssociatedTableStyleHasCustomValue_NormalStyleHasCustomValue()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            Table table = editor.InsertTable(2, 2);

            Paragraph paragraph = table.Rows.First().Cells.First().Blocks.AddParagraph();
            Run run = paragraph.Inlines.AddRun("hello");

            Style tableGridStyle = document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.TableGridStyleId);
            tableGridStyle.CharacterProperties.FontSize.LocalValue = 48;

            Style normalStyle = document.StyleRepository.GetDefaultStyle(StyleType.Paragraph);
            normalStyle.CharacterProperties.FontSize.LocalValue = 72;

            table.StyleId = tableGridStyle.Id;

            Assert.AreEqual(normalStyle.CharacterProperties.FontSize.LocalValue, run.FontSize);
        }

        [Description("uncomment when the style system is fixed. TP id: 200953, TFS id: 253887")]
        //[TestMethod]
        public void TestFontSizeEvaluationOfTextWhen_RunIsInATable_TheAssociatedTableStyleHasCustomValue_CustomStyleHasDefaultValue()
        {
            RadFlowDocument document = new RadFlowDocument();
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);

            Table table = editor.InsertTable(2, 2);

            Style customStyle = new Style("customStyle", StyleType.Paragraph);
            document.StyleRepository.Add(customStyle);

            Paragraph paragraph = table.Rows.First().Cells.First().Blocks.AddParagraph();
            paragraph.StyleId = customStyle.Id;

            Run run = paragraph.Inlines.AddRun("hello");

            Style tableGridStyle = document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.TableGridStyleId);
            tableGridStyle.CharacterProperties.FontSize.LocalValue = 48;

            table.StyleId = tableGridStyle.Id;

            Assert.AreEqual(tableGridStyle.CharacterProperties.FontSize.LocalValue, run.FontSize);
        }
    }
}
