﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Linq;
using Telerik.Windows.Documents.Flow.Model;

namespace Telerik.Windows.Documents.Flow.Tests.Model
{
    [TestClass]
    public class TableGridTest
    {
        [TestMethod]
        public void TestGridRowColumn_Count()
        {
            Table table = this.GetTable(3, 4);

            Assert.AreEqual(3, table.GridRowsCount);
            Assert.AreEqual(4, table.GridColumnsCount);
        }

        [TestMethod]
        public void TestGridRowColumn_AfterRowIsAdded_Count()
        {
            Table table = this.GetTable(3, 4);

            Assert.AreEqual(3, table.GridRowsCount);
            Assert.AreEqual(4, table.GridColumnsCount);

            table.Rows.AddTableRow().Cells.AddTableCell();

            Assert.AreEqual(4, table.GridRowsCount);
            Assert.AreEqual(4, table.GridColumnsCount);
        }

        [TestMethod]
        public void TestGridRowColumn_AfterRowIsRemoved_Count()
        {
            Table table = this.GetTable(3, 4);

            Assert.AreEqual(3, table.GridRowsCount);
            Assert.AreEqual(4, table.GridColumnsCount);

            table.Rows.Remove(table.Rows.Last());

            Assert.AreEqual(2, table.GridRowsCount);
            Assert.AreEqual(4, table.GridColumnsCount);
        }

        [TestMethod]
        public void TestGridRowColumn_AfterColumnIsAdded_Count()
        {
            Table table = this.GetTable(3, 4);

            Assert.AreEqual(3, table.GridRowsCount);
            Assert.AreEqual(4, table.GridColumnsCount);

            table.Rows.Last().Cells.AddTableCell();

            Assert.AreEqual(3, table.GridRowsCount);
            Assert.AreEqual(5, table.GridColumnsCount);
        }

        [TestMethod]
        public void TestGridRowColumn_AfterColumnIsRemoved_Count()
        {
            Table table = this.GetTable(3, 4);

            Assert.AreEqual(3, table.GridRowsCount);
            Assert.AreEqual(4, table.GridColumnsCount);

            foreach (TableRow row in table.Rows)
            {
                row.Cells.Remove(row.Cells.Last());
            }

            Assert.AreEqual(3, table.GridRowsCount);
            Assert.AreEqual(3, table.GridColumnsCount);
        }

        [TestMethod]
        public void TestGridRowColumn_AfterCellColSpanIsChanged_Count()
        {
            Table table = this.GetTable(3, 4);
            Assert.AreEqual(3, table.GridRowsCount);
            Assert.AreEqual(4, table.GridColumnsCount);

            table.Rows.First().Cells.First().ColumnSpan = 3;

            Assert.AreEqual(3, table.GridRowsCount);
            Assert.AreEqual(6, table.GridColumnsCount);
        }

        [TestMethod]
        public void TestGridRowColumn_AfterRowSpanIsChanged_Count()
        {
            Table table = this.GetTable(3, 4);
            Assert.AreEqual(3, table.GridRowsCount);
            Assert.AreEqual(4, table.GridColumnsCount);

            table.Rows.First().Cells.First().RowSpan = 3;

            Assert.AreEqual(3, table.GridRowsCount);
            Assert.AreEqual(5, table.GridColumnsCount);
        }

        [TestMethod]
        public void TestGridRowColumn_Count_WithSpannedCells()
        {
            // The table:
            //  _ _ _ _ 
            // |_|_|_|_|
            // |_|   |_|
            //   |_ _|

            Table table = this.GetTestTableWithSpannedCell();

            Assert.AreEqual(3, table.GridRowsCount);
            Assert.AreEqual(4, table.GridColumnsCount);

        }

        [TestMethod]
        public void TestCellAndRow_GridIndexes()
        {
            Table table = this.GetTable(3, 4);

            int rowIndex = 0;
            int colIndex = 0;
            foreach (TableRow row in table.Rows)
            {
                colIndex = 0;
                foreach (TableCell cell in row.Cells)
                {
                    Assert.AreEqual(colIndex, cell.GridColumnIndex);
                    Assert.AreEqual(rowIndex, cell.GridRowIndex);
                    colIndex++;
                }
                Assert.AreEqual(rowIndex, row.GridRowIndex);
                rowIndex++;
            }

        }

        [TestMethod]
        public void TestCell_GridIndexes_With_SpannedCells()
        {
            // The table:
            //  _ _ _ _ 
            // |_|_|_|_|
            // |_|   |_|
            // | |_ _|1|
            // |_|2|

            Table table = this.GetTable(0, 0);
            TableRow row = table.Rows.AddTableRow();
            row.Cells.AddTableCell();
            row.Cells.AddTableCell();
            row.Cells.AddTableCell();
            row.Cells.AddTableCell();

            row = table.Rows.AddTableRow();
            row.Cells.AddTableCell();
            TableCell spannedCell = row.Cells.AddTableCell();
            spannedCell.RowSpan = 2;
            spannedCell.ColumnSpan = 2;
            row.Cells.AddTableCell();

            row = table.Rows.AddTableRow();
            spannedCell = row.Cells.AddTableCell();
            spannedCell.RowSpan = 2;
            TableCell testCell1 = row.Cells.AddTableCell();

            row = table.Rows.AddTableRow();
            TableCell testCell2 = row.Cells.AddTableCell();

            Assert.AreEqual(3, testCell1.GridColumnIndex);
            Assert.AreEqual(2, testCell1.GridRowIndex);

            Assert.AreEqual(1, testCell2.GridColumnIndex);
            Assert.AreEqual(3, testCell2.GridRowIndex);
        }

        private Table GetTable(int rows, int cols)
        {
            Table table = new RadFlowDocument()
                .Sections.AddSection()
                .Blocks.AddTable();
            for (int i = 0; i < rows; i++)
            {
                TableRow row = table.Rows.AddTableRow();
                for (int j = 0; j < cols; j++)
                {
                    row.Cells.AddTableCell();
                }
            }

            return table;
        }

        private Table GetTestTableWithSpannedCell()
        {
            // The table:
            //  _ _ _ _ 
            // |_|_|_|_|
            // |_|   |_|
            //   |_ _|

            Table table = this.GetTable(0, 0);
            TableRow row = table.Rows.AddTableRow();
            row.Cells.AddTableCell();
            row.Cells.AddTableCell();
            row.Cells.AddTableCell();
            row.Cells.AddTableCell();

            row = table.Rows.AddTableRow();
            row.Cells.AddTableCell();
            TableCell spannedCell = row.Cells.AddTableCell();
            spannedCell.RowSpan = 2;
            spannedCell.ColumnSpan = 2;
            row.Cells.AddTableCell();
            return table;
        }


    }
}
