﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Media;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Watermarks;
using Telerik.Windows.Documents.Flow.Tests.Utilities;

namespace Telerik.Windows.Documents.Flow.Tests.Model
{
    [TestClass]
    public class WatermarkTests
    {
        [TestMethod]
        public void CreateAndSetImageWatermarkTest()
        {
            ImageWatermarkSettings originalImageSettings = new ImageWatermarkSettings()
            {
                Angle = 12,
                Width = 50,
                Height = 75,
                ImageSource = new Telerik.Windows.Documents.Media.ImageSource(TestHelper.GetImageStream("test.png"), "png")
            };

            Watermark originalWatermark = new Watermark(originalImageSettings);

            Assert.AreEqual(WatermarkType.Image, originalWatermark.WatermarkType);
            Assert.AreEqual(originalWatermark, originalImageSettings.Watermark);

            Header header = this.GetDefaultHeader();
            header.Watermarks.Add(originalWatermark);

            Watermark retrievedWatermark = header.Watermarks.First();

            StructureAsserter.AssertWatermarkImageSettingsAreEqual(originalImageSettings, retrievedWatermark.ImageSettings);
        }

        [TestMethod]
        public void CreateAndSetTextWatermarkTest()
        {
            TextWatermarkSettings originalTextSettings = new TextWatermarkSettings()
            {
                Angle = 12,
                Width = 50,
                Height = 75,
                Opacity = 0.4,
                FontFamily = new FontFamily("Verdana"),
                ForegroundColor = Colors.Yellow
            };

            Watermark originalWatermark = new Watermark(originalTextSettings);

            Assert.AreEqual(WatermarkType.Text, originalWatermark.WatermarkType);
            Assert.AreEqual(originalWatermark, originalTextSettings.Watermark);

            Header header = this.GetDefaultHeader();
            header.Watermarks.Add(originalWatermark);

            Watermark retrievedWatermark = header.Watermarks.First();

            StructureAsserter.AssertWatermarkTextSettingsAreEqual(originalTextSettings, retrievedWatermark.TextSettings);
        }

        [TestMethod]
        public void CreateTextWatermarkSettingsAndVerifyDefaultOpacityTest()
        {
            TextWatermarkSettings originalTextSettings = new TextWatermarkSettings()
            {
                Angle = 12,
                Width = 50,
                Height = 75,
                FontFamily = new FontFamily("Verdana"),
                ForegroundColor = Colors.Yellow
            };

            Assert.AreEqual(1, originalTextSettings.Opacity);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void CreateTextWatermarSettingskWithInvalidOpacityTest()
        {
            TextWatermarkSettings originalTextSettings = new TextWatermarkSettings()
            {
                Angle = 12,
                Width = 50,
                Height = 75,
                Opacity = 2,
                ForegroundColor = Colors.Yellow
            };
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void CreateTextWatermarkWithNullFontFamilyTest()
        {
            TextWatermarkSettings originalTextSettings = new TextWatermarkSettings()
            {
                Angle = 12,
                Width = 50,
                Height = 75,
                Opacity = 0.9,
                FontFamily = null,
                ForegroundColor = Colors.Yellow
            };
        }

        [TestMethod]
        public void CloneWatermarkTest()
        {
            TextWatermarkSettings originalTextSettings = new TextWatermarkSettings()
            {
                Angle = 12,
                Width = 50,
                Height = 75,
                Opacity = 0.4,
                FontFamily = new FontFamily("Verdana"),
                ForegroundColor = Colors.Yellow
            };

            Watermark originalWatermark = new Watermark(originalTextSettings);

            Header header = this.GetDefaultHeader();
            header.Watermarks.Add(originalWatermark);

            Watermark cloneWatermark = originalWatermark.Clone();

            StructureAsserter.AssertWatermarksAreEqual(originalWatermark, cloneWatermark);
            Assert.IsNull(cloneWatermark.Document);
        }

        [TestMethod]
        public void CloneWatermarkTextSettingsTest()
        {
            TextWatermarkSettings originalTextSettings = new TextWatermarkSettings()
            {
                Angle = 12,
                Width = 50,
                Height = 75,
                Opacity = 0.4,
                FontFamily = new FontFamily("Verdana"),
                ForegroundColor = Colors.Yellow
            };

            Watermark originalWatermark = new Watermark(originalTextSettings);

            Header header = this.GetDefaultHeader();
            header.Watermarks.Add(originalWatermark);

            TextWatermarkSettings cloneSettings = originalTextSettings.Clone();

            StructureAsserter.AssertWatermarkTextSettingsAreEqual(originalTextSettings, cloneSettings);
            Assert.IsNull(cloneSettings.Watermark);
        }

        [TestMethod]
        public void CloneWatermarkImageSettingsTest()
        {
            ImageWatermarkSettings originalImageSettings = new ImageWatermarkSettings()
            {
                Angle = 12,
                Width = 50,
                Height = 75,
                ImageSource = new Telerik.Windows.Documents.Media.ImageSource(TestHelper.GetImageStream("test.png"), "png")
            };

            Watermark originalWatermark = new Watermark(originalImageSettings);

            Header header = this.GetDefaultHeader();
            header.Watermarks.Add(originalWatermark);

            ImageWatermarkSettings cloneSettings = originalImageSettings.Clone();

            StructureAsserter.AssertWatermarkImageSettingsAreEqual(originalImageSettings, cloneSettings);
            Assert.IsNull(cloneSettings.Watermark);
        }

        private Header GetDefaultHeader(RadFlowDocument document = null)
        {
            if (document == null)
            {
                document = new RadFlowDocument();
            }

            Section section = document.Sections.AddSection();
            return section.Headers.Add(HeaderFooterType.Default);
        }
    }
}
