﻿using System;
using System.Linq;
using System.Text.RegularExpressions;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Tests.Utilities;
using Telerik.Windows.Documents.Flow.Model.Styles;
#if WPF || SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Flow.Tests.TextSearch
{
    [TestClass]
    public class ReplacePropertiesTests : ReplaceTestsBase
    {
        // Single run tests
        [TestMethod]
        public void ReplaceWithString_WhenAllContentIsSearched_ReplacesText_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            string text = "test";
            editor.InsertText(text);

            // Act
            editor.ReplaceStyling(text, true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, text);
            ReplaceTestsBase.AssertDocumentText(text, editor.Document);
        }

        [TestMethod]
        public void ReplaceWithString_WhenSingleCharWillBeReplacedWithOtherChar_ReplacesOnlyThatChar_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("alabala aa");

            // Act
            editor.ReplaceStyling("a", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertDocumentText("alabala aa", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "a");
        }

        [TestMethod]
        public void ReplaceWithString_WhenPartOfRunIsSearchedAndSearchedWordIsShorter_ReplacesText_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("test");

            // Act
            editor.ReplaceStyling("e", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertDocumentText("test", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "e");
        }

        [TestMethod]
        public void ReplaceWithString_WhenPartOfRunIsSearchedAndSearchedWordIsLonger_ReplacesText_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("telerikeleri");

            // Act
            editor.ReplaceStyling("eleri", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertDocumentText("telerikeleri", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "eleri");
        }

        [TestMethod]
        public void ReplaceWithString_WhenPartOfRunIsSearchedAndSearchedWordExistsMultipleTimesInTheMiddle_ReplacesText_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("telerik");

            // Act
            editor.ReplaceStyling("e", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertDocumentText("telerik", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "e");
        }

        [TestMethod]
        public void ReplaceWithString_WhenPartOfRunIsSearchedAndSearchedWordExistsMultipleTimesInTheBeginingOfTheRunAndInTheEnd_ReplacesText_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("todoryt");

            // Act
            editor.ReplaceStyling("t", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertDocumentText("todoryt", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "t");
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedCharIsMultipleTimesInStartAndEndOfRun_ReplacesText_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("ooohooooo");

            // Act
            editor.ReplaceStyling("o", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            Assert.AreEqual(3, editor.Document.EnumerateChildrenOfType<Run>().Count());
            ReplaceTestsBase.AssertDocumentText("ooohooooo", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "o");
        }


        // Multiple run tests
        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenTwoRuns_ReplacesText_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("aba");
            editor.InsertText("ba");

            // Act
            editor.ReplaceStyling("bab", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertDocumentText("ababa", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "bab");
        }

        [TestMethod]
        public void ReplaceWithString_WhenSingleCharWillBeReplacedWithOtherCharAndDocumentHasMultipleRuns_ReplacesOnlyThatChar_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("alabala");
            editor.InsertText("aa");
            editor.InsertText("makaron ala testa");

            // Act
            editor.ReplaceStyling("a", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertDocumentText("alabalaaamakaron ala testa", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "a");
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenTwoRunsAndTheReplaceContentIsPartOfFirstRunAndTheWholeSecond_ReplacesText_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("aba");
            editor.InsertText("ba");

            // Act
            editor.ReplaceStyling("baba", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertDocumentText("ababa", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "baba");
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenTwoRunsAndTheReplaceContentTheWholeSecondButFirstRunContainsItToo_ReplacesText_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("aba");
            editor.InsertText("ba");

            // Act
            editor.ReplaceStyling("ba", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            Assert.AreEqual(2, editor.Document.EnumerateChildrenOfType<Run>().Count());
            ReplaceTestsBase.AssertDocumentText("ababa", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "ba");
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenTwoRunsAndIncludesItAll_ReplacesText_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("aba");
            editor.InsertText("ba");

            // Act
            editor.ReplaceStyling("ababa", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Run>().Count());
            ReplaceTestsBase.AssertDocumentText("ababa", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "ababa");
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenTwoRunsAndMatchesTheBeginingAndEndCharacters_ReplacesText_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("abA");
            editor.InsertText("ba");

            // Act
            editor.ReplaceStyling("a", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            Assert.AreEqual(4, editor.Document.EnumerateChildrenOfType<Run>().Count());
            ReplaceTestsBase.AssertDocumentText("abAba", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "a");
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenThreeRuns_ReplacesText_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("admin operator tee");
            editor.InsertText("ee");
            editor.InsertText("st user.");

            // Act
            editor.ReplaceStyling("teeeest", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            Assert.AreEqual(3, editor.Document.EnumerateChildrenOfType<Run>().Count());
            ReplaceTestsBase.AssertDocumentText("admin operator teeeest user.", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "teeeest");
        }

        [TestMethod]
        public void ReplaceWithString_WhenReplaceTextContainsSearchedText_ReplacesSearchedWord_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            string text = "tests";
            editor.InsertText(text);
            editor.InsertText(text);

            // Act
            editor.ReplaceStyling(text, true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Run>().Count());
            ReplaceTestsBase.AssertDocumentText("teststests", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "tests");
        }

        [TestMethod]
        public void ReplaceWithString_WhenOnlyWholeWordsShouldBeReplace_ReplacesOnlyWholeWord_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("text Text searchedtext");
            editor.InsertText("thattext shouldnt be replaced, but this text should be replaced");
            editor.InsertText(" text");

            RadFlowDocument originalDocument = editor.Document.Clone();

            // Act
            editor.ReplaceStyling("text", false, true, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertDocumentText(originalDocument.Document, editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, run => run.Text.ToLower() == "text");
        }

        [TestMethod]
        public void ReplaceWithRegex_WhenSearchedContentShouldMatchAWordContainingUpperLetter_ReplacesWordsContainingUpperLetter_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            string text = @"Welcome to RegExr v1.0 by telerik!
Edit the Expression & Text to see matches. Roll over matches or the expression for details. Undo mistakes with ctrl-z. Save & Share expressions with friends or the Community. A full Reference & Help is available in the Library, or watch the video Tutorial.
Sample text for testing:
abcdefghijklmnopqrstuvwxyz ABCDEFGHIJKLMNOPQRSTUVWXYZ
0123456789 +-.,!@#$%^&*();\/|<>";

            editor.InsertText(text);
            RadFlowDocument originalDocument = editor.Document.Clone();

            // Act
            Regex regex = new Regex(@"([A-Z])\w+");
            editor.ReplaceStyling(regex, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            string notMatchedRunText = "A full";
            ReplaceTestsBase.AssertDocumentText(originalDocument.Document, editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, run => run.Text.Any(c => char.IsUpper(c)) && !run.Text.Contains(notMatchedRunText));
        }

        [TestMethod]
        public void ReplaceWithRegex_WhenSearchedContentShouldMatchWordContainingUpperLetter_AndDocumentContainsMultipleRunsOnDifferentParagraphs_ReplacesWordsContainingUpperLetter_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("Video provides a powerful way to help you prove your point. When you click Online Video, you can paste in the embed code for the video you want to add. You can also type a keyword to search online for the video that best fits your document. To make your document look professionally produced, Word provides header, footer, cover page, and text box designs that complement each other.");
            editor.InsertParagraph();
            editor.InsertText("For example, you can add a matching cover page, header, and sidebar. Click Insert and then choose the elements you want from the different galleries. Themes and styles also help keep your document coordinated. When you click Design and choose a new Theme, the pictures, charts, and SmartArt graphics change to match your new theme.");
            editor.InsertParagraph();
            editor.InsertText("When you apply styles, your headings change to match the new theme. Save time in Word with new buttons that show up where you need them. To change the way a picture fits in your document, click it and a button for layout options appears next to it. When you work on a table, click where you want to add a row or a column, and then click the plus sign.");
            editor.InsertParagraph();
            editor.InsertText("hello Wor");
            editor.InsertText("ld");
            editor.InsertText(" Telerik test ");
            editor.InsertText("testTest");

            RadFlowDocument originalDocument = editor.Document.Clone();

            // Act
            editor.ReplaceStyling(new Regex(@"([A-Z])\w+"), ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertDocumentText(originalDocument.Document, editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, this.runWithCapitalLettersFunc);
        }

        [TestMethod]
        public void ReplaceWithRegex_WhenEverySymbolsShouldGetNewPropertiesButHasDifferentCharacterProperties_AreNotMergedAfterReplace()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            var style = new Style("telerik", StyleType.Character);
            style.CharacterProperties.ForegroundColor.LocalValue = ThemableColor.FromArgb(255, 255, 0, 0);
            editor.Document.StyleRepository.Add(style);

            Run run1 = editor.InsertText("st");
            run1.StyleId = style.Id;

            Run run2 = editor.InsertText("op");
            run2.ForegroundColor = ThemableColor.FromArgb(255, 255, 0, 0);

            RadFlowDocument originalDocument = editor.Document.Clone();

            // Act
            editor.ReplaceStyling(new Regex("."), ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertDocumentText(originalDocument.Document, editor.Document);

            Run replacedRun1 = editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault();
            Run replacedRun2 = editor.Document.EnumerateChildrenOfType<Run>().LastOrDefault();

            Assert.AreEqual("st", replacedRun1.Text);
            Assert.AreEqual("op", replacedRun2.Text);
        }

        // Inlines between runs
        [TestMethod]
        public void ReplaceWithString_WhenImageIsBetweenRunsWhichMatchTheSearchedText_DoesntReplacesRunContent_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("te");
            editor.InsertInline(TestHelper.CreatePngImage(editor.Document));
            editor.InsertText("st");

            // Act
            editor.ReplaceStyling("test", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            Assert.AreEqual(0, editor.Document.EnumerateChildrenOfType<Run>().Where(r => r.Text == "test").Count());

            Paragraph paragraph = editor.Document.EnumerateChildrenOfType<Paragraph>().FirstOrDefault();
            Assert.AreEqual("te", ((Run)paragraph.Inlines.FirstOrDefault()).Text);
            Assert.AreEqual("st", ((Run)paragraph.Inlines.LastOrDefault()).Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenImageIsBetweenRunsWhichMatchTheSearchedText_DoesntReplacesRunContentButReplacesTextInPreviousRun_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("test");
            editor.InsertText("te");
            editor.InsertInline(TestHelper.CreatePngImage(editor.Document));
            editor.InsertText("st");

            // Act
            editor.ReplaceStyling("test", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Run>().Where(r => r.Text == "test").Count());

            Paragraph paragraph = editor.Document.EnumerateChildrenOfType<Paragraph>().FirstOrDefault();
            Assert.AreEqual("test", ((Run)paragraph.Inlines.FirstOrDefault()).Text);
            Assert.AreEqual("te", ((Run)paragraph.Inlines.ElementAt(1)).Text);
            Assert.AreEqual("st", ((Run)paragraph.Inlines.LastOrDefault()).Text);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "test");
        }


        // ReplaceStyling tests - Ignoring the text replacement
        [TestMethod]
        public void ReplaceStyling_WhenDocumentContainsOnlySearchedRun_PreservesDocumentTextAndChangeProperties_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("test");

            // Act
            editor.ReplaceStyling(@"test", ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            string expectedResult = "test";
            AssertDocumentText(expectedResult, editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "test");
        }

        [TestMethod]
        public void ReplaceStyling_WhenSearchedContentShouldMatchWordContainingUpperLetter_AndDocumentContainsSingleRunButItHasMultipleMatches_PreservesDocumentTextAndChangeProperties_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            string runText = "Video provides a powerful way to help you prove your point. When you click Online Video, you can paste in the embed code for the video you want to add. You can also type a keyword to search online for the video that best fits your document. To make your document look professionally produced, Word provides header, footer, cover page, and text box designs that complement each other.";
            editor.InsertText(runText);

            // Act
            editor.ReplaceStyling(new Regex(@"([A-Z])\w+"), ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertDocumentText(runText, editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, this.runWithCapitalLettersFunc);
        }

        [TestMethod]
        public void ReplaceStyling_WhenSearchedContentShouldMatchWordContainingUpperLetter_AndDocumentContainsMultipleRunsOnDifferentParagraphs_PreservesDocumentTextAndChangeProperties_Properties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("Video provides a powerful way to help you prove your point. When you click Online Video, you can paste in the embed code for the video you want to add. You can also type a keyword to search online for the video that best fits your document. To make your document look professionally produced, Word provides header, footer, cover page, and text box designs that complement each other.");
            editor.InsertParagraph();
            editor.InsertText("For example, you can add a matching cover page, header, and sidebar. Click Insert and then choose the elements you want from the different galleries. Themes and styles also help keep your document coordinated. When you click Design and choose a new Theme, the pictures, charts, and SmartArt graphics change to match your new theme.");
            editor.InsertParagraph();
            editor.InsertText("When you apply styles, your headings change to match the new theme. Save time in Word with new buttons that show up where you need them. To change the way a picture fits in your document, click it and a button for layout options appears next to it. When you work on a table, click where you want to add a row or a column, and then click the plus sign.");
            editor.InsertParagraph();
            editor.InsertText("hello Wor");
            editor.InsertText("ld");
            editor.InsertText(" Telerik test ");
            editor.InsertText("testTest");

            // Act
            editor.ReplaceStyling(new Regex(@"([A-Z])\w+"), ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            string expectedResult = "Video provides a powerful way to help you prove your point. When you click Online Video, you can paste in the embed code for the video you want to add. You can also type a keyword to search online for the video that best fits your document. To make your document look professionally produced, Word provides header, footer, cover page, and text box designs that complement each other." +
"For example, you can add a matching cover page, header, and sidebar. Click Insert and then choose the elements you want from the different galleries. Themes and styles also help keep your document coordinated. When you click Design and choose a new Theme, the pictures, charts, and SmartArt graphics change to match your new theme." +
"When you apply styles, your headings change to match the new theme. Save time in Word with new buttons that show up where you need them. To change the way a picture fits in your document, click it and a button for layout options appears next to it. When you work on a table, click where you want to add a row or a column, and then click the plus sign." +
"hello World Telerik test testTest";

            ReplaceTestsBase.AssertDocumentText(expectedResult, editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, this.runWithCapitalLettersFunc);
        }
    }
}