﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Linq;
using System.Text.RegularExpressions;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Tests.Utilities;

namespace Telerik.Windows.Documents.Flow.Tests.TextSearch
{
    [TestClass]
    public class ReplaceTextAndPropertiesTests : ReplaceTestsBase
    {
        // Single run tests
        [TestMethod]
        public void ReplaceWithString_WhenAllContentIsSearched_ReplacesText_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            string oldText = "test";
            editor.InsertText(oldText);

            // Act
            string replaceText = "telerik";
            editor.ReplaceText(oldText, replaceText, true, false);
            editor.ReplaceStyling(replaceText, true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, replaceText);
            ReplaceTestsBase.AssertDocumentText(replaceText, editor.Document);
        }

        [TestMethod]
        public void ReplaceWithString_WhenSingleCharWillBeReplacedWithOtherChar_ReplacesOnlyThatChar_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("alabala aa");

            // Act
            editor.ReplaceText("a", "o", true, false);
            editor.ReplaceStyling("o", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertDocumentText("olobolo oo", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "o");
        }

        [TestMethod]
        public void ReplaceWithString_WhenPartOfRunIsSearchedAndSearchedWordIsShorter_ReplacesText_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("test");

            // Act
            editor.ReplaceText("e", "telerik", true, false);
            editor.ReplaceStyling("telerik", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertDocumentText("ttelerikst", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "telerik");
        }

        [TestMethod]
        public void ReplaceWithString_WhenPartOfRunIsSearchedAndSearchedWordIsLonger_ReplacesText_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("telerikeleri");

            // Act
            editor.ReplaceStyling("eleri", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());
            editor.ReplaceText("eleri", "a", true, false);

            // Assert
            ReplaceTestsBase.AssertDocumentText("taka", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "a");
        }

        [TestMethod]
        public void ReplaceWithString_WhenPartOfRunIsSearchedAndSearchedWordExistsMultipleTimesInTheMiddle_ReplacesText_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("telerik");

            // Act
            editor.ReplaceStyling("e", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());
            editor.ReplaceText("e", "O", true, false);

            // Assert
            ReplaceTestsBase.AssertDocumentText("tOlOrik", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "O");
        }

        [TestMethod]
        public void ReplaceWithString_WhenPartOfRunIsSearchedAndSearchedWordExistsMultipleTimesInTheBeginingOfTheRunAndInTheEnd_ReplacesText_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("todoryt");

            // Act
            editor.ReplaceText("t", "f", true, false);
            editor.ReplaceStyling("f", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertDocumentText("fodoryf", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "f");
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedCharIsMultipleTimesInStartAndEndOfRun_ReplacesText_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("ooohooooo");

            // Act
            editor.ReplaceStyling("o", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());
            editor.ReplaceText("o", "e", true, false);

            // Assert
            Assert.AreEqual(3, editor.Document.EnumerateChildrenOfType<Run>().Count());
            ReplaceTestsBase.AssertDocumentText("eeeheeeee", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "e");
        }


        // Multiple run tests
        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenTwoRuns_ReplacesText_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("aba");
            editor.InsertText("ba");

            // Act
            editor.ReplaceStyling("bab", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());
            editor.ReplaceText("bab", "c", true, false);

            // Assert
            ReplaceTestsBase.AssertDocumentText("aca", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "c");
        }

        [TestMethod]
        public void ReplaceWithString_WhenSingleCharWillBeReplacedWithOtherCharAndDocumentHasMultipleRuns_ReplacesOnlyThatChar_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("alabala");
            editor.InsertText("aa");
            editor.InsertText("makaron ala testa");

            // Act
            editor.ReplaceStyling("a", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());
            editor.ReplaceText("a", "o", true, false);

            // Assert
            ReplaceTestsBase.AssertDocumentText("olobolooomokoron olo testo", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "o");
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenTwoRunsAndTheReplaceContentIsPartOfFirstRunAndTheWholeSecond_ReplacesText_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("aba");
            editor.InsertText("ba");

            // Act
            editor.ReplaceText("baba", "c", true, false);
            editor.ReplaceStyling("c", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            ReplaceTestsBase.AssertDocumentText("ac", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "c");
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenTwoRunsAndTheReplaceContentTheWholeSecondButFirstRunContainsItToo_ReplacesText_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("aba");
            editor.InsertText("ba");

            // Act
            editor.ReplaceStyling("ba", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());
            editor.ReplaceText("ba", "c", true, false);

            // Assert
            Assert.AreEqual(2, editor.Document.EnumerateChildrenOfType<Run>().Count());
            ReplaceTestsBase.AssertDocumentText("acc", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "c");
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenTwoRunsAndTheReplaceContentTheWholeSecondButFirstRunContainsItTooAndTheReplaceTextIsLongerThanSearched_ReplacesText_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("aba");
            editor.InsertText("ba");

            // Act
            editor.ReplaceText("ba", "telerik", true, false);
            editor.ReplaceStyling("telerik", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            Assert.AreEqual(2, editor.Document.EnumerateChildrenOfType<Run>().Count());
            ReplaceTestsBase.AssertDocumentText("ateleriktelerik", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "telerik");
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenTwoRunsAndIncludesItAll_ReplacesText_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("aba");
            editor.InsertText("ba");

            // Act
            editor.ReplaceText("ababa", "c", true, false);
            editor.ReplaceStyling("c", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Run>().Count());
            ReplaceTestsBase.AssertDocumentText("c", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "c");
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenTwoRunsAndMatchesTheBeginingAndEndCharacters_ReplacesText_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("abA");
            editor.InsertText("ba");

            // Act
            editor.ReplaceStyling("a", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());
            editor.ReplaceText("a", "c", true, false);

            // Assert
            Assert.AreEqual(4, editor.Document.EnumerateChildrenOfType<Run>().Count());
            ReplaceTestsBase.AssertDocumentText("cbAbc", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "c");
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenThreeRuns_ReplacesText_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("mega qkiq tee");
            editor.InsertText("ee");
            editor.InsertText("st prosto.");

            // Act
            editor.ReplaceStyling("teeeest", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());
            editor.ReplaceText("teeeest", "MEST", true, false);

            // Assert
            Assert.AreEqual(3, editor.Document.EnumerateChildrenOfType<Run>().Count());
            ReplaceTestsBase.AssertDocumentText("mega qkiq MEST prosto.", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "MEST");
        }

        [TestMethod]
        public void ReplaceWithString_WhenReplaceTextContainsSearchedText_ReplacesSearchedWord_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            string text = "tests";
            editor.InsertText(text);
            editor.InsertText(text);

            // Act
            string replaceText = "more tests";
            editor.ReplaceStyling(text, true, false, ReplaceTestsBase.GetReplacePropertiesHandler());
            editor.ReplaceText(text, replaceText, true, false);

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Run>().Count());
            ReplaceTestsBase.AssertDocumentText("more testsmore tests", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "more tests");
        }

        [TestMethod]
        public void ReplaceWithString_WhenOnlyWholeWordsShouldBeReplace_ReplacesOnlyWholeWord_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("text Text searchedtext");
            editor.InsertText("thattext shouldnt be replaced, but this text should be replaced");
            editor.InsertText("text");

            // Act
            editor.ReplaceStyling("text", false, true, ReplaceTestsBase.GetReplacePropertiesHandler());
            editor.ReplaceText("text", "NEW TEXT", false, true);

            // Assert
            string expectedDocumentText = @"NEW TEXT NEW TEXT searchedtextthattext shouldnt be replaced, but this NEW TEXT should be replacedtext";
            ReplaceTestsBase.AssertDocumentText(expectedDocumentText, editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "NEW TEXT");
        }


        // REGEX
        [TestMethod]
        public void ReplaceWithRegex_WhenSearchedContentIsBetweenThreeRuns_ReplacesText_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("mega qkiq tee");
            editor.InsertText("ee");
            editor.InsertText("st prosto.");

            // Act
            Regex regex = new Regex("teeeest");
            editor.ReplaceStyling(regex, ReplaceTestsBase.GetReplacePropertiesHandler());
            editor.ReplaceText(regex, "MEST");

            // Assert
            Assert.AreEqual(3, editor.Document.EnumerateChildrenOfType<Run>().Count());
            ReplaceTestsBase.AssertDocumentText("mega qkiq MEST prosto.", editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "MEST");
        }

        [TestMethod]
        public void ReplaceWithRegex_WhenSearchedContentShouldMatchAWordContainingUpperLetter_ReplacesWordsContainingUpperLetter_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            string text = @"Welcome to RegExr v1.0 by telerik!
Edit the Expression & Text to see matches. Roll over matches or the expression for details. Undo mistakes with ctrl-z. Save & Share expressions with friends or the Community. A full Reference & Help is available in the Library, or watch the video Tutorial.
Sample text for testing:
abcdefghijklmnopqrstuvwxyz ABCDEFGHIJKLMNOPQRSTUVWXYZ
0123456789 +-.,!@#$%^&*();\/|<>";

            editor.InsertText(text);

            // Act
            Regex regex = new Regex(@"([A-Z])\w+");
            string replaceText = "TELERIK";
            editor.ReplaceStyling(regex, ReplaceTestsBase.GetReplacePropertiesHandler());
            editor.ReplaceText(regex, replaceText);

            // Assert
            string expectedResult = @"TELERIK to TELERIK v1.0 by telerik!TELERIK the TELERIK & TELERIK to see matches. TELERIK over matches or the expression for details. TELERIK mistakes with ctrl-z. TELERIK & TELERIK expressions with friends or the TELERIK. A full TELERIK & TELERIK is available in the TELERIK, or watch the video TELERIK.TELERIK text for testing:abcdefghijklmnopqrstuvwxyz TELERIK0123456789 +-.,!@#$%^&*();\/|<>";
            ReplaceTestsBase.AssertDocumentText(expectedResult, editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, replaceText);
        }

        [TestMethod]
        public void ReplaceWithRegex_WhenSearchedContentShouldMatchWordContainingUpperLetter_AndDocumentContainsMultipleRunsOnDifferentParagraphs_ReplacesWordsContainingUpperLetter_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("Video provides a powerful way to help you prove your point. When you click Online Video, you can paste in the embed code for the video you want to add. You can also type a keyword to search online for the video that best fits your document. To make your document look professionally produced, Word provides header, footer, cover page, and text box designs that complement each other.");
            editor.InsertParagraph();
            editor.InsertText("For example, you can add a matching cover page, header, and sidebar. Click Insert and then choose the elements you want from the different galleries. Themes and styles also help keep your document coordinated. When you click Design and choose a new Theme, the pictures, charts, and SmartArt graphics change to match your new theme.");
            editor.InsertParagraph();
            editor.InsertText("When you apply styles, your headings change to match the new theme. Save time in Word with new buttons that show up where you need them. To change the way a picture fits in your document, click it and a button for layout options appears next to it. When you work on a table, click where you want to add a row or a column, and then click the plus sign.");
            editor.InsertParagraph();
            editor.InsertText("hello Wor");
            editor.InsertText("ld");
            editor.InsertText(" Telerik test ");
            editor.InsertText("testTest");

            // Act
            editor.ReplaceStyling(new Regex(@"([A-Z])\w+"), ReplaceTestsBase.GetReplacePropertiesHandler());
            editor.ReplaceText(new Regex(@"([A-Z])\w+"), "WORDSPROCESSING");

            // Assert
            string expectedResult = "WORDSPROCESSING provides a powerful way to help you prove your point. WORDSPROCESSING you click WORDSPROCESSING WORDSPROCESSING, you can paste in the embed" +
                " code for the video you want to add. WORDSPROCESSING can also type a keyword to search online for the video that best fits your document. WORDSPROCESSING make your document look" +
                " professionally produced, WORDSPROCESSING provides header, footer, cover page, and text box designs that complement each other.WORDSPROCESSING example, you can add a matching cover" +
                " page, header, and sidebar. WORDSPROCESSING WORDSPROCESSING and then choose the elements you want from the different galleries. WORDSPROCESSING and styles also help keep your document" +
                " coordinated. WORDSPROCESSING you click WORDSPROCESSING and choose a new WORDSPROCESSING, the pictures, charts, and WORDSPROCESSING graphics change to match your new theme.WORDSPROCESSING" +
                " you apply styles, your headings change to match the new theme. WORDSPROCESSING time in WORDSPROCESSING with new buttons that show up where you need them. WORDSPROCESSING change the way a picture" +
                " fits in your document, click it and a button for layout options appears next to it. WORDSPROCESSING you work on a table, click where you want to add a row or a column, and then click the plus sign." +
                "hello WORDSPROCESSING WORDSPROCESSING test testWORDSPROCESSING";
            AssertDocumentText(expectedResult, editor.Document);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "WORDSPROCESSING");
        }


        // Inlines between runs
        [TestMethod]
        public void ReplaceWithString_WhenImageIsBetweenRunsWhichMatchTheSearchedText_DoesntReplacesRunContent_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("te");
            editor.InsertInline(TestHelper.CreatePngImage(editor.Document));
            editor.InsertText("st");

            // Act
            editor.ReplaceStyling("test", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());
            editor.ReplaceText("test", "new text", true, false);

            // Assert
            Assert.AreEqual(0, editor.Document.EnumerateChildrenOfType<Run>().Where(r => r.Text == "new text").Count());

            Paragraph paragraph = editor.Document.EnumerateChildrenOfType<Paragraph>().FirstOrDefault();
            Assert.AreEqual("te", ((Run)paragraph.Inlines.FirstOrDefault()).Text);
            Assert.AreEqual("st", ((Run)paragraph.Inlines.LastOrDefault()).Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenImageIsBetweenRunsWhichMatchTheSearchedText_DoesntReplacesRunContentButReplacesTextInPreviousRun_TextAndProperties()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("test");
            editor.InsertText("te");
            editor.InsertInline(TestHelper.CreatePngImage(editor.Document));
            editor.InsertText("st");

            // Act
            editor.ReplaceStyling("test", true, false, ReplaceTestsBase.GetReplacePropertiesHandler());
            editor.ReplaceText("test", "new text", true, false);

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Run>().Where(r => r.Text == "new text").Count());

            Paragraph paragraph = editor.Document.EnumerateChildrenOfType<Paragraph>().FirstOrDefault();
            Assert.AreEqual("new text", ((Run)paragraph.Inlines.FirstOrDefault()).Text);
            Assert.AreEqual("te", ((Run)paragraph.Inlines.ElementAt(1)).Text);
            Assert.AreEqual("st", ((Run)paragraph.Inlines.LastOrDefault()).Text);
            ReplaceTestsBase.AssertCharacterPropertiesForRun(editor.Document, "new text");
        }
    }
}