﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Linq;
using System.Text.RegularExpressions;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Tests.Utilities;

namespace Telerik.Windows.Documents.Flow.Tests.TextSearch
{
    [TestClass]
    public class ReplaceTextTests : ReplaceTestsBase
    {
        // Single run tests
        [TestMethod]
        public void ReplaceWithString_WhenAllContentIsSearched_ReplacesText()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            string oldText = "test";
            editor.InsertText(oldText);

            // Act
            string replaceText = "telerik";
            editor.ReplaceText(oldText, replaceText);

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Run>().Count());
            Assert.AreEqual(replaceText, editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenSingleCharWillBeReplacedWithOtherChar_ReplacesOnlyThatChar()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("alabala aa");

            // Act
            editor.ReplaceText("a", "o");

            // Assert
            Assert.AreEqual("olobolo oo", editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenPartOfRunIsSearchedAndSearchedWordIsShorter_ReplacesText()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("test");

            // Act
            editor.ReplaceText("e", "telerik");

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Run>().Count());
            Assert.AreEqual("ttelerikst", editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenPartOfRunIsSearchedAndSearchedWordIsLonger_ReplacesText()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("telerikeleri");

            // Act
            editor.ReplaceText("eleri", "a");

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Run>().Count());
            Assert.AreEqual("taka", editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenPartOfRunIsSearchedAndSearchedWordExistsMultipleTimesInTheMiddle_ReplacesText()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("telerik");

            // Act
            editor.ReplaceText("e", "O");

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Run>().Count());
            Assert.AreEqual("tOlOrik", editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenPartOfRunIsSearchedAndSearchedWordExistsMultipleTimesInTheBeginingOfTheRunAndInTheEnd_ReplacesText()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("todoryt");

            // Act
            editor.ReplaceText("t", "f");

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Run>().Count());
            Assert.AreEqual("fodoryf", editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenReplaceTextIsEmptyString_DeletesTheSearchedTextFromRun()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("testeet");

            // Act
            editor.ReplaceText("e", "");

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Run>().Count());

            Run run = editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault();
            Assert.AreEqual("tstt", run.Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedCharIsMultipleTimesInStartAndEndOfRun_ReplacesText()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("ooohooooo");

            // Act
            editor.ReplaceText("o", "e");

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Run>().Count());
            Assert.AreEqual("eeeheeeee", editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
        }


        // Multiple run tests
        [TestMethod]
        public void ReplaceWithString_WhenSingleCharWillBeReplacedWithOtherCharAndDocumentHasMultipleRuns_ReplacesOnlyThatChar()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("alabala");
            editor.InsertText("aa");
            editor.InsertText("makaron ala testa");

            // Act
            editor.ReplaceText("a", "o");

            // Assert
            Assert.AreEqual("olobolo", editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
            Assert.AreEqual("oo", editor.Document.EnumerateChildrenOfType<Run>().ElementAt(1).Text);
            Assert.AreEqual("mokoron olo testo", editor.Document.EnumerateChildrenOfType<Run>().LastOrDefault().Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenTwoRuns_ReplacesText()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("aba");
            editor.InsertText("ba");

            // Act
            editor.ReplaceText("bab", "c");

            // Assert
            Assert.AreEqual(2, editor.Document.EnumerateChildrenOfType<Run>().Count());
            Assert.AreEqual("ac", editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
            Assert.AreEqual("a", editor.Document.EnumerateChildrenOfType<Run>().LastOrDefault().Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenTwoRunsAndTheReplaceContentIsPartOfFirstRunAndTheWholeSecond_ReplacesText()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("aba");
            editor.InsertText("ba");

            // Act
            editor.ReplaceText("baba", "c");

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Run>().Count());
            Assert.AreEqual("ac", editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenTwoRunsAndTheReplaceContentTheWholeSecondButFirstRunContainsItToo_ReplacesText()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("aba");
            editor.InsertText("ba");

            // Act
            editor.ReplaceText("ba", "c");

            // Assert
            Assert.AreEqual(2, editor.Document.EnumerateChildrenOfType<Run>().Count());
            Assert.AreEqual("ac", editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
            Assert.AreEqual("c", editor.Document.EnumerateChildrenOfType<Run>().LastOrDefault().Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenTwoRunsAndTheReplaceContentTheWholeSecondButFirstRunContainsItTooAndTheReplaceTextIsLongerThanSearched_ReplacesText()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("aba");
            editor.InsertText("ba");

            // Act
            editor.ReplaceText("ba", "telerik");

            // Assert
            Assert.AreEqual(2, editor.Document.EnumerateChildrenOfType<Run>().Count());
            Assert.AreEqual("atelerik", editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
            Assert.AreEqual("telerik", editor.Document.EnumerateChildrenOfType<Run>().LastOrDefault().Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenTwoRunsAndIncludesItAll_ReplacesText()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("aba");
            editor.InsertText("ba");

            // Act
            editor.ReplaceText("ababa", "c");

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Run>().Count());
            Assert.AreEqual("c", editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenTwoRunsAndMatchesTheBeginingAndEndCharacters_ReplacesText()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("abA");
            editor.InsertText("ba");

            // Act
            editor.ReplaceText("a", "c");

            // Assert
            Assert.AreEqual(2, editor.Document.EnumerateChildrenOfType<Run>().Count());
            Assert.AreEqual("cbA", editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
            Assert.AreEqual("bc", editor.Document.EnumerateChildrenOfType<Run>().LastOrDefault().Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenSearchedContentIsBetweenThreeRuns_ReplacesText()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("mega qkiq tee");
            editor.InsertText("ee");
            editor.InsertText("st prosto.");

            // Act
            editor.ReplaceText("teeeest", "MEST");

            // Assert
            Assert.AreEqual(2, editor.Document.EnumerateChildrenOfType<Run>().Count());
            Assert.AreEqual("mega qkiq MEST", editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
            Assert.AreEqual(" prosto.", editor.Document.EnumerateChildrenOfType<Run>().LastOrDefault().Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenReplaceTextContainsSearchedText_ReplacesSearchedWord()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            string text = "tests";
            editor.InsertText(text);
            editor.InsertText(text);

            // Act
            string replaceText = "more tests";
            editor.ReplaceText(text, replaceText);

            // Assert
            foreach (Run run in editor.Document.EnumerateChildrenOfType<Run>())
            {
                Assert.AreEqual(replaceText, run.Text);
            }
        }

        [TestMethod]
        public void ReplaceWithString_WhenOnlyWholeWordsShouldBeReplace_ReplacesOnlyWholeWord()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("text Text searchedtext");
            editor.InsertText("thattext shouldnt be replaced, but this text should be replaced");
            editor.InsertText("text");

            // Act
            editor.ReplaceText("text", "NEW TEXT", false, true);

            // Assert
            string expectedDocumentText = @"NEW TEXT NEW TEXT searchedtextthattext shouldnt be replaced, but this NEW TEXT should be replacedtext";
            ReplaceTestsBase.AssertDocumentText(expectedDocumentText, editor.Document);
        }


        // REGEX
        [TestMethod]
        public void ReplaceWithRegex_WhenSearchedContentIsBetweenThreeRuns_ReplacesText()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("mega qkiq tee");
            editor.InsertText("ee");
            editor.InsertText("st prosto.");

            // Act
            Regex regex = new Regex("teeeest");
            editor.ReplaceText(regex, "MEST");

            // Assert
            Assert.AreEqual(2, editor.Document.EnumerateChildrenOfType<Run>().Count());
            Assert.AreEqual("mega qkiq MEST", editor.Document.EnumerateChildrenOfType<Run>().FirstOrDefault().Text);
            Assert.AreEqual(" prosto.", editor.Document.EnumerateChildrenOfType<Run>().LastOrDefault().Text);
        }

        [TestMethod]
        public void ReplaceWithRegex_WhenSearchedContentShouldMatchAWordContainingUpperLetter_ReplacesWordsContainingUpperLetter()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            string text = @"Welcome to RegExr v1.0 by telerik!
Edit the Expression & Text to see matches. Roll over matches or the expression for details. Undo mistakes with ctrl-z. Save & Share expressions with friends or the Community. A full Reference & Help is available in the Library, or watch the video Tutorial.
Sample text for testing:
abcdefghijklmnopqrstuvwxyz ABCDEFGHIJKLMNOPQRSTUVWXYZ
0123456789 +-.,!@#$%^&*();\/|<>";

            editor.InsertText(text);

            // Act
            Regex regex = new Regex(@"([A-Z])\w+");
            string replaceText = "TELERIK";
            editor.ReplaceText(regex, replaceText);

            // Assert
            string expectedResult = @"TELERIK to TELERIK v1.0 by telerik!TELERIK the TELERIK & TELERIK to see matches. TELERIK over matches or the expression for details. TELERIK mistakes with ctrl-z. TELERIK & TELERIK expressions with friends or the TELERIK. A full TELERIK & TELERIK is available in the TELERIK, or watch the video TELERIK.TELERIK text for testing:abcdefghijklmnopqrstuvwxyz TELERIK0123456789 +-.,!@#$%^&*();\/|<>";
            ReplaceTestsBase.AssertDocumentText(expectedResult, editor.Document);
        }

        [TestMethod]
        public void ReplaceWithRegex_WhenSearchedContentShouldMatchWordContainingUpperLetter_AndDocumentContainsMultipleRunsOnDifferentParagraphs_ReplacesWordsContainingUpperLetter()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("Video provides a powerful way to help you prove your point. When you click Online Video, you can paste in the embed code for the video you want to add. You can also type a keyword to search online for the video that best fits your document. To make your document look professionally produced, Word provides header, footer, cover page, and text box designs that complement each other.");
            editor.InsertParagraph();
            editor.InsertText("For example, you can add a matching cover page, header, and sidebar. Click Insert and then choose the elements you want from the different galleries. Themes and styles also help keep your document coordinated. When you click Design and choose a new Theme, the pictures, charts, and SmartArt graphics change to match your new theme.");
            editor.InsertParagraph();
            editor.InsertText("When you apply styles, your headings change to match the new theme. Save time in Word with new buttons that show up where you need them. To change the way a picture fits in your document, click it and a button for layout options appears next to it. When you work on a table, click where you want to add a row or a column, and then click the plus sign.");
            editor.InsertParagraph();
            editor.InsertText("hello Wor");
            editor.InsertText("ld");
            editor.InsertText(" Telerik test ");
            editor.InsertText("testTest");

            // Act
            editor.ReplaceText(new Regex(@"([A-Z])\w+"), "WORDSPROCESSING");

            // Assert
            string expectedResult = "WORDSPROCESSING provides a powerful way to help you prove your point. WORDSPROCESSING you click WORDSPROCESSING WORDSPROCESSING, you can paste in the embed" +
                " code for the video you want to add. WORDSPROCESSING can also type a keyword to search online for the video that best fits your document. WORDSPROCESSING make your document look" +
                " professionally produced, WORDSPROCESSING provides header, footer, cover page, and text box designs that complement each other.WORDSPROCESSING example, you can add a matching cover" +
                " page, header, and sidebar. WORDSPROCESSING WORDSPROCESSING and then choose the elements you want from the different galleries. WORDSPROCESSING and styles also help keep your document" +
                " coordinated. WORDSPROCESSING you click WORDSPROCESSING and choose a new WORDSPROCESSING, the pictures, charts, and WORDSPROCESSING graphics change to match your new theme.WORDSPROCESSING" +
                " you apply styles, your headings change to match the new theme. WORDSPROCESSING time in WORDSPROCESSING with new buttons that show up where you need them. WORDSPROCESSING change the way a picture" +
                " fits in your document, click it and a button for layout options appears next to it. WORDSPROCESSING you work on a table, click where you want to add a row or a column, and then click the plus sign." +
                "hello WORDSPROCESSING WORDSPROCESSING test testWORDSPROCESSING";
            AssertDocumentText(expectedResult, editor.Document);
        }

        [TestMethod]
        public void ReplaceStyling_WhenDocumentContainsOnlySearchedRun_ReplacesWordsContainingUpperLetter()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("test");

            // Act
            editor.ReplaceText(@"test", "Telerik");

            // Assert
            string expectedResult = "Telerik";
            AssertDocumentText(expectedResult, editor.Document);
        }

        [TestMethod]
        public void ReplaceStyling_WhenSearchedContentShouldMatchWordContainingUpperLetter_AndDocumentContainsSingleRunButItHasMultipleMatches_ReplacesWordsContainingUpperLetter()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("Video provides a powerful way to help you prove your point. When you click Online Video, you can paste in the embed code for the video you want to add. You can also type a keyword to search online for the video that best fits your document. To make your document look professionally produced, Word provides header, footer, cover page, and text box designs that complement each other.");

            // Act
            editor.ReplaceText(new Regex(@"([A-Z])\w+"), "TELERIK");

            // Assert
            string expectedText = "TELERIK provides a powerful way to help you prove your point. TELERIK you click TELERIK TELERIK, you can paste in the embed code for the video you want to add. TELERIK can also type a keyword to search online for the video that best fits your document. TELERIK make your document look professionally produced, TELERIK provides header, footer, cover page, and text box designs that complement each other.";

            AssertDocumentText(expectedText, editor.Document);
        }

        [TestMethod]
        public void ReplaceWithRegex_WhenSearchedContentShouldMatchAWordsAndDigits_ReplacesWordsAndDigitsWithEmptyString()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            string text = @"Welcome to RegExr v1.0 by telerik!
Edit the Expression & Text to see matches. Roll over matches or the expression for details. Undo mistakes with ctrl-z. Save & Share expressions with friends or the Community. A full Reference & Help is available in the Library, or watch the video Tutorial.
Sample text for testing:
abcdefghijklmnopqrstuvwxyz ABCDEFGHIJKLMNOPQRSTUVWXYZ
0123456789 +-.,!@#$%^&*();\/|<>";

            editor.InsertText(text);

            // Act
            Regex regex = new Regex(@"\w+");
            string replaceText = "";
            editor.ReplaceText(regex, replaceText);

            // Assert
            string expectedResult = @"   .  !   &    .        .    -.  &       .    &      ,     .   :  +-.,!@#$%^&*();\/|<>";
            AssertDocumentText(expectedResult, editor.Document);
        }

        [TestMethod]
        public void ReplaceWithRegex_WhenSearchedContentShouldMatchEveryWord_AndDocumentContainsSingleRun_ReplacesAllWordsWithStringEmpty()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            string text = @"telerik.comtest2";

            editor.InsertText(text);

            // Act
            Regex regex = new Regex(@"\w+");
            string replaceText = "";
            editor.ReplaceText(regex, replaceText);

            // Assert
            string expectedResult = @".";
            AssertDocumentText(expectedResult, editor.Document);
        }

        [TestMethod]
        public void ReplaceWithRegex_WhenSearchedContentShouldMatchEveryWord_AndDocumentContainsMultipleRuns_ReplacesAllWordsWithStringEmpty()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());

            editor.InsertText("telerik.comtest");
            editor.InsertText("this is another text2");
            editor.InsertText(" text 3");

            // Act
            Regex regex = new Regex(@"\w+");
            string replaceText = "";
            editor.ReplaceText(regex, replaceText);

            // Assert
            string expectedResult = @".     ";
            AssertDocumentText(expectedResult, editor.Document);
        }

        [TestMethod]
        public void ReplaceWithRegex_WhenEverySingleCharIsMatchedAndReplaceTextIsLonger_ReplacesAllChars()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("tttt");
            editor.InsertParagraph();
            editor.InsertText("tttt");

            // Act
            editor.ReplaceText(new Regex("."), "TT");

            string expectedResult = @"TTTTTTTTTTTTTTTT";
            AssertDocumentText(expectedResult, editor.Document);
        }

        [TestMethod]
        public void ReplaceWithRegex_WhenEverySingleCharIsMatchedAndReplaceTextIsWithSameLength_ReplacesAllChars()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("tttt");
            editor.InsertParagraph();
            editor.InsertText("tttt");

            // Act
            editor.ReplaceText(new Regex("."), "T");

            string expectedResult = @"TTTTTTTT";
            AssertDocumentText(expectedResult, editor.Document);
        }

        [TestMethod]
        public void ReplaceWithRegex_WhenEverySingleCharIsMatchedAndReplaceTextIsShorter_ReplacesAllChars()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("tttt");
            editor.InsertParagraph();
            editor.InsertText("tttt");

            // Act
            editor.ReplaceText(new Regex("."), "");

            string expectedResult = "";
            AssertDocumentText(expectedResult, editor.Document);
        }

        // Inlines between runs
        [TestMethod]
        public void ReplaceWithString_WhenImageIsBetweenRunsWhichMatchTheSearchedText_DoesntReplacesRunContent()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("te");
            editor.InsertInline(TestHelper.CreatePngImage(editor.Document));
            editor.InsertText("st");

            // Act
            editor.ReplaceText("test", "new text");

            // Assert
            Assert.AreEqual(0, editor.Document.EnumerateChildrenOfType<Run>().Where(r => r.Text == "new text").Count());

            Paragraph paragraph = editor.Document.EnumerateChildrenOfType<Paragraph>().FirstOrDefault();
            Assert.AreEqual("te", ((Run)paragraph.Inlines.FirstOrDefault()).Text);
            Assert.AreEqual("st", ((Run)paragraph.Inlines.LastOrDefault()).Text);
        }

        [TestMethod]
        public void ReplaceWithString_WhenImageIsBetweenRunsWhichMatchTheSearchedText_DoesntReplacesRunContentButReplacesTextInPreviousRun()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("test");
            editor.InsertText("te");
            editor.InsertInline(TestHelper.CreatePngImage(editor.Document));
            editor.InsertText("st");

            // Act
            editor.ReplaceText("test", "new text");

            // Assert
            Assert.AreEqual(1, editor.Document.EnumerateChildrenOfType<Run>().Where(r => r.Text == "new text").Count());

            Paragraph paragraph = editor.Document.EnumerateChildrenOfType<Paragraph>().FirstOrDefault();
            Assert.AreEqual("new text", ((Run)paragraph.Inlines.FirstOrDefault()).Text);
            Assert.AreEqual("te", ((Run)paragraph.Inlines.ElementAt(1)).Text);
            Assert.AreEqual("st", ((Run)paragraph.Inlines.LastOrDefault()).Text);
        }

        // Text in header/footer
        [TestMethod]
        public void ReplaceWithString_WhenTextIsInHeaderFooter_ReplacesText()
        {
            // Arrange
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Paragraph paragraph = editor.Document.Sections.AddSection().Headers.Add().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("text");

            // Act
            editor.ReplaceText("text", "replaced TEXT");

            // Assert
            AssertDocumentText("replaced TEXT", editor.Document);
        }
    }
}