﻿using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Fields;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Model.Drawing.Shapes;
#if WPF || SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.Theming;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Flow.Model.Protection;
using Telerik.Windows.Documents.Flow.Model.Watermarks;
using Telerik.Windows.Documents.Media;

namespace Telerik.Windows.Documents.Flow.Tests.Utilities
{
    internal static class StructureAsserter
    {
        public static void AssertDocumentsAreEqual(RadFlowDocument expected, RadFlowDocument actual)
        {
            Assert.AreEqual(expected.Sections.Count, actual.Sections.Count);
            for (int i = 0; i < expected.Sections.Count; i++)
            {
                StructureAsserter.AssertSectionsAreEqual(expected.Sections[i], actual.Sections[i]);
            }
        }

        public static void AssertThemesAreEqual(DocumentTheme expected, DocumentTheme actual)
        {
            Assert.AreEqual(expected.Name, actual.Name);
            Assert.AreEqual(expected.ColorScheme, actual.ColorScheme);
            Assert.AreEqual(expected.FontScheme, actual.FontScheme);
        }

        public static void AssertSectionsAreEqual(Section expected, Section actual)
        {
            StructureAsserter.AssertHeadersAreEqual(expected.Headers.Default, actual.Headers.Default);
            StructureAsserter.AssertHeadersAreEqual(expected.Headers.Even, actual.Headers.Even);
            StructureAsserter.AssertHeadersAreEqual(expected.Headers.First, actual.Headers.First);

            StructureAsserter.AssertFootersAreEqual(expected.Footers.Default, actual.Footers.Default);
            StructureAsserter.AssertFootersAreEqual(expected.Footers.Even, actual.Footers.Even);
            StructureAsserter.AssertFootersAreEqual(expected.Footers.First, actual.Footers.First);

            StructureAsserter.AssertBlockContainersAreEqual(expected, actual);
        }

        public static void AssertHeadersAreEqual(Header header1, Header header2)
        {
            Assert.IsTrue(header1 == null && header2 == null || header1 != null && header2 != null);
            if (header1 != null)
            {
                StructureAsserter.AssertBlockContainersAreEqual(header1, header2);
                StructureAsserter.AssertWatermarkCollectionsAreEqual(header1.Watermarks, header2.Watermarks);
            }
        }

        public static void AssertFootersAreEqual(Footer footer1, Footer footer2)
        {
            Assert.IsTrue(footer1 == null && footer2 == null || footer1 != null && footer2 != null);

            if (footer1 != null)
            {
                StructureAsserter.AssertBlockContainersAreEqual(footer1, footer2);
            }
        }

        public static void AssertTablesAreEqual(Table expected, Table actual)
        {
            Assert.AreEqual(expected.PreferredWidth.Value, actual.PreferredWidth.Value);
            Assert.AreEqual(expected.PreferredWidth.Type, actual.PreferredWidth.Type);
            Assert.AreEqual(expected.Rows.Count, actual.Rows.Count);
            for (int i = 0; i < expected.Rows.Count; i++)
            {
                StructureAsserter.AssertTableRowsAreEqual(expected.Rows[i], actual.Rows[i]);
            }
        }

        public static void AssertTableRowsAreEqual(TableRow expected, TableRow actual)
        {
            Assert.AreEqual(expected.Cells.Count, actual.Cells.Count);
            for (int i = 0; i < expected.Cells.Count; i++)
            {
                StructureAsserter.AssertTableCellsAreEqual(expected.Cells[i], actual.Cells[i]);
            }
        }

        public static void AssertTableCellsAreEqual(TableCell expected, TableCell actual)
        {
            Assert.AreEqual(expected.ColumnSpan, actual.ColumnSpan);
            Assert.AreEqual(expected.RowSpan, actual.RowSpan);

            StructureAsserter.AssertBlockContainersAreEqual(expected, actual);
        }

        public static void AssertBlockContainersAreEqual(BlockContainerBase expectedContainer, BlockContainerBase actualContainer)
        {
            Assert.AreEqual(expectedContainer.Blocks.Count, actualContainer.Blocks.Count);

            for (int i = 0; i < expectedContainer.Blocks.Count; i++)
            {
                Paragraph expectedParagraph = expectedContainer.Blocks[i] as Paragraph;
                if (expectedParagraph != null)
                {
                    Assert.IsInstanceOfType(actualContainer.Blocks[i], typeof(Paragraph));
                    StructureAsserter.AssertParagraphsAreEqual(expectedParagraph, (Paragraph)actualContainer.Blocks[i]);
                }
                else
                {
                    StructureAsserter.AssertTablesAreEqual((Table)expectedContainer.Blocks[i], (Table)actualContainer.Blocks[i]);
                }
            }
        }

        public static void AssertParagraphsAreEqual(Paragraph expected, Paragraph actual)
        {
            Assert.AreEqual(expected.Inlines.Count, actual.Inlines.Count);
            for (int i = 0; i < expected.Inlines.Count; i++)
            {
                if (expected.Inlines[i] is Run)
                {
                    Run run1 = (Run)expected.Inlines[i];
                    Run run2 = (Run)actual.Inlines[i];
                    Assert.AreEqual(run1.Text, run2.Text);
                }
                else if (expected.Inlines[i] is FieldCharacter)
                {
                    FieldCharacter fc1 = (FieldCharacter)expected.Inlines[i];
                    FieldCharacter fc2 = (FieldCharacter)actual.Inlines[i];
                    Assert.AreEqual(fc1.FieldCharacterType, fc2.FieldCharacterType);

                    AssertFieldInfosAreEqual(fc1.FieldInfo, fc2.FieldInfo);
                    // TODO: Compare actual fields
                }
                else if (expected.Inlines[i] is ImageInline)
                {
                    ImageInline img1 = (ImageInline)expected.Inlines[i];
                    ImageInline img2 = (ImageInline)actual.Inlines[i];

                    AssertImageInlinesAreEqual(img1, img2);
                }
                else if (expected.Inlines[i] is FloatingImage)
                {
                    FloatingImage f1 = (FloatingImage)expected.Inlines[i];
                    FloatingImage f2 = (FloatingImage)actual.Inlines[i];

                    AssertFloatingImagesAreEqual(f1, f2);
                }
                else if (expected.Inlines[i] is BookmarkRangeStart)
                {
                    BookmarkRangeStart b1 = (BookmarkRangeStart)expected.Inlines[i];
                    BookmarkRangeStart b2 = (BookmarkRangeStart)actual.Inlines[i];

                    Assert.AreEqual(b1.Bookmark.Name, b2.Bookmark.Name);
                }
                else if (expected.Inlines[i] is BookmarkRangeEnd)
                {
                    BookmarkRangeEnd b1 = (BookmarkRangeEnd)expected.Inlines[i];
                    BookmarkRangeEnd b2 = (BookmarkRangeEnd)actual.Inlines[i];

                    Assert.AreEqual(b1.Bookmark.Name, b2.Bookmark.Name);
                }
                else if (expected.Inlines[i] is Break)
                {
                    Break br1 = (Break)expected.Inlines[i];
                    Break br2 = (Break)actual.Inlines[i];

                    Assert.AreEqual(br1.BreakType, br2.BreakType);
                    Assert.AreEqual(br1.TextWrappingRestartLocation, br2.TextWrappingRestartLocation);
                }
                else if (expected.Inlines[i] is CommentRangeStart)
                {
                    CommentRangeStart c1 = (CommentRangeStart)expected.Inlines[i];
                    CommentRangeStart c2 = (CommentRangeStart)actual.Inlines[i];

                    Assert.AreEqual(c1.Comment.Author, c2.Comment.Author);
                    Assert.AreEqual(c1.Comment.Initials, c2.Comment.Initials);

                    AssertBlockContainersAreEqual(c1.Comment, c2.Comment);
                }
                else if (expected.Inlines[i] is CommentRangeEnd)
                {
                    CommentRangeEnd c1 = (CommentRangeEnd)expected.Inlines[i];
                    CommentRangeEnd c2 = (CommentRangeEnd)actual.Inlines[i];

                    Assert.AreEqual(c1.Comment.Author, c2.Comment.Author);
                    Assert.AreEqual(c1.Comment.Initials, c2.Comment.Initials);

                    AssertBlockContainersAreEqual(c1.Comment, c2.Comment);
                }
                else
                {
                    Assert.AreEqual(expected.Inlines[i].GetType(), actual.Inlines[i].GetType());
                }
            }
        }

        public static void AsserProtectionSettingsAreEqual(RadFlowDocument expected, RadFlowDocument actual, bool excludeHashAndSalt = false)
        {
            ProtectionSettings expectedSettings = expected.Document.ProtectionSettings;
            ProtectionSettings actualSettings = actual.Document.ProtectionSettings;
            Assert.AreEqual(expectedSettings.AlgorithmName, actualSettings.AlgorithmName);
            Assert.AreEqual(expectedSettings.Enforced, actualSettings.Enforced);
            Assert.AreEqual(expectedSettings.ProtectionMode, actualSettings.ProtectionMode);
            Assert.AreEqual(expectedSettings.SpinCount, actualSettings.SpinCount);

            if (!excludeHashAndSalt)
            {
                Assert.AreEqual(expectedSettings.Salt, actualSettings.Salt);
                Assert.AreEqual(expectedSettings.Hash, actualSettings.Hash);
            }
        }

        public static void AssertWatermarksAreEqual(Watermark expectedWatermark, Watermark actualWatermark)
        {
            Assert.AreEqual(expectedWatermark.WatermarkType, actualWatermark.WatermarkType);

            if (expectedWatermark.WatermarkType == WatermarkType.Text)
            {
                StructureAsserter.AssertWatermarkTextSettingsAreEqual(expectedWatermark.TextSettings, actualWatermark.TextSettings);
            }
            else // watermark1.Type == WatermarkType.Image
            {
                StructureAsserter.AssertWatermarkImageSettingsAreEqual(expectedWatermark.ImageSettings, actualWatermark.ImageSettings);
            }
        }

        public static void AssertWatermarkImageSettingsAreEqual(ImageWatermarkSettings expectedWatermarkSettings, ImageWatermarkSettings actualWatermarkSettings)
        {
            Assert.AreEqual(expectedWatermarkSettings.Angle, actualWatermarkSettings.Angle);
            Assert.AreEqual(expectedWatermarkSettings.Height, actualWatermarkSettings.Height, 0.001);
            Assert.AreEqual(expectedWatermarkSettings.Width, actualWatermarkSettings.Width, 0.001);

            if (expectedWatermarkSettings.ImageSource == null || actualWatermarkSettings.ImageSource == null)
            {
                Assert.IsNull(expectedWatermarkSettings.ImageSource);
                Assert.IsNull(actualWatermarkSettings.ImageSource);
            }
            else
            {
                StructureAsserter.AssertImageSourcesAreEqual(expectedWatermarkSettings.ImageSource, actualWatermarkSettings.ImageSource);
            }
        }

        public static void AssertWatermarkTextSettingsAreEqual(TextWatermarkSettings expectedWatermarkSettings, TextWatermarkSettings actualWatermarkSettings)
        {
            Assert.AreEqual(expectedWatermarkSettings.Angle, actualWatermarkSettings.Angle);
            Assert.AreEqual(expectedWatermarkSettings.Height, actualWatermarkSettings.Height, 0.001);
            Assert.AreEqual(expectedWatermarkSettings.Width, actualWatermarkSettings.Width, 0.001);
            Assert.AreEqual(expectedWatermarkSettings.Text, actualWatermarkSettings.Text);
            Assert.AreEqual(expectedWatermarkSettings.Opacity, actualWatermarkSettings.Opacity, 0.001);
            Assert.AreEqual(expectedWatermarkSettings.ForegroundColor, actualWatermarkSettings.ForegroundColor);
            Assert.AreEqual(expectedWatermarkSettings.FontFamily, actualWatermarkSettings.FontFamily);
        }

        private static void AssertFloatingImagesAreEqual(FloatingImage f1, FloatingImage f2)
        {
            AssertImagesAreEqual(f1.Image, f2.Image);

            Assert.AreEqual(f1.Margin, f2.Margin);
            Assert.AreEqual(f1.AllowOverlap, f2.AllowOverlap);
            Assert.AreEqual(f1.HorizontalPosition, f2.HorizontalPosition);
            Assert.AreEqual(f1.VerticalPosition, f2.VerticalPosition);
            Assert.AreEqual(f1.IsBehindDocument, f2.IsBehindDocument);
            Assert.AreEqual(f1.LayoutInCell, f2.LayoutInCell);
            Assert.AreEqual(f1.Wrapping, f2.Wrapping);
            Assert.AreEqual(f1.ZIndex, f2.ZIndex);
        }

        private static void AssertImageInlinesAreEqual(ImageInline img1, ImageInline img2)
        {
            AssertImagesAreEqual(img1.Image, img2.Image);
        }

        private static void AssertImagesAreEqual(Image img1, Image img2)
        {
            Assert.AreEqual(img1.Width, img2.Width);
            Assert.AreEqual(img1.Height, img2.Height);
            Assert.AreEqual(img1.RotationAngle, img2.RotationAngle);
            Assert.AreEqual(img1.IsHorizontallyFlipped, img2.IsHorizontallyFlipped);
            Assert.AreEqual(img1.IsVerticallyFlipped, img2.IsVerticallyFlipped);
            if (img1.ImageSource == null)
            {
                Assert.IsNull(img2.ImageSource);
            }
            else
            {
                StructureAsserter.AssertImageSourcesAreEqual(img1.ImageSource, img2.ImageSource);
            }
        }

        private static void AssertImageSourcesAreEqual(ImageSource expectedImageSource, ImageSource actualImageSource)
        {
            Assert.AreEqual(expectedImageSource.Extension, actualImageSource.Extension, true);
            Assert.AreEqual(expectedImageSource.Data.Length, actualImageSource.Data.Length);
            Assert.IsTrue(expectedImageSource.Data.SequenceEqual(actualImageSource.Data));
        }

        private static void AssertFieldInfosAreEqual(FieldInfo f1, FieldInfo f2)
        {
            Assert.AreEqual(f1.IsLocked, f2.IsLocked);
            Assert.AreEqual(f1.IsDirty, f2.IsDirty);

            AssertFieldsAreEqual(f1.Field, f2.Field);
        }

        private static void AssertFieldsAreEqual(Field fd1, Field fd2)
        {
            Assert.AreEqual(fd1.GetType(), fd2.GetType());

            Assert.AreEqual(fd1.DateTimeFormatting, fd2.DateTimeFormatting);
            Assert.AreEqual(fd1.NumericFormatting, fd2.NumericFormatting);
            Assert.AreEqual(fd1.GeneralFormatting, fd2.GeneralFormatting);

            if (fd1 is Hyperlink)
            {
                Hyperlink h1 = (Hyperlink)fd1;
                Hyperlink h2 = (Hyperlink)fd2;
                Assert.AreEqual(h1.Uri, h2.Uri);
                Assert.AreEqual(h1.ToolTip, h2.ToolTip);
                Assert.AreEqual(h1.IsAnchor, h2.IsAnchor);
            }
        }

        private static void AssertWatermarkCollectionsAreEqual(WatermarkCollection expectedCollection, WatermarkCollection actualCollection)
        {
            Assert.AreEqual(expectedCollection.Count(), actualCollection.Count());

            for (int i = 0; i < expectedCollection.Count(); i++)
            {
                Watermark watermark1 = expectedCollection[i];
                Watermark watermark2 = actualCollection[i];

                StructureAsserter.AssertWatermarksAreEqual(watermark1, watermark2);
            }
        }
    }
}
