﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Lists;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Model.Styles.Core;

namespace Telerik.Windows.Documents.Flow.Tests.Utilities
{
    internal static class StylePropertyAsserter
    {
        public static void AssertListsAreEqual(RadFlowDocument expected, RadFlowDocument actual)
        {
            Assert.AreEqual(expected.Lists.Count, actual.Lists.Count);
            foreach (List expectedList in expected.Lists)
            {
                List actualList = actual.Lists.GetList(expectedList.Id);
                StylePropertyAsserter.AssertListsAreEqual(expectedList, actualList);
            }
        }

        public static void AssertListsAreEqual(List expected, List actual)
        {
            Assert.AreEqual(expected.MultilevelType, actual.MultilevelType);
            Assert.AreEqual(expected.StyleId, actual.StyleId);
            Assert.AreEqual(expected.Levels.Count, actual.Levels.Count);

            for (int i = 0; i < expected.Levels.Count; i++)
            {
                ListLevel expectedLevel = expected.Levels[i];
                ListLevel actualLevel = actual.Levels[i];

                StylePropertyAsserter.AssertListLevelsAreEqual(expectedLevel, actualLevel);
            }
        }

        public static void AssertListLevelsAreEqual(ListLevel expected, ListLevel actual)
        {
            Assert.AreEqual(expected.Alignment, actual.Alignment);
            Assert.AreEqual(expected.IsLegal, actual.IsLegal);
            Assert.AreEqual(expected.NumberingStyle, actual.NumberingStyle);
            Assert.AreEqual(expected.NumberTextFormat, actual.NumberTextFormat);
            Assert.AreEqual(expected.RestartAfterLevel, actual.RestartAfterLevel);
            Assert.AreEqual(expected.StartIndex, actual.StartIndex);
            Assert.AreEqual(expected.StyleId, actual.StyleId);
            StylePropertyAsserter.AssertPropertiesAreEqual(expected.CharacterProperties, actual.CharacterProperties);
            StylePropertyAsserter.AssertPropertiesAreEqual(expected.ParagraphProperties, actual.ParagraphProperties);
        }

        public static void AssertStylesAreEqual(RadFlowDocument expected, RadFlowDocument actual, bool assertOnlyActualValuesOnBaseStyles = false)
        {
            List<Style> expectedStyles = expected.StyleRepository.Styles.OrderBy(s => s.Id).ToList();
            List<Style> actualStyles = actual.StyleRepository.Styles.OrderBy(s => s.Id).ToList();

            Assert.AreEqual(expectedStyles.Count, actualStyles.Count);
            for (int i = 0; i < expectedStyles.Count; i++)
            {
                StylePropertyAsserter.AssertStylesAreEqual(expectedStyles[i], actualStyles[i], assertOnlyActualValuesOnBaseStyles);
            }
        }

        public static void AssertStylesAreEqual(Style expected, Style actual, bool assertOnlyActualValuesOnBaseStyle = false)
        {
            Assert.AreEqual(expected.Id, actual.Id);
            Assert.AreEqual(expected.Name, actual.Name);

            Assert.AreEqual(expected.StyleType, actual.StyleType);
            Assert.AreEqual(expected.BasedOnStyleId, actual.BasedOnStyleId);
            Assert.AreEqual(expected.LinkedStyleId, actual.LinkedStyleId);
            Assert.AreEqual(expected.NextStyleId, actual.NextStyleId);

            Assert.AreEqual(expected.IsCustom, actual.IsCustom);
            Assert.AreEqual(expected.IsDefault, actual.IsDefault);
            Assert.AreEqual(expected.IsPrimary, actual.IsPrimary);

            Assert.AreEqual(expected.UIPriority, actual.UIPriority);

            bool assertOnlyActualValues = assertOnlyActualValuesOnBaseStyle && (string.IsNullOrEmpty(actual.BasedOnStyleId) || actual.Id == BuiltInStyleNames.NormalWebStyleId);

            AssertPropertiesAreEqual(expected.CharacterProperties, actual.CharacterProperties, assertOnlyActualValues);
            if (expected.StyleType == StyleType.Paragraph || expected.StyleType == StyleType.Table)
            {
                AssertPropertiesAreEqual(expected.ParagraphProperties, actual.ParagraphProperties, assertOnlyActualValues);
            }

            if (expected.StyleType == StyleType.Table)
            {
                AssertPropertiesAreEqual(expected.TableProperties, actual.TableProperties, assertOnlyActualValues);
                AssertPropertiesAreEqual(expected.TableCellProperties, actual.TableCellProperties, assertOnlyActualValues);
                AssertPropertiesAreEqual(expected.TableRowProperties, actual.TableRowProperties, assertOnlyActualValues);
            }
        }

        public static void AssertPropertyValue(Style style, IStylePropertyDefinition propertyDefinition, object expectedValue)
        {
            object actualValue = style.GetPropertyValue(propertyDefinition);
            ConfirmValues(actualValue, expectedValue);
        }

        public static void AssertPropertyValue(IElementWithProperties element, IStylePropertyDefinition propertyDefinition, object value)
        {
            IStyleProperty styleProperty = element.Properties.GetStyleProperty(propertyDefinition);
            ConfirmStylePropertyValue(styleProperty, value);
        }

        public static void ConfirmStylePropertyValue(IStyleProperty styleProperty, object expectedValue)
        {
            object actualValue = styleProperty.GetActualValueAsObject();
            ConfirmValues(actualValue, expectedValue);
        }

        public static void ConfirmValues(object actualValue, object expectedValue)
        {
            Assert.AreEqual(expectedValue, actualValue, "The expected value doesn't match: [expected: {0}] [actual {1}]", expectedValue, actualValue);
        }

        public static void VerifyCopyProperties(DocumentElementPropertiesBase from, DocumentElementPropertiesBase to)
        {
            foreach (IStyleProperty toProperty in to.StyleProperties)
            {
                IStyleProperty fromProperty = from.GetStyleProperty(toProperty.PropertyDefinition);

                Assert.AreEqual(toProperty.HasLocalValue, fromProperty.HasLocalValue);
                Assert.AreEqual(toProperty.GetLocalValueAsObject(), fromProperty.GetLocalValueAsObject());
                Assert.AreEqual(toProperty.GetActualValueAsObject(), fromProperty.GetActualValueAsObject());
            }
        }

        public static void HasNoLocalProperties(IEnumerable<IStyleProperty> properties)
        {
            foreach (IStyleProperty styleProperty in properties)
            {
                Type type = styleProperty.GetType();

                // LocalPorperties always have local values - so skip them.
                if (!type.Name.StartsWith("LocalProperty"))
                {
                    Assert.IsFalse(styleProperty.HasLocalValue);
                }
                else
                {
                    Assert.AreEqual(styleProperty.GetActualValueAsObject(), styleProperty.PropertyDefinition.GetDefaultValueAsObject());
                }
            }
        }

        public static void AssertPropertiesAreEqual(IElementWithProperties expected, IElementWithProperties actual, StylePropertiesAsserterInfo asserterInfo = null)
        {
            Assert.IsNotNull(expected);
            Assert.IsNotNull(actual);

            AssertPropertiesAreEqual(expected.Properties, actual.Properties, asserterInfo);
        }

        public static void AssertPropertiesAreEqual(DocumentElementPropertiesBase expected, DocumentElementPropertiesBase actual)
        {
            AssertPropertiesAreEqual(expected, actual, false, null);
        }

        public static void AssertPropertiesAreEqual(DocumentElementPropertiesBase expected, DocumentElementPropertiesBase actual, StylePropertiesAsserterInfo asserterInfo = null)
        {
            AssertPropertiesAreEqual(expected, actual, false, asserterInfo);
        }

        private static void AssertPropertiesAreEqual(DocumentElementPropertiesBase expected, DocumentElementPropertiesBase actual, bool assertOnlyActualValues)
        {
            AssertPropertiesAreEqual(expected, actual, assertOnlyActualValues, null);
        }

        private static void AssertPropertiesAreEqual(DocumentElementPropertiesBase expected, DocumentElementPropertiesBase actual, bool assertOnlyActualValues, StylePropertiesAsserterInfo asserterInfo = null)
        {
            string comparisonResult = BuildComparisonString(expected, actual, string.Empty, assertOnlyActualValues, asserterInfo);

            if (expected is ParagraphProperties && actual is ParagraphProperties)
            {
                CharacterProperties expectedCharacterProperties = ((ParagraphProperties)expected).ParagraphMarkerProperties;
                CharacterProperties actualCharacterProperties = ((ParagraphProperties)actual).ParagraphMarkerProperties;

                if (expectedCharacterProperties == null ^ actualCharacterProperties == null)
                {
                    comparisonResult += "ParagraphMarkerProperties are different.";
                }
                else
                {
                    if (expectedCharacterProperties != null && actualCharacterProperties != null)
                    {
                        comparisonResult += BuildComparisonString(expectedCharacterProperties, actualCharacterProperties, "ParagraphMarkerProperties.", assertOnlyActualValues, asserterInfo);
                    }
                }
            }

            Assert.IsTrue(comparisonResult == string.Empty, Environment.NewLine + comparisonResult);
        }

        public static void AssertDefaultStylePropertiesAreEqual(DocumentDefaultStyle expected, DocumentDefaultStyle actual)
        {
            AssertPropertiesAreEqual(expected.CharacterProperties, actual.CharacterProperties);
            AssertPropertiesAreEqual(expected.ParagraphProperties, actual.ParagraphProperties);
        }

        private static string BuildComparisonString(
            DocumentElementPropertiesBase expected,
            DocumentElementPropertiesBase actual,
            string messagePrefix,
            bool compareOnlyActualValues,
            StylePropertiesAsserterInfo asserterInfo = null)
        {
            StringBuilder sb = new StringBuilder();

            asserterInfo = asserterInfo ?? new StylePropertiesAsserterInfo();
            bool shouldCheckStyleId = asserterInfo.ShouldCheckProperty("StyleID");
            if (shouldCheckStyleId && expected.StyleId != actual.StyleId)
            {
                sb.AppendLine(string.Format(
                    "{0}StyleId is different. Expected: {1} Actual: {2}",
                    messagePrefix,
                    expected.StyleId,
                    actual.StyleId));
            }

            foreach (IStyleProperty expectedProperty in expected.StyleProperties)
            {
                if (!asserterInfo.ShouldCheckProperty(expectedProperty.PropertyDefinition))
                {
                    continue;
                }

                IStyleProperty actualProperty = actual.GetStyleProperty(expectedProperty.PropertyDefinition);

                if (!compareOnlyActualValues)
                {
                    if (asserterInfo.ShouldCompareHasLocalValue(expectedProperty.PropertyDefinition.PropertyName))
                    {
                        if (expectedProperty.HasLocalValue != actualProperty.HasLocalValue)
                        {
                            sb.AppendLine(string.Format(
                                "HasLocalValue for {0}{1} is different. Expected: {2} Actual: {3}",
                                messagePrefix,
                                actualProperty.PropertyDefinition.PropertyName,
                                expectedProperty.HasLocalValue,
                                actualProperty.HasLocalValue));
                        }
                    }
                }

                if (expectedProperty.HasLocalValue || actualProperty.HasLocalValue)
                {
                    object expectedActualValue = expectedProperty.GetActualValueAsObject();
                    object actualActualValue = actualProperty.GetActualValueAsObject();

                    bool valuesAreCloseDoubles = TestHelper.AreClose(expectedActualValue, actualActualValue);

                    if (!valuesAreCloseDoubles && !object.Equals(expectedActualValue, actualActualValue))
                    {
                        sb.AppendLine(string.Format(
                            "ActualValue for {0}{1} is different. Expected: {2} Actual: {3}",
                            messagePrefix,
                            actualProperty.PropertyDefinition.PropertyName,
                            expectedActualValue,
                            actualActualValue));
                    }
                }
            }

            string comparisonResult = sb.ToString();
            return comparisonResult;
        }

        public static void AssertPropertiesAreEqualForAllElementsOfType<T>(RadFlowDocument expected, RadFlowDocument actual)
            where T : DocumentElementBase, IElementWithProperties
        {
            StructureAsserter.AssertDocumentsAreEqual(expected, actual);

            List<T> expectedElements = expected.EnumerateChildrenOfType<T>().ToList();
            List<T> actualElements = actual.EnumerateChildrenOfType<T>().ToList();

            for (int i = 0; i < expectedElements.Count; i++)
            {
                StylePropertyAsserter.AssertPropertiesAreEqual(expectedElements[i], actualElements[i]);
            }
        }
    }
}
