﻿using System;
using System.IO;
using System.Linq;
using System.Reflection;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Common.FormatProviders;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Media;
#if !NETSTANDART
using System.Windows;
#endif
using Telerik.Windows.Documents.Primitives;

namespace Telerik.Windows.Documents.Flow.Tests.Utilities
{
    internal static class TestHelper
    {
        private const string GoBackBookmarkName = "_GoBack";
        private const string TestResourcesFolderName = @"TestDocuments";
        private const string TestImagesFolderName = @"TestImages";

        public static RadFlowDocument GetDocumentFromFile(string fileName, IFormatProvider<RadFlowDocument> provider)
        {
            string path = string.Format("{0}/{1}", TestResourcesFolderName, fileName);

            RadFlowDocument result;
            using (Stream stream = GetResourceStream(path))
            {
                result = provider.Import(stream);
            }

            RemoveGoBackBookmark(result);

            foreach (Comment comment in result.Comments)
            {
                foreach (BlockBase blockBase in comment.Blocks)
                {
                    RemoveGoBackBookmark(blockBase);
                }
            }
            return result;
        }

        private static void RemoveGoBackBookmark(DocumentElementBase result)
        {
            BookmarkRangeStart goBackStart = result.EnumerateChildrenOfType<BookmarkRangeStart>()
                .Where(b => b.Bookmark.Name == GoBackBookmarkName).FirstOrDefault();

            if (goBackStart != null)
            {
                BookmarkRangeEnd goBackEnd = goBackStart.Bookmark.BookmarkRangeEnd;
                Paragraph pStart = goBackStart.Paragraph;
                pStart.Inlines.Remove(goBackStart);

                Paragraph pEnd = goBackEnd.Paragraph;
                if (pEnd != null) // The end element can be hanging
                {
                    pEnd.Inlines.Remove(goBackEnd);
                }
            }
        }

#if NETSTANDARD
        private static Assembly GetCurrentAssembly()
        {
            AssemblyName assemblyName = typeof(TestHelper).GetTypeInfo().Assembly.GetName();
            return Assembly.Load(new AssemblyName(assemblyName.Name));
        }

        public static Stream GetResourceStream(string resource)
        {
            resource = resource.Replace('/', '.');
            resource = resource.Replace('\\', '.');
            Assembly assembly = GetCurrentAssembly();
            string[] resourceNames = assembly.GetManifestResourceNames();
            string resourceName = resourceNames.Where(p => p.Contains(resource)).FirstOrDefault();

            if (string.IsNullOrEmpty(resourceName))
            {
                throw new ArgumentException("Missing resource with given name.", nameof(resource));
            }

            return assembly.GetManifestResourceStream(resourceName);
        }
#else

        public static Stream GetResourceStream(string resource)
        {
            var streamInfo = Application.GetResourceStream(GetResourceUri(resource));
            if (streamInfo != null)
            {
                return streamInfo.Stream;
            }

            return null;
        }
#endif

        public static Uri GetResourceUri(string resource)
        {
            AssemblyName assemblyName = new AssemblyName(typeof(TestHelper).Assembly.FullName);
            string resourcePath = string.Format("/{0};component/{1}", assemblyName.Name, resource);
            Uri resourceUri = new Uri(resourcePath, UriKind.Relative);

            return resourceUri;
        }

        public static ImageInline CreatePngImage(RadFlowDocument doc)
        {
            return GetInlineImage(doc, "png");
        }

        public static ImageInline CreateJpgImage(RadFlowDocument doc)
        {
            return GetInlineImage(doc, "jpg");
        }

        public static ImageInline CreateBmpImage(RadFlowDocument doc)
        {
            return GetInlineImage(doc, "bmp");
        }

        public static FloatingImage CreatePngFloatingImage(RadFlowDocument doc)
        {
            return GetFloatingImage(doc, "png");
        }

        public static FloatingImage CreateJpgFloatingImage(RadFlowDocument doc)
        {
            return GetFloatingImage(doc, "jpg");
        }

        public static FloatingImage CreateBmpFloatingImage(RadFlowDocument doc)
        {
            return GetFloatingImage(doc, "bmp");
        }

        public static void AssertExceptionIsThrown<TException>(Action action)
            where TException : Exception
        {
            try
            {
                action();
                Assert.Fail(string.Format("Expected exception type:{0} was not thrown.",
                    typeof(TException).Name));
            }
            catch (TException)
            {
            }
            catch (Exception ex)
            {
                Assert.Fail(string.Format(
                    "Expected exception type: '{0}', Actual: '{1}'",
                    typeof(TException).Name,
                    ex.GetType().Name));
            }
        }

        private static FloatingImage GetFloatingImage(RadFlowDocument doc, string ext)
        {
            FloatingImage res = new FloatingImage(doc);
            res.Image.Width = 90;
            res.Image.Height = 30;

            string path = string.Format("{0}/test.{1}", TestImagesFolderName, ext);
            using (Stream stream = GetResourceStream(path))
            {
                res.Image.ImageSource = new ImageSource(stream, ext);
            }

            return res;
        }

        private static ImageInline GetInlineImage(RadFlowDocument doc, string ext)
        {
            ImageInline res = new ImageInline(doc);
            res.Image.Width = 90;
            res.Image.Height = 30;

            string path = string.Format("{0}/test.{1}", TestImagesFolderName, ext);
            using (Stream stream = GetResourceStream(path))
            {
                res.Image.ImageSource = new ImageSource(stream, ext);
            }

            return res;
        }

        public static Stream GetImageStream(string imageName)
        {
            string path = string.Format("{0}/{1}", TestImagesFolderName, imageName);
            Stream stream = GetResourceStream(path);
            return stream;
        }

        public static bool AreClose(object x, object y, double eps = 0.1)
        {
            if (x is double)
            {
                return TestHelper.AreClose((double)x, (double)y, eps);
            }
            else if (x is Padding)
            {
                Padding xPadding = (Padding)x;
                Padding yPadding = (Padding)y;

                return TestHelper.AreClose(xPadding.Left, yPadding.Left, eps) &&
                    TestHelper.AreClose(xPadding.Top, yPadding.Top, eps) &&
                    TestHelper.AreClose(xPadding.Right, yPadding.Right, eps) &&
                    TestHelper.AreClose(xPadding.Bottom, yPadding.Bottom, eps);
            }

            return false;
        }

        public static bool AreClose(double x, double y, double eps = 0.1)
        {
            return Math.Abs(x - y) < eps;
        }
    }
}
