﻿using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Media;
#else
using System.Windows;
using System.Windows.Media;
#endif
using CsQuery;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using PreMailer.Net;
using Telerik.Windows.Documents.Flow.FormatProviders.Html;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;
#if WPF || SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Flow.InternalTests.Integration.Html.Export
{
    [TestClass]
    public class StyleExportTests
    {
        [TestMethod]
        public void ParagraphWithoutStyleId_IsExportedWithNormalStyle()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            document.Sections.AddSection().Blocks.AddParagraph().Inlines.AddRun("text");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            System.Collections.Generic.List<IDomObject> paragraphs = cq.Select("p").ToList();
            Assert.AreEqual(1, paragraphs.Count);
            Assert.AreEqual("TelerikNormal", paragraphs[0].Attributes["class"]);
        }

        [TestMethod]
        public void ParagraphWithNormalWebStyle_IsExportedWithoutClasses()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.NormalWebStyleId);

            Paragraph paragraph = document.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("text");
            paragraph.StyleId = BuiltInStyleNames.NormalWebStyleId;


            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            System.Collections.Generic.List<IDomObject> paragraphs = cq.Select("p").ToList();
            Assert.AreEqual(1, paragraphs.Count);
            Assert.AreEqual(string.Empty, paragraphs[0].Attributes["class"]);
        }

        [TestMethod]
        public void BuiltInStyle_IsExportedWithTelerikPrefix()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Style style = document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.CaptionStyleId);

            Paragraph paragraph = document.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("text");
            paragraph.StyleId = style.Id;

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            System.Collections.Generic.List<IDomObject> paragraphs = cq.Select("p").ToList();
            Assert.AreEqual(1, paragraphs.Count);
            Assert.AreEqual("TelerikNormal TelerikCaption", paragraphs[0].Attributes["class"]);
        }

        [TestMethod]
        public void BuiltInStyle_IsExportedAsInlineStyling()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Style style = document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.HyperlinkStyleId);
            style.CharacterProperties.ForegroundColor.LocalValue = new ThemableColor(Colors.Red);
            document.StyleRepository.Add(style);

            Paragraph paragraph = document.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("text");
            paragraph.StyleId = style.Id;

            // Act
            HtmlExportSettings exportSettings = new HtmlExportSettings();
            exportSettings.StylesExportMode = StylesExportMode.Inline;
            CQ cq = ExportHelper.ExportDocumentToCq(document, exportSettings);

            // Assert
            System.Collections.Generic.List<IDomObject> paragraphs = cq.Select("p").ToList();
            Assert.AreEqual(1, paragraphs.Count);
            Assert.IsTrue(paragraphs[0].Attributes["style"].Contains("color: #FF0000"));
        }

        [TestMethod]
        public void CustomStyle_IsExportedAsInlineStyling()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Style style = new Style("CustomStyle", StyleType.Paragraph);
            style.CharacterProperties.ForegroundColor.LocalValue = new ThemableColor(Colors.Red);
            document.StyleRepository.Add(style);

            Paragraph paragraph = document.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("text");
            paragraph.StyleId = style.Id;

            // Act
            HtmlExportSettings exportSettings = new HtmlExportSettings();
            exportSettings.StylesExportMode = StylesExportMode.Inline;
            CQ cq = ExportHelper.ExportDocumentToCq(document, exportSettings);

            // Assert
            System.Collections.Generic.List<IDomObject> paragraphs = cq.Select("p").ToList();
            Assert.AreEqual(1, paragraphs.Count);
            Assert.IsTrue(paragraphs[0].Attributes["style"].Contains("color: #FF0000"));
        }

        [TestMethod]
        public void PropertiesSetInDocumentDefaultStyles_AreExportedOnAllBaseStyles()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            document.DefaultStyle.ParagraphProperties.TextAlignment.LocalValue = Alignment.Center;
            document.DefaultStyle.CharacterProperties.FontFamily.LocalValue = new ThemableFontFamily("Comic Sans MS");

            Style paragraphBaseStyle = new Style("StyleBase", StyleType.Paragraph);
            document.StyleRepository.Add(paragraphBaseStyle);
            Style paragraphDerivedStyle = new Style("StyleDerived", StyleType.Paragraph);
            paragraphDerivedStyle.BasedOnStyleId = paragraphBaseStyle.Id;
            document.StyleRepository.Add(paragraphDerivedStyle);

            Paragraph paragraph = document.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("text");
            paragraph.StyleId = paragraphDerivedStyle.Id;

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            Assert.AreEqual("center", parser.Styles[".StyleBase"].Attributes["text-align"]);
            Assert.AreEqual("comic sans ms", parser.Styles[".StyleBase"].Attributes["font-family"]);

            Assert.IsFalse(parser.Styles[".StyleDerived"].Attributes.ContainsKey("text-align"));
            Assert.IsFalse(parser.Styles[".StyleDerived"].Attributes.ContainsKey("font-family"));
        }

        [TestMethod]
        public void PropertiesSetInNormal_AreExportedInNormalWeb()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            Style normalStyle = document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.NormalStyleId);
            normalStyle.ParagraphProperties.TextAlignment.LocalValue = Alignment.Center;
            normalStyle.CharacterProperties.FontWeight.LocalValue = FontWeights.Bold;

            document.StyleRepository.AddBuiltInStyle(BuiltInStyleNames.NormalWebStyleId);

            Paragraph paragraph = document.Sections.AddSection().Blocks.AddParagraph();
            paragraph.Inlines.AddRun("text");
            paragraph.StyleId = BuiltInStyleNames.NormalWebStyleId;

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            Assert.AreEqual("center", parser.Styles["p"].Attributes["text-align"]);
            Assert.AreEqual("bold", parser.Styles["p"].Attributes["font-weight"]);

            Assert.AreEqual("center", parser.Styles[".TelerikNormal"].Attributes["text-align"]);
            Assert.AreEqual("bold", parser.Styles[".TelerikNormal"].Attributes["font-weight"]);
        }
    }
}
