﻿using CsQuery;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using PreMailer.Net;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Media;
#else
using System.Windows.Media;
#endif
using Telerik.Windows.Documents.Flow.FormatProviders.Html;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Model.Styles;
#if WPF || SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Flow.InternalTests.Integration.Html.Export
{
    [TestClass]
    public class TableStyles
    {
        [TestMethod]
        public void TableStyle_WithAllSameBorders_IsExportedCorrectlyForAllCells()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            tableStyle.TableProperties.Borders.LocalValue = new TableBorders(redBorder);

            document.StyleRepository.Add(tableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            this.AssertTable_AllBorders(tables[0], tableStyle, parser);
        }

        // TODO: check the note in the asserts for the cell borders.
        [TestMethod]
        public void TableStyle_WithOuterBorders_IsExportedOnlyForOuterBorders()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders(redBorder, redBorder, redBorder, redBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;

            table.Rows.First().Cells.First().Blocks.AddParagraph().Inlines.AddRun("1-1");
            table.Rows.First().Cells.Last().Blocks.AddParagraph().Inlines.AddRun("1-2");
            table.Rows.Last().Cells.First().Blocks.AddParagraph().Inlines.AddRun("2-1");
            table.Rows.Last().Cells.Last().Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("4px solid #ff0000", styleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            Assert.IsFalse(styleCellPart.Attributes.ContainsKey("border"));

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(1, tables.Count);

            // Those borders are duplicating the table style. Think of refactoring the TableCellBordersConvert to exclude them.
            this.AssertTable_OuterBorders(tables[0], tableStyle, table.HasCellSpacing);
        }

        [TestMethod]
        public void TableStyle_WithInnerBorders_IsExportedOnlyForInnerBorders()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders();
            tableBorders = new TableBorders(tableBorders, insideHorizontalBorder: redBorder, insideVerticalBorder: redBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;

            table.Rows.First().Cells.First().Blocks.AddParagraph().Inlines.AddRun("1-1");
            table.Rows.First().Cells.Last().Blocks.AddParagraph().Inlines.AddRun("1-2");
            table.Rows.Last().Cells.First().Blocks.AddParagraph().Inlines.AddRun("2-1");
            table.Rows.Last().Cells.Last().Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);
            Assert.IsFalse(styleCellPart.Attributes.ContainsKey("border"));

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(1, tables.Count);
            Assert.IsTrue(string.IsNullOrEmpty(tables[0].Attributes["style"]));

            this.AssertTable_InnerBorders(tables[0], tableStyle, table.HasCellSpacing);
        }

        [TestMethod]
        public void TableStyle_WithDifferentOuterAndInnerBorders_AreOuterAndInnerBordersExportedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));
            Border greenBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Green));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders(redBorder, redBorder, redBorder, redBorder, greenBorder, greenBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;

            table.Rows.First().Cells.First().Blocks.AddParagraph().Inlines.AddRun("1-1");
            table.Rows.First().Cells.Last().Blocks.AddParagraph().Inlines.AddRun("1-2");
            table.Rows.Last().Cells.First().Blocks.AddParagraph().Inlines.AddRun("2-1");
            table.Rows.Last().Cells.Last().Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("4px solid #ff0000", styleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            Assert.IsFalse(styleCellPart.Attributes.ContainsKey("border"));

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(1, tables.Count);
            Assert.IsTrue(tables[0].Attributes["class"].Contains(tableStyle.Id));
            Assert.IsTrue(string.IsNullOrEmpty(tables[0].Attributes["style"]));

            System.Collections.Generic.List<IDomObject> cells = cq.Select("td").ToList();
            Assert.AreEqual(4, cells.Count);

            Assert.IsTrue(cells[0].Attributes["style"].Contains("border-left: 4px solid #FF0000;border-top: 4px solid #FF0000;border-right: 4px solid #008000;border-bottom: 4px solid #008000;"));
            Assert.IsTrue(cells[1].Attributes["style"].Contains("border-left: 4px solid #008000;border-top: 4px solid #FF0000;border-right: 4px solid #FF0000;border-bottom: 4px solid #008000;"));
            Assert.IsTrue(cells[2].Attributes["style"].Contains("border-left: 4px solid #FF0000;border-top: 4px solid #008000;border-right: 4px solid #008000;border-bottom: 4px solid #FF0000;"));
            Assert.IsTrue(cells[3].Attributes["style"].Contains("border-left: 4px solid #008000;border-top: 4px solid #008000;border-right: 4px solid #FF0000;border-bottom: 4px solid #FF0000;"));
        }

        [TestMethod]
        public void TableStyle_WithDifferentOuterAndInnerBordersAndCellsHasLocalBorders_AreCellBordersExportedLocaly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));
            Border greenBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Green));
            Border orangeBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Orange));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders(redBorder, redBorder, redBorder, redBorder, greenBorder, greenBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;

            TableCell cellOne = table.Rows.First().Cells.First();
            cellOne.Blocks.AddParagraph().Inlines.AddRun("1-1");
            cellOne.Borders = new TableCellBorders(orangeBorder);

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");
            cellTwo.Borders = new TableCellBorders(orangeBorder);

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");
            cellThree.Borders = new TableCellBorders(orangeBorder);

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");
            cellFour.Borders = new TableCellBorders(orangeBorder);

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("4px solid #ff0000", styleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            Assert.IsFalse(styleCellPart.Attributes.ContainsKey("border"));

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(1, tables.Count);
            Assert.IsTrue(tables[0].Attributes["class"].Contains(tableStyle.Id));
            Assert.IsTrue(string.IsNullOrEmpty(tables[0].Attributes["style"]));

            System.Collections.Generic.List<IDomObject> cells = cq.Select("td").ToList();
            Assert.AreEqual(4, cells.Count);

            Assert.IsTrue(cells[0].Attributes["style"].Contains("border: 4px solid #FFA500;"));
            Assert.IsTrue(cells[1].Attributes["style"].Contains("border: 4px solid #FFA500;"));
            Assert.IsTrue(cells[2].Attributes["style"].Contains("border: 4px solid #FFA500;"));
            Assert.IsTrue(cells[3].Attributes["style"].Contains("border: 4px solid #FFA500;"));
        }

        [TestMethod]
        public void TableStyle_WithDifferentOuterAndInnerBorders_TableWithCellSpacing_AreCellsWithInnderBordersApplied()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));
            Border greenBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Green));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders(redBorder, redBorder, redBorder, redBorder, greenBorder, greenBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;
            table.TableCellSpacing = 40;

            TableCell cellOne = table.Rows.First().Cells.First();
            cellOne.Blocks.AddParagraph().Inlines.AddRun("1-1");

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("4px solid #ff0000", styleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            Assert.IsFalse(styleCellPart.Attributes.ContainsKey("border"));

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(1, tables.Count);
            Assert.IsTrue(tables[0].Attributes["class"].Contains(tableStyle.Id));
            Assert.IsTrue(tables[0].Attributes["style"].Contains("border-collapse: separate;"));
            Assert.IsTrue(tables[0].Attributes["style"].Contains("border-spacing: 40px;"));

            System.Collections.Generic.List<IDomObject> cells = cq.Select("td").ToList();
            Assert.AreEqual(4, cells.Count);
            foreach (IDomObject cell in cells)
            {
                cell.Attributes["style"].Contains("4px solid #008000");
            }
        }

        [TestMethod]
        public void TableStyle_WithDifferentOuterAndInnerBorders_TableWithCellSpacingAndCellsHasLocalBorders_AreCellBordersExportedLocaly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));
            Border greenBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Green));
            Border orangeBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Orange));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders(redBorder, redBorder, redBorder, redBorder, greenBorder, greenBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;
            table.TableCellSpacing = 40;

            TableCell cellOne = table.Rows.First().Cells.First();
            cellOne.Blocks.AddParagraph().Inlines.AddRun("1-1");
            cellOne.Borders = new TableCellBorders(orangeBorder);

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");
            cellTwo.Borders = new TableCellBorders(orangeBorder);

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");
            cellThree.Borders = new TableCellBorders(orangeBorder);

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");
            cellFour.Borders = new TableCellBorders(orangeBorder);

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("4px solid #ff0000", styleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            Assert.IsFalse(styleCellPart.Attributes.ContainsKey("border"));

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(1, tables.Count);
            Assert.IsTrue(tables[0].Attributes["class"].Contains(tableStyle.Id));
            Assert.IsTrue(tables[0].Attributes["style"].Contains("border-collapse: separate;"));
            Assert.IsTrue(tables[0].Attributes["style"].Contains("border-spacing: 40px;"));

            System.Collections.Generic.List<IDomObject> cells = cq.Select("td").ToList();
            Assert.AreEqual(4, cells.Count);
            foreach (IDomObject cell in cells)
            {
                cell.Attributes["style"].Contains("4px solid #FFA500");
            }
        }

        [TestMethod]
        public void NestedTables_OuterTableWithTableStyle_WithAllSameBorders_IsExportedCorrectlyForAllCells()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            tableStyle.TableProperties.Borders.LocalValue = new TableBorders(redBorder);

            document.StyleRepository.Add(tableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;

            TableCell cellOne = table.Rows.First().Cells.First();
            Table nestedTable = cellOne.Blocks.AddTable();

            TableRow firstRow = nestedTable.Rows.AddTableRow();
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-1");
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-2");

            TableRow secondRow = nestedTable.Rows.AddTableRow();
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-1");
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-2");

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("4px solid #ff0000", styleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            Assert.AreEqual("4px solid #ff0000", styleCellPart.Attributes["border"]);

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(2, tables.Count);
            Assert.IsTrue(tables[0].Attributes["class"].Contains(tableStyle.Id));
            Assert.IsTrue(string.IsNullOrEmpty(tables[0].Attributes["style"]));

            this.AssertTable_AllBorders(tables[0], tableStyle, parser);

            IDomObject nestedHtmlTable = tables[1];
            Assert.IsFalse(nestedHtmlTable.Attributes["class"].Contains(tableStyle.Id));

            System.Collections.Generic.List<IDomElement> rows = nestedHtmlTable.ChildElements.ToList()[0].ChildElements.ToList();

            for (int i = 0; i < rows.Count; i++)
            {
                System.Collections.Generic.List<IDomElement> cells = rows[i].ChildElements.ToList();
                Assert.AreEqual(2, cells.Count);
                foreach (IDomElement cell in cells)
                {
                    Assert.IsFalse(cell.Attributes["style"].Contains("border"));
                }
            }
        }

        [TestMethod]
        public void NestedTables_OuterTableWithTableStyle_WithInnerBorders_IsExportedCorrectlyForAllCells()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders();
            tableBorders = new TableBorders(tableBorders, insideHorizontalBorder: redBorder, insideVerticalBorder: redBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;

            TableCell cellOne = table.Rows.First().Cells.First();
            Table nestedTable = cellOne.Blocks.AddTable();

            TableRow firstRow = nestedTable.Rows.AddTableRow();
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-1");
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-2");

            TableRow secondRow = nestedTable.Rows.AddTableRow();
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-1");
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-2");

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);
            Assert.IsFalse(styleCellPart.Attributes.ContainsKey("border"));

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(2, tables.Count);
            Assert.IsTrue(tables[0].Attributes["class"].Contains(tableStyle.Id));
            Assert.IsTrue(string.IsNullOrEmpty(tables[0].Attributes["style"]));

            System.Collections.Generic.List<IDomElement> rows = tables[0].ChildElements.ToList()[0].ChildElements.ToList();

            System.Collections.Generic.List<IDomElement> tableCells = rows[0].ChildElements.ToList();
            Assert.IsTrue(tableCells[0].Attributes["style"].Contains("border-right: 4px solid #FF0000;border-bottom: 4px solid #FF0000;"));
            Assert.IsTrue(tableCells[1].Attributes["style"].Contains("border-left: 4px solid #FF0000;border-bottom: 4px solid #FF0000;"));

            tableCells = rows[1].ChildElements.ToList();
            Assert.IsTrue(tableCells[0].Attributes["style"].Contains("border-top: 4px solid #FF0000;border-right: 4px solid #FF0000;"));
            Assert.IsTrue(tableCells[1].Attributes["style"].Contains("border-left: 4px solid #FF0000;border-top: 4px solid #FF0000;"));

            IDomObject nestedHtmlTable = tables[1];
            Assert.IsFalse(nestedHtmlTable.Attributes["class"].Contains(tableStyle.Id));

            rows = nestedHtmlTable.ChildElements.ToList()[0].ChildElements.ToList();
            for (int i = 0; i < rows.Count; i++)
            {
                System.Collections.Generic.List<IDomElement> cells = rows[i].ChildElements.ToList();
                Assert.AreEqual(2, cells.Count);
                foreach (IDomElement cell in cells)
                {
                    Assert.IsFalse(cell.Attributes["style"].Contains("border"));
                }
            }
        }

        [TestMethod]
        public void NestedTables_EachWithOwnStyle_AllBorders_AreStylesExportedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));
            Border orangeBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Orange));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            tableStyle.TableProperties.Borders.LocalValue = new TableBorders(redBorder);

            document.StyleRepository.Add(tableStyle);

            Style nestedTableStyle = new Style("NestedTableStyle", StyleType.Table);
            nestedTableStyle.TableProperties.Borders.LocalValue = new TableBorders(orangeBorder);

            document.StyleRepository.Add(nestedTableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;

            TableCell cellOne = table.Rows.First().Cells.First();
            Table nestedTable = cellOne.Blocks.AddTable();
            nestedTable.StyleId = nestedTableStyle.Id;

            TableRow firstRow = nestedTable.Rows.AddTableRow();
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-1");
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-2");

            TableRow secondRow = nestedTable.Rows.AddTableRow();
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-1");
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-2");

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("4px solid #ff0000", styleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            Assert.AreEqual("4px solid #ff0000", styleCellPart.Attributes["border"]);

            StyleClass nestedTableStyleTablePart = parser.Styles["table." + nestedTableStyle.Id];
            StyleClass nestedTableStyleCellPart = parser.Styles["td." + nestedTableStyle.Id];

            Assert.AreEqual("4px solid #ffa500", nestedTableStyleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", nestedTableStyleTablePart.Attributes["border-collapse"]);

            Assert.AreEqual("4px solid #ffa500", nestedTableStyleCellPart.Attributes["border"]);

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(2, tables.Count);
            Assert.IsTrue(tables[0].Attributes["class"].Contains(tableStyle.Id));
            Assert.IsTrue(string.IsNullOrEmpty(tables[0].Attributes["style"]));

            System.Collections.Generic.List<IDomElement> rows = tables[0].ChildElements.ToList()[0].ChildElements.ToList();

            for (int i = 0; i < rows.Count; i++)
            {
                System.Collections.Generic.List<IDomElement> cells = rows[i].ChildElements.ToList();
                Assert.AreEqual(2, cells.Count);
                foreach (IDomElement cell in cells)
                {
                    Assert.IsTrue(!cell.Attributes["style"].Contains("border"));
                }
            }

            IDomObject nestedHtmlTable = tables[1];
            Assert.IsTrue(nestedHtmlTable.Attributes["class"].Contains(nestedTableStyle.Id));

            rows = nestedHtmlTable.ChildElements.ToList()[0].ChildElements.ToList();

            for (int i = 0; i < rows.Count; i++)
            {
                System.Collections.Generic.List<IDomElement> cells = rows[i].ChildElements.ToList();
                Assert.AreEqual(2, cells.Count);
                foreach (IDomElement cell in cells)
                {
                    Assert.IsFalse(cell.Attributes["style"].Contains("border"));
                }
            }
        }

        [TestMethod]
        public void NestedTables_EachWithOwnStyle_InnerBorders_AreStylesExportedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));
            Border orangeBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Orange));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders();
            tableBorders = new TableBorders(tableBorders, insideHorizontalBorder: redBorder, insideVerticalBorder: redBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            Style nestedTableStyle = new Style("NestedTableStyle", StyleType.Table);
            tableBorders = new TableBorders();
            tableBorders = new TableBorders(tableBorders, insideHorizontalBorder: orangeBorder, insideVerticalBorder: orangeBorder);
            nestedTableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(nestedTableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;

            TableCell cellOne = table.Rows.First().Cells.First();
            Table nestedTable = cellOne.Blocks.AddTable();
            nestedTable.StyleId = nestedTableStyle.Id;

            TableRow firstRow = nestedTable.Rows.AddTableRow();
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-1");
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-2");

            TableRow secondRow = nestedTable.Rows.AddTableRow();
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-1");
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-2");

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);
            Assert.IsFalse(styleCellPart.Attributes.ContainsKey("border"));

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(2, tables.Count);
            Assert.IsTrue(tables[0].Attributes["class"].Contains(tableStyle.Id));
            Assert.IsTrue(string.IsNullOrEmpty(tables[0].Attributes["style"]));

            this.AssertTable_InnerBorders(tables[0], tableStyle, table.HasCellSpacing);
            this.AssertTable_InnerBorders(tables[1], nestedTableStyle, nestedTable.HasCellSpacing);
        }

        [TestMethod]
        public void NestedTablesWithCellSpacing_EachWithOwnStyle_AllBorders_AreStylesExportedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));
            Border orangeBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Orange));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            tableStyle.TableProperties.Borders.LocalValue = new TableBorders(redBorder);

            document.StyleRepository.Add(tableStyle);

            Style nestedTableStyle = new Style("NestedTableStyle", StyleType.Table);
            nestedTableStyle.TableProperties.Borders.LocalValue = new TableBorders(orangeBorder);

            document.StyleRepository.Add(nestedTableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;
            table.TableCellSpacing = 40;

            TableCell cellOne = table.Rows.First().Cells.First();
            Table nestedTable = cellOne.Blocks.AddTable();
            nestedTable.StyleId = nestedTableStyle.Id;
            nestedTable.TableCellSpacing = 40;

            TableRow firstRow = nestedTable.Rows.AddTableRow();
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-1");
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-2");

            TableRow secondRow = nestedTable.Rows.AddTableRow();
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-1");
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-2");

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("4px solid #ff0000", styleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            Assert.AreEqual("4px solid #ff0000", styleCellPart.Attributes["border"]);

            StyleClass nestedTableStyleTablePart = parser.Styles["table." + nestedTableStyle.Id];
            StyleClass nestedTableStyleCellPart = parser.Styles["td." + nestedTableStyle.Id];

            Assert.AreEqual("4px solid #ffa500", nestedTableStyleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", nestedTableStyleTablePart.Attributes["border-collapse"]);

            Assert.AreEqual("4px solid #ffa500", nestedTableStyleCellPart.Attributes["border"]);

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(2, tables.Count);
            Assert.IsTrue(tables[0].Attributes["class"].Contains(tableStyle.Id));
            Assert.IsTrue(tables[0].Attributes["style"].Contains("border-collapse: separate;"));
            Assert.IsTrue(tables[0].Attributes["style"].Contains("border-spacing: 40px;"));

            this.AssertTable_AllBorders(tables[0], tableStyle, parser);

            IDomObject nestedHtmlTable = tables[1];
            Assert.IsTrue(nestedHtmlTable.Attributes["class"].Contains(nestedTableStyle.Id));
            Assert.IsTrue(nestedHtmlTable.Attributes["style"].Contains("border-collapse: separate;"));
            Assert.IsTrue(nestedHtmlTable.Attributes["style"].Contains("border-spacing: 40px;"));

            this.AssertTable_AllBorders(nestedHtmlTable, nestedTableStyle, parser);
        }

        [TestMethod]
        public void NestedTablesWithCellSpacing_EachWithOwnStyle_InnerBorders_AreStylesExportedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));
            Border orangeBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Orange));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders();
            tableBorders = new TableBorders(tableBorders, insideHorizontalBorder: redBorder, insideVerticalBorder: redBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            Style nestedTableStyle = new Style("NestedTableStyle", StyleType.Table);
            tableBorders = new TableBorders();
            tableBorders = new TableBorders(tableBorders, insideHorizontalBorder: orangeBorder, insideVerticalBorder: orangeBorder);
            nestedTableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(nestedTableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;
            table.TableCellSpacing = 40;

            TableCell cellOne = table.Rows.First().Cells.First();
            Table nestedTable = cellOne.Blocks.AddTable();
            nestedTable.StyleId = nestedTableStyle.Id;
            nestedTable.TableCellSpacing = 40;

            TableRow firstRow = nestedTable.Rows.AddTableRow();
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-1");
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-2");

            TableRow secondRow = nestedTable.Rows.AddTableRow();
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-1");
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-2");

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            StyleClass nestedTableStyleTablePart = parser.Styles["table." + nestedTableStyle.Id];
            StyleClass nestedTableStyleCellPart = parser.Styles["td." + nestedTableStyle.Id];

            Assert.AreEqual("collapse", nestedTableStyleTablePart.Attributes["border-collapse"]);

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(2, tables.Count);
            this.AssertTable_InnerBorders(tables[0], tableStyle, table.HasCellSpacing);
            this.AssertTable_InnerBorders(tables[1], nestedTableStyle, nestedTable.HasCellSpacing);
        }

        [TestMethod]
        public void NestedTablesWithCellSpacing_EachWithOwnStyle_OuterBorders_AreStylesExportedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));
            Border orangeBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Orange));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders(redBorder, redBorder, redBorder, redBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            Style nestedTableStyle = new Style("NestedTableStyle", StyleType.Table);
            tableBorders = new TableBorders(orangeBorder, orangeBorder, orangeBorder, orangeBorder);
            nestedTableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(nestedTableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;
            table.TableCellSpacing = 40;

            TableCell cellOne = table.Rows.First().Cells.First();
            Table nestedTable = cellOne.Blocks.AddTable();
            nestedTable.StyleId = nestedTableStyle.Id;
            nestedTable.TableCellSpacing = 40;

            TableRow firstRow = nestedTable.Rows.AddTableRow();
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-1");
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-2");

            TableRow secondRow = nestedTable.Rows.AddTableRow();
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-1");
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-2");

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("4px solid #ff0000", styleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            Assert.IsFalse(styleCellPart.Attributes.ContainsKey("border"));

            StyleClass nestedTableStyleTablePart = parser.Styles["table." + nestedTableStyle.Id];
            StyleClass nestedTableStyleCellPart = parser.Styles["td." + nestedTableStyle.Id];

            Assert.AreEqual("4px solid #ffa500", nestedTableStyleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", nestedTableStyleTablePart.Attributes["border-collapse"]);

            Assert.IsFalse(nestedTableStyleCellPart.Attributes.ContainsKey("border"));

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(2, tables.Count);

            this.AssertTable_OuterBorders(tables[0], tableStyle, table.HasCellSpacing);
            this.AssertTable_OuterBorders(tables[1], nestedTableStyle, nestedTable.HasCellSpacing);
        }

        [TestMethod]
        public void NestedTables_TheOuterTableIsWithCellSpacing_EachWithOwnStyle_OuterBorderForOuterTable_OuterBordersForNestedTable_AreStylesExportedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));
            Border orangeBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Orange));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders(redBorder, redBorder, redBorder, redBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            Style nestedTableStyle = new Style("NestedTableStyle", StyleType.Table);
            tableBorders = new TableBorders(orangeBorder, orangeBorder, orangeBorder, orangeBorder);
            nestedTableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(nestedTableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;
            table.TableCellSpacing = 40;

            TableCell cellOne = table.Rows.First().Cells.First();
            Table nestedTable = cellOne.Blocks.AddTable();
            nestedTable.StyleId = nestedTableStyle.Id;
            //nestedTable.TableCellSpacing = 40;

            TableRow firstRow = nestedTable.Rows.AddTableRow();
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-1");
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-2");

            TableRow secondRow = nestedTable.Rows.AddTableRow();
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-1");
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-2");

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("4px solid #ff0000", styleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            Assert.IsFalse(styleCellPart.Attributes.ContainsKey("border"));

            StyleClass nestedTableStyleTablePart = parser.Styles["table." + nestedTableStyle.Id];
            StyleClass nestedTableStyleCellPart = parser.Styles["td." + nestedTableStyle.Id];

            Assert.AreEqual("4px solid #ffa500", nestedTableStyleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", nestedTableStyleTablePart.Attributes["border-collapse"]);

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(2, tables.Count);
            Assert.IsTrue(tables[0].Attributes["class"].Contains(tableStyle.Id));

            this.AssertTable_OuterBorders(tables[0], tableStyle, table.HasCellSpacing);
            this.AssertTable_OuterBorders(tables[1], nestedTableStyle, nestedTable.HasCellSpacing);
        }

        [TestMethod]
        public void NestedTables_TheOuterTableIsWithCellSpacing_EachWithOwnStyle_OuterBorderForOuterTable_InnerBordersForNestedTable_AreStylesExportedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));
            Border orangeBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Orange));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders(redBorder, redBorder, redBorder, redBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            Style nestedTableStyle = new Style("NestedTableStyle", StyleType.Table);
            tableBorders = new TableBorders();
            tableBorders = new TableBorders(tableBorders, insideHorizontalBorder: orangeBorder, insideVerticalBorder: orangeBorder);
            nestedTableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(nestedTableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;
            table.TableCellSpacing = 40;

            TableCell cellOne = table.Rows.First().Cells.First();
            Table nestedTable = cellOne.Blocks.AddTable();
            nestedTable.StyleId = nestedTableStyle.Id;
            //nestedTable.TableCellSpacing = 40;

            TableRow firstRow = nestedTable.Rows.AddTableRow();
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-1");
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-2");

            TableRow secondRow = nestedTable.Rows.AddTableRow();
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-1");
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-2");

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("4px solid #ff0000", styleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            Assert.IsFalse(styleCellPart.Attributes.ContainsKey("border"));

            StyleClass nestedTableStyleTablePart = parser.Styles["table." + nestedTableStyle.Id];
            StyleClass nestedTableStyleCellPart = parser.Styles["td." + nestedTableStyle.Id];

            Assert.AreEqual("collapse", nestedTableStyleTablePart.Attributes["border-collapse"]);

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(2, tables.Count);
            Assert.IsTrue(tables[0].Attributes["class"].Contains(tableStyle.Id));

            this.AssertTable_OuterBorders(tables[0], tableStyle, table.HasCellSpacing);
            this.AssertTable_InnerBorders(tables[1], nestedTableStyle, nestedTable.HasCellSpacing);
        }

        [TestMethod]
        public void NestedTables_TheOuterTableIsWithCellSpacing_EachWithOwnStyle_InnerBorderForOuterTable_InnerBordersForNestedTable_AreStylesExportedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));
            Border orangeBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Orange));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders();
            tableBorders = new TableBorders(tableBorders, insideHorizontalBorder: redBorder, insideVerticalBorder: redBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            Style nestedTableStyle = new Style("NestedTableStyle", StyleType.Table);
            tableBorders = new TableBorders();
            tableBorders = new TableBorders(tableBorders, insideHorizontalBorder: orangeBorder, insideVerticalBorder: orangeBorder);
            nestedTableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(nestedTableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;
            table.TableCellSpacing = 40;

            TableCell cellOne = table.Rows.First().Cells.First();
            Table nestedTable = cellOne.Blocks.AddTable();
            nestedTable.StyleId = nestedTableStyle.Id;
            //nestedTable.TableCellSpacing = 40;

            TableRow firstRow = nestedTable.Rows.AddTableRow();
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-1");
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-2");

            TableRow secondRow = nestedTable.Rows.AddTableRow();
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-1");
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-2");

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            Assert.IsFalse(styleCellPart.Attributes.ContainsKey("border"));

            StyleClass nestedTableStyleTablePart = parser.Styles["table." + nestedTableStyle.Id];
            StyleClass nestedTableStyleCellPart = parser.Styles["td." + nestedTableStyle.Id];

            Assert.AreEqual("collapse", nestedTableStyleTablePart.Attributes["border-collapse"]);

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(2, tables.Count);
            Assert.IsTrue(tables[0].Attributes["class"].Contains(tableStyle.Id));

            this.AssertTable_InnerBorders(tables[0], tableStyle, table.HasCellSpacing);
            this.AssertTable_InnerBorders(tables[1], nestedTableStyle, nestedTable.HasCellSpacing);
        }

        [TestMethod]
        public void NestedTables_TheOuterTableIsWithCellSpacing_EachWithOwnStyle_InnerBorderForOuterTable_OuterBordersForNestedTable_AreStylesExportedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));
            Border orangeBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Orange));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders();
            tableBorders = new TableBorders(tableBorders, insideHorizontalBorder: redBorder, insideVerticalBorder: redBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            Style nestedTableStyle = new Style("NestedTableStyle", StyleType.Table);
            tableBorders = new TableBorders();
            tableBorders = new TableBorders(orangeBorder, orangeBorder, orangeBorder, orangeBorder);
            nestedTableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(nestedTableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;
            table.TableCellSpacing = 40;

            TableCell cellOne = table.Rows.First().Cells.First();
            Table nestedTable = cellOne.Blocks.AddTable();
            nestedTable.StyleId = nestedTableStyle.Id;
            //nestedTable.TableCellSpacing = 40;

            TableRow firstRow = nestedTable.Rows.AddTableRow();
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-1");
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-2");

            TableRow secondRow = nestedTable.Rows.AddTableRow();
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-1");
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-2");

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            Assert.IsFalse(styleCellPart.Attributes.ContainsKey("border"));

            StyleClass nestedTableStyleTablePart = parser.Styles["table." + nestedTableStyle.Id];
            StyleClass nestedTableStyleCellPart = parser.Styles["td." + nestedTableStyle.Id];

            Assert.AreEqual("collapse", nestedTableStyleTablePart.Attributes["border-collapse"]);

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(2, tables.Count);
            Assert.IsTrue(tables[0].Attributes["class"].Contains(tableStyle.Id));

            this.AssertTable_InnerBorders(tables[0], tableStyle, table.HasCellSpacing);
            this.AssertTable_OuterBorders(tables[1], nestedTableStyle, nestedTable.HasCellSpacing);
        }

        [TestMethod]
        public void NestedTables_TheNestedTableIsWithCellSpacing_EachWithOwnStyle_OuterBorderForOuterTable_OuterBordersForNestedTable_AreStylesExportedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));
            Border orangeBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Orange));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders(redBorder, redBorder, redBorder, redBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            Style nestedTableStyle = new Style("NestedTableStyle", StyleType.Table);
            tableBorders = new TableBorders(orangeBorder, orangeBorder, orangeBorder, orangeBorder);
            nestedTableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(nestedTableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;
            //table.TableCellSpacing = 40;

            TableCell cellOne = table.Rows.First().Cells.First();
            Table nestedTable = cellOne.Blocks.AddTable();
            nestedTable.StyleId = nestedTableStyle.Id;
            nestedTable.TableCellSpacing = 40;

            TableRow firstRow = nestedTable.Rows.AddTableRow();
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-1");
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-2");

            TableRow secondRow = nestedTable.Rows.AddTableRow();
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-1");
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-2");

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("4px solid #ff0000", styleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            Assert.IsFalse(styleCellPart.Attributes.ContainsKey("border"));

            StyleClass nestedTableStyleTablePart = parser.Styles["table." + nestedTableStyle.Id];
            StyleClass nestedTableStyleCellPart = parser.Styles["td." + nestedTableStyle.Id];

            Assert.AreEqual("4px solid #ffa500", nestedTableStyleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", nestedTableStyleTablePart.Attributes["border-collapse"]);

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(2, tables.Count);
            Assert.IsTrue(tables[0].Attributes["class"].Contains(tableStyle.Id));

            this.AssertTable_OuterBorders(tables[0], tableStyle, table.HasCellSpacing);
            this.AssertTable_OuterBorders(tables[1], nestedTableStyle, nestedTable.HasCellSpacing);
        }

        [TestMethod]
        public void NestedTables_TheNestedTableIsWithCellSpacing_EachWithOwnStyle_OuterBorderForOuterTable_InnerBordersForNestedTable_AreStylesExportedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));
            Border orangeBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Orange));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders(redBorder, redBorder, redBorder, redBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            Style nestedTableStyle = new Style("NestedTableStyle", StyleType.Table);
            tableBorders = new TableBorders();
            tableBorders = new TableBorders(tableBorders, insideHorizontalBorder: orangeBorder, insideVerticalBorder: orangeBorder);
            nestedTableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(nestedTableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;
            //table.TableCellSpacing = 40;

            TableCell cellOne = table.Rows.First().Cells.First();
            Table nestedTable = cellOne.Blocks.AddTable();
            nestedTable.StyleId = nestedTableStyle.Id;
            nestedTable.TableCellSpacing = 40;

            TableRow firstRow = nestedTable.Rows.AddTableRow();
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-1");
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-2");

            TableRow secondRow = nestedTable.Rows.AddTableRow();
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-1");
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-2");

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("4px solid #ff0000", styleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            Assert.IsFalse(styleCellPart.Attributes.ContainsKey("border"));

            StyleClass nestedTableStyleTablePart = parser.Styles["table." + nestedTableStyle.Id];
            StyleClass nestedTableStyleCellPart = parser.Styles["td." + nestedTableStyle.Id];

            Assert.AreEqual("collapse", nestedTableStyleTablePart.Attributes["border-collapse"]);

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(2, tables.Count);
            Assert.IsTrue(tables[0].Attributes["class"].Contains(tableStyle.Id));

            this.AssertTable_OuterBorders(tables[0], tableStyle, table.HasCellSpacing);
            this.AssertTable_InnerBorders(tables[1], nestedTableStyle, nestedTable.HasCellSpacing);
        }

        [TestMethod]
        public void NestedTables_TheNestedTableIsWithCellSpacing_EachWithOwnStyle_InnerBorderForOuterTable_InnerBordersForNestedTable_AreStylesExportedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));
            Border orangeBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Orange));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders();
            tableBorders = new TableBorders(tableBorders, insideHorizontalBorder: redBorder, insideVerticalBorder: redBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            Style nestedTableStyle = new Style("NestedTableStyle", StyleType.Table);
            tableBorders = new TableBorders();
            tableBorders = new TableBorders(tableBorders, insideHorizontalBorder: orangeBorder, insideVerticalBorder: orangeBorder);
            nestedTableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(nestedTableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;
            table.TableCellSpacing = 40;

            TableCell cellOne = table.Rows.First().Cells.First();
            Table nestedTable = cellOne.Blocks.AddTable();
            nestedTable.StyleId = nestedTableStyle.Id;
            //nestedTable.TableCellSpacing = 40;

            TableRow firstRow = nestedTable.Rows.AddTableRow();
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-1");
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-2");

            TableRow secondRow = nestedTable.Rows.AddTableRow();
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-1");
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-2");

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            Assert.IsFalse(styleCellPart.Attributes.ContainsKey("border"));

            StyleClass nestedTableStyleTablePart = parser.Styles["table." + nestedTableStyle.Id];
            StyleClass nestedTableStyleCellPart = parser.Styles["td." + nestedTableStyle.Id];

            Assert.AreEqual("collapse", nestedTableStyleTablePart.Attributes["border-collapse"]);

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(2, tables.Count);
            Assert.IsTrue(tables[0].Attributes["class"].Contains(tableStyle.Id));

            this.AssertTable_InnerBorders(tables[0], tableStyle, table.HasCellSpacing);
            this.AssertTable_InnerBorders(tables[1], nestedTableStyle, nestedTable.HasCellSpacing);
        }

        [TestMethod]
        public void NestedTables_TheNestedTableIsWithCellSpacing_EachWithOwnStyle_InnerBorderForOuterTable_OuterBordersForNestedTable_AreStylesExportedCorrectly()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();

            Border redBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Red));
            Border orangeBorder = new Border(4, BorderStyle.Single, new ThemableColor(Colors.Orange));

            Style tableStyle = new Style("TableStyle", StyleType.Table);
            TableBorders tableBorders = new TableBorders();
            tableBorders = new TableBorders(tableBorders, insideHorizontalBorder: redBorder, insideVerticalBorder: redBorder);
            tableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(tableStyle);

            Style nestedTableStyle = new Style("NestedTableStyle", StyleType.Table);
            tableBorders = new TableBorders();
            tableBorders = new TableBorders(orangeBorder, orangeBorder, orangeBorder, orangeBorder);
            nestedTableStyle.TableProperties.Borders.LocalValue = tableBorders;

            document.StyleRepository.Add(nestedTableStyle);

            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(document);
            Table table = editor.InsertTable(2, 2);
            table.StyleId = tableStyle.Id;
            //table.TableCellSpacing = 40;

            TableCell cellOne = table.Rows.First().Cells.First();
            Table nestedTable = cellOne.Blocks.AddTable();
            nestedTable.StyleId = nestedTableStyle.Id;
            nestedTable.TableCellSpacing = 40;

            TableRow firstRow = nestedTable.Rows.AddTableRow();
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-1");
            firstRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 1-2");

            TableRow secondRow = nestedTable.Rows.AddTableRow();
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-1");
            secondRow.Cells.AddTableCell().Blocks.AddParagraph().Inlines.AddRun("nested 2-2");

            TableCell cellTwo = table.Rows.First().Cells.Last();
            cellTwo.Blocks.AddParagraph().Inlines.AddRun("1-2");

            TableCell cellThree = table.Rows.Last().Cells.First();
            cellThree.Blocks.AddParagraph().Inlines.AddRun("2-1");

            TableCell cellFour = table.Rows.Last().Cells.Last();
            cellFour.Blocks.AddParagraph().Inlines.AddRun("2-2");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(document);

            // Assert
            IDomObject style = cq.Select("style").ToList()[0];
            CssParser parser = new CssParser();
            parser.AddStyleSheet(style.InnerHTML);

            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);

            Assert.IsFalse(styleCellPart.Attributes.ContainsKey("border"));

            StyleClass nestedTableStyleTablePart = parser.Styles["table." + nestedTableStyle.Id];
            StyleClass nestedTableStyleCellPart = parser.Styles["td." + nestedTableStyle.Id];

            Assert.AreEqual("collapse", nestedTableStyleTablePart.Attributes["border-collapse"]);

            // Assert
            System.Collections.Generic.List<IDomObject> tables = cq.Select("table").ToList();
            Assert.AreEqual(2, tables.Count);
            Assert.IsTrue(tables[0].Attributes["class"].Contains(tableStyle.Id));

            this.AssertTable_InnerBorders(tables[0], tableStyle, table.HasCellSpacing);
            this.AssertTable_OuterBorders(tables[1], nestedTableStyle, nestedTable.HasCellSpacing);
        }

        [TestMethod]
        public void TableBorders_WhenBorderAndBordercolorAttributesAreDefined_ValuesAreExportedInStyleProperty_ImportExport()
        {
            // Arrange
            HtmlFormatProvider provider = new HtmlFormatProvider();
            RadFlowDocument imported = provider.Import("<html><body><table border=\"10\" bordercolor=\"#FF0000\"><tr><td><p><span>test</span></p></td></tr></table></body></html>");

            // Act
            CQ cq = ExportHelper.ExportDocumentToCq(imported);

            // Assert
            IDomObject table = cq.Select("table").FirstOrDefault();

            Assert.IsNull(table.Attributes["bordercolor"]);
            Assert.IsNull(table.Attributes["border"]);

            string expectedStylePropertyValue = "border: 10px solid #FF0000;";
            string actualStylePropertyValue = table.Attributes["style"];
            Assert.AreEqual(expectedStylePropertyValue, actualStylePropertyValue.ToString());
        }

        private void AssertTable_OuterBorders(IDomObject table, Style tableStyle, bool hasCellSpacing)
        {
            Assert.IsTrue(table.Attributes["class"].Contains(tableStyle.Id));

            System.Collections.Generic.List<IDomElement> rows = table.ChildElements.ToList()[0].ChildElements.ToList();

            if (!hasCellSpacing)
            {
                Assert.IsFalse(table.Attributes["style"].Contains("border-spacing"));

                // Those borders are duplicating the table style. Think of refactoring the TableCellBordersConvert to exclude them.
                System.Collections.Generic.List<IDomElement> cells = rows[0].ChildElements.ToList();
                TableBorders borders = tableStyle.TableProperties.Borders.LocalValue;

                string cell0Style = cells[0].Attributes["style"];
                Assert.IsTrue(cell0Style.Contains("border-left: " + borders.Left.Thickness + "px solid " + this.ColorToString(borders.Left.Color.LocalValue) + ";"));
                Assert.IsTrue(cell0Style.Contains("border-top: " + borders.Top.Thickness + "px solid " + this.ColorToString(borders.Top.Color.LocalValue) + ";"));

                string cell1Style = cells[1].Attributes["style"];
                Assert.IsTrue(cell1Style.Contains("border-top: " + borders.Top.Thickness + "px solid " + this.ColorToString(borders.Top.Color.LocalValue) + ";"));
                Assert.IsTrue(cell1Style.Contains("border-right: " + borders.Right.Thickness + "px solid " + this.ColorToString(borders.Right.Color.LocalValue) + ";"));

                cells = rows[1].ChildElements.ToList();

                cell0Style = cells[0].Attributes["style"];
                Assert.IsTrue(cell0Style.Contains("border-left: " + borders.Left.Thickness + "px solid " + this.ColorToString(borders.Left.Color.LocalValue) + ";"));
                Assert.IsTrue(cell0Style.Contains("border-bottom: " + borders.Bottom.Thickness + "px solid " + this.ColorToString(borders.Bottom.Color.LocalValue) + ";"));

                cell1Style = cells[1].Attributes["style"];
                Assert.IsTrue(cell1Style.Contains("border-right: " + borders.Right.Thickness + "px solid " + this.ColorToString(borders.Right.Color.LocalValue) + ";"));
                Assert.IsTrue(cell1Style.Contains("border-bottom: " + borders.Bottom.Thickness + "px solid " + this.ColorToString(borders.Bottom.Color.LocalValue) + ";"));
            }
            else
            {
                Assert.IsTrue(table.Attributes["style"].Contains("border-collapse: separate;"));
                Assert.IsTrue(table.Attributes["style"].Contains("border-spacing"));

                System.Collections.Generic.List<IDomElement> cells = rows[0].ChildElements.ToList();
                Assert.IsFalse(cells[0].Attributes["style"].Contains("border"));
                Assert.IsFalse(cells[1].Attributes["style"].Contains("border"));

                cells = rows[1].ChildElements.ToList();
                Assert.IsFalse(cells[0].Attributes["style"].Contains("border"));
                Assert.IsFalse(cells[1].Attributes["style"].Contains("border"));
            }
        }

        private void AssertTable_InnerBorders(IDomObject table, Style tableStyle, bool hasCellSpacing)
        {
            Assert.IsTrue(table.Attributes["class"].Contains(tableStyle.Id));

            System.Collections.Generic.List<IDomElement> rows = table.ChildElements.ToList()[0].ChildElements.ToList();
            if (!hasCellSpacing)
            {
                Assert.IsFalse(table.Attributes["style"].Contains("border-spacing"));

                System.Collections.Generic.List<IDomElement> cells = rows[0].ChildElements.ToList();
                TableBorders borders = tableStyle.TableProperties.Borders.LocalValue;

                string cell0Style = cells[0].Attributes["style"];
                Assert.IsTrue(cell0Style.Contains("border-right: " + borders.InsideVertical.Thickness + "px solid " + this.ColorToString(borders.InsideVertical.Color.LocalValue) + ";"));
                Assert.IsTrue(cell0Style.Contains("border-bottom: " + borders.InsideHorizontal.Thickness + "px solid " + this.ColorToString(borders.InsideHorizontal.Color.LocalValue) + ";"));

                string cell1Style = cells[1].Attributes["style"];
                Assert.IsTrue(cell1Style.Contains("border-left: " + borders.InsideVertical.Thickness + "px solid " + this.ColorToString(borders.InsideVertical.Color.LocalValue) + ";"));
                Assert.IsTrue(cell1Style.Contains("border-bottom: " + borders.InsideHorizontal.Thickness + "px solid " + this.ColorToString(borders.InsideHorizontal.Color.LocalValue) + ";"));

                cells = rows[1].ChildElements.ToList();

                cell0Style = cells[0].Attributes["style"];
                Assert.IsTrue(cell0Style.Contains("border-top: " + borders.InsideHorizontal.Thickness + "px solid " + this.ColorToString(borders.InsideHorizontal.Color.LocalValue) + ";"));
                Assert.IsTrue(cell0Style.Contains("border-right: " + borders.InsideVertical.Thickness + "px solid " + this.ColorToString(borders.InsideVertical.Color.LocalValue) + ";"));

                cell1Style = cells[1].Attributes["style"];
                Assert.IsTrue(cell1Style.Contains("border-left: " + borders.InsideVertical.Thickness + "px solid " + this.ColorToString(borders.InsideVertical.Color.LocalValue) + ";"));
                Assert.IsTrue(cell1Style.Contains("border-top: " + borders.InsideHorizontal.Thickness + "px solid " + this.ColorToString(borders.InsideHorizontal.Color.LocalValue) + ";"));
            }
            else
            {
                Assert.IsTrue(table.Attributes["style"].Contains("border-collapse: separate;"));
                Assert.IsTrue(table.Attributes["style"].Contains("border-spacing"));
            }
        }

        private void AssertTable_AllBorders(IDomObject table, Style tableStyle, CssParser parser)
        {
            StyleClass styleTablePart = parser.Styles["table." + tableStyle.Id];
            StyleClass styleCellPart = parser.Styles["td." + tableStyle.Id];

            TableBorders borders = tableStyle.TableProperties.Borders.LocalValue;

            Assert.AreEqual("" + borders.Left.Thickness + "px solid " + this.ColorToString(borders.Left.Color.LocalValue).ToLowerInvariant() + "", styleTablePart.Attributes["border"]);
            Assert.AreEqual("collapse", styleTablePart.Attributes["border-collapse"]);


            Assert.AreEqual("" + borders.Left.Thickness + "px solid " + this.ColorToString(borders.Left.Color.LocalValue).ToLowerInvariant() + "", styleCellPart.Attributes["border"]);
            //Assert.AreEqual("4px solid #ff0000", styleCellPart.Attributes["border"]);

            Assert.IsTrue(table.Attributes["class"].Contains(tableStyle.Id));

            System.Collections.Generic.List<IDomElement> rows = table.ChildElements.ToList()[0].ChildElements.ToList();
            Assert.IsTrue(table.Attributes["class"].Contains(tableStyle.Id));

            for (int i = 0; i < rows.Count; i++)
            {
                System.Collections.Generic.List<IDomElement> cells = rows[i].ChildElements.ToList();
                Assert.AreEqual(2, cells.Count);
                foreach (IDomElement cell in cells)
                {
                    Assert.IsFalse(cell.Attributes["style"].Contains("border"));
                    Assert.IsTrue(cell.Attributes["class"].Contains(tableStyle.Id));
                }
            }
        }

        private string ColorToString(Color color)
        {
            string result = "#" + color.R.ToString("X2") + color.G.ToString("X2") + color.B.ToString("X2");
            return result;
        }
    }
}
