﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Linq;
using Telerik.Windows.Documents.Flow.FormatProviders.Html;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Shapes;
using Telerik.Windows.Documents.Media;

namespace Telerik.Windows.Documents.Flow.InternalTests.Integration.Html.Import
{
    [TestClass]
    public class ImagesTests
    {
        private readonly string Base64String = "iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO9TXL0Y4OHwAAAABJRU5ErkJggg==";
        private readonly double OriginalImageWidth = 5;
        private readonly double OriginalImageHeight = 5;

        protected RadFlowDocument ImportHtmlFromString(string html)
        {
            return new HtmlFormatProvider().Import(html);
        }

        [TestMethod]
        public void Import_WhenImageHasWidth()
        {
            double imageWidth = 6;

            RadFlowDocument expected = new RadFlowDocument();
            ImageInline image = expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddImageInline();
            image.Image.Width = imageWidth;
            image.Image.Height = OriginalImageHeight;
            image.Image.ImageSource = new ImageSource(Convert.FromBase64String(Base64String), "png");

            string imageWidthAttribute = string.Format("width=\"{0}\"", imageWidth);
            RadFlowDocument imported = this.ImportHtmlFromString(string.Format("<img src='data:image/png;base64,{0}' {1}/>", Base64String, imageWidthAttribute));

            Paragraph paragraph = imported.Sections.First().Blocks.First() as Paragraph;
            ImageInline actualImage = paragraph.Inlines.First() as ImageInline;
            actualImage.Image.LockAspectRatio = false;
            actualImage.Image.RecalculateSize();

            Assert.IsNotNull(actualImage);
            Assert.AreEqual(actualImage.Image.Width, image.Image.Width);
            Assert.AreEqual(actualImage.Image.Height, image.Image.Height);
        }

        [TestMethod]
        public void ImportImage_WhenImageHasEmptyWidth()
        {
            RadFlowDocument expected = new RadFlowDocument();
            ImageInline image = expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddImageInline();
            image.Image.Width = OriginalImageWidth;
            image.Image.Height = OriginalImageHeight;
            image.Image.ImageSource = new ImageSource(Convert.FromBase64String(Base64String), "png");

            RadFlowDocument imported = this.ImportHtmlFromString(string.Format("<img src='data:image/png;base64,{0}' width=\"\"/>", Base64String));

            Paragraph paragraph = imported.Sections.First().Blocks.First() as Paragraph;
            ImageInline actualImage = paragraph.Inlines.First() as ImageInline;
            actualImage.Image.RecalculateSize();

            Assert.IsNotNull(actualImage);
            Assert.AreEqual(actualImage.Image.Width, image.Image.Width);
            Assert.AreEqual(actualImage.Image.Height, image.Image.Height);
        }

        [TestMethod]
        public void ImportImage_WhenImageHasInvalidWidth()
        {
            RadFlowDocument expected = new RadFlowDocument();
            ImageInline image = expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddImageInline();
            image.Image.Width = OriginalImageWidth;
            image.Image.Height = OriginalImageHeight;
            image.Image.ImageSource = new ImageSource(Convert.FromBase64String(Base64String), "png");

            RadFlowDocument imported = this.ImportHtmlFromString(string.Format("<img src='data:image/png;base64,{0}' width=\"test\"/>", Base64String));

            Paragraph paragraph = imported.Sections.First().Blocks.First() as Paragraph;
            ImageInline actualImage = paragraph.Inlines.First() as ImageInline;
            actualImage.Image.RecalculateSize();

            Assert.IsNotNull(actualImage);
            Assert.AreEqual(actualImage.Image.Width, image.Image.Width);
            Assert.AreEqual(actualImage.Image.Height, image.Image.Height);
        }

        [TestMethod]
        public void ImportImage_WhenImageHasWidthInPercentage()
        {
            RadFlowDocument expected = new RadFlowDocument();
            ImageInline image = expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddImageInline();
            image.Image.Width = OriginalImageWidth;
            image.Image.Height = OriginalImageHeight;
            image.Image.ImageSource = new ImageSource(Convert.FromBase64String(Base64String), "png");

            string imageWidthAttribute = string.Format("width=\"50%\"");
            RadFlowDocument imported = this.ImportHtmlFromString(string.Format("<img src='data:image/png;base64,{0}' {1}/>", Base64String, imageWidthAttribute));

            Paragraph paragraph = imported.Sections.First().Blocks.First() as Paragraph;
            ImageInline actualImage = paragraph.Inlines.First() as ImageInline;
            actualImage.Image.RecalculateSize();

            Assert.IsNotNull(actualImage);
            Assert.AreEqual(actualImage.Image.Width, image.Image.Width);
            Assert.AreEqual(actualImage.Image.Height, image.Image.Height);
        }

        [TestMethod]
        public void ImportImage_WhenImageHasWidthInPixels()
        {
            RadFlowDocument expected = new RadFlowDocument();
            ImageInline image = expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddImageInline();
            image.Image.Width = OriginalImageWidth;
            image.Image.Height = OriginalImageHeight;
            image.Image.ImageSource = new ImageSource(Convert.FromBase64String(Base64String), "png");

            string imageWidthAttribute = string.Format("width=\"50px\"");
            RadFlowDocument imported = this.ImportHtmlFromString(string.Format("<img src='data:image/png;base64,{0}' {1}/>", Base64String, imageWidthAttribute));

            Paragraph paragraph = imported.Sections.First().Blocks.First() as Paragraph;
            ImageInline actualImage = paragraph.Inlines.First() as ImageInline;
            actualImage.Image.RecalculateSize();

            Assert.IsNotNull(actualImage);
            Assert.AreEqual(actualImage.Image.Width, image.Image.Width);
            Assert.AreEqual(actualImage.Image.Height, image.Image.Height);
        }

        [TestMethod]
        public void ImportImage_WhenImageHasHeight()
        {
            double imageHeight = 15;

            RadFlowDocument expected = new RadFlowDocument();
            ImageInline image = expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddImageInline();
            image.Image.Width = OriginalImageWidth;
            image.Image.Height = imageHeight;
            image.Image.ImageSource = new ImageSource(Convert.FromBase64String(Base64String), "png");

            string imageHeightAttribute = string.Format("height=\"{0}\"", imageHeight);
            RadFlowDocument imported = this.ImportHtmlFromString(string.Format("<img src='data:image/png;base64,{0}' {1}/>", Base64String, imageHeightAttribute));

            Paragraph paragraph = imported.Sections.First().Blocks.First() as Paragraph;
            ImageInline actualImage = paragraph.Inlines.First() as ImageInline;
            actualImage.Image.LockAspectRatio = false;
            actualImage.Image.RecalculateSize();

            Assert.IsNotNull(actualImage);
            Assert.AreEqual(actualImage.Image.Width, image.Image.Width);
            Assert.AreEqual(actualImage.Image.Height, image.Image.Height);
        }

        [TestMethod]
        public void ImportImage_WhenImageHasEmptyHeight()
        {
            RadFlowDocument expected = new RadFlowDocument();
            ImageInline image = expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddImageInline();
            image.Image.Width = OriginalImageWidth;
            image.Image.Height = OriginalImageHeight;
            image.Image.ImageSource = new ImageSource(Convert.FromBase64String(Base64String), "png");

            RadFlowDocument imported = this.ImportHtmlFromString(string.Format("<img src='data:image/png;base64,{0}' height=\"\"/>", Base64String));

            Paragraph paragraph = imported.Sections.First().Blocks.First() as Paragraph;
            ImageInline actualImage = paragraph.Inlines.First() as ImageInline;
            actualImage.Image.RecalculateSize();

            Assert.IsNotNull(actualImage);
            Assert.AreEqual(actualImage.Image.Width, image.Image.Width);
            Assert.AreEqual(actualImage.Image.Height, image.Image.Height);
        }

        [TestMethod]
        public void ImportImage_WhenImageHasInvalidHeight()
        {
            RadFlowDocument expected = new RadFlowDocument();
            ImageInline image = expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddImageInline();
            image.Image.Width = OriginalImageWidth;
            image.Image.Height = OriginalImageHeight;
            image.Image.ImageSource = new ImageSource(Convert.FromBase64String(Base64String), "png");

            RadFlowDocument imported = this.ImportHtmlFromString(string.Format("<img src='data:image/png;base64,{0}' height=\"test\"/>", Base64String));

            Paragraph paragraph = imported.Sections.First().Blocks.First() as Paragraph;
            ImageInline actualImage = paragraph.Inlines.First() as ImageInline;
            actualImage.Image.RecalculateSize();

            Assert.IsNotNull(actualImage);
            Assert.AreEqual(actualImage.Image.Width, image.Image.Width);
            Assert.AreEqual(actualImage.Image.Height, image.Image.Height);
        }

        [TestMethod]
        public void ImportImage_WhenImageHasHeightInPercentage()
        {
            RadFlowDocument expected = new RadFlowDocument();
            ImageInline image = expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddImageInline();
            image.Image.Width = OriginalImageWidth;
            image.Image.Height = OriginalImageHeight;
            image.Image.ImageSource = new ImageSource(Convert.FromBase64String(Base64String), "png");

            string imageHeightAttribute = string.Format("height=\"50%\"");
            RadFlowDocument imported = this.ImportHtmlFromString(string.Format("<img src='data:image/png;base64,{0}' {1}/>", Base64String, imageHeightAttribute));

            Paragraph paragraph = imported.Sections.First().Blocks.First() as Paragraph;
            ImageInline actualImage = paragraph.Inlines.First() as ImageInline;
            actualImage.Image.RecalculateSize();

            Assert.IsNotNull(actualImage);
            Assert.AreEqual(actualImage.Image.Width, image.Image.Width);
            Assert.AreEqual(actualImage.Image.Height, image.Image.Height);
        }

        [TestMethod]
        public void ImportImage_WhenImageHasHeightInPixels()
        {
            RadFlowDocument expected = new RadFlowDocument();
            ImageInline image = expected.Sections.AddSection().Blocks.AddParagraph().Inlines.AddImageInline();
            image.Image.Width = OriginalImageWidth;
            image.Image.Height = OriginalImageHeight;
            image.Image.ImageSource = new ImageSource(Convert.FromBase64String(Base64String), "png");

            string imageHeightAttribute = string.Format("height=\"50px\"");
            RadFlowDocument imported = this.ImportHtmlFromString(string.Format("<img src='data:image/png;base64,{0}' {1}/>", Base64String, imageHeightAttribute));

            Paragraph paragraph = imported.Sections.First().Blocks.First() as Paragraph;
            ImageInline actualImage = paragraph.Inlines.First() as ImageInline;
            actualImage.Image.RecalculateSize();

            Assert.IsNotNull(actualImage);
            Assert.AreEqual(actualImage.Image.Width, image.Image.Width);
            Assert.AreEqual(actualImage.Image.Height, image.Image.Height);
        }
    }
}
