﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Collections;

namespace Telerik.Windows.Documents.Flow.InternalTests.Model.Collections.Common
{
    public abstract class DocumentElementCollectionTestsBase<T, TOwner>
        where T : DocumentElementBase
        where TOwner : DocumentElementBase
    {
        public abstract DocumentElementCollection<T, TOwner> GetDocumentElementCollection(TOwner owner);
        public abstract T CreateChild(RadFlowDocument document);
        public abstract TOwner CreateOwner(RadFlowDocument document);

        public void AssertAdd_WhenElementIsFromAnotherDocument_ThrowsArgumentException()
        {
            // Arrange
            TOwner owner = this.CreateOwner(new RadFlowDocument());
            DocumentElementCollection<T, TOwner> collection = this.GetDocumentElementCollection(owner);
            T childFromAnotherDocument = this.CreateChild(new RadFlowDocument());

            // Act
            try
            {
                collection.Add(childFromAnotherDocument);
                Assert.Fail("Should throw ArgumentException");
            }
            catch (ArgumentException)
            {
            }

            // Assert
            Assert.AreEqual(0, collection.Count);
        }

        public void AssertAdd_WhenElementIsFromTheSameDocument_AddsTheElement()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            TOwner owner = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection = this.GetDocumentElementCollection(owner);
            T child = this.CreateChild(document);

            // Act
            collection.Add(child);

            // Assert
            Assert.AreEqual(1, collection.Count);
            Assert.AreEqual(child, collection[0]);
        }

        public void AssertAdd_WhenElementIsNull_ThrowsArgumentNullException()
        {
            // Arrange
            TOwner owner = this.CreateOwner(new RadFlowDocument());
            DocumentElementCollection<T, TOwner> collection = this.GetDocumentElementCollection(owner);

            // Act
            try
            {
                collection.Add(null);
                Assert.Fail("Should throw ArgumentException");
            }
            catch (ArgumentNullException)
            {
            }

            // Assert
            Assert.AreEqual(0, collection.Count);
        }

        public void AssertIndexerSet_WhenElementIsFromAnotherDocument_ThrowsArgumentException()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            TOwner owner = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection = this.GetDocumentElementCollection(owner);
            T child = this.CreateChild(document);

            collection.Add(child);
            T childFromAnotherDocument = this.CreateChild(new RadFlowDocument());

            // Act
            try
            {
                collection[0] = childFromAnotherDocument;
                Assert.Fail("Should throw ArgumentException");
            }
            catch (ArgumentException)
            {
            }

            // Assert
            Assert.AreEqual(1, collection.Count);
            Assert.AreEqual(child, collection[0]);
        }

        public void AssertIndexerSet_WhenElementIsFromTheSameDocument_AddsTheElement()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            TOwner owner = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection = this.GetDocumentElementCollection(owner);
            T child = this.CreateChild(document);

            collection.Add(child);

            T child2 = this.CreateChild(document);

            // Act
            collection[0] = child2;

            // Assert
            Assert.AreEqual(1, collection.Count);
            Assert.AreEqual(child2, collection[0]);
        }

        public void AssertIndexerSet_WhenElementIsNull_ThrowsArgumentNullException()
        {
            RadFlowDocument document = new RadFlowDocument();
            TOwner owner = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection = this.GetDocumentElementCollection(owner);
            T child = this.CreateChild(document);

            collection.Add(child);

            try
            {
                collection[0] = null;
                Assert.Fail("Should throw ArgumentException");
            }
            catch (ArgumentNullException)
            {
            }

            Assert.AreEqual(1, collection.Count);
            Assert.AreEqual(child, collection[0]);
        }

        public void AssertInsert_WhenElementIsFromAnotherDocument_ThrowsArgumentException()
        {
            TOwner owner = this.CreateOwner(new RadFlowDocument());
            DocumentElementCollection<T, TOwner> collection = this.GetDocumentElementCollection(owner);
            T childFromAnotherDocument = this.CreateChild(new RadFlowDocument());

            try
            {
                collection.Insert(0, childFromAnotherDocument);
                Assert.Fail("Should throw ArgumentException");
            }
            catch (ArgumentException)
            {
            }

            Assert.AreEqual(0, collection.Count);
        }

        public void AssertInsert_WhenElementIsFromTheSameDocument_AddsTheElement()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            TOwner owner = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection = this.GetDocumentElementCollection(owner);
            T child = this.CreateChild(document);

            // Act
            collection.Insert(0, child);

            // Assert
            Assert.AreEqual(1, collection.Count);
            Assert.AreEqual(child, collection[0]);
        }

        public void AssertInsert_WhenElementIsNull_ThrowsArgumentNullException()
        {
            TOwner owner = this.CreateOwner(new RadFlowDocument());
            DocumentElementCollection<T, TOwner> collection = this.GetDocumentElementCollection(owner);

            try
            {
                collection.Insert(0, null);
                Assert.Fail("Should throw ArgumentException");
            }
            catch (ArgumentNullException)
            {
            }

            Assert.AreEqual(0, collection.Count);
        }

        public void AssertAdd_WhenElementIsFromTheSameDocument_SetsTheParent()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            TOwner owner = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection = this.GetDocumentElementCollection(owner);
            T child = this.CreateChild(document);

            // Act
            collection.Add(child);

            //Assert
            Assert.AreEqual(owner, child.Parent);
        }

        public void AssertIndexerSet_WhenElementIsFromTheSameDocument_SetsTheParent()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            TOwner owner = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection = this.GetDocumentElementCollection(owner);

            T child1 = this.CreateChild(document);
            T child2 = this.CreateChild(document);

            collection.Add(child1);

            // Act
            collection[0] = child2;

            //Assert
            Assert.AreEqual(null, child1.Parent);
            Assert.AreEqual(owner, child2.Parent);
        }

        public void AssertInsert_WhenElementIsFromTheSameDocument_SetsTheParentToBothElements()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            TOwner owner = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection = this.GetDocumentElementCollection(owner);

            T child = this.CreateChild(document);

            // Act
            collection.Insert(0, child);

            //Assert
            Assert.AreEqual(owner, child.Parent);
        }

        public void AssertRemove_WhenElementIsAdded_ClearsTheParent()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            TOwner owner = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection = this.GetDocumentElementCollection(owner);

            T child = this.CreateChild(document);
            collection.Add(child);

            // Act
            collection.Remove(child);

            //Assert
            Assert.AreEqual(null, child.Parent);
        }

        public void AssertRemoveAt_WhenElementIsAdded_ClearsTheParent()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            TOwner owner = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection = this.GetDocumentElementCollection(owner);

            T child = this.CreateChild(document);
            collection.Add(child);

            collection.RemoveAt(0);

            //Assert
            Assert.AreEqual(null, child.Parent);
        }

        public void AssertAdd_WhenElementIsInAnotherCollectionInTheSameDocument_ThrowsArgumentException()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            TOwner owner1 = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection1 = this.GetDocumentElementCollection(owner1);
            T child = this.CreateChild(document);

            collection1.Add(child);

            TOwner owner2 = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection2 = this.GetDocumentElementCollection(owner2);

            // Act
            try
            {
                collection2.Add(child);

                Assert.Fail("Should throw ArgumentException.");
            }
            catch (ArgumentException)
            {
            }

            // Assert
            Assert.AreEqual(0, collection2.Count);
        }

        public void AssertIndexerSet_WhenElementIsInAnotherCollectionInTheSameDocument_ThrowsArgumentException()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            TOwner owner1 = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection1 = this.GetDocumentElementCollection(owner1);
            T child1 = this.CreateChild(document);

            collection1.Add(child1);

            TOwner owner2 = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection2 = this.GetDocumentElementCollection(owner2);

            T child2 = this.CreateChild(document);
            collection2.Add(child2);

            // Act
            try
            {
                collection2[0] = child1;

                Assert.Fail("Should throw ArgumentException.");
            }
            catch (ArgumentException)
            {
            }

            // Assert
            Assert.AreEqual(1, collection2.Count);
            Assert.AreEqual(child2, collection2[0]);
        }

        public void AssertInsert_WhenElementIsInAnotherCollectionInTheSameDocument_ThrowsArgumentException()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            TOwner owner1 = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection1 = this.GetDocumentElementCollection(owner1);
            T child1 = this.CreateChild(document);

            collection1.Add(child1);

            TOwner owner2 = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection2 = this.GetDocumentElementCollection(owner2);

            T child2 = this.CreateChild(document);
            collection2.Add(child2);

            // Act
            try
            {
                collection2.Insert(0, child1);

                Assert.Fail("Should throw ArgumentException.");
            }
            catch (ArgumentException)
            {
            }

            // Assert
            Assert.AreEqual(1, collection2.Count);
            Assert.AreEqual(child2, collection2[0]);
        }


        public void AssertAdd_WhenElementIsAlreadyAddedToSameCollection_ThrowsArgumentException()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            TOwner owner = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection = this.GetDocumentElementCollection(owner);
            T child = this.CreateChild(document);

            collection.Add(child);

            // Act
            try
            {
                collection.Add(child);

                Assert.Fail("Should throw ArgumentException.");
            }
            catch (ArgumentException)
            {
            }

            // Assert
            Assert.AreEqual(1, collection.Count);
            Assert.AreEqual(child, collection[0]);
        }

        public void AssertInsert_WhenElementIsAlreadyAddedToSameCollection_ThrowsArgumentException()
        {
            // Arrange
            RadFlowDocument document = new RadFlowDocument();
            TOwner owner = this.CreateOwner(document);
            DocumentElementCollection<T, TOwner> collection = this.GetDocumentElementCollection(owner);
            T child = this.CreateChild(document);

            collection.Add(child);

            // Act
            try
            {
                collection.Insert(0, child);

                Assert.Fail("Should throw ArgumentException.");
            }
            catch (ArgumentException)
            {
            }

            // Assert
            Assert.AreEqual(1, collection.Count);
            Assert.AreEqual(child, collection[0]);
        }
    }
}
