﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Flow.InternalTests.Utilities;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Editing;
using Telerik.Windows.Documents.Flow.Model.Fields;

namespace Telerik.Windows.Documents.Flow.InternalTests.Model
{
    [TestClass]
    public class InlineRangeEditorTests
    {
        [TestMethod]
        public void Enumerate_WithNullFirstArgument_ThrowsException()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run end = editor.InsertText("end");

            TestHelper.AssertExceptionIsThrown<ArgumentException>(() =>
                {
                    InlineRangeEditor.EnumerateInlinesInRange(null, end).ToList();
                });
        }

        [TestMethod]
        public void Enumerate_WithNullSecondArgument_ThrowsException()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run start = editor.InsertText("start");

            TestHelper.AssertExceptionIsThrown<ArgumentException>(() =>
            {
                InlineRangeEditor.EnumerateInlinesInRange(start, null).ToList();
            });
        }

        [TestMethod]
        public void Enumerate_EndBeforeStart_InSameParagraph_ThrowsException()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run start = editor.InsertText("start");
            Run end = editor.InsertText("end");

            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
            {
                InlineRangeEditor.EnumerateInlinesInRange(end, start).ToList();
            });
        }

        [TestMethod]
        public void Enumerate_EndBeforeStart_InDifferentParagraphs_ThrowsException()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run start = editor.InsertText("start");
            editor.InsertParagraph();
            Run end = editor.InsertText("end");

            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
            {
                InlineRangeEditor.EnumerateInlinesInRange(end, start).ToList();
            });
        }

        [TestMethod]
        public void Enumerate_EndIsInTable_ThrowsException()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run start = editor.InsertText("start");
            Table table = editor.InsertTable(1, 1);
            Run end = table.Rows[0].Cells[0].Blocks.AddParagraph().Inlines.AddRun("end");

            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
            {
                InlineRangeEditor.EnumerateInlinesInRange(start, end).ToList();
            });
        }

        [TestMethod]
        public void Enumerate_StartIsInTable_ThrowsException()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Table table = editor.InsertTable(1, 1);
            Run start = table.Rows[0].Cells[0].Blocks.AddParagraph().Inlines.AddRun("start");
            Run end = editor.InsertText("end");

            TestHelper.AssertExceptionIsThrown<InvalidOperationException>(() =>
            {
                InlineRangeEditor.EnumerateInlinesInRange(start, end).ToList();
            });
        }

        [TestMethod]
        public void Enumerate_InSameParagraph_ThrowsException()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run start = editor.InsertText("start");
            Run middle1 = editor.InsertText("middle1");
            Run middle2 = editor.InsertText("middle2");
            Run end = editor.InsertText("end");

            List<InlineBase> result = InlineRangeEditor.EnumerateInlinesInRange(start, end).ToList();

            Assert.AreEqual(2, result.Count);
            Assert.AreEqual(middle1, result[0]);
            Assert.AreEqual(middle2, result[1]);
        }

        [TestMethod]
        public void Enumerate_InDifferentParagraph_ThrowsException()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run start = editor.InsertText("start");
            Run middle1 = editor.InsertText("middle1");
            editor.InsertParagraph();
            Run middle2 = editor.InsertText("middle2");
            Run end = editor.InsertText("end");

            List<InlineBase> result = InlineRangeEditor.EnumerateInlinesInRange(start, end).ToList();

            Assert.AreEqual(2, result.Count);
            Assert.AreEqual(middle1, result[0]);
            Assert.AreEqual(middle2, result[1]);
        }

        [TestMethod]
        public void Enumerate_InField_EnumeratesAllInlines()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            FieldInfo field = editor.InsertField("code fragment", "result fragment");

            List<InlineBase> result = InlineRangeEditor.EnumerateInlinesInRange(field.Start, field.End).ToList();

            Assert.AreEqual(3, result.Count);
            Assert.AreEqual("code fragment", ((Run)result[0]).Text);
            Assert.AreEqual(FieldCharacterType.Separator, ((FieldCharacter)result[1]).FieldCharacterType);
            Assert.AreEqual("result fragment", ((Run)result[2]).Text);
        }

        [TestMethod]
        public void EnumerateResult_InField_EnumeratesOnlyResultInlines()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            Run start = editor.InsertText("start");
            editor.InsertField("code fragment", "result fragment");
            Run end = editor.InsertText("end");

            List<InlineBase> result = InlineRangeEditor.EnumerateFieldInlinesInRange(start, end).ToList();

            // For now Start and End characters are enumerated when skipping field code fragment. 
            // This might be changed - in this case change the test.
            Assert.AreEqual(3, result.Count);
            Assert.AreEqual(FieldCharacterType.Start, ((FieldCharacter)result[0]).FieldCharacterType);
            Assert.AreEqual("result fragment", ((Run)result[1]).Text);
            Assert.AreEqual(FieldCharacterType.End, ((FieldCharacter)result[2]).FieldCharacterType);
        }

        [TestMethod]
        public void DeleteContent_EmptyRange_DoesNothing()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("before");
            Run start = editor.InsertText("start");
            Run end = editor.InsertText("end");
            editor.InsertText("after");

            InlineRangeEditor.DeleteContent(start, end);

            this.AssertDocumentAfterDelete(editor);
        }

        [TestMethod]
        public void DeleteContent_SameParagraph_Deletes()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("before");
            Run start = editor.InsertText("start");

            editor.InsertText("middle 1");
            editor.InsertText("middle 2");

            Run end = editor.InsertText("end");
            editor.InsertText("after");

            InlineRangeEditor.DeleteContent(start, end);

            this.AssertDocumentAfterDelete(editor);
        }

        [TestMethod]
        public void DeleteContent_ConsecutiveParagraphsNoContent_Deletes()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("before");
            Run start = editor.InsertText("start");

            editor.InsertParagraph();

            Run end = editor.InsertText("end");
            editor.InsertText("after");

            InlineRangeEditor.DeleteContent(start, end);

            this.AssertDocumentAfterDelete(editor);
        }

        [TestMethod]
        public void DeleteContent_ConsecutiveParagraphs_Deletes()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("before");
            Run start = editor.InsertText("start");

            editor.InsertText("middle 1");
            editor.InsertParagraph();
            editor.InsertText("middle 2");

            Run end = editor.InsertText("end");
            editor.InsertText("after");

            InlineRangeEditor.DeleteContent(start, end);

            this.AssertDocumentAfterDelete(editor);
        }

        [TestMethod]
        public void DeleteContent_NotConsecutiveParagraphs_Deletes()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("before");
            Run start = editor.InsertText("start");

            editor.InsertText("middle 1");
            editor.InsertParagraph();
            editor.InsertText("middle 2");
            editor.InsertParagraph();
            editor.InsertText("middle 3");

            Run end = editor.InsertText("end");
            editor.InsertText("after");

            InlineRangeEditor.DeleteContent(start, end);

            this.AssertDocumentAfterDelete(editor);
        }

        [TestMethod]
        public void DeleteContent_NotConsecutiveParagraphsAndTable_Deletes()
        {
            RadFlowDocumentEditor editor = new RadFlowDocumentEditor(new RadFlowDocument());
            editor.InsertText("before");
            Run start = editor.InsertText("start");

            editor.InsertText("middle 1");
            editor.InsertParagraph();
            editor.InsertTable(3, 3);
            editor.InsertParagraph();
            editor.InsertText("middle 3");

            Run end = editor.InsertText("end");
            editor.InsertText("after");

            InlineRangeEditor.DeleteContent(start, end);

            this.AssertDocumentAfterDelete(editor);
        }

        private void AssertDocumentAfterDelete(RadFlowDocumentEditor editor)
        {
            List<Paragraph> paragraphs = editor.Document.EnumerateChildrenOfType<Paragraph>().ToList();
            Assert.AreEqual(1, paragraphs.Count);

            Flow.Model.Collections.InlineCollection inlines = paragraphs[0].Inlines;
            Assert.AreEqual("before", ((Run)inlines[0]).Text);
            Assert.AreEqual("start", ((Run)inlines[1]).Text);
            Assert.AreEqual("end", ((Run)inlines[2]).Text);
            Assert.AreEqual("after", ((Run)inlines[3]).Text);
        }
    }
}
