﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.IO;
using System.Linq;
#if NETSTANDARD
using Telerik.Documents.Core.Fonts;
using Telerik.Documents.Media;
using Telerik.Windows.Documents.Media;
#else
using System.Windows.Media;
using ImageSource = Telerik.Windows.Documents.Media.ImageSource;
#endif
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;
using Telerik.Windows.Documents.Flow.Model.Watermarks;

namespace Telerik.Windows.Documents.Flow.InternalTests.Model
{
    [TestClass]
    public class WatermarkTests
    {
        [TestMethod]
        public void CreateWatermarkAndChangeSource()
        {
            ImageSource originalSource = new ImageSource(new MemoryStream(), "test");
            ImageWatermarkSettings imageSettings = new ImageWatermarkSettings()
            {
                Angle = 12,
                Width = 50,
                Height = 75,
                ImageSource = originalSource
            };

            Watermark watermark = new Watermark(imageSettings);

            RadFlowDocument document = new RadFlowDocument();
            Header header = document.Sections.AddSection().Headers.Add(HeaderFooterType.Default);
            header.Watermarks.Add(watermark);
           
            Assert.IsTrue(document.Resources.Contains(originalSource));

            ImageSource newSource = new ImageSource(new MemoryStream(), "test1");
            imageSettings.ImageSource = newSource;

            Assert.IsTrue(document.Resources.Contains(newSource));
            Assert.IsFalse(document.Resources.Contains(originalSource));
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void CreateWatermarkAndChangeDocument()
        {
            ImageSource originalSource = new ImageSource(new MemoryStream(), "test");
            ImageWatermarkSettings imageSettings = new ImageWatermarkSettings()
            {
                Angle = 12,
                Width = 50,
                Height = 75,
                ImageSource = originalSource
            };

            Watermark watermark = new Watermark(imageSettings);

            RadFlowDocument document = new RadFlowDocument();
            Header header = document.Sections.AddSection().Headers.Add(HeaderFooterType.Default);
            header.Watermarks.Add(watermark);

            RadFlowDocument newDocument = new RadFlowDocument();
            Header newHeader = newDocument.Sections.AddSection().Headers.Add(HeaderFooterType.Default);
            newHeader.Watermarks.Add(watermark);
        }

        [TestMethod]
        public void CreateWatermarkRemoveFromDocumentAndAddToAnother()
        {
            ImageSource originalSource = new ImageSource(new MemoryStream(), "test");
            ImageWatermarkSettings imageSettings = new ImageWatermarkSettings()
            {
                Angle = 12,
                Width = 50,
                Height = 75,
                ImageSource = originalSource
            };

            Watermark watermark = new Watermark(imageSettings);

            RadFlowDocument document = new RadFlowDocument();
            Header header = document.Sections.AddSection().Headers.Add(HeaderFooterType.Default);
            header.Watermarks.Add(watermark);

            Assert.IsTrue(document.Resources.Contains(originalSource));
            Assert.AreEqual(document, watermark.Document);

            header.Watermarks.Remove(watermark);

            Assert.IsFalse(document.Resources.Contains(originalSource));
            Assert.IsNull(watermark.Document);

            RadFlowDocument newDocument = new RadFlowDocument();
            Header newHeader = newDocument.Sections.AddSection().Headers.Add(HeaderFooterType.Default);
            newHeader.Watermarks.Add(watermark);

            Assert.IsTrue(newDocument.Resources.Contains(originalSource));
            Assert.AreEqual(newDocument, watermark.Document);
        }

        [TestMethod]
        public void ClearWatermarkCollection()
        {
            ImageSource originalSource = new ImageSource(new MemoryStream(), "test");
            ImageWatermarkSettings imageSettings = new ImageWatermarkSettings()
            {
                Angle = 12,
                Width = 50,
                Height = 75,
                ImageSource = originalSource
            };

            Watermark imageWatermark = new Watermark(imageSettings);

            TextWatermarkSettings textSettings = new TextWatermarkSettings();
            textSettings.Text = "test";
            textSettings.Angle = 45;
            textSettings.FontFamily = new FontFamily("Verdana");
            textSettings.Width = 100;
            textSettings.Height = 100;
            textSettings.Opacity = 1;
            textSettings.ForegroundColor = Colors.Red;

            Watermark textWatermark = new Watermark(textSettings);

            RadFlowDocument document = new RadFlowDocument();
            Header header = document.Sections.AddSection().Headers.Add(HeaderFooterType.Default);
            header.Watermarks.Add(imageWatermark);
            header.Watermarks.Add(textWatermark);

            Assert.IsTrue(document.Resources.Contains(originalSource));
            Assert.AreEqual(document, imageWatermark.Document);
            Assert.AreEqual(document, textWatermark.Document);

            header.Watermarks.Clear();

            Assert.IsFalse(document.Resources.Contains(originalSource));
            Assert.IsNull(imageWatermark.Document);
            Assert.IsNull(textWatermark.Document);
        }

        [TestMethod]
        public void CreateTextWatermarkSettingsAndVerifyDefaultFontFamilyTest()
        {
            TextWatermarkSettings originalTextSettings = new TextWatermarkSettings()
            {
                Angle = 12,
                Width = 50,
                Height = 75,
                Opacity = 0.4,
                ForegroundColor = Colors.Yellow
            };

            Assert.AreEqual(DocumentDefaultStyleSettings.FontFamily.LocalValue, originalTextSettings.FontFamily);
        }
    }
}
