﻿using System;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Fixed.Model.Graphics;
using Telerik.Windows.Documents.Flow.Model;
using Telerik.Windows.Documents.Flow.Model.Styles;
#if NETSTANDARD
using Telerik.Documents.Media;
#else
using System.Windows.Media;
#endif
#if WPF || SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.Model;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Fixed.Model.ColorSpaces;

namespace Telerik.Windows.Documents.Flow.FormatProviders.Pdf.Tests
{
    [TestClass]
    public class TableTests : TestsBase
    {
        [TestMethod]
        public void TestTableWithPercentageCellWidth()
        {
            RadFlowDocument document = new RadFlowDocument();
            Table table = document.Sections.AddSection().Blocks.AddTable();

            TableRow row = table.Rows.AddTableRow();
            TableCell cell1 = row.Cells.AddTableCell();
            TableCell cell2 = row.Cells.AddTableCell();

            cell1.PreferredWidth = new TableWidthUnit(TableWidthUnitType.Percent, 25);
            cell2.PreferredWidth = new TableWidthUnit(TableWidthUnitType.Percent, 75);

            PdfFormatProvider provider = new PdfFormatProvider();
            RadFixedDocument fixedDocument = provider.ExportToFixedDocument(document);

            RadFixedPage page = fixedDocument.Pages.First();

            Path path1 = page.Content.ElementAt(1) as Path;
            double firstCellWidth = path1.Geometry.Bounds.Width;

            Path path2 = page.Content.ElementAt(2) as Path;
            double secondCellWidth = path2.Geometry.Bounds.Width;

            Assert.AreEqual(156, firstCellWidth);
            Assert.AreEqual(468, secondCellWidth);
        }

        [TestMethod]
        public void TestTableWithTableWidthNoCellWidth()
        {
            RadFlowDocument document = new RadFlowDocument();
            Table table = document.Sections.AddSection().Blocks.AddTable();

            table.PreferredWidth = new TableWidthUnit(TableWidthUnitType.Percent, 100);

            TableRow row = table.Rows.AddTableRow();
            TableCell cell1 = row.Cells.AddTableCell();
            TableCell cell2 = row.Cells.AddTableCell();

            cell1.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Red);
            cell2.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Blue);

            PdfFormatProvider provider = new PdfFormatProvider();
            RadFixedDocument fixedDocument = provider.ExportToFixedDocument(document);

            RadFixedPage page = fixedDocument.Pages.First();

            double firstCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 255, 0, 0));
            double secondCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 0, 0, 255));

            Assert.AreEqual(312, firstCellWidth);
            Assert.AreEqual(312, secondCellWidth);
        }

        [TestMethod]
        public void TestTableWithTableWidthExceedingCellWidth()
        {
            RadFlowDocument document = new RadFlowDocument();
            Table table = document.Sections.AddSection().Blocks.AddTable();

            table.PreferredWidth = new TableWidthUnit(400);

            TableRow row = table.Rows.AddTableRow();
            TableCell cell1 = row.Cells.AddTableCell();
            TableCell cell2 = row.Cells.AddTableCell();

            cell1.PreferredWidth = new TableWidthUnit(75);
            cell2.PreferredWidth = new TableWidthUnit(75);

            cell1.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Red);
            cell2.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Blue);

            PdfFormatProvider provider = new PdfFormatProvider();
            RadFixedDocument fixedDocument = provider.ExportToFixedDocument(document);

            RadFixedPage page = fixedDocument.Pages.First();

            double firstCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 255, 0, 0));
            double secondCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 0, 0, 255));

            Assert.AreEqual(200, firstCellWidth);
            Assert.AreEqual(200, secondCellWidth);
        }

        [TestMethod]
        public void TestTableWithTableWidthExceedingCellWidthOneCellHasWidth()
        {
            RadFlowDocument document = new RadFlowDocument();
            Table table = document.Sections.AddSection().Blocks.AddTable();

            table.PreferredWidth = new TableWidthUnit(400);

            TableRow row = table.Rows.AddTableRow();
            TableCell cell1 = row.Cells.AddTableCell();
            TableCell cell2 = row.Cells.AddTableCell();
            TableCell cell3 = row.Cells.AddTableCell();

            cell1.PreferredWidth = new TableWidthUnit(75);

            cell1.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Red);
            cell2.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Blue);
            cell3.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Green);

            PdfFormatProvider provider = new PdfFormatProvider();
            RadFixedDocument fixedDocument = provider.ExportToFixedDocument(document);

            RadFixedPage page = fixedDocument.Pages.First();

            double firstCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 255, 0, 0));
            double secondCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 0, 0, 255));
            double thirdCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 0, 128, 0));

            Assert.AreEqual(75, firstCellWidth);
            Assert.AreEqual(162.5, secondCellWidth);
            Assert.AreEqual(162.5, thirdCellWidth);
        }

        [TestMethod]
        public void TestTableWithTableWidthLessThanCellWidth()
        {
            RadFlowDocument document = new RadFlowDocument();
            Table table = document.Sections.AddSection().Blocks.AddTable();

            table.PreferredWidth = new TableWidthUnit(150);

            TableRow row = table.Rows.AddTableRow();
            TableCell cell1 = row.Cells.AddTableCell();
            TableCell cell2 = row.Cells.AddTableCell();

            cell1.PreferredWidth = new TableWidthUnit(200);
            cell2.PreferredWidth = new TableWidthUnit(200);

            cell1.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Red);
            cell2.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Blue);

            PdfFormatProvider provider = new PdfFormatProvider();
            RadFixedDocument fixedDocument = provider.ExportToFixedDocument(document);

            RadFixedPage page = fixedDocument.Pages.First();

            double firstCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 255, 0, 0));
            double secondCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 0, 0, 255));

            Assert.AreEqual(75, firstCellWidth);
            Assert.AreEqual(75, secondCellWidth);
        }

        [TestMethod]
        public void TestTableWithCellPreferredWidthExceedingPage()
        {
            RadFlowDocument document = new RadFlowDocument();
            Table table = document.Sections.AddSection().Blocks.AddTable();

            TableRow row = table.Rows.AddTableRow();
            TableCell cell1 = row.Cells.AddTableCell();
            TableCell cell2 = row.Cells.AddTableCell();

            cell1.PreferredWidth = new TableWidthUnit(600);
            cell2.PreferredWidth = new TableWidthUnit(600);

            cell1.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Red);
            cell2.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Blue);

            PdfFormatProvider provider = new PdfFormatProvider();
            RadFixedDocument fixedDocument = provider.ExportToFixedDocument(document);

            RadFixedPage page = fixedDocument.Pages.First();

            double firstCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 255, 0, 0));
            double secondCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 0, 0, 255));

            Assert.AreEqual(312, firstCellWidth);
            Assert.AreEqual(312, secondCellWidth);
        }

        [TestMethod]
        public void TestTableWithTableWidthLessThanCellPreferredWidthExceedingPage()
        {
            RadFlowDocument document = new RadFlowDocument();
            Table table = document.Sections.AddSection().Blocks.AddTable();

            table.PreferredWidth = new TableWidthUnit(300);

            TableRow row = table.Rows.AddTableRow();
            TableCell cell1 = row.Cells.AddTableCell();
            TableCell cell2 = row.Cells.AddTableCell();

            cell1.PreferredWidth = new TableWidthUnit(600);
            cell2.PreferredWidth = new TableWidthUnit(600);

            cell1.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Red);
            cell2.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Blue);

            PdfFormatProvider provider = new PdfFormatProvider();
            RadFixedDocument fixedDocument = provider.ExportToFixedDocument(document);

            RadFixedPage page = fixedDocument.Pages.First();

            double firstCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 255, 0, 0));
            double secondCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 0, 0, 255));

            Assert.AreEqual(150, firstCellWidth);
            Assert.AreEqual(150, secondCellWidth);
        }

        [TestMethod]
        public void TestTableWithTableWidthExceedingPage()
        {
            RadFlowDocument document = new RadFlowDocument();
            Table table = document.Sections.AddSection().Blocks.AddTable();

            table.PreferredWidth = new TableWidthUnit(1200);

            TableRow row = table.Rows.AddTableRow();
            TableCell cell1 = row.Cells.AddTableCell();
            TableCell cell2 = row.Cells.AddTableCell();

            cell1.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Red);
            cell2.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Blue);

            PdfFormatProvider provider = new PdfFormatProvider();
            RadFixedDocument fixedDocument = provider.ExportToFixedDocument(document);

            RadFixedPage page = fixedDocument.Pages.First();

            double firstCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 255, 0, 0));
            double secondCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 0, 0, 255));

            Assert.AreEqual(600, firstCellWidth);
            Assert.AreEqual(600, secondCellWidth);
        }

        [TestMethod]
        public void TestTableWithTableWidthExceedingPageAndCellWidth()
        {
            RadFlowDocument document = new RadFlowDocument();
            Table table = document.Sections.AddSection().Blocks.AddTable();

            table.PreferredWidth = new TableWidthUnit(1200);

            TableRow row = table.Rows.AddTableRow();
            TableCell cell1 = row.Cells.AddTableCell();
            TableCell cell2 = row.Cells.AddTableCell();

            cell1.PreferredWidth = new TableWidthUnit(75);
            cell2.PreferredWidth = new TableWidthUnit(75);

            cell1.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Red);
            cell2.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Blue);

            PdfFormatProvider provider = new PdfFormatProvider();
            RadFixedDocument fixedDocument = provider.ExportToFixedDocument(document);

            RadFixedPage page = fixedDocument.Pages.First();

            double firstCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 255, 0, 0));
            double secondCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 0, 0, 255));

            Assert.AreEqual(600, firstCellWidth);
            Assert.AreEqual(600, secondCellWidth);
        }

        [TestMethod]
        public void TestTableWithTablePreferredWidthLongTextAndOneCellWithWidth()
        {
            RadFlowDocument document = new RadFlowDocument();
            Table table = document.Sections.AddSection().Blocks.AddTable();

            table.PreferredWidth = new TableWidthUnit(500);

            TableRow row = table.Rows.AddTableRow();
            TableCell cell1 = row.Cells.AddTableCell();
            TableCell cell2 = row.Cells.AddTableCell();
            TableCell cell3 = row.Cells.AddTableCell();

            cell1.Blocks.AddParagraph().Inlines.AddRun("Lorem ipsum dolor sit amet, consectetur adipiscing elit. Praesent sagittis dolor vitae ipsum venenatis, at fringilla eros porttitor. Integer commodo sed massa consequat dignissim. Phasellus ullamcorper tincidunt justo in volutpat. Phasellus feugiat neque diam, ac rhoncus augue suscipit vel. Mauris fringilla dolor nec arcu pellentesque sagittis. Donec viverra, erat sit amet gravida bibendum, erat metus commodo lorem, non venenatis dui nisl id dui. Ut convallis purus at sem tempor, nec tincidunt ipsum fermentum. Sed vel dolor purus. Mauris lacinia metus sit amet iaculis tempor. Mauris suscipit sed mauris et aliquet. Curabitur ullamcorper eros id nisl ultricies placerat. Morbi porttitor cursus neque a facilisis. Mauris fringilla consectetur nulla dapibus auctor. Suspendisse egestas mauris sed augue hendrerit, vitae rutrum neque interdum. Orci varius natoque penatibus et magnis dis parturient montes, nascetur ridiculus mus.");
            cell3.PreferredWidth = new TableWidthUnit(200);

            cell1.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Red);
            cell2.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Blue);
            cell3.Properties.BackgroundColor.LocalValue = new ThemableColor(Colors.Green);

            PdfFormatProvider provider = new PdfFormatProvider();
            RadFixedDocument fixedDocument = provider.ExportToFixedDocument(document);

            RadFixedPage page = fixedDocument.Pages.First();

            double firstCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 255, 0, 0));
            double secondCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 0, 0, 255));
            double thirdCellWidth = this.GetCellWidthByColor(page, new RgbColor(255, 0, 128, 0));

            Assert.AreEqual(299, Math.Round(firstCellWidth));
            Assert.AreEqual(1, Math.Round(secondCellWidth));
            Assert.AreEqual(200, thirdCellWidth);
        }

        private double GetCellWidthByColor(RadFixedPage page, RgbColor color)
        {
            Fixed.Model.Common.ContentElementBase cellElement = page.Content.First((el) =>
            {
                Path p = el as Path;
                return p != null && p.Fill.Equals(color);
            });

            double cellWidth = (cellElement as Path).Geometry.Bounds.Width;

            return cellWidth;
        }
    }
}