﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Commands
{
	internal abstract class UndoableWorkbookCommandBase<T> : WorkbookCommandBase<T>, IUndoableWorkbookCommand
		where T : WorkbookCommandContextBase
	{
		#region Methods

		protected override void OnExecuting(WorkbookCommandContextBase context)
		{
			base.OnExecuting(context);
			if (context.IsFirstExecution && context.Workbook.History.IsEnabled)
			{
				T specificContext = (T)context;
				this.PreserveStateBeforeExecute(specificContext);
			}
		}

		protected abstract void PreserveStateBeforeExecute(T context);

		public bool CanUndo(WorkbookCommandContextBase context)
		{
            return context.Workbook.History.IsEnabled && this.CanUndoOverride((T)context);
		}

		protected virtual bool CanUndoOverride(T context)
		{
			return true;
		}

		public bool Undo(WorkbookCommandContextBase context)
		{
			Guard.ThrowExceptionIfNull(context, "context");

			if (!this.CanUndo(context))
			{
				return false;
			}

			T specificContext = (T)context;
			CommandExecutingEventArgs e = new CommandExecutingEventArgs(this, context, true);
            context.Workbook.OnCommandExecuting(e);
			if (e.Canceled)
			{
				return false;
			}

			try
			{
				this.UndoOverride(specificContext);
			}
			catch (Exception ex)
			{
				CommandErrorEventArgs errorArgs = new CommandErrorEventArgs(ex, this, context, true);
                context.Workbook.OnCommandError(errorArgs);
				return false;
			}

            context.Workbook.OnCommandExecuted(new CommandExecutedEventArgs(this, context, true));
			
			return true;
		}

		protected abstract void UndoOverride(T context);

		public bool Redo(WorkbookCommandContextBase context)
		{
			T specificContext = (T)context;
			this.PrepareRedoExecute(specificContext);

            return this.Execute(context);
		}

		protected virtual void PrepareRedoExecute(T context)
		{
		}

		#endregion
	}
}