﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands
{
    internal abstract class SetBordersCommandContextBase : WorksheetCommandContextBase
    {
        #region Fields

        private readonly CellBorders newValue;
        private readonly CellRange cellRange;
        private readonly CellRange leftBorderCellRange;
        private readonly CellRange rightBorderCellRange;
        private readonly CellRange topBorderCellRange;
        private readonly CellRange bottomBorderCellRange;

        private ICompressedList<CellBorder> newLeftBorderValues;
        private ICompressedList<CellBorder> newRightBorderValues;
        private ICompressedList<CellBorder> newTopBorderValues;
        private ICompressedList<CellBorder> newBottomBorderValues;
        private ICompressedList<CellBorder> newDiagonalUpBorderValues;
        private ICompressedList<CellBorder> newDiagonalDownBorderValues;

        #endregion


        #region Properties

        public CellBorders NewValue
        {
            get
            {
                return this.newValue;
            }
        }

        public CellRange CellRange
        {
            get
            {
                return this.cellRange;
            }
        }

        internal CellRange LeftBorderCellRange
        {
            get
            {
                return this.leftBorderCellRange;
            }
        }

        internal CellRange RightBorderCellRange
        {
            get
            {
                return this.rightBorderCellRange;
            }
        }

        internal CellRange TopBorderCellRange
        {
            get
            {
                return this.topBorderCellRange;
            }
        }

        internal CellRange BottomBorderCellRange
        {
            get
            {
                return this.bottomBorderCellRange;
            }
        }

        internal ICompressedList<CellBorder> OldLeftBorderValues { get; set; }
        internal ICompressedList<CellBorder> OldRightBorderValues { get; set; }
        internal ICompressedList<CellBorder> OldTopBorderValues { get; set; }
        internal ICompressedList<CellBorder> OldBottomBorderValues { get; set; }
        internal ICompressedList<CellBorder> OldDiagonalUpBorderValues { get; set; }
        internal ICompressedList<CellBorder> OldDiagonalDownBorderValues { get; set; }

        internal ICompressedList<CellBorder> NewLeftBorderValues
        {
            get
            {
                if (this.newLeftBorderValues == null)
                {
                    this.newLeftBorderValues = this.CalculateNewLeftBorderValues();
                }

                return this.newLeftBorderValues;
            }
        }

        internal ICompressedList<CellBorder> NewRightBorderValues
        {
            get
            {
                if (this.newRightBorderValues == null)
                {
                    this.newRightBorderValues = this.CalculateNewRightBorderValues();
                }

                return this.newRightBorderValues;
            }
        }

        internal ICompressedList<CellBorder> NewTopBorderValues
        {
            get
            {
                if (this.newTopBorderValues == null)
                {
                    this.newTopBorderValues = this.CalculateNewTopBorderValues();
                }

                return this.newTopBorderValues;
            }
        }

        internal ICompressedList<CellBorder> NewBottomBorderValues
        {
            get
            {
                if (this.newBottomBorderValues == null)
                {
                    this.newBottomBorderValues = this.CalculateNewBottomBorderValues();
                }

                return this.newBottomBorderValues;
            }
        }

        internal ICompressedList<CellBorder> NewDiagonalUpBorderValues
        {
            get
            {
                if (this.newDiagonalUpBorderValues == null)
                {
                    this.newDiagonalUpBorderValues = this.CalculateNewDiagonalUpBorderValues();
                }

                return this.newDiagonalUpBorderValues;
            }
        }

        internal ICompressedList<CellBorder> NewDiagonalDownBorderValues
        {
            get
            {
                if (this.newDiagonalDownBorderValues == null)
                {
                    this.newDiagonalDownBorderValues = this.CalculateNewDiagonalDownBorderValues();
                }

                return this.newDiagonalDownBorderValues;
            }
        }

        #endregion


        #region Constructors

        public SetBordersCommandContextBase(Worksheet worksheet, CellRange cellRange, CellBorders newValue)
            : base(worksheet)
        {
            Guard.ThrowExceptionIfNull(cellRange, "cellRange");
            Guard.ThrowExceptionIfNull(newValue, "newValue");

            this.newValue = newValue;
            this.cellRange = cellRange;

            // Calculate LeftBorderRange.
            if (newValue.Right == null)
            {
                this.leftBorderCellRange = this.cellRange;
            }
            else
            {
                this.leftBorderCellRange = new CellRange(
                    cellRange.FromIndex.RowIndex,
                    cellRange.FromIndex.ColumnIndex,
                    cellRange.ToIndex.RowIndex,
                    Math.Min(SpreadsheetDefaultValues.ColumnCount - 1, cellRange.ToIndex.ColumnIndex + 1));
            }

            // Calculate RightBorderRange.
            if (newValue.Left == null)
            {
                this.rightBorderCellRange = this.cellRange;
            }
            else
            {
                this.rightBorderCellRange = new CellRange(
                    cellRange.FromIndex.RowIndex,
                    Math.Max(0, cellRange.FromIndex.ColumnIndex - 1),
                    cellRange.ToIndex.RowIndex,
                    cellRange.ToIndex.ColumnIndex);
            }

            // Calculate BottomBorderRange.
            if (newValue.Top == null)
            {
                this.bottomBorderCellRange = this.cellRange;
            }
            else
            {
                this.bottomBorderCellRange = new CellRange(
                    Math.Max(0, cellRange.FromIndex.RowIndex - 1),
                    cellRange.FromIndex.ColumnIndex,
                    cellRange.ToIndex.RowIndex,
                    cellRange.ToIndex.ColumnIndex);
            }

            // Calculate TopBorderRange.
            if (newValue.Bottom == null)
            {
                this.topBorderCellRange = this.cellRange;
            }
            else
            {
                this.topBorderCellRange = new CellRange(
                    cellRange.FromIndex.RowIndex,
                    cellRange.FromIndex.ColumnIndex,
                    Math.Min(SpreadsheetDefaultValues.RowCount - 1, cellRange.ToIndex.RowIndex + 1),
                    cellRange.ToIndex.ColumnIndex);
            }
        }

        #endregion


        #region Methods

        private ICompressedList<CellBorder> CalculateNewLeftBorderValues()
        {
            ICompressedList<CellBorder> newValues = null;
            Action ensureNewValuesInitialized = () =>
            {
                if (newValues == null)
                {
                    newValues = this.GetBorderProperty(CellPropertyDefinitions.LeftBorderProperty, this.LeftBorderCellRange);
                }
            };

            CellRange tempRange;
            if (this.NewValue.Left != null)
            {
                ensureNewValuesInitialized();

                // The left most column in the range should have left border.
                tempRange = new CellRange(
                    this.CellRange.FromIndex.RowIndex, this.CellRange.FromIndex.ColumnIndex,
                    this.CellRange.ToIndex.RowIndex, this.cellRange.FromIndex.ColumnIndex);

                SetValuesInRange(tempRange, newValues, this.NewValue.Left);
            }

            if (this.CellRange.ColumnCount > 1 && this.NewValue.InsideVertical != null)
            {
                ensureNewValuesInitialized();

                // The inside left borders - all except the first column.
                tempRange = new CellRange(
                    this.CellRange.FromIndex.RowIndex, this.CellRange.FromIndex.ColumnIndex + 1,
                    this.CellRange.ToIndex.RowIndex, this.cellRange.ToIndex.ColumnIndex);

                SetValuesInRange(tempRange, newValues, this.NewValue.InsideVertical);
            }

            // If we have right border, we should clear the left border of the first column after the selected range (if such exists).
            if (this.NewValue.Right != null && this.CellRange.ToIndex.ColumnIndex < SpreadsheetDefaultValues.ColumnCount - 1)
            {
                ensureNewValuesInitialized();

                tempRange = new CellRange(
                    this.CellRange.FromIndex.RowIndex, this.CellRange.ToIndex.ColumnIndex + 1,
                    this.CellRange.ToIndex.RowIndex, this.cellRange.ToIndex.ColumnIndex + 1);
                ClearValuesInRange(tempRange, newValues);
            }

            return newValues;
        }

        private ICompressedList<CellBorder> CalculateNewRightBorderValues()
        {
            ICompressedList<CellBorder> newValues = null;
            Action ensureNewValuesInitialized = () =>
            {
                if (newValues == null)
                {
                    newValues = this.GetBorderProperty(CellPropertyDefinitions.RightBorderProperty, this.RightBorderCellRange);
                }
            };

            CellRange tempRange;
            if (this.NewValue.Right != null)
            {
                ensureNewValuesInitialized();

                // The right most column in the range should have right border.
                tempRange = new CellRange(
                    this.CellRange.FromIndex.RowIndex, this.CellRange.ToIndex.ColumnIndex,
                    this.CellRange.ToIndex.RowIndex, this.cellRange.ToIndex.ColumnIndex);

                SetValuesInRange(tempRange, newValues, this.NewValue.Right);
            }

            if (this.CellRange.ColumnCount > 1 && this.NewValue.InsideVertical != null)
            {
                ensureNewValuesInitialized();

                // The inside right borders - all except the last column.
                tempRange = new CellRange(
                    this.CellRange.FromIndex.RowIndex, this.CellRange.FromIndex.ColumnIndex,
                    this.CellRange.ToIndex.RowIndex, this.cellRange.ToIndex.ColumnIndex - 1);

                SetValuesInRange(tempRange, newValues, this.NewValue.InsideVertical);
            }

            // If we have left border, we should clear the right border of the first column before the selected range (if such exists).
            if (this.NewValue.Left != null && this.CellRange.FromIndex.ColumnIndex > 0)
            {
                ensureNewValuesInitialized();

                tempRange = new CellRange(
                    this.CellRange.FromIndex.RowIndex, this.CellRange.FromIndex.ColumnIndex - 1,
                    this.CellRange.ToIndex.RowIndex, this.cellRange.FromIndex.ColumnIndex - 1);
                ClearValuesInRange(tempRange, newValues);
            }

            return newValues;
        }

        private ICompressedList<CellBorder> CalculateNewTopBorderValues()
        {
            ICompressedList<CellBorder> newValues = null;
            Action ensureNewValuesInitialized = () =>
            {
                if (newValues == null)
                {
                    newValues = this.GetBorderProperty(CellPropertyDefinitions.TopBorderProperty, this.TopBorderCellRange);
                }
            };

            CellRange tempRange;
            if (this.NewValue.Top != null)
            {
                ensureNewValuesInitialized();

                // The top most row in the range should have left border.
                tempRange = new CellRange(
                    this.CellRange.FromIndex.RowIndex, this.CellRange.FromIndex.ColumnIndex,
                    this.CellRange.FromIndex.RowIndex, this.cellRange.ToIndex.ColumnIndex);

                SetValuesInRange(tempRange, newValues, this.NewValue.Top);
            }

            if (this.CellRange.RowCount > 1 && this.NewValue.InsideHorizontal != null)
            {
                ensureNewValuesInitialized();

                // The inside top borders - all except the first row.
                tempRange = new CellRange(
                    this.CellRange.FromIndex.RowIndex + 1, this.CellRange.FromIndex.ColumnIndex,
                    this.CellRange.ToIndex.RowIndex, this.cellRange.ToIndex.ColumnIndex);

                SetValuesInRange(tempRange, newValues, this.NewValue.InsideHorizontal);
            }

            // If we have bottom border, we should clear the top border of the first row after the selected range (if such exists).
            if (this.NewValue.Bottom != null && this.CellRange.ToIndex.RowIndex < SpreadsheetDefaultValues.RowCount - 1)
            {
                ensureNewValuesInitialized();

                tempRange = new CellRange(
                    this.CellRange.ToIndex.RowIndex + 1, this.CellRange.FromIndex.ColumnIndex,
                    this.CellRange.ToIndex.RowIndex + 1, this.cellRange.ToIndex.ColumnIndex);
                ClearValuesInRange(tempRange, newValues);
            }

            return newValues;
        }

        private ICompressedList<CellBorder> CalculateNewBottomBorderValues()
        {
            ICompressedList<CellBorder> newValues = null;
            Action ensureNewValuesInitialized = () =>
            {
                if (newValues == null)
                {
                    newValues = this.GetBorderProperty(CellPropertyDefinitions.BottomBorderProperty, this.BottomBorderCellRange);
                }
            };

            CellRange tempRange;
            if (this.NewValue.Bottom != null)
            {
                ensureNewValuesInitialized();

                // The bottom most column in the range should have bottom border.
                tempRange = new CellRange(
                    this.CellRange.ToIndex.RowIndex, this.CellRange.FromIndex.ColumnIndex,
                    this.CellRange.ToIndex.RowIndex, this.cellRange.ToIndex.ColumnIndex);

                SetValuesInRange(tempRange, newValues, this.NewValue.Bottom);
            }

            if (this.CellRange.RowCount > 1 && this.NewValue.InsideHorizontal != null)
            {
                ensureNewValuesInitialized();

                // The inside bottom borders - all except the last row.
                tempRange = new CellRange(
                    this.CellRange.FromIndex.RowIndex, this.CellRange.FromIndex.ColumnIndex,
                    this.CellRange.ToIndex.RowIndex - 1, this.CellRange.ToIndex.ColumnIndex);

                SetValuesInRange(tempRange, newValues, this.NewValue.InsideHorizontal);
            }

            // If we have top border, we should clear the bottom border of the first row before the selected range (if such exists).
            if (this.NewValue.Top != null && this.CellRange.FromIndex.RowIndex > 0)
            {
                ensureNewValuesInitialized();

                tempRange = new CellRange(
                    this.CellRange.FromIndex.RowIndex - 1, this.CellRange.FromIndex.ColumnIndex,
                    this.CellRange.FromIndex.RowIndex - 1, this.cellRange.ToIndex.ColumnIndex);
                ClearValuesInRange(tempRange, newValues);
            }

            return newValues;
        }

        private ICompressedList<CellBorder> CalculateNewDiagonalUpBorderValues()
        {
            if (this.NewValue.DiagonalUp == null)
            {
                return null;
            }

            ICompressedList<CellBorder> newValues = this.GetBorderProperty(CellPropertyDefinitions.DiagonalUpBorderProperty, this.CellRange);
            SetValuesInRange(this.CellRange, newValues, this.NewValue.DiagonalUp);

            return newValues;
        }

        private ICompressedList<CellBorder> CalculateNewDiagonalDownBorderValues()
        {
            if (this.NewValue.DiagonalDown == null)
            {
                return null;
            }

            ICompressedList<CellBorder> newValues = this.GetBorderProperty(CellPropertyDefinitions.DiagonalDownBorderProperty, this.CellRange);
            SetValuesInRange(this.CellRange, newValues, this.NewValue.DiagonalDown);

            return newValues;
        }

        public abstract ICompressedList<CellBorder> GetBorderProperty(IPropertyDefinition<CellBorder> borderProperty, CellRange range);

        public abstract void SetBorderProperty(IPropertyDefinition<CellBorder> borderProperty, CellRange range, ICompressedList<CellBorder> values);

        protected abstract void SetValuesInRange(CellRange range, ICompressedList<CellBorder> collection, CellBorder value);

        protected abstract void ClearValuesInRange(CellRange range, ICompressedList collection);

        #endregion
    }
}