﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Collections;

namespace Telerik.Windows.Documents.Spreadsheet.Core.DataStructures
{
	internal class AATree<T> : IEnumerable<T>
		where T : IComparable<T>
	{
		#region AATreeNode

		internal class AATreeNode
		{
			internal int level;
			internal AATreeNode left;
			internal AATreeNode right;

			internal T value;

			internal AATreeNode()
			{
				this.level = 0;
				this.left = this;
				this.right = this;
			}

			internal AATreeNode(T value, AATreeNode sentinel)
			{
				this.level = 1;
				this.left = sentinel;
				this.right = sentinel;
				this.value = value;
			}
		}

		#endregion


		#region Fields

		protected AATreeNode root;
		protected readonly AATreeNode sentinel;
		private AATreeNode deleted;
		private bool isEnumerableCacheValid;
		private IEnumerable<T> enumerableCache;

		#endregion


		#region Constructors

		public AATree()
		{
			this.root = this.sentinel = new AATreeNode();
			this.deleted = null;
		}

		#endregion


		#region Methods

		private void Skew(ref AATreeNode node)
		{
			if (node.level == node.left.level)
			{
				// rotate right
				AATreeNode left = node.left;
				node.left = left.right;
				left.right = node;
				node = left;
			}
		}

		private void Split(ref AATreeNode node)
		{
			if (node.right.right.level == node.level)
			{
				// rotate left
				AATreeNode right = node.right;
				node.right = right.left;
				right.left = node;
				node = right;
				node.level++;
			}
		}

		private bool Insert(ref AATreeNode node, T value)
		{
			if (node == sentinel)
			{
				node = new AATreeNode(value, sentinel);
				return true;
			}

			int compare = value.CompareTo(node.value);
			if (compare < 0)
			{
				if (!this.Insert(ref node.left, value))
				{
					return false;
				}
			}
			else if (compare > 0)
			{
                if (!this.Insert(ref node.right, value))
				{
					return false;
				}
			}
			else
			{
				return false;
			}

            this.Skew(ref node);
            this.Split(ref node);

			this.InvalidateEnumerableCache();
			return true;
		}

		private bool Delete(ref AATreeNode node, T value)
		{
			if (node == sentinel)
			{
				return (deleted != null);
			}

			int compare = value.CompareTo(node.value);
			if (compare < 0)
			{
                if (!this.Delete(ref node.left, value))
				{
					return false;
				}
			}
			else
			{
				if (compare == 0)
				{
					deleted = node;
				}
                if (!this.Delete(ref node.right, value))
				{
					return false;
				}
			}

			if (deleted != null)
			{
				deleted.value = node.value;
				deleted = null;
				node = node.right;
			}
			else if (node.left.level < node.level - 1 || 
                node.right.level < node.level - 1)
			{
				node.level--;
				if (node.right.level > node.level)
				{
					node.right.level = node.level;
				}
                this.Skew(ref node);
                this.Skew(ref node.right);
                this.Skew(ref node.right.right);
                this.Split(ref node);
                this.Split(ref node.right);
			}

			this.InvalidateEnumerableCache();
			return true;
		}

		private AATreeNode Search(AATreeNode node, T value)
		{
			if (node == sentinel)
			{
				return null;
			}

			int compare = value.CompareTo(node.value);
			if (compare < 0)
			{
                return this.Search(node.left, value);
			}
			else if (compare > 0)
			{
                return this.Search(node.right, value);
			}
			else
			{
				return node;
			}
		}

        private void BuildEnumerateCache(AATreeNode node, ref Queue<T> result)
        {
            if (node != sentinel)
            {
                this.BuildEnumerateCache(node.left, ref result);
                result.Enqueue(node.value);
                this.BuildEnumerateCache(node.right, ref result);
            }
        }

		#endregion


		#region Public API

        protected void InvalidateEnumerableCache()
        {
            this.isEnumerableCacheValid = false;
        }

		public bool Add(T value)
		{
            return this.Insert(ref root, value);
		}

		public bool Delete(T value)
		{
			Debug.Assert(value != null);
            return this.Delete(ref root, value);
		}

		public bool Contains(T value)
		{
			return this.Search(root, value) != null;
		}

		public IEnumerator<T> GetEnumerator()
		{
			if (!this.isEnumerableCacheValid)
			{
				Queue<T> result = new Queue<T>();
				this.BuildEnumerateCache(this.root, ref result);
				enumerableCache = result;
				this.isEnumerableCacheValid = true;
			}

			return enumerableCache.GetEnumerator();
		}

		IEnumerator IEnumerable.GetEnumerator()
		{
			return this.GetEnumerator();
		}

		#endregion
	}
}