﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Core.DataStructures
{
    internal class CompressedList<TValue> : ICompressedList<TValue>
    {
        #region Constants

        private static readonly ValueBox<TValue> DefaultValueMarker = null;

        #endregion


        #region Fields

        private ValueBox<TValue> defaultValue;
        private readonly long fromIndex;
        private readonly long toIndex;
        private readonly RangeTreeCollection<long, ValueBox<TValue>> rangesTree;

        #endregion


        #region Properties

        public long FromIndex
        {
            get
            {
                return this.fromIndex;
            }
        }

        public long ToIndex
        {
            get
            {
                return this.toIndex;
            }
        }

        public long RangeLength
        {
            get
            {
                return this.toIndex - this.fromIndex + 1;
            }
        }

        #endregion


        #region Constructors

        public CompressedList(long fromIndex, long toIndex, TValue defaultValue)
        {
            Guard.ThrowExceptionIfLessThan(fromIndex, toIndex, "toIndex");
            this.defaultValue = new ValueBox<TValue>(defaultValue);

            this.fromIndex = fromIndex;
            this.toIndex = toIndex;
            this.rangesTree = new RangeTreeCollection<long, ValueBox<TValue>>();

            this.rangesTree.Add(new Range<long, ValueBox<TValue>>(fromIndex, toIndex, true));
        }

        public CompressedList(long fromIndex, long toIndex)
            : this(fromIndex, toIndex, default(TValue))
        {
        }

        internal CompressedList(CompressedList<TValue> original)
            : this(original.FromIndex, original.ToIndex, original.defaultValue.Value)
        {
        }

        internal CompressedList(ICompressedList<TValue> original)
            : this(original.FromIndex, original.ToIndex, original.GetDefaultValue())
        {
        }

        #endregion


        #region DefaultValue

        public void SetDefaultValue(TValue defaultValue)
        {
            this.defaultValue = new ValueBox<TValue>(defaultValue);
        }

        public TValue GetDefaultValue()
        {
            return this.defaultValue.Value;
        }

        #endregion


        #region Offsets

        ICompressedList ICompressedList.Offset(long delta)
        {
            return this.Offset(delta);
        }

        public CompressedList<TValue> Offset(long delta)
        {
            long newFromIndex = this.FromIndex + delta;
            long newToIndex = this.ToIndex + delta;
            CompressedList<TValue> toCompressedList = new CompressedList<TValue>(newFromIndex, newToIndex, this.GetDefaultValue());
            ((ITranslateableCollection)this).Copy(this.FromIndex, this.RangeLength, toCompressedList, newFromIndex);

            return toCompressedList;
        }

        #endregion


        #region SetValue

        public void SetValue(long index, TValue value)
        {
            this.SetValue(index, index, value);
        }

        public void SetValue(long fromIndex, IEnumerable<TValue> values)
        {
            Guard.ThrowExceptionIfNull(values, "values");
            TValue[] array = values.ToArray();

            this.ValidatePosition(fromIndex);
            this.ValidatePosition(fromIndex + array.Length - 1);

            long start = 0;
            long i = 0;
            while (i < array.Length)
            {
                if (!TelerikHelper.EqualsOfT<TValue>(array[start], array[i]))
                {
                    this.SetValue(start + fromIndex, i + fromIndex, array[start]);
                    start = i;
                }

                i++;
            }

            this.SetValue(start + fromIndex, array.Length - 1 + fromIndex, array[start]);
        }

        public void SetValue(long fromIndex, long toIndex, TValue value)
        {
            this.SetValueInternal(fromIndex, toIndex, new ValueBox<TValue>(value));
        }

        public void SetValue(CompressedList<TValue> compressedList)
        {
            foreach (var range in compressedList.rangesTree)
            {
                this.SetValueInternal(range.Start, range.End, range.Value);
            }
        }

        private void SetValueInternal(long index, ValueBox<TValue> value)
        {
            if (!TelerikHelper.EqualsOfT<ValueBox<TValue>>(this.GetValueInternal(index), value))
            {
                this.SetValueInternal(index, index, value);
            }
        }

        private void SetValueInternal(long fromIndex, long toIndex, ValueBox<TValue> value)
        {
            Guard.ThrowExceptionIfLessThan(fromIndex, toIndex, "toIndex");
            this.ValidatePosition(fromIndex);
            this.ValidatePosition(toIndex);

            bool isDefault = TelerikHelper.EqualsOfT(DefaultValueMarker, value);
            var ranges = this.rangesTree.GetIntersectingRanges(new Range<long, ValueBox<TValue>>(Math.Max(0, fromIndex - 1), toIndex + 1, isDefault, value));


            //if (ranges.Count == 1 && Equals(ranges[0].Value, value))
            //{
            //    return;
            //}

            if (ranges.Count > 0 && ranges.First().End.CompareTo(fromIndex) < 0)
            {
                var leftRange = ranges.First();
                if ((value == null && leftRange.Value == null) ||
                    (leftRange.Value != null && leftRange.Value.Equals(value)))
                {
                    fromIndex = leftRange.Start;
                }
                else
                {
                    ranges.RemoveAt(0);
                }
            }

            if (ranges.Count > 0 && ranges.Last().Start.CompareTo(toIndex) > 0)
            {
                var rightRange = ranges.Last();
                if ((value == null && rightRange.Value == null) ||
                    (rightRange.Value != null && rightRange.Value.Equals(value)))
                {
                    toIndex = rightRange.End;
                }
                else
                {
                    ranges.RemoveAt(ranges.Count - 1);
                }
            }

            foreach (var range in ranges)
            {
                this.rangesTree.Delete(range);

                if (range.Start.CompareTo(fromIndex) < 0)
                {
                    if (!TelerikHelper.EqualsOfT(range.Value, value))
                    {
                        var leftRange = new Range<long, ValueBox<TValue>>(range.Start, fromIndex - 1, range.IsDefault, range.Value);
                        this.rangesTree.Add(leftRange);
                    }
                    else
                    {
                        fromIndex = range.Start;
                    }
                }

                if (range.End.CompareTo(toIndex) > 0)
                {
                    if (!TelerikHelper.EqualsOfT(range.Value, value))
                    {
                        var rightRange = new Range<long, ValueBox<TValue>>(toIndex + 1, range.End, range.IsDefault, range.Value);
                        this.rangesTree.Add(rightRange);
                    }
                    else
                    {
                        toIndex = range.End;
                    }
                }
            }

            isDefault = TelerikHelper.EqualsOfT(DefaultValueMarker, value);
            var newRange = new Range<long, ValueBox<TValue>>(fromIndex, toIndex, isDefault, value);
            this.rangesTree.Add(newRange);

            this.ValidateStructure();
        }

        #endregion


        #region GetValue

        public TValue GetValue(long index)
        {
            ValueBox<TValue> valueBox = this.GetValueInternal(index);

            if (TelerikHelper.EqualsOfT(DefaultValueMarker, valueBox))
            {
                return this.defaultValue.Value;
            }

            return valueBox.Value;
        }

        public CompressedList<TValue> GetValue(long fromIndex, long toIndex)
        {
            Guard.ThrowExceptionIfLessThan(fromIndex, toIndex, "toIndex");

            this.ValidatePosition(fromIndex);
            this.ValidatePosition(toIndex);

            CompressedList<TValue> sublist = new CompressedList<TValue>(fromIndex, toIndex, this.defaultValue.Value);

            var ranges = this.rangesTree.GetIntersectingRanges(new Range<long, ValueBox<TValue>>(fromIndex, toIndex, true, new ValueBox<TValue>(default(TValue))));

            var leftRange = ranges.First();
            sublist.SetValueInternal(TelerikHelper.Max(fromIndex, leftRange.Start), TelerikHelper.Min(toIndex, leftRange.End), leftRange.Value);
            ranges.RemoveAt(0);

            if (ranges.Count > 0)
            {
                var rightRange = ranges.Last();
                sublist.SetValueInternal(rightRange.Start, TelerikHelper.Min(toIndex, rightRange.End), rightRange.Value);
                ranges.RemoveAt(ranges.Count - 1);
            }

            foreach (var range in ranges)
            {
                sublist.SetValueInternal(range.Start, range.End, range.Value);
            }

            return sublist;
        }

        ICompressedList<TValue> ICompressedList<TValue>.GetValue(long fromIndex, long toIndex)
        {
            return this.GetValue(fromIndex, toIndex);
        }

        public ValueBox<TValue> GetValueInternal(long index)
        {
            this.ValidatePosition(index);
            var range = this.rangesTree.GetContainingRange(index);

            return range.Value;
        }

        #endregion


        #region ClearValue

        public void ClearValue(long index)
        {
            this.ClearValue(index, index);
        }

        public void ClearValue(long fromIndex, long toIndex)
        {
            this.SetValueInternal(fromIndex, toIndex, DefaultValueMarker);
        }

        #endregion


        #region IEnumerable<Range<long, TValue>>

        public IEnumerable<Range<long, TValue>> GetNonDefaultRanges()
        {
            foreach (var range in this.rangesTree)
            {
                if (!TelerikHelper.EqualsOfT(DefaultValueMarker, range.Value))
                {
                    yield return new Range<long, TValue>(range.Start, range.End, range.IsDefault, range.Value.Value);
                }
            }
        }

        public IEnumerable<LongRange> GetRanges(bool getDefaultRanges)
        {
            foreach (Range<long, ValueBox<TValue>> range in this.rangesTree)
            {
                bool isDefault = TelerikHelper.EqualsOfT(DefaultValueMarker, range.Value);

                if ((!getDefaultRanges && !isDefault) || (getDefaultRanges && isDefault))
                {
                    yield return new LongRange(range.Start, range.End);
                }
            }
        }

        public IEnumerator<Range<long, TValue>> GetEnumerator()
        {
            foreach (Range<long, ValueBox<TValue>> range in this.rangesTree)
            {
                if (TelerikHelper.EqualsOfT(DefaultValueMarker, range.Value))
                {
                    yield return new Range<long, TValue>(range.Start, range.End, true, this.defaultValue.Value);
                }
                else
                {
                    yield return new Range<long, TValue>(range.Start, range.End, false, range.Value.Value);
                }
            }
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return ((IEnumerable<Range<long, TValue>>)this).GetEnumerator();
        }

        #endregion


        #region ITranslateableCollection

        void ITranslateableCollection.Copy(long fromIndex, long toIndex)
        {
            var fromValue = this.GetValueInternal(fromIndex);
            this.SetValueInternal(toIndex, fromValue);
        }

        void ITranslateableCollection.Copy(long fromIndex, long itemsCount, long toIndex)
        {
            ((ITranslateableCollection)this).Copy(fromIndex, itemsCount, this, toIndex);
        }

        void ITranslateableCollection.Copy(long fromIndex, long itemsCount, ITranslateableCollection toCollection, long toIndex)
        {
            CompressedList<TValue> toCompressedList = (CompressedList<TValue>)toCollection;

            List<Range<long, ValueBox<TValue>>> values = this.rangesTree.GetIntersectingRanges(new Range<long, ValueBox<TValue>>(fromIndex, fromIndex + itemsCount - 1, true));

            long delta = toIndex - fromIndex;
            foreach (var range in values)
            {
                long start = Math.Max(fromIndex, range.Start);
                long end = Math.Min(fromIndex + itemsCount - 1, range.End);

                toCompressedList.SetValueInternal(start + delta, end + delta, range.Value);
            }
        }

        void ITranslateableCollection.Clear(long index)
        {
            this.ClearValue(index);
        }

        void ITranslateableCollection.Clear(long fromIndex, long toIndex)
        {
            this.ClearValue(fromIndex, toIndex);
        }

        void ITranslateableCollection.Translate(long fromIndex, long toIndex, long offset, bool useSameValueAsPrevious)
        {
            Guard.ThrowExceptionIfLessThan(fromIndex, toIndex, "toIndex");
            this.ValidatePosition(fromIndex);
            this.ValidatePosition(toIndex);

            if (offset == 0)
            {
                return;
            }

            var rangeForTranslation = new Range<long, ValueBox<TValue>>(fromIndex, toIndex, true);

            // the ranges we are working on
            var ranges = this.rangesTree.GetIntersectingRanges(new Range<long, ValueBox<TValue>>(fromIndex, toIndex, true));

            Range<long, ValueBox<TValue>> firstRange = ranges.First();

            // TODO: Think of some kind of improvement.
            //if (ranges.Count == 1 && object.Equals(firstRange.Value, DefaultValueMarker))
            //{
            //    return;
            //}

            // splits the first range of the ranges we are working on, so there is space for translating
            if (firstRange.Start < fromIndex)
            {
                var tmpRange = new Range<long, ValueBox<TValue>>(firstRange.Start, fromIndex - 1, firstRange.IsDefault, firstRange.Value);
                firstRange.Start = fromIndex;
                this.rangesTree.Add(tmpRange);
            }

            // splits the last range of the ranges we are working on, so there is space for translating
            Range<long, ValueBox<TValue>> lastRange = ranges.Last();
            if (lastRange.End > toIndex)
            {
                var tmpRange = new Range<long, ValueBox<TValue>>(toIndex + 1, lastRange.End, lastRange.IsDefault, lastRange.Value);
                lastRange.End = toIndex;
                this.rangesTree.Add(tmpRange);
            }

            var rangeTrannslator = new Action<Range<long, ValueBox<TValue>>>(
                (range) =>
                {
                    Range<long, ValueBox<TValue>> translatedRange = new Range<long, ValueBox<TValue>>(range.Start + offset, range.End + offset, range.IsDefault, range.Value);
                    if (!rangeForTranslation.IntersectsWith(translatedRange))
                    {
                        // if range we want to translate do not intersects with current range that is translated
                        this.rangesTree.Delete(range);
                        ranges.Remove(range);
                    }
                    else
                    {
                        range.Start = Math.Max(fromIndex, translatedRange.Start);
                        range.End = Math.Min(translatedRange.End, toIndex);
                    }
                });

            // Translate the ranges.
            var rangesToTranslate = ranges.ToArray();
            if (offset > 0)
            {
                for (int i = rangesToTranslate.Length - 1; i >= 0; i--)
                {
                    rangeTrannslator(rangesToTranslate[i]);
                }
            }
            else
            {
                for (int i = 0; i < rangesToTranslate.Length; i++)
                {
                    rangeTrannslator(rangesToTranslate[i]);
                }
            }

            var beforeRange = this.rangesTree.GetContainingRange(fromIndex - 1);
            var afterRange = this.rangesTree.GetContainingRange(toIndex + 1);

            ValueBox<TValue> newRangeValue = DefaultValueMarker;
            bool isDefault = true;
            if (useSameValueAsPrevious && beforeRange != null)
            {
                newRangeValue = beforeRange.Value;
                isDefault = false;
            }

            // Add the new range.
            Range<long, ValueBox<TValue>> newRange;
            if (ranges.Count == 0)
            {
                newRange = new Range<long, ValueBox<TValue>>(fromIndex, toIndex, isDefault, newRangeValue);
                ranges.Add(newRange);
                this.rangesTree.Add(newRange);
            }
            else
            {
                if (offset > 0)
                {
                    newRange = new Range<long, ValueBox<TValue>>(fromIndex, ranges[0].Start - 1, isDefault, newRangeValue);
                    if (TelerikHelper.EqualsOfT(ranges[0].Value, newRange.Value))
                    {
                        ranges[0].Start = newRange.Start;
                    }
                    else
                    {
                        ranges.Insert(0, newRange);
                        this.rangesTree.Add(newRange);
                    }
                }
                else
                {
                    newRange = new Range<long, ValueBox<TValue>>(ranges.Last().End + 1, toIndex, isDefault, newRangeValue);
                    if (TelerikHelper.EqualsOfT(ranges.Last().Value, newRange.Value))
                    {
                        ranges.Last().End = newRange.End;
                    }
                    else
                    {
                        ranges.Add(newRange);
                        this.rangesTree.Add(newRange);
                    }
                }
            }

            var leftRange = ranges.FirstOrDefault();
            var rightRange = ranges.LastOrDefault();

            if (leftRange != null)
            {
                beforeRange = this.rangesTree.GetContainingRange(leftRange.Start - 1);
                if (beforeRange != null && TelerikHelper.EqualsOfT(beforeRange.Value, leftRange.Value))
                {
                    this.rangesTree.Delete(beforeRange);
                    leftRange.Start = beforeRange.Start;
                }

            }

            if (rightRange != null)
            {
                afterRange = this.rangesTree.GetContainingRange(rightRange.End + 1);
                if (afterRange != null && TelerikHelper.EqualsOfT(afterRange.Value, rightRange.Value))
                {
                    this.rangesTree.Delete(afterRange);
                    rightRange.End = afterRange.End;
                }
            }

            this.ValidateStructure();
        }

        #endregion


        #region ICompressedList Implementation

        void ICompressedList.Sort(long fromIndex, long toIndex, int[] sortedIndexes)
        {
            this.Sort(fromIndex, toIndex, sortedIndexes,
                (newIndex, value) =>
                {
                    return value;
                });
        }

        void ICompressedList<TValue>.Sort(long fromIndex, long toIndex, int[] sortedIndexes, Func<long, ValueBox<TValue>, ValueBox<TValue>> translateValue)
        {
            this.Sort(fromIndex, toIndex, sortedIndexes, translateValue);
        }

        ICompressedList ICompressedList.GetValue(long fromIndex, long toIndex)
        {
            return this.GetValue(fromIndex, toIndex);
        }

        bool ICompressedList.ContainsNonDefaultValues(long fromIndex, long toIndex)
        {
            var ranges = this.rangesTree.GetIntersectingRanges(new Range<long, ValueBox<TValue>>(fromIndex, toIndex, true, DefaultValueMarker));

            foreach (var range in ranges)
            {
                if (!TelerikHelper.EqualsOfT(range.Value, DefaultValueMarker))
                {
                    return true;
                }
            }

            return false;
        }

        void ICompressedList.SetValue(ICompressedList compressedList)
        {
            this.SetValue((CompressedList<TValue>)compressedList);
        }

        /// <summary>
        /// Sets a single value of compressed list to range starting from "fromIndex" and ending at "toIndex".
        /// </summary>
        /// <param name="fromIndex">From index.</param>
        /// <param name="toIndex">To index.</param>
        /// <param name="singleValueCompressedList">The single value compressed list.</param>
        void ICompressedList.SetValue(long fromIndex, long toIndex, ICompressedList singleValueCompressedList)
        {
            long count = singleValueCompressedList.ToIndex - singleValueCompressedList.FromIndex;
            Guard.ThrowExceptionIfGreaterThan(0, count, "count");

            CompressedList<TValue> values = (CompressedList<TValue>)singleValueCompressedList;
            TValue value = values.GetValue(values.FromIndex);
            this.SetValue(fromIndex, toIndex, value);
        }

        void ICompressedList.SetDefaultValue(long fromIndex, long toIndex)
        {
            this.SetValue(fromIndex, toIndex, this.GetDefaultValue());
        }

        long? ICompressedList.GetLastNonEmptyRangeEnd(long fromIndex, long toIndex)
        {
            var ranges = this.rangesTree.GetIntersectingRanges(new Range<long, ValueBox<TValue>>(fromIndex, toIndex, true, DefaultValueMarker));
            ranges.Reverse();
            foreach (var range in ranges)
            {
                if (!TelerikHelper.EqualsOfT(range.Value, DefaultValueMarker))
                {
                    return Math.Min(range.End, toIndex);
                }
            }

            return null;
        }

        ICompressedList ICompressedList.CreateInstance(long fromIndex, long toIndex)
        {
            return new CompressedList<TValue>(fromIndex, toIndex, this.GetDefaultValue());
        }

        #endregion


        #region Methods

        private void Sort(long fromLongIndex, long toLongIndex, int[] sortedIndexes, Func<long, ValueBox<TValue>, ValueBox<TValue>> translateValue)
        {
            ValueBox<TValue>[] values = new ValueBox<TValue>[sortedIndexes.Length];

            for (long i = fromLongIndex; i <= toLongIndex; i++)
            {
                values[i - fromLongIndex] = this.GetValueInternal(i);
            }

            for (long i = 0; i < sortedIndexes.Length; i++)
            {
                long newIndex = i + fromLongIndex;
                ValueBox<TValue> valueToTranslate = values[sortedIndexes[i]];
                ValueBox<TValue> translatedValue = translateValue(newIndex, valueToTranslate);
                this.SetValueInternal(newIndex, translatedValue);
            }
        }

        private void ValidatePosition(long position)
        {
            if (position < this.fromIndex && this.toIndex < position)
            {
                throw new IndexOutOfRangeException("position");
            }
        }

        private void ValidateStructure()
        {
            //#if DEBUG
            //            // NOTE: Commented because of some test slowliness.
            //            Range<long, ValueBox<TValue>> previousRange = null;
            //            foreach (var range in this.rangesTree)
            //            {
            //                if (previousRange == null)
            //                {
            //                    Debug.Assert(range.Start == this.range.Start);
            //                }
            //                else
            //                {
            //                    Debug.Assert(previousRange.End == range.Start - 1);
            //                    Debug.Assert(!Equals(previousRange.Value, range.Value));
            //                }

            //                previousRange = range;
            //            }

            //            Debug.Assert(previousRange == null || previousRange.End == this.range.End);
            //#endif
        }

        public override string ToString()
        {
            StringBuilder builder = new StringBuilder();
            foreach (var range in this.rangesTree)
            {
                builder.Append(string.Format("{0}->{1}\n", range, range.Value));
            }

            return builder.ToString();
        }

        #endregion
    }
}