﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Core.DataStructures
{
    /// <summary>
    /// Represents a range.
    /// </summary>
    public class Range
    {
        #region Fields

        private readonly int start;
        private readonly int end;
        
        #endregion


        #region Properties

        /// <summary>
        /// Gets the start.
        /// </summary>
        /// <value>The start.</value>
        public int Start
        {
            get
            {
                return this.start;
            }
        }

        /// <summary>
        /// Gets the end.
        /// </summary>
        /// <value>The end.</value>
        public int End
        {
            get
            {
                return this.end;
            }
        }

        /// <summary>
        /// Gets the length.
        /// </summary>
        /// <value>The length.</value>
        public long Length
        {
            get
            {
                return this.end - this.start + 1;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="Range" /> class.
        /// </summary>
        /// <param name="start">The start.</param>
        /// <param name="end">The end.</param>
        public Range(int start, int end)
        {
            Guard.ThrowExceptionIfLessThan(start, end, "end");

            this.start = start;
            this.end = end;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Expands by the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <returns>The expanded range.</returns>
        public Range Expand(int index)
        {
            return new Range(Math.Min(index, this.Start), Math.Max(index, this.End));
        }

        /// <summary>
        /// Creates range or expands an existing one by specified index.
        /// </summary>
        /// <param name="range">The range.</param>
        /// <param name="index">The index.</param>
        /// <returns>The result range.</returns>
        public static Range CreateOrExpand(Range range, int index)
        {
            if (range == null)
            {
                return new Range(index, index);
            }

            return range.Expand(index);
        }

        /// <summary>
        /// Returns the max range or null if both ranges are null.
        /// </summary>
        /// <param name="firstRange">The first range.</param>
        /// <param name="secondRange">The second range.</param>
        /// <returns>The maximum range or null.</returns>
        public static Range MaxOrNull(Range firstRange, Range secondRange)
        {
            if (firstRange == null)
            {
                return secondRange;
            }

            if (secondRange == null)
            {
                return firstRange;
            }

            return Max(firstRange, secondRange);
        }

        /// <summary>
        /// Finds the max range of two ranges.
        /// </summary>
        /// <param name="firstRange">The first range.</param>
        /// <param name="secondRange">The second range.</param>
        /// <returns>The result range.</returns>
        public static Range Max(Range firstRange, Range secondRange)
        {
            Guard.ThrowExceptionIfNull(firstRange, "firstRange");
            Guard.ThrowExceptionIfNull(secondRange, "secondRange");

            return new Range(Math.Min(firstRange.Start, secondRange.Start), Math.Max(firstRange.End, secondRange.end));
        }

        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current
        /// <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
            return string.Format("[{0}, {1}]", this.Start, this.End);
        }

        #endregion
    }
}