﻿using System;
using System.Collections;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Core
{
    /// <summary>
    /// Represents list of named objects.
    /// </summary>
    /// <typeparam name="T">The type of the T.</typeparam>
    public class NamedObjectList<T> : ICollection<T>
        where T : INamedObject
    {
        #region Fields

        private readonly Dictionary<string, T> nameToItem;
        private readonly LinkedList<T> itemList;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the number of elements contained in the <see cref="T:System.Collections.Generic.ICollection`1" />.
        /// </summary>
        /// <value>The number of elements contained in the <see cref="T:System.Collections.Generic.ICollection`1" />.</value>
        public int Count
        {
            get
            {
                return this.itemList.Count;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the <see cref="T:System.Collections.Generic.ICollection`1" />
        /// is read-only.
        /// </summary>
        /// <value>True if the <see cref="T:System.Collections.Generic.ICollection`1" />
        /// is read-only; otherwise, false.</value>
        public bool IsReadOnly
        {
            get
            {
                return ((ICollection<T>)this.itemList).IsReadOnly;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="NamedObjectList{T}" /> class.
        /// </summary>
        public NamedObjectList()
        {
            this.nameToItem = new Dictionary<string, T>();
            this.itemList = new LinkedList<T>();
        }

        #endregion


        #region Methods

        private void AddToDictionary(T item)
        {
            if (this.nameToItem.ContainsKey(item.Name))
            {
                throw new LocalizableException(string.Format("Item with name {0} already exists.", item.Name),
                    new InvalidOperationException(String.Format("Item with name {0} already exists.", item.Name)),
                    "Spreadsheet_ErrorExpressions_NamedObjectExistsError",
                    new string[] {item.Name.ToString()});
            }

            this.nameToItem.Add(item.Name, item);
        }

        private void RemoveFromDictionary(T item)
        {
            this.nameToItem.Remove(item.Name);
        }

        private LinkedListNode<T> GetNodeByName(string itemName)
        {
            LinkedListNode<T> node = this.itemList.First;
            while (node != null && node.Value.Name != itemName)
            {
                node = node.Next;
            }

            return node;
        }

        #endregion


        #region Public API

        /// <summary>
        /// Gets item by name.
        /// </summary>
        /// <param name="itemName">Name of the item.</param>
        /// <returns>The item if exists, else the default value of the returned type.</returns>
        public T GetByName(string itemName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(itemName, "itemName");

            T item = default(T);
            if (this.nameToItem.TryGetValue(itemName, out item))
            {
                return item;
            }

            return item;
        }

        /// <summary>
        /// Adds the specified item.
        /// </summary>
        /// <param name="item">The new item.</param>
        void ICollection<T>.Add(T item)
        {
            Guard.ThrowExceptionIfNull(item, "item");

            this.AddToDictionary(item);
            this.itemList.AddLast(item);
        }

        /// <summary>
        /// Adds the new item before an existing item.
        /// </summary>
        /// <param name="nextItemName">Name of the next item.</param>
        /// <param name="item">The new item.</param>
        /// <returns><c>True</c> if succeed.</returns>
        public bool AddBefore(string nextItemName, T item)
        {
            Guard.ThrowExceptionIfNullOrEmpty(nextItemName, "nextItemName");
            Guard.ThrowExceptionIfNull(item, "item");

            LinkedListNode<T> node = this.GetNodeByName(nextItemName);

            if (node != null)
            {
                this.AddToDictionary(item);
                this.itemList.AddBefore(node, item);
                return true;
            }

            return false;
        }

        /// <summary>
        /// Adds the new item after an existing item.
        /// </summary>
        /// <param name="previousItemName">Name of the item after which the new item will be added.</param>
        /// <param name="item">The new item.</param>
        /// <returns><c>True</c> if succeed.</returns>
        public bool AddAfter(string previousItemName, T item)
        {
            Guard.ThrowExceptionIfNullOrEmpty(previousItemName, "previousItemName");
            Guard.ThrowExceptionIfNull(item, "item");

            LinkedListNode<T> node = this.GetNodeByName(previousItemName);

            if (node != null)
            {
                this.AddToDictionary(item);
                this.itemList.AddAfter(node, item);
                return true;
            }

            return false;
        }

        /// <summary>
        /// Adds the item at the start of the collection.
        /// </summary>
        /// <param name="item">The item to add.</param>
        public void AddFirst(T item)
        {
            Guard.ThrowExceptionIfNull(item, "item");

            this.AddToDictionary(item);
            this.itemList.AddFirst(item);
        }

        /// <summary>
        /// Adds the item at the end of the collection.
        /// </summary>
        /// <param name="item">The item to add.</param>
        public void AddLast(T item)
        {
            Guard.ThrowExceptionIfNull(item, "item");

            this.AddToDictionary(item);
            this.itemList.AddLast(item);
        }

        /// <summary>
        /// Determines whether the specified item is contained in the collection.
        /// </summary>
        /// <param name="item">The item to locate in the collection.</param>
        /// <returns><c>True</c> if contained.</returns>
        public bool Contains(T item)
        {
            Guard.ThrowExceptionIfNull(item, "item");

            return this.Contains(item.Name);
        }

        /// <summary>
        /// Determines whether the the specified item name is contained in the collection.
        /// </summary>
        /// <param name="itemName">Name of the item.</param>
        /// <returns><c>True</c> if contained.</returns>
        public bool Contains(string itemName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(itemName, "itemName");

            return this.nameToItem.ContainsKey(itemName);
        }

        /// <summary>
        /// Called just before removing the item.
        /// </summary>
        /// <param name="item">The item to remove.</param>
        protected virtual void RemoveOverride(T item)
        {
        }

        /// <summary>
        /// Removes the specified item.
        /// </summary>
        /// <param name="item">The item to remove.</param>
        /// <returns><c>True</c> if succeed.</returns>
        public bool Remove(T item)
        {
            Guard.ThrowExceptionIfNull(item, "item");

            this.RemoveOverride(item);
            this.RemoveFromDictionary(item);
            return this.itemList.Remove(item);
        }

        /// <summary>
        /// Removes the specified item name.
        /// </summary>
        /// <param name="itemName">Name of the item.</param>
        /// <returns><c>True</c> if succeed.</returns>
        public bool Remove(string itemName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(itemName, "itemName");

            T item;
            if (this.nameToItem.TryGetValue(itemName, out item))
            {
                return this.Remove(item);
            }

            return false;
        }

        /// <summary>
        /// Copies to.
        /// </summary>
        /// <param name="array">The array.</param>
        /// <param name="arrayIndex">Index of the array.</param>
        public void CopyTo(T[] array, int arrayIndex)
        {
            this.itemList.CopyTo(array, arrayIndex);
        }

        /// <summary>
        /// Removes all items from the <see cref="T:System.Collections.Generic.ICollection`1" />.
        /// </summary>
        /// <exception cref="T:System.NotSupportedException">The <see cref="T:System.Collections.Generic.ICollection`1" />
        /// is read-only. </exception>
        public void Clear()
        {
            while(this.itemList.Count > 0)
            {
                this.Remove(this.itemList.First.Value.Name);
            }
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<T> GetEnumerator()
        {
            return this.itemList.GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return ((IEnumerable)this.itemList).GetEnumerator();
        }

        #endregion
    }
}