﻿using System;
using System.Collections;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Core
{
    /// <summary>
    /// Represents collection of named objects.
    /// </summary>
    /// <typeparam name="T">The type of the T.</typeparam>
    public class NamedObjects<T> : IEnumerable<T>
        where T : INamedObject
    {
        #region Fields

        private readonly QueueDictionary<string, T> nameToItem;
        private int suspendChangeEventsCount;

        #endregion


        #region Properties

        /// <summary>
        /// Gets a value indicating whether the change events are suspended.
        /// </summary>
        /// <value>The value indicating whether the change events are suspended.</value>
        private bool AreChangeEventsSuspended
        {
            get
            {
                return this.suspendChangeEventsCount > 0;
            }
        }

        /// <summary>
        /// Gets the count.
        /// </summary>
        /// <value>The count.</value>
        public int Count
        {
            get
            {
                return this.nameToItem.Count;
            }
        }

        #endregion


        #region Indexers

        /// <summary>
        /// Gets the INamedObject at the specified index.
        /// </summary>
        /// <param name="index">The index of the desired element.</param>
        /// <returns>The item at specified index.</returns>
        public T this[int index]
        {
            get
            {
                return this.nameToItem.GetAt(index);
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="NamedObjects{T}" /> class.
        /// </summary>
        public NamedObjects()
        {
            this.nameToItem = new QueueDictionary<string, T>();
        }

        #endregion


        #region Methods

        private void SuspendChangeEvents()
        {
            this.suspendChangeEventsCount++;
        }

        private void ResumeChangeEvents()
        {
            if (this.suspendChangeEventsCount == 0)
            {
                throw new InvalidOperationException("There is no active update to end.");
            }

            this.suspendChangeEventsCount--;
            if (this.suspendChangeEventsCount == 0)
            {
                this.OnChanged();
            }
        }

        /// <summary>
        /// Adds the specified item.
        /// </summary>
        /// <param name="item">The new item.</param>
        public void Add(T item)
        {
            Guard.ThrowExceptionIfNull(item, "item");

            this.nameToItem.Add(item.Name, item);
            this.OnChanged();
        }

        /// <summary>
        /// Replaces the item specified by name with the specified item.
        /// </summary>
        /// <param name="itemName">Name of the item to replace.</param>
        /// <param name="item">The item to replace with.</param>
        public void Replace(string itemName, T item)
        {
            Guard.ThrowExceptionIfNull(item, "item");
            Guard.ThrowExceptionIfNullOrEmpty(itemName, "itemName");

            using (new UpdateScope(this.SuspendChangeEvents, this.ResumeChangeEvents))
            {
                this.Remove(itemName);
                this.Add(item);
            }

            this.OnItemReplaced(itemName, item.Name);

            this.OnChanged();
        }

        /// <summary>
        /// Adds the range of items to the collection.
        /// </summary>
        /// <param name="items">The range of items.</param>
        public void AddRange(IEnumerable<T> items)
        {
            using (new UpdateScope(this.SuspendChangeEvents, this.ResumeChangeEvents))
            {
                foreach (var item in items)
                {
                    this.Add(item);
                }
            }

            this.OnChanged();
        }

        /// <summary>
        /// Tries to get item by name.
        /// </summary>
        /// <param name="name">The name of the searched item.</param>
        /// <param name="result">The result.</param>
        /// <returns><c>True</c> if succeed.</returns>
        public bool TryGetByName(string name, out T result)
        {
            return this.nameToItem.TryGetValue(name, out result);
        }

        /// <summary>
        /// Gets item by name.
        /// </summary>
        /// <param name="name">The name of the searched item.</param>
        /// <returns>The searched item.</returns>
        public T GetByName(string name)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");

            return this.nameToItem[name];
        }

        /// <summary>
        /// Determines whether item with the specified name is contained in the collection.
        /// </summary>
        /// <param name="name">The name of the searched item.</param>
        /// <returns><c>True</c> if contained.</returns>
        public bool Contains(string name)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");
            return this.nameToItem.ContainsKey(name);
        }

        /// <summary>
        /// Determines whether the specified item is contained in the collection.
        /// </summary>
        /// <param name="item">The searched item.</param>
        /// <returns><c>True</c> if contained.</returns>
        public bool Contains(T item)
        {
            Guard.ThrowExceptionIfNull(item, "item");

            T foundItem;
            if (!this.nameToItem.TryGetValue(item.Name, out foundItem))
            {
                return false;
            }

            return TelerikHelper.EqualsOfT(item, foundItem);
        }

        /// <summary>
        /// Removes the item with the specified name.
        /// </summary>
        /// <param name="name">The name of the item.</param>
        public void Remove(string name)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");

            this.nameToItem.Remove(name);
            this.OnChanged();
        }

        /// <summary>
        /// Clears this items from the collection.
        /// </summary>
        public void Clear()
        {
            this.nameToItem.Clear();
            this.OnChanged();
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<T> GetEnumerator()
        {
            foreach (var pair in this.nameToItem)
            {
                yield return pair.Value;
            }
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return ((IEnumerable<T>)this).GetEnumerator();
        }

        #endregion


        #region Event

        /// <summary>
        /// Called when an item is replaced.
        /// </summary>
        /// <param name="replacedItemName">Name of the replaced item.</param>
        /// <param name="newItemName">New name of the item.</param>
        private void OnItemReplaced(string replacedItemName, string newItemName)
        {
            this.OnItemReplaced(new NamedObjectsItemReplaceEventArgs(replacedItemName, newItemName));
        }

        /// <summary>
        /// Occurs when an item is replaced.
        /// </summary>
        public event EventHandler<NamedObjectsItemReplaceEventArgs> ItemReplaced;

        /// <summary>
        /// Raises the <see cref="E:ItemReplaced" /> event.
        /// </summary>
        /// <param name="args">The <see cref="Telerik.Windows.Documents.Spreadsheet.Core.NamedObjectsItemReplaceEventArgs" /> instance containing the event data.</param>
        protected virtual void OnItemReplaced(NamedObjectsItemReplaceEventArgs args)
        {
            if (this.ItemReplaced != null)
            {
                this.ItemReplaced(this, args);
            }
        }

        /// <summary>
        /// Occurs when the collection is changed.
        /// </summary>
        public event EventHandler Changed;

        /// <summary>
        /// Called when the collection is changed.
        /// </summary>
        protected virtual void OnChanged()
        {
            if (this.AreChangeEventsSuspended)
            {
                return;
            }

            if (this.Changed != null)
            {
                this.Changed(this, EventArgs.Empty);
            }
        }

        #endregion
    }
}