﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.DataSeries.AutoFill
{
    internal abstract class NumberRelatedAutoFillGeneratorBase : IAutoFillGenerator
    {
        #region Nested Types

        private class Region
        {
            #region Fields

            private readonly int startIndex;
            private List<double?> numberItems;

            #endregion


            #region Properties

            public int StartIndex
            {
                get
                {
                    return this.startIndex;
                }
            }

            public int EndIndex
            {
                get
                {
                    return startIndex + this.numberItems.Count;
                }
            }

            public List<double?> NumberItems
            {
                get
                {
                    if (this.numberItems == null)
                    {
                        this.numberItems = new List<double?>();
                    }

                    return this.numberItems;
                }
            }

            #endregion


            #region Constructors

            public Region(int startIndex)
            {
                this.startIndex = startIndex;
            }

            #endregion
        }

        #endregion


        #region Properties

        public abstract string Name { get; }
        protected abstract double ValuesRangeUpperBound { get; }
        protected abstract double ValuesRangeLowerBound { get; }
        protected abstract bool AllowReversingOfTheFitBelowTheLowerBoundValue { get; }
        protected virtual int MinAllowedInitialValuesCount
        {
            get
            {
                return 1;
            }
        }

        #endregion


        #region Methods

        public void FillSeries(Worksheet worksheet, CellIndex[] indexes, int initialIndexesCount)
        {
            if (initialIndexesCount < this.MinAllowedInitialValuesCount)
            {
                if (indexes.Length < 2 ||
                    ((indexes[0].Offset(1, 0).Equals(indexes[1]) || indexes[0].Offset(0, 1).Equals(indexes[1])) ||
                    (indexes[1].Offset(1, 0).Equals(indexes[0]) || indexes[1].Offset(0, 1).Equals(indexes[0]))))
                {
                    return;
                }
            }

            int step = initialIndexesCount;
            int resultSeriesLength = indexes.Length;
            ICellValue[] initialValues = new ICellValue[initialIndexesCount];

            for (int i = 0; i < initialIndexesCount; i++)
            {
                ICellValue value = worksheet.Cells[indexes[i]].GetValue().Value;
                initialValues[i] = value;
            }

            double?[] resultInNumbers = this.ExtractNumberValues(initialValues, resultSeriesLength);
            Region[] regions = this.ExtractRegions(initialValues, resultInNumbers);

            bool isGeneratorApplyed = false;
            foreach (var region in regions)
            {
                if (this.IsRegionIncreasing(region) ||
                    this.IsRegionDecreasing(region))
                {
                    this.ApplyGenerator(ref resultInNumbers, region, step);
                    isGeneratorApplyed = true;
                }
            }

            ICellValue[] result = new ICellValue[resultSeriesLength];

            if (isGeneratorApplyed)
            {
                for (int i = 0; i < resultInNumbers.Length; i++)
                {
                    if (resultInNumbers[i].HasValue &&
                        resultInNumbers[i].Value >= this.ValuesRangeLowerBound &&
                        resultInNumbers[i].Value <= this.ValuesRangeUpperBound)
                    {
                        result[i] = this.ValueToResultCellValue(resultInNumbers[i].Value, initialValues[i % initialIndexesCount]);
                    }
                }
            }

            for (int i = 0; i < indexes.Length; i++)
            {
                if (result[i] != null)
                {
                    worksheet.Cells[indexes[i]].SetValueAndExpandToFitNumberValuesWidth(result[i]);
                }
            }
        }

        private double?[] ExtractNumberValues(ICellValue[] initialValues, int resultSeriesLength)
        {
            double?[] result = new double?[resultSeriesLength];
            int numberCounter = 0;

            for (int i = 0; i < initialValues.Length; i++)
            {
                double number;
                if (this.TryParse(initialValues[i], out number))
                {
                    result[numberCounter] = number;
                }
                else
                {
                    result[numberCounter] = null;
                }
                numberCounter++;
            }

            return result;
        }

        private Region[] ExtractRegions(ICellValue[] initialValues, double?[] numberValues)
        {
            List<Region> regions = new List<Region>();
            int valuesCount = initialValues.Length;
            bool isNewRegion = true;
            int numberCounter = 0;
            ICellValue previousCellValue = null;
            double? previousNumberValue = numberValues[0];

            do
            {
                if (this.IsValueFitInDesiredBounds(numberValues[numberCounter]))
                {
                    if (!this.HaveTheSameOrgin(previousCellValue, initialValues[numberCounter], previousNumberValue, numberValues[numberCounter]))
                    {
                        isNewRegion = true;
                    }

                    if (isNewRegion)
                    {
                        Region region = new Region(numberCounter);
                        regions.Add(region);
                        isNewRegion = false;
                    }

                    regions[regions.Count - 1].NumberItems.Add(numberValues[numberCounter]);
                }
                else
                {
                    isNewRegion = true;
                }

                previousCellValue = initialValues[numberCounter];
                previousNumberValue = numberValues[numberCounter];

                numberCounter++;
            }
            while (numberCounter < valuesCount);

            List<Region> resultRegions = new List<Region>();
            foreach (var region in regions)
            {
                bool containsValues = false;
                foreach (var item in region.NumberItems)
                {
                    if (item.HasValue)
                    {
                        containsValues = true;
                        break;
                    }
                }

                if (containsValues)
                {
                    resultRegions.Add(region);
                }
            }

            return resultRegions.ToArray();
        }

        protected virtual bool IsValueFitInDesiredBounds(double? value)
        {
            return value >= this.ValuesRangeLowerBound &&
                   value <= this.ValuesRangeUpperBound;
        }

        private bool CheckSequence(Region region, Func<double, double, bool> checker)
        {
            bool result = true;

            if (region.NumberItems.Count > 1)
            {
                int firstValueIndex = 0;
                while (firstValueIndex < region.NumberItems.Count &&
                    !region.NumberItems[firstValueIndex].HasValue)
                {
                    firstValueIndex++;
                }

                int secondValueIndex = firstValueIndex + 1;
                while (secondValueIndex < region.NumberItems.Count &&
                    !region.NumberItems[secondValueIndex].HasValue)
                {
                    secondValueIndex++;
                }

                int previousValueIndex = firstValueIndex;
                for (int i = secondValueIndex; i < region.NumberItems.Count && !region.NumberItems[i].Equals(-1); i++)
                {
                    if (!region.NumberItems[i].HasValue)
                    {
                        continue;
                    }

                    bool nextIsIncreasing = checker(region.NumberItems[previousValueIndex].Value, region.NumberItems[i].Value);

                    previousValueIndex = i;
                    if (result != nextIsIncreasing)
                    {
                        return false;
                    }
                    result = nextIsIncreasing;
                }
            }

            if (region.NumberItems.Count == 1)
            {
                return true;
            }

            return result;
        }

        private bool IsRegionIncreasing(Region region)
        {
            return CheckSequence(region,
                (velue1, value2) =>
                {
                    return velue1 < value2;
                });
        }

        private bool IsRegionDecreasing(Region region)
        {
            return CheckSequence(region,
                (velue1, value2) =>
                {
                    return velue1 > value2;
                });
        }

        private double?[] RemoveNullsIfNeeded(double?[] initialValues)
        {
            int nullValuesCount = 0;

            var initialValuesList = this.TrimNullValues(initialValues);

            if (initialValuesList.Length < 2)
            {
                return initialValues;
            }

            int lastIndex = 0;

            for (int i = 1; i < initialValuesList.Length; i++)
            {
                if (initialValuesList[i] != null)
                {
                    lastIndex = i;
                    break;
                }

                nullValuesCount++;
            }

            int currentNullValuesCount = 0;
            for (int i = lastIndex + 1; i < initialValuesList.Length; i++)
            {
                if (initialValuesList[i] != null)
                {
                    if (nullValuesCount != currentNullValuesCount)
                    {
                        return initialValues;
                    }

                    currentNullValuesCount = 0;
                }
                else
                {
                    currentNullValuesCount++;
                }
            }

            return initialValuesList.Where(x => x != null).ToArray();
        }

        private int CountLeadingNullValues(double?[] initialValuesList)
        {
            for (int i = 0; i < initialValuesList.Length; i++)
            {
                if (initialValuesList[i] != null)
                {
                    return i;
                }
            }

            return initialValuesList.Length;
        }

        private int CountTrailingNullValues(double?[] initialValuesList)
        {
            for (int i = initialValuesList.Length - 1; i >= 0; i--)
            {
                if (initialValuesList[i] != null)
                {
                    return initialValuesList.Length - i - 1;
                }
            }

            return initialValuesList.Length;
        }

        private double?[] TrimNullValues(double?[] initialValuesList)
        {
            int leadingNullsCount = this.CountLeadingNullValues(initialValuesList);
            int trailingNullsCount = this.CountTrailingNullValues(initialValuesList);

            IEnumerable<double?> result = initialValuesList.Skip(leadingNullsCount);

            result = result.Take(initialValuesList.Length - leadingNullsCount - trailingNullsCount);

            return result.ToArray();
        }

        private void ApplyGenerator(ref double?[] resultValues, Region region, int step)
        {
            int resultSeriesLength = resultValues.Length;
            int repeatsCount = (resultSeriesLength - region.StartIndex) / (region.EndIndex - region.StartIndex);

            double?[] initialValues = region.NumberItems.ToArray();
            double?[] numberValues = this.RemoveNullsIfNeeded(initialValues);
            double linearStepValue = region.NumberItems.Count == 1 ? 1 : FillDataSeriesHelper.GetStepValue(numberValues);

            double?[] linearFitValues = this.ApplyLinearFill(initialValues, numberValues, resultSeriesLength, linearStepValue);

            int valuesCounter = 0;

            for (int i = 0; i < repeatsCount + 1; i++)
            {
                for (int j = 0; j < region.NumberItems.Count; j++)
                {
                    double? nextValue = linearFitValues[valuesCounter];

                    if (nextValue > this.ValuesRangeUpperBound)
                    {
                        if (nextValue % this.ValuesRangeUpperBound == 0)
                        {
                            nextValue = this.ValuesRangeUpperBound;
                        }
                        else
                        {
                            nextValue -= (int)(nextValue / this.ValuesRangeUpperBound) * this.ValuesRangeUpperBound;
                        }
                    }
                    else if (nextValue < this.ValuesRangeLowerBound)
                    {
                        if (!this.AllowReversingOfTheFitBelowTheLowerBoundValue)
                        {
                            nextValue += this.ValuesRangeUpperBound - (int)(nextValue / this.ValuesRangeUpperBound) * this.ValuesRangeUpperBound;
                        }
                        else
                        {
                            double?[] tempLinearFitValues = this.ToNullableDoubleArray(FillDataSeriesHelper.FillLinear(
                                Math.Abs(nextValue.Value),
                                linearFitValues.Length - valuesCounter,
                                Math.Abs(linearStepValue),
                                null));
                            Array.ConstrainedCopy(tempLinearFitValues, 0, linearFitValues, valuesCounter, tempLinearFitValues.Length);
                            nextValue = linearFitValues[valuesCounter];
                        }
                    }

                    int nextValueIndex = region.StartIndex + i * step + j;
                    if (nextValueIndex >= resultSeriesLength)
                    {
                        return;
                    }

                    resultValues[nextValueIndex] = nextValue;
                    valuesCounter++;

                    if (nextValueIndex >= resultSeriesLength - 1)
                    {
                        return;
                    }
                }
            }
        }

        private double?[] ApplyLinearFill(double?[] initialValues, double?[] numberValues, int resultSeriesLength, double linearStepValue)
        {
            bool containsNulls = false;
            foreach (double? value in initialValues)
            {
                if (!value.HasValue)
                {
                    containsNulls = true;
                    break;
                }
            }

            if (containsNulls)
            {
                int leadingNullValues = this.CountLeadingNullValues(initialValues);
                int trailingNullValues = this.CountTrailingNullValues(initialValues);
                int valuesCount = initialValues.Length - leadingNullValues - trailingNullValues;

                List<double?> result = new List<double?>();
                double?[] linearFitValues;

                var trendValues = FillDataSeriesHelper.FillLinearTrend(initialValues, resultSeriesLength + 1);
                var firstTrendValue = trendValues[0];
                var digit = (firstTrendValue * 100000) % 10;

                if (initialValues.Length == numberValues.Length ||
                    (initialValues.Length != numberValues.Length && digit != 0))
                {
                    linearFitValues = this.ToNullableDoubleArray(trendValues);

                    result.AddRange(initialValues);

                    IEnumerable<double?> values = (valuesCount == 1) ? linearFitValues.Skip(valuesCount) : linearFitValues.Skip(valuesCount + leadingNullValues);

                    do
                    {
                        result.AddRange(Enumerable.Repeat<double?>(null, leadingNullValues));
                        result.AddRange(values.Take(valuesCount));
                        result.AddRange(Enumerable.Repeat<double?>(null, trailingNullValues));
                        values = values.Skip(valuesCount);
                    }
                    while (result.Count < resultSeriesLength);

                    result = result.Take(resultSeriesLength + 1).ToList();
                }
                else
                {
                    var initialValue = initialValues.Where(x => x != null).First();
                    linearFitValues = this.ToNullableDoubleArray(FillDataSeriesHelper.FillLinear(initialValue.Value, resultSeriesLength + 1, linearStepValue));

                    result.AddRange(initialValues);

                    double?[] values = linearFitValues.Skip(numberValues.Length).ToArray();

                    int index = 0;
                    for (int i = result.Count; i < resultSeriesLength; i++)
                    {
                        double? newValue = null;
                        if (initialValues[i % initialValues.Length] != null)
                        {
                            newValue = values[index];
                            index++;
                        }

                        result.Add(newValue);
                    }

                    result = result.Take(resultSeriesLength + 1).ToList();
                }

                return result.ToArray();
            }
            else
            {
                return this.ToNullableDoubleArray(FillDataSeriesHelper.FillLinear(initialValues[0].Value, resultSeriesLength + 1, linearStepValue, null));
            }
        }

        private double?[] ToNullableDoubleArray(double[] sourceArray)
        {
            double?[] resultArray = new double?[sourceArray.Length];
            for (int i = 0; i < sourceArray.Length; i++)
            {
                resultArray[i] = sourceArray[i];
            }

            return resultArray;
        }

        protected abstract bool TryParse(ICellValue value, out double number);

        protected abstract ICellValue ValueToResultCellValue(double value, ICellValue firstInitialValue);

        protected virtual bool HaveTheSameOrgin(ICellValue previousCellValue, ICellValue currentCellValue, double? previousNumberValue, double? currentNumberValues)
        {
            if (!previousNumberValue.HasValue || !currentNumberValues.HasValue)
            {
                return false;
            }

            return true;
        }

        #endregion
    }
}