﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.DataSeries
{
    internal static class FillDataSeriesHelper
    {
        #region Methods

        public static double[] FillWithOperator(double initialValue, int resultSeriesLength, double stepValue, Func<double, double, double?> applyStepValue, double? stopValue)
        {
            if (!stopValue.HasValue)
            {
                stopValue = (stepValue >= 0) ? double.MaxValue : double.MinValue;
            }

            List<double> result = new List<double>();

            result.Add(initialValue);

            double? newValue = applyStepValue(result[result.Count - 1], stepValue);

            while (result.Count < resultSeriesLength &&
                   ((stepValue >= 0 && newValue.Value.LessOrEqualDouble(stopValue.Value)) ||
                   (stepValue < 0 && newValue.Value.GreaterOrEqualDouble(stopValue.Value))))
            {
                if (!newValue.HasValue)
                {
                    return null;
                }

                result.Add(newValue.Value);

                newValue = applyStepValue(result.Last(), stepValue);
            }

            return result.ToArray();
        }

        private static bool ContainsNegativeNumbers(double?[] initialValues)
        {
            for (int i = 0; i < initialValues.Length; i++)
            {
                if (initialValues[i].HasValue &&
                    initialValues[i].Value < 0)
                {
                    return true;
                }
            }

            return false;
        }

        #endregion


        #region Linear Fill

        public static double[] FillLinear(double initialValue, int resultSeriesLength, double stepValue, double? stopValue = null)
        {
            return FillWithOperator(initialValue, resultSeriesLength, stepValue, (value, step) => { return value + step; }, stopValue);
        }

        public static double[] FillLinearTrend(double?[] initialValues, int resultSeriesLength)
        {
            var initialNotNullValues = initialValues.Where(x => x != null);
            if (initialValues.Length == 1 ||
                initialNotNullValues.Count() == 1)
            {
                return FillLinear(initialNotNullValues.First().Value, resultSeriesLength, 1, null);
            }

            return InterpolationExtensions.FillLinear(initialValues, resultSeriesLength, 1d, true);
        }

        #endregion


        #region Exponential Fill

        public static double[] FillExponential(double initialValue, int resultSeriesLength, double stepValue, double? stopValue)
        {
            if (stepValue < 0 && stopValue.HasValue)
            {
                return null;
            }

            return FillWithOperator(initialValue, resultSeriesLength, stepValue, (value, step) => { return value * step; }, stopValue);
        }

        public static double[] FillExponentialTrend(double?[] initialValues, int resultSeriesLength)
        {
            if (ContainsNegativeNumbers(initialValues))
            {
                return null;
            }
            else
            {
                return InterpolationExtensions.FillExponential(initialValues, resultSeriesLength, 1d, true);
            }
        }

        #endregion


        #region Date Fill

        private static DateTime AddWeekDays(double days, DateTime dateTime)
        {
            dateTime = dateTime.AddDays(days);
            while (dateTime.DayOfWeek == DayOfWeek.Saturday ||
                   dateTime.DayOfWeek == DayOfWeek.Sunday)
            {
                dateTime = dateTime.AddDays(1);
            }
            return dateTime;
        }

        public static double[] FillDate(double initialValue, DateUnitType dateUnitType, int resultSeriesLength, double stepValue, double? stopValue)
        {
            Func<DateTime, double, double> transformDate = null;
            double lastResidue = 0;

            Func<double, double, double?> operatorFunc = (value, step) =>
            {
                DateTime? dateTime = FormatHelper.ConvertDoubleToDateTime(value);
                if (!dateTime.HasValue)
                {
                    return null;
                }

                // TODO: Rethink this round.
                double result = transformDate(dateTime.Value, step);

                return result;
            };

            switch (dateUnitType)
            {
                case DateUnitType.Day:
                    transformDate =
                        (dateTime, step) =>
                        {
                            double intStep = (int)Math.Floor(step);
                            dateTime = dateTime.AddDays(intStep);
                            double residue = step - intStep;

                            return FormatHelper.ConvertDateTimeToDouble(dateTime) + residue;
                        };
                    break;

                case DateUnitType.Weekday:
                    transformDate =
                        (dateTime, step) =>
                        {
                            int intStep = (int)Math.Floor(step);
                            double residue = step - intStep;

                            dateTime = AddWeekDays(residue, dateTime);

                            for (int i = 0; i < intStep; i++)
                            {
                                dateTime = AddWeekDays(1, dateTime);
                            }

                            return FormatHelper.ConvertDateTimeToDouble(dateTime);
                        };
                    break;

                case DateUnitType.Month:
                    transformDate =
                        (dateTime, step) =>
                        {
                            int intStep = (int)Math.Floor(step);
                            double residue = step - intStep;
                            lastResidue += residue;
                            if (lastResidue >= 1)
                            {
                                lastResidue--;
                                intStep++;
                            }
                            dateTime = dateTime.AddMonths(intStep);

                            return FormatHelper.ConvertDateTimeToDouble(dateTime);
                        };
                    break;

                case DateUnitType.Year:
                    transformDate =
                        (dateTime, step) =>
                        {
                            int intStep = (int)Math.Floor(step);
                            double residue = step - intStep;
                            lastResidue += residue;
                            if (lastResidue >= 1)
                            {
                                lastResidue--;
                                intStep++;
                            }
                            dateTime = dateTime.AddYears(intStep);

                            return FormatHelper.ConvertDateTimeToDouble(dateTime);
                        };
                    break;

                default:
                    throw new InvalidOperationException();
            }

            return FillWithOperator(initialValue, resultSeriesLength, stepValue, operatorFunc, stopValue);
        }

        #endregion


        #region Step Value

        public static double GetStepValue(double?[] initialValues)
        {
            var result = FillLinearTrend(initialValues, initialValues.Length + 2);

            double step = result[result.Length - 1] - result[result.Length - 2];

            return step == 0 ? 1 : step;
        }

        #endregion


        #region Repeat MergedRanges

        public static CellRange[] ExtractMergedRanges(Worksheet worksheet, CellIndex[] indexes, int initialIndexesCount)
        {
            List<CellRange> result = new List<CellRange>();

            for (int i = 0; i < initialIndexesCount; i++)
            {
                CellIndex index = indexes[i];

                CellRange mergedRange;
                if (worksheet.Cells.TryGetContainingMergedRange(index, out mergedRange))
                {
                    if (!result.Contains(mergedRange))
                    {
                        result.Add(mergedRange);
                    }
                }
            }

            return result.ToArray();
        }

        public static void UnmergeMergedRanges(Worksheet worksheet, CellRange[] mergedRanges)
        {
            for(int i = 0; i < mergedRanges.Length; i++)
            {
                CellRange range = mergedRanges[i];
                worksheet.Cells[range].Unmerge();
            }
        }

        private static bool Contains(this CellRange[] ranges, CellIndex cellIndex)
        {
            for(int i = 0; i < ranges.Length; i++)
            {
                CellRange range = ranges[i];
                if (range.Contains(cellIndex))
                {
                    return true;
                }
            }

            return false;
        }

        public static void RepeatMergedRanges(Worksheet worksheet, CellRange[] mergedRanges, CellIndex[] indexes, int initialIndexesCount)
        {
            bool isHorizontalSelection = indexes[indexes.Length - 1].ColumnIndex - indexes[0].ColumnIndex > 0;

            HashSet<CellIndex> mergedIndexes = new HashSet<CellIndex>();

            Action<CellIndex> addToMergedIndexes =
                (cellIndex) =>
                {
                    if (!mergedIndexes.Contains(cellIndex))
                    {
                        mergedIndexes.Add(cellIndex);
                    }
                };

            foreach (CellRange range in mergedRanges)
            {
                for (int i = 0; i < initialIndexesCount; i++)
                {
                    if (!mergedRanges.Contains(indexes[i]))
                    {
                        addToMergedIndexes(indexes[i]);
                    }
                    else if (isHorizontalSelection)
                    {
                        if (range.FromIndex.ColumnIndex == indexes[i].ColumnIndex)
                        {
                            addToMergedIndexes(indexes[i]);
                        }
                    }
                    else
                    {
                        if (range.FromIndex.RowIndex == indexes[i].RowIndex)
                        {
                            addToMergedIndexes(indexes[i]);
                        }
                    }
                }
            }

            int newInitialIndexesCount = mergedIndexes.Count;

            for (int i = initialIndexesCount; i < indexes.Length; i++)
            {
                addToMergedIndexes(indexes[i]);
            }

            CellRange[] ranges = RepeatMergedRanges(mergedRanges, mergedIndexes.ToArray(), newInitialIndexesCount);

            for(int i = 0; i < ranges.Length; i++)
            {
                CellRange range = ranges[i];
                if (range.ColumnCount > 1 ||
                    range.RowCount > 1)
                {
                    if (!worksheet.Cells.MergedCellRanges.IntersectsWithMergedRanges(range))
                    {
                        worksheet.Cells[range].Merge();
                    }
                }
            }
        }

        public static CellRange[] RepeatMergedRanges(CellRange[] mergedRanges, CellIndex[] indexes, int initialIndexesCount, bool isDirectionReversed = false)
        {
            List<CellRange> ranges = new List<CellRange>();
            List<CellIndex> result = new List<CellIndex>();

            if (isDirectionReversed)
            {
                CellIndex mergedIndexFrom = indexes.FirstOrDefault(cellIndex => cellIndex == mergedRanges.First().FromIndex);

                if (mergedIndexFrom != null)
                {
                    result.Add(mergedIndexFrom);
                }
                else
                {
                    result.Add(indexes[0]);
                }
            }
            else
            {
                for (int i = 0; i < initialIndexesCount; i++)
                {
                    result.Add(indexes[i]);
                }
            }

            for (int i = 0; i < result.Count; i++)
            {
                CellIndex mergedRangeToIndex = GetMergedRangeToIndex(mergedRanges, result[i]);
                if (mergedRangeToIndex == null)
                {
                    ranges.Add(new CellRange(result[i], result[i]));
                }
                else
                {
                    ranges.Add(new CellRange(result[i], mergedRangeToIndex));
                }
            }

            if (ranges.Count == 0)
            {
                return new CellRange[0];
            }

            int rowsOffset = GetRowsOffset(indexes);
            int columnsOffset = GetColumnsOffset(indexes);

            CellIndex lastIndex = indexes[indexes.Length - 1];
            CellRange newRange;
            int index = 0;

            do
            {
                CellRange lastRange = ranges.Last();

                CellIndex fromIndex = GetCellRangeFromIndex(rowsOffset, columnsOffset, isDirectionReversed, lastRange);
                if (fromIndex == null)
                {
                    break;
                }
                CellIndex toIndex = GetCellRangeToIndex(fromIndex, ranges[index]);

                newRange = new CellRange(fromIndex, toIndex);

                bool shouldAddToRanges = ShouldAddToRangeCollection(isDirectionReversed, newRange, lastIndex, rowsOffset, columnsOffset);

                if (shouldAddToRanges)
                {
                    ranges.Add(newRange);
                }

                index++;
                index = index % initialIndexesCount;


            } while (!isDirectionReversed ?
                    (newRange.ToIndex.RowIndex < lastIndex.RowIndex ||
                     newRange.ToIndex.ColumnIndex < lastIndex.ColumnIndex) :
                    (newRange.FromIndex.RowIndex > lastIndex.RowIndex ||
                     newRange.FromIndex.ColumnIndex > lastIndex.ColumnIndex));

            //indexes.Clear();
            //foreach (var range in ranges)
            //{
            //    indexes.Add(range.FromIndex);
            //}

            return ranges.ToArray();
        }

        private static CellIndex GetMergedRangeToIndex(CellRange[] mergedRanges, CellIndex result)
        {
            foreach (var range in mergedRanges)
            {
                if (range.Contains(result))
                {
                    return range.ToIndex;
                }
            }

            return null;
        }

        private static int GetColumnsOffset(CellIndex[] indexes)
        {
            int columnsOffset = Math.Abs(indexes[indexes.Length - 1].ColumnIndex - indexes[0].ColumnIndex);

            return (columnsOffset > 0) ? 1 : 0;
        }

        private static int GetRowsOffset(CellIndex[] indexes)
        {
            int rowsOffset = Math.Abs(indexes[indexes.Length - 1].RowIndex - indexes[0].RowIndex);

            return (rowsOffset > 0) ? 1 : 0;
        }

        private static CellIndex GetCellRangeToIndex(CellIndex fromIndex, CellRange cellRange)
        {
            CellIndex toIndex = new CellIndex(fromIndex.RowIndex + cellRange.RowCount - 1,
                                             fromIndex.ColumnIndex + cellRange.ColumnCount - 1);

            return toIndex;
        }

        private static CellIndex GetCellRangeFromIndex(int rowsOffset, int columnsOffset, bool isDirectionReversed, CellRange lastRange)
        {
            CellIndex fromIndex = null;

            if (rowsOffset > 0)
            {
                if (isDirectionReversed)
                {
                    if ((lastRange.FromIndex.RowIndex - lastRange.RowCount - 1 + rowsOffset >= 0))
                    {
                        fromIndex = new CellIndex(lastRange.FromIndex.RowIndex - lastRange.RowCount - 1 + rowsOffset,
                                                  lastRange.FromIndex.ColumnIndex);
                    }
                }
                else
                {
                    fromIndex = new CellIndex(lastRange.FromIndex.RowIndex + lastRange.RowCount - 1 + rowsOffset,
                                              lastRange.FromIndex.ColumnIndex);
                }
            }
            else
            {
                if (isDirectionReversed)
                {
                    if (lastRange.FromIndex.ColumnIndex - lastRange.ColumnCount - 1 + columnsOffset >= 0)
                    {
                        fromIndex = new CellIndex(lastRange.FromIndex.RowIndex,
                                                  lastRange.FromIndex.ColumnIndex - lastRange.ColumnCount - 1 + columnsOffset);
                    }
                }
                else
                {
                    fromIndex = new CellIndex(lastRange.FromIndex.RowIndex,
                                              lastRange.FromIndex.ColumnIndex + lastRange.ColumnCount - 1 + columnsOffset);
                }
            }

            return fromIndex;
        }


        private static bool ShouldAddToRangeCollection(bool isDirectionReversed, CellRange newRange, CellIndex lastIndex, int rowsOffset, int columnsOffset)
        {
            bool shouldAddToRanges = !isDirectionReversed ?
                                    ((newRange.ToIndex.RowIndex <= lastIndex.RowIndex && rowsOffset > 0) ||
                                     (newRange.ToIndex.ColumnIndex <= lastIndex.ColumnIndex && columnsOffset > 0)) :
                                    ((newRange.FromIndex.RowIndex >= lastIndex.RowIndex && rowsOffset > 0) ||
                                     (newRange.FromIndex.ColumnIndex >= lastIndex.ColumnIndex && columnsOffset > 0));
            return shouldAddToRanges;
        }

        public static void RepeatMergedRanges(Worksheet worksheet, CellRange[] mergedCellRanges, CellRange cellRange, int maxInitialValuesCount,
            CellOrientation cellOrientation, bool isDirectionReversed)
        {
            List<CellRange> rangesToMerge = new List<CellRange>();

            int repeatsCount;
            if (cellOrientation == CellOrientation.Horizontal)
            {
                repeatsCount = (int)Math.Ceiling((double)cellRange.ColumnCount / maxInitialValuesCount);
            }
            else
            {
                repeatsCount = (int)Math.Ceiling((double)cellRange.RowCount / maxInitialValuesCount);
            }

            int count = 0;
            do
            {
                for (int i = 0; i < mergedCellRanges.Length; i++)
                {
                    int offset = count * maxInitialValuesCount;
                    CellRange newRange;
                    if (cellOrientation == CellOrientation.Horizontal)
                    {
                        if (!isDirectionReversed)
                        {
                            newRange = mergedCellRanges[i].Offset(0, offset);
                        }
                        else
                        {
                            newRange = mergedCellRanges[i].Offset(0, -offset);
                        }
                    }
                    else
                    {
                        if (!isDirectionReversed)
                        {
                            newRange = mergedCellRanges[i].Offset(offset, 0);
                        }
                        else
                        {
                            newRange = mergedCellRanges[i].Offset(-offset, 0);
                        }
                    }

                    if (cellRange.Contains(newRange))
                    {
                        rangesToMerge.Add(newRange);
                    }
                }

                count++;
            }
            while (count < repeatsCount);

            foreach (var range in rangesToMerge)
            {
                worksheet.Cells[range].Merge();
            }
        }

        #endregion
    }
}