﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Text;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions
{
    /// <summary>
    /// Provides methods and properties for creating an ArrayExpression.
    /// </summary>
    public class ArrayExpression : ConstantExpression<RadExpression[,]>, IEnumerable<RadExpression>
    {
        #region Fields

        private readonly RadExpression[,] array;
        private readonly Dictionary<CultureInfo, string> arrayExpressionToStrings;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the row count of the array.
        /// </summary>
        /// <value>The row count.</value>
        /// <returns>The row count as Int32.</returns>
        public int RowCount
        {
            get
            {
                return this.array.GetLength(0);
            }
        }

        /// <summary>
        /// Gets the column count of the array.
        /// </summary>
        /// <value>The column count.</value>
        /// <returns>The column count as Int32.</returns>
        public int ColumnCount
        {
            get
            {
                return this.array.GetLength(1);
            }
        }

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Expressions.RadExpression" /> at the specified row and column index.
        /// </summary>
        /// <value>The <see cref="Telerik.Windows.Documents.Spreadsheet.Expressions.RadExpression" /> at the specified row and column index.</value>
        /// <returns>The <see cref="Telerik.Windows.Documents.Spreadsheet.Expressions.RadExpression" /> at the specified row and column index.</returns>
        public RadExpression this[int rowIndex, int columnIndex]
        {
            get
            {
                return this.array[rowIndex, columnIndex];
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="ArrayExpression" /> class.
        /// </summary>
        /// <param name="array">The array.</param>
        public ArrayExpression(RadExpression[,] array) 
            : base(array)
        {
            Guard.ThrowExceptionIfNull(array, "array");

            this.arrayExpressionToStrings = new Dictionary<CultureInfo, string>();
            this.array = array;
            this.AttachToChildrenEvent(this);
            this.ToConstant();
        }

        #endregion


        #region Methods

        /// <summary>
        /// Overridable method for getting expression value.
        /// </summary>
        /// <returns>The value as RadExpression.</returns>
        protected override RadExpression GetValueOverride()
        {
            RadExpression[,] resultArray = new RadExpression[this.RowCount, this.ColumnCount];

            for (int rowIndex = 0; rowIndex < this.RowCount; rowIndex++)
            {
                for (int columnIndex = 0; columnIndex < this.ColumnCount; columnIndex++)
                {
                    resultArray[rowIndex, columnIndex] = this[rowIndex, columnIndex].GetValueAsConstantExpression();
                }
            }

            return new ArrayExpression(resultArray);
        }

        internal ErrorExpression GetContainingError()
        {
            for (int i = 0; i < this.RowCount; i++)
            {
                for (int j = 0; j < this.ColumnCount; j++)
                {
                    ErrorExpression error = array[i, j] as ErrorExpression;
                    if (error != null)
                    {
                        return error;
                    }

                    ArrayExpression arrayValue = array[i, j].GetValueAsConstant() as ArrayExpression;
                    if (arrayValue != null)
                    {
                        ErrorExpression errorValue = arrayValue.GetContainingError();
                        if (errorValue != null)
                        {
                            return errorValue;
                        }
                    }
                }
            }

            return null;
        }

        internal ArrayExpression GetRowAsArrayExpression(int rowIndex)
        {
            RadExpression[,] resultArray = new RadExpression[1, this.ColumnCount];

            int columnCount = this.ColumnCount;

            for (int i = 0; i < columnCount; i++)
            {
                resultArray[0, i] = this[rowIndex, i];
            }

            return new ArrayExpression(resultArray);
        }

        internal ArrayExpression GetColumnAsArrayExpression(int columnIndex)
        {
            RadExpression[,] resultArray = new RadExpression[this.RowCount, 1];

            int rowCount = this.RowCount;

            for (int i = 0; i < rowCount; i++)
            {
                resultArray[i, 0] = this[i, columnIndex];
            }

            return new ArrayExpression(resultArray);
        }

        internal override string GetValueAsString(SpreadsheetCultureHelper cultureInfo)
        {
            return this.ToString(cultureInfo);
        }

        internal override string ToString(SpreadsheetCultureHelper cultureInfo)
        {
            string result;
            if (!this.arrayExpressionToStrings.TryGetValue(cultureInfo.CultureInfo, out result))
            {
                result = this.BuildExpressionToString(cultureInfo);
                this.arrayExpressionToStrings[cultureInfo.CultureInfo] = result;
            }

            return result;
        }

        private void ToConstant()
        {
            for (int rowIndex = 0; rowIndex < this.RowCount; rowIndex++)
            {
                for (int columnIndex = 0; columnIndex < this.ColumnCount; columnIndex++)
                {
                    this.array[rowIndex, columnIndex] = this.array[rowIndex, columnIndex].GetValue();
                }
            }
        }

        private string BuildExpressionToString(SpreadsheetCultureHelper cultureInfo)
        {
            StringBuilder builder = new StringBuilder(SpreadsheetCultureHelper.ArrayLeftBracket);

            for (int rowIndex = 0; rowIndex < this.RowCount; rowIndex++)
            {
                for (int columnIndex = 0; columnIndex < this.ColumnCount; columnIndex++)
                {
                    builder.Append(this[rowIndex, columnIndex].ToString(cultureInfo));
                    if (columnIndex != this.ColumnCount - 1)
                    {
                        builder.Append(cultureInfo.ArrayListSeparator);
                    }
                }
                if (rowIndex != this.RowCount - 1)
                {
                    builder.Append(cultureInfo.ArrayRowSeparator);
                }
            }

            builder.Append(SpreadsheetCultureHelper.ArrayRightBracket);

            return builder.ToString();
        }

        internal override bool SimilarEquals(object obj)
        {
            ArrayExpression other = obj as ArrayExpression;

            if (other == null)
            {
                return false;
            }

            if (this.RowCount != other.RowCount ||
                this.ColumnCount != other.ColumnCount)
            {
                return false;
            }

            for (int i = 0; i < this.RowCount; i++)
            {
                for (int j = 0; j < this.ColumnCount; j++)
                {
                    if (!this[i, j].SimilarEquals(other[i, j]))
                    {
                        return false;
                    }
                }
            }

            return true;
        }
      
        #endregion


        #region IEnumerable 

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<RadExpression> GetEnumerator()
        {
            for (int i = 0; i < array.GetLength(0); i++)
            {
                for (int j = 0; j < array.GetLength(1); j++)
                {
                    yield return array[i, j];
                }
            }
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.array.GetEnumerator();
        }

        #endregion
    }
}