﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Expressions.Functions;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions
{
    /// <summary>
    /// Abstract class providing methods and properties for using BinaryOperatorExpression of specified type T.
    /// </summary>
    /// <typeparam name="T">The type T of the operands.</typeparam>
    public abstract class BinaryOperatorExpression<T> : OperatorExpression
    {
        #region Fields

        private readonly RadExpression left;
        private readonly RadExpression right;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <value>The argument conversion rules.</value>
        /// <returns>The argument conversion rules.</returns>
        public virtual ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return ArgumentConversionRules.NumberFunctionConversion;
            }
        }

        /// <summary>
        /// Gets the left operand.
        /// </summary>
        /// <value>The left operand.</value>
        /// <returns>The left operand as RadExpression.</returns>
        public RadExpression Left
        {
            get
            {
                return this.left;
            }
        }

        /// <summary>
        /// Gets the right operand.
        /// </summary>
        /// <value>The right operand.</value>
        /// <returns>The right operand as RadExpression.</returns>
        public RadExpression Right
        {
            get
            {
                return this.right;
            }
        }

        /// <summary>
        /// Gets the type of the operands.
        /// </summary>
        /// <value>The type of the operands.</value>
        /// <returns>The type of the operands as ArgumentType.</returns>
        public abstract ArgumentType OperandsType { get; }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="BinaryOperatorExpression{T}" /> class.
        /// </summary>
        /// <param name="left">The left operand.</param>
        /// <param name="right">The right operand.</param>
        protected BinaryOperatorExpression(RadExpression left, RadExpression right)
        {
            Guard.ThrowExceptionIfNull(left, "left");
            Guard.ThrowExceptionIfNull(right, "right");

            this.left = left;
            this.right = right;

            base.AttachToChildEvent(this.Left);
            base.AttachToChildEvent(this.Right);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Overridable method for getting expression value.
        /// </summary>
        /// <returns>The value as RadExpression.</returns>
        protected sealed override RadExpression GetValueOverride()
        {
            ErrorExpression errorExpression = null;
            List<object> convertedArguments = new List<object>();

            errorExpression = FunctionHelper.TryConvertArgument(this.Left, this.ArgumentConversionRules, this.OperandsType, convertedArguments);

            if (errorExpression != null)
            {
                return errorExpression;
            }

            errorExpression = FunctionHelper.TryConvertArgument(this.Right, this.ArgumentConversionRules, this.OperandsType, convertedArguments);

            if (errorExpression != null)
            {
                return errorExpression;
            }

            T[] operands = new T[convertedArguments.Count];

            for (int i = 0; i < convertedArguments.Count; i++)
            {
                if (OperandsType == ArgumentType.Any)
                {
                    RadExpression constantArgument = (convertedArguments[i] as RadExpression).GetValueAsConstantExpression();

                    if (constantArgument is ErrorExpression)
                    {
                        return constantArgument;
                    }

                    convertedArguments[i] = constantArgument;
                }

                operands[i] = (T)convertedArguments[i];
            }

            return this.GetValueOverride(operands);
        }

        /// <summary>
        /// Gets the value.
        /// </summary>
        /// <param name="operands">The operands of type T.</param>
        /// <returns>Value as RadExpression.</returns>
        protected abstract RadExpression GetValueOverride(T[] operands);

        /// <summary>
        /// Toes the string.
        /// </summary>
        /// <param name="cultureInfo">The culture info.</param>
        /// <returns>The result string.</returns>
        internal override string ToString(SpreadsheetCultureHelper cultureInfo)
        {
            return base.GetOperandAsString(this.Left, cultureInfo) + this.OperatorInfo.Symbol + base.GetOperandAsString(this.Right, cultureInfo);
        }

        internal override void Translate(ExpressionTranslateContext context)
        {
            this.Left.Translate(context);
            this.Right.Translate(context);
        }

        #endregion
    }
}