﻿using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Cache
{
    internal class CellReferenceRangeExpressionCache
    {
        private static readonly int MaxCacheRecordsCount = 5000;
        private readonly Dictionary<CellReferenceRange, CellReferenceRangeExpression> cellReferenceRangeExpressionCache;
        private readonly Dictionary<Worksheet, Dictionary<CellRange, ICellValue[,]>> cellValuesCache;

        private readonly object objectLock = new object();
        private bool isEnabled = false;

        public CellReferenceRangeExpressionCache()
        {
            this.cellReferenceRangeExpressionCache = new Dictionary<CellReferenceRange, CellReferenceRangeExpression>();
            this.cellValuesCache = new Dictionary<Worksheet, Dictionary<CellRange, ICellValue[,]>>();
        }

        public bool IsEnabled
        {
            get
            {
                return this.isEnabled;
            }
            set
            {
                if (this.isEnabled != value)
                {
                    this.isEnabled = value;
                    if (!this.isEnabled)
                    {
                        this.Clear();
                    }
                }
            }
        }

        internal CellReferenceRangeExpression GetOrCreateExpression(CellReferenceRange cellReferenceRange)
        {
            if (!this.isEnabled)
            {
                return new CellReferenceRangeExpression(cellReferenceRange);
            }

            CellReferenceRangeExpression cellReferenceRangeExpression = null;

            lock (this.objectLock)
            {
                this.EnsureCellReferenceRangeExpressionCacheCountLimit();

                if (!this.cellReferenceRangeExpressionCache.TryGetValue(cellReferenceRange, out cellReferenceRangeExpression))
                {
                    cellReferenceRangeExpression = new CellReferenceRangeExpression(cellReferenceRange);
                    this.cellReferenceRangeExpressionCache[cellReferenceRange] = cellReferenceRangeExpression;
                }
            }

            return cellReferenceRangeExpression;
        }

        internal ICellValue[,] GetValuesCache(CellReferenceRange range)
        {
            Worksheet worksheet = range.Worksheet;

            if (worksheet == null)
            {
                return null;
            }

            CellRange cellRange = range.ToCellRange();

            if (!this.isEnabled)
            {
                return new ICellValue[cellRange.RowCount, cellRange.ColumnCount];
            }


            ICellValue[,] result = null;
            lock (this.objectLock)
            {
                this.EnsureCellValuesCacheCountLimit();

                Dictionary<CellRange, ICellValue[,]> childCache = null;
                if (!this.cellValuesCache.TryGetValue(worksheet, out childCache))
                {
                    childCache = new Dictionary<CellRange, ICellValue[,]>();
                    this.cellValuesCache[worksheet] = childCache;
                }

                if (!childCache.TryGetValue(cellRange, out result))
                {
                    result = new ICellValue[cellRange.RowCount, cellRange.ColumnCount];
                    childCache[cellRange] = result;
                }
            }

            return result;
        }

        internal void ClearValuesCache(CellReferenceRange range)
        {
            Worksheet worksheet = range.Worksheet;

            if (worksheet == null)
            {
                return;
            }

            lock (this.objectLock)
            {
                Dictionary<CellRange, ICellValue[,]> childCache = null;
                if (this.cellValuesCache.TryGetValue(worksheet, out childCache))
                {
                    CellRange cellRange = range.ToCellRange();
                    if (childCache.ContainsKey(cellRange))
                    {
                        childCache.Remove(cellRange);
                    }
                }
            }
        }

        private void Clear()
        {
            lock (this.objectLock)
            {
                this.cellReferenceRangeExpressionCache.Clear();
                this.cellValuesCache.Clear();
            }
        }

        private void EnsureCellReferenceRangeExpressionCacheCountLimit()
        {
            lock (this.objectLock)
            {
                if (this.cellReferenceRangeExpressionCache.Count > MaxCacheRecordsCount)
                {
                    this.cellReferenceRangeExpressionCache.Clear();
                }
            }
        }

        private void EnsureCellValuesCacheCountLimit()
        {
            lock (this.objectLock)
            {
                if (this.cellValuesCache.Count > MaxCacheRecordsCount ||
                    this.cellValuesCache.Sum(p => p.Value.Count) > MaxCacheRecordsCount)
                {
                    this.cellValuesCache.Clear();
                }
            }
        }
    }
}