﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Expressions.Cache;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;

#if WPF
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Core.WeakEventManagers;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.Expressions
{
    /// <summary>
    /// Provides methods and properties for using and CellReferenceRangeExpression.
    /// </summary>
    public class CellReferenceRangeExpression : RadExpression
#if WPF
        , IWeakEventListener
#endif
    {
        #region Fields

        private readonly ReadOnlyCollection<CellReferenceRange> cellReferenceRanges;
        private readonly CellReferenceRange firstCellReferenceRange;
        private readonly bool isValid;
        private readonly bool isInRange;
        private readonly bool hasSingleRange;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the cell reference range.
        /// </summary>
        /// <value>The cell reference range.</value>
        /// <returns>The cell reference range as CellReferenceRange.</returns>
        public CellReferenceRange CellReferenceRange
        {
            get
            {
                return this.firstCellReferenceRange;
            }
        }

        /// <summary>
        /// Gets the cell reference ranges.
        /// </summary>
        /// <value>The cell reference ranges.</value>
        public ReadOnlyCollection<CellReferenceRange> CellReferenceRanges
        {
            get
            {
                return this.cellReferenceRanges;
            }
        }

        internal Worksheet Worksheet
        {
            get
            {
                if (!this.IsValid)
                {
                    return null;
                }

                return this.firstCellReferenceRange.Worksheet;
            }
        }

        /// <summary>
        /// Gets a value indicating whether <see cref="CellReferenceRangeExpression" /> instance is valid depending on whether all its ranges are valid and belong to the same sheet.
        /// </summary>
        internal bool IsValid
        {
            get
            {
                return this.isValid &&
                    this.AreCellReferenceRangesInOneWorksheet();
            }
        }

        /// <summary>
        /// Gets a value indicating whether the <see cref="CellReferenceRangeExpression" /> instance is in range which is equivalent to when none of its ranges refers to a cell outside the boundaries of the worksheet.
        /// </summary>
        internal bool IsInRange
        {
            get
            {
                return this.isInRange;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the CellReferenceRanges count is equal to 1.
        /// </summary>
        internal bool HasSingleRange
        {
            get
            {
                return this.hasSingleRange;
            }
        }

        /// <summary>
        /// Gets the array value of the first range.
        /// </summary>
        internal ArrayExpression FirstRangeArray
        {
            get
            {
                ArrayExpression expressionValue = (ArrayExpression)(this.GetValue());
                ArrayExpression firstRangeValue = (ArrayExpression)(expressionValue.Value[0, 0]);

                return firstRangeValue;
            }
        }

        #endregion


        #region Constructors

        internal CellReferenceRangeExpression(Workbook workbook, string worksheetName, bool isWorksheetAbsolute, CellAreaReference cellReferenceRange)
            : this(new CellReferenceRange(workbook, worksheetName, isWorksheetAbsolute, cellReferenceRange))
        {
        }

        internal CellReferenceRangeExpression(CellReferenceRange cellReferenceRange)
            : this(new List<CellReferenceRange>() { cellReferenceRange })
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CellReferenceRangeExpression" /> class.
        /// </summary>
        /// <param name="cellReferenceRanges">The cell reference ranges.</param>
        internal CellReferenceRangeExpression(IEnumerable<CellReferenceRange> cellReferenceRanges)
        {
            Guard.ThrowExceptionIfNull(cellReferenceRanges, "cellReferenceRanges");
            Guard.ThrowExceptionIfLessThan(1, cellReferenceRanges.Count(), "cellReferenceRanges");

            this.cellReferenceRanges = new ReadOnlyCollection<CellReferenceRange>(cellReferenceRanges.ToList());
            this.firstCellReferenceRange = this.cellReferenceRanges[0];
            this.hasSingleRange = this.cellReferenceRanges.Count == 1;

            this.isValid = this.GetIsValid();
            this.isInRange = this.GetIsInRange();
            this.AttachWeekEventPropertyChanged();
        }

        #endregion


        #region Methods

        /// <summary>
        /// Gets the value of the expression.
        /// </summary>
        /// <returns>The value as RadExpression.</returns>
        protected override RadExpression GetValueOverride()
        {
            RadExpression[,] array = new RadExpression[1, this.cellReferenceRanges.Count];

            for (int i = 0; i < this.cellReferenceRanges.Count; i++)
            {
                CellReferenceRange range = this.cellReferenceRanges[i];

                RadExpression expression = this.GetValueOfRange(range);

                array[0, i] = expression;
            }

            return new ArrayExpression(array);
        }

        internal RadExpression GetValueOfRange(CellReferenceRange range)
        {
            if (!range.IsValid)
            {
                return ErrorExpressions.ReferenceError;
            }

            CellAreaReference cellAreaReference = range.CellAreaReference;
            int rowsCount = cellAreaReference.RowsCount;
            int columnsCount = cellAreaReference.ColumnsCount;

            RadExpression[,] expressionCache = new RadExpression[rowsCount, columnsCount];

            CellReference fromCellReference = cellAreaReference.FromCellReference;
            int fromCellReferenceRowIndex = fromCellReference.ActualRowIndex;
            int fromCellReferenceColumnIndex = fromCellReference.ActualColumnIndex;

            CellReference toCellReference = cellAreaReference.ToCellReference;
            int toCellReferenceRowIndex = toCellReference.ActualRowIndex;

            CellsPropertyBag propertyBag = range.Worksheet.Cells.PropertyBag;
            CellReferenceRangeExpressionCache cache = this.Worksheet.Workbook.CellReferenceRangeExpressionCache;

            ICellValue[,] valuesCache = cache.GetValuesCache(range);
            lock (valuesCache)
            {
                bool hasCachedValues = valuesCache[0, 0] != null;

                for (int rowIndex = 0; rowIndex < rowsCount; rowIndex++)
                {
                    for (int columnIndex = 0; columnIndex < columnsCount; columnIndex++)
                    {
                        expressionCache[rowIndex, columnIndex] = EmptyExpression.Empty;
                        if (!hasCachedValues)
                        {
                            valuesCache[rowIndex, columnIndex] = EmptyCellValue.EmptyValue;
                        }
                    }
                }

                for (int columnIndex = 0; columnIndex < range.CellAreaReference.ColumnsCount; columnIndex++)
                {
                    if (!hasCachedValues)
                    {
                        ICompressedList<ICellValue> cellValueCollection = propertyBag.GetPropertyValueCollection(CellPropertyDefinitions.ValueProperty);
                        long fromLongIndex = CellsPropertyBag.ConvertCellIndexToLong(fromCellReferenceRowIndex, columnIndex + fromCellReferenceColumnIndex);
                        long toLongIndex = CellsPropertyBag.ConvertCellIndexToLong(toCellReferenceRowIndex, columnIndex + fromCellReferenceColumnIndex);

                        ICompressedList<ICellValue> currentColumnCollection = cellValueCollection.GetValue(fromLongIndex, toLongIndex);

                        foreach (Range<long, ICellValue> nonDefaultRange in currentColumnCollection.GetNonDefaultRanges())
                        {
                            int fromRowIndex = 0;
                            int toRowIndex = 0;
                            int actualColumnIndex = 0;
                            CellsPropertyBag.ConvertLongToRowAndColumnIndexes(nonDefaultRange.Start, out fromRowIndex, out actualColumnIndex);
                            CellsPropertyBag.ConvertLongToRowAndColumnIndexes(nonDefaultRange.End, out toRowIndex, out actualColumnIndex);

                            ICellValue cellValue = nonDefaultRange.Value;

                            RadExpression resultExpression = this.GetExpression(cellValue);

                            for (int rowIndex = fromRowIndex; rowIndex <= toRowIndex; rowIndex++)
                            {
                                expressionCache[rowIndex - fromCellReferenceRowIndex, columnIndex] = resultExpression;
                                valuesCache[rowIndex - fromCellReferenceRowIndex, columnIndex] = cellValue;
                            }
                        }
                    }
                    else
                    {
                        for (int rowIndex = 0; rowIndex < range.CellAreaReference.RowsCount; rowIndex++)
                        {
                            ICellValue cellValue = null;

                            if (valuesCache != null)
                            {
                                cellValue = valuesCache[rowIndex, columnIndex];
                            }

                            RadExpression resultExpression = this.GetExpression(cellValue);

                            expressionCache[rowIndex, columnIndex] = resultExpression;
                        }
                    }
                }
            }

            return new ArrayExpression(expressionCache);
        }

        private RadExpression GetExpression(ICellValue cellValue)
        {
            RadExpression resultExpression = null;

            IExpressionAwareCellValue expressionAwareCellValue = cellValue as IExpressionAwareCellValue;
            if (expressionAwareCellValue != null)
            {
                resultExpression = expressionAwareCellValue.GetCellExpressionOrNull();
            }

            if (resultExpression == null)
            {
                resultExpression = this.GetExpressionFromValue(cellValue);
            }

            return resultExpression;
        }

        private RadExpression GetExpressionFromValue(ICellValue cellValue)
        {
            RadExpression resultExpression;
            switch (cellValue.ValueType)
            {
                case CellValueType.Empty:
                    resultExpression = EmptyExpression.Empty;
                    break;

                case CellValueType.Number:
                    double numberValue = ((NumberCellValue)cellValue).Value;
                    resultExpression = NumberExpression.CreateValidNumberOrErrorExpression(numberValue);
                    break;

                case CellValueType.Boolean:
                    bool booleanValue = ((BooleanCellValue)cellValue).Value;
                    resultExpression = booleanValue ? BooleanExpression.True : BooleanExpression.False;
                    break;

                case CellValueType.Text:
                    resultExpression = new StringExpression(cellValue.GetValueAsString(CellValueFormat.GeneralFormat));
                    break;

                case CellValueType.Formula:
                    RadExpression expression = this.InterpreteFormulaCellValue((FormulaCellValue)cellValue);

                    resultExpression = expression;
                    break;

                case CellValueType.Error:
                    resultExpression = ErrorExpressions.FindErrorExpression(cellValue.RawValue);
                    break;

                default:
                    // TODO: RichText support.
                    throw new NotImplementedException("TODO:");
            }

            return resultExpression;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            CellReferenceRangeExpression other = obj as CellReferenceRangeExpression;

            if (other == null || this.cellReferenceRanges.Count != other.cellReferenceRanges.Count)
            {
                return false;
            }

            for (int i = 0; i < this.cellReferenceRanges.Count; i++)
            {
                if (!TelerikHelper.EqualsOfT(this.cellReferenceRanges[i], other.cellReferenceRanges[i]))
                {
                    return false;
                }
            }

            return true;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            int hash = this.cellReferenceRanges[0].GetHashCodeOrZero();

            for (int i = 1; i < this.cellReferenceRanges.Count; i++)
            {
                hash = TelerikHelper.CombineHashCodes(hash, this.cellReferenceRanges[i].GetHashCodeOrZero());
            }

            return hash;
        }

        internal override bool SimilarEquals(object obj)
        {
            return this.Equals(obj);
        }

        private RadExpression InterpreteFormulaCellValue(FormulaCellValue formulaCellValue)
        {
            RadExpression formulaExpression = formulaCellValue.Value;

            this.DetachWeekEventValueInvalidated(formulaExpression);
            this.AttachWeekEventValueInvalidated(formulaExpression);

            RadExpression formulaExpressionValue = formulaExpression.GetValue();

            if (formulaExpression is CellReferenceRangeExpression)
            {
                ArrayExpression arrayExpression = formulaExpressionValue as ArrayExpression;
                if (arrayExpression != null && arrayExpression[0, 0] is EmptyExpression)
                {
                    formulaExpressionValue = NumberExpression.Zero;
                }
            }

            if (formulaExpression is BinaryOperatorExpression<CellReferenceRangeExpression> && this.Equals(formulaExpressionValue))
            {
                formulaExpressionValue = ErrorExpressions.CyclicReference;
            }

            return formulaExpressionValue;
        }

        internal override RadExpression CloneAndTranslate(ExpressionCloneAndTranslateContext context)
        {
            if (context.IsInsertingRemovingCells)
            {
                for (int i = 0; i < this.cellReferenceRanges.Count; i++)
                {
                    if (this.cellReferenceRanges[i].ShouldTranslateUponInsertRemoveCells(context))
                    {
                        return this.CloneAndTranslateInternal(context);
                    }
                }

                return this;
            }

            return this.CloneAndTranslateInternal(context);
        }

        private RadExpression CloneAndTranslateInternal(ExpressionCloneAndTranslateContext context)
        {
            CellReferenceRange[] translatedRanges = new CellReferenceRange[this.cellReferenceRanges.Count];

            for (int i = 0; i < this.cellReferenceRanges.Count; i++)
            {
                translatedRanges[i] = this.cellReferenceRanges[i].CloneAndTranslate(context);
            }

            CellReferenceRangeExpression translatedExpression = new CellReferenceRangeExpression(translatedRanges);

            if (context.OldFormatCollection != null && context.NewFormatCollection != null)
            {
                for (int i = 0; i < context.OldFormatCollection.Count; i++)
                {
                    CellReferenceInputString input = context.OldFormatCollection[i] as CellReferenceInputString;
                    if (input != null && object.ReferenceEquals(this, input.Expression))
                    {
                        ((CellReferenceInputString)context.NewFormatCollection[i]).Expression = translatedExpression;
                    }
                }
            }

            return translatedExpression;
        }

        internal override void Translate(ExpressionTranslateContext context)
        {
            bool shouldInvalidate = false;

            for (int i = 0; i < this.cellReferenceRanges.Count; i++)
            {
                if (context.IsWorksheetRenamed && context.RenamedWorksheet.Equals(this.Worksheet))
                {
                    this.cellReferenceRanges[i].Translate();
                    shouldInvalidate = true;
                }
            }

            if (shouldInvalidate)
            {
                this.InvalidateValue();
            }
        }

        /// <summary>
        /// Toes the string.
        /// </summary>
        /// <param name="cultureInfo">The culture info.</param>
        /// <returns></returns>
        internal override string ToString(SpreadsheetCultureHelper cultureInfo)
        {
            return this.BuildToString();
        }

        // TODO: if the expression is in a function, its string should be surrounded by ()
        private string BuildToString()
        {
            StringBuilder builder = new StringBuilder();

            for (int i = 0; i < this.cellReferenceRanges.Count; i++)
            {
                builder.Append(this.cellReferenceRanges[i].ToString());

                if (i < this.cellReferenceRanges.Count - 1)
                {
                    builder.Append(FormatHelper.DefaultSpreadsheetCulture.ListSeparator);
                }
            }

            return builder.ToString();
        }

        internal static bool AreCellReferenceRangesInOneWorksheet(CellReferenceRangeExpression left, CellReferenceRangeExpression right)
        {
            if (left.AreCellReferenceRangesInOneWorksheet() && right.AreCellReferenceRangesInOneWorksheet())
            {
                return TelerikHelper.EqualsOfT(left.firstCellReferenceRange.Worksheet, right.firstCellReferenceRange.Worksheet);
            }

            return false;
        }

        internal static bool TryCreateIntersection(CellReferenceRangeExpression left, CellReferenceRangeExpression right,
            out CellReferenceRangeExpression intersection)
        {
            intersection = null;

            if (!left.IsValid || !right.IsValid || !AreCellReferenceRangesInOneWorksheet(left, right))
            {
                return false;
            }

            List<CellReferenceRange> ranges = new List<CellReferenceRange>();

            for (int i = 0; i < left.cellReferenceRanges.Count; i++)
            {
                for (int j = 0; j < right.cellReferenceRanges.Count; j++)
                {
                    CellAreaReference intersectionReference = CellAreaReference.CreateIntersection(left.cellReferenceRanges[i].CellAreaReference,
                        right.cellReferenceRanges[j].CellAreaReference);

                    if (intersectionReference != null)
                    {
                        ranges.Add(new CellReferenceRange(left.cellReferenceRanges[i].WorksheetReference, intersectionReference));
                    }
                }
            }

            if (ranges.Count > 0)
            {
                intersection = new CellReferenceRangeExpression(ranges);
                return true;
            }

            return false;
        }

        internal static bool TryCreateUnion(CellReferenceRangeExpression left, CellReferenceRangeExpression right,
            out CellReferenceRangeExpression union)
        {
            union = null;

            if (!left.IsValid || !right.IsValid)
            {
                return false;
            }

            if (left.firstCellReferenceRange.IsInRange && right.firstCellReferenceRange.IsInRange &&
                AreCellReferenceRangesInOneWorksheet(left, right))
            {
                List<CellReferenceRange> ranges = new List<CellReferenceRange>();

                ranges.AddRange(left.cellReferenceRanges);
                ranges.AddRange(right.cellReferenceRanges);

                union = new CellReferenceRangeExpression(ranges);

                return true;
            }

            return false;
        }

        internal static bool TryCreateRange(CellReferenceRangeExpression left, CellReferenceRangeExpression right,
            out CellReferenceRangeExpression range)
        {
            range = null;

            if (!left.IsValid || !right.IsValid)
            {
                return false;
            }

            if (!AreCellReferenceRangesInOneWorksheet(left, right) &&
               !(left.firstCellReferenceRange.IsWorksheetAbsolute && !right.firstCellReferenceRange.IsWorksheetAbsolute))
            {
                return false;
            }

            CellAreaReference cellAreaReference = GetRangeArea(left, right);

            CellReferenceRange rangeInWorksheet = new CellReferenceRange(left.firstCellReferenceRange.Workbook,
                left.firstCellReferenceRange.WorksheetName, left.firstCellReferenceRange.IsWorksheetAbsolute, cellAreaReference);

            range = new CellReferenceRangeExpression(rangeInWorksheet);
            return true;
        }

        internal static CellReferenceRangeExpression CreateRangeWithSpecificDimensions(CellReferenceRangeExpression originalRange,
            int rowsCount, int columnsCount)
        {
            Guard.ThrowExceptionIfLessThan(1, rowsCount, "rowsCount");
            Guard.ThrowExceptionIfLessThan(1, columnsCount, "columnsCount");

            CellReference fromCellReference = originalRange.CellReferenceRange.FromCellReference;

            int rowIndex = Math.Min(fromCellReference.ActualRowIndex + rowsCount - 1, SpreadsheetDefaultValues.RowCount);
            int columnIndex = Math.Min(fromCellReference.ActualColumnIndex + columnsCount - 1, SpreadsheetDefaultValues.ColumnCount);

            CellReference toCellReference = new CellReference(
                RowColumnReference.CreateAbsoluteRowColumnReference(rowIndex),
                RowColumnReference.CreateAbsoluteRowColumnReference(columnIndex));

            CellAreaReference cellReferenceRange = new CellAreaReference(fromCellReference, toCellReference);

            return new CellReferenceRangeExpression(originalRange.cellReferenceRanges[0].Workbook, originalRange.cellReferenceRanges[0].WorksheetName,
                originalRange.cellReferenceRanges[0].IsWorksheetAbsolute, cellReferenceRange);
        }

        private void AttachWeekEventPropertyChanged()
        {
            if (this.IsValid)
            {
#if WPF
                CellsPropertyBagPropertyChangedWeakEventManager.AddListener(this.Worksheet.Cells.PropertyBag, this);
#else
                this.Worksheet.Cells.PropertyBag.PropertyChanged += this.Properties_PropertyChanged;
#endif
            }
        }

        private void DetachWeakEventPropertyChanged()
        {
            if (this.IsValid)
            {
#if WPF
                CellsPropertyBagPropertyChangedWeakEventManager.RemoveListener(this.Worksheet.Cells.PropertyBag, this);
#else
                this.Worksheet.Cells.PropertyBag.PropertyChanged -= this.Properties_PropertyChanged;
#endif
            }
        }

#if WPF
        /// <summary>
        /// Destructs the object.
        /// </summary>
        ~CellReferenceRangeExpression()
        {
            this.DetachWeakEventPropertyChanged();
        }
#endif

        private void AttachWeekEventValueInvalidated(RadExpression expression)
        {
#if WPF
            RadExpressionValueChangedWeakEventManager.AddListener(expression, this);
#else
            expression.ValueInvalidated += this.ExpressionCache_ValueInvalidated;
#endif
        }

        private void DetachWeekEventValueInvalidated(RadExpression expression)
        {
#if WPF
            RadExpressionValueChangedWeakEventManager.RemoveListener(expression, this);
#else
            expression.ValueInvalidated -= this.ExpressionCache_ValueInvalidated;
#endif
        }

#if WPF
        /// <summary>
        /// Receives events from the centralized event manager.
        /// </summary>
        /// <param name="managerType">The type of the <see cref="T:System.Windows.WeakEventManager" />
        /// calling this method.</param>
        /// <param name="sender">Object that originated the event.</param>
        /// <param name="e">Event data.</param>
        /// <returns>
        /// true if the listener handled the event. It is considered an error by
        /// the <see cref="T:System.Windows.WeakEventManager" /> handling in WPF to register
        /// a listener for an event that the listener does not handle. Regardless, the method
        /// should return false if it receives an event that it does not recognize or handle.
        /// </returns>
        public bool ReceiveWeakEvent(Type managerType, object sender, EventArgs e)
        {
            if (managerType == typeof(RadExpressionValueChangedWeakEventManager))
            {
                this.ExpressionCache_ValueInvalidated(sender, e);

                return true;
            }

            if (managerType == typeof(CellsPropertyBagPropertyChangedWeakEventManager))
            {
                this.Properties_PropertyChanged(sender, (CellPropertyChangedEventArgs)e);

                return true;
            }

            return false;
        }
#endif

        private void ExpressionCache_ValueInvalidated(object sender, EventArgs e)
        {
            base.InvalidateValue();
        }

        private void Properties_PropertyChanged(object sender, CellPropertyChangedEventArgs e)
        {
            if (e.Property != CellPropertyDefinitions.ValueProperty || !this.IsValid)
            {
                return;
            }

            for (int i = 0; i < this.cellReferenceRanges.Count; i++)
            {
                CellReference fromReference = this.cellReferenceRanges[i].FromCellReference;
                CellReference toReference = this.cellReferenceRanges[i].ToCellReference;

                int fromRowIndex = fromReference.ActualRowIndex;
                int fromColumnIndex = fromReference.ActualColumnIndex;
                int toRowIndex = toReference.ActualRowIndex;
                int toColumnIndex = toReference.ActualColumnIndex;

                if (e.CellRange.IntersectsWith(fromRowIndex, fromColumnIndex, toRowIndex, toColumnIndex))
                {
                    base.InvalidateValue();
                    break;
                }
            }
        }

        /// <summary>
        /// Called when the value is invalidated.
        /// </summary>
        protected override void OnValueInvalidated()
        {
            for (int i = 0; i < this.cellReferenceRanges.Count; i++)
            {
                CellReferenceRange range = this.cellReferenceRanges[i];
                this.Worksheet.Workbook.CellReferenceRangeExpressionCache.ClearValuesCache(range);
            }

            base.OnValueInvalidated();
        }

        private bool AreCellReferenceRangesInOneWorksheet()
        {
            if (this.cellReferenceRanges.Count < 2)
            {
                return true;
            }

            Worksheet worksheet = this.firstCellReferenceRange.Worksheet;

            for (int i = 1; i < this.cellReferenceRanges.Count; i++)
            {
                if (!TelerikHelper.EqualsOfT(worksheet, this.cellReferenceRanges[i].Worksheet))
                {
                    return false;
                }
            }

            return true;
        }

        private bool GetIsInRange()
        {
            for (int i = 0; i < this.cellReferenceRanges.Count; i++)
            {
                if (!this.cellReferenceRanges[i].IsInRange)
                {
                    return false;
                }
            }

            return true;
        }

        private bool GetIsValid()
        {
            for (int i = 0; i < this.cellReferenceRanges.Count; i++)
            {
                if (!this.cellReferenceRanges[i].IsValid)
                {
                    return false;
                }
            }

            return true;
        }

        private CellReference GetTopLeftCellReference()
        {
            RowColumnReference rowReference = this.cellReferenceRanges[0].CellAreaReference.FromCellReference.RowReference;
            RowColumnReference columnReference = this.cellReferenceRanges[0].CellAreaReference.FromCellReference.ColumnReference;

            for (int i = 0; i < this.cellReferenceRanges.Count; i++)
            {
                if (rowReference.ActualIndex > this.cellReferenceRanges[i].CellAreaReference.FromCellReference.ActualRowIndex)
                {
                    rowReference = this.cellReferenceRanges[i].CellAreaReference.FromCellReference.RowReference;
                }

                if (columnReference.ActualIndex > this.cellReferenceRanges[i].CellAreaReference.FromCellReference.ActualColumnIndex)
                {
                    columnReference = this.cellReferenceRanges[i].CellAreaReference.FromCellReference.ColumnReference;
                }
            }

            return new CellReference(rowReference, columnReference);
        }

        private CellReference GetBottomRightCellReference()
        {
            RowColumnReference rowReference = this.cellReferenceRanges[0].CellAreaReference.ToCellReference.RowReference;
            RowColumnReference columnReference = this.cellReferenceRanges[0].CellAreaReference.ToCellReference.ColumnReference;

            for (int i = 0; i < this.cellReferenceRanges.Count; i++)
            {
                if (rowReference.ActualIndex < this.cellReferenceRanges[i].CellAreaReference.ToCellReference.ActualRowIndex)
                {
                    rowReference = this.cellReferenceRanges[i].CellAreaReference.ToCellReference.RowReference;
                }

                if (columnReference.ActualIndex < this.cellReferenceRanges[i].CellAreaReference.ToCellReference.ActualColumnIndex)
                {
                    columnReference = this.cellReferenceRanges[i].CellAreaReference.ToCellReference.ColumnReference;
                }
            }

            return new CellReference(rowReference, columnReference);
        }

        private static CellAreaReference GetRangeArea(CellReferenceRangeExpression left, CellReferenceRangeExpression right)
        {
            CellReference leftFromCellReference = left.GetTopLeftCellReference();
            CellReference leftToCellReference = left.GetBottomRightCellReference();

            CellReference rightFromCellReference = right.GetTopLeftCellReference();
            CellReference rightToCellReference = right.GetBottomRightCellReference();

            CellReference fromCellReference = CellReference.GetTopLeftCellReference(leftFromCellReference, rightFromCellReference);
            CellReference toCellReference = CellReference.GetBottomRightCellReference(leftToCellReference, rightToCellReference);

            return new CellAreaReference(fromCellReference, toCellReference);
        }

        #endregion
    }
}