﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions
{
    /// <summary>
    /// Provides extension methods on RadExpression.
    /// </summary>
    public static class ExpressionExtensions
    {
        /// <summary>
        /// Gets the NumberValue of a RadExpression.
        /// </summary>
        /// <param name="expression">The expression.</param>
        /// <returns>The number value as double. If the expression is not NumberExpression returns 0.</returns>
        public static double NumberValue(this RadExpression expression)
        {
			NumberExpression numberExpression = expression as NumberExpression;
			if (numberExpression != null)
			{
				return numberExpression.Value;
			}

			return 0;
		}

        /// <summary>
        /// Gets the value of a RadExpression as a constant expression.
        /// </summary>
        /// <param name="expression">The expression.</param>
        /// <returns>The ConstantExpression result.</returns>
        public static ConstantExpression GetValueAsConstantExpression(this RadExpression expression)
        {
            return expression.GetValueAsNonArrayConstantExpression(true);
		}

        internal static ConstantExpression GetValueAsNonArrayConstantExpression(this RadExpression expression, bool popupArrayErrors)
        {
            ArrayExpression arrayExpression = expression as ArrayExpression;
            if (arrayExpression != null)
            {
                // TODO: Too much recursion for finding containing error!
                if (popupArrayErrors)
                {
                    ErrorExpression error = arrayExpression.GetContainingError();
                    if (error != null)
                    {
                        return error;
                    }
                }

                return arrayExpression[0, 0].GetValueAsNonArrayConstantExpression(false);
            }

            ConstantExpression constantExpression = expression as ConstantExpression;
            if (constantExpression != null)
            {
                return constantExpression;
            }

            RadExpression expressionValue = expression.GetValue();
            return GetValueAsNonArrayConstantExpression(expressionValue, popupArrayErrors);
        }

        //Calculates the value of a RadExpression and returns either a ConstantExpression or a CellReferenceRangeExpression
        internal static RadExpression GetValueAsConstantOrCellReference(this RadExpression expression)
        {
            if (expression is ConstantExpression || expression is CellReferenceRangeExpression)
            {
                return expression;
            }

            RadExpression expressionValue = expression.GetValue();
            return expressionValue.GetValueAsConstantOrCellReference();
        }

        //Calculates the value of a RadExpression and returns a ConstantExpression. 
        //Unlike the GetValueAsConstantExpression, this method returns the whole ArrayExpression instead of its first element
        internal static ConstantExpression GetValueAsConstant(this RadExpression expression)
        {
            ConstantExpression constantExpression = expression as ConstantExpression;
            if (constantExpression != null)
            {
                return constantExpression;
            }

            RadExpression expressionValue = expression.GetValue();
            return GetValueAsConstant(expressionValue);
        }

        /// <summary>
        /// Gets the integer value of a RadExpression.
        /// </summary>
        /// <param name="expression">The expression.</param>
        /// <exception cref="ArgumentNullException">Thrown when expression is not NumberExpression.</exception>
        /// <returns>The value as integer.</returns>
        public static int IntegerValue(this RadExpression expression)
        {
			NumberExpression numberExpression = expression as NumberExpression;
			Guard.ThrowExceptionIfNull(numberExpression, "numberExpression");

			return numberExpression.IntValue;
		}

        // TODO: Refactoring here.
		private static readonly HashSet<char> AllowedCharsInFunctionOrCellReferenceName = new HashSet<char>
		{
			'.', '_', '!', '$', '[', ']'
        };

        /// <summary>
        /// Determines whether some character is valid part of and function or cell reference name.
        /// </summary>
        /// <param name="character">The character.</param>
        /// <returns>True if valid, otherwise false.</returns>
        public static bool IsValidFunctionOrCellReferenceNamePart(this char character)
        {
			return 
				char.IsLetterOrDigit(character) ||
				AllowedCharsInFunctionOrCellReferenceName.Contains(character);
		}

        /// <summary>
        /// Converts Boolean value to BooleanExpression.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <returns>The result as BooleanExpression.</returns>
        public static BooleanExpression ToBooleanExpression(this bool value)
        {
            return value ? BooleanExpression.True : BooleanExpression.False;
        }

        /// <summary>
        /// Converts CellReferenceRange to CellRange.
        /// </summary>
        /// <param name="cellReferenceRange">The cell reference range.</param>
        /// <returns>The cell range.</returns>
        public static CellRange ToCellRange(this CellReferenceRange cellReferenceRange)
        {
            return new CellRange(cellReferenceRange.FromCellReference.ActualRowIndex,
                cellReferenceRange.FromCellReference.ActualColumnIndex,
                cellReferenceRange.ToCellReference.ActualRowIndex,
                cellReferenceRange.ToCellReference.ActualColumnIndex);
        }
	}
}