﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions
{
    internal class ExpressionQualifierInfo
    {
        #region Fields

        private readonly Workbook workbook;
        private readonly Worksheet currentWorksheet;
        
        private string qualifier;
        private string decodedQualifier;
        private bool isQualifierReferringWorkbook;

        #endregion


        #region Properties

        public Workbook Workbook
        {
            get
            {
                return this.workbook;
            }
        }

        public string Qualifier
        {
            get
            {
                return this.qualifier;
            }
        }

        public Worksheet CurrentWorksheet
        {
            get
            {
                return this.currentWorksheet;
            }
        }

        public string EscapedQualifier
        {
            get
            {
                if (this.IsQualifierReferringWorkbook)
                {
                    return TextHelper.EncodeWorksheetName(this.workbook.Name);
                }

                string referredWorksheetName = this.ReferredWorksheetName;
                if (!string.IsNullOrEmpty(referredWorksheetName))
                {
                    return referredWorksheetName;
                }

                return this.qualifier;
            }
        }

        public string ReferredWorksheetName
        {
            get
            {
                Worksheet referredWorksheet = SpreadsheetHelper.GetWorksheetByName(this.Workbook, this.decodedQualifier);

                if (referredWorksheet != null)
                {
                    return TextHelper.EncodeWorksheetName(referredWorksheet.Name, this.qualifier[0] == '\'');
                }

                return null;
            }
        }

        public bool HasQualifier
        {
            get
            {
                return !string.IsNullOrEmpty(this.qualifier);
            }
        }

        public bool IsQualifierReferringWorkbook
        {
            get
            {
                return this.isQualifierReferringWorkbook;
            }
        }

        #endregion


        #region Constructor

        internal ExpressionQualifierInfo(Workbook workbook, Worksheet currentWorksheet, string qualifier)
        {
            Guard.ThrowExceptionIfNull(workbook, "workbook");
            Guard.ThrowExceptionIfNull(currentWorksheet, "currentWorksheet");

            this.workbook = workbook;
            this.currentWorksheet = currentWorksheet;
            
            this.qualifier = qualifier;
            this.decodedQualifier = TextHelper.DecodeWorksheetName(this.qualifier);
            this.isQualifierReferringWorkbook = this.workbook.Name.Equals(this.decodedQualifier, StringComparison.CurrentCultureIgnoreCase);
        }

        #endregion


        #region Methods

        internal void UpdateQualifier(string newQualifier)
        {
            this.qualifier = newQualifier;
            this.decodedQualifier = TextHelper.DecodeWorksheetName(this.qualifier);
            this.isQualifierReferringWorkbook = this.workbook.Name.Equals(this.decodedQualifier, StringComparison.CurrentCultureIgnoreCase);
        }

        internal ExpressionQualifierInfo Translate(Worksheet worksheet)
        {
            return new ExpressionQualifierInfo(worksheet.Workbook, worksheet, this.qualifier);
        }

        internal ExpressionQualifierInfo TranslateOnWorkbookRenamed()
        {
            if (this.IsQualifierReferringWorkbook)
            {
                return new ExpressionQualifierInfo(workbook, currentWorksheet, workbook.Name);
            }

            return new ExpressionQualifierInfo(this.Workbook, this.CurrentWorksheet, this.Qualifier);
        }

        public override bool Equals(object obj)
        {
            ExpressionQualifierInfo other = obj as ExpressionQualifierInfo;

            return other != null &&
                TelerikHelper.EqualsOfT(this.workbook, other.workbook) &&
                TelerikHelper.EqualsOfT(this.currentWorksheet, other.currentWorksheet) &&
                TelerikHelper.EqualsOfT(this.qualifier, other.qualifier);
        }

        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                    this.workbook.GetHashCode(),
                    this.currentWorksheet.GetHashCode(),
                    this.qualifier.GetHashCodeOrZero());
        }

        #endregion
    }

}
