﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Text;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Expressions.Functions;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions
{
    /// <summary>
    /// Provides methods and properties for using FunctionExpression.
    /// </summary>
    public class FunctionExpression : RadExpression
    {
        #region Fields

        private readonly string functionName;
        private readonly FunctionBase function;
        private readonly RadExpression[] arguments;
        private readonly Worksheet worksheet;
        private readonly int rowIndex;
        private readonly int columnIndex;

        private readonly Dictionary<CultureInfo, string> functionExpressionToStrings;

        #endregion


        #region Properties

        internal FunctionBase Function
        {
            get
            {
                return this.function;
            }
        }

        internal bool IsValid
        {
            get
            {
                return this.function != null;
            }
        }

        #endregion


        #region Constructors

        internal FunctionExpression(string functionName, RadExpression[] arguments, Worksheet worksheet, int rowIndex, int columnIndex)
        {
            Guard.ThrowExceptionIfNullOrEmpty(functionName, "functionName");
            Guard.ThrowExceptionIfNull(arguments, "args");

            this.worksheet = worksheet;
            this.rowIndex = rowIndex;
            this.columnIndex = columnIndex;
            this.functionName = functionName;
            this.arguments = arguments;
            this.function = FunctionManager.GetFunctionByName(this.functionName);

            if (this.IsValid && !this.function.IsArgumentNumberValid(arguments.Length))
            {
                throw new ExpressionException(SpreadsheetStrings.GeneralErrorMessage, "Spreadsheet_ErrorExpressions_GeneralErrorMessage");
            }

            this.functionExpressionToStrings = new Dictionary<CultureInfo, string>();

            base.AttachToChildrenEvent(this.arguments);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Overridable method for getting expression value.
        /// </summary>
        /// <returns>The value as RadExpression.</returns>
        protected override RadExpression GetValueOverride()
        {
            RadExpression result = ErrorExpressions.NameError;

            if (this.function != null)
            {
                FunctionEvaluationContext<RadExpression> context = new FunctionEvaluationContext<RadExpression>(arguments, this.worksheet, this.rowIndex, this.columnIndex);
                result = this.function.Evaluate(context);

                CellReferenceRangeExpression cellReferenceRangeExpression = result as CellReferenceRangeExpression;
                if (cellReferenceRangeExpression != null)
                {
                    var cellReferenceRangeExpressionValueInvalidatedEventListener =
                        new RadWeakEventListener<FunctionExpression, CellReferenceRangeExpression, EventArgs>(this, cellReferenceRangeExpression);

                    cellReferenceRangeExpression.ValueInvalidated += cellReferenceRangeExpressionValueInvalidatedEventListener.OnEvent;

                    cellReferenceRangeExpressionValueInvalidatedEventListener.OnEventAction = CellReferenceRangeExpressionValueInvalidatedWeakEventAction;
                }
            }

            return result;
        }

        private static void CellReferenceRangeExpressionValueInvalidatedWeakEventAction(FunctionExpression instance, object source, EventArgs e)
        {
            instance.InvalidateValue();
        }

        internal override RadExpression CloneAndTranslate(ExpressionCloneAndTranslateContext context)
        {
            RadExpression[] translatedArguments = new RadExpression[this.arguments.Length];

            for (int i = 0; i < this.arguments.Length; i++)
            {
                translatedArguments[i] = this.arguments[i].CloneAndTranslate(context);
            }

            return new FunctionExpression(this.functionName, translatedArguments, context.Worksheet, context.RowIndex, context.ColumnIndex);
        }

        internal override void Translate(ExpressionTranslateContext context)
        {
            for (int i = 0; i < this.arguments.Length; i++)
            {
                this.arguments[i].Translate(context);
            }
        }

        internal override string ToString(SpreadsheetCultureHelper cultureInfo)
        {
            string result;
            if (!this.functionExpressionToStrings.TryGetValue(cultureInfo.CultureInfo, out result))
            {
                result = this.BuildExpressionToString(cultureInfo);
                this.functionExpressionToStrings[cultureInfo.CultureInfo] = result;
            }

            return result;
        }

        private string BuildExpressionToString(SpreadsheetCultureHelper cultureInfo)
        {
            StringBuilder builder = new StringBuilder(this.functionName);
            builder.Append(SpreadsheetCultureHelper.LeftParenthesis);

            for (int i = 0; i < arguments.Length; i++)
            {
                builder.Append(arguments[i].ToString(cultureInfo));

                if (i != arguments.Length - 1)
                {
                    builder.Append(cultureInfo.ListSeparator);
                }
            }

            builder.Append(SpreadsheetCultureHelper.RightParenthesis);

            return builder.ToString();
        }

        internal override bool SimilarEquals(object obj)
        {
            FunctionExpression other = obj as FunctionExpression;

            if (other == null)
            {
                return false;
            }

            if (this.functionName != other.functionName ||
                this.arguments.Length != other.arguments.Length)
            {
                return false;
            }

            for (int i = 0; i < this.arguments.Length; i++)
            {
                if (!this.arguments[i].SimilarEquals(other.arguments[i]))
                {
                    return false;
                }
            }

            return true;
        }

        #endregion
    }
}