﻿using System;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Enumeration describing the way an argument conversion should be interpreted.
    /// </summary>
    public enum ArgumentInterpretation
    {
        /// <summary>
        /// UseAsIs argument interpretation.
        /// </summary>
        UseAsIs,

        /// <summary>
        /// TreatAsError argument interpretation.
        /// </summary>
        TreatAsError,

        /// <summary>
        /// Ignore argument interpretation.
        /// </summary>
        Ignore,

        /// <summary>
        /// ConvertToDefault argument interpretation.
        /// </summary>
        ConvertToDefault
    }

    /// <summary>
    /// Enumeration describing the way an array argument conversion should be interpreted.
    /// </summary>
    public enum ArrayArgumentInterpretation
    {
        /// <summary>
        /// UseFirstElement array argument interpretation.
        /// </summary>
        UseFirstElement,

        /// <summary>
        /// UseFirstElement array argument interpretation.
        /// </summary>
        UseAllElements,

        /// <summary>
        /// UseFirstElement array argument interpretation.
        /// </summary>
        IgnoreAllElements
    }

    /// <summary>
    /// Provides properties for describing the argument conversion rules of some RadSpreadsheet FunctionBase instance.
    /// </summary>
    public class ArgumentConversionRules
    {
        #region Constants

        /// <summary>
        /// Represents number function argument conversion rule.
        /// </summary>
        public static readonly ArgumentConversionRules NumberFunctionConversion = new ArgumentConversionRules(
            nonTextNumberDirectArgument: ArgumentInterpretation.TreatAsError, 
            nonTextNumberIndirectArgument: ArgumentInterpretation.TreatAsError);

        /// <summary>
        /// Represents non boolean function argument conversion rule.
        /// </summary>
        public static readonly ArgumentConversionRules NonBoolNumberFunctionConversion = new ArgumentConversionRules(
            boolDirectArgument: ArgumentInterpretation.TreatAsError,
            boolIndirectArgument: ArgumentInterpretation.TreatAsError,
            nonTextNumberDirectArgument: ArgumentInterpretation.TreatAsError,
            nonTextNumberIndirectArgument: ArgumentInterpretation.TreatAsError);

        /// <summary>
        /// Represents nary number function argument conversion rule.
        /// </summary>
        public static readonly ArgumentConversionRules NaryNumberFunctionConversion = new ArgumentConversionRules(
            nonTextNumberDirectArgument: ArgumentInterpretation.TreatAsError,
            nonTextNumberIndirectArgument: ArgumentInterpretation.TreatAsError,
            arrayArgument: ArrayArgumentInterpretation.UseAllElements);

        /// <summary>
        /// Represents non boolean nary function argument conversion rule.
        /// </summary>
        public static readonly ArgumentConversionRules NonBoolNaryFunctionConversion = new ArgumentConversionRules(
            emptyIndirectArgument: ArgumentInterpretation.Ignore,
            boolDirectArgument: ArgumentInterpretation.TreatAsError,
            boolIndirectArgument : ArgumentInterpretation.TreatAsError,
            nonTextNumberDirectArgument : ArgumentInterpretation.TreatAsError,
            nonTextNumberIndirectArgument : ArgumentInterpretation.TreatAsError,
            arrayArgument : ArrayArgumentInterpretation.UseAllElements);

        /// <summary>
        /// Represents default value number function argument conversion rule.
        /// </summary>
        public static readonly ArgumentConversionRules DefaultValueNumberFunctionConversion = new ArgumentConversionRules(
            textNumberIndirectArgument: ArgumentInterpretation.ConvertToDefault,
            nonTextNumberDirectArgument: ArgumentInterpretation.TreatAsError,
            nonTextNumberIndirectArgument: ArgumentInterpretation.ConvertToDefault,
            arrayArgument : ArrayArgumentInterpretation.UseAllElements);

        /// <summary>
        /// Represents nary ingnore indirect number function argument conversion rule.
        /// </summary>
        public static readonly ArgumentConversionRules NaryIgnoreIndirectNumberFunctionConversion = new ArgumentConversionRules(
            emptyIndirectArgument: ArgumentInterpretation.Ignore,
            boolIndirectArgument: ArgumentInterpretation.Ignore,
            textNumberIndirectArgument: ArgumentInterpretation.Ignore,
            nonTextNumberDirectArgument: ArgumentInterpretation.TreatAsError,
            nonTextNumberIndirectArgument: ArgumentInterpretation.Ignore,
            arrayArgument: ArrayArgumentInterpretation.UseAllElements);

        /// <summary>
        /// Represents boolean function argument conversion rule.
        /// </summary>
        public static readonly ArgumentConversionRules BoolFunctionConversion = new ArgumentConversionRules(
            emptyIndirectArgument: ArgumentInterpretation.Ignore,
            textNumberDirectArgument: ArgumentInterpretation.TreatAsError,
            textNumberIndirectArgument: ArgumentInterpretation.Ignore,
            nonTextNumberDirectArgument: ArgumentInterpretation.TreatAsError,
            nonTextNumberIndirectArgument: ArgumentInterpretation.Ignore,
            arrayArgument: ArrayArgumentInterpretation.UseAllElements);

        #endregion


        #region Fields

        private readonly ArgumentInterpretation emptyDirectArgument;
        private readonly ArgumentInterpretation numberDirectArgument;
        private readonly ArgumentInterpretation boolDirectArgument;
        private readonly ArgumentInterpretation textNumberDirectArgument;
        private readonly ArgumentInterpretation nonTextNumberDirectArgument;

        private readonly ArgumentInterpretation emptyIndirectArgument;
        private readonly ArgumentInterpretation numberIndirectArgument;
        private readonly ArgumentInterpretation boolIndirectArgument;
        private readonly ArgumentInterpretation textNumberIndirectArgument;
        private readonly ArgumentInterpretation nonTextNumberIndirectArgument;

        private readonly ArrayArgumentInterpretation arrayArgument;

        private readonly bool convertDateTextToNumber;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the empty direct argument interpretation.
        /// </summary>
        /// <value>The empty direct argument interpretation.</value>
        /// <returns>The empty direct argument interpretation.</returns>
        public ArgumentInterpretation EmptyDirectArgument
        {
            get
            {
                return this.emptyDirectArgument;
            }
        }

        /// <summary>
        /// Gets the number direct argument interpretation.
        /// </summary>
        /// <value>The number direct argument interpretation.</value>
        /// <returns>The number direct argument interpretation.</returns>
        public ArgumentInterpretation NumberDirectArgument
        {
            get
            {
                return this.numberDirectArgument;
            }
        }

        /// <summary>
        /// Gets the boolean direct argument interpretation.
        /// </summary>
        /// <value>The boolean direct argument interpretation.</value>
        /// <returns>The boolean direct argument interpretation.</returns>
        public ArgumentInterpretation BoolDirectArgument
        {
            get
            {
                return this.boolDirectArgument;
            }
        }

        /// <summary>
        /// Gets the text number direct argument interpretation.
        /// </summary>
        /// <value>The text number direct argument interpretation.</value>
        /// <returns>The text number direct argument interpretation.</returns>
        public ArgumentInterpretation TextNumberDirectArgument
        {
            get
            {
                return this.textNumberDirectArgument;
            }
        }

        /// <summary>
        /// Gets the non text number direct argument interpretation.
        /// </summary>
        /// <value>The non text number direct argument interpretation.</value>
        /// <returns>The non text number direct argument interpretation.</returns>
        public ArgumentInterpretation NonTextNumberDirectArgument
        {
            get
            {
                return this.nonTextNumberDirectArgument;
            }
        }

        /// <summary>
        /// Gets the empty indirect argument interpretation.
        /// </summary>
        /// <value>The empty indirect argument interpretation.</value>
        /// <returns>The empty indirect argument interpretation.</returns>
        public ArgumentInterpretation EmptyIndirectArgument
        {
            get
            {
                return this.emptyIndirectArgument;
            }
        }

        /// <summary>
        /// Gets the number indirect argument interpretation.
        /// </summary>
        /// <value>The number indirect argument interpretation.</value>
        /// <returns>The number indirect argument interpretation.</returns>
        public ArgumentInterpretation NumberIndirectArgument
        {
            get
            {
                return this.numberIndirectArgument;
            }
        }

        /// <summary>
        /// Gets the boolean indirect argument interpretation.
        /// </summary>
        /// <value>The boolean indirect argument interpretation.</value>
        /// <returns>The boolean indirect argument interpretation.</returns>
        public ArgumentInterpretation BoolIndirectArgument
        {
            get
            {
                return this.boolIndirectArgument;
            }
        }

        /// <summary>
        /// Gets the text number indirect argument interpretation.
        /// </summary>
        /// <value>The text number indirect argument interpretation.</value>
        /// <returns>The text number indirect argument interpretation.</returns>
        public ArgumentInterpretation TextNumberIndirectArgument
        {
            get
            {
                return this.textNumberIndirectArgument;
            }
        }

        /// <summary>
        /// Gets the non text number indirect argument interpretation.
        /// </summary>
        /// <value>The non text number indirect argument interpretation.</value>
        /// <returns>The non text number indirect argument interpretation.</returns>
        public ArgumentInterpretation NonTextNumberIndirectArgument
        {
            get
            {
                return this.nonTextNumberIndirectArgument;
            }
        }

        /// <summary>
        /// Gets the array argument interpretation.
        /// </summary>
        /// <value>The array argument interpretation.</value>
        /// <returns>The array argument interpretation.</returns>
        public ArrayArgumentInterpretation ArrayArgument
        {
            get
            {
                return this.arrayArgument;
            }
        }

        /// <summary>
        /// Gets the value which determines whether the text of the argument should be converted as a date serial number or not.
        /// </summary>
        internal bool ConvertDateTextToNumber
        {
            get
            {
                return this.convertDateTextToNumber;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="ArgumentConversionRules" /> class.
        /// </summary>
        /// <param name="conversionRules">The conversion rules concerning the non-array argument interpretation.</param>
        /// <param name="arrayArgument">The array argument interpretation.</param>
        public ArgumentConversionRules(ArgumentConversionRules conversionRules, ArrayArgumentInterpretation arrayArgument) : this(emptyDirectArgument: conversionRules.EmptyDirectArgument,
        numberDirectArgument: conversionRules.NumberDirectArgument,
        boolDirectArgument: conversionRules.BoolDirectArgument,
        textNumberDirectArgument: conversionRules.TextNumberDirectArgument,
        nonTextNumberDirectArgument: conversionRules.NonTextNumberDirectArgument,
        emptyIndirectArgument: conversionRules.EmptyIndirectArgument,
        numberIndirectArgument: conversionRules.NumberIndirectArgument,
        boolIndirectArgument: conversionRules.BoolIndirectArgument,
        textNumberIndirectArgument: conversionRules.TextNumberIndirectArgument,
        nonTextNumberIndirectArgument: conversionRules.NonTextNumberIndirectArgument,
        arrayArgument: arrayArgument)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ArgumentConversionRules" /> class.
        /// </summary>
        /// <param name="emptyDirectArgument">The empty direct argument interpretation.</param>
        /// <param name="numberDirectArgument">The number direct argument interpretation.</param>
        /// <param name="boolDirectArgument">The boolean direct argument interpretation.</param>
        /// <param name="textNumberDirectArgument">The text number direct argument interpretation.</param>
        /// <param name="nonTextNumberDirectArgument">The non text number direct argument interpretation.</param>
        /// <param name="emptyIndirectArgument">The empty indirect argument interpretation.</param>
        /// <param name="numberIndirectArgument">The number indirect argument interpretation.</param>
        /// <param name="boolIndirectArgument">The boolean indirect argument interpretation.</param>
        /// <param name="textNumberIndirectArgument">The text number indirect argument interpretation.</param>
        /// <param name="nonTextNumberIndirectArgument">The non text number indirect argument interpretation.</param>
        /// <param name="arrayArgument">The array argument interpretation.</param>
        public ArgumentConversionRules(ArgumentInterpretation emptyDirectArgument = ArgumentInterpretation.ConvertToDefault,
            ArgumentInterpretation numberDirectArgument = ArgumentInterpretation.UseAsIs,
            ArgumentInterpretation boolDirectArgument = ArgumentInterpretation.UseAsIs,
            ArgumentInterpretation textNumberDirectArgument = ArgumentInterpretation.UseAsIs,
            ArgumentInterpretation nonTextNumberDirectArgument = ArgumentInterpretation.UseAsIs,
            ArgumentInterpretation emptyIndirectArgument = ArgumentInterpretation.UseAsIs,
            ArgumentInterpretation numberIndirectArgument = ArgumentInterpretation.UseAsIs,
            ArgumentInterpretation boolIndirectArgument = ArgumentInterpretation.UseAsIs,
            ArgumentInterpretation textNumberIndirectArgument = ArgumentInterpretation.UseAsIs,
            ArgumentInterpretation nonTextNumberIndirectArgument = ArgumentInterpretation.UseAsIs,
            ArrayArgumentInterpretation arrayArgument = ArrayArgumentInterpretation.UseFirstElement)
        {
            this.emptyDirectArgument = emptyDirectArgument;
            this.numberDirectArgument = numberDirectArgument;
            this.boolDirectArgument = boolDirectArgument;
            this.textNumberDirectArgument = textNumberDirectArgument;
            this.nonTextNumberDirectArgument = nonTextNumberDirectArgument;
            this.emptyIndirectArgument = emptyIndirectArgument;
            this.numberIndirectArgument = numberIndirectArgument;
            this.boolIndirectArgument = boolIndirectArgument;
            this.textNumberIndirectArgument = textNumberIndirectArgument;
            this.nonTextNumberIndirectArgument = nonTextNumberIndirectArgument;
            this.arrayArgument = arrayArgument;
        }


        internal ArgumentConversionRules(ArgumentConversionRules conversionRules, bool convertDateTextToNumber)
            : this(emptyDirectArgument: conversionRules.EmptyDirectArgument,
        numberDirectArgument: conversionRules.NumberDirectArgument,
        boolDirectArgument: conversionRules.BoolDirectArgument,
        textNumberDirectArgument: conversionRules.TextNumberDirectArgument,
        nonTextNumberDirectArgument: conversionRules.NonTextNumberDirectArgument,
        emptyIndirectArgument: conversionRules.EmptyIndirectArgument,
        numberIndirectArgument: conversionRules.NumberIndirectArgument,
        boolIndirectArgument: conversionRules.BoolIndirectArgument,
        textNumberIndirectArgument: conversionRules.TextNumberIndirectArgument,
        nonTextNumberIndirectArgument: conversionRules.NonTextNumberIndirectArgument,
        arrayArgument: conversionRules.ArrayArgument)
        {
            this.convertDateTextToNumber = convertDateTextToNumber;
        }

        #endregion
    }
}
