﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Describes the type of the argument.
    /// </summary>
    public enum ArgumentType
    {
        /// <summary>
        /// Any argument type.
        /// </summary>
        Any,

        /// <summary>
        /// Logical argument type.
        /// </summary>
        Logical,

        /// <summary>
        /// Number argument type.
        /// </summary>
        Number,

        /// <summary>
        /// Text argument type.
        /// </summary>
        Text,

        /// <summary>
        /// Reference argument type.
        /// </summary>
        Reference,

        /// <summary>
        /// Array argument type.
        /// </summary>
        Array
    }

    /// <summary>
    /// Provides methods and properties providing info for an argument.
    /// </summary>
    public class ArgumentInfo
    {
        #region Fields

        private readonly string name;
        private readonly string description;
        private readonly string nameLocalizationKey;
        private readonly string descriptionLocalizationKey;
        private readonly ArgumentType type;
        private readonly bool isRequired;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the argument.
        /// </summary>
        /// <value>The name.</value>
        /// <returns>The name as String.</returns>
        public string Name
        {
            get
            {
                return this.name;
            }
        }

        /// <summary>
        /// Gets the description of the argument.
        /// </summary>
        /// <value>The description.</value>
        /// <returns>The description as String.</returns>
        public string Description
        {
            get
            {
                return this.description;
            }
        }

        /// <summary>
        /// Gets the name localization key.
        /// </summary>
        /// <value>The name localization key.</value>
        public string NameLocalizationKey
        {
            get
            {
                return this.nameLocalizationKey;
            }
        }

        /// <summary>
        /// Gets the description localization key.
        /// </summary>
        /// <value>The description localization key.</value>
        public string DescriptionLocalizationKey
        {
            get
            {
                return this.descriptionLocalizationKey;
            }
        }

        /// <summary>
        /// Gets the type of the argument.
        /// </summary>
        /// <value>The type.</value>
        /// <value>The type as ArgumentType.</value>
        public ArgumentType Type
        {
            get
            {
                return this.type;
            }
        }

        internal bool IsRequired
        {
            get
            {
                return isRequired;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="ArgumentInfo" /> class.
        /// </summary>
        /// <param name="name">The name of the argument.</param>
        /// <param name="description">The description of the argument.</param>
        /// <param name="type">The type type of the argument.</param>
        /// <param name="isRequired">Boolean indicating whether the argument is required or optional. If it is required, it should be set to true.</param>
        /// <param name="nameLocalizationKey">The localization key of the name of the argument.</param>
        /// <param name="descriptionLocalizationKey">The localization key of the description of the argument.</param>
        public ArgumentInfo(string name, string description, ArgumentType type, bool isRequired = true,
            string nameLocalizationKey = null, string descriptionLocalizationKey = null)
        {
            this.name = name;
            this.description = description;
            this.nameLocalizationKey = nameLocalizationKey;
            this.descriptionLocalizationKey = descriptionLocalizationKey;
            this.type = type;
            this.isRequired = isRequired;
        }

        #endregion


        #region Methods

        // TODO: Rethink the usage of string.
        internal bool IsValid(string value)
        {
            value = value.Trim();

            if (string.IsNullOrEmpty(value))
            {
                return !this.IsRequired;
            }

            switch (this.Type)
            {
                case ArgumentType.Any:
                case ArgumentType.Text:
                    return true;

                case ArgumentType.Logical:
                    // TODO: Implement this!
                    throw new NotImplementedException();

                case ArgumentType.Reference:
                    // TODO: Implement this!
                    //return IsReference(value);

                case ArgumentType.Number:
                    // TODO: Numbers can be number values, empty cells, names, arrays, or references 
                    return IsValidNumber(value);// || IsReference(value);
            }

            throw new ArgumentException("Invalid argument");
        }

        private static bool IsValidNumber(string value)
        {
            double num;
            return double.TryParse(value, out num);
        }

        #endregion
    }
}