﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
using System.Text.RegularExpressions;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    internal class CriteriaEvaluator
    {
        #region Fields

        private readonly Worksheet worksheet;
        private readonly RadExpression criteria;
        private readonly Func<RadExpression, int, int, bool> evaluate;

        private static readonly string[] comperissonOperators;

        private static readonly Dictionary<string, Func<double, double, bool>> numberComparer;

        private string comparissonOperator = SpreadsheetCultureHelper.Equal;
        private double? parsedNumber = null;
        private bool? parsedBoolean = null;

        private string criteriaTrimmedFromComparissonOperator = null;
        private bool containsWildCharacter = false;
        private string criteriaTrimmedToFirstWildCharacter = null;
        private string criteriaAsRegex = null;

        #endregion


        #region Constructor

        static CriteriaEvaluator()
        {
            CriteriaEvaluator.comperissonOperators = new string[]
            { 
                SpreadsheetCultureHelper.GreaterThanOrEqualTo, 
                SpreadsheetCultureHelper.LessThanOrEqualTo, 
                SpreadsheetCultureHelper.NotEqual,
                SpreadsheetCultureHelper.GreaterThan, 
                SpreadsheetCultureHelper.LessThan, 
                SpreadsheetCultureHelper.Equal,
            };

            CriteriaEvaluator.numberComparer = new Dictionary<string, Func<double, double, bool>>();
            CriteriaEvaluator.numberComparer[SpreadsheetCultureHelper.GreaterThanOrEqualTo] = (x, y) => { return x >= y; };
            CriteriaEvaluator.numberComparer[SpreadsheetCultureHelper.LessThanOrEqualTo] = (x, y) => { return x <= y; };
            CriteriaEvaluator.numberComparer[SpreadsheetCultureHelper.NotEqual] = (x, y) => { return x != y; };
            CriteriaEvaluator.numberComparer[SpreadsheetCultureHelper.GreaterThan] = (x, y) => { return x > y; };
            CriteriaEvaluator.numberComparer[SpreadsheetCultureHelper.LessThan] = (x, y) => { return x < y; };
            CriteriaEvaluator.numberComparer[SpreadsheetCultureHelper.Equal] = (x, y) => { return x == y; };
        }

        public CriteriaEvaluator(RadExpression criteriaExpression, Worksheet worksheet = null)
        {         
            this.worksheet = worksheet ?? new Workbook().Worksheets.Add();

            this.criteria = criteriaExpression.GetValueAsNonArrayConstantExpression(false);

            if (this.criteria is NumberExpression)
            {
                this.evaluate = this.EvaluateNumberCriteria;
            }
            else if (this.criteria is BooleanExpression)
            {
                this.evaluate = this.EvaluateBooleanCriteria;
            }
            else if (this.criteria is StringExpression)
            {
                this.evaluate = this.EvaluateStringCriteria;
                this.CacheStringParsingValues((this.criteria as StringExpression).Value);
            }
            else if (this.criteria is ErrorExpression)
            {
                this.evaluate = this.EvaluateErrorCriteria;
            }
            else
            {
                this.evaluate = this.EvaluateFalse;
            }
        }

        #endregion


        #region Methods
        
        private void CacheStringParsingValues(string criteriaValue)
        {
            this.comparissonOperator = this.GetComparissonOperator(criteriaValue);

            string trimedFromComparissonOperator = CriteriaEvaluator.TrimFromComparissonOperator(criteriaValue, this.comparissonOperator);

            double parsedNumberValue;
            if (this.TryGetNumberCellValueFromString(0, 0, trimedFromComparissonOperator, out parsedNumberValue))
            {
                this.parsedNumber = parsedNumberValue;
            }

            bool parsedBooleanValue;
            if (bool.TryParse(trimedFromComparissonOperator, out parsedBooleanValue))
            {
                this.parsedBoolean = parsedBooleanValue;
            }

            this.criteriaTrimmedFromComparissonOperator = trimedFromComparissonOperator;

            this.criteriaAsRegex = CriteriaEvaluator.CreateRegexFrom(this.criteriaTrimmedFromComparissonOperator);

            if (this.comparissonOperator != SpreadsheetCultureHelper.Equal)
            {
                int firstWildCharacterIndex = this.IndexOfFirstWildcardCharacter();

                if (firstWildCharacterIndex >= 0)
                {
                    this.containsWildCharacter = true;
                    this.criteriaTrimmedToFirstWildCharacter = this.criteriaTrimmedFromComparissonOperator.Substring(0, firstWildCharacterIndex);
                }
            }
        }

        private int IndexOfFirstWildcardCharacter()
        {
            if(this.criteriaTrimmedFromComparissonOperator[0].IsNonTildeWildcardCharacter())
            {
                return 0;
            }

            for (int i = 1; i < this.criteriaTrimmedFromComparissonOperator.Length; i++)
            {
                if (this.criteriaTrimmedFromComparissonOperator[i].IsNonTildeWildcardCharacter() &&
                    this.criteriaTrimmedFromComparissonOperator[i - 1] != SpreadsheetCultureHelper.Tilde[0])
                {
                    return i;
                }
            }

            return -1;
        }

        private static string CreateRegexFrom(string excelWildcardString)
        {
            return string.Format("^{0}$", excelWildcardString.FromWildcardStringToRegex());
        }

        private bool Compare(string cellValue, string stringComparissonOperator)
        {
            if (stringComparissonOperator == SpreadsheetCultureHelper.Equal)
            {
                return this.GetEqualToComparissonResult(cellValue);
            }
            else if (stringComparissonOperator == SpreadsheetCultureHelper.NotEqual)
            {
                return !this.GetEqualToComparissonResult(cellValue);
            }
            else
            {
                return this.GetInequalityComparissonResult(cellValue, stringComparissonOperator);
            }
        }

        private bool GetInequalityComparissonResult(string cellValue, string stringComparissonOperator)
        {
            if (this.containsWildCharacter 
                && cellValue.Length > this.criteriaTrimmedToFirstWildCharacter.Length
                && cellValue.StartsWith(this.criteriaTrimmedToFirstWildCharacter))
            {
                return stringComparissonOperator.StartsWith(SpreadsheetCultureHelper.GreaterThan);
            }
            else
            {
                string stringComparisson = string.Format("=\"{0}\"{1}\"{2}\"", 
                    CriteriaEvaluator.EscapeQuotations(cellValue), 
                    stringComparissonOperator,
                    CriteriaEvaluator.EscapeQuotations(this.criteriaTrimmedFromComparissonOperator));

                return this.CalculateStringComparissonAsCellValue(stringComparisson);
            }
        }

        private static string EscapeQuotations(string stringCellValue)
        {
            return stringCellValue.Replace("\"", "\"\"");
        }

        private bool CalculateStringComparissonAsCellValue(string stringComparisson)
        {
            bool result;
            if (this.TryGetBooleanCellValueFromString(0, 0, stringComparisson, out result))
            {
                return result;
            }

            return false;
        }

        private bool GetEqualToComparissonResult(string cellValue)
        {
            return Regex.Match(cellValue, this.criteriaAsRegex).Success;
        }

        private static string TrimFromComparissonOperator(string criteriaToTrim, string operatorToTrim)
        {
            if (criteriaToTrim.StartsWith(operatorToTrim))
            {
                return criteriaToTrim.Substring(operatorToTrim.Length);
            }
            else
            {
                return criteriaToTrim;
            }
        }

        private string GetComparissonOperator(string criteria)
        {
            for (int index = 0; index < CriteriaEvaluator.comperissonOperators.Length; index++)
            {
                if(criteria.StartsWith(CriteriaEvaluator.comperissonOperators[index]))
                {
                    return CriteriaEvaluator.comperissonOperators[index];
                }
            }

            return SpreadsheetCultureHelper.Equal;
        }

        public bool Evaluate(RadExpression cellExpression, int rowIndex, int columnIndex)
        {
            return this.evaluate(cellExpression, rowIndex, columnIndex);
        }

        private bool EvaluateNumberCriteria(RadExpression cellExpression, int rowIndex, int columnIndex)
        {
            NumberExpression numberCriteria = this.criteria as NumberExpression;

            NumberExpression numberValue = cellExpression as NumberExpression;
            if (numberValue != null)
            {
                return numberValue.Value == numberCriteria.Value;
            }

            StringExpression stringValue = cellExpression as StringExpression;
            if (stringValue != null)
            {
                double number;
                if (this.TryGetNumberCellValueFromString(rowIndex, columnIndex, stringValue.Value, out number))
                {
                    return numberCriteria.Value == number;
                }
            }

            return false;
        }

        private bool TryGetNumberCellValueFromString(int rowIndex, int columnIndex, string stringValue, out double number)
        {
            ICellValue value = stringValue.ToCellValue(this.worksheet, rowIndex, columnIndex);

            number = 0;
            if (double.TryParse(value.RawValue, out number))
            {
                return true;
            }

            return false;
        }

        private bool TryGetBooleanCellValueFromString(int rowIndex, int columnIndex, string stringValue, out bool boolResult)
        {
            ICellValue value = stringValue.ToCellValue(this.worksheet, rowIndex, columnIndex);

            var formulaValue = value as FormulaCellValue;
            if (formulaValue != null)
            {
                value = formulaValue.GetResultValueAsCellValue();
            }

            boolResult = false;
            if (bool.TryParse(value.RawValue, out boolResult))
            {
                return true;
            }

            return false;
        }

        private bool EvaluateBooleanCriteria(RadExpression cellExpression, int rowIndex, int columnIndex)
        {
            BooleanExpression booleanCriteria = this.criteria as BooleanExpression;

            BooleanExpression boolValue = cellExpression as BooleanExpression;
            if (boolValue != null)
            {
                return booleanCriteria.Value == boolValue.Value;
            }

            return false;
        }

        private bool EvaluateStringCriteria(RadExpression cellExpression, int rowIndex, int columnIndex)
        {
            NumberExpression numberExpression = cellExpression as NumberExpression;
            if (numberExpression != null)
            {
                return this.EvaluateNumberWithStringCriteria(numberExpression.Value);
            }
            
            BooleanExpression booleanExpression = cellExpression as BooleanExpression;
            if (booleanExpression != null)
            {
                return this.EvaluateBooleanWithStringCriteria(booleanExpression.Value);
            }

            StringExpression stringExpression = cellExpression as StringExpression;
            if (stringExpression != null)
            {
                return this.EvaluateStringWithStringCriteria(stringExpression.Value, rowIndex, columnIndex);
            }

            if (cellExpression is EmptyExpression)
            {
                return this.EvaluateEmptyWithStringCriteria();
            }

            return this.StartsWithNotEqualComparissonOperator();
        }

        private bool EvaluateNumberWithStringCriteria(double cellValue)
        {
            if (this.parsedNumber != null)
            {
                return CriteriaEvaluator.numberComparer[this.comparissonOperator](cellValue, (double)this.parsedNumber);
            }

            return this.StartsWithNotEqualComparissonOperator();
        }

        private bool EvaluateBooleanWithStringCriteria(bool cellValue)
        {
            if (this.parsedBoolean != null)
            {
                int criteriaAsNumber = this.parsedBoolean == true ? 1 : 0;
                int cellAsNumber = cellValue ? 1 : 0;

                return CriteriaEvaluator.numberComparer[this.comparissonOperator](cellAsNumber, criteriaAsNumber);
            }

            return this.StartsWithNotEqualComparissonOperator();
        }

        private bool EvaluateStringWithStringCriteria(string cellValue, int rowIndex, int columnIndex)
        {
            if (this.parsedNumber != null)
            {
                if (this.comparissonOperator == SpreadsheetCultureHelper.Equal || this.comparissonOperator == SpreadsheetCultureHelper.NotEqual)
                {
                    double cellNumber;
                    if (TryGetNumberCellValueFromString(rowIndex, columnIndex, cellValue, out cellNumber))
                    {
                        return CriteriaEvaluator.numberComparer[this.comparissonOperator](cellNumber, (double)parsedNumber);
                    }
                }
            }
            else if (this.parsedBoolean != null && this.parsedBoolean.ToString().Length == this.criteriaTrimmedFromComparissonOperator.Length)
            {
                return this.StartsWithNotEqualComparissonOperator();
            }
            else
            {
                return this.Compare(cellValue, this.comparissonOperator);
            }

            return this.StartsWithNotEqualComparissonOperator();
        }

        private bool EvaluateEmptyWithStringCriteria()
        {
            if (this.comparissonOperator == SpreadsheetCultureHelper.Equal)
            {
                return string.IsNullOrEmpty(this.criteriaTrimmedFromComparissonOperator);
            }

            return this.StartsWithNotEqualComparissonOperator();
        }

        private bool EvaluateErrorCriteria(RadExpression cellValue, int rowIndex, int columnIndex)
        {
            return cellValue == criteria;
        }

        private bool EvaluateFalse(RadExpression cellExpression, int rowIndex, int columnIndex)   
        {
            return this.StartsWithNotEqualComparissonOperator();
        }

        private bool StartsWithNotEqualComparissonOperator()
        {
            return this.comparissonOperator == SpreadsheetCultureHelper.NotEqual;
        }

        #endregion
    }
}
