﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the DATE function.
    /// </summary>
    public class Date : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "DATE";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Date()
        {
            string description = "Returns the sequential serial number that represents a particular date. If the cell format was General before the function was entered, the result is formatted as a date.";
            string descriptionKey = "Spreadsheet_Functions_Date_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Year",
                                 "can be one to four digits. Microsoft Excel interprets the year argument according to the date system you are using. By default, Excel for Windows uses the 1900 date system; Excel for the Macintosh uses the 1904 date system.",
                                 ArgumentType.Number,
                                 nameLocalizationKey: "Spreadsheet_Functions_Args_Year",
                                 descriptionLocalizationKey: "Spreadsheet_Functions_Date_YearInfo"),
                new ArgumentInfo("Month",
                                 "is a positive or negative integer representing the month of the year from 1 to 12 (January to December).",
                                 ArgumentType.Number,
                                 nameLocalizationKey: "Spreadsheet_Functions_Args_Month",
                                 descriptionLocalizationKey: "Spreadsheet_Functions_Date_MonthInfo"),
                new ArgumentInfo("Day",
                                 "is a positive or negative integer representing the day of the month from 1 to 31.",
                                 ArgumentType.Number,
                                 nameLocalizationKey: "Spreadsheet_Functions_Args_Day",
                                 descriptionLocalizationKey: "Spreadsheet_Functions_Date_DayInfo"),
            };

            string formatString = FormatHelper.DefaultSpreadsheetCulture.CultureInfo.DateTimeFormat.ShortDatePattern.ToLower();
            Info = new FunctionInfo(FunctionName, FunctionCategory.DateTime, description, requiredArguments, new CellValueFormat(formatString), descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods
        
        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double years = context.Arguments[0];
            double months = context.Arguments[1];
            double days = context.Arguments[2];

            double result;

            if (MathUtility.GetDateSerialNumber(years, months, days, out result))
            {
                if (result >= 0)
                {
                    return NumberExpression.CreateValidNumberOrErrorExpression(result);
                }
            }

            return ErrorExpressions.NumberError;
        }

        #endregion
    }
}