﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the DATEVALUE function.
    /// </summary>
    public class DateValue : StringsInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "DATEVALUE";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static DateValue()
        {
            string description = "Returns the serial number of the date represented by date_text. Use DATEVALUE to convert a date represented by text to a serial number.";
            string descriptionKey = "Spreadsheet_Functions_DateValue_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("DateText",
                                 "is text that represents a date in a Microsoft Excel date format. For example, \"1/30/2008\" or \"30-Jan-2008\" are text strings within quotation marks that represent dates.",
                                 ArgumentType.Text,
                                 nameLocalizationKey: "Spreadsheet_Functions_Args_DateText",
                                 descriptionLocalizationKey: "Spreadsheet_Functions_Datevalue_DateTextInfo")
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.DateTime, description, requiredArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods
        
        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<string> context)
        {
            double dateSerialNumber;

            bool canParseAsDate = TryParseToDate(context.Arguments[0], out dateSerialNumber);

            if (!canParseAsDate)
            {
                return ErrorExpressions.ValueError;
            }

            dateSerialNumber = MathUtility.RoundDown(dateSerialNumber, 0);

            return NumberExpression.CreateValidNumberOrErrorExpression(dateSerialNumber);
        }

        internal static bool TryParseToDate(string value, out double dateSerialNumber)
        {
            string format = string.Empty;
            return  NumberValueParser.TryParse(value, null, null, out format, out dateSerialNumber) &&
                new CellValueFormat(format).FormatStringInfo.FormatType == FormatStringType.DateTime;
        }

        #endregion
    }
}