﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Spreadsheet.Formatting;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the DAYS function.
    /// </summary>
    public class Days : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "DAYS";
        private static readonly FunctionInfo Info;
        private static readonly ArgumentConversionRules argumentConversionRules = new ArgumentConversionRules(ArgumentConversionRules.NumberFunctionConversion, true);

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return argumentConversionRules;
            }
        }

        #endregion


        #region Constructors

        static Days()
        {
            string description = "Returns the number of days between the two dates.";
            string descriptionKey = "Spreadsheet_Functions_Days_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("End_date",
                                 "Start_date and end_date are the two dates between which you want to know the number of days.",
                                 ArgumentType.Number,
                                 nameLocalizationKey: "Spreadsheet_Functions_Args_EndDate",
                                 descriptionLocalizationKey: "Spreadsheet_Functions_Days_StartEndDateInfo"),
                new ArgumentInfo("Start_date",
                                 "Start_date and end_date are the two dates between which you want to know the number of days.",
                                 ArgumentType.Number,
                                 nameLocalizationKey: "Spreadsheet_Functions_Args_StartDate",
                                 descriptionLocalizationKey: "Spreadsheet_Functions_Days_StartEndDateInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.DateTime, description, requiredArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double endDateSerialNumber = context.Arguments[0];
            double startDateSerialNumber = context.Arguments[1];

            if (endDateSerialNumber < 0 || startDateSerialNumber < 0 || 
                endDateSerialNumber > FormatHelper.MaxDoubleValueTranslatableToDateTime ||
                startDateSerialNumber > FormatHelper.MaxDoubleValueTranslatableToDateTime)
            {
                return ErrorExpressions.NumberError;
            }

            int days = (int)endDateSerialNumber - (int)startDateSerialNumber;

            return NumberExpression.CreateValidNumberOrErrorExpression(days);
        }

        #endregion
    }
}
