﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Telerik.Windows.Documents.Spreadsheet.Formatting;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the MONTH function.
    /// </summary>
    public class Month : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "MONTH";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Month()
        {
            string description = "Returns the month of a date represented by a serial number. The month is given as an integer, ranging from 1 (January) to 12 (December).";
            string descriptionKey = "Spreadsheet_Functions_Month_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("SerialNumber",
                                 "is the date of the month you are trying to find. Dates should be entered by using the DATE function, or as results of other formulas or functions.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_SerialNumber",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Month_SerialNumberInfo")
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.DateTime, description, requiredArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods
        
        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double dateSerialNumber = context.Arguments[0];

            if (dateSerialNumber < 0)
            {
                return ErrorExpressions.NumberError;
            }

            DateTime? date = FormatHelper.ConvertDoubleToDateTime(dateSerialNumber);

            if (date == null)
            {
                return ErrorExpressions.NumberError;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(((DateTime)date).Month);
        }

        #endregion
    }
}