﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Telerik.Windows.Documents.Spreadsheet.Formatting;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the SECOND function.
    /// </summary>
    public class Second : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "SECOND";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Second()
        {
            string description = "Returns the seconds of a time value. The second is given as an integer in the range 0 (zero) to 59.";
            string descriptionKey = "Spreadsheet_Functions_Second_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("SerialNumber",
                                 "is the time that contains the minute you want to find. Times may be entered as text strings within quotation marks (for example, \"6:45 PM\"), as decimal numbers (for example, 0.78125, which represents 6:45 PM), or as results of other formulas or functions (for example, TIMEVALUE(\"6:45 PM\")).",
                                 ArgumentType.Number,
                                 nameLocalizationKey: "Spreadsheet_Functions_Args_SerialNumber",
                                 descriptionLocalizationKey: "Spreadsheet_Functions_Second_SerialNumberInfo")
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.DateTime, description, requiredArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods
        
        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double dateSerialNumber = context.Arguments[0];

            if (dateSerialNumber < 0)
            {
                return ErrorExpressions.NumberError;
            }

            DateTime? date = FormatHelper.ConvertDoubleToDateTime(dateSerialNumber);

            if (date == null)
            {
                return ErrorExpressions.NumberError;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(((DateTime)date).Second);
        }

        #endregion
    }
}