﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the TIMEVALUE function.
    /// </summary>
    public class TimeValue : StringsInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "TIMEVALUE";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static TimeValue()
        {
            string description = "Returns the decimal number of the time represented by a text string. The decimal number is a value ranging from 0 (zero) to 0.99999999, representing the times from 0:00:00 (12:00:00 AM) to 23:59:59 (11:59:59 P.M.).";
            string descriptionKey = "Spreadsheet_Functions_TimeValue_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("TimeText",
                                 "is a text string that represents a time in any one of the Microsoft Excel time formats; for example, \"6:45 PM\" and \"18:45\" text strings within quotation marks that represent time.",
                                 ArgumentType.Text,
                                 nameLocalizationKey: "Spreadsheet_Functions_Args_TimeText",
                                 descriptionLocalizationKey: "Spreadsheet_Functions_TimeValue_TimeTextInfo")
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.DateTime, description, requiredArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<string> context)
        {
            double timeSerialNumber;

            string format = string.Empty;
            bool canParseAsDate = NumberValueParser.TryParse(context.Arguments[0], null, null, out format, out timeSerialNumber) &&
                new CellValueFormat(format).FormatStringInfo.FormatType == FormatStringType.DateTime;

            if (!canParseAsDate)
            {
                return ErrorExpressions.ValueError;
            }

            timeSerialNumber = timeSerialNumber % 1;
            return NumberExpression.CreateValidNumberOrErrorExpression(timeSerialNumber);
        }

        #endregion
    }
}