﻿using System;
using System.Collections.Generic;
using System.ComponentModel;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the BITLSHIFT function.
    /// </summary>
    public class BitLShift : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "BITLSHIFT";
        private static readonly FunctionInfo Info;
        private const int BiggestBitIndex = 47;
        private const int MaxShiftAmount = 53;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static BitLShift()
        {
            string description = "Returns a number shifted left by the specified number of bits.";
            string descriptionKey = "Spreadsheet_Functions_BitLShift_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Number",
                                 "Number must be an integer greater than or equal to 0.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Number",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_BitShift_NumberInfo"),
                new ArgumentInfo("Shift_amount",
                                 "Shift_amount must be an integer.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_ShiftAmount",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_BitShift_ShiftAmountInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Engineering, description, requiredArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            long number;
            int shiftAmount;

            bool areValidNumbers = long.TryParse(context.Arguments[0].ToString(), out number);
            areValidNumbers &= int.TryParse(context.Arguments[1].ToString(), out shiftAmount);

            if (!areValidNumbers)
            {
                return ErrorExpressions.NumberError;
            }

            long topNumberBorder = (1L << (BitLShift.BiggestBitIndex + 1)) - 1;

            bool areArgumentsInRange = (0 <= number && number <= topNumberBorder) && (Math.Abs(shiftAmount) <= BitLShift.MaxShiftAmount);

            if (!areArgumentsInRange)
            {
                return Expressions.ErrorExpressions.NumberError;
            }

            bool willResultOverflow = (number != 0) && (MathUtility.IndexOfBiggestBit(number) + shiftAmount > BitLShift.BiggestBitIndex);

            if (willResultOverflow)
            {
                return Expressions.ErrorExpressions.NumberError;
            }

            long result = Telerik.Windows.Documents.Spreadsheet.Maths.EngineeringFunctions.BITLSHIFT(number, shiftAmount);

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

        #endregion
    }

}