﻿using System;
using System.Collections.Generic;
using System.ComponentModel;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the BITOR function.
    /// </summary>
    public class BitOr : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "BITOR";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static BitOr()
        {
            string description = "Returns a bitwise 'OR' of two numbers.";
            string descriptionKey = "Spreadsheet_Functions_BitOr_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Number1",
                                 "Must be in decimal form and greater than or equal to 0.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Number1",
                                 descriptionLocalizationKey: "Spreadsheet_Functions_BitOr_NumberInfo"),
                new ArgumentInfo("Number2",
                                 "Must be in decimal form and greater than or equal to 0.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Number2",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_BitOr_NumberInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Engineering, description, requiredArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            long number1, number2;

            bool areValidNumbers = long.TryParse(context.Arguments[0].ToString(), out number1);
            areValidNumbers &= long.TryParse(context.Arguments[1].ToString(), out number2);

            if (!areValidNumbers)
            {
                return ErrorExpressions.NumberError;
            }

            long topBorder = (1L << 48) - 1;

            bool areArgumentsInRange = (0 <= number1 && number1 <= topBorder) && (0 <= number2 && number2 <= topBorder);

            if (!areArgumentsInRange)
            {
                return Expressions.ErrorExpressions.NumberError;
            }

            long result = Telerik.Windows.Documents.Spreadsheet.Maths.EngineeringFunctions.BITOR(number1, number2);

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

        #endregion
    }
}