﻿using System;
using System.Collections.Generic;
using System.ComponentModel;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the COMPLEX function.
    /// </summary>
    public class Complex : StringsInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "COMPLEX";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Complex()
        {
            string description = "Converts real and imaginary coefficients into a complex number of the form x + yi or x + yj.";
            string descriptionKey = "Spreadsheet_Functions_Complex_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo(
                    "Real_num",
                    "The real coefficient of the complex number.",
                    ArgumentType.Text,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_Real_num",
                    descriptionLocalizationKey:"Spreadsheet_Functions_Complex_Real_numInfo"),

                new ArgumentInfo(
                    "I_num",
                    "The imaginary coefficient of the complex number.",
                    ArgumentType.Text,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_I_num",
                    descriptionLocalizationKey:"Spreadsheet_Functions_Complex_I_numInfo"),
            };
            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo(
                    "Suffix",
                    "The suffix for the imaginary component of the complex number. If omitted, suffix is assumed to be \"i\".",
                    ArgumentType.Text,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_Suffix",
                    descriptionLocalizationKey:"Spreadsheet_Functions_Complex_SuffixInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Engineering, description, requiredArguments, optionalArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<string> context)
        {
            double a, b;
            string i = (context.Arguments.Length < 3 || String.IsNullOrEmpty(context.Arguments[2]))
                ? "i"
                : context.Arguments[2];

            context.Arguments[0] = String.IsNullOrEmpty(context.Arguments[0]) ? "0" : context.Arguments[0];
            context.Arguments[1] = String.IsNullOrEmpty(context.Arguments[1]) ? "0" : context.Arguments[1];

            bool areValidArguments = double.TryParse(context.Arguments[0], out a);
            areValidArguments &= double.TryParse(context.Arguments[1], out b);
            areValidArguments &= (i == "i" || i == "j");

            if (!areValidArguments)
            {
                return ErrorExpressions.ValueError;
            }

            string result = Telerik.Windows.Documents.Spreadsheet.Maths.EngineeringFunctions.COMPLEX(a, b, i);

            return new StringExpression(result);
        }

        #endregion
    }
}