﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the CONVERT function.
    /// </summary>
    public class Convert : FunctionWithArguments
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "CONVERT";
        private static readonly FunctionInfo Info;
        private static readonly ArgumentConversionRules conversionRules = new ArgumentConversionRules(
            nonTextNumberDirectArgument: ArgumentInterpretation.TreatAsError,
            nonTextNumberIndirectArgument: ArgumentInterpretation.TreatAsError,
            textNumberDirectArgument: ArgumentInterpretation.TreatAsError,
            textNumberIndirectArgument: ArgumentInterpretation.TreatAsError,
            boolDirectArgument: ArgumentInterpretation.TreatAsError,
            boolIndirectArgument: ArgumentInterpretation.TreatAsError);

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return Convert.conversionRules;
            }
        }

        #endregion


        #region Constructors

        static Convert()
        {
            string description = "Converts a number from one measurement system to another. For example, CONVERT can translate a table of distances in miles to a table of distances in kilometers.";
            string descriptionKey = "Spreadsheet_Functions_Convert_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo(
                    "Number",
                    "A complex number you want to raise to a power.",
                    ArgumentType.Number,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_Number",
                    descriptionLocalizationKey:"Spreadsheet_Functions_Convert_NumberInfo"),

                new ArgumentInfo(
                    "From_unit",
                    "The units for the Number argument.",
                    ArgumentType.Text,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_From_unit",
                    descriptionLocalizationKey:"Spreadsheet_Functions_Convert_From_unitInfo"),

                new ArgumentInfo(
                    "To_unit",
                    "The units for the CONVERT function's result.",
                    ArgumentType.Text,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_To_unit",
                    descriptionLocalizationKey:"Spreadsheet_Functions_Convert_To_unitInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Engineering, description, requiredArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<object> context)
        {
            try
            {
                double number = (double)context.Arguments[0];
                string fromUnit = (string)context.Arguments[1];
                string toUnit = (string)context.Arguments[2];

                double result = EngineeringFunctions.CONVERT(number, fromUnit, toUnit);

                return NumberExpression.CreateValidNumberOrErrorExpression(result);
            }
            catch (Exception)
            {
                return ErrorExpressions.NotAvailableError;
            }
        }

        #endregion
    }
}