﻿using System;
using System.Collections.Generic;
using System.ComponentModel;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the ERF function.
    /// </summary>
    public class Erf : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "ERF";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return ArgumentConversionRules.NonBoolNumberFunctionConversion;
            }
        }

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Erf()
        {
            string description = "Returns the error function integrated between lower_limit and upper_limit.";
            string descriptionKey = "Spreadsheet_Functions_Erf_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("lower_limit",
                                 "The lower bound for integrating ERF.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_LowerLimit",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Erf_LowerLimitInfo"),
            };
            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("upper_limit",
                                 "The upper bound for integrating ERF. If omitted, ERF integrates between zero and lower_limit.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_UpperLimit",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Erf_UpperLimitInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Engineering, description, requiredArguments, optionalArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double result = context.Arguments.Length == 1 ?
                           Telerik.Windows.Documents.Spreadsheet.Maths.EngineeringFunctions.ERF(context.Arguments[0]) :
                           Telerik.Windows.Documents.Spreadsheet.Maths.EngineeringFunctions.ERF(context.Arguments[0], context.Arguments[1]);

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

        #endregion
    }
}