﻿using System;
using System.Collections.Generic;
using System.ComponentModel;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the GESTEP function.
    /// </summary>
    public class GEStep : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "GESTEP";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return ArgumentConversionRules.NonBoolNumberFunctionConversion;
            }
        }

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static GEStep()
        {
            string description = "Returns 1 if number ≥ step; returns 0 (zero) otherwise. Use this function to filter a set of values. For example, by summing several GESTEP functions you calculate the count of values that exceed a threshold.";
            string descriptionKey = "Spreadsheet_Functions_GEStep_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo(
                    "Number",
                    "The value to test against step.",
                    ArgumentType.Number,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_Number",
                    descriptionLocalizationKey:"Spreadsheet_Functions_GEStep_NumberInfo"),
            };
            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo(
                    "Step",
                    "The threshold value. If you omit a value for step, GESTEP uses zero.",
                    ArgumentType.Number,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_Step",
                    descriptionLocalizationKey:"Spreadsheet_Functions_GEStep_StepInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Engineering, description, requiredArguments, optionalArguments, descriptionLocalizationKey:descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double result = context.Arguments.Length == 1 ?
                          Telerik.Windows.Documents.Spreadsheet.Maths.EngineeringFunctions.GESTEP(context.Arguments[0]) :
                          Telerik.Windows.Documents.Spreadsheet.Maths.EngineeringFunctions.GESTEP(context.Arguments[0], context.Arguments[1]);

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

        #endregion
    }
}