﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the HEX2OCT function.
    /// </summary>
    public class Hex2Oct : StringsInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "HEX2OCT";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return ArgumentConversionRules.NonBoolNumberFunctionConversion;
            }
        }

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Hex2Oct()
        {
            string description = "Converts a hexadecimal number to octal.";
            string descriptionKey = "Spreadsheet_Functions_Hex2Oct_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Number",
                                 "The hexadecimal number you want to convert. Number cannot contain more than 10 characters. The most significant bit of number is the sign bit. The remaining 39 bits are magnitude bits. Negative numbers are represented using two's-complement notation.",
                                 ArgumentType.Text,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Number",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Hex2Oct_NumberInfo"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Places",
                                 "The number of characters to use. If places is omitted, HEX2OCT uses the minimum number of characters necessary. Places is useful for padding the return value with leading 0s (zeros).",
                                 ArgumentType.Text,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Places",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Hex2Oct_PlacesInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Engineering, description, requiredArguments, optionalArguments, descriptionLocalizationKey:descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<string> context)
        {
            string n = context.Arguments[0];

            if (n.Length > 10)
            {
                return Expressions.ErrorExpressions.NumberError;
            }

            if (string.IsNullOrEmpty(n))
            {
                n = "0";
            }

            long places = 0;
            bool hasPlaces = context.Arguments.Length > 1;

            try
            {
                if (hasPlaces)
                {
                    places = double.Parse(context.Arguments[1]).Truncate();

                    if (places <= 0 || places > 10)
                    {
                        return ErrorExpressions.NumberError;
                    }

                    if (EngineeringFunctions.HEX2DEC(n) > 0 && EngineeringFunctions.HEX2OCT(n).Length > places)
                    {
                        return ErrorExpressions.NumberError;
                    }

                    string result = EngineeringFunctions.HEX2OCT(n, (int)places);
                    return new StringExpression(result);
                }
                else
                {
                    string result = Telerik.Windows.Documents.Spreadsheet.Maths.EngineeringFunctions.HEX2OCT(n);
                    return new StringExpression(result);
                }
            }
            catch (Exception)
            {
                return Expressions.ErrorExpressions.NumberError;
            }
        }

        #endregion
    }
}