﻿using System;
using System.Collections.Generic;
using System.ComponentModel;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the IMARGUMENT function.
    /// </summary>
    public class ImArgument : StringsInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "IMARGUMENT";
        private static readonly FunctionInfo Info;
        private const double ZeroPrecision = 1E-300;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static ImArgument()
        {
            string description = "Returns the argument θ (theta), an angle expressed in radians, such that: x+yi = |x+yi|(cos(θ) + i sin(θ))";
            string descriptionKey = "Spreadsheet_Functions_ImArgument_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo(
                    "Inumber",
                    "A complex number for which you want the argument θ (theta).",
                    ArgumentType.Text,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_Inumber",
                    descriptionLocalizationKey:"Spreadsheet_Functions_ImArgument_InumberInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Engineering, description, requiredArguments, descriptionLocalizationKey:descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<string> context)
        {
            try
            {
                double complexModul = Telerik.Windows.Documents.Spreadsheet.Maths.EngineeringFunctions.IMABS(context.Arguments[0]);

                if (complexModul < ImArgument.ZeroPrecision)
                {
                    return ErrorExpressions.DivisionByZero;
                }

                double result = Telerik.Windows.Documents.Spreadsheet.Maths.EngineeringFunctions.IMARGUMENT(context.Arguments[0]);

                return NumberExpression.CreateValidNumberOrErrorExpression(result);
            }
            catch (Exception)
            {
                return ErrorExpressions.NumberError;
            }
        }

        #endregion
    }
}