﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the IMPOWER function.
    /// </summary>
    public class ImPower : FunctionWithArguments
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "IMPOWER";
        private static readonly FunctionInfo Info;
        private static readonly ArgumentConversionRules conversionRules = new ArgumentConversionRules(
            nonTextNumberDirectArgument: ArgumentInterpretation.TreatAsError,
            nonTextNumberIndirectArgument: ArgumentInterpretation.TreatAsError,
            textNumberDirectArgument: ArgumentInterpretation.TreatAsError,
            textNumberIndirectArgument: ArgumentInterpretation.TreatAsError,
            boolDirectArgument: ArgumentInterpretation.TreatAsError,
            boolIndirectArgument: ArgumentInterpretation.TreatAsError);

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return ImPower.conversionRules;
            }
        }

        #endregion


        #region Constructors

        static ImPower()
        {
            string description = "Returns a complex number in x + yi or x + yj text format raised to a power.";
            string descriptionKey = "Spreadsheet_Functions_ImPower_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo(
                    "Inumber",
                    "A complex number you want to raise to a power.",
                    ArgumentType.Text,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_Inumber",
                    descriptionLocalizationKey:"Spreadsheet_Functions_ImPower_InumberInfo"),

                new ArgumentInfo(
                    "Number",
                    "The power to which you want to raise the complex number.",
                    ArgumentType.Number,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_Number",
                    descriptionLocalizationKey:"Spreadsheet_Functions_ImPower_NumberInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Engineering, description, requiredArguments, descriptionLocalizationKey:descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<object> context)
        {
            try
            {
                string complexNumber = (string)context.Arguments[0];
                double power = (double)context.Arguments[1];

                if (EngineeringFunctions.IMABS(complexNumber) == 0)
                {
                    return ErrorExpressions.NumberError;
                }

                string imaginarySymbol = MathUtility.GetImaginarySymbol(complexNumber);
                var complexResult = EngineeringFunctions.IMPOWER(complexNumber, power);
                string result = EngineeringFunctions.COMPLEX(complexResult.Real, complexResult.Imaginary, imaginarySymbol);

                return new StringExpression(result);
            }
            catch (Exception)
            {
                return ErrorExpressions.NumberError;
            }
        }

        #endregion
    }
}