﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the IMPRODUCT function.
    /// </summary>
    public class ImProduct : StringsInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "IMPRODUCT";
        private static readonly FunctionInfo Info;
        private static readonly ArgumentConversionRules argumentConversionRules = new ArgumentConversionRules(
                    nonTextNumberDirectArgument: ArgumentInterpretation.TreatAsError,
                    nonTextNumberIndirectArgument: ArgumentInterpretation.TreatAsError,
                    arrayArgument: ArrayArgumentInterpretation.UseAllElements);

        #endregion


        #region Properties

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return ImProduct.argumentConversionRules;
            }
        }

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static ImProduct()
        {
            string description = "Returns the product of 1 to 255 complex numbers in x + yi or x + yj text format.";
            string descriptionKey = "Spreadsheet_Functions_ImProduct_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo(
                    "Inumber",
                    "Inumber1, Inumber2,... are the complex numbers to multiply.",
                    ArgumentType.Text,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_Inumber",
                    descriptionLocalizationKey:"Spreadsheet_Functions_ImProduct_InumberInfo"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo(
                    "Inumber",
                    "Inumber1, Inumber2,... are the complex numbers to multiply.",
                    ArgumentType.Text,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_Inumber",
                    descriptionLocalizationKey:"Spreadsheet_Functions_ImProduct_InumberInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Engineering, description, requiredArguments, optionalArguments, optionalArgumentsRepeatCount: 254, descriptionLocalizationKey:descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<string> context)
        {
            try
            {
                string imaginarySymbol;

                if (MathUtility.TryGetImaginarySymbol(context.Arguments, out imaginarySymbol))
                {
                    System.Numerics.Complex product = new System.Numerics.Complex(1, 0);

                    for (int i = 0; i < context.Arguments.Length; i++)
                    {
                        System.Numerics.Complex complexMultiplier = context.Arguments[i].ToComplex();
                        product = System.Numerics.Complex.Multiply(product, complexMultiplier);
                    }

                    string result = EngineeringFunctions.COMPLEX(product.Real, product.Imaginary, imaginarySymbol);

                    return new StringExpression(result);
                }
                else
                {
                    return ErrorExpressions.ValueError;
                }
            }
            catch (Exception)
            {
                return ErrorExpressions.NumberError;
            }
        }

        #endregion
    }
}