﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the OCT2BIN function.
    /// </summary>
    public class Oct2Bin : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "OCT2BIN";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return ArgumentConversionRules.NonBoolNumberFunctionConversion;
            }
        }

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Oct2Bin()
        {
            string description = "Converts an octal number to binary.";
            string descriptionKey = "Spreadsheet_Functions_Oct2Bin_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Number",
                                 "The octal number you want to convert. Number may not contain more than 10 characters. The most significant bit of number is the sign bit. The remaining 29 bits are magnitude bits. Negative numbers are represented using two's-complement notation.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Number",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Oct2Bin_NumberInfo"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Places",
                                 "The number of characters to use. If places is omitted, OCT2BIN uses the minimum number of characters necessary. Places is useful for padding the return value with leading 0s (zeros).",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Places",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Oct2Bin_PlacesInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Engineering, description, requiredArguments, optionalArguments, descriptionLocalizationKey:descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            string n = context.Arguments[0].ToString();

            if (string.IsNullOrEmpty(n) || n.Length > 10)
            {
                return Expressions.ErrorExpressions.NumberError;
            }

            long places = 0;
            bool hasPlaces = context.Arguments.Length > 1;

            try
            {
                if (hasPlaces)
                {
                    places = context.Arguments[1].Truncate();

                    if (places <= 0 || places > 10)
                    {
                        return ErrorExpressions.NumberError;
                    }

                    if (EngineeringFunctions.OCT2DEC(n) > 0 && EngineeringFunctions.OCT2BIN(n).Length > places)
                    {
                        return ErrorExpressions.NumberError;
                    }

                    string result = EngineeringFunctions.OCT2BIN(n, (int)places);
                    return new StringExpression(result);
                }
                else
                {
                    string result = Telerik.Windows.Documents.Spreadsheet.Maths.EngineeringFunctions.OCT2BIN(n);
                    return new StringExpression(result);
                }
            }
            catch (Exception)
            {
                return Expressions.ErrorExpressions.NumberError;
            }
        }

        #endregion
    }
}