﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the ACCRINTM function.
    /// </summary>
    public class AccrIntM : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "ACCRINTM";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return ArgumentConversionRules.NonBoolNumberFunctionConversion;
            }
        }

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static AccrIntM()
        {
            string description = "Returns the accrued interest for a security that pays interest at maturity.";
            string descriptionKey = "Spreadsheet_Functions_AccrIntM_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Issue",
                                 "is the security's issue date, expressed as a serial date number.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Issue",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_AccrIntM_IssueInfo"),
                new ArgumentInfo("Settlement",
                                 "is the security's maturity date, expressed as a serial date number.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Settlement",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_AccrIntM_SettlementInfo"),
                new ArgumentInfo("Rate",
                                 "is the security's annual coupon rate.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Rate",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_AccrIntM_RateInfo"),
                new ArgumentInfo("Par",
                                 "is the security's par value.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Par",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_AccrIntM_ParInfo"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Basis",
                                 "is the type of day count basis to use.", 
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Basis",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_AccrIntM_BasisInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Financial, description, requiredArguments, optionalArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods
        
        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            DateTime? issue = FormatHelper.ConvertDoubleToDateTime(context.Arguments[0]);
            DateTime? settlement = FormatHelper.ConvertDoubleToDateTime(context.Arguments[1]);
            double rate = context.Arguments[2];
            double par = context.Arguments[3];

            int basis = 0;
            if (context.Arguments.Length == 5)
            {
                basis = (int)context.Arguments[4];
            }

            if (issue == null || settlement == null || issue >= settlement || rate <= 0 || par <= 0 || basis < 0 || basis > 4)
            {
                return ErrorExpressions.NumberError;
            }

            double result;
            try
            {
                result = FinancialFunctions.ACCRINTM((DateTime)issue, (DateTime)settlement, rate, par, basis);
            }
            catch (Exception)
            {
                return ErrorExpressions.NumberError;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

        #endregion
    }
}