﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the AMORDEGRC function.
    /// </summary>
    public class AmorDegrc : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "AMORDEGRC";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return ArgumentConversionRules.NonBoolNumberFunctionConversion;
            }
        }

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static AmorDegrc()
        {
            string description = "Returns the depreciation for each accounting period. This function is provided for the French accounting system. If an asset is purchased in the middle of the accounting period, the prorated depreciation is taken into account. The function is similar to AMORLINC, except that a depreciation coefficient is applied in the calculation depending on the life of the assets.";
            string descriptionKey = "Spreadsheet_Functions_AmorDegrc_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Cost",
                                 "is the cost of the asset.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Cost",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_AmorDegrc_CostInfo"),
                new ArgumentInfo("DatePurchased",
                                 "is the date of the purchase of the asset.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_DatePurchased",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_AmorDegrc_DatePurchasedInfo"),
                new ArgumentInfo("FirstPeriod",
                                 "is the date of the end of the first period.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_FirstPeriod",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_AmorDegrc_FirstPeriodInfo"),
                new ArgumentInfo("Salvage",
                                 "is the salvage value at the end of the life of the asset.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Salvage",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_AmorDegrc_SalvageInfo"),
                new ArgumentInfo("Period",
                                 "is the period.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Period",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_AmorDegrc_PeriodInfo"),
                new ArgumentInfo("Rate",
                                 "is the rate of depreciation.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Rate",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_AmorDegrc_RateInfo"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Basis",
                                 "is the year basis to be used.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Basis",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_AmorDegrc_BasisInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Financial, description, requiredArguments, optionalArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods
        
        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double cost = context.Arguments[0];
            DateTime? datePurchased = FormatHelper.ConvertDoubleToDateTime(context.Arguments[1]);
            DateTime? firstPeriod = FormatHelper.ConvertDoubleToDateTime(context.Arguments[2]);
            double salvage = context.Arguments[3];
            double period = context.Arguments[4];
            double rate = context.Arguments[5];

            int basis = 0;
            if (context.Arguments.Length == 7)
            {
                basis = (int)context.Arguments[6];
            }

            if (datePurchased == null || firstPeriod == null || datePurchased >= firstPeriod)
            {
                return ErrorExpressions.NumberError;
            }

            if (cost < 0 || salvage < 0 || salvage >= cost || period < 0 || rate <= 0 || basis < 0 || basis > 4)
            {
                return ErrorExpressions.NumberError;
            }

            double result;
            try
            {
                result = FinancialFunctions.AMORDEGRC(cost, (DateTime)datePurchased, (DateTime)firstPeriod, salvage, period, rate, basis);
            }
            catch (Exception)
            {
                return ErrorExpressions.NumberError;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

        #endregion
    }
}