﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the CUMPRINC function.
    /// </summary>
    public class CumPrinc : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "CUMPRINC";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return ArgumentConversionRules.NonBoolNumberFunctionConversion;
            }
        }

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static CumPrinc()
        {
            string description = "Returns the cumulative principal paid on a loan between start_period and end_period.";
            string descriptionKey = "Spreadsheet_Functions_CumPrinc_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Rate",
                                 "is the interest rate.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Rate",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Cum_RateInfo"),
                new ArgumentInfo("Nper",
                                 "is the total number of payment periods.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Nper",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Cum_NperInfo"),
                new ArgumentInfo("Pv",
                                 "is the present value.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Pv",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Cum_PvInfo"),
                new ArgumentInfo("StartPeriod",
                                 "is the first period in the calculation. Payment periods are numbered beginning with 1.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_StartPeriod",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Cum_StartPeriodInfo"),
                new ArgumentInfo("EndPeriod",
                                 "is the last period in the calculation.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_EndPeriod",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Cum_EndPeriodInfo"),
                new ArgumentInfo("Type",
                                 "is the timing of the payment.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Type",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Cum_TypeInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Financial, description, requiredArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods
        
        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double rate = context.Arguments[0];
            double nper = context.Arguments[1];
            double pv = context.Arguments[2];
            double startPeriod = context.Arguments[3];
            double endPeriod = context.Arguments[4];
            int type = (int)context.Arguments[5];

            if (rate <= 0 || nper <= 0 || pv <= 0 || startPeriod < 1 || endPeriod < 1 || startPeriod > endPeriod || (type != 0 && type != 1))
            {
                return ErrorExpressions.NumberError;
            }

            double result;
            try
            {
                result = FinancialFunctions.CUMPRINC(rate, nper, pv, startPeriod, endPeriod, type);
            }
            catch (Exception)
            {
                return ErrorExpressions.NumberError;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

        #endregion
    }
}