﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings.Builders;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the DB function.
    /// </summary>
    public class Db : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "DB";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Db()
        {
            string description = "Returns the depreciation of an asset for a specified period using the fixed-declining balance method.";
            string descriptionKey = "Spreadsheet_Functions_Db_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Cost",
                                 "is the initial cost of the asset.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Cost",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Db_CostInfo"),
                new ArgumentInfo("Salvage",
                                 "is the value at the end of the depreciation (sometimes called the salvage value of the asset).",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Salvage",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Db_SalvageInfo"),
                new ArgumentInfo("Life",
                                 "is the number of periods over which the asset is being depreciated (sometimes called the useful life of the asset).",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Life",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Db_LifeInfo"),
                new ArgumentInfo("Period",
                                 "is the period for which you want to calculate the depreciation. Period must use the same units as life.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Period",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Db_PeriodInfo"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Month",
                                 "is the number of months in the first year. If month is omitted, it is assumed to be 12.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Month",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Db_MonthInfo"),
            };

            string formatString = CurrencyFormatStringBuilder.BuildFormatStrings(new Formatting.FormatStrings.CurrencyInfo(FormatHelper.DefaultSpreadsheetCulture.CultureInfo.NumberFormat.CurrencySymbol), 2).Last();
            Info = new FunctionInfo(FunctionName, FunctionCategory.Financial, description, requiredArguments, optionalArguments, new CellValueFormat(formatString), descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods
        
        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double cost = context.Arguments[0];
            double salvage = context.Arguments[1];
            double life = context.Arguments[2];
            double period = context.Arguments[3];

            int month = 12;
            if (context.Arguments.Length == 5)
            {
                month = (int)context.Arguments[4];
            }

            if (cost < 0 || salvage < 0 || life <= 0 || period <= 0 || period > life || month <= 0 || month > 12)
            {
                return ErrorExpressions.NumberError;
            }

            double result;
            try
            {
                result = FinancialFunctions.DB(cost, salvage, life, period, month);
            }
            catch (Exception)
            {
                return ErrorExpressions.NumberError;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

        #endregion
    }
}