﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings.Builders;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the FV function.
    /// </summary>
    public class Fv : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "FV";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Fv()
        {
            string description = "Returns the future value of an investment based on periodic, constant payments and a constant interest rate.";
            string descriptionKey = "Spreadsheet_Functions_Fv_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Rate",
                                 "is the interest rate per period.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Rate",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Fv_RateInfo"),
                new ArgumentInfo("Nper",
                                 "is the total number of payment periods in an annuity.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Nper",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Fv_NperInfo"),
                new ArgumentInfo("Pmt",
                                 "is the payment made each period; it cannot change over the life of the annuity. Typically, pmt contains principal and interest but no other fees or taxes. If pmt is omitted, you must include the pv argument.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Pmt",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Fv_PmtInfo"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Pv",
                                 "is the present value, or the lump-sum amount that a series of future payments is worth right now. If pv is omitted, it is assumed to be 0 (zero), and you must include the pmt argument.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Pv",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Fv_PvInfo"),
                new ArgumentInfo("Type",
                                 "is the indicates when payments are due. If type is omitted, it is assumed to be 0.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Type",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Fv_TypeInfo"),
            };

            string formatString = CurrencyFormatStringBuilder.BuildFormatStrings(new Formatting.FormatStrings.CurrencyInfo(FormatHelper.DefaultSpreadsheetCulture.CultureInfo.NumberFormat.CurrencySymbol), 2).Last();
            Info = new FunctionInfo(FunctionName, FunctionCategory.Financial, description, requiredArguments, optionalArguments, new CellValueFormat(formatString), descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods
        
        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double rate = context.Arguments[0];
            double nper = context.Arguments[1];
            double pmt = context.Arguments[2];

            int pv = 0;
            if (context.Arguments.Length > 3)
            {
                pv = (int)context.Arguments[3];
            }

            int type = 0;
            if (context.Arguments.Length == 5)
            {
                type = (int)context.Arguments[4];
            }

            if (type < 0 || type > 1)
            {
                return ErrorExpressions.NumberError;
            }

            double result;
            try
            {
                result = FinancialFunctions.FV(rate, nper, pmt, pv, type);
            }
            catch (Exception)
            {
                return ErrorExpressions.NumberError;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

        #endregion
    }
}