﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the NPER function.
    /// </summary>
    public class NPer : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "NPER";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static NPer()
        {
            string description = "Returns the number of periods for an investment based on periodic, constant payments and a constant interest rate.";
            string descriptionKey = "Spreadsheet_Functions_Nper_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Rate",
                                 "is the interest rate per period. For example, use 6%/4 for quarterly payments at 6% APR.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Rate",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Nper_RateInfo"),
                new ArgumentInfo("Pmt",
                                 "is the payment made each period; it cannot change over the life of the annuity. Typically, pmt contains principal and interest but no other fees or taxes.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Pmt",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Nper_PmtInfo"),
                new ArgumentInfo("Pv",
                                 "is the present value, or the lump-sum amount that a series of future payments is worth right now.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Pv",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Nper_PvInfo"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Fv",
                                 "is the future value, or a cash balance you want to attain after the last payment is made. If fv is omitted, it is assumed to be 0 (the future value of a loan, for example, is 0).",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Fv",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Nper_FvInfo"),
                new ArgumentInfo("Type",
                                 "is the number 0 or 1 and indicates when payments are due.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Type",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Nper_TypeInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Financial, description, requiredArguments, optionalArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods
        
        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double rate = context.Arguments[0];
            double pmt = context.Arguments[1];
            double pv = context.Arguments[2];

            double fv = 0;
            if (context.Arguments.Length > 3)
            {
                fv = context.Arguments[3];
            }

            int type = 0;
            if (context.Arguments.Length > 4)
            {
                type = (int)context.Arguments[4];
            }

            if (type < 0 || type > 1)
            {
                return ErrorExpressions.NumberError;
            }

            double result;
            try
            {
                result = FinancialFunctions.NPER(rate, pmt, pv, fv, type);
            }
            catch (Exception)
            {
                return ErrorExpressions.NumberError;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

        #endregion
    }
}