﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the ODDFPRICE function.
    /// </summary>
    public class OddFPrice : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "ODDFPRICE";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return ArgumentConversionRules.NonBoolNumberFunctionConversion;
            }
        }

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static OddFPrice()
        {
            string description = "Returns the price per $100 face value of a security having an odd (short or long) first period.";
            string descriptionKey = "Spreadsheet_Functions_OddFPrice_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Settlement",
                                 "is the security's settlement date. The security settlement date is the date after the issue date when the security is traded to the buyer.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Settlement",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_OddFL_SettlementInfo"),
                new ArgumentInfo("Maturity",
                                 "is the security's maturity date. The maturity date is the date when the security expires.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Maturity",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_OddFL_MaturityInfo"),
                new ArgumentInfo("Issue",
                                 "is the security's issue date.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Issue",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_OddF_IssueInfo"),
                new ArgumentInfo("FirstCoupon",
                                 "is the security's first coupon date.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_FirstCoupon",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_OddF_FirstCouponInfo"),
                new ArgumentInfo("Rate",
                                 "is the security's interest rate.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Rate",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_OddFL_RateInfo"),
                new ArgumentInfo("Yld",
                                 "is the security's annual yield.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Yld",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_OddPrice_YldInfo"),
                new ArgumentInfo("Redemption",
                                 "is the security's redemption value per $100 face value.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Redemption",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_OddFL_RedemptionInfo"),
                new ArgumentInfo("Frequency",
                                 "is the number of coupon payments per year. For annual payments, frequency = 1; for semiannual, frequency = 2; for quarterly, frequency = 4.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Frequency",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_OddFL_FrequencyInfo"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Basis",
                                 "is the type of day count basis to use.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Basis",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_OddFL_BasisInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Financial, description, requiredArguments, optionalArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods
        
        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            var arguments = context.Arguments;
            DateTime? settlement = FormatHelper.ConvertDoubleToDateTime(arguments[0]);
            DateTime? maturity = FormatHelper.ConvertDoubleToDateTime(arguments[1]);
            DateTime? issue = FormatHelper.ConvertDoubleToDateTime(arguments[2]);
            DateTime? firstCoupon = FormatHelper.ConvertDoubleToDateTime(arguments[3]);
            double rate = arguments[4];
            double yld = arguments[5];
            double redemption = arguments[6];
            int frequency = (int)arguments[7];

            int basis = 0;
            if (arguments.Length == 9)
            {
                basis = (int)arguments[8];
            }

            if (settlement == null || maturity == null || issue == null || firstCoupon == null ||
                maturity <= firstCoupon || firstCoupon <= settlement || settlement <= issue ||
                redemption <= 0 || rate < 0 || yld < 0 || basis < 0 || basis > 4)
            {
                return ErrorExpressions.NumberError;
            }

            double result;
            try
            {
                result = FinancialFunctions.ODDFPRICE((DateTime)settlement, (DateTime)maturity, (DateTime)issue, (DateTime)firstCoupon, rate, yld, redemption, frequency, basis);
            }
            catch (Exception)
            {
                return ErrorExpressions.NumberError;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

        #endregion
    }
}