﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings.Builders;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the PPMT function.
    /// </summary>
    public class Ppmt : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "PPMT";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Ppmt()
        {
            string description = "Returns the payment on the principal for a given period for an investment based on periodic, constant payments and a constant interest rate.";
            string descriptionKey = "Spreadsheet_Functions_Ppmt_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Rate",
                                 "is the interest rate per period. For example, use 6%/4 for quarterly payments at 6% APR.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Rate",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Pmt_RateInfo"),
                new ArgumentInfo("Per",
                                 "is the period and must be in the range 1 to nper.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Per",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Ppmt_PerInfo"),
                new ArgumentInfo("Nper",
                                 "is the total number of payment periods in an annuity.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Nper",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Ppmt_NperInfo"),
                new ArgumentInfo("Pv",
                                 "is the present value, or the lump-sum amount that a series of future payments is worth right now.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Pv",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Pmt_PvInfo"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Fv",
                                 "is the future value, or a cash balance you want to attain after the last payment is made. If fv is omitted, it is assumed to be 0 (the future value of a loan, for example, is 0).",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Fv",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Pmt_FvInfo"),
                new ArgumentInfo("Type",
                                 "is the number 0 or 1 and indicates when payments are due.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Type",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Pmt_TypeInfo"),
            };

            string formatString = CurrencyFormatStringBuilder.BuildFormatStrings(new Formatting.FormatStrings.CurrencyInfo(FormatHelper.DefaultSpreadsheetCulture.CultureInfo.NumberFormat.CurrencySymbol), 2).Last();
            Info = new FunctionInfo(FunctionName, FunctionCategory.Financial, description, requiredArguments, optionalArguments, new CellValueFormat(formatString), descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods
        
        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double rate = context.Arguments[0];
            double per = context.Arguments[1];
            double nper = context.Arguments[2];
            double pv = context.Arguments[3];
            double fv = 0;
            if (context.Arguments.Length > 4)
            {
                fv = context.Arguments[4];
            }

            int type = 0;
            if (context.Arguments.Length > 5)
            {
                type = (int)context.Arguments[5] == 0 ? 0 : 1;
            }

            double result;
            try
            {
                result = FinancialFunctions.PPMT(rate, per, nper, pv, fv, type);
            }
            catch (Exception)
            {
                return ErrorExpressions.NumberError;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

        #endregion
    }
}