﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the RECEIVED function.
    /// </summary>
    public class Received : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "RECEIVED";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return ArgumentConversionRules.NonBoolNumberFunctionConversion;
            }
        }

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Received()
        {
            string description = "Returns the amount received at maturity for a fully invested security.";
            string descriptionKey = "Spreadsheet_Functions_Received_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Settlement",
                                 "is the security's settlement date. The security settlement date is the date after the issue date when the security is traded to the buyer.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Settlement",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Received_SettlementInfo"),
                new ArgumentInfo("Maturity",
                                 "is the security's maturity date. The maturity date is the date when the security expires.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Maturity",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Received_MaturityInfo"),
                new ArgumentInfo("Investment",
                                 "is the amount invested in the security.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Investment",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Received_InvestmentInfo"),
                new ArgumentInfo("Discount",
                                 "is the security's discount rate.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Discount",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Received_DiscountInfo"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Basis",
                                 "is the type of day count basis to use.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Basis",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Received_BasisInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Financial, description, requiredArguments, optionalArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods
        
        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            DateTime? settlement = FormatHelper.ConvertDoubleToDateTime(context.Arguments[0]);
            DateTime? maturity = FormatHelper.ConvertDoubleToDateTime(context.Arguments[1]);
            double investment = context.Arguments[2];
            double discount = context.Arguments[3];

            int basis = 0;
            if (context.Arguments.Length == 5)
            {
                basis = (int)context.Arguments[4];
            }

            if (settlement == null || maturity == null || settlement >= maturity || investment <= 0 || discount <= 0 || basis < 0 || basis > 4)
            {
                return ErrorExpressions.NumberError;
            }

            double result;
            try
            {
                result = FinancialFunctions.Received((DateTime)settlement, (DateTime)maturity, investment, discount, basis);
            }
            catch (Exception)
            {
                return ErrorExpressions.NumberError;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

        #endregion
    }
}