﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the RRI function.
    /// </summary>
    public class Rri : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "RRI";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Rri()
        {
            string description = "Returns an equivalent interest rate for the growth of an investment.";
            string descriptionKey = "Spreadsheet_Functions_Rri_PvInfo";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Nper",
                                 "is the number of periods for the investment.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Nper",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Rri_NperInfo"),
                new ArgumentInfo("Pv",
                                 "is the present value of the investment.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Pv",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Rri_PvInfo"),
                new ArgumentInfo("Fv",
                                 "is the future value of the investment.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Fv",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Rri_FvInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Financial, description, requiredArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods
        
        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double nper = context.Arguments[0];
            double pv = context.Arguments[1];
            double fv = context.Arguments[2];

            if (nper <= 0 || (pv == 0 && fv != 0) || (pv > 0 && fv < 0) || (pv < 0 && fv > 0))
            {
                return ErrorExpressions.NumberError;
            }

            double result;
            try
            {
                result = FinancialFunctions.RRI(nper, pv, fv);
            }
            catch (Exception)
            {
                return ErrorExpressions.NumberError;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

        #endregion
    }
}