﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings.Builders;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the VDB function.
    /// </summary>
    public class Vdb : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "VDB";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Vdb()
        {
            string description = "Returns the depreciation for each accounting period. This function is provided for the French accounting system. If an asset is purchased in the middle of the accounting period, the prorated depreciation is taken into account. The function is similar to AMORLINC, except that a depreciation coefficient is applied in the calculation depending on the life of the assets.";
            string descriptionKey = "Spreadsheet_Functions_Vdb_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Cost",
                                 "is the initial cost of the asset.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Cost",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Vdb_CostInfo"),
                new ArgumentInfo("Salvage",
                                 "is the value at the end of the depreciation (sometimes called the salvage value of the asset). This value can be 0.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Salvage",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Vdb_SalvageInfo"),
                new ArgumentInfo("Life",
                                 "is the number of periods over which the asset is depreciated (sometimes called the useful life of the asset)..",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Life",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Vdb_LifeInfo"),
                new ArgumentInfo("StartPeriod",
                                 "is the starting period for which you want to calculate the depreciation. Start_period must use the same units as life.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_StartPeriod",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Vdb_StartPeriodInfo"),
                new ArgumentInfo("EndPeriod",
                                 "is the ending period for which you want to calculate the depreciation. End_period must use the same units as life.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_EndPeriod",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Vdb_EndPeriodInfo"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Factor",
                                 "is the rate at which the balance declines. If factor is omitted, it is assumed to be 2 (the double-declining balance method). Change factor if you do not want to use the double-declining balance method. For a description of the double-declining balance method, see DDB.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Factor",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Vdb_FactorInfo"),
                new ArgumentInfo("NoSwitch",
                                 "A logical value specifying whether to switch to straight-line depreciation when depreciation is greater than the declining balance calculation.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_NoSwitch",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Vdb_NoSwitchInfo"),
            };

            string formatString = CurrencyFormatStringBuilder.BuildFormatStrings(new Formatting.FormatStrings.CurrencyInfo(FormatHelper.DefaultSpreadsheetCulture.CultureInfo.NumberFormat.CurrencySymbol), 2).Last();
            Info = new FunctionInfo(FunctionName, FunctionCategory.Financial, description, requiredArguments, optionalArguments, new CellValueFormat(formatString), descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods
        
        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double cost = context.Arguments[0];
            double salvage = context.Arguments[1];
            double life = context.Arguments[2];
            double startPeriod = context.Arguments[3];
            double endPeriod = context.Arguments[4];

            double result;

            // TODO:

            try
            {
                if (context.Arguments.Length == 5)
                {
                    result = FinancialFunctions.VDB(cost, salvage, life, startPeriod, endPeriod);
                }
                else if (context.Arguments.Length == 6)
                {
                    result = FinancialFunctions.VDB(cost, salvage, life, startPeriod, endPeriod, context.Arguments[5]);
                }
                else
                {
                    int noSwitch = (int)context.Arguments[6];

                    if (noSwitch != 0 && noSwitch != 1)
                    {
                        return ErrorExpressions.NumberError;
                    }

                    result = FinancialFunctions.VDB(cost, salvage, life, startPeriod, endPeriod, context.Arguments[5], noSwitch);
                }
            }
            catch (Exception)
            {
                return ErrorExpressions.NumberError;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

        #endregion
    }
}