﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the YIELDMAT function.
    /// </summary>
    public class YieldMat : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "YIELDMAT";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return ArgumentConversionRules.NonBoolNumberFunctionConversion;
            }
        }

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static YieldMat()
        {
            string description = "Returns the annual yield of a security that pays interest at maturity.";
            string descriptionKey = "Spreadsheet_Functions_YieldMat_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Settlement",
                                 "is the security's settlement date. The security settlement date is the date after the issue date when the security is traded to the buyer.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Settlement",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Yield_SettlementInfo"),
                new ArgumentInfo("Maturity",
                                 "is the security's maturity date. The maturity date is the date when the security expires.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Maturity",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Yield_MaturityInfo"),
                new ArgumentInfo("Issue",
                                 "is the ssecurity's issue date, expressed as a serial date number.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Issue",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_YieldMat_IssueInfo"),
                new ArgumentInfo("Rate",
                                 "is the annual coupon rate.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Rate",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_YieldMat_RateInfo"),
                new ArgumentInfo("Pr",
                                 "is the type of day count basis to use",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Pr",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_YieldMat_PrInfo"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Basis",
                                 "is the type of day count basis to use.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Basis",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Yield_MaturityInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Financial, description, requiredArguments, optionalArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods
        
        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            DateTime? settlement = FormatHelper.ConvertDoubleToDateTime(context.Arguments[0]);
            DateTime? maturity = FormatHelper.ConvertDoubleToDateTime(context.Arguments[1]);
            DateTime? issue = FormatHelper.ConvertDoubleToDateTime(context.Arguments[2]);
            double rate = context.Arguments[3];
            double pr = context.Arguments[4];

            int basis = 0;
            if (context.Arguments.Length == 6)
            {
                basis = (int)context.Arguments[5];
            }

            if (settlement == null || maturity == null || issue == null || issue >= settlement ||
                settlement >= maturity || rate <= 0 || pr <= 0 || basis < 0 || basis > 4)
            {
                return ErrorExpressions.NumberError;
            }

            double result = 5;
            try
            {
                result = FinancialFunctions.YIELDMAT((DateTime)settlement, (DateTime)maturity, (DateTime)issue, rate, pr, basis);
            }
            catch (Exception)
            {
                return ErrorExpressions.NumberError;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

        #endregion
    }
}