﻿using System;
using System.ComponentModel;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Abstract class providing methods and properties for creating an RadSpreadsheet function.
    /// </summary>
    public abstract class FunctionBase
    {
        // TODO: The functions are instances is this thread safe calculations?

        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <value>The name.</value>
        /// <returns>The name as String.</returns>
        public abstract string Name { get; }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <value>The function info.</value>
        /// <returns>The function info as FunctionInfo.</returns>
        public abstract FunctionInfo FunctionInfo { get; }

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <value>The argument conversion rules.</value>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        public virtual ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return ArgumentConversionRules.NumberFunctionConversion;
            }
        }

        //public virtual bool IsVolatile
        //{
        //    get
        //    {
        //        return false;
        //    }
        //}

        #endregion


        #region Methods

        /// <summary>
        /// Evaluates the function with the specified arguments.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Function result as RadExpression.</returns>
        public RadExpression Evaluate(FunctionEvaluationContext<RadExpression> context)
        {
            return this.EvaluateOverride(context);
        }

        /// <summary>
        /// Virtual method that evaluates the function with the specified context.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Function result as RadExpression.</returns>
        protected virtual RadExpression EvaluateOverride(FunctionEvaluationContext<RadExpression> context)
        {
            throw new NotSupportedException();
        }

        /// <summary>
        /// Determines whether arguments number is valid.
        /// </summary>
        /// <param name="argumentsCount">The arguments count.</param>
        /// <returns>Boolean indicating whether the arguments number is valid.</returns>
        public bool IsArgumentNumberValid(int argumentsCount)
        {
            int optionalArgumentsCount = FunctionInfo.OptionalArgumentsCount;
            int optionalArgumentsRepetitionCount = FunctionInfo.OptionalArgumentsRepetitionCount > 0 ? FunctionInfo.OptionalArgumentsRepetitionCount : 1;

            int minArgumentsCount = FunctionInfo.RequiredArgumentsCount;
            int maxArgumentdCount = minArgumentsCount + optionalArgumentsCount * optionalArgumentsRepetitionCount;

            bool isValid = argumentsCount >= minArgumentsCount && argumentsCount <= maxArgumentdCount;

            if (optionalArgumentsCount > 1 && FunctionInfo.OptionalArgumentsRepetitionCount > 1)
            {
                isValid &= (argumentsCount - minArgumentsCount) % optionalArgumentsCount == 0;
            }

            return isValid;
        }

        #endregion
    }
}