﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Provides methods and properties used in RadSpreadsheet functions.
    /// </summary>
    internal static class FunctionHelper
    {
        #region Fields

        private static readonly ConstantExpressionConverter<double>[] numberConverters = new ConstantExpressionConverter<double>[] { FunctionHelper.EmptyToNumberExpression,
                                                                                                                 FunctionHelper.NumberToNumberExpression,
                                                                                                                 FunctionHelper.BoolToNumberExpression,
                                                                                                                 FunctionHelper.TextNumberToNumberExpression,
                                                                                                                 FunctionHelper.NonTextNumbersToNumberExpression };

        private static readonly ConstantExpressionConverter<bool>[] booleanConverters = new ConstantExpressionConverter<bool>[] { FunctionHelper.EmptyToBoolExpression,
                                                                                                                    FunctionHelper.NumberToBoolExpression,
                                                                                                                    FunctionHelper.BoolToBoolExpression,
                                                                                                                    FunctionHelper.TextNumberToBoolExpression,
                                                                                                                    FunctionHelper.NonTextNumbersToBoolExpression };

        private static readonly ConstantExpressionConverter<string>[] stringConverters = new ConstantExpressionConverter<string>[] { FunctionHelper.BoolToStringExpression,
                                                                                                                    FunctionHelper.AllToStringExpression };

        public delegate ErrorExpression ConstantExpressionConverter<T>(RadExpression expression, bool isDirect, ArgumentConversionRules conversionRules, out ConstantExpression<T> convertedExpression);

        #endregion


        #region Methods

        public static ErrorExpression TryConvertArgument(RadExpression expression, ArgumentConversionRules conversionRules, ArgumentType argumentType, List<object> convertedArguments)
        {
            ErrorExpression errorExpression = null;

            switch (argumentType)
            {
                case ArgumentType.Array:

                    errorExpression = TryConvertToArrayArgument(expression, convertedArguments);
                    break;

                case ArgumentType.Number:

                    errorExpression = TryConvertArgument<double>(expression, conversionRules, TryConvertArgumentToNumber, convertedArguments);
                    break;

                case ArgumentType.Logical:

                    errorExpression = TryConvertArgument<bool>(expression, conversionRules, TryConvertArgumentToBool, convertedArguments);
                    break;

                case ArgumentType.Text:

                    errorExpression = TryConvertArgument<string>(expression, conversionRules, TryConvertArgumentToString, convertedArguments);
                    break;

                case ArgumentType.Reference:

                    errorExpression = TryConvertToReferenceArgument(expression, convertedArguments);
                    break;

                case ArgumentType.Any:

                    errorExpression = TryConvertToAnyArgument(expression, convertedArguments);
                    break;
            }

            return errorExpression;
        }

        internal static ErrorExpression TryConvertFirstArgument(RadExpression expression, ArgumentConversionRules convertionRules, ArgumentType argumentType, out object convertedArgument)
        {
            convertedArgument = null;
            List<object> convertedArguments = new List<object>();

            ErrorExpression errorExpression = TryConvertArgument(expression, convertionRules, argumentType, convertedArguments);

            if (convertedArguments.Count > 0 && convertedArguments[0] != null)
            {
                convertedArgument = convertedArguments[0];
            }

            return errorExpression;
        }

        internal static ArrayExpression ConvertToArrayExpression(ConstantExpression constantExpression)
        {
            ArrayExpression arrayExpression = constantExpression as ArrayExpression;

            NumberExpression numberExpression = constantExpression as NumberExpression;
            if (numberExpression != null)
            {
                arrayExpression = new ArrayExpression(new RadExpression[,] { { numberExpression } });
            }

            return arrayExpression;
        }

        internal static ErrorExpression TryGetArrayFromFunctionArgument(object argument, out ArrayExpression array, out Worksheet worksheet, out CellIndex topLeftCellIndex)
        {
            array = null;
            worksheet = null;
            topLeftCellIndex = new CellIndex(0, 0);

            ArrayExpression arrayExpression = FunctionHelper.ConvertToArrayExpression(argument as ConstantExpression);

            CellReferenceRangeExpression referenceExpression = argument as CellReferenceRangeExpression;
            if (referenceExpression != null)
            {
                if (!referenceExpression.IsValid)
                {
                    return ErrorExpressions.ReferenceError;
                }

                if (!referenceExpression.HasSingleRange)
                {
                    return ErrorExpressions.NotAvailableError;
                }

                arrayExpression = referenceExpression.FirstRangeArray;
                worksheet = referenceExpression.Worksheet;
                CellReference topLeftReference = referenceExpression.CellReferenceRange.FromCellReference;
                topLeftCellIndex = new CellIndex(topLeftReference.ActualRowIndex, topLeftReference.ActualColumnIndex);
            }

            if (arrayExpression != null)
            {
                array = arrayExpression;

                return null;
            }

            return ErrorExpressions.ValueError;
        }

        internal static ErrorExpression TryGetCriteriaEvaluator(object argument, Worksheet criteriaWorksheet, ComparisonOperator comparisonOperator, out CriteriaEvaluator evaluator)
        {
            evaluator = null;

            ConstantExpression constant;
            ErrorExpression error = TryGetNonArrayConstantExpression(argument, out constant);
            if (error != null)
            {
                return error;
            }

            string criteriaOperator;

            switch (comparisonOperator)
            {
                case ComparisonOperator.LessThanOrEqualsTo:
                    criteriaOperator = SpreadsheetCultureHelper.LessThanOrEqualTo;
                    break;
                case ComparisonOperator.EqualsTo:
                    criteriaOperator = SpreadsheetCultureHelper.Equal;
                    break;
                case ComparisonOperator.GreaterThanOrEqualsTo:
                    criteriaOperator = SpreadsheetCultureHelper.GreaterThanOrEqualTo;
                    break;
                default:
                    throw new NotSupportedException("This type of comparison is not supported by the lookup and match functions.");
            }

            StringExpression criteria = new StringExpression(criteriaOperator + constant.GetValueAsString());

            evaluator = new CriteriaEvaluator(criteria, criteriaWorksheet);

            return null;
        }

        private static ErrorExpression TryGetNonArrayConstantExpression(object expression, out ConstantExpression constant)
        {
            constant = null;

            CellReferenceRangeExpression referenceExpression = expression as CellReferenceRangeExpression;
            if (referenceExpression != null)
            {
                if (!referenceExpression.IsInRange)
                {
                    return ErrorExpressions.ReferenceError;
                }

                bool isPointReferenceRange = referenceExpression.HasSingleRange &&
                    referenceExpression.CellReferenceRange.ColumnsCount == 1 &&
                    referenceExpression.CellReferenceRange.RowsCount == 1;

                if (!isPointReferenceRange)
                {
                    return ErrorExpressions.ValueError;
                }
            }

            constant = ((RadExpression)expression).GetValueAsNonArrayConstantExpression(false);

            CellValueType type = constant.GetCellValueType();
            if (type == CellValueType.Empty)
            {
                return ErrorExpressions.NotAvailableError;
            }

            ErrorExpression error = constant as ErrorExpression;
            if (error != null)
            {
                return error;
            }

            return null;
        }

        private static ErrorExpression TryConvertToArrayArgument(RadExpression expression, List<object> convertedArguments)
        {
            ConstantExpression constantExpression = expression.GetValueAsConstant();

            ArrayExpression arrayExpression = ConvertToArrayExpression(constantExpression);

            if (arrayExpression != null)
            {
                convertedArguments.Add(arrayExpression);
                return null;
            }

            ErrorExpression errorExpression = constantExpression as ErrorExpression;
            if (errorExpression != null)
            {
                return errorExpression;
            }

            return ErrorExpressions.ValueError;
        }

        private static ErrorExpression TryConvertToReferenceArgument(RadExpression expression, List<object> convertedArgumnets)
        {
            ErrorExpression errorExpression = null;

            RadExpression expressionValue = expression.GetValue();

            if (expression is CellReferenceRangeExpression)
            {
                convertedArgumnets.Add(expression);
            }
            else if (expressionValue is CellReferenceRangeExpression)
            {
                convertedArgumnets.Add(expressionValue);
            }
            else if (expressionValue.Equals(ErrorExpressions.ReferenceError))
            {
                errorExpression = ErrorExpressions.ReferenceError;
            }
            else if (expressionValue.Equals(ErrorExpressions.NullError))
            {
                errorExpression = ErrorExpressions.NullError;
            }
            else
            {
                errorExpression = ErrorExpressions.ValueError;
            }

            return errorExpression;
        }

        private static ErrorExpression TryConvertToAnyArgument(RadExpression expression, List<object> convertedArguments)
        {
            RadExpression expressionValue = expression.GetValueAsConstantOrCellReference();

            ErrorExpression errorExpression = expressionValue as ErrorExpression;

            if (errorExpression == null)
            {
                convertedArguments.Add(expressionValue);
            }

            return errorExpression;
        }

        private static ErrorExpression TryConvertArgument<T>(RadExpression expression, ArgumentConversionRules convertionRules, ConstantExpressionConverter<T> argumentConverter, List<object> convertedArguments)
        {
            ErrorExpression errorExpression;
            RadExpression expressionValue = expression.GetValueAsConstant();
            ArrayExpression arrayExpression = expressionValue as ArrayExpression;

            if (arrayExpression != null)
            {
                errorExpression = TryConvertArrayArgument<T>(arrayExpression, convertionRules, argumentConverter, convertedArguments);
            }
            else
            {
                bool isDirect = IsDirectArgument(expression);
                errorExpression = TryConvertSingleArgument<T>(expression, convertionRules, argumentConverter, convertedArguments, isDirect);
            }

            return errorExpression;
        }

        private static bool IsDirectArgument(RadExpression expression)
        {
            if (expression is CellReferenceRangeExpression || expression is ArrayExpression)
            {
                return false;
            }

            if (expression is ConstantExpression)
            {
                return true;
            }

            return IsDirectArgument(expression.GetValueAsConstantOrCellReference());
        }

        private static ErrorExpression TryConvertSingleArgument<T>(RadExpression expression, ArgumentConversionRules convertionRules, ConstantExpressionConverter<T> argumentConverter, List<object> convertedArguments, bool isDirect)
        {
            ConstantExpression<T> convertedArgument;

            ErrorExpression errorExpression = argumentConverter(expression, isDirect, convertionRules, out convertedArgument);

            if (errorExpression == null && convertedArgument != null)
            {
                convertedArguments.Add(convertedArgument.Value);
            }

            return errorExpression;
        }

        private static ErrorExpression TryConvertArrayArgument<T>(ArrayExpression arrayExpression, ArgumentConversionRules convertionRules, ConstantExpressionConverter<T> argumentConverter, List<object> convertedArguments)
        {
            ErrorExpression errorExpression = arrayExpression.GetContainingError();
            if (errorExpression != null)
            {
                return errorExpression;
            }

            if (convertionRules.ArrayArgument == ArrayArgumentInterpretation.UseAllElements)
            {
                foreach (RadExpression arrayElement in arrayExpression)
                {
                    CellReferenceRangeExpression cellReferenceRange = arrayElement as CellReferenceRangeExpression;
                    if (cellReferenceRange != null)
                    {
                        ArgumentConversionRules modifiedConvertionRules = new ArgumentConversionRules(convertionRules, ArrayArgumentInterpretation.UseFirstElement);
                        errorExpression = TryConvertArrayArgument<T>(((ArrayExpression)cellReferenceRange.GetValue()), modifiedConvertionRules, argumentConverter, convertedArguments);
                    }
                    else
                    {
                        ArrayExpression innerArrayExpression = arrayElement as ArrayExpression;
                        if (innerArrayExpression != null)
                        {
                            errorExpression = TryConvertArrayArgument<T>(innerArrayExpression, convertionRules, argumentConverter, convertedArguments);
                        }
                        else
                        {
                            errorExpression = TryConvertSingleArgument(arrayElement, convertionRules, argumentConverter, convertedArguments, isDirect: false);
                        }
                    }

                    if (errorExpression != null)
                    {
                        return errorExpression;
                    }
                }
            }
            else if (convertionRules.ArrayArgument == ArrayArgumentInterpretation.UseFirstElement)
            {
                RadExpression expressionValue = arrayExpression[0, 0].GetValueAsConstant();

                ArrayExpression innerArrayExpression = expressionValue as ArrayExpression;
                if (innerArrayExpression != null)
                {
                    errorExpression = TryConvertArrayArgument<T>(innerArrayExpression, convertionRules, argumentConverter, convertedArguments);
                }
                else
                {
                    errorExpression = TryConvertSingleArgument<T>(arrayExpression[0, 0], convertionRules, argumentConverter, convertedArguments, isDirect: false);
                }
            }

            return errorExpression;
        }

        #endregion


        #region ToNumber Convertion

        // TODO:
        private static ErrorExpression TryConvertArgumentToNumber(RadExpression expression, bool isDirectArgument, ArgumentConversionRules convertionRules, out ConstantExpression<double> convertedArgument)
        {
            ErrorExpression errorExpression = null;
            convertedArgument = null;

            RadExpression expressionValue = expression.GetValue();

            errorExpression = expressionValue as ErrorExpression;

            if (errorExpression == null)
            {
                for (int i = 0; i < numberConverters.Length; i++)
                {
                    errorExpression = numberConverters[i](expressionValue, isDirectArgument, convertionRules, out convertedArgument);
                    if (errorExpression != null)
                    {
                        return errorExpression;
                    }
                    else if (convertedArgument != null)
                    {
                        return null;
                    }
                }
            }

            return errorExpression;
        }

        private static ErrorExpression ToNumberExpression(double value, NumberExpression expression, bool isDirect, ArgumentInterpretation directOption, ArgumentInterpretation indirectOption, out ConstantExpression<double> convertedExpression)
        {
            ErrorExpression errorExpression = null;
            convertedExpression = null;

            if ((isDirect && directOption == ArgumentInterpretation.TreatAsError) || (!isDirect && indirectOption == ArgumentInterpretation.TreatAsError))
            {
                errorExpression = ErrorExpressions.ValueError;
            }
            else if ((isDirect && directOption == ArgumentInterpretation.UseAsIs) || (!isDirect && indirectOption == ArgumentInterpretation.UseAsIs))
            {
                if (expression != null)
                {
                    convertedExpression = expression;
                }
                else
                {
                    convertedExpression = new NumberExpression(value);
                }
            }
            else if ((isDirect && directOption == ArgumentInterpretation.ConvertToDefault) || (!isDirect && indirectOption == ArgumentInterpretation.ConvertToDefault))
            {
                convertedExpression = NumberExpression.Zero;
            }

            return errorExpression;
        }

        private static ErrorExpression EmptyToNumberExpression(RadExpression expressionValue, bool isDirect, ArgumentConversionRules convertionRules, out ConstantExpression<double> convertedArgument)
        {
            ErrorExpression errorExpression = null;
            convertedArgument = null;

            if (expressionValue is EmptyExpression)
            {
                errorExpression = ToNumberExpression(0, NumberExpression.Zero, isDirect, convertionRules.EmptyDirectArgument, convertionRules.EmptyIndirectArgument, out convertedArgument);
            }

            return errorExpression;
        }

        private static ErrorExpression NumberToNumberExpression(RadExpression expression, bool isDirect, ArgumentConversionRules convertionRules, out ConstantExpression<double> convertedExpression)
        {
            ErrorExpression errorExpression = null;
            convertedExpression = null;

            NumberExpression numberExpression = expression as NumberExpression;
            if (numberExpression != null)
            {
                errorExpression = ToNumberExpression(numberExpression.Value, numberExpression, isDirect, convertionRules.NumberDirectArgument, convertionRules.NumberIndirectArgument, out convertedExpression);
            }

            return errorExpression;
        }

        private static ErrorExpression BoolToNumberExpression(RadExpression expression, bool isDirect, ArgumentConversionRules convertionRules, out ConstantExpression<double> convertedExpression)
        {
            ErrorExpression errorExpression = null;
            convertedExpression = null;

            BooleanExpression booleanExpression = expression as BooleanExpression;
            if (booleanExpression != null)
            {
                bool booleanValue = booleanExpression.Value;
                double value = booleanValue ? 1 : 0;
                NumberExpression numberExpression = booleanValue ? NumberExpression.One : NumberExpression.Zero;
                errorExpression = ToNumberExpression(value, numberExpression, isDirect, convertionRules.BoolDirectArgument, convertionRules.BoolIndirectArgument, out convertedExpression);
            }

            return errorExpression;
        }

        private static ErrorExpression TextNumberToNumberExpression(RadExpression expression, bool isDirect, ArgumentConversionRules convertionRules, out ConstantExpression<double> convertedExpression)
        {
            ErrorExpression errorExpression = null;
            convertedExpression = null;

            StringExpression stringExpression = expression as StringExpression;
            if (stringExpression != null)
            {
                double stringAsDouble;
                bool succeededToParse = false;
                if (convertionRules.ConvertDateTextToNumber)
                {
                    bool canParseToDate = DateValue.TryParseToDate(stringExpression.Value, out stringAsDouble);
                    if (!canParseToDate)
                    {
                        return ErrorExpressions.ValueError;
                    }

                    succeededToParse = true;
                }
                else if (FormatHelper.DefaultSpreadsheetCulture.TryParseDouble(stringExpression.Value, out stringAsDouble))
                {
                    succeededToParse = true;
                }

                if (succeededToParse)
                {
                    errorExpression = ToNumberExpression(stringAsDouble, null, isDirect, convertionRules.TextNumberDirectArgument, convertionRules.TextNumberIndirectArgument, out convertedExpression);
                }
            }

            return errorExpression;
        }

        private static ErrorExpression NonTextNumbersToNumberExpression(RadExpression expression, bool isDirect, ArgumentConversionRules convertionRules, out ConstantExpression<double> convertedExpression)
        {
            ErrorExpression errorExpression = null;
            convertedExpression = null;

            StringExpression stringExpression = expression as StringExpression;
            if (stringExpression != null)
            {
                double stringAsDouble;
                bool isNonTextNumber = !(double.TryParse(stringExpression.Value, out stringAsDouble));

                if (isNonTextNumber)
                {
                    errorExpression = ToNumberExpression(0, NumberExpression.Zero, isDirect, convertionRules.NonTextNumberDirectArgument, convertionRules.NonTextNumberIndirectArgument, out convertedExpression);
                }
            }

            return errorExpression;
        }

        #endregion


        #region ToBool Convertion

        private static ErrorExpression TryConvertArgumentToBool(RadExpression expression, bool isDirectArgument, ArgumentConversionRules convertionRules, out ConstantExpression<bool> convertedArgument)
        {
            ErrorExpression errorExpression = null;
            convertedArgument = null;

            RadExpression expressionValue = expression.GetValue();

            errorExpression = expressionValue as ErrorExpression;

            if (errorExpression == null)
            {
                for (int i = 0; i < booleanConverters.Length; i++)
                {
                    errorExpression = booleanConverters[i](expressionValue, isDirectArgument, convertionRules, out convertedArgument);
                    if (errorExpression != null)
                    {
                        return errorExpression;
                    }
                    else if (convertedArgument != null)
                    {
                        return null;
                    }
                }
            }

            return errorExpression;
        }

        private static ErrorExpression ToBoolExpression(bool value, bool isDirect, ArgumentInterpretation directOption, ArgumentInterpretation indirectOption, out ConstantExpression<bool> convertedExpression)
        {
            ErrorExpression errorExpression = null;
            convertedExpression = null;

            if ((isDirect && directOption == ArgumentInterpretation.TreatAsError) || (!isDirect && indirectOption == ArgumentInterpretation.TreatAsError))
            {
                errorExpression = ErrorExpressions.ValueError;
            }
            else if ((isDirect && directOption == ArgumentInterpretation.UseAsIs) || (!isDirect && indirectOption == ArgumentInterpretation.UseAsIs))
            {
                convertedExpression = value.ToBooleanExpression();
            }
            else if ((isDirect && directOption == ArgumentInterpretation.ConvertToDefault) || (!isDirect && indirectOption == ArgumentInterpretation.ConvertToDefault))
            {
                convertedExpression = BooleanExpression.False;
            }

            return errorExpression;
        }

        private static ErrorExpression EmptyToBoolExpression(RadExpression expressionValue, bool isDirect, ArgumentConversionRules convertionRules, out ConstantExpression<bool> convertedExpression)
        {
            ErrorExpression errorExpression = null;
            convertedExpression = null;

            if (expressionValue is EmptyExpression)
            {
                errorExpression = ToBoolExpression(false, isDirect, convertionRules.EmptyDirectArgument, convertionRules.EmptyIndirectArgument, out convertedExpression);
            }

            return errorExpression;
        }

        private static ErrorExpression NumberToBoolExpression(RadExpression expression, bool isDirect, ArgumentConversionRules convertionRules, out ConstantExpression<bool> convertedExpression)
        {
            ErrorExpression errorExpression = null;
            convertedExpression = null;

            NumberExpression numberExpression = expression as NumberExpression;
            if (numberExpression != null)
            {
                bool value = numberExpression.Value == 0 ? false : true;
                errorExpression = ToBoolExpression(value, isDirect, convertionRules.NumberDirectArgument, convertionRules.NumberIndirectArgument, out convertedExpression);
            }

            return errorExpression;
        }

        private static ErrorExpression BoolToBoolExpression(RadExpression expression, bool isDirect, ArgumentConversionRules convertionRules, out ConstantExpression<bool> convertedExpression)
        {
            ErrorExpression errorExpression = null;
            convertedExpression = null;

            BooleanExpression booleanExpression = expression as BooleanExpression;
            if (booleanExpression != null)
            {
                errorExpression = ToBoolExpression(booleanExpression.Value, isDirect, convertionRules.BoolDirectArgument, convertionRules.BoolIndirectArgument, out convertedExpression);
            }

            return errorExpression;
        }

        private static ErrorExpression TextNumberToBoolExpression(RadExpression expression, bool isDirect, ArgumentConversionRules convertionRules, out ConstantExpression<bool> convertedExpression)
        {
            ErrorExpression errorExpression = null;
            convertedExpression = null;

            StringExpression stringExpression = expression as StringExpression;
            if (stringExpression != null)
            {
                double stringAsDouble;
                if (double.TryParse(stringExpression.Value, out stringAsDouble))
                {
                    bool value = stringAsDouble == 0 ? false : true;
                    errorExpression = ToBoolExpression(value, isDirect, convertionRules.TextNumberDirectArgument, convertionRules.TextNumberIndirectArgument, out convertedExpression);
                }
            }

            return errorExpression;
        }

        private static ErrorExpression NonTextNumbersToBoolExpression(RadExpression expression, bool isDirect, ArgumentConversionRules convertionRules, out ConstantExpression<bool> convertedExpression)
        {
            ErrorExpression errorExpression = null;
            convertedExpression = null;

            StringExpression stringExpression = expression as StringExpression;
            if (stringExpression != null)
            {
                double stringAsDouble;
                bool isNonTextNumber = !(double.TryParse(stringExpression.Value, out stringAsDouble));

                if (isNonTextNumber)
                {
                    bool value = stringAsDouble == 0 ? false : true;
                    errorExpression = ToBoolExpression(value, isDirect, convertionRules.NonTextNumberDirectArgument, convertionRules.NonTextNumberIndirectArgument, out convertedExpression);
                }
            }

            return errorExpression;
        }

        #endregion


        #region ToString Convertion

        private static ErrorExpression TryConvertArgumentToString(RadExpression expression, bool isDirectArgument, ArgumentConversionRules convertionRules, out ConstantExpression<string> convertedArgument)
        {
            ErrorExpression errorExpression = null;
            convertedArgument = null;

            RadExpression expressionValue = expression.GetValue();

            errorExpression = expressionValue as ErrorExpression;

            if (errorExpression == null)
            {
                for (int i = 0; i < stringConverters.Length; i++)
                {
                    errorExpression = stringConverters[i](expressionValue, isDirectArgument, convertionRules, out convertedArgument);
                    if (errorExpression != null)
                    {
                        return errorExpression;
                    }
                    else if (convertedArgument != null)
                    {
                        return null;
                    }
                }
            }

            return errorExpression;
        }

        private static ErrorExpression ToStringExpression(string value, bool isDirect, ArgumentInterpretation directOption, ArgumentInterpretation indirectOption, out ConstantExpression<string> convertedExpression)
        {
            ErrorExpression errorExpression = null;
            convertedExpression = null;

            if ((isDirect && directOption == ArgumentInterpretation.TreatAsError) || (!isDirect && indirectOption == ArgumentInterpretation.TreatAsError))
            {
                errorExpression = ErrorExpressions.ValueError;
            }
            else if ((isDirect && directOption == ArgumentInterpretation.UseAsIs) || (!isDirect && indirectOption == ArgumentInterpretation.UseAsIs))
            {
                convertedExpression = new StringExpression(value);
            }
            else if ((isDirect && directOption == ArgumentInterpretation.ConvertToDefault) || (!isDirect && indirectOption == ArgumentInterpretation.ConvertToDefault))
            {
                convertedExpression = new EmptyExpression();
            }

            return errorExpression;
        }

        private static ErrorExpression BoolToStringExpression(RadExpression expression, bool isDirect, ArgumentConversionRules convertionRules, out ConstantExpression<string> convertedExpression)
        {
            ErrorExpression errorExpression = null;
            convertedExpression = null;

            BooleanExpression booleanExpression = expression as BooleanExpression;
            if (booleanExpression != null)
            {
                string value = booleanExpression.ToString();
                errorExpression = ToStringExpression(value, isDirect, convertionRules.BoolDirectArgument, convertionRules.BoolIndirectArgument, out convertedExpression);
            }

            return errorExpression;
        }

        // TODO: Text arguments ignore the Conversion rules
        private static ErrorExpression AllToStringExpression(RadExpression expression, bool isDirect, ArgumentConversionRules convertionRules, out ConstantExpression<string> convertedExpression)
        {
            convertedExpression = null;

            RadExpression expressionValue = expression.GetValue();

            ErrorExpression errorExpression = expressionValue as ErrorExpression;
            if (errorExpression != null)
            {
                return errorExpression;
            }

            convertedExpression = new StringExpression(expressionValue.GetValueAsString());

            return null;
        }

        #endregion
    }
}