﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Describes the category of the function.
    /// </summary>
    public enum FunctionCategory
    {
        /// <summary>
        /// Financial function category.
        /// </summary>
        Financial,

        /// <summary>
        /// DateTime function category.
        /// </summary>
        DateTime,

        /// <summary>
        /// MathTrig function category.
        /// </summary>
        MathTrig,

        /// <summary>
        /// Statistical function category.
        /// </summary>
        Statistical,

        /// <summary>
        /// LookupReference function category.
        /// </summary>
        LookupReference,
        //Database,

        /// <summary>
        /// Text function category.
        /// </summary>
        Text,

        /// <summary>
        /// Logical function category.
        /// </summary>
        Logical,

        /// <summary>
        /// Information function category.
        /// </summary>
        Information,

        /// <summary>
        /// Engineering function category.
        /// </summary>
        Engineering,
        //Cube,
        //Compatibility
    }

    /// <summary>
    /// Provides properties for describing a FunctionBase.
    /// </summary>
    public class FunctionInfo
    {
        #region Fields

        private readonly string name;
        private readonly FunctionCategory category;
        private readonly string description;
        private readonly string descriptionLocalizationKey;
        private readonly bool isDefaultValueFunction;
        private readonly List<ArgumentInfo> requiredArgumentsInfos;
        private readonly List<ArgumentInfo> optionalArgumentsInfos;
        private readonly int optionalArgumentsRepetitionCount;
        private readonly CellValueFormat format;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the category.
        /// </summary>
        /// <value>The category.</value>
        /// <returns>The category as FunctionCategory.</returns>
        public FunctionCategory Category
        {
            get
            {
                return category;
            }
        }

        /// <summary>
        /// Gets the description.
        /// </summary>
        /// <value>The description.</value>
        /// <returns>The description as String.</returns>
        public string Description
        {
            get
            {
                return description;
            }
        }

        /// <summary>
        /// Gets the description localization key.
        /// </summary>
        /// <value>The description localization key.</value>
        public string DescriptionLocalizationKey
        {
            get
            {
                return descriptionLocalizationKey;
            }
        }

        /// <summary>
        /// Gets the required arguments count.
        /// </summary>
        /// <value>The required arguments count.</value>
        /// <returns>The required arguments count as Int32.</returns>
        public int RequiredArgumentsCount
        {
            get
            {
                return this.requiredArgumentsInfos.Count();
            }
        }

        /// <summary>
        /// Gets the optional arguments count.
        /// </summary>
        /// <value>The optional arguments count.</value>
        /// <returns>The optional arguments count as Int32.</returns>
        public int OptionalArgumentsCount
        {
            get
            {
                return this.optionalArgumentsInfos.Count();
            }
        }

        /// <summary>
        /// Gets the optional arguments repetition count.
        /// </summary>
        /// <value>The optional arguments repetition count.</value>
        /// <value>The optional arguments repetition count as Int32.</value>
        public int OptionalArgumentsRepetitionCount
        {
            get
            {
                return this.optionalArgumentsRepetitionCount;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the function is default value function.
        /// </summary>
        /// <value>Boolean indicating whether the function is default value function.</value>
        /// <returns>Boolean indicating whether the function is default value function.</returns>
        public bool IsDefaultValueFunction
        {
            get
            {
                return this.isDefaultValueFunction;
            }
        }

        /// <summary>
        /// Gets the format.
        /// </summary>
        /// <value>The format.</value>
        /// <returns>The format as CellValueFormat.</returns>
        public CellValueFormat Format
        {
            get
            {
                return this.format;
            }
        }

        internal string Name
        {
            get
            {
                return this.name;
            }
        }

        internal int TotalArgumentsCount
        {
            get
            {
                return this.RequiredArgumentsCount + this.OptionalArgumentsCount * this.OptionalArgumentsRepetitionCount;
            }
        }

        internal List<ArgumentInfo> RequiredArguments
        {
            get
            {
                return this.requiredArgumentsInfos;
            }
        }

        internal List<ArgumentInfo> OptionalArguments
        {
            get
            {
                return this.optionalArgumentsInfos;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="FunctionInfo" /> class.
        /// </summary>
        /// <param name="name">The name of the function.</param>
        /// <param name="category">The category of the function.</param>
        /// <param name="description">The description of the function.</param>
        /// <param name="descriptionLocalizationKey">The localization key of the function description.</param>
        public FunctionInfo(string name, FunctionCategory category, string description, string descriptionLocalizationKey = null)
            : this(name, category, description, Enumerable.Empty<ArgumentInfo>(), Enumerable.Empty<ArgumentInfo>(), CellValueFormat.GeneralFormat, descriptionLocalizationKey: descriptionLocalizationKey)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="FunctionInfo" /> class.
        /// </summary>
        /// <param name="name">The name of the function.</param>
        /// <param name="category">The category of the function.</param>
        /// <param name="description">The description of the function.</param>
        /// <param name="format">The format of the function.</param>
        /// <param name="descriptionLocalizationKey">The localization key of the function description.</param>
        public FunctionInfo(string name, FunctionCategory category, string description, CellValueFormat format, string descriptionLocalizationKey = null)
            : this(name, category, description, Enumerable.Empty<ArgumentInfo>(), Enumerable.Empty<ArgumentInfo>(), format, descriptionLocalizationKey: descriptionLocalizationKey)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="FunctionInfo" /> class.
        /// </summary>
        /// <param name="name">The name of the function.</param>
        /// <param name="category">The category of the function.</param>
        /// <param name="description">The description of the function.</param>
        /// <param name="requiredArgumentsInfos">The required arguments infos for the function.</param>
        /// <param name="isDefaultValueFunction">Boolean indicating whether the function is default value function.</param>
        /// <param name="descriptionLocalizationKey">The localization key of the function description.</param>
        public FunctionInfo(string name, FunctionCategory category, string description, IEnumerable<ArgumentInfo> requiredArgumentsInfos,
            bool isDefaultValueFunction = false, string descriptionLocalizationKey = null)
            : this(name, category, description, requiredArgumentsInfos, Enumerable.Empty<ArgumentInfo>(), CellValueFormat.GeneralFormat,
            isDefaultValueFunction: isDefaultValueFunction, descriptionLocalizationKey: descriptionLocalizationKey)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="FunctionInfo" /> class.
        /// </summary>
        /// <param name="name">The name of the function.</param>
        /// <param name="category">The category of the function.</param>
        /// <param name="description">The description of the function.</param>
        /// <param name="requiredArgumentsInfos">The required arguments infos for the function.</param>
        /// <param name="format">The format of the function.</param>
        /// <param name="isDefaultValueFunction">Boolean indicating whether the function is default value function.</param>
        /// <param name="descriptionLocalizationKey">The localization key of the function description.</param>
        public FunctionInfo(string name, FunctionCategory category, string description, IEnumerable<ArgumentInfo> requiredArgumentsInfos,
            CellValueFormat format, bool isDefaultValueFunction = false, string descriptionLocalizationKey = null)
            : this(name, category, description, requiredArgumentsInfos, Enumerable.Empty<ArgumentInfo>(), format,
            isDefaultValueFunction: isDefaultValueFunction, descriptionLocalizationKey: descriptionLocalizationKey)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="FunctionInfo" /> class.
        /// </summary>
        /// <param name="name">The name of the function.</param>
        /// <param name="category">The category of the function.</param>
        /// <param name="description">The description of the function.</param>
        /// <param name="requiredArgumentsInfos">The required arguments infos for the function.</param>
        /// <param name="optionalArgumentsInfos">The optional arguments infos for the function.</param>
        /// <param name="optionalArgumentsRepeatCount">The count of the optional arguments for the function.</param>
        /// <param name="isDefaultValueFunction">Boolean indicating whether the function is default value function.</param>
        /// <param name="descriptionLocalizationKey">The localization key of the function description.</param>
        public FunctionInfo(string name, FunctionCategory category, string description, IEnumerable<ArgumentInfo> requiredArgumentsInfos,
            IEnumerable<ArgumentInfo> optionalArgumentsInfos, int optionalArgumentsRepeatCount = 1, bool isDefaultValueFunction = false, string descriptionLocalizationKey = null)
            : this(name, category, description, requiredArgumentsInfos, optionalArgumentsInfos, CellValueFormat.GeneralFormat, optionalArgumentsRepeatCount,
            isDefaultValueFunction, descriptionLocalizationKey: descriptionLocalizationKey)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="FunctionInfo" /> class.
        /// </summary>
        /// <param name="name">The name of the function.</param>
        /// <param name="category">The category of the function.</param>
        /// <param name="description">The description of the function.</param>
        /// <param name="requiredArgumentsInfos">The required arguments infos for the function.</param>
        /// <param name="optionalArgumentsInfos">The optional arguments infos for the function.</param>
        /// <param name="format">The format of the function.</param>
        /// <param name="optionalArgumentsRepeatCount">The count of the optional arguments for the function.</param>
        /// <param name="isDefaultValueFunction">Boolean indicating whether the function is default value function.</param>
        /// <param name="descriptionLocalizationKey">The localization key of the function description.</param>
        public FunctionInfo(string name, FunctionCategory category, string description, IEnumerable<ArgumentInfo> requiredArgumentsInfos,
            IEnumerable<ArgumentInfo> optionalArgumentsInfos, CellValueFormat format, int optionalArgumentsRepeatCount = 1, bool isDefaultValueFunction = false, string descriptionLocalizationKey = null)
        {
            this.name = name;
            this.category = category;
            this.description = description;
            this.optionalArgumentsRepetitionCount = optionalArgumentsRepeatCount;
            this.descriptionLocalizationKey = descriptionLocalizationKey;
            this.requiredArgumentsInfos = new List<ArgumentInfo>(requiredArgumentsInfos);
            this.optionalArgumentsInfos = new List<ArgumentInfo>(optionalArgumentsInfos);
            this.format = format;
            this.isDefaultValueFunction = isDefaultValueFunction;
        }

        #endregion


        #region Methods

        internal IEnumerable<ArgumentInfo> GetAllArguments()
        {
            List<ArgumentInfo> result = new List<ArgumentInfo>();

            result.AddRange(this.requiredArgumentsInfos);

            for (int i = 0; i < this.OptionalArgumentsRepetitionCount; i++)
            {
                result.AddRange(this.optionalArgumentsInfos);
            }

            return result;
        }

        internal IEnumerable<ArgumentInfo> GetArguments(int currentArgumentIndex)
        {
            List<ArgumentInfo> result = new List<ArgumentInfo>();

            result.AddRange(this.requiredArgumentsInfos);

            int counter = 1;
            result.AddRange(this.GetOptionalArguments(counter));

            if (result.Count == 1)
            {
                counter++;
                result.AddRange(this.GetOptionalArguments(counter));
            }

            counter++;
            while (currentArgumentIndex >= result.Count &&
                this.optionalArgumentsInfos.Count() > 0)
            {
                result.AddRange(this.GetOptionalArguments(counter));
                counter++;
            }

            return result;
        }

        private List<ArgumentInfo> GetOptionalArguments(int count)
        {
            List<ArgumentInfo> result = new List<ArgumentInfo>();

            if (this.optionalArgumentsInfos.Count() == 0)
            {
                return result;
            }

            foreach (var argumentInfo in this.optionalArgumentsInfos)
            {
                if (count == 1)
                {
                    result.Add(new ArgumentInfo(
                        name: argumentInfo.Name + count,
                        description: argumentInfo.Description,
                        type: argumentInfo.Type));
                }
                else
                {
                    result.Add(new ArgumentInfo(
                        name: argumentInfo.Name + count,
                        description: argumentInfo.Description,
                        type: argumentInfo.Type,
                        isRequired: false));
                }
            }

            return result;
        }

        #endregion
    }
}