﻿using System.Collections.Generic;
using System;
using System.Linq;
using System.Reflection;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Provides methods for registering functions and accessing registered functions in a RadSpreadsheet instance.
    /// </summary>
    public static class FunctionManager
    {
        #region Fields

        private static readonly Dictionary<string, FunctionBase> functionNameToFunction;

        #endregion


        #region Constructors

        static FunctionManager()
        {
            functionNameToFunction = new Dictionary<string, FunctionBase>();

            foreach (Type type in Assembly.GetExecutingAssembly().GetTypes().Where(type => !type.IsAbstract && type.IsSubclassOf(typeof(FunctionBase))))
            {
                RegisterFunction((FunctionBase)Activator.CreateInstance(type));
            }
        }

        #endregion


        #region Methods

        /// <summary>
        /// Unregisters the function.
        /// </summary>
        /// <param name="function">The function.</param>
        public static void UnregisterFunction(FunctionBase function)
        {
            UnregisterFunction(function.Name);
        }

        /// <summary>
        /// Unregisters the function.
        /// </summary>
        /// <param name="functionName">Name of the function.</param>
        public static void UnregisterFunction(string functionName)
        {
            string upperFunctionName = functionName.ToUpper();
            if (functionNameToFunction.ContainsKey(upperFunctionName))
            {
                functionNameToFunction.Remove(upperFunctionName);
            }
        }

        /// <summary>
        /// Registers a function.
        /// </summary>
        /// <param name="function">The function.</param>
        public static void RegisterFunction(FunctionBase function)
        {
            functionNameToFunction[function.Name.ToUpper()] = function;
        }

        /// <summary>
        /// Gets a FunctionBase instance by a specified functionName.
        /// </summary>
        /// <param name="functionName">Name of the function.</param>
        /// <returns>An FunctionBase instance.</returns>
        public static FunctionBase GetFunctionByName(string functionName)
        {
            FunctionBase result;
            if (functionNameToFunction.TryGetValue(functionName.ToUpper(), out result))
            {
                return result;
            }

            return null;
        }

        /// <summary>
        /// Gets all registered functions.
        /// </summary>
        /// <returns>FunctionBase instances of all registered functions.</returns>
        public static IEnumerable<FunctionBase> GetAllFunctions()
        {
            return functionNameToFunction.Values.OrderBy(p => p.Name);
        }

        #endregion
    }
}
