﻿using System;
using System.Collections.Generic;
using System.ComponentModel;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the ERROR.TYPE function.
    /// </summary>
    public class ErrorType : FunctionBase
    {
        #region Constants

		///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "ERROR.TYPE";
        private static readonly FunctionInfo Info;

		#endregion


		#region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
			get
			{
				return FunctionName;
			}
		}

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

		#endregion


        #region Constructors

        static ErrorType()
        {
            string description = "Returns a number corresponding to one of the error values in Microsoft Excel or returns the #N/A error if no error exists. You can use ERROR.TYPE in an IF function to test for an error value and return a text string, such as a message, instead of the error value.";
            string descriptionKey = "Spreadsheet_Functions_ErrorType_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Value", 
                                 "is the error value whose identifying number you want to find. Although error_val can be the actual error value, it will usually be a reference to a cell containing a formula that you want to test.",
                                 ArgumentType.Any,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Value",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_ErrorType_ValueInfo")
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Information, description, requiredArguments, descriptionLocalizationKey:descriptionKey);
        }

        #endregion


		#region Methods

        /// <summary>
        /// Virtual method that evaluates the function with the specified context.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Function result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<RadExpression> context)
        {
            double result = 0;
            ErrorExpression errorExpression = context.Arguments[0].GetValueAsConstantExpression() as ErrorExpression;

            if (errorExpression != null)
            {
                if (errorExpression == ErrorExpressions.CyclicReference)
                {
                    return ErrorExpressions.CyclicReference;
                }
                else if (errorExpression == ErrorExpressions.NullError)
                {
                    result = 1;
                }
                else if (errorExpression == ErrorExpressions.DivisionByZero)
                {
                    result = 2;
                }
                else if (errorExpression == ErrorExpressions.ValueError)
                {
                    result = 3;
                }
                else if (errorExpression == ErrorExpressions.ReferenceError)
                {
                    result = 4;
                }
                else if (errorExpression == ErrorExpressions.NameError)
                {
                    result = 5;
                }
                else if (errorExpression == ErrorExpressions.NumberError)
                {
                    result = 6;
                }
                else if (errorExpression == ErrorExpressions.NotAvailableError)
                {
                    result = 7;
                }
            }

            if (result == 0)
            {
                return ErrorExpressions.NotAvailableError;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

		#endregion

    }
}