﻿using System;
using System.Collections.Generic;
using System.ComponentModel;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the IF function.
    /// </summary>
    public class IfFunction : FunctionBase
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "IF";
        private static readonly FunctionInfo Info;
        private static readonly ArgumentConversionRules ConvertionRules = new ArgumentConversionRules(
            emptyDirectArgument: ArgumentInterpretation.ConvertToDefault,
            emptyIndirectArgument: ArgumentInterpretation.ConvertToDefault,
            nonTextNumberDirectArgument: ArgumentInterpretation.TreatAsError,
            nonTextNumberIndirectArgument: ArgumentInterpretation.TreatAsError);

        #endregion


        #region Properties

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return ConvertionRules;
            }
        }

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static IfFunction()
        {
            string description = "The IF function returns one value if a condition you specify evaluates to TRUE, and another value if that condition evaluates to FALSE. For example, the formula =IF(A1>10,\"Over 10\",\"10 or less\") returns \"Over 10\" if A1 is greater than 10, and \"10 or less\" if A1 is less than or equal to 10.";
            string descriptionKey = "Spreadsheet_Functions_If_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Logical test",
                                 "Any value or expression that can be evaluated to TRUE or FALSE. For example, A10=100 is a logical expression; if the value in cell A10 is equal to 100, the expression evaluates to TRUE. Otherwise, the expression evaluates to FALSE. This argument can use any comparison calculation operator.",
                                 ArgumentType.Logical,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_LogicalTest",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_If_LogicalTestInfo"),
                new ArgumentInfo("Value if true",
                                 "The value that you want to be returned if the logical test argument evaluates to TRUE. For example, if the value of this argument is the text string \"Within budget\" and the logical test argument evaluates to TRUE, the IF function returns the text \"Within budget.\" If logical test evaluates to TRUE and the value if true argument is omitted (that is, there is only a comma following the logical test argument), the IF function returns 0 (zero). To display the word TRUE, use the logical value TRUE for the value if true argument.",
                                 ArgumentType.Any,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_ValueIfTrue",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_If_ValueIfTrueInfo"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Value if false",
                                 "The value that you want to be returned if the logical test argument evaluates to FALSE. For example, if the value of this argument is the text string \"Over budget\" and the logical test argument evaluates to FALSE, the IF function returns the text \"Over budget.\" If logical test evaluates to FALSE and the value if false argument is omitted, (that is, there is no comma following the value if true argument), the IF function returns the logical value FALSE. If logical test evaluates to FALSE and the value of the value if false argument is blank (that is, there is only a comma following the value if true argument), the IF function returns the value 0 (zero).",
                                 ArgumentType.Any,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_ValueIfFalse",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_If_ValueIfFalseInfo")
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Logical, description, requiredArguments, optionalArguments, 1, false, descriptionLocalizationKey:descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method that evaluates the function with the specified context.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Function result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<RadExpression> context)
        {
            object convertedArgument;
            ErrorExpression error = FunctionHelper.TryConvertFirstArgument(context.Arguments[0], this.ArgumentConversionRules, ArgumentType.Logical, out convertedArgument);

            if (error != null)
            {
                return error;
            }

            RadExpression trueExpression = context.Arguments[1];
            RadExpression falseExpression = BooleanExpression.False;

            if (context.Arguments.Length > 2)
            {
                falseExpression = context.Arguments[2];
            }

            RadExpression resultExpression = ((bool)convertedArgument ? trueExpression : falseExpression).GetValueAsConstantExpression();

            if (resultExpression is EmptyExpression)
            {
                resultExpression = NumberExpression.Zero;
            }

            return resultExpression;
        }

        #endregion
    }
}