﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the HYPERLINK function.
    /// </summary>
    public class HyperlinkFunction : FunctionWithArguments
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "HYPERLINK";
        private static readonly FunctionInfo Info;

        #endregion


        #region Fields

        private HyperlinkInfo hyperlinkInfo;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        internal HyperlinkInfo HyperlinkInfo
        {
            get
            {
                return this.hyperlinkInfo;
            }
        }

        #endregion


        #region Constructors

        static HyperlinkFunction()
        {
            string description = "Creates a shortcut or jump that opens a document stored on a network server, an intranet, or the Internet.";
            string descriptionKey = "Spreadsheet_Functions_Hyperlink_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                 new ArgumentInfo("Link Location",
                                  "is the path and file name to the document to be opened. Link Location can refer to a place in a document — such as a specific cell or named range in a worksheet or workbook. The path can be to a file that is stored on a hard disk drive.",
                                  ArgumentType.Text,
                                  nameLocalizationKey:"Spreadsheet_Functions_Args_LinkLocation",
                                  descriptionLocalizationKey:"Spreadsheet_Functions_Hyperlink_LinkLocationInfo"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                 new ArgumentInfo("Friendly Name",
                                  "specifies the jump text or numeric value that is displayed in the cell. Friendly Name is displayed in blue and is underlined. If Friendly Name is omitted, the cell displays the Link Location as the jump text.",
                                  ArgumentType.Any,
                                  nameLocalizationKey:"Spreadsheet_Functions_Args_FriendlyName",
                                  descriptionLocalizationKey:"Spreadsheet_Functions_Hyperlink_FriendlyNameInfo")
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.LookupReference, description, requiredArguments, optionalArguments, 1, false, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<object> context)
        {
            string address;
            string subAddress;
            string emailAddress;
            string emailSubject;

            SpreadsheetHelper.TryParseHyperlinkAddress(context.Arguments[0].ToString(), out address, out subAddress, out emailAddress, out emailSubject);
            string tooltip = string.Format("{0} - Click once to follow. Click and hold to select this cell.", context.Arguments[0].ToString());

            HyperlinkInfo hyperlinkInfo;
            if (!string.IsNullOrEmpty(emailAddress))
            {
                hyperlinkInfo = HyperlinkInfo.CreateMailtoHyperlink(emailAddress, emailSubject, tooltip);
            }
            else
            {
                hyperlinkInfo = HyperlinkInfo.CreateHyperlink(address, tooltip);
            }

            this.hyperlinkInfo = hyperlinkInfo;

            RadExpression result;
            if (context.Arguments.Length > 1)
            {
                result = (RadExpression)context.Arguments[1];
            }
            else
            {
                result = new StringExpression(context.Arguments[0].ToString());
            }

            return result;
        }

        #endregion
    }
}