﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the INDEX function.
    /// </summary>
    public class Index : FunctionWithArguments
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "INDEX";
        private static readonly FunctionInfo Info;
        private static readonly ArgumentConversionRules convertionRules = new ArgumentConversionRules(
            emptyDirectArgument: ArgumentInterpretation.ConvertToDefault,
            numberDirectArgument: ArgumentInterpretation.UseAsIs,
            boolDirectArgument: ArgumentInterpretation.UseAsIs,
            textNumberDirectArgument: ArgumentInterpretation.UseAsIs,
            nonTextNumberDirectArgument: ArgumentInterpretation.UseAsIs,
            emptyIndirectArgument: ArgumentInterpretation.ConvertToDefault,
            numberIndirectArgument: ArgumentInterpretation.UseAsIs,
            boolIndirectArgument: ArgumentInterpretation.UseAsIs,
            textNumberIndirectArgument: ArgumentInterpretation.UseAsIs,
            nonTextNumberIndirectArgument: ArgumentInterpretation.UseAsIs,
            arrayArgument: ArrayArgumentInterpretation.UseFirstElement);

        #endregion


        #region Properties

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return Index.convertionRules;
            }
        }

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Index()
        {
            string description = "Returns the reference of the cell at the intersection of a particular row and column. If the reference is made up of nonadjacent selections, you can pick the selection to look in.";
            string descriptionKey = "Spreadsheet_Functions_Index_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Reference",
                    "A reference to one or more cell ranges. If you are entering a nonadjacent range for the reference, enclose reference in parentheses. If each area in reference contains only one row or column, the Row_num or Column_num argument, respectively, is optional. For example, for a single row reference, use INDEX(reference,,column_num).",
                    ArgumentType.Any,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_Reference",
                    descriptionLocalizationKey:"Spreadsheet_Functions_Index_Reference"),

                new ArgumentInfo("Row_num",
                    "The number of the row in reference from which to return a reference.",
                    ArgumentType.Number,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_RowNum",
                    descriptionLocalizationKey:"Spreadsheet_Functions_Index_RowNum"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Column_num",
                    "The number of the column in reference from which to return a reference.",
                    ArgumentType.Number,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_ColumnNum",
                    descriptionLocalizationKey:"Spreadsheet_Functions_Index_ColumnNum"),

                new ArgumentInfo("Area_num",
                    "Selects a range in reference from which to return the intersection of Row_num and Column_num. The first area selected or entered is numbered 1, the second is 2, and so on. If Area_num is omitted, INDEX uses area 1.",
                    ArgumentType.Number,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_AreaNum",
                    descriptionLocalizationKey:"Spreadsheet_Functions_Index_AreaNum"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.LookupReference, description, requiredArguments, optionalArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<object> context)
        {
            int rowNum, columnNum, areaNum;

            if (!Index.TryGetNonNegativeRowColumnAreaNumsArguments(context.Arguments, out rowNum, out columnNum, out areaNum))
            {
                return ErrorExpressions.ValueError;
            }

            ArrayExpression arrayExpression;
            if (Index.TryGetArrayArgument(context.Arguments[0], out arrayExpression))
            {
                if (areaNum > 1)
                {
                    return ErrorExpressions.ReferenceError;
                }

                return Index.HandleArrayOverload(arrayExpression, rowNum, columnNum);
            }

            CellReferenceRangeExpression referenceRangeExpression;
            if (Index.TryGetReferenceArgument(context.Arguments[0], out referenceRangeExpression))
            {
                CellReferenceRange referenceRange;
                ErrorExpression error = TryGetCellReferenceRange(referenceRangeExpression, areaNum, out referenceRange);
                if (error != null)
                {
                    return error;
                }

                bool hasOptionalArguments = (context.Arguments.Length > 2);
                error = TryRecalculateZeroRowColumnArguments(referenceRange, hasOptionalArguments, ref rowNum, ref columnNum);
                if (error != null)
                {
                    return error;
                }

                return Index.HandleReferenceRangeOverload(referenceRange, rowNum, columnNum);
            }

            return ErrorExpressions.ValueError;
        }
        
        private static ErrorExpression TryGetCellReferenceRange(CellReferenceRangeExpression referenceRangeExpression, int areaNum, out CellReferenceRange referenceRange)
        {
            referenceRange = null;
            Guard.ThrowExceptionIfLessThan(1, areaNum, "areaNum");

            if (!referenceRangeExpression.IsInRange)
            {
                return ErrorExpressions.ReferenceError;
            }

            if (!referenceRangeExpression.IsValid)
            {
                return ErrorExpressions.ValueError;
            }

            if (areaNum > referenceRangeExpression.CellReferenceRanges.Count)
            {
                return ErrorExpressions.ReferenceError;
            }

            referenceRange = referenceRangeExpression.CellReferenceRanges[areaNum - 1];

            return null;
        }

        private static ErrorExpression TryRecalculateZeroRowColumnArguments(
            CellReferenceRange referenceRange, bool hasOptionalArguments, ref int rowNum, ref int columnNum)
        {
            if (referenceRange.ColumnsCount > 1 && referenceRange.RowsCount > 1)
            {
                if (rowNum == 0 || columnNum == 0)
                {
                    return hasOptionalArguments ? ErrorExpressions.ValueError : ErrorExpressions.ReferenceError;
                }
                else
                {
                    return null;
                }
            }

            if (columnNum == 0)
            {
                if (referenceRange.ColumnsCount == 1 || !hasOptionalArguments)
                {
                    columnNum = 1;
                }
                else
                {
                    return ErrorExpressions.ValueError;
                }
            }

            if (rowNum == 0)
            {
                if (referenceRange.RowsCount == 1)
                {
                    rowNum = 1;
                }
                else
                {
                    return ErrorExpressions.ValueError;
                }
            } 

            return null;
        }

        private static RadExpression HandleArrayOverload(ArrayExpression arrayExpression, int rowNum, int columnNum)
        {
            Guard.ThrowExceptionIfLessThan(0, rowNum, "rowNum");
            Guard.ThrowExceptionIfLessThan(0, columnNum, "columnNum");

            if (rowNum > 0 && columnNum > 0)
            {
                if (rowNum <= arrayExpression.RowCount && columnNum <= arrayExpression.ColumnCount)
                {
                    return arrayExpression[rowNum - 1, columnNum - 1];
                }
            }
            else if (rowNum > 0 && columnNum == 0)
            {
                if (rowNum <= arrayExpression.RowCount)
                {
                    return arrayExpression.GetRowAsArrayExpression(rowNum - 1);
                }
            }
            else if (columnNum > 0 && rowNum == 0)
            {
                if (columnNum <= arrayExpression.ColumnCount)
                {
                    return arrayExpression.GetColumnAsArrayExpression(columnNum - 1);
                }
            }
            else if (rowNum == 0 && columnNum == 0)
            {
                return arrayExpression;
            }

            return ErrorExpressions.ReferenceError;
        }

        private static RadExpression HandleReferenceRangeOverload(CellReferenceRange referenceRange, int rowNum, int columnNum)
        {
            Guard.ThrowExceptionIfLessThan(1, rowNum, "rowNum");
            Guard.ThrowExceptionIfLessThan(1, columnNum, "columnNum");

            int topLeftRowIndex = referenceRange.FromCellReference.ActualRowIndex;
            int topLeftColumnIndex = referenceRange.FromCellReference.ActualColumnIndex;

            int actualRowIndex = topLeftRowIndex + rowNum - 1;
            int actualColumnIndex = topLeftColumnIndex + columnNum - 1;

            if (rowNum <= referenceRange.RowsCount &&
                columnNum <= referenceRange.ColumnsCount)
            {
                var referenceArea = new CellAreaReference(new CellReference(
                    RowColumnReference.CreateAbsoluteRowColumnReference(actualRowIndex),
                    RowColumnReference.CreateAbsoluteRowColumnReference(actualColumnIndex)));

                return new CellReferenceRangeExpression(referenceRange.Worksheet.Workbook, referenceRange.Worksheet.Name, true, referenceArea);
            }

            return ErrorExpressions.ReferenceError;
        }

        private static bool TryGetReferenceArgument(object argument, out CellReferenceRangeExpression referenceRangeExpression)
        {
            referenceRangeExpression = argument as CellReferenceRangeExpression;

            return referenceRangeExpression != null;
        }

        private static bool TryGetArrayArgument(object argument, out ArrayExpression arrayExpression)
        {
            arrayExpression = FunctionHelper.ConvertToArrayExpression(argument as ConstantExpression);

            return arrayExpression != null;
        }

        private static bool TryGetNonNegativeRowColumnAreaNumsArguments(object[] arguments, out int rowNum, out int columnNum, out int areaNum)
        {
            rowNum = (int)Math.Floor((double)arguments[1]);
            columnNum = 0;
            areaNum = 1;

            if (arguments.Length > 2)
            {
                columnNum = (int)Math.Floor((double)arguments[2]);

                if (arguments.Length == 4)
                {
                    areaNum = (int)Math.Floor((double)arguments[3]);
                }
            }

            return rowNum >= 0 && columnNum >= 0 && areaNum > 0;
        }

        #endregion
    }
}