﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the INDIRECT function.
    /// </summary>
    public class Indirect : FunctionWithArguments
    {
        #region Constants
        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "INDIRECT";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Indirect()
        {
            string description = "Returns the reference of the cell specified by a text string.";
            string descriptionKey = "Spreadsheet_Functions_Indirect_Info";

            IEnumerable<Telerik.Windows.Documents.Spreadsheet.Expressions.Functions.ArgumentInfo> requiredArguments = new Telerik.Windows.Documents.Spreadsheet.Expressions.Functions.ArgumentInfo[]
            {
                new Telerik.Windows.Documents.Spreadsheet.Expressions.Functions.ArgumentInfo("Ref_text",
                    "A reference to a cell that contains an A1-style reference, a name defined as a reference, or a reference to a cell as a text string.",
                    Telerik.Windows.Documents.Spreadsheet.Expressions.Functions.ArgumentType.Text,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_RefText",
                    descriptionLocalizationKey:"Spreadsheet_Functions_Indirect_RefText"),
            };

            Info = new Telerik.Windows.Documents.Spreadsheet.Expressions.Functions.FunctionInfo(FunctionName, Telerik.Windows.Documents.Spreadsheet.Expressions.Functions.FunctionCategory.LookupReference, description, requiredArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<object> context)
        {
            string reference = context.Arguments[0].ToString();

            if (string.IsNullOrEmpty(reference))
            {
                return ErrorExpressions.ReferenceError;
            }

            List<CellReferenceRange> cellReferenceRanges = new List<CellReferenceRange>();

            CellReferenceRangeExpression expression;
            if (NameConverter.TryConvertNamesToCellReferenceRangeExpression(reference, context.Worksheet, context.RowIndex, context.ColumnIndex, out expression))
            {
                cellReferenceRanges.AddRange(expression.CellReferenceRanges);
            }

            if (cellReferenceRanges.Count == 1)
            {
                CellReferenceRange cellReferenceRange = cellReferenceRanges.First();
                if (cellReferenceRange.Worksheet == context.Worksheet)
                {
                    CellRange cellRange = cellReferenceRange.ToCellRange();
                    if (cellRange.Contains(context.RowIndex, context.ColumnIndex))
                    {
                        return ErrorExpressions.CyclicReference;
                    }
                }
            }

            if (cellReferenceRanges.Count == 0 || cellReferenceRanges.Count > 1)
            {
                return ErrorExpressions.ReferenceError;
            }

            return expression;
        }

        #endregion
    }
}