﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the LOOKUP function.
    /// </summary>
    public class Lookup : LookupFunctionBase
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "LOOKUP";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Lookup()
        {
            string description = "The LOOKUP function returns a value either from a one-row or one-column range or from an array. The LOOKUP function has two syntax forms: the vector form and the array form.";
            string descriptionKey = "Spreadsheet_Functions_Lookup_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Lookup_value",
                    "A value that LOOKUP searches for in the first vector. Lookup_value can be a number, text, a logical value, or a name or reference that refers to a value.",
                    ArgumentType.Any,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_LookupValue",
                    descriptionLocalizationKey:"Spreadsheet_Functions_Lookup_LookupValue"),

                new ArgumentInfo("Lookup_vector",
                    "A range that contains only one row or one column. The values in lookup_vector can be text, numbers, or logical values.",
                    ArgumentType.Any,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_LookupVector",
                    descriptionLocalizationKey:"Spreadsheet_Functions_Lookup_LookupVector"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Result_vector",
                    "A range that contains only one row or column. The result_vector argument must be the same size as lookup_vector.",
                    ArgumentType.Any,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_ResultVector",
                    descriptionLocalizationKey:"Spreadsheet_Functions_Lookup_ResultVector"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.LookupReference, description, requiredArguments, optionalArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods

        internal override ErrorExpression TryGetResultVector(ArrayExpression lookupMatrix, FunctionEvaluationContext<object> context, out ArrayExpression resultVector)
        {
            resultVector = null;
            if (context.Arguments.Length == 3)
            {
                int vectorDimension = Math.Max(lookupMatrix.ColumnCount, lookupMatrix.RowCount);

                ErrorExpression error = TryConstructResultVector(context.Arguments[2], vectorDimension, out resultVector);
                if (error != null)
                {
                    return error;
                }
            }
            else
            {
                resultVector = GetRowOrColumnVectorFromArray(lookupMatrix, false);
            }

            return null;
        }

        internal override ErrorExpression TryGetLookupVector(ArrayExpression lookupMatrix, out ArrayExpression lookupVector)
        {
            lookupVector = GetRowOrColumnVectorFromArray(lookupMatrix, true);
            return null;
        }

        internal override bool GetSearchForApproximateMatch(FunctionEvaluationContext<object> context)
        {
            return true;
        }

        private static ArrayExpression GetRowOrColumnVectorFromArray(ArrayExpression array, bool getVectorFromArrayStart)
        {
            int rows = array.RowCount;
            int columns = array.ColumnCount;
            ArrayExpression vector;

            if (rows >= columns)
            {
                int column = getVectorFromArrayStart ? 0 : columns - 1;
                vector = array.GetColumnAsArrayExpression(column);
            }
            else
            {
                int row = getVectorFromArrayStart ? 0 : rows - 1;
                vector = array.GetRowAsArrayExpression(row);
            }

            return vector;
        }

        private static ErrorExpression TryConstructResultVector(object argument, int maxVectorDimension, out ArrayExpression resultVector)
        {
            resultVector = null;

            ArrayExpression arrayExpression = FunctionHelper.ConvertToArrayExpression(argument as ConstantExpression);

            CellReferenceRangeExpression referenceExpression = argument as CellReferenceRangeExpression;
            if (referenceExpression != null)
            {
                if (!referenceExpression.IsValid)
                {
                    return ErrorExpressions.ReferenceError;
                }

                if (!referenceExpression.HasSingleRange)
                {
                    return ErrorExpressions.NotAvailableError;
                }

                referenceExpression = GetRangeWithSpecificDimensions(maxVectorDimension, referenceExpression);

                arrayExpression = referenceExpression.FirstRangeArray;
            }

            if (arrayExpression != null)
            {
                if (arrayExpression.RowCount > 1 && arrayExpression.ColumnCount > 1)
                {
                    return ErrorExpressions.NotAvailableError;
                }

                resultVector = arrayExpression;

                return null;
            }

            return ErrorExpressions.ValueError;
        }
  
        private static CellReferenceRangeExpression GetRangeWithSpecificDimensions(int maxVectorDimension, CellReferenceRangeExpression referenceExpression)
        {
            int columns = referenceExpression.CellReferenceRange.ColumnsCount;
            int rows = referenceExpression.CellReferenceRange.RowsCount;

            if (rows == 1)
            {
                return CellReferenceRangeExpression.CreateRangeWithSpecificDimensions(referenceExpression, 1, maxVectorDimension);
            }
            else if (columns == 1)
            {
                return CellReferenceRangeExpression.CreateRangeWithSpecificDimensions(referenceExpression, maxVectorDimension, 1);
            }

            return referenceExpression;
        }

        #endregion
    }
}