﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Abstract class providing a base for the VLOOKUP and HLOOKUP functions
    /// </summary>
    public abstract class LookupFunctionBase : FunctionWithArguments
    {
        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<object> context)
        {
            ArrayExpression lookupMatrix;
            Worksheet criteriaWorksheet;
            CellIndex topLeftCellIndex;
            ErrorExpression error = TryGetLookUpMatrix(context.Arguments[1], out lookupMatrix, out criteriaWorksheet, out topLeftCellIndex);
            if (error != null)
            {
                return error;
            }

            bool approximateMatch = this.GetSearchForApproximateMatch(context);

            CriteriaEvaluator evaluator;
            error = TryGetCriteriaEvaluator(context, approximateMatch, out evaluator);
            if (error != null)
            {
                return error;
            }

            ArrayExpression lookupVector;
            error = this.TryGetLookupVector(lookupMatrix, out lookupVector);
            if (error != null)
            {
                return error;
            }

            ArrayExpression resultVector;
            error = this.TryGetResultVector(lookupMatrix, context, out resultVector);
            if (error != null)
            {
                return error;
            }

            return HandleLookupFunction(evaluator, lookupVector, resultVector, topLeftCellIndex, approximateMatch);
        }

        internal abstract ErrorExpression TryGetLookupVector(ArrayExpression lookupMatrix, out ArrayExpression lookupVector);

        internal abstract ErrorExpression TryGetResultVector(ArrayExpression lookupMatrix, FunctionEvaluationContext<object> context, out ArrayExpression resultVector);

        internal abstract bool GetSearchForApproximateMatch(FunctionEvaluationContext<object> context);

        private static ErrorExpression TryGetCriteriaEvaluator(FunctionEvaluationContext<object> context, bool approximateMatch, out CriteriaEvaluator evaluator)
        {
            object argument = context.Arguments[0];
            Worksheet criteriaWorksheet = context.Worksheet;
            ComparisonOperator comparisonOperator = approximateMatch ? ComparisonOperator.LessThanOrEqualsTo : ComparisonOperator.EqualsTo;

            return FunctionHelper.TryGetCriteriaEvaluator(argument, criteriaWorksheet, comparisonOperator, out evaluator);
        }

        private static ErrorExpression TryGetLookUpMatrix(object argument, out ArrayExpression lookUpVector, out Worksheet worksheet, out CellIndex topLeftCellIndex)
        {
            return FunctionHelper.TryGetArrayFromFunctionArgument(argument, out lookUpVector, out worksheet, out topLeftCellIndex);
        }

        private static RadExpression HandleLookupFunction
            (CriteriaEvaluator criteriaEvaluator, ArrayExpression lookUpVector, ArrayExpression resultVector, CellIndex topLeftCellIndex, bool approximateMatch)
        {
            Guard.ThrowExceptionIfNull(criteriaEvaluator, "criteriaEvaluator");
            Guard.ThrowExceptionIfNull(lookUpVector, "lookUpVector");
            Guard.ThrowExceptionIfNull(resultVector, "resultVector");
            Guard.ThrowExceptionIfNull(topLeftCellIndex, "topLeftCellIndex");

            int lookupRowCount = lookUpVector.RowCount;
            int lookupColumnCount = lookUpVector.ColumnCount;
            bool searchFromStart = !approximateMatch;

            int rowStart = searchFromStart ? 0 : lookupRowCount - 1;
            int columnStart = searchFromStart ? 0 : lookupColumnCount - 1;
            int delta = searchFromStart ? 1 : -1;

            for (int row = rowStart; row >= 0 && row < lookupRowCount; row += delta)
            {
                for (int column = columnStart; column >= 0 && column < lookupColumnCount; column += delta)
                {
                    RadExpression value = lookUpVector[row, column];

                    if (criteriaEvaluator.Evaluate(value, topLeftCellIndex.RowIndex + row, topLeftCellIndex.ColumnIndex + column))
                    {
                        return GetCorrespondingResultValue(resultVector, row, column);
                    }
                }
            }

            return ErrorExpressions.NotAvailableError;
        }

        private static RadExpression GetCorrespondingResultValue(ArrayExpression resultVector, int rowIndex, int columnIndex)
        {
            Guard.ThrowExceptionIfLessThan(0, rowIndex, "rowIndex");
            Guard.ThrowExceptionIfLessThan(0, columnIndex, "columnIndex");

            int rowCount = resultVector.RowCount;
            int columnCount = resultVector.ColumnCount;
            int maxIndex = Math.Max(rowIndex, columnIndex);
            RadExpression result = null;

            if (rowCount == 1)
            {
                if (maxIndex < columnCount)
                {
                    result = resultVector[0, maxIndex];
                }
            }
            else if (columnCount == 1)
            {
                if (maxIndex < rowCount)
                {
                    result = resultVector[maxIndex, 0];
                }
            }

            return result ?? ErrorExpressions.NotAvailableError;
        }
    }
}
