﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions.LookupAndReference
{
    /// <summary>
    /// Represents the MATCH function.
    /// </summary>
    public class Match : FunctionWithArguments
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "MATCH";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }


        #endregion


        #region Constructors

        static Match()
        {
            string description = "Returns the relative position of an item in an array that matches a specified value in a specified order.";
            string descriptionKey = "Spreadsheet_Functions_Match_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Lookup_value",
                    "is the value you use to find the value you want in the array, a number, text, or logical value, or a reference to one of these.",
                    ArgumentType.Any,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_LookupValue",
                    descriptionLocalizationKey:"Spreadsheet_Functions_Match_LookupValueInfo"),

                new ArgumentInfo("Lookup_array",
                    "is a contiguous range of cells containing possible lookup values, an array of values, or a reference to an array.",
                    ArgumentType.Any,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_LookupArray",
                    descriptionLocalizationKey:"Spreadsheet_Functions_Match_LookupArrayInfo"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Match_type",
                    "is a number 1, 0, or -1 indicating which value to return.",
                    ArgumentType.Number,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_MatchType",
                    descriptionLocalizationKey:"Spreadsheet_Functions_Match_MatchTypeInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.LookupReference, description, requiredArguments, optionalArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion

        /// <summary>
        /// Virtual method that evaluates the function with the specified context.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Function result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<object> context)
        {
            ErrorExpression error = null;

            ArrayExpression lookupArray;
            CellIndex topLeftCellIndex;
            Worksheet criteriaWorksheet;
            error = TryGetLookupArray(context, out lookupArray, out criteriaWorksheet, out topLeftCellIndex);
            if(error != null)
            {
                return error;
            }

            int matchType = GetMatchType(context);

            CriteriaEvaluator lessOrGreaterEvaluator;
            error = TryGetCriteriaEvaluator(context, matchType, out lessOrGreaterEvaluator);
            if (error != null)
            {
                return error;
            }

            CriteriaEvaluator equalsEvaluator;
            error = TryGetCriteriaEvaluator(context, 0, out equalsEvaluator);
            if (error != null)
            {
                return error;
            }

            // There are three separate algorithms for the different match types in order for the results to be as close to the Excel results as possible.
            // Moreover, in invalid scenarios (e.g. the data is not ordered, or is ordered in order that does not correspond to the match type) these algorithms
            // return results differing from the results that other applications (e.g. Excel, OpenOffice, etc.) would return. This means that different
            // applications use different algorithms or structure their data differently.
            if (matchType > 0)
            {
                return HandleMatchFunctionLookingForLessOrEqual(lookupArray, lessOrGreaterEvaluator, topLeftCellIndex);
            }
            else if(matchType < 0)
            {
                return HandleMatchFunctionLookingForGreaterOrEqual(lookupArray, lessOrGreaterEvaluator, equalsEvaluator, topLeftCellIndex);
            }
            else
            {
                return HandleMatchFunctionLookingForEqual(lookupArray, equalsEvaluator, topLeftCellIndex);
            }
        }

        private static ErrorExpression TryGetLookupArray(FunctionEvaluationContext<object> context, out ArrayExpression lookUpArray, out Worksheet worksheet, out CellIndex topLeftCellIndex)
        {
            object lookupArrayArgument = context.Arguments[1];
            lookUpArray = null;
            worksheet = null;
            topLeftCellIndex = null;

            if (lookupArrayArgument is NumberExpression)
            {
                return ErrorExpressions.NotAvailableError;
            }

            return FunctionHelper.TryGetArrayFromFunctionArgument(lookupArrayArgument, out lookUpArray, out worksheet, out topLeftCellIndex);
        }

        private static ErrorExpression TryGetCriteriaEvaluator(FunctionEvaluationContext<object> context, int matchType, out CriteriaEvaluator evaluator)
        {
            evaluator = null;
            object argument = context.Arguments[0];
            Worksheet worksheet = context.Worksheet;
            ComparisonOperator comparisonOperator = GetComparisonOperatorFromMatchType(matchType);

            return FunctionHelper.TryGetCriteriaEvaluator(argument, worksheet, comparisonOperator, out evaluator);
        }

        private static int GetMatchType(FunctionEvaluationContext<object> context)
        {
            int matchType;
            if (context.Arguments.Length > 2)
            {
                double number = (double)context.Arguments[2];
                matchType = (int)number;
            }
            else
            {
                matchType = 1;
            }

            return matchType;
        }

        private static ComparisonOperator GetComparisonOperatorFromMatchType(int matchType)
        {
            if (matchType > 0)
            {
                return ComparisonOperator.LessThanOrEqualsTo;
            }
            else if (matchType < 0)
            {
                return ComparisonOperator.GreaterThanOrEqualsTo;
            }
            else //(matchType == 0)
            {
                return ComparisonOperator.EqualsTo;
            }
        }


        private static RadExpression HandleMatchFunctionLookingForLessOrEqual(ArrayExpression lookupArray, CriteriaEvaluator lessOrEqualEvaluator, CellIndex topLeftCellIndex)
        {
            bool isColumn;
            ErrorExpression error = IsLookupArrayColumn(lookupArray, out isColumn);
            if (error != null)
            {
                return error;
            }

            int lookupRowCount = lookupArray.RowCount;
            int lookupColumnCount = lookupArray.ColumnCount;

            for (int row = lookupRowCount - 1; row >= 0; row--)
            {
                for (int column = lookupColumnCount - 1; column >= 0; column--)
                {
                    RadExpression value = lookupArray[row, column];

                    if (lessOrEqualEvaluator.Evaluate(value, topLeftCellIndex.RowIndex + row, topLeftCellIndex.ColumnIndex + column))
                    {
                        return isColumn ?
                             NumberExpression.CreateValidNumberOrErrorExpression(row + 1) :
                             NumberExpression.CreateValidNumberOrErrorExpression(column + 1);
                    }
                }
            }

            return ErrorExpressions.NotAvailableError;
        }

        private static RadExpression HandleMatchFunctionLookingForGreaterOrEqual(ArrayExpression lookupArray, CriteriaEvaluator greaterOrEqualEvaluator, CriteriaEvaluator equalsEvaluator, CellIndex topLeftCellIndex)
        {
            bool isColumn;
            ErrorExpression error = IsLookupArrayColumn(lookupArray, out isColumn);
            if (error != null)
            {
                return error;
            }

            int lookupRowCount = lookupArray.RowCount;
            int lookupColumnCount = lookupArray.ColumnCount;

            for (int row = 0; row < lookupRowCount; row++)
            {
                for (int column = 0; column < lookupColumnCount; column++)
                {
                    RadExpression value = lookupArray[row, column];

                    if (equalsEvaluator.Evaluate(value, topLeftCellIndex.RowIndex + row, topLeftCellIndex.ColumnIndex + column))
                    {
                        return isColumn ?
                            NumberExpression.CreateValidNumberOrErrorExpression(row + 1) :
                            NumberExpression.CreateValidNumberOrErrorExpression(column + 1);
                    }

                    if (!greaterOrEqualEvaluator.Evaluate(value, topLeftCellIndex.RowIndex + row, topLeftCellIndex.ColumnIndex + column))
                    {
                        return isColumn ?
                            CreateValidRowOrColumnIndexOrErrorExpression(row) :
                            CreateValidRowOrColumnIndexOrErrorExpression(column);
                    }
                }
            }

            return isColumn ?
                    NumberExpression.CreateValidNumberOrErrorExpression(lookupRowCount) :
                    NumberExpression.CreateValidNumberOrErrorExpression(lookupColumnCount);
        }

        private static RadExpression HandleMatchFunctionLookingForEqual(ArrayExpression lookupArray, CriteriaEvaluator equalsEvaluator, CellIndex topLeftCellIndex)
        {
            bool isColumn;
            ErrorExpression error = IsLookupArrayColumn(lookupArray, out isColumn);
            if (error != null)
            {
                return error;
            }

            int lookupRowCount = lookupArray.RowCount;
            int lookupColumnCount = lookupArray.ColumnCount;

            for (int row = 0; row < lookupRowCount; row++)
            {
                for (int column = 0; column < lookupColumnCount; column++)
                {
                    RadExpression value = lookupArray[row, column];

                    if (equalsEvaluator.Evaluate(value, topLeftCellIndex.RowIndex + row, topLeftCellIndex.ColumnIndex + column))
                    {
                        return isColumn ?
                            NumberExpression.CreateValidNumberOrErrorExpression(row + 1) :
                            NumberExpression.CreateValidNumberOrErrorExpression(column + 1);
                    }
                }
            }

            return ErrorExpressions.NotAvailableError;
        }

        private static RadExpression CreateValidRowOrColumnIndexOrErrorExpression(int number)
        {
            if(number > 0)
            {
                return NumberExpression.CreateValidNumberOrErrorExpression(number);
            }
            else
            {
                return ErrorExpressions.NotAvailableError;
            }
        }

        private static ErrorExpression IsLookupArrayColumn(ArrayExpression lookupArray, out bool isColumn)
        {
            if (lookupArray.ColumnCount == 1)
            {
                isColumn = true;
                return null;
            }
            else if (lookupArray.RowCount == 1)
            {
                isColumn = false;
                return null;
            }
            else
            {
                isColumn = false;
                return ErrorExpressions.NotAvailableError;
            }
        }
    }
}
