﻿using System;
using System.Collections.Generic;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the VLOOKUP function.
    /// </summary>
    public class VLookup : LookupFunctionBase
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "VLOOKUP";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static VLookup()
        {
            string description = "The VLOOKUP function looks for a value in the leftmost column of a table or array of values and then returns a value in the same row from a column you specify. By default, the table must be sorted in an ascending order.";
            string descriptionKey = "Spreadsheet_Functions_VLookup_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Lookup_value",
                    "The value to be found in the first column of the table. It can be a value, a reference or a text string.",
                    ArgumentType.Any,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_LookupValue",
                    descriptionLocalizationKey:"Spreadsheet_Functions_VLookup_LookupValue"),

                new ArgumentInfo("Table_array",
                    "A table of text, numbers or logical values, in which the data is looked up. Table_array can be a reference to a range or a range name.",
                    ArgumentType.Any,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_TableArray",
                    descriptionLocalizationKey:"Spreadsheet_Functions_VHLookup_TableArray"),

                new ArgumentInfo("Col_index_num",
                    "The column number in table_array from which the matching value should be returned. The first column of values in the table is column 1.",
                    ArgumentType.Number,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_ColIndexNum",
                    descriptionLocalizationKey:"Spreadsheet_Functions_VLookup_ColIndexNum"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Range_lookup",
                    "A logical value: to find the closest match in the first column (sorted in ascending order) = TRUE or omitted; find an exact match = FALSE.",
                    ArgumentType.Logical,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_RangeLookup",
                    descriptionLocalizationKey:"Spreadsheet_Functions_VLookup_RangeLookup"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.LookupReference, description, requiredArguments, optionalArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion

        internal override ErrorExpression TryGetLookupVector(ArrayExpression lookupMatrix, out ArrayExpression lookupVector)
        {
            lookupVector = lookupMatrix.GetColumnAsArrayExpression(0);
            return null;
        }

        internal override ErrorExpression TryGetResultVector(ArrayExpression lookupMatrix, FunctionEvaluationContext<object> context, out ArrayExpression resultVector)
        {
            resultVector = null;
            int indexNum = (int)((double)context.Arguments[2]) - 1;

            if (indexNum < 0)
            {
                return ErrorExpressions.ValueError;
            }

            if (indexNum >= lookupMatrix.ColumnCount)
            {
                return ErrorExpressions.ReferenceError;
            }

            resultVector = lookupMatrix.GetColumnAsArrayExpression(indexNum);

            return null;
        }

        internal override bool GetSearchForApproximateMatch(FunctionEvaluationContext<object> context)
        {
            if (context.Arguments.Length == 4)
            {
                bool result = (bool)context.Arguments[3];
                return result;
            }
            else
            {
                return true;
            }
        }
    }
}
