﻿using System;
using System.Collections.Generic;
using System.ComponentModel;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the ATAN2 function.
    /// </summary>
    public class Atan2 : NumbersInFunction
    {
        
		#region Constants

		///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "ATAN2";
        private static readonly FunctionInfo Info;

		#endregion


		#region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
			get
			{
				return FunctionName;
			}
		}

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Atan2()
        {
            string description = "Returns the arctangent, or inverse tangent, of the specified x- and y-coordinates. The arctangent is the angle from the x-axis to a line containing the origin (0, 0) and a point with coordinates (x_num, y_num). The angle is given in radians between -pi and pi, excluding -pi.";
            string descriptionKey = "Spreadsheet_Functions_Atan2_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                 new ArgumentInfo("X_num",
                                  "is the x-coordinate of the point.",
                                  ArgumentType.Number,
                                  nameLocalizationKey:"Spreadsheet_Functions_Args_XNum",
                                  descriptionLocalizationKey:"Spreadsheet_Functions_Atan2_XNumInfo"),
                 new ArgumentInfo("Y_num",
                                  "is the y-coordinate of the point.",
                                  ArgumentType.Number,
                                  nameLocalizationKey:"Spreadsheet_Functions_Args_YNum",
                                  descriptionLocalizationKey:"Spreadsheet_Functions_Atan2_YNumInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.MathTrig, description, requiredArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


		#region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double x = context.Arguments[0];
            double y = context.Arguments[1];

            if (x == 0 && y == 0)
            {
                return ErrorExpressions.DivisionByZero;
            }

            double result = Math.Atan2(y, x);
            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

		#endregion
    }
}