﻿using System;
using System.Collections.Generic;
using System.ComponentModel;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the COMBIN function.
    /// </summary>
    public class Combin : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "COMBIN";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Combin()
        {
            string description = "Returns the number of combinations for a given number of items. Use COMBIN to determine the total possible number of groups for a given number of items.";
            string descriptionKey = "Spreadsheet_Functions_Combin_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Number",
                                 "is the number of items",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Number",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Combin_NumberInfo"),
                new ArgumentInfo("NumberChosen",
                                 "is the number of items in each combination.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_NumberChosen",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Combin_NumberChosenInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.MathTrig, description, requiredArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double candidates = MathUtility.RoundDown(context.Arguments[0], 0);
            double groups = MathUtility.RoundDown(context.Arguments[1], 0);

            if (candidates < 0 || groups < 0 || candidates < groups)
            {
                return ErrorExpressions.NumberError;
            }

            double numerator = MathUtility.Factorial(candidates);
            double denominator = MathUtility.Factorial(groups) * MathUtility.Factorial(candidates - groups);

            return NumberExpression.CreateValidNumberOrErrorExpression(numerator / denominator);
        }

        #endregion
    }
}