﻿using System;
using System.Collections.Generic;
using System.ComponentModel;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the FLOOR function.
    /// </summary>
    public class Floor : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "FLOOR";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties
        
        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

		#endregion


        #region Constructors

        static Floor()
        {
            string description = "Rounds number down, toward zero, to the nearest multiple of significance.";
            string descriptionKey = "Spreadsheet_Functions_Floor_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Number",
                                 "is the value you want to round.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Number",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Floor_NumberInfo"),
                new ArgumentInfo("Significance",
                                 "is the multiple to which you want to round.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Significance",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Floor_SignificanceInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.MathTrig, description, requiredArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


		#region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double number = context.Arguments[0];
            double significance = context.Arguments[1];

            if (number > 0 && significance < 0)
            {
                return ErrorExpressions.NumberError;
            }

            if (significance == 0 && number != 0)
            {
                return ErrorExpressions.DivisionByZero;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(MathUtility.Floor(number, significance));
        }

		#endregion
    }
}