﻿using System;
using System.Collections.Generic;
using System.ComponentModel;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the FLOOR.MATH function.
    /// </summary>
    public class FloorMath : NumbersInFunction
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "FLOOR.MATH";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties
        
        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

		#endregion


        #region Constructors

        static FloorMath()
        {
            string description = "Round a number down to the nearest integer or to the nearest multiple of significance.";
            string descriptionKey = "Spreadsheet_Functions_FloorMath_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Number",
                                 "is the value you want to round.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Number",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Floor_NumberInfo")
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Significance",
                                 "is the multiple to which you want to round.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Significance",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Floor_SignificanceInfo"),
                new ArgumentInfo("Mode",
                                 "is the direction (toward or away from 0) to round negative numbers.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Mode",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_FloorMath_ModeInfo")
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.MathTrig, description, requiredArguments, optionalArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


		#region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double number = context.Arguments[0];

            double significance = 1;
            if (context.Arguments.Length >= 2)
            {
                significance = context.Arguments[1];
            }

            bool roundDown = true;
            if (context.Arguments.Length > 2)
            {
                roundDown = context.Arguments[2] == 0;
            }

            if (number >= MathUtility.CeilingFloorUpperLimit || number <= MathUtility.CeilingFloorLowerLimit)
            {
                return ErrorExpressions.NumberError;
            }

            double result;
            if (number < 0 && !roundDown)
            {
                result = MathUtility.CeilingPrecise(number, significance);
            }
            else
            {
                result = MathUtility.FloorPrecise(number, significance);
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(result);
        }

		#endregion
    }
}