﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the SUMIF function.
    /// </summary>
    public class SumIf : FunctionWithArguments
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "SUMIF";
        private static readonly FunctionInfo Info;
        private static readonly ArgumentConversionRules convertionRules = new ArgumentConversionRules(
            emptyDirectArgument: ArgumentInterpretation.Ignore,
            numberDirectArgument: ArgumentInterpretation.UseAsIs,
            boolDirectArgument: ArgumentInterpretation.UseAsIs,
            textNumberDirectArgument: ArgumentInterpretation.UseAsIs,
            nonTextNumberDirectArgument: ArgumentInterpretation.UseAsIs,
            emptyIndirectArgument: ArgumentInterpretation.Ignore,
            numberIndirectArgument: ArgumentInterpretation.UseAsIs,
            boolIndirectArgument: ArgumentInterpretation.UseAsIs,
            textNumberIndirectArgument: ArgumentInterpretation.UseAsIs,
            nonTextNumberIndirectArgument: ArgumentInterpretation.UseAsIs,
            arrayArgument: ArrayArgumentInterpretation.UseFirstElement);

        #endregion


        #region Properties

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return SumIf.convertionRules;
            }
        }

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static SumIf()
        {
            string description = "You use the SUMIF function to sum the values in a range that meet criteria that you specify. For example, suppose that in a column that contains numbers, you want to sum only the values that are larger than 5. You can use the following formula: =SUMIF(B2:B25,\">5\"). In this example, the criteria is applied the same values that are being summed. If you want, you can apply the criteria to one range and sum the corresponding values in a different range. For example, the formula =SUMIF(B2:B5, \"John\", C2:C5) sums only the values in the range C2:C5, where the corresponding cells in the range B2:B5 equal \"John.\"";
            string descriptionKey = "Spreadsheet_Functions_SumIf_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Range",
                    "The range of cells that you want evaluated by criteria. Cells in each range must be numbers or names, arrays, or references that contain numbers. Blank and text values are ignored.",
                    ArgumentType.Reference,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_Range",
                    descriptionLocalizationKey:"Spreadsheet_Functions_SumIf_Range"),

                new ArgumentInfo("Criteria",
                    "The criteria in the form of a number, expression, a cell reference, text, or a function that defines which cells will be added. For example, criteria can be expressed as 32, \">32\", B5, 32, \"32\", \"apples\", or TODAY().",
                    ArgumentType.Any,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_Criteria",
                    descriptionLocalizationKey:"Spreadsheet_Functions_SumIf_Criteria"),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Sum_range",
                    "The actual cells to add, if you want to add cells other than those specified in the range argument. If the sum_range argument is omitted, RadSpreadsheet adds the cells that are specified in the range argument (the same cells to which the criteria is applied).",
                    ArgumentType.Reference,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_SumRange",
                    descriptionLocalizationKey:"Spreadsheet_Functions_SumIf_SumRange"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.MathTrig, description, requiredArguments, optionalArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<object> context)
        {
            CellReferenceRangeExpression criteriaRangeExpression, sumRangeExpression;
            if (!TryGetValidReferenceRangeExpressions(context.Arguments, out criteriaRangeExpression, out sumRangeExpression))
            {
                return ErrorExpressions.ReferenceError;
            }

            bool sumRangeExpressionHasSingleRange = sumRangeExpression.HasSingleRange;
            sumRangeExpression = CellReferenceRangeExpression.CreateRangeWithSpecificDimensions(sumRangeExpression,
             criteriaRangeExpression.CellReferenceRange.RowsCount,
             criteriaRangeExpression.CellReferenceRange.ColumnsCount);

            ArrayExpression criteriaValues, sumValues;
            if (!sumRangeExpressionHasSingleRange ||
                !TryGetArrayValues(criteriaRangeExpression, out criteriaValues) ||
                !TryGetArrayValues(sumRangeExpression, out sumValues))
            {
                return ErrorExpressions.ValueError;
            }

            RadExpression criteria = context.Arguments[1] as RadExpression;
            CriteriaEvaluator evaluator = new CriteriaEvaluator(criteria, criteriaRangeExpression.Worksheet);
            CellIndex topLeftCriteriaCell = new CellIndex(
                criteriaRangeExpression.CellReferenceRange.FromCellReference.ActualRowIndex,
                criteriaRangeExpression.CellReferenceRange.FromCellReference.ActualColumnIndex);

            double sum = 0;

            for (int i = 0; i < sumValues.RowCount; i++)
            {
                for (int j = 0; j < sumValues.ColumnCount; j++)
                {
                    if (evaluator.Evaluate(criteriaValues[i, j], i + topLeftCriteriaCell.RowIndex, j + topLeftCriteriaCell.ColumnIndex))
                    {
                        NumberExpression addent = sumValues[i, j] as NumberExpression;

                        if (addent != null)
                        {
                            sum += addent.Value;
                        }
                    }
                }
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(sum);
        }

        private static bool TryGetValidReferenceRangeExpressions(object[] arguments,
            out CellReferenceRangeExpression criteriaRangeExpression, out CellReferenceRangeExpression sumRangeExpression)
        {
            criteriaRangeExpression = (CellReferenceRangeExpression)arguments[0];

            if (arguments.Length == 3)
            {
                sumRangeExpression = (CellReferenceRangeExpression)arguments[2];
            }
            else
            {
                sumRangeExpression = criteriaRangeExpression;
            }

            return criteriaRangeExpression.IsValid && sumRangeExpression.IsValid;
        }

        private static bool TryGetArrayValues(CellReferenceRangeExpression referenceRangeExpression, out ArrayExpression arrayValues)
        {
            arrayValues = null;

            if (!referenceRangeExpression.HasSingleRange)
            {
                return false;
            }

            arrayValues = referenceRangeExpression.FirstRangeArray;

            return true;
        }

        #endregion
    }
}