﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the COUNTIF function.
    /// </summary>
    public class CountIf : FunctionWithArguments
    {
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "COUNTIF";
        private static readonly FunctionInfo Info;
        private static readonly ArgumentConversionRules convertionRules = new ArgumentConversionRules(
            emptyDirectArgument: ArgumentInterpretation.Ignore,
            numberDirectArgument: ArgumentInterpretation.UseAsIs,
            boolDirectArgument: ArgumentInterpretation.UseAsIs,
            textNumberDirectArgument: ArgumentInterpretation.UseAsIs,
            nonTextNumberDirectArgument: ArgumentInterpretation.UseAsIs,
            emptyIndirectArgument: ArgumentInterpretation.Ignore,
            numberIndirectArgument: ArgumentInterpretation.UseAsIs,
            boolIndirectArgument: ArgumentInterpretation.UseAsIs,
            textNumberIndirectArgument: ArgumentInterpretation.UseAsIs,
            nonTextNumberIndirectArgument: ArgumentInterpretation.UseAsIs,
            arrayArgument: ArrayArgumentInterpretation.UseFirstElement);

        #endregion


        #region Properties

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <returns>The argument conversion rules as ArgumentConversionRules.</returns>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return CountIf.convertionRules;
            }
        }

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static CountIf()
        {
            string description = "The COUNTIF function counts the number of cells within a range that meet a single criteria that you specify. For example, you can count all the cells that start with a certain letter, or you can count all the cells that contain a number that is larger or smaller than a number you specify. For example, suppose you have a worksheet that contains a list of tasks in column A, and the first name of the person assigned to each task in column B. You can use the COUNTIF function to count how many times a person's name appears in column B and, in that way, determine how many tasks are assigned to that person. For example: =COUNTIF(B2:B25,\"Nancy\")";
            string descriptionKey = "Spreadsheet_Functions_CountIf_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Range",
                    "One or more cells to count, passed as reference argument.",
                    ArgumentType.Reference,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_Range",
                    descriptionLocalizationKey:"Spreadsheet_Functions_CountIf_Range"),
                new ArgumentInfo("Criteria",
                    "A number, expression, cell reference, or text string that defines which cells will be counted. For example, criteria can be expressed as 32, \">32\", B4, \"apples\", or \"32\"",
                    ArgumentType.Any,
                    nameLocalizationKey:"Spreadsheet_Functions_Args_Criteria",
                    descriptionLocalizationKey:"Spreadsheet_Functions_CountIf_Criteria"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Statistical, description, requiredArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<object> context)
        {
            CellReferenceRangeExpression criteriaRangeExpression = (CellReferenceRangeExpression)context.Arguments[0];

            if (!criteriaRangeExpression.IsValid)
            {
                return ErrorExpressions.ReferenceError;
            }

            if (!criteriaRangeExpression.HasSingleRange)
            {
                return ErrorExpressions.ValueError;
            }

            ArrayExpression criteriaValues = criteriaRangeExpression.FirstRangeArray;

            RadExpression criteria = context.Arguments[1] as RadExpression;
            CriteriaEvaluator evaluator = new CriteriaEvaluator(criteria, criteriaRangeExpression.Worksheet);
            CellIndex topLeftCriteriaCell = new CellIndex(
                criteriaRangeExpression.CellReferenceRange.FromCellReference.ActualRowIndex,
                criteriaRangeExpression.CellReferenceRange.FromCellReference.ActualColumnIndex);

            double count = 0;

            for (int i = 0; i < criteriaValues.RowCount; i++)
            {
                for (int j = 0; j < criteriaValues.ColumnCount; j++)
                {
                    if (evaluator.Evaluate(criteriaValues[i, j], i + topLeftCriteriaCell.RowIndex, j + topLeftCriteriaCell.ColumnIndex))
                    {
                        count++;
                    }
                }
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(count);
        }

        #endregion
    }
}