﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the MODE function.
    /// </summary>
    public class Mode : NumbersInFunction
	{
		#region Constants

		///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "MODE";
        private static readonly FunctionInfo Info;
        private static readonly ArgumentConversionRules ConvertionRules = new ArgumentConversionRules(emptyIndirectArgument: ArgumentInterpretation.Ignore,
                                                                                                      boolDirectArgument: ArgumentInterpretation.TreatAsError,
                                                                                                      boolIndirectArgument: ArgumentInterpretation.TreatAsError,
                                                                                                      textNumberDirectArgument: ArgumentInterpretation.TreatAsError,
                                                                                                      textNumberIndirectArgument: ArgumentInterpretation.TreatAsError,
                                                                                                      nonTextNumberDirectArgument: ArgumentInterpretation.TreatAsError,
                                                                                                      nonTextNumberIndirectArgument: ArgumentInterpretation.TreatAsError);

		#endregion


		#region Properties

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <value>The argument conversion rules.</value>
        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return ConvertionRules;
            }
        }

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
			get
			{
				return FunctionName;
			}
		}

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

		#endregion


        #region Constructors

        static Mode()
        {
            string description = "Returns the most frequently occurring, or repetitive, value in an array or range of data.";
            string descriptionKey = "Spreadsheet_Functions_Mode_Info";
            
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Number",
                                 "number1, number2,... are 1 to 30 arguments for which you want to calculate the mode. You can also use a single array or a reference to an array instead of arguments separated by commas.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Number",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Mode_NumberInfo"),
            };
            
            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Number",
                                 "number1, number2,... are 1 to 30 arguments for which you want to calculate the mode. You can also use a single array or a reference to an array instead of arguments separated by commas.",
                                 ArgumentType.Number,
                                 nameLocalizationKey:"Spreadsheet_Functions_Args_Number",
                                 descriptionLocalizationKey:"Spreadsheet_Functions_Mode_NumberInfo"),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Statistical, description, requiredArguments, optionalArguments, 254, false, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


		#region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            Dictionary<double, int> values = new Dictionary<double, int>();

            for (int i = 0; i < context.Arguments.Length; i++)
            {
                if (values.ContainsKey(context.Arguments[i]))
                {
                    values[context.Arguments[i]]++;
                }
                else
                {
                    values.Add(context.Arguments[i], 1);
                }
            }

            double? result = null;
            int resultOccurrences = 0;

            foreach (var value in values)
            {
                if (result == null || value.Value > resultOccurrences)
                {
                    result = value.Key;
                    resultOccurrences = value.Value;
                }
            }

            if (resultOccurrences == 1)
            {
                return ErrorExpressions.NotAvailableError;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression((double)result);
        }

		#endregion
	}
}