﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the REPT function.
    /// </summary>
    public class Rept : FunctionWithArguments
    {  
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "REPT";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Rept()
        {
            string description = "Repeats text a given number of times. Use REPT to fill a cell with a number of instances of a text string.";
            string descriptionKey = "Spreadsheet_Functions_Rept_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                 new ArgumentInfo("Text",
                                  "is the text you want to repeat.",
                                  ArgumentType.Text,
                                  nameLocalizationKey:"Spreadsheet_Functions_Args_Text",
                                  descriptionLocalizationKey:"Spreadsheet_Functions_Rept_TextInfo"),
                 new ArgumentInfo("NumberTimes",
                                  "is a positive number specifying the number of times to repeat text.",
                                  ArgumentType.Number,
                                  nameLocalizationKey:"Spreadsheet_Functions_Args_NumberTimes",
                                  descriptionLocalizationKey:"Spreadsheet_Functions_Rept_NumberTimesInfo")
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Text, description, requiredArguments, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context"></param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<object> context)
        {
            string text = context.Arguments[0].ToString();
            double numberTimes = (double)context.Arguments[1];

            StringBuilder builder = new StringBuilder();

            int resultLength = text.Length * (int)numberTimes;

            if (numberTimes < 0 || resultLength > 32767)
            {
                return ErrorExpressions.ValueError;
            }

            for (int i = 0; i < numberTimes; i++)
            {
                builder.Append(text);
            }

            return new StringExpression(builder.ToString());
        }

        #endregion
    }
}