﻿using System;
using System.Collections.Generic;
using System.ComponentModel;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the RIGHT function.
    /// </summary>
    public class Right : FunctionWithArguments
    {  
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "RIGHT";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Right()
        {
            string description = "Returns the last character or characters in a text string, based on the number of characters you specify.";
            string descriptionKey = "Spreadsheet_Functions_Right_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                 new ArgumentInfo("Text",
                                  "is the text string that contains the characters you want to extract.",
                                  ArgumentType.Text,
                                  nameLocalizationKey:"Spreadsheet_Functions_Args_Text",
                                  descriptionLocalizationKey:"Spreadsheet_Functions_LeftRight_TextInfo")
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
            new ArgumentInfo("Number",
                             "specifies the number of characters you want RIGHT to extract.",
                             ArgumentType.Number,
                             nameLocalizationKey:"Spreadsheet_Functions_Args_Number",
                             descriptionLocalizationKey:"Spreadsheet_Functions_Right_NumberInfo")
                 
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Text, description, requiredArguments, optionalArguments, 1, false, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context"></param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<object> context)
        {
            string text = (string)context.Arguments[0];

            string result = string.Empty;

            if (context.Arguments.Length == 1)
            {
                if (text.Length > 0)
                {
                    result = text[text.Length - 1].ToString();
                }
            }
            else
            {
                double number = (double)context.Arguments[1];

                if (number < 0)
                {
                    return ErrorExpressions.ValueError;
                }

                int numberChars = (int)number;
                if (text.Length < numberChars)
                {
                    numberChars = text.Length;
                }

                result = text.Substring(text.Length - numberChars, numberChars);
            }

            return new StringExpression(result);
        }

        #endregion
    }
}