﻿using System;
using System.Collections.Generic;
using System.ComponentModel;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Functions
{
    /// <summary>
    /// Represents the SEARCH function.
    /// </summary>
    public class Search : FunctionWithArguments
    {  
        #region Constants

        ///<summary>
        /// The name of the function.
        ///</summary>
        public static readonly string FunctionName = "SEARCH";
        private static readonly FunctionInfo Info;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the function.
        /// </summary>
        /// <returns>The name as String.</returns>
        /// <value>The name.</value>
        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        /// <summary>
        /// Gets the function info.
        /// </summary>
        /// <returns>The function info as FunctionInfo.</returns>
        /// <value>The function info.</value>
        public override FunctionInfo FunctionInfo
        {
            get 
            {
                return Info;
            }
        }

        #endregion


        #region Constructors

        static Search()
        {
            string description = "Locates one text string within a second text string, and return the number of the starting position of the first text string from the first character of the second text string. SEARCH is not case sensitive. If you want to do a case sensitive search, you can use FIND.";
            string descriptionKey = "Spreadsheet_Functions_Search_Info";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                 new ArgumentInfo("FindText",
                                  "is the text you want to find.",
                                  ArgumentType.Text,
                                  nameLocalizationKey:"Spreadsheet_Functions_Args_FindText",
                                  descriptionLocalizationKey:"Spreadsheet_Functions_FindSearch_FindTextInfo"),
                 new ArgumentInfo("WithinText",
                                  "is the text containing the text you want to find.",
                                  ArgumentType.Text,
                                  nameLocalizationKey:"Spreadsheet_Functions_Args_WithinText",
                                  descriptionLocalizationKey:"Spreadsheet_Functions_FindSearch_WithinTextInfo")
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                 new ArgumentInfo("StartNumber",
                                  "specifies the character at which to start the search. The first character in within_text is character number 1. If you omit start_num, it is assumed to be 1.",
                                  ArgumentType.Number,
                                  nameLocalizationKey:"Spreadsheet_Functions_Args_StartNumber",
                                  descriptionLocalizationKey:"Spreadsheet_Functions_FindSearch_StartNumberInfo")
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.Text, description, requiredArguments, optionalArguments, 1, false, descriptionLocalizationKey: descriptionKey);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Virtual method evaluating the function with System.Object arguments array.
        /// </summary>
        /// <param name="context"></param>
        /// <returns>Functions result as RadExpression.</returns>
        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<object> context)
        {
            string findText = context.Arguments[0].ToString();
            string withinText = context.Arguments[1].ToString();

            findText = findText.ToLowerInvariant();
            withinText = withinText.ToLowerInvariant();

            int result = -1;

            if (context.Arguments.Length == 2)
            {
                result = withinText.IndexOf(findText, StringComparison.Ordinal);
            }
            else
            {
                double startNumber = (double)context.Arguments[2];

                if (startNumber < 1 || startNumber > withinText.Length)
                {
                    return ErrorExpressions.ValueError;
                }

                result = withinText.IndexOf(findText, (int)startNumber - 1, StringComparison.Ordinal);
            }

            if (result == -1)
            {
                return ErrorExpressions.ValueError;
            }

            return NumberExpression.CreateValidNumberOrErrorExpression(result + 1);
        }

        #endregion
    }
}