﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Formatting;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions
{
    internal static class MathUtility
    {
        #region Constants

        private static readonly DateTime Mar1st1900 = new DateTime(1900, 3, 1);
        private static readonly DateTime Jan1st1900 = new DateTime(1900, 1, 1);
        public static readonly double CotangentLimit = 134217728;
        public static readonly double CeilingFloorUpperLimit = 9.99E+307;
        public static readonly double CeilingFloorLowerLimit = -9.99E+307;

        #endregion


        #region Methods

        public static double Acosh(double value)
        {
            return Math.Log(value + Math.Sqrt(Math.Pow(value, 2) - 1));
        }

        public static double Asinh(double value)
        {
            return Math.Log(value + Math.Sqrt(Math.Pow(value, 2) + 1));
        }

        public static double Atanh(double value)
        {
            return 0.5 * Math.Log((1 + value) / (1 - value));
        }

        public static double Acoth(double value)
        {
            return 0.5 * Math.Log((value + 1) / (value - 1));
        }

        public static double Average(double[] values)
        {
            double sum = 0;

            for (int i = 0; i < values.Length; i++)
            {
                sum += values[i];
            }
            return sum / values.Length;
        }

        public static double RoundUp(double value, int decimals)
        {
            int sign = Math.Sign(value);
            value *= sign;

            return sign * Math.Ceiling(value * Math.Pow(10, decimals)) / Math.Pow(10, decimals);
        }

        public static double RoundDown(double value, int decimals)
        {
            int sign = Math.Sign(value);
            value *= sign;

            return sign * Math.Floor(value * Math.Pow(10, decimals)) / Math.Pow(10, decimals);
        }

        public static double Ceiling(double value, double significance)
        {
            if (significance == 0)
            {
                return 0;
            }

            double difference = value / significance;

            double result = value + (Math.Ceiling(difference) - difference) * significance;

            return result.DoubleWithPrecision();
        }

        public static double CeilingPrecise(double value, double significance)
        {
            if (significance == 0)
            {
                return 0;
            }

            return MathUtility.Ceiling(value, Math.Abs(significance));
        }

        public static double Floor(double value, double significance)
        {
            if (significance == 0)
            {
                return 0;
            }

            double difference = value / significance;

            double result = value - (difference - Math.Floor(difference)) * significance;

            return result.DoubleWithPrecision();
        }

        public static double FloorPrecise(double value, double significance)
        {
            if (significance == 0)
            {
                return 0;
            }

            return MathUtility.Floor(value, Math.Abs(significance));
        }

        public static double Mod(double number, double divisor)
        {
            double result = Math.Abs(number % divisor) * Math.Sign(divisor);
            return result.DoubleWithPrecision();
        }

        public static double MRound(double value, double multiple)
        {
            bool roundUp = Math.Abs(Mod(value, multiple)) >= Math.Abs(multiple / 2);

            if (roundUp)
            {
                return Ceiling(value, multiple);
            }

            return Floor(value, multiple);
        }

        public static double Round(double value, int numDigits)
        {
            decimal decimalValue = (decimal)value;

#if WPF
            if (numDigits > 0)
            {
                return (double)Math.Round(decimalValue, numDigits, MidpointRounding.AwayFromZero);
            }
#endif

            decimal multiplier = (decimal)Math.Pow(10, numDigits);
            return (double)(Math.Round(decimalValue * multiplier) / multiplier);
        }

        public static double Factorial(double number)
        {
            double result = 1;

            for (int i = 2; i <= number; i++)
            {
                result *= i;
            }

            return result;
        }

        public static int GreatestCommonDivisor(double[] arguments)
        {
            int result = LcmGcdCommon(arguments, GreatestCommonDivisor);
            return result;
        }

        public static int LeastCommonMultiple(double[] arguments)
        {
            int result = LcmGcdCommon(arguments, LeastCommonMultiple);
            return result;
        }

        private static int LcmGcdCommon(double[] arguments, Func<double, double, int> func)
        {
            List<double> args = arguments.ToList();

            if (args.Contains(0))
            {
                return 0;
            }

            if (args.Count == 1)
            {
                return (int)args[0];
            }

            args.Sort();

            int result = func(args[0], args[1]);

            for (int i = 2; i < args.Count; i++)
            {
                result = func(result, args[i]);
            }

            return result;
        }

        private static int GreatestCommonDivisor(double a, double b)
        {
            int result = 0;
            int numerator = (int)(a > b ? a : b);
            int divisor = (int)(a > b ? b : a);

            while (numerator > 0)
            {
                int reminder = numerator % divisor;
                if (reminder == 0)
                {
                    result = divisor;
                    break;
                }

                numerator = divisor;
                divisor = reminder;
            }

            return result;
        }

        private static int LeastCommonMultiple(double a, double b)
        {
            double denominator = GreatestCommonDivisor(a, b);
            int result = (int)(Math.Abs(a * b) / denominator);
            return result;
        }

        public static double StandardDeviation(double[] values, bool isEntirePopulation)
        {
            double mean = MathUtility.Average(values);
            double deviationsSum = 0;

            for (int i = 0; i < values.Length; i++)
            {
                double deviation = values[i] - mean;
                deviationsSum += Math.Pow(deviation, 2);
            }

            int divisor = isEntirePopulation ? values.Length : values.Length - 1;

            return Math.Sqrt(deviationsSum / divisor);
        }

        public static bool GetTimeSerialNumber(double hours, double minutes, double seconds, out double timeSerialNumber)
        {
            bool result = false;
            timeSerialNumber = 0;

            int hoursRounded = (int)hours;
            int minutesRounded = (int)minutes;
            int secondsRounded = (int)seconds;

            if (hoursRounded >= 0 && minutesRounded >= 0 && secondsRounded >= 0 && hoursRounded <= 32767 && minutesRounded <= 32767 && secondsRounded <= 32767)
            {
                DateTime time = new DateTime();

                try
                {
                    bool hoursAdded = false;
                    bool minutesAdded = false;
                    bool secondsAdded = false;

                    if (hoursRounded > 0)
                    {
                        time = time.AddHours(hoursRounded);
                        hoursAdded = true;
                    }

                    if (minutesRounded > 0)
                    {
                        time = time.AddMinutes(minutesRounded);
                        minutesAdded = true;
                    }

                    if (secondsRounded > 0)
                    {
                        time = time.AddSeconds(secondsRounded);
                        secondsAdded = true;
                    }

                    if (!hoursAdded)
                    {
                        time = time.AddHours(hoursRounded);
                    }

                    if (!minutesAdded)
                    {
                        time = time.AddMinutes(minutesRounded);
                    }

                    if (!secondsAdded)
                    {
                        time = time.AddSeconds(secondsRounded);
                    }

                    time = new DateTime(1, 1, 1, time.Hour, time.Minute, time.Second);
                    timeSerialNumber = FormatHelper.ConvertDateTimeToDouble(time);

                    result = true;
                }
                catch (ArgumentOutOfRangeException)
                { }
            }

            return result;
        }

        public static bool GetDateSerialNumber(double year, double month, double day, out double dateSerianNumber)
        {
            bool result = false;
            dateSerianNumber = 0;

            if (!(year <= 0 && month <= 0 && day <= 0))
            {
                int yearRounded = (int)year;
                int monthRounded = (int)month;
                int dayRounded = (int)day;

                if (yearRounded * 12 + monthRounded > 0)
                {
                    monthRounded = monthRounded - 1;
                    dayRounded = dayRounded >= 0 ? dayRounded - 1 : dayRounded;

                    if (yearRounded < 1900)
                    {
                        yearRounded += 1900;
                    }
                    yearRounded = yearRounded - 1;

                    DateTime date = new DateTime();

                    try
                    {
                        if (yearRounded > monthRounded)
                        {
                            date = date.AddYears(yearRounded);
                            date = date.AddMonths(monthRounded);
                        }
                        else
                        {
                            date = date.AddMonths(monthRounded);
                            date = date.AddYears(yearRounded);
                        }

                        if (!(date == FormatHelper.StartDate && dayRounded < 0))
                        {
                            date = date.AddDays(dayRounded);
                            dateSerianNumber = FormatHelper.ConvertDateTimeToDouble(date);
                            result = true;
                        }
                    }
                    catch (ArgumentOutOfRangeException)
                    { }
                }
            }

            return result;
        }

        public static double DoubleWithPrecision(this double d)
        {
            double result;

            // ToString and Parse are used for truncating purposes mimicking Excell behavior.
            string formattedDouble = d.ToString(CultureInfo.InvariantCulture);
            double.TryParse(formattedDouble, NumberStyles.Number | NumberStyles.AllowExponent, CultureInfo.InvariantCulture, out result);
            return result;
        }

        public static bool AreEqual(double a, double b)
        {
            double eps = 1E-8;
            return Math.Abs(a - b) <= ((Math.Abs(a) > Math.Abs(b) ? Math.Abs(b) : Math.Abs(a)) * eps);
        }

        public static bool IsOddEgyptianFraction(double number)
        {
            bool isOddEgyptianFraction = false;

            if (number != 0 && number > -0.5 && number < 0.5)
            {
                int nearestEgyptianFractionDemoninator = (int)Math.Round(1 / number);

                if (Math.Abs(nearestEgyptianFractionDemoninator % 2) == 1)
                {
                    isOddEgyptianFraction = MathUtility.AreEqual(number, 1.0 / nearestEgyptianFractionDemoninator);
                }
            }

            return isOddEgyptianFraction;
        }

        public static int IndexOfBiggestBit(long number)
        {
            int indexOfLastBit = 63;
            while (indexOfLastBit >= 0)
            {
                if (number.GetBitAtIndex(indexOfLastBit) == 1)
                {
                    break;
                }

                indexOfLastBit--;
            }

            return indexOfLastBit;
        }

        public static long GetBitAtIndex(this long number, int index)
        {
            return ((number >> index) & 1L);
        }

        public static string GetImaginarySymbol(string complexNumberString)
        {
            string imaginarySymbol = "i";

            if (string.IsNullOrEmpty(complexNumberString))
            {
                return imaginarySymbol;
            }

            string trimedComplexString = complexNumberString.Trim();
            if (trimedComplexString.Length > 0 && trimedComplexString.Last() == 'j')
            {
                imaginarySymbol = "j";
            }

            return imaginarySymbol;
        }

        public static bool TryGetImaginarySymbol(string[] complexNumberStrings, out string imaginarySymbol)
        {
            bool allComplexHaveSameImaginarySymbol = true;

            imaginarySymbol = "i";
            string firstMetSymbol = null;

            for (int i = 0; i < complexNumberStrings.Length; i++)
            {
                if (string.IsNullOrEmpty(complexNumberStrings[i]))
                {
                    continue;
                }

                string trimmedComplex = complexNumberStrings[i].Trim();
                if (string.IsNullOrEmpty(trimmedComplex))
                {
                    continue;
                }

                char lastSymbol = trimmedComplex.Last();
                if (lastSymbol == 'i' || lastSymbol == 'j')
                {
                    imaginarySymbol = lastSymbol.ToString();

                    if (string.IsNullOrEmpty(firstMetSymbol))
                    {
                        firstMetSymbol = imaginarySymbol;
                    }
                    else if (firstMetSymbol != imaginarySymbol)
                    {
                        allComplexHaveSameImaginarySymbol = false;
                        return allComplexHaveSameImaginarySymbol;
                    }
                }
            }

            return allComplexHaveSameImaginarySymbol;
        }

        #endregion
    }
}