﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Expressions.Parser;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions
{
    /// <summary>
    /// Provides a set of OperatorInfo instances in RadSpreadsheet.
    /// </summary>
    public static class OperatorInfos
    {
        #region Constants

        /// <summary>
        /// Represents Range operatior info.
        /// </summary>
        public static readonly OperatorInfo Range;

        /// <summary>
        /// Represents Union operatior info.
        /// </summary>
        public static readonly OperatorInfo Union;

        /// <summary>
        /// Represents Intersection operatior info.
        /// </summary>
        public static readonly OperatorInfo Intersection;

        /// <summary>
        /// Represents UnaryPlus operatior info.
        /// </summary>
        public static readonly OperatorInfo UnaryPlus;

        /// <summary>
        /// Represents UnaryMinus operatior info.
        /// </summary>
        public static readonly OperatorInfo UnaryMinus;

        /// <summary>
        /// Represents Percent operatior info.
        /// </summary>
        public static readonly OperatorInfo Percent;

        /// <summary>
        /// Represents Power operatior info.
        /// </summary>
        public static readonly OperatorInfo Power;

        /// <summary>
        /// Represents Multiply operatior info.
        /// </summary>
        public static readonly OperatorInfo Multiply;

        /// <summary>
        /// Represents Divide operatior info.
        /// </summary>
        public static readonly OperatorInfo Divide;

        /// <summary>
        /// Represents Plus operatior info.
        /// </summary>
        public static readonly OperatorInfo Plus;

        /// <summary>
        /// Represents Minus operatior info.
        /// </summary>
        public static readonly OperatorInfo Minus;

        /// <summary>
        /// Represents Ampersand operatior info.
        /// </summary>
        public static readonly OperatorInfo Ampersand;

        /// <summary>
        /// Represents Equal operatior info.
        /// </summary>
        public static readonly OperatorInfo Equal;

        /// <summary>
        /// Represents LessThan operatior info.
        /// </summary>
        public static readonly OperatorInfo LessThan;

        /// <summary>
        /// Represents GreaterThan operatior info.
        /// </summary>
        public static readonly OperatorInfo GreaterThan;

        /// <summary>
        /// Represents LessThanOrEqualTo operatior info.
        /// </summary>
        public static readonly OperatorInfo LessThanOrEqualTo;

        /// <summary>
        /// Represents GreaterThanOrEqualTo operatior info.
        /// </summary>
        public static readonly OperatorInfo GreaterThanOrEqualTo;

        /// <summary>
        /// Represents /// <summary>
        /// Represents Range operatior info.
        /// </summary> operatior info.
        /// </summary>
        public static readonly OperatorInfo NotEqual;

        private static readonly Dictionary<ExpressionTokenType, OperatorInfo> tokenTypeToOperatorInfo;

        #endregion


        #region Constructors

        static OperatorInfos()
        {
            tokenTypeToOperatorInfo = new Dictionary<ExpressionTokenType, OperatorInfo>();

            // NOTE: In MS Excel all operators are Left associative. More info in paper "A Spreadsheet Core Implementation in C#", University of Copenhagen.
            Range = RegisterOperatorInfo(ExpressionTokenType.Range, new OperatorInfo(SpreadsheetCultureHelper.Colon, 0, OperatorAssociativity.Left));
            Union = RegisterOperatorInfo(ExpressionTokenType.Union, new OperatorInfo(FormatHelper.DefaultSpreadsheetCulture.ListSeparator, 2, OperatorAssociativity.Left));
            Intersection = RegisterOperatorInfo(ExpressionTokenType.Intersection, new OperatorInfo(SpreadsheetCultureHelper.Space, 3, OperatorAssociativity.Left));
            UnaryPlus = RegisterOperatorInfo(ExpressionTokenType.UnaryPlus, new OperatorInfo(SpreadsheetCultureHelper.UnaryPlus, 5, OperatorAssociativity.Right));
            UnaryMinus = RegisterOperatorInfo(ExpressionTokenType.UnaryMinus, new OperatorInfo(SpreadsheetCultureHelper.UnaryMinus, 5, OperatorAssociativity.Right));
            Percent = RegisterOperatorInfo(ExpressionTokenType.Percent, new OperatorInfo(SpreadsheetCultureHelper.Percent, 7, OperatorAssociativity.Left));
            Power = RegisterOperatorInfo(ExpressionTokenType.Power, new OperatorInfo(SpreadsheetCultureHelper.Power, 10, OperatorAssociativity.Left));
            Multiply = RegisterOperatorInfo(ExpressionTokenType.Multiply, new OperatorInfo(SpreadsheetCultureHelper.Multiplication, 15, OperatorAssociativity.Left));
            Divide = RegisterOperatorInfo(ExpressionTokenType.Divide, new OperatorInfo(SpreadsheetCultureHelper.Division, 15, OperatorAssociativity.Left));
            Plus = RegisterOperatorInfo(ExpressionTokenType.Plus, new OperatorInfo(SpreadsheetCultureHelper.Plus, 20, OperatorAssociativity.Left));
            Minus = RegisterOperatorInfo(ExpressionTokenType.Minus, new OperatorInfo(SpreadsheetCultureHelper.Minus, 20, OperatorAssociativity.Left));
            Ampersand = RegisterOperatorInfo(ExpressionTokenType.Ampersand, new OperatorInfo(SpreadsheetCultureHelper.Ampersand, 20, OperatorAssociativity.Left));
            Equal = RegisterOperatorInfo(ExpressionTokenType.Equal, new OperatorInfo(SpreadsheetCultureHelper.Equal, 25, OperatorAssociativity.Left));
            LessThan = RegisterOperatorInfo(ExpressionTokenType.LessThan, new OperatorInfo(SpreadsheetCultureHelper.LessThan, 25, OperatorAssociativity.Left));
            GreaterThan = RegisterOperatorInfo(ExpressionTokenType.GreaterThan, new OperatorInfo(SpreadsheetCultureHelper.GreaterThan, 25, OperatorAssociativity.Left));
            LessThanOrEqualTo = RegisterOperatorInfo(ExpressionTokenType.LessThanOrEqualTo, new OperatorInfo(SpreadsheetCultureHelper.LessThanOrEqualTo, 25, OperatorAssociativity.Left));
            GreaterThanOrEqualTo = RegisterOperatorInfo(ExpressionTokenType.GreaterThanOrEqualTo, new OperatorInfo(SpreadsheetCultureHelper.GreaterThanOrEqualTo, 25, OperatorAssociativity.Left));
            NotEqual = RegisterOperatorInfo(ExpressionTokenType.NotEqual, new OperatorInfo(SpreadsheetCultureHelper.NotEqual, 25, OperatorAssociativity.Left));
        }

        #endregion


        #region Methods

        private static OperatorInfo RegisterOperatorInfo(ExpressionTokenType tokenType, OperatorInfo operatorInfo)
        {
            Guard.ThrowExceptionIfNull(operatorInfo, "operatorInfo");

            tokenTypeToOperatorInfo.Add(tokenType, operatorInfo);
            return operatorInfo;
        }

        internal static OperatorInfo GetOperatorInfo(ExpressionTokenType tokenType)
        {
            OperatorInfo operatorInfo;
            tokenTypeToOperatorInfo.TryGetValue(tokenType, out operatorInfo);

            return operatorInfo;
        }

        internal static bool IsOperator(ExpressionTokenType tokenType)
        {
            return GetOperatorInfo(tokenType) != null;
        }

        internal static OperatorAssociativity GetAssociativity(ExpressionTokenType tokenType)
        {
            return GetOperatorInfo(tokenType).Associativity;
        }

        internal static int ComparePrecedence(ExpressionTokenType tokenType1, ExpressionTokenType tokenType2)
        {
            return ComparePrecedence(GetOperatorInfo(tokenType1).Precedence, GetOperatorInfo(tokenType2).Precedence);
        }

        internal static int ComparePrecedence(int precedence1, int precedence2)
        {
            return precedence1 - precedence2;
        }

        internal static bool IsUnaryOperator(ExpressionTokenType tokenType)
        {
            return tokenType == ExpressionTokenType.UnaryMinus ||
                tokenType == ExpressionTokenType.UnaryPlus ||
                tokenType == ExpressionTokenType.Percent;
        }

        internal static bool IsBinaryOperator(ExpressionTokenType tokenType)
        {
            return IsOperator(tokenType) && !IsUnaryOperator(tokenType);
        }

        #endregion
    }
}