﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Parser
{
    /// <summary>
    /// The class reads tokens from the ExpressionTokenizer, modifies and validates the tokens, and passes them to the ExpressionParser class.
    /// The ExpressionLexer is designed to handle ambiguous tokens, insert tokens and merge tokens. The class saves information about the previous 
    /// tokens in the ExpressionLexerContext. Also, it can peek any number of tokens from the Tokenizer to resolve ambiguity. Unlike the ExpressionTokenizer,
    /// which performs simple validation based on the next char, the ExpressionLexer performs validation based on the previous and next tokens.
    /// Currently, the lexer solves the following issues:
    ///  - merges the tokens A1:B1 into a cell reference token, instead of treating this sequence as a range operator (Note that in the expression =A1:B1:C1:D1 the first and last : are not range operators)
    ///  - resolves ambiguity between list separator and union operator;
    ///  - resolves ambiguity between whitespace and intersection operator;
    ///  - distinguishes between unary and binary - and + operators;
    ///  - inserts missing value token (e.g. =INDEX(A1:C3,,2) is a valid expression);
    /// </summary>
    internal class ExpressionLexer : ExpressionLexerBase
    {

        #region Constructors

        public ExpressionLexer(string input, SpreadsheetCultureHelper cultureInfo)
            :base(input, cultureInfo)
        {
            this.RegisterModificationHandler(ExpressionTokenType.A1CellReference, CellReferenceRangeHandler);
            this.RegisterModificationHandler(ExpressionTokenType.Space, SpaceTokenIntersectionOperatorHandler);
            this.RegisterModificationHandler(ExpressionTokenType.ListSeparator, ListSeparatorUnionOperatorHandler);
            this.RegisterModificationHandler(ExpressionTokenType.Plus, UnaryPlusMinusOperatorsHandler);
            this.RegisterModificationHandler(ExpressionTokenType.Minus, UnaryPlusMinusOperatorsHandler);

            this.RegisterValidationHandler(new Func<ExpressionToken, ParseResult>(ValidateBinaryOperator));
            this.RegisterValidationHandler(new Func<ExpressionToken, ParseResult>(ValidatePostfixOperator));
        }

        #endregion


        #region Methods


        private ParseResult ListSeparatorUnionOperatorHandler(ExpressionToken token, out ExpressionToken modifiedToken)
        {
            modifiedToken = token;

            if (this.Context.IsUnionOperator)
            {
                modifiedToken = new ExpressionToken(ExpressionTokenType.Union, this.SpreadsheetCultureInfo.ListSeparator);
            }

            return ParseResult.Successful;
        }

        private ParseResult SpaceTokenIntersectionOperatorHandler(ExpressionToken token, out ExpressionToken modifiedToken)
        {
            modifiedToken = token;

            if (this.IsIntersectionOperator())
            {
                modifiedToken = new ExpressionToken(ExpressionTokenType.Intersection, token.Value);
            }

            return ParseResult.Successful;
        }

        private ParseResult UnaryPlusMinusOperatorsHandler(ExpressionToken token, out ExpressionToken modifiedToken)
        {
            modifiedToken = token;

            if (this.IsValidUnaryPlusMinusOperator())
            {
                if (token.TokenType == ExpressionTokenType.Minus)
                {
                    modifiedToken = ExpressionToken.UnaryMinus;
                }
                else
                {
                    modifiedToken = ExpressionToken.UnaryPlus;
                }
            }

            return ParseResult.Successful;
        }

        private ParseResult ValidateBinaryOperator(ExpressionToken token)
        {
            if (OperatorInfos.IsBinaryOperator(token.TokenType))
            {
                if (PreviousNonSpaceToken == null || !this.IsValidBinaryOperator())
                {
                    return ParseResult.Error;
                }
            }

            return ParseResult.Successful;
        }

        private ParseResult ValidatePostfixOperator(ExpressionToken token)
        {
            if (token.TokenType == ExpressionTokenType.Percent)
            {
                if (!this.IsValidPostFixOperator())
                {
                    return ParseResult.Error;
                }
            }

            return ParseResult.Successful;
        }

        private bool IsIntersectionOperator()
        {
            ExpressionToken nextToken;
            if (this.PreviousNonSpaceToken == null || this.PeekToken(out nextToken) != ParseResult.Successful)
            {
                return false;
            }

            if (OperatorInfos.IsOperator(this.PreviousNonSpaceToken.TokenType) ||
                OperatorInfos.IsOperator(nextToken.TokenType))
            {
                return false;
            }

            if (this.PreviousNonSpaceToken.TokenType == ExpressionTokenType.ListSeparator ||
                this.PreviousNonSpaceToken.TokenType == ExpressionTokenType.Number ||
                this.PreviousNonSpaceToken.TokenType == ExpressionTokenType.Text ||
                this.PreviousNonSpaceToken.TokenType == ExpressionTokenType.Array ||
                this.PreviousNonSpaceToken.TokenType == ExpressionTokenType.LeftParenthesis ||
                nextToken.TokenType == ExpressionTokenType.ListSeparator ||
                nextToken.TokenType == ExpressionTokenType.Number ||
                nextToken.TokenType == ExpressionTokenType.Text ||
                nextToken.TokenType == ExpressionTokenType.Array ||
                nextToken.TokenType == ExpressionTokenType.RightParenthesis)
            {
                return false;
            }

            return true;
        }

        private bool IsValidBinaryOperator()
        {
            ExpressionToken nextToken;
            if (this.PeekToken(out nextToken) != ParseResult.Successful)
            {
                return false;
            }

            if (nextToken == null)
            {
                return false;
            }

            return nextToken.TokenType != ExpressionTokenType.ListSeparator &&
                nextToken.TokenType != ExpressionTokenType.RightParenthesis &&
                nextToken.TokenType != ExpressionTokenType.Multiply &&
                nextToken.TokenType != ExpressionTokenType.Divide &&
                nextToken.TokenType != ExpressionTokenType.Power &&
                nextToken.TokenType != ExpressionTokenType.Ampersand &&
                nextToken.TokenType != ExpressionTokenType.Range;
        }

        private bool IsValidPostFixOperator()
        {
            if (PreviousNonSpaceToken == null)
            {
                return false;
            }

            return !this.IsValidUnaryOperatorPrecedingToken();
        }

        private bool IsValidUnaryPlusMinusOperator()
        {
            if (PreviousNonSpaceToken == null)
            {
                return true;
            }

            return this.IsValidUnaryOperatorPrecedingToken();
        }

        private bool IsValidUnaryOperatorPrecedingToken()
        {
            return this.PreviousNonSpaceToken.TokenType != ExpressionTokenType.Number &&
                   this.PreviousNonSpaceToken.TokenType != ExpressionTokenType.Array &&
                   this.PreviousNonSpaceToken.TokenType != ExpressionTokenType.Error &&
                   this.PreviousNonSpaceToken.TokenType != ExpressionTokenType.True &&
                   this.PreviousNonSpaceToken.TokenType != ExpressionTokenType.False &&
                   this.PreviousNonSpaceToken.TokenType != ExpressionTokenType.Text &&
                   this.PreviousNonSpaceToken.TokenType != ExpressionTokenType.A1CellReference &&
                   this.PreviousNonSpaceToken.TokenType != ExpressionTokenType.A1CellReferenceRange &&
                   this.PreviousNonSpaceToken.TokenType != ExpressionTokenType.RightParenthesis &&
                   this.PreviousNonSpaceToken.TokenType != ExpressionTokenType.DefinedName &&
                   this.PreviousNonSpaceToken.TokenType != ExpressionTokenType.Percent;
        }

        protected override bool ShouldBreakOnResult(ParseResult result)
        {
            if (result == ParseResult.Error)
            {
                return true;
            }

            return false;
        }

        #endregion
    }
}
