﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Parser
{
    internal abstract class ExpressionLexerBase
    {
        private readonly Dictionary<ExpressionTokenType, ExpressionTokenHandler> modificationHandlers;
        private readonly List<Func<ExpressionToken, ParseResult>> validationHandlers;
        
        private readonly ExpressionTokenizer tokenizer;
        private readonly ExpressionLexerContext context;
        private readonly SpreadsheetCultureHelper spreadsheetCultureInfo;
        private readonly List<ExpressionToken> peekedTokens;

        private ExpressionToken previousNonSpaceToken;

        protected ExpressionLexerContext Context
        {
            get
            {
                return this.context;
            }
        }

        protected SpreadsheetCultureHelper SpreadsheetCultureInfo
        {
            get
            {
                return this.spreadsheetCultureInfo;
            }
        }

        protected ExpressionToken PreviousNonSpaceToken
        {
            get
            {
                return this.previousNonSpaceToken;
            }
        }

        protected ExpressionLexerBase(string input, SpreadsheetCultureHelper cultureInfo)
        {
            this.tokenizer = new ExpressionTokenizer(input, cultureInfo);
            this.peekedTokens = new List<ExpressionToken>();
            this.context = new ExpressionLexerContext();
            this.spreadsheetCultureInfo = cultureInfo;

            this.modificationHandlers = new Dictionary<ExpressionTokenType, ExpressionTokenHandler>();
            this.validationHandlers = new List<Func<ExpressionToken, ParseResult>>();
        }

        /// <summary>
        /// Reads the next token from the tokenizer, modifies the token if necessary, and validates the token.
        /// </summary>
        /// <param name="token">The token.</param>
        /// <returns></returns>
        public ParseResult Read(out ExpressionToken token)
        {
            ParseResult readResult = this.PopToken(out token);

            if (readResult == ParseResult.Successful)
            {
                this.UpdateContext(token);

                ExpressionTokenHandler handler;
                if (modificationHandlers.TryGetValue(token.TokenType, out handler))
                {
                    ExpressionToken modifiedToken;
                    ParseResult result = handler(token, out modifiedToken);
                    if (this.ShouldBreakOnResult(result))
                    {
                        return result;
                    }
                    
                    token = modifiedToken;
                }

                foreach (var validationHandler in this.validationHandlers)
                {
                    ParseResult result = validationHandler(token);
                    if (this.ShouldBreakOnResult(result))
                    {
                        return result;
                    }
                }

                if (this.IsMissingValueTokenRequired(token.TokenType))
                {
                    this.peekedTokens.Add(token);
                    token = ExpressionToken.MissingValue;
                }

                this.UpdatePreviousToken(token);
            }

            return readResult;
        }

        protected abstract bool ShouldBreakOnResult(ParseResult result);

        protected ParseResult CellReferenceRangeHandler(ExpressionToken token, out ExpressionToken modifiedToken)
        {
            modifiedToken = token;

            ParseResult result = this.IsCellReferenceRangeToken();
            if (result != ParseResult.Successful)
            {
                return result;
            }

            CellReferenceExpressionToken fromCellReferenceToken = (CellReferenceExpressionToken)token;
            CellReferenceExpressionToken toCellReferenceToken = (CellReferenceExpressionToken)this.peekedTokens[this.peekedTokens.Count - 1];

            string fromSheetName = TextHelper.DecodeWorksheetName(fromCellReferenceToken.WorksheetName);
            string toSheetName = TextHelper.DecodeWorksheetName(toCellReferenceToken.WorksheetName);

            if (string.IsNullOrEmpty(toSheetName) || fromSheetName.Equals(toSheetName, StringComparison.CurrentCultureIgnoreCase))
            {
                modifiedToken = new CellReferenceRangeExpressionToken(fromCellReferenceToken, toCellReferenceToken);

                this.peekedTokens.RemoveRange(0, this.peekedTokens.Count);
                
                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult IsCellReferenceRangeToken()
        {
            ParseResult result = this.ReadUntilTokenOfType(ExpressionTokenType.Range);
            if (result == ParseResult.Successful)
            {
                result = this.ReadUntilTokenOfType(ExpressionTokenType.A1CellReference);
                if (result == ParseResult.Successful)
                {
                    CellReferenceExpressionToken token = this.peekedTokens[this.peekedTokens.Count - 1] as CellReferenceExpressionToken;
                    if (token == null || !string.IsNullOrEmpty(token.WorksheetName))
                    {
                        result = ParseResult.Unsuccessful;
                    }
                }
            }

            return result;
        }

        /// <summary>
        /// The method ignores the peeked tokens and reads tokens from the 
        /// ExpressionTokenizer until a token of the specified type is reached.
        /// </summary>
        /// <param name="tokenType">Type of the token.</param>
        /// <returns></returns>
        private ParseResult ReadUntilTokenOfType(ExpressionTokenType tokenType)
        {
            ExpressionToken nextToken;
            ParseResult result = this.tokenizer.Read(out nextToken);

            if (result == ParseResult.Successful)
            {
                this.peekedTokens.Add(nextToken);

                if (nextToken.TokenType == ExpressionTokenType.Space)
                {
                    return ReadUntilTokenOfType(tokenType);
                }

                return nextToken.TokenType == tokenType ? ParseResult.Successful : ParseResult.Unsuccessful;
            }

            return result;
        }

        /// <summary>
        /// Peeks the next token.
        /// </summary>
        /// <param name="nextToken">The next token.</param>
        /// <returns></returns>
        protected ParseResult PeekToken(out ExpressionToken nextToken)
        {
            if (this.peekedTokens.Count > 0)
            {
                nextToken = this.peekedTokens[0];
                return ParseResult.Successful;
            }

            ParseResult result = this.tokenizer.Read(out nextToken);

            if (result == ParseResult.Successful && nextToken != null)
            {
                this.peekedTokens.Add(nextToken);
            }

            return result;
        }

        protected void RegisterModificationHandler(ExpressionTokenType tokenType, ExpressionTokenHandler modificationHandler)
        {
            Guard.ThrowExceptionIfNull(modificationHandler, "modificationHandler");

            this.modificationHandlers.Add(tokenType, modificationHandler);
        }

        protected void RegisterValidationHandler(Func<ExpressionToken, ParseResult> validationHandler)
        {
            Guard.ThrowExceptionIfNull(validationHandler, "validationHandler");

            this.validationHandlers.Add(validationHandler);
        }

        /// <summary>
        /// Pops the next token. If there are already peeked tokens, the method returns the first element. 
        /// If there are no peeked tokens, the method reads from the ExpressionTokenizer.
        /// </summary>
        /// <param name="nextToken">The next token.</param>
        /// <returns></returns>
        private ParseResult PopToken(out ExpressionToken nextToken)
        {
            if (this.peekedTokens.Count > 0)
            {
                nextToken = this.peekedTokens[0];
                this.peekedTokens.RemoveAt(0);
                return ParseResult.Successful;
            }

            return this.tokenizer.Read(out nextToken);
        }

        private void UpdateContext(ExpressionToken token)
        {
            if (token.TokenType == ExpressionTokenType.LeftParenthesis)
            {
                if (this.previousNonSpaceToken != null && this.previousNonSpaceToken.TokenType == ExpressionTokenType.Function)
                {
                    this.context.RegisterFunction();
                }
                else
                {
                    this.context.RegisterLeftParenthesis();
                }
            }
            else if (token.TokenType == ExpressionTokenType.RightParenthesis)
            {
                this.context.RegisterRightParenthesis();
            }
        }

        private void UpdatePreviousToken(ExpressionToken token)
        {
            if (token.TokenType != ExpressionTokenType.Space)
            {
                this.previousNonSpaceToken = token;
            }
        }

        private bool IsMissingValueTokenRequired(ExpressionTokenType tokenType)
        {
            if (this.previousNonSpaceToken != null)
            {
                if ((this.previousNonSpaceToken.TokenType == ExpressionTokenType.LeftParenthesis && tokenType == ExpressionTokenType.ListSeparator) ||
                    (this.previousNonSpaceToken.TokenType == ExpressionTokenType.ListSeparator && tokenType == ExpressionTokenType.ListSeparator) ||
                    (this.previousNonSpaceToken.TokenType == ExpressionTokenType.ListSeparator && tokenType == ExpressionTokenType.RightParenthesis))
                {
                    return true;
                }
            }

            return false;
        }
    }
}
