﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Parser
{
    /// <summary>
    /// The class reads tokens from the ExpressionLexer class, parses them, and builds an expression tree and an InputStringCollection.
    /// The ExpressionParser maps the type of the expression token with a RadExpression. Note that the ExpressionParser does not deal 
    /// with ambiguous tokens - the ExpressionLexer handles such cases.
    /// </summary>
    internal class ExpressionParser
    {
        #region Fields

        private delegate ParseResult TokenHandler(ExpressionToken token, Stack<RadExpression> expressionStack, Worksheet worksheet, int rowIndex, int columnIndex,
            InputStringTokenCollection expressionTokenCollection);

        private readonly TokenHandler[] tokenHandlers;

        private readonly Dictionary<ExpressionTokenType, Func<RadExpression>> tokenToExpression;
        private readonly Dictionary<ExpressionTokenType, Func<RadExpression, RadExpression>> unaryOperatorToExpression;
        private readonly Dictionary<ExpressionTokenType, Func<RadExpression, RadExpression, RadExpression>> binaryOperatorToExpression;
        private readonly Dictionary<ExpressionTokenType, Func<InputStringToken, UserInputStringBase>> inputTokenToUserInput;

        private bool errorEncountered;
        private bool isParsingGlobalDefinedName;

        #endregion

        public static ExpressionParser ExpressionParserInstance
        {
            get;
            private set;
        }

        #region Constructor

        static ExpressionParser()
        {
            ExpressionParser.ExpressionParserInstance = new ExpressionParser();
        }

        private ExpressionParser()
        {
            this.tokenHandlers = new TokenHandler[]
            {
                this.ParseLiteral,
                this.ParseNumber,
                this.ParseText,
                this.ParseError,
                this.ParseUnaryOperator,
                this.ParseBinaryOperator,
                this.ParseCellReference,
                this.ParseCellReferenceRange,
                this.ParseVariable,
                this.ParseFunction,
                this.ParseArray
            };

            this.tokenToExpression = new Dictionary<ExpressionTokenType, Func<RadExpression>>();
            this.tokenToExpression.Add(ExpressionTokenType.True, () => { return BooleanExpression.True; });
            this.tokenToExpression.Add(ExpressionTokenType.False, () => { return BooleanExpression.False; });
            this.tokenToExpression.Add(ExpressionTokenType.MissingValue, () => { return EmptyExpression.Empty; });
            this.tokenToExpression.Add(ExpressionTokenType.FunctionStart, () => { return new FunctionStartExpression(); });

            this.unaryOperatorToExpression = new Dictionary<ExpressionTokenType, Func<RadExpression, RadExpression>>();
            this.unaryOperatorToExpression.Add(ExpressionTokenType.Percent, (p) => { return new PercentExpression(p); });
            this.unaryOperatorToExpression.Add(ExpressionTokenType.UnaryPlus, (p) => { return new UnaryPlusExpression(p); });
            this.unaryOperatorToExpression.Add(ExpressionTokenType.UnaryMinus, (p) => { return new UnaryMinusExpression(p); });

            this.binaryOperatorToExpression = new Dictionary<ExpressionTokenType, Func<RadExpression, RadExpression, RadExpression>>();
            this.binaryOperatorToExpression.Add(ExpressionTokenType.Plus, (p, q) => { return new AdditionExpression(p, q); });
            this.binaryOperatorToExpression.Add(ExpressionTokenType.Minus, (p, q) => { return new SubtractionExpression(p, q); });
            this.binaryOperatorToExpression.Add(ExpressionTokenType.Multiply, (p, q) => { return new MultiplicationExpression(p, q); });
            this.binaryOperatorToExpression.Add(ExpressionTokenType.Divide, (p, q) => { return new DivisionExpression(p, q); });
            this.binaryOperatorToExpression.Add(ExpressionTokenType.Power, (p, q) => { return new PowerExpression(p, q); });
            this.binaryOperatorToExpression.Add(ExpressionTokenType.Ampersand, (p, q) => { return new AmpersandExpression(p, q); });
            this.binaryOperatorToExpression.Add(ExpressionTokenType.Equal, (p, q) => { return new EqualExpression(p, q); });
            this.binaryOperatorToExpression.Add(ExpressionTokenType.NotEqual, (p, q) => { return new NotEqualExpression(p, q); });
            this.binaryOperatorToExpression.Add(ExpressionTokenType.GreaterThan, (p, q) => { return new GreaterThanExpression(p, q); });
            this.binaryOperatorToExpression.Add(ExpressionTokenType.GreaterThanOrEqualTo, (p, q) => { return new GreaterThanOrEqualToExpression(p, q); });
            this.binaryOperatorToExpression.Add(ExpressionTokenType.LessThan, (p, q) => { return new LessThanExpression(p, q); });
            this.binaryOperatorToExpression.Add(ExpressionTokenType.LessThanOrEqualTo, (p, q) => { return new LessThanOrEqualToExpression(p, q); });
            this.binaryOperatorToExpression.Add(ExpressionTokenType.Range, (p, q) => { return new RangeExpression(p, q); });
            this.binaryOperatorToExpression.Add(ExpressionTokenType.Intersection, (p, q) => { return new IntersectionExpression(p, q); });
            this.binaryOperatorToExpression.Add(ExpressionTokenType.Union, (p, q) => { return new UnionExpression(p, q); });

            this.inputTokenToUserInput = new Dictionary<ExpressionTokenType, Func<InputStringToken, UserInputStringBase>>();
            this.inputTokenToUserInput.Add(ExpressionTokenType.ListSeparator, (token) => { return new ListSeparatorInputString(); });
            this.inputTokenToUserInput.Add(ExpressionTokenType.Union, (token) => { return new ListSeparatorInputString(); });
            this.inputTokenToUserInput.Add(ExpressionTokenType.Number, (token) => { return new ExpressionInputString(token.TranslatableExpression); });
            this.inputTokenToUserInput.Add(ExpressionTokenType.Array, (token) => { return new ExpressionInputString(token.TranslatableExpression); });
            this.inputTokenToUserInput.Add(ExpressionTokenType.A1CellReference, (token) => { return new CellReferenceInputString((CellReferenceRangeExpression)token.TranslatableExpression); });
            this.inputTokenToUserInput.Add(ExpressionTokenType.A1CellReferenceRange, (token) => { return new CellReferenceInputString((CellReferenceRangeExpression)token.TranslatableExpression); });
            this.inputTokenToUserInput.Add(ExpressionTokenType.DefinedName, (token) => { return new DefinedNameInputString((SpreadsheetNameExpression)token.TranslatableExpression); });
        }

        #endregion


        #region Methods

        public ParseResult Parse(string input, Worksheet worksheet, int rowIndex, int columnIndex, SpreadsheetCultureHelper spreadsheetCultureInfo,
            out InputStringCollection stringExpressionCollection, out RadExpression expression, bool appendWorksheetName = false)
        {
            Guard.ThrowExceptionIfNullOrEmpty(input, "input");
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);
            Guard.ThrowExceptionIfNull(spreadsheetCultureInfo, "spreadsheetCultureInfo");

            expression = null;
            this.isParsingGlobalDefinedName = appendWorksheetName;
            stringExpressionCollection = null;
            Stack<RadExpression> expressionStack = new Stack<RadExpression>();
            ExpressionLexer phraseBuilder = new ExpressionLexer(input, spreadsheetCultureInfo);
            InputStringTokenCollection inputStringTokenCollection = new InputStringTokenCollection();

            IEnumerable<ExpressionToken> rpnTokens = this.InfixToReversedPolishNotation(phraseBuilder, inputStringTokenCollection);
            ParseResult result = ParseResult.Unsuccessful;

            foreach (ExpressionToken token in rpnTokens)
            {
                result = this.ParseToken(token, expressionStack, worksheet, rowIndex, columnIndex, inputStringTokenCollection);
                if (result == ParseResult.Error)
                {
                    return ParseResult.Error;
                }
            }

            if (this.errorEncountered || result == ParseResult.Unsuccessful || expressionStack.Count != 1)
            {
                this.errorEncountered = false;
                return ParseResult.Error;
            }

            stringExpressionCollection = this.BuildStringExpressionCollection(inputStringTokenCollection);
            expression = expressionStack.Pop();

            return ParseResult.Successful;
        }

        // TODO: refactor this method
        private IEnumerable<ExpressionToken> InfixToReversedPolishNotation(ExpressionLexer phraseBuilder, InputStringTokenCollection inputStringTokenCollection)
        {
            Stack<ExpressionToken> stack = new Stack<ExpressionToken>();

            ExpressionToken token;
            if (this.ReadAndSaveExpressionToken(phraseBuilder, inputStringTokenCollection, out token) == ParseResult.Error)
            {
                this.errorEncountered = true;
                yield break;
            }

            while (token != null)
            {
                if (token.TokenType == ExpressionTokenType.Function)
                {
                    yield return ExpressionToken.FunctionStart;
                    stack.Push(token);
                }
                else if (token.TokenType == ExpressionTokenType.ListSeparator)
                {
                    while (stack.Count > 0 && stack.Peek().TokenType != ExpressionTokenType.LeftParenthesis)
                    {
                        yield return stack.Pop();
                    }

                    if (stack.Count == 0 || stack.Peek().TokenType != ExpressionTokenType.LeftParenthesis)
                    {
                        this.errorEncountered = true;
                        yield break;
                    }
                }
                else if (OperatorInfos.IsOperator(token.TokenType))
                {
                    foreach (ExpressionToken expressionToken in this.GetOperandsFromStack(stack, token))
                    {
                        yield return expressionToken;
                    }

                    stack.Push(token);
                }
                else if (token.TokenType == ExpressionTokenType.LeftParenthesis)
                {
                    stack.Push(token);
                }
                else if (token.TokenType == ExpressionTokenType.RightParenthesis)
                {
                    while (stack.Count != 0 && stack.Peek().TokenType != ExpressionTokenType.LeftParenthesis)
                    {
                        yield return stack.Pop();
                    }

                    if (stack.Count == 0 || stack.Peek().TokenType != ExpressionTokenType.LeftParenthesis)
                    {
                        this.errorEncountered = true;
                        yield break;
                    }

                    stack.Pop();

                    if (stack.Count != 0 && stack.Peek().TokenType == ExpressionTokenType.Function)
                    {
                        yield return stack.Pop();
                    }
                }
                else
                {
                    if (token.TokenType != ExpressionTokenType.Space)
                    {
                        yield return token;
                    }
                }

                if (this.ReadAndSaveExpressionToken(phraseBuilder, inputStringTokenCollection, out token) == ParseResult.Error)
                {
                    this.errorEncountered = true;
                    yield break;
                }
            }

            while (stack.Count != 0)
            {
                yield return stack.Pop();
            }
        }

        private ParseResult ReadAndSaveExpressionToken(ExpressionLexer phraseBuilder, InputStringTokenCollection inputStringTokenCollection, out ExpressionToken token)
        {
            ParseResult result = phraseBuilder.Read(out token);

            if (result == ParseResult.Successful)
            {
                inputStringTokenCollection.Add(token);
            }

            return result;
        }

        private ParseResult ParseToken(ExpressionToken token, Stack<RadExpression> expressionStack, Worksheet worksheet, int rowIndex, int columnIndex,
            InputStringTokenCollection inputStringTokenCollection)
        {
            ParseResult result = ParseResult.Unsuccessful;

            foreach (TokenHandler tokenHandler in this.tokenHandlers)
            {
                result = tokenHandler(token, expressionStack, worksheet, rowIndex, columnIndex, inputStringTokenCollection);
                if (result != ParseResult.Unsuccessful)
                {
                    break;
                }
            }

            return result;
        }

        private ParseResult ParseLiteral(ExpressionToken token, Stack<RadExpression> expressionStack, Worksheet worksheet,
            int rowIndex, int columnIndex, InputStringTokenCollection inputStringTokenCollection)
        {
            Func<RadExpression> tokenHandler;
            if (this.tokenToExpression.TryGetValue(token.TokenType, out tokenHandler))
            {
                RadExpression expression = tokenHandler();
                expressionStack.Push(expression);

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ParseNumber(ExpressionToken token, Stack<RadExpression> expressionStack, Worksheet worksheet,
            int rowIndex, int columnIndex, InputStringTokenCollection inputStringTokenCollection)
        {
            if (token.TokenType == ExpressionTokenType.Number)
            {
                RadExpression expression = NumberExpression.CreateValidNumberOrErrorExpression(token.NumberValue);
                inputStringTokenCollection.UpdateTokenExpression(token, expression);
                expressionStack.Push(expression);

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ParseText(ExpressionToken token, Stack<RadExpression> expressionStack, Worksheet worksheet,
            int rowIndex, int columnIndex, InputStringTokenCollection inputStringTokenCollection)
        {
            if (token.TokenType == ExpressionTokenType.Text)
            {
                string encodedValue = TextHelper.DecodeValue(token.Value, SpreadsheetCultureHelper.DoubleQuote);
                RadExpression expression = new StringExpression(encodedValue);
                inputStringTokenCollection.UpdateTokenExpression(token, expression);
                expressionStack.Push(expression);

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ParseError(ExpressionToken token, Stack<RadExpression> expressionStack, Worksheet worksheet,
            int rowIndex, int columnIndex, InputStringTokenCollection inputStringTokenCollection)
        {
            if (token.TokenType == ExpressionTokenType.Error)
            {
                RadExpression value = ErrorExpressions.FindErrorExpression(token.Value);
                expressionStack.Push(value);

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ParseUnaryOperator(ExpressionToken token, Stack<RadExpression> expressionStack, Worksheet worksheet,
            int rowIndex, int columnIndex, InputStringTokenCollection inputStringTokenCollection)
        {
            RadExpression[] operands = null;

            if (OperatorInfos.IsUnaryOperator(token.TokenType))
            {
                if (!this.TryPopNOperands(expressionStack, 1, out operands))
                {
                    return ParseResult.Error;
                }
            }

            Func<RadExpression, RadExpression> unaryOperatorHandler;
            if (this.unaryOperatorToExpression.TryGetValue(token.TokenType, out unaryOperatorHandler))
            {
                RadExpression expression = unaryOperatorHandler(operands[0]);
                expressionStack.Push(expression);

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ParseBinaryOperator(ExpressionToken token, Stack<RadExpression> expressionStack, Worksheet worksheet,
            int rowIndex, int columnIndex, InputStringTokenCollection inputStringTokenCollection)
        {
            RadExpression[] operands = null;

            if (OperatorInfos.IsBinaryOperator(token.TokenType))
            {
                if (!this.TryPopNOperands(expressionStack, 2, out operands))
                {
                    return ParseResult.Error;
                }
            }

            Func<RadExpression, RadExpression, RadExpression> binaryOperatorHandler;
            if (this.binaryOperatorToExpression.TryGetValue(token.TokenType, out binaryOperatorHandler))
            {
                RadExpression expression = binaryOperatorHandler(operands[0], operands[1]);
                expressionStack.Push(expression);

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ParseFunction(ExpressionToken token, Stack<RadExpression> expressionStack, Worksheet worksheet,
            int rowIndex, int columnIndex, InputStringTokenCollection inputStringTokenCollection)
        {
            if (token.TokenType == ExpressionTokenType.Function)
            {
                RadExpression[] arguments;
                if (this.ReadFunctionArguments(expressionStack, out arguments))
                {
                    return ParseResult.Error;
                }

                try
                {
                    FunctionExpression functionExpression = new FunctionExpression(token.Value, arguments, worksheet, rowIndex, columnIndex);
                    expressionStack.Push(functionExpression);
                }
                catch (ExpressionException)
                {
                    return ParseResult.Error;
                }

                inputStringTokenCollection.UpdateTokenString(token, token.Value.ToUpper());

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private void GetWorksheetReference(CellReferenceExpressionToken token, Worksheet worksheet, out bool isWorksheetAbsolute, out string worksheetName)
        {
            isWorksheetAbsolute = !string.IsNullOrEmpty(token.WorksheetName);
            worksheetName = isWorksheetAbsolute ? token.WorksheetName : worksheet.Name;

            bool escapedByUser = worksheetName[0] == '\'';
            string decodedWorksheetName = TextHelper.DecodeWorksheetName(worksheetName);
            Worksheet referredWorksheet = SpreadsheetHelper.GetWorksheetByName(worksheet.Workbook, decodedWorksheetName);
            if (referredWorksheet != null)
            {
                worksheetName = TextHelper.EncodeWorksheetName(referredWorksheet.Name, escapedByUser);
            }

            if (string.Equals(decodedWorksheetName, worksheet.Workbook.Name, StringComparison.CurrentCultureIgnoreCase))
            {
                referredWorksheet = worksheet;
                worksheetName = TextHelper.EncodeWorksheetName(worksheet.Name);
                isWorksheetAbsolute = true;
            }
            else
            {
                isWorksheetAbsolute |= this.isParsingGlobalDefinedName;
            }
        }

        private ParseResult ParseCellReference(ExpressionToken token, Stack<RadExpression> expressionStack, Worksheet worksheet,
            int rowIndex, int columnIndex, InputStringTokenCollection inputStringTokenCollection)
        {
            CellReferenceExpressionToken cellReferenceToken = token as CellReferenceExpressionToken;
            if (cellReferenceToken != null)
            {
                string worksheetName;
                bool isWorksheetAbsolute;
                this.GetWorksheetReference(cellReferenceToken, worksheet, out isWorksheetAbsolute, out worksheetName);

                CellReferenceRange reference = new CellReferenceRange(worksheet.Workbook, worksheetName, isWorksheetAbsolute,
                     CellAreaReference.CreateFromCellName(cellReferenceToken.CellOrName, rowIndex, columnIndex));

                RadExpression value = worksheet.Workbook.CellReferenceRangeExpressionCache.GetOrCreateExpression(reference);

                inputStringTokenCollection.UpdateTokenStringAndExpression(token, value.ToString(), value);
                expressionStack.Push(value);

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ParseCellReferenceRange(ExpressionToken token, Stack<RadExpression> expressionStack, Worksheet worksheet,
           int rowIndex, int columnIndex, InputStringTokenCollection inputStringTokenCollection)
        {
            CellReferenceRangeExpressionToken cellReferenceRangeToken = token as CellReferenceRangeExpressionToken;
            if (cellReferenceRangeToken != null)
            {
                string worksheetName;
                bool isWorksheetAbsolute;
                this.GetWorksheetReference(cellReferenceRangeToken.Left, worksheet, out isWorksheetAbsolute, out worksheetName);

                CellReferenceRange reference = new CellReferenceRange(worksheet.Workbook, worksheetName, isWorksheetAbsolute,
                     CellAreaReference.CreateFromTwoCellNames(cellReferenceRangeToken.Left.CellOrName, cellReferenceRangeToken.Right.CellOrName, rowIndex, columnIndex, true));

                RadExpression value = worksheet.Workbook.CellReferenceRangeExpressionCache.GetOrCreateExpression(reference);

                inputStringTokenCollection.UpdateTokenStringAndExpression(token, value.ToString(), value);
                expressionStack.Push(value);

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ParseVariable(ExpressionToken token, Stack<RadExpression> expressionStack, Worksheet worksheet,
            int rowIndex, int columnIndex, InputStringTokenCollection inputStringTokenCollection)
        {
            DefinedNameExpressionToken variableToken = token as DefinedNameExpressionToken;
            if (variableToken != null)
            {
                string name = variableToken.CellOrName;

                DefinedName definedName = worksheet.Workbook.NameManager.FindSpreadsheetName(variableToken.WorksheetName, variableToken.CellOrName, worksheet, worksheet.Workbook) as DefinedName;
                if (definedName != null)
                {
                    name = definedName.Name;
                }

                ExpressionQualifierInfo info = new ExpressionQualifierInfo(worksheet.Workbook, worksheet, variableToken.WorksheetName);
                RadExpression expression = new SpreadsheetNameExpression(info, rowIndex, columnIndex, name);

                inputStringTokenCollection.UpdateTokenStringAndExpression(token, expression.ToString(), expression);
                expressionStack.Push(expression);

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ParseArray(ExpressionToken token, Stack<RadExpression> expressionStack, Worksheet worksheet,
            int rowIndex, int columnIndex, InputStringTokenCollection inputStringTokenCollection)
        {
            ArrayExpressionToken arrayToken = token as ArrayExpressionToken;
            if (token.TokenType == ExpressionTokenType.Array && arrayToken != null)
            {
                RadExpression[,] expressionsArray = new RadExpression[arrayToken.Array.GetLength(0), arrayToken.Array.GetLength(1)];

                for (int i = 0; i < arrayToken.Array.GetLength(0); i++)
                {
                    for (int j = 0; j < arrayToken.Array.GetLength(1); j++)
                    {
                        RadExpression arrayValue = null;
                        switch (arrayToken.Array[i, j].TokenType)
                        {
                            case ExpressionTokenType.Number:
                                arrayValue = NumberExpression.CreateValidNumberOrErrorExpression(arrayToken.Array[i, j].NumberValue);
                                break;

                            case ExpressionTokenType.Text:
                                arrayValue = new StringExpression(arrayToken.Array[i, j].Value);
                                break;

                            case ExpressionTokenType.Error:
                                arrayValue = ErrorExpressions.FindErrorExpression(arrayToken.Array[i, j].Value);
                                break;

                            case ExpressionTokenType.True:
                                arrayValue = BooleanExpression.True;
                                break;

                            case ExpressionTokenType.False:
                                arrayValue = BooleanExpression.False;
                                break;

                            default:
                                return ParseResult.Error;
                        }

                        expressionsArray[i, j] = arrayValue;
                    }
                }

                RadExpression value = new ArrayExpression(expressionsArray);

                inputStringTokenCollection.UpdateTokenStringAndExpression(token, value.ToString(), value);

                expressionStack.Push(value);
                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private bool ReadFunctionArguments(Stack<RadExpression> expressionStack, out RadExpression[] expressions)
        {
            expressions = null;

            Stack<RadExpression> result = new Stack<RadExpression>();
            while (expressionStack.Count > 0 && !(expressionStack.Peek() is FunctionStartExpression))
            {
                result.Push(expressionStack.Pop());
            }

            if (expressionStack.Count == 0 || !(expressionStack.Peek() is FunctionStartExpression))
            {
                return true;
            }

            expressionStack.Pop();
            expressions = result.ToArray();

            return false;
        }

        private IEnumerable<ExpressionToken> GetOperandsFromStack(Stack<ExpressionToken> stack, ExpressionToken token)
        {
            while (stack.Count > 0 && OperatorInfos.IsOperator(stack.Peek().TokenType))
            {
                OperatorAssociativity tokenAssociativity = OperatorInfos.GetAssociativity(token.TokenType);
                int precedenceCompare = OperatorInfos.ComparePrecedence(token.TokenType, stack.Peek().TokenType);
                if ((tokenAssociativity == OperatorAssociativity.Left && precedenceCompare >= 0) ||
                    (tokenAssociativity == OperatorAssociativity.Right && precedenceCompare > 0))
                {
                    yield return stack.Pop();
                }
                else
                {
                    break;
                }
            }
        }

        private bool TryPopNOperands(Stack<RadExpression> expressionStack, int n, out RadExpression[] operands)
        {
            operands = null;

            Stack<RadExpression> result = new Stack<RadExpression>();
            while (expressionStack.Count > 0 && n > 0)
            {
                result.Push(expressionStack.Pop());
                n--;
            }

            if (n != 0)
            {
                return false;
            }

            operands = result.ToArray();

            return true;
        }

        private InputStringCollection BuildStringExpressionCollection(InputStringTokenCollection inputStringTokenCollection)
        {
            InputStringCollection stringExpressionCollection = new InputStringCollection();

            for (int i = 0; i < inputStringTokenCollection.Count; i++)
            {
                InputStringToken token = inputStringTokenCollection[i];
                UserInputStringBase inputString;

                Func<InputStringToken, UserInputStringBase> func;
                if (this.inputTokenToUserInput.TryGetValue(token.Token.TokenType, out func))
                {
                    inputString = func(token);
                }
                else
                {
                    inputString = new FormatInputString(token.StringValue);
                }

                stringExpressionCollection.Add(inputString);
            }

            return stringExpressionCollection;
        }

        #endregion


        #region Private Classes

        private class InputStringToken
        {
            public ExpressionToken Token { get; set; }
            public string StringValue { get; set; }
            public RadExpression TranslatableExpression { get; set; }

            public override bool Equals(object obj)
            {
                InputStringToken other = obj as InputStringToken;
                if (other == null)
                {
                    return false;
                }

                return this.Token.Equals(other.Token);
            }

            public override int GetHashCode()
            {
                return this.Token.GetHashCode();
            }
        }

        /// <summary>
        /// This collection is resposible for updating the string casing of the user input
        /// </summary>
        private class InputStringTokenCollection
        {
            private readonly List<InputStringToken> innerList;

            public InputStringTokenCollection()
            {
                this.innerList = new List<InputStringToken>();
            }

            public InputStringToken this[int i]
            {
                get
                {
                    return this.innerList[i];
                }
            }

            public int Count
            {
                get
                {
                    return this.innerList.Count;
                }
            }

            public void Add(ExpressionToken token)
            {
                this.innerList.Add(new InputStringToken() { Token = token, StringValue = token.Value });
            }

            public void UpdateTokenString(ExpressionToken token, string updatedString)
            {
                InputStringToken inputToken = this.IndexOfStringExpressionToken(token);
                inputToken.StringValue = updatedString;
            }

            public void UpdateTokenStringAndExpression(ExpressionToken token, string updatedString, RadExpression updatedExpression)
            {
                InputStringToken inputToken = this.IndexOfStringExpressionToken(token);

                inputToken.StringValue = updatedString;
                inputToken.TranslatableExpression = updatedExpression;
            }

            public void UpdateTokenExpression(ExpressionToken token, RadExpression expression)
            {
                InputStringToken inputToken = this.IndexOfStringExpressionToken(token);
                inputToken.TranslatableExpression = expression;
            }

            private InputStringToken IndexOfStringExpressionToken(ExpressionToken token)
            {
                InputStringToken inputStringToken = this.innerList.First((t) =>
                {
                    return t.Token.Equals(token);
                });

                Guard.ThrowExceptionIfNull(inputStringToken, "inputStringToken");
                return inputStringToken;
            }
        }

        #endregion
    }
}