﻿using System;
using System.Text;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions.Parser
{
    /// <summary>
    /// The class reads chars from the ExpressionTokenReader class and attempts to convert them to tokens.
    /// It does not keep track of the preceding tokens and it cannot peek tokens. For this reason, the class 
    /// cannot handle ambiguities: it cannot determine whether the space should be treated as intersection operator, 
    /// whether the list separation should be union operator and whether the minus operator is a unary and binary.
    /// The class performs validation based only on the next char. For example, if it registers a multiplication 
    /// operator (*) the following char cannot be binary or postix operator, such as division (/) or percent (%).
    /// </summary>
    internal class ExpressionTokenizer
    {
        #region Fields

        private static readonly Dictionary<string, ExpressionToken> stringToOneCharExpressionToken;
        private static readonly Dictionary<string, ExpressionToken> stringToArithmeticOperatorExpressionToken;
        private static readonly Dictionary<string, ExpressionToken> stringToComparisonOperatorExpressionToken;
        private static readonly Dictionary<string, ExpressionToken> stringToErrorExpressionToken;
        private static readonly string globalNamesImportString = "[0]!";

        private delegate ParseResult ExpressionTokenHandler(out ExpressionToken token);
        private delegate ParseResult TextExpressionTokenHandler(string input, out ExpressionToken token);
        private delegate bool NumberParserHandler(string input, out double number);

        private readonly SpreadsheetCultureHelper spreadsheetCultureInfo;
        private readonly ExpressionTokenReader tokenReader;

        private readonly List<ExpressionTokenHandler> tokenHandlers;
        private readonly List<ExpressionTokenHandler> arrayTokenHandlers;
        private readonly List<TextExpressionTokenHandler> textTokenHandlers;
        private readonly List<TextExpressionTokenHandler> cellReferenceTokenHandlers;
        private readonly List<NumberParserHandler> numberParseHandlers;

        #endregion


        #region Constructors

        static ExpressionTokenizer()
        {
            stringToOneCharExpressionToken = new Dictionary<string, ExpressionToken>();
            stringToOneCharExpressionToken.Add(ExpressionToken.Percent.Value, ExpressionToken.Percent);
            stringToOneCharExpressionToken.Add(ExpressionToken.LeftParenthesis.Value, ExpressionToken.LeftParenthesis);
            stringToOneCharExpressionToken.Add(ExpressionToken.RightParenthesis.Value, ExpressionToken.RightParenthesis);

            stringToArithmeticOperatorExpressionToken = new Dictionary<string, ExpressionToken>();
            stringToArithmeticOperatorExpressionToken.Add(ExpressionToken.Plus.Value, ExpressionToken.Plus);
            stringToArithmeticOperatorExpressionToken.Add(ExpressionToken.Minus.Value, ExpressionToken.Minus);
            stringToArithmeticOperatorExpressionToken.Add(ExpressionToken.Multiply.Value, ExpressionToken.Multiply);
            stringToArithmeticOperatorExpressionToken.Add(ExpressionToken.Divide.Value, ExpressionToken.Divide);
            stringToArithmeticOperatorExpressionToken.Add(ExpressionToken.Power.Value, ExpressionToken.Power);
            stringToArithmeticOperatorExpressionToken.Add(ExpressionToken.Ampersand.Value, ExpressionToken.Ampersand);
            stringToArithmeticOperatorExpressionToken.Add(ExpressionToken.Range.Value, ExpressionToken.Range);

            stringToComparisonOperatorExpressionToken = new Dictionary<string, ExpressionToken>();
            stringToComparisonOperatorExpressionToken.Add(ExpressionToken.Equal.Value, ExpressionToken.Equal);
            stringToComparisonOperatorExpressionToken.Add(ExpressionToken.NotEqual.Value, ExpressionToken.NotEqual);
            stringToComparisonOperatorExpressionToken.Add(ExpressionToken.GreaterThan.Value, ExpressionToken.GreaterThan);
            stringToComparisonOperatorExpressionToken.Add(ExpressionToken.GreaterThanOrEqualTo.Value, ExpressionToken.GreaterThanOrEqualTo);
            stringToComparisonOperatorExpressionToken.Add(ExpressionToken.LessThan.Value, ExpressionToken.LessThan);
            stringToComparisonOperatorExpressionToken.Add(ExpressionToken.LessThanOrEqualTo.Value, ExpressionToken.LessThanOrEqualTo);

            stringToErrorExpressionToken = new Dictionary<string, ExpressionToken>();
            stringToErrorExpressionToken.Add(ExpressionToken.NullError.Value, ExpressionToken.NullError);
            stringToErrorExpressionToken.Add(ExpressionToken.DivisionByZeroError.Value, ExpressionToken.DivisionByZeroError);
            stringToErrorExpressionToken.Add(ExpressionToken.ValueError.Value, ExpressionToken.ValueError);
            stringToErrorExpressionToken.Add(ExpressionToken.ReferenceError.Value, ExpressionToken.ReferenceError);
            stringToErrorExpressionToken.Add(ExpressionToken.NameError.Value, ExpressionToken.NameError);
            stringToErrorExpressionToken.Add(ExpressionToken.NumberError.Value, ExpressionToken.NumberError);
            stringToErrorExpressionToken.Add(ExpressionToken.NotAvailableError.Value, ExpressionToken.NotAvailableError);
        }

        public ExpressionTokenizer(string input, SpreadsheetCultureHelper spreadsheetCultureInfo)
        {
            Guard.ThrowExceptionIfNull(spreadsheetCultureInfo, "spreadsheetCultureInfo");
            Guard.ThrowExceptionIfNullOrEmpty(input, "input");

            this.spreadsheetCultureInfo = spreadsheetCultureInfo;
            this.tokenReader = new ExpressionTokenReader(input);

            this.tokenHandlers = new List<ExpressionTokenHandler>();
            this.textTokenHandlers = new List<TextExpressionTokenHandler>();
            this.arrayTokenHandlers = new List<ExpressionTokenHandler>();
            this.cellReferenceTokenHandlers = new List<TextExpressionTokenHandler>();
            this.numberParseHandlers = new List<NumberParserHandler>();

            this.InitTokenHandlers();
        }

        #endregion


        #region Methods

        private void InitTokenHandlers()
        {
            this.tokenHandlers.Add(new ExpressionTokenHandler(this.ReadSpaceToken));
            this.tokenHandlers.Add(new ExpressionTokenHandler(this.ReadNumberToken));
            this.tokenHandlers.Add(new ExpressionTokenHandler(this.ReadOneCharToken));
            this.tokenHandlers.Add(new ExpressionTokenHandler(this.ReadArithmeticOperatorToken));
            this.tokenHandlers.Add(new ExpressionTokenHandler(this.ReadComparisonOperatorToken));
            this.tokenHandlers.Add(new ExpressionTokenHandler(this.ReadListSeparatorToken));
            this.tokenHandlers.Add(new ExpressionTokenHandler(this.ReadErrorToken));
            this.tokenHandlers.Add(new ExpressionTokenHandler(this.ReadTextToken));
            this.tokenHandlers.Add(new ExpressionTokenHandler(this.ReadArrayToken));
            this.tokenHandlers.Add(new ExpressionTokenHandler(this.ReadEscapedCellReferenceToken));
            this.tokenHandlers.Add(new ExpressionTokenHandler(this.ReadFunctionCellReferenceOrVariableToken));

            this.textTokenHandlers.Add(new TextExpressionTokenHandler(this.ReadLiteralToken));
            this.textTokenHandlers.Add(new TextExpressionTokenHandler(this.ReadFunctionToken));
            this.textTokenHandlers.Add(new TextExpressionTokenHandler(this.ReadCellReferenceToken));

            this.cellReferenceTokenHandlers.Add(new TextExpressionTokenHandler(this.ReadWorksheetRefErrorCellReference));
            this.cellReferenceTokenHandlers.Add(new TextExpressionTokenHandler(this.ReadWorksheetCellNameOrVariable));
            this.cellReferenceTokenHandlers.Add(new TextExpressionTokenHandler(this.ReadCellNameOrVariable));
            this.cellReferenceTokenHandlers.Add(new TextExpressionTokenHandler(this.ReadImportedVariable));

            this.arrayTokenHandlers.Add(new ExpressionTokenHandler(this.ReadNumberToken));
            this.arrayTokenHandlers.Add(new ExpressionTokenHandler(this.ReadSignedNumberToken));
            this.arrayTokenHandlers.Add(new ExpressionTokenHandler(this.ReadErrorToken));
            this.arrayTokenHandlers.Add(new ExpressionTokenHandler(this.ReadArrayTextToken));

            this.numberParseHandlers.Add(new NumberParserHandler(this.spreadsheetCultureInfo.TryParseNumber));
            this.numberParseHandlers.Add(new NumberParserHandler(this.spreadsheetCultureInfo.TryParseDouble));
            this.numberParseHandlers.Add(new NumberParserHandler(this.spreadsheetCultureInfo.TryParseScientific));
        }

        public ParseResult Read(out ExpressionToken resultToken)
        {
            resultToken = null;

            if (this.tokenReader.EndOfFile)
            {
                return ParseResult.Unsuccessful;
            }

            foreach (var tokenHandler in this.tokenHandlers)
            {
                ParseResult result = tokenHandler(out resultToken);
                if (result != ParseResult.Unsuccessful)
                {
                    return result;
                }
            }

            if (!this.tokenReader.EndOfFile && resultToken == null)
            {
                return ParseResult.Error;
            }

            return ParseResult.Unsuccessful;
        }

        private char ReadAppendAndPeekNextChar(char nextChar, StringBuilder builder)
        {
            this.tokenReader.Read();
            builder.Append(nextChar);
            return (char)this.tokenReader.Peek();
        }

        private ParseResult ReadSpaceToken(out ExpressionToken token)
        {
            token = null;
            char next = (char)this.tokenReader.Peek();

            if (SpreadsheetCultureHelper.IsCharEqualTo(next, SpreadsheetCultureHelper.Space))
            {
                StringBuilder builder = new StringBuilder();

                while (!this.tokenReader.EndOfFile && SpreadsheetCultureHelper.IsCharEqualTo(next, SpreadsheetCultureHelper.Space))
                {
                    builder.Append(next);
                    this.tokenReader.Read();
                    next = (char)this.tokenReader.Peek();
                }

                token = new ExpressionToken(ExpressionTokenType.Space, builder.ToString());

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ReadNumberToken(out ExpressionToken token)
        {
            token = null;
            double numberResult;

            ParseResult result = this.TryReadNumber(out numberResult);

            if (result == ParseResult.Successful)
            {
                string tokenStringValue = numberResult.ToString(this.spreadsheetCultureInfo.CultureInfo);
                token = new ExpressionToken(numberResult, tokenStringValue);
            }

            return result;
        }

        private ParseResult ReadOneCharToken(out ExpressionToken token)
        {
            token = null;

            char nextChar = (char)this.tokenReader.Peek();

            if (stringToOneCharExpressionToken.TryGetValue(nextChar.ToString(), out token))
            {
                this.tokenReader.Read();

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ReadListSeparatorToken(out ExpressionToken token)
        {
            token = null;

            char nextChar = (char)this.tokenReader.Peek();

            if (SpreadsheetCultureHelper.IsCharEqualTo(nextChar, this.spreadsheetCultureInfo.ListSeparator))
            {
                token = new ExpressionToken(ExpressionTokenType.ListSeparator, this.spreadsheetCultureInfo.ListSeparator);
                this.tokenReader.Read();

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ReadArithmeticOperatorToken(out ExpressionToken token)
        {
            token = null;

            char nextChar = (char)this.tokenReader.Peek();

            if (ExpressionTokenizer.stringToArithmeticOperatorExpressionToken.TryGetValue(nextChar.ToString(), out token))
            {
                this.tokenReader.Read();

                return this.IsValidBinaryOperator() ? ParseResult.Successful : ParseResult.Error;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ReadComparisonOperatorToken(out ExpressionToken token)
        {
            token = null;

            char nextChar = (char)this.tokenReader.Peek();

            if (ExpressionTokenizer.IsComparisonOperatorFirstChar(nextChar))
            {
                StringBuilder builder = new StringBuilder();
                nextChar = this.ReadAppendAndPeekNextChar(nextChar, builder);

                if (ExpressionTokenizer.IsComparisonOperatorSecondChar(builder[0], nextChar))
                {
                    builder.Append(nextChar);
                    this.tokenReader.Read();
                }

                if (ExpressionTokenizer.stringToComparisonOperatorExpressionToken.TryGetValue(builder.ToString(), out token))
                {
                    if (this.IsValidBinaryOperator())
                    {
                        return ParseResult.Successful;
                    }
                }

                return ParseResult.Error;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ReadErrorToken(out ExpressionToken token)
        {
            token = null;

            char nextChar = (char)this.tokenReader.Peek();

            if (SpreadsheetCultureHelper.IsCharEqualTo(nextChar, SpreadsheetCultureHelper.Hash))
            {
                StringBuilder builder = new StringBuilder();

                nextChar = this.ReadAppendAndPeekNextChar(nextChar, builder);

                bool isValidErrorChar = true;
                while (isValidErrorChar)
                {
                    nextChar = this.ReadAppendAndPeekNextChar(nextChar, builder);
                    isValidErrorChar = ExpressionTokenizer.IsValidErrorChar(nextChar, builder.ToString());
                }

                return (ExpressionTokenizer.stringToErrorExpressionToken.TryGetValue(builder.ToString().ToUpper(), out token)) ? ParseResult.Successful : ParseResult.Error;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ReadTextToken(out ExpressionToken token)
        {
            token = null;

            string escapedString;
            ParseResult result = this.TryReadEscapedValue(this.spreadsheetCultureInfo.TextQualifier, out escapedString);

            if (result == ParseResult.Successful)
            {
                token = new ExpressionToken(ExpressionTokenType.Text, escapedString);
            }

            return result;
        }

        private ParseResult ReadArrayTextToken(out ExpressionToken token)
        {
            token = null;

            string escapedString;

            ParseResult result = this.TryReadEscapedValue(this.spreadsheetCultureInfo.TextQualifier, out escapedString);

            if (result == ParseResult.Successful)
            {
                string tokenValue = TextHelper.DecodeValue(escapedString, this.spreadsheetCultureInfo.TextQualifier);
                token = new ExpressionToken(ExpressionTokenType.Text, tokenValue);
            }

            return result;
        }

        private ParseResult ReadArrayToken(out ExpressionToken token)
        {
            token = null;

            char nextChar = (char)this.tokenReader.Peek();

            if (SpreadsheetCultureHelper.IsCharEqualTo(nextChar, SpreadsheetCultureHelper.ArrayLeftBracket))
            {
                StringBuilder builder = new StringBuilder();
                builder.Append(nextChar);
                this.tokenReader.Read();
                nextChar = (char)this.tokenReader.Peek();

                List<List<ExpressionToken>> arrayTokens = new List<List<ExpressionToken>>();
                arrayTokens.Add(new List<ExpressionToken>());
                bool saveInNewRow = false;

                while (!this.tokenReader.EndOfFile && !SpreadsheetCultureHelper.IsCharEqualTo(nextChar, SpreadsheetCultureHelper.ArrayRightBracket))
                {
                    builder.Append(ReadArrayWhitespace());
                    ExpressionToken arrayToken;
                    if (this.ReadArrayValue(out arrayToken))
                    {
                        return ParseResult.Error;
                    }

                    if (arrayToken.TokenType == ExpressionTokenType.Text)
                    {
                        builder.Append(TextHelper.GetEscapedTextValue(arrayToken.Value, SpreadsheetCultureHelper.DoubleQuote));
                    }
                    else
                    {
                        builder.Append(arrayToken.Value);
                    }

                    builder.Append(ReadArrayWhitespace());

                    nextChar = (char)this.tokenReader.Peek();

                    if (!this.IsValidArrayChar(nextChar))
                    {
                        return ParseResult.Error;
                    }

                    builder.Append(nextChar);
                    this.tokenReader.Read();

                    if (saveInNewRow)
                    {
                        arrayTokens.Add(new List<ExpressionToken>());
                    }

                    arrayTokens[arrayTokens.Count - 1].Add(arrayToken);

                    saveInNewRow = SpreadsheetCultureHelper.IsCharEqualTo(nextChar, this.spreadsheetCultureInfo.ArrayRowSeparator);
                }

                ExpressionToken[,] expressionTokenArray;

                if (ExpressionTokenizer.CreateArrayToken(arrayTokens, out expressionTokenArray))
                {
                    return ParseResult.Error;
                }

                token = new ArrayExpressionToken(expressionTokenArray, builder.ToString());

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private string ReadArrayWhitespace()
        {
            StringBuilder builder = new StringBuilder();

            char nextChar = (char)this.tokenReader.Peek();
            while (char.IsWhiteSpace(nextChar))
            {
                builder.Append(nextChar);
                this.tokenReader.Read();
                nextChar = (char)this.tokenReader.Peek();
            }

            return builder.ToString();
        }

        private ParseResult ReadEscapedCellReferenceToken(out ExpressionToken token)
        {
            token = null;

            string escapedString;

            ParseResult result = this.TryReadEscapedValue(SpreadsheetCultureHelper.SingleQuote, out escapedString);

            if (result == ParseResult.Successful)
            {
                StringBuilder builder = new StringBuilder(escapedString);
                char nextChar = (char)this.tokenReader.Peek();
                bool refErrorFollows = false;

                while (nextChar.IsValidFunctionOrCellReferenceNamePart() || refErrorFollows)
                {
                    nextChar = this.ReadAppendAndPeekNextChar(nextChar, builder);

                    refErrorFollows = SpreadsheetCultureHelper.IsCharEqualTo(builder[builder.Length - 1], SpreadsheetCultureHelper.ExclamationMark) &&
                                      SpreadsheetCultureHelper.IsCharEqualTo(nextChar, SpreadsheetCultureHelper.Hash);
                }

                if (this.ReadCellReferenceToken(builder.ToString(), out token) == ParseResult.Error || token == null)
                {
                    return ParseResult.Error;
                }
            }

            return result;
        }

        private ParseResult ReadFunctionCellReferenceOrVariableToken(out ExpressionToken token)
        {
            token = null;
            char nextChar = (char)this.tokenReader.Peek();

            if (ExpressionTokenizer.IsValidFunctionChar(nextChar) ||
                ExpressionTokenizer.IsValidNonEscapedCellReferenceChar(nextChar) ||
                this.IsValidVariableChar(nextChar))
            {
                StringBuilder builder = new StringBuilder();
                bool isCharValid = true;

                while (!this.tokenReader.EndOfFile && isCharValid)
                {
                    nextChar = this.ReadAppendAndPeekNextChar(nextChar, builder);

                    isCharValid = this.IsValidFunctionCellReferenceOrVariableChar(nextChar);
                }

                string value = builder.ToString();

                if (string.IsNullOrEmpty(value))
                {
                    return ParseResult.Error;
                }

                foreach (var textTokenHandler in this.textTokenHandlers)
                {
                    ParseResult result = textTokenHandler(value, out token);
                    if (result != ParseResult.Unsuccessful)
                    {
                        return result;
                    }
                }

                if (token == null)
                {
                    return ParseResult.Error;
                }
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ReadLiteralToken(string value, out ExpressionToken token)
        {
            token = null;
            ParseResult result = ParseResult.Unsuccessful;

            if (SpreadsheetCultureHelper.IsLiteral(value, SpreadsheetCultureHelper.TrueLiteral))
            {
                token = ExpressionToken.True;
                result = ParseResult.Successful;
            }
            else if (SpreadsheetCultureHelper.IsLiteral(value, SpreadsheetCultureHelper.FalseLiteral))
            {
                token = ExpressionToken.False;
                result = ParseResult.Successful;
            }

            return result;
        }

        private ParseResult ReadFunctionToken(string value, out ExpressionToken token)
        {
            token = null;
            char nextChar = (char)this.tokenReader.Peek();

            if (SpreadsheetCultureHelper.IsCharEqualTo(nextChar, SpreadsheetCultureHelper.LeftParenthesis) && IsValidFunctionName(value))
            {
                token = new ExpressionToken(ExpressionTokenType.Function, value);
                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ReadCellReferenceToken(string value, out ExpressionToken token)
        {
            token = null;

            foreach (var handler in cellReferenceTokenHandlers)
            {
                ParseResult result = handler(value, out token);
                if (result != ParseResult.Unsuccessful)
                {
                    return result;
                }
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ReadWorksheetRefErrorCellReference(string value, out ExpressionToken token)
        {
            token = null;

            if (value.EndsWith(SpreadsheetCultureHelper.ReferenceErrorValue, StringComparison.CurrentCultureIgnoreCase))
            {
                string cellName = SpreadsheetCultureHelper.ReferenceErrorValue;

                string worksheetNameAndExclamationMark = value.Substring(0, value.Length - cellName.Length);

                if (string.IsNullOrEmpty(worksheetNameAndExclamationMark) ||
                    !worksheetNameAndExclamationMark.EndsWith(SpreadsheetCultureHelper.ExclamationMark))
                {
                    return ParseResult.Error;
                }

                string worksheetName = worksheetNameAndExclamationMark.Substring(0, worksheetNameAndExclamationMark.Length - 1);

                if (string.IsNullOrEmpty(worksheetName) ||
                    !SheetCollection.IsValidSheetName(TextHelper.DecodeWorksheetName(worksheetName)) ||
                    !ExpressionTokenizer.IsWorksheetNameEscapedCorrectly(worksheetName))
                {
                    return ParseResult.Error;
                }

                token = new CellReferenceExpressionToken(worksheetName, cellName);

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ReadWorksheetCellNameOrVariable(string value, out ExpressionToken token)
        {
            token = null;

            if (value.Contains(SpreadsheetCultureHelper.ExclamationMark))
            {
                int explamationMarkIndex = value.LastIndexOf(SpreadsheetCultureHelper.ExclamationMark);
                string cellOrVariableName = explamationMarkIndex == value.Length - 1 ? string.Empty : value.Substring(explamationMarkIndex + 1);
                string worksheetName = explamationMarkIndex == 0 ? string.Empty : value.Substring(0, explamationMarkIndex);

                if (string.IsNullOrEmpty(worksheetName) || string.IsNullOrEmpty(cellOrVariableName))
                {
                    return ParseResult.Error;
                }

                string decodedWorksheetName = TextHelper.DecodeWorksheetName(worksheetName);
                if (SheetCollection.IsValidSheetName(decodedWorksheetName) &&
                    ExpressionTokenizer.IsWorksheetNameEscapedCorrectly(worksheetName))
                {
                    if (NameConverter.IsValidA1CellName(cellOrVariableName))
                    {
                        token = new CellReferenceExpressionToken(worksheetName, cellOrVariableName);

                        return ParseResult.Successful;
                    }
                    else if (DefinedName.IsNameValid(cellOrVariableName))
                    {
                        token = new DefinedNameExpressionToken(worksheetName, cellOrVariableName);

                        return ParseResult.Successful;
                    }
                }
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ReadCellNameOrVariable(string value, out ExpressionToken token)
        {
            token = null;
            ParseResult result = ParseResult.Unsuccessful;

            if (NameConverter.IsValidA1CellName(value))
            {
                token = new CellReferenceExpressionToken(string.Empty, value);
                result = ParseResult.Successful;
            }
            else if (DefinedName.IsNameValid(value))
            {
                token = new DefinedNameExpressionToken(string.Empty, value);
                result = ParseResult.Successful;
            }

            return result;
        }

        private ParseResult ReadImportedVariable(string value, out ExpressionToken token)
        {
            token = null;

            if (value.StartsWith(SpreadsheetCultureHelper.LeftSquareBracket) &&
                value.Contains(SpreadsheetCultureHelper.RightSquareBracket))
            {
                if (value.StartsWith(globalNamesImportString))
                {
                    string sanitizedValue = value.Substring(globalNamesImportString.Length);
                    token = new DefinedNameExpressionToken("Book1", sanitizedValue);

                    return ParseResult.Successful;
                }
                else
                {
                    int index = value.IndexOf(SpreadsheetCultureHelper.RightSquareBracket) + 1;
                    if (index < value.Length)
                    {
                        string sanitizedValue = value.Substring(index);
                        return ReadWorksheetCellNameOrVariable(sanitizedValue, out token);
                    }
                }
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult ReadSignedNumberToken(out ExpressionToken token)
        {
            token = null;
            char nextChar = (char)this.tokenReader.Peek();

            if (SpreadsheetCultureHelper.IsCharEqualTo(nextChar, SpreadsheetCultureHelper.UnaryMinus, SpreadsheetCultureHelper.UnaryPlus))
            {
                int sign = SpreadsheetCultureHelper.IsCharEqualTo(nextChar, SpreadsheetCultureHelper.UnaryMinus) ? -1 : 1;
                this.tokenReader.Read();

                double numberResult;
                ParseResult result = this.TryReadNumber(out numberResult);

                if (result == ParseResult.Successful)
                {
                    numberResult *= sign;
                    string valueAsString = numberResult.ToString(this.spreadsheetCultureInfo.CultureInfo);
                    token = new ExpressionToken(numberResult, valueAsString);
                }

                return result;
            }

            return ParseResult.Unsuccessful;
        }

        private ParseResult TryReadNumber(out double numberResult)
        {
            numberResult = 0;
            char nextChar = (char)this.tokenReader.Peek();

            if (char.IsDigit(nextChar) || this.spreadsheetCultureInfo.IsDecimalSeparator(nextChar))
            {
                StringBuilder builder = new StringBuilder();

                while (char.IsDigit(nextChar) || this.spreadsheetCultureInfo.IsDecimalSeparator(nextChar))
                {
                    nextChar = this.ReadAppendAndPeekNextChar(nextChar, builder);
                }

                string exponent = this.ReadExponent();
                builder.Append(exponent);

                if (!this.TryParseNumber(builder.ToString(), out numberResult))
                {
                    return ParseResult.Error;
                }

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private string ReadExponent()
        {
            string result = string.Empty;
            char nextChar = (char)this.tokenReader.Peek();

            if (SpreadsheetCultureHelper.IsExponent(nextChar))
            {
                StringBuilder builder = new StringBuilder();
                nextChar = this.ReadAppendAndPeekNextChar(nextChar, builder);

                if (SpreadsheetCultureHelper.IsCharEqualTo(nextChar, SpreadsheetCultureHelper.Plus, SpreadsheetCultureHelper.Minus))
                {
                    nextChar = this.ReadAppendAndPeekNextChar(nextChar, builder);
                }

                while (char.IsDigit(nextChar))
                {
                    nextChar = this.ReadAppendAndPeekNextChar(nextChar, builder);
                }

                result = builder.ToString();
            }

            return result;
        }

        private bool TryParseNumber(string input, out double number)
        {
            number = 0;

            foreach (var handler in numberParseHandlers)
            {
                bool parsed = handler(input, out number);
                if (parsed)
                {
                    return true;
                }
            }

            return false;
        }

        private bool IsValidBinaryOperator()
        {
            if (this.tokenReader.EndOfFile)
            {
                return false;
            }

            char nextChar = (char)this.tokenReader.Peek();
            return !SpreadsheetCultureHelper.IsCharEqualTo(nextChar, spreadsheetCultureInfo.ListSeparator,
                                                                   SpreadsheetCultureHelper.RightParenthesis,
                                                                   SpreadsheetCultureHelper.Multiplication,
                                                                   SpreadsheetCultureHelper.Division,
                                                                   SpreadsheetCultureHelper.Power,
                                                                   SpreadsheetCultureHelper.Ampersand,
                                                                   SpreadsheetCultureHelper.Colon);
        }

        public static bool IsComparisonOperatorFirstChar(char firstChar)
        {
            return SpreadsheetCultureHelper.IsCharEqualTo(firstChar,
                   SpreadsheetCultureHelper.LessThan,
                   SpreadsheetCultureHelper.GreaterThan,
                   SpreadsheetCultureHelper.Equal);
        }

        private static bool IsComparisonOperatorSecondChar(char firstChar, char secondChar)
        {
            return (firstChar != SpreadsheetCultureHelper.Equal[0] && secondChar == SpreadsheetCultureHelper.Equal[0]) ||
                   (firstChar == SpreadsheetCultureHelper.LessThan[0] && secondChar == SpreadsheetCultureHelper.GreaterThan[0]);
        }

        private ParseResult TryReadEscapedValue(string escapeSymbol, out string escapedValue)
        {
            escapedValue = string.Empty;
            char nextChar = (char)this.tokenReader.Peek();

            if (SpreadsheetCultureHelper.IsCharEqualTo(nextChar, escapeSymbol))
            {
                StringBuilder builder = new StringBuilder();
                this.tokenReader.Read();
                builder.Append(nextChar);

                bool isQuotesNumberEven = false;
                bool isPreviousCharQuote = true;
                nextChar = (char)this.tokenReader.Peek();

                while (this.IsValidEscapedValueChar(nextChar, isQuotesNumberEven, isPreviousCharQuote, escapeSymbol))
                {
                    this.tokenReader.Read();
                    builder.Append(nextChar);

                    isPreviousCharQuote = SpreadsheetCultureHelper.IsCharEqualTo(nextChar, escapeSymbol);
                    if (isPreviousCharQuote)
                    {
                        isQuotesNumberEven = !isQuotesNumberEven;
                    }

                    nextChar = (char)this.tokenReader.Peek();
                }

                escapedValue = builder.ToString();

                if (!isQuotesNumberEven)
                {
                    return ParseResult.Error;
                }

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private bool IsValidEscapedValueChar(char character, bool isQuotesNumberEven, bool isPreviousCharQuote, string escapeSymbol)
        {
            bool isValid = true;
            bool isQuote = SpreadsheetCultureHelper.IsCharEqualTo(character, escapeSymbol);

            if (this.tokenReader.EndOfFile)
            {
                isValid = false;
            }
            else if (!isQuote)
            {
                if (isQuotesNumberEven)
                {
                    isValid = false;
                }
            }
            else if (isQuote && isQuotesNumberEven)
            {
                if (!isPreviousCharQuote)
                {
                    isValid = false;
                }
            }

            return isValid;
        }

        private static bool IsValueEscapedCorrectly(string value, char escapeSymbol)
        {
            bool isValid = false;
            bool expectQuote = false;

            if (value[0] == escapeSymbol && value[value.Length - 1] == escapeSymbol)
            {
                isValid = true;
                string valueContents = value.Substring(1, value.Length - 2);

                for (int i = 0; i < valueContents.Length; i++)
                {
                    bool isQuote = valueContents[i] == escapeSymbol;

                    if (!isQuote && expectQuote)
                    {
                        isValid = false;
                        break;
                    }

                    if (isQuote)
                    {
                        expectQuote = !expectQuote;
                    }
                }
            }

            return isValid;
        }

        private static bool IsValidErrorChar(char character, string previousSymbols)
        {
            if (character == '/')
            {
                return string.Equals(previousSymbols, "#DIV", StringComparison.CurrentCultureIgnoreCase) ||
                       string.Equals(previousSymbols, "#N", StringComparison.CurrentCultureIgnoreCase);
            }

            return char.IsLetter(character) || character == '!' || character == '?' || character == '0';
        }

        private bool IsValidArrayChar(char character)
        {
            return SpreadsheetCultureHelper.IsCharEqualTo(character, this.spreadsheetCultureInfo.ArrayListSeparator,
                        this.spreadsheetCultureInfo.ArrayRowSeparator,
                        SpreadsheetCultureHelper.ArrayRightBracket);
        }

        private bool ReadArrayValue(out ExpressionToken token)
        {
            token = null;
            ExpressionToken arrayToken = null;

            foreach (var handler in arrayTokenHandlers)
            {
                ParseResult result = handler(out arrayToken);
                if (result == ParseResult.Error)
                {
                    return true;
                }

                if (result == ParseResult.Successful)
                {
                    break;
                }
            }

            token = arrayToken;

            return token == null;
        }

        private static bool CreateArrayToken(List<List<ExpressionToken>> arrayTokens, out ExpressionToken[,] array)
        {
            array = null;

            if (arrayTokens.Count < 1 || arrayTokens[0].Count < 1)
            {
                return true;
            }

            ExpressionToken[,] arrayResult = new ExpressionToken[arrayTokens.Count, arrayTokens[0].Count];

            for (int i = 0; i < arrayTokens.Count; i++)
            {
                if (arrayTokens[i].Count != arrayTokens[0].Count)
                {
                    return true;
                }

                for (int j = 0; j < arrayTokens[i].Count; j++)
                {
                    arrayResult[i, j] = arrayTokens[i][j];
                }
            }

            array = arrayResult;

            return false;
        }

        private static bool IsWorksheetNameEscapedCorrectly(string worksheetName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(worksheetName, "worksheetName");

            if (!char.IsLetter(worksheetName[0]) && worksheetName[0] != '_' && worksheetName[0] != '\'')
            {
                return false;
            }

            for (int i = 0; i < worksheetName.Length; i++)
            {
                if (!char.IsLetterOrDigit(worksheetName[i]) && worksheetName[i] != '_' && worksheetName[i] != '\'' && worksheetName[i] != '.')
                {
                    return ExpressionTokenizer.IsValueEscapedCorrectly(worksheetName, '\'');
                }
            }

            return true;
        }

        private static bool IsValidFunctionChar(char nextChar)
        {
            return char.IsLetterOrDigit(nextChar) ||
                SpreadsheetCultureHelper.IsCharEqualTo(nextChar, SpreadsheetCultureHelper.Period, SpreadsheetCultureHelper.Underscore);
        }

        private static bool IsValidFunctionName(string function)
        {
            if (function.Length == 0 || char.IsDigit(function[0]))
            {
                return false;
            }

            for (int i = 0; i < function.Length; i++)
            {
                if (!IsValidFunctionChar(function[i]))
                {
                    return false;
                }
            }

            return true;
        }

        private static bool IsValidNonEscapedCellReferenceChar(char nextChar)
        {
            return char.IsLetterOrDigit(nextChar) ||
                SpreadsheetCultureHelper.IsCharEqualTo(nextChar, SpreadsheetCultureHelper.Underscore,
                SpreadsheetCultureHelper.Hash,
                SpreadsheetCultureHelper.ExclamationMark,
                "$");
        }

        private bool IsValidFunctionCellReferenceOrVariableChar(char nextChar)
        {
            if (SpreadsheetCultureHelper.IsCharEqualTo(nextChar, SpreadsheetCultureHelper.LeftParenthesis))
            {
                return false;
            }

            if (ExpressionTokenizer.IsValidFunctionChar(nextChar) || ExpressionTokenizer.IsValidNonEscapedCellReferenceChar(nextChar))
            {
                return true;
            }

            return this.IsValidVariableChar(nextChar);
        }

        private bool IsValidVariableChar(char nextChar)
        {
            return !ExpressionTokenizer.IsComparisonOperatorFirstChar(nextChar) &&
                   !ExpressionTokenizer.stringToArithmeticOperatorExpressionToken.ContainsKey(nextChar.ToString()) &&
                   !SpreadsheetCultureHelper.IsCharEqualTo(nextChar, this.spreadsheetCultureInfo.TextQualifier,
                                                                this.spreadsheetCultureInfo.NumberDecimalSeparator,
                                                                SpreadsheetCultureHelper.ArrayLeftBracket,
                                                                SpreadsheetCultureHelper.ArrayRightBracket,
                                                                SpreadsheetCultureHelper.LeftParenthesis,
                                                                SpreadsheetCultureHelper.RightParenthesis,
                                                                SpreadsheetCultureHelper.SingleQuote,
                                                                SpreadsheetCultureHelper.Hash,
                                                                SpreadsheetCultureHelper.ExclamationMark,
                                                                SpreadsheetCultureHelper.Space,
                                                                SpreadsheetCultureHelper.VerticalBar,
                                                                SpreadsheetCultureHelper.Percent,
                                                                "`", "~", ";", "@", ",");
        }

        #endregion
    }
}