﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Expressions.Parser;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using System.Globalization;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions
{
    /// <summary>
    /// Abstract class providing basic methods and properties for creating and using expressions in RadSpreadsheet.
    /// </summary>
    public abstract class RadExpression
    {
        #region Fields

        private ExpressionState state;
        private RadExpression value;

        #endregion


        #region Properties

        // TODO:
        //public virtual bool IsVolatile
        //{
        //    get
        //    {
        //        return false;
        //    }
        //}

        #endregion


        #region Methods

        /// <summary>
        /// Gets the value.
        /// </summary>
        /// <returns>The value as RadExpression.</returns>
        public RadExpression GetValue()
        {
            if (!this.Evaluate())
            {
                return ErrorExpressions.CyclicReference;
            }

            return this.value;
        }

        internal bool Evaluate()
        {
            if (this.state == ExpressionState.Invalid)
            {
                this.state = ExpressionState.Computing;
                this.value = this.GetValueOverride();
                this.state = ExpressionState.Valid;
            }
            else if (this.state == ExpressionState.Computing)
            {
                return false;
            }

            return true;
        }

        /// <summary>
        /// Gets the value as string.
        /// </summary>
        /// <returns>The value as string.</returns>
        public string GetValueAsString()
        {
            return this.GetValueAsString(FormatHelper.DefaultSpreadsheetCulture);
        }

        /// <summary>
        /// Gets the value as string by specified CultureInfo instance.
        /// </summary>
        /// <param name="cultureInfo">The culture info.</param>
        /// <returns>The value as string.</returns>
        public string GetValueAsString(CultureInfo cultureInfo)
        {
            return this.GetValueAsString(new SpreadsheetCultureHelper(cultureInfo));
        }

        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current
        /// <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
            return this.ToString(FormatHelper.DefaultSpreadsheetCulture);
        }

        /// <summary>
        /// Converts the expression to String by specified CultureInfo instance.
        /// </summary>
        /// <param name="cultureInfo">The culture info.</param>
        /// <returns>The expression as string.</returns>
        public string ToString(CultureInfo cultureInfo)
        {
            return this.ToString(new SpreadsheetCultureHelper(cultureInfo));
        }
        /// <summary>
        /// Overridable method for getting expression value.
        /// </summary>
        /// <returns>The value as RadExpression.</returns>
        protected abstract RadExpression GetValueOverride();

        /// <summary>
        /// Invalidates the value.
        /// </summary>
        protected internal void InvalidateValue()
        {
            if (this.state != ExpressionState.Invalid)
            {
                this.state = ExpressionState.Invalid;
                this.OnValueInvalidated();
            }
        }

        /// <summary>
        /// Attaches to children event.
        /// </summary>
        /// <param name="childExpressions">The child expressions.</param>
        protected void AttachToChildrenEvent(IEnumerable<RadExpression> childExpressions)
        {
            foreach (var childExpression in childExpressions)
            {
                this.AttachToChildEvent(childExpression);
            }
        }

        /// <summary>
        /// Attaches to child event.
        /// </summary>
        /// <param name="childExpression">The child expression.</param>
        protected void AttachToChildEvent(RadExpression childExpression)
        {
            Guard.ThrowExceptionIfNull(childExpression, "childExpression");

            if (!(childExpression is ConstantExpression))
            {
                childExpression.ValueInvalidated += this.ChildExpression_ValueInvalidated;
            }
        }

        internal static ParseResult TryParse(string value, Worksheet worksheet, CellIndex cellIndex, out RadExpression result)
        {
            return TryParse(value, worksheet, cellIndex.RowIndex, cellIndex.ColumnIndex, out result);
        }

        internal static ParseResult TryParse(string value, Worksheet worksheet, int rowIndex, int columnIndex, out RadExpression result)
        {
            InputStringCollection stringExpressionCollection;
            return TryParse(value, worksheet, rowIndex, columnIndex, out result, out stringExpressionCollection);
        }

        internal static ParseResult TryParse(string value, Worksheet worksheet, int rowIndex, int columnIndex, out RadExpression result,
            out InputStringCollection stringExpressionCollection, bool appendWorksheetName = false)
        {
            return TryParse(value, worksheet, rowIndex, columnIndex, FormatHelper.DefaultSpreadsheetCulture, out result,
                out stringExpressionCollection, appendWorksheetName);
        }

        internal static ParseResult TryParse(string value, Worksheet worksheet, int rowIndex, int columnIndex, SpreadsheetCultureHelper spreadsheetCultureInfo,
            out RadExpression result, out InputStringCollection stringExpressionCollection, bool appendWorksheetName = false)
        {
            Guard.ThrowExceptionIfNullOrEmpty(value, "value");
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);
            Guard.ThrowExceptionIfNull(spreadsheetCultureInfo, "spreadsheetCultureInfo");

            stringExpressionCollection = null;
            result = null;

            if (!RadExpression.StartsLikeExpression(value))
            {
                return ParseResult.Unsuccessful;
            }

            string clearedFormulaValue = SpreadsheetCultureHelper.ClearFormulaValue(value);
            if (!string.IsNullOrEmpty(clearedFormulaValue))
            {
                return ExpressionParser.ExpressionParserInstance.Parse(clearedFormulaValue, worksheet, rowIndex, columnIndex,
                    spreadsheetCultureInfo, out stringExpressionCollection, out result, appendWorksheetName);
            }

            return ParseResult.Error;
        }

        internal static bool StartsLikeExpression(string value)
        {
            if (string.IsNullOrEmpty(value))
            {
                return false;
            }

            return value.Length > 1 &&
                   SpreadsheetCultureHelper.IsCharEqualTo(value[0],
                   SpreadsheetCultureHelper.FunctionStart,
                   SpreadsheetCultureHelper.Plus,
                   SpreadsheetCultureHelper.Minus);
        }

        internal virtual string GetValueAsString(SpreadsheetCultureHelper cultureInfo)
        {
            return this.GetValue().ToString(cultureInfo);
        }

        internal abstract string ToString(SpreadsheetCultureHelper cultureInfo);

        internal abstract RadExpression CloneAndTranslate(ExpressionCloneAndTranslateContext context);

        internal abstract void Translate(ExpressionTranslateContext context);

        private void ChildExpression_ValueInvalidated(object sender, EventArgs e)
        {
            this.InvalidateValue();
        }

        internal virtual bool SimilarEquals(object obj)
        {
            RadExpression other = obj as RadExpression;

            if (other == null)
            {
                return false;
            }

            if (this.value == null && other.value == null)
            {
                return true;
            }

            if ((this.value != null && other.value == null) ||
                this.value == null && other.value != null)
            {
                return false;
            }

            return this.value.Equals(other.value);
        }

        #endregion


        #region Events

        /// <summary>
        /// Occurs when ExpressionState becomes Invalid.
        /// </summary>
        public event EventHandler ValueInvalidated;

        /// <summary>
        /// Called when the value is invalidated.
        /// </summary>
        protected virtual void OnValueInvalidated()
        {
            if (this.ValueInvalidated != null)
            {
                this.ValueInvalidated(this, EventArgs.Empty);
            }
        }

        #endregion
    }
}