﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Core;
using System.Text;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions
{
    /// <summary>
    /// Provides methods and properties for using SpreadsheetNameExpression.
    /// </summary>
    public class SpreadsheetNameExpression : RadExpression
    {
        #region Fields

        private readonly ExpressionQualifierInfo qualifierInfo;
        private readonly int rowIndex;
        private readonly int columnIndex;

        private string spreadsheetName;
        private RadExpression expressionCache;

        #endregion


        #region Properties

        internal bool IsNameReferringWorbook
        {
            get
            {
                return this.qualifierInfo.HasQualifier && this.qualifierInfo.IsQualifierReferringWorkbook;
            }
        }

        internal bool IsNameReferringWorksheet
        {
            get
            {
                return this.qualifierInfo.HasQualifier && !this.qualifierInfo.IsQualifierReferringWorkbook;
            }
        }

        #endregion


        #region Constructors

        internal SpreadsheetNameExpression(ExpressionQualifierInfo qualifierInfo, CellIndex cellIndex, string spreadsheetName)
            : this(qualifierInfo, cellIndex.RowIndex, cellIndex.ColumnIndex, spreadsheetName)
        {
        }

        internal SpreadsheetNameExpression(ExpressionQualifierInfo qualifierInfo, int rowIndex, int columnIndex, string spreadsheetName)
        {
            Guard.ThrowExceptionIfNull(qualifierInfo, "qualifierInfo");
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);
            Guard.ThrowExceptionIfNullOrEmpty(spreadsheetName, "spreadsheetName");

            this.qualifierInfo = qualifierInfo;
            this.spreadsheetName = spreadsheetName;
            this.rowIndex = rowIndex;
            this.columnIndex = columnIndex;

            this.AttachWeekEventNamesUpdated();
        }

        #endregion


        #region Methods

        private void AttachWeekEventNamesUpdated()
        {
            var nameCollectionsChangedEventListener =
            new RadWeakEventListener<SpreadsheetNameExpression, NameManager, EventArgs>(this, this.qualifierInfo.Workbook.NameManager);

            this.qualifierInfo.Workbook.NameManager.Changed += nameCollectionsChangedEventListener.OnEvent;

            nameCollectionsChangedEventListener.OnEventAction = InnerValueInvalidatedWeakEventAction;

            nameCollectionsChangedEventListener = null;
        }

        private static void InnerValueInvalidatedWeakEventAction(SpreadsheetNameExpression instance, object source, EventArgs e)
        {
            instance.InnerValueInvalidated(instance, e);
        }

        private void InnerValueInvalidated(object sender, EventArgs e)
        {
            base.InvalidateValue();
        }

        /// <summary>
        /// Overridable method for getting expression value.
        /// </summary>
        /// <returns>The value as RadExpression.</returns>
        protected override RadExpression GetValueOverride()
        {
            DefinedName variableCache = this.qualifierInfo.Workbook.NameManager.FindSpreadsheetName(this.qualifierInfo.Qualifier, this.spreadsheetName, this.qualifierInfo.CurrentWorksheet, this.qualifierInfo.Workbook) as DefinedName;

            if (variableCache != null)
            {
                RadExpression newValue = variableCache.FormulaCellValue.Value;
                RadExpression newValueResult = newValue.GetValue();

                if (newValueResult != ErrorExpressions.NameError && newValueResult != ErrorExpressions.CyclicReference)
                {
                    ExpressionCloneAndTranslateContext context = new ExpressionCloneAndTranslateContext(this.qualifierInfo.CurrentWorksheet, this.rowIndex, this.columnIndex, false);
                    RadExpression translatedExpression = newValue.CloneAndTranslate(context);

                    if (!translatedExpression.Equals(expressionCache))
                    {
                        if (this.expressionCache != null)
                        {
                            this.expressionCache.ValueInvalidated -= this.InnerValueInvalidated;
                        }

                        this.expressionCache = translatedExpression;

                        this.expressionCache.ValueInvalidated += this.InnerValueInvalidated;
                    }

                    return expressionCache.GetValueAsConstantOrCellReference();
                }
            }

            return ErrorExpressions.NameError;
        }

        internal override RadExpression CloneAndTranslate(ExpressionCloneAndTranslateContext context)
        {
            Worksheet toWorksheet = context.Worksheet;
            ExpressionQualifierInfo translatedQualifierInfo = this.qualifierInfo;
            if (this.qualifierInfo.CurrentWorksheet != toWorksheet)
            {
                translatedQualifierInfo = new ExpressionQualifierInfo(context.Worksheet.Workbook, toWorksheet, this.qualifierInfo.Qualifier);
            }

            RadExpression translatedExpression = new SpreadsheetNameExpression(translatedQualifierInfo, context.RowIndex, context.ColumnIndex, this.spreadsheetName);

            if (context.NewFormatCollection != null && context.OldFormatCollection != null)
            {
                for (int i = 0; i < context.OldFormatCollection.Count; i++)
                {
                    DefinedNameInputString input = context.OldFormatCollection[i] as DefinedNameInputString;
                    if (input != null && object.ReferenceEquals(this, input.Expression))
                    {
                        ((DefinedNameInputString)context.NewFormatCollection[i]).Expression = (SpreadsheetNameExpression)translatedExpression;
                    }
                }
            }

            return translatedExpression;
        }

        internal override void Translate(ExpressionTranslateContext context)
        {
            if (context.IsWorksheetRenamed && this.IsNameReferringWorksheet)
            {
                this.qualifierInfo.UpdateQualifier(TextHelper.EncodeWorksheetName(context.RenamedWorksheet.Name));
            }
            else if (context.IsNameRenamed)
            {
                if (this.spreadsheetName.Equals(context.OldSpreadsheetName))
                {
                    this.spreadsheetName = context.NewSpreadsheetName;
                }
            }
            else if (context.IsWorkbookRenamed && this.qualifierInfo.IsQualifierReferringWorkbook)
            {
                this.qualifierInfo.UpdateQualifier(TextHelper.EncodeWorksheetName(context.RenamedWorkbook.Name));
            }

            this.InvalidateValue();
        }

        internal override string ToString(SpreadsheetCultureHelper cultureInfo)
        {
            StringBuilder builder = new StringBuilder();

            if (this.qualifierInfo.HasQualifier)
            {
                builder.Append(this.qualifierInfo.EscapedQualifier);
                builder.Append(SpreadsheetCultureHelper.ExclamationMark);
            }

            builder.Append(this.spreadsheetName);
            // TODO: 
            return builder.ToString().ToString(cultureInfo.CultureInfo);
        }

        internal override bool SimilarEquals(object obj)
        {
            return this.Equals(obj);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            SpreadsheetNameExpression other = obj as SpreadsheetNameExpression;

            return other != null &&
                TelerikHelper.EqualsOfT(this.rowIndex, other.rowIndex) &&
                TelerikHelper.EqualsOfT(this.columnIndex, other.columnIndex) &&
                TelerikHelper.EqualsOfT(this.qualifierInfo, other.qualifierInfo);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                    this.rowIndex.GetHashCode(),
                    this.columnIndex.GetHashCode(),
                    this.qualifierInfo.GetHashCodeOrZero());
        }

        #endregion
    }
}
