﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Expressions.Functions;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Expressions
{
    /// <summary>
    /// Abstract class providing methods and properties for using UnaryOperatorExpression.
    /// </summary>
    public abstract class UnaryOperatorExpression : OperatorExpression
    {
		#region Fields

		private readonly RadExpression operand;

		#endregion


		#region Properties

        /// <summary>
        /// Gets the operand.
        /// </summary>
        /// <value>The operand.</value>
        /// <returns>The operand as RadExpression.</returns>
        public RadExpression Operand
        {
			get
			{
				return this.operand;
			}
		}

        /// <summary>
        /// Gets the argument conversion rules.
        /// </summary>
        /// <value>The argument conversion rules.</value>
        /// <returns>The argument conversion rules of the operand operation.</returns>
        public virtual ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return ArgumentConversionRules.NumberFunctionConversion;
            }
        }

		#endregion


		#region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="UnaryOperatorExpression" /> class.
        /// </summary>
        /// <param name="operand">The operand.</param>
        protected UnaryOperatorExpression(RadExpression operand)
        {
            Guard.ThrowExceptionIfNull(operand, "operand");

			this.operand = operand;

            base.AttachToChildEvent(this.operand);
		}

		#endregion


        #region Methods

        /// <summary>
        /// Overridable method for getting expression value.
        /// </summary>
        /// <returns>The value as RadExpression.</returns>
        protected sealed override RadExpression GetValueOverride()
        {
            List<object> convertedArguments = new List<object>();
           
            ErrorExpression errorExpression = FunctionHelper.TryConvertArgument(this.operand, this.ArgumentConversionRules, ArgumentType.Number, convertedArguments);

            if (errorExpression != null || convertedArguments.Count < 1 || !(convertedArguments[0] is double))
            {
                return errorExpression;
            }

            return GetValueOverride((double)convertedArguments[0]);
        }

        /// <summary>
        /// Gets the value of the expression.
        /// </summary>
        /// <param name="operand">The operand as double.</param>
        /// <returns>Value as RadExpression.</returns>
        protected abstract RadExpression GetValueOverride(double operand);

        internal override string ToString(SpreadsheetCultureHelper cultureInfo)
        {
            return this.OperatorInfo.Symbol + this.GetOperandAsString(this.Operand, cultureInfo);
        }

        internal override void Translate(ExpressionTranslateContext context)
        {
            this.Operand.Translate(context);
        }

        #endregion
    }
}