﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.Contexts
{
    /// <summary>
    /// Describes a cell.
    /// </summary>
    public class CellInfo
    {
        #region Fields

        private int rowIndex;
        private int columnIndex;
        private CellIndex cellIndex;
        private bool isCellIndexInvalidated;
        private int styleIndex;

        #endregion


        #region Properties

        /// <summary>
        /// Gets or sets the index of the row.
        /// </summary>
        /// <value>The index of the row.</value>
        public int RowIndex
        {
            get
            {
                return this.rowIndex;
            }
            set
            {
                this.rowIndex = value;
                this.isCellIndexInvalidated = true;
                this.IsCellIndexKnown = true;
            }
        }

        /// <summary>
        /// Gets or sets the index of the column.
        /// </summary>
        /// <value>The index of the column.</value>
        public int ColumnIndex
        {
            get
            {
                return this.columnIndex;
            }
            set
            {
                this.columnIndex = value;
                this.isCellIndexInvalidated = true;
                this.IsCellIndexKnown = true;
            }
        }

        /// <summary>
        /// Gets or sets the index of the cell.
        /// </summary>
        /// <value>The index of the cell.</value>
        public CellIndex CellIndex
        {
            get
            {
                if (this.isCellIndexInvalidated || this.cellIndex == null)
                {
                    this.cellIndex = new CellIndex(this.rowIndex, this.columnIndex);
                }

                return this.cellIndex;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "value");
                this.cellIndex = value;
                this.rowIndex = this.cellIndex.RowIndex;
                this.columnIndex = this.cellIndex.ColumnIndex;
                this.IsCellIndexKnown = true;
            }
        }

        /// <summary>
        /// Gets or sets the cell value.
        /// </summary>
        /// <value>The cell value.</value>
        public ICellValue CellValue { get; set; }

        /// <summary>
        /// Gets or sets the index of the style.
        /// </summary>
        /// <value>The index of the style.</value>
        public int StyleIndex
        {
            get
            {
                return this.styleIndex;
            }
            set
            {
                this.styleIndex = value;
            }
        }

        internal bool IsCellIndexKnown { get; set; }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="CellInfo" /> class.
        /// </summary>
        public CellInfo()
        {
            this.isCellIndexInvalidated = true;
            this.IsCellIndexKnown = false;
        }

        internal CellInfo(int rowIndex, int columnIndex, Cells cells)
            : this()
        {
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);
            Guard.ThrowExceptionIfNull(cells, "cells");

            long longIndex = CellsPropertyBag.ConvertCellIndexToLong(rowIndex, columnIndex);
            this.CellValue = cells.PropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.ValueProperty).GetValue(longIndex);
            this.rowIndex = rowIndex;
            this.columnIndex = columnIndex;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CellInfo" /> class.
        /// </summary>
        /// <param name="cellIndex">Index of the cell.</param>
        /// <param name="cellSelection">The cell selection.</param>
        public CellInfo(CellIndex cellIndex, CellSelection cellSelection)
            : this()
        {
            Guard.ThrowExceptionIfNull(cellSelection, "cellSelection");

            this.CellIndex = cellIndex;
            this.CellValue = cellSelection.GetValue().Value;
        }

        #endregion
    }
}