﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.Contexts
{
    /// <summary>
    /// Describes cell property data.
    /// </summary>
    internal class CellPropertyDataInfo
    {
#if WPF
        private readonly SortedList<int, Core.DataStructures.Range> rowUsedRange;
        private readonly SortedList<int, Core.DataStructures.Range> columnUsedRange;
#else
        private readonly Dictionary<int, Range> rowUsedRange;
        private readonly Dictionary<int, Range> columnUsedRange;
#endif

        private Core.DataStructures.Range maxRowUsedRange = null;
        private Core.DataStructures.Range maxColumnUsedRange = null;
        
        /// <summary>
        /// Initializes a new instance of the <see cref="CellPropertyDataInfo" /> class.
        /// </summary>
        public CellPropertyDataInfo()
        {
#if WPF
            this.rowUsedRange = new SortedList<int, Core.DataStructures.Range>();
            this.columnUsedRange = new SortedList<int, Core.DataStructures.Range>();
#else
            this.rowUsedRange = new Dictionary<int, Range>();
            this.columnUsedRange = new Dictionary<int, Range>();
#endif
        }

        /// <summary>
        /// Sets the row used range value.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="value">The value.</param>
        public void SetRowUsedRangeValue(int rowIndex, Core.DataStructures.Range value)
        {
            this.maxRowUsedRange = Core.DataStructures.Range.MaxOrNull(maxRowUsedRange, value);
            this.rowUsedRange[rowIndex] = value;
        }

        /// <summary>
        /// Sets the column used range value.
        /// </summary>
        /// <param name="columnIndex">Index of the column.</param>
        /// <param name="value">The value.</param>
        public void SetColumnUsedRangeValue(int columnIndex, Core.DataStructures.Range value)
        {
            this.maxColumnUsedRange = Core.DataStructures.Range.MaxOrNull(maxColumnUsedRange, value);
            this.columnUsedRange[columnIndex] = value;
        }

        /// <summary>
        /// Gets the row used range.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <returns>The row used range.</returns>
        public Core.DataStructures.Range GetRowUsedRange(int rowIndex)
        {
            if (this.rowUsedRange.ContainsKey(rowIndex))
            {
                return this.rowUsedRange[rowIndex];
            }

            return null;
        }

        /// <summary>
        /// Gets the column used range.
        /// </summary>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The column used range.</returns>
        public Core.DataStructures.Range GetColumnUsedRange(int columnIndex)
        {
            if (this.columnUsedRange.ContainsKey(columnIndex))
            {
                return this.columnUsedRange[columnIndex];
            }

            return null;
        }

        /// <summary>
        /// Gets a value indicating whether the row is empty.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <returns><c>True</c> if the row is empty.</returns>
        public bool GetRowIsEmpty(int rowIndex)
        {
            return this.GetRowUsedRange(rowIndex) == null;
        }

        /// <summary>
        /// Gets a value indicating whether the column is empty.
        /// </summary>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns><c>True</c> if the column is empty.</returns>
        public bool GetColumnIsEmpty(int columnIndex)
        {
            return this.GetColumnUsedRange(columnIndex) == null;
        }

        /// <summary>
        /// Gets the used cell range.
        /// </summary>
        /// <returns>The used cell range.</returns>
        public CellRange GetUsedCellRange()
        {

            if (this.maxColumnUsedRange == null || this.maxRowUsedRange == null)
            {
                return null;
            }

            return new CellRange(this.maxColumnUsedRange.Start, this.maxRowUsedRange.Start, this.maxColumnUsedRange.End, this.maxRowUsedRange.End);
        }
    }
}