﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.Contexts
{
    /// <summary>
    /// Represents worksheet export context.
    /// </summary>
    internal class WorksheetExportContext
    {
        #region Fields

        private readonly Worksheet worksheet;
        private readonly CellPropertyDataInfo valuePropertyDataInfo;
        
        #endregion


        #region Properties

        /// <summary>
        /// Gets the worksheet.
        /// </summary>
        /// <value>The worksheet.</value>
        public Worksheet Worksheet
        {
            get
            {
                return this.worksheet;
            }
        }

        /// <summary>
        /// Gets the value property data information.
        /// </summary>
        /// <value>The value property data information.</value>
        public CellPropertyDataInfo ValuePropertyDataInfo
        {
            get
            {
                return this.valuePropertyDataInfo;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="WorksheetExportContext" /> class.
        /// </summary>
        /// <param name="worksheet">The worksheet.</param>
        public WorksheetExportContext(Worksheet worksheet)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            
            this.worksheet = worksheet;
            this.valuePropertyDataInfo = new CellPropertyDataInfo();

            this.Initialize();
        }

        #endregion


        #region Initialization

        private void Initialize()
        {
            this.CalculateCellPropertyPropertyUsedValueRanges(this.valuePropertyDataInfo, CellPropertyDefinitions.ValueProperty, true);
        }

        private void CalculateCellPropertyPropertyUsedValueRanges<T>(
            CellPropertyDataInfo propertyDataInfo, IPropertyDefinition<T> propertyDefinition, bool ignoreDefaultValues)
        {
            ICompressedList<T> compressedList = this.worksheet.Cells.PropertyBag.GetPropertyValueCollection<T>(propertyDefinition);
            CalculateCellPropertyPropertyUsedValueRanges(propertyDataInfo, compressedList, ignoreDefaultValues);
        }

        internal static void CalculateCellPropertyPropertyUsedValueRanges<T>(
            CellPropertyDataInfo propertyDataInfo, ICompressedList<T> compressedList, bool ignoreDefaultValues)
        {
            T defaultValue = compressedList.GetDefaultValue();
            foreach (var range in compressedList.GetNonDefaultRanges())
            {
                if (ignoreDefaultValues && TelerikHelper.EqualsOfT(range.Value, defaultValue))
                {
                    continue;
                }

                for (long i = range.Start; i <= range.End; i++)
                {
                    CellIndex cellIndex = CellsPropertyBag.ConvertLongToCellIndex(i);
                    propertyDataInfo.SetRowUsedRangeValue(cellIndex.RowIndex,
                        Core.DataStructures.Range.CreateOrExpand(propertyDataInfo.GetRowUsedRange(cellIndex.RowIndex), cellIndex.ColumnIndex));

                    propertyDataInfo.SetColumnUsedRangeValue(cellIndex.ColumnIndex,
                        Core.DataStructures.Range.CreateOrExpand(propertyDataInfo.GetColumnUsedRange(cellIndex.ColumnIndex), cellIndex.RowIndex));
                }
            }
        }

        #endregion


        #region Methods

        internal static CellPropertyDataInfo GetValuePropertyDataInfo(Worksheet worksheet)
        {
            WorksheetExportContext context = new WorksheetExportContext(worksheet);
            return context.ValuePropertyDataInfo;
        }

        /// <summary>
        /// Gets the non empty cells in row.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <returns>The result collection containing information about the cells.</returns>
        public virtual IEnumerable<CellInfo> GetNonEmptyCellsInRow(int rowIndex)
        {
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);

            Core.DataStructures.Range usedRange = this.ValuePropertyDataInfo.GetRowUsedRange(rowIndex);
            if (usedRange == null)
            {
                yield break;
            }

            for (int columnIndex = usedRange.Start; columnIndex <= usedRange.End; columnIndex++)
            {
                yield return new CellInfo(rowIndex, columnIndex, worksheet.Cells);
            }
        }

        #endregion
    }
}