﻿using System.Collections.Generic;
using System.IO;
using System.Text;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.TextBased.Core
{
    internal class CsvParser
    {
        #region Fields
        
        private readonly CsvTokenizer tokenizer;
        
        #endregion


        #region Constructors

        public CsvParser(TextReader reader, CsvSettings settings = null)
        {
            this.tokenizer = new CsvTokenizer(reader, settings);
        }

        #endregion


        #region Methods

        public IEnumerable<CsvRecord> Parse()
        {
            while (!tokenizer.IsAtEndOfFile)
            {
                IEnumerable<CsvToken> recordTokens = this.ReadRecordTokens();
                yield return new CsvRecord(CsvParser.ParseRecordTokens(recordTokens));
            }
        }

        private IEnumerable<CsvToken> ReadRecordTokens()
        {
            bool isEndOfLineValid = true;
            
            CsvToken token = tokenizer.Read();
            while (token != null)
            {
                if (token.TokenType == CsvTokenType.EndOfLine && isEndOfLineValid)
                {
                    break;
                }

                if (token.TokenType == CsvTokenType.Quote)
                {
                    isEndOfLineValid = !isEndOfLineValid;
                }

                yield return token;

                token = tokenizer.Read();
            }
        }

        private static IEnumerable<string> ParseRecordTokens(IEnumerable<CsvToken> recordTokens)
        {            
            bool isSpecialCharValid = true;
            List<CsvToken> value = new List<CsvToken>();
            
            foreach (CsvToken token in recordTokens)
            {
                if (token.TokenType == CsvTokenType.Quote)
                {
                    isSpecialCharValid = !isSpecialCharValid;
                }

                if (token.TokenType == CsvTokenType.Delimiter && isSpecialCharValid)
                {
                    yield return ParseValueTokens(value);
                    value.Clear();
                }
                else
                {
                    value.Add(token);
                }
            }

            yield return ParseValueTokens(value);
        }

        private static string ParseValueTokens(List<CsvToken> valueTokens)
        {
            bool isPreviousTokenQuote = false;
            int quotesCounter = 0;

            StringBuilder builder = new StringBuilder();

            foreach (CsvToken token in valueTokens)
            {
                if (token.TokenType != CsvTokenType.Quote || (quotesCounter > 1 && isPreviousTokenQuote))
                {
                    builder.Append(token.Value);
                }

                if (token.TokenType == CsvTokenType.Quote)
                {
                    quotesCounter++;
                }

                if (isPreviousTokenQuote && quotesCounter > 2)
                {
                    isPreviousTokenQuote = false;
                }
                else
                {
                    isPreviousTokenQuote = (token.TokenType == CsvTokenType.Quote);
                }
            }

            return builder.ToString();
        }

        #endregion
    }
}