﻿using System.Text;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.TextBased.Core
{
    /// <summary>
    /// Represents CSV settings.
    /// </summary>
    public class CsvSettings
    {
        #region Fields

        private char? delimiter;
        private char? quote;
        private bool? hasHeaderRow;
        private Encoding encoding;

        #endregion


        #region Properties

        /// <summary>
        /// Gets or sets the delimiter symbol.
        /// </summary>
        /// <value>The delimiter symbol.</value>
        public char Delimiter
        {
            get
            {
                if (!this.delimiter.HasValue)
                {
                    string listSeparator = FormatHelper.DefaultSpreadsheetCulture.ListSeparator;
                    this.delimiter = (listSeparator.Length == 1) ? listSeparator[0] : '\0';
                }

                return this.delimiter.Value;
            }
            set
            {
                this.delimiter = value;
            }
        }

        /// <summary>
        /// Gets or sets the quote symbol.
        /// </summary>
        /// <value>The quote symbol.</value>
        public char Quote
        {
            get
            {
                if (!this.quote.HasValue)
                {
                    this.quote = '"';
                }

                return this.quote.Value;
            }
            set
            {
                this.quote = value;
            }
        }

        /// <summary>
        /// Gets or sets the has header row.
        /// </summary>
        /// <value>The has header row.</value>
        public bool HasHeaderRow
        {
            get
            {
                if (!this.hasHeaderRow.HasValue)
                {
                    this.hasHeaderRow = false;
                }

                return this.hasHeaderRow.Value;
            }
            set
            {
                this.hasHeaderRow = value;
            }
        }

        /// <summary>
        /// Gets or sets the encoding.
        /// </summary>
        /// <value>The encoding.</value>
        public Encoding Encoding
        {
            get
            {
                if (this.encoding == null)
                {
                    // The default encoding of Excel is UTF8 with BOM
                    this.encoding = new UTF8Encoding(true);
                }

                return this.encoding;
            }
            set
            {
                this.encoding = value;
            }
        }

        #endregion


        #region Methods

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The object to compare with the current object.</param>
        /// <returns>
        /// <c>true</c> if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            CsvSettings other = obj as CsvSettings;
            if (other == null)
            {
                return false;
            }

            return TelerikHelper.EqualsOfT<char?>(this.Delimiter, other.Delimiter) &&
                TelerikHelper.EqualsOfT<char?>(this.Quote, other.Quote) &&
                TelerikHelper.EqualsOfT<bool?>(this.HasHeaderRow, other.HasHeaderRow) &&
                TelerikHelper.EqualsOfT<Encoding>(this.Encoding, other.Encoding);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(this.Delimiter.GetHashCode(),
                                                    this.Quote.GetHashCode(),
                                                    this.HasHeaderRow.GetHashCode(),
                                                    this.Encoding.GetHashCodeOrZero());
        }

        /// <summary>
        /// Implements the less operator.
        /// </summary>
        /// <param name="first">The first value.</param>
        /// <param name="second">The second value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
        public static bool operator ==(CsvSettings first, CsvSettings second)
        {
            return first.Equals(second as object);
        }

        /// <summary>
        /// Implements the less operator.
        /// </summary>
        /// <param name="first">The first value.</param>
        /// <param name="second">The second value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
        public static bool operator !=(CsvSettings first, CsvSettings second)
        {
            return !first.Equals(second as object);
        }

        #endregion
    }
}