﻿using System.IO;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
using System.Collections.Generic;
using System.Text;
using System;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.TextBased.Core
{
    internal class CsvWriter : IDisposable
    {
        #region Fields

        private readonly TextWriter writer;
        private readonly CsvSettings settings;
        private bool isHeaderRow;
        private bool alreadyDisposed;

        #endregion


        #region Constructors

        public CsvWriter(Stream stream, CsvSettings settings = null) : this(settings)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");
            this.writer = new StreamWriter(stream, this.settings.Encoding);
        }

        public CsvWriter(TextWriter writer, CsvSettings settings = null)
            : this(settings)
        {
            Guard.ThrowExceptionIfNull(writer, "writer");
            this.writer = writer;
        }

        private CsvWriter(CsvSettings settings)
        {
            this.settings = settings ?? new CsvSettings();
            this.isHeaderRow = this.settings.HasHeaderRow;
        }

        #endregion


        #region Methods

        public void WriteRecord(IEnumerable<string> values)
        {
            bool isFirstValue = true;
            int emptyValueCount = 0;

            foreach (string value in values)
            {
                if (string.IsNullOrEmpty(value))
                {
                    emptyValueCount++;
                }
                else
                {
                    emptyValueCount = isFirstValue ? emptyValueCount : emptyValueCount + 1;
                    for (int i = 0; i < emptyValueCount; i++)
                    {
                        writer.Write(this.settings.Delimiter);
                    }

                    string escapedValue = this.EscapeValue(value);
                    writer.Write(escapedValue);

                    isFirstValue = false;
                    emptyValueCount = 0;
                }
            }

            this.isHeaderRow = false;

            writer.WriteLine();
            writer.Flush();
        }

        private string EscapeValue(string value)
        {
            StringBuilder builder = new StringBuilder(value);
            bool requitesQuotes = this.isHeaderRow;

            for (int i = 0; i < value.Length; i++)
            {
                char character = value[i];
                if (character == this.settings.Delimiter || character == '\n' || character == '\r')
                {
                    requitesQuotes = true;
                }

                if (character == this.settings.Quote)
                {
                    string oneQuote = this.settings.Quote.ToString();
                    string twoQuotes = oneQuote + oneQuote;

                    builder.Replace(oneQuote, twoQuotes);
                    requitesQuotes = true;
                    break;
                }
            }

            if (requitesQuotes)
            {
                builder.Insert(0, this.settings.Quote.ToString());
                builder.Append(this.settings.Quote);
            }

            return builder.ToString();
        }

        #endregion


        #region IDisposable

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing,
        /// or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Disposes the specified clean up managed resources.
        /// </summary>
        /// <param name="cleanUpManagedResources">The clean up managed resources.</param>
        protected virtual void Dispose(bool cleanUpManagedResources)
        {
            if (this.alreadyDisposed)
            {
                return;
            }

            //Place here the code that cleans up native resource.
            if (cleanUpManagedResources)
            {
                if (this.writer != null)
                {
                    this.writer.Dispose();
                }
            }

            this.alreadyDisposed = true;
        }

        #endregion
    }
}