﻿using System.Collections.Generic;
using System.IO;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.Contexts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.TextBased.Core;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.TextBased
{
    /// <summary>
    /// Provides the base class from which the classes that represent delimited values format providers are derived. 
    /// </summary>
    public abstract class DelimitedValuesFormatProviderBase : TextBasedWorkbookFormatProviderBase
    {
        #region Properties

        /// <summary>
        /// Gets the settings.
        /// </summary>
        /// <value>The settings.</value>
        public abstract CsvSettings Settings { get; }

        /// <summary>
        /// Gets a value indicating whether can import.
        /// </summary>
        /// <value>The value indicating whether can import.</value>
        public override bool CanImport
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Gets a value indicating whether can export.
        /// </summary>
        /// <value>The value indicating whether can export.</value>
        public override bool CanExport
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the leading empty rows and columns should be trimmed.
        /// </summary>
        /// <value>The value indicating whether the leading empty rows and columns should be trimmed.</value>
        public bool TrimLeadingEmptyRowsAndColumns { get; set; }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="DelimitedValuesFormatProviderBase" /> class.
        /// </summary>
        protected DelimitedValuesFormatProviderBase()
        {
            this.TrimLeadingEmptyRowsAndColumns = true;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Imports the specified input.
        /// </summary>
        /// <param name="input">The input.</param>
        /// <returns>The imported workbook.</returns>
        protected override Workbook ImportOverride(Stream input)
        {
            Workbook workbook = new Workbook();

            UpdateScope updateScope = new UpdateScope(
                () =>
                {
                    workbook.History.IsEnabled = false;
                    workbook.SuspendLayoutUpdate();
                    workbook.SuspendPropertyChanged();
                },
                () =>
                {
                    workbook.ResumePropertyChanged();
                    workbook.ResumeLayoutUpdate();
                    workbook.History.IsEnabled = true;
                });

            using (updateScope)
            {
                Worksheet worksheet = workbook.Worksheets.Add();

                using (StreamReader reader = new StreamReader(input, this.Settings.Encoding))
                {
                    int rowIndex = 0;
                    CsvParser parser = new CsvParser(reader, this.Settings);

                    ICompressedList<ICellValue> valuesCompressedList = worksheet.Cells.PropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.ValueProperty);
                    ICompressedList<CellValueFormat> formatsCompressedList = worksheet.Cells.PropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.FormatProperty);

                    foreach (CsvRecord record in parser.Parse())
                    {
                        int columnIndex = 0;
                        foreach (string value in record.GetValues())
                        {
                            ICellValue cellValue;
                            CellValueFormat newFormat;
                            CellValueFactory.CreateIgnoreErrors(value, worksheet, rowIndex, columnIndex, CellValueFormat.GeneralFormat, out cellValue, out newFormat);

                            long longIndex = CellsPropertyBag.ConvertCellIndexToLong(rowIndex, columnIndex);

                            valuesCompressedList.SetValue(longIndex, cellValue);
                            if (newFormat != CellValueFormat.GeneralFormat)
                            {
                                formatsCompressedList.SetValue(longIndex, newFormat);
                            }

                            columnIndex++;
                        }
                        rowIndex++;
                    }
                }
            }

            return workbook;
        }

        /// <summary>
        /// Exports the specified workbook.
        /// </summary>
        /// <param name="workbook">The workbook.</param>
        /// <param name="output">The output.</param>
        protected override void ExportOverride(Workbook workbook, Stream output)
        {
            CsvWriter csvWriter = new CsvWriter(output, this.Settings);
            LicenseCheck.Validate(csvWriter);

            Worksheet worksheet = workbook.ActiveWorksheet;

            WorksheetExportContext context = new WorksheetExportContext(worksheet);

            CellRange usedCellRange = context.ValuePropertyDataInfo.GetUsedCellRange();

            if (usedCellRange == null)
            {
                return;
            }

            if (!this.TrimLeadingEmptyRowsAndColumns)
            {
                usedCellRange = new CellRange(new CellIndex(0, 0), usedCellRange.ToIndex);
            }

            for (int rowIndex = usedCellRange.FromIndex.RowIndex; rowIndex <= usedCellRange.ToIndex.RowIndex; rowIndex++)
            {
                Range rowUsedRange = context.ValuePropertyDataInfo.GetRowUsedRange(rowIndex);
                if (rowUsedRange != null)
                {
                    rowUsedRange = rowUsedRange.Expand(usedCellRange.FromIndex.ColumnIndex);
                }

                csvWriter.WriteRecord(GetRowValuesToExport(worksheet, rowUsedRange, rowIndex));
            }
        }

        private IEnumerable<string> GetRowValuesToExport(Worksheet worksheet, Range usedRange, int rowIndex)
        {
            if (usedRange != null)
            {
                for (int columnIndex = usedRange.Start; columnIndex <= usedRange.End; columnIndex++)
                {
                    CellSelection selection = worksheet.Cells[rowIndex, columnIndex];
                    ICellValue cellValue = selection.GetValue().Value;
                    CellValueFormat formatting = selection.GetFormat().Value;

                    FormulaCellValue formulaCellValue = cellValue as FormulaCellValue;
                    if (formulaCellValue != null)
                    {
                        cellValue = formulaCellValue.GetResultValueAsCellValue();
                    }

                    yield return formatting.GetFormatResult(cellValue).ExtendedVisibleInfosText;
                }
            }
        }

        #endregion
    }
}