﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders
{
    /// <summary>
    /// Provides the base class from which the classes that represent workbook format providers are derived.
    /// </summary>
    public abstract class WorkbookFormatProviderBase : IWorkbookFormatProvider
    {
        #region Properties

        /// <summary>
        /// Gets the name.
        /// </summary>
        /// <value>The name of the provider.</value>
        public abstract string Name { get; }

        /// <summary>
        /// Gets the files description.
        /// </summary>
        /// <value>The files description.</value>
        public virtual string FilesDescription
        {
            get
            {
                return SupportedExtensions.First().Trim('.').ToUpper() + " Files";
            }
        }

        /// <summary>
        /// Gets the supported extensions.
        /// </summary>
        /// <value>The supported extensions.</value>
        public abstract IEnumerable<string> SupportedExtensions { get; }

        /// <summary>
        /// Gets a value indicating whether can import.
        /// </summary>
        /// <value>The value indicating whether can import.</value>
        public abstract bool CanImport { get; }

        /// <summary>
        /// Gets a value indicating whether can export.
        /// </summary>
        /// <value>The value indicating whether can export.</value>
        public abstract bool CanExport { get; }

        #endregion


        #region Methods

        /// <summary>
        /// Imports the specified input.
        /// </summary>
        /// <param name="input">The input.</param>
        /// <returns>The imported workbook.</returns>
        public Workbook Import(Stream input)
        {
            if (!this.CanImport)
            {
                throw new InvalidOperationException("Import not supported.");
            }

            return this.ImportOverride(input);
        }

        /// <summary>
        /// Imports the specified input.
        /// </summary>
        /// <param name="input">The input.</param>
        /// <returns>The imported workbook.</returns>
        protected virtual Workbook ImportOverride(Stream input)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Exports the specified workbook.
        /// </summary>
        /// <param name="workbook">The workbook.</param>
        /// <param name="output">The output.</param>
        public void Export(Workbook workbook, Stream output)
        {
            if (!this.CanExport)
            {
                throw new InvalidOperationException("Export not supported.");
            }

            this.ExportOverride(workbook, output);
        }

        /// <summary>
        /// Exports the specified workbook.
        /// </summary>
        /// <param name="workbook">The workbook.</param>
        /// <param name="output">The output.</param>
        protected virtual void ExportOverride(Workbook workbook, Stream output)
        {
            throw new NotImplementedException();
        }

        #endregion
    }
}