﻿using System.Collections.Generic;
using System.IO;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.TextBased.Csv;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.TextBased.Txt;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders
{
    /// <summary>
    /// Represents workbook format providers manager.
    /// </summary>
    public class WorkbookFormatProvidersManager
    {
        private static readonly WorkbookFormatProvidersManager Instance;

        private readonly List<IWorkbookFormatProvider> formatProviders;

        /// <summary>
        /// Gets the application-wide registered format providers.
        /// </summary>
        /// <value>The format providers.</value>
        public static IEnumerable<IWorkbookFormatProvider> FormatProviders
        {
            get
            {
                return WorkbookFormatProvidersManager.Instance.formatProviders;
            }
        }

        static WorkbookFormatProvidersManager()
        {
            WorkbookFormatProvidersManager.Instance = new WorkbookFormatProvidersManager();
        }

        private WorkbookFormatProvidersManager()
        {
            this.formatProviders = new List<IWorkbookFormatProvider>();

            this.InitBuiltInProviders();
        }

        internal static IEnumerable<string> GetSupportedExtensions(IEnumerable<IWorkbookFormatProvider> formatProviders)
        {
            return GetSupportedExtensionsInternal(formatProviders);
        }

        private void InitBuiltInProviders()
        {
            this.RegisterFormatProviderInternal(new TxtFormatProvider());
            this.RegisterFormatProviderInternal(new CsvFormatProvider());
        }

        private static IEnumerable<string> GetSupportedExtensionsInternal(IEnumerable<IWorkbookFormatProvider> formatProviders)
        {
            return formatProviders.Reverse().SelectMany(p => p.SupportedExtensions).Distinct();
        }

        private IWorkbookFormatProvider GetProviderByNameInternal(string providerName)
        {
            return this.formatProviders.FirstOrDefault(p => p.Name == providerName);
        }

        private static IWorkbookFormatProvider GetProviderByExtensionInternal(string extension, IEnumerable<IWorkbookFormatProvider> formatProviders)
        {
            Guard.ThrowExceptionIfNullOrEmpty(extension, "extension");

            extension = extension.ToLowerInvariant();
            if (!extension.StartsWith("."))
            {
                extension = "." + extension;
            }

            return formatProviders.FirstOrDefault(fp => fp.SupportedExtensions.Contains(extension));
        }

        private void RegisterFormatProviderInternal(IWorkbookFormatProvider provider)
        {
            Guard.ThrowExceptionIfNull(provider, "provider");

            this.RemoveFormatProviderByName(provider.Name);
            this.formatProviders.Add(provider);
        }

        private void UnregisterFormatProviderInternal(IWorkbookFormatProvider provider)
        {
            Guard.ThrowExceptionIfNull(provider, "provider");

            RemoveFormatProviderByName(provider.Name);
        }

        private void RemoveFormatProviderByName(string name)
        {
            IWorkbookFormatProvider formatProviderToRemove = this.GetProviderByNameInternal(name);
            if (formatProviderToRemove != null)
            {
                this.formatProviders.Remove(formatProviderToRemove);
            }
        }

        /// <summary>
        /// Gets the supported file extensions.
        /// </summary>
        /// <returns>The supported file extensions.</returns>
        public static IEnumerable<string> GetSupportedExtensions()
        {
            return GetSupportedExtensions(FormatProviders);
        }

        /// <summary>
        /// Gets format provider by name.
        /// </summary>
        /// <param name="providerName">Name of the provider.</param>
        /// <returns>The format provider.</returns>
        public static IWorkbookFormatProvider GetProviderByName(string providerName)
        {
            return WorkbookFormatProvidersManager.Instance.GetProviderByNameInternal(providerName);
        }

        /// <summary>
        /// Gets format provider by extension.
        /// </summary>
        /// <param name="extension">The extension.</param>
        /// <returns>The format provider.</returns>
        public static IWorkbookFormatProvider GetProviderByExtension(string extension)
        {
            return GetProviderByExtensionInternal(extension, FormatProviders);
        }

        /// <summary>
        /// Registers the format provider.
        /// </summary>
        /// <param name="provider">The provider.</param>
        public static void RegisterFormatProvider(IWorkbookFormatProvider provider)
        {
            WorkbookFormatProvidersManager.Instance.RegisterFormatProviderInternal(provider);
        }

        /// <summary>
        /// Unregisters the format provider.
        /// </summary>
        /// <param name="provider">The provider.</param>
        public static void UnregisterFormatProvider(IWorkbookFormatProvider provider)
        {
            WorkbookFormatProvidersManager.Instance.UnregisterFormatProviderInternal(provider);
        }

        /// <summary>
        /// Imports a workbook from stream.
        /// </summary>
        /// <param name="extension">The file extension.</param>
        /// <param name="input">The input stream.</param>
        /// <returns>The workbook.</returns>
        public static Workbook Import(string extension, Stream input)
        {
            return Import(extension, input, FormatProviders);
        }

        /// <summary>
        /// Imports a workbook from stream.
        /// </summary>
        /// <param name="extension">The file extension.</param>
        /// <param name="input">The input stream.</param>
        /// <param name="formatProviders">The format providers to be used.</param>
        /// <returns>
        /// The workbook.
        /// </returns>
        /// <exception cref="UnsupportedFileFormatException"></exception>
        public static Workbook Import(string extension, Stream input, IEnumerable<IWorkbookFormatProvider> formatProviders)
        {
            IWorkbookFormatProvider formatProvider = GetProviderByExtensionInternal(extension, formatProviders);
            if (formatProvider == null)
            {
                throw new UnsupportedFileFormatException(extension);
            }

            return formatProvider.Import(input);
        }

        /// <summary>
        /// Exports the specified workbook.
        /// </summary>
        /// <param name="workbook">The workbook.</param>
        /// <param name="extension">The file extension.</param>
        /// <param name="output">The output stream.</param>
        public static void Export(Workbook workbook, string extension, Stream output)
        {
            Export(workbook, extension, output, FormatProviders);
        }

        /// <summary>
        /// Exports the specified workbook.
        /// </summary>
        /// <param name="workbook">The workbook.</param>
        /// <param name="extension">The file extension.</param>
        /// <param name="output">The output stream.</param>
        /// <param name="formatProviders">The format providers to be used.</param>
        /// <exception cref="UnsupportedFileFormatException"></exception>
        public static void Export(Workbook workbook, string extension, Stream output, IEnumerable<IWorkbookFormatProvider> formatProviders)
        {
            IWorkbookFormatProvider formatProvider = GetProviderByExtensionInternal(extension, formatProviders);
            if (formatProvider == null)
            {
                throw new UnsupportedFileFormatException(extension);
            }

            formatProvider.Export(workbook, output);
        }
    }
}