﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings;

namespace Telerik.Windows.Documents.Spreadsheet.Formatting
{
    internal static class CellValueParser
    {
        #region Nested Types

        private delegate ParseResult TryParseCellValueDelegate(
            string value, Worksheet worksheet, int rowIndex, int columnIndex, CellValueFormat currentFormatValue,
            out ICellValue cellValue, out CellValueFormat newFormatValue);

        #endregion


        #region Constants

        private static readonly TryParseCellValueDelegate[] tryParseCellValueHandlers =
            new TryParseCellValueDelegate[]
            {
                TryParseEmptyCellValue,
                TryParseTextFormatCellValue,
                TryParseNumberCellValue,
                TryParseBooleanCellValue,
                TryParseFormulaCellValue,
                TryParseErrorCellValue,
                TryParseTextCellValue
            };

        #endregion


        #region Methods

        public static ParseResult Parse(string value, Worksheet worksheet, int rowIndex, int columnIndex, CellValueFormat currentFormatValue,
            out ICellValue cellValue, out CellValueFormat newFormatValue)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);
            Guard.ThrowExceptionIfNull(currentFormatValue, "currentFormatValue");

            foreach (TryParseCellValueDelegate tryParseCellValueHandler in tryParseCellValueHandlers)
            {
                ParseResult parseResult = tryParseCellValueHandler(value, worksheet, rowIndex, columnIndex, currentFormatValue,
                    out cellValue, out newFormatValue);

                if (parseResult != ParseResult.Unsuccessful)
                {
                    return parseResult;
                }
            }

            throw new LocalizableException("Cannot parse value to ICellValue.",
                new InvalidOperationException("Cannot parse value to ICellValue."),
                "Spreadsheet_ErrorExpressions_CannotParseValue");
        }

        #endregion


        #region Parse ICellValue Handlers

        private static ParseResult TryParseEmptyCellValue(string value, Worksheet worksheet, int rowIndex, int columnIndex, CellValueFormat currentFormatValue,
            out ICellValue cellValue, out CellValueFormat newFormatValue)
        {
            cellValue = null;
            newFormatValue = null;

            if (string.IsNullOrEmpty(value))
            {
                cellValue = EmptyCellValue.EmptyValue;
                newFormatValue = currentFormatValue;
                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private static ParseResult TryParseTextFormatCellValue(string value, Worksheet worksheet, int rowIndex, int columnIndex,
            CellValueFormat currentFormatValue, out ICellValue cellValue, out CellValueFormat newFormatValue)
        {
            cellValue = null;
            newFormatValue = null;

            if (currentFormatValue.FormatStringInfo.FormatType == FormatStringType.Text)
            {
                return TryParseTextCellValue(value, worksheet, rowIndex, columnIndex, currentFormatValue, out cellValue, out newFormatValue);
            }

            return ParseResult.Unsuccessful;
        }

        private static ParseResult TryParseNumberCellValue(string value, Worksheet worksheet, int rowIndex, int columnIndex, CellValueFormat currentFormatValue,
            out ICellValue cellValue, out CellValueFormat newFormatValue)
        {
            cellValue = null;
            newFormatValue = CellValueFormat.GeneralFormat;

            FormatStringType? forceParsingAsSpecificType = null;
            if (currentFormatValue.FormatStringInfo.FormatType == FormatStringType.Number && !string.IsNullOrEmpty(currentFormatValue.FormatString))
            {
                forceParsingAsSpecificType = FormatStringType.Number;
            }

            string newStringFormat;
            double doubleValue;
            if (NumberValueParser.TryParse(value, forceParsingAsSpecificType, null, out newStringFormat, out doubleValue))
            {
                cellValue = new NumberCellValue(doubleValue);
                if (!string.IsNullOrEmpty(newStringFormat))
                {
                    newFormatValue = new CellValueFormat(newStringFormat);
                }
                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private static ParseResult TryParseBooleanCellValue(string value, Worksheet worksheet, int rowIndex, int columnIndex, CellValueFormat currentFormatValue,
            out ICellValue cellValue, out CellValueFormat newFormatValue)
        {
            cellValue = null;
            newFormatValue = null;

            bool boolean;
            if (bool.TryParse(value.ToLower(), out boolean))
            {
                cellValue = new BooleanCellValue(boolean);
                newFormatValue = currentFormatValue;

                return ParseResult.Successful;
            }

            return ParseResult.Unsuccessful;
        }

        private static ParseResult TryParseFormulaCellValue(string value, Worksheet worksheet, int rowIndex, int columnIndex, CellValueFormat currentFormatValue,
            out ICellValue cellValue, out CellValueFormat newFormatValue)
        {
            cellValue = null;
            newFormatValue = null;

            RadExpression expression;
            InputStringCollection stringExpressionCollection;

            ParseResult parseResult = RadExpression.TryParse(value, worksheet, rowIndex, columnIndex, out expression, out stringExpressionCollection);

            if (parseResult == ParseResult.Successful)
            {
                FormulaCellValue formulaCellValue = new FormulaCellValue(stringExpressionCollection, expression, rowIndex, columnIndex);

                FunctionExpression functionExpression = formulaCellValue.Value as FunctionExpression;
                newFormatValue = currentFormatValue;
                if (functionExpression != null && functionExpression.IsValid)
                {
                    newFormatValue = functionExpression.Function.FunctionInfo.Format;
                }

                cellValue = formulaCellValue;
            }

            return parseResult;
        }

        private static ParseResult TryParseErrorCellValue(string value, Worksheet worksheet, int rowIndex, int columnIndex, CellValueFormat currentFormatValue,
            out ICellValue cellValue, out CellValueFormat newFormatValue)
        {
            cellValue = null;
            newFormatValue = null;

            if (SpreadsheetCultureHelper.IsCharEqualTo(value[0], SpreadsheetCultureHelper.Hash))
            {
                ErrorCellValue errorCellValue;
                if (ErrorCellValue.TryGetErrorCellValue(value, out errorCellValue))
                {
                    cellValue = errorCellValue;
                    newFormatValue = currentFormatValue;
                    return ParseResult.Successful;
                }
            }

            return ParseResult.Unsuccessful;
        }

        private static ParseResult TryParseTextCellValue(string value, Worksheet worksheet, int rowIndex, int columnIndex, CellValueFormat currentFormatValue,
            out ICellValue cellValue, out CellValueFormat newFormatValue)
        {
            cellValue = new TextCellValue(value);
            newFormatValue = currentFormatValue;
            return ParseResult.Successful;
        }

        #endregion
    }
}