﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;

namespace Telerik.Windows.Documents.Spreadsheet.Formatting
{
    internal static class DateTimeParser
    {
        #region Constants

        private static readonly int MaxYearValue = 9999;

        private static readonly string YearGroupName = "year";
        private static readonly string MonthGoupName = "month";
        private static readonly string DayGroupName = "day";
        private static readonly string HourGroupName = "hour";
        private static readonly string MinuteGroupName = "minute";
        private static readonly string SecondGroupName = "second";
        private static readonly string MillisecondGroupName = "millisecond";
        private static readonly string AmPmGroupName = "ampm";

        private static readonly Regex FullDateRegex = new Regex(string.Format(@"^(?<{0}>\d+)/(?<{1}>\d+)/(?<{2}>\d+)\s+(?<{3}>\d+):(?<{4}>\d+):(?<{5}>\d+).(?<{6}>\d+)$",
                YearGroupName, MonthGoupName, DayGroupName, HourGroupName, MinuteGroupName, SecondGroupName, MillisecondGroupName));
        private static readonly Regex DateRegex = new Regex(string.Format(@"^(?<{0}>\d{{1,2}})[-\s/](?<{1}>(\d{{1,2}}|\d{{4}}))$", MonthGoupName, DayGroupName));
        private static readonly Regex DateRegex1 = new Regex(string.Format(@"^(?<{0}>\d+)[-\s/](?<{1}>\D{{3,9}})$", DayGroupName, MonthGoupName));
        private static readonly Regex DateRegex2 = new Regex(string.Format(@"^(?<{0}>\D{{3,9}})[-\s/](?<{1}>\d+)$", MonthGoupName, DayGroupName));
        private static readonly Regex DateRegex3 = new Regex(string.Format(@"^(?<{0}>\d+)-(?<{1}>\D{{3,9}})-(?<{2}>\d{{1,2}}|\d{{4}})$", DayGroupName, MonthGoupName, YearGroupName));
        //private static readonly Regex DateRegex4 = new Regex(string.Format(@"^(?<{1}>\d{{1,2}})/(?<{0}>\d+)/(?<{2}>\d{{1,2}}|\d{{4}})$", DayGroupName, MonthGoupName, YearGroupName));
        private static readonly Regex TimeRegex = new Regex(string.Format(@"^(?<{0}>\d+):(?<{1}>\d+)\s*$", HourGroupName, MinuteGroupName));
        private static readonly Regex TimeRegex1 = new Regex(string.Format(@"^(?<{0}>\d+):(?<{1}>\d+):(?<{2}>\d+)$\s*", HourGroupName, MinuteGroupName, SecondGroupName));
        private static readonly Regex TimeRegex2 = new Regex(string.Format(@"^(?<{0}>\d+):(?<{1}>\d+)\s+[AP]M\s*$", HourGroupName, MinuteGroupName));
        private static readonly Regex TimeRegex3 = new Regex(string.Format(@"^(?<{0}>\d+):(?<{1}>\d+):(?<{2}>\d+)\s+[AP]M\s*$", HourGroupName, MinuteGroupName, SecondGroupName));
        private static readonly Regex TimeMillisecondsRegex = new Regex(string.Format(@"^(?<{0}>\d+):(?<{1}>\d+).(?<{2}>\d+)\s*$", MinuteGroupName, SecondGroupName, MillisecondGroupName));
        private static readonly Regex TimeMillisecondsAmPmRegex = new Regex(string.Format(@"^(?<{0}>\d+):(?<{1}>\d+).(?<{2}>\d+)\s+(?<{3}>[AP]M)?\s*$", MinuteGroupName, SecondGroupName, MillisecondGroupName, AmPmGroupName));
        private static readonly Regex TimeMillisecondsRegex1 = new Regex(string.Format(@"^(?<{0}>\d+):(?<{1}>\d+):(?<{2}>\d+).(?<{3}>\d+)$", HourGroupName, MinuteGroupName, SecondGroupName, MillisecondGroupName));
        private static readonly Regex TimeMillisecondsAmPmRegex1 = new Regex(string.Format(@"^(?<{0}>\d+):(?<{1}>\d+):(?<{2}>\d+).(?<{3}>\d+)\s+(?<{4}>[AP]M)?\s*$",
                HourGroupName, MinuteGroupName, SecondGroupName, MillisecondGroupName, AmPmGroupName));

        // TODO: All of these should not be hardcoded
        private static readonly string DefaultDayMonthYearFormat1 = "d-mmm-yy";
        private static readonly string DefaultDayMonthFormat = "d-mmm"; 
        private static readonly string DefaultMonthYearFormat = "mmm-yy"; 
        private static readonly string DefaultHourMinuteFormat = "h:mm"; 
        private static readonly string DefaultHourMinuteFormat1 = "h:mm AM/PM"; 
        private static readonly string DefaultHourMinuteSecondsFormat = "h:mm:ss"; 
        private static readonly string DefaultHourMinuteSecondsFormat1 = "h:mm:ss AM/PM"; 
        private static readonly string DefaultMinuteSecondsMillisecondsFormat = "mm:ss.0"; 
        
        #endregion


        #region Properties

        private static string DefaultDayMonthYearFormat
        {
            get
            {
                return FormatHelper.ShortDatePattern.ToLower();
            }
        }

        private static string DefaultDateTimeFormat
        {
            get
            {
                return string.Format("{0} {1}", DefaultDayMonthYearFormat, DefaultHourMinuteFormat);
            }
        }

        #endregion


        #region Methods

        public static bool TryParse(string value, out DateTime result, out string format)
        {
            format = FormatHelper.GeneralFormatString;
            DateTime startDate = FormatHelper.StartDate.AddDays(-1);

            bool canParseAsDate =
               TryParseMonthYearDateTime(value, out result) ||
               TryParseMonthNameYearDateTime(value, out result);

            bool systemParsed = false;
            if (!canParseAsDate)
            {
                systemParsed = FormatHelper.DefaultSpreadsheetCulture.TryParseDateTime(value, out result);
                //DateTime.TryParse(value, FormatHelper.GetCurrentCulture(), DateTimeStyles.NoCurrentDateDefault, out result);
                canParseAsDate = systemParsed;
            }

            if (canParseAsDate && result.Date == DateTime.MinValue)
            {
                result = new DateTime(startDate.Year, startDate.Month, startDate.Day, 
                    result.Hour, result.Minute, result.Second, result.Millisecond);

                canParseAsDate &= result > startDate;
                if (canParseAsDate)
                {
                    result = result.RoundMilliseconds(1);
                    format = GetDateTimeFormat(result, value, false);
                }
            }

            TimeSpan timeSpan;
            bool canParseAsTimeSpan = FormatHelper.DefaultSpreadsheetCulture.TryParseTimeSpan(value, out timeSpan);
            if (!canParseAsDate && !canParseAsTimeSpan)
            {
                canParseAsTimeSpan = 
                    TryParseMinuteSecondMillisecondTimeSpan(value, out timeSpan) ||
                    TryParseMinuteSecondMillisecondAmPmTimeSpan(value, out timeSpan);
            }

            if (!canParseAsDate || (canParseAsDate && canParseAsTimeSpan))
            {
                if (canParseAsTimeSpan)
                {
                    result = startDate;
                    result = result.Add(timeSpan);

                    canParseAsTimeSpan &= result >= startDate;
                    if (canParseAsTimeSpan)
                    {
                        result = result.RoundMilliseconds(1);
                        format = GetDateTimeFormat(result, value, false);
                    }

                    return canParseAsTimeSpan;
                }
            }

            canParseAsDate &= (result >= startDate);

            if (canParseAsDate)
            {
                result = result.RoundMilliseconds(1);
                format = GetDateTimeFormat(result, value, systemParsed);
            }

            return canParseAsDate;
        }

        // NOTE: The case we have "3:44.567 AM" or "3:44:55.678 AM" which is not parsed by .NET parsing engine.
        private static bool TryParseMinuteSecondMillisecondAmPmTimeSpan(string value, out TimeSpan parsedTimeSpan)
        {
            parsedTimeSpan = TimeSpan.MinValue;

            if (!TimeMillisecondsAmPmRegex.IsMatch(value))
            {
                return false;
            }
            
            Match match = TimeMillisecondsAmPmRegex.Match(value);

            string minute = match.Groups[MinuteGroupName].Value;
            string second = match.Groups[SecondGroupName].Value;
            string millisecond = match.Groups[MillisecondGroupName].Value;
            string ampm = match.Groups[AmPmGroupName].Value;

            string newValue = string.Format("{0}:{1}.{2}", minute, second, millisecond);

            bool success = TryParseMinuteSecondMillisecondTimeSpan(newValue, out parsedTimeSpan);
            if (success && ampm == "PM")
            {
                parsedTimeSpan = parsedTimeSpan.Add(TimeSpan.FromHours(12));
            }

            return success;
        }

        // NOTE: Special case not supported by .NET DateTime and TimeSpan parsing engine: 
        // 4:55.6789 where 44 is minutes, 55 is seconds and 6789 is milliseconds.
        private static bool TryParseMinuteSecondMillisecondTimeSpan(string value, out TimeSpan parsedTimeSpan)
        {
            parsedTimeSpan = TimeSpan.Zero;
            if (!TimeMillisecondsRegex.IsMatch(value))
            {
                return false;
            }

            Match match = TimeMillisecondsRegex.Match(value);

            string minutes = match.Groups[MinuteGroupName].Value;
            string seconds = match.Groups[SecondGroupName].Value;
            string milliseconds = match.Groups[MillisecondGroupName].Value;

            string formatString = 
                string.Concat(Enumerable.Repeat('m', minutes.Length)) + "\\:" +
                string.Concat(Enumerable.Repeat('s', seconds.Length)) + "\\." +
                string.Concat(Enumerable.Repeat('f', milliseconds.Length));

            return TimeSpan.TryParseExact(value, formatString, null, out parsedTimeSpan);
        }

        private static bool TryParseMonthYearDateTime(string value, out DateTime parsedDateTime)
        {
            string newFormat;
            return TryParseMonthYearDateTime(value, out parsedDateTime, out newFormat);
        }

        // NOTE: Special case like 12/34 where 12 is a month and 34 is a year. 
        // This translates to Dec-1934, a case that is not parsed by .NET's DateTiem parsing engine.
        private static bool TryParseMonthYearDateTime(string value, out DateTime parsedDateTime, out string newFormat)
        {
            parsedDateTime = DateTime.MinValue;
            DateTime startDate = FormatHelper.StartDate;//.AddDays(1);
            newFormat = FormatHelper.GeneralFormatString;

            if (!DateRegex.IsMatch(value))
            {
                return false;
            }

            Match match = DateRegex.Match(value);

            int month = int.Parse(match.Groups[MonthGoupName].Value);
            int dayYear = int.Parse(match.Groups[DayGroupName].Value);

            if (month > 12 || month < 1 || (dayYear > 99 && dayYear < startDate.Year) || dayYear > MaxYearValue)
            {
                return false;
            }

            bool dayYearIsDay = dayYear > 0 && dayYear <= GetNumberOfDaysByMonth(month);
            if (dayYearIsDay && match.Groups[DayGroupName].Value.Length > 2)
            {
                return false;
            }

            if (dayYear >= startDate.Year)
            {
                parsedDateTime = new DateTime(dayYear, month, startDate.Day);
                newFormat = DefaultMonthYearFormat;
            }
            else if (dayYearIsDay)
            {
                parsedDateTime = new DateTime(DateTime.Now.Year, month, dayYear);
                newFormat = DefaultDayMonthFormat;
            }
            else
            {
                parsedDateTime = new DateTime(startDate.Year + dayYear, month, 1);
                newFormat = DefaultMonthYearFormat;
            }

            return true;
        }

        private static bool TryParseMonthNameYearDateTime(string value, out DateTime parsedDateTime)
        {
            string newFormat;
            return TryParseMonthNameYearDateTime(value, out parsedDateTime, out newFormat);
        }

        private static bool TryParseMonthNameYearDateTime(string value, out DateTime parsedDateTime, out string newFormat)
        {
            parsedDateTime = DateTime.MinValue;
            DateTime startDate = FormatHelper.StartDate;
            newFormat = FormatHelper.GeneralFormatString;
            
            if (!DateRegex2.IsMatch(value))
            {
                return false;
            }

            Match match = DateRegex2.Match(value);

            string monthName = match.Groups[MonthGoupName].Value;
            int month = GetMonthNumberByName(monthName);
            if (month == -1)
            {
                return false;
            }

            int dayYear = int.Parse(match.Groups[DayGroupName].Value);
            if (dayYear > 9999 || dayYear < 0)
            {
                return false;
            }

            bool dayYearIsDay = dayYear <= GetNumberOfDaysByMonth(month);
            if (dayYearIsDay)
            {
                parsedDateTime = new DateTime(DateTime.Now.Year, month, dayYear);
                newFormat = DefaultDayMonthFormat;
            }
            else
            {
                if (dayYear.ToString().Length == 2)
                {
                    parsedDateTime = new DateTime(startDate.Year + dayYear, month, 1);
                }
                else
                {
                    parsedDateTime = new DateTime(dayYear, month, 1);
                }
                newFormat = DefaultMonthYearFormat;
            }

            return true;
        }

        private static string GetDateTimeFormat(DateTime? dateTime, string value, bool systemParsed)
        {
            string newFormat = FormatHelper.GeneralFormatString;
            
            if (TryParseUsingRegex(value, out newFormat))
            {
                return newFormat;
            }

            DateTime startDate = FormatHelper.StartDate;

            bool hasYear = dateTime.Value.Year != 1 || systemParsed; 
            bool hasMonth = dateTime.Value.Month != startDate.Month || systemParsed;
            bool hasDay = dateTime.Value.Day != startDate.Day || systemParsed;
            bool hasHour = dateTime.Value.Hour != startDate.Hour;
            bool hasMinute = dateTime.Value.Minute != startDate.Minute;
            bool hasSecond = dateTime.Value.Second != startDate.Second;
            bool hasMilliSeconds = dateTime.Value.Millisecond != startDate.Millisecond;

            if ((hasYear || hasMonth || hasDay) && (hasHour || hasMinute || hasSecond))
            {
                return DefaultDateTimeFormat;
            }
            else if (hasYear && hasMonth && hasDay)
            {
                return DefaultDayMonthYearFormat;
            }
            else if (hasYear && hasMonth)
            {
                return DefaultMonthYearFormat;
            }
            else if (hasMonth && hasDay)
            {
                return DefaultDayMonthFormat;
            }
            else if ((hasHour || hasMinute || hasSecond) && hasMilliSeconds)
            {
                return DefaultMinuteSecondsMillisecondsFormat;
            }
            else if (hasHour && hasMinute && hasSecond)
            {
                return DefaultHourMinuteSecondsFormat;
            }
            else if (hasHour && hasMinute)
            {
                return DefaultHourMinuteFormat;
            }

            return DefaultDateTimeFormat;
        }

        private static bool TryParseUsingRegex(string value, out string newFormat)
        {
            newFormat = FormatHelper.GeneralFormatString;

            DateTime dateTime;
            if (TryParseMonthYearDateTime(value, out dateTime, out newFormat) ||
                TryParseMonthNameYearDateTime(value, out dateTime, out newFormat))
            {
                return true;
            }

            foreach (Tuple<Regex, string> regexTuple in GetPredefinedDateTimeRegexes())
            {
                if (TryParseUsingPattern(value, regexTuple.Item1, regexTuple.Item2, out newFormat))
                {
                    return true;
                }
            }

            return false;
        }

        // NOTE: The tuple values represent: regex, newFormat
        private static List<Tuple<Regex, string>> GetPredefinedDateTimeRegexes()
        {
            List<Tuple<Regex, string>> result = new List<Tuple<Regex, string>>();

            result.Add(new Tuple<Regex, string>(FullDateRegex, DefaultMinuteSecondsMillisecondsFormat));
            result.Add(new Tuple<Regex, string>(TimeRegex, DefaultHourMinuteFormat));
            result.Add(new Tuple<Regex, string>(TimeRegex1, DefaultHourMinuteSecondsFormat));
            result.Add(new Tuple<Regex, string>(TimeMillisecondsRegex, DefaultMinuteSecondsMillisecondsFormat));
            result.Add(new Tuple<Regex, string>(TimeMillisecondsRegex1, DefaultMinuteSecondsMillisecondsFormat));
            result.Add(new Tuple<Regex, string>(TimeRegex2, DefaultHourMinuteFormat1));
            result.Add(new Tuple<Regex, string>(TimeRegex3, DefaultHourMinuteSecondsFormat1));
            result.Add(new Tuple<Regex, string>(TimeMillisecondsAmPmRegex, DefaultMinuteSecondsMillisecondsFormat));
            result.Add(new Tuple<Regex, string>(TimeMillisecondsAmPmRegex1, DefaultMinuteSecondsMillisecondsFormat));
            result.Add(new Tuple<Regex, string>(DateRegex1, DefaultDayMonthFormat));
            result.Add(new Tuple<Regex, string>(DateRegex3, DefaultDayMonthYearFormat1));
           // result.Add(new Tuple<Regex, string>(DateRegex4, DefaultDayMonthYearFormat));

            return result;
        }

        private static bool TryParseUsingPattern(string value, Regex regex, string newFormatSuggestedValue, out string newFormatValue)
        {
            newFormatValue = FormatHelper.GeneralFormatString;

            if (regex.IsMatch(value))
            {
                newFormatValue = newFormatSuggestedValue;
                return true;
            }

            return false;
        }

        #endregion


        #region Helper Methods

        private static int GetMonthNumberByName(string monthName)
        {
            switch (monthName.ToLower())
            {
                case "jan":
                case "january":
                    return 1;

                case "feb":
                case "february":
                    return 2;

                case "mar":
                case "march":
                    return 3;

                case "apr":
                case "april":
                    return 4;

                case "may":
                    return 5;

                case "jun":
                case "june":
                    return 6;

                case "jul":
                case "july":
                    return 7;

                case "aug":
                case "august":
                    return 8;

                case "sep":
                case "september":
                    return 9;

                case "oct":
                case "october":
                    return 10;

                case "nov":
                case "november":
                    return 11;

                case "dec":
                case "december":
                    return 12;

                default:
                    return -1;
            }
        }

        private static int GetNumberOfDaysByMonth(int month)
        {
            switch (month)
            {
                case 1:
                case 3:
                case 5:
                case 7:
                case 8:
                case 10:
                case 12:
                    return 31;

                case 4:
                case 6:
                case 9:
                case 11:
                    return 30;

                case 2:
                    if (DateTime.IsLeapYear(DateTime.Now.Year))
                    {
                        return 29;
                    }
                    else
                    {
                        return 28;
                    }
            }

            throw new ArgumentException("Invalid month.");
        }

        #endregion
    }
}