﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Windows.Media;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings.Builders;
using Telerik.Windows.Documents.Spreadsheet.Measurement;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
using TextMeasurementInfo = Telerik.Windows.Documents.Core.TextMeasurer.TextMeasurementInfo;

namespace Telerik.Windows.Documents.Spreadsheet.Formatting
{
    /// <summary>
    /// Defines auxiliary formatting methods.
    /// </summary>
    public static class FormatHelper
    {
        #region Constants

        // NOTE: According to an article in MSDN: http://msdn.microsoft.com/en-us/library/system.globalization.numberformatinfo.currencypositivepattern.aspx
        internal static readonly string[] PositivePatternStrings =
                               {
                                 "$n", "n$", "$ n", "n $" 
                               };

        // NOTE: According to an article in MSDN: http://msdn.microsoft.com/en-us/library/system.globalization.numberformatinfo.currencynegativepattern.aspx
        internal static readonly string[] NegativePatternStrings = 
                               { 
                                 "($n)", "-$n", "$-n", "$n-", "(n$)", "-n$", "n-$", "n$-", 
                                 "-n $", "-$ n", "n $-", "$ n-", "$ -n", "n- $", "($ n)", "(n $)" 
                               };
        internal static readonly string ZeroCountGroupName = "zeroCount";
        internal static readonly string UseThousandsSeparatorGroupName = "useThousandsSeparator";

        /// <summary>
        /// Defines the earliest date the RadSpreadsheet recognizes.
        /// </summary>
        public static readonly DateTime StartDate = new DateTime(1900, 1, 1, 0, 0, 0);

        internal static readonly string IsRtlRegExPattern = @"[\p{IsArabic}\p{IsHebrew}]";
        internal static readonly string GeneralFormatString = string.Empty;
        internal static readonly string GeneralFormatName = "General";
        internal static readonly string TextPlaceholder = "@";
        internal static readonly double MinimumScientificValue = 100000000000;
        internal static readonly char FractionsInvisibleChar = '1';
        internal static readonly char MilliSecondsChar = 'f';
        internal static readonly string InvisibleDivider = "q";
        internal static readonly double InvisibleDividerFontsize = 0.01;

        internal static readonly char[] NumberChars = new char[] { '0', '#' };
        internal static readonly char[] DateChars = new char[] { 'y', 'm', 'd' };
        internal static readonly char[] TimeChars = new char[] { 'h', 's' };
        internal static readonly double MaxDoubleValueTranslatableToDateTime = 2958465d;

        internal static readonly Color? DefaultFormatForeground = null;
        internal static readonly CultureInfo DefaultFormatCulture = null;
        internal static readonly Predicate<double> DefaultFormatCondition = null;

        internal static readonly double MaxGeneralNumber = 999999999999999999999d;
        internal static readonly double MinGeneralNumber = 0.0000000000000000001;
        internal static readonly string GeneralNumberEditFormatString = "0.###################";
        internal static readonly string GeneralNumberScientificEditFormatString = "0.##############E+00";

        #endregion


        #region Properties

        internal static string ShortDatePattern
        {
            get
            {
                return FormatHelper.DefaultSpreadsheetCulture.CultureInfo.DateTimeFormat.ShortDatePattern;
            }
        }

        internal static string LongTimePattern
        {
            get
            {
                return FormatHelper.DefaultSpreadsheetCulture.CultureInfo.DateTimeFormat.LongTimePattern;
            }
        }

        internal static string ShortDateLongTimePattern
        {
            get
            {
                return string.Format("{0} {1}", ShortDatePattern, LongTimePattern);
            }
        }

        internal static SpreadsheetCultureHelper DefaultSpreadsheetCulture
        {
            get;
            set;
        }

        #endregion


        #region Constructors

        static FormatHelper()
        {
            DefaultSpreadsheetCulture = new SpreadsheetCultureHelper();
        }

        #endregion


        #region Public Methods

        /// <summary>
        /// Converts a double number to DateTime.
        /// </summary>
        /// <param name="doubleValue">The double value.</param>
        /// <returns>The DateTime instance.</returns>
        public static DateTime? ConvertDoubleToDateTime(double doubleValue)
        {
            DateTime? dateTime;
            if (doubleValue < 0)
            {
                return null;
            }

            double value = doubleValue;

            // NOTE: There is known bug in Excel - 1900 is consedered to be a leap year. 
            // That's why numbers before 3/1/1900 appear wrong and that's why
            // we need that adjustment - http://www.ozgrid.com/Excel/ExcelDateandTimes.htm
            if (value <= 60d)
            {
                value += 1;
            }

            long ticks;
            if (TryConvertDoubleDateToTicks(value, out ticks))
            {
                dateTime = new DateTime(ticks, DateTimeKind.Unspecified);
            }
            else
            {
                dateTime = null;
            }

            return dateTime;
        }

        /// <summary>
        /// Converts a DateTime instance to double.
        /// </summary>
        /// <param name="dateTime">The DateTime instance.</param>
        /// <returns>The double number.</returns>
        public static double ConvertDateTimeToDouble(DateTime dateTime)
        {
            double dateAsNumber = dateTime.ToOADate();

            if (dateAsNumber <= 60d && dateAsNumber >= 1d)
            {
                dateAsNumber--;
            }

            return dateAsNumber;
        }

        #endregion


        #region Extension Methods

        /// <summary>
        /// Rounds the minutes of a DateTime instance.
        /// </summary>
        /// <param name="dateTimeValue">The DateTime value.</param>
        /// <returns>The DateTime instance with rounded minutes.</returns>
        public static DateTime RoundMinutes(this DateTime dateTimeValue)
        {
            int milliseconds = dateTimeValue.Millisecond;

            DateTime result = new DateTime(dateTimeValue.Year, dateTimeValue.Month, dateTimeValue.Day,
                dateTimeValue.Hour, dateTimeValue.Minute, dateTimeValue.Second);

            if (milliseconds >= 500)
            {
                result = result.AddSeconds(1);
            }

            return result;
        }

        /// <summary>
        /// Rounds the milliseconds of a DateTime instance.
        /// </summary>
        /// <param name="dateTimeValue">The DateTime value.</param>
        /// <param name="precision">The precision.</param>
        /// <returns>The DateTime instance with rounded milliseconds.</returns>
        public static DateTime RoundMilliseconds(this DateTime dateTimeValue, int precision)
        {
            int milliseconds = dateTimeValue.Millisecond;
            string millisecondsStr = milliseconds.ToString();
            DateTime dateTimeResult = dateTimeValue;

            if (precision > 0 && millisecondsStr.Length > precision)
            {
                string formattingStr = "0." + string.Concat(Enumerable.Repeat('0', precision));
                double valToFormat = milliseconds / Math.Pow(10, millisecondsStr.Length);
                string result = valToFormat.ToString(formattingStr);
                int newMilliseconds = int.Parse(result.Substring(2));

                int multiplyBy = millisecondsStr.Length - newMilliseconds.ToString().Length;

                newMilliseconds *= (int)Math.Pow(10, multiplyBy);

                dateTimeResult = new DateTime(dateTimeResult.Year, dateTimeResult.Month, dateTimeResult.Day,
                    dateTimeResult.Hour, dateTimeResult.Minute, dateTimeResult.Second, newMilliseconds);
            }

            return dateTimeResult;
        }

        /// <summary>
        /// Converts a NumberCellValue instance to DateTime.
        /// </summary>
        /// <param name="numberCellValue">The number cell value.</param>
        /// <returns>The DateTime instance.</returns>
        public static DateTime? ToDateTime(this NumberCellValue numberCellValue)
        {
            return FormatHelper.ConvertDoubleToDateTime(numberCellValue.Value);
        }

        internal static Color? GetForegroundOrDefault(this FormatDescriptor formatDescriptor)
        {
            if (formatDescriptor == null)
            {
                return FormatHelper.DefaultFormatForeground;
            }

            return formatDescriptor.Foreground;
        }

        internal static CultureInfo GetCultureOrDefault(this FormatDescriptor formatDescriptor)
        {
            if (formatDescriptor == null)
            {
                return FormatHelper.DefaultFormatCulture;
            }

            return formatDescriptor.Culture;
        }

        internal static Predicate<double> GetConditionOrDefault(this FormatDescriptor formatDescriptor)
        {
            if (formatDescriptor == null)
            {
                return FormatHelper.DefaultFormatCondition;
            }

            return formatDescriptor.Condition;
        }

        #endregion


        #region Methods

        internal static double Round(double value)
        {
            if (value > 0)
            {
                return Math.Floor(value);
            }

            return Math.Ceiling(value);
        }

        internal static string ReplaceCultureSpecificSeparatorsWithDefaults(string formatString)
        {
            return ReplaceSeparators(formatString,
                (c) =>
                {
                    return SpreadsheetCultureHelper.IsCharEqualTo(c, FormatHelper.DefaultSpreadsheetCulture.NumberGroupSeparator, FormatHelper.DefaultSpreadsheetCulture.CurrencyGroupSeparator);
                },
                (c) =>
                {
                    return SpreadsheetCultureHelper.IsCharEqualTo(c, FormatHelper.DefaultSpreadsheetCulture.NumberDecimalSeparator, FormatHelper.DefaultSpreadsheetCulture.CurrencyDecimalSeparator);
                },
                CultureInfo.InvariantCulture.NumberFormat.NumberGroupSeparator, CultureInfo.InvariantCulture.NumberFormat.NumberDecimalSeparator);
        }

        internal static string ReplaceDefaultSeparatorsWithCultureSpecific(string formatString)
        {
            return ReplaceSeparators(formatString,
                (c) =>
                {
                    return SpreadsheetCultureHelper.IsCharEqualTo(c, CultureInfo.InvariantCulture.NumberFormat.NumberGroupSeparator);
                },
                (c) =>
                {
                    return SpreadsheetCultureHelper.IsCharEqualTo(c, CultureInfo.InvariantCulture.NumberFormat.NumberDecimalSeparator);
                },
                FormatHelper.DefaultSpreadsheetCulture.NumberGroupSeparator, FormatHelper.DefaultSpreadsheetCulture.NumberDecimalSeparator);
        }

        private static string ReplaceSeparators(string formatString, Predicate<char> shouldReplaceWithGroupSeparator, Predicate<char> shouldReplaceWithDecimalSeparator, string newGroupSeparator, string newDecimalSeparator)
        {
            string currencySymbol = FormatHelper.DefaultSpreadsheetCulture.CultureInfo.NumberFormat.CurrencySymbol;
            string invisibleCurrencySymbol = BuildersHelper.MakeCurrencySymbolInvisible(currencySymbol);

            int[] currencySymbolStartIndeces = FindAllOccurrences(formatString, currencySymbol);
            int[] invisibleCurrencySymbolStartIndeces = FindAllOccurrences(formatString, invisibleCurrencySymbol);

            StringBuilder sb = new StringBuilder();
            bool isInBrackets = false;
            bool isLiteral = false;
            for (int i = 0; i < formatString.Length; i++)
            {
                char currentChar = formatString[i];

                if (CheckIfIsCurrencySymbolPart(currencySymbolStartIndeces, currencySymbol, i) ||
                    CheckIfIsCurrencySymbolPart(invisibleCurrencySymbolStartIndeces, invisibleCurrencySymbol, i))
                {
                    sb.Append(currentChar);
                    continue;
                }

                if (currentChar == '\"')
                {
                    isLiteral = !isLiteral;
                    sb.Append(currentChar);
                }
                else if (!isLiteral)
                {
                    if (currentChar == '[')
                    {
                        isInBrackets = true;
                    }
                    else if (currentChar == ']')
                    {
                        isInBrackets = false;
                    }

                    if (!isInBrackets && shouldReplaceWithGroupSeparator(currentChar))
                    {
                        sb.Append(newGroupSeparator);
                    }
                    else if (!isInBrackets && shouldReplaceWithDecimalSeparator(currentChar))
                    {
                        sb.Append(newDecimalSeparator);
                    }
                    else
                    {
                        sb.Append(currentChar);
                    }
                }
                else
                {
                    sb.Append(currentChar);
                }
            }

            return sb.ToString();
        }

        internal static bool CheckIfIsCurrencySymbolPart(int[] startIndeces, string currencySymbol, int currentCharIndex)
        {
            bool isCurrencySymbolPart = false;
            for (int i = 0; i < startIndeces.Length; i++)
            {
                int currencySymbolStart = startIndeces[i];
                int currencySymbolEnd = currencySymbolStart + currencySymbol.Length - 1;
                if (currencySymbolStart <= currentCharIndex && currentCharIndex <= currencySymbolEnd)
                {
                    isCurrencySymbolPart = true;
                    break;
                }
            }

            return isCurrencySymbolPart;
        }

        internal static int[] FindAllOccurrences(string value, string part)
        {
            List<int> list = new List<int>();

            int startPosition = value.IndexOf(part, StringComparison.Ordinal);

            while (startPosition >= 0)
            {
                list.Add(list.Count * part.Length + startPosition);
                value = value.Remove(startPosition, part.Length);
                startPosition = value.IndexOf(part);
            }

            return list.ToArray();
        }

        internal static CultureInfo GetCurrentCulture()
        {
            return Thread.CurrentThread.CurrentCulture;
        }

        internal static DateTime? RoundUpValue(string formatString, DateTime? dateTimeValue)
        {
            DateTime? newDateTimeValue = dateTimeValue;
            if (!newDateTimeValue.HasValue || newDateTimeValue.Value.Millisecond == 0)
            {
                return newDateTimeValue;
            }

            if (!formatString.Contains(FormatHelper.MilliSecondsChar))
            {
                newDateTimeValue = newDateTimeValue.Value.RoundMinutes();
            }
            else
            {
                int zerosCount = formatString.Where((c) => { return c == FormatHelper.MilliSecondsChar; }).Count();
                newDateTimeValue = newDateTimeValue.Value.RoundMilliseconds(zerosCount);
            }

            return newDateTimeValue;
        }

        internal static string ExpandSymbol(string symbol, double availableWidth, FontProperties fontProperties)
        {
            TextMeasurementInfo measurementInfo = RadTextMeasurer.Measure(symbol, fontProperties);
            double symbolWidth = measurementInfo.Size.Width;

            int symbolCount = (int)(availableWidth / symbolWidth);

            string result = string.Concat(Enumerable.Repeat(symbol, symbolCount));
            double resultWidth = RadTextMeasurer.Measure(result, fontProperties).Size.Width;
            while ((resultWidth - availableWidth) > 0.1 && result.Length > 0)
            {
                result = result.Substring(0, result.Length - 1);
                resultWidth = RadTextMeasurer.Measure(result, fontProperties).Size.Width;
            }

            return result;
        }

        internal static string StripUnderscores(string text)
        {
            StringBuilder result = new StringBuilder();

            int i = 0;
            while (i < text.Length)
            {
                if (text[i] != '_' || i == (text.Length - 1))
                {
                    result.Append(text[i]);
                    i++;
                }
                else
                {
                    i += 2;
                }
            }

            return result.ToString();
        }

        internal static string StripBrackets(string text)
        {
            if (string.IsNullOrEmpty(text))
            {
                return text;
            }

            text = text.Trim();

            int startIndex = text.IndexOf('[');

            if (startIndex != -1)
            {
                int endIndex = text.IndexOf(']', startIndex);

                while (startIndex != -1 && startIndex < endIndex)
                {
                    text = text.Remove(startIndex, endIndex - startIndex + 1);
                    text = text.Trim();

                    startIndex = text.IndexOf('[');
                    if (startIndex != -1)
                    {
                        endIndex = text.IndexOf(']', startIndex);
                    }
                }
            }

            return text;
        }

        internal static string StripCurrencyEscapingsAndSlashes(string text)
        {
            string currencySymbol = FormatHelper.DefaultSpreadsheetCulture.CultureInfo.NumberFormat.CurrencySymbol;
            string invisibleCurrencySymbol = BuildersHelper.MakeCurrencySymbolInvisible(currencySymbol);

            text = RemoveCurrencySymbolEscapingsAndSlashes(text, currencySymbol);
            text = RemoveCurrencySymbolEscapingsAndSlashes(text, invisibleCurrencySymbol);

            return text;
        }
 
        private static string RemoveCurrencySymbolEscapingsAndSlashes(string text, string currencySymbol)
        {
            int startPosition = text.IndexOf(currencySymbol, StringComparison.Ordinal);

            while (startPosition != -1)
            {
                int currencySymbolLength = currencySymbol.Length;

                while (startPosition > 0 && (text[startPosition - 1] == '\\' || text[startPosition - 1] == '\"'))
                {
                    startPosition--;
                    currencySymbolLength++;
                }

                while ((startPosition + currencySymbolLength + 1) < text.Length &&
                    text[startPosition + currencySymbolLength] == '\"')
                {
                    currencySymbolLength++;
                }

                text = text.Remove(startPosition, currencySymbolLength);
                startPosition = text.IndexOf(currencySymbol, StringComparison.Ordinal);
            }

            return text;
        }

        internal static bool ContainsTotalHoursMinutesSecondsModifiers(string formatString)
        {
            StringBuilder sb = new StringBuilder(formatString);

            if (sb.Length == 0)
            {
                return false;
            }

            sb.Replace(sb.ToString(), sb.ToString().Trim());
            int index;
            while ((index = sb.ToString().IndexOf('[')) != -1)
            {
                int index1 = sb.ToString().IndexOf(']', index);
                if (index1 == -1)
                {
                    break;
                }

                string textInBrackets = sb.ToString().Substring(index + 1, index1 - index - 1);
                if (textInBrackets == "h" || textInBrackets == "m" || textInBrackets == "s")
                {
                    return true;
                }

                sb.Remove(index, index1 - index + 1);

                if (!string.IsNullOrEmpty(sb.ToString()))
                {
                    sb.Replace(sb.ToString(), sb.ToString().Trim());
                }
            }

            return false;
        }

        internal static bool TryGetDecimalPlacesForFormatString(string formatString, CultureInfo cultureInfo, out int decimalPlaces)
        {
            bool useThousandsSeparator;
            return TryGetDecimalPlacesForFormatString(formatString, cultureInfo, out decimalPlaces, out useThousandsSeparator);
        }

        internal static bool TryGetDecimalPlacesForFormatString(string formatString, CultureInfo cultureInfo,
            out int decimalPlaces, out bool useThousandsSeparator)
        {
            string pattern = string.Format(@"(?<{0}>#{1}##)?0({2}(?<{3}>0+))?",
                UseThousandsSeparatorGroupName,
                cultureInfo.NumberFormat.NumberGroupSeparator,
                cultureInfo.NumberFormat.NumberDecimalSeparator,
                ZeroCountGroupName);
            decimalPlaces = -1;
            useThousandsSeparator = false;

            Regex regex = new Regex(pattern);
            MatchCollection matches = regex.Matches(formatString);
            if (matches.Count > 0)
            {
                Match match = matches[0];
                foreach (Match currentMatch in matches)
                {
                    if (currentMatch.Length > match.Length)
                    {
                        match = currentMatch;
                    }
                }

                decimalPlaces = match.Groups[ZeroCountGroupName].Value.Length;
                useThousandsSeparator = match.Groups[UseThousandsSeparatorGroupName].Success;
                return true;
            }

            return false;
        }

        internal static string GetValuePatternConsiderDecimalPlaces(int decimalPlaces, bool useThousandsSeparator = true)
        {
            StringBuilder sb = new StringBuilder();
            if (useThousandsSeparator)
            {
                sb.Append(string.Format("#{0}##", FormatHelper.DefaultSpreadsheetCulture.CultureInfo.NumberFormat.NumberGroupSeparator));
            }
            sb.Append("0");

            if (decimalPlaces > 0)
            {
                sb.Append(FormatHelper.DefaultSpreadsheetCulture.CultureInfo.NumberFormat.NumberDecimalSeparator);
                for (int i = 0; i < decimalPlaces; i++)
                {
                    sb.Append("0");
                }
            }

            return sb.ToString();
        }

        internal static ICellValue GetCellValueRespectFormat(CellValueFormat format, ICellValue currentCellValue)
        {
            ICellValue resultCellValue = currentCellValue;
            if (format.FormatStringInfo.FormatType == FormatStringType.Text)
            {
                NumberCellValue numberCellValue = currentCellValue as NumberCellValue;
                if (numberCellValue != null)
                {
                    resultCellValue = new TextCellValue(numberCellValue.Value.ToString(FormatHelper.DefaultSpreadsheetCulture.CultureInfo));
                }

                BooleanCellValue booleanCellValue = currentCellValue as BooleanCellValue;
                if (booleanCellValue != null)
                {
                    resultCellValue = new TextCellValue(booleanCellValue.GetResultValueAsString(CellValueFormat.GeneralFormat).ToString());
                }

                FormulaCellValue formulaCellValue = currentCellValue as FormulaCellValue;
                if (formulaCellValue != null)
                {
                    resultCellValue = new TextCellValue(formulaCellValue.GetValueAsString(format));
                }
            }
            else if (currentCellValue.ValueType == CellValueType.Text)
            {
                double doubleValue;
                string newFormat;
                if (NumberValueParser.TryParse(currentCellValue.GetValueAsString(format), null, null, out newFormat, out doubleValue))
                {
                    resultCellValue = new NumberCellValue(doubleValue);
                }
            }

            return resultCellValue;
        }

        private static bool TryConvertDoubleDateToTicks(double value, out long ticks)
        {
            ticks = -1;

            double num;
            if (value >= 2958466 || value <= -657435)
            {
                return false;
            }
            else
            {
                double num1 = value * 86400000;
                if (value >= 0)
                {
                    num = 0.5;
                }
                else
                {
                    num = -0.5;
                }
                long num2 = (long)(num1 + num);
                if (num2 < (long)0)
                {
                    num2 = num2 - num2 % (long)86400000 * (long)2;
                }
                num2 = num2 + 59926435200000L;
                if (num2 < (long)0 || num2 >= 315537897600000L)
                {
                    return false;
                }

                ticks = num2 * (long)10000;
                return true;
            }
        }

        #endregion
    }
}