﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings.CategoryManagers;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings.Builders
{
    /// <summary>
    /// Represents the accounting format string builder class.
    /// </summary>
    public static class AccountingFormatStringBuilder
    {
        #region Constants

        private static readonly string ExpandSymbol = "* ";
        private static readonly string ZeroText = "\"-\"{0}";

        #endregion


        #region Fields

        private static readonly List<Func<CurrencyInfo, int, string>> formatStringHandlers;

        #endregion


        #region Constructors

        static AccountingFormatStringBuilder()
        {
            formatStringHandlers = new List<Func<CurrencyInfo, int, string>>();
            formatStringHandlers.Add(DefaultAccountingFormatStringsHandler);
            formatStringHandlers.Add(CurrencyWithCultureFormatStringsHandler);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Builds the format string.
        /// </summary>
        /// <param name="currencyInfo">The currency info.</param>
        /// <param name="decimalPlaces">The decimal places.</param>
        /// <returns></returns>
        public static string BuildFormatString(CurrencyInfo currencyInfo, int decimalPlaces)
        {
            string result = null;
            foreach (var handler in formatStringHandlers)
            {
                result = handler(currencyInfo, decimalPlaces);
                if (result != null)
                {
                    break;
                }
            }

            return result;
        }

        private static string DefaultAccountingFormatStringsHandler(CurrencyInfo currencyInfo, int decimalPlaces)
        {
            if (TelerikHelper.EqualsOfT(currencyInfo, CurrencyInfo.None))
            {
                CultureInfo cultureInfo = FormatHelper.DefaultSpreadsheetCulture.CultureInfo;
                string currencySymbol =  BuildersHelper.MakeCurrencySymbolInvisible(cultureInfo.NumberFormat.CurrencySymbol);

                return GetAccountingStrignsForCurrentCulture(cultureInfo, currencySymbol, decimalPlaces);
            }

            if (TelerikHelper.EqualsOfT(currencyInfo, CurrencyInfo.Symbol))
            {
                CultureInfo cultureInfo = FormatHelper.DefaultSpreadsheetCulture.CultureInfo;
                string currencySymbol = currencyInfo.EscapedCurrencySymbol;

                return GetAccountingStrignsForCurrentCulture(cultureInfo, currencySymbol, decimalPlaces);
            }

            return null;
        }

        private static string CurrencyWithCultureFormatStringsHandler(CurrencyInfo currencyInfo, int decimalPlaces)
        {
            string currencyCode = CurrencyFormatCategoryManager.CultureInfoToCurrencyCode[currencyInfo.Culture];
            return GetAccountingStrignsForCurrentCulture(currencyInfo.Culture, currencyCode, decimalPlaces);
        }

        private static string GetAccountingStrignsForCurrentCulture(CultureInfo cultureInfo, string currencySymbol, int decimalPlaces)
        {
            if (cultureInfo == null)
            {
                return null;
            }

            string valuePattern = FormatHelper.GetValuePatternConsiderDecimalPlaces(decimalPlaces);

            int currencyNegativePatternIndex = cultureInfo.NumberFormat.CurrencyNegativePattern;
            string positiveSystemPattern = FormatHelper.PositivePatternStrings[cultureInfo.NumberFormat.CurrencyPositivePattern];
            string negativeSystemPattern = FormatHelper.NegativePatternStrings[currencyNegativePatternIndex];

            string start;
            string middle;
            string end;
            AccountingFormatStringBuilder.GetStartMiddleEnd(negativeSystemPattern, out start, out middle, out end);

            string positivePattern = GetPositivePattern(positiveSystemPattern, start, middle, end, currencySymbol, valuePattern);
            string negativePattern = AccountingFormatStringBuilder.GetNegativePattern(negativeSystemPattern, start, end, currencySymbol, valuePattern);
            string zeroPattern = AccountingFormatStringBuilder.GetZeroPattern(positiveSystemPattern, start, middle, end, currencySymbol, decimalPlaces);
            string textPattern = AccountingFormatStringBuilder.GetTextPattern(start, end);

            string stringBuilt = string.Format("{0};{1};{2};{3}", positivePattern, negativePattern, zeroPattern, textPattern);

            return stringBuilt;
        }

        private static string GetPositivePattern(string positivePatternSystem, string start, string middle, string end, string currencySymbol,
            string valuePattern)
        {
            int indexOfDollarSign = positivePatternSystem.IndexOf('$');
            string result = positivePatternSystem;

            if (indexOfDollarSign > positivePatternSystem.IndexOf('n'))
            {
                result = result.Insert(0, ExpandSymbol);
            }
            else
            {
                result = result.Insert(indexOfDollarSign + 1, ExpandSymbol);
            }

            result = result.Insert(0, string.Format("_{0}", start));

            result += string.Format("_{0}", end);
            result = result.Insert(result.IndexOf('n') + 1, string.IsNullOrEmpty(middle) ? string.Empty : string.Format("_{0}", middle));

            result = result.Replace("n", valuePattern);
            result = result.Replace("$", currencySymbol);

            return result;
        }

        private static string GetNegativePattern(string negativePatternSystem, string start, string end, string currencySymbol, string valuePattern)
        {
            int indexOfDollarSign = negativePatternSystem.IndexOf('$');
            string result = negativePatternSystem;

            if (indexOfDollarSign > negativePatternSystem.IndexOf('n'))
            {
                int indexOf_n = result.IndexOf('n');
                result = result.Insert(indexOf_n, ExpandSymbol);
            }
            else
            {
                result = result.Insert(indexOfDollarSign + 1, ExpandSymbol);
            }

            if (result.Contains('('))
            {
                result = result.Replace("(", string.Empty);
                result = result.Replace(")", string.Empty);
                int indexOf_n = result.IndexOf('n');
                result = result.Insert(indexOf_n + 1, ")");
                result = result.Insert(indexOf_n, "(");
            }

            if (!string.IsNullOrEmpty(start) && result.First() != start[0])
            {
                result = result.Insert(0, string.Format("_{0}", start));
            }

            if (!string.IsNullOrEmpty(end) && result.Last() != end[0])
            {
                result += string.Format("_{0}", end);
            }

            result = result.Replace("n", valuePattern);
            result = result.Replace("$", currencySymbol);

            return result;
        }

        private static string GetZeroPattern(string positivePatternSystem, string start, string middle, string end, string currencySymbol, int decimalPlaces)
        {
            int indexOfDollarSign = positivePatternSystem.IndexOf('$');

            string result = positivePatternSystem;

            if (indexOfDollarSign > positivePatternSystem.IndexOf('n'))
            {
                result = result.Insert(0, ExpandSymbol);
            }
            else
            {
                result = result.Insert(indexOfDollarSign + 1, ExpandSymbol);
            }

            result = result.Insert(0, string.Format("_{0}", start));

            result += string.Format("_{0}", end);
            result = result.Insert(result.IndexOf('n') + 1, string.IsNullOrEmpty(middle) ? string.Empty : string.Format("_{0}", middle));

            result = result.Replace("n", string.Format(ZeroText, string.Concat(Enumerable.Repeat('?', decimalPlaces))));
            result = result.Replace("$", currencySymbol);

            return result;
        }

        private static string GetTextPattern(string start, string end)
        {
            return string.Format("_{0}@_{1}", start, end);
        }

        private static void GetStartMiddleEnd(string negativePattern, out string start, out string middle, out string end)
        {
            start = string.Empty;
            middle = string.Empty;
            end = string.Empty;

            if (negativePattern.Contains("(") && negativePattern.IndexOf('$') > negativePattern.IndexOf('n'))
            {
                start = end = " ";
                middle = ")";
            }
            else if (negativePattern.Contains("("))
            {
                start = "(";
                end = ")";
            }
            else if (negativePattern[0] == '-' || negativePattern.Last() == '-')
            {
                start = end = "-";
            }
            else
            {
                start = end = " ";
            }
        }

        #endregion
    }
}