﻿using System;
using System.Collections.Generic;
using System.Globalization;

namespace Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings.CategoryManagers
{
    /// <summary>
    /// Represents the currency format category manager class.
    /// </summary>
    public static class CurrencyFormatCategoryManager
    {
        #region Fields

        private const int CurrencyInfosSymbolIndex = 1;
        internal const int CurrencyInfosNoneIndex = 0;
        private static readonly Dictionary<CultureInfo, string> cultureInfoToCurrencyCode;
        private static readonly List<CurrencyInfo> currencyInfos;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the culture info to currency code.
        /// </summary>
        /// <value>The culture info to currency code.</value>
        public static Dictionary<CultureInfo, string> CultureInfoToCurrencyCode
        {
            get
            {
                return cultureInfoToCurrencyCode;
            }
        }

        /// <summary>
        /// Gets the currency infos.
        /// </summary>
        /// <value>The currency infos.</value>
        public static IEnumerable<CurrencyInfo> CurrencyInfos
        {
            get
            {
                if (!currencyInfos[CurrencyInfosSymbolIndex].Equals(CurrencyInfo.Symbol))
                {
                    currencyInfos[CurrencyInfosSymbolIndex] = CurrencyInfo.Symbol;
                }

                return currencyInfos;
            }
        }

        #endregion


        #region Constructors

        static CurrencyFormatCategoryManager()
        {
            cultureInfoToCurrencyCode = new Dictionary<CultureInfo, string>();
            InitCultureInfoToCurrencyCode();

            currencyInfos = new List<CurrencyInfo>();
            InitCurrencyInfos();
        }

        private static void InitCurrencyInfos()
        {
            AddCurrencyInfo("None", null);
            AddCurrencyInfo(FormatHelper.DefaultSpreadsheetCulture.CultureInfo.NumberFormat.CurrencySymbol, null);

            foreach (CultureInfo cultureInfo in CultureInfoToCurrencyCode.Keys)
            {
                AddCurrencyInfo(cultureInfo.NumberFormat.CurrencySymbol, cultureInfo);
            }
        }

        private static void AddCurrencyInfo(string currencySymbol, CultureInfo culture)
        {
            currencyInfos.Add(new CurrencyInfo(currencySymbol, culture));
        }

        private static void InitCultureInfoToCurrencyCode()
        {
            AddCultureInfoToCurrencyCode("af", "[$R-436]");
            AddCultureInfoToCurrencyCode("gsw", "[$€-484]");
            AddCultureInfoToCurrencyCode("am", "[$ETB-45E]");
            AddCultureInfoToCurrencyCode("ar-DZ", "[$د.ج.‏-1401]");
            AddCultureInfoToCurrencyCode("ar-SA", "[$ر.س.‏-401]");
            AddCultureInfoToCurrencyCode("hy", "[$դր.-42B]");
            AddCultureInfoToCurrencyCode("as", "[$ট-44D]");
            AddCultureInfoToCurrencyCode("az-Cyrl", "[$ман.-82C]");
            AddCultureInfoToCurrencyCode("az-Latn", "[$man.-42C]");
            AddCultureInfoToCurrencyCode("ba", "[$һ.-46D]");
            AddCultureInfoToCurrencyCode("eu", "[$€-42D]");
            AddCultureInfoToCurrencyCode("be", "[$р.-423]");
            AddCultureInfoToCurrencyCode("bn-BD", "[$৳-845]");
            AddCultureInfoToCurrencyCode("bn-IN", "[$টা-445]");
            AddCultureInfoToCurrencyCode("bs-Cyrl", "[$КМ-201A]");
            AddCultureInfoToCurrencyCode("bs-Latn", "[$KM-141A]");
            AddCultureInfoToCurrencyCode("br", "[$€-47E]");
            AddCultureInfoToCurrencyCode("bg", "[$лв.-402]");
            AddCultureInfoToCurrencyCode("ca", "[$€-403]");
            AddCultureInfoToCurrencyCode("zh-HK", "[$HK$-C04]");
            AddCultureInfoToCurrencyCode("zh-MO", "[$MOP-1404]");
            AddCultureInfoToCurrencyCode("zh-CN", "[$¥-804]");
            AddCultureInfoToCurrencyCode("zh-SG", "[$$-1004]");
            AddCultureInfoToCurrencyCode("zh-TW", "[$NT$-404]");
            AddCultureInfoToCurrencyCode("co", "[$€-483]");
            AddCultureInfoToCurrencyCode("hr-BA", "[$KM-101A]");
            AddCultureInfoToCurrencyCode("hr-HR", "[$kn-41A]");
            AddCultureInfoToCurrencyCode("cs", "[$Kč-405]");
            AddCultureInfoToCurrencyCode("da", "[$kr.-406]");
            AddCultureInfoToCurrencyCode("prs-AF", "[$؋-48C]");
            AddCultureInfoToCurrencyCode("dv", "[$ރ.-465]");
            AddCultureInfoToCurrencyCode("nl-BE", "[$€-813]");
            AddCultureInfoToCurrencyCode("nl-NL", "[$€-413]");
            AddCultureInfoToCurrencyCode("en-AU", "[$$-C09]");
            AddCultureInfoToCurrencyCode("en-BZ", "[$BZ$-2809]");
            AddCultureInfoToCurrencyCode("en-CA", "[$$-1009]");
            AddCultureInfoToCurrencyCode("en-029", "[$$-2409]");
            AddCultureInfoToCurrencyCode("en-IN", "[$Rs.-4009]");
            AddCultureInfoToCurrencyCode("en-IE", "[$€-1809]");
            AddCultureInfoToCurrencyCode("en-JM", "[$J$-2009]");
            AddCultureInfoToCurrencyCode("en-MY", "[$RM-4409]");
            AddCultureInfoToCurrencyCode("en-NZ", "[$$-1409]");
            AddCultureInfoToCurrencyCode("en-PH", "[$Php-3409]");
            AddCultureInfoToCurrencyCode("en-SG", "[$$-4809]");
            AddCultureInfoToCurrencyCode("en-US", "[$$-409]");
            AddCultureInfoToCurrencyCode("en-ZA", "[$R-1C09]");
            AddCultureInfoToCurrencyCode("en-TT", "[$TT$-2C09]");
            AddCultureInfoToCurrencyCode("en-GB", "[$£-809]");
            AddCultureInfoToCurrencyCode("en-ZW", "[$Z$-3009]");
            AddCultureInfoToCurrencyCode("et", "[$kr-425]");
            AddCultureInfoToCurrencyCode("fo", "[$kr.-438]");
            AddCultureInfoToCurrencyCode("fil", "[$PhP-464]");
            AddCultureInfoToCurrencyCode("fi", "[$€-40B]");
            AddCultureInfoToCurrencyCode("fr-BE", "[$€-80C]");
            AddCultureInfoToCurrencyCode("fr-CA", "[$$-C0C]");
            AddCultureInfoToCurrencyCode("fr-FR", "[$€-40C]");
            AddCultureInfoToCurrencyCode("fr-LU", "[$€-140C]");
            AddCultureInfoToCurrencyCode("fr-MC", "[$€-180C]");
            AddCultureInfoToCurrencyCode("fr-CH", "[$fr.-100C]");
            AddCultureInfoToCurrencyCode("fy", "[$€-462]");
            AddCultureInfoToCurrencyCode("gl", "[$€-456]");
            AddCultureInfoToCurrencyCode("ka", "[$Lari-437]");
            AddCultureInfoToCurrencyCode("de-AT", "[$€-C07]");
            AddCultureInfoToCurrencyCode("de-DE", "[$€-407]");
            AddCultureInfoToCurrencyCode("de-LI", "[$CHF-1407]");
            AddCultureInfoToCurrencyCode("de-LU", "[$€-1007]");
            AddCultureInfoToCurrencyCode("de-CH", "[$Fr.-807]");
            AddCultureInfoToCurrencyCode("el", "[$€-408]");
            AddCultureInfoToCurrencyCode("kl", "[$kr.-46F]");
            AddCultureInfoToCurrencyCode("gu", "[$રૂ-447]");
            AddCultureInfoToCurrencyCode("ha", "[$N-468]");
            AddCultureInfoToCurrencyCode("he", "[$₪-40D]");
            AddCultureInfoToCurrencyCode("hi", "[$रु-439]");
            AddCultureInfoToCurrencyCode("hu", "[$Ft-40E]");
            AddCultureInfoToCurrencyCode("is", "[$kr.-40F]");
            AddCultureInfoToCurrencyCode("ig", "[$N-470]");
            AddCultureInfoToCurrencyCode("id", "[$Rp-421]");
            AddCultureInfoToCurrencyCode("iu-Latn", "[$$-85D]");
            AddCultureInfoToCurrencyCode("iu-Cans", "[$$-45D]");
            AddCultureInfoToCurrencyCode("ga", "[$€-83C]");
            AddCultureInfoToCurrencyCode("xh", "[$R-434]");
            AddCultureInfoToCurrencyCode("zu", "[$R-435]");
            AddCultureInfoToCurrencyCode("it-IT", "[$€-410]");
            AddCultureInfoToCurrencyCode("it-CH", "[$fr.-810]");
            AddCultureInfoToCurrencyCode("ja", "[$¥-411]");
            AddCultureInfoToCurrencyCode("kn", "[$ರೂ-44B]");
            AddCultureInfoToCurrencyCode("kk", "[$Т-43F]");
            AddCultureInfoToCurrencyCode("km", "[$៛-453]");
            AddCultureInfoToCurrencyCode("qut", "[$Q-486]");
            AddCultureInfoToCurrencyCode("rw", "[$RWF-487]");
            AddCultureInfoToCurrencyCode("kok", "[$रु-457]");
            AddCultureInfoToCurrencyCode("ko", "[$₩-412]");
            AddCultureInfoToCurrencyCode("ky", "[$сом-440]");
            AddCultureInfoToCurrencyCode("lo", "[$₭-454]");
            AddCultureInfoToCurrencyCode("lv", "[$Ls-426]");
            AddCultureInfoToCurrencyCode("lt", "[$Lt-427]");
            AddCultureInfoToCurrencyCode("dsb", "[$€-82E]");
            AddCultureInfoToCurrencyCode("lb", "[$€-46E]");
            AddCultureInfoToCurrencyCode("mk", "[$ден.-42F]");
            AddCultureInfoToCurrencyCode("ms-BN", "[$$-83E]");
            AddCultureInfoToCurrencyCode("ms-MY", "[$RM-43E]");
            AddCultureInfoToCurrencyCode("ml", "[$ക-44C]");
            AddCultureInfoToCurrencyCode("mt", "[$€-43A]");
            AddCultureInfoToCurrencyCode("mi", "[$$-481]");
            AddCultureInfoToCurrencyCode("arn", "[$$-47A]");
            AddCultureInfoToCurrencyCode("mr", "[$रु-44E]");
            AddCultureInfoToCurrencyCode("moh", "[$$-47C]");
            AddCultureInfoToCurrencyCode("mn-Cyrl", "[$₮-450]");
            AddCultureInfoToCurrencyCode("mn-Mong-CN", "[$¥-850]");
            AddCultureInfoToCurrencyCode("ne", "[$रु-461]");
            AddCultureInfoToCurrencyCode("nb", "[$kr-414]");
            AddCultureInfoToCurrencyCode("nn", "[$kr-814]");
            AddCultureInfoToCurrencyCode("oc", "[$€-482]");
            AddCultureInfoToCurrencyCode("or", "[$ଟ-448]");
            AddCultureInfoToCurrencyCode("ps", "[$؋-463]");
            AddCultureInfoToCurrencyCode("fa", "[$ريال-429]");
            AddCultureInfoToCurrencyCode("pl", "[$zł-415]");
            AddCultureInfoToCurrencyCode("pt-BR", "[$R$-416]");
            AddCultureInfoToCurrencyCode("pt-PT", "[$€-816]");
            AddCultureInfoToCurrencyCode("pa", "[$ਰੁ-446]");
            AddCultureInfoToCurrencyCode("quz-BO", "[$$b-46B]");
            AddCultureInfoToCurrencyCode("quz-EC", "[$$-86B]");
            AddCultureInfoToCurrencyCode("quz-PE", "[$S/.-C6B]");
            AddCultureInfoToCurrencyCode("ro", "[$lei-418]");
            AddCultureInfoToCurrencyCode("rm", "[$fr.-417]");
            AddCultureInfoToCurrencyCode("ru", "[$р.-419]");
            AddCultureInfoToCurrencyCode("smn-FI", "[$€-243B]");
            AddCultureInfoToCurrencyCode("smj-NO", "[$kr-103B]");
            AddCultureInfoToCurrencyCode("smj-SE", "[$kr-143B]");
            AddCultureInfoToCurrencyCode("se-FI", "[$€-C3B]");
            AddCultureInfoToCurrencyCode("se-NO", "[$kr-43B]");
            AddCultureInfoToCurrencyCode("se-SE", "[$kr-83B]");
            AddCultureInfoToCurrencyCode("sms-FI", "[$€-203B]");
            AddCultureInfoToCurrencyCode("sma-NO", "[$kr-183B]");
            AddCultureInfoToCurrencyCode("sma-SE", "[$kr-1C3B]");
            AddCultureInfoToCurrencyCode("sa", "[$रु-44F]");
            AddCultureInfoToCurrencyCode("gd-GB", "[$£-491]");
            AddCultureInfoToCurrencyCode("sr-Cyrl-BA", "[$КМ-1C1A]");
            AddCultureInfoToCurrencyCode("sr-Cyrl-ME", "[$€-301A]");
            AddCultureInfoToCurrencyCode("sr-Latn-BA", "[$KM-181A]");
            AddCultureInfoToCurrencyCode("sr-Latn-ME", "[$€-2C1A]");
            AddCultureInfoToCurrencyCode("nso", "[$R-46C]");
            AddCultureInfoToCurrencyCode("tn", "[$R-432]");
            AddCultureInfoToCurrencyCode("si", "[$රු.-45B]");
            AddCultureInfoToCurrencyCode("sk", "[$€-41B]");
            AddCultureInfoToCurrencyCode("sl", "[$€-424]");
            AddCultureInfoToCurrencyCode("es-AR", "[$$-2C0A]");
            AddCultureInfoToCurrencyCode("es-BO", "[$$b-400A]");
            AddCultureInfoToCurrencyCode("es-CL", "[$$-340A]");
            AddCultureInfoToCurrencyCode("es-CO", "[$$-240A]");
            AddCultureInfoToCurrencyCode("es-CR", "[$₡-140A]");
            AddCultureInfoToCurrencyCode("es-DO", "[$RD$-1C0A]");
            AddCultureInfoToCurrencyCode("es-EC", "[$$-300A]");
            AddCultureInfoToCurrencyCode("es-SV", "[$$-440A]");
            AddCultureInfoToCurrencyCode("es-GT", "[$Q-100A]");
            AddCultureInfoToCurrencyCode("es-HN", "[$L.-480A]");
            AddCultureInfoToCurrencyCode("es-MX", "[$$-80A]");
            AddCultureInfoToCurrencyCode("es-NI", "[$C$-4C0A]");
            AddCultureInfoToCurrencyCode("es-PA", "[$B/.-180A]");
            AddCultureInfoToCurrencyCode("es-PY", "[$Gs-3C0A]");
            AddCultureInfoToCurrencyCode("es-PE", "[$S/.-280A]");
            AddCultureInfoToCurrencyCode("es-PR", "[$$-500A]");
            AddCultureInfoToCurrencyCode("es-ES_tradnl", "[$€-40A]");
            AddCultureInfoToCurrencyCode("es-US", "[$$-540A]");
            AddCultureInfoToCurrencyCode("es-UY", "[$$U-380A]");
            AddCultureInfoToCurrencyCode("sw", "[$S-441]");
            AddCultureInfoToCurrencyCode("sv-FI", "[$€-81D]");
            AddCultureInfoToCurrencyCode("sv-SE", "[$kr-41D]");
            AddCultureInfoToCurrencyCode("syr", "[$ل.س.‏-45A]");
            AddCultureInfoToCurrencyCode("tg", "[$т.р.-428]");
            AddCultureInfoToCurrencyCode("tzm-Latn-DZ", "[$DZD-85F]");
            AddCultureInfoToCurrencyCode("ta", "[$ரூ-449]");
            AddCultureInfoToCurrencyCode("tt", "[$р.-444]");
            AddCultureInfoToCurrencyCode("te", "[$రూ-44A]");
            AddCultureInfoToCurrencyCode("th", "[$฿-41E]");
            AddCultureInfoToCurrencyCode("bo-CN", "[$¥-451]");
            AddCultureInfoToCurrencyCode("tr", "[$TL-41F]");
            AddCultureInfoToCurrencyCode("tk", "[$m.-442]");
            AddCultureInfoToCurrencyCode("uk", "[$₴-422]");
            AddCultureInfoToCurrencyCode("hsb", "[$€-42E]");
            AddCultureInfoToCurrencyCode("ur", "[$Rs-420]");
            AddCultureInfoToCurrencyCode("ug", "[$¥-480]");
            AddCultureInfoToCurrencyCode("uz-Cyrl", "[$сўм-843]");
            AddCultureInfoToCurrencyCode("uz-Latn", "[$so'm-443]");
            AddCultureInfoToCurrencyCode("vi", "[$₫-42A]");
            AddCultureInfoToCurrencyCode("cy", "[$£-452]");
            AddCultureInfoToCurrencyCode("wo", "[$XOF-488]");
            AddCultureInfoToCurrencyCode("ii-CN", "[$¥-478]");
            AddCultureInfoToCurrencyCode("yo", "[$N-46A]");
        }

        private static void AddCultureInfoToCurrencyCode(string cultureCode, string currencyCode)
        {
            try
            {
                cultureInfoToCurrencyCode.Add(new CultureInfo(cultureCode), currencyCode);
            }
            catch (CultureNotFoundException)
            { }
        }

        #endregion
    }
}