﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Windows.Media;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings
{
    /// <summary>
    /// Represents the cell value format result.
    /// </summary>
    public class CellValueFormatResult
    {
        #region Fields

        private readonly CultureInfo culture;
        private readonly Color? foreground;
        private readonly Predicate<double> condition;
        private readonly IEnumerable<CellValueFormatResultItem> infos;
        private readonly string infosText;
        private readonly string visibleInfosText;
        private readonly string extendedVisibleInfosText;
        private readonly bool containsExpandableInfo;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the culture.
        /// </summary>
        /// <value>The culture.</value>
        public CultureInfo Culture
        {
            get
            {
                return this.culture;
            }
        }

        /// <summary>
        /// Gets the foreground.
        /// </summary>
        /// <value>The foreground.</value>
        public Color? Foreground
        {
            get
            {
                return this.foreground;
            }
        }

        /// <summary>
        /// Gets the condition.
        /// </summary>
        /// <value>The condition.</value>
        public Predicate<double> Condition
        {
            get
            {
                return this.condition;
            }
        }

        /// <summary>
        /// Gets the infos.
        /// </summary>
        /// <value>The infos.</value>
        public IEnumerable<CellValueFormatResultItem> Infos
        {
            get
            {
                return this.infos;
            }
        }

        /// <summary>
        /// Gets the infos text.
        /// </summary>
        /// <value>The infos text.</value>
        public string InfosText
        {
            get
            {
                return this.infosText;
            }
        }

        /// <summary>
        /// Gets the visible infos text.
        /// </summary>
        /// <value>The visible infos text.</value>
        public string VisibleInfosText
        {
            get
            {
                return this.visibleInfosText;
            }
        }

        /// <summary>
        /// Gets the extended visible infos text.
        /// </summary>
        /// <value>The extended visible infos text.</value>
        internal string ExtendedVisibleInfosText
        {
            get
            {
                return this.extendedVisibleInfosText;
            }
        }

        /// <summary>
        /// Gets the contains expandable info.
        /// </summary>
        /// <value>The contains expandable info.</value>
        internal bool ContainsExpandableInfo
        {
            get
            {
                return this.containsExpandableInfo;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="CellValueFormatResult" /> class.
        /// </summary>
        /// <param name="infos">The infos.</param>
        /// <param name="foreground">The foreground.</param>
        /// <param name="condition">The condition.</param>
        /// <param name="culture">The culture.</param>
        public CellValueFormatResult(IEnumerable<CellValueFormatResultItem> infos, Color? foreground, Predicate<double> condition, CultureInfo culture)
        {
            Guard.ThrowExceptionIfNullOrEmpty(infos, "infos");
            Guard.ThrowExceptionIfContainsNull(infos, "infos");

            this.infos = new List<CellValueFormatResultItem>(infos);
            this.infosText = infos.Aggregate(string.Empty, (res, next) => res + next.Text);
            this.visibleInfosText = infos.Aggregate(string.Empty, (res, next) => res + (!next.IsTransparent ? next.Text : string.Empty));
            this.extendedVisibleInfosText = infos.Aggregate(string.Empty, (res, next) => 
            {
                if (next.IsTransparent)
                {
                    return res + SpreadsheetCultureHelper.Space;
                }

                if (next.ShouldExpand)
                {
                    return res;
                }

                return res + next.Text;
            });
            this.containsExpandableInfo = infos.Any(fspi => fspi.ShouldExpand);
            this.foreground = foreground;
            this.condition = condition;
            this.culture = culture;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CellValueFormatResult" /> class.
        /// </summary>
        /// <param name="infos">The infos.</param>
        public CellValueFormatResult(IEnumerable<CellValueFormatResultItem> infos) 
            : this(infos, null, null, null)
        {
        }

        #endregion
    }
}