﻿using System.Globalization;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings
{
    /// <summary>
    /// Represents currency info.
    /// </summary>
    public class CurrencyInfo
    {
        #region Constants

        /// <summary>
        /// Represents none currency info.
        /// </summary>
        public static readonly CurrencyInfo None = new CurrencyInfo("None");

        #endregion


        #region Fields

        private readonly string currencySymbol;
        private readonly string escapedCurrencySymbol;
        private readonly CultureInfo culture;
        private static CurrencyInfo symbol;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the currency symbol.
        /// </summary>
        /// <value>The currency symbol.</value>
        public string CurrencySymbol
        {
            get
            {
                return this.currencySymbol;
            }
        }

        internal string EscapedCurrencySymbol
        {
            get
            {
                return this.escapedCurrencySymbol;
            }
        }

        /// <summary>
        /// Gets the culture.
        /// </summary>
        /// <value>The culture.</value>
        public CultureInfo Culture
        {
            get
            {
                return this.culture;
            }
        }

        /// <summary>
        /// Gets the symbol.
        /// </summary>
        /// <value>The symbol.</value>
        public static CurrencyInfo Symbol
        {
            get
            {
                if (symbol == null ||
                    symbol.CurrencySymbol != FormatHelper.DefaultSpreadsheetCulture.CultureInfo.NumberFormat.CurrencySymbol)
                {
                    symbol = new CurrencyInfo(FormatHelper.DefaultSpreadsheetCulture.CultureInfo.NumberFormat.CurrencySymbol);
                }

                return symbol;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="CurrencyInfo" /> class.
        /// </summary>
        /// <param name="currencySymbol">The currency symbol.</param>
        /// <param name="culture">The culture.</param>
        public CurrencyInfo(string currencySymbol, CultureInfo culture = null)
        {
            // currencySymbol can be string.Empty. For example: new CultureInfo("ps").NumberFormat.CurrencySymbol == "";
            Guard.ThrowExceptionIfNull(currencySymbol, "currencySymbol");

            this.currencySymbol = currencySymbol;
            this.escapedCurrencySymbol = string.Format("\"{0}\"", this.currencySymbol);
            this.culture = culture;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            CurrencyInfo other = obj as CurrencyInfo;
            if (other == null)
            {
                return false;
            }

            return this.CurrencySymbol == other.CurrencySymbol &&
                this.Culture == other.Culture;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(this.CurrencySymbol.GetHashCode(), 
                this.Culture != null ? this.Culture.GetHashCode() : 0);
        }

        #endregion
    }
}