﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings
{
    internal class DateTimeFormatStringParser : TextFormatStringParser
    {
        #region Constants

        private static readonly char HourChar = 'h';
        private static readonly char MinuteChar = 'm';
        private static readonly char SecondsChar = 's';

        #endregion


        #region Constructors

        public DateTimeFormatStringParser(string formatString, FormatStringInfo formatStringInfo)
            : base(formatString, formatStringInfo)
        {
        }

        #endregion


        #region Methods

        private void HandleZeroInDateTimePart(string sbPart, int index, out int newIndex)
        {
            this.GetConsecutiveSymbolRange(sbPart, index, out newIndex);
            int count = newIndex - index + 1;

            this.currentSequence += string.Concat(Enumerable.Repeat<char>(FormatHelper.MilliSecondsChar, count));
        }

        private void HandleAmPm(string sbPart, int index, out int newIndex)
        {
            newIndex = index;
            if (index + 4 < sbPart.Length &&
                (sbPart[index + 1] == 'm') &&
                sbPart[index + 2] == '/' &&
                sbPart[index + 3] == 'p' &&
                (sbPart[index + 1] == 'm'))
            {
                newIndex = index + 4;
                this.currentSequence += "tt";
            }
            else
            {
                this.currentSequence += sbPart[index];
            }
        }

        private void HandleHourSymbol(string sbPart, int index, out int newIndex)
        {
            base.GetConsecutiveSymbolRange(sbPart, index, out newIndex);

            int count = newIndex - index + 1;
            string part = sbPart.ToString().Substring(index, count);
            if (!(sbPart.ToString().Contains("am/pm") || sbPart.ToString().Contains("tt")))
            {
                part = part.ToUpper();
            }

            this.currentSequence += part;
        }

        private void HandleMonthMinuteSymbol(string sbPart, int index, out int newIndex, FormatDescriptor formatDescriptor)
        {
            base.GetConsecutiveSymbolRange(sbPart, index, out newIndex);

            int count = newIndex - index + 1;
            bool isDate = this.IsPositionInDatePart(sbPart, index, newIndex);

            string part = sbPart.Substring(index, count);

            if (isDate && count == 5)
            {
                this.AddCurrentSequenceToItemsAndClear(formatDescriptor);

                formatDescriptor.AddItem(new FormatDescriptorItem(this.ApplyFirstLetterOfMonthToValue, false, false, true));
            }
            else if (isDate)
            {
                this.currentSequence += part.ToUpper();
            }
            else
            {
                this.currentSequence += part.ToLower();
            }
        }

        private string ApplyFirstLetterOfMonthToValue(DateTime? dateTimeValue)
        {
            return this.EscapeString(dateTimeValue.Value.ToString("MMM").Substring(0, 1));
        }

        private string EscapeString(string str)
        {
            return string.Format("\"{0}\"", str);
        }

        private bool IsPositionInDatePart(string sbPart, int start, int end)
        {
            bool isDate = true;
            if (start - 2 >= 0 && SpreadsheetCultureHelper.IsCharEqualTo(sbPart[start - 1], FormatHelper.DefaultSpreadsheetCulture.TimeSeparator) &&
                (sbPart[start - 2] == DateTimeFormatStringParser.HourChar ||
                sbPart[start - 2] == DateTimeFormatStringParser.SecondsChar))
            {
                isDate = false;
            }
            else if (end + 2 < sbPart.Length && SpreadsheetCultureHelper.IsCharEqualTo(sbPart[end + 1], FormatHelper.DefaultSpreadsheetCulture.TimeSeparator) &&
                (sbPart[end + 2] == DateTimeFormatStringParser.SecondsChar))
            {
                isDate = false;
            }

            return isDate;
        }

        private bool TryParseTotalHoursMinutesSeconds(string modificator, FormatDescriptor formatDescriptor)
        {
            Func<DateTime?, string> action = null;

            if (modificator == DateTimeFormatStringParser.HourChar.ToString())
            {
                action = this.GetTotalHours;
            }
            else if (modificator == DateTimeFormatStringParser.MinuteChar.ToString())
            {
                action = this.GetTotalMinutes;
            }
            else if (modificator == DateTimeFormatStringParser.SecondsChar.ToString())
            {
                action = this.GetTotalSeconds;
            }
            else
            {
                throw new ArgumentException("Invalid modificator.");
            }

            this.AddCurrentSequenceToItemsAndClear(formatDescriptor);
            formatDescriptor.AddItem(new FormatDescriptorItem(action, false, false, true));

            return true;
        }

        // NOTE: We subtract days here because the actual start date in Excel is 12/31/1899
        // For the days before 29 feb 1900 we substract 1 day, because of the actual start date and for dates after that
        // we substract 2 days, because in Excel 29 feb 1900 is a valid day, while in .NET it doesn't exist.
        private string GetTotalHours(DateTime? dateTimeValue)
        {
            DateTime leap1900 = new DateTime(1900, 2, 28);
            DateTime startDate = dateTimeValue <= leap1900 ? FormatHelper.StartDate.AddDays(-1) : FormatHelper.StartDate.AddDays(-2);

            TimeSpan timePassedFromStartDate = dateTimeValue.Value.Subtract(startDate);
            return ((Int64)timePassedFromStartDate.TotalHours).ToString();
        }

        private string GetTotalMinutes(DateTime? dateTimeValue)
        {
            DateTime leap1900 = new DateTime(1900, 2, 28);
            DateTime startDate = dateTimeValue <= leap1900 ? FormatHelper.StartDate.AddDays(-1) : FormatHelper.StartDate.AddDays(-2);

            TimeSpan timePassedFromStartDate = dateTimeValue.Value.Subtract(startDate);
            return ((Int64)timePassedFromStartDate.TotalMinutes).ToString();
        }

        private string GetTotalSeconds(DateTime? dateTimeValue)
        {
            DateTime leap1900 = new DateTime(1900, 2, 28);
            DateTime startDate = dateTimeValue <= leap1900 ? FormatHelper.StartDate.AddDays(-1) : FormatHelper.StartDate.AddDays(-2);

            TimeSpan timePassedFromStartDate = dateTimeValue.Value.Subtract(startDate);

            Int64 seconds = (Int64)timePassedFromStartDate.TotalSeconds;
            if (timePassedFromStartDate.Milliseconds > 0.5)
            {
                seconds++;
            }

            return seconds.ToString();
        }

        private void ValidateDateTimeFormatDescriptor(List<FormatDescriptor> formatStringParts)
        {
            if ((formatStringParts == null || formatStringParts.Count == 0) ||
                (formatStringParts.Count > 2) ||
                (formatStringParts.Count == 2 && !formatStringParts[1].Text.Contains(FormatHelper.TextPlaceholder)))
            {
                throw new LocalizableException("Format string is not in the correct format.",
                    new ArgumentException("Format string is not in the correct format."),
                    "Spreadsheet_ErrorExpressions_FormatStringIncorrectFormat");
            }
        }

        #endregion


        #region Overrides

        protected override FormatDescriptor ChooseFormatDescriptor(double? doubleValue, DateTime? dateTimeValue)
        {
            if (dateTimeValue == null && doubleValue != null)
            {
                FormatDescriptor formatDescriptor = new FormatDescriptor();
                formatDescriptor.AddItem(new FormatDescriptorItem("#", false, true));
                return formatDescriptor;
            }

            if (dateTimeValue == null && this.TextFormatDescriptor != null)
            {
                return this.TextFormatDescriptor;
            }

            return this.MainFormatDescriptor;
        }

        protected override string ApplyFormatToString(FormatDescriptorItem item, double? doubleValue, DateTime? dateTimeValue, string stringValue)
        {
            if (doubleValue == null && dateTimeValue == null && !item.Text.Contains(FormatHelper.TextPlaceholder))
            {
                return stringValue;
            }

            if (dateTimeValue == null)
            {
                return base.ApplyFormatToString(item, doubleValue, dateTimeValue, stringValue);
            }

            string format = TextHelper.EscapeInvariantDateTimeSeparators(item.ApplyFormatToValue(dateTimeValue));
            int result;
            if (int.TryParse(format, out result))
            {
                return format;
            }
            else
            {
                return dateTimeValue.Value.ToString(format, FormatHelper.DefaultSpreadsheetCulture.CultureInfo);
            }
        }

        protected override void SetFormatDescriptors(List<FormatDescriptor> formatDescriptors)
        {
            this.ValidateDateTimeFormatDescriptor(formatDescriptors);

            this.MainFormatDescriptor = formatDescriptors[0];
        }

        protected override FormatDescriptor ConstructFormatDescriptor(string sbPart)
        {
            if (sbPart.ToString().Contains(FormatHelper.TextPlaceholder))
            {
                return base.ConstructFormatDescriptor(sbPart);
            }

            sbPart = sbPart.ToLower();

            FormatDescriptor formatDescriptor = new FormatDescriptor();

            for (int i = 0; i < sbPart.Length; i++)
            {
                if (sbPart[i] == '[')
                {
                    this.HandleSquareBracket(sbPart, i, out i, formatDescriptor);
                }
                else if (sbPart[i] == 'h')
                {
                    this.HandleHourSymbol(sbPart, i, out i);
                }
                else if (sbPart[i] == 'm')
                {
                    this.HandleMonthMinuteSymbol(sbPart, i, out i, formatDescriptor);
                }
                else if (sbPart[i] == 'a')
                {
                    this.HandleAmPm(sbPart, i, out i);
                }
                else if (sbPart[i] == '0')
                {
                    this.HandleZeroInDateTimePart(sbPart, i, out i);
                }
                else
                {
                    this.currentSequence += sbPart[i];
                }
            }

            this.AddCurrentSequenceToItemsAndClear(formatDescriptor);
            return formatDescriptor;
        }

        protected override void HandleSquareBracket(string sb, int index, out int newIndex, FormatDescriptor formatDescriptor)
        {
            string modificator = string.Empty;
            newIndex = index;
            bool found = false;

            for (int i = index; i < sb.Length; i++)
            {
                if (sb[i] == ']')
                {
                    modificator = sb.Substring(index + 1, i - index - 1);

                    newIndex = i;
                    found = true;
                    break;
                }
            }

            if (!found)
            {
                return;
            }

            if (base.TryParseSwitchArguments(modificator, formatDescriptor))
            {
                return;
            }

            bool success = base.TryParseCulture(modificator, formatDescriptor);
            if (!success)
            {
                success = this.TryParseTotalHoursMinutesSeconds(modificator, formatDescriptor);
            }
        }

        #endregion
    }
}
