﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings.CategoryManagers;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings.Infos;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings
{
    /// <summary>
    /// Represents the format string category manager class.
    /// </summary>
    public static class FormatStringCategoryManager
    {
        #region Fields

        private static readonly QueueDictionary<FormatStringCategory, Predicate<string>> formatStringCategoryPredicates;

        #endregion


        #region Constructors

        static FormatStringCategoryManager()
        {
            formatStringCategoryPredicates = new QueueDictionary<FormatStringCategory, Predicate<string>>();

            RegisterFormatStringCategoryPredicate(FormatStringCategory.General, BelongsToGeneralCategory);
            RegisterFormatStringCategoryPredicate(FormatStringCategory.Number, BelongsToNumberCategory);
            RegisterFormatStringCategoryPredicate(FormatStringCategory.Currency, BelongsToCurrencyCategory);
            RegisterFormatStringCategoryPredicate(FormatStringCategory.Accounting, BelongsToAccountingCategory);
            RegisterFormatStringCategoryPredicate(FormatStringCategory.Date, BelongsToDateCategory);
            RegisterFormatStringCategoryPredicate(FormatStringCategory.Time, BelongsToTimeCategory);
            RegisterFormatStringCategoryPredicate(FormatStringCategory.Percentage, BelongsToPercentageCategory);
            RegisterFormatStringCategoryPredicate(FormatStringCategory.Fraction, BelongsToFractionCategory);
            RegisterFormatStringCategoryPredicate(FormatStringCategory.Scientific, BelongsToScientificCategory);
            RegisterFormatStringCategoryPredicate(FormatStringCategory.Text, BelongsToTextCategory);
            RegisterFormatStringCategoryPredicate(FormatStringCategory.Special, BelongsToSpecialCategory);
            RegisterFormatStringCategoryPredicate(FormatStringCategory.Custom, BelongsToCustomCategory);
        }

        #endregion


        #region Public API

        /// <summary>
        /// Belongs to format string category.
        /// </summary>
        /// <param name="formatString">The format string.</param>
        /// <param name="category">The category.</param>
        /// <returns></returns>
        public static bool BelongsToFormatStringCategory(string formatString, FormatStringCategory category)
        {
            return formatStringCategoryPredicates[category](formatString);
        }

        /// <summary>
        /// Gets the category from format string.
        /// </summary>
        /// <param name="localFormatString">The local format string.</param>
        /// <returns></returns>
        public static FormatStringCategory GetCategoryFromFormatString(string localFormatString)
        {
            Guard.ThrowExceptionIfNull(localFormatString, "localFormatString");

            foreach (var categoryToPredicate in formatStringCategoryPredicates)
            {
                if (categoryToPredicate.Value(localFormatString))
                {
                    return categoryToPredicate.Key;
                }
            }

            throw new InvalidOperationException();
        }

        #endregion


        #region Methods

        private static void RegisterFormatStringCategoryPredicate(FormatStringCategory category, Predicate<string> predicate)
        {
            formatStringCategoryPredicates.Add(category, predicate);
        }

        #endregion


        #region BelongsTo Methods

        private static bool BelongsToGeneralCategory(string formatString)
        {
            return string.IsNullOrEmpty(formatString);
        }

        private static bool BelongsToNumberCategory(string formatString)
        {
            NumberFormatStringInfo formatStringInfo;
            return TryGetNumberFormatStringInfo(formatString, out formatStringInfo);
        }

        private static bool BelongsToCurrencyCategory(string formatString)
        {
            CurrencyFormatStringInfo formatStringInfo;
            return TryGetCurrencyFormatStringInfo(formatString, out formatStringInfo);
        }

        private static bool BelongsToAccountingCategory(string formatString)
        {
            AccountingFormatStringInfo formatStringInfo;
            return TryGetAccountingFormatStringInfo(formatString, out formatStringInfo);
        }

        private static bool BelongsToDateCategory(string formatString)
        {
            DateFormatStringInfo formatStringInfo;
            return TryGetDateFormatStringInfo(formatString, out formatStringInfo);
        }

        private static bool BelongsToTimeCategory(string formatString)
        {
            TimeFormatStringInfo formatStringInfo;
            return TryGetTimeFormatStringInfo(formatString, out formatStringInfo);
        }

        private static bool BelongsToPercentageCategory(string formatString)
        {
            PercentageFormatStringInfo formatStringInfo;
            return TryGetPercentageFormatStringInfo(formatString, out formatStringInfo);
        }

        private static bool BelongsToFractionCategory(string formatString)
        {
            return FractionFormatCategoryManager.FractionTypeToFormatString.Values.Contains(formatString);
        }

        private static bool BelongsToScientificCategory(string formatString)
        {
            ScientificFormatStringInfo formatStringInfo;
            return TryGetScientificFormatStringInfo(formatString, out formatStringInfo);
        }

        private static bool BelongsToTextCategory(string formatString)
        {
            return formatString == FormatHelper.TextPlaceholder;
        }

        private static bool BelongsToSpecialCategory(string formatString)
        {
            SpecialFormatStringInfo specialFormatStringInfo;
            return TryGetSpecialFormatStringInfo(formatString, out specialFormatStringInfo);
        }

        private static bool BelongsToCustomCategory(string formatString)
        {
            return true;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Tries the get number format string info.
        /// </summary>
        /// <param name="formatString">The format string.</param>
        /// <param name="formatStringInfo">The format string info.</param>
        /// <returns></returns>
        public static bool TryGetNumberFormatStringInfo(string formatString, out NumberFormatStringInfo formatStringInfo)
        {
            return NumberFormatStringInfo.TryCreate(formatString, out formatStringInfo);
        }

        /// <summary>
        /// Tries the get currency format string info.
        /// </summary>
        /// <param name="formatString">The format string.</param>
        /// <param name="formatStringInfo">The format string info.</param>
        /// <returns></returns>
        public static bool TryGetCurrencyFormatStringInfo(string formatString, out CurrencyFormatStringInfo formatStringInfo)
        {
            return CurrencyFormatStringInfo.TryCreate(formatString, out formatStringInfo);
        }

        /// <summary>
        /// Tries the get accounting format string info.
        /// </summary>
        /// <param name="formatString">The format string.</param>
        /// <param name="formatStringInfo">The format string info.</param>
        /// <returns></returns>
        public static bool TryGetAccountingFormatStringInfo(string formatString, out AccountingFormatStringInfo formatStringInfo)
        {
            return AccountingFormatStringInfo.TryCreate(formatString, out formatStringInfo);
        }

        /// <summary>
        /// Tries the get date format string info.
        /// </summary>
        /// <param name="formatString">The format string.</param>
        /// <param name="formatStringInfo">The format string info.</param>
        /// <returns></returns>
        public static bool TryGetDateFormatStringInfo(string formatString, out DateFormatStringInfo formatStringInfo)
        {
            return DateFormatStringInfo.TryCreate(formatString, out formatStringInfo);
        }

        /// <summary>
        /// Tries the get time format string info.
        /// </summary>
        /// <param name="formatString">The format string.</param>
        /// <param name="formatStringInfo">The format string info.</param>
        /// <returns></returns>
        public static bool TryGetTimeFormatStringInfo(string formatString, out TimeFormatStringInfo formatStringInfo)
        {
            return TimeFormatStringInfo.TryCreate(formatString, out formatStringInfo);
        }

        /// <summary>
        /// Tries the get percentage format string info.
        /// </summary>
        /// <param name="formatString">The format string.</param>
        /// <param name="formatStringInfo">The format string info.</param>
        /// <returns></returns>
        public static bool TryGetPercentageFormatStringInfo(string formatString, out PercentageFormatStringInfo formatStringInfo)
        {
            return PercentageFormatStringInfo.TryCreate(formatString, out formatStringInfo);
        }

        /// <summary>
        /// Tries the get scientific format string info.
        /// </summary>
        /// <param name="formatString">The format string.</param>
        /// <param name="formatStringInfo">The format string info.</param>
        /// <returns></returns>
        public static bool TryGetScientificFormatStringInfo(string formatString, out ScientificFormatStringInfo formatStringInfo)
        {
            return ScientificFormatStringInfo.TryCreate(formatString, out formatStringInfo);
        }

        /// <summary>
        /// Tries the get special format string info.
        /// </summary>
        /// <param name="formatString">The format string.</param>
        /// <param name="formatStringInfo">The format string info.</param>
        /// <returns></returns>
        public static bool TryGetSpecialFormatStringInfo(string formatString, out SpecialFormatStringInfo formatStringInfo)
        {
            return SpecialFormatStringInfo.TryCreate(formatString, out formatStringInfo);
        }

        #endregion
    }
}