﻿using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings.Builders;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings.CategoryManagers;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings.Infos
{
    /// <summary>
    /// Represents the currency format string info class.
    /// </summary>
    public class CurrencyFormatStringInfo : ICategorizedFormatStringInfo
    {
        #region Fields

        private readonly int decimalPlaces;
        private readonly CurrencyInfo currencyInfo;
        private readonly List<string> formatStrings;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the decimal places.
        /// </summary>
        /// <value>The decimal places.</value>
        public int DecimalPlaces
        {
            get
            {
                return this.decimalPlaces;
            }
        }

        /// <summary>
        /// Gets the currency info.
        /// </summary>
        /// <value>The currency info.</value>
        public CurrencyInfo CurrencyInfo
        {
            get
            {
                return this.currencyInfo;
            }
        }

        /// <summary>
        /// Gets the format strings.
        /// </summary>
        /// <value>The format strings.</value>
        public IList<string> FormatStrings
        {
            get
            {
                return this.formatStrings;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="CurrencyFormatStringInfo" /> class.
        /// </summary>
        /// <param name="decimalPlaces">The decimal places.</param>
        /// <param name="currencyInfo">The currency info.</param>
        /// <param name="formatStrings">The format strings.</param>
        public CurrencyFormatStringInfo(int decimalPlaces, CurrencyInfo currencyInfo, IEnumerable<string> formatStrings)
        {
            Guard.ThrowExceptionIfLessThan(0, decimalPlaces, "decimalPlaces");
            Guard.ThrowExceptionIfNull(currencyInfo, "currencyInfo");
            Guard.ThrowExceptionIfNull(formatStrings, "formatStrings");

            this.decimalPlaces = decimalPlaces;
            this.currencyInfo = currencyInfo;
            this.formatStrings = new List<string>(formatStrings);
        }

        #endregion


        #region Methods

        internal static bool TryCreate(string formatString, out CurrencyFormatStringInfo formatStringInfo)
        {
            int decimalPlaces;
            if (FormatHelper.TryGetDecimalPlacesForFormatString(formatString,
                FormatHelper.DefaultSpreadsheetCulture.CultureInfo, out decimalPlaces))
            {
                foreach (CurrencyInfo currencyInfo in CurrencyFormatCategoryManager.CurrencyInfos)
                {
                    List<string> formatStrings = CurrencyFormatStringBuilder.BuildFormatStrings(currencyInfo, decimalPlaces);

                    for (int i = 0; i < formatStrings.Count; i++)
                    {
                        if (formatStrings[i].CompareIgnoringSpecialSymbols(formatString))
                        {
                            formatStringInfo = new CurrencyFormatStringInfo(decimalPlaces, currencyInfo, formatStrings);
                            return true;
                        }
                    }
                }
            }

            formatStringInfo = null;
            return false;
        }

        #endregion
    }
}