﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings.Builders;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings
{
    internal class NumberFormatStringParser : TextFormatStringParser
    {
        #region Fields

        private List<FormatDescriptor> ConditionalFormatDescriptors { get; set; }
        private int valueToAdd = 0;

        #endregion


        #region Constructors

        public NumberFormatStringParser(string formatString, FormatStringInfo formatStringInfo)
            : base(formatString, formatStringInfo)
        {
        }

        #endregion


        #region Overrides

        protected override string ApplyFormatToString(FormatDescriptorItem item, double? doubleValue, DateTime? dateTimeValue, string stringValue)
        {
            if (doubleValue == null)
            {
                return base.ApplyFormatToString(item, doubleValue, dateTimeValue, stringValue);
            }

            //the case when we have very big number and we should display it in scientific notation
            if (doubleValue.Value >= FormatHelper.MinimumScientificValue && string.IsNullOrEmpty(item.ApplyFormatToValue(doubleValue)))
            {
                return doubleValue.Value.ToString(this.GetFormatForScientificNumber(doubleValue.Value), FormatHelper.DefaultSpreadsheetCulture.CultureInfo);
            }

            double valueToApplyTo = doubleValue.Value;
            if (this.FormatDescriptor == this.NegativeValueFormatDescriptor)
            {
                valueToApplyTo *= -1;
            }

            valueToApplyTo += this.valueToAdd;

            string returnValue = string.Empty;
            try
            {
                string format = item.ApplyFormatToValue(doubleValue);
                returnValue = valueToApplyTo.ToString(format, FormatHelper.DefaultSpreadsheetCulture.CultureInfo);
            }
            catch (FormatException)
            {
                returnValue = item.ApplyFormatToValue(doubleValue);
            }

            return returnValue;
        }

        protected override FormatDescriptor ChooseFormatDescriptor(double? doubleValue, DateTime? dateTimeValue)
        {
            if (doubleValue == null)
            {
                return this.TextFormatDescriptor;
            }

            if (this.MainFormatDescriptor == null && this.ConditionalFormatDescriptors != null)
            {
                foreach (FormatDescriptor formatDescriptor in this.ConditionalFormatDescriptors)
                {
                    if (formatDescriptor.Condition != null)
                    {
                        if (formatDescriptor.Condition(doubleValue.Value))
                        {
                            return formatDescriptor;
                        }
                    }
                    else
                    {
                        return formatDescriptor;
                    }
                }
            }

            if (doubleValue.Value == 0 && this.ZeroValueFormatDescriptor != null)
            {
                return this.ZeroValueFormatDescriptor;
            }

            if (doubleValue.Value < 0 && this.NegativeValueFormatDescriptor != null)
            {
                return this.NegativeValueFormatDescriptor;
            }

            if (this.MainFormatDescriptor != null)
            {
                return this.MainFormatDescriptor;
            }

            throw new InvalidFormatStringException();
        }

        protected override void SetFormatDescriptors(List<FormatDescriptor> formatDescriptors)
        {
            this.ValidateNumberFormatDescriptors(formatDescriptors);

            if (formatDescriptors.Count == 0)
            {
                return;
            }

            if (formatDescriptors[0].Condition != null)
            {
                this.ConditionalFormatDescriptors = formatDescriptors;
                return;
            }

            if (formatDescriptors.Count >= 1)
            {
                this.MainFormatDescriptor = formatDescriptors[0];
            }

            if (formatDescriptors.Count >= 2)
            {
                this.NegativeValueFormatDescriptor = formatDescriptors[1];
            }

            if (formatDescriptors.Count >= 3)
            {
                this.ZeroValueFormatDescriptor = formatDescriptors[2];
            }
        }

        protected override FormatDescriptor ConstructFormatDescriptor(string sbPart)
        {
            if (sbPart.Contains(FormatHelper.TextPlaceholder))
            {
                return base.ConstructFormatDescriptor(sbPart);
            }

            FormatDescriptor formatDescriptor = new FormatDescriptor();

            if (sbPart.Length == 0)
            {
                formatDescriptor.AddItem(new FormatDescriptorItem(string.Empty, false));
            }
            else
            {
                string currencySymbol = FormatHelper.DefaultSpreadsheetCulture.CultureInfo.NumberFormat.CurrencySymbol;
                int[] currencySymbolIndexes = FormatHelper.FindAllOccurrences(sbPart, currencySymbol);

                string invisibleCurrencySymbol = BuildersHelper.MakeCurrencySymbolInvisible(currencySymbol);
                int[] invisibleCurrencySymbolIndexes = FormatHelper.FindAllOccurrences(sbPart, invisibleCurrencySymbol);

                bool isLiteral = false;

                for (int i = 0; i < sbPart.Length; i++)
                {
                    char currentChar = sbPart[i];
                    if (currentChar == '\"')
                    {
                        isLiteral = !isLiteral;
                        this.currentSequence += currentChar;
                        continue;
                    }

                    if (isLiteral)
                    {
                        this.currentSequence += currentChar;
                        continue;
                    }

                    if (FormatHelper.CheckIfIsCurrencySymbolPart(currencySymbolIndexes, currencySymbol, i) ||
                        FormatHelper.CheckIfIsCurrencySymbolPart(invisibleCurrencySymbolIndexes, invisibleCurrencySymbol, i))
                    {
                        if (currentChar == '_')
                        {
                            this.HandleUnderscore(sbPart, i, out i, formatDescriptor, true);
                        }
                        else if (i > 0 && !(sbPart[i - 1] == '\\'))
                        {
                            this.currentSequence += "\\" + currentChar;
                        }
                        else
                        {
                            this.currentSequence += currentChar;
                        }

                        continue;
                    }

                    if (currentChar == '_')
                    {
                        this.HandleUnderscore(sbPart, i, out i, formatDescriptor);
                    }
                    else if (currentChar == '[')
                    {
                        this.HandleSquareBracket(sbPart, i, out i, formatDescriptor);
                    }
                    else if (currentChar == '*')
                    {
                        this.HandleStar(sbPart, i, out i, formatDescriptor);
                    }
                    else if (currentChar == '?')
                    {
                        this.HandleQuestionMarkSymbol(sbPart, i, out i, formatDescriptor);
                    }
                    //the case when the currency symbol contains dot('.') or zero('0') 
                    else if (currentChar == '.' && i - 1 > 0 && sbPart[i - 1] != '#' && sbPart[i - 1] != '0')
                    {
                        this.AddCurrentSequenceToItemsAndClear(formatDescriptor, true);
                        this.currentSequence += currentChar;
                        this.AddCurrentSequenceToItemsAndClear(formatDescriptor, false);
                    }
                    else
                    {
                        this.currentSequence += currentChar;
                    }
                }
            }

            this.AddCurrentSequenceToItemsAndClear(formatDescriptor);

            return formatDescriptor;
        }

        protected override void HandleSquareBracket(string sb, int index, out int newIndex, FormatDescriptor formatDescriptor)
        {
            string modificator = string.Empty;
            newIndex = index;
            bool found = false;
            for (int i = index; i < sb.Length; i++)
            {
                if (sb[i] == ']')
                {
                    modificator = sb.Substring(index + 1, i - index - 1);

                    newIndex = i;
                    found = true;
                    break;
                }
            }

            if (!found)
            {
                this.currentSequence += sb[index];
                return;
            }

            bool success = this.TryParseColor(modificator, formatDescriptor);
            if (!success)
            {
                success = this.TryParseCondition(modificator, formatDescriptor);
            }
            if (!success)
            {
                success = this.TryParseSymbolCulture(modificator, formatDescriptor);
            }
            if (!success)
            {
                success = this.TryParseCulture(modificator, formatDescriptor);
            }
        }

        #endregion


        #region Methods

        private void HandleStar(string sbPart, int index, out int newIndex, FormatDescriptor formatDescriptor)
        {
            newIndex = index;
            if (index + 1 < sbPart.Length)
            {
                this.AddCurrentSequenceToItemsAndClear(formatDescriptor);
                newIndex = index + 1;
                formatDescriptor.AddItem(new FormatDescriptorItem(sbPart[newIndex].ToString(), false, true));
            }
        }

        private void HandleQuestionMarkSymbol(string sbPart, int index, out int newIndex, FormatDescriptor formatDescriptor)
        {
            this.GetConsecutiveSymbolRange(sbPart, index, out newIndex);
            int count = newIndex - index + 1;
            int nominatorRequiredLength = count;

            //this is the case where we have a zero value and it should be raplaced(most times) by dash. The dash should be aligned with the other values
            //and that's why we make the question marks invisible
            if (!(newIndex + 1 < sbPart.Length) || sbPart[newIndex + 1] != '/')
            {
                this.AddCurrentSequenceToItemsAndClear(formatDescriptor);
                string text = string.Concat(Enumerable.Repeat(sbPart[index], count));
                formatDescriptor.AddItem(new FormatDescriptorItem(text, true));
            }
            //this is the case when we have fractions and we should calculate them
            else
            {
                Debug.Assert(sbPart[newIndex + 1] == '/');
                int newStartIndex = newIndex + 2;

                string upToDigitsString = string.Empty;
                int denominatorRequiredLength = -1;
                bool isNextSymbolQuestionMark = sbPart[newStartIndex] == '?';

                if (isNextSymbolQuestionMark)
                {
                    this.GetConsecutiveSymbolRange(sbPart, newStartIndex, out newIndex);
                }
                else
                {
                    this.GetConsecutiveDigitRange(sbPart, newStartIndex, out newIndex);
                }
                int newIndexCopy = newIndex;

                Func<double?, string, IEnumerable<FormatDescriptorItem>> action =
                    new Func<double?, string, IEnumerable<FormatDescriptorItem>>((doubleValue, startSequence) =>
                    {
                        List<FormatDescriptorItem> items = new List<FormatDescriptorItem>();

                        double whole = FormatHelper.Round(doubleValue.Value);
                        double rest = doubleValue.Value - whole;

                        if (isNextSymbolQuestionMark)
                        {
                            int count1 = newIndexCopy - newStartIndex + 1;
                            denominatorRequiredLength = count1;
                            upToDigitsString = this.GetUpToDigits(Math.Abs(rest), count1);
                        }
                        else
                        {
                            denominatorRequiredLength = newIndexCopy - newStartIndex + 1;
                            string numberAsString = sbPart.Substring(newStartIndex, denominatorRequiredLength);
                            upToDigitsString = this.GetFractionStringByDenominator(Math.Abs(rest), int.Parse(numberAsString), true);
                        }

                        //if the approximate fraction is 0 or 1
                        if (upToDigitsString.IndexOf('/') == -1)
                        {
                            //the case when we have to display only zero(the "#" formatting string doesn't display anything if value is smaller than 0.5)
                            if (upToDigitsString == "0" && whole == 0)
                            {
                                startSequence = startSequence.Replace('#', '0');
                            }

                            if (!string.IsNullOrEmpty(startSequence))
                            {
                                items.Add(new FormatDescriptorItem(startSequence, false));
                                startSequence = string.Empty;
                            }

                            //here we calculate how long the invisiblefraction has to be(so the number can be aligned to the other fractions)
                            string left = string.Concat(Enumerable.Repeat(FormatHelper.FractionsInvisibleChar, nominatorRequiredLength));
                            string right = string.Concat(Enumerable.Repeat(FormatHelper.FractionsInvisibleChar, denominatorRequiredLength));
                            //we use 1 as the universal number width (i assume every number can be used)
                            items.Add(new FormatDescriptorItem(left + "/" + right, true, false, false));
                        }
                        else
                        {
                            string stringToAdd = upToDigitsString;
                            int dashIndex = stringToAdd.IndexOf('/');
                            string nominator = stringToAdd.Substring(0, dashIndex);
                            string denominator = stringToAdd.Substring(dashIndex + 1);

                            if (rest >= 0.5d && whole > 0)
                            {
                                this.valueToAdd = -1;
                            }
                            else if (rest <= -0.5d)
                            {
                                this.valueToAdd = 1;
                            }

                            //this is the case where we have only a fraction without a whole number. Then we clear the "# " before it
                            if (whole == 0 && rest > 0)
                            {
                                startSequence = string.Empty;
                            }
                            //the case when the number is negative, we just add the minus sign
                            else if (whole == 0 && rest < 0)
                            {
                                startSequence = "- ";
                            }

                            if (nominator.Length < nominatorRequiredLength || denominator.Length < denominatorRequiredLength)
                            {
                                if (!string.IsNullOrEmpty(startSequence))
                                {
                                    items.Add(new FormatDescriptorItem(startSequence, false));
                                    startSequence = string.Empty;
                                }

                                //we use 1 as the universal number width (i assume every number can be used)
                                if (nominator.Length < nominatorRequiredLength)
                                {
                                    string ones = string.Concat(Enumerable.Repeat(FormatHelper.FractionsInvisibleChar,
                                        nominatorRequiredLength - nominator.Length));
                                    items.Add(new FormatDescriptorItem(ones, true, false, false));
                                }

                                items.Add(new FormatDescriptorItem(stringToAdd, false, false, false));
                                if (denominator.Length < denominatorRequiredLength)
                                {
                                    string ones = string.Concat(Enumerable.Repeat(FormatHelper.FractionsInvisibleChar,
                                        denominatorRequiredLength - denominator.Length));
                                    items.Add(new FormatDescriptorItem(ones, true, false, false));
                                }
                            }
                            else
                            {
                                if (!string.IsNullOrEmpty(startSequence))
                                {
                                    items.Add(new FormatDescriptorItem(startSequence, false));
                                    startSequence = string.Empty;
                                }

                                items.Add(new FormatDescriptorItem(stringToAdd, false, false, false));
                            }
                        }

                        if (!string.IsNullOrEmpty(startSequence))
                        {
                            items.Add(new FormatDescriptorItem(startSequence, false, false, false));
                            startSequence = string.Empty;
                        }

                        return items;
                    });

                formatDescriptor.AddItem(new FormatDescriptorItemComposite(action, this.currentSequence));
                this.currentSequence = string.Empty;
            }
        }

        private string GetFractionStringByDenominator(double value, int denom, bool isResultStringEmpty)
        {
            int valueToMultiplyBy = denom.ToString().Length;

            double multipliedValue = value * denom;
            double roundedNumber = Math.Round(multipliedValue, valueToMultiplyBy);

            double smallestInterval = 1d / (denom * 2);
            if (value < smallestInterval)
            {
                if (!isResultStringEmpty)
                {
                    return "";
                }
                return "0";
            }
            else if (value >= 1d - smallestInterval)
            {
                if (!isResultStringEmpty)
                {
                    return "";
                }
                return "1";
            }

            //this is needed because 0.5 is rounded to 0 instead of 1
            if (roundedNumber - 0.5d < 0.000001d) roundedNumber += 0.1d;

            return Math.Round(roundedNumber).ToString() + "/" + denom.ToString();
        }

        protected void GetConsecutiveDigitRange(string sbPart, int index, out int newIndex)
        {
            newIndex = index;
            bool stopped = false;
            for (int i = index; i < sbPart.Length; i++)
            {
                if (!char.IsDigit(sbPart[i]))
                {
                    newIndex = i - 1;
                    stopped = true;
                    break;
                }
            }

            if (!stopped)
            {
                newIndex = sbPart.Length - 1;
            }
        }

        private string GetUpToDigits(double number, int upTo)
        {
            Tuple<int, int> selectedTuple = null;
            List<Tuple<int, int>> fractions = this.GetApproximateFractions((double)number);
            for (int i = fractions.Count - 1; i >= 0; i--)
            {
                if (fractions[i].Item2.ToString().Length <= upTo)
                {
                    selectedTuple = fractions[i];
                    break;
                }
            }

            if (selectedTuple.Item1 == 0)
            {
                return "0";
            }
            else if (selectedTuple.Item1 == selectedTuple.Item2)
            {
                return "1";
            }

            return selectedTuple.Item1.ToString() + "/" + selectedTuple.Item2.ToString();
        }

        private List<Tuple<int, int>> GetApproximateFractions(double number)
        {
            List<Tuple<int, int>> result = new List<Tuple<int, int>>();
            double d = number;

            double[] numerators = new double[1000];
            numerators[0] = 0;
            numerators[1] = 1;
            double[] denominators = new double[1000];
            denominators[0] = 1;
            denominators[1] = 0;

            var maxNumerator = GetMaxNumerator(d);
            var d2 = d;
            double calcD, prevCalcD = -1;
            for (var i = 2; i < 1000; i++)
            {
                var l2 = Math.Floor(d2);
                numerators[i] = l2 * numerators[i - 1] + numerators[i - 2];
                if (Math.Abs(numerators[i]) > maxNumerator)
                {
                    return result;
                }

                denominators[i] = l2 * denominators[i - 1] + denominators[i - 2];

                calcD = numerators[i] / denominators[i];
                if (calcD == prevCalcD)
                {
                    return result;
                }

                result.Add(new Tuple<int, int>((int)numerators[i], (int)denominators[i]));

                if (calcD == d)
                {
                    return result;
                }

                prevCalcD = calcD;

                d2 = 1 / (d2 - l2);
            }

            return result;
        }

        private UInt64 GetMaxNumerator(double f)
        {
            string f2 = null;
            string fToString = f.ToString(FormatHelper.DefaultSpreadsheetCulture.CultureInfo);
            int ixe = fToString.IndexOf("E");
            if (ixe == -1)
            {
                ixe = fToString.IndexOf("e");
            }
            if (ixe == -1)
            {
                f2 = fToString;
            }
            else
            {
                f2 = fToString.Substring(0, ixe);
            }

            string digits = null;
            string separator = FormatHelper.DefaultSpreadsheetCulture.NumberDecimalSeparator;
            var ix = f2.ToString(FormatHelper.DefaultSpreadsheetCulture.CultureInfo).IndexOf(separator);
            if (ix == -1)
            {
                digits = f2;
            }
            else if (ix == 0)
            {
                digits = f2.Substring(1, f2.Length);
            }
            else if (ix < f2.Length)
            {
                int indexOfDot = ix + 1;
                digits = f2.Substring(0, ix) + f2.Substring(indexOfDot, f2.Length - (indexOfDot));
            }

            var l = digits;

            var numDigits = l.ToString().Length;
            var l2 = f;
            var numIntDigits = l2.ToString(FormatHelper.DefaultSpreadsheetCulture.CultureInfo).Length;
            if (l2 == 0)
            {
                numIntDigits = 0;
            }
            var numDigitsPastDecimal = numDigits - numIntDigits;

            UInt64 l1 = UInt64.Parse(l);

            for (var i = numDigitsPastDecimal; i > 0 && l1 % 2 == 0; i--)
            {
                l1 /= 2;
            }
            for (var i = numDigitsPastDecimal; i > 0 && l1 % 5 == 0; i--)
            {
                l1 /= 5;
            }

            return l1;
        }

        private bool TryParseSymbolCulture(string modificator, FormatDescriptor formatDescriptor)
        {
            if (string.IsNullOrEmpty(modificator) || modificator[0] != '$')
            {
                return false;
            }

            if (modificator.Length == 4 && !modificator.Contains('-'))
            {
                this.currentSequence += modificator.Substring(1);

                return true;
            }

            int dashIndex = modificator.IndexOf('-');
            if (dashIndex != -1 && dashIndex != 1)
            {
                string textToAdd = modificator.Substring(1, dashIndex - 1);

                if (textToAdd.Contains('.') || textToAdd.Contains('\'') || Regex.IsMatch(textToAdd, FormatHelper.IsRtlRegExPattern) ||
                    textToAdd.Equals("R"))
                {
                    this.AddCurrentSequenceToItemsAndClear(formatDescriptor);
                    formatDescriptor.AddItem(new FormatDescriptorItem(textToAdd, false, false, false));
                }
                else
                {
                    this.currentSequence += textToAdd;
                }

                //string culture = modificator.Substring(dashIndex);
                //base.TryParseCulture("$" + culture, formatDescriptor, false);

                return true;
            }

            return false;
        }

        private bool TryParseCondition(string modificator, FormatDescriptor formatDescriptor)
        {
            StringBuilder sb = new StringBuilder(modificator);
            decimal number;
            int index = 0;
            if ((index = modificator.IndexOf("<=")) > -1)
            {
                string numStr = sb.Remove(index, 2).ToString();
                if (decimal.TryParse(numStr, out number))
                {
                    formatDescriptor.Condition =
                        (num) =>
                        {
                            return (decimal)num <= number;
                        };
                    return true;
                }
                return false;
            }
            else if ((index = modificator.IndexOf(">=")) > -1)
            {
                string numStr = sb.Remove(index, 2).ToString();
                if (decimal.TryParse(numStr, out number))
                {
                    formatDescriptor.Condition =
                        (num) =>
                        {
                            return (decimal)num >= number;
                        };
                    return true;
                }
                return false;
            }
            else if ((index = modificator.IndexOf(">")) > -1)
            {
                string numStr = sb.Remove(index, 1).ToString();
                if (decimal.TryParse(numStr, out number))
                {
                    formatDescriptor.Condition =
                        (num) =>
                        {
                            return (decimal)num > number;
                        };
                    return true;
                }
                return false;
            }
            else if ((index = modificator.IndexOf("<")) > -1)
            {
                string numStr = sb.Remove(index, 1).ToString();
                if (decimal.TryParse(numStr, out number))
                {
                    formatDescriptor.Condition =
                        (num) =>
                        {
                            return (decimal)num < number;
                        };
                    return true;
                }
                return false;
            }
            else if ((index = modificator.IndexOf("=")) > -1)
            {
                string numStr = sb.Remove(index, 1).ToString();
                if (decimal.TryParse(numStr, out number))
                {
                    formatDescriptor.Condition =
                        (num) =>
                        {
                            return (decimal)num == number;
                        };
                    return true;
                }
                return false;
            }

            return false;
        }

        //Here we replace the culture-specific CurrencyGroupSeparator and CurrencyDecimalSeparator with comma(',') and dot('.'). We need this because 
        //in some cases the culture-specific CurrencyGroupSeparator and CurrencyDecimalSeparator are different from dot('.') and comma(','). This
        //is a problem because the custom formatting engine in .NET accepts only those two as separators. 
        private void EscapeItemsText(FormatDescriptor formatDescriptor)
        {
            CultureInfo ci = formatDescriptor.Culture;
            if (ci == null)
            {
                ci = Thread.CurrentThread.CurrentCulture;
            }
            foreach (FormatDescriptorItem item in formatDescriptor.Items.Where((fpi) => { return fpi.ApplyFormat; }))
            {
                item.Text = item.Text.Replace(ci.NumberFormat.CurrencyGroupSeparator, ",").Replace(ci.NumberFormat.CurrencyDecimalSeparator, ".");
            }
        }

        private string GetFormatForScientificNumber(double number)
        {
            StringBuilder format = new StringBuilder("0");
            string numberString = number.ToString();
            int decimalPlaces = 0;
            for (int i = 1; i < numberString.Length; i++)
            {
                double currentDigit = char.GetNumericValue(numberString[i]);
                if (currentDigit == 0 || i == numberString.Length - 1 || i == 6)
                {
                    decimalPlaces = i - 1;
                    break;
                }
            }

            if (decimalPlaces > 0)
            {
                if (decimalPlaces >= 5)
                {
                    decimalPlaces = 5;
                }
                format.Append(".");
                format.Append(string.Concat(Enumerable.Repeat("0", decimalPlaces)));
            }
            format.Append("E+00");

            return format.ToString();
        }

        private void ValidateNumberFormatDescriptors(List<FormatDescriptor> formatStringParts)
        {
            string errorKey = "Spreadsheet_ErrorExpressions_FormatStringIncorrectFormat";
            string errorString = "Format string is not in the correct format.";

            if ((formatStringParts.Count == 1 && formatStringParts[0].Condition != null) ||
                formatStringParts.Count > 4)
            {
                throw new LocalizableException(errorString, new ArgumentException(errorString), errorKey);
            }

            if (formatStringParts.Count > 1 && formatStringParts[0].Condition != null)
            {
                for (int i = formatStringParts.Count - 2; i >= 0; i--)
                {
                    if (formatStringParts[i].Condition == null)
                    {
                        throw new LocalizableException(errorString, new ArgumentException(errorString), errorKey);
                    }
                }
            }

            if (formatStringParts.Count > 1 && formatStringParts[0].Condition == null)
            {
                for (int i = 1; i < formatStringParts.Count; i++)
                {
                    if (formatStringParts[i].Condition != null)
                    {
                        throw new LocalizableException(errorString, new ArgumentException(errorString), errorKey);
                    }
                }
            }

            if (formatStringParts.Count > 1 && formatStringParts[formatStringParts.Count - 1].Text.Contains(FormatHelper.TextPlaceholder))
            {
                for (int i = 0; i < formatStringParts.Count - 1; i++)
                {
                    if (formatStringParts[i].Text.Contains(FormatHelper.TextPlaceholder))
                    {
                        throw new LocalizableException(errorString, new ArgumentException(errorString), errorKey);
                    }
                }
            }
        }

        #endregion
    }
}
