﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Media;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings
{
    internal class TextFormatStringParser
    {
        #region Constants

        private static readonly string DBNUM1 = "dbnum1";

        #endregion


        #region Fields

        private readonly string formatString;
        protected string currentSequence;
        protected FormatDescriptor formatDescriptor;
        private readonly FormatStringInfo formatStringInfo;

        #endregion


        #region Properties

        public FormatStringInfo FormatStringInfo
        {
            get
            {
                return this.formatStringInfo;
            }
        }

        internal string FormatString
        {
            get
            {
                return this.formatString;
            }
        }

        internal FormatDescriptor FormatDescriptor
        {
            get
            {
                return this.formatDescriptor;
            }
        }

        protected FormatDescriptor MainFormatDescriptor { get; set; }
        protected FormatDescriptor NegativeValueFormatDescriptor { get; set; }
        protected FormatDescriptor ZeroValueFormatDescriptor { get; set; }
        protected FormatDescriptor TextFormatDescriptor { get; set; }

        #endregion


        #region Constructors

        public TextFormatStringParser(string formatString, FormatStringInfo formatStringInfo)
        {
            this.formatString = formatString;
            this.formatStringInfo = formatStringInfo;
            List<FormatDescriptor> formatDescriptors = new List<FormatDescriptor>();

            string[] parts = this.formatString.Split(new char[] { ';' });
            foreach (string part in parts)
            {
                FormatDescriptor formatDescriptor = this.ConstructFormatDescriptor(part);
                if (formatDescriptor != null)
                {
                    formatDescriptors.Add(formatDescriptor);
                }
            }

            FormatDescriptor lastFormatDescriptor = formatDescriptors.Count > 0 ? formatDescriptors.Last() : null;
            if (lastFormatDescriptor != null && lastFormatDescriptor.Text.Contains(FormatHelper.TextPlaceholder))
            {
                this.TextFormatDescriptor = lastFormatDescriptor;
                formatDescriptors.Remove(lastFormatDescriptor);
            }

            bool isGeneralFormatDescriptor = formatDescriptors.Count == 1 && lastFormatDescriptor.ItemsCount == 1 && 
                lastFormatDescriptor.Text.Equals(FormatHelper.GeneralFormatName, StringComparison.OrdinalIgnoreCase);

            if (isGeneralFormatDescriptor)
            {
                lastFormatDescriptor.Items.First().Text = FormatHelper.GeneralFormatString;
            }

            this.SetFormatDescriptors(formatDescriptors);
        }

        #endregion


        #region Virtual Methods

        protected virtual string ApplyFormatToString(FormatDescriptorItem item, double? doubleValue, DateTime? dateTimeValue, string stringValue)
        {
            return string.Format(item.GetTextReplacePlaceholder(), stringValue);
        }

        protected virtual void HandleSquareBracket(string sb, int index, out int newIndex, FormatDescriptor formatDescriptor)
        {
            string modificator = string.Empty;
            newIndex = index;
            bool found = false;
            for (int i = index; i < sb.Length; i++)
            {
                if (sb[i] == ']')
                {
                    modificator = sb.Substring(index + 1, i - index - 1);

                    newIndex = i;
                    found = true;
                    break;
                }
            }

            if (!found)
            {
                this.currentSequence += sb[index];
                return;
            }

            this.TryParseColor(modificator, formatDescriptor);
        }

        protected virtual void SetFormatDescriptors(List<FormatDescriptor> formatDescriptors)
        {
        }

        protected virtual FormatDescriptor ChooseFormatDescriptor(double? doubleValue, DateTime? dateTimeValue)
        {
            return this.TextFormatDescriptor;
        }

        protected virtual FormatDescriptor ConstructFormatDescriptor(string sbPartString)
        {
            if (sbPartString.Contains(FormatHelper.TextPlaceholder))
            {
                FormatDescriptor formatDescriptor = new FormatDescriptor();

                for (int i = 0; i < sbPartString.Length; i++)
                {
                    if (sbPartString[i] == '_')
                    {
                        this.HandleUnderscore(sbPartString, i, out i, formatDescriptor);
                    }
                    else if (sbPartString[i] == '[')
                    {
                        this.HandleSquareBracket(sbPartString, i, out i, formatDescriptor);
                    }
                    else
                    {
                        this.currentSequence += sbPartString[i];
                    }
                }

                this.AddCurrentSequenceToItemsAndClear(formatDescriptor);
                return formatDescriptor;
            }

            return null;
        }

        #endregion


        #region Methods

        public List<CellValueFormatResultItem> ApplyFormatToValues(ICellValue cellValue, CellValueFormat format)
        {
            double? doubleValue = null;
            DateTime? dateTimeValue = null;
            string stringValue = cellValue.GetValueAsString(format);

            NumberCellValue numberCellValue = cellValue as NumberCellValue;
            if (numberCellValue != null)
            {
                doubleValue = numberCellValue.Value;

                string strippedFormatString = FormatHelper.StripBrackets(this.FormatString);
                dateTimeValue = numberCellValue.ToDateTime();
                if (dateTimeValue != null && format.FormatStringInfo.FormatType == FormatStringType.DateTime && !strippedFormatString.Contains("0"))
                {
                    dateTimeValue = dateTimeValue.Value.RoundMinutes();
                }
            }

            List<CellValueFormatResultItem> result = new List<CellValueFormatResultItem>();

            this.formatDescriptor = this.ChooseFormatDescriptor(doubleValue, dateTimeValue);
            if (this.FormatDescriptor == null)
            {
                result.Add(new CellValueFormatResultItem(stringValue, false, false, true));
                return result;
            }

            foreach (FormatDescriptorItem item in this.FormatDescriptor.Items.Reverse())
            {
                foreach (FormatDescriptorItem item1 in item.GetItems(doubleValue).Reverse())
                {
                    if (item1.ApplyFormat)
                    {
                        result.Insert(0, new CellValueFormatResultItem(this.ApplyFormatToString(item1, doubleValue, dateTimeValue, stringValue),
                            item1.IsTransparent, item1.ShouldExpand, item1.ApplyFormat));
                    }
                    else
                    {
                        result.Insert(0, new CellValueFormatResultItem(item1.Text, item1.IsTransparent, item1.ShouldExpand, item1.ApplyFormat));
                    }
                }
            }

            return result;
        }

        protected void HandleUnderscore(string sb, int index, out int newIndex, FormatDescriptor formatDescriptor, bool isCurrency = false)
        {
            newIndex = index;
            if (!(index + 1 < sb.Length))
            {
                return;
            }

            this.AddCurrentSequenceToItemsAndClear(formatDescriptor);
            newIndex = index + 1;

            char nextChar = sb[index + 1];
            bool applyFormat = isCurrency ? false : nextChar != '.';

            formatDescriptor.AddItem(new FormatDescriptorItem(nextChar.ToString(), true, false, applyFormat));
        }

        protected bool TryParseColor(string modificator, FormatDescriptor formatDescriptor)
        {
            switch (modificator)
            {
                case "Red":
                    formatDescriptor.Foreground = Colors.Red;
                    return true;
                case "Blue":
                    formatDescriptor.Foreground = Colors.Blue;
                    return true;
                case "Green":
                    formatDescriptor.Foreground = Colors.Green;
                    return true;
                case "White":
                    formatDescriptor.Foreground = Colors.White;
                    return true;
                case "Magenta":
                    formatDescriptor.Foreground = Colors.Magenta;
                    return true;
                case "Yellow":
                    formatDescriptor.Foreground = Colors.Yellow;
                    return true;
                case "Cyan":
                    formatDescriptor.Foreground = Colors.Cyan;
                    return true;
                default:
                    return false;
            }
        }

        protected bool TryParseCulture(string modificator, FormatDescriptor formatDescriptor, bool getDefault = true)
        {
            if (string.IsNullOrEmpty(modificator) || modificator.IndexOf("$-") == -1)
            {
                return false;
            }

            if (getDefault)
            {
                formatDescriptor.Culture = LocaleCodeToCultureInfoResolver.GetCultureOrDefault(modificator);
            }
            else
            {
                formatDescriptor.Culture = LocaleCodeToCultureInfoResolver.GetCulture(modificator);
            }

            return true;
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Usage", "CA1801:ReviewUnusedParameters", MessageId = "formatDescriptor")]
        public bool TryParseSwitchArguments(string modificator, FormatDescriptor formatDescriptor)
        {
            if (modificator == DBNUM1)
            {
                return true;
            }

            return false;
        }

        protected void AddCurrentSequenceToItemsAndClear(FormatDescriptor formatDescriptor, bool applyFormatting = true)
        {
            if (!string.IsNullOrEmpty(this.currentSequence))
            {
                formatDescriptor.AddItem(new FormatDescriptorItem(this.currentSequence, false, false, applyFormatting));
                this.currentSequence = string.Empty;
            }
        }

        protected void GetConsecutiveSymbolRange(string sbPart, int index, out int newIndex)
        {
            newIndex = index;
            char currentSymbol = sbPart[index];
            bool stopped = false;
            for (int i = index; i < sbPart.Length; i++)
            {
                if (sbPart[i] != currentSymbol)
                {
                    newIndex = i - 1;
                    stopped = true;
                    break;
                }
            }

            if (!stopped)
            {
                newIndex = sbPart.Length - 1;
            }
        }

        #endregion


        #region Static Methods

        public static TextFormatStringParser Create(string invariantFormatString, string localizedFormatString)
        {
            Guard.ThrowExceptionIfNull(invariantFormatString, "invariantFormatString");
            Guard.ThrowExceptionIfNull(localizedFormatString, "localizedFormatString");

            FormatStringInfo formatStringInfo = RetrieveFormatStringInfo(invariantFormatString, localizedFormatString);

            switch (formatStringInfo.FormatType)
            {
                case FormatStringType.Number:
                    return new NumberFormatStringParser(invariantFormatString, formatStringInfo);

                case FormatStringType.DateTime:
                    return new DateTimeFormatStringParser(invariantFormatString, formatStringInfo);

                case FormatStringType.Text:
                    return new TextFormatStringParser(invariantFormatString, formatStringInfo);

                default:
                    return new NumberFormatStringParser(invariantFormatString, formatStringInfo);
            }
        }

        private static FormatStringInfo RetrieveFormatStringInfo(string invariantFormatString, string localizedFormatString)
        {
            Guard.ThrowExceptionIfNull(invariantFormatString, "invariantFormatString");
            Guard.ThrowExceptionIfNull(localizedFormatString, "localizedFormatString");

            FormatStringInfo formatStringInfo = new FormatStringInfo(null, FormatStringCategoryManager.GetCategoryFromFormatString(localizedFormatString), false, false);

            if (formatStringInfo.Category == FormatStringCategory.Custom ||
                formatStringInfo.Category == FormatStringCategory.Date ||
                formatStringInfo.Category == FormatStringCategory.Time)
            {
                if (FormatHelper.ContainsTotalHoursMinutesSecondsModifiers(invariantFormatString))
                {
                    formatStringInfo.FormatType = FormatStringType.DateTime;
                }
            }

            //the case when we input a simple double value
            if (string.IsNullOrEmpty(invariantFormatString) && formatStringInfo.FormatType == null)
            {
                formatStringInfo.FormatType = FormatStringType.Number;
                return formatStringInfo;
            }

            string[] parts = invariantFormatString.Split(';');

            if (parts.Count() == 1 && parts[0].Contains(FormatHelper.TextPlaceholder))
            {
                formatStringInfo.FormatType = FormatStringType.Text;
                return formatStringInfo;
            }

            for (int i = 0; i < parts.Length; i++)
            {
                GetPartInfo(parts[i], formatStringInfo);
            }

            return formatStringInfo;
        }

        private static void GetPartInfo(string part, FormatStringInfo formatStringInfo)
        {
            if (part.Contains(FormatHelper.TextPlaceholder))
            {
                return;
            }

            if (!string.IsNullOrEmpty(part))
            {
                part = FormatHelper.StripBrackets(part);
                part = FormatHelper.StripUnderscores(part);
                part = FormatHelper.StripCurrencyEscapingsAndSlashes(part);
            }

            bool isLiteral = false;
            for (int i = 0; i < part.Length; i++)
            {
                if (part[i] == '\"')
                {
                    isLiteral = !isLiteral;
                }
                else if (!isLiteral)
                {
                    if (FormatHelper.NumberChars.Contains(part[i]))
                    {
                        if (formatStringInfo.FormatType == null)
                        {
                            formatStringInfo.FormatType = FormatStringType.Number;
                        }
                    }
                    else if (FormatHelper.DateChars.Contains(part[i]) || FormatHelper.TimeChars.Contains(part[i]))
                    {
                        if (FormatHelper.TimeChars.Contains(part[i]))
                        {
                            formatStringInfo.HasTimeChars = true;
                        }

                        if (FormatHelper.DateChars.Contains(part[i]) && !IsMinute(part, i))
                        {
                            formatStringInfo.HasDateChars = true;
                        }

                        if (formatStringInfo.FormatType == null)
                        {
                            formatStringInfo.FormatType = FormatStringType.DateTime;
                        }
                    }
                }
            }
        }

        private static bool IsMinute(string part, int index)
        {
            char lower = char.ToLower(part[index]);
            char upper = char.ToUpper(part[index]);

            bool isMinute = false;

            for (int i = index; i < part.Length; i++)
            {
                if (SpreadsheetCultureHelper.IsCharEqualTo(part[i], FormatHelper.DefaultSpreadsheetCulture.TimeSeparator))
                {
                    isMinute = true;
                }

                if (part[i] != lower && part[i] != upper && !SpreadsheetCultureHelper.IsCharEqualTo(part[i], FormatHelper.DefaultSpreadsheetCulture.TimeSeparator))
                {
                    break;
                }
            }

            for (int i = index; i >= 0; i--)
            {
                if (SpreadsheetCultureHelper.IsCharEqualTo(part[i], FormatHelper.DefaultSpreadsheetCulture.TimeSeparator))
                {
                    isMinute = true;
                }

                if (part[i] != lower && part[i] != upper && !SpreadsheetCultureHelper.IsCharEqualTo(part[i], FormatHelper.DefaultSpreadsheetCulture.TimeSeparator))
                {
                    break;
                }
            }

            return isMinute;
        }

        #endregion
    }
}