﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Formatting
{
    // TODO:: Rename this class.
    internal static class NumberValueParser
    {
        #region Constants

        private static readonly string DefaultNumberFormat = "#{0}##0";//,
        private static readonly string DefaultNumberDecimalFormat = "#{0}##0{1}00";

        private static readonly string DefaultPercentFormat = "0%";
        private static readonly string DefaultPercentDecimalFormat = "0{0}00%";

        private static readonly string DefaultScientificFormat = "0{0}00E+00";

        private delegate bool TryParseNumberValueDelegate(string value, FormatStringType? forceParsingAsSpecificType, SpreadsheetCultureHelper cultureInfo,
            out double parsedDoubleValue, out string format);

        private static TryParseNumberValueDelegate[] TryParseFunctions =
            new TryParseNumberValueDelegate[]
            {
                TryParseGeneral,
                TryParseNumber,
                TryParseCurrency,
                TryParseFraction,
                TryParseDate,
                TryParsePercent,
                TryParseScientific
            };

        #endregion


        #region Methods

        public static bool TryParse(string value, FormatStringType? forceParsingAsSpecificType, SpreadsheetCultureHelper cultureInfo, out string format, out double doubleValue)
        {
            doubleValue = 0;
            format = FormatHelper.GeneralFormatString;
            cultureInfo = cultureInfo ?? FormatHelper.DefaultSpreadsheetCulture;

            foreach (TryParseNumberValueDelegate tryParse in TryParseFunctions)
            {
                if (tryParse(value, forceParsingAsSpecificType, cultureInfo, out doubleValue, out format))
                {
                    return true;
                }
            }

            return false;
        }

        private static bool TryParseGeneral(string value, FormatStringType? forceParsingAsSpecificType, SpreadsheetCultureHelper cultureInfo, 
            out double parsedDoubleValue, out string format)
        {
            format = string.Empty;
            return cultureInfo.TryParseGeneral(value, out parsedDoubleValue);
        }

        private static bool TryParseNumber(string value, FormatStringType? forceParsingAsSpecificType, SpreadsheetCultureHelper cultureInfo, 
            out double result, out string format)
        {
            format = FormatHelper.GeneralFormatString;
            result = 0;

            bool canParse = cultureInfo.TryParseNumber(value, out result);

            if (canParse)
            {
                format = NumberValueParser.GetNumberFormat(value, cultureInfo);
            }

            return canParse;
        }

        private static bool TryParseCurrency(string value, FormatStringType? forceParsingAsSpecificType, SpreadsheetCultureHelper cultureInfo, 
            out double result, out string format)
        {
            return CurrencyParser.TryParse(value, out result, out format);
        }

        private static bool TryParseFraction(string value, FormatStringType? forceParsingAsSpecificType, SpreadsheetCultureHelper cultureInfo, 
            out double result, out string format)
        {
            return FractionParser.TryParse(value, forceParsingAsSpecificType, out result, out format);
        }

        private static bool TryParseDate(string value, FormatStringType? forceParsingAsSpecificType, SpreadsheetCultureHelper cultureInfo, 
            out double result, out string format)
        {
            DateTime resultAsDateTime;
            if (DateTimeParser.TryParse(value, out resultAsDateTime, out format))
            {
                result = FormatHelper.ConvertDateTimeToDouble(resultAsDateTime);

                return true;
            }

            result = 0;
            format = FormatHelper.GeneralFormatString;

            return false;
        }

        private static bool TryParsePercent(string value, FormatStringType? forceParsingAsSpecificType, SpreadsheetCultureHelper cultureInfo, 
            out double result, out string format)
        {
            format = FormatHelper.GeneralFormatString;

            bool canParse = cultureInfo.TryParsePercent(value, out result);

            if (canParse)
            {
                format = NumberValueParser.GetPercentFormat(value, cultureInfo);
                // NOTE: When inputing 12% it should be presented as 12% and not 1200%. That's why we devide by 100.
                //result /= 100;
            }

            return canParse;
        }

        private static bool TryParseScientific(string value, FormatStringType? forceParsingAsSpecificType, SpreadsheetCultureHelper cultureInfo, 
            out double result, out string format)
        {
            format = FormatHelper.GeneralFormatString;

            bool canParse = cultureInfo.TryParseScientific(value, out result);
           
            if (canParse)
            {
                format = string.Format(NumberValueParser.DefaultScientificFormat, cultureInfo.NumberDecimalSeparator);
            }

            return canParse;
        }

        private static string GetNumberFormat(string value, SpreadsheetCultureHelper cultureInfo)
        {
            if (value.IndexOf(cultureInfo.CultureInfo.NumberFormat.NumberDecimalSeparator) != -1)
            {
                return string.Format(NumberValueParser.DefaultNumberDecimalFormat, cultureInfo.NumberGroupSeparator, cultureInfo.NumberDecimalSeparator);
            }

            return string.Format(NumberValueParser.DefaultNumberFormat, cultureInfo.NumberGroupSeparator);
        }

        private static string GetPercentFormat(string value, SpreadsheetCultureHelper cultureInfo)
        {
            int index;
            if ((index = value.IndexOf(cultureInfo.CultureInfo.NumberFormat.PercentDecimalSeparator)) != -1)
            {
                int zerosCount = 0;
                for (int i = index + 1; i < value.Length; i++)
                {
                    if (char.IsDigit(value[i]))
                    {
                        zerosCount++;
                    }
                    else
                    {
                        break;
                    }
                }

                return string.Format(NumberValueParser.DefaultPercentDecimalFormat, cultureInfo.NumberDecimalSeparator);
            }

            return NumberValueParser.DefaultPercentFormat;
        }

        #endregion
    }
}