﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Core;

namespace Telerik.Windows.Documents.Spreadsheet.History
{
    /// <summary>
    /// Represents the history of a workbook.
    /// </summary>
    public class WorkbookHistory
    {
        #region Types

        private enum HistoryState
        {
            Listening,
            InUndo,
            InRedo,
            InCancelUndoGroup
        }

        #endregion


        #region Constants

        // TODO: Rethink this.
        /// <summary>
        /// The default depth of the workbook history.
        /// </summary>
        public const int DefaultDepth = int.MaxValue;

        #endregion


        #region Fields

        private readonly Workbook workbook;
        private int depth;
        private bool isEnabled;
        private HistoryState state;
        //private int beginUndoGroupCount;
        private readonly BeginEndCounter undoGroupCounter;
        private readonly WorkbookChangesCollection changes;

        private readonly LinkedList<WorkbookHistoryAction> undoStack;
        private readonly Stack<WorkbookHistoryAction> redoStack;

        #endregion


        #region Properties

        /// <summary>
        /// Gets a value indicating whether undo can be performed.
        /// </summary>
        /// <value>The value indicating whether undo can be performed.</value>
        public bool CanUndo
        {
            get
            {
                return (this.undoStack.Count > 0);
            }
        }

        /// <summary>
        /// Gets a value indicating whether redo can be performed.
        /// </summary>
        /// <value>The value indicating whether redo can be performed.</value>
        public bool CanRedo
        {
            get
            {
                return (this.redoStack.Count > 0);
            }
        }

        /// <summary>
        /// Gets or sets the depth of the history.
        /// </summary>
        /// <value>The depth of the history.</value>
        public int Depth
        {
            get
            {
                return this.depth;
            }
            set
            {
                Guard.ThrowExceptionIfLessThan(1, value, "History depth should be a positive number");

                if (this.depth != value)
                {
                    this.depth = value;
                    this.OnDepthChanged();
                }
            }
        }

        /// <summary>
        /// Gets a value indicating whether the history is currently recording an undo group.
        /// </summary>
        /// <value>The value indicating whether the history is currently recording an undo group.</value>
        public bool IsInUndoGroup
        {
            get
            {
                return this.undoGroupCounter.IsUpdateInProgress;
            }
        }

        internal bool IsActionMergeSuspended { get; set; }

        /// <summary>
        /// Gets or sets a value indicating whether history is enabled.
        /// </summary>
        /// <value>The value indicating whether history is enabled.</value>
        public bool IsEnabled
        {
            get
            {
                return this.isEnabled;
            }
            set
            {
                Guard.ThrowExceptionIfTrue(this.IsInUndoGroup, "IsInUndoGroup");

                if (this.isEnabled != value)
                {
                    this.isEnabled = value;
                    this.OnIsEnabledChanged();
                }
            }
        }

        /// <summary>
        /// Gets the workbook.
        /// </summary>
        /// <value>The workbook.</value>
        public Workbook Workbook
        {
            get
            {
                return this.workbook;
            }
        }

        #endregion


        #region Constructors

        internal WorkbookHistory(Workbook workbook)
        {
            Guard.ThrowExceptionIfNull(workbook, "workbook");

            this.workbook = workbook;
            this.depth = WorkbookHistory.DefaultDepth;
            this.undoStack = new LinkedList<WorkbookHistoryAction>();
            this.redoStack = new Stack<WorkbookHistoryAction>();
            this.changes = new WorkbookChangesCollection();
            this.undoGroupCounter = new BeginEndCounter();
        }

        #endregion


        #region Methods

        private void OnDepthChanged()
        {
            this.EnsureUndoStackDepth();
            this.redoStack.Clear();
        }

        private void OnIsEnabledChanged()
        {
            if (!this.isEnabled)
            {
                this.Clear();
            }
        }

        /// <summary>
        /// Begins the undo group.
        /// </summary>
        public void BeginUndoGroup()
        {
            if (!this.IsEnabled)
            {
                return;
            }

            if (!this.IsInUndoGroup)
            {
                this.Workbook.SuspendLayoutUpdate();
                this.OnRecordExecuting();
            }

            this.undoGroupCounter.BeginUpdate();
        }

        /// <summary>
        /// Ends the undo group.
        /// </summary>
        public void EndUndoGroup()
        {
            this.EndUndoGroup(false);
        }

        /// <summary>
        /// Cancels the undo group.
        /// </summary>
        public void CancelUndoGroup()
        {
            this.EndUndoGroup(true);
        }

        private void EndUndoGroup(bool isCanceled)
        {
            if (!this.IsEnabled)
            {
                return;
            }

            int undoGroupLevel = this.undoGroupCounter.BeginUpdateCounter;
            if (undoGroupLevel == 1)
            {
                foreach (var worksheet in this.Workbook.Worksheets)
                {
                    worksheet.ExpandCachedColumnWidths();
                }
            }

            this.undoGroupCounter.EndUpdate();

            // If we have changes to Cancel/Commit.
            if (!this.changes.IsEmpty)
            {
                if (isCanceled)
                {
                    IEnumerable<WorkbookChange> changes = this.changes.PopChangesFromUndoGroupLevel(undoGroupLevel);
                    WorkbookHistoryAction historyAction = new WorkbookHistoryAction(changes, string.Empty);

                    Debug.Assert(this.state == HistoryState.Listening, "Problem");
                    this.state = HistoryState.InCancelUndoGroup;

                    historyAction.Undo();

                    if (!this.IsInUndoGroup)
                    {
                        this.Workbook.ResumeLayoutUpdate();
                    }

                    this.state = HistoryState.Listening;
                }
                else
                {
                    if (!this.IsInUndoGroup)
                    {
                        IEnumerable<WorkbookChange> changes = this.changes.PopChangesFromUndoGroupLevel(undoGroupLevel);

                        // TODO: Set proper display text.
                        WorkbookHistoryAction compositeHistoryAction = new WorkbookHistoryAction(changes, string.Empty);
                        Debug.Assert(this.changes.IsEmpty, "Changes must be empty!");
                        //this.changes = new List<TextChange>();

                        WorkbookHistoryAction previousHistoryAction = null;
                        if (this.undoStack.Count > 0)
                        {
                            previousHistoryAction = this.undoStack.Last.Value;
                        }

                        bool isMerged = !this.IsActionMergeSuspended && (previousHistoryAction != null) && previousHistoryAction.Merge();
                        if (!isMerged)
                        {
                            this.undoStack.AddLast(compositeHistoryAction);
                            this.EnsureUndoStackDepth();
                        }

                        this.Workbook.ResumeLayoutUpdate();

                        // Clear redo stack.
                        this.redoStack.Clear();

                        this.OnRecordExecuted(compositeHistoryAction, isMerged);
                    }
                }
            }
            else
            {
                if (!this.IsInUndoGroup)
                {
                    this.Workbook.ResumeLayoutUpdate();
                }
            }
        }

        internal bool RecordSheetChange(WorkbookChange change)
        {
            if (!this.IsEnabled)
            {
                return false;
            }

            if (this.state != HistoryState.Listening)
            {
                return false;
            }

            if (this.undoStack.Count > this.depth)
            {
                return false;
            }

            bool hasOpenedUndoGroup = this.IsInUndoGroup;

            if (!hasOpenedUndoGroup)
            {
                this.BeginUndoGroup();
            }

            try
            {
                this.changes.Add(this.undoGroupCounter.BeginUpdateCounter, change);
            }
            finally
            {
                if (!hasOpenedUndoGroup)
                {
                    this.EndUndoGroup();
                }
            }

            return true;
        }

        /// <summary>
        /// Performs the undo action.
        /// </summary>
        /// <returns>The value indicating whether the undo was successfully performed.</returns>
        public bool Undo()
        {
            if (this.IsInUndoGroup)
            {
                throw new InvalidOperationException("Cannot Undo while in begin undo group.");
            }

            if (!this.CanUndo)
            {
                return false;
            }

            this.state = HistoryState.InUndo;

            WorkbookHistoryAction action = this.undoStack.Last.Value;

            this.RestoreActiveSheet(action);

            this.OnUndoExecuting(action);
            bool result = this.UndoCore();
            if (result)
            {
                this.OnUndoExecuted(action);
            }

            this.state = HistoryState.Listening;

            return result;
        }

        private void RestoreActiveSheet(WorkbookHistoryAction action)
        {
            this.Workbook.ActiveSheet = action.ActiveSheet;
        }

        /// <summary>
        /// Performs the redo action.
        /// </summary>
        /// <returns>The value indicating whether the redo was successfully performed.</returns>
        public bool Redo()
        {
            if (this.IsInUndoGroup)
            {
                throw new InvalidOperationException("Cannot Redo while in begin undo group.");
            }

            if (!this.CanRedo)
            {
                return false;
            }

            this.state = HistoryState.InRedo;

            WorkbookHistoryAction action = this.redoStack.Peek();

            this.RestoreActiveSheet(action);

            this.OnRedoExecuting(action);
            bool result = this.RedoCore();
            if (result)
            {
                this.OnRedoExecuted(action);
            }

            this.state = HistoryState.Listening;

            return result;
        }

        private bool UndoCore()
        {
            WorkbookHistoryAction action = this.undoStack.Last.Value;
            this.undoStack.RemoveLast();

            action.Undo();

            if (this.redoStack.Count >= this.depth)
            {
                return false;
            }

            this.redoStack.Push(action);

            return true;
        }

        private bool RedoCore()
        {
            WorkbookHistoryAction action = this.redoStack.Pop();
            action.Redo();

            if (this.undoStack.Count >= this.depth)
            {
                return false;
            }

            this.undoStack.AddLast(action);

            return true;
        }

        /// <summary>
        /// Clears the workbook history.
        /// </summary>
        public void Clear()
        {
            if (this.IsInUndoGroup)
            {
                throw new InvalidOperationException("Clear Undo while in begin undo group");
            }

            this.undoStack.Clear();
            this.redoStack.Clear();

            this.OnCleared();
        }

        private void EnsureUndoStackDepth()
        {
            while (this.undoStack.Count > this.depth)
            {
                this.undoStack.RemoveFirst();
            }
        }

        internal void Clear(Sheet associatedSheetToClear)
        {
            var linkedListEnumerator = this.undoStack.GetEnumerator();
            LinkedListNode<WorkbookHistoryAction> currentNode = this.undoStack.First;

            List<WorkbookHistoryAction> actionsToRemove = new List<WorkbookHistoryAction>();
            while (linkedListEnumerator.MoveNext())
            {
                var nextNode = currentNode.Next;
                var currentHistoryAction = currentNode.Value;

                if (currentHistoryAction.ActiveSheet.Equals(associatedSheetToClear))
                {
                    actionsToRemove.Add(currentHistoryAction);
                }

                if (nextNode != null)
                {
                    currentNode = nextNode;
                }
            }

            foreach (var historyActionToRemove in actionsToRemove)
            {
                this.undoStack.Remove(historyActionToRemove);
            }

            this.redoStack.Clear();
        }

        #endregion


        #region Events

        /// <summary>
        /// Occurs when a record is executing.
        /// </summary>
        public event EventHandler RecordExecuting;

        /// <summary>
        /// Called when a record is executing.
        /// </summary>
        protected virtual void OnRecordExecuting()
        {
            if (this.RecordExecuting != null)
            {
                this.RecordExecuting(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Occurs when a record has executed.
        /// </summary>
        public event EventHandler<WorkbookHistoryRecordExecutedEventArgs> RecordExecuted;

        /// <summary>
        /// Called when a record is executed.
        /// </summary>
        /// <param name="historyAction">The history action.</param>
        /// <param name="isMerged">This parameter is not used and is reserved for future use.</param>
        protected virtual void OnRecordExecuted(WorkbookHistoryAction historyAction, bool isMerged)
        {
            if (this.RecordExecuted != null)
            {
                WorkbookHistoryRecordExecutedEventArgs args = new WorkbookHistoryRecordExecutedEventArgs(historyAction, isMerged);
                this.RecordExecuted(this, args);
            }
        }

        /// <summary>
        /// Occurs when undo is being executed.
        /// </summary>
        public event EventHandler<WorkbookHistoryEventArgs> UndoExecuting;

        /// <summary>
        /// Called when undo is being executed.
        /// </summary>
        /// <param name="historyAction">The history action.</param>
        protected virtual void OnUndoExecuting(WorkbookHistoryAction historyAction)
        {
            if (this.UndoExecuting != null)
            {
                WorkbookHistoryEventArgs args = new WorkbookHistoryEventArgs(historyAction);
                this.UndoExecuting(this, args);
            }
        }

        /// <summary>
        /// Occurs when undo has been executed.
        /// </summary>
        public event EventHandler<WorkbookHistoryEventArgs> UndoExecuted;

        /// <summary>
        /// Called when undo has been executed.
        /// </summary>
        /// <param name="historyAction">The history action.</param>
        protected virtual void OnUndoExecuted(WorkbookHistoryAction historyAction)
        {
            if (this.UndoExecuted != null)
            {
                WorkbookHistoryEventArgs args = new WorkbookHistoryEventArgs(historyAction);
                this.UndoExecuted(this, args);
            }
        }

        /// <summary>
        /// Occurs when redo is being executed.
        /// </summary>
        public event EventHandler<WorkbookHistoryEventArgs> RedoExecuting;

        /// <summary>
        /// Called when redo is being executed.
        /// </summary>
        /// <param name="historyAction">The history action.</param>
        protected virtual void OnRedoExecuting(WorkbookHistoryAction historyAction)
        {
            if (this.RedoExecuting != null)
            {
                WorkbookHistoryEventArgs args = new WorkbookHistoryEventArgs(historyAction);
                this.RedoExecuting(this, args);
            }
        }

        /// <summary>
        /// Occurs when redo has been executed.
        /// </summary>
        public event EventHandler<WorkbookHistoryEventArgs> RedoExecuted;

        /// <summary>
        /// Called when redo has been executed.
        /// </summary>
        /// <param name="historyAction">The history action.</param>
        protected virtual void OnRedoExecuted(WorkbookHistoryAction historyAction)
        {
            if (this.RedoExecuted != null)
            {
                WorkbookHistoryEventArgs args = new WorkbookHistoryEventArgs(historyAction);
                this.RedoExecuted(this, args);
            }
        }

        /// <summary>
        /// Occurs when history has been cleared.
        /// </summary>
        public event EventHandler Cleared;

        /// <summary>
        /// Called when history has been cleared.
        /// </summary>
        protected virtual void OnCleared()
        {
            if (this.Cleared != null)
            {
                this.Cleared(this, EventArgs.Empty);
            }
        }

        #endregion
    }
}