﻿using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;

namespace Telerik.Windows.Documents.Spreadsheet.Layout
{
    /// <summary>
    /// Represents a cell layout box.
    /// </summary>
    public class CellLayoutBox : LayoutBox
    {
        #region Fields

        private CellIndex index;
        private readonly int columnIndex;
        private readonly int rowIndex;
        private readonly long longIndex;
        private readonly CellMergeState mergeState;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the index of the row.
        /// </summary>
        /// <value>The index of the row.</value>
        public int RowIndex
        {
            get
            {
                return this.rowIndex;
            }
        }

        /// <summary>
        /// Gets the index of the column.
        /// </summary>
        /// <value>The index of the column.</value>
        public int ColumnIndex
        {
            get
            {
                return this.columnIndex;
            }
        }

        /// <summary>
        /// Gets the cell index.
        /// </summary>
        /// <value>The index.</value>
        public CellIndex Index
        {
            get
            {
                if (this.index == null)
                {
                    this.index = CellsPropertyBag.ConvertLongToCellIndex(this.longIndex);
                }

                return this.index;
            }
        }

        internal long LongIndex
        {
            get
            {
                return this.longIndex;
            }
        }

        /// <summary>
        /// Gets the merge state of the cell.
        /// </summary>
        /// <value>The merge state of the cell.</value>
        public CellMergeState MergeState
        {
            get
            {
                return this.mergeState;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="CellLayoutBox" /> class.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <param name="rect">The rectangle.</param>
        /// <param name="mergeState">The merge.</param>
        public CellLayoutBox(int rowIndex, int columnIndex, Rect rect, CellMergeState mergeState)
            : base(rect)
        {
            this.columnIndex = columnIndex;
            this.rowIndex = rowIndex;
            this.longIndex = CellsPropertyBag.ConvertCellIndexToLong(rowIndex, columnIndex);
            this.mergeState = mergeState;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current
        /// <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
            return string.Format("({0}, {1})", this.RowIndex, this.ColumnIndex);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            CellLayoutBox other = obj as CellLayoutBox;
            if (other == null)
            {
                return false;
            }

            return this.LongIndex == other.LongIndex;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return this.LongIndex.GetHashCode();
        }

        #endregion
    }
}