﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Media;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;

#if SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.Maths;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.Layout.Layers
{
    internal class CellBordersLayerBase : WorksheetLayerBase
    {
        #region Fields

        private readonly bool shouldShowTopLeftMostBorders;
        private readonly bool supportDiagonalBorders;
        private readonly string name;
        private readonly Func<CellBorder, bool> shouldShowBorder;
        private readonly IRenderer<LineRenderable> lineRenderer;

        #endregion


        #region Properties
        
        /// <summary>
        /// Gets the value indicating if the layer supports diagonal borders.
        /// </summary>
        /// <value>The value indicating if the layer supports diagonal borders.</value>
        public bool SupportDiagonalBorders
        {
            get
            {
                return this.supportDiagonalBorders;
            }
        }

        public override string Name
        {
            get
            {
                return this.name;
            }
        }

        internal IRenderer<LineRenderable> LineRenderer
        {
            get
            {
                return this.lineRenderer;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="CellBordersLayerBase" /> class.
        /// </summary>
        public CellBordersLayerBase(IRenderer<LineRenderable> lineRenderer, bool shouldShowTopLeftMostBorders, bool supportDiagonalBorders, Func<CellBorder, bool> shouldShowBorder, string name)
        {
            this.shouldShowTopLeftMostBorders = shouldShowTopLeftMostBorders;
            this.supportDiagonalBorders = supportDiagonalBorders;
            this.shouldShowBorder = shouldShowBorder;
            this.lineRenderer = lineRenderer;
            this.name = name;
        }

        #endregion


        #region Methods

        protected override void UpdateRenderOverride(WorksheetRenderUpdateContext worksheetUpdateContext)
        {
            this.UpdateHorizontalBorders(worksheetUpdateContext);
            this.UpdateVerticalBorders(worksheetUpdateContext);

            if (this.SupportDiagonalBorders)
            {
                this.UpdateDiagonalBorders(worksheetUpdateContext);
            }
        }

        #endregion


        #region Horizontal Borders

        /// <summary>
        /// Updates the horizontal borders.
        /// </summary>
        /// <param name="updateContext">The update context.</param>
        private void UpdateHorizontalBorders(WorksheetRenderUpdateContext updateContext)
        {
            foreach (ViewportPane viewportPane in updateContext.SheetViewport.ViewportPanes)
            {
                CellRange visibleRange = viewportPane.VisibleRange;

                int fromRowIndex = visibleRange.FromIndex.RowIndex;
                int fromColumnIndex = visibleRange.FromIndex.ColumnIndex;
                int toRowIndex = visibleRange.ToIndex.RowIndex;
                int toColumnIndex = visibleRange.ToIndex.ColumnIndex;

                for (int rowIndex = fromRowIndex; rowIndex <= toRowIndex; rowIndex++)
                {
                    ICompressedList<CellBorder> borders =
                        new CompressedList<CellBorder>(fromColumnIndex, toColumnIndex, null);

                    for (int columnIndex = fromColumnIndex; columnIndex <= toColumnIndex; columnIndex++)
                    {
                        this.AddToBorders(updateContext,
                                (border) => borders.SetValue(columnIndex, border),
                                (mergedRange) => { return rowIndex < mergedRange.ToIndex.RowIndex; },
                                () => { return BorderRenderHelper.GetBottomBorderToDisplay(updateContext, rowIndex, columnIndex); },
                                false,
                                rowIndex, columnIndex);
                    }

                    this.UpdateBordersInRow(viewportPane.ViewportPaneType, rowIndex, borders, updateContext,
                        (boundingRect) =>
                        {
                            return boundingRect.Bottom;
                        });

                    if (rowIndex == fromRowIndex && this.shouldShowTopLeftMostBorders)
                    {
                        borders = new CompressedList<CellBorder>(fromColumnIndex, toColumnIndex, null);

                        for (int columnIndex = fromColumnIndex; columnIndex <= toColumnIndex; columnIndex++)
                        {
                            this.AddToBorders(updateContext,
                                (border) => borders.SetValue(columnIndex, border),
                                (mergedRange) => { return rowIndex >= mergedRange.FromIndex.RowIndex && rowIndex < mergedRange.ToIndex.RowIndex; },
                                () => { return updateContext.GetTopBorder(CellsPropertyBag.ConvertCellIndexToLong(rowIndex, columnIndex)); },
                                false,
                                rowIndex, columnIndex);
                        }

                        this.UpdateBordersInRow(viewportPane.ViewportPaneType, rowIndex, borders, updateContext,
                            (boundingRect) =>
                            {
                                return boundingRect.Top;
                            });
                    }
                }
            }
        }

        /// <summary>
        /// Updates the bottom borders in row.
        /// </summary>
        /// <param name="paneType">Type of the pane.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="borders">The borders.</param>
        /// <param name="updateContext">The update context.</param>
        /// <param name="getYCoordinate">The get Y coordinate.</param>
        private void UpdateBordersInRow(ViewportPaneType paneType,
            int rowIndex, ICompressedList<CellBorder> borders, WorksheetRenderUpdateContext updateContext, Func<Rect, double> getYCoordinate)
        {
            CellBorder verticalBorder;
            foreach (Range<long, CellBorder> range in borders)
            {
                if (range.Value == null)
                {
                    continue;
                }

                CellLayoutBox leftCellBox = updateContext.GetVisibleCellBox(rowIndex, (int)range.Start);
                CellLayoutBox rightCellBox = updateContext.GetVisibleCellBox(rowIndex, (int)range.End);

                double leftDelta = 0;
                if (leftCellBox.ColumnIndex > 0)
                {
                    verticalBorder = BorderRenderHelper.GetRightBorderToDisplay(updateContext, leftCellBox.RowIndex, leftCellBox.ColumnIndex - 1);
                    leftDelta = verticalBorder.Thickness / 2;
                }

                double rightDelta = 0;
                verticalBorder = BorderRenderHelper.GetRightBorderToDisplay(updateContext, rightCellBox.RowIndex, rightCellBox.ColumnIndex);
                rightDelta = verticalBorder.Thickness / 2;

                LineRenderable line = new LineRenderable();
                SetLinePropertiesForBorder(line, range.Value, updateContext);
                line.X1 = leftCellBox.BoundingRectangle.Left - leftDelta;
                line.X2 = rightCellBox.BoundingRectangle.Right + rightDelta;
                line.Y1 = getYCoordinate(leftCellBox.BoundingRectangle);//leftCellBox.BoundingRectangle.Bottom;
                line.Y2 = line.Y1;

                this.ContainerManager.Add(line, paneType);
            }
        }

        #endregion


        #region Vertical Borders

        /// <summary>
        /// Updates the vertical borders.
        /// </summary>
        /// <param name="updateContext">The update context.</param>
        private void UpdateVerticalBorders(WorksheetRenderUpdateContext updateContext)
        {
            HashSet<long> verticalBordersToSkip = this.GetBordersToSkip(updateContext);
            this.UpdateVerticalBordersForVisibleRanges(updateContext, verticalBordersToSkip);
        }

        private void UpdateVerticalBordersForVisibleRanges(WorksheetRenderUpdateContext updateContext, HashSet<long> verticalBordersToSkip)
        {
            foreach (ViewportPane viewportPane in updateContext.SheetViewport.ViewportPanes)
            {
                this.UpdateVerticalBordersForCellRange(updateContext, viewportPane, verticalBordersToSkip);
            }
        }

        private void UpdateVerticalBordersForCellRange(WorksheetRenderUpdateContext updateContext, ViewportPane viewportPane, HashSet<long> verticalBordersToSkip)
        {
            CellRange range = viewportPane.VisibleRange;

            int fromRowIndex = range.FromIndex.RowIndex;
            int fromColumnIndex = range.FromIndex.ColumnIndex;
            int toRowIndex = range.ToIndex.RowIndex;
            int toColumnIndex = range.ToIndex.ColumnIndex;

            for (int columnIndex = fromColumnIndex; columnIndex <= toColumnIndex; columnIndex++)
            {
                ICompressedList<CellBorder> borders =
                    new CompressedList<CellBorder>(fromRowIndex, toRowIndex, null);

                for (int rowIndex = fromRowIndex; rowIndex <= toRowIndex; rowIndex++)
                {
                    this.AddToBorders(updateContext,
                        (border) => borders.SetValue(rowIndex, border),
                        (mergedRange) => { return columnIndex < mergedRange.ToIndex.ColumnIndex; },
                        () => { return BorderRenderHelper.GetRightBorderToDisplay(updateContext, rowIndex, columnIndex); },
                        verticalBordersToSkip.Contains(CellsPropertyBag.ConvertCellIndexToLong(rowIndex, columnIndex)),
                        rowIndex, columnIndex);
                }

                this.UpdateBordersInColumn(viewportPane.ViewportPaneType, columnIndex, borders, updateContext,
                    (boundingRect) =>
                    {
                        return boundingRect.Right;
                    });

                if (columnIndex == fromColumnIndex && this.shouldShowTopLeftMostBorders)
                {
                    borders = new CompressedList<CellBorder>(fromRowIndex, toRowIndex, null);

                    for (int rowIndex = fromRowIndex; rowIndex <= toRowIndex; rowIndex++)
                    {
                        long index = CellsPropertyBag.ConvertCellIndexToLong(rowIndex, columnIndex);
                        this.AddToBorders(updateContext,
                            (border) => borders.SetValue(rowIndex, border),
                            (mergedRange) => { return columnIndex >= mergedRange.FromIndex.ColumnIndex && columnIndex < mergedRange.ToIndex.ColumnIndex; },
                            () => { return updateContext.GetLeftBorder(index); },
                            verticalBordersToSkip.Contains(index),
                            rowIndex, columnIndex);
                    }

                    this.UpdateBordersInColumn(viewportPane.ViewportPaneType, columnIndex, borders, updateContext,
                        (boundingRect) =>
                        {
                            return boundingRect.Left;
                        });
                }
            }
        }

        private void AddToBorders(WorksheetRenderUpdateContext updateContext, Action<CellBorder> addToBorders,
            Func<CellRange, bool> testIsInsideMergedRange, Func<CellBorder> getCellBorder, bool shouldSkip,
            int rowIndex, int columnIndex)
        {
            Worksheet worksheet = updateContext.Worksheet;

            bool isInsideMergedRange = false;
            CellRange mergedRange;
            if (worksheet.Cells.TryGetContainingMergedRange(rowIndex, columnIndex, out mergedRange))
            {
                isInsideMergedRange = testIsInsideMergedRange(mergedRange);
            }

            if (!isInsideMergedRange && !shouldSkip)
            {
                CellBorder topBorder = getCellBorder();
                if (this.ShouldShowBorder(topBorder))
                {
                    addToBorders(topBorder);
                }
            }
        }

        private HashSet<long> GetBordersToSkip(WorksheetRenderUpdateContext updateContext)
        {
            HashSet<long> verticalBordersToSkip = new HashSet<long>();
            foreach (var viewportPane in updateContext.SheetViewport.ViewportPanes)
            {
                long[] bordersToSkip = this.GetBordersToSkipFromCellRange(viewportPane.VisibleRange, updateContext);
                verticalBordersToSkip.UnionWith(bordersToSkip);
            }

            return verticalBordersToSkip;
        }

        private long[] GetBordersToSkipFromCellRange(CellRange range, WorksheetRenderUpdateContext updateContext)
        {
            List<long> bordersToSkip = new List<long>();

            int fromRowIndex = range.FromIndex.RowIndex;
            int fromColumnIndex = range.FromIndex.ColumnIndex;
            int toRowIndex = range.ToIndex.RowIndex;
            int toColumnIndex = range.ToIndex.ColumnIndex;

            // Calculate which borders should be skipped because of cell values, spanning in more than one column.
            for (int rowIndex = fromRowIndex; rowIndex <= toRowIndex; rowIndex++)
            {
                for (int columnIndex = fromColumnIndex; columnIndex <= toColumnIndex; columnIndex++)
                {
                    long index = CellsPropertyBag.ConvertCellIndexToLong(rowIndex, columnIndex);
                    CellLayoutBox cellBox = updateContext.GetVisibleCellBox(index);

                    if (cellBox.MergeState == CellMergeState.NonTopLeftCellInMergedRange)
                    {
                        continue;
                    }

                    ICellValue cellValue = updateContext.GetCellValue(index);
                    if (cellValue == EmptyCellValue.EmptyValue)
                    {
                        continue;
                    }

                    Rect clippedBoundingRectangle = CellBoxRenderHelper.CalculateCellClippedBoundingRectangle(cellBox, updateContext);
                    clippedBoundingRectangle.X += 0.1;
                    clippedBoundingRectangle.Y = cellBox.Top;
                    clippedBoundingRectangle.Width = Math.Max(0, clippedBoundingRectangle.Width - SpreadsheetDefaultValues.RightCellMargin);
                    CellLayoutBox prevCellBox;
                    CellLayoutBox nextCellBox;
                    updateContext.GetPreviousNextNonEmptyNonMergedCells(cellBox.LongIndex, out prevCellBox, out nextCellBox);
                    for (int i = cellBox.ColumnIndex - 1; i >= prevCellBox.ColumnIndex; i--)
                    {
                        CellLayoutBox currentCellBox = updateContext.GetVisibleCellBox(rowIndex, i);
                        if (clippedBoundingRectangle.IntersectsWith(currentCellBox.BoundingRectangle))
                        {
                            bordersToSkip.Add(currentCellBox.LongIndex);
                        }
                        else
                        {
                            break;
                        }
                    }

                    for (int i = cellBox.ColumnIndex + 1; i <= nextCellBox.ColumnIndex; i++)
                    {
                        CellLayoutBox currentCellBox = updateContext.GetVisibleCellBox(rowIndex, i);
                        if (clippedBoundingRectangle.IntersectsWith(currentCellBox.BoundingRectangle))
                        {
                            bordersToSkip.Add(CellsPropertyBag.ConvertCellIndexToLong(rowIndex, i - 1));
                        }
                        else
                        {
                            break;
                        }
                    }
                }
            }

            return bordersToSkip.ToArray();
        }

        /// <summary>
        /// Updates the right borders in column.
        /// </summary>
        /// <param name="paneType">Type of the pane.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <param name="borders">The borders.</param>
        /// <param name="updateContext">The update context.</param>
        /// <param name="getXCoordinate">The get X coordinate.</param>
        private void UpdateBordersInColumn(ViewportPaneType paneType,
            int columnIndex, ICompressedList<CellBorder> borders, WorksheetRenderUpdateContext updateContext, Func<Rect, double> getXCoordinate)
        {
            CellBorder horizontalBorder;
            foreach (var range in borders)
            {
                if (range.Value == null)
                {
                    continue;
                }

                CellLayoutBox topCellBox = updateContext.GetVisibleCellBox((int)range.Start, columnIndex);
                CellLayoutBox bottomCellBox = updateContext.GetVisibleCellBox((int)range.End, columnIndex);

                LineRenderable line = new LineRenderable();
                SetLinePropertiesForBorder(line, range.Value, updateContext);

                double topDelta = 0;
                if (topCellBox.RowIndex > 0)
                {
                    horizontalBorder = BorderRenderHelper.GetBottomBorderToDisplay(updateContext, topCellBox.RowIndex - 1, topCellBox.ColumnIndex);
                    topDelta = horizontalBorder.Thickness / 2;
                }

                double bottomDelta = 0;
                horizontalBorder = BorderRenderHelper.GetBottomBorderToDisplay(updateContext, bottomCellBox.RowIndex, bottomCellBox.ColumnIndex);
                bottomDelta = horizontalBorder.Thickness / 2;

                line.X1 = getXCoordinate(topCellBox.BoundingRectangle); //topCellBox.BoundingRectangle.Right;
                line.X2 = line.X1;
                line.Y1 = topCellBox.BoundingRectangle.Top - topDelta;
                line.Y2 = bottomCellBox.BoundingRectangle.Bottom + bottomDelta;

                this.ContainerManager.Add(line, paneType);
            }
        }

        #endregion


        #region Diagonal Borders

        /// <summary>
        /// Updates the diagonal borders.
        /// </summary>
        /// <param name="updateContext">The update context.</param>
        private void UpdateDiagonalBorders(WorksheetRenderUpdateContext updateContext)
        {
            foreach (ViewportPane viewportPane in updateContext.SheetViewport.ViewportPanes)
            {
                CellRange visibleRange = viewportPane.VisibleRange;

                int fromRowIndex = visibleRange.FromIndex.RowIndex;
                int fromColumnIndex = visibleRange.FromIndex.ColumnIndex;
                int toRowIndex = visibleRange.ToIndex.RowIndex;
                int toColumnIndex = visibleRange.ToIndex.ColumnIndex;

                for (int rowIndex = fromRowIndex; rowIndex <= toRowIndex; rowIndex++)
                {
                    for (int columnIndex = fromColumnIndex; columnIndex <= toColumnIndex; columnIndex++)
                    {
                        CellLayoutBox cellBox = updateContext.GetVisibleCellBox(rowIndex, columnIndex);

                        // Skip the non top left merged cells.
                        if (cellBox.MergeState == CellMergeState.NonTopLeftCellInMergedRange)
                        {
                            continue;
                        }

                        CellBorder diagonalUp = updateContext.GetDiagonalUpBorder(cellBox.LongIndex);//worksheet.Cells[cellBox.Index].GetDiagonalUpBorder().Value;
                        if (diagonalUp != CellBorder.Default)
                        {
                            LineRenderable line = new LineRenderable();
                            SetLinePropertiesForBorder(line, diagonalUp, updateContext);
                            line.X1 = cellBox.BoundingRectangle.Left;
                            line.X2 = cellBox.BoundingRectangle.Right;
                            line.Y1 = cellBox.BoundingRectangle.Bottom;
                            line.Y2 = cellBox.BoundingRectangle.Top;

                            this.ContainerManager.Add(line, viewportPane.ViewportPaneType);
                        }

                        CellBorder diagonalDown = updateContext.GetDiagonalDownBorder(cellBox.LongIndex);//worksheet.Cells[cellBox.Index].GetDiagonalDownBorder().Value;
                        if (diagonalDown != CellBorder.Default)
                        {
                            LineRenderable line = new LineRenderable();
                            SetLinePropertiesForBorder(line, diagonalDown, updateContext);
                            line.X1 = cellBox.BoundingRectangle.Left;
                            line.X2 = cellBox.BoundingRectangle.Right;
                            line.Y1 = cellBox.BoundingRectangle.Top;
                            line.Y2 = cellBox.BoundingRectangle.Bottom;

                            this.ContainerManager.Add(line, viewportPane.ViewportPaneType);
                        }
                    }
                }
            }
        }

        #endregion


        #region Methods

        /// <summary>
        /// Computes if the given border have to be shown.
        /// </summary>
        /// <param name="border">The border.</param>
        /// <returns><c>True</c> if the border have to be shown.</returns>
        public bool ShouldShowBorder(CellBorder border)
        {
            return this.shouldShowBorder(border);
        }

        /// <summary>
        /// Sets the line properties for border.
        /// </summary>
        /// <param name="line">The line used to show the border.</param>
        /// <param name="border">The cell border properties.</param>
        /// <param name="updateContext">The update context.</param>
        private static void SetLinePropertiesForBorder(LineRenderable line, CellBorder border, WorksheetRenderUpdateContext updateContext)
        {
            line.Stroke = border.Color.GetActualValue(updateContext.CurrentTheme.ColorScheme);
            line.CellBorderStyle = border.Style;
            //DoubleCollection doubleCollection = BorderUIHelper.GetStrokeDashArrayByBorderStyle(border.Style);
            //line.StrokeDashArray = doubleCollection;
            line.StrokeThickness = border.Thickness;

            // NOTE: The ZIndex is for the edges.
            int zIndex = border.GetZIndex(updateContext.CurrentTheme.ColorScheme);
            line.ZIndex = zIndex;
        }

        /// <summary>
        /// Translates and scales the UI elements of the layer.
        /// </summary>
        /// <param name="updateContext">The update context.</param>
        protected override void TranslateAndScale(WorksheetRenderUpdateContext updateContext)
        {
            ScaleTransform scaleTransform = new ScaleTransform
            {
                ScaleX = updateContext.ScaleFactor.Width,
                ScaleY = updateContext.ScaleFactor.Height,
            };

            foreach (ViewportPaneType viewportPaneType in Enum.GetValues(typeof(ViewportPaneType)))
            {
                foreach (IRenderable child in this.ContainerManager.GetElementsForViewportPane(viewportPaneType))
                {
                    LineRenderable line = (LineRenderable)child;
                    Rect translatedRect = this.Translate(line, viewportPaneType, updateContext);
                    Point topLeft = new Point(translatedRect.Left, translatedRect.Top);
                    topLeft = scaleTransform.Transform(topLeft);

                    Point widthHeight = scaleTransform.Transform(new Point(translatedRect.Width, translatedRect.Height));
                    translatedRect.Width = widthHeight.X;
                    translatedRect.Height = widthHeight.Y;

                    line.X2 = topLeft.X + (Math.Sign(line.X2 - line.X1) * translatedRect.Width);
                    line.X1 = topLeft.X;

                    line.Y2 = topLeft.Y + (Math.Sign(line.Y2 - line.Y1) * translatedRect.Height);
                    line.Y1 = topLeft.Y;

                    this.LineRenderer.Render(line, viewportPaneType);
                }
            }
        }

        private Rect Translate(LineRenderable line, ViewportPaneType viewportPaneType, RenderUpdateContext updateContext)
        {
            // TODO: Remove this commented lines in two weeks if nothing goes wrong! :D
            //if (line.Y1 < line.Y2)
            //{
                Rect rect = new Rect(line.X1, line.Y1, Math.Abs(line.X2 - line.X1), Math.Abs(line.Y2 - line.Y1));
                Rect translatedRect = base.Translate(rect, viewportPaneType, updateContext);

                return translatedRect;
            //}
            //else
            //{
            //    Rect rect = new Rect(line.X1, line.Y2, Math.Abs(line.X2 - line.X1), Math.Abs(line.Y1 - line.Y2));
            //    Rect translatedRect = base.Translate(rect, viewportPaneType, updateContext);

            //    return new Rect(translatedRect.Left, translatedRect.Bottom, translatedRect.Width, translatedRect.Height);
            //}
        }

        #endregion
    }
}
