﻿using System;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Layout.Layers
{
    internal static class CellBoxRenderHelper
    {
        public static bool ShouldDisplayContentAsSharps(long index, WorksheetRenderUpdateContext updateContext)
        {
            ICellValue cellValue = updateContext.GetCellValue(index);
            NumberCellValue numberCellValue = cellValue as NumberCellValue;
            FormulaCellValue formulaCellValue = cellValue as FormulaCellValue;
            if (numberCellValue == null)
            {
                if (formulaCellValue != null)
                {
                    numberCellValue = formulaCellValue.GetResultValueAsCellValue() as NumberCellValue;
                }

                if (numberCellValue == null)
                {
                    return false;
                }
            }

            CellLayoutBox cellBox = updateContext.GetVisibleCellBox(index);
            Size cellContentSize = updateContext.GetCellContentSize(cellBox);
            CellValueFormat format = updateContext.GetCellFormat(index);

            if (!(cellContentSize.Width.LessOrEqualDouble(cellBox.Size.Width)))
            {
                return true;
            }

            if (format.FormatStringInfo.FormatType == FormatStringType.DateTime &&
                numberCellValue.ToDateTime() == null)
            {
                return true;
            }

            return false;
        }

        public static Thickness CalculateIndentPadding(long index, WorksheetRenderUpdateContext updateContext)
        {
            Guard.ThrowExceptionIfNull(updateContext, "updateContext");

            double leftCellMargin = 0;
            double rightCellMargin = 0;
            if (updateContext.GetCellValue(index) != EmptyCellValue.EmptyValue)
            {
                leftCellMargin = SpreadsheetDefaultValues.LeftCellMargin;

                bool isWrapped = updateContext.GetIsWrapped(index);
                if (isWrapped)
                {
                    rightCellMargin = SpreadsheetDefaultValues.RightCellMargin;
                }
            }

            Thickness indentPadding = new Thickness();
            if (!CellBoxRenderHelper.ShouldDisplayContentAsSharps(index, updateContext))
            {
                double indent = updateContext.GetIndentPropertyValue(index) * SpreadsheetDefaultValues.IndentStep;
                if (indent != 0)
                {
                    RadHorizontalAlignment horizontalAlignment = updateContext.GetHorizontalAlignment(index);

                    if (horizontalAlignment == RadHorizontalAlignment.Left)
                    {
                        indentPadding = new Thickness(indent, 0, 0, 0);
                    }
                    else if (horizontalAlignment == RadHorizontalAlignment.Right)
                    {
                        indentPadding = new Thickness(0, 0, indent, 0);
                    }
                }
            }

            indentPadding.Left += leftCellMargin;
            indentPadding.Right += rightCellMargin;

            return indentPadding;
        }

        public static Rect CalculateActualBoundingRectangleByContentAlignment(CellLayoutBox cellBox, WorksheetRenderUpdateContext updateContext)
        {
            Guard.ThrowExceptionIfNull(updateContext, "updateContext");
            Guard.ThrowExceptionIfNull(cellBox, "cellBox");

            Size cellContentSize = updateContext.GetCellContentSize(cellBox);

            if (ShouldDisplayContentAsSharps(cellBox.LongIndex, updateContext))
            {
                cellContentSize.Width = cellBox.Size.Width;
            }

            RadHorizontalAlignment radAlignment = updateContext.GetHorizontalAlignment(cellBox.LongIndex);
            ICellValue value = updateContext.GetCellValue(cellBox.LongIndex);

            Rect rect = new Rect();

            var horizontalAlignment = CellBoxRenderHelper.GetHorizontalAlignment(radAlignment, value);
            double cellWidth = cellBox.Width;
            if (updateContext.GetIsWrapped(cellBox.LongIndex))
            {
                rect.X = cellBox.Left;
                rect.Width = cellWidth;
            }
            else
            {
                switch (horizontalAlignment)
                {
                    case TextAlignment.Right:
                        rect.X = cellBox.Left + (cellWidth - cellContentSize.Width);
                        break;

                    case TextAlignment.Center:
                        rect.X = cellBox.Left + (cellWidth - cellContentSize.Width) / 2;
                        break;

                    // TODO: Respect the other options.
                    case TextAlignment.Justify:
                    case TextAlignment.Left:
                        rect.X = cellBox.Left;
                        break;

                    default:
                        break;
                }
                rect.Width = cellContentSize.Width;
            }

            var verticalAlignment = updateContext.GetVerticalAlignment(cellBox.LongIndex);
            double cellHeight = cellBox.Height;
            switch (verticalAlignment)
            {
                case RadVerticalAlignment.Top:
                    rect.Y = cellBox.Top;
                    break;

                case RadVerticalAlignment.Center:
                    rect.Y = cellBox.Top + (cellHeight - cellContentSize.Height) / 2;
                    break;

                // TODO: Respect the other options.
                case RadVerticalAlignment.Justify:
                case RadVerticalAlignment.Distributed:
                case RadVerticalAlignment.Undetermined:
                case RadVerticalAlignment.Bottom:
                    rect.Y = cellBox.Top + (cellHeight - cellContentSize.Height);
                    break;

                default:
                    break;
            }
            rect.Height = cellContentSize.Height;

            return rect;
        }

        public static Rect CalculateCellClipping(CellLayoutBox cellBox, WorksheetRenderUpdateContext updateContext)
        {
            CellLayoutBox prevCellBox;
            CellLayoutBox nextCellBox;
            updateContext.GetPreviousNextNonEmptyNonMergedCells(cellBox.LongIndex, out prevCellBox, out nextCellBox);

            Rect actualBoundingRectangle = updateContext.GetActualBoundingRectangleByContentAlignment(cellBox);

            double clippingLeft = Math.Max(0, prevCellBox.BoundingRectangle.Left - actualBoundingRectangle.Left);
            double clippingRight = Math.Min(actualBoundingRectangle.Width, nextCellBox.BoundingRectangle.Right - actualBoundingRectangle.Left);

            double height = cellBox.Height - CellBorder.Default.Thickness;
            height = Math.Max(0, height);

            return new Rect(clippingLeft, Math.Max(0, cellBox.Top - actualBoundingRectangle.Top), Math.Max(0, clippingRight - clippingLeft), height);
        }

        public static Rect CalculateCellClippedBoundingRectangle(CellLayoutBox cellBox, WorksheetRenderUpdateContext updateContext)
        {
            CellLayoutBox prevCellBox;
            CellLayoutBox nextCellBox;
            updateContext.GetPreviousNextNonEmptyNonMergedCells(cellBox.LongIndex, out prevCellBox, out nextCellBox);

            Rect actualBoundingRectangle = updateContext.GetActualBoundingRectangleByContentAlignment(cellBox);

            double clippingLeft = Math.Max(0, prevCellBox.BoundingRectangle.Left - actualBoundingRectangle.Left);
            double clippingRight = Math.Min(actualBoundingRectangle.Width, nextCellBox.BoundingRectangle.Right - actualBoundingRectangle.Left);

            double width = Math.Max(0, clippingRight - clippingLeft);
            double height = cellBox.Height - CellBorder.Default.Thickness;
            height = Math.Max(0, height);

            return new Rect(actualBoundingRectangle.Left + clippingLeft, actualBoundingRectangle.Top + 0, width, height);
        }

        public static Thickness GetPaddingForVerticalAlignment(RadVerticalAlignment verticalAlignment, double desiredHeight, double finalHeight)
        {
            finalHeight -= 1;
            double paddingTop = 0;
            double paddingBottom = 0;
            double paddingDelta = Math.Max(0, (finalHeight - desiredHeight));

            switch (verticalAlignment)
            {
                case RadVerticalAlignment.Top:
                    paddingTop = 0;
                    paddingBottom = paddingDelta;
                    break;
                case RadVerticalAlignment.Center:
                    paddingTop = paddingDelta / 2;
                    paddingBottom = paddingTop;
                    break;
                case RadVerticalAlignment.Bottom:
                case RadVerticalAlignment.Justify:
                case RadVerticalAlignment.Distributed:
                    paddingTop = paddingDelta;
                    break;
                default:
                    break;
            }

            return new Thickness(0, paddingTop, 0, paddingBottom);
        }

        public static TextAlignment GetHorizontalAlignment(RadHorizontalAlignment horizontalAlignment, ICellValue value)
        {
            TextAlignment textAlignment = TextAlignment.Left;

            if (horizontalAlignment == RadHorizontalAlignment.General)
            {
                switch (value.ResultValueType)
                {
                    case CellValueType.Number:
                        textAlignment = TextAlignment.Right;
                        break;

                    case CellValueType.Boolean:
                        textAlignment = TextAlignment.Center;
                        break;

                    case CellValueType.Text:
                        textAlignment = TextAlignment.Left;
                        break;

                    case CellValueType.Error:
                        FormulaCellValue formulaCellValue = value as FormulaCellValue;
                        if (formulaCellValue != null && formulaCellValue.IsValueCyclicReferenceError())
                        {
                            textAlignment = TextAlignment.Right;
                        }
                        else
                        {
                            textAlignment = TextAlignment.Center;
                        }
                        break;

                    default:
                        textAlignment = TextAlignment.Justify;
                        break;
                }
            }
            else
            {
                switch (horizontalAlignment)
                {
                    case RadHorizontalAlignment.Left:
                        textAlignment = TextAlignment.Left;
                        break;

                    case RadHorizontalAlignment.Center:
                    case RadHorizontalAlignment.CenterContinuous:
                        textAlignment = TextAlignment.Center;
                        break;

                    case RadHorizontalAlignment.Right:
                        textAlignment = TextAlignment.Right;
                        break;

                    case RadHorizontalAlignment.Justify:
                        textAlignment = TextAlignment.Justify;
                        break;

                    default:
                        break;
                }
            }

            return textAlignment;
        }
    }
}
