﻿using System;
using System.Collections.Generic;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;

namespace Telerik.Windows.Documents.Spreadsheet.Layout.Layers
{
    internal class CellFillLayer : WorksheetLayerBase
    {
        private readonly IRenderer<PatternFillRenderable> patternFillRenderer;
        private readonly IRenderer<GradientFillRenderable> gradientFillRenderer;

        public CellFillLayer(IRenderer<PatternFillRenderable> patternFillRenderer, IRenderer<GradientFillRenderable> gradientFillRenderer)
        {
            this.patternFillRenderer = patternFillRenderer;
            this.gradientFillRenderer = gradientFillRenderer;
        }

        public override string Name
        {
            get
            {
                return WorksheetPredefinedLayers.CellFill;
            }
        }

        internal IRenderer<PatternFillRenderable> PatternFillRenderer
        {
            get
            {
                return this.patternFillRenderer;
            }
        }

        internal IRenderer<GradientFillRenderable> GradientFillRenderer
        {
            get
            {
                return this.gradientFillRenderer;
            }
        }

        protected override void UpdateRenderOverride(WorksheetRenderUpdateContext worksheetUpdateContext)
        {
            this.UpdatePatternFillsRender(worksheetUpdateContext);
            this.UpdateGradientFillsRender(worksheetUpdateContext);
        }

        private void UpdatePatternFillsRender(WorksheetRenderUpdateContext worksheetUpdateContext)
        {
            foreach (ViewportPane viewportPane in worksheetUpdateContext.SheetViewport.ViewportPanes)
            {
                Dictionary<IFill, PatternFillRenderable> fillToRenderable = new Dictionary<IFill, PatternFillRenderable>();

                foreach (CellLayoutBox cellBox in worksheetUpdateContext.VisibleCellLayoutBoxes[viewportPane.ViewportPaneType])
                {
                    if (cellBox.MergeState == CellMergeState.NonTopLeftCellInMergedRange)
                    {
                        continue;
                    }

                    IFill fill = worksheetUpdateContext.GetFill(cellBox.LongIndex);
                    if (fill == CellPropertyDefinitions.FillProperty.DefaultValue)
                    {
                        continue;
                    }

                    PatternFill patternFill = fill as PatternFill;

                    if (patternFill != null)
                    {
                        PatternFillRenderable renderable;
                        if (!fillToRenderable.TryGetValue(fill, out renderable))
                        {
                            renderable = new PatternFillRenderable()
                            {
                                ColorScheme = worksheetUpdateContext.CurrentTheme.ColorScheme,
                                PatternFill = patternFill
                            };
                            
                            fillToRenderable.Add(fill, renderable);
                            this.ContainerManager.Add(renderable, viewportPane.ViewportPaneType);
                        }

                        renderable.AddVisibleBox(cellBox.BoundingRectangle);
                    }
                }
            }
        }

        private void UpdateGradientFillsRender(WorksheetRenderUpdateContext worksheetUpdateContext)
        {
            foreach (ViewportPane viewportPane in worksheetUpdateContext.SheetViewport.ViewportPanes)
            {
                foreach (CellLayoutBox cellBox in worksheetUpdateContext.VisibleCellLayoutBoxes[viewportPane.ViewportPaneType])
                {
                    if (cellBox.MergeState == CellMergeState.NonTopLeftCellInMergedRange)
                    {
                        continue;
                    }

                    IFill fill = worksheetUpdateContext.GetFill(cellBox.LongIndex);
                    if (fill == CellPropertyDefinitions.FillProperty.DefaultValue)
                    {
                        continue;
                    }

                    GradientFill gradientFill = fill as GradientFill;

                    if (gradientFill != null)
                    {
                        GradientFillRenderable renderable = new GradientFillRenderable()
                        {
                            GradientFill = gradientFill,
                            ColorScheme = worksheetUpdateContext.CurrentTheme.ColorScheme,
                            BoundingRectangle = cellBox.BoundingRectangle
                        };

                        this.ContainerManager.Add(renderable, viewportPane.ViewportPaneType);
                    }
                }
            }
        }

        protected override void TranslateAndScale(WorksheetRenderUpdateContext worksheetUpdateContext)
        {
            foreach (ViewportPane viewportPane in worksheetUpdateContext.SheetViewport.ViewportPanes)
            {
                ViewportPaneType paneType = viewportPane.ViewportPaneType;

                foreach (IRenderable renderable in this.ContainerManager.GetElementsForViewportPane(paneType))
                {
                    PatternFillRenderable patternFillRenderable = renderable as PatternFillRenderable;
                    GradientFillRenderable gradientFillRenderable = renderable as GradientFillRenderable;

                    if (patternFillRenderable != null)
                    {
                        this.TranslateAndScale(patternFillRenderable, paneType, worksheetUpdateContext);
                    }
                    else if (gradientFillRenderable != null)
                    {
                        this.TranslateAndScale(gradientFillRenderable, paneType, worksheetUpdateContext);
                    }
                    else
                    {
                        throw new InvalidCastException("Invalid fill renderable type!");
                    }
                }
            }
        }

        private void TranslateAndScale(PatternFillRenderable patternFillRenderable, ViewportPaneType paneType, WorksheetRenderUpdateContext worksheetUpdateContext)
        {
            List<Rect> translatedRectangles = new List<Rect>();

            foreach (Rect rectangle in patternFillRenderable.VisibleBoxes)
            {
                translatedRectangles.Add(base.TranslateAndScale(rectangle, paneType, worksheetUpdateContext));
            }

            patternFillRenderable.ClearVisibleBoxes();
            patternFillRenderable.AddVisibleBoxes(translatedRectangles);

            this.patternFillRenderer.Render(patternFillRenderable, paneType);
        }

        private void TranslateAndScale(GradientFillRenderable gradientFillRenderable, ViewportPaneType paneType, WorksheetRenderUpdateContext worksheetUpdateContext)
        {
            gradientFillRenderable.BoundingRectangle = this.TranslateAndScale(gradientFillRenderable.BoundingRectangle, paneType, worksheetUpdateContext);

            this.gradientFillRenderer.Render(gradientFillRenderable, paneType);
        }
    }
}
