﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Media;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings;
using Telerik.Windows.Documents.Spreadsheet.Measurement;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Layout.Layers
{
    internal class CellValuesLayer : CellBasedLayerBase
    {
        private readonly IRenderer<TextBlockRenderable> textRenderer;

        public CellValuesLayer(IRenderer<TextBlockRenderable> textRenderer)
        {
            this.textRenderer = textRenderer;
        }

        public override string Name
        {
            get
            {
                return WorksheetPredefinedLayers.CellValues;
            }
        }

        internal IRenderer<TextBlockRenderable> TextRenderer
        {
            get
            {
                return this.textRenderer;
            }
        }

        internal static bool ShouldCreateElementForBox(CellLayoutBox box, WorksheetRenderUpdateContext updateContext)
        {
            CellValueFormat format = updateContext.GetCellFormat(box.LongIndex);
            string value = updateContext.GetCellValue(box.LongIndex).GetAsStringOrNull(format);

            return !string.IsNullOrEmpty(value);
        }

        protected override IRenderable CreateRenderableElementForBox(ViewportPaneType viewportPaneType, CellLayoutBox box, WorksheetRenderUpdateContext updateContext)
        {
            if (!ShouldCreateElementForBox(box, updateContext))
            {
                return null;
            }

            return CellValuesLayer.CreateTextBlockRenderableForBox(box, updateContext);
        }

        internal static TextBlockRenderable CreateTextBlockRenderableForBox(CellLayoutBox box, WorksheetRenderUpdateContext updateContext)
        {
            TextBlockRenderable textBlockRenderable = new TextBlockRenderable();
            
            ICellValue cellValue = updateContext.GetCellValue(box.LongIndex);
            CellValueFormat cellFormat = updateContext.GetCellFormat(box.LongIndex);//.Worksheet.Cells[cellBox.Index].GetFormat().Value;

            CellValuesLayer.SetTextBlockContent(textBlockRenderable, cellValue, cellFormat, box, updateContext);

            FontProperties fontProperties = updateContext.GetFontProperties(box.LongIndex);

            RadHorizontalAlignment horizontalAlignment = updateContext.GetHorizontalAlignment(box.LongIndex);
            textBlockRenderable.FontFamily = fontProperties.FontFamily;
            textBlockRenderable.FontSize = fontProperties.FontSize;
            textBlockRenderable.FontWeight = fontProperties.IsBold ? FontWeights.Bold : FontWeights.Normal;
            textBlockRenderable.FontStyle = fontProperties.IsItalic ? FontStyles.Italic : FontStyles.Normal;
            textBlockRenderable.TextDecorations = (fontProperties.Underline != UnderlineType.None) ? TextDecorations.Underline : null;

            TextAlignment textAlignment = CellBoxRenderHelper.GetHorizontalAlignment(horizontalAlignment, cellValue);
            if (textAlignment == TextAlignment.Justify)
            {
                // NOTE: NOT SUPPORTED IN SL.
                textAlignment = TextAlignment.Left;
            }
            textBlockRenderable.TextAlignment = textAlignment;

            bool isWrapped = updateContext.GetIsWrapped(box.LongIndex);
            if (isWrapped)
            {
                textBlockRenderable.TextWrapping = TextWrapping.Wrap;
            }
            else
            {
                textBlockRenderable.TextWrapping = TextWrapping.NoWrap;
            }

            Thickness indentPadding = CellBoxRenderHelper.CalculateIndentPadding(box.LongIndex, updateContext);
            Thickness padding = new Thickness(indentPadding.Left, 0, indentPadding.Right, 0);
            textBlockRenderable.Padding = padding;

            return textBlockRenderable;
        }

        private static void SetTextBlockContent(TextBlockRenderable textBlockRenderable, ICellValue cellValue, CellValueFormat format, CellLayoutBox cellBox, WorksheetRenderUpdateContext updateContext)
        {
            double cellBoxWidth = cellBox.Width;
            FontProperties fontProperties = updateContext.GetFontProperties(cellBox.LongIndex);

            textBlockRenderable.Foreground = fontProperties.ForeColor.GetActualValue(updateContext.CurrentTheme);

            try
            {
                FormulaCellValue formulaCellValue = cellValue as FormulaCellValue;
                if (formulaCellValue != null)
                {
                    cellValue = formulaCellValue.GetResultValueAsCellValue();
                }

                CellValueFormatResult cellValueFormatResult = format.GetFormatResult(cellValue);
                IEnumerable<CellValueFormatResultItem> infos = cellValueFormatResult.Infos;
                string infosText = cellValueFormatResult.InfosText;
                int infosCount = infos.Count();
                bool shouldExpand = cellValueFormatResult.ContainsExpandableInfo;

                Size cellContentSize = updateContext.GetCellContentSize(cellBox);

                if (CellBoxRenderHelper.ShouldDisplayContentAsSharps(cellBox.LongIndex, updateContext))
                {
                    textBlockRenderable.Text = FormatHelper.ExpandSymbol("#", cellBoxWidth, fontProperties);
                    return;
                }

                if (cellValueFormatResult.Foreground.HasValue)
                {
                    textBlockRenderable.Foreground = cellValueFormatResult.Foreground.Value;
                }

                textBlockRenderable.ShouldFitToCellSize = false;

                if (infosCount == 1)
                {
                    textBlockRenderable.Text = infosText;
                    if (cellValueFormatResult.Foreground.HasValue)
                    {
                        textBlockRenderable.Foreground = cellValueFormatResult.Foreground.Value;
                    }

                    return;
                }

                RunRenderable runToBeExpanded = null;
                textBlockRenderable.ClearRuns();
                foreach (CellValueFormatResultItem info in infos)
                {
                    RunRenderable run = new RunRenderable();
                    run.Text = info.Text;
                    if (info.IsTransparent)
                    {
                        run.Foreground = new SolidColorBrush(Colors.Transparent);
                    }

                    textBlockRenderable.AddRun(run);

                    if (info.ShouldExpand)
                    {
                        runToBeExpanded = run;
                    }

                    if (!info.ApplyFormat && shouldExpand)
                    {
                        RunRenderable dividingRun = new RunRenderable();
                        dividingRun.Text = FormatHelper.InvisibleDivider;
                        dividingRun.FontSize = FormatHelper.InvisibleDividerFontsize;
                        textBlockRenderable.AddRun(dividingRun);
                    }
                }

                if (runToBeExpanded != null)
                {
                    // TODO: Rething the use of RadTextMeasurer.Measure method instead of WorksheetRenderUpdateContext.measureMultilineText field!
                    // The same issue exists in FormatHelper.ExpandSymbol and RenderHelper.CreateSpaceRun methods, which are used below!
                    double symbolWidth = RadTextMeasurer.Measure(runToBeExpanded.Text, fontProperties).Size.Width;

                    double availableWidth = Math.Max(0, cellBoxWidth - (cellContentSize.Width - symbolWidth + CellBorder.Default.Thickness));
                    runToBeExpanded.Text = FormatHelper.ExpandSymbol(runToBeExpanded.Text, availableWidth, fontProperties);

                    double gapWidth = availableWidth - RadTextMeasurer.Measure(runToBeExpanded.Text, fontProperties).Size.Width;
                    gapWidth = Math.Round(gapWidth, 4);
                    if (gapWidth > 0)
                    {
                        RunRenderable spaceRun = RenderHelper.CreateSpaceRun(gapWidth);
                        textBlockRenderable.InsertRunAfter(runToBeExpanded, spaceRun);
                    }

                    textBlockRenderable.ShouldFitToCellSize = true;
                }
            }
            finally
            {
                textBlockRenderable.CellValue = cellValue;
                textBlockRenderable.CellValueFormat = format;
                textBlockRenderable.CellBox = cellBox;
            }
        }

        protected override void TranslateAndScale(WorksheetRenderUpdateContext worksheetUpdateContext)
        {
            ScaleTransform scaleTransform = new ScaleTransform
            {
                ScaleX = worksheetUpdateContext.ScaleFactor.Width,
                ScaleY = worksheetUpdateContext.ScaleFactor.Height,
            };

            foreach (var pair in worksheetUpdateContext.VisibleCellLayoutBoxes)
            {
                ViewportPaneType paneType = pair.Key;

                foreach (IRenderable child in this.ContainerManager.GetElementsForViewportPane(paneType))
                {
                    TextBlockRenderable textRenderable = (TextBlockRenderable)child;
                    CellLayoutBox box = textRenderable.CellBox;
                    Rect boundingRectangle = worksheetUpdateContext.GetActualBoundingRectangleByContentAlignment(box);

                    Rect clipRectangle = worksheetUpdateContext.GetCellClipping(box);
                    if (textRenderable.ShouldFitToCellSize)
                    {
                        boundingRectangle.X = box.BoundingRectangle.Left;
                        boundingRectangle.Width = box.BoundingRectangle.Width;

                        clipRectangle.X = 0;
                        clipRectangle.Width = boundingRectangle.Width;
                    }

                    textRenderable.Clip = clipRectangle;

                    boundingRectangle = this.Translate(boundingRectangle, pair.Key, worksheetUpdateContext);
                    textRenderable.TopLeft = new Point(boundingRectangle.Left, boundingRectangle.Top);
                    textRenderable.ScaleFactor = worksheetUpdateContext.ScaleFactor;

                    this.textRenderer.Render(textRenderable, paneType);
                }
            }

            foreach (var pair in worksheetUpdateContext.VisibleCellLayoutBoxes)
            {
                foreach (CellLayoutBox cellBox in pair.Value)
                {
                    HyperlinkInfo hyperlinkInfo;

                    long index = cellBox.LongIndex;

                    if (worksheetUpdateContext.CellIndexToHyperlinkInfo.TryGetValue(index, out hyperlinkInfo))
                    {
                        Rect rect = worksheetUpdateContext.CellIndexToHyperlinkArea[index];

                        Point topLeft = new Point(rect.Left, rect.Top);
                        topLeft = this.Translate(topLeft, pair.Key, worksheetUpdateContext);

                        topLeft = scaleTransform.Transform(topLeft);

                        Point widthHeight = scaleTransform.Transform(new Point(rect.Width, rect.Height));

                        rect = new Rect(topLeft.X, topLeft.Y, widthHeight.X, widthHeight.Y);

                        worksheetUpdateContext.CellIndexToHyperlinkArea[index] = rect;
                    }
                }
            }
        }
    }
}
