﻿using System;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Model.Printing;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Layout.Layers
{
    internal class HeaderFooterLayer : PrintWorksheetLayerBase
    {
        private readonly HeaderFooterSectionRenderer sectionRenderer;

        public HeaderFooterLayer(HeaderFooterSectionRenderer renderer)
        {
            Guard.ThrowExceptionIfNull(renderer, "renderer");

            this.sectionRenderer = renderer;
        }

        public override string Name
        {
            get
            {
                return WorksheetPredefinedLayers.HeaderFooter;
            }
        }

        protected override void UpdateRenderOverride(PrintWorksheetRenderUpdateContext worksheetUpdateContext)
        {
            base.UpdateRenderOverride(worksheetUpdateContext);

            HeaderFooterRenderContext context = worksheetUpdateContext.HeaderFooterContext;
            HeaderFooterSettings settings = context.Worksheet.WorksheetPageSetup.HeaderFooterSettings;
            int pageNumber = context.PageNumber;
            HeaderFooterContent header, footer;

            if (settings.DifferentFirstPage && pageNumber == 1)
            {
                header = settings.FirstPageHeader;
                footer = settings.FirstPageFooter;
            }
            else if (settings.DifferentOddAndEvenPages && ((pageNumber & 1) == 0))
            {
                header = settings.EvenPageHeader;
                footer = settings.EvenPageFooter;
            }
            else
            {
                header = settings.Header;
                footer = settings.Footer;
            }
            
            this.UpdateHeaderContentRender(header, context);
            this.UpdateFooterContentRender(footer, context);
        }

        private void UpdateHeaderContentRender(HeaderFooterContent content, HeaderFooterRenderContext context)
        {
            this.UpdateHeaderSectionRender(content.LeftSection, TextAlignment.Left, context);
            this.UpdateHeaderSectionRender(content.CenterSection, TextAlignment.Center, context);
            this.UpdateHeaderSectionRender(content.RightSection, TextAlignment.Right, context);
        }

        private void UpdateFooterContentRender(HeaderFooterContent content, HeaderFooterRenderContext context)
        {
            this.UpdateFooterSectionRender(content.LeftSection, TextAlignment.Left, context);
            this.UpdateFooterSectionRender(content.CenterSection, TextAlignment.Center, context);
            this.UpdateFooterSectionRender(content.RightSection, TextAlignment.Right, context);
        }

        private void UpdateHeaderSectionRender(HeaderFooterSection headerSection, TextAlignment textAlignment, HeaderFooterRenderContext context)
        {
            if (!headerSection.IsEmpty)
            {
                double top = context.Worksheet.WorksheetPageSetup.Margins.Header;
                HeaderSectionRenderable renderable = new HeaderSectionRenderable() { Top = top };
                HeaderFooterLayer.SetHeaderFooterSectionProperties(renderable, headerSection, textAlignment, context);
                this.sectionRenderer.Render(renderable, ViewportPaneType.Scrollable);
            }
        }

        private void UpdateFooterSectionRender(HeaderFooterSection footerSection, TextAlignment textAlignment, HeaderFooterRenderContext context)
        {
            if (!footerSection.IsEmpty)
            {
                double bottom = context.Worksheet.WorksheetPageSetup.RotatedPageSize.Height - context.Worksheet.WorksheetPageSetup.Margins.Footer;
                FooterSectionRenderable renderable = new FooterSectionRenderable() { Bottom =  bottom };
                HeaderFooterLayer.SetHeaderFooterSectionProperties(renderable, footerSection, textAlignment, context);
                this.sectionRenderer.Render(renderable, ViewportPaneType.Scrollable);
            }
        }

        private static void SetHeaderFooterSectionProperties(HeaderFooterSectionRenderable renderable, HeaderFooterSection headerSection, TextAlignment textAlignment, HeaderFooterRenderContext context)
        {
            renderable.Alignment = textAlignment;
            renderable.BlockWidth = HeaderFooterLayer.CalculateHeaderFooterSectionScaledWidth(context.Worksheet.WorksheetPageSetup);
            renderable.HeaderFooterContext = context;
            renderable.Left = HeaderFooterLayer.GetHeaderFooterLeft(context.Worksheet.WorksheetPageSetup);
            renderable.ScaleFactor = HeaderFooterLayer.GetHeaderFooterScale(context.Worksheet.WorksheetPageSetup);
            renderable.Section = headerSection;
        }

        private static double CalculateHeaderFooterSectionScaledWidth(WorksheetPageSetup pageSetup)
        {
            double left = GetHeaderFooterLeft(pageSetup);
            double right = GetHeaderFooterRigth(pageSetup);
            double pageWidth = pageSetup.RotatedPageSize.Width;
            double width = pageWidth - left - right;
            double scale = HeaderFooterLayer.GetHeaderFooterScale(pageSetup);
            double scaledWidth = width / scale;

            return scaledWidth;
        }

        private static double GetHeaderFooterScale(WorksheetPageSetup pageSetup)
        {
            return pageSetup.HeaderFooterSettings.ScaleWithDocument ? pageSetup.ScaleFactor.Width : 1;
        }

        private static double GetHeaderFooterLeft(WorksheetPageSetup pageSetup)
        {
            PageMargins margins = pageSetup.HeaderFooterSettings.AlignWithPageMargins ? pageSetup.Margins : PageMargins.NormalMargins;

            return margins.Left;
        }

        private static double GetHeaderFooterRigth(WorksheetPageSetup pageSetup)
        {
            PageMargins margins = pageSetup.HeaderFooterSettings.AlignWithPageMargins ? pageSetup.Margins : PageMargins.NormalMargins;

            return margins.Right;
        }
    }
}
