﻿using System;
using System.Windows;
using System.Windows.Media;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Printing;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Layout.Layers
{
    internal abstract class HeaderFooterSectionRenderer : HeaderFooterSectionTextRendererBase, IRenderer<HeaderSectionRenderable>, IRenderer<FooterSectionRenderable>
    {
        private HeaderFooterSectionRenderable renderable;

        protected sealed override Sheet SheetContext
        {
            get
            {
                return this.renderable.HeaderFooterContext.Worksheet;
            }
        }

        public void Render(HeaderSectionRenderable renderable, ViewportPaneType paneType)
        {
            this.Render(renderable, this.CalculateMatrix); 
        }

        public void Render(FooterSectionRenderable renderable, ViewportPaneType paneType)
        {
            this.Render(renderable, this.CalculateMatrix);
        }

        public static string GetDateContent(DateTime dateTime)
        {
            return dateTime.ToShortDateString();
        }

        public static string GetTimeContent(DateTime dateTime)
        {
            return dateTime.ToShortTimeString();
        }  

        protected abstract void PrepareForRendering(TextAlignment alignment);
        protected abstract double MeasureHeightBeforeRendering(double blockWidth);
        protected abstract void RenderOverride(Matrix positionMarix, double blockWidth);

        protected sealed override void OnAppendingAmpersandOverride()
        {
            this.OnAppendingTextFragmentOverride("&");
        }

        protected sealed override void OnAppendingFileNameOverride()
        {
            this.OnAppendingTextFragmentOverride(this.renderable.HeaderFooterContext.Worksheet.Workbook.Name);
        }

        protected sealed override void OnAppendingPageNumberOverride()
        {
            this.OnAppendingTextFragmentOverride(this.renderable.HeaderFooterContext.PageNumber.ToString());
        }

        protected sealed override void OnAppendingAddToPageNumberOverride()
        {
            // Do nothing.
        }

        protected sealed override void OnAppendingSubstractFromPageNumberOverride()
        {
            // Do nothing
        }

        protected sealed override void OnAppendingNumberOfPagesOverride()
        {
            this.OnAppendingTextFragmentOverride(this.renderable.HeaderFooterContext.NumberOfPages.ToString());
        }

        protected sealed override void OnAppendingSheetNameOverride()
        {
            this.OnAppendingTextFragmentOverride(this.renderable.HeaderFooterContext.Worksheet.Name);
        }

        protected sealed override void OnAppendingFilePathOverride()
        {
            // Do nothing because in RadSpreadProcessing the file is not linked to a file path.
        }

        protected sealed override void OnAppendingPictureOverride()
        {
            // Do nothing because pictures are not supported yet.
            // Only first picture should be rendered.
            // All other pictures should be skipped.
        }

        protected sealed override void OnAppendingDateOverride()
        {
            string dateText = HeaderFooterSectionRenderer.GetDateContent(this.renderable.HeaderFooterContext.DateTime);
            this.OnAppendingTextFragmentOverride(dateText);
        }

        protected sealed override void OnAppendingTimeOverride()
        {
            string timeText = HeaderFooterSectionRenderer.GetTimeContent(this.renderable.HeaderFooterContext.DateTime);
            this.OnAppendingTextFragmentOverride(timeText);
        }

        private void Render<T>(T renderable, Func<T, Matrix> calculateMatrix)
            where T : HeaderFooterSectionRenderable
        {
            Guard.ThrowExceptionIfNull(renderable, "renderable");
            Guard.ThrowExceptionIfNotNull(this.renderable, "this.renderable");
            this.renderable = renderable;
            
            this.PrepareForRendering(this.renderable.Alignment);
            this.Render(this.renderable.Section.Text);
            Matrix positionMatrix = calculateMatrix(renderable);
            this.RenderOverride(positionMatrix, this.renderable.BlockWidth);

            this.renderable = null;
        }

        private Matrix CalculateMatrix(HeaderSectionRenderable renderable)
        {
            return this.CalculateMatrix(renderable.Left, renderable.Top, renderable.ScaleFactor);
        }

        private Matrix CalculateMatrix(FooterSectionRenderable renderable)
        {
            double height = this.MeasureHeightBeforeRendering(renderable.BlockWidth);
            double scaledHeight = height * renderable.ScaleFactor;
            double top = renderable.Bottom - scaledHeight;

            return this.CalculateMatrix(renderable.Left, top, renderable.ScaleFactor);
        }

        private Matrix CalculateMatrix(double left, double top, double scale)
        {
            return new Matrix(scale, 0, 0, scale, left, top);
        }
    }
}
