﻿using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Layout.Layers
{
    internal abstract class LayerBase : INamedObject
    {
        private readonly LayerContainerManager containerManager;

        public LayerBase()
        {
            this.containerManager = new LayerContainerManager();
        }

        public abstract string Name { get; }

        public LayerContainerManager ContainerManager
        {
            get
            {
                return this.containerManager;
            }
        }

        /// <summary>
        /// Updates the UI.
        /// </summary>
        /// <param name="updateContext">The update context.</param>
        public void UpdateRender(RenderUpdateContext updateContext)
        {
            this.Clear();

            this.UpdateRenderOverride(updateContext);
            this.OnRenderUpdated();
            this.TranslateAndScale(updateContext);
        }

        protected virtual void UpdateRenderOverride(RenderUpdateContext updateContext)
        {
        }

        protected virtual void OnRenderUpdated()
        {
        }

        protected virtual void TranslateAndScale(RenderUpdateContext updateContext)
        {
        }

        protected virtual void Clear()
        {
            this.ContainerManager.Clear();
        }

        /// <summary>
        /// Translates and scales a point.
        /// </summary>
        /// <param name="point">The point.</param>
        /// <param name="viewportPaneType">Type of the viewport pane.</param>
        /// <param name="updateContext">The update context.</param>
        /// <returns>The translated and scaled point.</returns>
        protected Point TranslateAndScale(Point point, ViewportPaneType viewportPaneType, RenderUpdateContext updateContext)
        {
            point = this.Translate(point, viewportPaneType, updateContext);
            point = updateContext.ScaleTransform.Transform(point);

            return point;
        }

        /// <summary>
        /// Scales the specified size.
        /// </summary>
        /// <param name="size">The size to scale.</param>
        /// <param name="updateContext">The update context.</param>
        /// <returns>The scaled size.</returns>
        protected Size Scale(Size size, RenderUpdateContext updateContext)
        {
            size.Width = SpreadsheetHelper.Transform(updateContext.ScaleTransform, size.Width);
            size.Height = SpreadsheetHelper.Transform(updateContext.ScaleTransform, size.Height);

            return size;
        }

        /// <summary>
        /// Scales the specified point.
        /// </summary>
        /// <param name="point">The point.</param>
        /// <param name="updateContext">The update context.</param>
        /// <returns>The scaled point.</returns>
        protected Point Scale(Point point, RenderUpdateContext updateContext)
        {
            point = updateContext.ScaleTransform.Transform(point);

            return point;
        }

        /// <summary>
        /// Scales the specified value.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <param name="updateContext">The update context.</param>
        /// <returns>Scaled value.</returns>
        protected double Scale(double value, RenderUpdateContext updateContext)
        {
            return SpreadsheetHelper.Transform(updateContext.ScaleTransform, value);
        }

        /// <summary>
        /// Translates the specified point.
        /// </summary>
        /// <param name="point">The point.</param>
        /// <param name="viewportPaneType">Type of the viewport pane.</param>
        /// <param name="updateContext">The update context.</param>
        /// <returns>The translated point.</returns>
        protected Point Translate(Point point, ViewportPaneType viewportPaneType, RenderUpdateContext updateContext)
        {
            return updateContext.Translate(point, viewportPaneType);
        }

        /// <summary>
        /// Translates the specified rectangle.
        /// </summary>
        /// <param name="rect">The rectangle.</param>
        /// <param name="viewportPaneType">Type of the viewport pane.</param>
        /// <param name="updateContext">The update context.</param>
        /// <returns>The translated rectangle.</returns>
        protected Rect Translate(Rect rect, ViewportPaneType viewportPaneType, RenderUpdateContext updateContext)
        {
            return updateContext.Translate(rect, viewportPaneType);
        }

        protected Rect TranslateAndScale(Rect rect, ViewportPaneType viewportPaneType, RenderUpdateContext updateContext)
        {
            return updateContext.TranslateAndScale(rect, viewportPaneType);
        }
    }
}
