﻿using System;
using System.Windows;
using System.Windows.Media;

namespace Telerik.Windows.Documents.Spreadsheet.Layout.Layers
{
    internal class RenderUpdateContext
    {
        private readonly SheetViewport sheetViewport;
        private readonly Size scaleFactor;
        private ScaleTransform scaleTransform;

        public RenderUpdateContext(SheetViewport sheetViewport, Size scaleFactor)
        {
            this.sheetViewport = sheetViewport;
            this.scaleFactor = scaleFactor;
        }

        /// <summary>
        /// Gets the sheet viewport.
        /// </summary>
        /// <value>The sheet viewport.</value>
        public SheetViewport SheetViewport
        {
            get
            {
                return this.sheetViewport;
            }
        }

        /// <summary>
        /// Gets the scale factor.
        /// </summary>
        /// <value>The scale factor.</value>
        public Size ScaleFactor
        {
            get
            {
                return this.scaleFactor;
            }
        }

        /// <summary>
        /// Gets the scale transform.
        /// </summary>
        /// <value>The scale transform.</value>
        public ScaleTransform ScaleTransform
        {
            get
            {
                if (this.scaleTransform == null)
                {
                    this.scaleTransform = new ScaleTransform
                    {
                        ScaleX = this.ScaleFactor.Width,
                        ScaleY = this.ScaleFactor.Height,
                    };
                }

                return this.scaleTransform;
            }
        }

        /// <summary>
        /// Translates the specified point.
        /// </summary>
        /// <param name="point">The point.</param>
        /// <param name="viewportPaneType">Type of the viewport pane.</param>
        /// <returns>The translated point.</returns>
        public Point Translate(Point point, ViewportPaneType viewportPaneType)
        {
            return this.SheetViewport.Translate(point, viewportPaneType);
        }

        /// <summary>
        /// Translates the specified rectangle.
        /// </summary>
        /// <param name="rect">The rectangle.</param>
        /// <param name="viewportPaneType">Type of the viewport pane.</param>
        /// <returns>The translated rectangle.</returns>
        public Rect Translate(Rect rect, ViewportPaneType viewportPaneType)
        {
            Point bottomRight = new Point(rect.Right, rect.Bottom);
            Point translatedBottomRight = this.Translate(bottomRight, viewportPaneType);

            Point topLeft = new Point(translatedBottomRight.X - rect.Width, translatedBottomRight.Y - rect.Height);

            return new Rect(topLeft, translatedBottomRight);
        }

        public Rect TranslateAndScale(Rect rect, ViewportPaneType viewportPaneType)
        {
            Rect translatedRect = this.Translate(rect, viewportPaneType);
            ScaleTransform scaleTransform = new ScaleTransform()
            {
                ScaleX = this.ScaleFactor.Width,
                ScaleY = this.ScaleFactor.Height
            };

            Point topLeft = scaleTransform.Transform(new Point(translatedRect.Left, translatedRect.Top));
            Point widthHeight = scaleTransform.Transform(new Point(translatedRect.Width, translatedRect.Height));

            return new Rect(topLeft.X, topLeft.Y, widthHeight.X, widthHeight.Y);
        }
    }
}
