﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Media;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.DataValidation;
using Telerik.Windows.Documents.Spreadsheet.Model.Shapes;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Theming;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Spreadsheet.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Maths;

namespace Telerik.Windows.Documents.Spreadsheet.Layout.Layers
{
    internal class WorksheetRenderUpdateContext : RenderUpdateContext
    {
        #region Constants

        private static readonly int MaxColumnsToCheckForLongText = 30;
        private static readonly Size HiddenCellSize = new Size(0, 0);

        #endregion


        #region Fields

        private readonly RadWorksheetLayout layout;
        private readonly List<RowLayoutBox> visibleRowBoxes;
        private readonly List<ColumnLayoutBox> visibleColumnBoxes;
        private readonly Dictionary<ViewportPaneType, IEnumerable<CellLayoutBox>> visibleCellBoxes;
        private Dictionary<ViewportPaneType, Dictionary<long, CellLayoutBox>> cellIndexToCellLayoutBox;

        private readonly Dictionary<long, ICellValue> cellIndexToValue;
        private readonly Dictionary<long, CellValueFormat> cellIndexToFormat;
        private readonly Dictionary<long, bool> cellindexToIsWrapped;
        private readonly Dictionary<long, FontProperties> cellIndexToFontProperties;
        private readonly Dictionary<long, IFill> cellIndexToFill;
        private readonly Dictionary<long, Rect> cellIndexToActualBoundingRectangle;
        private readonly Dictionary<long, Rect> cellIndexToClipping;
        private readonly Dictionary<long, int> cellIndexToIndentPropertyValue;
        private readonly Dictionary<long, RadHorizontalAlignment> cellIndexToHorizontalAlignment;
        private readonly Dictionary<long, RadVerticalAlignment> cellIndexToVerticalAlignment;
        private readonly Dictionary<long, Size> cellIndexToContentSize;
        private readonly Dictionary<long, CellBorder> cellIndexToLeftBorder;
        private readonly Dictionary<long, CellBorder> cellIndexToTopBorder;
        private readonly Dictionary<long, CellBorder> cellIndexToRightBorder;
        private readonly Dictionary<long, CellBorder> cellIndexToBottomBorder;
        private readonly Dictionary<long, CellBorder> cellIndexToDiagonalUpBorder;
        private readonly Dictionary<long, CellBorder> cellIndexToDiagonalDownBorder;
        private readonly Dictionary<long, Tuple<CellLayoutBox, CellLayoutBox>> cellIndexToPreviousNextNonEmptyNonMergedCells;
        private readonly Dictionary<long, HyperlinkInfo> cellIndexToHyperlinkInfo;
        private readonly Dictionary<long, Rect> cellIndexToHyperlinkArea;
        private readonly Dictionary<long, IDataValidationRule> cellIndexToDataValidationRule;
        private readonly Dictionary<long, bool> cellIndexToDataValidationResult;
        private readonly List<FloatingShapeBase> shapes;
        private readonly RowHeight defaultRowHeight;
        private readonly ColumnWidth defaultColumnWidth;

        private CellRange[] invalidatedCellRanges;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the worksheet layout.
        /// </summary>
        /// <value>The worksheet layout.</value>
        internal RadWorksheetLayout Layout
        {
            get
            {
                return this.layout;
            }
        }

        /// <summary>
        /// Gets the worksheet.
        /// </summary>
        /// <value>The worksheet.</value>
        public Worksheet Worksheet
        {
            get
            {
                return this.layout.Worksheet;
            }
        }

        /// <summary>
        /// Gets the current theme.
        /// </summary>
        /// <value>The current theme.</value>
        public DocumentTheme CurrentTheme
        {
            get
            {
                return this.Worksheet.Workbook.Theme;
            }
        }

        /// <summary>
        /// Gets the cell index to hyperlink info.
        /// </summary>
        /// <value>The cell index to hyperlink info.</value>
        internal Dictionary<long, HyperlinkInfo> CellIndexToHyperlinkInfo
        {
            get
            {
                return this.cellIndexToHyperlinkInfo;
            }
        }

        /// <summary>
        /// Gets the cell index to hyperlink area.
        /// </summary>
        /// <value>The cell index to hyperlink area.</value>
        internal Dictionary<long, Rect> CellIndexToHyperlinkArea
        {
            get
            {
                return this.cellIndexToHyperlinkArea;
            }
        }

        /// <summary>
        /// Gets the visible row boxes.
        /// </summary>
        /// <value>The visible row boxes.</value>
        public IEnumerable<RowLayoutBox> VisibleRowBoxes
        {
            get
            {
                return this.visibleRowBoxes;
            }
        }

        /// <summary>
        /// Gets the visible column boxes.
        /// </summary>
        /// <value>The visible column boxes.</value>
        public IEnumerable<ColumnLayoutBox> VisibleColumnBoxes
        {
            get
            {
                return this.visibleColumnBoxes;
            }
        }

        /// <summary>
        /// Gets the layout boxes of the cells which are currently visible.
        /// </summary>
        /// <value>The layout boxes of the cells which are currently visible.</value>
        public Dictionary<ViewportPaneType, IEnumerable<CellLayoutBox>> VisibleCellLayoutBoxes
        {
            get
            {
                return this.visibleCellBoxes;
            }
        }

        /// <summary>
        /// Gets the visible pictures.
        /// </summary>
        /// <value>The visible pictures.</value>
        public IEnumerable<FloatingShapeBase> Shapes
        {
            get
            {
                return this.shapes;
            }
        }

        #endregion


        #region Constructors

        internal WorksheetRenderUpdateContext(RadWorksheetLayout worksheetLayout, SheetViewport sheetViewport, Size scaleFactor, WorksheetRenderUpdateContext oldContext = null)
            : base(sheetViewport, scaleFactor)
        {
            this.layout = worksheetLayout;

            this.Layout.GetVisibleBoxes(this.SheetViewport, out this.visibleRowBoxes, out this.visibleColumnBoxes, out this.visibleCellBoxes);

            this.defaultColumnWidth = this.Worksheet.DefaultColumnWidth;
            this.defaultRowHeight = this.Worksheet.DefaultRowHeight;

            this.InitializeInvalidatedRanges(oldContext);

            this.cellIndexToValue = new Dictionary<long, ICellValue>();
            this.cellIndexToFormat = new Dictionary<long, CellValueFormat>();
            this.cellIndexToIndentPropertyValue = new Dictionary<long, int>();
            this.cellIndexToHorizontalAlignment = new Dictionary<long, RadHorizontalAlignment>();
            this.cellIndexToVerticalAlignment = new Dictionary<long, RadVerticalAlignment>();
            this.cellindexToIsWrapped = new Dictionary<long, bool>();
            this.cellIndexToFontProperties = new Dictionary<long, FontProperties>();
            this.cellIndexToContentSize = new Dictionary<long, Size>();
            this.cellIndexToActualBoundingRectangle = new Dictionary<long, Rect>();
            this.cellIndexToClipping = new Dictionary<long, Rect>();

            this.cellIndexToLeftBorder = new Dictionary<long, CellBorder>();
            this.cellIndexToTopBorder = new Dictionary<long, CellBorder>();
            this.cellIndexToRightBorder = new Dictionary<long, CellBorder>();
            this.cellIndexToBottomBorder = new Dictionary<long, CellBorder>();
            this.cellIndexToDiagonalUpBorder = new Dictionary<long, CellBorder>();
            this.cellIndexToDiagonalDownBorder = new Dictionary<long, CellBorder>();

            this.cellIndexToFill = new Dictionary<long, IFill>();

            this.cellIndexToPreviousNextNonEmptyNonMergedCells = new Dictionary<long, Tuple<CellLayoutBox, CellLayoutBox>>();

            this.shapes = new List<FloatingShapeBase>();

            this.cellIndexToHyperlinkInfo = new Dictionary<long, HyperlinkInfo>();
            this.cellIndexToHyperlinkArea = new Dictionary<long, Rect>();
            this.cellIndexToDataValidationRule = new Dictionary<long, IDataValidationRule>();
            this.cellIndexToDataValidationResult = new Dictionary<long, bool>();

            this.InitCaches(oldContext);

            this.AddCellsWithLargeContentToVisibleCellBoxes();
        }

        private void InitializeInvalidatedRanges(WorksheetRenderUpdateContext oldContext)
        {
            List<CellRange> invalidatedCellRanges = new List<CellRange>();

            foreach (ViewportPaneType viewportPaneType in Enum.GetValues(typeof(ViewportPaneType)))
            {
                ViewportPane newSheetViewport = this.SheetViewport[viewportPaneType];
                if (newSheetViewport.IsEmpty)
                {
                    continue;
                }

                CellRange newVisibleRange = newSheetViewport.VisibleRange;

                if (oldContext != null)
                {
                    ViewportPane oldSheetViewport = oldContext.SheetViewport[viewportPaneType];
                    if (oldSheetViewport.IsEmpty)
                    {
                        continue;
                    }

                    CellRange oldVisibleRange = oldSheetViewport.VisibleRange;

                    if (this.AreVisibleRowsOrColumnsChanged(oldContext, oldVisibleRange, newVisibleRange))
                    {
                        if (this.ContainsRowsColumnsWithChangedSize(oldContext, newSheetViewport))
                        {
                            // This is used in case of filtering. In all other cases there is no old context
                            invalidatedCellRanges.Add(newVisibleRange);
                        }
                        else
                        {
                            invalidatedCellRanges.AddRange(CellRange.GetDifference(newVisibleRange, oldVisibleRange));
                        }
                    }
                    else if (oldVisibleRange.RowCount != newVisibleRange.RowCount || oldVisibleRange.ColumnCount != newVisibleRange.ColumnCount)
                    {
                        invalidatedCellRanges.Add(newVisibleRange);
                    }
                    else
                    {
                        invalidatedCellRanges.AddRange(CellRange.GetDifference(newVisibleRange, oldVisibleRange));
                    }
                }
                else
                {
                    invalidatedCellRanges.Add(newVisibleRange);
                }
            }

            this.invalidatedCellRanges = invalidatedCellRanges.Distinct().ToArray();
        }

        private bool ContainsRowsColumnsWithChangedSize(WorksheetRenderUpdateContext oldContext, ViewportPane newSheetViewport)
        {
            CellRange newVisibleRange = newSheetViewport.VisibleRange;

            for (int row = newVisibleRange.FromIndex.RowIndex; row <= newVisibleRange.ToIndex.RowIndex; row++)
            {
                RowLayoutBox newBox = this.GetVisibleRowLayoutBox(row);
                RowLayoutBox oldBox = oldContext.GetVisibleRowLayoutBox(row);

                if (oldBox != null && !newBox.BoundingRectangle.Equals(oldBox.BoundingRectangle))
                {
                    return true;
                }
            }

            for (int column = newVisibleRange.FromIndex.ColumnIndex; column <= newVisibleRange.ToIndex.ColumnIndex; column++)
            {
                ColumnLayoutBox newBox = this.GetVisibleColumnLayoutBox(column);
                ColumnLayoutBox oldBox = oldContext.GetVisibleColumnLayoutBox(column);

                if (oldBox != null && !newBox.BoundingRectangle.Equals(oldBox.BoundingRectangle))
                {
                    return true;
                }
            }

            return false;
        }

        private bool AreVisibleRowsOrColumnsChanged(WorksheetRenderUpdateContext oldContext, CellRange oldVisibleRange, CellRange newVisibleRange)
        {
            if (newVisibleRange.IntersectsWith(oldVisibleRange))
            {
                if (oldVisibleRange.RowCount == newVisibleRange.RowCount &&
                    oldVisibleRange.FromIndex.RowIndex != newVisibleRange.FromIndex.RowIndex)
                {
                    return true;
                }

                if (oldVisibleRange.ColumnCount == newVisibleRange.ColumnCount &&
                    oldVisibleRange.FromIndex.ColumnIndex != newVisibleRange.FromIndex.ColumnIndex)
                {
                    return true;
                }

                CellRange intersectionRange = newVisibleRange.Intersect(oldVisibleRange);

                bool visibleRowsHasChanged = this.AreVisibleRowsChanged(oldContext, intersectionRange) || !this.defaultColumnWidth.Equals(oldContext.defaultColumnWidth);
                bool visibleColumnsHasChanged = this.AreVisibleColumnsChanged(oldContext, intersectionRange) || !this.defaultRowHeight.Equals(oldContext.defaultRowHeight);

                return visibleRowsHasChanged || visibleColumnsHasChanged;
            }

            return false;
        }

        private bool AreVisibleRowsChanged(WorksheetRenderUpdateContext oldContext, CellRange intersectionRange)
        {
            List<int> newVisibleRows = this.GetSortedVisibleRowIndexesInRange(intersectionRange);
            List<int> oldVisibleRows = oldContext.GetSortedVisibleRowIndexesInRange(intersectionRange);

            if (newVisibleRows.Count != oldVisibleRows.Count)
            {
                return true;
            }

            for (int i = 0; i < newVisibleRows.Count; i++)
            {
                if (newVisibleRows[i] != oldVisibleRows[i])
                {
                    return true;
                }
            }

            return false;
        }

        private bool AreVisibleColumnsChanged(WorksheetRenderUpdateContext oldContext, CellRange intersectionRange)
        {
            List<int> newVisibleColumns = this.GetSortedVisibleColumnIndexesInRange(intersectionRange);
            List<int> oldVisibleColumns = oldContext.GetSortedVisibleColumnIndexesInRange(intersectionRange);

            if (newVisibleColumns.Count != oldVisibleColumns.Count)
            {
                return true;
            }

            for (int i = 0; i < newVisibleColumns.Count; i++)
            {
                if (newVisibleColumns[i] != oldVisibleColumns[i])
                {
                    return true;
                }
            }

            return false;
        }

        private List<int> GetSortedVisibleRowIndexesInRange(CellRange range)
        {
            List<int> visibleRows = new List<int>();

            foreach (RowLayoutBox rowBox in this.VisibleRowBoxes)
            {
                if (range.FromIndex.RowIndex <= rowBox.RowIndex && rowBox.RowIndex <= range.ToIndex.RowIndex)
                {
                    visibleRows.Add(rowBox.RowIndex);
                }
            }

            return visibleRows;
        }

        private List<int> GetSortedVisibleColumnIndexesInRange(CellRange range)
        {
            List<int> visibleColumns = new List<int>();

            foreach (ColumnLayoutBox columnBox in this.VisibleColumnBoxes)
            {
                if (range.FromIndex.ColumnIndex <= columnBox.ColumnIndex && columnBox.ColumnIndex <= range.ToIndex.ColumnIndex)
                {
                    visibleColumns.Add(columnBox.ColumnIndex);
                }
            }

            return visibleColumns;
        }

        #endregion


        #region Init Caches

        /// <summary>
        /// Initializes the caches.
        /// </summary>
        private void InitCaches(WorksheetRenderUpdateContext oldContext)
        {
            List<Action> actions = new List<Action>();

            actions.Add(new Action(() =>
                {
                    this.InitCellValues(oldContext);
                    this.InitPreviousNextNonEmptyNonMergeCells();
                }));

            //QUESTION: Why there are differences getting different property values.
            actions.Add(new Action(() => { this.InitCellFontProperties(oldContext); }));
            actions.Add(new Action(() =>
            {
                this.InitPropertyCache(CellPropertyDefinitions.FormatProperty, this.cellIndexToFormat,
                    (oldContext == null) ? null : oldContext.cellIndexToFormat);
            }));
            actions.Add(new Action(() =>
            {
                this.InitPropertyCache(CellPropertyDefinitions.IndentProperty, this.cellIndexToIndentPropertyValue,
                    (oldContext == null) ? null : oldContext.cellIndexToIndentPropertyValue);
            }));
            actions.Add(new Action(() =>
            {
                this.InitPropertyCache(CellPropertyDefinitions.IsWrappedProperty, this.cellindexToIsWrapped,
                    (oldContext == null) ? null : oldContext.cellindexToIsWrapped);
            }));
            actions.Add(new Action(() =>
            {
                this.InitPropertyCache(CellPropertyDefinitions.HorizontalAlignmentProperty, this.cellIndexToHorizontalAlignment,
                    (oldContext == null) ? null : oldContext.cellIndexToHorizontalAlignment);
            }));
            actions.Add(new Action(() =>
            {
                this.InitPropertyCache(CellPropertyDefinitions.VerticalAlignmentProperty, this.cellIndexToVerticalAlignment,
                    (oldContext == null) ? null : oldContext.cellIndexToVerticalAlignment);
            }));
            actions.Add(new Action(() =>
            {
                this.InitPropertyCache(CellPropertyDefinitions.FillProperty, this.cellIndexToFill,
                    (oldContext == null) ? null : oldContext.cellIndexToFill);
            }));
            actions.Add(new Action(() =>
            {
                this.InitPropertyCache(CellPropertyDefinitions.LeftBorderProperty, this.cellIndexToLeftBorder,
                    (oldContext == null) ? null : oldContext.cellIndexToLeftBorder);
            }));
            actions.Add(new Action(() =>
            {
                this.InitPropertyCache(CellPropertyDefinitions.TopBorderProperty, this.cellIndexToTopBorder,
                    (oldContext == null) ? null : oldContext.cellIndexToTopBorder);
            }));
            actions.Add(new Action(() =>
            {
                this.InitPropertyCache(CellPropertyDefinitions.RightBorderProperty, this.cellIndexToRightBorder,
                    (oldContext == null) ? null : oldContext.cellIndexToRightBorder);
            }));
            actions.Add(new Action(() =>
            {
                this.InitPropertyCache(CellPropertyDefinitions.BottomBorderProperty, this.cellIndexToBottomBorder,
                    (oldContext == null) ? null : oldContext.cellIndexToBottomBorder);
            }));
            actions.Add(new Action(() =>
            {
                this.InitPropertyCache(CellPropertyDefinitions.DiagonalUpBorderProperty, this.cellIndexToDiagonalUpBorder,
                    (oldContext == null) ? null : oldContext.cellIndexToDiagonalUpBorder);
            }));
            actions.Add(new Action(() =>
            {
                this.InitPropertyCache(CellPropertyDefinitions.DiagonalDownBorderProperty, this.cellIndexToDiagonalDownBorder,
                    (oldContext == null) ? null : oldContext.cellIndexToDiagonalDownBorder);
            }));
            if (this.Worksheet.ViewState.CircleInvalidData)
            {
                actions.Add(new Action(() =>
                {
                    this.InitPropertyCache(CellPropertyDefinitions.DataValidationRuleProperty, this.cellIndexToDataValidationRule,
                        (oldContext == null) ? null : oldContext.cellIndexToDataValidationRule);
                }));
            }
#if WPF
            Parallel.Invoke(actions.ToArray());
#else
            Telerik.Windows.Documents.Utilities.TasksHelper.DoAsync(actions);
#endif

            if (this.Worksheet.ViewState.CircleInvalidData)
            {
                // This is accessing cell values. If you want to run it on a separate thread from values initialization lock the values collection!
                this.InitDataValidationRuleResults(oldContext);
            }
            this.InitCellContentSizes(oldContext);
            this.InitCellActualBoundingRectangleByHorizontalAlignment(oldContext);
            this.InitCellClippings(oldContext);

            this.InitHyperlinks();
            this.InitShapes();
        }

        private bool IsContainedInInvalidatedCellRange(long index)
        {
            int rowIndex;
            int columnIndex;
            CellsPropertyBag.ConvertLongToRowAndColumnIndexes(index, out rowIndex, out columnIndex);

            for (int i = 0; i < this.invalidatedCellRanges.Length; i++)
            {
                if (this.invalidatedCellRanges[i].Contains(rowIndex, columnIndex))
                {
                    return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Initializes the cell values.
        /// </summary>
        private void InitCellValues(WorksheetRenderUpdateContext oldContext)
        {
            foreach (KeyValuePair<ViewportPaneType, IEnumerable<CellLayoutBox>> pair in this.VisibleCellLayoutBoxes)
            {
                foreach (CellLayoutBox cellBox in pair.Value)
                {
                    if (cellBox.ColumnIndex == 1 && cellBox.RowIndex == 1)
                    {
                    }

                    long index = cellBox.LongIndex;

                    if (!this.cellIndexToValue.ContainsKey(index))
                    {
                        if (!this.IsContainedInInvalidatedCellRange(index) &&
                            oldContext != null && oldContext.cellIndexToValue.ContainsKey(index))
                        {
                            this.cellIndexToValue.Add(index, oldContext.cellIndexToValue[index]);
                        }
                        else
                        {
                            this.cellIndexToValue.Add(index,
                                this.Worksheet.Cells.PropertyBag.GetPropertyValueRespectingStyle(CellPropertyDefinitions.ValueProperty, this.Worksheet, index));
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Gets the previous and next non empty and non merged cells.
        /// </summary>
        /// <param name="cellBox">The cell box.</param>
        /// <param name="updateContext">The update context.</param>
        /// <param name="prevCellBox">The previous cell box.</param>
        /// <param name="nextCellBox">The next cell box.</param>
        private static void GetPreviousNextNonEmptyNonMergedCells(CellLayoutBox cellBox, WorksheetRenderUpdateContext updateContext, out CellLayoutBox prevCellBox, out CellLayoutBox nextCellBox)
        {
            if (cellBox.MergeState == CellMergeState.NonTopLeftCellInMergedRange)
            {
                CellRange mergedRange;
                updateContext.Worksheet.Cells.TryGetContainingMergedRange(cellBox.RowIndex, cellBox.ColumnIndex, out mergedRange);
                cellBox = updateContext.GetVisibleCellBox(mergedRange.FromIndex.RowIndex, mergedRange.FromIndex.ColumnIndex);
            }

            if (cellBox.MergeState == CellMergeState.TopLeftCellInMergedRange)
            {
                prevCellBox = cellBox;
                nextCellBox = cellBox;
                return;
            }

            int leftMostColumnIndex = SpreadsheetDefaultValues.ColumnCount - 1;
            int rightMostColumnIndex = 0;
            foreach (ViewportPane pane in updateContext.SheetViewport.ViewportPanes)
            {
                leftMostColumnIndex = Math.Min(leftMostColumnIndex, pane.VisibleRange.FromIndex.ColumnIndex);
                rightMostColumnIndex = Math.Max(rightMostColumnIndex, pane.VisibleRange.ToIndex.ColumnIndex);
            }

            int columnIndexOfNextNonEmptyColumn = cellBox.ColumnIndex + 1;
            while (columnIndexOfNextNonEmptyColumn >= leftMostColumnIndex && columnIndexOfNextNonEmptyColumn <= rightMostColumnIndex)
            {
                nextCellBox = updateContext.GetVisibleCellBox(cellBox.RowIndex, columnIndexOfNextNonEmptyColumn);
                if (updateContext.GetCellValue(nextCellBox.LongIndex) == EmptyCellValue.EmptyValue &&
                    nextCellBox.MergeState == CellMergeState.NotMerged)
                {
                    columnIndexOfNextNonEmptyColumn++;
                }
                else
                {
                    break;
                }
            }

            columnIndexOfNextNonEmptyColumn--;
            nextCellBox = updateContext.GetVisibleCellBox(cellBox.RowIndex, columnIndexOfNextNonEmptyColumn);

            int columnIndexOfPrevNonEmptyColumn = cellBox.ColumnIndex - 1;
            while (columnIndexOfPrevNonEmptyColumn >= leftMostColumnIndex && columnIndexOfPrevNonEmptyColumn <= rightMostColumnIndex)
            {
                prevCellBox = updateContext.GetVisibleCellBox(cellBox.RowIndex, columnIndexOfPrevNonEmptyColumn);
                if (updateContext.GetCellValue(prevCellBox.LongIndex) == EmptyCellValue.EmptyValue &&
                    prevCellBox.MergeState == CellMergeState.NotMerged)
                {
                    columnIndexOfPrevNonEmptyColumn--;
                }
                else
                {
                    break;
                }
            }

            columnIndexOfPrevNonEmptyColumn++;
            prevCellBox = updateContext.GetVisibleCellBox(cellBox.RowIndex, columnIndexOfPrevNonEmptyColumn);
        }

        /// <summary>
        /// Initializes the previous and next non empty and non merged cells.
        /// </summary>
        private void InitPreviousNextNonEmptyNonMergeCells()
        {
            foreach (KeyValuePair<ViewportPaneType, IEnumerable<CellLayoutBox>> pair in this.VisibleCellLayoutBoxes)
            {
                foreach (CellLayoutBox cellBox in pair.Value)
                {
                    long index = cellBox.LongIndex;

                    if (this.cellIndexToPreviousNextNonEmptyNonMergedCells.ContainsKey(index))
                    {
                        continue;
                    }

                    CellLayoutBox previousBox;
                    CellLayoutBox nextBox;
                    GetPreviousNextNonEmptyNonMergedCells(cellBox, this, out previousBox, out nextBox);
                    this.cellIndexToPreviousNextNonEmptyNonMergedCells.Add(index, new Tuple<CellLayoutBox, CellLayoutBox>(previousBox, nextBox));
                }
            }
        }

        /// <summary>
        /// Initializes the hyperlinks.
        /// </summary>
        private void InitHyperlinks()
        {
            foreach (KeyValuePair<ViewportPaneType, IEnumerable<CellLayoutBox>> pair in this.VisibleCellLayoutBoxes)
            {
                foreach (CellLayoutBox cellBox in pair.Value)
                {
                    SpreadsheetHyperlink hyperlink;
                    HyperlinkInfo hyperlinkInfo;
                    long index = cellBox.LongIndex;

                    ICellValue cellValue = this.GetCellValue(index);

                    if (this.Worksheet.Hyperlinks.TryGetHyperlink(cellBox.RowIndex, cellBox.ColumnIndex, out hyperlink))
                    {
                        this.CellIndexToHyperlinkInfo.Add(index, hyperlink.HyperlinkInfo);
                        hyperlinkInfo = hyperlink.HyperlinkInfo;
                    }
                    else if (SpreadsheetHelper.TryGetHyperLinkInfo(cellValue, out hyperlinkInfo))
                    {
                        this.CellIndexToHyperlinkInfo.Add(index, hyperlinkInfo);
                    }

                    if (hyperlinkInfo != null)
                    {
                        CellValueFormat format = this.GetCellFormat(index);
                        string stringValue = cellValue.GetValueAsString(format);
                        Rect rect;
                        if (string.IsNullOrEmpty(stringValue))
                        {
                            rect = cellBox.BoundingRectangle;
                        }
                        else
                        {
                            rect = this.GetActualBoundingRectangleByContentAlignment(cellBox);
                        }

                        this.CellIndexToHyperlinkArea.Add(index, rect);
                    }
                }
            }
        }

        private void InitShapes()
        {
            HashSet<FloatingShapeBase> shapesToAdd = new HashSet<FloatingShapeBase>();

            foreach (FloatingShapeBase shape in this.Worksheet.Shapes)
            {
                ShapeLayoutBox shapeLayoutBox = this.Layout.GetShapeLayoutBox(shape.Id);

                if (this.SheetViewport.ViewportPanes.Any(p => p.Rect.IntersectsWith(shapeLayoutBox.BoundingRectangle)) &&
                    !shapesToAdd.Contains(shape))
                {
                    shapesToAdd.Add(shape);
                }
            }

            this.shapes.AddRange(shapesToAdd);
        }

        /// <summary>
        /// Initializes the cell font properties.
        /// </summary>
        private void InitCellFontProperties(WorksheetRenderUpdateContext oldContext)
        {
            Dictionary<CellRange, CellRangeFontProperties> cellRangeToCellRangeFontProperties = new Dictionary<CellRange, CellRangeFontProperties>();

            foreach (ViewportPane pane in this.SheetViewport.ViewportPanes)
            {
                CellRange cellRange = pane.VisibleRange;
                cellRangeToCellRangeFontProperties.Add(cellRange, this.Worksheet.Cells.GetFontProperties(cellRange));
            }

            foreach (KeyValuePair<ViewportPaneType, IEnumerable<CellLayoutBox>> pair in this.VisibleCellLayoutBoxes)
            {
                foreach (CellLayoutBox cellBox in pair.Value)
                {
                    long index = cellBox.LongIndex;

                    if (!this.SheetViewport.Contains(cellBox.RowIndex, cellBox.ColumnIndex) ||
                        this.cellIndexToFontProperties.ContainsKey(index))
                    {
                        continue;
                    }

                    if (!this.IsContainedInInvalidatedCellRange(index) &&
                            oldContext != null && oldContext.cellIndexToFontProperties.ContainsKey(index))
                    {
                        this.cellIndexToFontProperties.Add(index, oldContext.cellIndexToFontProperties[index]);
                    }
                    else
                    {
                        CellRange visibleRange = this.SheetViewport.GetCellRangeContainingCellIndex(cellBox.RowIndex, cellBox.ColumnIndex);
                        CellRangeFontProperties visibleCellRangeFontProperties = cellRangeToCellRangeFontProperties[visibleRange];

                        this.cellIndexToFontProperties.Add(index, visibleCellRangeFontProperties.GetFontProperties(index, this.CurrentTheme));
                    }
                }
            }
        }

        /// <summary>
        /// Initializes the property cache.
        /// </summary>
        /// <typeparam name="T">The type of the T.</typeparam>
        /// <param name="property">The property.</param>
        /// <param name="cellIndexToPropertyValue">The cell index to property value.</param>
        /// <param name="oldCellIndexToPropertyValue">The old cell index to property value.</param>
        private void InitPropertyCache<T>(IPropertyDefinition<T> property, Dictionary<long, T> cellIndexToPropertyValue, Dictionary<long, T> oldCellIndexToPropertyValue)
        {
            Dictionary<CellRange, ICompressedList<T>> cellRangeToPropertyCollection = new Dictionary<CellRange, ICompressedList<T>>();

            foreach (ViewportPane pane in this.SheetViewport.ViewportPanes)
            {
                CellRange cellRange = pane.VisibleRange;
                cellRangeToPropertyCollection.Add(cellRange, this.Worksheet.Cells.GetPropertyValueRespectingStyle(property, cellRange));
            }

            foreach (KeyValuePair<ViewportPaneType, IEnumerable<CellLayoutBox>> pair in this.VisibleCellLayoutBoxes)
            {
                foreach (CellLayoutBox cellBox in pair.Value)
                {
                    long index = cellBox.LongIndex;

                    if (!this.SheetViewport.Contains(cellBox.LongIndex) ||
                        cellIndexToPropertyValue.ContainsKey(index))
                    {
                        continue;
                    }

                    if (!this.IsContainedInInvalidatedCellRange(index) &&
                            oldCellIndexToPropertyValue != null && oldCellIndexToPropertyValue.ContainsKey(index))
                    {
                        cellIndexToPropertyValue.Add(index, oldCellIndexToPropertyValue[index]);
                    }
                    else
                    {
                        CellRange visibleRange = this.SheetViewport.GetCellRangeContainingCellIndex(cellBox.RowIndex, cellBox.ColumnIndex);
                        ICompressedList<T> cellValuesCollection = cellRangeToPropertyCollection[visibleRange];

                        cellIndexToPropertyValue.Add(index, cellValuesCollection.GetValue(index));
                    }
                }
            }
        }

        /// <summary>
        /// Initializes the cell content sizes.
        /// </summary>
        private void InitCellContentSizes(WorksheetRenderUpdateContext oldContext)
        {
            foreach (KeyValuePair<ViewportPaneType, IEnumerable<CellLayoutBox>> pair in this.VisibleCellLayoutBoxes)
            {
                foreach (CellLayoutBox cellBox in pair.Value)
                {
                    long index = cellBox.LongIndex;

                    if (this.cellIndexToContentSize.ContainsKey(index))
                    {
                        continue;
                    }

                    if (!this.IsContainedInInvalidatedCellRange(index) &&
                            oldContext != null && oldContext.cellIndexToContentSize.ContainsKey(index))
                    {
                        this.cellIndexToContentSize.Add(index, oldContext.cellIndexToContentSize[index]);
                    }
                    else
                    {
                        Size cellContentSize = this.Layout.GetCellContentSize(index, cellBox,
                        this.GetIsWrapped(index),
                        this.GetFontProperties(index));

                        this.cellIndexToContentSize.Add(index, cellContentSize);
                    }
                }
            }
        }

        /// <summary>
        /// Initializes the cell clippings.
        /// </summary>
        private void InitCellClippings(WorksheetRenderUpdateContext oldContext)
        {
            foreach (KeyValuePair<ViewportPaneType, IEnumerable<CellLayoutBox>> pair in this.VisibleCellLayoutBoxes)
            {
                foreach (CellLayoutBox cellBox in pair.Value)
                {
                    long index = cellBox.LongIndex;

                    if (this.cellIndexToClipping.ContainsKey(index))
                    {
                        continue;
                    }

                    if (!this.IsContainedInInvalidatedCellRange(index) &&
                            oldContext != null && oldContext.cellIndexToClipping.ContainsKey(index))
                    {
                        this.cellIndexToClipping.Add(index, oldContext.cellIndexToClipping[index]);
                    }
                    else
                    {
                        Rect clipping = CellBoxRenderHelper.CalculateCellClipping(cellBox, this);
                        this.cellIndexToClipping.Add(index, clipping);
                    }
                }
            }
        }

        /// <summary>
        /// Initializes the cell actual bounding rectangle by horizontal alignment.
        /// </summary>
        private void InitCellActualBoundingRectangleByHorizontalAlignment(WorksheetRenderUpdateContext oldContext)
        {
            Dictionary<ViewportPaneType, HashSet<CellLayoutBox>> cellBoxesToAdd = new Dictionary<ViewportPaneType, HashSet<CellLayoutBox>>();
            foreach (ViewportPaneType viewportPaneType in Enum.GetValues(typeof(ViewportPaneType)))
            {
                cellBoxesToAdd.Add(viewportPaneType, new HashSet<CellLayoutBox>());
            }

            foreach (KeyValuePair<ViewportPaneType, IEnumerable<CellLayoutBox>> pair in this.VisibleCellLayoutBoxes)
            {
                foreach (CellLayoutBox cellBox in pair.Value)
                {
                    long index = cellBox.LongIndex;

                    if (this.cellIndexToActualBoundingRectangle.ContainsKey(index))
                    {
                        continue;
                    }

                    if (!this.IsContainedInInvalidatedCellRange(index) &&
                            oldContext != null && oldContext.cellIndexToActualBoundingRectangle.ContainsKey(index))
                    {
                        this.cellIndexToActualBoundingRectangle.Add(index, oldContext.cellIndexToActualBoundingRectangle[index]);
                    }
                    else
                    {
                        Rect boundingRectangle = CellBoxRenderHelper.CalculateActualBoundingRectangleByContentAlignment(cellBox, this);

                        this.AddCellBoxToTheOtherViewportIfNeeded(cellBox, boundingRectangle, cellBoxesToAdd);

                        this.cellIndexToActualBoundingRectangle.Add(index, boundingRectangle);
                    }
                }
            }

            foreach (KeyValuePair<ViewportPaneType, HashSet<CellLayoutBox>> pair in cellBoxesToAdd)
            {
                HashSet<CellLayoutBox> cellBoxes = this.visibleCellBoxes[pair.Key] as HashSet<CellLayoutBox>;
                foreach (CellLayoutBox cellBox in pair.Value)
                {
                    cellBoxes.Add(cellBox);
                }
            }
        }

        private void InitDataValidationRuleResults(WorksheetRenderUpdateContext oldContext)
        {
            foreach (KeyValuePair<ViewportPaneType, IEnumerable<CellLayoutBox>> pair in this.VisibleCellLayoutBoxes)
            {
                foreach (CellLayoutBox cellBox in pair.Value)
                {
                    long index = cellBox.LongIndex;

                    if (this.cellIndexToDataValidationResult.ContainsKey(index))
                    {
                        continue;
                    }

                    if (!this.IsContainedInInvalidatedCellRange(index) &&
                            oldContext != null && oldContext.cellIndexToDataValidationResult.ContainsKey(index))
                    {
                        this.cellIndexToDataValidationResult.Add(index, oldContext.cellIndexToDataValidationResult[index]);
                    }
                    else
                    {
                        IDataValidationRule rule = this.GetDataValidationRule(index);
                        int row;
                        int column;
                        CellsPropertyBag.ConvertLongToRowAndColumnIndexes(index, out row, out column);
                        this.cellIndexToDataValidationResult.Add(index, rule.Evaluate(this.Worksheet, row, column));
                    }
                }
            }
        }

        private void AddCellBoxToTheOtherViewportIfNeeded(CellLayoutBox cellBox, Rect boundingRectangle, Dictionary<ViewportPaneType, HashSet<CellLayoutBox>> cellBoxesToAdd)
        {
            Rect rect = new Rect(boundingRectangle.X, boundingRectangle.Y, Math.Max(boundingRectangle.Width - 1, 0), Math.Max(boundingRectangle.Height - 1, 0));
            Point topLeft = new Point(rect.Left, rect.Top);
            Point topRight = new Point(rect.Right, rect.Top);
            Point bottomRight = new Point(rect.Right, rect.Bottom);
            Point bottomLeft = new Point(rect.Left, rect.Bottom);

            ViewportPaneType topLeftPane = this.SheetViewport.GetViewportPaneFromDocumentPoint(topLeft).ViewportPaneType;
            ViewportPaneType topRightPane = this.SheetViewport.GetViewportPaneFromDocumentPoint(topRight).ViewportPaneType;
            ViewportPaneType bottomRightPane = this.SheetViewport.GetViewportPaneFromDocumentPoint(bottomRight).ViewportPaneType;
            ViewportPaneType bottomLeftPane = this.SheetViewport.GetViewportPaneFromDocumentPoint(bottomLeft).ViewportPaneType;

            if (topLeftPane != topRightPane &&
                this.SheetViewport[topRightPane].Rect.IntersectsWith(rect) &&
                !cellBoxesToAdd[topRightPane].Contains(cellBox))
            {
                cellBoxesToAdd[topRightPane].Add(cellBox);
            }

            if (topLeftPane != bottomRightPane &&
                this.SheetViewport[bottomRightPane].Rect.IntersectsWith(rect) &&
                !cellBoxesToAdd[bottomRightPane].Contains(cellBox))
            {
                cellBoxesToAdd[bottomRightPane].Add(cellBox);
            }

            if (topLeftPane != bottomLeftPane &&
                this.SheetViewport[bottomLeftPane].Rect.IntersectsWith(rect) &&
                !cellBoxesToAdd[bottomLeftPane].Contains(cellBox))
            {
                cellBoxesToAdd[bottomLeftPane].Add(cellBox);
            }
        }

        #endregion


        #region Methods

        private void AddCellsWithLargeContentToVisibleCellBoxes()
        {
            foreach (ViewportPane pane in this.SheetViewport.ViewportPanes)
            {
                int minRow = pane.VisibleRange.FromIndex.RowIndex;
                int maxRow = Math.Min(pane.VisibleRange.ToIndex.RowIndex, this.Worksheet.UsedCellRange.ToIndex.RowIndex);

                for (int row = minRow; row <= maxRow; row++)
                {
                    if (this.Layout.GetRowHeight(row) == HiddenCellSize.Height)
                    {
                        continue;
                    }

                    int currentColumnIndex = pane.VisibleRange.FromIndex.ColumnIndex - 1;
                    int counter = 0;

                    while (currentColumnIndex >= 0 &&
                        !this.SheetViewport.ContainsColumnIndex(currentColumnIndex) &&
                        this.GetCellValue(CellsPropertyBag.ConvertCellIndexToLong(row, currentColumnIndex)) is EmptyCellValue)
                    {
                        if (counter == MaxColumnsToCheckForLongText)
                        {
                            break;
                        }

                        counter++;
                        currentColumnIndex--;
                    }

                    if (currentColumnIndex < 0 ||
                        this.SheetViewport.ContainsColumnIndex(currentColumnIndex) ||
                        this.Layout.GetColumnWidth(currentColumnIndex) == HiddenCellSize.Width)
                    {
                        continue;
                    }

                    this.AddCellBoxToVisibleBoxesIfPartOfItIntersectsWithTheViewport(row, currentColumnIndex, pane);
                }
            }
        }

        private void AddCellBoxToVisibleBoxesIfPartOfItIntersectsWithTheViewport(int row, int column, ViewportPane pane)
        {
            CellLayoutBox box = this.GetVisibleCellBox(row, column);
            if (box.Width == HiddenCellSize.Width || box.Height == HiddenCellSize.Height)
            {
                return;
            }

            Rect boundingRect = CellBoxRenderHelper.CalculateActualBoundingRectangleByContentAlignment(box, this);

            if (pane.Rect.IntersectsWith(boundingRect))
            {
                HashSet<CellLayoutBox> cellBoxes = this.visibleCellBoxes[pane.ViewportPaneType] as HashSet<CellLayoutBox>;
                cellBoxes.Add(box);
            }
        }

        /// <summary>
        /// Gets the visible row layout box.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <returns>The visible row layout box.</returns>
        public RowLayoutBox GetVisibleRowLayoutBox(int rowIndex)
        {
            RowLayoutBox result = null;

            if (this.SheetViewport.ContainsRowIndex(rowIndex))
            {
                foreach (RowLayoutBox rowBox in this.VisibleRowBoxes)
                {
                    if (rowBox.RowIndex == rowIndex)
                    {
                        result = rowBox;
                        break;
                    }
                }

                result = result ?? this.Layout.GetRowLayoutBox(rowIndex);
            }

            return result;
        }

        /// <summary>
        /// Gets the visible column layout box.
        /// </summary>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The visible column layout box.</returns>
        public ColumnLayoutBox GetVisibleColumnLayoutBox(int columnIndex)
        {
            ColumnLayoutBox result = null;

            if (this.SheetViewport.ContainsColumnIndex(columnIndex))
            {
                foreach (ColumnLayoutBox columnBox in this.VisibleColumnBoxes)
                {
                    if (columnBox.ColumnIndex == columnIndex)
                    {
                        result = columnBox;
                        break;
                    }
                }

                result = result ?? this.Layout.GetColumnLayoutBox(columnIndex);
            }

            return result;
        }

        /// <summary>
        /// Gets the visible cell box.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <returns>The visible cell box.</returns>
        public CellLayoutBox GetVisibleCellBox(long index)
        {
            int rowIndex;
            int columnIndex;

            CellsPropertyBag.ConvertLongToRowAndColumnIndexes(index, out rowIndex, out columnIndex);
            return this.GetVisibleCellBox(rowIndex, columnIndex);
        }

        /// <summary>
        /// Gets the visible cell box.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The visible cell box.</returns>
        public CellLayoutBox GetVisibleCellBox(int rowIndex, int columnIndex)
        {
            if (this.cellIndexToCellLayoutBox == null)
            {
                this.cellIndexToCellLayoutBox = new Dictionary<ViewportPaneType, Dictionary<long, CellLayoutBox>>();

                foreach (KeyValuePair<ViewportPaneType, IEnumerable<CellLayoutBox>> pair in this.VisibleCellLayoutBoxes)
                {
                    Dictionary<long, CellLayoutBox> cellBoxes = new Dictionary<long, CellLayoutBox>();

                    foreach (CellLayoutBox cellBox in pair.Value)
                    {
                        cellBoxes.Add(cellBox.LongIndex, cellBox);
                    }

                    this.cellIndexToCellLayoutBox.Add(pair.Key, cellBoxes);
                }
            }

            ViewportPaneType cellIndexViewportPaneType = this.SheetViewport.GetViewportPaneContainingCellIndex(rowIndex, columnIndex).ViewportPaneType;

            long desiredIndex = CellsPropertyBag.ConvertCellIndexToLong(rowIndex, columnIndex);

            CellLayoutBox result = null;
            if (!this.cellIndexToCellLayoutBox[cellIndexViewportPaneType].TryGetValue(desiredIndex, out result))
            {
                result = this.Layout.GetCellLayoutBox(rowIndex, columnIndex);
                this.cellIndexToCellLayoutBox[cellIndexViewportPaneType].Add(desiredIndex, result);
            }

            return result;
        }

        /// <summary>
        /// Gets the actual bounding rectangle by content alignment.
        /// </summary>
        /// <param name="cellBox">The cell box.</param>
        /// <returns>The actual bounding rectangle.</returns>
        public Rect GetActualBoundingRectangleByContentAlignment(CellLayoutBox cellBox)
        {
            long index = cellBox.LongIndex;

            Rect boundingRectangle;
            if (!this.cellIndexToActualBoundingRectangle.TryGetValue(index, out boundingRectangle))
            {
                boundingRectangle = CellBoxRenderHelper.CalculateActualBoundingRectangleByContentAlignment(cellBox, this);
                this.cellIndexToActualBoundingRectangle.Add(index, boundingRectangle);
            }

            return boundingRectangle;
        }

        public HyperlinkInfo GetHyperlinkInfo(long index)
        {
            HyperlinkInfo result;
            if (!this.CellIndexToHyperlinkInfo.TryGetValue(index, out result))
            {
                return null;
            }

            return result;
        }

        public Rect GetHyperlinkArea(long index)
        {
            Rect result;
            if (!this.CellIndexToHyperlinkArea.TryGetValue(index, out result))
            {
                return Rect.Empty;
            }

            return result;
        }

        public void AddHyperlinkArea(long index, Rect rect)
        {
            this.CellIndexToHyperlinkArea[index] = rect;
        }

        /// <summary>
        /// Gets the cell clipping.
        /// </summary>
        /// <param name="cellBox">The cell box.</param>
        /// <returns>The cell clipping rectangle.</returns>
        public Rect GetCellClipping(CellLayoutBox cellBox)
        {
            long index = cellBox.LongIndex;

            Rect clipping;
            if (!this.cellIndexToClipping.TryGetValue(index, out clipping))
            {
                clipping = CellBoxRenderHelper.CalculateCellClipping(cellBox, this);
                this.cellIndexToClipping.Add(index, clipping);
            }

            return clipping;
        }

        /// <summary>
        /// Gets the size of the cell content.
        /// </summary>
        /// <param name="cellBox">The cell box.</param>
        /// <returns>The size of the cell content.</returns>
        public Size GetCellContentSize(CellLayoutBox cellBox)
        {
            Size cellContentSize;
            long index = cellBox.LongIndex;
            if (!this.cellIndexToContentSize.TryGetValue(index, out cellContentSize))
            {
                cellContentSize = this.Layout.GetCellContentSize(index, cellBox, this.GetIsWrapped(index), this.GetFontProperties(index));
            }

            return cellContentSize;
        }

        //internal Size GetCellContentSize(CellLayoutBox cellBox, bool useDefaultTextMeasuringMethod)
        //{
        //Func<string, FontProperties, double?, Size> measuringMethod = useDefaultTextMeasuringMethod ? LayoutHelper.DefaultTextMeasuringMethod : this.measureMultilineText;
        //bool shouldForceTextMeasuring = measuringMethod != this.measureMultilineText;

        //long index = cellBox.LongIndex;

        //Size cellContentSize;
        //if (shouldForceTextMeasuring || !this.cellIndexToContentSize.TryGetValue(index, out cellContentSize))
        //{
        //    cellContentSize = LayoutHelper.CalculateCellContentSize(this.Worksheet, cellBox, this.GetIsWrapped(index), this.GetFontProperties(index), measuringMethod);

        //    if (!shouldForceTextMeasuring)
        //    {
        //        this.cellIndexToContentSize.Add(index, cellContentSize);
        //    }
        //}

        //return cellContentSize;
        //}

        public int CalculateIndent(long index)
        {
            return LayoutHelper.CalculateIndent(this.GetHorizontalAlignment(index), this.GetIndentPropertyValue(index));
        }

        private T GetPropertyValue<T>(IPropertyDefinition<T> propertyDefinition, Dictionary<long, T> cache, long index)
        {
            T result;
            if (!cache.TryGetValue(index, out result))
            {
                result = this.Worksheet.Cells.PropertyBag.GetPropertyValueRespectingStyle(propertyDefinition, this.Worksheet, index);
                cache.Add(index, result);
            }

            return result;
        }

        public RadHorizontalAlignment GetHorizontalAlignment(long index)
        {
            return this.GetPropertyValue(CellPropertyDefinitions.HorizontalAlignmentProperty, this.cellIndexToHorizontalAlignment, index);
        }

        public RadVerticalAlignment GetVerticalAlignment(long index)
        {
            return this.GetPropertyValue(CellPropertyDefinitions.VerticalAlignmentProperty, this.cellIndexToVerticalAlignment, index);
        }

        public IFill GetFill(long index)
        {
            return this.GetPropertyValue(CellPropertyDefinitions.FillProperty, this.cellIndexToFill, index);
        }

        public CellBorder GetLeftBorder(long index)
        {
            return this.GetPropertyValue(CellPropertyDefinitions.LeftBorderProperty, this.cellIndexToLeftBorder, index);
        }

        public CellBorder GetTopBorder(long index)
        {
            return this.GetPropertyValue(CellPropertyDefinitions.TopBorderProperty, this.cellIndexToTopBorder, index);
        }

        public CellBorder GetRightBorder(long index)
        {
            return this.GetPropertyValue(CellPropertyDefinitions.RightBorderProperty, this.cellIndexToRightBorder, index);
        }

        public CellBorder GetBottomBorder(long index)
        {
            return this.GetPropertyValue(CellPropertyDefinitions.BottomBorderProperty, this.cellIndexToBottomBorder, index);
        }

        public CellBorder GetDiagonalUpBorder(long index)
        {
            return this.GetPropertyValue(CellPropertyDefinitions.DiagonalUpBorderProperty, this.cellIndexToDiagonalUpBorder, index);
        }

        public CellBorder GetDiagonalDownBorder(long index)
        {
            return this.GetPropertyValue(CellPropertyDefinitions.DiagonalDownBorderProperty, this.cellIndexToDiagonalDownBorder, index);
        }

        public ICellValue GetCellValue(long index)
        {
            return this.GetPropertyValue(CellPropertyDefinitions.ValueProperty, this.cellIndexToValue, index);
        }

        public CellValueFormat GetCellFormat(long index)
        {
            return this.GetPropertyValue(CellPropertyDefinitions.FormatProperty, this.cellIndexToFormat, index);
        }

        public FontProperties GetFontProperties(long index)
        {
            FontProperties fontProperties;
            if (!this.cellIndexToFontProperties.TryGetValue(index, out fontProperties))
            {
                fontProperties = this.Worksheet.Cells.GetFontProperties(index);
                this.cellIndexToFontProperties.Add(index, fontProperties);
            }

            return fontProperties;
        }

        public bool GetIsWrapped(long index)
        {
            return this.GetPropertyValue(CellPropertyDefinitions.IsWrappedProperty, this.cellindexToIsWrapped, index);
        }

        /// <summary>
        /// Gets the indent property value.
        /// </summary>
        /// <param name="index">Index of the cell.</param>
        /// <returns>The indent.</returns>
        public int GetIndentPropertyValue(long index)
        {
            return this.GetPropertyValue(CellPropertyDefinitions.IndentProperty, this.cellIndexToIndentPropertyValue, index);
        }

        /// <summary>
        /// Gets the previous and the next non empty and non merged cells.
        /// </summary>
        /// <param name="index">Index of the cell.</param>
        /// <param name="previousCellBox">The previous cell box.</param>
        /// <param name="nextCellBox">The next cell box.</param>
        public void GetPreviousNextNonEmptyNonMergedCells(long index,
            out CellLayoutBox previousCellBox, out CellLayoutBox nextCellBox)
        {
            CellLayoutBox cellBox = this.GetVisibleCellBox(index);
            if (cellBox.MergeState == CellMergeState.NonTopLeftCellInMergedRange)
            {
                CellRange mergedRange;
                this.Worksheet.Cells.TryGetContainingMergedRange(cellBox.RowIndex, cellBox.ColumnIndex, out mergedRange);
                cellBox = this.GetVisibleCellBox(mergedRange.FromIndex.RowIndex, mergedRange.FromIndex.ColumnIndex);
            }

            //this.cellIndexToPreviousNextNonEmptyNonMergedCells
            Tuple<CellLayoutBox, CellLayoutBox> tuple = null;
            if (!this.cellIndexToPreviousNextNonEmptyNonMergedCells.TryGetValue(index, out tuple))
            {
                GetPreviousNextNonEmptyNonMergedCells(this.GetVisibleCellBox(index), this, out previousCellBox, out nextCellBox);
                tuple = new Tuple<CellLayoutBox, CellLayoutBox>(previousCellBox, nextCellBox);
                this.cellIndexToPreviousNextNonEmptyNonMergedCells.Add(index, tuple);
            }

            previousCellBox = tuple.Item1;
            nextCellBox = tuple.Item2;
        }

        /// <summary>
        /// Gets the shape which contains a given point of the layout if there is such.
        /// </summary>
        /// <param name="point">The point.</param>
        /// <returns>Returns a shape if such is found, otherwise returns null.</returns>
        public FloatingShapeBase GetShapeFromPoint(Point point)
        {
            foreach (FloatingShapeBase shape in this.Worksheet.Shapes.ReverseEnumerate())
            {
                Point position = this.Layout.GetShapeTopLeft(shape);
                double x = position.X;
                double y = position.Y;

                double width = shape.DoesRotationAngleRequireCellIndexChange() ? shape.Height : shape.Width;
                double height = shape.DoesRotationAngleRequireCellIndexChange() ? shape.Width : shape.Height;
                double rotationAngle = shape.DoesRotationAngleRequireCellIndexChange() ? -(shape.RotationAngle - 90) : -shape.RotationAngle;

                Matrix m = Matrix.Identity;
                m = m.RotateMatrixAt(rotationAngle, x + width / 2, y + height / 2);

                Point rotatedPoint = m.Transform(point);

                Rect shapeRect = new Rect(x, y, width, height);
                if (shapeRect.Contains(rotatedPoint))
                {
                    return shape;
                }
            }

            return null;
        }

        public IDataValidationRule GetDataValidationRule(long index)
        {
            return this.GetPropertyValue(CellPropertyDefinitions.DataValidationRuleProperty, this.cellIndexToDataValidationRule, index);
        }

        public bool GetDataValidationRuleResult(long index)
        {
            bool result;
            if (!this.cellIndexToDataValidationResult.TryGetValue(index, out result))
            {
                int row;
                int column;
                CellsPropertyBag.ConvertLongToRowAndColumnIndexes(index, out row, out column);
                result = this.GetDataValidationRule(index).Evaluate(this.Worksheet, row, column);
                this.cellIndexToDataValidationResult.Add(index, result);
            }

            return result;
        }

        #endregion
    }
}
