﻿using System;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Measurement;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Layout
{
    /// <summary>
    /// Represents a layout hellper class.
    /// </summary>
    public static class LayoutHelper
    {
        #region Methods

        internal static Size DefaultTextMeasuringMethod(string text, FontProperties fontProperties, double? wrappingWidth)
        {
            return RadTextMeasurer.MeasureMultiline(text, fontProperties, wrappingWidth);
        }

        internal static double? GetWrappingWidth(double cellBoxWidth, double cellIndent, bool isWrapped)
        {
            double? wrappingWidth = null;
            if (isWrapped)
            {
                wrappingWidth = Math.Max(0, cellBoxWidth - cellIndent);
            }

            return wrappingWidth;
        }

        internal static bool ShouldRespectIndent(RadHorizontalAlignment horizontalAlignment)
        {
            return horizontalAlignment == RadHorizontalAlignment.Left ||
                horizontalAlignment == RadHorizontalAlignment.Right ||
                horizontalAlignment == RadHorizontalAlignment.Distributed;
        }

        internal static int CalculateIndent(RadHorizontalAlignment horizontalAlignment, int indent)
        {
            if (LayoutHelper.ShouldRespectIndent(horizontalAlignment))
            {
                return indent;
            }

            return 0;
        }

        /// <summary>
        /// Calculates the size of the cell content.
        /// </summary>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="cellIndex">Index of the cell.</param>
        /// <returns>The size of the cell content.</returns>
        public static Size CalculateCellContentSize(Worksheet worksheet, CellIndex cellIndex)
        {
            return CalculateCellContentSize(worksheet, cellIndex.RowIndex, cellIndex.ColumnIndex);
        }

        /// <summary>
        /// Calculates the size of the cell content.
        /// </summary>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The size of the cell content.</returns>
        public static Size CalculateCellContentSize(Worksheet worksheet, int rowIndex, int columnIndex)
        {
            CellLayoutBox cellBox = CalculateCellLayoutBox(worksheet, rowIndex, columnIndex);
            bool isWrapped = worksheet.Cells[rowIndex, columnIndex].GetIsWrapped().Value;
            FontProperties fontProperties = worksheet.Cells.GetFontProperties(rowIndex, columnIndex);

            return CalculateCellContentSize(worksheet, cellBox, isWrapped, fontProperties);
        }

        /// <summary>
        /// Calculates the cell layout box.
        /// </summary>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="cellIndex">Index of the cell.</param>
        /// <returns>The cell layout box.</returns>
        public static CellLayoutBox CalculateCellLayoutBox(Worksheet worksheet, CellIndex cellIndex)
        {
            return CalculateCellLayoutBox(worksheet, cellIndex.RowIndex, cellIndex.ColumnIndex);
        }

        /// <summary>
        /// Calculates the cell layout box.
        /// </summary>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The cell layout box.</returns>
        public static CellLayoutBox CalculateCellLayoutBox(Worksheet worksheet, int rowIndex, int columnIndex)
        {
            RadWorksheetLayout worksheetLayout = worksheet.Workbook.GetWorksheetLayout(worksheet, false);

            CellRange mergedRange = null;
            if (worksheet.Cells.TryGetContainingMergedRange(rowIndex, columnIndex, out mergedRange) &&
                mergedRange.FromIndex.RowIndex == rowIndex && mergedRange.FromIndex.ColumnIndex == columnIndex)
            {
                return worksheetLayout.GetTopLeftMergedCellLayoutBox(mergedRange);
            }

            return worksheetLayout.GetCellLayoutBox(rowIndex, columnIndex);
        }

        internal static Size CalculateCellContentSize(Worksheet worksheet, CellLayoutBox cellBox, bool isWrapped,
            FontProperties? fontProperties = null, Func<string, FontProperties, double?, Size> measureMultilineText = null)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfNull(cellBox, "cellBox");
            
            if (!fontProperties.HasValue)
            {
                fontProperties = worksheet.Cells.GetFontProperties(cellBox.LongIndex, true);
            }

            measureMultilineText = measureMultilineText ?? LayoutHelper.DefaultTextMeasuringMethod;

            CellsPropertyBag propertyBag = worksheet.Cells.PropertyBag;
            long index = cellBox.LongIndex;

            RadHorizontalAlignment horizontalAlignment = propertyBag.GetPropertyValueRespectingStyle(CellPropertyDefinitions.HorizontalAlignmentProperty, worksheet, index);
            int indent = propertyBag.GetPropertyValueRespectingStyle(CellPropertyDefinitions.IndentProperty, worksheet, index);
            ICellValue cellValue = propertyBag.GetPropertyValueRespectingStyle(CellPropertyDefinitions.ValueProperty, worksheet, index);
            CellValueFormat format = propertyBag.GetPropertyValueRespectingStyle(CellPropertyDefinitions.FormatProperty, worksheet, index);

            int indentFactor = CalculateIndent(horizontalAlignment, indent);
            double cellIndent = indentFactor * SpreadsheetDefaultValues.IndentStep;
            double? wrappingWidth = GetWrappingWidth(cellBox.Width, cellIndent, isWrapped);

            return LayoutHelper.CalculateCellContentSize(cellValue, format, fontProperties.Value, cellIndent, wrappingWidth, measureMultilineText);
        }

        internal static Size CalculateCellContentSize(ICellValue cellValue, CellValueFormat format, FontProperties fontProperties, double cellIndent,
            double? wrappingWidth, Func<string, FontProperties, double?, Size> measureMultilineText)
        {
            Guard.ThrowExceptionIfNull(cellValue, "cellValue");
            Guard.ThrowExceptionIfNull(fontProperties, "fontProperties");
            Guard.ThrowExceptionIfNull(measureMultilineText, "measureMultilineText");

            string resultString = GetResultValueStringIncludingInvisibleSymbols(cellValue, format);
            if (string.IsNullOrEmpty(resultString))
            {
                return new Size(0, 0);
            }

            double indentPlusMargins = GetIndentPlusMargin(cellIndent);

            double? calculatedWrappingWidth = wrappingWidth - indentPlusMargins;
            double? actualWrappingWidth = calculatedWrappingWidth < 0 ? 0 : calculatedWrappingWidth;
            Size size = measureMultilineText(resultString, fontProperties, actualWrappingWidth);

            size.Width += indentPlusMargins;

            return size;
        }

        internal static string GetResultValueStringIncludingInvisibleSymbols(ICellValue value, CellValueFormat format)
        {
            string infos;

            lock (format)
            {
                FormulaCellValue formulaCellValue = value as FormulaCellValue;
                if (formulaCellValue != null)
                {
                    return GetResultValueStringIncludingInvisibleSymbols(formulaCellValue.GetResultValueAsCellValue(), format);
                }

                infos = format.GetFormatResult(value).InfosText;
            }

            return infos;
        }

        internal static ColumnWidth CalculateAutoColumnWidth(Worksheet worksheet, int columnIndex)
        {
            return LayoutHelper.CalculateAutoColumnWidth(worksheet, columnIndex, 0, SpreadsheetDefaultValues.RowCount - 1);
        }

        internal static ColumnWidth CalculateAutoColumnWidth(Worksheet worksheet, 
            int columnIndex, int fromRowIndex, int toRowIndex, bool respectNumberValuesOnly = false)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            double? columnWidth = null;

            long fromIndex = CellsPropertyBag.ConvertCellIndexToLong(fromRowIndex, columnIndex);
            long toIndex = CellsPropertyBag.ConvertCellIndexToLong(toRowIndex, columnIndex);

            ICompressedList<ICellValue> cellValues =
                worksheet.Cells.PropertyBag.GetPropertyValueRespectingStyle(CellPropertyDefinitions.ValueProperty, worksheet, fromIndex, toIndex);

            ICompressedList<int> cellIndents =
                worksheet.Cells.PropertyBag.GetPropertyValueRespectingStyle(CellPropertyDefinitions.IndentProperty, worksheet, fromIndex, toIndex);

            ICompressedList<bool> cellsIsWrapped =
                worksheet.Cells.PropertyBag.GetPropertyValueRespectingStyle(CellPropertyDefinitions.IsWrappedProperty, worksheet, fromIndex, toIndex);

            ICompressedList<CellValueFormat> cellFormats =
                worksheet.Cells.PropertyBag.GetPropertyValueRespectingStyle(CellPropertyDefinitions.FormatProperty, worksheet, fromIndex, toIndex);

            bool hasWrappedCell = false;
            foreach (var range in cellValues.GetNonDefaultRanges())
            {
                if (range.Value.ValueType == CellValueType.Empty ||
                    (respectNumberValuesOnly && range.Value.ResultValueType != CellValueType.Number))
                {
                    continue;
                }

                for (long i = range.Start; i <= range.End; i++)
                {
                    CellIndex cellIndex = CellsPropertyBag.ConvertLongToCellIndex(i);

                    bool isMerged = false;
                    bool isColumnOrRowMerge = false;
                    CellRange mergedRange;
                    if (worksheet.Cells.TryGetContainingMergedRange(cellIndex, out mergedRange))
                    {
                        isMerged = true;
                        isColumnOrRowMerge = (mergedRange.RowCount >= 1 && mergedRange.ColumnCount == 1) || (mergedRange.ColumnCount >= 1 && mergedRange.RowCount == 1);
                    }

                    if (isMerged && isColumnOrRowMerge)
                    {
                        continue;
                    }

                    FontProperties fontProperties = worksheet.Cells.GetFontProperties(cellIndex, true);
                    bool isWrapped = cellsIsWrapped.GetValue(i);
                    double cellWidth;
                    //??
                    if (isWrapped)
                    {
                        hasWrappedCell = true;
                    }
                    else
                    {
                        double cellIndent = cellIndents.GetValue(i) * SpreadsheetDefaultValues.IndentStep;
                        cellWidth = CalculateCellContentSize(range.Value, cellFormats.GetValue(i), fontProperties, cellIndent, null, LayoutHelper.DefaultTextMeasuringMethod).Width;
                        columnWidth = Math.Max(columnWidth.HasValue ? columnWidth.Value : 0, cellWidth);    
                    }   
                }
            }

            if (columnWidth.HasValue)
            {
                columnWidth = Math.Max(columnWidth.Value, SpreadsheetDefaultValues.MinColumnWidth);
                columnWidth = Math.Min(columnWidth.Value, SpreadsheetDefaultValues.MaxColumnWidth);
            }
            else
            {
                if (hasWrappedCell)
                {
                    columnWidth = worksheet.Columns[columnIndex].GetWidth().Value.Value;
                }
                else
                {
                    return null;
                }
            }

            return new ColumnWidth(columnWidth.Value, false);
        }

        private static double GetIndentPlusMargin(double cellIndent)
        {
            return cellIndent + SpreadsheetDefaultValues.TotalHorizontalCellMargin;
        }

        #endregion
    }
}