﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Printing;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Layout
{
    internal class PageLayout
    {
        #region Nested Types

        private class RangeDimensionInfo
        {
            public int StartIndex { get; private set; }
            public int EndIndex { get; private set; }
            public double Length { get; private set; }

            public RangeDimensionInfo(int startIndex, int endIndex, double length)
            {
                this.StartIndex = startIndex;
                this.EndIndex = endIndex;
                this.Length = length;
            }
        }

        #endregion


        #region Fields

        private readonly List<WorksheetPageLayoutBox> pages;
        private bool isPreparingPages;

        #endregion


        #region Properties

        internal int PagesCount
        {
            get
            {
                return this.pages.Count;
            }
        }

        internal IEnumerable<WorksheetPageLayoutBox> Pages
        {
            get
            {
                return this.pages;
            }
        }

        #endregion


        #region Constructor

        public PageLayout()
        {
            this.pages = new List<WorksheetPageLayoutBox>();
            this.isPreparingPages = false;
        }

        #endregion


        #region Methods

        internal WorksheetPageLayoutBox GetPage(int pageIndex)
        {
            return this.pages[pageIndex];
        }

        internal void PreparePages(IEnumerable<WorksheetPageLayoutBox> pages, int fromPageIndex, int toPageIndex, int copies, bool areOrderedCollated)
        {
            List<WorksheetPageLayoutBox> allPages = new List<WorksheetPageLayoutBox>(pages);
            List<WorksheetPageLayoutBox> orderedPages = new List<WorksheetPageLayoutBox>();

            using (this.BeginPrintPagesPreparation())
            {
                if (areOrderedCollated)
                {
                    for (int copy = 0; copy < copies; copy++)
                    {
                        for (int pageIndex = fromPageIndex; pageIndex <= toPageIndex; pageIndex++)
                        {
                            orderedPages.Add(new WorksheetPageLayoutBox(allPages[pageIndex]));
                        }
                    }
                }
                else
                {
                    for (int pageIndex = fromPageIndex; pageIndex <= toPageIndex; pageIndex++)
                    {
                        for (int copy = 0; copy < copies; copy++)
                        {
                            orderedPages.Add(new WorksheetPageLayoutBox(allPages[pageIndex]));
                        }
                    }
                }

                this.pages.AddRange(orderedPages);
            }
        }

        internal void PreparePages(Worksheet worksheet, IEnumerable<CellRange> selectedRanges)
        {
            using (this.BeginPrintPagesPreparation())
            {
                foreach (CellRange selectedRange in selectedRanges)
                {
                    this.pages.AddRange(this.CalculatePages(worksheet, selectedRange, true));
                }
            }
        }

        internal void PreparePages(Worksheet worksheet, bool ignorePrintArea)
        {
            using (this.BeginPrintPagesPreparation())
            {
                this.AddPages(worksheet, ignorePrintArea);
            }
        }

        internal void PreparePages(Workbook workbook, bool ignorePrintArea)
        {
            using (this.BeginPrintPagesPreparation())
            {
                foreach (Worksheet worksheet in workbook.Worksheets)
                {
                    this.AddPages(worksheet, ignorePrintArea);
                }
            }
        }

        internal Size GetMaximumRowColumnHeadingsContent(RadWorksheetLayout layout, Size scaleFactor)
        {
            CellRange visibleCellRange = new CellRange(0, 0, SpreadsheetDefaultValues.RowCount - 1, SpreadsheetDefaultValues.ColumnCount - 1);
            double width = layout.GetRowHeadingSize(visibleCellRange, SpreadsheetDefaultValues.RowCount - 1).Width;
            double height = layout.GetColumnHeadingSize(SpreadsheetDefaultValues.ColumnCount - 1).Height;

            return new Size(width * scaleFactor.Width, height * scaleFactor.Height);
        }

        private void AddPages(Worksheet worksheet, bool ignorePrintArea)
        {
            WorksheetPageSetup pageSetup = worksheet.WorksheetPageSetup;

            if (ignorePrintArea || !pageSetup.PrintArea.HasPrintAreaRanges)
            {
                this.pages.AddRange(this.CalculatePages(worksheet, worksheet.TotalUsedCellRange, true));
            }
            else
            {
                foreach (CellRange printAreaRange in pageSetup.PrintArea.Ranges)
                {
                    this.pages.AddRange(this.CalculatePages(worksheet, printAreaRange, false));
                }
            }
        }

        private IEnumerable<WorksheetPageLayoutBox> CalculatePages(Worksheet worksheet, CellRange printAreaRange, bool intersectPageBreaksAsInfinite)
        {
            WorksheetPageSetup pageSetup = worksheet.WorksheetPageSetup;

            IEnumerable<PageBreak> sortedHorizontalPageBreaks = pageSetup.PageBreaks.SortedHorizontalPageBreaks;
            IEnumerable<PageBreak> sortedVerticalPageBreaks = pageSetup.PageBreaks.SortedVerticalPageBreaks;
            PageOrder pageOrder = pageSetup.PageOrder;

            IEnumerable<CellRange> subPrintAreaRanges =
                PageLayout.GetSubCellRanges(printAreaRange, sortedHorizontalPageBreaks, sortedVerticalPageBreaks, pageOrder, intersectPageBreaksAsInfinite);

            foreach (CellRange subPrintAreaRange in subPrintAreaRanges)
            {
                foreach (WorksheetPageLayoutBox printPage in this.CalculatePagesIgnoringPageBreaks(worksheet, subPrintAreaRange))
                {
                    yield return printPage;
                }
            }
        }

        private static IEnumerable<CellRange> GetSubCellRanges(CellRange printAreaRange, IEnumerable<PageBreak> sortedHorizontalPageBreaks,
            IEnumerable<PageBreak> sortedVerticalPageBreaks, PageOrder pageOrder, bool intersectPageBreaksAsInfinite)
        {
            List<int> horizontalBoundaryIndexes =
                PageLayout.GetRangeBoundaryIndexes(sortedVerticalPageBreaks, printAreaRange.FromIndex.ColumnIndex, printAreaRange.ToIndex.ColumnIndex,
                (pageBreak) =>
                {
                    return pageBreak.FromIndex <= printAreaRange.FromIndex.RowIndex && printAreaRange.ToIndex.RowIndex <= pageBreak.ToIndex;
                }, intersectPageBreaksAsInfinite).ToList();

            List<int> verticalBoundaryIndexes =
                PageLayout.GetRangeBoundaryIndexes(sortedHorizontalPageBreaks, printAreaRange.FromIndex.RowIndex, printAreaRange.ToIndex.RowIndex,
                (pageBreak) =>
                {
                    return pageBreak.FromIndex <= printAreaRange.FromIndex.ColumnIndex && printAreaRange.ToIndex.ColumnIndex <= pageBreak.ToIndex;
                }, intersectPageBreaksAsInfinite).ToList();

            if (pageOrder == PageOrder.DownThenOver)
            {
                return PageLayout.GetSubRangesDownThenOver(horizontalBoundaryIndexes, verticalBoundaryIndexes);
            }
            else
            {
                return PageLayout.GetSubRangesOverThenDown(horizontalBoundaryIndexes, verticalBoundaryIndexes);
            }
        }

        private static IEnumerable<int> GetRangeBoundaryIndexes(IEnumerable<PageBreak> sortedPageBreaksCollection, int rangeStart, int rangeEnd,
            Func<PageBreak, bool> isIntersectingPageBreak, bool intersectPageBreaksAsInfinite)
        {
            int lastBoundaryIndex = rangeStart;
            yield return rangeStart;

            foreach (PageBreak pageBreak in sortedPageBreaksCollection)
            {
                int currentIndex = pageBreak.Index;

                if (currentIndex > rangeEnd)
                {
                    break;
                }

                bool intersects = (currentIndex > lastBoundaryIndex) && (intersectPageBreaksAsInfinite || isIntersectingPageBreak(pageBreak));

                if (intersects)
                {
                    lastBoundaryIndex = currentIndex;
                    yield return lastBoundaryIndex - 1;
                    yield return lastBoundaryIndex;
                }
            }

            yield return rangeEnd;
        }

        private static IEnumerable<CellRange> GetSubRangesDownThenOver(List<int> horizontalBoundaryIndexes, List<int> verticalBoundaryIndexes)
        {
            for (int horizontalIndex = 0; horizontalIndex < horizontalBoundaryIndexes.Count; horizontalIndex += 2)
            {
                for (int verticalIndex = 0; verticalIndex < verticalBoundaryIndexes.Count; verticalIndex += 2)
                {
                    int fromRow = verticalBoundaryIndexes[verticalIndex];
                    int toRow = verticalBoundaryIndexes[verticalIndex + 1];
                    int fromColumn = horizontalBoundaryIndexes[horizontalIndex];
                    int toColumn = horizontalBoundaryIndexes[horizontalIndex + 1];

                    yield return new CellRange(fromRow, fromColumn, toRow, toColumn);
                }
            }
        }

        private static IEnumerable<CellRange> GetSubRangesOverThenDown(List<int> horizontalBoundaryIndexes, List<int> verticalBoundaryIndexes)
        {
            for (int verticalIndex = 0; verticalIndex < verticalBoundaryIndexes.Count; verticalIndex += 2)
            {
                for (int horizontalIndex = 0; horizontalIndex < horizontalBoundaryIndexes.Count; horizontalIndex += 2)
                {
                    int fromRow = verticalBoundaryIndexes[verticalIndex];
                    int toRow = verticalBoundaryIndexes[verticalIndex + 1];
                    int fromColumn = horizontalBoundaryIndexes[horizontalIndex];
                    int toColumn = horizontalBoundaryIndexes[horizontalIndex + 1];

                    yield return new CellRange(fromRow, fromColumn, toRow, toColumn);
                }
            }
        }

        private IEnumerable<WorksheetPageLayoutBox> CalculatePagesIgnoringPageBreaks(Worksheet worksheet, CellRange cellRange)
        {
            RadWorksheetLayout layout = worksheet.Workbook.GetWorksheetLayout(worksheet, true);
            WorksheetPageSetup pageSetup = worksheet.WorksheetPageSetup;

            Size scaleFactor = pageSetup.FitToPages ?
                PageScaleFactorCalculator.CalculateScaleAccordingToFitToPages(worksheet) :
                pageSetup.ScaleFactor;

            Size rowColumnHeadingsSize = PageScaleFactorCalculator.CalculateRowHeadingsWidthColumnHeadingsHeight(worksheet, cellRange, scaleFactor);

            if (!PageMargins.AreValidMarginsWithinPageSize(pageSetup.Margins, pageSetup.RotatedPageSize, rowColumnHeadingsSize))
            {
                throw new PrintingException("Margins do not fit page size.",
                    "Spreadsheet_ErrorExpressions_MarginsDoNotFitPageSize");
            }

            double maxViewPortHeight = pageSetup.RotatedPageSize.Height - pageSetup.Margins.Top - pageSetup.Margins.Bottom - rowColumnHeadingsSize.Height;
            List<RangeDimensionInfo> viewportHeightDimensions = PageLayout.GetSplittedDimensionRanges(cellRange.FromIndex.RowIndex, cellRange.ToIndex.RowIndex,
                maxViewPortHeight,
                (index) =>
                {
                    return layout.GetRowHeight(index) * scaleFactor.Height;
                });

            double maxViewPortWidth = pageSetup.RotatedPageSize.Width - pageSetup.Margins.Left - pageSetup.Margins.Right - rowColumnHeadingsSize.Width;
            List<RangeDimensionInfo> viewportWidthDimensions = PageLayout.GetSplittedDimensionRanges(cellRange.FromIndex.ColumnIndex, cellRange.ToIndex.ColumnIndex,
                maxViewPortWidth,
                (index) =>
                {
                    double contentWidth = 0;
                    if (index == cellRange.ToIndex.ColumnIndex)
                    {
                        for (int rowIndex = cellRange.FromIndex.RowIndex; rowIndex <= cellRange.ToIndex.RowIndex; rowIndex++)
                        {
                            Size cellContentSize = LayoutHelper.CalculateCellContentSize(worksheet, rowIndex, index);
                            contentWidth = Math.Max(contentWidth, cellContentSize.Width);
                        }
                    }

                    return Math.Max(contentWidth, layout.GetColumnWidth(index)) * scaleFactor.Width;
                });

            Point initialOffset = new Point(layout.GetColumnLeft(cellRange.FromIndex.ColumnIndex), layout.GetRowTop(cellRange.FromIndex.RowIndex));

            if (pageSetup.PageOrder == PageOrder.DownThenOver)
            {
                return PageLayout.GetPagesDownThenOver(worksheet, viewportHeightDimensions, viewportWidthDimensions, initialOffset, scaleFactor);
            }
            else
            {
                return PageLayout.GetPagesOverThenDown(worksheet, viewportHeightDimensions, viewportWidthDimensions, initialOffset, scaleFactor);
            }
        }

        private static CellRange RangeDimensionsToCellRange(RangeDimensionInfo rowPrintRange, RangeDimensionInfo columnPrintRange)
        {
            return new CellRange(rowPrintRange.StartIndex, columnPrintRange.StartIndex, rowPrintRange.EndIndex, columnPrintRange.EndIndex);
        }

        private static IEnumerable<WorksheetPageLayoutBox> GetPagesOverThenDown(Worksheet worksheet, List<RangeDimensionInfo> viewportHeightDimensions,
            List<RangeDimensionInfo> viewportWidthDimensions, Point initialOffset, Size scaleFactor)
        {
            double topOffset = initialOffset.Y;

            for (int rowIndex = 0; rowIndex < viewportHeightDimensions.Count; rowIndex++)
            {
                double leftOffset = initialOffset.X;

                for (int columnIndex = 0; columnIndex < viewportWidthDimensions.Count; columnIndex++)
                {
                    RangeDimensionInfo rowPrintRange = viewportHeightDimensions[rowIndex];
                    RangeDimensionInfo columnPrintRange = viewportWidthDimensions[columnIndex];

                    CellRange range = RangeDimensionsToCellRange(rowPrintRange, columnPrintRange);
                    Rect viewPortRectangle = new Rect(leftOffset, topOffset, columnPrintRange.Length, rowPrintRange.Length);

                    yield return new WorksheetPageLayoutBox(worksheet, range, viewPortRectangle, scaleFactor);

                    leftOffset += viewportWidthDimensions[columnIndex].Length / scaleFactor.Width;
                }

                topOffset += viewportHeightDimensions[rowIndex].Length / scaleFactor.Height;
            }
        }

        private static IEnumerable<WorksheetPageLayoutBox> GetPagesDownThenOver(Worksheet worksheet, List<RangeDimensionInfo> viewportHeightDimensions,
            List<RangeDimensionInfo> viewportWidthDimensions, Point initialOffset, Size scaleFactor)
        {
            double leftOffset = initialOffset.X;

            for (int columnIndex = 0; columnIndex < viewportWidthDimensions.Count; columnIndex++)
            {
                double topOffset = initialOffset.Y;

                for (int rowIndex = 0; rowIndex < viewportHeightDimensions.Count; rowIndex++)
                {
                    RangeDimensionInfo rowPrintRange = viewportHeightDimensions[rowIndex];
                    RangeDimensionInfo columnPrintRange = viewportWidthDimensions[columnIndex];

                    CellRange range = RangeDimensionsToCellRange(rowPrintRange, columnPrintRange);
                    Rect viewPortRectangle = new Rect(leftOffset, topOffset, columnPrintRange.Length, rowPrintRange.Length);

                    yield return new WorksheetPageLayoutBox(worksheet, range, viewPortRectangle, scaleFactor);

                    topOffset += viewportHeightDimensions[rowIndex].Length / scaleFactor.Height;
                }

                leftOffset += viewportWidthDimensions[columnIndex].Length / scaleFactor.Width;
            }
        }

        private static List<RangeDimensionInfo> GetSplittedDimensionRanges(int startIndex, int endIndex, double maxDimensionLength, Func<int, double> indexToDimensionLength)
        {
            List<RangeDimensionInfo> result = new List<RangeDimensionInfo>();
            double sum = 0;

            for (int index = startIndex; index <= endIndex; index++)
            {
                double scaledValue = indexToDimensionLength(index);
                sum += scaledValue;
                if (sum > maxDimensionLength)
                {
                    if (index == startIndex)
                    {
                        result.Add(new RangeDimensionInfo(startIndex, index, maxDimensionLength));
                        startIndex = index + 1;
                        sum = 0;
                    }
                    else
                    {
                        result.Add(new RangeDimensionInfo(startIndex, index - 1, sum - scaledValue));
                        startIndex = index;
                        sum = scaledValue;
                    }
                }
            }

            if (startIndex <= endIndex)
            {
                result.Add(new RangeDimensionInfo(startIndex, endIndex, sum));
            }

            return result;
        }

        private IDisposable BeginPrintPagesPreparation()
        {
            Guard.ThrowExceptionIfTrue(this.isPreparingPages, "isPreparingPages");

            this.isPreparingPages = true;
            this.pages.Clear();

            return new DisposableObject(this.EndPrintPagesPreparation);
        }

        private void EndPrintPagesPreparation()
        {
            Guard.ThrowExceptionIfFalse(this.isPreparingPages, "isPreparingPages");

            DateTime dateTime = DateTime.Now;

            for (int pageIndex = 0; pageIndex < this.PagesCount; pageIndex++)
            {
                this.pages[pageIndex].PrepareHeaderFooterContext(pageIndex + 1, this.PagesCount, dateTime);
            }

            this.isPreparingPages = false;
        }

        #endregion
    }
}
