﻿using System;
using System.Collections.Generic;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Printing;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Layout
{
    /// <summary>
    /// Provides method for calculating the scale factor that needs to be set to a worksheet in order for it to fit in a specified number of pages when printed or exported to PDF.
    /// </summary>
    public static class PageScaleFactorCalculator
    {
        /// <summary>
        /// Calculates the maximum scale factor that can be set to a worksheet in order for it to fit into the number of pages specified in the <see cref="WorksheetPageSetup.FitToPagesWide"/> and <see cref="WorksheetPageSetup.FitToPagesTall"/>.
        /// </summary>
        /// <param name="worksheet">The worksheet.</param>
        /// <returns>The resulting scale factor.</returns>
        public static Size CalculateScaleAccordingToFitToPages(Worksheet worksheet)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");

            double usedCellRangeScaleFactor = GetScaleAccordingToFitToPages(worksheet, worksheet.UsedCellRange);
            return new Size(usedCellRangeScaleFactor, usedCellRangeScaleFactor);
        }

        /// <summary>
        /// Calculates the maximum scale factor that can be set to a worksheet in order for the specified ranges to fit into the number of pages specified in the <see cref="WorksheetPageSetup.FitToPagesWide"/> and <see cref="WorksheetPageSetup.FitToPagesTall"/>.
        /// </summary>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="includedRanges">All the ranges that will be included in the page.</param>
        /// <returns>The resulting scale factor.</returns>
        public static Size CalculateScaleAccordingToFitToPages(Worksheet worksheet, IEnumerable<CellRange> includedRanges)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfNull(includedRanges, "includedRanges");

            double minScaleFactor = 1;

            foreach (CellRange includedRange in includedRanges)
            {
                double rangeScaleFactor = GetScaleAccordingToFitToPages(worksheet, includedRange);
                minScaleFactor = Math.Min(rangeScaleFactor, minScaleFactor);
            }

            return new Size(minScaleFactor, minScaleFactor);
        }

        internal static Size CalculateRowHeadingsWidthColumnHeadingsHeight(Worksheet worksheet, CellRange cellRange, Size scaleFactor)
        {
            double rowHeadingsWidth = 0;
            double columnHeadingsHeight = 0;
            WorksheetPrintOptions printOptions = worksheet.WorksheetPageSetup.PrintOptions;

            if (printOptions.PrintRowColumnHeadings)
            {
                RadWorksheetLayout layout = worksheet.Workbook.GetWorksheetLayout(worksheet, true);

                for (int i = cellRange.FromIndex.ColumnIndex; i <= cellRange.ToIndex.ColumnIndex; i++)
                {
                    columnHeadingsHeight = Math.Max(columnHeadingsHeight, scaleFactor.Height * layout.GetColumnHeadingSize(i).Height);
                }

                if (!printOptions.IsDraftQuality)
                {
                    for (int i = cellRange.FromIndex.RowIndex; i <= cellRange.ToIndex.RowIndex; i++)
                    {
                        rowHeadingsWidth = Math.Max(rowHeadingsWidth, scaleFactor.Width * layout.GetRowHeadingSize(cellRange, i).Width);
                    }
                }
            }

            return new Size(rowHeadingsWidth, columnHeadingsHeight);
        }

        private static double GetScaleAccordingToFitToPages(Worksheet worksheet, CellRange range)
        {
            RadWorksheetLayout worksheetLayout = worksheet.Workbook.GetWorksheetLayout(worksheet, true);
            WorksheetPageSetup worksheetPageSetup = worksheet.WorksheetPageSetup;

            int fitToPagesWide = worksheetPageSetup.FitToPagesWide;
            int fitToPagesTall = worksheetPageSetup.FitToPagesTall;

            Size headingsSize = CalculateRowHeadingsWidthColumnHeadingsHeight(worksheet, range, new Size(1,1));

            double pageWidth = worksheetPageSetup.RotatedPageSize.Width - worksheetPageSetup.Margins.Left - worksheetPageSetup.Margins.Right - headingsSize.Width;
            double pageHeight = worksheetPageSetup.RotatedPageSize.Height - worksheetPageSetup.Margins.Top - worksheetPageSetup.Margins.Bottom - headingsSize.Height;

            int firstColumnIndex = range.FromIndex.ColumnIndex;
            int lastColumnIndex = range.ToIndex.ColumnIndex;
            double totalColumnWidth = worksheetLayout.GetColumnLeft(lastColumnIndex) - worksheetLayout.GetColumnLeft(firstColumnIndex) + worksheetLayout.GetColumnWidth(lastColumnIndex);

            int firstRowIndex = range.FromIndex.RowIndex;
            int lastRowIndex = range.ToIndex.RowIndex;
            double totalRowHeight = worksheetLayout.GetRowTop(lastRowIndex) - worksheetLayout.GetRowTop(firstRowIndex) + worksheetLayout.GetRowHeight(lastRowIndex);

            double widthScaleFactor = 1;
            double heightScaleFactor = 1;

            IEnumerable<double> columns = GetColumnWidths(range, worksheetLayout);
            IEnumerable<double> rows = GetRowHeights(range, worksheetLayout);

            if (fitToPagesWide != 0 || fitToPagesTall != 0)
            {
                if (fitToPagesWide == 0)
                {
                    heightScaleFactor = GetScaleAccordingToFitToPages(rows, totalRowHeight, pageHeight, fitToPagesTall);
                }
                else if (fitToPagesTall == 0)
                {
                    widthScaleFactor = GetScaleAccordingToFitToPages(columns, totalColumnWidth, pageWidth, fitToPagesWide);
                }
                else
                {
                    widthScaleFactor = GetScaleAccordingToFitToPages(columns, totalColumnWidth, pageWidth, fitToPagesWide);
                    heightScaleFactor = GetScaleAccordingToFitToPages(rows, totalRowHeight, pageHeight, fitToPagesTall);
                }
            }

            double minScaleFactor = Math.Min(widthScaleFactor, heightScaleFactor);
            return MathUtility.RoundDown(minScaleFactor, 2);
        }

        private static IEnumerable<double> GetColumnWidths(CellRange range, RadWorksheetLayout layout)
        {
            for (int col = range.FromIndex.ColumnIndex; col <= range.ToIndex.ColumnIndex; col++)
            {
                yield return layout.GetColumnWidth(col);
            }
        }

        private static IEnumerable<double> GetRowHeights(CellRange range, RadWorksheetLayout layout)
        {
            for (int row = range.FromIndex.RowIndex; row <= range.ToIndex.RowIndex; row++)
            {
                yield return layout.GetRowHeight(row);
            }
        }

        private static double GetScaleAccordingToFitToPages(IEnumerable<double> lengths, double totalLength, double availablePageLength, int pagesCount)
        {
            if (pagesCount == 1)
            {
                double scaleFactor = availablePageLength / totalLength;

                return Math.Max(Math.Min(1, scaleFactor), SpreadsheetDefaultValues.MinPageScaleFactor);
            }

            double minimalScaleFactor = 1;
            double averageLengthPerPage = totalLength / pagesCount;
            double currentPageRemainingLength = averageLengthPerPage;
            int currentPageNumber = 1;

            foreach (double length in lengths)
            {
                bool isLastPage = currentPageNumber == pagesCount;

                if (!isLastPage && currentPageRemainingLength < length)
                {
                    bool shouldBeMovedToNextPage = currentPageRemainingLength < (length / 2);

                    double currentPageLength = shouldBeMovedToNextPage ? (averageLengthPerPage - currentPageRemainingLength) : (length + averageLengthPerPage - currentPageRemainingLength);
                    minimalScaleFactor = Math.Min(minimalScaleFactor, availablePageLength / currentPageLength);
                    currentPageNumber++;

                    currentPageRemainingLength = shouldBeMovedToNextPage ? (averageLengthPerPage - length) : averageLengthPerPage;
                }
                else
                {
                    currentPageRemainingLength -= length;
                }
            }

            minimalScaleFactor = Math.Min(minimalScaleFactor, availablePageLength / (averageLengthPerPage - currentPageRemainingLength));

            return minimalScaleFactor;
        }
    }
}
