﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Shapes;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Layout
{
    internal class RadWorksheetLayout
    {
        #region Fields

        private Worksheet worksheet;
        private Size size;
        private Size scrollableSize;
        private CellIndex frozenCellIndex;
        private bool isPrintingLayout;
        private bool isMeasureValid;
        private readonly double[] rowsTop;
        private readonly double[] columnsLeft;
        private SizeI visibleSize;

        private readonly bool[] rowsIsEmpty;
        private readonly bool[] columnsIsEmpty;
        private Dictionary<long, Size> indexToCellContentSizeCache;
        private readonly Dictionary<int, Rect> shapeIdToShapeRectCache;
        private CellRange maxInvalidatedCellRange;

        #endregion


        #region Properties

        public double Width
        {
            get
            {
                return this.size.Width;
            }
        }

        public double Height
        {
            get
            {
                return this.size.Height;
            }
        }

        public Size Size
        {
            get
            {
                return this.size;
            }
        }

        public Size ScrollableSize
        {
            get
            {
                return this.scrollableSize;
            }
        }

        public bool IsMeasureValid
        {
            get
            {
                return this.isMeasureValid;
            }
        }

        internal Worksheet Worksheet
        {
            get
            {
                return this.worksheet;
            }
            private set
            {
                this.worksheet = value;
            }
        }

        internal SizeI VisibleSize
        {
            get
            {
                return this.visibleSize;
            }
            set
            {
                if (this.visibleSize != value)
                {
                    Telerik.Windows.Documents.Spreadsheet.Utilities.Guard.ThrowExceptionIfOutOfRange(0, SpreadsheetDefaultValues.RowCount, value.Height, "visibleSize");
                    Telerik.Windows.Documents.Spreadsheet.Utilities.Guard.ThrowExceptionIfOutOfRange(0, SpreadsheetDefaultValues.ColumnCount, value.Width, "visibleSize");

                    this.visibleSize = value;
                }
            }
        }

        #endregion


        #region Constructors

        public RadWorksheetLayout(SizeI visibleSize)
        {
            Telerik.Windows.Documents.Spreadsheet.Utilities.Guard.ThrowExceptionIfOutOfRange(0, SpreadsheetDefaultValues.RowCount, visibleSize.Height, "visibleSize");
            Telerik.Windows.Documents.Spreadsheet.Utilities.Guard.ThrowExceptionIfOutOfRange(0, SpreadsheetDefaultValues.ColumnCount, visibleSize.Width, "visibleSize");

            this.visibleSize = visibleSize;

            this.rowsTop = new double[visibleSize.Height];
            this.columnsLeft = new double[visibleSize.Width];

            this.rowsIsEmpty = new bool[visibleSize.Height];
            this.columnsIsEmpty = new bool[visibleSize.Width];
            this.ResetIsRowsColumnsEmpty();

            this.indexToCellContentSizeCache = new Dictionary<long, Size>();
            this.shapeIdToShapeRectCache = new Dictionary<int, Rect>();
        }

        private void ResetIsRowsColumnsEmpty()
        {
            for (int i = 0; i < this.visibleSize.Height; i++)
            {
                this.rowsIsEmpty[i] = true;
            }
            for (int i = 0; i < this.visibleSize.width; i++)
            {
                this.columnsIsEmpty[i] = true;
            }
        }

        #endregion


        #region Worksheet EventHandlers

        private void Cells_MergedCellsChanged(object sender, MergedCellRangesChangedEventArgs e)
        {
            this.InvalidateCellContentSizeCache(e.CellRange);
        }

        private void RowsProperties_PropertyChanged(object sender, RowColumnPropertyChangedEventArgs e)
        {
            if (e.Property.AffectsLayout)
            {
                this.InvalidateCellContentSizeCache(CellRange.FromRowRange(e.FromIndex, e.ToIndex));
            }
        }

        private void ColumnsProperties_PropertyChanged(object sender, RowColumnPropertyChangedEventArgs e)
        {
            if (e.Property.AffectsLayout)
            {
                this.InvalidateCellContentSizeCache(CellRange.FromColumnRange(e.FromIndex, e.ToIndex));
            }
        }

        private void Properties_PropertyChanged(object sender, CellPropertyChangedEventArgs e)
        {
            if (e.ShouldRespectAffectLayoutProperty && e.Property.AffectsLayout)
            {
                this.InvalidateCellContentSizeCache(e.CellRange);
            }
        }

        private void Worksheet_LayoutInvalidated(object sender, EventArgs e)
        {
            this.InvalidateLayout();
        }

        #endregion


        #region Methods

        private void InvalidateLayout()
        {
            this.isMeasureValid = false;
        }

        private void InvalidateCellContentSizeCache(CellRange cellRange)
        {
            CellRange restrictedCellRange = cellRange;

            if (cellRange != null)
            {
                restrictedCellRange = CellRange.RestrictCellRange(cellRange, this.visibleSize);
            }
            this.maxInvalidatedCellRange = CellRange.MaxOrNull(this.maxInvalidatedCellRange, restrictedCellRange);
        }

        internal void SetWorksheet(Worksheet worksheet, bool isPrintingLayout)
        {
            this.isPrintingLayout = isPrintingLayout;

            if (worksheet == null ||
                this.Worksheet == worksheet)
            {
                return;
            }

            this.DetachFromWorksheetEvents();

            this.Worksheet = worksheet;

            this.AttachToWorksheetEvents();

            this.ResetIsRowsColumnsEmpty();

            var usedCellRange = this.Worksheet.UsedCellRange;
            if (usedCellRange.IsSingleCell &&
                this.Worksheet.Cells.IsDefault(usedCellRange.FromIndex))
            {
                this.InvalidateCellContentSizeCache(null);
            }
            else
            {
                this.InvalidateCellContentSizeCache(CellRange.RestrictCellRange(this.Worksheet.UsedCellRange, this.visibleSize));
            }

            this.InvalidateLayout();
        }

        private void AttachToWorksheetEvents()
        {
            if (this.Worksheet != null)
            {
                this.Worksheet.LayoutInvalidated += this.Worksheet_LayoutInvalidated;
                this.Worksheet.Cells.MergedCellsChanged += this.Cells_MergedCellsChanged;
                this.Worksheet.Cells.PropertyBag.PropertyChanged += this.Properties_PropertyChanged;
                this.Worksheet.Rows.PropertyBag.PropertyChanged += this.RowsProperties_PropertyChanged;
                this.Worksheet.Columns.PropertyBag.PropertyChanged += this.ColumnsProperties_PropertyChanged;
            }
        }

        private void DetachFromWorksheetEvents()
        {
            if (this.Worksheet != null)
            {
                this.Worksheet.LayoutInvalidated -= this.Worksheet_LayoutInvalidated;
                this.Worksheet.Cells.MergedCellsChanged -= this.Cells_MergedCellsChanged;
                this.Worksheet.Cells.PropertyBag.PropertyChanged -= this.Properties_PropertyChanged;
                this.Worksheet.Rows.PropertyBag.PropertyChanged -= this.RowsProperties_PropertyChanged;
                this.Worksheet.Columns.PropertyBag.PropertyChanged -= this.ColumnsProperties_PropertyChanged;
            }
        }

        private void UpdateCellContentSizesCache()
        {
            if (this.maxInvalidatedCellRange == null)
            {
                return;
            }

            for (int i = this.maxInvalidatedCellRange.FromIndex.RowIndex; i <= this.maxInvalidatedCellRange.ToIndex.RowIndex; i++)
            {
                this.rowsIsEmpty[i] = true;
            }

            for (int i = this.maxInvalidatedCellRange.FromIndex.ColumnIndex; i <= this.maxInvalidatedCellRange.ToIndex.ColumnIndex; i++)
            {
                this.columnsIsEmpty[i] = true;
            }

            Dictionary<long, Size> oldDictionary = this.indexToCellContentSizeCache;
            this.indexToCellContentSizeCache = new Dictionary<long, Size>();

            foreach (var pair in oldDictionary)
            {
                int rowIndex, columnIndex;
                CellsPropertyBag.ConvertLongToRowAndColumnIndexes(pair.Key, out rowIndex, out columnIndex);

                if (!this.maxInvalidatedCellRange.Contains(rowIndex, columnIndex))
                {
                    this.indexToCellContentSizeCache.Add(pair.Key, pair.Value);
                }
            }

            List<long> cellsToCalculateContentSize = new List<long>();
            foreach (var range in this.Worksheet.Cells.PropertyBag.GetPropertyValueCollection<ICellValue>(CellPropertyDefinitions.ValueProperty).GetNonDefaultRanges())
            {
                for (long i = range.Start; i <= range.End; i++)
                {
                    int rowIndex, columnIndex;
                    CellsPropertyBag.ConvertLongToRowAndColumnIndexes(i, out rowIndex, out columnIndex);

                    if (columnIndex >= this.visibleSize.Width ||
                        rowIndex >= this.visibleSize.Height)
                    {
                        break;
                    }

                    bool isMerged = false;
                    bool isRowMerge = false;
                    CellRange mergedRange;
                    if (this.Worksheet.Cells.TryGetContainingMergedRange(rowIndex, columnIndex, out mergedRange))
                    {
                        isMerged = true;
                        isRowMerge = mergedRange.ColumnCount >= 1 && mergedRange.RowCount == 1;
                    }

                    if ((this.maxInvalidatedCellRange.Contains(rowIndex, columnIndex) && (!isMerged || (isMerged && isRowMerge))) ||
                        this.IsFormulaCellValue(i))
                    {
                        cellsToCalculateContentSize.Add(i);
                    }

                    this.rowsIsEmpty[rowIndex] = false;
                    this.columnsIsEmpty[columnIndex] = false;
                }
            }

            CellProperties[] cellsFontProperties = this.CalculateCellProperties(cellsToCalculateContentSize);

            this.CalculateCellContentSizes(cellsToCalculateContentSize, cellsFontProperties);

            this.maxInvalidatedCellRange = null;
        }

        private bool IsFormulaCellValue(long index)
        {
            return (this.Worksheet.Cells.PropertyBag.GetPropertyValue(CellPropertyDefinitions.ValueProperty, index) as FormulaCellValue) != null;
        }

        private void CalculateCellContentSizes(List<long> cellsToCalculateContentSize, CellProperties[] cellsFontProperties)
        {
            if (cellsToCalculateContentSize.Count == 0)
            {
                return;
            }

            int itemsPerTaskCount = cellsToCalculateContentSize.Count;
            int tasksCount = 1;

            if (cellsToCalculateContentSize.Count > 8)
            {
                itemsPerTaskCount = (int)Math.Ceiling((float)cellsToCalculateContentSize.Count / 8);
                tasksCount = cellsToCalculateContentSize.Count / itemsPerTaskCount + ((cellsToCalculateContentSize.Count % itemsPerTaskCount > 0) ? 1 : 0);
            }

            List<Action> actions = new List<Action>(cellsToCalculateContentSize.Count);
            for (int j = 0; j < tasksCount; j++)
            {
                int fromIndex = j * itemsPerTaskCount;
                int toIndex = Math.Min(cellsToCalculateContentSize.Count, fromIndex + itemsPerTaskCount);

                actions.Add(new Action(
                    () =>
                    {
                        for (int i = fromIndex; i < toIndex; i++)
                        {
                            CellProperties cellProperties = cellsFontProperties[i];

                            long cellIndex = cellsToCalculateContentSize[i];
                            int indentFactor = LayoutHelper.CalculateIndent(cellProperties.HorizontalAlignment, cellProperties.Indent);
                            double cellIndent = indentFactor * SpreadsheetDefaultValues.IndentStep;

                            int rowIndex, columnIndex;
                            CellsPropertyBag.ConvertLongToRowAndColumnIndexes(cellIndex, out rowIndex, out columnIndex);

                            double columnWidths = this.GetColumnWidth(columnIndex);
                            CellRange mergedRange;
                            if (this.Worksheet.Cells.TryGetContainingMergedRange(rowIndex, columnIndex, out mergedRange))
                            {
                                columnWidths = 0;
                                for (int col = mergedRange.FromIndex.ColumnIndex; col <= mergedRange.ToIndex.ColumnIndex; col++)
                                {
                                    columnWidths += this.GetColumnWidth(col);
                                }
                            }

                            double? wrappingWidth = LayoutHelper.GetWrappingWidth(columnWidths, cellIndent, cellProperties.IsWrapped);

                            Size cellSize = LayoutHelper.CalculateCellContentSize(cellProperties.Value, cellProperties.Format, cellProperties.FontProperties,
                                cellIndent, wrappingWidth, LayoutHelper.DefaultTextMeasuringMethod);

                            lock (this.indexToCellContentSizeCache)
                            {
                                this.indexToCellContentSizeCache[cellIndex] = cellSize;
                            }
                        }
                    }));
            }

#if WPF
            TasksHelper.DoAsync(actions);
#else
            foreach (var action in actions)
            {
                action();
            }
#endif
        }

        private void UpdateShapesSizeAndPosition()
        {
            foreach (FloatingShapeBase shape in this.Worksheet.Shapes)
            {
                Rect rect = CalculateShapeBoundingRect(shape);
                this.shapeIdToShapeRectCache[shape.Id] = rect;
            }
        }

        internal Size GetRowHeadingSize(CellRange visibleRange, int rowIndex)
        {
            RowLayoutBox rowBox = this.GetRowLayoutBox(rowIndex);
            double height = rowBox.Height;
            string rowName = this.Worksheet.HeaderNameRenderingConverter.ConvertRowIndexToName(new HeaderNameRenderingConverterContext(this.Worksheet, visibleRange), rowIndex);
            double width = RadWorksheetLayout.GetRowHeadingDesiredWidth(rowName);

            return new Size(width, height);
        }

        internal static double GetRowHeadingDesiredWidth(string rowName)
        {
            int symbolsCount = rowName.Length;
            double width = (int)(SpreadsheetDefaultValues.RowColumnHeadingSymbolWidth * symbolsCount +
                SpreadsheetDefaultValues.RowColumnHeadingsPadding.Left + SpreadsheetDefaultValues.RowColumnHeadingsPadding.Right);

            width = Math.Max(SpreadsheetDefaultValues.RowColumnHeadingMinimumSize.Width, width);

            return width;
        }

        internal Size GetColumnHeadingSize(int columnIndex)
        {
            ColumnLayoutBox columnBox = this.GetColumnLayoutBox(columnIndex);

            return new Size(columnBox.Width, SpreadsheetDefaultValues.RowColumnHeadingMinimumSize.Height);
        }

        private Rect CalculateShapeBoundingRect(FloatingShapeBase shape)
        {
            return SpreadsheetHelper.CalculateShapeBoundingRect(this.GetShapeTopLeft(shape), shape);
        }

        public Point GetShapeTopLeft(FloatingShapeBase shape)
        {
            CellLayoutBox cellLayoutBox = this.GetCellLayoutBox(shape.CellIndex);
            double y = cellLayoutBox.Top;
            double x = cellLayoutBox.Left;
            y += shape.OffsetY;
            x += shape.OffsetX;

            return new Point(x, y);
        }

        private CellProperties[] CalculateCellProperties(List<long> cellIndexes)
        {
            CellProperties[] cellProperties = new CellProperties[cellIndexes.Count];

            int itemsPerTaskCount = 100;
            int tasksCount = cellProperties.Length / itemsPerTaskCount + ((cellProperties.Length % itemsPerTaskCount > 0) ? 1 : 0);

            Action[] actions = new Action[tasksCount];
            for (int i = 0; i < tasksCount; i++)
            {
                int fromIndex = i * itemsPerTaskCount;
                int toIndex = Math.Min(cellIndexes.Count, fromIndex + itemsPerTaskCount);
                var action = new Action(
                    () =>
                    {
                        int minRowIndex = int.MaxValue;
                        int minColumnIndex = int.MaxValue;
                        int maxRowIndex = 0;
                        int maxColumnIndex = 0;
                        for (int k = fromIndex; k < toIndex; k++)
                        {
                            int rowIndex, columnIndex;
                            CellsPropertyBag.ConvertLongToRowAndColumnIndexes(cellIndexes[k], out rowIndex, out columnIndex);

                            minRowIndex = Math.Min(rowIndex, minRowIndex);
                            minColumnIndex = Math.Min(columnIndex, minColumnIndex);
                            maxRowIndex = Math.Max(rowIndex, maxRowIndex);
                            maxColumnIndex = Math.Max(columnIndex, maxColumnIndex);
                        }

                        CellRange cellRange = new CellRange(minRowIndex, minColumnIndex, maxRowIndex, maxColumnIndex);
                        CellRangeFontProperties rangeFontProperties = this.Worksheet.Cells.GetFontProperties(cellRange, true);
                        ICompressedList<bool> isWrappedValues = this.Worksheet.Cells.GetPropertyValueRespectingStyle(CellPropertyDefinitions.IsWrappedProperty, cellRange);
                        ICompressedList<ICellValue> cellValues = this.Worksheet.Cells.PropertyBag.GetPropertyValue(CellPropertyDefinitions.ValueProperty, cellRange);
                        ICompressedList<RadHorizontalAlignment> horizontalAlignmentValues = this.Worksheet.Cells.GetPropertyValueRespectingStyle(CellPropertyDefinitions.HorizontalAlignmentProperty, cellRange);
                        ICompressedList<int> indentValues = this.Worksheet.Cells.GetPropertyValueRespectingStyle(CellPropertyDefinitions.IndentProperty, cellRange);
                        ICompressedList<CellValueFormat> formatValues = this.Worksheet.Cells.GetPropertyValueRespectingStyle(CellPropertyDefinitions.FormatProperty, cellRange);

                        for (int k = fromIndex; k < toIndex; k++)
                        {
                            long longCellIndex = cellIndexes[k];

                            int rowIndex, columnIndex;
                            CellsPropertyBag.ConvertLongToRowAndColumnIndexes(longCellIndex, out rowIndex, out columnIndex);

                            cellProperties[k].FontProperties = rangeFontProperties.GetFontProperties(longCellIndex, this.Worksheet.Workbook.Theme);
                            cellProperties[k].IsWrapped = isWrappedValues.GetValue(longCellIndex);
                            cellProperties[k].Value = cellValues.GetValue(longCellIndex);
                            cellProperties[k].HorizontalAlignment = horizontalAlignmentValues.GetValue(longCellIndex);
                            cellProperties[k].Indent = indentValues.GetValue(longCellIndex);
                            cellProperties[k].Format = formatValues.GetValue(longCellIndex);
                        }
                    });

                actions[i] = action;
            }

            TasksHelper.DoAsync(actions);

            for (int i = 0; i < cellProperties.Length; i++)
            {
                FormulaCellValue formulaCellValue = cellProperties[i].Value as FormulaCellValue;
                if (formulaCellValue != null)
                {
                    cellProperties[i].Value = formulaCellValue.GetResultValueAsCellValue();
                }
            }

            return cellProperties;
        }

        public void Measure(CellIndex frozenCellIndex)
        {
            this.frozenCellIndex = frozenCellIndex;

            this.Measure();
        }

        public void Measure()
        {
            if (this.isMeasureValid)
            {
                return;
            }

            Size desiredSize = new Size();
            if (this.Worksheet == null)
            {
                desiredSize = new Size(0, 0);
            }
            else
            {
                double totalWidth = this.CalculateFinalColumnWidths();

                this.UpdateCellContentSizesCache();

                double totalHeight = this.CalculateFinalRowHeights();

                desiredSize.Width = totalWidth;
                desiredSize.Height = totalHeight;

                this.UpdateShapesSizeAndPosition();
            }

            this.size = desiredSize;

            double width = (this.frozenCellIndex == null) ? this.Width : this.Width - this.columnsLeft[this.frozenCellIndex.ColumnIndex];
            double height = (this.frozenCellIndex == null) ? this.Height : this.Height - this.rowsTop[this.frozenCellIndex.RowIndex];
            this.scrollableSize = new Size(width, height);

            this.isMeasureValid = true;

            this.OnMeasureExecuted();
        }

        private double CalculateFinalColumnWidths()
        {
            // Calculate final column widths.
            ICompressedList<ColumnWidth> columnWidthsList =
                this.Worksheet.Columns.PropertyBag.GetColumnWidthPropertyValueRespectingHidden();

            double totalWidth = 0;
            foreach (Range<long, ColumnWidth> range in columnWidthsList)
            {
                double columnWidth;
                for (int i = (int)range.Start; i <= range.End && i < SpreadsheetDefaultValues.ColumnCount; i++)
                {
                    columnWidth = range.Value.Value;

                    this.columnsLeft[i] = totalWidth;
                    totalWidth += columnWidth;
                }
            }

            return totalWidth;
        }

        private double CalculateFinalRowHeights()
        {
            // Calculate final row heights.
            ICompressedList<RowHeight> rowHeightsList =
                this.Worksheet.Rows.PropertyBag.GetRowHeightPropertyValueRespectingHidden();

            double totalHeight = 0;

            foreach (var range in rowHeightsList)
            {
                double rowHeight = 0;
                if (range.Value.IsCustom)
                {
                    rowHeight = range.Value.Value;

                    for (int i = (int)range.Start; i <= range.End && i < SpreadsheetDefaultValues.RowCount; i++)
                    {
                        this.rowsTop[i] = totalHeight;
                        totalHeight += rowHeight;
                    }

                }
                else
                {
                    for (int i = (int)range.Start; i <= range.End && i < SpreadsheetDefaultValues.RowCount; i++)
                    {
                        rowHeight = this.CalculateAutoRowHeight(i, rowHeightsList.GetDefaultValue().Value);

                        this.rowsTop[i] = totalHeight;
                        totalHeight += rowHeight;
                    }
                }
            }

            return totalHeight;
        }

        private double CalculateAutoRowHeight(int rowIndex, double rowHeight)
        {
            if (!this.rowsIsEmpty[rowIndex])
            {
                Size cellSize;
                CellRange cellRange = CellRange.RestrictCellRange(this.Worksheet.UsedCellRange, this.VisibleSize);

                for (int columnIndex = 0; columnIndex <= cellRange.ToIndex.ColumnIndex; columnIndex++)
                {
                    if (!this.columnsIsEmpty[columnIndex])
                    {
                        if (this.indexToCellContentSizeCache.TryGetValue(
                            CellsPropertyBag.ConvertCellIndexToLong(rowIndex, columnIndex), out cellSize))
                        {
                            rowHeight = Math.Max(cellSize.Height, rowHeight);
                        }
                    }
                }
            }

            return rowHeight;
        }

        private static CellLayoutBox GetCellLayoutBox(RowLayoutBox rowBox, ColumnLayoutBox columnBox, CellMergeState mergeState)
        {
            return new CellLayoutBox(
                rowBox.RowIndex, columnBox.ColumnIndex,
                new Rect(columnBox.Left, rowBox.Top, columnBox.Width, rowBox.Height), mergeState);
        }

        internal CellLayoutBox GetTopLeftMergedCellLayoutBox(CellRange mergedRange)
        {
            RowLayoutBox fromRowBox = this.GetRowLayoutBox(mergedRange.FromIndex.RowIndex);
            ColumnLayoutBox fromColumnBox = this.GetColumnLayoutBox(mergedRange.FromIndex.ColumnIndex);
            RowLayoutBox toRowBox = this.GetRowLayoutBox(mergedRange.ToIndex.RowIndex);
            ColumnLayoutBox toColumnBox = this.GetColumnLayoutBox(mergedRange.ToIndex.ColumnIndex);


            double width = toColumnBox.BoundingRectangle.Right - fromColumnBox.Left;
            double height = toRowBox.BoundingRectangle.Bottom - fromRowBox.Top;

            return new CellLayoutBox(fromRowBox.RowIndex, fromColumnBox.ColumnIndex, new Rect(fromColumnBox.Left, fromRowBox.Top, width, height), CellMergeState.TopLeftCellInMergedRange);
        }

        public RowLayoutBox GetRowLayoutBox(int rowIndex)
        {
            Telerik.Windows.Documents.Spreadsheet.Utilities.Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);

            double top = this.rowsTop[rowIndex];
            double left = 0;
            double width = this.Width;
            double height;
            if (rowIndex < this.rowsTop.Length - 1)
            {
                height = this.rowsTop[rowIndex + 1] - this.rowsTop[rowIndex];
            }
            else
            {
                height = this.Height - this.rowsTop[rowIndex];
            }

            return new RowLayoutBox(rowIndex, new Rect(left, top, width, height));
        }

        public ColumnLayoutBox GetColumnLayoutBox(int columnIndex)
        {
            Telerik.Windows.Documents.Spreadsheet.Utilities.Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            double top = 0;
            double left = this.columnsLeft[columnIndex];
            double width;
            if (columnIndex < this.columnsLeft.Length - 1)
            {
                width = this.columnsLeft[columnIndex + 1] - this.columnsLeft[columnIndex];
            }
            else
            {
                width = this.Width - this.columnsLeft[columnIndex];
            }

            double height = this.Height;

            return new ColumnLayoutBox(columnIndex, new Rect(left, top, width, height));
        }

        public CellLayoutBox GetCellLayoutBox(CellIndex cellIndex)
        {
            return this.GetCellLayoutBox(cellIndex.RowIndex, cellIndex.ColumnIndex);
        }

        public CellLayoutBox GetCellLayoutBox(int rowIndex, int columnIndex)
        {
            RowLayoutBox rowBox = this.GetRowLayoutBox(rowIndex);
            ColumnLayoutBox columnBox = this.GetColumnLayoutBox(columnIndex);

            CellMergeState mergeState = this.Worksheet.Cells.GetMergeState(rowIndex, columnIndex);

            return GetCellLayoutBox(rowBox, columnBox, mergeState);
        }

        public ShapeLayoutBox GetShapeLayoutBox(int shapeId)
        {
            Rect rect;
            if (this.shapeIdToShapeRectCache.TryGetValue(shapeId, out rect))
            {
                return new ShapeLayoutBox(shapeId, rect);
            }

            return null;
        }

        private IEnumerable<RowLayoutBox> GetVisibleRowBoxes(SheetViewport sheetViewport)
        {
            //fixed
            ViewportPane fixedPane = sheetViewport[ViewportPaneType.Fixed];
            fixedPane = fixedPane.IsEmpty ? sheetViewport[ViewportPaneType.HorizontalScrollable] : fixedPane;
            if (!fixedPane.IsEmpty)
            {
                foreach (RowLayoutBox item in this.GetVisibleRowBoxes(fixedPane))
                {
                    yield return item;
                }
            }

            //scrollable
            ViewportPane scrollablePane = sheetViewport[ViewportPaneType.Scrollable];
            scrollablePane = scrollablePane.IsEmpty ? sheetViewport[ViewportPaneType.VerticalScrollable] : scrollablePane;
            if (!scrollablePane.IsEmpty)
            {
                foreach (RowLayoutBox item in this.GetVisibleRowBoxes(scrollablePane))
                {
                    yield return item;
                }
            }
        }

        private IEnumerable<ColumnLayoutBox> GetVisibleColumnBoxes(SheetViewport sheetViewport)
        {
            //fixed
            ViewportPane fixedPane = sheetViewport[ViewportPaneType.Fixed];
            fixedPane = fixedPane.IsEmpty ? sheetViewport[ViewportPaneType.VerticalScrollable] : fixedPane;
            if (!fixedPane.IsEmpty)
            {
                foreach (ColumnLayoutBox item in this.GetVisibleColumnBoxes(fixedPane))
                {
                    yield return item;
                }
            }

            //scrollable
            ViewportPane scrollablePane = sheetViewport[ViewportPaneType.Scrollable];
            scrollablePane = scrollablePane.IsEmpty ? sheetViewport[ViewportPaneType.HorizontalScrollable] : scrollablePane;
            if (!scrollablePane.IsEmpty)
            {
                foreach (ColumnLayoutBox item in this.GetVisibleColumnBoxes(scrollablePane))
                {
                    yield return item;
                }
            }
        }

        public void GetVisibleBoxes(SheetViewport sheetViewport,
            out List<RowLayoutBox> rowLayoutBoxes,
            out List<ColumnLayoutBox> columnLayoutBoxes,
            out Dictionary<ViewportPaneType, IEnumerable<CellLayoutBox>> cellBoxes)
        {
            rowLayoutBoxes = new List<RowLayoutBox>(this.GetVisibleRowBoxes(sheetViewport));
            columnLayoutBoxes = new List<ColumnLayoutBox>(this.GetVisibleColumnBoxes(sheetViewport));

            cellBoxes = new Dictionary<ViewportPaneType, IEnumerable<CellLayoutBox>>();
            Dictionary<ViewportPaneType, HashSet<CellIndex>> mergedCellBoxes = new Dictionary<ViewportPaneType, HashSet<CellIndex>>();
            foreach (ViewportPaneType paneType in Enum.GetValues(typeof(ViewportPaneType)))
            {
                cellBoxes.Add(paneType, new HashSet<CellLayoutBox>());
                mergedCellBoxes.Add(paneType, new HashSet<CellIndex>());
            }

            foreach (RowLayoutBox rowBox in rowLayoutBoxes)
            {
                foreach (ColumnLayoutBox columnBox in columnLayoutBoxes)
                {
                    ViewportPane viewportPane = sheetViewport.GetViewportPaneFromDocumentPoint(new Point(columnBox.Left, rowBox.Top));
                    ViewportPaneType viewportPaneType = viewportPane.ViewportPaneType;

                    HashSet<CellLayoutBox> cellBoxesHashSet = cellBoxes[viewportPaneType] as HashSet<CellLayoutBox>;

                    CellRange mergedRange;
                    if (this.Worksheet.Cells.TryGetContainingMergedRange(rowBox.RowIndex, columnBox.ColumnIndex, out mergedRange))
                    {
                        CellRange restrictedMergedRange = CellRange.RestrictCellRange(mergedRange, this.visibleSize);
                        if (!mergedCellBoxes[viewportPaneType].Contains(restrictedMergedRange.FromIndex))
                        {
                            cellBoxesHashSet.Add(this.GetTopLeftMergedCellLayoutBox(restrictedMergedRange));
                            mergedCellBoxes[viewportPaneType].Add(restrictedMergedRange.FromIndex);
                        }
                        else if (restrictedMergedRange.FromIndex.RowIndex == rowBox.RowIndex &&
                            restrictedMergedRange.FromIndex.ColumnIndex == columnBox.ColumnIndex)
                        {
                            cellBoxesHashSet.Add(GetCellLayoutBox(rowBox, columnBox, CellMergeState.NonTopLeftCellInMergedRange));
                        }
                    }
                    else
                    {
                        cellBoxesHashSet.Add(GetCellLayoutBox(rowBox, columnBox, CellMergeState.NotMerged));
                    }
                }
            }
        }

        private IEnumerable<RowLayoutBox> GetVisibleRowBoxes(ViewportPane viewportPane)
        {
            // TODO: rethink this.isPrintingLayout
            double bottomDelta = (viewportPane.ViewportPaneType == ViewportPaneType.Fixed ||
                viewportPane.ViewportPaneType == ViewportPaneType.HorizontalScrollable ||
                this.isPrintingLayout) ? 0.1 : 0;

            int firstVisibleRowIndex = viewportPane.VisibleRange.FromIndex.RowIndex;
            double top = this.rowsTop[firstVisibleRowIndex];

            double left = 0;
            double width = this.Width;
            double height;

            int i = firstVisibleRowIndex;
            if (i == this.VisibleSize.Height)
            {
                yield break;
            }

            do
            {
                if (!worksheet.Rows.GetIsHidden(i))
                {
                    height = this.GetRowHeight(i);
                    if (!height.IsZero())
                    {
                        yield return new RowLayoutBox(i, new Rect(left, top, width, height));
                    }

                    top += height;
                }

                i++;
            }
            while (top <= viewportPane.Rect.Bottom - bottomDelta && i < this.visibleSize.Height && i <= viewportPane.VisibleRange.ToIndex.RowIndex);
        }

        private IEnumerable<ColumnLayoutBox> GetVisibleColumnBoxes(ViewportPane viewportPane)
        {
            // TODO: rethink this.isPrintingLayout
            double rightDelta = (viewportPane.ViewportPaneType == ViewportPaneType.Fixed ||
                viewportPane.ViewportPaneType == ViewportPaneType.VerticalScrollable ||
                this.isPrintingLayout) ? 0.1 : 0;

            int firstVisibleColumnIndex = viewportPane.VisibleRange.FromIndex.ColumnIndex;
            double left = this.columnsLeft[firstVisibleColumnIndex];

            double top = 0;
            double width;
            double height = this.Height;

            int i = firstVisibleColumnIndex;
            if (i == this.VisibleSize.Width)
            {
                yield break;
            }

            do
            {
                if (!worksheet.Columns.GetIsHidden(i))
                {
                    width = this.GetColumnWidth(i);
                    if (!width.IsZero())
                    {
                        yield return new ColumnLayoutBox(i, new Rect(left, top, width, height));
                    }

                    left += width;
                }

                i++;
            }
            while (left <= viewportPane.Rect.Right - rightDelta && i < this.visibleSize.Width && i <= viewportPane.VisibleRange.ToIndex.ColumnIndex);
        }

        internal void GetNearestTopLeftViewportContainingCell(int rowIndex, int columnIndex, Rect viewport, out int resultRowIndex, out int resultColumnIndex)
        {
            double columnWidth = this.GetColumnWidth(columnIndex);
            double width = viewport.Width - columnWidth;

            if (width > 0 && columnIndex > 0)
            {
                columnWidth = this.GetColumnWidth(columnIndex - 1);
                while (width > columnWidth)
                {
                    width -= columnWidth;
                    columnIndex--;
                    if (columnIndex == 0)
                    {
                        break;
                    }
                    columnWidth = this.GetColumnWidth(columnIndex - 1);
                }
            }

            double rowHeight = this.GetRowHeight(rowIndex);
            double height = viewport.Height - rowHeight;

            if (height > 0 && rowIndex > 0)
            {
                rowHeight = this.GetRowHeight(rowIndex - 1);
                while (height > rowHeight)
                {
                    height -= rowHeight;
                    rowIndex--;
                    if (rowIndex == 0)
                    {
                        break;
                    }
                    rowHeight = this.GetRowHeight(rowIndex - 1);
                }
            }

            resultRowIndex = rowIndex;
            resultColumnIndex = columnIndex;
        }

        internal CellLayoutBox GetNearestViewPortContainingCell(SheetViewport viewport, CellIndex cellIndex)
        {
            ViewportPane scrollablePane = viewport[ViewportPaneType.Scrollable];
            int resultColumnIndex = scrollablePane.VisibleRange.FromIndex.ColumnIndex;
            int resultRowIndex = scrollablePane.VisibleRange.FromIndex.RowIndex;

            double left = scrollablePane.Rect.Left;
            double top = scrollablePane.Rect.Top;
            CellLayoutBox cellBox = this.GetCellLayoutBox(cellIndex);

            int nearestTopLeftViewportRowIndex;
            int nearestTopLeftViewportColumnIndex;
            this.GetNearestTopLeftViewportContainingCell(cellIndex.RowIndex, cellIndex.ColumnIndex, scrollablePane.Rect, out nearestTopLeftViewportRowIndex, out nearestTopLeftViewportColumnIndex);

            if (cellBox.BoundingRectangle.Right > scrollablePane.Rect.Right)
            {
                resultColumnIndex = nearestTopLeftViewportColumnIndex;
                left = this.GetColumnLeft(resultColumnIndex);
            }

            if (cellBox.Left < left)
            {
                resultColumnIndex = cellBox.ColumnIndex;
                left = cellBox.Left;
            }

            if (cellBox.BoundingRectangle.Bottom > scrollablePane.Rect.Bottom)
            {
                resultRowIndex = nearestTopLeftViewportRowIndex;
                top = this.GetRowTop(resultRowIndex);
            }

            if (cellBox.Top < top)
            {
                resultRowIndex = cellBox.RowIndex;
                top = cellBox.Top;
            }

            return this.GetCellLayoutBox(resultRowIndex, resultColumnIndex);
        }

        private int FindColumnIndexAtOffset(double offset, bool isPreviousWithPriority)
        {
            int startIndex = 0;
            int endIndex = this.columnsLeft.Length - 1;

            Func<int, double> getRangeStart =
                (index) =>
                {
                    return this.columnsLeft[index];
                };

            Func<int, double> getRangeEnd =
                (index) =>
                {
                    return this.GetColumnRight(index);
                };

            double rangeStart = getRangeStart(startIndex);
            double rangeEnd = getRangeEnd(endIndex);

            offset = Math.Max(rangeStart, offset);
            offset = Math.Min(rangeEnd, offset);

            if (offset.EqualsDouble(rangeStart))
            {
                return startIndex;
            }

            if (offset.EqualsDouble(rangeEnd))
            {
                return endIndex;
            }

            return this.FindRangeIndexAtOffset(offset, startIndex, endIndex, getRangeStart, getRangeEnd, isPreviousWithPriority);
        }

        private int FindRowIndexAtOffset(double offset, bool isPreviousWithPriority)
        {
            int startIndex = 0;
            int endIndex = this.rowsTop.Length - 1;

            Func<int, double> getRangeStart =
                (index) =>
                {
                    return Math.Round(this.rowsTop[index], 3);
                };

            Func<int, double> getRangeEnd =
                (index) =>
                {
                    return Math.Round(this.GetRowBottom(index), 3);
                };

            double rangeStart = getRangeStart(startIndex);
            double rangeEnd = getRangeEnd(endIndex);

            offset = Math.Max(rangeStart, offset);
            offset = Math.Min(rangeEnd, offset);

            offset = Math.Round(offset, 3);

            if (offset.EqualsDouble(rangeStart))
            {
                return startIndex;
            }

            if (offset.EqualsDouble(rangeEnd))
            {
                return endIndex;
            }

            return this.FindRangeIndexAtOffset(offset, startIndex, endIndex, getRangeStart, getRangeEnd, isPreviousWithPriority);
        }

        private int FindRangeIndexAtOffset(double offset, int startIndex, int endIndex, Func<int, double> getRangeStart, Func<int, double> getRangeEnd, bool isPreviousWithPriority)
        {
            // This is binary search
            int initialStartIndex = startIndex;
            int initialEndIndex = endIndex;
            int itemsCount = endIndex - startIndex;
            int middleIndex = startIndex + itemsCount / 2;

            double rangeStart = getRangeStart(middleIndex);
            double rangeEnd = getRangeEnd(middleIndex);

            while (rangeStart > offset || offset.GreaterOrEqualDouble(rangeEnd))
            {
                if (offset.EqualsDouble(rangeStart))
                {
                    return (isPreviousWithPriority) ? Math.Max(initialStartIndex, middleIndex - 1) : Math.Min(initialEndIndex, middleIndex);
                }

                if (offset.EqualsDouble(rangeEnd))
                {
                    return (isPreviousWithPriority) ? Math.Max(initialStartIndex, middleIndex) : Math.Min(initialEndIndex, middleIndex + 1);
                }

                if (rangeStart < offset && offset < rangeEnd)
                {
                    return middleIndex;
                }

                if (offset < rangeStart)
                {
                    endIndex = middleIndex - 1;
                }

                if (rangeEnd < offset)
                {
                    startIndex = middleIndex + 1;
                }

                itemsCount = endIndex - startIndex;

                if (itemsCount <= 0)
                {
                    return startIndex;
                }

                middleIndex = startIndex + itemsCount / 2;

                rangeStart = getRangeStart(middleIndex);
                rangeEnd = getRangeEnd(middleIndex);
            }

            return middleIndex;
        }

        private double GetRowBottom(int index)
        {
            double rowBottom = 0;

            if (index < this.rowsTop.Length - 1)
            {
                rowBottom = this.rowsTop[index + 1];
            }
            else
            {
                rowBottom = this.rowsTop[index] + this.GetRowHeight(index) - 0.1;
            }

            return rowBottom;
        }

        private double GetColumnRight(int index)
        {
            double columnRight = 0;

            if (index < this.columnsLeft.Length - 1)
            {
                columnRight = this.columnsLeft[index + 1];
            }
            else
            {
                columnRight = this.columnsLeft[index] + this.GetColumnWidth(index) - 0.1;
            }

            return columnRight;
        }

        public CellIndex GetCellIndexFromPoint(double x, double y, bool getLeftCellWithPriority, bool getTopCellWithPriority, bool respectHiddenCollumns = false, bool respectHiddenRows = false)
        {
            int rowIndex;
            int columnIndex;

            this.GetCellIndexFromPoint(x, y, getLeftCellWithPriority, getTopCellWithPriority, out rowIndex, out columnIndex, respectHiddenCollumns, respectHiddenRows);

            return new CellIndex(rowIndex, columnIndex);
        }

        public void GetCellIndexFromPoint(double x, double y, bool getLeftCellWithPriority, bool getTopCellWithPriority, out int rowIndex, out int columnIndex, bool respectHiddenCollumns = false, bool respectHiddenRows = false)
        {
            rowIndex = this.GetRowColumnIndexFromOffset(y, SpreadsheetDefaultValues.RowCount - 1, getTopCellWithPriority, respectHiddenCollumns, this.FindRowIndexAtOffset, this.GetRowHeight);
            columnIndex = this.GetRowColumnIndexFromOffset(x, SpreadsheetDefaultValues.ColumnCount - 1, getLeftCellWithPriority, respectHiddenRows, this.FindColumnIndexAtOffset, this.GetColumnWidth);
        }

        private int GetRowColumnIndexFromOffset(double offset, int maxValue, bool getPreviousIndexWithPriority, bool respectHiddenIndexes, Func<double, bool, int> findIndexAtOffset, Func<int, double> getWidthHeight)
        {
            int index = findIndexAtOffset(offset, getPreviousIndexWithPriority);

            if (respectHiddenIndexes)
            {
                int newIndex = index;
                if (getPreviousIndexWithPriority)
                {
                    while (newIndex - 1 >= 0 && getWidthHeight(newIndex) == 0)
                    {
                        newIndex--;
                    }
                }
                else
                {
                    while (newIndex + 1 <= maxValue && getWidthHeight(newIndex) == 0)
                    {
                        newIndex++;
                    }
                }

                if (getWidthHeight(newIndex) == 0)
                {
                    if (newIndex == 0)
                    {
                        while (newIndex + 1 <= maxValue && getWidthHeight(newIndex) == 0)
                        {
                            newIndex++;
                        }
                    }
                    if (newIndex == maxValue)
                    {
                        while (newIndex - 1 >= 0 && getWidthHeight(newIndex) == 0)
                        {
                            newIndex--;
                        }
                    }
                }
                index = newIndex;
            }

            return index;
        }

        public Point GetTopLeftPointFromCellIndex(int rowIndex, int columnIndex, bool ignoreMergedState)
        {
            if (ignoreMergedState)
            {
                return new Point(this.GetColumnLeft(columnIndex), this.GetRowTop(rowIndex));
            }
            else
            {
                CellLayoutBox layoutBox = this.GetCellLayoutBox(rowIndex, columnIndex);
                return new Point(layoutBox.Left, layoutBox.Top);
            }
        }

        public Point GetTopLeftPointFromCellIndex(CellIndex cellIndex)
        {
            return this.GetTopLeftPointFromCellIndex(cellIndex.RowIndex, cellIndex.ColumnIndex);
        }

        public Point GetTopLeftPointFromCellIndex(int rowIndex, int columnIndex)
        {
            return this.GetTopLeftPointFromCellIndex(rowIndex, columnIndex, false);
        }

        public Point GetPointFromCellIndexAndOffset(CellIndex cellIndex, double offsetX, double offsetY)
        {
            return this.GetPointFromCellIndexAndOffset(cellIndex.RowIndex, cellIndex.ColumnIndex, offsetX, offsetY);
        }

        public Point GetPointFromCellIndexAndOffset(int rowIndex, int columnIndex, double offsetX, double offsetY)
        {
            Point topLeft = this.GetTopLeftPointFromCellIndex(rowIndex, columnIndex);
            return new Point(topLeft.X + offsetX, topLeft.Y + offsetY);
        }

        public CellIndex GetCellIndexAndOffsetFromPoint(Point point, out double offsetX, out double offsetY)
        {
            int rowIndex;
            int columnIndex;

            this.GetCellIndexFromPoint(point.X, point.Y, false, false, out rowIndex, out columnIndex, false, false);
            Point topLeft = this.GetTopLeftPointFromCellIndex(rowIndex, columnIndex);

            offsetX = point.X - topLeft.X;
            offsetY = point.Y - topLeft.Y;

            return new CellIndex(rowIndex, columnIndex);
        }

        internal bool TryGetCellContentSize(long cellIndex, out Size result)
        {
            return this.indexToCellContentSizeCache.TryGetValue(cellIndex, out result);
        }

        internal Size GetCellContentSize(long cellIndex, CellLayoutBox cellBox, bool isWrapped, FontProperties fontProperties)
        {
            this.UpdateCellContentSizesCache();

            Size result;
            if (!this.TryGetCellContentSize(cellIndex, out result))
            {
                int rowIndex;
                int columnIndex;
                CellsPropertyBag.ConvertLongToRowAndColumnIndexes(cellIndex, out rowIndex, out columnIndex);
                result = LayoutHelper.CalculateCellContentSize(worksheet, cellBox, isWrapped, fontProperties);
            }

            return result;
        }

        internal double GetRowTop(int rowIndex)
        {
            return this.rowsTop[rowIndex];
        }

        internal double GetColumnLeft(int columnIndex)
        {
            return this.columnsLeft[columnIndex];
        }

        internal double GetRowHeight(int rowIndex)
        {
            if (rowIndex == SpreadsheetDefaultValues.RowCount - 1)
            {
                return this.Height - this.rowsTop[rowIndex];
            }

            return this.rowsTop[rowIndex + 1] - this.rowsTop[rowIndex];
        }

        internal double GetColumnWidth(int columnIndex)
        {
            if (columnIndex == SpreadsheetDefaultValues.ColumnCount - 1)
            {
                return this.Width - this.columnsLeft[columnIndex];
            }

            return this.columnsLeft[columnIndex + 1] - this.columnsLeft[columnIndex];
        }

        internal Point GetMaxBottomRight(SizeI visibleSize)
        {
            Point maxBottomRight = new Point(this.Width, this.Height);

            if (visibleSize.Height < SpreadsheetDefaultValues.RowCount)
            {
                maxBottomRight.Y = this.GetRowTop(visibleSize.Height);
            }

            if (visibleSize.Width < SpreadsheetDefaultValues.ColumnCount)
            {
                maxBottomRight.X = this.GetColumnLeft(visibleSize.Width);
            }

            return maxBottomRight;
        }

        #endregion


        #region Events

        public event EventHandler MeasureExecuted;
        protected virtual void OnMeasureExecuted()
        {
            if (this.MeasureExecuted != null)
            {
                this.MeasureExecuted(this, EventArgs.Empty);
            }
        }

        #endregion
    }
}