﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

#if SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.Maths;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.Layout
{
    /// <summary>
    /// Represents the sheet viewport.
    /// </summary>
    public class SheetViewport
    {
        #region Fields

        private RadWorksheetLayout worksheetLayout;
        private readonly Size displaySize;
        private readonly Point fixedTopLeft;
        private readonly Point horizontalTopLeft;
        private readonly Point verticalTopLeft;
        private readonly Point scrollableTopLeft;
        private readonly Point maxBottomRight;
        private Point frozenPoint;
        private CellIndex frozenCellIndex;

        private readonly Size scaleFactor;
        private Dictionary<ViewportPaneType, ViewportPane> paneToViewport;
        private static readonly List<ViewportPaneType> PaneTypes = Enum.GetValues(typeof(ViewportPaneType)).Cast<ViewportPaneType>().ToList();
        private ViewportPane[] viewportPanes;
        private CellRange[] visisbleRanges;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the width.
        /// </summary>
        /// <value>The width.</value>
        public double Width
        {
            get
            {
                return this[ViewportPaneType.VerticalScrollable].Rect.Width + this[ViewportPaneType.Scrollable].Rect.Width;
            }
        }

        /// <summary>
        /// Gets the height.
        /// </summary>
        /// <value>The height.</value>
        public double Height
        {
            get
            {
                return this[ViewportPaneType.HorizontalScrollable].Rect.Height + this[ViewportPaneType.Scrollable].Rect.Height;
            }
        }

        /// <summary>
        /// Gets the viewport panes count.
        /// </summary>
        /// <value>The viewport panes count.</value>
        public int ViewportPanesCount
        {
            get
            {
                return this.ViewportPanes.Count();
            }
        }

        /// <summary>
        /// Gets the viewport panes.
        /// </summary>
        /// <value>The viewport panes.</value>
        public IEnumerable<ViewportPane> ViewportPanes
        {
            get
            {
                //TODO: Consider removing empty panes form dicitionary on init.
                if (this.viewportPanes == null)
                {
                    this.viewportPanes = this.paneToViewport.Values.Where(p => !p.IsEmpty).ToArray();
                }

                return this.viewportPanes;
            }
        }

        private CellRange[] VisibleRanges
        {
            get
            {
                if (this.visisbleRanges == null)
                {
                    this.visisbleRanges = this.ViewportPanes.Select(p => p.VisibleRange).ToArray();
                }
                return this.visisbleRanges;
            }
        }

        internal CellIndex TopLeftFixedCellIndex
        {
            get
            {
                ViewportPane fixedPane = this[ViewportPaneType.Fixed];
                ViewportPane horizontalScrollablePane = this[ViewportPaneType.HorizontalScrollable];
                ViewportPane verticalScrollablePane = this[ViewportPaneType.VerticalScrollable];
                ViewportPane scrollablePane = this[ViewportPaneType.Scrollable];

                CellIndex fixedPaneTopLeft = fixedPane.VisibleRange.FromIndex;
                CellIndex verticalScrollableTopLeft = verticalScrollablePane.VisibleRange.FromIndex;
                CellIndex horizontalScrollableTopLeft = horizontalScrollablePane.VisibleRange.FromIndex;
                CellIndex scrollableTopLeft = scrollablePane.VisibleRange.FromIndex;

                if (this.worksheetLayout == null)
                {
                    return scrollableTopLeft;
                }

                Func<CellIndex, bool> isCellHidden = new Func<CellIndex, bool>(
                    (cellIndex) =>
                    {
                        return this.worksheetLayout.GetColumnWidth(cellIndex.ColumnIndex) == 0 ||
                            this.worksheetLayout.GetRowHeight(cellIndex.RowIndex) == 0;
                    });

                if (!fixedPane.IsEmpty || isCellHidden(fixedPaneTopLeft))
                {
                    return fixedPaneTopLeft;
                }
                else if (!verticalScrollablePane.IsEmpty || isCellHidden(verticalScrollableTopLeft))
                {
                    return verticalScrollableTopLeft;
                }
                else if (!horizontalScrollablePane.IsEmpty || isCellHidden(horizontalScrollableTopLeft))
                {
                    return horizontalScrollableTopLeft;
                }
                else
                {
                    return scrollableTopLeft;
                }
            }
        }

        internal CellIndex TopLeftScrollableCellIndex
        {
            get
            {
                return this[ViewportPaneType.Scrollable].VisibleRange.FromIndex;
            }
        }

        internal CellIndex FrozenCellIndex
        {
            get
            {
                if (this.frozenCellIndex == null)
                {
                    this.frozenCellIndex = new CellIndex(0, 0);
                }

                return this.frozenCellIndex;
            }
        }

        internal Size DisplaySize
        {
            get
            {
                return this.displaySize;
            }
        }

        internal Size ScaleFactor
        {
            get
            {
                return this.scaleFactor;
            }
        }

        #endregion


        #region Indexers

        /// <summary>
        /// Gets or sets the <see cref="Telerik.Windows.Documents.Spreadsheet.Layout.ViewportPane" /> with the specified pane.
        /// </summary>
        /// <value>The viewport pane.</value>
        public ViewportPane this[ViewportPaneType pane]
        {
            get
            {
                return this.paneToViewport[pane];
            }
            set
            {
                this.paneToViewport[pane] = value;
            }
        }

        #endregion


        #region Constructors

        internal SheetViewport()
        {
            this.scaleFactor = new Size(1, 1);
            this.BeforeInitialize();
            this.AfterInitialize();
        }

        internal SheetViewport(RadWorksheetLayout worksheetLayout,
            Size displaySize, Size scaleFactor,
            Point fixedTopLeft, Point scrollableTopLeft, CellIndex frozenCellIndex,
            Point maxBottomRight)
        {
            this.worksheetLayout = worksheetLayout;

            this.displaySize = displaySize;
            this.scaleFactor = scaleFactor;
            this.fixedTopLeft = fixedTopLeft;
            this.horizontalTopLeft = new Point(scrollableTopLeft.X, fixedTopLeft.Y);
            this.verticalTopLeft = new Point(fixedTopLeft.X, scrollableTopLeft.Y);
            this.scrollableTopLeft = scrollableTopLeft;
            this.maxBottomRight = maxBottomRight;

            double frozenTop = this.worksheetLayout.GetRowTop(frozenCellIndex.RowIndex);
            double frozenLeft = this.worksheetLayout.GetColumnLeft(frozenCellIndex.ColumnIndex);
            this.frozenPoint = new Point(frozenLeft, frozenTop);
            this.frozenCellIndex = frozenCellIndex;

            this.BeforeInitialize();
            this.InitializeViewPorts();
            this.InitializeVisibleRanges();
            this.AfterInitialize();
        }

        #endregion


        #region Viewport Methods

        private void InitializeViewPorts()
        {
            double width = this.displaySize.Width / this.scaleFactor.Width;
            double height = this.displaySize.Height / this.scaleFactor.Height;

            double fixedWidth = Math.Max(this.frozenPoint.X - this.fixedTopLeft.X, 0);
            fixedWidth = Math.Min(fixedWidth, width);
            double maxFixedWidth = Math.Max(0, this.maxBottomRight.X - this.fixedTopLeft.X);
            fixedWidth = Math.Min(maxFixedWidth, fixedWidth);

            double fixedHeight = Math.Max(this.frozenPoint.Y - this.fixedTopLeft.Y, 0);
            fixedHeight = Math.Min(fixedHeight, height);
            double maxFixedHeight = Math.Max(0, this.maxBottomRight.Y - this.fixedTopLeft.Y);
            fixedHeight = Math.Min(maxFixedHeight, fixedHeight);

            double horizontalWidth = width - fixedWidth;
            double maxHorizontalWidth = Math.Max(0, this.maxBottomRight.X - this.horizontalTopLeft.X);
            horizontalWidth = Math.Min(maxHorizontalWidth, horizontalWidth);

            double horizontalHeight = fixedHeight;
            double verticalWidth = fixedWidth;

            double verticalHeight = height - fixedHeight;
            double maxverticalHeight = Math.Max(0, this.maxBottomRight.Y - this.verticalTopLeft.Y);
            verticalHeight = Math.Min(maxverticalHeight, verticalHeight);

            double scrollableWidth = horizontalWidth;
            double scrollableHeight = verticalHeight;

            this[ViewportPaneType.Fixed].Rect = new Rect(
                this.fixedTopLeft.X,
                this.fixedTopLeft.Y,
                fixedWidth,
                fixedHeight);

            this[ViewportPaneType.HorizontalScrollable].Rect = new Rect(
                this.horizontalTopLeft.X,
                this.horizontalTopLeft.Y,
                horizontalWidth,
                horizontalHeight);

            this[ViewportPaneType.VerticalScrollable].Rect = new Rect(
                this.verticalTopLeft.X,
                this.verticalTopLeft.Y,
                verticalWidth,
                verticalHeight);

            this[ViewportPaneType.Scrollable].Rect = new Rect(
                this.scrollableTopLeft.X,
                this.scrollableTopLeft.Y,
                scrollableWidth,
                scrollableHeight);
        }

        private void BeforeInitialize()
        {
            this.paneToViewport = new Dictionary<ViewportPaneType, ViewportPane>();

            foreach (ViewportPaneType pane in PaneTypes)
            {
                this[pane] = new ViewportPane(pane);
            }
        }

        private void InitializeVisibleRanges()
        {
            foreach (ViewportPane viewportPane in this.ViewportPanes)
            {
                if (viewportPane.IsEmpty)
                {
                    continue;
                }

                Rect rect = viewportPane.Rect;
                Point topLeftPoint = new Point(rect.Left, rect.Top);
                Point bottomRightPoint = new Point(rect.Right, rect.Bottom);
                var viewportPaneType = viewportPane.ViewportPaneType;
                CellIndex topLeftCellIndex = this.GetPaneTopLeftCellIndex(viewportPaneType, topLeftPoint);
                CellIndex bottomRightCellIndex = this.GetPaneBottomRightCellIndex(viewportPaneType, bottomRightPoint);

                viewportPane.VisibleRange = new CellRange(topLeftCellIndex, bottomRightCellIndex);
                viewportPane.TopLeftPoint = this.worksheetLayout.GetPointFromCellIndexAndOffset(topLeftCellIndex, 0, 0);
            }
        }

        private CellIndex GetPaneTopLeftCellIndex(ViewportPaneType paneType, Point topLeftPoint)
        {
            int rowIndex;
            int columnIndex;
            this.worksheetLayout.GetCellIndexFromPoint(topLeftPoint.X, topLeftPoint.Y, false, false, out rowIndex, out columnIndex);

            if (this.frozenCellIndex != null)
            {
                if (paneType == ViewportPaneType.HorizontalScrollable || paneType == ViewportPaneType.Scrollable)
                {
                    columnIndex = Math.Max(columnIndex, this.frozenCellIndex.ColumnIndex);
                }
                if (paneType == ViewportPaneType.VerticalScrollable || paneType == ViewportPaneType.Scrollable)
                {
                    rowIndex = Math.Max(rowIndex, this.frozenCellIndex.RowIndex);
                }
            }

            return new CellIndex(rowIndex, columnIndex);
        }

        private CellIndex GetPaneBottomRightCellIndex(ViewportPaneType paneType, Point bottomRightPoint)
        {
            int rowIndex;
            int columnIndex;
            this.worksheetLayout.GetCellIndexFromPoint(bottomRightPoint.X, bottomRightPoint.Y, true, true, out rowIndex, out columnIndex, true, true);

            if (this.frozenCellIndex != null)
            {
                if (paneType == ViewportPaneType.HorizontalScrollable || paneType == ViewportPaneType.Fixed)
                {
                    rowIndex = Math.Min(rowIndex, Math.Max(this.frozenCellIndex.RowIndex - 1, 0));
                }
                if (paneType == ViewportPaneType.VerticalScrollable || paneType == ViewportPaneType.Fixed)
                {
                    columnIndex = Math.Min(columnIndex, Math.Max(this.frozenCellIndex.ColumnIndex - 1, 0));
                }
            }

            return new CellIndex(rowIndex, columnIndex);
        }

        private void AfterInitialize()
        {
            ViewportPane fixedPane = this[ViewportPaneType.Fixed];
            ViewportPane horizontalScrollablePane = this[ViewportPaneType.HorizontalScrollable];
            ViewportPane verticalScrollablePane = this[ViewportPaneType.VerticalScrollable];
            ViewportPane scrollablePane = this[ViewportPaneType.Scrollable];

            fixedPane.BoundingRect = new Rect(
                0,
                0,
                fixedPane.Rect.Width * this.scaleFactor.Width,
                fixedPane.Rect.Height * this.scaleFactor.Height);

            horizontalScrollablePane.BoundingRect = new Rect(
                fixedPane.BoundingRect.Width,
                0,
                horizontalScrollablePane.Rect.Width * this.scaleFactor.Width,
                horizontalScrollablePane.Rect.Height * this.scaleFactor.Height);

            verticalScrollablePane.BoundingRect = new Rect(
                0,
                fixedPane.BoundingRect.Height,
                verticalScrollablePane.Rect.Width * this.scaleFactor.Width,
                verticalScrollablePane.Rect.Height * this.scaleFactor.Height);

            scrollablePane.BoundingRect = new Rect(
                fixedPane.BoundingRect.Width,
                fixedPane.BoundingRect.Height,
                scrollablePane.Rect.Width * this.scaleFactor.Width,
                scrollablePane.Rect.Height * this.scaleFactor.Height);
        }

        /// <summary>
        /// Gets the viewport pane from viewport point.
        /// </summary>
        /// <param name="point">The view point.</param>
        /// <returns>The viewport pane.</returns>
        private ViewportPane GetViewportPaneFromViewportPoint(Point point)
        {
            ViewportPane fixedPane = this[ViewportPaneType.Fixed];

            Point crossPoint = new Point(fixedPane.Rect.Width, fixedPane.Rect.Height);

            return this.GetViewportPaneForPointRelativeToCrossPoint(point, crossPoint);
        }

        /// <summary>
        /// Gets the viewport pane from document point.
        /// </summary>
        /// <param name="point">The document point.</param>
        /// <returns>The viewport pane.</returns>
        public ViewportPane GetViewportPaneFromDocumentPoint(Point point)
        {
            return this.GetViewportPaneForPointRelativeToCrossPoint(point, this.frozenPoint);
        }

        private ViewportPane GetViewportPaneForPointRelativeToCrossPoint(Point point, Point crossPoint)
        {
            ViewportPaneType pointContainingPane = ViewportPaneType.Scrollable;

            if (point.X < crossPoint.X)
            {
                if (point.Y < crossPoint.Y)
                {
                    pointContainingPane = ViewportPaneType.Fixed;
                }
                else
                {
                    pointContainingPane = ViewportPaneType.VerticalScrollable;
                }
            }
            else
            {
                if (point.Y < crossPoint.Y)
                {
                    pointContainingPane = ViewportPaneType.HorizontalScrollable;
                }
                else
                {
                    pointContainingPane = ViewportPaneType.Scrollable;
                }
            }

            if (pointContainingPane == ViewportPaneType.Fixed && this[pointContainingPane].IsEmpty)
            {
                if (this[ViewportPaneType.HorizontalScrollable].IsEmpty)
                {
                    pointContainingPane = ViewportPaneType.VerticalScrollable;
                }
                else
                {
                    pointContainingPane = ViewportPaneType.HorizontalScrollable;
                }
            }

            if (this[pointContainingPane].IsEmpty)
            {
                pointContainingPane = ViewportPaneType.Scrollable;
            }

            return this[pointContainingPane];
        }

        /// <summary>
        /// Gets the document point from view point.
        /// </summary>
        /// <param name="point">The view point point.</param>
        /// <returns>The document point.</returns>
        public Point GetDocumentPointFromViewPoint(Point point)
        {
            point = new Point(point.X / this.scaleFactor.Width, point.Y / this.scaleFactor.Height);

            ViewportPane pointContainingPane = this.GetViewportPaneFromViewportPoint(point);
            ViewportPaneType type = pointContainingPane.ViewportPaneType;

            double horizontalOffset = (type == ViewportPaneType.HorizontalScrollable || type == ViewportPaneType.Scrollable) ? this[ViewportPaneType.VerticalScrollable].Rect.Width : 0;
            double verticalOffset = (type == ViewportPaneType.VerticalScrollable || type == ViewportPaneType.Scrollable) ? this[ViewportPaneType.HorizontalScrollable].Rect.Height : 0;

            Point scrollPosition = new Point(
                (point.X - horizontalOffset + pointContainingPane.Rect.Left),
                (point.Y - verticalOffset + pointContainingPane.Rect.Top));

            return scrollPosition;
        }

        /// <summary>
        /// Gets the view point from document point.
        /// </summary>
        /// <param name="point">The document point.</param>
        /// <returns>The view point.</returns>
        public Point GetViewPointFromDocumentPoint(Point point)
        {
            ViewportPane pointContainingPane = this.GetViewportPaneFromDocumentPoint(point);
            ViewportPaneType type = pointContainingPane.ViewportPaneType;

            ViewportPane fixedPane = this[ViewportPaneType.Fixed];
            ViewportPane horizontalScrollablePane = this[ViewportPaneType.HorizontalScrollable];
            ViewportPane verticalScrollablePane = this[ViewportPaneType.VerticalScrollable];
            ViewportPane scrollablePane = this[ViewportPaneType.Scrollable];

            double x = 0;
            double y = 0;

            switch (type)
            {
                case ViewportPaneType.Fixed:
                    x = point.X - fixedPane.Rect.Left;
                    y = point.Y - fixedPane.Rect.Top;
                    break;
                case ViewportPaneType.HorizontalScrollable:
                    x = point.X - horizontalScrollablePane.Rect.Left + fixedPane.Rect.Width;
                    y = point.Y - horizontalScrollablePane.Rect.Top;
                    break;
                case ViewportPaneType.VerticalScrollable:
                    x = point.X - verticalScrollablePane.Rect.Left;
                    y = point.Y - verticalScrollablePane.Rect.Top + fixedPane.Rect.Height;
                    break;
                case ViewportPaneType.Scrollable:
                    x = point.X - scrollablePane.Rect.Left + verticalScrollablePane.Rect.Width;
                    y = point.Y - scrollablePane.Rect.Top + horizontalScrollablePane.Rect.Height;
                    break;
            }

            return new Point(x * scaleFactor.Width, y * scaleFactor.Height);
        }

        /// <summary>
        /// Translates the specified point.
        /// </summary>
        /// <param name="point">The point.</param>
        /// <param name="pointContainingPaneType">Type of the pane that contains the point.</param>
        /// <returns>The translated point.</returns>
        public Point Translate(Point point, ViewportPaneType pointContainingPaneType)
        {
            ViewportPane pointContainingPane = this[pointContainingPaneType];

            double newX = 0;
            double newY = 0;

            if (pointContainingPaneType == ViewportPaneType.Fixed || pointContainingPaneType == ViewportPaneType.VerticalScrollable)
            {
                newX = point.X - pointContainingPane.Rect.Left;
            }
            else
            {
                newX = point.X - pointContainingPane.Rect.Left + this[ViewportPaneType.VerticalScrollable].Rect.Width;
            }

            if (pointContainingPaneType == ViewportPaneType.Fixed || pointContainingPaneType == ViewportPaneType.HorizontalScrollable)
            {
                newY = point.Y - pointContainingPane.Rect.Top;
            }
            else
            {
                newY = point.Y - pointContainingPane.Rect.Top + this[ViewportPaneType.HorizontalScrollable].Rect.Height;
            }

            return new Point(newX, newY);
        }

        /// <summary>
        /// Determines whether [contains] [the specified bounding rectangle].
        /// </summary>
        /// <param name="boundingRectangle">The bounding rectangle.</param>
        /// <returns></returns>
        public bool Contains(Rect boundingRectangle)
        {
            foreach (var pane in this.ViewportPanes)
            {
                if (pane.Rect.Contains(boundingRectangle))
                {
                    return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Determines whether the SheetViewport contains the specified point.
        /// </summary>
        /// <param name="point">The point.</param>
        /// <returns>True if contains the point.</returns>
        public bool Contains(Point point)
        {
            foreach (var pane in this.ViewportPanes)
            {
                if (pane.Rect.Contains(point))
                {
                    return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Gets the top left point.
        /// </summary>
        /// <returns></returns>
        public Point GetTopLeftPoint()
        {
            ViewportPane topLeftPane = this.GetTopLeftPane();

            return new Point(topLeftPane.Rect.Left, topLeftPane.Rect.Top);
        }

        #endregion


        #region VisibleRanges Methods

        /// <summary>
        /// Determines whetherthe SheetViewport contains the specified column index.
        /// </summary>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>True if contains the column index.</returns>
        public bool ContainsColumnIndex(int columnIndex)
        {
            for (int i = 0; i < this.VisibleRanges.Length; i++)
            {
                CellRange range = this.VisibleRanges[i];
                if (range.FromIndex.ColumnIndex <= columnIndex && range.ToIndex.ColumnIndex >= columnIndex)
                {
                    return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Determines whether the SheetViewport contains the specified row index.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <returns>True if contains the row index.</returns>
        public bool ContainsRowIndex(int rowIndex)
        {
            for (int i = 0; i < this.VisibleRanges.Length; i++)
            {
                CellRange range = this.VisibleRanges[i];
                if (range.FromIndex.RowIndex <= rowIndex && range.ToIndex.RowIndex >= rowIndex)
                {
                    return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Determines whether the SheetViewport contains the specified cell index.
        /// </summary>
        /// <param name="cellIndex">Index of the cell.</param>
        /// <returns>True if contains the cell index.</returns>
        public bool Contains(CellIndex cellIndex)
        {
            Guard.ThrowExceptionIfNull(cellIndex, "cellIndex");

            return this.Contains(cellIndex.RowIndex, cellIndex.ColumnIndex);
        }

        internal bool Contains(long index)
        {
            int rowIndex;
            int columnIndex;

            CellsPropertyBag.ConvertLongToRowAndColumnIndexes(index, out rowIndex, out columnIndex);
            return this.Contains(rowIndex, columnIndex);
        }

        /// <summary>
        /// Determines whether the sheet viewport contains the specified row and column index.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns></returns>
        public bool Contains(int rowIndex, int columnIndex)
        {
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            for (int i = 0; i < this.VisibleRanges.Length; i++)
            {
                if (this.VisibleRanges[i].Contains(rowIndex, columnIndex))
                {
                    return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Gets the cell range containing cell index.
        /// </summary>
        /// <param name="cellIndex">Index of the cell.</param>
        /// <returns>The cell range.</returns>
        public CellRange GetCellRangeContainingCellIndex(CellIndex cellIndex)
        {
            Guard.ThrowExceptionIfNull(cellIndex, "cellIndex");

            return this.GetCellRangeContainingCellIndex(cellIndex.RowIndex, cellIndex.ColumnIndex);
        }

        /// <summary>
        /// Gets the index of the cell range containing cell.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns></returns>
        public CellRange GetCellRangeContainingCellIndex(int rowIndex, int columnIndex)
        {
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            return this.GetViewportPaneContainingCellIndex(rowIndex, columnIndex).VisibleRange;
        }

        /// <summary>
        /// Gets the index of the viewport pane containing cell.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <returns>The viewport pane containing cell.</returns>
        public ViewportPane GetViewportPaneContainingCellIndex(CellIndex index)
        {
            return this.GetViewportPaneContainingCellIndex(index.RowIndex, index.ColumnIndex);
        }

        internal ViewportPane GetViewportPaneContainingCellIndex(int rowIndex, int columnIndex)
        {
            ViewportPane fixedPane = this[ViewportPaneType.Fixed];
            ViewportPane horizontalScrollablePane = this[ViewportPaneType.HorizontalScrollable];
            ViewportPane verticalScrollablePane = this[ViewportPaneType.VerticalScrollable];
            ViewportPane scrollablePane = this[ViewportPaneType.Scrollable];

            if ((rowIndex <= fixedPane.VisibleRange.ToIndex.RowIndex && !fixedPane.IsEmpty) ||
                (rowIndex <= horizontalScrollablePane.VisibleRange.ToIndex.RowIndex && !horizontalScrollablePane.IsEmpty))
            {
                if ((columnIndex <= fixedPane.VisibleRange.ToIndex.ColumnIndex && !fixedPane.IsEmpty) ||
                    (columnIndex <= verticalScrollablePane.VisibleRange.ToIndex.ColumnIndex && !verticalScrollablePane.IsEmpty))
                {
                    return fixedPane;
                }
                else
                {
                    return horizontalScrollablePane;
                }
            }
            else
            {
                if ((columnIndex <= fixedPane.VisibleRange.ToIndex.ColumnIndex && !fixedPane.IsEmpty) ||
                    (columnIndex <= verticalScrollablePane.VisibleRange.ToIndex.ColumnIndex && !verticalScrollablePane.IsEmpty))
                {
                    return verticalScrollablePane;
                }
                else
                {
                    return scrollablePane;
                }
            }
        }

        /// <summary>
        /// Gets the top most viewport pane containing column index.
        /// </summary>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The viewport pane containing column index.</returns>
        public ViewportPane GetTopMostViewportPaneContainingColumnIndex(int columnIndex)
        {
            IEnumerable<ViewportPane> viewportPanes = this.ViewportPanes.Where(p =>
                p.VisibleRange.FromIndex.ColumnIndex <= columnIndex &&
                p.VisibleRange.ToIndex.ColumnIndex >= columnIndex).OrderBy(p => p.VisibleRange.FromIndex.RowIndex);

            return viewportPanes.FirstOrDefault();
        }

        /// <summary>
        /// Gets the left most viewport pane containing row index.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <returns>The viewport pane containing row index.</returns>
        public ViewportPane GetLeftMostViewportPaneContainingRowIndex(int rowIndex)
        {
            IEnumerable<ViewportPane> viewportPanes = this.ViewportPanes.Where(p =>
                p.VisibleRange.FromIndex.RowIndex <= rowIndex &&
                p.VisibleRange.ToIndex.RowIndex >= rowIndex).OrderBy(p => p.VisibleRange.FromIndex.ColumnIndex);

            return viewportPanes.FirstOrDefault();
        }

        internal ViewportPane GetTopLeftPane()
        {
            ViewportPane topLeftViewportPane = this[ViewportPaneType.Fixed];
            if (topLeftViewportPane.IsEmpty)
            {
                topLeftViewportPane = this[ViewportPaneType.HorizontalScrollable];
            }
            if (topLeftViewportPane.IsEmpty)
            {
                topLeftViewportPane = this[ViewportPaneType.VerticalScrollable];
            }
            if (topLeftViewportPane.IsEmpty)
            {
                topLeftViewportPane = this[ViewportPaneType.Scrollable];
            }

            return topLeftViewportPane;
        }

        #endregion


        #region Method Overrides

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The object to compare with the current object.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            SheetViewport other = obj as SheetViewport;

            if (other == null)
            {
                return false;
            }

            if (this.frozenPoint != other.frozenPoint)
            {
                return false;
            }

            foreach (ViewportPaneType pane in PaneTypes)
            {
                if (!this[pane].Equals(other[pane]))
                {
                    return false;
                }
            }

            return true;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            ViewportPaneType firstPane = PaneTypes[0];

            int hashCode = this[firstPane].GetHashCodeOrZero();

            for (int i = 1; i < PaneTypes.Count; i++)
            {
                ViewportPaneType currentPane = PaneTypes[i];
                hashCode = TelerikHelper.CombineHashCodes(hashCode, this[currentPane].GetHashCodeOrZero());
            }

            return hashCode;
        }

        /// <summary>
        /// Implements the equals operator.
        /// </summary>
        /// <param name="first">The first value.</param>
        /// <param name="second">The second value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
        public static bool operator ==(SheetViewport first, SheetViewport second)
        {
            return TelerikHelper.EqualsOfT<SheetViewport>(first, second);
        }

        /// <summary>
        /// Implements the less operator.
        /// </summary>
        /// <param name="first">The first value.</param>
        /// <param name="second">The second value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
        public static bool operator !=(SheetViewport first, SheetViewport second)
        {
            return !TelerikHelper.EqualsOfT<SheetViewport>(first, second);
        }

        #endregion
    }
}