﻿using System;
using System.Linq;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.Layout
{
    internal static class SheetViewportFactory
    {
        internal static SheetViewport CreateSheetViewport(RadWorksheetLayout worksheetLayout, SheetViewport oldSheetViewport, Size scaleFactor, Size finalSize, SizeI visibleSize, CellIndex frozenCellIndex,
            Point offset, bool horizontalScrollModeIsItemBased, bool verticalScrollModeIsItemBased)
        {
            ViewportPane fixedPane = oldSheetViewport[ViewportPaneType.Fixed];

            Point fixedTopLeft = new Point(fixedPane.Rect.Left, fixedPane.Rect.Top);
            Point scrollableTopLeft = offset;

            if (horizontalScrollModeIsItemBased || verticalScrollModeIsItemBased)
            {
                int scrollableRowIndex;
                int scrollableColumnIndex;

                worksheetLayout.GetCellIndexFromPoint(scrollableTopLeft.X, scrollableTopLeft.Y, false, false, out scrollableRowIndex, out scrollableColumnIndex, true, true);
                scrollableRowIndex = Math.Max(scrollableRowIndex, frozenCellIndex.RowIndex);
                scrollableColumnIndex = Math.Max(scrollableColumnIndex, frozenCellIndex.ColumnIndex);

                CellLayoutBox scrollableTopLeftCellBox = worksheetLayout.GetCellLayoutBox(scrollableRowIndex, scrollableColumnIndex);

                if (horizontalScrollModeIsItemBased)
                {
                    scrollableTopLeft = new Point(scrollableTopLeftCellBox.BoundingRectangle.Left, scrollableTopLeft.Y);
                }

                if (verticalScrollModeIsItemBased)
                {
                    scrollableTopLeft = new Point(scrollableTopLeft.X, scrollableTopLeftCellBox.BoundingRectangle.Top);
                }
            }

            Point maxBottomRight = worksheetLayout.GetMaxBottomRight(visibleSize);

            SheetViewport viewport = new SheetViewport(worksheetLayout,
                finalSize, scaleFactor,
                fixedTopLeft, scrollableTopLeft, frozenCellIndex, maxBottomRight);

            return viewport;
        }

        internal static SheetViewport CreateSheetViewport(RadWorksheetLayout worksheetLayout, Size displaySize, Point offset, Size scaleFactor)
        {
            SizeI visibleSize = SpreadsheetDefaultValues.VisibleSize;
            CellIndex frozenCellIndex = new CellIndex(0, 0);
            Point fixedTopLeft = new Point(0, 0);
            Point scrollableTopLeft = offset;
            var maxBottomRight = worksheetLayout.GetMaxBottomRight(visibleSize);

            return new SheetViewport(worksheetLayout, displaySize, scaleFactor, fixedTopLeft, scrollableTopLeft, frozenCellIndex, maxBottomRight);
        }

        internal static SheetViewport CreateSheetViewport(RadWorksheetLayout worksheetLayout, Size finalSize, SizeI visibleSize,
            WorksheetViewState viewstate)
        {
            double fixedTop = worksheetLayout.GetRowTop(viewstate.TopLeftCellIndex.RowIndex);
            double fixedLeft = worksheetLayout.GetColumnLeft(viewstate.TopLeftCellIndex.ColumnIndex);
            Point fixedTopLeft = new Point(fixedLeft, fixedTop);

            double scrollableTop = worksheetLayout.GetRowTop(viewstate.Pane.TopLeftCellIndex.RowIndex);
            double scrollableLeft = worksheetLayout.GetColumnLeft(viewstate.Pane.TopLeftCellIndex.ColumnIndex);
            Point scrollableTopLeft = new Point(scrollableLeft, scrollableTop);

            CellIndex frozenCellIndex = viewstate.GetFrozenCellIndex();
            Point maxBottomRight = worksheetLayout.GetMaxBottomRight(visibleSize);

            SheetViewport viewport = new SheetViewport(worksheetLayout,
                finalSize, viewstate.ScaleFactor,
                fixedTopLeft, scrollableTopLeft, frozenCellIndex, maxBottomRight);

            return viewport;
        }

        internal static bool GetFrozenViewport(RadWorksheetLayout worksheetLayout, SheetViewport oldSheetViewport, SizeI visibleSize, CellIndex frozenCellIndex, out SheetViewport result)
        {
            result = oldSheetViewport;

            CellIndex topLeftFixedCellIndex = oldSheetViewport[ViewportPaneType.Scrollable].VisibleRange.FromIndex;
            frozenCellIndex = ValidateFrozenCellIndex(oldSheetViewport, frozenCellIndex);

            int frozenRowIndex = frozenCellIndex.RowIndex;
            int frozenColumnIndex = frozenCellIndex.ColumnIndex;

            if (frozenCellIndex.RowIndex == topLeftFixedCellIndex.RowIndex)
            {
                frozenRowIndex = 0;
            }

            if (frozenCellIndex.ColumnIndex == topLeftFixedCellIndex.ColumnIndex)
            {
                frozenColumnIndex = 0;
            }

            if (frozenCellIndex.RowIndex == 0 && frozenCellIndex.ColumnIndex == 0)
            {
                return false;
            }

            frozenCellIndex = new CellIndex(frozenRowIndex, frozenColumnIndex);

            Rect scrollableRect = oldSheetViewport[ViewportPaneType.Scrollable].Rect;
            Point fixedTopLeft = new Point(scrollableRect.Left, scrollableRect.Top);

            Rect scrollableTopLeftCellBoundingRectangle = worksheetLayout.GetCellLayoutBox(frozenCellIndex).BoundingRectangle;
            Point scrollableTopLeft = new Point(scrollableTopLeftCellBoundingRectangle.Left, scrollableTopLeftCellBoundingRectangle.Top);
            Point maxBottomRight = worksheetLayout.GetMaxBottomRight(visibleSize);

            result = new SheetViewport(worksheetLayout, oldSheetViewport.DisplaySize, oldSheetViewport.ScaleFactor, fixedTopLeft, scrollableTopLeft, frozenCellIndex, maxBottomRight);

            return true;
        }

        internal static SheetViewport GetUnfrozenViewport(RadWorksheetLayout worksheetLayout, SheetViewport oldSheetViewport)
        {
            Point fixedTopLeft = oldSheetViewport.GetTopLeftPoint();
            return CreateSheetViewport(worksheetLayout, oldSheetViewport.DisplaySize, fixedTopLeft, oldSheetViewport.ScaleFactor);
        }

        private static CellIndex ValidateFrozenCellIndex(SheetViewport sheetViewport, CellIndex cellIndex)
        {
            CellIndex result = cellIndex;

            if ((sheetViewport.Contains(cellIndex) && sheetViewport[ViewportPaneType.Scrollable].VisibleRange.FromIndex != cellIndex) ||
                (sheetViewport.ContainsRowIndex(cellIndex.RowIndex) && cellIndex.ColumnIndex == 0) ||
                (sheetViewport.ContainsColumnIndex(cellIndex.ColumnIndex) && cellIndex.RowIndex == 0))
            {
                return result;
            }

            if ((!sheetViewport.Contains(cellIndex) ||
                sheetViewport[ViewportPaneType.Scrollable].VisibleRange.FromIndex == cellIndex))
            {
                CellRange fixedVisibleRange = sheetViewport[ViewportPaneType.Fixed].VisibleRange;
                CellRange scrollableVisibleRange = sheetViewport[ViewportPaneType.Scrollable].VisibleRange;

                CellIndex fromIndex = (sheetViewport[ViewportPaneType.Fixed].IsEmpty) ? scrollableVisibleRange.FromIndex : fixedVisibleRange.FromIndex;
                int rowCount = fixedVisibleRange.RowCount + scrollableVisibleRange.RowCount;
                int columnCount = fixedVisibleRange.ColumnCount + scrollableVisibleRange.ColumnCount;

                result = new CellIndex(
                    fromIndex.RowIndex + rowCount / 2,
                    fromIndex.ColumnIndex + columnCount / 2);
            }

            return result;
        }
    }
}