﻿using System;
using System.Linq;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Layout
{
    /// <summary>
    /// Represents the viewport pane class.
    /// </summary>
    public class ViewportPane
    {
        #region Fields

        private Rect rect;
        private CellRange visibleRange;
        private readonly ViewportPaneType viewportPaneType;

        #endregion


        #region Properties

        /// <summary>
        /// Gets or sets the rect.
        /// </summary>
        /// <value>The rect.</value>
        public Rect Rect
        {
            get
            {
                return this.rect;
            }
            internal set
            {
                if (this.rect != value)
                {
                    this.rect = value;
                }
            }
        }

        /// <summary>
        /// Gets or sets the bounding rect.
        /// </summary>
        /// <value>The bounding rect.</value>
        public Rect BoundingRect { get; internal set; }

        /// <summary>
        /// Gets or sets the visible range.
        /// </summary>
        /// <value>The visible range.</value>
        public CellRange VisibleRange
        {
            get
            {
                if (this.visibleRange == null)
                {
                    this.visibleRange = CellRange.Empty;
                }

                return this.visibleRange;
            }
            internal set
            {
                if (this.visibleRange != value)
                {
                    this.visibleRange = value;
                }
            }
        }

        /// <summary>
        /// Gets the type of the viewport pane.
        /// </summary>
        /// <value>The type of the viewport pane.</value>
        public ViewportPaneType ViewportPaneType
        {
            get
            {
                return this.viewportPaneType;
            }
        }

        /// <summary>
        /// Gets the is empty.
        /// </summary>
        /// <value>The is empty.</value>
        public bool IsEmpty
        {
            get
            {
                return this.rect.Width == 0 || this.rect.Height == 0;
            }
        }

        internal Point TopLeftPoint { get; set; }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="ViewportPane" /> class.
        /// </summary>
        /// <param name="viewportPaneType">Type of the viewport pane.</param>
        public ViewportPane(ViewportPaneType viewportPaneType)
        {
            this.viewportPaneType = viewportPaneType;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ViewportPane" /> class.
        /// </summary>
        /// <param name="rect">The rect.</param>
        /// <param name="viewportPaneType">Type of the viewport pane.</param>
        public ViewportPane(Rect rect, ViewportPaneType viewportPaneType)
            : this(viewportPaneType)
        {
            this.rect = rect;
        }

        #endregion


        #region Methods

        internal void SetX(double newX)
        {
            this.rect = new Rect(newX, this.rect.Y, this.rect.Width, this.rect.Height);
        }

        internal void SetY(double newY)
        {
            this.rect = new Rect(this.rect.X, newY, this.rect.Width, this.rect.Height);
        }

        #endregion


        #region Method Overrides

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            ViewportPane other = obj as ViewportPane;

            if (other == null)
            {
                return false;
            }

            return this.ViewportPaneType.Equals(other.ViewportPaneType) &&
                this.Rect.Equals(other.Rect) &&
                this.VisibleRange.Equals(other.VisibleRange);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                this.Rect.GetHashCodeOrZero(),
                this.ViewportPaneType.GetHashCodeOrZero());
        }

        #endregion
    }
}