﻿using System;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Layout.Layers;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Printing;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Layout
{
    internal class WorksheetPageLayoutBox : LayoutBox
    {
        private HeaderFooterRenderContext headerFooterContext;

        public WorksheetPageLayoutBox(Worksheet worksheet, CellRange range, Rect viewportRectangle, Size scaleFactor)
            : base(viewportRectangle)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfNull(range, "range");

            this.Worksheet = worksheet;
            this.Range = range;
            this.RowColumnHeadingsSize = CalculateRowColumnHeadings(worksheet, range, scaleFactor);
            this.PageContentBox = CalculatePageContentBox(this.Worksheet.WorksheetPageSetup, this.RowColumnHeadingsSize, this.Size);
            this.ScaleFactor = scaleFactor;
        }

        internal WorksheetPageLayoutBox(WorksheetPageLayoutBox other)
            : base(new Rect(other.Left, other.Top, other.Width, other.Height))
        {
            this.Worksheet = other.Worksheet;
            this.Range = other.Range;
            this.RowColumnHeadingsSize = other.RowColumnHeadingsSize;
            this.PageContentBox = other.PageContentBox;
            this.ScaleFactor = other.ScaleFactor;
        }

        public Size RowColumnHeadingsSize { get; private set; }
        public Worksheet Worksheet { get; private set; }
        public CellRange Range { get; private set; }
        public Rect PageContentBox { get; private set; }
        public Size ScaleFactor { get; private set; }

        public HeaderFooterRenderContext HeaderFooterContext
        {
            get
            {
                return this.headerFooterContext;
            }
        }

        public void PrepareHeaderFooterContext(int pageNumber, int numberOfPages, DateTime dateAndTime)
        {
            Guard.ThrowExceptionIfNotNull(this.headerFooterContext, "headerFooterContext");

            this.headerFooterContext = new HeaderFooterRenderContext(this.Worksheet, dateAndTime, pageNumber, numberOfPages);
        }

        private static Rect CalculatePageContentBox(WorksheetPageSetup pageSetup, Size rowColumnHeadings, Size viewportSize)
        {
            double contentWidth = viewportSize.Width + rowColumnHeadings.Width;
            double contentHeight = viewportSize.Height + rowColumnHeadings.Height;

            PageMargins margins = pageSetup.Margins;
            Size pageSize = pageSetup.RotatedPageSize;

            double centeringLeftOffset = pageSetup.CenterHorizontally ? ((pageSize.Width - margins.Left - margins.Right - contentWidth) / 2) : 0;
            double centeringTopOffset = pageSetup.CenterVertically ? ((pageSize.Height - margins.Top - margins.Bottom - contentHeight) / 2) : 0;

            double left = margins.Left + centeringLeftOffset;
            double top = margins.Top + centeringTopOffset;

            return new Rect(left, top, contentWidth, contentHeight);
        }

        private static Size CalculateRowColumnHeadings(Worksheet worksheet, CellRange range, Size scaleFactor)
        {
            if (!worksheet.WorksheetPageSetup.PrintOptions.PrintRowColumnHeadings)
            {
                return new Size();
            }

            RadWorksheetLayout layout = worksheet.Workbook.GetWorksheetLayout(worksheet, true);

            double width = 0;
            for (int i = range.FromIndex.RowIndex; i <= range.ToIndex.RowIndex; i++)
            {
                width = Math.Max(width, layout.GetRowHeadingSize(range, i).Width * scaleFactor.Width);
            }

            double height = 0;
            for (int i = range.FromIndex.ColumnIndex; i <= range.ToIndex.ColumnIndex; i++)
            {
                height = Math.Max(height, layout.GetColumnHeadingSize(range.ToIndex.ColumnIndex).Height * scaleFactor.Height);
            }

            return new Size(width, height);
        }

        public override bool Equals(object obj)
        {
            WorksheetPageLayoutBox other = obj as WorksheetPageLayoutBox;

            if (other == null)
            {
                return false;
            }

            return
                this.RowColumnHeadingsSize.Equals(other.RowColumnHeadingsSize) &&
                this.BoundingRectangle.Equals(other.BoundingRectangle) &&
                this.Worksheet.Name.Equals(other.Worksheet.Name) &&
                this.Range.Equals(other.Range) &&
                this.HeaderFooterContext.DateTime.Equals(other.HeaderFooterContext.DateTime) &&
                this.HeaderFooterContext.NumberOfPages.Equals(other.HeaderFooterContext.NumberOfPages) &&
                this.HeaderFooterContext.PageNumber.Equals(other.HeaderFooterContext.PageNumber);
        }

        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                this.RowColumnHeadingsSize.GetHashCode(),
                this.BoundingRectangle.GetHashCode(),
                this.Worksheet.Name.GetHashCode(),
                this.Range.GetHashCode(),
                this.HeaderFooterContext.DateTime.GetHashCode(),
                this.HeaderFooterContext.NumberOfPages.GetHashCode(),
                this.HeaderFooterContext.PageNumber.GetHashCode());
        }
    }
}
