﻿using System;
using System.Windows;
using System.Windows.Media;

namespace Telerik.Windows.Documents.Spreadsheet.Maths
{
    // TODO: Use AlgebraExtensions from Telerik.Windows.Documents.Core instead this class!
    internal static class AlgebraExtensions
    {
        /// <summary>
        /// Gauss-Jordan elimination method to solve Ax = b for x.
        /// </summary>
        /// <param name="A">A real-valued matrix.</param>
        /// <param name="b">The real-valued vector.</param>
        /// <param name="MaxIterations">The maximum amount of iterations.</param>
        /// <param name="tolerance">The tolerance.</param>
        public static RVector GaussJacobi(RMatrix A, RVector b, int MaxIterations = 10, double tolerance = 1E-4)
        {
            var bSize = b.GetVectorSize;
            var x = new RVector(bSize);
            for (var nIteration = 0; nIteration < MaxIterations; nIteration++)
            {
                var xOld = x.Clone();
                for (var i = 0; i < bSize; i++)
                {
                    var entry = b[i];
                    var diagonal = A[i, i];
                    if (System.Math.Abs(diagonal) < Constants.Epsilon) throw new Exception("Diagonal element went into epsilon.");
                    for (var j = 0; j < bSize; j++) if (j != i) entry -= A[i, j] * xOld[j];
                    x[i] = entry / diagonal;
                }
                var dx = x - xOld;
                if (dx.GetNorm() < tolerance) return x;
            }
            return x;
        }

        /// <summary>
        /// Gauss-Jordan elimination method solving the Ax=b inear set of equations.
        /// </summary>
        /// <param name="A">A real-valued matrix.</param>
        /// <param name="b">A real-valued vector.</param>
        ///<remarks>http://en.wikipedia.org/wiki/Gauss-Jordan</remarks>
        public static RVector GaussJordan(RMatrix A, RVector b)
        {
            Triangulate(A, b);
            var bSize = b.GetVectorSize;
            var x = new RVector(bSize);
            for (var i = bSize - 1; i >= 0; i--)
            {
                var aii = A[i, i];
                if (System.Math.Abs(aii) < Constants.Epsilon) throw new Exception("Diagonal element is too small!");
                x[i] = (b[i] - RVector.DotProduct(A.GetRowVector(i), x)) / aii;
            }
            return x;
        }

        public static RVector GaussSeidel(RMatrix A, RVector b, int maxIterations = 10, double tolerance = 1E-4)
        {
            var size = b.GetVectorSize;
            var x = new RVector(size);
            for (var nIteration = 0; nIteration < maxIterations; nIteration++)
            {
                var xOld = x.Clone();
                for (var i = 0; i < size; i++)
                {
                    var entry = b[i];
                    var diagonal = A[i, i];
                    if (System.Math.Abs(diagonal) < Constants.Epsilon) throw new Exception("Diagonal element is too small!");
                    for (var j = 0; j < i; j++) entry -= A[i, j] * x[j];
                    for (var j = i + 1; j < size; j++) entry -= A[i, j] * xOld[j];
                    x[i] = entry / diagonal;
                }
                var dx = x - xOld;
                if (dx.GetNorm() < tolerance) return x;
            }
            return x;
        }

        /// <summary>
        /// Triangulation of matrix A using LU decomposition on the basis of the Crout algorithm with pivoting.
        /// </summary>
        /// <param name="A">The real-valued matrix.</param>
        /// <param name="b">The real-valued vector.</param>
        /// <returns></returns>
        public static double LUCrout(RMatrix A, RVector b)
        {
            LUDecompose(A);
            return LUSubstitute(A, b);
        }

        /// <summary>
        /// Returns the LU Matrix Inverse.
        /// </summary>
        /// <param name="matrix"></param>
        /// <returns></returns>
        public static RMatrix LUInverse(RMatrix matrix)
        {
            var nRows = matrix.GetnRows;
            var u = matrix.IdentityMatrix();
            LUDecompose(matrix);
            var uv = new RVector(nRows);
            for (var i = 0; i < nRows; i++)
            {
                uv = u.GetRowVector(i);
                LUSubstitute(matrix, uv);
                u.ReplaceRow(uv, i);
            }
            var inverse = u.GetTranspose();
            return inverse;
        }

        /// <summary>
        /// Returns the LU decomposition of the given matrix.
        /// </summary>
        /// <param name="matrix">A matrix.</param>
        /// <remarks>http://en.wikipedia.org/wiki/LU_decomposition</remarks>
        public static RMatrix LUDecompose(RMatrix matrix)
        {
            var nRows = matrix.GetnRows;
            for (var i = 0; i < nRows; i++)
            {
                for (var j = 0; j < nRows; j++)
                {
                    var w = matrix[i, j];
                    for (var k = 0; k < System.Math.Min(i, j); k++) w -= matrix[i, k] * matrix[k, j];
                    if (j > i)
                    {
                        var s = matrix[i, i];
                        if (System.Math.Abs(w) < Constants.Epsilon) throw new Exception("Diagonal element is too small!");
                        w /= s;
                    }
                    matrix[i, j] = w;
                }
            }
            return matrix;
        }

        public static Point MultiplyBy(this Point a, double number)
        {
            return new Point(a.X * number, a.Y * number);
        }

        public static double MultiplyBy(this Point a, Point other)
        {
            return a.X * other.X + a.Y * other.Y;
        }

        public static double Length(this Point a)
        {
            return Math.Sqrt(a.MultiplyBy(a));
        }

        public static bool IsZero(this Point a)
        {
            return a.Length() < Constants.Epsilon;
        }

        public static Point UnitVector(this Point a)
        {
            if (a.IsZero())
            {
                throw new ArgumentException("Cannot calculate unit vector of a point with zero length!");
            }

            return a.MultiplyBy(1 / a.Length());
        }

        public static Matrix MultiplyBy(this Matrix m1, Matrix m2)
        {
            return new Matrix(
                (m1.M11 * m2.M11) + (m1.M12 * m2.M21), (m1.M11 * m2.M12) + (m1.M12 * m2.M22),
                (m1.M21 * m2.M11) + (m1.M22 * m2.M21), (m1.M21 * m2.M12) + (m1.M22 * m2.M22),
                (m1.OffsetX * m2.M11) + (m1.OffsetY * m2.M21) + m2.OffsetX, (m1.OffsetX * m2.M12) + (m1.OffsetY * m2.M22) + m2.OffsetY);
        }

        public static Matrix TranslateMatrix(this Matrix m, double deltaX, double deltaY)
        {
            return m.MultiplyBy(new Matrix(1, 0, 0, 1, deltaX, deltaY));
        }

        public static Matrix ScaleMatrix(this Matrix m, double scaleX, double scaleY)
        {
            return m.ScaleMatrixAt(scaleX, scaleY, 0, 0);
        }

        public static Matrix ScaleMatrixAt(this Matrix m, double scaleX, double scaleY, double centerX, double centerY)
        {
            return m.MultiplyBy(new Matrix(scaleX, 0, 0, scaleY, 0, 0).GetTransformationAt(centerX, centerY));
        }

        public static Matrix RotateMatrix(this Matrix m, double angleInDegrees)
        {
            return m.RotateMatrixAt(angleInDegrees, 0, 0);
        }

        public static Matrix RotateMatrixAt(this Matrix m, double angleInDegrees, double centerX, double centerY)
        {
            double radians = angleInDegrees * Math.PI / 180;
            double sin = Math.Sin(radians);
            double cos = Math.Cos(radians);

            return m.MultiplyBy(new Matrix(cos, sin, -sin, cos, 0, 0).GetTransformationAt(centerX, centerY));
        }

        public static Matrix InverseMatrix(this Matrix m)
        {
            ////just basic algebra
            var det = (m.M11 * m.M22) - (m.M12 * m.M21);
            var det1 = 1 / det;
            return new Matrix(m.M22 * det1, -m.M12 * det1, -m.M21 * det1, m.M11 * det1, ((m.M21 * m.OffsetY) - (m.OffsetX * m.M22)) * det1, ((m.OffsetX * m.M12) - (m.M11 * m.OffsetY)) * det1);
        }

        private static Matrix GetTransformationAt(this Matrix zeroCenteredTransform, double centerX, double centerY)
        {
            double offsetX = zeroCenteredTransform.OffsetX + (1 - zeroCenteredTransform.M11) * centerX - zeroCenteredTransform.M21 * centerY;
            double offsetY = zeroCenteredTransform.OffsetY + (1 - zeroCenteredTransform.M22) * centerY - zeroCenteredTransform.M12 * centerX;

            return new Matrix(zeroCenteredTransform.M11, zeroCenteredTransform.M12, zeroCenteredTransform.M21, zeroCenteredTransform.M22, offsetX, offsetY);
        }

        private static double LUSubstitute(RMatrix matrix, RVector vec)
        {
            var size = vec.GetVectorSize;
            var det = 1.0;
            for (var i = 0; i < size; i++)
            {
                var w = vec[i];
                for (var j = 0; j < i; j++) w -= matrix[i, j] * vec[j];
                var p = matrix[i, i];
                if (System.Math.Abs(w) < Constants.Epsilon) throw new Exception("Diagonal element is too small!");
                w /= p;
                vec[i] = w;
                det *= matrix[i, i];
            }
            for (var i = size - 1; i >= 0; i--)
            {
                var s = vec[i];
                for (var j = i + 1; j < size; j++) s -= matrix[i, j] * vec[j];
                vec[i] = s;
            }
            return det;
        }

        private static void Triangulate(RMatrix A, RVector b)
        {
            var nRows = A.GetnRows;
            for (var i = 0; i < nRows - 1; i++)
            {
                var diagonalElement = pivotGaussJordan(A, b, i);
                if (System.Math.Abs(diagonalElement) < Constants.Epsilon) throw new Exception("Diagonal element is too small!");
                for (var j = i + 1; j < nRows; j++)
                {
                    var w = A[j, i] / diagonalElement;
                    for (var k = i + 1; k < nRows; k++) A[j, k] -= w * A[i, k];
                    b[j] -= w * b[i];
                }
            }
        }

        private static double pivotGaussJordan(RMatrix A, RVector b, int q)
        {
            var bSize = b.GetVectorSize;
            var c = q;
            var d = 0.0;
            for (var j = q; j < bSize; j++)
            {
                var w = System.Math.Abs(A[j, q]);
                if (w > d)
                {
                    d = w;
                    c = j;
                }
            }
            if (c > q)
            {
                A.SwapRMatrixow(q, c);
                b.SwapVectorEntries(q, c);
            }
            return A[q, q];
        }
    }
}