﻿using System;
using System.Linq;
using System.Globalization;
using System.Text.RegularExpressions;
using System.Collections.Generic;
using System.Numerics;
namespace Telerik.Windows.Documents.Spreadsheet.Maths
{
    /// <summary>
    /// Extension methods 
    /// </summary>
    internal static class ComplexExtensions
    {
        /// <summary>
        /// Return whether the <see cref="Complex"/> number is zero (in an <see cref="Constants.Epsilon"/> sense).
        /// </summary>
        public static bool IsZero(this Complex complex)
        {
            return complex.Real.IsZero() && complex.Imaginary.IsZero();
        }
        /// <summary>
        /// Calculates the modulus of the complex number in polar form.
        /// </summary>
        /// <returns>The modulus of the complex number in polar form.</returns>
        public static double PolarFormModulus(this Complex c)
        {
            var a = c.Real;
            var b = c.Imaginary;

            return System.Math.Sqrt((System.Math.Pow(a, 2) + System.Math.Pow(b, 2)));
        }

        /// <summary>
        /// Calculates the angle in radians for the complex number in polar form.
        /// </summary>
        /// <returns>The angle in radians for the complex number in polar form.</returns>
        public static double PolarFormAngle(this Complex c)
        {
            var a = c.Real;
            var b = c.Imaginary;
            return System.Math.Atan2(a, b);
        }

        /// <summary>
        /// Inverts the sign of the imaginary number part.
        /// </summary>
        public static Complex NegateImaginaryNumberPart(this Complex c)
        {
            return new Complex(c.Real, -c.Imaginary);
        }

        /// <summary>
        /// Inverts the sign of the real number part.
        /// </summary>
        public static Complex NegateRealNumberPart(this Complex c)
        {
            return new Complex(-c.Real, c.Imaginary);
        }

        /// <summary>
        /// Returns whether the <see cref="Complex"/> number is one (in an <see cref="Constants.Epsilon"/> sense).
        /// </summary>
        public static bool IsOne(this Complex complex)
        {
            return complex.Real.IsEqualTo(1.0) && complex.Imaginary.IsZero();
        }

        /// <summary>
        /// Returns whether the <see cref="Complex"/> number is the imaginary unit.
        /// </summary>
        public static bool IsImaginaryOne(this Complex complex)
        {
            return complex.Real.IsZero() && complex.Imaginary.IsEqualTo(1.0);
        }

        /// <summary>
        /// Returns whether either one or both parts of the <see cref="Complex"/> number is not a number (NaN).
        /// </summary>
        public static bool IsNaN(this Complex complex)
        {
            return double.IsNaN(complex.Real) || double.IsNaN(complex.Imaginary);
        }

        /// <summary>
        /// Returns whether either one or both parts of the <see cref="Complex"/> number is infinite (<see cref="double.IsInfinity"/>).
        /// </summary>
        public static bool IsInfinity(this Complex complex)
        {
            return double.IsInfinity(complex.Real) || double.IsInfinity(complex.Imaginary);
        }

        /// <summary>
        /// Returns whether the <see cref="Complex"/> number is purely real.
        /// </summary>
        public static bool IsReal(this Complex complex)
        {
            return complex.Imaginary.IsZero();
        }

        /// <summary>
        /// Returns whether the <see cref="Complex"/> number is purely real and negative.
        /// </summary>
        public static bool IsRealNonNegative(this Complex complex)
        {
            return complex.Imaginary.IsZero() && complex.Real >= 0;
        }

        /// <summary>
        /// Returns the <see cref="Complex"/> conjugate of the number.
        /// </summary>
        public static Complex Conjugate(this Complex complex)
        {
            return new Complex(complex.Real, -complex.Imaginary);
        }

        /// <summary>
        /// Gets the squared norm of the <see cref="Complex"/> number.
        /// </summary>
        public static double NormSquared(this Complex complex)
        {
            return (complex.Real * complex.Real) + (complex.Imaginary * complex.Imaginary);
        }


        /// <summary>
        /// Returns the (natural) exponent of the given <see cref="Complex"/> number.
        /// </summary>
        public static Complex Exp(this Complex complex)
        {
            var exp = System.Math.Exp(complex.Real);
            return complex.IsReal() ? new Complex(exp, 0.0) : new Complex(exp * System.Math.Cos(complex.Imaginary), exp * System.Math.Sin(complex.Imaginary));
        }

        /// <summary>
        /// Returns the (natural) logarithm of the given <see cref="Complex"/> number.
        /// </summary>
        public static Complex Log(this Complex complex)
        {
            return complex.IsRealNonNegative() ? new Complex(System.Math.Log(complex.Real), 0.0) : new Complex(0.5 * System.Math.Log(complex.NormSquared()), complex.Phase);
        }

        /// <summary>
        /// Returns the given <see cref="Complex"/> number to a power.
        /// </summary>
        public static Complex Pow(this Complex complex, Complex exponent)
        {
            if (complex.IsZero())
            {
                if (exponent.IsZero()) return Complex.One;
                if (exponent.Real > 0.0) return Complex.Zero;
                if (exponent.Real < 0)
                    return exponent.Imaginary.IsZero()
                            ? new Complex(double.PositiveInfinity, 0.0)
                            : new Complex(double.PositiveInfinity, double.PositiveInfinity);

                return double.NaN;
            }

            return (exponent * complex.Log()).Exp();
        }

        /// <summary>
        /// Returns the given <see cref="Complex"/> number to root.
        /// </summary>
        public static Complex Root(this Complex complex, Complex rootExponent)
        {
            return Pow(complex, 1 / rootExponent);
        }

        /// <summary>
        /// The square of the given <see cref="Complex"/> number.
        /// </summary>
        public static Complex Squared(this Complex complex)
        {
            return complex.IsReal() ? new Complex(complex.Real * complex.Real, 0.0) : new Complex((complex.Real * complex.Real) - (complex.Imaginary * complex.Imaginary), 2 * complex.Real * complex.Imaginary);
        }

        /// <summary>
        /// The square root of the given <see cref="Complex"/> number.
        /// </summary>
        public static Complex SquareRoot(this Complex complex)
        {
            if (complex.IsRealNonNegative()) return new Complex(System.Math.Sqrt(complex.Real), 0.0);
            Complex result;
            var absReal = System.Math.Abs(complex.Real);
            var absImag = System.Math.Abs(complex.Imaginary);
            double w;
            if (absReal >= absImag)
            {
                var ratio = complex.Imaginary / complex.Real;
                w = System.Math.Sqrt(absReal) * System.Math.Sqrt(0.5 * (1.0 + System.Math.Sqrt(1.0 + (ratio * ratio))));
            }
            else
            {
                var ratio = complex.Real / complex.Imaginary;
                w = System.Math.Sqrt(absImag) * System.Math.Sqrt(0.5 * (System.Math.Abs(ratio) + System.Math.Sqrt(1.0 + (ratio * ratio))));
            }

            if (complex.Real >= 0.0) result = new Complex(w, complex.Imaginary / (2.0 * w));
            else if (complex.Imaginary >= 0.0) result = new Complex(absImag / (2.0 * w), w);
            else result = new Complex(absImag / (2.0 * w), -w);

            return result;
        }

        /// <summary>
        /// The norm of the given <see cref="Complex"/> number.
        /// </summary>
        public static double Norm(this Complex complex)
        {
            return System.Math.Sqrt((complex.Real * complex.Real) + (complex.Imaginary * complex.Imaginary));
        }

        /// <summary>
        /// The squared norm of the difference of the given <see cref="Complex"/> numbers.
        /// </summary>
        public static double NormOfDifference(this Complex u, Complex v)
        {
            return (u - v).NormSquared();
        }

        /// <summary>
        /// Parses the complex number in form of 'a + ib'. If the parsing fails an exception will be thrown.
        /// </summary>
        /// <param name="complexString">The complex number to parse.</param>
        /// <returns>The parsed complex number.</returns>
        public static Complex ToComplex(this string complexString)
        {
            if (string.IsNullOrEmpty(complexString))
            {
                return new Complex(0, 0);
            }

            complexString = complexString.Trim();

            try
            {
                bool hasImaginaryPart = complexString.Last() == 'i' || complexString.Last() == 'j';

                if (!hasImaginaryPart)
                {
                    double realPart = double.Parse(complexString.TrimWhiteSpaceBetweenSignAndDigits());

                    return new Complex(realPart, 0);
                }
                else
                {
                    return ParseToComplexWithImaginaryPart(complexString);
                }
            }
            catch (Exception exception)
            {
                throw new ArgumentException("Cannot parse complexString to Complex number! ", exception);
            }
        }

        private static Complex ParseToComplexWithImaginaryPart(string complexString)
        {
            if (complexString.Length == 1)
            {
                return new Complex(0, 1);
            }

            int imaginarySignIndex = IndexOfImaginarySign(complexString);

            double realPart = 0;
            if (imaginarySignIndex > 0)
            {
                string realString = complexString
                    .Substring(0, imaginarySignIndex)
                    .TrimWhiteSpaceBetweenSignAndDigits();
                realPart = double.Parse(realString);
            }

            double imaginaryPart = 0;

            if (imaginarySignIndex == -1)
            {
                imaginarySignIndex = 0;
            }

            string imaginaryString = complexString
                .Substring(imaginarySignIndex, complexString.Length - imaginarySignIndex - 1)
                .TrimWhiteSpaceBetweenSignAndDigits();

            if (imaginaryString.Length == 1 && imaginaryString[0].IsPlusOrMinus())
            {
                imaginaryString += "1";
            }

            imaginaryPart = double.Parse(imaginaryString);

            return new Complex(realPart, imaginaryPart);
        }

        private static string TrimWhiteSpaceBetweenSignAndDigits(this string numberString)
        {
            if (numberString.Length < 2)
            {
                return numberString;
            }

            string sign = string.Empty;

            if (numberString[0].IsPlusOrMinus())
            {
                sign = numberString[0].ToString();
                numberString = numberString.Substring(1);
            }

            return string.Format("{0}{1}", sign, numberString.Trim());
        }

        private static int IndexOfImaginarySign(string complexString)
        {
            int signIndex = -1;

            for (int i = complexString.Length - 2; i > 0; i--)
            {
                if (complexString[i].IsPlusOrMinus() && !complexString[i - 1].IsScientificE())
                {
                    signIndex = i;
                    break;
                }
            }

            if (signIndex == -1 && complexString[0].IsPlusOrMinus())
            {
                signIndex = 0;
            }

            return signIndex;
        }

        private static bool IsPlusOrMinus(this char symbol)
        {
            return symbol == '+' || symbol == '-';
        }

        private static bool IsScientificE(this char symbol)
        {
            return symbol == 'e' || symbol == 'E';
        }
    }
}