﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Globalization;
using System.Linq;
using System.Windows;
using System.Windows.Media;

namespace Telerik.Windows.Documents.Spreadsheet.Maths
{
#pragma warning disable 3006
    /// <summary>
    /// Diverse extension method which do not fit in the more specialized extension classes.
    /// </summary>
    internal static class Utils
    {
        /// <summary>
        /// Tells you whether the given value sits in the specified interval.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <param name="lower">The lower.</param>
        /// <param name="upper">The upper.</param>
        /// <returns></returns>
        public static bool BetweenOrEqual(double value, double lower, double upper)
        {
            Sort(ref lower, ref upper);
            return BetweenOrEqualSorted(value, lower, upper);
        }

        /// <summary>
        /// The opposite of the approximation method, this attempts to smoothen out the polyline to a Bezier curve.
        /// </summary>
        /// <param name="points">The polyline.</param>
        /// <returns></returns>
        public static IEnumerable<Point> PolylineToBezier(Collection<Point> points)
        {
            var list = new Collection<Point>();
            Point p1;
            Point p2;
            list.Add(points[0]);
            for (var i = 0; i < points.Count - 2; i++)
            {
                p2 = points[i + 1];

                list.Add(p2);
                list.Add(p2);
                if (i != points.Count - 3)
                {
                    var ptf3 = points[i + 2];
                    p1 = new Point((p2.X + ptf3.X) / 2, (p2.Y + ptf3.Y) / 2);
                    list.Add(p1);
                }
                else
                {
                    list.Add(points[i + 2]);
                }
            }
            if (list.Count == 1)
            {
                list = new Collection<Point>();
                var point = points[0];
                p2 = points[points.Count - 1];
                p1 = new Point((point.X + p2.X) / 2, (point.Y + p2.Y) / 2);
                list.Add(point);
                list.Add(p1);
                list.Add(p1);
                list.Add(p2);
            }
            return list;
        }

        public static void Sort(ref double a, ref double b)
        {
            if (b >= a) return;
            var tmp = a;
            a = b;
            b = tmp;
        }

        /// <summary>
        /// Converts the Cartesian coordinates to polar coordinates.
        /// </summary>
        /// <param name="rootPoint">The root point.</param>
        /// <param name="otherPoint">The other point.</param>
        /// <param name="angle">The angle.</param>
        /// <param name="rho">The rho.</param>
        public static void CartesianToPolar(Point rootPoint, Point otherPoint, ref double angle, ref double rho)
        {
            if (rootPoint == otherPoint)
            {
                angle = 0;
                rho = 0;
                return;
            }
            var dx = otherPoint.X - rootPoint.X;
            var dy = otherPoint.Y - rootPoint.Y;
            rho = DistanceExstensions.Distance(dx, dy);
            angle = System.Math.Atan(-dy / dx) * 180 / System.Math.PI;
            if (dx < 0) angle += 180;
        }

        /// <summary>
        /// Polars to cartesian coordinates conversion.
        /// </summary>
        /// <param name="coordCenter">The coordinate center.</param>
        /// <param name="angle">The angle.</param>
        /// <param name="rho">The polar radius.</param>
        /// <returns></returns>
        /// <seealso cref="CartesianToPolar"/>
        public static Point PolarToCartesian(Point coordCenter, double angle, double rho)
        {
            var radians = angle.ToRadians();
            return new Point(coordCenter.X + (System.Math.Cos(radians) * rho), coordCenter.Y - (System.Math.Sin(radians) * rho));
        }

        /// <summary>
        /// Clones the specified list.
        /// </summary>
        /// <param name="doubles">The list to clone.</param>
        /// <returns>The cloned list.</returns>
        public static DoubleCollection Clone(this IEnumerable<double> doubles)
        {
            var dc = new DoubleCollection();
            doubles.ToList().ForEach(dc.Add);
            return dc;
        }

        /// <summary>
        /// Constrains the specified value to the given interval.
        /// </summary>
        /// <param name="value">The value to constrain.</param>
        /// <param name="min">The min.</param>
        /// <param name="max">The max.</param>
        /// <returns></returns>
        public static double Constrain(this double value, double min, double max)
        {
            return System.Math.Max(min, System.Math.Min(value, max));
        }

        public static bool AreIntersectingIntervals(int firstStart, int firstEnd, int secondStart, int secondEnd)
        {
            bool areIntersecting = (secondStart <= firstStart && firstStart <= secondEnd);
            areIntersecting |= (secondStart <= firstEnd && firstEnd <= secondEnd);
            areIntersecting |= (firstStart <= secondStart && secondStart <= firstEnd);
            areIntersecting |= (firstStart <= secondEnd && secondEnd <= firstEnd);

            return areIntersecting;
        }

        public static Tuple<int, int> GetIntersection(int firstStart, int firstEnd, int secondStart, int secondEnd)
        {
            if (!AreIntersectingIntervals(firstStart, firstEnd, secondStart, secondEnd))
            {
                return null;
            }

            int start = Math.Max(firstStart, secondStart);
            int end = Math.Min(firstEnd, secondEnd);

            return new Tuple<int, int>(start, end);
        }

#if SILVERLIGHT
		/// <summary>
		/// Returns the coordinates of the given point with respect to the root application. 
		/// The given point should be the coordinates with respect to the given framework element.
		/// </summary>
		/// <param name="frameworkElement">A framework element with respect to which the point is given.</param>
		/// <param name="point">A point with respect to the given framework element.</param>
		/// <returns>The coordinates of the point with respect to the root visual.</returns>
		/// <seealso cref="PointToScreen">Another alias for the same method.</seealso>
		public static Point ToRoot(this FrameworkElement frameworkElement, Point point)
		{
			return PointToScreen(frameworkElement, point);
		}

		/// <summary>
		/// Returns the coordinates of the given point with respect to the root application. 
		/// The given point should be the coordinates with respect to the given framework element.
		/// </summary>
		/// <param name="element">An element with respect to which the point is given.</param>
		/// <param name="point">A point with respect to the given framework element.</param>
		/// <returns>The coordinates of the point with respect to the root visual.</returns>
		/// <seealso cref="ToRoot">Another alias for the same method.</seealso>
		public static Point PointToScreen(this UIElement element, Point point)
		{
			return element.TransformToVisual(Application.Current.RootVisual).Transform(point);
		}

		/// <summary>
		/// Returns the coordinates of the given point with respect to the given element.
		/// The given coordinates should be with respect to the root visual.
		/// </summary>
		/// <param name="element">A UI element.</param>
		/// <param name="point">A point with coordinates with respect to the root visual.</param>
		/// <returns>The coordinates with respect to the given element.</returns>
		/// <seealso cref="FromRoot">Another alias for the same method.</seealso>
		public static Point PointFromScreen(this UIElement element, Point point)
		{
			return Application.Current.RootVisual.TransformToVisual(element).Transform(point);
		}

		/// <summary>
		/// Returns the coordinates of the given point with respect to the given element.
		/// The given coordinates should be with respect to the root visual.
		/// </summary>
		/// <param name="element">A UI element.</param>
		/// <param name="point">A point with coordinates with respect to the root visual.</param>
		/// <returns>The coordinates with respect to the given element.</returns>
		/// <seealso cref="PointFromScreen">Another alias for the same method.</seealso>
		public static Point FromRoot(this FrameworkElement element, Point point)
		{
			return PointToScreen(element, point);
		}

		/// <summary>
		/// Iterative/recursive search for the point between the two given points where the given element is crossed.
		/// The inside and outside points represent a line which is iterative subdivided in two to find where the given
		/// element is hit. The recursion stops when the division is less than five pixels (i.e. the error Margin).
		/// </summary>
		/// <param name="element">The element to search for.</param>
		/// <param name="insidePoint">The inside point.</param>
		/// <param name="outsidePoint">The outside point.</param>
		/// <returns></returns>
		public static Point BinarySearchIntersection(FrameworkElement element, Point insidePoint, Point outsidePoint)
		{
			var middlePoint =PointExtensions.MiddlePoint(insidePoint, outsidePoint);
			if (DistanceExstensions.Distance(insidePoint, outsidePoint) < 5) return middlePoint;
			var hit = VisualTreeHelper.FindElementsInHostCoordinates(middlePoint, Application.Current.RootVisual);
			if (hit.Any(e => e == element)) return BinarySearchIntersection(element, middlePoint, outsidePoint);
			return BinarySearchIntersection(element, insidePoint, middlePoint);
		}
#endif

        /// <summary>
        /// Gets the projected points of the specified point on the given rectangle.
        /// </summary>
        public static void GetProjections(Point point, Rect rectangle, Point[] projections)
        {
            projections[0] = projections[1] = projections[2] = projections[3] = point;
            projections[0].Y = rectangle.Top;
            projections[1].Y = rectangle.Bottom;
            projections[2].X = rectangle.Right;
            projections[3].X = rectangle.Left;
        }

        /// <summary>
        /// Gets the Bezier point at the specified value.
        /// </summary>
        /// <param name="points">The points.</param>
        /// <param name="segment">The segment.</param>
        /// <param name="value">A value in the interval [0,1].</param>
        /// <returns></returns>
        public static Point GetBezierPoint(Collection<Point> points, int segment, double value)
        {
            if (segment >= int.MaxValue / 3 || segment * 3 > points.Count - 3)
                throw new ArgumentOutOfRangeException("segment");

            var x0 = points[(segment * 3) + 0].X;
            var y0 = points[(segment * 3) + 0].Y;
            var x1 = points[(segment * 3) + 1].X;
            var y1 = points[(segment * 3) + 1].Y;
            var x2 = points[(segment * 3) + 2].X;
            var y2 = points[(segment * 3) + 2].Y;
            var x3 = points[(segment * 3) + 3].X;
            var y3 = points[(segment * 3) + 3].Y;
            var q0 = (1 - value) * (1 - value) * (1 - value);
            var q1 = 3 * value * (1 - value) * (1 - value);
            var q2 = 3 * value * value * (1 - value);
            var q3 = value * value * value;
            var xt = (q0 * x0) + (q1 * x1) + (q2 * x2) + (q3 * x3);
            var yt = (q0 * y0) + (q1 * y1) + (q2 * y2) + (q3 * y3);
            return new Point(xt, yt);
        }



        /// <summary>
        /// Gets the equal sides of a triange from the Pythagor theorem.
        /// </summary>
        /// <param name="hypotenuse">The hypotenuse.</param>
        /// <returns></returns>
        public static double GetPythagorEqualSide(double hypotenuse)
        {
            return System.Math.Sqrt(System.Math.Pow(hypotenuse, 2) / 2);
        }

        /// <summary>
        /// Gets the hypotenuse according to the Pythagor theorem.
        /// </summary>
        /// <param name="sideA">The first side.</param>
        /// <param name="sideB">The second side.</param>
        /// <returns></returns>
        public static double GetPythagorHypotenuse(double sideA, double sideB)
        {
            return System.Math.Sqrt(System.Math.Pow(sideA, 2) + System.Math.Pow(sideB, 2));
        }

        /// <summary>
        /// Determines whether [has valid area] [the specified size].
        /// </summary>
        /// <param name="size">The size.</param>
        /// <returns>
        ///   <c>true</c> if [has valid area] [the specified size]; otherwise, <c>false</c>.
        /// </returns>
        public static bool HasValidArea(this Size size)
        {
            if (size.Width <= 0 || size.Height <= 0 || double.IsInfinity(size.Width)) return false;
            return !double.IsInfinity(size.Height);
        }

        /// <summary>
        /// Hypotenuses the specified x.
        /// </summary>
        /// <param name="x">The x.</param>
        /// <param name="y">The y.</param>
        /// <returns></returns>
        public static double Hypotenuse(double x, double y)
        {
            return System.Math.Sqrt((x * x) + (y * y));
        }



        /// <summary>
        /// Determines whether [is point in rectangle] [the specified pt].
        /// </summary>
        /// <param name="pt">The pt.</param>
        /// <param name="rc">The rc.</param>
        /// <returns>
        ///   <c>true</c> if [is point in rectangle] [the specified pt]; otherwise, <c>false</c>.
        /// </returns>
        public static bool IsPointInRectangle(Point pt, Rect rc)
        {
            return rc.Contains(pt);
        }

        /// <summary>
        /// Determines whether [is in right open interval] [the specified v].
        /// </summary>
        /// <param name="v">The v.</param>
        /// <param name="min">The min.</param>
        /// <param name="max">The max.</param>
        /// <returns>
        ///   <c>True</c> f [is in right open interval] [the specified v]; otherwise, <c>false</c>.
        /// </returns>
        public static bool IsInRightOpenInterval(this double v, double min, double max)
        {
            return min <= v && v < max;
        }

        /// <summary>
        /// Determines whether [is in closed interval] [the specified v].
        /// </summary>
        /// <param name="v">The v.</param>
        /// <param name="min">The min.</param>
        /// <param name="max">The max.</param>
        /// <returns>
        ///   <c>True</c> if [is in closed interval] [the specified v]; otherwise, <c>false</c>.
        /// </returns>
        public static bool IsInClosedInterval(this double v, double min, double max)
        {
            return min <= v && v <= max;
        }

        /// <summary>
        /// Inverts the matrix.
        /// </summary>
        /// <param name="m">The matrix.</param>
        /// <returns></returns>
        public static Matrix Invert(this Matrix m)
        {
            return m.InverseMatrix();
        }


        /// <summary>
        /// Moves the given point into the rectangle by taking the rectangle's intervals as limiting values for
        /// the point's coordinates.
        /// </summary>
        /// <param name="p">Any point.</param>
        /// <param name="rectangle">A rectangle which acts as limiting container.</param>
        /// <returns></returns>
        public static Point Limit(this Point p, Rect rectangle)
        {
            return new Point(System.Math.Min(System.Math.Max(p.X, rectangle.Left), rectangle.Right), System.Math.Min(System.Math.Max(p.Y, rectangle.Top), rectangle.Bottom));
        }

        /// <summary>
        /// Linear interpolation between the given values.
        /// </summary>
        /// <remarks>See http://en.wikipedia.org/wiki/Linear_interpolation .</remarks>
        /// <param name="x">A value.</param>
        /// <param name="y">Another value.</param>
        /// <param name="fraction">A value in the [0,1] interval. At zero the interpolation returns the first value, at one it results in the second value.</param>
        /// <returns></returns>
        public static double Lerp(double x, double y, double fraction)
        {
            return (x * (1.0 - fraction)) + (y * fraction);
        }

        /// <summary>
        /// Multiplies the given matrices.
        /// </summary>
        /// <param name="m1">A matrix.</param>
        /// <param name="m2">Another matrix.</param>
        /// <returns></returns>
        public static Matrix Multiply(this Matrix m1, Matrix m2)
        {
            return m1.MultiplyBy(m2);
        }

        /// <summary>
        /// Normalizes the specified angle into the [0, 2Pi] interval.
        /// </summary>
        /// <param name="angle">
        /// The angle.
        /// </param>
        /// <returns>
        /// The normalize.
        /// </returns>
        public static double NormalizeAngle(double angle)
        {
            while (angle > System.Math.PI * 2) angle -= 2 * System.Math.PI;
            while (angle < 0) angle += System.Math.PI * 2;
            return angle;
        }

        /// <summary>
        /// Offsets the point collection.
        /// </summary>
        /// <param name="points">The points.</param>
        /// <param name="originalPoints">The original points.</param>
        /// <param name="offset">The offset.</param>
        public static void OffsetPointCollection(Collection<Point> points, Collection<Point> originalPoints, Vector offset)
        {
            if (points.Count != originalPoints.Count) return;
            for (var i = 0; i < points.Count; ++i) points[i] = originalPoints[i] + offset;
        }

        /// <summary>
        /// Pushes the specified start point onwards in the direction specified by the vector and a length given.
        /// </summary>
        /// <param name="start">The start.</param>
        /// <param name="unitVector">The unit vector.</param>
        /// <param name="distance">The distance to push.</param>
        /// <returns></returns>
        public static Point Push(Point start, Vector unitVector, double distance)
        {
            return new Point(start.X + (unitVector.X * distance), start.Y + (unitVector.Y * distance));
        }

        /// <summary>
        /// Safes the divide.
        /// </summary>
        /// <param name="value1">The value1.</param>
        /// <param name="value2">The value2.</param>
        /// <param name="fallback">The fallback.</param>
        /// <returns></returns>
        public static double SafeDivide(this double value1, double value2, double fallback)
        {
            if (!value2.IsVerySmall()) return value1 / value2;
            return fallback;
        }

        /// <summary>
        /// Shrinks the specified outer bounds.
        /// </summary>
        /// <param name="outerBounds">The outer bounds.</param>
        /// <param name="left">The left.</param>
        /// <param name="top">The top.</param>
        /// <param name="right">The right.</param>
        /// <param name="bottom">The bottom.</param>
        /// <returns></returns>
        public static Rect Shrink(Rect outerBounds, double left = 0, double top = 0, double right = 0, double bottom = 0)
        {
            return new Rect(outerBounds.Left + left, outerBounds.Top + top, outerBounds.Width - left - right, outerBounds.Height - top - bottom);
        }

        /// <summary>
        /// Returns zero within the given interval and +/-1 whether the value is above or below the interval.
        /// </summary>
        /// <param name="value">The argument of the function.</param>
        /// <param name="lower">The lower value of the interval.</param>
        /// <param name="upper">The upper value of the interval.</param>
        /// <returns>Zero, -1 or +1 whether the value is below, resp. inside, resp. above the interval.</returns>
        public static int StairValue(double value, double lower, double upper)
        {
            if (value < lower) return -1;
            return value < upper ? 0 : 1;
        }

        /// <summary>
        /// Returns the <see cref="StairValue(double,double,double)"/> of the x,y values of the given point with.
        /// respect to the boundaries of the specified rectangle.
        /// </summary>
        /// <param name="p">The point to test.</param>
        /// <param name="rectangle">The rectangle which acts as the domain of the stair function.</param>
        /// <returns>Zero, -1 or +1 whether the coordinate is below, resp. inside, resp. above the boundary of the rectangle.</returns>
        public static Vector StairValue(Point p, Rect rectangle)
        {
            return new Vector(StairValue(p.X, rectangle.Left, rectangle.Right), StairValue(p.Y, rectangle.Top, rectangle.Bottom));
        }

        /// <summary>
        /// Swaps the values of the two numbers.
        /// </summary>
        /// <param name="a">A value.</param>
        /// <param name="b">Another value.</param>
        public static void Swap(ref double a, ref double b)
        {
            var t = a;
            a = b;
            b = t;
        }

        /// <summary>
        /// Snaps the specified value.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <param name="snappingValue">The snapping value.</param>
        /// <returns></returns>
        public static double Snap(this double value, int snappingValue)
        {
            return value - (value % snappingValue);
        }

        /// <summary>
        /// Transforms the specified tr.
        /// </summary>
        /// <param name="tr">The tr.</param>
        /// <param name="r">The r.</param>
        /// <returns></returns>
        public static Rect Transform(this GeneralTransform tr, Rect r)
        {
            return new Rect(tr.Transform(r.TopLeft()), tr.Transform(r.BottomRight()));
        }

        /// <summary>
        /// Transforms the specified m.
        /// </summary>
        /// <param name="m">The m.</param>
        /// <param name="r">The r.</param>
        /// <returns></returns>
        public static Rect Transform(this Matrix m, Rect r)
        {
            return new Rect(m.Transform(r.TopLeft()), m.Transform(r.BottomRight()));
        }

        /// <summary>
        /// Transforms the size of the percent to.
        /// </summary>
        /// <param name="rect">The rect.</param>
        /// <param name="size">The size.</param>
        /// <returns></returns>
        public static Rect TransformPercentToSize(Rect rect, Size size)
        {
            return new Rect(PointExtensions.PointFromBarycentricPercentage(rect.TopLeft(), size), PointExtensions.PointFromBarycentricPercentage(rect.BottomRight(), size));
        }

        /// <summary>
        /// Toes the matrix.
        /// </summary>
        /// <param name="transform">The transform.</param>
        /// <returns></returns>
        public static Matrix ToMatrix(this Transform transform)
        {
            var matrixTransform = transform as MatrixTransform;
            if (matrixTransform != null) return matrixTransform.Matrix;
            var transformGroup = transform as TransformGroup;
            if (transformGroup != null) return transformGroup.Value;
            transformGroup = new TransformGroup();
            transformGroup.Children.Add(transform);
            return transformGroup.Value;
        }



        /// <summary>
        /// Converts the point to a string representation.
        /// </summary>
        /// <param name="size">The size.</param>
        /// <returns></returns>
        public static string ToInvariantString(this Size size)
        {
            var width = double.IsNaN(size.Width) ? "Auto" : size.Width.ToString(CultureInfo.InvariantCulture);
            var height = double.IsNaN(size.Height) ? "Auto" : size.Height.ToString(CultureInfo.InvariantCulture);
            return string.Format("{0};{1}", width, height);
        }

        /// <summary>
        /// Converts the double to a string representation.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <returns></returns>
        public static string ToInvariantString(this double value)
        {
            return double.IsNaN(value) ? "Auto" : value.ToString(CultureInfo.InvariantCulture);
        }

        /// <summary>
        /// Converts the int to a string representation.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <returns></returns>
        public static string ToInvariantString(this int value)
        {
            return value.ToString(CultureInfo.InvariantCulture);
        }

        /// <summary>
        /// Sizes from string.
        /// </summary>
        /// <param name="s">The value.</param>
        /// <returns></returns>
        public static Size? ToSize(string s)
        {
            if (string.IsNullOrEmpty(s)) return null;
            var parts = s.Split(';');
            if (parts.Length < 2) return null;
            double x, y;
            if (parts[0].ToLower() == "auto") x = double.NaN;
            else if (!double.TryParse(parts[0], NumberStyles.Float, CultureInfo.InvariantCulture, out x)) return null;

            if (parts[1].ToLower() == "auto") y = double.NaN;
            else if (!double.TryParse(parts[1], NumberStyles.Float, CultureInfo.InvariantCulture, out y)) return null;

            return new Size(x, y);
        }

        /// <summary>
        /// Converts the given value to an enum type.
        /// </summary>
        /// <typeparam name="T">The type of the enumeration.</typeparam>
        /// <param name="value">The value.</param>
        /// <returns></returns>
        public static T ToEnum<T>(object value)
        {
            if (value == null)
                return default(T);
            return (T)Enum.Parse(typeof(T), value.ToString(), true);
        }

        internal static Matrix ToMatix(this GeneralTransform tr)
        {
            var mt = tr as MatrixTransform;
            return mt != null ? mt.Matrix : Matrix.Identity;
        }

        internal static Point RectanglePointFromPercent(Point pointPercent, Rect rc)
        {
            return new Point(rc.Left + (pointPercent.X / 100 * rc.Width),
                             rc.Top + (pointPercent.Y / 100 * rc.Height));
        }

        internal static double ReplaceZero(double suspect, double defaultValue)
        {
            return suspect.IsEqualTo(0) ? defaultValue : suspect;
        }

        internal static List<object> Clone(this IEnumerable<object> list)
        {
            var clone = new List<object>();
            clone.AddRange(list);
            return clone;
        }

        internal static Point PercentagePoint(Point pt, Rect rc)
        {
            var percentagePoint = new Point(50, 50);

            var w = rc.Right - rc.Left;
            var h = rc.Bottom - rc.Top;

            if (w.IsNotEqualTo(0) && h.IsNotEqualTo(0))
            {
                percentagePoint.X = (pt.X - rc.Left) * 100 / w;
                percentagePoint.Y = (pt.Y - rc.Top) * 100 / h;
            }

            return percentagePoint;
        }

        internal static Size Scale(this Size size, double scale)
        {
            return new Size(size.Width * scale, size.Height * scale);
        }

        /// <summary>
        /// Gets the bezier coefficients, see http://processingjs.nihongoresources.com/bezierinfo/ .
        /// </summary>
        /// <param name="a0">The a0.</param>
        /// <param name="a1">The a1.</param>
        /// <param name="a2">The a2.</param>
        /// <param name="a3">The a3.</param>
        /// <param name="b0">The b0.</param>
        /// <param name="b1">The b1.</param>
        /// <param name="b2">The b2.</param>
        /// <param name="b3">The b3.</param>
        /// <param name="u">The u.</param>
        /// <param name="s">The s.</param>
        /// <param name="z">The z.</param>
        /// <param name="x4">The x4.</param>
        /// <param name="y4">The y4.</param>
        public static void GetBezierCoefficients(ref double a0, ref double a1, ref double a2, ref double a3, ref double b0, ref double b1, ref double b2, ref double b3, ref double u, ref double s, ref double z, ref double x4, ref double y4)
        {
            var x = a0 + (u * (a1 + (u * (a2 + (u * a3)))));
            var dx4 = x - x4;
            var dy = b1 + (u * ((2 * b2) + (3 * u * b3)));
            var y = b0 + (u * (b1 + (u * (b2 + (u * b3)))));
            var dy4 = y - y4;
            var dx = a1 + (u * ((2 * a2) + (3 * u * a3)));
            s = (dx4 * dx4) + (dy4 * dy4);
            z = (dx * dx4) + (dy * dy4);
        }

        public static bool BetweenOrEqualSorted(double n, double boundary1, double boundary2)
        {
            return boundary1 <= n && n <= boundary2;
        }

        private static char GetPointSeparator(IFormatProvider provider)
        {
            var ch = ',';
            var instance = NumberFormatInfo.GetInstance(provider);
            if ((instance.NumberDecimalSeparator.Length > 0) && (ch == instance.NumberDecimalSeparator[0])) ch = ';';
            return ch;
        }

        public static bool ArraysEqual<T>(T[] a1, T[] a2)
        {
            if (ReferenceEquals(a1, a2)) return true;

            if (a1 == null || a2 == null) return false;

            if (a1.Length != a2.Length) return false;

            var comparer = EqualityComparer<T>.Default;
            return !a1.Where((t, i) => !comparer.Equals(t, a2[i])).Any();
        }

        public static bool ArraysEqual<T>(T[,] a1, T[,] a2)
        {
            if (ReferenceEquals(a1, a2)) return true;

            if (a1 == null || a2 == null) return false;

            if (a1.Length != a2.Length) return false;
            for (var i = 0; i < 2; i++) if (a1.GetLength(i) != a2.GetLength(i)) return false;
            var comparer = EqualityComparer<T>.Default;
            for (var m = 0; m < a1.GetLength(0); m++)
            {
                for (var n = 0; n < a1.GetLength(1); n++)
                {
                    if (!comparer.Equals(a1[m, n], a2[m, n])) return false;
                }
            }
            return true;
        }
    }
}