﻿using System.Windows;

namespace Telerik.Windows.Documents.Spreadsheet.Maths
{
    internal static class VectorExtensions
	{
		/// <summary>
		/// Linear interpolation between the given vectors.
		/// </summary>
		/// <param name="u">A point.</param>
		/// <param name="v">Another point.</param>
		/// <param name="fraction">A value in the [0,1] interval. At zero the interpolation returns the first vector, at one it results in the second vector.</param>
		/// <returns></returns>
		public static Vector Lerp(Vector u, Vector v, double fraction)
		{
			return new Vector(Utils.Lerp(u.X, v.X, fraction), Utils.Lerp(u.Y, v.Y, fraction));
		}

		/// <summary>
		/// Returns the mirrored vector with respect to the X-coordinate.
		/// </summary>
		/// <param name="v">The vector to mirror.</param>
		/// <returns>The mirrored vector.</returns>
		public static Vector MirrorHorizontally(this Vector v)
		{
			return new Vector(v.Y, -v.X);
		}

		/// <summary>
		/// Returns the mirrored vector with respect to the Y-coordinate.
		/// </summary>
		/// <param name="v">The vector to mirror.</param>
		/// <returns>The mirrored vector.</returns>
		public static Vector MirrorVertically(this Vector v)
		{
			return new Vector(-v.Y, v.X);
		}

		/// <summary>
		/// Returns a unit vector in the direction specified by the angle.
		/// </summary>
		/// <param name="degrees">The angle in degrees.</param>
		/// <returns>A unit vector.</returns>
		public static Vector UnitVector(double degrees)
		{
			var rad = Trigonometry.DegreesToRadians(degrees);
			return new Vector(System.Math.Cos(rad), System.Math.Sin(rad));
		}

		/// <summary>
		/// Returns the perpendicular of the specified vector.
		/// </summary>
		/// <param name="v">A vector.</param>
		/// <returns></returns>
		public static Vector Perpendicular(Vector v)
		{
			return new Vector(v.Y, -v.X);
		}

		/// <summary>
		/// Normalizeds the specified vector.
		/// </summary>
		/// <param name="vector">The vector.</param>
		/// <returns></returns>
		public static Vector Normalized(this Vector vector)
		{
			var v = new Vector(vector.X, vector.Y);
			var length = v.Length;
			return !length.IsVerySmall() ? v / length : new Vector(0, 1);
		}
	}
}