﻿using System;
using System.Collections.Generic;
using System.Text;

namespace Telerik.Windows.Documents.Spreadsheet.Maths.Number
{
    internal class FiniteIntervalsUnion<T> 
        where T : IComparable
    {
        private const string EmptySetSymbol = "Ø";
        private const string UnionSymbol = "∪";
        private readonly List<FiniteInterval<T>> sortedIntervals;

        public FiniteIntervalsUnion()
        {
            this.sortedIntervals = new List<FiniteInterval<T>>();
        }

        public IEnumerable<T> IntervalEndPointValues
        {
            get
            {
                foreach (FiniteInterval<T> interval in this.sortedIntervals)
                {
                    yield return interval.Start;
                    yield return interval.End;
                }
            }
        }

        public IEnumerable<FiniteInterval<T>> Intervals
        {
            get
            {
                foreach (var interval in this.sortedIntervals)
                {
                    yield return interval;
                }
            }
        }

        public int IntervalsCount
        {
            get
            {
                return this.sortedIntervals.Count;
            }
        }

        public bool IsEmpty
        {
            get
            {
                return this.sortedIntervals.Count == 0;
            }
        }

        public void Clear()
        {
            this.sortedIntervals.Clear();
        }

        public FiniteInterval<T> GetContainingInterval(T point)
        {
            int smallerOrEqualIntervalIndex;

            return this.TryGetContainingIntervalOrSmallerIntervalIndex(point, out smallerOrEqualIntervalIndex);
        }

        public void AddInterval(T start, T end, bool isStartOpened = false, bool isEndOpened = false)
        {
            if (start.CompareTo(end) > 0)
            {
                throw new ArgumentException("Interval start have value less than or equal to interval end value!");
            }

            if (start.CompareTo(end) == 0)
            {
                if (isStartOpened && isEndOpened)
                {
                    return;
                }
                else
                {
                    isStartOpened = false;
                    isEndOpened = false;
                }
            }
            
            int startIndexToRemove;
            T insertionIntervalStartPoint;
            bool insertionStartIsOpened;
            this.CalculateIntervalsStartChanges(start, isStartOpened, out startIndexToRemove, out insertionIntervalStartPoint, out insertionStartIsOpened);

            int endIndexToRemove;
            T insertionIntervalEndPoint;
            bool insertionEndIsOpened;
            this.CalculateIntervalsEndChanges(end, isEndOpened, out endIndexToRemove, out insertionIntervalEndPoint, out insertionEndIsOpened);

            if (startIndexToRemove <= endIndexToRemove)
            {
                int elementsToRemove = endIndexToRemove - startIndexToRemove + 1;
                this.sortedIntervals.RemoveRange(startIndexToRemove, elementsToRemove);
            }

            FiniteInterval<T> intervalToInsert = new FiniteInterval<T>(insertionIntervalStartPoint, insertionIntervalEndPoint, insertionStartIsOpened, insertionEndIsOpened);
            this.sortedIntervals.Insert(startIndexToRemove, intervalToInsert);
        }

        private void CalculateIntervalsStartChanges(T startPoint, bool isStartOpened,
            out int startIndexToRemove, out T insertionIntervalStartPoint, out bool insertionStartIsOpened)
        {
            int smallerOrEqualIntervalIndex;
            FiniteInterval<T> containingInterval = this.TryGetContainingIntervalOrSmallerIntervalIndex(startPoint, out smallerOrEqualIntervalIndex);

            startIndexToRemove = smallerOrEqualIntervalIndex + 1;
            insertionIntervalStartPoint = startPoint;
            insertionStartIsOpened = isStartOpened;

            if (containingInterval == null)
            {
                if (smallerOrEqualIntervalIndex > -1)
                {
                    FiniteInterval<T> smallerInterval = this.sortedIntervals[smallerOrEqualIntervalIndex];
                    bool shouldUniteIntervals = !isStartOpened && (smallerInterval.End.CompareTo(startPoint) == 0);

                    if (shouldUniteIntervals)
                    {
                        startIndexToRemove = smallerOrEqualIntervalIndex;
                        insertionIntervalStartPoint = smallerInterval.Start;
                        insertionStartIsOpened = smallerInterval.IsStartOpened;
                    }
                }
            }
            else
            {
                startIndexToRemove = smallerOrEqualIntervalIndex;
                insertionIntervalStartPoint = containingInterval.Start;
                insertionStartIsOpened = containingInterval.IsStartOpened;
            }
        }

        private void CalculateIntervalsEndChanges(T endPoint, bool isEndOpened,
            out int endIndexToRemove, out T insertionIntervalEndPoint, out bool insertionEndIsOpened)
        {
            int smallerOrEqualIntervalIndex;
            FiniteInterval<T> containingInterval = this.TryGetContainingIntervalOrSmallerIntervalIndex(endPoint, out smallerOrEqualIntervalIndex);

            endIndexToRemove = smallerOrEqualIntervalIndex;
            insertionIntervalEndPoint = endPoint;
            insertionEndIsOpened = isEndOpened;

            if (containingInterval == null)
            {
                int biggerIntervalIndex = smallerOrEqualIntervalIndex + 1;
                if (biggerIntervalIndex < this.sortedIntervals.Count)
                {
                    FiniteInterval<T> biggerInterval = this.sortedIntervals[biggerIntervalIndex];
                    bool shouldUniteIntervals = !isEndOpened && (biggerInterval.Start.CompareTo(endPoint) == 0);

                    if (shouldUniteIntervals)
                    {
                        endIndexToRemove = biggerIntervalIndex;
                        insertionIntervalEndPoint = biggerInterval.End;
                        insertionEndIsOpened = biggerInterval.IsEndOpened;
                    }
                }
            }
            else
            {
                endIndexToRemove = smallerOrEqualIntervalIndex;
                insertionIntervalEndPoint = containingInterval.End;
                insertionEndIsOpened = containingInterval.IsEndOpened;
            }
        }

        private FiniteInterval<T> TryGetContainingIntervalOrSmallerIntervalIndex(T point, out int smallerOrEqualIntervalIndex)
        {
            smallerOrEqualIntervalIndex = -1;

            if(this.sortedIntervals.Count == 0)
            {
                return null;
            }

            int startIndex = 0;
            int endIndex = this.sortedIntervals.Count - 1;

            while (startIndex < endIndex)
            {
                int currentIndex = (startIndex + endIndex) / 2;

                int intervalPointComparisson = this.sortedIntervals[currentIndex].CompareTo(point);

                if (intervalPointComparisson > 0)
                {
                    endIndex = currentIndex - 1;
                    smallerOrEqualIntervalIndex = endIndex;
                }
                else if (intervalPointComparisson < 0)
                {
                    startIndex = currentIndex + 1;
                }
                else
                {
                    smallerOrEqualIntervalIndex = currentIndex;
                    return this.sortedIntervals[currentIndex];
                }
            }

            int lastIndexToCheck = startIndex;
            int lastIntervalPointComparisson = this.sortedIntervals[lastIndexToCheck].CompareTo(point);
            smallerOrEqualIntervalIndex = lastIntervalPointComparisson > 0 ? lastIndexToCheck - 1 : lastIndexToCheck;

            return lastIntervalPointComparisson == 0 ? this.sortedIntervals[lastIndexToCheck] : null;
        }

        public override string ToString()
        {
            if (this.IsEmpty)
            {
                return EmptySetSymbol;
            }

            StringBuilder builder = new StringBuilder();
            builder.Append(this.sortedIntervals[0].ToString());

            for (int i = 1; i < this.sortedIntervals.Count; i++)
            {
                builder.AppendFormat(" {0} {1}", UnionSymbol, this.sortedIntervals[i]);
            }

            return builder.ToString();
        }

        public override bool Equals(object obj)
        {
            FiniteIntervalsUnion<T> other = obj as FiniteIntervalsUnion<T>;

            if (other == null)
            {
                return false;
            }

            int intervalsCount = this.sortedIntervals.Count;
            if (!other.sortedIntervals.Count.Equals(intervalsCount))
            {
                return false;
            }

            for (int i = 0; i < intervalsCount; i++)
            {
                if (!this.sortedIntervals[i].Equals(other.sortedIntervals[i]))
                {
                    return false;
                }
            }

            return true;
        }

        public override int GetHashCode()
        {
            if (this.IsEmpty)
            {
                return 0;
            }

            int hash = this.sortedIntervals[0].GetHashCode();

            for (int i = 1; i < this.sortedIntervals.Count; i++)
            {
                hash ^= this.sortedIntervals[i].GetHashCode();
            }

            return hash;
        }
    }
}
