﻿using System;

namespace Telerik.Windows.Documents.Spreadsheet.Maths.Number
{
    internal class TwoComplementNumber
    {
        private static readonly string bitsCountOutOfRangeMessage = String.Format("bitsCount must be in range from {0} to {1}!", MinSupportedBits, MaxSupportedBits);
        public const int MaxSupportedBits = 64;
        public const int MinSupportedBits = 2;

        private readonly long bitsArray;
        private readonly int bitsCount;
        private int maxBitIndex;

        public int NegativeBitIndex
        {
            get
            {
                return this.bitsCount - 1;
            }
        }

        private long this[int bitIndex]
        {
            get
            {
                return ((this.bitsArray >> bitIndex) & 1L);
            }
        }

        private TwoComplementNumber(long bitsArray, int bitsCount)
        {
            bool hasValidBitsCount = HasValidBitsCount(bitsCount);
            if (!hasValidBitsCount)
            {
                throw new ArgumentException(bitsCountOutOfRangeMessage, "negativeBitIndex");
            }

            this.bitsCount = bitsCount;
            this.bitsArray = bitsArray;

            this.CalculateMaxBitIndex();

            if (this.maxBitIndex > this.NegativeBitIndex)
            {
                throw new ArgumentException(String.Format("Invalid bits array - the maxIndex of a non-zero bit must be equal to (bitsCount-1)!"), "bitsArray");
            }
        }

        private int CalculateMaxBitIndex()
        {
            this.maxBitIndex = -1;
            for (int i = MaxSupportedBits - 1; i >= 0; i--)
            {
                if (this[i] == 1)
                {
                    this.maxBitIndex = i;

                    return this.maxBitIndex;
                }
            }

            return this.maxBitIndex;
        }

        private static bool HasValidBitsCount(int bitsCount)
        {
            return MinSupportedBits <= bitsCount && bitsCount <= MaxSupportedBits;
        }

        public static TwoComplementNumber FromDec(long decNumber, int bitsCount)
        {
            long minNumber, maxNumber;
            if (TryGetDecRange(bitsCount, out minNumber, out maxNumber))
            {
                if (decNumber > maxNumber || decNumber < minNumber)
                {
                    throw new ArgumentOutOfRangeException("decNumber must be in range corresponding to the given bitsCount", "decNumber");
                }

                if (bitsCount == MaxSupportedBits || decNumber >= 0)
                {
                    return new TwoComplementNumber(decNumber, bitsCount);
                }
                else
                {
                    long positiveBits = decNumber - minNumber;
                    long negativeBit = (1L << (bitsCount - 1));
                    long bitsArray = negativeBit | positiveBits;

                    return new TwoComplementNumber(bitsArray, bitsCount);
                }
            }
            else
            {
                throw new ArgumentException(bitsCountOutOfRangeMessage, "negativeBitIndex");
            }
        }

        private static bool TryGetDecRange(int bitsCount, out long minNumber, out long maxNumber)
        {
            if (bitsCount == MaxSupportedBits)
            {
                minNumber = long.MinValue;
                maxNumber = long.MaxValue;
            }
            else
            {
                minNumber = -(1L << (bitsCount - 1));
                maxNumber = -minNumber - 1;
            }

            return HasValidBitsCount(bitsCount);
        }

        public static TwoComplementNumber FromBin(string binNumber, int bitsCount)
        {
            long bitArray = Convert.ToInt64(binNumber, 2);

            return new TwoComplementNumber(bitArray, bitsCount);
        }

        public static TwoComplementNumber FromOct(string hexNumber, int bitsCount)
        {
            long bitsArray = Convert.ToInt64(hexNumber, 8);

            return new TwoComplementNumber(bitsArray, bitsCount);
        }

        public static TwoComplementNumber FromHex(string hexNumber, int bitsCount)
        {
            long bitsArray = Convert.ToInt64(hexNumber, 16);

            return new TwoComplementNumber(bitsArray, bitsCount);
        }

        public long ToDec()
        {
            if (this.bitsCount == MaxSupportedBits)
            {
                return this.bitsArray;
            }
            else
            {
                long negative = (this.bitsArray & (1L << this.NegativeBitIndex));
                long positive = (this.bitsArray & (~(1L << this.NegativeBitIndex)));
                long result = positive - negative;

                return result;
            }
        }

        public string ToBin()
        {
            return Convert.ToString(this.bitsArray, 2);
        }

        public string ToOct()
        {
            return Convert.ToString(this.bitsArray, 8);
        }

        public string ToHex()
        {
            return Convert.ToString(this.bitsArray, 16).ToUpper();
        }

        public override string ToString()
        {
            return string.Format("TwoComplement: dec -> {0}, bin -> {1}, oct -> {2}, hex -> {3}", this.ToDec(), this.ToBin(), this.ToOct(), this.ToHex());
        }
    }
}