﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;

namespace Telerik.Windows.Documents.Spreadsheet.Maths
{
	/// <summary>
	/// A real-valued matrix.
	/// </summary>
    internal struct RMatrix
#if WPF
	: ICloneable
#endif
	{
		private readonly int rowCount;
		private readonly int colCount;
		private double[,] matrix;

		public RMatrix(int rowCount, int colCount)
		{
			this.rowCount = rowCount;
			this.colCount = colCount;
			this.matrix = new double[rowCount, colCount];
			for (var i = 0; i < rowCount; i++)
				for (var j = 0; j < colCount; j++)
					this.matrix[i, j] = 0.0;
		}

		public RMatrix(double[,] matrix)
		{
			this.rowCount = matrix.GetLength(0);
			this.colCount = matrix.GetLength(1);
			this.matrix = matrix;
		}

		public RMatrix(RMatrix m)
		{
			this.rowCount = m.GetnRows;
			this.colCount = m.GetnCols;
			this.matrix = m.matrix;
		}

		public RMatrix IdentityMatrix()
		{
			var m = new RMatrix(this.rowCount, this.colCount);
			for (var i = 0; i < this.rowCount; i++)
				for (var j = 0; j < this.colCount; j++)
					if (i == j) m[i, j] = 1;
			return m;
		}

		public double this[int m, int n]
		{
			get
			{
				if (m < 0 || m > this.rowCount) throw new Exception("m-th row is out of range!");
				if (n < 0 || n > this.colCount) throw new Exception("n-th col is out of range!");
				return this.matrix[m, n];
			}
			set { this.matrix[m, n] = value; }
		}

		public int GetnRows
		{
			get { return this.rowCount; }
		}

		public int GetnCols
		{
			get { return this.colCount; }
		}

		public RMatrix Clone()
		{
			var m = new RMatrix(this.matrix) { matrix = (double[,])this.matrix.Clone() };
			return m;
		}

#if WPF
		object ICloneable.Clone()
		{
			return this.Clone();
		}
#endif

		public override string ToString()
		{
			var strMatrix = "(";
			for (var i = 0; i < this.rowCount; i++)
			{
				var str = "";
				for (var j = 0; j < this.colCount - 1; j++)
				{
					str += this.matrix[i, j].ToString(CultureInfo.InvariantCulture) + ", ";
				}
				str += this.matrix[i, this.colCount - 1].ToString(CultureInfo.InvariantCulture);
				if (i != this.rowCount - 1 && i == 0)
					strMatrix += str + "\n";
				else if (i != this.rowCount - 1 && i != 0)
					strMatrix += " " + str + "\n";
				else
					strMatrix += " " + str + ")";
			}
			return strMatrix;
		}

		public override bool Equals(object obj)
		{
			return (obj is RMatrix) && this.Equals((RMatrix)obj);
		}

		public bool Equals(RMatrix m)
		{
			return Utils.ArraysEqual(this.matrix,m.matrix);
		}

		public override int GetHashCode()
		{
			return this.matrix.GetHashCode();
		}

		public static bool operator ==(RMatrix m1, RMatrix m2)
		{
			return m1.Equals(m2);
		}

		public static bool operator !=(RMatrix m1, RMatrix m2)
		{
			return !m1.Equals(m2);
		}

		public static RMatrix operator +(RMatrix m)
		{
			return m;
		}

		public static RMatrix operator +(RMatrix m1, RMatrix m2)
		{
			if (!CompareDimension(m1, m2))
			{
				throw new Exception("The dimensions of two matrices must be the same!");
			}
			var result = new RMatrix(m1.GetnRows, m1.GetnCols);
			for (var i = 0; i < m1.GetnRows; i++)
			{
				for (var j = 0; j < m1.GetnCols; j++) 
					result[i, j] = m1[i, j] + m2[i, j];
			}
			return result;
		}

		public static RMatrix operator -(RMatrix m)
		{
			for (var i = 0; i < m.GetnRows; i++)
			{
				for (var j = 0; j < m.GetnCols; j++)
				{
					m[i, j] = -m[i, j];
				}
			}
			return m;
		}

		public static RMatrix operator -(RMatrix m1, RMatrix m2)
		{
			if (!CompareDimension(m1, m2))
			{
				throw new Exception("The dimensions of two matrices must be the same!");
			}
			var result = new RMatrix(m1.GetnRows, m1.GetnCols);
			for (var i = 0; i < m1.GetnRows; i++)
			{
				for (var j = 0; j < m1.GetnCols; j++)
				{
					result[i, j] = m1[i, j] - m2[i, j];
				}
			}
			return result;
		}

		public static RMatrix operator *(RMatrix m, double d)
		{
			var result = new RMatrix(m.GetnRows, m.GetnCols);
			for (var i = 0; i < m.GetnRows; i++)
			{
				for (var j = 0; j < m.GetnCols; j++)
				{
					result[i, j] = m[i, j] * d;
				}
			}
			return result;
		}

		public static RMatrix operator *(double d, RMatrix m)
		{
			var result = new RMatrix(m.GetnRows, m.GetnCols);
			for (var i = 0; i < m.GetnRows; i++)
			{
				for (var j = 0; j < m.GetnCols; j++)
				{
					result[i, j] = m[i, j] * d;
				}
			}
			return result;
		}

		public static RMatrix operator /(RMatrix m, double d)
		{
			var result = new RMatrix(m.GetnRows, m.GetnCols);
			for (var i = 0; i < m.GetnRows; i++)
			{
				for (var j = 0; j < m.GetnCols; j++)
				{
					result[i, j] = m[i, j] / d;
				}
			}
			return result;
		}

		public static RMatrix operator /(double d, RMatrix m)
		{
			var result = new RMatrix(m.GetnRows, m.GetnCols);
			for (var i = 0; i < m.GetnRows; i++)
			{
				for (var j = 0; j < m.GetnCols; j++)
				{
					result[i, j] = m[i, j] / d;
				}
			}
			return result;
		}

		public static RMatrix operator *(RMatrix m1, RMatrix m2)
		{
			if (m1.GetnCols != m2.GetnRows)
			{
				throw new Exception("The numbers of columns of the" +
				                    " first matrix must be equal to the number of " +
				                    " rows of the second matrix!");
			}
			var result = new RMatrix(m1.GetnRows, m2.GetnCols);
			for (var i = 0; i < m1.GetnRows; i++)
			{
				for (var j = 0; j < m2.GetnCols; j++)
				{
					var tmp = result[i, j];
					for (var k = 0; k < result.GetnRows; k++) 
						tmp += m1[i, k] * m2[k, j];
					result[i, j] = tmp;
				}
			}
			return result;
		}

		public RMatrix GetTranspose()
		{
			var m = this;
			m.Transpose();
			return m;
		}

		public void Transpose()
		{
			var m = new RMatrix(this.colCount, this.rowCount);
			for (var i = 0; i < this.rowCount; i++)
			{
				for (var j = 0; j < this.colCount; j++)
				{
					m[j, i] = this.matrix[i, j];
				}
			}
			this = m;
		}

		public double GetTrace()
		{
			var sumOfDiag = 0.0;
			for (var i = 0; i < this.rowCount; i++)
			{
				if (i < this.colCount)
					sumOfDiag += this.matrix[i, i];
			}
			return sumOfDiag;
		}

		public bool IsSquared()
		{
			return this.rowCount == this.colCount;
		}

		public static bool CompareDimension(RMatrix m1, RMatrix m2)
		{
			return m1.GetnRows == m2.GetnRows && m1.GetnCols == m2.GetnCols;
		}

		public RVector GetRowVector(int m)
		{
			if (m < 0 || m > this.rowCount) throw new Exception("m-th row is out of range!");
			var rowVector = new RVector(this.colCount);
			for (var i = 0; i < this.colCount; i++) rowVector[i] = this.matrix[m, i];
			return rowVector;
		}

		public RVector GetColVector(int n)
		{
			if (n < 0 || n > this.colCount) throw new Exception("n-th col is out of range!");
			var colVector = new RVector(this.rowCount);
			for (var i = 0; i < this.rowCount; i++) colVector[i] = this.matrix[i, n];
			return colVector;
		}

		public RMatrix ReplaceRow(RVector vec, int m)
		{
			if (m < 0 || m > this.rowCount) throw new Exception("m-th row is out of range!");
			if (vec.GetVectorSize != this.colCount) throw new Exception("Vector ndim is out of range!");
			for (var i = 0; i < this.colCount; i++) this.matrix[m, i] = vec[i];
			return new RMatrix(this.matrix);
		}

		public RMatrix ReplaceCol(RVector vec, int n)
		{
			if (n < 0 || n > this.colCount) throw new Exception("n-th col is out of range!");
			if (vec.GetVectorSize != this.rowCount) throw new Exception("Vector ndim is out of range!");
			for (var i = 0; i < this.rowCount; i++) this.matrix[i, n] = vec[i];
			return new RMatrix(this.matrix);
		}

		public RMatrix SwapRMatrixow(int m, int n)
		{
			for (var i = 0; i < this.colCount; i++)
			{
				var temp = this.matrix[m, i];
				this.matrix[m, i] = this.matrix[n, i];
				this.matrix[n, i] = temp;
			}
			return new RMatrix(this.matrix);
		}

		public RMatrix SwapMatrixColumn(int m, int n)
		{
			for (var i = 0; i < this.rowCount; i++)
			{
				var temp = this.matrix[i, m];
				this.matrix[i, m] = this.matrix[i, n];
				this.matrix[i, n] = temp;
			}
			return new RMatrix(this.matrix);
		}

		public static RVector Transform(RMatrix mat, RVector vec)
		{
			var result = new RVector(vec.GetVectorSize);
			if (!mat.IsSquared())
			{
				throw new Exception("The matrix must be squared!");
			}
			if (mat.GetnCols != vec.GetVectorSize)
			{
				throw new Exception("The ndim of the vector must be equal "
				                    + "to the number of cols of the matrix!");
			}
			for (var i = 0; i < mat.GetnRows; i++)
			{
				result[i] = 0.0;
				for (var j = 0; j < mat.GetnCols; j++) result[i] += mat[i, j] * vec[j];
			}
			return result;
		}

		public static RVector Transform(RVector vec, RMatrix mat)
		{
			var result = new RVector(vec.GetVectorSize);
			if (!mat.IsSquared()) throw new Exception("The matrix must be squared!");
			if (mat.GetnRows != vec.GetVectorSize) throw new Exception("The ndim of the vector must be equal " + "to the number of rows of the matrix!");
			for (var i = 0; i < mat.GetnRows; i++)
			{
				result[i] = 0.0;
				for (var j = 0; j < mat.GetnCols; j++) result[i] += vec[j] * mat[j, i];
			}
			return result;
		}

		public static RMatrix Transform(RVector v1, RVector v2)
		{
			if (v1.GetVectorSize != v2.GetVectorSize)
			{
				throw new Exception("The vectors must have the same ndim!");
			}
			var result = new RMatrix(v1.GetVectorSize, v1.GetVectorSize);
			for (var i = 0; i < v1.GetVectorSize; i++)
			{
				for (var j = 0; j < v1.GetVectorSize; j++) result[j, i] = v1[i] * v2[j];
			}
			return result;
		}

		public static double Determinant(RMatrix mat)
		{
			var result = 0.0;
			if (!mat.IsSquared())
			{
				throw new Exception("The matrix must be squared!");
			}
			if (mat.GetnRows == 1)
				result = mat[0, 0];
			else
			{
				for (var i = 0; i < mat.GetnRows; i++) result += System.Math.Pow(-1, i) * mat[0, i] * Determinant(Minor(mat, 0, i));
			}
			return result;
		}

		public static RMatrix Minor(RMatrix mat, int row, int col)
		{
			var mm = new RMatrix(mat.GetnRows - 1, mat.GetnCols - 1);
			var ii = 0;
			for (var i = 0; i < mat.GetnRows; i++)
			{
				if (i == row)
					continue;
				var jj = 0;
				for (var j = 0; j < mat.GetnCols; j++)
				{
					if (j == col)
						continue;
					mm[ii, jj] = mat[i, j];
					jj++;
				}
				ii++;
			}
			return mm;
		}

		public static RMatrix Adjoint(RMatrix mat)
		{
			if (!mat.IsSquared())
			{
				throw new Exception("The matrix must be squared!");
			}
			var ma = new RMatrix(mat.GetnRows, mat.GetnCols);
			for (var i = 0; i < mat.GetnRows; i++)
			{
				for (var j = 0; j < mat.GetnCols; j++) ma[i, j] = System.Math.Pow(-1, i + j) * (Determinant(Minor(mat, i, j)));
			}
			return ma.GetTranspose();
		}

		public static RMatrix Inverse(RMatrix mat)
		{
			if (System.Math.Abs(Determinant(mat) ) < Constants.Epsilon) throw new Exception("Cannot inverse a matrix with a zero determinant!");
			return (Adjoint(mat) / Determinant(mat));
		}
	}
}