﻿using System;
using System.ComponentModel;
using System.Linq;

namespace Telerik.Windows.Documents.Spreadsheet.Maths
{
	/// <summary>
	/// A  real-valued vector.
	/// </summary>    
    internal struct RVector 
	#if WPF
	: ICloneable
	#endif
	{
		private readonly int dimension;
		private double[] vector;

		/// <summary>
		/// Initializes a new instance of the <see cref="RVector" /> struct.
		/// </summary>
		/// <param name="dimension">The dimension.</param>
		public RVector(int dimension)
		{
			this.dimension = dimension;
			this.vector = new double[dimension];

			// initialization
			for (var i = 0; i < dimension; i++) this.vector[i] = 0.0;
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="RVector" /> struct.
		/// </summary>
		/// <param name="vector">The vector.</param>
		public RVector(double[] vector)
		{
			this.dimension = vector.Length;
			this.vector = vector;
		}

		/// <summary>
		/// Gets or sets the <see cref="System.Double" /> with the specified i.
		/// </summary>
		/// <value>
		/// The <see cref="System.Double" />.
		/// </value>
		/// <param name="i">The entry number.</param>
		/// <returns></returns>
		/// <exception cref="System.Exception">Requested vector index is out of range!</exception>
		public double this[int i]
		{
			get
			{
				if (i < 0 || i > this.dimension) throw new ArgumentOutOfRangeException("i");
				return this.vector[i];
			}
			set { this.vector[i] = value; }
		}

		public int GetVectorSize
		{
			get { return this.dimension; }
		}

		/// <summary>
		/// Clones this instance.
		/// </summary>
		/// <returns></returns>
		public RVector Clone()
		{
			var v = new RVector(this.vector) { vector = (double[])this.vector.Clone() };
			return v;
		}
#if WPF
		object ICloneable.Clone()
		{
			return this.Clone();
		}
#endif
		public RVector SwapVectorEntries(int m, int n)
		{
			var temp = this.vector[m];
			this.vector[m] = this.vector[n];
			this.vector[n] = temp;
			return new RVector(this.vector);
		}

		/// <summary>
		/// Returns a <see cref="System.String" /> that represents this instance.
		/// </summary>
		/// <returns>
		/// A <see cref="System.String" /> that represents this instance.
		/// </returns>
		public override string ToString()
		{
			return this.vector.Length > 5 ? "(" + String.Join(",", this.vector.Take(5)) + "...)" : "(" + String.Join(",", this.vector) + ")";
		}

		public override bool Equals(object obj)
		{
			return (obj is RVector) && this.Equals((RVector)obj);
		}

		public bool Equals(RVector v)
		{
			return this.vector == v.vector;
		}

		public override int GetHashCode()
		{
			return this.vector.GetHashCode();
		}

		public static bool operator ==(RVector v1, RVector v2)
		{
			return v1.Equals(v2);
		}

		public static bool operator !=(RVector v1, RVector v2)
		{
			return !v1.Equals(v2);
		}

		public static RVector operator +(RVector v)
		{
			return v;
		}

		public static RVector operator +(RVector v1, RVector v2)
		{
			var result = new RVector(v1.dimension);
			for (var i = 0; i < v1.dimension; i++) result[i] = v1[i] + v2[i];
			return result;
		}

		public static RVector operator -(RVector v)
		{
			var result = new double[v.dimension];
			for (var i = 0; i < v.dimension; i++) result[i] = -v[i];
			return new RVector(result);
		}

		public static RVector operator -(RVector v1, RVector v2)
		{
			var result = new RVector(v1.dimension);
			for (var i = 0; i < v1.dimension; i++) result[i] = v1[i] - v2[i];
			return result;
		}

		public static RVector operator *(RVector v, double d)
		{
			var result = new RVector(v.dimension);
			for (var i = 0; i < v.dimension; i++) result[i] = v[i] * d;
			return result;
		}

		public static RVector operator *(double d, RVector v)
		{
			var result = new RVector(v.dimension);
			for (var i = 0; i < v.dimension; i++) result[i] = d * v[i];
			return result;
		}

		public static RVector operator /(RVector v, double d)
		{
			var result = new RVector(v.dimension);
			for (var i = 0; i < v.dimension; i++) result[i] = v[i] / d;
			return result;
		}

		public static RVector operator /(double d, RVector v)
		{
			var result = new RVector(v.dimension);
			for (var i = 0; i < v.dimension; i++) result[i] = v[i] / d;
			return result;
		}

		public static double DotProduct(RVector v1, RVector v2)
		{
			var result = 0.0;
			for (var i = 0; i < v1.dimension; i++) result += v1[i] * v2[i];
			return result;
		}

		public double GetNorm()
		{
			var result = 0.0;
			for (var i = 0; i < this.dimension; i++) result += this.vector[i] * this.vector[i];
			return System.Math.Sqrt(result);
		}

		public double GetNormSquare()
		{
			var result = 0.0;
			for (var i = 0; i < this.dimension; i++) result += this.vector[i] * this.vector[i];
			return result;
		}

		public void Normalize()
		{
			var norm = this.GetNorm();
			if (System.Math.Abs(norm) < Constants.Epsilon) throw new Exception("Tried to normalize a vector with norm of zero!");
			for (var i = 0; i < this.dimension; i++) this.vector[i] /= norm;
		}

		public RVector GetUnitVector()
		{
			var result = new RVector(this.vector);
			result.Normalize();
			return result;
		}

		public static RVector CrossProduct(RVector v1, RVector v2)
		{
			if (v1.dimension != 3) throw new Exception("Vector v1 must be 3 dimensional!");
			if (v2.dimension != 3) throw new Exception("Vector v2 must be 3 dimensional!");
			var result = new RVector(3);
			result[0] = v1[1] * v2[2] - v1[2] * v2[1];
			result[1] = v1[2] * v2[0] - v1[0] * v2[2];
			result[2] = v1[0] * v2[1] - v1[1] * v2[0];
			return result;
		}
	}
}