﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    internal class CellAreaReference
    {
        #region Fields

        private readonly CellReference fromCellReference;
        private readonly CellReference toCellReference;
        private readonly bool useBothReferencesInString;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the start cell reference.
        /// </summary>
        /// <value>The start cell reference.</value>
        public CellReference FromCellReference
        {
            get
            {
                return this.fromCellReference;
            }
        }

        /// <summary>
        /// Gets the end cell reference.
        /// </summary>
        /// <value>The end cell reference.</value>
        public CellReference ToCellReference
        {
            get
            {
                return this.toCellReference;
            }
        }

        /// <summary>
        /// Gets the rows count.
        /// </summary>
        /// <value>The rows count.</value>
        public int RowsCount
        {
            get
            {
                return toCellReference.ActualRowIndex - fromCellReference.ActualRowIndex + 1;
            }
        }

        /// <summary>
        /// Gets the columns count.
        /// </summary>
        /// <value>The columns count.</value>
        public int ColumnsCount
        {
            get
            {
                return toCellReference.ActualColumnIndex - fromCellReference.ActualColumnIndex + 1;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the range is within the valid boundaries.
        /// </summary>
        /// <value>The value indicating whether the range is within the valid boundaries.</value>
        public bool IsValid
        {
            get
            {
                return TelerikHelper.IsValidRowIndex(this.fromCellReference.ActualRowIndex) &&
                       TelerikHelper.IsValidRowIndex(this.toCellReference.ActualRowIndex) &&
                       TelerikHelper.IsValidColumnIndex(this.fromCellReference.ActualColumnIndex) &&
                       TelerikHelper.IsValidColumnIndex(this.toCellReference.ActualColumnIndex);
            }
        }

        #endregion


        #region Constructors

        internal CellAreaReference(CellReference cellReference)
            : this(cellReference, cellReference)
        {
        }

        internal CellAreaReference(CellReference fromCellReference, CellReference toCellReference, bool useBothReferencesInString = false)
        {
            Guard.ThrowExceptionIfNull(fromCellReference, "fromCellReference");
            Guard.ThrowExceptionIfNull(toCellReference, "toCellReference");

            this.fromCellReference = CellReference.GetTopLeftCellReference(fromCellReference, toCellReference);
            this.toCellReference = CellReference.GetBottomRightCellReference(fromCellReference, toCellReference);

            this.useBothReferencesInString = useBothReferencesInString;
        }

        #endregion


        #region Methods

        internal static CellAreaReference CreateRange(CellAreaReference firstRange, CellAreaReference secondRange)
        {
            Guard.ThrowExceptionIfNull(firstRange, "firstRange");
            Guard.ThrowExceptionIfNull(secondRange, "secondRange");

            CellReference validFromCellReference = CellReference.GetTopLeftCellReference(firstRange.FromCellReference, secondRange.FromCellReference);
            CellReference validToCellReference = CellReference.GetBottomRightCellReference(firstRange.ToCellReference, secondRange.ToCellReference);

            return new CellAreaReference(validFromCellReference, validToCellReference);
        }

        internal static CellAreaReference CreateIntersection(CellAreaReference firstRange, CellAreaReference secondRange)
        {
            Guard.ThrowExceptionIfNull(firstRange, "firstRange");
            Guard.ThrowExceptionIfNull(secondRange, "secondRange");

            if (AreRangesIntersecting(firstRange, secondRange))
            {
                CellReference validFromCellReference = CellReference.GetTopLeftCellReference(firstRange.ToCellReference, secondRange.ToCellReference);
                CellReference validToCellReference = CellReference.GetBottomRightCellReference(firstRange.FromCellReference, secondRange.FromCellReference);

                return new CellAreaReference(validFromCellReference, validToCellReference);
            }

            return null;
        }

        internal static CellAreaReference CreateFromCellName(string cellName, CellIndex cellIndex)
        {
            return CreateFromCellName(cellName, cellIndex.RowIndex, cellIndex.ColumnIndex);
        }

        internal static CellAreaReference CreateFromCellName(string cellName, int rowIndex, int columnIndex)
        {
            Guard.ThrowExceptionIfNull(cellName, "cellName");
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            CellReference cellReference = CellReference.CreateFromCellName(cellName, rowIndex, columnIndex);

            if (cellReference != null)
            {
                return new CellAreaReference(cellReference);
            }

            return null;
        }

        internal static CellAreaReference CreateFromTwoCellNames(string leftCellName, string rightCellName, int rowIndex, int columnIndex, bool useTwoReferences)
        {
            Guard.ThrowExceptionIfNullOrEmpty(leftCellName, "leftCellName");
            Guard.ThrowExceptionIfNullOrEmpty(rightCellName, "rightCellName");
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            CellReference fromCellReference = CellReference.CreateFromCellName(leftCellName, rowIndex, columnIndex);
            CellReference toCellReference = CellReference.CreateFromCellName(rightCellName, rowIndex, columnIndex);
            if (leftCellName != null && toCellReference != null)
            {
                return new CellAreaReference(fromCellReference, toCellReference, useTwoReferences);
            }

            return null;
        }

        private static bool AreRangesIntersecting(CellAreaReference range1, CellAreaReference range2)
        {
            return AreCellReferencesInRange(range1, range2) || AreCellReferencesInRange(range2, range1);
        }

        private static bool AreCellReferencesInRange(CellAreaReference range1, CellAreaReference range2)
        {
            int fromColumnIndex = range1.FromCellReference.ActualColumnIndex;
            int toColumnIndex = range1.ToCellReference.ActualColumnIndex;
            int fromRowIndex = range1.FromCellReference.ActualRowIndex;
            int toRowIndex = range1.ToCellReference.ActualRowIndex;

            return IsCellReferenceInRange(fromRowIndex, fromColumnIndex, range2) ||
                IsCellReferenceInRange(fromRowIndex, toColumnIndex, range2) ||
                IsCellReferenceInRange(toRowIndex, toColumnIndex, range2) ||
                IsCellReferenceInRange(toRowIndex, fromColumnIndex, range2);
        }

        private static bool IsCellReferenceInRange(int rowIndex, int columnIndex, CellAreaReference cellReferenceRange)
        {
            return cellReferenceRange.FromCellReference.ActualColumnIndex <= columnIndex &&
                cellReferenceRange.ToCellReference.ActualColumnIndex >= columnIndex &&
                cellReferenceRange.FromCellReference.ActualRowIndex <= rowIndex &&
                cellReferenceRange.ToCellReference.ActualRowIndex >= rowIndex;
        }

        internal CellAreaReference CloneAndTranslate(int rowIndex, int columnIndex, bool onOverflowStartOver)
        {
            return new CellAreaReference(
                this.FromCellReference.CloneAndTranslate(rowIndex, columnIndex, onOverflowStartOver),
                this.ToCellReference.CloneAndTranslate(rowIndex, columnIndex, onOverflowStartOver), this.useBothReferencesInString);
        }

        internal CellAreaReference CloneAndTranslateUponInsertRemoveCells(ExpressionCloneAndTranslateContext context, int rangeFromIndex, int rangeToIndex, int referenceFromIndex, int referenceToIndex,
            bool translateSource, bool translateTarget)
        {
            int rowColumnCount = context.IsHorizontal ? context.Range.ColumnCount : context.Range.RowCount;

            if (context.IsRemove)
            {
                if (translateTarget && rangeFromIndex <= referenceFromIndex && referenceToIndex <= rangeToIndex)
                {
                    return null;
                }

                return this.CloneAndTranslateUponRemoveCells(translateSource, translateTarget, context.IsHorizontal,
                    rowColumnCount, rangeFromIndex, rangeToIndex, referenceFromIndex, referenceToIndex, context.OnOverflowStartOver);
            }

            return this.CloneAndTranslateUponInsertCells(translateSource, translateTarget, context.IsHorizontal, rowColumnCount, rangeFromIndex, referenceFromIndex, context.OnOverflowStartOver);
        }

        internal CellAreaReference CloneAndTranslateUponRemoveCells(bool translateSource, bool translateTarget, bool isHorizontal, int rowColumnCount,
            int rangeFromIndex, int rangeToIndex, int referenceFromIndex, int referenceToIndex, bool onOverflowStartOver)
        {
            int fromReferenceIndexShift = 0;
            int fromReferenceOffsetShift = 0;
            int toReferenceIndexShift = 0;
            int toReferenceOffsetShift = 0;

            if (translateSource && translateTarget)
            {
                fromReferenceIndexShift = -rowColumnCount;
                toReferenceIndexShift = -rowColumnCount;

                if (rangeToIndex >= referenceFromIndex)
                {
                    if (rangeFromIndex >= referenceFromIndex)
                    {
                        fromReferenceOffsetShift = rowColumnCount;
                    }
                    else
                    {
                        fromReferenceOffsetShift = referenceFromIndex - rangeToIndex;
                        toReferenceOffsetShift = fromReferenceOffsetShift;
                    }
                }
            }
            else if (translateTarget)
            {
                if (rangeToIndex < referenceFromIndex)
                {
                    fromReferenceOffsetShift = -rowColumnCount;
                    toReferenceOffsetShift = -rowColumnCount;
                }
                else
                {
                    if (rangeFromIndex >= referenceFromIndex)
                    {
                        toReferenceOffsetShift = rangeFromIndex - Math.Min(referenceToIndex, rangeToIndex) - 1;
                    }
                    else
                    {
                        fromReferenceOffsetShift = rangeFromIndex - referenceFromIndex;
                        toReferenceOffsetShift = -rowColumnCount;
                    }
                }
            }
            else if (translateSource)
            {
                fromReferenceIndexShift = -rowColumnCount;
                fromReferenceOffsetShift = rowColumnCount;
                toReferenceIndexShift = -rowColumnCount;
                toReferenceOffsetShift = rowColumnCount;
            }

            return new CellAreaReference(this.fromCellReference.CloneAndTranslateUponInsertRemoveCells(isHorizontal, fromReferenceIndexShift, fromReferenceOffsetShift, onOverflowStartOver),
                this.toCellReference.CloneAndTranslateUponInsertRemoveCells(isHorizontal, toReferenceIndexShift, toReferenceOffsetShift, onOverflowStartOver),
                this.useBothReferencesInString);
        }

        internal CellAreaReference CloneAndTranslateUponInsertCells(bool translateSource, bool translateTarget, bool isHorizontal, int rowColumnCount,
            int rangeFromIndex, int referenceFromIndex, bool onOverflowStartOver)
        {
            int fromReferenceIndexShift = 0;
            int fromReferenceOffsetShift = 0;
            int toReferenceIndexShift = 0;
            int toReferenceOffsetShift = 0;

            if (translateSource && translateTarget)
            {
                fromReferenceIndexShift = rowColumnCount;
                toReferenceIndexShift = rowColumnCount;
               
                if (rangeFromIndex > referenceFromIndex)
                {
                    fromReferenceOffsetShift = -rowColumnCount;
                }
            }
            else if (translateTarget)
            {
                toReferenceOffsetShift = rowColumnCount;
                
                if (rangeFromIndex <= referenceFromIndex)
                {
                    fromReferenceOffsetShift = rowColumnCount;
                }
            }
            else if (translateSource)
            {
                fromReferenceIndexShift = rowColumnCount;
                fromReferenceOffsetShift = -rowColumnCount;
                toReferenceIndexShift = rowColumnCount;
                toReferenceOffsetShift = -rowColumnCount;
            }

            return new CellAreaReference(this.fromCellReference.CloneAndTranslateUponInsertRemoveCells(isHorizontal, fromReferenceIndexShift, fromReferenceOffsetShift, onOverflowStartOver),
                this.toCellReference.CloneAndTranslateUponInsertRemoveCells(isHorizontal, toReferenceIndexShift, toReferenceOffsetShift, onOverflowStartOver),
                this.useBothReferencesInString);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            CellAreaReference cellReferenceRange = obj as CellAreaReference;
            if (cellReferenceRange == null)
            {
                return false;
            }

            return this.FromCellReference.Equals(cellReferenceRange.FromCellReference) &&
                   this.ToCellReference.Equals(cellReferenceRange.ToCellReference);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(this.FromCellReference.GetHashCode(), this.ToCellReference.GetHashCode());
        }

        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current
        /// <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
            if (!this.IsValid)
            {
                return ErrorExpressions.ReferenceError.Value;
            }

            if (!this.useBothReferencesInString && this.FromCellReference.Equals(this.ToCellReference))
            {
                return NameConverter.ConvertCellReferenceToName(this.FromCellReference);
            }

            return NameConverter.ConvertCellReferenceRangeToName(this.FromCellReference, this.ToCellReference);
        }

        #endregion
    }
}
