﻿using System;
using System.Windows.Media;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Theming;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a cell border.
    /// </summary>
    public class CellBorder
    {
        #region Constants

        private static readonly double DefaultThickness = 1;

        /// <summary>
        /// The default cell border.
        /// </summary>
        public static readonly CellBorder Default = new CellBorder(CellBorderStyle.None, new ThemableColor(Colors.LightGray));

        #endregion


        #region Fields

        private readonly CellBorderStyle style;
        private readonly ThemableColor color;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the style of the border.
        /// </summary>
        /// <value>The style.</value>
        public CellBorderStyle Style
        {
            get
            {
                return this.style;
            }
        }

        /// <summary>
        /// Gets the color.
        /// </summary>
        /// <value>The color.</value>
        public ThemableColor Color
        {
            get
            {
                return this.color;
            }
        }

        /// <summary>
        /// Gets the thickness.
        /// </summary>
        /// <value>The thickness.</value>
        public double Thickness
        {
            get
            {
                return GetThicknessByBorderStyle(this.Style);
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="CellBorder" /> class.
        /// </summary>
        /// <param name="style">The style.</param>
        /// <param name="color">The color.</param>
        public CellBorder(CellBorderStyle style, ThemableColor color)
        {
            this.style = style;

            if (CellBorder.Default != null &&
                this.style == CellBorderStyle.None &&
                color != new ThemableColor(Colors.Transparent))
            {
                color = CellBorder.Default.Color;
            }

            this.color = color;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Implements the equals operator.
        /// </summary>
        /// <param name="first">The first value.</param>
        /// <param name="second">The second value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
        public static bool operator ==(CellBorder first, CellBorder second)
        {
            if (object.ReferenceEquals(first, second))
            {
                return true;
            }

            if ((object)first == null)
            {
                return false;
            }

            return first.Equals(second);
        }

        /// <summary>
        /// Implements the not equals operator.
        /// </summary>
        /// <param name="first">The first value.</param>
        /// <param name="second">The second value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
        public static bool operator !=(CellBorder first, CellBorder second)
        {
            return !(first == second);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            CellBorder other = obj as CellBorder;

            if (other == null)
            {
                return false;
            }

            return this.style == other.style &&
                TelerikHelper.EqualsOfT(this.color, other.color);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(this.style.GetHashCode(), this.color.GetHashCode());
        }

        /// <summary>
        /// Gets the border with higher priority according to the color scheme.
        /// </summary>
        /// <param name="leftBorder">The left border.</param>
        /// <param name="rightBorder">The right border.</param>
        /// <param name="colorScheme">The color scheme.</param>
        /// <returns>The border with higher priority according to the color scheme.</returns>
        public static CellBorder GetWithMaxPriority(CellBorder leftBorder, CellBorder rightBorder, ThemeColorScheme colorScheme)
        {
            Guard.ThrowExceptionIfNull(leftBorder, "leftBorder");
            Guard.ThrowExceptionIfNull(rightBorder, "rightBorder");
            Guard.ThrowExceptionIfNull(colorScheme, "colorScheme");

            return leftBorder.GetZIndex(colorScheme) > rightBorder.GetZIndex(colorScheme) ? leftBorder : rightBorder;
        }

        internal int GetZIndex(ThemeColorScheme colorScheme)
        {
            if (CellBorder.Default.Equals(this))
            {
                return 0;
            }

            int result = 255 * 5 * (int)Math.Round(this.Thickness);

            Color color = this.Color.GetActualValue(colorScheme);
            result += (color.R + color.B * 2 + color.G * 4);

            return result;
        }

        /// <summary>
        /// Gets the thickness by border style.
        /// </summary>
        /// <param name="borderStyle">The border style.</param>
        /// <returns>The thickness.</returns>
        public static double GetThicknessByBorderStyle(CellBorderStyle borderStyle)
        {
            switch (borderStyle)
            {
                case CellBorderStyle.None:
                case CellBorderStyle.DashDot:
                case CellBorderStyle.DashDotDot:
                case CellBorderStyle.Dashed:
                case CellBorderStyle.Dotted:
                case CellBorderStyle.Hair:
                case CellBorderStyle.Thin:
                    return DefaultThickness;

                // TODO:
                case CellBorderStyle.Double:
                case CellBorderStyle.Medium:
                case CellBorderStyle.SlantDashDot:
                case CellBorderStyle.MediumDashDot:
                case CellBorderStyle.MediumDashDotDot:
                case CellBorderStyle.MediumDashed:
                    return DefaultThickness * 2;

                case CellBorderStyle.Thick:
                    return DefaultThickness * 3;

                default:
                    return DefaultThickness;
            }
        }

        #endregion
    }
}