﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents all the borders of a cell.
    /// </summary>
    public class CellBorders
    {
        #region Constants

        /// <summary>
        /// The default borders of a cell.
        /// </summary>
        public static readonly CellBorders Default = new CellBorders(CellBorder.Default);

        #endregion


        #region Properties

        /// <summary>
        /// Gets or sets the left border.
        /// </summary>
        /// <value>The left border.</value>
        public CellBorder Left { get; set; }

        /// <summary>
        /// Gets or sets the top border.
        /// </summary>
        /// <value>The top border.</value>
        public CellBorder Top { get; set; }

        /// <summary>
        /// Gets or sets the right border.
        /// </summary>
        /// <value>The right border.</value>
        public CellBorder Right { get; set; }

        /// <summary>
        /// Gets or sets the bottom border.
        /// </summary>
        /// <value>The bottom border.</value>
        public CellBorder Bottom { get; set; }

        /// <summary>
        /// Gets or sets the inside horizontal border.
        /// </summary>
        /// <value>The inside horizontal border.</value>
        public CellBorder InsideHorizontal { get; set; }

        /// <summary>
        /// Gets or sets the inside vertical border.
        /// </summary>
        /// <value>The inside vertical border.</value>
        public CellBorder InsideVertical { get; set; }

        /// <summary>
        /// Gets or sets the diagonal up border.
        /// </summary>
        /// <value>The diagonal up border.</value>
        public CellBorder DiagonalUp { get; set; }

        /// <summary>
        /// Gets or sets the diagonal down border.
        /// </summary>
        /// <value>The diagonal down border.</value>
        public CellBorder DiagonalDown { get; set; }

		#endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="CellBorders" /> class.
        /// </summary>
        public CellBorders()
        {
		}

        /// <summary>
        /// Initializes a new instance of the <see cref="CellBorders" /> class.
        /// </summary>
        /// <param name="all">The cell border to be applied on all sides of the cell.</param>
        public CellBorders(CellBorder all) 
            : this(all, all, all, all, all, all, all, all)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CellBorders" /> class.
        /// </summary>
        /// <param name="left">The left border.</param>
        /// <param name="top">The top border.</param>
        /// <param name="right">The right border.</param>
        /// <param name="bottom">The bottom border.</param>
        /// <param name="insideHorizontal">The inside horizontal border.</param>
        /// <param name="insideVertical">The inside vertical border.</param>
        /// <param name="diagonalUp">The diagonal up border.</param>
        /// <param name="diagonalDown">The diagonal down border.</param>
        public CellBorders(
            CellBorder left, CellBorder top,
            CellBorder right, CellBorder bottom,
            CellBorder insideHorizontal, CellBorder insideVertical,
            CellBorder diagonalUp, CellBorder diagonalDown)
		{
			this.Left = left;
			this.Top = top;
			this.Right = right;
			this.Bottom = bottom;
			this.InsideHorizontal = insideHorizontal;
			this.InsideVertical = insideVertical;
			this.DiagonalUp = diagonalUp;
			this.DiagonalDown = diagonalDown;
		}

		#endregion


		#region Methods

        /// <summary>
        /// Implements the equals operator.
        /// </summary>
        /// <param name="first">The first value.</param>
        /// <param name="second">The second value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
        public static bool operator ==(CellBorders first, CellBorders second)
        {
            if (object.ReferenceEquals(first, second))
            {
                return true;
            }

            if ((object)first == null)
            {
                return false;
            }

            return first.Equals(second);
        }

        /// <summary>
        /// Implements the not equals operator.
        /// </summary>
        /// <param name="first">The first value.</param>
        /// <param name="second">The second value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
        public static bool operator !=(CellBorders first, CellBorders second)
        {
            return !(first == second);
        }

        internal void MergeWith(CellBorders other)
        {
            Func<CellBorder, CellBorder, CellBorder> nullIfDifferent =
                (firstBorder, secondBorder) =>
                {
                    return (firstBorder == secondBorder) ? firstBorder : null;
                };

            this.Left = nullIfDifferent(this.Left, other.Left);
            this.Top = nullIfDifferent(this.Top, other.Top);
            this.Right = nullIfDifferent(this.Right, other.Right);
            this.Bottom = nullIfDifferent(this.Bottom, other.Bottom);
            this.InsideHorizontal = nullIfDifferent(this.InsideHorizontal, other.InsideHorizontal);
            this.InsideVertical = nullIfDifferent(this.InsideVertical, other.InsideVertical);
            this.DiagonalUp = nullIfDifferent(this.DiagonalUp, other.DiagonalUp);
            this.DiagonalDown = nullIfDifferent(this.DiagonalDown, other.DiagonalDown);
        }

        /// <summary>
        /// Creates an outline.
        /// </summary>
        /// <param name="all">The border to be applied to the top, right, bottom and left sides of the cell.</param>
        /// <returns>The outline.</returns>
        public static CellBorders CreateOutline(CellBorder all)
        {
			return CreateOutline(all, all, all, all);
		}

        /// <summary>
        /// Creates an outline.
        /// </summary>
        /// <param name="left">The left border.</param>
        /// <param name="top">The top border.</param>
        /// <param name="right">The right border.</param>
        /// <param name="bottom">The bottom border.</param>
        /// <returns>The outline.</returns>
        public static CellBorders CreateOutline(
            CellBorder left, CellBorder top,
            CellBorder right, CellBorder bottom)
		{
			CellBorders outline = new CellBorders();
			outline.Left = left;
			outline.Top = top;
			outline.Right = right;
			outline.Bottom = bottom;

			return outline;
		}

        /// <summary>
        /// Creates an inside group of borders.
        /// </summary>
        /// <param name="all">The border to be applied to the inner borders of a cell.</param>
        /// <returns>The inside group of borders.</returns>
        public static CellBorders CreateInside(CellBorder all)
        {
			return CreateInside(all, all);
		}

        /// <summary>
        /// Creates an inside group of borders.
        /// </summary>
        /// <param name="insideHorizontal">The inside horizontal border.</param>
        /// <param name="insideVertical">The inside vertical border.</param>
        /// <returns>The inside group of borders.</returns>
        public static CellBorders CreateInside(
            CellBorder insideHorizontal, CellBorder insideVertical)
		{
			CellBorders inside = new CellBorders();
			inside.InsideHorizontal = insideHorizontal;
			inside.InsideVertical = insideVertical;

			return inside;
		}

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
			CellBorders other = obj as CellBorders;

			if (other == null)
			{
				return false;
			}

			return TelerikHelper.EqualsOfT(this.Left, other.Left) &&
				TelerikHelper.EqualsOfT(this.Top, other.Top) &&
				TelerikHelper.EqualsOfT(this.Right, other.Right) &&
				TelerikHelper.EqualsOfT(this.Bottom, other.Bottom) &&
				TelerikHelper.EqualsOfT(this.InsideHorizontal, other.InsideHorizontal) &&
				TelerikHelper.EqualsOfT(this.InsideVertical, other.InsideVertical) &&
				TelerikHelper.EqualsOfT(this.DiagonalUp, other.DiagonalUp) &&
				TelerikHelper.EqualsOfT(this.DiagonalDown, other.DiagonalDown);
		}

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                this.Left.GetHashCodeOrZero(),
                this.Top.GetHashCodeOrZero(),
                this.Right.GetHashCodeOrZero(),
                this.Bottom.GetHashCodeOrZero(),
                this.InsideHorizontal.GetHashCodeOrZero(),
                this.InsideVertical.GetHashCodeOrZero(),
                this.DiagonalUp.GetHashCodeOrZero(),
                this.DiagonalDown.GetHashCodeOrZero());
		}

		#endregion
	}
}