﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a cell index.
    /// </summary>
    public class CellIndex : CellIndexBase
    {
        #region Fields

        private readonly int rowIndex;
        private readonly int columnIndex;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the index of the row.
        /// </summary>
        /// <value>The index of the row.</value>
        public override int RowIndex
        {
            get
            {
                return this.rowIndex;
            }
        }

        /// <summary>
        /// Gets the index of the column.
        /// </summary>
        /// <value>The index of the column.</value>
        public override int ColumnIndex
        {
            get
            {
                return this.columnIndex;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="CellIndex" /> class.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        public CellIndex(int rowIndex, int columnIndex)
        {
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            this.rowIndex = rowIndex;
            this.columnIndex = columnIndex;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Creates a new cell index based on the current instance and a specified offset.
        /// </summary>
        /// <param name="rowOffset">The row offset.</param>
        /// <param name="columnOffset">The column offset.</param>
        /// <returns>The new cell index.</returns>
        public CellIndex Offset(int rowOffset, int columnOffset)
        {
            int rowIndex = this.rowIndex + rowOffset;
            int columnIndex = this.columnIndex + columnOffset;
            if (!TelerikHelper.IsValidRowIndex(rowIndex) ||
                !TelerikHelper.IsValidColumnIndex(columnIndex))
            {
                return null;
            }

            return new CellIndex(rowIndex, columnIndex);
        }

        internal CellRange ToCellRange()
        {
            return new CellRange(this, this);
        }

        internal static CellIndex Min(CellIndex first, CellIndex second)
        {
            return new CellIndex(Math.Min(first.RowIndex, second.RowIndex), Math.Min(first.ColumnIndex, second.ColumnIndex));
        }

        internal static CellIndex Max(CellIndex first, CellIndex second)
        {
            return new CellIndex(Math.Max(first.RowIndex, second.RowIndex), Math.Max(first.ColumnIndex, second.ColumnIndex));
        }

        internal static CellIndex RestrictCellIndex(CellIndex cellIndex, SizeI size)
        {
            int maxRowIndex = size.Height - 1;
            int maxColumnIndex = size.Width - 1;

            Guard.ThrowExceptionIfNull(cellIndex, "cellIndex");
            Guard.ThrowExceptionIfOutOfRange(0, SpreadsheetDefaultValues.RowCount - 1, maxRowIndex, "maxRowIndex");
            Guard.ThrowExceptionIfOutOfRange(0, SpreadsheetDefaultValues.ColumnCount - 1, maxColumnIndex, "maxColumnIndex");

            return new CellIndex(
                Math.Min(cellIndex.RowIndex, maxRowIndex),
                Math.Min(cellIndex.ColumnIndex, maxColumnIndex));
        }

        /// <summary>
        /// Creates a numeric representation of the current index.
        /// </summary>
        /// <returns>A numeric representation of the current index.</returns>
        public long ToNumber()
        {
            return (long)this.RowIndex * SpreadsheetDefaultValues.ColumnCount + this.ColumnIndex;
        }

        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current
        /// <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
            return string.Format("{0},{1}", this.RowIndex, this.ColumnIndex);
        }

        #endregion
    }
}