﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a base for the cell index and position classes.
    /// </summary>
    public abstract class CellIndexBase : IComparable<CellIndexBase>
    {
		#region Properties

        /// <summary>
        /// Gets the index of the row.
        /// </summary>
        /// <value>The index of the row.</value>
        public abstract int RowIndex { get; }

        /// <summary>
        /// Gets the index of the column.
        /// </summary>
        /// <value>The index of the column.</value>
        public abstract int ColumnIndex { get; }

		#endregion


		#region Operators & Overrides

        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current
        /// <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
			return string.Format("[RowIndex={0}, ColumnIndex={1}]", this.RowIndex, this.ColumnIndex);
		}

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
			CellIndexBase other = obj as CellIndexBase;
			if (other == null)
			{
				return false;
			}

			return (this == other);
		}

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
			return TelerikHelper.CombineHashCodes(this.RowIndex.GetHashCode(), this.ColumnIndex.GetHashCode());
		}

        /// <summary>
        /// Implements the equals operator.
        /// </summary>
        /// <param name="left">The left value.</param>
        /// <param name="right">The right value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
		public static bool operator ==(CellIndexBase left, CellIndexBase right)
		{
			if (object.ReferenceEquals(left, right))
			{
				return true;
			}

			if ((object)left == null)
			{
				return (object)right == null;
			}

			if ((object)right == null)
			{
				return false;
			}

			return left.RowIndex == right.RowIndex && left.ColumnIndex == right.ColumnIndex;
		}

        /// <summary>
        /// Implements the not equals operator.
        /// </summary>
        /// <param name="left">The left value.</param>
        /// <param name="right">The right value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
		public static bool operator !=(CellIndexBase left, CellIndexBase right)
		{
			return !(left == right);
		}

        /// <summary>
        /// Implements the less operator.
        /// </summary>
        /// <param name="left">The left value.</param>
        /// <param name="right">The right value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
		public static bool operator <(CellIndexBase left, CellIndexBase right)
		{
			if (left == null)
			{
				return false;
			}

			if (left.RowIndex < right.RowIndex)
			{
				return true;
			}

			if (left.RowIndex == right.RowIndex)
			{
				return left.ColumnIndex < right.ColumnIndex;
			}

			return false;
		}

        /// <summary>
        /// Implements the greater operator.
        /// </summary>
        /// <param name="left">The left value.</param>
        /// <param name="right">The right value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
		public static bool operator >(CellIndexBase left, CellIndexBase right)
		{
			if (left == null)
			{
				return false;
			}

			if (left.RowIndex > right.RowIndex)
			{
				return true;
			}

			if (left.RowIndex == right.RowIndex)
			{
				return left.ColumnIndex > right.ColumnIndex;
			}

			return false;
		}

        /// <summary>
        /// Implements the less or equal operator.
        /// </summary>
        /// <param name="left">The left value.</param>
        /// <param name="right">The right value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
		public static bool operator <=(CellIndexBase left, CellIndexBase right)
		{
			return !(left > right);
		}

        /// <summary>
        /// Implements the greater or equal operator.
        /// </summary>
        /// <param name="left">The left value.</param>
        /// <param name="right">The right value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
		public static bool operator >=(CellIndexBase left, CellIndexBase right)
		{
			return !(left < right);
		}

        /// <summary>
        /// Compares to.
        /// </summary>
        /// <param name="other">The other.</param>
        /// <returns></returns>
        public int CompareTo(CellIndexBase other)
        {
			if (this == other)
			{
				return 0;
			}

			if (this < other)
			{
				return -1;
			}

			return 1;
		}

		#endregion
	}
}