﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Maths;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a range of cells.
    /// </summary>
    public class CellRange
    {
        #region Constants

        /// <summary>
        /// An empty range.
        /// </summary>
        public static readonly CellRange Empty = new CellRange(0, 0, 0, 0);

        #endregion


        #region Fields

        private readonly CellIndex fromIndex;
        private readonly CellIndex toIndex;
        private readonly bool isTopToBottom;
        private readonly bool isLeftToRight;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the start index.
        /// </summary>
        /// <value>The start index.</value>
        public CellIndex FromIndex
        {
            get
            {
                return this.fromIndex;
            }
        }

        /// <summary>
        /// Gets the end index.
        /// </summary>
        /// <value>The end index.</value>
        public CellIndex ToIndex
        {
            get
            {
                return this.toIndex;
            }
        }

        /// <summary>
        /// Gets the row count.
        /// </summary>
        /// <value>The row count.</value>
        public int RowCount
        {
            get
            {
                return this.ToIndex.RowIndex - this.FromIndex.RowIndex + 1;
            }
        }

        /// <summary>
        /// Gets the column count.
        /// </summary>
        /// <value>The column count.</value>
        public int ColumnCount
        {
            get
            {
                return this.ToIndex.ColumnIndex - this.FromIndex.ColumnIndex + 1;
            }
        }

        internal bool IsTopToBottom
        {
            get
            {
                return this.isTopToBottom;
            }
        }

        internal bool IsLeftToRight
        {
            get
            {
                return this.isLeftToRight;
            }
        }

        /// <summary>
        /// Gets a value indicating if the cell range is single cell range.
        /// </summary>
        /// <value>The value indicating if the cell range is single cell range.</value>
        public bool IsSingleCell
        {
            get
            {
                return this.RowCount == 1 && this.ColumnCount == 1;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="CellRange" /> class.
        /// </summary>
        /// <param name="fromRowIndex">Start row index.</param>
        /// <param name="fromColumnIndex">Start column index.</param>
        /// <param name="toRowIndex">End row index.</param>
        /// <param name="toColumnIndex">End column index.</param>
        public CellRange(int fromRowIndex, int fromColumnIndex, int toRowIndex, int toColumnIndex)
        {
            this.isLeftToRight = (fromColumnIndex <= toColumnIndex);
            this.isTopToBottom = (fromRowIndex <= toRowIndex);

            this.fromIndex = new CellIndex(
                Math.Min(fromRowIndex, toRowIndex),
                Math.Min(fromColumnIndex, toColumnIndex));

            this.toIndex = new CellIndex(
                Math.Max(fromRowIndex, toRowIndex),
                Math.Max(fromColumnIndex, toColumnIndex));
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CellRange" /> class.
        /// </summary>
        /// <param name="fromIndex">Start index.</param>
        /// <param name="toIndex">End index.</param>
        public CellRange(CellIndex fromIndex, CellIndex toIndex)
            : this(fromIndex.RowIndex, fromIndex.ColumnIndex, toIndex.RowIndex, toIndex.ColumnIndex)
        {
        }

        #endregion


        #region Methods

        /// <summary>
        /// Creates a new cell range using the present instance and a specified offset.
        /// </summary>
        /// <param name="rowOffset">The row offset.</param>
        /// <param name="columnOffset">The column offset.</param>
        /// <returns>The created cell range.</returns>
        public CellRange Offset(int rowOffset, int columnOffset)
        {
            CellIndex fromIndex = this.FromIndex.Offset(rowOffset, columnOffset);
            CellIndex toIndex = this.ToIndex.Offset(rowOffset, columnOffset);

            if (fromIndex == null || toIndex == null)
            {
                return null;
            }

            return new CellRange(fromIndex, toIndex);
        }

        internal CellRange OffsetFromRow(int offset)
        {
            int newFromRowIndex = this.FromIndex.RowIndex + offset;
            int newToRowIndex = this.ToIndex.RowIndex + offset;

            if (!TelerikHelper.IsValidRowIndex(newFromRowIndex) ||
                !TelerikHelper.IsValidRowIndex(newToRowIndex))
            {
                return null;
            }

            return new CellRange(newFromRowIndex, this.FromIndex.ColumnIndex, newToRowIndex, this.ToIndex.ColumnIndex);
        }

        internal CellRange OffsetFromColumn(int offset)
        {
            int newFromColumnIndex = this.FromIndex.ColumnIndex + offset;
            int newToColumnIndex = this.ToIndex.ColumnIndex + offset;

            if (!TelerikHelper.IsValidColumnIndex(newFromColumnIndex) ||
                !TelerikHelper.IsValidColumnIndex(newToColumnIndex))
            {
                return null;
            }

            return new CellRange(this.FromIndex.RowIndex, newFromColumnIndex, this.ToIndex.RowIndex, newToColumnIndex);
        }

        /// <summary>
        /// Gets the first row.
        /// </summary>
        /// <returns>The cell range of the first row.</returns>
        public CellRange GetFirstRow()
        {
            return new CellRange(this.FromIndex.RowIndex, this.FromIndex.ColumnIndex, this.FromIndex.RowIndex, this.ToIndex.ColumnIndex);
        }

        /// <summary>
        /// Gets the last row.
        /// </summary>
        /// <returns>The cell range of the last row.</returns>
        public CellRange GetLastRow()
        {
            return new CellRange(this.ToIndex.RowIndex, this.FromIndex.ColumnIndex, this.ToIndex.RowIndex, this.ToIndex.ColumnIndex);
        }

        /// <summary>
        /// Gets the first column.
        /// </summary>
        /// <returns>The cell range of the first column.</returns>
        public CellRange GetFirstColumn()
        {
            return new CellRange(this.FromIndex.RowIndex, this.FromIndex.ColumnIndex, this.ToIndex.RowIndex, this.FromIndex.ColumnIndex);
        }

        /// <summary>
        /// Gets the last column.
        /// </summary>
        /// <returns>The cell range of the last column.</returns>
        public CellRange GetLastColumn()
        {
            return new CellRange(this.FromIndex.RowIndex, this.ToIndex.ColumnIndex, this.ToIndex.RowIndex, this.ToIndex.ColumnIndex);
        }

        internal CellRange GetRow(int rowIndex)
        {
            Guard.ThrowExceptionIfOutOfRange(this.FromIndex.RowIndex, this.ToIndex.RowIndex, rowIndex, "rowIndex");

            return new CellRange(rowIndex, this.FromIndex.ColumnIndex, rowIndex, this.ToIndex.ColumnIndex);
        }

        internal CellRange GetColumn(int columnIndex)
        {
            Guard.ThrowExceptionIfOutOfRange(this.FromIndex.ColumnIndex, this.ToIndex.ColumnIndex, columnIndex, "columnIndex");

            return new CellRange(this.FromIndex.RowIndex, columnIndex, this.ToIndex.RowIndex, columnIndex);
        }

        internal bool ContainsColumnsProjection(CellRange other)
        {
            return this.IsFirstIntervalContainingSecond(
                        this.FromIndex.ColumnIndex, this.ToIndex.ColumnIndex,
                        other.FromIndex.ColumnIndex, other.ToIndex.ColumnIndex);
        }

        internal bool ContainsRowsProjection(CellRange other)
        {
            return this.IsFirstIntervalContainingSecond(
                        this.FromIndex.RowIndex, this.ToIndex.RowIndex,
                        other.FromIndex.RowIndex, other.ToIndex.RowIndex);
        }

        internal bool ContainsProjection(CellRange other, bool isRowProjection)
        {
            return isRowProjection ? this.ContainsRowsProjection(other) : this.ContainsColumnsProjection(other);
        }

        private bool IsFirstIntervalContainingSecond(int firstStart, int firstEnd, int secondStart, int secondEnd)
        {
            return firstStart <= secondStart && secondEnd <= firstEnd;
        }

        /// <summary>
        /// Determines whether the range contains the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <returns>The value indicating whether the range contains the specified index.</returns>
        public bool Contains(CellIndex index)
        {
            Guard.ThrowExceptionIfNull(index, "index");

            return this.Contains(index.RowIndex, index.ColumnIndex);
        }

        /// <summary>
        /// Determines whether the range contains the specified index.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The value indicating whether the range contains the specified index.</returns>
        public bool Contains(int rowIndex, int columnIndex)
        {
            return this.fromIndex.RowIndex <= rowIndex && rowIndex <= this.toIndex.RowIndex &&
                this.fromIndex.ColumnIndex <= columnIndex && columnIndex <= this.toIndex.ColumnIndex;
        }

        /// <summary>
        /// Determines whether the range contains the specified cell range.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <returns>The value indicating whether the range contains the specified range.</returns>
        public bool Contains(CellRange cellRange)
        {
            return this.Contains(cellRange.FromIndex) && this.Contains(cellRange.ToIndex);
        }

        internal bool IntersectsRowsProjections(CellRange other)
        {
            return Utils.AreIntersectingIntervals(
                       this.FromIndex.RowIndex, this.ToIndex.RowIndex,
                       other.FromIndex.RowIndex, other.ToIndex.RowIndex);
        }

        internal bool IntersectsColumnsProjections(CellRange other)
        {
            return Utils.AreIntersectingIntervals(
                        this.FromIndex.ColumnIndex, this.ToIndex.ColumnIndex,
                        other.FromIndex.ColumnIndex, other.ToIndex.ColumnIndex);
        }

        /// <summary>
        /// Determines whether the present instance intersects with another cell range.
        /// </summary>
        /// <param name="other">The other cell range.</param>
        /// <returns>The value indicating whether the present instance intersects with another cell range.</returns>
        public bool IntersectsWith(CellRange other)
        {
            Guard.ThrowExceptionIfNull(other, "other");

            return (other.FromIndex.ColumnIndex <= this.ToIndex.ColumnIndex) &&
                (other.ToIndex.ColumnIndex >= this.FromIndex.ColumnIndex) &&
                (other.FromIndex.RowIndex <= this.ToIndex.RowIndex) &&
                (other.ToIndex.RowIndex >= this.FromIndex.RowIndex);
        }

        /// <summary>
        /// Intersects the present instance with another cell range.
        /// </summary>
        /// <param name="other">The other cell range.</param>
        /// <returns>The intersection.</returns>
        public CellRange Intersect(CellRange other)
        {
            if (!this.IntersectsWith(other))
            {
                return CellRange.Empty;
            }
            else
            {
                return new CellRange(CellIndex.Max(this.FromIndex, other.FromIndex), CellIndex.Min(this.ToIndex, other.ToIndex));
            }
        }

        internal static CellRange MaxOrNull(CellRange firstCellRange, CellRange secondCellRange)
        {
            if (firstCellRange == null)
            {
                return secondCellRange;
            }

            if (secondCellRange == null)
            {
                return firstCellRange;
            }

            return CellRange.Max(firstCellRange, secondCellRange);
        }

        internal static CellRange Max(CellRange first, CellRange second)
        {
            Guard.ThrowExceptionIfNull(first, "first");
            Guard.ThrowExceptionIfNull(second, "second");

            return new CellRange(
                Math.Min(first.FromIndex.RowIndex, second.FromIndex.RowIndex),
                Math.Min(first.FromIndex.ColumnIndex, second.FromIndex.ColumnIndex),
                Math.Max(first.ToIndex.RowIndex, second.ToIndex.RowIndex),
                Math.Max(first.ToIndex.ColumnIndex, second.ToIndex.ColumnIndex));
        }

        internal static CellRange RestrictCellRange(CellRange cellRange, SizeI size)
        {
            Guard.ThrowExceptionIfNull(cellRange, "cellRange");
            Guard.ThrowExceptionIfOutOfRange(0, SpreadsheetDefaultValues.RowCount - 1, size.Height - 1, "maxRowIndex");
            Guard.ThrowExceptionIfOutOfRange(0, SpreadsheetDefaultValues.ColumnCount - 1, size.Width - 1, "maxColumnIndex");

            CellIndex restrictedFromIndex = CellIndex.RestrictCellIndex(cellRange.FromIndex, size);
            CellIndex restrictedToIndex = CellIndex.RestrictCellIndex(cellRange.ToIndex, size);

            if (cellRange.IsLeftToRight && cellRange.IsTopToBottom)
            {
                return new CellRange(restrictedFromIndex, restrictedToIndex);
            }
            else if (cellRange.IsLeftToRight && !cellRange.IsTopToBottom)
            {
                return new CellRange(restrictedToIndex.RowIndex, restrictedFromIndex.ColumnIndex,
                    restrictedFromIndex.RowIndex, restrictedToIndex.ColumnIndex);
            }
            else if (!cellRange.IsLeftToRight && cellRange.IsTopToBottom)
            {
                return new CellRange(restrictedFromIndex.RowIndex, restrictedToIndex.ColumnIndex,
                    restrictedToIndex.RowIndex, restrictedFromIndex.ColumnIndex);
            }
            else
            {
                return new CellRange(restrictedToIndex, restrictedFromIndex);
            }
        }

        /// <summary>
        /// Creates a cell range of the specified row.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <returns>The cell range of the specified row.</returns>
        public static CellRange FromRow(int rowIndex)
        {
            return FromRowRange(rowIndex, rowIndex);
        }

        /// <summary>
        /// Creates a cell range of a range of rows.
        /// </summary>
        /// <param name="fromRowIndex">Start row.</param>
        /// <param name="toRowIndex">End row.</param>
        /// <returns>The cell range.</returns>
        public static CellRange FromRowRange(int fromRowIndex, int toRowIndex)
        {
            Guard.ThrowExceptionIfInvalidRowIndex(fromRowIndex);
            Guard.ThrowExceptionIfInvalidRowIndex(toRowIndex);
            return new CellRange(fromRowIndex, 0, toRowIndex, SpreadsheetDefaultValues.ColumnCount - 1);
        }

        /// <summary>
        /// Creates a cell range of the specified column.
        /// </summary>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The cell range of the specified column.</returns>
        public static CellRange FromColumn(int columnIndex)
        {
            return FromColumnRange(columnIndex, columnIndex);
        }

        /// <summary>
        /// Creates a cell range of a range of columns.
        /// </summary>
        /// <param name="fromColumnIndex">Start column.</param>
        /// <param name="toColumnIndex">End column.</param>
        /// <returns>The cell range.</returns>
        public static CellRange FromColumnRange(int fromColumnIndex, int toColumnIndex)
        {
            Guard.ThrowExceptionIfInvalidColumnIndex(fromColumnIndex);
            return new CellRange(0, fromColumnIndex, SpreadsheetDefaultValues.RowCount - 1, toColumnIndex);
        }

        internal IEnumerable<CellRange> SplitToRangesSurroundingRange(CellRange cellRange)
        {
            if (!this.IntersectsWith(cellRange))
            {
                yield return this;
            }
            else
            {
                if (this.FromIndex.ColumnIndex < cellRange.FromIndex.ColumnIndex)
                {
                    yield return new CellRange(this.FromIndex.RowIndex, this.FromIndex.ColumnIndex, this.ToIndex.RowIndex, cellRange.FromIndex.ColumnIndex - 1);
                }

                if (this.ToIndex.ColumnIndex > cellRange.ToIndex.ColumnIndex)
                {
                    yield return new CellRange(this.FromIndex.RowIndex, cellRange.ToIndex.ColumnIndex + 1, this.ToIndex.RowIndex, this.ToIndex.ColumnIndex);
                }

                if (this.FromIndex.RowIndex < cellRange.FromIndex.RowIndex)
                {
                    yield return new CellRange(this.FromIndex.RowIndex, cellRange.FromIndex.ColumnIndex, cellRange.FromIndex.RowIndex - 1, cellRange.ToIndex.ColumnIndex);
                }

                if (this.ToIndex.RowIndex > cellRange.ToIndex.RowIndex)
                {
                    yield return new CellRange(cellRange.ToIndex.RowIndex + 1, cellRange.FromIndex.ColumnIndex, this.ToIndex.RowIndex, cellRange.ToIndex.ColumnIndex);
                }
            }
        }

        internal static IEnumerable<CellRange> GetDifference(CellRange firstRange, CellRange secondRange)
        {
            Guard.ThrowExceptionIfNull(firstRange, "firstRange");
            Guard.ThrowExceptionIfNull(secondRange, "secondRange");

            if (!firstRange.IntersectsWith(secondRange))
            {
                yield return firstRange;
            }
            else
            {
                CellRange intersectionRange = firstRange.Intersect(secondRange);

                int innerRangeFromRow = intersectionRange.FromIndex.RowIndex - 1;
                int innerRangeFromColumn = intersectionRange.FromIndex.ColumnIndex - 1;
                int innerRangeToRow = intersectionRange.ToIndex.RowIndex + 1;
                int innerRangeToColumn = intersectionRange.ToIndex.ColumnIndex + 1;

                if (firstRange.FromIndex.ColumnIndex <= innerRangeFromColumn)
                {
                    yield return new CellRange(firstRange.FromIndex.RowIndex, firstRange.FromIndex.ColumnIndex, intersectionRange.ToIndex.RowIndex, innerRangeFromColumn);
                }

                if (firstRange.FromIndex.RowIndex <= innerRangeFromRow)
                {
                    yield return new CellRange(firstRange.FromIndex.RowIndex, intersectionRange.FromIndex.ColumnIndex, innerRangeFromRow, firstRange.ToIndex.ColumnIndex);
                }

                if (innerRangeToColumn <= firstRange.ToIndex.ColumnIndex)
                {
                    yield return new CellRange(intersectionRange.FromIndex.RowIndex, innerRangeToColumn, firstRange.ToIndex.RowIndex, firstRange.ToIndex.ColumnIndex);
                }

                if (innerRangeToRow <= firstRange.ToIndex.RowIndex)
                {
                    yield return new CellRange(innerRangeToRow, firstRange.FromIndex.ColumnIndex, firstRange.ToIndex.RowIndex, intersectionRange.ToIndex.ColumnIndex);
                }
            }
        }

        internal int StartIndex(bool isRowIndex)
        {
            return isRowIndex ? this.FromIndex.RowIndex : this.FromIndex.ColumnIndex;
        }

        internal int EndIndex(bool isRowIndex)
        {
            return isRowIndex ? this.ToIndex.RowIndex : this.ToIndex.ColumnIndex;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            CellRange other = obj as CellRange;

            if (other == null)
            {
                return false;
            }

            return this.FromIndex.Equals(other.FromIndex) &&
                this.ToIndex.Equals(other.ToIndex);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(this.FromIndex.GetHashCode(), this.ToIndex.GetHashCode());
        }

        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current
        /// <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
            return string.Format("{0}:{1}", this.FromIndex, this.ToIndex);
        }

        #endregion
    }
}