﻿using System;
using System.Collections;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{    
    /// <summary>
    /// A collection which can process the results of insertions and removals of cell ranges.
    /// </summary>
    public abstract class CellRangeInsertedOrRemovedAwareCollection : CellRangeInsertedOrRemovedAwareCollectionBase, IEnumerable<CellRange>
    {
        #region Fields

        private readonly HashSet<CellRange> cellRanges;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the cell ranges.
        /// </summary>
        /// <value>The cell ranges.</value>
        protected override ICollection<CellRange> CellRanges
        {
            get
            {
                return this.cellRanges;
            }
        }

        /// <summary>
        /// Gets the count.
        /// </summary>
        /// <value>The count.</value>
        public int Count
        {
            get 
            {
                return this.cellRanges.Count;
            }
        }

        #endregion


        #region Constructors

        internal CellRangeInsertedOrRemovedAwareCollection(Cells cells)
            : base(cells)
        {
            Guard.ThrowExceptionIfNull(cells, "cells");

            this.cellRanges = new HashSet<CellRange>();
        }

        #endregion


        #region Methods

        /// <summary>
        /// Translates the range.
        /// </summary>
        /// <param name="oldRange">The old range.</param>
        /// <param name="newRange">The new range.</param>
        protected override void TranslateRange(CellRange oldRange, CellRange newRange)
        {
            Guard.ThrowExceptionIfNull(oldRange, "oldRange");
            Guard.ThrowExceptionIfNull(newRange, "newRange");

            this.Remove(oldRange);

            if (newRange != null)
            {
                this.Add(newRange);
            }
        }

        /// <summary>
        /// Adds the specified cell range.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        protected virtual void Add(CellRange cellRange)
        {
            Guard.ThrowExceptionIfNull(cellRange, "cellRange");

            this.CellRanges.Add(cellRange);
        }

        /// <summary>
        /// Removes the specified cell range.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        protected virtual void Remove(CellRange cellRange)
        {
            Guard.ThrowExceptionIfNull(cellRange, "cellRange");

            this.CellRanges.Remove(cellRange);
        }

        #endregion


        #region IEnumerable<T> implementation
        
        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<CellRange> GetEnumerator()
        {
            return this.cellRanges.GetEnumerator();
        }
        
        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return ((IEnumerable<CellRange>)this).GetEnumerator();
        }

        #endregion
    }
}
