﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents base for the CellRangeInsertedOrRemovedAwareCollection class.
    /// </summary>
    public abstract class CellRangeInsertedOrRemovedAwareCollectionBase
    {
        #region Properties

        /// <summary>
        /// Gets the cell ranges.
        /// </summary>
        /// <value>The cell ranges.</value>
        protected abstract ICollection<CellRange> CellRanges { get; }

        /// <summary>
        /// Gets or sets the cells.
        /// </summary>
        /// <value>The cells.</value>
        protected Cells Cells { get; set; }

        #endregion


        #region Constructors

        // TODO: Remove cells from this constructor!
        internal CellRangeInsertedOrRemovedAwareCollectionBase(Cells cells)
        {
            Guard.ThrowExceptionIfNull(cells, "cells");

            this.Cells = cells;
        }

        #endregion
        

        #region Methods

        internal void Update(CellRangeInsertedOrRemovedEventArgs eventArgs)
        {
            ShiftType shiftType;
            if (this.ShouldTranslateUpDown(eventArgs.RangeType) || this.ShouldTranslateLeftRight(eventArgs.RangeType))
            {
                shiftType = eventArgs.RangeType.ToShiftType(eventArgs.IsRemove);
                this.TranslateRanges(eventArgs.Range, shiftType);
            }
        }

        /// <summary>
        /// Determines whether the specified selected range can be inserted or removed from this instance.
        /// </summary>
        /// <param name="selectedRange">The selected range.</param>
        /// <param name="shiftType">Type of the shift.</param>
        /// <returns>The value indicating whether the specified selected range can be inserted or removed from this instance.</returns>
        internal bool CanInsertOrRemove(CellRange selectedRange, ShiftType shiftType)
        {
            Guard.ThrowExceptionIfNull(selectedRange, "selectedRange");

            foreach (CellRange range in this.CellRanges)
            {
                if (!this.CanInsertOrRemove(shiftType, range, selectedRange))
                {
                    return false;
                }
            }

            return true;
        }

        /// <summary>
        /// Determines whether this instance can insert or remove the selected range depending on some existing range and a given shift type.
        /// </summary>
        /// <param name="shiftType">Type of the shift.</param>
        /// <param name="range">The range.</param>
        /// <param name="selectedRange">The selected range.</param>
        /// <returns>True if the selected range can be inserted or removed.</returns>
        protected virtual bool CanInsertOrRemove(ShiftType shiftType, CellRange range, CellRange selectedRange)
        {
            Guard.ThrowExceptionIfNull(range, "range");
            Guard.ThrowExceptionIfNull(selectedRange, "selectedRange");

            return true;
        }

        /// <summary>
        /// Translates the range.
        /// </summary>
        /// <param name="oldRange">The old range.</param>
        /// <param name="newRange">The new range.</param>
        protected abstract void TranslateRange(CellRange oldRange, CellRange newRange);

        /// <summary>
        /// Checks if should translate up or down.
        /// </summary>
        /// <param name="rangeType">Type of the range.</param>
        /// <returns>A value indicating whether should translate up or down.</returns>
        protected virtual bool ShouldTranslateUpDown(RangeType rangeType)
        {
            return rangeType == RangeType.Rows || rangeType == RangeType.CellsInColumn;
        }

        /// <summary>
        /// Checks if should translate left or right.
        /// </summary>
        /// <param name="rangeType">Type of the range.</param>
        /// <returns>A value indicating whether should translate left or right.</returns>
        protected virtual bool ShouldTranslateLeftRight(RangeType rangeType)
        {
            return rangeType == RangeType.Columns || rangeType == RangeType.CellsInRow;
        }
  
        /// <summary>
        /// Translates the ranges.
        /// </summary>
        /// <param name="oldAndNewTranslatedRangesPositions">The old and new translated ranges positions.</param>
        protected virtual void TranslateRanges(Dictionary<CellRange, CellRange> oldAndNewTranslatedRangesPositions)
        {
            foreach (KeyValuePair<CellRange, CellRange> item in oldAndNewTranslatedRangesPositions)
            {
                this.TranslateRange(item.Key, item.Value);
            }
        }

        /// <summary>
        /// Called before translation.
        /// </summary>
        /// <param name="oldAndNewTranslatedRangesPositions">The old and new translated ranges positions.</param>
        /// <param name="rangeToInsertOrRemove">The range to insert or remove.</param>
        /// <param name="shiftType">Type of the shift.</param>
        protected virtual void OnBeforeTranslation(
            Dictionary<CellRange, CellRange> oldAndNewTranslatedRangesPositions, CellRange rangeToInsertOrRemove, ShiftType shiftType)
        {
        }

        /// <summary>
        /// Called after translation.
        /// </summary>
        /// <param name="oldAndNewTranslatedRangesPositions">The old and new translated ranges positions.</param>
        /// <param name="rangeToInsertOrRemove">The range to insert or remove.</param>
        /// <param name="shiftType">Type of the shift.</param>
        protected virtual void OnAfterTranslation(
            Dictionary<CellRange, CellRange> oldAndNewTranslatedRangesPositions, CellRange rangeToInsertOrRemove, ShiftType shiftType)
        {
        }
  
        private Dictionary<CellRange, CellRange> CalculateOldAndNewTranslatedRangesPositions(CellRange rangeToInsertOrRemove, ShiftType shiftType)
        {
            Guard.ThrowExceptionIfNull(rangeToInsertOrRemove, "rangeToInsertOrRemove");

            Dictionary<CellRange, CellRange> translatedCellRanges = new Dictionary<CellRange, CellRange>();

            foreach (CellRange range in this.CellRanges)
            {
                CellRange translatedRange;
                if (range.TryTranslate(rangeToInsertOrRemove, shiftType, out translatedRange))
                {
                    translatedCellRanges[range] = translatedRange;
                }
            }

            return translatedCellRanges;
        }

        private void TranslateRanges(CellRange rangeToInsertOrRemove, ShiftType shiftType)
        {
            Dictionary<CellRange, CellRange> translatedCellRanges = this.CalculateOldAndNewTranslatedRangesPositions(rangeToInsertOrRemove, shiftType);

            this.OnBeforeTranslation(translatedCellRanges, rangeToInsertOrRemove, shiftType);
            this.TranslateRanges(translatedCellRanges);
            this.OnAfterTranslation(translatedCellRanges, rangeToInsertOrRemove, shiftType);
        }

        #endregion
    }
}
