﻿using System;
using System.Collections;
using System.Collections.Generic;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// A collection which can process the results of insertions and removals of cell ranges.
    /// </summary>
    public abstract class CellRangeInsertedOrRemovedAwareCollection<T> : CellRangeInsertedOrRemovedAwareCollectionBase, IEnumerable<T>
    {
        #region Fields

        private readonly Worksheet worksheet;
        private readonly Dictionary<CellRange, T> elements;

        #endregion


        #region Properties

        
        /// <summary>
        /// Gets the count.
        /// </summary>
        /// <value>The count.</value>
        public int Count
        {
            get
            {
                return this.elements.Count;
            }
        }

        /// <summary>
        /// Gets the worksheet.
        /// </summary>
        /// <value>The worksheet.</value>
        protected Worksheet Worksheet
        {
            get
            {
                return this.worksheet;
            }
        }

        /// <summary>
        /// Gets the cell ranges.
        /// </summary>
        /// <value>The cell ranges.</value>
        protected override ICollection<CellRange> CellRanges
        {
            get
            {
                return this.elements.Keys;
            }
        }

        /// <summary>
        /// Gets the elements.
        /// </summary>
        /// <value>The elements.</value>
        protected IDictionary<CellRange, T> Elements
        {
            get
            {
                return this.elements;
            }
        }

        #endregion


        #region Constructors

        internal CellRangeInsertedOrRemovedAwareCollection(Worksheet worksheet)
            : base(worksheet.Cells)
        {
            this.worksheet = worksheet;
            this.elements = new Dictionary<CellRange, T>();
        }

        #endregion


        #region Methods

        /// <summary>
        /// Adds the specified element when cellRange is inserted.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <param name="element">The element.</param>
        /// <returns></returns>
        protected abstract T Add(CellRange cellRange, T element);

        /// <summary>
        /// Removes elements when cellRange is removed.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <returns></returns>
        protected abstract bool Remove(CellRange cellRange);

        /// <summary>
        /// Translates the range.
        /// </summary>
        /// <param name="oldRange">The old range.</param>
        /// <param name="newRange">The new range.</param>
        protected override void TranslateRange(CellRange oldRange, CellRange newRange)
        {
            T item = this.elements[oldRange];

            this.Remove(oldRange);

            if (newRange != null)
            {
                this.Add(newRange, item);
            }
        }

        #endregion


        #region IEnumerable<T> implementation

        
        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<T> GetEnumerator()
        {
            return this.Elements.Values.GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return ((IEnumerable<T>)this).GetEnumerator();
        }

        #endregion
    }
}
