﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Model.Printing;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    internal static class InsertRemoveCellsHelper
    {
        private static readonly Dictionary<ShiftType, object> shiftTypeToInsertRemoveShiftType;

        static InsertRemoveCellsHelper()
        {
            shiftTypeToInsertRemoveShiftType = new Dictionary<ShiftType, object>();

            shiftTypeToInsertRemoveShiftType[ShiftType.Down] = InsertShiftType.Down;
            shiftTypeToInsertRemoveShiftType[ShiftType.Right] = InsertShiftType.Right;
            shiftTypeToInsertRemoveShiftType[ShiftType.Up] = RemoveShiftType.Up;
            shiftTypeToInsertRemoveShiftType[ShiftType.Left] = RemoveShiftType.Left;
        }

        public static RangeType ToRangeType(this InsertShiftType insertShiftType, bool isFullRowColumn)
        {
            if (isFullRowColumn)
            {
                return insertShiftType == InsertShiftType.Right ? RangeType.Columns : RangeType.Rows;
            }
            else
            {
                return insertShiftType == InsertShiftType.Right ? RangeType.CellsInRow : RangeType.CellsInColumn;
            }
        }

        public static RangeType ToRangeType(this RemoveShiftType removeShiftType, bool isFullRowColumn)
        {
            if (isFullRowColumn)
            {
                return removeShiftType == RemoveShiftType.Left ? RangeType.Columns : RangeType.Rows;
            }
            else
            {
                return removeShiftType == RemoveShiftType.Left ? RangeType.CellsInRow : RangeType.CellsInColumn;
            }
        }

        public static ShiftType ToShiftType(this RangeType rangeType, bool isRemove)
        {
            switch(rangeType)
            {
                case RangeType.Rows:
                case RangeType.CellsInColumn:
                    return isRemove ? ShiftType.Up : ShiftType.Down;
                case RangeType.Columns:
                case RangeType.CellsInRow:
                    return isRemove ? ShiftType.Left : ShiftType.Right;
                default:
                    throw new NotSupportedException(String.Format("Not supported RangeType enumeration: {0}!", rangeType));
            }
        }

        public static ShiftType ToShiftType(this RemoveShiftType removeShiftType)
        {
            return removeShiftType == RemoveShiftType.Left ? ShiftType.Left : ShiftType.Up;
        }

        public static ShiftType ToShiftType(this InsertShiftType insertShiftType)
        {
            return insertShiftType == InsertShiftType.Down ? ShiftType.Down : ShiftType.Right;
        }

        public static InsertShiftType ReverseShiftType(this RemoveShiftType shiftType)
        {
            switch (shiftType)
            {
                case RemoveShiftType.Left:
                    return InsertShiftType.Right;
                case RemoveShiftType.Up:
                    return InsertShiftType.Down;
                default:
                    throw new NotSupportedException();
            }
        }

        public static RemoveShiftType ReverseShiftType(this InsertShiftType shiftType)
        {
            switch (shiftType)
            {
                case InsertShiftType.Right:
                    return RemoveShiftType.Left;
                case InsertShiftType.Down:
                    return RemoveShiftType.Up;
                default:
                    throw new NotSupportedException();
            }
        }

        public static bool TryGetInsertShiftType(this ShiftType shiftType, out InsertShiftType insertShiftType)
        {
            insertShiftType = InsertShiftType.Down;

            object value;
            if (shiftTypeToInsertRemoveShiftType.TryGetValue(shiftType, out value) && value is InsertShiftType)
            {
                insertShiftType = (InsertShiftType)value;

                return true;
            }

            return false;
        }

        public static bool TryGetRemoveShiftType(this ShiftType shiftType, out RemoveShiftType removeShiftType)
        {
            removeShiftType = RemoveShiftType.Left;

            object value;
            if (shiftTypeToInsertRemoveShiftType.TryGetValue(shiftType, out value) && value is RemoveShiftType)
            {
                removeShiftType = (RemoveShiftType)value;

                return true;
            }

            return false;
        }

        public static bool ShouldTranslate(this CellRange range, CellRange insertedRange, InsertShiftType shiftType)
        {
            bool isShiftingRows = shiftType == InsertShiftType.Down;
            int insertionIndex = insertedRange.StartIndex(isShiftingRows);
            int insertedRowsColumnsCount = isShiftingRows ? insertedRange.RowCount : insertedRange.ColumnCount;
            int totalRowsColumnsCount = isShiftingRows ? SpreadsheetDefaultValues.RowCount : SpreadsheetDefaultValues.ColumnCount;

            int rangeEndIndex = range.EndIndex(isShiftingRows);

            bool shouldTranslate = insertedRange.ContainsProjection(range, !isShiftingRows) &&
                insertionIndex <= rangeEndIndex &&
                rangeEndIndex + insertedRowsColumnsCount < totalRowsColumnsCount;

            return shouldTranslate;
        }

        public static bool ShouldTranslate(this CellRange range, CellRange removedRange, RemoveShiftType shiftType)
        {
            bool isShiftingRows = shiftType == RemoveShiftType.Up;
            int removedRangeStartIndex = removedRange.StartIndex(isShiftingRows);
            int rangeEndIndex = range.EndIndex(isShiftingRows);

            bool shouldTranslate = removedRange.ContainsProjection(range, !isShiftingRows) && removedRangeStartIndex <= rangeEndIndex;

            return shouldTranslate;
        }

        public static bool TryTranslate(this CellRange range, CellRange insertedRange, InsertShiftType shiftType, out CellRange translatedObject)
        {
            Guard.ThrowExceptionIfNull(range, "range");
            Guard.ThrowExceptionIfNull(insertedRange, "insertedRange");

            translatedObject = null;

            if (range.ShouldTranslate(insertedRange, shiftType))
            {
                bool isShiftingRows = shiftType == InsertShiftType.Down;
                int insertionIndex = isShiftingRows ? insertedRange.FromIndex.RowIndex : insertedRange.FromIndex.ColumnIndex;
                int insertedRowsColumnsCount = isShiftingRows ? insertedRange.RowCount : insertedRange.ColumnCount;
                int rangeStartIndex = isShiftingRows ? range.FromIndex.RowIndex : range.FromIndex.ColumnIndex;

                bool isInsideRange = insertionIndex > rangeStartIndex;

                int rowsOffset = isShiftingRows ? insertedRowsColumnsCount : 0;
                int columnsOffset = isShiftingRows ? 0 : insertedRowsColumnsCount;

                if (isInsideRange)
                {
                    CellIndex translatedEndIndex = range.ToIndex.Offset(rowsOffset, columnsOffset);

                    if(translatedEndIndex != null)
                    {
                        translatedObject = new CellRange(range.FromIndex, translatedEndIndex);
                    }
                }
                else
                {
                    translatedObject = range.Offset(rowsOffset, columnsOffset);
                }

                return true;
            }

            return false;
        }

        public static bool TryTranslate(this CellRange range, CellRange removedRange, RemoveShiftType shiftType, out CellRange translatedObject)
        {
            Guard.ThrowExceptionIfNull(range, "range");
            Guard.ThrowExceptionIfNull(removedRange, "removedRange");

            translatedObject = null;

            if (range.ShouldTranslate(removedRange, shiftType))
            {
                bool isShiftingRows = shiftType == RemoveShiftType.Up;
                int removedRangeStartIndex = isShiftingRows ? removedRange.FromIndex.RowIndex : removedRange.FromIndex.ColumnIndex;
                int removedRowsColumnsCount = isShiftingRows ? removedRange.RowCount : removedRange.ColumnCount;

                CellIndex translatedStartIndex = range.FromIndex;
                int rangeStartIndex = range.StartIndex(isShiftingRows);
                int rowsColumnsToDeleteAtStart = Math.Min(rangeStartIndex - removedRangeStartIndex, removedRowsColumnsCount);

                if (rowsColumnsToDeleteAtStart > 0)
                {
                    int startRowsOffset = isShiftingRows ? -rowsColumnsToDeleteAtStart : 0;
                    int startColumnsOffset = isShiftingRows ? 0 : -rowsColumnsToDeleteAtStart;

                    translatedStartIndex = translatedStartIndex.Offset(startRowsOffset, startColumnsOffset);
                }

                CellIndex translatedEndIndex = range.ToIndex;
                int rangeEndIndex = range.EndIndex(isShiftingRows);
                int rowsColumnsToDeleteAtEnd = Math.Min(rangeEndIndex - removedRangeStartIndex + 1, removedRowsColumnsCount);
                int endRowsOffset = isShiftingRows ? -rowsColumnsToDeleteAtEnd : 0;
                int endColumnsOffset = isShiftingRows ? 0 : -rowsColumnsToDeleteAtEnd;
                translatedEndIndex = translatedEndIndex.Offset(endRowsOffset, endColumnsOffset);
                
                bool translatedRangeExists = false;

                if (translatedStartIndex != null && translatedEndIndex != null)
                {
                    translatedRangeExists = 
                        translatedStartIndex.RowIndex <= translatedEndIndex.RowIndex &&
                        translatedStartIndex.ColumnIndex <= translatedEndIndex.ColumnIndex;
                }

                if (translatedRangeExists)
                {
                    translatedObject = new CellRange(translatedStartIndex, translatedEndIndex);
                }

                return true;
            }

            return false;
        }

        public static bool TryTranslate(this CellRange range, CellRange insertedOrRemovedRange, ShiftType shiftType, out CellRange translatedObject)
        {
            Guard.ThrowExceptionIfNull(range, "range");
            Guard.ThrowExceptionIfNull(insertedOrRemovedRange, "insertedOrRemovedRange");

            InsertShiftType insertShiftType;
            RemoveShiftType removeShiftType;

            if (shiftType.TryGetInsertShiftType(out insertShiftType))
            {
                return range.TryTranslate(insertedOrRemovedRange, insertShiftType, out translatedObject);
            }
            else if (shiftType.TryGetRemoveShiftType(out removeShiftType))
            {
                return range.TryTranslate(insertedOrRemovedRange, removeShiftType, out translatedObject);
            }
            else
            {
                throw new NotSupportedException("Not supported ShiftType!");
            }
        }

        public static bool ShouldTranslate(this PageBreak pageBreak, CellRange insertedRange, InsertShiftType shiftType)
        {
            return pageBreak.CellRange.ShouldTranslate(insertedRange, shiftType);
        }

        public static bool ShouldTranslate(this PageBreak pageBreak, CellRange removedRange, RemoveShiftType shiftType)
        {
            return pageBreak.CellRange.ShouldTranslate(removedRange, shiftType);
        }

        public static bool TryTranslate(this PageBreak pageBreak, CellRange insertedRange, InsertShiftType shiftType, out PageBreak translatedObject)
        {
            translatedObject = null;

            CellRange translatedCellRange;
            if (pageBreak.CellRange.TryTranslate(insertedRange, shiftType, out translatedCellRange))
            {
                if (translatedCellRange != null)
                {
                    translatedObject = GetPageBreakFromCellRange(translatedCellRange, pageBreak.Type);
                }

                return true;
            }

            return false;
        }

        public static bool TryTranslate(this PageBreak pageBreak, CellRange removedRange, RemoveShiftType shiftType, out PageBreak translatedObject)
        {
            translatedObject = null;

            CellRange translatedCellRange;
            if (pageBreak.CellRange.TryTranslate(removedRange, shiftType, out translatedCellRange))
            {
                if (translatedCellRange != null)
                {
                    translatedObject = GetPageBreakFromCellRange(translatedCellRange, pageBreak.Type);
                }

                return true;
            }

            return false;
        }

        public static bool TryTranslate(this PageBreak pageBreak, CellRange insertedOrRemovedRange, ShiftType shiftType, out PageBreak translatedObject)
        {
            InsertShiftType insertShiftType;
            RemoveShiftType removeShiftType;

            if (shiftType.TryGetInsertShiftType(out insertShiftType))
            {
                return pageBreak.TryTranslate(insertedOrRemovedRange, insertShiftType, out translatedObject);
            }
            else if (shiftType.TryGetRemoveShiftType(out removeShiftType))
            {
                return pageBreak.TryTranslate(insertedOrRemovedRange, removeShiftType, out translatedObject);
            }
            else
            {
                throw new NotSupportedException("Not supported ShiftType!");
            }
        }

        public static bool CanInsertOrRemove(this CellRange range, ShiftType shiftType, CellRange selectedRange)
        {
            switch (shiftType)
            {
                case ShiftType.Left:
                case ShiftType.Right:
                    bool isSelectedColumnsAfterRangeColumns = selectedRange.FromIndex.ColumnIndex > range.ToIndex.ColumnIndex;

                    bool areIntersectingRowsProjections = selectedRange.IntersectsRowsProjections(range);

                    bool isSelectedRowProjectionContainingRangeRowProjection = selectedRange.ContainsRowsProjection(range);

                    return isSelectedColumnsAfterRangeColumns || !areIntersectingRowsProjections || isSelectedRowProjectionContainingRangeRowProjection;

                case ShiftType.Up:
                case ShiftType.Down:
                    bool isSelectedRowsAfterRangeRows = selectedRange.FromIndex.RowIndex > range.ToIndex.RowIndex;

                    bool areIntersectingColumnsProjections = selectedRange.IntersectsColumnsProjections(range);

                    bool isSelectedColumnProjectionContainingRangeColumnProjection = selectedRange.ContainsColumnsProjection(range);

                    return isSelectedRowsAfterRangeRows || !areIntersectingColumnsProjections || isSelectedColumnProjectionContainingRangeColumnProjection;
            }

            return true;
        }
        
        public static int GetRelativeRearrangeIndex(CellRange range, CellRange selectedRange)
        {
            int relativeCutOffIndex = range.ColumnCount;
            bool conditionsAreRearranged = range.FromIndex.ColumnIndex < selectedRange.FromIndex.ColumnIndex;
            if (conditionsAreRearranged)
            {
                relativeCutOffIndex = selectedRange.ToIndex.ColumnIndex - range.FromIndex.ColumnIndex;
            }

            return relativeCutOffIndex;
        }

        public static int GetRearrangeDelta(CellRange range, ShiftType shiftType)
        {
            int delta = range.ColumnCount;

            RemoveShiftType removeType;
            bool shiftTypeIsRemove = shiftType.TryGetRemoveShiftType(out removeType);

            if (shiftTypeIsRemove)
            {
                delta *= -1;
            }

            return delta;
        }

        private static PageBreak GetPageBreakFromCellRange(CellRange range, PageBreakType type)
        {
            bool isHorizontal = type == PageBreakType.Horizontal;

            int index = range.StartIndex(isHorizontal);
            int fromIndex = range.StartIndex(!isHorizontal);
            int toIndex = range.EndIndex(!isHorizontal);

            return new PageBreak(type, index, fromIndex, toIndex);
        }
    }
}
