﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Describes the types of ranges.
    /// </summary>
    public enum RangeType
    {
        /// <summary>
        /// Rows range type.
        /// </summary>
        Rows,

        /// <summary>
        /// Columns range type.
        /// </summary>
        Columns,

        /// <summary>
        /// CellsInRow range type.
        /// </summary>
        CellsInRow,

        /// <summary>
        /// CellsInColumn range type.
        /// </summary>
        CellsInColumn
    }

    /// <summary>
    /// Arguments for the CellRangeInsertedOrRemoved event.
    /// </summary>
    public class CellRangeInsertedOrRemovedEventArgs : EventArgs
    {
        #region Properties

        /// <summary>
        /// Gets the affected range.
        /// </summary>
        /// <value>The affected range.</value>
        public CellRange AffectedRange { get; private set; }

        /// <summary>
        /// Gets the range.
        /// </summary>
        /// <value>The range.</value>
        public CellRange Range { get; private set; }

        /// <summary>
        /// Gets the type of the range.
        /// </summary>
        /// <value>The type of the range.</value>
        public RangeType RangeType { get; private set; }

        /// <summary>
        /// Gets a value indicating whether the type of change is remove.
        /// </summary>
        /// <value>The value indicating whether the type of change is remove.</value>
        public bool IsRemove { get; private set; }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="CellRangeInsertedOrRemovedEventArgs" /> class.
        /// </summary>
        /// <param name="range">The range.</param>
        /// <param name="rangeType">Type of the range.</param>
        /// <param name="isRemove">The value indicating whether the type of change is remove.</param>
        public CellRangeInsertedOrRemovedEventArgs(CellRange range, RangeType rangeType, bool isRemove)
        {
            Guard.ThrowExceptionIfNull(range, "range");

            this.Range = range;
            this.RangeType = rangeType;
            this.IsRemove = isRemove;

            switch (this.RangeType)
            {
                case RangeType.Rows:
                    this.AffectedRange = new CellRange(0, range.FromIndex.ColumnIndex, SpreadsheetDefaultValues.RowCount - 1, SpreadsheetDefaultValues.ColumnCount - 1);
                    break;
                case RangeType.Columns:
                    this.AffectedRange = new CellRange(range.FromIndex.RowIndex, 0, SpreadsheetDefaultValues.RowCount - 1, SpreadsheetDefaultValues.ColumnCount - 1);
                    break;
                case RangeType.CellsInRow:
                    this.AffectedRange = new CellRange(range.FromIndex.RowIndex, range.FromIndex.ColumnIndex, range.ToIndex.RowIndex, SpreadsheetDefaultValues.ColumnCount - 1);
                    break;
                case RangeType.CellsInColumn:
                    this.AffectedRange = new CellRange(range.FromIndex.RowIndex, range.FromIndex.ColumnIndex, SpreadsheetDefaultValues.RowCount - 1, range.ToIndex.ColumnIndex);
                    break;
            }
        }

        #endregion
    }
}