﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a reference to a cell.
    /// </summary>
    public class CellReference
    {
        #region Fields

        private RowColumnReference rowReference;
        private RowColumnReference columnReference;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the actual index of the row.
        /// </summary>
        /// <value>The actual index of the row.</value>
        public int ActualRowIndex
        {
            get
            {
                return this.rowReference.ActualIndex;
            }
        }

        /// <summary>
        /// Gets the actual index of the column.
        /// </summary>
        /// <value>The actual index of the column.</value>
        public int ActualColumnIndex
        {
            get
            {
                return this.columnReference.ActualIndex;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the row is absolute.
        /// </summary>
        /// <value>A value indicating whether the row is absolute.</value>
        public bool IsRowAbsolute
        {
            get
            {
                return this.RowReference.IsAbsolute;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the column is absolute.
        /// </summary>
        /// <value>A value indicating whether the column is absolute.</value>
        public bool IsColumnAbsolute
        {
            get
            {
                return this.ColumnReference.IsAbsolute;
            }
        }

        internal RowColumnReference RowReference
        {
            get
            {
                return this.rowReference;
            }
            private set
            {
                this.rowReference = value;
            }
        }

        internal RowColumnReference ColumnReference
        {
            get
            {
                return this.columnReference;
            }
            private set
            {
                this.columnReference = value;
            }
        }

        #endregion


        #region Constructors

        internal CellReference(RowColumnReference rowReference, RowColumnReference columnReference)
        {
            Guard.ThrowExceptionIfNull(rowReference, "rowReference");
            Guard.ThrowExceptionIfNull(columnReference, "columnReference");

            this.rowReference = rowReference;
            this.columnReference = columnReference;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            CellReference cellReference = obj as CellReference;
            if (cellReference == null)
            {
                return false;
            }

            return this.RowReference.Equals(cellReference.RowReference) &&
                   this.ColumnReference.Equals(cellReference.ColumnReference);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(this.RowReference.GetHashCode(), this.ColumnReference.GetHashCode());
        }

        internal CellReference CloneAndTranslate(int rowIndex, int columnIndex, bool onOverflowStartOver)
        {
            return new CellReference(
                this.rowReference.CloneAndTranslate(rowIndex, onOverflowStartOver, SpreadsheetDefaultValues.RowCount),
                this.columnReference.CloneAndTranslate(columnIndex, onOverflowStartOver, SpreadsheetDefaultValues.ColumnCount));
        }

        internal CellReference CloneAndTranslateUponInsertRemoveCells(bool isHorizontal, int indexShift, int offsetShift, bool onOverflowStartOver)
        {
            if (isHorizontal)
            {
                return new CellReference(
                    this.rowReference,
                    this.columnReference.CloneAndTranslateUponInsertRemoveCells(indexShift, offsetShift, onOverflowStartOver, SpreadsheetDefaultValues.ColumnCount));
            }

            return new CellReference(
                this.rowReference.CloneAndTranslateUponInsertRemoveCells(indexShift, offsetShift, onOverflowStartOver, SpreadsheetDefaultValues.RowCount),
                this.columnReference);
        }

        internal static CellReference GetTopLeftCellReference(CellReference inputFromCellReference, CellReference inputToCellReference)
        {
            RowColumnReference rowFromCellReference = inputFromCellReference.RowReference;
            if (inputFromCellReference.ActualRowIndex > inputToCellReference.ActualRowIndex)
            {
                rowFromCellReference = inputToCellReference.RowReference;
            }

            RowColumnReference columnFromCellReference = inputFromCellReference.ColumnReference;
            if (inputFromCellReference.ActualColumnIndex > inputToCellReference.ActualColumnIndex)
            {
                columnFromCellReference = inputToCellReference.ColumnReference;
            }

            return new CellReference(rowFromCellReference, columnFromCellReference);
        }

        internal static CellReference GetBottomRightCellReference(CellReference inputFromCellReference, CellReference inputToCellReference)
        {
            RowColumnReference rowToCellReference = inputToCellReference.RowReference;
            if (inputFromCellReference.ActualRowIndex > inputToCellReference.ActualRowIndex)
            {
                rowToCellReference = inputFromCellReference.RowReference;
            }

            RowColumnReference columnToCellReference = inputToCellReference.ColumnReference;
            if (inputFromCellReference.ActualColumnIndex > inputToCellReference.ActualColumnIndex)
            {
                columnToCellReference = inputFromCellReference.ColumnReference;
            }

            return new CellReference(rowToCellReference, columnToCellReference);
        }

        internal static CellReference CreateFromCellName(string cellName, CellIndex cellIndex)
        {
            return CreateFromCellName(cellName, cellIndex.RowIndex, cellIndex.ColumnIndex);
        }

        internal static CellReference CreateFromCellName(string cellName, int row, int column)
        {
            Guard.ThrowExceptionIfNull(cellName, "cellName");
            Guard.ThrowExceptionIfInvalidRowIndex(row);
            Guard.ThrowExceptionIfInvalidColumnIndex(column);

            int columnIndex;
            int rowIndex;
            bool isColumnAbsolute;
            bool isRowAbsolute;

            if (NameConverter.TryConvertCellNameToIndex(cellName, out isRowAbsolute, out rowIndex, out isColumnAbsolute, out columnIndex))
            {
                RowColumnReference rowReference = RowColumnReference.CreateRowColumnReference(isRowAbsolute, rowIndex, row, false, SpreadsheetDefaultValues.RowCount);
                RowColumnReference columnReference = RowColumnReference.CreateRowColumnReference(isColumnAbsolute, columnIndex, column, false, SpreadsheetDefaultValues.ColumnCount);

                return new CellReference(rowReference, columnReference);
            }

            return null;
        }

        #endregion
    }
}
