﻿using System;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a cell reference range.
    /// </summary>
    public class CellReferenceRange
    {
        #region Fields

        private readonly WorksheetReference worksheetReference;
        private readonly CellAreaReference cellAreaReference;

        #endregion


        #region Inherited properties

        /// <summary>
        /// Gets the start cell reference.
        /// </summary>
        /// <value>The start cell reference.</value>
        public CellReference FromCellReference
        {
            get
            {
                return this.cellAreaReference.FromCellReference;
            }
        }

        /// <summary>
        /// Gets the end cell reference.
        /// </summary>
        /// <value>The end cell reference.</value>
        public CellReference ToCellReference
        {
            get
            {
                return this.cellAreaReference.ToCellReference;
            }
        }

        /// <summary>
        /// Gets the rows count.
        /// </summary>
        /// <value>The rows count.</value>
        public int RowsCount
        {
            get
            {
                return this.cellAreaReference.RowsCount;
            }
        }

        /// <summary>
        /// Gets the columns count.
        /// </summary>
        /// <value>The columns count.</value>
        public int ColumnsCount
        {
            get
            {
                return this.cellAreaReference.ColumnsCount;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the range is within the valid boundaries.
        /// </summary>
        /// <value>The value indicating whether the range is within the valid boundaries.</value>
        public bool IsInRange
        {
            get
            {
                return this.cellAreaReference != null &&
                    this.cellAreaReference.IsValid;
            }
        }

        /// <summary>
        /// Gets the worksheet.
        /// </summary>
        /// <value>The worksheet.</value>
        public Worksheet Worksheet
        {
            get
            {
                return this.worksheetReference.Worksheet;
            }
        }

        #endregion


        #region Properties

        /// <summary>
        /// Gets the cell reference range.
        /// </summary>
        /// <value>The cell reference range.</value>
        /// <returns>The cell reference range as CellReferenceRange.</returns>
        internal CellAreaReference CellAreaReference
        {
            get
            {
                return this.cellAreaReference;
            }
        }

        internal bool IsWorksheetAbsolute
        {
            get
            {
                return this.worksheetReference.IsWorksheetAbsolute;
            }
        }

        internal bool IsValid
        {
            get
            {
                return this.IsInRange && this.Worksheet != null;
            }
        }

        internal string WorksheetName
        {
            get
            {
                return this.worksheetReference.WorksheetName;
            }
        }

        internal Workbook Workbook
        {
            get
            {
                return this.worksheetReference.Workbook;
            }
        }

        internal WorksheetReference WorksheetReference
        {
            get
            {
                return this.worksheetReference;
            }
        }

        #endregion


        #region Constructors

        internal CellReferenceRange(Workbook workbook, string worksheetName, bool isWorksheetAbsolute, CellAreaReference cellReferenceRange)
            : this(new WorksheetReference(workbook, worksheetName, isWorksheetAbsolute), cellReferenceRange)
        {

        }

        internal CellReferenceRange(WorksheetReference worksheetReference, CellAreaReference cellReferenceRange)
        {
            Guard.ThrowExceptionIfNull(worksheetReference, "worksheetReference");

            this.worksheetReference = worksheetReference;
            this.cellAreaReference = cellReferenceRange;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            CellReferenceRange other = obj as CellReferenceRange;

            return other != null &&
                TelerikHelper.EqualsOfT(this.worksheetReference, other.worksheetReference) &&
                TelerikHelper.EqualsOfT(this.cellAreaReference, other.cellAreaReference);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                this.worksheetReference.GetHashCode(),
                this.cellAreaReference.GetHashCodeOrZero());
        }

        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current
        /// <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
            return this.BuildToString();
        }

        internal string BuildToString()
        {
            StringBuilder builder = new StringBuilder();

            builder.Append(this.worksheetReference.ToString());

            if (this.cellAreaReference == null || !this.cellAreaReference.IsValid)
            {
                builder.Append(ErrorExpressions.ReferenceError.Value);
            }
            else
            {
                builder.Append(this.cellAreaReference.ToString());
            }

            return builder.ToString();
        }

        internal CellReferenceRange CloneAndTranslate(ExpressionCloneAndTranslateContext context)
        {
            CellAreaReference translatedRange = null;

            if (this.cellAreaReference != null)
            {
                if (context.Range != null)
                {
                    return this.CloneAndTranslateUponInsertRemoveCells(context);
                }
                else
                {
                    translatedRange = this.cellAreaReference.CloneAndTranslate(context.RowIndex, context.ColumnIndex, context.OnOverflowStartOver);
                }

                translatedRange = translatedRange.IsValid ? translatedRange : null;
            }

            return new CellReferenceRange(this.worksheetReference.CloneAndTranslate(context.Worksheet), translatedRange);
        }

        internal void Translate()
        {
            this.worksheetReference.Translate();
        }

        internal bool ShouldTranslateUponInsertRemoveCells(ExpressionCloneAndTranslateContext context)
        {
            if (this.cellAreaReference != null)
            {
                if (context.IsHorizontal)
                {
                    if (context.Range.FromIndex.RowIndex <= this.FromCellReference.ActualRowIndex &&
                        this.ToCellReference.ActualRowIndex <= context.Range.ToIndex.RowIndex &&
                        context.Range.FromIndex.ColumnIndex <= this.ToCellReference.ActualColumnIndex)
                    {
                        return true;
                    }

                    if (context.Range.FromIndex.RowIndex <= context.RowIndex &&
                        context.RowIndex <= context.Range.ToIndex.RowIndex &&
                        context.Range.FromIndex.ColumnIndex <= context.ColumnIndex)
                    {
                        return true;
                    }
                }
                else
                {
                    if (context.Range.FromIndex.ColumnIndex <= this.FromCellReference.ActualColumnIndex &&
                        this.ToCellReference.ActualColumnIndex <= context.Range.ToIndex.ColumnIndex &&
                        context.Range.FromIndex.RowIndex <= this.ToCellReference.ActualRowIndex)
                    {
                        return true;
                    }

                    if (context.Range.FromIndex.ColumnIndex <= context.ColumnIndex &&
                        context.ColumnIndex <= context.Range.ToIndex.ColumnIndex &&
                        context.Range.FromIndex.RowIndex <= context.RowIndex)
                    {
                        return true;
                    }
                }
            }

            return false;
        }

        private CellReferenceRange CloneAndTranslateUponInsertRemoveCells(ExpressionCloneAndTranslateContext context)
        {
            if (this.cellAreaReference != null)
            {
                bool translateTarget = false;
                bool translateSource = false;

                int rangeFromIndex;
                int rangeToIndex;
                int referenceFromIndex;
                int referenceToIndex;

                if (context.IsHorizontal)
                {
                    translateTarget = context.Range.FromIndex.RowIndex <= this.FromCellReference.ActualRowIndex &&
                        this.ToCellReference.ActualRowIndex <= context.Range.ToIndex.RowIndex &&
                        context.Range.FromIndex.ColumnIndex <= this.ToCellReference.ActualColumnIndex;

                    translateSource = context.Range.FromIndex.RowIndex <= context.RowIndex &&
                        context.RowIndex <= context.Range.ToIndex.RowIndex &&
                        context.Range.FromIndex.ColumnIndex <= context.ColumnIndex;

                    rangeFromIndex = context.Range.FromIndex.ColumnIndex;
                    rangeToIndex = context.Range.ToIndex.ColumnIndex;
                    referenceFromIndex = this.FromCellReference.ActualColumnIndex;
                    referenceToIndex = this.ToCellReference.ActualColumnIndex;
                }
                else
                {
                    translateTarget = context.Range.FromIndex.ColumnIndex <= this.FromCellReference.ActualColumnIndex &&
                        this.ToCellReference.ActualColumnIndex <= context.Range.ToIndex.ColumnIndex &&
                        context.Range.FromIndex.RowIndex <= this.ToCellReference.ActualRowIndex;

                    translateSource = context.Range.FromIndex.ColumnIndex <= context.ColumnIndex &&
                        context.ColumnIndex <= context.Range.ToIndex.ColumnIndex &&
                        context.Range.FromIndex.RowIndex <= context.RowIndex;

                    rangeFromIndex = context.Range.FromIndex.RowIndex;
                    rangeToIndex = context.Range.ToIndex.RowIndex;
                    referenceFromIndex = this.FromCellReference.ActualRowIndex;
                    referenceToIndex = this.ToCellReference.ActualRowIndex;
                }

                return new CellReferenceRange(this.worksheetReference, this.cellAreaReference.CloneAndTranslateUponInsertRemoveCells(context,
                    rangeFromIndex, rangeToIndex, referenceFromIndex, referenceToIndex, translateSource, translateTarget));
            }

            return null;
        }

        #endregion
    }
}
