﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.DataSeries;
using Telerik.Windows.Documents.Spreadsheet.DataSeries.AutoFill;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings;
using Telerik.Windows.Documents.Spreadsheet.Model.DataValidation;
using Telerik.Windows.Documents.Spreadsheet.Model.Filtering;
using Telerik.Windows.Documents.Spreadsheet.Model.Shapes;
using Telerik.Windows.Documents.Spreadsheet.Model.Sorting;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
#if NETCORE || NETSTANDARD
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a selection of cells.
    /// </summary>
    public class CellSelection : SelectionBase
    {
        #region Nested Types

        // TODO: Remove this after fixing copy - paste with merged cells
        private struct CellPropertiesValues
        {
            public CellBorders Borders { get; set; }
            public IFill Fill { get; set; }
            public ThemableFontFamily FontFamily { get; set; }
            public double FontSize { get; set; }
            public ThemableColor ForeColor { get; set; }
            public CellValueFormat Format { get; set; }
            public RadHorizontalAlignment HorizontalAlignment { get; set; }
            public int Indent { get; set; }
            public bool IsBold { get; set; }
            public bool IsItalic { get; set; }
            public bool IsWrapped { get; set; }
            public string StyleName { get; set; }
            public UnderlineType Underline { get; set; }
            public RadVerticalAlignment VerticalAlignment { get; set; }
        }

        #endregion


        #region Constants

        private static readonly CellValueFormat shortDateFormat = new CellValueFormat(FormatHelper.ShortDatePattern);

        #endregion


        #region Properties

        internal override WorksheetEntityBase WorksheetEntity
        {
            get
            {
                return this.Cells;
            }
        }

        private Cells Cells
        {
            get
            {
                return this.Worksheet.Cells;
            }
        }

        #endregion


        #region Constructors

        internal CellSelection(Worksheet worksheet, IEnumerable<CellRange> cellRanges)
            : base(worksheet, cellRanges)
        {
        }

        #endregion


        #region Methods

        /// <summary>
        /// Sets the borders.
        /// </summary>
        /// <param name="value">The value.</param>
        public override void SetBorders(CellBorders value)
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate(
                (cellRange) =>
                {
                    SetCellBordersCommandContext context = new SetCellBordersCommandContext(this.Worksheet, cellRange, value);
                    this.Worksheet.ExecuteCommand(WorkbookCommands.SetCellBorders, context);
                });
        }

        internal override void PreserveOldStylePropertiesAsLocalIfNeeded<T>(CellRange cellRange, IProperty<T> property, T value)
        {
            if (property.PropertyDefinition == CellPropertyDefinitions.StyleNameProperty)
            {
                ISheet sheet = this.Worksheet;
                using (new UpdateScope(sheet.SuspendLayoutUpdate, sheet.ResumeLayoutUpdate))
                {
                    CellStyle newStyle = this.Worksheet.Workbook.Styles.GetByName(value as string);

                    Action<CellSelection> action = new Action<CellSelection>(
                        (selection) =>
                        {
                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.BottomBorderProperty,
                                CellPropertyDefinitions.BottomBorderProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.DiagonalDownBorderProperty,
                                CellPropertyDefinitions.DiagonalDownBorderProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.DiagonalUpBorderProperty,
                                CellPropertyDefinitions.DiagonalUpBorderProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.FillProperty,
                                CellPropertyDefinitions.FillProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.FontFamilyProperty,
                                CellPropertyDefinitions.FontFamilyProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.FontFamilyProperty,
                                CellPropertyDefinitions.FontFamilyProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.FontSizeProperty,
                                CellPropertyDefinitions.FontSizeProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.ForeColorProperty,
                                CellPropertyDefinitions.ForeColorProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.FormatProperty,
                                CellPropertyDefinitions.FormatProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.HorizontalAlignmentProperty,
                                CellPropertyDefinitions.HorizontalAlignmentProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.IndentProperty,
                                CellPropertyDefinitions.IndentProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.IsBoldProperty,
                                CellPropertyDefinitions.IsBoldProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.IsItalicProperty,
                                CellPropertyDefinitions.IsItalicProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.IsWrappedProperty,
                                CellPropertyDefinitions.IsWrappedProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.LeftBorderProperty,
                                CellPropertyDefinitions.LeftBorderProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.RightBorderProperty,
                                CellPropertyDefinitions.RightBorderProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.TopBorderProperty,
                                CellPropertyDefinitions.TopBorderProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.UnderlineProperty,
                                CellPropertyDefinitions.UnderlineProperty);

                            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.Cells.VerticalAlignmentProperty,
                                CellPropertyDefinitions.VerticalAlignmentProperty);
                        });

                    RangePropertyValue<string> oldStyleName = this.GetStyleName();

                    if (oldStyleName.IsIndeterminate)
                    {
                        ICompressedList<string> propertyValueCollection = this.Cells[cellRange].Cells.PropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.StyleNameProperty);

                        foreach (Range<long, string> range in propertyValueCollection)
                        {
                            CellRange longRangeAsCellRange = CellsPropertyBag.ConvertLongCellRangeToCellRange(range.Start, range.End);
                            if (cellRange.IntersectsWith(longRangeAsCellRange))
                            {
                                longRangeAsCellRange = cellRange.Intersect(longRangeAsCellRange);
                                action(this.Worksheet.Cells[longRangeAsCellRange]);
                            }
                        }
                    }
                    else
                    {
                        action(this);
                    }
                }
            }
        }

        #region Validation for future use

        private bool ValidateSelection(IEnumerable<CellRange> cellRanges)
        {
            int rowCount = this.CellRanges.First().RowCount;
            int firstRowIndex = this.CellRanges.First().GetFirstRow().FromIndex.RowIndex;

            int columnCount = this.CellRanges.First().ColumnCount;
            int firstColumnIndex = this.CellRanges.First().GetFirstColumn().FromIndex.ColumnIndex;

            bool useRows = true;
            bool useColumns = true;

            if (cellRanges.First().ColumnCount != cellRanges.Last().ColumnCount)
            {
                useColumns = false;
            }

            if (cellRanges.First().RowCount != cellRanges.Last().RowCount)
            {
                useRows = false;
            }

            if (!useColumns && !useRows)
            {
                return false;
            }

            foreach (CellRange cellRange in this.CellRanges)
            {
                if (useColumns && (cellRange.ColumnCount != columnCount || cellRange.GetFirstColumn().FromIndex.ColumnIndex != firstColumnIndex))
                {
                    return false;
                }

                if (useRows && (cellRange.RowCount != rowCount || cellRange.GetFirstRow().FromIndex.RowIndex != firstRowIndex))
                {
                    return false;
                }
            }

            return true;
        }

        #endregion

        #endregion


        #region Value Property

        /// <summary>
        /// Gets the value.
        /// </summary>
        /// <returns>Returns the value of the selection.</returns>
        public RangePropertyValue<ICellValue> GetValue()
        {
            return this.GetPropertyValue(this.Cells.ValueProperty);
        }

        /// <summary>
        /// Sets the value.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetValue(DateTime value)
        {
            this.SetValue(value.ToCellValue());

            CellValueFormat currentCellValueFormat = this.GetFormat().Value;
            bool isDefaultCellValueFormat = currentCellValueFormat == CellPropertyDefinitions.FormatProperty.DefaultValue;
            if (isDefaultCellValueFormat)
            {
                this.SetFormat(shortDateFormat);
            }
        }

        internal void SetValueAndExpandToFitNumberValuesWidth(DateTime value)
        {
            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
            {
                this.SetValue(value);
            });
        }

        /// <summary>
        /// Sets the value.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetValue(double value)
        {
            this.SetValue(value.ToCellValue());
        }

        internal void SetValueAndExpandToFitNumberValuesWidth(double value)
        {
            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
            {
                this.SetValue(value);
            });
        }

        /// <summary>
        /// Sets the value.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetValue(bool value)
        {
            this.SetValue(value.ToCellValue());
        }

        internal void SetValueAndExpandToFitNumberValuesWidth(bool value)
        {
            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
                {
                    this.SetValue(value);
                });
        }

        /// <summary>
        /// Sets the value.
        /// </summary>
        /// <remarks>Please note that the string value is parsed to find its type. This process is slow.
        /// If you know the type of your data use the other overloads of this method.</remarks>
        /// <param name="value">The value.</param>
        public void SetValue(string value)
        {
            ICellValue cellValue;
            CellValueFormat newFormat;

            this.CreateValue(value, out cellValue, out newFormat);

            this.SetCellValueInternal(cellValue, newFormat);
        }

        internal void SetValueAndExpandToFitNumberValuesWidth(string value)
        {
            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
            {
                this.SetValue(value);
            });
        }

        internal void SetValueIgnoreErrorsInternal(string value, bool shouldExpandToFitNumberValuesWidth)
        {
            ICellValue cellValue;
            CellValueFormat newFormat;

            this.CreateValueIgnoreErrors(value, out cellValue, out newFormat);

            this.SetCellValueInternal(cellValue, newFormat);

            if (shouldExpandToFitNumberValuesWidth)
            {
                this.Worksheet.Columns[this.CellRanges].ExpandToFitNumberValuesWidth();
            }
        }

        /// <summary>
        /// Sets the value as text.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetValueAsText(string value)
        {
            Guard.ThrowExceptionIfNull(value, "value");
            this.SetValue(value.ToTextCellValue());
        }

        /// <summary>
        /// Sets the value as formula.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetValueAsFormula(string value)
        {
            Guard.ThrowExceptionIfNullOrEmpty(value, "value");

            CellIndex topLeftCellIndex = this.GetTopLeftCellIndex();
            this.SetValue(value.ToFormulaCellValue(this.Worksheet, topLeftCellIndex.RowIndex, topLeftCellIndex.ColumnIndex));
        }

        /// <summary>
        /// Sets the value.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetValue(ICellValue value)
        {
            Guard.ThrowExceptionIfNull(value, "value");

            this.SetCellValueInternal(value, CellValueFormat.GeneralFormat);
        }

        internal void SetValueAndExpandToFitNumberValuesWidth(ICellValue value)
        {
            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
            {
                this.SetValue(value);
            });
        }

        private void SetCellValueInternal(ICellValue value, CellValueFormat format)
        {
            Guard.ThrowExceptionIfNull(value, "value");

            FormulaCellValue formulaCellValue = value as FormulaCellValue;
            if (formulaCellValue != null)
            {
                this.SetFormulaValueInternal(formulaCellValue);
            }
            else
            {
                this.SetValueInternal(value, format);
            }
        }

        private void SetFormulaValueInternal(FormulaCellValue value)
        {
            FunctionExpression functionExpression = value.Value as FunctionExpression;
            CellValueFormat newFormat = CellValueFormat.GeneralFormat;

            if (functionExpression != null && functionExpression.IsValid)
            {
                newFormat = functionExpression.Function.FunctionInfo.Format;
            }

            this.ExecuteForEachRangeInsideBeginEndUpdate((range) =>
            {
                for (int rowIndex = range.FromIndex.RowIndex; rowIndex <= range.ToIndex.RowIndex; rowIndex++)
                {
                    for (int columnIndex = range.FromIndex.ColumnIndex; columnIndex <= range.ToIndex.ColumnIndex; columnIndex++)
                    {
                        this.Cells[rowIndex, columnIndex].SetValueInternal(value.CloneAndTranslate(this.Worksheet, rowIndex, columnIndex), newFormat);
                    }
                }
            });
        }

        internal void SetValueInternal(ICellValue cellValue, CellValueFormat newFormat)
        {
            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                if (SpreadsheetHelper.IsHyperlinkFormula(cellValue))
                {
                    this.SetStyleName(SpreadsheetDefaultValues.HyperlinkStyleName);
                }

                if (!TelerikHelper.EqualsOfT(newFormat, CellValueFormat.GeneralFormat))
                {
                    RangePropertyValue<CellValueFormat> currentFormatRangeValue = this.GetFormat();
                    CellValueFormat currentFormat = currentFormatRangeValue.Value;

                    if (TelerikHelper.EqualsOfT(currentFormat, CellValueFormat.GeneralFormat) && !currentFormatRangeValue.IsIndeterminate)
                    {
                        this.SetFormat(newFormat);
                    }
                    else
                    {
                        this.UpdateCellsFormat(newFormat);
                    }
                }

                this.SetPropertyValue(this.Cells.ValueProperty, cellValue);
            }
        }

        private void UpdateCellsFormat(CellValueFormat newFormat)
        {
            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                this.ExecuteForEachRangeInsideBeginEndUpdate(
                    (CellRange cellRange) =>
                    {
                        UpdateCellPropertyCommandContext<CellValueFormat> context = new UpdateCellPropertyCommandContext<CellValueFormat>(
                            this.Worksheet, CellPropertyDefinitions.FormatProperty, cellRange,
                            (CellValueFormat value) =>
                            {
                                bool oldFormatIsGeneral = value == CellValueFormat.GeneralFormat;
                                bool textFormatIsOverwritten = value.FormatStringInfo.Category == FormatStringCategory.Text &&
                                    newFormat.FormatStringInfo.Category != FormatStringCategory.General &&
                                    newFormat.FormatStringInfo.Category != FormatStringCategory.Number;

                                if (oldFormatIsGeneral || textFormatIsOverwritten)
                                {
                                    return newFormat;
                                }

                                return value;
                            });

                        this.Worksheet.ExecuteCommand(WorkbookCommands.UpdateCellFormat, context);
                    });
            }
        }

        private void CreateValue(string value, out ICellValue cellValue, out CellValueFormat newFormat)
        {
            CellValueFormat currentFormat = this.GetFormat().Value;
            CellIndex topLeftCellIndex = this.GetTopLeftCellIndex();

            CellValueFactory.Create(value, this.Worksheet, topLeftCellIndex.RowIndex, topLeftCellIndex.ColumnIndex, currentFormat, out cellValue, out newFormat);
        }

        internal void CreateValueIgnoreErrors(string value, out ICellValue cellValue, out CellValueFormat newFormat)
        {
            CellValueFormat currentFormat = this.GetFormat().Value;
            CellIndex topLeftCellIndex = this.GetTopLeftCellIndex();

            CellValueFactory.CreateIgnoreErrors(value, this.Worksheet, topLeftCellIndex.RowIndex, topLeftCellIndex.ColumnIndex, currentFormat, out cellValue, out newFormat);
        }

        // TODO: 
        private void CreateValueIgnoreErrorsAndFormat(string value, out ICellValue cellValue, out CellValueFormat newFormat)
        {
            CellIndex topLeftCellIndex = this.GetTopLeftCellIndex();

            CellValueFactory.CreateIgnoreErrors(value, this.Worksheet, topLeftCellIndex.RowIndex, topLeftCellIndex.ColumnIndex, CellValueFormat.GeneralFormat,
                out cellValue, out newFormat);
        }

        internal void CreateValueIgnoreFormat(string value, out ICellValue cellValue, out CellValueFormat newFormat)
        {
            CellValueFormat currentFormat = CellValueFormat.GeneralFormat;

            CellIndex topLeftCellIndex = this.GetTopLeftCellIndex();
            CellValueFactory.Create(value, this.Worksheet, topLeftCellIndex.RowIndex, topLeftCellIndex.ColumnIndex, currentFormat, out cellValue, out newFormat);
        }

        /// <summary>
        /// Clears the value.
        /// </summary>
        public void ClearValue()
        {
            this.ClearPropertyValue(this.Cells.ValueProperty);
        }

        #endregion


        #region Clear

        internal override void Clear(ClearType type, bool clearIsLocked)
        {
            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                if (type == ClearType.All)
                {
                    foreach (IProperty property in this.Cells.Properties)
                    {
                        if (property.PropertyDefinition != CellPropertyDefinitions.IsLockedProperty || clearIsLocked)
                        {
                            foreach (CellRange cellRange in this.CellRanges)
                            {
                                // the selection is whole row selection
                                if (cellRange.ColumnCount == SpreadsheetDefaultValues.ColumnCount)
                                {
                                    IProperty rowProperty;
                                    if (this.Worksheet.Rows.TryGetProperyFromPropertyDefinition(property.PropertyDefinition, out rowProperty))
                                    {
                                        this.Worksheet.Rows[cellRange.FromIndex.RowIndex, cellRange.ToIndex.RowIndex].ClearPropertyValue(rowProperty);
                                    }
                                }

                                // the selection is whole column selection
                                if (cellRange.RowCount == SpreadsheetDefaultValues.RowCount)
                                {
                                    IProperty rowProperty;
                                    if (this.Worksheet.Columns.TryGetProperyFromPropertyDefinition(property.PropertyDefinition, out rowProperty))
                                    {
                                        this.Worksheet.Columns[cellRange.FromIndex.ColumnIndex, cellRange.ToIndex.ColumnIndex].ClearPropertyValue(rowProperty);
                                    }
                                }
                            }

                            this.ClearPropertyValue(property);
                        }
                    }

                    this.ClearHyperlinks();
                }
                else if (type == ClearType.Formats)
                {
                    foreach (IProperty property in this.Cells.Properties)
                    {
                        if (property != this.Cells.ValueProperty)
                        {
                            this.ClearPropertyValue(property);
                        }
                    }
                }
                else if (type == ClearType.Contents)
                {
                    this.ClearValue();
                    this.ClearHyperlinks();
                }
                else if (type == ClearType.Hyperlinks)
                {
                    this.ClearHyperlinks();
                }
                else
                {
                    throw new InvalidOperationException();
                }
            }
        }

        internal override void ClearBySettingDefaultValuesOnRowColumnPropertyIntersection(CellRange cellRange, IProperty property)
        {
            Action<RowsColumnsBase, PropertyBagBase, Func<int, int, CellRange>> clearAction = new Action<RowsColumnsBase, PropertyBagBase, Func<int, int, CellRange>>(
                (rowsColumns, propertyBag, createRowColumnRangeAction) =>
                {
                    IProperty rowColumnProperty;
                    if (rowsColumns.TryGetProperyFromPropertyDefinition(property.PropertyDefinition, out rowColumnProperty))
                    {
                        ICompressedList propertyValueCollection = propertyBag.GetPropertyValueCollection(rowColumnProperty.PropertyDefinition);
                        foreach (LongRange range in propertyValueCollection.GetRanges(false))
                        {
                            CellRange rowRange = createRowColumnRangeAction((int)range.Start, (int)range.End);
                            if (cellRange.IntersectsWith(rowRange))
                            {
                                CellRange rangeToClear = cellRange.Intersect(rowRange);
                                property.SetDefaultValue(rangeToClear);
                            }
                        }
                    }
                });

            clearAction(this.Worksheet.Rows, this.Worksheet.Rows.PropertyBag, CellRange.FromRowRange);
            clearAction(this.Worksheet.Columns, this.Worksheet.Columns.PropertyBag, CellRange.FromColumnRange);
        }

        private void ClearHyperlinks()
        {
            this.Worksheet.Hyperlinks.RemoveRange(this.Worksheet.Hyperlinks.GetContainingHyperlinks(this.CellRanges), false);
        }

        #endregion


        #region Merge/Unmerge

        /// <summary>
        /// Merges this instance.
        /// </summary>
        /// <returns>The value indicating whether the merge was successful.</returns>
        public bool Merge()
        {
            if (this.CellRanges.ContainsOverlappingRanges())
            {
                return false;
            }


            using (new UpdateScope(this.Cells.MergedCellRanges.BeginUpdate, this.Cells.MergedCellRanges.EndUpdate))
            {
                this.ExecuteForEachRangeInsideBeginEndUpdate(
                    (cellRange) =>
                    {
                        this.MergeInternal(cellRange);
                    });
            }

            return true;
        }

        private void MergeInternal(CellRange cellRange)
        {
            this.PrepareForMerge(cellRange);

            MergeCellsCommandContext context = new MergeCellsCommandContext(this.Worksheet, cellRange);
            this.Worksheet.ExecuteCommand(WorkbookCommands.MergeCells, context);
        }

        private void PrepareForMerge(CellRange cellRange)
        {
            CellSelection sourceCells = this.Cells[cellRange.FromIndex];

            CellPropertiesValues values = this.GetFormatProperties(sourceCells);

            foreach (CellRange range in cellRange.SplitToRangesSurroundingRange(cellRange.FromIndex.ToCellRange()))
            {
                CellSelection destinationCells = this.Cells[range];
                using (new UpdateScope(destinationCells.BeginUpdate, destinationCells.EndUpdate))
                {
                    destinationCells.Clear(ClearType.Contents);
                    this.SetFormatProperties(values, destinationCells);
                }
            }
            // TODO: Use Copy - Paste
            //WorksheetFragment fragment = this.Cells[cellRange.FromIndex].Copy();
            //foreach (var range in cellRange.SplitToRangesSurroundingRange(cellRange.FromIndex.ToCellRange()))
            //{
            //    this.Cells[range].Clear(ClearType.Contents);
            //    this.Cells[range].Paste(fragment, new PasteOptions(PasteType.Formats));
            //}
        }

        private CellPropertiesValues GetFormatProperties(CellSelection sourceCells)
        {
            CellPropertiesValues values = new CellPropertiesValues();

            values.Borders = sourceCells.GetBorders();
            values.Fill = sourceCells.GetFill().Value;
            values.FontFamily = sourceCells.GetFontFamily().Value;
            values.FontSize = sourceCells.GetFontSize().Value;
            values.ForeColor = sourceCells.GetForeColor().Value;
            values.Format = sourceCells.GetFormat().Value;
            values.HorizontalAlignment = sourceCells.GetHorizontalAlignment().Value;
            values.Indent = sourceCells.GetIndent().Value;
            values.IsBold = sourceCells.GetIsBold().Value;
            values.IsItalic = sourceCells.GetIsItalic().Value;
            values.IsWrapped = sourceCells.GetIsWrapped().Value;
            values.StyleName = sourceCells.GetStyleName().Value;
            values.Underline = sourceCells.GetUnderline().Value;
            values.VerticalAlignment = sourceCells.GetVerticalAlignment().Value;

            return values;
        }

        private void SetFormatProperties(CellPropertiesValues values, CellSelection destinationCells)
        {
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.StyleNameProperty, values.StyleName);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.BottomBorderProperty, values.Borders.Bottom);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.DiagonalDownBorderProperty, values.Borders.DiagonalDown);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.DiagonalUpBorderProperty, values.Borders.DiagonalUp);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.LeftBorderProperty, values.Borders.Left);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.RightBorderProperty, values.Borders.Right);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.TopBorderProperty, values.Borders.Top);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.FillProperty, values.Fill);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.FontFamilyProperty, values.FontFamily);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.FontSizeProperty, values.FontSize);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.ForeColorProperty, values.ForeColor);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.FormatProperty, values.Format);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.IndentProperty, values.Indent);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.HorizontalAlignmentProperty, values.HorizontalAlignment);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.IsBoldProperty, values.IsBold);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.IsItalicProperty, values.IsItalic);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.IsWrappedProperty, values.IsWrapped);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.UnderlineProperty, values.Underline);
            this.SetFormatPropertyIfNeeded(destinationCells, CellPropertyDefinitions.VerticalAlignmentProperty, values.VerticalAlignment);

            destinationCells.Worksheet.Columns[destinationCells.CellRanges].ExpandToFitNumberValuesWidth();
        }

        private void SetFormatPropertyIfNeeded<T>(CellSelection destinationCells, IPropertyDefinition<T> propertyDefinition, T value)
        {
            IProperty<T> property;
            this.Cells.TryGetProperyFromPropertyDefinition(propertyDefinition, out property);

            if (TelerikHelper.EqualsOfT(value, propertyDefinition.DefaultValue))
            {
                ICompressedList<T> valuesCompressedList = this.Cells.PropertyBag.GetPropertyValueCollection(propertyDefinition);
                foreach (CellRange range in destinationCells.CellRanges)
                {
                    foreach (LongRange longRange in CellsPropertyBag.ConvertCellRangeToLongRanges(range))
                    {
                        ICompressedList<T> longRangeValues = valuesCompressedList.GetValue(longRange.Start, longRange.End);
                        foreach (Range<long, T> valueRange in longRangeValues.GetNonDefaultRanges())
                        {
                            CellRange clearCellRange = CellsPropertyBag.ConvertLongCellRangeToCellRange(valueRange.Start, valueRange.End);
                            property.ClearValue(clearCellRange);
                        }
                    }
                }
            }
            else
            {
                destinationCells.SetPropertyValue(property, value);
            }
        }

        /// <summary>
        /// Merges the instance across.
        /// </summary>
        /// <returns>The value indicating whether the merge was successful.</returns>
        public bool MergeAcross()
        {
            if (this.CellRanges.ContainsOverlappingRanges())
            {
                return false;
            }

            this.ExecuteForEachRangeInsideBeginEndUpdate(
                (cellRange) =>
                {
                    for (int rowIndex = cellRange.FromIndex.RowIndex; rowIndex <= cellRange.ToIndex.RowIndex; rowIndex++)
                    {
                        CellRange cellRangePart = new CellRange(rowIndex, cellRange.FromIndex.ColumnIndex, rowIndex, cellRange.ToIndex.ColumnIndex);

                        this.PrepareForMerge(cellRangePart);

                        MergeCellsCommandContext context = new MergeCellsCommandContext(this.Worksheet, cellRangePart);
                        this.Worksheet.ExecuteCommand(WorkbookCommands.MergeCells, context);
                    }
                });

            return true;
        }

        /// <summary>
        /// Unmerges this instance.
        /// </summary>
        public void Unmerge()
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate(
                (cellRange) =>
                {
                    this.UnmergeInternal(cellRange);
                });
        }

        private void UnmergeInternal(CellRange cellRange)
        {
            UnmergeCellsCommandContext context = new UnmergeCellsCommandContext(this.Worksheet, cellRange);
            this.Worksheet.ExecuteCommand(WorkbookCommands.UnmergeCells, context);
        }

        #endregion


        #region Copy Paste

        /// <summary>
        /// Copies this instance.
        /// </summary>
        /// <returns>The worksheet fragment.</returns>
        public WorksheetFragment Copy()
        {
            return this.CopyInternal().Fragment;
        }

        internal CopyResult CopyInternal()
        {
            bool success = true;
            string errorMessage = string.Empty;
            string errorMessageLocalizationKey = string.Empty;
            WorksheetFragment fragment = null;

            if (this.CellRanges.Count() > 1)
            {
                success = false;
                errorMessage = "That command cannot be used on multiple selections.";
                errorMessageLocalizationKey = "Spreadsheet_MessageBox_CommandCannotBeUsed";
                fragment = null;
            }

            if (this.Cells.MergedCellRanges.MergedCellIsSplitByHiddenRowColumn(this.CellRanges.First()))
            {
                success = false;
                errorMessage = "This cannot be done to a merged cell.";
                fragment = null;
            }

            if (success)
            {
                CellRange range = this.Cells.MergedCellRanges.ExpandRangeToNotIntersectMergedCells(this.CellRanges.First());
                fragment = new WorksheetFragment(range, null, this.Worksheet, false, true);
            }

            return new CopyResult(success, fragment, errorMessage, errorMessageLocalizationKey);
        }

        /// <summary>
        /// Determines whether the specified fragment can be copied.
        /// </summary>
        /// <param name="fragment">The fragment.</param>
        /// <returns>The value indicating whether the specified fragment can be copied.</returns>
        public bool CanCopy(WorksheetFragment fragment)
        {
            if (fragment == null)
            {
                return false;
            }

            return true;
        }

        /// <summary>
        /// Pastes the specified fragment.
        /// </summary>
        /// <param name="fragment">The fragment.</param>
        /// <param name="pasteOptions">The paste options.</param>
        /// <returns>The value indicating whether the paste was successful.</returns>
        public bool Paste(WorksheetFragment fragment, PasteOptions pasteOptions)
        {
            IEnumerable<CellRange> affectedcellRangesList;
            IEnumerable<FloatingShapeBase> newShapes;
            return this.Paste(fragment, pasteOptions, out affectedcellRangesList, out newShapes);
        }

        /// <summary>
        /// Pastes the specified fragment.
        /// </summary>
        /// <param name="fragment">The fragment.</param>
        /// <param name="pasteOptions">The paste options.</param>
        /// <param name="newShapes">The new shapes.</param>
        /// <returns></returns>
        public bool Paste(WorksheetFragment fragment, PasteOptions pasteOptions, out IEnumerable<FloatingShapeBase> newShapes)
        {
            IEnumerable<CellRange> affectedcellRangesList;
            return this.Paste(fragment, pasteOptions, out affectedcellRangesList, out newShapes);
        }

        /// <summary>
        /// Determines whether the specified fragment can be pasted.
        /// </summary>
        /// <param name="fragment">The fragment.</param>
        /// <returns>The paste result.</returns>
        public PasteResult CanPaste(WorksheetFragment fragment)
        {
            bool success = true;
            string errorMessage = string.Empty;
            string errorMessageLocalizationKey = string.Empty;
            bool shouldDisableHistory = false;

            if (fragment == null)
            {
                success = false;
                errorMessage = "The information cannot be pasted.";
                errorMessageLocalizationKey = "Spreadsheet_ErrorExpressions_CannotPasteGeneral";
            }
            else if (fragment.IsCellCopyPaste)
            {
                CellRange pasteCellRange = this.GetPasteCellRange(fragment);

                if (pasteCellRange == null)
                {
                    success = false;
                    errorMessage = "The information cannot be pasted beacause the Copy area and \r\n the paste area are not the same size and shape.";
                    errorMessageLocalizationKey = "Spreadsheet_ErrorExpressions_CannotPasteDueToSize";
                }
                else if (!this.CanChangeCellContent(pasteCellRange))
                {
                    success = false;
                    errorMessage = "The cell you are trying to change is on a protected sheet. To make changes, unprotect the sheet.";
                    errorMessageLocalizationKey = "Spreadsheet_ProtectedWorksheet_Error";
                }
                else if (this.Cells.MergedCellRanges.GetIntersectingMergedRanges(pasteCellRange).Count > 0)
                {
                    success = false;
                    errorMessage = "Cannot change part of a merged cell.";
                    errorMessageLocalizationKey = "Spreadsheet_ErrorExpressions_CannotChangeMergedCell";
                }
                else if (WorksheetFragment.GetPasteCellRangesCount(fragment.FromCellRanges, pasteCellRange) > WorksheetFragment.MaxAllowedAffectedRanges)
                {
                    success = false;
                    errorMessage = "The operation you requested affects a large number of cells and may not be possible with the available resources.";
                }
                else if (WorksheetFragment.GetPasteCellRangesCount(fragment.FromCellRanges, pasteCellRange) > WorksheetFragment.MaxAllowedAffectedRangesWithUndo)
                {
                    success = false;
                    shouldDisableHistory = true;
                    errorMessage = "The operation you requested affects a large number of cells and may not be possible with the available resources. Would you like to continue without undo?";
                }
            }

            return new PasteResult(success, errorMessage, errorMessageLocalizationKey, shouldDisableHistory);
        }

        internal CellRange GetPasteCellRange(WorksheetFragment fragment)
        {
            Guard.ThrowExceptionIfNull(fragment, "fragment");

            if (this.CellRanges.Count() > 1)
            {
                return null;
            }

            CellRange currentCellRange = this.CellRanges.First();
            return WorksheetFragment.EnsureProportionalPasteCellRange(fragment.FromCellRanges, currentCellRange);
        }

        /// <summary>
        /// Pastes the specified fragment.
        /// </summary>
        /// <param name="fragment">The fragment.</param>
        /// <param name="pasteOptions">The paste options.</param>
        /// <param name="affectedCellRanges">The affected cell ranges.</param>
        /// <param name="newShapes">The new shapes.</param>
        /// <returns>The value indicating whether the paste was successful.</returns>
        public bool Paste(WorksheetFragment fragment, PasteOptions pasteOptions, out IEnumerable<CellRange> affectedCellRanges, out IEnumerable<FloatingShapeBase> newShapes)
        {
            Guard.ThrowExceptionIfNull(fragment, "fragment");

            List<CellRange> affectedcellRangesList;
            bool result = fragment.Paste(this.Worksheet, this.CellRanges.First(), null, pasteOptions, out affectedcellRangesList, out newShapes, fragment.IsCellCopyPaste);
            affectedCellRanges = affectedcellRangesList;

            return result;
        }

        internal bool CanChangeCellContent(CellRange cellRange)
        {
            RangePropertyValue<bool> isLocked = this.Cells[cellRange].GetIsLocked();

            return !(this.Worksheet.IsProtected && (isLocked.IsIndeterminate || isLocked.Value));
        }

        internal bool CanChangeCellContent()
        {
            return !this.Worksheet.IsProtected || !this.SelectedCellRangesContainsLockedCell();
        }

        private bool SelectedCellRangesContainsLockedCell()
        {
            foreach (CellRange cellRange in this.CellRanges)
            {
                if (this.Cells[cellRange].GetIsLocked().Value)
                {
                    return true;
                }
            }

            return false;
        }

        #endregion


        #region Insert/Remove

        /// <summary>
        /// Inserts this instance using the specified shift type.
        /// </summary>
        /// <param name="shiftType">Type of the shift.</param>
        public void Insert(InsertShiftType shiftType)
        {
            History.WorkbookHistory history = this.Cells.Worksheet.Workbook.History;
            using (new UpdateScope(history.BeginUndoGroup, history.EndUndoGroup))
            {
                this.ExecuteForEachRangeInsideBeginEndUpdate((cellRange) =>
                {
                    this.Worksheet.Cells.Insert(cellRange, shiftType);
                });
            }
        }

        /// <summary>
        /// Removes the instance using the specified shift type.
        /// </summary>
        /// <param name="shiftType">Type of the shift.</param>
        public void Remove(RemoveShiftType shiftType)
        {
            History.WorkbookHistory history = this.Cells.Worksheet.Workbook.History;
            using (new UpdateScope(history.BeginUndoGroup, history.EndUndoGroup))
            {
                this.ExecuteForEachRangeInsideBeginEndUpdate((cellRange) =>
                {
                    this.Worksheet.Cells.Remove(cellRange, shiftType);
                });
            }
        }

        /// <summary>
        /// Determines whether the selected range can be inserted or removed using the specified shift type.
        /// </summary>
        /// <param name="selectedRange">The selected range.</param>
        /// <param name="shiftType">Type of the shift.</param>
        /// <returns>The value indicating whether the selected range can be inserted or removed using the specified shift type.</returns>
        public bool CanInsertOrRemove(CellRange selectedRange, ShiftType shiftType)
        {
            // TODO:
            //return this.Worksheet.Cells.MergedCellRanges.CanInsertOrRemove(selectedRange, shiftType);
            CanInsertRemoveResult canInsert = this.CanInsertOrRemoveInternal(selectedRange, shiftType);

            return canInsert == CanInsertRemoveResult.Success;
        }

        internal CanInsertRemoveResult CanInsertOrRemoveInternal(CellRange selectedRange, ShiftType shiftType)
        {
            CanInsertRemoveResult canInsertRemove = CanInsertRemoveResult.Success;

            InsertShiftType insertShift;
            if (InsertRemoveCellsHelper.TryGetInsertShiftType(shiftType, out insertShift))
            {
                canInsertRemove = this.Worksheet.Cells.CanInsertInternal(selectedRange, insertShift);
            }

            RemoveShiftType removeShift;
            if (InsertRemoveCellsHelper.TryGetRemoveShiftType(shiftType, out removeShift))
            {
                canInsertRemove = this.Worksheet.Cells.CanRemoveInternal(selectedRange, removeShift);
            }

            return canInsertRemove;
        }

        #endregion


        #region Data Series

        private int GetInitialValuesCount(ICellValue[] initialValues, CellIndex[] indexes)
        {
            for (int i = initialValues.Length - 1; i >= 0; i--)
            {
                CellRange mergedRange;
                this.Worksheet.Cells.TryGetContainingMergedRange(indexes[i], out mergedRange);

                if (mergedRange != null && indexes.Contains(mergedRange.FromIndex))
                {
                    int index = Array.IndexOf(indexes, mergedRange.FromIndex);
                    if (initialValues[index].ValueType != CellValueType.Empty)
                    {
                        return i + 1;
                    }
                }

                if (initialValues[i].ValueType != CellValueType.Empty)
                {
                    return i + 1;
                }
            }

            return initialValues.Length;
        }

        private void RepeatFormatProperties(CellIndex[] indexes, int initialValuesCount)
        {
            if (indexes.Length == 0)
            {
                return;
            }

            for (int i = 0; i < initialValuesCount; i++)
            {
                CellSelection sourceCellSelection = this.Worksheet.Cells[indexes[i]];

                CellPropertiesValues values = this.GetFormatProperties(sourceCellSelection);
                for (int j = i + initialValuesCount; j < indexes.Length; j += initialValuesCount)
                {
                    CellSelection targetCellSelection = this.Worksheet.Cells[indexes[j]];
                    this.SetFormatProperties(values, targetCellSelection);
                }
            }

            // TODO: Use Copy - Paste. The reason why it is not used at the moment is that Paste requires the initial range to fit evenly into the new range.
            // Filling doesn't have such requirement.

            //CellRange sourceCellRange = new CellRange(indexes[0], indexes[initialValuesCount - 1]);

            //var fragment = this.Worksheet.Cells[sourceCellRange].Copy();

            //CellRange targetCellRange = new CellRange(indexes[initialValuesCount], indexes[indexes.Length - 1]);
            //this.Worksheet.Cells[targetCellRange].Paste(fragment, new PasteOptions(PasteType.Formats));
        }

        private void FillDataSeries(CellRange cellRange, CellOrientation seriesOrientation, Action<CellIndex[], int> fillGenerator, int? initialValuesCount = null, bool isDirectionReversed = false)
        {
            this.FillDataSeries(null, cellRange, seriesOrientation, fillGenerator, initialValuesCount, isDirectionReversed);
        }

        private void FillDataSeries(ValuesInfo valuesInfo, CellRange cellRange, CellOrientation seriesOrientation, Action<CellIndex[], int> fillGenerator, int? initialValuesCount = null, bool isDirectionReversed = false)
        {
            if (valuesInfo == null)
            {
                valuesInfo = this.GetValuesInfo(cellRange, seriesOrientation, isDirectionReversed, initialValuesCount);
            }

            for (int i = 0; i < valuesInfo.CellIndexLists.Count; i++)
            {
                fillGenerator(valuesInfo.CellIndexLists[i], valuesInfo.InitialValuesCounts[i]);
            }
        }

        private ValuesInfo GetValuesInfo(CellRange cellRange, CellOrientation seriesOrientation, bool isDirectionReversed, int? initialValuesCount)
        {
            CellRangeEnumerator cellRangeEnumerator = new CellRangeEnumerator(cellRange, seriesOrientation, isDirectionReversed);

            ValuesInfo result = new ValuesInfo();

            List<CellIndex> currentList = null;
            while (cellRangeEnumerator.MoveNext())
            {
                if (cellRangeEnumerator.IsAtRowColumnStart)
                {
                    if (currentList != null)
                    {
                        result.CellIndexLists.Add(currentList.ToArray());
                    }
                    currentList = new List<CellIndex>();
                }

                currentList.Add(cellRangeEnumerator.Current);
            }

            result.CellIndexLists.Add(currentList.ToArray());

            foreach (CellIndex[] indexList in result.CellIndexLists)
            {
                ICellValue[] initialValues = new ICellValue[indexList.Length];
                for (int i = 0; i < indexList.Length; i++)
                {
                    CellIndex cellIndex = indexList[i];
                    ICellValue cellValue = this.Worksheet.Cells[cellIndex].GetValue().Value;
                    initialValues[i] = cellValue;
                    if (i == initialValuesCount - 1)
                    {
                        break;
                    }
                }

                result.InitialValuesList.Add(initialValues);
                if (initialValuesCount == null)
                {
                    result.InitialValuesCounts.Add(this.GetInitialValuesCount(initialValues, indexList));
                }
                else
                {
                    result.InitialValuesCounts.Add(initialValuesCount.Value);
                }
            }

            return result;
        }

        private CellIndex[] RebuildCellIdexesRespectingMergedCells(CellIndex[] indexes)
        {
            List<CellIndex> newIndexes = new List<CellIndex>();
            for (int i = 0; i < indexes.Length; i++)
            {
                if (newIndexes.Count == 0)
                {
                    newIndexes.Add(indexes[i]);
                }
                else
                {
                    CellRange mergedRange;
                    if (this.Worksheet.Cells.TryGetContainingMergedRange(indexes[i], out mergedRange))
                    {
                        CellRange previousIndexMergedRange;
                        if (this.Worksheet.Cells.TryGetContainingMergedRange(indexes[i - 1], out previousIndexMergedRange))
                        {
                            if (mergedRange.IntersectsWith(previousIndexMergedRange))
                            {
                                continue;
                            }
                        }
                    }

                    newIndexes.Add(indexes[i]);
                }
            }

            return newIndexes.ToArray();
        }

        private void ApplyChanges(double[] doubleResult, CellIndex[] indexes)
        {
            if (doubleResult == null)
            {
                return;
            }

            for (int i = 0; i < doubleResult.Length; i++)
            {
                this.Worksheet.Cells[indexes[i]].SetValueAndExpandToFitNumberValuesWidth(doubleResult[i]);
            }
        }

        private bool InitializeBeforeFill(int initialIndexesCount, ref CellIndex[] indexes, out double?[] initialDoubleValues)
        {
            initialDoubleValues = new double?[initialIndexesCount];

            if (!indexes[0].ToCellValue(this.Worksheet).GetAsDoubleOrNull().HasValue)
            {
                return true;
            }

            CellRange mergedRange = null;
            this.Worksheet.Cells.TryGetContainingMergedRange(indexes[0], out mergedRange);

            if (mergedRange != null)
            {
                FillDataSeriesHelper.RepeatMergedRanges(this.Worksheet, new CellRange[] { mergedRange }, indexes, 1);
            }

            indexes = this.RebuildCellIdexesRespectingMergedCells(indexes);

            int initialDoubleValuesCounter = 0;

            if (indexes.Length > initialIndexesCount)
            {
                for (int i = 0; i < initialIndexesCount; i++)
                {
                    double? doubleValue = indexes[i].ToCellValue(this.Worksheet).GetAsDoubleOrNull();

                    initialDoubleValues[i] = doubleValue;
                    initialDoubleValuesCounter++;
                }
            }

            if (initialDoubleValuesCounter == 0)
            {
                return true;
            }

            return false;
        }

        private void FillRespectingMergedCells(int initialIndexesCount, double stepValue, double? stopValue, CellIndex[] indexes, Func<double, int, double, double?, double[]> generator)
        {
            double?[] initialDoubleValues;
            if (this.InitializeBeforeFill(initialIndexesCount, ref indexes, out initialDoubleValues))
            {
                return;
            }

            double[] doubleResult = generator(initialDoubleValues[0].Value, indexes.Length, stepValue, stopValue);

            this.ApplyChanges(doubleResult, indexes);
        }

        private void FillRespectingMergedCells(int initialIndexesCount, double stepValue, double? stopValue, CellIndex[] indexes, DateUnitType dateUnitType, Func<double, DateUnitType, int, double, double?, double[]> generator)
        {
            double?[] initialDoubleValues;
            if (this.InitializeBeforeFill(initialIndexesCount, ref indexes, out initialDoubleValues))
            {
                return;
            }

            double[] doubleResult = generator(initialDoubleValues[0].Value, dateUnitType, indexes.Length, stepValue, stopValue);

            this.ApplyChanges(doubleResult, indexes);
        }

        private void FillRespectingMergedCells(int initialIndexesCount, CellIndex[] indexes, Func<double?[], int, double[]> generator)
        {
            double?[] initialDoubleValues;
            if (this.InitializeBeforeFill(initialIndexesCount, ref indexes, out initialDoubleValues))
            {
                return;
            }

            double[] doubleResult = generator(initialDoubleValues, indexes.Length);

            this.ApplyChanges(doubleResult, indexes);
        }

        internal CellRange GetRangeAffectedByFillData(CellRange newSelection, FillDirection direction, int initialValuesCount)
        {
            int fromRowIndex = newSelection.FromIndex.RowIndex;
            int fromColumnIndex = newSelection.FromIndex.ColumnIndex;
            int toRowIndex = newSelection.ToIndex.RowIndex;
            int toColumnIndex = newSelection.ToIndex.ColumnIndex;

            switch (direction)
            {
                case FillDirection.Down:
                    fromRowIndex += initialValuesCount; break;
                case FillDirection.Up:
                    toRowIndex -= initialValuesCount; break;
                case FillDirection.Right:
                    fromColumnIndex += initialValuesCount; break;
                case FillDirection.Left:
                    toColumnIndex -= initialValuesCount; break;
            }

            return new CellRange(fromRowIndex, fromColumnIndex, toRowIndex, toColumnIndex);
        }

        /// <summary>
        /// Does linear data series fill.
        /// </summary>
        /// <param name="seriesOrientation">The series orientation.</param>
        /// <param name="stepValue">The step value.</param>
        /// <param name="stopValue">The stop value.</param>
        public void FillDataSeriesLinear(CellOrientation seriesOrientation, double stepValue, double? stopValue = null)
        {
            Action<CellRange> fillRangeAction =
                (cellRange) =>
                {
                    this.FillDataSeries(cellRange, seriesOrientation,
                    (indexes, initialIndexesCount) =>
                    {
                        this.FillRespectingMergedCells(initialIndexesCount, stepValue, stopValue, indexes, FillDataSeriesHelper.FillLinear);

                        CellIndex[] indexList = this.RebuildCellIdexesRespectingMergedCells(indexes);
                        this.RepeatFormatProperties(indexList, 1);
                    },
                    1);
                };

            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
                {
                    this.ExecuteForEachRangeInsideBeginEndUpdate(fillRangeAction);
                });
        }

        /// <summary>
        /// Does trend data series fill.
        /// </summary>
        /// <param name="seriesOrientation">The series orientation.</param>
        public void FillDataSeriesLinearTrend(CellOrientation seriesOrientation)
        {
            Action<CellRange> fillRangeAction =
                (cellRange) =>
                {
                    this.FillDataSeries(cellRange, seriesOrientation,
                        (indexes, initialIndexesCount) =>
                        {
                            this.FillRespectingMergedCells(initialIndexesCount, indexes, FillDataSeriesHelper.FillLinearTrend);

                            CellIndex[] indexList = this.RebuildCellIdexesRespectingMergedCells(indexes);
                            this.RepeatFormatProperties(indexList, 1);
                        });
                };

            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
                {
                    this.ExecuteForEachRangeInsideBeginEndUpdate(fillRangeAction);
                });
        }

        /// <summary>
        /// Does exponential data series fill.
        /// </summary>
        /// <param name="seriesOrientation">The series orientation.</param>
        /// <param name="stepValue">The step value.</param>
        /// <param name="stopValue">The stop value.</param>
        public void FillDataSeriesExponential(CellOrientation seriesOrientation, double stepValue, double? stopValue = null)
        {
            Action<CellRange> fillRangeAction =
                (cellRange) =>
                {
                    this.FillDataSeries(cellRange, seriesOrientation,
                        (indexes, initialIndexesCount) =>
                        {
                            this.FillRespectingMergedCells(initialIndexesCount, stepValue, stopValue, indexes, FillDataSeriesHelper.FillExponential);

                            CellIndex[] indexList = this.RebuildCellIdexesRespectingMergedCells(indexes);
                            this.RepeatFormatProperties(indexList, 1);
                        },
                        1);
                };

            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
                {
                    this.ExecuteForEachRangeInsideBeginEndUpdate(fillRangeAction);
                });
        }

        /// <summary>
        /// Does exponential and trend data series fill.
        /// </summary>
        /// <param name="seriesOrientation">The series orientation.</param>
        public void FillDataSeriesExponentialTrend(CellOrientation seriesOrientation)
        {
            Action<CellRange> fillRangeAction =
                (cellRange) =>
                {
                    this.FillDataSeries(cellRange, seriesOrientation,
                        (indexes, initialIndexesCount) =>
                        {
                            this.FillRespectingMergedCells(initialIndexesCount, indexes, FillDataSeriesHelper.FillExponentialTrend);

                            CellIndex[] indexList = this.RebuildCellIdexesRespectingMergedCells(indexes);
                            this.RepeatFormatProperties(indexList, 1);
                        });
                };

            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
                {
                    this.ExecuteForEachRangeInsideBeginEndUpdate(fillRangeAction);
                });
        }

        /// <summary>
        /// Does date data series fill.
        /// </summary>
        /// <param name="seriesOrientation">The series orientation.</param>
        /// <param name="dateUnitType">Type of the date unit.</param>
        /// <param name="stepValue">The step value.</param>
        /// <param name="stopValue">The stop value.</param>
        public void FillDataSeriesDate(CellOrientation seriesOrientation, DateUnitType dateUnitType, double stepValue, double? stopValue = null)
        {
            Action<CellRange> fillRangeAction =
                (cellRange) =>
                {
                    this.FillDataSeries(cellRange, seriesOrientation,
                        (indexes, initialIndexesCount) =>
                        {
                            this.FillRespectingMergedCells(initialIndexesCount, stepValue, stopValue, indexes, dateUnitType, FillDataSeriesHelper.FillDate);

                            CellIndex[] indexList = this.RebuildCellIdexesRespectingMergedCells(indexes);
                            this.RepeatFormatProperties(indexList, 1);
                        },
                        1);
                };

            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
                {
                    this.ExecuteForEachRangeInsideBeginEndUpdate(fillRangeAction);
                });
        }

        /// <summary>
        /// Does auto data series fill.
        /// </summary>
        /// <param name="seriesOrientation">The series orientation.</param>
        /// <param name="respectRangesDirection">The respect ranges direction.</param>
        /// <param name="initialValueCount">The initial value count.</param>
        public void FillDataSeriesAuto(CellOrientation seriesOrientation, bool respectRangesDirection, int? initialValueCount = null)
        {
            int maxInitialValuesCount = 0;
            ValuesInfo valuesInfo = new ValuesInfo();
            Dictionary<CellRange, HashSet<CellRange>> mergedRanges = new Dictionary<CellRange, HashSet<CellRange>>();

            Func<CellRange, bool> getIsDirectionReserved =
                (cellRange) =>
                {
                    return respectRangesDirection &&
                        ((seriesOrientation == CellOrientation.Horizontal && !cellRange.IsLeftToRight) ||
                        (seriesOrientation == CellOrientation.Vertical && !cellRange.IsTopToBottom));
                };

            Action<CellRange> fillRangeAction =
                (cellRange) =>
                {
                    HashSet<CellRange> currentRangeMergedCells = new HashSet<CellRange>();
                    mergedRanges.Add(cellRange, currentRangeMergedCells);
                    bool isDirectionReversed = getIsDirectionReserved(cellRange);

                    valuesInfo = this.GetValuesInfo(cellRange, seriesOrientation, isDirectionReversed, initialValueCount);
                    maxInitialValuesCount = valuesInfo.MaxInitialValuesCount;

                    foreach (CellIndex[] indexes in valuesInfo.CellIndexLists)
                    {
                        // TODO: Preserve HorizontalAlignment
                        CellRange[] mergedRangesArray = FillDataSeriesHelper.ExtractMergedRanges(this.Worksheet, indexes, maxInitialValuesCount);
                        if (mergedRangesArray.Length > 0)
                        {
                            FillDataSeriesHelper.UnmergeMergedRanges(this.Worksheet, mergedRangesArray);
                            for (int i = 0; i < mergedRangesArray.Length; i++)
                            {
                                CellRange range = mergedRangesArray[i];
                                if (!currentRangeMergedCells.Contains(range))
                                {
                                    currentRangeMergedCells.Add(range);
                                }
                            }
                        }
                    }

                    this.Worksheet.Cells[cellRange].Unmerge();

                    this.FillDataSeries(valuesInfo, cellRange, seriesOrientation,
                        (indexes, initialIndexesCount) =>
                        {
                            AutoFillDataSeriesManager autoFillManger = new AutoFillDataSeriesManager(this.Worksheet);
                            autoFillManger.FillAuto(indexes, maxInitialValuesCount);
                        }, null, isDirectionReversed);
                };

            Action<CellRange> mergeAction =
                (cellRange) =>
                {
                    bool isDirectionReversed = getIsDirectionReserved(cellRange);

                    if (mergedRanges.ContainsKey(cellRange))
                    {
                        CellRange[] mergedCellRanges = mergedRanges[cellRange].ToArray();
                        FillDataSeriesHelper.RepeatMergedRanges(this.Worksheet, mergedCellRanges, cellRange, maxInitialValuesCount, seriesOrientation, isDirectionReversed);
                    }
                };

            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
            {
                this.ExecuteForEachRange(fillRangeAction);

                foreach (var indexes in valuesInfo.CellIndexLists)
                {
                    this.RepeatFormatProperties(indexes, valuesInfo.MaxInitialValuesCount);
                }

                this.ExecuteForEachRange(mergeAction);
            });
        }

        /// <summary>
        /// Fills data.
        /// </summary>
        /// <param name="direction">The direction.</param>
        public void FillData(FillDirection direction)
        {
            CellOrientation orientation;
            bool isDirectionReversed = false;
            switch (direction)
            {
                case FillDirection.Left:
                    orientation = CellOrientation.Horizontal;
                    isDirectionReversed = true;
                    break;

                case FillDirection.Up:
                    orientation = CellOrientation.Vertical;
                    isDirectionReversed = true;
                    break;

                case FillDirection.Right:
                    orientation = CellOrientation.Horizontal;
                    break;

                case FillDirection.Down:
                    orientation = CellOrientation.Vertical;
                    break;

                default:
                    throw new InvalidOperationException();
            }

            Action<CellRange> fillRangeAction =
                (cellRange) =>
                {
                    int usedMergedCellRangesIndexer = 0;
                    bool shouldGetNextMergedCellRange = false;
                    CellRange mergedCells = null;
                    CellRange[] mergedCellRanges = null;
                    CellIndex eventualMergedCellIndex = isDirectionReversed ? cellRange.ToIndex : cellRange.FromIndex;

                    if (isDirectionReversed && orientation != CellOrientation.Vertical)
                    {
                        this.TryGetMergedCellRangesInReversedSelection(cellRange, out mergedCells, out mergedCellRanges);
                    }
                    else
                    {
                        this.Worksheet.Cells.TryGetContainingMergedRange(eventualMergedCellIndex, out mergedCells);
                    }

                    if ((cellRange.ColumnCount == 1 && (direction == FillDirection.Left || direction == FillDirection.Right)) ||
                        (cellRange.RowCount == 1 && (direction == FillDirection.Down || direction == FillDirection.Up)) ||
                        (mergedCells != null && mergedCells.Equals(cellRange)))
                    {
                        WorksheetFragment value = null;

                        CellRange[] mergedRanges = this.PreserveMergedRanges(cellRange);

                        this.Worksheet.Cells[cellRange].Unmerge();

                        switch (direction)
                        {
                            case FillDirection.Left:
                                if (cellRange.FromIndex.ColumnIndex + 1 < SpreadsheetDefaultValues.ColumnCount)
                                {
                                    CellRange sourceRange = this.GetSourceRange(
                                        cellRange.FromIndex.RowIndex, cellRange.FromIndex.ColumnIndex + 1,
                                        cellRange.ToIndex.RowIndex, cellRange.ToIndex.ColumnIndex + 1);

                                    CellSelection cellSelection = this.Worksheet.Cells[sourceRange];
                                    value = cellSelection.Copy();
                                }
                                break;

                            case FillDirection.Up:
                                if (cellRange.FromIndex.RowIndex + 1 < SpreadsheetDefaultValues.RowCount)
                                {
                                    CellRange sourceRange = this.GetSourceRange(
                                        cellRange.FromIndex.RowIndex + 1, cellRange.FromIndex.ColumnIndex,
                                        cellRange.ToIndex.RowIndex + 1, cellRange.ToIndex.ColumnIndex);

                                    CellSelection cellSelection = this.Worksheet.Cells[sourceRange];
                                    value = cellSelection.Copy();
                                }
                                break;

                            case FillDirection.Right:
                                if (cellRange.FromIndex.ColumnIndex - 1 >= 0)
                                {
                                    CellRange sourceRange = this.GetSourceRange(
                                        cellRange.FromIndex.RowIndex, cellRange.FromIndex.ColumnIndex - 1,
                                        cellRange.ToIndex.RowIndex, cellRange.ToIndex.ColumnIndex - 1);

                                    CellSelection cellSelection = this.Worksheet.Cells[sourceRange];
                                    value = cellSelection.Copy();
                                }
                                break;

                            case FillDirection.Down:
                                if (cellRange.FromIndex.RowIndex - 1 >= 0)
                                {
                                    CellRange sourceRange = this.GetSourceRange(cellRange.FromIndex.RowIndex - 1, cellRange.FromIndex.ColumnIndex,
                                        cellRange.ToIndex.RowIndex - 1, cellRange.ToIndex.ColumnIndex);

                                    CellSelection cellSelection = this.Worksheet.Cells[sourceRange];
                                    value = cellSelection.Copy();
                                }
                                break;

                            default:
                                throw new InvalidOperationException();
                        }

                        if (value != null)
                        {
                            this.Worksheet.Cells[cellRange].Paste(value, new PasteOptions(PasteType.All, false, true));
                        }

                        this.RestoreMergedRanges(mergedRanges);
                    }
                    else
                    {
                        this.FillDataSeries(cellRange, orientation,
                            (indexes, initialIndexesCount) =>
                            {
                                if (mergedCellRanges != null && shouldGetNextMergedCellRange)
                                {
                                    mergedCells = mergedCellRanges.Length > usedMergedCellRangesIndexer ? mergedCellRanges[usedMergedCellRangesIndexer] : null;
                                    shouldGetNextMergedCellRange = false;
                                }

                                CellIndex mergedCellFromIndex = (mergedCells != null &&
                                                                 isDirectionReversed &&
                                                                 (indexes.Contains(mergedCells.FromIndex) || indexes.Contains(mergedCells.ToIndex)))
                                                                 ? mergedCells.FromIndex : indexes[0];

                                if (mergedCells != null)
                                {
                                    if (indexes.Contains(mergedCells.ToIndex) && !indexes.Contains(mergedCells.FromIndex))
                                    {
                                        mergedCellFromIndex = mergedCells.ToIndex;
                                        shouldGetNextMergedCellRange = true;
                                    }
                                }

                                CellRange range = new CellRange(mergedCellFromIndex, indexes[0]);

                                CellMergeState mergeState = this.Worksheet.Cells.GetMergeState(mergedCellFromIndex);
                                if (mergeState == CellMergeState.NonTopLeftCellInMergedRange)
                                {
                                    return;
                                }

                                CellRange[] resultRanges = null;
                                CellRange mergedRange;
                                if (this.Worksheet.Cells.TryGetContainingMergedRange(mergedCellFromIndex, out mergedRange))
                                {
                                    if (mergedCellRanges != null && mergedCellRanges.Contains(mergedRange))
                                    {
                                        shouldGetNextMergedCellRange = true;
                                        usedMergedCellRangesIndexer++;
                                    }

                                    range = mergedRange;

                                    CellRange[] mergedRanges = new CellRange[] { mergedRange };
                                    resultRanges = FillDataSeriesHelper.RepeatMergedRanges(mergedRanges, indexes, initialIndexesCount, isDirectionReversed);
                                }

                                CellRange destination;
                                if (resultRanges != null)
                                {
                                    int resultRangesDestinationIndex = resultRanges.Length > 1 ? 1 : 0;

                                    destination = !isDirectionReversed ?
                                            new CellRange(resultRanges[resultRangesDestinationIndex].FromIndex, resultRanges[resultRanges.Length - 1].ToIndex) :
                                            new CellRange(resultRanges[resultRanges.Length - 1].FromIndex, resultRanges[resultRangesDestinationIndex].ToIndex);
                                }
                                else
                                {
                                    destination = new CellRange(
                                        indexes[1], new CellIndex(indexes[indexes.Length - 1].RowIndex + range.RowCount - 1,
                                        indexes[indexes.Length - 1].ColumnIndex + range.ColumnCount - 1));
                                }

                                WorksheetFragment value = this.Worksheet.Cells[range].Copy();

                                this.Worksheet.Cells[destination].Unmerge();

                                this.Worksheet.Cells[destination].Paste(value, new PasteOptions(PasteType.All));

                            },
                            1, isDirectionReversed);
                    }
                };

            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
                {
                    this.ExecuteForEachRangeInsideBeginEndUpdate(fillRangeAction);
                });
        }

        private bool TryGetMergedCellRangesInReversedSelection(CellRange cellRange, out CellRange mergedCellRange, out CellRange[] mergedCellRanges)
        {
            mergedCellRange = null;
            mergedCellRanges = this.Worksheet.Cells.MergedCellRanges.GetIntersectingMergedRanges(cellRange).ToArray();

            if (mergedCellRanges != null)
            {
                if (mergedCellRanges.Length > 0)
                {
                    mergedCellRange = mergedCellRanges[0];
                }
            }

            return mergedCellRange != null;
        }

        private CellRange[] PreserveMergedRanges(CellRange cellRange)
        {
            List<CellRange> result = new List<CellRange>();

            foreach (CellRange range in this.Worksheet.Cells.MergedCellRanges.Ranges)
            {
                if (range.IntersectsWith(cellRange))
                {
                    result.Add(range);
                }
            }

            return result.ToArray();
        }

        private void RestoreMergedRanges(CellRange[] mergedRanges)
        {
            for (int i = 0; i < mergedRanges.Length; i++)
            {
                CellRange mergedRange = mergedRanges[i];
                this.Worksheet.Cells[mergedRange].Merge();
            }
        }

        private CellRange GetSourceRange(int fromRowIndex, int fromColumnIndex, int toRowIndex, int toColumnIndex)
        {
            CellIndex fromIndex = new CellIndex(fromRowIndex, fromColumnIndex);
            CellIndex toIndex = new CellIndex(toRowIndex, toColumnIndex);

            ISet<CellRange> mergedCellRanges = this.Worksheet.Cells.MergedCellRanges.GetIntersectingMergedRanges(new CellRange(fromIndex, toIndex));
            if (mergedCellRanges.Count > 0)
            {
                CellRange intersectingMergedCellRange = mergedCellRanges.First();
                toIndex = toIndex.Offset(intersectingMergedCellRange.FromIndex.RowIndex - fromIndex.RowIndex,
                    intersectingMergedCellRange.FromIndex.ColumnIndex - fromIndex.ColumnIndex);
                fromIndex = intersectingMergedCellRange.FromIndex;
            }

            return new CellRange(fromIndex, toIndex);
        }

        #endregion


        #region Data Validation

        /// <summary>
        /// Gets the data validation rule.
        /// </summary>
        /// <returns></returns>
        public RangePropertyValue<IDataValidationRule> GetDataValidationRule()
        {
            return this.GetPropertyValue(this.Cells.DataValidationRuleProperty);
        }

        /// <summary>
        /// Sets the data validation rule.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetDataValidationRule(IDataValidationRule value)
        {
            this.SetPropertyValue(this.Cells.DataValidationRuleProperty, value);
        }

        /// <summary>
        /// Clears the data validation rule.
        /// </summary>
        public void ClearDataValidationRule()
        {
            this.ClearPropertyValue(this.Cells.DataValidationRuleProperty);
        }

        #endregion


        #region Sorting

        /// <summary>
        /// Sorts the selected cells using the specified sort conditions.
        /// </summary>
        /// <param name="sortConditions">The sort conditions.</param>
        public void Sort(params ISortCondition[] sortConditions)
        {
            CellRange sortRange = this.CellRanges.First();

            if (!SortAndFilterHelper.IsValidSelection(this.CellRanges))
            {
                throw new SortingException("Wrong range selected.", "Spreadsheet_Sorting_WrongSelection");
            }

            if (SortManager.AreSortConditionsDuplicating(sortConditions))
            {
                throw new SortingException("Duplicated sort conditions.", "Spreadsheet_Sorting_DuplicatedSortConditions");
            }

            int mergedCellsCount = this.Cells.GetContainingMergedRanges(sortRange).Count();

            if (mergedCellsCount > 0)
            {
                throw new SortingException("Cannot sort a range containing merged cells.", "Spreadsheet_Sorting_MergedCells");
            }

            CellSelection cells = this.Cells[sortRange];

            if (SortAndFilterHelper.IsEmptySortRange(cells))
            {
                throw new SortingException("Cannot sort empty range!", "Spreadsheet_Sorting_NoData");
            }

            foreach (ISortCondition sortCondition in sortConditions)
            {
                Guard.ThrowExceptionIfOutOfRange(0, sortRange.ColumnCount - 1, sortCondition.RelativeIndex, "RelativeIndex");
            }

            using (new UpdateScope(this.Worksheet.BeginUndoGroup, this.Worksheet.EndUndoGroup))
            {
                this.Worksheet.SortState.ClearInternal();
                this.Worksheet.SortState.SortRange = sortRange;

                sortConditions = sortConditions.Reverse().ToArray();

                SortManager.Sort(this.Cells, sortRange, sortConditions);
            }
        }

        #endregion


        #region Filtering

        /// <summary>
        /// Sets the specified filters to the cell selection.
        /// </summary>
        /// <param name="filters">The filters.</param>
        public void Filter(params IFilter[] filters)
        {
            if (this.CellRanges.Count() > 1)
            {
                throw new LocalizableException("This action cannot be performed on multiple ranges");
            }

            CellRange filterRange = this.CellRanges.First();

            using (new UpdateScope(this.Worksheet.BeginUndoGroup, this.Worksheet.EndUndoGroup))
            {
                this.Worksheet.Filter.FilterRange = filterRange;
                this.Worksheet.Filter.SetFilters(filters);
            }
        }

        #endregion
    }
}
