﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Provides base for the cell value types classes.
    /// </summary>
    /// <typeparam name="T">The type of the underlying value.</typeparam>
    public abstract class CellValueBase<T> : ICellValue
    {
        #region Fields

        private readonly T value;
        private string valueAsString;
        private bool isEditFormatStringValid;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the type of the value.
        /// </summary>
        /// <value>The type of the value.</value>
        public abstract CellValueType ValueType { get; }

        /// <summary>
        /// Gets the type of the result value.
        /// </summary>
        /// <value>The type of the result value.</value>
        public virtual CellValueType ResultValueType
        {
            get
            {
                return this.ValueType;
            }
        }

        /// <summary>
        /// Gets the raw value.
        /// </summary>
        /// <value>The raw value.</value>
        public virtual string RawValue
        {
            get
            {
                return this.value.ToString();
            }
        }

        /// <summary>
        /// Gets the value.
        /// </summary>
        /// <value>The value.</value>
        public T Value
        {
            get
            {
                return this.value;
            }
        }

        // TODO:: Remove this property. We need some kind of invalidation? - consider it!
        //protected virtual bool AlwaysResetValueAsString
        //{
        //    get
        //    {
        //        return false;
        //    }
        //}

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="CellValueBase{T}" /> class.
        /// </summary>
        /// <param name="value">The value.</param>
        protected CellValueBase(T value)
        {
            Guard.ThrowExceptionIfNull(value, "value");
            this.value = value;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Gets the value as string.
        /// </summary>
        /// <param name="format">The format.</param>
        /// <returns></returns>
        public string GetValueAsString(CellValueFormat format)
        {
            if (this.valueAsString == null || !this.isEditFormatStringValid)
            {
                this.valueAsString = this.GetValueAsStringOverride(format);
            }

            return this.valueAsString;
        }

        /// <summary>
        /// Invalidates the edit format string.
        /// </summary>
        public void InvalidateEditFormatString()
        {
            this.isEditFormatStringValid = false;
        }

        /// <summary>
        /// Gets the value as string override.
        /// </summary>
        /// <param name="format">The format.</param>
        /// <returns></returns>
        protected virtual string GetValueAsStringOverride(CellValueFormat format = null)
        {
            return this.value.ToString();
        }

        /// <summary>
        /// Gets the result value as string.
        /// </summary>
        /// <param name="format">The format.</param>
        /// <returns></returns>
        public virtual string GetResultValueAsString(CellValueFormat format)
        {
            return format.GetFormatResult(this).VisibleInfosText;
        }

        #endregion


        #region Overrides

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            CellValueBase<T> other = obj as CellValueBase<T>;
            if (other == null)
            {
                return false;
            }

            return this.ValueType == other.ValueType &&
                this.GetValueAsStringOverride() == other.GetValueAsStringOverride();
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(this.Value.GetHashCode(), this.GetValueAsStringOverride().GetHashCode());
        }

        #endregion
    }
}