﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Produces instances of ICellValue type.
    /// </summary>
    public static class CellValueFactory
    {
        #region Extension Methods

        /// <summary>
        /// Produces an ICellValue instance.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="cellIndex">Index of the cell.</param>
        /// <remarks>Please note that the string value is parsed to find its type. This process is slow.
        /// If you know the type of your data use the other overloads of this method.</remarks>
        /// <returns>The ICellValue instance.</returns>
        public static ICellValue ToCellValue(this string value, Worksheet worksheet, CellIndex cellIndex)
        {
            return ToCellValue(value, worksheet, cellIndex.RowIndex, cellIndex.ColumnIndex);
        }

        /// <summary>
        /// Produces an ICellValue instance.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <remarks>Please note that the string value is parsed to find its type. This process is slow.
        /// If you know the type of your data use the other overloads of this method.</remarks>
        /// <returns>The ICellValue instance.</returns>
        public static ICellValue ToCellValue(this string value, Worksheet worksheet, int rowIndex, int columnIndex)
        {
            CellValueFormat currentFormat = worksheet.Cells[rowIndex, columnIndex].GetFormat().Value;

            ICellValue cellValue = ToCellValue(value, worksheet, rowIndex, columnIndex, currentFormat);

            return cellValue;
        }

        internal static ICellValue ToCellValue(this string value, Worksheet worksheet, int rowIndex, int columnIndex, CellValueFormat currentFormat)
        {
            CellValueFormat format;
            ICellValue cellValue;
            CellValueFactory.Create(value, worksheet, rowIndex, columnIndex, currentFormat, out cellValue, out format);

            return cellValue;
        }

        /// <summary>
        /// Produces an ICellValue instance.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <param name="worksheet">The worksheet.</param>
        /// <returns>The ICellValue instance.</returns>
        public static ICellValue ToCellValue(this CellIndex value, Worksheet worksheet)
        {
            return worksheet.Cells[value].GetValue().Value;
        }

        /// <summary>
        /// Produces an ICellValue instance.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <returns>The ICellValue instance.</returns>
        public static ICellValue ToCellValue(this double value)
        {
            return CellValueFactory.Create(value);
        }

        /// <summary>
        /// Produces an ICellValue instance.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <returns>The ICellValue instance.</returns>
        public static ICellValue ToCellValue(this bool value)
        {
            return CellValueFactory.Create(value);
        }

        /// <summary>
        /// Produces an ICellValue instance.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <remarks>Please note the date is stored as number value. If a date number format is not set to the cell the value will be shown as number.</remarks>
        /// <returns>The ICellValue instance.</returns>
        public static ICellValue ToCellValue(this DateTime value)
        {
            return CellValueFactory.Create(value);
        }

        /// <summary>
        /// Produces an TextCellValue instance.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <returns>The TextCellValue instance.</returns>
        public static ICellValue ToTextCellValue(this string value)
        {
            return CellValueFactory.CreateTextCellValue(value);
        }

        /// <summary>
        /// Produces an FormulaCellValue instance.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The FormulaCellValue instance.</returns>
        public static ICellValue ToFormulaCellValue(this string value, Worksheet worksheet, int rowIndex, int columnIndex)
        {
            return CellValueFactory.CreateFormulaCellValue(value, worksheet, rowIndex, columnIndex);
        }

        /// <summary>
        /// Gets the type of the cell value resulting from an expression.
        /// </summary>
        /// <param name="value">The expression.</param>
        /// <returns>The cell value type.</returns>
        public static CellValueType GetCellValueType(this ConstantExpression value)
        {
            if (value is NumberExpression)
            {
                return CellValueType.Number;
            }

            if (value is StringExpression)
            {
                return CellValueType.Text;
            }

            if (value is BooleanExpression)
            {
                return CellValueType.Boolean;
            }

            if (value is ErrorExpression)
            {
                return CellValueType.Error;
            }

            if (value is EmptyExpression)
            {
                return CellValueType.Empty;
            }

            throw new NotImplementedException();
        }

        #endregion


        #region Methods

        internal static ICellValue Create(string value, Worksheet worksheet, int rowIndex, int columnIndex, SpreadsheetCultureHelper cultureInfo)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfNull(cultureInfo, "cultureInfo");
            Guard.ThrowExceptionIfNull(value, "value");
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            RadExpression expression;
            InputStringCollection inputStringCollection;
            ParseResult result = RadExpression.TryParse(value, worksheet, rowIndex, columnIndex, cultureInfo, out expression, out inputStringCollection);
            if (result != ParseResult.Successful)
            {
                return new TextCellValue(value);
            }

            return new FormulaCellValue(inputStringCollection, expression, rowIndex, columnIndex);
        }

        /// <summary>
        /// Creates the specified value.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="cellIndex">Index of the cell.</param>
        /// <param name="currentFormatValue">The current format value.</param>
        /// <param name="cellValue">The cell value.</param>
        /// <param name="newFormatValue">The new format value.</param>
        public static void Create(string value, Worksheet worksheet, CellIndex cellIndex, CellValueFormat currentFormatValue,
            out ICellValue cellValue, out CellValueFormat newFormatValue)
        {
            Create(value, worksheet, cellIndex.RowIndex, cellIndex.ColumnIndex, currentFormatValue, out cellValue, out newFormatValue);
        }

        /// <summary>
        /// Creates an ICellValue instance.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <param name="currentFormatValue">The current format value.</param>
        /// <param name="cellValue">The cell value.</param>
        /// <param name="newFormatValue">The format value.</param>
        public static void Create(string value, Worksheet worksheet, int rowIndex, int columnIndex, CellValueFormat currentFormatValue,
            out ICellValue cellValue, out CellValueFormat newFormatValue)
        {
            ParseResult parseResult = CellValueParser.Parse(value, worksheet, rowIndex, columnIndex, currentFormatValue, out cellValue, out newFormatValue);
            if (parseResult == ParseResult.Error)
            {
                ExpressionException ex = new ExpressionException(SpreadsheetStrings.GeneralErrorMessage, "Spreadsheet_ErrorExpressions_GeneralErrorMessage");

                throw new ParseException(ex.Message, ex, ex.LocalizationKey, ex.FormatStringArguments);
            }
        }

        internal static void CreateIgnoreErrors(string value, Worksheet worksheet, int rowIndex, int columnIndex, CellValueFormat currentFormatValue,
            out ICellValue cellValue, out CellValueFormat newFormatValue)
        {
            ParseResult parseResult = CellValueParser.Parse(value, worksheet, rowIndex, columnIndex, currentFormatValue, out cellValue, out newFormatValue);
            if (parseResult != ParseResult.Successful)
            {
                newFormatValue = currentFormatValue;
                cellValue = new TextCellValue(value);
            }
        }

        internal static ParseResult TryCreate(string value, Worksheet worksheet, int rowIndex, int columnIndex, CellValueFormat currentFormatValue,
            out ICellValue cellValue, out CellValueFormat newFormatValue)
        {
            return CellValueParser.Parse(value, worksheet, rowIndex, columnIndex, currentFormatValue, out cellValue, out newFormatValue);
        }

        /// <summary>
        /// Creates an ICellValue instance from a double value.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <returns>The ICellValue instance.</returns>
        public static ICellValue Create(double value)
        {
            return new NumberCellValue(value);
        }

        /// <summary>
        /// Creates an ICellValue instance from a bool value.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <returns>The ICellValue instance.</returns>
        public static ICellValue Create(bool value)
        {
            return new BooleanCellValue(value);
        }

        /// <summary>
        /// Creates an ICellValue instance from a DateTime value.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <returns>The ICellValue instance.</returns>
        public static ICellValue Create(DateTime value)
        {
            double date = FormatHelper.ConvertDateTimeToDouble(value);
            return CellValueFactory.Create(date);
        }

        /// <summary>
        /// Creates a text cell value.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <returns>The ICellValue instance.</returns>
        public static ICellValue CreateTextCellValue(string value)
        {
            return new TextCellValue(value);
        }

        /// <summary>
        /// Creates a formula cell value.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The ICellValue instance.</returns>
        public static ICellValue CreateFormulaCellValue(string value, Worksheet worksheet, int rowIndex, int columnIndex)
        {
            RadExpression expression;
            InputStringCollection inputStringCollection;
            ParseResult result = RadExpression.TryParse(value, worksheet, rowIndex, columnIndex, out expression, out inputStringCollection);
            if (result != ParseResult.Successful)
            {
                throw new InvalidOperationException();
            }

            return new FormulaCellValue(inputStringCollection, expression, rowIndex, columnIndex);
        }

        // TODO: Rich text.

        #endregion
    }
}