﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Windows.Media;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Describes the format that can be applied to a cell value.
    /// </summary>
    public class CellValueFormat
    {
        #region Constants

        /// <summary>
        /// The general format.
        /// </summary>
        public static readonly CellValueFormat GeneralFormat = new CellValueFormat(FormatHelper.GeneralFormatString);

        #endregion


        #region Fields

        private readonly TextFormatStringParser formatStringParser;
        private readonly string formatString;
        private readonly string invariantFormatString;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the format string.
        /// </summary>
        /// <value>The format string.</value>
        public string FormatString
        {
            get
            {
                return this.formatString;
            }
        }

        /// <summary>
        /// Gets the information about the format string.
        /// </summary>
        /// <value>The format string information.</value>
        public FormatStringInfo FormatStringInfo
        {
            get
            {
                return this.formatStringParser.FormatStringInfo;
            }
        }

        internal string InvariantFormatString
        {
            get
            {
                return this.invariantFormatString;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="CellValueFormat" /> class.
        /// </summary>
        /// <param name="formatString">The format string.</param>
        public CellValueFormat(string formatString) 
            : this(formatString, true)
        {
        }

        internal CellValueFormat(string formatString, bool convertToInvariant)
        {
            Guard.ThrowExceptionIfNull(formatString, "formatString");

            FormatStringType? formatStringType;
            FormatStringValidator.ValidateAndGetFormatStringType(formatString, out formatStringType);
            bool isFormatStringANumberType = formatStringType.HasValue && formatStringType.Value == FormatStringType.Number;

            string invariantFormatString = formatString;
            if (isFormatStringANumberType)
            {
                if (convertToInvariant)
                {
                    invariantFormatString = FormatHelper.ReplaceCultureSpecificSeparatorsWithDefaults(formatString);
                }
                else
                {
                    formatString = FormatHelper.ReplaceDefaultSeparatorsWithCultureSpecific(invariantFormatString);
                }
            }

            this.formatStringParser = TextFormatStringParser.Create(invariantFormatString, formatString);
            this.invariantFormatString = invariantFormatString;
            this.formatString = formatString;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Gets the format result.
        /// </summary>
        /// <param name="cellValue">The cell value.</param>
        /// <returns>The format result.</returns>
        public CellValueFormatResult GetFormatResult(ICellValue cellValue)
        {
            List<CellValueFormatResultItem> infos = this.formatStringParser.ApplyFormatToValues(cellValue, this);

            Color? foreground = this.formatStringParser.FormatDescriptor.GetForegroundOrDefault();
            CultureInfo culture = this.formatStringParser.FormatDescriptor.GetCultureOrDefault();
            Predicate<double> condition = this.formatStringParser.FormatDescriptor.GetConditionOrDefault();

            return new CellValueFormatResult(infos, foreground, condition, culture);
        }

        #endregion


        #region Overrides

        /// <summary>
        /// Implements the equals operator.
        /// </summary>
        /// <param name="first">The first value.</param>
        /// <param name="second">The second value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
        public static bool operator ==(CellValueFormat first, CellValueFormat second)
        {
            if (object.ReferenceEquals(first, second))
            {
                return true;
            }

            if ((object)first == null)
            {
                return false;
            }

            return first.Equals(second);
        }

        /// <summary>
        /// Implements the less operator.
        /// </summary>
        /// <param name="first">The first value.</param>
        /// <param name="second">The second value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
        public static bool operator !=(CellValueFormat first, CellValueFormat second)
        {
            return !(first == second);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            CellValueFormat other = obj as CellValueFormat;
            if (other == null)
            {
                return false;
            }

            return TelerikHelper.EqualsOfT(this.FormatString, other.FormatString);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return this.FormatString.GetHashCode();
        }

        #endregion
    }
}