﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents error cell value class.
    /// </summary>
    public class ErrorCellValue : CellValueBase<string>
    {
        #region Constants

        private static readonly Dictionary<string, ErrorCellValue> errorCellValues;

        /// <summary>
        /// Represents null error cell value.
        /// </summary>
        public static readonly ErrorCellValue NullErrorCellValue;

        /// <summary>
        /// Represents division by zero error cell value.
        /// </summary>
        public static readonly ErrorCellValue DivisionByZeroErrorCellValue;

        /// <summary>
        /// Represents value error cell value.
        /// </summary>
        public static readonly ErrorCellValue ValueErrorCellValue;

        /// <summary>
        /// Represents reference error cell value.
        /// </summary>
        public static readonly ErrorCellValue ReferenceErrorCellValue;

        /// <summary>
        /// Represents name error cell value.
        /// </summary>
        public static readonly ErrorCellValue NameErrorCellValue;

        /// <summary>
        /// Represents number error cell value.
        /// </summary>
        public static readonly ErrorCellValue NumberErrorCellValue;

        /// <summary>
        /// Represents not available error cell value.
        /// </summary>
        public static readonly ErrorCellValue NotAvailableErrorCellValue;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the type of the value.
        /// </summary>
        /// <value>The type of the value.</value>
        public override CellValueType ValueType
        {
            get
            {
                return CellValueType.Error;
            }
        }

        #endregion


        #region Constructors

        static ErrorCellValue()
        {
            errorCellValues = new Dictionary<string, ErrorCellValue>();

            NullErrorCellValue = CreateErrorCellValue(SpreadsheetCultureHelper.NullErrorValue);
            DivisionByZeroErrorCellValue = CreateErrorCellValue(SpreadsheetCultureHelper.DivisionByZeroErrorValue);
            ValueErrorCellValue = CreateErrorCellValue(SpreadsheetCultureHelper.ValueErrorValue);
            ReferenceErrorCellValue = CreateErrorCellValue(SpreadsheetCultureHelper.ReferenceErrorValue);
            NameErrorCellValue = CreateErrorCellValue(SpreadsheetCultureHelper.NameErrorValue);
            NumberErrorCellValue = CreateErrorCellValue(SpreadsheetCultureHelper.NumberErrorValue);
            NotAvailableErrorCellValue = CreateErrorCellValue(SpreadsheetCultureHelper.NotAvailableErrorValue);
        }

        private ErrorCellValue(string value)
            : base(value)
        {
        }

        #endregion


        #region Methods

        internal static ErrorCellValue GetErrorCellValue(string key)
        {
            Guard.ThrowExceptionIfNullOrEmpty(key, "key");

            return errorCellValues[key];
        }

        internal static bool TryGetErrorCellValue(string key, out ErrorCellValue value)
        {
            Guard.ThrowExceptionIfNullOrEmpty(key, "key");

            string sanitizedKey = key.ToUpper(FormatHelper.DefaultSpreadsheetCulture.CultureInfo);

            return errorCellValues.TryGetValue(sanitizedKey, out value);
        }

        private static ErrorCellValue CreateErrorCellValue(string value)
        {
            Guard.ThrowExceptionIfNullOrEmpty(value, "value");

            ErrorCellValue errorCellValue = new ErrorCellValue(value);
            errorCellValues[value] = errorCellValue;

            return errorCellValue;
        }

        #endregion
    }
}
