﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Copying;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    ///Represents the value of a cell which contains a formula.
    /// </summary>
    public class FormulaCellValue : CellValueBase<RadExpression>, ICopyable<FormulaCellValue>
    {
        #region Fields

        private string valueAsString;
        private readonly InputStringCollection inputStringCollection;
        private readonly int rowIndex;
        private readonly int columnIndex;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the raw value.
        /// </summary>
        /// <value>The raw value.</value>
        public override string RawValue
        {
            get
            {
                return this.valueAsString;
            }
        }

        /// <summary>
        /// Gets the type of the value.
        /// </summary>
        /// <value>The type of the value.</value>
        public override CellValueType ValueType
        {
            get
            {
                return CellValueType.Formula;
            }
        }

        /// <summary>
        /// Gets the type of the result value.
        /// </summary>
        /// <value>The type of the result value.</value>
        public override CellValueType ResultValueType
        {
            get
            {
                return InterpretEmptyExpression(this.Value.GetValueAsConstantExpression()).GetCellValueType();
            }
        }

        internal int RowIndex
        {
            get
            {
                return this.rowIndex;
            }
        }

        internal int ColumnIndex
        {
            get
            {
                return this.columnIndex;
            }
        }

        internal bool IsSheetNameDependent
        {
            get
            {
                return this.inputStringCollection.IsSheetNameDependent;
            }
        }

        internal bool IsWorkbookNameDependent
        {
            get
            {
                return this.inputStringCollection.IsWorkbookNameDependent;
            }
        }

        internal bool IsCellsNameDependent
        {
            get
            {
                return this.inputStringCollection.IsCellsNameDependent;
            }
        }

        internal bool ContainsSpreadsheetName
        {
            get
            {
                return this.inputStringCollection.ContainsSpreadsheetName;
            }
        }

        #endregion


        #region Constructors

        internal FormulaCellValue(InputStringCollection translatableString, RadExpression value, CellIndex cellIndex)
            : this(translatableString, value, cellIndex.RowIndex, cellIndex.ColumnIndex)
        {
        }

        internal FormulaCellValue(InputStringCollection translatableString, RadExpression value, int rowIndex, int columnIndex)
            : base(value)
        {
            Guard.ThrowExceptionIfNull(translatableString, "translatableString");
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            this.inputStringCollection = translatableString;
            this.rowIndex = rowIndex;
            this.columnIndex = columnIndex;
            this.valueAsString = SpreadsheetCultureHelper.PrepareFormulaValue(this.inputStringCollection.BuildString());
        }

        #endregion


        #region Methods

        /// <summary>
        /// Gets the result value as string.
        /// </summary>
        /// <param name="format">The format.</param>
        /// <returns>The result value string.</returns>
        public override string GetResultValueAsString(CellValueFormat format)
        {
            return this.GetResultValueAsCellValue().GetResultValueAsString(format);
        }

        /// <summary>
        /// Gets the value as string override.
        /// </summary>
        /// <param name="format">The format.</param>
        /// <returns>The value as string.</returns>
        protected override string GetValueAsStringOverride(CellValueFormat format = null)
        {
            return this.RawValue;
        }

        /// <summary>
        /// Gets the result value as cell value.
        /// </summary>
        /// <returns>The cell value.</returns>
        public ICellValue GetResultValueAsCellValue()
        {
            ConstantExpression resultValue = InterpretEmptyExpression(this.Value.GetValueAsConstantExpression());

            if (resultValue is NumberExpression)
            {
                return new NumberCellValue(resultValue.NumberValue());
            }

            ErrorCellValue value;
            if (resultValue is ErrorExpression && ErrorCellValue.TryGetErrorCellValue(resultValue.GetValueAsString(), out value))
            {
                return value;
            }

            return new TextCellValue(resultValue.GetValueAsString());
        }

        internal string GetExportString()
        {
            return this.inputStringCollection.BuildStringInCulture(SpreadsheetCultureHelper.InvariantSpreadsheetCultureInfo);
        }

        internal FormulaCellValue CloneAndTranslate(Worksheet targetWorksheet, CellIndex cellIndex, bool onOverflowStarOver = false)
        {
            return CloneAndTranslate(targetWorksheet, cellIndex.RowIndex, cellIndex.ColumnIndex, onOverflowStarOver);
        }

        /// <summary>
        /// Creates a new instance of FormulaCellValue and, if necessary, translates it using the targetCellIndex.
        /// The method is used for when copying, pasting and setting new value.
        /// </summary>
        /// <param name="targetWorksheet">The target worksheet.</param>
        /// <param name="targetRowIndex">Index of the target row.</param>
        /// <param name="targetColumnIndex">Index of the target column.</param>
        /// <param name="onOverflowStartOver">The on overflow start over.</param>
        /// <returns>New instance of FormulaCellValue</returns>
        internal FormulaCellValue CloneAndTranslate(Worksheet targetWorksheet, int targetRowIndex, int targetColumnIndex, bool onOverflowStartOver = false)
        {
            // TODO: the copied cell value could be with the same CellIndex but on another sheet
            //if (this.cellIndex.ColumnIndex == targetCellIndex.ColumnIndex && this.cellIndex.RowIndex == targetCellIndex.RowIndex)
            //{
            //    return this;
            //}

            InputStringCollection newCollection = this.inputStringCollection.Clone();

            ExpressionCloneAndTranslateContext context = new ExpressionCloneAndTranslateContext(targetWorksheet, targetRowIndex, targetColumnIndex, this.inputStringCollection, newCollection, onOverflowStartOver);
            RadExpression expression = this.Value.CloneAndTranslate(context);

            return new FormulaCellValue(newCollection, expression, targetRowIndex, targetColumnIndex);
        }

        internal FormulaCellValue CloneAndTranslate(Worksheet worksheet, int rowIndex, int columnIndex, CellRangeInsertedOrRemovedEventArgs eventArgs, bool onOverflowStarOver = false)
        {
            InputStringCollection newCollection = this.inputStringCollection.Clone();

            bool isHorizontal = eventArgs.RangeType == RangeType.Columns || eventArgs.RangeType == RangeType.CellsInRow;

            ExpressionCloneAndTranslateContext context = new ExpressionCloneAndTranslateContext(worksheet, rowIndex, columnIndex, this.inputStringCollection,
                newCollection, isHorizontal, eventArgs.Range, eventArgs.IsRemove, onOverflowStarOver);

            RadExpression expression = this.Value.CloneAndTranslate(context);

            // TODO: there is no mechanism to determine whether a FCV should be changed upon cells insertion/removal
            // curretnly, if the value should not be changed, the same instance is returned.
            if (expression.Equals(this.Value))
            {
                return this;
            }

            return new FormulaCellValue(newCollection, expression, this.rowIndex, this.columnIndex);
        }

        /// <summary>
        /// Updates the worksheet name of its CellReferenceRange and SpreadsheetName expressions.
        /// </summary>
        /// <param name="worksheet"></param>
        internal void Translate(Worksheet worksheet)
        {
            ExpressionTranslateContext context = new ExpressionTranslateContext(worksheet);
            this.Value.Translate(context);
            this.valueAsString = SpreadsheetCultureHelper.PrepareFormulaValue(this.inputStringCollection.BuildString());
        }

        /// <summary>
        /// Updates the workbook name of its SpreadsheetName expressions.
        /// </summary>
        /// <param name="workbook"></param>
        internal void Translate(Workbook workbook)
        {
            ExpressionTranslateContext context = new ExpressionTranslateContext(workbook);
            this.Value.Translate(context);
            this.valueAsString = SpreadsheetCultureHelper.PrepareFormulaValue(this.inputStringCollection.BuildString());
        }

        /// <summary>
        /// Updates the SpreadsheetName name of its expression tree.
        /// </summary>
        /// <param name="oldName"></param>
        /// <param name="newName"></param>
        internal void Translate(string oldName, string newName)
        {
            ExpressionTranslateContext context = new ExpressionTranslateContext(oldName, newName);
            this.Value.Translate(context);
            this.valueAsString = SpreadsheetCultureHelper.PrepareFormulaValue(this.inputStringCollection.BuildString());
        }

        internal static ConstantExpression InterpretEmptyExpression(ConstantExpression expressionValue)
        {
            if (expressionValue is EmptyExpression)
            {
                return NumberExpression.Zero;
            }

            return expressionValue;
        }

        internal bool IsValueCyclicReferenceError()
        {
            return this.Value.GetValueAsConstantExpression().Equals(ErrorExpressions.CyclicReference);
        }

        FormulaCellValue ICopyable<FormulaCellValue>.Copy(CopyContext context)
        {
            string newValue = this.valueAsString.Replace(context.SourceWorksheet.Name, context.TargetWorksheet.Name);
            FormulaCellValue formulaCellValueCopied = (FormulaCellValue)CellValueFactory.CreateFormulaCellValue(newValue, context.TargetWorksheet, this.RowIndex, this.ColumnIndex);

            return formulaCellValueCopied;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            return object.ReferenceEquals(this, obj);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return this.Value.GetHashCode();
        }

        #endregion
    }
}