﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents the value of a cell which contains a number.
    /// </summary>
    public class NumberCellValue : CellValueBase<double>, IExpressionAwareCellValue
    {
        #region Fields

        private RadExpression expression;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the type of the value.
        /// </summary>
        /// <value>The type of the value.</value>
        public override CellValueType ValueType
        {
            get
            {
                return CellValueType.Number;
            }
        }

        /// <summary>
        /// Gets the raw value.
        /// </summary>
        /// <value>The raw value.</value>
        public override string RawValue
        {
            get
            {
                return this.Value.ToString(FormatHelper.DefaultSpreadsheetCulture.CultureInfo);
            }
        }

        #endregion


        #region Constructors

        internal NumberCellValue(double doubleValue)
            : base(doubleValue)
        {
        }

        #endregion


        #region Methods

        internal string ToString(FormatStringType formatType)
        {
            if (formatType == FormatStringType.Number)
            {
                return this.Value.ToString();
            }
            else if (formatType == FormatStringType.DateTime)
            {
                DateTime? dateTimeValue = this.ToDateTime();
                if (!dateTimeValue.HasValue)
                {
                    return this.Value.ToString();
                }

                return this.ToDateTime().ToString();
            }

            throw new ArgumentException("Invalid format string type.");
        }

        private string GetEditModeFormatString(string value, CellValueFormat format)
        {
            double doubleRepresentation = double.Parse(value, FormatHelper.DefaultSpreadsheetCulture.CultureInfo);
            bool hasTime = (doubleRepresentation - Math.Floor(doubleRepresentation)) != 0;
            bool hasDate = Math.Floor(doubleRepresentation) != 0;

            if (format.FormatStringInfo.FormatType == FormatStringType.DateTime && !hasDate)
            {
                return FormatHelper.LongTimePattern;
            }

            if (format.FormatStringInfo.FormatType == FormatStringType.DateTime && hasTime)
            {
                return FormatHelper.ShortDateLongTimePattern;
            }

            if (format.FormatStringInfo.FormatType == FormatStringType.DateTime)
            {
                return FormatHelper.ShortDatePattern;
            }

            if (format.FormatStringInfo.FormatType == FormatStringType.Number && format.FormatString.Contains("%"))
            {
                double doubleValue;
                bool canParse = double.TryParse(value, System.Globalization.NumberStyles.Any, FormatHelper.DefaultSpreadsheetCulture.CultureInfo, out doubleValue);
                if (canParse)
                {
                    int index;
                    string value1 = (doubleValue * 100).ToString(FormatHelper.DefaultSpreadsheetCulture.CultureInfo);
                    string editModeFormatString = string.Empty;
                    if ((index = value1.IndexOf(FormatHelper.DefaultSpreadsheetCulture.CultureInfo.NumberFormat.PercentDecimalSeparator)) != -1)
                    {
                        int zerosCount = 0;
                        for (int i = index + 1; i < value1.Length; i++)
                        {
                            if (char.IsDigit(value1[i]))
                            {
                                zerosCount++;
                            }
                            else
                            {
                                break;
                            }
                        }

                        editModeFormatString = string.Format("0.{0}%", string.Concat(Enumerable.Repeat('0', zerosCount)));
                    }
                    else
                    {
                        editModeFormatString = "0%";
                    }

                    return editModeFormatString;
                }
            }

            if (format.FormatStringInfo.FormatType == FormatStringType.Number)
            {
                double absoluteDoubleRepresentation = Math.Abs(doubleRepresentation);

                if (absoluteDoubleRepresentation == 0 ||
                   (absoluteDoubleRepresentation >= FormatHelper.MinGeneralNumber &&
                    absoluteDoubleRepresentation <= FormatHelper.MaxGeneralNumber))
                {
                    return FormatHelper.GeneralNumberEditFormatString;
                }
                else
                {
                    return FormatHelper.GeneralNumberScientificEditFormatString;
                }
            }

            return string.Empty;
        }

        #endregion


        #region Method Overrides

        /// <summary>
        /// Gets the value as string override.
        /// </summary>
        /// <param name="format">The format.</param>
        /// <returns></returns>
        protected override string GetValueAsStringOverride(CellValueFormat format)
        {
            if (format.FormatStringInfo.FormatType == FormatStringType.DateTime)
            {
                string valueAsString = this.GetValueAsString(CellValueFormat.GeneralFormat);
                string editModeFormatString = this.GetEditModeFormatString(valueAsString, format);

                if (this.Value < 0 || this.Value > FormatHelper.MaxDoubleValueTranslatableToDateTime)
                {
                    return this.ToString(format.FormatStringInfo.FormatType.Value);
                }
                else
                {
                    return this.ToDateTime().Value.RoundMinutes().ToString(editModeFormatString, FormatHelper.DefaultSpreadsheetCulture.CultureInfo);
                }
            }
            else if (format.FormatStringInfo.FormatType == FormatStringType.Number)
            {
                string editModeFormatString = this.GetEditModeFormatString(this.RawValue, format);
                return this.Value.ToString(editModeFormatString, FormatHelper.DefaultSpreadsheetCulture.CultureInfo);
            }

            return this.Value.ToString();
            //throw new ArgumentException("Invalid format type.");
        }

        RadExpression IExpressionAwareCellValue.GetCellExpressionOrNull()
        {
            if (this.expression == null)
            {
                this.expression = NumberExpression.CreateValidNumberOrErrorExpression(this.Value);
            }

            return this.expression;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            NumberCellValue numberCellValue = obj as NumberCellValue;
            if (numberCellValue == null)
            {
                return false;
            }

            return this.Value.Equals(numberCellValue.Value);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return this.Value.GetHashCode();
        }

        #endregion
    }
}