﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.Copying;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Model.DataValidation;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
#if NETCORE || NETSTANDARD
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents the cells of the worksheet.
    /// </summary>
    public class Cells : WorksheetEntityBase
    {
        #region Fields

        private readonly CellsPropertyBag propertyBag;
        private readonly MergedCellRanges mergedCellRanges;

        private readonly IProperty<ICellValue> valueProperty;
        private readonly IProperty<IDataValidationRule> dataValidationRuleProperty;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the row count.
        /// </summary>
        /// <value>The row count.</value>
        public int RowCount
        {
            get
            {
                return SpreadsheetDefaultValues.RowCount;
            }
        }

        /// <summary>
        /// Gets the column count.
        /// </summary>
        /// <value>The column count.</value>
        public int ColumnCount
        {
            get
            {
                return SpreadsheetDefaultValues.ColumnCount;
            }
        }

        internal override PropertyBagBase PropertyBagBase
        {
            get
            {
                return this.PropertyBag;
            }
        }

        internal CellsPropertyBag PropertyBag
        {
            get
            {
                return this.propertyBag;
            }
        }

        internal MergedCellRanges MergedCellRanges
        {
            get
            {
                return this.mergedCellRanges;
            }
        }

        #endregion


        #region Cell Properties

        internal IProperty<ICellValue> ValueProperty
        {
            get
            {
                return this.valueProperty;
            }
        }

        internal IProperty<IDataValidationRule> DataValidationRuleProperty
        {
            get
            {
                return this.dataValidationRuleProperty;
            }
        }

        #endregion


        #region Indexers

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.CellSelection" /> with the specified row and column indices.
        /// </summary>
        /// <value>The cell selection.</value>
        public CellSelection this[int rowIndex, int columnIndex]
        {
            get
            {
                return this.GetCellSelection(rowIndex, columnIndex);
            }
        }

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.CellSelection" /> with the specified cell index.
        /// </summary>
        /// <value>The cell selection.</value>
        public CellSelection this[CellIndex cellIndex]
        {
            get
            {
                return this.GetCellSelection(cellIndex);
            }
        }

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.CellSelection" /> with the specified start row index, start column index, end row index and end column index.
        /// </summary>
        /// <value>The cell selection.</value>
        public CellSelection this[int fromRowIndex, int fromColumnIndex, int toRowIndex, int toColumnIndex]
        {
            get
            {
                return this.GetCellSelection(fromRowIndex, fromColumnIndex, toRowIndex, toColumnIndex);
            }
        }

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.CellSelection" /> with the specified start and end indices.
        /// </summary>
        /// <value>The cell selection.</value>
        public CellSelection this[CellIndex fromIndex, CellIndex toIndex]
        {
            get
            {
                return this.GetCellSelection(fromIndex, toIndex);
            }
        }

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.CellSelection" /> with the specified cell range.
        /// </summary>
        /// <value>The cell selection.</value>
        public CellSelection this[CellRange cellRange]
        {
            get
            {
                return this.GetCellSelection(cellRange);
            }
        }

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.CellSelection" /> with the specified cell ranges.
        /// </summary>
        /// <value>The cell selection.</value>
        public CellSelection this[IEnumerable<CellRange> cellRanges]
        {
            get
            {
                return this.GetCellSelection(cellRanges);
            }
        }

        #endregion


        #region Constructors

        internal Cells(Worksheet worksheet)
            : base(worksheet)
        {
            this.propertyBag = new CellsPropertyBag();
            this.propertyBag.PropertyChanged += this.PropertyBag_PropertyChanged;
            this.mergedCellRanges = new MergedCellRanges(this);
            this.mergedCellRanges.Changing += this.MergedCellRanges_Changing;
            this.mergedCellRanges.Changed += this.MergedCellRanges_Changed;

            this.valueProperty = this.CreateProperty(CellPropertyDefinitions.ValueProperty);
            this.dataValidationRuleProperty = this.CreateProperty(CellPropertyDefinitions.DataValidationRuleProperty);
        }

        #endregion


        #region Event Handlers

        private void MergedCellRanges_Changing(object sender, EventArgs e)
        {
            this.OnMergedCellsChanging();
        }

        private void MergedCellRanges_Changed(object sender, MergedCellRangesChangedEventArgs args)
        {
            this.OnMergedCellsChanged(args);
        }

        private void PropertyBag_PropertyChanged(object sender, CellPropertyChangedEventArgs e)
        {
            this.OnCellPropertyChanged(e);
        }

        #endregion


        #region Methods

        internal bool IsDefault(CellIndex index)
        {
            long longIndex = CellsPropertyBag.ConvertCellIndexToLong(index);

            return this.PropertyBag.IsDefault(longIndex);
        }

        internal override IProperty<T> CreatePropertyOverride<T>(IPropertyDefinition<T> propertyDefinition)
        {
            CellProperty<T> property = new CellProperty<T>(this.Worksheet, propertyDefinition);

            return property;
        }

        /// <summary>
        /// Determines whether cells can be inserted at the specified row index.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <param name="itemCount">The item count.</param>
        /// <param name="shiftType">Type of the shift.</param>
        /// <returns>The value indicating whether cells can be inserted at the specified row index.</returns>
        public bool CanInsert(int rowIndex, int columnIndex, int itemCount, InsertShiftType shiftType)
        {
            CellRange range;
            if (shiftType == InsertShiftType.Right)
            {
                int newColumnIndex = columnIndex + itemCount - 1;
                if (!TelerikHelper.IsValidColumnIndex(newColumnIndex))
                {
                    return false;
                }
                range = new CellRange(rowIndex, columnIndex, rowIndex, newColumnIndex);
            }
            else
            {
                int newRowIndex = columnIndex + itemCount - 1;
                if (!TelerikHelper.IsValidRowIndex(newRowIndex))
                {
                    return false;
                }
                range = new CellRange(rowIndex, columnIndex, newRowIndex, columnIndex);
            }

            return this.CanInsert(range, shiftType);
        }

        /// <summary>
        /// Determines whether cells can be inserted at the specified cell range.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <param name="shiftType">Type of the shift.</param>
        /// <returns>The value indicating whether cells can be inserted at the specified cell range.</returns>
        public bool CanInsert(CellRange cellRange, InsertShiftType shiftType)
        {
            CanInsertRemoveResult canInsert = this.CanInsertInternal(cellRange, shiftType);

            return canInsert == CanInsertRemoveResult.Success || canInsert == CanInsertRemoveResult.MergedCells;
        }

        internal CanInsertRemoveResult CanInsertInternal(CellRange cellRange, InsertShiftType shiftType)
        {
            bool filterRangeAllowsInsert = this.CheckForFilterRangeBeforeInsert(cellRange, shiftType);

            if (!filterRangeAllowsInsert)
            {
                return CanInsertRemoveResult.Filtering;
            }

            bool valuesAllowInsert = this.PropertyBag.CanInsertCells(cellRange, shiftType);

            if (!valuesAllowInsert)
            {
                return CanInsertRemoveResult.ValueLoss;
            }

            ShiftType insertRemoveShiftType = shiftType.ToShiftType();
            bool mergedCellsAllowInsert = this.MergedCellRanges.CanInsertOrRemove(cellRange, insertRemoveShiftType);

            if (!mergedCellsAllowInsert)
            {
                return CanInsertRemoveResult.MergedCells;
            }

            return CanInsertRemoveResult.Success;
        }

        internal bool CanRemove(CellRange cellRange, RemoveShiftType shiftType)
        {
            CanInsertRemoveResult canRemove = this.CanRemoveInternal(cellRange, shiftType);

            return canRemove == CanInsertRemoveResult.Success || canRemove == CanInsertRemoveResult.MergedCells;
        }

        internal CanInsertRemoveResult CanRemoveInternal(CellRange cellRange, RemoveShiftType shiftType)
        {
            bool filterRangeAllowsRemove = this.CheckForFilterRangeBeforeRemove(cellRange, shiftType);

            if (!filterRangeAllowsRemove)
            {
                return CanInsertRemoveResult.Filtering;
            }

            ShiftType insertRemoveShiftType = shiftType.ToShiftType();
            bool mergedCellsAllowInsert = this.MergedCellRanges.CanInsertOrRemove(cellRange, insertRemoveShiftType);

            if (!mergedCellsAllowInsert)
            {
                return CanInsertRemoveResult.MergedCells;
            }

            return CanInsertRemoveResult.Success;
        }

        /// <summary>
        /// Inserts cells at the specified cell range.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <param name="shiftType">Type of the shift.</param>
        /// <returns>The value indicating whether the insertion was successful.</returns>
        public bool Insert(CellRange cellRange, InsertShiftType shiftType)
        {
            bool isSuccessful = false;
            using (new UpdateScope(this.Worksheet.BeginUndoGroup, this.Worksheet.EndUndoGroup))
            {
                this.UnmergeBeforeInsert(cellRange, shiftType);

                InsertCellCommandContext insertContext = new InsertCellCommandContext(this.Worksheet, cellRange, shiftType);
                isSuccessful = this.Worksheet.ExecuteCommand(WorkbookCommands.InsertCell, insertContext);

                if (isSuccessful)
                {
                    this.UpdateCellRangeInsertedOrRemovedDependentCollections(cellRange, shiftType.ToRangeType(false), false);
                }
            }

            return isSuccessful;
        }

        /// <summary>
        /// Inserts the specified number of items at the specified location.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <param name="itemCount">The item count.</param>
        /// <param name="shiftType">Type of the shift.</param>
        /// <returns>The value indicating whether the insertion was successful.</returns>
        public bool Insert(int rowIndex, int columnIndex, int itemCount, InsertShiftType shiftType)
        {
            CellRange range;
            if (shiftType == InsertShiftType.Right)
            {
                range = new CellRange(rowIndex, columnIndex, rowIndex, columnIndex + itemCount - 1);
            }
            else
            {
                range = new CellRange(rowIndex, columnIndex, rowIndex + itemCount - 1, columnIndex);
            }

            return this.Insert(range, shiftType);
        }

        /// <summary>
        /// Removes cells at the specified cell range.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <param name="shiftType">Type of the shift.</param>
        public void Remove(CellRange cellRange, RemoveShiftType shiftType)
        {
            using (new UpdateScope(this.Worksheet.BeginUndoGroup, this.Worksheet.EndUndoGroup))
            {
                this.UnmergeBeforeRemove(cellRange, shiftType);

                RemoveCellCommandContext context = new RemoveCellCommandContext(this.Worksheet, cellRange, shiftType);
                bool isSuccessful = this.Worksheet.ExecuteCommand(WorkbookCommands.RemoveCell, context);

                if (isSuccessful)
                {
                    this.UpdateCellRangeInsertedOrRemovedDependentCollections(cellRange, shiftType.ToRangeType(false), true);
                }
            }
        }

        /// <summary>
        /// Removes the specified number of items at the specified location.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <param name="itemCount">The item count.</param>
        /// <param name="shiftType">Type of the shift.</param>
        public void Remove(int rowIndex, int columnIndex, int itemCount, RemoveShiftType shiftType)
        {
            CellRange range;
            if (shiftType == RemoveShiftType.Left)
            {
                range = new CellRange(rowIndex, columnIndex, rowIndex, columnIndex + itemCount - 1);
            }
            else
            {
                range = new CellRange(rowIndex, columnIndex, rowIndex + itemCount - 1, columnIndex);
            }

            this.Remove(range, shiftType);
        }

        internal void InsertInternal(CellRange cellRange, InsertShiftType shiftType, CellsPropertyBag oldValues = null)
        {
            if (oldValues == null)
            {
                this.PropertyBag.InsertEmptyCells(cellRange, shiftType);
            }
            else
            {
                this.PropertyBag.InsertCellsAndSetValues(cellRange, shiftType, oldValues);
            }

            RangeType rangeType = shiftType.ToRangeType(false);

            this.OnCellRangeInsertedOrRemoved(cellRange, rangeType, false);
        }

        internal void RemoveInternal(CellRange cellRange, RemoveShiftType shiftType)
        {
            this.PropertyBag.RemoveCells(cellRange, shiftType);

            RangeType rangeType = shiftType.ToRangeType(false);

            this.OnCellRangeInsertedOrRemoved(cellRange, rangeType, true);
        }

        internal void InsertColumn(int index, int itemCount, CellsPropertyBag oldValues = null)
        {
            if (oldValues == null)
            {
                this.PropertyBag.InsertEmptyColumn(index, itemCount);
            }
            else
            {
                this.PropertyBag.InsertColumnAndSetValues(index, itemCount, oldValues);
            }

            CellRange range = CellRange.FromColumnRange(index, index + itemCount - 1);
            this.OnCellRangeInsertedOrRemoved(range, RangeType.Columns, false);
        }

        internal void RemoveColumn(int index, int itemCount)
        {
            this.PropertyBag.RemoveColumn(index, itemCount);

            CellRange range = CellRange.FromColumnRange(index, index + itemCount - 1);
            this.OnCellRangeInsertedOrRemoved(range, RangeType.Columns, true);
        }

        internal void InsertRow(int index, int itemCount, CellsPropertyBag oldValues = null)
        {
            if (oldValues == null)
            {
                this.PropertyBag.InsertEmptyRow(index, itemCount);
            }
            else
            {
                this.PropertyBag.InsertRowAndSetValues(index, itemCount, oldValues);
            }

            CellRange range = CellRange.FromRowRange(index, index + itemCount - 1);
            this.OnCellRangeInsertedOrRemoved(range, RangeType.Rows, false);
        }

        internal void RemoveRow(int index, int itemCount)
        {
            this.PropertyBag.RemoveRow(index, itemCount);

            CellRange range = CellRange.FromRowRange(index, index + itemCount - 1);
            this.OnCellRangeInsertedOrRemoved(range, RangeType.Rows, true);
        }

        internal FontProperties GetFontProperties(int rowIndex, int columnIndex, bool onlyPropertiesAffectingLayout = false)
        {
            long index = CellsPropertyBag.ConvertCellIndexToLong(rowIndex, columnIndex);

            return this.GetFontProperties(index, onlyPropertiesAffectingLayout);
        }

        internal FontProperties GetFontProperties(CellIndex cellIndex, bool onlyPropertiesAffectingLayout = false)
        {
            Guard.ThrowExceptionIfNull(cellIndex, "cellIndex");

            long index = CellsPropertyBag.ConvertCellIndexToLong(cellIndex);

            return this.GetFontProperties(index, onlyPropertiesAffectingLayout);
        }

        internal FontProperties GetFontProperties(long index, bool onlyPropertiesAffectingLayout = false)
        {
            ThemableFontFamily fontFamily = this.PropertyBag.GetPropertyValueRespectingStyle(CellPropertyDefinitions.FontFamilyProperty, this.Worksheet, index);
            double fontSize = this.PropertyBag.GetPropertyValueRespectingStyle(CellPropertyDefinitions.FontSizeProperty, this.Worksheet, index);
            bool isBold = this.PropertyBag.GetPropertyValueRespectingStyle(CellPropertyDefinitions.IsBoldProperty, this.Worksheet, index);

            FontProperties fontProperties = new FontProperties(
                fontFamily.GetActualValue(this.Worksheet.Workbook.Theme),
                fontSize,
                isBold);

            if (!onlyPropertiesAffectingLayout)
            {
                fontProperties.IsItalic = this.PropertyBag.GetPropertyValueRespectingStyle(CellPropertyDefinitions.IsItalicProperty, this.Worksheet, index);
                fontProperties.Underline = this.PropertyBag.GetPropertyValueRespectingStyle(CellPropertyDefinitions.UnderlineProperty, this.Worksheet, index);
                fontProperties.ForeColor = this.PropertyBag.GetPropertyValueRespectingStyle(CellPropertyDefinitions.ForeColorProperty, this.Worksheet, index);
            }

            return fontProperties;
        }

        internal CellRangeFontProperties GetFontProperties(CellRange cellRange, bool onlyPropertiesAffectingLayout = false)
        {
            return new CellRangeFontProperties(this, cellRange, onlyPropertiesAffectingLayout);
        }

        internal ICompressedList<T> GetPropertyValueRespectingStyle<T>(IPropertyDefinition<T> property, CellRange cellRange)
        {
            return this.propertyBag.GetPropertyValueRespectingStyle(property, this.Worksheet, cellRange);
        }

        internal ICompressedList<T> GetPropertyValueRespectingStyle<T>(IPropertyDefinition<T> property, long fromIndex, long toIndex)
        {
            return this.propertyBag.GetPropertyValueRespectingStyle(property, this.Worksheet, fromIndex, toIndex);
        }

        internal void UpdateCellRangeInsertedOrRemovedDependentCollections(CellRange cellRange, RangeType rangeType, bool isRemove)
        {
            CellRangeInsertedOrRemovedEventArgs eventArgs = new CellRangeInsertedOrRemovedEventArgs(cellRange, rangeType, isRemove);

            this.MergedCellRanges.Update(eventArgs);
            this.Worksheet.Hyperlinks.Update(eventArgs);
            this.Worksheet.WorksheetPageSetup.PrintArea.Update(eventArgs);
            this.Worksheet.Filter.Update(eventArgs);
            this.Worksheet.SortState.Update(eventArgs);

            this.UpdateCellsInsertedOrRemovedDependentCellValues(eventArgs);
            this.UpdateCellsInsertedOrRemovedDependentDataValidationRules(eventArgs);
        }

        private void UpdateCellsInsertedOrRemovedDependentDataValidationRules(CellRangeInsertedOrRemovedEventArgs eventArgs)
        {
            using (new UpdateScope(this.PropertyBag.SuspendPropertyChanged, this.PropertyBag.ResumePropertyChanged))
            {
                foreach (LongRange range in CellsPropertyBag.ConvertCellRangeToLongRanges(eventArgs.AffectedRange))
                {
                    ICompressedList<IDataValidationRule> compressedList = this.Worksheet.Cells.PropertyBag.GetPropertyValue(CellPropertyDefinitions.DataValidationRuleProperty, range.Start, range.End);
                    foreach (Range<long, IDataValidationRule> item in compressedList.GetNonDefaultRanges())
                    {
                        SingleArgumentDataValidationRuleBase rule = item.Value as SingleArgumentDataValidationRuleBase;
                        if (rule != null)
                        {
                            CellIndex fromCellIndex = CellsPropertyBag.ConvertLongToCellIndex(item.Start);
                            CellIndex toCellIndex = CellsPropertyBag.ConvertLongToCellIndex(item.End);

                            IDataValidationRule value = rule.CloneAndTranslate(this.Worksheet, fromCellIndex.RowIndex, fromCellIndex.ColumnIndex, false);
                            if (!rule.Equals(value))
                            {
                                this[fromCellIndex, toCellIndex].SetDataValidationRule(value);
                            }
                        }
                    }
                }
            }
        }

        internal void UpdateCellsInsertedOrRemovedDependentCellValues(CellRangeInsertedOrRemovedEventArgs eventArgs)
        {
            ICompressedList<ICellValue> compressedList = this.Worksheet.Cells.PropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.ValueProperty);

            using (new UpdateScope(this.PropertyBag.SuspendPropertyChanged, this.PropertyBag.ResumePropertyChanged))
            {
                foreach (Range<long, ICellValue> item in compressedList.GetNonDefaultRanges())
                {
                    FormulaCellValue formulaCellValue = item.Value as FormulaCellValue;
                    if (formulaCellValue != null && formulaCellValue.IsCellsNameDependent)
                    {
                        CellIndex fromCellIndex = CellsPropertyBag.ConvertLongToCellIndex(item.Start);
                        CellIndex toCellIndex = CellsPropertyBag.ConvertLongToCellIndex(item.End);

                        ICellValue value = formulaCellValue.CloneAndTranslate(this.Worksheet, fromCellIndex.RowIndex, fromCellIndex.ColumnIndex, eventArgs);
                        if (!formulaCellValue.Equals(value))
                        {
                            this[fromCellIndex, toCellIndex].SetValue(value);
                        }
                    }
                }
            }
        }

        private void UnmergeBeforeInsert(CellRange range, InsertShiftType shiftType)
        {
            switch (shiftType)
            {
                case InsertShiftType.Right:
                    this.UnmergeBeforeInsertRemove(range, ShiftType.Right);
                    break;
                case InsertShiftType.Down:
                    this.UnmergeBeforeInsertRemove(range, ShiftType.Down);
                    break;
            }
        }

        private void UnmergeBeforeRemove(CellRange range, RemoveShiftType shiftType)
        {
            switch (shiftType)
            {
                case RemoveShiftType.Left:
                    this.UnmergeBeforeInsertRemove(range, ShiftType.Left);
                    break;
                case RemoveShiftType.Up:
                    this.UnmergeBeforeInsertRemove(range, ShiftType.Up);
                    break;
            }
        }

        private void UnmergeBeforeInsertRemove(CellRange range, ShiftType shiftType)
        {
            List<CellRange> possibleRangesToUnmerge = new List<CellRange>(this.MergedCellRanges);

            for (int i = 0; i < possibleRangesToUnmerge.Count; i++)
            {
                if (!this.MergedCellRanges.CanInsertOrRemoveInternal(shiftType, possibleRangesToUnmerge[i], range))
                {
                    this.Worksheet.Cells[possibleRangesToUnmerge[i]].Unmerge();
                }
            }
        }

        private bool CheckForFilterRangeBeforeInsert(CellRange cellRange, InsertShiftType shiftType)
        {
            switch (shiftType)
            {
                case InsertShiftType.Right:
                    return this.Worksheet.Filter.CanInsertOrRemove(ShiftType.Right, cellRange);
                case InsertShiftType.Down:
                    return this.Worksheet.Filter.CanInsertOrRemove(ShiftType.Down, cellRange);
            }

            throw new NotSupportedException("Invalid insert shift type.");
        }

        private bool CheckForFilterRangeBeforeRemove(CellRange cellRange, RemoveShiftType shiftType)
        {
            switch (shiftType)
            {
                case RemoveShiftType.Left:
                    return this.Worksheet.Filter.CanInsertOrRemove(ShiftType.Left, cellRange);
                case RemoveShiftType.Up:
                    return this.Worksheet.Filter.CanInsertOrRemove(ShiftType.Up, cellRange);
            }

            throw new NotSupportedException("Invalid remove shift type.");
        }

        internal void UpdateSpreadsheetNameDependentCellValues(string oldName, string newName)
        {
            this.UpdateFormulaCellValues((value) => { return value.ContainsSpreadsheetName; }, (value) => { value.Translate(oldName, newName); });
        }

        internal void UpdateSheetNameDependentCellValues(Worksheet renamedWorksheet)
        {
            this.UpdateFormulaCellValues((value) => { return value.IsSheetNameDependent; }, (value) => { value.Translate(renamedWorksheet); });
        }

        internal void UpdateWorkbookNameDependentCellValues(Workbook renamedWorkbook)
        {
            this.UpdateFormulaCellValues((value) => { return value.IsWorkbookNameDependent; }, (value) => { value.Translate(renamedWorkbook); });
        }

        private void UpdateFormulaCellValues(Func<FormulaCellValue, bool> condition, Action<FormulaCellValue> action)
        {
            ICompressedList<ICellValue> compressedList = this.Worksheet.Cells.PropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.ValueProperty);

            using (new UpdateScope(this.PropertyBag.SuspendPropertyChanged, this.PropertyBag.ResumePropertyChanged))
            {
                foreach (Range<long, ICellValue> item in compressedList.GetNonDefaultRanges())
                {
                    FormulaCellValue formulaCellValue = item.Value as FormulaCellValue;
                    if (formulaCellValue != null && condition(formulaCellValue))
                    {
                        action(formulaCellValue);
                    }
                }
            }
        }

        internal override void CopyFromOverride(WorksheetEntityBase fromWorksheetEntity, CopyContext context, IEnumerable<IProperty> properties)
        {
            base.CopyFromOverride(fromWorksheetEntity, context, properties.Except(new IProperty[] { this.ValueProperty }));

            IPropertyDefinition<string> styleNamePrpertyDefinition = CellPropertyDefinitions.StyleNameProperty;
            ICompressedList<string> styleNamesCollection = this.PropertyBag.GetPropertyValueCollection(styleNamePrpertyDefinition);

            Range<long, string>[] nonDefaultRanges = styleNamesCollection.GetNonDefaultRanges().ToArray();

            for (long rangeIndex = 0; rangeIndex < nonDefaultRanges.Length; rangeIndex++)
            {
                Range<long, string> range = nonDefaultRanges[rangeIndex];
                string renamedStyleName;
                if (context.OriginalToRenamedStyleNames.TryGetValue(range.Value, out renamedStyleName))
                {
                    styleNamesCollection.SetValue(range.Start, range.End, renamedStyleName);
                }
            }

            IPropertyDefinition<ICellValue> valuePropertyDefinition = CellPropertyDefinitions.ValueProperty;
            ICompressedList<ICellValue> valuePropertyCollection = fromWorksheetEntity.PropertyBagBase.GetPropertyValueCollection(valuePropertyDefinition);

            ICompressedList<ICellValue> targetValuePropertyCollection = context.TargetWorksheet.Cells.PropertyBag.GetPropertyValueCollection(valuePropertyDefinition);

            foreach (Range<long, ICellValue> range in valuePropertyCollection.GetNonDefaultRanges())
            {
                if (range.Value.ValueType == CellValueType.Formula)
                {
                    FormulaCellValue formulaCellValue = (FormulaCellValue)range.Value;
                    for (long i = range.Start; i <= range.End; i++)
                    {
                        int rowIndex;
                        int columnIndex;
                        CellsPropertyBag.ConvertLongToRowAndColumnIndexes(i, out rowIndex, out columnIndex);

                        FormulaCellValue formulaCellValueCopied = ((ICopyable<FormulaCellValue>)formulaCellValue).Copy(context);
                        targetValuePropertyCollection.SetValue(i, formulaCellValueCopied);
                    }
                }
                else
                {
                    targetValuePropertyCollection.SetValue(range.Start, range.End, range.Value);
                }
            }

            this.MergedCellRanges.CopyFrom(((Cells)fromWorksheetEntity).MergedCellRanges);
        }

        internal override void Clear()
        {
            base.Clear();

            this.MergedCellRanges.RemoveMergedRanges(this.MergedCellRanges.Ranges.ToList());
        }

        #endregion


        #region GetSelection Methods

        /// <summary>
        /// Gets cell selection.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The cell selection.</returns>
        public CellSelection GetCellSelection(int rowIndex, int columnIndex)
        {
            return this.GetCellSelection(new CellIndex(rowIndex, columnIndex));
        }

        /// <summary>
        /// Gets cell selection.
        /// </summary>
        /// <param name="cellIndex">Index of the cell.</param>
        /// <returns>The cell selection.</returns>
        public CellSelection GetCellSelection(CellIndex cellIndex)
        {
            return this.GetCellSelection(cellIndex, cellIndex);
        }

        /// <summary>
        /// Gets cell selection.
        /// </summary>
        /// <param name="fromRowIndex">Start row index.</param>
        /// <param name="fromColumnIndex">Start column index.</param>
        /// <param name="toRowIndex">End row index.</param>
        /// <param name="toColumnIndex">End column index.</param>
        /// <returns>The cell selection.</returns>
        public CellSelection GetCellSelection(int fromRowIndex, int fromColumnIndex, int toRowIndex, int toColumnIndex)
        {
            return this.GetCellSelection(new CellRange(fromRowIndex, fromColumnIndex, toRowIndex, toColumnIndex));
        }

        /// <summary>
        /// Gets cell selection.
        /// </summary>
        /// <param name="fromIndex">Start index.</param>
        /// <param name="toIndex">End index.</param>
        /// <returns>The cell selection.</returns>
        public CellSelection GetCellSelection(CellIndex fromIndex, CellIndex toIndex)
        {
            return this.GetCellSelection(new CellRange(fromIndex, toIndex));
        }

        /// <summary>
        /// Gets cell selection.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <returns>The cell selection.</returns>
        public CellSelection GetCellSelection(CellRange cellRange)
        {
            return this.GetCellSelection(new CellRange[] { cellRange });
        }

        /// <summary>
        /// Gets the cell selection.
        /// </summary>
        /// <param name="cellRanges">The cell ranges.</param>
        /// <returns>The cell selection.</returns>
        public CellSelection GetCellSelection(IEnumerable<CellRange> cellRanges)
        {
            return new CellSelection(this.Worksheet, cellRanges);
        }

        #endregion


        #region Merge/Unmerge

        /// <summary>
        /// Tries to get the merged range containing the index.
        /// </summary>
        /// <param name="cellIndex">Index of the cell.</param>
        /// <param name="mergedRange">The merged range.</param>
        /// <returns>The value indicating whether precisely a range was found.</returns>
        public bool TryGetContainingMergedRange(CellIndex cellIndex, out CellRange mergedRange)
        {
            Guard.ThrowExceptionIfNull(cellIndex, "cellIndex");

            return this.MergedCellRanges.TryGetContainingMergedRange(cellIndex, out mergedRange);
        }

        internal bool TryGetContainingMergedRange(int rowIndex, int columnIndex, out CellRange mergedRange)
        {
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            return this.MergedCellRanges.TryGetContainingMergedRange(rowIndex, columnIndex, out mergedRange);
        }

        /// <summary>
        /// Gets the value indicating whether the cell index is contained in a merged range.
        /// </summary>
        /// <param name="cellIndex">Index of the cell.</param>
        /// <returns>The value indicating whether the cell index is contained in a merged range.</returns>
        public bool GetIsMerged(CellIndex cellIndex)
        {
            Guard.ThrowExceptionIfNull(cellIndex, "cellIndex");

            CellRange mergedRange;
            return this.TryGetContainingMergedRange(cellIndex, out mergedRange);
        }

        /// <summary>
        /// Gets the merge state.
        /// </summary>
        /// <param name="cellIndex">Index of the cell.</param>
        /// <returns>The merge state.</returns>
        public CellMergeState GetMergeState(CellIndex cellIndex)
        {
            Guard.ThrowExceptionIfNull(cellIndex, "cellIndex");

            return this.GetMergeState(cellIndex.RowIndex, cellIndex.ColumnIndex);
        }

        /// <summary>
        /// Gets the merge state.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The merge state.</returns>
        public CellMergeState GetMergeState(int rowIndex, int columnIndex)
        {
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            CellMergeState mergeState = CellMergeState.NotMerged;
            CellRange mergedRange;
            if (this.Worksheet.Cells.TryGetContainingMergedRange(rowIndex, columnIndex, out mergedRange))
            {
                if (mergedRange.FromIndex.RowIndex == rowIndex &&
                    mergedRange.FromIndex.ColumnIndex == columnIndex)
                {
                    mergeState = CellMergeState.TopLeftCellInMergedRange;
                }
                else
                {
                    mergeState = CellMergeState.NonTopLeftCellInMergedRange;
                }
            }

            return mergeState;
        }

        /// <summary>
        /// Gets the merged cell ranges.
        /// </summary>
        /// <returns>The cell ranges.</returns>
        public IEnumerable<CellRange> GetMergedCellRanges()
        {
            return this.MergedCellRanges.Ranges;
        }

        /// <summary>
        /// Gets the merged ranges containing the specified cell range.
        /// </summary>
        /// <param name="range">The range.</param>
        /// <returns>The merged ranges containing the specified cell range.</returns>
        public IEnumerable<CellRange> GetContainingMergedRanges(CellRange range)
        {
            ISet<CellRange> ranges = this.mergedCellRanges.GetIntersectingMergedRanges(range);

            foreach (CellRange mergedRange in ranges)
            {
                if (range.Contains(mergedRange))
                {
                    yield return mergedRange;
                }
            }
        }

        #endregion


        #region Events

        internal void OnCellRangeInsertedOrRemoved(CellRange range, RangeType rangeType, bool isRemove)
        {
            this.OnCellRangeInsertedOrRemoved(new CellRangeInsertedOrRemovedEventArgs(range, rangeType, isRemove));
        }

        /// <summary>
        /// Occurs when a cell range is inserted or removed.
        /// </summary>
        public event EventHandler<CellRangeInsertedOrRemovedEventArgs> CellRangeInsertedOrRemoved;

        /// <summary>
        /// Raises the <see cref="E:CellRangeInsertedOrRemoved" /> event.
        /// </summary>
        /// <param name="args">The <see cref="Telerik.Windows.Documents.Spreadsheet.Model.CellRangeInsertedOrRemovedEventArgs" /> instance containing the event data.</param>
        protected virtual void OnCellRangeInsertedOrRemoved(CellRangeInsertedOrRemovedEventArgs args)
        {
            if (this.CellRangeInsertedOrRemoved != null)
            {
                this.CellRangeInsertedOrRemoved(this, args);
            }
        }

        internal event EventHandler MergedCellsChanging;

        private void OnMergedCellsChanging()
        {
            if (this.MergedCellsChanging != null)
            {
                this.MergedCellsChanging(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Occurs when merged cells are changed.
        /// </summary>
        public event EventHandler<MergedCellRangesChangedEventArgs> MergedCellsChanged;

        /// <summary>
        /// Raises the <see cref="E:MergedCellsChanged" /> event.
        /// </summary>
        /// <param name="args">The <see cref="Telerik.Windows.Documents.Spreadsheet.Model.MergedCellRangesChangedEventArgs" /> instance containing the event data.</param>
        protected virtual void OnMergedCellsChanged(MergedCellRangesChangedEventArgs args)
        {
            if (this.MergedCellsChanged != null)
            {
                this.MergedCellsChanged(this, args);
            }
        }

        /// <summary>
        /// Occurs when a cell property is changed.
        /// </summary>
        public event EventHandler<CellPropertyChangedEventArgs> CellPropertyChanged;

        /// <summary>
        /// Raises the <see cref="E:CellPropertyChanged" /> event.
        /// </summary>
        /// <param name="args">The <see cref="Telerik.Windows.Documents.Spreadsheet.PropertySystem.CellPropertyChangedEventArgs" /> instance containing the event data.</param>
        protected virtual void OnCellPropertyChanged(CellPropertyChangedEventArgs args)
        {
            if (this.CellPropertyChanged != null)
            {
                this.CellPropertyChanged(this, args);
            }
        }

        #endregion
    }
}
