﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Model.Drawing.Charts;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Charts
{
    /// <summary>
    /// Provides methods for creation of DocumentChart object.
    /// </summary>
    internal static class DocumentChartFactory
    {
        /// <summary>
        /// Creates a DocumentChart with structure based on the passed data.
        /// </summary>
        /// <param name="worksheet">The worksheet containing the chart data.</param>
        /// <param name="initialRange">The range containing the chart data.</param>
        /// <param name="seriesOrientation">A value indicating whether the series of the chart will refer to vertical or horizontal ranges or it will be determined automatically.</param>
        /// <param name="chartTypes">The types of chart that will be created. Passing more than one type will create a combo chart.</param>
        /// <returns>The result DocumentChart object.</returns>
        public static DocumentChart CreateChartFromRangeOfType(Worksheet worksheet, CellRange initialRange, SeriesRangesOrientation seriesOrientation, params ChartType[] chartTypes)
        {
            bool? seriesAreVertical;

            switch (seriesOrientation)
            {
                case SeriesRangesOrientation.Automatic: seriesAreVertical = null; break;
                case SeriesRangesOrientation.Vertical: seriesAreVertical = true; break;
                case SeriesRangesOrientation.Horizontal: seriesAreVertical = false; break;
                default: throw new NotSupportedException();
            }

            RangeParseResultInfo parseResult = RangeChartParser.ParseRange(worksheet, initialRange, seriesAreVertical);
            return CreateChartFromParsedRange(worksheet, parseResult, chartTypes);
        }

        private static DocumentChart CreateChartFromParsedRange(Worksheet worksheet, RangeParseResultInfo parseResult, params ChartType[] chartTypes)
        {
            DocumentChart newChart = new DocumentChart();

            PopulateSeriesGroups(newChart, chartTypes);

            PopulateSeries(worksheet, newChart, parseResult);
            PopulateAxes(newChart, parseResult, worksheet, chartTypes);
            PopulateLegend(newChart);

            return newChart;
        }

        private static void PopulateSeriesGroups(DocumentChart newChart, params ChartType[] chartTypes)
        {
            if (chartTypes.Length == 0)
            {
                newChart.SeriesGroups.Add(SeriesGroupFactory.GetSeriesGroup(ChartType.Bar));
            }

            foreach (ChartType chartType in chartTypes)
            {
                newChart.SeriesGroups.Add(SeriesGroupFactory.GetSeriesGroup(chartType));
            }
        }

        private static void PopulateAxes(DocumentChart newChart, RangeParseResultInfo parseResult, Worksheet worksheet, ChartType[] chartTypes)
        {
            if (!newChart.SeriesGroups.Any((c) => c is ISupportAxes))
            {
                return;
            }

            Axis firstAxis = AxisFactory.GetHorizontalAxis(parseResult, worksheet, chartTypes);
            Axis secondAxis = AxisFactory.GetVerticalAxis(parseResult, worksheet);

            AxisGroup primaryGroup = new AxisGroup(firstAxis, secondAxis);

            newChart.PrimaryAxes = primaryGroup;

            foreach (SeriesGroup seriesGroup in newChart.SeriesGroups)
            {
                ISupportAxes seriesGroupWithAxes = seriesGroup as ISupportAxes;

                if (seriesGroupWithAxes != null)
                {
                    seriesGroupWithAxes.AxisGroupName = AxisGroupName.Primary;
                }
            }
        }

        private static void PopulateSeries(Worksheet worksheet, DocumentChart newChart, RangeParseResultInfo parseResult)
        {
            int dataRangesStartIndex = parseResult.SeriesRangesAreVertical ? parseResult.DataRange.FromIndex.ColumnIndex : parseResult.DataRange.FromIndex.RowIndex;
            int dataRangesEndIndex = parseResult.SeriesRangesAreVertical ? parseResult.DataRange.ToIndex.ColumnIndex : parseResult.DataRange.ToIndex.RowIndex;
            int currentDataRangesIndex = dataRangesStartIndex;

            Dictionary<int, int> seriesGroupIndexToSeriesCount = new Dictionary<int, int>();

            List<SeriesGroup> seriesGroups = newChart.SeriesGroups.ToList();
            while (currentDataRangesIndex <= dataRangesEndIndex)
            {
                bool success = false;
                for (int i = 0; i < seriesGroups.Count; i++)
                {
                    SeriesGroup seriesGroup = seriesGroups[i];
                    int rangesPerSeries = GetDataRangesPerSeries(parseResult, seriesGroup);

                    currentDataRangesIndex += rangesPerSeries;
                    if (currentDataRangesIndex > dataRangesEndIndex + 1)
                    {
                        currentDataRangesIndex -= rangesPerSeries;
                        continue;
                    }

                    if (!seriesGroupIndexToSeriesCount.ContainsKey(i))
                    {
                        seriesGroupIndexToSeriesCount[i] = 0;
                    }

                    seriesGroupIndexToSeriesCount[i]++;
                    success = true;
                }

                if (!success)
                {
                    break;
                }
            }

            currentDataRangesIndex = dataRangesStartIndex;

            for (int seriesIndex = 0; seriesIndex < seriesGroups.Count; seriesIndex++)
            {
                SeriesGroup seriesGroup = seriesGroups[seriesIndex];
                int dataRangesPerSeries = GetDataRangesPerSeries(parseResult, seriesGroup);
                int seriesCount = 1;

                if (seriesGroupIndexToSeriesCount.ContainsKey(seriesIndex))
                {
                    seriesCount = seriesGroupIndexToSeriesCount[seriesIndex];
                }

                for (int i = 0; i < seriesCount; i++)
                {
                    AddSeries(worksheet, parseResult, currentDataRangesIndex, seriesGroup);
                    currentDataRangesIndex += dataRangesPerSeries;
                }
            }
        }

        private static void AddSeries(Worksheet worksheet, RangeParseResultInfo parseResult, int seriesIndex, SeriesGroup seriesGroup)
        {
            int dataRangesPerSeries = GetDataRangesPerSeries(parseResult, seriesGroup);

            CellRange[] valuesData = new CellRange[dataRangesPerSeries];

            bool areSeriesVertical = parseResult.SeriesRangesAreVertical;

            for (int i = 0; i < dataRangesPerSeries; i++)
            {
                int dataRangeIndex = seriesIndex + i;
                CellIndex fromIndex = parseResult.DataRange.FromIndex;
                CellIndex toIndex = parseResult.DataRange.ToIndex;

                CellRange seriesValuesRange = null;

                if (areSeriesVertical)
                {
                    seriesValuesRange = new CellRange(fromIndex.RowIndex, dataRangeIndex, toIndex.RowIndex, dataRangeIndex);
                }
                else
                {
                    seriesValuesRange = new CellRange(dataRangeIndex, fromIndex.ColumnIndex, dataRangeIndex, toIndex.ColumnIndex);
                }

                if (parseResult.DataRange.Contains(seriesValuesRange))
                {
                    valuesData[i] = seriesValuesRange;
                }
            }

            CellRange seriesTitleRange = null;

            CellRange titleRanges = parseResult.SeriesTitlesRange;
            if (titleRanges != null)
            {
                CellIndex fromIndex = titleRanges.FromIndex;
                CellIndex toIndex = titleRanges.ToIndex;

                if (areSeriesVertical)
                {
                    seriesTitleRange = new CellRange(fromIndex.RowIndex, seriesIndex, toIndex.RowIndex, seriesIndex);
                }
                else
                {
                    seriesTitleRange = new CellRange(seriesIndex, fromIndex.ColumnIndex, seriesIndex, toIndex.ColumnIndex);
                }
            }

            CellRange categoriesRange = parseResult.CategoriesRange;
            if (seriesGroup.HasCategories && parseResult.CategoriesRange == null && parseResult.SeriesTitlesRange != null)
            {
                categoriesRange = parseResult.SeriesTitlesRange;
            }

            CellRange firstValuesRange = valuesData.First();
            if (categoriesRange != null)
            {
                if (!areSeriesVertical && categoriesRange.ColumnCount < firstValuesRange.ColumnCount)
                {
                    categoriesRange = null;
                }
                else if (areSeriesVertical && categoriesRange.RowCount < firstValuesRange.RowCount)
                {
                    categoriesRange = null;
                }
            }

            IChartData xValues = categoriesRange == null ? null : new WorkbookFormulaChartData(worksheet, categoriesRange);
            Title seriesTitle = seriesTitleRange != null ? new FormulaTitle(new WorkbookFormulaChartData(worksheet, seriesTitleRange)) : null;

            IChartData values = null;
            IChartData bubbleSize = null;

            if (seriesGroup.HasCategories && xValues == null)
            {
                xValues = GenerateIncreasingAutoValues(parseResult, areSeriesVertical);
            }

            if (valuesData.Length == 1)
            {
                values = new WorkbookFormulaChartData(worksheet, valuesData[0]);
            }
            else if (valuesData.Length == 2 && xValues == null)
            {
                if (valuesData[0] != null && valuesData[1] != null)
                {
                    xValues = new WorkbookFormulaChartData(worksheet, valuesData[0]);
                    values = new WorkbookFormulaChartData(worksheet, valuesData[1]);
                }
                else if (valuesData[0] != null && valuesData[1] == null)
                {
                    xValues = GenerateIncreasingAutoValues(parseResult, areSeriesVertical);
                    values = new WorkbookFormulaChartData(worksheet, valuesData[0]);
                }
            }
            else if (valuesData.Length == 2 && xValues != null)
            {
                if (valuesData[0] != null && valuesData[1] != null)
                {
                    values = new WorkbookFormulaChartData(worksheet, valuesData[0]);
                    xValues = GenerateIncreasingAutoValues(parseResult, areSeriesVertical);
                    bubbleSize = new WorkbookFormulaChartData(worksheet, valuesData[1]);
                }
                else if (valuesData[0] != null && valuesData[1] == null)
                {
                    values = new WorkbookFormulaChartData(worksheet, valuesData[0]);
                    xValues = GenerateIncreasingAutoValues(parseResult, areSeriesVertical);
                    bubbleSize = GenerateConstantAutoValues(parseResult, areSeriesVertical, 1);
                }
            }
            else if (valuesData.Length == 3)
            {
                if (valuesData[0] != null && valuesData[1] != null && valuesData[2] != null)
                {
                    xValues = new WorkbookFormulaChartData(worksheet, valuesData[0]);
                    values = new WorkbookFormulaChartData(worksheet, valuesData[1]);
                    bubbleSize = new WorkbookFormulaChartData(worksheet, valuesData[2]);
                }
                else if (valuesData[0] != null && valuesData[1] == null && valuesData[2] == null)
                {
                    xValues = GenerateIncreasingAutoValues(parseResult, areSeriesVertical);
                    values = new WorkbookFormulaChartData(worksheet, valuesData[0]);
                    bubbleSize = GenerateConstantAutoValues(parseResult, areSeriesVertical, 1);
                }
                else if (valuesData[0] != null && valuesData[1] != null && valuesData[2] == null)
                {
                    xValues = GenerateIncreasingAutoValues(parseResult, areSeriesVertical);
                    values = new WorkbookFormulaChartData(worksheet, valuesData[0]);
                    bubbleSize = new WorkbookFormulaChartData(worksheet, valuesData[1]);
                }
            }

            seriesGroup.Series.AddOverride(xValues, values, bubbleSize, seriesTitle);
        }

        private static IChartData GenerateIncreasingAutoValues(RangeParseResultInfo parseResult, bool areSeriesVertical)
        {
            IChartData xValues;
            CellRange dataRange = parseResult.DataRange;
            int valuesCount = areSeriesVertical ? dataRange.RowCount : dataRange.ColumnCount;
            xValues = new NumericChartData(Enumerable.Range(1, valuesCount).Select(p => (double)p));
            return xValues;
        }

        private static IChartData GenerateConstantAutoValues(RangeParseResultInfo parseResult, bool areSeriesVertical, double value)
        {
            IChartData xValues;
            CellRange dataRange = parseResult.DataRange;
            int valuesCount = areSeriesVertical ? dataRange.RowCount : dataRange.ColumnCount;
            xValues = new NumericChartData(Enumerable.Repeat(value, valuesCount));
            return xValues;
        }

        private static int GetDataRangesPerSeries(RangeParseResultInfo parseResult, SeriesGroup seriesGroup)
        {
            int rangesPerSeries = seriesGroup.DataRangesPerSeries;
            if (!seriesGroup.HasCategories && parseResult.CategoriesRange != null)
            {
                rangesPerSeries--;
            }

            return rangesPerSeries;
        }

        private static void PopulateLegend(DocumentChart newChart)
        {
            newChart.Legend = new Legend();
            newChart.Legend.Position = LegendPosition.Bottom;
        }
    }
}
