﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Charts
{
    internal static class RangeChartParser
    {
        public static RangeParseResultInfo ParseRange(Worksheet worksheet, CellRange range, bool? seriesAreVertical = null)
        {
            CellRange dataRange;

            bool success = TryGetDataRangeFromDataPositioning(worksheet, range, out dataRange);

            if (!success)
            {
                dataRange = GetDataRangeFromContent(worksheet, range);
            }

            CellRange categoriesRange;
            CellRange seriesTitlesRange;

            if (!seriesAreVertical.HasValue)
            {
                seriesAreVertical = DetermineWhetherSeriesAreVertical(dataRange);
            }

            GetCategoriesAndSeriesTitlesRangesFromKnownDataRange(range, dataRange, out categoriesRange, out seriesTitlesRange, seriesAreVertical.Value);

            return new RangeParseResultInfo()
            {
                DataRange = dataRange,
                SeriesTitlesRange = seriesTitlesRange,
                CategoriesRange = categoriesRange,
                SeriesRangesAreVertical = seriesAreVertical.Value
            };
        }

        private static CellRange CombineRanges(CellRange categoriesRange, CellRange dataRange)
        {
            if (categoriesRange == null)
            {
                return dataRange;
            }

            int fromRowIndex = Math.Min(categoriesRange.FromIndex.RowIndex, dataRange.FromIndex.RowIndex);
            int fromColumnIndex = Math.Min(categoriesRange.FromIndex.ColumnIndex, dataRange.FromIndex.ColumnIndex);

            int toRowIndex = Math.Max(categoriesRange.ToIndex.RowIndex, dataRange.ToIndex.RowIndex);
            int toColumnIndex = Math.Max(categoriesRange.ToIndex.ColumnIndex, dataRange.ToIndex.ColumnIndex);

            return new CellRange(fromRowIndex, fromColumnIndex, toRowIndex, toColumnIndex);
        }

        /// <summary>
        /// This method looks for an empty range at the top left part of the range. If the top left cell is empty, it first goes along the top row,
        /// to try and find the end of the empty range. If it finds non-empty cell, it concludes that this is the start of the top title range and the 
        /// left title range starts from the second row below. If the entire first row is empty, it repeats the same along the first column. If the entire
        /// first row and first column are empty, the first row and column are considered to be the title ranges.
        /// </summary>
        private static bool TryGetDataRangeFromDataPositioning(Worksheet worksheet, CellRange range, out CellRange dataRange)
        {
            dataRange = null;

            if (worksheet.Cells[range.FromIndex].GetValue().Value.ValueType != CellValueType.Empty)
            {
                return false;
            }

            int resultColumnIndex;
            int resultRowIndex;

            GetCellIndexDeterminingEmptyRangeInTopLeftCorner(worksheet, range, out resultColumnIndex, out resultRowIndex);
            dataRange = ConstructDataRange(range, resultColumnIndex, resultRowIndex);

            return true;
        }

        private static void GetCellIndexDeterminingEmptyRangeInTopLeftCorner(Worksheet worksheet, CellRange range, out int resultColumnIndex, out int resultRowIndex)
        {
            int rangeStartRow = range.FromIndex.RowIndex;
            int rangeStartColumn = range.FromIndex.ColumnIndex;

            int rangeEndRow = range.ToIndex.RowIndex;
            int rangeEndColumn = range.ToIndex.ColumnIndex;

            resultColumnIndex = rangeStartColumn;
            resultRowIndex = rangeStartRow;

            for (int col = rangeStartColumn + 1; col <= rangeEndColumn; col++)
            {
                if (worksheet.Cells[rangeStartRow, col].GetValue().Value.ValueType != CellValueType.Empty)
                {
                    resultColumnIndex = col - 1;
                    break;
                }
            }

            for (int row = rangeStartRow + 1; row <= rangeEndRow; row++)
            {
                if (worksheet.Cells[row, resultColumnIndex].GetValue().Value.ValueType != CellValueType.Empty)
                {
                    resultRowIndex = row - 1;
                    break;
                }
            }
        }

        /// <summary>
        /// The range will be divided into a data range and "titles" ranges, starting from the bottom right cell and working our way to the top and left. The titles part contain the
        /// ranges for the categories and the series titles. Where the data range ends is determined by going along the bottom row in a right-to-left direction and finding the first cell
        /// that is considered title cell (that is, non-data). These are cells containing text or dates, in the case when the bottom right cell is not a date. If the bottom right cell
        /// contains a date, dates are considered part of the value range.
        /// </summary>
        private static CellRange GetDataRangeFromContent(Worksheet worksheet, CellRange range)
        {
            int lastRowLastTitleCellColumnIndex;
            int lastColumnLastTitleCellRowIndex;

            GetLastTitleCellsInLastRowAndLastColumnOfRange(worksheet, range, out lastRowLastTitleCellColumnIndex, out lastColumnLastTitleCellRowIndex);
            CellRange dataRange = ConstructDataRange(range, lastRowLastTitleCellColumnIndex, lastColumnLastTitleCellRowIndex);

            return dataRange;
        }

        private static void GetLastTitleCellsInLastRowAndLastColumnOfRange(Worksheet worksheet, CellRange range, out int lastRowLastTitleCellColumnIndex, out int lastColumnLastTitleCellRowIndex)
        {
            bool datesAreConsideredTitles = !SelectionContainsDates(worksheet.Cells[range.ToIndex]);

            int rangeStartRow = range.FromIndex.RowIndex;
            int rangeStartColumn = range.FromIndex.ColumnIndex;

            int rangeEndRow = range.ToIndex.RowIndex;
            int rangeEndColumn = range.ToIndex.ColumnIndex;

            lastRowLastTitleCellColumnIndex = rangeStartColumn - 1;
            lastColumnLastTitleCellRowIndex = rangeStartRow - 1;

            for (int row = rangeEndRow; row >= rangeStartRow; row--)
            {
                if (CellIsNonData(worksheet, row, rangeEndColumn, datesAreConsideredTitles))
                {
                    lastColumnLastTitleCellRowIndex = row;
                    break;
                }
            }

            for (int col = rangeEndColumn; col >= rangeStartColumn; col--)
            {
                if (CellIsNonData(worksheet, rangeEndRow, col, datesAreConsideredTitles))
                {
                    lastRowLastTitleCellColumnIndex = col;
                    break;
                }
            }
        }

        private static bool CellIsNonData(Worksheet worksheet, int rowIndex, int columnIndex, bool datesAreConsideredTitles)
        {
            ICellValue lastColumnCurrentValue = worksheet.Cells[rowIndex, columnIndex].GetValue().Value;

            if (lastColumnCurrentValue.ValueType == CellValueType.Formula)
            {
                lastColumnCurrentValue = (lastColumnCurrentValue as FormulaCellValue).GetResultValueAsCellValue();
            }

            bool result = lastColumnCurrentValue is TextCellValue ||
                (datesAreConsideredTitles && SelectionContainsDates(worksheet.Cells[rowIndex, columnIndex]));

            return result;
        }

        private static bool SelectionContainsDates(CellSelection selection)
        {
            ICellValue value = selection.GetValue().Value;

            if (value.ValueType == CellValueType.Formula)
            {
                value = (value as FormulaCellValue).GetResultValueAsCellValue();
            }

            return value.ValueType == CellValueType.Number &&
                selection.GetFormat().Value.FormatStringInfo.Category == FormatStringCategory.Date;
        }

        private static CellRange ConstructDataRange(CellRange range, int lastRowLastTextCellColumnIndex, int lastColumnLastTextCellRowIndex)
        {
            int dataRangeColumnFromIndex = Math.Min(lastRowLastTextCellColumnIndex + 1, range.ToIndex.ColumnIndex);
            int dataRangeRowFromIndex = Math.Min(lastColumnLastTextCellRowIndex + 1, range.ToIndex.RowIndex);

            CellRange dataRange = new CellRange(dataRangeRowFromIndex, dataRangeColumnFromIndex, range.ToIndex.RowIndex, range.ToIndex.ColumnIndex);
            return dataRange;
        }

        private static bool DetermineWhetherSeriesAreVertical(CellRange dataRange)
        {
            return dataRange.RowCount > dataRange.ColumnCount;
        }

        private static void GetCategoriesAndSeriesTitlesRangesFromKnownDataRange(CellRange range, CellRange dataRange, out CellRange categoriesRange, out CellRange seriesTitlesRange, bool seriesAreVertical)
        {
            CellRange topRange, leftRange;
            GetTopAndLeftRanges(range, dataRange, out topRange, out leftRange);
            DetermineCategoriesAndSeriesTitlesRangesPosition(topRange, leftRange, seriesAreVertical, out categoriesRange, out seriesTitlesRange);
        }

        private static void GetTopAndLeftRanges(CellRange range, CellRange dataRange, out CellRange topRange, out CellRange leftRange)
        {
            int topRangeRowCount = dataRange.FromIndex.RowIndex - range.FromIndex.RowIndex;
            int leftRangeColumnCount = dataRange.FromIndex.ColumnIndex - range.FromIndex.ColumnIndex;

            if (topRangeRowCount == 0)
            {
                topRange = null;
            }
            else
            {
                int topRangeFromIndexColumn = leftRangeColumnCount == 0 ? range.FromIndex.ColumnIndex : range.FromIndex.ColumnIndex + leftRangeColumnCount;
                topRange = new CellRange(range.FromIndex.RowIndex, topRangeFromIndexColumn, range.FromIndex.RowIndex + topRangeRowCount - 1, range.ToIndex.ColumnIndex);
            }

            if (leftRangeColumnCount == 0)
            {
                leftRange = null;
            }
            else
            {
                int leftRangeFromIndexRow = topRangeRowCount == 0 ? range.FromIndex.RowIndex : range.FromIndex.RowIndex + topRangeRowCount;
                leftRange = new CellRange(leftRangeFromIndexRow, range.FromIndex.ColumnIndex, range.ToIndex.RowIndex, range.FromIndex.ColumnIndex + leftRangeColumnCount - 1);
            }
        }

        private static void DetermineCategoriesAndSeriesTitlesRangesPosition(CellRange topRange, CellRange leftRange, bool seriesAreVertical, out CellRange categoriesRange, out CellRange seriesTitlesRange)
        {
            if (seriesAreVertical)
            {
                seriesTitlesRange = topRange;
                categoriesRange = leftRange;
            }
            else
            {
                seriesTitlesRange = leftRange;
                categoriesRange = topRange;
            }
        }
    }
}
