﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Model.Drawing.Charts;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Charts
{
    /// <summary>
    /// Represents workbook formula chart data. This type of chart data holds a formula which refers to the actual data in a workbook.
    /// </summary>
    public class WorkbookFormulaChartData : FormulaChartData
    {
        private readonly Workbook workbook;

        /// <summary>
        /// Initializes a new instance of the <see cref="WorkbookFormulaChartData" /> class.
        /// </summary>
        /// <param name="worksheet">The worksheet the ranges refer to.</param>
        /// <param name="ranges">The ranges that refer to the chart data.</param>
        public WorkbookFormulaChartData(Worksheet worksheet, params CellRange[] ranges)
            : base(NameConverter.ConvertCellRangesToName(worksheet, ranges, true))
        {
            this.workbook = worksheet.Workbook;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="WorkbookFormulaChartData" /> class.
        /// </summary>
        /// <param name="workbook">The workbook the formula refer to.</param>
        /// <param name="formula">The formula which refers to the chart data.</param>
        public WorkbookFormulaChartData(Workbook workbook, string formula)
            : base(formula)
        {
            this.workbook = workbook;
        }

        /// <summary>
        /// Gets the workbook that the formula refers to get the chart data.
        /// </summary>
        public Workbook Workbook
        {
            get
            {
                return this.workbook;
            }
        }

        /// <summary>
        /// Creates a deep copy of the object.
        /// </summary>
        /// <returns>The deep copy of the object.</returns>
        public override IChartData Clone()
        {
            WorkbookFormulaChartData newChartData = new WorkbookFormulaChartData(this.workbook, this.Formula);
            return newChartData;
        }

        /// <summary>
        /// Enumerates the cell ranges that contain the actual chart data.
        /// </summary>
        /// <returns>The result cell ranges.</returns>
        /// <exception cref="FormatException">Thrown when the formula referring to the data cannot be parsed to cell ranges.</exception>
        public IEnumerable<CellRange> EnumerateCellRanges(out Worksheet worksheet)
        {
            IEnumerable<CellRange> resultCellRanges;
            if (this.TryEnumerateCellRanges(out resultCellRanges, out worksheet))
            {
                return resultCellRanges;
            }

            throw new FormatException("The formula cannot be parsed to cell ranges.");
        }

        /// <summary>
        /// Enumerates the cell ranges that contain the actual chart data. Returns a value indicating whether the enumeration was successful.
        /// </summary>
        /// <param name="resultCellRanges">The result cell ranges.</param>
        /// <param name="worksheet">The worksheet the cell ranges are located in.</param>
        /// <returns>A value indicating whether the ranges have been retrieved.</returns>
        public bool TryEnumerateCellRanges(out IEnumerable<CellRange> resultCellRanges, out Worksheet worksheet)
        {
            resultCellRanges = Enumerable.Empty<CellRange>();
            bool result = TryGetCellRangesFromFormula(this.Formula, this.workbook, out resultCellRanges, out worksheet);

            return result;
        }

        private static bool TryGetCellRangesFromFormula(string formula, Workbook workbook, out IEnumerable<CellRange> resultCellRanges, out Worksheet worksheet)
        {
            RadExpression result;
            InputStringCollection stringExpressionCollection;
            RadExpression.TryParse(SpreadsheetCultureHelper.FunctionStart + formula, workbook.Worksheets.First(), 0, 0, out result, out stringExpressionCollection);

            RadExpression expression = result.GetValueAsConstantOrCellReference();
            CellReferenceRangeExpression cellReference = expression as CellReferenceRangeExpression;

            resultCellRanges = Enumerable.Empty<CellRange>();
            worksheet = null;

            if (cellReference != null && cellReference.IsValid)
            {
                worksheet = cellReference.Worksheet;

                List<CellRange> cellRangesList = new List<CellRange>();

                foreach (CellReferenceRange cellReferenceRange in cellReference.CellReferenceRanges)
                {
                    cellRangesList.Add(cellReferenceRange.ToCellRange());
                }

                resultCellRanges = cellRangesList;

                return true;
            }

            return false;
        }
    }
}
