﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Layout;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a column selection.
    /// </summary>
    public class ColumnSelection : RowColumnSelectionBase
    {
        #region Properties

        internal override RowColumnPropertyBagBase PropertyBag
        {
            get
            {
                return this.Worksheet.Columns.PropertyBag;
            }
        }

        internal override WorksheetEntityBase WorksheetEntity
        {
            get
            {
                return this.Worksheet.Columns;
            }
        }

        #endregion


        #region Constructors

        internal ColumnSelection(Worksheet worksheet, IEnumerable<CellRange> cellRanges)
            : base(worksheet, cellRanges)
        {
        }

        #endregion


        #region Methods

        /// <summary>
        /// Sets the borders.
        /// </summary>
        /// <param name="value">The value.</param>
        public override void SetBorders(CellBorders value)
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate(
                (cellRange) =>
                {
                    var context = new SetColumnBordersCommandContext(this.Worksheet, cellRange, value);
                    this.Worksheet.ExecuteCommand(WorkbookCommands.SetColumnBorders, context);
                });
        }

        /// <summary>
        /// Gets from to index from range.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <param name="fromIndex">From index.</param>
        /// <param name="toIndex">To index.</param>
        protected override void GetFromToIndexFromRange(CellRange cellRange, out int fromIndex, out int toIndex)
        {
            fromIndex = cellRange.FromIndex.ColumnIndex;
            toIndex = cellRange.ToIndex.ColumnIndex;
        }

        internal override void PreserveOldRowColumnPropertyAsLocalIfNeeded<T>(CellRange cellRange, IProperty<T> property, T value)
        {
            if (property.PropertyDefinition != CellPropertyDefinitions.StyleNameProperty)
            {
                IPropertyDefinition<T> propertyDefinition = property.PropertyDefinition as IPropertyDefinition<T>;
                this.PreserveOldRowColumnPropertyAsLocalIfNeededInternal(cellRange, propertyDefinition, value);
            }
            else
            {
                string styleName = value as string;
                var style = this.Worksheet.Workbook.Styles.GetByName(styleName);

                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.BottomBorderProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.BottomBorderProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.DiagonalDownBorderProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.DiagonalUpBorderProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.FillProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.FontFamilyProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.FontSizeProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.ForeColorProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.FormatProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.HorizontalAlignmentProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.IndentProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.IsBoldProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.IsItalicProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.IsLockedProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.IsWrappedProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.LeftBorderProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.RightBorderProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.StyleNameProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.TopBorderProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.UnderlineProperty);
                this.PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded(cellRange, style, CellPropertyDefinitions.VerticalAlignmentProperty);
            }
        }

        internal override void PreserveAsLocalPropertyIfStyleIsApplied<T>(CellRange cellRange, IProperty<T> property, T value)
        {
            IPropertyDefinition<T> propertyDefinition = property.PropertyDefinition as IPropertyDefinition<T>;
            IProperty<T> cellProperty;
            if (this.Worksheet.Cells.TryGetProperyFromPropertyDefinition(propertyDefinition, out cellProperty))
            {
                CellsPropertyBag cellsPropertyBag = this.Worksheet.Cells.PropertyBag;
                RowsPropertyBag rowsPropertyBag = this.Worksheet.Rows.PropertyBag;

                ICompressedList<string> cellsStylePropertyCollection = cellsPropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.StyleNameProperty);

                ICompressedList<string> rowsStylePropertyCollection = rowsPropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.StyleNameProperty);

                foreach (LongRange longRange in CellsPropertyBag.ConvertCellRangeToLongRanges(cellRange))
                {
                    ICompressedList<string> cellsStyleValues = cellsStylePropertyCollection.GetValue(longRange.Start, longRange.End);
                    Range<long, string>[] nonDefaultCellStyleValueRanges = cellsStyleValues.GetNonDefaultRanges().ToArray();

                    foreach (Range<long, string> range in nonDefaultCellStyleValueRanges)
                    {
                        CellStyle cellStyle = this.Worksheet.Workbook.Styles.GetByName(range.Value);
                        if (cellStyle.GetIsPropertyIncluded(propertyDefinition))
                        {
                            CellRange newCellRange = CellsPropertyBag.ConvertLongCellRangeToCellRange(range.Start, range.End);
                            cellProperty.SetValue(newCellRange, value);
                        }
                    }
                }

                ICompressedList<string> rowStyleValues = rowsStylePropertyCollection.GetValue(cellRange.FromIndex.RowIndex, cellRange.ToIndex.RowIndex);
                Range<long, string>[] nonDefaultRowStyleValues = rowStyleValues.GetNonDefaultRanges().ToArray();
                foreach (Range<long, string> range in nonDefaultRowStyleValues)
                {
                    CellStyle cellStyle = this.Worksheet.Workbook.Styles.GetByName(range.Value);
                    if (cellStyle.GetIsPropertyIncluded(propertyDefinition))
                    {
                        for (int columnIndex = cellRange.FromIndex.ColumnIndex; columnIndex <= cellRange.ToIndex.ColumnIndex; columnIndex++)
                        {

                            CellRange newCellRange = new CellRange((int)range.Start, columnIndex, (int)range.End, columnIndex);
                            cellProperty.SetValue(newCellRange, value);
                        }
                    }
                }
            }
        }

        private void PreserveOldRowColumnPropertyAsLocalFromStyleIfNeeded<T>(CellRange cellRange, CellStyle style, IPropertyDefinition<T> propertyDefinition)
        {
            if (style.GetIsPropertyIncluded(propertyDefinition))
            {
                T styleValue = style.GetPropertyValue(propertyDefinition);
                this.PreserveOldRowColumnPropertyAsLocalIfNeededInternal<T>(cellRange, propertyDefinition, styleValue);
            }
        }

        private void PreserveOldRowColumnPropertyAsLocalIfNeededInternal<T>(CellRange cellRange, IPropertyDefinition<T> propertyDefinition, T value)
        {
            if (!this.Worksheet.Rows.ContainsPropertyDefinition(propertyDefinition))
            {
                return;
            }

            ICompressedList<T> rowPropertyValueCollection = this.Worksheet.Rows.PropertyBag.GetPropertyValueRespectingStyle(propertyDefinition, this.Worksheet, (long)0, (long)SpreadsheetDefaultValues.RowCount - 1);
            var nonDefaultRanges = rowPropertyValueCollection.GetNonDefaultRanges().ToArray();
            foreach (Range<long, T> range in nonDefaultRanges)
            {
                CellRange rowRage = new CellRange((int)range.Start, 0, (int)range.End, SpreadsheetDefaultValues.ColumnCount - 1);
                if (cellRange.IntersectsWith(rowRage))
                {
                    CellRange intersectionRange = cellRange.Intersect(rowRage);
                    IProperty<T> cellProperty;
                    if (this.Worksheet.Cells.TryGetProperyFromPropertyDefinition<T>(propertyDefinition, out cellProperty))
                    {
                        this.Worksheet.Cells[intersectionRange].SetPropertyValue(cellProperty, value);
                    }
                }
            }
        }

        #endregion


        #region Width Property

        /// <summary>
        /// Gets the width of the selection.
        /// </summary>
        /// <returns>The width for the selection.</returns>
        public RangePropertyValue<ColumnWidth> GetWidth()
        {
            return this.GetPropertyValue(this.Worksheet.Columns.WidthProperty);
        }

        /// <summary>
        /// Sets the width.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetWidth(ColumnWidth value)
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate((cellRange) =>
            {
                this.SetWidth(cellRange.FromIndex.ColumnIndex, cellRange.ToIndex.ColumnIndex, value);
            });
        }

        internal void SetWidth(ColumnWidth value, bool considerHidden)
        {
            if (!considerHidden)
            {
                this.SetWidth(value);
            }

            if (value.Value == 0)
            {
                this.SetHidden(true);
            }
            else
            {
                using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
                {
                    this.ClearHidden();
                    this.SetWidth(value);
                }
            }
        }

        private void SetWidth(int fromIndex, int toIndex, ColumnWidth value)
        {
            var context = new SetRowColumnPropertyCommandContext<ColumnWidth>(
                    this.Worksheet, ColumnsPropertyBag.WidthProperty,
                    fromIndex, toIndex, value);

            this.Worksheet.ExecuteCommand(WorkbookCommands.SetColumnWidth, context);
        }

        /// <summary>
        /// Clears the width.
        /// </summary>
        public void ClearWidth()
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate((cellRange) =>
            {
                this.ClearWidth(cellRange.FromIndex.ColumnIndex, cellRange.ToIndex.ColumnIndex);
            });
        }

        private void ClearWidth(int fromIndex, int toIndex)
        {
            var context = new ClearRowColumnPropertyCommandContext<ColumnWidth>(
                    this.Worksheet, ColumnsPropertyBag.WidthProperty, fromIndex, toIndex);

            this.Worksheet.ExecuteCommand(WorkbookCommands.SetColumnWidth, context);
        }

        /// <summary>
        /// Automatically adjusts the width.
        /// </summary>
        public void AutoFitWidth()
        {
            this.AutoFitWidth(false);
        }

        /// <summary>
        /// Expands the width of to fit the width of the number value.
        /// </summary>
        public void ExpandToFitNumberValuesWidth()
        {
            if (!this.Worksheet.IsLayoutUpdateSuspended)
            {
                this.AutoFitWidth(true);
            }
            else
            {
                this.Worksheet.CacheExpandToFitColumns(this.CellRanges);
            }
        }

        internal void AutoFitWidth(bool expandOnly, bool considerHidden = false)
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate(
                (cellRange) =>
                {
                    ColumnWidth[] columnWidths = new ColumnWidth[cellRange.ColumnCount];
                    List<Action> actions = new List<Action>();

                    for (int columnIndex = cellRange.FromIndex.ColumnIndex; columnIndex <= cellRange.ToIndex.ColumnIndex; columnIndex++)
                    {
                        int currentColumnIndex = columnIndex;
                        actions.Add(() =>
                            {
                                ColumnWidth autoColumnWidth;
                                if (expandOnly)
                                {
                                    ColumnWidth currentWidth = this.Worksheet.Columns[currentColumnIndex].GetWidth().Value;
                                    if (currentWidth.IsCustom)
                                    {
                                        return;
                                    }

                                    autoColumnWidth = LayoutHelper.CalculateAutoColumnWidth(
                                        this.Worksheet, currentColumnIndex, cellRange.FromIndex.RowIndex, cellRange.ToIndex.RowIndex, true);
                                    if (autoColumnWidth == null ||
                                        currentWidth.Value > autoColumnWidth.Value)
                                    {
                                        return;
                                    }
                                }
                                else
                                {
                                    autoColumnWidth = LayoutHelper.CalculateAutoColumnWidth(this.Worksheet, currentColumnIndex);
                                }

                                columnWidths[currentColumnIndex - cellRange.FromIndex.ColumnIndex] = autoColumnWidth;//.Add(currentColumnIndex, autoColumnWidth);
                            });
                    }

#if WPF
                    if (actions.Count > 1)
                    {
                        TasksHelper.DoAsync(actions);
                    }
                    else
                    {
#endif
                        for (int i = 0; i < actions.Count; i++)
                        {
                            actions[i]();
                        }
#if WPF
                    }
#endif

                    for (int columnIndex = cellRange.FromIndex.ColumnIndex; columnIndex <= cellRange.ToIndex.ColumnIndex; columnIndex++)
                    {
                        if (columnWidths[columnIndex - cellRange.FromIndex.ColumnIndex] != null)
                        {
                            if (considerHidden)
                            {
                                this.ClearHidden(columnIndex, columnIndex);
                            }

                            this.SetWidth(columnIndex, columnIndex, columnWidths[columnIndex - cellRange.FromIndex.ColumnIndex]);
                        }
                    }
                });
        }

        #endregion


        #region Outline Level Property

        internal override void SetOutlineLevelOnRange(CellRange cellRange, int value)
        {
            var context = new SetRowColumnPropertyCommandContext<int>(
                this.Worksheet, ColumnsPropertyBag.OutlineLevelProperty,
                cellRange.FromIndex.ColumnIndex, cellRange.ToIndex.ColumnIndex, value);

            this.Worksheet.ExecuteCommand(WorkbookCommands.SetColumnOutlineLevel, context);
        }

        /// <summary>
        /// Gets the outline level.
        /// </summary>
        /// <returns>The outline level.</returns>
        public override RangePropertyValue<int> GetOutlineLevel()
        {
            return this.GetPropertyValue(ColumnsPropertyBag.OutlineLevelProperty);
        }

        internal override void ClearOutlineLevelOnRange(CellRange cellRange)
        {
            var context = new ClearRowColumnPropertyCommandContext<int>(
                this.Worksheet, ColumnsPropertyBag.OutlineLevelProperty, cellRange.FromIndex.ColumnIndex, cellRange.ToIndex.ColumnIndex);

            this.Worksheet.ExecuteCommand(WorkbookCommands.SetColumnOutlineLevel, context);
        }

        internal override RowColumnSelectionBase GetRowColumnSelection(int fromIndex, int toIndex)
        {
            return this.Worksheet.Columns[fromIndex, toIndex];
        }

        #endregion


        #region Hidden Property

        /// <summary>
        /// Sets the hidden property.
        /// </summary>
        /// <param name="value">The value indicating whether the selected columns are hidden.</param>
        public void SetHidden(bool value)
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate((cellRange) =>
            {
                var context = new SetRowColumnPropertyCommandContext<bool>(
                    this.Worksheet, ColumnsPropertyBag.HiddenProperty,
                    cellRange.FromIndex.ColumnIndex, cellRange.ToIndex.ColumnIndex, value);

                this.Worksheet.ExecuteCommand(WorkbookCommands.SetColumnHidden, context);
            });
        }

        /// <summary>
        /// Gets the value of the hidden property.
        /// </summary>
        /// <returns>The value of the hidden property</returns>
        public RangePropertyValue<bool> GetHidden()
        {
            return this.GetPropertyValue(ColumnsPropertyBag.HiddenProperty);
        }

        /// <summary>
        /// Clears the hidden property.
        /// </summary>
        public void ClearHidden()
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate((cellRange) =>
            {
                this.ClearHidden(cellRange.FromIndex.ColumnIndex, cellRange.ToIndex.ColumnIndex);
            });
        }

        private void ClearHidden(int fromIndex, int toIndex)
        {
            var context = new ClearRowColumnPropertyCommandContext<bool>(
                this.Worksheet, ColumnsPropertyBag.HiddenProperty, fromIndex, toIndex);

            this.Worksheet.ExecuteCommand(WorkbookCommands.SetColumnHidden, context);
        }

        #endregion


        #region Insert/Remove

        /// <summary>
        /// Inserts this instance in the workbook.
        /// </summary>
        /// <returns>The value indicating whether the insert was successful.</returns>
        public bool Insert()
        {
            var sortedSelectedRanges =
                from r in this.CellRanges
                orderby r.FromIndex.ColumnIndex descending
                select r;

            return this.ExecuteForEachRangeInsideBeginEndUpdate(
                sortedSelectedRanges,
                (cellRange) =>
                {
                    this.Worksheet.Columns.Insert(cellRange.FromIndex.ColumnIndex, cellRange.ColumnCount);
                },
                (cellRange) => { return this.Worksheet.Columns.CanInsert(cellRange.FromIndex.ColumnIndex, cellRange.ColumnCount); });
        }

        /// <summary>
        /// Removes this instance from the workbook.
        /// </summary>
        public void Remove()
        {
            ICompressedList<bool> columnsToRemove = new CompressedList<bool>(0, SpreadsheetDefaultValues.ColumnCount - 1);
            foreach (var selectedRange in this.CellRanges)
            {
                columnsToRemove.SetValue(selectedRange.FromIndex.ColumnIndex, selectedRange.ToIndex.ColumnIndex, true);
            }

            List<CellRange> cellRanges = new List<CellRange>();
            foreach (var range in columnsToRemove)
            {
                if (range.Value)
                {
                    cellRanges.Add(new CellRange(0, (int)range.Start, 0, (int)range.End));
                }
            }
            cellRanges.Reverse();

            this.ExecuteForEachRangeInsideBeginEndUpdate(
                cellRanges,
                (cellRange) =>
                {
                    this.Worksheet.Columns.Remove(cellRange.FromIndex.ColumnIndex, cellRange.ColumnCount);
                });
        }

        #endregion

    }
}
