﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a collection of columns.
    /// </summary>
    public class Columns : RowsColumnsBase
    {
        #region Fields

        private readonly ColumnsPropertyBag propertyBag;
        private readonly IProperty<ColumnWidth> widthProperty;

        #endregion


        #region Properties

        internal override PropertyBagBase PropertyBagBase
        {
            get
            {
                return this.PropertyBag;
            }
        }

        /// <summary>
        /// Gets the row / column count.
        /// </summary>
        /// <value>The row / column count.</value>
        public override int Count
        {
            get
            {
                return this.Worksheet.Cells.ColumnCount;
            }
        }

        internal ColumnsPropertyBag PropertyBag
        {
            get
            {
                return this.propertyBag;
            }
        }

        internal IProperty<ColumnWidth> WidthProperty
        {
            get
            {
                return this.widthProperty;
            }
        }

        #endregion


        #region Indexers

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.ColumnSelection" /> at the specified index.
        /// </summary>
        /// <value>The column selection.</value>
        public ColumnSelection this[int index]
        {
            get
            {
                return this.GetColumnSelection(index);
            }
        }

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.ColumnSelection" /> with the specified from index and to index.
        /// </summary>
        /// <value>The column selection.</value>
        public ColumnSelection this[int fromIndex, int toIndex]
        {
            get
            {
                return this.GetColumnSelection(fromIndex, toIndex);
            }
        }

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.ColumnSelection" /> with the specified cell index.
        /// </summary>
        /// <value>The column selection.</value>
        public ColumnSelection this[CellIndex cellIndex]
        {
            get
            {
                return this.GetColumnSelection(cellIndex);
            }
        }

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.ColumnSelection" /> with the specified cell range.
        /// </summary>
        /// <value>The column selection.</value>
        public ColumnSelection this[CellRange cellRange]
        {
            get
            {
                return this.GetColumnSelection(cellRange);
            }
        }

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.ColumnSelection" /> with the specified cell ranges.
        /// </summary>
        /// <value>The column selection.</value>
        public ColumnSelection this[IEnumerable<CellRange> cellRanges]
        {
            get
            {
                return this.GetColumnSelection(cellRanges);
            }
        }

        #endregion


        #region Constructors

        internal Columns(Worksheet worksheet)
            : base(worksheet)
        {
            this.propertyBag = new ColumnsPropertyBag();
            this.PropertyBag.PropertyChanged += this.PropertyBag_PropertyChanged;

            this.widthProperty = this.CreateProperty(ColumnsPropertyBag.WidthProperty);
        }

        #endregion


        #region Event Handlers

        private void PropertyBag_PropertyChanged(object sender, RowColumnPropertyChangedEventArgs e)
        {
            if (e.Property == ColumnsPropertyBag.WidthProperty)
            {
                this.OnColumnsWidthChanged(e);
            }
        }

        #endregion


        #region Methods

        /// <summary>
        /// Determines whether this instance can be inserted at the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <param name="itemCount">The item count.</param>
        /// <returns>
        /// The value indicating whether this instance can be inserted at the specified index
        /// </returns>
        public override bool CanInsert(int index, int itemCount)
        {
            return this.PropertyBag.CanInsert() &&
                this.Cells.PropertyBag.CanInsertColumn(itemCount);
        }

        /// <summary>
        /// Inserts the override.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <param name="itemCount">The item count.</param>
        /// <returns></returns>
        protected override bool InsertOverride(int index, int itemCount)
        {
            var context = new InsertColumnCommandContext(this.Worksheet, index, itemCount);
            bool isSuccessful = this.Worksheet.ExecuteCommand(WorkbookCommands.InsertColumn, context);

            CellRange range = CellRange.FromColumnRange(index, index + itemCount - 1);
            this.Cells.UpdateCellRangeInsertedOrRemovedDependentCollections(range, RangeType.Columns, false);

            return isSuccessful;
        }

        /// <summary>
        /// Removes the override.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <param name="itemCount">The item count.</param>
        protected override void RemoveOverride(int index, int itemCount)
        {
            var context = new RemoveColumnCommandContext(this.Worksheet, index, itemCount);
            this.Worksheet.ExecuteCommand(WorkbookCommands.RemoveColumn, context);

            CellRange range = CellRange.FromColumnRange(index, index + itemCount - 1);
            this.Cells.UpdateCellRangeInsertedOrRemovedDependentCollections(range, RangeType.Columns, true);
        }

        internal bool GetIsHidden(int columnIndex)
        {
            return this.PropertyBag.GetPropertyValue(ColumnsPropertyBag.HiddenProperty, columnIndex);
        }

        internal override IProperty<T> CreatePropertyOverride<T>(IPropertyDefinition<T> propertyDefinition)
        {
            return new ColumnProperty<T>(this.Worksheet, propertyDefinition);
        }

        #endregion


        #region GetSelection Methods

        /// <summary>
        /// Gets a column selection at the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <returns>The column selection.</returns>
        public ColumnSelection GetColumnSelection(int index)
        {
            return this.GetColumnSelection(index, index);
        }

        /// <summary>
        /// Gets a column selection between two indices.
        /// </summary>
        /// <param name="fromIndex">From index.</param>
        /// <param name="toIndex">To index.</param>
        /// <returns>The column selection.</returns>
        public ColumnSelection GetColumnSelection(int fromIndex, int toIndex)
        {
            return this.GetColumnSelection(CellRange.FromColumnRange(fromIndex, toIndex));
        }

        /// <summary>
        /// Gets a column selection at the specified index.
        /// </summary>
        /// <param name="cellIndex">Index of the cell.</param>
        /// <returns>The column selection.</returns>
        public ColumnSelection GetColumnSelection(CellIndex cellIndex)
        {
            return this.GetColumnSelection(new CellRange(cellIndex, cellIndex));
        }

        /// <summary>
        /// Gets a column selection of a specified cell range.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <returns>The column selection.</returns>
        public ColumnSelection GetColumnSelection(CellRange cellRange)
        {
            return this.GetColumnSelection(new CellRange[] { cellRange });
        }

        /// <summary>
        /// Gets a column selection of a specified collection of cell ranges.
        /// </summary>
        /// <param name="cellRanges">The cell ranges.</param>
        /// <returns>The column selection.</returns>
        public ColumnSelection GetColumnSelection(IEnumerable<CellRange> cellRanges)
        {
            return new ColumnSelection(this.Worksheet, cellRanges);
        }

        #endregion


        #region Column Properties

        /// <summary>
        /// Gets the default width.
        /// </summary>
        /// <returns>The column width</returns>
        public ColumnWidth GetDefaultWidth()
        {
            return this.PropertyBag.GetDefaultPropertyValue(ColumnsPropertyBag.WidthProperty);
        }

        /// <summary>
        /// Sets the default width.
        /// </summary>
        /// <param name="width">The width.</param>
        public void SetDefaultWidth(ColumnWidth width)
        {
            var context = new SetDefaultPropertyValueCommandContext<ColumnWidth>(
                this.Worksheet, ColumnsPropertyBag.WidthProperty, this.PropertyBag, width);

            this.Worksheet.ExecuteCommand(WorkbookCommands.SetDefaultColumnWidth, context);
        }

        #endregion


        #region Events

        /// <summary>
        /// Occurs when the columns' widths have changed.
        /// </summary>
        public event EventHandler<RowColumnPropertyChangedEventArgs> ColumnsWidthChanged;

        private void OnColumnsWidthChanged(RowColumnPropertyChangedEventArgs args)
        {
            if (this.ColumnsWidthChanged != null)
            {
                this.ColumnsWidthChanged(this, args);
            }
        }

        #endregion
    }
}