﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Model.Shapes;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    internal class CopyPasteContext
    {
        private readonly Worksheet fromWorksheet;
        private readonly List<CellRange> fromCellRange;
        private readonly Worksheet toWorksheet;
        private readonly CellRange toCellRange;
        private readonly IEnumerable<FloatingShapeBase> fromShapes;
        private readonly IEnumerable<FloatingShapeBase> toShapes;
        private readonly PasteOptions pasteOptions;
        private readonly bool isCopying;
        private readonly bool shouldPasteAsText;
        private readonly bool isCellCopyPaste;
        private readonly int[] hiddenRows;
        private readonly int[] hiddenColumns;
        List<CellRange> affectedCellRanges;

        public Worksheet FromWorksheet
        {
            get
            {
                return this.fromWorksheet;
            }
        }

        public List<CellRange> FromCellRange
        {
            get
            {
                return this.fromCellRange;
            }
        }

        public Worksheet ToWorksheet
        {
            get
            {
                return this.toWorksheet;
            }
        }

        public CellRange ToCellRange
        {
            get
            {
                return this.toCellRange;
            }
        }

        public IEnumerable<FloatingShapeBase> FromShapes
        {
            get
            {
                return this.fromShapes;
            }
        }

        public IEnumerable<FloatingShapeBase> ToShapes
        {
            get
            {
                return this.toShapes;
            }
        }

        public PasteOptions PasteOptions
        {
            get
            {
                return this.pasteOptions;
            }
        }


        public bool IsCopying
        {
            get
            {
                return this.isCopying;
            }
        }

        public bool ShouldPasteAsText
        {
            get
            {
                return this.shouldPasteAsText;
            }
        }

        public bool IsCellCopyPaste
        {
            get
            {
                return this.isCellCopyPaste;
            }
        }

        public List<CellRange> AffectedCellRanges
        {
            get 
            {
                return this.affectedCellRanges;
            }
            set
            {
                Guard.ThrowExceptionIfNotNull(this.affectedCellRanges, "affectedCellRanges");
                this.affectedCellRanges = value;
            }
        }

        public int[] HiddenRows
        {
            get
            {
                return this.hiddenRows;
            }
        }

        public int[] HiddenColumns
        {
            get
            {
                return this.hiddenColumns;
            }
        }

        public CopyPasteContext(Worksheet fromWorksheet, CellRange fromCellRange, Worksheet toWorksheet, CellRange toCellRange, IEnumerable<FloatingShapeBase> fromShapes,
            IEnumerable<FloatingShapeBase> toShapes, PasteOptions pasteOptions, bool isCellCopyPaste, bool isCopying, bool shouldPasteAsText)
        {
            if (isCellCopyPaste)
            {
                Guard.ThrowExceptionIfNull(fromWorksheet, "fromWorksheet");
                Guard.ThrowExceptionIfNull(fromCellRange, "fromCellRange");
                Guard.ThrowExceptionIfNull(toWorksheet, "toWorksheet");
                Guard.ThrowExceptionIfNull(toCellRange, "toCellRange");
            }
            else
            {
                Guard.ThrowExceptionIfNull(fromShapes, "fromShapes");
                Guard.ThrowExceptionIfNull(fromWorksheet, "fromWorksheet");
                Guard.ThrowExceptionIfNull(toWorksheet, "toWorksheet");
            }

            this.fromWorksheet = fromWorksheet;
            this.fromCellRange = GetFromCellRange(this.fromWorksheet, fromCellRange, out this.hiddenRows, out this.hiddenColumns);
            this.toWorksheet = toWorksheet;
            this.toCellRange = toCellRange;
            this.fromShapes = fromShapes;
            this.toShapes = toShapes;
            this.isCopying = isCopying;
            this.shouldPasteAsText = shouldPasteAsText;
            this.isCellCopyPaste = isCellCopyPaste;

            this.pasteOptions = GetPasteOptions(pasteOptions, shouldPasteAsText);
        }

        public static List<CellRange> GetFromCellRange(Worksheet fromWorksheet, CellRange fromCellRange)
        {
            int[] hiddenRows;
            int[] hiddenColumns;

            return GetFromCellRange(fromWorksheet, fromCellRange, out hiddenRows, out hiddenColumns);
        }

        private static List<CellRange> GetFromCellRange(Worksheet fromWorksheet, CellRange fromCellRange, out int[] hiddenRows, out int[] hiddenColumns)
        {
            List<CellRange> result;
            
            if (fromCellRange == null)
            {
                result = null;
                hiddenRows = null;
                hiddenColumns = null;
            }

            bool copyHiddenRanges = !fromWorksheet.Filter.FilterIsApplied;

            if (copyHiddenRanges)
            {
                result = new List<CellRange>() { fromCellRange };
                hiddenRows = new int[0];
                hiddenColumns = new int[0];
            }
            else
            {
                result = SpreadsheetHelper.SplitIntoRangesAccordingToHidden(fromCellRange, fromWorksheet, out hiddenRows, out hiddenColumns);
            }

            return result;
        }

        private static PasteOptions GetPasteOptions(PasteOptions pasteOptions, bool shouldPasteAsText)
        {
            if (shouldPasteAsText)
            {
                return new PasteOptions(PasteType.Values);
            }

            if (pasteOptions != null)
            {
                return pasteOptions;
            }

            return PasteOptions.All;
        }

        public bool ShouldCopyPasteHyperlinks()
        {
            return this.IsCopying || this.PasteOptions.PasteType == PasteType.All;
        }

        internal bool ShouldCopyPasteColumnWidths()
        {
            return this.PasteOptions.PasteType == PasteType.ColumnWidths || this.IsCopying;
        }
    }
}
