﻿using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.DataValidation
{
    /// <summary>
    /// Represents a base class for implementing a data validation rule.
    /// </summary>
    public abstract class DataValidationRuleBase : IDataValidationRule
    {
        #region Fields

        private readonly bool showInputMessage;
        private readonly string inputMessageTitle;
        private readonly string inputMessageContent;
        private readonly bool showErrorMessage;
        private readonly ErrorStyle errorStyle;
        private readonly string errorAlertTitle;
        private readonly string errorMessageContent;

        #endregion


        #region Properties

        /// <summary>
        /// Gets or sets a value indicating whether an input message should be shown.
        /// </summary>
        /// <value>The value indicating whether an input message should be shown.</value>
        public bool ShowInputMessage
        {
            get
            {
                return this.showInputMessage;
            }
        }

        /// <summary>
        /// Gets or sets the input message title.
        /// </summary>
        /// <value>The input message title.</value>
        public string InputMessageTitle
        {
            get
            {
                return this.inputMessageTitle;
            }
        }

        /// <summary>
        /// Gets or sets the content of the input message.
        /// </summary>
        /// <value>The content of the input message.</value>
        public string InputMessageContent
        {
            get
            {
                return this.inputMessageContent;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether an error message should be shown.
        /// </summary>
        /// <value>The value indicating whether an input message should be shown.</value>
        public bool ShowErrorMessage
        {
            get
            {
                return this.showErrorMessage;
            }
        }

        /// <summary>
        /// Gets or sets the error style.
        /// </summary>
        /// <value>The error style.</value>
        public ErrorStyle ErrorStyle
        {
            get
            {
                return this.errorStyle;
            }
        }

        /// <summary>
        /// Gets the error alert title.
        /// </summary>
        /// <value>The error alert title.</value>
        public string ErrorAlertTitle
        {
            get
            {
                return this.errorAlertTitle;
            }
        }

        /// <summary>
        /// Gets the content of the error alert.
        /// </summary>
        /// <value>The content of the error alert.</value>
        public string ErrorAlertContent
        {
            get
            {
                return this.errorMessageContent;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="DataValidationRuleBase" /> class.
        /// </summary>
        /// <param name="context">The context.</param>
        protected DataValidationRuleBase(DataValidationRuleContextBase context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            this.showInputMessage = context.ShowInputMessage;
            this.inputMessageTitle = context.InputMessageTitle;
            this.inputMessageContent = context.InputMessageContent;

            this.showErrorMessage = context.ShowErrorMessage;
            this.errorStyle = context.ErrorStyle;
            this.errorAlertTitle = context.ErrorAlertTitle;
            this.errorMessageContent = context.ErrorAlertContent;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Evaluates the data validation rule.
        /// </summary>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <param name="cellValue"></param>
        /// <returns>Retrurns True if the rule is satisfied, else returns False.</returns>
        public bool Evaluate(Worksheet worksheet, int rowIndex, int columnIndex, ICellValue cellValue)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfNull(cellValue, "cellValue");

            ICompressedList<ICellValue> cellValues = worksheet.Cells.PropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.ValueProperty);
            long index = CellsPropertyBag.ConvertCellIndexToLong(rowIndex, columnIndex);

            ICellValue currentValue = cellValues.GetValue(index);
            cellValues.SetValue(index, cellValue);

            bool result = false;

            try
            {
                result = this.EvaluateOverride(worksheet, rowIndex, columnIndex, cellValue);
            }
            finally
            {
                cellValues.SetValue(index, currentValue);
            }

            return result;
        }

        /// <summary>
        /// Evaluates the data validation rule.
        /// </summary>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <param name="cellValue">The cell value.</param>
        /// <returns>Retrurns True if the rule is satisfied, else returns False.</returns>
        protected abstract bool EvaluateOverride(Worksheet worksheet, int rowIndex, int columnIndex, ICellValue cellValue);

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            DataValidationRuleBase other = obj as DataValidationRuleBase;

            if (other == null)
            {
                return false;
            }

            return TelerikHelper.EqualsOfT(this.ShowInputMessage, other.ShowInputMessage) &&
                TelerikHelper.EqualsOfT(this.InputMessageTitle, other.InputMessageTitle) &&
                TelerikHelper.EqualsOfT(this.InputMessageContent, other.InputMessageContent) &&
                TelerikHelper.EqualsOfT(this.ShowErrorMessage, other.ShowErrorMessage) &&
                TelerikHelper.EqualsOfT(this.ErrorStyle, other.ErrorStyle) &&
                TelerikHelper.EqualsOfT(this.ErrorAlertTitle, other.ErrorAlertTitle) &&
                TelerikHelper.EqualsOfT(this.ErrorAlertContent, other.ErrorAlertContent) &&
                TelerikHelper.EqualsOfT(this.GetType(), other.GetType());
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                TelerikHelper.CombineHashCodes(
                    this.ShowInputMessage.GetHashCodeOrZero(),
                    this.InputMessageTitle.GetHashCodeOrZero(),
                    this.InputMessageContent.GetHashCodeOrZero(),
                    this.GetType().GetHashCodeOrZero()),
                TelerikHelper.CombineHashCodes(
                    this.ShowErrorMessage.GetHashCodeOrZero(),
                    this.ErrorStyle.GetHashCodeOrZero(),
                    this.ErrorAlertTitle.GetHashCodeOrZero(),
                    this.ErrorAlertContent.GetHashCodeOrZero()));
        }

        #endregion
    }
}