﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.DataValidation
{
    /// <summary>
    /// Represents a base class for implementing number data validation rule.
    /// </summary>
    public abstract class NumberDataValidationRuleBase : SingleArgumentDataValidationRuleBase
    {
        #region Fileds

        private readonly ComparisonOperator comparisonOperator;
        private readonly ICellValue argument2;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the comparison operator.
        /// </summary>
        /// <value>The comparison operator.</value>
        public ComparisonOperator ComparisonOperator
        {
            get
            {
                return this.comparisonOperator;
            }
        }

        /// <summary>
        /// Gets the second argument.
        /// </summary>
        /// <value>The second argument.</value>
        public ICellValue Argument2
        {
            get
            {
                return this.argument2;
            }
        }

        /// <summary>
        /// Gets a value indicating if the rule requires whole numbers as a result of the argument expression.
        /// </summary>
        /// <value>The value indicating if the rule requires whole numbers as a result of the argument expression.</value>
        protected virtual bool RequireWholeNumbers
        {
            get
            {
                return false;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="NumberDataValidationRuleBase" /> class.
        /// </summary>
        /// <param name="context">The context.</param>
        protected NumberDataValidationRuleBase(NumberDataValidationRuleContext context)
            : base(context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            this.comparisonOperator = context.ComparisonOperator;
            ICellValue result = this.ConvertToCellValue(context.Argument2, context.Worksheet, context.CultureInfo);

            this.argument2 = result;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Evaluates the data validation rule.
        /// </summary>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <param name="cellValue">The cell value.</param>
        /// <returns>Retrurns True if the rule is satisfied, else returns False.</returns>
        protected override bool EvaluateOverride(Worksheet worksheet, int rowIndex, int columnIndex, ICellValue cellValue)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfNull(cellValue, "cellValue");
            Guard.ThrowExceptionIfNull(this.Argument1, "Argument1");

            if (this.ShouldIgnore(cellValue))
            {
                return true;
            }

            double? cellValueResult = cellValue.GetAsDoubleOrNull();
            double? argument1Result = base.EvaluateArgument(this.Argument1, worksheet, rowIndex, columnIndex);
            double? argument2Result = (this.Argument2 == null) ? null : base.EvaluateArgument(this.Argument2, worksheet, rowIndex, columnIndex);

            if (!cellValueResult.HasValue)
            {
                return false;
            }

            if (this.RequireWholeNumbers && !this.IsWholeNumber(cellValueResult.Value))
            {
                return false;
            }

            return this.CompareValues(cellValueResult, argument1Result, argument2Result);
        }

        /// <summary>
        /// Compares the values.
        /// </summary>
        /// <param name="cellValueResult">The cell value result.</param>
        /// <param name="argument1">The argument1.</param>
        /// <param name="argument2">The argument2.</param>
        /// <returns>Retrurns True if the rule is satisfied, else returns False.</returns>
        protected bool CompareValues(double? cellValueResult, double? argument1, double? argument2)
        {
            if (!argument1.HasValue)
            {
                return true;
            }

            switch (this.ComparisonOperator)
            {
                case ComparisonOperator.EqualsTo:
                    return cellValueResult.Value == argument1.Value;

                case ComparisonOperator.GreaterThan:
                    return cellValueResult.Value > argument1.Value;

                case ComparisonOperator.GreaterThanOrEqualsTo:
                    return cellValueResult.Value >= argument1.Value;

                case ComparisonOperator.LessThan:
                    return cellValueResult.Value < argument1.Value;

                case ComparisonOperator.LessThanOrEqualsTo:
                    return cellValueResult.Value <= argument1.Value;

                case ComparisonOperator.NotEqualsTo:
                    return cellValueResult.Value != argument1.Value;

                case ComparisonOperator.Between:
                    if (argument2.HasValue)
                    {
                        return argument1.Value <= cellValueResult.Value &&
                                      cellValueResult.Value <= argument2.Value;
                    }

                    return true;

                case ComparisonOperator.NotBetween:
                    if (argument2.HasValue)
                    {
                        return argument1.Value > cellValueResult.Value ||
                                      cellValueResult.Value > argument2.Value;
                    }

                    return true;
            }

            return false;
        }

        private bool IsWholeNumber(double value)
        {
            int intValue = (int)value;
            return value.EqualsDouble(intValue);
        }

        internal override void CloneAndTranslateOverride(DataValidationRuleContextBase context, Worksheet worksheet, int rowIndex, int columnIndex, bool shouldTranslateArguments)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            NumberDataValidationRuleContext numberContext = context as NumberDataValidationRuleContext;
            Guard.ThrowExceptionIfNull(numberContext, "numberContext");

            numberContext.Argument2 = (shouldTranslateArguments) ?
                this.GetExpressionString(this.Argument2, worksheet, rowIndex, columnIndex) :
                this.GetExpressionString(this.Argument2, worksheet, this.CellIndex.RowIndex, this.CellIndex.ColumnIndex);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            NumberDataValidationRuleBase other = obj as NumberDataValidationRuleBase;

            if (other == null)
            {
                return false;
            }

            return base.Equals(obj) &&
                TelerikHelper.EqualsOfT(this.ComparisonOperator, other.ComparisonOperator) &&
                this.SimilarEquals(this.Argument2, other.Argument2);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                base.GetHashCode(),
                this.ComparisonOperator.GetHashCodeOrZero(),
                this.Argument2.GetHashCodeOrZero());
        }

        #endregion
    }
}