﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.DataValidation
{
    /// <summary>
    /// Represents a base class for implementing an single argument data validation rule.
    /// </summary>
    public abstract class SingleArgumentDataValidationRuleBase : DataValidationRuleBase
    {
        #region Fields

        private readonly bool ignoreBlank;
        private readonly ICellValue argument1;
        private readonly CellIndex cellIndex;
        private readonly Worksheet worksheet;
        private readonly SingleArgumentDataValidationRuleContext context;

        #endregion


        #region Properties

        /// <summary>
        /// Gets or sets a value indicating if blank cells should be ignored.
        /// </summary>
        /// <value>The value indicating if blank cells should be ignored.</value>
        public bool IgnoreBlank
        {
            get
            {
                return this.ignoreBlank;
            }
        }

        /// <summary>
        /// Gets the index of the cell in which the Validation rule is created.
        /// </summary>
        /// <value>The index of the cell.</value>
        public CellIndex CellIndex
        {
            get
            {
                return this.cellIndex;
            }
        }

        /// <summary>
        /// Gets the first argument.
        /// </summary>
        /// <value>The first argument.</value>
        public ICellValue Argument1
        {
            get
            {
                return this.argument1;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="SingleArgumentDataValidationRuleBase" /> class.
        /// </summary>
        /// <param name="context">The context.</param>
        protected SingleArgumentDataValidationRuleBase(SingleArgumentDataValidationRuleContext context)
            : base(context)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(context.Worksheet, "Worksheet");
            Guard.ThrowExceptionIfNull(context.CellIndex, "CellIndex");

            this.context = context;
            this.ignoreBlank = context.IgnoreBlank;
            this.worksheet = context.Worksheet;
            this.cellIndex = context.CellIndex;

            this.argument1 = this.ConvertToCellValue(context.Argument1, context.Worksheet, context.CultureInfo);
        }

        #endregion


        #region Methods

        internal ICellValue ConvertToCellValue(string argument, Worksheet worksheet, SpreadsheetCultureHelper cultureInfo)
        {
            ICellValue result;

            if (RadExpression.StartsLikeExpression(argument))
            {
                result = CellValueFactory.Create(argument, worksheet, this.CellIndex.RowIndex, this.CellIndex.ColumnIndex, cultureInfo);
            }
            else
            {
                result = argument.ToCellValue(worksheet, this.CellIndex.RowIndex, this.CellIndex.ColumnIndex, CellValueFormat.GeneralFormat);
            }

            return result;
        }

        internal bool ShouldIgnore(ICellValue cellValue)
        {
            Guard.ThrowExceptionIfNull(cellValue, "cellValue");

            return this.IgnoreBlank && cellValue is EmptyCellValue;
        }

        /// <summary>
        /// Gets the expression string.
        /// </summary>
        /// <param name="argument">The expression result.</param>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns></returns>
        internal string GetExpressionString(ICellValue argument, Worksheet worksheet, int rowIndex, int columnIndex)
        {
            Guard.ThrowExceptionIfNull(argument, "argument");
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            ICellValue translatedArgument = this.GetTranslatedArgument(argument, worksheet, rowIndex, columnIndex);

            return translatedArgument.RawValue;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            SingleArgumentDataValidationRuleBase other = obj as SingleArgumentDataValidationRuleBase;

            if (other == null)
            {
                return false;
            }

            return base.Equals(other) &&
                   TelerikHelper.EqualsOfT(this.IgnoreBlank, other.IgnoreBlank) &&
                   this.SimilarEquals(this.Argument1, other.Argument1);
        }

        internal bool SimilarEquals(ICellValue first, ICellValue second)
        {
            FormulaCellValue firstFormula = first as FormulaCellValue;
            FormulaCellValue secondFormula = second as FormulaCellValue;

            if (firstFormula != null && secondFormula != null)
            {
                first = firstFormula.CloneAndTranslate(this.worksheet, this.cellIndex, true);
                second = secondFormula.CloneAndTranslate(this.worksheet, this.cellIndex, true);
            }

            return TelerikHelper.EqualsOfT(first.RawValue, second.RawValue);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                base.GetHashCode(),
                this.IgnoreBlank.GetHashCodeOrZero(),
                this.argument1.GetHashCodeOrZero());
        }

        /// <summary>
        /// Evaluates the argument.
        /// </summary>
        /// <param name="argument">The argument.</param>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns></returns>
        internal double? EvaluateArgument(ICellValue argument, Worksheet worksheet, int rowIndex, int columnIndex)
        {
            Guard.ThrowExceptionIfNull(argument, "argument");
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            ICellValue argumentCellValue = this.GetTranslatedArgument(argument, worksheet, rowIndex, columnIndex);

            return argumentCellValue.GetAsDoubleOrNull();
        }

        internal ICellValue GetTranslatedArgument(ICellValue argument, Worksheet worksheet, int rowIndex, int columnIndex)
        {
            Guard.ThrowExceptionIfNull(argument, "argument");
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            FormulaCellValue formulaValue = argument as FormulaCellValue;

            if (formulaValue != null)
            {
                argument = formulaValue.CloneAndTranslate(worksheet, rowIndex, columnIndex, true);
            }

            return argument;
        }

        internal IDataValidationRule CloneAndTranslate(Worksheet worksheet, int rowIndex, int columnIndex, bool shouldTranslateArguments)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            context.Worksheet = worksheet;
            context.CellIndex = new CellIndex(rowIndex, columnIndex);
            context.Argument1 = (shouldTranslateArguments) ?
                this.GetExpressionString(this.Argument1, worksheet, rowIndex, columnIndex) :
                this.GetExpressionString(this.Argument1, worksheet, this.CellIndex.RowIndex, this.CellIndex.ColumnIndex);

            this.CloneAndTranslateOverride(context, worksheet, rowIndex, columnIndex, shouldTranslateArguments);

            return (IDataValidationRule)Activator.CreateInstance(this.GetType(), context);
        }

        internal virtual void CloneAndTranslateOverride(DataValidationRuleContextBase context, Worksheet worksheet, int rowIndex, int columnIndex, bool shouldTranslateArguments)
        {
        }

        #endregion
    }
}