﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.DataValidation
{
    /// <summary>
    /// Represents list data validation rule.
    /// </summary>
    public sealed class ListDataValidationRule : SingleArgumentDataValidationRuleBase
    {
        #region Properties

        /// <summary>
        /// Gets or sets a value indicating if an cell dropdown should be shown.
        /// </summary>
        /// <value>The a value indicating if an cell dropdown should be shown.</value>
        public bool InCellDropdown { get; set; }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="ListDataValidationRule" /> class.
        /// </summary>
        /// <param name="context">The context.</param>
        public ListDataValidationRule(ListDataValidationRuleContext context)
            : base(context)
        {
            this.InCellDropdown = context.InCellDropdown;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Evaluates the data validation rule.
        /// </summary>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <param name="cellValue">The cell value.</param>
        /// <returns>Retrurns True if the rule is satisfied, else returns False.</returns>
        protected override bool EvaluateOverride(Worksheet worksheet, int rowIndex, int columnIndex, ICellValue cellValue)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfNull(cellValue, "cellValue");
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            if (this.ShouldIgnore(cellValue) || this.Argument1 is EmptyCellValue)
            {
                return true;
            }

            IEnumerable<string> listItems = this.ExtractListItems(worksheet, rowIndex, columnIndex);

            string currentCellValue = cellValue.GetResultValueAsString(CellValueFormat.GeneralFormat);

            return listItems.Contains(currentCellValue);
        }

        /// <summary>
        /// Extracts the list items.
        /// </summary>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="cellIndex">Index of the cell.</param>
        /// <returns></returns>
        public IEnumerable<string> ExtractListItems(Worksheet worksheet, CellIndex cellIndex)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfNull(cellIndex, "cellIndex");

            return this.ExtractListItems(worksheet, cellIndex.RowIndex, cellIndex.ColumnIndex);
        }

        /// <summary>
        /// Extracts the list items.
        /// </summary>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns></returns>
        public IEnumerable<string> ExtractListItems(Worksheet worksheet, int rowIndex, int columnIndex)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            ICellValue cellValue = this.GetTranslatedArgument(this.Argument1, worksheet, rowIndex, columnIndex);

            FormulaCellValue formulaValue = cellValue as FormulaCellValue;
            if (formulaValue != null)
            {
                RadExpression value = formulaValue.Value.GetValue();
                ArrayExpression arrayExpression = value as ArrayExpression;

                if (arrayExpression == null)
                {
                    arrayExpression = value.GetValue() as ArrayExpression;
                }

                if (arrayExpression != null)
                {
                    foreach (var item in this.GetArrayExpressionItemsAsStrings(arrayExpression))
                    {
                        yield return item;
                    }
                }
            }
            else
            {
                string[] items = cellValue.RawValue.Split(SpreadsheetCultureInfo.ListSeparator[0]);
                foreach (var item in items)
                {
                    yield return item.Trim();
                }
            }
        }

        private IEnumerable<string> GetArrayExpressionItemsAsStrings(ArrayExpression arrayExpression)
        {
            if (arrayExpression != null)
            {
                foreach (var expression in arrayExpression)
                {
                    ArrayExpression array = expression as ArrayExpression;

                    if (array != null)
                    {
                        foreach (var item in this.GetArrayExpressionItemsAsStrings(array))
                        {
                            yield return item;
                        }
                    }
                    else
                    {
                        yield return expression.GetValueAsString();
                    }
                }
            }
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            ListDataValidationRule other = obj as ListDataValidationRule;

            if (other == null)
            {
                return false;
            }

            return base.Equals(obj) &&
                TelerikHelper.EqualsOfT(this.InCellDropdown, other.InCellDropdown);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(base.GetHashCode(), this.InCellDropdown.GetHashCodeOrZero());
        }

        #endregion
    }
}