﻿using System;
using System.Collections;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a base for classes representing collections of a specific type of sheets.
    /// </summary>
    /// <typeparam name="T">The type of the sheet.</typeparam>
    public abstract class FilteredSheetCollection<T> : IEnumerable<T>
        where T : Sheet
    {
        #region Fields

        private readonly SheetCollection sheetCollection;
        private readonly SheetType sheetType;
        private List<T> allSheets;
        private bool allSheetsInvalidated;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the workbook.
        /// </summary>
        /// <value>The workbook.</value>
        public Workbook Workbook
        {
            get
            {
                return this.sheetCollection.Workbook;
            }
        }

        /// <summary>
        /// Gets the item count.
        /// </summary>
        /// <value>The item count.</value>
        public int Count
        {
            get
            {
                return this.GetCount();
            }
        }

        #endregion


        #region Indexers

        /// <summary>
        /// Gets the sheet at the specified index.
        /// </summary>
        /// <value></value>
        public T this[int index]
        {
            get
            {
                return this.GetAllOfT()[index];
            }
        }

        /// <summary>
        /// Gets the sheet with the specified sheet name.
        /// </summary>
        /// <value></value>
        public T this[string sheetName]
        {
            get
            {
                return this.GetByName(sheetName);
            }
        }

        #endregion


        #region Constructors

        internal FilteredSheetCollection(SheetCollection sheetCollection)
        {
            Guard.ThrowExceptionIfNull(sheetCollection, "sheetCollection");

            this.Invalidate();

            this.sheetCollection = sheetCollection;
            this.sheetCollection.Changed += this.SheetCollection_Changed;
            this.sheetType = SpreadsheetHelper.GetSheetType<T>();
        }

        #endregion


        #region Event Handlers

        private void SheetCollection_Changed(object sender, SheetCollectionChangedEventArgs e)
        {
            this.Invalidate();

            this.OnSheetCollectionChanged(e);
        }

        internal event EventHandler<SheetCollectionChangedEventArgs> Changed;

        private void OnSheetCollectionChanged(SheetCollectionChangedEventArgs e)
        {
            T t = e.Sheet as T;

            if (e.Sheet != null && t == null)
            {
                return;
            }

            if (this.Changed != null)
            {
                this.Changed(this, e);
            }
        }

        #endregion


        #region Methods

        internal void Invalidate()
        {
            this.allSheetsInvalidated = true;
        }

        private List<T> GetAllOfT()
        {
            if (this.allSheetsInvalidated)
            {
                this.allSheets = new List<T>();
                this.allSheets.AddRange(this);

                this.allSheetsInvalidated = false;
            }

            return this.allSheets;
        }

        private int GetCount()
        {
            return this.GetAllOfT().Count;
        }

        /// <summary>
        /// Creates a new sheet and adds it to the collection.
        /// </summary>
        /// <returns>The created sheet.</returns>
        public T Add()
        {
            return this.Insert(this.Count);
        }

        /// <summary>
        /// Creates a new sheet and inserts it in the collection at the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <returns>The created sheet.</returns>
        public T Insert(int index)
        {
            List<T> list = this.GetAllOfT();
            Guard.ThrowExceptionIfOutOfRange(0, list.Count, index, "index");

            int indexToInsert;
            if (index == list.Count)
            {
                if (index == 0)
                {
                    // If there aren't any Worksheets, we add it at the end of the sheetsCollection.
                    indexToInsert = this.sheetCollection.Count;
                }
                else
                {
                    // The index after the last element of T.
                    indexToInsert = this.sheetCollection.IndexOf(list[index - 1]) + 1;
                }
            }
            else
            {
                indexToInsert = this.sheetCollection.IndexOf(list[index]);
            }

            return (T)this.sheetCollection.Insert(indexToInsert, this.sheetType);
        }

        /// <summary>
        /// Searches for the specified item and returns the zero-based index of the first occurrence within the entire collection. 
        /// </summary>
        /// <param name="item">The item.</param>
        /// <returns>The index of the item.</returns>
        public int IndexOf(T item)
        {
            List<T> list = this.GetAllOfT();
            return list.IndexOf(item);
        }

        /// <summary>
        /// Searches for the specified item by its name and returns the zero-based index of the first occurrence within the entire collection. .
        /// </summary>
        /// <param name="sheetName">Name of the sheet.</param>
        /// <returns>The index of the item</returns>
        public int IndexOf(string sheetName)
        {
            List<T> list = this.GetAllOfT();
            for (int i = 0; i < list.Count; i++)
            {
                if (SheetCollection.AreEqualIgnoreCaseSheetNames(list[i].Name, sheetName))
                {
                    return i;
                }
            }

            return -1;
        }

        /// <summary>
        /// Determines whether the collection contains a sheet with the specified sheet name.
        /// </summary>
        /// <param name="sheetName">Name of the sheet.</param>
        /// <returns>The value indicating whether the collection contains a sheet with the specified sheet name.</returns>
        public bool Contains(string sheetName)
        {
            return this.IndexOf(sheetName) >= 0;
        }

        /// <summary>
        /// Determines whether the collection contains the specified sheet.
        /// </summary>
        /// <param name="item">The item.</param>
        /// <returns>The value indicating whether the collection contains the specified sheet.</returns>
        public bool Contains(T item)
        {
            return this.IndexOf(item) >= 0;
        }

        /// <summary>
        /// Gets the sheet by name. The search is case insensitive.
        /// </summary>
        /// <param name="sheetName">Name of the sheet.</param>
        /// <returns>The sheet.</returns>
        public T GetByName(string sheetName)
        {
            foreach (T sheet in this)
            {
                if (SheetCollection.AreEqualIgnoreCaseSheetNames(sheet.Name, sheetName))
                {
                    return sheet;
                }
            }

            return null;
        }

        /// <summary>
        /// Removes a sheet at the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        public void RemoveAt(int index)
        {
            List<T> list = this.GetAllOfT();
            T item = list[index];

            this.sheetCollection.Remove(item);
        }

        /// <summary>
        /// Removes a sheet by the specified sheet name.
        /// </summary>
        /// <param name="sheetName">Name of the sheet.</param>
        /// <returns>The value indicating whether a sheet by the specified name was found.</returns>
        public bool Remove(string sheetName)
        {
            T sheet = this.GetByName(sheetName);
            return this.Remove(sheet);
        }

        /// <summary>
        /// Removes the specified item.
        /// </summary>
        /// <param name="item">The item.</param>
        /// <returns>The value indicating whether the specified sheet was found.</returns>
        public bool Remove(T item)
        {
            return this.sheetCollection.Remove(item);
        }

        /// <summary>
        /// Clears this instance.
        /// </summary>
        public void Clear()
        {
            List<T> list = this.GetAllOfT();
            for (int i = list.Count - 1; i >= 0; i--)
            {
                this.sheetCollection.Remove(list[i]);
            }
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<T> GetEnumerator()
        {
            foreach (Sheet sheet in this.sheetCollection)
            {
                T t = sheet as T;
                if (t != null)
                {
                    yield return t;
                }
            }
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return ((IEnumerable<T>)this).GetEnumerator();
        }

        #endregion
    }
}