﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.Copying;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Filtering
{
    /// <summary>
    /// Represents auto filter class.
    /// </summary>
    public class AutoFilter
    {
        #region Fields

        private CellRange filterRange;

        private readonly FilterManager filterManager;
        private readonly Worksheet worksheet;
        private readonly BeginEndCounter collectionChangedCounter;
        private List<IFilter> filters;
        private Dictionary<int, IFilter> relativeColumnIndexToFilter;

        #endregion


        #region Properties

        /// <summary>
        /// Gets or sets the range of the filter, including the column heading.
        /// </summary>
        /// <value>The range of the filter.</value>
        public CellRange FilterRange
        {
            get
            {
                return this.filterRange;
            }
            set
            {
                if (this.filterRange != value)
                {
                    using (new UpdateScope(this.worksheet.BeginUndoGroup, this.worksheet.EndUndoGroup))
                    {
                        if (this.filterRange != null)
                        {
                            this.ClearFilters();
                        }

                        SetFilterRangeCommandContext context = new SetFilterRangeCommandContext(this.worksheet, value);
                        WorkbookCommands.SetFilterRange.Execute(context);
                    }
                }
            }
        }

        internal CellRange ActualFilteredRange
        {
            get
            {
                return SortAndFilterHelper.GetActualFilteredRange(this.filterRange, this.worksheet);
            }
        }

        internal List<IFilter> Filters
        {
            get
            {
                return this.filters;
            }
            set
            {
                if (this.filters != value)
                {
                    this.filters = value;
                    this.SyncFilterColumnListToDictionary();
                }
            }
        }

        internal bool FilterIsApplied
        {
            get
            {
                return this.Filters.Count != 0;
            }
        }

        #endregion


        #region Constructors

        internal AutoFilter(Worksheet owner)
        {
            Guard.ThrowExceptionIfNull(owner, "owner");

            this.worksheet = owner;
            this.filterManager = new FilterManager(this.worksheet);
            this.filters = new List<IFilter>();
            this.relativeColumnIndexToFilter = new Dictionary<int, IFilter>();
            this.collectionChangedCounter = new BeginEndCounter(this.OnFiltersChanged);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Sets the specified filters to the filtered range.
        /// </summary>
        /// <param name="filters">The filters.</param>
        public void SetFilters(IEnumerable<IFilter> filters)
        {
            using (new UpdateScope(this.worksheet.BeginUndoGroup, this.worksheet.EndUndoGroup))
            {
                foreach (IFilter filter in filters)
                {
                    this.SetFilter(filter);
                }
            }
        }

        /// <summary>
        /// Sets the specified filter to the filter range.
        /// </summary>
        /// <param name="filter">The filter.</param>
        public void SetFilter(IFilter filter)
        {
            Guard.ThrowExceptionIfNull(filter, "filter");
            this.CheckIfRangeIsNull();
            Guard.ThrowExceptionIfOutOfRange(0, this.filterRange.ColumnCount - 1, filter.RelativeColumnIndex, "relativeColumnIndex");

            IFilter oldFilter = this.relativeColumnIndexToFilter.ContainsKey(filter.RelativeColumnIndex) ?
                this.relativeColumnIndexToFilter[filter.RelativeColumnIndex] :
                null;

            SetRemoveFilterCommandContext context = new SetRemoveFilterCommandContext(this.worksheet, filter, oldFilter);
            WorkbookCommands.SetFilter.Execute(context);
        }

        internal void SetFilterInternal(IFilter filter)
        {
            Guard.ThrowExceptionIfNull(filter, "filter");
            this.CheckIfRangeIsNull();
            Guard.ThrowExceptionIfOutOfRange(0, this.filterRange.ColumnCount - 1, filter.RelativeColumnIndex, "relativeColumnIndex");

            this.SetFilterWithoutApplying(filter);

            this.filterManager.FilterSheet(this.ActualFilteredRange, filter);
        }

        internal void SetFilterWithoutApplying(IFilter filter)
        {
            Guard.ThrowExceptionIfNull(filter, "filter");
            this.CheckIfRangeIsNull();

            if (this.relativeColumnIndexToFilter.ContainsKey(filter.RelativeColumnIndex))
            {
                IFilter oldFilter = this.relativeColumnIndexToFilter[filter.RelativeColumnIndex];
                this.filterManager.ClearFilter(oldFilter);
                this.filters.Remove(oldFilter);
                this.relativeColumnIndexToFilter.Remove(oldFilter.RelativeColumnIndex);
            }

            IWorksheetFilter worksheetFilter = filter as IWorksheetFilter;
            if (worksheetFilter != null)
            {
                worksheetFilter.SetWorksheet(this.worksheet);
            }

            IRangeFilter rangeFilter = filter as IRangeFilter;
            if (rangeFilter != null)
            {
                CellRange columnRange = this.ActualFilteredRange == null ?
                    null :
                    this.ActualFilteredRange.GetColumn(this.filterRange.FromIndex.ColumnIndex + filter.RelativeColumnIndex);
                rangeFilter.SetColumnRange(columnRange);
            }

            this.filters.Add(filter);
            this.relativeColumnIndexToFilter.Add(filter.RelativeColumnIndex, filter);

            this.DoOnFiltersChanged();
        }

        /// <summary>
        /// Gets the filter applied to the specified column index.
        /// </summary>
        /// <param name="relativeColumnIndex">The column index relative to the filter range.</param>
        /// <returns>The filter applied to the column, or null if the column is not filtered.</returns>
        public IFilter GetFilter(int relativeColumnIndex)
        {
            if (this.relativeColumnIndexToFilter.ContainsKey(relativeColumnIndex))
            {
                return this.relativeColumnIndexToFilter[relativeColumnIndex];
            }
            else
            {
                return null;
            }
        }

        /// <summary>
        /// Removes the specified filter.
        /// </summary>
        /// <param name="filter">The filter.</param>
        /// <returns>A value indicating whether the filter was successfully found and removed.</returns>
        public bool RemoveFilter(IFilter filter)
        {
            return this.RemoveFilter(filter.RelativeColumnIndex);
        }

        /// <summary>
        /// Removes the filter applied to the specified column.
        /// </summary>
        /// <param name="relativeColumnIndex">The column index relative to the filter range.</param>
        /// <returns>A value indicating whether the filter was successfully found and removed.</returns>
        public bool RemoveFilter(int relativeColumnIndex)
        {
            this.CheckIfRangeIsNull();
            Guard.ThrowExceptionIfOutOfRange(0, this.filterRange.ColumnCount - 1, relativeColumnIndex, "relativeColumnIndex");

            if (!this.relativeColumnIndexToFilter.ContainsKey(relativeColumnIndex))
            {
                return false;
            }

            SetRemoveFilterCommandContext context = new SetRemoveFilterCommandContext(this.worksheet, null, this.relativeColumnIndexToFilter[relativeColumnIndex]);
            return WorkbookCommands.RemoveFilter.Execute(context);
        }

        internal void RemoveFilterInternal(int relativeColumnIndex)
        {
            this.CheckIfRangeIsNull();
            Guard.ThrowExceptionIfOutOfRange(0, this.filterRange.ColumnCount - 1, relativeColumnIndex, "relativeColumnIndex");

            if (!this.relativeColumnIndexToFilter.ContainsKey(relativeColumnIndex))
            {
                throw new FilteringException("This column is not filtered.",
                    new InvalidOperationException("This column is not filtered."),
                    "Spreadsheet_Filtering_UnfilteredColumn");
            }


            IFilter filter = this.relativeColumnIndexToFilter[relativeColumnIndex];
            this.RemoveFilterWithoutUnhidingRows(relativeColumnIndex);
            this.filterManager.ClearFilter(filter);
        }

        private void RemoveFilterWithoutUnhidingRows(int relativeColumnIndex)
        {
            IFilter filter = this.relativeColumnIndexToFilter[relativeColumnIndex];

            this.relativeColumnIndexToFilter.Remove(relativeColumnIndex);
            this.filters.Remove(filter);
            this.DoOnFiltersChanged();
        }

        /// <summary>
        /// Clears all the filters.
        /// </summary>
        public void ClearFilters()
        {
            UpdateScope updateScope = new UpdateScope(
                () =>
                {
                    this.worksheet.BeginUndoGroup();
                    this.collectionChangedCounter.BeginUpdate();
                },
                () =>
                {
                    this.collectionChangedCounter.EndUpdate();
                    this.worksheet.EndUndoGroup();
                });

            using (updateScope)
            {
                foreach (IFilter filter in this.filters.ToList())
                {
                    this.RemoveFilter(filter.RelativeColumnIndex);
                }
            }
        }

        /// <summary>
        /// Reapplies the specified filter.
        /// </summary>
        /// <param name="filter">The filter.</param>
        public void ReapplyFilter(IFilter filter)
        {
            this.ReapplyFilter(filter.RelativeColumnIndex);
        }

        /// <summary>
        /// Reapplies the filter applied to the specified column.
        /// </summary>
        /// <param name="relativeColumnIndex">The column index relative to the filter range.</param>
        public void ReapplyFilter(int relativeColumnIndex)
        {
            if (!this.relativeColumnIndexToFilter.ContainsKey(relativeColumnIndex))
            {
                throw new FilteringException("No filter was found to reapply.",
                    new InvalidOperationException("No filter was found to reapply."),
                    "Spreadsheet_Filtering_NoFilterFound");
            }

            IFilter filter = this.relativeColumnIndexToFilter[relativeColumnIndex];
            this.SetFilter(filter);
        }

        internal void SetFilterRangeInternal(CellRange newFilterRange)
        {
            if (this.filterRange != newFilterRange)
            {
                this.filterRange = newFilterRange;
                
                this.OnFilterRangeChanged();
            }
        }

        internal void CopyFrom(AutoFilter fromAutoFilter, CopyContext context)
        {
            // We set the filterRange field instead the property in order the change to be NOT recorded in the history.
            this.filterRange = fromAutoFilter.FilterRange;

            List<IFilter> filtersCopied = new List<IFilter>();
            foreach (ICopyable<IFilter> filter in fromAutoFilter.Filters.Cast<ICopyable<IFilter>>())
            {
                filtersCopied.Add(filter.Copy(context));
            }

            this.filterManager.CopyFrom(fromAutoFilter.filterManager);
            this.Filters = filtersCopied;
        }

        #endregion


        #region Hidden state

        internal ICompressedList<bool> GetRowsHiddenUniquelyByFilter(int relativeColumnIndex)
        {
            return this.filterManager.GetRowsHiddenUniquelyByFilter(relativeColumnIndex);
        }

        internal ICompressedList<bool> GetHiddenRowsState(int relativeColumnIndex)
        {
            return this.filterManager.GetHiddenRowsState(relativeColumnIndex);
        }

        internal void RestoreHiddenRowsState(int relativeColumnIndex, ICompressedList<bool> state)
        {
            this.filterManager.RestoreHiddenRowsState(relativeColumnIndex, state);
        }

        #endregion


        #region Update on insert / remove

        internal bool CanInsertOrRemove(ShiftType shiftType, CellRange selectedRange)
        {
            if (this.filterRange == null)
            {
                return true;
            }

            if (this.FilterIsApplied)
            {
                return this.InsertionIsAfterFilteredRange(selectedRange);
            }

            return this.FilterRange.CanInsertOrRemove(shiftType, selectedRange);
        }

        private bool InsertionIsAfterFilteredRange(CellRange selectedRange)
        {
            if (selectedRange.FromIndex.RowIndex > this.FilterRange.ToIndex.RowIndex ||
                selectedRange.FromIndex.ColumnIndex > this.FilterRange.ToIndex.ColumnIndex)
            {
                return true;
            }

            return false;
        }

        internal void Update(CellRangeInsertedOrRemovedEventArgs eventArgs)
        {
            ShiftType shiftType = eventArgs.RangeType.ToShiftType(eventArgs.IsRemove);
            CellRange insertRemoveRange = eventArgs.Range;

            if (this.filterRange == null)
            {
                return;
            }

            using (new UpdateScope(this.worksheet.BeginUndoGroup, this.worksheet.EndUndoGroup))
            {
                bool translateFiltersVertically =  this.FilterIsApplied &&
                                                   insertRemoveRange.FromIndex.RowIndex <= this.FilterRange.ToIndex.RowIndex &&
                                                   (shiftType == ShiftType.Down || shiftType == ShiftType.Up);

                bool translateFiltersHorizontally = this.FilterIsApplied && 
                                                    insertRemoveRange.FromIndex.ColumnIndex > this.FilterRange.FromIndex.ColumnIndex &&
                                                    insertRemoveRange.FromIndex.ColumnIndex <= this.FilterRange.ToIndex.ColumnIndex &&
                                                    (shiftType == ShiftType.Left || shiftType == ShiftType.Right);

                this.TranslateRange(insertRemoveRange, shiftType);

                if (translateFiltersHorizontally)
                {
                    this.TranslateFiltersHorizontally(insertRemoveRange, shiftType);
                }

                if (translateFiltersVertically)
                {
                    this.TranslateFiltersVertically(insertRemoveRange, shiftType);
                }
            }
        }

        private void TranslateRange(CellRange insertRemoveRange, ShiftType shiftType)
        {
            CellRange resultRange;
            bool success = this.FilterRange.TryTranslate(insertRemoveRange, shiftType, out resultRange);

            if (success)
            {
                if (resultRange == null)
                {
                    this.FilterRange = resultRange;
                }
                else
                {
                    SetFilterRangeCommandContext filterRangeContext = new SetFilterRangeCommandContext(this.worksheet, resultRange);
                    WorkbookCommands.SetFilterRange.Execute(filterRangeContext);
                }
            }
        }

        private void TranslateFiltersHorizontally(CellRange insertRemoveRange, ShiftType shiftType)
        {
            List<IFilter> copyFilters = new List<IFilter>(this.filters);
            int relativeRearrangeIndex = InsertRemoveCellsHelper.GetRelativeRearrangeIndex(this.filterRange, insertRemoveRange);
            int delta = InsertRemoveCellsHelper.GetRearrangeDelta(insertRemoveRange, shiftType);

            RemoveShiftType removeType;
            bool shiftTypeIsRemove = shiftType.TryGetRemoveShiftType(out removeType);

            foreach (IFilter filter in copyFilters)
            {
                if (shiftTypeIsRemove && this.ShouldRemoveFilter(filter.RelativeColumnIndex, insertRemoveRange))
                {
                    this.RemoveFilter(filter);
                }
                else if (filter.RelativeColumnIndex >= relativeRearrangeIndex)
                {
                    int newIndex = filter.RelativeColumnIndex;
                    newIndex = filter.RelativeColumnIndex + delta;

                    MoveFilterCommandContext moveContext = new MoveFilterCommandContext(this.worksheet, filter, newIndex);
                    WorkbookCommands.MoveFilter.Execute(moveContext);
                }
            }
        }

        private void TranslateFiltersVertically(CellRange insertRemoveRange, ShiftType shiftType)
        {
            int absoluteRearrangeIndex = insertRemoveRange.FromIndex.RowIndex;
            int delta = insertRemoveRange.RowCount;

            RearrangeFilterRowsCommandContext context = new RearrangeFilterRowsCommandContext(this.worksheet, delta, absoluteRearrangeIndex, shiftType);
            WorkbookCommands.RearrangeFilterRows.Execute(context);
        }

        internal void RearrangeFilterRows(int absoluteRearrangeIndex, int delta, ShiftType shiftType)
        {
            this.filterManager.RearrangeRows(absoluteRearrangeIndex, delta, shiftType);
        }

        internal IFilter MoveFilter(IFilter filter, int newIndex)
        {
            IFilter newFilter = (IFilter)((ITranslatable)filter).Copy(newIndex);

            this.RemoveFilterWithoutUnhidingRows(filter.RelativeColumnIndex);
            this.SetFilterWithoutApplying(newFilter);

            this.filterManager.MoveFilter(newFilter.RelativeColumnIndex, filter.RelativeColumnIndex);

            return newFilter;
        }

        private bool ShouldRemoveFilter(int filterRelativeIndex, CellRange removedRange)
        {
            int filterAbsoluteIndex = this.FilterRange.FromIndex.ColumnIndex + filterRelativeIndex;
            return filterAbsoluteIndex >= removedRange.FromIndex.ColumnIndex && filterAbsoluteIndex <= removedRange.ToIndex.ColumnIndex;
        }

        #endregion


        #region Validation and syncing

        private void CheckIfRangeIsNull()
        {
            if (this.FilterRange == null)
            {
                throw new FilteringException("The filter range must be assigned before a filter is applied.",
                    new InvalidOperationException("The filter range must be assigned before a filter is applied."),
                    "Spreadsheet_Filtering_RangeMustBeAssigned");
            }
        }

        private void SyncFilterColumnListToDictionary()
        {
            this.relativeColumnIndexToFilter = new Dictionary<int, IFilter>();
            foreach (var column in this.filters)
            {
                this.relativeColumnIndexToFilter.Add(column.RelativeColumnIndex, column);
            }
        }

        #endregion


        #region Events

        private void DoOnFiltersChanged()
        {
            if (!this.collectionChangedCounter.IsUpdateInProgress)
            {
                this.OnFiltersChanged();
            }
        }

        internal event EventHandler FiltersChanged;

        private void OnFiltersChanged()
        {
            if (this.FiltersChanged != null)
            {
                this.FiltersChanged(this, EventArgs.Empty);
            }
        }

        internal event EventHandler FilterRangeChanged;

        private void OnFilterRangeChanged()
        {
            if (this.FilterRangeChanged != null)
            {
                this.FilterRangeChanged(this, EventArgs.Empty);
            }
        }

        #endregion
    }
}
