﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Filtering
{
    /// <summary>
    /// Describes a filter which compares the cell value to one or two critera in order to determine whether it should be shown.
    /// </summary>
    public class CustomFilter : CellValuesFilterBase
    {
        private readonly CustomFilterCriteria criteria1;
        private readonly CustomFilterCriteria criteria2;
        private readonly LogicalOperator logicalOperator;

        /// <summary>
        /// Gets the first criteria.
        /// </summary>
        /// <value>The first criteria.</value>
        public CustomFilterCriteria Criteria1
        {
            get
            {
                return this.criteria1;
            }
        }

        /// <summary>
        /// Gets the second criteria.
        /// </summary>
        /// <value>The second criteria.</value>
        public CustomFilterCriteria Criteria2
        {
            get
            {
                return this.criteria2;
            }
        }

        /// <summary>
        /// Gets the logical operator.
        /// </summary>
        /// <value>The logical operator.</value>
        public LogicalOperator LogicalOperator
        {
            get
            {
                return this.logicalOperator;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CustomFilter" /> class.
        /// </summary>
        /// <param name="relativeColumnIndex">Relative index of the column.</param>
        /// <param name="criteria">The criteria.</param>
        public CustomFilter(int relativeColumnIndex, CustomFilterCriteria criteria) : this(relativeColumnIndex, criteria, LogicalOperator.Or, null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CustomFilter" /> class.
        /// </summary>
        /// <param name="relativeColumnIndex">Relative index of the column.</param>
        /// <param name="criteria1">The first criteria.</param>
        /// <param name="logicalOperator">The logical operator.</param>
        /// <param name="criteria2">The second criteria.</param>
        public CustomFilter(int relativeColumnIndex, CustomFilterCriteria criteria1, LogicalOperator logicalOperator, CustomFilterCriteria criteria2) : base(relativeColumnIndex)
        {
            Guard.ThrowExceptionIfNull(criteria1, "criteria1");

            this.criteria1 = criteria1;
            this.criteria2 = criteria2;
            this.logicalOperator = logicalOperator;
        }

        internal override IFilter Copy(int newRelativeColumnIndex)
        {
            return new CustomFilter(newRelativeColumnIndex, this.Criteria1, this.LogicalOperator, this.Criteria2);
        }

        /// <summary>
        /// Determines whether the row which contains the specified value will be shown.
        /// </summary>
        /// <param name="value">The value retrieved by the GetValue method.</param>
        /// <returns>
        /// A value indicating whether the row which contains the specified value will be shown
        /// </returns>
        public override bool ShouldShowValue(object value)
        {
            ICellValue cellValueObject = value as ICellValue;

            bool cellValueSatisfiesCriteria1 = this.CellValueSatisifiesCriteria(this.criteria1, cellValueObject);
            var cellValueSatisfiesCriteria2 = this.CellValueSatisifiesCriteria(this.criteria2, cellValueObject);

            return (this.logicalOperator == LogicalOperator.And) ?
                cellValueSatisfiesCriteria1 && cellValueSatisfiesCriteria2 :
                cellValueSatisfiesCriteria1 || cellValueSatisfiesCriteria2;
        }

        private bool CellValueSatisifiesCriteria(CustomFilterCriteria criteria, ICellValue cellValue)
        {
            if (criteria == null)
            {
                return false;
            }

            bool satisfiesCriteria = false;

            IComparable value = this.GetComparableValueFromCellValue(criteria.ComparableFilterValue, cellValue);

            if (value == null)
            {
                return false;
            }

            IComparable loweredFilterValue = this.ToLowerString(criteria.ComparableFilterValue);

            switch (criteria.ComparisonOperator)
            {
                case ComparisonOperator.GreaterThan:
                    satisfiesCriteria = value.CompareTo(loweredFilterValue) > 0;
                    break;
                case ComparisonOperator.EqualsTo:
                    // TODO: consider using number format
                    satisfiesCriteria = value.CompareTo(loweredFilterValue) == 0;
                    break;
                case ComparisonOperator.GreaterThanOrEqualsTo:
                    satisfiesCriteria = value.CompareTo(loweredFilterValue) >= 0;
                    break;
                case ComparisonOperator.LessThan:
                    satisfiesCriteria = value.CompareTo(loweredFilterValue) < 0;
                    break;
                case ComparisonOperator.LessThanOrEqualsTo:
                    satisfiesCriteria = value.CompareTo(loweredFilterValue) <= 0;
                    break;
                case ComparisonOperator.NotEqualsTo:
                    satisfiesCriteria = value.CompareTo(loweredFilterValue) != 0;
                    break;
                default: throw new FilteringException("Unknown compare operator.",
                    new InvalidOperationException("Unknown compare operator."),
                    "Spreadsheet_Filtering_UnknownCompareOperator");
            }

            return satisfiesCriteria;
        }

        private IComparable ToLowerString(IComparable filterValue)
        {
            string stringFilterValue = filterValue as string;

            if (stringFilterValue != null)
            {
                return stringFilterValue.ToLower();
            }

            return filterValue;
        }

        private IComparable GetComparableValueFromCellValue(IComparable filterValue, ICellValue cellValue)
        {
            IComparable result = null;
            FormulaCellValue formulaCellValue = cellValue as FormulaCellValue;

            if (filterValue is double)
            {
                NumberCellValue numberCellValue = cellValue as NumberCellValue;

                if (numberCellValue == null)
                {
                    if (formulaCellValue != null)
                    {
                        numberCellValue = formulaCellValue.GetResultValueAsCellValue() as NumberCellValue;
                    }
                }

                if (numberCellValue != null)
                {
                    result = numberCellValue.Value;
                }
            }
            else if (filterValue is string)
            {
                TextCellValue textCellValue = cellValue as TextCellValue;

                if (textCellValue == null)
                {
                    if (formulaCellValue != null)
                    {
                        textCellValue = formulaCellValue.GetResultValueAsCellValue() as TextCellValue;
                    }
                }

                if (textCellValue != null)
                {
                    result = textCellValue.Value.ToLower();
                }
            }
            else if (filterValue is bool)
            {
                BooleanCellValue booleanCellValue = cellValue as BooleanCellValue;

                if (booleanCellValue == null)
                {
                    if (formulaCellValue != null)
                    {
                        booleanCellValue = formulaCellValue.GetResultValueAsCellValue() as BooleanCellValue;
                    }
                }

                if (booleanCellValue != null)
                {
                    result = booleanCellValue.Value;
                }
            }

            return result;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            CustomFilter other = obj as CustomFilter;

            if (other == null)
            {
                return false;
            }

            return TelerikHelper.EqualsOfT(this.Criteria1, other.Criteria1) &&
                TelerikHelper.EqualsOfT(this.Criteria2, other.Criteria2) &&
                this.LogicalOperator.Equals(other.LogicalOperator);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                this.Criteria1.GetHashCodeOrZero(),
                this.Criteria2.GetHashCodeOrZero(),
                this.LogicalOperator.GetHashCode());
        }
    }
}