﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Filtering
{
    /// <summary>
    /// Describes an item used by the ValuesCollectionFilter to filter date values.
    /// </summary>
    public class DateGroupItem
    {
        private readonly DateTimeGroupingType dateTimeGroupingType;

        private readonly int year;
        private readonly int month;
        private readonly int day;
        private readonly int hour;
        private readonly int minute;
        private readonly int second;

        /// <summary>
        /// Gets the grouping type of the item.
        /// </summary>
        /// <value>The grouping type of the item.</value>
        public DateTimeGroupingType DateTimeGroupingType
        {
            get
            {
                return this.dateTimeGroupingType;
            }
        }

        /// <summary>
        /// Gets the year.
        /// </summary>
        /// <value>The year.</value>
        public int Year
        {
            get
            {
                return this.year;
            }
        }

        /// <summary>
        /// Gets the month.
        /// </summary>
        /// <value>The month.</value>
        public int Month
        {
            get
            {
                return this.month;
            }
        }

        /// <summary>
        /// Gets the day.
        /// </summary>
        /// <value>The day.</value>
        public int Day
        {
            get
            {
                return this.day;
            }
        }

        /// <summary>
        /// Gets the hour.
        /// </summary>
        /// <value>The hour.</value>
        public int Hour
        {
            get
            {
                return this.hour;
            }
        }

        /// <summary>
        /// Gets the minute.
        /// </summary>
        /// <value>The minute.</value>
        public int Minute
        {
            get
            {
                return this.minute;
            }
        }

        /// <summary>
        /// Gets the second.
        /// </summary>
        /// <value>The second.</value>
        public int Second
        {
            get
            {
                return this.second;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="DateGroupItem" /> class.
        /// </summary>
        /// <param name="year">The year.</param>
        public DateGroupItem(int year) : this(DateTimeGroupingType.Year, year, 0, 0, 0, 0, 0)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="DateGroupItem" /> class.
        /// </summary>
        /// <param name="year">The year.</param>
        /// <param name="month">The month.</param>
        public DateGroupItem(int year, int month) : this(DateTimeGroupingType.Month, year, month, 0, 0, 0, 0)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="DateGroupItem" /> class.
        /// </summary>
        /// <param name="year">The year.</param>
        /// <param name="month">The month.</param>
        /// <param name="day">The day.</param>
        public DateGroupItem(int year, int month, int day) : this(DateTimeGroupingType.Day, year, month, day, 0, 0, 0)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="DateGroupItem" /> class.
        /// </summary>
        /// <param name="year">The year.</param>
        /// <param name="month">The month.</param>
        /// <param name="day">The day.</param>
        /// <param name="hour">The hour.</param>
        public DateGroupItem(int year, int month, int day, int hour) : this(DateTimeGroupingType.Hour, year, month, day, hour, 0, 0)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="DateGroupItem" /> class.
        /// </summary>
        /// <param name="year">The year.</param>
        /// <param name="month">The month.</param>
        /// <param name="day">The day.</param>
        /// <param name="hour">The hour.</param>
        /// <param name="minute">The minute.</param>
        public DateGroupItem(int year, int month, int day, int hour, int minute) : this(DateTimeGroupingType.Minute, year, month, day, hour, minute, 0)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="DateGroupItem" /> class.
        /// </summary>
        /// <param name="year">The year.</param>
        /// <param name="month">The month.</param>
        /// <param name="day">The day.</param>
        /// <param name="hour">The hour.</param>
        /// <param name="minute">The minute.</param>
        /// <param name="seconds">The seconds.</param>
        public DateGroupItem(int year, int month, int day, int hour, int minute, int seconds) : this(DateTimeGroupingType.Second, year, month, day, hour, minute, seconds)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="DateGroupItem" /> class.
        /// </summary>
        /// <param name="dateTimeGroupingType">Type of the date time grouping.</param>
        /// <param name="year">The year.</param>
        /// <param name="month">The month.</param>
        /// <param name="day">The day.</param>
        /// <param name="hour">The hour.</param>
        /// <param name="minute">The minute.</param>
        /// <param name="second">The second.</param>
        internal DateGroupItem(DateTimeGroupingType dateTimeGroupingType, int year, int month, int day, int hour, int minute, int second)
        {
            Guard.ThrowExceptionIfLessThan(0, year, "year");
            Guard.ThrowExceptionIfOutOfRange(0, 12, month, "month");
            Guard.ThrowExceptionIfOutOfRange(0, 31, day, "day");
            Guard.ThrowExceptionIfOutOfRange(0, 23, hour, "hour");
            Guard.ThrowExceptionIfOutOfRange(0, 59, minute, "minute");
            Guard.ThrowExceptionIfOutOfRange(0, 59, second, "second");

            this.dateTimeGroupingType = dateTimeGroupingType;
            this.year = year;
            this.month = month;
            this.day = day;
            this.hour = hour;
            this.minute = minute;
            this.second = second;
        }

        /// <summary>
        /// Determine whether the specified date satisfies the necessary conditions of the item. All dates whose year, month, day, hour, minute and/or second
        /// are the same as the properties of the item satisfy the item and will be shown by the filter. Which properties will be used in the comparison is
        /// determined by the grouping type of the item.
        /// </summary>
        /// <param name="date">The date.</param>
        /// <returns>A value indicating whether the specified date satisfies the item.</returns>
        public bool DateSatisfiesItem(DateTime date)
        {
            if (date.Year != this.year)
            {
                return false;
            }

            if (this.dateTimeGroupingType >= DateTimeGroupingType.Month &&
                date.Month != this.month)
            {
                return false;
            }

            if (this.dateTimeGroupingType >= DateTimeGroupingType.Day &&
                date.Day != this.day)
            {
                return false;
            }

            if (this.dateTimeGroupingType >= DateTimeGroupingType.Hour &&
                date.Hour != this.hour)
            {
                return false;
            }

            if (this.dateTimeGroupingType >= DateTimeGroupingType.Minute &&
                date.Minute != this.minute)
            {
                return false;
            }

            if (this.dateTimeGroupingType >= DateTimeGroupingType.Second &&
                date.Second != this.second)
            {
                return false;
            }

            return true;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            DateGroupItem other = obj as DateGroupItem;

            if (other == null)
            {
                return false;
            }

            return this.Year.Equals(other.Year) &&
                this.Month.Equals(other.Month) &&
                this.Day.Equals(other.Day) &&
                this.Hour.Equals(other.Hour) &&
                this.Minute.Equals(other.Minute) &&
                this.Second.Equals(other.Second) &&
                this.DateTimeGroupingType.Equals(other.DateTimeGroupingType);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(this.Year.GetHashCode(), this.Month.GetHashCode(), this.Day.GetHashCode(),
                this.Hour.GetHashCode(), this.Minute.GetHashCode(), this.Second.GetHashCode(),
                this.DateTimeGroupingType.GetHashCode());
        }
    }
}
