﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Filtering
{
    /// <summary>
    /// Describes a filter which hides or shows the cells based on a criteria specified by the filter type.
    /// </summary>
    public class DynamicFilter : CellValuesFilterBase, IWorksheetFilter, IRangeFilter
    {
        private readonly DynamicFilterType filterType;
        private double? average = null;

        private Dictionary<DynamicFilterType, Func<DateTime, bool>> shouldShowForDate;
        private Dictionary<DynamicFilterType, Func<double, bool>> shouldShowForNumber;
        private Worksheet worksheet;

        /// <summary>
        /// Gets the type of the dynamic filter.
        /// </summary>
        /// <value>The type of the dynamic filter.</value>
        public DynamicFilterType DynamicFilterType
        {
            get
            {
                return this.filterType;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="DynamicFilter" /> class.
        /// </summary>
        /// <param name="relativeColumnIndex">Relative index of the column.</param>
        /// <param name="dynamicFilterType">Type of the dynamic filter.</param>
        public DynamicFilter(int relativeColumnIndex, DynamicFilterType dynamicFilterType) : base(relativeColumnIndex)
        {
            this.filterType = dynamicFilterType;

            this.InitializeShowForDatePredicates();
            this.InitializeShowForNumberPredicates();
        }

        internal override IFilter Copy(int newRelativeColumnIndex)
        {
            return new DynamicFilter(newRelativeColumnIndex, this.DynamicFilterType);
        }

        void IWorksheetFilter.SetWorksheet(Worksheet worksheet)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");

            this.worksheet = worksheet;
        }

        void IRangeFilter.SetColumnRange(CellRange columnRange)
        {
            this.SetColumnRange(columnRange);
        }

        private void SetColumnRange(CellRange columnRange)
        {
            Guard.ThrowExceptionIfNull(this.worksheet, "worksheet");
            if (columnRange != null)
            {
                Guard.ThrowExceptionIfNotEqual(1, columnRange.ColumnCount, "range");
            }

            if (this.filterType == DynamicFilterType.AboveAverage ||
                this.filterType == DynamicFilterType.BelowAverage)
            {
                this.average = this.CalculateAverage(this.worksheet, columnRange);
            }
        }

        private double CalculateAverage(Worksheet worksheet, CellRange range)
        {
            if (range == null)
            {
                return 0;
            }

            double sum = 0;
            int number = 0;

            for (int row = range.FromIndex.RowIndex; row <= range.ToIndex.RowIndex; row++)
            {
                long index = CellsPropertyBag.ConvertCellIndexToLong(row, range.FromIndex.ColumnIndex);
                ICellValue cellValue = worksheet.Cells.PropertyBag.GetPropertyValue(CellPropertyDefinitions.ValueProperty, index);
                NumberCellValue numberCellValue = cellValue as NumberCellValue;

                if(numberCellValue != null)
                {
                    sum += numberCellValue.Value;
                    number++;
                }
            }

            if (number != 0)
            {
                return sum / number;
            }
            else
            {
                throw new InvalidOperationException();
            }
        }

        /// <summary>
        /// Gets the value of the cell at the specified index. This value is used to determine whether the row will be hidden by the filter.
        /// </summary>
        /// <param name="cells">The cells of the worksheet.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The value of the cell.</returns>
        public override object GetValue(Cells cells, int rowIndex, int columnIndex)
        {
            ICellValue cellValue = base.GetValue(cells, rowIndex, columnIndex) as ICellValue;

            FormulaCellValue formulaCellValue = cellValue as FormulaCellValue;

            if (formulaCellValue != null)
            {
                cellValue = formulaCellValue.GetResultValueAsCellValue();
            }

            NumberCellValue numberCellValue = cellValue as NumberCellValue;

            if (numberCellValue == null)
            {
                return null;
            }

            long index = CellsPropertyBag.ConvertCellIndexToLong(rowIndex, columnIndex);
            CellValueFormat format = cells.PropertyBag.GetPropertyValue(CellPropertyDefinitions.FormatProperty, index);

            if (format.IsDateFormat())
            {
                DateTime? result = numberCellValue.ToDateTime();
                if (result.HasValue && numberCellValue.Value >= 1)
                {
                    return result.Value;
                }
            }

            return numberCellValue.Value;
        }

        /// <summary>
        /// Determines whether the row which contains the specified value will be shown.
        /// </summary>
        /// <param name="value">The value retrieved by the GetValue method.</param>
        /// <returns>
        /// A value indicating whether the row which contains the specified value will be shown
        /// </returns>
        public override bool ShouldShowValue(object value)
        {
            if (value is DateTime)
            {
                return ShouldShowValueDate((DateTime)value);
            }
            else if (value is double)
            {
                return ShouldShowValueNumber((double)value);
            }
            else
            {
                return false;
            }
        }

        private bool ShouldShowValueNumber(double number)
        {
            if (!this.shouldShowForNumber.ContainsKey(this.filterType))
            {
                throw new FilteringException("Invalid dynamic filter type",
                        new InvalidOperationException("Invalid dynamic filter type"),
                        "Spreadsheet_Filtering_InvalidDynamicFilter");
            }

            return this.shouldShowForNumber[this.filterType](number);
        }

        private bool ShouldShowValueDate(DateTime date)
        {
            if (!this.shouldShowForDate.ContainsKey(this.filterType))
            {
                throw new FilteringException("Invalid dynamic filter type",
                        new InvalidOperationException("Invalid dynamic filter type"),
                        "Spreadsheet_Filtering_InvalidDynamicFilter");
            }

            return this.shouldShowForDate[this.filterType](date);
        }

        private void InitializeShowForNumberPredicates()
        {
            this.shouldShowForNumber = new Dictionary<DynamicFilterType, Func<double, bool>>();

            this.shouldShowForNumber.Add(DynamicFilterType.BelowAverage,
                new Func<double,bool>(
                    (number) =>
                        {
                            if (!this.average.HasValue)
                            {
                                throw new FilteringException("This filter is not yet assigned to a column.",
                                    new InvalidOperationException("This filter is not yet assigned to a column."),
                                    "Spreadsheet_Filtering_RangeFilterNotAssigned");
                            }

                            return number < this.average;
                        }));
            this.shouldShowForNumber.Add(DynamicFilterType.AboveAverage,
                new Func<double,bool>(
                    (number) =>
                        {
                            if (!this.average.HasValue)
                            {
                                throw new FilteringException("This filter is not yet assigned to a column.",
                                    new InvalidOperationException("This filter is not yet assigned to a column."),
                                    "Spreadsheet_Filtering_RangeFilterNotAssigned");
                            }

                            return number > this.average;
                        }));

            this.shouldShowForNumber.Add(DynamicFilterType.None,
                new Func<double, bool>(
                (number) =>
                {
                    throw new FilteringException("Invalid dynamic filter type",
                        new InvalidOperationException("Invalid dynamic filter type"),
                        "Spreadsheet_Filtering_InvalidDynamicFilter");
                }));

            foreach (DynamicFilterType filterType in Enum.GetValues(typeof(DynamicFilterType)))
            {
                if (!this.shouldShowForNumber.ContainsKey(filterType))
                {
                    this.shouldShowForNumber.Add(filterType, (number) => { return false; });
                }
            }
        }

        private void InitializeShowForDatePredicates()
        {
            this.shouldShowForDate = new Dictionary<DynamicFilterType, Func<DateTime, bool>>();

            this.shouldShowForDate.Add(DynamicFilterType.BelowAverage,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return FormatHelper.ConvertDateTimeToDouble(date) < this.average;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.AboveAverage,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return FormatHelper.ConvertDateTimeToDouble(date) > this.average;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.Tomorrow,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.AddDays(-1) == DateTime.Today;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.Today,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date == DateTime.Today;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.Yesterday,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                           return date.AddDays(1) == DateTime.Today; 
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.NextWeek,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.GetBeginningOfWeek().AddDays(-7) == DateTime.Today.GetBeginningOfWeek();
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.ThisWeek,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.GetBeginningOfWeek() == DateTime.Today.GetBeginningOfWeek();
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.LastWeek,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.GetBeginningOfWeek().AddDays(7) == DateTime.Today.GetBeginningOfWeek();
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.NextMonth,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.AddMonths(-1).Month == DateTime.Today.Month;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.ThisMonth,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Month == DateTime.Today.Month;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.LastMonth,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.AddMonths(1).Month == DateTime.Today.Month;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.NextQuarter,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.DateIsNextQuarter();
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.ThisQuarter,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Year == DateTime.Today.Year &&
                                    date.GetQuarter() == DateTime.Today.GetQuarter();
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.LastQuarter,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.DateIsLastQuarter();
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.NextYear,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Year - 1 == DateTime.Today.Year;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.ThisYear,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Year == DateTime.Today.Year;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.LastYear,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Year + 1 == DateTime.Today.Year;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.YearToDate,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Year == DateTime.Today.Year && date <= DateTime.Today;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.Quarter1,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.GetQuarter() == 1;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.Quarter2,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.GetQuarter() == 2;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.Quarter3,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.GetQuarter() == 3;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.Quarter4,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.GetQuarter() == 4;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.January,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Month == 1;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.February,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Month == 2;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.March,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Month == 3;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.April,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Month == 4;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.May,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Month == 5;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.June,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Month == 6;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.July,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Month == 7;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.August,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Month == 8;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.September,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Month == 9;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.October,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Month == 10;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.November,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Month == 11;
                        }));
            this.shouldShowForDate.Add(DynamicFilterType.December,
                new Func<DateTime,bool>(
                    (date) =>
                        {
                            return date.Month == 12;
                        }));
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            DynamicFilter other = obj as DynamicFilter;

            if (other == null)
            {
                return false;
            }

            return this.DynamicFilterType.Equals(other.DynamicFilterType);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return this.DynamicFilterType.GetHashCode();
        }
    }
}