﻿using System;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Filtering
{
    /// <summary>
    /// Describes a filter which hides or shows the cells based on the cell fill color.
    /// </summary>
    public class FillColorFilter : FilterBase<IFill>, IWorksheetFilter
    {
        private readonly IFill fill;
        private IFill actualFill;
        private Worksheet worksheet;

        /// <summary>
        /// Gets the fill which the cells displayed by the filter should have.
        /// </summary>
        /// <value>The fill.</value>
        public IFill Fill
        {
            get
            {
                return this.fill;
            }
        }

        /// <summary>
        /// Gets the property definition for the value which will be retrieved by the GetValue method.
        /// </summary>
        /// <value>The property definition.</value>
        protected override IPropertyDefinition<IFill> PropertyDefinition
        {
            get 
            {
                return CellPropertyDefinitions.FillProperty;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="FillColorFilter" /> class.
        /// </summary>
        /// <param name="relativeColumnIndex">Relative index of the column.</param>
        /// <param name="fill">The fill.</param>
        public FillColorFilter(int relativeColumnIndex, IFill fill) 
            : base(relativeColumnIndex)
        {
            Guard.ThrowExceptionIfNull(fill, "fill");
            
            this.fill = fill;
        }

        internal override IFilter Copy(int newRelativeColumnIndex)
        {
            return new FillColorFilter(newRelativeColumnIndex, this.Fill);
        }

        void IWorksheetFilter.SetWorksheet(Worksheet worksheet)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");

            this.worksheet = worksheet;
            this.actualFill = SortAndFilterHelper.GetIFillWithLocalColors(this.fill, this.worksheet.Workbook.Theme.ColorScheme);
        }

        /// <summary>
        /// Determines whether the row which contains the specified value will be shown.
        /// </summary>
        /// <param name="value">The value retrieved by the GetValue method.</param>
        /// <returns>
        /// A value indicating whether the row which contains the specified value will be shown
        /// </returns>
        public override bool ShouldShowValue(object value)
        {
            IFill fillValue = value as IFill;

            Guard.ThrowExceptionIfNull(this.worksheet, "worksheet");
            fillValue = SortAndFilterHelper.GetIFillWithLocalColors(fillValue, this.worksheet.Workbook.Theme.ColorScheme);

            return fillValue.Equals(this.actualFill);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            FillColorFilter other = obj as FillColorFilter;

            if (other == null)
            {
                return false;
            }

            return this.Fill.Equals(other.Fill);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return this.Fill.GetHashCode();
        }
    }
}