﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Copying;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Filtering
{
    /// <summary>
    /// Provides a base class for the filter classes.
    /// </summary>
    /// <typeparam name="T">The type of the property definition of the filter.</typeparam>
    public abstract class FilterBase<T> : IFilter, ITranslatable, ICopyable<IFilter>
    {
        /// <summary>
        /// Gets the index of the column to which the filter is applied. The index is relative to the beginning of the filter range.
        /// </summary>
        /// <value>The relative index of the column.</value>
        public int RelativeColumnIndex { get; private set; }

        /// <summary>
        /// Gets the property definition for the value which will be retrieved by the GetValue method.
        /// </summary>
        /// <value>The property definition.</value>
        protected abstract IPropertyDefinition<T> PropertyDefinition { get; }

        /// <summary>
        /// Initializes a new instance of the <see cref="FilterBase{T}" /> class.
        /// </summary>
        /// <param name="relativeColumnIndex">Index of the relative column.</param>
        protected FilterBase(int relativeColumnIndex)
        {
            this.RelativeColumnIndex = relativeColumnIndex;
        }

        /// <summary>
        /// Gets the value of the cell at the specified index. This value is used to determine whether the row will be hidden by the filter.
        /// </summary>
        /// <param name="cells">The cells of the worksheet.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The value of the cell.</returns>
        public virtual object GetValue(Cells cells, int rowIndex, int columnIndex)
        {
            long index = CellsPropertyBag.ConvertCellIndexToLong(rowIndex, columnIndex);
            return cells.PropertyBag.GetPropertyValueRespectingStyle(this.PropertyDefinition, cells.Worksheet, index);
        }

        /// <summary>
        /// Determines whether the row which contains the specified value will be shown.
        /// </summary>
        /// <param name="value">The value retrieved by the GetValue method.</param>
        /// <returns>
        /// A value indicating whether the row which contains the specified value will be shown
        /// </returns>
        public abstract bool ShouldShowValue(object value);

        /// <summary>
        /// Copies the specified item with a new relative index.
        /// </summary>
        /// <param name="newRelativeColumnIndex">New relative index.</param>
        /// <returns>The new item</returns>
        object ITranslatable.Copy(int newRelativeColumnIndex)
        {
            return this.Copy(newRelativeColumnIndex);
        }

        internal abstract IFilter Copy(int newRelativeColumnIndex);

        IFilter ICopyable<IFilter>.Copy(CopyContext context)
        {
            IFilter filterCopied = this.Copy(this.RelativeColumnIndex);
            IWorksheetFilter worksheetFilter = filterCopied as IWorksheetFilter;
            if (worksheetFilter != null)
            {
                worksheetFilter.SetWorksheet(context.TargetWorksheet);
            }

            return filterCopied;
        }
    }
}
