﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Filtering
{
    internal class FilterManager
    {
        private readonly Worksheet worksheet;

        private Dictionary<int, ICompressedList<bool>> relativeIndexToHiddenRowsState;
        private Dictionary<int, int> hiddenRows;

        public FilterManager(Worksheet worksheet)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");

            this.worksheet = worksheet;
            this.relativeIndexToHiddenRowsState = new Dictionary<int, ICompressedList<bool>>();
            this.hiddenRows = new Dictionary<int, int>();
        }

        public void FilterSheet(CellRange range, IFilter filter)
        {
            Guard.ThrowExceptionIfNull(filter, "filter");

            int relativeColumnIndex = filter.RelativeColumnIndex;

            if (this.relativeIndexToHiddenRowsState.ContainsKey(relativeColumnIndex))
            {
                throw new FilteringException("This column is already filtered. Remove the filter before applying a new one.",
                    new InvalidOperationException("This column is already filtered. Remove the filter before applying a new one."),
                    "Spreadsheet_Filtering_ColumnFiltered");
            }

            CompressedList<bool> allHiddenRowsByFilter = this.GetEmptyCompressedList();
            this.relativeIndexToHiddenRowsState.Add(relativeColumnIndex, allHiddenRowsByFilter);

            if (range == null)
            {
                return;
            }

            int columnIndex = range.FromIndex.ColumnIndex + relativeColumnIndex;

            for (int row = range.FromIndex.RowIndex; row <= range.ToIndex.RowIndex; row++)
            {
                object value = filter.GetValue(this.worksheet.Cells, row, columnIndex);

                if (!filter.ShouldShowValue(value))
                {
                    if (!this.hiddenRows.ContainsKey(row))
                    {
                        this.SetHiddenOutsideCommand(row);
                        this.hiddenRows.Add(row, 1);
                    }
                    else
                    {
                        this.hiddenRows[row]++;
                    }

                    allHiddenRowsByFilter.SetValue(row, true);
                }
            }
        }

        public void ClearFilter(IFilter filter)
        {
            Guard.ThrowExceptionIfNull(filter, "filter");

            ICompressedList<bool> allHiddenRowsForFilter = this.relativeIndexToHiddenRowsState[filter.RelativeColumnIndex];

            IEnumerable<Range<long, bool>> hiddenRanges = allHiddenRowsForFilter.GetNonDefaultRanges();

            this.UnhideHiddenRanges(hiddenRanges);

            this.relativeIndexToHiddenRowsState.Remove(filter.RelativeColumnIndex);
        }

        private void UnhideHiddenRanges(IEnumerable<Range<long, bool>> allHiddenRanges)
        {
            foreach (Range<long, bool> range in allHiddenRanges)
            {
                int startRowIndex = (int)range.Start;
                int endRowIndex = (int)range.End;

                for (int rowIndex = startRowIndex; rowIndex <= endRowIndex; rowIndex++)
                {
                    this.hiddenRows[rowIndex]--;
                    if (this.hiddenRows[rowIndex] == 0)
                    {
                        this.hiddenRows.Remove(rowIndex);
                        this.ClearHiddenOutsideCommand(rowIndex);
                    }
                }
            }
        }

        internal ICompressedList<bool> GetHiddenRowsState(int relativeColumnIndex)
        {
            ICompressedList<bool> result;

            if (!this.relativeIndexToHiddenRowsState.ContainsKey(relativeColumnIndex))
            {
                result = this.GetEmptyCompressedList();
            }
            else
            {
                result = this.relativeIndexToHiddenRowsState[relativeColumnIndex];
            }

            return result;
        }

        public void RestoreHiddenRowsState(int relativeColumnIndex, ICompressedList<bool> filterState)
        {
            this.relativeIndexToHiddenRowsState[relativeColumnIndex] = filterState;

            IEnumerable<Range<long, bool>> allHiddenRanges = filterState.GetNonDefaultRanges();

            foreach (Range<long, bool> range in allHiddenRanges)
            {
                int startRowIndex = (int)range.Start;
                int endRowIndex = (int)range.End;

                for (int rowIndex = startRowIndex; rowIndex <= endRowIndex; rowIndex++)
                {
                    if (this.hiddenRows.ContainsKey(rowIndex))
                    {
                        this.hiddenRows[rowIndex]++;
                    }
                    else
                    {
                        this.hiddenRows.Add(rowIndex, 1);
                        this.SetHiddenOutsideCommand(rowIndex);
                    }
                }
            }
        }

        public ICompressedList<bool> GetRowsHiddenUniquelyByFilter(int relativeColumnIndex)
        {
            ICompressedList<bool> result = new CompressedList<bool>(0, SpreadsheetDefaultValues.RowCount - 1, false);

            if (!this.relativeIndexToHiddenRowsState.ContainsKey(relativeColumnIndex))
            {
                return result;
            }

            ICompressedList<bool> allHiddenRowsForFilter = this.relativeIndexToHiddenRowsState[relativeColumnIndex];
            IEnumerable<Range<long, bool>> hiddenRanges = allHiddenRowsForFilter.GetNonDefaultRanges();

            foreach (Range<long, bool> range in hiddenRanges)
            {
                int startRowIndex = (int)range.Start;
                int endRowIndex = (int)range.End;

                for (int rowIndex = startRowIndex; rowIndex <= endRowIndex; rowIndex++)
                {
                    if (this.hiddenRows[rowIndex] == 1)
                    {
                        result.SetValue(rowIndex, true);
                    }
                }
            }

            return result;
        }

        public void MoveFilter(int newIndex, int oldIndex)
        {
            var list = this.relativeIndexToHiddenRowsState[oldIndex];
            this.relativeIndexToHiddenRowsState[newIndex] = list;
            this.relativeIndexToHiddenRowsState.Remove(oldIndex);
        }

        public void RearrangeRows(int absoluteRearrangeIndex, int delta, ShiftType shiftType)
        {
            if (shiftType != ShiftType.Down && shiftType != ShiftType.Up)
            {
                throw new ArgumentException("The shift type must be ShiftType.Down or ShiftType.Up", "shiftType");
            }

            this.RearrangeIndexToHiddenRowStateDictionary(absoluteRearrangeIndex, delta, shiftType);
            this.RearrangeHiddenRowsDictionary(absoluteRearrangeIndex, delta, shiftType);
        }

        private void RearrangeIndexToHiddenRowStateDictionary(int absoluteRearrangeIndex, int delta, ShiftType shiftType)
        {
            Dictionary<int, ICompressedList<bool>> newIndexToRowStateDictionary = new Dictionary<int, ICompressedList<bool>>();

            foreach (KeyValuePair<int, ICompressedList<bool>> pair in this.relativeIndexToHiddenRowsState)
            {
                ICompressedList<bool> list = pair.Value;
                IEnumerable<Range<long, bool>> hiddenRanges = list.GetNonDefaultRanges();
                CompressedList<bool> newList = this.GetEmptyCompressedList();

                foreach (Range<long, bool> range in hiddenRanges)
                {
                    int startRowIndex = (int)range.Start;
                    int endRowIndex = (int)range.End;

                    for (int rowIndex = startRowIndex; rowIndex <= endRowIndex; rowIndex++)
                    {
                        if (rowIndex < absoluteRearrangeIndex)
                        {
                            newList.SetValue(rowIndex, true);
                        }
                        else if (rowIndex >= absoluteRearrangeIndex)
                        {
                            if (shiftType == ShiftType.Down)
                            {
                                newList.SetValue(rowIndex + delta, true);
                            }
                            else if (rowIndex > absoluteRearrangeIndex + delta)
                            {
                                newList.SetValue(rowIndex - delta, true);
                            }
                        }
                    }
                }

                newIndexToRowStateDictionary.Add(pair.Key, newList);
            }

            this.relativeIndexToHiddenRowsState = newIndexToRowStateDictionary;
        }

        private void RearrangeHiddenRowsDictionary(int absoluteRearrangeIndex, int delta, ShiftType shiftType)
        {
            Dictionary<int, int> newHiddenRowsDictionary = new Dictionary<int, int>();

            foreach (var pair in this.hiddenRows)
            {
                int row = pair.Key;

                if (row < absoluteRearrangeIndex)
                {
                    newHiddenRowsDictionary.Add(row, pair.Value);
                }
                else
                {
                    if (row < absoluteRearrangeIndex)
                    {
                        newHiddenRowsDictionary.Add(row, pair.Value);
                    }
                    else if (row >= absoluteRearrangeIndex)
                    {
                        if (shiftType == ShiftType.Down)
                        {
                            newHiddenRowsDictionary.Add(row + delta, pair.Value);
                        }
                        else if (row > absoluteRearrangeIndex + delta)
                        {
                            newHiddenRowsDictionary.Add(row - delta, pair.Value);
                        }
                    }
                }
            }

            this.hiddenRows = newHiddenRowsDictionary;
        }

        private CompressedList<bool> GetEmptyCompressedList()
        {
            return new CompressedList<bool>(0, SpreadsheetDefaultValues.RowCount - 1, false);
        }

        private void SetHiddenOutsideCommand(int row)
        {
            this.SetHiddenOutsideCommand(row, row);
        }

        private void SetHiddenOutsideCommand(int startRow, int endRow)
        {
            this.worksheet.Rows.PropertyBag.SetPropertyValue(RowColumnPropertyBagBase.HiddenProperty, startRow, endRow, true);
        }

        private void ClearHiddenOutsideCommand(int row)
        {
            this.ClearHiddenOutsideCommand(row, row);
        }

        private void ClearHiddenOutsideCommand(int startRow, int endRow)
        {
            this.worksheet.Rows.PropertyBag.ClearPropertyValue(RowColumnPropertyBagBase.HiddenProperty, startRow, endRow);
        }

        internal void CopyFrom(FilterManager other)
        {
            this.relativeIndexToHiddenRowsState = new Dictionary<int, ICompressedList<bool>>(other.relativeIndexToHiddenRowsState);
            this.hiddenRows = new Dictionary<int, int>(other.hiddenRows);

            this.relativeIndexToHiddenRowsState = new Dictionary<int, ICompressedList<bool>>();
            foreach (KeyValuePair<int, ICompressedList<bool>> pair in other.relativeIndexToHiddenRowsState)
            {
                ICompressedList<bool> source = pair.Value;

                CompressedList<bool> copy = new CompressedList<bool>(source);
                this.relativeIndexToHiddenRowsState[pair.Key] = copy;

                foreach (Range<long, bool> range in source.GetNonDefaultRanges())
                {
                    copy.SetValue(range.Start, range.End, range.Value);
                }
            }
        }
    }
}
