﻿using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
#if NETCORE || NETSTANDARD
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.Model.Filtering
{
    /// <summary>
    /// Describes a filter which hides or shows the cells based on the cell fore color.
    /// </summary>
    public class ForeColorFilter : FilterBase<ThemableColor>, IWorksheetFilter
    {
        private readonly ThemableColor color;
        private ThemableColor actualColor;
        private Worksheet worksheet;

        /// <summary>
        /// Gets the fore color which the cells displayed by the filter should have.
        /// </summary>
        /// <value>The fore color.</value>
        public ThemableColor Color
        {
            get
            {
                return this.color;
            }
        }

        void IWorksheetFilter.SetWorksheet(Worksheet worksheet)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");

            this.worksheet = worksheet;
            this.actualColor = SortAndFilterHelper.GetThemableColorWithLocalColors(this.color, this.worksheet.Workbook.Theme.ColorScheme);
        }

        /// <summary>
        /// Gets the property definition for the value which will be retrieved by the GetValue method.
        /// </summary>
        /// <value>The property definition.</value>
        protected override IPropertyDefinition<ThemableColor> PropertyDefinition
        {
            get
            {
                return CellPropertyDefinitions.ForeColorProperty;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ForeColorFilter" /> class.
        /// </summary>
        /// <param name="relativeColumnIndex">Index of the relative column.</param>
        /// <param name="color">The color.</param>
        public ForeColorFilter(int relativeColumnIndex, ThemableColor color) : base(relativeColumnIndex)
        {
            Guard.ThrowExceptionIfNull(color, "color");

            this.color = color;
        }

        internal override IFilter Copy(int newRelativeColumnIndex)
        {
            return new ForeColorFilter(newRelativeColumnIndex, this.Color);
        }

        /// <summary>
        /// Determines whether the row which contains the specified value will be shown.
        /// </summary>
        /// <param name="value">The value retrieved by the GetValue method.</param>
        /// <returns>
        /// A value indicating whether the row which contains the specified value will be shown
        /// </returns>
        public override bool ShouldShowValue(object value)
        {
            ThemableColor colorValue = value as ThemableColor;

            Guard.ThrowExceptionIfNull(this.worksheet, "worksheet");
            colorValue = SortAndFilterHelper.GetThemableColorWithLocalColors(colorValue, this.worksheet.Workbook.Theme.ColorScheme);

            return colorValue.Equals(this.actualColor);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            ForeColorFilter other = obj as ForeColorFilter;

            if (other == null)
            {
                return false;
            }

            return this.Color.Equals(other.Color);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return this.GetHashCode();
        }
    }
}