﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Filtering
{
    /// <summary>
    /// Describes a filter which hides or shows the cells based on the cell fill color.
    /// </summary>
    public class TopFilter : CellValuesFilterBase, IWorksheetFilter, IRangeFilter
    {
        private HashSet<double> valuesToShow;
        private readonly TopFilterType topFilterType;
        private readonly double value;
        private Worksheet worksheet;

        /// <summary>
        /// Gets the type of the top filter.
        /// </summary>
        /// <value>The type of the top filter.</value>
        public TopFilterType TopFilterType
        {
            get
            {
                return this.topFilterType;
            }
        }

        /// <summary>
        /// Gets the value which indicates how many or what percent of the items covered by the filter should be shown.
        /// </summary>
        /// <value>The value.</value>
        public double Value
        {
            get
            {
                return this.value;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TopFilter" /> class.
        /// </summary>
        /// <param name="relativeColumnIndex">The relative index of the column.</param>
        public TopFilter(int relativeColumnIndex) : this(relativeColumnIndex, TopFilterType.TopNumber, 10)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TopFilter" /> class.
        /// </summary>
        /// <param name="relativeColumnIndex">The relative index of the column.</param>
        /// <param name="topFilterType">Type of the top filter.</param>
        /// <param name="value">Gets the value which indicates how many or what percent of the items covered by the filter should be shown.</param>
        public TopFilter(int relativeColumnIndex, TopFilterType topFilterType, double value) : base(relativeColumnIndex)
        {
            this.topFilterType = topFilterType;
            this.value = value;
        }

        internal override IFilter Copy(int newRelativeColumnIndex)
        {
            return new TopFilter(newRelativeColumnIndex, this.TopFilterType, this.Value);
        }

        /// <summary>
        /// Determines whether the row which contains the specified value will be shown.
        /// </summary>
        /// <param name="value">The value retrieved by the GetValue method.</param>
        /// <returns>
        /// A value indicating whether the row which contains the specified value will be shown
        /// </returns>
        public override bool ShouldShowValue(object value)
        {
            if (this.valuesToShow == null)
            {
                throw new FilteringException("This filter is not yet assigned to a column.", 
                    new InvalidOperationException("This filter is not yet assigned to a column."), 
                    "Spreadsheet_Filtering_RangeFilterNotAssigned");
            }

            FormulaCellValue formulaCellValue = value as FormulaCellValue;

            if (formulaCellValue != null)
            {
                value = formulaCellValue.GetResultValueAsCellValue();
            }

            NumberCellValue numberCellValue = value as NumberCellValue;

            if (numberCellValue == null)
            {
                return false;
            }

            if (valuesToShow.Contains(numberCellValue.Value))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        void IWorksheetFilter.SetWorksheet(Worksheet worksheet)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");

            this.worksheet = worksheet;
        }

        void IRangeFilter.SetColumnRange(CellRange columnRange)
        {
            this.SetColumnRange(columnRange);
        }

        private void SetColumnRange(CellRange columnRange)
        {
            Guard.ThrowExceptionIfNull(this.worksheet, "worksheet");
            if (columnRange != null)
            {
                Guard.ThrowExceptionIfNotEqual(1, columnRange.ColumnCount, "range");
            }

            this.valuesToShow = this.GetValuesToBeShownArray(this.worksheet, columnRange, topFilterType, value);
        }

        private HashSet<double> GetValuesToBeShownArray(Worksheet worksheet, CellRange range, TopFilterType topFilterType, double value)
        {
            if (range == null)
            {
                return new HashSet<double>();
            }

            List<double> result = new List<double>();

            for (int row = range.FromIndex.RowIndex; row <= range.ToIndex.RowIndex; row++)
            {
                long index = CellsPropertyBag.ConvertCellIndexToLong(row, range.FromIndex.ColumnIndex);
                ICellValue cellValue = worksheet.Cells.PropertyBag.GetPropertyValue(CellPropertyDefinitions.ValueProperty, index);
                FormulaCellValue formulaCellValue = cellValue as FormulaCellValue;

                if (formulaCellValue != null)
                {
                    cellValue = formulaCellValue.GetResultValueAsCellValue();
                }

                NumberCellValue numberCellValue = cellValue as NumberCellValue;

                if (numberCellValue != null)
                {
                    result.Add(numberCellValue.Value);
                }
            }

            result.Sort();

            if (topFilterType == TopFilterType.TopNumber ||
                topFilterType == TopFilterType.TopPercent)
            {
                result.Reverse();
            }

            int numberOfValues;

            if (topFilterType == TopFilterType.TopPercent ||
                topFilterType == TopFilterType.BottomPercent)
            {
                numberOfValues = (int)(result.Count * value / 100);
            }
            else
            {
                numberOfValues = (int)value;
            }

            return new HashSet<double>(result.Take(numberOfValues));
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            TopFilter other = obj as TopFilter;

            if (other == null)
            {
                return false;
            }

            return this.TopFilterType.Equals(other.TopFilterType) &&
                this.Value.Equals(other.Value);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                this.TopFilterType.GetHashCode(),
                this.Value.GetHashCode());
        }
    }
}
