﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Filtering
{
    /// <summary>
    /// Describes a filter which hides or shows the cells based on whether the result value of the cell can be found in a predefined list of values.
    /// </summary>
    public class ValuesCollectionFilter : CellValuesFilterBase
    {
        private readonly string[] stringValues;
        private readonly DateGroupItem[] dateValues;
        private readonly bool blank;

        /// <summary>
        /// Gets the string values of the filter.
        /// </summary>
        /// <value>The string values.</value>
        public IEnumerable<string> StringValues
        {
            get
            {
                return this.stringValues;
            }
        }

        /// <summary>
        /// Gets the date items of the filter.
        /// </summary>
        /// <value>The date items.</value>
        public IEnumerable<DateGroupItem> DateItems
        {
            get
            {
                return this.dateValues;
            }
        }

        /// <summary>
        /// Gets the value indicating whether the cells containing blank values should be shown by the filter.
        /// </summary>
        /// <value>The value indicating whether the cells containing blank values should be shown by the filter.</value>
        public bool Blank
        {
            get
            {
                return this.blank;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ValuesCollectionFilter" /> class.
        /// </summary>
        /// <param name="relativeColumnIndex">Relative index of the column.</param>
        /// <param name="stringValues">The string values.</param>
        public ValuesCollectionFilter(int relativeColumnIndex, IEnumerable<string> stringValues) : this(relativeColumnIndex, stringValues, null, false)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ValuesCollectionFilter" /> class.
        /// </summary>
        /// <param name="relativeColumnIndex">Relative index of the column.</param>
        /// <param name="dateValues">The date values.</param>
        public ValuesCollectionFilter(int relativeColumnIndex, IEnumerable<DateGroupItem> dateValues) : this(relativeColumnIndex, null, dateValues, false)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ValuesCollectionFilter" /> class.
        /// </summary>
        /// <param name="relativeColumnIndex">Relative index of the column.</param>
        /// <param name="stringValues">The string values.</param>
        /// <param name="blank">The value indicating whether the cells containing blank values should be shown by the filter.</param>
        public ValuesCollectionFilter(int relativeColumnIndex, IEnumerable<string> stringValues, bool blank) : this(relativeColumnIndex, stringValues, null, blank)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ValuesCollectionFilter" /> class.
        /// </summary>
        /// <param name="relativeColumnIndex">Relative index of the column.</param>
        /// <param name="dateValues">The date values.</param>
        /// <param name="blank">The value indicating whether the cells containing blank values should be shown by the filter.</param>
        public ValuesCollectionFilter(int relativeColumnIndex, IEnumerable<DateGroupItem> dateValues, bool blank) : this(relativeColumnIndex, null, dateValues, blank)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ValuesCollectionFilter" /> class.
        /// </summary>
        /// <param name="relativeColumnIndex">Relative index of the column.</param>
        /// <param name="stringValues">The string values.</param>
        /// <param name="dateValues">The date values.</param>
        public ValuesCollectionFilter(int relativeColumnIndex, IEnumerable<string> stringValues, IEnumerable<DateGroupItem> dateValues) : this(relativeColumnIndex, stringValues, dateValues, false)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ValuesCollectionFilter" /> class.
        /// </summary>
        /// <param name="relativeColumnIndex">Relative index of the column.</param>
        /// <param name="stringValues">The string values.</param>
        /// <param name="dateValues">The date values.</param>
        /// <param name="blank">The value indicating whether the cells containing blank values should be shown by the filter.</param>
        public ValuesCollectionFilter(int relativeColumnIndex, IEnumerable<string> stringValues, IEnumerable<DateGroupItem> dateValues, bool blank) : base(relativeColumnIndex)
        {
            this.dateValues = dateValues == null ? new DateGroupItem[0] : dateValues.ToArray();
            this.stringValues = stringValues == null ? new string[0] : stringValues.ToArray();
            this.blank = blank;
        }

        internal override IFilter Copy(int newRelativeColumnIndex)
        {
            return new ValuesCollectionFilter(newRelativeColumnIndex, this.StringValues, this.DateItems, this.Blank);
        }

        /// <summary>
        /// Gets the value of the cell at the specified index. This value is used to determine whether the row will be hidden by the filter.
        /// </summary>
        /// <param name="cells">The cells of the worksheet.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The value of the cell.</returns>
        public override object GetValue(Cells cells, int rowIndex, int columnIndex)
        {
            ICellValue value = base.GetValue(cells, rowIndex, columnIndex) as ICellValue;

            long index = CellsPropertyBag.ConvertCellIndexToLong(rowIndex, columnIndex);
            CellValueFormat format = cells.PropertyBag.GetPropertyValueRespectingStyle(CellPropertyDefinitions.FormatProperty, cells.Worksheet, index);

            FormulaCellValue formulaCellValue = value as FormulaCellValue;

            if (formulaCellValue != null)
            {
                value = formulaCellValue.GetResultValueAsCellValue();
            }

            NumberCellValue numberCellValue = value as NumberCellValue;
            DateTime? result;

            if (format.IsDateFormat() &&
                numberCellValue != null &&
                numberCellValue.Value >= 1)
            {
                result = numberCellValue.ToDateTime();
                if (result.HasValue)
                {
                    return result.Value;
                }
            }

            return value.GetResultValueAsString(format);
        }

        /// <summary>
        /// Determines whether the row which contains the specified value will be shown.
        /// </summary>
        /// <param name="value">The value retrieved by the GetValue method.</param>
        /// <returns>
        /// A value indicating whether the row which contains the specified value will be shown
        /// </returns>
        public override bool ShouldShowValue(object value)
        {
            string stringValue = value as string;

            if (stringValue != null)
            {
                return ShouldShowValue(stringValue);
            }

            DateTime? dateTimeValue = value as DateTime?;

            if (dateTimeValue.HasValue)
            {
                return ShouldShowValue(dateTimeValue.Value);
            }

            return false;
        }

        private bool ShouldShowValue(string value)
        {
            if (this.stringValues.Any(x => string.Equals(x, value, StringComparison.OrdinalIgnoreCase)))
            {
                return true;
            }

            if (string.IsNullOrWhiteSpace(value) && this.blank)
            {
                return true;
            }

            return false;
        }

        private bool ShouldShowValue(DateTime value)
        {
            if (this.dateValues.Any(x => x.DateSatisfiesItem(value)))
            {
                return true;
            }

            return false;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            ValuesCollectionFilter other = obj as ValuesCollectionFilter;

            if (other == null)
            {
                return false;
            }

            return this.Blank.Equals(other.Blank) &&
                Enumerable.SequenceEqual(this.StringValues, other.StringValues) &&
                Enumerable.SequenceEqual(this.DateItems, other.DateItems);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                this.Blank.GetHashCode(),
                this.StringValues.GetHashCodeOrZero(),
                this.DateItems.GetHashCodeOrZero());
        }
    }
}