﻿using System.Collections.Generic;
using System.Text.RegularExpressions;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Describes the find options.
    /// </summary>
    public class FindOptions
    {
        #region Fields

        private string findWhat;
        private FindWithin findWithin;
        private FindBy findBy;
        private FindInContentType findIn;
        private bool matchCase;
        private bool matchEntireCellContents;
        private WorksheetCellIndex startCell;
        private IEnumerable<CellRange> searchRange;

        #endregion


        #region Properties

        /// <summary>
        /// Gets or sets the searched string.
        /// </summary>
        /// <value>The searched string.</value>
        public string FindWhat
        {
            get
            {
                if (this.findWhat == null)
                {
                    this.findWhat = string.Empty;
                }

                return this.findWhat;
            }
            set
            {
                if (this.findWhat != value)
                {
                    this.findWhat = value;
                }

                this.SetFindWhatRegex();
            }
        }

        /// <summary>
        /// Gets the searched string as regex.
        /// </summary>
        /// <value>The searched string regex.</value>
        internal string FindWhatRegex { get; private set; }

        /// <summary>
        /// Gets the find what regex options.
        /// </summary>
        /// <value>The find what regex options.</value>
        internal RegexOptions FindWhatRegexOptions { get; private set; }

        /// <summary>
        /// Gets or sets the find within option.
        /// </summary>
        /// <value>The find within option.</value>
        public FindWithin FindWithin
        {
            get
            {
                return this.findWithin;
            }
            set
            {
                if (this.findWithin != value)
                {
                    this.findWithin = value;
                }
            }
        }

        /// <summary>
        /// Gets or sets the find by option.
        /// </summary>
        /// <value>The find by option.</value>
        public FindBy FindBy
        {
            get
            {
                return this.findBy;
            }
            set
            {
                if (this.findBy != value)
                {
                    this.findBy = value;
                }
            }
        }

        /// <summary>
        /// Gets or sets the find in option.
        /// </summary>
        /// <value>The find in option.</value>
        public FindInContentType FindIn
        {
            get
            {
                return this.findIn;
            }
            set
            {
                if (this.findIn != value)
                {
                    this.findIn = value;
                }
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the match case option should be used.
        /// </summary>
        /// <value>The value indicating whether the match case option should be used.</value>
        public bool MatchCase
        {
            get
            {
                return this.matchCase;
            }
            set
            {
                if (this.matchCase != value)
                {
                    this.matchCase = value;
                }

                this.FindWhatRegexOptions = this.matchCase ? RegexOptions.None : RegexOptions.IgnoreCase;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the option to match entire cell contents should be used.
        /// </summary>
        /// <value>The value indicating whether the option to match entire cell contents should be used.</value>
        public bool MatchEntireCellContents
        {
            get
            {
                return this.matchEntireCellContents;
            }
            set
            {
                if (this.matchEntireCellContents != value)
                {
                    this.matchEntireCellContents = value;
                }

                this.SetFindWhatRegex();
            }
        }

        /// <summary>
        /// Gets or sets the start cell.
        /// </summary>
        /// <value>The start cell.</value>
        public WorksheetCellIndex StartCell
        {
            get
            {
                return this.startCell;
            }
            set
            {
                if (this.startCell != value)
                {
                    this.startCell = value;
                }
            }
        }

        /// <summary>
        /// Gets or sets the search ranges.
        /// </summary>
        /// <value>The search ranges.</value>
        public IEnumerable<CellRange> SearchRanges
        {
            get
            {
                return this.searchRange;
            }
            set
            {
                if (this.searchRange != value)
                {
                    this.searchRange = value;
                }
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="FindOptions" /> class.
        /// </summary>
        public FindOptions()
        {
            this.findWithin = FindWithin.Sheet;
            this.findBy = FindBy.Rows;
            this.findIn = FindInContentType.Formulas;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Copies the properties from the specified options instance.
        /// </summary>
        /// <param name="options">The options.</param>
        public void CopyPropertiesFrom(FindOptions options)
        {
            this.StartCell = options.StartCell;
            this.FindBy = options.FindBy;
            this.FindIn = options.FindIn;
            this.FindWhat = options.FindWhat;
            this.FindWithin = options.FindWithin;
            this.MatchCase = options.MatchCase;
            this.MatchEntireCellContents = options.MatchEntireCellContents;
            this.SearchRanges = options.SearchRanges;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            FindOptions other = obj as FindOptions;

            if (other == null)
            {
                return false;
            }

            return this.StartCell.Equals(other.StartCell) &&
                this.FindWhat.Equals(other.FindWhat) &&
                this.FindWithin.Equals(other.FindWithin) &&
                this.FindBy.Equals(other.FindBy) &&
                this.FindIn.Equals(other.FindIn) &&
                this.MatchCase.Equals(other.MatchCase) &&
                this.MatchEntireCellContents.Equals(this.MatchEntireCellContents);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(this.StartCell.GetHashCodeOrZero(), this.FindWhat.GetHashCodeOrZero(), this.FindWithin.GetHashCodeOrZero(),
                this.FindBy.GetHashCodeOrZero(), this.FindIn.GetHashCodeOrZero(), this.MatchCase.GetHashCodeOrZero(), this.MatchEntireCellContents.GetHashCodeOrZero());
        }

        /// <summary>
        /// Sets the search string regex.
        /// </summary>
        private void SetFindWhatRegex()
        {
            string format = this.MatchEntireCellContents ? "^{0}$" : "{0}";
            this.FindWhatRegex = string.Format(format, this.FindWhat.FromWildcardStringToRegex());
        }

        #endregion
    }
}