﻿using System.Windows;
using System.Windows.Media;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
#if NETCORE || NETSTANDARD
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Describes the properties of a font.
    /// </summary>
    public struct FontProperties : IFontProperties
    {
        #region Fields

        private bool isMonospaced;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the font family.
        /// </summary>
        /// <value>The font family.</value>
        public FontFamily FontFamily { get; internal set; }

        /// <summary>
        /// Gets the size of the font.
        /// </summary>
        /// <value>The size of the font.</value>
        public double FontSize { get; internal set; }

        /// <summary>
        /// Gets a value indicating whether the font is bold.
        /// </summary>
        /// <value>The value indicating whether the font is bold.</value>
        public bool IsBold { get; internal set; }

        /// <summary>
        /// Gets a value indicating whether the font is italic.
        /// </summary>
        /// <value>The value indicating whether the font is italic.</value>
        public bool IsItalic { get; internal set; }

        /// <summary>
        /// Gets the underline.
        /// </summary>
        /// <value>The underline.</value>
        public UnderlineType Underline { get; internal set; }

        /// <summary>
        /// Gets the font color.
        /// </summary>
        /// <value>The font color.</value>
        public ThemableColor ForeColor { get; internal set; }

        /// <summary>
        /// Gets the font weight.
        /// </summary>
        /// <value>The font weight.</value>
        public FontWeight FontWeight
        {
            get
            {
                return this.IsBold ? FontWeights.Bold : FontWeights.Normal;
            }
        }

        /// <summary>
        /// Gets the font style.
        /// </summary>
        /// <value>The font style.</value>
        public FontStyle FontStyle
        {
            get
            {
                return this.IsItalic ? FontStyles.Italic : FontStyles.Normal;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the text is monospaced.
        /// </summary>
        /// <value>The value indicating whether the text is monospaced.</value>
        public bool IsMonospaced
        {
            get
            {
                return this.isMonospaced || Telerik.Windows.Documents.Core.Fonts.SystemFontsManager.IsMonospaced(this.FontFamily.Source);
            }
            set
            {
                this.isMonospaced = value;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="FontProperties" /> struct.
        /// </summary>
        /// <param name="fontFamily">The font family.</param>
        /// <param name="fontSize">Size of the font.</param>
        /// <param name="isBold">The value indicating whether the font is bold.</param>
        public FontProperties(FontFamily fontFamily, double fontSize, bool isBold)
            : this()
        {
            this.FontFamily = fontFamily;
            this.FontSize = fontSize;
            this.IsBold = isBold;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            if (obj is FontProperties)
            {
                return (this == (FontProperties)obj);
            }

            return false;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                this.FontFamily.GetHashCode(),
                this.FontSize.GetHashCode(),
                this.IsBold.GetHashCode(),
                this.IsItalic.GetHashCode(),
                this.Underline.GetHashCode(),
                this.ForeColor.GetHashCode());
        }

        /// <summary>
        /// Implements the equals operator.
        /// </summary>
        /// <param name="left">The left value.</param>
        /// <param name="right">The right value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
        public static bool operator ==(FontProperties left, FontProperties right)
        {
            return left.FontFamily.GetSourceOrNull() == right.FontFamily.GetSourceOrNull() &&
                left.FontSize == right.FontSize &&
                left.IsBold == right.IsBold &&
                left.IsItalic == right.IsItalic &&
                left.Underline == right.Underline &&
                left.ForeColor == right.ForeColor;
        }

        /// <summary>
        /// Implements the not equals operator.
        /// </summary>
        /// <param name="left">The left value.</param>
        /// <param name="right">The right value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
        public static bool operator !=(FontProperties left, FontProperties right)
        {
            return !(left == right);
        }

        #endregion
    }
}