﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a collection of hyperlinks.
    /// </summary>
    public class HyperlinkCollection : CellRangeInsertedOrRemovedAwareCollection<SpreadsheetHyperlink>
    {
        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="HyperlinkCollection" /> class.
        /// </summary>
        /// <param name="worksheet">The worksheet.</param>
        public HyperlinkCollection(Worksheet worksheet)
            : base(worksheet)
        {
        }

        #endregion


        #region Methods

        internal void Sort(CellRange sortedRange, int[] sortedIndexes)
        {
            Guard.ThrowExceptionIfNotEqual(sortedRange.RowCount, sortedIndexes.Length, "RowCount");
            // TODO: Sort by rows???

            this.OnChanging();

            SpreadsheetHyperlink[] hyperlinks = this.GetContainingHyperlinks(sortedRange).Where(p => p.Range.RowCount == 1).ToArray();

            List<SpreadsheetHyperlink>[] hyperlinksToSort = new List<SpreadsheetHyperlink>[sortedIndexes.Length];

            for (int i = 0; i < hyperlinksToSort.Length; i++)
            {
                hyperlinksToSort[i] = new List<SpreadsheetHyperlink>();
            }

            int fromIndex = sortedRange.FromIndex.RowIndex;
            int toIndex = sortedRange.ToIndex.RowIndex;

            for (int row = fromIndex; row <= toIndex; row++)
            {
                hyperlinksToSort[row - fromIndex].AddRange(hyperlinks.Where(p => p.Range.FromIndex.RowIndex == row));
            }

            foreach(var hyperlink in hyperlinks)
            {
                this.Elements.Remove(hyperlink.Range);
            }

            for (int i = 0; i < sortedIndexes.Length; i++)
            {
                foreach (var hyperlink in hyperlinksToSort[sortedIndexes[i]])
                {
                    int newRowIndex = i + fromIndex;
                    CellRange newCellRange = new CellRange(newRowIndex, hyperlink.Range.FromIndex.ColumnIndex, newRowIndex, hyperlink.Range.ToIndex.ColumnIndex);
                    SpreadsheetHyperlink newHyperlink = new SpreadsheetHyperlink(newCellRange, hyperlink.HyperlinkInfo);
                    this.Elements.Add(newCellRange, newHyperlink);
                }
            }

            this.OnChanged();
        }

        /// <summary>
        /// Adds a new hyperlink to the collection.
        /// </summary>
        /// <param name="cellIndex">Index of the cell.</param>
        /// <param name="hyperlinkInfo">The hyperlink info.</param>
        /// <returns>The new hyperlink.</returns>
        public SpreadsheetHyperlink Add(CellIndex cellIndex, HyperlinkInfo hyperlinkInfo)
        {
            return this.Add(cellIndex.ToCellRange(), hyperlinkInfo);
        }

        /// <summary>
        /// Adds a new hyperlink to the collection.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <param name="hyperlinkInfo">The hyperlink info.</param>
        /// <returns>The new hyperlink.</returns>
        public SpreadsheetHyperlink Add(CellRange cellRange, HyperlinkInfo hyperlinkInfo)
        {
            SpreadsheetHyperlink hyperlink = new SpreadsheetHyperlink(cellRange, hyperlinkInfo);

            AddRemoveHyperlinkCommandContext addContext = new AddRemoveHyperlinkCommandContext(this.Worksheet, hyperlink);

            using (new UpdateScope(this.Worksheet.BeginUndoGroup, this.Worksheet.EndUndoGroup))
            {
                SpreadsheetHyperlink hyperlinkToRemove;
                if (this.TryGetHyperlinkExact(cellRange, out hyperlinkToRemove))
                {
                    AddRemoveHyperlinkCommandContext removeContext = new AddRemoveHyperlinkCommandContext(this.Worksheet, hyperlinkToRemove);
                    this.Worksheet.ExecuteCommand(WorkbookCommands.RemoveHyperlink, removeContext);
                }
                this.Worksheet.ExecuteCommand(WorkbookCommands.AddHyperlink, addContext);
                this.Worksheet.Cells[cellRange].SetStyleNameAndExpandToFitNumberValuesWidth(SpreadsheetDefaultValues.HyperlinkStyleName);
            }

            return hyperlink;
        }

        /// <summary>
        /// Removes the specified hyperlink.
        /// </summary>
        /// <param name="hyperlink">The hyperlink.</param>
        /// <returns>The value indicating whether the remove was successful.</returns>
        public bool Remove(SpreadsheetHyperlink hyperlink)
        {
            return this.Remove(hyperlink, true);
        }

        internal bool Remove(SpreadsheetHyperlink hyperlink, bool removeFormatting)
        {
            AddRemoveHyperlinkCommandContext context = new AddRemoveHyperlinkCommandContext(this.Worksheet, hyperlink);

            bool hasRemove = false;
            using (new UpdateScope(this.Worksheet.BeginUndoGroup, this.Worksheet.EndUndoGroup))
            {
                hasRemove = this.Worksheet.ExecuteCommand(WorkbookCommands.RemoveHyperlink, context);
                if (removeFormatting && hasRemove)
                {
                    this.Worksheet.Cells[context.Hyperlink.Range].SetStyleNameAndExpandToFitNumberValuesWidth(SpreadsheetDefaultValues.DefaultStyleName);
                }
            }

            return hasRemove;
        }

        internal bool RemoveRange(IEnumerable<SpreadsheetHyperlink> hyperlinks, bool removeFormatting)
        {
            bool hasRemove = true;

            using (new UpdateScope(this.Worksheet.BeginUndoGroup, this.Worksheet.EndUndoGroup))
            {
                foreach (SpreadsheetHyperlink hyperlink in hyperlinks)
                {
                    hasRemove &= this.Remove(hyperlink, removeFormatting);
                    if (!hasRemove)
                    {
                        break;
                    }
                }
            }

            return hasRemove;
        }

        /// <summary>
        /// Determines whether the collection contains the specified hyperlink.
        /// </summary>
        /// <param name="hyperlink">The hyperlink.</param>
        /// <returns>A value indicating whether the collection contains the specified hyperlink.</returns>
        public bool Contains(SpreadsheetHyperlink hyperlink)
        {
            return this.Elements.ContainsKey(hyperlink.Range);
        }

        internal IEnumerable<SpreadsheetHyperlink> GetIntersectingHyperlinks(int rowIndex, int columnIndex)
        {
            return this.Elements.Values.Where(sh => sh.Range.IntersectsWith(rowIndex, columnIndex, rowIndex, columnIndex));
        }

        /// <summary>
        /// Gets the hyperlinks intersecting with the specified cell range.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <returns>The collection of intersecting hyperlinks.</returns>
        public IEnumerable<SpreadsheetHyperlink> GetIntersectingHyperlinks(CellRange cellRange)
        {
            return this.Elements.Values.Where(sh => sh.Range.IntersectsWith(cellRange));
        }

        /// <summary>
        /// Gets the hyperlinks contained in the specified cell range.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <returns>The collection of contained hyperlinks.</returns>
        public IEnumerable<SpreadsheetHyperlink> GetContainingHyperlinks(CellRange cellRange)
        {
            return this.Elements.Values.Where(sh => cellRange.Contains(sh.Range));
        }

        /// <summary>
        /// Gets the hyperlinks contained in the specified cell ranges.
        /// </summary>
        /// <param name="cellRanges">The cell ranges.</param>
        /// <returns>The collection of contained hyperlinks.</returns>
        public IEnumerable<SpreadsheetHyperlink> GetContainingHyperlinks(IEnumerable<CellRange> cellRanges)
        {
            List<SpreadsheetHyperlink> result = new List<SpreadsheetHyperlink>();
            foreach (CellRange cellRange in cellRanges)
            {
                result.AddRange(this.Worksheet.Hyperlinks.GetContainingHyperlinks(cellRange));
            }

            return result;
        }

        /// <summary>
        /// Tries to get a hyperlink from the specified cell range.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <param name="hyperlink">The hyperlink.</param>
        /// <returns>The value indicating whether a hyperlink was found.</returns>
        public bool TryGetHyperlink(CellRange cellRange, out SpreadsheetHyperlink hyperlink)
        {
            hyperlink = this.GetIntersectingHyperlinks(cellRange).LastOrDefault();
            return hyperlink != null;
        }

        /// <summary>
        /// Tries to get a hyperlink from the specified cell index.
        /// </summary>
        /// <param name="cellIndex">Index of the cell.</param>
        /// <param name="hyperlink">The hyperlink.</param>
        /// <returns>The value indicating whether a hyperlink was found.</returns>
        public bool TryGetHyperlink(CellIndex cellIndex, out SpreadsheetHyperlink hyperlink)
        {
            return this.TryGetHyperlink(cellIndex.RowIndex, cellIndex.ColumnIndex, out hyperlink);
        }

        internal bool TryGetHyperlink(int rowIndex, int columnIndex, out SpreadsheetHyperlink hyperlink)
        {
            hyperlink = this.GetIntersectingHyperlinks(rowIndex, columnIndex).LastOrDefault();
            return hyperlink != null;
        }

        /// <summary>
        /// Tries to get the hyperlink from the specified cell range where the cell range has to match the hyperlink's cell range exactly.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <param name="hyperlink">The hyperlink.</param>
        /// <returns>The value indicating whether a hyperlink was found.</returns>
        public bool TryGetHyperlinkExact(CellRange cellRange, out SpreadsheetHyperlink hyperlink)
        {
            hyperlink = this.Elements.Values.Where(sh => sh.Range.Equals(cellRange)).LastOrDefault();
            return hyperlink != null;
        }

        internal void AddInternal(SpreadsheetHyperlink hyperlink)
        {
            Guard.ThrowExceptionIfNull(hyperlink, "hyperlink");

            this.OnChanging();

            this.Elements.Add(hyperlink.Range, hyperlink);

            this.OnChanged();
        }

        internal void RemoveInternal(SpreadsheetHyperlink hyperlink)
        {
            Guard.ThrowExceptionIfNull(hyperlink, "hyperlink");

            this.OnChanging();

            this.Elements.Remove(hyperlink.Range);

            this.OnChanged();
        }

        /// <summary>
        /// Adds the specified element when cellRange is inserted.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <param name="element">The element.</param>
        /// <returns></returns>
        protected override SpreadsheetHyperlink Add(CellRange cellRange, SpreadsheetHyperlink element)
        {
            Guard.ThrowExceptionIfNull(cellRange, "cellRange");
            Guard.ThrowExceptionIfNull(element, "element");

            return this.Add(cellRange, element.HyperlinkInfo);
        }

        /// <summary>
        /// Removes elements when cellRange is removed.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <returns></returns>
        protected override bool Remove(CellRange cellRange)
        {
            SpreadsheetHyperlink hyperlink = this.Elements[cellRange];
            return this.Remove(hyperlink, true);
        }

        internal void CopyFrom(HyperlinkCollection fromHyperlinkCollection)
        {
            foreach (SpreadsheetHyperlink hyperlink in fromHyperlinkCollection)
            {
                this.Elements.Add(hyperlink.Range, hyperlink);
            }
        }

        internal void Clear()
        {
            this.Elements.Clear();
        }

        #endregion


        #region Events

        internal event EventHandler Changing;

        private void OnChanging()
        {
            if (this.Changing != null)
            {
                this.Changing(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Occurs when the collection is changed.
        /// </summary>
        public event EventHandler Changed;

        /// <summary>
        /// Called when the collection is changed.
        /// </summary>
        protected virtual void OnChanged()
        {
            if (this.Changed != null)
            {
                this.Changed(this, EventArgs.Empty);
            }
        }

        #endregion
    }
}