﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents information about a hyperlink.
    /// </summary>
    public class HyperlinkInfo
    {
        #region Fields

        private readonly string address;
        private readonly string emailSubject;
        private readonly string screenTip;
        private readonly string subAddress;
        private readonly HyperlinkType type;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the address.
        /// </summary>
        /// <value>The address.</value>
        public string Address
        {
            get
            {
                return this.address;
            }
        }

        /// <summary>
        /// Gets the email subject.
        /// </summary>
        /// <value>The email subject.</value>
        public string EmailSubject
        {
            get
            {
                return this.emailSubject;
            }
        }

        /// <summary>
        /// Gets the screen tip.
        /// </summary>
        /// <value>The screen tip.</value>
        public string ScreenTip
        {
            get
            {
                return this.screenTip;
            }
        }

        /// <summary>
        /// Gets the sub address.
        /// </summary>
        /// <value>The sub address.</value>
        public string SubAddress
        {
            get
            {
                return this.subAddress;
            }
        }

        /// <summary>
        /// Gets the type of the hyperlink.
        /// </summary>
        /// <value>The type.</value>
        public HyperlinkType Type
        {
            get
            {
                return this.type;
            }
        }

        #endregion


        #region Constructors

        private HyperlinkInfo(string address, string subAddress, string emailSubject, HyperlinkType type, string screenTip = null)
        {
            this.address = address;
            this.subAddress = subAddress;
            this.emailSubject = emailSubject;
            this.screenTip = screenTip;
            this.type = type;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Creates a hyperlink.
        /// </summary>
        /// <param name="address">The address.</param>
        /// <param name="screenTip">The screen tip.</param>
        /// <returns>The hyperlink information.</returns>
        public static HyperlinkInfo CreateHyperlink(string address, string screenTip = null)
        {
            return new HyperlinkInfo(address, string.Empty, string.Empty, HyperlinkType.Url, screenTip);
        }

        /// <summary>
        /// Creates a mailto hyperlink.
        /// </summary>
        /// <param name="emailAddress">The email address.</param>
        /// <returns>The hyperlink information.</returns>
        public static HyperlinkInfo CreateMailtoHyperlink(string emailAddress)
        {
            return new HyperlinkInfo(emailAddress, string.Empty, string.Empty, HyperlinkType.Mailto);
        }

        /// <summary>
        /// Creates a mailto hyperlink.
        /// </summary>
        /// <param name="emailAddress">The email address.</param>
        /// <param name="emailSubject">The email subject.</param>
        /// <returns>The hyperlink information.</returns>
        public static HyperlinkInfo CreateMailtoHyperlink(string emailAddress, string emailSubject)
        {
            return new HyperlinkInfo(emailAddress, string.Empty, emailSubject, HyperlinkType.Mailto);
        }

        /// <summary>
        /// Creates a mailto hyperlink.
        /// </summary>
        /// <param name="emailAddress">The email address.</param>
        /// <param name="emailSubject">The email subject.</param>
        /// <param name="screenTip">The screen tip.</param>
        /// <returns>The hyperlink information.</returns>
        public static HyperlinkInfo CreateMailtoHyperlink(string emailAddress, string emailSubject, string screenTip)
        {
            return new HyperlinkInfo(emailAddress, string.Empty, emailSubject, HyperlinkType.Mailto, screenTip);
        }

        /// <summary>
        /// Creates an in-document hyperlink.
        /// </summary>
        /// <param name="subAddress">The sub address.</param>
        /// <param name="screenTip">The screen tip.</param>
        /// <returns>The hyperlink information.</returns>
        public static HyperlinkInfo CreateInDocumentHyperlink(string subAddress, string screenTip = null)
        {
            return new HyperlinkInfo(string.Empty, subAddress, string.Empty, HyperlinkType.InDocument, screenTip);
        }

        internal string GetHyperlinkAsString()
        {
            switch (this.type)
            {
                case HyperlinkType.Url:
                    return this.Address;
                case HyperlinkType.InDocument:
                    return this.SubAddress;
                case HyperlinkType.Mailto:
                    return this.GetEmailAddress();
                default:
                    throw new InvalidOperationException();
            }
        }

        internal string GetEmailAddress()
        {
            return GetEmailAddress(this.Address, this.emailSubject);
        }

        internal static string GetEmailAddress(string email, string emailSubject)
        {
            string emailAddress = string.Empty;
            if (!string.IsNullOrEmpty(email))
            {
                emailAddress += string.Format("mailto:{0}", email);
                if (!string.IsNullOrEmpty(emailSubject))
                {
                    emailAddress += string.Format("?Subject={0}", emailSubject);
                }
            }

            return emailAddress;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            HyperlinkInfo other = obj as HyperlinkInfo;

            if (other == null)
            {
                return false;
            }

            return TelerikHelper.EqualsOfT(this.Address, other.Address) &&
                   TelerikHelper.EqualsOfT(this.EmailSubject, other.EmailSubject) &&
                   TelerikHelper.EqualsOfT(this.ScreenTip, other.ScreenTip) &&
                   TelerikHelper.EqualsOfT(this.SubAddress, other.SubAddress) &&
                   this.Type.Equals(other.Type);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                this.Address.GetHashCodeOrZero(),
                this.EmailSubject.GetHashCodeOrZero(),
                this.ScreenTip.GetHashCodeOrZero(),
                this.SubAddress.GetHashCodeOrZero(),
                this.Type.GetHashCodeOrZero());
        }

        #endregion
    }
}
