﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    internal class MergedCellRanges : CellRangeInsertedOrRemovedAwareCollection
    {
        #region Fields

        private int beginUpdateCount;
        private CellRange changedCellRange;

        #endregion


        #region Properties

        public IEnumerable<CellRange> Ranges
        {
            get
            {
                return this.CellRanges;
            }
        }

        #endregion


        #region Constructors

        public MergedCellRanges(Cells cells)
            :base(cells)
        {
        }

        #endregion


        #region Methods

        protected override void TranslateRanges(Dictionary<CellRange, CellRange> oldAndNewTranslatedRangesPositions)
        {
            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            using (new UpdateScope(this.Cells.Worksheet.BeginUndoGroup, this.Cells.Worksheet.EndUndoGroup))
            {
                this.Cells[oldAndNewTranslatedRangesPositions.Keys].Unmerge();

                foreach (CellRange translatedRange in oldAndNewTranslatedRangesPositions.Values)
                {
                    if (translatedRange != null)
                    {
                        this.Cells[translatedRange].Merge();
                    }
                }
            }
        }

        protected override void Add(CellRange cellRange)
        {
            base.Add(cellRange);
            this.UpdateChangedCellRange(cellRange);
        }

        protected override void Remove(CellRange cellRange)
        {
            base.Remove(cellRange);
            this.UpdateChangedCellRange(cellRange);
        }

        internal bool CanInsertOrRemoveInternal(ShiftType shiftType, CellRange range, CellRange selectedRange)
        {
            return this.CanInsertOrRemove(shiftType, range, selectedRange);
        }

        internal bool IsSingleMergedRange(CellRange range)
        {
            Guard.ThrowExceptionIfNull(range, "range");

            CellRange mergedRange;
            if (this.TryGetContainingMergedRange(range.ToIndex, out mergedRange))
            {
                return mergedRange.Equals(range);
            }

            return false;
        }        

        /// <summary>
        /// Determines whether this instance can insert or remove the selected range depending on some existing range and a given shift type.
        /// </summary>
        /// <param name="shiftType">Type of the shift.</param>
        /// <param name="range">The range.</param>
        /// <param name="selectedRange">The selected range.</param>
        /// <returns>True if the selected range can be inserted or removed.</returns>
        protected override bool CanInsertOrRemove(ShiftType shiftType, CellRange range, CellRange selectedRange)
        {
            return range.CanInsertOrRemove(shiftType, selectedRange);
        }

        private static ISet<CellRange> GetIntersectingMergedRanges(CellRange cellRange, IEnumerable<CellRange> mergedCellRanges)
        {
            HashSet<CellRange> intersectingRanges = new HashSet<CellRange>();

            foreach (var range in mergedCellRanges)
            {
                if (cellRange.IntersectsWith(range))
                {
                    intersectingRanges.Add(range);
                }
            }

            return intersectingRanges;
        }

        private ISet<CellRange> GetIntersectingMergedRanges(int rowIndex, int columnIndex, ICollection<CellRange> mergedCellRanges)
        {
            HashSet<CellRange> intersectingRanges = new HashSet<CellRange>();

            foreach (var range in mergedCellRanges)
            {
                if (range.Contains(rowIndex, columnIndex))
                {
                    intersectingRanges.Add(range);
                }
            }

            return intersectingRanges;
        }

        public bool IntersectsWithMergedRanges(CellRange cellRange)
        {
            Guard.ThrowExceptionIfNull(cellRange, "cellRange");

            foreach (var range in this.CellRanges)
            {
                if (range.IntersectsWith(cellRange))
                {
                    return true;
                }
            }

            return false;
        }

        public ISet<CellRange> GetIntersectingMergedRanges(CellRange cellRange)
        {
            Guard.ThrowExceptionIfNull(cellRange, "cellRange");

            return GetIntersectingMergedRanges(cellRange, this.CellRanges);
        }

        private ISet<CellRange> GetIntersectingMergedRanges(int rowIndex, int columnIndex)
        {
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            return GetIntersectingMergedRanges(rowIndex, columnIndex, this.CellRanges);
        }

        public CellRange ExpandRangeToNotIntersectMergedCells(CellRange cellRange)
        {
            List<CellRange> mergedCellRangesCopy = new List<CellRange>(this.CellRanges);
            CellRange finalCellRange = cellRange;
            ISet<CellRange> intersectingRanges;
            do
            {
                intersectingRanges = GetIntersectingMergedRanges(finalCellRange, mergedCellRangesCopy);
                foreach (var range in intersectingRanges)
                {
                    mergedCellRangesCopy.Remove(range);
                    finalCellRange = CellRange.Max(finalCellRange, range);
                }
            } while (intersectingRanges.Count > 0);

            return finalCellRange;
        }

        /// <summary>
        /// Determines whether any of the merged ranges within a cell range are intersected 
        /// by a hidden row or column in a way which causes it to be split or reduced in size.
        /// If a merged range is entirely hidden, it is not considered split.
        /// </summary>
        /// <param name="cellRange">The cell range</param>
        /// <returns>True if any of the merged ranges within a cell range are split or reduced in size
        /// by hidden rows or columns. False if all merged ranges are entirely visible or entirely hidden.</returns>
        public bool MergedCellIsSplitByHiddenRowColumn(CellRange cellRange)
        {
            Worksheet worksheet = this.Cells.Worksheet;

            ISet<CellRange> mergedRanges = this.GetIntersectingMergedRanges(cellRange);

            bool isIntersected = false;
            foreach (var range in mergedRanges)
            {
                isIntersected |= worksheet.Rows[range].GetHidden().IsIndeterminate;
                isIntersected |= worksheet.Columns[range].GetHidden().IsIndeterminate;
            }

            return isIntersected;
        }

        public void AddMergedRange(CellRange cellRange)
        {
            Guard.ThrowExceptionIfNull(cellRange, "cellRange");

            if (cellRange.RowCount == 1 && cellRange.ColumnCount == 1)
            {
                throw new LocalizableException("Cannot merge a single cell.", 
                    new InvalidOperationException("Cannot merge a single cell."),
                        "Spreadsheet_ErrorExpressions_CantMergeSingle");
            }


            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                CellRange finalCellRange = cellRange;
                ISet<CellRange> intersectingRanges;
                do
                {
                    intersectingRanges = this.GetIntersectingMergedRanges(finalCellRange);
                    foreach (var range in intersectingRanges)
                    {
                        this.Remove(range);
                        finalCellRange = CellRange.Max(finalCellRange, range);
                    }
                } while (intersectingRanges.Count > 0);

                this.Add(finalCellRange);
            }
        }

        public void AddMergedRanges(IEnumerable<CellRange> cellRanges)
        {
            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                foreach (var range in cellRanges)
                {
                    this.AddMergedRange(range);
                }
            }
        }

        public void RemoveMergedRange(CellRange cellRange)
        {
            Guard.ThrowExceptionIfNull(cellRange, "cellRange");

            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                ISet<CellRange> intersectingRanges = this.GetIntersectingMergedRanges(cellRange);
                foreach (var range in intersectingRanges)
                {
                    this.Remove(range);
                }
            }
        }

        public void RemoveMergedRanges(IEnumerable<CellRange> cellRanges)
        {
            if (!cellRanges.Any())
            {
                return;
            }

            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                foreach (var range in cellRanges)
                {
                    this.RemoveMergedRange(range);
                }
            }
        }

        public bool TryGetContainingMergedRange(CellIndex cellIndex, out CellRange mergedRange)
        {
            Guard.ThrowExceptionIfNull(cellIndex, "cellIndex");

            return this.TryGetContainingMergedRange(cellIndex.RowIndex, cellIndex.ColumnIndex, out mergedRange);
        }

        internal bool TryGetContainingMergedRange(int rowIndex, int columnIndex, out CellRange mergedRange)
        {
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            ISet<CellRange> intersectingRanges = this.GetIntersectingMergedRanges(rowIndex, columnIndex);

            mergedRange = intersectingRanges.FirstOrDefault();
            return intersectingRanges.Count == 1;
        }

        public bool TryGetContainingMergedRange(CellRange cellRange, out CellRange mergedRange)
        {
            Guard.ThrowExceptionIfNull(cellRange, "cellRange");

            ISet<CellRange> intersectingRanges = this.GetIntersectingMergedRanges(cellRange);

            mergedRange = intersectingRanges.FirstOrDefault();
            return intersectingRanges.Count == 1;
        }

        public void BeginUpdate()
        {
            this.OnChanging();
            this.beginUpdateCount++;
        }

        public void EndUpdate()
        {
            if (this.beginUpdateCount == 0)
            {
                // TODO: We should handle this case!
                return;
                //throw new InvalidOperationException("There is no active update to end.");
            }

            this.beginUpdateCount--;
            this.OnChanged();
        }

        internal void CopyFrom(MergedCellRanges fromMergedCellRanges)
        {
            foreach (CellRange mergedCellRange in fromMergedCellRanges.CellRanges)
            {
                this.CellRanges.Add(mergedCellRange);
            }
        }

        private void UpdateChangedCellRange(CellRange cellRange)
        {
            if (this.changedCellRange == null)
            {
                this.changedCellRange = cellRange;
            }
            else
            {
                this.changedCellRange = CellRange.Max(this.changedCellRange, cellRange);
            }
        }

        #endregion


        #region Events

        internal event EventHandler Changing;
        protected virtual void OnChanging()
        {
            if (this.beginUpdateCount > 0)
            {
                return;
            }

            if (this.Changing != null)
            {
                this.Changing(this, EventArgs.Empty);
            }
        }

        private void OnChanged()
        {
            if (this.beginUpdateCount > 0)
            {
                return;
            }

            if (this.changedCellRange != null)
            {
                this.OnChanged(new MergedCellRangesChangedEventArgs(this.changedCellRange));
                this.changedCellRange = null;
            }
        }

        public event EventHandler<MergedCellRangesChangedEventArgs> Changed;
        protected virtual void OnChanged(MergedCellRangesChangedEventArgs args)
        {
            if (this.beginUpdateCount > 0)
            {
                return;
            }

            if (this.Changed != null)
            {
                this.Changed(this, args);
            }
        }

        #endregion
    }
}