﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Spreadsheet.Copying;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a defined name in the RadSpreadsheet.
    /// </summary>
    public class DefinedName : ISpreadsheetName, ICopyable<ISpreadsheetName>
    {
        #region Constants

        internal static readonly string PrintAreaDefinedName = "_xlnm.Print_Area";
        internal static readonly string FilterDefinedName = "_xlnm._FilterDatabase";
        private static readonly string ArrayString = "{...}";

        #endregion


        #region Fields

        private readonly string name;
        private readonly SpreadsheetNameCollectionScope scope;
        private readonly FormulaCellValue formulaCellValue;
        private readonly string comment;
        private readonly bool isVisible;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name.
        /// </summary>
        /// <value>The name.</value>
        public string Name
        {
            get
            {
                return this.name;
            }
        }

        /// <summary>
        /// Gets the scope of the name.
        /// </summary>
        /// <value>The scope.</value>
        public SpreadsheetNameCollectionScope Scope
        {
            get
            {
                return this.scope;
            }
        }

        /// <summary>
        /// Gets the comment.
        /// </summary>
        /// <value>The comment.</value>
        public string Comment
        {
            get
            {
                return this.comment;
            }
        }

        /// <summary>
        /// Gets the string representation of the object the name refers to.
        /// </summary>
        /// <value>The string representation of the object the name refers to.</value>
        public string RefersTo
        {
            get
            {
                return this.formulaCellValue.GetValueAsString(CellValueFormat.GeneralFormat);
            }
        }

        /// <summary>
        /// Gets the value.
        /// </summary>
        /// <value>The value.</value>
        public string Value
        {
            get
            {
                return this.GetDefinedNameValueAsString(int.MaxValue);
            }
        }

        /// <summary>
        /// Gets the value indicating whether the name is visible to the user.
        /// </summary>
        /// <value>The value indicating whether the name is visible to the user.</value>
        public bool IsVisible
        {
            get
            {
                return this.isVisible;
            }
        }

        internal CellReferenceRangeExpression CellReferenceRangeExpression
        {
            get
            {
                return this.GetCellReferenceRangeExpression(this.formulaCellValue.Value);
            }
        }

        internal FormulaCellValue FormulaCellValue
        {
            get
            {
                return this.formulaCellValue;
            }
        }

        #endregion


        #region Constructor

        private DefinedName(string name, SpreadsheetNameCollectionScope scope, FormulaCellValue cellValue, string comment, bool isVisible)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");
            Guard.ThrowExceptionIfNull(scope, "scope");
            Guard.ThrowExceptionIfNull(cellValue, "cellValue");

            this.name = name;
            this.scope = scope;
            this.formulaCellValue = cellValue;
            this.comment = comment;
            this.isVisible = isVisible;
        }

        #endregion


        #region Methods

        private CellReferenceRangeExpression GetCellReferenceRangeExpression(RadExpression expression)
        {
            CellReferenceRangeExpression cellReference = expression.GetValueAsConstantOrCellReference() as CellReferenceRangeExpression;

            if (cellReference != null && cellReference.IsValid)
            {
                return cellReference;
            }

            return null;
        }

        internal static DefinedName Create(string name, SpreadsheetNameCollectionScope scope, FormulaCellValue formulaCellValue, string comment, bool isVisible)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");
            Guard.ThrowExceptionIfNull(scope, "scope");
            Guard.ThrowExceptionIfNull(formulaCellValue, "formulaCellValue");

            if (DefinedName.IsNameValid(name))
            {
                return new DefinedName(name, scope, formulaCellValue, comment, isVisible);
            }

            return null;
        }

        internal string GetDefinedNameValueAsString(int maxStringLength)
        {
            RadExpression expressionValue = this.formulaCellValue.Value.GetValue();

            ArrayExpression arrayExpression = expressionValue as ArrayExpression;
            if (arrayExpression != null)
            {
                if (this.formulaCellValue.Value is CellReferenceRangeExpression)
                {
                    return this.GetArrayExpressionString(arrayExpression, maxStringLength);
                }

                return ArrayString;
            }

            CellReferenceRangeExpression cellReferenceExpression = expressionValue as CellReferenceRangeExpression;
            if (cellReferenceExpression != null)
            {
                return this.GetCellReferenceExpressionString(cellReferenceExpression, maxStringLength);
            }

            if (expressionValue is EmptyExpression)
            {
                return string.Empty;
            }

            return this.formulaCellValue.GetResultValueAsString(CellValueFormat.GeneralFormat);
        }

        private string GetCellReferenceExpressionString(CellReferenceRangeExpression expression, int maxStringValueLength)
        {
            RadExpression expersionValue = expression.GetValue();

            ArrayExpression arrayExpression = expersionValue as ArrayExpression;
            if (arrayExpression != null)
            {
                return this.GetArrayExpressionString(arrayExpression, maxStringValueLength);
            }

            return expersionValue.ToString(FormatHelper.DefaultSpreadsheetCulture);
        }

        private string GetArrayExpressionString(ArrayExpression arrayExpression, int maxStringValueLength)
        {
            Guard.ThrowExceptionIfNull(arrayExpression, "arrayExpression");

            if (arrayExpression.ColumnCount == 1 && arrayExpression.RowCount == 1)
            {
                if (arrayExpression[0, 0] is EmptyExpression)
                {
                    return string.Empty;
                }

                ArrayExpression innerArray = arrayExpression[0, 0] as ArrayExpression;
                if (innerArray != null)
                {
                    return this.GetArrayExpressionString(innerArray, maxStringValueLength);
                }

                return this.formulaCellValue.GetResultValueAsString(CellValueFormat.GeneralFormat);
            }

            StringBuilder builder = new StringBuilder();
            builder.Append(SpreadsheetCultureHelper.ArrayLeftBracket);

            for (int rowIndex = 0; rowIndex < arrayExpression.RowCount; rowIndex++)
            {
                for (int columnIndex = 0; columnIndex < arrayExpression.ColumnCount; columnIndex++)
                {
                    builder.Append(SpreadsheetCultureHelper.DoubleQuote);
                    builder.Append(arrayExpression[rowIndex, columnIndex].ToString(FormatHelper.DefaultSpreadsheetCulture));
                    builder.Append(SpreadsheetCultureHelper.DoubleQuote);

                    if (columnIndex != arrayExpression.ColumnCount - 1)
                    {
                        builder.Append(FormatHelper.DefaultSpreadsheetCulture.ArrayListSeparator);
                    }
                }
                if (rowIndex != arrayExpression.RowCount - 1)
                {
                    builder.Append(FormatHelper.DefaultSpreadsheetCulture.ArrayRowSeparator);
                }

                if (builder.Length >= maxStringValueLength)
                {
                    builder.Append(SpreadsheetCultureHelper.Ellipsis);
                    return builder.ToString();
                }
            }

            builder.Append(SpreadsheetCultureHelper.ArrayRightBracket);

            return builder.ToString();
        }

        // TODO: Extract a validator that holds all validation cases of Sheet names, defined names etc.
        internal static bool IsNameValid(string name)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "value");

            char firstChar = name[0];

            //The name should start with a letter, underscore or backslash
            if (!char.IsLetter(firstChar) && firstChar != '_' && firstChar != '\\')
            {
                return false;
            }

            //The name of a vairable cannot collide with the name of a cell
            if (NameConverter.IsValidA1CellName(name))
            {
                return false;
            }

            //The name of a variable should not contain '%', '$', '!', '[', ']', '#', '<', '>', '='
            HashSet<char> forbiddenSymbolsInVariables = new HashSet<char>() { '%', '$', '!', '[', ']', '#', '<', '>', '=' };

            for (int i = 0; i < name.Length; i++)
            {
                if (forbiddenSymbolsInVariables.Contains(name[i]) || char.IsWhiteSpace(name[i]))
                {
                    return false;
                }
            }

            //The name of a variable must not be a number
            double number;
            if (FormatHelper.DefaultSpreadsheetCulture.TryParseDouble(name, out number))
            {
                return false;
            }

            //The name of a variable should not collide with any literal
            if (AreNamesEqual(name, "R") ||
                AreNamesEqual(name, "C") ||
                AreNamesEqual(name, "TRUE") ||
                AreNamesEqual(name, "FALSE") ||
                AreNamesEqual(name, PrintAreaDefinedName) ||
                AreNamesEqual(name, FilterDefinedName))
            {
                return false;
            }

            return true;
        }

        internal static bool AreNamesEqual(string first, string second)
        {
            return string.Equals(first, second, StringComparison.CurrentCultureIgnoreCase);
        }

        ISpreadsheetName ICopyable<ISpreadsheetName>.Copy(CopyContext context)
        {
            context.SpreadsheetNameExistsInTargetWorkbook = context.TargetWorksheet.Workbook.Names.Contains(this.Name);

            SpreadsheetNameCollectionScope scopeCopied = ((ICopyable<SpreadsheetNameCollectionScope>)this.Scope).Copy(context);
            FormulaCellValue formulaCellValueCopied = ((ICopyable<FormulaCellValue>)this.FormulaCellValue).Copy(context);

            return new DefinedName(this.Name, scopeCopied, formulaCellValueCopied, this.Comment, this.IsVisible);
        }

        #endregion
    }
}