﻿using System;
using System.Linq;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    internal class NameManager
    {
        #region Fields

        private readonly Dictionary<Worksheet, NameCollection> worksheetToCollection;
        private readonly Dictionary<SpreadsheetNameCollectionScope, Worksheet> ownerToWorksheet;
        private readonly NameCollection workbookCollection;

        #endregion


        #region Properties

        public NameCollection WorkbookNameCollection
        {
            get
            {
                return this.workbookCollection;
            }
        }

        public SpreadsheetNameCollectionScope WorkbookScope
        {
            get
            {
                return this.workbookCollection.Owner;
            }
        }

        public Dictionary<int, IEnumerable<ISpreadsheetName>> NameWithWorksheetIdCollections
        {
            get
            {
                Dictionary<int, IEnumerable<ISpreadsheetName>> result = new Dictionary<int, IEnumerable<ISpreadsheetName>>();

                foreach (KeyValuePair<Worksheet, NameCollection> pair in this.worksheetToCollection)
                {
                    result.Add(pair.Key.Workbook.Worksheets.IndexOf(pair.Key), pair.Value);
                }

                result.Add(-1, this.workbookCollection);

                return result;
            }
        }

        public IEnumerable<ISpreadsheetName> Names
        {
            get
            {
                // TODO: yield
                IEnumerable<ISpreadsheetName> result = this.workbookCollection;
                foreach (NameCollection collection in this.worksheetToCollection.Values)
                {
                    result = result.Concat(collection);
                }

                return result;
            }
        }


        public IEnumerable<SpreadsheetNameCollectionScope> Scopes
        {
            get
            {
                yield return this.WorkbookScope;

                foreach (SpreadsheetNameCollectionScope owner in this.ownerToWorksheet.Keys)
                {
                    yield return owner;
                }
            }
        }

        #endregion


        #region Constructors

        public NameManager(Workbook workbook)
        {
            Guard.ThrowExceptionIfNull(workbook, "workbook");

            this.worksheetToCollection = new Dictionary<Worksheet, NameCollection>();
            this.ownerToWorksheet = new Dictionary<SpreadsheetNameCollectionScope, Worksheet>();
            this.workbookCollection = new NameCollection(new SpreadsheetNameCollectionScope(workbook));
            this.workbookCollection.Changing += this.NameCollectionChanging;
            this.workbookCollection.Changed += this.NameCollectionChanged;
        }

        #endregion


        #region Methods

        private ISpreadsheetName GetGlobalName(string name)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");

            ISpreadsheetName result = null;
            this.workbookCollection.TryGetSpreadsheetName(name, out result);

            return result;
        }

        private ISpreadsheetName GetLocalName(Worksheet worksheet, string name)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");

            ISpreadsheetName result = null;
            if (this.worksheetToCollection.ContainsKey(worksheet))
            {
                this.worksheetToCollection[worksheet].TryGetSpreadsheetName(name, out result);
            }

            return result;
        }

        public IEnumerable<ISpreadsheetName> GetOrderedNameBoxDefinedNames(Worksheet worksheet)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfFalse(worksheetToCollection.ContainsKey(worksheet), "WorksheetToCollection does not contain worksheet");

            IEnumerable<ISpreadsheetName> result = this.GetNameBoxDefinedNames(worksheet);
            return result.OrderBy(name => name.Name);
        }

        private IEnumerable<ISpreadsheetName> GetNameBoxDefinedNames(Worksheet worksheet)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfFalse(worksheetToCollection.ContainsKey(worksheet), "WorksheetToCollection does not contain worksheet");

            NameCollection nameCollection;
            if (this.worksheetToCollection.TryGetValue(worksheet, out nameCollection))
            {
                IEnumerable<ISpreadsheetName> localNames = nameCollection.GetNameBoxDefinedNames();
                foreach (ISpreadsheetName name in localNames)
                {
                    yield return name;
                }
            }

            IEnumerable<ISpreadsheetName> globalNames = this.workbookCollection.GetNameBoxDefinedNames();
            foreach (ISpreadsheetName globalName in globalNames)
            {
                if (!nameCollection.Contains(globalName.Name))
                {
                    yield return globalName;
                }
            }
        }

        public ISpreadsheetName Add(SpreadsheetNameCollectionScope scope, string name, string refersTo, CellIndex cellIndex, string comment = null)
        {
            NameCollection nameCollection = this.FindNameCollection(scope);
            return nameCollection.Add(name, refersTo, cellIndex, comment);
        }

        public void Remove(ISpreadsheetName name)
        {
            Guard.ThrowExceptionIfNull(name, "name");

            NameCollection nameCollection = this.FindNameCollection(name.Scope);
            nameCollection.Remove(name.Name);
        }

        public void Update(ISpreadsheetName oldName, string name, string refersTo, CellIndex cellIndex, string comment = null)
        {
            NameCollection nameCollection = this.FindNameCollection(oldName.Scope);
            nameCollection.Update(oldName, name, refersTo, cellIndex, comment);
        }

        private NameCollection FindNameCollection(SpreadsheetNameCollectionScope scope)
        {
            if (scope == this.WorkbookScope)
            {
                return this.workbookCollection;
            }

            Worksheet worksheet;
            NameCollection nameCollection;
            if (this.ownerToWorksheet.TryGetValue(scope, out worksheet) && this.worksheetToCollection.TryGetValue(worksheet, out nameCollection))
            {
                return nameCollection;
            }

            return null;
        }

        public NameCollection CreateCollection(Worksheet worksheet)
        {
            SpreadsheetNameCollectionScope owner = new SpreadsheetNameCollectionScope(worksheet);
            NameCollection collection = new NameCollection(owner);

            this.worksheetToCollection.Add(worksheet, collection);
            this.ownerToWorksheet.Add(owner, worksheet);

            collection.Changing += this.NameCollectionChanging;
            collection.Changed += this.NameCollectionChanged;

            return collection;
        }

        public void RemoveOwner(Worksheet worksheet)
        {
            NameCollection nameCollection;
            if (this.worksheetToCollection.TryGetValue(worksheet, out nameCollection))
            {
                nameCollection.Changing -= this.NameCollectionChanging;
                nameCollection.Changed -= this.NameCollectionChanged;
            }
            this.worksheetToCollection.Remove(worksheet);
        }

        private void NameCollectionChanging(object sender, NameCollectionChangingEventArgs args)
        {
            this.OnChanging(args);
        }

        private void NameCollectionChanged(object sender, EventArgs e)
        {
            this.OnChanged();
        }

        public ISpreadsheetName FindSpreadsheetName(string qualifier, string name, Worksheet currentWorksheet, Workbook workbook)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");
            Guard.ThrowExceptionIfNull(currentWorksheet, "currentWorksheet");
            Guard.ThrowExceptionIfNull(workbook, "workbook");

            if (!string.IsNullOrEmpty(qualifier))
            {
                string decodedQualifier = TextHelper.DecodeWorksheetName(qualifier);
                if (decodedQualifier.Equals(workbook.Name, StringComparison.CurrentCultureIgnoreCase))
                {
                    return this.GetGlobalName(name);
                }

                Worksheet referredWorksheet = SpreadsheetHelper.GetWorksheetByName(workbook, decodedQualifier);

                if (referredWorksheet != null)
                {
                    return this.GetLocalName(referredWorksheet, name);
                }
            }
            else
            {
                ISpreadsheetName definedName = this.GetLocalName(currentWorksheet, name);
                if (definedName == null)
                {
                    definedName = workbook.NameManager.GetGlobalName(name);
                }

                return definedName;
            }

            return null;
        }

        #endregion


        #region Events

        internal event EventHandler<NameCollectionChangingEventArgs> Changing;
        private void OnChanging(NameCollectionChangingEventArgs args)
        {
            if (this.Changing != null)
            {
                this.Changing(this, args);
            }
        }

        public event EventHandler Changed;
        private void OnChanged()
        {
            if (this.Changed != null)
            {
                this.Changed(this, EventArgs.Empty);
            }
        }

        #endregion
    }
}
