﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.Copying;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a collection of names.
    /// </summary>
    public class NameCollection : IEnumerable<ISpreadsheetName>
    {
        #region Fields

        private readonly Dictionary<string, ISpreadsheetName> innerList;
        private readonly SpreadsheetNameCollectionScope owner;
        private int suspendChangeEventsCount;

        #endregion


        #region Properties

        internal SpreadsheetNameCollectionScope Owner
        {
            get
            {
                return owner;
            }
        }

        #endregion


        #region Constructors

        internal NameCollection(SpreadsheetNameCollectionScope owner)
        {
            Guard.ThrowExceptionIfNull(owner, "owner");

            this.owner = owner;
            this.innerList = new Dictionary<string, ISpreadsheetName>();
        }

        #endregion


        #region Indexer

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.ISpreadsheetName" /> with the specified name.
        /// </summary>
        /// <value>The defined name.</value>
        public ISpreadsheetName this[string name]
        {
            get
            {
                return this.innerList[name.ToLowerInvariant()];
            }
        }

        #endregion


        #region Methods

        /// <summary>
        /// Adds the specified name.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="refersTo">The refers to.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <param name="comment">The comment.</param>
        /// <param name="isVisible">The is visible.</param>
        /// <returns></returns>
        public ISpreadsheetName Add(string name, string refersTo, int rowIndex, int columnIndex, string comment = null, bool isVisible = true)
        {
            if (string.IsNullOrEmpty(name) || !DefinedName.IsNameValid(name))
            {
                throw new SpreadsheetNameException("The name you have entered is not valid",
                "Spreadsheet_ErrorExpressions_InvalidName");
            }

            if (string.IsNullOrEmpty(refersTo))
            {
                throw new SpreadsheetNameException("The value that you entered is not valid.",
                    "Spreadsheet_ErrorExpressions_InvalidValue");
            }

            if (this.innerList.ContainsKey(name.ToLowerInvariant()))
            {
                throw new SpreadsheetNameException("The name entered already exists. Enter a unique name.",
                    "Spreadsheet_ErrorExpressions_NameExists");
            }

            this.EnsureNonEmptyOwner();

            FormulaCellValue formulaCellValue = this.CreateFormulaCellValue(refersTo, rowIndex, columnIndex);
            DefinedName definedName = DefinedName.Create(name, this.owner, formulaCellValue, comment, isVisible);
            AddRemoveSpreadsheetNameCommandContext addContext = new AddRemoveSpreadsheetNameCommandContext(this.owner, definedName);

            this.owner.CurrentWorksheet.ExecuteCommand(WorkbookCommands.AddSpreadsheetName, addContext);

            return definedName;
        }

        /// <summary>
        /// Adds the specified name.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="refersTo">The refers to.</param>
        /// <param name="cellIndex">Index of the cell.</param>
        /// <param name="comment">The comment.</param>
        /// <param name="isVisible">The is visible.</param>
        /// <returns></returns>
        public ISpreadsheetName Add(string name, string refersTo, CellIndex cellIndex, string comment = null, bool isVisible = true)
        {
            if (cellIndex == null)
            {
                throw new SpreadsheetNameException("The cell index that you provided is not valid",
                    "Spreadsheet_ErrorExpressions_InvalidCellIndex");
            }

            return this.Add(name, refersTo, cellIndex.RowIndex, cellIndex.ColumnIndex, comment, isVisible);
        }

        /// <summary>
        /// Removes the specified name.
        /// </summary>
        /// <param name="name">The name.</param>
        public void Remove(string name)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");

            ISpreadsheetName definedName = this[name];
            AddRemoveSpreadsheetNameCommandContext removeContext = new AddRemoveSpreadsheetNameCommandContext(this.owner, definedName);
            this.owner.CurrentWorksheet.ExecuteCommand(WorkbookCommands.RemoveSpreadsheetName, removeContext);
        }

        /// <summary>
        /// Determines whether the collection contains the specified name.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <returns>The value indicating whether the collection contains the specified name.</returns>
        public bool Contains(string name)
        {
            return this.innerList.ContainsKey(name.ToLowerInvariant());
        }

        /// <summary>
        /// Get the SpreadsheetName associated with the specified name.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="spreadsheetName">The resulting name.</param>
        /// <returns>The value indicating whether the name was successfully found.</returns>
        public bool TryGetSpreadsheetName(string name, out ISpreadsheetName spreadsheetName)
        {
            return this.innerList.TryGetValue(name.ToLowerInvariant(), out spreadsheetName);
        }

        internal void AddInternal(ISpreadsheetName name)
        {
            Guard.ThrowExceptionIfNull(name, "name");

            this.OnChanging(new NameCollectionChangingEventArgs(null, null, name.Name, name.RefersTo));

            this.innerList.Add(name.Name.ToLowerInvariant(), name);

            this.OnChanged();
        }

        internal void RemoveInternal(string name)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");

            this.OnChanging(new NameCollectionChangingEventArgs(name, null, null, null));

            this.innerList.Remove(name.ToLowerInvariant());

            this.OnChanged();
        }

        internal void RemoveInternal(ISpreadsheetName name)
        {
            Guard.ThrowExceptionIfNull(name, "name");

            this.RemoveInternal(name.Name);
        }

        internal void Update(ISpreadsheetName oldName, string name, string refersTo, CellIndex cellIndex, string comment = null)
        {
            Guard.ThrowExceptionIfNull(oldName, "oldName");

            this.OnChanging(new NameCollectionChangingEventArgs(oldName.Name, oldName.RefersTo, name, refersTo));
            this.SuspendChangeEvents();

            this.Owner.Workbook.History.BeginUndoGroup();

            try
            {
                this.Remove(oldName.Name);
                this.Add(name, refersTo, cellIndex, comment);
                this.Owner.Workbook.History.EndUndoGroup();
            }
            catch
            {
                this.Owner.Workbook.History.CancelUndoGroup();
                throw;
            }
            finally
            {
                this.ResumeChangeEvents();
                this.OnChanged();
            }
        }

        internal IEnumerable<ISpreadsheetName> GetNameBoxDefinedNames()
        {
            foreach (ISpreadsheetName name in this.innerList.Values)
            {
                DefinedName definedName = name as DefinedName;
                if (definedName != null)
                {
                    if (definedName.CellReferenceRangeExpression != null)
                    {
                        yield return definedName;
                    }
                }
            }
        }

        private void UpdateDependentNames(Action<DefinedName> action)
        {
            foreach (ISpreadsheetName name in this.innerList.Values)
            {
                DefinedName definedName = name as DefinedName;
                if (definedName != null)
                {
                    action(definedName);
                }
            }
        }

        internal void UpdateSheetNameDependentNames(Worksheet renamedWorksheet)
        {
            this.UpdateDependentNames((name) =>
            {
                if (name.FormulaCellValue.IsSheetNameDependent)
                {
                    name.FormulaCellValue.Translate(renamedWorksheet);
                }
            });
        }

        internal void UpdateWorkbookNameDependentNames(Workbook renamedWorkbook)
        {
            this.UpdateDependentNames((name) =>
            {
                if (name.FormulaCellValue.IsWorkbookNameDependent)
                {
                    name.FormulaCellValue.Translate(renamedWorkbook);
                }
            });
        }

        internal void UpdateSpreadsheetNameDependentNames(string oldName, string newName)
        {
            this.UpdateDependentNames((name) =>
            {
                if (name.FormulaCellValue.ContainsSpreadsheetName)
                {
                    name.FormulaCellValue.Translate(oldName, newName);
                }
            });
        }

        internal void CopyFrom(NameCollection fromNameCollection, CopyContext context)
        {
            foreach (KeyValuePair<string, ISpreadsheetName> nameToSpreadsheetName in fromNameCollection.innerList)
            {
                ISpreadsheetName spreadsheetNameCopied = ((ICopyable<ISpreadsheetName>)nameToSpreadsheetName.Value).Copy(context);

                if (!context.SpreadsheetNameExistsInTargetWorkbook)
                {
                    this.innerList.Add(nameToSpreadsheetName.Key, spreadsheetNameCopied);
                }
                else
                {
                    context.TargetWorksheet.Names.AddInternal(spreadsheetNameCopied);
                }
            }
        }

        internal void Clear()
        {
            this.innerList.Clear();
        }

        private void EnsureNonEmptyOwner()
        {
            if (owner.CurrentWorksheet == null)
            {
                throw new SpreadsheetNameException("Workbook should contain at least one worksheet", "Spreadsheet_NameCollection_EmptyWorkbookError");
            }
        }

        private FormulaCellValue CreateFormulaCellValue(string refersTo, int rowIndex, int columnIndex)
        {
            Guard.ThrowExceptionIfNullOrEmpty(refersTo, "refersTo");
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            ParseResult parseResult = ParseResult.Unsuccessful;
            RadExpression expression = null;
            InputStringCollection stringExpressionCollection = null;

            if (SpreadsheetCultureHelper.IsCharEqualTo(refersTo[0], SpreadsheetCultureHelper.Equal) && refersTo.Length > 1)
            {
                parseResult = this.TryParseRadExpression(refersTo, rowIndex, columnIndex, out expression, out stringExpressionCollection);
            }
            else
            {
                if (!char.IsLetter(refersTo[0]))
                {
                    StringBuilder builder = new StringBuilder();
                    builder.Append(SpreadsheetCultureHelper.Equal);
                    builder.Append(refersTo);

                    parseResult = this.TryParseRadExpression(builder.ToString(), rowIndex, columnIndex, out expression, out stringExpressionCollection);
                }

                if (parseResult != ParseResult.Successful)
                {
                    parseResult = this.TryParseRadExpression(GetEscapedFormulaString(refersTo), rowIndex, columnIndex,
                        out expression, out stringExpressionCollection);
                }
            }

            if (parseResult == ParseResult.Error)
            {
                throw new ParseException(SpreadsheetStrings.GeneralErrorMessage);
            }

            return new FormulaCellValue(stringExpressionCollection, expression, rowIndex, columnIndex);
        }

        private static string GetEscapedFormulaString(string value)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append(SpreadsheetCultureHelper.Equal);
            builder.Append(TextHelper.EncodeValue(value));

            return builder.ToString();
        }

        private ParseResult TryParseRadExpression(string value, int rowIndex, int columnIndex, out RadExpression expression,
            out InputStringCollection stringExpressionCollection)
        {
            return RadExpression.TryParse(value, owner.CurrentWorksheet, rowIndex, columnIndex,
                   out expression, out stringExpressionCollection, this.owner.IsGlobal);
        }

        private void SuspendChangeEvents()
        {
            this.suspendChangeEventsCount++;
        }

        private void ResumeChangeEvents()
        {
            if (this.suspendChangeEventsCount == 0)
            {
                throw new SpreadsheetNameException("There is no active suspend to resume.",
                new InvalidOperationException("There is no active suspend to resume."),
                "Spreadsheet_ErrorExpressions_NoActiveSuspendToResume");
            }

            if (this.suspendChangeEventsCount > 0)
            {
                this.suspendChangeEventsCount--;
            }
        }

        private bool AreChangeEventsSuspended()
        {
            return this.suspendChangeEventsCount > 0;
        }

        #endregion


        #region Events

        internal event EventHandler<NameCollectionChangingEventArgs> Changing;
        private void OnChanging(NameCollectionChangingEventArgs args)
        {
            if (this.AreChangeEventsSuspended())
            {
                return;
            }

            if (this.Changing != null)
            {
                this.Changing(this, args);
            }
        }

        internal event EventHandler Changed;
        private void OnChanged()
        {
            if (this.AreChangeEventsSuspended())
            {
                return;
            }

            if (this.Changed != null)
            {
                this.Changed(this, EventArgs.Empty);
            }
        }

        #endregion


        #region IEnumerable

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return ((IEnumerable<ISpreadsheetName>)this).GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<ISpreadsheetName> GetEnumerator()
        {
            return this.innerList.Values.GetEnumerator();
        }

        #endregion

    }
}
