﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Layout;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// The class represents the pane.
    /// </summary>
    public class Pane
    {
        #region Fields

        private readonly int xSplit;
        private readonly int ySplit;
        private readonly CellIndex topLeftCellIndex;
        private readonly ViewportPaneType activePane;
        private readonly PaneState state;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the X split.
        /// </summary>
        /// <value>The X split.</value>
        public int XSplit
        {
            get
            {
                return this.xSplit;
            }
        }

        /// <summary>
        /// Gets the Y split.
        /// </summary>
        /// <value>The Y split.</value>
        public int YSplit
        {
            get
            {
                return this.ySplit;
            }
        }

        /// <summary>
        /// Gets the index of the top left cell of the scrollable pane.
        /// </summary>
        /// <value>The index of the top left cell of the scrollable pane.</value>
        public CellIndex TopLeftCellIndex
        {
            get
            {
                return this.topLeftCellIndex;
            }
        }

        /// <summary>
        /// Gets the active pane.
        /// </summary>
        /// <value>The active pane.</value>
        public ViewportPaneType ActivePane
        {
            get
            {
                return this.activePane;
            }
        }

        /// <summary>
        /// Gets the state.
        /// </summary>
        /// <value>The state.</value>
        public PaneState State
        {
            get
            {
                return this.state;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="Pane" /> class.
        /// </summary>
        /// <param name="topLeftCellIndex">Index of the top left cell.</param>
        /// <param name="xSplit">The x split.</param>
        /// <param name="ySplit">The y split.</param>
        /// <param name="activePane">The active pane.</param>
        /// <param name="state">The state.</param>
        public Pane(CellIndex topLeftCellIndex, int xSplit, int ySplit, ViewportPaneType activePane, PaneState state = PaneState.Frozen)
        {
            Guard.ThrowExceptionIfNull(topLeftCellIndex, "topLeftCellIndex");
            Guard.ThrowExceptionIfLessThan(0, xSplit, "xSplit");
            Guard.ThrowExceptionIfLessThan(0, ySplit, "ySplit");

            this.topLeftCellIndex = topLeftCellIndex;
            this.xSplit = xSplit;
            this.ySplit = ySplit;
            this.activePane = activePane;
            this.state = state;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Clones this instance.
        /// </summary>
        /// <returns>The result pane.</returns>
        public Pane Clone()
        {
            return new Pane(this.TopLeftCellIndex, this.XSplit, this.YSplit, this.ActivePane, this.State);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The object to compare with the current object.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            Pane other = obj as Pane;

            if (other == null)
            {
                return false;
            }

            return this.XSplit.Equals(other.XSplit) &&
                this.YSplit.Equals(other.YSplit) &&
                this.TopLeftCellIndex.Equals(other.TopLeftCellIndex) &&
                this.ActivePane.Equals(other.ActivePane) &&
                this.State.Equals(other.State);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                this.XSplit.GetHashCodeOrZero(),
                this.YSplit.GetHashCodeOrZero(),
                this.TopLeftCellIndex.GetHashCodeOrZero(),
                this.ActivePane.GetHashCodeOrZero(),
                this.State.GetHashCodeOrZero());
        }

        #endregion
    }
}