﻿using System.Windows.Media;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
#if NETCORE || NETSTANDARD
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Describes a pattern type of fill.
    /// </summary>
    public class PatternFill : IFill
    {
        #region Constants

        /// <summary>
        /// Default pattern fill.
        /// </summary>
        public static readonly PatternFill Default = (PatternFill)PatternFill.CreateSolidFill(Colors.Transparent);

        #endregion


        #region Fields

        private readonly PatternType patternType;
        private readonly ThemableColor patternColor;
        private readonly ThemableColor backgroundColor;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the type of the pattern.
        /// </summary>
        /// <value>The type of the pattern.</value>
        public PatternType PatternType
        {
            get
            {
                return this.patternType;
            }
        }

        /// <summary>
        /// Gets the color of the pattern.
        /// </summary>
        /// <value>The color of the pattern.</value>
        public ThemableColor PatternColor
        {
            get
            {
                return this.patternColor;
            }
        }

        /// <summary>
        /// Gets the color of the background.
        /// </summary>
        /// <value>The color of the background.</value>
        public ThemableColor BackgroundColor
        {
            get
            {
                return this.backgroundColor;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="PatternFill" /> class.
        /// </summary>
        /// <param name="patternType">Type of the pattern.</param>
        /// <param name="patternColor">Color of the pattern.</param>
        /// <param name="backgroundColor">Color of the background.</param>
        public PatternFill(PatternType patternType, Color patternColor, Color backgroundColor) : this(patternType, new ThemableColor(patternColor), new ThemableColor(backgroundColor))
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="PatternFill" /> class.
        /// </summary>
        /// <param name="patternType">Type of the pattern.</param>
        /// <param name="patternColor">Color of the pattern.</param>
        /// <param name="backgroundColor">Color of the background.</param>
        public PatternFill(PatternType patternType, ThemableColor patternColor, ThemableColor backgroundColor)
        {
            this.patternType = patternType;
            this.patternColor = patternColor;
            this.backgroundColor = backgroundColor;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Creates a solid fill.
        /// </summary>
        /// <param name="color">The color.</param>
        /// <returns>The fill.</returns>
        public static IFill CreateSolidFill(Color color)
        {
            return new PatternFill(PatternType.Solid, color, Colors.Black);
        }

        /// <summary>
        /// Creates a solid fill.
        /// </summary>
        /// <param name="color">The themable color.</param>
        /// <returns>The fill.</returns>
        public static IFill CreateSolidFill(ThemableColor color)
        {
            return new PatternFill(PatternType.Solid, color, new ThemableColor(Colors.Black));
        }

        internal ThemableColor GetBackgroundDependingOnPatternType()
        {
            return this.PatternType == Model.PatternType.Solid ? this.PatternColor : this.BackgroundColor;
        }

        #endregion


        #region Operators & Overrides

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            PatternFill other = obj as PatternFill;
            if (other == null)
            {
                return false;
            }

            return this.patternType == other.patternType &&
                TelerikHelper.EqualsOfT(this.patternColor, other.patternColor) &&
                TelerikHelper.EqualsOfT(this.backgroundColor, other.backgroundColor);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                this.patternType.GetHashCode(),
                this.patternColor.GetHashCodeOrZero(),
                this.backgroundColor.GetHashCodeOrZero());
        }

        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current
        /// <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
            return string.Format("pt:{0} fg:{1} bg:{2}", this.PatternType, this.PatternColor, this.BackgroundColor);
        }

        #endregion
    }
}