﻿using System;
using System.Collections.Generic;
using System.Text;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Printing
{
    /// <summary>
    /// Represents the content of a header or a footer.
    /// </summary>
    public class HeaderFooterContent
    {
        internal const int MaxLength = 255;
        private readonly HeaderFooterSection leftSection;
        private readonly HeaderFooterSection centerSection;
        private readonly HeaderFooterSection rightSection;

        internal HeaderFooterContent(Action updateOnChange)
        {
            this.leftSection = new HeaderFooterSection(updateOnChange);
            this.centerSection = new HeaderFooterSection(updateOnChange);
            this.rightSection = new HeaderFooterSection(updateOnChange);
        }

        internal HeaderFooterContent(HeaderFooterContent other, Action updateOnChange)
        {
            this.leftSection = new HeaderFooterSection(other.leftSection, updateOnChange);
            this.centerSection = new HeaderFooterSection(other.centerSection, updateOnChange);
            this.rightSection = new HeaderFooterSection(other.rightSection, updateOnChange);
        }

        /// <summary>
        /// Gets the left section of the header. This section has horizontal alignment left.
        /// </summary>
        /// <value>The left section.</value>
        public HeaderFooterSection LeftSection
        {
            get
            {
                return this.leftSection;
            }
        }

        /// <summary>
        /// Gets the center section of the header. This section has horizontal alignment center.
        /// </summary>
        /// <value>The center section.</value>
        public HeaderFooterSection CenterSection
        {
            get
            {
                return this.centerSection;
            }
        }

        /// <summary>
        /// Gets the right section of the header. This section has horizontal alignment right.
        /// </summary>
        /// <value>The right section.</value>
        public HeaderFooterSection RightSection
        {
            get
            {
                return this.rightSection;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the length of total content text is valid. 
        /// The maximum characters allowed for exporting HeaderFooterContent is 255.
        /// </summary>
        /// <value>Value indicating whether the length of the content is valid.</value>
        public bool HasValidLength
        {
            get
            {
                string text = this.BuildContentText();

                return text.Length <= HeaderFooterContent.MaxLength;
            }
        }

        internal IEnumerable<HeaderFooterSection> Sections
        {
            get
            {
                yield return this.LeftSection;
                yield return this.CenterSection;
                yield return this.RightSection;
            }
        }

        internal bool IsEmpty
        {
            get
            {
                string text = this.BuildContentText();

                return string.IsNullOrEmpty(text);
            }
        }

        internal string BuildContentText()
        {
            HeaderFooterSectionTextSanitizer sanitizer = new HeaderFooterSectionTextSanitizer();
            string leftContent = sanitizer.Sanitize(this.LeftSection.Text);
            string centerContent = sanitizer.Sanitize(this.CenterSection.Text);
            string rightContent = sanitizer.Sanitize(this.RightSection.Text);

            StringBuilder contentText = new StringBuilder();

            if (!string.IsNullOrEmpty(leftContent))
            {
                contentText.AppendFormat("{0}{1}{2}", HeaderFooterFieldSymbols.FieldStartSymbol, HeaderFooterFieldSymbols.LeftSectionSymbol, leftContent);
            }

            if (!string.IsNullOrEmpty(centerContent))
            {
                contentText.AppendFormat("{0}{1}{2}", HeaderFooterFieldSymbols.FieldStartSymbol, HeaderFooterFieldSymbols.CenterSectionSymbol, centerContent);
            }

            if (!string.IsNullOrEmpty(rightContent))
            {
                contentText.AppendFormat("{0}{1}{2}", HeaderFooterFieldSymbols.FieldStartSymbol, HeaderFooterFieldSymbols.RightSectionSymbol, rightContent);
            }

            return contentText.ToString();
        }

        internal void SetFromContentText(string contentText)
        {
            contentText = contentText ?? string.Empty;
            StringBuilder leftBuilder = new StringBuilder(), centerBuilder = new StringBuilder(), rightBuilder = new StringBuilder();
            StringBuilder currentSectionBuilder = null;
            bool isEscaping = false;

            foreach (char symbol in contentText)
            {
                if (isEscaping)
                {
                    switch (symbol)
                    {
                        case HeaderFooterFieldSymbols.LeftSectionSymbol:
                            currentSectionBuilder = leftBuilder;
                            break;
                        case HeaderFooterFieldSymbols.CenterSectionSymbol:
                            currentSectionBuilder = centerBuilder;
                            break;
                        case HeaderFooterFieldSymbols.RightSectionSymbol:
                            currentSectionBuilder = rightBuilder;
                            break;
                        default:
                            if (currentSectionBuilder != null)
                            {
                                currentSectionBuilder.AppendFormat("{0}{1}", HeaderFooterFieldSymbols.FieldStartSymbol, symbol);
                            }
                            break;
                    }

                    isEscaping = false;
                }
                else if (symbol == HeaderFooterFieldSymbols.FieldStartSymbol)
                {
                    isEscaping = true;
                }
                else if (currentSectionBuilder != null)
                {
                    currentSectionBuilder.Append(symbol);
                }
            }

            HeaderFooterSectionTextSanitizer sanitizer = new HeaderFooterSectionTextSanitizer();
            this.LeftSection.Text = sanitizer.Sanitize(leftBuilder.ToString());
            this.CenterSection.Text = sanitizer.Sanitize(centerBuilder.ToString());
            this.RightSection.Text = sanitizer.Sanitize(rightBuilder.ToString());
        }

        internal void CopyFrom(HeaderFooterContent fromHeaderFooterContent)
        {
            this.LeftSection.CopyFrom(fromHeaderFooterContent.LeftSection);
            this.CenterSection.CopyFrom(fromHeaderFooterContent.CenterSection);
            this.RightSection.CopyFrom(fromHeaderFooterContent.RightSection);
        }

        internal void Clear()
        {
            this.LeftSection.Clear();
            this.CenterSection.Clear();
            this.RightSection.Clear();
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            HeaderFooterContent other = obj as HeaderFooterContent;

            if (other == null)
            {
                return false;
            }

            return this.LeftSection.Equals(other.LeftSection) &&
                this.CenterSection.Equals(other.CenterSection) &&
                this.RightSection.Equals(other.RightSection);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(this.LeftSection.GetHashCode(), this.CenterSection.GetHashCode(), this.RightSection.GetHashCode());
        }
    }
}