﻿using System;
using System.Collections.Generic;
using System.Text;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Printing
{
    internal class HeaderFooterSectionTextParser
    {
        public const int ColorTextLength = 6;
        private readonly Dictionary<char, Action> fieldHandlers;
        private readonly StringBuilder textFragmentBuilder;
        private readonly IHeaderFooterSectionTextParseHandler parseHandler;
        private bool isParsing;
        private int currentIndex;
        private string text;

        public HeaderFooterSectionTextParser(IHeaderFooterSectionTextParseHandler parseHandler)
        {
            Guard.ThrowExceptionIfNull(parseHandler, "parseHandler");

            this.parseHandler = parseHandler;
            this.textFragmentBuilder = new StringBuilder();
            this.isParsing = false;

            this.fieldHandlers = new Dictionary<char, Action>();
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.ColorSymbol, this.OnParsingColor);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.FontPropertiesStartSymbol, this.OnParsingFontProperties);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.BoldSymbol, this.parseHandler.OnBoldParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.ItalicSymbol, this.parseHandler.OnItalicParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.UnderlineSymbol, this.parseHandler.OnUnderlineParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.DoubleUnderlineSymbol, this.parseHandler.OnDoubleUnderlineParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.StrikeThroughSymbol, this.parseHandler.OnStrikeThroughParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.SubscriptSymbol, this.parseHandler.OnSubscriptParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.SuperscriptSymbol, this.parseHandler.OnSuperscriptParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.PageNumberSymbol, this.parseHandler.OnPageNumberParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.AddToPageNumberSymbol, this.parseHandler.OnAddToPageNumberParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.SubstractFromPageNumberSymbol, this.parseHandler.OnSubstractFromPageNumberParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.NumberOfPagesSymbol, this.parseHandler.OnNumberOfPagesParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.DateSymbol, this.parseHandler.OnDateParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.TimeSymbol, this.parseHandler.OnTimeParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.FilePathSymbol, this.parseHandler.OnFilePathParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.FileNameSymbol, this.parseHandler.OnFileNameParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.SheetNameSymbol, this.parseHandler.OnSheetNameParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.PictureSymbol, this.parseHandler.OnPictureParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.AmpersandSymbol, this.parseHandler.OnAmpersandParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.OutlineStyle, this.parseHandler.OnOutlineStyleParse);
            this.fieldHandlers.Add(HeaderFooterFieldSymbols.ShadowStyle, this.parseHandler.OnShadowStyleParse);
        }

        public void Parse(string headerFooterSectionText)
        {
            this.OnBeforeParse(headerFooterSectionText);

            if (!string.IsNullOrEmpty(headerFooterSectionText))
            {
                this.ParseText();
            }

            this.OnAfterParse();
        }        

        private void OnBeforeParse(string headerFooterSectionText)
        {
            Guard.ThrowExceptionIfTrue(this.isParsing, "isParsing");
            Guard.ThrowExceptionIfGreaterThan(0, this.textFragmentBuilder.Length, "textFragmentBuilder.Length");
            this.isParsing = true;
            this.currentIndex = 0;
            this.text = headerFooterSectionText;
        }

        private void OnAfterParse()
        {
            Guard.ThrowExceptionIfFalse(this.isParsing, "isParsing");
            Guard.ThrowExceptionIfGreaterThan(0, this.textFragmentBuilder.Length, "textFragmentBuilder.Length");

            this.text = null;
            this.currentIndex = 0;
            this.isParsing = false;
        }

        private void ParseText()
        {
            while (this.currentIndex < this.text.Length)
            {
                char symbol = this.text[this.currentIndex];

                if (symbol == HeaderFooterFieldSymbols.FieldStartSymbol)
                {
                    this.PopTextFragment();
                    this.currentIndex++;
                    this.OnParsingSpecialSymbol();
                }
                else
                {
                    this.textFragmentBuilder.Append(symbol);
                    this.currentIndex++;
                }
            }

            this.PopTextFragment();
        }

        private void OnParsingSpecialSymbol()
        {
            if (this.currentIndex < this.text.Length)
            {
                char symbol = this.text[this.currentIndex];

                Action fieldhandler;
                if (this.fieldHandlers.TryGetValue(symbol, out fieldhandler))
                {
                    fieldhandler();
                    this.currentIndex++;
                }
                else if (char.IsDigit(symbol))
                {
                    this.OnParsingFontSize();
                }
            }
        }

        private void OnParsingColor()
        {
            this.currentIndex++;
            int indexAfterColor = this.currentIndex + ColorTextLength;
            bool shouldParseColor = indexAfterColor < this.text.Length;

            if (shouldParseColor)
            {
                string colorText = this.text.Substring(this.currentIndex, ColorTextLength);
                this.parseHandler.OnColorParse(colorText);
                this.currentIndex = indexAfterColor - 1;
            }
            else
            {
                this.currentIndex = this.text.Length - 1;
            }
        }

        private void OnParsingFontProperties()
        {
            this.currentIndex++;
            int endSymbolIndex = this.text.IndexOf(HeaderFooterFieldSymbols.FontPropertiesEndSymbol, this.currentIndex);
            bool shouldParseFontProperties = -1 < endSymbolIndex &&  endSymbolIndex < this.text.Length - 1;

            if (shouldParseFontProperties)
            {
                int fontPropertiesTextLength = endSymbolIndex - this.currentIndex;
                string fontProperties = this.text.Substring(this.currentIndex, fontPropertiesTextLength);
                this.parseHandler.OnFontPropertiesParse(fontProperties.Trim());
                this.currentIndex = endSymbolIndex;
            }
            else
            {
                this.currentIndex = this.text.Length - 1;
            }
        }

        private void OnParsingFontSize()
        {
            StringBuilder numberBuilder = new StringBuilder();
            char symbol = this.text[this.currentIndex];

            while (this.currentIndex < this.text.Length && char.IsDigit(symbol))
            {
                numberBuilder.Append(symbol);
                this.currentIndex++;

                if (this.currentIndex < this.text.Length)
                {
                    symbol = this.text[this.currentIndex];
                }
            }

            bool shouldParseFontSize = this.currentIndex < this.text.Length;

            if (shouldParseFontSize)
            {
                int fontSize = int.Parse(numberBuilder.ToString());
                this.parseHandler.OnFontSizeParse(fontSize);
            }
        }

        private void PopTextFragment()
        {
            if (this.textFragmentBuilder.Length > 0)
            {
                this.parseHandler.OnTextFragmentParse(this.textFragmentBuilder.ToString());
                this.textFragmentBuilder.Clear();
            }
        }
    }
}
