﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Printing
{
    /// <summary>
    /// The setting describing the headers and footers presentation.
    /// </summary>
    public class HeaderFooterSettings
    {
        internal const bool DefaultDifferentOddAndEvenPages = false;
        internal const bool DefaultDifferentFirstPage = false;
        internal const bool DefaultScaleWithDocument = true;
        internal const bool DefaultAlignWithPageMargins = true;
        private readonly Action updateOnChange;
        private readonly HeaderFooterContentCollection headerContents;
        private readonly HeaderFooterContentCollection footerContents;
        private bool differentOddAndEvenPages;
        private bool differentFirstPage;
        private bool scaleWithDocument;
        private bool alignWithPageMargins;

        internal HeaderFooterSettings(Action updateOnChange)
        {
            Guard.ThrowExceptionIfNull(updateOnChange, "updateOnChange");

            this.updateOnChange = updateOnChange;
            this.headerContents = new HeaderFooterContentCollection(updateOnChange);
            this.footerContents = new HeaderFooterContentCollection(updateOnChange);
            this.differentOddAndEvenPages = HeaderFooterSettings.DefaultDifferentOddAndEvenPages;
            this.differentFirstPage = HeaderFooterSettings.DefaultDifferentFirstPage;
            this.scaleWithDocument = HeaderFooterSettings.DefaultScaleWithDocument;
            this.alignWithPageMargins = HeaderFooterSettings.DefaultAlignWithPageMargins;
        }

        internal HeaderFooterSettings(HeaderFooterSettings other, Action updateOnChange)
        {
            Guard.ThrowExceptionIfNull(updateOnChange, "updateOnChange");

            this.updateOnChange = updateOnChange;
            this.headerContents = new HeaderFooterContentCollection(other.headerContents, updateOnChange);
            this.footerContents = new HeaderFooterContentCollection(other.footerContents, updateOnChange);
            this.differentOddAndEvenPages = other.differentOddAndEvenPages;
            this.differentFirstPage = other.differentFirstPage;
            this.scaleWithDocument = other.scaleWithDocument;
            this.alignWithPageMargins = other.alignWithPageMargins;
        }

        /// <summary>
        /// Gets or sets value indicating whether there should be different headers and footers for odd and even pages.
        /// </summary>
        /// <value>The value indicating whether there should be different headers and footers for odd and even pages.</value>
        public bool DifferentOddAndEvenPages
        {
            get
            {
                return this.differentOddAndEvenPages;
            }
            set
            {
                if (this.differentOddAndEvenPages != value)
                {
                    this.differentOddAndEvenPages = value;
                    this.updateOnChange();
                }
            }
        }

        /// <summary>
        /// Gets or sets value indicating whether there should be different headers and footers on the first page.
        /// </summary>
        /// <value>The value indicating whether there should be different headers and footers on the first page.</value>
        public bool DifferentFirstPage
        {
            get
            {
                return this.differentFirstPage;
            }
            set
            {
                if (this.differentFirstPage != value)
                {
                    this.differentFirstPage = value;
                    this.updateOnChange();
                }
            }
        }

        /// <summary>
        /// Gets or sets value indicating whether headers and footers should scale using the document content scale factor.
        /// </summary>
        /// <value>The value indicating whether headers and footers should scale using the document content scale factor.</value>
        public bool ScaleWithDocument
        {
            get
            {
                return this.scaleWithDocument;
            }
            set
            {
                if (this.scaleWithDocument != value)
                {
                    this.scaleWithDocument = value;
                    this.updateOnChange();
                }
            }
        }

        /// <summary>
        /// Gets or sets value indicating whether headers and footers should be aligned with the page margins.
        /// </summary>
        /// <value>The value indicating whether headers and footers should be aligned with the page margins.</value>
        public bool AlignWithPageMargins
        {
            get
            {
                return this.alignWithPageMargins;
            }
            set
            {
                if (this.alignWithPageMargins != value)
                {
                    this.alignWithPageMargins = value;
                    this.updateOnChange();
                }
            }
        }

        /// <summary>
        /// Gets the header content. 
        /// When DifferentOddAndEvenPages is true this content is used for the odd pages.
        /// Otherwise it is used for both even and odd pages.
        /// </summary>
        /// <value>The header content.</value>
        public HeaderFooterContent Header
        {
            get
            {
                return this.headerContents.DefaultContent;
            }
        }

        /// <summary>
        /// Gets the footer content.
        /// When DifferentOddAndEvenPages is true this content is used for the odd pages.
        /// Otherwise it is used for both even and odd pages.
        /// </summary>
        /// <value>The footer content.</value>
        public HeaderFooterContent Footer
        {
            get
            {
                return this.footerContents.DefaultContent;
            }
        }

        /// <summary>
        /// Gets the first header content.
        /// This content is used for the first page when DifferentFirstPage property is set to true. 
        /// </summary>
        /// <value>The first header content.</value>
        public HeaderFooterContent FirstPageHeader
        {
            get
            {
                return this.headerContents.FirstPageContent;
            }
        }

        /// <summary>
        /// Gets the first footer content.
        /// This content is used for the first page when DifferentFirstPage property is set to true. 
        /// </summary>
        /// <value>The first footer content.</value>
        public HeaderFooterContent FirstPageFooter
        {
            get
            {
                return this.footerContents.FirstPageContent;
            }
        }

        /// <summary>
        /// Gets the even header content.
        /// This content is used for the even pages when DifferentOddAndEvenPages property is set to true.
        /// </summary>
        /// <value>The even header content.</value>
        public HeaderFooterContent EvenPageHeader
        {
            get
            {
                return this.headerContents.EvenPageContent;
            }
        }

        /// <summary>
        /// Gets the even footer content.
        /// This content is used for the even pages when DifferentOddAndEvenPages property is set to true.
        /// </summary>
        /// <value>The even footer content.</value>
        public HeaderFooterContent EvenPageFooter
        {
            get
            {
                return this.footerContents.EvenPageContent;
            }
        }

        internal HeaderFooterContentCollection HeaderContents
        {
            get
            {
                return this.headerContents;
            }
        }

        internal HeaderFooterContentCollection FooterContents
        {
            get
            {
                return this.footerContents;
            }
        }

        internal IEnumerable<HeaderFooterContent> Contents
        {
            get
            {
                foreach (HeaderFooterContent content in this.headerContents)
                {
                    yield return content;
                }
                
                foreach (HeaderFooterContent content in this.footerContents)
                {
                    yield return content;
                }
            }
        }

        internal void CopyFrom(HeaderFooterSettings fromHeaderFooterSettings)
        {
            this.DifferentOddAndEvenPages = fromHeaderFooterSettings.DifferentOddAndEvenPages;
            this.DifferentFirstPage = fromHeaderFooterSettings.DifferentFirstPage;
            this.ScaleWithDocument = fromHeaderFooterSettings.ScaleWithDocument;
            this.AlignWithPageMargins = fromHeaderFooterSettings.AlignWithPageMargins;

            this.Header.CopyFrom(fromHeaderFooterSettings.Header);
            this.Footer.CopyFrom(fromHeaderFooterSettings.Footer);
            this.FirstPageHeader.CopyFrom(fromHeaderFooterSettings.FirstPageHeader);
            this.FirstPageFooter.CopyFrom(fromHeaderFooterSettings.FirstPageFooter);
            this.EvenPageHeader.CopyFrom(fromHeaderFooterSettings.EvenPageHeader);
            this.EvenPageFooter.CopyFrom(fromHeaderFooterSettings.EvenPageFooter);
        }

        internal void Clear()
        {
            this.DifferentOddAndEvenPages = HeaderFooterSettings.DefaultDifferentOddAndEvenPages;
            this.DifferentFirstPage = HeaderFooterSettings.DefaultDifferentFirstPage;
            this.ScaleWithDocument = HeaderFooterSettings.DefaultScaleWithDocument;
            this.AlignWithPageMargins = HeaderFooterSettings.DefaultAlignWithPageMargins;

            foreach (HeaderFooterContent content in this.Contents)
            {
                content.Clear();
            }
        }
    }
}