﻿using System;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Printing
{
    /// <summary>
    /// Represents a single page break.
    /// </summary>
    public class PageBreak
    {
        private readonly PageBreakType type;
        private readonly int index;
        private readonly int fromIndex;
        private readonly int toIndex;
        private readonly CellRange range;

        internal static readonly int HorizontalPageBreakMaximumIndex = SpreadsheetDefaultValues.ColumnCount - 1;
        internal static readonly int VerticalPageBreakMaximumIndex = SpreadsheetDefaultValues.RowCount - 1;

        internal PageBreak(PageBreakType type, int placementIndex, int fromIndex, int toIndex)
        {
            this.type = type;
            this.index = placementIndex;
            this.fromIndex = fromIndex;
            this.toIndex = toIndex;

            int fromRowIndex;
            int fromColumnIndex;
            int toRowIndex;
            int toColumnIndex;

            if (this.Type == PageBreakType.Horizontal)
            {
                fromRowIndex = this.Index;
                fromColumnIndex = this.FromIndex;
                toRowIndex = this.Index;
                toColumnIndex = this.ToIndex;
            }
            else
            {
                fromRowIndex = this.FromIndex;
                fromColumnIndex = this.Index;
                toRowIndex = this.ToIndex;
                toColumnIndex = this.Index;
            }

            fromRowIndex = Math.Max(fromRowIndex, 0);
            fromColumnIndex = Math.Max(fromColumnIndex, 0);
            toRowIndex = Math.Min(toRowIndex, SpreadsheetDefaultValues.RowCount - 1);
            toColumnIndex = Math.Min(toColumnIndex, SpreadsheetDefaultValues.ColumnCount - 1);

            this.range = new CellRange(fromRowIndex, fromColumnIndex, toRowIndex, toColumnIndex);
        }

        /// <summary>
        /// Gets the index.
        /// </summary>
        /// <value>The index.</value>
        public int Index
        {
            get
            {
                return this.index;
            }
        }

        /// <summary>
        /// Gets the type.
        /// </summary>
        /// <value>The type.</value>
        public PageBreakType Type
        {
            get
            {
                return this.type;
            }
        }

        /// <summary>
        /// Gets the starting index.
        /// </summary>
        /// <value>The starting index.</value>
        public int FromIndex
        {
            get
            {
                return this.fromIndex;
            }
        }

        /// <summary>
        /// Gets the ending index.
        /// </summary>
        /// <value>The ending index.</value>
        public int ToIndex
        {
            get
            {
                return this.toIndex;
            }
        }

        internal CellRange CellRange
        {
            get
            {
                return this.range;
            }
        }

        internal bool IsInfinite
        {
            get
            {
                return this.FromIndex == 0 &&
                    this.ToIndex == (this.Type == PageBreakType.Horizontal ? PageBreak.HorizontalPageBreakMaximumIndex : PageBreak.VerticalPageBreakMaximumIndex);
            }
        }

        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current
        /// <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
            return string.Format("PageBreakType: {0}, Index: {1}, FromIndex: {2}, ToIndex: {3}", this.Type, this.Index, this.FromIndex, this.ToIndex);
        }
    }
}