﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Maths.Number;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Printing
{
    /// <summary>
    /// Represents a collection of page breaks.
    /// </summary>
    public class PageBreaks
    {
        #region Fields

        private readonly Worksheet worksheet;
        private readonly Dictionary<int, List<PageBreak>> horizontalPageBreaks;
        private readonly Dictionary<int, List<PageBreak>> verticalPageBreaks;
        private readonly BeginEndCounter pageBreaksChangedUpdateCounter;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the horizontal page breaks.
        /// </summary>
        /// <value>The horizontal page breaks.</value>
        public IEnumerable<PageBreak> HorizontalPageBreaks
        {
            get
            {
                return PageBreaks.EnumeratePageBreaks(this.horizontalPageBreaks);
            }
        }

        /// <summary>
        /// Gets the vertical page breaks.
        /// </summary>
        /// <value>The vertical page breaks.</value>
        public IEnumerable<PageBreak> VerticalPageBreaks
        {
            get
            {
                return PageBreaks.EnumeratePageBreaks(this.verticalPageBreaks);
            }
        }

        internal IEnumerable<PageBreak> SortedHorizontalPageBreaks
        {
            get
            {
                return PageBreaks.GetSortedPageBreaks(this.HorizontalPageBreaks);
            }
        }

        internal IEnumerable<PageBreak> SortedVerticalPageBreaks
        {
            get
            {
                return PageBreaks.GetSortedPageBreaks(this.VerticalPageBreaks);
            }
        }

        internal bool HasAnyPageBreaks
        {
            get
            {
                return this.HorizontalPageBreaks.FirstOrDefault() != null || this.VerticalPageBreaks.FirstOrDefault() != null;
            }
        }

        internal BeginEndCounter PageBreaksChangedUpdateCounter
        {
            get
            {
                return this.pageBreaksChangedUpdateCounter;
            }
        }

        private PrintArea PrintArea
        {
            get
            {
                return this.worksheet.WorksheetPageSetup.PrintArea;
            }
        }

        #endregion


        #region Constructors

        internal PageBreaks(Worksheet worksheet)
        {
            this.worksheet = worksheet;
            this.horizontalPageBreaks = new Dictionary<int, List<PageBreak>>();
            this.verticalPageBreaks = new Dictionary<int, List<PageBreak>>();
            this.pageBreaksChangedUpdateCounter = new BeginEndCounter(this.OnPageBreaksChanged);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Tries to insert page breaks.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>True if succeeded.</returns>
        public bool TryInsertPageBreaks(int rowIndex, int columnIndex)
        {
            bool insertHorizontal = false;
            bool insertVertical = false;

            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                insertHorizontal = this.TryInsertHorizontalPageBreakInternal(rowIndex, columnIndex);
                insertVertical = this.TryInsertVerticalPageBreakInternal(rowIndex, columnIndex);
            }

            return insertHorizontal || insertVertical;
        }

        /// <summary>
        /// Tries to insert horizontal page break.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>True if succeeded.</returns>
        public bool TryInsertHorizontalPageBreak(int rowIndex, int columnIndex)
        {
            return this.TryInsertHorizontalPageBreakInternal(rowIndex, columnIndex);
        }

        /// <summary>
        /// Tries to insert vertical page break.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>True if succeeded.</returns>
        public bool TryInsertVerticalPageBreak(int rowIndex, int columnIndex)
        {
            return this.TryInsertVerticalPageBreakInternal(rowIndex, columnIndex);
        }

        /// <summary>
        /// Tries to remove page breaks.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>True if succeeded.</returns>
        public bool TryRemovePageBreaks(int rowIndex, int columnIndex)
        {
            bool removeHorizontal = false;
            bool removeVertical = false;

            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                removeHorizontal = this.TryRemoveHorizontalPageBreakInternal(rowIndex, columnIndex);
                removeVertical = this.TryRemoveVerticalPageBreakInternal(rowIndex, columnIndex);
            }

            return removeHorizontal || removeVertical;
        }

        /// <summary>
        /// Tries to remove horizontal page break.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>True if succeeded.</returns>
        public bool TryRemoveHorizontalPageBreak(int rowIndex, int columnIndex)
        {
            return this.TryRemoveHorizontalPageBreakInternal(rowIndex, columnIndex);
        }

        /// <summary>
        /// Tries to remove vertical page break.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>True if succeeded.</returns>
        public bool TryRemoveVerticalPageBreak(int rowIndex, int columnIndex)
        {
            return this.TryRemoveVerticalPageBreakInternal(rowIndex, columnIndex);
        }

        /// <summary>
        /// Clears this instance.
        /// </summary>
        public void Clear()
        {
            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                List<PageBreak> breaksToRemove = new List<PageBreak>();
                breaksToRemove.AddRange(this.HorizontalPageBreaks);
                breaksToRemove.AddRange(this.VerticalPageBreaks);

                this.RemovePageBreaks(breaksToRemove);
            }
        }

        internal bool TryInsertInfinitePageBreak(int pageBreakId, PageBreakType breakType)
        {
            if (pageBreakId <= 0)
            {
                return false;
            }

            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                List<PageBreak> breaksCollection = (breakType == PageBreakType.Horizontal) ?
                    this.HorizontalBreaksCollection(pageBreakId) :
                    this.VerticalBreaksCollection(pageBreakId);

                this.RemovePageBreaks(breaksCollection);
                this.InsertInfinitePageBreak(pageBreakId, breakType);
            }

            return true;
        }

        internal void MakeAllPageBreaksInfinite()
        {
            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                foreach (KeyValuePair<int, List<PageBreak>> keyValuePair in this.horizontalPageBreaks)
                {
                    List<PageBreak> breaksCollection = keyValuePair.Value;
                    if (breaksCollection != null && breaksCollection.Count > 0 && !breaksCollection[0].IsInfinite)
                    {
                        this.TryInsertInfinitePageBreak(keyValuePair.Key, PageBreakType.Horizontal);
                    }
                }

                foreach (KeyValuePair<int, List<PageBreak>> keyValuePair in this.verticalPageBreaks)
                {
                    List<PageBreak> breaksCollection = keyValuePair.Value;
                    if (breaksCollection != null && breaksCollection.Count > 0 && !breaksCollection[0].IsInfinite)
                    {
                        this.TryInsertInfinitePageBreak(keyValuePair.Key, PageBreakType.Vertical);
                    }
                }
            }
        }

        internal void AddPageBreak(PageBreak pageBreak)
        {
            var context = new AddRemovePageBreakCommandContext(this.worksheet, pageBreak);
            this.worksheet.ExecuteCommand(WorkbookCommands.AddPageBreak, context);
        }

        internal void RemovePageBreak(PageBreak pageBreak)
        {
            var context = new AddRemovePageBreakCommandContext(this.worksheet, pageBreak);
            this.worksheet.ExecuteCommand(WorkbookCommands.RemovePageBreak, context);
        }

        internal void AddPageBreakInternal(PageBreak pageBreak)
        {
            if (pageBreak.Type == PageBreakType.Horizontal)
            {
                this.HorizontalBreaksCollection(pageBreak.Index).Add(pageBreak);
            }
            else
            {
                this.VerticalBreaksCollection(pageBreak.Index).Add(pageBreak);
            }

            this.OnPageBreaksChanged();
        }

        internal void RemovePageBreakInternal(PageBreak pageBreak)
        {
            if (pageBreak.Type == PageBreakType.Horizontal)
            {
                this.HorizontalBreaksCollection(pageBreak.Index).Remove(pageBreak);
            }
            else
            {
                this.VerticalBreaksCollection(pageBreak.Index).Remove(pageBreak);
            }

            this.OnPageBreaksChanged();
        }

        internal void OnBeforePrintAreaTranslation(IEnumerable<CellRange> printAreaRangesToTranslate, CellRange rangeToInsertOrRemove, ShiftType shiftType)
        {
            this.BeginUpdate();

            List<PageBreak> breaksToTranslate = this.GetPageBreaksToTranslate(printAreaRangesToTranslate, shiftType);

            this.TranslatePageBreaks(breaksToTranslate, rangeToInsertOrRemove, shiftType);
        }

        internal void OnAfterPrintAreaTranslation()
        {
            this.TrimExtendPageBreaks(true);
            this.TrimExtendPageBreaks(false);

            this.EndUpdate();
        }
        
        private void TrimExtendPageBreaks(bool areBreaksHorizontal)
        {
            List<PageBreak> breaksToTrim = new List<PageBreak>(areBreaksHorizontal ? this.HorizontalPageBreaks : this.VerticalPageBreaks);

            foreach (PageBreak pageBreak in breaksToTrim)
            {
                if (!pageBreak.IsInfinite)
                {
                    int startRowIndex = areBreaksHorizontal ? pageBreak.Index : pageBreak.FromIndex;
                    int startColumnIndex = areBreaksHorizontal ? pageBreak.FromIndex : pageBreak.Index;
                    FiniteInterval<int> fromIndexInterval = this.GetPrintAreaContainedInterval(startRowIndex, startColumnIndex, areBreaksHorizontal);

                    int endRowIndex = areBreaksHorizontal ? pageBreak.Index : pageBreak.ToIndex;
                    int endColumnIndex = areBreaksHorizontal ? pageBreak.ToIndex : pageBreak.Index;
                    FiniteInterval<int> toIndexInterval = this.GetPrintAreaContainedInterval(endRowIndex, endColumnIndex, areBreaksHorizontal);

                    if (fromIndexInterval != null || toIndexInterval != null)
                    {
                        this.RemovePageBreak(pageBreak);

                        if (fromIndexInterval != null)
                        {
                            this.AddPageBreak(new PageBreak(pageBreak.Type, pageBreak.Index, fromIndexInterval.Start, fromIndexInterval.End));
                        }

                        if (toIndexInterval != null && !toIndexInterval.Equals(fromIndexInterval))
                        {
                            this.AddPageBreak(new PageBreak(pageBreak.Type, pageBreak.Index, toIndexInterval.Start, toIndexInterval.End));
                        }
                    }
                }
            }
        }

        private FiniteInterval<int> GetPrintAreaContainedInterval(int rowIndex, int columnIndex, bool isHorizontalInterval)
        {
            bool isOnlyOnAreasStart;

            if (isHorizontalInterval)
            {
                return this.TryGetHorizontalPrintAreaContainedInterval(rowIndex, columnIndex, out isOnlyOnAreasStart);
            }
            else
            {
                return this.TryGetVerticalPrintAreaContainedInterval(rowIndex, columnIndex, out isOnlyOnAreasStart);
            }
        }

        private List<PageBreak> GetPageBreaksToTranslate(IEnumerable<CellRange> printAreaRangesToTranslate, ShiftType shiftType)
        {
            List<PageBreak> breaksToTranslate = new List<PageBreak>();

            bool isShiftingRows = shiftType == ShiftType.Down || shiftType == ShiftType.Up;

            IEnumerable<PageBreak> breaksToTranslateDirectly = isShiftingRows ? this.HorizontalPageBreaks : this.VerticalPageBreaks;
            IEnumerable<PageBreak> breaksToCheck = isShiftingRows ? this.VerticalPageBreaks : this.HorizontalPageBreaks;

            breaksToTranslate.AddRange(breaksToTranslateDirectly);

            foreach (PageBreak pageBreak in breaksToCheck)
            {
                if (PageBreaks.ShouldTranslate(pageBreak, printAreaRangesToTranslate))
                {
                    breaksToTranslate.Add(pageBreak);
                }
            }

            return breaksToTranslate;
        }

        private static bool ShouldTranslate(PageBreak pageBreak, IEnumerable<CellRange> printAreaRangesToTranslate)
        {
            bool isHorizontal = pageBreak.Type == PageBreakType.Horizontal;

            foreach (CellRange printRange in printAreaRangesToTranslate)
            {
                int rangeStart = printRange.StartIndex(!isHorizontal);
                int rangeEnd = printRange.EndIndex(!isHorizontal);

                bool isStartingOrEndingAtPrintRange = (rangeStart == pageBreak.FromIndex || rangeEnd == pageBreak.ToIndex) &&
                    (printRange.StartIndex(isHorizontal) <= pageBreak.Index && pageBreak.Index <= printRange.EndIndex(isHorizontal));

                if (isStartingOrEndingAtPrintRange)
                {
                    return true;
                }
            }

            return false;
        }

        private void TranslatePageBreaks(List<PageBreak> pageBreaks, CellRange rangeToInsertOrRemove, ShiftType shiftType)
        {   
            foreach (PageBreak pageBreak in pageBreaks)
            {
                PageBreak translatedBreak;
                if (pageBreak.TryTranslate(rangeToInsertOrRemove, shiftType, out translatedBreak))
                {
                    this.RemovePageBreak(pageBreak);

                    if (translatedBreak != null)
                    {
                        this.AddPageBreak(translatedBreak);
                    }
                }
            }
        }
        
        private bool TryRemoveHorizontalPageBreakInternal(int rowIndex, int columnIndex)
        {
            List<PageBreak> breaksCollection = this.HorizontalBreaksCollection(rowIndex);
            PageBreak existingPageBreak = PageBreaks.TryGetExistingPageBreak(breaksCollection, columnIndex);

            if (existingPageBreak != null)
            {
                this.RemovePageBreak(existingPageBreak);

                return true;
            }

            return false;
        }

        private bool TryRemoveVerticalPageBreakInternal(int rowIndex, int columnIndex)
        {
            List<PageBreak> breaksCollection = this.VerticalBreaksCollection(columnIndex);
            PageBreak existingPageBreak = PageBreaks.TryGetExistingPageBreak(breaksCollection, rowIndex);

            if (existingPageBreak != null)
            {
                this.RemovePageBreak(existingPageBreak);

                return true;
            }

            return false;
        }
        
        private bool TryInsertHorizontalPageBreakInternal(int rowIndex, int columnIndex)
        {
            List<PageBreak> breaksCollection = this.HorizontalBreaksCollection(rowIndex);
            PageBreak existingPageBreak = PageBreaks.TryGetExistingPageBreak(breaksCollection, columnIndex);

            if (existingPageBreak != null)
            {
                return false;
            }

            bool isOnlyOnTopRowsInAreas;
            FiniteInterval<int> containingInterval = this.TryGetHorizontalPrintAreaContainedInterval(rowIndex, columnIndex, out isOnlyOnTopRowsInAreas);

            if (containingInterval != null)
            {
                if (!isOnlyOnTopRowsInAreas)
                {
                    this.AddPageBreak(new PageBreak(PageBreakType.Horizontal, rowIndex, containingInterval.Start, containingInterval.End));
                    return true;
                }
            }
            else
            {
                return this.TryInsertInfinitePageBreak(rowIndex, PageBreakType.Horizontal);
            }

            return false;
        }

        private FiniteInterval<int> TryGetHorizontalPrintAreaContainedInterval(int rowIndex, int columnIndex, out bool isOnlyOnTopRowsInAreas)
        {
            isOnlyOnTopRowsInAreas = true;
            FiniteIntervalsUnion<int> intervals = new FiniteIntervalsUnion<int>();

            foreach (var range in this.PrintArea.Ranges)
            {
                if (range.Contains(rowIndex, columnIndex) && range.FromIndex.RowIndex < rowIndex)
                {
                    isOnlyOnTopRowsInAreas = false;
                }

                if (range.FromIndex.RowIndex <= rowIndex && rowIndex <= range.ToIndex.RowIndex)
                {
                    intervals.AddInterval(range.FromIndex.ColumnIndex, range.ToIndex.ColumnIndex);
                }
            }

            return intervals.GetContainingInterval(columnIndex);
        }

        private bool TryInsertVerticalPageBreakInternal(int rowIndex, int columnIndex)
        {
            List<PageBreak> breaksCollection = this.VerticalBreaksCollection(columnIndex);
            PageBreak existingPageBreak = PageBreaks.TryGetExistingPageBreak(breaksCollection, rowIndex);
            if (existingPageBreak != null)
            {
                return false;
            }

            bool isOnlyOnLeftColumnsInAreas;
            FiniteInterval<int> containingInterval = this.TryGetVerticalPrintAreaContainedInterval(rowIndex, columnIndex, out isOnlyOnLeftColumnsInAreas);

            if (containingInterval != null)
            {
                if (!isOnlyOnLeftColumnsInAreas)
                {
                    this.AddPageBreak(new PageBreak(PageBreakType.Vertical, columnIndex, containingInterval.Start, containingInterval.End));
                    return true;
                }
            }
            else
            {
                return this.TryInsertInfinitePageBreak(columnIndex, PageBreakType.Vertical);
            }

            return false;
        }

        private FiniteInterval<int> TryGetVerticalPrintAreaContainedInterval(int rowIndex, int columnIndex, out bool isOnlyOnLeftColumnsInAreas)
        {
            isOnlyOnLeftColumnsInAreas = true;
            FiniteIntervalsUnion<int> intervals = new FiniteIntervalsUnion<int>();

            foreach (var range in this.PrintArea.Ranges)
            {
                if (range.Contains(rowIndex, columnIndex) && range.FromIndex.ColumnIndex < columnIndex)
                {
                    isOnlyOnLeftColumnsInAreas = false;
                }

                if (range.FromIndex.ColumnIndex <= columnIndex && columnIndex <= range.ToIndex.ColumnIndex)
                {
                    intervals.AddInterval(range.FromIndex.RowIndex, range.ToIndex.RowIndex);
                }
            }

            return intervals.GetContainingInterval(rowIndex);
        }

        private void RemovePageBreaks(IEnumerable<PageBreak> pageBreaks)
        {
            List<PageBreak> breaksToRemove = new List<PageBreak>(pageBreaks);

            foreach (PageBreak pageBreak in breaksToRemove)
            {
                this.RemovePageBreak(pageBreak);
            }
        }

        private void InsertInfinitePageBreak(int pageBreakId, PageBreakType breakType)
        {
            int maxIndex = breakType == PageBreakType.Horizontal ? PageBreak.HorizontalPageBreakMaximumIndex : PageBreak.VerticalPageBreakMaximumIndex;

            this.AddPageBreak(new PageBreak(breakType, pageBreakId, 0, maxIndex));
        }
              
        private List<PageBreak> HorizontalBreaksCollection(int rowIndex)
        {
            List<PageBreak> breaksCollection;
            if (!this.horizontalPageBreaks.TryGetValue(rowIndex, out breaksCollection))
            {
                breaksCollection = new List<PageBreak>();
                this.horizontalPageBreaks.Add(rowIndex, breaksCollection);
            }

            return breaksCollection;
        }

        private List<PageBreak> VerticalBreaksCollection(int columnIndex)
        {
            List<PageBreak> breaksCollection;
            if (!this.verticalPageBreaks.TryGetValue(columnIndex, out breaksCollection))
            {
                breaksCollection = new List<PageBreak>();
                this.verticalPageBreaks.Add(columnIndex, breaksCollection);
            }

            return breaksCollection;
        }

        private void BeginUpdate()
        {
            this.PageBreaksChangedUpdateCounter.BeginUpdate();
            this.worksheet.BeginUndoGroup();
        }

        private void EndUpdate()
        {
            this.worksheet.EndUndoGroup();
            this.PageBreaksChangedUpdateCounter.EndUpdate();
        }

        private static PageBreak TryGetExistingPageBreak(List<PageBreak> pageBreakCollection, int breakInsertionPoint)
        {
            for (int i = 0; i < pageBreakCollection.Count; i++)
            {
                PageBreak pageBreak = pageBreakCollection[i];
                if (pageBreak.FromIndex <= breakInsertionPoint && breakInsertionPoint <= pageBreak.ToIndex)
                {
                    return pageBreak;
                }
            }

            return null;                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  
        }

        private static IEnumerable<PageBreak> EnumeratePageBreaks(Dictionary<int, List<PageBreak>> pageBreaks)
        {
            foreach (List<PageBreak> breaksCollection in pageBreaks.Values)
            {
                foreach (PageBreak pageBreak in breaksCollection)
                {
                    yield return pageBreak;
                }
            }
        }

        private static PageBreak[] GetSortedPageBreaks(IEnumerable<PageBreak> pageBreaks)
        {
            return pageBreaks.OrderBy(pageBreak => pageBreak.Index).ToArray();
        }

        internal void CopyFrom(PageBreaks fromPageBreaks)
        {
            foreach (KeyValuePair<int, List<PageBreak>> indexToHorizontalPageBreak in fromPageBreaks.horizontalPageBreaks)
            {
                List<PageBreak> horizontalPageBreaksCopied = new List<PageBreak>();
                foreach (PageBreak horizontalPageBreak in indexToHorizontalPageBreak.Value)
                {
                    horizontalPageBreaksCopied.Add(horizontalPageBreak);
                }

                this.horizontalPageBreaks[indexToHorizontalPageBreak.Key] = horizontalPageBreaksCopied;
            }

            foreach (KeyValuePair<int, List<PageBreak>> indexToVerticalPageBreak in fromPageBreaks.verticalPageBreaks)
            {
                List<PageBreak> verticalPageBreaksCopied = new List<PageBreak>();
                foreach (PageBreak verticalPageBreak in indexToVerticalPageBreak.Value)
                {
                    verticalPageBreaksCopied.Add(verticalPageBreak);
                }

                this.verticalPageBreaks[indexToVerticalPageBreak.Key] = new List<PageBreak>(indexToVerticalPageBreak.Value);
            }
        }

        #endregion


        #region Events

        // TODO: Think of firing PageSetupChanged event
        internal event EventHandler PageBreaksChanged;

        private void OnPageBreaksChanged()
        {
            if (this.PageBreaksChangedUpdateCounter.BeginUpdateCounter > 0)
            {
                return;
            }

            if (this.PageBreaksChanged != null)
            {
                this.PageBreaksChanged(this, EventArgs.Empty);
            }
        }

        #endregion
    }
}