﻿using System;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Printing
{
    /// <summary>
    /// Represents a collection of page margins.
    /// </summary>
    public class PageMargins
    {
        #region Fields

        /// <summary>
        /// The normal margins.
        /// </summary>
        public static readonly PageMargins NormalMargins;

        /// <summary>
        /// The wide margins.
        /// </summary>
        public static readonly PageMargins WideMargins;

        /// <summary>
        /// The narrow margins.
        /// </summary>
        public static readonly PageMargins NarrowMargins;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the left margin.
        /// </summary>
        /// <value>The left margin.</value>
        public double Left { get; private set; }

        /// <summary>
        /// Gets the bottom margin.
        /// </summary>
        /// <value>The bottom margin.</value>
        public double Bottom { get; private set; }

        /// <summary>
        /// Gets the right margin.
        /// </summary>
        /// <value>The right margin.</value>
        public double Right { get; private set; }

        /// <summary>
        /// Gets the top margin.
        /// </summary>
        /// <value>The top margin.</value>
        public double Top { get; private set; }

        /// <summary>
        /// Gets or sets the header margin.
        /// </summary>
        /// <value>The header margin.</value>
        public double Header { get; private set; }

        /// <summary>
        /// Gets or sets the footer margin.
        /// </summary>
        /// <value>The footer margin.</value>
        public double Footer { get; private set; }

        #endregion


        #region Constructor

        static PageMargins()
        {
            PageMargins.NormalMargins = new PageMargins(UnitHelper.InchToDip(0.7), UnitHelper.InchToDip(0.75), UnitHelper.InchToDip(0.3));
            PageMargins.WideMargins = new PageMargins(UnitHelper.InchToDip(1), UnitHelper.InchToDip(1), UnitHelper.InchToDip(0.5));
            PageMargins.NarrowMargins = new PageMargins(UnitHelper.InchToDip(0.25), UnitHelper.InchToDip(0.75), UnitHelper.InchToDip(0.3));
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="PageMargins" /> class.
        /// </summary>
        public PageMargins() 
            : this(0)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="PageMargins" /> class.
        /// </summary>
        /// <param name="leftRight">The left and the right margin.</param>
        /// <param name="topBottom">The top and the bottom margin.</param>
        public PageMargins(double leftRight, double topBottom)
            : this(leftRight, topBottom, 0)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="PageMargins" /> class.
        /// </summary>
        /// <param name="left">The left margin.</param>
        /// <param name="top">The top margin.</param>
        /// <param name="right">The right margin.</param>
        /// <param name="bottom">The bottom margin.</param>
        public PageMargins(double left, double top, double right, double bottom) 
            : this(left, top, right, bottom, 0, 0)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="PageMargins" /> class.
        /// </summary>
        /// <param name="all">All margins.</param>
        public PageMargins(double all) 
            : this(all, all, all)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="PageMargins" /> class.
        /// </summary>
        /// <param name="leftRight">The left and right margins.</param>
        /// <param name="topBottom">The top and bottom margins.</param>
        /// <param name="headerFooter">The header and footer margins.</param>
        public PageMargins(double leftRight, double topBottom, double headerFooter) 
            : this(leftRight, topBottom, leftRight, topBottom, headerFooter, headerFooter)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="PageMargins" /> class.
        /// </summary>
        /// <param name="left">The left margin.</param>
        /// <param name="top">The top margin.</param>
        /// <param name="right">The right margin.</param>
        /// <param name="bottom">The bottom margin.</param>
        /// <param name="header">The header margin.</param>
        /// <param name="footer">The footer margin.</param>
        public PageMargins(double left, double top, double right, double bottom, double header, double footer)
        {
            this.Left = left;
            this.Right = right;
            this.Top = top;
            this.Bottom = bottom;
            this.Header = header;
            this.Footer = footer;
        }

        #endregion


        #region Methods

        internal static bool AreValidMarginsWithinPageSize(PageMargins pageMargins, Size rotatedPageSize, Size minimumContent)
        {
            return IsValidMarginInequation(pageMargins.Left, pageMargins.Right, rotatedPageSize.Width, minimumContent.Width) && 
                IsValidMarginInequation(pageMargins.Top, pageMargins.Bottom, rotatedPageSize.Height, minimumContent.Height) && 
                IsValidMarginInequation(pageMargins.Header, pageMargins.Footer, rotatedPageSize.Height);
        }

        private static bool IsValidMarginInequation(double firstMargin, double secondMargin, double maxMarginSum, double minimumContentSize = 0)
        {
            return firstMargin >= 0 &&
                secondMargin >= 0 &&
                minimumContentSize >= 0 &&
                (firstMargin + secondMargin + minimumContentSize) < maxMarginSum;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            PageMargins other = obj as PageMargins;

            if (other == null)
            {
                return false;
            }

            return this.Bottom == other.Bottom &&
                this.Top == other.Top &&
                this.Right == other.Right &&
                this.Left == other.Left &&
                this.Header == other.Header &&
                this.Footer == other.Footer;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return this.Bottom.GetHashCode() ^
                this.Top.GetHashCode() ^
                this.Left.GetHashCode() ^
                this.Right.GetHashCode() ^
                this.Footer.GetHashCode() ^
                this.Header.GetHashCode();
        }

        #endregion
    }
}