﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.Core;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Printing
{
    /// <summary>
    /// Represents a print area.
    /// </summary>
    public class PrintArea : CellRangeInsertedOrRemovedAwareOrderedCollection
    {
        #region Fields

        private readonly Worksheet worksheet;
        private readonly BeginEndCounter printAreaChangedUpdateCounter;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the ranges to print.
        /// </summary>
        /// <value>The ranges to print.</value>
        public IEnumerable<CellRange> Ranges
        {
            get
            {
                for (int i = 0; i < this.PrintRanges.Count; i++)
                {
                    yield return this.PrintRanges[i];
                }
            }
        }

        /// <summary>
        /// Gets the value indicating whether there are print ranges.
        /// </summary>
        /// <value>The value indicating whether there are print ranges.</value>
        public bool HasPrintAreaRanges
        {
            get
            {
                return this.PrintRanges.Count > 0;
            }
        }

        internal int RangesCount
        {
            get
            {
                return this.PrintRanges.Count;
            }
        }

        internal BeginEndCounter PrintAreaChangedUpdateCounter
        {
            get
            {
                return this.printAreaChangedUpdateCounter;
            }
        }

        private List<CellRange> PrintRanges
        {
            get
            {
                return base.RangesList;
            }
        }

        private PageBreaks PageBreaks
        {
            get
            {
                return this.worksheet.WorksheetPageSetup.PageBreaks;
            }
        }

        #endregion


        #region Constructors

        internal PrintArea(Worksheet worksheet)
            : base(worksheet.Cells)
        {
            this.worksheet = worksheet;
            this.printAreaChangedUpdateCounter = new BeginEndCounter(this.OnPrintAreaChanged);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Sets the print area.
        /// </summary>
        /// <param name="range">The range.</param>
        public void SetPrintArea(CellRange range)
        {
            this.SetPrintArea(new CellRange[] { range });
        }

        /// <summary>
        /// Sets the print area.
        /// </summary>
        /// <param name="ranges">The ranges.</param>
        public void SetPrintArea(IEnumerable<CellRange> ranges)
        {
            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                this.Clear();
                this.TryAddToPrintArea(ranges);
            }
        }

        /// <summary>
        /// Tries to add to print area.
        /// </summary>
        /// <param name="range">The range.</param>
        /// <returns></returns>
        public bool TryAddToPrintArea(CellRange range)
        {
            return this.TryAddToPrintArea(new CellRange[] { range });
        }

        /// <summary>
        /// Tries to add to print area.
        /// </summary>
        /// <param name="ranges">The ranges.</param>
        /// <returns></returns>
        public bool TryAddToPrintArea(IEnumerable<CellRange> ranges)
        {
            if (this.CanAddToPrintArea(ranges))
            {
                this.InsertRange(this.RangesCount, ranges);

                return true;
            }

            return false;
        }

        /// <summary>
        /// Determines whether the specified range can be added.
        /// </summary>
        /// <param name="range">The range.</param>
        /// <returns>The value indicating wheter the specified range can be added.</returns>
        public bool CanAddToPrintArea(CellRange range)
        {
            return this.CanAddToPrintArea(new CellRange[] { range });
        }

        /// <summary>
        /// Determines whether the specified ranges can be added.
        /// </summary>
        /// <param name="ranges">The ranges.</param>
        /// <returns>The value indicating wheter the specified ranges can be added.</returns>
        public bool CanAddToPrintArea(IEnumerable<CellRange> ranges)
        {
            foreach (var range in ranges)
            {
                foreach (var existingRange in this.Ranges)
                {
                    if(range.IntersectsWith(existingRange))
                    {
                        return false;
                    }
                }
            }

            return true;
        }

        /// <summary>
        /// Clears this instance.
        /// </summary>
        public void Clear()
        {
            if (this.HasPrintAreaRanges)
            {
                using (new UpdateScope(this.worksheet.BeginUndoGroup, this.worksheet.EndUndoGroup))
                {
                    this.RemoveRange(0, this.RangesCount);
                }
            }
        }

        internal void InsertAt(int index, CellRange range)
        {
            this.InsertRange(index, new CellRange[] { range });
        }

        internal void InsertRange(int index, IEnumerable<CellRange> ranges)
        {
            var context = new AddToRemoveFromPrintAreaCommandContext(this.worksheet, ranges, index);
            this.worksheet.ExecuteCommand(WorkbookCommands.AddToPrintArea, context);
        }

        internal void RemoveAt(int index)
        {
            this.RemoveRange(index, 1);
        }

        internal void RemoveRange(int index, int count)
        {
            IEnumerable<CellRange> rangesToRemove = this.PrintRanges.GetRange(index, count);
            var context = new AddToRemoveFromPrintAreaCommandContext(this.worksheet, rangesToRemove, index);
            this.worksheet.ExecuteCommand(WorkbookCommands.RemoveFromPrintArea, context);
        }

        internal void InsertRangesInternal(IEnumerable<CellRange> range, int index)
        {
            this.PrintRanges.InsertRange(index, range);
            this.OnPrintAreaChanged();
        }

        internal void RemoveRangesInternal(int index, int count)
        {
            this.PrintRanges.RemoveRange(index, count);
            this.OnPrintAreaChanged();
        }

        internal void CopyFrom(PrintArea fromPrintArea)
        {
            foreach (CellRange cellRange in fromPrintArea.CellRanges)
            {
                this.CellRanges.Add(cellRange);
            }
        }

        /// <summary>
        /// Inserts the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <param name="range">The range.</param>
        protected override void Insert(int index, CellRange range)
        {
            this.InsertAt(index, range);
        }

        /// <summary>
        /// Removes the cell range at.
        /// </summary>
        /// <param name="index">The index.</param>
        protected override void RemoveCellRangeAt(int index)
        {
            this.RemoveAt(index);
        }

        /// <summary>
        /// Called before translation.
        /// </summary>
        /// <param name="oldAndNewTranslatedRangesPositions">The old and new translated ranges positions.</param>
        /// <param name="rangeToInsertOrRemove">The range to insert or remove.</param>
        /// <param name="shiftType">Type of the shift.</param>
        protected override void OnBeforeTranslation(Dictionary<CellRange, CellRange> oldAndNewTranslatedRangesPositions, CellRange rangeToInsertOrRemove, ShiftType shiftType)
        {
            this.BeginUpdate();
            this.PageBreaks.OnBeforePrintAreaTranslation(oldAndNewTranslatedRangesPositions.Keys, rangeToInsertOrRemove, shiftType);
        }

        /// <summary>
        /// Called after translation.
        /// </summary>
        /// <param name="oldAndNewTranslatedRangesPositions">The old and new translated ranges positions.</param>
        /// <param name="rangeToInsertOrRemove">The range to insert or remove.</param>
        /// <param name="shiftType">Type of the shift.</param>
        protected override void OnAfterTranslation(Dictionary<CellRange, CellRange> oldAndNewTranslatedRangesPositions, CellRange rangeToInsertOrRemove, ShiftType shiftType)
        {
            this.PageBreaks.OnAfterPrintAreaTranslation();
            this.EndUpdate();
        }

        private void BeginUpdate()
        {
            this.PrintAreaChangedUpdateCounter.BeginUpdate();
            this.worksheet.BeginUndoGroup();
        }

        private void EndUpdate()
        {
            this.worksheet.EndUndoGroup();
            this.PrintAreaChangedUpdateCounter.EndUpdate();
        }

        #endregion


        #region Events

        // TODO: Think of firing PageSetupChanged when this event gets fired?
        internal event EventHandler PrintAreaChanged;

        private void OnPrintAreaChanged()
        {
            if (this.PrintAreaChangedUpdateCounter.BeginUpdateCounter > 0)
            {
                return;
            }

            if (this.PrintAreaChanged != null)
            {
                this.PrintAreaChanged(this, EventArgs.Empty);
            }
        }

        #endregion
    }
}