﻿using System;
using System.Windows;
using Telerik.Windows.Documents.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Printing
{
    /// <summary>
    /// Represents a base class for sheet page setup.
    /// </summary>
    public abstract class SheetPageSetupBase
    {
        #region Fields

        private readonly HeaderFooterSettings headerFooterSettings;
        private PaperTypes paperType;
        private PageOrientation pageOrientation;
        private Size pageSize;
        private Size rotatedPageSize;
        private int? firstPageNumber;
        private readonly ISheet sheet;
        private PageMargins pageMargins;
        private int beginUpdateCounter;
        private bool shouldFirePageSetupChanged;

        #endregion


        #region Properties

        /// <summary>
        /// Gets or sets the type of the paper.
        /// </summary>
        /// <value>The type of the paper.</value>
        public PaperTypes PaperType
        {
            get
            {
                return this.paperType;
            }
            set
            {
                if (this.paperType != value)
                {
                    this.paperType = value;
                    this.OnPaperTypeChanged();
                }
            }
        }

        /// <summary>
        /// Gets or sets the page orientation.
        /// </summary>
        /// <value>The page orientation.</value>
        public PageOrientation PageOrientation
        {
            get
            {
                return this.pageOrientation;
            }
            set
            {
                if (this.pageOrientation != value)
                {
                    this.pageOrientation = value;
                    this.OnPageOrientationChanged();
                }
            }
        }

        /// <summary>
        /// Gets the size of the page.
        /// </summary>
        /// <value>The size of the page.</value>
        public Size PageSize
        {
            get
            {
                return this.pageSize;
            }
        }

        internal Size RotatedPageSize
        {
            get
            {
                return this.rotatedPageSize;
            }
            set
            {
                this.rotatedPageSize = value;
            }
        }

        internal int? FirstPageNumber
        {
            get
            {
                return this.firstPageNumber;
            }
            set
            {
                if (this.firstPageNumber != value)
                {
                    if (value != null)
                    {
                        Guard.ThrowExceptionIfOutOfRange(
                            SpreadsheetDefaultValues.MinFirstPageNumber,
                            SpreadsheetDefaultValues.MaxFirstPageNumber,
                            (int)value,
                            "firstPageNumber");
                    }   

                    this.firstPageNumber = value;
                    this.OnPageSetupChanged();
                }
            }
        }

        /// <summary>
        /// Gets or sets the margins.
        /// </summary>
        /// <value>The margins.</value>
        public PageMargins Margins 
        {
            get
            {
                return this.pageMargins;
            }
            set
            {
                if (!TelerikHelper.EqualsOfT(this.pageMargins, value))
                {
                    this.pageMargins = value;
                    this.OnPageSetupChanged();
                }
            }
        }

        /// <summary>
        /// Gets the header footer settings.
        /// </summary>
        /// <value>The header footer settings.</value>
        public HeaderFooterSettings HeaderFooterSettings 
        {
            get
            {
                return this.headerFooterSettings;
            }
        }

        #endregion


        #region Constructor

        internal SheetPageSetupBase(Sheet sheet)
        {
            this.headerFooterSettings = new HeaderFooterSettings(this.OnPageSetupChanged);
            this.paperType = PaperTypes.Letter;
            this.pageSize = PaperTypeConverter.ToSize(this.PaperType);
            this.PageOrientation = Documents.Model.PageOrientation.Portrait;
            this.rotatedPageSize = 
                (this.PageOrientation == PageOrientation.Portrait) ? this.PageSize : new Size(this.PageSize.Height, this.PageSize.Width);
            this.Margins = PageMargins.NormalMargins;
            this.sheet = sheet;
            this.beginUpdateCounter = 0;
        }

        #endregion


        #region Methods

        private void OnPaperTypeChanged()
        {
            this.pageSize = PaperTypeConverter.ToSize(this.PaperType);
            this.UpdateRotatedPageSize();

            this.OnPageSetupChanged();
        }

        private void OnPageOrientationChanged()
        {
            this.UpdateRotatedPageSize();

            this.OnPageSetupChanged();
        }

        private void UpdateRotatedPageSize()
        {
            this.RotatedPageSize = CalculateRotatedPageSize(this.PageSize, this.PageOrientation);
        }

        /// <summary>
        /// Invalidates the sheet layout.
        /// </summary>
        protected void InvalidateSheetLayout()
        {
            this.sheet.InvalidateLayout();
        }

        /// <summary>
        /// Suspends the sheet layout update.
        /// </summary>
        protected void SuspendSheetLayoutUpdate()
        {
            this.sheet.SuspendLayoutUpdate();
        }

        /// <summary>
        /// Resumes the sheet layout update.
        /// </summary>
        protected void ResumeSheetLayoutUpdate()
        {
            this.sheet.ResumeLayoutUpdate();
        }

        internal static Size CalculateRotatedPageSize(PaperTypes paperType, PageOrientation pageOrientation)
        {
            return CalculateRotatedPageSize(PaperTypeConverter.ToSize(paperType), pageOrientation);
        }

        internal static Size CalculateRotatedPageSize(Size pageSize, PageOrientation pageOrientation)
        {
            return (pageOrientation == PageOrientation.Portrait) ? pageSize : new Size(pageSize.Height, pageSize.Width);
        }

        internal void BeginUpdate()
        {
            this.beginUpdateCounter++;
        }

        internal void EndUpdate()
        {
            this.beginUpdateCounter--;

            if (this.beginUpdateCounter < 0)
            {
                throw new InvalidOperationException("There is no active update to end.");
            }

            if (beginUpdateCounter == 0 && this.shouldFirePageSetupChanged)
            {
                this.OnPageSetupChanged();
            }
        }

        internal virtual void CopyFrom(SheetPageSetupBase fromSheetPageSetupBase)
        {
            this.paperType = fromSheetPageSetupBase.PaperType;
            this.pageOrientation = fromSheetPageSetupBase.PageOrientation;
            this.pageSize = fromSheetPageSetupBase.PageSize;
            
            this.headerFooterSettings.CopyFrom(fromSheetPageSetupBase.HeaderFooterSettings);
        }

        #endregion


        #region Events

        internal event EventHandler PageSetupChanged;

        // TODO: Think whether we need to know what exactly has changed!
        internal void OnPageSetupChanged()
        {
            this.shouldFirePageSetupChanged = true;

            if (this.beginUpdateCounter == 0 && this.PageSetupChanged != null)
            {
                this.PageSetupChanged(this, EventArgs.Empty);
                this.shouldFirePageSetupChanged = false;
            }
        }

        #endregion
    }
}