﻿using System;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Printing
{
    /// <summary>
    /// Represents worksheet page setup.
    /// </summary>
    public class WorksheetPageSetup : SheetPageSetup<WorksheetPrintOptions>
    {
        #region Fields

        private bool fitToPages;
        private int fitToPagesTall;
        private int fitToPagesWide;
        private Size scaleFactor;
        private readonly PrintArea printArea;
        private readonly PageBreaks pageBreaks;
        private readonly WorksheetPrintOptions printOptions;
        private PageOrder pageOrder;
        private bool centerHorizontally;
        private bool centerVertically;

        #endregion


        #region Properties

        /// <summary>
        /// Gets or sets the page order.
        /// </summary>
        /// <value>The page order.</value>
        public PageOrder PageOrder
        {
            get
            {
                return this.pageOrder;
            }
            set
            {
                if (this.pageOrder != value)
                {
                    this.pageOrder = value;
                    this.OnPageSetupChanged();
                }
            }
        }

        /// <summary>
        /// Gets or sets the value indicating whether the print content should be centered horizontally.
        /// </summary>
        /// <value>The value indicating whether the print content should be centered horizontally.</value>
        public bool CenterHorizontally
        {
            get
            {
                return this.centerHorizontally;
            }
            set
            {
                if (this.centerHorizontally != value)
                {
                    this.centerHorizontally = value;
                    this.OnPageSetupChanged();
                }
            }
        }

        /// <summary>
        /// Gets or sets the value indicating whether the print content should be centered vertically.
        /// </summary>
        /// <value>The value indicating whether the print content should be centered vertically.</value>
        public bool CenterVertically
        {
            get
            {
                return this.centerVertically;
            }
            set
            {
                if (this.centerVertically != value)
                {
                    this.centerVertically = value;
                    this.OnPageSetupChanged();
                }
            }
        }

        /// <summary>
        /// Gets the print area.
        /// </summary>
        /// <value>The print area.</value>
        public PrintArea PrintArea
        {
            get
            {
                return this.printArea;
            }
        }

        /// <summary>
        /// Gets the page breaks.
        /// </summary>
        /// <value>The page breaks.</value>
        public PageBreaks PageBreaks
        {
            get
            {
                return this.pageBreaks;
            }
        }

        /// <summary>
        /// Gets or sets the scale factor.
        /// </summary>
        /// <value>The scale factor.</value>
        public Size ScaleFactor
        {
            get
            {
                return this.scaleFactor;
            }
            set
            {
                if (this.scaleFactor != value)
                {
                    Guard.ThrowExceptionIfOutOfRange(SpreadsheetDefaultValues.MinPageScaleFactor,
                        SpreadsheetDefaultValues.MaxPageScaleFactor, value.Width, "scaleFactor");
                    Guard.ThrowExceptionIfOutOfRange(SpreadsheetDefaultValues.MinPageScaleFactor,
                        SpreadsheetDefaultValues.MaxPageScaleFactor, value.Height, "scaleFactor");

                    this.scaleFactor = value;
                    this.OnPageSetupChanged();
                }
            }
        }

        /// <summary>
        /// Gets or sets the number of pages tall the worksheet will be scaled to when it's printed.
        /// </summary>
        public int FitToPagesTall
        {
            get
            {
                return this.fitToPagesTall;
            }
            set
            {
                if (this.fitToPagesTall != value)
                {
                    Guard.ThrowExceptionIfOutOfRange(SpreadsheetDefaultValues.MinFitToPagesCount,
                        SpreadsheetDefaultValues.MaxFitToPagesCount, value, "fitToPagesTall");

                    this.fitToPagesTall = value;
                    this.OnPageSetupChanged();
                }
            }
        }

        /// <summary>
        /// Gets or sets the number of pages wide the worksheet will be scaled to when it's printed.
        /// </summary>
        public int FitToPagesWide
        {
            get
            {
                return this.fitToPagesWide;
            }
            set
            {
                if (this.fitToPagesWide != value)
                {
                    Guard.ThrowExceptionIfOutOfRange(SpreadsheetDefaultValues.MinFitToPagesCount,
                        SpreadsheetDefaultValues.MaxFitToPagesCount, value, "fitToPagesWide");

                    this.fitToPagesWide = value;
                    this.OnPageSetupChanged();
                }
            }
        }

        /// <summary>
        /// Gets or sets the value indicating whether the worksheet will be scaled according to a number of pages.
        /// </summary>
        /// <value>
        /// If <c>true</c>, the worksheet will be scaled according to the FitToPagesWide and FitToPagesTall values. If <c>false</c>, it will be scaled according to the ScaleFactor value. 
        /// </value>
        public bool FitToPages
        {
            get
            {
                return this.fitToPages;
            }
            set
            {
                if (this.fitToPages != value)
                {
                    this.fitToPages = value;
                    this.OnPageSetupChanged();
                }
            }
        }

        /// <summary>
        /// Gets the print options.
        /// </summary>
        /// <value>The print options.</value>
        public override WorksheetPrintOptions PrintOptions
        {
            get
            {
                return this.printOptions;
            }
        }

        #endregion


        #region Constructor

        internal WorksheetPageSetup(Worksheet worksheet)
            : base(worksheet)
        {
            this.ScaleFactor = SpreadsheetDefaultValues.DefaultScaleFactorSize;
            this.printOptions = new WorksheetPrintOptions(this);
            this.fitToPages = false;
            this.fitToPagesTall = 1;
            this.fitToPagesWide = 1;
            this.printArea = new PrintArea(worksheet);
            this.pageBreaks = new PageBreaks(worksheet);
            this.AttachToEvents();
            this.PageOrder = PageOrder.DownThenOver;
        }

        #endregion


        #region Methods

        internal override void CopyFrom(SheetPageSetupBase fromSheetPageSetupBase)
        {
            base.CopyFrom(fromSheetPageSetupBase);

            WorksheetPageSetup fromWorksheetPageSetup = (WorksheetPageSetup)fromSheetPageSetupBase;
            this.printOptions.CopyFrom(fromWorksheetPageSetup.PrintOptions);
            this.printArea.CopyFrom(fromWorksheetPageSetup.PrintArea);
            this.pageBreaks.CopyFrom(fromWorksheetPageSetup.PageBreaks);

            this.pageOrder = fromWorksheetPageSetup.PageOrder;
            this.centerHorizontally = fromWorksheetPageSetup.CenterHorizontally;
            this.centerVertically = fromWorksheetPageSetup.CenterVertically;
            this.scaleFactor = fromWorksheetPageSetup.ScaleFactor;
            this.RotatedPageSize = fromWorksheetPageSetup.RotatedPageSize;
          
            this.fitToPages = fromWorksheetPageSetup.FitToPages;
            this.fitToPagesTall = fromWorksheetPageSetup.FitToPagesTall;
            this.fitToPagesWide = fromWorksheetPageSetup.FitToPagesWide;
        }

        internal void Clear()
        {
            this.PrintArea.Clear();
            this.PageBreaks.Clear();
            this.HeaderFooterSettings.Clear();
        }

        private void AttachToEvents()
        {
            this.PrintArea.PrintAreaChanged += this.PrintAreaChanged;
            this.PageBreaks.PageBreaksChanged += this.PageBreaksChanged;
        }

        private void PrintAreaChanged(object sender, EventArgs args)
        {
            using (new UpdateScope(this.SuspendSheetLayoutUpdate, this.ResumeSheetLayoutUpdate))
            {
                if (!this.PrintArea.HasPrintAreaRanges)
                {
                    this.PageBreaks.MakeAllPageBreaksInfinite();
                }

                this.InvalidateSheetLayout();
            }
        }

        private void PageBreaksChanged(object sender, EventArgs args)
        {
            this.InvalidateSheetLayout();
        }

        #endregion
    }
}