﻿using System;
using Telerik.Windows.Documents.Common.Model.Protection;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Protection
{
    /// <summary>
    /// Contains all data requried to set and remove protection.
    /// </summary>
    internal class ProtectionData
    {
        internal static string DefaultProtectionAlgorithm = SHA512.Name;

        private const int DefaultSpinCount = 100000;
        private bool enforced;
        private string algorithmName;
        private string hash;
        private string salt;
        private int spinCount;
        private string password;

        public ProtectionData()
        {
            this.AlgorithmName = ProtectionData.DefaultProtectionAlgorithm;
            this.SpinCount = ProtectionData.DefaultSpinCount;
            this.Salt = SpreadsheetProtectionHelper.Instance.GenerateSaltBase64();
        }

        /// <summary>
        /// Gets or sets the enforced.
        /// </summary>
        /// <value>The enforced.</value>
        public bool Enforced
        {
            get
            {
                return this.enforced;
            }
            set
            {
                this.enforced = value;
            }
        }

        /// <summary>
        /// Gets or sets the name of the algorithm.
        /// </summary>
        /// <value>The name of the algorithm.</value>
        public string AlgorithmName
        {
            get
            {
                return this.algorithmName;
            }
            set
            {
                this.algorithmName = value;
            }
        }

        /// <summary>
        /// Gets or sets the hash.
        /// </summary>
        /// <value>The hash.</value>
        public string Hash
        {
            get
            {
                return this.hash;
            }
            set
            {
                this.hash = value;
            }
        }

        /// <summary>
        /// Gets or sets the salt.
        /// </summary>
        /// <value>The salt.</value>
        public string Salt
        {
            get
            {
                return this.salt;
            }
            set
            {
                this.salt = value;
            }
        }

        /// <summary>
        /// Gets or sets the spin count.
        /// </summary>
        /// <value>The spin count.</value>
        public int SpinCount
        {
            get
            {
                return this.spinCount;
            }
            set
            {
                this.spinCount = value;
            }
        }

        /// <summary>
        /// Gets or sets the password.
        /// </summary>
        /// <value>The password.</value>
        public string Password
        {
            get
            {
                return this.password;
            }
            set
            {
                this.password = value;
            }
        }

        /// <summary>
        /// Indicates whether the data is produced using legacy protection scheme.
        /// </summary>
        /// <value>The is using legacy protection scheme.</value>
        public bool IsUsingLegacyProtectionScheme
        {
            get
            {
                return !string.IsNullOrEmpty(this.password);
            }
        }

        /// <summary>
        /// Indicates whether a user password is required to remove protection.
        /// </summary>
        /// <value>The requires user password.</value>
        public bool RequiresUserPassword
        {
            get
            {
                return !string.IsNullOrEmpty(this.Hash) || !string.IsNullOrEmpty(this.Password);
            }
        }

        /// <summary>
        /// Tries the remove protection.
        /// </summary>
        /// <param name="input">The input.</param>
        /// <returns></returns>
        public bool TryRemoveProtection(string input)
        {
            bool isSuccessful = false;

            if (this.IsUsingLegacyProtectionScheme)
            {
                if (string.IsNullOrEmpty(input))
                {
                    isSuccessful = string.IsNullOrEmpty(this.Password);
                }
                else
                {
                    string legacyPasswordHash = SpreadsheetProtectionHelper.Instance.GeneratePasswordLegacyScheme(input);
                    isSuccessful = string.Equals(this.Password, legacyPasswordHash, StringComparison.Ordinal);
                }
            }
            else
            {
                isSuccessful = SpreadsheetProtectionHelper.Instance.IsPasswordCorrect(input, this.Hash, this.Salt, this.AlgorithmName, this.SpinCount);
            }

            if (isSuccessful)
            {
                this.RemoveProtection();
            }

            return isSuccessful;
        }

        /// <summary>
        /// Enforces protection.
        /// </summary>
        /// <param name="input">The input.</param>
        public void EnforceProtection(string input)
        {
            this.Enforced = true;

            if (!string.IsNullOrEmpty(input))
            {
                if (string.IsNullOrEmpty(this.AlgorithmName))
                {
                    this.AlgorithmName = DefaultProtectionAlgorithm;
                }

                if (this.SpinCount < 1)
                {
                    this.SpinCount = DefaultSpinCount;
                }
                this.Salt = SpreadsheetProtectionHelper.Instance.GenerateSaltBase64();
                this.hash = SpreadsheetProtectionHelper.Instance.GenerateHashBase64(input, this.Salt, this.AlgorithmName, this.SpinCount);
            }
        }

        private void RemoveProtection()
        {
            this.enforced = false;
            this.hash = null;
            this.salt = null;
            this.password = null;
        }

        internal void CopyFrom(ProtectionData fromProtectionData)
        {
            this.AlgorithmName = fromProtectionData.AlgorithmName;
            this.Enforced = fromProtectionData.Enforced;
            this.Hash = fromProtectionData.Hash;
            this.Salt = fromProtectionData.Salt;
            this.SpinCount = fromProtectionData.SpinCount;
            this.Password = fromProtectionData.Password;
        }
    }
}
