﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Protection
{
    /// <summary>
    /// Represents the protection options of a worksheet.
    /// </summary>
    public class WorksheetProtectionOptions
    {
        /// <summary>
        /// Default protection options.
        /// </summary>
        public static readonly WorksheetProtectionOptions Default = new WorksheetProtectionOptions();

        private readonly bool allowFormatCells;
        private readonly bool allowFormatColumns;
        private readonly bool allowFormatRows;
        private readonly bool allowDeleteRows;
        private readonly bool allowInsertRows;
        private readonly bool allowDeleteColumns;
        private readonly bool allowInsertColumns;
        private readonly bool allowFiltering;
        private readonly bool allowSorting;

        /// <summary>
        /// Gets a value indicating whether the user can format cells when the worksheet is protected.
        /// </summary>
        /// <value>The allow format cells.</value>
        public bool AllowFormatCells
        {
            get
            {
                return this.allowFormatCells;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the user can format columns when the worksheet is protected.
        /// </summary>
        /// <value>The allow format columns.</value>
        public bool AllowFormatColumns
        {
            get
            {
                return this.allowFormatColumns;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the user can format rows when the worksheet is protected.
        /// </summary>
        /// <value>The allow format rows.</value>
        public bool AllowFormatRows
        {
            get
            {
                return this.allowFormatRows;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the user can delete rows when the worksheet is protected.
        /// </summary>
        /// <value>The allow delete rows.</value>
        public bool AllowDeleteRows
        {
            get
            {
                return this.allowDeleteRows;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the user can insert rows when the worksheet is protected.
        /// </summary>
        /// <value>The allow insert rows.</value>
        public bool AllowInsertRows
        {
            get
            {
                return this.allowInsertRows;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the user can delete columns when the worksheet is protected.
        /// </summary>
        /// <value>The allow delete columns.</value>
        public bool AllowDeleteColumns
        {
            get
            {
                return this.allowDeleteColumns;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the user can insert columns when the worksheet is protected.
        /// </summary>
        /// <value>The allow insert columns.</value>
        public bool AllowInsertColumns
        {
            get
            {
                return this.allowInsertColumns;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the user can use autofilter when the worksheet is protected.
        /// </summary>
        /// <value>The allow filtering.</value>
        public bool AllowFiltering
        {
            get
            {
                return this.allowFiltering;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the user can sort when the worksheet is protected.
        /// </summary>
        /// <value>The allow sorting.</value>
        public bool AllowSorting
        {
            get
            {
                return this.allowSorting;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="WorksheetProtectionOptions" /> class.
        /// </summary>
        /// <param name="allowDeleteRows">The allow delete rows.</param>
        /// <param name="allowInsertRows">The allow insert rows.</param>
        /// <param name="allowDeleteColumns">The allow delete columns.</param>
        /// <param name="allowInsertColumns">The allow insert columns.</param>
        /// <param name="allowFormatCells">The allow format cells.</param>
        /// <param name="allowFormatColumns">The allow format columns.</param>
        /// <param name="allowFormatRows">The allow format rows.</param>
        /// <param name="allowFiltering">The allow use autofilter.</param>
        /// <param name="allowSorting">The allow sorting.</param>
        public WorksheetProtectionOptions(bool allowDeleteRows = false, bool allowInsertRows = false, bool allowDeleteColumns = false, bool allowInsertColumns = false,
            bool allowFormatCells = false, bool allowFormatColumns = false, bool allowFormatRows = false, bool allowFiltering = false, bool allowSorting = false)
        {
            this.allowFormatCells = allowFormatCells;
            this.allowFormatColumns = allowFormatColumns;
            this.allowFormatRows = allowFormatRows;
            this.allowDeleteRows = allowDeleteRows;
            this.allowInsertRows = allowInsertRows;
            this.allowDeleteColumns = allowDeleteColumns;
            this.allowInsertColumns = allowInsertColumns;
            this.allowFiltering = allowFiltering;
            this.allowSorting = allowSorting;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            WorksheetProtectionOptions other = obj as WorksheetProtectionOptions;
            if (other == null)
            {
                return false;
            }

            return this.AllowDeleteColumns == other.AllowDeleteColumns &&
                this.AllowDeleteRows == other.AllowDeleteRows &&
                this.AllowInsertColumns == other.AllowInsertColumns &&
                this.AllowInsertRows == other.AllowInsertRows &&
                this.AllowFormatCells == other.AllowFormatCells &&
                this.AllowFormatColumns == other.AllowFormatColumns &&
                this.AllowFormatRows == other.AllowFormatRows &&
                this.AllowFiltering == other.AllowFiltering &&
                this.AllowSorting == other.AllowSorting;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(this.AllowDeleteColumns.GetHashCode(),
                this.AllowDeleteRows.GetHashCode(),
                this.AllowInsertColumns.GetHashCode(),
                this.AllowInsertRows.GetHashCode(),
                this.AllowFormatCells.GetHashCode(),
                this.AllowFormatColumns.GetHashCode(),
                this.AllowFormatRows.GetHashCode(),
                this.AllowFiltering.GetHashCode(),
                this.AllowSorting.GetHashCode());
        }

        /// <summary>
        /// Clones this instance.
        /// </summary>
        /// <returns>The cloned instance.</returns>
        public WorksheetProtectionOptions Clone()
        {
            return new WorksheetProtectionOptions(this.AllowDeleteRows, this.AllowInsertRows, this.AllowDeleteColumns, this.AllowInsertColumns,
                this.AllowFormatCells, this.AllowFormatColumns, this.AllowFormatRows, this.AllowFiltering, this.AllowSorting);
        }
    }
}
