﻿using Telerik.Windows.Documents.Spreadsheet.Utilities;
namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a reference to a row or column.
    /// </summary>
    public class RowColumnReference
    {
        #region Fields

        private readonly bool onOverflowStartOver;
        private readonly int rowsColumnsCount;
        private bool isAbsolute;
        private int index;
        private int offset;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the actual index.
        /// </summary>
        /// <value>The actual index.</value>
        public int ActualIndex
        {
            get
            {
                if (this.IsAbsolute)
                {
                    return this.index;
                }
                else
                {
                    int newIndex = this.index + this.offset;
                    if (this.onOverflowStartOver)
                    {
                        if (newIndex < 0)
                        {
                            return this.rowsColumnsCount + newIndex;
                        }
                        if (newIndex >= this.rowsColumnsCount)
                        {
                            return newIndex - this.rowsColumnsCount;
                        }
                    }

                    return newIndex;
                }
            }
        }

        /// <summary>
        /// Gets a value indicating whether the reference is absolute.
        /// </summary>
        /// <value>The value indicating whether the reference is absolut.</value>
        public bool IsAbsolute
        {
            get
            {
                return this.isAbsolute;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="RowColumnReference" /> class.
        /// </summary>
        public RowColumnReference()
            : this(false, 0)
        {
        }

        internal RowColumnReference(bool onOverflowStartOver, int rowsColumnsCount)
        {
            this.onOverflowStartOver = onOverflowStartOver;
            this.rowsColumnsCount = rowsColumnsCount;
        }

        #endregion


        #region Methods

        internal static RowColumnReference CreateRelativeRowColumnReference(int startIndex, int offset)
        {
            return CreateRelativeRowColumnReference(startIndex, offset, false, 0);
        }

        internal static RowColumnReference CreateRelativeRowColumnReference(int startIndex, int offset, bool onOverflowStartOver, int rowsColumnsCount)
        {
            RowColumnReference res = new RowColumnReference(onOverflowStartOver, rowsColumnsCount);
            res.isAbsolute = false;
            res.index = startIndex;
            res.offset = offset;

            return res;
        }

        internal static RowColumnReference CreateAbsoluteRowColumnReference(int targetIndex)
        {
            RowColumnReference res = new RowColumnReference(false, 0);
            res.isAbsolute = true;
            res.index = targetIndex;

            return res;
        }

        internal static RowColumnReference CreateRowColumnReference(bool isAbsolute, int index, int cellIndex, bool onOverflowStartOver, int rowsColumnsCount)
        {
            if (isAbsolute)
            {
                return RowColumnReference.CreateAbsoluteRowColumnReference(index);
            }

            return RowColumnReference.CreateRelativeRowColumnReference(cellIndex, index - cellIndex, onOverflowStartOver, rowsColumnsCount);
        }

        internal RowColumnReference CloneAndTranslate(int referenceIndex, bool onOverflowStartOver, int rowsColumnsCount)
        {
            RowColumnReference translatedIndex = new RowColumnReference(onOverflowStartOver, rowsColumnsCount);

            translatedIndex.isAbsolute = this.IsAbsolute;
            translatedIndex.index = IsAbsolute ? this.index : referenceIndex;
            translatedIndex.offset = this.offset;

            return translatedIndex;
        }

        internal RowColumnReference CloneAndTranslateUponInsertRemoveCells(int indexShift, int offsetShift, bool onOverflowStartOver, int rowsColumnsCount)
        {
            RowColumnReference translatedReference = new RowColumnReference(onOverflowStartOver, rowsColumnsCount);

            translatedReference.isAbsolute = this.isAbsolute;
            translatedReference.index = this.index + indexShift;
            translatedReference.offset = this.offset + offsetShift;

            return translatedReference;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            RowColumnReference rowColumnReference = obj as RowColumnReference;
            if (rowColumnReference == null)
            {
                return false;
            }

            return this.IsAbsolute == rowColumnReference.IsAbsolute &&
                   this.index == rowColumnReference.index &&
                   this.offset == rowColumnReference.offset;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(this.IsAbsolute.GetHashCode(), this.index.GetHashCode(), this.offset.GetHashCode());
        }

        #endregion
    }
}
