﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.History;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Provides base for the row and column selection classes.
    /// </summary>
    public abstract class RowColumnSelectionBase : SelectionBase
    {
        #region Properties

        internal abstract RowColumnPropertyBagBase PropertyBag { get; }

        #endregion


        #region Constructors

        internal RowColumnSelectionBase(Worksheet worksheet, IEnumerable<CellRange> cellRanges)
            : base(worksheet, cellRanges)
        {
        }

        #endregion


        #region Methods

        private RangePropertyValue<T> GetRangePropertyValue<T>(IPropertyDefinition<T> property, int fromIndex, int toIndex)
        {
            T value = default(T);
            bool isIndeterminate = false;
            var rangeValues = this.PropertyBag.GetPropertyValue(property, fromIndex, toIndex);
            int rangeIndex = 0;
            foreach (var rangeValue in rangeValues)
            {
                if (rangeIndex < 1)
                {
                    value = rangeValue.Value;
                }
                else
                {
                    value = this.PropertyBag.GetDefaultPropertyValue(property);
                    isIndeterminate = true;
                    break;
                }

                rangeIndex++;
            }

            return new RangePropertyValue<T>(isIndeterminate, value);
        }

        /// <summary>
        /// Gets the property value.
        /// </summary>
        /// <typeparam name="T">The type of the T.</typeparam>
        /// <param name="property">The property.</param>
        /// <returns></returns>
        protected RangePropertyValue<T> GetPropertyValue<T>(IPropertyDefinition<T> property)
        {
            return base.GetPropertyValue(
                (CellRange cellRange) =>
                {
                    int fromIndex;
                    int toIndex;
                    GetFromToIndexFromRange(cellRange, out fromIndex, out toIndex);
                    return GetRangePropertyValue(property, fromIndex, toIndex);
                },
                this.CellRanges,
                this.PropertyBag.GetDefaultPropertyValue(property));
        }

        /// <summary>
        /// Gets from to index from range.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <param name="fromIndex">From index.</param>
        /// <param name="toIndex">To index.</param>
        protected abstract void GetFromToIndexFromRange(CellRange cellRange, out int fromIndex, out int toIndex);

        internal RowColumnSelectionBase GetRowColumnSelection(int index)
        {
            return this.GetRowColumnSelection(index, index);
        }

        internal abstract RowColumnSelectionBase GetRowColumnSelection(int fromIndex, int toIndex);

        internal override void ClearCellPropertiesIfNeeded<T>(CellRange cellRange, IProperty<T> property)
        {
            IProperty<T> cellProperty;
            if (this.Worksheet.Cells.TryGetProperyFromPropertyDefinition<T>(property.PropertyDefinition, out cellProperty))
            {
                this.Worksheet.Cells[cellRange].ClearPropertyValue(cellProperty, false);
            }
        }

        internal override void PreserveOldStylePropertiesAsLocalIfNeeded<T>(CellRange cellRange, IProperty<T> property, T value)
        {
            if (property.PropertyDefinition == CellPropertyDefinitions.StyleNameProperty)
            {
                Workbook workbook = this.Worksheet.Workbook;
                using (new UpdateScope(workbook.SuspendLayoutUpdate, workbook.ResumeLayoutUpdate))
                {
                    CellStyle newStyle = this.Worksheet.Workbook.Styles.GetByName(value as string);

                    RangePropertyValue<string> oldStyleName = this.GetStyleName();

                    if (oldStyleName.IsIndeterminate)
                    {
                        ICompressedList<string> propertyValueCollection = this.PropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.StyleNameProperty);

                        int fromSelectedIndex = this.PropertyBag.GetRowColumnIndex(cellRange.FromIndex);
                        int toSelectedIndex = this.PropertyBag.GetRowColumnIndex(cellRange.ToIndex);

                        foreach (Range<long, string> range in propertyValueCollection)
                        {
                            if (fromSelectedIndex >= range.Start && toSelectedIndex <= range.End)
                            {
                                int intersectedFromIndex = (int)Math.Max(fromSelectedIndex, range.Start);
                                int intersectedToIndex = (int)Math.Max(toSelectedIndex, range.End);

                                RowColumnSelectionBase selection = this.GetRowColumnSelection(intersectedFromIndex, intersectedToIndex);

                                this.PreserveAllOldStylePropertiesAsLocalIfNeeded(selection, newStyle);
                            }
                        }
                    }
                    else
                    {
                        this.PreserveAllOldStylePropertiesAsLocalIfNeeded(this, newStyle);
                    }
                }
            }
        }

        internal override void Clear(ClearType type, bool clearIsLocked)
        {
            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                if (type == ClearType.All)
                {
                    foreach (IProperty property in this.WorksheetEntity.Properties)
                    {
                        if (property.PropertyDefinition != CellPropertyDefinitions.IsLockedProperty || clearIsLocked)
                        {
                            this.ClearPropertyValue(property);
                        }
                    }
                }
                else if (type == ClearType.Formats)
                {
                    foreach (IProperty property in this.WorksheetEntity.Properties)
                    {
                        this.ClearPropertyValue(property);
                    }
                }
            }
        }

        #endregion


        #region Outline Level property methods

        /// <summary>
        /// Groups the current selection. The selection cannot include multiple ranges.
        /// </summary>
        /// <returns>A value indicating whether at least part of the selection was grouped successfully.</returns>
        public bool Group()
        {
            return this.ShiftOutlineLevel((rowColumnSelection) =>
            {
                int outLineLevel = rowColumnSelection.GetOutlineLevel().Value;

                if (outLineLevel < RowColumnPropertyBagBase.MaxOutlineLevel)
                {
                    rowColumnSelection.SetOutlineLevel(outLineLevel + 1);
                    return true;
                }

                return false;
            });
        }

        /// <summary>
        /// Ungroups the current selection. The selection cannot include multiple ranges.
        /// </summary>
        /// <returns>
        /// A value indicating whether at least part of the selection was ungrouped successfully.
        /// </returns>
        public bool Ungroup()
        {
            return this.ShiftOutlineLevel((rowColumnSelection) =>
                {
                    int outLineLevel = rowColumnSelection.GetOutlineLevel().Value;

                    if (outLineLevel > 0)
                    {
                        int newOutLineLevel = outLineLevel - 1;
                        if (newOutLineLevel != 0)
                        {
                            rowColumnSelection.SetOutlineLevel(outLineLevel - 1);
                        }
                        else
                        {
                            rowColumnSelection.ClearOutlineLevel();
                        }

                        return true;
                    }

                    return false;
                });
        }

        private bool ShiftOutlineLevel(Func<RowColumnSelectionBase, bool> shiftAction)
        {
            if (this.CellRanges.Count() > 1)
            {
                return false;
            }

            bool newOutlineIsSet = false;

            CellRange range = this.CellRanges.First();

            WorkbookHistory history = this.Worksheet.Workbook.History;
            using (new UpdateScope(history.BeginUndoGroup, history.EndUndoGroup))
            {
                int startIndex = this.PropertyBag.GetRowColumnIndex(range.FromIndex);
                int endIndex = this.PropertyBag.GetRowColumnIndex(range.ToIndex);

                for (int index = startIndex; index <= endIndex; index++)
                {
                    RowColumnSelectionBase rowColumnSelection = this.GetRowColumnSelection(index);
                    newOutlineIsSet = shiftAction(rowColumnSelection);
                }
            }

            return newOutlineIsSet;
        }

        /// <summary>
        /// Sets the outline level.
        /// </summary>
        /// <param name="value">The outline level.</param>
        public void SetOutlineLevel(int value)
        {
            Guard.ThrowExceptionIfOutOfRange(0, RowColumnPropertyBagBase.MaxOutlineLevel, value, "value");

            this.ExecuteForEachRangeInsideBeginEndUpdate((cellRange) =>
            {
                this.SetOutlineLevelOnRange(cellRange, value);
            });
        }

        /// <summary>
        /// Gets the outline level.
        /// </summary>
        /// <returns>The outline level.</returns>
        public abstract RangePropertyValue<int> GetOutlineLevel();

        /// <summary>
        /// Clears the outline level.
        /// </summary>
        public void ClearOutlineLevel()
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate((cellRange) =>
            {
                this.ClearOutlineLevelOnRange(cellRange);
            });
        }

        internal abstract void SetOutlineLevelOnRange(CellRange cellRange, int value);

        internal abstract void ClearOutlineLevelOnRange(CellRange cellRange);

        #endregion
    }
}