﻿using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a row selection.
    /// </summary>
    public class RowSelection : RowColumnSelectionBase
    {
        #region Properties

        internal override RowColumnPropertyBagBase PropertyBag
        {
            get
            {
                return this.Worksheet.Rows.PropertyBag;
            }
        }

        internal override WorksheetEntityBase WorksheetEntity
        {
            get
            {
                return this.Worksheet.Rows;
            }
        }

        #endregion


        #region Constructors

        internal RowSelection(Worksheet worksheet, IEnumerable<CellRange> cellRanges)
            : base(worksheet, cellRanges)
        {
        }

        #endregion


        #region Methods

        /// <summary>
        /// Sets the borders.
        /// </summary>
        /// <param name="value">The value.</param>
        public override void SetBorders(CellBorders value)
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate(
                (cellRange) =>
                {
                    var context = new SetRowBordersCommandContext(this.Worksheet, cellRange, value);
                    this.Worksheet.ExecuteCommand(WorkbookCommands.SetRowBorders, context);
                });
        }

        /// <summary>
        /// Gets from to index from range.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <param name="fromIndex">From index.</param>
        /// <param name="toIndex">To index.</param>
        protected override void GetFromToIndexFromRange(CellRange cellRange, out int fromIndex, out int toIndex)
        {
            fromIndex = cellRange.FromIndex.RowIndex;
            toIndex = cellRange.ToIndex.RowIndex;
        }

        internal override void PreserveAsLocalPropertyIfStyleIsApplied<T>(CellRange cellRange, IProperty<T> property, T value)
        {
            IPropertyDefinition<T> propertyDefinition = property.PropertyDefinition as IPropertyDefinition<T>;
            IProperty<T> cellProperty;
            if (this.Worksheet.Cells.TryGetProperyFromPropertyDefinition(propertyDefinition, out cellProperty))
            {
                CellsPropertyBag cellsPropertyBag = this.Worksheet.Cells.PropertyBag;

                ICompressedList<string> cellsStylePropertyCollection = cellsPropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.StyleNameProperty);

                foreach (LongRange longRange in CellsPropertyBag.ConvertCellRangeToLongRanges(cellRange))
                {
                    ICompressedList<string> cellsStyleValues = cellsStylePropertyCollection.GetValue(longRange.Start, longRange.End);
                    Range<long, string>[] nonDefaultCellStyleValueRanges = cellsStyleValues.GetNonDefaultRanges().ToArray();

                    foreach (Range<long, string> range in nonDefaultCellStyleValueRanges)
                    {
                        CellStyle cellStyle = this.Worksheet.Workbook.Styles.GetByName(range.Value);
                        if (cellStyle.GetIsPropertyIncluded(propertyDefinition))
                        {
                            CellRange newCellRange = CellsPropertyBag.ConvertLongCellRangeToCellRange(range.Start, range.End);
                            cellProperty.SetValue(newCellRange, value);
                        }
                    }
                }
            }
        }

        #endregion


        #region Height Property

        /// <summary>
        /// Gets the height.
        /// </summary>
        /// <returns>The height for the selection.</returns>
        public RangePropertyValue<RowHeight> GetHeight()
        {
            return this.GetPropertyValue(RowsPropertyBag.HeightProperty);
        }

        /// <summary>
        /// Sets the height.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetHeight(RowHeight value)
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate((cellRange) =>
            {
                var context = new SetRowColumnPropertyCommandContext<RowHeight>(
                    this.Worksheet, RowsPropertyBag.HeightProperty,
                    cellRange.FromIndex.RowIndex, cellRange.ToIndex.RowIndex, value);

                this.Worksheet.ExecuteCommand(WorkbookCommands.SetRowHeight, context);
            });
        }

        internal void SetHeight(RowHeight value, bool considerHidden)
        {
            if (!considerHidden)
            {
                this.SetHeight(value);
            }

            if (value.Value == 0)
            {
                this.SetHidden(true);
            }
            else
            {
                using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
                {
                    this.ClearHidden();
                    this.SetHeight(value);
                }
            }
        }

        /// <summary>
        /// Clears the height.
        /// </summary>
        public void ClearHeight()
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate((cellRange) =>
            {
                var context = new ClearRowColumnPropertyCommandContext<RowHeight>(
                    this.Worksheet, RowsPropertyBag.HeightProperty, cellRange.FromIndex.RowIndex, cellRange.ToIndex.RowIndex);

                this.Worksheet.ExecuteCommand(WorkbookCommands.SetRowHeight, context);
            });
        }

        /// <summary>
        /// Automatically adjusts the height.
        /// </summary>
        public void AutoFitHeight()
        {
            this.SetHeight(RowHeight.AutoFit);
        }

        #endregion


        #region Outline Level Property

        internal override void SetOutlineLevelOnRange(CellRange cellRange, int value)
        {
            var context = new SetRowColumnPropertyCommandContext<int>(
                this.Worksheet, RowsPropertyBag.OutlineLevelProperty,
                cellRange.FromIndex.RowIndex, cellRange.ToIndex.RowIndex, value);

            this.Worksheet.ExecuteCommand(WorkbookCommands.SetRowOutlineLevel, context);
        }

        /// <summary>
        /// Gets the outline level.
        /// </summary>
        /// <returns>The outline level.</returns>
        public override RangePropertyValue<int> GetOutlineLevel()
        {
            return this.GetPropertyValue(RowsPropertyBag.OutlineLevelProperty);
        }

        internal override void ClearOutlineLevelOnRange(CellRange cellRange)
        {
            var context = new ClearRowColumnPropertyCommandContext<int>(
                this.Worksheet, RowsPropertyBag.OutlineLevelProperty, cellRange.FromIndex.RowIndex, cellRange.ToIndex.RowIndex);

            this.Worksheet.ExecuteCommand(WorkbookCommands.SetRowOutlineLevel, context);
        }

        internal override RowColumnSelectionBase GetRowColumnSelection(int fromIndex, int toIndex)
        {
            return this.Worksheet.Rows[fromIndex, toIndex];
        }

        #endregion


        #region Hidden Property

        /// <summary>
        /// Sets the hidden property.
        /// </summary>
        /// <param name="value">The value indicating whether the selected rows are hidden.</param>
        public void SetHidden(bool value)
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate((cellRange) =>
            {
                var context = new SetRowColumnPropertyCommandContext<bool>(
                    this.Worksheet, RowsPropertyBag.HiddenProperty,
                    cellRange.FromIndex.RowIndex, cellRange.ToIndex.RowIndex, value);

                this.Worksheet.ExecuteCommand(WorkbookCommands.SetRowHidden, context);
            });
        }

        /// <summary>
        /// Gets the value of the hidden property.
        /// </summary>
        /// <returns>The value of the hidden property</returns>
        public RangePropertyValue<bool> GetHidden()
        {
            return this.GetPropertyValue(RowsPropertyBag.HiddenProperty);
        }

        /// <summary>
        /// Clears the hidden property.
        /// </summary>
        public void ClearHidden()
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate((cellRange) =>
            {
                var context = new ClearRowColumnPropertyCommandContext<bool>(
                    this.Worksheet, RowsPropertyBag.HiddenProperty, cellRange.FromIndex.RowIndex, cellRange.ToIndex.RowIndex);

                this.Worksheet.ExecuteCommand(WorkbookCommands.SetRowHidden, context);
            });
        }

        #endregion


        #region Insert/Remove

        /// <summary>
        /// Inserts this instance in the workbook.
        /// </summary>
        /// <returns>The value indicating whether the insert was successful.</returns>
        public bool Insert()
        {
            var sortedSelectedRanges =
                from r in this.CellRanges
                orderby r.FromIndex.RowIndex descending
                select r;

            return this.ExecuteForEachRangeInsideBeginEndUpdate(
                sortedSelectedRanges,
                (cellRange) =>
                {
                    this.Worksheet.Rows.Insert(cellRange.FromIndex.RowIndex, cellRange.RowCount);
                },
                (cellRange) => { return this.Worksheet.Rows.CanInsert(cellRange.FromIndex.RowIndex, cellRange.RowCount); });
        }

        /// <summary>
        /// Removes this instance from the workbook.
        /// </summary>
        public void Remove()
        {
            ICompressedList<bool> rowsToRemove = new CompressedList<bool>(0, SpreadsheetDefaultValues.RowCount - 1);

            foreach (var selectedRange in this.CellRanges)
            {
                rowsToRemove.SetValue(selectedRange.FromIndex.RowIndex, selectedRange.ToIndex.RowIndex, true);
            }

            List<CellRange> cellRanges = new List<CellRange>();
            foreach (var range in rowsToRemove)
            {
                if (range.Value)
                {
                    cellRanges.Add(new CellRange((int)range.Start, 0, (int)range.End, 0));
                }
            }
            cellRanges.Reverse();

            this.ExecuteForEachRangeInsideBeginEndUpdate(
                cellRanges,
                (cellRange) =>
                {
                    this.Worksheet.Rows.Remove(cellRange.FromIndex.RowIndex, cellRange.RowCount);
                });
        }

        #endregion
    }
}