﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a collection of rows.
    /// </summary>
    public class Rows : RowsColumnsBase
    {
		#region Fields

		private readonly RowsPropertyBag propertyBag;
        private readonly IProperty<RowHeight> heightProperty;

		#endregion


		#region Properties

        /// <summary>
        /// Gets the row / column count.
        /// </summary>
        /// <value>The row / column count.</value>
        public override int Count
        {
			get
			{
				return this.Worksheet.Cells.RowCount;
			}
		}

		internal RowsPropertyBag PropertyBag
		{
			get
			{
				return this.propertyBag;
			}
		}

        internal override PropertyBagBase PropertyBagBase
        {
            get
            {
                return this.PropertyBag;
            }
        }

        internal IProperty<RowHeight> HeightProperty
        {
            get
            {
                return this.heightProperty;
            }
        }

		#endregion


        #region Indexers

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.RowSelection" /> at the specified index.
        /// </summary>
        /// <value>The row selection.</value>
        public RowSelection this[int index]
        {
            get
            {
                return this.GetRowSelection(index);
            }
        }

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.RowSelection" /> with the specified start and end indices.
        /// </summary>
        /// <value>The row selection.</value>
        public RowSelection this[int fromIndex, int toIndex]
        {
            get
            {
                return this.GetRowSelection(fromIndex, toIndex);
            }
        }

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.RowSelection" /> with the specified cell index.
        /// </summary>
        /// <value>The row selection.</value>
        public RowSelection this[CellIndex cellIndex]
        {
            get
            {
                return this.GetRowSelection(cellIndex);
            }
        }

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.RowSelection" /> with the specified cell range.
        /// </summary>
        /// <value>The row selection.</value>
        public RowSelection this[CellRange cellRange]
        {
            get
            {
                return this.GetRowSelection(cellRange);
            }
        }

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.RowSelection" /> with the specified cell ranges.
        /// </summary>
        /// <value>The row selection.</value>
        public RowSelection this[IEnumerable<CellRange> cellRanges]
        {
            get
            {
                return this.GetRowSelection(cellRanges);
            }
        }

        #endregion


        #region Constructors

        internal Rows(Worksheet worksheet)
			: base(worksheet)
		{
			this.propertyBag = new RowsPropertyBag();
            this.PropertyBag.PropertyChanged += this.PropertyBag_PropertyChanged;

            this.heightProperty = this.CreateProperty(RowsPropertyBag.HeightProperty);
		}

		#endregion


        #region Event Handlers

        private void PropertyBag_PropertyChanged(object sender, RowColumnPropertyChangedEventArgs e)
        {
            if (e.Property == RowsPropertyBag.HeightProperty)
            {
                this.OnRowsHeightChanged(e);
            }
        }

        #endregion


		#region Methods

        /// <summary>
        /// Determines whether this instance can be inserted at the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <param name="itemCount">The item count.</param>
        /// <returns>
        /// The value indicating whether this instance can be inserted at the specified index
        /// </returns>
        public override bool CanInsert(int index, int itemCount)
        {
            return this.PropertyBag.CanInsert() &&
				this.Cells.PropertyBag.CanInsertRow(itemCount);
		}

        /// <summary>
        /// Inserts the override.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <param name="itemCount">The item count.</param>
        /// <returns></returns>
        protected override bool InsertOverride(int index, int itemCount)
        {
			var context = new InsertRowCommandContext(this.Worksheet, index, itemCount);
			bool isSuccessful = this.Worksheet.ExecuteCommand(WorkbookCommands.InsertRow, context);

            CellRange range = CellRange.FromRowRange(index, index + itemCount - 1);
            this.Cells.UpdateCellRangeInsertedOrRemovedDependentCollections(range, RangeType.Rows, false);

            return isSuccessful;
		}

        /// <summary>
        /// Removes the override.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <param name="itemCount">The item count.</param>
        protected override void RemoveOverride(int index, int itemCount)
        {
			var context = new RemoveRowCommandContext(this.Worksheet, index, itemCount);
			this.Worksheet.ExecuteCommand(WorkbookCommands.RemoveRow, context);

            CellRange range = CellRange.FromRowRange(index, index + itemCount - 1);
            this.Cells.UpdateCellRangeInsertedOrRemovedDependentCollections(range, RangeType.Rows, true);
		}

        internal bool GetIsHidden(int rowIndex)
        {
            return this.PropertyBag.GetPropertyValue(RowsPropertyBag.HiddenProperty, rowIndex);
        }

        internal override IProperty<T> CreatePropertyOverride<T>(IPropertyDefinition<T> propertyDefinition)
        {
            return new RowProperty<T>(this.Worksheet, propertyDefinition);
        }

		#endregion


        #region GetSelection Methods

        /// <summary>
        /// Gets a row selection at the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <returns>The row selection</returns>
        public RowSelection GetRowSelection(int index)
        {
            return this.GetRowSelection(index, index);
        }

        /// <summary>
        /// Gets a row selection between two indices.
        /// </summary>
        /// <param name="fromIndex">From index.</param>
        /// <param name="toIndex">To index.</param>
        /// <returns>The row selection</returns>
        public RowSelection GetRowSelection(int fromIndex, int toIndex)
        {
            return this.GetRowSelection(CellRange.FromRowRange(fromIndex, toIndex));
        }

        /// <summary>
        /// Gets a row selection at the specified index.
        /// </summary>
        /// <param name="cellIndex">Index of the cell.</param>
        /// <returns>The row selection</returns>
        public RowSelection GetRowSelection(CellIndex cellIndex)
        {
            return this.GetRowSelection(new CellRange(cellIndex, cellIndex));
        }

        /// <summary>
        /// Gets a row selection of a specified cell range.
        /// </summary>
        /// <param name="cellRange">The cell range.</param>
        /// <returns>The row selection</returns>
        public RowSelection GetRowSelection(CellRange cellRange)
        {
            return this.GetRowSelection(new CellRange[] { cellRange });
        }

        /// <summary>
        /// Gets a row selection of a specified collection of cell ranges.
        /// </summary>
        /// <param name="cellRanges">The cell ranges.</param>
        /// <returns>The row selection</returns>
        public RowSelection GetRowSelection(IEnumerable<CellRange> cellRanges)
        {
            return new RowSelection(this.Worksheet, cellRanges);
        }

        #endregion


        #region Default Height

        /// <summary>
        /// Gets the default height.
        /// </summary>
        /// <returns>The row height</returns>
        public RowHeight GetDefaultHeight()
        {
            return this.PropertyBag.GetDefaultPropertyValue(RowsPropertyBag.HeightProperty);
        }

        /// <summary>
        /// Sets the default height.
        /// </summary>
        /// <param name="height">The height.</param>
        public void SetDefaultHeight(RowHeight height)
        {
            var context = new SetDefaultPropertyValueCommandContext<RowHeight>(
                this.Worksheet, RowsPropertyBag.HeightProperty, this.PropertyBag, height);

            this.Worksheet.ExecuteCommand(WorkbookCommands.SetDefaultRowHeight, context);
        }

        #endregion


        #region Events

        /// <summary>
        /// Occurs when the rows' heights have changed.
        /// </summary>
        public event EventHandler<RowColumnPropertyChangedEventArgs> RowsHeightsChanged;

        private void OnRowsHeightChanged(RowColumnPropertyChangedEventArgs args)
        {
            if (this.RowsHeightsChanged != null)
            {
                this.RowsHeightsChanged(this, args);
            }
        }

        #endregion
    }
}