﻿using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Provides bas for the Rows and Columns classes.
    /// </summary>
    public abstract class RowsColumnsBase : WorksheetEntityBase
    {
        #region Fields

        private readonly IProperty<bool> hiddenProperty;
        private readonly IProperty<int> outlineLevelProperty;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the cells.
        /// </summary>
        /// <value>The cells.</value>
        protected Cells Cells
        {
			get
			{
				return this.Worksheet.Cells;
			}
		}

        /// <summary>
        /// Gets the row / column count.
        /// </summary>
        /// <value>The row / column count.</value>
        public abstract int Count { get; }

        internal IProperty<bool> HiddenProperty
        {
            get
            {
                return this.hiddenProperty;
            }
        }

        internal IProperty<int> OutlineLevelProperty
        {
            get
            {
                return this.outlineLevelProperty;
            }
        }

		#endregion


		#region Constructors

		internal RowsColumnsBase(Worksheet worksheet)
            : base(worksheet)
		{
            this.hiddenProperty = this.CreateProperty(RowColumnPropertyBagBase.HiddenProperty);
            this.outlineLevelProperty = this.CreateProperty(RowColumnPropertyBagBase.OutlineLevelProperty);
		}

		#endregion


		#region Methods

        /// <summary>
        /// Inserts the item at the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <returns>The value indicating whether the insert command was successfully executed.</returns>
        public bool Insert(int index)
        {
			return this.Insert(index, 1);
		}


        /// <summary>
        /// Inserts the specified number of items at the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <param name="itemCount">The item count.</param>
        /// <returns>
        /// The value indicating whether the insert command was successfully executed.
        /// </returns>
        public bool Insert(int index, int itemCount)
		{
            bool result = false;
            using (new UpdateScope(this.Worksheet.BeginUndoGroup, this.Worksheet.EndUndoGroup))
            {
                result = this.InsertOverride(index, itemCount);
            }

            return result;
		}

        /// <summary>
        /// Removes the item from the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        public void Remove(int index)
        {
			this.Remove(index, 1);
		}

        /// <summary>
        /// Removes the specified number of items from the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <param name="itemCount">The item count.</param>
        public void Remove(int index, int itemCount)
        {
            using (new UpdateScope(this.Worksheet.BeginUndoGroup, this.Worksheet.EndUndoGroup))
            {
                this.RemoveOverride(index, itemCount);
            }
		}

        /// <summary>
        /// Determines whether this instance can be inserted at the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <param name="itemCount">The item count.</param>
        /// <returns>The value indicating whether this instance can be inserted at the specified index</returns>
        public abstract bool CanInsert(int index, int itemCount);

        /// <summary>
        /// Inserts the override.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <param name="itemCount">The item count.</param>
        /// <returns></returns>
        protected abstract bool InsertOverride(int index, int itemCount);

        /// <summary>
        /// Removes the override.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <param name="itemCount">The item count.</param>
        protected abstract void RemoveOverride(int index, int itemCount);
		
        #endregion
	}
}