﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Layout;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Theming;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents base for the selection classes.
    /// </summary>
    public abstract class SelectionBase
    {
        #region Fields

        private readonly Worksheet worksheet;
        private readonly IEnumerable<CellRange> cellRanges;
        private int beginUpdateCount;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the cell ranges.
        /// </summary>
        /// <value>The cell ranges.</value>
        public IEnumerable<CellRange> CellRanges
        {
            get
            {
                return this.cellRanges;
            }
        }

        /// <summary>
        /// Gets the worksheet.
        /// </summary>
        /// <value>The worksheet.</value>
        public Worksheet Worksheet
        {
            get
            {
                return this.worksheet;
            }
        }

        internal abstract WorksheetEntityBase WorksheetEntity
        {
            get;
        }

        #endregion


        #region Constructors

        internal SelectionBase(Worksheet worksheet, IEnumerable<CellRange> cellRanges)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfNull(cellRanges, "cellRanges");
            Guard.ThrowExceptionIfContainsNull(cellRanges, "cellRanges");

            this.worksheet = worksheet;
            this.cellRanges = cellRanges;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Begins the update.
        /// </summary>
        protected void BeginUpdate()
        {
            if (this.beginUpdateCount == 0)
            {
                this.Worksheet.Workbook.History.BeginUndoGroup();
                this.Worksheet.Workbook.SuspendLayoutUpdate();
            }

            this.beginUpdateCount++;
        }

        /// <summary>
        /// Ends the update.
        /// </summary>
        protected void EndUpdate()
        {
            if (this.beginUpdateCount == 0)
            {
                throw new InvalidOperationException("There is no active update to end.");
            }

            this.beginUpdateCount--;
            if (this.beginUpdateCount == 0)
            {
                this.Worksheet.Workbook.ResumeLayoutUpdate();
                this.Worksheet.Workbook.History.EndUndoGroup();
            }
        }

        internal void ExecuteAndExpandToFitNumberValuesWidth(Action action)
        {
            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                action();
                this.Worksheet.Columns[this.CellRanges].ExpandToFitNumberValuesWidth();
            }
        }

        /// <summary>
        /// Executes for each range inside begin end update.
        /// </summary>
        /// <param name="action">The action.</param>
        protected void ExecuteForEachRangeInsideBeginEndUpdate(Action<CellRange> action)
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate(this.CellRanges, action);
        }

        /// <summary>
        /// Executes for each range inside begin end update.
        /// </summary>
        /// <param name="cellRanges">The cell ranges.</param>
        /// <param name="action">The action.</param>
        /// <param name="canExecute">The can execute.</param>
        /// <returns></returns>
        protected bool ExecuteForEachRangeInsideBeginEndUpdate(
            IEnumerable<CellRange> cellRanges,
            Action<CellRange> action,
            Predicate<CellRange> canExecute = null)
        {
            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                bool result = this.ExecuteForEachRange(cellRanges, action, canExecute);

                return result;
            }
        }

        /// <summary>
        /// Executes for each range.
        /// </summary>
        /// <param name="action">The action.</param>
        protected void ExecuteForEachRange(Action<CellRange> action)
        {
            this.ExecuteForEachRange(this.CellRanges, action);
        }

        /// <summary>
        /// Executes for each range.
        /// </summary>
        /// <param name="cellRanges">The cell ranges.</param>
        /// <param name="action">The action.</param>
        /// <param name="canExecute">The can execute.</param>
        /// <returns></returns>
        protected bool ExecuteForEachRange(
            IEnumerable<CellRange> cellRanges,
            Action<CellRange> action,
            Predicate<CellRange> canExecute = null)
        {
            if (canExecute != null)
            {
                foreach (CellRange cellRange in cellRanges)
                {
                    if (!canExecute(cellRange))
                    {
                        return false;
                    }
                }
            }

            foreach (CellRange cellRange in cellRanges)
            {
                action(cellRange);
            }

            return true;
        }

        internal RangePropertyValue<T> GetPropertyValue<T>(IProperty<T> property)
        {
            return this.GetPropertyValue(property, this.CellRanges);
        }

        internal RangePropertyValue<T> GetPropertyValue<T>(IProperty<T> property, CellRange cellRange)
        {
            return this.GetPropertyValue(property, new CellRange[] { cellRange });
        }

        internal RangePropertyValue<T> GetPropertyValue<T>(IProperty<T> property, IEnumerable<CellRange> cellRanges)
        {
            return this.GetPropertyValue(property.GetValue, cellRanges, property.GetDefaultValue());
        }

        /// <summary>
        /// Gets the property value.
        /// </summary>
        /// <typeparam name="T">The type of the T.</typeparam>
        /// <param name="getRangePropertyValue">The get range property value.</param>
        /// <param name="cellRanges">The cell ranges.</param>
        /// <param name="defaultValue">The default value.</param>
        /// <returns></returns>
        protected RangePropertyValue<T> GetPropertyValue<T>(
            Func<CellRange, RangePropertyValue<T>> getRangePropertyValue,
            IEnumerable<CellRange> cellRanges,
            T defaultValue)
        {
            T resultValue = default(T);
            bool isIndeterminate = false;

            bool isFirstTime = true;
            foreach (CellRange cellRange in cellRanges)
            {
                RangePropertyValue<T> currenetRangeValue = getRangePropertyValue(cellRange);
                if (currenetRangeValue.IsIndeterminate)
                {
                    isIndeterminate = true;
                    break;
                }

                if (isFirstTime)
                {
                    resultValue = currenetRangeValue.Value;
                    isFirstTime = false;
                }
                else
                {
                    if (!TelerikHelper.EqualsOfT(currenetRangeValue.Value, resultValue))
                    {
                        isIndeterminate = true;
                        break;
                    }
                }
            }

            if (isIndeterminate)
            {
                resultValue = defaultValue;
            }

            return new RangePropertyValue<T>(isIndeterminate, resultValue);
        }

        internal void SetPropertyValue<T>(IProperty<T> property, T value)
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate(
                (cellRange) =>
                {
                    this.ClearCellPropertiesIfNeeded<T>(cellRange, property);
                    this.PreserveOldStylePropertiesAsLocalIfNeeded<T>(cellRange, property, value);
                    this.PreserveOldRowColumnPropertyAsLocalIfNeeded<T>(cellRange, property, value);
                    this.PreserveAsLocalPropertyIfStyleIsApplied(cellRange, property, value);

                    property.SetValue(cellRange, value);
                });
        }

        internal virtual void PreserveAsLocalPropertyIfStyleIsApplied<T>(CellRange cellRange, IProperty<T> property, T value)
        {
        }

        internal virtual void ClearCellPropertiesIfNeeded<T>(CellRange cellRange, IProperty<T> property)
        {
        }

        internal virtual void PreserveOldRowColumnPropertyAsLocalIfNeeded<T>(CellRange cellRange, IProperty<T> property, T value)
        {
        }

        internal abstract void PreserveOldStylePropertiesAsLocalIfNeeded<T>(CellRange cellRange, IProperty<T> property, T value);

        internal void PreserveAllOldStylePropertiesAsLocalIfNeeded(SelectionBase selection, CellStyle newStyle)
        {
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.BottomBorderProperty, CellPropertyDefinitions.BottomBorderProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.DiagonalDownBorderProperty, CellPropertyDefinitions.DiagonalDownBorderProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.DiagonalUpBorderProperty, CellPropertyDefinitions.DiagonalUpBorderProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.FillProperty, CellPropertyDefinitions.FillProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.FontFamilyProperty, CellPropertyDefinitions.FontFamilyProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.FontFamilyProperty, CellPropertyDefinitions.FontFamilyProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.FontSizeProperty, CellPropertyDefinitions.FontSizeProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.ForeColorProperty, CellPropertyDefinitions.ForeColorProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.FormatProperty, CellPropertyDefinitions.FormatProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.HorizontalAlignmentProperty, CellPropertyDefinitions.HorizontalAlignmentProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.IndentProperty, CellPropertyDefinitions.IndentProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.IsBoldProperty, CellPropertyDefinitions.IsBoldProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.IsItalicProperty, CellPropertyDefinitions.IsItalicProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.IsWrappedProperty, CellPropertyDefinitions.IsWrappedProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.LeftBorderProperty, CellPropertyDefinitions.LeftBorderProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.RightBorderProperty, CellPropertyDefinitions.RightBorderProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.TopBorderProperty, CellPropertyDefinitions.TopBorderProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.UnderlineProperty, CellPropertyDefinitions.UnderlineProperty);
            this.PreserveOldStylePropertiyAsLocalIfNeeded(newStyle, selection, selection.WorksheetEntity.VerticalAlignmentProperty, CellPropertyDefinitions.VerticalAlignmentProperty);
        }

        internal void PreserveOldStylePropertiyAsLocalIfNeeded<T>(CellStyle newStyle, SelectionBase selection,
            IProperty<T> property, IPropertyDefinition propertyDefinition)
        {
            if (!newStyle.GetIsPropertyIncluded(propertyDefinition))
            {
                RangePropertyValue<T> propertyValue = selection.GetPropertyValue(property);
                if (!propertyValue.IsIndeterminate &&
                    !propertyValue.Value.Equals(property.GetDefaultValue()))
                {
                    selection.SetPropertyValue(property, propertyValue.Value);
                }
            }
        }

        internal void ClearPropertyValue(IProperty property, bool shouldSetDefaultValuesIfIntersectsWithRowColumn = true)
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate(
                (cellRange) =>
                {
                    property.ClearValue(cellRange);
                    if (shouldSetDefaultValuesIfIntersectsWithRowColumn)
                    {
                        this.ClearBySettingDefaultValuesOnRowColumnPropertyIntersection(cellRange, property);
                    }
                });
        }

        internal virtual void ClearBySettingDefaultValuesOnRowColumnPropertyIntersection(CellRange cellRange, IProperty property)
        {
        }

        private void UpdateValueRespectFormat(CellValueFormat format)
        {
            this.ExecuteForEachRangeInsideBeginEndUpdate(
               (CellRange cellRange) =>
               {
                   UpdateCellPropertyCommandContext<ICellValue> context = new UpdateCellPropertyCommandContext<ICellValue>(
                       this.Worksheet, CellPropertyDefinitions.ValueProperty, cellRange,
                       (ICellValue value) =>
                       {
                           return FormatHelper.GetCellValueRespectFormat(format, value);
                       });

                   this.Worksheet.ExecuteCommand(WorkbookCommands.UpdateCellValue, context);
               });
        }

        /// <summary>
        /// Gets the index of the top left cell.
        /// </summary>
        /// <returns></returns>
        protected CellIndex GetTopLeftCellIndex()
        {
            int rowIndex = int.MaxValue;
            int columnIndex = int.MaxValue;

            foreach (CellRange range in this.CellRanges)
            {
                if (rowIndex > range.FromIndex.RowIndex)
                {
                    rowIndex = range.FromIndex.RowIndex;
                }

                if (columnIndex > range.FromIndex.ColumnIndex)
                {
                    columnIndex = range.FromIndex.ColumnIndex;
                }
            }

            return new CellIndex(rowIndex, columnIndex);
        }

        /// <summary>
        /// Clears the properties using the specified type of clear.
        /// </summary>
        /// <param name="type">The type of clear.</param>
        public void Clear(ClearType type)
        {
            this.Clear(type, true);
        }

        internal abstract void Clear(ClearType type, bool clearIsLocked);

        #endregion


        #region Format Property

        internal void SetFormatAndExpandToFitNumberValuesWidth(CellValueFormat value)
        {
            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
            {
                this.SetFormat(value);
            });
        }

        /// <summary>
        /// Gets the format.
        /// </summary>
        /// <returns>The format for the selection.</returns>
        public RangePropertyValue<CellValueFormat> GetFormat()
        {
            return this.GetPropertyValue(this.WorksheetEntity.FormatProperty);
        }

        /// <summary>
        /// Sets the format.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetFormat(CellValueFormat value)
        {
            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                this.UpdateValueRespectFormat(value);
                this.SetPropertyValue(this.WorksheetEntity.FormatProperty, value);
            }
        }

        /// <summary>
        /// Clears the format.
        /// </summary>
        public void ClearFormat()
        {
            this.ClearPropertyValue(this.WorksheetEntity.FormatProperty);
        }

        #endregion


        #region StyleName Property

        /// <summary>
        /// Gets the name of the style.
        /// </summary>
        /// <returns>The name of the style for the selection.</returns>
        public RangePropertyValue<string> GetStyleName()
        {
            return this.GetPropertyValue(this.WorksheetEntity.StyleNameProperty);
        }

        /// <summary>
        /// Sets the name of the style.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetStyleName(string value)
        {
            bool styleExists = this.Worksheet.Workbook.Styles.Contains(value);

            if (!styleExists)
            {
                throw new ArgumentException("This style does not exist in the styles collection of the workbook.");
            }

            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                this.ClearLocallySetPropertiesDefinedByStyle(value);
                this.SetPropertyValue(this.WorksheetEntity.StyleNameProperty, value);
            }
        }

        internal void SetStyleNameAndExpandToFitNumberValuesWidth(string value)
        {
            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
            {
                this.SetStyleName(value);
            });
        }

        private void ClearLocallySetPropertiesDefinedByStyle(string styleName)
        {
            CellStyle cellStyle = this.Worksheet.Workbook.Styles[styleName];
            if (cellStyle == null)
            {
                return;
            }

            WorksheetEntityBase[] worksheetEntities = new WorksheetEntityBase[]
            {
                this.Worksheet.Cells,
                this.Worksheet.Rows,
                this.Worksheet.Columns
            };

            foreach (WorksheetEntityBase entity in worksheetEntities)
            {
                foreach (IProperty property in entity.Properties)
                {
                    if (cellStyle.GetIsPropertyIncluded(property.PropertyDefinition))
                    {
                        if (property.PropertyDefinition == CellPropertyDefinitions.IsLockedProperty)
                        {
                            this.SetIsLockedLocalValueIfSheetIsProtected();
                        }
                        else
                        {
                            this.ClearPropertyValue(property, false);
                        }
                    }
                }
            }
        }

        private void SetIsLockedLocalValueIfSheetIsProtected()
        {
            if (this.Worksheet.IsProtected)
            {
                bool isLockedValue = this.GetIsLocked().Value;
                // When the sheet is protected the current value of the property should override the style
                this.SetIsLocked(isLockedValue);
            }
        }

        /// <summary>
        /// Clears the name of the style.
        /// </summary>
        public void ClearStyleName()
        {
            this.ClearPropertyValue(this.WorksheetEntity.StyleNameProperty);
        }

        #endregion


        #region Fill Property

        /// <summary>
        /// Gets the fill.
        /// </summary>
        /// <returns>The fill value for the selection.</returns>
        public RangePropertyValue<IFill> GetFill()
        {
            return this.GetPropertyValue(this.WorksheetEntity.FillProperty);
        }

        /// <summary>
        /// Sets the fill.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetFill(IFill value)
        {
            this.SetPropertyValue(this.WorksheetEntity.FillProperty, value);
        }

        /// <summary>
        /// Clears the fill.
        /// </summary>
        public void ClearFill()
        {
            this.ClearPropertyValue(this.WorksheetEntity.FillProperty);
        }

        #endregion


        #region FontFamily Property

        /// <summary>
        /// Gets the font family.
        /// </summary>
        /// <returns>The font family for the selection.</returns>
        public RangePropertyValue<ThemableFontFamily> GetFontFamily()
        {
            return this.GetPropertyValue(this.WorksheetEntity.FontFamilyProperty);
        }

        /// <summary>
        /// Sets the font family.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetFontFamily(ThemableFontFamily value)
        {
            this.SetPropertyValue(this.WorksheetEntity.FontFamilyProperty, value);
        }

        internal void SetFontFamilyAndExpandToFitNumberValuesWidth(ThemableFontFamily value)
        {
            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
            {
                this.SetFontFamily(value);
            });
        }

        /// <summary>
        /// Clears the font family.
        /// </summary>
        public void ClearFontFamily()
        {
            this.ClearPropertyValue(this.WorksheetEntity.FontFamilyProperty);
        }

        #endregion


        #region FontSize Property

        /// <summary>
        /// Gets the size of the font.
        /// </summary>
        /// <returns>The font size for the selection.</returns>
        public RangePropertyValue<double> GetFontSize()
        {
            return this.GetPropertyValue(this.WorksheetEntity.FontSizeProperty);
        }

        /// <summary>
        /// Sets the size of the font.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetFontSize(double value)
        {
            Guard.ThrowExceptionIfOutOfRange<double>(UnitHelper.PointToDip(1), UnitHelper.PointToDip(409), value, "value");
            this.SetPropertyValue(this.WorksheetEntity.FontSizeProperty, value);
        }

        internal void SetFontSizeAndExpandToFitNumberValuesWidth(double value)
        {
            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
            {
                this.SetFontSize(value);
            });
        }

        /// <summary>
        /// Clears the size of the font.
        /// </summary>
        public void ClearFontSize()
        {
            this.ClearPropertyValue(this.WorksheetEntity.FontSizeProperty);
        }

        #endregion


        #region IsBold Property

        /// <summary>
        /// Gets the value indicating whether the font is bold.
        /// </summary>
        /// <returns>The value indicating whether the font is bold for the selection.</returns>
        public RangePropertyValue<bool> GetIsBold()
        {
            return this.GetPropertyValue(this.WorksheetEntity.IsBoldProperty);
        }

        /// <summary>
        /// Sets the value indicating whether the font is bold.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetIsBold(bool value)
        {
            this.SetPropertyValue(this.WorksheetEntity.IsBoldProperty, value);
        }

        internal void SetIsBoldAndExpandToFitNumberValuesWidth(bool value)
        {
            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
            {
                this.SetIsBold(value);
            });
        }

        /// <summary>
        /// Clears the bold property.
        /// </summary>
        public void ClearIsBold()
        {
            this.ClearPropertyValue(this.WorksheetEntity.IsBoldProperty);
        }

        #endregion


        #region IsItalic Property

        /// <summary>
        /// Gets the value indicating whether the font is italic.
        /// </summary>
        /// <returns>The value indicating whether the font is italic for the selection.</returns>
        public RangePropertyValue<bool> GetIsItalic()
        {
            return this.GetPropertyValue(this.WorksheetEntity.IsItalicProperty);
        }

        /// <summary>
        /// Sets the value indicating whether the font is italic.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetIsItalic(bool value)
        {
            this.SetPropertyValue(this.WorksheetEntity.IsItalicProperty, value);
        }

        internal void SetIsItalicAndExpandToFitNumberValuesWidth(bool value)
        {
            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
            {
                this.SetIsItalic(value);
            });
        }

        /// <summary>
        /// Clears the italic property.
        /// </summary>
        public void ClearIsItalic()
        {
            this.ClearPropertyValue(this.WorksheetEntity.IsItalicProperty);
        }

        #endregion


        #region Underline Property

        /// <summary>
        /// Gets the underline.
        /// </summary>
        /// <returns>The underline property for the selection.</returns>
        public RangePropertyValue<UnderlineType> GetUnderline()
        {
            return this.GetPropertyValue(this.WorksheetEntity.UnderlineProperty);
        }

        /// <summary>
        /// Sets the underline.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetUnderline(UnderlineType value)
        {
            this.SetPropertyValue(this.WorksheetEntity.UnderlineProperty, value);
        }

        internal void SetUnderlineAndExpandToFitNumberValuesWidth(UnderlineType value)
        {
            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
            {
                this.SetUnderline(value);
            });
        }

        /// <summary>
        /// Clears the underline.
        /// </summary>
        public void ClearUnderline()
        {
            this.ClearPropertyValue(this.WorksheetEntity.UnderlineProperty);
        }

        #endregion


        #region ForeColor Property

        /// <summary>
        /// Gets the font color.
        /// </summary>
        /// <returns>The font color for the selection.</returns>
        public RangePropertyValue<ThemableColor> GetForeColor()
        {
            return this.GetPropertyValue(this.WorksheetEntity.ForeColorProperty);
        }

        /// <summary>
        /// Sets the font color.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetForeColor(ThemableColor value)
        {
            this.SetPropertyValue(this.WorksheetEntity.ForeColorProperty, value);
        }

        /// <summary>
        /// Clears the font color.
        /// </summary>
        public void ClearForeColor()
        {
            this.ClearPropertyValue(this.WorksheetEntity.ForeColorProperty);
        }

        #endregion


        #region HorizontalAlignment Property

        /// <summary>
        /// Gets the horizontal alignment.
        /// </summary>
        /// <returns>The horizontal alignment for the selection.</returns>
        public RangePropertyValue<RadHorizontalAlignment> GetHorizontalAlignment()
        {
            return this.GetPropertyValue(this.WorksheetEntity.HorizontalAlignmentProperty);
        }

        /// <summary>
        /// Sets the horizontal alignment.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetHorizontalAlignment(RadHorizontalAlignment value)
        {
            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                this.EnsureIndentThatRespectsHorizontalAlignment(value);
                this.SetPropertyValue(this.WorksheetEntity.HorizontalAlignmentProperty, value);
            }
        }

        private void EnsureIndentThatRespectsHorizontalAlignment(RadHorizontalAlignment value)
        {
            if (value != RadHorizontalAlignment.Left &&
                value != RadHorizontalAlignment.Right &&
                value != RadHorizontalAlignment.Distributed)
            {
                this.ClearIndent();
            }
        }

        /// <summary>
        /// Clears the horizontal alignment.
        /// </summary>
        public void ClearHorizontalAlignment()
        {
            this.ClearPropertyValue(this.WorksheetEntity.HorizontalAlignmentProperty);
        }

        #endregion


        #region VerticalAlignment Property

        /// <summary>
        /// Gets the vertical alignment.
        /// </summary>
        /// <returns>The vertical alignment for the selection.</returns>
        public RangePropertyValue<RadVerticalAlignment> GetVerticalAlignment()
        {
            return this.GetPropertyValue(this.WorksheetEntity.VerticalAlignmentProperty);
        }

        /// <summary>
        /// Sets the vertical alignment.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetVerticalAlignment(RadVerticalAlignment value)
        {
            this.SetPropertyValue(this.WorksheetEntity.VerticalAlignmentProperty, value);
        }

        /// <summary>
        /// Clears the vertical alignment.
        /// </summary>
        public void ClearVerticalAlignment()
        {
            this.ClearPropertyValue(this.WorksheetEntity.VerticalAlignmentProperty);
        }

        #endregion


        #region Indent Property

        /// <summary>
        /// Gets the indent.
        /// </summary>
        /// <returns>The indent for the selection.</returns>
        public RangePropertyValue<int> GetIndent()
        {
            return this.GetPropertyValue(this.WorksheetEntity.IndentProperty);
        }

        /// <summary>
        /// Sets the indent.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetIndent(int value)
        {
            Guard.ThrowExceptionIfOutOfRange<int>(0, 250, value, "value");
            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                this.EnsureHorizontalAlignmentThatRespectsIndent();
                this.SetPropertyValue(this.WorksheetEntity.IndentProperty, value);
            }
        }

        internal void SetIndentAndExpandToFitNumberValuesWidth(int value)
        {
            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
            {
                this.SetIndent(value);
            });
        }

        /// <summary>
        /// Clears the indent.
        /// </summary>
        public void ClearIndent()
        {
            this.ClearPropertyValue(this.WorksheetEntity.IndentProperty);
        }

        /// <summary>
        /// Increases the indent.
        /// </summary>
        public void IncreaseIndent()
        {
            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                this.EnsureHorizontalAlignmentThatRespectsIndent();
                this.ExecuteAndExpandToFitNumberValuesWidth(() =>
                {
                    this.ExecuteForEachRangeInsideBeginEndUpdate(
                        (cellRange) =>
                        {
                            UpdateCellPropertyCommandContext<int> context = new UpdateCellPropertyCommandContext<int>(
                                this.Worksheet, CellPropertyDefinitions.IndentProperty, cellRange,
                                (i) =>
                                {
                                    i++;
                                    return Math.Min(i, SpreadsheetDefaultValues.MaxIndent);
                                });

                            this.Worksheet.ExecuteCommand(WorkbookCommands.UpdateIndent, context);
                        });
                });
            }
        }

        /// <summary>
        /// Decreases the indent.
        /// </summary>
        public void DecreaseIndent()
        {
            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                this.EnsureHorizontalAlignmentThatRespectsIndent();

                this.ExecuteAndExpandToFitNumberValuesWidth(() =>
                {
                    this.ExecuteForEachRangeInsideBeginEndUpdate(
                        (cellRange) =>
                        {
                            UpdateCellPropertyCommandContext<int> context = new UpdateCellPropertyCommandContext<int>(
                                this.Worksheet, CellPropertyDefinitions.IndentProperty, cellRange,
                                (i) =>
                                {
                                    i--;
                                    return Math.Max(i, 0);
                                });

                            this.Worksheet.ExecuteCommand(WorkbookCommands.UpdateIndent, context);
                        });
                });
            }
        }

        private void EnsureHorizontalAlignmentThatRespectsIndent()
        {
            RangePropertyValue<RadHorizontalAlignment> horizontalAlignment = this.GetHorizontalAlignment();
            if (horizontalAlignment.IsIndeterminate ||
                !LayoutHelper.ShouldRespectIndent(horizontalAlignment.Value))
            {
                this.SetHorizontalAlignment(RadHorizontalAlignment.Left);
            }
        }

        #endregion


        #region IsWrapped Property

        /// <summary>
        /// Gets a value indicating whether text wrapping is used.
        /// </summary>
        /// <returns>The value indicating whether text wrapping is used.</returns>
        public RangePropertyValue<bool> GetIsWrapped()
        {
            return this.GetPropertyValue(this.WorksheetEntity.IsWrappedProperty);
        }

        /// <summary>
        /// Sets a value indicating whether text wrapping is used.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetIsWrapped(bool value)
        {
            this.SetPropertyValue(this.WorksheetEntity.IsWrappedProperty, value);
        }

        internal void SetIsWrappedAndExpandToFitNumberValuesWidth(bool value)
        {
            this.ExecuteAndExpandToFitNumberValuesWidth(() =>
            {
                this.SetIsWrapped(value);
            });
        }

        /// <summary>
        /// Clears the text wrapping.
        /// </summary>
        public void ClearIsWrapped()
        {
            this.ClearPropertyValue(this.WorksheetEntity.IsWrappedProperty);
        }

        #endregion


        #region Borders

        /// <summary>
        /// Gets the left border.
        /// </summary>
        /// <returns>The left border for the selection.</returns>
        internal RangePropertyValue<CellBorder> GetLeftBorder()
        {
            return this.GetPropertyValue(this.WorksheetEntity.LeftBorderProperty);
        }

        /// <summary>
        /// Gets the right border.
        /// </summary>
        /// <returns>The right border for the selection.</returns>
        internal RangePropertyValue<CellBorder> GetRightBorder()
        {
            return this.GetPropertyValue(this.WorksheetEntity.RightBorderProperty);
        }

        /// <summary>
        /// Gets the top border.
        /// </summary>
        /// <returns>The top border for the selection.</returns>
        internal RangePropertyValue<CellBorder> GetTopBorder()
        {
            return this.GetPropertyValue(this.WorksheetEntity.TopBorderProperty);
        }

        internal RangePropertyValue<CellBorder> GetBottomBorder()
        {
            return this.GetPropertyValue(this.WorksheetEntity.BottomBorderProperty);
        }

        internal RangePropertyValue<CellBorder> GetDiagonalUpBorder()
        {
            return this.GetPropertyValue(this.WorksheetEntity.DiagonalUpBorderProperty);
        }

        internal RangePropertyValue<CellBorder> GetDiagonalDownBorder()
        {
            return this.GetPropertyValue(this.WorksheetEntity.DiagonalDownBorderProperty);
        }

        private static CellBorder GetActualBorderByPriority(RangePropertyValue<CellBorder> firstBorder, RangePropertyValue<CellBorder> secondBorder, ThemeColorScheme colorScheme)
        {
            if (firstBorder == null)
            {
                if (!secondBorder.IsIndeterminate)
                {
                    return secondBorder.Value;
                }
                else
                {
                    return null;
                }
            }

            if (secondBorder == null)
            {
                if (!firstBorder.IsIndeterminate)
                {
                    return firstBorder.Value;
                }
                else
                {
                    return null;
                }
            }

            if (firstBorder.IsIndeterminate || secondBorder.IsIndeterminate)
            {
                return null;
            }

            return CellBorder.GetWithMaxPriority(firstBorder.Value, secondBorder.Value, colorScheme);
        }

        private static CellBorder GetActualBorder(RangePropertyValue<CellBorder> border, ThemeColorScheme colorScheme)
        {
            return GetActualBorderByPriority(border, border, colorScheme);
        }

        private CellBorder GetRowHorizontalBorder(CellRange currentRow, bool isBottom = false)
        {
            Guard.ThrowExceptionIfGreaterThan(1, currentRow.RowCount, "RowCount");

            CellRange previousRow;
            if (isBottom)
            {
                previousRow = currentRow;
                currentRow = currentRow.OffsetFromRow(1);
            }
            else
            {
                previousRow = currentRow.OffsetFromRow(-1);
            }

            RangePropertyValue<CellBorder> currentRowTopBorder = null;
            if (currentRow != null)
            {
                currentRowTopBorder = this.GetPropertyValue(this.WorksheetEntity.TopBorderProperty, currentRow);
            }

            RangePropertyValue<CellBorder> previousRowBottomBorder = null;
            if (previousRow != null)
            {
                previousRowBottomBorder = this.GetPropertyValue(this.WorksheetEntity.BottomBorderProperty, previousRow);
            }

            return GetActualBorderByPriority(currentRowTopBorder, previousRowBottomBorder, this.Worksheet.Workbook.Theme.ColorScheme);
        }

        private CellBorder GetColumnVerticalBorder(CellRange currentColumn, bool isRight = false)
        {
            Guard.ThrowExceptionIfGreaterThan(1, currentColumn.ColumnCount, "ColumnCount");

            CellRange previousColumn;
            if (isRight)
            {
                previousColumn = currentColumn;
                currentColumn = currentColumn.OffsetFromColumn(1);
            }
            else
            {
                previousColumn = currentColumn.OffsetFromColumn(-1);
            }

            RangePropertyValue<CellBorder> currentColumnLeftBorder = null;
            if (currentColumn != null)
            {
                currentColumnLeftBorder = this.GetPropertyValue(this.WorksheetEntity.LeftBorderProperty, currentColumn);
            }

            RangePropertyValue<CellBorder> previousColumnRightBorder = null;
            if (previousColumn != null)
            {
                previousColumnRightBorder = this.GetPropertyValue(this.WorksheetEntity.RightBorderProperty, previousColumn);
            }

            return GetActualBorderByPriority(currentColumnLeftBorder, previousColumnRightBorder, this.Worksheet.Workbook.Theme.ColorScheme);
        }

        private CellBorder GetInsideHorizontalBorder(CellRange cellRange)
        {
            if (cellRange.RowCount == 1)
            {
                return null;
            }

            CellRange bottomBorderCellRange = new CellRange(cellRange.FromIndex.RowIndex, cellRange.FromIndex.ColumnIndex, cellRange.ToIndex.RowIndex - 1, cellRange.ToIndex.ColumnIndex);
            RangePropertyValue<CellBorder> bottomBorder = this.GetPropertyValue(this.WorksheetEntity.BottomBorderProperty, bottomBorderCellRange);

            RangePropertyValue<CellBorder> topBorder = null;
            if (TelerikHelper.IsValidRowIndex(cellRange.FromIndex.RowIndex + 1))
            {
                CellRange topBorderCellRange = new CellRange(cellRange.FromIndex.RowIndex + 1, cellRange.FromIndex.ColumnIndex, cellRange.ToIndex.RowIndex, cellRange.ToIndex.ColumnIndex);
                topBorder = this.GetPropertyValue(this.WorksheetEntity.TopBorderProperty, topBorderCellRange);
            }

            return GetActualBorderByPriority(bottomBorder, topBorder, this.Worksheet.Workbook.Theme.ColorScheme);
        }

        private CellBorder GetInsideVerticalBorder(CellRange cellRange)
        {
            if (cellRange.ColumnCount == 1)
            {
                return null;
            }

            CellRange rightBorderCellRange = new CellRange(cellRange.FromIndex.RowIndex, cellRange.FromIndex.ColumnIndex, cellRange.ToIndex.RowIndex, cellRange.ToIndex.ColumnIndex - 1);
            RangePropertyValue<CellBorder> rightBorder = this.GetPropertyValue(this.WorksheetEntity.RightBorderProperty, rightBorderCellRange);

            RangePropertyValue<CellBorder> leftBorder = null;
            if (TelerikHelper.IsValidColumnIndex(cellRange.FromIndex.ColumnIndex + 1))
            {
                CellRange leftBorderCellRange = new CellRange(cellRange.FromIndex.RowIndex, cellRange.FromIndex.ColumnIndex + 1, cellRange.ToIndex.RowIndex, cellRange.ToIndex.ColumnIndex);
                leftBorder = this.GetPropertyValue(this.WorksheetEntity.LeftBorderProperty, leftBorderCellRange);
            }

            return GetActualBorderByPriority(rightBorder, leftBorder, this.Worksheet.Workbook.Theme.ColorScheme);
        }

        private CellBorder GetDiagonalUpBorder(CellRange cellRange)
        {
            return GetActualBorder(this.GetPropertyValue(this.WorksheetEntity.DiagonalUpBorderProperty, cellRange), this.Worksheet.Workbook.Theme.ColorScheme);
        }

        private CellBorder GetDiagonalDownBorder(CellRange cellRange)
        {
            return GetActualBorder(this.GetPropertyValue(this.WorksheetEntity.DiagonalDownBorderProperty, cellRange), this.Worksheet.Workbook.Theme.ColorScheme);
        }

        private CellBorders GetCellBorders(CellRange cellRange)
        {
            Guard.ThrowExceptionIfNull(cellRange, "cellRange");

            CellBorder leftBorder = this.GetColumnVerticalBorder(cellRange.GetFirstColumn());
            CellBorder topBorder = this.GetRowHorizontalBorder(cellRange.GetFirstRow());
            CellBorder rightBorder = this.GetColumnVerticalBorder(cellRange.GetLastColumn(), true);
            CellBorder bottomBorder = this.GetRowHorizontalBorder(cellRange.GetLastRow(), true);
            CellBorder insideHorizontalBorder = this.GetInsideHorizontalBorder(cellRange);
            CellBorder insideVerticalBorder = this.GetInsideVerticalBorder(cellRange);
            CellBorder diagonalUpBorder = this.GetDiagonalUpBorder(cellRange);
            CellBorder diagonalDownBorder = this.GetDiagonalDownBorder(cellRange);

            return new CellBorders(leftBorder, topBorder, rightBorder, bottomBorder, insideHorizontalBorder, insideVerticalBorder, diagonalUpBorder, diagonalDownBorder);
        }

        /// <summary>
        /// Gets the borders.
        /// </summary>
        /// <returns>The cell borders.</returns>
        public CellBorders GetBorders()
        {
            CellBorders resultCellBorders = null;
            foreach (CellRange cellRange in this.CellRanges)
            {
                CellBorders currentRangeBorders = this.GetCellBorders(cellRange);
                if (resultCellBorders == null)
                {
                    resultCellBorders = currentRangeBorders;
                }
                else if (resultCellBorders != currentRangeBorders)
                {
                    resultCellBorders.MergeWith(currentRangeBorders);
                }
            }

            return resultCellBorders;
        }

        /// <summary>
        /// Sets the borders.
        /// </summary>
        /// <param name="value">The value.</param>
        public virtual void SetBorders(CellBorders value)
        {
        }

        /// <summary>
        /// Clears the borders.
        /// </summary>
        public void ClearBorders()
        {
            // TODO: Implement clear command.
            //this.SetBorders(CellBorders.Default);

            this.ClearPropertyValue(this.WorksheetEntity.LeftBorderProperty);
            this.ClearPropertyValue(this.WorksheetEntity.TopBorderProperty);
            this.ClearPropertyValue(this.WorksheetEntity.RightBorderProperty);
            this.ClearPropertyValue(this.WorksheetEntity.BottomBorderProperty);
            this.ClearPropertyValue(this.WorksheetEntity.DiagonalUpBorderProperty);
            this.ClearPropertyValue(this.WorksheetEntity.DiagonalDownBorderProperty);
        }

        #endregion


        #region Protection

        /// <summary>
        /// Gets the value indicating whether the selection is locked.
        /// </summary>
        /// <returns>The value indicating whether the selection is locked.</returns>
        public RangePropertyValue<bool> GetIsLocked()
        {
            return this.GetPropertyValue(this.WorksheetEntity.IsLockedProperty);
        }

        /// <summary>
        /// Sets the value indicating whether the selection is locked.
        /// </summary>
        /// <param name="value">The value.</param>
        public void SetIsLocked(bool value)
        {
            this.SetPropertyValue(this.WorksheetEntity.IsLockedProperty, value);
        }

        /// <summary>
        /// Clears the is locked property.
        /// </summary>
        public void ClearIsLocked()
        {
            this.ClearPropertyValue(this.WorksheetEntity.IsLockedProperty);
        }

        #endregion
    }
}