﻿using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Layout;
using Telerik.Windows.Documents.Spreadsheet.Model.Shapes;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents the state of the selection in the worksheet.
    /// </summary>
    public class SelectionState
    {
		#region Fields

		private readonly List<CellRange> selectedRanges;
		private readonly CellIndex activeCellIndex;
        private readonly ViewportPaneType pane;
        private readonly List<FloatingShapeBase> selectedShapes;
        private readonly bool isCellSelection;

		#endregion


		#region Properties

        /// <summary>
        /// Gets the selected ranges.
        /// </summary>
        /// <value>The selected ranges.</value>
        public IEnumerable<CellRange> SelectedRanges
        {
			get
			{
				return this.selectedRanges;
			}
		}

        /// <summary>
        /// Gets the index of the active cell.
        /// </summary>
        /// <value>The index of the active cell.</value>
        public CellIndex ActiveCellIndex
        {
			get
			{
				return this.activeCellIndex;
			}
		}

        /// <summary>
        /// Gets the pane.
        /// </summary>
        /// <value>The pane.</value>
        public ViewportPaneType Pane
        {
            get
            {
                return this.pane;
            }
        }

        internal IEnumerable<FloatingShapeBase> SelectedShapes
        {
            get
            {
                return this.selectedShapes;
            }
        }

        internal bool IsCellSelection
        {
            get
            {
                return this.isCellSelection;
            }
        }

		#endregion


		#region Constructors

        internal SelectionState()
            : this(new CellRange[] { new CellRange(0, 0, 0, 0) }, new CellIndex(0, 0), ViewportPaneType.Scrollable, new List<FloatingShapeBase>(), true)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SelectionState" /> class.
        /// </summary>
        /// <param name="selectedRanges">The selected ranges.</param>
        /// <param name="activeCellIndex">Index of the active cell.</param>
        /// <param name="pane">The pane.</param>
        public SelectionState(IEnumerable<CellRange> selectedRanges, CellIndex activeCellIndex, ViewportPaneType pane) : this(selectedRanges, activeCellIndex, pane, new List<FloatingShapeBase>(), true)
        {
        }

        internal SelectionState(IEnumerable<CellRange> selectedRanges, CellIndex activeCellIndex, ViewportPaneType pane, IEnumerable<FloatingShapeBase> selectedShapes, bool isCellSelection)
        {
            Guard.ThrowExceptionIfNull(selectedRanges, "selectedRanges");
            Guard.ThrowExceptionIfNull(activeCellIndex, "activeCellIndex");
            Guard.ThrowExceptionIfNull(selectedShapes, "selectedShapes");

            this.selectedShapes = new List<FloatingShapeBase>(selectedShapes);
            this.selectedRanges = new List<CellRange>(selectedRanges);
            this.activeCellIndex = activeCellIndex;
            this.pane = pane;
            this.isCellSelection = isCellSelection;
        }

		#endregion


        #region Methods

        /// <summary>
        /// Clones this instance.
        /// </summary>
        /// <returns>The cloned selection state.</returns>
        public SelectionState Clone()
        {
            return new SelectionState(this.SelectedRanges, this.ActiveCellIndex, this.Pane, this.SelectedShapes, this.IsCellSelection);
        }

        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current
        /// <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
            return string.Format("Pane: {0}; ActiveCellIndex: {1}; SelectedRanges: {2}", this.Pane, this.ActiveCellIndex, string.Join("\n", this.SelectedRanges));
        }

        #endregion
    }
}