﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Model.Drawing.Charts;
using Telerik.Windows.Documents.Model.Drawing.Shapes;
using Telerik.Windows.Documents.Spreadsheet.Model.Charts;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Shapes
{
    /// <summary>
    /// Represents a shape which can contain a chart.
    /// </summary>
    public class FloatingChartShape : FloatingShapeBase
    {
        private readonly ChartShape chartShape;

        /// <summary>
        /// Gets the type of the shape.
        /// </summary>
        public override FloatingShapeType FloatingShapeType
        {
            get
            {
                return FloatingShapeType.Chart;
            }
        }

        /// <summary>
        /// Gets or sets the chart contained in the shape.
        /// </summary>
        public DocumentChart Chart
        {
            get
            {
                return this.chartShape.Chart;
            }
            set
            {
                if (this.chartShape.Chart != value)
                {
                    Guard.ThrowExceptionIfNull(value, "value");
                    this.chartShape.Chart.ChartChanged -= this.Chart_ChartChanged;
                    this.chartShape.Chart = value;
                    this.chartShape.Chart.ChartChanged += this.Chart_ChartChanged;

                    this.OnShapeChanged();
                }
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="FloatingChartShape"/> class.
        /// </summary>
        /// <param name="worksheet">The worksheet that the shape is associated with.</param>
        /// <param name="cellIndex">The cell index where the top left corner of the shape is positioned.</param>
        /// <param name="chartDataRange">The range containing the chart data.</param>
        /// <param name="chartTypes">The types of chart that will be created. Passing more than one type will create a combo chart.</param>
        public FloatingChartShape(Worksheet worksheet, CellIndex cellIndex, CellRange chartDataRange, params ChartType[] chartTypes)
            : this(worksheet, cellIndex, chartDataRange, SeriesRangesOrientation.Automatic, chartTypes)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="FloatingChartShape"/> class.
        /// </summary>
        /// <param name="worksheet">The worksheet that the shape is associated with.</param>
        /// <param name="cellIndex">The cell index where the top left corner of the shape is positioned.</param>
        /// <param name="chartDataRange">The range containing the chart data.</param>
        /// <param name="seriesRangesOrientation">A value indicating whether the series of the chart will refer to vertical or horizontal ranges or the direction will be determined automatically.</param>
        /// <param name="chartTypes">The types of chart that will be created. Passing more than one type will create a combo chart.</param>
        public FloatingChartShape(Worksheet worksheet, CellIndex cellIndex, CellRange chartDataRange, SeriesRangesOrientation seriesRangesOrientation, params ChartType[] chartTypes)
            : this(worksheet, cellIndex, new ChartShape())
        {
            DocumentChart chart = DocumentChartFactory.CreateChartFromRangeOfType(worksheet, chartDataRange, seriesRangesOrientation, chartTypes);

            this.chartShape.Chart = chart;
            this.chartShape.Chart.ChartChanged += this.Chart_ChartChanged;

            this.Outline.Fill = SpreadsheetDefaultValues.ChartLinesDefaultFill;
            this.Outline.Width = SpreadsheetDefaultValues.ChartLinesDefaultWidth;
            this.Fill = SpreadsheetDefaultValues.ChartDefaultFill;
        }

        internal FloatingChartShape(Worksheet worksheet, CellIndex cellIndex, ChartShape chartShape)
            : base(worksheet, chartShape, cellIndex, 0, 0)
        {
            this.chartShape = chartShape;
        }

        private void Chart_ChartChanged(object sender, EventArgs e)
        {
            this.Worksheet.InvalidateLayout();
        }

        internal override FloatingShapeBase Copy(Worksheet worksheet, CellIndex cellIndex, double offsetX, double offsetY)
        {
            return new FloatingChartShape(worksheet, cellIndex, new ChartShape(this.chartShape))
            {
                OffsetX = offsetX,
                OffsetY = offsetY
            };
        }

        internal void OnWorksheetCellPropertyChanged(Worksheet source, CellPropertyChangedEventArgs e)
        {
            string propertyName = e.Property.Name;

            if (propertyName != CellPropertyDefinitions.FormatProperty.Name &&
                propertyName != CellPropertyDefinitions.StyleNameProperty.Name)
            {
                return;
            }

            CellRange cellRange = e.CellRange;

            SeriesGroup firstSeriesGroup = this.Chart.SeriesGroups.First();
            SeriesBase firstSeries = firstSeriesGroup.Series.First();

            WorkbookFormulaChartData seriesData = firstSeries.VerticalSeriesData as WorkbookFormulaChartData;

            this.UpdateSeriesFormat(source, cellRange, seriesData);
        }

        private void UpdateSeriesFormat(Worksheet source, CellRange cellRange, WorkbookFormulaChartData seriesData)
        {
            if (seriesData != null)
            {
                IEnumerable<CellRange> cellRanges;
                Worksheet worksheet;
                if (seriesData.TryEnumerateCellRanges(out cellRanges, out worksheet))
                {
                    CellRange firstRange = cellRanges.First();
                    CellIndex numberFormatIndex = firstRange.FromIndex;
                    if (worksheet == source && cellRange.Contains(numberFormatIndex))
                    {
                        long numberFormatLongIndex = WorksheetPropertyBagBase.ConvertCellIndexToLong(numberFormatIndex);
                        CellValueFormat numberFormat = worksheet.Cells.GetPropertyValueRespectingStyle(CellPropertyDefinitions.FormatProperty, numberFormatLongIndex, numberFormatLongIndex).Single().Value;
                        this.Chart.PrimaryAxes.ValueAxis.NumberFormat = numberFormat.FormatString;
                    }
                }
            }
        }
    }
}
