﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Model.Data;
using Telerik.Windows.Documents.Model.Drawing.Shapes;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.Copying;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Shapes
{
    /// <summary>
    /// Represents a picture element in the RadSpreadsheet.
    /// </summary>
    public class FloatingImage : FloatingShapeBase, ICopyable<FloatingShapeBase>
    {
        private readonly Image image;

        /// <summary>
        /// Gets the type of the shape.
        /// </summary>
        public override FloatingShapeType FloatingShapeType
        {
            get
            {
                return FloatingShapeType.Image;
            }
        }

        /// <summary>
        /// Gets or sets the image source.
        /// </summary>
        /// <value>The image source.</value>
        public ImageSource ImageSource
        {
            get
            {
                return this.image.ImageSource;
            }
            set
            {
                SetImageImageSourceCommandContext context = new SetImageImageSourceCommandContext(this.Worksheet, this, value);
                WorkbookCommands.SetImageImageSource.Execute(context);
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether to resize relative to the original size.
        /// </summary>
        /// <value>The prefer relative to original resize.</value>
        public bool PreferRelativeToOriginalResize
        {
            get
            {
                return this.image.PreferRelativeToOriginalResize;
            }
            set
            {
                if (this.image.PreferRelativeToOriginalResize != value)
                {
                    SetImagePreferRelativeToOriginalResizeCommandContext context = new SetImagePreferRelativeToOriginalResizeCommandContext(this.Worksheet, this, value);
                    WorkbookCommands.SetImagePreferRelativeToOriginalResizeCommand.Execute(context);
                }
            }
        }

        internal Image Image
        {
            get
            {
                return this.image;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="FloatingImage" /> class.
        /// </summary>
        /// <param name="worksheet">The worksheet assigned to the image.</param>
        /// <param name="cellIndex">The cell index where the top left corner of the shape is positioned.</param>
        /// <param name="offsetX">The left offset of the top left corner of the shape relative to the top left corner of the cell index.</param>
        /// <param name="offsetY">The top offset of the top left corner of the shape relative to the top left corner of the cell index.</param>
        public FloatingImage(Worksheet worksheet, CellIndex cellIndex, double offsetX, double offsetY) 
            : this(worksheet, cellIndex, offsetX, offsetY, new Image())
        {
        }

        internal FloatingImage(Worksheet worksheet, CellIndex cellIndex, double offsetX, double offsetY, Image image)
            : base(worksheet, image, cellIndex, offsetX, offsetY)
        {
            this.image = image;
            this.image.ImageSourceChanged += this.Image_ImageSourceChanged;
        }

        internal override FloatingShapeBase Copy(Worksheet worksheet, CellIndex cellIndex, double offsetX, double offsetY)
        {
            return new FloatingImage(worksheet, cellIndex, offsetX, offsetY, new Image(this.Image));
        }

        FloatingShapeBase ICopyable<FloatingShapeBase>.Copy(CopyContext context)
        {
            FloatingShapeBase sourceImage = context.SourceFloatingShape;
            FloatingImage clonedImage = 
                new FloatingImage(context.TargetWorksheet, sourceImage.CellIndex, sourceImage.OffsetX, sourceImage.OffsetY, this.Image.Clone(true));
            
            context.TargetWorksheet.Workbook.Resources.RegisterResource(clonedImage.ImageSource);

            return clonedImage;
        }

        private void Image_ImageSourceChanged(object sender, ResourceChangedEventArgs e)
        {
            if (e.OldValue != null)
            {
                this.Worksheet.Workbook.Resources.ReleaseResource(e.OldValue);
            }

            if (e.NewValue != null)
            {
                this.Worksheet.Workbook.Resources.RegisterResource(e.NewValue);
            }
        }
    }
}
