﻿using System;
using System.Windows;
using Telerik.Windows.Documents.Model.Drawing.Shapes;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Model.Drawing.Theming;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.Layout;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Shapes
{
    /// <summary>
    /// Represents a base for all shape elements in the RadSpreadsheet.
    /// </summary>
    public abstract class FloatingShapeBase
    {
        private readonly ShapeBase shape;
        private readonly Worksheet worksheet;

        private CellIndex cellIndex;
        private double offsetX;
        private double offsetY;

        private bool cellIndexAdjustedForRotation;

        /// <summary>
        /// Gets the type of the shape.
        /// </summary>
        public abstract FloatingShapeType FloatingShapeType
        {
            get;
        }

        /// <summary>
        /// Gets the worksheet assigned to the shape.
        /// </summary>
        /// <value>The worksheet.</value>
        public Worksheet Worksheet
        {
            get
            {
                return this.worksheet;
            }
        }

        /// <summary>
        /// Gets or sets the cell index where the top left corner of the shape is positioned.
        /// </summary>
        /// <value>The cell index where the top left corner of the shape is positioned.</value>
        public CellIndex CellIndex
        {
            get
            {
                return this.cellIndex;
            }
            set
            {
                if (this.cellIndex != value)
                {
                    SetShapePositionCommandContext context = new SetShapePositionCommandContext(this.worksheet, this, value, this.OffsetX, this.OffsetY);
                    WorkbookCommands.SetShapePosition.Execute(context);
                }
            }
        }

        /// <summary>
        /// Gets or sets the left offset of the top left corner of the shape relative to the top left corner of the cell index.
        /// </summary>
        /// <value>The left offset.</value>
        public double OffsetX
        {
            get
            {
                return this.offsetX;
            }
            set
            {
                if (this.offsetX != value)
                {
                    SetShapePositionCommandContext context = new SetShapePositionCommandContext(this.worksheet, this, this.CellIndex, value, this.OffsetY);
                    WorkbookCommands.SetShapePosition.Execute(context);
                }
            }
        }

        /// <summary>
        /// Gets or sets the top offset of the top left corner of the shape relative to the top left corner of the cell index.
        /// </summary>
        /// <value>The top offset.</value>
        public double OffsetY
        {
            get
            {
                return this.offsetY;
            }
            set
            {
                if (this.offsetY != value)
                {
                    SetShapePositionCommandContext context = new SetShapePositionCommandContext(this.worksheet, this, this.CellIndex, this.OffsetX, value);
                    WorkbookCommands.SetShapePosition.Execute(context);
                }
            }
        }

        /// <summary>
        /// Gets the id of the shape, which is unique for the worksheet it belongs to.
        /// </summary>
        /// <value>The id of the shape.</value>
        public int Id
        {
            get
            {
                return this.shape.Id;
            }
        }

        /// <summary>
        /// Gets or sets the name of the shape.
        /// </summary>
        /// <value>The name of the shape.</value>
        public string Name
        {
            get
            {
                return this.shape.Name;
            }
            set
            {
                if (this.shape.Name != value)
                {
                    SetShapeNameCommandContext context = new SetShapeNameCommandContext(this.worksheet, this, value);
                    WorkbookCommands.SetShapeName.Execute(context);

                    this.OnShapeChanged();
                }
            }
        }

        /// <summary>
        /// Gets or sets the width of the shape.
        /// </summary>
        /// <value>The width of the shape.</value>
        public double Width
        {
            get
            {
                return this.shape.Width;
            }
            set
            {
                this.SetWidth(false, value);
            }
        }

        /// <summary>
        /// Gets or sets the height of the shape.
        /// </summary>
        /// <value>The height of the shape.</value>
        public double Height
        {
            get
            {
                return this.shape.Height;
            }
            set
            {
                this.SetHeight(false, value);
            }
        }

        /// <summary>
        /// Gets or sets the value indicating whether the shape is horizontally flipped.
        /// </summary>
        /// <value>The value indicating whether the shape is horizontally flipped.</value>
        public bool IsHorizontallyFlipped
        {
            get
            {
                return this.shape.IsHorizontallyFlipped;
            }
            set
            {
                if (this.shape.IsHorizontallyFlipped != value)
                {
                    SetShapeFlipCommandContext context = new SetShapeFlipCommandContext(this.worksheet, this, value, this.IsVerticallyFlipped);
                    WorkbookCommands.SetShapeFlip.Execute(context);

                    this.OnShapeChanged();
                }
            }
        }

        /// <summary>
        /// Gets or sets the value indicating whether the shape is vertically flipped.
        /// </summary>
        /// <value>The value indicating whether the shape is vertically flipped.</value>
        public bool IsVerticallyFlipped
        {
            get
            {
                return this.shape.IsVerticallyFlipped;
            }
            set
            {
                if (this.shape.IsVerticallyFlipped != value)
                {
                    SetShapeFlipCommandContext context = new SetShapeFlipCommandContext(this.worksheet, this, this.IsHorizontallyFlipped, value);
                    WorkbookCommands.SetShapeFlip.Execute(context);

                    this.OnShapeChanged();
                }
            }
        }

        /// <summary>
        /// Gets or sets the rotation angle of the shape.
        /// </summary>
        /// <value>The rotation angle of the shape.</value>
        public double RotationAngle
        {
            get
            {
                return this.shape.RotationAngle;
            }
            set
            {
                this.SetRotationAngle(value, false);
            }
        }

        /// <summary>
        /// Gets or sets the value indicating whether the aspect ratio between the width and height should remain constant.
        /// </summary>
        /// <value>The value indicating whether the aspect ratio between the width and height should remain constant.</value>
        public bool LockAspectRatio
        {
            get
            {
                return this.shape.LockAspectRatio;
            }
            set
            {
                if (this.shape.LockAspectRatio != value)
                {
                    SetShapeLockAspectRatioCommandContext context = new SetShapeLockAspectRatioCommandContext(this.worksheet, this, value);
                    WorkbookCommands.SetShapeLockAspectRatio.Execute(context);
                }
            }
        }

        // At this point this property works for FloatingChartShape only in the UI.
        /// <summary>
        /// Gets or sets the fill of the shape.
        /// </summary>
        public Fill Fill
        {
            get
            {
                return this.shape.Fill;
            }
            set
            {
                this.shape.Fill = value;
            }
        }

        // At this point this property works for FloatingChartShape only in the UI.
        /// <summary>
        /// Gets the outline of the shape.
        /// </summary>
        public Outline Outline
        {
            get
            {
                return this.shape.Outline;
            }
        }

        internal ShapeBase Shape
        {
            get
            {
                return this.shape;
            }
        }

        internal CellIndex CellIndexInternal
        {
            get
            {
                return this.cellIndex;
            }
            set
            {
                if (this.cellIndex != value)
                {
                    this.cellIndex = value;
                    this.OnShapeChanged();
                }
            }
        }

        internal double OffsetXInternal
        {
            get
            {
                return this.offsetX;
            }
            set
            {
                if (this.offsetX != value)
                {
                    this.offsetX = value;
                    this.OnShapeChanged();
                }
            }
        }

        internal double OffsetYInternal
        {
            get
            {
                return this.offsetY;
            }
            set
            {
                if (this.offsetY != value)
                {
                    this.offsetY = value;
                    this.OnShapeChanged();
                }
            }
        }

        internal FloatingShapeBase(Worksheet worksheet, ShapeBase shape, CellIndex cellIndex, double offsetX, double offsetY)
        {
            this.worksheet = worksheet;
            this.CellIndexInternal = cellIndex;
            this.OffsetXInternal = offsetX;
            this.OffsetYInternal = offsetY;
            this.shape = shape;

            this.cellIndexAdjustedForRotation = this.DoesRotationAngleRequireCellIndexChange();
        }

        /// <summary>
        /// Sets the width of the shape.
        /// </summary>
        /// <param name="respectLockAspectRatio">A value indicating whether the aspect ratio lock should be respected.</param>
        /// <param name="width">The new width.</param>
        /// <param name="adjustCellIndex">A value indicating whether the top cell index and offset of the image should be adjusted.</param>
        public void SetWidth(bool respectLockAspectRatio, double width, bool adjustCellIndex = false)
        {
            SetShapeSizeCommandContext context = new SetShapeSizeCommandContext(this.worksheet, this, width, respectLockAspectRatio, adjustCellIndex);
            WorkbookCommands.SetShapeWidth.Execute(context);
        }

        internal void SetWidthInternal(bool respectLockAspectRatio, double width, bool adjustCellIndex)
        {
            if (!this.Width.EqualsDouble(width))
            {
                if (adjustCellIndex)
                {
                    double oldWidth = this.Width;
                    double oldHeight = this.Height;
                    double newWidth = width;

                    RadWorksheetLayout layout = this.Worksheet.Workbook.GetWorksheetLayout(this.worksheet, false);
                    Point oldTopLeftPoint = ShapesResizeHelper.GetShapeTopLeftConsideringAdjustmentForRotation(this, layout);

                    this.Shape.SetWidth(respectLockAspectRatio, width);

                    double newHeight = this.Height;

                    this.AdjustCellIndexAndOffsetForSize(layout, oldWidth, oldHeight, newWidth, newHeight, oldTopLeftPoint);
                }
                else
                {
                    this.Shape.SetWidth(respectLockAspectRatio, width);
                }

                this.OnShapeChanged();
            }
        }

        /// <summary>
        /// Sets the height of the shape.
        /// </summary>
        /// <param name="respectLockAspectRatio">A value indicating whether the aspect ratio lock should be respected.</param>
        /// <param name="height">The new height.</param>
        /// <param name="adjustCellIndex">A value indicating whether the top cell index and offset of the image should be adjusted.</param>
        public void SetHeight(bool respectLockAspectRatio, double height, bool adjustCellIndex = false)
        {
            SetShapeSizeCommandContext context = new SetShapeSizeCommandContext(this.worksheet, this, height, respectLockAspectRatio, adjustCellIndex);
            WorkbookCommands.SetShapeHeight.Execute(context);
        }

        internal void SetHeightInternal(bool respectLockAspectRatio, double height, bool adjustCellIndex)
        {
            if (!this.Height.EqualsDouble(height))
            {
                if (adjustCellIndex)
                {
                    double oldHeight = this.Height;
                    double oldWidth = this.Width;
                    double newHeight = height;

                    RadWorksheetLayout layout = this.Worksheet.Workbook.GetWorksheetLayout(this.worksheet, false);
                    Point oldTopLeftPoint = ShapesResizeHelper.GetShapeTopLeftConsideringAdjustmentForRotation(this, layout);

                    this.Shape.SetHeight(respectLockAspectRatio, height);

                    double newWidth = this.Width;

                    this.AdjustCellIndexAndOffsetForSize(layout, oldWidth, oldHeight, newWidth, newHeight, oldTopLeftPoint);
                }
                else
                {
                    this.Shape.SetHeight(respectLockAspectRatio, height);
                }

                this.OnShapeChanged();
            }
        }

        /// <summary>
        /// Sets the rotation angle.
        /// </summary>
        /// <param name="rotationAngle">The rotation angle.</param>
        /// <param name="adjustCellIndex">A value indicating whether the top cell index and offset of the image should be adjusted.</param>
        public void SetRotationAngle(double rotationAngle, bool adjustCellIndex = false)
        {
            SetShapeRotationAngleCommandContext context = new SetShapeRotationAngleCommandContext(this.worksheet, this, rotationAngle, adjustCellIndex);
            WorkbookCommands.SetShapeRotationAngle.Execute(context);
        }

        internal void SetRotationAngleInternal(double rotationAngle, bool adjustCellIndex)
        {
            if (this.shape.RotationAngle != rotationAngle)
            {
                this.Shape.RotationAngle = rotationAngle;

                this.OnShapeChanged();

                if (adjustCellIndex)
                {
                    this.AdjustCellIndexAndOffsetForRotation();
                }
                else
                {
                    this.cellIndexAdjustedForRotation = this.DoesRotationAngleRequireCellIndexChange();
                }
            }
        }

        internal abstract FloatingShapeBase Copy(Worksheet worksheet, CellIndex cellIndex, double offsetX, double offsetY);

        /// <summary>
        /// When the rotation angle of a shape is in the ranges 45 - 135 and 225 - 315, the cell index changes to the cell index where
        /// the top left point of the shape is located when it is rotated at 90 degrees. This method ensures the cell index
        /// is adjusted when the rotation is changed.
        /// </summary>
        private void AdjustCellIndexAndOffsetForRotation()
        {
            bool rotationAngleRequiresCellIndexChange = this.DoesRotationAngleRequireCellIndexChange();

            if ((rotationAngleRequiresCellIndexChange && !this.cellIndexAdjustedForRotation) ||
                (!rotationAngleRequiresCellIndexChange && this.cellIndexAdjustedForRotation))
            {
                RadWorksheetLayout worksheetLayout = this.Worksheet.Workbook.GetWorksheetLayout(this.worksheet, false);
                Point topleftPoint = worksheetLayout.GetPointFromCellIndexAndOffset(this.CellIndex, this.OffsetX, this.OffsetY);

                double width = this.shape.Width;
                double height = this.shape.Height;

                Point rotatedPoint;

                if (rotationAngleRequiresCellIndexChange)
                {
                    rotatedPoint = ShapesResizeHelper.AdjustTopLeftForRotation(topleftPoint, width, height);
                    this.cellIndexAdjustedForRotation = true;
                }
                else
                {
                    rotatedPoint = ShapesResizeHelper.ReverseAdjustmentForRotation(topleftPoint, width, height);
                    this.cellIndexAdjustedForRotation = false;
                }

                rotatedPoint.X = Math.Max(rotatedPoint.X, 0);
                rotatedPoint.Y = Math.Max(rotatedPoint.Y, 0);

                CellIndex cellIndex;
                double offsetX;
                double offsetY;

                cellIndex = worksheetLayout.GetCellIndexAndOffsetFromPoint(rotatedPoint, out offsetX, out offsetY);

                this.CellIndexInternal = cellIndex;
                this.OffsetXInternal = offsetX;
                this.OffsetYInternal = offsetY;
            }
        }

        /// <summary>
        /// Adjusts the CellIndex and the offset when the shape is rotated and then resized
        /// in order for the visual top left corner to remain in place. Validates the size in
        /// order for the image not to go out of the worksheet.
        /// </summary>
        private void AdjustCellIndexAndOffsetForSize(RadWorksheetLayout worksheetLayout, double oldWidth, double oldHeight, double newWidth, double newHeight, Point oldTopleftPoint)
        {
            Rect oldRectangle = new Rect(oldTopleftPoint.X, oldTopleftPoint.Y, oldWidth, oldHeight);
            Point newTopLeft = ShapesResizeHelper.DetermineNewTopLeftWhenResized(oldRectangle, this.RotationAngle, newWidth, newHeight);

            if (this.cellIndexAdjustedForRotation)
            {
                newTopLeft = ShapesResizeHelper.AdjustTopLeftForRotation(newTopLeft, newWidth, newHeight);
            }

            newTopLeft = ShapesResizeHelper.AdjustBoundingRectangleWhenOutOfSpreadsheet(worksheetLayout, new Rect(newTopLeft, new Size(newWidth, newHeight)), newTopLeft);
            newTopLeft = ShapesResizeHelper.AdjustPointWhenOutOfSpreadsheet(worksheetLayout, newTopLeft);

            CellIndex cellIndex;
            double offsetX;
            double offsetY;

            cellIndex = worksheetLayout.GetCellIndexAndOffsetFromPoint(newTopLeft, out offsetX, out offsetY);

            this.CellIndexInternal = cellIndex;
            this.OffsetXInternal = offsetX;
            this.OffsetYInternal = offsetY;
        }

        internal bool DoesRotationAngleRequireCellIndexChange()
        {
#if !SILVERLIGHT
            if (this.FloatingShapeType == FloatingShapeType.Chart)
            {
                return false;
            }
#endif

            double rotationAngle = SpreadsheetHelper.RestrictRotationAngle(this.RotationAngle);
            return (rotationAngle >= 45 && rotationAngle < 135) || (rotationAngle >= 225 && rotationAngle < 315);
        }

        /// <summary>
        /// Occurs when the value of any of the shape properties is changed.
        /// </summary>
        public event EventHandler ShapeChanged;

        /// <summary>
        /// Called when the value of any of the shape properties is changed.
        /// </summary>
        protected void OnShapeChanged()
        {
            if (this.ShapeChanged != null)
            {
                this.ShapeChanged(this, EventArgs.Empty);
            }
        }
    }
}
