﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.Copying;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Layout;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
#if !SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Model.Drawing.Charts;
using Telerik.Windows.Documents.Spreadsheet.Model.Charts;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.Model.Shapes
{
    /// <summary>
    /// Represents the collection containing the shapes included in the sheet.
    /// </summary>
    public class ShapeCollection : IEnumerable<FloatingShapeBase>
    {
        #region Fields

        private const string ImagesName = "Picture {0}";
        private const string ChartsName = "Chart {0}";

        /// <summary>
        /// Represents the offset between the pictures when they are inserted through the InsertPicture command.
        /// </summary>
        public static readonly double InsertDistance = 16;

        private int shapeIdCounter = 0;

        private readonly List<FloatingShapeBase> innerList;
        private readonly Worksheet worksheet;

        private readonly BeginEndCounter collectionChangeCounter;
        private readonly BeginEndCounter shapeChangeCounter;
        private bool usedCellRangeInvalidated;

        private CellRange usedCellRange;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the count of the shapes.
        /// </summary>
        /// <value>The count of the shapes.</value>
        public int Count
        {
            get
            {
                return this.innerList.Count;
            }
        }

        /// <summary>
        /// Gets all the images contained in the collection.
        /// </summary>
        public IEnumerable<FloatingImage> Images
        {
            get
            {
                foreach (FloatingShapeBase s in this.innerList)
                {
                    if (s.FloatingShapeType == FloatingShapeType.Image)
                    {
                        yield return s as FloatingImage;
                    }
                }
            }
        }

#if !SILVERLIGHT
        /// <summary>
        /// Gets all the chart shapes in teh collection.
        /// </summary>
        public IEnumerable<FloatingChartShape> Charts
        {
            get
            {
                foreach (FloatingShapeBase s in this.innerList)
                {
                    if (s.FloatingShapeType == FloatingShapeType.Chart)
                    {
                        yield return s as FloatingChartShape;
                    }
                }
            }
        }
#endif

        internal BeginEndCounter ShapeChangeCounter
        {
            get
            {
                return this.shapeChangeCounter;
            }
        }

        internal CellRange UsedCellRange
        {
            get
            {
                if (this.usedCellRangeInvalidated)
                {
                    this.RecalculateUsedCellRange();
                    this.usedCellRangeInvalidated = false;
                }

                return this.usedCellRange;
            }
        }

        /// <summary>
        /// Gets the worksheet.
        /// </summary>
        /// <value>The worksheet.</value>
        internal Worksheet Worksheet
        {
            get
            {
                return this.worksheet;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="ShapeCollection" /> class.
        /// </summary>
        /// <param name="worksheet">The worksheet.</param>
        public ShapeCollection(Worksheet worksheet)
        {
            this.innerList = new List<FloatingShapeBase>();
            this.worksheet = worksheet;
            this.usedCellRangeInvalidated = true;
            this.collectionChangeCounter = new BeginEndCounter(this.OnChanged);
            this.shapeChangeCounter = new BeginEndCounter(this.OnShapeChanged);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Adds the specified shape to the collection.
        /// </summary>
        /// <param name="shape">The shape.</param>
        public void Add(FloatingShapeBase shape)
        {
            Guard.ThrowExceptionIfNull(shape, "shape");

            if (shape.Worksheet != this.Worksheet)
            {
                throw new InvalidOperationException("The worksheet of the shape and of the collection it belongs to should be the same.");
            }

            AddRemoveShapeCommandContext addContext = new AddRemoveShapeCommandContext(this.Worksheet, shape);
            this.Worksheet.ExecuteCommand(WorkbookCommands.AddShape, addContext);
        }

        /// <summary>
        /// Adds the specified shapes to the collection.
        /// </summary>
        /// <param name="shapes">The shapes.</param>
        public void Add(IEnumerable<FloatingShapeBase> shapes)
        {
            Guard.ThrowExceptionIfNull(shapes, "shapes");
            Guard.ThrowExceptionIfContainsNull(shapes, "shapes");

            if (shapes.Any(shape => shape.Worksheet != this.Worksheet))
            {
                throw new InvalidOperationException("The worksheet of the shape and of the collection it belongs to should be the same.");
            }

            UpdateScope updateScope = new UpdateScope(
                () =>
                {
                    this.collectionChangeCounter.BeginUpdate();
                    this.Worksheet.BeginUndoGroup();
                },
                () =>
                {
                    this.Worksheet.EndUndoGroup();
                    this.collectionChangeCounter.EndUpdate();
                });

            using (updateScope)
            {
                foreach (FloatingShapeBase shape in shapes)
                {
                    AddRemoveShapeCommandContext addContext = new AddRemoveShapeCommandContext(this.Worksheet, shape);
                    this.Worksheet.ExecuteCommand(WorkbookCommands.AddShape, addContext);
                }
            }
        }

#if !SILVERLIGHT
        /// <summary>
        /// Creates and adds a chart to the collection.
        /// </summary>
        /// <param name="cellIndex">The cell index where the top left corner of the shape is positioned.</param>
        /// <param name="chartDataRange">The range containing the data that the chart will be based on.</param>
        /// <param name="chartTypes">The types of chart that will be created. Passing more than one type will create a combo chart.</param>
        /// <returns>The result chart shape.</returns>
        public FloatingChartShape AddChart(CellIndex cellIndex, CellRange chartDataRange, params ChartType[] chartTypes)
        {
            return this.AddChart(cellIndex, chartDataRange, SeriesRangesOrientation.Automatic, chartTypes);
        }

        /// <summary>
        /// Creates and adds a chart to the collection.
        /// </summary>
        /// <param name="cellIndex">The cell index where the top left corner of the shape is positioned.</param>
        /// <param name="chartDataRange">The range containing the data that the chart will be based on.</param>
        /// <param name="seriesRangesOrientation">A value indicating whether the series of the chart will refer to vertical or horizontal ranges or the direction will be determined automatically.</param>
        /// <param name="chartTypes">The types of chart that will be created. Passing more than one type will create a combo chart.</param>
        /// <returns>The result chart shape.</returns>
        public FloatingChartShape AddChart(CellIndex cellIndex, CellRange chartDataRange, SeriesRangesOrientation seriesRangesOrientation, params ChartType[] chartTypes)
        {
            FloatingChartShape chartShape = new FloatingChartShape(this.Worksheet, cellIndex, chartDataRange, seriesRangesOrientation, chartTypes);
            chartShape.Width = SpreadsheetDefaultValues.ChartDefaultSize.Width;
            chartShape.Height = SpreadsheetDefaultValues.ChartDefaultSize.Height;

            this.Add(chartShape);

            return chartShape;
        }
#endif

        /// <summary>
        /// Removes the specified shape from the collection.
        /// </summary>
        /// <param name="shape">The shape.</param>
        /// <returns></returns>
        public bool Remove(FloatingShapeBase shape)
        {
            Guard.ThrowExceptionIfNull(shape, "shape");

            AddRemoveShapeCommandContext removeContext = new AddRemoveShapeCommandContext(this.Worksheet, shape);
            bool hasRemoved = this.Worksheet.ExecuteCommand(WorkbookCommands.RemoveShape, removeContext);
            return hasRemoved;
        }

        /// <summary>
        /// Removes the specified shapes from the collection.
        /// </summary>
        /// <param name="shapes">The shapes.</param>
        public void Remove(IEnumerable<FloatingShapeBase> shapes)
        {
            Guard.ThrowExceptionIfNull(shapes, "shapes");
            Guard.ThrowExceptionIfContainsNull(shapes, "shapes");

            UpdateScope updateScope = new UpdateScope(
                () =>
                {
                    this.collectionChangeCounter.BeginUpdate();
                    this.Worksheet.BeginUndoGroup();
                },
                () =>
                {
                    this.Worksheet.EndUndoGroup();
                    this.collectionChangeCounter.EndUpdate();
                });

            using (updateScope)
            {
                foreach (FloatingShapeBase shape in shapes)
                {
                    AddRemoveShapeCommandContext removeContext = new AddRemoveShapeCommandContext(this.Worksheet, shape);
                    this.Worksheet.ExecuteCommand(WorkbookCommands.RemoveShape, removeContext);
                }
            }
        }

        /// <summary>
        /// Removes all elements from the collection.
        /// </summary>
        public void Clear()
        {
            UpdateScope updateScope = new UpdateScope(
                () =>
                {
                    this.collectionChangeCounter.BeginUpdate();
                    this.Worksheet.BeginUndoGroup();
                },
                () =>
                {
                    this.Worksheet.EndUndoGroup();
                    this.collectionChangeCounter.EndUpdate();
                });

            using (updateScope)
            {
                List<FloatingShapeBase> innerListCopy = new List<FloatingShapeBase>(this.innerList);
                foreach (FloatingShapeBase shape in innerListCopy)
                {
                    AddRemoveShapeCommandContext removeContext = new AddRemoveShapeCommandContext(this.Worksheet, shape);
                    this.Worksheet.ExecuteCommand(WorkbookCommands.RemoveShape, removeContext);
                }
            }
        }

        /// <summary>
        /// Gets or sets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.Shapes.FloatingShapeBase" /> at the specified index.
        /// </summary>
        /// <value>The <see cref="Telerik.Windows.Documents.Spreadsheet.Model.Shapes.FloatingShapeBase" /> at the specified index.</value>
        public FloatingShapeBase this[int index]
        {
            get
            {
                return this.innerList[index];
            }
            set
            {
                Guard.ThrowExceptionIfOutOfRange(0, this.innerList.Count - 1, index, "index");

                FloatingShapeBase existingShape = this.innerList[index];

                UpdateScope updateScope = new UpdateScope(
                    () =>
                    {
                        this.collectionChangeCounter.BeginUpdate();
                        this.Worksheet.BeginUndoGroup();
                    },
                    () =>
                    {
                        this.Worksheet.EndUndoGroup();
                        this.collectionChangeCounter.EndUpdate();
                    }
                );

                using (updateScope)
                {
                    this.Remove(existingShape);
                    this.Add(value);
                    this.MoveShapesToIndices(new Dictionary<FloatingShapeBase, int>()
                    {
                        { value, index }
                    });
                }
            }
        }

        internal void AddInternal(FloatingShapeBase shape)
        {
            Guard.ThrowExceptionIfNull(shape, "shape");

            this.OnChanging();

            this.SetUniqueIdAndNameToShape(shape);
            this.innerList.Add(shape);
            shape.ShapeChanged += this.Shape_ShapeChanged;

            this.DoOnChanged();
        }

        internal void RemoveInternal(FloatingShapeBase shape)
        {
            Guard.ThrowExceptionIfNull(shape, "shape");

            this.OnChanging();

            this.innerList.Remove(shape);
            shape.ShapeChanged -= this.Shape_ShapeChanged;

            this.DoOnChanged();
        }

        internal void BringForward(IEnumerable<FloatingShapeBase> shapes)
        {
            MoveShapeInDepthCommandContext context = new MoveShapeInDepthCommandContext(this.Worksheet, shapes);
            this.Worksheet.ExecuteCommand(WorkbookCommands.BringShapeForward, context);
        }

        internal void BringForwardInternal(IEnumerable<FloatingShapeBase> shapes)
        {
            Guard.ThrowExceptionIfNull(shapes, "shapes");

            this.OnChanging();

            var orderedShapes = shapes.OrderByDescending((x) => this.GetZIndex(x));

            foreach (FloatingShapeBase shape in orderedShapes)
            {
                int shapeIndex = this.innerList.IndexOf(shape);
                if (shapeIndex == this.innerList.Count - 1)
                {
                    continue;
                }

                this.innerList.Remove(shape);
                this.innerList.Insert(shapeIndex + 1, shape);
            }

            this.DoOnChanged();
        }

        internal void BringToFront(IEnumerable<FloatingShapeBase> shapes)
        {
            MoveShapeInDepthCommandContext context = new MoveShapeInDepthCommandContext(this.Worksheet, shapes);
            this.Worksheet.ExecuteCommand(WorkbookCommands.BringShapeToFront, context);
        }

        internal void BringToFrontInternal(IEnumerable<FloatingShapeBase> shapes)
        {
            Guard.ThrowExceptionIfNull(shapes, "shapes");

            this.OnChanging();

            var orderedShapes = shapes.OrderBy(x => this.GetZIndex(x));

            foreach (FloatingShapeBase shape in orderedShapes)
            {
                int shapeIndex = this.innerList.IndexOf(shape);
                if (shapeIndex == this.innerList.Count - 1)
                {
                    return;
                }

                this.innerList.Remove(shape);
                this.innerList.Insert(this.innerList.Count, shape);
            }

            this.DoOnChanged();
        }

        internal void SendBackward(IEnumerable<FloatingShapeBase> shapes)
        {
            MoveShapeInDepthCommandContext context = new MoveShapeInDepthCommandContext(this.Worksheet, shapes);
            this.Worksheet.ExecuteCommand(WorkbookCommands.SendShapeBackward, context);
        }

        internal void SendBackwardInternal(IEnumerable<FloatingShapeBase> shapes)
        {
            Guard.ThrowExceptionIfNull(shapes, "shapes");

            this.OnChanging();

            var orderedShapes = shapes.OrderBy((x) => this.GetZIndex(x));

            foreach (FloatingShapeBase shape in orderedShapes)
            {
                int shapeIndex = this.innerList.IndexOf(shape);
                if (shapeIndex == 0)
                {
                    continue;
                }

                this.innerList.Remove(shape);
                this.innerList.Insert(shapeIndex - 1, shape);
            }

            this.DoOnChanged();
        }

        internal void SendToBack(IEnumerable<FloatingShapeBase> shapes)
        {
            MoveShapeInDepthCommandContext context = new MoveShapeInDepthCommandContext(this.Worksheet, shapes);
            this.Worksheet.ExecuteCommand(WorkbookCommands.SendShapeToBack, context);
        }

        internal void SendToBackInternal(IEnumerable<FloatingShapeBase> shapes)
        {
            Guard.ThrowExceptionIfNull(shapes, "shapes");

            this.OnChanging();

            var orderedShapes = shapes.OrderByDescending(x => this.GetZIndex(x));

            foreach (FloatingShapeBase shape in orderedShapes)
            {
                int shapeIndex = this.innerList.IndexOf(shape);
                if (shapeIndex == 0)
                {
                    return;
                }

                this.innerList.Remove(shape);
                this.innerList.Insert(0, shape);
            }

            this.DoOnChanged();
        }

        internal void MoveShapesToIndices(Dictionary<FloatingShapeBase, int> shapeToIndex)
        {
            Guard.ThrowExceptionIfNull(shapeToIndex, "shapeToIndex");

            this.OnChanging();

            foreach (FloatingShapeBase shape in shapeToIndex.Keys)
            {
                if (!this.innerList.Contains(shape))
                {
                    throw new ArgumentException("One or more of the shapes is not present in the collection.");
                }

                this.innerList.Remove(shape);
            }

            var orderedPairs = shapeToIndex.OrderBy(x => x.Value);

            foreach (var pair in orderedPairs)
            {
                FloatingShapeBase shape = pair.Key;
                int index = pair.Value;
                this.innerList.Insert(index, shape);
            }

            this.DoOnChanged();
        }

        internal void Rotate(IEnumerable<FloatingShapeBase> shapes, double degree, bool adjustCellIndex)
        {
            UpdateScope updateScope = new UpdateScope(
                () =>
                {
                    this.worksheet.BeginUndoGroup();
                    this.shapeChangeCounter.BeginUpdate();
                },
                () =>
                {
                    this.shapeChangeCounter.EndUpdate();
                    this.worksheet.EndUndoGroup();
                });

            using (updateScope)
            {
                foreach (FloatingShapeBase shape in shapes)
                {
                    double newAngle = SpreadsheetHelper.RestrictRotationAngle(shape.RotationAngle + degree);
                    shape.SetRotationAngle(newAngle, adjustCellIndex);
                }
            }
        }

        internal void SetRotation(IEnumerable<FloatingShapeBase> shapes, double degree, bool adjustCellIndex)
        {
            UpdateScope updateScope = new UpdateScope(
                () =>
                {
                    this.worksheet.BeginUndoGroup();
                    this.shapeChangeCounter.BeginUpdate();
                },
                () =>
                {
                    this.shapeChangeCounter.EndUpdate();
                    this.worksheet.EndUndoGroup();
                });

            using (updateScope)
            {
                foreach (FloatingShapeBase shape in shapes)
                {
                    shape.SetRotationAngle(degree, adjustCellIndex);
                }
            }
        }

        internal void FlipHorizontal(IEnumerable<FloatingShapeBase> shapes)
        {
            UpdateScope updateScope = new UpdateScope(
                () =>
                {
                    this.worksheet.BeginUndoGroup();
                    this.shapeChangeCounter.BeginUpdate();
                },
                () =>
                {
                    this.shapeChangeCounter.EndUpdate();
                    this.worksheet.EndUndoGroup();
                });

            using (updateScope)
            {
                foreach (FloatingShapeBase shape in shapes)
                {
                    shape.IsHorizontallyFlipped = !shape.IsHorizontallyFlipped;
                }
            }
        }

        internal void FlipVertical(IEnumerable<FloatingShapeBase> shapes)
        {
            UpdateScope updateScope = new UpdateScope(
                () =>
                {
                    this.worksheet.BeginUndoGroup();
                    this.shapeChangeCounter.BeginUpdate();
                },
                () =>
                {
                    this.shapeChangeCounter.EndUpdate();
                    this.worksheet.EndUndoGroup();
                });

            using (updateScope)
            {
                foreach (FloatingShapeBase shape in shapes)
                {
                    shape.IsVerticallyFlipped = !shape.IsVerticallyFlipped;
                }
            }
        }


        internal void SetWidth(IEnumerable<FloatingShapeBase> shapes, double width, bool respectLockAspectRatio, bool adjustCellIndex)
        {
            UpdateScope updateScope = new UpdateScope(
                () =>
                {
                    this.worksheet.BeginUndoGroup();
                    this.shapeChangeCounter.BeginUpdate();
                },
                () =>
                {
                    this.shapeChangeCounter.EndUpdate();
                    this.worksheet.EndUndoGroup();
                });

            using (updateScope)
            {
                foreach (FloatingShapeBase shape in shapes)
                {
                    shape.SetWidth(respectLockAspectRatio, width, adjustCellIndex);
                }
            }
        }

        internal void SetHeight(IEnumerable<FloatingShapeBase> shapes, double height, bool respectLockAspectRatio, bool adjustCellIndex)
        {
            UpdateScope updateScope = new UpdateScope(
                () =>
                {
                    this.worksheet.BeginUndoGroup();
                    this.shapeChangeCounter.BeginUpdate();
                },
                () =>
                {
                    this.shapeChangeCounter.EndUpdate();
                    this.worksheet.EndUndoGroup();
                });

            using (updateScope)
            {
                foreach (FloatingShapeBase shape in shapes)
                {
                    shape.SetHeight(respectLockAspectRatio, height, adjustCellIndex);
                }
            }
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<FloatingShapeBase> GetEnumerator()
        {
            return ((IEnumerable<FloatingShapeBase>)this.innerList).GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.innerList.GetEnumerator();
        }

        /// <summary>
        /// Enumerates the collection in reversed order.
        /// </summary>
        /// <returns>A reversed enumeration of the elements in the collection.</returns>
        public IEnumerable<FloatingShapeBase> ReverseEnumerate()
        {
            for (int i = this.innerList.Count() - 1; i >= 0; i--)
            {
                FloatingShapeBase current = this.innerList.ElementAt(i);
                yield return current;
            }
        }

        /// <summary>
        /// Gets the ZIndex of the specified shape.
        /// </summary>
        /// <param name="shape"></param>
        /// <returns>The zero-based ZIndex. If the item is not found in the collection, -1 is returned.</returns>
        public int GetZIndex(FloatingShapeBase shape)
        {
            return this.innerList.IndexOf(shape);
        }

        /// <summary>
        /// Gets the ZIndex of the specified shape by its id.
        /// </summary>
        /// <param name="shapeId">The shape id.</param>
        /// <returns>The zero-based ZIndex. If the item is not found in the collection, -1 is returned.</returns>
        public int GetZIndexById(int shapeId)
        {
            FloatingShapeBase shape = this.innerList.FirstOrDefault(x => x.Id == shapeId);

            if (shape == null)
            {
                return -1;
            }

            return this.innerList.IndexOf(shape);
        }

        internal void CopyFrom(ShapeCollection fromShapeCollection, CopyContext context)
        {
            foreach (FloatingShapeBase floatingShape in fromShapeCollection.innerList)
            {
                context.SourceFloatingShape = floatingShape;

                this.innerList.Add(((ICopyable<FloatingShapeBase>)floatingShape).Copy(context));
            }

            context.SourceFloatingShape = null;
        }

        private void SetUniqueIdAndNameToShape(FloatingShapeBase shape)
        {
            shape.Shape.Id = this.shapeIdCounter;
            this.shapeIdCounter++;

            if (shape.Name == null)
            {
                switch (shape.FloatingShapeType)
                {
                    case FloatingShapeType.Image:
                        shape.Name = string.Format(ImagesName, shape.Id);
                        break;
#if !SILVERLIGHT
                    case FloatingShapeType.Chart:
                        shape.Name = string.Format(ChartsName, shape.Id);
                        break;
#endif
                }
            }
        }

        private void RecalculateUsedCellRange()
        {
            var topLeftIndex = this.GetTopLeftCellIndex();
            var bottomRightIndex = this.GetBottomRightCellIndex();

            this.usedCellRange = new CellRange(topLeftIndex, bottomRightIndex);
        }

        private CellIndex GetTopLeftCellIndex()
        {
            Point groupTopleft = this.GetTopLeftPoint();

            groupTopleft.X = Math.Max(groupTopleft.X, 0);
            groupTopleft.Y = Math.Max(groupTopleft.Y, 0);

            RadWorksheetLayout worksheetLayout = this.Worksheet.Workbook.GetWorksheetLayout(worksheet, false);
            return worksheetLayout.GetCellIndexFromPoint(groupTopleft.X, groupTopleft.Y, false, false);
        }

        private CellIndex GetBottomRightCellIndex()
        {
            RadWorksheetLayout worksheetLayout = this.Worksheet.Workbook.GetWorksheetLayout(worksheet, false);

            Point groupBottomRight = this.GetBottomRightPoint();

            groupBottomRight.X = Math.Min(groupBottomRight.X, worksheetLayout.Width);
            groupBottomRight.Y = Math.Min(groupBottomRight.Y, worksheetLayout.Height);

            return worksheetLayout.GetCellIndexFromPoint(groupBottomRight.X, groupBottomRight.Y, true, true);
        }

        private Point GetTopLeftPoint()
        {
            RadWorksheetLayout worksheetLayout = this.Worksheet.Workbook.GetWorksheetLayout(worksheet, false);

            double groupLeft = Double.MaxValue;
            double groupTop = Double.MaxValue;

            foreach (FloatingShapeBase shape in this.innerList)
            {
                Rect boundingRect = worksheetLayout.GetShapeLayoutBox(shape.Id).BoundingRectangle;
                double shapeLeft = boundingRect.Left;
                double shapeTop = boundingRect.Top;

                if (shapeLeft < groupLeft)
                {
                    groupLeft = shapeLeft;
                }

                if (shapeTop < groupTop)
                {
                    groupTop = shapeTop;
                }
            }

            if (groupLeft == Double.MaxValue)
            {
                groupLeft = 0;
            }

            if (groupTop == Double.MaxValue)
            {
                groupTop = 0;
            }

            return new Point(groupLeft, groupTop);
        }

        private Point GetBottomRightPoint()
        {
            RadWorksheetLayout worksheetLayout = this.Worksheet.Workbook.GetWorksheetLayout(worksheet, false);

            double groupRight = 0;
            double groupBottom = 0;

            foreach (FloatingShapeBase shape in this.innerList)
            {
                Rect boundingRect = worksheetLayout.GetShapeLayoutBox(shape.Id).BoundingRectangle;
                double shapeRight = boundingRect.Right;
                double shapeBottom = boundingRect.Bottom;

                if (shapeRight > groupRight)
                {
                    groupRight = shapeRight;
                }

                if (shapeBottom > groupBottom)
                {
                    groupBottom = shapeBottom;
                }
            }

            return new Point(groupRight, groupBottom);
        }

        private void DoOnChanged()
        {
            this.usedCellRangeInvalidated = true;

            if (!this.collectionChangeCounter.IsUpdateInProgress)
            {
                this.OnChanged();
            }
        }

        #endregion


        #region Event Handlers

        private void Shape_ShapeChanged(object sender, EventArgs e)
        {
            this.usedCellRangeInvalidated = true;

            if (!this.shapeChangeCounter.IsUpdateInProgress)
            {
                this.OnShapeChanged();
            }
        }

        #endregion


        #region Events

#if !SILVERLIGHT
        internal void OnWorksheetCellPropertyChanged(Worksheet source, CellPropertyChangedEventArgs e)
        {
            foreach (FloatingChartShape chartShape in this.Charts)
            {
                chartShape.OnWorksheetCellPropertyChanged(source, e);
            }
        }
#endif

        internal event EventHandler Changing;

        private void OnChanging()
        {
            if (this.Changing != null)
            {
                this.Changing(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Occurs when the collection has been changed.
        /// </summary>
        public event EventHandler Changed;

        /// <summary>
        /// Called when [changed].
        /// </summary>
        protected virtual void OnChanged()
        {
            if (this.Changed != null)
            {
                this.Changed(this, EventArgs.Empty);
            }
        }

        internal event EventHandler ShapeChanged;

        private void OnShapeChanged()
        {
            if (this.ShapeChanged != null)
            {
                this.ShapeChanged(this, EventArgs.Empty);
            }
        }

        #endregion
    }
}