﻿using System;
using Telerik.Windows.Documents.Spreadsheet.Commands;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Model.Printing;
using Telerik.Windows.Documents.Spreadsheet.Model.Protection;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a sheet.
    /// </summary>
    public abstract class Sheet : NotifyPropertyChangedBase, ISheet, IDisposable
    {
        #region Fields

        private string name;
        private Workbook workbook;
        private SheetVisibility visibility;
        //private int suspendLayoutUpdateCount;
        private bool hasLayoutInvalidation;
        private readonly ISheetViewState viewState;
        private bool alreadyDisposed;
        private readonly BeginEndCounter suspendLayoutUpdateCounter;
        private readonly ProtectionData protectionData;

        #endregion


        #region Properties

        /// <summary>
        /// Gets or sets the name of the sheet.
        /// </summary>
        /// <value>The name of the sheet.</value>
        public string Name
        {
            get
            {
                return this.name;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "value");

                if (this.name != value)
                {
                    if (this.Workbook != null)
                    {
                        if (!SheetCollection.IsValidSheetName(value))
                        {
                            throw new LocalizableException(SpreadsheetStrings.InvalidSheetNameMessage,
                                new InvalidOperationException(SpreadsheetStrings.InvalidSheetNameMessage),
                                "Spreadsheet_ErrorExpressions_InvalidSheetNameErrorMessage");
                        }
                        else if (!this.Workbook.Sheets.IsUniqueSheetNameInputed(value, this.name))
                        {
                            throw new LocalizableException(string.Format(SpreadsheetStrings.ExistingSheetName, this.name),
                                new InvalidOperationException(string.Format("Sheet with Name={0} already exist.", this.name)),
                                "Spreadsheet_ErrorExpressions_ExistingSheetName", new string[] { this.name });
                        }
                    }

                    this.name = value;

                    this.OnNameChanged();
                    this.OnPropertyChanged("Name");
                }
            }
        }


        /// <summary>
        /// Gets the workbook.
        /// </summary>
        /// <value>The workbook.</value>
        public Workbook Workbook
        {
            get
            {
                return this.workbook;
            }
            private set
            {
                if (this.workbook != value)
                {
                    if (value != null && this.workbook != null)
                    {
                        throw new LocalizableException(SpreadsheetStrings.SheetAlreadyAdded,
                            new InvalidOperationException(SpreadsheetStrings.SheetAlreadyAdded),
                            "Spreadsheet_ErrorExpressions_SheetAlreadyAdded");
                    }

                    this.workbook = value;
                    this.OnPropertyChanged("Workbook");
                }
            }
        }


        /// <summary>
        /// Gets the visibility.
        /// </summary>
        /// <value>The visibility.</value>
        public SheetVisibility Visibility
        {
            get
            {
                return this.visibility;
            }
            set
            {
                if (this.visibility != value)
                {
                    this.visibility = value;
                    this.OnPropertyChanged("Visibility");
                }
            }
        }


        /// <summary>
        /// Gets the type of the sheet.
        /// </summary>
        /// <value>The type of the sheet.</value>
        public abstract SheetType Type { get; }


        /// <summary>
        /// Gets the value indicating whether the layout update is suspended.
        /// </summary>
        /// <value>The value indicating whether the layout update is suspended.</value>
        public bool IsLayoutUpdateSuspended
        {
            get
            {
                return this.suspendLayoutUpdateCounter.IsUpdateInProgress;
            }
        }

        ISheetViewState ISheet.ViewState
        {
            get
            {
                if (this.viewState.IsInvalidated)
                {
                    this.viewState.IsInvalidated = false;

                    this.OnViewStateUpdateNeeded();
                }

                return this.viewState;
            }
        }

        internal ProtectionData ProtectionData
        {
            get
            {
                return this.protectionData;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the sheet is protected.
        /// </summary>
        /// <value>The is protected.</value>
        public bool IsProtected
        {
            get
            {
                return this.ProtectionData.Enforced;
            }
        }

        /// <summary>
        /// Gets the sheet page setup.
        /// </summary>
        /// <value>The sheet page setup.</value>
        protected abstract SheetPageSetupBase SheetPageSetup { get; }

        internal bool IsResumingLayoutUpdate { get; set; }

        #endregion


        #region Constructors

        internal Sheet(string name, Workbook workbook)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");
            Guard.ThrowExceptionIfNull(workbook, "workbook");

            this.Name = name;
            this.workbook = workbook;
            this.viewState = this.CreateViewState();
            this.suspendLayoutUpdateCounter = new BeginEndCounter();
            this.protectionData = new ProtectionData();
        }

        #endregion


        #region Methods

        /// <summary>
        /// Creates the state of the view.
        /// </summary>
        /// <returns></returns>
        protected abstract ISheetViewState CreateViewState();

        void ISheet.SuspendLayoutUpdate()
        {
            this.SuspendLayoutUpdate();
        }

        void ISheet.ResumeLayoutUpdate()
        {
            this.ResumeLayoutUpdate();
        }

        /// <summary>
        /// Suspends the layout update.
        /// </summary>
        protected virtual void SuspendLayoutUpdate()
        {
            this.suspendLayoutUpdateCounter.BeginUpdate();
        }

        /// <summary>
        /// Resumes the layout update.
        /// </summary>
        protected virtual void ResumeLayoutUpdate()
        {
            this.suspendLayoutUpdateCounter.EndUpdate();

            if (!this.IsLayoutUpdateSuspended)
            {
                if (this.hasLayoutInvalidation)
                {
                    this.RaiseOnLayoutInvalidated();
                    this.IsResumingLayoutUpdate = true;
                }

                this.OnLayoutUpdateResumed();
            }
        }

        internal virtual void OnLayoutUpdateResumed()
        {
        }

        /// <summary>
        /// Marks the beginning of an undo group.
        /// </summary>
        public void BeginUndoGroup()
        {
            this.Workbook.History.BeginUndoGroup();
        }

        /// <summary>
        /// Marks the end of the undo group.
        /// </summary>
        public void EndUndoGroup()
        {
            this.Workbook.History.EndUndoGroup();
        }

        internal void EnsureSheetNotProtected()
        {
            if (this.ProtectionData.Enforced)
            {
                throw new ProtectionException("Cannot protect a sheet that is already protected.",
                    "Spreadsheet_Protection_ProtectAProtectedSheet_Error");
            }
        }

        internal void EnsureSheetProtected()
        {
            if (!this.ProtectionData.Enforced)
            {
                throw new ProtectionException("Cannot unprotect a sheet that is not protected.",
                    "Spreadsheet_Protection_UnprotectAnUnprotectedSheet_Error");
            }
        }

        internal void ProtectSheet(string password)
        {
            this.ProtectionData.EnforceProtection(password);

            this.Workbook.History.Clear(this);

            this.OnIsProtectedChanged();
        }

        internal bool UnprotectSheet(string password)
        {
            this.EnsureSheetProtected();

            bool isSuccessful = this.ProtectionData.TryRemoveProtection(password);

            if (isSuccessful)
            {
                this.OnIsProtectedChanged();
            }

            return isSuccessful;
        }

        void ISheet.InvalidateLayout()
        {
            this.InvalidateLayout();
        }

        internal void InvalidateLayout()
        {
            this.hasLayoutInvalidation = true;
            if (this.IsLayoutUpdateSuspended)
            {
                return;
            }

            this.IsResumingLayoutUpdate = false;
            this.RaiseOnLayoutInvalidated();
        }

        internal bool ExecuteCommand<T>(WorkbookCommandBase<T> command, T context)
            where T : SheetCommandContextBase
        {
            return command.Execute(context);
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing,
        /// or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Disposes the specified disposing.
        /// </summary>
        /// <param name="disposing">The disposing.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (!alreadyDisposed)
            {
                if (disposing)
                {
                    this.workbook = null;
                }

                this.alreadyDisposed = true;
            }
        }

        #endregion


        #region Events

        private void RaiseOnLayoutInvalidated()
        {
            this.hasLayoutInvalidation = false;
            this.OnLayoutInvalidated();
        }

        /// <summary>
        /// Occurs when the layout is invalidated.
        /// </summary>
        public event EventHandler LayoutInvalidated;

        /// <summary>
        /// Called when [layout invalidated].
        /// </summary>
        protected virtual void OnLayoutInvalidated()
        {
            Guard.ThrowExceptionIfTrue(this.IsLayoutUpdateSuspended, "IsUpdateInProgress");

            if (this.LayoutInvalidated != null)
            {
                this.LayoutInvalidated(this, EventArgs.Empty);
            }
        }

        internal event EventHandler ViewStateUpdateNeeded;

        private void OnViewStateUpdateNeeded()
        {
            if (this.ViewStateUpdateNeeded != null)
            {
                this.ViewStateUpdateNeeded(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Occurs when name is changed.
        /// </summary>
        public event EventHandler NameChanged;

        private void OnNameChanged()
        {
            if (this.NameChanged != null)
            {
                this.NameChanged(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Occurs when the current protection state has changed.
        /// </summary>
        public event EventHandler IsProtectedChanged;

        /// <summary>
        /// Called when [is protected changed].
        /// </summary>
        protected virtual void OnIsProtectedChanged()
        {
            if (this.IsProtectedChanged != null)
            {
                this.IsProtectedChanged(this, EventArgs.Empty);
            }
        }

        #endregion
    }
}