﻿using System;
using System.Collections;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
	/// <summary>
	/// Collection of Sheet.
	/// </summary>
	public class SheetCollection : IEnumerable<Sheet>
	{
		#region Fields

		private readonly Workbook workbook;
		private readonly List<Sheet> innerList;
        private Sheet activeSheet;
        private int activeSheetIndex;
        private static readonly HashSet<char> forbiddenSheetNameSymbols = new HashSet<char>(new char[] { '\\', '/', '?', '*', '[', ']', ':' });
        private const int MaxSheetNameLength = 31;

		#endregion


		#region Properties

        /// <summary>
        /// Gets the workbook.
        /// </summary>
        /// <value>The workbook.</value>
        public Workbook Workbook
        {
			get
			{
				return this.workbook;
			}
		}
		
        /// <summary>
        /// Gets or sets the active sheet.
        /// </summary>
        /// <value>The active sheet.</value>
        public Sheet ActiveSheet
        {
			get
			{
                return this.activeSheet;
			}
            set 
            {
                int index = this.IndexOf(value);

                if (index == -1)
                {
                    throw new InvalidOperationException("The value must belong to the same workbook.");
                }

                this.UpdateActiveSheetAndIndex(value, index);
            }
		}

        /// <summary>
        /// Gets or sets the index of the active sheet.
        /// </summary>
        /// <value>The index of the active sheet.</value>
        public int ActiveSheetIndex
        {
            get
            {
                return this.activeSheetIndex;
            }
            set
            {
                Guard.ThrowExceptionIfOutOfRange(-1, this.Count - 1, value, "value");
                
                Sheet sheet = value != -1 ? this.innerList[value] : null;
                this.UpdateActiveSheetAndIndex(sheet, value);
            }
        }

        /// <summary>
        /// Gets the count of the sheets.
        /// </summary>
        /// <value>The count of the sheets.</value>
        public int Count
        {
			get
			{
				return this.innerList.Count;
			}
		}

		#endregion


		#region Indexers

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.Sheet" /> at the specified index.
        /// </summary>
        /// <value>The sheet.</value>
        public Sheet this[int index]
        {
			get
			{ 
				return this.innerList[index];
			}			
		}

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.Model.Sheet" /> with the specified sheet name.
        /// </summary>
        /// <value>The sheet.</value>
        public Sheet this[string sheetName]
        {
			get
			{
				return this.GetByName(sheetName);
			}
		}

		#endregion


		#region Constructors

		internal SheetCollection(Workbook workbook)
		{
			Guard.ThrowExceptionIfNull(workbook, "workbook");

            this.activeSheetIndex = -1;
			this.workbook = workbook;
			this.innerList = new List<Sheet>();
		}

		#endregion


		#region Helper Methods

		private string GetNextSheetName(SheetType type)
		{
			int count = 0;
			foreach (Sheet sheet in this)
			{
				if (sheet.Type == type)
				{
					count++;
				}
			}
			count++;

			string sheetNamePrefix = SpreadsheetHelper.GetSheetNamePrefixBySheetType(type);

			string sheetName = sheetNamePrefix + count;
			while (this.Contains(sheetName))
			{
				count++;
				sheetName = sheetNamePrefix + count;
			}

			return sheetName;
		}

		private Sheet CreateSheet(SheetType type)
		{
			string name = this.GetNextSheetName(type);

            return SheetFactory.Create(type, name, workbook);
		}

		#endregion


		#region Methods

        private void UpdateActiveSheetAndIndex(Sheet sheet, int index)
        {
            if (this.activeSheetIndex != index)
            {
                this.activeSheetIndex = index;
                this.OnActiveSheetIndexChanged();
            }

            if (this.activeSheet != sheet)
            {
                this.activeSheet = sheet;
                this.OnActiveSheetChanged();
            }
        }

        private void UpdateActiveSheetIndex()
        {
            this.UpdateActiveSheetAndIndex(this.ActiveSheet, this.IndexOf(this.ActiveSheet));
        }

        private void UpdateActiveSheet()
        {
            this.UpdateActiveSheetAndIndex(this[this.ActiveSheetIndex], this.ActiveSheetIndex);
        }

        /// <summary>
        /// Creates a new sheet of the specified type and adds it to the collection.
        /// </summary>
        /// <param name="type">The type of sheet.</param>
        /// <returns>The newly created sheet.</returns>
        public Sheet Add(SheetType type)
        {
			return this.Insert(this.Count, type);
		}

        /// <summary>
        /// Creates a new sheet of the specified type and inserts it in the collection at the active sheet index.
        /// </summary>
        /// <param name="type">The type.</param>
        /// <returns>The newly created sheet.</returns>
        public Sheet Insert(SheetType type)
        {
            return this.Insert(this.ActiveSheetIndex, type);
        }

        /// <summary>
        /// Creates a new sheet of the specified type and inserts it in the collection at specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <param name="type">The type.</param>
        /// <returns>The newly created sheet.</returns>
        public Sheet Insert(int index, SheetType type)
        {
			Sheet sheet = this.CreateSheet(type);

            if (this.Workbook.IsLayoutUpdateSuspended)
            {
                ISheet sheetToSuspend = sheet;
                sheetToSuspend.SuspendLayoutUpdate();
            }

            if (type == SheetType.Worksheet &&
                this.Workbook.IsPropertyChangeSuspended)
            {
                Worksheet wokrsheet = (Worksheet)sheet;
                wokrsheet.Cells.PropertyBag.SuspendPropertyChanged();
            }

            this.OnItemAdding(sheet);
            this.innerList.Insert(index, sheet);
            this.OnItemAdded(sheet);

            if (this.ActiveSheetIndex == -1)
            {
                this.ActiveSheetIndex = 0;
            }
            else if (this.workbook.ActiveTabIndex == index)
            {
                this.UpdateActiveSheetAndIndex(sheet, index);
            }

			return sheet;
		}

        /// <summary>
        /// Searches for the specified sheet and returns the zero-based index of the first occurrence within the SheetCollection. 
        /// </summary>
        /// <param name="sheet">The sheet.</param>
        /// <returns>The index of the sheet.</returns>
        public int IndexOf(Sheet sheet)
        {
			return this.innerList.IndexOf(sheet);
		}

        /// <summary>
        /// Searches for the specified sheet by name and returns the zero-based index of the first occurrence within the SheetCollection. The search is case insensitive.
        /// </summary>
        /// <param name="sheetName">The name of the sheet.</param>
        /// <returns>The index of the sheet.</returns>
        public int IndexOf(string sheetName)
        {
			for (int i = 0; i < this.innerList.Count; i++)
			{
				if (SheetCollection.AreEqualIgnoreCaseSheetNames(this.innerList[i].Name, sheetName))
				{
					return i;
				}
			}

			return -1;
		}

        /// <summary>
        /// Determines whether the collection contains a sheet with the specified sheet name.
        /// </summary>
        /// <param name="sheetName">The name of the sheet.</param>
        /// <returns>The value indicating whether a sheet with the specified name is contained in the collection.</returns>
        public bool Contains(string sheetName)
        {
			return this.IndexOf(sheetName) >= 0;
		}

        /// <summary>
        /// Determines whether the sheet collection contains the specified sheet.
        /// </summary>
        /// <param name="sheet">The sheet.</param>
        /// <returns>The value indicating whether the sheet is contained in the collection.</returns>
        public bool Contains(Sheet sheet)
        {
			return this.IndexOf(sheet) >= 0;
		}

        /// <summary>
        /// Gets a sheet by name.
        /// </summary>
        /// <param name="sheetName">The name of the sheet.</param>
        /// <returns>The sheet.</returns>
        public Sheet GetByName(string sheetName)
        {
			foreach (Sheet sheet in this)
			{
				if (SheetCollection.AreEqualIgnoreCaseSheetNames(sheet.Name, sheetName))
				{
					return sheet;
				}
			}

			return null;
		}

        /// <summary>
        /// Removes the active sheet.
        /// </summary>
        public void Remove()
        {
            this.RemoveAt(this.ActiveSheetIndex);
        }

        /// <summary>
        /// Removes the sheet with the specified name.
        /// </summary>
        /// <param name="sheetName">The name of the sheet.</param>
        /// <returns>The value indicating whether a sheet with a specified name was found.</returns>
        public bool Remove(string sheetName)
        {
            Sheet sheet = this.GetByName(sheetName);
            return this.Remove(sheet);
        }

        /// <summary>
        /// Removes the specified sheet.
        /// </summary>
        /// <param name="sheet">The sheet.</param>
        /// <returns>The value indicating whether the specified sheet was found.</returns>
        public bool Remove(Sheet sheet)
        {
            int sheetIndex = this.innerList.IndexOf(sheet);

            if (sheetIndex == -1)
            {
                return false;
            }

            this.RemoveAt(sheetIndex);
            return true;
		}

        /// <summary>
        /// Removes the sheet at the specified index.
        /// </summary>
        /// <param name="index">The index.</param>
        public void RemoveAt(int index)
        {
            Sheet sheet = this.innerList[index];
            this.OnItemRemoving(sheet);
            this.innerList.RemoveAt(index);

            if (this.ActiveSheetIndex == index)
            {
                if (this.ActiveSheetIndex == this.Count)
                {
                    this.ActiveSheetIndex--;
                }
                else
                {
                    this.UpdateActiveSheet();
                }
            }
            else
            {
                this.UpdateActiveSheetIndex();
            }

            this.OnItemRemoved(sheet);
        }

        /// <summary>
        /// Moves the specified number of items from the specified index to a specified index.
        /// </summary>
        /// <param name="fromIndex">From index.</param>
        /// <param name="itemCount">The item count.</param>
        /// <param name="toIndex">To index.</param>
        public void Move(int fromIndex, int itemCount, int toIndex)
        {
            if (fromIndex == toIndex)
            {
                return;
            }

            Stack<Sheet> removedItems = new Stack<Sheet>();
            for (int i = itemCount - 1; i >= 0; i--)
            {
                removedItems.Push(this.innerList[fromIndex + i]);
                this.innerList.RemoveAt(fromIndex + i);
            }

            if (toIndex > fromIndex)
            {
                toIndex -= itemCount;
            }

            this.innerList.InsertRange(toIndex, removedItems);

            this.OnChanged(SheetCollectionChangeType.Move);
            this.UpdateActiveSheetIndex();
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<Sheet> GetEnumerator()
        {
			return this.innerList.GetEnumerator();
		}

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
			return ((IEnumerable)this.innerList).GetEnumerator();
		}

		internal static bool IsValidSheetName(string sheetName)
		{
            bool isValid = !string.IsNullOrEmpty(sheetName);

            if (isValid)
            {
                isValid &= (sheetName[0] != '\'');
                isValid &= (sheetName[sheetName.Length - 1] != '\'');
                isValid &= (sheetName.Length <= MaxSheetNameLength);

                foreach (char symbol in sheetName)
                {
                    if (!isValid)
                    {
                        break;
                    }

                    isValid &= !forbiddenSheetNameSymbols.Contains(symbol);
                }
            }

            return isValid;
		}

        internal bool IsUniqueSheetName(string sheetName)
        {
            return !this.Contains(sheetName);
        }

        internal bool IsUniqueSheetNameInputed(string inputedName, string oldName)
        {
            return SheetCollection.AreEqualIgnoreCaseSheetNames(inputedName, oldName) || this.IsUniqueSheetName(inputedName);
        }

        internal static bool AreEqualIgnoreCaseSheetNames(string firstName, string secondName)
        {
            return firstName.Equals(secondName, StringComparison.CurrentCultureIgnoreCase);
        }

        private void OnItemAdding(Sheet sheet)
        {
            this.OnChanging(SheetCollectionChangeType.Add, sheet);
        }

		private void OnItemAdded(Sheet sheet)
		{
            this.OnChanged(SheetCollectionChangeType.Add, sheet);
		}

        private void OnItemRemoving(Sheet sheet)
        {
            this.OnChanging(SheetCollectionChangeType.Remove, sheet);
        }

		private void OnItemRemoved(Sheet sheet)
		{
			sheet.Dispose();
            this.OnChanged(SheetCollectionChangeType.Remove);
		}

		#endregion


        #region Events

        /// <summary>
        /// Occurs when the active sheet index has changed.
        /// </summary>
        public event EventHandler ActiveSheetIndexChanged;

        /// <summary>
        /// Called when [active sheet index changed].
        /// </summary>
        protected virtual void OnActiveSheetIndexChanged()
        {
            if (this.ActiveSheetIndexChanged != null)
            {
                this.ActiveSheetIndexChanged(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Occurs when the active sheet has changed.
        /// </summary>
        public event EventHandler ActiveSheetChanged;

        /// <summary>
        /// Called when [active sheet changed].
        /// </summary>
        protected virtual void OnActiveSheetChanged()
        {
            if (this.ActiveSheetChanged != null)
            {
                this.ActiveSheetChanged(this, EventArgs.Empty);
            }
        }

        private void OnChanging(SheetCollectionChangeType change, Sheet sheet)
        {
            this.OnChanging(new SheetCollectionChangedEventArgs(change, sheet));
        }

        /// <summary>
        /// Occurs when the collection has changed.
        /// </summary>
        internal event EventHandler<SheetCollectionChangedEventArgs> Changing;

        /// <summary>
        /// Raises the <see cref="E:Changing" /> event.
        /// </summary>
        /// <param name="args">The <see cref="Telerik.Windows.Documents.Spreadsheet.Model.SheetCollectionChangedEventArgs" /> instance containing the event data.</param>
        private void OnChanging(SheetCollectionChangedEventArgs args)
        {
            if (this.Changing != null)
            {
                this.Changing(this, args);
            }
        }

        private void OnChanged(SheetCollectionChangeType change, Sheet sheet = null)
        {
            this.OnChanged(new SheetCollectionChangedEventArgs(change, sheet));
        }

        /// <summary>
        /// Occurs when the collection has changed.
        /// </summary>
        public event EventHandler<SheetCollectionChangedEventArgs> Changed;

        /// <summary>
        /// Raises the <see cref="E:Changed" /> event.
        /// </summary>
        /// <param name="args">The <see cref="Telerik.Windows.Documents.Spreadsheet.Model.SheetCollectionChangedEventArgs" /> instance containing the event data.</param>
        protected virtual void OnChanged(SheetCollectionChangedEventArgs args)
        {
            if (this.Changed != null)
            {
                this.Changed(this, args);
            }
        }

        #endregion
    }
}