﻿using System;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents size.
    /// </summary>
    public struct SizeI
    {
        #region Constants

        private readonly static SizeI emptySizeI = SizeI.CreateEmptySize();

        #endregion


        #region Fields

        internal int width;
        internal int height;

        #endregion


        #region Properties

        /// <summary>
        /// Gets an empty size value.
        /// </summary>
        /// <value>The empty size value.</value>
        public static SizeI Empty
        {
            get
            {
                return SizeI.emptySizeI;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the size is empty.
        /// </summary>
        /// <value>The value indicating whether the size is empty.</value>
        public bool IsEmpty
        {
            get
            {
                return this.width < 0;
            }
        }

        /// <summary>
        /// Gets or sets the height.
        /// </summary>
        /// <value>The height.</value>
        public int Height
        {
            get
            {
                return this.height;
            }
            set
            {
                if (value < 0)
                {
                    throw new ArgumentException("Height must be greater or equal to 0");
                }

                this.height = value;
            }
        }

        /// <summary>
        /// Gets or sets the width.
        /// </summary>
        /// <value>The width.</value>
        public int Width
        {
            get
            {
                return this.width;
            }
            set
            {
                if (value < 0)
                {
                    throw new ArgumentException("Width must be greater or equal to 0");
                }

                this.width = value;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="SizeI" /> class.
        /// </summary>
        /// <param name="width">The width.</param>
        /// <param name="height">The height.</param>
        public SizeI(int width, int height)
        {
            if (width < 0 || height < 0)
            {
                throw new ArgumentException("Width and height must be greater or equal to 0");
            }

            this.width = width;
            this.height = height;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Creates an empty size value.
        /// </summary>
        /// <returns>The empty size value.</returns>
        private static SizeI CreateEmptySize()
        {
            SizeI size = new SizeI();
            size.width = int.MinValue;
            size.height = int.MinValue;

            return size;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// True if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            if (obj == null || !(obj is SizeI))
            {
                return false;
            }

            return SizeI.Equals(this, (SizeI)obj);
        }
        
        /// <summary>
        /// Determines whether the specified size is equal to this instance.
        /// </summary>
        /// <param name="value">The other size.</param>
        /// <returns>The value indicating whether the specified size is equal to this instance.</returns>
        public bool Equals(SizeI value)
        {
            return SizeI.Equals(this, value);
        }

        private static bool Equals(SizeI size1, SizeI size2)
        {
            if (size1.IsEmpty)
            {
                return size2.IsEmpty;
            }

            if (!size1.Width.Equals(size2.Width))
            {
                return false;
            }

            int height = size1.Height;

            return height.Equals(size2.Height);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            if (this.IsEmpty)
            {
                return 0;
            }

            int hashCode = this.Width.GetHashCode();

            return hashCode ^ this.Height.GetHashCode();
        }

        /// <summary>
        /// Implements the equals operator.
        /// </summary>
        /// <param name="size1">The size1 value.</param>
        /// <param name="size2">The size2 value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
        public static bool operator ==(SizeI size1, SizeI size2)
        {
            if (size1.Width != size2.Width)
            {
                return false;
            }

            return size1.Height == size2.Height;
        }

        /// <summary>
        /// Implements the not equals operator.
        /// </summary>
        /// <param name="size1">The size1 value.</param>
        /// <param name="size2">The size2 value.</param>
        /// <returns>
        /// The result of the operator.
        /// </returns>
        public static bool operator !=(SizeI size1, SizeI size2)
        {
            return !(size1 == size2);
        }

        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current
        /// <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
            if (this.IsEmpty)
            {
                return "Empty";
            }

            return string.Format("{0},{1}", this.width, this.height);
        }

        #endregion
    }
}
