﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Sorting
{
    /// <summary>
    /// Represents a sort condition for sorting values by custom list.
    /// </summary>
    public sealed class CustomValuesSortCondition : OrderedSortConditionBase<ICellValue>
    {
        #region Fields

        private readonly string[] customList;
        private readonly CustomValuesSortConditionComparer comparer;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the property definition of the property which is used for sorting.
        /// </summary>
        /// <value>
        /// The property definition of the property which is used for sorting.
        /// </value>
        protected override IPropertyDefinition<ICellValue> PropertyDefinition
        {
            get
            {
                return CellPropertyDefinitions.ValueProperty;
            }
        }

        /// <summary>
        /// Gets the comparer used for sorting.
        /// </summary>
        /// <value>The comparer used for sorting.</value>
        public override IComparer<SortValue> Comparer
        {
            get
            {
                return this.comparer;
            }
        }

        /// <summary>
        /// Gets the custom list.
        /// </summary>
        /// <value>The custom list.</value>
        public string[] CustomList
        {
            get
            {
                return this.customList;
            }
        }

        #endregion


        #region Constructros

        /// <summary>
        /// Initializes a new instance of the <see cref="CustomValuesSortCondition" /> class.
        /// </summary>
        /// <param name="relativeIndex">Gets the index of the row or column that is key for sorting relative to the sorted range.</param>
        /// <param name="customList">The custom list.</param>
        /// <param name="sortOrder">The sort order.</param>
        public CustomValuesSortCondition(int relativeIndex, string[] customList, SortOrder sortOrder) 
            : base(relativeIndex, sortOrder)
        {
            this.customList = customList;
            this.comparer = new CustomValuesSortConditionComparer(this.CustomList, sortOrder);
        }

        #endregion


        #region Methods

        internal override ISortCondition Copy(int newRelativeColumnIndex)
        {
            return new CustomValuesSortCondition(newRelativeColumnIndex, this.CustomList, this.SortOrder);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The object to compare with the current object.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            CustomValuesSortCondition other = obj as CustomValuesSortCondition;
            if (other == null)
            {
                return false;
            }

            if (this.customList.Length != other.customList.Length)
            {
                return false;
            }

            for (int i = 0; i < this.customList.Length; i++)
            {
                if (!TelerikHelper.EqualsOfT(this.customList[i], other.customList[i]))
                {
                    return false;
                }
            }

            return base.Equals(obj);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return 
                TelerikHelper.CombineHashCodes(
                base.GetHashCode(),
                this.customList.GetHashCodeOrZero());
        }

        #endregion
    }
}