﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Model.Sorting.Conditions;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Sorting
{
    /// <summary>
    /// Represents a sort condition for sorting by cell fill color.
    /// </summary>
    public sealed class FillColorSortCondition : OrderedSortConditionBase<IFill>, IWorksheetSortCondition
    {
        #region Fields

        private readonly IFill fill;
        private readonly SortOrder sortOrder;
        private IFill actualFill;
        private IComparer<SortValue> comparer;
        private Worksheet worksheet;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the fill.
        /// </summary>
        /// <value>The fill.</value>
        public IFill Fill
        {
            get
            {
                return this.fill;
            }
        }

        /// <summary>
        /// Gets the property definition of the property which is used for sorting.
        /// </summary>
        /// <value>
        /// The property definition of the property which is used for sorting.
        /// </value>
        protected override IPropertyDefinition<IFill> PropertyDefinition
        {
            get
            {
                return CellPropertyDefinitions.FillProperty;
            }
        }

        /// <summary>
        /// Gets the comparer used for sorting.
        /// </summary>
        /// <value>The comparer used for sorting.</value>
        public override IComparer<SortValue> Comparer
        {
            get 
            {
                if (this.comparer == null)
                {
                    this.comparer = new FillColorComparer(this.actualFill, this.sortOrder);
                }

                return this.comparer;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="FillColorSortCondition" /> class.
        /// </summary>
        /// <param name="relativeIndex">Gets the index of the row or column that is key for sorting relative to the sorted range.</param>
        /// <param name="fill">The fill.</param>
        /// <param name="sortOrder">The sort order.</param>
        public FillColorSortCondition(int relativeIndex, IFill fill, SortOrder sortOrder)
            : base(relativeIndex, sortOrder)
        {
            this.sortOrder = sortOrder;
            this.fill = fill;
        }

        #endregion


        #region Methods

        void IWorksheetSortCondition.SetWorksheet(Worksheet worksheet)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");

            this.worksheet = worksheet;
            this.actualFill = SortAndFilterHelper.GetIFillWithLocalColors(this.fill, this.worksheet.Workbook.Theme.ColorScheme);
        }

        /// <summary>
        /// Gets the value.
        /// </summary>
        /// <param name="cells">The cells.</param>
        /// <param name="rowIndex">The row index.</param>
        /// <param name="columnIndex">The column index.</param>
        /// <returns>The value of the specified cell.</returns>
        public override object GetValue(Cells cells, int rowIndex, int columnIndex)
        {
            IFill fill = (IFill)base.GetValue(cells, rowIndex, columnIndex);

            return SortAndFilterHelper.GetIFillWithLocalColors(fill, this.worksheet.Workbook.Theme.ColorScheme);
        }

        internal override ISortCondition Copy(int newRelativeColumnIndex)
        {
            return new FillColorSortCondition(newRelativeColumnIndex, this.Fill, this.SortOrder);
        }

        /// <summary>
        /// Equalses the specified obj.
        /// </summary>
        /// <param name="obj">The obj.</param>
        /// <returns></returns>
        public override bool Equals(object obj)
        {
            FillColorSortCondition other = obj as FillColorSortCondition;

            if (other == null)
            {
                return false;
            }

            return this.Fill.Equals(other.Fill) && base.Equals(obj);
        }

        /// <summary>
        /// Gets the hash code.
        /// </summary>
        /// <returns></returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                this.Fill.GetHashCode(),
                base.GetHashCode());
        }

        #endregion
    }
}