﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Model.Sorting.Conditions;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
#if NETCORE || NETSTANDARD
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.Model.Sorting
{
    /// <summary>
    /// Represents a sort condition for sorting by texts fore color.
    /// </summary>
    public sealed class ForeColorSortCondition : OrderedSortConditionBase<ThemableColor>, IWorksheetSortCondition
    {
        #region Fields

        private readonly ThemableColor color;
        private readonly SortOrder sortOrder;
        private IComparer<SortValue> comparer;
        private ThemableColor actualColor;
        private Worksheet worksheet;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the color.
        /// </summary>
        /// <value>The color.</value>
        public ThemableColor Color
        {
            get
            {
                return this.color;
            }
        }

        /// <summary>
        /// Gets the property definition of the property which is used for sorting.
        /// </summary>
        /// <value>
        /// The property definition of the property which is used for sorting.
        /// </value>
        protected override IPropertyDefinition<ThemableColor> PropertyDefinition
        {
            get
            {
                return CellPropertyDefinitions.ForeColorProperty;
            }
        }

        /// <summary>
        /// Gets the comparer used for sorting.
        /// </summary>
        /// <value>The comparer used for sorting.</value>
        public override IComparer<SortValue> Comparer
        {
            get
            {
                if (this.comparer == null)
                {
                    this.comparer = new ForeColorComparer(this.actualColor, this.sortOrder);
                }

                return this.comparer;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="ForeColorSortCondition" /> class.
        /// </summary>
        /// <param name="relativeIndex">Gets the index of the row or column that is key for sorting relative to the sorted range.</param>
        /// <param name="color">The color.</param>
        /// <param name="sortOrder">The sort order.</param>
        public ForeColorSortCondition(int relativeIndex, ThemableColor color, SortOrder sortOrder)
            : base(relativeIndex, sortOrder)
        {
            this.color = color;
            this.sortOrder = sortOrder;
        }

        #endregion


        #region Methods

        void IWorksheetSortCondition.SetWorksheet(Worksheet worksheet)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");

            this.worksheet = worksheet;
            this.actualColor = SortAndFilterHelper.GetThemableColorWithLocalColors(this.color, this.worksheet.Workbook.Theme.ColorScheme);
        }

        /// <summary>
        /// Gets the value.
        /// </summary>
        /// <param name="cells">The cells.</param>
        /// <param name="rowIndex">The row index.</param>
        /// <param name="columnIndex">The column index.</param>
        /// <returns>The value of the specified cell.</returns>
        public override object GetValue(Cells cells, int rowIndex, int columnIndex)
        {
            ThemableColor themableColor = (ThemableColor)base.GetValue(cells, rowIndex, columnIndex);

            return SortAndFilterHelper.GetThemableColorWithLocalColors(themableColor, this.worksheet.Workbook.Theme.ColorScheme);
        }

        internal override ISortCondition Copy(int newRelativeColumnIndex)
        {
            return new ForeColorSortCondition(newRelativeColumnIndex, this.Color, this.SortOrder);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The object to compare with the current object.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            ForeColorSortCondition other = obj as ForeColorSortCondition;

            if (other == null)
            {
                return false;
            }

            return this.Color.Equals(other.Color) && base.Equals(obj);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                this.Color.GetHashCode(),
                base.GetHashCode());
        }

        #endregion
    }
}