﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Copying;
using Telerik.Windows.Documents.Spreadsheet.Model.Filtering;
using Telerik.Windows.Documents.Spreadsheet.Model.Sorting.Conditions;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Sorting
{
    /// <summary>
    /// Represents a base class for sort conditions.
    /// </summary>
    /// <typeparam name="T">The type of the T.</typeparam>
    public abstract class SortConditionBase<T> : ISortCondition, ITranslatable, ICopyable<ISortCondition>
    {
        #region Fields

        private readonly int relativeIndex;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the index of the row or column that it key for sorting relative to the sorted range.
        /// </summary>
        /// <value>
        /// The index of the row or column that it key for sorting relative to the sorted range.
        /// </value>
        public int RelativeIndex
        {
            get
            {
                return this.relativeIndex;
            }
        }

        /// <summary>
        /// Gets the property definition of the property which is used for sorting.
        /// </summary>
        /// <value>The property definition of the property which is used for sorting.</value>
        protected abstract IPropertyDefinition<T> PropertyDefinition { get; }

        /// <summary>
        /// Gets the comparer used for sorting.
        /// </summary>
        /// <value>The comparer used for sorting.</value>
        public abstract IComparer<SortValue> Comparer { get; }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="SortConditionBase{T}" /> class.
        /// </summary>
        /// <param name="relativeIndex">Gets the index of the row or column that is key for sorting relative to the sorted range.</param>
        protected SortConditionBase(int relativeIndex)
        {
            this.relativeIndex = relativeIndex;
        }

        #endregion


        #region Methods

        /// <summary>
        /// Gets the value.
        /// </summary>
        /// <param name="cells">The cells.</param>
        /// <param name="rowIndex">The row index.</param>
        /// <param name="columnIndex">The column index.</param>
        /// <returns>The value of the specified cell.</returns>
        public virtual object GetValue(Cells cells, int rowIndex, int columnIndex)
        {
            long longIndex = CellsPropertyBag.ConvertCellIndexToLong(rowIndex, columnIndex);
            return cells.PropertyBag.GetPropertyValueRespectingStyle(this.PropertyDefinition, cells.Worksheet, longIndex);
        }

        object ITranslatable.Copy(int newRelativeColumnIndex)
        {
            return this.Copy(newRelativeColumnIndex);
        }

        internal abstract ISortCondition Copy(int newRelativeColumnIndex);

        ISortCondition ICopyable<ISortCondition>.Copy(CopyContext context)
        {
            ISortCondition copiedSortCondition = this.Copy(this.RelativeIndex);
            IWorksheetSortCondition copiedWorksheetSortCondition = copiedSortCondition as IWorksheetSortCondition;
            if (copiedWorksheetSortCondition != null)
            {
                copiedWorksheetSortCondition.SetWorksheet(context.TargetWorksheet);
            }

            return copiedSortCondition;
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The object to compare with the current object.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            SortConditionBase<T> other = obj as SortConditionBase<T>;

            if (other == null)
            {
                return false;
            }

            return this.RelativeIndex.Equals(other.RelativeIndex);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return this.RelativeIndex.GetHashCode();
        }

        #endregion
    }
}
