﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Model.Sorting.Conditions;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Sorting
{
    internal static class SortManager
    {
        public static void Sort(Cells cells, CellRange range, ISortCondition[] sortConditions)
        {
            using (new UpdateScope(cells.PropertyBag.BeginSort, cells.PropertyBag.EndSort))
            {
                range = range.Intersect(cells.PropertyBag.GetUsedCellRange());

                foreach (ISortCondition sortCondition in sortConditions)
                {
                    IWorksheetSortCondition worksheetSortCondition = sortCondition as IWorksheetSortCondition;
                    if (worksheetSortCondition != null)
                    {
                        worksheetSortCondition.SetWorksheet(cells.Worksheet);
                    }

                    ICompressedList<bool> hiddenRows = cells.Worksheet.Rows.PropertyBag.GetPropertyValueCollection(RowColumnPropertyBagBase.HiddenProperty);

                    int hiddenRowsCount = GetHiddenRowsCount(range, hiddenRows);

                    SortValue[] array = ExtractValuesToSort(range, hiddenRowsCount, sortCondition, hiddenRows, cells);

                    Array.Sort(array, sortCondition.Comparer);

                    int[] sortedIndexes = ExtractSortedIndexes(range, hiddenRows, array);

                    SetSortCommandContext context = new SetSortCommandContext(cells.Worksheet, range, sortCondition, sortedIndexes);
                    cells.Worksheet.ExecuteCommand(WorkbookCommands.SetSort, context);
                }
            }
        }
  
        private static SortValue[] ExtractValuesToSort(CellRange range, int hiddenRowsCount, ISortCondition sortCondition, ICompressedList<bool> hiddenRows, Cells cells)
        {
            SortValue[] array = new SortValue[range.RowCount - hiddenRowsCount];
            int absoluteSortIndex = range.FromIndex.ColumnIndex + sortCondition.RelativeIndex;

            // TODO: Add merged cells logic. It may use step for the indexes!
            int index = 0;
            for (int i = range.FromIndex.RowIndex; i <= range.ToIndex.RowIndex; i++)
            {
                if (!hiddenRows.GetValue(i))
                {
                    array[index] = new SortValue(i - range.FromIndex.RowIndex, sortCondition.GetValue(cells, i, absoluteSortIndex));
                    index++;
                }
            }

            return array;
        }
  
        private static int[] ExtractSortedIndexes(CellRange range, ICompressedList<bool> hiddenRows, SortValue[] array)
        {
            int[] sortedIndexes = new int[range.RowCount];
            int arrayIndex = 0;
            for (int i = range.FromIndex.RowIndex; i <= range.ToIndex.RowIndex; i++)
            {
                int sortedIndex = i - range.FromIndex.RowIndex;
                if (hiddenRows.GetValue(i))
                {
                    sortedIndexes[sortedIndex] = sortedIndex;
                }
                else
                {
                    sortedIndexes[sortedIndex] = array[arrayIndex].Index;
                    arrayIndex++;
                }
            }

            return sortedIndexes;
        }
  
        private static int GetHiddenRowsCount(CellRange range, ICompressedList<bool> hiddenRows)
        {
            long fromLongIndex = CellsPropertyBag.ConvertCellIndexToLong(range.FromIndex.RowIndex, 0);
            long toLongIndex = CellsPropertyBag.ConvertCellIndexToLong(range.ToIndex.RowIndex, 0);
            ICompressedList<bool> hiddenRowsForRange = hiddenRows.GetValue(fromLongIndex, toLongIndex);
            IEnumerable<Range<long, bool>> hiddenRanges = hiddenRowsForRange.GetNonDefaultRanges();

            int hiddenRowsCount = 0;
            foreach (Range<long, bool> hiddenRange in hiddenRanges)
            {
                hiddenRowsCount += (int)(hiddenRange.End - hiddenRange.Start) + 1;
            }

            return hiddenRowsCount;
        }

        public static bool AreSortConditionsDuplicating(ISortCondition[] sortConditions)
        {
            for (int i = 0; i < sortConditions.Length - 1; i++)
            {
                for (int j = i + 1; j < sortConditions.Length; j++)
                {
                    if (sortConditions[i].Equals(sortConditions[j]))
                    {
                        return true;
                    }
                }
            }

            return false;
        }
    }
}