﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.Copying;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model.Sorting
{
    /// <summary>
    /// Represents the sort state.
    /// </summary>
    public class SortState
    {
        #region Fields

        private readonly Worksheet worksheet;
        private readonly List<ISortCondition> innerList;
        private CellRange sortRange;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the count.
        /// </summary>
        /// <value>The count.</value>
        public int Count
        {
            get
            {
                return this.innerList.Count;
            }
        }

        /// <summary>
        /// Gets the sort conditions.
        /// </summary>
        /// <value>The sort conditions.</value>
        public IEnumerable<ISortCondition> SortConditions
        {
            get
            {
                return this.innerList;
            }
        }

        /// <summary>
        /// Gets the sort range.
        /// </summary>
        /// <value>The sort range.</value>
        public CellRange SortRange
        {
            get
            {
                return this.sortRange;
            }
            internal set
            {
                if (this.sortRange != value)
                {
                    SetSortRangeCommandContext context = new SetSortRangeCommandContext(this.worksheet, value);
                    WorkbookCommands.SetSortRange.Execute(context);
                }
            }
        }

        #endregion


        #region Constructors

        internal SortState(Worksheet worksheet)
        {
            this.worksheet = worksheet;
            this.innerList = new List<ISortCondition>();
        }

        #endregion


        #region Methods

        /// <summary>
        /// Sets the sort conditions for the specified sort range.
        /// </summary>
        /// <param name="sortRange">The sort range.</param>
        /// <param name="sortConditions">The sort conditions.</param>
        public void Set(CellRange sortRange, params ISortCondition[] sortConditions)
        {
            Guard.ThrowExceptionIfNull(sortRange, "sortRange");

            this.worksheet.Cells[sortRange].Sort(sortConditions);
        }

        /// <summary>
        /// Clears the sort range and sort conditions.
        /// </summary>
        public void Clear()
        {
            ClearSortCommandContext context = new ClearSortCommandContext(this.worksheet);
            this.worksheet.ExecuteCommand(WorkbookCommands.ClearSort, context);
        }

        internal void CopyFrom(SortState fromSortState, CopyContext context)
        {
            List<ISortCondition> sortConditionsCopied = new List<ISortCondition>();
            foreach (ISortCondition sortCondition in fromSortState.SortConditions)
            {
                ISortCondition sortConditionCopied = ((ICopyable<ISortCondition>)sortCondition).Copy(context);
                sortConditionsCopied.Add(sortConditionCopied);
            }

            this.SetInternal(fromSortState.SortRange, sortConditionsCopied.ToArray());
        }

        internal void SetInternal(CellRange sortRange, params ISortCondition[] sortConditions)
        {
            this.sortRange = sortRange;
            this.innerList.AddRange(sortConditions);
        }

        internal void Remove(ISortCondition sortCondition)
        {
            this.innerList.Remove(sortCondition);
        }

        internal ISortCondition this[int index]
        {
            get
            {
                return this.innerList[index];
            }
            set
            {
                this.innerList[index] = value;
            }
        }

        internal void Insert(int index, ISortCondition condition)
        {
            this.innerList.Insert(index, condition);
        }

        internal void AddRange(IEnumerable<ISortCondition> sortConditions)
        {
            this.innerList.AddRange(sortConditions);
        }

        internal void AddFirstInternal(ISortCondition sortCondition)
        {
            this.innerList.Insert(0, sortCondition);
        }

        internal void RemoveFirst()
        {
            if (innerList.Count > 0)
            {
                this.innerList.RemoveAt(0);
            }
        }

        internal void ClearInternal()
        {
            this.SetSortRangeInternal(null);
            this.innerList.Clear();
        }

        internal void SetSortRangeInternal(CellRange range)
        {
            if (this.sortRange != range)
            {
                this.sortRange = range;
            }
        }

        #endregion


        #region Update on insert / remove

        internal void Update(CellRangeInsertedOrRemovedEventArgs eventArgs)
        {
            ShiftType shiftType = eventArgs.RangeType.ToShiftType(eventArgs.IsRemove);
            CellRange insertRemoveRange = eventArgs.Range;

            if (this.SortRange == null)
            {
                return;
            }

            using (new UpdateScope(this.worksheet.BeginUndoGroup, this.worksheet.EndUndoGroup))
            {
                bool translateConditions = this.Count != 0 &&
                                           insertRemoveRange.FromIndex.ColumnIndex > this.SortRange.FromIndex.ColumnIndex &&
                                           insertRemoveRange.FromIndex.ColumnIndex <= this.SortRange.ToIndex.ColumnIndex &&
                                           (shiftType == ShiftType.Left || shiftType == ShiftType.Right) &&
                                           this.SortRange.CanInsertOrRemove(shiftType, insertRemoveRange);

                this.TranslateRange(eventArgs.Range, shiftType);

                if (translateConditions)
                {
                    this.TranslateConditions(insertRemoveRange, shiftType);
                }
            }
        }

        private void TranslateRange(CellRange insertRemoveRange, ShiftType shiftType)
        {
            CellRange resultRange;
            bool success = this.SortRange.TryTranslate(insertRemoveRange, shiftType, out resultRange);

            if (success)
            {
                if (resultRange != null)
                {
                    this.SortRange = resultRange;
                }
                else
                {
                    this.Clear();
                }
            }
        }

        private void TranslateConditions(CellRange insertRemoveRange, ShiftType shiftType)
        {
            List<ISortCondition> copyConditions = new List<ISortCondition>(this.SortConditions);
            int relativeRearrangeIndex = InsertRemoveCellsHelper.GetRelativeRearrangeIndex(this.sortRange, insertRemoveRange);
            int delta = InsertRemoveCellsHelper.GetRearrangeDelta(insertRemoveRange, shiftType);

            RemoveShiftType removeType;
            bool shiftTypeIsRemove = shiftType.TryGetRemoveShiftType(out removeType);

            foreach (var condition in copyConditions)
            {
                int orderIndex = this.innerList.IndexOf(condition);

                if (shiftTypeIsRemove && this.ShouldRemoveCondition(condition.RelativeIndex, insertRemoveRange))
                {
                    RemoveSortConditionCommandContext removeContext = new RemoveSortConditionCommandContext(this.worksheet, condition, orderIndex);
                    WorkbookCommands.RemoveSortCondition.Execute(removeContext);
                }
                else if (condition.RelativeIndex >= relativeRearrangeIndex)
                {
                    int newIndex = condition.RelativeIndex;
                    newIndex = condition.RelativeIndex + delta;

                    MoveSortConditionCommandContext moveContext = new MoveSortConditionCommandContext(this.worksheet, condition, newIndex, orderIndex);
                    WorkbookCommands.MoveSortCondition.Execute(moveContext);
                }
            }
        }

        private bool ShouldRemoveCondition(int conditionRelativeIndex, CellRange removedRange)
        {
            int conditionAbsoluteIndex = this.sortRange.FromIndex.ColumnIndex + conditionRelativeIndex;
            return conditionAbsoluteIndex >= removedRange.FromIndex.ColumnIndex && conditionAbsoluteIndex <= removedRange.ToIndex.ColumnIndex;
        }

        #endregion

    }
}