﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Common.Model.Data;
using Telerik.Windows.Documents.Spreadsheet.Commands;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorkbookCommands;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Expressions.Cache;
using Telerik.Windows.Documents.Spreadsheet.History;
using Telerik.Windows.Documents.Spreadsheet.Layout;
using Telerik.Windows.Documents.Spreadsheet.Model.Protection;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Theming;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// A collection of sheets.
    /// </summary>
    public class Workbook : NotifyPropertyChangedBase, IDisposable
    {
        #region Constants

        private static readonly int defaultContentChangedInterval = 30;

        #endregion


        #region Fields

        private readonly NameManager nameManager;
        private readonly SheetCollection sheets;
        private readonly WorksheetCollection worksheets;
        private readonly CellStyleCollection styles;
        private readonly WorkbookHistory history;
        private DocumentTheme theme;
        private readonly BeginEndCounter suspendLayoutUpdateCount;
        private readonly BeginEndCounter suspendPropertyChangedCount;
        private readonly RadWorksheetLayout worksheetLayout;
        private Worksheet lastActiveWorksheet;
        private bool isMergedCellsChanging;
        private bool isHyperlinksChanging;
        private TimeSpan workbookContentChangedInterval;
        private DelayedExecution contentChangedExecutor;
        private bool alreadyDisposed;
        private string name;
        private WorkbookProtectionOptions protectionOptions;
        private readonly ProtectionData protectionData;
        private int activeTabIndex;
        private bool containsProtectedSheets;
        private readonly CellReferenceRangeExpressionCache cellReferenceRangeExpressionCache;
        private readonly ResourceManager resources;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the name of the workbook.
        /// </summary>
        /// <value>The name of the workbook.</value>
        public string Name
        {
            get
            {
                return this.name;
            }
            set
            {
                if (this.name != value)
                {
                    this.name = value;
                    this.UpdateWorkbookNameDependentValuesAndNames();

                    this.OnNameChanged();
                    this.OnPropertyChanged("Name");
                }
            }
        }

        /// <summary>
        /// Gets or sets the index of the active tab.
        /// </summary>
        /// <value>The index of the active tab.</value>
        public int ActiveTabIndex
        {
            get
            {
                return this.activeTabIndex;
            }
            set
            {
                this.activeTabIndex = value;
            }
        }

        internal NameManager NameManager
        {
            get
            {
                return this.nameManager;
            }
        }

        /// <summary>
        /// Gets the collection of defined names of the workbook.
        /// </summary>
        /// <value>The collection of defined names of the workbook.</value>
        public NameCollection Names
        {
            get
            {
                return this.nameManager.WorkbookNameCollection;
            }
        }

        /// <summary>
        /// Gets the sheets of the workbook.
        /// </summary>
        /// <value>The sheets of the workbook.</value>
        public SheetCollection Sheets
        {
            get
            {
                return this.sheets;
            }
        }

        /// <summary>
        /// Gets the worksheets of the workbook.
        /// </summary>
        /// <value>The worksheets of the workbook.</value>
        public WorksheetCollection Worksheets
        {
            get
            {
                return this.worksheets;
            }
        }

        /// <summary>
        /// Gets or sets the active sheet.
        /// </summary>
        /// <value>The active sheet.</value>
        public Sheet ActiveSheet
        {
            get
            {
                return this.Sheets.ActiveSheet;
            }
            set
            {
                this.Sheets.ActiveSheet = value;
            }
        }

        /// <summary>
        /// Gets or sets the active worksheet.
        /// </summary>
        /// <value>The active worksheet.</value>
        public Worksheet ActiveWorksheet
        {
            get
            {
                return this.ActiveSheet as Worksheet;
            }
            set
            {
                this.ActiveSheet = value;
            }
        }

        /// <summary>
        /// Gets the collection of styles for this workbook.
        /// </summary>
        /// <value>The collection of styles for this workbook.</value>
        public CellStyleCollection Styles
        {
            get
            {
                return this.styles;
            }
        }

        /// <summary>
        /// Gets the history for this workbook.
        /// </summary>
        /// <value>The history for this workbook.</value>
        public WorkbookHistory History
        {
            get
            {
                return this.history;
            }
        }

        /// <summary>
        /// Gets or sets the workbook theme.
        /// </summary>
        /// <value>The workbook theme.</value>
        public DocumentTheme Theme
        {
            get
            {
                return this.theme;
            }
            set
            {
                if (!TelerikHelper.EqualsOfT(this.theme, value))
                {
                    this.ExecuteCommand(WorkbookCommands.SetWorkbookTheme, new SetWorkbookPropertyCommandContext<DocumentTheme>(this, value));
                }
            }
        }

        /// <summary>
        /// Gets a value indicating whether the layout update is suspended.
        /// </summary>
        /// <value>The value indicating whether the layout update is suspended.</value>
        public bool IsLayoutUpdateSuspended
        {
            get
            {
                return this.suspendLayoutUpdateCount.BeginUpdateCounter > 0;
            }
        }

        internal bool IsPropertyChangeSuspended
        {
            get
            {
                return this.suspendPropertyChangedCount.IsUpdateInProgress;
            }
        }

        /// <summary>
        /// Gets or sets the interval on which the workbook content changes.
        /// </summary>
        /// <value>The interval on which the workbook content changes.</value>
        public TimeSpan WorkbookContentChangedInterval
        {
            get
            {
                return this.workbookContentChangedInterval;
            }
            set
            {
                this.workbookContentChangedInterval = value;
                if (this.contentChangedExecutor != null)
                {
                    this.contentChangedExecutor.WaitInterval = this.workbookContentChangedInterval;
                }
            }
        }

        internal ProtectionData ProtectionData
        {
            get
            {
                return this.protectionData;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the workbook is protected.
        /// </summary>
        /// <value>The is protected.</value>
        public bool IsProtected
        {
            get
            {
                return this.ProtectionData.Enforced;
            }
        }

        internal WorkbookProtectionOptions ProtectionOptions
        {
            get
            {
                return this.protectionOptions;
            }
            set
            {
                this.protectionOptions = value;
            }
        }

        internal bool ContainsProtectedSheets
        {
            get
            {
                return this.containsProtectedSheets;
            }
            private set
            {
                if (this.containsProtectedSheets != value)
                {
                    this.containsProtectedSheets = value;
                    this.OnPropertyChanged("ContainsProtectedSheets");
                    this.OnContainsProtectedSheetsChanged();
                }
            }
        }

        internal ResourceManager Resources
        {
            get
            {
                return this.resources;
            }
        }

        internal CellReferenceRangeExpressionCache CellReferenceRangeExpressionCache
        {
            get
            {
                return this.cellReferenceRangeExpressionCache;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="Workbook" /> class.
        /// </summary>
        public Workbook()
        {
            this.cellReferenceRangeExpressionCache = new CellReferenceRangeExpressionCache();
            this.suspendLayoutUpdateCount = new BeginEndCounter();
            this.suspendPropertyChangedCount = new BeginEndCounter();
            this.activeTabIndex = -1;

            this.nameManager = new NameManager(this);
            this.nameManager.Changing += this.NameManager_Changing;
            this.sheets = new SheetCollection(this);
            this.sheets.ActiveSheetChanged += this.Sheets_ActiveSheetChanged;
            this.sheets.Changing += this.Sheets_Changing;
            this.sheets.Changed += this.Sheets_Changed;
            this.worksheets = new WorksheetCollection(this.sheets);
            this.styles = new CellStyleCollection(this);
            this.history = new WorkbookHistory(this);
            this.protectionOptions = WorkbookProtectionOptions.Default;
            this.protectionData = new ProtectionData();
            this.name = "Book1";
            this.resources = new ResourceManager();

            this.theme = PredefinedThemeSchemes.DefaultTheme;

            this.worksheetLayout = new RadWorksheetLayout(new SizeI(SpreadsheetDefaultValues.ColumnCount, SpreadsheetDefaultValues.RowCount));

            this.workbookContentChangedInterval = TimeSpan.FromMilliseconds(defaultContentChangedInterval);

            this.ThemeChanged += this.Workbook_ThemeChanged;
            this.NameChanged += this.Workbook_NameChanged;
        }

        #endregion


        #region Event Handlers

        private void Sheets_ActiveSheetChanged(object sender, EventArgs e)
        {
            this.DetachFromActiveWorksheetEvents();
            this.lastActiveWorksheet = this.ActiveWorksheet;
            this.AttachToActiveWorksheetEvents();

            this.OnActiveSheetChanged();
        }

        private void Sheets_Changing(object sender, SheetCollectionChangedEventArgs e)
        {
            if (e.ChangeType == SheetCollectionChangeType.Remove)
            {
                e.Sheet.IsProtectedChanged -= this.Sheet_ProtectionStateChanged;
                e.Sheet.PropertyChanged -= this.Sheet_PropertyChanged;
            }
        }

        private void Sheets_Changed(object sender, SheetCollectionChangedEventArgs e)
        {
            if (e.ChangeType == SheetCollectionChangeType.Remove)
            {
                this.History.Clear();
            }

            if (e.ChangeType == SheetCollectionChangeType.Add)
            {
                e.Sheet.IsProtectedChanged += this.Sheet_ProtectionStateChanged;
                e.Sheet.PropertyChanged += this.Sheet_PropertyChanged;
            }

            this.CallOnWorkbookContentChanged();
        }

        private void Worksheet_NameChanged(object sender, EventArgs e)
        {
            this.CallOnWorkbookContentChanged();
        }

        private void Cells_CellPropertyChanged(object sender, CellPropertyChangedEventArgs e)
        {
            if (!this.isMergedCellsChanging &&
                !this.isHyperlinksChanging)
            {
                this.CallOnWorkbookContentChanged();
            }
#if !SILVERLIGHT
            this.OnWorksheetCellPropertyChanged(((Cells)sender).Worksheet, e);
#endif
        }

        private void Cells_MergedCellsChanging(object sender, EventArgs e)
        {
            this.isMergedCellsChanging = true;
        }

        private void Cells_MergedCellsChanged(object sender, MergedCellRangesChangedEventArgs e)
        {
            this.isMergedCellsChanging = false;
            this.CallOnWorkbookContentChanged();
        }

        private void Cells_CellRangeInsertedOrRemoved(object sender, CellRangeInsertedOrRemovedEventArgs e)
        {
            this.CallOnWorkbookContentChanged();
        }

        private void Rows_RowsHeightsChanged(object sender, RowColumnPropertyChangedEventArgs e)
        {
            this.CallOnWorkbookContentChanged();
        }

        private void Columns_ColumnsWidthChanged(object sender, RowColumnPropertyChangedEventArgs e)
        {
            this.CallOnWorkbookContentChanged();
        }

        private void Hyperlinks_Changing(object sender, EventArgs e)
        {
            this.isHyperlinksChanging = true;
        }

        private void Hyperlinks_Changed(object sender, EventArgs e)
        {
            this.isHyperlinksChanging = false;
            this.CallOnWorkbookContentChanged();
        }

        private void Filter_FilterRangeChanged(object sender, EventArgs e)
        {
            this.CallOnWorkbookContentChanged();
        }

        private void Filter_FiltersChanged(object sender, EventArgs e)
        {
            this.CallOnWorkbookContentChanged();
        }

        private void ShapesCollection_Changed(object sender, EventArgs e)
        {
            this.CallOnWorkbookContentChanged();
        }

        private void Shape_ShapeChanged(object sender, EventArgs e)
        {
            this.CallOnWorkbookContentChanged();
        }

        private void Workbook_ThemeChanged(object sender, EventArgs e)
        {
            this.CallOnWorkbookContentChanged();
        }

        private void Workbook_NameChanged(object sender, EventArgs e)
        {
            this.CallOnWorkbookContentChanged();
        }

        #endregion


        #region Methods

        internal RadWorksheetLayout GetWorksheetLayout(Worksheet worksheet, bool isForPrinting)
        {
            this.UpdateWorksheetLayout(this.worksheetLayout, worksheet, isForPrinting);

            return this.worksheetLayout;
        }

        private void UpdateWorksheetLayout(RadWorksheetLayout worksheetLayout, Worksheet worksheet, bool isForPrinting)
        {
            worksheetLayout.SetWorksheet(worksheet, isForPrinting);
            if (worksheet != null)
            {
                WorksheetViewState viewState = (WorksheetViewState)((ISheet)worksheet).ViewState;
                CellIndex frozenCellIndex = viewState.GetFrozenCellIndex();
                worksheetLayout.Measure(frozenCellIndex);
            }
        }

        internal void SetThemeInternal(DocumentTheme value)
        {
            this.EnsureNotDisposed();

            if (!TelerikHelper.EqualsOfT(this.theme, value))
            {
                this.theme = value;
                this.OnThemeChanged();
            }
        }

        /// <summary>
        /// Suspends the layout update.
        /// </summary>
        public void SuspendLayoutUpdate()
        {
            this.EnsureNotDisposed();

            if (this.suspendLayoutUpdateCount.BeginUpdateCounter == 0)
            {
                foreach (ISheet sheet in this.Sheets)
                {
                    sheet.SuspendLayoutUpdate();
                }
            }

            this.suspendLayoutUpdateCount.BeginUpdate();
        }

        /// <summary>
        /// Resumes the layout update.
        /// </summary>
        public void ResumeLayoutUpdate()
        {
            this.EnsureNotDisposed();

            this.suspendLayoutUpdateCount.EndUpdate();

            if (this.suspendLayoutUpdateCount.BeginUpdateCounter == 0)
            {
                foreach (ISheet sheet in this.Sheets)
                {
                    sheet.ResumeLayoutUpdate();
                }
            }
        }

        internal void SuspendPropertyChanged()
        {
            this.EnsureNotDisposed();

            if (this.suspendPropertyChangedCount.BeginUpdateCounter == 0)
            {
                foreach (Worksheet sheet in this.Worksheets)
                {
                    sheet.Cells.PropertyBag.SuspendPropertyChanged();
                }
            }

            this.suspendPropertyChangedCount.BeginUpdate();
        }

        internal void ResumePropertyChanged()
        {
            this.EnsureNotDisposed();

            this.suspendPropertyChangedCount.EndUpdate();

            if (this.suspendPropertyChangedCount.BeginUpdateCounter == 0)
            {
                foreach (Worksheet sheet in this.Worksheets)
                {
                    sheet.Cells.PropertyBag.ResumePropertyChanged();
                }
            }
        }

        internal void InvalidateLayout()
        {
            this.EnsureNotDisposed();

            foreach (ISheet sheet in this.Sheets)
            {
                sheet.InvalidateLayout();
            }
        }

        internal bool ExecuteCommand<T>(WorkbookCommandBase<T> command, T context)
            where T : WorkbookCommandContextBase
        {
            this.EnsureNotDisposed();

            return command.Execute(context);
        }

        /// <summary>
        /// Executes find action according to the specified find options.
        /// </summary>
        /// <param name="findOptions">The find options.</param>
        /// <returns>The find results.</returns>
        public FindResult Find(FindOptions findOptions)
        {
            this.EnsureNotDisposed();

            IEnumerable<FindResult> findResults = this.FindAll(findOptions);
            findResults = FindAndReplaceHelper.OrderResults(findOptions, findResults);

            return findResults.FirstOrDefault();
        }

        /// <summary>
        /// Executes find all action according to the specified find options.
        /// </summary>
        /// <param name="findOptions">The find options.</param>
        /// <returns>The find results.</returns>
        public IEnumerable<FindResult> FindAll(FindOptions findOptions)
        {
            this.EnsureNotDisposed();

            List<FindResult> result = new List<FindResult>();
            IEnumerable<CellRange> searchRanges = findOptions.SearchRanges;

            if (findOptions.FindWithin == FindWithin.Workbook)
            {
                foreach (Worksheet worksheet in this.Worksheets)
                {
                    findOptions.SearchRanges = (findOptions.FindWithin == FindWithin.Sheet && this.ActiveWorksheet == worksheet) ? searchRanges : null;
                    result.AddRange(worksheet.FindAll(findOptions));
                }
            }
            else
            {
                result.AddRange(this.ActiveWorksheet.FindAll(findOptions));
            }

            findOptions.SearchRanges = searchRanges;

            return result;
        }

        /// <summary>
        /// Executes replace action in the active worksheet according to the specified replace options.
        /// </summary>
        /// <param name="replaceOptions">The replace options.</param>
        /// <returns>The value indicating whether a value was found to replace.</returns>
        public bool Replace(ReplaceOptions replaceOptions)
        {
            this.EnsureNotDisposed();

            return this.ActiveWorksheet.Replace(replaceOptions);
        }

        /// <summary>
        /// Executes replace all action according to the specified replace options.
        /// </summary>
        /// <param name="replaceOptions">The replace options.</param>
        /// <returns>The number of replacements made.</returns>
        public int ReplaceAll(ReplaceOptions replaceOptions)
        {
            this.EnsureNotDisposed();

            int replacementsCount = 0;

            using (new UpdateScope(this.History.BeginUndoGroup, this.History.EndUndoGroup))
            {
                if (replaceOptions.FindWithin == FindWithin.Workbook)
                {
                    foreach (Worksheet worksheet in this.Worksheets)
                    {
                        replacementsCount += worksheet.ReplaceAll(replaceOptions);
                    }
                }
                else
                {
                    replacementsCount = this.ActiveWorksheet.ReplaceAll(replaceOptions);
                }
            }

            return replacementsCount;
        }

        /// <summary>
        /// Protects the workbook with the specified password.
        /// </summary>
        /// <param name="password">The password.</param>
        public void Protect(string password)
        {
            if (this.ProtectionData.Enforced)
            {
                throw new ProtectionException("Cannot protect workbook that is already protected",
                    new InvalidOperationException(),
                    "Spreadsheet_Protection_ProtectAProtectedWorkbook_Error");
            }

            this.ProtectionData.EnforceProtection(password);
            this.OnIsProtectedChanged();
        }

        /// <summary>
        /// Unprotects the workbook using the specified password.
        /// </summary>
        /// <param name="password">The password.</param>
        /// <returns></returns>
        public bool Unprotect(string password)
        {
            if (!this.protectionData.Enforced)
            {
                throw new ProtectionException("Cannot unprotect workbook that is not protected",
                    new InvalidOperationException(),
                    "Spreadsheet_Protection_UnprotectAnUnprotectedWorkbook_Error");
            }

            bool isSucccessful = this.ProtectionData.TryRemoveProtection(password);
            if (isSucccessful)
            {
                this.OnIsProtectedChanged();
            }

            return isSucccessful;
        }

        private void AttachToActiveWorksheetEvents()
        {
            if (this.lastActiveWorksheet != null)
            {
                Worksheet worksheet = this.lastActiveWorksheet;

                worksheet.NameChanged += this.Worksheet_NameChanged;
                worksheet.Cells.CellPropertyChanged += this.Cells_CellPropertyChanged;
                worksheet.Cells.MergedCellsChanging += this.Cells_MergedCellsChanging;
                worksheet.Cells.MergedCellsChanged += this.Cells_MergedCellsChanged;
                worksheet.Cells.CellRangeInsertedOrRemoved += this.Cells_CellRangeInsertedOrRemoved;
                worksheet.Columns.ColumnsWidthChanged += this.Columns_ColumnsWidthChanged;
                worksheet.Rows.RowsHeightsChanged += this.Rows_RowsHeightsChanged;
                worksheet.Hyperlinks.Changing += this.Hyperlinks_Changing;
                worksheet.Hyperlinks.Changed += this.Hyperlinks_Changed;
                worksheet.Filter.FilterRangeChanged += this.Filter_FilterRangeChanged;
                worksheet.Filter.FiltersChanged += this.Filter_FiltersChanged;
                worksheet.Shapes.Changed += this.ShapesCollection_Changed;
                worksheet.Shapes.ShapeChanged += this.Shape_ShapeChanged;
            }
        }

        private void DetachFromActiveWorksheetEvents()
        {
            if (this.lastActiveWorksheet != null)
            {
                Worksheet worksheet = this.lastActiveWorksheet;
                worksheet.NameChanged -= this.Worksheet_NameChanged;
                worksheet.Cells.CellPropertyChanged -= this.Cells_CellPropertyChanged;
                worksheet.Cells.MergedCellsChanging -= this.Cells_MergedCellsChanging;
                worksheet.Cells.MergedCellsChanged -= this.Cells_MergedCellsChanged;
                worksheet.Cells.CellRangeInsertedOrRemoved -= this.Cells_CellRangeInsertedOrRemoved;
                worksheet.Columns.ColumnsWidthChanged -= this.Columns_ColumnsWidthChanged;
                worksheet.Rows.RowsHeightsChanged -= this.Rows_RowsHeightsChanged;
                worksheet.Hyperlinks.Changing -= this.Hyperlinks_Changing;
                worksheet.Hyperlinks.Changed -= this.Hyperlinks_Changed;
                worksheet.Filter.FilterRangeChanged -= this.Filter_FilterRangeChanged;
                worksheet.Filter.FiltersChanged -= this.Filter_FiltersChanged;
                worksheet.Shapes.Changed -= this.ShapesCollection_Changed;
                worksheet.Shapes.ShapeChanged -= this.Shape_ShapeChanged;
            }
        }

        private void EnsureContentChangedExecutorInitialized()
        {
            if (this.contentChangedExecutor == null)
            {
                this.contentChangedExecutor = new DelayedExecution(() =>
                {
                    this.OnWorkbookContentChanged();
                },
                (long)this.WorkbookContentChangedInterval.TotalMilliseconds);
            }
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing,
        /// or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Disposes the specified clean up managed resources.
        /// </summary>
        /// <param name="cleanUpManagedResources">The clean up managed resources.</param>
        protected virtual void Dispose(bool cleanUpManagedResources)
        {
            if (this.alreadyDisposed)
            {
                return;
            }

            //Place here the code that cleans up native resource.
            if (cleanUpManagedResources)
            {
                if (this.contentChangedExecutor != null)
                {
                    this.contentChangedExecutor.Dispose();
                }
            }

            this.alreadyDisposed = true;
        }

        private void EnsureNotDisposed()
        {
            if (this.alreadyDisposed)
            {
                throw new ObjectDisposedException("Workbook");
            }
        }

        private void Sheet_ProtectionStateChanged(object sender, EventArgs e)
        {
            this.UpdateContainsProtectedSheets();
        }

        internal void UpdateContainsProtectedSheets()
        {
            bool contains = false;

            for (int i = 0; i < this.sheets.Count; i++)
            {
                if (this.sheets[i].IsProtected)
                {
                    contains = true;
                    break;
                }
            }

            this.ContainsProtectedSheets = contains;
        }

        private void UpdateWorkbookNameDependentValuesAndNames()
        {
            for (int i = 0; i < this.Worksheets.Count; i++)
            {
                this.Worksheets[i].Cells.UpdateWorkbookNameDependentCellValues(this);
                this.Worksheets[i].Names.UpdateWorkbookNameDependentNames(this);
            }

            this.Names.UpdateWorkbookNameDependentNames(this);
        }

        internal void UpdateWorksheetDependentValuesAndNames(Sheet renamedSheet)
        {
            Worksheet renamedWorksheet = (Worksheet)renamedSheet;
            for (int i = 0; i < this.Worksheets.Count; i++)
            {
                this.Worksheets[i].Cells.UpdateSheetNameDependentCellValues(renamedWorksheet);
                this.Worksheets[i].Names.UpdateSheetNameDependentNames(renamedWorksheet);
            }

            this.Names.UpdateSheetNameDependentNames(renamedWorksheet);
        }

        private void NameManager_Changing(object sender, NameCollectionChangingEventArgs args)
        {
            if (!string.IsNullOrEmpty(args.OldName) && !string.IsNullOrEmpty(args.NewName) && !args.OldName.Equals(args.NewName, StringComparison.CurrentCultureIgnoreCase))
            {
                for (int i = 0; i < this.Worksheets.Count; i++)
                {
                    this.Worksheets[i].Cells.UpdateSpreadsheetNameDependentCellValues(args.OldName, args.NewName);
                    this.Worksheets[i].Names.UpdateSpreadsheetNameDependentNames(args.OldName, args.NewName);
                }

                this.Names.UpdateSpreadsheetNameDependentNames(args.OldName, args.NewName);
            }
        }

        private void Sheet_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
        {
            if (e.PropertyName == "Name")
            {
                this.UpdateWorksheetDependentValuesAndNames((Sheet)sender);
            }
        }

        #endregion


        #region Events

        /// <summary>
        /// Occurs when a command is in a state of execution.
        /// </summary>
        public event EventHandler<CommandExecutingEventArgs> CommandExecuting;

        /// <summary>
        /// Raises the <see cref="E:CommandExecuting" /> event.
        /// </summary>
        /// <param name="args">The <see cref="Telerik.Windows.Documents.Spreadsheet.Commands.CommandExecutingEventArgs" /> instance containing the event data.</param>
        protected internal virtual void OnCommandExecuting(CommandExecutingEventArgs args)
        {
            EventHandler<CommandExecutingEventArgs> commandExecuting = this.CommandExecuting;

            if (commandExecuting != null)
            {
                foreach (EventHandler<CommandExecutingEventArgs> handler in commandExecuting.GetInvocationList())
                {
                    handler(this, args);
                    if (args.Canceled)
                    {
                        break;
                    }
                }
            }
        }

        /// <summary>
        /// Occurs when a command has executed.
        /// </summary>
        public event EventHandler<CommandExecutedEventArgs> CommandExecuted;

        /// <summary>
        /// Raises the <see cref="E:CommandExecuted" /> event.
        /// </summary>
        /// <param name="args">The <see cref="Telerik.Windows.Documents.Spreadsheet.Commands.CommandExecutedEventArgs" /> instance containing the event data.</param>
        protected internal virtual void OnCommandExecuted(CommandExecutedEventArgs args)
        {
            if (this.History.IsEnabled && !args.IsInUndo)
            {
                IUndoableWorkbookCommand undoableSheetCommand = args.Command as IUndoableWorkbookCommand;
                if (undoableSheetCommand != null)
                {
                    this.history.RecordSheetChange(new WorkbookChange(undoableSheetCommand, args.CommandContext));
                }
            }

            if (args.Command.AffectsLayout(args.CommandContext))
            {
                args.CommandContext.InvalidateLayout();
            }

            if (this.CommandExecuted != null)
            {
                this.CommandExecuted(this, args);
            }
        }

        /// <summary>
        /// Occurs in a command error.
        /// </summary>
        public event EventHandler<CommandErrorEventArgs> CommandError;

        /// <summary>
        /// Raises the <see cref="E:CommandError" /> event.
        /// </summary>
        /// <param name="args">The <see cref="Telerik.Windows.Documents.Spreadsheet.Commands.CommandErrorEventArgs" /> instance containing the event data.</param>
        protected internal virtual void OnCommandError(CommandErrorEventArgs args)
        {
            if (this.CommandError != null)
            {
                this.CommandError(this, args);
            }
        }

        /// <summary>
        /// Occurs when a theme is changed.
        /// </summary>
        public event EventHandler ThemeChanged;

        /// <summary>
        /// Called when [theme changed].
        /// </summary>
        protected virtual void OnThemeChanged()
        {
            if (this.ThemeChanged != null)
            {
                this.ThemeChanged(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Occurs when the active sheet is changed.
        /// </summary>
        public event EventHandler ActiveSheetChanged;

        /// <summary>
        /// Called when the active sheet is changed.
        /// </summary>
        protected virtual void OnActiveSheetChanged()
        {
            if (this.ActiveSheetChanged != null)
            {
                this.ActiveSheetChanged(this, EventArgs.Empty);
            }
        }

        private event EventHandler workbookContentChanged;

        /// <summary>
        /// Occurs when the workbook content is changed.
        /// </summary>
        public event EventHandler WorkbookContentChanged
        {
            add
            {
                this.EnsureContentChangedExecutorInitialized();
                this.workbookContentChanged += value;
            }
            remove
            {
                this.workbookContentChanged -= value;
            }
        }

        /// <summary>
        /// Called when [workbook content changed].
        /// </summary>
        protected virtual void OnWorkbookContentChanged()
        {
            if (this.workbookContentChanged != null)
            {
                this.workbookContentChanged(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Calls the on workbook content changed.
        /// </summary>
        protected void CallOnWorkbookContentChanged()
        {
            if (this.contentChangedExecutor != null)
            {
                this.contentChangedExecutor.Execute();
            }
        }

        /// <summary>
        /// Occurs when the current protection state has changed.
        /// </summary>
        public event EventHandler IsProtectedChanged;

        /// <summary>
        /// Called when [is protected changed].
        /// </summary>
        protected virtual void OnIsProtectedChanged()
        {
            if (this.IsProtectedChanged != null)
            {
                this.IsProtectedChanged(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Occurs when name is changed.
        /// </summary>
        public event EventHandler NameChanged;

        private void OnNameChanged()
        {
            if (this.NameChanged != null)
            {
                this.NameChanged(this, EventArgs.Empty);
            }
        }

        internal event EventHandler ContainsProtectedSheetsChanged;

        private void OnContainsProtectedSheetsChanged()
        {
            if (this.ContainsProtectedSheetsChanged != null)
            {
                this.ContainsProtectedSheetsChanged(this, EventArgs.Empty);
            }
        }

#if !SILVERLIGHT
        private void OnWorksheetCellPropertyChanged(Worksheet source, CellPropertyChangedEventArgs e)
        {
            foreach (Worksheet worksheet in this.Worksheets)
            {
                worksheet.Shapes.OnWorksheetCellPropertyChanged(source, e);
            }
        }
#endif

        #endregion
    }
}