﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using Telerik.Windows.Documents.Spreadsheet.Copying;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.History;
using Telerik.Windows.Documents.Spreadsheet.Model.Filtering;
using Telerik.Windows.Documents.Spreadsheet.Model.Printing;
using Telerik.Windows.Documents.Spreadsheet.Model.Protection;
using Telerik.Windows.Documents.Spreadsheet.Model.Shapes;
using Telerik.Windows.Documents.Spreadsheet.Model.Sorting;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// A two-dimensional grid of cells that are organized into rows and columns.
    /// </summary>
    public class Worksheet : Sheet
    {
        #region Fields

        private readonly Cells cells;
        private readonly Rows rows;
        private readonly Columns columns;

        private HyperlinkCollection hyperlinks;
        private NameCollection names;
        private readonly ShapeCollection shapes;
        private WorksheetProtectionOptions protectionOptions;
        private readonly AutoFilter filter;
        private readonly GroupingProperties groupingProperties;

        private bool alreadyDisposed;

        private readonly Dictionary<int, CellRange> expandToFitCache;
        private readonly WorksheetPageSetup worksheetPageSetup;
        private readonly SheetPageSetupBase sheetPageSetup;

        private readonly SortState sortState;
        private HeaderNameRenderingConverterBase headerNameRenderingConverter;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the cells of the worksheet.
        /// </summary>
        /// <value>The cells of the worksheet.</value>
        public Cells Cells
        {
            get
            {
                return this.cells;
            }
        }

        /// <summary>
        /// Gets the rows of the worksheet.
        /// </summary>
        /// <value>The rows of the worksheet.</value>
        public Rows Rows
        {
            get
            {
                return this.rows;
            }
        }

        /// <summary>
        /// Gets the columns of the worksheet.
        /// </summary>
        /// <value>The columns of the worksheet.</value>
        public Columns Columns
        {
            get
            {
                return this.columns;
            }
        }


        /// <summary>
        /// Gets the type of the sheet.
        /// </summary>
        /// <value>The type of the sheet.</value>
        public override SheetType Type
        {
            get
            {
                return SheetType.Worksheet;
            }
        }

        /// <summary>
        /// Gets or sets the default width of the column.
        /// </summary>
        /// <value>The default width of the column.</value>
        public ColumnWidth DefaultColumnWidth
        {
            get
            {
                return this.Columns.GetDefaultWidth();
            }
            set
            {
                this.Columns.SetDefaultWidth(value);
            }
        }

        /// <summary>
        /// Gets or sets the default height of the row.
        /// </summary>
        /// <value>The default height of the row.</value>
        public RowHeight DefaultRowHeight
        {
            get
            {
                return this.Rows.GetDefaultHeight();
            }
            set
            {
                this.Rows.SetDefaultHeight(value);
            }
        }

        /// <summary>
        /// Gets the used cell range.
        /// </summary>
        /// <value>The used cell range.</value>
        public CellRange UsedCellRange
        {
            get
            {
                return this.cells.PropertyBag.GetUsedCellRange();
            }
        }

        /// <summary>
        /// Gets the all hyperlinks in the worksheet.
        /// </summary>
        /// <value>The hyperlinks.</value>
        public HyperlinkCollection Hyperlinks
        {
            get
            {
                return this.hyperlinks;
            }
        }

        /// <summary>
        /// Gets the collection of defined names of the worksheet.
        /// </summary>
        /// <value>The collection of defined names of the worksheet.</value>
        public NameCollection Names
        {
            get
            {
                return this.names;
            }
        }

        /// <summary>
        /// Gets the collection of shapes inserted in the current worksheet.
        /// </summary>
        /// <value>The shapes.</value>
        public ShapeCollection Shapes
        {
            get
            {
                return this.shapes;
            }
        }

        /// <summary>
        /// Gets the worksheet page setup.
        /// </summary>
        /// <value>The worksheet page setup.</value>
        public WorksheetPageSetup WorksheetPageSetup
        {
            get
            {
                return this.worksheetPageSetup;
            }
        }

        /// <summary>
        /// Gets the sheet page setup.
        /// </summary>
        /// <value>The sheet page setup.</value>
        protected sealed override SheetPageSetupBase SheetPageSetup
        {
            get
            {
                return this.sheetPageSetup;
            }
        }

        /// <summary>
        /// Gets the sort state.
        /// </summary>
        /// <value>The sort state.</value>
        public SortState SortState
        {
            get
            {
                return this.sortState;
            }
        }

        /// <summary>
        /// Gets or sets the protection options.
        /// </summary>
        /// <value>The protection options.</value>
        public WorksheetProtectionOptions ProtectionOptions
        {
            get
            {
                return this.protectionOptions;
            }
            internal set
            {
                this.protectionOptions = value;
            }
        }

        /// <summary>
        /// Gets the filtering of the worksheet.
        /// </summary>
        /// <value>The filtering of the worksheet.</value>
        public AutoFilter Filter
        {
            get
            {
                return this.filter;
            }
        }

        /// <summary>
        /// Gets the view state of the worksheet.
        /// </summary>
        /// <value>The view state of the worksheet.</value>
        public WorksheetViewState ViewState
        {
            get
            {
                return (WorksheetViewState)((ISheet)this).ViewState;
            }
        }

        /// <summary>
        /// Gets the grouping properties for the worksheet.
        /// </summary>
        /// <value>The grouping properties.</value>
        public GroupingProperties GroupingProperties
        {
            get
            {
                return this.groupingProperties;
            }
        }

        internal CellRange ShapeUsedCellRange
        {
            get
            {
                return this.shapes.UsedCellRange;
            }
        }

        internal CellRange TotalUsedCellRange
        {
            get
            {
                CellRange usedCellRange = this.UsedCellRange;
                CellRange usedShapeRange = this.ShapeUsedCellRange;

                int fromRowIndex = Math.Min(usedCellRange.FromIndex.RowIndex, usedShapeRange.FromIndex.RowIndex);
                int fromColumnIndex = Math.Min(usedCellRange.FromIndex.ColumnIndex, usedShapeRange.FromIndex.ColumnIndex);

                int toRowIndex = Math.Max(usedCellRange.ToIndex.RowIndex, usedShapeRange.ToIndex.RowIndex);
                int toColumnIndex = Math.Max(usedCellRange.ToIndex.ColumnIndex, usedShapeRange.ToIndex.ColumnIndex);

                return new CellRange(fromRowIndex, fromColumnIndex, toRowIndex, toColumnIndex);
            }
        }

        /// <summary>
        /// Gets or sets the header name rendering converter.
        /// </summary>
        /// <value>The header name rendering converter.</value>
        public HeaderNameRenderingConverterBase HeaderNameRenderingConverter
        {
            get
            {
                if (this.headerNameRenderingConverter == null)
                {
                    this.headerNameRenderingConverter = new DefaultHeaderRenderingNameConverter();
                }

                return this.headerNameRenderingConverter;
            }
            set
            {
                if (this.headerNameRenderingConverter != value)
                {
                    this.headerNameRenderingConverter = value;
                    this.InvalidateLayout();
                }
            }
        }

        #endregion


        #region Constructors

        internal Worksheet(string name, Workbook workbook)
            : base(name, workbook)
        {
            this.cells = new Cells(this);
            this.rows = new Rows(this);
            this.columns = new Columns(this);
            this.hyperlinks = new HyperlinkCollection(this);
            this.names = this.Workbook.NameManager.CreateCollection(this);
            this.protectionOptions = WorksheetProtectionOptions.Default;
            this.filter = new AutoFilter(this);
            this.expandToFitCache = new Dictionary<int, CellRange>();
            this.worksheetPageSetup = new WorksheetPageSetup(this);
            this.sheetPageSetup = this.WorksheetPageSetup;
            this.shapes = new ShapeCollection(this);
            this.sortState = new SortState(this);
            this.groupingProperties = new GroupingProperties();

            this.Cells.PropertyBag.Sorted += this.PropertyBag_Sorted;
        }

        #endregion


        #region Methods

        private void PropertyBag_Sorted(object sender, SortedEventArgs e)
        {
            CellRange sortedRange = e.Range;
            int[] sortedIndexes = e.SortedIndexes;

            this.Hyperlinks.Sort(sortedRange, sortedIndexes);
            this.Rows.PropertyBag.Sort(sortedRange, sortedIndexes);
        }

        /// <summary>
        /// Creates the state of the view.
        /// </summary>
        /// <returns></returns>
        protected override ISheetViewState CreateViewState()
        {
            return new WorksheetViewState();
        }

        /// <summary>
        /// Executes find action according to the specified find options.
        /// </summary>
        /// <param name="findOptions">The find options.</param>
        /// <returns>The find results.</returns>
        public FindResult Find(FindOptions findOptions)
        {
            IEnumerable<FindResult> findResults = this.FindAll(findOptions);

            findResults = FindAndReplaceHelper.OrderResults(findOptions, findResults);

            return findResults.FirstOrDefault();
        }

        /// <summary>
        /// Executes find all action according to the specified find options.
        /// </summary>
        /// <param name="findOptions">The find options.</param>
        /// <returns>The find results.</returns>
        public IEnumerable<FindResult> FindAll(FindOptions findOptions)
        {
            if (string.IsNullOrEmpty(findOptions.FindWhat))
            {
                yield break;
            }

            if (findOptions.FindWithin == FindWithin.Workbook)
            {
                findOptions.SearchRanges = null;
            }

            ICompressedList<ICellValue> cellProperyValueCompressedList = this.Cells.GetPropertyValueRespectingStyle(CellPropertyDefinitions.ValueProperty, this.UsedCellRange);
            ICompressedList<bool> hiddenRows = this.Rows.PropertyBag.GetPropertyValueCollection(RowColumnPropertyBagBase.HiddenProperty);
            IEnumerable<Range<long, ICellValue>> nonDefaultRanges = cellProperyValueCompressedList.GetNonDefaultRanges();

            IEnumerable<Range<CellIndex, ICellValue>> singleCellNonDefaultRanges = this.SplitToSingleCellRanges(nonDefaultRanges);

            IEnumerable<Range<CellIndex, ICellValue>> findResults = singleCellNonDefaultRanges.Where(
                (range) =>
                {
                    bool isMatch = false;

                    CellIndex index = range.Start;

                    if ((findOptions.SearchRanges == null ||
                        findOptions.SearchRanges.Where(p => p.Contains(index)).Count() > 0) &&
                        !hiddenRows.GetValue(index.RowIndex))
                    {
                        switch (findOptions.FindIn)
                        {
                            case FindInContentType.Formulas:
                                isMatch = FindAndReplaceHelper.FindInRawValue(range, findOptions);
                                isMatch = isMatch || FindAndReplaceHelper.FindInResultValue(this.Cells, range, findOptions);
                                break;
                            case FindInContentType.Values:
                                isMatch = FindAndReplaceHelper.FindInResultValue(this.Cells, range, findOptions);
                                break;
                        }
                    }

                    return isMatch;
                });

            if (findOptions.FindBy == FindBy.Rows)
            {
                findResults = from p in findResults orderby p.Start.RowIndex, p.Start.ColumnIndex select p;
            }

            foreach (Range<CellIndex, ICellValue> range in findResults)
            {
                yield return new FindResult(
                    new WorksheetCellIndex(this, range.Start));
            }
        }

        private IEnumerable<Range<CellIndex, ICellValue>> SplitToSingleCellRanges(IEnumerable<Range<long, ICellValue>> nonDefaultRanges)
        {
            foreach (Range<long, ICellValue> range in nonDefaultRanges)
            {
                for (long i = range.Start; i <= range.End; i++)
                {
                    CellIndex index = CellsPropertyBag.ConvertLongToCellIndex(i);
                    yield return new Range<CellIndex, ICellValue>(index, index, range.IsDefault, range.Value);
                }
            }
        }

        /// <summary>
        /// Executes replace action according to the specified replace options.
        /// </summary>
        /// <param name="replaceOptions">The replace options.</param>
        /// <returns>The value indicating whether a value was found to replace.</returns>
        public bool Replace(ReplaceOptions replaceOptions)
        {
            bool currentCellContainsSearchedValue = this.CheckIfCurrentCellContainsSearchedValue(replaceOptions);

            if (!currentCellContainsSearchedValue)
            {
                return false;
            }

            CellSelection currentCell = this.Cells[replaceOptions.StartCell.CellIndex];
            string cellValue = currentCell.GetValue().Value.RawValue;
            string newValue = Regex.Replace(cellValue, replaceOptions.FindWhatRegex, m =>
                    {
                        if (String.IsNullOrEmpty(m.Value))
                        {
                            return string.Empty;
                        }

                        return replaceOptions.ReplaceWith;
                    },
                    replaceOptions.FindWhatRegexOptions);

            currentCell.SetValueIgnoreErrorsInternal(newValue, true);

            return true;
        }

        private bool CheckIfCurrentCellContainsSearchedValue(ReplaceOptions replaceOptions)
        {
            string value = this.Cells[replaceOptions.StartCell.CellIndex].GetValue().Value.RawValue;

            return FindAndReplaceHelper.FindInString(value, replaceOptions);
        }

        /// <summary>
        /// Executes replace all action according to the specified replace options.
        /// </summary>
        /// <param name="replaceOptions">The replace options.</param>
        /// <returns>The number of replacements made.</returns>
        public int ReplaceAll(ReplaceOptions replaceOptions)
        {
            int replacesCount = 0;

            using (new UpdateScope(this.Workbook.History.BeginUndoGroup, this.Workbook.History.EndUndoGroup))
            {
                List<FindResult> findResults = this.FindAll(replaceOptions).ToList();

                foreach (FindResult result in findResults)
                {
                    CellSelection currentCell = this.Cells[result.FoundCell.CellIndex];
                    string cellValue = currentCell.GetValue().Value.RawValue;
                    string newValue = Regex.Replace(cellValue, replaceOptions.FindWhatRegex, m =>
                    {
                        if (String.IsNullOrEmpty(m.Value))
                        {
                            return string.Empty;
                        }

                        replacesCount++;
                        return replaceOptions.ReplaceWith;
                    },
                        replaceOptions.FindWhatRegexOptions);
                    currentCell.SetValueIgnoreErrorsInternal(newValue, true);
                }
            }

            return replacesCount;
        }

        internal override void OnLayoutUpdateResumed()
        {
            this.ExpandCachedColumnWidths();
        }

        internal void ExpandCachedColumnWidths()
        {
            if (this.expandToFitCache.Count > 0)
            {
                CellRange[] values = this.expandToFitCache.Values.ToArray();
                this.expandToFitCache.Clear();

                List<CellRange> autofitCellRanges = new List<CellRange>();

                CellRange currentCellRange = values[0];
                for (int i = 1; i < values.Length; i++)
                {
                    int columnIndex = values[i].FromIndex.ColumnIndex;
                    int fromRowIndex = values[i].FromIndex.RowIndex;
                    int toRowIndex = values[i].ToIndex.RowIndex;

                    if (currentCellRange.FromIndex.ColumnIndex > columnIndex)
                    {
                        currentCellRange = new CellRange(fromRowIndex, columnIndex, toRowIndex, currentCellRange.ToIndex.ColumnIndex);
                    }
                    else if (currentCellRange.ToIndex.ColumnIndex < columnIndex)
                    {
                        currentCellRange = new CellRange(fromRowIndex, currentCellRange.FromIndex.ColumnIndex, toRowIndex, columnIndex);
                    }
                    else if (!currentCellRange.Contains(fromRowIndex, columnIndex) ||
                        !currentCellRange.Contains(toRowIndex, columnIndex))
                    {
                        autofitCellRanges.Add(currentCellRange);
                        i++;
                        if (i < values.Length)
                        {
                            columnIndex = values[i].FromIndex.ColumnIndex;
                            fromRowIndex = values[i].FromIndex.RowIndex;
                            toRowIndex = values[i].ToIndex.RowIndex;
                            currentCellRange = new CellRange(fromRowIndex, columnIndex, toRowIndex, columnIndex);
                        }
                    }
                }

                if (autofitCellRanges.Count == 0)
                {
                    autofitCellRanges.Add(currentCellRange);
                }

                foreach (CellRange range in autofitCellRanges)
                {
                    this.Columns[range].AutoFitWidth(true);
                }
            }
        }

        /// <summary>
        /// Caches the expand to fit ranges.
        /// </summary>
        /// <param name="cellRanges">The cell ranges.</param>
        internal void CacheExpandToFitColumns(IEnumerable<CellRange> cellRanges)
        {
            if (cellRanges != null)
            {
                foreach (CellRange range in cellRanges)
                {
                    for (int i = range.FromIndex.ColumnIndex; i <= range.ToIndex.ColumnIndex; i++)
                    {
                        if (!this.expandToFitCache.ContainsKey(i))
                        {
                            this.expandToFitCache.Add(i, new CellRange(range.FromIndex.RowIndex, i, range.ToIndex.RowIndex, i));
                        }
                        else
                        {
                            CellRange rangeToUpdate = this.expandToFitCache[i];
                            CellRange updatedRange = new CellRange(
                                Math.Min(rangeToUpdate.FromIndex.RowIndex, range.FromIndex.RowIndex), i,
                                Math.Max(rangeToUpdate.ToIndex.RowIndex, range.ToIndex.RowIndex), i);

                            this.expandToFitCache[i] = updatedRange;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Protects the worksheet.
        /// </summary>
        /// <param name="password">The password.</param>
        /// <param name="options">The protection options of the worksheet.</param>
        public void Protect(string password, WorksheetProtectionOptions options)
        {
            base.EnsureSheetNotProtected();

            this.protectionOptions = options ?? WorksheetProtectionOptions.Default;

            base.ProtectSheet(password);
        }

        /// <summary>
        /// Unprotects the worksheet.
        /// </summary>
        /// <param name="password">The password.</param>
        public bool Unprotect(string password)
        {
            return base.UnprotectSheet(password);
        }

        /// <summary>
        /// Copies <see cref="Worksheet"/> contents from a source <see cref="Worksheet"/> into this instance. 
        /// This <see cref="Worksheet"/> instance contents are overriden from the source <see cref="Worksheet"/> contents. 
        /// </summary>
        /// <param name="sourceWorksheet">The worksheet to copy from.</param>
        public void CopyFrom(Worksheet sourceWorksheet)
        {
            Guard.ThrowExceptionIfNull(sourceWorksheet, "sourceWorksheet");

            if (!this.Equals(sourceWorksheet))
            {
                CopyContext copyContext = new CopyContext(this, sourceWorksheet);

                this.Clear();

                if (sourceWorksheet.Workbook != this.Workbook)
                {
                    CellStyleCollectionsMerger stylesMerger = new CellStyleCollectionsMerger(this.Workbook.Styles, sourceWorksheet.Workbook.Styles, copyContext);
                    stylesMerger.Merge();

                    this.Workbook.Names.CopyFrom(sourceWorksheet.Workbook.Names, copyContext);
                }

                this.Names.CopyFrom(sourceWorksheet.Names, copyContext);

                this.Shapes.CopyFrom(sourceWorksheet.Shapes, copyContext);
                this.Hyperlinks.CopyFrom(sourceWorksheet.Hyperlinks);

                this.Cells.CopyFrom(sourceWorksheet.Cells, copyContext);
                this.Rows.CopyFrom(sourceWorksheet.Rows, copyContext);
                this.Columns.CopyFrom(sourceWorksheet.Columns, copyContext);

                this.ProtectionData.CopyFrom(sourceWorksheet.ProtectionData);
                this.ProtectionOptions = sourceWorksheet.ProtectionOptions.Clone();

                this.WorksheetPageSetup.CopyFrom(sourceWorksheet.WorksheetPageSetup);
                this.ViewState.CopyFrom(sourceWorksheet.ViewState);
                this.SortState.CopyFrom(sourceWorksheet.SortState, copyContext);
                this.Filter.CopyFrom(sourceWorksheet.Filter, copyContext);
                this.GroupingProperties.CopyFrom(sourceWorksheet.GroupingProperties);

                this.Columns.PropertyBag.SetDefaultPropertyValue<ColumnWidth>(ColumnsPropertyBag.WidthProperty, sourceWorksheet.DefaultColumnWidth);
                this.Rows.PropertyBag.SetDefaultPropertyValue<RowHeight>(RowsPropertyBag.HeightProperty, sourceWorksheet.DefaultRowHeight);
                this.HeaderNameRenderingConverter = sourceWorksheet.HeaderNameRenderingConverter;
                this.Visibility = sourceWorksheet.Visibility;

                this.OnCopied();
                this.Workbook.InvalidateLayout();
            }
        }

        private void Clear()
        {
            WorkbookHistory history = this.Workbook.History;
            bool isHistoryEnabled = history.IsEnabled;

            using (new UpdateScope(() => history.IsEnabled = false, () => history.IsEnabled = isHistoryEnabled))
            {
                this.Names.Clear();
                this.Shapes.Clear();
                this.Hyperlinks.Clear();

                this.Cells.Clear();
                this.Rows.Clear();
                this.Columns.Clear();

                this.WorksheetPageSetup.Clear();
                this.SortState.ClearInternal();
                this.Filter.ClearFilters();
            }
        }

        internal bool CanInsertRows
        {
            get
            {
                return this.ShouldEnforceProtectionOption(this.ProtectionOptions.AllowInsertRows);
            }
        }

        internal bool CanDeleteRows
        {
            get
            {
                return this.ShouldEnforceProtectionOption(this.ProtectionOptions.AllowDeleteRows);
            }
        }

        internal bool CanInsertColumns
        {
            get
            {
                return this.ShouldEnforceProtectionOption(this.ProtectionOptions.AllowInsertColumns);
            }
        }

        internal bool CanDeleteColumns
        {
            get
            {
                return this.ShouldEnforceProtectionOption(this.ProtectionOptions.AllowDeleteColumns);
            }
        }

        internal bool CanFormatCells
        {
            get
            {
                return this.ShouldEnforceProtectionOption(this.ProtectionOptions.AllowFormatCells);
            }
        }

        internal bool CanFormatColumns
        {
            get
            {
                return this.ShouldEnforceProtectionOption(this.ProtectionOptions.AllowFormatColumns);
            }
        }

        internal bool CanFormatRows
        {
            get
            {
                return this.ShouldEnforceProtectionOption(this.ProtectionOptions.AllowFormatRows);
            }
        }

        internal bool CanFilter
        {
            get
            {
                return this.ShouldEnforceProtectionOption(this.ProtectionOptions.AllowFiltering);
            }
        }

        internal bool CanSort
        {
            get
            {
                return this.ShouldEnforceProtectionOption(this.ProtectionOptions.AllowSorting);
            }
        }

        private bool ShouldEnforceProtectionOption(bool protectionOption)
        {
            return !this.IsProtected || protectionOption;
        }

        /// <summary>
        /// Gets the used cell range for specific cell property definitions.
        /// </summary>
        /// <param name="propertyDefinitions">The cell property definitions from <see cref="Telerik.Windows.Documents.Spreadsheet.PropertySystem.CellPropertyDefinitions"/></param>
        /// <returns>The used cell range for specific cell property definitions.</returns>
        public CellRange GetUsedCellRange(IEnumerable<IPropertyDefinition> propertyDefinitions)
        {
            return this.Cells.PropertyBag.GetUsedCellRange(propertyDefinitions);
        }

        /// <summary>
        /// Disposes the specified disposing.
        /// </summary>
        /// <param name="disposing">The disposing.</param>
        protected override void Dispose(bool disposing)
        {
            if (!this.alreadyDisposed)
            {
                if (disposing)
                {
                    this.Workbook.NameManager.RemoveOwner(this);
                    this.names = null;
                    this.hyperlinks = null;
                }

                this.alreadyDisposed = true;
            }

            base.Dispose(disposing);
        }

        #endregion


        #region Events

        internal event EventHandler Copied;
        private void OnCopied()
        {
            if (this.Copied != null)
            {
                this.Copied(this, EventArgs.Empty);
            }
        }

        #endregion
    }
}
