﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Copying;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
#if NETCORE || NETSTANDARD
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a base class for table organization entities like rows, columns and cells.
    /// </summary>
    public abstract class WorksheetEntityBase
    {
        #region Fields

        private readonly Worksheet worksheet;

        private readonly Dictionary<IPropertyDefinition, IProperty> properties;
        private readonly IProperty<string> styleNameProperty;
        private readonly IProperty<CellValueFormat> formatProperty;
        private readonly IProperty<IFill> fillProperty;
        private readonly IProperty<ThemableFontFamily> fontFamilyProperty;
        private readonly IProperty<double> fontSizeProperty;
        private readonly IProperty<bool> isBoldProperty;
        private readonly IProperty<bool> isItalicProperty;
        private readonly IProperty<UnderlineType> underlineTypeProperty;
        private readonly IProperty<ThemableColor> foreColorProperty;
        private readonly IProperty<RadHorizontalAlignment> horizontalAlignmentProperty;
        private readonly IProperty<RadVerticalAlignment> verticalAlignmentProperty;
        private readonly IProperty<int> indentProperty;
        private readonly IProperty<bool> isWrappedProperty;
        private readonly IProperty<CellBorder> leftBorderProperty;
        private readonly IProperty<CellBorder> rightBorderProperty;
        private readonly IProperty<CellBorder> topBorderProperty;
        private readonly IProperty<CellBorder> bottomBorderProperty;
        private readonly IProperty<CellBorder> diagonalUpBorderProperty;
        private readonly IProperty<CellBorder> diagonalDownBorderProperty;
        private readonly IProperty<bool> isLockedProperty;

        #endregion


        #region Properties

        internal abstract PropertyBagBase PropertyBagBase
        {
            get;
        }

        /// <summary>
        /// Gets the worksheet.
        /// </summary>
        /// <value>The worksheet.</value>
        public Worksheet Worksheet
        {
            get
            {
                return this.worksheet;
            }
        }

        internal IEnumerable<IProperty> Properties
        {
            get
            {
                return this.properties.Values;
            }
        }

        internal IProperty<string> StyleNameProperty
        {
            get
            {
                return this.styleNameProperty;
            }
        }

        internal IProperty<CellValueFormat> FormatProperty
        {
            get
            {
                return this.formatProperty;
            }
        }

        internal IProperty<IFill> FillProperty
        {
            get
            {
                return this.fillProperty;
            }
        }

        internal IProperty<ThemableFontFamily> FontFamilyProperty
        {
            get
            {
                return this.fontFamilyProperty;
            }
        }

        internal IProperty<double> FontSizeProperty
        {
            get
            {
                return this.fontSizeProperty;
            }
        }

        internal IProperty<bool> IsBoldProperty
        {
            get
            {
                return this.isBoldProperty;
            }
        }

        internal IProperty<bool> IsItalicProperty
        {
            get
            {
                return this.isItalicProperty;
            }
        }

        internal IProperty<UnderlineType> UnderlineProperty
        {
            get
            {
                return this.underlineTypeProperty;
            }
        }

        internal IProperty<ThemableColor> ForeColorProperty
        {
            get
            {
                return this.foreColorProperty;
            }
        }

        internal IProperty<RadHorizontalAlignment> HorizontalAlignmentProperty
        {
            get
            {
                return this.horizontalAlignmentProperty;
            }
        }

        internal IProperty<RadVerticalAlignment> VerticalAlignmentProperty
        {
            get
            {
                return this.verticalAlignmentProperty;
            }
        }

        internal IProperty<int> IndentProperty
        {
            get
            {
                return this.indentProperty;
            }
        }

        internal IProperty<bool> IsWrappedProperty
        {
            get
            {
                return this.isWrappedProperty;
            }
        }

        internal IProperty<CellBorder> LeftBorderProperty
        {
            get
            {
                return this.leftBorderProperty;
            }
        }

        internal IProperty<CellBorder> RightBorderProperty
        {
            get
            {
                return this.rightBorderProperty;
            }
        }

        internal IProperty<CellBorder> TopBorderProperty
        {
            get
            {
                return this.topBorderProperty;
            }
        }

        internal IProperty<CellBorder> BottomBorderProperty
        {
            get
            {
                return this.bottomBorderProperty;
            }
        }

        internal IProperty<CellBorder> DiagonalUpBorderProperty
        {
            get
            {
                return this.diagonalUpBorderProperty;
            }
        }

        internal IProperty<CellBorder> DiagonalDownBorderProperty
        {
            get
            {
                return this.diagonalDownBorderProperty;
            }
        }

        internal IProperty<bool> IsLockedProperty
        {
            get
            {
                return this.isLockedProperty;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="WorksheetEntityBase" /> class.
        /// </summary>
        /// <param name="worksheet">The worksheet.</param>
        protected WorksheetEntityBase(Worksheet worksheet)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            this.worksheet = worksheet;

            this.properties = new Dictionary<IPropertyDefinition, IProperty>();

            this.formatProperty = this.CreateProperty(CellPropertyDefinitions.FormatProperty);
            this.styleNameProperty = this.CreateProperty(CellPropertyDefinitions.StyleNameProperty);
            this.fillProperty = this.CreateProperty(CellPropertyDefinitions.FillProperty);
            this.fontFamilyProperty = this.CreateProperty(CellPropertyDefinitions.FontFamilyProperty);
            this.fontSizeProperty = this.CreateProperty(CellPropertyDefinitions.FontSizeProperty);
            this.isBoldProperty = this.CreateProperty(CellPropertyDefinitions.IsBoldProperty);
            this.isItalicProperty = this.CreateProperty(CellPropertyDefinitions.IsItalicProperty);
            this.underlineTypeProperty = this.CreateProperty(CellPropertyDefinitions.UnderlineProperty);
            this.foreColorProperty = this.CreateProperty(CellPropertyDefinitions.ForeColorProperty);
            this.horizontalAlignmentProperty = this.CreateProperty(CellPropertyDefinitions.HorizontalAlignmentProperty);
            this.verticalAlignmentProperty = this.CreateProperty(CellPropertyDefinitions.VerticalAlignmentProperty);
            this.indentProperty = this.CreateProperty(CellPropertyDefinitions.IndentProperty);
            this.isWrappedProperty = this.CreateProperty(CellPropertyDefinitions.IsWrappedProperty);
            this.isLockedProperty = this.CreateProperty(CellPropertyDefinitions.IsLockedProperty);
            this.leftBorderProperty = this.CreateProperty(CellPropertyDefinitions.LeftBorderProperty);
            this.rightBorderProperty = this.CreateProperty(CellPropertyDefinitions.RightBorderProperty);
            this.topBorderProperty = this.CreateProperty(CellPropertyDefinitions.TopBorderProperty);
            this.bottomBorderProperty = this.CreateProperty(CellPropertyDefinitions.BottomBorderProperty);
            this.diagonalUpBorderProperty = this.CreateProperty(CellPropertyDefinitions.DiagonalUpBorderProperty);
            this.diagonalDownBorderProperty = this.CreateProperty(CellPropertyDefinitions.DiagonalDownBorderProperty);
        }

        #endregion


        #region Methods

        internal IProperty<T> CreateProperty<T>(IPropertyDefinition<T> propertyDefinition)
        {
            Guard.ThrowExceptionIfNull(propertyDefinition, "propertyDefinition");

            IProperty<T> property = this.CreatePropertyOverride<T>(propertyDefinition);
            this.AddToProperties(property);

            return property;
        }

        internal abstract IProperty<T> CreatePropertyOverride<T>(IPropertyDefinition<T> propertyDefinition);

        private void AddToProperties(IProperty property)
        {
            this.properties.Add(property.PropertyDefinition, property);
        }

        internal bool TryGetProperyFromPropertyDefinition<T>(IPropertyDefinition propertyDefinition, out IProperty<T> property)
        {
            IProperty prop;
            bool result = this.TryGetProperyFromPropertyDefinition(propertyDefinition, out prop);
            property = prop as IProperty<T>;

            return result;
        }

        internal bool TryGetProperyFromPropertyDefinition(IPropertyDefinition propertyDefinition, out IProperty property)
        {
            return this.properties.TryGetValue(propertyDefinition, out property);
        }

        internal bool ContainsPropertyDefinition(IPropertyDefinition property)
        {
            return this.properties.ContainsKey(property);
        }

        internal void CopyFrom(WorksheetEntityBase fromWorksheetEntity, CopyContext cloneContext)
        {
            this.CopyFromOverride(fromWorksheetEntity, cloneContext, this.Properties);
        }

        internal virtual void CopyFromOverride(WorksheetEntityBase fromWorksheetEntity, CopyContext cloneContext, IEnumerable<IProperty> properties)
        {
            foreach (IProperty property in properties)
            {
                ICompressedList sourcePropertyCollection = fromWorksheetEntity.PropertyBagBase.GetPropertyValueCollection(property.PropertyDefinition);
                ICompressedList targetPropertyCollection = this.PropertyBagBase.GetPropertyValueCollection(property.PropertyDefinition);

                foreach (LongRange range in sourcePropertyCollection.GetRanges(false))
                {
                    ICompressedList value = sourcePropertyCollection.GetValue(range.Start, range.End);
                    targetPropertyCollection.SetValue(value);
                }
            }
        }

        internal virtual void Clear()
        {
            foreach (IProperty property in this.Properties)
            {
                this.PropertyBagBase.ClearPropertyValue(property.PropertyDefinition);
            }
        }

        #endregion
    }
}