﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorksheetCommands;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Layout;
using Telerik.Windows.Documents.Spreadsheet.Model.Shapes;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

#if SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.Maths;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents a fragment of a worksheet.
    /// </summary>
    public class WorksheetFragment : IDisposable
    {
        #region Nested Types

        private static class WorkbookPool
        {
            private static readonly int maxCount = 5;
            private static readonly Queue<Workbook> innerList = new Queue<Workbook>(maxCount);

            private static readonly Object lockObject = new Object();

            public static Workbook GetWorkbook()
            {
                if (innerList.Count > 0)
                {
                    lock (lockObject)
                    {
                        if (innerList.Count > 0)
                        {
                            return innerList.Dequeue();
                        }
                        else
                        {
                            return CreateWorkbook();
                        }
                    }
                }
                else
                {
                    return CreateWorkbook();
                }
            }

            private static Workbook CreateWorkbook()
            {
                Workbook workbook = new Workbook();
                workbook.History.IsEnabled = false;
                return workbook;
            }

            public static void ReturnWorkbook(Workbook workbook)
            {
                if (innerList.Count >= maxCount)
                {
                    workbook.Dispose();
                }
                else
                {
                    workbook.Worksheets.Clear();
                    innerList.Enqueue(workbook);
                }
            }
        }

        #endregion


        #region Constants

        internal static readonly int MaxAllowedAffectedRanges = 500000;
        internal static readonly int MaxAllowedAffectedRangesWithUndo = 200000;

        #endregion


        #region Fields

        private readonly Workbook workbook;
        private readonly Worksheet worksheet;
        private readonly CellRange cellRange;
        private readonly IEnumerable<CellRange> cellRanges;
        private readonly IEnumerable<FloatingShapeBase> shapes;
        private readonly bool shouldPasteAsText;
        private bool alreadyDisposed;
        private readonly bool isCellCopyPaste;

        // TODO: Try not to use static
        private static int fromCellRangesRowCount;
        private static int fromCellRangesColumnCount;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the cell range of the fragment.
        /// </summary>
        /// <value>The cell range of the fragment.</value>
        public CellRange CellRange
        {
            get
            {
                return this.cellRange;
            }
        }

        internal IEnumerable<CellRange> FromCellRanges
        {
            get
            {
                return this.cellRanges;
            }
        }

        /// <summary>
        /// Gets the shapes included in the fragment.
        /// </summary>
        /// <value>The shapes.</value>
        internal IEnumerable<FloatingShapeBase> Shapes
        {
            get
            {
                return this.shapes;
            }
        }

        /// <summary>
        /// Gets the workbook.
        /// </summary>
        /// <value>The workbook.</value>
        public Workbook Workbook
        {
            get
            {
                return this.workbook;
            }
        }

        /// <summary>
        /// Gets the worksheet.
        /// </summary>
        /// <value>The worksheet.</value>
        public Worksheet Worksheet
        {
            get
            {
                return this.worksheet;
            }
        }

        // The property is set to true when the source of the fragment is the external clipboard
        // In this case the values in the fragment should be pasted as literal text, e.g. =B1 should not be translated
        internal bool ShouldPasteAsText
        {
            get
            {
                return this.shouldPasteAsText;
            }
        }

        internal bool IsCellCopyPaste
        {
            get
            {
                return this.isCellCopyPaste;
            }
        }

        #endregion


        #region Constructors

        internal WorksheetFragment(CellRange cellRange, IEnumerable<FloatingShapeBase> shapes, Worksheet fromWorksheet, bool shouldPasteAsText, bool isCellCopyPaste)
        {
            if (isCellCopyPaste)
            {
                Guard.ThrowExceptionIfNull(cellRange, "cellRange");

                List<CellRange> fromCellRanges = CopyPasteContext.GetFromCellRange(fromWorksheet, cellRange);

                this.cellRanges = fromCellRanges;
                fromCellRangesRowCount = SpreadsheetHelper.CountRows(this.FromCellRanges);
                fromCellRangesColumnCount = SpreadsheetHelper.CountColumns(this.FromCellRanges);
                this.cellRange = new CellRange(0, 0, fromCellRangesRowCount - 1, fromCellRangesColumnCount - 1);
                this.shapes = new List<FloatingShapeBase>();
            }
            else
            {
                Guard.ThrowExceptionIfNull(shapes, "shapes");

                this.shapes = new List<FloatingShapeBase>(shapes);
            }

            this.workbook = WorkbookPool.GetWorkbook();

            this.worksheet = this.workbook.Worksheets.Add();
            this.shouldPasteAsText = shouldPasteAsText;
            this.isCellCopyPaste = isCellCopyPaste;

            CopyPasteContext context = new CopyPasteContext(fromWorksheet, cellRange, this.worksheet, this.cellRange, this.shapes, null, null,
                isCellCopyPaste: isCellCopyPaste, isCopying: true, shouldPasteAsText: this.shouldPasteAsText);

            CopyPaste(context);
        }

        #endregion


        #region Methods

        internal bool Paste(Worksheet toWorksheet, CellRange toCellRange, IEnumerable<FloatingShapeBase> toShapes, PasteOptions pasteOptions, out List<CellRange> affectedCellRanges, out IEnumerable<FloatingShapeBase> newShapes, bool isCellCopyPaste)
        {
            this.EnsureNotDisposed();

            Guard.ThrowExceptionIfNull(toWorksheet, "toWorksheet");
            Guard.ThrowExceptionIfTrue(toCellRange == null && toShapes == null, "toCellRange && toShapes");

            CopyPasteContext context = new CopyPasteContext(this.worksheet, this.cellRange, toWorksheet, toCellRange, this.shapes, toShapes, pasteOptions,
                isCellCopyPaste: isCellCopyPaste, isCopying: false, shouldPasteAsText: this.ShouldPasteAsText);

            return CopyPaste(context, out affectedCellRanges, out newShapes);
        }

        internal static CellRange EnsureProportionalPasteCellRange(IEnumerable<CellRange> fromCellRange, CellRange toCellRange)
        {
            if (fromCellRange != null)
            {
                if (toCellRange.RowCount % fromCellRangesRowCount != 0 ||
                    toCellRange.ColumnCount % fromCellRangesColumnCount != 0)
                {
                    CellIndex toIndex = toCellRange.FromIndex.Offset(fromCellRangesRowCount - 1, fromCellRangesColumnCount - 1);
                    if (toIndex == null)
                    {
                        return null;
                    }

                    return new CellRange(toCellRange.FromIndex, toIndex);
                }
            }

            return toCellRange;
        }

        internal static bool TrySplitToPasteCellRanges(IEnumerable<CellRange> fromCellRange, CellRange toCellRange, out List<CellRange> cellRanges)
        {
            cellRanges = new List<CellRange>();

            toCellRange = EnsureProportionalPasteCellRange(fromCellRange, toCellRange);
            if (toCellRange == null)
            {
                return false;
            }

            for (int rowIndex = toCellRange.FromIndex.RowIndex; rowIndex <= toCellRange.ToIndex.RowIndex; rowIndex += fromCellRangesRowCount)
            {
                for (int columnIndex = toCellRange.FromIndex.ColumnIndex; columnIndex <= toCellRange.ToIndex.ColumnIndex; columnIndex += fromCellRangesColumnCount)
                {
                    int fromRowIndex = rowIndex;
                    int fromColumnIndex = columnIndex;
                    int toRowIndex = fromRowIndex + fromCellRangesRowCount - 1;
                    int toColumnIndex = fromColumnIndex + fromCellRangesColumnCount - 1;

                    if (!TelerikHelper.IsValidRowIndex(toRowIndex) || !TelerikHelper.IsValidColumnIndex(toColumnIndex))
                    {
                        return false;
                    }

                    CellRange newRange = new CellRange(fromRowIndex, fromColumnIndex, toRowIndex, toColumnIndex);
                    cellRanges.Add(newRange);
                }
            }

            return true;
        }

        internal static long GetPasteCellRangesCount(IEnumerable<CellRange> fromCellRange, CellRange proportionalToCellRange)
        {
            int fromCellRangesRowCount = SpreadsheetHelper.CountRows(fromCellRange);
            int fromCellRangesColumnCount = SpreadsheetHelper.CountColumns(fromCellRange);

            long rows = proportionalToCellRange.RowCount / fromCellRangesRowCount;
            long columns = proportionalToCellRange.ColumnCount / fromCellRangesColumnCount;

            long rangeCount = rows * columns;
            return rangeCount;
        }

        private static bool CopyPaste(CopyPasteContext context)
        {
            List<CellRange> affectedCellRanges;
            IEnumerable<FloatingShapeBase> newShapes;
            return CopyPaste(context, out affectedCellRanges, out newShapes);
        }

        private static bool CopyPaste(CopyPasteContext context, out List<CellRange> affectedCellRanges, out IEnumerable<FloatingShapeBase> newShapes)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            newShapes = new List<FloatingShapeBase>();

            if (context.IsCellCopyPaste)
            {
                if (!TrySplitToPasteCellRanges(context.FromCellRange, context.ToCellRange, out affectedCellRanges))
                {
                    return false;
                }

                context.AffectedCellRanges = affectedCellRanges;

                using (new UpdateScope(context.ToWorksheet.BeginUndoGroup, context.ToWorksheet.EndUndoGroup))
                {
                    CopyPasteMergedCellRanges(context);
                    CopyPasteHyperlinks(context);
                    
                    // The following method calls order respects the style system priorities
                    CopyPasteColumnProperties(context);
                    CopyPasteRowProperties(context);
                    CopyPasteCellProperties(context);
                }

                return true;
            }
            else
            {
                CopyPasteShapes(context, out newShapes);

                affectedCellRanges = new List<CellRange>() { new CellRange(0, 0, 0, 0) };
                context.AffectedCellRanges = affectedCellRanges;
                return true;
            }
        }

        private static void CopyPasteMergedCellRanges(CopyPasteContext context)
        {
            if (context.PasteOptions.SkipMergedCells)
            {
                return;
            }

            ISet<CellRange> fromMergedCellRanges = new HashSet<CellRange>();
            foreach (CellRange fromCellRange in context.FromCellRange)
            {
                ISet<CellRange> currentFromMergedCellRanges = context.FromWorksheet.Cells.MergedCellRanges.GetIntersectingMergedRanges(fromCellRange);
                fromMergedCellRanges.UnionWith(currentFromMergedCellRanges);
            }

            if (fromMergedCellRanges.Count == 0)
            {
                return;
            }

            foreach (var toCellRange in context.AffectedCellRanges)
            {
                WorkbookCommands.UnmergeCells.Execute(new UnmergeCellsCommandContext(context.ToWorksheet, toCellRange));

                int rowOffset = toCellRange.FromIndex.RowIndex - context.FromCellRange.First().FromIndex.RowIndex;
                int columnOffset = toCellRange.FromIndex.ColumnIndex - context.FromCellRange.First().FromIndex.ColumnIndex;
                foreach (CellRange fromMergedCellRange in fromMergedCellRanges)
                {
                    context.ToWorksheet.Cells[fromMergedCellRange.Offset(rowOffset, columnOffset)].Merge();
                }
            }
        }

        private static void CopyPasteHyperlinks(CopyPasteContext context)
        {
            if (!context.ShouldCopyPasteHyperlinks())
            {
                return;
            }

            List<SpreadsheetHyperlink> hyperlinksToCopy = new List<SpreadsheetHyperlink>();

            IEnumerable<SpreadsheetHyperlink> intersectingHyperlinks = new List<SpreadsheetHyperlink>();
            foreach (CellRange fromRange in context.FromCellRange)
            {
                IEnumerable<SpreadsheetHyperlink> currentIntersectingHyperlinks = context.FromWorksheet.Hyperlinks.GetIntersectingHyperlinks(fromRange);
                intersectingHyperlinks = intersectingHyperlinks.Union(currentIntersectingHyperlinks);
            }

            foreach (SpreadsheetHyperlink hyperlink in intersectingHyperlinks)
            {
                if (context.FromCellRange.Any(x => x.Contains(hyperlink.Range)))
                {
                    hyperlinksToCopy.Add(hyperlink);
                }
                else
                {
                    foreach (CellRange fromCellRange in context.FromCellRange)
                    {
                        CellRange intersect = fromCellRange.Intersect(hyperlink.Range);
                        if (intersect != CellRange.Empty)
                        {
                            hyperlinksToCopy.Add(new SpreadsheetHyperlink(intersect, hyperlink.HyperlinkInfo));
                            break;
                        }
                    }
                }
            }

            int rowOffset = context.ToCellRange.FromIndex.RowIndex - context.FromCellRange.First().FromIndex.RowIndex;
            int columnOffset = context.ToCellRange.FromIndex.ColumnIndex - context.FromCellRange.First().FromIndex.ColumnIndex;

            foreach (SpreadsheetHyperlink hyperlink in hyperlinksToCopy)
            {
                context.ToWorksheet.Hyperlinks.Add(hyperlink.Range.Offset(rowOffset, columnOffset), hyperlink.HyperlinkInfo);
            }
        }

        private static void CopyPasteShapes(CopyPasteContext context, out IEnumerable<FloatingShapeBase> newShapes)
        {
            Point fromGroupTopLeft = GetTopLeftOfShapeGroup(context.FromWorksheet, context.FromShapes);

            RadWorksheetLayout toWorksheetLayout = context.ToWorksheet.Workbook.GetWorksheetLayout(context.ToWorksheet, false);
            RadWorksheetLayout fromWorksheetLayout = context.FromWorksheet.Workbook.GetWorksheetLayout(context.FromWorksheet, false);

            Point pointToPlace;
            if (context.ToCellRange != null)
            {
                CellIndex indexToPlace;
                indexToPlace = context.ToCellRange.FromIndex;
                pointToPlace = toWorksheetLayout.GetTopLeftPointFromCellIndex(indexToPlace);
            }
            else if (context.ToShapes != null)
            {
                Point toGroupTopLeft = GetTopLeftOfShapeGroup(context.ToWorksheet, context.ToShapes);
                pointToPlace = toGroupTopLeft;
                pointToPlace.Offset(ShapeCollection.InsertDistance, ShapeCollection.InsertDistance);
            }
            else
            {
                pointToPlace = new Point(0, 0);
            }

            List<FloatingShapeBase> result = new List<FloatingShapeBase>();
            foreach (FloatingShapeBase shape in context.FromShapes)
            {
                Point shapeTopLeft = fromWorksheetLayout.GetShapeTopLeft(shape);
                double leftOffsetFromGroup = shapeTopLeft.X - fromGroupTopLeft.X;
                double topOffsetFromGroup = shapeTopLeft.Y - fromGroupTopLeft.Y;

                Point newShapeTopLeft = new Point(pointToPlace.X + leftOffsetFromGroup, pointToPlace.Y + topOffsetFromGroup);
                CellIndex newCellIndex = toWorksheetLayout.GetCellIndexFromPoint(newShapeTopLeft.X, newShapeTopLeft.Y, false, false);
                Point cellTopleft = toWorksheetLayout.GetTopLeftPointFromCellIndex(newCellIndex);
                double offsetX = newShapeTopLeft.X - cellTopleft.X;
                double offsetY = newShapeTopLeft.Y - cellTopleft.Y;

                result.Add(shape.Copy(context.ToWorksheet, newCellIndex, offsetX, offsetY));
            }

            context.ToWorksheet.Shapes.Add(result);
            newShapes = result;
        }

        private static Point GetTopLeftOfShapeGroup(Worksheet worksheet, IEnumerable<FloatingShapeBase> shapes)
        {
            if (shapes.Count() == 0)
            {
                return new Point(0, 0);
            }

            double groupLeft = Double.MaxValue;
            double groupTop = Double.MaxValue;

            foreach (var shape in shapes)
            {
                Point shapeTopLeft = worksheet.Workbook.GetWorksheetLayout(worksheet, false).GetShapeTopLeft(shape);
                double shapeLeft = shapeTopLeft.X;
                double shapeTop = shapeTopLeft.Y;

                if (shapeLeft < groupLeft)
                {
                    groupLeft = shapeLeft;
                }

                if (shapeTop < groupTop)
                {
                    groupTop = shapeTop;
                }
            }

            return new Point(groupLeft, groupTop);
        }

        private static void CopyPasteCellProperties(CopyPasteContext context)
        {
            if (context.PasteOptions.PasteType == PasteType.ColumnWidths)
            {
                return;
            }

            CellsPropertyBag targetPropertyBag = context.ToWorksheet.Cells.PropertyBag;
            using (new UpdateScope(targetPropertyBag.SuspendPropertyChanged, () =>
            {
                targetPropertyBag.ResumePropertyChanged(context.ToCellRange);
            }))
            {

                List<LongRange> fromCellRangeParts = new List<LongRange>();
                foreach (CellRange fromCellRange in context.FromCellRange)
                {
                    Range<long, ICellValue> range = null;
                    foreach (LongRange longCellRange in CellsPropertyBag.ConvertCellRangeToLongRanges(fromCellRange))
                    {
                        var rangeValues = context.FromWorksheet.Cells.PropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.ValueProperty)
                            .GetValue(longCellRange.Start, longCellRange.End);

                        if (!context.PasteOptions.SkipBlanks)
                        {
                            fromCellRangeParts.Add(longCellRange);
                        }
                        else
                        {
                            foreach (var nonDefaultRange in rangeValues.GetNonDefaultRanges())
                            {
                                if (TelerikHelper.EqualsOfT(nonDefaultRange.Value, EmptyCellValue.EmptyValue))
                                {
                                    continue;
                                }

                                if (range == null)
                                {
                                    range = nonDefaultRange;
                                }
                                else if ((nonDefaultRange.Start - 1) == range.End)
                                {
                                    range.End = nonDefaultRange.End;
                                }
                                else
                                {
                                    var longRange = new LongRange(range.Start, range.End);
                                    fromCellRangeParts.Add(longRange);
                                    range = nonDefaultRange;
                                }
                            }
                            if (range != null)
                            {
                                var longRange = new LongRange(range.Start, range.End);
                                fromCellRangeParts.Add(longRange);
                            }
                        }
                    }
                }

                foreach (CellRange toCellRange in context.AffectedCellRanges)
                {
                    foreach (LongRange fromCellRangePart in fromCellRangeParts)
                    {
                        int fromRowIndex;
                        int fromColumnIndex;
                        int toRowIndex;
                        int toColumnIndex;
                        CellsPropertyBag.ConvertLongToRowAndColumnIndexes(fromCellRangePart.Start, out fromRowIndex, out fromColumnIndex);
                        CellsPropertyBag.ConvertLongToRowAndColumnIndexes(fromCellRangePart.End, out toRowIndex, out toColumnIndex);

                        int rowOffset = toCellRange.FromIndex.RowIndex - context.FromCellRange.First().FromIndex.RowIndex;
                        int columnOffset = toCellRange.FromIndex.ColumnIndex - context.FromCellRange.First().FromIndex.ColumnIndex;

                        for (int i = 0; i < context.HiddenRows.Length; i++)
                        {
                            int hiddenRowIndex = context.HiddenRows[i];
                            if (hiddenRowIndex < fromRowIndex)
                            {
                                rowOffset--;
                            }
                        }

                        for (int i = 0; i < context.HiddenColumns.Length; i++)
                        {
                            int hiddenColumnIndex = context.HiddenColumns[i];
                            if (hiddenColumnIndex < fromColumnIndex)
                            {
                                columnOffset--;
                            }
                        }

                        fromRowIndex += rowOffset;
                        toRowIndex += rowOffset;
                        fromColumnIndex += columnOffset;
                        toColumnIndex += columnOffset;

                        long fromIndex = CellsPropertyBag.ConvertCellIndexToLong(fromRowIndex, fromColumnIndex);
                        long toIndex = CellsPropertyBag.ConvertCellIndexToLong(toRowIndex, toColumnIndex);

                        CopyPasteCellProperties(context, fromCellRangePart, new LongRange(fromIndex, toIndex));
                    }
                }
            }
        }

        private static void CopyPasteCellProperties(CopyPasteContext context, LongRange fromLongRange, LongRange toLongRange)
        {
            Guard.ThrowExceptionIfNotEqual(fromLongRange.Length, toLongRange.Length, "Length");

            HashSet<IPropertyDefinition> nonDefaultProperties = GetNonDefaultProperties(context, fromLongRange);

            switch (context.PasteOptions.PasteType)
            {
                case PasteType.Formulas:
                    CopyPasteCellValueProperty(context, nonDefaultProperties, fromLongRange, toLongRange, false);
                    break;
                case PasteType.FormulasAndNumberFormats:
                    CopyPasteCellValueProperty(context, nonDefaultProperties, fromLongRange, toLongRange, false);
                    CopyPasteFormatCellProperty(context, nonDefaultProperties, fromLongRange, toLongRange);
                    break;
                case PasteType.Values:
                    CopyPasteCellValueProperty(context, nonDefaultProperties, fromLongRange, toLongRange, true);
                    break;
                case PasteType.ValuesAndNumberFormats:
                    CopyPasteCellValueProperty(context, nonDefaultProperties, fromLongRange, toLongRange, true);
                    CopyPasteFormatCellProperty(context, nonDefaultProperties, fromLongRange, toLongRange);
                    break;
                case PasteType.Formats:
                    CopyPasteWorkbookStyles(context, fromLongRange);
                    CopyPasteIsLockedCellProperty(context, fromLongRange, toLongRange);
                    CopyPastePropertiesButIsLockedAndValues(context, nonDefaultProperties, fromLongRange, toLongRange);
                    break;
                default:
                    CopyPasteCellValueProperty(context, nonDefaultProperties, fromLongRange, toLongRange, false);
                    CopyPasteWorkbookStyles(context, fromLongRange);
                    CopyPasteIsLockedCellProperty(context, fromLongRange, toLongRange);
                    CopyPastePropertiesButIsLockedAndValues(context, nonDefaultProperties, fromLongRange, toLongRange);
                    break;
            }
        }

        private static HashSet<IPropertyDefinition> GetNonDefaultProperties(CopyPasteContext context, LongRange fromLongRange)
        {
            HashSet<IPropertyDefinition> nonDefaultproperties = new HashSet<IPropertyDefinition>();

            foreach (IProperty cellProperty in context.FromWorksheet.Cells.Properties)
            {
                var propertyBag = context.FromWorksheet.Cells.PropertyBag;
                bool containsNonDefaultValues = propertyBag.GetPropertyValueCollection(cellProperty.PropertyDefinition)
                        .ContainsNonDefaultValues(fromLongRange.Start, fromLongRange.End);

                if (containsNonDefaultValues)
                {
                    nonDefaultproperties.Add(cellProperty.PropertyDefinition);
                }
            }

            return nonDefaultproperties;
        }

        private static void CopyPastePropertiesButIsLockedAndValues(CopyPasteContext context, HashSet<IPropertyDefinition> nonDefaultProperties,
            LongRange fromLongRange, LongRange toLongRange)
        {
            foreach (var cellProperty in context.FromWorksheet.Cells.Properties)
            {
                if (nonDefaultProperties.Contains(cellProperty.PropertyDefinition) &&
                    cellProperty.PropertyDefinition != CellPropertyDefinitions.ValueProperty &&
                    cellProperty.PropertyDefinition != CellPropertyDefinitions.IsLockedProperty)
                {
                    CopyPasteCellProperty(context, cellProperty, fromLongRange, toLongRange);
                }
            }
        }

        private static void CopyPasteWorkbookStyles(CopyPasteContext context, LongRange fromLongRange)
        {
            var cells = context.FromWorksheet.Cells;
            IEnumerable<Range<long, string>> fromWorksheetStyleNameRanges = cells.PropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.StyleNameProperty).GetValue(fromLongRange.Start, fromLongRange.End).GetNonDefaultRanges();

            CellStyleCollection fromStyles = context.FromWorksheet.Workbook.Styles;
            CellStyleCollection toStyles = context.ToWorksheet.Workbook.Styles;

            foreach (Range<long, string> styleRange in fromWorksheetStyleNameRanges)
            {
                string styleName = styleRange.Value;

                if (!toStyles.Contains(styleName))
                {
                    CellStyle fromStyle = fromStyles[styleName];
                    toStyles.Add(fromStyle.Name, fromStyle.Category, fromStyle.IsRemovable).CopyPropertiesFrom(fromStyle);
                }
            }
        }

        private static void CopyPasteCellValueProperty(CopyPasteContext context, HashSet<IPropertyDefinition> nonDefaultProperties, LongRange fromLongRange,
            LongRange toLongRange, bool transformValueIfFormulaCellValue)
        {
            if (!nonDefaultProperties.Contains(CellPropertyDefinitions.ValueProperty))
            {
                return;
            }

            Func<ICellValue, ICellValue> transformValue = new Func<ICellValue, ICellValue>(
            (cellValue) =>
            {
                FormulaCellValue formulaCellValue = cellValue as FormulaCellValue;
                if (formulaCellValue != null && transformValueIfFormulaCellValue)
                {
                    return formulaCellValue.GetResultValueAsCellValue();
                }

                return cellValue;
            });

            var fromRangeValues = context.FromWorksheet.Cells.PropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.ValueProperty)
                .GetValue(fromLongRange.Start, fromLongRange.End);

            var toRangeValues = new CompressedList<ICellValue>(toLongRange.Start, toLongRange.End, fromRangeValues.GetDefaultValue());
            foreach (Range<long, ICellValue> rangeValue in fromRangeValues)
            {
                ICellValue transformedValue = transformValue(rangeValue.Value);

                long offsetStart = rangeValue.Start - fromLongRange.Start;
                long rangeLength = rangeValue.End - rangeValue.Start + 1;

                long toRangeStart = toLongRange.Start + offsetStart;
                long toRangeEnd = toRangeStart + rangeLength - 1;

                bool shouldSetValues = false;
                var currentToRangeValues = context.ToWorksheet.Cells.PropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.ValueProperty)
                    .GetValue(toRangeStart, toRangeEnd);

                foreach (var currentToRange in currentToRangeValues)
                {
                    if (!TelerikHelper.EqualsOfT(transformedValue, currentToRange.Value))
                    {
                        shouldSetValues = true;
                        transformedValue = GetTransformedValue(context, transformedValue, currentToRange.Start, rangeValue.Value, toRangeStart);
                        break;
                    }
                }

                if (shouldSetValues || transformedValue != currentToRangeValues.GetDefaultValue())
                {
                    toRangeValues.SetValue(toRangeStart, toRangeEnd, transformedValue);
                }
            }

            WorkbookCommands.SetCellProperty.Execute(new SetPropertyValuesCommandContext(context.ToWorksheet, CellPropertyDefinitions.ValueProperty, toRangeValues));
        }

        private static ICellValue GetTransformedValue(CopyPasteContext context, ICellValue transformedValue,
            long currentToRangeStart, ICellValue currentCellValue, long toRangeStart)
        {
            if (!context.IsCopying)
            {
                FormulaCellValue formulaValue = transformedValue as FormulaCellValue;
                if (formulaValue != null)
                {
                    int rowIndex;
                    int columnIndex;
                    CellsPropertyBag.ConvertLongToRowAndColumnIndexes(currentToRangeStart, out rowIndex, out columnIndex);
                    return formulaValue.CloneAndTranslate(context.ToWorksheet, rowIndex, columnIndex);
                }
            }

            if (context.ShouldPasteAsText)
            {
                ICellValue newCellValue;
                CellValueFormat newCellValueFormat;

                int rowIndex;
                int columnIndex;
                CellsPropertyBag.ConvertLongToRowAndColumnIndexes(toRangeStart, out rowIndex, out columnIndex);
                CellValueFactory.CreateIgnoreErrors(currentCellValue.RawValue, context.ToWorksheet, rowIndex, columnIndex,
                    CellValueFormat.GeneralFormat, out newCellValue, out newCellValueFormat);

                return newCellValue;
            }

            return transformedValue;
        }

        private static void CopyPasteCellProperty(CopyPasteContext context, IProperty cellProperty,
            LongRange fromLongRange, LongRange toLongRange)
        {
            var propertyBag = context.FromWorksheet.Cells.PropertyBag;
            ICompressedList fromRangeValues = propertyBag.GetPropertyValueCollection(cellProperty.PropertyDefinition).GetValue(fromLongRange.Start, fromLongRange.End);

            var toRangeValues = fromRangeValues.Offset(toLongRange.Start - fromLongRange.Start);

            WorkbookCommands.SetCellProperty.Execute(new SetPropertyValuesCommandContext(context.ToWorksheet, cellProperty.PropertyDefinition, toRangeValues));
        }

        private static void CopyPasteIsLockedCellProperty(CopyPasteContext context, LongRange fromLongRange, LongRange toLongRange)
        {
            ICompressedList toRangeValues;

            var propertyBag = context.ToWorksheet.Cells.PropertyBag;
            if (context.ToWorksheet.IsProtected)
            {
                toRangeValues = propertyBag.GetPropertyValueRespectingStyle(CellPropertyDefinitions.IsLockedProperty, context.ToWorksheet, toLongRange.Start, toLongRange.End);
            }
            else
            {
                ICompressedList fromRangeValues = propertyBag.GetPropertyValueCollection(context.FromWorksheet.Cells.IsLockedProperty.PropertyDefinition)
                        .GetValue(fromLongRange.Start, fromLongRange.End);

                toRangeValues = fromRangeValues.Offset(toLongRange.Start - fromLongRange.Start);
            }

            WorkbookCommands.SetCellProperty.Execute(new SetPropertyValuesCommandContext(context.ToWorksheet, context.FromWorksheet.Cells.IsLockedProperty.PropertyDefinition, toRangeValues));
        }

        private static void CopyPasteFormatCellProperty(CopyPasteContext context, HashSet<IPropertyDefinition> nonDefaultProperties,
            LongRange fromLongRange, LongRange toLongRange)
        {
            if (!nonDefaultProperties.Contains(CellPropertyDefinitions.ValueProperty))
            {
                return;
            }

            ICompressedList<CellValueFormat> fromRangeValues = context.FromWorksheet.Cells.PropertyBag.
                GetPropertyValueRespectingStyle(CellPropertyDefinitions.FormatProperty, context.FromWorksheet, fromLongRange.Start, fromLongRange.End);
            ICompressedList toRangeValues = fromRangeValues.Offset(toLongRange.Start - fromLongRange.Start);

            WorkbookCommands.SetCellProperty.Execute(new SetPropertyValuesCommandContext(context.ToWorksheet, context.FromWorksheet.Cells.FormatProperty.PropertyDefinition, toRangeValues));
        }

        private static void CopyPasteColumnProperties(CopyPasteContext context)
        {
            CopyPasteColumnWidth(context);

            Columns columns = context.FromWorksheet.Columns;
            CopyPasteEntireRowColumnProperties(context, columns.PropertyBag, columns.Properties, WorkbookCommands.SetColumnProperty,
                (range) =>
                {
                    return range.RowCount == SpreadsheetDefaultValues.RowCount || context.IsCopying;
                });

            CopyPasteRowColumnProperties(context, columns.PropertyBag, columns.Properties, (range) =>
                {
                    return range.RowCount != SpreadsheetDefaultValues.RowCount && !context.IsCopying;
                }, SetColumnValuesForCells);
        }

        private static void CopyPasteRowProperties(CopyPasteContext context)
        {
            Rows rows = context.FromWorksheet.Rows;
            CopyPasteEntireRowColumnProperties(context, rows.PropertyBag, rows.Properties, WorkbookCommands.SetRowProperty,
                (range) =>
                {
                    return range.ColumnCount == SpreadsheetDefaultValues.ColumnCount || context.IsCopying;
                });

            CopyPasteRowColumnProperties(context, rows.PropertyBag, rows.Properties, (range) =>
            {
                return range.ColumnCount != SpreadsheetDefaultValues.ColumnCount && !context.IsCopying;
            }, SetRowValuesForCells);
        }

        private static void CopyPasteEntireRowColumnProperties(CopyPasteContext context, RowColumnPropertyBagBase propertyBag, IEnumerable<IProperty> properties,
            UndoableWorksheetCommandBase<SetPropertyValuesCommandContext> command, Func<CellRange, bool> shouldCopyPasteColumn)
        {
            Dictionary<IProperty, ICompressedList> propertyToNewValues = new Dictionary<IProperty, ICompressedList>();
            foreach (CellRange fromCellRange in context.FromCellRange)
            {
                if (!shouldCopyPasteColumn(fromCellRange))
                {
                    continue;
                }

                foreach (IProperty property in properties)
                {
                    ICompressedList propertyValueCollection = propertyBag.GetPropertyValueCollection(property.PropertyDefinition);
                    ICompressedList currentValues = propertyValueCollection.GetValue(
                        propertyBag.GetRowColumnIndex(fromCellRange.FromIndex), propertyBag.GetRowColumnIndex(fromCellRange.ToIndex));

                    if (!propertyToNewValues.ContainsKey(property))
                    {
                        propertyToNewValues.Add(property, currentValues);
                    }
                    else
                    {
                        propertyToNewValues[property].SetValue(currentValues);
                    }
                }
            }

            foreach (CellRange range in context.AffectedCellRanges)
            {
                foreach (KeyValuePair<IProperty, ICompressedList> pair in propertyToNewValues)
                {
                    IProperty property = pair.Key;
                    ICompressedList newValues = pair.Value;

                    int offset = propertyBag.GetRowColumnIndex(range.FromIndex) - propertyBag.GetRowColumnIndex(context.FromCellRange[0].FromIndex);
                    ICompressedList translatedNewValues = newValues.Offset(offset);

                    command.Execute(new SetPropertyValuesCommandContext(context.ToWorksheet, property.PropertyDefinition, translatedNewValues));
                }
            }
        }

        private static void CopyPasteRowColumnProperties(CopyPasteContext context, RowColumnPropertyBagBase propertyBag,
            IEnumerable<IProperty> properties, Func<CellRange, bool> shouldCopyPasteColumn,
            Action<CopyPasteContext, RowColumnPropertyBagBase, ICompressedList, CellRange, IProperty> setRowColumnValuesForCellsAction)
        {
            Dictionary<IProperty, ICompressedList> propertyToNewValues = new Dictionary<IProperty, ICompressedList>();
            foreach (CellRange fromCellRange in context.FromCellRange)
            {
                if (!shouldCopyPasteColumn(fromCellRange))
                {
                    continue;
                }

                foreach (IProperty property in properties)
                {
                    ICompressedList propertyValueCollection = propertyBag.GetPropertyValueCollection(property.PropertyDefinition);
                    ICompressedList currentValues = propertyValueCollection.GetValue(
                        propertyBag.GetRowColumnIndex(fromCellRange.FromIndex), propertyBag.GetRowColumnIndex(fromCellRange.ToIndex));

                    if (!propertyToNewValues.ContainsKey(property))
                    {
                        propertyToNewValues.Add(property, currentValues);
                    }
                    else
                    {
                        propertyToNewValues[property].SetValue(currentValues);
                    }
                }
            }

            foreach (CellRange range in context.AffectedCellRanges)
            {
                foreach (KeyValuePair<IProperty, ICompressedList> pair in propertyToNewValues)
                {
                    IProperty property = pair.Key;
                    ICompressedList newValues = pair.Value;
                    if (context.ToWorksheet.Cells.ContainsPropertyDefinition(property.PropertyDefinition))
                    {
                        setRowColumnValuesForCellsAction(context, propertyBag, newValues, range, property);
                    }
                }
            }
        }

        private static void SetColumnValuesForCells(CopyPasteContext context, RowColumnPropertyBagBase propertyBag,
            ICompressedList newValues, CellRange range, IProperty property)
        {
            if (!newValues.ContainsNonDefaultValues(newValues.FromIndex, newValues.ToIndex))
            {
                return;
            }

            int offset = propertyBag.GetRowColumnIndex(range.FromIndex) - propertyBag.GetRowColumnIndex(context.FromCellRange[0].FromIndex);
            ICompressedList translatedNewValues = newValues.Offset(offset);

            for (int columnIndex = range.FromIndex.ColumnIndex; columnIndex <= range.ToIndex.ColumnIndex; columnIndex++)
            {
                // skip default values
                if (!translatedNewValues.ContainsNonDefaultValues(columnIndex, columnIndex))
                {
                    continue;
                }

                long rangeStart = WorksheetPropertyBagBase.ConvertCellIndexToLong(range.FromIndex.RowIndex, columnIndex);
                long rangeEnd = WorksheetPropertyBagBase.ConvertCellIndexToLong(range.ToIndex.RowIndex, columnIndex);

                ICompressedList cellValues = context.ToWorksheet.Cells.PropertyBag.GetPropertyValueCollection(property.PropertyDefinition).CreateInstance(rangeStart, rangeEnd);

                ICompressedList columnValue = translatedNewValues.GetValue(columnIndex, columnIndex);
                cellValues.SetValue(rangeStart, rangeEnd, columnValue);

                WorkbookCommands.SetCellProperty.Execute(new SetPropertyValuesCommandContext(context.ToWorksheet, property.PropertyDefinition, cellValues));
            }
        }

        private static void SetRowValuesForCells(CopyPasteContext context, RowColumnPropertyBagBase propertyBag,
            ICompressedList newValues, CellRange range, IProperty property)
        {
            // skip default values
            if (!newValues.ContainsNonDefaultValues(newValues.FromIndex, newValues.ToIndex))
            {
                return;
            }

            int offset = propertyBag.GetRowColumnIndex(range.FromIndex) - propertyBag.GetRowColumnIndex(context.FromCellRange[0].FromIndex);
            ICompressedList translatedNewValues = newValues.Offset(offset);

            for (int rowIndex = range.FromIndex.RowIndex; rowIndex <= range.ToIndex.RowIndex; rowIndex++)
            {
                if (!translatedNewValues.ContainsNonDefaultValues(rowIndex, rowIndex))
                {
                    continue;
                }

                for (int columnIndex = range.FromIndex.ColumnIndex; columnIndex <= range.ToIndex.ColumnIndex; columnIndex++)
                {
                    long index = WorksheetPropertyBagBase.ConvertCellIndexToLong(rowIndex, columnIndex);

                    ICompressedList cellValues = context.ToWorksheet.Cells.PropertyBag.GetPropertyValueCollection(property.PropertyDefinition).CreateInstance(index, index);

                    ICompressedList rowValue = translatedNewValues.GetValue(rowIndex, rowIndex);
                    cellValues.SetValue(index, index, rowValue);

                    WorkbookCommands.SetCellProperty.Execute(new SetPropertyValuesCommandContext(context.ToWorksheet, property.PropertyDefinition, cellValues));
                }
            }
        }

        private static void CopyPasteColumnWidth(CopyPasteContext context)
        {
            if (!context.ShouldCopyPasteColumnWidths())
            {
                return;
            }

            ICompressedList<ColumnWidth> newValues = null;
            foreach (CellRange fromCellRange in context.FromCellRange)
            {
                ICompressedList<ColumnWidth> currentNewValues = context.FromWorksheet.Columns.PropertyBag.GetPropertyValue(ColumnsPropertyBag.WidthProperty,
                    fromCellRange.FromIndex.ColumnIndex, fromCellRange.ToIndex.ColumnIndex);

                if (newValues == null)
                {
                    newValues = currentNewValues;
                }
                else
                {
                    newValues.SetValue(currentNewValues);
                }
            }

            foreach (CellRange range in context.AffectedCellRanges)
            {
                var translatedNewValues = newValues.Offset(range.FromIndex.ColumnIndex - context.FromCellRange[0].FromIndex.ColumnIndex);
                WorkbookCommands.SetColumnProperty.Execute(new SetPropertyValuesCommandContext(context.ToWorksheet,
                    ColumnsPropertyBag.WidthProperty, translatedNewValues));
            }
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing,
        /// or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Disposes the specified clean up managed resources.
        /// </summary>
        /// <param name="cleanUpManagedResources">The clean up managed resources.</param>
        protected virtual void Dispose(bool cleanUpManagedResources)
        {
            if (this.alreadyDisposed)
            {
                return;
            }

            if (cleanUpManagedResources)
            {
                if (this.workbook != null)
                {
                    WorkbookPool.ReturnWorkbook(this.workbook);
                }
            }

            this.alreadyDisposed = true;
        }

        private void EnsureNotDisposed()
        {
            if (this.alreadyDisposed)
            {
                throw new ObjectDisposedException("WorksheetFragment");
            }
        }

        #endregion
    }
}