﻿using System;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Layout;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Model
{
    /// <summary>
    /// Represents the visible section of the worksheet.
    /// </summary>
    public class WorksheetViewState : SheetViewStateBase
    {
        #region Fields

        private Size scaleFactor;
        private CellIndex topLeftCellIndex;
        private SelectionState selectionState;
        private Pane pane;
        private bool isSelected;
        private bool showGridLines;
        private bool showRowColHeaders;
        private bool circleInvalidData;

        #endregion


        #region Properties

        /// <summary>
        /// Gets or sets the scale factor.
        /// </summary>
        /// <value>The scale factor.</value>
        public Size ScaleFactor
        {
            get
            {
                return this.scaleFactor;
            }
            set
            {
                this.scaleFactor = SpreadsheetHelper.EnsureValidScaleFactor(value);
            }
        }

        /// <summary>
        /// Gets or sets the index of the top left cell of the screen.
        /// </summary>
        /// <value>The index of the top left cell of the screen.</value>
        public CellIndex TopLeftCellIndex
        {
            get
            {
                return this.topLeftCellIndex;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "value");
                this.topLeftCellIndex = value;
            }
        }

        /// <summary>
        /// Gets or sets the state of the selection.
        /// </summary>
        /// <value>The state of the selection.</value>
        public SelectionState SelectionState
        {
            get
            {
                return this.selectionState;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "value");
                this.selectionState = value;
            }
        }

        /// <summary>
        /// Gets or sets value indicating if the sheet is selected.
        /// </summary>
        /// <value>The value indicating if the sheet is selected.</value>
        public bool IsSelected
        {
            get
            {
                return this.isSelected;
            }
            set
            {
                this.isSelected = value;
            }
        }

        /// <summary>
        /// Gets or sets value indicating if the gridlines should be shown.
        /// </summary>
        /// <value>The value indicating if the gridlines should be shown.</value>
        public bool ShowGridLines
        {
            get
            {
                return this.showGridLines;
            }
            set
            {
                this.showGridLines = value;
            }
        }

        /// <summary>
        /// Gets or sets value indicating if the row and column headings should be shown.
        /// </summary>
        /// <value>The value indicating if the row and column headings should be shown.</value>
        public bool ShowRowColHeaders
        {
            get
            {
                return this.showRowColHeaders;
            }
            set
            {
                this.showRowColHeaders = value;
            }
        }

        /// <summary>
        /// Gets or sets the pane.
        /// </summary>
        /// <value>The pane.</value>
        public Pane Pane
        {
            get
            {
                return this.pane;
            }
            set
            {
                this.pane = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether to circle the invalid data.
        /// </summary>
        /// <value>The value indicating whether to circle the invalid data.</value>
        public bool CircleInvalidData
        {
            get
            {
                return this.circleInvalidData;
            }
            set
            {
                this.circleInvalidData = value;
            }
        }

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="WorksheetViewState" /> class.
        /// </summary>
        public WorksheetViewState()
        {
            this.scaleFactor = new Size(1, 1);
            this.topLeftCellIndex = new CellIndex(0, 0);
            this.selectionState = new SelectionState();
            this.showGridLines = true;
            this.showRowColHeaders = true;
            //this.pane = new Pane(new CellIndex(0, 0), new CellIndex(0, 0), ViewportPaneType.Scrollable);
        }

        #endregion


        #region Methods

        /// <summary>
        /// Freezes the sheet panes.
        /// </summary>
        /// <param name="frozenRowsCount">The number of rows which will be frozen.</param>
        /// <param name="frozenColumnsCount">The number of columns which will be frozen.</param>
        public void FreezePanes(int frozenRowsCount, int frozenColumnsCount)
        {
            this.FreezePanes(this.topLeftCellIndex, frozenRowsCount, frozenColumnsCount);
        }

        /// <summary>
        /// Freezes the sheet panes.
        /// </summary>
        /// <param name="fixedPaneTopLeftCellIndex">The top left cell index of the fixed pane.</param>
        /// <param name="frozenRowsCount">The number of rows which will be frozen.</param>
        /// <param name="frozenColumnsCount">The number of columns which will be frozen.</param>
        public void FreezePanes(CellIndex fixedPaneTopLeftCellIndex, int frozenRowsCount, int frozenColumnsCount)
        {
            this.FreezePanes(fixedPaneTopLeftCellIndex, frozenRowsCount, frozenColumnsCount, null);
        }

        /// <summary>
        /// Freezes the sheet panes.
        /// </summary>
        /// <param name="fixedPaneTopLeftCellIndex">The top left cell index of the fixed pane.</param>
        /// <param name="frozenRowsCount">The number of rows which will be frozen.</param>
        /// <param name="frozenColumnsCount">The number of columns which will be frozen.</param>
        /// <param name="scrollableTopLeftCellIndex">The top left cell index of the scrollable pane.</param>
        public void FreezePanes(CellIndex fixedPaneTopLeftCellIndex, int frozenRowsCount, int frozenColumnsCount, CellIndex scrollableTopLeftCellIndex)
        {
            Guard.ThrowExceptionIfLessThan(0, frozenRowsCount, "frozenRows");
            Guard.ThrowExceptionIfLessThan(0, frozenColumnsCount, "frozenColumns");

            if (scrollableTopLeftCellIndex != null && scrollableTopLeftCellIndex.RowIndex <= fixedPaneTopLeftCellIndex.RowIndex + frozenRowsCount)
            {
                throw new ArgumentException("The top left cell index of the scrollable pane must be within the scrollable pane.");
            }

            if (scrollableTopLeftCellIndex != null && scrollableTopLeftCellIndex.ColumnIndex <= fixedPaneTopLeftCellIndex.ColumnIndex + frozenColumnsCount)
            {
                throw new ArgumentException("The top left cell index of the scrollable pane must be within the scrollable pane.");
            }

            this.TopLeftCellIndex = fixedPaneTopLeftCellIndex;

            if (scrollableTopLeftCellIndex == null)
            {
                scrollableTopLeftCellIndex = new CellIndex(fixedPaneTopLeftCellIndex.RowIndex + frozenRowsCount,
                    fixedPaneTopLeftCellIndex.ColumnIndex + frozenColumnsCount);
            }

            Pane pane;
            if (frozenRowsCount == 0 && frozenColumnsCount == 0)
            {
                pane = null;
            }
            else
            {
                pane = new Pane(scrollableTopLeftCellIndex, frozenColumnsCount, frozenRowsCount, ViewportPaneType.Scrollable);
            }

            this.Pane = pane;
        }

        internal CellIndex GetFrozenCellIndex()
        {
            CellIndex frozenCellIndex = null;

            Pane pane = this.Pane;

            if (pane != null && pane.State == PaneState.Frozen)
            {
                int xSplit = pane.XSplit;
                int ySplit = pane.YSplit;

                int frozenRowIndex = (ySplit == 0) ? 0 : pane.YSplit + this.TopLeftCellIndex.RowIndex;
                int frozenColumnIndex = (xSplit == 0) ? 0 : pane.XSplit + this.TopLeftCellIndex.ColumnIndex;

                frozenCellIndex = new CellIndex(frozenRowIndex, frozenColumnIndex);
            }

            return frozenCellIndex;
        }

        internal WorksheetViewState Clone()
        {
            WorksheetViewState result = new WorksheetViewState();
            result.IsSelected = this.IsSelected;
            result.ScaleFactor = this.ScaleFactor;
            result.TabColor = this.TabColor;
            result.TopLeftCellIndex = this.TopLeftCellIndex;
            result.ShowGridLines = this.ShowGridLines;
            result.ShowRowColHeaders = this.ShowRowColHeaders;
            result.CircleInvalidData = this.CircleInvalidData;

            if (this.Pane != null)
            {
                result.Pane = this.Pane.Clone();
            }

            if (this.SelectionState != null)
            {
                result.SelectionState = this.SelectionState.Clone();
            }

            return result;
        }

        internal void CopyFrom(WorksheetViewState fromWorksheetViewState)
        {
            this.isSelected = fromWorksheetViewState.IsSelected;
            this.scaleFactor = fromWorksheetViewState.ScaleFactor;
            this.topLeftCellIndex = fromWorksheetViewState.TopLeftCellIndex;
            this.showGridLines = fromWorksheetViewState.ShowGridLines;
            this.showRowColHeaders = fromWorksheetViewState.ShowRowColHeaders;
            this.circleInvalidData = fromWorksheetViewState.CircleInvalidData;
            this.TabColor = fromWorksheetViewState.TabColor;

            this.pane = fromWorksheetViewState.Pane != null ? fromWorksheetViewState.Pane.Clone() : null;
            this.selectionState = fromWorksheetViewState.SelectionState != null ? fromWorksheetViewState.SelectionState.Clone() : null;
        }

        /// <summary>
        /// Returns a <see cref="T:System.String" /> that represents the current
        /// <see cref="T:System.Object" />.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String" /> that represents the current <see cref="T:System.Object" />.
        /// </returns>
        public override string ToString()
        {
            return string.Format("ScaleFactor: {0}; TopLeftCellIndex: {1}; SelectionState: {2}",
                this.scaleFactor, this.topLeftCellIndex, this.selectionState);
        }

        #endregion
    }
}