﻿using System.Collections.ObjectModel;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.DataValidation;
#if NETCORE || NETSTANDARD
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.PropertySystem
{
    /// <summary>
    /// Provides the cell property definitions.
    /// </summary>
    public static class CellPropertyDefinitions
    {
        /// <summary>
        /// Gets format property definition.
        /// </summary>
        public static readonly IPropertyDefinition<CellValueFormat> FormatProperty = new PropertyDefinition<CellValueFormat>(
            "Format", true, StylePropertyGroup.Number, CellValueFormat.GeneralFormat, false);

        /// <summary>
        /// Gets value property definition.
        /// </summary>
        public static readonly IPropertyDefinition<ICellValue> ValueProperty = new PropertyDefinition<ICellValue>(
            "CellValue", true, StylePropertyGroup.None, EmptyCellValue.EmptyValue, false);

        /// <summary>
        /// Gets style name property definition.
        /// </summary>
        public static readonly IPropertyDefinition<string> StyleNameProperty = new PropertyDefinition<string>(
            "StyleName", true, StylePropertyGroup.None, SpreadsheetDefaultValues.DefaultStyleName);

        /// <summary>
        /// Gets left border property definition.
        /// </summary>
        public static readonly IPropertyDefinition<CellBorder> LeftBorderProperty = new PropertyDefinition<CellBorder>(
            "LeftBorder", true, StylePropertyGroup.Border, CellBorder.Default);

        /// <summary>
        /// Gets top border property definition.
        /// </summary>
        public static readonly IPropertyDefinition<CellBorder> TopBorderProperty = new PropertyDefinition<CellBorder>(
            "TopBorder", true, StylePropertyGroup.Border, CellBorder.Default);

        /// <summary>
        /// Gets right border property definition.
        /// </summary>
        public static readonly IPropertyDefinition<CellBorder> RightBorderProperty = new PropertyDefinition<CellBorder>(
            "RightBorder", true, StylePropertyGroup.Border, CellBorder.Default);

        /// <summary>
        /// Gets bottom border property definition.
        /// </summary>
        public static readonly IPropertyDefinition<CellBorder> BottomBorderProperty = new PropertyDefinition<CellBorder>(
            "BottomBorder", true, StylePropertyGroup.Border, CellBorder.Default);

        /// <summary>
        /// Gets diagonal up border property definition.
        /// </summary>
        public static readonly IPropertyDefinition<CellBorder> DiagonalUpBorderProperty = new PropertyDefinition<CellBorder>(
            "DiagonalUpBorder", true, StylePropertyGroup.Border, CellBorder.Default);

        /// <summary>
        /// Gets diagonal down border property definition.
        /// </summary>
        public static readonly IPropertyDefinition<CellBorder> DiagonalDownBorderProperty = new PropertyDefinition<CellBorder>(
            "DiagonalDownBorder", true, StylePropertyGroup.Border, CellBorder.Default);

        /// <summary>
        /// Gets fill property definition.
        /// </summary>
        public static readonly IPropertyDefinition<IFill> FillProperty = new PropertyDefinition<IFill>(
            "Fill", true, StylePropertyGroup.Fill, PatternFill.Default);

        /// <summary>
        /// Gets font family property definition.
        /// </summary>
        public static readonly IPropertyDefinition<ThemableFontFamily> FontFamilyProperty = new PropertyDefinition<ThemableFontFamily>(
            "FontFamily", true, StylePropertyGroup.Font, SpreadsheetDefaultValues.DefaultFontFamily);

        /// <summary>
        /// Gets fonts size property definition.
        /// </summary>
        public static readonly IPropertyDefinition<double> FontSizeProperty = new PropertyDefinition<double>(
            "FontSize", true, StylePropertyGroup.Font, SpreadsheetDefaultValues.DefaultFontSize);

        /// <summary>
        /// Gets is bold property definition.
        /// </summary>
        public static readonly IPropertyDefinition<bool> IsBoldProperty = new PropertyDefinition<bool>(
            "IsBold", true, StylePropertyGroup.Font);

        /// <summary>
        /// Gets is italic property definition.
        /// </summary>
        public static readonly IPropertyDefinition<bool> IsItalicProperty = new PropertyDefinition<bool>(
            "IsItalic", true, StylePropertyGroup.Font);

        /// <summary>
        /// Gets underline property definition.
        /// </summary>
        public static readonly IPropertyDefinition<UnderlineType> UnderlineProperty = new PropertyDefinition<UnderlineType>(
            "Underline", true, StylePropertyGroup.Font);

        /// <summary>
        /// Gets fore color property definition.
        /// </summary>
        public static readonly IPropertyDefinition<ThemableColor> ForeColorProperty = new PropertyDefinition<ThemableColor>(
            "ForeColor", true, StylePropertyGroup.Font, SpreadsheetDefaultValues.DefaultForeColor);

        /// <summary>
        /// Gets horizontal alignment property definition.
        /// </summary>
        public static readonly IPropertyDefinition<RadHorizontalAlignment> HorizontalAlignmentProperty = new PropertyDefinition<RadHorizontalAlignment>(
            "HorizontalAlignment", true, StylePropertyGroup.Alignment);

        /// <summary>
        /// Gets vertical alignment property definition.
        /// </summary>
        public static readonly IPropertyDefinition<RadVerticalAlignment> VerticalAlignmentProperty = new PropertyDefinition<RadVerticalAlignment>(
            "VerticalAlignment", true, StylePropertyGroup.Alignment, RadVerticalAlignment.Bottom);

        /// <summary>
        /// Gets indent property definition.
        /// </summary>
        public static readonly IPropertyDefinition<int> IndentProperty = new PropertyDefinition<int>(
            "Indent", true, StylePropertyGroup.Alignment);

        /// <summary>
        /// Gets is wrapped property definition.
        /// </summary>
        public static readonly IPropertyDefinition<bool> IsWrappedProperty = new PropertyDefinition<bool>(
            "IsWrapped", true, StylePropertyGroup.Alignment);

        /// <summary>
        /// Gets is locked property definition.
        /// </summary>
        public static readonly IPropertyDefinition<bool> IsLockedProperty = new PropertyDefinition<bool>(
            "IsLocked", false, StylePropertyGroup.Protection, true);

        /// <summary>
        /// Gets data validation property definition.
        /// </summary>
        public static readonly IPropertyDefinition<IDataValidationRule> DataValidationRuleProperty = new PropertyDefinition<IDataValidationRule>(
            "DataValidationRule", true, StylePropertyGroup.None, new AnyValueDataValidationRule(new AnyValueDataValidationRuleContext()));

        /// <summary>
        /// Gets all property definitions.
        /// </summary>
        public static readonly ReadOnlyCollection<IPropertyDefinition> AllPropertyDefinitions = new ReadOnlyCollection<IPropertyDefinition>(new IPropertyDefinition[]
        {
            FormatProperty,
            ValueProperty,
            StyleNameProperty,
            LeftBorderProperty,
            TopBorderProperty,
            RightBorderProperty,
            BottomBorderProperty,
            DiagonalUpBorderProperty,
            DiagonalDownBorderProperty,
            FillProperty,
            FontFamilyProperty,
            FontSizeProperty,
            IsBoldProperty,
            IsItalicProperty,
            UnderlineProperty,
            ForeColorProperty,
            HorizontalAlignmentProperty,
            VerticalAlignmentProperty,
            IndentProperty,
            IsWrappedProperty,
            IsLockedProperty,
            DataValidationRuleProperty
        });
    }
}