﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorkbookCommands;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
#if NETCORE || NETSTANDARD
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.PropertySystem
{
    /// <summary>
    /// Represents cell style.
    /// </summary>
    public class CellStyle
    {
        #region Constants

        private static readonly Dictionary<IPropertyDefinition, Func<CellStyle, StylePropertyBase>> propertyDefinitionToStyleProperty;
        private readonly Dictionary<StylePropertyGroup, bool> stylePropertyGroupToIsIncluded;

        #endregion


        #region Fields

        private readonly Workbook workbook;
        private readonly string name;
        private readonly CellStyleCategory category;
        private readonly bool isRemovable;
        private readonly int? builtInId;
        private int beginEndUpdateCount;
        private int suspendForcingLayoutCount;
        private readonly HashSet<IPropertyDefinition> changedProperties;

        private readonly StyleProperty<CellBorder> leftBorderProperty;
        private readonly StyleProperty<CellBorder> topBorderProperty;
        private readonly StyleProperty<CellBorder> rightBorderProperty;
        private readonly StyleProperty<CellBorder> bottomBorderProperty;
        private readonly StyleProperty<CellBorder> diagonalUpBorderProperty;
        private readonly StyleProperty<CellBorder> diagonalDownBorderProperty;
        private readonly StyleProperty<IFill> fillProperty;
        private readonly StyleProperty<ThemableFontFamily> fontFamilyProperty;
        private readonly StyleProperty<double> fontSizeProperty;
        private readonly StyleProperty<bool> isBoldProperty;
        private readonly StyleProperty<bool> isItalicProperty;
        private readonly StyleProperty<UnderlineType> underlineProperty;
        private readonly StyleProperty<ThemableColor> foreColorProperty;
        private readonly StyleProperty<RadHorizontalAlignment> horizontalAlignmentProperty;
        private readonly StyleProperty<RadVerticalAlignment> verticalAlignmentProperty;
        private readonly StyleProperty<int> indentProperty;
        private readonly StyleProperty<bool> isWrappedProperty;
        private readonly StyleProperty<CellValueFormat> formatProperty;
        private readonly StyleProperty<bool> isLockedProperty;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the workbook.
        /// </summary>
        /// <value>The workbook.</value>
        public Workbook Workbook
        {
            get
            {
                return this.workbook;
            }
        }

        /// <summary>
        /// Gets the name.
        /// </summary>
        /// <value>The name of the style.</value>
        public string Name
        {
            get
            {
                return this.name;
            }
        }

        /// <summary>
        /// Gets the category.
        /// </summary>
        /// <value>The category.</value>
        public CellStyleCategory Category
        {
            get
            {
                return this.category;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the style is removable.
        /// </summary>
        /// <value>The value indicating whether the style is removable.</value>
        public bool IsRemovable
        {
            get
            {
                return this.isRemovable;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the style is removed.
        /// </summary>
        /// <value>The value indicating whether the style is removed.</value>
        public bool IsRemoved
        {
            get
            {
                return this.Workbook == null ||
                    this.Workbook.Styles == null ||
                    !this.Workbook.Styles.Contains(this);
            }
        }

        /// <summary>
        /// Gets the built in id.
        /// </summary>
        /// <value>The built in id.</value>
        public int? BuiltInId
        {
            get
            {
                return this.builtInId;
            }
        }

        /// <summary>
        /// Gets a value indicating whether an update is in progress.
        /// </summary>
        /// <value>The value indicating whether an update is in progress.</value>
        public bool IsUpdateInProgress
        {
            get
            {
                return this.beginEndUpdateCount > 0;
            }
        }

        #endregion


        #region Style Properties

        /// <summary>
        /// Gets or sets the left border.
        /// </summary>
        /// <value>The left border.</value>
        public CellBorder LeftBorder
        {
            get
            {
                return this.leftBorderProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.leftBorderProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the top border.
        /// </summary>
        /// <value>The top border.</value>
        public CellBorder TopBorder
        {
            get
            {
                return this.topBorderProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.topBorderProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the right border.
        /// </summary>
        /// <value>The right border.</value>
        public CellBorder RightBorder
        {
            get
            {
                return this.rightBorderProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.rightBorderProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the bottom border.
        /// </summary>
        /// <value>The bottom border.</value>
        public CellBorder BottomBorder
        {
            get
            {
                return this.bottomBorderProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.bottomBorderProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the diagonal up border.
        /// </summary>
        /// <value>The diagonal up border.</value>
        public CellBorder DiagonalUpBorder
        {
            get
            {
                return this.diagonalUpBorderProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.diagonalUpBorderProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the diagonal down border.
        /// </summary>
        /// <value>The diagonal down border.</value>
        public CellBorder DiagonalDownBorder
        {
            get
            {
                return this.diagonalDownBorderProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.diagonalDownBorderProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the fill.
        /// </summary>
        /// <value>The fill value.</value>
        public IFill Fill
        {
            get
            {
                return this.fillProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.fillProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the font family.
        /// </summary>
        /// <value>The font family.</value>
        public ThemableFontFamily FontFamily
        {
            get
            {
                return this.fontFamilyProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.fontFamilyProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the size of the font.
        /// </summary>
        /// <value>The size of the font.</value>
        public double FontSize
        {
            get
            {
                return this.fontSizeProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.fontSizeProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the text affected by the style is bold.
        /// </summary>
        /// <value>The value indicating whether the text affected by the style is bold..</value>
        public bool IsBold
        {
            get
            {
                return this.isBoldProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.isBoldProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the text affected by the style is italic.
        /// </summary>
        /// <value>The value indicating whether the text affected by the style is italic.</value>
        public bool IsItalic
        {
            get
            {
                return this.isItalicProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.isItalicProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the underline.
        /// </summary>
        /// <value>The underline.</value>
        public UnderlineType Underline
        {
            get
            {
                return this.underlineProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.underlineProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the fore color.
        /// </summary>
        /// <value>The fore color.</value>
        public ThemableColor ForeColor
        {
            get
            {
                return this.foreColorProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.foreColorProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the horizontal alignment.
        /// </summary>
        /// <value>The horizontal alignment.</value>
        public RadHorizontalAlignment HorizontalAlignment
        {
            get
            {
                return this.horizontalAlignmentProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.horizontalAlignmentProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the vertical alignment.
        /// </summary>
        /// <value>The vertical alignment.</value>
        public RadVerticalAlignment VerticalAlignment
        {
            get
            {
                return this.verticalAlignmentProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.verticalAlignmentProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the indent.
        /// </summary>
        /// <value>The indent.</value>
        public int Indent
        {
            get
            {
                return this.indentProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.indentProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the text affected by the style is wrapped.
        /// </summary>
        /// <value>The value indicating whether the text affected by the style is wrapped.</value>
        public bool IsWrapped
        {
            get
            {
                return this.isWrappedProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.isWrappedProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets the format.
        /// </summary>
        /// <value>The format.</value>
        public CellValueFormat Format
        {
            get
            {
                return this.formatProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.formatProperty, value);
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the cell affected by the style is locked.
        /// </summary>
        /// <value>The value indicating whether the cell affected by the style is locked.</value>
        public bool IsLocked
        {
            get
            {
                return this.isLockedProperty.GetValue();
            }
            set
            {
                this.SetStyleProperty(this.isLockedProperty, value);
            }
        }

        #endregion


        #region IsIncluded Properties

        /// <summary>
        /// Gets or sets a value indicating whether the number property group is included.
        /// </summary>
        /// <value>The value indicating whether the number property group is included.</value>
        public bool IncludeNumber
        {
            get
            {
                return this.GetIsStylePropertyGroupIncluded(StylePropertyGroup.Number);
            }
            set
            {
                if (this.GetIsStylePropertyGroupIncluded(StylePropertyGroup.Number) != value)
                {
                    this.SetIsStylePropertyGroupIncluded(StylePropertyGroup.Number, value);
                }
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the alignment property group is included.
        /// </summary>
        /// <value>The value indicating whether the alignment property group is included.</value>
        public bool IncludeAlignment
        {
            get
            {
                return this.GetIsStylePropertyGroupIncluded(StylePropertyGroup.Alignment);
            }
            set
            {
                if (this.GetIsStylePropertyGroupIncluded(StylePropertyGroup.Alignment) != value)
                {
                    this.SetIsStylePropertyGroupIncluded(StylePropertyGroup.Alignment, value);
                }
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the font property group is included.
        /// </summary>
        /// <value>The value indicating whether the font property group is included.</value>
        public bool IncludeFont
        {
            get
            {
                return this.GetIsStylePropertyGroupIncluded(StylePropertyGroup.Font);
            }
            set
            {
                if (this.GetIsStylePropertyGroupIncluded(StylePropertyGroup.Font) != value)
                {
                    this.SetIsStylePropertyGroupIncluded(StylePropertyGroup.Font, value);
                }
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the border property group is included.
        /// </summary>
        /// <value>The value indicating whether the border property group is included.</value>
        public bool IncludeBorder
        {
            get
            {
                return this.GetIsStylePropertyGroupIncluded(StylePropertyGroup.Border);
            }
            set
            {
                if (this.GetIsStylePropertyGroupIncluded(StylePropertyGroup.Border) != value)
                {
                    this.SetIsStylePropertyGroupIncluded(StylePropertyGroup.Border, value);
                }
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the fill property group is included.
        /// </summary>
        /// <value>The value indicating whether the fill property group is included.</value>
        public bool IncludeFill
        {
            get
            {
                return this.GetIsStylePropertyGroupIncluded(StylePropertyGroup.Fill);
            }
            set
            {
                if (this.GetIsStylePropertyGroupIncluded(StylePropertyGroup.Fill) != value)
                {
                    this.SetIsStylePropertyGroupIncluded(StylePropertyGroup.Fill, value);
                }
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the protection property group is included.
        /// </summary>
        /// <value>The value indicating whether the protection property group is included.</value>
        public bool IncludeProtection
        {
            get
            {
                return this.GetIsStylePropertyGroupIncluded(StylePropertyGroup.Protection);
            }
            set
            {
                if (this.GetIsStylePropertyGroupIncluded(StylePropertyGroup.Protection) != value)
                {
                    this.SetIsStylePropertyGroupIncluded(StylePropertyGroup.Protection, value);
                }
            }
        }

        #endregion


        #region Constructors

        static CellStyle()
        {
            propertyDefinitionToStyleProperty = new Dictionary<IPropertyDefinition, Func<CellStyle, StylePropertyBase>>();

            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.LeftBorderProperty, (style) => { return style.leftBorderProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.TopBorderProperty, (style) => { return style.topBorderProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.RightBorderProperty, (style) => { return style.rightBorderProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.BottomBorderProperty, (style) => { return style.bottomBorderProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.DiagonalUpBorderProperty, (style) => { return style.diagonalUpBorderProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.DiagonalDownBorderProperty, (style) => { return style.diagonalDownBorderProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.FillProperty, (style) => { return style.fillProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.FontFamilyProperty, (style) => { return style.fontFamilyProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.FontSizeProperty, (style) => { return style.fontSizeProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.IsBoldProperty, (style) => { return style.isBoldProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.IsItalicProperty, (style) => { return style.isItalicProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.UnderlineProperty, (style) => { return style.underlineProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.ForeColorProperty, (style) => { return style.foreColorProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.HorizontalAlignmentProperty, (style) => { return style.horizontalAlignmentProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.VerticalAlignmentProperty, (style) => { return style.verticalAlignmentProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.IndentProperty, (style) => { return style.indentProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.IsWrappedProperty, (style) => { return style.isWrappedProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.FormatProperty, (style) => { return style.formatProperty; });
            propertyDefinitionToStyleProperty.Add(CellPropertyDefinitions.IsLockedProperty, (style) => { return style.isLockedProperty; });
        }

        internal CellStyle(Workbook workbook, string name, CellStyleCategory category, bool isRemoveable = true, int? builtinId = null)
        {
            Guard.ThrowExceptionIfNullOrEmpty(name, "name");

            this.workbook = workbook;
            this.name = name;
            this.category = category;
            this.isRemovable = isRemoveable;
            this.builtInId = builtinId;

            this.changedProperties = new HashSet<IPropertyDefinition>();

            this.leftBorderProperty = new StyleProperty<CellBorder>(this, CellPropertyDefinitions.LeftBorderProperty);
            this.topBorderProperty = new StyleProperty<CellBorder>(this, CellPropertyDefinitions.TopBorderProperty);
            this.rightBorderProperty = new StyleProperty<CellBorder>(this, CellPropertyDefinitions.RightBorderProperty);
            this.bottomBorderProperty = new StyleProperty<CellBorder>(this, CellPropertyDefinitions.BottomBorderProperty);
            this.diagonalUpBorderProperty = new StyleProperty<CellBorder>(this, CellPropertyDefinitions.DiagonalUpBorderProperty);
            this.diagonalDownBorderProperty = new StyleProperty<CellBorder>(this, CellPropertyDefinitions.DiagonalDownBorderProperty);
            this.fillProperty = new StyleProperty<IFill>(this, CellPropertyDefinitions.FillProperty);
            this.fontFamilyProperty = new StyleProperty<ThemableFontFamily>(this, CellPropertyDefinitions.FontFamilyProperty);
            this.fontSizeProperty = new StyleProperty<double>(this, CellPropertyDefinitions.FontSizeProperty);
            this.isBoldProperty = new StyleProperty<bool>(this, CellPropertyDefinitions.IsBoldProperty);
            this.isItalicProperty = new StyleProperty<bool>(this, CellPropertyDefinitions.IsItalicProperty);
            this.underlineProperty = new StyleProperty<UnderlineType>(this, CellPropertyDefinitions.UnderlineProperty);
            this.foreColorProperty = new StyleProperty<ThemableColor>(this, CellPropertyDefinitions.ForeColorProperty);
            this.horizontalAlignmentProperty = new StyleProperty<RadHorizontalAlignment>(this, CellPropertyDefinitions.HorizontalAlignmentProperty);
            this.verticalAlignmentProperty = new StyleProperty<RadVerticalAlignment>(this, CellPropertyDefinitions.VerticalAlignmentProperty);
            this.indentProperty = new StyleProperty<int>(this, CellPropertyDefinitions.IndentProperty);
            this.isWrappedProperty = new StyleProperty<bool>(this, CellPropertyDefinitions.IsWrappedProperty);
            this.formatProperty = new StyleProperty<CellValueFormat>(this, CellPropertyDefinitions.FormatProperty);
            this.isLockedProperty = new StyleProperty<bool>(this, CellPropertyDefinitions.IsLockedProperty);

            this.stylePropertyGroupToIsIncluded = new Dictionary<StylePropertyGroup, bool>();

            foreach (StylePropertyGroup group in Enum.GetValues(typeof(StylePropertyGroup)))
            {
                this.stylePropertyGroupToIsIncluded.Add(group, false);
            }
        }

        #endregion


        #region Methods

        private void SetStyleProperty<T>(StyleProperty<T> property, T value)
        {
            if (this.Workbook != null &&
                this.Workbook.History != null)
            {
                this.Workbook.History.BeginUndoGroup();
            }

            property.SetValue(value);
            this.SetIsStylePropertyGroupIncluded(property.PropertyDefinition.StylePropertyGroup, true);

            if (this.Workbook != null &&
                this.Workbook.History != null)
            {
                this.Workbook.History.EndUndoGroup();
            }
        }

        /// <summary>
        /// Creates temporary style.
        /// </summary>
        /// <returns>The temporary style.</returns>
        [Browsable(false)]
        [EditorBrowsable(EditorBrowsableState.Never)]
        public static CellStyle CreateTempStyle()
        {
            return new CellStyle(null, Guid.NewGuid().ToString(), CellStyleCategory.Custom);
        }

        /// <summary>
        /// Copies the style properties from another style.
        /// </summary>
        /// <param name="fromStyle">From style.</param>
        public void CopyPropertiesFrom(CellStyle fromStyle)
        {
            Guard.ThrowExceptionIfNull(fromStyle, "fromStyle");

            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                foreach (KeyValuePair<IPropertyDefinition, Func<CellStyle, StylePropertyBase>> propertyToStyleProeprty in propertyDefinitionToStyleProperty)
                {
                    Func<CellStyle, StylePropertyBase> styleProperty = propertyToStyleProeprty.Value;
                    object fromStyleValue = styleProperty(fromStyle).GetValueAsObject();
                    styleProperty(this).SetValue(fromStyleValue);
                }

                foreach (StylePropertyGroup stylePropertyGroup in Enum.GetValues(typeof(StylePropertyGroup)))
                {
                    bool isStylePropertyGroupIncluded = fromStyle.GetIsStylePropertyGroupIncluded(stylePropertyGroup);
                    this.SetIsStylePropertyGroupIncluded(stylePropertyGroup, isStylePropertyGroupIncluded);
                }
            }
        }

        /// <summary>
        /// Determines whether the specified other style has same properties.
        /// </summary>
        /// <param name="otherStyle">The other style.</param>
        /// <returns><c>True</c> if has same properties.</returns>
        public bool HasSameProperties(CellStyle otherStyle)
        {
            foreach (KeyValuePair<IPropertyDefinition, Func<CellStyle, StylePropertyBase>> propertyToStyleProeprty in propertyDefinitionToStyleProperty)
            {
                Func<CellStyle, StylePropertyBase> styleProperty = propertyToStyleProeprty.Value;
                if (!TelerikHelper.EqualsOfT(
                    styleProperty(this).GetValueAsObject(),
                    styleProperty(otherStyle).GetValueAsObject()))
                {
                    return false;
                }
            }

            return true;
        }

        /// <summary>
        /// Gets all properties.
        /// </summary>
        /// <returns>All properties.</returns>
        public static IEnumerable<IPropertyDefinition> GetAllProperties()
        {
            return propertyDefinitionToStyleProperty.Keys;
        }

        /// <summary>
        /// Begins the update.
        /// </summary>
        public void BeginUpdate()
        {
            this.beginEndUpdateCount++;
        }

        /// <summary>
        /// Ends the update.
        /// </summary>
        public void EndUpdate()
        {
            if (this.beginEndUpdateCount == 0)
            {
                throw new InvalidOperationException("There is no active update to end.");
            }

            this.beginEndUpdateCount--;
            if (this.beginEndUpdateCount == 0)
            {
                if (this.changedProperties.Count > 0)
                {
                    this.OnChanged();
                    this.changedProperties.Clear();
                }
            }
        }

        internal void SuspendForcingLayoutUpdate()
        {
            this.suspendForcingLayoutCount++;
        }

        internal void ResumeForcingLayoutUpdate()
        {
            if (this.suspendForcingLayoutCount == 0)
            {
                throw new InvalidOperationException("There is no active update to end.");
            }

            this.suspendForcingLayoutCount--;
        }

        /// <summary>
        /// Determines whether the specified property definition is supported by the style.
        /// </summary>
        /// <param name="propertyDefinition">The property definition.</param>
        /// <returns><c>True</c> if supported.</returns>
        public static bool IsSupportedProperty(IPropertyDefinition propertyDefinition)
        {
            Guard.ThrowExceptionIfNull(propertyDefinition, "propertyDefinition");

            return propertyDefinitionToStyleProperty.ContainsKey(propertyDefinition);
        }

        /// <summary>
        /// Determines whether a value has been set for the specified property.
        /// </summary>
        /// <typeparam name="T">The type of the T.</typeparam>
        /// <param name="propertyDefinition">The property definition.</param>
        /// <returns><c>True</c> if has been set.</returns>
        public bool IsPropertyValueSet<T>(IPropertyDefinition<T> propertyDefinition)
        {
            Guard.ThrowExceptionIfNull(propertyDefinition, "propertyDefinition");

            StylePropertyBase property = this.GetStylePropertyFromPropertyDefinition(propertyDefinition);

            return ((StyleProperty<T>)property).IsValueSet;
        }

        /// <summary>
        /// Gets a value indicating if the property is included in the style.
        /// </summary>
        /// <param name="propertyDefinition">The property definition.</param>
        /// <returns>The value indicating if the property is included in the style.</returns>
        public bool GetIsPropertyIncluded(IPropertyDefinition propertyDefinition)
        {
            return this.stylePropertyGroupToIsIncluded[propertyDefinition.StylePropertyGroup];
        }

        private StylePropertyBase GetStylePropertyFromPropertyDefinition(IPropertyDefinition propertyDefinition)
        {
            Func<CellStyle, StylePropertyBase> styleProperty;
            if (!propertyDefinitionToStyleProperty.TryGetValue(propertyDefinition, out styleProperty))
            {
                throw new InvalidOperationException(string.Format("The {0} property is not supported in styles",
                    propertyDefinition.Name));
            }

            StylePropertyBase property = styleProperty(this);
            return property;
        }

        /// <summary>
        /// Gets a value indicating whether the property group is included in cell style.
        /// </summary>
        /// <param name="stylePropertyGroup">The style property group.</param>
        /// <returns>The value indicating whether the property group is included in cell style.</returns>
        public bool GetIsStylePropertyGroupIncluded(StylePropertyGroup stylePropertyGroup)
        {
            bool result = this.stylePropertyGroupToIsIncluded[stylePropertyGroup];
            return result;
        }

        /// <summary>
        /// Sets a value indicating whether the property group is included in cell style.
        /// </summary>
        /// <param name="stylePropertyGroup">The style property group.</param>
        /// <param name="value">The value indicating whether the property group is included in cell style.</param>
        public void SetIsStylePropertyGroupIncluded(StylePropertyGroup stylePropertyGroup, bool value)
        {
            if (!this.IsRemoved &&
                this.suspendForcingLayoutCount == 0)
            {
                using (new UpdateScope(this.Workbook.History.BeginUndoGroup, this.Workbook.History.EndUndoGroup))
                {
                    if (!TelerikHelper.EqualsOfT(this.GetIsStylePropertyGroupIncluded(stylePropertyGroup), value))
                    {
                        SetStylePropertyGroupIncludedCommandContext setStylePropertyIncludedCommandContext =
                            new SetStylePropertyGroupIncludedCommandContext(
                                this.Workbook,
                                this,
                                stylePropertyGroup,
                                this.GetIsStylePropertyGroupIncluded(stylePropertyGroup),
                                value);
                        this.Workbook.ExecuteCommand(WorkbookCommands.SetStylePropertyIsIncluded, setStylePropertyIncludedCommandContext);

                        if (this.Name != SpreadsheetDefaultValues.DefaultStyleName)
                        {
                            ReapplyCellStyleCommandContext reapplyCellStyleCommandContext = new ReapplyCellStyleCommandContext(this.Workbook, this.Name);
                            this.Workbook.ExecuteCommand(WorkbookCommands.ReapplyCellStyleCommand, reapplyCellStyleCommandContext);
                        }

                        if (!value)
                        {
                            this.ClearLocalPropertiesThatAreNotNeeded(stylePropertyGroup);
                        }
                    }
                }
            }
            else
            {
                this.SetIsStylePropertyGroupIncludedInternal(stylePropertyGroup, value);
            }
        }

        private void ClearLocalPropertiesThatAreNotNeeded(StylePropertyGroup stylePropertyGroup)
        {
            if (stylePropertyGroup == StylePropertyGroup.None)
            {
                return;
            }

            using (new UpdateScope(this.workbook.History.BeginUndoGroup, this.workbook.History.EndUndoGroup))
            {
                IEnumerable<IPropertyDefinition> propertyDefinitions = this.GetPropertyGroupProperties(stylePropertyGroup);

                foreach (Worksheet sheet in this.workbook.Worksheets)
                {
                    Core.DataStructures.ICompressedList<string> styleNamePropertyCollection = sheet.Cells.PropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.StyleNameProperty);
                    IEnumerable<Core.DataStructures.Range<long, string>> ranges = styleNamePropertyCollection.Where(x => x.Value == this.Name);

                    foreach (Core.DataStructures.Range<long, string> range in ranges)
                    {
                        CellRange cellRange = CellsPropertyBag.ConvertLongCellRangeToCellRange(range.Start, range.End);

                        foreach (IPropertyDefinition propertyDefinition in propertyDefinitions)
                        {
                            IProperty property = this.GetCellPropertyFromPropertyDefinition(sheet, propertyDefinition);

                            if (property != null)
                            {
                                sheet.Cells[cellRange].ClearPropertyValue(property);
                            }
                        }
                    }
                }
            }
        }

        private IProperty GetCellPropertyFromPropertyDefinition(Worksheet sheet, IPropertyDefinition propertyDefinition)
        {
            IProperty property = (from p in sheet.Cells.Properties
                                  where p.PropertyDefinition == propertyDefinition
                                  select p).FirstOrDefault();

            return property;
        }

        internal void SetIsStylePropertyGroupIncludedInternal(StylePropertyGroup stylePropertyGroup, bool value)
        {
            if (this.stylePropertyGroupToIsIncluded[stylePropertyGroup] == value)
            {
                return;
            }

            this.stylePropertyGroupToIsIncluded[stylePropertyGroup] = value;
            IEnumerable<IPropertyDefinition> propertyDefinitions = this.GetPropertyGroupProperties(stylePropertyGroup);

            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                foreach (IPropertyDefinition property in propertyDefinitions)
                {
                    this.OnPropertyChanged(property);
                }
            }
        }

        private IEnumerable<IPropertyDefinition> GetPropertyGroupProperties(StylePropertyGroup stylePropertyGroup)
        {
            return CellPropertyDefinitions.AllPropertyDefinitions.Where(x => x.StylePropertyGroup == stylePropertyGroup);
        }

        /// <summary>
        /// Gets the property value.
        /// </summary>
        /// <typeparam name="T">The type of the T.</typeparam>
        /// <param name="propertyDefinition">The property definition.</param>
        /// <returns>The property value.</returns>
        public T GetPropertyValue<T>(IPropertyDefinition<T> propertyDefinition)
        {
            Guard.ThrowExceptionIfNull(propertyDefinition, "propertyDefinition");

            Func<CellStyle, StylePropertyBase> styleProperty;
            if (!propertyDefinitionToStyleProperty.TryGetValue(propertyDefinition, out styleProperty))
            {
                throw new InvalidOperationException(string.Format("The {0} property is not supported in styles",
                    propertyDefinition.Name));
            }

            return ((StyleProperty<T>)styleProperty(this)).GetValue();
        }

        internal void OnPropertyChanged(IPropertyDefinition proeprtyDefinition)
        {
            using (new UpdateScope(this.BeginUpdate, this.EndUpdate))
            {
                this.changedProperties.Add(proeprtyDefinition);
            }
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The object to compare with the current object.</param>
        /// <returns>
        /// <c>True</c> if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            CellStyle other = obj as CellStyle;

            if (other == null)
            {
                return false;
            }

            return this.Name.Equals(other.Name) &&
                this.Category.Equals(other.Category) &&
                this.LeftBorder.Equals(other.LeftBorder) &&
                this.TopBorder.Equals(other.TopBorder) &&
                this.RightBorder.Equals(other.RightBorder) &&
                this.BottomBorder.Equals(other.BottomBorder) &&
                this.DiagonalUpBorder.Equals(other.DiagonalUpBorder) &&
                this.DiagonalDownBorder.Equals(other.DiagonalDownBorder) &&
                this.Fill.Equals(other.Fill) &&
                this.FontFamily.Equals(other.FontFamily) &&
                this.FontSize.Equals(other.FontSize) &&
                this.IsBold.Equals(other.IsBold) &&
                this.IsItalic.Equals(other.IsItalic) &&
                this.Underline.Equals(other.Underline) &&
                this.ForeColor.Equals(other.ForeColor) &&
                this.HorizontalAlignment.Equals(other.HorizontalAlignment) &&
                this.VerticalAlignment.Equals(other.VerticalAlignment) &&
                this.Indent.Equals(other.Indent) &&
                this.IsWrapped.Equals(other.IsWrapped) &&
                this.IsLocked.Equals(other.IsLocked) &&
                this.IncludeNumber.Equals(other.IncludeNumber) &&
                this.IncludeAlignment.Equals(other.IncludeAlignment) &&
                this.IncludeFont.Equals(other.IncludeFont) &&
                this.IncludeBorder.Equals(other.IncludeBorder) &&
                this.IncludeFill.Equals(other.IncludeFill) &&
                this.IncludeProtection.Equals(other.IncludeProtection);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return TelerikHelper.CombineHashCodes(
                TelerikHelper.CombineHashCodes(this.Name.GetHashCode(), this.Category.GetHashCode(), this.LeftBorder.GetHashCode(),
                    this.TopBorder.GetHashCode(), this.RightBorder.GetHashCode(), this.BottomBorder.GetHashCode(),
                    this.DiagonalUpBorder.GetHashCode(), this.DiagonalDownBorder.GetHashCode(), this.Fill.GetHashCode()),

                TelerikHelper.CombineHashCodes(this.FontFamily.GetHashCode(), this.FontSize.GetHashCode(), this.IsBold.GetHashCode(),
                    this.IsItalic.GetHashCode(), this.Underline.GetHashCode(), this.ForeColor.GetHashCode(),
                    this.HorizontalAlignment.GetHashCode(), this.VerticalAlignment.GetHashCode(), this.Indent.GetHashCode()),

                TelerikHelper.CombineHashCodes(this.IsWrapped.GetHashCode(), this.IsLocked.GetHashCode(), this.IncludeNumber.GetHashCode(),
                    this.IncludeAlignment.GetHashCode(), this.IncludeFont.GetHashCode(), this.IncludeBorder.GetHashCode(),
                    this.IncludeFill.GetHashCode(), this.IncludeProtection.GetHashCode()));
        }

        #endregion


        #region Events

        /// <summary>
        /// Occurs when the style is changed.
        /// </summary>
        public event EventHandler<StyleChangedEventArgs> Changed;

        /// <summary>
        /// Called when the style is changed.
        /// </summary>
        protected virtual void OnChanged()
        {
            Guard.ThrowExceptionIfTrue(this.IsUpdateInProgress, "IsUpdateInProgress");
            if (this.Changed != null)
            {
                this.Changed(this, new StyleChangedEventArgs(this.changedProperties));
            }
        }

        #endregion
    }
}