﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Windows.Media;
using Telerik.Windows.Documents.Spreadsheet.Commands.WorkbookCommands;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Model;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Theming;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.PropertySystem
{
    /// <summary>
    /// Represents cell style collection.
    /// </summary>
    public class CellStyleCollection : IEnumerable<CellStyle>
    {
        #region Constants

        private static readonly Dictionary<string, int> StyleToBuiltinId;

        #endregion


        #region Fields

        private readonly Workbook workbook;
        private readonly Dictionary<string, CellStyle> nameToStyles;

        #endregion


        #region Properties

        /// <summary>
        /// Gets the workbook.
        /// </summary>
        /// <value>The workbook.</value>
        public Workbook Workbook
        {
            get
            {
                return this.workbook;
            }
        }

        /// <summary>
        /// Gets the count of the styles.
        /// </summary>
        /// <value>The count of the styles.</value>
        public int Count
        {
            get
            {
                return this.nameToStyles.Keys.Count;
            }
        }

        #endregion


        #region Indexers

        /// <summary>
        /// Gets the <see cref="Telerik.Windows.Documents.Spreadsheet.PropertySystem.CellStyle" /> with the specified style name.
        /// </summary>
        /// <param name="styleName">The name of the desired style.</param>
        /// <returns>The style.</returns>
        public CellStyle this[string styleName]
        {
            get
            {
                return this.GetByName(styleName);
            }
        }

        #endregion


        #region Constructors

        static CellStyleCollection()
        {
            StyleToBuiltinId = new Dictionary<string, int>();
            RegisterStyleBuiltinId(SpreadsheetDefaultValues.DefaultStyleName, 0);
            RegisterStyleBuiltinId("Comma", 3);
            RegisterStyleBuiltinId("Currency", 4);
            RegisterStyleBuiltinId("Percent", 5);
            RegisterStyleBuiltinId("Comma [0]", 6);
            RegisterStyleBuiltinId("Currency [0]", 7);
            RegisterStyleBuiltinId(SpreadsheetDefaultValues.HyperlinkStyleName, 8);
            RegisterStyleBuiltinId("Followed Hyperlink", 9);
            RegisterStyleBuiltinId("Note", 10);
            RegisterStyleBuiltinId("Warning Text", 11);
            RegisterStyleBuiltinId("Title", 15);
            RegisterStyleBuiltinId("Heading 1", 16);
            RegisterStyleBuiltinId("Heading 2", 17);
            RegisterStyleBuiltinId("Heading 3", 18);
            RegisterStyleBuiltinId("Heading 4", 19);
            RegisterStyleBuiltinId("Input", 20);
            RegisterStyleBuiltinId("Output", 21);
            RegisterStyleBuiltinId("Calculation", 22);
            RegisterStyleBuiltinId("Check Cell", 23);
            RegisterStyleBuiltinId("Linked Cell", 24);
            RegisterStyleBuiltinId("Total", 25);
            RegisterStyleBuiltinId("Good", 26);
            RegisterStyleBuiltinId("Bad", 27);
            RegisterStyleBuiltinId("Neutral", 28);

            RegisterStyleBuiltinId("Accent1", 29);
            RegisterStyleBuiltinId("20% - Accent1", 30);
            RegisterStyleBuiltinId("40% - Accent1", 31);
            RegisterStyleBuiltinId("60% - Accent1", 32);

            RegisterStyleBuiltinId("Accent2", 33);
            RegisterStyleBuiltinId("20% - Accent2", 34);
            RegisterStyleBuiltinId("40% - Accent2", 35);
            RegisterStyleBuiltinId("60% - Accent2", 36);

            RegisterStyleBuiltinId("Accent3", 37);
            RegisterStyleBuiltinId("20% - Accent3", 38);
            RegisterStyleBuiltinId("40% - Accent3", 39);
            RegisterStyleBuiltinId("60% - Accent3", 40);

            RegisterStyleBuiltinId("Accent4", 41);
            RegisterStyleBuiltinId("20% - Accent4", 42);
            RegisterStyleBuiltinId("40% - Accent4", 43);
            RegisterStyleBuiltinId("60% - Accent4", 44);

            RegisterStyleBuiltinId("Accent5", 45);
            RegisterStyleBuiltinId("20% - Accent5", 46);
            RegisterStyleBuiltinId("40% - Accent5", 47);
            RegisterStyleBuiltinId("60% - Accent5", 48);

            RegisterStyleBuiltinId("Accent6", 49);
            RegisterStyleBuiltinId("20% - Accent6", 50);
            RegisterStyleBuiltinId("40% - Accent6", 51);
            RegisterStyleBuiltinId("60% - Accent6", 52);
            RegisterStyleBuiltinId("Explanatory Text", 53);
        }

        internal CellStyleCollection(Workbook workbook)
        {
            Guard.ThrowExceptionIfNull(workbook, "cells");

            this.workbook = workbook;
            this.nameToStyles = new Dictionary<string, CellStyle>();

            this.InitDefaultStyles();
        }

        #endregion


        #region Methods

        private static void RegisterStyleBuiltinId(string styleName, int builtinId)
        {
            StyleToBuiltinId.Add(styleName, builtinId);
        }

        private void InitDefaultStyles()
        {
            this.AddGoodBadAndNeutralStyles();
            this.AddDataAndModelStyles();
            this.AddTitlesAndHeadingsStyles();
            this.AddThemedStyles();
            this.AddNumberFormatStyles();
        }

        private static int? GetStyleBuiltinId(string styleName)
        {
            int id;
            if (!StyleToBuiltinId.TryGetValue(styleName, out id))
            {
                return null;
            }

            return id;
        }

        private CellStyle CreateCellStyle(Workbook workbook, string styleName, CellStyleCategory category, bool isRemoveable = true)
        {
            int? builtinId = GetStyleBuiltinId(styleName);
            return new CellStyle(workbook, styleName, category, isRemoveable, builtinId);
        }

        private void AddGoodBadAndNeutralStyles()
        {
            CellStyle normalStyle = this.CreateCellStyle(this.Workbook, SpreadsheetDefaultValues.DefaultStyleName, CellStyleCategory.GoodBadAndNeutral, false);
            normalStyle.IncludeAlignment = true;
            normalStyle.IncludeBorder = true;
            normalStyle.IncludeFill = true;
            normalStyle.IncludeFont = true;
            normalStyle.IncludeNumber = true;
            normalStyle.IncludeProtection = true;
            this.AddInternal(normalStyle);

            CellStyle badStyle = this.CreateCellStyle(this.Workbook, "Bad", CellStyleCategory.GoodBadAndNeutral);
            badStyle.ForeColor = ThemableColor.FromArgb(255, 156, 0, 6);
            badStyle.Fill = PatternFill.CreateSolidFill(Color.FromArgb(255, 255, 199, 206));
            badStyle.IncludeFont = true;
            badStyle.IncludeFill = true;
            this.AddInternal(badStyle);

            CellStyle goodStyle = this.CreateCellStyle(this.Workbook, "Good", CellStyleCategory.GoodBadAndNeutral);
            goodStyle.ForeColor = ThemableColor.FromArgb(255, 0, 97, 0);
            goodStyle.Fill = PatternFill.CreateSolidFill(Color.FromArgb(255, 198, 239, 206));
            goodStyle.IncludeFont = true;
            goodStyle.IncludeFill = true;
            this.AddInternal(goodStyle);

            CellStyle neutralStyle = this.CreateCellStyle(this.Workbook, "Neutral", CellStyleCategory.GoodBadAndNeutral);
            neutralStyle.ForeColor = ThemableColor.FromArgb(255, 156, 101, 0);
            neutralStyle.Fill = PatternFill.CreateSolidFill(Color.FromArgb(255, 255, 235, 156));
            neutralStyle.IncludeFont = true;
            neutralStyle.IncludeFill = true;
            this.AddInternal(neutralStyle);
        }

        private void AddDataAndModelStyles()
        {
            CellStyle calculationStyle = this.CreateCellStyle(this.Workbook, "Calculation", CellStyleCategory.DataAndModel);
            calculationStyle.ForeColor = ThemableColor.FromArgb(255, 250, 125, 0);
            CellBorder cellBorder = new CellBorder(CellBorderStyle.Thin, new ThemableColor(Color.FromArgb(255, 127, 127, 127)));
            calculationStyle.LeftBorder = cellBorder;
            calculationStyle.TopBorder = cellBorder;
            calculationStyle.RightBorder = cellBorder;
            calculationStyle.BottomBorder = cellBorder;
            calculationStyle.Fill = PatternFill.CreateSolidFill(Color.FromArgb(255, 242, 242, 242));
            calculationStyle.IncludeFont = true;
            calculationStyle.IncludeBorder = true;
            calculationStyle.IncludeFill = true;
            this.AddInternal(calculationStyle);

            CellStyle checkCellStyle = this.CreateCellStyle(this.Workbook, "Check Cell", CellStyleCategory.DataAndModel);
            checkCellStyle.ForeColor = new ThemableColor(ThemeColorType.Background1);
            // TODO:
            CellBorder checkCellBorder = new CellBorder(CellBorderStyle.Thick, new ThemableColor(Color.FromArgb(255, 127, 127, 127)));//new CellBorder(CellBorderStyle.DoubleLine, Color.FromArgb(255, 127, 127, 127));
            checkCellStyle.LeftBorder = checkCellBorder;
            checkCellStyle.TopBorder = checkCellBorder;
            checkCellStyle.RightBorder = checkCellBorder;
            checkCellStyle.BottomBorder = checkCellBorder;
            checkCellStyle.Fill = PatternFill.CreateSolidFill(Color.FromArgb(255, 165, 165, 165));
            checkCellStyle.IncludeFont = true;
            checkCellStyle.IncludeBorder = true;
            checkCellStyle.IncludeFill = true;

            this.AddInternal(checkCellStyle);

            CellStyle explanatoryStyle = this.CreateCellStyle(this.Workbook, "Explanatory Text", CellStyleCategory.DataAndModel);
            explanatoryStyle.ForeColor = ThemableColor.FromArgb(255, 127, 127, 127);
            explanatoryStyle.IsItalic = true;
            explanatoryStyle.IncludeFont = true;
            this.AddInternal(explanatoryStyle);

            CellStyle followedHyperlink = this.CreateCellStyle(this.Workbook, "Followed Hyperlink", CellStyleCategory.DataAndModel);
            followedHyperlink.FontFamily = new ThemableFontFamily(ThemeFontType.Minor);
            followedHyperlink.Underline = UnderlineType.Single;
            followedHyperlink.ForeColor = new ThemableColor(ThemeColorType.FollowedHyperlink);
            followedHyperlink.IncludeFont = true;
            this.AddInternal(followedHyperlink);

            CellStyle hyperlink = this.CreateCellStyle(this.Workbook, SpreadsheetDefaultValues.HyperlinkStyleName, CellStyleCategory.DataAndModel);
            hyperlink.FontFamily = new ThemableFontFamily(ThemeFontType.Minor);
            hyperlink.Underline = UnderlineType.Single;
            hyperlink.ForeColor = new ThemableColor(ThemeColorType.Hyperlink);
            hyperlink.IncludeFont = true;
            this.AddInternal(hyperlink);

            CellStyle inputStyle = this.CreateCellStyle(this.Workbook, "Input", CellStyleCategory.DataAndModel);
            inputStyle.ForeColor = ThemableColor.FromArgb(255, 63, 63, 118);
            CellBorder inputCellBorder = new CellBorder(CellBorderStyle.Thin, new ThemableColor(Color.FromArgb(255, 127, 127, 127)));
            inputStyle.LeftBorder = inputCellBorder;
            inputStyle.TopBorder = inputCellBorder;
            inputStyle.RightBorder = inputCellBorder;
            inputStyle.BottomBorder = inputCellBorder;
            inputStyle.Fill = PatternFill.CreateSolidFill(Color.FromArgb(255, 255, 204, 153));
            inputStyle.IncludeFont = true;
            inputStyle.IncludeBorder = true;
            inputStyle.IncludeFill = true;
            this.AddInternal(inputStyle);

            CellStyle linkedCellStyle = this.CreateCellStyle(this.Workbook, "Linked Cell", CellStyleCategory.DataAndModel);
            linkedCellStyle.ForeColor = ThemableColor.FromArgb(255, 250, 125, 0);
            // TODO:
            CellBorder linkedCellBorder = new CellBorder(CellBorderStyle.Thick, new ThemableColor(Color.FromArgb(255, 255, 128, 1))); //new CellBorder(CellBorderStyle.DoubleLine, Color.FromArgb(255, 255, 128, 1));
            linkedCellStyle.BottomBorder = linkedCellBorder;
            linkedCellStyle.IncludeFont = true;
            linkedCellStyle.IncludeBorder = true;
            this.AddInternal(linkedCellStyle);

            CellStyle noteStyle = this.CreateCellStyle(this.Workbook, "Note", CellStyleCategory.DataAndModel);
            // TODO:
            CellBorder noteBorder = new CellBorder(CellBorderStyle.Thick, new ThemableColor(Color.FromArgb(255, 178, 178, 178))); //new CellBorder(CellBorderStyle.DoubleLine, Color.FromArgb(255, 178, 178, 178));
            noteStyle.LeftBorder = noteBorder;
            noteStyle.TopBorder = noteBorder;
            noteStyle.RightBorder = noteBorder;
            noteStyle.BottomBorder = noteBorder;
            noteStyle.Fill = PatternFill.CreateSolidFill(Color.FromArgb(255, 255, 255, 204));
            noteStyle.IncludeBorder = true;
            noteStyle.IncludeFill = true;
            this.AddInternal(noteStyle);

            CellStyle outputStyle = this.CreateCellStyle(this.Workbook, "Output", CellStyleCategory.DataAndModel);
            outputStyle.ForeColor = (ThemableColor)Color.FromArgb(255, 63, 63, 63);
            outputStyle.IsBold = true;
            CellBorder outputBorder = new CellBorder(CellBorderStyle.Thin, new ThemableColor(Color.FromArgb(255, 63, 63, 63)));
            outputStyle.LeftBorder = outputBorder;
            outputStyle.TopBorder = outputBorder;
            outputStyle.RightBorder = outputBorder;
            outputStyle.BottomBorder = outputBorder;
            outputStyle.Fill = PatternFill.CreateSolidFill(Color.FromArgb(255, 242, 242, 242));
            outputStyle.IncludeFont = true;
            outputStyle.IncludeBorder = true;
            outputStyle.IncludeFill = true;
            this.AddInternal(outputStyle);

            CellStyle warningTextStyle = this.CreateCellStyle(this.Workbook, "Warning Text", CellStyleCategory.DataAndModel);
            warningTextStyle.ForeColor = (ThemableColor)Colors.Red;
            warningTextStyle.IncludeFont = true;
            this.AddInternal(warningTextStyle);
        }

        private void AddTitlesAndHeadingsStyles()
        {
            CellStyle heading1Style = this.CreateCellStyle(this.Workbook, "Heading 1", CellStyleCategory.TitlesAndHeadings);
            heading1Style.IsBold = true;
            heading1Style.FontSize = UnitHelper.PointToDip(15);
            heading1Style.ForeColor = new ThemableColor(ThemeColorType.Text2);
            CellBorder heading1Border = new CellBorder(CellBorderStyle.Thick, new ThemableColor(ThemeColorType.Accent1));
            heading1Style.BottomBorder = heading1Border;
            heading1Style.IncludeFont = true;
            heading1Style.IncludeBorder = true;
            this.AddInternal(heading1Style);

            CellStyle heading2Style = this.CreateCellStyle(this.Workbook, "Heading 2", CellStyleCategory.TitlesAndHeadings);
            heading2Style.IsBold = true;
            heading2Style.FontSize = UnitHelper.PointToDip(13);
            heading2Style.ForeColor = new ThemableColor(ThemeColorType.Text2);
            CellBorder heading2Border = new CellBorder(CellBorderStyle.Thick, new ThemableColor(ThemeColorType.Accent1, ColorShadeType.Shade2));
            heading2Style.BottomBorder = heading2Border;
            heading2Style.IncludeFont = true;
            heading2Style.IncludeBorder = true;
            this.AddInternal(heading2Style);

            CellStyle heading3Style = this.CreateCellStyle(this.Workbook, "Heading 3", CellStyleCategory.TitlesAndHeadings);
            heading3Style.IsBold = true;
            heading3Style.ForeColor = new ThemableColor(ThemeColorType.Text2);
            CellBorder heading3Border = new CellBorder(CellBorderStyle.Medium, new ThemableColor(ThemeColorType.Accent1, ColorShadeType.Shade2));
            heading3Style.BottomBorder = heading3Border;
            heading3Style.IncludeFont = true;
            heading3Style.IncludeBorder = true;
            this.AddInternal(heading3Style);

            CellStyle heading4Style = this.CreateCellStyle(this.Workbook, "Heading 4", CellStyleCategory.TitlesAndHeadings);
            heading4Style.IsBold = true;
            heading4Style.ForeColor = new ThemableColor(ThemeColorType.Text2);
            heading4Style.IncludeFont = true;
            this.AddInternal(heading4Style);

            CellStyle titleStyle = this.CreateCellStyle(this.Workbook, "Title", CellStyleCategory.TitlesAndHeadings);
            titleStyle.FontFamily = new ThemableFontFamily(ThemeFontType.Major);
            titleStyle.FontSize = UnitHelper.PointToDip(18);
            titleStyle.IsBold = true;
            titleStyle.ForeColor = new ThemableColor(ThemeColorType.Text2);
            titleStyle.IncludeFont = true;
            this.AddInternal(titleStyle);

            CellStyle totalStyle = this.CreateCellStyle(this.Workbook, "Total", CellStyleCategory.TitlesAndHeadings);
            totalStyle.IsBold = true;
            totalStyle.ForeColor = new ThemableColor(ThemeColorType.Text1);
            CellBorder totalStyleTopBorder = new CellBorder(CellBorderStyle.Thin, new ThemableColor(ThemeColorType.Accent1, ColorShadeType.Shade3));
            totalStyle.TopBorder = totalStyleTopBorder;
            // TODO:
            CellBorder totalStyleBottomBorder = new CellBorder(CellBorderStyle.Thick, new ThemableColor(ThemeColorType.Accent1, ColorShadeType.Shade3)); //new CellBorder(CellBorderStyle.DoubleLine, new ThemableColor(ThemeColorType.Accent1, 2));
            totalStyle.BottomBorder = totalStyleBottomBorder;
            totalStyle.IncludeFont = true;
            totalStyle.IncludeBorder = true;
            this.AddInternal(totalStyle);
        }

        private void AddThemedStyles()
        {
            double tintAndShade = 0.8;
            for (int i = 0; i < 4; i++)
            {
                ThemableColor foreColor = new ThemableColor((i < 2) ? ThemeColorType.Text1 : ThemeColorType.Background1);
                for (int j = (int)ThemeColorType.Accent1; j <= (int)ThemeColorType.Accent6; j++)
                {
                    ThemeColorType themeColorType = (ThemeColorType)j;
                    string styleName = themeColorType.ToString();
                    if (tintAndShade != 0)
                    {
                        styleName = string.Format("{0}% - ", ((1 - tintAndShade) * 100)) + styleName;
                    }

                    CellStyle style = this.CreateCellStyle(this.Workbook, styleName, CellStyleCategory.ThemedCellStyles);
                    style.ForeColor = foreColor;
                    style.Fill = PatternFill.CreateSolidFill(new ThemableColor(themeColorType, tintAndShade));
                    style.IncludeFont = true;
                    style.IncludeFill = true;
                    this.AddInternal(style);
                }

                if (i < 2)
                {
                    tintAndShade -= 0.2;
                }
                else
                {
                    tintAndShade = 0;
                }
            }
        }

        private void AddNumberFormatStyles()
        {
            CellStyle commaStyle = this.CreateCellStyle(this.Workbook, "Comma", CellStyleCategory.NumberFormat);
            commaStyle.Format = new CellValueFormat("_(* #,##0.00_);_(* (#,##0.00);_(* \"-\"??_);_(@_)");
            commaStyle.IncludeNumber = true;
            this.AddInternal(commaStyle);

            CellStyle commaZeroStyle = this.CreateCellStyle(this.Workbook, "Comma [0]", CellStyleCategory.NumberFormat);
            commaZeroStyle.Format = new CellValueFormat("_(* #,##0_);_(* (#,##0);_(* \"-\"_);_(@_)");
            commaZeroStyle.IncludeNumber = true;
            this.AddInternal(commaZeroStyle);

            CellStyle currencyStyle = this.CreateCellStyle(this.Workbook, "Currency", CellStyleCategory.NumberFormat);
            currencyStyle.Format = new CellValueFormat("_($* #,##0.00_);_($* (#,##0.00);_($* \"-\"??_);_(@_)");
            currencyStyle.IncludeNumber = true;
            this.AddInternal(currencyStyle);

            CellStyle currencyZeroStyle = this.CreateCellStyle(this.Workbook, "Currency [0]", CellStyleCategory.NumberFormat);
            currencyZeroStyle.Format = new CellValueFormat("_($* #,##0_);_($* (#,##0);_($* \"-\"_);_(@_)");
            currencyZeroStyle.IncludeNumber = true;
            this.AddInternal(currencyZeroStyle);

            CellStyle percentStyle = this.CreateCellStyle(this.Workbook, "Percent", CellStyleCategory.NumberFormat);
            percentStyle.Format = new CellValueFormat("0%");
            percentStyle.IncludeNumber = true;
            this.AddInternal(percentStyle);
        }

        /// <summary>
        /// Determines whether the style with the specified name is contained.
        /// </summary>
        /// <param name="styleName">Name of the style.</param>
        /// <returns><c>True</c> if is contained.</returns>
        public bool Contains(string styleName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(styleName, "styleName");
            return this.nameToStyles.ContainsKey(styleName);
        }

        /// <summary>
        /// Determines whether the specified style is contained.
        /// </summary>
        /// <param name="style">The style.</param>
        /// <returns><c>True</c> if is contained.</returns>
        public bool Contains(CellStyle style)
        {
            Guard.ThrowExceptionIfNull(style, "style");
            return style.Workbook == this.Workbook && this.Contains(style.Name);
        }

        /// <summary>
        /// Gets the style by name.
        /// </summary>
        /// <param name="styleName">Name of the style.</param>
        /// <returns>The style.</returns>
        public CellStyle GetByName(string styleName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(styleName, "styleName");

            CellStyle result = null;
            this.nameToStyles.TryGetValue(styleName, out result);

            return result;
        }

        /// <summary>
        /// Adds the specified style.
        /// </summary>
        /// <param name="styleName">Name of the style.</param>
        /// <param name="category">The category.</param>
        /// <param name="isRemovable">A value indicating if the style can be removed later.</param>
        /// <returns>The style.</returns>
        public CellStyle Add(string styleName, CellStyleCategory category = CellStyleCategory.Custom, bool isRemovable = true)
        {
            Guard.ThrowExceptionIfNullOrEmpty(styleName, "styleName");

            if (this.Contains(styleName))
            {
                throw new InvalidOperationException(
                    string.Format("Style with name '{0}' already exists",
                    styleName));
            }

            CellStyle style = this.CreateCellStyle(this.workbook, styleName, category, isRemovable);
            AddRemoveStyleCommandContext context = new AddRemoveStyleCommandContext(this.Workbook, style);
            this.Workbook.ExecuteCommand(WorkbookCommands.AddStyle, context);

            return style;
        }

        /// <summary>
        /// Removes the specified style by name.
        /// </summary>
        /// <param name="styleName">Name of the style.</param>
        /// <returns><c>True</c> if succeed.</returns>
        public bool Remove(string styleName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(styleName, "styleName");

            CellStyle style = this.GetByName(styleName);
            if (style == null)
            {
                return false;
            }

            return this.Remove(style);
        }

        /// <summary>
        /// Removes the specified style.
        /// </summary>
        /// <param name="style">The style.</param>
        /// <returns><c>True</c> if succeed.</returns>
        public bool Remove(CellStyle style)
        {
            Guard.ThrowExceptionIfNull(style, "style");

            AddRemoveStyleCommandContext context = new AddRemoveStyleCommandContext(this.Workbook, style);
            return this.Workbook.ExecuteCommand(WorkbookCommands.RemoveStyle, context);
        }

        internal void AddInternal(CellStyle style)
        {
            Guard.ThrowExceptionIfNull(style, "style");
            Guard.ThrowExceptionIfNotEqual(this.Workbook, style.Workbook, "style.Workbook");

            this.nameToStyles.Add(style.Name, style);
            this.OnStyleAdded(style);
        }

        internal void RemoveInternal(CellStyle style)
        {
            Guard.ThrowExceptionIfNull(style, "style");
            Guard.ThrowExceptionIfNotEqual(this.Workbook, style.Workbook, "style.Workbook");

            this.nameToStyles.Remove(style.Name);
            this.OnStyleRemoved(style);
        }

        private void OnStyleAdded(CellStyle style)
        {
            this.OnChanged();
            this.OnStyleChanged(style, CellStyle.GetAllProperties());
            style.Changed += this.Style_Changed;
        }

        private void OnStyleRemoved(CellStyle style)
        {
            style.Changed -= this.Style_Changed;
            this.OnChanged();
            this.OnStyleChanged(style, CellStyle.GetAllProperties());
        }

        private void Style_Changed(object sender, StyleChangedEventArgs e)
        {
            this.OnStyleChanged((CellStyle)sender, e.ChangedProperties);
        }

        private void OnStyleChanged(CellStyle style, IEnumerable<IPropertyDefinition> changedProperties)
        {
            using (new UpdateScope(this.workbook.SuspendLayoutUpdate, this.workbook.ResumeLayoutUpdate))
            {
                foreach (Worksheet worksheet in this.workbook.Worksheets)
                {
                    worksheet.Cells.PropertyBag.OnStyleChanged(style, changedProperties);
                }
            }
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<CellStyle> GetEnumerator()
        {
            return this.nameToStyles.Values.GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return ((IEnumerable<CellStyle>)this).GetEnumerator();
        }

        #endregion


        #region Events

        /// <summary>
        /// Occurs when the collection is changed.
        /// </summary>
        public event EventHandler Changed;

        /// <summary>
        /// Called when the collection is changed.
        /// </summary>
        protected virtual void OnChanged()
        {
            if (this.Changed != null)
            {
                this.Changed(this, EventArgs.Empty);
            }
        }

        #endregion
    }
}