﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.PropertySystem
{
    internal class CellsPropertyBag : WorksheetPropertyBagBase
    {
        #region Fields

        private readonly BeginEndCounter suspendPropertyChangedCounter;
        private readonly HashSet<IPropertyDefinition> changedProperties;
        private readonly BeginEndCounter sortCounter;
        private readonly HashSet<CellRange> sortedRanges;

        #endregion


        #region Properties

        internal bool IsSortInProgress
        {
            get
            {
                return this.sortCounter.IsUpdateInProgress;
            }
        }

        protected override long ToItemIndex
        {
            get
            {
                return SpreadsheetDefaultValues.CellCount - 1;
            }
        }

        #endregion


        #region Constructors

        public CellsPropertyBag()
        {
            this.suspendPropertyChangedCounter = new BeginEndCounter();
            this.changedProperties = new HashSet<IPropertyDefinition>();
            this.sortCounter = new BeginEndCounter(this.OnSortCompleted);
            this.sortedRanges = new HashSet<CellRange>();

            this.RegisterProperty(CellPropertyDefinitions.ValueProperty);
            this.RegisterProperty(CellPropertyDefinitions.DataValidationRuleProperty);
        }

        #endregion


        #region Translation

        internal bool CanInsertRow(/*int index,*/ int itemCount)
        {
            return this.EndLostRowRangeIsNonEmpty(itemCount);
        }

        private bool EndLostRowRangeIsNonEmpty(int itemCount)
        {
            CellRange endRange = new CellRange(SpreadsheetDefaultValues.RowCount - itemCount, 0, SpreadsheetDefaultValues.RowCount - 1, SpreadsheetDefaultValues.ColumnCount - 1);
            return this.IsCellRangeEmpty(endRange);
        }

        private void InsertRowInternal(int index, int itemCount)
        {
            this.TranslateRow(index, itemCount, true);
            this.InvalidateAllPropertiesUsedCellRange();
        }

        internal void InsertEmptyRow(int index, int itemCount)
        {
            CellRange usedCellRange = this.GetUsedCellRange();
            CellRange affectedRange = new CellRange(index,
                  usedCellRange.FromIndex.ColumnIndex,
                  Math.Min(usedCellRange.ToIndex.RowIndex + itemCount, SpreadsheetDefaultValues.RowCount - 1),
                  usedCellRange.ToIndex.ColumnIndex);

            this.InsertRowInternal(index, itemCount);

            if (index <= usedCellRange.ToIndex.RowIndex)
            {
                this.RaisePropertyChangedForRegisteredProperties(affectedRange);
            }
        }

        internal void InsertRowAndSetValues(int index, int itemCount, CellsPropertyBag oldValues)
        {
            this.InsertRowInternal(index, itemCount);
            CellRange cellRange = CellRange.FromRowRange(index, index + itemCount - 1);
            this.CopyFromWithoutPropertyChanged(cellRange, oldValues);

            CellRange usedCellRange = this.GetUsedCellRange();
            CellRange affectedRange = new CellRange(index,
                 usedCellRange.FromIndex.ColumnIndex,
                 Math.Min(usedCellRange.ToIndex.RowIndex + itemCount, SpreadsheetDefaultValues.RowCount - 1),
                 usedCellRange.ToIndex.ColumnIndex);

            if (index <= usedCellRange.ToIndex.RowIndex)
            {
                this.RaisePropertyChangedForRegisteredProperties(affectedRange);
            }
        }

        internal void RemoveRow(int index, int itemCount)
        {
            CellRange usedCellRange = this.GetUsedCellRange();
            Dictionary<IPropertyDefinition, CellRange> changedProperties = null;

            if (index <= usedCellRange.ToIndex.RowIndex)
            {
                CellRange cellRange = CellRange.FromRowRange(index, index + itemCount - 1);

                changedProperties = this.GetChangedProperties((propertyUsedCellRange) =>
                {
                    return new CellRange(cellRange.FromIndex.RowIndex,
                        cellRange.FromIndex.ColumnIndex,
                        propertyUsedCellRange.ToIndex.RowIndex,
                        Math.Min(cellRange.ToIndex.ColumnIndex, propertyUsedCellRange.ToIndex.ColumnIndex));
                });
            }

            this.TranslateRow(index, -itemCount, false);
            this.Clear(new CellRange(SpreadsheetDefaultValues.RowCount - itemCount, 0, SpreadsheetDefaultValues.RowCount - 1, SpreadsheetDefaultValues.ColumnCount - 1));

            this.RaisePropertyChangedForRegisteredProperties(changedProperties);
            this.InvalidateAllPropertiesUsedCellRange();
        }

        internal bool CanInsertColumn(/*int index, */int itemCount)
        {
            return this.EndLostColumnRangeIsNonEmpty(itemCount);
        }

        private bool EndLostColumnRangeIsNonEmpty(int itemCount)
        {
            CellRange endRange = new CellRange(0, SpreadsheetDefaultValues.ColumnCount - itemCount, SpreadsheetDefaultValues.RowCount - 1, SpreadsheetDefaultValues.ColumnCount - 1);
            return this.IsCellRangeEmpty(endRange);
        }

        private void InsertColumnInternal(int index, int itemCount)
        {
            this.TranslateColumn(index, itemCount);

            if (index > 0)
            {
                CellRange rangeToCopyFrom = new CellRange(0, index - 1, SpreadsheetDefaultValues.RowCount - 1, index - 1);
                long fromIndex = ConvertCellIndexToLong(rangeToCopyFrom.FromIndex);
                long toIndex;
                for (int i = index; i < index + itemCount; i++)
                {
                    toIndex = ConvertCellIndexToLong(0, i);
                    base.Copy(fromIndex, SpreadsheetDefaultValues.RowCount, toIndex, true);
                }
            }

            this.InvalidateAllPropertiesUsedCellRange();
        }

        internal void InsertEmptyColumn(int index, int itemCount)
        {
            CellRange usedCellRange = this.GetUsedCellRange();
            CellRange affectedRange = new CellRange(usedCellRange.FromIndex.RowIndex,
                  index,
                  usedCellRange.ToIndex.RowIndex,
                  Math.Min(usedCellRange.ToIndex.ColumnIndex + itemCount, SpreadsheetDefaultValues.ColumnCount - 1));

            this.InsertColumnInternal(index, itemCount);

            if (index <= usedCellRange.ToIndex.ColumnIndex)
            {
                this.RaisePropertyChangedForRegisteredProperties(affectedRange);
            }
        }

        internal void InsertColumnAndSetValues(int index, int itemCount, CellsPropertyBag oldValues)
        {
            this.InsertColumnInternal(index, itemCount);
            CellRange cellRange = CellRange.FromColumnRange(index, index + itemCount - 1);
            this.CopyFromWithoutPropertyChanged(cellRange, oldValues);

            CellRange usedCellRange = this.GetUsedCellRange();
            CellRange affectedRange = new CellRange(usedCellRange.FromIndex.RowIndex,
                  index,
                  usedCellRange.ToIndex.RowIndex,
                  Math.Min(usedCellRange.ToIndex.ColumnIndex + itemCount, SpreadsheetDefaultValues.ColumnCount - 1));

            if (index <= usedCellRange.ToIndex.ColumnIndex)
            {
                this.RaisePropertyChangedForRegisteredProperties(affectedRange);
            }
        }

        internal void RemoveColumn(int index, int itemCount)
        {
            CellRange usedCellRange = this.GetUsedCellRange();
            Dictionary<IPropertyDefinition, CellRange> changedProperties = null;

            if (index <= usedCellRange.ToIndex.ColumnIndex)
            {
                CellRange cellRange = CellRange.FromColumnRange(index, index + itemCount - 1);
                changedProperties = this.GetChangedProperties((propertyUsedCellRange) =>
                    {
                        return new CellRange(cellRange.FromIndex.RowIndex,
                            cellRange.FromIndex.ColumnIndex,
                            Math.Min(cellRange.ToIndex.RowIndex, propertyUsedCellRange.ToIndex.RowIndex),
                            propertyUsedCellRange.ToIndex.ColumnIndex);
                    });
            }

            this.TranslateColumn(index, -itemCount);

            this.RaisePropertyChangedForRegisteredProperties(changedProperties);

            this.InvalidateAllPropertiesUsedCellRange();
        }

        internal bool CanInsertCells(/*int rowIndex, int columnIndex, int itemCount,*/ CellRange range, InsertShiftType shiftType)
        {
            return this.EndLostRangeIsNonEmpty(range, shiftType);
        }

        private bool EndLostRangeIsNonEmpty(CellRange range, InsertShiftType shiftType)
        {
            CellRange lostRange;

            if (shiftType == InsertShiftType.Down)
            {
                lostRange = new CellRange(SpreadsheetDefaultValues.RowCount - range.RowCount, range.FromIndex.ColumnIndex, SpreadsheetDefaultValues.RowCount - 1, range.ToIndex.ColumnIndex);
            }
            else if (shiftType == InsertShiftType.Right)
            {
                lostRange = new CellRange(range.FromIndex.RowIndex, SpreadsheetDefaultValues.ColumnCount - range.ColumnCount, range.ToIndex.RowIndex, SpreadsheetDefaultValues.ColumnCount - 1);
            }
            else
            {
                throw new NotSupportedException("Shift type not supported.");
            }

            return this.IsCellRangeEmpty(lostRange);
        }

        private void InsertCellsInternal(CellRange cellRange, InsertShiftType insertShiftType)
        {
            CellRange usedCellRange = this.GetUsedCellRange();

            if (insertShiftType == InsertShiftType.Right)
            {
                CellRange rangeToMove = new CellRange(cellRange.FromIndex.RowIndex,
                    cellRange.FromIndex.ColumnIndex,
                    cellRange.ToIndex.RowIndex,
                    Math.Min(usedCellRange.ToIndex.ColumnIndex + cellRange.ColumnCount, SpreadsheetDefaultValues.ColumnCount - 1));

                int offset = SpreadsheetDefaultValues.RowCount * cellRange.ColumnCount;
                foreach (var range in ConvertCellRangeToLongRanges(rangeToMove).Reverse())
                {
                    if (range.End + offset <= SpreadsheetDefaultValues.CellCount)
                    {
                        this.TranslateWithoutPropertyChanged(offset, range);
                    }
                }

                this.Clear(cellRange);

                if (rangeToMove.FromIndex.ColumnIndex > 0)
                {
                    LongRange rangeToCopyPropertiesFrom = new LongRange(
                        CellsPropertyBag.ConvertCellIndexToLong(rangeToMove.FromIndex.RowIndex, rangeToMove.FromIndex.ColumnIndex - 1),
                        CellsPropertyBag.ConvertCellIndexToLong(rangeToMove.ToIndex.RowIndex, rangeToMove.FromIndex.ColumnIndex - 1));

                    foreach (var range in ConvertCellRangeToLongRanges(cellRange).Reverse())
                    {
                        if (range.End + offset <= SpreadsheetDefaultValues.CellCount)
                        {
                            this.Copy(rangeToCopyPropertiesFrom.Start, rangeToCopyPropertiesFrom.Length, range.Start, true);
                        }
                    }
                }
            }
            else
            {
                CellRange rangeToMove = new CellRange(cellRange.FromIndex.RowIndex,
                    cellRange.FromIndex.ColumnIndex,
                    Math.Min(usedCellRange.ToIndex.RowIndex + cellRange.RowCount, SpreadsheetDefaultValues.RowCount - 1),
                    cellRange.ToIndex.ColumnIndex);

                this.Translate(rangeToMove, cellRange.RowCount, true);
            }

            this.InvalidateAllPropertiesUsedCellRange();
        }

        internal void InsertEmptyCells(CellRange cellRange, InsertShiftType insertShiftType)
        {
            CellRange usedCellRange = this.GetUsedCellRange();

            CellRange extendedUsedCellRange = new CellRange(
                usedCellRange.FromIndex.RowIndex,
                usedCellRange.FromIndex.ColumnIndex,
                usedCellRange.ToIndex.RowIndex + 1,
                usedCellRange.ToIndex.ColumnIndex + 1);

            if (!extendedUsedCellRange.IntersectsWith(cellRange))
            {
                return;
            }

            int toRowIndex;
            int toColumnIndex;
            if (insertShiftType == InsertShiftType.Right)
            {
                toRowIndex = Math.Min(cellRange.ToIndex.RowIndex, usedCellRange.ToIndex.RowIndex);
                toColumnIndex = Math.Min(usedCellRange.ToIndex.ColumnIndex + cellRange.ColumnCount, SpreadsheetDefaultValues.ColumnCount - 1);
            }
            else
            {
                toRowIndex = Math.Min(usedCellRange.ToIndex.RowIndex + cellRange.RowCount, SpreadsheetDefaultValues.RowCount - 1);
                toColumnIndex = Math.Min(cellRange.ToIndex.ColumnIndex, usedCellRange.ToIndex.ColumnIndex);
            }

            CellRange affectedRange = new CellRange(cellRange.FromIndex.RowIndex,
                cellRange.FromIndex.ColumnIndex,
                toRowIndex,
                toColumnIndex);

            this.InsertCellsInternal(cellRange, insertShiftType);

            this.RaisePropertyChangedForRegisteredProperties(affectedRange);
        }

        internal void InsertCellsAndSetValues(CellRange cellRange, InsertShiftType insertShiftType, CellsPropertyBag oldValues)
        {
            this.InsertCellsInternal(cellRange, insertShiftType);
            this.CopyFromWithoutPropertyChanged(cellRange, oldValues);

            int toRowIndex;
            int toColumnIndex;
            CellRange usedCellRange = this.GetUsedCellRange();
            if (insertShiftType == InsertShiftType.Right)
            {
                toRowIndex = Math.Min(cellRange.ToIndex.RowIndex, usedCellRange.ToIndex.RowIndex);
                toColumnIndex = Math.Min(usedCellRange.ToIndex.ColumnIndex + cellRange.ColumnCount, SpreadsheetDefaultValues.ColumnCount - 1);
            }
            else
            {
                toRowIndex = Math.Min(usedCellRange.ToIndex.RowIndex + cellRange.RowCount, SpreadsheetDefaultValues.RowCount - 1);
                toColumnIndex = Math.Min(cellRange.ToIndex.ColumnIndex, usedCellRange.ToIndex.ColumnIndex);
            }

            CellRange affectedRange = new CellRange(cellRange.FromIndex.RowIndex,
                cellRange.FromIndex.ColumnIndex,
                toRowIndex,
                toColumnIndex);

            this.RaisePropertyChangedForRegisteredProperties(affectedRange);
        }

        internal void RemoveCells(CellRange cellRange, RemoveShiftType shiftType)
        {
            CellRange usedCellRange = this.GetUsedCellRange();

            if (!usedCellRange.IntersectsWith(cellRange))
            {
                return;
            }

            Dictionary<IPropertyDefinition, CellRange> changedProperties;

            if (shiftType == RemoveShiftType.Left)
            {
                changedProperties = this.GetChangedProperties((propertyUsedCellRange) =>
                    {
                        return new CellRange(cellRange.FromIndex.RowIndex,
                            cellRange.FromIndex.ColumnIndex,
                            Math.Min(cellRange.ToIndex.RowIndex, propertyUsedCellRange.ToIndex.RowIndex),
                            propertyUsedCellRange.ToIndex.ColumnIndex);
                    });

                this.RemoveCellsLeft(cellRange, usedCellRange);
            }
            else
            {
                changedProperties = this.GetChangedProperties((propertyUsedCellRange) =>
                {
                    return new CellRange(cellRange.FromIndex.RowIndex,
                        cellRange.FromIndex.ColumnIndex,
                        propertyUsedCellRange.ToIndex.RowIndex,
                        Math.Min(cellRange.ToIndex.ColumnIndex, propertyUsedCellRange.ToIndex.ColumnIndex));
                });

                this.RemoveCellsUp(cellRange, usedCellRange);
            }

            this.RaisePropertyChangedForRegisteredProperties(changedProperties);
            this.InvalidateAllPropertiesUsedCellRange();
        }

        private void RemoveCellsLeft(CellRange cellRange, CellRange usedCellRange)
        {
            int fromColumnIndex = Math.Min(cellRange.ToIndex.ColumnIndex + 1, SpreadsheetDefaultValues.ColumnCount - 1);
            int toColumnIndex = Math.Max(usedCellRange.ToIndex.ColumnIndex, cellRange.ToIndex.ColumnIndex + 1);

            CellRange rangeToMove = new CellRange(
                cellRange.FromIndex.RowIndex,
                fromColumnIndex,
                cellRange.ToIndex.RowIndex,
                toColumnIndex);

            if (cellRange.ToIndex.ColumnIndex != SpreadsheetDefaultValues.ColumnCount - 1)
            {
                int offset = -SpreadsheetDefaultValues.RowCount * cellRange.ColumnCount;
                foreach (var range in ConvertCellRangeToLongRanges(rangeToMove))
                {
                    if (range.Start + offset >= 0)
                    {
                        this.TranslateWithoutPropertyChanged(offset, range);
                    }
                }
            }

            var rangeToClear = new CellRange(
                rangeToMove.FromIndex.RowIndex,
                rangeToMove.ToIndex.ColumnIndex - cellRange.ColumnCount + 1,
                rangeToMove.ToIndex.RowIndex,
                Math.Min(usedCellRange.ToIndex.ColumnIndex, rangeToMove.ToIndex.ColumnIndex + 1));

            this.Clear(rangeToClear);
        }

        private void RemoveCellsUp(CellRange cellRange, CellRange usedCellRange)
        {
            CellRange rangeToMove = new CellRange(
                cellRange.FromIndex.RowIndex,
                cellRange.FromIndex.ColumnIndex,
                Math.Max(cellRange.ToIndex.RowIndex, usedCellRange.ToIndex.RowIndex),
                cellRange.ToIndex.ColumnIndex);

            this.Translate(rangeToMove, -cellRange.RowCount, false);

            CellRange rangeToClear = new CellRange(
                Math.Max(usedCellRange.ToIndex.RowIndex - cellRange.RowCount + 1, cellRange.FromIndex.RowIndex),
                cellRange.FromIndex.ColumnIndex,
                Math.Max(usedCellRange.ToIndex.RowIndex, cellRange.ToIndex.RowIndex),
                cellRange.ToIndex.ColumnIndex);

            this.Clear(rangeToClear);
        }

        internal void RemoveCell(int rowIndex, int columnIndex, int itemCount, RemoveShiftType shiftType)
        {
            if (shiftType == RemoveShiftType.Left)
            {
                for (int i = columnIndex; i < SpreadsheetDefaultValues.ColumnCount - itemCount; i++)
                {
                    base.Copy(
                        ConvertCellIndexToLong(rowIndex, i + itemCount),
                        ConvertCellIndexToLong(rowIndex, i), false);
                }

                for (int i = SpreadsheetDefaultValues.ColumnCount - itemCount; i < SpreadsheetDefaultValues.ColumnCount; i++)
                {
                    base.Clear(ConvertCellIndexToLong(rowIndex, i));
                }
            }
            else
            {
                CellRange range = new CellRange(rowIndex, columnIndex, SpreadsheetDefaultValues.RowCount - 1, columnIndex);
                this.Translate(range, -itemCount, false);
                this.Clear(new CellRange(SpreadsheetDefaultValues.RowCount - itemCount, columnIndex, SpreadsheetDefaultValues.RowCount - 1, columnIndex));
            }

            //this.RaisePropertyChangedForRegisteredProperties(rangeToMove);            
            this.InvalidateAllPropertiesUsedCellRange();
        }

        private void TranslateWithoutPropertyChanged(int offset, LongRange range)
        {
            foreach (IPropertyDefinition propertyDefinition in this.GetRegisteredProperties())
            {
                ICompressedList compressedList = this.GetPropertyValueCollection(propertyDefinition);
                ICompressedList valuesToMove = compressedList.GetValue(range.Start, range.End);
                ICompressedList movedValues = valuesToMove.Offset(offset);
                compressedList.SetValue(movedValues);
            }
        }

        private void CopyFromWithoutPropertyChanged(CellRange cellRange, CellsPropertyBag oldValues)
        {
            foreach (LongRange range in ConvertCellRangeToLongRanges(cellRange))
            {
                foreach (IPropertyDefinition property in this.GetRegisteredProperties())
                {
                    ICompressedList compressedList = oldValues.GetPropertyValueCollection(property);
                    ICompressedList propertyValueCollection = this.GetPropertyValueCollection(property);
                    propertyValueCollection.SetValue(compressedList.GetValue(range.Start, range.End));
                }
            }
        }

        private Dictionary<IPropertyDefinition, CellRange> GetChangedProperties(Func<CellRange, CellRange> getCellRange)
        {
            Dictionary<IPropertyDefinition, CellRange> changedProperties = new Dictionary<IPropertyDefinition, CellRange>();

            foreach (IPropertyDefinition propertyDefinition in this.GetRegisteredProperties())
            {
                CellRange propertyUsedCellRange = this.GetUsedCellRange(propertyDefinition);
                if (!propertyUsedCellRange.Equals(CellRange.Empty) || !this.IsPropertyDefault(propertyDefinition, 0))
                {
                    CellRange cellRange = getCellRange(propertyUsedCellRange);
                    changedProperties.Add(propertyDefinition, cellRange);
                }
            }

            return changedProperties;
        }

        private void Clear(CellRange cellRange)
        {
            foreach (LongRange range in ConvertCellRangeToLongRanges(cellRange))
            {
                this.Clear(range.Start, range.End);
            }
        }

        private void TranslateRow(int index, int offset, bool useSameValueAsPrevious)
        {
            CellRange range = new CellRange(index, 0, SpreadsheetDefaultValues.RowCount - 1, SpreadsheetDefaultValues.ColumnCount - 1);
            this.Translate(range, offset, useSameValueAsPrevious);
        }

        private void TranslateColumn(int index, int offset)
        {
            CellRange range = new CellRange(0, index, SpreadsheetDefaultValues.RowCount - 1, SpreadsheetDefaultValues.ColumnCount - 1);
            //this.Translate(range, (offset * SpreadsheetDefaultValues.RowCount), false);
            long fromIndex = ConvertCellIndexToLong(range.FromIndex);
            long toIndex = ConvertCellIndexToLong(range.ToIndex);
            this.Translate(fromIndex, toIndex, (offset * SpreadsheetDefaultValues.RowCount), false);
        }

        private void Translate(CellRange cellRange, long offset, bool useSameValueAsPrevious)
        {
            foreach (LongRange range in ConvertCellRangeToLongRanges(cellRange))
            {
                this.Translate(range.Start, range.End, offset, useSameValueAsPrevious);
            }
        }

        private void RaisePropertyChangedForRegisteredProperties(CellRange cellRange)
        {
            foreach (IPropertyDefinition propertyDefinition in this.GetRegisteredProperties())
            {
                CellRange propertyUsedCellRange = this.GetUsedCellRange(propertyDefinition);
                if ((!propertyUsedCellRange.Equals(CellRange.Empty) || !this.IsPropertyDefault(propertyDefinition, 0))
                    && cellRange.IntersectsWith(propertyUsedCellRange))
                {
                    CellRange changedRange = new CellRange(cellRange.FromIndex, cellRange.ToIndex);
                    this.OnPropertyChanged(new CellPropertyChangedEventArgs(propertyDefinition, changedRange));
                }
            }
        }

        private void RaisePropertyChangedForRegisteredProperties(Dictionary<IPropertyDefinition, CellRange> changedProperties)
        {
            if (changedProperties != null)
            {
                foreach (KeyValuePair<IPropertyDefinition, CellRange> pair in changedProperties)
                {
                    this.OnPropertyChanged(new CellPropertyChangedEventArgs(pair.Key, pair.Value));
                }
            }
        }

        #endregion


        #region Methods

        private bool IsCellRangeEmpty(CellRange cellRange)
        {
            IEnumerable<LongRange> longRanges = ConvertCellRangeToLongRanges(cellRange);

            foreach (var range in longRanges)
            {
                long fromIndex = range.Start;
                long toIndex = range.End;

                foreach (IPropertyDefinition property in this.GetRegisteredProperties())
                {
                    ICompressedList collection = this.GetPropertyValueCollection(property);

                    if (collection.GetLastNonEmptyRangeEnd(fromIndex, toIndex) != null)
                    {
                        return false;
                    }
                }
            }

            return true;
        }

        internal bool IsDefault(long index)
        {
            foreach (var property in CellPropertyDefinitions.AllPropertyDefinitions)
            {
                long? range = this.GetPropertyValueCollection(property).GetLastNonEmptyRangeEnd(index, index);
                if (range != null)
                {
                    return false;
                }
            }

            return true;
        }

        private bool IsPropertyDefault(IPropertyDefinition property, long index)
        {
            long? range = this.GetPropertyValueCollection(property).GetLastNonEmptyRangeEnd(index, index);
            return range == null;
        }

        internal void SuspendPropertyChanged()
        {
            this.suspendPropertyChangedCounter.BeginUpdate();
        }

        /// <summary>
        /// This method raises PropertyChanged for the entire used cell range for each property that 
        /// has been changed. It is used upon Import and cell references update. Use with caution!
        /// </summary>
        internal void ResumePropertyChanged()
        {
            this.ResumePropertyChanged(null);
        }

        /// <summary>
        /// This method raises PropertyChanged for the entire used cell range for each property that 
        /// has been changed. It is used upon Import and cell references update. Use with caution!
        /// </summary>
        internal void ResumePropertyChanged(CellRange updatedRange)
        {
            this.suspendPropertyChangedCounter.EndUpdate();

            if (!this.suspendPropertyChangedCounter.IsUpdateInProgress &&
                this.changedProperties.Count > 0)
            {
                foreach (var property in this.changedProperties)
                {
                    if (updatedRange == null)
                    {
                        updatedRange = this.GetUsedCellRange(property);
                    }

                    this.OnPropertyChanged(new CellPropertyChangedEventArgs(property, updatedRange));
                }

                this.changedProperties.Clear();
            }
        }

        public ICompressedList<T> GetPropertyValue<T>(IPropertyDefinition<T> property, CellRange cellRange)
        {
            Guard.ThrowExceptionIfNull(cellRange, "cellRange");

            long fromIndex = ConvertCellIndexToLong(cellRange.FromIndex);
            long toIndex = ConvertCellIndexToLong(cellRange.ToIndex);

            ICompressedList<T> result = new CompressedList<T>(fromIndex, toIndex, property.DefaultValue);
            ICompressedList<T> propertyValueCollection = base.GetPropertyValueCollection(property);

            foreach (LongRange longRange in ConvertCellRangeToLongRanges(cellRange))
            {
                ICompressedList<T> value = propertyValueCollection.GetValue(longRange.Start, longRange.End);
                result.SetValue(value);
            }

            return result;
        }

        public T GetPropertyValue<T>(IPropertyDefinition<T> property, CellIndex cellIndex)
        {
            Guard.ThrowExceptionIfNull(cellIndex, "cellIndex");

            long index = ConvertCellIndexToLong(cellIndex);
            return base.GetPropertyValue(property, index);
        }

        public void SetPropertyValue<T>(IPropertyDefinition<T> property, CellIndex cellIndex, T value)
        {
            Guard.ThrowExceptionIfNull(cellIndex, "index");

            this.SetPropertyValue(property, cellIndex.ToCellRange(), value);
        }

        public void SetPropertyValue<T>(IPropertyDefinition<T> property, CellRange cellRange, T value)
        {
            Guard.ThrowExceptionIfNull(cellRange, "range");

            foreach (LongRange range in ConvertCellRangeToLongRanges(cellRange))
            {
                base.SetPropertyValue(property, range.Start, range.End, value);
            }
        }

        public void SetPropertyValue<T>(IPropertyDefinition<T> property, CellRange cellRange, ICompressedList values)
        {
            Guard.ThrowExceptionIfNull(property, "property");
            Guard.ThrowExceptionIfNull(values, "values");

            ICompressedList propertyValueCollection = this.GetPropertyValueCollection(property);
            foreach (LongRange longRange in ConvertCellRangeToLongRanges(cellRange))
            {
                ICompressedList value = values.GetValue(longRange.Start, longRange.End);
                propertyValueCollection.SetValue(value);
            }

            this.OnPropertyChanged(new CellPropertyChangedEventArgs(property, cellRange));
        }

        public void UpdatePropertyValue<T>(IPropertyDefinition<T> property, CellRange cellRange, Func<T, T> newValueTransform)
        {
            Guard.ThrowExceptionIfNull(cellRange, "cellRange");
            Guard.ThrowExceptionIfNull(newValueTransform, "newValueTransform");

            foreach (LongRange range in ConvertCellRangeToLongRanges(cellRange))
            {
                base.UpdatePropertyValue(property, range.Start, range.End, newValueTransform);
            }
        }

        public void ClearPropertyValue<T>(IPropertyDefinition<T> property, CellIndex cellIndex)
        {
            long index = ConvertCellIndexToLong(cellIndex);
            base.ClearPropertyValue(property, index);
        }

        public void ClearPropertyValue<T>(IPropertyDefinition<T> property, CellRange cellRange)
        {
            Guard.ThrowExceptionIfNull(cellRange, "range");

            foreach (LongRange range in ConvertCellRangeToLongRanges(cellRange))
            {
                base.ClearPropertyValue(property, range.Start, range.End);
            }
        }

        public void CopyPropertiesFrom(CellsPropertyBag fromProperties, CellRange cellRange)
        {
            foreach (LongRange range in ConvertCellRangeToLongRanges(cellRange))
            {
                base.CopyPropertiesFrom(fromProperties, range.Start, range.End);
            }
        }

        protected override void OnPropertyChanged(IPropertyDefinition property, long fromIndex, long toIndex)
        {
            this.InvalidatePropertyUsedCellRange(property);

            if (this.suspendPropertyChangedCounter.IsUpdateInProgress)
            {
                if (!this.changedProperties.Contains(property))
                {
                    this.changedProperties.Add(property);
                }
            }
            else
            {
                int fromRowIndex;
                int fromColumnIndex;
                int toRowIndex;
                int toColumnIndex;

                ConvertLongToRowAndColumnIndexes(fromIndex, out fromRowIndex, out fromColumnIndex);
                ConvertLongToRowAndColumnIndexes(toIndex, out toRowIndex, out toColumnIndex);

                CellRange cellRange = new CellRange(fromRowIndex, fromColumnIndex, toRowIndex, toColumnIndex);
                this.OnPropertyChanged(new CellPropertyChangedEventArgs(property, cellRange));
            }
        }

        internal IEnumerable<CellRange> GetValueContainingCellRanges<T>(IPropertyDefinition<T> properyDefinition, T value)
        {
            ICompressedList<T> dataValidationValueCollection = this.GetPropertyValueCollection(properyDefinition);
            IEnumerable<Range<long, T>> nonDefaultRanges = dataValidationValueCollection.GetNonDefaultRanges();
            IEnumerable<Range<long, T>> rangesWithTheSameSettings = nonDefaultRanges.Where(
                (p) =>
                {
                    return p.Value.Equals(value);
                });

            IOrderedEnumerable<Range<long, T>> orderedRanges = rangesWithTheSameSettings.OrderBy(p => p.Start);

            List<CellRange> usedRanges = new List<CellRange>();
            foreach (Range<long, T> range in orderedRanges)
            {
                usedRanges.Add(CellsPropertyBag.ConvertLongCellRangeToCellRange(range.Start, range.End));
            }

            bool result;
            do
            {
                result = this.MergeCellRanges(usedRanges);
            }
            while (result);

            return usedRanges;
        }

        private bool MergeCellRanges(List<CellRange> usedRanges)
        {
            for (int i = 0; i < usedRanges.Count; i++)
            {
                for (int j = i + 1; j < usedRanges.Count; j++)
                {
                    CellRange firstRange = usedRanges[i];
                    CellRange secondRange = usedRanges[j];
                    if (this.CanMergeCellRanges(firstRange, secondRange))
                    {
                        usedRanges.Remove(firstRange);
                        usedRanges.Remove(secondRange);
                        usedRanges.Insert(0, this.MergeCellRanges(firstRange, secondRange));
                        return true;
                    }
                }
            }

            return false;
        }

        private bool CanMergeCellRanges(CellRange firstRange, CellRange secondRange)
        {
            if (firstRange.RowCount != secondRange.RowCount)
            {
                return false;
            }

            if (firstRange.FromIndex.RowIndex != secondRange.FromIndex.RowIndex)
            {
                return false;
            }

            if (firstRange.ToIndex.ColumnIndex != secondRange.FromIndex.ColumnIndex - 1 &&
                firstRange.FromIndex.ColumnIndex != secondRange.ToIndex.ColumnIndex + 1)
            {
                return false;
            }

            return true;
        }

        private CellRange MergeCellRanges(CellRange firstRange, CellRange secondRange)
        {
            return new CellRange(
                Math.Min(firstRange.FromIndex.RowIndex, secondRange.FromIndex.RowIndex),
                Math.Min(firstRange.FromIndex.ColumnIndex, secondRange.FromIndex.ColumnIndex),
                Math.Max(firstRange.ToIndex.RowIndex, secondRange.ToIndex.RowIndex),
                Math.Max(firstRange.ToIndex.ColumnIndex, secondRange.ToIndex.ColumnIndex));
        }

        protected override ICompressedList<T> GetLocalPropertyValueRespectingRowsColumns<T>(IPropertyDefinition<T> property, Worksheet worksheet, long fromIndex, long toIndex)
        {
            int fromRowIndex;
            int fromColumnIndex;
            int toRowIndex;
            int toColumnIndex;

            CellsPropertyBag.ConvertLongToRowAndColumnIndexes(fromIndex, out fromRowIndex, out fromColumnIndex);
            CellsPropertyBag.ConvertLongToRowAndColumnIndexes(toIndex, out toRowIndex, out toColumnIndex);

            ICompressedList<T> localPropertyValue = this.GetPropertyValue(property, fromIndex, toIndex);
            ICompressedList<T> resultCellPropertyValues = new CompressedList<T>(localPropertyValue.FromIndex, localPropertyValue.ToIndex, property.DefaultValue);

            if (worksheet.Columns.ContainsPropertyDefinition(property))
            {
                ICompressedList<T> columnPropertyValues = worksheet.Columns.PropertyBag.GetPropertyValueCollection(property).GetValue(fromColumnIndex, toColumnIndex);
                foreach (Range<long, T> columnRange in columnPropertyValues.GetNonDefaultRanges())
                {
                    for (int column = (int)columnRange.Start; column <= columnRange.End; column++)
                    {
                        long startIndex = CellsPropertyBag.ConvertCellIndexToLong(fromRowIndex, column);
                        long endIndex = CellsPropertyBag.ConvertCellIndexToLong(toRowIndex, column);

                        resultCellPropertyValues.SetValue(startIndex, endIndex, columnRange.Value);
                    }
                }
            }

            if (worksheet.Rows.ContainsPropertyDefinition(property))
            {
                ICompressedList<T> rowPropertyValues = worksheet.Rows.PropertyBag.GetPropertyValueCollection(property).GetValue(fromRowIndex, toRowIndex);
                foreach (Range<long, T> rowRange in rowPropertyValues.GetNonDefaultRanges())
                {
                    for (int column = fromColumnIndex; column <= toColumnIndex; column++)
                    {
                        long startIndex = CellsPropertyBag.ConvertCellIndexToLong((int)rowRange.Start, column);
                        long endIndex = CellsPropertyBag.ConvertCellIndexToLong((int)rowRange.End, column);

                        resultCellPropertyValues.SetValue(startIndex, endIndex, rowRange.Value);
                    }
                }
            }

            foreach (Range<long, T> range in localPropertyValue.GetNonDefaultRanges())
            {
                resultCellPropertyValues.SetValue(range.Start, range.End, range.Value);
            }

            return resultCellPropertyValues;
        }

        protected override bool TryGetRowOrColumnPropertyValueRespectingStyle<T>(IPropertyDefinition<T> property, Worksheet worksheet, long index, out T result)
        {
            int rowIndex;
            int columnIndex;
            CellsPropertyBag.ConvertLongToRowAndColumnIndexes(index, out rowIndex, out columnIndex);

            result = worksheet.Rows.PropertyBag.GetPropertyValueRespectingStyle(property, worksheet, rowIndex);

            if (TelerikHelper.EqualsOfT(result, property.DefaultValue))
            {
                result = worksheet.Columns.PropertyBag.GetPropertyValueRespectingStyle(property, worksheet, columnIndex);
                if (TelerikHelper.EqualsOfT(result, property.DefaultValue))
                {
                    return false;
                }
            }

            return true;
        }

        protected override bool TryGetRowOrColumnPropertyValueRespectingStyle<T>(IPropertyDefinition<T> property, Worksheet worksheet, long fromIndex, long toIndex, out ICompressedList<T> rowOrColumnPropertyValuesRespectingStyle)
        {
            rowOrColumnPropertyValuesRespectingStyle = new CompressedList<T>(fromIndex, toIndex, property.DefaultValue);

            int fromRowIndex;
            int fromColumnIndex;
            CellsPropertyBag.ConvertLongToRowAndColumnIndexes(fromIndex, out fromRowIndex, out fromColumnIndex);

            int toRowIndex;
            int toColumnIndex;
            CellsPropertyBag.ConvertLongToRowAndColumnIndexes(toIndex, out toRowIndex, out toColumnIndex);

            for (int column = fromColumnIndex; column <= toColumnIndex; column++)
            {
                T result = worksheet.Columns.PropertyBag.GetPropertyValueRespectingStyle(property, worksheet, column);
                lock (property)
                {
                    if (!TelerikHelper.EqualsOfT(result, property.DefaultValue))
                    {
                        long startIndex = CellsPropertyBag.ConvertCellIndexToLong(fromRowIndex, column);
                        long endIndex = CellsPropertyBag.ConvertCellIndexToLong(toRowIndex, column);
                        rowOrColumnPropertyValuesRespectingStyle.SetValue(startIndex, endIndex, result);
                    }
                }
            }

            IEnumerable<Range<long, T>> propertyNonDefaultRanges = worksheet.Rows.PropertyBag.GetPropertyValueCollection(property).GetValue(fromRowIndex, toRowIndex).GetNonDefaultRanges();
            IEnumerable<Range<long, string>> styleNonDefaultRanges = worksheet.Rows.PropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.StyleNameProperty).GetValue(fromRowIndex, toRowIndex).GetNonDefaultRanges();

            foreach (Range<long, string> styleRange in styleNonDefaultRanges)
            {
                CellStyle cellStyle = worksheet.Workbook.Styles.GetByName(styleRange.Value);
                if (cellStyle.GetIsPropertyIncluded(property))
                {
                    for (int column = fromColumnIndex; column <= toColumnIndex; column++)
                    {
                        long startIndex = CellsPropertyBag.ConvertCellIndexToLong((int)styleRange.Start, column);
                        long endIndex = CellsPropertyBag.ConvertCellIndexToLong((int)styleRange.End, column);
                        rowOrColumnPropertyValuesRespectingStyle.SetValue(startIndex, endIndex, cellStyle.GetPropertyValue(property));
                    }
                }
            }

            foreach (Range<long, T> propertyRange in propertyNonDefaultRanges)
            {
                for (int column = fromColumnIndex; column <= toColumnIndex; column++)
                {
                    long startIndex = CellsPropertyBag.ConvertCellIndexToLong((int)propertyRange.Start, column);
                    long endIndex = CellsPropertyBag.ConvertCellIndexToLong((int)propertyRange.End, column);
                    rowOrColumnPropertyValuesRespectingStyle.SetValue(startIndex, endIndex, propertyRange.Value);
                }
            }

            return true;
        }

        #endregion


        #region Sorting

        internal void BeginSort()
        {
            this.sortCounter.BeginUpdate();
        }

        internal void EndSort()
        {
            this.sortCounter.EndUpdate();
        }

        internal void Sort(Worksheet worksheet, CellRange range, int[] sortedIndexes)
        {
            List<Action> actions = new List<Action>();

            foreach (var propertyDefinition in CellPropertyDefinitions.AllPropertyDefinitions)
            {
                actions.Add(new Action(() =>
                {
                    for (int col = range.FromIndex.ColumnIndex; col <= range.ToIndex.ColumnIndex; col++)
                    {
                        var fromLongIndex = CellsPropertyBag.ConvertCellIndexToLong(range.FromIndex.RowIndex, col);
                        var toLongIndex = CellsPropertyBag.ConvertCellIndexToLong(range.ToIndex.RowIndex, col);

                        if (propertyDefinition == CellPropertyDefinitions.ValueProperty)
                        {
                            this.GetPropertyValueCollection(CellPropertyDefinitions.ValueProperty).Sort(fromLongIndex, toLongIndex, sortedIndexes,
                                (newIndex, value) =>
                                {
                                    if (value == null)
                                    {
                                        return value;
                                    }

                                    FormulaCellValue formulaCellValue = value.Value as FormulaCellValue;
                                    if (formulaCellValue != null)
                                    {
                                        int rowIndex;
                                        int columnIndex;
                                        CellsPropertyBag.ConvertLongToRowAndColumnIndexes(newIndex, out rowIndex, out columnIndex);
                                        return new ValueBox<ICellValue>(formulaCellValue.CloneAndTranslate(worksheet, rowIndex, columnIndex));
                                    }

                                    return value;
                                });
                        }
                        else
                        {
                            this.GetPropertyValueCollection(propertyDefinition).Sort(fromLongIndex, toLongIndex, sortedIndexes);
                        }
                    }
                }));
            }

            Telerik.Windows.Documents.Utilities.TasksHelper.DoAsync(actions);

            this.OnAfterSort(range);

            this.OnSorted(new SortedEventArgs(range, sortedIndexes));
        }

        private void OnAfterSort(CellRange range)
        {
            if (this.IsSortInProgress)
            {
                if (!this.sortedRanges.Contains(range))
                {
                    this.sortedRanges.Add(range);
                }
            }
            else
            {
                this.FireSortPropertyChanged(range);
            }
        }

        private void OnSortCompleted()
        {
            foreach (var range in this.sortedRanges)
            {
                this.FireSortPropertyChanged(range);
            }

            this.sortedRanges.Clear();
        }

        private void FireSortPropertyChanged(CellRange range)
        {
            foreach (var propertyDefinition in CellPropertyDefinitions.AllPropertyDefinitions)
            {
                var eventArgs = new CellPropertyChangedEventArgs(propertyDefinition, range);
                eventArgs.ShouldRespectAffectLayoutProperty = false;
                this.OnPropertyChanged(eventArgs);
            }
        }

        #endregion


        #region Style Related

        internal void OnStyleChanged(CellStyle style, IEnumerable<IPropertyDefinition> changedProperties)
        {
            List<Tuple<long, long>> rangesWithStyle = new List<Tuple<long, long>>();
            foreach (var range in this.GetPropertyValueCollection(CellPropertyDefinitions.StyleNameProperty))
            {
                if (range.Value == style.Name)
                {
                    rangesWithStyle.Add(new Tuple<long, long>(range.Start, range.End));
                }
            }

            foreach (var property in changedProperties)
            {
                foreach (var range in rangesWithStyle)
                {
                    this.OnPropertyChanged(property, range.Item1, range.Item2);
                }
            }
        }

        #endregion


        #region Events

        public event EventHandler<CellPropertyChangedEventArgs> PropertyChanged;
        protected virtual void OnPropertyChanged(CellPropertyChangedEventArgs args)
        {
            Guard.ThrowExceptionIfNull(args, "args");

            if (this.suspendPropertyChangedCounter.IsUpdateInProgress)
            {
                if (!this.changedProperties.Contains(args.Property))
                {
                    this.changedProperties.Add(args.Property);
                }
            }
            else if (this.PropertyChanged != null)
            {
                this.PropertyChanged(this, args);
            }
        }

        internal event EventHandler<SortedEventArgs> Sorted;
        internal void OnSorted(SortedEventArgs args)
        {
            Guard.ThrowExceptionIfNull(args, "args");

            if (this.Sorted != null)
            {
                this.Sorted(this, args);
            }
        }

        #endregion
    }
}
