﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.PropertySystem
{
    internal abstract class PropertyBagBase
    {
        #region Fields

        private readonly Dictionary<IPropertyDefinition, ICompressedList> propertyToValueCollection;

        #endregion


        #region Properties

        protected virtual long FromItemIndex
        {
            get
            {
                return 0;
            }
        }

        protected abstract long ToItemIndex { get; }

        #endregion


        #region Constructors

        public PropertyBagBase()
        {
            this.propertyToValueCollection = new Dictionary<IPropertyDefinition, ICompressedList>();
        }

        #endregion


        #region Methods

        protected IEnumerable<IPropertyDefinition> GetRegisteredProperties()
        {
            return this.propertyToValueCollection.Keys;
        }

        protected void Copy(long fromIndex, long toIndex, bool respectUseSameValueAsPrevious)
        {
            foreach (var keyValuePair in this.propertyToValueCollection)
            {
                bool useSameValueAsPreviousOnInsert = keyValuePair.Key.UseSameValueAsPreviousOnInsert;
                ITranslateableCollection valueCollection = keyValuePair.Value;
                if (!respectUseSameValueAsPrevious || useSameValueAsPreviousOnInsert)
                {
                    valueCollection.Copy(fromIndex, toIndex);
                }
            }
        }

        protected void Copy(long fromIndex, long itemsCount, long toIndex, bool respectUseSameValueAsPrevious)
        {
            foreach (var keyValuePair in this.propertyToValueCollection)
            {
                bool useSameValueAsPreviousOnInsert = keyValuePair.Key.UseSameValueAsPreviousOnInsert;
                ITranslateableCollection valueCollection = keyValuePair.Value;
                if (!respectUseSameValueAsPrevious || useSameValueAsPreviousOnInsert)
                {
                    valueCollection.Copy(fromIndex, itemsCount, toIndex);
                }
            }
        }

        protected void Copy(IPropertyDefinition propertyDefinition, long fromIndex, long toIndex)
        {
            Guard.ThrowExceptionIfNull(propertyDefinition, "propertyDefinition");
            ITranslateableCollection valueCollection = this.propertyToValueCollection[propertyDefinition];
            valueCollection.Copy(fromIndex, toIndex);
        }

        protected void Clear(long index)
        {
            foreach (ITranslateableCollection valueCollection in this.propertyToValueCollection.Values)
            {
                valueCollection.Clear(index);
            }
        }

        protected void Clear(long fromIndex, long toIndex)
        {
            foreach (ITranslateableCollection valueCollection in this.propertyToValueCollection.Values)
            {
                valueCollection.Clear(fromIndex, toIndex);
            }
        }

        protected void Translate(long fromIndex, long toIndex, long offset, bool respectUseSameValueAsPrevious)
        {
            //foreach (ITranslateableCollection valueCollection in this.propertyToValueCollection.Values)
            foreach (var keyValuePair in this.propertyToValueCollection)
            {
                bool useSameValueAsPreviousOnInsert = keyValuePair.Key.UseSameValueAsPreviousOnInsert;
                ITranslateableCollection valueCollection = keyValuePair.Value;
                valueCollection.Translate(fromIndex, toIndex, offset, !respectUseSameValueAsPrevious || useSameValueAsPreviousOnInsert);
            }
        }


        protected virtual void RegisterProperty<T>(IPropertyDefinition<T> property)
        {
            Guard.ThrowExceptionIfNull(property, "property");

            this.propertyToValueCollection.Add(property,
                new CompressedList<T>(this.FromItemIndex, this.ToItemIndex, property.DefaultValue));
        }

        protected internal ICompressedList GetPropertyValueCollection(IPropertyDefinition property)
        {
            Guard.ThrowExceptionIfNull(property, "property");

            return this.propertyToValueCollection[property];
        }

        protected internal ICompressedList<T> GetPropertyValueCollection<T>(IPropertyDefinition<T> property)
        {
            Guard.ThrowExceptionIfNull(property, "property");

            return (ICompressedList<T>)this.propertyToValueCollection[property];
        }


        public T GetDefaultPropertyValue<T>(IPropertyDefinition<T> property)
        {
            Guard.ThrowExceptionIfNull(property, "property");

            ICompressedList<T> propertyValueCollection = this.GetPropertyValueCollection(property);
            return propertyValueCollection.GetDefaultValue();
        }

        public void SetDefaultPropertyValue<T>(IPropertyDefinition<T> property, T defaultValue)
        {
            Guard.ThrowExceptionIfNull(property, "property");

            ICompressedList<T> propertyValueCollection = this.GetPropertyValueCollection(property);
            propertyValueCollection.SetDefaultValue(defaultValue);
        }

        protected internal T GetPropertyValue<T>(IPropertyDefinition<T> property, long index)
        {
            Guard.ThrowExceptionIfNull(property, "property");

            ICompressedList<T> propertyValueCollection = this.GetPropertyValueCollection(property);
            return propertyValueCollection.GetValue(index);
        }

        protected internal ICompressedList<T> GetPropertyValue<T>(IPropertyDefinition<T> property, long fromIndex, long toIndex)
        {
            Guard.ThrowExceptionIfNull(property, "property");

            ICompressedList<T> propertyValueCollection = this.GetPropertyValueCollection(property);
            return propertyValueCollection.GetValue(fromIndex, toIndex);
        }

        protected internal ICompressedList GetNonTypedPropertyValue(IPropertyDefinition property, long fromIndex, long toIndex)
        {
            Guard.ThrowExceptionIfNull(property, "property");

            ICompressedList propertyValueCollection = this.GetPropertyValueCollection(property);
            return propertyValueCollection.GetValue(fromIndex, toIndex);
        }

        protected void SetPropertyValue<T>(IPropertyDefinition<T> property, long index, T value)
        {
            this.SetPropertyValue(property, index, index, value);
        }

        protected void SetPropertyValue<T>(IPropertyDefinition<T> property, long fromIndex, long toIndex, T value)
        {
            Guard.ThrowExceptionIfNull(property, "property");

            ICompressedList<T> propertyValueCollection = this.GetPropertyValueCollection(property);
            propertyValueCollection.SetValue(fromIndex, toIndex, value);

            this.OnPropertyChanged(property, fromIndex, toIndex);
        }

        internal void SetNonTypedPropertyValue(IPropertyDefinition property, ICompressedList values)
        {
            Guard.ThrowExceptionIfNull(property, "property");
            Guard.ThrowExceptionIfNull(values, "values");

            ICompressedList propertyValueCollection = this.GetPropertyValueCollection(property);
            propertyValueCollection.SetValue(values);

            this.OnPropertyChanged(property, values.FromIndex, values.ToIndex);
        }

        protected void UpdatePropertyValue<T>(IPropertyDefinition<T> property, long fromIndex, long toIndex, Func<T, T> newValueTransform)
        {
            Guard.ThrowExceptionIfNull(property, "property");
            Guard.ThrowExceptionIfNull(newValueTransform, "newValueTransform");

            ICompressedList<T> propertyValueCollection = this.GetPropertyValueCollection(property);

            ICompressedList<T> valuesInRange = propertyValueCollection.GetValue(fromIndex, toIndex);
            ICompressedList<T> updatedValuesInRange = new CompressedList<T>(valuesInRange);
            foreach (var range in valuesInRange)
            {
                T newValue = newValueTransform(range.Value);
                if (!TelerikHelper.EqualsOfT(newValue, updatedValuesInRange.GetDefaultValue()))
                {
                    updatedValuesInRange.SetValue(range.Start, range.End, newValue);
                }
            }

            propertyValueCollection.SetValue(updatedValuesInRange);

            this.OnPropertyChanged(property, fromIndex, toIndex);
        }

        protected void SetPropertyValue(IPropertyDefinition property, ICompressedList values)
        {
            Guard.ThrowExceptionIfNull(property, "property");
            Guard.ThrowExceptionIfNull(values, "values");

            ICompressedList propertyValueCollection = this.GetPropertyValueCollection(property);
            propertyValueCollection.SetValue(values);

            this.OnPropertyChanged(property, values.FromIndex, values.ToIndex);
        }

        protected void ClearPropertyValue<T>(IPropertyDefinition<T> property, long index)
        {
            this.ClearPropertyValue(property, index, index);
        }

        protected void ClearPropertyValue<T>(IPropertyDefinition<T> property, long fromIndex, long toIndex)
        {
            Guard.ThrowExceptionIfNull(property, "property");

            ICompressedList<T> propertyValueCollection = this.GetPropertyValueCollection(property);
            propertyValueCollection.ClearValue(fromIndex, toIndex);

            this.OnPropertyChanged(property, fromIndex, toIndex);
        }

        internal void ClearPropertyValue(IPropertyDefinition property)
        {
            Guard.ThrowExceptionIfNull(property, "property");

            ICompressedList propertyValueCollection = this.GetPropertyValueCollection(property);
            long fromIndex = propertyValueCollection.FromIndex;
            long toIndex = propertyValueCollection.ToIndex;

            propertyValueCollection.ClearValue(fromIndex, toIndex);

            this.OnPropertyChanged(property, fromIndex, toIndex);
        }

        protected void CopyPropertiesFrom(PropertyBagBase fromProperties, long fromIndex, long toIndex)
        {
            foreach (IPropertyDefinition property in this.GetRegisteredProperties())
            {
                ICompressedList compressedList = fromProperties.GetPropertyValueCollection(property);
                this.SetPropertyValue(property, compressedList.GetValue(fromIndex, toIndex));
            }
        }

        protected virtual void OnPropertyChanged(IPropertyDefinition property, long fromIndex, long toIndex)
        {
        }

        #endregion
    }
}