﻿using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
using System;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.PropertySystem
{
    internal abstract class RowColumnPropertyBagBase : WorksheetPropertyBagBase
    {
        public static readonly IPropertyDefinition<bool> HiddenProperty =
            new PropertyDefinition<bool>("Hidden", true, StylePropertyGroup.None);

        public static readonly IPropertyDefinition<int> OutlineLevelProperty =
            new PropertyDefinition<int>("OutlineLevel", false, StylePropertyGroup.None);

        public static readonly int MaxOutlineLevel = 7;

        protected abstract WidthHeightBase HiddenSize { get; }

        #region Constructors

        public RowColumnPropertyBagBase()
        {
            base.RegisterProperty(RowColumnPropertyBagBase.HiddenProperty);
            base.RegisterProperty(RowColumnPropertyBagBase.OutlineLevelProperty);
        }

        #endregion


        #region Translation

        // TODO: Implement this method to prevent data loss.
        internal bool CanInsert(/*int index, int itemsCount*/)
        {
            return true;
        }

        public void Insert(int index, int itemsCount)
        {
            base.Translate(index, this.ToItemIndex, itemsCount, true);

            this.ClearPropertyValue(HiddenProperty, index, index + itemsCount - 1);
        }

        public void Remove(int index, int itemsCount)
        {
            base.Translate(index, this.ToItemIndex, -itemsCount, false);
        }

        #endregion


        #region Methods

        internal abstract int GetRowColumnIndex(int rowIndex, int columnIndex);

        internal int GetRowColumnIndex(CellIndex index)
        {
            return this.GetRowColumnIndex(index.RowIndex, index.ColumnIndex);
        }

        public T GetPropertyValue<T>(IPropertyDefinition<T> property, int index)
        {
            return base.GetPropertyValue(property, index);
        }

        public ICompressedList<T> GetPropertyValue<T>(IPropertyDefinition<T> property, int fromIndex, int toIndex)
        {
            Guard.ThrowExceptionIfLessThan(fromIndex, toIndex, "toIndex");
            return base.GetPropertyValue(property, fromIndex, toIndex);
        }

        public void SetPropertyValue<T>(IPropertyDefinition<T> property, int index, T value)
        {
            base.SetPropertyValue(property, index, value);
        }

        public void SetPropertyValue<T>(IPropertyDefinition<T> property, int fromIndex, int toIndex, T value)
        {
            base.SetPropertyValue(property, fromIndex, toIndex, value);
        }

        public void SetPropertyValue<T>(IPropertyDefinition<T> property, ICompressedList<T> values)
        {
            base.SetPropertyValue(property, values);
        }

        public void ClearPropertyValue<T>(IPropertyDefinition<T> property, int index)
        {
            base.ClearPropertyValue(property, index);
        }

        public void ClearPropertyValue<T>(IPropertyDefinition<T> property, int fromIndex, int toIndex)
        {
            Guard.ThrowExceptionIfLessThan(fromIndex, toIndex, "toIndex");
            base.ClearPropertyValue(property, fromIndex, toIndex);
        }

        protected override ICompressedList<T> GetLocalPropertyValueRespectingRowsColumns<T>(IPropertyDefinition<T> property, Worksheet worksheet, long fromIndex, long toIndex)
        {
            ICompressedList<T> localPropertyValue = this.GetPropertyValue(property, fromIndex, toIndex);

            return localPropertyValue;
        }

        public ICompressedList<T> GetSizePropertyValueRespectingHidden<T>(IPropertyDefinition<T> property)
            where T : WidthHeightBase
        {
            ICompressedList<T> actualPropertyValue = this.GetPropertyValueCollection(property);

            if (property != RowsPropertyBag.HeightProperty && property != ColumnsPropertyBag.WidthProperty)
            {
                return actualPropertyValue;
            }

            ICompressedList<bool> hiddenList = this.GetPropertyValueCollection(RowColumnPropertyBagBase.HiddenProperty);
            ICompressedList<T> result = new CompressedList<T>(actualPropertyValue);

            foreach (var range in actualPropertyValue.GetNonDefaultRanges())
            {
                result.SetValue(range.Start, range.End, range.Value);
            }

            foreach (Range<long, bool> range in hiddenList)
            {
                if (range.Value)
                {
                    T size = this.HiddenSize as T;
                    result.SetValue(range.Start, range.End, size);
                }
            }

            return result;
        }

        protected override void OnPropertyChanged(IPropertyDefinition property, long fromIndex, long toIndex)
        {
            RowColumnPropertyChangedEventArgs args = new RowColumnPropertyChangedEventArgs(property, (int)fromIndex, (int)toIndex);
            this.OnPropertyChanged(args);
        }

        #endregion


        #region Events

        public event EventHandler<RowColumnPropertyChangedEventArgs> PropertyChanged;
        private void OnPropertyChanged(RowColumnPropertyChangedEventArgs args)
        {
            Guard.ThrowExceptionIfNull(args, "args");

            if (this.PropertyChanged != null)
            {
                this.PropertyChanged(this, args);
            }
        }

        #endregion
    }
}