﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.PropertySystem
{
    internal abstract class WorksheetPropertyBagBase : PropertyBagBase
    {
        #region Fields

        private readonly Dictionary<IPropertyDefinition, CellRange> propertyToUsedCellRanges;

        #endregion


        #region Constructors

        public WorksheetPropertyBagBase()
        {
            this.propertyToUsedCellRanges = new Dictionary<IPropertyDefinition, CellRange>();

            this.RegisterProperty(CellPropertyDefinitions.FormatProperty);
            this.RegisterProperty(CellPropertyDefinitions.StyleNameProperty);
            this.RegisterProperty(CellPropertyDefinitions.LeftBorderProperty);
            this.RegisterProperty(CellPropertyDefinitions.TopBorderProperty);
            this.RegisterProperty(CellPropertyDefinitions.RightBorderProperty);
            this.RegisterProperty(CellPropertyDefinitions.BottomBorderProperty);
            this.RegisterProperty(CellPropertyDefinitions.DiagonalUpBorderProperty);
            this.RegisterProperty(CellPropertyDefinitions.DiagonalDownBorderProperty);
            this.RegisterProperty(CellPropertyDefinitions.FillProperty);
            this.RegisterProperty(CellPropertyDefinitions.FontFamilyProperty);
            this.RegisterProperty(CellPropertyDefinitions.FontSizeProperty);
            this.RegisterProperty(CellPropertyDefinitions.IsBoldProperty);
            this.RegisterProperty(CellPropertyDefinitions.IsItalicProperty);
            this.RegisterProperty(CellPropertyDefinitions.UnderlineProperty);
            this.RegisterProperty(CellPropertyDefinitions.ForeColorProperty);
            this.RegisterProperty(CellPropertyDefinitions.HorizontalAlignmentProperty);
            this.RegisterProperty(CellPropertyDefinitions.VerticalAlignmentProperty);
            this.RegisterProperty(CellPropertyDefinitions.IndentProperty);
            this.RegisterProperty(CellPropertyDefinitions.IsWrappedProperty);
            this.RegisterProperty(CellPropertyDefinitions.IsLockedProperty);
        }

        #endregion


        #region Methods

        internal static long ConvertCellIndexToLong(CellIndex index)
        {
            return ConvertCellIndexToLong(index.RowIndex, index.ColumnIndex);
        }

        internal static long ConvertCellIndexToLong(int rowIndex, int columnIndex)
        {
            return (long)columnIndex * SpreadsheetDefaultValues.RowCount + rowIndex;
        }

        internal static CellIndex ConvertLongToCellIndex(long index)
        {
            int rowIndex;
            int columnIndex;

            ConvertLongToRowAndColumnIndexes(index, out rowIndex, out columnIndex);

            return new CellIndex(rowIndex, columnIndex);
        }

        internal static void ConvertLongToRowAndColumnIndexes(long index, out int rowIndex, out int columnIndex)
        {
            rowIndex = (int)(index % SpreadsheetDefaultValues.RowCount);
            columnIndex = (int)(index / SpreadsheetDefaultValues.RowCount);
        }

        internal static IEnumerable<LongRange> ConvertCellRangeToLongRanges(CellRange cellRange)
        {
            int fromRowIndex = cellRange.FromIndex.RowIndex;
            int toRowIndex = cellRange.ToIndex.RowIndex;
            for (int i = cellRange.FromIndex.ColumnIndex; i <= cellRange.ToIndex.ColumnIndex; i++)
            {
                long fromIndex = ConvertCellIndexToLong(fromRowIndex, i);
                long toIndex = ConvertCellIndexToLong(toRowIndex, i);

                yield return new LongRange(fromIndex, toIndex);
            }
        }

        internal static IEnumerable<CellRange> ConvertCellRangeToLongCellRanges(CellRange cellRange)
        {
            int fromRowIndex = cellRange.FromIndex.RowIndex;
            int toRowIndex = cellRange.ToIndex.RowIndex;
            for (int i = cellRange.FromIndex.ColumnIndex; i <= cellRange.ToIndex.ColumnIndex; i++)
            {
                yield return new CellRange(fromRowIndex, i, toRowIndex, i);
            }
        }

        internal static CellRange ConvertLongCellRangeToCellRange(long fromIndex, long toIndex)
        {
            Guard.ThrowExceptionIfLessThan(fromIndex, toIndex, "toIndex");
            return new CellRange(ConvertLongToCellIndex(fromIndex), ConvertLongToCellIndex(toIndex));
        }

        public ICompressedList<T> GetPropertyValueRespectingStyle<T>(IPropertyDefinition<T> property, Worksheet worksheet, CellRange cellRange)
        {
            Guard.ThrowExceptionIfNull(cellRange, "cellRange");
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");

            long fromIndex = ConvertCellIndexToLong(cellRange.FromIndex);
            long toIndex = ConvertCellIndexToLong(cellRange.ToIndex);

            ICompressedList<T> result = new CompressedList<T>(fromIndex, toIndex, property.DefaultValue);

            foreach (LongRange longRange in ConvertCellRangeToLongRanges(cellRange))
            {
                ICompressedList<T> value = this.GetPropertyValueRespectingStyle(property, worksheet, longRange.Start, longRange.End);
                result.SetValue(value);
            }

            return result;
        }

        public T GetPropertyValueRespectingStyle<T>(IPropertyDefinition<T> property, Worksheet worksheet, long index)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");

            var propertyValueCollection = this.GetPropertyValueCollection(property);
            T value = propertyValueCollection.GetValue(index);
            bool hasValue = propertyValueCollection.ContainsNonDefaultValues(index, index);

            //return localPropertyValue;
            if (!CellStyle.IsSupportedProperty(property) || hasValue)
            {
                return value;
            }

            var styleNameCollection = this.GetPropertyValueCollection(CellPropertyDefinitions.StyleNameProperty);
            var styleName = styleNameCollection.GetValue(index);

            CellStyle style = worksheet.Workbook.Styles.GetByName(styleName);
            // TODO: What happens if the style does not exist?
            if (style != null)
            {
                if (style.GetIsPropertyIncluded(property) &&
                    !TelerikHelper.EqualsOfT(style.GetPropertyValue(property), this.GetDefaultPropertyValue(property)))
                {
                    return style.GetPropertyValue(property);
                }
                else
                {
                    T result;
                    if (this.TryGetRowOrColumnPropertyValueRespectingStyle<T>(property, worksheet, index, out result))
                    {
                        return result;
                    }
                }
            }

            return value;
        }

        public ICompressedList<T> GetPropertyValueRespectingStyle<T>(IPropertyDefinition<T> property, Worksheet worksheet, long fromIndex, long toIndex)
        {
            Guard.ThrowExceptionIfLessThan(fromIndex, toIndex, "toIndex");
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");

            //return localPropertyValue;
            if (!CellStyle.IsSupportedProperty(property))
            {
                return this.GetLocalPropertyValueRespectingRowsColumns<T>(property, worksheet, fromIndex, toIndex);
            }

            ICompressedList<string> styleNameList = this.GetPropertyValue(CellPropertyDefinitions.StyleNameProperty, fromIndex, toIndex);

            ICompressedList<T> localPropertyValue = this.GetPropertyValue(property, fromIndex, toIndex);
            ICompressedList<T> result = new CompressedList<T>(localPropertyValue);

            foreach (Range<long, string> range in styleNameList)
            {
                CellStyle style = worksheet.Workbook.Styles.GetByName(range.Value);
                // TODO: What happens if the style does not exist?
                if (style == null)
                {
                    continue;
                }

                bool isDefaultStyle = range.IsDefault;
                if (style.GetIsPropertyIncluded(property) && !isDefaultStyle)
                {
                    result.SetValue(range.Start, range.End, style.GetPropertyValue(property));
                }
                else
                {
                    ICompressedList<T> rowOrColumnPropertyValuesRespectingStyle;
                    if (this.TryGetRowOrColumnPropertyValueRespectingStyle<T>(property, worksheet, range.Start, range.End, out rowOrColumnPropertyValuesRespectingStyle))
                    {
                        foreach (Range<long, T> rowOrColumnPropertyValueRange in rowOrColumnPropertyValuesRespectingStyle.GetNonDefaultRanges())
                        {
                            result.SetValue(rowOrColumnPropertyValueRange.Start, rowOrColumnPropertyValueRange.End, rowOrColumnPropertyValueRange.Value);
                        }
                    }
                }
            }

            foreach (var range in localPropertyValue.GetNonDefaultRanges())
            {
                result.SetValue(range.Start, range.End, range.Value);
            }

            return result;
        }

        protected abstract ICompressedList<T> GetLocalPropertyValueRespectingRowsColumns<T>(IPropertyDefinition<T> property, Worksheet worksheet, long fromIndex, long toIndex);

        // This have sense to be overridden only in CellsPropertyBag.
        protected virtual bool TryGetRowOrColumnPropertyValueRespectingStyle<T>(IPropertyDefinition<T> property, Worksheet worksheet, long index, out T result)
        {
            result = default(T);
            return false;
        }

        // This have sense to be overridden only in CellsPropertyBag.
        protected virtual bool TryGetRowOrColumnPropertyValueRespectingStyle<T>(IPropertyDefinition<T> property, Worksheet worksheet, long fromIndex, long toIndex, out ICompressedList<T> rowOrColumnPropertyValuesRespectingStyle)
        {
            rowOrColumnPropertyValuesRespectingStyle = default(ICompressedList<T>);
            return false;
        }

        #endregion


        #region Used Cell Range Calculation

        internal void InvalidateAllPropertiesUsedCellRange()
        {
            foreach (IPropertyDefinition property in this.propertyToUsedCellRanges.Keys.ToList())
            {
                this.InvalidatePropertyUsedCellRange(property);
            }
        }

        protected void InvalidatePropertyUsedCellRange(IPropertyDefinition property)
        {
            this.propertyToUsedCellRanges.Remove(property);
        }

        protected CellRange GetUsedCellRange(IPropertyDefinition property)
        {
            int currentColumnIndex = SpreadsheetDefaultValues.ColumnCount - 1;
            CellRange cellRange = new CellRange(0, 0, SpreadsheetDefaultValues.RowCount - 1, SpreadsheetDefaultValues.ColumnCount - 1);
            List<LongRange> longRanges = new List<LongRange>(ConvertCellRangeToLongRanges(cellRange));
            longRanges.Reverse();

            int rowIndex = 0;
            int columnIndex = 0;
            foreach (var range in longRanges)
            {
                ICompressedList compressedList = this.GetPropertyValueCollection(property);
                long? lastNonDefaultIndexInRange = compressedList.GetLastNonEmptyRangeEnd(range.Start, range.End);
                if (lastNonDefaultIndexInRange.HasValue)
                {
                    rowIndex = (int)Math.Max(rowIndex, lastNonDefaultIndexInRange.Value - range.Start);
                    columnIndex = Math.Max(columnIndex, currentColumnIndex);
                }
                currentColumnIndex--;
            }

            return new CellRange(0, 0, rowIndex, columnIndex);
        }

        public CellRange GetUsedCellRange(IEnumerable<IPropertyDefinition> propertyDefinitions)
        {
            Guard.ThrowExceptionIfNull(propertyDefinitions, "propertyDefinitions");

            List<Action> actions = new List<Action>();

            foreach (IPropertyDefinition property in propertyDefinitions)
            {
                if (!this.propertyToUsedCellRanges.ContainsKey(property))
                {
                    Action action = new Action(() =>
                    {
                        CellRange temp = this.GetUsedCellRange(property);
                        lock (this.propertyToUsedCellRanges)
                        {
                            this.propertyToUsedCellRanges[property] = temp;
                        }
                    });

                    actions.Add(action);
                }
            }

            if (actions.Count > 0)
            {
                Documents.Utilities.TasksHelper.DoAsync(actions);
            }

            int toRowIndex = 0;
            int toColumnIndex = 0;
            foreach (var range in this.propertyToUsedCellRanges)
            {
                if (propertyDefinitions.Contains(range.Key) && range.Value != null)
                {
                    toRowIndex = Math.Max(toRowIndex, range.Value.ToIndex.RowIndex);
                    toColumnIndex = Math.Max(toColumnIndex, range.Value.ToIndex.ColumnIndex);
                }
            }

            return new CellRange(0, 0, toRowIndex, toColumnIndex);
        }

        public CellRange GetUsedCellRange()
        {
            IEnumerable<IPropertyDefinition> properties = this.GetRegisteredProperties();
            return this.GetUsedCellRange(properties);
        }

        #endregion
    }
}