﻿using System;
using System.Windows.Threading;
using System.Threading;

namespace Telerik.Windows.Documents.Spreadsheet.Utilities
{
	internal class DelayedExecution : IDisposable
	{
		private class HaltSwitch
		{
			public bool halted;
		}

		#region Fields

		private readonly Action action;
		private readonly DispatcherTimer timer;
		private Timer threadingTimer;
		private readonly TimeSpan timerInterval;
		private readonly bool useThreadingTimer;
		private readonly Dispatcher dispatcher;
		private HaltSwitch halt;
        private bool alreadyDisposed;

		#endregion


		#region Properties

		public TimeSpan WaitInterval
		{
			get
			{
				return this.timer.Interval;
			}
			set
			{
				this.timer.Interval = value;
			}
		}

		public DelayedExecution NextAction { get; set; }

		#endregion


		#region Constructors

		public DelayedExecution(Action action, long milliseconds)
			: this(action, milliseconds, false, null)
		{
		}

		public DelayedExecution(Action action, long milliseconds, bool useThreadingTimer, Dispatcher dispatcher)
		{
			this.dispatcher = dispatcher;
			this.action = action;
			this.useThreadingTimer = useThreadingTimer;
			this.timerInterval = TimeSpan.FromMilliseconds(milliseconds);

			if (!useThreadingTimer)
			{
				this.timer = new DispatcherTimer();
				this.timer.Interval = timerInterval;
				this.timer.Tick += Timer_Tick;
			}
		}

		#endregion


        #region Methods

        public void Execute()
        {
            this.EnsureNotDisposed();

            this.Execute(true);
        }

        public void Execute(bool executeAsynch)
        {
            this.EnsureNotDisposed();

            if (executeAsynch)
            {
                if (useThreadingTimer)
                {
                    this.threadingTimer = new System.Threading.Timer(OnThreadingTimer, null, timerInterval, timerInterval);
                }
                else if (!this.timer.IsEnabled)
                {
                    this.timer.Start();
                }
            }
            else
            {
                this.ExecuteAction(false);
            }
        }

        private void Timer_Tick(object sender, EventArgs e)
        {
            timer.Stop();

            if (this.GetHaltSwitch().halted)
            {
                return;
            }

            this.ExecuteAction(true);
        }


        private void OnThreadingTimer(object state)
        {
            threadingTimer.Dispose();

            if (this.GetHaltSwitch().halted)
            {
                return;
            }

            try
            {
                this.ExecuteAction(true);
            }
            catch
            {
            }
        }

        private void ExecuteAction(bool asynch)
        {
            this.action();

            if (this.NextAction != null)
            {
                NextAction.halt = this.GetHaltSwitch();
                Action executeNext = () =>
                {
                    this.NextAction.Execute(asynch);
                };

                if (asynch && this.dispatcher != null)
                {
                    dispatcher.BeginInvoke(executeNext);
                }
                else
                {
                    executeNext();
                }
            }
        }

        private HaltSwitch GetHaltSwitch()
        {
            if (this.halt == null)
            {
                this.halt = new HaltSwitch();
            }

            return this.halt;
        }

        internal void Halt()
        {
            this.EnsureNotDisposed();

            this.GetHaltSwitch().halted = true;
        }

        public void Dispose()
        {
            if (this.alreadyDisposed)
            {
                return;
            }

            if (this.threadingTimer != null)
            {
                this.threadingTimer.Dispose();
            }

            if (this.timer != null)
            {
                this.timer.Stop();
                this.timer.Tick -= Timer_Tick;
            }
            GC.SuppressFinalize(this);

            this.alreadyDisposed = true;
        }

        private void EnsureNotDisposed()
        {
            if (this.alreadyDisposed)
            {
                throw new ObjectDisposedException("DelayedExecution");
            }
        }

        #endregion
    }
}