﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Media;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.Utilities
{
    internal static class ExtensionMethods
    {
        #region Methods

        internal static bool IsDateFormat(this CellValueFormat format)
        {
            return format.FormatStringInfo.FormatType == FormatStringType.DateTime;
        }

        public static string GetAsStringOrNull(this ICellValue cellValue, CellValueFormat format)
        {
            return cellValue != null ? cellValue.GetValueAsString(format) : null;
        }

        public static double? GetAsDoubleOrNull(this ICellValue cellValue)
        {
            NumberCellValue numberValue = cellValue as NumberCellValue;
            if (numberValue != null)
            {
                return numberValue.Value;
            }

            FormulaCellValue formulaValue = cellValue as FormulaCellValue;
            if (formulaValue != null)
            {
                RadExpression expressionValue = formulaValue.Value.GetValue();
                if (expressionValue is NumberExpression)
                {
                    return expressionValue.NumberValue();
                }

                ArrayExpression arrayExoression = expressionValue as ArrayExpression;
                if (arrayExoression != null && arrayExoression.Value.Length == 1)
                {
                    return GetSingleValueArrayValueAsDoubleOrNull(arrayExoression);
                }
            }

            return null;
        }

        private static double? GetSingleValueArrayValueAsDoubleOrNull(RadExpression expressionValue)
        {
            ArrayExpression arrayExoression = expressionValue as ArrayExpression;
            if (arrayExoression != null && arrayExoression.Value.Length == 1)
            {
                RadExpression value = arrayExoression.Value[0, 0].GetValue();
                return GetSingleValueArrayValueAsDoubleOrNull(value);
            }

            if (expressionValue is NumberExpression)
            {
                return expressionValue.NumberValue();
            }

            return null;
        }

        public static Color Darker(this Color color)
        {
            return color.Darker(24);
        }

        public static Color Darker(this Color color, byte faktor)
        {
            if (color.R - faktor >= 0)
            {
                color.R -= faktor;
            }
            else
            {
                color.R = 0;
            }

            if (color.G - faktor >= 0)
            {
                color.G -= faktor;
            }
            else
            {
                color.G = 0;
            }

            if (color.B - faktor >= 0)
            {
                color.B -= faktor;
            }
            else
            {
                color.B = 0;
            }

            return color;
        }

        public static Color Lighter(this Color color)
        {
            return color.Lighter(24);
        }

        public static Color Lighter(this Color color, byte faktor)
        {
            if (color.R + faktor < 255)
            {
                color.R += faktor;
            }
            else
            {
                color.R = 255;
            }

            if (color.G + faktor < 255)
            {
                color.G += faktor;
            }
            else
            {
                color.G = 255;
            }

            if (color.B + faktor < 255)
            {
                color.B += faktor;
            }
            else
            {
                color.B = 255;
            }

            return color;
        }

        public static string GetSourceOrNull(this FontFamily family)
        {
            if (family == null)
            {
                return null;
            }

            return family.Source;
        }

        public static int GetHashCodeOrZero(this object obj)
        {
            if (obj == null)
            {
                return 0;
            }

            return obj.GetHashCode();
        }

        public static bool EqualsDouble(this double d1, double d2)
        {
            //double diff = Math.Abs(Math.Abs(d1) - Math.Abs(d2));

            //return diff <= 1E-8;
            return Math.Abs(d1 - d2) <= ((Math.Abs(d1) > Math.Abs(d2) ? Math.Abs(d2) : Math.Abs(d1)) * 1E-8);
        }

        public static bool GreaterOrEqualDouble(this double d1, double d2)
        {
            return d1.EqualsDouble(d2) || d1 > d2;
        }

        public static bool LessOrEqualDouble(this double d1, double d2)
        {
            return d1.EqualsDouble(d2) || d1 < d2;
        }

        #endregion


        #region String operations
        
        /// <summary>
        /// Performs a case sensitive comparison of two strings. It ignores the different types of space separators that may be included in the string.
        /// </summary>
        /// <param name="value">The string.</param>
        /// <param name="otherValue">The string the string is compared to.</param>
        /// <returns>The value indicating whether the strings are the same.</returns>
        public static bool CompareIgnoringSpecialSymbols(this string value, string otherValue)
        {
            Guard.ThrowExceptionIfNull(value, "value");
            Guard.ThrowExceptionIfNull(otherValue, "otherValue");

            List<char> specialSymbols = new List<char> { '"', '_', '\\' };

            StringBuilder sbStrippedValue = new StringBuilder();
            for (int i = 0; i < value.Length; i++)
            {
                if (!specialSymbols.Contains(value[i]))
                {
                    sbStrippedValue.Append(value[i]);
                }
            }

            StringBuilder sbStrippedOtherValue = new StringBuilder();
            for (int i = 0; i < otherValue.Length; i++)
            {
                if (!specialSymbols.Contains(otherValue[i]))
                {
                    sbStrippedOtherValue.Append(otherValue[i]);
                }
            }

            string strippedValue1 = sbStrippedValue.ToString();
            string strippedValue = sbStrippedOtherValue.ToString();

            if (strippedValue1.Length != strippedValue.Length)
            {
                return false;
            }

            for (int i = 0; i < strippedValue1.Length; i++)
            {
                char char1 = strippedValue1[i];
                char char2 = strippedValue[i];
                
                if (char1 != char2)
                {
                    if (char.GetUnicodeCategory(char1) != UnicodeCategory.SpaceSeparator ||
                        char.GetUnicodeCategory(char2) != UnicodeCategory.SpaceSeparator)
                    {
                        return false;
                    }
                }
            }

            return true;
        }


        #endregion


        #region Stream operations

        internal static MemoryStream ToMemoryStream(this Stream stream)
        {
            MemoryStream result = new MemoryStream();
            stream.CopyTo(result);
            result.Seek(0, SeekOrigin.Begin);
            return result;
        }

        #endregion


        #region DateTime operations

        public static DateTime GetBeginningOfWeek(this DateTime date)
        {
            DayOfWeek day = date.DayOfWeek;
            return date.AddDays(-(int)day);
        }

        public static int GetQuarter(this DateTime date)
        {
            if(date.Month >= 1 && date.Month <= 3)
            {
                return 1;
            }
		    else if(date.Month >= 4 && date.Month <= 6)
            {
                return 2;
            }
            else if (date.Month >= 7 && date.Month <= 9)
            {
                return 3;
            }
            else //date.Month >= 10 && date.Month <= 12;
            {
                return 4;
            }
        }

        public static bool DateIsNextQuarter(this DateTime otherDate)
        {
            DateTime currentDate = DateTime.Today;

            int currentQuarter = currentDate.GetQuarter();
            int otherQuarter = otherDate.GetQuarter();

            int difference = otherQuarter - currentQuarter;

            if ((difference == 1 && currentDate.Year == otherDate.Year) ||
                (difference == -3 && currentDate.Year + 1 == otherDate.Year))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        public static bool DateIsLastQuarter(this DateTime otherDate)
        {
            DateTime currentDate = DateTime.Today;

            int currentQuarter = currentDate.GetQuarter();
            int otherQuarter = otherDate.GetQuarter();

            int difference = currentQuarter - otherQuarter;

            if ((difference == 1 && currentDate.Year == otherDate.Year) ||
                (difference == -3 && currentDate.Year - 1 == otherDate.Year))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        #endregion


        #region CellRange operations

        public static bool ContainsOverlappingRanges(this IEnumerable<CellRange> cellRanges)
        {
            var cellRangesList = cellRanges.ToList();
            for (int i = 0; i < cellRangesList.Count; i++)
            {
                for (int j = i + 1; j < cellRangesList.Count; j++)
                {
                    if (cellRangesList[i].IntersectsWith(cellRangesList[j]))
                    {
                        return true;
                    }
                }
            }

            return false;
        }

        public static bool IntersectsWith(this CellRange cellrange, int fromRowIndex, int fromColumnIndex, int toRowIndex, int toColumnIndex)
        {
            Guard.ThrowExceptionIfNull(cellrange, "cellrange");

            return (fromColumnIndex <= cellrange.ToIndex.ColumnIndex) &&
                (toColumnIndex >= cellrange.FromIndex.ColumnIndex) &&
                (fromRowIndex <= cellrange.ToIndex.RowIndex) &&
                (toRowIndex >= cellrange.FromIndex.RowIndex);
        }

        #endregion
    }
}