﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.Utilities
{
    /// <summary>
    /// Provides methods for checking whether the parameters of methods fall within the expected conditions.
    /// </summary>
    public static class Guard
    {
        #region Methods

        /// <summary>
        /// Asserts the specified condition.
        /// </summary>
        /// <param name="condition">The condition.</param>
        /// <param name="message">The message to throw in case the condition is not met.</param>
        public static void Assert(bool condition, string message)
        {
            if (!condition)
            {
                throw new InvalidOperationException(message);
            }
        }

        /// <summary>
        /// Throws an exception if the parameter is less than the lower bound.
        /// </summary>
        /// <typeparam name="T">The type of the parameter.</typeparam>
        /// <param name="lowerBound">The lower bound.</param>
        /// <param name="param">The parameter.</param>
        /// <param name="paramName">Name of the parameter.</param>
        public static void ThrowExceptionIfLessThan<T>(T lowerBound, T param, string paramName)
            where T : IComparable<T>
        {
            if (lowerBound.CompareTo(param) > 0)
            {
                throw new ArgumentOutOfRangeException(paramName,
                    string.Format("{0} should be greater or equal to {1}.",
                    paramName, lowerBound));
            }
        }

        /// <summary>
        /// Throws an exception if the parameters is greater than the upper bound.
        /// </summary>
        /// <typeparam name="T">The type of the parameter.</typeparam>
        /// <param name="upperBound">The upper bound.</param>
        /// <param name="param">The parameter.</param>
        /// <param name="paramName">Name of the parameter.</param>
        public static void ThrowExceptionIfGreaterThan<T>(T upperBound, T param, string paramName)
            where T : IComparable<T>
        {
            if (upperBound.CompareTo(param) < 0)
            {
                throw new ArgumentOutOfRangeException(paramName,
                    string.Format("{0} should be less or equal to {1}.",
                    paramName, upperBound));
            }
        }

        /// <summary>
        /// Throws an exception if out of range.
        /// </summary>
        /// <typeparam name="T">The type of the parameters.</typeparam>
        /// <param name="lowerBound">The lower bound.</param>
        /// <param name="upperBound">The upper bound.</param>
        /// <param name="param">The parameter.</param>
        /// <param name="paramName">Name of the parameter.</param>
        public static void ThrowExceptionIfOutOfRange<T>(T lowerBound, T upperBound, T param, string paramName)
            where T : IComparable<T>
        {
            if (param.CompareTo(lowerBound) < 0 || param.CompareTo(upperBound) > 0)
            {
                throw new ArgumentOutOfRangeException(paramName,
                    string.Format("{0} should be greater or equal than {1} and less or equal than {2}.",
                    paramName, lowerBound, upperBound));
            }
        }

        /// <summary>
        /// Throws an exception if the parameter is not equal to a value.
        /// </summary>
        /// <typeparam name="T">The type of the parameters.</typeparam>
        /// <param name="expected">The expected value.</param>
        /// <param name="param">The parameters.</param>
        /// <param name="paramName">Name of the parameter.</param>
        public static void ThrowExceptionIfNotEqual<T>(T expected, T param, string paramName)
        {
            if (!TelerikHelper.EqualsOfT(expected, param))
            {
                throw new ArgumentException(paramName, string.Format("{0} should be equal to {1}.",
                    paramName, param));
            }
        }

        /// <summary>
        /// Throws an exception if the parameter is null.
        /// </summary>
        /// <typeparam name="T">The type of the parameter.</typeparam>
        /// <param name="param">The parameter.</param>
        /// <param name="paramName">Name of the parameter.</param>
        public static void ThrowExceptionIfNull<T>(T param, string paramName)
        {
            if (param == null)
            {
                throw new ArgumentNullException(paramName);
            }
        }

        /// <summary>
        /// Throws an exception if the parameter is not null.
        /// </summary>
        /// <typeparam name="T">The type of the parameter.</typeparam>
        /// <param name="param">The parameter.</param>
        /// <param name="paramName">Name of the parameter.</param>
        public static void ThrowExceptionIfNotNull<T>(T param, string paramName)
        {
            if (param != null)
            {
                throw new ArgumentException(paramName + " should be null");
            }
        }

        /// <summary>
        /// Throws an exception if the parameter contains null or empty values.
        /// </summary>
        /// <param name="param">The parameter.</param>
        /// <param name="paramName">Name of the parameter.</param>
        public static void ThrowExceptionIfContainsNullOrEmpty(IEnumerable<string> param, string paramName)
        {
            foreach (var p in param)
            {
                if (string.IsNullOrEmpty(p))
                {
                    throw new ArgumentException(paramName);
                }
            }
        }

        /// <summary>
        /// Throws an exception if the parameter is null or empty.
        /// </summary>
        /// <typeparam name="T">The type of the parameter.</typeparam>
        /// <param name="param">The parameter.</param>
        /// <param name="paramName">Name of the parameter.</param>
        public static void ThrowExceptionIfNullOrEmpty<T>(IEnumerable<T> param, string paramName)
            where T : class
        {
            if (param == null || param.Count() == 0)
            {
                throw new ArgumentException(paramName);
            }
        }

        /// <summary>
        /// Throws an exception if the parameter contains null values.
        /// </summary>
        /// <typeparam name="T">The type of the parameter.</typeparam>
        /// <param name="param">The parameter.</param>
        /// <param name="paramName">Name of the parameter.</param>
        public static void ThrowExceptionIfContainsNull<T>(IEnumerable<T> param, string paramName)
            where T : class
        {
            foreach (var p in param)
            {
                if (p == null)
                {
                    throw new ArgumentException(paramName);
                }
            }
        }

        /// <summary>
        /// Throws the exception if the parameter is null or empty.
        /// </summary>
        /// <param name="param">The parameter.</param>
        /// <param name="paramName">Name of the parameter.</param>
        public static void ThrowExceptionIfNullOrEmpty(string param, string paramName)
        {
            Guard.ThrowExceptionIfNull(param, paramName);

            if (string.IsNullOrEmpty(param))
            {
                throw new ArgumentException(paramName);
            }
        }

        internal static void ThrowExceptionIfTrue(bool param, string paramName)
        {
            if (param)
            {
                throw new InvalidOperationException(string.Format("{0} must be false.",
                    paramName));
            }
        }

        internal static void ThrowExceptionIfFalse(bool param, string paramName)
        {
            if (!param)
            {
                throw new InvalidOperationException(string.Format( "{0} must be true.", paramName));
            }
        }

        internal static void ThrowExceptionIfInvalidDouble(double argument, string argumentName)
        {
            if (double.IsInfinity(argument) || double.IsNaN(argument))
            {
                throw new ArgumentException(string.Format("{0} must be a valid double.", argumentName), argumentName);
            }
        }

        internal static void ThrowExceptionIfNotSupportedType<T>(object param, string paramName)
            where T : class
        {
            if (!(param is T))
            {
                throw new NotSupportedException(string.Format("{0} is not supported. The expected value type is: {1}", 
                    paramName,
                    typeof(T)));
            }
        }

        #endregion


        #region Spreadsheet Specific

        /// <summary>
        /// Throws an exception if the row index is invalid.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        public static void ThrowExceptionIfInvalidRowIndex(int rowIndex)
        {
            Guard.ThrowExceptionIfOutOfRange(0, SpreadsheetDefaultValues.RowCount - 1, rowIndex, "rowIndex");
        }

        /// <summary>
        /// Throws an exception if the column index is invalid.
        /// </summary>
        /// <param name="columnIndex">Index of the column.</param>
        public static void ThrowExceptionIfInvalidColumnIndex(int columnIndex)
        {
            Guard.ThrowExceptionIfOutOfRange(0, SpreadsheetDefaultValues.ColumnCount - 1, columnIndex, "columnIndex");
        }

        internal static void ThrowExceptionIfInvalidColumnWidth(double columnWidth)
        {
            Guard.ThrowExceptionIfOutOfRange(0, SpreadsheetDefaultValues.MaxColumnWidth, columnWidth, "columnWidth");
        }

        internal static void ThrowExceptionIfInvalidRowHeight(double rowHeight)
        {
            Guard.ThrowExceptionIfOutOfRange(0, SpreadsheetDefaultValues.MaxRowHeight, rowHeight, "rowHeight");
        }

        internal static void ThrowExceptionIfInvalidIndent(int indent)
        {
            Guard.ThrowExceptionIfOutOfRange(0, SpreadsheetDefaultValues.MaxIndent, indent, "indent");
        }

        internal static void ThrowExceptionIfInvalidScaleFactor(Size scaleFactor)
        {
            ThrowExceptionIfInvalidScaleFactor(scaleFactor.Width);
            ThrowExceptionIfInvalidScaleFactor(scaleFactor.Height);
        }

        internal static void ThrowExceptionIfInvalidScaleFactor(double scaleFactor)
        {
            Guard.ThrowExceptionIfOutOfRange(SpreadsheetDefaultValues.MinScaleFactor, SpreadsheetDefaultValues.MaxScaleFactor, scaleFactor, "scaleFactor");
        }

        #endregion
    }
}