﻿using System;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.Utilities
{
    /// <summary>
    /// Provides methods for name conversion.
    /// </summary>
    public static class NameConverter
    {
        #region Constants

        private static readonly int LatinAlphabetLettersCount = ('Z' - 'A' + 1);

        #endregion


        #region Methods

        /// <summary>
        /// Converts the row index to name.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <returns>The name.</returns>
        public static string ConvertRowIndexToName(int rowIndex)
        {
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);
            rowIndex++;
            return rowIndex.ToString();
        }

        /// <summary>
        /// Converts the row name to index.
        /// </summary>
        /// <param name="rowName">Name of the row.</param>
        /// <returns>The index.</returns>
        public static int ConvertRowNameToIndex(string rowName)
        {
            int rowIndex;
            if (!int.TryParse(rowName, out rowIndex))
            {
                throw new LocalizableException(string.Format("'{0}' is invalid row name.", rowName),
                    new InvalidOperationException(string.Format("'{0}' is invalid row name.", rowName)),
                    "Spreadsheet_ErrorExpressions_InvalidRowName",
                    new string[] { rowName });
            }

            rowIndex--;
            Guard.ThrowExceptionIfInvalidRowIndex(rowIndex);

            return rowIndex;
        }

        /// <summary>
        /// Converts the row name to index.
        /// </summary>
        /// <param name="rowName">Name of the row.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <returns>Boolean value indicating whether the convertion is successful.</returns>
        internal static bool TryConvertRowNameToIndex(string rowName, out int rowIndex)
        {
            rowIndex = -1;

            if (!IsValidRowName(rowName))
            {
                return false;
            }

            if (int.TryParse(rowName, out rowIndex))
            {
                rowIndex--;
                return rowIndex >= 0 && rowIndex <= SpreadsheetDefaultValues.RowCount - 1;
            }

            return false;
        }

        /// <summary>
        /// Converts the column index to name.
        /// </summary>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The name.</returns>
        public static string ConvertColumnIndexToName(int columnIndex)
        {
            Guard.ThrowExceptionIfInvalidColumnIndex(columnIndex);

            string result = string.Empty;
            while (columnIndex >= 26)
            {
                result = (char)((columnIndex % LatinAlphabetLettersCount) + 'A') + result;
                columnIndex = (columnIndex / LatinAlphabetLettersCount) - 1;
            }

            result = (char)((columnIndex % LatinAlphabetLettersCount) + 'A') + result;

            return result;
        }

        /// <summary>
        /// Converts the column name to index.
        /// </summary>
        /// <param name="columnName">Name of the column.</param>
        /// <returns>The index.</returns>
        public static int ConvertColumnNameToIndex(string columnName)
        {
            int result;
            if (!TryConvertColumnNameToIndexInternal(columnName, out result))
            {
                throw new LocalizableException(string.Format("'{0}' is invalid column name.", columnName),
                    new InvalidOperationException(string.Format("'{0}' is invalid column name.", columnName)),
                    "Spreadsheet_ErrorExpressions_InvalidColumnName",
                    new string[] { columnName });
            }

            Guard.ThrowExceptionIfInvalidColumnIndex(result);

            return result;
        }

        /// <summary>
        /// Converts the cell ranges names to cell reference ranges.
        /// </summary>
        /// <param name="cellRangesNames">The cell ranges names.</param>
        /// <param name="worksheet">The worksheet.</param>
        /// <param name="rowIndex">The row index of the cell for which the expression is created.</param>
        /// <param name="columnIndex">The column index of the cell for which the expression is created.</param>
        /// <param name="expression">The expression.</param>
        /// <returns>True if converter successfuly, else false.</returns>
        public static bool TryConvertNamesToCellReferenceRangeExpression(
            string cellRangesNames, Worksheet worksheet, int rowIndex, int columnIndex, out CellReferenceRangeExpression expression)
        {
            expression = null;

            if (string.IsNullOrEmpty(cellRangesNames))
            {
                return false;
            }

            string equalSign = SpreadsheetCultureHelper.Equal;
            if (cellRangesNames.Length > 1 && cellRangesNames[0].ToString() != equalSign)
            {
                cellRangesNames = equalSign + cellRangesNames;
            }

            expression = null;
            RadExpression parseResultExpression;
            var parseResult = RadExpression.TryParse(cellRangesNames, worksheet, rowIndex, columnIndex, out parseResultExpression);
            if (parseResult == ParseResult.Successful)
            {
                expression = parseResultExpression as CellReferenceRangeExpression;
                if (expression == null)
                {
                    UnionExpression unionExpression = parseResultExpression as UnionExpression;
                    if (unionExpression != null)
                    {
                        expression = unionExpression.GetValue() as CellReferenceRangeExpression;
                    }
                }
            }

            return expression != null;
        }

        /// <summary>
        /// Converts the column name to index.
        /// </summary>
        /// <param name="columnName">Name of the column.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The index.</returns>
        internal static bool TryConvertColumnNameToIndex(string columnName, out int columnIndex)
        {
            if (!TryConvertColumnNameToIndexInternal(columnName, out columnIndex))
            {
                return false;
            }

            return columnIndex >= 0 && columnIndex <= SpreadsheetDefaultValues.ColumnCount - 1;
        }

        private static bool TryConvertColumnNameToIndexInternal(string columnName, out int index)
        {
            index = -1;

            if (!IsValidColumnName(columnName))
            {
                return false;
            }

            StringBuilder stringBuilder = new StringBuilder();
            for (int i = 0; i < columnName.Length; i++)
            {
                stringBuilder.Append(char.ToUpperInvariant(columnName[i]));
            }

            string columnNameToUpper = stringBuilder.ToString();

            int multiplier = LatinAlphabetLettersCount;
            index = (columnNameToUpper[columnNameToUpper.Length - 1] - 'A');
            for (int i = columnNameToUpper.Length - 2; i >= 0; i--)
            {
                index += multiplier * (columnNameToUpper[i] - 'A' + 1);
                multiplier *= LatinAlphabetLettersCount;
            }

            return index <= SpreadsheetDefaultValues.ColumnCount - 1;
        }

        /// <summary>
        /// Converts the cell index to name.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The name.</returns>
        public static string ConvertCellIndexToName(int rowIndex, int columnIndex)
        {
            string rowName = ConvertRowIndexToName(rowIndex);
            string columnName = ConvertColumnIndexToName(columnIndex);

            return columnName + rowName;
        }

        internal static string ConvertCellIndexToAbsoluteName(CellIndex cellIndex)
        {
            string rowName = ConvertRowIndexToName(cellIndex.RowIndex);
            string columnName = ConvertColumnIndexToName(cellIndex.ColumnIndex);

            StringBuilder builder = new StringBuilder();

            builder.Append("$");
            builder.Append(columnName);
            builder.Append("$");
            builder.Append(rowName);

            return builder.ToString();
        }

        /// <summary>
        /// Converts the cell reference to name.
        /// </summary>
        /// <param name="cellReference">The cell reference.</param>
        /// <returns>The name.</returns>
        public static string ConvertCellReferenceToName(CellReference cellReference)
        {
            StringBuilder builder = new StringBuilder();

            if (cellReference.ColumnReference.IsAbsolute)
            {
                builder.Append('$');
            }
            builder.Append(ConvertColumnIndexToName(cellReference.ColumnReference.ActualIndex));

            if (cellReference.RowReference.IsAbsolute)
            {
                builder.Append('$');
            }
            builder.Append(ConvertRowIndexToName(cellReference.RowReference.ActualIndex));

            return builder.ToString();

        }

        /// <summary>
        /// Converts the cell range to a name.
        /// </summary>
        /// <param name="fromIndex">From index.</param>
        /// <param name="toIndex">To index.</param>
        /// <returns>The name.</returns>
        public static string ConvertCellRangeToName(CellIndex fromIndex, CellIndex toIndex)
        {
            Guard.ThrowExceptionIfNull(fromIndex, "fromIndex");
            Guard.ThrowExceptionIfNull(toIndex, "toIndex");

            return ConvertCellIndexesToName(fromIndex.RowIndex, fromIndex.ColumnIndex, toIndex.RowIndex, toIndex.ColumnIndex);
        }

        /// <summary>
        /// Converts a range defined between cell references to a name.
        /// </summary>
        /// <param name="fromCellReference">The start cell reference.</param>
        /// <param name="toCellReference">The end cell reference.</param>
        /// <returns>The name.</returns>
        public static string ConvertCellReferenceRangeToName(CellReference fromCellReference, CellReference toCellReference)
        {
            Guard.ThrowExceptionIfNull(fromCellReference, "fromCellReference");
            Guard.ThrowExceptionIfNull(toCellReference, "toCellReference");
            Guard.ThrowExceptionIfInvalidRowIndex(fromCellReference.ActualRowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(fromCellReference.ActualColumnIndex);
            Guard.ThrowExceptionIfInvalidRowIndex(toCellReference.ActualRowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(toCellReference.ActualColumnIndex);

            StringBuilder builder = new StringBuilder();

            builder.Append(ConvertCellReferenceToName(fromCellReference));
            builder.Append(':');
            builder.Append(ConvertCellReferenceToName(toCellReference));

            return builder.ToString();
        }

        /// <summary>
        /// Converts the cell indexes to a name.
        /// </summary>
        /// <param name="fromRowIndex">The start row index.</param>
        /// <param name="fromColumnIndex">The start column index.</param>
        /// <param name="toRowIndex">The end row index.</param>
        /// <param name="toColumnIndex">The end column index.</param>
        /// <returns>The name.</returns>
        public static string ConvertCellIndexesToName(int fromRowIndex, int fromColumnIndex, int toRowIndex, int toColumnIndex)
        {
            Guard.ThrowExceptionIfInvalidRowIndex(fromRowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(fromColumnIndex);
            Guard.ThrowExceptionIfInvalidRowIndex(toRowIndex);
            Guard.ThrowExceptionIfInvalidColumnIndex(toColumnIndex);

            return string.Format("{0}:{1}", ConvertCellIndexToName(fromRowIndex, fromColumnIndex), ConvertCellIndexToName(toRowIndex, toColumnIndex));
        }

        /// <summary>
        /// Converts the cell index to name.
        /// </summary>
        /// <param name="cellIndex">The cell index.</param>
        /// <returns>The name.</returns>
        public static string ConvertCellIndexToName(CellIndex cellIndex)
        {
            return ConvertCellIndexToName(cellIndex.RowIndex, cellIndex.ColumnIndex);
        }

        /// <summary>
        /// Converts the cell name to a cell index.
        /// </summary>
        /// <param name="cellName">Name of the cell.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        public static void ConvertCellNameToIndex(string cellName, out int rowIndex, out int columnIndex)
        {
            Guard.ThrowExceptionIfNullOrEmpty(cellName, "cellName");

            int firstDigitIndex = GetFirstDigitIndex(cellName);

            rowIndex = ConvertRowNameToIndex(cellName.Substring(firstDigitIndex));
            columnIndex = ConvertColumnNameToIndex(cellName.Substring(0, firstDigitIndex));
        }

        /// <summary>
        /// Tries to convert the cell name to index.
        /// </summary>
        /// <param name="cellName">Name of the cell.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The value indicating whether the conversion was successful.</returns>
        public static bool TryConvertCellNameToIndex(string cellName, out int rowIndex, out int columnIndex)
        {
            rowIndex = 0;
            columnIndex = 0;

            if (string.IsNullOrEmpty(cellName))
            {
                return false;
            }

            int firstDigitIndex = GetFirstDigitIndex(cellName);

            if (firstDigitIndex == 0)
            {
                return false;
            }

            return TryConvertRowNameToIndex(cellName.Substring(firstDigitIndex), out rowIndex) &&
                TryConvertColumnNameToIndex(cellName.Substring(0, firstDigitIndex), out columnIndex);
        }

        private static int GetFirstDigitIndex(string cellName)
        {
            for (int i = 0; i < cellName.Length; i++)
            {
                if (char.IsDigit(cellName[i]))
                {
                    return i;
                }
            }

            return 0;
        }

        internal static void ConvertCellNameToCellRange(string cellName, out CellRange cellRange)
        {
            Guard.ThrowExceptionIfNullOrEmpty(cellName, "cellName");

            string[] cellNames = cellName.Split(':');
            int row;
            int column;

            if (cellNames.Length == 1)
            {
                ConvertCellNameToIndex(cellName, out row, out column);
                CellIndex index = new CellIndex(row, column);
                cellRange = new CellRange(index, index);
            }
            else
            {
                int row1 = 0;
                int column1 = 0;
                bool isRowAbsolute;
                bool isColumnAbsolute;
                bool isRowAbsolute1;
                bool isColumnAbsolute1;

                ConvertCellNameToIndex(cellNames[0], out isRowAbsolute, out row, out isColumnAbsolute, out column);
                ConvertCellNameToIndex(cellNames[1], out isRowAbsolute1, out row1, out isColumnAbsolute1, out column1);

                cellRange = new CellRange(new CellIndex(row, column), new CellIndex(row1, column1));
            }
        }

        internal static bool TryConvertCellNameToCellRange(string cellName, out CellRange cellRange)
        {
            cellRange = null;
            try
            {
                ConvertCellNameToCellRange(cellName, out cellRange);
                return true;
            }
            catch
            {
                return false;
            }
        }

        /// <summary>
        /// Converts the cell name to a cell index.
        /// </summary>
        /// <param name="cellName">Name of the cell.</param>
        /// <param name="isRowAbsolute">The value indicating whether the row is absolute.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="isColumnAbsolute">The value indicating whether the column is absolute.</param>
        /// <param name="columnIndex">Index of the column.</param>
        public static void ConvertCellNameToIndex(string cellName, out bool isRowAbsolute, out int rowIndex, out bool isColumnAbsolute, out int columnIndex)
        {
            Guard.ThrowExceptionIfNullOrEmpty(cellName, "cellName");
            string sanitizedCellName = cellName.Replace("$", string.Empty);
            ConvertCellNameToIndex(sanitizedCellName, out rowIndex, out columnIndex);

            int dollarCount = cellName.Length - sanitizedCellName.Length;

            isColumnAbsolute = cellName.StartsWith("$");
            if (isColumnAbsolute)
            {
                dollarCount--;
            }

            isRowAbsolute = (dollarCount > 0);
        }

        /// <summary>
        /// Tries to convert the cell name to index.
        /// </summary>
        /// <param name="cellName">Name of the cell.</param>
        /// <param name="isRowAbsolute">The is row absolute.</param>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="isColumnAbsolute">The is column absolute.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns></returns>
        public static bool TryConvertCellNameToIndex(string cellName, out bool isRowAbsolute, out int rowIndex, out bool isColumnAbsolute, out int columnIndex)
        {
            isRowAbsolute = false;
            isColumnAbsolute = false;
            rowIndex = 0;
            columnIndex = 0;

            if (string.IsNullOrEmpty(cellName))
            {
                return false;
            }

            string sanitizedCellName = cellName.Replace("$", string.Empty);

            if (TryConvertCellNameToIndex(sanitizedCellName, out rowIndex, out columnIndex))
            {
                int dollarCount = cellName.Length - sanitizedCellName.Length;

                isColumnAbsolute = cellName.StartsWith("$");
                if (isColumnAbsolute)
                {
                    dollarCount--;
                }

                isRowAbsolute = (dollarCount > 0);

                return true;
            }

            return false;
        }

        internal static string ConvertCellRangesToName(Worksheet worksheet, CellRange[] ranges, bool referencesAreAbsolute)
        {
            CellRange lastRange = ranges.Last();

            InputStringCollection stringCollection = new InputStringCollection();

            string encodedWorksheetName = TextHelper.EncodeWorksheetName(worksheet.Name);

            foreach (CellRange range in ranges)
            {
                RowColumnReference fromRowIndex;
                RowColumnReference fromColumnIndex;
                RowColumnReference toRowIndex;
                RowColumnReference toColumnIndex;

                if (referencesAreAbsolute)
                {
                    fromRowIndex = RowColumnReference.CreateAbsoluteRowColumnReference(range.FromIndex.RowIndex);
                    fromColumnIndex = RowColumnReference.CreateAbsoluteRowColumnReference(range.FromIndex.ColumnIndex);
                    toRowIndex = RowColumnReference.CreateAbsoluteRowColumnReference(range.ToIndex.RowIndex);
                    toColumnIndex = RowColumnReference.CreateAbsoluteRowColumnReference(range.ToIndex.ColumnIndex);
                }
                else
                {
                    fromRowIndex = RowColumnReference.CreateRelativeRowColumnReference(range.FromIndex.RowIndex, 0);
                    fromColumnIndex = RowColumnReference.CreateRelativeRowColumnReference(range.FromIndex.ColumnIndex, 0);
                    toRowIndex = RowColumnReference.CreateRelativeRowColumnReference(range.ToIndex.RowIndex, 0);
                    toColumnIndex = RowColumnReference.CreateRelativeRowColumnReference(range.ToIndex.ColumnIndex, 0);
                }

                CellAreaReference cellAreaReference = new CellAreaReference(new CellReference(fromRowIndex, fromColumnIndex), new CellReference(toRowIndex, toColumnIndex));
                CellReferenceRangeExpression cellReferenceRangeExpression = new CellReferenceRangeExpression(worksheet.Workbook, encodedWorksheetName, true, cellAreaReference);
                CellReferenceInputString inputString = new CellReferenceInputString(cellReferenceRangeExpression);

                stringCollection.Add(inputString);

                if (range != lastRange)
                {
                    stringCollection.Add(new ListSeparatorInputString());
                }
            }

            string formula = stringCollection.BuildStringInCulture(SpreadsheetCultureHelper.InvariantSpreadsheetCultureInfo);

            return formula;
        }

        /// <summary>
        /// Determines whether the name of the cell is valid.
        /// </summary>
        /// <param name="cellName">Name of the cell.</param>
        /// <returns>The value indicating whether the name of the cell is valid.</returns>
        public static bool IsValidA1CellName(string cellName)
        {
            int rowIndex;
            int columnIndex;
            bool isRowAbsolute;
            bool isColumnAbsolute;

            if (string.IsNullOrEmpty(cellName))
            {
                return false;
            }

            return TryConvertCellNameToIndex(cellName, out isRowAbsolute, out rowIndex, out isColumnAbsolute, out columnIndex);
        }

        private static bool IsValidColumnName(string columnName)
        {
            if (string.IsNullOrEmpty(columnName))
            {
                return false;
            }

            return columnName.All(x => char.IsLetter(x));
        }

        private static bool IsValidRowName(string rowName)
        {
            return rowName.All(x => char.IsDigit(x));
        }

        #endregion
    }
}