﻿using System;
using System.ComponentModel;

namespace Telerik.Windows.Documents.Spreadsheet.Utilities.Performance
{
    /// <summary>
    /// Defines functionality for testing a code snippet's performance.
    /// </summary>
    [Browsable(false)]
    [EditorBrowsable(EditorBrowsableState.Never)]
    public class SnippetPerformanceCounter
	{
		#region Fields

		private ILogger logger;
		private string snippetName;
		private double totalMilliseconds;
		private bool isStartCalled;
		private DateTime currentStartTime;
		private bool hasBeenStarted;
		private int hitCounter;

		#endregion


		#region Properties

        /// <summary>
        /// Gets the hit counter.
        /// </summary>
        /// <value>The hit counter.</value>
        public int HitCounter
        {
            get
            {
                return this.hitCounter;
            }
        }

        /// <summary>
        /// Gets the total milliseconds.
        /// </summary>
        /// <value>The total milliseconds.</value>
        public double TotalMilliseconds
        {
			get
			{
				return totalMilliseconds;
			}
		}

		#endregion


		#region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="SnippetPerformanceCounter" /> class.
        /// </summary>
        /// <param name="snippetName">Name of the snippet.</param>
        /// <param name="useLogging">The value indicating whether logging should be used.</param>
        public SnippetPerformanceCounter(string snippetName, bool useLogging = true)
        {
			if (useLogging)
			{
				this.logger = new DebugLogger();
			}
			this.snippetName = snippetName;
			this.Reset();
		}

		#endregion


		#region Methods

        /// <summary>
        /// Starts the counter.
        /// </summary>
        public void Start()
        {
			if (!this.hasBeenStarted)
			{
				if (this.logger != null)
				{
					this.logger.Log(string.Format("[{0}]: FIRST TIME STARTED", this.snippetName));
				}
				this.hasBeenStarted = true;
			}

			if (!this.isStartCalled)
			{
				this.currentStartTime = DateTime.Now;
				this.isStartCalled = true;
			}

			hitCounter++;
		}

        /// <summary>
        /// Stops the counter.
        /// </summary>
        public void Stop()
        {
			if (this.isStartCalled)
			{
				totalMilliseconds += (DateTime.Now - this.currentStartTime).TotalMilliseconds;
				this.isStartCalled = false;
			}
		}

        /// <summary>
        /// Ends the counter.
        /// </summary>
        public void End()
        {
			this.Stop();

			if (this.hasBeenStarted)
			{
				if (this.logger != null)
				{
					logger.Log(string.Format("[{0}]: ENDED WITH TOTAL:{1}; TIMES HIT: {2}", this.snippetName, totalMilliseconds, hitCounter));
				}
			}
		}

        /// <summary>
        /// Resets the counter.
        /// </summary>
        public void Reset()
        {
			this.totalMilliseconds = 0;
			this.hitCounter = 0;
			this.isStartCalled = false;
			this.hasBeenStarted = false;
		}

		#endregion
	}
}
