﻿using System;
using System.Linq;
using System.Windows;
using System.Windows.Media;
using Telerik.Windows.Documents.Spreadsheet.Layout;
using Telerik.Windows.Documents.Spreadsheet.Maths;
using Telerik.Windows.Documents.Spreadsheet.Model.Shapes;

namespace Telerik.Windows.Documents.Spreadsheet.Utilities
{
    internal static class ShapesResizeHelper
    {
        /// <summary>
        /// Determines the new top left of an adorner or shape when the adorner or shape are resized. When the shape or
        /// its adorner are resized while there is rotation angle applied, the top left corner must be changed.
        /// This is done in order to maintain the visual top left to remain in place.
        /// </summary>
        /// <param name="shapeCurrentRectangle">The shape current rectangle.</param>
        /// <param name="rotationAngle">The rotation angle.</param>
        /// <param name="newWidth">The new width. Negative value if the resizing causes the shape to flip horizontally. Otherwise positive.</param>
        /// <param name="newHeight">The new height. Negative value if the resizing causes the shape to flip vertically. Otherwise positive.</param>
        /// <param name="isLeftResizing">Boolean value indicating whether the left side of the shape is being resized.</param>
        /// <param name="isTopResizing">Boolean value indicating whether the top side of the shape is being resized.</param>
        /// <returns>The new top left coordinate.</returns>
        public static Point DetermineNewTopLeftWhenResized(Rect shapeCurrentRectangle, double rotationAngle, double newWidth, double newHeight, bool isLeftResizing = false, bool isTopResizing = false)
        {
            double newRotatedCenterX = isLeftResizing ? (shapeCurrentRectangle.Width - newWidth / 2) : (newWidth / 2);
            double newRotatedCenterY = isTopResizing ? (shapeCurrentRectangle.Height - newHeight / 2) : (newHeight / 2);
            Point newRotatedCenter = new Point(newRotatedCenterX, newRotatedCenterY);

            Matrix m = new Matrix();
            m = m.RotateMatrix(rotationAngle);

            Point currentTopLeft = new Point(shapeCurrentRectangle.X, shapeCurrentRectangle.Y);
            Point center = currentTopLeft.Add(new Point(shapeCurrentRectangle.Width / 2, shapeCurrentRectangle.Height / 2));
            Point rotatedTopleft = currentTopLeft.Rotate(center, rotationAngle);
            m = m.TranslateMatrix(rotatedTopleft.X, rotatedTopleft.Y);

            Point newCenter = m.Transform(newRotatedCenter);

            Point newTopLeftCornerCoordinates = new Point(newCenter.X - Math.Abs(newWidth / 2), newCenter.Y - Math.Abs(newHeight / 2));

            return newTopLeftCornerCoordinates;
        }

        /// <summary>
        /// Gets the lock aspect ratio scale factor which should be applied to the proposed width and height.
        /// </summary>
        /// <param name="widthHeightAspectRatio">The width:height aspect ratio.</param>
        /// <param name="proposedWidth">The proposed new width.</param>
        /// <param name="proposedHeight">The proposed new height.</param>
        /// <returns>The scale factor.</returns>
        public static Size GetLockAspectRatioScaleFactor(double widthHeightAspectRatio, double proposedWidth, double proposedHeight)
        {
            double proposedWidthHeightRatio = Math.Abs(proposedWidth / proposedHeight);

            return proposedWidthHeightRatio > widthHeightAspectRatio ?
                new Size(1, proposedWidthHeightRatio / widthHeightAspectRatio) :
                new Size(widthHeightAspectRatio / proposedWidthHeightRatio, 1);
        }

        public static Point AdjustBoundingRectangleWhenOutOfSpreadsheet(RadWorksheetLayout layout, Rect boundingRectangle, Point topLeftCornerCoordinates)
        {
            Guard.ThrowExceptionIfNull(layout, "layout");

            double newTopLeftLeft = topLeftCornerCoordinates.X;
            double newTopLeftTop = topLeftCornerCoordinates.Y;

            if (boundingRectangle.Left < 0)
            {
                newTopLeftLeft = topLeftCornerCoordinates.X + Math.Abs(boundingRectangle.X);
            }

            if (boundingRectangle.Right >= layout.Width)
            {
                newTopLeftLeft = topLeftCornerCoordinates.X - (boundingRectangle.Right - layout.Width) - 0.1;
            }

            if (boundingRectangle.Top < 0)
            {
                newTopLeftTop = topLeftCornerCoordinates.Y + Math.Abs(boundingRectangle.Y);
            }

            if (boundingRectangle.Bottom >= layout.Height)
            {
                newTopLeftTop = topLeftCornerCoordinates.Y - (boundingRectangle.Bottom - layout.Height) - 0.1;
            }

            return new Point(newTopLeftLeft, newTopLeftTop);
        }

        public static Point AdjustPointWhenOutOfSpreadsheet(RadWorksheetLayout layout, Point point)
        {
            double newPointLeft = point.X;
            double newPointTop = point.Y;

            newPointLeft = Math.Max(0, newPointLeft);
            newPointTop = Math.Max(0, newPointTop);
            newPointLeft = Math.Min(layout.Width - 0.1, newPointLeft);
            newPointTop = Math.Min(layout.Height - 0.1, newPointTop);

            return new Point(newPointLeft, newPointTop);
        }

        public static Point AdjustTopLeftForRotation(Point point, double width, double height)
        {
            Matrix m = Matrix.Identity;
            m = m.RotateMatrixAt(90, point.X + (width / 2), point.Y + (height / 2));

            Point rotatedPoint = m.Transform(point);
            rotatedPoint.X -= height;

            return rotatedPoint;
        }

        public static Point ReverseAdjustmentForRotation(Point point, double width, double height)
        {
            Matrix m = Matrix.Identity;
            m = m.RotateMatrixAt(-90, point.X + (height / 2), point.Y + (width / 2));

            Point rotatedPoint = m.Transform(point);
            rotatedPoint.Y -= height;

            return rotatedPoint;
        }

        public static Point GetShapeTopLeftConsideringAdjustmentForRotation(FloatingShapeBase shape, RadWorksheetLayout layout)
        {
            Guard.ThrowExceptionIfNull(shape, "shape");
            Point topLeft = layout.GetShapeTopLeft(shape);

            if (shape.DoesRotationAngleRequireCellIndexChange())
            {
                return ReverseAdjustmentForRotation(topLeft, shape.Width, shape.Height);
            }
            else
            {
                return topLeft;
            }
        }

        public static Point GetDocumentCoordinateFromRotatedPoint(Point point, double rotationAngle, Point topLeftCoordinate, double width, double height)
        {
            Matrix m = GetRotatedToDocumentTransoformMatrix(rotationAngle, topLeftCoordinate, width, height);
            return m.Transform(point);
        }

        public static Point GetRotatedCoordinateFromDocumentPoint(Point point, double rotationAngle, Point topLeftCoordinate, double width, double height)
        {
            Matrix m = GetRotatedToDocumentTransoformMatrix(rotationAngle, topLeftCoordinate, width, height);
            m = m.InverseMatrix();

            return m.Transform(point);
        }

        private static Matrix GetRotatedToDocumentTransoformMatrix(double rotationAngle, Point topLeftCoordinate, double width, double height)
        {
            Matrix m = new Matrix();
            m = m.RotateMatrix(rotationAngle);
            Point center = topLeftCoordinate.Add(new Point(width / 2, height / 2));
            Point rotatedTopleft = topLeftCoordinate.Rotate(center, rotationAngle);
            m = m.TranslateMatrix(rotatedTopleft.X, rotatedTopleft.Y);

            return m;
        }
    }
}
