﻿using System;
using System.Globalization;
using System.Text;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings.Builders;
using Telerik.Windows.Documents.Spreadsheet.Formatting.FormatStrings;

namespace Telerik.Windows.Documents.Spreadsheet.Utilities
{
    internal class SpreadsheetCultureHelper
    {
        #region Constants

        internal const string NullErrorValue = "#NULL!";
        internal const string DivisionByZeroErrorValue = "#DIV/0!";
        internal const string ValueErrorValue = "#VALUE!";
        internal const string ReferenceErrorValue = "#REF!";
        internal const string NameErrorValue = "#NAME?";
        internal const string NumberErrorValue = "#NUM!";
        internal const string NotAvailableErrorValue = "#N/A";

        // TODO: Why here?
        internal const string ArrayLeftBracket = "{";
        internal const string ArrayRightBracket = "}";
        internal const string Ampersand = "&";
        internal const string Colon = ":";
        internal const string Exponent = "E";

        internal const string UnaryPlus = "+";
        internal const string UnaryMinus = "-";
        internal const string Percent = "%";

        internal const string Plus = "+";
        internal const string Minus = "-";
        internal const string Multiplication = "*";
        internal const string Division = "/";
        internal const string Power = "^";
        internal const string Equal = "=";
        internal const string GreaterThan = ">";
        internal const string GreaterThanOrEqualTo = ">=";
        internal const string LessThan = "<";
        internal const string LessThanOrEqualTo = "<=";
        internal const string NotEqual = "<>";

        internal const string LeftParenthesis = "(";
        internal const string RightParenthesis = ")";
        internal const string LeftSquareBracket = "[";
        internal const string RightSquareBracket = "]";

        internal const string TrueLiteral = "TRUE";
        internal const string FalseLiteral = "FALSE";

        internal const string DoubleQuote = "\"";
        internal const string SingleQuote = "'";
        internal const string Comma = ",";
        internal const string Semicolon = ";";

        // TODO: function start should be =+-
        internal const string FunctionStart = "=";

        // TODO: remove these
        internal const string Hash = "#";
        internal const string ExclamationMark = "!";
        internal const string QuestionMark = "?";
        internal const string Period = ".";
        internal const string Ellipsis = "...";
        internal const string Space = " ";
        internal const string Underscore = "_";
        internal const string Prime = "`";
        internal const string VerticalBar = "|";
        internal const string Tilde = "~";
        internal const string DollarSign = "$";
        internal const string Asterisk = "*";

        private const string Backslash = "\\";

        internal static readonly SpreadsheetCultureHelper InvariantSpreadsheetCultureInfo = new SpreadsheetCultureHelper(CultureInfo.InvariantCulture);

        #endregion


        #region Fields

        private readonly CultureInfo cultureInfo;

        #endregion


        #region Properties

        public CultureInfo CultureInfo
        {
            get
            {
                return this.cultureInfo;
            }
        }

        public string NumberDecimalSeparator
        {
            get
            {
                return this.cultureInfo.NumberFormat.NumberDecimalSeparator;
            }
        }

        public string NumberGroupSeparator
        {
            get
            {
                return this.cultureInfo.NumberFormat.NumberGroupSeparator;
            }
        }

        public string CurrencyDecimalSeparator
        {
            get
            {
                return this.cultureInfo.NumberFormat.CurrencyDecimalSeparator;
            }
        }

        public string CurrencyGroupSeparator
        {
            get
            {
                return this.cultureInfo.NumberFormat.CurrencyGroupSeparator;
            }
        }

        public string ListSeparator
        {
            get
            {
                string listSeparator = this.cultureInfo.TextInfo.ListSeparator;

                if (listSeparator == this.NumberDecimalSeparator)
                {
                    if (listSeparator == Comma)
                    {
                        return Semicolon;
                    }
                }

                return listSeparator;
            }
        }

        public string ArrayListSeparator
        {
            get
            {
                if (this.ListSeparator == this.ArrayRowSeparator)
                {
                    return Backslash;
                }

                return ListSeparator;
            }
        }

        public string ArrayRowSeparator
        {
            get
            {
                return Semicolon;
            }
        }

        public string TextQualifier
        {
            get
            {
                return DoubleQuote;
            }
        }

        public string TimeSeparator
        {
            get
            {
#if SILVERLIGHT
                string shortTimePattern = this.cultureInfo.DateTimeFormat.ShortTimePattern;

                for (int i = 0; i < shortTimePattern.Length; i++)
                {
                    char currentChar = char.ToLower(shortTimePattern[i]);
                    if (currentChar != 'h' && currentChar != 'm')
                    {
                        return currentChar.ToString();
                    }
                }

                return ":";
#else
                return cultureInfo.DateTimeFormat.TimeSeparator;
#endif
            }
        }

        public string DateSeparator
        {
            get
            {
                string dateSeparator = string.Empty;
#if SILVERLIGHT
                string shortDatePattern = this.cultureInfo.DateTimeFormat.ShortDatePattern;

                for (int i = 0; i < shortDatePattern.Length; i++)
                {
                    char currentChar = char.ToLower(shortDatePattern[i]);
                    if (currentChar != 'd' && currentChar != 'm' && currentChar != 'y')
                    {
                        dateSeparator = currentChar.ToString();
                        break;
                    }
                }

#else
                dateSeparator = cultureInfo.DateTimeFormat.DateSeparator;
#endif
                dateSeparator = dateSeparator == "/" ? "-" : dateSeparator;
                return dateSeparator;
            }
        }

        #endregion


        #region Constructors

        public SpreadsheetCultureHelper()
            : this(FormatHelper.GetCurrentCulture())
        {
        }

        public SpreadsheetCultureHelper(CultureInfo cultureInfo)
        {
            this.cultureInfo = cultureInfo;
        }

        #endregion


        #region Methods

        public static string ClearFormulaValue(string formula)
        {
            StringBuilder sb = new StringBuilder(formula.Trim());
            if (sb.Length > 0 && IsCharEqualTo(sb[0], FunctionStart))
            {
                sb.Remove(0, 1);
            }

            return sb.ToString();
        }

        public static string PrepareFormulaValue(string formula)
        {
            formula = formula.Trim();
            if (!formula.StartsWith(FunctionStart))
            {
                formula = FunctionStart + formula;
            }

            return formula;
        }

        public static bool IsExponent(char ch)
        {
            return char.ToUpper(ch).ToString() == Exponent;
        }

        public static bool IsLiteral(string input, params string[] literals)
        {
            string inputToUpper = input.ToUpper();
            for (int i = 0; i < literals.Length; i++)
            {
                if (inputToUpper == literals[i])
                {
                    return true;
                }
            }

            return false;
        }

        public static bool IsCharEqualTo(char ch, params string[] values)
        {
            string character = ch.ToString();
            for (int i = 0; i < values.Length; i++)
            {
                if (character == values[i])
                {
                    return true;
                }
            }

            return false;
        }

        public bool TryParseGeneral(string value, out double result)
        {
            value = value.Trim(' ');
            NumberStyles numberStyles = NumberStyles.AllowDecimalPoint | NumberStyles.AllowLeadingSign;
            return double.TryParse(value, numberStyles, this.cultureInfo, out result);
        }

        public bool TryParseNumber(string value, out double result)
        {
            result = 0;

            bool containsNumberGroupSeparator = value.Contains(this.NumberGroupSeparator);

            bool canParse = containsNumberGroupSeparator &&
                double.TryParse(value, NumberStyles.Number, this.cultureInfo, out result) &&
                this.ApplyAdditionalNumberConstraints(value);

            return canParse;
        }

        public bool TryParseDouble(string value, out double result)
        {
            return this.TryParseGeneral(value, out result) ||
                this.TryParseNumber(value, out result);
        }

        public bool TryParseCurrency(string value, out double result)
        {
            result = 0;

            NumberStyles styles = NumberStyles.Number | NumberStyles.AllowCurrencySymbol;
            bool containsCurrencySymbol = value.Contains(this.cultureInfo.NumberFormat.CurrencySymbol);

            bool canParse = containsCurrencySymbol &&
                double.TryParse(value, styles, this.cultureInfo, out result) &&
                this.CheckGroupSeparatorsPositionedCorrectly(value, this.CurrencyGroupSeparator, this.CurrencyDecimalSeparator, this.CultureInfo.NumberFormat.CurrencyGroupSizes);

            return canParse;
        }

        public bool TryParseDateTime(string value, out DateTime result)
        {
            return DateTime.TryParse(value, this.cultureInfo, DateTimeStyles.NoCurrentDateDefault, out result) &&
                !value.Contains(Environment.NewLine)
                && (value.Contains(this.DateSeparator) ||
                    value.Contains("/") ||
                    value.Contains("-") ||
                    value.Contains(" "));
        }

        public bool TryParseTimeSpan(string value, out TimeSpan timeSpan)
        {
            return TimeSpan.TryParse(value, out timeSpan) && !value.Contains(Environment.NewLine);
        }

        public bool TryParsePercent(string value, out double result)
        {
            result = 0;
            int percentIndex = value.IndexOf(this.cultureInfo.NumberFormat.PercentSymbol);
            if (percentIndex != value.Length - 1)
            {
                return false;
            }

            string percentStrippedString = value.Replace(this.cultureInfo.NumberFormat.PercentSymbol, "");
            bool canParse = double.TryParse(percentStrippedString, NumberStyles.Any, this.cultureInfo, out result);

            // NOTE: When inputing 12% it should be presented as 12% and not 1200%. That's why we devide by 100.
            result /= 100;

            return canParse;
        }

        public bool TryParseScientific(string value, out double result)
        {
            NumberStyles numberStyles = NumberStyles.AllowLeadingWhite |
                NumberStyles.AllowTrailingWhite |
                NumberStyles.AllowLeadingSign |
                NumberStyles.AllowDecimalPoint |
                NumberStyles.AllowThousands |
                NumberStyles.AllowExponent;

            return double.TryParse(value, numberStyles, this.cultureInfo, out result) &&
                this.CheckGroupSeparatorsPositionedCorrectly(value, this.NumberGroupSeparator, this.NumberDecimalSeparator, this.CultureInfo.NumberFormat.NumberGroupSizes) &&
                !value.Contains(Environment.NewLine);
        }

        public bool IsDecimalSeparator(char ch)
        {
            return ch.ToString() == this.NumberDecimalSeparator;
        }

        public bool IsNumberGroupSeparator(char ch)
        {
            return ch.ToString() == this.NumberGroupSeparator;
        }

        public string ToString(double value)
        {
            return value.ToString(this.CultureInfo);
        }

        public string GetDefaultNumberFormatString()
        {
            int decimalDigits = this.CultureInfo.NumberFormat.NumberDecimalDigits;
            List<string> numberStrings = NumberFormatStringBuilder.BuildFormatStrings(decimalDigits, false);
            return numberStrings.First();
        }

        public string GetDefaultCurrencyFormatString()
        {
            CurrencyInfo currencySymbol = CurrencyInfo.Symbol;
            int decimalDigits = this.CultureInfo.NumberFormat.NumberDecimalDigits;
            List<string> currencyStrings = CurrencyFormatStringBuilder.BuildFormatStrings(currencySymbol, decimalDigits);
            return currencyStrings.First();
        }

        public string GetDefaultAccountingFormatString()
        {
            CurrencyInfo currencySymbol = CurrencyInfo.Symbol;
            int decimalDigits = this.CultureInfo.NumberFormat.NumberDecimalDigits;
            string accountingString = AccountingFormatStringBuilder.BuildFormatString(currencySymbol, decimalDigits);
            return accountingString;
        }

        public string GetDefaultShortDateFormatString()
        {
            return FormatHelper.DefaultSpreadsheetCulture.CultureInfo.DateTimeFormat.ShortDatePattern;
        }

        public string GetDefaultLongDateFormatString()
        {
            return FormatHelper.DefaultSpreadsheetCulture.CultureInfo.DateTimeFormat.LongDatePattern;
        }

        public string GetDefaultTimeFormatString()
        {
            return FormatHelper.DefaultSpreadsheetCulture.CultureInfo.DateTimeFormat.LongTimePattern;
        }

        // These constraints are apllied by Excel. .NET parses those values if not constrained.
        internal bool ApplyAdditionalNumberConstraints(string value)
        {
            if (value.IndexOf(this.NumberGroupSeparator) != -1)
            {
                return CheckGroupSeparatorsPositionedCorrectly(value, this.NumberGroupSeparator, this.NumberDecimalSeparator, this.CultureInfo.NumberFormat.NumberGroupSizes);
            }

            return false;
        }

        // Checking if the group separator is applied correctly.
        // This is to prevent strings like "12.2.2003" to be interpreted as numbers
        internal bool CheckGroupSeparatorsPositionedCorrectly(string value, string groupSeparator, string decimalSeparator, int[] groupSizes)
        {
            string plainString = this.GetValueWithoutFractionalPart(value, decimalSeparator);
            plainString = this.RemoveLeadingAndTrailingNonDigitSymbols(plainString, groupSeparator, decimalSeparator);

            string[] groups = plainString.Split(new string[] { groupSeparator }, StringSplitOptions.None);

            if (groups.Length == 1)
            {
                return true;
            }

            Func<int, int, bool> gteComparer = new Func<int, int, bool>((i, j) => i >= j);
            Func<int, int, bool> eComparer = new Func<int, int, bool>((i, j) => i == j);

            if (!this.CompareValues(groupSizes, groups[0].Length, gteComparer))
            {
                return false;
            }

            for (int i = 1; i < groups.Length; i++)
            {
                if (!this.CompareValues(groupSizes, groups[i].Length, eComparer))
                {
                    return false;
                }
            }

            return true;
        }

        private bool CompareValues(int[] availableValues, int requiredValue, Func<int, int, bool> comparer)
        {
            bool returnValue = false;
            foreach (int value in availableValues)
            {
                if (comparer(value, requiredValue))
                {
                    returnValue = true;
                    break;
                }
            }

            if (!returnValue)
            {
                return false;
            }

            return true;
        }

        private string GetValueWithoutFractionalPart(string value, string decimalSeparator)
        {
            string result = value;
            int decimalSeparatorIndex = value.IndexOf(decimalSeparator);
            if (decimalSeparatorIndex != -1)
            {
                result = value.Substring(0, decimalSeparatorIndex);
            }

            return result;
        }

        private string RemoveLeadingAndTrailingNonDigitSymbols(string value, string groupSeparator, string decimalSeparator)
        {
            value = FormatHelper.StripCurrencyEscapingsAndSlashes(value);

            string result = value;

            int firstDigitIndex = 0;
            for (int i = 0; i < value.Length; i++)
            {
                if (IsValidSymbol(value[i], groupSeparator, decimalSeparator))
                {
                    firstDigitIndex = i;
                    break;
                }
            }

            int lastDigitIndex = value.Length - 1;
            for (int i = lastDigitIndex; i >= 0; i--)
            {
                if (IsValidSymbol(value[i], groupSeparator, decimalSeparator))
                {
                    lastDigitIndex = i;
                    break;
                }
            }

            result = value.Substring(firstDigitIndex, lastDigitIndex - firstDigitIndex + 1);

            return result;
        }

        private bool IsValidSymbol(char c, string groupSeparator, string decimalSeparator)
        {
            return Char.IsDigit(c) || IsCharEqualTo(c, groupSeparator, decimalSeparator);
        }

        #endregion
    }
}
