﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Expressions.Functions;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Shapes;
using Telerik.Windows.Documents.Spreadsheet.Maths;
#if !SILVERLIGHT
using System.Drawing.Imaging;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.Utilities
{
    internal static class SpreadsheetHelper
    {
        private const string EmailMatch = @"([a-zA-Z0-9._%-]+)@([a-zA-Z0-9.-]+\.[a-zA-Z]{2,6})";
        private const string EmailGroupName = "email";
        private const string SubjectGroupName = "subject";
        private static readonly string mailtoMatch =
            string.Format("mailto:(?<{0}>{1})(\\?[Ss]ubject=(?<{2}>.+))?", EmailGroupName, EmailMatch, SubjectGroupName);
        private static readonly Regex mailtoRegex = new Regex(mailtoMatch);
        internal static readonly Regex EmailRegex = new Regex(EmailMatch);

        private static BitmapSource noImagePlaceholder;

        private static BitmapSource NoImagePlaceholder
        {
            get
            {
                if (noImagePlaceholder == null)
                {
                    using (Stream noImageStream = System.Windows.Application.GetResourceStream(new Uri(@"/Telerik.Windows.Documents.Spreadsheet;component/Resources/imageUnavailablePlaceholder.png", UriKind.Relative)).Stream)
                    {
                        noImagePlaceholder = GetImageSource(noImageStream);
                    }
                }

                return noImagePlaceholder;
            }
        }

        public static string GetSheetNamePrefixBySheetType(SheetType type)
        {
            switch (type)
            {
                case SheetType.Worksheet:
                    return "Sheet";
                default:
                    throw new NotSupportedException();
            }
        }

        public static SheetType GetSheetType<T>()
            where T : Sheet
        {
            Type typeOfT = typeof(T);
            if (typeOfT == typeof(Worksheet))
            {
                return SheetType.Worksheet;
            }

            throw new NotSupportedException();
        }

        public static double EnsureValidScaleFactor(double value)
        {
            value = Math.Max(SpreadsheetDefaultValues.MinScaleFactor, value);
            value = Math.Min(SpreadsheetDefaultValues.MaxScaleFactor, value);

            return value;
        }

        public static Size EnsureValidScaleFactor(Size value)
        {
            value.Width = EnsureValidScaleFactor(value.Width);
            value.Height = EnsureValidScaleFactor(value.Height);

            return value;
        }

        private static double Min(params double[] arguments)
        {
            double min = arguments[0];
            for (int i = 1; i < arguments.Length; i++)
            {
                if (min > arguments[i])
                {
                    min = arguments[i];
                }
            }

            return min;
        }

        private static double Max(params double[] arguments)
        {
            double max = arguments[0];
            for (int i = 1; i < arguments.Length; i++)
            {
                if (max < arguments[i])
                {
                    max = arguments[i];
                }
            }

            return max;
        }

        public static Rect CalculateShapeBoundingRect(Point position, FloatingShapeBase shape)
        {
            double x = position.X;
            double y = position.Y;

            double width = shape.DoesRotationAngleRequireCellIndexChange() ? shape.Height : shape.Width;
            double height = shape.DoesRotationAngleRequireCellIndexChange() ? shape.Width : shape.Height;
            double rotationAngle = shape.DoesRotationAngleRequireCellIndexChange() ? (shape.RotationAngle - 90) : shape.RotationAngle;

            Matrix m = Matrix.Identity;
            m = m.RotateMatrixAt(rotationAngle, x + width / 2, y + height / 2);

            Point topLeft = m.Transform(new Point(x, y));
            Point topRight = m.Transform(new Point(x + width, y));
            Point bottomRight = m.Transform(new Point(x + width, y + height));
            Point bottomLeft = m.Transform(new Point(x, y + height));

            return new Rect(
                new Point(Min(topLeft.X, topRight.X, bottomRight.X, bottomLeft.X),
                Min(topLeft.Y, topRight.Y, bottomRight.Y, bottomLeft.Y)),
                new Point(Max(topLeft.X, topRight.X, bottomRight.X, bottomLeft.X),
                Max(topLeft.Y, topRight.Y, bottomRight.Y, bottomLeft.Y)));
        }


        public static bool TryParseMailToAddress(string address, out string emailAddress, out string emailSubject)
        {
            emailAddress = null;
            emailSubject = null;
            Match match = mailtoRegex.Match(address);
            bool canParse = match.Success;

            if (canParse)
            {
                emailAddress = match.Groups[EmailGroupName].ToString();

                if (match.Groups[SubjectGroupName].Success)
                {
                    emailSubject = match.Groups[SubjectGroupName].ToString();
                }
            }

            return canParse;
        }

        public static bool TryParseHyperlinkAddress(string name, out string address, out string subAddress, out string emailAddress,
            out string emailSubject)
        {
            address = string.Empty;
            subAddress = string.Empty;
            emailAddress = string.Empty;
            emailSubject = string.Empty;
            bool canParse = TryParseMailToAddress(name, out emailAddress, out emailSubject);

            if (!canParse)
            {
                string[] adresses = name.Split('!');
                address = adresses[0];
                if (adresses.Length > 1)
                {
                    subAddress = adresses[1];
                }
                canParse = true;
            }

            return canParse;
        }

        public static bool TryGetHyperLinkInfo(ICellValue cellValue, out HyperlinkInfo hyperlinkInfo)
        {
            hyperlinkInfo = null;
            FormulaCellValue formulaCellValue = cellValue as FormulaCellValue;
            if (formulaCellValue != null)
            {
                FunctionExpression functionExpression = formulaCellValue.Value as FunctionExpression;
                if (functionExpression != null && functionExpression.IsValid)
                {
                    HyperlinkFunction hyperlinkFunction = functionExpression.Function as HyperlinkFunction;
                    if (hyperlinkFunction != null)
                    {
                        hyperlinkInfo = hyperlinkFunction.HyperlinkInfo;
                        return true;
                    }
                }
            }

            return false;
        }

        public static bool IsHyperlinkFormula(ICellValue cellValue)
        {
            HyperlinkInfo hyperlinkInfo;
            return SpreadsheetHelper.TryGetHyperLinkInfo(cellValue, out hyperlinkInfo);
        }

        public static Worksheet GetWorksheetByName(Workbook workbook, string worksheetName)
        {
            Worksheet worksheet = null;

            for (int i = 0; i < workbook.Worksheets.Count; i++)
            {
                if (string.Equals(workbook.Worksheets[i].Name, worksheetName, StringComparison.CurrentCultureIgnoreCase))
                {
                    worksheet = workbook.Worksheets[i];
                    break;
                }
            }

            return worksheet;
        }

        public static BitmapSource GetImageSourceOrNoImagePlaceHolder(Stream stream)
        {
            BitmapSource result;

            try
            {
                result = GetImageSource(stream);
            }
            catch (SystemException)
            {
                result = NoImagePlaceholder;
            }

            return result;
        }

        public static BitmapSource GetImageSource(Stream stream)
        {
            BitmapImage bitmapImage = new BitmapImage();

            try
            {
                InitBitmapImage(bitmapImage, stream);
            }
            catch (NotSupportedException)
            {
#if SILVERLIGHT
                throw;
#else
                // WMF and EMF images which are not supported by BitmapImage.
                System.Drawing.Image image = System.Drawing.Image.FromStream(stream);
                MemoryStream ms = new MemoryStream();
                image.Save(ms, ImageFormat.Png);
                ms.Seek(0, SeekOrigin.Begin);
                stream = ms;

                bitmapImage = new BitmapImage();
                InitBitmapImage(bitmapImage, stream);
#endif
            }

            return bitmapImage;
        }

        public static void InitBitmapImage(BitmapImage bitmapImage, Stream stream)
        {
#if SILVERLIGHT
            bitmapImage.SetSource(stream);
#else
            bitmapImage.BeginInit();
            bitmapImage.StreamSource = stream;
            bitmapImage.EndInit();
#endif
        }

#if !SILVERLIGHT
        public static Stream StreamFromBitmapSource(BitmapSource writeBmp)
        {
            Stream bmp = new MemoryStream();

            BitmapEncoder enc = new BmpBitmapEncoder();
            enc.Frames.Add(BitmapFrame.Create(writeBmp));
            enc.Save(bmp);

            return bmp;
        }
#endif

        public static T GetNearestValidValue<T>(T value, T minValue, T maxValue)
            where T : IComparable<T>
        {
            if (value.CompareTo(minValue) < 0)
            {
                return minValue;
            }
            else if (value.CompareTo(maxValue) > 0)
            {
                return maxValue;
            }
            else
            {
                return value;
            }
        }

        public static int PercentFromScaleFactor(Size scaleFactor)
        {
            return PercentFromScaleFactor(scaleFactor.Width);
        }

        public static int PercentFromScaleFactor(double scaleFactor)
        {
            return (int)(scaleFactor * 100);
        }

        public static Size ScaleFactorFromPercent(int? percent)
        {
            double scaleFactor = (percent ?? 0) / 100.0;

            return new Size(scaleFactor, scaleFactor);
        }

        public static double RestrictRotationAngle(double angle)
        {
            double restrictedAngle = angle % 360;

            if (restrictedAngle < 0)
            {
                restrictedAngle = 360 + restrictedAngle;
            }

            return restrictedAngle;
        }

        /// <summary>
        /// Transforms the specified value using current general transform.
        /// </summary>
        /// <param name="generalTransform">The general transform.</param>
        /// <param name="value">The value.</param>
        /// <returns>The transformed value.</returns>
        public static double Transform(GeneralTransform generalTransform, double value)
        {
            return generalTransform.Transform(new Point(value, 0)).X;
        }

        /// <summary>
        /// Transforms the specified value using current general transform.
        /// </summary>
        /// <param name="generalTransform">The general transform.</param>
        /// <param name="value">The value.</param>
        /// <returns>The transformed value.</returns>
        public static Size Transform(GeneralTransform generalTransform, Size value)
        {
            Point point = generalTransform.Transform(new Point(value.Width, value.Height));
            return new Size(point.X, point.Y);
        }

        public static List<CellRange> SplitIntoRangesAccordingToHidden(CellRange cellRange, Worksheet worksheet)
        {
            int[] hiddenRows;
            int[] hiddenColumns;
            return SplitIntoRangesAccordingToHidden(cellRange, worksheet, out hiddenRows, out hiddenColumns);
        }

        public static List<CellRange> SplitIntoRangesAccordingToHidden(CellRange cellRange, Worksheet worksheet, out int[] hiddenRows, out int[] hiddenColumns)
        {
            List<CellRange> columns = SplitAccordingToHiddenColumns(cellRange, worksheet, out hiddenColumns);
            List<CellRange> result = SplitAccordingToHiddenRows(columns, cellRange, worksheet, out hiddenRows);

            return result;
        }

        public static int CountRows(IEnumerable<CellRange> cellRanges)
        {
            HashSet<int> rows = new HashSet<int>();

            foreach (CellRange range in cellRanges)
            {
                for (int i = range.FromIndex.RowIndex; i <= range.ToIndex.RowIndex; i++)
                {
                    rows.Add(i);
                }
            }

            return rows.Count;
        }

        public static int CountColumns(IEnumerable<CellRange> cellRanges)
        {
            HashSet<int> columns = new HashSet<int>();

            foreach (CellRange range in cellRanges)
            {
                for (int i = range.FromIndex.ColumnIndex; i <= range.ToIndex.ColumnIndex; i++)
                {
                    columns.Add(i);
                }
            }

            return columns.Count;
        }

        private static List<CellRange> SplitAccordingToHiddenColumns(CellRange cellRange, Worksheet worksheet, out int[] hiddenColumns)
        {
            List<CellRange> result = new List<CellRange>();
            var hiddenColumnsList = new List<int>();

            RangePropertyValue<bool> entireHiddenValue = worksheet.Columns[cellRange].GetHidden();
            if (!entireHiddenValue.IsIndeterminate && !entireHiddenValue.Value)
            {
                hiddenColumns = new int[0];
                return new List<CellRange>() { cellRange };
            }

            bool isInRange = false;
            int rangeStart = 0;
            int rangeEnd = 0;

            for (int i = cellRange.FromIndex.ColumnIndex; i <= cellRange.ToIndex.ColumnIndex; i++)
            {
                if (worksheet.Columns[i].GetHidden().Value)
                {
                    if (isInRange)
                    {
                        rangeEnd = i - 1;
                        isInRange = false;
                        result.Add(new CellRange(cellRange.FromIndex.RowIndex, rangeStart, cellRange.ToIndex.RowIndex, rangeEnd));
                    }

                    hiddenColumnsList.Add(i);
                }
                else
                {
                    if (!isInRange)
                    {
                        rangeStart = i;
                        isInRange = true;
                    }
                }
            }

            if (isInRange)
            {
                result.Add(new CellRange(cellRange.FromIndex.RowIndex, rangeStart, cellRange.ToIndex.RowIndex, cellRange.ToIndex.ColumnIndex));
            }

            hiddenColumns = hiddenColumnsList.ToArray();

            return result;
        }

        private static List<CellRange> SplitAccordingToHiddenRows(List<CellRange> columnRanges, CellRange cellRange, Worksheet worksheet, out int[] hiddenRows)
        {
            List<CellRange> result = new List<CellRange>();
            HashSet<int> hiddenRowsHash = new HashSet<int>();

            RangePropertyValue<bool> entireHiddenValue = worksheet.Rows[cellRange].GetHidden();
            if (!entireHiddenValue.IsIndeterminate && !entireHiddenValue.Value)
            {
                hiddenRows = new int[0];
                return columnRanges;
            }

            foreach (CellRange columnRange in columnRanges)
            {
                bool isInRange = false;
                int rangeStart = 0;
                int rangeEnd = 0;

                for (int i = cellRange.FromIndex.RowIndex; i <= columnRange.ToIndex.RowIndex; i++)
                {
                    if (worksheet.Rows[i].GetHidden().Value)
                    {
                        if (isInRange)
                        {
                            rangeEnd = i - 1;
                            isInRange = false;
                            result.Add(new CellRange(rangeStart, columnRange.FromIndex.ColumnIndex, rangeEnd, columnRange.ToIndex.ColumnIndex));
                        }

                        hiddenRowsHash.Add(i);
                    }
                    else
                    {
                        if (!isInRange)
                        {
                            rangeStart = i;
                            isInRange = true;
                        }
                    }
                }

                if (isInRange)
                {
                    result.Add(new CellRange(rangeStart, columnRange.FromIndex.ColumnIndex, columnRange.ToIndex.RowIndex, columnRange.ToIndex.ColumnIndex));
                }
            }

            hiddenRows = hiddenRowsHash.ToArray();

            return result;
        }
    }
}