﻿using System.Collections.Generic;
using System.Linq;
using System.Text;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Measurement;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.Utilities
{
    internal static class TextHelper
    {
        #region Methods

        public static string SanitizeNewLines(string input)
        {
            StringBuilder result = new StringBuilder(input.Length);
            bool hasSeenCr = false;

            for (int i = 0, l = input.Length; i < l; ++i)
            {
                if (input[i] == '\r')
                {
                    if (hasSeenCr)
                    {
                        result.Append(LineBreak.NewLine);
                    }

                    hasSeenCr = true;
                }
                else if (input[i] == '\n')
                {
                    result.Append(LineBreak.NewLine);
                    hasSeenCr = false;
                }
                else
                {
                    if (hasSeenCr)
                    {
                        result.Append(LineBreak.NewLine);
                        hasSeenCr = false;
                    }

                    result.Append(input[i]);
                }
            }

            if (hasSeenCr)
            {
                result.Append(LineBreak.NewLine);
            }

            return result.ToString();
        }

        public static string GetClipboardEncodedText(string input)
        {
            StringBuilder result = new StringBuilder();

            foreach (string value in GetValues(input))
            {
                result.Append(DecodeValue(value, SpreadsheetCultureHelper.DoubleQuote));
            }

            return result.ToString();
        }

        public static string GetClipboardDecodedText(string input)
        {
            StringBuilder result = new StringBuilder();

            foreach (string value in GetValues(input))
            {
                result.Append(EncodeValue(value, SpreadsheetCultureHelper.DoubleQuote));
            }

            return result.ToString();
        }

        private static IEnumerable<string> GetValues(string input)
        {
            StringBuilder builder = new StringBuilder();
            bool isValueEscapedCorrectly = false;
            bool requireDoubleQuote = false;

            for (int i = 0; i < input.Length; i++)
            {
                if (IsStartOfValue(input, i, isValueEscapedCorrectly))
                {
                    isValueEscapedCorrectly = input[i] == SpreadsheetCultureHelper.DoubleQuote[0];
                }
                else
                {
                    if (input[i] == SpreadsheetCultureHelper.DoubleQuote[0])
                    {
                        if (isValueEscapedCorrectly)
                        {
                            requireDoubleQuote = !requireDoubleQuote;
                        }
                    }
                    else
                    {
                        if (requireDoubleQuote)
                        {
                            requireDoubleQuote = false;
                            isValueEscapedCorrectly = false;
                        }

                        if (IsEndOfValue(input, i, isValueEscapedCorrectly))
                        {
                            if (builder.Length > 0)
                            {
                                yield return builder.ToString();

                                builder.Clear();
                                isValueEscapedCorrectly = false;
                            }

                            yield return input[i].ToString();
                        }
                    }
                }

                if (!IsEndOfValue(input, i, isValueEscapedCorrectly))
                {
                    builder.Append(input[i]);
                }
            }

            if (builder.Length != 0)
            {
                yield return builder.ToString();
            }
        }

        internal static string DecodeValue(string value, string textQualifier)
        {
            Guard.ThrowExceptionIfNullOrEmpty(value, "value");

            if (value.Length < 2 || value.Contains("\n") || value.Contains("\t"))
            {
                return value;
            }

            string result = StripSurroundingQuotes(value, textQualifier);

            return result.Replace(textQualifier + textQualifier, textQualifier);
        }

        internal static string EncodeValue(string value, string textQualifier)
        {
            Guard.ThrowExceptionIfNullOrEmpty(value, "value");

            if (value.Contains("\n") || value.Contains("\t") || !value.Contains(textQualifier))
            {
                return value;
            }

            return GetEscapedTextValue(value, textQualifier);
        }

        internal static string GetEscapedTextValue(string value, string textQualifier)
        {
            StringBuilder builder = new StringBuilder();
            builder.Append(textQualifier);
            builder.Append(value.Replace(textQualifier, textQualifier + textQualifier));
            builder.Append(textQualifier);

            return builder.ToString();
        }

        internal static string EncodeValue(string value)
        {
            Guard.ThrowExceptionIfNullOrEmpty(value, "value");

            string quote = SpreadsheetCultureHelper.DoubleQuote;

            StringBuilder builder = new StringBuilder();

            builder.Append(quote);
            builder.Append(value.Replace(quote, quote + quote));
            builder.Append(quote);

            return builder.ToString();
        }

        private static bool IsStartOfValue(string value, int position, bool escapeSpecialSymbols)
        {
            if (position == 0)
            {
                return !IsSpecialSymbol(value[position]);
            }

            return !IsSpecialSymbol(value[position]) &&
                    IsSpecialSymbol(value[position - 1]) &&
                   !escapeSpecialSymbols;
        }

        private static bool IsEndOfValue(string value, int position, bool escapeSpecialSymbols)
        {
            if (position == 0)
            {
                return IsSpecialSymbol(value[position]);
            }

            return IsSpecialSymbol(value[position]) && !escapeSpecialSymbols;
        }

        private static bool IsSpecialSymbol(char character)
        {
            return character == '\t' || character == '\n';
        }

        private static string StripSurroundingQuotes(string value, string textQualifier)
        {
            string result = value;
            if (value.Length > 1 && result[0] == textQualifier[0] && result[result.Length - 1] == textQualifier[0])
            {
                result = result.Substring(1, result.Length - 2);
            }

            return result;
        }

        public static string DecodeWorksheetName(string input)
        {
            char quote = SpreadsheetCultureHelper.SingleQuote[0];

            if (!string.IsNullOrEmpty(input) &&
                input[0] == quote &&
                input[input.Length - 1] == quote)
            {
                StringBuilder builder = new StringBuilder();

                bool skipQuote = false;
                for (int i = 1; i < input.Length - 1; i++)
                {
                    if (!skipQuote)
                    {
                        builder.Append(input[i]);

                        if (i < input.Length - 2 && input[i] == quote && input[i + 1] == quote)
                        {
                            skipQuote = true;
                        }
                    }
                    else
                    {
                        skipQuote = false;
                    }
                }

                return builder.ToString();
            }

            return input;
        }

        public static string EncodeWorksheetName(string input, bool encodeRegardlessOfContent = false)
        {
            string quote = SpreadsheetCultureHelper.SingleQuote;

            if (ShouldEscapeSheetName(input) || encodeRegardlessOfContent)
            {
                StringBuilder builder = new StringBuilder();

                builder.Append(quote);
                builder.Append(input.Replace(quote, quote + quote));
                builder.Append(quote);

                return builder.ToString();
            }

            return input;
        }

        internal static string EscapeInvariantDateTimeSeparators(string value)
        {
            StringBuilder builder = new StringBuilder();

            int quotesCount = 0;
            for (int i = 0; i < value.Length; i++)
            {
                char currentChar = value[i];

                if (currentChar == '"')
                {
                    quotesCount++;
                }

                bool isPreviousCharBackSlash = (i >= 1) && (value[i - 1] == '\\');
                bool isPreviousCharClosingQuote = (i >= 1) && (value[i - 1] == '"') && (quotesCount % 2 == 0);

                bool shouldEscape = !(isPreviousCharBackSlash || isPreviousCharClosingQuote);

                if (SpreadsheetCultureHelper.IsCharEqualTo(value[i], "/", SpreadsheetCultureHelper.Colon) && shouldEscape)
                {
                    builder.Append(SpreadsheetCultureHelper.DoubleQuote);
                    builder.Append(currentChar);
                    builder.Append(SpreadsheetCultureHelper.DoubleQuote);
                }
                else
                {
                    builder.Append(currentChar);
                }
            }

            return builder.ToString();
        }

        private static bool ShouldEscapeSheetName(string input)
        {
            Guard.ThrowExceptionIfNullOrEmpty(input, "input");

            if (!char.IsLetter(input[0]) &&
                !SpreadsheetCultureHelper.IsCharEqualTo(input[0], SpreadsheetCultureHelper.Underscore))
            {
                return true;
            }

            for (int i = 0; i < input.Length; i++)
            {
                if (!char.IsLetter(input[i]) && !char.IsDigit(input[i]) &&
                    !SpreadsheetCultureHelper.IsCharEqualTo(input[i], SpreadsheetCultureHelper.Underscore, SpreadsheetCultureHelper.Period))
                {
                    return true;
                }
            }

            return false;
        }

        public static string BuildAbsoluteCellReferenceString(string worksheetName, CellRange selectedRange, SpreadsheetCultureHelper cultureInfo = null)
        {
            return TextHelper.BuildAbsoluteCellReferenceString(worksheetName, Enumerable.Repeat(selectedRange, 1), cultureInfo);
        }

        public static string BuildAbsoluteCellReferenceString(string worksheetName, IEnumerable<CellRange> selectedRanges, SpreadsheetCultureHelper cultureInfo = null)
        {
            Guard.ThrowExceptionIfNullOrEmpty(worksheetName, "worksheetName");
            Guard.ThrowExceptionIfNull(selectedRanges, "selectedRanges");
            Guard.ThrowExceptionIfLessThan(1, selectedRanges.Count(), "selectedRanges");

            cultureInfo = cultureInfo ?? FormatHelper.DefaultSpreadsheetCulture;

            StringBuilder builder = new StringBuilder();

            builder.Append(SpreadsheetCultureHelper.Equal);

            List<CellRange> ranges = selectedRanges.ToList();
            for (int i = 0; i < ranges.Count; i++)
            {
                CellRange selectedRange = ranges[i];

                builder.Append(TextHelper.EncodeWorksheetName(worksheetName));
                builder.Append(SpreadsheetCultureHelper.ExclamationMark);

                builder.Append(NameConverter.ConvertCellIndexToAbsoluteName(selectedRange.FromIndex));

                if (selectedRange.ColumnCount > 1 || selectedRange.RowCount > 1)
                {
                    builder.Append(SpreadsheetCultureHelper.Colon);
                    builder.Append(NameConverter.ConvertCellIndexToAbsoluteName(selectedRange.ToIndex));
                }

                if (i < ranges.Count - 1)
                {
                    builder.Append(cultureInfo.ListSeparator);
                    builder.Append(SpreadsheetCultureHelper.Space);
                }
            }

            return builder.ToString();
        }

        #endregion
    }
}