﻿using System;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Measurement;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;

namespace Telerik.Windows.Documents.Spreadsheet.Utilities
{
    /// <summary>
    /// Provides methods for conversion between units.
    /// </summary>
    public static class UnitHelper
    {
		#region Constants

		private readonly static double DPI = 96.0;
		private readonly static double CmToInch = 2.54;
		private readonly static double PointToInch = 72.0;
		private readonly static double PicaToInch = 6.0;
		private readonly static double TwipToInch = 1440.0;
		private readonly static double EmuToInch = 914400.0;

		#endregion


		#region Convert From Dip to Unit

		/// <summary>
		/// Converts dips to points.
		/// </summary>
		/// <param name="value">Pixels.</param>
		/// <returns>Points.</returns>
		public static double DipToPoint(double value)
		{
			return (value * UnitHelper.PointToInch) / UnitHelper.DPI;
		}

		/// <summary>
		/// Converts dips to points.
		/// </summary>
		/// <param name="value">Pixels.</param>
		/// <returns>Points.</returns>
		public static int DipToPointI(double value)
		{
			return (int)Math.Round(UnitHelper.DipToPoint(value));
		}

		/// <summary>
		/// Converts dips to picas.
		/// </summary>
		/// <param name="value">Pixels.</param>
		/// <returns>Picas.</returns>
		public static double DipToPica(double value)
		{
			return (value * UnitHelper.PicaToInch) / UnitHelper.DPI;
		}

		/// <summary>
		/// Converts dips to centimeters.
		/// </summary>
        /// <param name="value">Pixels.</param>
		/// <returns>Centimeters.</returns>
		public static double DipToCm(double value)
		{
			return (value * UnitHelper.CmToInch) / UnitHelper.DPI;
		}

		/// <summary>
		/// Converts dips to milimeters.
		/// </summary>
        /// <param name="value">Pixels.</param>
		/// <returns>Millimeters.</returns>
		public static double DipToMm(double value)
		{
			return DipToCm(value * 10);
		}

		/// <summary>
		/// Converts dips to inches.
		/// </summary>
        /// <param name="value">Pixels.</param>
		/// <returns>Inches.</returns>
		public static double DipToInch(double value)
		{
			return (value / UnitHelper.DPI);
		}

		/// <summary>
		/// Converts dips to twips.
		/// </summary>
        /// <param name="value">Pixels.</param>
		/// <returns>Twips.</returns>
		public static double DipToTwip(double value)
		{
			return (value * UnitHelper.TwipToInch) / UnitHelper.DPI;
		}

        /// <summary>
        /// Converts dips to EMUs.
        /// </summary>
        /// <param name="value">Dips.</param>
        /// <returns>EMUs.</returns>
        public static double DipToEmu(double value)
        {
			return (value * UnitHelper.EmuToInch) / UnitHelper.DPI;
		}

		/// <summary>
		/// Converts dips to twips.
		/// </summary>
        /// <param name="value">Pixels.</param>
		/// <returns>Twips.</returns>
		public static int DipToTwipI(double value)
		{
			return (int)Math.Round(UnitHelper.DipToTwip(value));
		}

		/// <summary>
		/// Converts dips to twips.
		/// </summary>
        /// <param name="value">Pixels.</param>
		/// <returns>Twips.</returns>
		public static float DipToTwipF(double value)
		{
			return (float)UnitHelper.DipToTwip(value);
		}

		/// <summary>
		/// Converts twips to dips.
		/// </summary>
        /// <param name="value">Twips.</param>
		/// <returns>Dips.</returns>
		public static float TwipToDipF(double value)
		{
			return (float)UnitHelper.TwipToDip(value);
		}

		/// <summary>
		/// Converts twips to dips.
		/// </summary>
        /// <param name="value">Twips.</param>
        /// <returns>Dips.</returns>
		public static int TwipToDipI(double value)
		{
			return (int)UnitHelper.TwipToDip(value);
		}
		/// <summary>
		/// Converts DIPs to units.
		/// </summary>
        /// <param name="value">Pixels.</param>
		/// <param name="type">Unit type.</param>
		/// <returns>Units.</returns>
		public static double DipToUnit(double value, UnitType type)
		{
			switch (type)
			{
				case UnitType.Dip:
					return value;
				case UnitType.Point:
					return UnitHelper.DipToPoint(value);
				case UnitType.Pica:
					return UnitHelper.DipToPica(value);
				case UnitType.Inch:
					return UnitHelper.DipToInch(value);
				case UnitType.Mm:
					return UnitHelper.DipToMm(value);
				case UnitType.Cm:
					return UnitHelper.DipToCm(value);
				case UnitType.Twip:
					return UnitHelper.DipToTwip(value);
				default:
                    throw new ArgumentException("Unknown UnitType");
			}
		}

		#endregion


		#region Convert From Unit to Dip

		/// <summary>
		/// Converts points to dips.
		/// </summary>
		/// <param name="value">Points.</param>
		/// <returns>Pixels.</returns>
		public static double PointToDip(double value)
		{
			return (value * UnitHelper.DPI) / UnitHelper.PointToInch;
		}

		/// <summary>
		/// Converts to points dips.
		/// </summary>
		/// <param name="value">Picas.</param>
		/// <returns>Pixels.</returns>
		public static double PicaToDip(double value)
		{
			return (value * UnitHelper.DPI) / UnitHelper.PicaToInch;
		}

		/// <summary>
		/// Converts EMUs to dips.
		/// </summary>
		/// <param name="value">EMUs.</param>
		/// <returns>Pixels.</returns>
		public static double EmuToDip(double value)
		{
			return (value * UnitHelper.DPI) / UnitHelper.EmuToInch;
		}

		/// <summary>
		/// Converts centimeters to dips.
		/// </summary>
		/// <param name="value">Centimeters.</param>
        /// <returns>Pixels.</returns>
		public static double CmToDip(double value)
		{
			return (value * UnitHelper.DPI) / UnitHelper.CmToInch;
		}

		/// <summary>
		/// Converts milimeteres to dips.
		/// </summary>
		/// <param name="value">Millimeteres.</param>
        /// <returns>Pixels.</returns>
		public static double MmToDip(double value)
		{
			return CmToDip(value / 10);
		}

		/// <summary>
		/// Converts inches to dips.
		/// </summary>
		/// <param name="value">Inches.</param>
        /// <returns>Pixels.</returns>
		public static double InchToDip(double value)
		{
			return (value * UnitHelper.DPI);
		}

		/// <summary>
		/// Converts twips to dips.
		/// </summary>
		/// <param name="value">Twips.</param>
        /// <returns>Pixels.</returns>
		public static double TwipToDip(double value)
		{
			return (value * UnitHelper.DPI) / UnitHelper.TwipToInch;
		}

		/// <summary>
		/// Converts Units to dips.
		/// </summary>
		/// <param name="value">Units.</param>
		/// <param name="type">Unit type.</param>
        /// <returns>Pixels.</returns>
		public static double UnitToDip(double value, UnitType type)
		{
			switch (type)
			{
				case UnitType.Dip:
					return value;
				case UnitType.Point:
					return UnitHelper.PointToDip(value);
				case UnitType.Pica:
					return UnitHelper.PicaToDip(value);
				case UnitType.Inch:
					return UnitHelper.InchToDip(value);
				case UnitType.Mm:
					return UnitHelper.MmToDip(value);
				case UnitType.Cm:
					return UnitHelper.CmToDip(value);
				case UnitType.Twip:
					return UnitHelper.TwipToDip(value);
				default:
                    throw new ArgumentException("Unknown UnitType");
			}
		}

		#endregion


        #region ColumnWidth converters

        private static double GetMaxDigitWidthInNormalStyle(Workbook workbook)
        {
            Guard.ThrowExceptionIfNull(workbook, "workbook");

            CellStyle cellStyle = workbook.Styles["Normal"];

            FontProperties fontProperties = new FontProperties();
            fontProperties.FontFamily = cellStyle.FontFamily.GetActualValue(workbook.Theme);
            fontProperties.FontSize = cellStyle.FontSize;
            fontProperties.IsBold = cellStyle.IsBold;
            fontProperties.IsItalic = cellStyle.IsItalic;

            double maxDigitWidth = 0;
            for (int i = 0; i <= 9; i++)
            {
                maxDigitWidth = Math.Max(maxDigitWidth, RadTextMeasurer.Measure(i.ToString(), fontProperties).Size.Width);
            }

            return maxDigitWidth;
        }

        private static double PixelsToCharacterCount(double pixels, double maxDigitWidth)
        {
            double marginPixels = SpreadsheetDefaultValues.LeftCellMargin + SpreadsheetDefaultValues.RightCellMargin;
            return Math.Floor((pixels - marginPixels) / maxDigitWidth * 100 + 0.5) / 100;
        }

        private static double CharacterCountToExcelColumnWidth(double charCount, double maxDigitWidth)
        {
            double marginPixels = SpreadsheetDefaultValues.LeftCellMargin + SpreadsheetDefaultValues.RightCellMargin;
            return Math.Floor((charCount * maxDigitWidth + marginPixels) / maxDigitWidth * 256) / 256;
        }

        /// <summary>
        /// Converts column width in pixels to excel column width.
        /// </summary>
        /// <param name="workbook">The workbook.</param>
        /// <param name="pixels">The pixels.</param>
        /// <returns></returns>
        public static double PixelWidthToExcelColumnWidth(Workbook workbook, double pixels)
        {
            double maxDigitWidth = Math.Floor(GetMaxDigitWidthInNormalStyle(workbook));
            double charCount = PixelsToCharacterCount(pixels, maxDigitWidth);
            return CharacterCountToExcelColumnWidth(charCount, maxDigitWidth);
        }

        /// <summary>
        /// Converts excel column width to pixels width.
        /// </summary>
        /// <param name="workbook">The workbook.</param>
        /// <param name="width">The width.</param>
        /// <returns></returns>
        public static double ExcelColumnWidthToPixelWidth(Workbook workbook, double width)
        {
            double maxDigitWidth = Math.Floor(GetMaxDigitWidthInNormalStyle(workbook));
            return Math.Floor(((256 * width + Math.Floor(128 / maxDigitWidth)) / 256) * maxDigitWidth);
        }

        #endregion
    }
}