﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.Contexts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Contexts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Types;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Theming;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Data
{
    internal class XlsxDataWorkbookExportContext : IXlsxWorkbookExportContext
    {
        private readonly StyleSheet styleSheet;
        private readonly DataWorkbook workbook;
        private readonly Dictionary<string, int> styleNameToFormattingRecordIndex;
        private readonly Dictionary<string, XlsxDataWorksheetExportContext> exportContexts;

        public XlsxDataWorkbookExportContext(DataWorkbook workbook)
        {
            Guard.ThrowExceptionIfNull(workbook, "workbook");

            this.workbook = workbook;

            this.exportContexts = new Dictionary<string, XlsxDataWorksheetExportContext>();
            foreach (DataWorksheet worksheet in this.workbook.Worksheets)
            {
                this.exportContexts[worksheet.Name] = new XlsxDataWorksheetExportContext(this, worksheet);
            }

            this.styleSheet = new StyleSheet(true);
            this.styleNameToFormattingRecordIndex = new Dictionary<string, int>();

            this.Initialize();
        }

        public ResourceIndexedTable<SharedString> SharedStrings
        {
            get { return new ResourceIndexedTable<SharedString>(); }
        }

        public DocumentTheme Theme
        {
            get
            {
                return PredefinedThemeSchemes.DefaultTheme;
            }
        }

        public StyleSheet StyleSheet
        {
            get
            {
                return this.styleSheet;
            }
        }

        public IEnumerable<IXlsxWorksheetExportContext> WorksheetContexts 
        {
            get
            {
                return this.exportContexts.Values;
            }
        }

        private void Initialize()
        {
            this.InitializeSharedStrings();
            this.InitializeStyleSheet();
        }

        private void InitializeSharedStrings()
        {
            ISet<SharedString> sharedStringsSet = new HashSet<SharedString>();
            foreach (XlsxDataWorksheetExportContext worksheetContext in this.WorksheetContexts)
            {
                CollectWorksheetSharedStrings(worksheetContext.Worksheet, sharedStringsSet);
            }

            this.SharedStrings.AddRange(sharedStringsSet);
        }

        private void InitializeStyleSheet()
        {
            this.InitDefaultFormatting();
            this.InitStyleFormatting();
            this.InitDirectFormatting();
        }

        private static void CollectWorksheetSharedStrings(DataWorksheet worksheet, ISet<SharedString> sharedStringsSet)
        {
            sharedStringsSet = new HashSet<SharedString>();
            foreach (RowInfo row in worksheet.GetNonEmptyRows())
            {
                foreach (CellInfo cell in worksheet.GetNonEmptyCellsInRow(row.RowIndex))
                {
                    if (cell.CellValue.ValueType == Model.CellValueType.Text)
                    {
                        sharedStringsSet.Add(new TextSharedString(cell.CellValue.RawValue));
                    }
                }
            }
        }

        private void InitDefaultFormatting()
        {
            this.StyleSheet.FillTable.Add(NoneFill.Instance);
            this.StyleSheet.FillTable.Add(new PatternFill(PatternType.Gray12Percent, null, null));

            this.InitDefaultNumberFormatting();
        }

        private void InitDefaultNumberFormatting()
        {
            foreach (var formatStringPair in NumberFormatTypes.BuiltInFormatStrings)
            {
                this.StyleSheet.CellValueFormatTable.Add(new CellValueFormat(formatStringPair.Key), formatStringPair.Value);
            }
        }

        private void InitStyleFormatting()
        {
            ResourceIndexedTable<string> styleNameTable = new ResourceIndexedTable<string>();
            this.InitStyleNameTable(styleNameTable);

            foreach (var styleName in styleNameTable)
            {
                this.InitStyleFormatting(styleName);
            }
        }

        private void InitStyleFormatting(string styleName)
        {
            FormattingRecord formattingRecord = new FormattingRecord();

            CellStyle cellStyle = this.workbook.Styles[styleName];

            if (cellStyle.IsPropertyValueSet(CellPropertyDefinitions.FillProperty))
            {
                if (!TelerikHelper.EqualsOfT(cellStyle.Fill, CellPropertyDefinitions.FillProperty.DefaultValue))
                {
                    formattingRecord.FillId = this.StyleSheet.FillTable.Add(cellStyle.Fill);
                }
            }

            if (cellStyle.IsPropertyValueSet(CellPropertyDefinitions.FormatProperty))
            {
                formattingRecord.NumberFormatId = this.StyleSheet.CellValueFormatTable.Add(cellStyle.Format);
            }

            if (cellStyle.IsPropertyValueSet(CellPropertyDefinitions.HorizontalAlignmentProperty))
            {
                formattingRecord.HorizontalAlignment = cellStyle.HorizontalAlignment;
            }

            if (cellStyle.IsPropertyValueSet(CellPropertyDefinitions.VerticalAlignmentProperty))
            {
                formattingRecord.VerticalAlignment = cellStyle.VerticalAlignment;
            }

            if (cellStyle.IsPropertyValueSet(CellPropertyDefinitions.IndentProperty))
            {
                formattingRecord.Indent = cellStyle.Indent;
            }

            if (cellStyle.IsPropertyValueSet(CellPropertyDefinitions.IsWrappedProperty))
            {
                formattingRecord.WrapText = cellStyle.IsWrapped;
            }

            formattingRecord.ApplyNumberFormat = cellStyle.IncludeNumber;
            formattingRecord.ApplyAlignment = cellStyle.IncludeAlignment;
            formattingRecord.ApplyFont = cellStyle.IncludeFont;
            formattingRecord.ApplyBorder = cellStyle.IncludeBorder;
            formattingRecord.ApplyFill = cellStyle.IncludeFill;

            formattingRecord.FontInfoId = this.StyleSheet.FontInfoTable.Add(new FontInfo(cellStyle));
            formattingRecord.BordersInfoId = this.StyleSheet.BordersInfoTable.Add(new BordersInfo(cellStyle));


            int formattingRecordIndex = this.StyleSheet.StyleFormattingTable.Add(formattingRecord);
            this.styleNameToFormattingRecordIndex.Add(styleName, formattingRecordIndex);
            this.StyleSheet.DirectFormattingTable.Add(new FormattingRecord(formattingRecord) { StyleFormattingRecordId = formattingRecordIndex });

            this.StyleSheet.StyleInfoTable.Add(
                new StyleInfo()
                {
                    Name = styleName,
                    BuiltInId = cellStyle.BuiltInId,
                    FormattingRecordId = formattingRecordIndex
                });
        }

        private void InitStyleNameTable(ResourceIndexedTable<string> styleNameTable)
        {
            styleNameTable.Add("Normal");
        }

        private void InitDirectFormatting()
        {
            this.InitDirectFormattingFontInfoTable();
            this.InitDirectFormattingFills();
            this.InitDirectFormattingCellValueFormats();
            this.InitDirectFormattingBordersInfoTable();
            this.InitDirectFormattingTable();
        }

        private void InitDirectFormattingFontInfoTable()
        {
            foreach (XlsxDataWorksheetExportContext worksheetContext in this.WorksheetContexts)
            {
                this.StyleSheet.FontInfoTable.AddRange(worksheetContext.GetFontInfos());
            }
        }

        private void InitDirectFormattingFills()
        {
            foreach (XlsxDataWorksheetExportContext worksheetContext in this.WorksheetContexts)
            {
                this.StyleSheet.FillTable.AddRange(worksheetContext.GetFills());
            }
        }

        private void InitDirectFormattingCellValueFormats()
        {
            foreach (XlsxDataWorksheetExportContext worksheetContext in this.WorksheetContexts)
            {
                this.StyleSheet.CellValueFormatTable.AddRange(worksheetContext.GetCellValueFormats());
            }
        }

        private void InitDirectFormattingBordersInfoTable()
        {
            foreach (XlsxDataWorksheetExportContext worksheetContext in this.WorksheetContexts)
            {
                this.StyleSheet.BordersInfoTable.AddRange(worksheetContext.GetBorderInfos());
            }
        }

        private void InitDirectFormattingTable()
        {
            foreach (XlsxDataWorksheetExportContext worksheetContext in this.WorksheetContexts)
            {
                this.StyleSheet.DirectFormattingTable.AddRange(worksheetContext.GetDirectFormattings());
            }
        }

        public ResourceIndexedTable<DefinedNameInfo> DefinedNames
        {
            get
            {
                return new ResourceIndexedTable<DefinedNameInfo>();
            }
        }

        public IXlsxWorksheetExportContext GetWorksheetContextByRelationshipId(string worksheetName)
        {
            return this.exportContexts[worksheetName];
        }


        public string GetRelationshipIdFromWorksheetContext(IXlsxWorksheetExportContext context)
        {
            throw new NotImplementedException();
        }


        public void RegisterWorksheetContext(string relationshipId, IXlsxWorksheetExportContext context)
        {
            throw new NotImplementedException();
        }
    }
}
