﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
#if NETCORE || NETSTANDARD
using Telerik.Documents.Common.Model;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Contexts
{
    internal class GradientFillInfo
    {
        private static readonly Dictionary<GradientType, Func<GradientFill, GradientFillInfo>> gradientTypeToInfoFactory;
        private List<GradientStop> stops;

        static GradientFillInfo()
        {
            gradientTypeToInfoFactory = new Dictionary<GradientType, Func<GradientFill, GradientFillInfo>>();
            InitGradientInfoFactories();
        }

        public GradientFillInfo()
        {
            this.Stops = new List<GradientStop>();
        }

        public GradientFillInfo(IEnumerable<GradientStop> stops,
            double? degree = null,
            GradientInfoType? infoType = null,
            double? left = null,
            double? top = null,
            double? right = null,
            double? bottom = null)
        {
            Guard.ThrowExceptionIfNull(stops, "stops");

            this.Stops = stops.ToList();
            this.Degree = degree;
            this.Left = left;
            this.Top = top;
            this.Right = right;
            this.Bottom = bottom;
            this.InfoType = infoType;
        }

        public List<GradientStop> Stops
        {
            get
            {
                return this.stops;
            }
            set
            {
                Guard.ThrowExceptionIfNull(value, "value");
                this.stops = value;
            }
        }

        public double? Degree { get; set; }

        public double? Left { get; set; }

        public double? Top { get; set; }

        public double? Right { get; set; }

        public double? Bottom { get; set; }

        public GradientInfoType? InfoType { get; set; }

        public static GradientFillInfo Create(GradientFill gradientFill)
        {
            return gradientTypeToInfoFactory[gradientFill.GradientType](gradientFill);
        }

        public GradientFill ToFill()
        {
            ThemableColor color1 = null;
            ThemableColor color2 = null;
            GradientType gradientType = GradientType.Horizontal;

            foreach (GradientStop gradeintStop in this.Stops)
            {
                if (color1 == null)
                {
                    color1 = gradeintStop.ThemableColor;
                }
                else if (color2 == null)
                {
                    color2 = gradeintStop.ThemableColor;
                }
            }

            Guard.ThrowExceptionIfNull(color1, "color1");
            Guard.ThrowExceptionIfNull(color2, "color2");

            GradientFill fakeGradientFill = new GradientFill(GradientType.Horizontal, color1, color2);
            foreach (KeyValuePair<GradientType, Func<GradientFill, GradientFillInfo>> pair in gradientTypeToInfoFactory)
            {
                Func<GradientFill, GradientFillInfo> fillInfoFactory = pair.Value;
                GradientFillInfo other = fillInfoFactory(fakeGradientFill);

                if (this.Stops.Count == other.Stops.Count &&
                    TelerikHelper.EqualsOfT(this.Degree, other.Degree) &&
                    TelerikHelper.EqualsOfT(this.Left, other.Left) &&
                    TelerikHelper.EqualsOfT(this.Right, other.Right) &&
                    TelerikHelper.EqualsOfT(this.Top, other.Top) &&
                    TelerikHelper.EqualsOfT(this.Bottom, other.Bottom) &&
                    TelerikHelper.EqualsOfT(this.InfoType, other.InfoType))
                {
                    bool areSame = true;
                    for (int i = 0; i < this.Stops.Count; i++)
                    {
                        if (this.stops[i].Position != other.stops[i].Position)
                        {
                            areSame = false;
                            break;
                        }
                    }

                    if (areSame)
                    {
                        gradientType = pair.Key;
                        break;
                    }
                }
            }

            return new GradientFill(gradientType, color1, color2);
        }

        private static void InitGradientInfoFactories()
        {
            RegisterGradientCreationMethod(GradientType.Horizontal,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops, 90);
                });

            RegisterGradientCreationMethod(GradientType.HorizontalReversed,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops, 270);
                });

            RegisterGradientCreationMethod(GradientType.HorizontalRepeated,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(0.5, gradientFill.Color2),
                        new GradientStop(1, gradientFill.Color1)
                    };

                    return new GradientFillInfo(gradientStops, 90);
                });

            RegisterGradientCreationMethod(GradientType.Vertical,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops);
                });

            RegisterGradientCreationMethod(GradientType.VerticalReversed,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops, 180);
                });

            RegisterGradientCreationMethod(GradientType.VerticalRepeated,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(0.5, gradientFill.Color2),
                        new GradientStop(1, gradientFill.Color1)
                    };

                    return new GradientFillInfo(gradientStops);
                });

            RegisterGradientCreationMethod(GradientType.DiagonalUp,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2),
                    };

                    return new GradientFillInfo(gradientStops, 45);
                });

            RegisterGradientCreationMethod(GradientType.DiagonalUpReversed,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2),
                    };

                    return new GradientFillInfo(gradientStops, 225);
                });

            RegisterGradientCreationMethod(GradientType.DiagonalUpRepeated,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(0.5, gradientFill.Color2),
                        new GradientStop(1, gradientFill.Color1),
                    };

                    return new GradientFillInfo(gradientStops, 45);
                });

            RegisterGradientCreationMethod(GradientType.DiagonalDown,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2),
                    };

                    return new GradientFillInfo(gradientStops, 135);
                });

            RegisterGradientCreationMethod(GradientType.DiagonalDownReversed,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2),
                    };

                    return new GradientFillInfo(gradientStops, 315);
                });

            RegisterGradientCreationMethod(GradientType.DiagonalDownRepeated,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(0.5, gradientFill.Color2),
                        new GradientStop(1, gradientFill.Color1),
                    };

                    return new GradientFillInfo(gradientStops, 135);
                });

            RegisterGradientCreationMethod(GradientType.FromTopLeftCorner,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops, infoType: GradientInfoType.Path);
                });

            RegisterGradientCreationMethod(GradientType.FromTopRightCorner,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops, infoType: GradientInfoType.Path, left: 1, right: 1);
                });

            RegisterGradientCreationMethod(GradientType.FromBottomLeftCorner,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops, infoType: GradientInfoType.Path, top: 1, bottom: 1);
                });

            RegisterGradientCreationMethod(GradientType.FromBottomRightCorner,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops, infoType: GradientInfoType.Path, left: 1, top: 1, right: 1, bottom: 1);
                });

            RegisterGradientCreationMethod(GradientType.FromCenter,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops, infoType: GradientInfoType.Path, left: 0.5, top: 0.5, right: 0.5, bottom: 0.5);
                });
        }

        private static void RegisterGradientCreationMethod(GradientType gradientType, Func<GradientFill, GradientFillInfo> factory)
        {
            gradientTypeToInfoFactory.Add(gradientType, factory);
        }
    }
}