﻿using System.Collections.Generic;
using Telerik.Windows.Documents.Common.Model.Data;
using Telerik.Windows.Documents.Core.Data;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Parts;
#if !SILVERLIGHT
using Telerik.Windows.Documents.Model.Drawing.Shapes;
#endif
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Parts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Types;
using Telerik.Windows.Documents.Spreadsheet.Model;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Theming;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Contexts
{
    internal abstract class XlsxWorkbookContextBase<TWorksheetContext>
        where TWorksheetContext : XlsxWorksheetContextBase
    {
        private readonly Workbook workbook;
        private readonly Dictionary<Worksheet, TWorksheetContext> worksheetToContext;
        private readonly ResourceIndexedTable<SharedString> sharedStrings;
        private readonly StyleSheet styleSheet;
        private readonly ResourceIndexedTable<DefinedNameInfo> definedNames;
        private readonly ValueMapper<Worksheet, WorksheetPart> worksheetToWorksheetPartMapping;
        private readonly ValueMapper<Worksheet, DrawingPart> worksheetToDrawingPartMapping;
#if !SILVERLIGHT
        private readonly ValueMapper<ChartShape, ChartPart> chartToChartPartMapping;
#endif

        public XlsxWorkbookContextBase(bool isExport, Workbook workbook)
        {
            Guard.ThrowExceptionIfNull(workbook, "workbook");

            this.workbook = workbook;
            this.worksheetToContext = new Dictionary<Worksheet, TWorksheetContext>();
            this.worksheetToWorksheetPartMapping = new ValueMapper<Worksheet, WorksheetPart>();
            this.worksheetToDrawingPartMapping = new ValueMapper<Worksheet, DrawingPart>();

            this.sharedStrings = new ResourceIndexedTable<SharedString>(isExport);
            this.styleSheet = new StyleSheet(isExport);
            this.definedNames = new ResourceIndexedTable<DefinedNameInfo>();
#if !SILVERLIGHT
            this.chartToChartPartMapping = new ValueMapper<ChartShape, ChartPart>();
#endif
        }

        public Workbook Workbook
        {
            get
            {
                return this.workbook;
            }
        }

        public ResourceManager Resources
        {
            get
            {
                return this.Workbook.Resources;
            }
        }

        public ResourceIndexedTable<SharedString> SharedStrings
        {
            get
            {
                return this.sharedStrings;
            }
        }

        public StyleSheet StyleSheet
        {
            get
            {
                return this.styleSheet;
            }
        }

        public DocumentTheme Theme
        {
            get
            {
                return this.workbook.Theme;
            }
            set
            {
                this.workbook.Theme = value;
            }
        }

        public ResourceIndexedTable<DefinedNameInfo> DefinedNames
        {
            get
            {
                return this.definedNames;
            }
        }

        public TWorksheetContext GetWorksheetContext(Worksheet worksheet)
        {
            return this.worksheetToContext[worksheet];
        }

        public TWorksheetContext GetWorksheetContext(string worksheetName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(worksheetName, "worksheetName");

            Worksheet worksheet = this.Workbook.Worksheets.GetByName(worksheetName);

            TWorksheetContext worksheetContext = null;
            if (worksheet != null)
            {
                worksheetContext = this.worksheetToContext[worksheet];
            }

            return worksheetContext;
        }

        public IEnumerable<TWorksheetContext> GetWorksheetContexts()
        {
            return this.worksheetToContext.Values;
        }

        public void AddDefinedNameInfo(DefinedNameInfo definedNameInfo)
        {
            this.definedNames.Add(definedNameInfo);
        }

        public void RegisterWorksheetPart(Worksheet worksheet, WorksheetPart worksheetPart)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfNull(worksheetPart, "worksheetPart");

            this.worksheetToWorksheetPartMapping.AddPair(worksheet, worksheetPart);
        }

        public WorksheetPart GetWorksheetPartFromWorksheet(Worksheet worksheet)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");

            return this.worksheetToWorksheetPartMapping.GetToValue(worksheet);
        }

        public Worksheet GetWorksheetFromWorksheetPart(WorksheetPart worksheetPart)
        {
            Guard.ThrowExceptionIfNull(worksheetPart, "worksheetPart");

            return this.worksheetToWorksheetPartMapping.GetFromValue(worksheetPart);
        }

        public void RegisterDrawingPart(Worksheet worksheet, DrawingPart drawingPart)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");
            Guard.ThrowExceptionIfNull(drawingPart, "drawingPart");

            this.worksheetToDrawingPartMapping.AddPair(worksheet, drawingPart);
        }

        public bool HasRegisteredDrawingPart(DrawingPart drawingPart)
        {
            Guard.ThrowExceptionIfNull(drawingPart, "drawingPart");

            return this.worksheetToDrawingPartMapping.ContainsToValue(drawingPart);
        }

        public DrawingPart GetDrawingPartFromWorksheet(Worksheet worksheet)
        {
            Guard.ThrowExceptionIfNull(worksheet, "worksheet");

            return this.worksheetToDrawingPartMapping.GetToValue(worksheet);
        }

        public Worksheet GetWorksheetFromDrawingPart(DrawingPart drawingPart)
        {
            Guard.ThrowExceptionIfNull(drawingPart, "drawingPart");

            return this.worksheetToDrawingPartMapping.GetFromValue(drawingPart);
        }

#if !SILVERLIGHT
        public void RegisterChartPartForChart(ChartShape chart, ChartPart chartPart)
        {
            Guard.ThrowExceptionIfNull(chart, "chart");
            Guard.ThrowExceptionIfNull(chartPart, "chartPart");

            this.chartToChartPartMapping.AddPair(chart, chartPart);
        }

        public ChartPart GetChartPartForChart(ChartShape chart)
        {
            Guard.ThrowExceptionIfNull(chart, "chart");

            return this.chartToChartPartMapping.GetToValue(chart);
        }

        public ChartShape GetChartForChartPart(ChartPart chartPart)
        {
            Guard.ThrowExceptionIfNull(chartPart, "chartPart");

            ChartShape result;
            this.chartToChartPartMapping.TryGetFromValue(chartPart, out result);
            return result;
        }
#endif

        protected void InitDefaultNumberFormatting()
        {
            foreach (KeyValuePair<string, int> formatStringPair in NumberFormatTypes.BuiltInFormatStrings)
            {
                this.StyleSheet.CellValueFormatTable.Add(new CellValueFormat(formatStringPair.Key), formatStringPair.Value);
            }
        }

        protected void AddWorksheetContext(TWorksheetContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            this.worksheetToContext.Add(context.Worksheet, context);
        }
    }
}