﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Core.Data;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.Contexts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Export;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Parts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Filtering;
using Telerik.Windows.Documents.Spreadsheet.Model.Printing;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Contexts
{
    internal class XlsxWorkbookExportContext : XlsxWorkbookContextBase<XlsxWorksheetExportContext>, IXlsxWorkbookExportContext
    {
        private readonly Dictionary<string, int> styleNameToFormattingRecordIndex;
        private readonly Dictionary<string, IXlsxWorksheetExportContext> relationshipIdToWorksheetContextMapping;
        private readonly Dictionary<IXlsxWorksheetExportContext, string> worksheetContextToRelationshipMapping;
        private readonly DifferentialFormatsExportContext formatsContext;

        public XlsxWorkbookExportContext(Workbook workbook)
            : base(true, workbook)
        {
            this.styleNameToFormattingRecordIndex = new Dictionary<string, int>();
            this.worksheetContextToRelationshipMapping = new Dictionary<IXlsxWorksheetExportContext, string>();
            this.relationshipIdToWorksheetContextMapping = new Dictionary<string, IXlsxWorksheetExportContext>();
            this.formatsContext = new DifferentialFormatsExportContext();

            this.Initialize();
        }

        public IEnumerable<IXlsxWorksheetExportContext> WorksheetContexts
        {
            get
            {
                return this.GetWorksheetContexts();
            }
        }

        public DifferentialFormatsExportContext DifferentialFormatsContext
        {
            get
            {
                return this.formatsContext;
            }
        }

        public int GetStyleFormattingRecordId(string styleName)
        {
            return this.styleNameToFormattingRecordIndex[styleName];
        }

        public string GetRelationshipIdByResource(Common.Model.Data.IResource resource)
        {
            throw new NotImplementedException();
        }

        public void RegisterResource(string relationshipId, Common.Model.Data.IResource resource)
        {
            throw new NotImplementedException();
        }

        WorkbookProtectionInfo IXlsxWorkbookExportContext.GetWorkbookProtectionInfo()
        {
            WorkbookProtectionInfo info = new WorkbookProtectionInfo();

            info.Enforced = this.Workbook.IsProtected;
            info.AlgorithmName = this.Workbook.ProtectionData.AlgorithmName;
            info.HashValue = this.Workbook.ProtectionData.Hash;
            info.SaltValue = this.Workbook.ProtectionData.Salt;
            info.SpinCount = this.Workbook.ProtectionData.SpinCount;
            info.Password = this.Workbook.ProtectionData.Password;
            info.LockStructure = this.Workbook.ProtectionOptions.LockStructure;
            info.LockWindows = this.Workbook.ProtectionOptions.LockWindows;

            return info;
        }

        void IXlsxWorkbookExportContext.RegisterWorksheetContext(string relationshipId, IXlsxWorksheetExportContext context)
        {
            this.relationshipIdToWorksheetContextMapping[relationshipId] = context;
            this.worksheetContextToRelationshipMapping[context] = relationshipId;
        }

        IXlsxWorksheetExportContext IXlsxWorkbookExportContext.GetWorksheetContextByRelationshipId(string relationshipId)
        {
            return this.relationshipIdToWorksheetContextMapping[relationshipId];
        }

        string IXlsxWorkbookExportContext.GetRelationshipIdFromWorksheetContext(IXlsxWorksheetExportContext context)
        {
            return this.worksheetContextToRelationshipMapping[context];
        }

        public IXlsxWorksheetExportContext GetWorksheetContextFromWorksheetPart(WorksheetPart worksheetPart)
        {
            Guard.ThrowExceptionIfNull(worksheetPart, "worksheetPart");

            return this.GetWorksheetContext(this.GetWorksheetFromWorksheetPart(worksheetPart));
        }

        public IXlsxWorksheetExportContext GetWorksheetContextFromDrawingPart(Model.Parts.DrawingPart drawingPart)
        {
            Guard.ThrowExceptionIfNull(drawingPart, "drawingPart");

            return this.GetWorksheetContext(this.GetWorksheetFromDrawingPart(drawingPart));
        }

        public int GetActiveTabIndex()
        {
            return this.Workbook.Sheets.IndexOf(this.Workbook.ActiveSheet);
        }

        public IEnumerable<WorksheetEntityBase> GetWorksheetEntitiesFromWorksheet(Worksheet worksheet)
        {
            return new WorksheetEntityBase[]
            {
                worksheet.Cells,
                worksheet.Rows,
                worksheet.Columns
            };
        }

        private static void CollectWorksheetSharedStrings(Worksheet worksheet, ISet<SharedString> sharedStringsSet)
        {
            ICompressedList<ICellValue> cellValueRanges = worksheet.Cells.PropertyBag.GetPropertyValueCollection(CellPropertyDefinitions.ValueProperty);
            foreach (var range in cellValueRanges)
            {
                TextCellValue textCellValue = range.Value as TextCellValue;
                if (textCellValue != null)
                {
                    sharedStringsSet.Add(new TextSharedString(textCellValue.Value));
                }
            }
        }

        private void Initialize()
        {
            this.InitializeWorksheetContexts();
            this.InitializeSharedStrings();
            this.InitializeStyleSheet();
            this.InitializeDefinedNames();
        }

        private void InitializeWorksheetContexts()
        {
            for (int i = 0; i < this.Workbook.Worksheets.Count; i++)
            {
                this.AddWorksheetContext(new XlsxWorksheetExportContext(this, this.Workbook.Worksheets[i], i + 1));
            }
        }

        private void InitializeSharedStrings()
        {
            // NOTE: Multithreading here is slower!
            List<Worksheet> worksheets = new List<Worksheet>(this.Workbook.Worksheets);

            ISet<SharedString> sharedStringsSet = new HashSet<SharedString>();
            for (int i = 0; i < worksheets.Count; i++)
            {
                int worksheetIndex = i;
                CollectWorksheetSharedStrings(worksheets[worksheetIndex], sharedStringsSet);
            }

            this.SharedStrings.AddRange(sharedStringsSet);
        }

        private void InitializeStyleSheet()
        {
            this.InitDefaultFormatting();
            this.InitStyleFormatting();
            this.InitDirectFormatting();
        }

        private void InitDefaultFormatting()
        {
            this.StyleSheet.FillTable.Add(NoneFill.Instance);
            this.StyleSheet.FillTable.Add(new PatternFill(PatternType.Gray12Percent, null, null));

            this.InitDefaultNumberFormatting();
        }

        private void InitStyleFormatting()
        {
            ResourceIndexedTable<string> styleNameTable = new ResourceIndexedTable<string>();
            this.InitStyleNameTable(styleNameTable);

            foreach (var styleName in styleNameTable)
            {
                this.InitStyleFormatting(styleName);
            }
        }

        private void InitStyleFormatting(string styleName)
        {
            FormattingRecord formattingRecord = new FormattingRecord();

            CellStyle cellStyle = this.Workbook.Styles[styleName];

            if (cellStyle.IsPropertyValueSet(CellPropertyDefinitions.FillProperty))
            {
                if (!TelerikHelper.EqualsOfT(cellStyle.Fill, CellPropertyDefinitions.FillProperty.DefaultValue))
                {
                    formattingRecord.FillId = this.StyleSheet.FillTable.Add(cellStyle.Fill);
                }
            }

            formattingRecord.NumberFormatId = this.StyleSheet.CellValueFormatTable.Add(cellStyle.Format);

            if (cellStyle.IsPropertyValueSet(CellPropertyDefinitions.HorizontalAlignmentProperty))
            {
                formattingRecord.HorizontalAlignment = cellStyle.HorizontalAlignment;
            }

            if (cellStyle.IsPropertyValueSet(CellPropertyDefinitions.VerticalAlignmentProperty))
            {
                formattingRecord.VerticalAlignment = cellStyle.VerticalAlignment;
            }

            if (cellStyle.IsPropertyValueSet(CellPropertyDefinitions.IndentProperty))
            {
                formattingRecord.Indent = cellStyle.Indent;
            }

            if (cellStyle.IsPropertyValueSet(CellPropertyDefinitions.IsWrappedProperty))
            {
                formattingRecord.WrapText = cellStyle.IsWrapped;
            }

            if (cellStyle.IsPropertyValueSet(CellPropertyDefinitions.IsLockedProperty))
            {
                formattingRecord.IsLocked = cellStyle.IsLocked;
            }

            formattingRecord.ApplyNumberFormat = cellStyle.IncludeNumber;
            formattingRecord.ApplyAlignment = cellStyle.IncludeAlignment;
            formattingRecord.ApplyFont = cellStyle.IncludeFont;
            formattingRecord.ApplyBorder = cellStyle.IncludeBorder;
            formattingRecord.ApplyFill = cellStyle.IncludeFill;
            formattingRecord.ApplyProtection = cellStyle.IncludeProtection;

            formattingRecord.FontInfoId = this.StyleSheet.FontInfoTable.Add(new FontInfo(cellStyle));
            formattingRecord.BordersInfoId = this.StyleSheet.BordersInfoTable.Add(new BordersInfo(cellStyle));

            int formattingRecordIndex = this.StyleSheet.StyleFormattingTable.Add(formattingRecord);
            this.styleNameToFormattingRecordIndex.Add(styleName, formattingRecordIndex);
            this.StyleSheet.DirectFormattingTable.Add(new FormattingRecord(formattingRecord) { StyleFormattingRecordId = formattingRecordIndex });

            this.StyleSheet.StyleInfoTable.Add(
                new StyleInfo()
                {
                    Name = styleName,
                    BuiltInId = cellStyle.BuiltInId,
                    FormattingRecordId = formattingRecordIndex
                });
        }

        private void InitStyleNameTable(ResourceIndexedTable<string> styleNameTable)
        {
            styleNameTable.Add("Normal");

            foreach (var worksheetContext in this.GetWorksheetContexts())
            {
                foreach (var entity in this.GetWorksheetEntitiesFromWorksheet(worksheetContext.Worksheet))
                {
                    var compressedList = worksheetContext.StyleNameCompressedList[entity];
                    foreach (var styleName in compressedList.GetNonDefaultRanges())
                    {
                        this.MarkRangeAsNonDefaultFormatted(worksheetContext, entity, compressedList, styleName);

                        styleNameTable.Add(styleName.Value);
                    }
                }
            }
        }
 
        private void MarkRangeAsNonDefaultFormatted<T>(XlsxWorksheetExportContext worksheetContext, WorksheetEntityBase entity, ICompressedList<T> compressedList, Range<long, T> range)
        {
            T defaultValue = compressedList.GetDefaultValue();
            if (!TelerikHelper.EqualsOfT(range.Value, defaultValue))
            {
                worksheetContext.NonDefaultFormattingCompressedList[entity].SetValue(range.Start, range.End, true);
            }
        }

        private void InitDirectFormatting()
        {
            this.InitDirectFormattingFontInfoTable();
            this.InitDirectFormattingFills();
            this.InitDirectFormattingCellValueFormats();
            this.InitDirectFormattingBordersInfoTable();
            this.InitDirectFormattingTable();
        }

        private void InitDirectFormattingFontInfoTable()
        {
            foreach (var worksheetContext in this.GetWorksheetContexts())
            {
                worksheetContext.InitFontInfos();

                foreach (var entity in this.GetWorksheetEntitiesFromWorksheet(worksheetContext.Worksheet))
                {
                    var compressedList = worksheetContext.FontInfoCompressedList[entity];
                    foreach (var fontInfo in compressedList.GetNonDefaultRanges())
                    {
                        this.MarkRangeAsNonDefaultFormatted(worksheetContext, entity, compressedList, fontInfo);
                        this.StyleSheet.FontInfoTable.Add(fontInfo.Value);
                    }
                }
            }
        }

        private void InitDirectFormattingFills()
        {
            foreach (var worksheetContext in this.GetWorksheetContexts())
            {
                foreach (var entity in this.GetWorksheetEntitiesFromWorksheet(worksheetContext.Worksheet))
                {
                    var compressedList = worksheetContext.FillCompressedList[entity];
                    foreach (var fill in worksheetContext.FillCompressedList[entity].GetNonDefaultRanges())
                    {
                        this.MarkRangeAsNonDefaultFormatted(worksheetContext, entity, compressedList, fill);
                        this.StyleSheet.FillTable.Add(fill.Value);
                    }
                }
            }
        }

        private void InitDirectFormattingCellValueFormats()
        {
            foreach (var worksheetContext in this.GetWorksheetContexts())
            {
                foreach (var entity in this.GetWorksheetEntitiesFromWorksheet(worksheetContext.Worksheet))
                {
                    var compressedList = worksheetContext.CellValueFormatCompressedList[entity];
                    foreach (var format in compressedList.GetNonDefaultRanges())
                    {
                        this.MarkRangeAsNonDefaultFormatted(worksheetContext, entity, compressedList, format);
                        this.StyleSheet.CellValueFormatTable.Add(format.Value);
                    }
                }
            }
        }

        private void InitDirectFormattingBordersInfoTable()
        {
            foreach (var worksheetContext in this.GetWorksheetContexts())
            {
                worksheetContext.InitBordersInfos();

                foreach (var entity in this.GetWorksheetEntitiesFromWorksheet(worksheetContext.Worksheet))
                {
                    var compressedList = worksheetContext.BordersInfoCompressedList[entity];
                    foreach (var borders in compressedList.GetNonDefaultRanges())
                    {
                        this.MarkRangeAsNonDefaultFormatted(worksheetContext, entity, compressedList, borders);
                        this.StyleSheet.BordersInfoTable.Add(borders.Value);
                    }
                }
            }
        }

        private void InitDirectFormattingTable()
        {
            foreach (var worksheetContext in this.GetWorksheetContexts())
            {
                worksheetContext.InitDirectFormattingRecords();

                foreach (var entity in this.GetWorksheetEntitiesFromWorksheet(worksheetContext.Worksheet))
                {
                    var compressedList = worksheetContext.DirectFormattingCompressedList[entity];
                    foreach (var formattingRecord in compressedList.GetNonDefaultRanges())
                    {
                        this.MarkRangeAsNonDefaultFormatted(worksheetContext, entity, compressedList, formattingRecord);
                        this.StyleSheet.DirectFormattingTable.Add(formattingRecord.Value);
                    }
                }
            }
        }

        private void InitializeDefinedNames()
        {
            foreach (KeyValuePair<int, IEnumerable<ISpreadsheetName>> pair in this.Workbook.NameManager.NameWithWorksheetIdCollections)
            {
                foreach (ISpreadsheetName spreadsheetName in pair.Value)
                {
                    DefinedName definedName = spreadsheetName as DefinedName;
                    if (definedName != null)
                    {
                        string value = definedName.FormulaCellValue.Value.ToString(XlsxHelper.CultureInfo);
                        DefinedNameInfo definedNameInfo = new DefinedNameInfo(definedName.Name, value, pair.Key, definedName.Comment, !definedName.IsVisible);

                        this.AddDefinedNameInfo(definedNameInfo);
                    }
                }
            }

            this.ExportRequiredDefinedNames();
        }

        private void ExportRequiredDefinedNames()
        {
            for (int i = 0; i < this.Workbook.Worksheets.Count; i++)
            {
                Worksheet worksheet = this.Workbook.Worksheets[i];

                PrintArea printArea = worksheet.WorksheetPageSetup.PrintArea;
                if (printArea.HasPrintAreaRanges)
                {
                    string absoluteCellReference = TextHelper.BuildAbsoluteCellReferenceString(worksheet.Name, printArea.Ranges, XlsxHelper.CultureInfo);
                    string definedNameValue = SpreadsheetCultureHelper.ClearFormulaValue(absoluteCellReference);
                    DefinedNameInfo definedNameInfo = new DefinedNameInfo(DefinedName.PrintAreaDefinedName, definedNameValue, i, string.Empty, false);

                    this.AddDefinedNameInfo(definedNameInfo);
                }

                AutoFilter filter = worksheet.Filter;
                if (filter.FilterRange != null)
                {
                    string absoluteCellReference = TextHelper.BuildAbsoluteCellReferenceString(worksheet.Name, filter.FilterRange, XlsxHelper.CultureInfo);
                    string definedNameValue = SpreadsheetCultureHelper.ClearFormulaValue(absoluteCellReference);
                    DefinedNameInfo definedNameInfo = new DefinedNameInfo(DefinedName.FilterDefinedName, definedNameValue, i, string.Empty, true);

                    this.AddDefinedNameInfo(definedNameInfo);
                }
            }
        }
    }
}