﻿using System;
using System.Linq;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements.Drawing;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Export;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Import;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Elements.Drawing
{
    internal class MarkerElement : WorksheetDrawingElementBase
    {
        private readonly string elementName;

        private readonly OpenXmlChildElement<ColumnRowElement> column;
        private readonly OpenXmlChildElement<OneDimentionOffsetElement> columnOffset;
        private readonly OpenXmlChildElement<ColumnRowElement> row;
        private readonly OpenXmlChildElement<OneDimentionOffsetElement> rowOffset;

        public MarkerElement(XlsxPartsManager partsManager, string elementName)
            : base(partsManager)
        {
            Guard.ThrowExceptionIfNullOrEmpty(elementName, "elementName");

            this.elementName = elementName;

            this.column = this.RegisterChildElement<ColumnRowElement>(XlsxElementNames.Column);
            this.columnOffset = this.RegisterChildElement<OneDimentionOffsetElement>(XlsxElementNames.ColumnOffset);
            this.row = this.RegisterChildElement<ColumnRowElement>(XlsxElementNames.Row);
            this.rowOffset = this.RegisterChildElement<OneDimentionOffsetElement>(XlsxElementNames.RowOffset);
        }

        public ColumnRowElement ColumnElement
        {
            get
            {
                return this.column.Element;
            }
        }

        public OneDimentionOffsetElement ColumnOffsetElement
        {
            get
            {
                return this.columnOffset.Element;
            }
        }

        public ColumnRowElement RowElement
        {
            get
            {
                return this.row.Element;
            }
        }

        public OneDimentionOffsetElement RowOffsetElement
        {
            get
            {
                return this.rowOffset.Element;
            }
        }

        public override string ElementName
        {
            get
            {
                return elementName;
            }
        }

        public void CopyPropertiesFrom(IXlsxWorksheetExportContext context, CellIndex cellIndex, double offsetX, double offsetY)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(cellIndex, "cellIndex");

            this.CreateElement(this.column);
            this.ColumnElement.InnerText = cellIndex.ColumnIndex.ToString();

            this.CreateElement(this.columnOffset);
            this.ColumnOffsetElement.InnerText = ((int)Unit.DipToEmu(offsetX)).ToString();

            this.CreateElement(this.row);
            this.RowElement.InnerText = cellIndex.RowIndex.ToString();

            this.CreateElement(this.rowOffset);
            this.RowOffsetElement.InnerText = ((int)Unit.DipToEmu(offsetY)).ToString();
        }

        public void CopyPropertiesTo(IXlsxWorksheetImportContext context, out CellIndex cellIndex, out double offsetX, out double offsetY)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            int cellRow = int.Parse(this.RowElement.InnerText);
            int cellCol = int.Parse(this.ColumnElement.InnerText);
            cellIndex = new CellIndex(cellRow, cellCol);

            offsetX = Unit.EmuToDip(int.Parse(this.ColumnOffsetElement.InnerText));
            offsetY = Unit.EmuToDip(int.Parse(this.RowOffsetElement.InnerText));

            this.ReleaseElement(this.column);
            this.ReleaseElement(this.columnOffset);
            this.ReleaseElement(this.row);
            this.ReleaseElement(this.rowOffset);
        }

        protected override OpenXmlElementBase CreateElement(string elementName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(elementName, "elementName");

            ColumnRowElement result;
            switch (elementName)
            {
                case XlsxElementNames.Column:

                    result = new ColumnRowElement(this.PartsManager, XlsxElementNames.Column);
                    result.Part = this.Part;

                    return result;

                case XlsxElementNames.Row:

                    result = new ColumnRowElement(this.PartsManager, XlsxElementNames.Row);
                    result.Part = this.Part;

                    return result;
            }

            return base.CreateElement(elementName);
        }

        protected override void ReleaseElementOverride(OpenXmlElementBase element)
        {
            Guard.ThrowExceptionIfNull(element, "element");

            switch (element.ElementName)
            {
                case XlsxElementNames.Column:
                case XlsxElementNames.Row:

                    return;
            }

            base.ReleaseElementOverride(element);
        }
    }
}
