﻿using System;
using System.Linq;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Contexts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Export;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Import;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Elements.StyleSheet
{
    internal class BorderElement : StyleSheetElementBase
    {
        private readonly BoolOpenXmlAttribute diagonalDown;
        private readonly BoolOpenXmlAttribute diagonalUp;

        private readonly OpenXmlChildElement<BottomBorderElement> bottom;
        private readonly OpenXmlChildElement<DiagonalBorderElement> diagonal;
        private readonly OpenXmlChildElement<LeftBorderElement> left;
        private readonly OpenXmlChildElement<RightBorderElement> right;
        private readonly OpenXmlChildElement<TopBorderElement> top;

        public BorderElement(XlsxPartsManager partsManager)
            : base(partsManager)
        {
            this.diagonalDown = this.RegisterAttribute(new BoolOpenXmlAttribute("diagonalDown"));
            this.diagonalUp = this.RegisterAttribute(new BoolOpenXmlAttribute("diagonalUp"));

            this.left = this.RegisterChildElement<LeftBorderElement>(XlsxElementNames.LeftBorderElementName);
            this.right = this.RegisterChildElement<RightBorderElement>(XlsxElementNames.RightBorderElementName);
            this.top = this.RegisterChildElement<TopBorderElement>(XlsxElementNames.TopBorderElementName);
            this.bottom = this.RegisterChildElement<BottomBorderElement>(XlsxElementNames.BottomBorderElementName);
            this.diagonal = this.RegisterChildElement<DiagonalBorderElement>(XlsxElementNames.DiagonalBorderElementName);
        }

        public override string ElementName
        {
            get
            {
                return XlsxElementNames.BorderElementName;
            }
        }

        public bool DiagonalDown
        {
            get
            {
                return this.diagonalDown.Value;
            }
            set
            {
                this.diagonalDown.Value = value;
            }
        }

        public bool DiagonalUp
        {
            get
            {
                return this.diagonalUp.Value;
            }
            set
            {
                this.diagonalUp.Value = value;
            }
        }

        public LeftBorderElement LeftBorder
        {
            get
            {
                return this.left.Element;
            }
        }

        public RightBorderElement RightBorder
        {
            get
            {
                return this.right.Element;
            }
        }

        public TopBorderElement TopBorder
        {
            get
            {
                return this.top.Element;
            }
        }

        public BottomBorderElement BottomBorder
        {
            get
            {
                return this.bottom.Element;
            }
        }

        public DiagonalBorderElement DiagonalBorder
        {
            get
            {
                return this.diagonal.Element;
            }
        }

        public void CopyPropertiesFrom(IXlsxWorkbookExportContext context, BordersInfo border)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(border, "border");

            if (border.Left != null)
            {
                this.CreateElement(this.left);
                this.LeftBorder.CopyPropertiesFrom(context, border.Left);
            }

            if (border.Top != null)
            {
                this.CreateElement(this.top);
                this.TopBorder.CopyPropertiesFrom(context, border.Top);
            }

            if (border.Right != null)
            {
                this.CreateElement(this.right);
                this.RightBorder.CopyPropertiesFrom(context, border.Right);
            }

            if (border.Bottom != null)
            {
                this.CreateElement(this.bottom);
                this.BottomBorder.CopyPropertiesFrom(context, border.Bottom);
            }

            if (!IsNullOrNone(border.DiagonalDown))
            {
                this.CreateElement(this.diagonal);
                this.DiagonalBorder.CopyPropertiesFrom(context, border.DiagonalDown);
                this.DiagonalDown = true;
            }

            if (!IsNullOrNone(border.DiagonalUp))
            {
                this.CreateElement(this.diagonal);
                this.DiagonalBorder.CopyPropertiesFrom(context, border.DiagonalUp);
                this.DiagonalUp = true;
            }
            
            // TODO: Refactor diagonal border.
            // TODO: Implement horizontal border?
            // TODO: Implement vertical border?
        }

        protected override void OnAfterRead(IXlsxWorkbookImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            BordersInfo bordersInfo = new BordersInfo();
            if (this.LeftBorder != null)
            {
                bordersInfo.Left = this.LeftBorder.CreateCellBorder(context);
                this.ReleaseElement(this.left);
            }

            if (this.TopBorder != null)
            {
                bordersInfo.Top = this.TopBorder.CreateCellBorder(context);
                this.ReleaseElement(this.top);
            }

            if (this.RightBorder != null)
            {
                bordersInfo.Right = this.RightBorder.CreateCellBorder(context);
                this.ReleaseElement(this.right);
            }

            if (this.BottomBorder != null)
            {
                bordersInfo.Bottom = this.BottomBorder.CreateCellBorder(context);
                this.ReleaseElement(this.bottom);
            }

            if (this.DiagonalBorder != null)
            {
                CellBorder diagonalBorder = this.DiagonalBorder.CreateCellBorder(context);
                if (this.DiagonalUp)
                {
                    bordersInfo.DiagonalUp = diagonalBorder;
                    this.ReleaseElement(this.diagonal);
                }

                if (this.DiagonalDown)
                {
                    bordersInfo.DiagonalDown = diagonalBorder;
                    this.ReleaseElement(this.diagonal);
                }
            }

            //// TODO: Refactor diagonal border.
            //// TODO: Implement horizontal border?
            //// TODO: Implement vertical border?

            context.StyleSheet.BordersInfoTable.Add(bordersInfo);
        }

        private static bool IsNullOrNone(CellBorder border)
        {
            return border == null || border.Style == CellBorderStyle.None;
        }
    }
}