﻿using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Contexts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Export;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Elements.Common;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Elements.SharedStrings;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Types;
#if (WPF || SILVERLIGHT) && !NETCORE
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Theming;
#else
using Telerik.Documents.Common.Model;
#endif
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Elements.StyleSheet
{
    internal class FontElement : StyleSheetElementBase
    {
        private readonly OpenXmlChildElement<BoldElement> bold;
        private readonly OpenXmlChildElement<FontFamilyElement> family;
        private readonly OpenXmlChildElement<ColorElement> color;
        private readonly OpenXmlChildElement<ItalicElement> italic;
        private readonly OpenXmlChildElement<FontNameElement> name;
        private readonly OpenXmlChildElement<SchemeElement> scheme;
        private readonly OpenXmlChildElement<FontSizeElement> fontSize;
        private readonly OpenXmlChildElement<UnderlineElement> underline;
        private readonly OpenXmlChildElement<VerticalAlignmentElement> vertAlign;
        private FontInfo? fontInfo;

        public FontElement(XlsxPartsManager partsManager)
            : base(partsManager)
        {
            this.bold = this.RegisterChildElement<BoldElement>(XlsxElementNames.BoldElementName);
            this.italic = this.RegisterChildElement<ItalicElement>(XlsxElementNames.ItalicElementName);
            this.underline = this.RegisterChildElement<UnderlineElement>(XlsxElementNames.UnderlineElementName);
            this.vertAlign = this.RegisterChildElement<VerticalAlignmentElement>(XlsxElementNames.VerticalAlignmentElementName);
            this.fontSize = this.RegisterChildElement<FontSizeElement>(XlsxElementNames.FontSizeElementName);
            this.color = this.RegisterChildElement<ColorElement>(XlsxElementNames.ColorElementName);
            this.name = this.RegisterChildElement<FontNameElement>(XlsxElementNames.FontNameElementName);
            this.family = this.RegisterChildElement<FontFamilyElement>(XlsxElementNames.FontFamilyElementName);
            this.scheme = this.RegisterChildElement<SchemeElement>(XlsxElementNames.SchemeElementName);
        }

        public override string ElementName
        {
            get
            {
                return XlsxElementNames.FontElementName;
            }
        }

        public FontInfo? FontInfo
        {
            get
            {
                return this.fontInfo;
            }
        }

        public FontNameElement NameElement
        {
            get
            {
                return this.name.Element;
            }
            set
            {
                this.name.Element = value;
            }
        }

        public FontFamilyElement FamilyElement
        {
            get
            {
                return this.family.Element;
            }
            set
            {
                this.family.Element = value;
            }
        }

        public BoldElement BoldElement
        {
            get
            {
                return this.bold.Element;
            }
            set
            {
                this.bold.Element = value;
            }
        }

        public ItalicElement ItalicElement
        {
            get
            {
                return this.italic.Element;
            }
            set
            {
                this.italic.Element = value;
            }
        }

        public ColorElement ColorElement
        {
            get
            {
                return this.color.Element;
            }
            set
            {
                this.color.Element = value;
            }
        }

        public FontSizeElement FontSizeElement
        {
            get
            {
                return this.fontSize.Element;
            }
            set
            {
                this.fontSize.Element = value;
            }
        }

        public UnderlineElement UnderlineElement
        {
            get
            {
                return this.underline.Element;
            }
            set
            {
                this.underline.Element = value;
            }
        }

        public VerticalAlignmentElement VertAlignElement
        {
            get
            {
                return this.vertAlign.Element;
            }
            set
            {
                this.vertAlign.Element = value;
            }
        }

        public SchemeElement SchemeElement
        {
            get
            {
                return this.scheme.Element;
            }
            set
            {
                this.scheme.Element = value;
            }
        }

        public void CopyPropertiesFrom(IXlsxWorkbookExportContext context, FontInfo font)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(font, "font");

            if (font.Bold.HasValue)
            {
                this.CreateElement(this.bold);
                this.BoldElement.Value = font.Bold.Value;
            }

            if (font.Italic.HasValue)
            {
                this.CreateElement(this.italic);
                this.ItalicElement.Value = font.Italic.Value;
            }

            if (font.ForeColor != null)
            {
                this.CreateElement(this.color);
                this.ColorElement.CopyPropertiesFrom(context, font.ForeColor);
            }

            if (font.FontSize.HasValue)
            {
                this.CreateElement(this.fontSize);
                this.FontSizeElement.Value = UnitHelper.DipToUnit(font.FontSize.Value, UnitType.Point);
            }

            if (font.FontFamily != null)
            {
                this.CreateElement(this.name);
                this.NameElement.Value = font.FontFamily.GetActualValue(context.Theme).Source;

                if (font.FontFamily.IsFromTheme)
                {
                    this.CreateElement(this.scheme);
                    this.SchemeElement.Value = FontSchemes.GetFontSchemeName(font.FontFamily.ThemeFontType);
                }
            }

            if (font.UnderlineType.HasValue)
            {
                this.CreateElement(this.underline);
                this.UnderlineElement.Value = UnderlineValues.GetUnderlineValueName(font.UnderlineType.Value);
            }

            // TODO: Implement vert align.
        }

        protected override void OnAfterRead(Import.IXlsxWorkbookImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            this.fontInfo = this.GetFontInfo(context);

            this.ReleaseElement(this.bold);
            this.ReleaseElement(this.italic);
            this.ReleaseElement(this.color);
            this.ReleaseElement(this.fontSize);
            this.ReleaseElement(this.scheme);
            this.ReleaseElement(this.name);
            this.ReleaseElement(this.underline);
        }

        protected override void ClearOverride()
        {
            this.fontInfo = null;
        }

        private FontInfo GetFontInfo(Import.IXlsxWorkbookImportContext context)
        {
            FontInfo fontInfo = new FontInfo();
            if (this.BoldElement != null)
            {
                fontInfo.Bold = this.BoldElement.Value;
            }

            if (this.ItalicElement != null)
            {
                fontInfo.Italic = this.ItalicElement.Value;
            }

            if (this.ColorElement != null)
            {
                fontInfo.ForeColor = this.ColorElement.CreateThemableColor(context);
            }

            if (this.FontSizeElement != null)
            {
                fontInfo.FontSize = UnitHelper.UnitToDip(this.FontSizeElement.Value, UnitType.Point);
            }

            if (this.SchemeElement != null)
            {
                ThemeFontType? themeFontType = FontSchemes.GetFontScheme(this.SchemeElement.Value);
                if (themeFontType != null)
                {
                    fontInfo.FontFamily = new ThemableFontFamily((ThemeFontType)themeFontType);
                }
            }
            else if (this.NameElement != null)
            {
                fontInfo.FontFamily = new ThemableFontFamily(this.NameElement.Value);
            }

            if (this.UnderlineElement != null)
            {
                fontInfo.UnderlineType = UnderlineValues.GetUnderlineValue(this.UnderlineElement.Value);
            }

            return fontInfo;
        }
    }
}