﻿using System;
using System.Linq;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Contexts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Export;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Import;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Elements.StyleSheet
{
    internal class FormatElement : StyleSheetElementBase
    {
        private readonly BoolOpenXmlAttribute applyAlignment;
        private readonly BoolOpenXmlAttribute applyBorder;
        private readonly BoolOpenXmlAttribute applyFill;
        private readonly BoolOpenXmlAttribute applyFont;
        private readonly BoolOpenXmlAttribute applyNumberFormat;
        private readonly BoolOpenXmlAttribute applyProtection;
        private readonly IntOpenXmlAttribute borderId;
        private readonly IntOpenXmlAttribute fillId;
        private readonly IntOpenXmlAttribute fontId;
        private readonly IntOpenXmlAttribute numberFormatId;
        private readonly IntOpenXmlAttribute formatId;

        private readonly OpenXmlChildElement<AlignmentElement> alignments;
        private readonly OpenXmlChildElement<ProtectionElement> protection;

        public FormatElement(XlsxPartsManager partsManager)
            : base(partsManager)
        {
            this.numberFormatId = this.RegisterAttribute(new IntOpenXmlAttribute("numFmtId"));
            this.fontId = this.RegisterAttribute(new IntOpenXmlAttribute("fontId"));
            this.fillId = this.RegisterAttribute(new IntOpenXmlAttribute("fillId"));
            this.borderId = this.RegisterAttribute(new IntOpenXmlAttribute("borderId"));
            this.formatId = this.RegisterAttribute(new IntOpenXmlAttribute("xfId"));
            this.applyNumberFormat = this.RegisterAttribute(new BoolOpenXmlAttribute("applyNumberFormat"));
            this.applyFont = this.RegisterAttribute(new BoolOpenXmlAttribute("applyFont"));
            this.applyFill = this.RegisterAttribute(new BoolOpenXmlAttribute("applyFill"));
            this.applyBorder = this.RegisterAttribute(new BoolOpenXmlAttribute("applyBorder"));
            this.applyAlignment = this.RegisterAttribute(new BoolOpenXmlAttribute("applyAlignment"));
            this.applyProtection = this.RegisterAttribute(new BoolOpenXmlAttribute("applyProtection"));

            this.alignments = this.RegisterChildElement<AlignmentElement>(XlsxElementNames.AlignmentElementName);
            this.protection = this.RegisterChildElement<ProtectionElement>(XlsxElementNames.ProtectionElementName);
        }

        public bool ApplyAlignment
        {
            get
            {
                return this.applyAlignment.Value;
            }
            set
            {
                this.applyAlignment.Value = value;
            }
        }

        public bool ApplyBorder
        {
            get
            {
                return this.applyBorder.Value;
            }
            set
            {
                this.applyBorder.Value = value;
            }
        }

        public bool ApplyFill
        {
            get
            {
                return this.applyFill.Value;
            }
            set
            {
                this.applyFill.Value = value;
            }
        }

        public bool ApplyFont
        {
            get
            {
                return this.applyFont.Value;
            }
            set
            {
                this.applyFont.Value = value;
            }
        }

        public bool ApplyNumberFormat
        {
            get
            {
                return this.applyNumberFormat.Value;
            }
            set
            {
                this.applyNumberFormat.Value = value;
            }
        }

        public bool ApplyProtection
        {
            get
            {
                return this.applyProtection.Value;
            }
            set
            {
                this.applyProtection.Value = value;
            }
        }

        public int BorderId
        {
            get
            {
                return this.borderId.Value;
            }
            set
            {
                this.borderId.Value = value;
            }
        }

        public int FillId
        {
            get
            {
                return this.fillId.Value;
            }
            set
            {
                this.fillId.Value = value;
            }
        }

        public int FontId
        {
            get
            {
                return this.fontId.Value;
            }
            set
            {
                this.fontId.Value = value;
            }
        }

        public int NumberFormatId
        {
            get
            {
                return this.numberFormatId.Value;
            }
            set
            {
                this.numberFormatId.Value = value;
            }
        }

        public int FormatId
        {
            get
            {
                return this.formatId.Value;
            }
            set
            {
                this.formatId.Value = value;
            }
        }

        public AlignmentElement AlignmentsElement
        {
            get
            {
                return this.alignments.Element;
            }
            set
            {
                this.alignments.Element = value;
            }
        }

        public ProtectionElement ProtectionElement
        {
            get
            {
                return this.protection.Element;
            }
            set
            {
                this.protection.Element = value;
            }
        }

        public override string ElementName
        {
            get
            {
                return XlsxElementNames.FormatElementName;
            }
        }

        public void CopyPropertiesFrom(IXlsxWorkbookExportContext context, FormattingRecord format)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(format, "format");

            this.CreateElement(this.alignments);
            this.CreateElement(this.protection);

            this.FormatId = 0;

            if (format.StyleFormattingRecordId.HasValue)
            {
                this.FormatId = format.StyleFormattingRecordId.Value;
            }

            FormattingRecord styleFormatting = context.StyleSheet.StyleFormattingTable[this.FormatId];
            this.CopyPropertiesFromStyleInternal(context, styleFormatting);

            this.CopyPropertiesFromDirectFormatInternal(context, format);
        }

        public FormattingRecord CreateFormattingRecord(IXlsxWorkbookImportContext context)
        {
            FormattingRecord formattingRecord = new FormattingRecord();
            if (this.formatId.HasValue)
            {
                formattingRecord.StyleFormattingRecordId = this.FormatId;
            }

            if (this.fontId.HasValue)
            {
                formattingRecord.FontInfoId = this.FontId;
            }

            if (this.borderId.HasValue)
            {
                formattingRecord.BordersInfoId = this.BorderId;
            }

            if (this.fillId.HasValue)
            {
                formattingRecord.FillId = this.FillId;
            }

            if (this.numberFormatId.HasValue)
            {
                formattingRecord.NumberFormatId = this.NumberFormatId;
            }

            if (this.applyNumberFormat.HasValue)
            {
                formattingRecord.ApplyNumberFormat = this.ApplyNumberFormat;
            }

            if (this.applyAlignment.HasValue)
            {
                formattingRecord.ApplyAlignment = this.ApplyAlignment;
            }

            if (this.applyFont.HasValue)
            {
                formattingRecord.ApplyFont = this.ApplyFont;
            }

            if (this.applyBorder.HasValue)
            {
                formattingRecord.ApplyBorder = this.ApplyBorder;
            }

            if (this.applyFill.HasValue)
            {
                formattingRecord.ApplyFill = this.ApplyFill;
            }

            if (this.applyProtection.HasValue)
            {
                formattingRecord.ApplyProtection = this.ApplyProtection;
            }

            if (this.AlignmentsElement != null)
            {
                this.AlignmentsElement.ApplyAllignment(context, ref formattingRecord);
                this.ReleaseElement(this.alignments);
            }

            if (this.ProtectionElement != null)
            {
                this.ProtectionElement.ApplyProtection(context, ref formattingRecord);
                this.ReleaseElement(this.protection);
            }

            return formattingRecord;
        }

        private void CopyPropertiesFromDirectFormatInternal(IXlsxWorkbookExportContext context, FormattingRecord format)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(format, "format");

            if (format.FontInfoId.HasValue)
            {
                this.FontId = format.FontInfoId.Value;
                this.ApplyFont = true;
            }

            if (format.BordersInfoId.HasValue)
            {
                this.BorderId = format.BordersInfoId.Value;
                this.ApplyBorder = true;
            }

            if (format.FillId.HasValue)
            {
                this.FillId = format.FillId.Value;
                this.ApplyFill = true;
            }

            if (format.NumberFormatId.HasValue)
            {
                this.NumberFormatId = format.NumberFormatId.Value;
                this.ApplyNumberFormat = true;
            }

            this.AlignmentsElement.CopyPropertiesFrom(context, format);
            this.ProtectionElement.CopyPropertiesFrom(context, format);

            if (format.ApplyNumberFormat.HasValue)
            {
                this.ApplyNumberFormat = format.ApplyNumberFormat.Value;
            }

            if (format.ApplyAlignment.HasValue)
            {
                this.ApplyAlignment = format.ApplyAlignment.Value;
            }

            if (format.ApplyFont.HasValue)
            {
                this.ApplyFont = format.ApplyFont.Value;
            }

            if (format.ApplyBorder.HasValue)
            {
                this.ApplyBorder = format.ApplyBorder.Value;
            }

            if (format.ApplyFill.HasValue)
            {
                this.ApplyFill = format.ApplyFill.Value;
            }

            if (format.ApplyProtection.HasValue)
            {
                this.ApplyProtection = format.ApplyProtection.Value;
            }
        }

        private void CopyPropertiesFromStyleInternal(IXlsxWorkbookExportContext context, FormattingRecord format)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(format, "format");

            if (format.FontInfoId.HasValue && format.ApplyFont == true)
            {
                this.FontId = format.FontInfoId.Value;
            }

            if (format.BordersInfoId.HasValue && format.ApplyBorder == true)
            {
                this.BorderId = format.BordersInfoId.Value;
            }

            if (format.FillId.HasValue && format.ApplyFill == true)
            {
                this.FillId = format.FillId.Value;
            }

            if (format.NumberFormatId.HasValue && format.ApplyNumberFormat == true)
            {
                this.NumberFormatId = format.NumberFormatId.Value;
            }

            if (format.ApplyAlignment == true)
            {
                this.AlignmentsElement.CopyPropertiesFrom(context, format);
            }

            if (format.ApplyProtection == true)
            {
                this.ProtectionElement.CopyPropertiesFrom(context, format);
            }
        }
    }
}