﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Elements;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Contexts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Export;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Import;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Types;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Elements.StyleSheet
{
    internal class GradientFillElement : StyleSheetElementBase
    {
        private readonly OpenXmlAttribute<double> bottom;
        private readonly OpenXmlAttribute<double> degree;
        private readonly OpenXmlAttribute<double> left;
        private readonly OpenXmlAttribute<double> right;
        private readonly OpenXmlAttribute<double> top;
        private readonly OpenXmlAttribute<string> type;

        private List<GradientStop> gradientStops;

        public GradientFillElement(XlsxPartsManager partsManager)
            : base(partsManager)
        {
            this.type = this.RegisterAttribute<string>("type", GradientTypes.Linear);
            this.degree = this.RegisterAttribute<double>("degree", 0);
            this.left = this.RegisterAttribute<double>("left", 0);
            this.right = this.RegisterAttribute<double>("right", 0);
            this.top = this.RegisterAttribute<double>("top", 0);
            this.bottom = this.RegisterAttribute<double>("bottom", 0);

            this.gradientStops = new List<GradientStop>();
        }

        public string Type
        {
            get
            {
                return this.type.Value;
            }
            set
            {
                this.type.Value = value;
            }
        }

        public double Bottom
        {
            get
            {
                return this.bottom.Value;
            }
            set
            {
                this.bottom.Value = value;
            }
        }

        public double Degree
        {
            get
            {
                return this.degree.Value;
            }
            set
            {
                this.degree.Value = value;
            }
        }

        public double Left
        {
            get
            {
                return this.left.Value;
            }
            set
            {
                this.left.Value = value;
            }
        }

        public double Right
        {
            get
            {
                return this.right.Value;
            }
            set
            {
                this.right.Value = value;
            }
        }

        public double Top
        {
            get
            {
                return this.top.Value;
            }
            set
            {
                this.top.Value = value;
            }
        }

        public override string ElementName
        {
            get
            {
                return XlsxElementNames.GradientFillElementName;
            }
        }

        public void CopyPropertiesFrom(IXlsxWorkbookExportContext context, GradientFill gradientFill)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(gradientFill, "gradientFill");

            GradientFillInfo gradientFillInfo = GradientFillInfo.Create(gradientFill);
            
            if (gradientFillInfo.InfoType.HasValue)
            {
                this.Type = GradientTypes.GetGradientTypeName(gradientFillInfo.InfoType.Value);
            }

            if (gradientFillInfo.Degree.HasValue)
            {
                this.Degree = gradientFillInfo.Degree.Value;
            }

            if (gradientFillInfo.Left.HasValue)
            {
                this.Left = gradientFillInfo.Left.Value;
            }

            if (gradientFillInfo.Right.HasValue)
            {
                this.Right = gradientFillInfo.Right.Value;
            }

            if (gradientFillInfo.Top.HasValue)
            {
                this.Top = gradientFillInfo.Top.Value;
            }

            if (gradientFillInfo.Bottom.HasValue)
            {
                this.Bottom = gradientFillInfo.Bottom.Value;
            }

            this.gradientStops = gradientFillInfo.Stops;
        }

        public IFill CreateFill(IXlsxWorkbookImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            GradientFillInfo gradientFillInfo = new GradientFillInfo();

            if (this.type.HasValue)
            {
                gradientFillInfo.InfoType = GradientTypes.GetGradientType(this.Type);
            }

            if (this.degree.HasValue)
            {
                gradientFillInfo.Degree = this.Degree;
            }

            if (this.left.HasValue)
            {
                gradientFillInfo.Left = this.Left;
            }

            if (this.right.HasValue)
            {
                gradientFillInfo.Right = this.Right;
            }

            if (this.top.HasValue)
            {
                gradientFillInfo.Top = this.Top;
            }

            if (this.bottom.HasValue)
            {
                gradientFillInfo.Bottom = this.Bottom;
            }

            gradientFillInfo.Stops = new List<GradientStop>(this.gradientStops);
            
            this.gradientStops.Clear();

            return gradientFillInfo.ToFill();
        }

        protected override bool ShouldExport(IXlsxWorkbookExportContext context)
        {
            return this.gradientStops.Count > 0 || base.ShouldExport(context);
        }

        protected override IEnumerable<OpenXmlElementBase> EnumerateChildElements(IXlsxWorkbookExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            foreach (GradientStop stop in this.gradientStops)
            {
                StopElement stopElement = this.CreateElement<StopElement>(XlsxElementNames.StopElementName);
                stopElement.CopyPropertiesFrom(context, stop);

                yield return stopElement;
            }
        }

        protected override void OnAfterReadChildElement(IXlsxWorkbookImportContext context, OpenXmlElementBase element)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(element, "element");

            this.gradientStops.Add(((StopElement)element).ToGradientStop(context));
        }
    }
}