﻿using System;
using System.Linq;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.Contexts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Export;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Import;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Elements.Workbooks
{
    internal class WorkbookProtectionElement : WorkbookElementBase
    {
        private readonly OpenXmlAttribute<string> workbookAlgorithmName;
        private readonly OpenXmlAttribute<string> workbookHashValue;
        private readonly OpenXmlAttribute<string> workbookSaltValue;
        private readonly IntOpenXmlAttribute workbookSpinCount;
        private readonly OpenXmlAttribute<string> password;

        private readonly BoolOpenXmlAttribute lockStructure;
        private readonly BoolOpenXmlAttribute lockWindows;

        public WorkbookProtectionElement(XlsxPartsManager partsManager) 
            : base(partsManager)
        {
            this.workbookAlgorithmName = this.RegisterAttribute<string>("workbookAlgorithmName", string.Empty);
            this.workbookHashValue = this.RegisterAttribute<string>("workbookHashValue", string.Empty);
            this.workbookSaltValue = this.RegisterAttribute<string>("workbookSaltValue", string.Empty);
            this.workbookSpinCount = this.RegisterAttribute(new IntOpenXmlAttribute("workbookSpinCount"));
            this.password = this.RegisterAttribute<string>("workbookPassword", string.Empty);

            this.lockStructure = this.RegisterAttribute(new BoolOpenXmlAttribute("lockStructure"));
            this.lockWindows = this.RegisterAttribute(new BoolOpenXmlAttribute("lockWindows"));
        }

        public override string ElementName
        {
            get
            {
                return XlsxElementNames.WorkbookProtectionElementName;
            }
        }

        public bool LockStructure
        {
            get
            {
                return this.lockStructure.Value;
            }
            set
            {
                this.lockStructure.Value = value;
            }
        }

        public bool LockWindows
        {
            get
            {
                return this.lockWindows.Value;
            }
            set
            {
                this.lockWindows.Value = value;
            }
        }

        public string WorkbookAlgorithmName
        {
            get
            {
                return this.workbookAlgorithmName.Value;
            }
            set
            {
                this.workbookAlgorithmName.Value = value;
            }
        }

        public string WorkbookHashValue
        {
            get
            {
                return this.workbookHashValue.Value;
            }
            set
            {
                this.workbookHashValue.Value = value;
            }
        }

        public string WorkbookSaltValue
        {
            get
            {
                return this.workbookSaltValue.Value;
            }
            set
            {
                this.workbookSaltValue.Value = value;
            }
        }

        public int WorkbookSpinCount
        {
            get
            {
                return this.workbookSpinCount.Value;
            }
            set
            {
                this.workbookSpinCount.Value = value;
            }
        }

        public string Password
        {
            get
            {
                return this.password.Value;
            }
            set
            {
                this.password.Value = value;
            }
        }

        protected override bool ShouldExport(IXlsxWorkbookExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            return context.GetWorkbookProtectionInfo().Enforced;
        }

        protected override void OnBeforeWrite(IXlsxWorkbookExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            WorkbookProtectionInfo info = context.GetWorkbookProtectionInfo();

            if (!string.IsNullOrEmpty(info.Password))
            {
                this.Password = info.Password;
            }
            else if (!string.IsNullOrEmpty(info.HashValue))
            {
                this.WorkbookAlgorithmName = info.AlgorithmName;
                this.WorkbookHashValue = info.HashValue;
                this.WorkbookSpinCount = info.SpinCount;

                if (!string.IsNullOrEmpty(info.SaltValue))
                {
                    this.WorkbookSaltValue = info.SaltValue;
                }
            }

            if (this.lockStructure.DefaultValue != info.LockStructure)
            {
                this.LockStructure = info.LockStructure;
            }

            if (this.lockWindows.DefaultValue != info.LockWindows)
            {
                this.LockWindows = info.LockWindows;
            }
        }

        protected override void OnAfterRead(IXlsxWorkbookImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            WorkbookProtectionInfo workbookProtection = new WorkbookProtectionInfo();

            if (string.IsNullOrEmpty(this.Password))
            {
                workbookProtection.AlgorithmName = this.WorkbookAlgorithmName;
                workbookProtection.HashValue = this.WorkbookHashValue;
                workbookProtection.SaltValue = this.WorkbookSaltValue;
                workbookProtection.SpinCount = this.WorkbookSpinCount;
            }
            else
            {
                workbookProtection.Password = this.Password;
            }

            workbookProtection.LockStructure = this.LockStructure;
            workbookProtection.LockWindows = this.LockWindows;
            workbookProtection.Enforced = workbookProtection.LockStructure || workbookProtection.LockWindows;

            context.ApplyWorkbookProtectionInfo(workbookProtection);
        }
    }
}
