﻿using System;
using System.Linq;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.Contexts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Export;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Import;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Types;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Elements.Worksheets
{
    internal class CellElement : WorksheetElementBase
    {
        private readonly ConvertedOpenXmlAttribute<CellRef> reference;
        private readonly IntOpenXmlAttribute styleIndex;
        private readonly OpenXmlAttribute<string> cellDataType;

        private readonly OpenXmlChildElement<FormulaElement> formula;
        private readonly OpenXmlChildElement<CellValueElement> cellValue;
        private readonly OpenXmlChildElement<RichTextInlineElement> richTextInline;

        public CellElement(XlsxPartsManager partsManager)
            : base(partsManager)
        {
            this.reference = this.RegisterAttribute(new ConvertedOpenXmlAttribute<CellRef>("r", XlsxConverters.CellRefConverter));
            this.styleIndex = this.RegisterAttribute(new IntOpenXmlAttribute("s", 0));
            this.cellDataType = this.RegisterAttribute<string>("t", CellTypes.Number);

            this.formula = this.RegisterChildElement<FormulaElement>(XlsxElementNames.FormulaElementName, XlsxElementUniquePoolIds.FormulaElementUniquePoolId);
            this.cellValue = this.RegisterChildElement<CellValueElement>(XlsxElementNames.CellValueElementName, XlsxElementUniquePoolIds.CellValueElementUniquePoolId);
            this.richTextInline = this.RegisterChildElement<RichTextInlineElement>(XlsxElementNames.RichTextInlineElementName);
        }

        public CellRef Reference
        {
            get
            {
                return this.reference.Value;
            }
            set
            {
                this.reference.Value = value;
            }
        }

        public int StyleIndex
        {
            get
            {
                return this.styleIndex.Value;
            }
            set
            {
                this.styleIndex.Value = value;
            }
        }

        public string CellDataType
        {
            get
            {
                return this.cellDataType.Value;
            }
            set
            {
                this.cellDataType.Value = value;
            }
        }

        public override string ElementName
        {
            get
            {
                return XlsxElementNames.CellElementName;
            }
        }

        public FormulaElement FormulaElement
        {
            get
            {
                return this.formula.Element;
            }
            set
            {
                this.formula.Element = value;
            }
        }

        public CellValueElement CellValueElement
        {
            get
            {
                return this.cellValue.Element;
            }
            set
            {
                this.cellValue.Element = value;
            }
        }

        public RichTextInlineElement RichTextInlineElement
        {
            get
            {
                return this.richTextInline.Element;
            }
            set
            {
                this.richTextInline.Element = value;
            }
        }

        public void CopyPropertiesFrom(IXlsxWorksheetExportContext context, CellInfo cell)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(cell, "cell");

            this.Reference = new CellRef(cell.RowIndex, cell.ColumnIndex);

            if (cell.CellValue != null && cell.CellValue.ValueType != CellValueType.Empty)
            {
                this.CellDataType = GetCellDataType(cell.CellValue);
                this.CreateElement(this.cellValue);
                this.CellValueElement.InnerText = GetValueAsString(context, cell.CellValue);

                if (cell.CellValue.ValueType == CellValueType.Formula)
                {
                    this.CreateElement(this.formula);
                    this.FormulaElement.CopyPropertiesFrom(context, (FormulaCellValue)cell.CellValue);
                }
            }

            var longCellIndex = CellsPropertyBag.ConvertCellIndexToLong(cell.RowIndex, cell.ColumnIndex);
            bool containsNonDefaultFormattingRecordIndex = context.ContainsNonDefaultFormattingRecordIndex(context.Worksheet.Cells, longCellIndex);
            if (containsNonDefaultFormattingRecordIndex)
            {
                int? styleIndex = context.GetFormattingRecordIndex(context.Worksheet.Cells, longCellIndex);
                if (styleIndex.HasValue)
                {
                    this.StyleIndex = styleIndex.Value;
                }
            }
        }

        protected override void OnAfterRead(IXlsxWorksheetImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            CellInfo cellInfo = new CellInfo();

            if (this.Reference != null)
            {
                cellInfo.RowIndex = this.Reference.RowIndex;
                cellInfo.ColumnIndex = this.Reference.ColumnIndex;
            }

            cellInfo.StyleIndex = this.StyleIndex;

            ICellValue cellValue = null;
            if (this.FormulaElement != null)
            {
                if (!this.FormulaElement.CopyPropertiesTo(context, cellInfo.RowIndex, cellInfo.ColumnIndex, out cellValue) && this.CellValueElement != null)
                {
                    cellValue = new TextCellValue(this.CellValueElement.InnerText);
                }
            }
            else if (this.CellValueElement != null)
            {
                if (this.CellDataType == CellTypes.SharedString)
                {
                    string value = context.GetSharedStringValue(Convert.ToInt32(this.CellValueElement.InnerText));
                    if (value == null)
                    {
                        cellValue = EmptyCellValue.EmptyValue;
                    }
                    else
                    {
                        cellValue = new TextCellValue(value);
                    }
                }
                else if (this.CellDataType == CellTypes.Boolean)
                {
                    byte numberCellValue;
                    bool successParseToNum = byte.TryParse(this.CellValueElement.InnerText, out numberCellValue);

                    if (successParseToNum)
                    {
                        cellValue = new BooleanCellValue(Convert.ToBoolean(numberCellValue));
                    }
                    else
                    {
                        cellValue = new BooleanCellValue(Convert.ToBoolean(this.CellValueElement.InnerText));
                    }
                }
                else if (this.CellDataType == CellTypes.Error)
                {
                    cellValue = this.CellValueElement.InnerText.ToCellValue(context.Worksheet, cellInfo.RowIndex, cellInfo.ColumnIndex);
                }
                else if (this.CellDataType == CellTypes.Number)
                {
                    double res;
                    bool parsed = XlsxHelper.CultureInfo.TryParseDouble(this.CellValueElement.InnerText, out res) ||
                        XlsxHelper.CultureInfo.TryParseScientific(this.CellValueElement.InnerText, out res);

                    if (parsed)
                    {
                        cellValue = new NumberCellValue(res);
                    }
                }
                else if (this.CellDataType == CellTypes.String)
                {
                    cellValue = new TextCellValue(this.CellValueElement.InnerText);
                }
                else if (this.CellDataType == CellTypes.Date)
                {
                    DateTime cellValueDate = DateTime.Parse(this.CellValueElement.InnerText);
                    double numberValue = FormatHelper.ConvertDateTimeToDouble(cellValueDate);
                    cellValue = new NumberCellValue(numberValue);
                }
            }
            else if (this.RichTextInlineElement != null && this.CellDataType == CellTypes.InlineString)
            {
                cellValue = new TextCellValue(this.RichTextInlineElement.ResultText);
            }

            this.ReleaseElement(this.formula);
            this.ReleaseElement(this.cellValue);
            this.ReleaseElement(this.richTextInline);
            cellInfo.CellValue = cellValue;

            context.ApplyCellInfo(cellInfo);
        }

        private static string GetFormulaCellValueType(FormulaCellValue formula)
        {
            Guard.ThrowExceptionIfNull(formula, "formula");

            ConstantExpression constantExpressoin = formula.Value.GetValue().GetValueAsConstantExpression();

            if (constantExpressoin is StringExpression)
            {
                return CellTypes.String;
            }

            if (constantExpressoin is NumberExpression)
            {
                return CellTypes.Number;
            }

            if (constantExpressoin is ErrorExpression)
            {
                return CellTypes.Error;
            }

            if (constantExpressoin is BooleanExpression)
            {
                return CellTypes.Boolean;
            }

            return CellTypes.Number;
        }

        private static string GetFormulaValueAsString(FormulaCellValue formula)
        {
            Guard.ThrowExceptionIfNull(formula, "formula");

            ConstantExpression constantExpression = formula.Value.GetValue().GetValueAsConstantExpression();

            BooleanExpression booleanExpression = constantExpression as BooleanExpression;
            if (booleanExpression != null)
            {
                return booleanExpression.Value.ToValueString();
            }

            return constantExpression.GetValueAsString(XlsxHelper.CultureInfo);
        }

        private static string GetCellDataType(ICellValue value)
        {
            Guard.ThrowExceptionIfNull(value, "value");

            switch (value.ValueType)
            {
                case CellValueType.Boolean:
                case CellValueType.Error:
                case CellValueType.Number:
                case CellValueType.Text:
                    return CellTypes.GetCellTypeName(value.ValueType);
                case CellValueType.Formula:
                    return GetFormulaCellValueType((FormulaCellValue)value);
                default:
                    throw new NotSupportedException("Value type is not supported.");
            }
        }

        private static string GetValueAsString(IXlsxWorksheetExportContext context, ICellValue value)
        {
            switch (value.ValueType)
            {
                case CellValueType.Boolean:
                    BooleanCellValue boolValue = (BooleanCellValue)value;

                    return boolValue.Value.ToValueString();

                case CellValueType.Error:

                    return value.RawValue;

                case CellValueType.Number:

                    return XlsxHelper.CultureInfo.ToString(((NumberCellValue)value).Value);

                case CellValueType.Formula:

                    return GetFormulaValueAsString((FormulaCellValue)value);

                case CellValueType.Text:

                    return context.GetSharedStringIndex((TextCellValue)value).ToString();

                default:
                    throw new NotSupportedException("Value type is not supported.");
            }
        }
    }
}