﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Contexts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Export;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Import;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Types;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.DataValidation;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Elements.Worksheets
{
    internal class DataValidationElement : WorksheetElementBase
    {
        private readonly IntOpenXmlAttribute allowBlank;
        private readonly OpenXmlAttribute<string> error;
        private readonly ConvertedOpenXmlAttribute<ErrorStyle> errorStyle;
        private readonly OpenXmlAttribute<string> errorTitle;
        private readonly ConvertedOpenXmlAttribute<ComparisonOperator> comparisonOperator;
        private readonly OpenXmlAttribute<string> prompt;
        private readonly OpenXmlAttribute<string> promptTitle;
        private readonly IntOpenXmlAttribute showDropDown;
        private readonly IntOpenXmlAttribute showErrorMessage;
        private readonly IntOpenXmlAttribute showInputMessage;
        private readonly ConvertedOpenXmlAttribute<CellRefRangeSequence> sqref;
        private readonly ConvertedOpenXmlAttribute<DataValidationRuleType> type;

        private readonly OpenXmlChildElement<Formula1Element> formula1Element;
        private readonly OpenXmlChildElement<Formula2Element> formula2Element;

        private readonly OpenXmlChildElement<SqRefElement> sqrefElement;

        private readonly List<Action<IXlsxWorksheetExportContext, IDataValidationRule, CellIndex, IEnumerable<CellRange>>> copyActions;
        private readonly Dictionary<Type, DataValidationRuleType> typeToDataValidationRuleType;
        private readonly Dictionary<DataValidationRuleType, Func<IXlsxWorksheetImportContext, IDataValidationRule>> ruleTypeToFactory;

        public DataValidationElement(XlsxPartsManager partsManager)
            : base(partsManager)
        {
            this.type = this.RegisterAttribute(new ConvertedOpenXmlAttribute<DataValidationRuleType>("type", XlsxConverters.DataValidationRuleTypeConverter, DataValidationRuleType.None));
            this.errorStyle = this.RegisterAttribute(new ConvertedOpenXmlAttribute<ErrorStyle>("errorStyle", XlsxConverters.ErrorStyleConverter, ErrorStyle.Stop));
            this.comparisonOperator = this.RegisterAttribute(new ConvertedOpenXmlAttribute<ComparisonOperator>("operator", XlsxConverters.ComparisonOperatorConverter, ComparisonOperator.Between));
            this.allowBlank = this.RegisterAttribute(new IntOpenXmlAttribute("allowBlank", 0));
            this.showDropDown = this.RegisterAttribute(new IntOpenXmlAttribute("showDropDown", 0));
            this.showInputMessage = this.RegisterAttribute(new IntOpenXmlAttribute("showInputMessage", 0));
            this.showErrorMessage = this.RegisterAttribute(new IntOpenXmlAttribute("showErrorMessage", 0));
            this.errorTitle = this.RegisterAttribute(new OpenXmlAttribute<string>("errorTitle", string.Empty));
            this.error = this.RegisterAttribute(new OpenXmlAttribute<string>("error", string.Empty));
            this.promptTitle = this.RegisterAttribute(new OpenXmlAttribute<string>("promptTitle", string.Empty));
            this.prompt = this.RegisterAttribute(new OpenXmlAttribute<string>("prompt", string.Empty));
            this.sqref = this.RegisterAttribute(new ConvertedOpenXmlAttribute<CellRefRangeSequence>("sqref", XlsxConverters.CellRefRangeSequenceConverter, true));

            this.formula1Element = this.RegisterChildElement<Formula1Element>(XlsxElementNames.Formula1ElementName);
            this.formula2Element = this.RegisterChildElement<Formula2Element>(XlsxElementNames.Formula2ElementName);

            this.sqrefElement = this.RegisterChildElement<SqRefElement>(XlsxElementNames.SqRefElementName);

            this.copyActions = new List<Action<IXlsxWorksheetExportContext, IDataValidationRule, CellIndex, IEnumerable<CellRange>>>();
            this.InitializeCopyActions();

            this.typeToDataValidationRuleType = new Dictionary<Type, DataValidationRuleType>();
            this.InitializeTypeToDataValidationRuleType();

            this.ruleTypeToFactory = new Dictionary<DataValidationRuleType, Func<IXlsxWorksheetImportContext, IDataValidationRule>>();
            this.InitializeRuleTypeToFactory();
        }

        public override string ElementName
        {
            get
            {
                return XlsxElementNames.DataValidationElementName;
            }
        }

        public bool AllowBlank
        {
            get
            {
                return (this.allowBlank.Value == 0) ? false : true;
            }
            set
            {
                this.allowBlank.Value = value ? 1 : 0;
            }
        }

        public string Error
        {
            get
            {
                return this.error.Value;
            }
            set
            {
                this.error.Value = value;
            }
        }

        public ErrorStyle ErrorStyle
        {
            get
            {
                return this.errorStyle.Value;
            }
            set
            {
                this.errorStyle.Value = value;
            }
        }

        public string ErrorTitle
        {
            get
            {
                return this.errorTitle.Value;
            }
            set
            {
                this.errorTitle.Value = value;
            }
        }

        public ComparisonOperator ComparisonOperator
        {
            get
            {
                return this.comparisonOperator.Value;
            }
            set
            {
                this.comparisonOperator.Value = value;
            }
        }

        public string Prompt
        {
            get
            {
                return this.prompt.Value;
            }
            set
            {
                this.prompt.Value = value;
            }
        }

        public string PromptTitle
        {
            get
            {
                return this.promptTitle.Value;
            }
            set
            {
                this.promptTitle.Value = value;
            }
        }

        public bool ShowDropDown
        {
            get
            {
                return (this.showDropDown.Value == 0) ? true : false;
            }
            set
            {
                this.showDropDown.Value = value ? 0 : 1;
            }
        }

        public bool ShowErrorMessage
        {
            get
            {
                return (this.showErrorMessage.Value == 0) ? false : true;
            }
            set
            {
                this.showErrorMessage.Value = value ? 1 : 0;
            }
        }

        public bool ShowInputMessage
        {
            get
            {
                return (this.showInputMessage.Value == 0) ? false : true;
            }
            set
            {
                this.showInputMessage.Value = value ? 1 : 0;
            }
        }

        public CellRefRangeSequence SqRef
        {
            get
            {
                CellRefRangeSequence reference = this.sqref.Value;
                if (reference == null)
                {
                    reference = new CellRefRangeSequence(this.sqrefElement.Element.InnerText);
                }

                return reference;
            }
            set
            {
                this.sqref.Value = value;
            }
        }

        public DataValidationRuleType Type
        {
            get
            {
                return this.type.Value;
            }
            set
            {
                this.type.Value = value;
            }
        }

        public Formula1Element Formula1Element
        {
            get
            {
                return this.formula1Element.Element;
            }
            set
            {
                this.formula1Element.Element = value;
            }
        }

        public Formula2Element Formula2Element
        {
            get
            {
                return this.formula2Element.Element;
            }
            set
            {
                this.formula2Element.Element = value;
            }
        }

        public void CopyPropertiesFrom(IXlsxWorksheetExportContext context, IDataValidationRule rule)
        {
            this.CreateElement(this.formula1Element);
            this.CreateElement(this.formula2Element);

            IEnumerable<CellRange> ruleCellRanges = context.GetDataValidationRuleCellRanges(rule);

            CellIndex topLeftCellIndex = null;
            foreach (CellRange range in ruleCellRanges)
            {
                if (topLeftCellIndex == null)
                {
                    topLeftCellIndex = range.FromIndex;
                }
                else if (range.FromIndex.ColumnIndex <= topLeftCellIndex.ColumnIndex ||
                        range.FromIndex.RowIndex <= topLeftCellIndex.RowIndex)
                {
                    topLeftCellIndex = range.FromIndex;
                }
            }

            ruleCellRanges = from p in ruleCellRanges orderby p.FromIndex == topLeftCellIndex select p;

            foreach (var action in this.copyActions)
            {
                action(context, rule, topLeftCellIndex, ruleCellRanges);
            }
        }

        protected override void OnAfterRead(IXlsxWorksheetImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            IDataValidationRule rule = this.ruleTypeToFactory[this.Type](context);

            foreach (var range in this.SqRef.CellRanges)
            {
                context.Worksheet.Cells[range].SetDataValidationRule(rule);
            }

            this.ReleaseElement(this.formula1Element);
            this.ReleaseElement(this.formula2Element);
        }

        private void InitializeRuleTypeToFactory()
        {
            this.ruleTypeToFactory.Add(DataValidationRuleType.Custom, this.CustomRuleFactory);
            this.ruleTypeToFactory.Add(DataValidationRuleType.Date, this.DateRuleFactory);
            this.ruleTypeToFactory.Add(DataValidationRuleType.Decimal, this.DecimalRuleFactory);
            this.ruleTypeToFactory.Add(DataValidationRuleType.List, this.ListRuleFactory);
            this.ruleTypeToFactory.Add(DataValidationRuleType.None, this.AnyRuleFactory);
            this.ruleTypeToFactory.Add(DataValidationRuleType.TextLength, this.TextLengthRuleFactory);
            this.ruleTypeToFactory.Add(DataValidationRuleType.Time, this.TimeRuleFactory);
            this.ruleTypeToFactory.Add(DataValidationRuleType.Whole, this.WholeRuleFactory);
        }

        private IDataValidationRule CustomRuleFactory(IXlsxWorksheetImportContext context)
        {
            SingleArgumentDataValidationRuleContext dataValidationRuleContext = new SingleArgumentDataValidationRuleContext(context.Worksheet, this.SqRef.CellRanges.First().FromIndex);
            this.CopyPropertiesToDataValidationRuleContext(dataValidationRuleContext);
            CustomDataValidationRule rule = new CustomDataValidationRule(dataValidationRuleContext);

            return rule;
        }

        private IDataValidationRule DateRuleFactory(IXlsxWorksheetImportContext context)
        {
            NumberDataValidationRuleContext dataValidationRuleContext = new NumberDataValidationRuleContext(context.Worksheet, this.SqRef.CellRanges.First().FromIndex);
            this.CopyPropertiesToDataValidationRuleContext(dataValidationRuleContext);
            DateDataValidationRule rule = new DateDataValidationRule(dataValidationRuleContext);

            return rule;
        }

        private IDataValidationRule DecimalRuleFactory(IXlsxWorksheetImportContext context)
        {
            NumberDataValidationRuleContext dataValidationRuleContext = new NumberDataValidationRuleContext(context.Worksheet, this.SqRef.CellRanges.First().FromIndex);
            this.CopyPropertiesToDataValidationRuleContext(dataValidationRuleContext);
            DecimalDataValidationRule rule = new DecimalDataValidationRule(dataValidationRuleContext);

            return rule;
        }

        private IDataValidationRule ListRuleFactory(IXlsxWorksheetImportContext context)
        {
            ListDataValidationRuleContext dataValidationRuleContext = new ListDataValidationRuleContext(context.Worksheet, this.SqRef.CellRanges.First().FromIndex);
            this.CopyPropertiesToDataValidationRuleContext(dataValidationRuleContext);
            ListDataValidationRule rule = new ListDataValidationRule(dataValidationRuleContext);

            return rule;
        }

        private IDataValidationRule AnyRuleFactory(IXlsxWorksheetImportContext arg)
        {
            AnyValueDataValidationRuleContext dataValidationRuleContext = new AnyValueDataValidationRuleContext();
            this.CopyPropertiesToDataValidationRuleContext(dataValidationRuleContext);
            AnyValueDataValidationRule rule = new AnyValueDataValidationRule(dataValidationRuleContext);

            return rule;
        }

        private IDataValidationRule TextLengthRuleFactory(IXlsxWorksheetImportContext context)
        {
            NumberDataValidationRuleContext dataValidationRuleContext = new NumberDataValidationRuleContext(context.Worksheet, this.SqRef.CellRanges.First().FromIndex);
            this.CopyPropertiesToDataValidationRuleContext(dataValidationRuleContext);
            TextLengthDataValidationRule rule = new TextLengthDataValidationRule(dataValidationRuleContext);

            return rule;
        }

        private IDataValidationRule TimeRuleFactory(IXlsxWorksheetImportContext context)
        {
            NumberDataValidationRuleContext dataValidationRuleContext = new NumberDataValidationRuleContext(context.Worksheet, this.SqRef.CellRanges.First().FromIndex);
            this.CopyPropertiesToDataValidationRuleContext(dataValidationRuleContext);
            TimeDataValidationRule rule = new TimeDataValidationRule(dataValidationRuleContext);

            return rule;
        }

        private IDataValidationRule WholeRuleFactory(IXlsxWorksheetImportContext context)
        {
            NumberDataValidationRuleContext dataValidationRuleContext = new NumberDataValidationRuleContext(context.Worksheet, this.SqRef.CellRanges.First().FromIndex);
            this.CopyPropertiesToDataValidationRuleContext(dataValidationRuleContext);
            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(dataValidationRuleContext);

            return rule;
        }

        private void CopyPropertiesToDataValidationRuleContext(DataValidationRuleContextBase context)
        {
            context.ErrorAlertContent = this.Error;
            context.ErrorAlertTitle = this.ErrorTitle;
            context.ErrorStyle = this.ErrorStyle;
            context.InputMessageContent = this.Prompt;
            context.InputMessageTitle = this.PromptTitle;
            context.ShowErrorMessage = this.ShowErrorMessage;
            context.ShowInputMessage = this.ShowInputMessage;

            SingleArgumentDataValidationRuleContext singleArgumentContext = context as SingleArgumentDataValidationRuleContext;
            if (singleArgumentContext != null)
            {
                this.CopyPropertiesToSingleArgumentContext(singleArgumentContext);
            }

            ListDataValidationRuleContext listContext = context as ListDataValidationRuleContext;
            if (listContext != null)
            {
                this.CopyPropertiesToListDataValidationRuleContext(listContext);
            }

            NumberDataValidationRuleContext numberContext = context as NumberDataValidationRuleContext;
            if (numberContext != null)
            {
                this.CopyPropertiesToNumberDataValidationRuleContext(numberContext);
            }
        }

        private void CopyPropertiesToSingleArgumentContext(SingleArgumentDataValidationRuleContext context)
        {
            string formulaText = null;

            if (this.Formula1Element != null)
            {
                formulaText = this.Formula1Element.GetValue();
            }

            context.Argument1 = formulaText;
            context.IgnoreBlank = this.AllowBlank;
            context.CultureInfo = XlsxHelper.CultureInfo;
        }

        private void CopyPropertiesToListDataValidationRuleContext(ListDataValidationRuleContext context)
        {
            context.InCellDropdown = this.ShowDropDown;
        }

        private void CopyPropertiesToNumberDataValidationRuleContext(NumberDataValidationRuleContext context)
        {
            context.ComparisonOperator = this.ComparisonOperator;

            if (this.Formula2Element != null)
            {
                string formulaText = this.Formula2Element.GetValue();
                context.Argument2 = formulaText;
            }
        }

        private void InitializeTypeToDataValidationRuleType()
        {
            this.typeToDataValidationRuleType.Add(typeof(CustomDataValidationRule), DataValidationRuleType.Custom);
            this.typeToDataValidationRuleType.Add(typeof(DateDataValidationRule), DataValidationRuleType.Date);
            this.typeToDataValidationRuleType.Add(typeof(DecimalDataValidationRule), DataValidationRuleType.Decimal);
            this.typeToDataValidationRuleType.Add(typeof(ListDataValidationRule), DataValidationRuleType.List);
            this.typeToDataValidationRuleType.Add(typeof(AnyValueDataValidationRule), DataValidationRuleType.None);
            this.typeToDataValidationRuleType.Add(typeof(TextLengthDataValidationRule), DataValidationRuleType.TextLength);
            this.typeToDataValidationRuleType.Add(typeof(TimeDataValidationRule), DataValidationRuleType.Time);
            this.typeToDataValidationRuleType.Add(typeof(WholeNumberDataValidationRule), DataValidationRuleType.Whole);
        }

        private void InitializeCopyActions()
        {
            this.copyActions.Add(this.CopyPropertiesFromDataValidationRule);
            this.copyActions.Add(this.CopyPropertiesFromSingleArgumentRule);
            this.copyActions.Add(this.CopyPropertiesFromListRule);
            this.copyActions.Add(this.CopyPropertiesFromNumberRule);
        }

        private void CopyPropertiesFromDataValidationRule(IXlsxWorksheetExportContext context, IDataValidationRule rule, CellIndex topLeftCellIndex, IEnumerable<CellRange> ruleCellRanges)
        {
            this.Error = rule.ErrorAlertContent;
            this.ErrorStyle = rule.ErrorStyle;
            this.ErrorTitle = rule.ErrorAlertTitle;
            this.Prompt = rule.InputMessageContent;
            this.PromptTitle = rule.InputMessageTitle;
            this.ShowErrorMessage = rule.ShowErrorMessage;
            this.ShowInputMessage = rule.ShowInputMessage;
            this.SqRef = new CellRefRangeSequence(ruleCellRanges.Select((range) =>
            {
                return new CellRefRange(range);
            }));

            this.Type = this.typeToDataValidationRuleType[rule.GetType()];
        }

        private void CopyPropertiesFromSingleArgumentRule(IXlsxWorksheetExportContext context, IDataValidationRule rule, CellIndex topLeftCellIndex, IEnumerable<CellRange> ruleCellRanges)
        {
            SingleArgumentDataValidationRuleBase singleArgumentRule = rule as SingleArgumentDataValidationRuleBase;
            if (singleArgumentRule != null)
            {
                this.AllowBlank = singleArgumentRule.IgnoreBlank;

                this.Formula1Element.InnerText = this.GetFormulaString(singleArgumentRule.Argument1, context, topLeftCellIndex, singleArgumentRule);
            }
        }

        private void CopyPropertiesFromListRule(IXlsxWorksheetExportContext context, IDataValidationRule rule, CellIndex topLeftCellIndex, IEnumerable<CellRange> ruleCellRanges)
        {
            ListDataValidationRule listRule = rule as ListDataValidationRule;
            if (listRule != null)
            {
                this.ShowDropDown = listRule.InCellDropdown;
            }
        }

        private void CopyPropertiesFromNumberRule(IXlsxWorksheetExportContext context, IDataValidationRule rule, CellIndex topLeftCellIndex, IEnumerable<CellRange> ruleCellRanges)
        {
            NumberDataValidationRuleBase numberRule = rule as NumberDataValidationRuleBase;
            if (numberRule != null)
            {
                this.ComparisonOperator = numberRule.ComparisonOperator;
                this.Formula2Element.InnerText = this.GetFormulaString(numberRule.Argument2, context, topLeftCellIndex, numberRule);
            }
        }

        private string GetFormulaString(ICellValue argument, IXlsxWorksheetExportContext context, CellIndex topLeftCellIndex, SingleArgumentDataValidationRuleBase singleArgumentRule)
        {
            string value;
            if (argument.ValueType == CellValueType.Formula)
            {
                FormulaCellValue formulaCellValue = argument as FormulaCellValue;
                formulaCellValue = singleArgumentRule.GetTranslatedArgument(argument, context.Worksheet, topLeftCellIndex.RowIndex, topLeftCellIndex.ColumnIndex) as FormulaCellValue;
                value = formulaCellValue.GetExportString();
            }
            else if (argument.ValueType == CellValueType.Number)
            {
                value = XlsxHelper.CultureInfo.ToString(((NumberCellValue)argument).Value);
            }
            else
            {
                value = singleArgumentRule.GetExpressionString(argument, context.Worksheet, topLeftCellIndex.RowIndex, topLeftCellIndex.ColumnIndex);
            }

            if (argument.ValueType == CellValueType.Text)
            {
                if (singleArgumentRule is ListDataValidationRule)
                {
                    value = value.Replace(FormatHelper.DefaultSpreadsheetCulture.ListSeparator, SpreadsheetCultureHelper.InvariantSpreadsheetCultureInfo.ListSeparator);
                }

                value = string.Format("\"{0}\"", value);
            }

            return value;
        }
    }
}