﻿using System;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Export;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Import;
using Telerik.Windows.Documents.Spreadsheet.Model.Printing;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Elements.Worksheets
{
    internal class HeaderFooterElement : WorksheetElementBase
    {
        private readonly BoolOpenXmlAttribute differentFirst;
        private readonly BoolOpenXmlAttribute differentOddEven;
        private readonly BoolOpenXmlAttribute scaleWithDocument;
        private readonly BoolOpenXmlAttribute alignWithMargins;
        private readonly OpenXmlChildElement<OddHeaderElement> oddHeader;
        private readonly OpenXmlChildElement<OddFooterElement> oddFooter;
        private readonly OpenXmlChildElement<EvenHeaderElement> evenHeader;
        private readonly OpenXmlChildElement<EvenFooterElement> evenFooter;
        private readonly OpenXmlChildElement<FirstHeaderElement> firstHeader;
        private readonly OpenXmlChildElement<FirstFooterElement> firstFooter;

        public HeaderFooterElement(XlsxPartsManager partsManager)
            : base(partsManager)
        {
            this.differentFirst = this.RegisterAttribute(new BoolOpenXmlAttribute("differentFirst", HeaderFooterSettings.DefaultDifferentFirstPage, false));
            this.differentOddEven = this.RegisterAttribute(new BoolOpenXmlAttribute("differentOddEven", HeaderFooterSettings.DefaultDifferentOddAndEvenPages, false));
            this.scaleWithDocument = this.RegisterAttribute(new BoolOpenXmlAttribute("scaleWithDoc", HeaderFooterSettings.DefaultScaleWithDocument, false));
            this.alignWithMargins = this.RegisterAttribute(new BoolOpenXmlAttribute("alignWithMargins", HeaderFooterSettings.DefaultAlignWithPageMargins, false));

            this.oddHeader = this.RegisterChildElement<OddHeaderElement>(XlsxElementNames.OddHeaderElementName);
            this.oddFooter = this.RegisterChildElement<OddFooterElement>(XlsxElementNames.OddFooterElementName);
            this.evenHeader = this.RegisterChildElement<EvenHeaderElement>(XlsxElementNames.EvenHeaderElementName);
            this.evenFooter = this.RegisterChildElement<EvenFooterElement>(XlsxElementNames.EvenFooterElementName);
            this.firstHeader = this.RegisterChildElement<FirstHeaderElement>(XlsxElementNames.FirstHeaderElementName);
            this.firstFooter = this.RegisterChildElement<FirstFooterElement>(XlsxElementNames.FirstFooterElementName);
        }

        public override string ElementName
        {
            get
            {
                return XlsxElementNames.HeaderFooterElementName;
            }
        }

        public bool DifferentFirst
        {
            get
            {
                return this.differentFirst.Value;
            }
            set
            {
                this.differentFirst.Value = value;
            }
        }

        public bool DifferentOddEven
        {
            get
            {
                return this.differentOddEven.Value;
            }
            set
            {
                this.differentOddEven.Value = value;
            }
        }

        public bool ScaleWithDocument
        {
            get
            {
                return this.scaleWithDocument.Value;
            }
            set
            {
                this.scaleWithDocument.Value = value;
            }
        }

        public bool AlignWithMargins
        {
            get
            {
                return this.alignWithMargins.Value;
            }
            set
            {
                this.alignWithMargins.Value = value;
            }
        }

        public OddHeaderElement OddHeader
        {
            get
            {
                return this.oddHeader.Element;
            }
            set
            {
                this.oddHeader.Element = value;
            }
        }

        public OddFooterElement OddFooter
        {
            get
            {
                return this.oddFooter.Element;
            }
            set
            {
                this.oddFooter.Element = value;
            }
        }

        public EvenHeaderElement EvenHeader
        {
            get
            {
                return this.evenHeader.Element;
            }
            set
            {
                this.evenHeader.Element = value;
            }
        }

        public EvenFooterElement EvenFooter
        {
            get
            {
                return this.evenFooter.Element;
            }
            set
            {
                this.evenFooter.Element = value;
            }
        }

        public FirstHeaderElement FirstHeader
        {
            get
            {
                return this.firstHeader.Element;
            }
            set
            {
                this.firstHeader.Element = value;
            }
        }

        public FirstFooterElement FirstFooter
        {
            get
            {
                return this.firstFooter.Element;
            }
            set
            {
                this.firstFooter.Element = value;
            }
        }

        protected override void OnAfterReadAttributes(IXlsxWorksheetImportContext context)
        {
            base.OnAfterReadAttributes(context);

            HeaderFooterSettings settings = context.Worksheet.WorksheetPageSetup.HeaderFooterSettings;

            if (this.differentFirst.HasValue)
            {
                settings.DifferentFirstPage = this.DifferentFirst;
            }

            if (this.differentOddEven.HasValue)
            {
                settings.DifferentOddAndEvenPages = this.DifferentOddEven;
            }

            if (this.alignWithMargins.HasValue)
            {
                settings.AlignWithPageMargins = this.AlignWithMargins;
            }

            if (this.scaleWithDocument.HasValue)
            {
                settings.ScaleWithDocument = this.ScaleWithDocument;
            }
        }

        protected override void OnAfterRead(IXlsxWorksheetImportContext context)
        {
            base.OnAfterRead(context);

            HeaderFooterSettings settings = context.Worksheet.WorksheetPageSetup.HeaderFooterSettings;

            this.ReadChild(this.OddHeader, settings.Header);
            this.ReadChild(this.OddFooter, settings.Footer);

            if (settings.DifferentOddAndEvenPages)
            {
                this.ReadChild(this.EvenHeader, settings.EvenPageHeader);
                this.ReadChild(this.EvenFooter, settings.EvenPageFooter);
            }

            if (settings.DifferentFirstPage)
            {
                this.ReadChild(this.FirstHeader, settings.FirstPageHeader);
                this.ReadChild(this.FirstFooter, settings.FirstPageFooter);
            }
        }

        protected override void OnBeforeWrite(IXlsxWorksheetExportContext context)
        {
            base.OnBeforeWrite(context);

            HeaderFooterSettings settings = context.Worksheet.WorksheetPageSetup.HeaderFooterSettings;

            this.WriteAttributes(settings);
            this.WriteChildren(settings);
        }

        protected override bool ShouldExport(IXlsxWorksheetExportContext context)
        {
            bool shouldExport = false;

            HeaderFooterSettings settings = context.Worksheet.WorksheetPageSetup.HeaderFooterSettings;

            shouldExport |= !settings.Header.IsEmpty;
            shouldExport |= !settings.Footer.IsEmpty;

            if (settings.DifferentOddAndEvenPages)
            {                
                shouldExport |= !settings.EvenPageHeader.IsEmpty;
                shouldExport |= !settings.EvenPageFooter.IsEmpty;
            }

            if (settings.DifferentFirstPage)
            {
                shouldExport |= !settings.FirstPageHeader.IsEmpty;
                shouldExport |= !settings.FirstPageFooter.IsEmpty;
            }

            return shouldExport;
        }

        private void ReadChild(HeaderFooterChildElementBase child, HeaderFooterContent content)
        {
            if (child != null)
            {
                content.SetFromContentText(child.InnerText);
            }
        }

        private void WriteChildren(HeaderFooterSettings settings)
        {
            this.WriteChild(this.oddHeader, settings.Header);
            this.WriteChild(this.oddFooter, settings.Footer);

            if (settings.DifferentOddAndEvenPages)
            {
                this.WriteChild(this.evenHeader, settings.EvenPageHeader);
                this.WriteChild(this.evenFooter, settings.EvenPageFooter);
            }

            if (settings.DifferentFirstPage)
            {
                this.WriteChild(this.firstHeader, settings.FirstPageHeader);
                this.WriteChild(this.firstFooter, settings.FirstPageFooter);
            }
        }

        private void WriteChild<T>(OpenXmlChildElement<T> child, HeaderFooterContent content)
            where T : HeaderFooterChildElementBase
        {
            string contentText = content.BuildContentText();

            if (!string.IsNullOrEmpty(contentText))
            {
                this.CreateElement(child);

                if (!content.HasValidLength)
                {
                    contentText = contentText.Substring(0, HeaderFooterContent.MaxLength);
                }

                child.Element.InnerText = contentText;
            }
        }

        private void WriteAttributes(HeaderFooterSettings settings)
        {
            if (settings.DifferentFirstPage != this.differentFirst.DefaultValue)
            {
                this.DifferentFirst = settings.DifferentFirstPage;
            }

            if (settings.DifferentOddAndEvenPages != this.differentOddEven.DefaultValue)
            {
                this.DifferentOddEven = settings.DifferentOddAndEvenPages;
            }

            if (settings.AlignWithPageMargins != this.alignWithMargins.DefaultValue)
            {
                this.AlignWithMargins = settings.AlignWithPageMargins;
            }

            if (settings.ScaleWithDocument != this.scaleWithDocument.DefaultValue)
            {
                this.ScaleWithDocument = settings.ScaleWithDocument;
            }
        }
    }
}
