﻿using System;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Contexts;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Export;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Parts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Types;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Elements.Worksheets
{
    internal class HyperlinkElement : WorksheetElementBase
    {
        private readonly ConvertedOpenXmlAttribute<CellRefRange> reference;
        private readonly OpenXmlAttribute<string> relationId;
        private readonly OpenXmlAttribute<string> tooltip;
        private readonly OpenXmlAttribute<string> location;
        private readonly OpenXmlAttribute<string> display;

        public HyperlinkElement(XlsxPartsManager partsManager)
            : base(partsManager)
        {
            this.reference = this.RegisterAttribute(new ConvertedOpenXmlAttribute<CellRefRange>("ref", XlsxConverters.CellRefRangeConverter, true));
            this.relationId = this.RegisterAttribute(new OpenXmlAttribute<string>("id", OpenXmlNamespaces.OfficeDocumentRelationshipsNamespace));
            this.tooltip = this.RegisterAttribute(new OpenXmlAttribute<string>("tooltip"));
            this.location = this.RegisterAttribute(new OpenXmlAttribute<string>("location"));
            this.display = this.RegisterAttribute(new OpenXmlAttribute<string>("display"));
        }

        public override string ElementName
        {
            get
            {
                return XlsxElementNames.HyperlinkElementName;
            }
        }

        public CellRefRange Reference
        {
            get
            {
                return this.reference.Value;
            }
            set
            {
                this.reference.Value = value;
            }
        }

        public string RelationshipId
        {
            get
            {
                return this.relationId.Value;
            }
            set
            {
                this.relationId.Value = value;
            }
        }

        public string Tooltip
        {
            get
            {
                return this.tooltip.Value;
            }
            set
            {
                this.tooltip.Value = value;
            }
        }

        public string Location
        {
            get
            {
                return this.location.Value;
            }
            set
            {
                this.location.Value = value;
            }
        }

        public string Display
        {
            get
            {
                return this.display.Value;
            }
            set
            {
                this.display.Value = value;
            }
        }

        public void CopyPropertiesFrom(IXlsxWorksheetExportContext context, SpreadsheetHyperlink hyperlink)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            this.Reference = new CellRefRange(hyperlink.Range);

            if (!string.IsNullOrEmpty(hyperlink.HyperlinkInfo.SubAddress))
            {
                this.Location = hyperlink.HyperlinkInfo.SubAddress;
            }
            else
            {
                string address;
                if (SpreadsheetHelper.EmailRegex.IsMatch(hyperlink.HyperlinkInfo.Address))
                {
                    address = hyperlink.HyperlinkInfo.GetEmailAddress();
                }
                else
                {
                    address = hyperlink.HyperlinkInfo.Address;
                }

                this.RelationshipId = this.PartsManager.CreateWorksheetRelationship((WorksheetPart)this.Part, address, XlsxRelationshipTypes.HyperlinkRelationshipType, TargetModes.External);
            }

            if (!string.IsNullOrEmpty(hyperlink.HyperlinkInfo.ScreenTip))
            {
                this.Tooltip = hyperlink.HyperlinkInfo.ScreenTip;
            }
        }

        protected override void OnAfterRead(Import.IXlsxWorksheetImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            string address = string.Empty;
            string emailAddress = string.Empty;
            string emailSubject = string.Empty;
            string location = this.Location;
            string tooltip = this.Tooltip;

            if (this.RelationshipId != null)
            {
                string target = this.PartsManager.GetWorksheetRelationshipTarget((WorksheetPart)this.Part, this.RelationshipId);
                if (!SpreadsheetHelper.TryParseMailToAddress(target, out emailAddress, out emailSubject))
                {
                    address = target;
                }
            }

            HyperlinkInfo hyperlinkInfo;
            if (!string.IsNullOrEmpty(location))
            {
                hyperlinkInfo = HyperlinkInfo.CreateInDocumentHyperlink(location, tooltip);
            }
            else if (!string.IsNullOrEmpty(emailAddress))
            {
                hyperlinkInfo = HyperlinkInfo.CreateMailtoHyperlink(emailAddress, emailSubject, tooltip);
            }
            else
            {
                hyperlinkInfo = HyperlinkInfo.CreateHyperlink(address, tooltip);
            }

            context.Worksheet.Hyperlinks.Add(this.Reference.ToCellRange(), hyperlinkInfo);
        }
    }
}
