﻿using System;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Types;
using Telerik.Windows.Documents.Model;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Contexts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Export;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Import;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Types;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Printing;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Elements.Worksheets
{
    internal class PageSetupElement : WorksheetElementBase
    {
        private readonly ConvertedOpenXmlAttribute<PaperTypes> paperSize;
        private readonly ConvertedOpenXmlAttribute<PageOrientation> orientation;
        private readonly ConvertedOpenXmlAttribute<PageOrder> pageOrder;
        private readonly ConvertedOpenXmlAttribute<ErrorsPrintStyle> errors;
        private readonly IntOpenXmlAttribute firstPageNumber;
        private readonly IntOpenXmlAttribute fitToHeight;
        private readonly IntOpenXmlAttribute fitToWidth;
        private readonly IntOpenXmlAttribute scale;

        public PageSetupElement(XlsxPartsManager partsManager)
            : base(partsManager)
        {
            this.paperSize = this.RegisterAttribute(new ConvertedOpenXmlAttribute<PaperTypes>("paperSize", null, XlsxConverters.PaperTypesConverter, PaperTypesConverter.DefaultExportPaperType));
            this.orientation = this.RegisterAttribute(new ConvertedOpenXmlAttribute<PageOrientation>("orientation", null, Converters.PageOrientationConverter, PageOrientation.Portrait));
            this.pageOrder = this.RegisterAttribute(new ConvertedOpenXmlAttribute<PageOrder>("pageOrder", null, XlsxConverters.PageOrderConverter, PageOrder.DownThenOver));
            this.errors = this.RegisterAttribute(new ConvertedOpenXmlAttribute<ErrorsPrintStyle>("errors", null, XlsxConverters.ErrorsPrintStyleConverter, ErrorsPrintStyle.AsDisplayed));
            this.firstPageNumber = this.RegisterAttribute(new IntOpenXmlAttribute("firstPageNumber", 1));
            this.fitToHeight = this.RegisterAttribute(new IntOpenXmlAttribute("fitToHeight", 1));
            this.fitToWidth = this.RegisterAttribute(new IntOpenXmlAttribute("fitToWidth", 1));
            this.scale = this.RegisterAttribute(new IntOpenXmlAttribute("scale", 100));
        }

        public PaperTypes PaperSize
        {
            get
            {
                return this.paperSize.Value;
            }
            set
            {
                this.paperSize.Value = value;
            }
        }

        public PageOrientation Orientation
        {
            get
            {
                return this.orientation.Value;
            }
            set
            {
                this.orientation.Value = value;
            }
        }

        public PageOrder PageOrder
        {
            get
            {
                return this.pageOrder.Value;
            }
            set
            {
                this.pageOrder.Value = value;
            }
        }
        
        public ErrorsPrintStyle Errors
        {
            get
            {
                return this.errors.Value;
            }
            set
            {
                this.errors.Value = value;
            }
        }

        public int FirstPageNumber
        {
            get
            {
                return this.firstPageNumber.Value;
            }
            set
            {
                this.firstPageNumber.Value = value;
            }
        }

        public int FitToHeight
        {
            get
            {
                return this.fitToHeight.Value;
            }
            set
            {
                this.fitToHeight.Value = value;
            }
        }

        public int FitToWidth
        {
            get
            {
                return this.fitToWidth.Value;
            }
            set
            {
                this.fitToWidth.Value = value;
            }
        }

        public int Scale
        {
            get
            {
                return this.scale.Value;
            }
            set
            {
                this.scale.Value = value;
            }
        }

        public override string ElementName
        {
            get
            {
                return XlsxElementNames.PageSetupElementName;
            }
        }

        protected override void OnAfterRead(IXlsxWorksheetImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            PageSetupInfo pageSetupInfo = new PageSetupInfo();

            pageSetupInfo.PaperType = this.PaperSize;

            pageSetupInfo.PageOrientation = this.Orientation;

            if (this.pageOrder.HasValue)
            {
                pageSetupInfo.PageOrder = this.PageOrder;
            }

            if (this.errors.HasValue)
            {
                pageSetupInfo.Errors = this.Errors;
            }

            if (this.firstPageNumber.HasValue)
            {
                pageSetupInfo.FirstPageNumber = this.FirstPageNumber;
            }

            if (this.fitToHeight.HasValue)
            {
                pageSetupInfo.FitToHeight = this.FitToHeight;
            }

            if (this.fitToWidth.HasValue)
            {
                pageSetupInfo.FitToWidth = this.FitToWidth;
            }

            pageSetupInfo.Scale = 
                (int)SpreadsheetHelper.GetNearestValidValue(this.Scale, SpreadsheetDefaultValues.MinScaleFactor * 100, SpreadsheetDefaultValues.MaxScaleFactor * 100);

            context.ApplyPageSetup(pageSetupInfo);
        }

        protected override void OnBeforeWrite(IXlsxWorksheetExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            PageSetupInfo pageSetupInfo = context.GetPageSetupInfo();
            bool shouldExportElement = false;

            if (pageSetupInfo.PageOrder != this.pageOrder.DefaultValue)
            {
                this.PageOrder = pageSetupInfo.PageOrder;
                shouldExportElement = true;
            }

            if (pageSetupInfo.Errors != this.errors.DefaultValue)
            {
                this.Errors = pageSetupInfo.Errors;
                shouldExportElement = true;
            }

            if (pageSetupInfo.FirstPageNumber != null && pageSetupInfo.FirstPageNumber != this.firstPageNumber.DefaultValue)
            {
                this.FirstPageNumber = (int)pageSetupInfo.FirstPageNumber;
                shouldExportElement = true;
            }

            if (pageSetupInfo.FitToHeight != this.fitToHeight.DefaultValue)
            {
                this.FitToHeight = pageSetupInfo.FitToHeight;
                shouldExportElement = true;
            }

            if (pageSetupInfo.FitToWidth != this.fitToWidth.DefaultValue)
            {
                this.FitToWidth = pageSetupInfo.FitToWidth;
                shouldExportElement = true;
            }

            if (pageSetupInfo.Scale != this.scale.DefaultValue)
            {
                this.Scale = pageSetupInfo.Scale;
                shouldExportElement = true;
            }

            if (shouldExportElement || pageSetupInfo.PaperType != this.paperSize.DefaultValue)
            {
                this.PaperSize = pageSetupInfo.PaperType;
                shouldExportElement = true;
            }

            if (shouldExportElement || pageSetupInfo.PageOrientation != this.orientation.DefaultValue)
            {
                this.Orientation = pageSetupInfo.PageOrientation;
            }
        }
    }
}