﻿using System.Collections.Generic;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Types;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.Contexts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Contexts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Export;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Import;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Elements.Common;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Elements.Worksheets
{
    internal class RowElement : WorksheetElementBase
    {
        private readonly OpenXmlAttribute<double> rowHeight;
        private readonly ConvertedOpenXmlAttribute<int?> rowIndex;
        private readonly BoolOpenXmlAttribute customFormat;
        private readonly BoolOpenXmlAttribute customHeight;
        private readonly IntOpenXmlAttribute styleIndex;
        private readonly BoolOpenXmlAttribute hidden;
        private readonly IntOpenXmlAttribute outlineLevel;

        public RowElement(XlsxPartsManager partsManager)
            : base(partsManager)
        {
            this.rowIndex = this.RegisterAttribute(new ConvertedOpenXmlAttribute<int?>("r", Converters.NullableIntValueConverter, false));
            this.rowHeight = this.RegisterAttribute<double>("ht");
            this.styleIndex = this.RegisterAttribute(new IntOpenXmlAttribute("s", 0));
            this.customFormat = this.RegisterAttribute(new BoolOpenXmlAttribute("customFormat"));
            this.customHeight = this.RegisterAttribute(new BoolOpenXmlAttribute("customHeight"));
            this.hidden = this.RegisterAttribute(new BoolOpenXmlAttribute("hidden"));
            this.outlineLevel = this.RegisterAttribute(new IntOpenXmlAttribute("outlineLevel"));
        }

        public double RowHeight
        {
            get
            {
                return this.rowHeight.Value;
            }
            set
            {
                this.rowHeight.Value = value;
            }
        }

        public int? RowIndex
        {
            get
            {
                return this.rowIndex.Value;
            }
            set
            {
                this.rowIndex.Value = value;
            }
        }

        public bool CustomFormat
        {
            get
            {
                return this.customFormat.Value;
            }
            set
            {
                this.customFormat.Value = value;
            }
        }

        public bool CustomHeight
        {
            get
            {
                return this.customHeight.Value;
            }
            set
            {
                this.customHeight.Value = value;
            }
        }

        public int StyleIndex
        {
            get
            {
                return this.styleIndex.Value;
            }
            set
            {
                this.styleIndex.Value = value;
            }
        }

        public bool Hidden
        {
            get
            {
                return this.hidden.Value;
            }
            set
            {
                this.hidden.Value = value;
            }
        }

        public int OutlineLevel
        {
            get
            {
                return this.outlineLevel.Value;
            }
            set
            {
                this.outlineLevel.Value = value;
            }
        }

        public override string ElementName
        {
            get
            {
                return XlsxElementNames.RowElementName;
            }
        }

        public void CopyPropertiesFrom(IXlsxWorksheetExportContext context, RowInfo row)
        {
            Guard.ThrowExceptionIfNull(context, "context");
            Guard.ThrowExceptionIfNull(row, "row");

            this.RowIndex = row.RowIndex + 1;

            if (row.IsCustom)
            {
                this.CustomHeight = true;
            }

            if (row.IsCustom || (!row.IsDefault && row.Height.HasValue))
            {
                this.RowHeight = UnitHelper.DipToPoint(row.Height.Value);
            }

            if (row.OutlineLevel != this.outlineLevel.DefaultValue)
            {
                this.OutlineLevel = row.OutlineLevel;
            }

            if (row.Hidden != this.hidden.DefaultValue)
            {
                this.Hidden = row.Hidden;
            }

            int? styleIndex = context.GetFormattingRecordIndex(context.Worksheet.Rows, (long)row.RowIndex);
            if (styleIndex.HasValue)
            {
                this.StyleIndex = styleIndex.Value;
                this.CustomFormat = true;
            }
        }

        protected override IEnumerable<XlsxElementBase> EnumerateChildElements(IXlsxWorksheetExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            foreach (CellInfo cell in context.GetNonEmptyCellsInRow(this.RowIndex.Value - 1))
            {
                CellElement cellElement = this.CreateElement<CellElement>(XlsxElementNames.CellElementName);
                cellElement.CopyPropertiesFrom(context, cell);

                yield return cellElement;
            }
        }

        protected override void OnAfterReadAttributes(IXlsxWorksheetImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            RowInfo rowInfo = new RowInfo();
            rowInfo.RowIndex = this.RowIndex - 1;
            rowInfo.IsCustom = this.CustomHeight;
            rowInfo.Hidden = this.Hidden;
            rowInfo.OutlineLevel = this.OutlineLevel;
            rowInfo.StyleIndex = this.StyleIndex;

            if (this.rowHeight.HasValue)
            {
                rowInfo.Height = UnitHelper.PointToDip(this.RowHeight);
                rowInfo.IsCustom = true;
            }
            else if (rowInfo.IsCustom)
            {
                rowInfo.Height = SpreadsheetDefaultValues.DefaultRowHeight;
            }

            context.ApplyRowInfo(rowInfo);
        }
    }
}