﻿using System;
using System.Collections.Generic;
using System.Windows;
using Telerik.Windows.Documents.FormatProviders.OpenXml.Model.Core;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Export;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Import;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Types;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Model.Elements.Worksheets
{
    internal class SheetViewElement : WorksheetElementBase
    {
        private readonly OpenXmlChildElement<PaneElement> paneChildElement;
        private readonly OpenXmlChildElement<SelectionElement> selectionChildElement;

        private readonly OpenXmlAttribute<double> zoomScale;
        private readonly ConvertedOpenXmlAttribute<CellRef> topLeftCell;
        private readonly IntOpenXmlAttribute workbookViewId;
        private readonly BoolOpenXmlAttribute showGridLines;
        private readonly BoolOpenXmlAttribute showRowColHeaders;

        public SheetViewElement(XlsxPartsManager partsManager)
            : base(partsManager)
        {
            this.paneChildElement = this.RegisterChildElement<PaneElement>(XlsxElementNames.PaneElementName);
            this.selectionChildElement = this.RegisterChildElement<SelectionElement>(XlsxElementNames.SelectionElementName);

            this.zoomScale = this.RegisterAttribute<double>("zoomScale", 100);
            this.topLeftCell = this.RegisterAttribute(new ConvertedOpenXmlAttribute<CellRef>("topLeftCell", XlsxConverters.CellRefConverter, new CellRef(0, 0)));
            this.workbookViewId = this.RegisterAttribute(new IntOpenXmlAttribute("workbookViewId", 0, true));
            this.showGridLines = this.RegisterAttribute(new BoolOpenXmlAttribute("showGridLines", true, false));
            this.showRowColHeaders = this.RegisterAttribute(new BoolOpenXmlAttribute("showRowColHeaders", true, false));
        }

        public double ZoomScale
        {
            get
            {
                return this.zoomScale.Value;
            }
            set
            {
                this.zoomScale.Value = value;
            }
        }

        public CellRef TopLeftCell
        {
            get
            {
                return this.topLeftCell.Value;
            }
            set
            {
                this.topLeftCell.Value = value;
            }
        }

        public int WorkbookViewId
        {
            get
            {
                return this.workbookViewId.Value;
            }
            set
            {
                this.workbookViewId.Value = value;
            }
        }

        public bool ShowGridLines
        {
            get
            {
                return this.showGridLines.Value;
            }
            set
            {
                this.showGridLines.Value = value;
            }
        }

        public bool ShowRowColHeaders
        {
            get
            {
                return this.showRowColHeaders.Value;
            }
            set
            {
                this.showRowColHeaders.Value = value;
            }
        }

        public SelectionElement SelectionElement
        {
            get
            {
                return this.selectionChildElement.Element;
            }
            set
            {
                this.selectionChildElement.Element = value;
            }
        }

        public PaneElement PaneElement
        {
            get
            {
                return this.paneChildElement.Element;
            }
            set
            {
                this.paneChildElement.Element = value;
            }
        }

        public override string ElementName
        {
            get
            {
                return XlsxElementNames.SheetViewElementName;
            }
        }

        public override bool AlwaysExport
        {
            get
            {
                return true;
            }
        }

        protected override void OnAfterRead(IXlsxWorksheetImportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            ISheet currentWorksheet = context.Worksheet;
            WorksheetViewState worksheetViewState = (WorksheetViewState)currentWorksheet.ViewState;

            worksheetViewState.TopLeftCellIndex = this.TopLeftCell.ToCellIndex();

            worksheetViewState.ShowGridLines = this.ShowGridLines;
            worksheetViewState.ShowRowColHeaders = this.ShowRowColHeaders;

            if (this.SelectionElement != null)
            {
                CellIndex cellIndex = this.SelectionElement.ActiveCell != null ? this.SelectionElement.ActiveCell.ToCellIndex() : new CellIndex(0, 0);

                Layout.ViewportPaneType viewPortPaneType = this.SelectionElement.Pane;

                IEnumerable<CellRange> selectedRanges = new CellRange[] { new CellRange(0, 0, 0, 0) };
                if (this.SelectionElement.SqRef != null)
                {
                    selectedRanges = this.SelectionElement.SqRef.CellRanges;
                }
                SelectionState selectionState = new SelectionState(selectedRanges, cellIndex, viewPortPaneType);

                worksheetViewState.SelectionState = selectionState;
            }

            worksheetViewState.ScaleFactor = this.GetScaleFactor();

            if (this.PaneElement != null)
            {
                Pane pane = this.PaneElement.ToPane();
                worksheetViewState.Pane = pane;
            }

            this.ReleaseElement(this.paneChildElement);
            this.ReleaseElement(this.selectionChildElement);
        }

        protected override void OnBeforeWrite(IXlsxWorksheetExportContext context)
        {
            Guard.ThrowExceptionIfNull(context, "context");

            CellIndex topLeftCellIndex = context.WorksheetViewState.TopLeftCellIndex;
            this.TopLeftCell = new CellRef(topLeftCellIndex.RowIndex, topLeftCellIndex.ColumnIndex);
            this.ZoomScale = Math.Floor(context.WorksheetViewState.ScaleFactor.Height * 100);
            this.ShowGridLines = context.WorksheetViewState.ShowGridLines;
            this.ShowRowColHeaders = context.WorksheetViewState.ShowRowColHeaders;

            this.CreateElement(this.paneChildElement);
            this.CreateElement(this.selectionChildElement);
        }

        private Size GetScaleFactor()
        {
            double value = this.ZoomScale / 100;

            double sizeX;
            double sizeY;

            sizeX = sizeY = Math.Max(0.5, value);

            return new Size(sizeX, sizeY);
        }
    }
}