﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Contexts;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Export;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx.Import;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx
{
    /// <summary>
    /// Binary format provider to export Spreadsheet data to XLSX.
    /// </summary>
    public class XlsxFormatProvider : BinaryWorkbookFormatProviderBase
    {
        private static readonly string name = "XlsxFormatProvider";
        private static readonly string fileDescription = "Excel Workbook";
        private static readonly IEnumerable<string> supportedExtensions = new string[] { ".xlsx" };

        /// <summary>
        /// Initializes a new instance of the <see cref="XlsxFormatProvider" /> class.
        /// </summary>
        public XlsxFormatProvider()
        {
            this.ExportSettings = new XlsxExportSettings();
            this.ImportSettings = new XlsxImportSettings();
        }

        /// <summary>
        /// Gets the name.
        /// </summary>
        /// <value>The name of the provider.</value>
        public override string Name
        {
            get
            {
                return name;
            }
        }

        /// <summary>
        /// Gets the files description.
        /// </summary>
        /// <value>The files description.</value>
        public override string FilesDescription
        {
            get
            {
                return fileDescription;
            }
        }

        /// <summary>
        /// Gets the supported extensions.
        /// </summary>
        /// <value>The supported extensions.</value>
        public override IEnumerable<string> SupportedExtensions
        {
            get
            {
                return supportedExtensions;
            }
        }

        /// <summary>
        /// Gets a value indicating whether can import.
        /// </summary>
        /// <value>The value indicating whether can import.</value>
        public override bool CanImport
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Gets a value indicating whether can export.
        /// </summary>
        /// <value>The value indicating whether can export.</value>
        public override bool CanExport
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Reserved for future use.
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public XlsxExportSettings ExportSettings { get; set; }

        /// <summary>
        /// Reserved for future use.
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        public XlsxImportSettings ImportSettings { get; set; }

        /// <summary>
        /// Imports the specified input.
        /// </summary>
        /// <param name="input">The input.</param>
        /// <returns>The imported workbook.</returns>
        protected override Workbook ImportOverride(Stream input)
        {
            XlsxImporter importer = new XlsxImporter();
            XlsxWorkbookImportContext importContext = new XlsxWorkbookImportContext();

            Action beginUpdateAction = () =>
            {
                importContext.Workbook.CellReferenceRangeExpressionCache.IsEnabled = true;
            };

            Action endUpdateAction = () =>
            {
                importContext.Workbook.CellReferenceRangeExpressionCache.IsEnabled = false;
            };

            using (new UpdateScope(beginUpdateAction, endUpdateAction))
            {
                importer.Import(input, importContext);
            }

            return importContext.Workbook;
        }

        /// <summary>
        /// Exports the specified workbook.
        /// </summary>
        /// <param name="workbook">The workbook.</param>
        /// <param name="output">The output.</param>
        protected override void ExportOverride(Workbook workbook, Stream output)
        {
            XlsxExporter exporter = new XlsxExporter();

            using (new LicenseCheck(workbook))
            {
                exporter.Export(output, new XlsxWorkbookExportContext(workbook), this.ExportSettings);
            }
        }
    }
}