﻿using System;
using System.Collections.Generic;
using System.IO;
using Telerik.Windows.Documents.Fixed.Model;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.Pdf.Export;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.FormatProviders.Pdf
{
    /// <summary>
    /// Represents the PdfFormatProvider.
    /// </summary>
    public class PdfFormatProvider : BinaryWorkbookFormatProviderBase
    {
        private static readonly string name = "PdfFormatProvider";
        private static readonly string fileDescription = "PDF Documents";
        private static readonly string[] supportedExtensions = new string[] { ".pdf" };
        private readonly PdfExporter exporter;

        /// <summary>
        /// Initializes a new instance of the <see cref="PdfFormatProvider" /> class.
        /// </summary>
        public PdfFormatProvider()
        {
            this.exporter = new PdfExporter();
            this.ExportSettings = new PdfExportSettings();
        }

        /// <summary>
        /// Gets the name.
        /// </summary>
        /// <value>The name of the provider.</value>
        public override string Name
        {
            get
            {
                return name;
            }
        }

        /// <summary>
        /// Gets the files description.
        /// </summary>
        /// <value>The files description.</value>
        public override string FilesDescription
        {
            get
            {
                return fileDescription;
            }
        }

        /// <summary>
        /// Gets a value indicating whether can export.
        /// </summary>
        /// <value>The value indicating whether can export.</value>
        public override bool CanExport
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Gets a value indicating whether can import.
        /// </summary>
        /// <value>The value indicating whether can import.</value>
        public override bool CanImport
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Gets the supported extensions.
        /// </summary>
        /// <value>The supported extensions.</value>
        public override IEnumerable<string> SupportedExtensions
        {
            get
            {
                return supportedExtensions;
            }
        }

        /// <summary>
        /// Gets or sets the export settings.
        /// </summary>
        /// <value>The export settings.</value>
        public PdfExportSettings ExportSettings { get; set; }

        /// <summary>
        /// Exports to fixed document.
        /// </summary>
        /// <param name="workbook">The workbook.</param>
        /// <returns>The exported fixed document.</returns>
        public RadFixedDocument ExportToFixedDocument(Workbook workbook)
        {
#if !SILVERLIGHT
            this.exporter.ChartRenderer = this.ExportSettings.ChartRenderer;
#endif
            RadFixedDocument document = null;

            switch (this.ExportSettings.ExportWhat)
            {
                case ExportWhat.ActiveSheet:
                    document = this.exporter.CreateFixedDocument(workbook.ActiveWorksheet, this.ExportSettings.IgnorePrintArea);
                    break;
                case ExportWhat.EntireWorkbook:
                    document = this.exporter.CreateFixedDocument(workbook, this.ExportSettings.IgnorePrintArea);
                    break;
                case ExportWhat.Selection:
                    document = this.exporter.CreateFixedDocument(workbook.ActiveWorksheet, this.ExportSettings.SelectedRanges);
                    break;
                default:
                    throw new NotSupportedException("Not supported ExportWhat option!");
            }

            return document;
        }

        /// <summary>
        /// Exports the specified workbook.
        /// </summary>
        /// <param name="workbook">The workbook to export.</param>
        /// <param name="output">The output.</param>
        protected override void ExportOverride(Workbook workbook, Stream output)
        {
            this.exporter.PdfExportSettings = this.ExportSettings.PdfFileSettings;
#if !SILVERLIGHT
            this.exporter.ChartRenderer = this.ExportSettings.ChartRenderer;
#endif

            switch (this.ExportSettings.ExportWhat)
            {
                case ExportWhat.ActiveSheet:
                    this.exporter.Export(workbook.ActiveWorksheet, this.ExportSettings.IgnorePrintArea, output);
                    break;
                case ExportWhat.EntireWorkbook:
                    this.exporter.Export(workbook, this.ExportSettings.IgnorePrintArea, output);
                    break;
                case ExportWhat.Selection:
                    this.exporter.Export(workbook.ActiveWorksheet, this.ExportSettings.SelectedRanges, output);
                    break;
                default:
                    throw new NotSupportedException("Not supported ExportWhat option!");
            }
        }
    }
}
