﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests
{
	[TestClass]
	public class CellValueFormulaTests
    {
        #region Helper Methods

        private static void AssertCellFormula(Worksheet worksheet, int rowIndex, int columnIndex, double expectedValue)
		{
			ICellValue cellValue = worksheet.Cells[rowIndex, columnIndex].GetValue().Value;

			Assert.AreEqual(cellValue.ValueType, CellValueType.Formula);
			FormulaCellValue formula = (FormulaCellValue)cellValue;

			Assert.AreEqual(expectedValue, formula.Value.GetValue().NumberValue());
		}

        private static void AssertReferringToCyclicReference(string input)
        {
            ICellValue cellValue;
            CellValueFormat newFormat;
            RadExpression valueExpression;

            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Cells[0, 0].SetValue("=A1");
            worksheet.Cells[0, 1].SetValue("=B1:B2");

            CellValueFactory.Create(input.ReplaceCultureDependentSeparators(), worksheet, new CellIndex(1, 1), CellValueFormat.GeneralFormat, out cellValue, out newFormat);
            worksheet.Cells[1, 1].SetValue(cellValue);
            valueExpression = ((FormulaCellValue)worksheet.Cells[1, 1].GetValue().Value).Value.GetValueAsConstantExpression();
            
            Assert.AreEqual(valueExpression, ErrorExpressions.CyclicReference);
        }

        private static void AssertCyclicReference(string input, CellIndex cellIndex)
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            ICellValue cellValue;
            CellValueFormat newFormat;
            RadExpression valueExpression;

            CellValueFactory.Create(input.ReplaceCultureDependentSeparators(), worksheet, cellIndex, CellValueFormat.GeneralFormat, out cellValue, out newFormat);
            worksheet.Cells[cellIndex].SetValue(cellValue);
            valueExpression = ((FormulaCellValue)worksheet.Cells[cellIndex].GetValue().Value).Value.GetValueAsConstantExpression();
            Assert.AreEqual(valueExpression, ErrorExpressions.CyclicReference);
        }

        private static void AssertFormulaCellValueAsString(string expectedString, string inputString, CellIndex cellIndex, Worksheet worksheet = null)
        {
            ICellValue cellValue;
            CellValueFormat newFormat;
            
            if (worksheet == null)
            {
                worksheet = TestHelper.CreateWorksheet();
            }

            CellValueFactory.Create(inputString.ReplaceCultureDependentSeparators(), worksheet, cellIndex, CellValueFormat.GeneralFormat, out cellValue, out newFormat);
            
            FormulaCellValue formulaCellValue = cellValue as FormulaCellValue;
            Assert.IsNotNull(formulaCellValue);
            Assert.AreEqual<string>(expectedString.ReplaceCultureDependentSeparators(), formulaCellValue.GetValueAsString(CellValueFormat.GeneralFormat));
        }

		#endregion


		#region Formula Tests

        [TestMethod]
        public void FormulaCellValueAsStringTestParenthesis()
        {
            string testValue = "=((1))";
            AssertFormulaCellValueAsString(testValue, testValue, new CellIndex(0, 0));
        }

        [TestMethod]
        public void FormulaCellValueAsStringTestParenthesisSpaces()
        {
            string testValue = "=( - ( 1 ) )";
            AssertFormulaCellValueAsString(testValue, testValue, new CellIndex(0, 0));
        }
        
        [TestMethod]
        public void FormulaCellValueAsStringTestArraySpaces()
        {
            string testValue = "=- {1} + ({2})";
            AssertFormulaCellValueAsString(testValue, testValue, new CellIndex(0, 0));
        }

        [TestMethod]
        public void FormulaCellValueAsStringTestAmpersand()
        {
            string testValue = "=\"string\" & \"\"";
            AssertFormulaCellValueAsString(testValue, testValue, new CellIndex(0, 0));
        }

        [TestMethod]
        public void FormulaCellValueAsStringTestEmptyStrings()
        {
            string testValue = "=\"\" & \"\" & \"\"";
            AssertFormulaCellValueAsString(testValue, testValue, new CellIndex(0, 0));
        }

        [TestMethod]
        public void FormulaCellValueAsStringTestCellRanges1()
        {
            string testValue = "=Sheet1!A1";
            AssertFormulaCellValueAsString(testValue, testValue, new CellIndex(1, 1));
        }

        [TestMethod]
        public void FormulaCellValueAsStringTestCellRanges2()
        {
            string testValue = "='Sheet1'!A1";
            AssertFormulaCellValueAsString(testValue, testValue, new CellIndex(1, 1));
        }

        [TestMethod]
        public void FormulaCellValueAsStringTestCellRanges3()
        {
            string testValue = "=sHEET1!a1";
            string expectedValue = "=Sheet1!A1";
            AssertFormulaCellValueAsString(expectedValue, testValue, new CellIndex(1, 1));
        }

        [TestMethod]
        public void FormulaCellValueAsStringTestCellRanges4()
        {
            string testValue = "=sHEET1!a1";
            string expectedValue = "=Sheet1!A1";
            AssertFormulaCellValueAsString(expectedValue, testValue, new CellIndex(1, 1));
        }

        [TestMethod]
        public void FormulaCellValueAsStringTestCellRanges5()
        {
            string testValue = "='sHEET1'!a1";
            string expectedValue = "='Sheet1'!A1";
            AssertFormulaCellValueAsString(expectedValue, testValue, new CellIndex(1, 1));
        }

        [TestMethod]
        public void FormulaCellValueAsStringTestCellRanges6()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Name = "Sam's Worksheet";

            string testValue = "='Sam''s Worksheet'!A1";
            AssertFormulaCellValueAsString(testValue, testValue, new CellIndex(1, 1), worksheet);
        }

        [TestMethod]
        public void FormulaCellValueAsStringTestCellRanges7()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Name = "Sam's Worksheet";

            string testValue = "='sAM''S wORKSHEET'!a1";
            string expectedValue = "='Sam''s Worksheet'!A1";
            AssertFormulaCellValueAsString(expectedValue, testValue, new CellIndex(1, 1), worksheet);
        }

        [TestMethod]
        public void FormulaCellValueAsStringTestCellRanges8()
        {
            string testValue = "=A1 + (Sheet1!B1)";
            AssertFormulaCellValueAsString(testValue, testValue, new CellIndex(1, 1));
        }

        [TestMethod]
        public void FormulaCellValueAsStringTestFunctions1()
        {
            string testValue = "=sum( abs( average( {ls} ) ) )";
            string expectedValue = "=SUM( ABS( AVERAGE( {ls} ) ) )";
            AssertFormulaCellValueAsString(expectedValue, testValue, new CellIndex(1, 1));
        }

        [TestMethod]
        public void FormulaCellValueAsStringTestFunctions2()
        {
            string testValue = "=error.type( atan2( A1 {ls} B1 ) )";
            string expectedValue = "=ERROR.TYPE( ATAN2( A1 {ls} B1 ) )";
            AssertFormulaCellValueAsString(expectedValue, testValue, new CellIndex(1, 1));
        }

        [TestMethod]
        public void FormulaCellValueAsStringTestFunctions3()
        {
            string testValue = "=concatenate( \"abc\" & \"def\" )";
            string expectedValue = "=CONCATENATE( \"abc\" & \"def\" )";
            AssertFormulaCellValueAsString(expectedValue, testValue, new CellIndex(1, 1));
        }

        [TestMethod]
        public void FormulaCellValueAsStringTestUnionOperator()
        {
            string testValue = "=SUM((A1:B1){ls} (B1))";
            AssertFormulaCellValueAsString(testValue, testValue, new CellIndex(1, 1));
        }

        [TestMethod]
        public void CyclicReferenceInFunctionsTest1()
        {
            AssertReferringToCyclicReference("=A1");
        }

        [TestMethod]
        public void CyclicReferenceInFunctionsTest2()
        {
            AssertReferringToCyclicReference("=B1");
        }

        [TestMethod]
        public void CyclicReferenceInOneCellUnion()
        {
            AssertReferringToCyclicReference("=A1{ls}A1");
        }

        [TestMethod]
        public void CyclicReferenceInUnion()
        {
            AssertReferringToCyclicReference("=A1{ls}B1");
        }

        [TestMethod]
        public void CyclicReferenceInFunctionsTest3()
        {
            AssertReferringToCyclicReference("=SUM(5{ls} A1)");
        }

        [TestMethod]
        public void CyclicReferenceUnionInFunction()
        {
            AssertReferringToCyclicReference("=SUM((A1{ls}B1))");
        }

        [TestMethod]
        public void CyclicReferenceInFunctionsCellReferenceTest()
        {
            AssertReferringToCyclicReference("=SUM(B1)");
        }

        [TestMethod]
        public void CyclicReferenceInFunctionsCellReferenceTest2()
        {
            AssertCyclicReference("=SUM(A1)", new CellIndex(0, 0));
        }

        [TestMethod]
        public void CyclicReferenceInFunctionCellReferenceRangeTest()
        {
            AssertCyclicReference("=SUM(A1:A1)", new CellIndex(0, 0));
        }

        [TestMethod]
        public void CyclicReferenceInFunctionIntersectionTest()
        {
            AssertCyclicReference("=SUM(A1 A1)", new CellIndex(0, 0));
        }

        [TestMethod]
        public void CyclicReferenceInFunctionOneCellUnionTest()
        {
            AssertCyclicReference("=SUM((A1{ls}A1))", new CellIndex(0, 0));
        }

        [TestMethod]
        public void CyclicReferenceInFunctionTwoCellUnionTest()
        {
            AssertCyclicReference("=SUM((A1{ls}B1))", new CellIndex(0, 1));
        }

        [TestMethod]
        public void CyclicReferenceCellRefToItselfTest()
        {
            AssertCyclicReference("=A1", new CellIndex(0, 0));
        }

        [TestMethod]
        public void CyclicReferenceCellRefRangeTest()
        {
            AssertCyclicReference("=A1:A1", new CellIndex(0, 0));
        }

        [TestMethod]
        public void CyclicReferenceCellRefRange2Test()
        {
            AssertCyclicReference("=A1:B1", new CellIndex(0, 0));
        }

        [TestMethod]
        public void CyclicReferenceCellRefRangeEndCellTest()
        {
            AssertCyclicReference("=A1:B1", new CellIndex(0, 1));
        }

        [TestMethod]
        public void CyclicReferenceCellRefRangeMiddleCellTest()
        {
            AssertCyclicReference("=A1:C1", new CellIndex(0, 1));
        }

        [TestMethod]
        public void CyclicReferenceCellRefRange4Test()
        {
            AssertCyclicReference("=A1:B2", new CellIndex(0, 0));
        }

        [TestMethod]
        public void CyclicReferenceCellRefRange1ToItselfTest()
        {
            AssertCyclicReference("=C3:C3", new CellIndex(2, 2));
        }

        [TestMethod]
        public void CyclicReferenceCellRefRangeToItselfTest()
        {
            AssertCyclicReference("=C3:D5", new CellIndex(2, 2));
        }
         
        [TestMethod]
        public void CyclicReferenceUnnormalizedCellRefRangeToItselfTest()
        {
            AssertCyclicReference("=C5:E3", new CellIndex(2, 2));
        }

        [TestMethod]
        public void CyclicReferenceIntersectionToItselfTest()
        {
            AssertCyclicReference("=A1 A1", new CellIndex(0, 0));
        }

        [TestMethod]
        public void CyclicReferenceUnionToItselfTest()
        {
            AssertCyclicReference("=A1{ls}A1", new CellIndex(0, 0));
        }

        [TestMethod]
        public void CyclicReferenceRange1AndIntersection()
        {
            AssertCyclicReference("=A1:A1 A1", new CellIndex(0, 0));
        }

        [TestMethod]
        public void CyclicReferenceRangeAndUnion()
        {
            AssertCyclicReference("=A1:A1{ls}A1", new CellIndex(0, 0));
        }

        [TestMethod]
        public void CyclicReferenceTwoCellRangeAndUnion()
        {
            AssertCyclicReference("=A1:B1{ls}A1", new CellIndex(0, 0));
        }

        [TestMethod]
        public void CyclicReferenceRange2AndIntersection2()
        {
            AssertCyclicReference("=A1:B1 B1", new CellIndex(0, 1));
        }

        [TestMethod]
        public void CyclicReferenceTwoCellRangeAndUnion2()
        {
            AssertCyclicReference("=A1:B1{ls}B1", new CellIndex(0, 1));
        }

        [TestMethod]
        public void CyclicReferenceTwoCellRangeAndUnion3()
        {
            AssertCyclicReference("=A1:B1{ls}C1:D1", new CellIndex(0, 3));
        }

        [TestMethod]
        public void CyclicReferenceIntersectionOfRanges()
        {
            AssertCyclicReference("=A1:B1 B1:C1", new CellIndex(0, 1));
        }

        [TestMethod]
        public void CyclicReferenceUnionOfRanges()
        {
            AssertCyclicReference("=A1:B1{ls}B1:C1", new CellIndex(0, 1));
        }

		[TestMethod]
		public void SimpleFormulaTest()
		{
            Worksheet worksheet = TestHelper.CreateWorksheet();

			worksheet.Cells[0, 0].SetValue("= 1 + 3");
			AssertCellFormula(worksheet, 0, 0, 4);
		}

		[TestMethod]
		public void CellDependencyTest()
		{
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Cells[0, 0].SetValue("5");
            worksheet.Cells[0, 1].SetValue("= 3 + A1");
			AssertCellFormula(worksheet, 0, 1, 8);
		}

		[TestMethod]
		public void CellDependencyToAnotherCellWithNumberTest()
		{
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Cells[0, 0].SetValue("5");
            worksheet.Cells[0, 1].SetValue("= 3 + A1");
			AssertCellFormula(worksheet, 0, 1, 8);

            worksheet.Cells[0, 2].SetValue("= 3 + B1");
			AssertCellFormula(worksheet, 0, 2, 11);

            worksheet.Cells[0, 0].SetValue("1");
			AssertCellFormula(worksheet, 0, 1, 4);
			AssertCellFormula(worksheet, 0, 2, 7);
		}

		[TestMethod]
		public void NCellDependencyToAnotherCellTest()
		{
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Cells[0, 0].SetValue("5");
            worksheet.Cells[0, 1].SetValue("= 3 + A1");
			AssertCellFormula(worksheet, 0, 1, 8);

            worksheet.Cells[0, 2].SetValue("= 3 + B1");
            worksheet.Cells[0, 3].SetValue("= 3 + B1");
            worksheet.Cells[0, 4].SetValue("= 3 + B1");

			AssertCellFormula(worksheet, 0, 2, 11);
			AssertCellFormula(worksheet, 0, 3, 11);
			AssertCellFormula(worksheet, 0, 4, 11);

            worksheet.Cells[0, 0].SetValue("1");
			AssertCellFormula(worksheet, 0, 1, 4);
			AssertCellFormula(worksheet, 0, 2, 7);
			AssertCellFormula(worksheet, 0, 3, 7);
			AssertCellFormula(worksheet, 0, 4, 7);
		}

		[TestMethod]
		public void CellChainDependencyChangeLeafTest()
		{
			Worksheet worksheet = TestHelper.CreateWorksheet();
			worksheet.Workbook.History.IsEnabled = false;

            worksheet.Cells[0, 0].SetValue("5");
			for (int i = 1; i <= 2000; i++)
			{
                worksheet.Cells[i, 0].SetValue("= 1 + A" + i);
				AssertCellFormula(worksheet, i, 0, 5 + i);
			}

            worksheet.Cells[0, 0].SetValue("15");
            for (int i = 1; i <= 2000; i++)
            {
                AssertCellFormula(worksheet, i, 0, 15 + i);
            }
		}

		[TestMethod]
		public void CellChainDependencyChangeInnerNodeTest()
		{
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Workbook.History.IsEnabled = false;

            worksheet.Cells[0, 0].SetValue("5");
			for (int i = 1; i <= 5; i++)
			{
                worksheet.Cells[i, 0].SetValue("= 1 + A" + i);
				AssertCellFormula(worksheet, i, 0, 5 + i);
			}

            worksheet.Cells[3, 0].SetValue("3");

			AssertCellFormula(worksheet, 4, 0, 4);
			AssertCellFormula(worksheet, 5, 0, 5);
		}

        [TestMethod]
        public void CellValueTextValueInArrayFormulaTest()
        {
            string testValue = "=SUM({\"a\"{als}1{als}\"value\"\"with\"\"quotes\"{als}\"\"\"\"})";
            AssertFormulaCellValueAsString(testValue, testValue, new CellIndex(0, 0));
        }

		#endregion
	}
}