﻿using System;
using System.Windows.Media;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Theming;
using System.Collections.Generic;

namespace Telerik.Windows.Documents.Spreadsheet.Tests
{
    [TestClass]
    public class CopyPasteTests
    {
        #region Nested Types

        private class CellProperties
        {
            #region Properties

            public string Value { get; set; }
            public CellBorders Borders { get; set; }
            public IFill Fill { get; set; }
            public ThemableFontFamily FontFamily { get; set; }
            public double FontSize { get; set; }
            public ThemableColor ForeColor { get; set; }
            public CellValueFormat Format { get; set; }
            public RadHorizontalAlignment HorizontalAlignment { get; set; }
            public int Indent { get; set; }
            public bool IsBold { get; set; }
            public bool IsItalic { get; set; }
            public bool IsWrapped { get; set; }
            public UnderlineType Underline { get; set; }
            public RadVerticalAlignment VerticalAlignment { get; set; }
            public bool IsLocked { get; set; }

            #endregion
        }

        #endregion


        #region Methods

        private void AssertCopyPasteSingleValue(string value, string expectedValue, PasteType pasteType, CellIndex source, CellIndex target)
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            workbook.Worksheets.Add();

            Worksheet worksheet = workbook.Worksheets[0];
            CellSelection selection = worksheet.Cells[source];

            selection.SetValue(value.ReplaceCultureDependentSeparators());
            WorksheetFragment fragment = selection.Copy();

            CellSelection pasteSelection = worksheet.Cells[target];
            pasteSelection.Paste(fragment, new PasteOptions(pasteType));

            string actualValue = pasteSelection.GetValue().Value.RawValue;

            Assert.AreEqual(expectedValue.ReplaceCultureDependentSeparators(), actualValue);

            if (pasteType == PasteType.ColumnWidths)
            {
                Assert.AreEqual(worksheet.Columns[source].GetWidth().Value, worksheet.Columns[target].GetWidth().Value);
            }
        }

        private void InitializeProperties(CellProperties properties)
        {
            properties.Value = "1";
            properties.Borders = new CellBorders(
                new CellBorder(CellBorderStyle.DashDot, new ThemableColor(Colors.Red)),
                new CellBorder(CellBorderStyle.DashDotDot, new ThemableColor(Colors.Blue)),
                new CellBorder(CellBorderStyle.Dashed, new ThemableColor(ThemeColorType.Accent1)),
                new CellBorder(CellBorderStyle.Dotted, new ThemableColor(ThemeColorType.Accent2)),
                null,
                null,
                new CellBorder(CellBorderStyle.Hair, new ThemableColor(ThemeColorType.Accent3)),
                new CellBorder(CellBorderStyle.Medium, new ThemableColor(ThemeColorType.Accent4)));
            properties.Fill = new PatternFill(PatternType.Gray12Percent, Colors.Red, Colors.Blue);
            properties.FontFamily = new ThemableFontFamily("Arial");
            properties.FontSize = 15;
            properties.ForeColor = new ThemableColor(ThemeColorType.Accent5);
            properties.Format = new CellValueFormat("\\$#,##0.00");
            properties.HorizontalAlignment = RadHorizontalAlignment.Left;
            properties.Indent = 3;
            properties.IsBold = true;
            properties.IsItalic = true;
            properties.IsWrapped = true;
            properties.Underline = UnderlineType.Single;
            properties.VerticalAlignment = RadVerticalAlignment.Center;
            properties.IsLocked = false;
        }

        private void SetCellSelectionProperties(CellProperties properties, SelectionBase selection)
        {
            CellSelection cellSelection = selection as CellSelection;

            if(cellSelection != null)
            {
                cellSelection.SetValue(properties.Value);
            }

            selection.SetBorders(properties.Borders);
            selection.SetFill(properties.Fill);
            selection.SetFontFamily(properties.FontFamily);
            selection.SetFontSize(properties.FontSize);
            selection.SetForeColor(properties.ForeColor);
            selection.SetFormat(properties.Format);
            selection.SetHorizontalAlignment(properties.HorizontalAlignment);
            selection.SetIndent(properties.Indent);
            selection.SetIsBold(properties.IsBold);
            selection.SetIsItalic(properties.IsItalic);
            selection.SetIsWrapped(properties.IsWrapped);
            selection.SetUnderline(properties.Underline);
            selection.SetVerticalAlignment(properties.VerticalAlignment);
            selection.SetIsLocked(properties.IsLocked);
        }

        #endregion


        #region Paste All Tests

        [TestMethod]
        public void CopyStyleAndValuePasteAllDifferentWorkbookTest()
        {
            Worksheet fromWorksheet = TestHelper.CreateWorksheet();
            CellStyleCollection fromStyles = fromWorksheet.Workbook.Styles;

            string styleName = "myCustomStyle";
            CellStyle fromStyle = fromStyles.Add(styleName);
            fromStyle.Fill = PatternFill.CreateSolidFill(Colors.Purple);
            fromStyle.ForeColor = new ThemableColor(Colors.Yellow);

            string cellValue = "myTestValue";
            CellSelection fromSelection = fromWorksheet.Cells[1, 1];
            fromSelection.SetValue(cellValue);
            fromSelection.SetStyleName(styleName);

            WorksheetFragment fragment = fromSelection.Copy();

            Worksheet toWorksheet = TestHelper.CreateWorksheet();
            CellStyleCollection toCollection = toWorksheet.Workbook.Styles;
            CellSelection toSelection = toWorksheet.Cells[3, 3];
            toSelection.Paste(fragment, PasteOptions.All);

            string pastedStyleName = toSelection.GetStyleName().Value;
            Assert.AreEqual(styleName, pastedStyleName);

            string pastedValue = toSelection.GetValue().Value.RawValue;
            Assert.AreEqual(cellValue, pastedValue);

            CellStyle pastedStyle = toCollection[styleName];
            Assert.AreEqual(fromStyle, pastedStyle);
        }

        [TestMethod]
        public void CopyStyleAndValuePasteAllTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            string value = "1";
            string styleName = "Good";

            worksheet.Cells[0, 0].SetValue(value);
            worksheet.Cells[0, 0].SetStyleName(styleName);
            CellValueFormat expectedFormat = worksheet.Cells[0, 0].GetFormat().Value;

            WorksheetFragment fragment = worksheet.Cells[0, 0].Copy();
            worksheet.Cells[3, 3].Paste(fragment, new PasteOptions(PasteType.All));

            CellValueFormat format = worksheet.Cells[3, 3].GetFormat().Value;
            Assert.AreEqual(expectedFormat, format);

            string actualValue = worksheet.Cells[3, 3].GetValue().Value.GetValueAsString(format);
            string actualStyleName = worksheet.Cells[3, 3].GetStyleName().Value;

            Assert.AreEqual(value, actualValue);
            Assert.AreEqual(styleName, actualStyleName);
        }

        [TestMethod]
        public void CopyNumberFormatAndValuePasteAllTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            string value = "1";
            string numberFormat = "$0.00";

            worksheet.Cells[0, 0].SetValue(value);
            worksheet.Cells[0, 0].SetFormat(new CellValueFormat(numberFormat));

            WorksheetFragment fragment = worksheet.Cells[0, 0].Copy();
            worksheet.Cells[3, 3].Paste(fragment, new PasteOptions(PasteType.All));

            CellValueFormat format = worksheet.Cells[3, 3].GetFormat().Value;
            Assert.AreEqual(numberFormat, format.FormatString);

            string actualValue = worksheet.Cells[3, 3].GetValue().Value.GetValueAsString(format);

            Assert.AreEqual(value, actualValue);
        }

        [TestMethod]
        public void CopyLocalPropertiesPasteAllTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellProperties properties = new CellProperties();

            this.InitializeProperties(properties);
            CellSelection selection = worksheet.Cells[0, 0];
            this.SetCellSelectionProperties(properties, selection);

            var pasteSelection = worksheet.Cells[3, 3];
            WorksheetFragment fragment = selection.Copy();
            pasteSelection.Paste(fragment, new PasteOptions(PasteType.All));

            string actualValue = pasteSelection.GetValue().Value.GetValueAsString(properties.Format);
            Assert.AreEqual(properties.Value, actualValue);

            CellBorders actualBorders = pasteSelection.GetBorders();
            Assert.AreEqual(properties.Borders, actualBorders);

            IFill actualFill = pasteSelection.GetFill().Value;
            Assert.AreEqual(properties.Fill, actualFill);

            ThemableFontFamily actualFontFamily = pasteSelection.GetFontFamily().Value;
            Assert.AreEqual(properties.FontFamily, actualFontFamily);

            double actualFontSize = pasteSelection.GetFontSize().Value;
            Assert.AreEqual(properties.FontSize, actualFontSize);

            ThemableColor actualForeColor = pasteSelection.GetForeColor().Value;
            Assert.AreEqual(properties.ForeColor, actualForeColor);

            CellValueFormat actualFormat = pasteSelection.GetFormat().Value;
            Assert.AreEqual(properties.Format, actualFormat);

            RadHorizontalAlignment actualHorizontalAlignment = pasteSelection.GetHorizontalAlignment().Value;
            Assert.AreEqual(properties.HorizontalAlignment, actualHorizontalAlignment);

            int actualIndent = pasteSelection.GetIndent().Value;
            Assert.AreEqual(properties.Indent, actualIndent);

            bool actualIsBold = pasteSelection.GetIsBold().Value;
            Assert.AreEqual(properties.IsBold, actualIsBold);

            bool actualIsItalic = pasteSelection.GetIsItalic().Value;
            Assert.AreEqual(properties.IsItalic, actualIsItalic);

            bool actualIsWrapped = pasteSelection.GetIsWrapped().Value;
            Assert.AreEqual(properties.IsWrapped, actualIsWrapped);

            UnderlineType actualUnderline = pasteSelection.GetUnderline().Value;
            Assert.AreEqual(properties.Underline, actualUnderline);

            RadVerticalAlignment actualVerticalAlignment = pasteSelection.GetVerticalAlignment().Value;
            Assert.AreEqual(properties.VerticalAlignment, actualVerticalAlignment);

            bool actualIsLocked = pasteSelection.GetIsLocked().Value;
            Assert.AreEqual(properties.IsLocked, actualIsLocked);
        }

        [TestMethod]
        public void CopyPasteOperatorExpressionWithSimpleQuotes()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(1, 0);
            string value = "=\"Rad\"&\"Spreadsheet\"";
            this.AssertCopyPasteSingleValue(value, value, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyPasteOperatorExpressionWithInnerQuotes()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(1, 0);
            string value = "=\"\"\"Rad\"\"\"&\"\"\"Sprea\"\"dsheet\"\"\"";
            this.AssertCopyPasteSingleValue(value, value, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyPasteFunctionWithSimpleQuotes()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(1, 0);
            string value = "=UPPER(\"abc\")";
            this.AssertCopyPasteSingleValue(value, value, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyPasteFunctionWithInnerQuotes()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(1, 0);
            string value = "=CONCATENATE(\"a\"{ls}\"\"{ls}\"b\"\"c\")";
            this.AssertCopyPasteSingleValue(value, value, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyPasteOperatorExpressionWithSimpleQuotesAndCellReference()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(1, 0);
            string value = "=\"3\"&B10";
            string expectedResult = "=\"3\"&B11";
            this.AssertCopyPasteSingleValue(value, expectedResult, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyPasteFunctionWithInnerQuotesAndCellReference()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(1, 0);
            string value = "=CONCATENATE(\"\"\"this\"\"\"{ls}\" is \"{ls}B1{ls}B2)";
            string expectedResult = "=CONCATENATE(\"\"\"this\"\"\"{ls}\" is \"{ls}B2{ls}B3)";
            this.AssertCopyPasteSingleValue(value, expectedResult, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyPasteTwoFunctionsWithInnerQuotesAndCellReference()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(1, 0);
            string value = "=CONCATENATE(\"\"\"this\"\"\"{ls}\" is \"{ls}B1{ls}B2)&CONCATENATE(\"\"\"this\"\"\"{ls}\" is \"{ls}B1{ls}B2)";
            string expectedResult = "=CONCATENATE(\"\"\"this\"\"\"{ls}\" is \"{ls}B2{ls}B3)&CONCATENATE(\"\"\"this\"\"\"{ls}\" is \"{ls}B2{ls}B3)";
            this.AssertCopyPasteSingleValue(value, expectedResult, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyPasteReferenceError()
        {
            CellIndex sourceIndex = new CellIndex(1, 0);
            CellIndex targetIndex = new CellIndex(0, 0);
            string value = "=\"3\"&B1";
            string expectedResult = "=\"3\"&#REF!";
            this.AssertCopyPasteSingleValue(value, expectedResult, PasteType.All, sourceIndex, targetIndex);
        }
        [TestMethod]
        public void CopyFormulaPasteAllTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=SUM(1,5)";
            this.AssertCopyPasteSingleValue(value, value, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteAllTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=D1";
            string expectedValue = "=G4";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteAll1Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=$D1";
            string expectedValue = "=$D4";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteAll2Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=D$1";
            string expectedValue = "=G$1";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteAll3Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=$D$1";
            string expectedValue = "=$D$1";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopySheetAndCellReferenceDifferentSheetPasteAllTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=Sheet2!A1";
            string expectedValue = "=Sheet2!D4";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyQuotedStringPasteAllTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "\"ab\"c";
            string expectedValue = "\"ab\"c";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingQuotesPasteAllTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=CONCATENATE(\"abs\"{ls}\"bf\")";
            string expectedValue = "=CONCATENATE(\"abs\"{ls}\"bf\")";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsPasteAllTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=A+B";
            string expectedValue = "=A+B";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteAllTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=(A+B)";
            string expectedValue = "=(A+B)";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteAll1Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=(1+2)";
            string expectedValue = "=(1+2)";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteAll2Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=(1+2)*3";
            string expectedValue = "=(1+2)*3";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteAll3Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=((1+2)*3)";
            string expectedValue = "=((1+2)*3)";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.All, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaAcrossWorkbookTest()
        {
            Worksheet sourceWorksheet = TestHelper.CreateWorksheet();
            Worksheet targetWorksheet = TestHelper.CreateWorksheet();

            sourceWorksheet.Cells[0, 1].SetValue("1");
            targetWorksheet.Cells[0, 1].SetValue("2");

            sourceWorksheet.Cells[0, 0].SetValue("=B1");
            WorksheetFragment fragment = sourceWorksheet.Cells[0, 0].Copy();
            targetWorksheet.Cells[0, 0].Paste(fragment, PasteOptions.All);

            string contentsOfA1 = targetWorksheet.Cells[0, 0].GetValue().Value.GetResultValueAsString(CellValueFormat.GeneralFormat);

            Assert.AreEqual("2", contentsOfA1);
        }

        #endregion


        #region Paste Formats Tests

        [TestMethod]
        public void CopyStyleAndValuePasteFormatsDifferentWorkbooksTest()
        {
            Worksheet fromWorksheet = TestHelper.CreateWorksheet();
            CellStyleCollection fromStyles = fromWorksheet.Workbook.Styles;

            string styleName = "myCustomStyle";
            CellStyle fromStyle = fromStyles.Add(styleName);
            fromStyle.Fill = PatternFill.CreateSolidFill(Colors.Purple);
            fromStyle.ForeColor = new ThemableColor(Colors.Yellow);

            string cellValue = "myTestValue";
            CellSelection fromSelection = fromWorksheet.Cells[1, 1];
            fromSelection.SetValue(cellValue);
            fromSelection.SetStyleName(styleName);

            WorksheetFragment fragment = fromSelection.Copy();

            Worksheet toWorksheet = TestHelper.CreateWorksheet();
            CellStyleCollection toCollection = toWorksheet.Workbook.Styles;
            CellSelection toSelection = toWorksheet.Cells[3, 3];
            toSelection.Paste(fragment, new PasteOptions(PasteType.Formats));

            string pastedStyleName = toSelection.GetStyleName().Value;
            Assert.AreEqual(styleName, pastedStyleName);

            string expectedValue = string.Empty;
            string pastedValue = toSelection.GetValue().Value.RawValue;
            Assert.AreEqual(expectedValue, pastedValue);

            CellStyle pastedStyle = toCollection[styleName];
            Assert.AreEqual(fromStyle, pastedStyle);
        }

        [TestMethod]
        public void CopyStyleAndValuePasteFormatsTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            string value = "1";
            string styleName = "Good";

            worksheet.Cells[0, 0].SetValue(value);
            worksheet.Cells[0, 0].SetStyleName(styleName);
            CellValueFormat expectedFormat = worksheet.Cells[0, 0].GetFormat().Value;

            WorksheetFragment fragment = worksheet.Cells[0, 0].Copy();
            worksheet.Cells[3, 3].Paste(fragment, new PasteOptions(PasteType.Formats));

            CellValueFormat format = worksheet.Cells[3, 3].GetFormat().Value;
            Assert.AreEqual(expectedFormat, format);

            string actualValue = worksheet.Cells[3, 3].GetValue().Value.GetValueAsString(format);
            string actualStyleName = worksheet.Cells[3, 3].GetStyleName().Value;

            string expectedValue = string.Empty;
            Assert.AreEqual(expectedValue, actualValue);
            Assert.AreEqual(styleName, actualStyleName);
        }

        [TestMethod]
        public void CopyNumberFormatAndValuePasteFormatsTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            string value = "1";
            string numberFormat = "$0.00";

            worksheet.Cells[0, 0].SetValue(value);
            worksheet.Cells[0, 0].SetFormat(new CellValueFormat(numberFormat));

            WorksheetFragment fragment = worksheet.Cells[0, 0].Copy();
            worksheet.Cells[3, 3].Paste(fragment, new PasteOptions(PasteType.Formats));

            CellValueFormat format = worksheet.Cells[3, 3].GetFormat().Value;
            Assert.AreEqual(numberFormat, format.FormatString);

            string actualValue = worksheet.Cells[3, 3].GetValue().Value.GetValueAsString(format);

            string expectedValue = string.Empty;
            Assert.AreEqual(expectedValue, actualValue);
        }

        [TestMethod]
        public void CopyLocalPropertiesPasteFormatsTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellProperties properties = new CellProperties();

            this.InitializeProperties(properties);
            CellSelection selection = worksheet.Cells[0, 0];
            this.SetCellSelectionProperties(properties, selection);

            var pasteSelection = worksheet.Cells[3, 3];
            WorksheetFragment fragment = selection.Copy();
            pasteSelection.Paste(fragment, new PasteOptions(PasteType.Formats));

            string actualValue = pasteSelection.GetValue().Value.GetValueAsString(properties.Format);
            string expectedValue = string.Empty;
            Assert.AreEqual(expectedValue, actualValue);

            CellBorders actualBorders = pasteSelection.GetBorders();
            Assert.AreEqual(properties.Borders, actualBorders);

            IFill actualFill = pasteSelection.GetFill().Value;
            Assert.AreEqual(properties.Fill, actualFill);

            ThemableFontFamily actualFontFamily = pasteSelection.GetFontFamily().Value;
            Assert.AreEqual(properties.FontFamily, actualFontFamily);

            double actualFontSize = pasteSelection.GetFontSize().Value;
            Assert.AreEqual(properties.FontSize, actualFontSize);

            ThemableColor actualForeColor = pasteSelection.GetForeColor().Value;
            Assert.AreEqual(properties.ForeColor, actualForeColor);

            CellValueFormat actualFormat = pasteSelection.GetFormat().Value;
            Assert.AreEqual(properties.Format, actualFormat);

            RadHorizontalAlignment actualHorizontalAlignment = pasteSelection.GetHorizontalAlignment().Value;
            Assert.AreEqual(properties.HorizontalAlignment, actualHorizontalAlignment);

            int actualIndent = pasteSelection.GetIndent().Value;
            Assert.AreEqual(properties.Indent, actualIndent);

            bool actualIsBold = pasteSelection.GetIsBold().Value;
            Assert.AreEqual(properties.IsBold, actualIsBold);

            bool actualIsItalic = pasteSelection.GetIsItalic().Value;
            Assert.AreEqual(properties.IsItalic, actualIsItalic);

            bool actualIsWrapped = pasteSelection.GetIsWrapped().Value;
            Assert.AreEqual(properties.IsWrapped, actualIsWrapped);

            UnderlineType actualUnderline = pasteSelection.GetUnderline().Value;
            Assert.AreEqual(properties.Underline, actualUnderline);

            RadVerticalAlignment actualVerticalAlignment = pasteSelection.GetVerticalAlignment().Value;
            Assert.AreEqual(properties.VerticalAlignment, actualVerticalAlignment);

            bool actualIsLocked = pasteSelection.GetIsLocked().Value;
            Assert.AreEqual(properties.IsLocked, actualIsLocked);
        }

        [TestMethod]
        public void CopyFormulaPasteFormatsTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=SUM(1,5)";
            string expectedValue = string.Empty;
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Formats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormattingWhenSourceColumnHasFormatting()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            CellProperties columnProperties = new CellProperties();
            this.InitializeProperties(columnProperties);
            columnProperties.Borders = new CellBorders(new CellBorder(CellBorderStyle.Thin, new ThemableColor(Colors.Red)));
            columnProperties.Fill = new PatternFill(PatternType.Solid, Colors.Yellow, Colors.Yellow);
            columnProperties.ForeColor = new ThemableColor(Colors.Red);
            ColumnSelection column = worksheet.Columns[0];
            this.SetCellSelectionProperties(columnProperties, column);

            CellProperties cellProperties = new CellProperties();
            this.InitializeProperties(cellProperties);
            CellSelection cellSelection = worksheet.Cells[0, 0];
            this.SetCellSelectionProperties(cellProperties, cellSelection);

            var pasteSelection = worksheet.Cells[3, 3];
            WorksheetFragment fragment = cellSelection.Copy();
            pasteSelection.Paste(fragment, PasteOptions.All);

            string actualValue = pasteSelection.GetValue().Value.GetValueAsString(cellProperties.Format);
            Assert.AreEqual(cellProperties.Value, actualValue);

            CellBorders actualBorders = pasteSelection.GetBorders();
            Assert.AreEqual(cellProperties.Borders, actualBorders);

            IFill actualFill = pasteSelection.GetFill().Value;
            Assert.AreEqual(cellProperties.Fill, actualFill);

            ThemableColor actualForeColor = pasteSelection.GetForeColor().Value;
            Assert.AreEqual(cellProperties.ForeColor, actualForeColor);
        }

        #endregion


        #region Paste Formulas Tests

        [TestMethod]
        public void CopyStyleAndValuePasteFormulasTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            string value = "1";
            string styleName = "Good";

            worksheet.Cells[0, 0].SetValue(value);
            worksheet.Cells[0, 0].SetStyleName(styleName);
            CellValueFormat expectedFormat = worksheet.Cells[0, 0].GetFormat().Value;

            WorksheetFragment fragment = worksheet.Cells[0, 0].Copy();
            worksheet.Cells[3, 3].Paste(fragment, new PasteOptions(PasteType.Formulas));

            CellValueFormat format = worksheet.Cells[3, 3].GetFormat().Value;
            Assert.AreEqual(expectedFormat, format);

            string actualValue = worksheet.Cells[3, 3].GetValue().Value.GetValueAsString(format);
            string actualStyleName = worksheet.Cells[3, 3].GetStyleName().Value;
            string expectedStyleName = SpreadsheetDefaultValues.DefaultStyleName;

            Assert.AreEqual(value, actualValue);
            Assert.AreEqual(expectedStyleName, actualStyleName);
        }

        [TestMethod]
        public void CopyNumberFormatAndValuePasteFormulasTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            string value = "1";
            string numberFormat = "$0.00";

            worksheet.Cells[0, 0].SetValue(value);
            worksheet.Cells[0, 0].SetFormat(new CellValueFormat(numberFormat));

            WorksheetFragment fragment = worksheet.Cells[0, 0].Copy();
            worksheet.Cells[3, 3].Paste(fragment, new PasteOptions(PasteType.Formulas));

            CellValueFormat format = worksheet.Cells[3, 3].GetFormat().Value;
            string expectedFormat = string.Empty;

            Assert.AreEqual(expectedFormat, format.FormatString);

            string actualValue = worksheet.Cells[3, 3].GetValue().Value.GetValueAsString(format);

            Assert.AreEqual(value, actualValue);
        }

        [TestMethod]
        public void CopyLocalPropertiesPasteFormulasTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellProperties properties = new CellProperties();

            this.InitializeProperties(properties);
            CellSelection selection = worksheet.Cells[0, 0];
            this.SetCellSelectionProperties(properties, selection);

            var pasteSelection = worksheet.Cells[3, 3];
            WorksheetFragment fragment = selection.Copy();
            pasteSelection.Paste(fragment, new PasteOptions(PasteType.Formulas));

            string actualValue = pasteSelection.GetValue().Value.GetValueAsString(properties.Format);
            Assert.AreEqual(properties.Value, actualValue);

            CellBorders actualBorders = pasteSelection.GetBorders();
            CellBorders expectedBorders = CellBorders.Default;
            expectedBorders.InsideHorizontal = null;
            expectedBorders.InsideVertical = null;
            Assert.AreEqual(expectedBorders, actualBorders);

            IFill actualFill = pasteSelection.GetFill().Value;
            IFill expectedFill = PatternFill.Default;
            Assert.AreEqual(expectedFill, actualFill);

            ThemableFontFamily actualFontFamily = pasteSelection.GetFontFamily().Value;
            ThemableFontFamily expectedFontFamily = SpreadsheetDefaultValues.DefaultFontFamily;
            Assert.AreEqual(expectedFontFamily, actualFontFamily);

            double actualFontSize = pasteSelection.GetFontSize().Value;
            double expectedFontSize = SpreadsheetDefaultValues.DefaultFontSize;
            Assert.AreEqual(expectedFontSize, actualFontSize);

            ThemableColor actualForeColor = pasteSelection.GetForeColor().Value;
            ThemableColor expectedForeColor = SpreadsheetDefaultValues.DefaultForeColor;
            Assert.AreEqual(expectedForeColor, actualForeColor);

            CellValueFormat actualFormat = pasteSelection.GetFormat().Value;
            CellValueFormat expectedFormat = CellValueFormat.GeneralFormat;
            Assert.AreEqual(expectedFormat, actualFormat);

            RadHorizontalAlignment actualHorizontalAlignment = pasteSelection.GetHorizontalAlignment().Value;
            RadHorizontalAlignment expectedHorizontalAlignment = RadHorizontalAlignment.General;
            Assert.AreEqual(expectedHorizontalAlignment, actualHorizontalAlignment);

            int actualIndent = pasteSelection.GetIndent().Value;
            int expectedIndent = 0;
            Assert.AreEqual(expectedIndent, actualIndent);

            bool actualIsBold = pasteSelection.GetIsBold().Value;
            bool expectedIsBold = false;
            Assert.AreEqual(expectedIsBold, actualIsBold);

            bool actualIsItalic = pasteSelection.GetIsItalic().Value;
            bool expectedIsItalic = false;
            Assert.AreEqual(expectedIsItalic, actualIsItalic);

            bool actualIsWrapped = pasteSelection.GetIsWrapped().Value;
            bool expectedIsWrapped = false;
            Assert.AreEqual(expectedIsWrapped, actualIsWrapped);

            UnderlineType actualUnderline = pasteSelection.GetUnderline().Value;
            UnderlineType expectedUnderline = UnderlineType.None;
            Assert.AreEqual(expectedUnderline, actualUnderline);

            RadVerticalAlignment actualVerticalAlignment = pasteSelection.GetVerticalAlignment().Value;
            RadVerticalAlignment expectedVerticalAlignment = RadVerticalAlignment.Bottom;
            Assert.AreEqual(expectedVerticalAlignment, actualVerticalAlignment);

            bool actualIsLocked = pasteSelection.GetIsLocked().Value;
            bool expectedIsLocked = true;
            Assert.AreEqual(expectedIsLocked, actualIsLocked);
        }

        [TestMethod]
        public void CopyFormulaPasteFormulasTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=SUM(1,5)";
            this.AssertCopyPasteSingleValue(value, value, PasteType.Formulas, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteFormulasTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=D1";
            string expectedValue = "=G4";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Formulas, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteFormulas1Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=$D1";
            string expectedValue = "=$D4";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Formulas, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteFormulas2Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=D$1";
            string expectedValue = "=G$1";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Formulas, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteFormulas3Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=$D$1";
            string expectedValue = "=$D$1";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Formulas, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopySheetAndCellReferenceDifferentSheetPasteFormulasTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=Sheet2!A1";
            string expectedValue = "=Sheet2!D4";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Formulas, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyQuotedStringPasteFormulasTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "\"ab\"c";
            string expectedValue = "\"ab\"c";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Formulas, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingQuotesPasteFormulasTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=CONCATENATE(\"abs\"{ls}\"bf\")";
            string expectedValue = "=CONCATENATE(\"abs\"{ls}\"bf\")";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Formulas, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsPasteFormulasTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=A+B";
            string expectedValue = "=A+B";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Formulas, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteFormulasTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=(A+B)";
            string expectedValue = "=(A+B)";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Formulas, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteFormulas1Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=(1+2)";
            string expectedValue = "=(1+2)";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Formulas, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteFormulas2Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=(1+2)*3";
            string expectedValue = "=(1+2)*3";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Formulas, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteFormulas3Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=((1+2)*3)";
            string expectedValue = "=((1+2)*3)";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Formulas, sourceIndex, targetIndex);
        }

        #endregion


        #region Paste FormulasAndNumberFormats Tests

        [TestMethod]
        public void CopyStyleAndValuePasteFormulasAndNumberFormatsTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            string value = "1";
            string styleName = "Good";

            worksheet.Cells[0, 0].SetValue(value);
            worksheet.Cells[0, 0].SetStyleName(styleName);
            CellValueFormat expectedFormat = worksheet.Cells[0, 0].GetFormat().Value;

            WorksheetFragment fragment = worksheet.Cells[0, 0].Copy();
            worksheet.Cells[3, 3].Paste(fragment, new PasteOptions(PasteType.FormulasAndNumberFormats));

            CellValueFormat format = worksheet.Cells[3, 3].GetFormat().Value;
            Assert.AreEqual(expectedFormat, format);

            string actualValue = worksheet.Cells[3, 3].GetValue().Value.GetValueAsString(format);
            string actualStyleName = worksheet.Cells[3, 3].GetStyleName().Value;
            string expectedStyleName = SpreadsheetDefaultValues.DefaultStyleName;

            Assert.AreEqual(value, actualValue);
            Assert.AreEqual(expectedStyleName, actualStyleName);
        }

        [TestMethod]
        public void CopyNumberFormatAndValuePasteFormulasAndNumberFormatsTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            string value = "1";
            string numberFormat = "$0.00";

            worksheet.Cells[0, 0].SetValue(value);
            worksheet.Cells[0, 0].SetFormat(new CellValueFormat(numberFormat));

            WorksheetFragment fragment = worksheet.Cells[0, 0].Copy();
            worksheet.Cells[3, 3].Paste(fragment, new PasteOptions(PasteType.FormulasAndNumberFormats));

            CellValueFormat format = worksheet.Cells[3, 3].GetFormat().Value;
            Assert.AreEqual(numberFormat, format.FormatString);

            string actualValue = worksheet.Cells[3, 3].GetValue().Value.GetValueAsString(format);

            Assert.AreEqual(value, actualValue);
        }

        [TestMethod]
        public void CopyLocalPropertiesPasteFormulasAndNumberFormatsTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellProperties properties = new CellProperties();

            this.InitializeProperties(properties);
            CellSelection selection = worksheet.Cells[0, 0];
            this.SetCellSelectionProperties(properties, selection);

            var pasteSelection = worksheet.Cells[3, 3];
            WorksheetFragment fragment = selection.Copy();
            pasteSelection.Paste(fragment, new PasteOptions(PasteType.FormulasAndNumberFormats));

            string actualValue = pasteSelection.GetValue().Value.GetValueAsString(properties.Format);
            Assert.AreEqual(properties.Value, actualValue);

            CellBorders actualBorders = pasteSelection.GetBorders();
            CellBorders expectedBorders = CellBorders.Default;
            expectedBorders.InsideHorizontal = null;
            expectedBorders.InsideVertical = null;
            Assert.AreEqual(expectedBorders, actualBorders);

            IFill actualFill = pasteSelection.GetFill().Value;
            IFill expectedFill = PatternFill.Default;
            Assert.AreEqual(expectedFill, actualFill);

            ThemableFontFamily actualFontFamily = pasteSelection.GetFontFamily().Value;
            ThemableFontFamily expectedFontFamily = SpreadsheetDefaultValues.DefaultFontFamily;
            Assert.AreEqual(expectedFontFamily, actualFontFamily);

            double actualFontSize = pasteSelection.GetFontSize().Value;
            double expectedFontSize = SpreadsheetDefaultValues.DefaultFontSize;
            Assert.AreEqual(expectedFontSize, actualFontSize);

            ThemableColor actualForeColor = pasteSelection.GetForeColor().Value;
            ThemableColor expectedForeColor = SpreadsheetDefaultValues.DefaultForeColor;
            Assert.AreEqual(expectedForeColor, actualForeColor);

            CellValueFormat actualFormat = pasteSelection.GetFormat().Value;
            Assert.AreEqual(properties.Format, actualFormat);

            RadHorizontalAlignment actualHorizontalAlignment = pasteSelection.GetHorizontalAlignment().Value;
            RadHorizontalAlignment expectedHorizontalAlignment = RadHorizontalAlignment.General;
            Assert.AreEqual(expectedHorizontalAlignment, actualHorizontalAlignment);

            int actualIndent = pasteSelection.GetIndent().Value;
            int expectedIndent = 0;
            Assert.AreEqual(expectedIndent, actualIndent);

            bool actualIsBold = pasteSelection.GetIsBold().Value;
            bool expectedIsBold = false;
            Assert.AreEqual(expectedIsBold, actualIsBold);

            bool actualIsItalic = pasteSelection.GetIsItalic().Value;
            bool expectedIsItalic = false;
            Assert.AreEqual(expectedIsItalic, actualIsItalic);

            bool actualIsWrapped = pasteSelection.GetIsWrapped().Value;
            bool expectedIsWrapped = false;
            Assert.AreEqual(expectedIsWrapped, actualIsWrapped);

            UnderlineType actualUnderline = pasteSelection.GetUnderline().Value;
            UnderlineType expectedUnderline = UnderlineType.None;
            Assert.AreEqual(expectedUnderline, actualUnderline);

            RadVerticalAlignment actualVerticalAlignment = pasteSelection.GetVerticalAlignment().Value;
            RadVerticalAlignment expectedVerticalAlignment = RadVerticalAlignment.Bottom;
            Assert.AreEqual(expectedVerticalAlignment, actualVerticalAlignment);

            bool actualIsLocked = pasteSelection.GetIsLocked().Value;
            bool expectedIsLocked = true;
            Assert.AreEqual(expectedIsLocked, actualIsLocked);
        }

        [TestMethod]
        public void CopyFormulaPasteFormulasAndNumberFormatsTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=SUM(1,5)";
            this.AssertCopyPasteSingleValue(value, value, PasteType.FormulasAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteFormulasAndNumberFormatsTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=D1";
            string expectedValue = "=G4";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.FormulasAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteFormulasAndNumberFormats1Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=$D1";
            string expectedValue = "=$D4";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.FormulasAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteFormulasAndNumberFormats2Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=D$1";
            string expectedValue = "=G$1";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.FormulasAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteFormulasAndNumberFormats3Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=$D$1";
            string expectedValue = "=$D$1";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.FormulasAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopySheetAndCellReferenceDifferentSheetPasteFormulasAndNumberFormatsTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=Sheet2!A1";
            string expectedValue = "=Sheet2!D4";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.FormulasAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyQuotedStringPasteFormulasAndNumberFormatsTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "\"ab\"c";
            string expectedValue = "\"ab\"c";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.FormulasAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingQuotesPasteFormulasAndNumberFormatsTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=CONCATENATE(\"abs\"{ls}\"bf\")";
            string expectedValue = "=CONCATENATE(\"abs\"{ls}\"bf\")";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.FormulasAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsPasteFormulasAndNumberFormatsTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=A+B";
            string expectedValue = "=A+B";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.FormulasAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteFormulasAndNumberFormatsTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=(A+B)";
            string expectedValue = "=(A+B)";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.FormulasAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteFormulasAndNumberFormats1Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=(1+2)";
            string expectedValue = "=(1+2)";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.FormulasAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteFormulasAndNumberFormats2Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=(1+2)*3";
            string expectedValue = "=(1+2)*3";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.FormulasAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteFormulasAndNumberFormats3Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=((1+2)*3)";
            string expectedValue = "=((1+2)*3)";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.FormulasAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaAndStyleAffectingNumberFormatPasteValuesAndNumberFormatsTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            string value = "=((1+2)*3)";
            string styleName = "Percent";

            worksheet.Cells[0, 0].SetValue(value);
            worksheet.Cells[0, 0].SetStyleName(styleName);
            CellValueFormat expectedFormat = new CellValueFormat("0%");

            WorksheetFragment fragment = worksheet.Cells[0, 0].Copy();
            worksheet.Cells[3, 3].Paste(fragment, new PasteOptions(PasteType.ValuesAndNumberFormats));

            CellValueFormat format = worksheet.Cells[3, 3].GetFormat().Value;
            Assert.AreEqual(expectedFormat, format);
        }

        #endregion


        #region Paste Values Tests

        [TestMethod]
        public void CopyStyleAndValuePasteValuesTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            string value = "1";
            string styleName = "Good";

            worksheet.Cells[0, 0].SetValue(value);
            worksheet.Cells[0, 0].SetStyleName(styleName);
            CellValueFormat expectedFormat = worksheet.Cells[0, 0].GetFormat().Value;

            WorksheetFragment fragment = worksheet.Cells[0, 0].Copy();
            worksheet.Cells[3, 3].Paste(fragment, new PasteOptions(PasteType.Values));

            CellValueFormat format = worksheet.Cells[3, 3].GetFormat().Value;
            Assert.AreEqual(expectedFormat, format);

            string actualValue = worksheet.Cells[3, 3].GetValue().Value.GetValueAsString(format);
            string actualStyleName = worksheet.Cells[3, 3].GetStyleName().Value;
            string expectedStyleName = SpreadsheetDefaultValues.DefaultStyleName;

            Assert.AreEqual(value, actualValue);
            Assert.AreEqual(expectedStyleName, actualStyleName);
        }

        [TestMethod]
        public void CopyNumberFormatAndValuePasteValuesTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            string value = "1";
            string numberFormat = "$0.00";

            worksheet.Cells[0, 0].SetValue(value);
            worksheet.Cells[0, 0].SetFormat(new CellValueFormat(numberFormat));

            WorksheetFragment fragment = worksheet.Cells[0, 0].Copy();
            worksheet.Cells[3, 3].Paste(fragment, new PasteOptions(PasteType.Values));

            CellValueFormat format = worksheet.Cells[3, 3].GetFormat().Value;
            CellValueFormat expectedFormat = CellValueFormat.GeneralFormat;
            Assert.AreEqual(expectedFormat, format);

            string actualValue = worksheet.Cells[3, 3].GetValue().Value.GetValueAsString(format);

            Assert.AreEqual(value, actualValue);
        }

        [TestMethod]
        public void CopyLocalPropertiesPasteValuesTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellProperties properties = new CellProperties();

            this.InitializeProperties(properties);
            CellSelection selection = worksheet.Cells[0, 0];
            this.SetCellSelectionProperties(properties, selection);

            var pasteSelection = worksheet.Cells[3, 3];
            WorksheetFragment fragment = selection.Copy();
            pasteSelection.Paste(fragment, new PasteOptions(PasteType.Values));

            string actualValue = pasteSelection.GetValue().Value.GetValueAsString(properties.Format);
            Assert.AreEqual(properties.Value, actualValue);

            CellBorders actualBorders = pasteSelection.GetBorders();
            CellBorders expectedBorders = CellBorders.Default;
            expectedBorders.InsideHorizontal = null;
            expectedBorders.InsideVertical = null;
            Assert.AreEqual(expectedBorders, actualBorders);

            IFill actualFill = pasteSelection.GetFill().Value;
            IFill expectedFill = PatternFill.Default;
            Assert.AreEqual(expectedFill, actualFill);

            ThemableFontFamily actualFontFamily = pasteSelection.GetFontFamily().Value;
            ThemableFontFamily expectedFontFamily = SpreadsheetDefaultValues.DefaultFontFamily;
            Assert.AreEqual(expectedFontFamily, actualFontFamily);

            double actualFontSize = pasteSelection.GetFontSize().Value;
            double expectedFontSize = SpreadsheetDefaultValues.DefaultFontSize;
            Assert.AreEqual(expectedFontSize, actualFontSize);

            ThemableColor actualForeColor = pasteSelection.GetForeColor().Value;
            ThemableColor expectedForeColor = SpreadsheetDefaultValues.DefaultForeColor;
            Assert.AreEqual(expectedForeColor, actualForeColor);

            CellValueFormat actualFormat = pasteSelection.GetFormat().Value;
            CellValueFormat expectedFormat = CellValueFormat.GeneralFormat;
            Assert.AreEqual(expectedFormat, actualFormat);

            RadHorizontalAlignment actualHorizontalAlignment = pasteSelection.GetHorizontalAlignment().Value;
            RadHorizontalAlignment expectedHorizontalAlignment = RadHorizontalAlignment.General;
            Assert.AreEqual(expectedHorizontalAlignment, actualHorizontalAlignment);

            int actualIndent = pasteSelection.GetIndent().Value;
            int expectedIndent = 0;
            Assert.AreEqual(expectedIndent, actualIndent);

            bool actualIsBold = pasteSelection.GetIsBold().Value;
            bool expectedIsBold = false;
            Assert.AreEqual(expectedIsBold, actualIsBold);

            bool actualIsItalic = pasteSelection.GetIsItalic().Value;
            bool expectedIsItalic = false;
            Assert.AreEqual(expectedIsItalic, actualIsItalic);

            bool actualIsWrapped = pasteSelection.GetIsWrapped().Value;
            bool expectedIsWrapped = false;
            Assert.AreEqual(expectedIsWrapped, actualIsWrapped);

            UnderlineType actualUnderline = pasteSelection.GetUnderline().Value;
            UnderlineType expectedUnderline = UnderlineType.None;
            Assert.AreEqual(expectedUnderline, actualUnderline);

            RadVerticalAlignment actualVerticalAlignment = pasteSelection.GetVerticalAlignment().Value;
            RadVerticalAlignment expectedVerticalAlignment = RadVerticalAlignment.Bottom;
            Assert.AreEqual(expectedVerticalAlignment, actualVerticalAlignment);

            bool actualIsLocked = pasteSelection.GetIsLocked().Value;
            bool expectedIsLocked = true;
            Assert.AreEqual(expectedIsLocked, actualIsLocked);
        }

        [TestMethod]
        public void CopyFormulaPasteValuesTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=SUM(1{ls}5)";
            string expectedValue = "6";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Values, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteValuesTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=D1";
            string expectedValue = "0";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Values, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteValues1Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=$D1";
            string expectedValue = "0";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Values, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteValues2Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=D$1";
            string expectedValue = "0";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Values, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteValues3Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=$D$1";
            string expectedValue = "0";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Values, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopySheetAndCellReferenceDifferentSheetPasteValuesTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=Sheet2!A1";
            string expectedValue = "0";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Values, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyQuotedStringPasteValuesTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "\"ab\"c";
            string expectedValue = "\"ab\"c";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Values, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingQuotesPasteValuesTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=CONCATENATE(\"abs\"{ls}\"bf\")";
            string expectedValue = "absbf";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Values, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsPasteValuesTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=A+B";
            string expectedValue = "#NAME?";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Values, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteValuesTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=(A+B)";
            string expectedValue = "#NAME?";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Values, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteValues1Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=(1+2)";
            string expectedValue = "3";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Values, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteValues2Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=(1+2)*3";
            string expectedValue = "9";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Values, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteValues3Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=((1+2)*3)";
            string expectedValue = "9";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.Values, sourceIndex, targetIndex);
        }

        #endregion


        #region Paste ValuesAndNumberFormats Tests

        [TestMethod]
        public void CopyStyleAndValuePasteValuesAndNumberFormatsTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            string value = "1";
            string styleName = "Good";

            worksheet.Cells[0, 0].SetValue(value);
            worksheet.Cells[0, 0].SetStyleName(styleName);
            CellValueFormat expectedFormat = worksheet.Cells[0, 0].GetFormat().Value;

            WorksheetFragment fragment = worksheet.Cells[0, 0].Copy();
            worksheet.Cells[3, 3].Paste(fragment, new PasteOptions(PasteType.ValuesAndNumberFormats));

            CellValueFormat format = worksheet.Cells[3, 3].GetFormat().Value;
            Assert.AreEqual(expectedFormat, format);

            string actualValue = worksheet.Cells[3, 3].GetValue().Value.GetValueAsString(format);
            string actualStyleName = worksheet.Cells[3, 3].GetStyleName().Value;
            string expectedStyleName = SpreadsheetDefaultValues.DefaultStyleName;

            Assert.AreEqual(value, actualValue);
            Assert.AreEqual(expectedStyleName, actualStyleName);
        }

        [TestMethod]
        public void CopyNumberFormatAndValuePasteValuesAndNumberFormatsTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            string value = "1";
            string numberFormat = "$0.00";

            worksheet.Cells[0, 0].SetValue(value);
            worksheet.Cells[0, 0].SetFormat(new CellValueFormat(numberFormat));

            WorksheetFragment fragment = worksheet.Cells[0, 0].Copy();
            worksheet.Cells[3, 3].Paste(fragment, new PasteOptions(PasteType.ValuesAndNumberFormats));

            CellValueFormat format = worksheet.Cells[3, 3].GetFormat().Value;
            Assert.AreEqual(numberFormat, format.FormatString);

            string actualValue = worksheet.Cells[3, 3].GetValue().Value.GetValueAsString(format);

            Assert.AreEqual(value, actualValue);
        }

        [TestMethod]
        public void CopyLocalPropertiesPasteValuesAndNumberFormatsTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellProperties properties = new CellProperties();

            this.InitializeProperties(properties);
            CellSelection selection = worksheet.Cells[0, 0];
            this.SetCellSelectionProperties(properties, selection);

            var pasteSelection = worksheet.Cells[3, 3];
            WorksheetFragment fragment = selection.Copy();
            pasteSelection.Paste(fragment, new PasteOptions(PasteType.ValuesAndNumberFormats));

            string actualValue = pasteSelection.GetValue().Value.GetValueAsString(properties.Format);
            Assert.AreEqual(properties.Value, actualValue);

            CellBorders actualBorders = pasteSelection.GetBorders();
            CellBorders expectedBorders = CellBorders.Default;
            expectedBorders.InsideHorizontal = null;
            expectedBorders.InsideVertical = null;
            Assert.AreEqual(expectedBorders, actualBorders);

            IFill actualFill = pasteSelection.GetFill().Value;
            IFill expectedFill = PatternFill.Default;
            Assert.AreEqual(expectedFill, actualFill);

            ThemableFontFamily actualFontFamily = pasteSelection.GetFontFamily().Value;
            ThemableFontFamily expectedFontFamily = SpreadsheetDefaultValues.DefaultFontFamily;
            Assert.AreEqual(expectedFontFamily, actualFontFamily);

            double actualFontSize = pasteSelection.GetFontSize().Value;
            double expectedFontSize = SpreadsheetDefaultValues.DefaultFontSize;
            Assert.AreEqual(expectedFontSize, actualFontSize);

            ThemableColor actualForeColor = pasteSelection.GetForeColor().Value;
            ThemableColor expectedForeColor = SpreadsheetDefaultValues.DefaultForeColor;
            Assert.AreEqual(expectedForeColor, actualForeColor);

            CellValueFormat actualFormat = pasteSelection.GetFormat().Value;
            Assert.AreEqual(properties.Format, actualFormat);

            RadHorizontalAlignment actualHorizontalAlignment = pasteSelection.GetHorizontalAlignment().Value;
            RadHorizontalAlignment expectedHorizontalAlignment = RadHorizontalAlignment.General;
            Assert.AreEqual(expectedHorizontalAlignment, actualHorizontalAlignment);

            int actualIndent = pasteSelection.GetIndent().Value;
            int expectedIndent = 0;
            Assert.AreEqual(expectedIndent, actualIndent);

            bool actualIsBold = pasteSelection.GetIsBold().Value;
            bool expectedIsBold = false;
            Assert.AreEqual(expectedIsBold, actualIsBold);

            bool actualIsItalic = pasteSelection.GetIsItalic().Value;
            bool expectedIsItalic = false;
            Assert.AreEqual(expectedIsItalic, actualIsItalic);

            bool actualIsWrapped = pasteSelection.GetIsWrapped().Value;
            bool expectedIsWrapped = false;
            Assert.AreEqual(expectedIsWrapped, actualIsWrapped);

            UnderlineType actualUnderline = pasteSelection.GetUnderline().Value;
            UnderlineType expectedUnderline = UnderlineType.None;
            Assert.AreEqual(expectedUnderline, actualUnderline);

            RadVerticalAlignment actualVerticalAlignment = pasteSelection.GetVerticalAlignment().Value;
            RadVerticalAlignment expectedVerticalAlignment = RadVerticalAlignment.Bottom;
            Assert.AreEqual(expectedVerticalAlignment, actualVerticalAlignment);

            bool actualIsLocked = pasteSelection.GetIsLocked().Value;
            bool expectedIsLocked = true;
            Assert.AreEqual(expectedIsLocked, actualIsLocked);
        }

        [TestMethod]
        public void CopyFormulaPasteValuesAndNumberFormatsTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=SUM(1{ls}5)";
            string expectedValue = "6";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.ValuesAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteValuesAndNumberFormatsTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=D1";
            string expectedValue = "0";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.ValuesAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteValuesAndNumberFormats1Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=$D1";
            string expectedValue = "0";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.ValuesAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteValuesAndNumberFormats2Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=D$1";
            string expectedValue = "0";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.ValuesAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteValuesAndNumberFormats3Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=$D$1";
            string expectedValue = "0";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.ValuesAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopySheetAndCellReferenceDifferentSheetPasteValuesAndNumberFormatsTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=Sheet2!A1";
            string expectedValue = "0";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.ValuesAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyQuotedStringPasteValuesAndNumberFormatsTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "\"ab\"c";
            string expectedValue = "\"ab\"c";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.ValuesAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingQuotesPasteValuesAndNumberFormatsTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=CONCATENATE(\"abs\"{ls}\"bf\")";
            string expectedValue = "absbf";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.ValuesAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsPasteValuesAndNumberFormatsTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=A+B";
            string expectedValue = "#NAME?";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.ValuesAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteValuesAndNumberFormatsTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=(A+B)";
            string expectedValue = "#NAME?";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.ValuesAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteValuesAndNumberFormats1Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=(1+2)";
            string expectedValue = "3";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.ValuesAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteValuesAndNumberFormats2Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=(1+2)*3";
            string expectedValue = "9";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.ValuesAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyFormulaContainingMissingConstantsAndBracketsPasteValuesAndNumberFormats3Test()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=((1+2)*3)";
            string expectedValue = "9";
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.ValuesAndNumberFormats, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyValueAndStyleAffectingNumberFormatPasteValuesAndNumberFormatsTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            string value = "1";
            string styleName = "Percent";

            worksheet.Cells[0, 0].SetValue(value);
            worksheet.Cells[0, 0].SetStyleName(styleName);
            CellValueFormat expectedFormat = new CellValueFormat("0%");

            WorksheetFragment fragment = worksheet.Cells[0, 0].Copy();
            worksheet.Cells[3, 3].Paste(fragment, new PasteOptions(PasteType.ValuesAndNumberFormats));

            CellValueFormat format = worksheet.Cells[3, 3].GetFormat().Value;
            Assert.AreEqual(expectedFormat, format);
        }

        #endregion


        #region Paste ColumnWidths Tests

        [TestMethod]
        public void CopyStyleAndValuePasteColumnWidthsTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            string value = "1";
            string styleName = "Good";

            worksheet.Cells[0, 0].SetValue(value);
            worksheet.Cells[0, 0].SetStyleName(styleName);
            CellValueFormat expectedFormat = worksheet.Cells[0, 0].GetFormat().Value;

            WorksheetFragment fragment = worksheet.Cells[0, 0].Copy();
            worksheet.Cells[3, 3].Paste(fragment, new PasteOptions(PasteType.ColumnWidths));

            CellValueFormat format = worksheet.Cells[3, 3].GetFormat().Value;
            Assert.AreEqual(expectedFormat, format);

            string actualValue = worksheet.Cells[3, 3].GetValue().Value.GetValueAsString(format);
            string expectedValue = string.Empty;
            string actualStyleName = worksheet.Cells[3, 3].GetStyleName().Value;
            string expectedStyleName = SpreadsheetDefaultValues.DefaultStyleName;

            Assert.AreEqual(expectedValue, actualValue);
            Assert.AreEqual(expectedStyleName, actualStyleName);
            Assert.AreEqual(worksheet.Columns[0, 0].GetWidth().Value, worksheet.Columns[3, 3].GetWidth().Value);
        }

        [TestMethod]
        public void CopyNumberFormatAndValuePasteColumnWidthsTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            string value = "1";
            string numberFormat = "$0.00";

            worksheet.Cells[0, 0].SetValue(value);
            worksheet.Cells[0, 0].SetFormat(new CellValueFormat(numberFormat));

            WorksheetFragment fragment = worksheet.Cells[0, 0].Copy();
            worksheet.Cells[3, 3].Paste(fragment, new PasteOptions(PasteType.ColumnWidths));

            CellValueFormat format = worksheet.Cells[3, 3].GetFormat().Value;
            CellValueFormat expectedFormat = CellValueFormat.GeneralFormat;
            Assert.AreEqual(expectedFormat, format);

            string actualValue = worksheet.Cells[3, 3].GetValue().Value.GetValueAsString(format);
            string expectedValue = string.Empty;
            Assert.AreEqual(expectedValue, actualValue);

            Assert.AreEqual(worksheet.Columns[0, 0].GetWidth().Value, worksheet.Columns[3, 3].GetWidth().Value);
        }

        [TestMethod]
        public void CopyLocalPropertiesPasteColumnWidthsTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellProperties properties = new CellProperties();

            this.InitializeProperties(properties);
            CellSelection selection = worksheet.Cells[0, 0];
            this.SetCellSelectionProperties(properties, selection);

            var pasteSelection = worksheet.Cells[3, 3];
            WorksheetFragment fragment = selection.Copy();
            pasteSelection.Paste(fragment, new PasteOptions(PasteType.ColumnWidths));

            string actualValue = pasteSelection.GetValue().Value.GetValueAsString(properties.Format);
            string expectedValue = string.Empty;
            Assert.AreEqual(expectedValue, actualValue);

            CellBorders actualBorders = pasteSelection.GetBorders();
            CellBorders expectedBorders = CellBorders.Default;
            expectedBorders.InsideHorizontal = null;
            expectedBorders.InsideVertical = null;
            Assert.AreEqual(expectedBorders, actualBorders);

            IFill actualFill = pasteSelection.GetFill().Value;
            IFill expectedFill = PatternFill.Default;
            Assert.AreEqual(expectedFill, actualFill);

            ThemableFontFamily actualFontFamily = pasteSelection.GetFontFamily().Value;
            ThemableFontFamily expectedFontFamily = SpreadsheetDefaultValues.DefaultFontFamily;
            Assert.AreEqual(expectedFontFamily, actualFontFamily);

            double actualFontSize = pasteSelection.GetFontSize().Value;
            double expectedFontSize = SpreadsheetDefaultValues.DefaultFontSize;
            Assert.AreEqual(expectedFontSize, actualFontSize);

            ThemableColor actualForeColor = pasteSelection.GetForeColor().Value;
            ThemableColor expectedForeColor = SpreadsheetDefaultValues.DefaultForeColor;
            Assert.AreEqual(expectedForeColor, actualForeColor);

            CellValueFormat actualFormat = pasteSelection.GetFormat().Value;
            CellValueFormat expectedFormat = CellValueFormat.GeneralFormat;
            Assert.AreEqual(expectedFormat, actualFormat);

            RadHorizontalAlignment actualHorizontalAlignment = pasteSelection.GetHorizontalAlignment().Value;
            RadHorizontalAlignment expectedHorizontalAlignment = RadHorizontalAlignment.General;
            Assert.AreEqual(expectedHorizontalAlignment, actualHorizontalAlignment);

            int actualIndent = pasteSelection.GetIndent().Value;
            int expectedIndent = 0;
            Assert.AreEqual(expectedIndent, actualIndent);

            bool actualIsBold = pasteSelection.GetIsBold().Value;
            bool expectedIsBold = false;
            Assert.AreEqual(expectedIsBold, actualIsBold);

            bool actualIsItalic = pasteSelection.GetIsItalic().Value;
            bool expectedIsItalic = false;
            Assert.AreEqual(expectedIsItalic, actualIsItalic);

            bool actualIsWrapped = pasteSelection.GetIsWrapped().Value;
            bool expectedIsWrapped = false;
            Assert.AreEqual(expectedIsWrapped, actualIsWrapped);

            UnderlineType actualUnderline = pasteSelection.GetUnderline().Value;
            UnderlineType expectedUnderline = UnderlineType.None;
            Assert.AreEqual(expectedUnderline, actualUnderline);

            RadVerticalAlignment actualVerticalAlignment = pasteSelection.GetVerticalAlignment().Value;
            RadVerticalAlignment expectedVerticalAlignment = RadVerticalAlignment.Bottom;
            Assert.AreEqual(expectedVerticalAlignment, actualVerticalAlignment);

            bool actualIsLocked = pasteSelection.GetIsLocked().Value;
            bool expectedIsLocked = true;
            Assert.AreEqual(expectedIsLocked, actualIsLocked);

            Assert.AreEqual(worksheet.Columns[0, 0].GetWidth().Value, worksheet.Columns[3, 3].GetWidth().Value);
        }

        [TestMethod]
        public void CopyFormulaPasteColumnWidthsTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=SUM(10000000,5000000)";
            string expectedValue = string.Empty;
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.ColumnWidths, sourceIndex, targetIndex);
        }

        [TestMethod]
        public void CopyCellReferencePasteColumnWidthsTest()
        {
            CellIndex sourceIndex = new CellIndex(0, 0);
            CellIndex targetIndex = new CellIndex(3, 3);
            string value = "=D1";
            string expectedValue = string.Empty;
            this.AssertCopyPasteSingleValue(value, expectedValue, PasteType.ColumnWidths, sourceIndex, targetIndex);
        }

        #endregion


        #region Copy Paste Merged cells

        [TestMethod]
        public void CopyInsideMergedCellsTest()
        {
            Cells cells = new Workbook().Worksheets.Add().Cells;

            cells[1, 1, 4, 4].Merge();

            WorksheetFragment copy = cells[2, 2].Copy();
            CellRange expectedRange = new CellRange(0, 0, 3, 3);

            Assert.AreEqual(expectedRange, copy.CellRange);
        }

        [TestMethod]
        public void CopyInsideMergedCellsAndPaste()
        {
            Cells cells = new Workbook().Worksheets.Add().Cells;

            cells[1, 1, 4, 4].Merge();
            cells[5, 5, 8, 8].Merge();

            WorksheetFragment copy = cells[4, 4, 5, 5].Copy();
            CellRange expectedRange = new CellRange(0, 0, 7, 7);

            Assert.AreEqual(expectedRange, copy.CellRange);

            cells[10, 10].Paste(copy, PasteOptions.All);
            List<CellRange> mergedRanges = new List<CellRange>(cells.GetMergedCellRanges().OrderBy((range) => range.FromIndex.RowIndex));
            List<CellRange> expectedRanges = new List<CellRange>(new CellRange[]
            {
                new CellRange(1, 1, 4, 4),
                new CellRange(5, 5, 8, 8),
                new CellRange(10, 10, 13, 13),
                new CellRange(14, 14, 17, 17)
            });

            Assert.AreEqual(expectedRanges.Count, mergedRanges.Count);

            for (int index = 0; index < expectedRanges.Count; index++)
            {
                CellRange expected = expectedRanges[index];
                CellRange actual = mergedRanges[index];

                Assert.AreEqual(expected, actual);
            }
        }

        #endregion
    }
}