﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Windows.Documents.Spreadsheet.DataSeries;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests.FillSeriesTests
{
    [TestClass]
    public class AutoFillAllDirectionsValues
    {
        private Worksheet worksheet;

        [TestInitialize]
        public void TestInitialize()
        {
            this.worksheet = TestHelper.CreateWorksheet();
        }


        #region Helper Methods

        private CellSelection MergeCells(int fromRowIndex, int fromColumnIndex, int toRowIndex, int toColumnIndex, int mergedCellsValue)
        {
            this.worksheet.Cells[fromRowIndex, fromColumnIndex, toRowIndex, toColumnIndex].Merge();
            this.worksheet.Cells[fromRowIndex, fromColumnIndex, toRowIndex, toColumnIndex].SetValue(mergedCellsValue);

            return this.worksheet.Cells.GetCellSelection(fromRowIndex, fromColumnIndex, toRowIndex, toColumnIndex);
        }

        private static bool IsExpectedRangeMerged(CellRange cellRange)
        {
            return cellRange.ColumnCount > 1 || cellRange.RowCount > 1;
        }

        private static bool IsActualRangeMerged(Worksheet worksheet, CellRange cellRange)
        {
            CellRange mergedRange;
            if (!worksheet.Cells.TryGetContainingMergedRange(cellRange.FromIndex, out mergedRange))
            {
                return false;
            }

            return mergedRange.Equals(cellRange);
        }

        #endregion


        #region Assert Methods

        private void AssertValueRanges(CellRangeValue[] expectedValues)
        {
            foreach (CellRangeValue expectedValue in expectedValues)
            {
                bool isExpectedRangeMerged = IsExpectedRangeMerged(expectedValue.CellRange);
                bool isActualRangeMerged = IsActualRangeMerged(this.worksheet, expectedValue.CellRange);

                Assert.AreEqual(isExpectedRangeMerged, isActualRangeMerged);

                CellSelection valueSelection = this.worksheet.Cells[expectedValue.CellRange.FromIndex];
                CellValueFormat format = valueSelection.GetFormat().Value;
                string actualValue = valueSelection.GetValue().Value.GetResultValueAsString(format);
                Assert.AreEqual(expectedValue.Value, actualValue);
            }
        }

        private void AssertValuesHorizontally(CellSelection expectedCellSelection, bool isDirectionInversed, int toColumnIndex = 0, int selectionToRowIndex = 0)
        {
            CellRange cellRange = expectedCellSelection.CellRanges.First();
            int columnCount = cellRange.ColumnCount;

            List<CellSelection> actualCellSelections = new List<CellSelection>();

            if (isDirectionInversed)
            {
                int startIndex = cellRange.FromIndex.ColumnIndex - columnCount;

                for (int i = startIndex; i >= 0; i = i - columnCount)
                {
                    actualCellSelections.Add(this.worksheet.Cells.GetCellSelection(0, i, selectionToRowIndex, i + columnCount - 1));
                }
            }
            else
            {
                for (int i = cellRange.ToIndex.ColumnIndex + 1; i < toColumnIndex; i = i + columnCount)
                {
                    actualCellSelections.Add(this.worksheet.Cells.GetCellSelection(0, i, selectionToRowIndex, i + columnCount - 1));
                }
            }

            var expectedCellValue = expectedCellSelection.GetValue().Value;

            foreach (var cellSelection in actualCellSelections)
            {
                Assert.AreEqual(expectedCellValue, cellSelection.GetValue().Value);
            }

        }

        #endregion


        #region Horizontal Fill Tests

        [TestMethod]
        public void FillLeft_WhenTwoHorizontallyMergedCells_CellsAreFilled()
        {
            int rowIndex = 0;
            int colIndex = 9;
            int mergedCellsCount = 2;
            var fillDirection = FillDirection.Left;

            CellSelection expectedCellSelection = this.MergeCells(rowIndex, colIndex, rowIndex, colIndex + mergedCellsCount - 1, 2);
            this.worksheet.Cells[rowIndex, colIndex, 0, 0].FillData(fillDirection);

            bool isReversed = fillDirection == FillDirection.Left;

            this.AssertValuesHorizontally(expectedCellSelection, isReversed);
        }

        [TestMethod]
        public void FillRight_WhenTwoHorizontallyMergedCells_CellsAreFilled()
        {
            int rowIndex = 0;
            int colIndex = 0;
            int mergedCellsCount = 2;
            int toColIndex = 9;
            var fillDirection = FillDirection.Right;

            CellSelection expectedCellSelection = this.MergeCells(rowIndex, colIndex, rowIndex, colIndex + mergedCellsCount - 1, 2);
            this.worksheet.Cells[rowIndex, colIndex, 0, toColIndex].FillData(fillDirection);

            bool isReversed = fillDirection == FillDirection.Left;

            this.AssertValuesHorizontally(expectedCellSelection, isReversed, toColIndex);
        }

        [TestMethod]
        public void FillLeft_WhenThreeHorizontallyMergedCells_CellsAreFilled()
        {
            int rowIndex = 0;
            int colIndex = 9;
            int mergedCellsCount = 3;
            var fillDirection = FillDirection.Left;

            CellSelection expectedCellSelection = this.MergeCells(rowIndex, colIndex, rowIndex, colIndex + mergedCellsCount - 1, 2);
            //CellSelection expectedCellSelection = this.MergeCellsHorizontally(rowIndex, colIndex, mergedCellsCount);
            this.worksheet.Cells[rowIndex, colIndex, rowIndex, 0].FillData(fillDirection);

            bool isReversed = fillDirection == FillDirection.Left;

            this.AssertValuesHorizontally(expectedCellSelection, isReversed);
        }

        [TestMethod]
        public void FillRight_WhenThreeHorizontallyMergedCells_CellsAreFilled()
        {
            int rowIndex = 0;
            int colIndex = 0;
            int mergedCellsCount = 3;
            int toColIndex = 9;
            var fillDirection = FillDirection.Right;

            CellSelection expectedCellSelection = this.MergeCells(rowIndex, colIndex, rowIndex, colIndex + mergedCellsCount - 1, 2);
            this.worksheet.Cells[rowIndex, colIndex, rowIndex, toColIndex].FillData(fillDirection);

            bool isReversed = fillDirection == FillDirection.Left;

            this.AssertValuesHorizontally(expectedCellSelection, isReversed, toColIndex);
        }

        [TestMethod]
        public void FillLeft_WhenTwoMergedCellsAndTwoNotMergedBelow_CellsAreFilled()
        {
            var fillDirection = FillDirection.Left;

            this.MergeCells(0, 5, 0, 6, 2);
            this.worksheet.Cells[1, 5, 1, 6].SetValue(1);

            this.worksheet.Cells[0, 0, 1, 6].FillData(fillDirection);


            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                // first
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 5, 0, 6),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 5, 1, 5),
                    Value = "1"
                },
                 new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 6, 1, 6),
                    Value = "1"
                },
                // second
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 3, 0, 4),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 3, 1, 3),
                    Value = "1"
                },
                 new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 4, 1, 4),
                    Value = "1"
                },
                // third
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 1, 0, 2),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 1, 1, 1),
                    Value = "1"
                },
                 new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 2, 1, 2),
                    Value = "1"
                },
                // last not merged
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 0, 0),
                    Value = ""
                },
                 new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 0, 1, 0),
                    Value = "1"
                },
            };

            this.AssertValueRanges(expectedValues);
        }

        [TestMethod]
        public void FillRight_WhenTwoMergedCellsAndTwoNotMergedBelow_CellsAreFilled()
        {
            var fillDirection = FillDirection.Right;

            this.MergeCells(0, 0, 0, 1, 2);
            this.worksheet.Cells[1, 0, 1, 1].SetValue(1);

            this.worksheet.Cells[0, 0, 1, 6].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                // first
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 0, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 0, 1, 0),
                    Value = "1"
                },
                 new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 1, 1, 1),
                    Value = "1"
                },
                // second
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 2, 0, 3),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 2, 1, 2),
                    Value = "1"
                },
                 new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 3, 1, 3),
                    Value = "1"
                },
                // third
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 4, 0, 5),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 4, 1, 4),
                    Value = "1"
                },
                 new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 5, 1, 5),
                    Value = "1"
                },
                // last not merged
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 6, 0, 6),
                    Value = ""
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 6, 1, 6),
                    Value = "1"
                },
            };

            this.AssertValueRanges(expectedValues);
        }

        [TestMethod]
        public void FillLeft_WhenTwoMergedCellsAndTwoNotMergedAbove_CellsAreFilled()
        {
            var fillDirection = FillDirection.Left;

            this.worksheet.Cells[0, 5, 0, 6].SetValue(1);
            this.MergeCells(1, 5, 1, 6, 2);

            this.worksheet.Cells[0, 6, 1, 0].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                // first
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 5, 0, 5),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 6, 0, 6),
                    Value = "1"
                },
                 new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 5, 1, 6),
                    Value = "2"
                },
                // second
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 3, 0, 3),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 4, 0, 4),
                    Value = "1"
                },
                 new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 3, 1, 4),
                    Value = "2"
                },
                // third
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 1, 0, 1),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 2, 0, 2),
                    Value = "1"
                },
                 new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 1, 1, 2),
                    Value = "2"
                },
                // last not merged
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 0, 0),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 0, 1, 0),
                    Value = ""
                },
            };

            this.AssertValueRanges(expectedValues);

        }

        [TestMethod]
        public void FillRight_WhenTwoMergedCellsAndTwoNotMergedAbove_CellsAreFilled()
        {
            var fillDirection = FillDirection.Right;

            this.worksheet.Cells[0, 0, 0, 1].SetValue(1);
            this.MergeCells(1, 0, 1, 1, 2);

            this.worksheet.Cells[0, 0, 1, 6].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                // first
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 0, 0),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 1, 0, 1),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 0, 1, 1),
                    Value = "2"
                },
                // second
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 2, 0, 2),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 3, 0, 3),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 2, 1, 3),
                    Value = "2"
                },
                // third
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 4, 0, 4),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 5, 0, 5),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 4, 1, 5),
                    Value = "2"
                },
                // last with no merged
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 6, 0, 6),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 6, 1, 6),
                    Value = ""
                },
            };

            this.AssertValueRanges(expectedValues);
        }

        [TestMethod]
        public void FillLeft_WhenSingleCellIsMergedOnMultipleRows_CellsAreFilled()
        {
            int rowIndex = 0;
            int colIndex = 9;
            int mergedCellsCount = 3;
            var fillDirection = FillDirection.Left;

            CellSelection expectedSelection = this.MergeCells(rowIndex, colIndex, rowIndex + mergedCellsCount - 1, colIndex, 3);

            this.worksheet.Cells[rowIndex, colIndex, rowIndex + mergedCellsCount - 1, 0].FillData(fillDirection);

            bool isReversed = fillDirection == FillDirection.Left;

            this.AssertValuesHorizontally(expectedSelection, isReversed, 0, mergedCellsCount - 1);
        }

        [TestMethod]
        public void FillRight_WhenSingleCellIsMergedOnMultipleRows_CellsAreFilled()
        {
            int rowIndex = 0;
            int colIndex = 0;
            int mergedCellsCount = 3;
            var fillDirection = FillDirection.Right;

            CellSelection expectedSelection = this.MergeCells(rowIndex, colIndex, rowIndex + mergedCellsCount - 1, colIndex, 3);

            this.worksheet.Cells[rowIndex, colIndex, rowIndex + mergedCellsCount - 1, 9].FillData(fillDirection);

            bool isReversed = fillDirection == FillDirection.Left;

            this.AssertValuesHorizontally(expectedSelection, isReversed, 9, mergedCellsCount - 1);
        }

        [TestMethod]
        public void FillLeft_WhenCellAreMergedOnMultipleRowsAndColumns_CellsAreFilled()
        {
            int rowIndex = 0;
            int colIndex = 9;
            int mergedCellsCount = 3;
            var fillDirection = FillDirection.Left;

            CellSelection expectedSelection = this.MergeCells(rowIndex, colIndex, rowIndex + mergedCellsCount - 1, colIndex + mergedCellsCount - 1, 3);

            this.worksheet.Cells[rowIndex, colIndex, rowIndex + mergedCellsCount - 1, 0].FillData(fillDirection);

            bool isReversed = fillDirection == FillDirection.Left;

            this.AssertValuesHorizontally(expectedSelection, isReversed, 0, mergedCellsCount - 1);
        }

        [TestMethod]
        public void FillRight_WhenCellAreMergedOnMultipleRowsAndColumns_CellsAreFilled()
        {
            int rowIndex = 0;
            int colIndex = 0;
            int mergedCellsCount = 3;
            var fillDirection = FillDirection.Right;

            CellSelection expectedSelection = this.MergeCells(rowIndex, colIndex, rowIndex + mergedCellsCount - 1, colIndex + mergedCellsCount - 1, 3);

            this.worksheet.Cells[rowIndex, colIndex, rowIndex + mergedCellsCount - 1, 9].FillData(fillDirection);

            bool isReversed = fillDirection == FillDirection.Left;

            this.AssertValuesHorizontally(expectedSelection, isReversed, 9, mergedCellsCount - 1);
        }

        [TestMethod]
        public void FillLeft_WhenCellIsMergedOn2RowsAnd2ColsAndBelowIsCellMergedOn2ColsAndBelowAre2NotMergedCells_CellsAreFilled()
        {
            var fillDirection = FillDirection.Left;

            this.MergeCells(0, 5, 1, 6, 4);
            this.MergeCells(2, 5, 2, 6, 2);
            this.worksheet.Cells[3, 5, 3, 6].SetValue(1);

            this.worksheet.Cells[0, 6, 3, 0].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                // first
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 5, 1, 6),
                    Value = "4"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 5, 2, 6),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 5, 3, 5),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 6, 3, 6),
                    Value = "1"
                },
                 // second
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 3, 1, 4),
                    Value = "4"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 3, 2, 4),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 3, 3, 3),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 4, 3, 4),
                    Value = "1"
                },
                 // third
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 1, 1, 2),
                    Value = "4"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 1, 2, 2),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 1, 3, 1),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 2, 3, 2),
                    Value = "1"
                },
                // last with not merged
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 0, 0),
                    Value = ""
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 0, 1, 0),
                    Value = ""
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 0, 2, 0),
                    Value = ""
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 0, 3, 0),
                    Value = "1"
                }
            };

            this.AssertValueRanges(expectedValues);
        }

        [TestMethod]
        public void FillRight_WhenCellIsMergedOn2RowsAnd2ColsAndBelowIsCellMergedOn2ColsAndBelowAre2NotMergedCells_CellsAreFilled()
        {
            var fillDirection = FillDirection.Right;

            this.MergeCells(0, 0, 1, 1, 4);
            this.MergeCells(2, 0, 2, 1, 2);
            this.worksheet.Cells[3, 0, 3, 1].SetValue(1);

            this.worksheet.Cells[0, 0, 3, 8].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                // first
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 1, 1),
                    Value = "4"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 0, 2, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {
                    CellRange = new CellRange(3, 0, 3, 0),
                    Value = "1"
                },
                new CellRangeValue()
                {
                    CellRange = new CellRange(3, 1, 3, 1),
                    Value="1"
                },
                // second
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 2, 1, 3),
                    Value = "4"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 2, 2, 3),
                    Value = "2"
                },
                new CellRangeValue()
                {
                    CellRange = new CellRange(3, 2, 3, 2),
                    Value = "1"
                },
                new CellRangeValue()
                {
                    CellRange = new CellRange(3, 3, 3, 3),
                    Value="1"
                },
                // third
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 4, 1, 5),
                    Value = "4"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 4, 2, 5),
                    Value = "2"
                },
                new CellRangeValue()
                {
                    CellRange = new CellRange(3, 4, 3, 4),
                    Value = "1"
                },
                new CellRangeValue()
                {
                    CellRange = new CellRange(3, 5, 3, 5),
                    Value="1"
                },
                // fourth
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 6, 1, 7),
                    Value = "4"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 6, 2, 7),
                    Value = "2"
                },
                new CellRangeValue()
                {
                    CellRange = new CellRange(3, 6, 3, 6),
                    Value = "1"
                },
                new CellRangeValue()
                {
                    CellRange = new CellRange(3, 7, 3, 7),
                    Value="1"
                },
                // last not with merged
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 8, 0, 8),
                    Value = ""
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 8, 1, 8),
                    Value = ""
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 8, 2, 8),
                    Value = ""
                },
                new CellRangeValue()
                {
                    CellRange = new CellRange(3, 8, 3, 8),
                    Value="1"
                }
            };

            this.AssertValueRanges(expectedValues);
        }

        [TestMethod]
        public void FillLeft_WhenCellIsMergedOn2RowsAnd2ColsAndAboveIsCellMergedOn2ColsAndAboveAre2NotMergedCells_CellsAreFilled()
        {
            var fillDirection = FillDirection.Left;

            this.worksheet.Cells[0, 5, 0, 6].SetValue(1);
            this.MergeCells(1, 5, 1, 6, 2);            
            this.MergeCells(2, 5, 3, 6, 4);

            this.worksheet.Cells[0, 6, 3, 0].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                // first
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 5, 0, 5),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 6, 0, 6),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 5, 1, 6),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 5, 3, 6),
                    Value = "4"
                },
                // second
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 3, 0, 3),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 4, 0, 4),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 3, 1, 4),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 3, 3, 4),
                    Value = "4"
                },
                // third
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 1, 0, 1),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 2, 0, 2),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 1, 1, 2),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 1, 3, 2),
                    Value = "4"
                },
                // last not merged
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 0, 0),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 0, 1, 0),
                    Value = ""
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 0, 2, 0),
                    Value = ""
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 0, 3, 0),
                    Value = ""
                },
                
            };

            this.AssertValueRanges(expectedValues);
        }

        #endregion


        #region Vertical Fill Tests

        [TestMethod]
        public void FillUp_WhenTwoHorizontallyMergedCells_CellsAreFilled()
        {
            var fillDirection = FillDirection.Up;

            this.MergeCells(3, 0, 3, 1, 2);

            this.worksheet.Cells[3, 0, 0, 1].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 0, 3, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 0, 2, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 0, 1, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 0, 1),
                    Value = "2"
                },
            };

            this.AssertValueRanges(expectedValues);
        }

        [TestMethod]
        public void FillDown_WhenTwoHorizontallyMergedCells_CellsAreFilled()
        {
            var fillDirection = FillDirection.Down;

            this.MergeCells(0, 0, 0, 1, 2);

            this.worksheet.Cells[0, 0, 3, 1].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 0, 3, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 0, 2, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 0, 1, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 0, 1),
                    Value = "2"
                },
            };

            this.AssertValueRanges(expectedValues);
        }

        [TestMethod]
        public void FillUp_WhenSingleCellIsMergedOnMultipleRows_CellsAreFilled()
        {
            var fillDirection = FillDirection.Up;

            this.MergeCells(7, 0, 9, 0, 3);

            this.worksheet.Cells[9, 0, 0, 0].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                new CellRangeValue()
                {   
                    CellRange = new CellRange(7, 0, 9, 0),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(4, 0, 6, 0),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 0, 3, 0),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 0, 0),
                    Value = ""
                },
            };

            this.AssertValueRanges(expectedValues);
        }

        [TestMethod]
        public void FillDown_WhenSingleCellIsMergedOnMultipleRows_CellsAreFilled()
        {
            var fillDirection = FillDirection.Down;

            this.MergeCells(0, 0, 2, 0, 3);

            this.worksheet.Cells[0, 0, 9, 0].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 2, 0),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 0, 5, 0),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(6, 0, 8, 0),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(9, 0, 9, 0),
                    Value = ""
                },
            };

            this.AssertValueRanges(expectedValues);
        }

        [TestMethod]
        public void FillUp_WhenCellAreMergedOnMultipleRowsAndColumns_CellsAreFilled()
        {
            var fillDirection = FillDirection.Up;

            this.MergeCells(7, 0, 9, 2, 3);

            this.worksheet.Cells[9, 0, 0, 2].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                new CellRangeValue()
                {   
                    CellRange = new CellRange(7, 0, 9, 2),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(4, 0, 6, 2),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 0, 3, 2),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 0, 0),
                    Value = ""
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 1, 0, 1),
                    Value = ""
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 2, 0, 2),
                    Value = ""
                },
            };

            this.AssertValueRanges(expectedValues);
        }

        [TestMethod]
        public void FillDown_WhenCellAreMergedOnMultipleRowsAndColumns_CellsAreFilled()
        {
            var fillDirection = FillDirection.Down;

            this.MergeCells(0, 0, 2, 2, 3);

            this.worksheet.Cells[0, 0, 9, 2].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 2, 2),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 0, 5, 2),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(6, 0, 8, 2),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(9, 0, 9, 0),
                    Value = ""
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(9, 1, 9, 1),
                    Value = ""
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(9, 2, 9, 2),
                    Value = ""
                },
            };

            this.AssertValueRanges(expectedValues);
        }

        [TestMethod]
        public void FillUp_WhenTwoMergedCellsAndTwoNotMergedBelow_CellsAreFilledWithNotMergedCellValues()
        {
            var fillDirection = FillDirection.Up;

            this.MergeCells(1, 0, 1, 1, 2);
            this.worksheet.Cells[2, 0, 2, 1].SetValue(1);

            this.worksheet.Cells[2, 0, 0, 1].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 0, 0),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 1, 0, 1),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 0, 1, 0),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 0, 1, 0),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 0, 2, 0),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 1, 2, 1),
                    Value = "1"
                },
            };

            this.AssertValueRanges(expectedValues);
        }

        [TestMethod]
        public void FillDown_WhenTwoMergedCellsAndTwoNotMergedBelow_CellsAreFilledWithMergedCellValue()
        {
            var fillDirection = FillDirection.Down;

            this.MergeCells(0, 0, 0, 1, 2);
            this.worksheet.Cells[1, 0, 1, 1].SetValue(1);

            this.worksheet.Cells[0, 0, 2, 1].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 0, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 0, 1, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 0, 2, 1),
                    Value = "2"
                },
            };

            this.AssertValueRanges(expectedValues);
        }

        [TestMethod]
        public void FillUp_WhenTwoMergedCellsAndTwoNotMergedAbove_CellsAreFilledWithTheMergedCellsValue()
        {
            var fillDirection = FillDirection.Up;

            this.MergeCells(4, 0, 4, 1, 2);
            this.worksheet.Cells[3, 0, 3, 1].SetValue(1);

            this.worksheet.Cells[4, 0, 0, 1].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                new CellRangeValue()
                {   
                    CellRange = new CellRange(4, 0, 4, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 0, 3, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 0, 2, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 0, 1, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 0, 1),
                    Value = "2"
                },
            };

            this.AssertValueRanges(expectedValues);
        }

        [TestMethod]
        public void FillDown_WhenTwoMergedCellsAndTwoNotMergedAbove_CellsAreFilledWithTheNotMergedCellsValue()
        {
            var fillDirection = FillDirection.Down;

            this.worksheet.Cells[0, 0, 0, 1].SetValue(1);
            this.MergeCells(1, 0, 1, 1, 2);

            this.worksheet.Cells[0, 0, 3, 1].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 0, 0),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 1, 0, 1),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 0, 1, 0),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 1, 1, 1),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 0, 2, 0),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 1, 2, 1),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 0, 3, 0),
                    Value = "1"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 1, 3, 1),
                    Value = "1"
                },
            };

            this.AssertValueRanges(expectedValues);
        }

        [TestMethod]
        public void FillUp_When2ColumnsAreMergedAndNextTheCellIsAnotherWith3MergedColumns_CellsAreFilledWithThe2TypesOfSameMergedCells()
        {
            var fillDirection = FillDirection.Up;

            this.MergeCells(3, 0, 3, 1, 2);
            this.MergeCells(3, 2, 3, 4, 3);

            this.worksheet.Cells[3, 0, 0, 4].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 0, 3, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 2, 3, 4),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 0, 2, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 2, 2, 4),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 0, 1, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 2, 1, 4),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 0, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 2, 0, 4),
                    Value = "3"
                },
            };

            this.AssertValueRanges(expectedValues);
        }

        [TestMethod]
        public void FillDown_When2ColumnsAreMergedAndNextTheCellIsAnotherWith3MergedColumns_CellsAreFilledWithThe2TypesOfSameMergedCells()
        {
            var fillDirection = FillDirection.Down;

            this.MergeCells(0, 0, 0, 1, 2);
            this.MergeCells(0, 2, 0, 4, 3);

            this.worksheet.Cells[0, 0, 3, 4].FillData(fillDirection);

            CellRangeValue[] expectedValues = new CellRangeValue[]
            {
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 0, 0, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(0, 2, 0, 4),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 0, 1, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(1, 2, 1, 4),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 0, 2, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(2, 2, 2, 4),
                    Value = "3"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 0, 3, 1),
                    Value = "2"
                },
                new CellRangeValue()
                {   
                    CellRange = new CellRange(3, 2, 3, 4),
                    Value = "3"
                },
            };

            this.AssertValueRanges(expectedValues);
        }

        #endregion

    }
}