﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Filtering;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests.Filtering
{
    [TestClass]
    public class AutoFilterTests
    {
        public static readonly CellRange DefaultCellRange = new CellRange(5, 5, 10, 10);

        #region CellRange

        [TestMethod]
        public void AddNewRangeTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            
            worksheet.Filter.FilterRange = DefaultCellRange;

            this.AssertNoColumnHasFilter(worksheet);
        }

        [TestMethod]
        public void AddRangeFilterAndNewRangeTests()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            
            worksheet.Filter.FilterRange = DefaultCellRange;
            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter);
            worksheet.Filter.FilterRange = null;

            this.AssertNoColumnHasFilter(worksheet);
        }

        #endregion


        #region SetFilter

        [TestMethod]
        public void SetFilterTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            
            worksheet.Filter.FilterRange = DefaultCellRange;
            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter);

            IFilter retrievedFilter = worksheet.Filter.GetFilter(0);
            Assert.AreEqual(filter, retrievedFilter);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void SetFilterOutOfRangeTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            
            worksheet.Filter.FilterRange = DefaultCellRange;
            int outOfRangeColumn = DefaultCellRange.ToIndex.ColumnIndex - DefaultCellRange.FromIndex.ColumnIndex + 1;
            IFilter filter = new DynamicFilter(outOfRangeColumn, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter);
        }

        [TestMethod]
        [ExpectedException(typeof(FilteringException))]
        public void SetFilterBeforeRangeTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            
            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter);
        }

        [TestMethod]
        public void SetRangeFilterOnSingleRowRange()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Filter.FilterRange = new CellRange(0, 0, 0, 0);
            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter);

            IFilter retrievedFilter = worksheet.Filter.GetFilter(0);
            Assert.AreEqual(filter, retrievedFilter);
        }

        [TestMethod]
        public void SetRangeFilterOnSingleRowRange2()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Filter.FilterRange = new CellRange(0, 0, 0, 0);
            IFilter filter = new TopFilter(0);
            worksheet.Filter.SetFilter(filter);

            IFilter retrievedFilter = worksheet.Filter.GetFilter(0);
            Assert.AreEqual(filter, retrievedFilter);
        }

        #endregion


        #region RemoveFilter

        [TestMethod]
        public void RemoveFilterTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Filter.FilterRange = DefaultCellRange;
            IFilter filter = new DynamicFilter(1, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter);

            worksheet.Filter.RemoveFilter(1);

            IFilter retrievedFilter = worksheet.Filter.GetFilter(1);
            Assert.IsNull(retrievedFilter);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void RemoveFilterOutOfRangeTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Filter.FilterRange = DefaultCellRange;

            int outOfRangeColumn = DefaultCellRange.ToIndex.ColumnIndex - DefaultCellRange.FromIndex.ColumnIndex + 1;
            worksheet.Filter.RemoveFilter(outOfRangeColumn);
        }

        [TestMethod]
        [ExpectedException(typeof(FilteringException))]
        public void RemoveFilterBeforeRangeTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Filter.RemoveFilter(1);
        }

        [TestMethod]
        public void RemoveUnexistingFilterTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Filter.FilterRange = DefaultCellRange;

            bool sucess = worksheet.Filter.RemoveFilter(1);
            Assert.IsFalse(sucess);
        }

        #endregion


        #region ClearFilters

        [TestMethod]
        public void ClearFilterTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Filter.FilterRange = DefaultCellRange;
            IFilter filter1 = new DynamicFilter(1, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter1);
            IFilter filter0 = new DynamicFilter(0, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter0);

            worksheet.Filter.ClearFilters();

            this.AssertNoColumnHasFilter(worksheet);
        }

        #endregion


        #region Filter through cell selection

        [TestMethod]
        public void FilterThroughCellSelection()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection selection = worksheet.Cells[0, 0, 5, 5];

            IFilter filter0 = new DynamicFilter(0, DynamicFilterType.April);
            IFilter filter1 = new DynamicFilter(1, DynamicFilterType.April);
            selection.Filter(new IFilter[] { filter0, filter1 });

            IFilter retrievedFilter0 = worksheet.Filter.GetFilter(0);
            IFilter retrievedFilter1 = worksheet.Filter.GetFilter(1);
            Assert.AreEqual(filter0, retrievedFilter0);
            Assert.AreEqual(filter1, retrievedFilter1);
        }

        #endregion

        private void AssertNoColumnHasFilter(Worksheet worksheet)
        {
            for (int i = 0; i < DefaultCellRange.ColumnCount; i++)
            {
                IFilter filter = worksheet.Filter.GetFilter(i);
                Assert.IsNull(filter);
            }
        }
    }
}
