﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Filtering;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests.Filtering
{
    [TestClass]
    public class DynamicFilterTests
    {
        private static readonly CellRange range = new CellRange(0, 0, 10, 0);

        #region Use filter before assigning it to sheet

        [TestMethod]
        [ExpectedException(typeof(FilteringException))]
        public void UseFilterBeforeAssigningAboveAverageType()
        {
            double value = 2;
            IFilter filter = new DynamicFilter(0, DynamicFilterType.AboveAverage);

            filter.ShouldShowValue(value);
        }

        [TestMethod]
        public void UseFilterBeforeAssigningMonthType()
        {
            DateTime date = new DateTime(2013, 4, 2);
            IFilter filter = new DynamicFilter(0, DynamicFilterType.AboveAverage);

            bool shouldShow = filter.ShouldShowValue(date);
            Assert.IsFalse(shouldShow);
        }

        #endregion


        #region Get value

        [TestMethod]
        public void GetValueNumberCellValueTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[0, 0, 0, 0].SetValue(2);

            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            object value = filter.GetValue(worksheet.Cells, 0, 0);

            Assert.IsTrue(value is double);
            Assert.AreEqual(2d, value);
        }

        [TestMethod]
        public void GetValueNumberCellValueDateTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[0, 0, 0, 0].SetValue(new DateTime(1986, 4, 2));

            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            object value = filter.GetValue(worksheet.Cells, 0, 0);

            Assert.IsTrue(value is DateTime);
            Assert.AreEqual(new DateTime(1986, 4, 2), value);
        }

        [TestMethod]
        public void GetValueNumberCellValueDateCustomFormatTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[0, 0, 0, 0].SetValue(31504);
            worksheet.Cells[0, 0, 0, 0].SetFormat(new CellValueFormat("dd/mm/yyyy h:mm"));

            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            object value = filter.GetValue(worksheet.Cells, 0, 0);

            Assert.IsTrue(value is DateTime);
            Assert.AreEqual(new DateTime(1986, 4, 2), value);
        }

        [TestMethod]
        public void GetValueBoolCellValueTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[0, 0, 0, 0].SetValue(true);

            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            object value = filter.GetValue(worksheet.Cells, 0, 0);

            Assert.IsNull(value);
        }

        [TestMethod]
        public void GetValueTextCellValueTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[0, 0, 0, 0].SetValue("test");

            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            object value = filter.GetValue(worksheet.Cells, 0, 0);

            Assert.IsNull(value);
        }

        [TestMethod]
        public void GetValueFormulaCellValueTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[0, 0, 0, 0].SetValue("=\"test\"");

            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            object value = filter.GetValue(worksheet.Cells, 0, 0);

            Assert.IsNull(value);
        }

        [TestMethod]
        public void GetValueFormulaCellValueWithNumberResultTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[0, 0, 0, 0].SetValue("=2+2");

            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            object value = filter.GetValue(worksheet.Cells, 0, 0);

            Assert.IsTrue(value is double);
            Assert.AreEqual(4d, value);
        }

        [TestMethod]
        public void GetValueBlankCellValueTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            object value = filter.GetValue(worksheet.Cells, 0, 0);

            Assert.IsNull(value);
        }

        #endregion


        #region Should show value

        [TestMethod]
        public void ShouldShowValueAboveAveragePositive()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FilterTestHelper.SetValuesToWorksheetColumn(worksheet, 1, 0, new List<object>()
            {
                1d, 2d, 3d, 4d, 5d, 6d, 7d, "test1", "test2", true
            });

            IFilter filter = new DynamicFilter(0, DynamicFilterType.AboveAverage);
            worksheet.Filter.FilterRange = range;
            worksheet.Filter.SetFilter(filter);

            var shouldShow = filter.ShouldShowValue(5d);
            Assert.IsTrue(shouldShow);
        }

        [TestMethod]
        public void ShouldShowValueAboveAverageDatePositive()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FilterTestHelper.SetValuesToWorksheetColumn(worksheet, 1, 0, new List<object>()
            {
                1d, 2d, 3d, 4d, 5d, 6d, 7d, "test1", new DateTime(2013, 2, 2), true
            });

            IFilter filter = new DynamicFilter(0, DynamicFilterType.AboveAverage);
            worksheet.Filter.FilterRange = range;
            worksheet.Filter.SetFilter(filter);

            var shouldShow = filter.ShouldShowValue(new DateTime(2013, 2, 2));
            Assert.IsTrue(shouldShow);
        }

        [TestMethod]
        public void ShouldShowValueAboveAverageNegative()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FilterTestHelper.SetValuesToWorksheetColumn(worksheet, 1, 0, new List<object>()
            {
                1d, 2d, 3d, 4d, 5d, 6d, 7d, "test1", "test2", true
            });

            IFilter filter = new DynamicFilter(0, DynamicFilterType.AboveAverage);
            worksheet.Filter.FilterRange = range;
            worksheet.Filter.SetFilter(filter);

            var shouldShow = filter.ShouldShowValue("test1");
            Assert.IsFalse(shouldShow);
        }

        [TestMethod]
        public void ShouldShowValueAboveAverageNegative2()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FilterTestHelper.SetValuesToWorksheetColumn(worksheet, 1, 0, new List<object>()
            {
                1d, 2d, 3d, 4d, 5d, 6d, 7d, "test1", "test2", true
            });

            IFilter filter = new DynamicFilter(0, DynamicFilterType.AboveAverage);
            worksheet.Filter.FilterRange = range;
            worksheet.Filter.SetFilter(filter);

            var shouldShow = filter.ShouldShowValue(4);
            Assert.IsFalse(shouldShow);
        }

        [TestMethod]
        public void ShouldShowValueTomorrowPositive()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            IFilter filter = new DynamicFilter(0, DynamicFilterType.Tomorrow);
            worksheet.Filter.FilterRange = range;
            worksheet.Filter.SetFilter(filter);

            var shouldShow = filter.ShouldShowValue(DateTime.Today.AddDays(1));
            Assert.IsTrue(shouldShow);
        }

        [TestMethod]
        public void ShouldShowValuePreviousWeekNegative()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            IFilter filter = new DynamicFilter(0, DynamicFilterType.LastWeek);
            worksheet.Filter.FilterRange = range;
            worksheet.Filter.SetFilter(filter);

            var shouldShow = filter.ShouldShowValue(new DateTime(1986, 4, 2));
            Assert.IsFalse(shouldShow);
        }

        [TestMethod]
        public void ShouldShowValueYearToDatePositive()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            IFilter filter = new DynamicFilter(0, DynamicFilterType.YearToDate);
            worksheet.Filter.FilterRange = range;
            worksheet.Filter.SetFilter(filter);

            var shouldShow = filter.ShouldShowValue(DateTime.Today);
            Assert.IsTrue(shouldShow);
        }

        [TestMethod]
        public void ShouldShowValueLastMonthPositive()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            IFilter filter = new DynamicFilter(0, DynamicFilterType.LastMonth);
            worksheet.Filter.FilterRange = range;
            worksheet.Filter.SetFilter(filter);

            var shouldShow = filter.ShouldShowValue(DateTime.Today.AddMonths(-1));
            Assert.IsTrue(shouldShow);
        }

        [TestMethod]
        public void ShouldShowValueNextQuarterPositive()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            IFilter filter = new DynamicFilter(0, DynamicFilterType.NextQuarter);
            worksheet.Filter.FilterRange = range;
            worksheet.Filter.SetFilter(filter);

            var shouldShow = filter.ShouldShowValue(DateTime.Today.AddMonths(3));
            Assert.IsTrue(shouldShow);
        }

        [TestMethod]
        public void ShouldShowValueQuarter1Positive()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            IFilter filter = new DynamicFilter(0, DynamicFilterType.Quarter1);
            worksheet.Filter.FilterRange = range;
            worksheet.Filter.SetFilter(filter);

            var shouldShow = filter.ShouldShowValue(new DateTime(2013, 1, 1));
            Assert.IsTrue(shouldShow);
        }

        [TestMethod]
        public void ShouldShowValueAprilNegative()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            worksheet.Filter.FilterRange = range;
            worksheet.Filter.SetFilter(filter);

            var shouldShow = filter.ShouldShowValue(new DateTime(2013, 1, 1));
            Assert.IsFalse(shouldShow);
        }

        #endregion
    }
}
