﻿using System;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Model.Filtering;

namespace Telerik.Windows.Documents.Spreadsheet.Tests.Filtering
{
    [TestClass]
    public class ValuesCollectionFilterTests
    {
        private static readonly List<string> StringValues = new List<string>() {"1", "1.2", "2", "test"};
        private static readonly List<DateGroupItem> DateValues = new List<DateGroupItem>() 
        {
            new DateGroupItem(1986, 4, 2),
            new DateGroupItem(2013)
        };

        #region Constructors

        [TestMethod]
        public void CreateFilterTest()
        {
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, null, null, true);

            Assert.IsNotNull(filter.StringValues);
            Assert.AreEqual(0, filter.StringValues.Count());
            Assert.IsNotNull(filter.DateItems);
            Assert.AreEqual(0, filter.DateItems.Count());
            Assert.IsTrue(filter.Blank);
        }

        #endregion


        #region Should show value

        [TestMethod]
        public void ShouldShowValueTestString1()
        {
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, StringValues, DateValues, true);

            string testString = "1";
            bool shouldShow = filter.ShouldShowValue(testString);

            Assert.IsTrue(shouldShow);
        }

        [TestMethod]
        public void ShouldShowValueTestString2()
        {
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, StringValues, DateValues, true);

            string testString = "3";
            bool shouldShow = filter.ShouldShowValue(testString);

            Assert.IsFalse(shouldShow);
        }

        [TestMethod]
        public void ShouldShowValueTestBlank()
        {
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, StringValues, DateValues, true);

            string testString = String.Empty;
            bool shouldShow = filter.ShouldShowValue(testString);

            Assert.IsTrue(shouldShow);
        }

        [TestMethod]
        public void ShouldShowValueTestBlank2()
        {
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, StringValues, DateValues, true);

            string testString = "   ";
            bool shouldShow = filter.ShouldShowValue(testString);

            Assert.IsTrue(shouldShow);
        }

        [TestMethod]
        public void ShouldShowValueTestDate1()
        {
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, StringValues, DateValues, true);

            DateTime date = new DateTime(1986, 4, 2);
            bool shouldShow = filter.ShouldShowValue(date);

            Assert.IsTrue(shouldShow);
        }

        [TestMethod]
        public void ShouldShowValueTestDate2()
        {
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, StringValues, DateValues, true);

            DateTime date = new DateTime(1986, 4, 2, 13, 2, 0);
            bool shouldShow = filter.ShouldShowValue(date);

            Assert.IsTrue(shouldShow);
        }

        [TestMethod]
        public void ShouldShowValueTestDate3()
        {
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, StringValues, DateValues, true);

            DateTime date = new DateTime(1986, 4, 1);
            bool shouldShow = filter.ShouldShowValue(date);

            Assert.IsFalse(shouldShow);
        }

        [TestMethod]
        public void ShouldShowValueTestDate4()
        {
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, StringValues, DateValues, true);

            DateTime date = new DateTime(2013, 4, 2);
            bool shouldShow = filter.ShouldShowValue(date);

            Assert.IsTrue(shouldShow);
        }

        #endregion


        #region Get value

        [TestMethod]
        public void GetValueTestEmpty()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, null, null, false);

            object value = filter.GetValue(worksheet.Cells, 0, 0);

            string stringValue = value as string;

            Assert.IsNotNull(stringValue);
            Assert.AreEqual("", stringValue);
        }

        [TestMethod]
        public void GetValueTestString1()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[0, 0, 0, 0].SetValue(1);
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, null, null, false);

            object value = filter.GetValue(worksheet.Cells, 0, 0);

            string stringValue = value as string;

            Assert.IsNotNull(stringValue);
            Assert.AreEqual("1", stringValue);
        }

        [TestMethod]
        public void GetValueTestString2()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[0, 0, 0, 0].SetValue("test");
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, null, null, false);

            object value = filter.GetValue(worksheet.Cells, 0, 0);

            string stringValue = value as string;

            Assert.IsNotNull(stringValue);
            Assert.AreEqual("test", stringValue);
        }

        [TestMethod]
        public void GetValueTestString3()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[0, 0, 0, 0].SetValue(true);
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, null, null, false);

            object value = filter.GetValue(worksheet.Cells, 0, 0);

            string stringValue = value as string;

            Assert.IsNotNull(stringValue);
            Assert.AreEqual("TRUE", stringValue);
        }

        [TestMethod]
        public void GetValueTestDate1()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[0, 0, 0, 0].SetValue(41527);
            worksheet.Cells[0, 0, 0, 0].SetFormat(new CellValueFormat("dd/mm/yyyy h:mm"));
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, null, null, false);

            object value = filter.GetValue(worksheet.Cells, 0, 0);

            DateTime? dateValue = value as DateTime?;

            Assert.IsNotNull(dateValue);
            Assert.AreEqual(new DateTime(2013, 9, 10, 0, 0, 0), dateValue.Value);
        }

        [TestMethod]
        public void GetValueTestDate2()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[0, 0, 0, 0].SetValue(41527.52);
            worksheet.Cells[0, 0, 0, 0].SetFormat(new CellValueFormat("dd/mm/yyyy"));
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, null, null, false);

            object value = filter.GetValue(worksheet.Cells, 0, 0);

            DateTime? dateValue = value as DateTime?;

            Assert.IsNotNull(dateValue);
            Assert.AreEqual(new DateTime(2013, 9, 10, 12, 28, 48), dateValue.Value);
        }

        [TestMethod]
        public void GetValueTestDate3()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[0, 0, 0, 0].SetValue(new DateTime(2013, 9, 10, 12, 28, 48));
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, null, null, false);

            object value = filter.GetValue(worksheet.Cells, 0, 0);

            DateTime? dateValue = value as DateTime?;

            Assert.IsNotNull(dateValue);
            Assert.AreEqual(new DateTime(2013, 9, 10, 12, 28, 48), dateValue.Value);
        }

        [TestMethod]
        public void GetValueTestTime()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[0, 0, 0, 0].SetValue(0.6);
            worksheet.Cells[0, 0, 0, 0].SetFormat(new CellValueFormat("h:mm:ss AM/PM"));
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, null, null, false);

            object value = filter.GetValue(worksheet.Cells, 0, 0);

            string stringValue = value as string;

            Assert.IsNotNull(stringValue);
            Assert.AreEqual("2:24:00 PM", stringValue);
        }

        [TestMethod]
        public void GetValueTestFormulaCellValue()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[0, 0, 0, 0].SetValue("=TODAY()");
            ValuesCollectionFilter filter = new ValuesCollectionFilter(0, null, null, false);

            object value = filter.GetValue(worksheet.Cells, 0, 0);

            Assert.IsTrue(value is DateTime);
        }

        #endregion
    }
}
