﻿using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Expressions.Functions;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests
{
    [TestClass]
    public class CustomFunctionsTests
    {
        #region Custom Functions Test Methods

        [TestMethod]
        public void NullaryFunctionTest()
        {
            Workbook workbook = new Workbook();
            workbook.Sheets.Add(SheetType.Worksheet);
            Worksheet worksheet = workbook.ActiveWorksheet;

            FunctionManager.RegisterFunction(new EFunction());

            ICellValue cellValue;
            CellValueFormat newFormat;
            CellValueFactory.Create("=e()", worksheet, new CellIndex(0, 0), CellValueFormat.GeneralFormat, out cellValue, out newFormat);
            ConstantExpression valueExpression = ((FormulaCellValue)cellValue).Value.GetValueAsConstantExpression();
            Assert.AreEqual(valueExpression.NumberValue(), Math.E);
        }

        [TestMethod]
        public void OneOptionalParameterFunctionTest()
        {
            Workbook workbook = new Workbook();
            workbook.Sheets.Add(SheetType.Worksheet);
            Worksheet worksheet = workbook.ActiveWorksheet;

            FunctionManager.RegisterFunction(new RepeatFunction());

            ICellValue cellValue;
            CellValueFormat newFormat;
            CellValueFactory.Create("=repeat.string(\"Text\"{ls} 2)".ReplaceCultureDependentSeparators(), worksheet, new CellIndex(0, 0), CellValueFormat.GeneralFormat, out cellValue, out newFormat);
            ConstantExpression valueExpression = ((FormulaCellValue)cellValue).Value.GetValueAsConstantExpression();
            Assert.AreEqual(valueExpression.GetValueAsString(), "TextText");
        }

        [TestMethod]
        public void NaryFunctionTest()
        {
            Workbook workbook = new Workbook();
            workbook.Sheets.Add(SheetType.Worksheet);
            Worksheet worksheet = workbook.ActiveWorksheet;

            FunctionManager.RegisterFunction(new AddFunction());
            FunctionManager.RegisterFunction(new AddNumbersFunction());

            ICellValue cellValue;
            CellValueFormat newFormat;
            ConstantExpression valueExpression;

            CellValueFactory.Create("=Add(1{ls} 2{ls} 3{ls} 4{ls} \"5\")".ReplaceCultureDependentSeparators(), worksheet, new CellIndex(0, 0), CellValueFormat.GeneralFormat, out cellValue, out newFormat);
            valueExpression = ((FormulaCellValue)cellValue).Value.GetValueAsConstantExpression();
            Assert.AreEqual(valueExpression.NumberValue(), 15);

            CellValueFactory.Create("=AddNumbers(1{ls} 2{ls} 3{ls} 4{ls} \"5\")".ReplaceCultureDependentSeparators(), worksheet, new CellIndex(0, 0), CellValueFormat.GeneralFormat, out cellValue, out newFormat);
            valueExpression = ((FormulaCellValue)cellValue).Value.GetValueAsConstantExpression();
            Assert.AreEqual(valueExpression.NumberValue(), 10);
        }

        #endregion
    }
    
    #region Custom Function Classes

    //Nullary function
    public class EFunction : FunctionBase
    {
        public static readonly string FunctionName = "E";
        private static readonly FunctionInfo Info;

        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        static EFunction()
        {
            string description = "Description of REPEAT.STRING comes here.";

            Info = new FunctionInfo(FunctionName, FunctionCategory.Text, description);
        }

        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<RadExpression> context)
        {
            return new NumberExpression(Math.E);
        }
    }

    //Function with 1 required and 1 optional parameter
    public class RepeatFunction : FunctionWithArguments
    {
        public static readonly string FunctionName = "REPEAT.STRING";
        private static readonly FunctionInfo Info;

        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        static RepeatFunction()
        {
            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                 new ArgumentInfo("Text", "is the text you want to repeat.", ArgumentType.Text)
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("NumberTimes", "is a positive number specifying the number of times to repeat text.", ArgumentType.Number)
            };

            string description = "Description of REPEAT.STRING comes here.";

            Info = new FunctionInfo(FunctionName, FunctionCategory.Text, description, requiredArguments, optionalArguments, 1, false);
        }

        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<object> context)
        {
            string text = context.Arguments[0].ToString();

            double numberTimes = 1;
            if (context.Arguments.Length > 1)
            {
                numberTimes = (double)context.Arguments[1];
            }

            StringBuilder builder = new StringBuilder();

            if (numberTimes < 0)
            {
                return ErrorExpressions.ValueError;
            }

            for (int i = 0; i < numberTimes; i++)
            {
                builder.Append(text);
            }

            return new StringExpression(builder.ToString());
        }
    }

    //Nary function
    public class AddFunction : NumbersInFunction
    {
		public static readonly string FunctionName = "ADD";
        private static readonly FunctionInfo Info;

		public override string Name
		{
			get
			{
				return FunctionName;
			}
		}

        public override FunctionInfo FunctionInfo
        {
            get
            {
                return Info;
            }
        }

        static AddFunction()
        {
            string description = "Adds all the numbers in range of cells.";

            IEnumerable<ArgumentInfo> requiredArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Number", "number1, number2,... are the numbers to sum. Logical values and text are ignored in cells, included if typed as arguments.", ArgumentType.Number),
            };

            IEnumerable<ArgumentInfo> optionalArguments = new ArgumentInfo[]
            {
                new ArgumentInfo("Number", "number1, number2,... are the numbers to sum. Logical values and text are ignored in cells, included if typed as arguments.", ArgumentType.Number),
            };

            Info = new FunctionInfo(FunctionName, FunctionCategory.MathTrig, description, requiredArguments, optionalArguments, 254, true);
        }

        protected override RadExpression EvaluateOverride(FunctionEvaluationContext<double> context)
        {
            double result = 0;

            for (int i = 0; i < context.Arguments.Length; i++)
            {
                result += context.Arguments[i];
            }

            return new NumberExpression(result);
        }
    }

    //Nary function that ignores text representation of numbers
    public class AddNumbersFunction : AddFunction
    {
        public static readonly new string FunctionName = "ADDNUMBERS";

        private static ArgumentConversionRules argumentConvertionRules = new ArgumentConversionRules(
            emptyIndirectArgument: ArgumentInterpretation.Ignore,
            boolIndirectArgument: ArgumentInterpretation.Ignore,
            textNumberDirectArgument: ArgumentInterpretation.Ignore,
            textNumberIndirectArgument: ArgumentInterpretation.Ignore,
            nonTextNumberDirectArgument: ArgumentInterpretation.TreatAsError,
            nonTextNumberIndirectArgument: ArgumentInterpretation.Ignore,
            arrayArgument: ArrayArgumentInterpretation.UseAllElements);

        public override string Name
        {
            get
            {
                return FunctionName;
            }
        }

        public override ArgumentConversionRules ArgumentConversionRules
        {
            get
            {
                return argumentConvertionRules;
            }
        }
    }

    #endregion
}
