﻿using System;
using System.Linq;
using System.Collections.Generic;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Expressions.Functions;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Model;
using System.Text;

namespace Telerik.Windows.Documents.Spreadsheet.Tests
{
    public static class FunctionsTestHelper
    {
        #region Constants

        private const double Epsilon = 1E-8;

        #endregion


        #region Methods

        public static void AssertCellValue(string cellInput, string expectedResult, Worksheet worksheet = null, CellIndex cellIndex = null)
        {
            if (worksheet == null)
            {
                worksheet = new Workbook().Worksheets.Add();
            }

            if (cellIndex == null)
            {
                cellIndex = new CellIndex(0, 0);
            }

            string cultureDependentInput = cellInput.ReplaceCultureDependentSeparators();
            worksheet.Cells[cellIndex].SetValue(cultureDependentInput);

            string actualResult;
            ICellValue cellValue = worksheet.Cells[cellIndex].GetValue().Value;
            FormulaCellValue formulaValue = cellValue as FormulaCellValue;

            if (formulaValue != null)
            {
                actualResult = formulaValue.GetResultValueAsCellValue().RawValue;
            }
            else
            {
                actualResult = cellValue.RawValue;
            }
            
            Assert.AreEqual(expectedResult, actualResult);
        }

        public static void AssertErrorCellValue(string cellInput, ErrorExpression error, Worksheet worksheet = null)
        {
            AssertCellValue(cellInput, error.Value, worksheet);
        }

        public static void AssertFunctionResult(FunctionBase function, object expectedResult, params object[] arguments)
        {
            RadExpression[] expressionArguments = ConvertArgumentsToExpressions(arguments);

            RadExpression actualResultExpression = function.Evaluate(new FunctionEvaluationContext<RadExpression>(expressionArguments, null, 0, 0)).GetValue();
            RadExpression expectedResultExpression = ConvertArgumentToExpression(expectedResult);

            Assert.IsTrue(CompareExpressions(expectedResultExpression, actualResultExpression));
        }

        public static void AssertAreEssentiallyEqualFunctionResult(FunctionBase function, object expectedResult, double epsilon, params object[] arguments)
        {
            RadExpression[] expressionArguments = ConvertArgumentsToExpressions(arguments);

            RadExpression actualResultExpression = function.Evaluate(new FunctionEvaluationContext<RadExpression>(expressionArguments, null, 0, 0)).GetValue();
            RadExpression expectedResultExpression = ConvertArgumentToExpression(expectedResult);

            NumberExpression actualNumber = actualResultExpression as NumberExpression;
            NumberExpression expectedNumber = expectedResultExpression as NumberExpression;

            if (actualNumber != null && expectedNumber != null)
            {
                double a = actualNumber.Value;
                double b = expectedNumber.Value;
                bool isTrue = TestHelper.AreDoubleValuesEssentiallyEqual(a, b, epsilon);
                Assert.IsTrue(isTrue);
            }
            else
            {
                Assert.Fail();
            }            
        }

        public static void AssertAreEssentiallyEqualComplexNumberFunctionResult(FunctionBase function, string expectedComplexNumberResult,
            double epsilon, params object[] arguments)
        {
            RadExpression[] expressionArguments = ConvertArgumentsToExpressions(arguments);

            RadExpression actualResultExpression = function.Evaluate(new FunctionEvaluationContext<RadExpression>(expressionArguments, null, 0, 0)).GetValue();
            StringExpression actualResult = actualResultExpression as StringExpression;
            StringExpression expectedResult = new StringExpression(expectedComplexNumberResult);

            if (actualResult != null && expectedResult != null)
            {
                string a = actualResult.Value;
                string b = expectedResult.Value;

                bool areEqual = TestHelper.AreComplexNumbersEssentiallyEqual(a, b, epsilon);

                Assert.IsTrue(areEqual);
            }
            else
            {
                Assert.Fail();
            }
        }

        public static void AssertMultipleTestsForAComplexComplexFunction(FunctionBase complexFunction, double accuracy,
            IEnumerable<KeyValuePair<string, string>> expectedResultPlusFunctionArgumentPairs)
        {
            foreach (KeyValuePair<string, string> test in expectedResultPlusFunctionArgumentPairs)
            {
                FunctionsTestHelper.AssertAreEssentiallyEqualComplexNumberFunctionResult(complexFunction,
                test.Key.ReplaceCultureDependentSeparators(), accuracy,
                test.Value.ReplaceCultureDependentSeparators());
            }
        }

        public static void AssertFunctionError(FunctionBase function, ErrorExpression errorExpression, params object[] arguments)
        {
            RadExpression[] expressionArguments = ConvertArgumentsToExpressions(arguments);
            RadExpression result = function.Evaluate(new FunctionEvaluationContext<RadExpression>(expressionArguments, null, 0, 0));

            Assert.AreEqual(errorExpression, result);
        }

        public static void AssertFunctionReferenceErrorVariationsTests(string functionName, int minArgumentsCount, int maxArgumentsCount, string b2CellValue)
        {
            Worksheet worksheet = new Workbook().Worksheets.Add();
            worksheet.Cells[1, 1].SetValue(b2CellValue);

            for (int argumentsCount = minArgumentsCount; argumentsCount <= maxArgumentsCount; argumentsCount++)
            {
                string[] functionVariations = FunctionsTestHelper.GenerateFunctionsWithTwoArgumentVariations(functionName, argumentsCount, "#REF!", "B2").ToArray();

                for (int i = 1; i < functionVariations.Length; i++)
                {
                    string functionExpression = functionVariations[i];
                    FunctionsTestHelper.AssertErrorCellValue(functionExpression, ErrorExpressions.ReferenceError, worksheet);
                }
            }
        }

        public static IEnumerable<string> GenerateFunctionsWithTwoArgumentVariations(string functionName, int numberOfArgumentPositions, string firstArgument, string secondArgument)
        {
            StringBuilder builder = new StringBuilder();

            foreach (bool[] variation in TestHelper.GenerateBinaryVariations(numberOfArgumentPositions))
            {
                string[] args = FunctionsTestHelper.GetFunctionArgumentsFromVariation(variation, firstArgument, secondArgument);

                builder.Clear();
                builder.AppendFormat("={0}({1}", functionName, args[0]);

                for (int i = 1; i < numberOfArgumentPositions; i++)
                {
                    builder.AppendFormat("{0}{1}", "{ls}", args[i]); 
                }

                builder.Append(")");

                yield return builder.ToString();
            }            
        }

        public static void AssertFunctionResultWithReferenceToArgumentsInSeparateCells(FunctionBase function, string expectedResult, params string[] argumentInputs)
        {
            Worksheet worksheet = CreateWorksheetWithInputInColumnB(argumentInputs);
            string a1CellInput = ConstructFunctionArgumentStringForColumnB(function.Name, argumentInputs.Length);
            FunctionsTestHelper.AssertCellValue(a1CellInput, expectedResult, worksheet);
        }

        private static string ConstructFunctionArgumentStringForColumnB(string functionName, int argumentCount)
        {
            string resultString = "=";

            resultString += functionName;
            resultString += "(";

            for (int i = 1; i <= argumentCount; i++)
            {
                resultString += string.Format("B{0}", i);

                if (i != argumentCount)
                {
                    resultString += ",";
                }
            }

            resultString += ")";

            return resultString;
        }

        private static Worksheet CreateWorksheetWithInputInColumnB(string[] argumentInputs)
        {
            Worksheet worksheet = new Workbook().Worksheets.Add();

            for (int i = 0; i < argumentInputs.Length; i++)
            {
                worksheet.Cells[i, 1].SetValue(argumentInputs[i]);
            }

            return worksheet;
        }

        private static string[] GetFunctionArgumentsFromVariation(bool[] variation, string argumentWhenTrue, string argumentWhenFalse)
        {
            int count = variation.Length;
            string[] arguments = new string[count];

            for (int i = 0; i < count; i++)
            {
                arguments[i] = variation[i] ? argumentWhenTrue : argumentWhenFalse;
            }

            return arguments;
        }

        private static RadExpression[] ConvertArgumentsToExpressions(params object[] arguments)
        {
            List<RadExpression> expressionArguments = new List<RadExpression>();

            for (int i = 0; i < arguments.Length; i++)
            {
                RadExpression expression = ConvertArgumentToExpression(arguments[i]);
                if (expression != null)
                {
                    expressionArguments.Add(expression);
                }
            }

            return expressionArguments.ToArray();
        }

        private static RadExpression ConvertArgumentToExpression(object argument)
        {
            RadExpression argAsExpression = argument as RadExpression;
            if (argAsExpression != null)
            {
                return argAsExpression;
            }

            object[,] argAsArray = argument as object[,];
            if (argAsArray != null)
            {
                int rowCount = argAsArray.GetLength(0);
                int columnCount = argAsArray.GetLength(1);

                RadExpression[,] expressionArray = new RadExpression[rowCount, columnCount];

                for (int rowIndex = 0; rowIndex < rowCount; rowIndex++)
                {
                    for (int columnIndex = 0; columnIndex < columnCount; columnIndex++)
                    {
                        RadExpression result = ConvertArgumentToExpression(argAsArray[rowIndex, columnIndex]);
                        expressionArray[rowIndex, columnIndex] = result;
                    }
                }

                return new ArrayExpression(expressionArray);
            }

            string argumentToString = argument.ToString().ReplaceCultureDependentSeparators();

            string argAsString = argument as string;
            if (argAsString != null)
            {
                return new StringExpression(argumentToString);
            }

            double argAsDouble;
            if (Double.TryParse(argumentToString, out argAsDouble))
            {
                return new NumberExpression(argAsDouble);
            }

            bool argAsBool;
            if (Boolean.TryParse(argumentToString, out argAsBool))
            {
                return argAsBool ? BooleanExpression.True : BooleanExpression.False;
            }

            return new StringExpression(argument.ToString());
        }

        private static bool CompareExpressions(RadExpression expected, RadExpression actual)
        {
            NumberExpression actualNumber = actual as NumberExpression;
            NumberExpression expectedNumber = expected as NumberExpression;

            if (actualNumber != null && expectedNumber != null)
            {
                double a = actualNumber.Value;
                double b = expectedNumber.Value;

                // TODO:
                return Math.Abs(a - b) <= ((Math.Abs(a) > Math.Abs(b) ? Math.Abs(b) : Math.Abs(a)) * Epsilon);
            }

            StringExpression actualString = actual as StringExpression;
            StringExpression expectedString = expected as StringExpression;

            if (actualString != null && expectedString != null)
            {
                return String.Compare(actualString.Value, expectedString.Value) == 0;
            }

            BooleanExpression actualBoolean = actual as BooleanExpression;
            BooleanExpression expectedBoolean = expected as BooleanExpression;

            if (actualBoolean != null && expectedBoolean != null)
            {
                return actualBoolean.Value == expectedBoolean.Value;
            }

            return false;
        }        

        #endregion
    }
}
