﻿using System;
using System.Text;
using System.Windows;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;
using System.Diagnostics;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx;
using System.IO;

namespace Telerik.Windows.Documents.Spreadsheet.Tests.FunctionsTests
{
    [TestClass]
    public class LookupFunctionsTests
    {
        #region Constants

        private static readonly string[,] sampleIndexFunctionCellInput = new string[,]
            {
                {"1", "2", "4"},
                {"8", "16", "32"},
            };

        private const string SampleIndexFunctionArrayInput = "{1{als}2{als}4{ars}8{als}16{als}32}";

        private static readonly string[,] sampleLookupFunctionCellInput = new string[,]
            {
                {"0", "2", "a", "Fu", "s", "Ts"},
                {"1", null, "b", null, null, null},
                {"2", null, "c", null, null, null},
                {"2", null, "d", null, null, null},
            };

        private static readonly string[,] invertedSampleLookupFunctionCellInput = new string[,]
            {
                {"0", "1", "2", "2"},
                {"2", null, null, null},
                {"a", "b", "c", "d"},
                {"Fu", null, null, null},
                {"s", null, null, null},
                {"Ts", null, null, null},
            };

        private static readonly string[,] sampleDescendingLookupFunctionCellInput = new string[,]
            {
                {"2", null, "d", null, null, null},
                {"2", null, "c", null, null, null},
                {"1", null, "b", null, null, null},
                {"0", "2", "a", "Fu", "s", "Ts"},
            };

        #endregion

        #region Choose function tests

        [TestMethod]
        public void ChooseFunctionTestWholeNumber()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cell = worksheet.Cells[2, 3];
            string function = "=CHOOSE(1,10)".ReplaceCultureDependentSeparators();
            cell.SetValue(function);

            ICellValue value = cell.GetValue().Value;
            Assert.AreEqual("10", value.GetResultValueAsString(CellValueFormat.GeneralFormat));
        }

        [TestMethod]
        public void ChooseFunctionTestNumberWithDecimalSeparator()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            string function = "=CHOOSE(1,0.10)".ReplaceCultureDependentSeparators();
            CellSelection cell = worksheet.Cells[2, 3];
            cell.SetValue(function);

            ICellValue value = cell.GetValue().Value;
            Assert.AreEqual("0.1", value.GetResultValueAsString(CellValueFormat.GeneralFormat));
        }

        [TestMethod]
        public void ChooseFunctionTestTooBigIndex()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            string function = "=CHOOSE(10,2,3,4)".ReplaceCultureDependentSeparators();
            CellSelection cell = worksheet.Cells[12, 33];
            cell.SetValue(function);

            ICellValue value = cell.GetValue().Value;
            Assert.AreEqual("#VALUE!", value.GetResultValueAsString(CellValueFormat.GeneralFormat));
        }

        [TestMethod]
        public void ChooseFunctionTestStringCollection()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            string function = "=CHOOSE(1,\"first\",\"second\",\"third\")".ReplaceCultureDependentSeparators();
            CellSelection cell = worksheet.Cells[7, 31];
            cell.SetValue(function);

            ICellValue value = cell.GetValue().Value;
            Assert.AreEqual("first", value.GetResultValueAsString(CellValueFormat.GeneralFormat));
        }

        #endregion

        #region Indirect function tests

        [TestMethod]
        public void IndirectFunctionCreateingSingleCellReferenceToAnotherWorksheet()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            string function = "=Indirect(\"A1\")";
            worksheet.Cells[2, 3].SetValue(function);

            worksheet.Cells[0, 0].SetValue(5);

            ICellValue value = worksheet.Cells[2, 3].GetValue().Value;
            Assert.AreEqual("5", value.GetResultValueAsString(CellValueFormat.GeneralFormat));

            worksheet.Cells[0, 0].SetValue(3);

            value = worksheet.Cells[2, 3].GetValue().Value;
            Assert.AreEqual("3", value.GetResultValueAsString(CellValueFormat.GeneralFormat));
        }

        [TestMethod]
        public void IndirectFunctionCreateingSingleCellReference()
        {
            Worksheet worksheet1 = TestHelper.CreateWorksheet();
            var worksheet2 = worksheet1.Workbook.Worksheets.Add();

            string function = "=Indirect(\"Sheet2!A1\")";
            worksheet1.Cells[2, 3].SetValue(function);

            worksheet2.Cells[0, 0].SetValue(5);

            ICellValue value = worksheet1.Cells[2, 3].GetValue().Value;
            Assert.AreEqual("5", value.GetResultValueAsString(CellValueFormat.GeneralFormat));

            worksheet2.Cells[0, 0].SetValue(3);

            value = worksheet1.Cells[2, 3].GetValue().Value;
            Assert.AreEqual("3", value.GetResultValueAsString(CellValueFormat.GeneralFormat));
        }

        [TestMethod]
        public void IndirectFunctionCreateingCellRangeReference()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            string function = "=Indirect(\"A1:B1\")";
            worksheet.Cells[2, 3].SetValue(function);

            worksheet.Cells[0, 0].SetValue(5);

            ICellValue value = worksheet.Cells[2, 3].GetValue().Value;
            Assert.AreEqual("5", value.GetResultValueAsString(CellValueFormat.GeneralFormat));

            worksheet.Cells[0, 0].SetValue(3);

            value = worksheet.Cells[2, 3].GetValue().Value;
            Assert.AreEqual("3", value.GetResultValueAsString(CellValueFormat.GeneralFormat));
        }

        [TestMethod]
        public void IndirectFunctionCreateingCellRangeSumContainingValuesReference()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            string function = "=Sum(Indirect(\"A1:B1\"))";
            worksheet.Cells[2, 3].SetValue(function);

            worksheet.Cells[0, 0].SetValue(5);
            worksheet.Cells[0, 1].SetValue(5);

            ICellValue value = worksheet.Cells[2, 3].GetValue().Value;
            Assert.AreEqual("10", value.GetResultValueAsString(CellValueFormat.GeneralFormat));

            worksheet.Cells[0, 0].SetValue(3);

            value = worksheet.Cells[2, 3].GetValue().Value;
            Assert.AreEqual("8", value.GetResultValueAsString(CellValueFormat.GeneralFormat));
        }

        [TestMethod]
        public void IndirectFunctionCreateingCellRangesSumContainingValuesReference()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            string function = "=Sum(Indirect(\"A1:B1,C1:D1\"))";
            worksheet.Cells[2, 3].SetValue(function);

            ICellValue value = worksheet.Cells[2, 3].GetValue().Value;
            Assert.AreEqual("#REF!", value.GetResultValueAsString(CellValueFormat.GeneralFormat));

        }

        [TestMethod]
        public void IndirectFunctionCreateingInvalidReferenceString()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            string function = "=Indirect(\"qwerty\")";
            worksheet.Cells[2, 3].SetValue(function);

            ICellValue value = worksheet.Cells[2, 3].GetValue().Value;
            Assert.AreEqual("#REF!", value.GetResultValueAsString(CellValueFormat.GeneralFormat));
        }

        [TestMethod]
        public void IndirectFunctionCreateingEmptyReferenceString()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            string function = "=Indirect(\"\")";
            worksheet.Cells[2, 3].SetValue(function);

            ICellValue value = worksheet.Cells[2, 3].GetValue().Value;
            Assert.AreEqual("#REF!", value.GetResultValueAsString(CellValueFormat.GeneralFormat));
        }

        [TestMethod]
        public void IndirectFunctionCreateingEmptyReferenceStringFromCellReference()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            string function = "=Indirect(A1)";
            worksheet.Cells[2, 3].SetValue(function);

            ICellValue value = worksheet.Cells[2, 3].GetValue().Value;
            Assert.AreEqual("#REF!", value.GetResultValueAsString(CellValueFormat.GeneralFormat));
        }

        [TestMethod]
        [ExpectedException(typeof(ParseException))]
        public void IndirectFunctionCreateingNoParameters()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            string function = "=Indirect()";
            worksheet.Cells[2, 3].SetValue(function);

            ICellValue value = worksheet.Cells[2, 3].GetValue().Value;
            Assert.AreEqual("#REF!", value.GetResultValueAsString(CellValueFormat.GeneralFormat));
        }

        [TestMethod]
        public void IndirectFunctionCreateingCyclicReference()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            string function = "=Indirect(\"A1\")";
            worksheet.Cells[0, 0].SetValue(function);

            ICellValue value = worksheet.Cells[0, 0].GetValue().Value;

            Assert.AreEqual(CellValueType.Error, value.ResultValueType);
            Assert.AreEqual("0", value.GetResultValueAsString(CellValueFormat.GeneralFormat));
        }

        #endregion


        #region Index function tests

        [TestMethod]
        public void IndexFunctionReferenceCheckChangingTheReferedValueInvalidateTheFunction()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            string function = "=INDEX(A1:B2{ls}1{ls}1)+10".ReplaceCultureDependentSeparators();
            worksheet.Cells[2, 3].SetValue(function);

            ICellValue value = worksheet.Cells[2, 3].GetValue().Value;
            Assert.AreEqual("10", value.GetResultValueAsString(CellValueFormat.GeneralFormat));

            worksheet.Cells[0, 0].SetValue(3);

            value = worksheet.Cells[2, 3].GetValue().Value;
            Assert.AreEqual("13", value.GetResultValueAsString(CellValueFormat.GeneralFormat));
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadCorrectnessTests()
        {
            string[,] cellsInput = LookupFunctionsTests.sampleIndexFunctionCellInput;

            Worksheet worksheet = LookupFunctionsTests.CreateWorksheetScenario(cellsInput);

            for (int row = 0; row < cellsInput.GetLength(0); row++)
            {
                for (int column = 0; column < cellsInput.GetLength(1); column++)
                {
                    string a1CellInput = String.Format("=Index(B2:D3{0}{1}{0}{2})", "{ls}", row + 1, column + 1);
                    FunctionsTestHelper.AssertCellValue(a1CellInput, cellsInput[row, column], worksheet);
                }
            }
        }

        [TestMethod]
        public void IndexFunctionFractionValuesCorrectnessTests()
        {
            string[,] cellsInput = LookupFunctionsTests.sampleIndexFunctionCellInput;

            Worksheet worksheet = LookupFunctionsTests.CreateWorksheetScenario(cellsInput);

            double delta = 0.2;

            for (int row = 0; row < cellsInput.GetLength(0); row++)
            {
                for (int column = 0; column < cellsInput.GetLength(1); column++)
                {
                    double rowFraction = (row + 1) * delta - (int) ((row + 1) * delta);
                    double columnFraction = (column + 1) * delta - (int) ((column + 1) * delta);
                    string a1CellInput = String.Format("=Index(B2:D3{0}{1}{0}{2})", "{ls}", row + 1 + rowFraction, column + 1 + columnFraction);
                    FunctionsTestHelper.AssertCellValue(a1CellInput, cellsInput[row, column], worksheet);
                }
            }
        }

        [TestMethod]
        public void IndexFunctionArrayOverloadCorrectnessTests()
        {
            Worksheet worksheet = new Workbook().Worksheets.Add();

            string[,] expectedCellValues = LookupFunctionsTests.sampleIndexFunctionCellInput;
            string arrayInput = LookupFunctionsTests.SampleIndexFunctionArrayInput;

            for (int row = 0; row < expectedCellValues.GetLength(0); row++)
            {
                for (int column = 0; column < expectedCellValues.GetLength(1); column++)
                {
                    string a1CellInput = String.Format("=Index({0}{1}{2}{1}{3})", arrayInput, "{ls}", row + 1, column + 1);
                    FunctionsTestHelper.AssertCellValue(a1CellInput, expectedCellValues[row, column], worksheet);
                }
            }
        }

        [TestMethod]
        public void IndexFunctionReferencingArrayCellValueTest()
        {
            string b2CellInput = string.Format("={0}", LookupFunctionsTests.SampleIndexFunctionArrayInput);
            Worksheet worksheet = CreateWorksheetScenario(b2CellInput);
            string a1CellInput = "=Index(B2{ls}1{ls}1)";
            string expectedValue = "1";

            FunctionsTestHelper.AssertCellValue(a1CellInput, expectedValue, worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferencingArrayCellValueErrorTest()
        {
            string b2CellInput = string.Format("={0}", LookupFunctionsTests.SampleIndexFunctionArrayInput);
            Worksheet worksheet = CreateWorksheetScenario(b2CellInput);
            string a1CellInput = "=Index(B2{ls}2{ls}2)";

            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ReferenceError, worksheet);
        }

        [TestMethod]
        public void IndexFunctionArrayOverloadArrayRowColumnArgumentTest()
        {
            string arrayFirstArgument = LookupFunctionsTests.SampleIndexFunctionArrayInput;
            string arraySecondArgument = "{2{als}2}";
            string a1CellInput = String.Format("=Index({0}{1}{2})", arrayFirstArgument, "{ls}", arraySecondArgument);
            string expectedValue = "8";

            FunctionsTestHelper.AssertCellValue(a1CellInput, expectedValue);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadArrayRowColumnArgumentTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            string arraySecondArgument = "{2{als}2}";
            string a1CellInput = String.Format("=Index(B2:D3{0}{1})", "{ls}", arraySecondArgument);

            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ReferenceError, worksheet);
        }

        [TestMethod]
        public void IndexFunctionArrayOverloadSkippedRowArgumentTest()
        {
            Worksheet worksheet = new Workbook().Worksheets.Add();

            string[] indexArguments = new string[]
            {
                LookupFunctionsTests.SampleIndexFunctionArrayInput,
                string.Empty,
                "2"
            };

            string a1CellInput = String.Format("=Index({0}{1}{2}{1}{3})", indexArguments[0], "{ls}", indexArguments[1], indexArguments[2]);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "2", worksheet);
        }

        // TODO: Not working scenario - Excel gets from A1 only the first array element, while currently RadSpreadsheet gets all elements.
        //[TestMethod]
        //public void IndexFunctionWithArrayArgumentIndirectUseInSumFunctionTest()
        //{
        //    string[] indexArguments = new string[]
        //    {
        //        LookupFunctionsTests.SampleIndexFunctionArrayInput,
        //        string.Empty,
        //        "2"
        //    };

        //    string a1CellInput = String.Format("=Index({0}{1}{2}{1}{3})", indexArguments[0], "{ls}", indexArguments[1], indexArguments[2]);
        //    Worksheet worksheet = new Workbook().Worksheets.Add();
        //    worksheet.Cells[0, 0].SetValue(a1CellInput.ReplaceCultureDependentSeparators());

        //    string b2CellInput = "=Sum(A1)";
        //    FunctionsTestHelper.AssertCellValue(b2CellInput, "2", worksheet, new CellIndex(1,1)); // Currently Spreadsheet returns "18" instead of "2" as Sum function result
        //}

        [TestMethod]
        public void IndexFunctionWithArrayArgumentDirectUseInSumFunctionTest()
        {
            string[] indexArguments = new string[]
            {
                LookupFunctionsTests.SampleIndexFunctionArrayInput,
                string.Empty,
                "2"
            };

            string a1CellInput = String.Format("=Sum(Index({0}{1}{2}{1}{3}))", indexArguments[0], "{ls}", indexArguments[1], indexArguments[2]);

            FunctionsTestHelper.AssertCellValue(a1CellInput, "18");
        }

        [TestMethod]
        public void IndexFunctionNegativeRowColumnErrorTest()
        {
            FunctionsTestHelper.AssertErrorCellValue("=Index(B2:D3{ls}-1{ls}0)", ErrorExpressions.ValueError);
            FunctionsTestHelper.AssertErrorCellValue("=Index(B2:D3{ls}-1{ls}-2)", ErrorExpressions.ValueError);
            FunctionsTestHelper.AssertErrorCellValue("=Index({1{als}2}{ls}0{ls}-1)", ErrorExpressions.ValueError);
            FunctionsTestHelper.AssertErrorCellValue("=Index(B2:D3{ls}1{ls}-1)", ErrorExpressions.ValueError);
            FunctionsTestHelper.AssertErrorCellValue("=Index({3{als}2}{ls}-1{ls}1)", ErrorExpressions.ValueError);
        }

        [TestMethod]
        public void IndexFunctionNegativeFractionCloseToZeroErrorTest()
        {
            FunctionsTestHelper.AssertErrorCellValue("=Index({3{als}2}{ls}-0{ds}1{ls}1)", ErrorExpressions.ValueError);
            FunctionsTestHelper.AssertErrorCellValue("=Index(B2:D3{ls}-0{ds}1{ls}1)", ErrorExpressions.ValueError);
        }

        [TestMethod]
        public void IndexFunctionInvalidArgumentTypeTest()
        {
            FunctionsTestHelper.AssertErrorCellValue("=Index(\"some string value\"{ls}1{ls}1)", ErrorExpressions.ValueError);
            FunctionsTestHelper.AssertErrorCellValue("=Index(TRUE{ls}1{ls}1)", ErrorExpressions.ValueError);
            FunctionsTestHelper.AssertErrorCellValue("=Index(FALSE{ls}1{ls}1)", ErrorExpressions.ValueError);
        }

        [TestMethod]
        public void IndexFunctionNumberArgumentTest()
        {
            FunctionsTestHelper.AssertCellValue("=Index(3{ls}1{ls}1)", "3");
            FunctionsTestHelper.AssertCellValue("=Index(3{ls}{ls}1)", "3");
            FunctionsTestHelper.AssertCellValue("=Index(3{ls}0{ls}1)", "3");
            FunctionsTestHelper.AssertCellValue("=Index(3{ls}1{ls}0)", "3");
            FunctionsTestHelper.AssertCellValue("=Index(3{ls}1)", "3");
        }

        [TestMethod]
        public void IndexFunctionReferenceErrorTests()
        {
            FunctionsTestHelper.AssertErrorCellValue("=Index(C4:F7{ls}5{ls}1)", ErrorExpressions.ReferenceError);
            FunctionsTestHelper.AssertErrorCellValue("=Index(C4:F7{ls}1{ls}5)", ErrorExpressions.ReferenceError);

            FunctionsTestHelper.AssertErrorCellValue(
                String.Format("=Index({0}{1}{2}{1}{3})", LookupFunctionsTests.SampleIndexFunctionArrayInput, "{ls}", 3, 1), ErrorExpressions.ReferenceError);

            FunctionsTestHelper.AssertErrorCellValue(
                String.Format("=Index({0}{1}{2}{1}{3})", LookupFunctionsTests.SampleIndexFunctionArrayInput, "{ls}", 1, 4), ErrorExpressions.ReferenceError);
        }

        [TestMethod]
        public void IndexFunctionArrayOverloadZeroArgumentTests()
        {
            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index({0}{1}{2}{1}{3}))", LookupFunctionsTests.SampleIndexFunctionArrayInput, "{ls}", 1, 0), "7");
            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index({0}{1}{2}{1}{3}))", LookupFunctionsTests.SampleIndexFunctionArrayInput, "{ls}", 2, 0), "56");
            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index({0}{1}{2}{1}{3}))", LookupFunctionsTests.SampleIndexFunctionArrayInput, "{ls}", 0, 1), "9");
            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index({0}{1}{2}{1}{3}))", LookupFunctionsTests.SampleIndexFunctionArrayInput, "{ls}", 0, 2), "18");
            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index({0}{1}{2}{1}{3}))", LookupFunctionsTests.SampleIndexFunctionArrayInput, "{ls}", 0, 3), "36");
            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index({0}{1}{2}{1}{3}))", LookupFunctionsTests.SampleIndexFunctionArrayInput, "{ls}", 0, 0), "63");
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadZeroArgumentValueErrorTests()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index({0}{1}{2}{1}{3}))", "B2:D3", "{ls}", 1, 0), ErrorExpressions.ValueError, worksheet);
            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index({0}{1}{2}{1}{3}))", "B2:D3", "{ls}", 2, 0), ErrorExpressions.ValueError, worksheet);
            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index({0}{1}{2}{1}{3}))", "B2:D3", "{ls}", 0, 1), ErrorExpressions.ValueError, worksheet);
            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index({0}{1}{2}{1}{3}))", "B2:D3", "{ls}", 0, 2), ErrorExpressions.ValueError, worksheet);
            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index({0}{1}{2}{1}{3}))", "B2:D3", "{ls}", 0, 3), ErrorExpressions.ValueError, worksheet);
            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index({0}{1}{2}{1}{3}))", "B2:D3", "{ls}", 0, 0), ErrorExpressions.ValueError, worksheet);
            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index({0}{1}{2}))", "B2:D3", "{ls}", -1), ErrorExpressions.ValueError, worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadZeroArgumentValueWithSkippedArgumentErrorTests()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index({0}{1}{1}{2}))", "B2:D3", "{ls}", 0), ErrorExpressions.ValueError, worksheet);
            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index({0}{1}{1}{2}))", "B2:D3", "{ls}", 1), ErrorExpressions.ValueError, worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadZeroArgumentReferenceErrorTests()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index({0}{1}{2}))", "B2:D3", "{ls}", 0), ErrorExpressions.ReferenceError, worksheet);
            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index({0}{1}{2}))", "B2:D3", "{ls}", 1), ErrorExpressions.ReferenceError, worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceResultUsedInSumFunctionTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertCellValue("=Sum(Index(B2{ls}1{ls}1):Index(D3{ls}1{ls}1))", "63", worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadZeroArgumentWithRowVectorReferenceCorrectnessTests()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index(B3:D3{0}{1}{0}{2}))", "{ls}", 0, 1), "8", worksheet);
            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index(B3:D3{0}{1}{0}{2}))", "{ls}", string.Empty, 1), "8", worksheet);
            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index(B3:D3{0}{1}))", "{ls}", 1), "8", worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadZeroArgumentWithRowVectorReferenceErrorTests()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index(B3:D3{0}{1}{0}{2}))", "{ls}", 1, 0), ErrorExpressions.ValueError, worksheet);
            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index(B3:D3{0}{1}{0}{2}))", "{ls}", 1, string.Empty), ErrorExpressions.ValueError, worksheet);
            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index(B3:D3{0}{1}{0}{2}))", "{ls}", 0, 0), ErrorExpressions.ValueError, worksheet);
            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index(B3:D3{0}{1}{0}{2}))", "{ls}", 0, string.Empty), ErrorExpressions.ValueError, worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadZeroArgumentWithRowVectorSkippedArgumentsReferenceErrorTests()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index(B3:D3{0}{1}{0}{2}))", "{ls}", string.Empty, string.Empty), ErrorExpressions.ValueError, worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadZeroArgumentWithColumnVectorReferenceCorrectnessTests()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index(D2:D3{0}{1}))", "{ls}", 1), "4", worksheet);
            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index(D2:D3{0}{1}{0}{2}))", "{ls}", 1, 0), "4", worksheet);
            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index(D2:D3{0}{1}{0}{2}))", "{ls}", 1, string.Empty), "4", worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadZeroArgumentWithColumnVectorReferenceErrorTests()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index(D2:D3{0}{1}{0}{2}))", "{ls}", 0, 1), ErrorExpressions.ValueError, worksheet);
            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index(D2:D3{0}{1}{0}{2}))", "{ls}", 0, 0), ErrorExpressions.ValueError, worksheet);
            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index(D2:D3{0}{1}{0}{2}))", "{ls}", 0, string.Empty), ErrorExpressions.ValueError, worksheet);
            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index(D2:D3{0}{1}{0}{2}))", "{ls}", string.Empty, string.Empty), ErrorExpressions.ValueError, worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadZeroArgumentWithColumnVectorSkippedArgumentsReferenceErrorTests()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Sum(Index(D2:D3{0}{1}{0}{2}))", "{ls}", string.Empty, 1), ErrorExpressions.ValueError, worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadZeroArgumentWithPointReferenceTests()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index(D2{0}{1}))", "{ls}", 1), "4", worksheet);
            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index(D2{0}{1}{0}{2}))", "{ls}", 1, 0), "4", worksheet);
            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index(D2{0}{1}{0}{2}))", "{ls}", 1, string.Empty), "4", worksheet);
            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index(D2{0}{1}{0}{2}))", "{ls}", 0, 1), "4", worksheet);
            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index(D2{0}{1}{0}{2}))", "{ls}", string.Empty, 1), "4", worksheet);
            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index(D2{0}{1}{0}{2}))", "{ls}", 0, 0), "4", worksheet);
            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index(D2{0}{1}{0}{2}))", "{ls}", 0, string.Empty), "4", worksheet);
            FunctionsTestHelper.AssertCellValue(String.Format("=Sum(Index(D2{0}{1}{0}{2}))", "{ls}", string.Empty, string.Empty), "4", worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadAreaArgumentOneAreaCorrectnessTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertCellValue(String.Format("=Index(D2{0}{1}{0}{2}{0}{3})", "{ls}", 1, 1, 1), "4", worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadAreaArgumentOneAreaErrorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Index(D2{0}{1}{0}{2}{0}{3})", "{ls}", 1, 2, 1), ErrorExpressions.ReferenceError, worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadAreaArgumentMultupleAreasFirstAreaUsedCorrectnessTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertCellValue(String.Format("=Index((D2{0}B2:D3{0}C3{0}C4{0}B3{0}B4{0}C2:D2){0}{1}{0}{2}{0}{3})", "{ls}", 1, 1, 1), "4", worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadAreaArgumentMultupleAreasFirstAreaUsedErrorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Index((D2{0}B2:D3{0}C3{0}C4{0}B3{0}B4{0}C2:D2){0}{1}{0}{2}{0}{3})", "{ls}", 1, 2, 1), ErrorExpressions.ReferenceError, worksheet);
        }

        [TestMethod]
        public void IndexFunctionArrayOverloadAreaArgumentOneAreaCorrectnessTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertCellValue(String.Format("=Index({4}{0}{1}{0}{2}{0}{3})", "{ls}", 2, 2, 1, SampleIndexFunctionArrayInput), "16", worksheet);
        }

        [TestMethod]
        public void IndexFunctionArrayOverloadAreaArgumentOneAreaErrorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Index({4}{0}{1}{0}{2}{0}{3})", "{ls}", 2, 2, 2, SampleIndexFunctionArrayInput), ErrorExpressions.ReferenceError, worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadAreaArgumentMultupleAreasSecondAreaUsedTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertCellValue(String.Format("=Index((D2{0}B2:D3{0}C3{0}C4{0}B3{0}B4{0}C2:D2){0}{1}{0}{2}{0}{3})", "{ls}", 1, 2, 2), "2", worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadAreaArgumentMultupleAreasBoundaryAreaUsedTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertCellValue(String.Format("=Index((D2{0}B2:D3{0}C3{0}C4{0}B3{0}B4{0}C2:D2){0}{1}{0}{2}{0}{3})", "{ls}", 1, 2, 7), "4", worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadAreaArgumentMultupleAreasOutOfBoundaryAreaErrorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertErrorCellValue(String.Format("=Index((D2{0}B2:D3{0}C3{0}C4{0}B3{0}B4{0}C2:D2){0}{1}{0}{2}{0}{3})", "{ls}", 1, 2, 8), ErrorExpressions.ReferenceError, worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadAreaArgumentMultupleAreasRepeatingSheetNameCorrectnessTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertCellValue(String.Format("=Index(({4}D2{0}{4}B2:D3){0}{1}{0}{2}{0}{3})", "{ls}", 1, 2, 2, worksheet.Name + "!"), "2", worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadAreaArgumentMultupleAreasSingleSheetNameFirstCorrectnessTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertCellValue(String.Format("=Index(({4}D2{0}B2:D3){0}{1}{0}{2}{0}{3})", "{ls}", 1, 2, 2, worksheet.Name + "!"), "2", worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadAreaArgumentMultupleAreasSingleSheetNameSecondCorrectnessTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);

            FunctionsTestHelper.AssertCellValue(String.Format("=Index((D2{0}{4}B2:D3){0}{1}{0}{2}{0}{3})", "{ls}", 1, 2, 2, worksheet.Name + "!"), "2", worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadAreaArgumentMultupleAreasDifferentSheetNamesValueErrorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);
            Worksheet secondWorksheet = worksheet.Workbook.Worksheets.Add();
            string cellFormulaValue = String.Format("=Index(({4}D2{0}{5}B2:D3){0}{1}{0}{2}{0}{3})", "{ls}", 1, 2, 2, worksheet.Name + "!", secondWorksheet.Name + "!");

            FunctionsTestHelper.AssertErrorCellValue(cellFormulaValue, ErrorExpressions.ValueError, worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceOverloadAreaArgumentMultupleAreasDifferentSheetNameOutOfAreaRangeStillValueTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(LookupFunctionsTests.sampleIndexFunctionCellInput);
            Worksheet secondWorksheet = worksheet.Workbook.Worksheets.Add();
            string cellFormulaValue = String.Format("=Index(({4}D2{0}{5}B2:D3){0}{1}{0}{2}{0}{3})", "{ls}", 1, 2, 20, worksheet.Name + "!", secondWorksheet.Name + "!");

            FunctionsTestHelper.AssertErrorCellValue(cellFormulaValue, ErrorExpressions.ValueError, worksheet);
        }

        [TestMethod]
        public void IndexFunctionReferenceErrorVariationsTests()
        {
            FunctionsTestHelper.AssertFunctionReferenceErrorVariationsTests("Index", 2, 3, "1");
        }

        #endregion


        #region Lookup function Two Reference arguments tests

        [TestMethod]
        public void LookupFunctionReferenceArgumentsRowVectorPointVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:G5{ls}D2)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Fu", worksheet);
        }

        [TestMethod]
        public void LookupFunctionReferenceArgumentsRowVectorColumnVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:G5{ls}D2:D3)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "b", worksheet);
        }

        [TestMethod]
        public void LookupFunctionReferenceArgumentsRowVectorRowVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:G5{ls}D2:E2)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Fu", worksheet);
        }

        [TestMethod]
        public void LookupFunctionReferenceArgumentsRowVectorMatrixResultTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:G5{ls}D2:E3)";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        [TestMethod]
        public void LookupFunctionReferenceArgumentsColumnVectorPointVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:B5{ls}D2)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Ts", worksheet);
        }

        [TestMethod]
        public void LookupFunctionReferenceArgumentsColumnVectorRowVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:B5{ls}D2:E2)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Ts", worksheet);
        }

        [TestMethod]
        public void LookupFunctionReferenceArgumentsColumnVectorColumnVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:B5{ls}D2:D3)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "d", worksheet);
        }

        [TestMethod]
        public void LookupFunctionReferenceArgumentsColumnVectorMatrixResultTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:B5{ls}D2:E3)";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        [TestMethod]
        public void LookupFunctionReferenceArgumentsFewColumnsMatrixPointVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:C5{ls}D2)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Ts", worksheet);
        }

        [TestMethod]
        public void LookupFunctionReferenceArgumentsFewColumnsMatrixRowVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:C5{ls}D2:E2)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Ts", worksheet);
        }

        [TestMethod]
        public void LookupFunctionReferenceArgumentsFewColumnsMatrixColumnVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:C5{ls}D2:D3)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "d", worksheet);
        }

        [TestMethod]
        public void LookupFunctionReferenceArgumentsMoreColumnsMatrixPointVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:F5{ls}D2)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Fu", worksheet);
        }

        [TestMethod]
        public void LookupFunctionReferenceArgumentsMoreColumnsMatrixRowVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:F5{ls}D2:E2)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Fu", worksheet);
        }

        [TestMethod]
        public void LookupFunctionReferenceArgumentsMoreColumnsMatrixColumnVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:F5{ls}D2:D3)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "b", worksheet);
        }

        [TestMethod]
        public void LookupFunctionReferenceArgumentsQuadraticMatrixPointVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:E5{ls}D2)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Ts", worksheet);
        }

        [TestMethod]
        public void LookupFunctionReferenceArgumentsSmallQuadraticMatrixPointVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:D4{ls}D2)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "s", worksheet);
        }

        [TestMethod]
        public void LookupFunctionReferenceArgumentsColumnVectorBigMatrixResultTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:B5{ls}D2:F5)";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        [TestMethod]
        public void LookupFunctionReferenceArgumentsRowVectorBigMatrixResultTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Lookup(2{ls}B2:F2{ls}D2:F5)";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        #endregion


        #region Lookup function Array lookup argument Reference second argument tests

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsRowVectorPointVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Fu", worksheet);
        }

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsRowVectorColumnVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2:D3)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "b", worksheet);
        }

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsRowVectorRowVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2:E2)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Fu", worksheet);
        }

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsRowVectorMatrixResultTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2:E3)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsColumnVectorPointVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 0));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Ts", worksheet);
        }

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsColumnVectorRowVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 0));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2:E2)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Ts", worksheet);
        }

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsColumnVectorColumnVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 0));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2:D3)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "d", worksheet);
        }

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsColumnVectorMatrixResultTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 0));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2:E3)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsFewColumnsMatrixPointVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 1));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Ts", worksheet);
        }

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsFewColumnsMatrixRowVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 1));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2:E2)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Ts", worksheet);
        }

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsFewColumnsMatrixColumnVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 1));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2:D3)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "d", worksheet);
        }

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsMoreColumnsMatrixPointVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 4));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Fu", worksheet);
        }

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsMoreColumnsMatrixRowVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 4));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2:E2)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Fu", worksheet);
        }

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsMoreColumnsMatrixColumnVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 4));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2:D3)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "b", worksheet);
        }

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsQuadraticMatrixPointVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 3));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "Ts", worksheet);
        }

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsSmallQuadraticMatrixPointVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(2, 2));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "s", worksheet);
        }

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsColumnVectorBigMatrixResultTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 0));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2:F5)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        [TestMethod]
        public void LookupFunctionArrayReferenceArgumentsRowVectorBigMatrixResultTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 4));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}D2:F5)", "{ls}", lookupArray);
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        #endregion


        #region Lookup function Two Array arguments tests

        [TestMethod]
        public void LookupFunctionArrayArgumentsRowVectorPointVectorTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 2), new CellIndex(0, 5));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertCellValue(a1CellInput, "Fu");
        }

        [TestMethod]
        public void LookupFunctionArrayArgumentsRowVectorColumnVectorTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 2), new CellIndex(3, 2));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertCellValue(a1CellInput, "b");
        }

        [TestMethod]
        public void LookupFunctionArrayArgumentsRowVectorRowVectorTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 2), new CellIndex(0, 5));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertCellValue(a1CellInput, "Fu");
        }

        [TestMethod]
        public void LookupFunctionArrayArgumentsRowVectorMatrixResultTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 2), new CellIndex(2, 4));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError);
        }

        [TestMethod]
        public void LookupFunctionNotFullArrayArgumentsRowVectorPointVectorTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 2), new CellIndex(0, 2));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError);
        }

        [TestMethod]
        public void LookupFunctionNotFullArrayArgumentsRowVectorColumnVectorTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 2), new CellIndex(1, 2));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertCellValue(a1CellInput, "b");
        }

        [TestMethod]
        public void LookupFunctionNotFullArrayArgumentsRowVectorRowVectorTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 2), new CellIndex(0, 3));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertCellValue(a1CellInput, "Fu");
        }

        [TestMethod]
        public void LookupFunctionNotFullArrayArgumentsRowVectorMatrixResultTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 2), new CellIndex(1, 3));
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError);
        }

        [TestMethod]
        public void LookupFunctionArrayNumberArgumentsRowVectorPointVectorTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = "314";
            string a1CellInput = String.Format("=Lookup(1{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertCellValue(a1CellInput, "314");
        }

        [TestMethod]
        public void LookupFunctionArrayNumberArgumentsRowVectorColumnVectorTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = "314";
            string a1CellInput = String.Format("=Lookup(1{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertCellValue(a1CellInput, "314");
        }

        [TestMethod]
        public void LookupFunctionArrayNumberArgumentsRowVectorRowVectorTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = "314";
            string a1CellInput = String.Format("=Lookup(1{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertCellValue(a1CellInput, "314");
        }

        [TestMethod]
        public void LookupFunctionArrayNumberArgumentsRowVectorMatrixResultTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = "314";
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError);
        }

        [TestMethod]
        public void LookupFunctionArrayStringArgumentsRowVectorPointVectorTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = "\"some text\"";
            string a1CellInput = String.Format("=Lookup(1{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ValueError);
        }

        [TestMethod]
        public void LookupFunctionArrayStringArgumentsRowVectorColumnVectorTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = "\"some text\"";
            string a1CellInput = String.Format("=Lookup(1{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ValueError);
        }

        [TestMethod]
        public void LookupFunctionArrayStringArgumentsRowVectorRowVectorTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = "\"some text\"";
            string a1CellInput = String.Format("=Lookup(1{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ValueError);
        }

        [TestMethod]
        public void LookupFunctionArrayStringArgumentsRowVectorMatrixResultTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = "\"some text\"";
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ValueError);
        }

        [TestMethod]
        public void LookupFunctionArrayBooleanArgumentsRowVectorPointVectorTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = "TRUE";
            string a1CellInput = String.Format("=Lookup(1{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ValueError);
        }

        [TestMethod]
        public void LookupFunctionArrayBooleanArgumentsRowVectorColumnVectorTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = "TRUE";
            string a1CellInput = String.Format("=Lookup(1{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ValueError);
        }

        [TestMethod]
        public void LookupFunctionArrayBooleanArgumentsRowVectorRowVectorTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = "FALSE";
            string a1CellInput = String.Format("=Lookup(1{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ValueError);
        }

        [TestMethod]
        public void LookupFunctionArrayBooleanArgumentsRowVectorMatrixResultTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = "FALSE";
            string a1CellInput = String.Format("=Lookup(2{0}{1}{0}{2})", "{ls}", lookupArray, resultArray);

            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ValueError);
        }

        #endregion


        #region Lookup function different criteria tests

        [TestMethod]
        public void LookupFunctionDirectArrayCriteriaTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = "314";
            string criteria = "{1{als}2{als}3}";
            string a1CellInput = String.Format("=Lookup({3}{0}{1}{0}{2})", "{ls}", lookupArray, resultArray, criteria);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "314");
        }

        [TestMethod]
        public void LookupNotPointReferenceCriteriaTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = "314";
            string criteria = "B2:F2";
            string a1CellInput = String.Format("=Lookup({3}{0}{1}{0}{2})", "{ls}", lookupArray, resultArray, criteria);
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ValueError, worksheet);
        }

        [TestMethod]
        public void LookupPointReferenceCriteriaTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = "314";
            string criteria = "B2";
            string a1CellInput = String.Format("=Lookup({3}{0}{1}{0}{2})", "{ls}", lookupArray, resultArray, criteria);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "314", worksheet);
        }

        [TestMethod]
        public void LookupBooleanCriteriaTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string resultArray = "314";
            string criteria = "TRUE";
            string a1CellInput = String.Format("=Lookup({3}{0}{1}{0}{2})", "{ls}", lookupArray, resultArray, criteria);
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        #endregion


        #region Lookup function only required arguments with Array lookup tests

        [TestMethod]
        public void LookupFunctionOnlyRequiredArgumentsArrayLookupRowVectorTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 4));
            string a1CellInput = String.Format("=Lookup(2{0}{1})", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "2");
        }

        [TestMethod]
        public void LookupFunctionOnlyRequiredArgumentsArrayLookupColumnVectorTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 0));
            string a1CellInput = String.Format("=Lookup(2{0}{1})", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "2");
        }

        [TestMethod]
        public void LookupFunctionOnlyRequiredArgumentsArrayLookupMoreRowsThanColumnsMatrixTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 1));
            string a1CellInput = String.Format("=Lookup(2{0}{1})", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "");
        }

        [TestMethod]
        public void LookupFunctionOnlyRequiredArgumentsArrayLookupMoreRowsThanColumnsLongerMatrixTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 2));
            string a1CellInput = String.Format("=Lookup(2{0}{1})", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "d");
        }

        [TestMethod]
        public void LookupFunctionOnlyRequiredArgumentsArrayLookupMoreColumnsThanRowsMatrixTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 4));
            string a1CellInput = String.Format("=Lookup(2{0}{1})", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "");
        }

        [TestMethod]
        public void LookupFunctionOnlyRequiredArgumentsArrayLookupQuadraticMatrixTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 3));
            string a1CellInput = String.Format("=Lookup(2{0}{1})", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "");
        }

        [TestMethod]
        public void LookupFunctionOnlyRequiredArgumentsArrayLookupQuadraticSmallerMatrixTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(2, 2));
            string a1CellInput = String.Format("=Lookup(2{0}{1})", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "c");
        }

        [TestMethod]
        public void LookupFunctionOnlyRequiredArgumentsArrayLookupQuadraticSmallerMatrixEarlierStopTest()
        {
            string[,] earlierStopScenario = new string[,]
            {
                { "0", "1", "2" },
                { "2", null, "b" },
                { "3", null, "c" },
            };

            string lookupArray = GetSubArrayStringFromStringArray(earlierStopScenario,
                new CellIndex(0, 0), new CellIndex(2, 2));
            string a1CellInput = String.Format("=Lookup(2{0}{1})", "{ls}", lookupArray);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "b");
        }

        #endregion


        #region Lookup function only required arguments with Reference lookup tests

        [TestMethod]
        public void LookupFunctionOnlyRequiredArgumentsReferenceLookupRowVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = String.Format("=Lookup(2{0}B2:F2)", "{ls}");
            FunctionsTestHelper.AssertCellValue(a1CellInput, "2", worksheet);
        }

        [TestMethod]
        public void LookupFunctionOnlyRequiredArgumentsReferenceLookupColumnVectorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = String.Format("=Lookup(2{0}B2:B5)", "{ls}");
            FunctionsTestHelper.AssertCellValue(a1CellInput, "2", worksheet);
        }

        [TestMethod]
        public void LookupFunctionOnlyRequiredArgumentsReferenceLookupMoreRowsThanColumnsMatrixTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = String.Format("=Lookup(2{0}B2:C5)", "{ls}");
            FunctionsTestHelper.AssertCellValue(a1CellInput, "0", worksheet);
        }

        [TestMethod]
        public void LookupFunctionOnlyRequiredArgumentsReferenceLookupMoreRowsThanColumnsLongerMatrixTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = String.Format("=Lookup(2{0}B2:D5)", "{ls}");
            FunctionsTestHelper.AssertCellValue(a1CellInput, "d", worksheet);
        }

        [TestMethod]
        public void LookupFunctionOnlyRequiredArgumentsReferenceLookupMoreColumnsThanRowsMatrixTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = String.Format("=Lookup(2{0}B2:F5)", "{ls}");
            FunctionsTestHelper.AssertCellValue(a1CellInput, "0", worksheet);
        }

        [TestMethod]
        public void LookupFunctionOnlyRequiredArgumentsReferenceLookupQuadraticMatrixTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = String.Format("=Lookup(2{0}B2:E5)", "{ls}");
            FunctionsTestHelper.AssertCellValue(a1CellInput, "0", worksheet);
        }

        [TestMethod]
        public void LookupFunctionOnlyRequiredArgumentsReferenceLookupQuadraticSmallerMatrixTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = String.Format("=Lookup(2{0}B2:D4)", "{ls}");
            FunctionsTestHelper.AssertCellValue(a1CellInput, "c", worksheet);
        }

        [TestMethod]
        public void LookupFunctionOnlyRequiredArgumentsReferenceLookupToOtherSheetTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet secondWorksheet = worksheet.Workbook.Worksheets.Add();
            secondWorksheet.CopyFrom(worksheet);

            CellRange usedCellRange = worksheet.UsedCellRange;
            for(int row = 0; row < usedCellRange.RowCount; row++)
            {
                for(int col = 0; col < usedCellRange.ColumnCount; col++)
                {
                    worksheet.Cells[row, col].Clear(ClearType.All);
                }
            }

            string a1CellInput = String.Format("=Lookup(2{0}Sheet2!B2:D4)", "{ls}");
            FunctionsTestHelper.AssertCellValue(a1CellInput, "c", worksheet);
        }

        [TestMethod]
        public void LookupFunctionOnlyRequiredArgumentsReferenceLookupQuadraticSmallerMatrixEarlierStopTest()
        {
            string[,] earlierStopScenario = new string[,]
            {
                { "0", "1", "2" },
                { "2", null, "b" },
                { "3", null, "c" },
            };

            Worksheet worksheet = CreateWorksheetScenario(earlierStopScenario);

            string a1CellInput = String.Format("=Lookup(2{0}B2:D4)", "{ls}");
            FunctionsTestHelper.AssertCellValue(a1CellInput, "b", worksheet);
        }

        #endregion


        #region Lookup function complex criteria tests

        [TestMethod]
        public void LookupFunctionPointReferenceCriteriaTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = String.Format("=Lookup(C2{0}B2:G2)", "{ls}");
            FunctionsTestHelper.AssertCellValue(a1CellInput, "2", worksheet);
        }

        [TestMethod]
        public void LookupFunctionRowReferenceCriteriaTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = String.Format("=Lookup(C2:D2{0}B2:G2)", "{ls}");
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ValueError, worksheet);
        }

        [TestMethod]
        public void LookupFunctionColumnReferenceCriteriaTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = String.Format("=Lookup(C2:C3{0}B2:G2)", "{ls}");
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ValueError, worksheet);
        }

        [TestMethod]
        public void LookupFunctionMatrixReferenceCriteriaTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = String.Format("=Lookup(C2:D3{0}B2:G2)", "{ls}");
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ValueError, worksheet);
        }

        [TestMethod]
        public void LookupFunctionMatrixReferenceCriteriaNotPoppingUpErrorsTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            worksheet.Cells[2, 2].SetValue(ErrorExpressions.NumberError.Value);
            worksheet.Cells[2, 3].SetValue(ErrorExpressions.DivisionByZero.Value);
            worksheet.Cells[1, 3].SetValue(ErrorExpressions.NotAvailableError.Value);

            string a1CellInput = String.Format("=Lookup(C2:D3{0}B2:G2)", "{ls}");
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ValueError, worksheet);
        }

        [TestMethod]
        public void LookupFunctionPointReferenceCriteriaPoppingUpErrorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            worksheet.Cells[1, 2].SetValue(ErrorExpressions.NumberError.Value);

            string a1CellInput = String.Format("=Lookup(C2{0}B2:G2)", "{ls}");
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NumberError, worksheet);
        }

        [TestMethod]
        public void LookupFunctionMatrixArrayCriteriaNotPoppingUpErrorsTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            string[,] arrayWithErrors = new string[,]
            {
                { "3", ErrorExpressions.NumberError.Value},
                {ErrorExpressions.NotAvailableError.Value, ErrorExpressions.DivisionByZero.Value}
            };

            string arrayExpression = GetSubArrayStringFromStringArray(arrayWithErrors, new CellIndex(0, 0), new CellIndex(1, 1));

            string a1CellInput = String.Format("=Lookup({0}{1}B2:G2)", arrayExpression, "{ls}");
            FunctionsTestHelper.AssertCellValue(a1CellInput, "2", worksheet);
        }

        [TestMethod]
        public void LookupFunctionMatrixArrayCriteriaPoppingUpFirstArrayElementErrorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            string[,] arrayWithErrors = new string[,]
            {
                { ErrorExpressions.NumberError.Value, ErrorExpressions.NullError.Value },
                { ErrorExpressions.NotAvailableError.Value, ErrorExpressions.DivisionByZero.Value }
            };

            string arrayExpression = GetSubArrayStringFromStringArray(arrayWithErrors, new CellIndex(0, 0), new CellIndex(1, 1));

            string a1CellInput = String.Format("=Lookup({0}{1}B2:G2)", arrayExpression, "{ls}");
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NumberError, worksheet);
        }

        #endregion


        #region Lookup function error tests

        [TestMethod]
        public void LookupFunctionReferenceErrorVariationsTests()
        {
            FunctionsTestHelper.AssertFunctionReferenceErrorVariationsTests("Lookup", 2, 3, "1");
        }

        #endregion


        #region VLookup and HLookup function reference table argument tests

        [TestMethod]
        public void VHLookupFunctionReferenceTableArgumentMatrixTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G5";
            string hLookupRange = "B2:E7";

            string a1CellInput = "={fn}(2{ls}{range}{ls}3)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "d", vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionReferenceTableArgumentRowColumnTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G2";
            string hLookupRange = "B2:B7";

            string a1CellInput = "={fn}(0{ls}{range}{ls}3)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "a", vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionReferenceTableArgumentColumnRowTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:B5";
            string hLookupRange = "B2:E2";

            string a1CellInput = "={fn}(2{ls}{range}{ls}1)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "2", vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionReferenceTableArgumentFewerColumnsRowsTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:C5";
            string hLookupRange = "B2:E3";

            string a1CellInput = "={fn}(1{ls}{range}{ls}2)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "0", vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionReferenceTableArgumentMoreColumnsRowsTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G5";
            string hLookupRange = "B2:E7";

            string a1CellInput = "={fn}(1{ls}{range}{ls}3)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "b", vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionReferenceTableArgumentQuadraticMatrixTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:E5";
            string hLookupRange = "B2:E5";

            string a1CellInput = "={fn}(2{ls}{range}{ls}4)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "0", vLookupworksheet, hLookupworksheet);
        }

        #endregion


        #region VLookup and HLookup functions array table argument tests

        [TestMethod]
        public void VHLookupFunctionArrayTableArgumentMatrixTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupTableArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 5));
            string hLookupTableArray = GetSubArrayStringFromStringArray(invertedSampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(5, 3));
            string a1CellInput = String.Format("={2}(2{0}{1}{0}3)", "{ls}", "{range}", "{fn}");

            AssertBothVAndHLookup(a1CellInput, vLookupTableArray, hLookupTableArray, "d", vLookupWorksheet, hLookupWorksheet);
        }

        [TestMethod]
        public void VHLookupFunctionArrayTableArgumentRowColumnTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupTableArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 5));
            string hLookupTableArray = GetSubArrayStringFromStringArray(invertedSampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(5, 0));

            string a1CellInput = String.Format("={2}(0{0}{1}{0}3)", "{ls}", "{range}", "{fn}");

            AssertBothVAndHLookup(a1CellInput, vLookupTableArray, hLookupTableArray, "a", vLookupWorksheet, hLookupWorksheet);
        }

        [TestMethod]
        public void VHLookupFunctionArrayTableArgumentColumnRowTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupTableArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 0));
            string hLookupTableArray = GetSubArrayStringFromStringArray(invertedSampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(0, 3));

            string a1CellInput = String.Format("={2}(2{0}{1}{0}1)", "{ls}", "{range}", "{fn}");

            AssertBothVAndHLookup(a1CellInput, vLookupTableArray, hLookupTableArray, "2", vLookupWorksheet, hLookupWorksheet);
        }

        [TestMethod]
        public void VHLookupFunctionArrayTableArgumentFewerColumnsRowsTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupTableArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 1));
            string hLookupTableArray = GetSubArrayStringFromStringArray(invertedSampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(1, 3));

            string a1CellInput = String.Format("={2}(1{0}{1}{0}2)", "{ls}", "{range}", "{fn}");

            AssertBothVAndHLookup(a1CellInput, vLookupTableArray, hLookupTableArray, String.Empty, vLookupWorksheet, hLookupWorksheet);
        }

        [TestMethod]
        public void VHLookupFunctionArrayTableArgumentMoreColumnsRowsTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupTableArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 5));
            string hLookupTableArray = GetSubArrayStringFromStringArray(invertedSampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(5, 3));

            string a1CellInput = String.Format("={2}(1{0}{1}{0}3)", "{ls}", "{range}", "{fn}");

            AssertBothVAndHLookup(a1CellInput, vLookupTableArray, hLookupTableArray, "b", vLookupWorksheet, hLookupWorksheet);
        }

        [TestMethod]
        public void VHLookupFunctionArrayTableArgumentQuadraticMatrixTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupTableArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 3));
            string hLookupTableArray = GetSubArrayStringFromStringArray(invertedSampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 3));

            string a1CellInput = String.Format("={2}(2{0}{1}{0}4)", "{ls}", "{range}", "{fn}");

            AssertBothVAndHLookup(a1CellInput, vLookupTableArray, hLookupTableArray, String.Empty, vLookupWorksheet, hLookupWorksheet);
        }

        #endregion


        #region VLookup and HLookup functions criteria tests

        [TestMethod]
        public void VHLookupFunctionDirectArrayCriteriaInArrayTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 5));
            string hLookupArray = GetSubArrayStringFromStringArray(invertedSampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(5, 3));

            string criteria = "{1{als}2{als}3}";
            string indexNumber = "3";
            string a1CellInput = String.Format("={4}({3}{0}{1}{0}{2})", "{ls}", "{range}", indexNumber, criteria, "{fn}");

            AssertBothVAndHLookup(a1CellInput, vLookupArray, hLookupArray, "b", vLookupWorksheet, hLookupWorksheet);
        }

        [TestMethod]
        public void VHLookupNotPointReferenceCriteriaInArrayTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 5));
            string hLookupArray = GetSubArrayStringFromStringArray(invertedSampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(5, 3));

            string criteria = "B2:F2";
            string indexNumber = "3";
            string a1CellInput = String.Format("={4}({3}{0}{1}{0}{2})", "{ls}", "{range}", indexNumber, criteria, "{fn}");

            AssertErrorForBothVAndHLookup(a1CellInput, vLookupArray, hLookupArray, ErrorExpressions.ValueError, vLookupWorksheet, hLookupWorksheet);
        }

        [TestMethod]
        public void VHLookupPointReferenceCriteriaInArrayTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 5));
            string hLookupArray = GetSubArrayStringFromStringArray(invertedSampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(5, 3));

            string criteria = "B2";
            string indexNumber = "3";
            string a1CellInput = String.Format("={4}({3}{0}{1}{0}{2})", "{ls}", "{range}", indexNumber, criteria, "{fn}");

            AssertBothVAndHLookup(a1CellInput, vLookupArray, hLookupArray, "a", vLookupWorksheet, hLookupWorksheet);
        }

        [TestMethod]
        public void VHLookupBooleanCriteriaInArrayTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(3, 5));
            string hLookupArray = GetSubArrayStringFromStringArray(invertedSampleLookupFunctionCellInput,
                new CellIndex(0, 0), new CellIndex(5, 3));

            string criteria = "TRUE";
            string indexNumber = "3";
            string a1CellInput = String.Format("={4}({3}{0}{1}{0}{2})", "{ls}", "{range}", indexNumber, criteria, "{fn}");

            AssertErrorForBothVAndHLookup(a1CellInput, vLookupArray, hLookupArray, ErrorExpressions.NotAvailableError, vLookupWorksheet, hLookupWorksheet);
        }

        [TestMethod]
        public void VHLookupFunctionPointReferenceCriteriaTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G2";
            string hLookupRange = "B2:B7";

            string a1CellInput = "={fn}(B2{ls}{range}{ls}3)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "a", vLookupWorksheet, hLookupWorksheet);
        }

        [TestMethod]
        public void VHLookupFunctionRowReferenceCriteriaTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G2";
            string hLookupRange = "B2:B7";

            string a1CellInput = "={fn}(C2:D2{ls}{range}{ls}3)";
            AssertErrorForBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, ErrorExpressions.ValueError, vLookupWorksheet, hLookupWorksheet);
        }

        [TestMethod]
        public void VHLookupFunctionColumnReferenceCriteriaTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G2";
            string hLookupRange = "B2:B7";

            string a1CellInput = "={fn}(C2:C3{ls}{range}{ls}3)";
            AssertErrorForBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, ErrorExpressions.ValueError, vLookupWorksheet, hLookupWorksheet);
        }

        [TestMethod]
        public void VHLookupFunctionMatrixReferenceCriteriaTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G2";
            string hLookupRange = "B2:B7";

            string a1CellInput = "={fn}(C2:D3{ls}{range}{ls}3)";
            AssertErrorForBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, ErrorExpressions.ValueError, vLookupWorksheet, hLookupWorksheet);
        }

        [TestMethod]
        public void VHLookupFunctionMatrixReferenceCriteriaErrorsTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            vLookupWorksheet.Cells[2, 2].SetValue(ErrorExpressions.NumberError.Value);
            vLookupWorksheet.Cells[2, 3].SetValue(ErrorExpressions.DivisionByZero.Value);
            vLookupWorksheet.Cells[1, 3].SetValue(ErrorExpressions.NotAvailableError.Value);

            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);
            vLookupWorksheet.Cells[2, 2].SetValue(ErrorExpressions.NumberError.Value);
            vLookupWorksheet.Cells[2, 3].SetValue(ErrorExpressions.DivisionByZero.Value);
            vLookupWorksheet.Cells[1, 3].SetValue(ErrorExpressions.NotAvailableError.Value);

            string vLookupRange = "B2:G2";
            string hLookupRange = "B2:B7";

            string a1CellInput = "={fn}(C2:D3{ls}{range}{ls})";
            AssertErrorForBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, ErrorExpressions.ValueError, vLookupWorksheet, hLookupWorksheet);
        }

        [TestMethod]
        public void VHLookupFunctionMatrixArrayCriteriaErrorsTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string[,] arrayWithErrors = new string[,]
            {
                { "3", ErrorExpressions.NumberError.Value},
                {ErrorExpressions.NotAvailableError.Value, ErrorExpressions.DivisionByZero.Value}
            };

            string vLookupRange = "B2:G2";
            string hLookupRange = "B2:B7";

            string arrayExpression = GetSubArrayStringFromStringArray(arrayWithErrors, new CellIndex(0, 0), new CellIndex(1, 1));

            string a1CellInput = String.Format("={2}({0}{1}{3}{1}3)", arrayExpression, "{ls}", "{fn}", "{range}");
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "a", vLookupWorksheet, hLookupWorksheet);
        }

        [TestMethod]
        public void VHLookupFunctionReferenceToAnotherSheetCriteriaTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet vLookupSecondWorksheet = vLookupWorksheet.Workbook.Worksheets.Add();
            vLookupSecondWorksheet.Cells[1, 1].SetValue(sampleLookupFunctionCellInput[0, 0]);

            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);
            Worksheet hLookupSecondWorksheet = hLookupWorksheet.Workbook.Worksheets.Add();
            hLookupSecondWorksheet.Cells[1, 1].SetValue(invertedSampleLookupFunctionCellInput[0, 0]);

            string vLookupRange = "B2:G2";
            string hLookupRange = "B2:B7";

            string a1CellInput = "={fn}(Sheet2!B2{ls}{range}{ls}3)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "a", vLookupWorksheet, hLookupWorksheet);
        }

        [TestMethod]
        public void VHLookupFunctionReferenceToEmptyCellValueCriteriaTest()
        {
            Worksheet vLookupWorksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupWorksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G2";
            string hLookupRange = "B2:B7";

            string a1CellInput = "={fn}(C3{ls}{range}{ls}3)";
            AssertErrorForBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, ErrorExpressions.NotAvailableError, vLookupWorksheet, hLookupWorksheet);
        }

        #endregion


        #region VLookup and HLookup functions range lookup tests

        [TestMethod]
        public void VHLookupFunctionRangeLookupArgumentExactMatchCriteriaFoundTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G5";
            string hLookupRange = "B2:E7";

            string a1CellInput = "={fn}(2{ls}{range}{ls}3{ls}FALSE)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "c", vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionRangeLookupArgumentExactMatchCriteriaNotFoundTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G5";
            string hLookupRange = "B2:E7";

            string a1CellInput = "={fn}(3{ls}{range}{ls}3{ls}FALSE)";
            AssertErrorForBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, ErrorExpressions.NotAvailableError, vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionRangeLookupArgumentApproximateMatchSearchOrderTest()
        {
            // Note the difference between this test and VLookupFunctionRangeLookupArgumentExactMatchCriteriaFoundTest is only in the value of the range lookup argument value.
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G5";
            string hLookupRange = "B2:E7";

            string a1CellInput = "={fn}(2{ls}{range}{ls}3{ls}TRUE)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "d", vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionRangeLookupArgumentApproximateMatchNotAscendingSortedValuesTest()
        {
            string[,] notSortedArray = new string[,]
            {
                {"2", "2", "a", "Fu", "s", "Ts"},
                {"1", null, "b", null, null, null},
                {"0", null, "c", null, null, null},
                {"2", null, "d", null, null, null},
            };

            string[,] invertedNotSortedArray = new string[,]
            {
                {"2", "1", "0", "2"},
                {"2", null, null, null},
                {"a", "b", "c", "d"},
                {"Fu", null, null, null},
                {"s", null, null, null},
                {"Ts", null, null, null},
            };

            Worksheet vLookupworksheet = CreateWorksheetScenario(notSortedArray);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedNotSortedArray);

            string vLookupRange = "B2:G5";
            string hLookupRange = "B2:E7";

            string a1CellInput = "={fn}(1.5{ls}{range}{ls}3{ls}TRUE)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "c", vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionRangeLookupArgumentApproximateMatchCriteriaSmallerThanFirstTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G5";
            string hLookupRange = "B2:E7";

            string a1CellInput = "={fn}(-1{ls}{range}{ls}3{ls}TRUE)";
            AssertErrorForBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, ErrorExpressions.NotAvailableError, vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionRangeLookupArgumentApproximateMatchCriteriaLargerThanLastTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G5";
            string hLookupRange = "B2:E7";

            string a1CellInput = "={fn}(3{ls}{range}{ls}3{ls}TRUE)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "d", vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionRangeLookupArgumentPointReferenceTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string b1CellInput = "false";
            vLookupworksheet.Cells[0, 1].SetValue(b1CellInput);
            hLookupworksheet.Cells[0, 1].SetValue(b1CellInput);

            string vLookupRange = "B2:G5";
            string hLookupRange = "B2:E7";

            string a1CellInput = "={fn}(1{ls}{range}{ls}3{ls}B1)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "b", vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionRangeLookupArgumentRefersToOtherWorksheetTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet vLookupSecondWorksheet = vLookupworksheet.Workbook.Worksheets.Add();

            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);
            Worksheet hLookupSecondWorksheet = hLookupworksheet.Workbook.Worksheets.Add();

            string vLookupRange = "Sheet1!B2:G5";
            string hLookupRange = "Sheet1!B2:E7";

            string a1CellInput = "={fn}(2{ls}{range}{ls}3{ls}FALSE)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "c", vLookupSecondWorksheet, hLookupSecondWorksheet);
        }

        #endregion


        #region VLookup and HLookup functions column index num tests

        [TestMethod]
        public void VHLookupFunctionIndexArgumentInRangeTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G5";
            string hLookupRange = "B2:E7";

            string a1CellInput = "={fn}(2{ls}{range}{ls}2)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "0", vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionIndexArgumentSmallerThanRangeTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G5";
            string hLookupRange = "B2:E7";

            string a1CellInput = "={fn}(2{ls}{range}{ls}-1)";
            AssertErrorForBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, ErrorExpressions.ValueError, vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionIndexArgumentLargerThanRangeTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G5";
            string hLookupRange = "B2:E7";

            string a1CellInput = "={fn}(2{ls}{range}{ls}7)";
            AssertErrorForBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, ErrorExpressions.ReferenceError, vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionIndexArgumentFractionTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G5";
            string hLookupRange = "B2:E7";

            string a1CellInput = "={fn}(2{ls}{range}{ls}3.5)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "d", vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionIndexArgumentPointReferenceTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string b1CellInput = "3";
            vLookupworksheet.Cells[0, 1].SetValue(b1CellInput);
            hLookupworksheet.Cells[0, 1].SetValue(b1CellInput);

            string vLookupRange = "B2:G5";
            string hLookupRange = "B2:E7";

            string a1CellInput = "={fn}(2{ls}{range}{ls}B1)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "d", vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionIndexArgumentStringTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G5";
            string hLookupRange = "B2:E7";

            string a1CellInput = "={fn}(2{ls}{range}{ls}\"test\")";
            AssertErrorForBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, ErrorExpressions.ValueError, vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionIndexArgumentBooleanFalseTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G5";
            string hLookupRange = "B2:E7";

            string a1CellInput = "={fn}(2{ls}{range}{ls}FALSE)";
            AssertErrorForBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, ErrorExpressions.ValueError, vLookupworksheet, hLookupworksheet);
        }

        [TestMethod]
        public void VHLookupFunctionIndexArgumentBooleanTrueTest()
        {
            Worksheet vLookupworksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet hLookupworksheet = CreateWorksheetScenario(invertedSampleLookupFunctionCellInput);

            string vLookupRange = "B2:G5";
            string hLookupRange = "B2:E7";

            string a1CellInput = "={fn}(2{ls}{range}{ls}TRUE)";
            AssertBothVAndHLookup(a1CellInput, vLookupRange, hLookupRange, "2", vLookupworksheet, hLookupworksheet);
        }

        #endregion


        #region VLookup and HLookup functions error tests

        [TestMethod]
        public void VLookupFunctionReferenceErrorVariationsTests()
        {
            FunctionsTestHelper.AssertFunctionReferenceErrorVariationsTests("VLookup", 3, 4, "1");
        }

        [TestMethod]
        public void HLookupFunctionReferenceErrorVariationsTests()
        {
            FunctionsTestHelper.AssertFunctionReferenceErrorVariationsTests("HLookup", 3, 4, "1");
        }

        #endregion


        #region Match function lookup value tests

        [TestMethod]
        public void MatchFunctionLookUpValueNumberTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(2{ls}B2:B5)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "4", worksheet);
        }

        [TestMethod]
        public void MatchFunctionLookUpValueTextTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(\"c\"{ls}D2:D5)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "3", worksheet);
        }

        [TestMethod]
        public void MatchFunctionLookUpEmptyValueTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match({ls}C2:C5)";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        [TestMethod]
        public void MatchFunctionLookUpValuePointReferenceTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            worksheet.Cells[0, 1].SetValue(2);

            string a1CellInput = "=Match(B1{ls}B2:B5)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "4", worksheet);
        }

        [TestMethod]
        public void MatchFunctionLookUpValueRangeReferenceTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            worksheet.Cells[0, 1].SetValue(2);
            worksheet.Cells[0, 2].SetValue(2);

            string a1CellInput = "=Match(B1:C1{ls}B2:B5)";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ValueError, worksheet);
        }

        [TestMethod]
        public void MatchFunctionLookUpValueDirectArrayPointTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupValue = "{2}";

            string a1CellInput = String.Format("=Match({1}{0}B2:B5)", "{ls}", lookupValue);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "4", worksheet);
        }

        [TestMethod]
        public void MatchFunctionLookUpValueDirectArrayMatrixTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string lookupValue = GetSubArrayStringFromStringArray(new string[,] { {"2", "2"}, {"2", "3"} }, new CellIndex(0, 0), new CellIndex(1, 1));

            string a1CellInput = String.Format("=Match({1}{0}B2:B5)", "{ls}", lookupValue);
            FunctionsTestHelper.AssertCellValue(a1CellInput, "4", worksheet);
        }

        [TestMethod]
        public void MatchFunctionLookUpValueReferenceToErrorTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            worksheet.Cells[0,1].SetValue(ErrorExpressions.NumberError.Value);

            string a1CellInput = "=Match(B1{ls}B2:B5)";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NumberError, worksheet);
        }

        [TestMethod]
        public void MatchFunctionLookUpValueReferenceToDifferentSheetTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet secondWorksheet = worksheet.Workbook.Worksheets.Add();
            secondWorksheet.Cells[0, 0].SetValue(2);

            string a1CellInput = "=Match(Sheet2!A1{ls}B2:B5)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "4", worksheet);
        }

        #endregion


        #region Match function lookup array tests

        [TestMethod]
        public void MatchFunctionLookUpArrayColumnTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(1{ls}B2:B5)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "2", worksheet);
        }

        [TestMethod]
        public void MatchFunctionLookUpArrayRowTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(\"s\"{ls}B2:G2)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "5", worksheet);
        }

        [TestMethod]
        public void MatchFunctionLookUpArrayMatrixTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(2{ls}B2:G5)";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        [TestMethod]
        public void MatchFunctionLookUpArrayDirectArrayFromMatrixTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput, new CellIndex(0, 0), new CellIndex(3, 5));

            string a1CellInput = String.Format("=Match(1{1}{0})", lookupArray, "{ls}");
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError);
        }

        [TestMethod]
        public void MatchFunctionLookUpArrayDirectArrayFromColumnTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput, new CellIndex(0, 0), new CellIndex(3, 0));

            string a1CellInput = String.Format("=Match(1{1}{0})", lookupArray, "{ls}");
            FunctionsTestHelper.AssertCellValue(a1CellInput, "2");
        }

        [TestMethod]
        public void MatchFunctionLookUpArrayDirectArrayFromRowTest()
        {
            string lookupArray = GetSubArrayStringFromStringArray(sampleLookupFunctionCellInput, new CellIndex(0, 0), new CellIndex(0, 5));

            string a1CellInput = String.Format("=Match(1{1}{0})", lookupArray, "{ls}");
            FunctionsTestHelper.AssertCellValue(a1CellInput, "1");
        }

        [TestMethod]
        public void MatchFunctionLookUpArrayOnDifferentSheetTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);
            Worksheet secondWorksheet = worksheet.Workbook.Worksheets.Add();

            string a1CellInput = "=Match(1{ls}Sheet1!B2:B5)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "2", secondWorksheet);
        }

        [TestMethod]
        public void MatchFunctionLookUpArrayNumberValueTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(1{ls}1)";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        [TestMethod]
        public void MatchFunctionLookUpArrayTextValueTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(1{ls}\"test\")";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ValueError, worksheet);
        }

        #endregion


        #region Match function match type tests

        [TestMethod]
        public void MatchFunctionMatchType1AscendingNumberDataExactMatchPresentTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(2{ls}B2:B5{ls}1)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "4", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchType1AscendingNumberDataExactMatchPresentTest2()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(1{ls}B2:B5{ls}1)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "2", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchType1AscendingNumberDataExactMatchNotPresentTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(2.5{ls}B2:B5{ls}1)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "4", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchType1AscendingNumberDataExactMatchNotPresentTest2()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(1.5{ls}B2:B5{ls}1)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "2", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchType1AscendingNumberDataExactMatchNotPresentTest3()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(-1{ls}B2:B5{ls}1)";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchType1AscendingTextDataExactMatchPresentTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(\"b\"{ls}D2:D5{ls}1)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "2", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchType1AscendingTextDataExactMatchNotPresentTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(\"ba\"{ls}D2:D5{ls}1)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "2", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchType1AscendingTextDataExactMatchNotPresentTest2()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(\"da\"{ls}D2:D5{ls}1)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "4", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchType1LookingForEmptyTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(B1{ls}C2:C5{ls}1)";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchTypeMinus1DescendingNumberDataExactMatchPresentTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleDescendingLookupFunctionCellInput);

            string a1CellInput = "=Match(2{ls}B2:B5{ls}-1)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "1", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchTypeMinus1DescendingNumberDataExactMatchPresentTest2()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleDescendingLookupFunctionCellInput);

            string a1CellInput = "=Match(1{ls}B2:B5{ls}-1)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "3", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchTypeMinus1DescendingNumberDataExactMatchNotPresentTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleDescendingLookupFunctionCellInput);

            string a1CellInput = "=Match(1.5{ls}B2:B5{ls}-1)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "2", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchTypeMinus1DescendingNumberDataExactMatchNotPresentTest2()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleDescendingLookupFunctionCellInput);

            string a1CellInput = "=Match(2.5{ls}B2:B5{ls}-1)";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchTypeMinus1DescendingNumberDataExactMatchNotPresentTest3()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleDescendingLookupFunctionCellInput);

            string a1CellInput = "=Match(-1{ls}B2:B5{ls}-1)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "4", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchTypeMinus1DescendingTextDataExactMatchPresentTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleDescendingLookupFunctionCellInput);

            string a1CellInput = "=Match(\"c\"{ls}D2:D5{ls}-1)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "2", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchTypeMinus1DescendingTextDataExactMatchNotPresentTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleDescendingLookupFunctionCellInput);

            string a1CellInput = "=Match(\"ca\"{ls}D2:D5{ls}-1)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "1", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchTypeMinus1DescendingTextDataExactMatchNotPresentTest2()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleDescendingLookupFunctionCellInput);

            string a1CellInput = "=Match(\"f\"{ls}D2:D5{ls}-1)";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchType0NumberDataExactMatchPresentTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(1{ls}B2:B5{ls}0)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "2", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchType0NumberDataExactMatchNotPresentTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(1.5{ls}B2:B5{ls}0)";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchType0TextDataExactMatchPresentTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(\"d\"{ls}D2:D5{ls}0)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "4", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchType0TextDataExactMatchNotPresentTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(\"ca\"{ls}B2:B5{ls}0)";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchType0LookingForEmptyTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(B1{ls}C2:C5{ls}0)";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchType100AscendingNumberDataExactMatchPresentTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(2{ls}B2:B5{ls}100)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "4", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchTypeMinus100DescendingNumberDataExactMatchPresentTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleDescendingLookupFunctionCellInput);

            string a1CellInput = "=Match(2{ls}B2:B5{ls}-100)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "1", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchTypeFractionAscendingNumberDataExactMatchPresentTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(2{ls}B2:B5{ls}3.5)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "4", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchTypeTextTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(2{ls}B2:B5{ls}\"text\")";
            FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.ValueError, worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchTypeBoolTrueTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(2{ls}B2:B5{ls}TRUE)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "4", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchTypeBoolFalseTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(2{ls}B2:B5{ls}FALSE)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "3", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchTypeEmptyTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

            string a1CellInput = "=Match(2{ls}B2:B5{ls}B1)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "3", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchTypeReferenceToPointTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleDescendingLookupFunctionCellInput);
            worksheet.Cells[0, 1].SetValue(-1);

            string a1CellInput = "=Match(2{ls}B2:B5{ls}B1)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "1", worksheet);
        }

        [TestMethod]
        public void MatchFunctionMatchTypeReferenceToDifferentSheetTest()
        {
            Worksheet worksheet = CreateWorksheetScenario(sampleDescendingLookupFunctionCellInput);
            Worksheet secondWorksheet = worksheet.Workbook.Worksheets.Add();
            secondWorksheet.Cells[0, 1].SetValue(-1);

            string a1CellInput = "=Match(2{ls}B2:B5{ls}Sheet2!B1)";
            FunctionsTestHelper.AssertCellValue(a1CellInput, "1", worksheet);
        }

        #endregion

        #region Invalid scenarios tests

        // See bug 226334 in TeamPulse Match function returns results different from the ones in Excel in invalid scenarios
        //[TestMethod]
        //public void MatchFunctionMatchType1DescendingNumberDataExactMatchPresentTest()
        //{
        //    Worksheet worksheet = CreateWorksheetScenario(sampleDescendingLookupFunctionCellInput);

        //    string a1CellInput = "=Match(1{ls}B2:B5{ls}1)";
        //    FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        //}

        //[TestMethod]
        //public void MatchFunctionMatchType1DescendingNumberDataExactMatchPresentTest2()
        //{
        //    Worksheet worksheet = CreateWorksheetScenario(sampleDescendingLookupFunctionCellInput);

        //    string a1CellInput = "=Match(2{ls}B2:B5{ls}1)";
        //    FunctionsTestHelper.AssertCellValue(a1CellInput, "2", worksheet);
        //}

        //[TestMethod]
        //public void MatchFunctionMatchTypeMinus1AscendingNumberDataExactMatchPresentTest()
        //{
        //    Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

        //    string a1CellInput = "=Match(2{ls}B2:B5{ls}-1)";
        //    FunctionsTestHelper.AssertErrorCellValue(a1CellInput, ErrorExpressions.NotAvailableError, worksheet);
        //}

        //[TestMethod]
        //public void MatchFunctionMatchTypeMinus1AscendingNumberDataExactMatchPresentTest2()
        //{
        //    Worksheet worksheet = CreateWorksheetScenario(sampleLookupFunctionCellInput);

        //    string a1CellInput = "=Match(0{ls}B2:B5{ls}-1)";
        //    FunctionsTestHelper.AssertCellValue(a1CellInput, "1", worksheet);
        //}

        #endregion


        #region Helper methods

        private static void AssertBothVAndHLookup(string cellInput, string vLookupTable, string hLookupTable, string expectedResult, Worksheet vLookupWorksheet, Worksheet hLookupWorksheet)
        {
            string vLookupCellInput = ReplaceFunctionSpecificSubstrings(cellInput, "Vlookup", vLookupTable);
            FunctionsTestHelper.AssertCellValue(vLookupCellInput, expectedResult, vLookupWorksheet);

            string hLookupCellInput = ReplaceFunctionSpecificSubstrings(cellInput, "Hlookup", hLookupTable);
            FunctionsTestHelper.AssertCellValue(hLookupCellInput, expectedResult, hLookupWorksheet);
        }

        private static void AssertErrorForBothVAndHLookup(string cellInput, string vLookupTable, string hLookupTable, ErrorExpression expectedError, Worksheet vLookupWorksheet, Worksheet hLookupWorksheet)
        {
            AssertBothVAndHLookup(cellInput, vLookupTable, hLookupTable, expectedError.Value, vLookupWorksheet, hLookupWorksheet);
        }

        private static string ReplaceFunctionSpecificSubstrings(string cellInput, string functionName, string rangeName)
        {
            cellInput = cellInput.Replace("{fn}", functionName);
            cellInput = cellInput.Replace("{range}", rangeName);

            return cellInput;
        }

        private static string GetSubArrayStringFromStringArray(string[,] array, CellIndex fromIndex, CellIndex toIndex)
        {
            StringBuilder result = new StringBuilder();
            result.Append("{");

            for (int row = fromIndex.RowIndex; row <= toIndex.RowIndex; row++)
            {
                for (int column = fromIndex.ColumnIndex; column <= toIndex.ColumnIndex; column++)
                {
                    string element = array[row, column];

                    if (element == null)
                    {
                        element = string.Empty;
                    }

                    double number;
                    bool boolean;
                    if (double.TryParse(element, out number) ||
                        bool.TryParse(element, out boolean) ||
                        IsError(element))
                    {
                        result.Append(element);
                    }
                    else
                    {
                        result.AppendFormat("\"{0}\"", element);
                    }

                    if (column < toIndex.ColumnIndex)
                    {
                        result.Append("{als}");
                    }
                    else if (row < toIndex.RowIndex)
                    {
                        result.Append("{ars}");
                    }
                }
            }

            result.Append("}");

            return result.ToString();
        }

        // TODO: Think of a public method in ErrorExpressions class that gives this info!
        private static bool IsError(string value)
        {
            string upper = value.ToUpper();

            return upper == ErrorExpressions.CyclicReference.Value ||
                upper == ErrorExpressions.DivisionByZero.Value ||
                upper == ErrorExpressions.NameError.Value ||
                upper == ErrorExpressions.NotAvailableError.Value ||
                upper == ErrorExpressions.NullError.Value ||
                upper == ErrorExpressions.NumberError.Value ||
                upper == ErrorExpressions.ReferenceError.Value ||
                upper == ErrorExpressions.ValueError.Value;
        }

        private static Worksheet CreateWorksheetScenario(string b2CellValue)
        {
            return CreateWorksheetScenario(new string[,] { { b2CellValue } });
        }

        private static Worksheet CreateWorksheetScenario(string[,] cellValues)
        {
            Workbook wb = new Workbook();
            Worksheet worksheet = wb.Worksheets.Add();
            Cells cells = worksheet.Cells;

            for (int row = cellValues.GetLength(0); row > 0; row--)
            {
                for (int column = cellValues.GetLength(1); column > 0; column--)
                {
                    if (cellValues[row - 1, column - 1] != null)
                    {
                        cells[row, column].SetValue(cellValues[row - 1, column - 1].ReplaceCultureDependentSeparators());
                    }
                }
            }

            return worksheet;
        }

        #endregion
    }
}
