﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests
{
    [TestClass]
    public class GroupingTests
    {
        public static readonly int MaxOutlineLevel = 7;

        [TestMethod]
        public void GroupRowsBySettingOutlineLevelTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            int expectedRowOutlineLevel = 1;
            int expectedColumnOutlineLevel = 1;

            worksheet.Rows[0, 2].SetOutlineLevel(expectedRowOutlineLevel);
            worksheet.Columns[0, 2].SetOutlineLevel(expectedColumnOutlineLevel);

            int actualRowOutlineLevel = worksheet.Rows[0, 2].GetOutlineLevel().Value;
            int actualColumnOutlineLevel = worksheet.Columns[0, 2].GetOutlineLevel().Value;

            Assert.AreEqual(expectedRowOutlineLevel, actualRowOutlineLevel);
            Assert.AreEqual(expectedColumnOutlineLevel, actualColumnOutlineLevel);
        }

        [TestMethod]
        public void ClearGroupingByClearingOutlineLevelTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Rows[0, 2].SetOutlineLevel(1);
            worksheet.Columns[0, 2].SetOutlineLevel(1);

            worksheet.Rows[0, 2].ClearOutlineLevel();
            worksheet.Columns[0, 2].ClearOutlineLevel();

            RangePropertyValue<int> actualRowOutlineLevelRange = worksheet.Rows[0, 2].GetOutlineLevel();
            RangePropertyValue<int> actualColumnOutlineLevelRange = worksheet.Columns[0, 2].GetOutlineLevel();

            int actualRowOutlineLevel = actualRowOutlineLevelRange.Value;
            int actualColumnOutlineLevel = actualColumnOutlineLevelRange.Value;

            Assert.AreEqual(0, actualRowOutlineLevel);
            Assert.AreEqual(0, actualColumnOutlineLevel);

            Assert.IsFalse(actualRowOutlineLevelRange.IsIndeterminate);
            Assert.IsFalse(actualColumnOutlineLevelRange.IsIndeterminate);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void SetRowOutlineLevelPositiveOutOfRangeTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Rows[0, 2].SetOutlineLevel(8);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void SetRowOutlineLevelNegativeOutOfRangeTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Rows[0, 2].SetOutlineLevel(-1);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void SetColumnOutlineLevelPositiveOutOfRangeTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Columns[0, 2].SetOutlineLevel(8);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void SetColumnOutlineLevelNegativeOutOfRangeTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Columns[0, 2].SetOutlineLevel(-1);
        }

        [TestMethod]
        public void GroupThroughSelectionTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Rows[0, 2].Group();
            worksheet.Columns[0, 2].Group();

            int actualRowOutlineLevel = worksheet.Rows[0, 2].GetOutlineLevel().Value;
            int actualColumnOutlineLevel = worksheet.Columns[0, 2].GetOutlineLevel().Value;

            Assert.AreEqual(1, actualRowOutlineLevel);
            Assert.AreEqual(1, actualColumnOutlineLevel);
        }

        [TestMethod]
        public void UngroupThroughSelectionTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Rows[0, 2].Group();
            worksheet.Columns[0, 2].Group();

            worksheet.Rows[0, 2].Ungroup();
            worksheet.Columns[0, 2].Ungroup();

            int actualRowOutlineLevel = worksheet.Rows[0, 2].GetOutlineLevel().Value;
            int actualColumnOutlineLevel = worksheet.Columns[0, 2].GetOutlineLevel().Value;

            Assert.AreEqual(0, actualRowOutlineLevel);
            Assert.AreEqual(0, actualColumnOutlineLevel);
        }

        [TestMethod]
        public void UngroupPartOfExistingGroupThroughSelectionTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Rows[0, 2].Group();
            worksheet.Columns[0, 2].Group();

            worksheet.Rows[1].Ungroup();
            worksheet.Columns[1].Ungroup();

            int rowOutlineLevel0 = worksheet.Rows[0].GetOutlineLevel().Value;
            int columnOutlineLevel0 = worksheet.Columns[0].GetOutlineLevel().Value;

            Assert.AreEqual(1, rowOutlineLevel0);
            Assert.AreEqual(1, columnOutlineLevel0);

            int rowOutlineLevel1 = worksheet.Rows[1].GetOutlineLevel().Value;
            int columnOutlineLevel1 = worksheet.Columns[1].GetOutlineLevel().Value;

            Assert.AreEqual(0, rowOutlineLevel1);
            Assert.AreEqual(0, columnOutlineLevel1);

            int rowOutlineLevel2 = worksheet.Rows[2].GetOutlineLevel().Value;
            int columnOutlineLevel2 = worksheet.Columns[2].GetOutlineLevel().Value;

            Assert.AreEqual(1, rowOutlineLevel2);
            Assert.AreEqual(1, columnOutlineLevel2);
        }

        [TestMethod]
        public void GroupMultipleLevelsThroughSelectionTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Rows[0, 4].Group();
            worksheet.Rows[1, 3].Group();
            worksheet.Rows[2].Group();

            worksheet.Columns[0, 4].Group();
            worksheet.Columns[1, 3].Group();
            worksheet.Columns[2].Group();
            
            int rowOutlineLevel0 = worksheet.Rows[0].GetOutlineLevel().Value;
            Assert.AreEqual(1, rowOutlineLevel0);

            int rowOutlineLevel1 = worksheet.Rows[1].GetOutlineLevel().Value;
            Assert.AreEqual(2, rowOutlineLevel1);

            int rowOutlineLevel2 = worksheet.Rows[2].GetOutlineLevel().Value;
            Assert.AreEqual(3, rowOutlineLevel2);

            int rowOutlineLevel3 = worksheet.Rows[3].GetOutlineLevel().Value;
            Assert.AreEqual(2, rowOutlineLevel3);

            int rowOutlineLevel4 = worksheet.Rows[4].GetOutlineLevel().Value;
            Assert.AreEqual(1, rowOutlineLevel4);

            int columnOutlineLevel0 = worksheet.Columns[0].GetOutlineLevel().Value;
            Assert.AreEqual(1, columnOutlineLevel0);

            int columnOutlineLevel1 = worksheet.Columns[1].GetOutlineLevel().Value;
            Assert.AreEqual(2, columnOutlineLevel1);

            int columnOutlineLevel2 = worksheet.Columns[2].GetOutlineLevel().Value;
            Assert.AreEqual(3, columnOutlineLevel2);

            int columnOutlineLevel3 = worksheet.Columns[3].GetOutlineLevel().Value;
            Assert.AreEqual(2, columnOutlineLevel3);

            int columnOutlineLevel4 = worksheet.Columns[4].GetOutlineLevel().Value;
            Assert.AreEqual(1, columnOutlineLevel4);
        }

        [TestMethod]
        public void UngroupMultipleLevelsThroughSelectionTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Rows[0, 4].Group();
            worksheet.Rows[0, 4].Group();
            worksheet.Rows[0, 4].Group();
            worksheet.Rows[1, 3].Ungroup();
            worksheet.Rows[2].Ungroup();

            worksheet.Columns[0, 4].Group();
            worksheet.Columns[0, 4].Group();
            worksheet.Columns[0, 4].Group();
            worksheet.Columns[1, 3].Ungroup();
            worksheet.Columns[2].Ungroup();
            
            int rowOutlineLevel0 = worksheet.Rows[0].GetOutlineLevel().Value;
            Assert.AreEqual(3, rowOutlineLevel0);

            int rowOutlineLevel1 = worksheet.Rows[1].GetOutlineLevel().Value;
            Assert.AreEqual(2, rowOutlineLevel1);

            int rowOutlineLevel2 = worksheet.Rows[2].GetOutlineLevel().Value;
            Assert.AreEqual(1, rowOutlineLevel2);

            int rowOutlineLevel3 = worksheet.Rows[3].GetOutlineLevel().Value;
            Assert.AreEqual(2, rowOutlineLevel3);

            int rowOutlineLevel4 = worksheet.Rows[4].GetOutlineLevel().Value;
            Assert.AreEqual(3, rowOutlineLevel4);

            int columnOutlineLevel0 = worksheet.Columns[0].GetOutlineLevel().Value;
            Assert.AreEqual(3, columnOutlineLevel0);

            int columnOutlineLevel1 = worksheet.Columns[1].GetOutlineLevel().Value;
            Assert.AreEqual(2, columnOutlineLevel1);

            int columnOutlineLevel2 = worksheet.Columns[2].GetOutlineLevel().Value;
            Assert.AreEqual(1, columnOutlineLevel2);

            int columnOutlineLevel3 = worksheet.Columns[3].GetOutlineLevel().Value;
            Assert.AreEqual(2, columnOutlineLevel3);

            int columnOutlineLevel4 = worksheet.Columns[4].GetOutlineLevel().Value;
            Assert.AreEqual(3, columnOutlineLevel4);
        }

        [TestMethod]
        public void GroupOutOfRangeThroughSelectionTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            for (int i = 0; i < MaxOutlineLevel; i++)
            {
                worksheet.Rows[0, 4].Group();
            }

            Assert.IsFalse(worksheet.Rows[0, 4].Group());

            int rowOutlineLevel = worksheet.Rows[0, 4].GetOutlineLevel().Value;
            Assert.AreEqual(7, rowOutlineLevel);

            for (int i = 0; i < MaxOutlineLevel; i++)
            {
                worksheet.Columns[0, 4].Group();
            }

            Assert.IsFalse(worksheet.Columns[0, 4].Group());
            
            int columnOutlineLevel = worksheet.Columns[0, 4].GetOutlineLevel().Value;
            Assert.AreEqual(7, columnOutlineLevel);
        }

        [TestMethod]
        public void UngroupOutOfRangeThroughSelectionTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            bool rowUngroupResult = worksheet.Rows[0].Ungroup();
            int rowOutlineLevel = worksheet.Rows[0].GetOutlineLevel().Value;

            Assert.IsFalse(rowUngroupResult);
            Assert.AreEqual(0, rowOutlineLevel);

            bool columnUngroupResult = worksheet.Columns[0].Ungroup();
            int columnOutlineLevel = worksheet.Columns[0].GetOutlineLevel().Value;

            Assert.IsFalse(columnUngroupResult);
            Assert.AreEqual(0, columnOutlineLevel);
        }

        [TestMethod]
        public void GroupOnMultipleSelectionTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            List<CellRange> ranges = new List<CellRange> { new CellRange(0, 0, 0, 0), new CellRange(2, 2, 2, 2) };

            bool rowResult = worksheet.Rows[ranges].Group();
            bool columnResult = worksheet.Columns[ranges].Group();

            int rowOutlineLevel0 = worksheet.Rows[0].GetOutlineLevel().Value;
            int columnOutlineLevel0 = worksheet.Columns[0].GetOutlineLevel().Value;

            int rowOutlineLevel2 = worksheet.Rows[2].GetOutlineLevel().Value;
            int columnOutlineLevel2 = worksheet.Columns[2].GetOutlineLevel().Value;

            Assert.IsFalse(rowResult);
            Assert.IsFalse(columnResult);

            Assert.AreEqual(0, rowOutlineLevel0);
            Assert.AreEqual(0, columnOutlineLevel0);
            Assert.AreEqual(0, rowOutlineLevel2);
            Assert.AreEqual(0, columnOutlineLevel2);
        }
    }
}
