﻿using System;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests
{
    [TestClass]
    public class HyperlinkTests
    {
        private SpreadsheetHyperlink CreateAndAddHyperlink(Worksheet worksheet)
        {
            CellRange hyperlinkRange = new CellRange(2, 2, 4, 4);
            return worksheet.Hyperlinks.Add(hyperlinkRange, HyperlinkInfo.CreateInDocumentHyperlink("U10:V15"));
        }

        private Worksheet CreateWorksheetWithHyperlinkRange()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellRange hyperlinkRange = new CellRange(2, 2, 4, 4);

            worksheet.Hyperlinks.Add(hyperlinkRange, HyperlinkInfo.CreateInDocumentHyperlink("U10:V15"));

            return worksheet;
        }

        public bool TryFireHyperlinkCollectionChanged(Worksheet worksheet, Action[] actions, out int counter)
        {
            bool isEventFired = false;
            int innerCounter = 0;

            worksheet.Hyperlinks.Changed += (s, e) =>
            {
                isEventFired = !isEventFired;
                innerCounter++;
            };

            foreach (Action action in actions)
            {
                action();
            }

            counter = innerCounter;

            return isEventFired;
        }

        private void InsertRowOrColumnTest(int insertIndex, CellRange expectedRange, bool isRow)
        {
            Worksheet worksheet = this.CreateWorksheetWithHyperlinkRange();

            if (isRow)
            {
                Assert.IsTrue(worksheet.Rows.CanInsert(insertIndex, 1));
                worksheet.Rows.Insert(insertIndex);
            }
            else
            {
                Assert.IsTrue(worksheet.Columns.CanInsert(insertIndex, 1));
                worksheet.Columns.Insert(insertIndex);
            }

            Assert.AreEqual(1, worksheet.Hyperlinks.Count);
            Assert.AreEqual(expectedRange, worksheet.Hyperlinks.First().Range);
        }

        private void DeleteRowOrColumnTest(int removeIndex, CellRange expectedRange, bool isRow)
        {
            Worksheet worksheet = this.CreateWorksheetWithHyperlinkRange();

            if (isRow)
            {
                worksheet.Rows.Remove(removeIndex);
            }
            else
            {
                worksheet.Columns.Remove(removeIndex);
            }

            Assert.AreEqual(1, worksheet.Hyperlinks.Count);
            Assert.AreEqual(expectedRange, worksheet.Hyperlinks.First().Range);
        }

        private void InsertCellTest(int rowIndex, int columnIndex, InsertShiftType shiftType, CellRange expectedRange)
        {
            Worksheet worksheet = this.CreateWorksheetWithHyperlinkRange();
            worksheet.Cells.Insert(rowIndex, columnIndex, 1, shiftType);

            Assert.AreEqual(1, worksheet.Hyperlinks.Count);
            Assert.AreEqual(expectedRange, worksheet.Hyperlinks.First().Range);
        }

        private void DeleteCellTest(int rowIndex, int columnIndex, RemoveShiftType shiftType, CellRange expectedRange)
        {
            Worksheet worksheet = this.CreateWorksheetWithHyperlinkRange();
            worksheet.Cells.Remove(rowIndex, columnIndex, 1, shiftType);

            Assert.AreEqual(1, worksheet.Hyperlinks.Count);
            Assert.AreEqual(expectedRange, worksheet.Hyperlinks.First().Range);
        }

        [TestMethod]
        public void TestInsertRowOnTopOfHyperlinkRange()
        {
            CellRange expectedCellRange = new CellRange(3, 2, 5, 4);
            this.InsertRowOrColumnTest(1, expectedCellRange, true);
        }

        [TestMethod]
        public void TestInsertColumnToTheLeftOfHyperlinkRange()
        {
            CellRange expectedCellRange = new CellRange(2, 3, 4, 5);
            this.InsertRowOrColumnTest(1, expectedCellRange, false);
        }

        [TestMethod]
        public void TestInsertRowInTheMiddleOfHyperlinkRange()
        {
            CellRange expectedCellRange = new CellRange(2, 2, 5, 4);
            this.InsertRowOrColumnTest(3, expectedCellRange, true);
        }

        [TestMethod]
        public void TestInsertColumnInTheMiddleOfHyperlinkRange()
        {
            CellRange expectedCellRange = new CellRange(2, 2, 4, 5);
            this.InsertRowOrColumnTest(3, expectedCellRange, false);
        }

        [TestMethod]
        public void TestDeleteRowOnTopOfHyperlinkRange()
        {
            CellRange expectedCellRange = new CellRange(1, 2, 3, 4);
            this.DeleteRowOrColumnTest(1, expectedCellRange, true);
        }

        [TestMethod]
        public void TestDeleteColumnToTheLeftOfHyperlinkRange()
        {
            CellRange expectedCellRange = new CellRange(2, 1, 4, 3);
            this.DeleteRowOrColumnTest(1, expectedCellRange, false);
        }

        [TestMethod]
        public void TestDeleteRowInTheMiddleOfHyperlinkRange()
        {
            CellRange expectedCellRange = new CellRange(2, 2, 3, 4);
            this.DeleteRowOrColumnTest(3, expectedCellRange, true);
        }

        [TestMethod]
        public void TestDeleteColumnInTheMiddleOfHyperlinkRange()
        {
            CellRange expectedCellRange = new CellRange(2, 2, 4, 3);
            this.DeleteRowOrColumnTest(3, expectedCellRange, false);
        }

        [TestMethod]
        public void TestInsertCellOnTopOfHyperlinkRange()
        {
            CellRange expectedRange = new CellRange(2, 2, 4, 4);
            this.InsertCellTest(1, 3, InsertShiftType.Down, expectedRange);
        }

        [TestMethod]
        public void TestInsertCellToTheLeftOfHyperlinkRange()
        {
            CellRange expectedRange = new CellRange(2, 2, 4, 4);
            this.InsertCellTest(3, 1, InsertShiftType.Right, expectedRange);
        }

        [TestMethod]
        public void TestInsertCellDownInTheMiddleOfHyperlinkRange()
        {
            CellRange expectedRange = new CellRange(2, 2, 4, 4);
            this.InsertCellTest(3, 3, InsertShiftType.Down, expectedRange);
        }

        [TestMethod]
        public void TestInsertCellRightInTheMiddleOfHyperlinkRange()
        {
            CellRange expectedRange = new CellRange(2, 2, 4, 4);
            this.InsertCellTest(3, 3, InsertShiftType.Right, expectedRange);
        }

        [TestMethod]
        public void TestDeleteCellOnTopOfHyperlinkRange()
        {
            CellRange expectedRange = new CellRange(2, 2, 4, 4);
            this.DeleteCellTest(1, 3, RemoveShiftType.Up, expectedRange);
        }

        [TestMethod]
        public void TestDeleteCellToTheLeftOfHyperlinkRange()
        {
            CellRange expectedRange = new CellRange(2, 2, 4, 4);
            this.DeleteCellTest(3, 1, RemoveShiftType.Left, expectedRange);
        }

        [TestMethod]
        public void TestDeleteCellUpInTheMiddleOfHyperlinkRange()
        {
            CellRange expectedRange = new CellRange(2, 2, 4, 4);
            this.DeleteCellTest(3, 3, RemoveShiftType.Up, expectedRange);
        }

        [TestMethod]
        public void TestDeleteCellLeftInTheMiddleOfHyperlinkRange()
        {
            CellRange expectedRange = new CellRange(2, 2, 4, 4);
            this.DeleteCellTest(3, 3, RemoveShiftType.Left, expectedRange);
        }

        [TestMethod]
        public void HyperlinkCollectionChangedEvent_WhenHyperlinkIsAddedAndUndoExecuted_IsFired()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            this.CreateAndAddHyperlink(worksheet);

            int actualEventFiredCount;
            Action invokeUndo = new Action(() => worksheet.Workbook.History.Undo());
            
            bool isChangedEventThrown = this.TryFireHyperlinkCollectionChanged(worksheet, new Action[] { invokeUndo }, out actualEventFiredCount);

            int expectedEventFiredCount = 1;

            Assert.AreEqual(expectedEventFiredCount, actualEventFiredCount);
            Assert.IsTrue(isChangedEventThrown);
        }

        [TestMethod]
        public void HyperlinkCollectionChangedEvent_WhenHyperlinkIsAddedRedoExecuted_IsFired()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            this.CreateAndAddHyperlink(worksheet);
            worksheet.Workbook.History.Undo();

            int actualEventFiredCount;
            Action invokeRedo = new Action(() => worksheet.Workbook.History.Redo());

            bool isChangedEventThrown = this.TryFireHyperlinkCollectionChanged(worksheet, new Action[] { invokeRedo }, out actualEventFiredCount);

            int expectedEventFiredCount = 1;

            Assert.AreEqual(expectedEventFiredCount, actualEventFiredCount);
            Assert.IsTrue(isChangedEventThrown);
        }

        [TestMethod]
        public void HyperlinkCollectionChangedEvent_WhenAddRemoveUndoRedoAreExecuted_IsFiredProperTimes()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            SpreadsheetHyperlink hyperlink = this.CreateAndAddHyperlink(worksheet);

            int actualEventFiredCount;
            Action[] actions = new Action[]
            {
                () => worksheet.Hyperlinks.Remove(hyperlink),
                () => worksheet.Workbook.History.Undo(),
                () => worksheet.Workbook.History.Redo()
            };

            this.TryFireHyperlinkCollectionChanged(worksheet, actions, out actualEventFiredCount);

            int expectedEventFiredCount = 3;

            Assert.AreEqual(expectedEventFiredCount, actualEventFiredCount);
        }
    }
}