﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Collections.Generic;
using System.IO;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Filtering;
using Telerik.Windows.Documents.Spreadsheet.Model.Printing;
using Telerik.Windows.Documents.Spreadsheet.Model.Sorting;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests.ImportExportTests
{
    [TestClass]
    public class EscapedStringsTests
    {
        // All characters not allowed in XML.
        private char[] escapedChars = new char[]
 {
            (char)0x00,
            (char)0x01,
            (char)0x02,
            (char)0x03,
            (char)0x04,
            (char)0x05,
            (char)0x06,
            (char)0x07,
            (char)0x08,
            (char)0x0B,
            (char)0x0C,
            (char)0x0E,
            (char)0x0F,
            (char)0x10,
            (char)0x11,
            (char)0x12,
            (char)0x13,
            (char)0x14,
            (char)0x15,
            (char)0x16,
            (char)0x17,
            (char)0x18,
            (char)0x19,
            (char)0x1A,
            (char)0x1B,
            (char)0x1C,
            (char)0x1D,
            (char)0x1E,
            (char)0x1F,
            (char)0xFFFE,
            (char)0xFFFF,
 };

        [TestMethod]
        [TestCategory("ImportExport")]
        public void ExportAndImportWorksheetThatHasIllegalChars_CharsAreSkipped()
        {
            Workbook originalWorkbook = this.ArrangeWorkbook(((char)27).ToString());
            Workbook expectedWorkbook = this.ArrangeWorkbook(string.Empty);

            using (Stream stream = new MemoryStream())
            {
                XlsxFormatProvider provider = new XlsxFormatProvider();
                provider.Export(originalWorkbook, stream);

                Workbook actualWorkbook = provider.Import(stream);

                TestHelper.AssertWorkbooksAreEqual(expectedWorkbook, actualWorkbook);
            }
        }

        public Workbook ArrangeWorkbook(string value)
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets["Sheet1"];

            // Text element contents have to be escaped.
            worksheet.Cells[0, 0].SetValue(value);

            // Format contents have to be escaped.
            worksheet.Cells[0, 1].SetFormat(new CellValueFormat(value));

            // Formula content should be escaped.
            worksheet.Cells[0, 2].SetValue(string.Format("=\"{0}\"", value));

            // Comment and refersTo value of the name have to be escaped. The name attribute also can contain escaped char according to the specification, but Excel does not
            // accept those, but instead notifies that the workbook is corrupted.
            worksheet.Names.Add("name", string.Format("=\"{0}\"", value), new CellIndex(0, 0), string.Format("comment {0}", value));

            // Custom filter and values collection filter values have to be escaped.
            worksheet.Filter.FilterRange = new CellRange(0, 1, 1, 2);
            CustomFilterCriteria critera = new CustomFilterCriteria(ComparisonOperator.EqualsTo, string.Format("filter {0}", value));
            CustomFilter customFilter = new CustomFilter(0, critera);
            worksheet.Filter.SetFilter(customFilter);
            ValuesCollectionFilter valuesFilter = new ValuesCollectionFilter(1, new List<string>() { string.Format("filter {0}", value) });
            worksheet.Filter.SetFilter(valuesFilter);
            worksheet.Rows[1].SetHidden(true);

            // Sort condition custom list values have to be escaped.
            worksheet.Cells[2, 1, 2, 1].SetValue(string.Format("some value: {0}", value));
            CustomValuesSortCondition condition = new CustomValuesSortCondition(0, new string[] { string.Format("sort condition: {0}", value) }, SortOrder.Ascending);
            worksheet.SortState.Set(new CellRange(2, 1, 2, 1), condition);

            // Hyperlink location and tooltip have to be escaped.
            HyperlinkInfo hyperlink = HyperlinkInfo.CreateHyperlink(value, value);
            worksheet.Hyperlinks.Add(new CellIndex(2, 1), hyperlink);

            // Headers and footers content has to be escaped.
            HeaderFooterSettings settings = worksheet.WorksheetPageSetup.HeaderFooterSettings;
            settings.Header.RightSection.Text = value;

            // Adding all chars that have to be escaped.
            for (int i = 0; i < this.escapedChars.Length; i++)
            {
                char character = this.escapedChars[i];
                value = value == string.Empty ? string.Empty : character.ToString();
                worksheet.Cells[i, 3].SetValue(value);
            }

            return workbook;
        }
    }
}