﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests.ImportExportTests
{
    internal enum SharedStringDocumentType
    {
        SharedStrings,
        SingleRichTextElement,
        MultipleRichTextElements, 
        SharedStringDifferentProperties,
        EmptyStringTextElement
    }

    [TestClass]
    public class SharedStringsTests : ImportExportTestClassBase
    {
        private const string EmptyString = "";
        private const string TestString = "Test";
        private const string RichTextString = "RichText";
        private const string DifferentRichTextString = "Different Text";
        private const string NormalTextString = "Normal text";
        private const string LeadingSpacesText = "             Leading spaces text";
        private const string TrailingSpacesText = "Trailing spaces text               ";

        private const string ImportSharedStringsDocument = "SharedStringsDocuments/ImportSharedStringsDocument.xlsx";
        private const string ImportSingleRichTextElementDocument = "SharedStringsDocuments/ImportSingleRichTextElementDocument.xlsx";
        private const string ImportDiffrentRichTextElementsEqualTextDocument = "SharedStringsDocuments/ImportDiffrentRichTextElementsWithEqualTextDocument.xlsx";
        private const string ImportSharedStringDifferentPropertiesDocument = "SharedStringsDocuments/ImportSharedStringDifferentPropertiesDocument.xlsx";
        private const string ImportEmptyStringDocument = "SharedStringsDocuments/ImportEmptyString.xlsx";

        private SharedStringDocumentType documentType;

        private List<CellIndex> CreateCellIndicesCollection(int length, int rowIndex)
        {
            List<CellIndex> cellIndices = new List<CellIndex>();

            for (int i = 0; i < length; i++)
            {
                cellIndices.Add(new CellIndex(rowIndex, i));
            }

            return cellIndices;
        }

        private void PopulateCellValues(Worksheet worksheet, int rowIndex, int length, string textValue)
        {
            List<CellIndex> cellIndices = this.CreateCellIndicesCollection(length, rowIndex);

            foreach (var cellIndex in cellIndices)
            {
                worksheet.Cells[cellIndex].SetValue(textValue);
            }
        }

        private void AssertSharedStringsAreImportedCorrectly(Worksheet worksheet, CellValueType expectedCellValueType, string expectedRawValue,
            int rowIndex, int length)
        {
            List<CellIndex> cellIndices = this.CreateCellIndicesCollection(length, rowIndex);

            foreach (var cellIndex in cellIndices)
            {
                var cellValue = worksheet.Cells[cellIndex].GetValue().Value;

                var actualCellValueType = cellValue.ValueType;
                var actualRawValue = cellValue.RawValue;

                Assert.AreEqual(expectedCellValueType, actualCellValueType);
                Assert.AreEqual(expectedRawValue, actualRawValue);
            }
        }

        public override Workbook ArrangeWorkbook()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets["Sheet1"];

            this.PopulateCellValues(worksheet, 0, 7, RichTextString);
            this.PopulateCellValues(worksheet, 1, 7, DifferentRichTextString);
            this.PopulateCellValues(worksheet, 2, 7, NormalTextString);
            this.PopulateCellValues(worksheet, 3, 7, LeadingSpacesText);
            this.PopulateCellValues(worksheet, 4, 7, TrailingSpacesText);

            worksheet.Cells[5, 7].SetValueAsText(EmptyString);

            return workbook;
        }

        public override void AssertWorkbookImportedCorrectly(Workbook workbook)
        {
            Worksheet worksheet = workbook.Worksheets[0];

            switch (this.documentType)
            {
                case SharedStringDocumentType.SharedStrings:
                    {
                        this.AssertSharedStringsAreImportedCorrectly(worksheet, CellValueType.Text, RichTextString, 0, 7);
                        this.AssertSharedStringsAreImportedCorrectly(worksheet, CellValueType.Text, DifferentRichTextString, 1, 7);
                        this.AssertSharedStringsAreImportedCorrectly(worksheet, CellValueType.Text, NormalTextString, 2, 7);
                    }
                    break;

                case SharedStringDocumentType.SingleRichTextElement:
                    {
                        this.AssertSharedStringsAreImportedCorrectly(worksheet, CellValueType.Text, RichTextString, 0, 1);
                    }
                    break;

                case SharedStringDocumentType.MultipleRichTextElements:
                    {
                        this.AssertSharedStringsAreImportedCorrectly(worksheet, CellValueType.Text, RichTextString, 0, 1);
                        this.AssertSharedStringsAreImportedCorrectly(worksheet, CellValueType.Text, RichTextString, 1, 1);
                    }
                    break;

                case SharedStringDocumentType.SharedStringDifferentProperties:
                    {
                        this.AssertSharedStringsAreImportedCorrectly(worksheet, CellValueType.Text, TestString, 0, 1);
                        this.AssertSharedStringsAreImportedCorrectly(worksheet, CellValueType.Text, TestString, 1, 1);
                    }
                    break;

                case SharedStringDocumentType.EmptyStringTextElement:
                    {
                        this.AssertSharedStringsAreImportedCorrectly(worksheet, CellValueType.Text, EmptyString, 0, 1);
                    }
                    break;

                default:
                    break;
            }
        }

        [TestMethod]
        [TestCategory("ImportExport")]
        public void Document_WhenContainsDifferentRichTextElementsWithEqualText_IsImportedCorrectly()
        {
            this.documentType = SharedStringDocumentType.MultipleRichTextElements;

            this.ImportFileAndAssert(ImportDiffrentRichTextElementsEqualTextDocument);
        }

        [TestMethod]
        [TestCategory("ImportExport")]
        public void Document_WhenContainsSingleRichTextElement_IsImportedCorrectly()
        {
            this.documentType = SharedStringDocumentType.SingleRichTextElement;

            this.ImportFileAndAssert(ImportSingleRichTextElementDocument);
        }

        [TestMethod]
        [TestCategory("ImportExport")]
        public void Document_WhenContainsSharedStrings_IsImportedCorrectly()
        {
            this.documentType = SharedStringDocumentType.SharedStrings;

            this.ImportFileAndAssert(ImportSharedStringsDocument);
        }

        [TestMethod]
        [TestCategory("ImportExport")]
        public void Document_WhenContainsSharedStrings_IsExportedAndImportedCorrectly()
        {
            this.ArrangeWorkbookExportImportAssert();
        }

        [TestMethod]
        [TestCategory("ImportExport")]
        public void Document_WhenContainsSharedStringDifferentProperties_IsImportedCorrectly()
        {
            this.documentType = SharedStringDocumentType.SharedStringDifferentProperties;

            this.ImportFileAndAssert(ImportSharedStringDifferentPropertiesDocument);
        }

        [TestMethod]
        [TestCategory("ImportExport")]
        public void Document_WhenContainsEmptyStringSharedString_IsImportedCorrectly()
        {
            this.documentType = SharedStringDocumentType.EmptyStringTextElement;

            this.ImportFileAndAssert(ImportEmptyStringDocument);
        }
    }
}