﻿using System.Collections.Generic;
using System.Linq;
using System.Windows.Media;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests
{
    [TestClass]
    public class MergeTests
    {
        [TestMethod]
        public void MergeCellsWithDifferentFillUnmerge()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            CellSelection cellSelection = worksheet.Cells[0, 0, 4, 2];
            cellSelection.SetFill(PatternFill.CreateSolidFill(Colors.Red));

            CellSelection cellSelection1 = worksheet.Cells[4, 2, 4, 4];
            cellSelection1.SetFill(PatternFill.CreateSolidFill(Colors.Green));

            CellSelection cellSelection2 = worksheet.Cells[0, 0, 4, 4];
            cellSelection2.Merge();

            cellSelection2.Unmerge();

            IFill fill = cellSelection1.GetFill().Value;
            PatternFill patternFill = (PatternFill)fill;

            Assert.AreEqual(PatternFill.CreateSolidFill(Colors.Red), patternFill);
        }


        #region GetContainingMergedRanges method tests

        private void GenerateMergedCellsExample(out Worksheet worksheet, out CellRange[] mergedRanges)
        {
            worksheet = TestHelper.CreateWorksheet();

            mergedRanges = new CellRange[]
            {
                new CellRange(1, 1, 3, 3),
                new CellRange(1, 4, 5, 8),
                new CellRange(6, 1, 13, 7)
            };

            foreach (var range in mergedRanges)
            {
                worksheet.Cells[range].Merge();
            }
        }

        [TestMethod]
        public void GetMergedRangesTest()
        {
            Worksheet worksheet;
            CellRange[] mergedRanges;
            GenerateMergedCellsExample(out worksheet, out mergedRanges);

            CellRange selection = new CellRange(0, 0, 5, 3);
            IEnumerable<CellRange> ranges = worksheet.Cells.GetContainingMergedRanges(selection);

            Assert.AreEqual(1, ranges.Count());

            CellRange mergedRange = ranges.First();

            Assert.AreEqual(mergedRanges[0], mergedRange);
        }

        [TestMethod]
        public void GetMergedRangesTestForNotEntirelyContainedRanges()
        {
            Worksheet worksheet;
            CellRange[] mergedRanges;
            GenerateMergedCellsExample(out worksheet, out mergedRanges);

            CellRange selection = new CellRange(0, 0, 13, 7);
            IEnumerable<CellRange> ranges = worksheet.Cells.GetContainingMergedRanges(selection);
            int count = ranges.Count();
            Assert.AreEqual(2, count);

            CellRange mergedRangeFirst = ranges.First();
            Assert.AreEqual(mergedRanges[0], mergedRangeFirst);

            CellRange mergedRangeSecond = ranges.Last();
            Assert.AreEqual(mergedRanges[2], mergedRangeSecond);
        }

        [TestMethod]
        public void GetMergedRangesTestNoContaingMergedRanges()
        {
            Worksheet worksheet;
            CellRange[] mergedRanges;
            GenerateMergedCellsExample(out worksheet, out mergedRanges);

            CellRange selection = new CellRange(1, 2, 13, 7);
            IEnumerable<CellRange> ranges = worksheet.Cells.GetContainingMergedRanges(selection);

            int count = ranges.Count();
            Assert.AreEqual(0, count);
        }

        [TestMethod]
        public void GetMergedRangesTestAllContaingMergedRanges()
        {
            Worksheet worksheet;
            CellRange[] mergedRanges;
            GenerateMergedCellsExample(out worksheet, out mergedRanges);

            CellRange selection = new CellRange(0, 0, 50, 50);
            IEnumerable<CellRange> ranges = worksheet.Cells.GetContainingMergedRanges(selection);

            int count = ranges.Count();
            Assert.AreEqual(3, count);
        }

        #endregion


        #region Insert Cells

        [TestMethod]
        public void InsertCellsOnTheTopOfMergedRange()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            CellSelection cellSelection = worksheet.Cells[1, 1, 4, 4];
            cellSelection.Merge();

            CellSelection cellSelection1 = worksheet.Cells[0, 1, 0, 4];

            Assert.IsTrue(cellSelection1.CanInsertOrRemove(cellSelection1.CellRanges.First(), ShiftType.Down));

            cellSelection1.Insert(InsertShiftType.Down);

            CellRange expectedRange = new CellRange(2, 1, 5, 4);

            CellRange mergedRange;
            worksheet.Cells.TryGetContainingMergedRange(new CellIndex(2, 1), out mergedRange);

            Assert.AreEqual(expectedRange, mergedRange);
        }

        [TestMethod]
        public void InsertCellsInTheMiddleOfMergedRange()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            CellSelection cellSelection = worksheet.Cells[1, 1, 4, 4];
            cellSelection.Merge();

            CellSelection cellSelection1 = worksheet.Cells[2, 1, 2, 4];

            Assert.IsTrue(cellSelection1.CanInsertOrRemove(cellSelection1.CellRanges.First(), ShiftType.Down));

            cellSelection1.Insert(InsertShiftType.Down);

            CellRange expectedRange = new CellRange(1, 1, 5, 4);

            CellRange mergedRange;
            worksheet.Cells.TryGetContainingMergedRange(new CellIndex(1, 1), out mergedRange);

            Assert.AreEqual(expectedRange, mergedRange);
        }

        [TestMethod]
        public void InsertCellsInTheMiddleOfTwoMergedRanges()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Cells[1, 1, 4, 4].Merge();
            worksheet.Cells[1, 6, 4, 10].Merge();

            worksheet.Columns[3].Insert();
            worksheet.Columns[6].Insert();

            CellRange mergedRange;
            worksheet.Cells.TryGetContainingMergedRange(new CellIndex(1, 1), out mergedRange);

            CellRange mergedRange1;
            worksheet.Cells.TryGetContainingMergedRange(new CellIndex(1, 8), out mergedRange1);

            CellRange expectedRange = new CellRange(1, 1, 4, 5);
            CellRange expectedRange1 = new CellRange(1, 8, 4, 12);

            Assert.AreEqual(expectedRange, mergedRange);
            Assert.AreEqual(expectedRange1, mergedRange1);
        }

        [TestMethod]
        public void InsertCellsTwoMergedRanges()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Cells[4, 0, 10, 3].Merge();
            worksheet.Cells[1, 4, 11, 7].Merge();

            worksheet.Cells[1, 0, 11, 7].Insert(InsertShiftType.Down);

            CellRange mergedRange;
            worksheet.Cells.TryGetContainingMergedRange(new CellIndex(15, 0), out mergedRange);

            CellRange mergedRange1;
            worksheet.Cells.TryGetContainingMergedRange(new CellIndex(12, 4), out mergedRange1);

            CellRange expectedRange = new CellRange(15, 0, 21, 3);
            CellRange expectedRange1 = new CellRange(12, 4, 22, 7);

            Assert.AreEqual(expectedRange, mergedRange);
            Assert.AreEqual(expectedRange1, mergedRange1);
        }

        #endregion


        #region Remove Cells

        [TestMethod]
        public void RemoveCellsOnTheTopOfMergedRange()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            CellSelection cellSelection = worksheet.Cells[2, 1, 5, 4];
            cellSelection.Merge();

            CellSelection cellSelection1 = worksheet.Cells[0, 1, 0, 4];

            Assert.IsTrue(cellSelection1.CanInsertOrRemove(cellSelection1.CellRanges.First(), ShiftType.Up));

            cellSelection1.Remove(RemoveShiftType.Up);

            CellRange expectedRange = new CellRange(1, 1, 4, 4);

            CellRange mergedRange;
            worksheet.Cells.TryGetContainingMergedRange(new CellIndex(1, 1), out mergedRange);

            Assert.AreEqual(expectedRange, mergedRange);
        }

        [TestMethod]
        public void RemoveCellsInTheMiddleOfMergedRange()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            CellSelection cellSelection = worksheet.Cells[1, 1, 5, 4];
            cellSelection.Merge();

            CellSelection cellSelection1 = worksheet.Cells[2, 1, 2, 4];

            Assert.IsTrue(cellSelection1.CanInsertOrRemove(cellSelection1.CellRanges.First(), ShiftType.Down));

            cellSelection1.Remove(RemoveShiftType.Up);

            CellRange expectedRange = new CellRange(1, 1, 4, 4);

            CellRange mergedRange;
            worksheet.Cells.TryGetContainingMergedRange(new CellIndex(1, 1), out mergedRange);

            Assert.AreEqual(expectedRange, mergedRange);
        }

        [TestMethod]
        public void RemoveCellsInTheMiddleOfTwoMergedRanges()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Cells[1, 1, 4, 5].Merge();
            worksheet.Cells[1, 8, 4, 12].Merge();

            worksheet.Columns[3].Remove();
            worksheet.Columns[6].Remove();

            CellRange mergedRange;
            worksheet.Cells.TryGetContainingMergedRange(new CellIndex(1, 1), out mergedRange);

            CellRange mergedRange1;
            worksheet.Cells.TryGetContainingMergedRange(new CellIndex(1, 6), out mergedRange1);

            CellRange expectedRange = new CellRange(1, 1, 4, 4);
            CellRange expectedRange1 = new CellRange(1, 6, 4, 10);

            Assert.AreEqual(expectedRange, mergedRange);
            Assert.AreEqual(expectedRange1, mergedRange1);
        }

        [TestMethod]
        public void RemoveCellsTwoMergedRanges()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Cells[15, 0, 21, 3].Merge();
            worksheet.Cells[12, 4, 22, 7].Merge();

            worksheet.Cells[1, 0, 11, 7].Remove(RemoveShiftType.Up);

            CellRange mergedRange;
            worksheet.Cells.TryGetContainingMergedRange(new CellIndex(4, 0), out mergedRange);

            CellRange mergedRange1;
            worksheet.Cells.TryGetContainingMergedRange(new CellIndex(1, 4), out mergedRange1);

            CellRange expectedRange = new CellRange(4, 0, 10, 3);
            CellRange expectedRange1 = new CellRange(1, 4, 11, 7);

            Assert.AreEqual(expectedRange, mergedRange);
            Assert.AreEqual(expectedRange1, mergedRange1);
        }

        #endregion
    }
}