﻿using System;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Theming;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests
{
    [TestClass]
    public class CellPropertiesTests
    {
        #region Test UsedCellRange

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void GetUsedCellRangeForNoProperties()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            CellRange expected = new CellRange(0, 0, 0, 0);
            CellRange result = worksheet.GetUsedCellRange(Enumerable.Empty<IPropertyDefinition>());

            Assert.AreEqual(expected, result);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void GetUsedCellRangeForOneProperty()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            CellRange expected = new CellRange(0, 0, 4, 4);

            worksheet.Cells[4, 4].SetValue(1);
            worksheet.Cells[5, 5].SetIsBold(true);
            CellRange result = worksheet.GetUsedCellRange(new IPropertyDefinition[] { CellPropertyDefinitions.ValueProperty });

            Assert.AreEqual(expected, result);
        }

        #endregion


        #region Test CellSelection Properties

        #region Generic Property Methods

        private void AssertCellRangeProperty<T>(RangePropertyValue<T> rangePropertyValue, bool expectedIsIndeterminate, T expectedValue)
        {
            bool isIndeterminate = rangePropertyValue.IsIndeterminate;
            T value = rangePropertyValue.Value;

            Assert.AreEqual(expectedIsIndeterminate, isIndeterminate);
            Assert.AreEqual(expectedValue, value);
        }

        private void AssertCellRangeFontSize(RangePropertyValue<double> rangePropertyValue, bool expectedIsIndeterminate, double expectedValue)
        {
            bool isIndeterminate = rangePropertyValue.IsIndeterminate;
            double value = rangePropertyValue.Value;

            Assert.AreEqual(expectedIsIndeterminate, isIndeterminate);
            Assert.AreEqual(expectedValue, Math.Round(value, 2));
        }

        private void AssertCellRangeValue(RangePropertyValue<ICellValue> rangePropertyValue, string expectedValue, bool expectedIsIndeterminate)
        {
            bool isIndeterminate = rangePropertyValue.IsIndeterminate;
            ICellValue value = rangePropertyValue.Value;

            CellValueFormat cellValueFormat = new CellValueFormat(string.Empty);

            Assert.AreEqual(expectedIsIndeterminate, isIndeterminate);
            Assert.AreEqual(expectedValue, value.GetValueAsString(cellValueFormat));
        }

        private void AssertCellRangeFormat(RangePropertyValue<CellValueFormat> rangePropertyValue, CellValueFormat expectedValue, bool expectedIsIndeterminate)
        {
            bool isIndeterminate = rangePropertyValue.IsIndeterminate;
            CellValueFormat value = rangePropertyValue.Value;

            Assert.AreEqual(expectedIsIndeterminate, isIndeterminate);
            Assert.AreEqual(expectedValue, value);
        }


        private void TestOneSelectionRangeGetSetProperty<T>(
            T defaultValue, T newValue,
            Func<CellSelection, RangePropertyValue<T>> getPropertyValue,
            Action<CellSelection, T> setPropertyValue)
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, newValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            setPropertyValue(cellSelection, defaultValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);
        }

        private void TestTwoSelectionRangesGetSetProperty<T>(
            T defaultValue, T newValue,
            Func<CellSelection, RangePropertyValue<T>> getPropertyValue,
            Action<CellSelection, T> setPropertyValue)
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[
                new CellRange[]
                {
                    new CellRange(0, 0, 4, 4),
                    new CellRange(6, 6, 10, 10)
                }];
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, newValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            setPropertyValue(cellSelection, defaultValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);
        }

        private void TestTwoCrossSelectionRangesGetSetProperty<T>(
            T defaultValue, T newValue,
            Func<CellSelection, RangePropertyValue<T>> getPropertyValue,
            Action<CellSelection, T> setPropertyValue)
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[
                new CellRange[]
                {
                    new CellRange(0, 0, 4, 4),
                    new CellRange(3, 3, 10, 10)
                }];
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, newValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            setPropertyValue(cellSelection, defaultValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);
        }

        private void TestOneSelectionRangeClearProperty<T>(
             T defaultValue, T newValue,
             Func<CellSelection, RangePropertyValue<T>> getPropertyValue,
             Action<CellSelection, T> setPropertyValue,
             Action<CellSelection> clearPropertyValue)
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, newValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            clearPropertyValue(cellSelection);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, defaultValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            clearPropertyValue(cellSelection);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);
        }

        private void TestTwoSelectionRangesClearProperty<T>(
             T defaultValue, T newValue,
             Func<CellSelection, RangePropertyValue<T>> getPropertyValue,
             Action<CellSelection, T> setPropertyValue,
             Action<CellSelection> clearPropertyValue)
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[
                new CellRange[]
                {
                    new CellRange(0, 0, 4, 4),
                    new CellRange(6, 6, 10, 10)
                }];
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, newValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            clearPropertyValue(cellSelection);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, defaultValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            clearPropertyValue(cellSelection);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);
        }

        private void TestTwoCrossSelectionRangesClearProperty<T>(
             T defaultValue, T newValue,
             Func<CellSelection, RangePropertyValue<T>> getPropertyValue,
             Action<CellSelection, T> setPropertyValue,
             Action<CellSelection> clearPropertyValue)
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[
                new CellRange[]
                {
                    new CellRange(0, 0, 4, 4),
                    new CellRange(3, 3, 10, 10)
                }];
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, newValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            clearPropertyValue(cellSelection);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, defaultValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            clearPropertyValue(cellSelection);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);
        }

        private void TestOneSelectionRangeUndoRedoProperty<T>(
            T defaultValue, T newValue,
            Func<CellSelection, RangePropertyValue<T>> getPropertyValue,
            Action<CellSelection, T> setPropertyValue)
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, newValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            workbook.History.Undo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            workbook.History.Redo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            setPropertyValue(cellSelection, defaultValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            workbook.History.Undo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            workbook.History.Redo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);
        }

        private void TestTwoSelectionRangesUndoRedoProperty<T>(
            T defaultValue, T newValue,
            Func<CellSelection, RangePropertyValue<T>> getPropertyValue,
            Action<CellSelection, T> setPropertyValue)
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[
                new CellRange[]
                {
                    new CellRange(0, 0, 4, 4),
                    new CellRange(6, 6, 10, 10)
                }];
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, newValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            workbook.History.Undo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            workbook.History.Redo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            setPropertyValue(cellSelection, defaultValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            workbook.History.Undo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            workbook.History.Redo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);
        }

        private void TestTwoCrossSelectionRangesUndoRedoProperty<T>(
            T defaultValue, T newValue,
            Func<CellSelection, RangePropertyValue<T>> getPropertyValue,
            Action<CellSelection, T> setPropertyValue)
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[
                new CellRange[]
                {
                    new CellRange(0, 0, 4, 4),
                    new CellRange(3, 3, 10, 10)
                }];
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, newValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            workbook.History.Undo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            workbook.History.Redo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            setPropertyValue(cellSelection, defaultValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            workbook.History.Undo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            workbook.History.Redo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);
        }

        private void TestOneSelectionRangeClearUndoRedoProperty<T>(
            T defaultValue, T newValue,
            Func<CellSelection, RangePropertyValue<T>> getPropertyValue,
            Action<CellSelection, T> setPropertyValue,
            Action<CellSelection> clearPropertyValue)
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, newValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            clearPropertyValue(cellSelection);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            workbook.History.Undo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            workbook.History.Redo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, defaultValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            clearPropertyValue(cellSelection);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            workbook.History.Undo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            workbook.History.Redo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);
        }

        private void TestTwoSelectionRangesClearUndoRedoProperty<T>(
            T defaultValue, T newValue,
            Func<CellSelection, RangePropertyValue<T>> getPropertyValue,
            Action<CellSelection, T> setPropertyValue,
            Action<CellSelection> clearPropertyValue)
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[
                new CellRange[]
                {
                    new CellRange(0, 0, 4, 4),
                    new CellRange(6, 6, 10, 10)
                }];
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, newValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            clearPropertyValue(cellSelection);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            workbook.History.Undo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            workbook.History.Redo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, defaultValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            clearPropertyValue(cellSelection);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            workbook.History.Undo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            workbook.History.Redo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);
        }

        private void TestTwoCrossSelectionRangesClearUndoRedoProperty<T>(
            T defaultValue, T newValue,
            Func<CellSelection, RangePropertyValue<T>> getPropertyValue,
            Action<CellSelection, T> setPropertyValue,
            Action<CellSelection> clearPropertyValue)
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[
                new CellRange[]
                {
                    new CellRange(0, 0, 4, 4),
                    new CellRange(3, 3, 10, 10)
                }];
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, newValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            clearPropertyValue(cellSelection);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            workbook.History.Undo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, newValue);

            workbook.History.Redo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            setPropertyValue(cellSelection, defaultValue);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            clearPropertyValue(cellSelection);
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            workbook.History.Undo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);

            workbook.History.Redo();
            AssertCellRangeProperty(getPropertyValue(cellSelection), false, defaultValue);
        }

        #endregion

        // TODO: Add tests for all SetValue method overloads
        #region Value Property

        const string defaultValue = "";
        const string newValue = "test";

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeGetSetValue()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];
            RangePropertyValue<ICellValue> rangePropertyValue = cellSelection.GetValue();
            AssertCellRangeValue(rangePropertyValue, defaultValue, false);

            cellSelection.SetValue(newValue);
            rangePropertyValue = cellSelection.GetValue();
            AssertCellRangeValue(rangePropertyValue, newValue, false);

            cellSelection.SetValue(defaultValue);
            rangePropertyValue = cellSelection.GetValue();
            AssertCellRangeValue(rangePropertyValue, defaultValue, false);
        }

        #endregion

        // TODO: Add Format Property tests for all selection cases. Please, refer to Generic Property Methods region.
        #region Format Property

        CellValueFormat defaultFormat = new CellValueFormat(string.Empty);
        CellValueFormat newFormat = new CellValueFormat("test");

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeGetSetFormat()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];
            RangePropertyValue<CellValueFormat> rangePropertyFormat = cellSelection.GetFormat();
            AssertCellRangeFormat(rangePropertyFormat, defaultFormat, false);

            cellSelection.SetFormat(newFormat);
            rangePropertyFormat = cellSelection.GetFormat();
            AssertCellRangeFormat(rangePropertyFormat, newFormat, false);

            cellSelection.SetFormat(defaultFormat);
            rangePropertyFormat = cellSelection.GetFormat();
            AssertCellRangeFormat(rangePropertyFormat, defaultFormat, false);
        }

        #endregion


        #region StyleName Property

        const string normalStyleName = "Normal";
        const string calculationStyleName = "Calculation";

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeGetSetStyleName()
        {
            TestOneSelectionRangeGetSetProperty(normalStyleName, calculationStyleName,
                (cellSelection) =>
                {
                    return cellSelection.GetStyleName();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetStyleName(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesGetSetStyleName()
        {
            TestTwoSelectionRangesGetSetProperty(normalStyleName, calculationStyleName,
                (cellSelection) =>
                {
                    return cellSelection.GetStyleName();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetStyleName(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesGetSetStyleName()
        {
            TestTwoCrossSelectionRangesGetSetProperty(normalStyleName, calculationStyleName,
                (cellSelection) =>
                {
                    return cellSelection.GetStyleName();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetStyleName(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearStyleName()
        {
            TestOneSelectionRangeClearProperty(normalStyleName, calculationStyleName,
                (cellSelection) =>
                {
                    return cellSelection.GetStyleName();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetStyleName(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearStyleName();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearStyleName()
        {
            TestTwoSelectionRangesClearProperty(normalStyleName, calculationStyleName,
                (cellSelection) =>
                {
                    return cellSelection.GetStyleName();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetStyleName(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearStyleName();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearStyleName()
        {
            TestTwoCrossSelectionRangesClearProperty(normalStyleName, calculationStyleName,
                (cellSelection) =>
                {
                    return cellSelection.GetStyleName();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetStyleName(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearStyleName();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeUndoRedoStyleName()
        {
            TestOneSelectionRangeUndoRedoProperty(normalStyleName, calculationStyleName,
                (cellSelection) =>
                {
                    return cellSelection.GetStyleName();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetStyleName(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesUndoRedoStyleName()
        {
            TestTwoSelectionRangesUndoRedoProperty(normalStyleName, calculationStyleName,
                (cellSelection) =>
                {
                    return cellSelection.GetStyleName();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetStyleName(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesUndoRedoStyleName()
        {
            TestTwoCrossSelectionRangesUndoRedoProperty(normalStyleName, calculationStyleName,
                (cellSelection) =>
                {
                    return cellSelection.GetStyleName();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetStyleName(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearUndoRedoStyleName()
        {
            TestOneSelectionRangeClearUndoRedoProperty(normalStyleName, calculationStyleName,
                (cellSelection) =>
                {
                    return cellSelection.GetStyleName();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetStyleName(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearStyleName();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearUndoRedoStyleName()
        {
            TestTwoSelectionRangesClearUndoRedoProperty(normalStyleName, calculationStyleName,
                (cellSelection) =>
                {
                    return cellSelection.GetStyleName();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetStyleName(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearStyleName();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearUndoRedoStyleName()
        {
            TestTwoCrossSelectionRangesClearUndoRedoProperty(normalStyleName, calculationStyleName,
                (cellSelection) =>
                {
                    return cellSelection.GetStyleName();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetStyleName(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearStyleName();
                });
        }

        #endregion

        // TODO: Add Fill Property tests both for gradient and pattern fills for all selection cases. Please, refer to Generic Property Methods region.
        #region Fill Property



        #endregion


        #region FontFamily Property

        static ThemeFontType themeFontTypeMinor = new ThemeFontType();
        ThemableFontFamily themableFontFamilyMinor = new ThemableFontFamily(themeFontTypeMinor);
        ThemableFontFamily themableFontFamilyArial = new ThemableFontFamily("Arial");

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeGetSetFontFamily()
        {
            TestOneSelectionRangeGetSetProperty(themableFontFamilyMinor, themableFontFamilyArial,
                (cellSelection) =>
                {
                    return cellSelection.GetFontFamily();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetFontFamily(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesGetSetFontFamily()
        {
            TestTwoSelectionRangesGetSetProperty(themableFontFamilyMinor, themableFontFamilyArial,
                (cellSelection) =>
                {
                    return cellSelection.GetFontFamily();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetFontFamily(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesGetSetFontFamily()
        {
            TestTwoCrossSelectionRangesGetSetProperty(themableFontFamilyMinor, themableFontFamilyArial,
                (cellSelection) =>
                {
                    return cellSelection.GetFontFamily();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetFontFamily(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearFontFamily()
        {
            TestOneSelectionRangeClearProperty(themableFontFamilyMinor, themableFontFamilyArial,
                (cellSelection) =>
                {
                    return cellSelection.GetFontFamily();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetFontFamily(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearFontFamily();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearFontFamily()
        {
            TestTwoSelectionRangesClearProperty(themableFontFamilyMinor, themableFontFamilyArial,
                (cellSelection) =>
                {
                    return cellSelection.GetFontFamily();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetFontFamily(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearFontFamily();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearFontFamily()
        {
            TestTwoCrossSelectionRangesClearProperty(themableFontFamilyMinor, themableFontFamilyArial,
                (cellSelection) =>
                {
                    return cellSelection.GetFontFamily();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetFontFamily(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearFontFamily();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeUndoRedoFontFamily()
        {
            TestOneSelectionRangeUndoRedoProperty(themableFontFamilyMinor, themableFontFamilyArial,
                (cellSelection) =>
                {
                    return cellSelection.GetFontFamily();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetFontFamily(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesUndoRedoFontFamily()
        {
            TestTwoSelectionRangesUndoRedoProperty(themableFontFamilyMinor, themableFontFamilyArial,
                (cellSelection) =>
                {
                    return cellSelection.GetFontFamily();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetFontFamily(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesUndoRedoFontFamily()
        {
            TestTwoCrossSelectionRangesUndoRedoProperty(themableFontFamilyMinor, themableFontFamilyArial,
                (cellSelection) =>
                {
                    return cellSelection.GetFontFamily();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetFontFamily(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearUndoRedoFontFamily()
        {
            TestOneSelectionRangeClearUndoRedoProperty(themableFontFamilyMinor, themableFontFamilyArial,
                (cellSelection) =>
                {
                    return cellSelection.GetFontFamily();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetFontFamily(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearFontFamily();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearUndoRedoFontFamily()
        {
            TestTwoSelectionRangesClearUndoRedoProperty(themableFontFamilyMinor, themableFontFamilyArial,
                (cellSelection) =>
                {
                    return cellSelection.GetFontFamily();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetFontFamily(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearFontFamily();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearUndoRedoFontFamily()
        {
            TestTwoCrossSelectionRangesClearUndoRedoProperty(themableFontFamilyMinor, themableFontFamilyArial,
                (cellSelection) =>
                {
                    return cellSelection.GetFontFamily();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetFontFamily(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearFontFamily();
                });
        }

        #endregion


        #region FontSize Property

        const double defaultFontSize = 14.67;
        const double newFontSize = 13.34;

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void TestFontSizeNegativeThrowsException()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];

            cellSelection.SetFontSize(UnitHelper.PointToDip(-1));
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void TestSetFontSizeZeroThrowsException()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];

            cellSelection.SetFontSize(UnitHelper.PointToDip(0));
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void TestSetFontSizeDoubleMaxValueThrowsException()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];

            cellSelection.SetFontSize(UnitHelper.PointToDip(410));
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeGetSetFontSize()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];
            RangePropertyValue<double> rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(newFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            cellSelection.SetFontSize(defaultFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesGetSetFontSize()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[
                new CellRange[]
                {
                    new CellRange(0, 0, 4, 4),
                    new CellRange(6, 6, 10, 10)
                }];
            RangePropertyValue<double> rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(newFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            cellSelection.SetFontSize(defaultFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesGetSetFontSize()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[
                new CellRange[]
                {
                    new CellRange(0, 0, 4, 4),
                    new CellRange(3, 3, 10, 10)
                }];
            RangePropertyValue<double> rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(newFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            cellSelection.SetFontSize(defaultFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearFontSize()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];
            RangePropertyValue<double> rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(newFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            cellSelection.ClearFontSize();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(defaultFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.ClearFontSize();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearFontSize()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[
                new CellRange[]
                {
                    new CellRange(0, 0, 4, 4),
                    new CellRange(6, 6, 10, 10)
                }];
            RangePropertyValue<double> rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(newFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            cellSelection.ClearFontSize();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(defaultFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.ClearFontSize();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearFontSize()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[
                new CellRange[]
                {
                    new CellRange(0, 0, 4, 4),
                    new CellRange(3, 3, 10, 10)
                }];
            RangePropertyValue<double> rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(newFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            cellSelection.ClearFontSize();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(defaultFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.ClearFontSize();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeUndoRedoFontSize()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];
            RangePropertyValue<double> rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(newFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            cellSelection.SetFontSize(defaultFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesUndoRedoFontSize()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[
                new CellRange[] 
                { 
                    new CellRange(0, 0, 4, 4), 
                    new CellRange(6, 10, 10, 6)
                }];
            RangePropertyValue<double> rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(newFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            cellSelection.SetFontSize(defaultFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesUndoRedoFontSize()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[
                new CellRange[] 
                { 
                    new CellRange(0, 0, 4, 4), 
                    new CellRange(3, 3, 10, 10)
                }];
            RangePropertyValue<double> rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(newFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            cellSelection.SetFontSize(defaultFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearUndoRedoFontSize()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];
            RangePropertyValue<double> rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(newFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            cellSelection.ClearFontSize();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(defaultFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.ClearFontSize();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearUndoRedoFontSize()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[
                new CellRange[] 
                { 
                    new CellRange(0, 0, 4, 4), 
                    new CellRange(6, 10, 10, 6)
                }];
            RangePropertyValue<double> rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(newFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            cellSelection.ClearFontSize();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(defaultFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.ClearFontSize();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearUndoRedoFontSize()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[
                new CellRange[] 
                { 
                    new CellRange(0, 0, 4, 4), 
                    new CellRange(3, 3, 10, 10)
                }];
            RangePropertyValue<double> rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(newFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            cellSelection.ClearFontSize();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, newFontSize);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.SetFontSize(defaultFontSize);
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            cellSelection.ClearFontSize();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetFontSize();
            AssertCellRangeFontSize(rangePropertyValue, false, defaultFontSize);
        }

        #endregion


        #region IsBold Property

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeGetSetIsBold()
        {
            TestOneSelectionRangeGetSetProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsBold();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsBold(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesGetSetIsBold()
        {
            TestTwoSelectionRangesGetSetProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsBold();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsBold(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesGetSetIsBold()
        {
            TestTwoCrossSelectionRangesGetSetProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsBold();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsBold(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearIsBold()
        {
            TestOneSelectionRangeClearProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsBold();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsBold(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsBold();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearIsBold()
        {
            TestTwoSelectionRangesClearProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsBold();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsBold(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsBold();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearIsBold()
        {
            TestTwoCrossSelectionRangesClearProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsBold();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsBold(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsBold();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeUndoRedoIsBold()
        {
            TestOneSelectionRangeUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsBold();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsBold(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesUndoRedoIsBold()
        {
            TestTwoSelectionRangesUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsBold();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsBold(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesUndoRedoIsBold()
        {
            TestTwoCrossSelectionRangesUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsBold();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsBold(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearUndoRedoIsBold()
        {
            TestOneSelectionRangeClearUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsBold();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsBold(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsBold();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearUndoRedoIsBold()
        {
            TestTwoSelectionRangesClearUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsBold();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsBold(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsBold();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearUndoRedoIsBold()
        {
            TestTwoCrossSelectionRangesClearUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsBold();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsBold(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsBold();
                });
        }

        #endregion


        #region IsItalic Property

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeGetSetIsItalic()
        {
            TestOneSelectionRangeGetSetProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsItalic();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsItalic(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesGetSetIsItalic()
        {
            TestTwoSelectionRangesGetSetProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsItalic();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsItalic(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesGetSetIsItalic()
        {
            TestTwoCrossSelectionRangesGetSetProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsItalic();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsItalic(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearIsItalic()
        {
            TestOneSelectionRangeClearProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsItalic();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsItalic(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsItalic();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearIsItalic()
        {
            TestTwoSelectionRangesClearProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsItalic();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsItalic(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsItalic();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearIsItalic()
        {
            TestTwoCrossSelectionRangesClearProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsItalic();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsItalic(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsItalic();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeUndoRedoIsItalic()
        {
            TestOneSelectionRangeUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsItalic();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsItalic(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesUndoRedoIsItalic()
        {
            TestTwoSelectionRangesUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsItalic();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsItalic(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesUndoRedoIsItalic()
        {
            TestTwoCrossSelectionRangesUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsItalic();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsItalic(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearUndoRedoIsItalic()
        {
            TestOneSelectionRangeClearUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsItalic();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsItalic(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsItalic();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearUndoRedoIsItalic()
        {
            TestTwoSelectionRangesClearUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsItalic();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsItalic(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsItalic();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearUndoRedoIsItalic()
        {
            TestTwoCrossSelectionRangesClearUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsItalic();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsItalic(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsItalic();
                });
        }

        #endregion


        #region Underline Property

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeGetSetUnderline()
        {
            TestOneSelectionRangeGetSetProperty(UnderlineType.None, UnderlineType.Single,
                (cellSelection) =>
                {
                    return cellSelection.GetUnderline();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetUnderline(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesGetSetUnderline()
        {
            TestTwoSelectionRangesGetSetProperty(UnderlineType.None, UnderlineType.Single,
                (cellSelection) =>
                {
                    return cellSelection.GetUnderline();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetUnderline(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesGetSetUnderline()
        {
            TestTwoCrossSelectionRangesGetSetProperty(UnderlineType.None, UnderlineType.Single,
                (cellSelection) =>
                {
                    return cellSelection.GetUnderline();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetUnderline(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearUnderline()
        {
            TestOneSelectionRangeClearProperty(UnderlineType.None, UnderlineType.Single,
                (cellSelection) =>
                {
                    return cellSelection.GetUnderline();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetUnderline(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearUnderline();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearUnderline()
        {
            TestTwoSelectionRangesClearProperty(UnderlineType.None, UnderlineType.Single,
                (cellSelection) =>
                {
                    return cellSelection.GetUnderline();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetUnderline(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearUnderline();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearUnderline()
        {
            TestTwoCrossSelectionRangesClearProperty(UnderlineType.None, UnderlineType.Single,
                (cellSelection) =>
                {
                    return cellSelection.GetUnderline();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetUnderline(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearUnderline();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeUndoRedoUnderline()
        {
            TestOneSelectionRangeUndoRedoProperty(UnderlineType.None, UnderlineType.Single,
                (cellSelection) =>
                {
                    return cellSelection.GetUnderline();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetUnderline(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesUndoRedoUnderline()
        {
            TestTwoSelectionRangesUndoRedoProperty(UnderlineType.None, UnderlineType.Single,
               (cellSelection) =>
               {
                   return cellSelection.GetUnderline();
               },
               (cellSelection, value) =>
               {
                   cellSelection.SetUnderline(value);
               });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesUndoRedoUnderline()
        {
            TestTwoCrossSelectionRangesUndoRedoProperty(UnderlineType.None, UnderlineType.Single,
               (cellSelection) =>
               {
                   return cellSelection.GetUnderline();
               },
               (cellSelection, value) =>
               {
                   cellSelection.SetUnderline(value);
               });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearUndoRedoUnderline()
        {
            TestOneSelectionRangeClearUndoRedoProperty(UnderlineType.None, UnderlineType.Single,
                (cellSelection) =>
                {
                    return cellSelection.GetUnderline();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetUnderline(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearUnderline();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearUndoRedoUnderline()
        {
            TestTwoSelectionRangesClearUndoRedoProperty(UnderlineType.None, UnderlineType.Single,
                (cellSelection) =>
                {
                    return cellSelection.GetUnderline();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetUnderline(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearUnderline();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearUndoRedoUnderline()
        {
            TestTwoCrossSelectionRangesClearUndoRedoProperty(UnderlineType.None, UnderlineType.Single,
                (cellSelection) =>
                {
                    return cellSelection.GetUnderline();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetUnderline(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearUnderline();
                });
        }

        #endregion


        #region ForeColor Property

        ThemableColor themableColorText1 = new ThemableColor(ThemeColorType.Text1);
        ThemableColor themableColorHyperlink = new ThemableColor(ThemeColorType.Hyperlink);

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeGetSetForeColor()
        {
            TestOneSelectionRangeGetSetProperty(themableColorText1, themableColorHyperlink,
                (cellSelection) =>
                {
                    return cellSelection.GetForeColor();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetForeColor(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesGetSetForeColor()
        {
            TestTwoSelectionRangesGetSetProperty(themableColorText1, themableColorHyperlink,
                (cellSelection) =>
                {
                    return cellSelection.GetForeColor();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetForeColor(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesGetSetForeColor()
        {
            TestTwoCrossSelectionRangesGetSetProperty(themableColorText1, themableColorHyperlink,
                (cellSelection) =>
                {
                    return cellSelection.GetForeColor();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetForeColor(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearForeColor()
        {
            TestOneSelectionRangeClearProperty(themableColorText1, themableColorHyperlink,
                (cellSelection) =>
                {
                    return cellSelection.GetForeColor();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetForeColor(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearForeColor();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearForeColor()
        {
            TestTwoSelectionRangesClearProperty(themableColorText1, themableColorHyperlink,
                (cellSelection) =>
                {
                    return cellSelection.GetForeColor();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetForeColor(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearForeColor();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearForeColor()
        {
            TestTwoCrossSelectionRangesClearProperty(themableColorText1, themableColorHyperlink,
                (cellSelection) =>
                {
                    return cellSelection.GetForeColor();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetForeColor(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearForeColor();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeUndoRedoForeColor()
        {
            TestOneSelectionRangeUndoRedoProperty(themableColorText1, themableColorHyperlink,
                 (cellSelection) =>
                 {
                     return cellSelection.GetForeColor();
                 },
                 (cellSelection, value) =>
                 {
                     cellSelection.SetForeColor(value);
                 });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesUndoRedoForeColor()
        {
            TestTwoSelectionRangesUndoRedoProperty(themableColorText1, themableColorHyperlink,
                 (cellSelection) =>
                 {
                     return cellSelection.GetForeColor();
                 },
                 (cellSelection, value) =>
                 {
                     cellSelection.SetForeColor(value);
                 });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesUndoRedoForeColor()
        {
            TestTwoCrossSelectionRangesUndoRedoProperty(themableColorText1, themableColorHyperlink,
                 (cellSelection) =>
                 {
                     return cellSelection.GetForeColor();
                 },
                 (cellSelection, value) =>
                 {
                     cellSelection.SetForeColor(value);
                 });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearUndoRedoForeColor()
        {
            TestOneSelectionRangeClearUndoRedoProperty(themableColorText1, themableColorHyperlink,
                (cellSelection) =>
                {
                    return cellSelection.GetForeColor();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetForeColor(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearForeColor();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearUndoRedoForeColor()
        {
            TestTwoSelectionRangesClearUndoRedoProperty(themableColorText1, themableColorHyperlink,
                (cellSelection) =>
                {
                    return cellSelection.GetForeColor();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetForeColor(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearForeColor();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearUndoRedoForeColor()
        {
            TestTwoCrossSelectionRangesClearUndoRedoProperty(themableColorText1, themableColorHyperlink,
                (cellSelection) =>
                {
                    return cellSelection.GetForeColor();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetForeColor(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearForeColor();
                });
        }

        #endregion


        #region HorizontalAlignment Property

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeGetSetHorizontalAlignment()
        {
            TestOneSelectionRangeGetSetProperty(RadHorizontalAlignment.General, RadHorizontalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetHorizontalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetHorizontalAlignment(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesGetSetHorizontalAlignment()
        {
            TestTwoSelectionRangesGetSetProperty(RadHorizontalAlignment.General, RadHorizontalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetHorizontalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetHorizontalAlignment(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesGetSetHorizontalAlignment()
        {
            TestTwoCrossSelectionRangesGetSetProperty(RadHorizontalAlignment.General, RadHorizontalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetHorizontalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetHorizontalAlignment(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearHorizontalAlignment()
        {
            TestOneSelectionRangeClearProperty(RadHorizontalAlignment.General, RadHorizontalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetHorizontalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetHorizontalAlignment(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearHorizontalAlignment();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearHorizontalAlignment()
        {
            TestTwoSelectionRangesClearProperty(RadHorizontalAlignment.General, RadHorizontalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetHorizontalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetHorizontalAlignment(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearHorizontalAlignment();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearHorizontalAlignment()
        {
            TestTwoCrossSelectionRangesClearProperty(RadHorizontalAlignment.General, RadHorizontalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetHorizontalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetHorizontalAlignment(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearHorizontalAlignment();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeUndoRedoHorizontalAlignment()
        {
            TestOneSelectionRangeUndoRedoProperty(RadHorizontalAlignment.General, RadHorizontalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetHorizontalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetHorizontalAlignment(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesUndoRedoHorizontalAlignment()
        {
            TestTwoSelectionRangesUndoRedoProperty(RadHorizontalAlignment.General, RadHorizontalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetHorizontalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetHorizontalAlignment(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesUndoRedoHorizontalAlignment()
        {
            TestTwoCrossSelectionRangesUndoRedoProperty(RadHorizontalAlignment.General, RadHorizontalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetHorizontalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetHorizontalAlignment(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearUndoRedoHorizontalAlignment()
        {
            TestOneSelectionRangeClearUndoRedoProperty(RadHorizontalAlignment.General, RadHorizontalAlignment.Center,
                 (cellSelection) =>
                 {
                     return cellSelection.GetHorizontalAlignment();
                 },
                 (cellSelection, value) =>
                 {
                     cellSelection.SetHorizontalAlignment(value);
                 },
                 (cellSelection) =>
                 {
                     cellSelection.ClearHorizontalAlignment();
                 });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearUndoRedoHorizontalAlignment()
        {
            TestTwoSelectionRangesClearUndoRedoProperty(RadHorizontalAlignment.General, RadHorizontalAlignment.Center,
                 (cellSelection) =>
                 {
                     return cellSelection.GetHorizontalAlignment();
                 },
                 (cellSelection, value) =>
                 {
                     cellSelection.SetHorizontalAlignment(value);
                 },
                 (cellSelection) =>
                 {
                     cellSelection.ClearHorizontalAlignment();
                 });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearUndoRedoHorizontalAlignment()
        {
            TestTwoCrossSelectionRangesClearUndoRedoProperty(RadHorizontalAlignment.General, RadHorizontalAlignment.Center,
                 (cellSelection) =>
                 {
                     return cellSelection.GetHorizontalAlignment();
                 },
                 (cellSelection, value) =>
                 {
                     cellSelection.SetHorizontalAlignment(value);
                 },
                 (cellSelection) =>
                 {
                     cellSelection.ClearHorizontalAlignment();
                 });
        }

        #endregion


        #region VerticalAlignment Property

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeGetSetVerticalAlignment()
        {
            TestOneSelectionRangeGetSetProperty(RadVerticalAlignment.Bottom, RadVerticalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetVerticalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetVerticalAlignment(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesGetSetVerticalAlignment()
        {
            TestTwoSelectionRangesGetSetProperty(RadVerticalAlignment.Bottom, RadVerticalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetVerticalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetVerticalAlignment(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesGetSetVerticalAlignment()
        {
            TestTwoCrossSelectionRangesGetSetProperty(RadVerticalAlignment.Bottom, RadVerticalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetVerticalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetVerticalAlignment(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearVerticalAlignment()
        {
            TestOneSelectionRangeClearProperty(RadVerticalAlignment.Bottom, RadVerticalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetVerticalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetVerticalAlignment(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearVerticalAlignment();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearVerticalAlignment()
        {
            TestTwoSelectionRangesClearProperty(RadVerticalAlignment.Bottom, RadVerticalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetVerticalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetVerticalAlignment(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearVerticalAlignment();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearVerticalAlignment()
        {
            TestTwoCrossSelectionRangesClearProperty(RadVerticalAlignment.Bottom, RadVerticalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetVerticalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetVerticalAlignment(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearVerticalAlignment();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeUndoRedoVerticalAlignment()
        {
            TestOneSelectionRangeUndoRedoProperty(RadVerticalAlignment.Bottom, RadVerticalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetVerticalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetVerticalAlignment(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesUndoRedoVerticalAlignment()
        {
            TestTwoSelectionRangesUndoRedoProperty(RadVerticalAlignment.Bottom, RadVerticalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetVerticalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetVerticalAlignment(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesUndoRedoVerticalAlignment()
        {
            TestTwoCrossSelectionRangesUndoRedoProperty(RadVerticalAlignment.Bottom, RadVerticalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetVerticalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetVerticalAlignment(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearUndoRedoVerticalAlignment()
        {
            TestOneSelectionRangeClearUndoRedoProperty(RadVerticalAlignment.Bottom, RadVerticalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetVerticalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetVerticalAlignment(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearVerticalAlignment();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearUndoRedoVerticalAlignment()
        {
            TestTwoSelectionRangesClearUndoRedoProperty(RadVerticalAlignment.Bottom, RadVerticalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetVerticalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetVerticalAlignment(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearVerticalAlignment();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearUndoRedoVerticalAlignment()
        {
            TestTwoCrossSelectionRangesClearUndoRedoProperty(RadVerticalAlignment.Bottom, RadVerticalAlignment.Center,
                (cellSelection) =>
                {
                    return cellSelection.GetVerticalAlignment();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetVerticalAlignment(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearVerticalAlignment();
                });
        }

        #endregion


        #region Indent Property

        const int defaultIndent = 0;
        const int newIndent = 1;
        const int maxIndent = 250;
        const int increasedIndent = 2;
        const int decreasedIndent = 249;

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void TestSetIndentNegativeThrowsException()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];

            cellSelection.SetIndent(-1);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void TestSetIndentGreaterThanMaxValueThrowsException()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];

            cellSelection.SetIndent(251);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeGetSetIndent()
        {
            TestOneSelectionRangeGetSetProperty(defaultIndent, newIndent,
                (cellSelection) =>
                {
                    return cellSelection.GetIndent();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIndent(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesGetSetIndent()
        {
            TestTwoSelectionRangesGetSetProperty(defaultIndent, newIndent,
                (cellSelection) =>
                {
                    return cellSelection.GetIndent();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIndent(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesGetSetIndent()
        {
            TestTwoCrossSelectionRangesGetSetProperty(defaultIndent, newIndent,
                (cellSelection) =>
                {
                    return cellSelection.GetIndent();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIndent(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearIndent()
        {
            TestOneSelectionRangeClearProperty(defaultIndent, newIndent,
                (cellSelection) =>
                {
                    return cellSelection.GetIndent();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIndent(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIndent();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearIndent()
        {
            TestTwoSelectionRangesClearProperty(defaultIndent, newIndent,
                (cellSelection) =>
                {
                    return cellSelection.GetIndent();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIndent(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIndent();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearIndent()
        {
            TestTwoCrossSelectionRangesClearProperty(defaultIndent, newIndent,
                (cellSelection) =>
                {
                    return cellSelection.GetIndent();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIndent(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIndent();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeUndoRedoIndent()
        {
            TestOneSelectionRangeUndoRedoProperty(defaultIndent, newIndent,
                (cellSelection) =>
                {
                    return cellSelection.GetIndent();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIndent(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesUndoRedoIndent()
        {
            TestTwoSelectionRangesUndoRedoProperty(defaultIndent, newIndent,
                (cellSelection) =>
                {
                    return cellSelection.GetIndent();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIndent(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesUndoRedoIndent()
        {
            TestTwoCrossSelectionRangesUndoRedoProperty(defaultIndent, newIndent,
                (cellSelection) =>
                {
                    return cellSelection.GetIndent();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIndent(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearUndoRedoIndent()
        {
            TestOneSelectionRangeClearUndoRedoProperty(defaultIndent, newIndent,
                (cellSelection) =>
                {
                    return cellSelection.GetIndent();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIndent(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIndent();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearUndoRedoIndent()
        {
            TestTwoSelectionRangesClearUndoRedoProperty(defaultIndent, newIndent,
                (cellSelection) =>
                {
                    return cellSelection.GetIndent();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIndent(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIndent();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearUndoRedoIndent()
        {
            TestTwoCrossSelectionRangesClearUndoRedoProperty(defaultIndent, newIndent,
                (cellSelection) =>
                {
                    return cellSelection.GetIndent();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIndent(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIndent();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestIncreaseIndentMaxValueThrowsException()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];
            RangePropertyValue<int> rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, defaultIndent);

            cellSelection.SetIndent(maxIndent);
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, maxIndent);

            cellSelection.IncreaseIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, maxIndent);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestDecreaseIndentDefaultValueThrowsException()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];
            RangePropertyValue<int> rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, defaultIndent);

            cellSelection.DecreaseIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, defaultIndent);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeIncreaseIndent()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];
            RangePropertyValue<int> rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, defaultIndent);

            cellSelection.IncreaseIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, newIndent);

            cellSelection.SetIndent(newIndent);
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, newIndent);

            cellSelection.IncreaseIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, increasedIndent);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, newIndent);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, increasedIndent);

            cellSelection.ClearIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, defaultIndent);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangeIncreaseIndent()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[
                new CellRange[]
                {
                    new CellRange(0, 0, 4, 4),
                    new CellRange(6, 6, 10, 10)
                }];
            RangePropertyValue<int> rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, defaultIndent);

            cellSelection.IncreaseIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, newIndent);

            cellSelection.SetIndent(newIndent);
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, newIndent);

            cellSelection.IncreaseIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, increasedIndent);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, newIndent);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, increasedIndent);

            cellSelection.ClearIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, defaultIndent);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangeIncreaseIndent()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[
                new CellRange[]
                {
                    new CellRange(0, 0, 4, 4),
                    new CellRange(3, 3, 10, 10)
                }];
            RangePropertyValue<int> rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, defaultIndent);

            cellSelection.IncreaseIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, true, defaultIndent);

            cellSelection.SetIndent(newIndent);
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, newIndent);

            cellSelection.IncreaseIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, true, defaultIndent);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, newIndent);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, true, defaultIndent);

            cellSelection.ClearIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, defaultIndent);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeDecreaseIndent()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[0, 0, 1, 1];
            RangePropertyValue<int> rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, defaultIndent);

            cellSelection.SetIndent(maxIndent);
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, maxIndent);

            cellSelection.DecreaseIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, decreasedIndent);

            cellSelection.ClearIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, defaultIndent);

            cellSelection.SetIndent(maxIndent);
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, maxIndent);

            cellSelection.DecreaseIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, decreasedIndent);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, maxIndent);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, decreasedIndent);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangeDecreaseIndent()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[
                new CellRange[]
                {
                    new CellRange(0, 0, 4, 4),
                    new CellRange(6, 6, 10, 10)
                }];
            RangePropertyValue<int> rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, defaultIndent);

            cellSelection.SetIndent(maxIndent);
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, maxIndent);

            cellSelection.DecreaseIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, decreasedIndent);

            cellSelection.ClearIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, defaultIndent);

            cellSelection.SetIndent(maxIndent);
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, maxIndent);

            cellSelection.DecreaseIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, decreasedIndent);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, maxIndent);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, decreasedIndent);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangeDecreaseIndent()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Workbook workbook = worksheet.Workbook;

            CellSelection cellSelection = worksheet.Cells[
                new CellRange[]
                {
                    new CellRange(0, 0, 4, 4),
                    new CellRange(3, 3, 10, 10)
                }];
            RangePropertyValue<int> rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, defaultIndent);

            cellSelection.SetIndent(maxIndent);
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, maxIndent);

            cellSelection.DecreaseIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, true, defaultIndent);

            cellSelection.ClearIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, defaultIndent);

            cellSelection.SetIndent(maxIndent);
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, maxIndent);

            cellSelection.DecreaseIndent();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, true, defaultIndent);

            workbook.History.Undo();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, false, maxIndent);

            workbook.History.Redo();
            rangePropertyValue = cellSelection.GetIndent();
            AssertCellRangeProperty(rangePropertyValue, true, defaultIndent);
        }

        #endregion


        #region IsWrapped Property

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeGetSetIsWrapped()
        {
            TestOneSelectionRangeGetSetProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsWrapped();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsWrapped(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesGetSetIsWrapped()
        {
            TestTwoSelectionRangesGetSetProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsWrapped();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsWrapped(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesGetSetIsWrapped()
        {
            TestTwoCrossSelectionRangesGetSetProperty(false, true,
               (cellSelection) =>
               {
                   return cellSelection.GetIsWrapped();
               },
               (cellSelection, value) =>
               {
                   cellSelection.SetIsWrapped(value);
               });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearIsWrapped()
        {
            TestOneSelectionRangeClearProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsWrapped();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsWrapped(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsWrapped();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearIsWrapped()
        {
            TestTwoSelectionRangesClearProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsWrapped();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsWrapped(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsWrapped();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearIsWrapped()
        {
            TestTwoCrossSelectionRangesClearProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsWrapped();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsWrapped(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsWrapped();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeUndoRedoIsWrapped()
        {
            TestOneSelectionRangeUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsWrapped();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsWrapped(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesUndoRedoIsWrapped()
        {
            TestTwoSelectionRangesUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsWrapped();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsWrapped(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesUndoRedoIsWrapped()
        {
            TestTwoCrossSelectionRangesUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsWrapped();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsWrapped(value);
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeClearUndoRedoIsWrapped()
        {
            TestOneSelectionRangeClearUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsWrapped();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsWrapped(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsWrapped();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSelectionRangesClearUndoRedoIsWrapped()
        {
            TestTwoSelectionRangesClearUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsWrapped();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsWrapped(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsWrapped();
                });
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoCrossSelectionRangesClearUndoRedoIsWrapped()
        {
            TestTwoCrossSelectionRangesClearUndoRedoProperty(false, true,
                (cellSelection) =>
                {
                    return cellSelection.GetIsWrapped();
                },
                (cellSelection, value) =>
                {
                    cellSelection.SetIsWrapped(value);
                },
                (cellSelection) =>
                {
                    cellSelection.ClearIsWrapped();
                });
        }

        #endregion

        #endregion


        #region Test CellSelection Methods
        // TODO: Add remaining tests for all ramining methods of CellSelection for all selection cases. Please, refer to Generic Property Methods region.

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestOneSelectionRangeGetSetBorders()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellSelection cellSelection = worksheet.Cells[0, 0, 5, 5];

            CellBorder cellBorder = new CellBorder(CellBorderStyle.Dotted, new ThemableColor(ThemeColorType.Accent1, ColorShadeType.Shade3));
            CellBorders cellBorders = new CellBorders(cellBorder);

            cellSelection.SetBorders(cellBorders);

            CellBorders result = cellSelection.GetBorders();

            Assert.AreEqual(cellBorders, result);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestTwoSellectionRangesGetSetBorders()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            CellSelection cellSelection = worksheet.Cells[
                new CellRange[] 
                { 
                    new CellRange(0, 0, 4, 5), 
                    new CellRange(5, 0, 10, 5)
                }];

            CellBorder cellBorder = new CellBorder(CellBorderStyle.Dotted, new ThemableColor(ThemeColorType.Accent1, ColorShadeType.Shade3));
            CellBorders cellBorders = new CellBorders(cellBorder);

            cellSelection.SetBorders(cellBorders);

            CellBorders result = cellSelection.GetBorders();

            Assert.AreEqual(cellBorders, result);
        }

        #endregion


        #region SetCellValue Methods

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        [ExpectedException(typeof(ParseException))]
        public void SetValue()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Cells[0, 0].SetValue("=3-");
        }

        #endregion
    }
}