﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;
using System;
using System.Windows.Media;
using Telerik.Windows.Documents.Spreadsheet.Theming;

namespace Telerik.Windows.Documents.Spreadsheet.Tests
{
    [TestClass]
    public class CellStyleTests
    {
        #region Methods

        private void ApplyStyle(Workbook workbook, CellRange range, string styleName)
        {
            workbook.ActiveWorksheet.Cells[range].SetStyleName(styleName);
        }

        private void ClearCellStyle(Workbook workbook, CellRange range)
        {
            workbook.ActiveWorksheet.Cells[range].ClearStyleName();
        }

        private void AssertStyleApplied(Workbook workbook, CellRange range, string styleName)
        {
            var appliedStyleName = workbook.ActiveWorksheet.Cells[range].GetStyleName().Value;
            Assert.AreEqual(styleName, appliedStyleName);
        }

        private void AssertStyleProperties(Workbook workbook, CellRange range, params string[] styleName)
        {
            CellStyle resultStyle = CellStyle.CreateTempStyle();
            resultStyle.IncludeNumber = false;
            resultStyle.IncludeAlignment = false;
            resultStyle.IncludeFont = false;
            resultStyle.IncludeBorder = false;
            resultStyle.IncludeFill = false;

            foreach (string name in styleName)
            {
                CellStyle cellStyle = workbook.Styles[name];

                if (cellStyle.IncludeNumber)
                {
                    resultStyle.IncludeFill = true;
                    resultStyle.Format = cellStyle.Format;
                }

                if (cellStyle.IncludeAlignment)
                {
                    resultStyle.IncludeAlignment = true;
                    resultStyle.HorizontalAlignment = cellStyle.HorizontalAlignment;
                    resultStyle.VerticalAlignment = cellStyle.VerticalAlignment;
                    resultStyle.Indent = cellStyle.Indent;
                    resultStyle.IsWrapped = cellStyle.IsWrapped;
                }

                if (cellStyle.IncludeFont)
                {
                    resultStyle.IncludeFont = true;
                    resultStyle.FontFamily = cellStyle.FontFamily;
                    resultStyle.FontSize = cellStyle.FontSize;
                    resultStyle.IsBold = cellStyle.IsBold;
                    resultStyle.IsItalic = cellStyle.IsItalic;
                    resultStyle.Underline = cellStyle.Underline;
                    resultStyle.ForeColor = cellStyle.ForeColor;
                }

                if (cellStyle.IncludeBorder)
                {
                    resultStyle.IncludeBorder = true;
                    resultStyle.LeftBorder = cellStyle.LeftBorder;
                    resultStyle.TopBorder = cellStyle.TopBorder;
                    resultStyle.RightBorder = cellStyle.RightBorder;
                    resultStyle.BottomBorder = cellStyle.BottomBorder;
                    resultStyle.DiagonalDownBorder = cellStyle.DiagonalDownBorder;
                    resultStyle.DiagonalUpBorder = cellStyle.DiagonalUpBorder;
                }

                if (cellStyle.IncludeFill)
                {
                    resultStyle.IncludeFill = true;
                    resultStyle.Fill = cellStyle.Fill;
                }
            }

            CellSelection selection = workbook.ActiveWorksheet.Cells[range];

            if (resultStyle.IncludeNumber)
            {
                CellValueFormat format = selection.GetFormat().Value;
                Assert.AreEqual(resultStyle.Format, format);
            }

            if (resultStyle.IncludeAlignment)
            {
                RadHorizontalAlignment horizontalAlignment = selection.GetHorizontalAlignment().Value;
                Assert.AreEqual(resultStyle.HorizontalAlignment, horizontalAlignment);

                RadVerticalAlignment verticalAlignment = selection.GetVerticalAlignment().Value;
                Assert.AreEqual(resultStyle.VerticalAlignment, verticalAlignment);

                int indent = selection.GetIndent().Value;
                Assert.AreEqual(resultStyle.Indent, indent);

                bool isWrapped = selection.GetIsWrapped().Value;
                Assert.AreEqual(resultStyle.IsWrapped, isWrapped);
            }

            if (resultStyle.IncludeFont)
            {
                ThemableFontFamily fontFamily = selection.GetFontFamily().Value;
                Assert.AreEqual(resultStyle.FontFamily, fontFamily);

                double fontSize = selection.GetFontSize().Value;
                Assert.AreEqual(resultStyle.FontSize, fontSize);

                bool isBold = selection.GetIsBold().Value;
                Assert.AreEqual(resultStyle.IsBold, isBold);

                bool isItalic = selection.GetIsItalic().Value;
                Assert.AreEqual(resultStyle.IsItalic, isItalic);

                UnderlineType underline = selection.GetUnderline().Value;
                Assert.AreEqual(resultStyle.Underline, underline);

                ThemableColor foreColor = selection.GetForeColor().Value;
                Assert.AreEqual(resultStyle.ForeColor, foreColor);
            }

            if (resultStyle.IncludeBorder)
            {
                CellBorders borders = selection.GetBorders();
                Assert.AreEqual(resultStyle.LeftBorder, borders.Left);
                Assert.AreEqual(resultStyle.TopBorder, borders.Top);
                Assert.AreEqual(resultStyle.RightBorder, borders.Right);
                Assert.AreEqual(resultStyle.BottomBorder, borders.Bottom);
                Assert.AreEqual(resultStyle.DiagonalDownBorder, borders.DiagonalDown);
                Assert.AreEqual(resultStyle.DiagonalUpBorder, borders.DiagonalUp);
            }

            if (resultStyle.IncludeFill)
            {
                var fill = selection.GetFill().Value;
                Assert.AreEqual(resultStyle.Fill, fill);
            }
        }

        #endregion


        #region CellStyleCollection Tests

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestAddNewStyle()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();

            int styleCount = workbook.Styles.Count;

            CellStyle style1 = workbook.Styles.Add("Style1");
            styleCount++;

            Assert.AreEqual(styleCount, workbook.Styles.Count);
            Assert.IsTrue(workbook.Styles.Contains(style1));
            Assert.AreEqual(style1, workbook.Styles.GetByName(style1.Name));
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        [ExpectedException(typeof(InvalidOperationException))]
        public void TestAddStyleWithExistingNameThrowsException()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();

            CellStyle style1 = workbook.Styles.Add("Style1");
            CellStyle style2 = workbook.Styles.Add("Style1");
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestRemoveStyle()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();

            int styleCount = workbook.Styles.Count;

            CellStyle style1 = workbook.Styles.Add("Style1");

            bool result = workbook.Styles.Remove(style1);

            Assert.IsTrue(result);
            Assert.AreEqual(styleCount, workbook.Styles.Count);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestRemoveStyleByName()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();

            int styleCount = workbook.Styles.Count;

            CellStyle style1 = workbook.Styles.Add("Style1");

            bool result = workbook.Styles.Remove(style1.Name);
            
            Assert.IsTrue(result);
            Assert.AreEqual(styleCount, workbook.Styles.Count);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestRemoveNonExistingStyle()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();

            int styleCount = workbook.Styles.Count;

            CellStyle style1 = workbook.Styles.Add("Style1");
            workbook.Styles.Remove("Style1");
            Assert.AreEqual(styleCount, workbook.Styles.Count);

            bool result = workbook.Styles.Remove(style1);
            Assert.IsFalse(result);
            Assert.AreEqual(styleCount, workbook.Styles.Count);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestRemoveNonExistingStyleByName()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();

            int styleCount = workbook.Styles.Count;

            CellStyle style1 = workbook.Styles.Add("Style1");
            styleCount++;

            bool result = workbook.Styles.Remove("NonExistingStyleName");

            Assert.IsFalse(result);
            Assert.AreEqual(styleCount, workbook.Styles.Count);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestRemoveOfStyleThatCannotBeRemoved()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();

            int styleCount = workbook.Styles.Count;

            bool result = workbook.Styles.Remove(SpreadsheetDefaultValues.DefaultStyleName);

            Assert.IsFalse(result);
            Assert.AreEqual(styleCount, workbook.Styles.Count);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestAddRemoveStylesTriggersCollectionChanged()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();

            int styleCount = workbook.Styles.Count;

            int collectionChangedCount = 0;
            workbook.Styles.Changed += 
                (s, a) => 
                {
                    collectionChangedCount++;
                };

            for (int i = 0; i < 10; i++)
            {
                workbook.Styles.Add("Style" + i);
            }

            for (int i = 0; i < 10; i++)
            {
                workbook.Styles.Remove("Style" + i);
            }

            Assert.AreEqual(styleCount, workbook.Styles.Count);
            Assert.AreEqual(20, collectionChangedCount);
        }

        #endregion


        #region Style Property Evaluation Tests

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestCellPropertyDefaultValue()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            var selection = worksheet.Cells[1, 1];
            
            Assert.AreEqual(ThemeColorType.Text1, selection.GetForeColor().Value.ThemeColorType);
            Assert.AreEqual(Colors.Black, selection.GetForeColor().Value.GetActualValue(worksheet.Workbook.Theme));
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestCellPropertyEvaluation()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.ActiveWorksheet;
            
            var selection = worksheet.Cells[1, 1];
            Assert.AreEqual(ThemeColorType.Text1, selection.GetForeColor().Value.ThemeColorType);
            Assert.AreEqual(Colors.Black, selection.GetForeColor().Value.GetActualValue(worksheet.Workbook.Theme));

            CellStyle style1 = workbook.Styles.Add("style1");

            Assert.IsFalse(style1.IncludeFont);

            style1.ForeColor = (ThemableColor)Colors.Blue;

            Assert.IsTrue(style1.IncludeFont);

            Assert.AreEqual(Colors.Blue, style1.ForeColor.LocalValue);

            selection.SetStyleName("style1");

            Assert.AreEqual(style1.ForeColor, selection.GetForeColor().Value);

            selection.SetForeColor((ThemableColor)Colors.Yellow);
            
            Assert.AreEqual(Colors.Yellow, selection.GetForeColor().Value.LocalValue);

            selection.ClearForeColor();

            Assert.AreEqual(style1.ForeColor, selection.GetForeColor().Value);

            workbook.Styles.Remove(style1);

            Assert.AreEqual(ThemeColorType.Text1, selection.GetForeColor().Value.ThemeColorType);
            Assert.AreEqual(Colors.Black, selection.GetForeColor().Value.GetActualValue(worksheet.Workbook.Theme));
        }

        #endregion


        #region Style Undo/Redo Tests

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestAddStyleUndoRedo()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();

            int styleCount = workbook.Styles.Count;
            int stylesToAddCount = 10;

            for (int i = 0; i < stylesToAddCount; i++)
            {
                workbook.Styles.Add("Style" + i);
            }

            Assert.AreEqual(styleCount + stylesToAddCount, workbook.Styles.Count);

            while (workbook.History.Undo())
            {
            }

            Assert.AreEqual(styleCount, workbook.Styles.Count);

            while (workbook.History.Redo())
            {
            }

            Assert.AreEqual(styleCount + stylesToAddCount, workbook.Styles.Count);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestRemoveStyleUndoRedo()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();

            int styleCount = workbook.Styles.Count;
            int stylesToAddCount = 10;

            workbook.History.IsEnabled = false;
            for (int i = 0; i < stylesToAddCount; i++)
            {
                workbook.Styles.Add("Style" + i);
            }
            workbook.History.IsEnabled = true;

            Assert.IsTrue(!workbook.History.CanUndo && !workbook.History.CanRedo);

            Assert.AreEqual(styleCount + stylesToAddCount, workbook.Styles.Count);

            for (int i = 0; i < stylesToAddCount; i++)
            {
                workbook.Styles.Remove("Style" + i);
            }

            Assert.AreEqual(styleCount, workbook.Styles.Count);

            while (workbook.History.Undo())
            {
            }

            Assert.AreEqual(styleCount + stylesToAddCount, workbook.Styles.Count);

            while (workbook.History.Redo())
            {
            }

            Assert.AreEqual(styleCount, workbook.Styles.Count);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void SetBoldStylePropertyUndoRedo()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            CellStyle style = workbook.Styles.Add("Style1");

            style.IsBold = true;
            Assert.IsTrue(style.IsBold);

            workbook.History.Undo();
            Assert.IsFalse(style.IsBold);

            workbook.History.Redo();
            Assert.IsTrue(style.IsBold);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void SetFontFamilyStylePropertyUndoRedo()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            CellStyle style = workbook.Styles.Add("Style1");

            var defaultFontFamily = style.FontFamily;
            var fontFamily = new ThemableFontFamily("Arial");

            style.FontFamily = fontFamily;
            Assert.AreEqual(fontFamily, style.FontFamily);

            workbook.History.Undo();
            Assert.AreEqual(defaultFontFamily, style.FontFamily);

            workbook.History.Redo();
            Assert.AreEqual(fontFamily, style.FontFamily);
        }

        #endregion


        #region Tests

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestCopyPropertiesFromTempStyle()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();

            CellStyle style1 = CellStyle.CreateTempStyle();
            style1.LeftBorder = new CellBorder(CellBorderStyle.DashDot, new ThemableColor(Colors.Blue));
            style1.TopBorder = new CellBorder(CellBorderStyle.DashDot, new ThemableColor(Colors.Blue));
            style1.RightBorder = new CellBorder(CellBorderStyle.DashDot, new ThemableColor(Colors.Blue));
            style1.BottomBorder = new CellBorder(CellBorderStyle.DashDot, new ThemableColor(Colors.Blue));
            style1.DiagonalUpBorder = new CellBorder(CellBorderStyle.DashDot, new ThemableColor(Colors.Blue));
            style1.DiagonalDownBorder = new CellBorder(CellBorderStyle.DashDot, new ThemableColor(Colors.Blue));
            style1.Fill = PatternFill.CreateSolidFill(Colors.Blue);
            style1.FontFamily = new ThemableFontFamily("Arial");
            style1.FontSize = 13;
            style1.IsBold = true;
            style1.IsItalic = true;
            style1.Underline = UnderlineType.Single;
            style1.ForeColor = (ThemableColor)Colors.Green;
            style1.HorizontalAlignment = RadHorizontalAlignment.Right;
            style1.VerticalAlignment = RadVerticalAlignment.Center;
            style1.Indent = 5;
            style1.IsWrapped = true;

            CellStyle style2 = workbook.Styles.Add("Style2");

            Assert.IsFalse(style1.HasSameProperties(style2));

            style2.CopyPropertiesFrom(style1);

            Assert.IsTrue(style1.HasSameProperties(style2));
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void TestSetStylePropertyOfRemovedStyleDoesNotHaveUndo()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            CellStyle style1 = workbook.Styles.Add("Style1");
            workbook.History.Clear();

            Assert.IsFalse(workbook.History.CanUndo);
            style1.FontSize++;
            Assert.IsTrue(workbook.History.CanUndo);
            
            workbook.Styles.Remove(style1);
            workbook.History.Clear();

            Assert.IsFalse(workbook.History.CanUndo);
            style1.FontSize++;
            Assert.IsFalse(workbook.History.CanUndo);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void ApplyCellStyleTest()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();

            foreach (var style in workbook.Styles)
            {
                var styleName = style.Name;

                CellRange range = new CellRange(0, 0, 10, 10);

                this.ClearCellStyle(workbook, range);
                this.ApplyStyle(workbook, range, styleName);
                this.AssertStyleApplied(workbook, range, styleName);
                this.AssertStyleProperties(workbook, range, styleName);
            }
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void ChangeCellStyleTest()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();

            string styleName = "Bad";
            CellRange range = new CellRange(0, 0, 10, 10);

            this.ApplyStyle(workbook, range, styleName);
            this.AssertStyleApplied(workbook, range, styleName);
            this.AssertStyleProperties(workbook, range, styleName);

            CellStyle style = workbook.Styles[styleName];
            style.IncludeFill = false;

            this.AssertStyleProperties(workbook, range, styleName);

            style.IncludeFill = true;
            style.Fill = new PatternFill(
                PatternType.Gray50Percent,
                new ThemableColor(ThemeColorType.Accent2, ColorShadeType.Shade3),
                new ThemableColor(ThemeColorType.Accent1, ColorShadeType.Shade5));

            this.AssertStyleProperties(workbook, range, styleName);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void ApplyTwoStylesTest()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();

            string styleName = "Bad";
            CellRange range = new CellRange(0, 0, 10, 10);

            this.ApplyStyle(workbook, range, styleName);
            this.AssertStyleApplied(workbook, range, styleName);
            this.AssertStyleProperties(workbook, range, styleName);

            string styleName1 = "Explanatory Text";
            CellRange range1 = new CellRange(5, 5, 15, 15);

            this.ApplyStyle(workbook, range1, styleName1);
            this.AssertStyleApplied(workbook, range1, styleName1);

            CellRange firstSingleStyleRange = new CellRange(0, 0, 4, 4);
            this.AssertStyleProperties(workbook, firstSingleStyleRange, styleName);

            CellRange secondSingleStyleRange = new CellRange(11, 11, 15, 15);
            this.AssertStyleProperties(workbook, secondSingleStyleRange, styleName1);

            CellRange multipleStylesRange = new CellRange(5, 5, 10, 10);
            this.AssertStyleProperties(workbook, multipleStylesRange, styleName, styleName1);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void ApplyThreeStylesTest()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();

            string styleName = "Bad";
            string styleName1 = "Linked Cell";
            string styleName2 = "Explanatory Text";
            string styleName3 = "Currency";

            CellRange range = new CellRange(0, 0, 10, 10);

            this.ApplyStyle(workbook, range, styleName);
            this.AssertStyleApplied(workbook, range, styleName);

            this.ApplyStyle(workbook, range, styleName1);
            this.AssertStyleApplied(workbook, range, styleName1);

            this.ApplyStyle(workbook, range, styleName2);
            this.AssertStyleApplied(workbook, range, styleName2);

            this.ApplyStyle(workbook, range, styleName3);
            this.AssertStyleApplied(workbook, range, styleName3);

            this.AssertStyleProperties(workbook, range, styleName, styleName1, styleName2, styleName3);
        }

        [TestMethod]
        [TestCategory("PropertySystemTests")]
        public void CellStyleIsIncludedPropertyUndoRedoTest()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();

            string styleName = "Bad";
            CellRange range = new CellRange(0, 0, 10, 10);

            this.ApplyStyle(workbook, range, styleName);
            this.AssertStyleApplied(workbook, range, styleName);
            this.AssertStyleProperties(workbook, range, styleName);

            CellStyle style = workbook.Styles[styleName];
            style.IncludeFill = false;

            this.AssertStyleProperties(workbook, range, styleName);

            style.Fill = new PatternFill(
                PatternType.Gray50Percent,
                new ThemableColor(ThemeColorType.Accent2, ColorShadeType.Shade3),
                new ThemableColor(ThemeColorType.Accent1, ColorShadeType.Shade5));

            this.AssertStyleProperties(workbook, range, styleName);

            workbook.History.Undo();

            Assert.IsFalse(style.IncludeFill);

            workbook.History.Undo();

            Assert.IsTrue(style.IncludeFill);

            workbook.History.Redo();

            Assert.IsFalse(style.IncludeFill);

            workbook.History.Redo();

            Assert.IsTrue(style.IncludeFill);
        }

        #endregion
    }
}