﻿using System;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests
{
    [TestClass]
    public class CellsTests
    {

        #region Fields

        private Worksheet worksheet;

        #endregion


        #region Initialize Methods

        [TestInitialize]
        public void TestInitializers()
        {
            this.worksheet = TestHelper.CreateWorksheet();
        }

        #endregion


        #region Helper Methods

        public bool TryFireCellRangeInsertedOrRemoved(Worksheet worksheet, Action[] actions, out int counter)
        {
            bool isEventFired = false;
            int innerCounter = 0;

            worksheet.Cells.CellRangeInsertedOrRemoved += (s, e) =>
            {
                isEventFired = !isEventFired;
                innerCounter++;
            };

            foreach (Action action in actions)
            {
                action();
            }

            counter = innerCounter;

            return isEventFired;
        }

        #endregion


        #region Tests

        [TestMethod]
        public void OnCellRangeInsertedOrRemoved_WhenCellsAreInserted_IsFired()
        {
            Action insertCells = new Action(() => this.worksheet.Cells.Insert(0, 0, 1, InsertShiftType.Right));

            int actualEventFiredCount;
            bool isEventFired = this.TryFireCellRangeInsertedOrRemoved(this.worksheet, new Action[] { insertCells }, out actualEventFiredCount);

            int expectedEventFiredCount = 1;

            Assert.AreEqual(expectedEventFiredCount, actualEventFiredCount);
            Assert.IsTrue(isEventFired);
        }

        [TestMethod]
        public void OnCellRangeInsertedOrRemoved_WhenCellsAreRemoved_IsFired()
        {
            Action removeCells = new Action(() => this.worksheet.Cells.Remove(0, 0, 1, RemoveShiftType.Up));

            int actualEventFiredCount;
            bool isEventFired = this.TryFireCellRangeInsertedOrRemoved(this.worksheet, new Action[] { removeCells }, out actualEventFiredCount);

            int expectedEventFiredCount = 1;

            Assert.AreEqual(expectedEventFiredCount, actualEventFiredCount);
            Assert.IsTrue(isEventFired);
        }

        [TestMethod]
        public void OnCellRangeInsertedOrRemoved_WhenUndoIsInvoked_IsFired()
        {
            this.worksheet.Cells.Insert(0, 0, 1, InsertShiftType.Down);
            Action invokeUndo = new Action(() => this.worksheet.Workbook.History.Undo());

            int actualEventFiredCount;
            bool isEventFired = this.TryFireCellRangeInsertedOrRemoved(this.worksheet, new Action[] { invokeUndo }, out actualEventFiredCount);

            int expectedEventFiredCount = 1;

            Assert.AreEqual(expectedEventFiredCount, actualEventFiredCount);
            Assert.IsTrue(isEventFired);
        }

        [TestMethod]
        public void OnCellRangeInsertedOrRemoved_WhenRedoIsInvoked_IsFired()
        {
            this.worksheet.Cells.Insert(0, 0, 1, InsertShiftType.Down);
            this.worksheet.Workbook.History.Undo();
            Action invokeRedo = new Action(() => this.worksheet.Workbook.History.Redo());

            int actualEventFiredCount;
            bool isEventFired = this.TryFireCellRangeInsertedOrRemoved(this.worksheet, new Action[] { invokeRedo }, out actualEventFiredCount);

            int expectedEventFiredCount = 1;

            Assert.AreEqual(expectedEventFiredCount, actualEventFiredCount);
            Assert.IsTrue(isEventFired);
        }

        [TestMethod]
        public void OnCellRangeInsertedOrRemoved_WhenAddRemoveUndoRedoAreInvoked_IsFiredProperTimes()
        {
            int firedTimesCounter;
            Action[] actions = new Action[]
            {
                () => this.worksheet.Cells.Insert(0, 0, 1, InsertShiftType.Down),
                () => this.worksheet.Workbook.History.Undo(),
                () => this.worksheet.Workbook.History.Redo(),
                () => this.worksheet.Cells.Remove(0, 0, 1, RemoveShiftType.Up),
            };

            this.TryFireCellRangeInsertedOrRemoved(this.worksheet, actions, out firedTimesCounter);

            int expectedFiredTimesCounter = 4;

            Assert.AreEqual(expectedFiredTimesCounter, firedTimesCounter);
        }

        #endregion

    }
}