﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Protection;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests
{
    [TestClass]
    public class HistoryTests
    {
        #region Fields

        private Workbook workbook;

        #endregion


        #region Initialize and Cleanup Methods

        [TestInitialize]
        public void TestInitializers()
        {
            this.workbook = TestHelper.CreateNonEmptyWorkbook();
        }

        [TestCleanup]
        public void TestCleanup()
        {
            this.workbook = null;
        }

        #endregion


        #region Helper Methods

        private void AssertHistoryIsEmpty()
        {
            Assert.IsFalse(this.workbook.History.CanUndo);
            Assert.IsFalse(this.workbook.History.CanRedo);
            Assert.IsFalse(this.workbook.History.IsInUndoGroup);
        }

        private void PerformUndoableAction(int count)
        {
            Worksheet worksheet = this.workbook.Worksheets[0];

            for (int i = 0; i < count; i++)
            {
                worksheet.Cells[0, i].SetValue("test" + count);
            }
        }

        private void PerformUndoableAction(int count, Worksheet worksheet)
        {
            for (int i = 0; i < count; i++)
            {
                worksheet.Cells[0, i].SetValue("test" + count);
            }
        }

        #endregion


        #region History Test Methods

        [TestMethod]
        public void TestIsHistoryEmptyWhenWorkbookIsCreated()
        {
            this.AssertHistoryIsEmpty();
        }

        [TestMethod]
        public void TestIsHistoryDisabledWhenWorkbookIsCreated()
        {
            Workbook workbook = new Workbook();
            Assert.IsFalse(workbook.History.IsEnabled);
        }

        [TestMethod]
        public void TestHistoryUndoRedo()
        {
            this.PerformUndoableAction(1);

            Assert.IsTrue(this.workbook.History.CanUndo);
            Assert.IsFalse(this.workbook.History.CanRedo);

            this.workbook.History.Undo();
            Assert.IsFalse(this.workbook.History.CanUndo);
            Assert.IsTrue(this.workbook.History.CanRedo);

            this.workbook.History.Redo();
            Assert.IsTrue(this.workbook.History.CanUndo);
            Assert.IsFalse(this.workbook.History.CanRedo);

            this.PerformUndoableAction(1);

            this.workbook.History.Undo();
            Assert.IsTrue(this.workbook.History.CanUndo);
            Assert.IsTrue(this.workbook.History.CanRedo);

            this.workbook.History.Undo();
            Assert.IsFalse(this.workbook.History.CanUndo);
            Assert.IsTrue(this.workbook.History.CanRedo);
        }
     
        [TestMethod]
        public void TestHistoryClear()
        {
            this.PerformUndoableAction(1);

            Assert.IsTrue(this.workbook.History.CanUndo);
            Assert.IsFalse(this.workbook.History.CanRedo);

            this.workbook.History.Clear();

            this.AssertHistoryIsEmpty();
        }

        [TestMethod]
        public void TestHistoryDepthWhenIsDecreased()
        {
            this.PerformUndoableAction(2);

            Assert.IsTrue(this.workbook.History.CanUndo);
            Assert.IsFalse(this.workbook.History.CanRedo);

            this.workbook.History.Undo();
            Assert.IsTrue(this.workbook.History.CanUndo);
            Assert.IsTrue(this.workbook.History.CanRedo);

            this.workbook.History.Undo();
            Assert.IsFalse(this.workbook.History.CanUndo);

            this.workbook.History.Redo();
            this.workbook.History.Redo();

            this.workbook.History.Depth = 1;
            Assert.IsFalse(this.workbook.History.CanRedo);
            Assert.IsTrue(this.workbook.History.CanUndo);

            this.workbook.History.Undo();
            Assert.IsFalse(this.workbook.History.CanUndo);
            Assert.IsTrue(this.workbook.History.CanRedo);
        }

        [TestMethod]
        public void TestCanRedoWhenDepthIsDecreased()
        {
            this.PerformUndoableAction(1);

            this.workbook.History.Undo();
            Assert.IsTrue(this.workbook.History.CanRedo);

            this.workbook.History.Depth = 1;
            Assert.IsFalse(this.workbook.History.CanRedo);
        }

        [TestMethod]
        public void TestCanRedoWhenDepthIsIncreased()
        {
            this.workbook.History.Depth = 1;
            Assert.IsFalse(this.workbook.History.CanRedo);

            this.PerformUndoableAction(1);

            this.workbook.History.Undo();
            Assert.IsTrue(this.workbook.History.CanRedo);

            this.workbook.History.Depth = 1000;
            Assert.IsFalse(this.workbook.History.CanRedo);
        }

        [TestMethod]
        public void TestActionsAreSavedWhenHistoryIsEnabled()
        {
            this.PerformUndoableAction(1);

            Assert.IsTrue(this.workbook.History.CanUndo);
            Assert.IsFalse(this.workbook.History.CanRedo);

            this.workbook.History.Undo();
            Assert.IsFalse(this.workbook.History.CanUndo);
            Assert.IsTrue(this.workbook.History.CanRedo);
        }

        [TestMethod]
        public void TestActionsAreSavedWhenHistoryIsNotEnabled()
        {
            this.workbook.History.IsEnabled = false;

            this.PerformUndoableAction(2);

            this.AssertHistoryIsEmpty();
        }

        #endregion


        #region Protection History Test Methods

        [TestMethod]
        public void ActionInWorksheet_WhenWorkbookIsProtected_CanBeUndone()
        {
            this.PerformUndoableAction(1);

            this.workbook.Protect(string.Empty);

            Assert.IsTrue(this.workbook.History.CanUndo);
        }

        [TestMethod]
        public void ActionInWorksheet_WhenIsProtected_CannotBeUndone()
        {
            this.PerformUndoableAction(1);

            this.workbook.Worksheets[0].Protect(string.Empty, WorksheetProtectionOptions.Default);

            Assert.IsFalse(this.workbook.History.CanUndo);
            Assert.IsFalse(this.workbook.History.CanRedo);
        }

        [TestMethod]
        public void ActionInWorksheet_WhenIsProtectedAndUnprotected_CannotBeUndone()
        {
            this.PerformUndoableAction(1);

            this.workbook.Worksheets[0].Protect(string.Empty, WorksheetProtectionOptions.Default);
            this.workbook.Worksheets[0].Unprotect(string.Empty);

            Assert.IsFalse(this.workbook.History.CanUndo);
            Assert.IsFalse(this.workbook.History.CanRedo);
        }

        [TestMethod]
        public void ActionInWorksheet_WhenOtherIsInserted_CannotBeUndone()
        {
            this.PerformUndoableAction(1);

            Worksheet unprotectedWorksheet = this.workbook.Worksheets.Add();
            this.workbook.ActiveWorksheet = unprotectedWorksheet;
            this.PerformUndoableAction(1, unprotectedWorksheet);

            Worksheet protectedWorksheet = this.workbook.Worksheets[0];
            protectedWorksheet.Protect(string.Empty, WorksheetProtectionOptions.Default);

            Assert.IsTrue(this.workbook.History.CanUndo);
            Assert.IsFalse(this.workbook.History.CanRedo);

            this.workbook.History.Undo();

            Assert.IsFalse(this.workbook.History.CanUndo);
            Assert.IsTrue(this.workbook.History.CanRedo);

            var actualProtectedCellValue = protectedWorksheet.Cells[0, 0].GetValue().Value.RawValue;
            string expectedProtectedCellValue = "test1";

            Assert.AreEqual(expectedProtectedCellValue, actualProtectedCellValue);
        }

        [TestMethod]
        public void ActionsInUnlockedCells_WhenWorksheetIsProtected_CannotBeUndone()
        {
            Worksheet worksheet = this.workbook.ActiveWorksheet;
            worksheet.Cells[0, 0, 5, 5].SetIsLocked(false);

            this.PerformUndoableAction(10);
            worksheet.Protect(string.Empty, WorksheetProtectionOptions.Default);

            Assert.IsFalse(this.workbook.History.CanUndo);
            Assert.IsFalse(this.workbook.History.CanRedo);
        }

        [TestMethod]
        public void ActionsInUnlockedCells_WhenWorksheetIsProtectedAndUnprotected_CannotBeUndone()
        {
            Worksheet worksheet = this.workbook.ActiveWorksheet;
            worksheet.Cells[0, 0, 5, 5].SetIsLocked(false);

            this.PerformUndoableAction(10);
            worksheet.Protect(string.Empty, WorksheetProtectionOptions.Default);
            worksheet.Unprotect(string.Empty);

            Assert.IsFalse(this.workbook.History.CanUndo);
            Assert.IsFalse(this.workbook.History.CanRedo);
        }

        [TestMethod]
        public void ActionsInUnlockedCells_WhenWorksheetIsProtectedAndUnprotected_CanBeUndone()
        {
            Worksheet worksheet = this.workbook.ActiveWorksheet;
            worksheet.Cells[0, 0, 5, 5].SetIsLocked(false);

            this.PerformUndoableAction(10);

            worksheet.Protect(string.Empty, WorksheetProtectionOptions.Default);
            worksheet.Unprotect(string.Empty);

            this.PerformUndoableAction(1); // Edit the unlocked cells.

            Assert.IsTrue(this.workbook.History.CanUndo);
            Assert.IsFalse(this.workbook.History.CanRedo);

            this.workbook.History.Undo();

            Assert.IsFalse(this.workbook.History.CanUndo);
            Assert.IsTrue(this.workbook.History.CanRedo);
        }

        [TestMethod]
        public void ActionsInUnlockedCellsAndUnlockedWorksheet_WhenFirstWorksheetIsProtectedAndUnprotected_CanBeUndone()
        {
            Worksheet unprotectedWorksheet = this.workbook.Worksheets.Add();
            this.workbook.ActiveWorksheet = unprotectedWorksheet;
            
            this.PerformUndoableAction(1, unprotectedWorksheet);

            Worksheet protectedWorksheet = this.workbook.Worksheets[0];
            this.workbook.ActiveWorksheet = protectedWorksheet;

            protectedWorksheet.Cells[0, 0, 5, 5].SetIsLocked(false);
            this.PerformUndoableAction(10);

            protectedWorksheet.Protect(string.Empty, WorksheetProtectionOptions.Default);
            protectedWorksheet.Unprotect(string.Empty);

            Assert.IsTrue(this.workbook.History.CanUndo);
            Assert.IsFalse(this.workbook.History.CanRedo);

            this.workbook.History.Undo();
            
            Assert.IsFalse(this.workbook.History.CanUndo);

            this.PerformUndoableAction(1); // Edit the unlocked cells.

            Assert.IsTrue(this.workbook.History.CanUndo);
        }

        #endregion
    }
}