﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Media;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests
{
    [TestClass]
    public class InsertDeleteCellRangesTests
    {
        #region InsertRemoveCells Test Helper Methods

        private void AssertCellsInsertion(CellRange valuesRange, CellRange insertRange, InsertShiftType insertShiftType,
            CellRange resultRange, bool insertRangeIntersectsValuesRange)
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            Cells cells = worksheet.Cells;

            Func<int, int, string> rowsColumnsFormula = (row, col) =>
            {
                return (row * (valuesRange.ColumnCount - 1) + col).ToString();
            };

            this.FillCellsInRangeWithFormula(cells, valuesRange, rowsColumnsFormula);

            cells.Insert(insertRange, insertShiftType);
            AssertInsertCellsCorectness(cells, resultRange, insertRange, insertRangeIntersectsValuesRange, insertShiftType, rowsColumnsFormula);
        }

        private void AssertValueAfterMergeAndInsertRemove(CellRange rangeToMerge, CellIndex cellToSetValue, CellIndex movedCellToTestValue, Action<Cells> insertRemoveAction)
        {
            Cells cells = TestHelper.CreateWorksheet().Cells;
            string value = "test";

            cells[rangeToMerge].Merge();
            cells[cellToSetValue].SetValue(value);
            insertRemoveAction(cells);

            string actualValue = cells[movedCellToTestValue].GetValue().Value.RawValue;
            Assert.AreEqual(value, actualValue);
        }

        private void AssertInsertCellsCorectness(Cells cells, CellRange resultRange, CellRange insertRange, bool insertRangeIntersectsValuesRange,
            InsertShiftType shiftType, Func<int, int, string> rowsColumnsFormula)
        {
            if (insertRangeIntersectsValuesRange)
            {
                int valuesRangeRowCounter = 0;
                int valuesRangeColumnCounter = 0;

                for (int r = resultRange.FromIndex.RowIndex; r <= resultRange.ToIndex.RowIndex; r++)
                {
                    if (shiftType == InsertShiftType.Down &&
                        insertRange.Contains(r, insertRange.FromIndex.ColumnIndex))
                    {
                        continue;
                    }

                    valuesRangeColumnCounter = 0;
                    for (int c = resultRange.FromIndex.ColumnIndex; c <= resultRange.ToIndex.ColumnIndex; c++)
                    {
                        if (shiftType == InsertShiftType.Right &&
                            insertRange.Contains(insertRange.FromIndex.RowIndex, c))
                        {
                            continue;
                        }

                        string expected = rowsColumnsFormula(valuesRangeRowCounter, valuesRangeColumnCounter);
                        string actual = cells[r, c].GetValue().Value.RawValue;
                        Assert.AreEqual(expected, actual);

                        valuesRangeColumnCounter++;
                    }

                    valuesRangeRowCounter++;
                }
            }
            else
            {
                for (int r = resultRange.FromIndex.RowIndex; r <= resultRange.ToIndex.RowIndex; r++)
                {
                    for (int c = resultRange.FromIndex.ColumnIndex; c <= resultRange.ToIndex.ColumnIndex; c++)
                    {
                        int row = r;
                        int column = c;
                        if (shiftType == InsertShiftType.Right)
                        {
                            column -= insertRange.ColumnCount;
                        }
                        else
                        {
                            row -= insertRange.RowCount;
                        }

                        string expected = rowsColumnsFormula(row, column);
                        string actual = cells[r, c].GetValue().Value.RawValue;
                        Assert.AreEqual(expected, actual);
                    }
                }
            }
        }

        private void AssertRemoveCells(CellRange valuesRange, CellRange removeRange, RemoveShiftType removeShiftType, CellRange resultRange,
            bool removeRangeIntersectsValuesRange)
        {
            Cells cells = TestHelper.CreateWorksheet().Cells;

            Func<int, int, string> rowsColumnsFormula = (row, col) =>
            {
                return (row * (valuesRange.ColumnCount - 1) + col).ToString();
            };

            this.FillCellsInRangeWithFormula(cells, valuesRange, rowsColumnsFormula);
            cells.Remove(removeRange, removeShiftType);
            AssertRemoveCellsCorectness(cells, resultRange, removeRange, removeRangeIntersectsValuesRange, removeShiftType, rowsColumnsFormula);
        }

        private void AssertRemoveCellsCorectness(Cells cells, CellRange resultRange, CellRange removeRange, bool removeRangeIntersectsValuesRange, RemoveShiftType removeShiftType,
            Func<int, int, string> rowsColumnsFormula)
        {
            if (removeRangeIntersectsValuesRange)
            {
                int valuesRangeRowCounter = 0;
                int valuesRangeColumnCounter = 0;

                for (int r = resultRange.FromIndex.RowIndex; r <= resultRange.ToIndex.RowIndex; r++)
                {
                    if (removeShiftType == RemoveShiftType.Up &&
                        removeRange.Contains(r, removeRange.FromIndex.ColumnIndex))
                    {
                        valuesRangeRowCounter++;
                    }

                    valuesRangeColumnCounter = 0;

                    for (int c = resultRange.FromIndex.ColumnIndex; c <= resultRange.ToIndex.ColumnIndex; c++)
                    {
                        if (removeShiftType == RemoveShiftType.Left &&
                            removeRange.Contains(removeRange.FromIndex.RowIndex, c))
                        {
                            valuesRangeColumnCounter++;
                        }

                        string expected = rowsColumnsFormula(valuesRangeRowCounter, valuesRangeColumnCounter);
                        string actual = cells[r, c].GetValue().Value.RawValue;
                        Assert.AreEqual(expected, actual);

                        valuesRangeColumnCounter++;
                    }

                    valuesRangeRowCounter++;
                }
            }
            else
            {
                for (int r = resultRange.FromIndex.RowIndex; r <= resultRange.ToIndex.RowIndex; r++)
                {
                    for (int c = resultRange.FromIndex.ColumnIndex; c <= resultRange.ToIndex.ColumnIndex; c++)
                    {
                        int row = r;
                        int column = c;
                        if (removeShiftType == RemoveShiftType.Left)
                        {
                            column += removeRange.ColumnCount;
                        }
                        else
                        {
                            row += removeRange.RowCount;
                        }

                        string expected = rowsColumnsFormula(row, column);
                        string actual = cells[r, c].GetValue().Value.RawValue;
                        Assert.AreEqual(expected, actual);
                    }
                }
            }
        }

        private void FillCellsInRangeWithFormula(Cells cells, CellRange valuesRange, Func<int, int, string> rowsColumnsFormula)
        {
            for (int r = valuesRange.FromIndex.RowIndex; r <= valuesRange.ToIndex.RowIndex; r++)
            {
                for (int c = valuesRange.FromIndex.ColumnIndex; c <= valuesRange.ToIndex.ColumnIndex; c++)
                {
                    cells[r, c].SetValue(rowsColumnsFormula(r, c));
                }
            }
        }

        private Cells GenerateMergedRangesTestScenario()
        {
            Cells cells = TestHelper.CreateWorksheet().Cells;

            CellRange valuesRange = new CellRange(0, 0, 10, 10);

            this.FillCellsInRangeWithFormula(cells, valuesRange, (row, column) =>
            {
                return (row * (valuesRange.ColumnCount - 1) + column).ToString();
            });

            cells[2, 2, 4, 5].Merge();
            cells[6, 3, 8, 7].Merge();
            cells[3, 6, 5, 9].Merge();

            return cells;
        }

        private bool CanInsertInMergeRangesTestScenario(int fromIndexRow, int fromIndexColumn, int toIndexRow, int toIndexColumn, ShiftType shiftType)
        {
            CellSelection selection = this.GenerateMergedRangesTestScenario().GetCellSelection(fromIndexRow, fromIndexColumn, toIndexRow, toIndexColumn);
            bool canInsertRemove = selection.CanInsertOrRemove(selection.CellRanges.First(), shiftType);

            return canInsertRemove;
        }

        private Cells UnmergeWhenInsertingCellsInMergedRangesTestScenario(IEnumerable<CellRange> rangesToInsert, InsertShiftType shiftType)
        {
            Cells cells = this.GenerateMergedRangesTestScenario();
            cells[rangesToInsert].Insert(shiftType);

            return cells;
        }

        private Cells UnmergeWhenRemovingCellsInMergedRangesTestScenario(IEnumerable<CellRange> rangesToRemove, RemoveShiftType shiftType)
        {
            Cells cells = this.GenerateMergedRangesTestScenario();
            cells[rangesToRemove].Remove(shiftType);

            return cells;
        }

        private void AssertMergedCells(Cells cells, IEnumerable<CellIndex> indexes)
        {
            foreach (var index in indexes)
            {
                Assert.IsTrue(cells.GetIsMerged(index));
            }
        }

        private void AssertNotMergedCells(Cells cells, IEnumerable<CellIndex> indexes)
        {
            foreach (var index in indexes)
            {
                Assert.IsFalse(cells.GetIsMerged(index));
            }
        }

        //private void InsertWithValuesOnTheEndScenario(CellRange valuesRange, CellRange insertRange, InsertShiftType shiftType, bool expectedSuccess)
        //{
        //    RadSpreadsheet radSpreadsheet = new RadSpreadsheet();
        //    Cells cells = radSpreadsheet.Workbook.ActiveWorksheet.Cells;

        //    Func<int, int, string> rowsColumnsFormula = (row, col) =>
        //    {
        //        return "test";
        //    };

        //    this.FillCellsInRangeWithFormula(cells, valuesRange, rowsColumnsFormula);

        //    bool success = cells.Insert(insertRange, shiftType);

        //    Assert.AreEqual(expectedSuccess, success);

        //    if (!expectedSuccess)
        //    {
        //        var value = cells[valuesRange].GetValue();
        //        Assert.IsFalse(value.IsIndeterminate);
        //    }
        //}

        #endregion


        #region Insert or remove cells that will affect merged ranges

        [TestMethod]
        public void InsertCellLeftOfMergedRange()
        {
            this.AssertValueAfterMergeAndInsertRemove(new CellRange(0, 5, 5, 10), new CellIndex(0, 5), new CellIndex(0, 6), (cells) =>
                {
                    cells[0, 0].Insert(InsertShiftType.Right);
                });
        }

        [TestMethod]
        public void InsertCellAboveMergedRange()
        {
            this.AssertValueAfterMergeAndInsertRemove(new CellRange(5, 0, 10, 5), new CellIndex(5, 0), new CellIndex(6, 0), (cells) =>
                {
                    cells[0, 0].Insert(InsertShiftType.Down);
                });
        }

        [TestMethod]
        public void RemoveCellLeftOfMergedRange()
        {
            this.AssertValueAfterMergeAndInsertRemove(new CellRange(0, 5, 5, 10), new CellIndex(0, 5), new CellIndex(0, 4), (cells) =>
                {
                    cells[0, 0].Remove(RemoveShiftType.Left);
                });
        }

        [TestMethod]
        public void RemoveCellAboveMergedRange()
        {
            this.AssertValueAfterMergeAndInsertRemove(new CellRange(5, 0, 10, 5), new CellIndex(5, 0), new CellIndex(4, 0), (cells) =>
                {
                    cells[0, 0].Remove(RemoveShiftType.Up);
                });
        }

        [TestMethod]
        public void RemoveCellsAboveMergedRangesAndUndo()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[5, 7, 5, 8].Merge();
            worksheet.Cells[6, 4, 6, 6].Merge();

            IFill fill = new PatternFill(PatternType.Solid, Colors.Red, Colors.Red);
            CellRange range = new CellRange(4, 0, 4, 10);
            worksheet.Cells[range].SetFill(fill);

            worksheet.Cells[3, 0, 3, 10].Remove(RemoveShiftType.Up);

            worksheet.Workbook.History.Undo();

            IFill resultFill = worksheet.Cells[range].GetFill().Value;

            Assert.AreEqual(fill, resultFill);
        }

        [TestMethod]
        public void RemoveCellsContainingMergedCellAtTheBeginnigOfWorksheet()
        {
            Cells cells = TestHelper.CreateWorksheet().Cells;
            int startColumnMerge = 4;
            int endColumnToMerge = 8;

            cells[new CellRange(0, startColumnMerge, 0, endColumnToMerge)].Merge();
            cells[new CellRange(0, startColumnMerge - 1, 0, endColumnToMerge + 1)].Remove(RemoveShiftType.Left);

            for (int column = 0; column <= endColumnToMerge + 1; column++)
            {
                Assert.IsFalse(cells.GetIsMerged(new CellIndex(0, column)));
            }

            cells.Worksheet.Workbook.History.Undo();

            for (int column = 0; column < endColumnToMerge + 1; column++)
            {
                if (startColumnMerge <= column && column <= endColumnToMerge)
                {
                    Assert.IsTrue(cells.GetIsMerged(new CellIndex(0, column)));
                }
                else
                {
                    Assert.IsFalse(cells.GetIsMerged(new CellIndex(0, column)));
                }
            }
        }

        [TestMethod]
        public void InsertCellsContainingMergedCellAtTheEndingOfWorksheetShouldNotInsertCells()
        {
            Cells cells = new Workbook().Worksheets.Add().Cells;
            int maxColumnIndex = SpreadsheetDefaultValues.ColumnCount - 1;
            int startColumnMerge = maxColumnIndex - 8;
            int endColumnToMerge = maxColumnIndex - 3;

            cells[new CellRange(0, startColumnMerge, 0, endColumnToMerge)].Merge();
            cells[new CellRange(0, startColumnMerge - 1, 0, endColumnToMerge + 1)].Insert(InsertShiftType.Right);

            for (int column = 0; column < endColumnToMerge + 1; column++)
            {
                if (startColumnMerge <= column && column <= endColumnToMerge)
                {
                    Assert.IsTrue(cells.GetIsMerged(new CellIndex(0, column)));
                }
                else
                {
                    Assert.IsFalse(cells.GetIsMerged(new CellIndex(0, column)));
                }
            }
        }

        #endregion


        #region Insert or remove cells that will affect hyperlinks

        [TestMethod]
        public void RemovingCellsWithHyperlinkAtWorksheetBeginning()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            CellRange addHyperlinkRange = new CellRange(1, 1, 3, 3);
            worksheet.Hyperlinks.Add(addHyperlinkRange, HyperlinkInfo.CreateHyperlink("adress"));

            CellRange removeCellsRange = new CellRange(0, 0, 4, 4);
            worksheet.Cells[removeCellsRange].Remove(RemoveShiftType.Left);

            int actualHyperlinksCount = worksheet.Hyperlinks.GetContainingHyperlinks(removeCellsRange).Count();
            Assert.AreEqual(0, actualHyperlinksCount);

            worksheet.Workbook.History.Undo();

            IEnumerable<SpreadsheetHyperlink> hyperlinks = worksheet.Hyperlinks.GetContainingHyperlinks(removeCellsRange);
            Assert.AreEqual(1, hyperlinks.Count());

            CellRange hyperlinkRange = hyperlinks.First().Range;
            Assert.AreEqual(addHyperlinkRange, hyperlinkRange);
        }

        #endregion


        #region Cell values after inserting tests

        [TestMethod]
        public void InsertCellsInTheMiddleOfValuesShiftRightTest()
        {
            CellRange valuesRange = new CellRange(0, 0, 5, 5);
            CellRange insertRange = new CellRange(0, 1, 5, 1);
            InsertShiftType insertShiftType = InsertShiftType.Right;
            CellRange resultRange = new CellRange(valuesRange.FromIndex, valuesRange.ToIndex.Offset(0, insertRange.ColumnCount));
            bool insertRangeIntersectsValuesRange = true;

            this.AssertCellsInsertion(valuesRange, insertRange, insertShiftType, resultRange, insertRangeIntersectsValuesRange);
        }

        [TestMethod]
        public void RemoveCellsInTheMiddleOfValuesShiftLeftTest()
        {
            CellRange valuesRange = new CellRange(0, 0, 5, 5);
            CellRange removeRange = new CellRange(0, 1, 5, 1);
            RemoveShiftType removeShiftType = RemoveShiftType.Left;
            CellRange resultRange = new CellRange(valuesRange.FromIndex, valuesRange.ToIndex.Offset(0, -removeRange.ColumnCount));
            bool removeRangeIntersectsValuesRange = true;

            this.AssertRemoveCells(valuesRange, removeRange, removeShiftType, resultRange, removeRangeIntersectsValuesRange);
        }

        [TestMethod]
        public void InserteCellsInTheMiddleOfValuesShiftDownTest()
        {
            CellRange valuesRange = new CellRange(0, 0, 5, 5);
            CellRange insertRange = new CellRange(1, 0, 1, 5);
            InsertShiftType insertShiftType = InsertShiftType.Down;
            CellRange resultRange = new CellRange(valuesRange.FromIndex, valuesRange.ToIndex.Offset(insertRange.RowCount, 0));
            bool insertRangeIntersectsValuesRange = true;

            this.AssertCellsInsertion(valuesRange, insertRange, insertShiftType, resultRange, insertRangeIntersectsValuesRange);
        }

        [TestMethod]
        public void RemoveCellsInTheMiddleOfValuesShiftUpTest()
        {
            CellRange valuesRange = new CellRange(0, 0, 5, 5);
            CellRange removeRange = new CellRange(1, 0, 1, 5);
            RemoveShiftType removeShiftType = RemoveShiftType.Up;
            CellRange resultRange = new CellRange(valuesRange.FromIndex, valuesRange.ToIndex.Offset(-removeRange.RowCount, 0));
            bool removeRangeIntersectsValuesRange = true;

            this.AssertRemoveCells(valuesRange, removeRange, removeShiftType, resultRange, removeRangeIntersectsValuesRange);
        }

        [TestMethod]
        public void InsertCellsOnTheLeftOfValuesShiftRightTest()
        {
            CellRange valuesRange = new CellRange(0, 2, 5, 7);
            CellRange insertRange = new CellRange(0, 0, 5, 0);
            InsertShiftType insertShiftType = InsertShiftType.Right;
            CellRange resultRange = new CellRange(valuesRange.FromIndex.Offset(0, insertRange.ColumnCount), valuesRange.ToIndex.Offset(0, insertRange.ColumnCount));
            bool insertRangeIntersectsValuesRange = false;

            this.AssertCellsInsertion(valuesRange, insertRange, insertShiftType, resultRange, insertRangeIntersectsValuesRange);
        }

        [TestMethod]
        public void RemoveCellsOnTheLeftOfValuesShiftLeftTest()
        {
            CellRange valuesRange = new CellRange(0, 2, 5, 7);
            CellRange removeRange = new CellRange(0, 0, 5, 0);
            RemoveShiftType removeShiftType = RemoveShiftType.Left;
            CellRange resultRange = new CellRange(valuesRange.FromIndex.Offset(0, -removeRange.ColumnCount), valuesRange.ToIndex.Offset(0, -removeRange.ColumnCount));
            bool removeRangeIntersectsValuesRange = false;

            this.AssertRemoveCells(valuesRange, removeRange, removeShiftType, resultRange, removeRangeIntersectsValuesRange);
        }

        [TestMethod]
        public void InsertCellsOnTheTopOfValuesShiftDownTest()
        {
            CellRange valuesRange = new CellRange(2, 0, 7, 5);
            CellRange insertRange = new CellRange(0, 0, 0, 5);
            InsertShiftType insertShiftType = InsertShiftType.Down;
            CellRange resultRange = new CellRange(valuesRange.FromIndex.Offset(insertRange.RowCount, 0), valuesRange.ToIndex.Offset(insertRange.RowCount, 0));
            bool insertRangeIntersectsValuesRange = false;

            this.AssertCellsInsertion(valuesRange, insertRange, insertShiftType, resultRange, insertRangeIntersectsValuesRange);
        }

        [TestMethod]
        public void RemoveCellsOnTheTopOfValuesShiftUpTest()
        {
            CellRange valuesRange = new CellRange(2, 0, 7, 5);
            CellRange removeRange = new CellRange(0, 0, 0, 5);
            RemoveShiftType removeShiftType = RemoveShiftType.Up;
            CellRange resultRange = new CellRange(valuesRange.FromIndex.Offset(-removeRange.RowCount, 0), valuesRange.ToIndex.Offset(-removeRange.RowCount, 0));
            bool removeRangeIntersectsValuesRange = false;

            this.AssertRemoveCells(valuesRange, removeRange, removeShiftType, resultRange, removeRangeIntersectsValuesRange);
        }

        #endregion


        #region CanInsertOrRemove Tests

        [TestMethod]
        public void CanInsertOrRemoveTest1()
        {
            bool canInsertRemove = CanInsertInMergeRangesTestScenario(1, 3, 1, 7, ShiftType.Down);
            Assert.IsFalse(canInsertRemove);

            canInsertRemove = CanInsertInMergeRangesTestScenario(1, 3, 1, 7, ShiftType.Up);
            Assert.IsFalse(canInsertRemove);
        }

        [TestMethod]
        public void CanInsertOrRemoveTest2()
        {
            bool canInsertRemove = CanInsertInMergeRangesTestScenario(9, 3, 9, 7, ShiftType.Up);
            Assert.IsTrue(canInsertRemove);

            canInsertRemove = CanInsertInMergeRangesTestScenario(9, 3, 9, 7, ShiftType.Down);
            Assert.IsTrue(canInsertRemove);
        }

        [TestMethod]
        public void CanInsertOrRemoveTest3()
        {
            bool canInsertRemove = CanInsertInMergeRangesTestScenario(1, 2, 1, 9, ShiftType.Up);
            Assert.IsTrue(canInsertRemove);

            canInsertRemove = CanInsertInMergeRangesTestScenario(1, 2, 1, 9, ShiftType.Down);
            Assert.IsTrue(canInsertRemove);
        }

        [TestMethod]
        public void CanInsertOrRemoveTest4()
        {
            bool canInsertRemove = CanInsertInMergeRangesTestScenario(1, 3, 2, 9, ShiftType.Up);
            Assert.IsFalse(canInsertRemove);

            canInsertRemove = CanInsertInMergeRangesTestScenario(1, 3, 2, 9, ShiftType.Down);
            Assert.IsFalse(canInsertRemove);
        }

        [TestMethod]
        public void CanInsertOrRemoveTest5()
        {
            bool canInsertRemove = CanInsertInMergeRangesTestScenario(6, 2, 8, 2, ShiftType.Left);
            Assert.IsTrue(canInsertRemove);

            canInsertRemove = CanInsertInMergeRangesTestScenario(6, 2, 8, 2, ShiftType.Right);
            Assert.IsTrue(canInsertRemove);
        }

        [TestMethod]
        public void CanInsertOrRemoveTest6()
        {
            bool canInsertRemove = CanInsertInMergeRangesTestScenario(5, 2, 8, 2, ShiftType.Left);
            Assert.IsFalse(canInsertRemove);

            canInsertRemove = CanInsertInMergeRangesTestScenario(5, 2, 8, 2, ShiftType.Right);
            Assert.IsFalse(canInsertRemove);
        }

        [TestMethod]
        public void CanInsertOrRemoveTest7()
        {
            bool canInsertRemove = CanInsertInMergeRangesTestScenario(5, 10, 8, 10, ShiftType.Left);
            Assert.IsTrue(canInsertRemove);

            canInsertRemove = CanInsertInMergeRangesTestScenario(5, 10, 8, 10, ShiftType.Right);
            Assert.IsTrue(canInsertRemove);
        }

        [TestMethod]
        public void CanInsertOrRemoveTest8()
        {
            bool canInsertRemove = CanInsertInMergeRangesTestScenario(1, 1, 9, 1, ShiftType.Left);
            Assert.IsTrue(canInsertRemove);

            canInsertRemove = CanInsertInMergeRangesTestScenario(1, 1, 9, 1, ShiftType.Right);
            Assert.IsTrue(canInsertRemove);
        }

        [TestMethod]
        public void CanInsertOrRemoveTest9()
        {
            bool canInsertRemove = CanInsertInMergeRangesTestScenario(3, 1, 9, 1, ShiftType.Left);
            Assert.IsFalse(canInsertRemove);

            canInsertRemove = CanInsertInMergeRangesTestScenario(3, 1, 9, 1, ShiftType.Right);
            Assert.IsFalse(canInsertRemove);
        }

        #endregion


        #region Unmerge before insert/remove tests

        [TestMethod]
        public void UnmergeBeforeInsertRemoveCellsTest1()
        {
            CellRange[] rangesToInsert = { new CellRange(1, 3, 1, 7) };

            Cells cells = this.UnmergeWhenInsertingCellsInMergedRangesTestScenario(rangesToInsert, InsertShiftType.Down);
            Assert.AreEqual(1, cells.GetMergedCellRanges().Count());

            CellIndex[] unmerged = { new CellIndex(2, 2), new CellIndex(3, 6) };
            CellIndex[] merged = { new CellIndex(7, 3) };

            this.AssertMergedCells(cells, merged);
            this.AssertNotMergedCells(cells, unmerged);
        }

        [TestMethod]
        public void UnmergeBeforeInsertRemoveCellsTest2()
        {
            CellRange[] rangesToInsert = { new CellRange(2, 1, 8, 1) };

            Cells cells = this.UnmergeWhenInsertingCellsInMergedRangesTestScenario(rangesToInsert, InsertShiftType.Right);
            Assert.AreEqual(3, cells.GetMergedCellRanges().Count());

            CellIndex[] merged = { new CellIndex(6, 4), new CellIndex(2, 3), new CellIndex(3, 7) };

            this.AssertMergedCells(cells, merged);
        }

        [TestMethod]
        public void UnmergeBeforeInsertRemoveCellsTest3()
        {
            CellRange[] rangesToRemove = { new CellRange(1, 2, 1, 9) };

            Cells cells = this.UnmergeWhenRemovingCellsInMergedRangesTestScenario(rangesToRemove, RemoveShiftType.Up);
            Assert.AreEqual(3, cells.GetMergedCellRanges().Count());

            CellIndex[] merged = { new CellIndex(5, 3), new CellIndex(1, 2), new CellIndex(2, 6) };

            this.AssertMergedCells(cells, merged);
        }

        [TestMethod]
        public void UnmergeBeforeInsertRemoveCellsTest4()
        {
            CellRange[] rangesToRemove = { new CellRange(3, 0, 9, 1) };

            Cells cells = this.UnmergeWhenRemovingCellsInMergedRangesTestScenario(rangesToRemove, RemoveShiftType.Left);
            Assert.AreEqual(2, cells.GetMergedCellRanges().Count());

            CellIndex[] merged = { new CellIndex(6, 1), new CellIndex(3, 4) };
            CellIndex[] unmerged = { new CellIndex(2, 2) };

            this.AssertMergedCells(cells, merged);
            this.AssertNotMergedCells(cells, unmerged);
        }

        [TestMethod]
        public void UnmergeBeforeInsertRemoveCellsUndoTest1()
        {
            CellRange[] rangesToInsert = { new CellRange(1, 3, 1, 7) };

            Cells cells = this.UnmergeWhenInsertingCellsInMergedRangesTestScenario(rangesToInsert, InsertShiftType.Down);
            Assert.AreEqual(1, cells.GetMergedCellRanges().Count());

            cells.Worksheet.Workbook.History.Undo();

            CellIndex[] merged = { new CellIndex(2, 2), new CellIndex(3, 6), new CellIndex(6, 3) };

            Assert.AreEqual(3, cells.GetMergedCellRanges().Count());
            this.AssertMergedCells(cells, merged);
        }

        [TestMethod]
        public void UnmergeBeforeInsertRemoveCellsUndoTest2()
        {
            CellRange[] rangesToInsert = { new CellRange(2, 1, 8, 1) };

            Cells cells = this.UnmergeWhenInsertingCellsInMergedRangesTestScenario(rangesToInsert, InsertShiftType.Right);
            Assert.AreEqual(3, cells.GetMergedCellRanges().Count());

            cells.Worksheet.Workbook.History.Undo();

            CellIndex[] merged = { new CellIndex(2, 2), new CellIndex(3, 6), new CellIndex(6, 3) };

            Assert.AreEqual(3, cells.GetMergedCellRanges().Count());
            this.AssertMergedCells(cells, merged);
        }

        [TestMethod]
        public void UnmergeBeforeInsertRemoveCellsUndoTest3()
        {
            CellRange[] rangesToRemove = { new CellRange(1, 2, 1, 9) };

            Cells cells = this.UnmergeWhenRemovingCellsInMergedRangesTestScenario(rangesToRemove, RemoveShiftType.Up);
            Assert.AreEqual(3, cells.GetMergedCellRanges().Count());

            cells.Worksheet.Workbook.History.Undo();

            CellIndex[] merged = { new CellIndex(2, 2), new CellIndex(3, 6), new CellIndex(6, 3) };

            Assert.AreEqual(3, cells.GetMergedCellRanges().Count());
            this.AssertMergedCells(cells, merged);
        }

        [TestMethod]
        public void UnmergeBeforeInsertRemoveCellsUndoTest4()
        {
            CellRange[] rangesToRemove = { new CellRange(3, 0, 9, 1) };

            Cells cells = this.UnmergeWhenRemovingCellsInMergedRangesTestScenario(rangesToRemove, RemoveShiftType.Left);
            Assert.AreEqual(2, cells.GetMergedCellRanges().Count());

            cells.Worksheet.Workbook.History.Undo();

            CellIndex[] merged = { new CellIndex(2, 2), new CellIndex(3, 6), new CellIndex(6, 3) };

            Assert.AreEqual(3, cells.GetMergedCellRanges().Count());
            this.AssertMergedCells(cells, merged);
        }

        #endregion


        #region Remove cells and rows just above the last used cell

        [TestMethod]
        public void RemoveCellsJustAboveTheLastUsedCellTest()
        {
            CellRange valuesRange = new CellRange(4, 0, 4, 0);
            CellRange removeRange = new CellRange(3, 0, 3, 0);
            RemoveShiftType removeShiftType = RemoveShiftType.Up;
            CellRange resultRange = new CellRange(valuesRange.FromIndex.Offset(-removeRange.RowCount, 0), valuesRange.ToIndex.Offset(-removeRange.RowCount, 0));
            bool removeRangeIntersectsValuesRange = false;

            this.AssertRemoveCells(valuesRange, removeRange, removeShiftType, resultRange, removeRangeIntersectsValuesRange);
        }

        [TestMethod]
        public void RemoveCellsJustAboveTheLastUsedCellWithMoreThanOneValueTest()
        {
            CellRange valuesRange = new CellRange(4, 0, 4, 15);
            CellRange removeRange = new CellRange(3, 0, 3, 20);
            RemoveShiftType removeShiftType = RemoveShiftType.Up;
            CellRange resultRange = new CellRange(valuesRange.FromIndex.Offset(-removeRange.RowCount, 0), valuesRange.ToIndex.Offset(-removeRange.RowCount, 0));
            bool removeRangeIntersectsValuesRange = false;

            this.AssertRemoveCells(valuesRange, removeRange, removeShiftType, resultRange, removeRangeIntersectsValuesRange);
        }

        [TestMethod]
        public void RemoveCellsJustAboveTheLastUsedCellRemovingWholeRowTest()
        {
            CellRange valuesRange = new CellRange(4, 0, 4, 15);
            CellRange removeRange = new CellRange(3, 0, 3, SpreadsheetDefaultValues.ColumnCount - 1);
            RemoveShiftType removeShiftType = RemoveShiftType.Up;
            CellRange resultRange = new CellRange(valuesRange.FromIndex.Offset(-removeRange.RowCount, 0), valuesRange.ToIndex.Offset(-removeRange.RowCount, 0));
            bool removeRangeIntersectsValuesRange = false;

            this.AssertRemoveCells(valuesRange, removeRange, removeShiftType, resultRange, removeRangeIntersectsValuesRange);
        }

        [TestMethod]
        public void RemoveCellsJustOnTheLeftOfTheLastUsedCellTest()
        {
            CellRange valuesRange = new CellRange(0, 4, 0, 4);
            CellRange removeRange = new CellRange(0, 3, 0, 3);
            RemoveShiftType removeShiftType = RemoveShiftType.Left;
            CellRange resultRange = new CellRange(valuesRange.FromIndex.Offset(0, -removeRange.ColumnCount), valuesRange.ToIndex.Offset(0, -removeRange.ColumnCount));
            bool removeRangeIntersectsValuesRange = false;

            this.AssertRemoveCells(valuesRange, removeRange, removeShiftType, resultRange, removeRangeIntersectsValuesRange);
        }

        [TestMethod]
        public void RemoveCellsJustOnTheLeftOfTheLastUsedCellWithMoreThanOneValueTest()
        {
            CellRange valuesRange = new CellRange(0, 4, 15, 4);
            CellRange removeRange = new CellRange(0, 3, 20, 3);
            RemoveShiftType removeShiftType = RemoveShiftType.Left;
            CellRange resultRange = new CellRange(valuesRange.FromIndex.Offset(0, -removeRange.ColumnCount), valuesRange.ToIndex.Offset(0, -removeRange.ColumnCount));
            bool removeRangeIntersectsValuesRange = false;

            this.AssertRemoveCells(valuesRange, removeRange, removeShiftType, resultRange, removeRangeIntersectsValuesRange);
        }

        [TestMethod]
        public void RemoveCellsJustOnTheLeftOfTheLastUsedCellRemovingWholeRowTest()
        {
            CellRange valuesRange = new CellRange(0, 4, 15, 4);
            CellRange removeRange = new CellRange(0, 3, SpreadsheetDefaultValues.RowCount - 1, 3);
            RemoveShiftType removeShiftType = RemoveShiftType.Left;
            CellRange resultRange = new CellRange(valuesRange.FromIndex.Offset(0, -removeRange.ColumnCount), valuesRange.ToIndex.Offset(0, -removeRange.ColumnCount));
            bool removeRangeIntersectsValuesRange = false;

            this.AssertRemoveCells(valuesRange, removeRange, removeShiftType, resultRange, removeRangeIntersectsValuesRange);
        }

        #endregion


        #region Insert or remove cells with values on the end of the worksheet

        [TestMethod]
        public void InsertCellsWithValuesInTheLastColumn()
        {
            CellRange valuesRange = new CellRange(0, SpreadsheetDefaultValues.ColumnCount - 3, 3, SpreadsheetDefaultValues.ColumnCount - 1);
            CellRange insertRange = new CellRange(0, SpreadsheetDefaultValues.ColumnCount - 3, 3, SpreadsheetDefaultValues.ColumnCount - 2);

            Cells cells = TestHelper.CreateWorksheet().Cells;

            Func<int, int, string> rowsColumnsFormula = (row, col) =>
            {
                return "test";
            };

            this.FillCellsInRangeWithFormula(cells, valuesRange, rowsColumnsFormula);

            bool success = cells.Insert(insertRange, InsertShiftType.Right);

            Assert.IsFalse(success);
            Assert.IsFalse(cells[valuesRange].GetValue().IsIndeterminate);
        }

        [TestMethod]
        public void InsertColumnsWithValuesInTheLastColumn()
        {
            CellRange valuesRange = new CellRange(0, SpreadsheetDefaultValues.ColumnCount - 3, 3, SpreadsheetDefaultValues.ColumnCount - 1);
            Worksheet worksheet = TestHelper.CreateWorksheet();

            Cells cells = worksheet.Cells;

            Func<int, int, string> rowsColumnsFormula = (row, col) =>
            {
                return "test";
            };

            this.FillCellsInRangeWithFormula(cells, valuesRange, rowsColumnsFormula);

            bool success = worksheet.Columns.Insert(0);

            Assert.IsFalse(success);
            Assert.IsFalse(cells[valuesRange].GetValue().IsIndeterminate);
        }

        [TestMethod]
        public void InsertRowsWithValuesInTheLastRow()
        {
            CellRange valuesRange = new CellRange(SpreadsheetDefaultValues.RowCount - 3, 0, SpreadsheetDefaultValues.RowCount - 1, 3);

            Worksheet worksheet = TestHelper.CreateWorksheet();
            Cells cells = worksheet.Cells;

            Func<int, int, string> rowsColumnsFormula = (row, col) =>
            {
                return "test";
            };

            this.FillCellsInRangeWithFormula(cells, valuesRange, rowsColumnsFormula);

            bool success = worksheet.Rows.Insert(0);

            Assert.IsFalse(success);
            Assert.IsFalse(cells[valuesRange].GetValue().IsIndeterminate);
        }

        [TestMethod]
        public void InsertRowsWithValuesInTheLastButOneRow()
        {
            CellRange valuesRange = new CellRange(SpreadsheetDefaultValues.RowCount - 3, 0, SpreadsheetDefaultValues.RowCount - 2, 3);

            Worksheet worksheet = TestHelper.CreateWorksheet();
            Cells cells = worksheet.Cells;

            Func<int, int, string> rowsColumnsFormula = (row, col) =>
            {
                return "test";
            };

            this.FillCellsInRangeWithFormula(cells, valuesRange, rowsColumnsFormula);

            bool success = worksheet.Rows.Insert(0);

            Assert.IsTrue(success);
            Assert.IsFalse(cells[valuesRange.Offset(1, 0)].GetValue().IsIndeterminate);
        }

        #endregion


        #region Insert columns and rows Undo and Redo

        [TestMethod]
        public void InsertColumnSelectCellInsertCellUndoRedo()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellRange valuesRange = new CellRange(0, 0, 8, 8);
            CellSelection selection = worksheet.Cells[valuesRange];
            IFill fill = PatternFill.CreateSolidFill(Colors.Blue);
            selection.SetFill(fill);
            selection.SetValue(1);

            worksheet.Columns.Insert(1);

            for (int row = valuesRange.FromIndex.RowIndex; row <= valuesRange.ToIndex.RowIndex; row++)
            {
                CellSelection cell = worksheet.Cells[row, 0];
                Assert.AreEqual("1", cell.GetValue().Value.RawValue);
                Assert.AreEqual(fill, cell.GetFill().Value);
            }

            for (int row = valuesRange.FromIndex.RowIndex; row <= valuesRange.ToIndex.RowIndex; row++)
            {
                CellSelection cell = worksheet.Cells[row, 1];
                Assert.AreEqual(string.Empty, cell.GetValue().Value.RawValue);
                Assert.AreEqual(fill, cell.GetFill().Value);
            }

            for (int row = valuesRange.FromIndex.RowIndex; row <= valuesRange.ToIndex.RowIndex; row++)
            {
                for (int col = 2; col <= valuesRange.ToIndex.ColumnIndex; col++)
                {
                    CellSelection cell = worksheet.Cells[row, col];
                    Assert.AreEqual("1", cell.GetValue().Value.RawValue);
                    Assert.AreEqual(fill, cell.GetFill().Value);
                }
            }
        }

        [TestMethod]
        public void InsertRowsSelectCellInsertCellUndoRedo()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellRange valuesRange = new CellRange(0, 0, 8, 8);
            CellSelection selection = worksheet.Cells[valuesRange];
            IFill fill = PatternFill.CreateSolidFill(Colors.Blue);
            selection.SetFill(fill);
            selection.SetValue(1);

            worksheet.Rows.Insert(1);

            for (int col = valuesRange.FromIndex.ColumnIndex; col < valuesRange.ToIndex.ColumnIndex; col++)
            {
                CellSelection cell = worksheet.Cells[0, col];
                Assert.AreEqual("1", cell.GetValue().Value.RawValue);
                Assert.AreEqual(fill, cell.GetFill().Value);
            }

            for (int col = valuesRange.FromIndex.ColumnIndex; col < valuesRange.ToIndex.ColumnIndex; col++)
            {
                CellSelection cell = worksheet.Cells[1, col];
                Assert.AreEqual(string.Empty, cell.GetValue().Value.RawValue);
                Assert.AreEqual(fill, cell.GetFill().Value);
            }
        }

        #endregion


        #region Insert columns around cells with Fill

        [TestMethod]
        public void InsertColumnsAfterCellsWithAppliedFill()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellRange valuesRange = new CellRange(0, 3, 8, 3);
            CellSelection selection = worksheet.Cells[valuesRange];
            IFill fill = PatternFill.CreateSolidFill(Colors.Blue);
            selection.SetFill(fill);
            selection.SetValue(1);

            worksheet.Columns.Insert(4);

            for (int row = valuesRange.FromIndex.RowIndex; row <= valuesRange.ToIndex.RowIndex; row++)
            {
                for (int col = valuesRange.FromIndex.ColumnIndex; col <= valuesRange.ToIndex.ColumnIndex; col++)
                {
                    CellSelection cell = worksheet.Cells[row, col];
                    Assert.AreEqual("1", cell.GetValue().Value.RawValue);
                    Assert.AreEqual(fill, cell.GetFill().Value);
                }
            }

            CellRange newRange = valuesRange.Offset(0, 1);
            for (int row = newRange.FromIndex.RowIndex; row <= newRange.ToIndex.RowIndex; row++)
            {
                for (int col = newRange.FromIndex.ColumnIndex; col <= newRange.ToIndex.ColumnIndex; col++)
                {
                    CellSelection cell = worksheet.Cells[row, col];
                    Assert.AreEqual(string.Empty, cell.GetValue().Value.RawValue);
                    Assert.AreEqual(fill, cell.GetFill().Value);
                }
            }
        }

        [TestMethod]
        public void InsertColumnsBeforeCellsWithAppliedFill()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellRange valuesRange = new CellRange(0, 3, 8, 3);
            CellSelection selection = worksheet.Cells[valuesRange];
            IFill fill = PatternFill.CreateSolidFill(Colors.Blue);
            selection.SetFill(fill);
            selection.SetValue(1);

            worksheet.Columns.Insert(1);

            CellRange newRange = valuesRange.Offset(0, 1);
            for (int row = newRange.FromIndex.RowIndex; row <= newRange.ToIndex.RowIndex; row++)
            {
                for (int col = newRange.FromIndex.ColumnIndex; col <= newRange.ToIndex.ColumnIndex; col++)
                {
                    CellSelection cell = worksheet.Cells[row, col];
                    Assert.AreEqual("1", cell.GetValue().Value.RawValue);
                    Assert.AreEqual(fill, cell.GetFill().Value);
                }
            }
        }

        #endregion

        #region Insert cells around cells with Fill

        [TestMethod]
        public void InsertCellsAfterCellsWithAppliedFill()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellRange valuesRange = new CellRange(0, 3, 8, 3);
            CellSelection selection = worksheet.Cells[valuesRange];
            IFill fill = PatternFill.CreateSolidFill(Colors.Blue);
            selection.SetFill(fill);
            selection.SetValue(1);

            worksheet.Cells[0, 4, 8, 4].Insert(InsertShiftType.Right);

            for (int row = valuesRange.FromIndex.RowIndex; row <= valuesRange.ToIndex.RowIndex; row++)
            {
                for (int col = valuesRange.FromIndex.ColumnIndex; col <= valuesRange.ToIndex.ColumnIndex; col++)
                {
                    CellSelection cell = worksheet.Cells[row, col];
                    Assert.AreEqual("1", cell.GetValue().Value.RawValue);
                    Assert.AreEqual(fill, cell.GetFill().Value);
                }
            }

            CellRange newRange = valuesRange.Offset(0, 1);
            for (int row = newRange.FromIndex.RowIndex; row <= newRange.ToIndex.RowIndex; row++)
            {
                for (int col = newRange.FromIndex.ColumnIndex; col <= newRange.ToIndex.ColumnIndex; col++)
                {
                    CellSelection cell = worksheet.Cells[row, col];
                    Assert.AreEqual(string.Empty, cell.GetValue().Value.RawValue);
                    Assert.AreEqual(fill, cell.GetFill().Value);
                }
            }
        }

        [TestMethod]
        public void InsertCellsAfterCellsWithAppliedFillFollowingCellsWithFill()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            CellIndex redCell = new CellIndex(2, 2);
            CellIndex blueCell = new CellIndex(4, 4);

            IFill redFill = PatternFill.CreateSolidFill(Colors.Red);
            IFill blueFill = PatternFill.CreateSolidFill(Colors.Blue);

            worksheet.Cells[redCell].SetFill(redFill);
            worksheet.Cells[blueCell].SetFill(blueFill);

            worksheet.Cells[0, 3, SpreadsheetDefaultValues.RowCount - 1, 3].Insert(InsertShiftType.Right);

            CellIndex newRedCell = new CellIndex(2, 3);
            CellIndex newBlueCell = new CellIndex(4, 5);

            Assert.AreEqual(redFill, worksheet.Cells[redCell].GetFill().Value);
            Assert.AreEqual(redFill, worksheet.Cells[newRedCell].GetFill().Value);
            Assert.AreEqual(blueFill, worksheet.Cells[newBlueCell].GetFill().Value);

            CellRange usedRange = worksheet.UsedCellRange;
            for (int i = usedRange.FromIndex.RowIndex; i <= usedRange.ToIndex.RowIndex; i++)
            {
                for (int j = usedRange.FromIndex.ColumnIndex; j <= usedRange.ToIndex.ColumnIndex; j++)
                {
                    CellIndex currentCellIndex = new CellIndex(i, j);
                    if (currentCellIndex == redCell || currentCellIndex == newRedCell || currentCellIndex == newBlueCell)
                    {
                        continue;
                    }

                    Assert.AreEqual(PatternFill.Default, worksheet.Cells[currentCellIndex].GetFill().Value);
                }
            }

            worksheet.Workbook.History.Undo();

            Assert.AreEqual(redFill, worksheet.Cells[redCell].GetFill().Value);
            Assert.AreEqual(blueFill, worksheet.Cells[blueCell].GetFill().Value);

            usedRange = worksheet.UsedCellRange;
            for (int i = usedRange.FromIndex.RowIndex; i <= usedRange.ToIndex.RowIndex; i++)
            {
                for (int j = usedRange.FromIndex.ColumnIndex; j <= usedRange.ToIndex.ColumnIndex; j++)
                {
                    CellIndex currentCellIndex = new CellIndex(i, j);
                    if (currentCellIndex == redCell || currentCellIndex == blueCell)
                    {
                        continue;
                    }

                    Assert.AreEqual(PatternFill.Default, worksheet.Cells[currentCellIndex].GetFill().Value);
                }
            }

            worksheet.Workbook.History.Redo();

            Assert.AreEqual(redFill, worksheet.Cells[redCell].GetFill().Value);
            Assert.AreEqual(redFill, worksheet.Cells[newRedCell].GetFill().Value);
            Assert.AreEqual(blueFill, worksheet.Cells[newBlueCell].GetFill().Value);

            usedRange = worksheet.UsedCellRange;
            for (int i = usedRange.FromIndex.RowIndex; i <= usedRange.ToIndex.RowIndex; i++)
            {
                for (int j = usedRange.FromIndex.ColumnIndex; j <= usedRange.ToIndex.ColumnIndex; j++)
                {
                    CellIndex currentCellIndex = new CellIndex(i, j);
                    if (currentCellIndex == redCell || currentCellIndex == newRedCell || currentCellIndex == newBlueCell)
                    {
                        continue;
                    }

                    Assert.AreEqual(PatternFill.Default, worksheet.Cells[currentCellIndex].GetFill().Value);
                }
            }
        }

        #endregion
    }
}