﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests
{
    public static class ExpressionTestHelper
    {
        private static Worksheet worksheet;

        static ExpressionTestHelper()
        {
            InitializeWorksheet();
        }

        public static Worksheet Worksheet
        {
            get
            {
                return worksheet;
            }
        }

        public static Type TryParseExpression(string cellInput, Worksheet worksheet = null, CellIndex cellIndex = null, bool addFunctionStart = false)
        {
            try
            {
                ParseExpression(cellInput, worksheet, cellIndex, addFunctionStart);
                return null;
            }
            catch (Exception ex)
            {
                return ex.InnerException.GetType();
            }
        }

        public static RadExpression ParseExpression(string cellInput, Worksheet worksheet = null, CellIndex cellIndex = null, bool addFunctionStart = false)
        {
            if (worksheet == null)
            {
                worksheet = TestHelper.CreateWorksheet();
            }

            if (cellIndex == null)
            {
                cellIndex = new CellIndex(0, 0);
            }

            string stringToParse = addFunctionStart ? "=" + cellInput : cellInput;

            worksheet.Cells[cellIndex].SetValue(stringToParse);
            FormulaCellValue formulaCellValue = worksheet.Cells[cellIndex].GetValue().Value as FormulaCellValue;

            if (formulaCellValue != null)
            {
                return formulaCellValue.Value;
            }

            return null;
        }

        public static void AssertAreEssentiallyEqualExpressionResult(string input, Type expectedExpressionType, double expectedResult, Worksheet worksheet = null)
        {
            RadExpression expression = ParseExpression(input, addFunctionStart: true);

            Assert.IsInstanceOfType(expression, expectedExpressionType);
            NumberExpression result = expression.GetValue() as NumberExpression;

            Assert.IsNotNull(result);
            Assert.IsTrue(TestHelper.AreDoubleValuesEssentiallyEqual(expectedResult, result.Value));
        }

        public static void AssertExpressionResult(string input, Type expectedExpressionType, double expectedResult, Worksheet worksheet = null, bool addFunctionStart = true)
        {
            AssertExpressionResult<double, NumberExpression>(input, expectedExpressionType, expectedResult, worksheet, addFunctionStart);
        }

        public static void AssertExpressionResult(string input, Type expectedExpressionType, string expectedResult, Worksheet worksheet = null, bool addFunctionStart = true)
        {
            AssertExpressionResult<string, StringExpression>(input, expectedExpressionType, expectedResult, worksheet, addFunctionStart);
        }

        public static void AssertExpressionResult(string input, Type expectedExpressionType, bool expectedResult, Worksheet worksheet = null, bool addFunctionStart = true)
        {
            AssertExpressionResult<bool, BooleanExpression>(input, expectedExpressionType, expectedResult, worksheet, addFunctionStart);
        }

        public static void AssertErrorExpressionResult(string input, RadExpression expectedErrorExpression, Worksheet worksheet = null)
        {
            RadExpression actualErrorExpression = ParseExpression(input.ReplaceCultureDependentSeparators(), worksheet, new CellIndex(100, 100), addFunctionStart: true);
            Assert.AreEqual(expectedErrorExpression, actualErrorExpression.GetValue());
        }

        private static void AssertExpressionResult<T, U>(string input, Type expectedExpressionType, T expectedResult, Worksheet worksheet = null, bool addFunctionStart = true) where U : ConstantExpression<T>
        {
            CellIndex cellIndex = new CellIndex(100, 100);
            RadExpression expression = ParseExpression(input.ReplaceCultureDependentSeparators(), worksheet, cellIndex, addFunctionStart);
            Assert.IsNotNull(expression);
            Assert.IsInstanceOfType(expression, expectedExpressionType);

            U expressionValue = expression.GetValue() as U;
            Assert.IsNotNull(expressionValue);
            Assert.AreEqual(expectedResult, expressionValue.Value);
        }

        public static void AssertThrownException(string input, Type expectedExceptionType)
        {
            Type exceptionType = TryParseExpression(input, addFunctionStart: true);

            Assert.AreEqual(expectedExceptionType, exceptionType);
        }

        public static void AssertExpressionToStringCorrectness(string expressionAsString)
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            RadExpression expression = ParseExpression(expressionAsString, worksheet, new CellIndex(0, 0));
            RadExpression expression2 = ParseExpression(expression.ToString(), worksheet, new CellIndex(1, 1), addFunctionStart: true);

            Assert.AreEqual(expression.GetValue().NumberValue(), expression2.GetValue().NumberValue());
        }

        public static void AssertCellValueWithName(Worksheet worksheet, CellIndex cellIndex, string name, string expectedRawValue, string expectedResultValue)
        {
            Assert.AreEqual(expectedRawValue, worksheet.Names[name].RefersTo);
            AssertCellValue(worksheet, cellIndex, String.Format("={0}", name), expectedResultValue);
        }

        public static void AssertCellValue(Worksheet worksheet, CellIndex cellIndex, string expectedRawValue, string expectedResultValue)
        {
            ICellValue cellValue = worksheet.Cells[cellIndex].GetValue().Value;

            string rawValue = cellValue.RawValue;
            Assert.AreEqual(expectedRawValue, rawValue);

            string resultValue = cellValue.GetResultValueAsString(CellValueFormat.GeneralFormat);
            Assert.AreEqual(expectedResultValue, resultValue);
        }

        // TODO: Insert random intervals at random positions
        public static string InsertIntervalsIntoExpression(string expression)
        {
            for (int i = 0; i < expression.Length; i += 2)
            {
                expression = expression.Insert(i, " ");
            }
            return expression;
        }

        public static void InitializeWorksheet()
        {
            Workbook workbook = new Workbook();
            workbook.Sheets.Add(SheetType.Worksheet);
            worksheet = workbook.ActiveWorksheet;

            worksheet.Cells[0, 0].SetValue(10);
            worksheet.Cells[0, 1].SetValue(-5);
            worksheet.Cells[0, 2].SetValue("=concatenate(20)");
            worksheet.Cells[0, 3].SetValue("string");
            worksheet.Cells[0, 4].SetValue("true");
            worksheet.Cells[0, 5].SetValue("false");

            worksheet.Cells[1, 0].SetValue(10);
            worksheet.Cells[1, 1].SetValue("=concatenate(20)");
            worksheet.Cells[1, 2].SetValue("string");
            worksheet.Cells[1, 3].SetValue("=5/0");
            worksheet.Cells[1, 4].SetValue("=LN(0)");
            worksheet.Cells[1, 5].SetValue("true");
            worksheet.Cells[1, 6].SetValue("ODD(\"test\")");

            worksheet.Cells[2, 0].SetValue("=B3");
            worksheet.Cells[2, 1].SetValue("=C3");
            worksheet.Cells[2, 2].SetValue("=D3");

            worksheet.Cells[3, 0].SetValue("=B3:B5");
            worksheet.Cells[3, 1].SetValue("=C3:C5");
            worksheet.Cells[3, 2].SetValue("=D3:D5");

            worksheet.Cells[2, 3].SetValue(2);
            worksheet.Cells[3, 3].SetValue(3);
            worksheet.Cells[4, 3].SetValue(4);
        }
    }
}
