﻿using System;
using System.Linq;
using System.Windows.Media;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Sorting;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Theming;

namespace Telerik.Windows.Documents.Spreadsheet.Tests.Sorting
{
    [TestClass]
    public class Sorting_ForeColor_Should
    {
        #region Constants

        private static readonly ThemableColor color1 = new ThemableColor(Colors.Red);
        private static readonly ThemableColor color2 = new ThemableColor(Colors.Green);
        private static readonly ThemableColor color3 = new ThemableColor(Colors.Blue);

        #endregion


        #region Properties

        public Worksheet Worksheet { get; set; }

        public CellRange UsedRange
        {
            get
            {
                return new CellRange(0, 0, 10, 0);
            }
        }

        #endregion


        #region Methods

        [TestInitialize]
        public void TestInitialize()
        {
            this.Worksheet = TestHelper.CreateWorksheet();
            Color[] colors = new Color[] 
            {
                Colors.Green,
                Colors.Red,
                Colors.Blue
            };

            int colorsCounter = 0;
            for (int row = this.UsedRange.FromIndex.RowIndex; row <= this.UsedRange.ToIndex.RowIndex; row++)
            {
                for (int column = this.UsedRange.FromIndex.ColumnIndex; column <= this.UsedRange.ToIndex.ColumnIndex; column++)
                {
                    this.Worksheet.Cells[row, column].SetValue(row + column);
                    this.Worksheet.Cells[row, column].SetForeColor(new ThemableColor(colors[colorsCounter]));

                    colorsCounter++;
                    colorsCounter = (colorsCounter == 3) ? 0 : colorsCounter;
                }
            }
        }

        private void AssertColorsAndValues(ThemableColor[] expectedColors, int[] expectedValues)
        {
            var cells = this.Worksheet.Cells;

            for (int i = 0; i < this.UsedRange.RowCount; i++)
            {
                Assert.AreEqual(expectedColors[i], cells[i, 0].GetForeColor().Value);
                Assert.AreEqual(expectedValues[i].ToString(), cells[i, 0].GetValue().Value.RawValue);
            }
        }

        #endregion


        [TestMethod]
        public void SortRedOnTheTopByTheFirstColumn_WhenTheSortConditionUsesAccendingOrder()
        {
            SortOrder sortOrder = SortOrder.Ascending;

            ForeColorSortCondition sortCondition1 = new ForeColorSortCondition(0, color1, sortOrder);

            var cells = this.Worksheet.Cells;
            cells[this.UsedRange].Sort(sortCondition1);

            ThemableColor[] expectedColors = new ThemableColor[]
            {
                color1,
                color1,
                color1,
                color1,
                color2,
                color3,
                color2,
                color3,
                color2,
                color3,
                color2,
            };

            int[] expectedValues = new int[]
            {
                1,4,7,10,0,2,3,5,6,8,9
            };

            this.AssertColorsAndValues(expectedColors, expectedValues);
        }

        [TestMethod]
        public void SortGreenOnTheTopByTheFirstColumn_WhenTheSortConditionUsesAccendingOrder()
        {
            SortOrder sortOrder = SortOrder.Ascending;

            ForeColorSortCondition sortCondition1 = new ForeColorSortCondition(0, color2, sortOrder);

            var cells = this.Worksheet.Cells;
            cells[this.UsedRange].Sort(sortCondition1);

            ThemableColor[] expectedColors = new ThemableColor[]
            {
                color2,
                color2,
                color2,
                color2,
                color1,
                color3,
                color1,
                color3,
                color1,
                color3,
                color1,
            };

            int[] expectedValues = new int[]
            {
                0,3,6,9,1,2,4,5,7,8,10
            };

            this.AssertColorsAndValues(expectedColors, expectedValues);
        }

        [TestMethod]
        public void SortBlueOnTheTopByTheFirstColumn_WhenTheSortConditionUsesAccendingOrder()
        {
            SortOrder sortOrder = SortOrder.Ascending;

            ForeColorSortCondition sortCondition1 = new ForeColorSortCondition(0, color3, sortOrder);

            var cells = this.Worksheet.Cells;
            cells[this.UsedRange].Sort(sortCondition1);

            ThemableColor[] expectedColors = new ThemableColor[]
            {
                color3,
                color3,
                color3,
                color2,
                color1,
                color2,
                color1,
                color2,
                color1,
                color2,
                color1,
            };

            int[] expectedValues = new int[]
            {
                2,5,8,0,1,3,4,6,7,9,10
            };

            this.AssertColorsAndValues(expectedColors, expectedValues);
        }

        [TestMethod]
        public void SortRedOnTheBottomByTheFirstColumn_WhenTheSortConditionUsesDescendingOrder()
        {
            SortOrder sortOrder = SortOrder.Descending;

            ForeColorSortCondition sortCondition1 = new ForeColorSortCondition(0, color1, sortOrder);

            var cells = this.Worksheet.Cells;
            cells[this.UsedRange].Sort(sortCondition1);

            ThemableColor[] expectedColors = new ThemableColor[]
            {
                color2,
                color3,
                color2,
                color3,
                color2,
                color3,
                color2,
                color1,
                color1,
                color1,
                color1,
            };

            int[] expectedValues = new int[]
            {
                0,2,3,5,6,8,9,1,4,7,10
            };

            this.AssertColorsAndValues(expectedColors, expectedValues);
        }

        [TestMethod]
        public void SortGreenOnTheBottomByTheFirstColumn_WhenTheSortConditionUsesDescendingOrder()
        {
            SortOrder sortOrder = SortOrder.Descending;

            ForeColorSortCondition sortCondition1 = new ForeColorSortCondition(0, color2, sortOrder);

            var cells = this.Worksheet.Cells;
            cells[this.UsedRange].Sort(sortCondition1);

            ThemableColor[] expectedColors = new ThemableColor[]
            {
                color1,
                color3,
                color1,
                color3,
                color1,
                color3,
                color1,
                color2,
                color2,
                color2,
                color2,
            };

            int[] expectedValues = new int[]
            {
                1,2,4,5,7,8,10,0,3,6,9
            };

            this.AssertColorsAndValues(expectedColors, expectedValues);
        }

        [TestMethod]
        public void SortBlueOnTheBottomByTheFirstColumn_WhenTheSortConditionUsesDescendingOrder()
        {
            SortOrder sortOrder = SortOrder.Descending;

            ForeColorSortCondition sortCondition1 = new ForeColorSortCondition(0, color3, sortOrder);

            var cells = this.Worksheet.Cells;
            cells[this.UsedRange].Sort(sortCondition1);

            ThemableColor[] expectedColors = new ThemableColor[]
            {
                color2,
                color1,
                color2,
                color1,
                color2,
                color1,
                color2,
                color1,
                color3,
                color3,
                color3,
            };

            int[] expectedValues = new int[]
            {
                0,1,3,4,6,7,9,10,2,5,8
            };

            this.AssertColorsAndValues(expectedColors, expectedValues);
        }

        [TestMethod]
        public void SortRedGreenBlueByTheFirstColumn_WhenTheSortConditionUsesAccendingOrder()
        {
            SortOrder sortOrder = SortOrder.Ascending;

            ForeColorSortCondition sortCondition1 = new ForeColorSortCondition(0, color1, sortOrder);
            ForeColorSortCondition sortCondition2 = new ForeColorSortCondition(0, color2, sortOrder);
            ForeColorSortCondition sortCondition3 = new ForeColorSortCondition(0, color3, sortOrder);

            var cells = this.Worksheet.Cells;
            cells[this.UsedRange].Sort(new ISortCondition[] { sortCondition1, sortCondition2, sortCondition3 });

            ThemableColor[] expectedColors = new ThemableColor[]
            {
                color1,
                color1,
                color1,
                color1,
                color2,
                color2,
                color2,
                color2,
                color3,
                color3,
                color3,
            };

            int[] expectedValues = new int[]
            {
                1,4,7,10,0,3,6,9,2,5,8
            };

            this.AssertColorsAndValues(expectedColors, expectedValues);
        }

        [TestMethod]
        public void SortRedGreenBlueByTheFirstColumn_WhenTheSortConditionUsesDescendingOrder()
        {
            SortOrder sortOrder = SortOrder.Descending;

            ForeColorSortCondition sortCondition1 = new ForeColorSortCondition(0, color1, sortOrder);
            ForeColorSortCondition sortCondition2 = new ForeColorSortCondition(0, color2, sortOrder);
            ForeColorSortCondition sortCondition3 = new ForeColorSortCondition(0, color3, sortOrder);

            var cells = this.Worksheet.Cells;
            cells[this.UsedRange].Sort(new ISortCondition[] { sortCondition1, sortCondition2, sortCondition3 });

            ThemableColor[] expectedColors = new ThemableColor[]
            {
                color3,
                color3,
                color3,
                color2,
                color2,
                color2,
                color2,
                color1,
                color1,
                color1,
                color1,
            };

            int[] expectedValues = new int[]
            {
                2,5,8,0,3,6,9,1,4,7,10
            };

            this.AssertColorsAndValues(expectedColors, expectedValues);
        }


        #region Duplicating Sort Conditions

        [TestMethod]
        [ExpectedException(typeof(SortingException))]
        public void SortThrowsException_WhenDpulicatingForeColorSortConditionsAreApplied()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex = 0;
            SortOrder sortOrder = SortOrder.Ascending;

            ForeColorSortCondition sortCondition = new ForeColorSortCondition(columnIndex, color2, sortOrder);
            ForeColorSortCondition sortCondition1 = new ForeColorSortCondition(columnIndex, color2, sortOrder);

            this.Worksheet.Cells[range].Sort(sortCondition, sortCondition1);
        }

        #endregion
    }
}
