﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Core;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Sorting;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests.Sorting
{
    [TestClass]
    public class Sorting_Values_Should
    {
        #region Properties

        public Worksheet Worksheet { get; set; }

        #endregion


        #region Methods

        [TestInitialize]
        public void TestInitialize()
        {
            this.Worksheet = TestHelper.CreateWorksheet();
        }

        private void InitializeValues(CellRange range, List<string>[] values)
        {
            using (new UpdateScope(this.Worksheet.BeginUndoGroup, this.Worksheet.EndUndoGroup))
            {
                for (int column = range.FromIndex.ColumnIndex; column <= range.ToIndex.ColumnIndex; column++)
                {
                    for (int row = range.FromIndex.RowIndex; row <= range.ToIndex.RowIndex; row++)
                    {
                        this.Worksheet.Cells[row, column].SetValue(values[column][row]);
                    }
                }
            }
        }

        private void AssertValues(CellRange range, List<string>[] expectedValues)
        {
            var cells = this.Worksheet.Cells;

            for (int column = range.FromIndex.ColumnIndex; column <= range.ToIndex.ColumnIndex; column++)
            {
                for (int row = range.FromIndex.RowIndex; row <= range.ToIndex.RowIndex; row++)
                {
                    Assert.AreEqual(expectedValues[column][row].ToString(), cells[row, column].GetValue().Value.GetResultValueAsString(CellValueFormat.GeneralFormat));
                }
            }
        }

        private void AssertSorting(CellRange range, List<string>[] values, List<string>[] expectedValues, params ISortCondition[] sortConditions)
        {
            this.AssertSorting(range, range, values, expectedValues, sortConditions);
        }

        private void AssertSorting(CellRange sortRange, CellRange valuesRange, List<string>[] values, List<string>[] expectedValues, params ISortCondition[] sortConditions)
        {
            this.InitializeValues(valuesRange, values);

            this.Worksheet.Cells[sortRange].Sort(sortConditions);

            this.AssertValues(sortRange, expectedValues);
        }

        private void AssertRowHeights(CellRange range, double[] expectedRowHeights)
        {
            for (int row = range.FromIndex.RowIndex; row <= range.ToIndex.RowIndex; row++)
            {
                for (int column = range.FromIndex.ColumnIndex; column <= range.ToIndex.ColumnIndex; column++)
                {
                    Assert.AreEqual(expectedRowHeights[row - range.FromIndex.RowIndex], this.Worksheet.Rows[row].GetHeight().Value.Value);
                }
            }
        }

        private void SetRowHeights(CellRange range, double[] rowHeights)
        {
            for (int row = range.FromIndex.RowIndex; row <= range.ToIndex.RowIndex; row++)
            {
                this.Worksheet.Rows[row].SetHeight(new RowHeight(rowHeights[row - range.FromIndex.RowIndex], true));
            }
        }

        #endregion


        #region Numbers Sort Tests

        [TestMethod]
        public void SortNumbersAscendingByFirstColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex = 0;
            SortOrder sortOrder = SortOrder.Ascending;

            List<string>[] values = new List<string>[]
            {
                new List<string> {"5", "3", "7", "1", "8", "9", "4", "2", "0", "6"},
                new List<string> {"1", "4", "2", "3", "5", "7", "9", "0", "8", "6"},
                new List<string> {"3", "1", "5", "8", "7", "9", "2", "4", "0", "6"},
                new List<string> {"8", "7", "0", "1", "2", "3", "4", "5", "9", "6"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string>() {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9"},
                new List<string>() {"8", "3", "0", "4", "9", "1", "6", "2", "5", "7"},
                new List<string>() {"0", "8", "4", "1", "2", "3", "6", "5", "7", "9"},
                new List<string>() {"9", "1", "5", "7", "4", "8", "6", "0", "2", "3"}
            };

            this.AssertSorting(range, values, expectedValues, new ValuesSortCondition(columnIndex, sortOrder));
        }

        [TestMethod]
        public void SortNumbersDescendingByFirstColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex = 0;
            SortOrder sortOrder = SortOrder.Descending;

            List<string>[] values = new List<string>[]
            {
                new List<string> {"5", "3", "7", "1", "8", "9", "4", "2", "0", "6"},
                new List<string> {"1", "4", "2", "3", "5", "7", "9", "0", "8", "6"},
                new List<string> {"3", "1", "5", "8", "7", "9", "2", "4", "0", "6"},
                new List<string> {"8", "7", "0", "1", "2", "3", "4", "5", "9", "6"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string>(),
                new List<string>(),
                new List<string>(),
                new List<string>(),
            };

            expectedValues[0].AddRange((new string[] { "0", "1", "2", "3", "4", "5", "6", "7", "8", "9" }).Reverse());
            expectedValues[1].AddRange((new string[] { "8", "3", "0", "4", "9", "1", "6", "2", "5", "7" }).Reverse());
            expectedValues[2].AddRange((new string[] { "0", "8", "4", "1", "2", "3", "6", "5", "7", "9" }).Reverse());
            expectedValues[3].AddRange((new string[] { "9", "1", "5", "7", "4", "8", "6", "0", "2", "3" }).Reverse());

            this.AssertSorting(range, values, expectedValues, new ValuesSortCondition(columnIndex, sortOrder));
        }

        [TestMethod]
        public void SortNumbersAscendingBySecondColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex = 1;
            SortOrder sortOrder = SortOrder.Ascending;

            List<string>[] values = new List<string>[]
            {
                new List<string> {"5", "3", "7", "1", "8", "9", "4", "2", "0", "6"},
                new List<string> {"1", "4", "2", "3", "5", "7", "9", "0", "8", "6"},
                new List<string> {"3", "1", "5", "8", "7", "9", "2", "4", "0", "6"},
                new List<string> {"8", "7", "0", "1", "2", "3", "4", "5", "9", "6"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"2", "5", "7", "1", "3", "8", "6", "9", "0", "4"},
                new List<string> {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9"},
                new List<string> {"4", "3", "5", "8", "1", "7", "6", "9", "0", "2"},
                new List<string> {"5", "8", "0", "1", "7", "2", "6", "3", "9", "4"}
            };

            this.AssertSorting(range, values, expectedValues, new ValuesSortCondition(columnIndex, sortOrder));
        }

        [TestMethod]
        public void SortNumbersDescendingBySecondColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex = 1;
            SortOrder sortOrder = SortOrder.Descending;

            List<string>[] values = new List<string>[]
            {
                new List<string> {"5", "3", "7", "1", "8", "9", "4", "2", "0", "6"},
                new List<string> {"1", "4", "2", "3", "5", "7", "9", "0", "8", "6"},
                new List<string> {"3", "1", "5", "8", "7", "9", "2", "4", "0", "6"},
                new List<string> {"8", "7", "0", "1", "2", "3", "4", "5", "9", "6"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string>(),
                new List<string>(),
                new List<string>(),
                new List<string>(),
            };

            expectedValues[0].AddRange((new string[] { "2", "5", "7", "1", "3", "8", "6", "9", "0", "4" }).Reverse());
            expectedValues[1].AddRange((new string[] { "0", "1", "2", "3", "4", "5", "6", "7", "8", "9" }).Reverse());
            expectedValues[2].AddRange((new string[] { "4", "3", "5", "8", "1", "7", "6", "9", "0", "2" }).Reverse());
            expectedValues[3].AddRange((new string[] { "5", "8", "0", "1", "7", "2", "6", "3", "9", "4" }).Reverse());

            this.AssertSorting(range, values, expectedValues, new ValuesSortCondition(columnIndex, sortOrder));
        }

        [TestMethod]
        public void SortNumbersAscendingByThirdColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex = 2;
            SortOrder sortOrder = SortOrder.Ascending;

            List<string>[] values = new List<string>[]
            {
                new List<string> {"5", "3", "7", "1", "8", "9", "4", "2", "0", "6"},
                new List<string> {"1", "4", "2", "3", "5", "7", "9", "0", "8", "6"},
                new List<string> {"3", "1", "5", "8", "7", "9", "2", "4", "0", "6"},
                new List<string> {"8", "7", "0", "1", "2", "3", "4", "5", "9", "6"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"0", "3", "4", "5", "2", "7", "6", "8", "1", "9"},
                new List<string> {"8", "4", "9", "1", "0", "2", "6", "5", "3", "7"},
                new List<string> {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9"},
                new List<string> {"9", "7", "4", "8", "5", "0", "6", "2", "1", "3"}
            };

            this.AssertSorting(range, values, expectedValues, new ValuesSortCondition(columnIndex, sortOrder));
        }

        [TestMethod]
        public void SortNumbersDescendingByThirdColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex = 2;
            SortOrder sortOrder = SortOrder.Descending;

            List<string>[] values = new List<string>[]
            {
                new List<string> {"5", "3", "7", "1", "8", "9", "4", "2", "0", "6"},
                new List<string> {"1", "4", "2", "3", "5", "7", "9", "0", "8", "6"},
                new List<string> {"3", "1", "5", "8", "7", "9", "2", "4", "0", "6"},
                new List<string> {"8", "7", "0", "1", "2", "3", "4", "5", "9", "6"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string>(),
                new List<string>(),
                new List<string>(),
                new List<string>(),
            };

            expectedValues[0].AddRange((new string[] { "0", "3", "4", "5", "2", "7", "6", "8", "1", "9" }).Reverse());
            expectedValues[1].AddRange((new string[] { "8", "4", "9", "1", "0", "2", "6", "5", "3", "7" }).Reverse());
            expectedValues[2].AddRange((new string[] { "0", "1", "2", "3", "4", "5", "6", "7", "8", "9" }).Reverse());
            expectedValues[3].AddRange((new string[] { "9", "7", "4", "8", "5", "0", "6", "2", "1", "3" }).Reverse());

            this.AssertSorting(range, values, expectedValues, new ValuesSortCondition(columnIndex, sortOrder));
        }

        [TestMethod]
        public void SortNumbersAscendingByFirstThenByThirdColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex1 = 0;
            SortOrder sortOrder1 = SortOrder.Ascending;
            int columnIndex2 = 2;
            SortOrder sortOrder2 = SortOrder.Ascending;

            ValuesSortCondition sortCondition1 = new ValuesSortCondition(columnIndex1, sortOrder1);
            ValuesSortCondition sortCondition2 = new ValuesSortCondition(columnIndex2, sortOrder2);

            List<string>[] values = new List<string>[]
            {
                new List<string> {"5", "3", "7", "1", "8", "9", "4", "2", "0", "6"},
                new List<string> {"1", "4", "2", "3", "5", "7", "9", "0", "8", "6"},
                new List<string> {"4", "1", "2", "3", "4", "3", "1", "2", "3", "4"},
                new List<string> {"8", "7", "0", "1", "2", "3", "4", "5", "9", "6"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9"},
                new List<string> {"8", "3", "0", "4", "9", "1", "6", "2", "5", "7"},
                new List<string> {"3", "3", "2", "1", "1", "4", "4", "2", "4", "3"},
                new List<string> {"9", "1", "5", "7", "4", "8", "6", "0", "2", "3"}
            };

            this.AssertSorting(range, values, expectedValues, sortCondition1, sortCondition2);
        }

        [TestMethod]
        public void SortNumbersDescendingByByFirstThenByThirdColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex1 = 0;
            SortOrder sortOrder1 = SortOrder.Descending;
            int columnIndex2 = 2;
            SortOrder sortOrder2 = SortOrder.Descending;

            ValuesSortCondition sortCondition1 = new ValuesSortCondition(columnIndex1, sortOrder1);
            ValuesSortCondition sortCondition2 = new ValuesSortCondition(columnIndex2, sortOrder2);

            List<string>[] values = new List<string>[]
            {
                new List<string> {"5", "3", "7", "1", "8", "9", "4", "2", "0", "6"},
                new List<string> {"1", "4", "2", "3", "5", "7", "9", "0", "8", "6"},
                new List<string> {"4", "1", "2", "3", "4", "3", "1", "2", "3", "4"},
                new List<string> {"8", "7", "0", "1", "2", "3", "4", "5", "9", "6"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string>(),
                new List<string>(),
                new List<string>(),
                new List<string>(),
            };

            expectedValues[0].AddRange((new string[] { "0", "1", "2", "3", "4", "5", "6", "7", "8", "9" }).Reverse());
            expectedValues[1].AddRange((new string[] { "8", "3", "0", "4", "9", "1", "6", "2", "5", "7" }).Reverse());
            expectedValues[2].AddRange((new string[] { "3", "3", "2", "1", "1", "4", "4", "2", "4", "3" }).Reverse());
            expectedValues[3].AddRange((new string[] { "9", "1", "5", "7", "4", "8", "6", "0", "2", "3" }).Reverse());

            this.AssertSorting(range, values, expectedValues, sortCondition1, sortCondition2);
        }

        [TestMethod]
        public void SortNumbersAscendingByThirdThenByFirstColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex1 = 2;
            SortOrder sortOrder1 = SortOrder.Ascending;
            int columnIndex2 = 0;
            SortOrder sortOrder2 = SortOrder.Ascending;

            ValuesSortCondition sortCondition1 = new ValuesSortCondition(columnIndex1, sortOrder1);
            ValuesSortCondition sortCondition2 = new ValuesSortCondition(columnIndex2, sortOrder2);

            List<string>[] values = new List<string>[]
            {
                new List<string> {"5", "3", "7", "1", "8", "9", "4", "2", "0", "6"},
                new List<string> {"1", "4", "2", "3", "5", "7", "9", "0", "8", "6"},
                new List<string> {"4", "1", "2", "3", "4", "3", "1", "2", "3", "4"},
                new List<string> {"8", "7", "0", "1", "2", "3", "4", "5", "9", "6"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"3", "4", "2", "7", "0", "1", "9", "5", "6", "8"},
                new List<string> {"4", "9", "0", "2", "8", "3", "7", "1", "6", "5"},
                new List<string> {"1", "1", "2", "2", "3", "3", "3", "4", "4", "4"},
                new List<string> {"7", "4", "5", "0", "9", "1", "3", "8", "6", "2"}
            };

            this.AssertSorting(range, values, expectedValues, sortCondition1, sortCondition2);
        }

        [TestMethod]
        public void SortNumbersDescendingByThirdThenByFirstColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex1 = 2;
            SortOrder sortOrder1 = SortOrder.Descending;
            int columnIndex2 = 0;
            SortOrder sortOrder2 = SortOrder.Descending;

            ValuesSortCondition sortCondition1 = new ValuesSortCondition(columnIndex1, sortOrder1);
            ValuesSortCondition sortCondition2 = new ValuesSortCondition(columnIndex2, sortOrder2);

            List<string>[] values = new List<string>[]
            {
                new List<string> {"5", "3", "7", "1", "8", "9", "4", "2", "0", "6"},
                new List<string> {"1", "4", "2", "3", "5", "7", "9", "0", "8", "6"},
                new List<string> {"4", "1", "2", "3", "4", "3", "1", "2", "3", "4"},
                new List<string> {"8", "7", "0", "1", "2", "3", "4", "5", "9", "6"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string>(),
                new List<string>(),
                new List<string>(),
                new List<string>(),
            };

            expectedValues[0].AddRange((new string[] { "3", "4", "2", "7", "0", "1", "9", "5", "6", "8" }).Reverse());
            expectedValues[1].AddRange((new string[] { "4", "9", "0", "2", "8", "3", "7", "1", "6", "5" }).Reverse());
            expectedValues[2].AddRange((new string[] { "1", "1", "2", "2", "3", "3", "3", "4", "4", "4" }).Reverse());
            expectedValues[3].AddRange((new string[] { "7", "4", "5", "0", "9", "1", "3", "8", "6", "2" }).Reverse());

            this.AssertSorting(range, values, expectedValues, sortCondition1, sortCondition2);
        }

        [TestMethod]
        public void SortNumbersAscendingBySecondThenByFourthColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex1 = 1;
            SortOrder sortOrder1 = SortOrder.Ascending;
            int columnIndex2 = 3;
            SortOrder sortOrder2 = SortOrder.Ascending;

            ValuesSortCondition sortCondition1 = new ValuesSortCondition(columnIndex1, sortOrder1);
            ValuesSortCondition sortCondition2 = new ValuesSortCondition(columnIndex2, sortOrder2);

            List<string>[] values = new List<string>[]
            {
                new List<string> {"5", "3", "7", "1", "8", "9", "4", "2", "0", "6"},
                new List<string> {"2", "4", "3", "1", "4", "2", "3", "2", "1", "4"},
                new List<string> {"4", "1", "2", "3", "4", "3", "1", "2", "3", "4"},
                new List<string> {"3", "3", "1", "4", "4", "2", "1", "2", "4", "2"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"1", "0", "9", "2", "5", "7", "4", "6", "3", "8"},
                new List<string> {"1", "1", "2", "2", "2", "3", "3", "4", "4", "4"},
                new List<string> {"3", "3", "3", "2", "4", "2", "1", "4", "1", "4"},
                new List<string> {"4", "4", "2", "2", "3", "1", "1", "2", "3", "4"}
            };

            this.AssertSorting(range, values, expectedValues, sortCondition1, sortCondition2);
        }

        [TestMethod]
        public void SortNumbersDescendingBySecondThenByFourthColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex1 = 1;
            SortOrder sortOrder1 = SortOrder.Descending;
            int columnIndex2 = 3;
            SortOrder sortOrder2 = SortOrder.Descending;

            ValuesSortCondition sortCondition1 = new ValuesSortCondition(columnIndex1, sortOrder1);
            ValuesSortCondition sortCondition2 = new ValuesSortCondition(columnIndex2, sortOrder2);

            List<string>[] values = new List<string>[]
            {
                new List<string> {"5", "3", "7", "1", "8", "9", "4", "2", "0", "6"},
                new List<string> {"2", "4", "3", "1", "4", "2", "3", "2", "1", "4"},
                new List<string> {"4", "1", "2", "3", "4", "3", "1", "2", "3", "4"},
                new List<string> {"3", "3", "1", "4", "4", "2", "1", "2", "4", "2"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"8", "3", "6", "7", "4", "5", "9", "2", "1", "0"},
                new List<string> {"4", "4", "4", "3", "3", "2", "2", "2", "1", "1"},
                new List<string> {"4", "1", "4", "2", "1", "4", "3", "2", "3", "3"},
                new List<string> {"4", "3", "2", "1", "1", "3", "2", "2", "4", "4"}
            };

            this.AssertSorting(range, values, expectedValues, sortCondition1, sortCondition2);
        }

        [TestMethod]
        public void SortNumbersAscendingByFourthThenBySecondColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex1 = 3;
            SortOrder sortOrder1 = SortOrder.Ascending;
            int columnIndex2 = 1;
            SortOrder sortOrder2 = SortOrder.Ascending;

            ValuesSortCondition sortCondition1 = new ValuesSortCondition(columnIndex1, sortOrder1);
            ValuesSortCondition sortCondition2 = new ValuesSortCondition(columnIndex2, sortOrder2);

            List<string>[] values = new List<string>[]
            {
                new List<string> {"5", "3", "7", "1", "8", "9", "4", "2", "0", "6"},
                new List<string> {"2", "4", "3", "1", "4", "2", "3", "2", "1", "4"},
                new List<string> {"4", "1", "2", "3", "4", "3", "1", "2", "3", "4"},
                new List<string> {"3", "3", "1", "4", "4", "2", "1", "2", "4", "2"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"7", "4", "9", "2", "6", "5", "3", "1", "0", "8"},
                new List<string> {"3", "3", "2", "2", "4", "2", "4", "1", "1", "4"},
                new List<string> {"2", "1", "3", "2", "4", "4", "1", "3", "3", "4"},
                new List<string> {"1", "1", "2", "2", "2", "3", "3", "4", "4", "4"}
            };

            this.AssertSorting(range, values, expectedValues, sortCondition1, sortCondition2);
        }

        [TestMethod]
        public void SortNumbersDescendingByFourthThenBySecondColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex1 = 3;
            SortOrder sortOrder1 = SortOrder.Descending;
            int columnIndex2 = 1;
            SortOrder sortOrder2 = SortOrder.Descending;

            ValuesSortCondition sortCondition1 = new ValuesSortCondition(columnIndex1, sortOrder1);
            ValuesSortCondition sortCondition2 = new ValuesSortCondition(columnIndex2, sortOrder2);

            List<string>[] values = new List<string>[]
            {
                new List<string> {"5", "3", "7", "1", "8", "9", "4", "2", "0", "6"},
                new List<string> {"2", "4", "3", "1", "4", "2", "3", "2", "1", "4"},
                new List<string> {"4", "1", "2", "3", "4", "3", "1", "2", "3", "4"},
                new List<string> {"3", "3", "1", "4", "4", "2", "1", "2", "4", "2"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"8", "1", "0", "3", "5", "6", "9", "2", "7", "4"},
                new List<string> {"4", "1", "1", "4", "2", "4", "2", "2", "3", "3"},
                new List<string> {"4", "3", "3", "1", "4", "4", "3", "2", "2", "1"},
                new List<string> {"4", "4", "4", "3", "3", "2", "2", "2", "1", "1"}
            };

            this.AssertSorting(range, values, expectedValues, sortCondition1, sortCondition2);
        }

        #endregion


        #region Formulas Sort Tests

        [TestMethod]
        public void SortFormulasDescendingByFirstColumn_WhenThereAreFormulasInMultipleColumns()
        {
            CellRange valuesRange = new CellRange(0, 0, 2, 1);
            CellRange sortRange = new CellRange(0, 0, 2, 0);
            int columnIndex = 0;
            SortOrder sortOrder = SortOrder.Descending;

            List<string>[] values = new List<string>[]
            {
                new List<string> {"=B1", "=B2 + 1", "=B3 + 2"},
                new List<string> {"1", "2", "3"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string>() {"3", "3", "3"}
            };

            this.AssertSorting(sortRange, valuesRange, values, expectedValues, new ValuesSortCondition(columnIndex, sortOrder));
        }

        #endregion


        #region Text Sort Tests

        [TestMethod]
        public void SortTextValuesAscendingByFirstColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 26, 3);
            int columnIndex1 = 0;
            SortOrder sortOrder1 = SortOrder.Ascending;

            ValuesSortCondition sortCondition1 = new ValuesSortCondition(columnIndex1, sortOrder1);

            List<string>[] values = new List<string>[]
            {
                new List<string> {"220-B", "105", "220-A", "220-B", "105", "105", "220-B", "105", "220-B", "135", "110", "135", "135", "105", "110", "220-B", "135", "135", "220-A", "220-A", "105", "110", "220-B", "135", "110", "220-B", "220-A"},
                new List<string> {"Malik", "Melissa", "Juan", "Karla", "Christiana", "Sidney", "Windy", "Esther", "Tyrese", "Lia", "Kris", "Jordan", "Alex", "Nathan", "Gabriel", "Avery", "James", "Chantal", "Chevonne", "Brigid", "Derek", "Matt", "Michael", "Anisa", "Regina", "Samantha", "Christopher"},
                new List<string> {"Reynolds", "White", "Flores", "Nichols", "Chen", "Kelly", "Shaw", "Yaron", "Hanlon", "Richards", "Ackerman", "Weller", "Yuen", "Albee", "Del Toro", "Kelly", "Panarello", "Weller", "Means", "Ellison", "MacDonald", "Benson", "Lazar", "Naser", "Olivera", "Bell", "Peyton-Gomez"},
                new List<string> {"Small", "Small", "X-Large", "X-Large", "Medium", "Medium", "Small", "Small", "X-Large", "X-Large", "Large", "Large", "Large", "Medium", "Medium", "Medium", "Medium", "Medium", "Medium", "Small", "Large", "Medium", "Small", "Small", "Large", "Medium", "Small"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"105", "105", "105", "105", "105", "105", "110", "110", "110", "110", "135", "135", "135", "135", "135", "135", "220-A", "220-A", "220-A", "220-A", "220-B", "220-B", "220-B", "220-B", "220-B", "220-B", "220-B"},
                new List<string> {"Melissa", "Christiana", "Sidney", "Esther", "Nathan", "Derek", "Kris", "Gabriel", "Matt", "Regina", "Lia", "Jordan", "Alex", "James", "Chantal", "Anisa", "Juan", "Chevonne", "Brigid", "Christopher", "Malik", "Karla", "Windy", "Tyrese", "Avery", "Michael", "Samantha"},
                new List<string> {"White", "Chen", "Kelly", "Yaron", "Albee", "MacDonald", "Ackerman", "Del Toro", "Benson", "Olivera", "Richards", "Weller", "Yuen", "Panarello", "Weller", "Naser", "Flores", "Means", "Ellison", "Peyton-Gomez", "Reynolds", "Nichols", "Shaw", "Hanlon", "Kelly", "Lazar", "Bell"},
                new List<string> {"Small", "Medium", "Medium", "Small", "Medium", "Large", "Large", "Medium", "Medium", "Large", "X-Large", "Large", "Large", "Medium", "Medium", "Small", "X-Large", "Medium", "Small", "Small", "Small", "X-Large", "Small", "X-Large", "Medium", "Small", "Medium"}
            };

            this.AssertSorting(range, values, expectedValues, sortCondition1);
        }

        [TestMethod]
        public void SortTextValuesAscendingByFirstThenByFourthColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 26, 3);
            int columnIndex1 = 0;
            SortOrder sortOrder1 = SortOrder.Ascending;
            int columnIndex2 = 3;
            SortOrder sortOrder2 = SortOrder.Ascending;

            ValuesSortCondition sortCondition1 = new ValuesSortCondition(columnIndex1, sortOrder1);
            ValuesSortCondition sortCondition2 = new ValuesSortCondition(columnIndex2, sortOrder2);

            List<string>[] values = new List<string>[]
            {
                new List<string> {"220-B", "105", "220-A", "220-B", "105", "105", "220-B", "105", "220-B", "135", "110", "135", "135", "105", "110", "220-B", "135", "135", "220-A", "220-A", "105", "110", "220-B", "135", "110", "220-B", "220-A"},
                new List<string> {"Malik", "Melissa", "Juan", "Karla", "Christiana", "Sidney", "Windy", "Esther", "Tyrese", "Lia", "Kris", "Jordan", "Alex", "Nathan", "Gabriel", "Avery", "James", "Chantal", "Chevonne", "Brigid", "Derek", "Matt", "Michael", "Anisa", "Regina", "Samantha", "Christopher"},
                new List<string> {"Reynolds", "White", "Flores", "Nichols", "Chen", "Kelly", "Shaw", "Yaron", "Hanlon", "Richards", "Ackerman", "Weller", "Yuen", "Albee", "Del Toro", "Kelly", "Panarello", "Weller", "Means", "Ellison", "MacDonald", "Benson", "Lazar", "Naser", "Olivera", "Bell", "Peyton-Gomez"},
                new List<string> {"Small", "Small", "X-Large", "X-Large", "Medium", "Medium", "Small", "Small", "X-Large", "X-Large", "Large", "Large", "Large", "Medium", "Medium", "Medium", "Medium", "Medium", "Medium", "Small", "Large", "Medium", "Small", "Small", "Large", "Medium", "Small"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"105", "105", "105", "105", "105", "105", "110", "110", "110", "110", "135", "135", "135", "135", "135", "135", "220-A", "220-A", "220-A", "220-A", "220-B", "220-B", "220-B", "220-B", "220-B", "220-B", "220-B"},
                new List<string> {"Derek", "Christiana", "Sidney", "Nathan", "Melissa", "Esther", "Kris", "Regina", "Gabriel", "Matt", "Jordan", "Alex", "James", "Chantal", "Anisa", "Lia", "Chevonne", "Brigid", "Christopher", "Juan", "Avery", "Samantha", "Malik", "Windy", "Michael", "Karla", "Tyrese"},
                new List<string> {"MacDonald", "Chen", "Kelly", "Albee", "White", "Yaron", "Ackerman", "Olivera", "Del Toro", "Benson", "Weller", "Yuen", "Panarello", "Weller", "Naser", "Richards", "Means", "Ellison", "Peyton-Gomez", "Flores", "Kelly", "Bell", "Reynolds", "Shaw", "Lazar", "Nichols", "Hanlon"},
                new List<string> {"Large", "Medium", "Medium", "Medium", "Small", "Small", "Large", "Large", "Medium", "Medium", "Large", "Large", "Medium", "Medium", "Small", "X-Large", "Medium", "Small", "Small", "X-Large", "Medium", "Medium", "Small", "Small", "Small", "X-Large", "X-Large"}
            };

            this.AssertSorting(range, values, expectedValues, sortCondition1, sortCondition2);
        }

        #endregion


        #region Mixed values Sort Tests

        [TestMethod]
        public void SortNumbersAndTextValuesAscending_WhenThereAreValuesInSingleColumn()
        {
            CellRange range = new CellRange(0, 0, 9, 0);
            int columnIndex1 = 0;
            SortOrder sortOrder1 = SortOrder.Ascending;

            ValuesSortCondition sortCondition1 = new ValuesSortCondition(columnIndex1, sortOrder1);

            List<string>[] values = new List<string>[]
            {
                new List<string> { "3", "2", "1", "5", "4", "=\"3\"", "=\"2\"", "=\"1\"", "=\"5\"", "=\"4\"" }
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"1", "2", "3", "4", "5", "1", "2", "3", "4", "5", },
            };

            this.AssertSorting(range, values, expectedValues, sortCondition1);
        }

        [TestMethod]
        public void SortNumbersAndTextValuesDescending_WhenThereAreValuesInSingleColumn()
        {
            CellRange range = new CellRange(0, 0, 9, 0);
            int columnIndex1 = 0;
            SortOrder sortOrder1 = SortOrder.Descending;

            ValuesSortCondition sortCondition1 = new ValuesSortCondition(columnIndex1, sortOrder1);

            List<string>[] values = new List<string>[]
            {
                new List<string> { "3", "2", "1", "5", "4", "=\"3\"", "=\"2\"", "=\"1\"", "=\"5\"", "=\"4\"" }
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string>( (new string[] {"1", "2", "3", "4", "5", "1", "2", "3", "4", "5", }).Reverse())
            };

            this.AssertSorting(range, values, expectedValues, sortCondition1);
        }

        [TestMethod]
        public void SortMixedValuesAscending_WhenThereAreValuesInSingleColumn()
        {
            CellRange range = new CellRange(0, 0, 15, 0);
            int columnIndex1 = 0;
            SortOrder sortOrder1 = SortOrder.Ascending;

            ValuesSortCondition sortCondition1 = new ValuesSortCondition(columnIndex1, sortOrder1);

            List<string>[] values = new List<string>[]
            {
                new List<string> { "3", "2", "1", "5", "4", "FALSE", "=5/0", "", "=#REF!", "=\"3\"", "=\"2\"", "=\"1\"", "=\"5\"", "=\"4\"", "TRUE", "=4/0" }
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"1", "2", "3", "4", "5", "1", "2", "3", "4", "5", "FALSE", "TRUE", "#DIV/0!", "#REF!", "#DIV/0!", "" },
            };

            this.AssertSorting(range, values, expectedValues, sortCondition1);
        }

        [TestMethod]
        public void SortMixedValuesDescending_WhenThereAreValuesInSingleColumn()
        {
            CellRange range = new CellRange(0, 0, 15, 0);
            int columnIndex1 = 0;
            SortOrder sortOrder1 = SortOrder.Descending;

            ValuesSortCondition sortCondition1 = new ValuesSortCondition(columnIndex1, sortOrder1);

            List<string>[] values = new List<string>[]
            {
                new List<string> { "3", "2", "1", "5", "4", "FALSE", "=5/0", "", "=#REF!", "=\"3\"", "=\"2\"", "=\"1\"", "=\"5\"", "=\"4\"", "TRUE", "=4/0" }
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string>((new string[] {"1", "2", "3", "4", "5", "1", "2", "3", "4", "5", "FALSE", "TRUE", "#DIV/0!", "#REF!", "#DIV/0!" }).Reverse())
            };

            expectedValues[0].Add(string.Empty);

            this.AssertSorting(range, values, expectedValues, sortCondition1);
        }

        #endregion


        #region Sorting affecting cell sizes Tests

        [TestMethod]
        public void SortNumbersAscendingCheckCellSizes_WhenThereAreThreeValuesInSingleColumn()
        {
            CellRange range = new CellRange(0, 0, 2, 0);
            int columnIndex1 = 0;
            SortOrder sortOrder1 = SortOrder.Ascending;

            ValuesSortCondition sortCondition1 = new ValuesSortCondition(columnIndex1, sortOrder1);

            List<string>[] values = new List<string>[]
            {
                new List<string> { "3", "2", "1" }
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"1", "2", "3" },
            };

            this.InitializeValues(range, values);

            double[] rowHeights = new double[]
            {
                35, 40, 45
            };

            double[] expectedRowHeights = new double[]
            {
                45, 40, 35
            };

            this.SetRowHeights(range, rowHeights);
            this.AssertSorting(range, values, expectedValues, sortCondition1);
            this.AssertRowHeights(range, expectedRowHeights);
        }

        [TestMethod]
        public void SortNumbersAscendingCheckCellSizes_WhenThereAreSevenValuesInSingleColumn()
        {
            CellRange range = new CellRange(0, 0, 5, 0);
            int columnIndex1 = 0;
            SortOrder sortOrder1 = SortOrder.Ascending;

            ValuesSortCondition sortCondition1 = new ValuesSortCondition(columnIndex1, sortOrder1);

            List<string>[] values = new List<string>[]
            {
                new List<string> { "5", "7", "3", "1", "2", "6" }
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"1", "2", "3", "5", "6", "7"},
            };

            this.InitializeValues(range, values);

            double[] rowHeights = new double[]
            {
                50, 70, 30, 10, 20, 60
            };

            double[] expectedRowHeights = new double[]
            {
                10, 20, 30, 50, 60, 70
            };

            this.SetRowHeights(range, rowHeights);
            this.AssertSorting(range, values, expectedValues, sortCondition1);
            this.AssertRowHeights(range, expectedRowHeights);
        }

        #endregion


        #region Sorting cells containing hyperlinks Tests



        #endregion


        #region Sorting cells containing merged cells Tests



        #endregion


        #region Sorting Undo / Redo

        [TestMethod]
        public void SortNumbersAscending_UndoCorrectly()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex = 0;
            SortOrder sortOrder = SortOrder.Ascending;

            List<string>[] values = new List<string>[]
            {
                new List<string> {"5", "3", "7", "1", "8", "9", "4", "2", "0", "6"},
                new List<string> {"1", "4", "2", "3", "5", "7", "9", "0", "8", "6"},
                new List<string> {"3", "1", "5", "8", "7", "9", "2", "4", "0", "6"},
                new List<string> {"8", "7", "0", "1", "2", "3", "4", "5", "9", "6"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9"},
                new List<string> {"8", "3", "0", "4", "9", "1", "6", "2", "5", "7"},
                new List<string> {"0", "8", "4", "1", "2", "3", "6", "5", "7", "9"},
                new List<string> {"9", "1", "5", "7", "4", "8", "6", "0", "2", "3"}
            };

            this.AssertSorting(range, values, expectedValues, new ValuesSortCondition(columnIndex, sortOrder));

            this.Worksheet.Workbook.History.Undo();

            this.AssertValues(range, values);

            this.Worksheet.Workbook.History.Redo();

            this.AssertValues(range, expectedValues);
        }

        [TestMethod]
        public void ClearSortState_UndoCorrectly()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex = 0;
            SortOrder sortOrder = SortOrder.Ascending;

            List<string>[] values = new List<string>[]
            {
                new List<string> {"5", "3", "7", "1", "8", "9", "4", "2", "0", "6"},
                new List<string> {"1", "4", "2", "3", "5", "7", "9", "0", "8", "6"},
                new List<string> {"3", "1", "5", "8", "7", "9", "2", "4", "0", "6"},
                new List<string> {"8", "7", "0", "1", "2", "3", "4", "5", "9", "6"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9"},
                new List<string> {"8", "3", "0", "4", "9", "1", "6", "2", "5", "7"},
                new List<string> {"0", "8", "4", "1", "2", "3", "6", "5", "7", "9"},
                new List<string> {"9", "1", "5", "7", "4", "8", "6", "0", "2", "3"}
            };

            ValuesSortCondition sortCondition = new ValuesSortCondition(columnIndex, sortOrder);
            this.AssertSorting(range, values, expectedValues, sortCondition);

            this.Worksheet.SortState.Clear();

            Assert.IsNull(this.Worksheet.SortState.SortRange);
            Assert.AreEqual(0, this.Worksheet.SortState.SortConditions.Count());

            this.Worksheet.Workbook.History.Undo();

            Assert.AreEqual(range, this.Worksheet.SortState.SortRange);
            Assert.AreEqual(sortCondition, this.Worksheet.SortState.SortConditions.First());

            this.Worksheet.Workbook.History.Redo();

            Assert.IsNull(this.Worksheet.SortState.SortRange);
            Assert.AreEqual(0, this.Worksheet.SortState.SortConditions.Count());
        }

        // TODO: Test SortRange undo

        #endregion


        #region Duplicating Sort Conditions

        [TestMethod]
        [ExpectedException(typeof(SortingException))]
        public void SortThrowsException_WhenDpulicatingValuesSortConditionsAreApplied()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex = 0;
            SortOrder sortOrder = SortOrder.Ascending;

            var sortCondition = new ValuesSortCondition(columnIndex, sortOrder);
            var sortCondition1 = new ValuesSortCondition(columnIndex, sortOrder);

            this.Worksheet.Cells[range].Sort(sortCondition, sortCondition1);
        }

        [TestMethod]
        [ExpectedException(typeof(SortingException))]
        public void SortThrowsException_WhenDpulicatingCustomSortConditionsAreApplied()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            int columnIndex = 0;
            SortOrder sortOrder = SortOrder.Ascending;
            SortOrder sortOrder1 = SortOrder.Ascending;

            CustomValuesSortCondition sortCondition = new CustomValuesSortCondition(
                columnIndex,
                new string[]
                {
                    "Small",
                    "Medium",
                    "Large",
                    "X-Large"
                }, sortOrder);
            CustomValuesSortCondition sortCondition1 = new CustomValuesSortCondition(
                columnIndex,
                new string[]
                {
                    "Small",
                    "Medium",
                    "Large",
                    "X-Large"
                }, sortOrder1);

            this.Worksheet.Cells[range].Sort(sortCondition, sortCondition1);
        }

        #endregion


        #region Sort multiple ranges

        [TestMethod]
        [ExpectedException(typeof(SortingException))]
        public void SortThrowsException_MultipleRangesAreSorted()
        {
            CellRange range = new CellRange(0, 0, 9, 3);
            CellRange range1 = new CellRange(0, 0, 5, 3);
            SortOrder sortOrder1 = SortOrder.Ascending;
            int columnIndex = 0;

            CustomValuesSortCondition sortCondition = new CustomValuesSortCondition(
                columnIndex,
                new string[]
                {
                    "Small",
                    "Medium",
                    "Large",
                    "X-Large"
                }, sortOrder1);


            this.Worksheet.Cells.GetCellSelection(new CellRange[] {range, range1}).Sort(sortCondition);
        }

        #endregion


        #region Custom Sort Tests

        [TestMethod]
        public void SortTextValuesAscendingByFirstThenCustomSortByFourthColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 26, 3);
            int columnIndex1 = 0;
            SortOrder sortOrder1 = SortOrder.Ascending;
            SortOrder sortOrder2 = SortOrder.Ascending;
            int columnIndex2 = 3;

            ValuesSortCondition sortCondition1 = new ValuesSortCondition(columnIndex1, sortOrder1);
            CustomValuesSortCondition sortCondition2 = new CustomValuesSortCondition(
                columnIndex2,
                new string[]
                {
                    "Small",
                    "Medium",
                    "Large",
                    "X-Large"
                }, sortOrder2);

            List<string>[] values = new List<string>[]
            {
                new List<string> {"220-B", "105", "220-A", "220-B", "105", "105", "220-B", "105", "220-B", "135", "110", "135", "135", "105", "110", "220-B", "135", "135", "220-A", "220-A", "105", "110", "220-B", "135", "110", "220-B", "220-A"},
                new List<string> {"Malik", "Melissa", "Juan", "Karla", "Christiana", "Sidney", "Windy", "Esther", "Tyrese", "Lia", "Kris", "Jordan", "Alex", "Nathan", "Gabriel", "Avery", "James", "Chantal", "Chevonne", "Brigid", "Derek", "Matt", "Michael", "Anisa", "Regina", "Samantha", "Christopher"},
                new List<string> {"Reynolds", "White", "Flores", "Nichols", "Chen", "Kelly", "Shaw", "Yaron", "Hanlon", "Richards", "Ackerman", "Weller", "Yuen", "Albee", "Del Toro", "Kelly", "Panarello", "Weller", "Means", "Ellison", "MacDonald", "Benson", "Lazar", "Naser", "Olivera", "Bell", "Peyton-Gomez"},
                new List<string> {"Small", "Small", "X-Large", "X-Large", "Medium", "Medium", "Small", "Small", "X-Large", "X-Large", "Large", "Large", "Large", "Medium", "Medium", "Medium", "Medium", "Medium", "Medium", "Small", "Large", "Medium", "Small", "Small", "Large", "Medium", "Small"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"105", "105", "105", "105", "105", "105", "110", "110", "110", "110", "135", "135", "135", "135", "135", "135", "220-A", "220-A", "220-A", "220-A", "220-B", "220-B", "220-B", "220-B", "220-B", "220-B", "220-B"},
                new List<string> {"Melissa", "Esther", "Christiana", "Sidney", "Nathan", "Derek", "Gabriel", "Matt", "Kris", "Regina", "Anisa", "James", "Chantal", "Jordan", "Alex", "Lia", "Brigid", "Christopher", "Chevonne", "Juan", "Malik", "Windy", "Michael", "Avery", "Samantha", "Karla", "Tyrese"},
                new List<string> {"White", "Yaron", "Chen", "Kelly", "Albee", "MacDonald", "Del Toro", "Benson", "Ackerman", "Olivera", "Naser", "Panarello", "Weller", "Weller", "Yuen", "Richards", "Ellison", "Peyton-Gomez", "Means", "Flores", "Reynolds", "Shaw", "Lazar", "Kelly", "Bell", "Nichols", "Hanlon"},
                new List<string> {"Small", "Small", "Medium", "Medium", "Medium", "Large", "Medium", "Medium", "Large", "Large", "Small", "Medium", "Medium", "Large", "Large", "X-Large", "Small", "Small", "Medium", "X-Large", "Small", "Small", "Small", "Medium", "Medium", "X-Large", "X-Large"}
            };

            this.AssertSorting(range, values, expectedValues, sortCondition1, sortCondition2);
        }

        [TestMethod]
        public void SortTextValuesAscendingByFirstThenCustomSortDescendingByFourthColumn_WhenThereAreValuesInMultipleColumns()
        {
            CellRange range = new CellRange(0, 0, 3, 0);
            SortOrder sortOrder = SortOrder.Descending;
            int columnIndex = 0;

            CustomValuesSortCondition sortCondition = new CustomValuesSortCondition(
                columnIndex,
                new string[]
                {
                    "Small",
                    "Medium",
                    "Large",
                    "X-Large"
                }, sortOrder);

            List<string>[] values = new List<string>[]
            {
                new List<string> {"Small", "Medium", "Large", "X-Large"}
            };

            List<string>[] expectedValues = new List<string>[]
            {
                new List<string> {"X-Large", "Large", "Medium", "Small"}
            };

            this.AssertSorting(range, values, expectedValues, sortCondition);
        }

        #endregion
    }
}