﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests.Utilities.Charts
{
    internal abstract class SafeEqualityComparerBase<T> : EqualityComparer<T>
    {
        private static readonly Dictionary<string, Func<T, T, bool>> propertyNameToEqualityComparerFunction = new Dictionary<string, Func<T, T, bool>>();
        private static readonly Dictionary<string, IEqualityComparer> typeNameToEqualityComparer = new Dictionary<string, IEqualityComparer>();

        protected static void RegisterComparerFunction(string propertyName, Func<T, T, bool> func)
        {
            propertyNameToEqualityComparerFunction[propertyName] = func;
        }

        protected static void RegisterConcreteComparer<K>(IEqualityComparer<K> comparer)
        {
            var type = typeof(K);
            typeNameToEqualityComparer[type.FullName] = (IEqualityComparer)comparer;
        }

        public sealed override bool Equals(T x, T y)
        {
            if (x == null && y == null)
            {
                return true;
            }

            if (x == null || y == null)
            {
                return false;
            }

            Type type = typeof(T);
            string genericTypeFullName = type.FullName;
            string xParameterFullName = x.GetType().FullName;
            if (genericTypeFullName != xParameterFullName)
            {
                var yParameterFullName = y.GetType().FullName;
                if (xParameterFullName != yParameterFullName)
                {
#if DEBUG
                    System.Diagnostics.Debug.Assert(false);
#endif

                    return false;
                }

                IEqualityComparer comparer = typeNameToEqualityComparer[xParameterFullName];

                return comparer.Equals(x, y);
            }

            bool result = true;

            PropertyInfo[] properties = type.GetProperties(BindingFlags.Public | BindingFlags.Instance);
            for (int i = 0; i < properties.Length; i++)
            {
                PropertyInfo property = properties[i];
                string propertyName = property.Name;

                Func<T, T, bool> comparerFunction = this.GetComparerFunction(propertyName);

                Guard.ThrowExceptionIfNull(comparerFunction, "comparerFunction");

                result &= comparerFunction(x, y);

#if DEBUG
                System.Diagnostics.Debug.Assert(result);
#endif
            }

            return result;
        }

        public sealed override int GetHashCode(T obj)
        {
            return obj.GetHashCode();
        }

        private Func<T, T, bool> GetComparerFunction(string propertyName)
        {
            Func<T, T, bool> func = null;

            if (propertyNameToEqualityComparerFunction.TryGetValue(propertyName, out func))
            {
                return func;
            }

            return null;
        }
    }
}
