﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Windows;
using System.Windows.Media;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Expressions.Functions;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.DataValidation;
using Telerik.Windows.Documents.Spreadsheet.Model.Filtering;
using Telerik.Windows.Documents.Spreadsheet.Model.Printing;
using Telerik.Windows.Documents.Spreadsheet.Model.Shapes;
using Telerik.Windows.Documents.Spreadsheet.Model.Sorting;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Tests.ImportExportTests;
using Telerik.Windows.Documents.Spreadsheet.Theming;
using Telerik.Windows.Documents.Spreadsheet.Utilities;
using System.Text.RegularExpressions;
#if !SILVERLIGHT
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities.Charts;
using Telerik.Windows.Documents.Model.Drawing.Charts;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.Tests.Utilities
{
    public static class TestHelper
    {
        #region Constants

        internal static string FractionsOne = "1";
        internal static string FractionsTwo = "11";
        internal static string FractionsThree = "111";
        internal static FontProperties DefaultFontProperties = new FontProperties(new FontFamily("Calibri"), 14.666666, false);
        internal const double Epsilon = 1E-8;
        private const string TestResourcesFolderName = "TestDataResources";

        #endregion


        #region Constructors

        static TestHelper()
        {
            WorkbookFormatProvidersManager.RegisterFormatProvider(new XlsxFormatProvider());
        }

        #endregion

        #region Helper Methods

        public static Uri GetResourceUri(string resource, string assemblyFullName = null)
        {
            string currentAssemblyFullName = string.IsNullOrEmpty(assemblyFullName) ? typeof(TestHelper).Assembly.FullName : assemblyFullName;

            AssemblyName assemblyName = new AssemblyName(currentAssemblyFullName);
            string resourcePath = "/" + assemblyName.Name + ";component/" + resource;
            Uri resourceUri = new Uri(resourcePath, UriKind.Relative);

            return resourceUri;
        }

        public static Stream GetResourceStream(string resource, string assemblyFullName = null)
        {
            return Application.GetResourceStream(TestHelper.GetResourceUri(resource, assemblyFullName)).Stream;
        }

        public static Stream GetStreamFromFile(string fileName, string assemblyFullName = null)
        {
            string path = TestResourcesFolderName + "/" + fileName;

            return TestHelper.GetResourceStream(path, assemblyFullName);
        }

        public static Workbook GetWorkbookFromFile(string fileName, string assemblyFullName = null)
        {
            string path = TestResourcesFolderName + "/" + fileName;

            Workbook workbook;
            using (Stream stream = TestHelper.GetResourceStream(path, assemblyFullName))
            {
                string extension = Path.GetExtension(path);
                IWorkbookFormatProvider formatProvider = WorkbookFormatProvidersManager.GetProviderByExtension(extension);

                workbook = formatProvider.Import(stream);
            }

            return workbook;
        }

        public static Workbook CreateNonEmptyWorkbook()
        {
            Workbook workbook = new Workbook();
            workbook.History.IsEnabled = true;
            workbook.Sheets.Add(SheetType.Worksheet);
            return workbook;
        }

        public static Worksheet CreateWorksheet()
        {
            Workbook workbook = CreateNonEmptyWorkbook();
            return workbook.ActiveWorksheet;
        }

        public static void AssertWorkbooksAreEqual(Workbook workbook1, Workbook workbook2)
        {
            Assert.AreEqual(workbook1.ActiveSheet.Name, workbook2.ActiveSheet.Name);
            Assert.AreEqual(workbook1.Name, workbook2.Name);
            Assert.AreEqual(workbook1.IsProtected, workbook2.IsProtected);

            AssertNamesAreEqual(workbook1.Names, workbook2.Names);
            AssertSheetsAreEqual(workbook1.Sheets, workbook2.Sheets);
            AssertStylesAreEqual(workbook1.Styles, workbook2.Styles);
            AssertThemesAreEqual(workbook1.Theme, workbook2.Theme);
            AssertProtectionInfosAreEqual(workbook1, workbook2);
            AssertWorksheetsAreEqual(workbook1.Worksheets, workbook2.Worksheets);
        }

        internal static void AssertCellRangeCollectionsAreEqual(IEnumerable<CellRange> firstCollection, IEnumerable<CellRange> secondCollection)
        {
            CellRange[] firstRanges = firstCollection.ToArray();
            CellRange[] secondRanges = secondCollection.ToArray();

            Assert.AreEqual(firstRanges.Length, secondRanges.Length);

            for (int index = 0; index < firstRanges.Length; index++)
            {
                CellRange firstRange = firstRanges[index];
                CellRange secondRange = secondRanges[index];

                Assert.AreEqual(firstRange, secondRange);
            }
        }

        /// <summary>
        /// Asserts the themes are equal.
        /// </summary>
        /// <param name="theme1">The theme1.</param>
        /// <param name="theme2">The theme2.</param>
        private static void AssertThemesAreEqual(DocumentTheme theme1, DocumentTheme theme2)
        {
            Assert.AreEqual(theme1.ColorScheme, theme2.ColorScheme);
            Assert.AreEqual(theme1.FontScheme, theme2.FontScheme);
        }

        private static void AssertProtectionInfosAreEqual(Workbook workbook1, Workbook workbook2)
        {
            Assert.AreEqual(workbook1.IsProtected, workbook2.IsProtected);
        }

        /// <summary>
        /// Asserts the worksheets are equal.
        /// </summary>
        /// <param name="worksheets1">The worksheets1.</param>
        /// <param name="worksheets2">The worksheets2.</param>
        private static void AssertWorksheetsAreEqual(WorksheetCollection worksheets1, WorksheetCollection worksheets2)
        {
            List<Worksheet> worksheetsList1 = worksheets1.ToList();
            List<Worksheet> worksheetsList2 = worksheets2.ToList();

            Assert.AreEqual(worksheets1.Count, worksheets2.Count);

            for (int i = 0; i < worksheetsList1.Count; i++)
            {
                AssertWorksheetsPropertiesAreEqual(worksheetsList1[i], worksheetsList2[i]);
            }
        }

        /// <summary>
        /// Asserts the styles are equal.
        /// </summary>
        /// <param name="styles1">The styles1.</param>
        /// <param name="styles2">The styles2.</param>
        private static void AssertStylesAreEqual(CellStyleCollection styles1, CellStyleCollection styles2)
        {
            List<CellStyle> stylesList1 = styles1.ToList();
            List<CellStyle> stylesList2 = styles2.ToList();

            Assert.AreEqual(stylesList1.Count, stylesList2.Count);

            for (int i = 0; i < stylesList1.Count; i++)
            {
                Assert.AreEqual(stylesList1[i], stylesList2[i]);
            }
        }

        /// <summary>
        /// Asserts the sheets are equal.
        /// </summary>
        /// <param name="sheets1">The sheets1.</param>
        /// <param name="sheets2">The sheets2.</param>
        private static void AssertSheetsAreEqual(SheetCollection sheets1, SheetCollection sheets2)
        {
            List<Sheet> sheetsList1 = sheets1.ToList();
            List<Sheet> sheetsList2 = sheets2.ToList();

            Assert.AreEqual(sheetsList1.Count, sheetsList2.Count);

            for (int i = 0; i < sheetsList1.Count; i++)
            {
                Assert.AreEqual(sheetsList1[i].Name, sheetsList2[i].Name);
                Assert.AreEqual(sheetsList1[i].Type, sheetsList2[i].Type);
                Assert.AreEqual(sheetsList1[i].Visibility, sheetsList2[i].Visibility);
                Assert.AreEqual(sheetsList1[i].IsProtected, sheetsList2[i].IsProtected);
            }
        }

        internal static void AssertWorksheetsPropertiesAreEqual(Worksheet worksheet1, Worksheet worksheet2, bool assertSheetNames = true)
        {
            WorksheetViewState worksheetViewState1 = (WorksheetViewState)((ISheet)worksheet1).ViewState;
            WorksheetViewState worksheetViewState2 = (WorksheetViewState)((ISheet)worksheet2).ViewState;

            if (assertSheetNames)
            {
                Assert.AreEqual(worksheet1.Name, worksheet2.Name);
            }

            AssertWorksheetViewStatesAreEqual(worksheetViewState1, worksheetViewState2);

            AssertWorksheetRowsAreEqual(worksheet1.Rows, worksheet2.Rows);
            AssertWorksheetColumnsAreEqual(worksheet1.Columns, worksheet2.Columns);

            Assert.AreEqual(worksheet1.DefaultColumnWidth, worksheet2.DefaultColumnWidth);
            Assert.AreEqual(worksheet1.DefaultRowHeight, worksheet2.DefaultRowHeight);

            AssertWorksheetHyperlinksAreEqual(worksheet1.Hyperlinks, worksheet2.Hyperlinks);

            AssertWorksheetShapesAreEqual(worksheet1.Shapes, worksheet2.Shapes);

            AssertNamesAreEqual(worksheet1.Names, worksheet2.Names);

            Assert.AreEqual(worksheet1.Type, worksheet2.Type);
            Assert.AreEqual(worksheet1.UsedCellRange, worksheet2.UsedCellRange);
            Assert.AreEqual(worksheet1.Visibility, worksheet2.Visibility);

            CellRange usedCellRange1 = worksheet1.UsedCellRange;
            CellRange usedCellRange2 = worksheet2.UsedCellRange;

            Assert.AreEqual(usedCellRange1, usedCellRange2);

            Assert.AreEqual(worksheet1.Cells.ColumnCount, worksheet2.Cells.ColumnCount);
            Assert.AreEqual(worksheet1.Cells.RowCount, worksheet2.Cells.RowCount);

            AssertWorksheetMergedRangesAreEqual(worksheet1.Cells.GetMergedCellRanges(), worksheet2.Cells.GetMergedCellRanges());

            for (int row = 0; row < usedCellRange1.RowCount; row++)
            {
                for (int column = 0; column < usedCellRange1.ColumnCount; column++)
                {
                    CellSelection cellSelection1 = worksheet1.Cells[row, column];
                    CellSelection cellSelection2 = worksheet2.Cells[row, column];

                    AssertPropertiesAreEqual(cellSelection1, cellSelection2);
                }
            }

            AssertWorksheetPageSetupsAreEqual(worksheet1.WorksheetPageSetup, worksheet2.WorksheetPageSetup);

            Assert.AreEqual(worksheet1.IsProtected, worksheet2.IsProtected);
            Assert.AreEqual(worksheet1.ProtectionOptions, worksheet2.ProtectionOptions);

            AssertFiltersAreEqual(worksheet1.Filter, worksheet2.Filter);
            AssertSortStatesAreEqual(worksheet1.SortState, worksheet2.SortState);
            AssertGroupingPropertiesAreEqual(worksheet1.GroupingProperties, worksheet2.GroupingProperties);
        }

        internal static void AssertHeaderFooterSectionsTextAreEqual(string headerFooterSectionText1, string headerFooterSectionText2)
        {
            string[] lines1 = GetLines(headerFooterSectionText1);
            string[] lines2 = GetLines(headerFooterSectionText2);

            Assert.AreEqual(lines1.Length, lines2.Length);

            for (int index = 0; index < lines1.Length; index++)
            {
                string line1 = lines1[index];
                string line2 = lines2[index];

                Assert.AreEqual(line1, line2);
            }
        }

        private static void AssertWorksheetPageSetupsAreEqual(WorksheetPageSetup worksheetPageSetup1, WorksheetPageSetup worksheetPageSetup2)
        {
            AssertPageBreaksAreEqual(worksheetPageSetup1.PageBreaks, worksheetPageSetup2.PageBreaks);
            AssertPrintAreasAreEqual(worksheetPageSetup1.PrintArea, worksheetPageSetup2.PrintArea);
            AssertWorksheetPrintOptionsAreEqual(worksheetPageSetup1.PrintOptions, worksheetPageSetup2.PrintOptions);
            AssertHeaderFooterSettingsAreEqual(worksheetPageSetup1.HeaderFooterSettings, worksheetPageSetup2.HeaderFooterSettings);

            Assert.AreEqual(worksheetPageSetup1.CenterHorizontally, worksheetPageSetup2.CenterHorizontally);
            Assert.AreEqual(worksheetPageSetup1.CenterVertically, worksheetPageSetup2.CenterVertically);
            Assert.AreEqual(worksheetPageSetup1.Margins, worksheetPageSetup2.Margins);
            Assert.AreEqual(worksheetPageSetup1.PageOrder, worksheetPageSetup2.PageOrder);
            Assert.AreEqual(worksheetPageSetup1.PageOrientation, worksheetPageSetup2.PageOrientation);
            Assert.AreEqual(worksheetPageSetup1.PaperType, worksheetPageSetup2.PaperType);
            Assert.AreEqual(worksheetPageSetup1.PageSize, worksheetPageSetup2.PageSize);
            Assert.AreEqual(worksheetPageSetup1.ScaleFactor, worksheetPageSetup2.ScaleFactor);
            Assert.AreEqual(worksheetPageSetup1.FitToPages, worksheetPageSetup2.FitToPages);
            Assert.AreEqual(worksheetPageSetup1.FitToPagesWide, worksheetPageSetup2.FitToPagesWide);
            Assert.AreEqual(worksheetPageSetup1.FitToPagesTall, worksheetPageSetup2.FitToPagesTall);
        }

        private static void AssertHeaderFooterSettingsAreEqual(HeaderFooterSettings headerFooterSettings1, HeaderFooterSettings headerFooterSettings2)
        {
            Assert.AreEqual(headerFooterSettings1.DifferentFirstPage, headerFooterSettings2.DifferentFirstPage);
            Assert.AreEqual(headerFooterSettings1.DifferentOddAndEvenPages, headerFooterSettings2.DifferentOddAndEvenPages);
            Assert.AreEqual(headerFooterSettings1.ScaleWithDocument, headerFooterSettings2.ScaleWithDocument);
            Assert.AreEqual(headerFooterSettings1.AlignWithPageMargins, headerFooterSettings2.AlignWithPageMargins);

            AssertHeaderFooterContentAreEqual(headerFooterSettings1.Header, headerFooterSettings2.Header);
            AssertHeaderFooterContentAreEqual(headerFooterSettings1.Footer, headerFooterSettings2.Footer);
            AssertHeaderFooterContentAreEqual(headerFooterSettings1.FirstPageHeader, headerFooterSettings2.FirstPageHeader);
            AssertHeaderFooterContentAreEqual(headerFooterSettings1.FirstPageFooter, headerFooterSettings2.FirstPageFooter);
            AssertHeaderFooterContentAreEqual(headerFooterSettings1.EvenPageHeader, headerFooterSettings2.EvenPageHeader);
            AssertHeaderFooterContentAreEqual(headerFooterSettings1.EvenPageFooter, headerFooterSettings2.EvenPageFooter);
        }

        private static void AssertHeaderFooterContentAreEqual(HeaderFooterContent headerFooterContent1, HeaderFooterContent headerFooterContent2)
        {
            AssertHeaderFooterSectionsAreEqual(headerFooterContent1.LeftSection, headerFooterContent2.LeftSection);
            AssertHeaderFooterSectionsAreEqual(headerFooterContent1.CenterSection, headerFooterContent2.CenterSection);
            AssertHeaderFooterSectionsAreEqual(headerFooterContent1.RightSection, headerFooterContent2.RightSection);
        }

        private static void AssertHeaderFooterSectionsAreEqual(HeaderFooterSection headerFooterSection1, HeaderFooterSection headerFooterSection2)
        {
            TestHelper.AssertHeaderFooterSectionsTextAreEqual(headerFooterSection1.Text, headerFooterSection2.Text);

            // TODO: Add section image comparison.
        }

        private static string[] GetLines(string text)
        {
            return Regex.Split(text, "\r\n|\r|\n");
        }

        private static void AssertWorksheetPrintOptionsAreEqual(WorksheetPrintOptions worksheetPrintOptions1, WorksheetPrintOptions worksheetPrintOptions2)
        {
            Assert.AreEqual(worksheetPrintOptions1.PrintGridlines, worksheetPrintOptions2.PrintGridlines);
            Assert.AreEqual(worksheetPrintOptions1.PrintRowColumnHeadings, worksheetPrintOptions2.PrintRowColumnHeadings);
        }

        private static void AssertPageBreaksAreEqual(PageBreaks pageBreaks1, PageBreaks pageBreaks2)
        {
            IEnumerable<PageBreakInfo> firstHorizontalBreaks = PageBreakInfo.GetPageBreakInfos(pageBreaks1.HorizontalPageBreaks);
            IEnumerable<PageBreakInfo> secondHorizontalBreaks = PageBreakInfo.GetPageBreakInfos(pageBreaks2.HorizontalPageBreaks);

            Assert.IsTrue(PageBreakInfo.AreEqualPageBreaksCollections(firstHorizontalBreaks, secondHorizontalBreaks));

            IEnumerable<PageBreakInfo> firstVerticalBreaks = PageBreakInfo.GetPageBreakInfos(pageBreaks1.VerticalPageBreaks);
            IEnumerable<PageBreakInfo> secondVerticalBreaks = PageBreakInfo.GetPageBreakInfos(pageBreaks2.VerticalPageBreaks);

            Assert.IsTrue(PageBreakInfo.AreEqualPageBreaksCollections(firstVerticalBreaks, secondVerticalBreaks));
        }

        private static void AssertPrintAreasAreEqual(PrintArea printArea1, PrintArea printArea2)
        {
            AssertCellRangeCollectionsAreEqual(printArea1.Ranges, printArea2.Ranges);
        }

        private static void AssertWorksheetShapesAreEqual(ShapeCollection shapeCollection1, ShapeCollection shapeCollection2)
        {
            Assert.AreEqual(shapeCollection1.Count(), shapeCollection2.Count());

            for (int i = 0; i < shapeCollection1.Count; i++)
            {
                FloatingShapeBase shape1 = shapeCollection1.ElementAt(i);
                FloatingShapeBase shape2 = shapeCollection2.ElementAt(i);

                Assert.AreEqual(shape1.Name, shape2.Name);
                Assert.AreEqual(shape1.Width, shape2.Width);
                Assert.AreEqual(shape1.Height, shape2.Height);
                Assert.AreEqual(shape1.RotationAngle, shape2.RotationAngle);
                Assert.AreEqual(shape1.CellIndex, shape2.CellIndex);
                Assert.AreEqual(shape1.OffsetX, shape2.OffsetX);
                Assert.AreEqual(shape1.OffsetY, shape2.OffsetY);
                Assert.AreEqual(shape1.IsHorizontallyFlipped, shape2.IsHorizontallyFlipped);
                Assert.AreEqual(shape1.IsVerticallyFlipped, shape2.IsVerticallyFlipped);
                Assert.AreEqual(shape1.LockAspectRatio, shape2.LockAspectRatio);

                if (shape1.FloatingShapeType == FloatingShapeType.Image)
                {
                    Assert.AreEqual(FloatingShapeType.Image, shape2.FloatingShapeType);

                    Assert.IsTrue(CompareArrays((shape1 as FloatingImage).ImageSource.Data, (shape2 as FloatingImage).ImageSource.Data));
                    Assert.AreEqual((shape1 as FloatingImage).PreferRelativeToOriginalResize, (shape2 as FloatingImage).PreferRelativeToOriginalResize);
                }

#if !SILVERLIGHT
                if (shape1.FloatingShapeType == FloatingShapeType.Chart)
                {
                    Assert.AreEqual(FloatingShapeType.Chart, shape2.FloatingShapeType);
                    AssertChartsAreEqual((shape1 as FloatingChartShape).Chart, (shape2 as FloatingChartShape).Chart);
                }
#endif
            }
        }

#if !SILVERLIGHT
        public static void AssertChartsAreEqual(DocumentChart chart1, DocumentChart chart2)
        {
            AssertSeriesGroupCollectionsAreEqual(chart1.SeriesGroups, chart2.SeriesGroups);
            AssertAxesCollectionsAreEqual(chart1.PrimaryAxes, chart2.PrimaryAxes);
            AssertAxesCollectionsAreEqual(chart1.SecondaryAxes, chart2.SecondaryAxes);
        }

        private static void AssertSeriesGroupCollectionsAreEqual(SeriesGroupCollection seriesGroups1, SeriesGroupCollection seriesGroups2)
        {
            Assert.AreEqual(seriesGroups1.Count(), seriesGroups2.Count());

            for (int i = 0; i < seriesGroups1.Count(); i++)
            {
                AssertSeriesGroupsAreEqual(seriesGroups1.ElementAt(i), seriesGroups2.ElementAt(i));
            }
        }

        private static void AssertSeriesGroupsAreEqual(SeriesGroup seriesGroup1, SeriesGroup seriesGroup2)
        {
            GenericSeriesGroupEqualityComparer equalityComparer = new GenericSeriesGroupEqualityComparer();
            Assert.IsTrue(equalityComparer.Equals(seriesGroup1, seriesGroup2));
        }

        private static void AssertAxesCollectionsAreEqual(AxisGroup axes1, AxisGroup axes2)
        {
            if (axes1 == null)
            {
                Assert.IsNull(axes2);
                return;
            }

            if (axes2 == null)
            {
                Assert.IsNull(axes1);
                return;
            }

            AssertAxesAreEqual(axes1.CategoryAxis, axes2.CategoryAxis);
            AssertAxesAreEqual(axes1.ValueAxis, axes2.ValueAxis);
        }

        private static void AssertAxesAreEqual(Axis axis1, Axis axis2)
        {
            if (axis1 == null)
            {
                Assert.IsNull(axis2);
                return;
            }

            if (axis2 == null)
            {
                Assert.IsNull(axis1);
                return;
            }

            Assert.AreEqual(axis1.IsVisible, axis2.IsVisible);
            Assert.AreEqual(axis1.Min, axis2.Min);
            Assert.AreEqual(axis1.Max, axis2.Max);
            Assert.AreEqual(axis1.NumberFormat, axis2.NumberFormat);

            Assert.AreEqual(axis1.AxisType, axis2.AxisType);
        }
#endif

        private static void AssertFiltersAreEqual(AutoFilter autoFilter1, AutoFilter autoFilter2)
        {
            Assert.AreEqual(autoFilter1.FilterRange, autoFilter2.FilterRange);

            if (autoFilter1.FilterRange != null)
            {
                for (int i = 0; i < autoFilter1.FilterRange.ColumnCount; i++)
                {
                    IFilter filter = autoFilter1.GetFilter(i);
                    if (filter == null)
                    {
                        Assert.IsNull(autoFilter2.GetFilter(i));
                    }
                    else
                    {
                        Type filter1Type = filter.GetType();
                        Assert.IsInstanceOfType(autoFilter2.GetFilter(i), filter1Type);
                    }

                    ValuesCollectionFilter valuesCollectionFilter = filter as ValuesCollectionFilter;
                    if (valuesCollectionFilter != null)
                    {
                        Assert.AreEqual(valuesCollectionFilter, (ValuesCollectionFilter)autoFilter2.GetFilter(i));
                    }

                    CustomFilter customFilter = filter as CustomFilter;
                    if (customFilter != null)
                    {
                        Assert.AreEqual(customFilter, (CustomFilter)autoFilter2.GetFilter(i));
                    }

                    DynamicFilter dynamicFilter = filter as DynamicFilter;
                    if (dynamicFilter != null)
                    {
                        Assert.AreEqual(dynamicFilter, (DynamicFilter)autoFilter2.GetFilter(i));
                    }

                    TopFilter topFilter = filter as TopFilter;
                    if (topFilter != null)
                    {
                        Assert.AreEqual(topFilter, (TopFilter)autoFilter2.GetFilter(i));
                    }

                    ForeColorFilter foreColorFilter = filter as ForeColorFilter;
                    if (foreColorFilter != null)
                    {
                        Assert.AreEqual(foreColorFilter, (ForeColorFilter)autoFilter2.GetFilter(i));
                    }

                    FillColorFilter fillColorFilter = filter as FillColorFilter;
                    if (fillColorFilter != null)
                    {
                        Assert.AreEqual(fillColorFilter, (FillColorFilter)autoFilter2.GetFilter(i));
                    }
                }
            }
        }

        private static void AssertSortStatesAreEqual(SortState state1, SortState state2)
        {
            Assert.AreEqual(state1.SortRange, state2.SortRange);

            if (state1.SortRange != null)
            {
                ISortCondition[] conditions1 = state1.SortConditions.ToArray();
                ISortCondition[] conditions2 = state2.SortConditions.ToArray();

                Assert.AreEqual(conditions1.Length, conditions2.Length);

                for (int i = 0; i < conditions1.Length; i++)
                {
                    Assert.IsInstanceOfType(conditions1[i], conditions2[i].GetType());
                    Assert.AreEqual(conditions1[i].RelativeIndex, conditions2[i].RelativeIndex);

                    ValuesSortCondition valuesSort = conditions1[i] as ValuesSortCondition;
                    if (valuesSort != null)
                    {
                        Assert.AreEqual(valuesSort.SortOrder, ((ValuesSortCondition)conditions2[i]).SortOrder);
                    }

                    CustomValuesSortCondition customSort = conditions1[i] as CustomValuesSortCondition;
                    if (customSort != null)
                    {
                        CollectionAssert.AreEquivalent(customSort.CustomList, ((CustomValuesSortCondition)conditions2[i]).CustomList);
                    }

                    FillColorSortCondition fillColorSort = conditions1[i] as FillColorSortCondition;
                    if (fillColorSort != null)
                    {
                        Assert.AreEqual(fillColorSort.Fill, ((FillColorSortCondition)conditions2[i]).Fill);
                        Assert.AreEqual(fillColorSort.SortOrder, ((FillColorSortCondition)conditions2[i]).SortOrder);
                    }

                    ForeColorSortCondition foreColorSort = conditions1[i] as ForeColorSortCondition;
                    if (foreColorSort != null)
                    {
                        Assert.AreEqual(foreColorSort.Color, ((ForeColorSortCondition)conditions2[i]).Color);
                        Assert.AreEqual(foreColorSort.SortOrder, ((ForeColorSortCondition)conditions2[i]).SortOrder);
                    }
                }
            }
        }

        private static void AssertGroupingPropertiesAreEqual(GroupingProperties properties1, GroupingProperties properties2)
        {
            Assert.AreEqual(properties1.SummaryRowIsBelow, properties2.SummaryRowIsBelow);
            Assert.AreEqual(properties1.SummaryColumnIsToRight, properties2.SummaryColumnIsToRight);
        }

        private static bool CompareArrays(byte[] data1, byte[] data2)
        {
            if (data1.Length != data2.Length)
            {
                return false;
            }

            for (int i = 0; i < data1.Length; i++)
            {
                if (data1[i] != data2[i])
                {
                    return false;
                }
            }

            return true;
        }

        private static void AssertWorksheetViewStatesAreEqual(WorksheetViewState expectedWorksheetViewState, WorksheetViewState actualWorksheetViewState)
        {
            ViewStateTestHelper viewStateTestHelper = new ViewStateTestHelper(expectedWorksheetViewState);

            viewStateTestHelper.AssertScaleFactorIsEqualTo(actualWorksheetViewState.ScaleFactor);
            viewStateTestHelper.AssertTopLeftCellIndexIsEqualTo(actualWorksheetViewState.TopLeftCellIndex);
            viewStateTestHelper.AssertPaneIsEqualTo(actualWorksheetViewState.Pane);
            viewStateTestHelper.AssertSelectionIsEqualTo(actualWorksheetViewState.SelectionState);
        }

        /// <summary>
        /// Asserts the worksheet merged ranges are equal.
        /// </summary>
        /// <param name="mergedCellRanges1">The merged cell ranges.</param>
        /// <param name="mergedCellRanges2">The merged cell ranges1.</param>
        private static void AssertWorksheetMergedRangesAreEqual(IEnumerable<CellRange> mergedCellRanges1, IEnumerable<CellRange> mergedCellRanges2)
        {
            List<CellRange> mergedCellRangesList1 = mergedCellRanges1.ToList();
            List<CellRange> mergedCellRangesList2 = mergedCellRanges2.ToList();

            Assert.AreEqual(mergedCellRangesList1.Count, mergedCellRangesList2.Count);

            for (int i = 0; i < mergedCellRangesList1.Count; i++)
            {
                Assert.AreEqual(mergedCellRangesList1[i], mergedCellRangesList2[i]);
            }
        }

        /// <summary>
        /// Asserts the worksheet names are equal.
        /// </summary>
        /// <param name="names1">The names1.</param>
        /// <param name="names2">The names2.</param>
        private static void AssertNamesAreEqual(NameCollection names1, NameCollection names2)
        {
            List<ISpreadsheetName> namesList1 = names1.ToList();
            List<ISpreadsheetName> namesList2 = names2.ToList();

            Assert.AreEqual(namesList1.Count, namesList2.Count);

            for (int i = 0; i < namesList1.Count; i++)
            {
                AssertISpreadsheetNameEqulas(namesList1[i], namesList2[i]);
            }
        }

        /// <summary>
        /// Asserts the ISpreadsheet names are equlas.
        /// </summary>
        /// <param name="name1">The name1.</param>
        /// <param name="name2">The name2.</param>
        private static void AssertISpreadsheetNameEqulas(ISpreadsheetName name1, ISpreadsheetName name2)
        {
            Assert.AreEqual(name1.Comment, name2.Comment);
            Assert.AreEqual(name1.Name, name2.Name);
            Assert.AreEqual(name1.RefersTo, name2.RefersTo);
            AssertISpreadsheetNamaScopeEquals(name1.Scope, name2.Scope);
            Assert.AreEqual(name1.Value, name2.Value);
        }

        /// <summary>
        /// Asserts the I spreadsheet nama scope equals.
        /// </summary>
        /// <param name="scope1">The scope1.</param>
        /// <param name="scope2">The scope2.</param>
        private static void AssertISpreadsheetNamaScopeEquals(SpreadsheetNameCollectionScope scope1, SpreadsheetNameCollectionScope scope2)
        {
            Assert.AreEqual(scope1.IsGlobal, scope2.IsGlobal);
        }

        /// <summary>
        /// Asserts the worksheet hyperlinks are equal.
        /// </summary>
        /// <param name="hyperlinks1">The hyperlinks1.</param>
        /// <param name="hyperlinks2">The hyperlinks2.</param>
        private static void AssertWorksheetHyperlinksAreEqual(HyperlinkCollection hyperlinks1, HyperlinkCollection hyperlinks2)
        {
            List<SpreadsheetHyperlink> hyperlinksList1 = hyperlinks1.ToList();
            List<SpreadsheetHyperlink> hyperlinksList2 = hyperlinks2.ToList();

            Assert.AreEqual(hyperlinksList1.Count, hyperlinksList2.Count);

            for (int i = 0; i < hyperlinksList1.Count; i++)
            {
                Assert.AreEqual(hyperlinksList1[i], hyperlinksList2[i]);
            }
        }

        /// <summary>
        /// Asserts the worksheet columns are equal.
        /// </summary>
        /// <param name="columns1">The columns1.</param>
        /// <param name="columns2">The columns2.</param>
        private static void AssertWorksheetColumnsAreEqual(Columns columns1, Columns columns2)
        {
            Assert.AreEqual(columns1.Count, columns2.Count);

            ColumnWidth defaultWidth1 = columns1.GetDefaultWidth();
            ColumnWidth defaultWidth2 = columns2.GetDefaultWidth();
            Assert.AreEqual(defaultWidth1, defaultWidth2);

            CellRange usedCellRange1 = columns1.Worksheet.UsedCellRange;

            for (int column = 0; column < usedCellRange1.ColumnCount; column++)
            {
                AssertIndividualWorksheetColumnsAreEqual(columns1[column], columns2[column]);
            }
        }

        private static void AssertIndividualWorksheetColumnsAreEqual(ColumnSelection column1, ColumnSelection column2)
        {
            ColumnWidth columnWidth1 = column1.GetWidth().Value;
            ColumnWidth columnWidth2 = column2.GetWidth().Value;

            Assert.AreEqual(columnWidth1.IsCustom, columnWidth2.IsCustom);
            Assert.AreEqual(columnWidth1.Value, columnWidth2.Value);

            int columnOutlineLevel1 = column1.GetOutlineLevel().Value;
            int columnOutlineLevel2 = column2.GetOutlineLevel().Value;

            Assert.AreEqual(columnOutlineLevel1, columnOutlineLevel2);

            bool columnHiddenState1 = column1.GetHidden().Value;
            bool columnHiddenState2 = column2.GetHidden().Value;

            Assert.AreEqual(columnHiddenState1, columnHiddenState2);
        }

        /// <summary>
        /// Asserts the worksheet rows are equal.
        /// </summary>
        /// <param name="rows1">The rows1.</param>
        /// <param name="rows2">The rows2.</param>
        private static void AssertWorksheetRowsAreEqual(Rows rows1, Rows rows2)
        {
            Assert.AreEqual(rows1.Count, rows2.Count);

            RowHeight defaultHeight1 = rows1.GetDefaultHeight();
            RowHeight defaultHeight2 = rows2.GetDefaultHeight();
            Assert.AreEqual(defaultHeight1, defaultHeight2);

            CellRange usedCellRange1 = rows1.Worksheet.UsedCellRange;

            for (int row = 0; row < usedCellRange1.RowCount; row++)
            {
                AssertIndividualWorksheetRowsAreEqual(rows1[row], rows2[row]);
            }
        }

        private static void AssertIndividualWorksheetRowsAreEqual(RowSelection row1, RowSelection row2)
        {
            RowHeight rowHeight1 = row1.GetHeight().Value;
            RowHeight rowHeight2 = row2.GetHeight().Value;

            Assert.AreEqual(rowHeight1.IsCustom, rowHeight2.IsCustom);

            if (rowHeight1.IsCustom)
            {
                // The non-custom height of the rows is not exported.
                Assert.AreEqual(rowHeight1.Value, rowHeight2.Value);
            }

            int rowOutlineLevel1 = row1.GetOutlineLevel().Value;
            int rowOutlineLevel2 = row2.GetOutlineLevel().Value;

            Assert.AreEqual(rowOutlineLevel1, rowOutlineLevel2);

            bool rowHiddenState1 = row1.GetHidden().Value;
            bool rowHiddenState2 = row2.GetHidden().Value;

            Assert.AreEqual(rowHiddenState1, rowHiddenState2);
        }

        /// <summary>
        /// Asserts the properties are equal.
        /// </summary>
        /// <param name="cellSelection1">The cell selection1.</param>
        /// <param name="cellSelection2">The cell selection2.</param>
        private static void AssertPropertiesAreEqual(CellSelection cellSelection1, CellSelection cellSelection2)
        {
            CellBorders borders1 = cellSelection1.GetBorders();
            CellBorders borders2 = cellSelection2.GetBorders();
            Assert.AreEqual(borders1, borders2);

            IFill fill1 = cellSelection1.GetFill().Value;
            IFill fill2 = cellSelection2.GetFill().Value;
            Assert.AreEqual(fill1, fill2);

            ThemableFontFamily fontFamily1 = cellSelection1.GetFontFamily().Value;
            ThemableFontFamily fontFamily2 = cellSelection2.GetFontFamily().Value;
            Assert.AreEqual(fontFamily1, fontFamily2);

            double fontSize1 = cellSelection1.GetFontSize().Value;
            double fontSize2 = cellSelection2.GetFontSize().Value;
            Assert.AreEqual(fontSize1, fontSize2);

            ThemableColor foreColor1 = cellSelection1.GetForeColor().Value;
            ThemableColor foreColor2 = cellSelection2.GetForeColor().Value;
            Assert.AreEqual(foreColor1, foreColor2);

            CellValueFormat format1 = cellSelection1.GetFormat().Value;
            CellValueFormat format2 = cellSelection2.GetFormat().Value;
            Assert.AreEqual(format1, format2);

            RadHorizontalAlignment horizontalAlignment1 = cellSelection1.GetHorizontalAlignment().Value;
            RadHorizontalAlignment horizontalAlignment2 = cellSelection2.GetHorizontalAlignment().Value;
            Assert.AreEqual(horizontalAlignment1, horizontalAlignment2);

            int indent1 = cellSelection1.GetIndent().Value;
            int indent2 = cellSelection2.GetIndent().Value;
            Assert.AreEqual(indent1, indent2);

            bool isBold1 = cellSelection1.GetIsBold().Value;
            bool isBold2 = cellSelection2.GetIsBold().Value;
            Assert.AreEqual(isBold1, isBold2);

            bool isItalic1 = cellSelection1.GetIsItalic().Value;
            bool isItalic2 = cellSelection2.GetIsItalic().Value;
            Assert.AreEqual(isItalic1, isItalic2);

            bool isWrapped1 = cellSelection1.GetIsWrapped().Value;
            bool isWrapped2 = cellSelection2.GetIsWrapped().Value;
            Assert.AreEqual(isWrapped1, isWrapped2);

            string styleName1 = cellSelection1.GetStyleName().Value;
            string styleName2 = cellSelection2.GetStyleName().Value;
            Assert.AreEqual(styleName1, styleName2);

            UnderlineType underline1 = cellSelection1.GetUnderline().Value;
            UnderlineType underline2 = cellSelection2.GetUnderline().Value;
            Assert.AreEqual(underline1, underline2);

            string value1 = cellSelection1.GetValue().Value.GetResultValueAsString(cellSelection1.GetFormat().Value);
            string value2 = cellSelection2.GetValue().Value.GetResultValueAsString(cellSelection2.GetFormat().Value);
            Assert.AreEqual(value1, value2);

            RadVerticalAlignment verticalAlignment1 = cellSelection1.GetVerticalAlignment().Value;
            RadVerticalAlignment verticalAlignment2 = cellSelection2.GetVerticalAlignment().Value;
            Assert.AreEqual(verticalAlignment1, verticalAlignment2);

            IDataValidationRule rule1 = cellSelection1.GetDataValidationRule().Value;
            IDataValidationRule rule2 = cellSelection2.GetDataValidationRule().Value;
            Assert.AreEqual(rule1, rule2);
        }

        public static bool AreDoubleValuesEssentiallyEqual(double a, double b, double epsilon = Epsilon)
        {
            if (double.IsNaN(a) || double.IsInfinity(a) || double.IsNaN(b) || double.IsInfinity(b))
            {
                throw new ArgumentException("Cannot compare NaN or Infinity value for equality");
            }

            double aAbsolute = Math.Abs(a);
            double bAbsolute = Math.Abs(b);

            if (Math.Abs(a) < epsilon || Math.Abs(b) < epsilon)
            {
                return Math.Abs(a - b) < epsilon;
            }

            return Math.Abs(a - b) <= (Math.Min(aAbsolute, bAbsolute) * epsilon);
        }

        public static bool AreComplexNumbersEssentiallyEqual(string firstComplex, string secondComplex, double epsilon = Epsilon)
        {
            if (!HaveSameImaginarySymbol(firstComplex, secondComplex))
            {
                return false;
            }

            RadExpression[] firstAsArguments = { new StringExpression(firstComplex) };
            RadExpression[] secondAsArguments = { new StringExpression(secondComplex) };

            ImReal imRealFunction = new ImReal();
            FunctionEvaluationContext<RadExpression> firstContext = new FunctionEvaluationContext<RadExpression>(firstAsArguments, null, 0, 0);
            FunctionEvaluationContext<RadExpression> secondContext = new FunctionEvaluationContext<RadExpression>(secondAsArguments, null, 0, 0);

            NumberExpression firstRealPart = imRealFunction.Evaluate(firstContext) as NumberExpression;
            NumberExpression secondRealPart = imRealFunction.Evaluate(secondContext) as NumberExpression;

            Imaginary imaginaryFunction = new Imaginary();
            NumberExpression firstImaginaryPart = imaginaryFunction.Evaluate(firstContext) as NumberExpression;
            NumberExpression secondImaginaryPart = imaginaryFunction.Evaluate(secondContext) as NumberExpression;

            if (firstRealPart != null && firstImaginaryPart != null && secondRealPart != null && secondImaginaryPart != null)
            {
                bool areRealPartsEqual = AreDoubleValuesEssentiallyEqual(firstRealPart.Value, secondRealPart.Value, epsilon);
                bool areImaginaryPartsEqual = AreDoubleValuesEssentiallyEqual(firstImaginaryPart.Value, secondImaginaryPart.Value, epsilon);

                return areRealPartsEqual && areImaginaryPartsEqual;
            }
            else
            {
                return false;
            }
        }

        private static bool HaveSameImaginarySymbol(string firstComplex, string secondComplex)
        {
            if (string.IsNullOrEmpty(firstComplex) || string.IsNullOrEmpty(secondComplex))
            {
                return true;
            }

            string firstTrimed = firstComplex.Trim();
            string secondTrimed = secondComplex.Trim();

            bool hasFirstDefinedSymbol = firstTrimed.Length > 0 && (firstTrimed.Last() == 'i' || firstTrimed.Last() == 'j');
            bool hasSecondDefinedSymbol = secondTrimed.Length > 0 && (secondTrimed.Last() == 'i' || secondTrimed.Last() == 'j');

            if (hasFirstDefinedSymbol && hasSecondDefinedSymbol)
            {
                return firstTrimed.Last() == secondTrimed.Last();
            }

            return true;
        }

        public static FormulaCellValue CreateFormulaCellValue(string value, Worksheet worksheet, CellIndex cellIndex, bool addFunctionStart = true)
        {
            string input = addFunctionStart ? "=" + value : value;

            return (FormulaCellValue)input.ToCellValue(worksheet, cellIndex);
        }

        public static string ReplaceCultureDependentSeparators(this string value)
        {
            return CultureStringHelper.Replace(value);
        }

        public static string ToCultureDependentString(this double value)
        {
            return value.ToString();
        }

        public static IEnumerable<bool[]> GenerateBinaryVariations(int positionsCount)
        {
            Guard.ThrowExceptionIfGreaterThan(31, positionsCount, "positionsCount");

            int variationsCount = 1 << positionsCount;

            for (int i = 0; i < variationsCount; i++)
            {
                bool[] variation = new bool[positionsCount];

                for (int position = 0; position < positionsCount; position++)
                {
                    variation[position] = ((i >> position) & 1) == 1;
                }

                yield return variation;
            }
        }

        public static void AssertCollectionsAreEqual<T>(IEnumerable<T> expected, IEnumerable<T> actual)
        {
            Assert.AreEqual(expected.Count(), actual.Count());

            for (int i = 0; i < expected.Count(); i++)
            {
                Assert.AreEqual(expected.ElementAt(i), actual.ElementAt(i));
            }
        }

        #endregion
    }
}
