﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests
{
	[TestClass]
	public class WorkbookSheetCollectionTests
    {
        #region Manipulating with SheetCollection's sheets

        [TestMethod]
		public void CreateEmptyWorkbookTest()
		{
			Workbook workbook = new Workbook();
			Assert.AreEqual(0, workbook.Sheets.Count);
		}

		[TestMethod]
		public void AddTest()
		{
			Workbook workbook = new Workbook();
			Assert.AreEqual(0, workbook.Sheets.Count);

			for (int i = 1; i <= 10; i++)
			{
				Sheet sheet = workbook.Sheets.Add(SheetType.Worksheet);

				Assert.AreEqual(workbook, sheet.Workbook);
				Assert.AreEqual(SheetType.Worksheet, sheet.Type);
				Assert.AreEqual(i, workbook.Sheets.Count);
			}
		}

		[TestMethod]
		public void InsertTest()
		{
			Workbook workbook = new Workbook();
			Assert.AreEqual(0, workbook.Sheets.Count);

			for (int i = 1; i <= 10; i++)
			{
				Sheet sheet = workbook.Sheets.Insert(0, SheetType.Worksheet);

				Assert.AreEqual("Sheet" + i, sheet.Name);
				Assert.AreEqual(i, workbook.Sheets.Count);
			}

			for (int i = 1; i <= 10; i++)
			{
				Sheet sheet = workbook.Sheets[10 - i];

				string expectedSheetName = ("Sheet" + i);
				Assert.AreEqual(expectedSheetName, sheet.Name);
				Assert.AreEqual(sheet, workbook.Sheets[expectedSheetName]);
			}
		}

		[TestMethod]
		public void AddAndGetByNameTest()
		{
			Workbook workbook = new Workbook();
			Assert.AreEqual(0, workbook.Sheets.Count);

			for (int i = 1; i <= 10; i++)
			{
				Sheet sheet = workbook.Sheets.Add(SheetType.Worksheet);

				Assert.AreEqual("Sheet" + i, sheet.Name);
				Assert.AreEqual(i, workbook.Sheets.Count);
			}

			for (int i = 1; i <= 10; i++)
			{
				Sheet sheet = workbook.Sheets[i - 1];

				string expectedSheetName = ("Sheet" + i);
				Assert.AreEqual(expectedSheetName, sheet.Name);
				Assert.AreEqual(sheet, workbook.Sheets[expectedSheetName]);
			}
		}

		[TestMethod]
		public void Remove()
		{
			Workbook workbook = new Workbook();

			Assert.AreEqual(0, workbook.Sheets.Count);

			Sheet sheet = workbook.Sheets.Add(SheetType.Worksheet);
			Assert.AreEqual(1, workbook.Sheets.Count);

			workbook.Sheets.Remove(sheet);
			Assert.AreEqual(0, workbook.Sheets.Count);
		}

		[TestMethod]
		[ExpectedException(typeof(ArgumentNullException))]
		public void SetSheetNameToNullTest()
		{
			Workbook workbook = new Workbook();
			Sheet sheet = workbook.Sheets.Add(SheetType.Worksheet);
			sheet.Name = null;
		}

		[TestMethod]
		[ExpectedException(typeof(LocalizableException))]
		public void ChangeNameToExistingNameTest()
		{
			Workbook workbook = new Workbook();

			Sheet sheet = workbook.Sheets.Add(SheetType.Worksheet);
			Assert.AreEqual(1, workbook.Sheets.Count);
			Assert.AreEqual("Sheet1", sheet.Name);

			Sheet sheet1 = workbook.Sheets.Add(SheetType.Worksheet);
			Assert.AreEqual("Sheet2", sheet1.Name);
			sheet1.Name = "Sheet1";
		}

        [TestMethod]
        public void MoveForwardTest()
        {
            Workbook workbook = new Workbook();
            Assert.AreEqual(0, workbook.Sheets.Count);

            Sheet sheet;
            string expectedSheetName;
            for (int i = 1; i <= 10; i++)
            {
                sheet = workbook.Sheets.Add(SheetType.Worksheet);

                Assert.AreEqual("Sheet" + i, sheet.Name);
                Assert.AreEqual(i, workbook.Sheets.Count);
            }

            workbook.Sheets.Move(5, 2, 8);

            int[] expectedIndexes = new int[]
            {
                1,
                2,
                3,
                4,
                5,
                8,
                6,
                7,
                9,
                10
            };

            for (int i = 0; i < workbook.Sheets.Count; i++)
            {
                sheet = workbook.Sheets[i];

                expectedSheetName = ("Sheet" + expectedIndexes[i]);
                Assert.AreEqual(expectedSheetName, sheet.Name);
            }
        }

        #endregion


        #region Sheet name validation

        [TestMethod]
        public void ValidSheetNameTests()
        {
            Workbook workbook = new Workbook();
            Worksheet worksheet = workbook.Worksheets.Add();

            string[] validNames = { "Pesho", "Pesho's", "-+=&@#%^()!.<>~;  __", " ", "1111122222333334444455555666667", "_Sheet_Ябадабадуу64_", "=EqualSign" };

            foreach (string validName in validNames)
            {
                worksheet.Name = validName;
            }
        }

        [TestMethod]
        public void InvalidSheetNameTests()
        {
            Workbook workbook = new Workbook();
            Worksheet worksheet = workbook.Worksheets.Add();

            string[] invalidNames = { "Pesho'", "'Pesho'", "11111222223333344444555556666677", "", 
                                        @"P/esho", @"P\esho", @"P/\esho", @"P?esho", @"P*esho", @"P:esho", @"P[esho", @"P]esho", @"'P/\?*:[]esho"};

            foreach (string invalidName in invalidNames)
            {
                try
                {
                    worksheet.Name = invalidName;
                    Assert.Fail();
                }
                catch (Exception ex)
                {
                    Assert.AreEqual(typeof(LocalizableException), ex.GetType());
                }                
            }            
        }    

        [TestMethod]
        [ExpectedException(typeof(LocalizableException))]
        public void RepeatedSheetNameExceptionsTests()
        {
            Workbook workbook = new Workbook();

            Worksheet worksheet1 = workbook.Worksheets.Add();
            worksheet1.Name = "Pesho";

            Worksheet worksheet2 = workbook.Worksheets.Add();
            worksheet2.Name = "Pesho";
        }

        [TestMethod]
        [ExpectedException(typeof(LocalizableException))]
        public void RepeatedCaseInsensitiveSheetDefaultNameExceptionsTests()
        {
            Workbook workbook = new Workbook();
            workbook.Worksheets.Add();

            Worksheet worksheet = workbook.Worksheets.Add();
            worksheet.Name = "ShEeT1";
        }

        [TestMethod]
        [ExpectedException(typeof(LocalizableException))]
        public void RepeatedCaseInsensitiveCyrillicSheetNameExceptionsTests()
        {
            Workbook workbook = new Workbook();

            Worksheet worksheet1 = workbook.Worksheets.Add();
            worksheet1.Name = "Иван Иванов";

            Worksheet worksheet2 = workbook.Worksheets.Add();
            worksheet2.Name = "иВан иВаноВ";
        }

        [TestMethod]
        public void ChangingToTheIgnoreCaseSameNameTests()
        {
            Workbook workbook = new Workbook();

            string firstName = "Иван Иванов";
            string secondName = "иВан иВаноВ";

            Sheet sheet = workbook.Sheets.Add(SheetType.Worksheet);
            sheet.Name = firstName;

            //Changing to the same name with different casing
            sheet.Name = secondName;

            Assert.AreEqual(secondName, sheet.Name);
            Assert.AreNotEqual(firstName, sheet.Name);
        }

        [TestMethod]
        [ExpectedException(typeof(LocalizableException))]
        public void ChangeNameToExistingNameIgnoreCaseTest()
        {
            Workbook workbook = new Workbook();

            workbook.Sheets.Add(SheetType.Worksheet);
            Sheet sheet2 = workbook.Sheets.Add(SheetType.Worksheet);

            sheet2.Name = "sHeEt1";
        }

        [TestMethod]
        public void IndexOfIgnoreCaseTest()
        {
            Workbook workbook = new Workbook();

            workbook.Sheets.Add(SheetType.Worksheet);
            workbook.Sheets.Add(SheetType.Worksheet).Name = "Pesho";
            workbook.Sheets.Add(SheetType.Worksheet);

            int resultIndex = workbook.Sheets.IndexOf("pEsHo");
            int expectedIndex = 1;

            Assert.AreEqual(expectedIndex, resultIndex);
        }

        [TestMethod]
        public void GetByNameIgnoreCaseTest()
        {
            Workbook workbook = new Workbook();

            string sheetName = "Pesho";
            string differentCasingName = "pEsHo";

            workbook.Sheets.Add(SheetType.Worksheet);
            workbook.Sheets.Add(SheetType.Worksheet).Name = sheetName;
            workbook.Sheets.Add(SheetType.Worksheet);

            Sheet sheet = workbook.Sheets.GetByName(differentCasingName);

            Assert.AreEqual(sheetName, sheet.Name);
            Assert.AreNotEqual(differentCasingName, sheet.Name);
        }

        #endregion
    }
}