﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Printing;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.Tests
{
    [TestClass]
    public class WorksheetPageSetupTests
    {
        [TestMethod]
        public void TestAddToPrintArea()
        {
            PrintArea printArea = GeneratePrintAreaScenario().WorksheetPageSetup.PrintArea;

            AssertPrintArea(printArea, new CellRange[]
            {
                new CellRange(4, 2, 14, 5),
                new CellRange(8, 2, 16, 7),
                new CellRange(10, 4, 22, 12),
                new CellRange(6, 11, 15, 12),
                new CellRange(5, 13, 5, 13),
                new CellRange(6, 14, 15, 16),
                new CellRange(4, 17, 18, 18),
            });
        }

        [TestMethod]
        public void TestPrintAreaUndo()
        {
            Worksheet worksheet = GeneratePrintAreaScenario();
            PrintArea printArea = worksheet.WorksheetPageSetup.PrintArea;

            worksheet.Workbook.History.Undo();
            worksheet.Workbook.History.Undo();

            AssertPrintArea(printArea, new CellRange[]
            {
                new CellRange(4, 2, 14, 5),
                new CellRange(8, 2, 16, 7),
                new CellRange(10, 4, 22, 12),
                new CellRange(6, 11, 15, 12),
            });
        }

        [TestMethod]
        public void TestPrintAreaUndoRedo()
        {
            Worksheet worksheet = GeneratePrintAreaScenario();
            PrintArea printArea = worksheet.WorksheetPageSetup.PrintArea;

            worksheet.Workbook.History.Undo();
            worksheet.Workbook.History.Undo();
            worksheet.Workbook.History.Redo();

            AssertPrintArea(printArea, new CellRange[]
            {
                new CellRange(4, 2, 14, 5),
                new CellRange(8, 2, 16, 7),
                new CellRange(10, 4, 22, 12),
                new CellRange(6, 11, 15, 12),
                new CellRange(5, 13, 5, 13),
            });
        }

        [TestMethod]
        public void TestPageBreaksInsertion()
        {
            PageBreaks pageBreaks = GeneratePageBreaksInPrintAreaScenario().WorksheetPageSetup.PageBreaks;

            AssertPageBreaks(pageBreaks,
            new PageBreakInfo[]
            {
                new PageBreakInfo(PageBreakType.Horizontal, 13, 2, 12),
                new PageBreakInfo(PageBreakType.Horizontal, 6, 2, 5),
                new PageBreakInfo(PageBreakType.Horizontal, 18),
            },
            new PageBreakInfo[]
            {
                new PageBreakInfo(PageBreakType.Vertical, 7, 8, 22),
                new PageBreakInfo(PageBreakType.Vertical, 13),
            });
        }

        [TestMethod]
        public void TestPageBreaksUndo()
        {
            Worksheet worksheet = GeneratePageBreaksInPrintAreaScenario();
            PageBreaks pageBreaks = worksheet.WorksheetPageSetup.PageBreaks;

            worksheet.Workbook.History.Undo();
            worksheet.Workbook.History.Undo();

            AssertPageBreaks(pageBreaks,
            new PageBreakInfo[]
            {
                new PageBreakInfo(PageBreakType.Horizontal, 13, 2, 12),
            },
            new PageBreakInfo[]
            {
                new PageBreakInfo(PageBreakType.Vertical, 7, 8, 22),
            });
        }

        [TestMethod]
        public void TestPageBreaksUndoRedo()
        {
            Worksheet worksheet = GeneratePageBreaksInPrintAreaScenario();
            PageBreaks pageBreaks = worksheet.WorksheetPageSetup.PageBreaks;

            worksheet.Workbook.History.Undo();
            worksheet.Workbook.History.Undo();
            worksheet.Workbook.History.Redo();

            AssertPageBreaks(pageBreaks,
            new PageBreakInfo[]
            {
                new PageBreakInfo(PageBreakType.Horizontal, 13, 2, 12),
                new PageBreakInfo(PageBreakType.Horizontal, 6, 2, 5),
            },
            new PageBreakInfo[]
            {
                new PageBreakInfo(PageBreakType.Vertical, 7, 8, 22),
            });
        }

        [TestMethod]
        public void TestPrintAreaClearedWhenPageBreaksUndoRedo()
        {
            Worksheet worksheet = GeneratePageBreaksInPrintAreaScenario();

            PageBreaks pageBreaks = worksheet.WorksheetPageSetup.PageBreaks;
            PrintArea printArea = worksheet.WorksheetPageSetup.PrintArea;

            CellRange[] expectedPrintArea = new CellRange[]
            {
                new CellRange(4, 2, 14, 5),
                new CellRange(8, 2, 16, 7),
                new CellRange(10, 4, 22, 12),
                new CellRange(6, 11, 15, 12),
                new CellRange(5, 13, 5, 13),
                new CellRange(6, 14, 15, 16),
                new CellRange(4, 17, 18, 18),
            };

            PageBreakInfo[] finiteHorizontalBreaks = new PageBreakInfo[]
            {
                new PageBreakInfo(PageBreakType.Horizontal, 13, 2, 12),
                new PageBreakInfo(PageBreakType.Horizontal, 6, 2, 5),
                new PageBreakInfo(PageBreakType.Horizontal, 18),
            };

            PageBreakInfo[] infiniteHorizontalBreaks = new PageBreakInfo[]
            {
                new PageBreakInfo(PageBreakType.Horizontal, 13),
                new PageBreakInfo(PageBreakType.Horizontal, 6),
                new PageBreakInfo(PageBreakType.Horizontal, 18),
            };

            PageBreakInfo[] finiteVerticalBreaks = new PageBreakInfo[]
            {
                new PageBreakInfo(PageBreakType.Vertical, 7, 8, 22),
                new PageBreakInfo(PageBreakType.Vertical, 13),
            };

            PageBreakInfo[] infiniteVerticalBreaks = new PageBreakInfo[]
            {
                new PageBreakInfo(PageBreakType.Vertical, 7),
                new PageBreakInfo(PageBreakType.Vertical, 13),
            };

            Action assertFiniteBreaks = () =>
            {
                AssertPrintArea(printArea, expectedPrintArea);
                AssertPageBreaks(pageBreaks, finiteHorizontalBreaks, finiteVerticalBreaks);
            };

            Action assertInfiniteBreaks = () =>
            {
                AssertPrintArea(printArea, new CellRange[] { });
                AssertPageBreaks(pageBreaks, infiniteHorizontalBreaks, infiniteVerticalBreaks);
            };

            assertFiniteBreaks();
            printArea.Clear();
            assertInfiniteBreaks();
            worksheet.Workbook.History.Undo();
            assertFiniteBreaks();
            worksheet.Workbook.History.Redo();
            assertInfiniteBreaks();
            worksheet.Workbook.History.Undo();
            assertFiniteBreaks();  
        }

        [TestMethod]
        public void TestPageBreaksPrintAreaOnCellsInsertedRightWithUndoRedo()
        {
            Worksheet worksheet = GeneratePageBreaksInPrintAreaScenario();

            CellRange translatedRange = new CellRange(6, 7, 19, 9);

            worksheet.Cells.Insert(translatedRange, InsertShiftType.Right);

            PageBreaks pageBreaks = worksheet.WorksheetPageSetup.PageBreaks;
            PrintArea printArea = worksheet.WorksheetPageSetup.PrintArea;

            Action<int> assertWithTranslation = (translation) =>
            {
                AssertPrintArea(printArea, new CellRange[]
                {
                    new CellRange(4, 2, 14, 5),
                    new CellRange(8, 2, 16, 7 + translation),
                    new CellRange(10, 4, 22, 12),
                    new CellRange(6, 11 + translation, 15, 12 + translation),
                    new CellRange(5, 13, 5, 13),
                    new CellRange(6, 14 + translation, 15, 16 + translation),
                    new CellRange(4, 17, 18, 18),
                });

                PageBreakInfo[] expectedHorizontals = translation == 0 ?
                    new PageBreakInfo[]
                    {
                        new PageBreakInfo(PageBreakType.Horizontal, 13, 2, 12),
                        new PageBreakInfo(PageBreakType.Horizontal, 6, 2, 5),
                        new PageBreakInfo(PageBreakType.Horizontal, 18),
                    } :
                    new PageBreakInfo[]
                    {
                        new PageBreakInfo(PageBreakType.Horizontal, 13, 2, 12),
                        new PageBreakInfo(PageBreakType.Horizontal, 13, 11 + translation, 12 + translation),
                        new PageBreakInfo(PageBreakType.Horizontal, 6, 2, 5),
                        new PageBreakInfo(PageBreakType.Horizontal, 18),
                    };


                AssertPageBreaks(pageBreaks, expectedHorizontals,
                new PageBreakInfo[]
                {
                    new PageBreakInfo(PageBreakType.Vertical, 7, 8, 22),
                    new PageBreakInfo(PageBreakType.Vertical, 13),
                });
            };

            assertWithTranslation(translatedRange.ColumnCount);
            worksheet.Workbook.History.Undo();
            assertWithTranslation(0);
            worksheet.Workbook.History.Redo();
            assertWithTranslation(translatedRange.ColumnCount);
        }

        [TestMethod]
        public void TestPageBreaksPrintAreaOnCellsRemovedLeftWithUndoRedo()
        {
            Worksheet worksheet = GeneratePageBreaksInPrintAreaScenario();

            CellRange translatedRange = new CellRange(6, 7, 19, 9);

            worksheet.Cells.Remove(translatedRange, RemoveShiftType.Left);

            PageBreaks pageBreaks = worksheet.WorksheetPageSetup.PageBreaks;
            PrintArea printArea = worksheet.WorksheetPageSetup.PrintArea;

            Action<int> assertWithTranslation = (translation) =>
            {
                int rangeCutting = translation == 0 ? 0 : 1;

                AssertPrintArea(printArea, new CellRange[]
                {
                    new CellRange(4, 2, 14, 5),
                    new CellRange(8, 2, 16, 7 - rangeCutting),
                    new CellRange(10, 4, 22, 12),
                    new CellRange(6, 11 - translation, 15, 12 - translation),
                    new CellRange(5, 13, 5, 13),
                    new CellRange(6, 14 - translation, 15, 16 - translation),
                    new CellRange(4, 17, 18, 18),
                });


                int horizontalBreakExtension = translation == 0 ? 0 : 1;
                int verticalBreakTrimming = translation == 0 ? 0 : 2;

                AssertPageBreaks(pageBreaks,
                new PageBreakInfo[]
                {
                    new PageBreakInfo(PageBreakType.Horizontal, 13, 2, 12 + horizontalBreakExtension),
                    new PageBreakInfo(PageBreakType.Horizontal, 6, 2, 5),
                    new PageBreakInfo(PageBreakType.Horizontal, 18),
                },
                new PageBreakInfo[]
                {
                    new PageBreakInfo(PageBreakType.Vertical, 7, 8 + verticalBreakTrimming, 22),
                    new PageBreakInfo(PageBreakType.Vertical, 13),
                });
            };

            assertWithTranslation(translatedRange.ColumnCount);
            worksheet.Workbook.History.Undo();
            assertWithTranslation(0);
            worksheet.Workbook.History.Redo();
            assertWithTranslation(translatedRange.ColumnCount);
        }

        private static Worksheet GeneratePageBreaksInPrintAreaScenario()
        {
            Worksheet worksheet = GeneratePrintAreaScenario();

            PageBreaks pageBreaks = worksheet.WorksheetPageSetup.PageBreaks;

            pageBreaks.TryInsertPageBreaks(13, 7);
            pageBreaks.TryInsertPageBreaks(6, 2);
            pageBreaks.TryInsertPageBreaks(18, 13);

            return worksheet;
        }

        private static Worksheet GeneratePrintAreaScenario()
        {
            return CreateWorksheetPrintArea(new CellRange[][]
            {
                new CellRange[] 
                {
                    new CellRange(4, 2, 14, 5),
                    new CellRange(8, 2, 16, 7),
                    new CellRange(10, 4, 22, 12),
                    new CellRange(6, 11, 15, 12),
                },
                new CellRange[] 
                {
                    new CellRange(0, 0, 1, 1),
                    new CellRange(3, 12, 6, 13),
                },
                new CellRange[] 
                {
                    new CellRange(5, 13, 5, 13),
                },
                new CellRange[] 
                {
                    new CellRange(6, 14, 15, 16),
                    new CellRange(4, 17, 18, 18),
                },
                new CellRange[] 
                {
                    new CellRange(5, 13, 100, 13),
                },
            });
        }

        private static Worksheet CreateWorksheetPrintArea(CellRange[][] rangesToAdd)
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            PrintArea printArea = worksheet.WorksheetPageSetup.PrintArea;

            foreach (IEnumerable<CellRange> ranges in rangesToAdd)
            {
                printArea.TryAddToPrintArea(ranges);
            }

            return worksheet;
        }

        private static void AssertPrintArea(PrintArea printArea, IEnumerable<CellRange> expectedRanges)
        {
            List<CellRange> actualCollection = new List<CellRange>(printArea.Ranges);
            List<CellRange> expectedCollection = new List<CellRange>(expectedRanges);

            Assert.AreEqual(expectedCollection.Count, actualCollection.Count);

            for (int index = 0; index < expectedCollection.Count; index++)
            {
                CellRange expectedRange = expectedCollection[index];
                CellRange actualRange = actualCollection[index];

                Assert.AreEqual(expectedRange, actualRange);
            }
        }

        private static void AssertPageBreaks(PageBreaks pageBreaks, IEnumerable<PageBreakInfo> expectedHorizontalPageBreaks, IEnumerable<PageBreakInfo> expectedVerticalPageBreaks)
        {
            IEnumerable<PageBreakInfo> actualHorizontalPageBreaks = PageBreakInfo.GetPageBreakInfos(pageBreaks.HorizontalPageBreaks);
            Assert.IsTrue(PageBreakInfo.AreEqualPageBreaksCollections(expectedHorizontalPageBreaks, actualHorizontalPageBreaks));

            IEnumerable<PageBreakInfo> actualVerticalPageBreaks = PageBreakInfo.GetPageBreakInfos(pageBreaks.VerticalPageBreaks);
            Assert.IsTrue(PageBreakInfo.AreEqualPageBreaksCollections(expectedVerticalPageBreaks, actualVerticalPageBreaks));
        }
    }
}
