﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;
using Telerik.Windows.Documents.Spreadsheet.TestsInternal.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal.Expressions
{
    [TestClass]
    public class AbsoluteWorksheetCellReferenceRangeTests
    {
        #region Fields

        private Workbook workbook;
        private Worksheet worksheet1;
        private Worksheet worksheet2;

        #endregion


        #region TestInitialize

        [TestInitialize]
        public void TestInitialize()
        {
            this.workbook = new Workbook();
            
            this.workbook.Sheets.Add(SheetType.Worksheet);
            this.workbook.Sheets.Add(SheetType.Worksheet);

            this.worksheet1 = this.workbook.Worksheets["Sheet1"];
            this.worksheet2 = this.workbook.Worksheets["Sheet2"];
        }

        [TestCleanup]
        public void TestCleanup()
        {
            this.worksheet1 = null;
            this.worksheet2 = null;
            this.workbook = null;
        }

        #endregion


        #region Helper Methods

        public void AssertWorksheetCellReferenceRangeCreated(string expressionString, Worksheet worksheet, CellIndex cellIndex, 
            double expectedValue, string exprectedValueString = null)
        {
            RadExpression expression = ParseExpression(expressionString.ReplaceListSeparator().ReplaceGroupDecimalSeparators(), worksheet, cellIndex);
            CellReferenceTestHelper.AssertNumberExpressionValue(expression, expectedValue);

            if (!string.IsNullOrEmpty(exprectedValueString))
            {
                Assert.AreEqual(exprectedValueString, expression.ToString());
            }
        }

        public void AssertWorksheetCellReferenceRangeError(string expressionString, Worksheet worksheet, CellIndex cellIndex, ErrorExpression expectedErrorExpression)
        {
            RadExpression expression = ParseExpression(expressionString.ReplaceListSeparator().ReplaceGroupDecimalSeparators(), worksheet, cellIndex);
            Assert.AreEqual(expectedErrorExpression, expression.GetValueAsConstantExpression());
        }

        public void AssertTranslatedWorksheetCellReferenceRange(RadExpression expression, Worksheet targetWorksheet, CellIndex targetCell, double expectedValue, 
            string exptectedString,  int expectedFromIndexRow, int expectedFromIndexColumn, int expectedToIndexRow, int expectedToIndexColumn)
        {
            RadExpression translatedExpression = TranslateExpression(expression, targetWorksheet, targetCell);

            CellReferenceTestHelper.AssertTranslation(translatedExpression, exptectedString, expectedFromIndexRow, expectedFromIndexColumn, expectedToIndexRow, expectedToIndexColumn);
            CellReferenceTestHelper.AssertNumberExpressionValue(translatedExpression, expectedValue);
        }

        public void AssertTranslatedWorksheetCellReferenceRangeError(RadExpression expression, Worksheet targetWorksheet, CellIndex targetCell, string exptectedString)
        {
            RadExpression translatedExpression = TranslateExpression(expression, targetWorksheet, targetCell);

            CellReferenceTestHelper.AssertToString(translatedExpression, exptectedString);
            Assert.AreEqual(ErrorExpressions.ReferenceError, translatedExpression.GetValueAsConstantExpression());
        }

        private RadExpression ParseExpression(string expressionString, Worksheet worksheet, CellIndex cellIndex)
        {
            RadExpression expression;
            RadExpression.TryParse(expressionString, worksheet, cellIndex, out expression);
            return expression;
        }

        public void AssertTranslateExpressionStringCorrectness(string value, Worksheet sourceWorksheet,
            CellIndex sourceCell, string expectedResult, Worksheet targetWorksheet, CellIndex targetCell)
        {
            sourceWorksheet.Cells[sourceCell].SetValue(value.ReplaceListSeparator().ReplaceGroupDecimalSeparators());
            WorksheetFragment fragment = sourceWorksheet.Cells[sourceCell].Copy();
            targetWorksheet.Cells[targetCell].Paste(fragment, PasteOptions.All);

            Assert.AreEqual(expectedResult.ReplaceListSeparator().ReplaceGroupDecimalSeparators(), targetWorksheet.Cells[targetCell].GetValue().Value.GetValueAsString(CellValueFormat.GeneralFormat));
        }

        private RadExpression TranslateExpression(RadExpression expression, Worksheet targetWorksheet, CellIndex targetCell)
        {
            RadExpression translatedExpression;
            ExpressionCloneAndTranslateContext context = new ExpressionCloneAndTranslateContext(targetWorksheet, targetCell.RowIndex, targetCell.ColumnIndex, false);
            translatedExpression = expression.CloneAndTranslate(context);
            return translatedExpression;
        }

        #endregion


        #region Creating Cell Reference Ranges with Absolute Worksheet

        [TestMethod]
        public void CreateCellReferenceRangeWithAbsoluteWorkbookTest1()
        {
            this.workbook.ActiveWorksheet = this.worksheet1;
            this.worksheet1.Cells[0, 0].SetValue(10);

            AssertWorksheetCellReferenceRangeCreated("=Book1!A1", this.worksheet1, new CellIndex(1, 1), 10, "Sheet1!A1");
        }

        [TestMethod]
        public void CreateCellReferenceRangeWithAbsoluteWorkbookTest2()
        {
            this.workbook.ActiveWorksheet = this.worksheet2;
            this.worksheet2.Cells[0, 0].SetValue(10);
            this.worksheet2.Name = "123Tr'alala";

            AssertWorksheetCellReferenceRangeCreated("=Book1!A1", this.worksheet2, new CellIndex(1, 1), 10, "'123Tr''alala'!A1");
        }

        [TestMethod]
        public void CreateCellReferenceRangeWithAbsoluteWorksheetTest1()
        {
            this.worksheet1.Cells[0, 0].SetValue(10);

            AssertWorksheetCellReferenceRangeCreated("=Sheet1!A1", this.worksheet1, new CellIndex(1, 1), 10);
        }

        [TestMethod]
        public void CreateCellReferenceRangeWithAbsoluteWorksheetTest2()
        {
            this.worksheet1.Cells[0, 0].SetValue(10);

            AssertWorksheetCellReferenceRangeCreated("=Sheet1!$A$1", this.worksheet1, new CellIndex(1, 1), 10);
        }

        [TestMethod]
        public void CreateCellReferenceRangeWithAbsoluteWorksheetTest3()
        {
            this.worksheet2.Cells[0, 0].SetValue(20);

            AssertWorksheetCellReferenceRangeCreated("=Sheet2!A1", this.worksheet1, new CellIndex(1, 1), 20);
        }

        [TestMethod]
        public void CreateCellReferenceRangeWithAbsoluteWorksheetTest4()
        {
            this.worksheet2.Cells[0, 0].SetValue(20);

            AssertWorksheetCellReferenceRangeCreated("=Sheet2!$A$1", this.worksheet1, new CellIndex(1, 1), 20);
        }

        [TestMethod]
        public void CreateCellReferenceRangeWithAbsoluteWorksheetTest5()
        {
            this.worksheet1.Cells[0, 0].SetValue(10);

            AssertWorksheetCellReferenceRangeCreated("='Sheet1'!A1", this.worksheet1, new CellIndex(1, 1), 10);
        }

        [TestMethod]
        public void CreateInvalidRangeWithAbsoluteWorksheetTest()
        {
            this.worksheet1.Cells[0, 0].SetValue(10);

            AssertWorksheetCellReferenceRangeError("=Sheet2!#REF!", this.worksheet1, new CellIndex(1, 1), ErrorExpressions.ReferenceError);
        }

        [TestMethod]
        public void CreateCellReferenceRangeWithInvalidWorksheetTest()
        {
            AssertWorksheetCellReferenceRangeError("=NoSheet!A1", this.worksheet1, new CellIndex(1, 1), ErrorExpressions.ReferenceError);
        }

        [TestMethod]
        public void CreateCellReferenceRangeWithInvalidWorksheetTest2()
        {
            AssertWorksheetCellReferenceRangeError("='NoSheet'!A1", this.worksheet1, new CellIndex(1, 1), ErrorExpressions.ReferenceError);
        }

        [TestMethod]
        public void CreateCellReferenceRangeWithSpecialWorksheetName1()
        {
            this.worksheet2.Name = " ";
            this.worksheet2.Cells[0, 0].SetValue(123);

            AssertWorksheetCellReferenceRangeCreated("=' '!A1", this.worksheet1, new CellIndex(1, 1), 123);
        }

        [TestMethod]
        public void CreateCellReferenceRangeWithSpecialWorksheetName2()
        {
            this.worksheet2.Name = "-#!";
            this.worksheet2.Cells[0, 0].SetValue(123);

            AssertWorksheetCellReferenceRangeCreated("='-#!'!A1", this.worksheet1, new CellIndex(1, 1), 123);
        }

        [TestMethod]
        public void CreateCellReferenceRangeWithSpecialWorksheetName3()
        {
            this.worksheet2.Name = "-+=&@#%^()!.<>~;  __";
            this.worksheet2.Cells[0, 0].SetValue(123);

            AssertWorksheetCellReferenceRangeCreated("='-+=&@#%^()!.<>~;  __'!A1", this.worksheet1, new CellIndex(1, 1), 123);
        }

        [TestMethod]
        public void CreateCellReferenceRangeWithSpecialWorksheetName4()
        {
            this.worksheet2.Name = "_Sheet";
            this.worksheet2.Cells[0, 0].SetValue(123);

            AssertWorksheetCellReferenceRangeCreated("=_Sheet!A1", this.worksheet1, new CellIndex(1, 1), 123);
        }

        [TestMethod]
        public void CreateCellReferenceRangeWithSpecialWorksheetName5()
        {
            this.worksheet2.Name = "_Sheet";
            this.worksheet2.Cells[0, 0].SetValue(123);

            AssertWorksheetCellReferenceRangeCreated("='_Sheet'!A1", this.worksheet1, new CellIndex(1, 1), 123);
        }

        [TestMethod]
        public void CreateCellReferenceRangeWithSpecialWorksheetName6()
        {
            this.worksheet2.Name = "_";
            this.worksheet2.Cells[0, 0].SetValue(123);

            AssertWorksheetCellReferenceRangeCreated("=_!A1", this.worksheet1, new CellIndex(1, 1), 123);
        }

        [TestMethod]
        public void CreateCellReferenceRangeWithSpecialWorksheetName7()
        {
            this.worksheet2.Name = "_123";
            this.worksheet2.Cells[0, 0].SetValue(123);

            AssertWorksheetCellReferenceRangeCreated("=_123!A1", this.worksheet1, new CellIndex(1, 1), 123);
        }

        #endregion


        #region Cell Reference Range With Absolute and Relative Worksheets

        [TestMethod]
        public void CreateRangeLeftAbsoluteRightAbsoluteSameWorksheet()
        {
            this.worksheet1.Cells[0, 0].SetValue(6);
            this.worksheet1.Cells[1, 0].SetValue(5);

            AssertWorksheetCellReferenceRangeCreated("=SUM(Sheet1!A1:Sheet1!A2)", this.worksheet1, new CellIndex(1, 1), 11);
        }

        [TestMethod]
        public void CreateRangeLeftAbsoluteRightAbsoluteSameWorksheet2()
        {
            this.worksheet1.Cells[0, 0].SetValue(6);
            this.worksheet1.Cells[1, 0].SetValue(5);

            AssertWorksheetCellReferenceRangeCreated("=SUM(Sheet1!A1:Sheet1!A2)", this.worksheet2, new CellIndex(1, 1), 11);
        }

        [TestMethod]
        public void CreateRangeLeftAbsoluteRightAbsoluteDifferentWorksheets()
        {
            AssertWorksheetCellReferenceRangeError("=SUM(Sheet1!A1:Sheet2!A2)", this.worksheet1, new CellIndex(1, 1), ErrorExpressions.ValueError);
        }

        [TestMethod]
        public void CreateRangeLeftAbsoluteRightAbsoluteDifferentWorksheets2()
        {
            AssertWorksheetCellReferenceRangeError("=SUM(Sheet1!A1:Sheet2!A2)", this.worksheet2, new CellIndex(1, 1), ErrorExpressions.ValueError);
        }

        [TestMethod]
        public void CreateRangeLeftAbsoluteRightAbsoluteDifferentWorksheets3()
        {
            AssertWorksheetCellReferenceRangeError("=SUM(Sheet1!A1:NoSuchSheet!A2)", this.worksheet2, new CellIndex(1, 1), ErrorExpressions.ValueError);
        }

        [TestMethod]
        public void CreateRangeLeftAbsoluteRightRelativeSameWorksheet()
        {
            this.worksheet1.Cells[0, 0].SetValue(6);
            this.worksheet1.Cells[1, 0].SetValue(5);

            AssertWorksheetCellReferenceRangeCreated("=SUM(Sheet1!A1:A2)", this.worksheet1, new CellIndex(1, 1), 11);
        }

        [TestMethod]
        public void CreateRangeLeftAbsoluteRightRelativeSameWorksheet2()
        {
            this.worksheet1.Cells[0, 0].SetValue(6);
            this.worksheet1.Cells[1, 0].SetValue(5);

            AssertWorksheetCellReferenceRangeCreated("=SUM(Sheet1!A1:A2)", this.worksheet2, new CellIndex(1, 1), 11);
        }

        [TestMethod]
        public void CreateRangeLeftRelativeRightAbsoluteDifferentWorksheets()
        {
            AssertWorksheetCellReferenceRangeError("=SUM(A1:Sheet2!A2)", this.worksheet1, new CellIndex(1, 1), ErrorExpressions.ValueError);
        }

        [TestMethod]
        public void CreateRangeLeftRelativeRightAbsoluteDifferentWorksheets2()
        {
            this.worksheet2.Cells[0, 0].SetValue(6);
            this.worksheet2.Cells[1, 0].SetValue(5);

            AssertWorksheetCellReferenceRangeCreated("=SUM(A1:Sheet2!A2)", this.worksheet2, new CellIndex(1, 1), 11);
        }

        #endregion


        #region Creating Ranges, Unions and Intersections

        [TestMethod]
        public void CreateIntersectionWithOneCellRanges()
        {
            AssertWorksheetCellReferenceRangeError("=SUM(A1 B1)", this.worksheet1, new CellIndex(1, 1), ErrorExpressions.NullError);
        }

        [TestMethod]
        public void CreateUnionWithCrossWorksheetReferences()
        {
            AssertWorksheetCellReferenceRangeError("=SUM((A1 {ls} Sheet2!A1))", this.worksheet1, new CellIndex(1, 1), ErrorExpressions.ValueError);
        }

        [TestMethod]
        public void CreateUnionWithNonExistingWorksheetReferences()
        {
            AssertWorksheetCellReferenceRangeError("=SUM((A1 {ls} NonExistingSheet!A1))", this.worksheet1, new CellIndex(1, 1), ErrorExpressions.ValueError);
        }

        [TestMethod]
        public void CreateUnionWithRef()
        {
            AssertWorksheetCellReferenceRangeError("=SUM((#REF! {ls} Sheet2!A1))", this.worksheet1, new CellIndex(1, 1), ErrorExpressions.ReferenceError);
            AssertWorksheetCellReferenceRangeError("=SUM((#REF! {ls} #REF!))", this.worksheet1, new CellIndex(1, 1), ErrorExpressions.ReferenceError);
            AssertWorksheetCellReferenceRangeError("=SUM((Sheet1!#REF! {ls} Sheet2!#REF!))", this.worksheet1, new CellIndex(1, 1), ErrorExpressions.ReferenceError);
        }

        [TestMethod]
        public void CreateUnionWithRef2()
        {
            AssertWorksheetCellReferenceRangeError("=SUM((#REF! {ls} Sheet2!A1))", this.worksheet1, new CellIndex(1, 1), ErrorExpressions.ReferenceError);
        }

        [TestMethod]
        public void CreateSameCellUnion()
        {
            this.worksheet2.Cells[0, 0].SetValue(1);

            AssertWorksheetCellReferenceRangeCreated("=SUM((A1 {ls} Sheet2!A1))", this.worksheet2, new CellIndex(1, 1), 2);
        }

        #endregion


        #region Cell Reference Ranges Cross Worksheet Translation

        [TestMethod]
        public void CellReferenceRangeWithRelativeWorksheetTranslationTest1()
        {
            this.worksheet2.Cells[0, 0].SetValue(20);

            RadExpression expression = ParseExpression("=A1", this.worksheet1, new CellIndex(1, 1));
            AssertTranslatedWorksheetCellReferenceRange(expression, this.worksheet2, new CellIndex(1, 1), 20, "A1", 0, 0, 0, 0);
        }

        [TestMethod]
        public void CellReferenceRangeWithRelativeWorksheetTranslationTest2()
        {
            this.worksheet2.Cells[0, 0].SetValue(20);

            RadExpression expression = ParseExpression("=$A$1", this.worksheet1, new CellIndex(1, 1));
            AssertTranslatedWorksheetCellReferenceRange(expression, this.worksheet2, new CellIndex(5, 5), 20, "$A$1", 0, 0, 0, 0);
        }

        [TestMethod]
        public void CellReferenceRangeWithAbsoluteWorksheetTranslationTest1()
        {
            this.worksheet1.Cells[0, 0].SetValue(10);

            RadExpression expression = ParseExpression("=Sheet1!A1", this.worksheet1, new CellIndex(1, 1));
            AssertTranslatedWorksheetCellReferenceRange(expression, this.worksheet2, new CellIndex(1, 1), 10, "Sheet1!A1", 0, 0, 0, 0);
        }

        [TestMethod]
        public void CellReferenceRangeWithAbsoluteWorksheetTranslationTest2()
        {
            this.worksheet1.Cells[0, 0].SetValue(10);

            RadExpression expression = ParseExpression("=Sheet1!$A$1", this.worksheet1, new CellIndex(1, 1));
            AssertTranslatedWorksheetCellReferenceRange(expression, this.worksheet2, new CellIndex(10, 10), 10, "Sheet1!$A$1", 0, 0, 0, 0);
        }

        [TestMethod]
        public void CellReferenceRangeWithAbsoluteWorksheetTranslationTest3()
        {
            this.worksheet2.Cells[0, 0].SetValue(20);

            RadExpression expression = ParseExpression("=Sheet2!A1", this.worksheet1, new CellIndex(1, 1));
            AssertTranslatedWorksheetCellReferenceRange(expression, this.worksheet2, new CellIndex(1, 1), 20, "Sheet2!A1", 0, 0, 0, 0);
        }

        [TestMethod]
        public void CellReferenceRangeWithAbsoluteWorksheetTranslationTest4()
        {
            this.worksheet2.Cells[0, 0].SetValue(20);

            RadExpression expression = ParseExpression("=Sheet2!$A$1", this.worksheet1, new CellIndex(1, 1));
            AssertTranslatedWorksheetCellReferenceRange(expression, this.worksheet2, new CellIndex(15, 15), 20, "Sheet2!$A$1", 0, 0, 0, 0);
        }

        [TestMethod]
        public void CellReferenceRangeWithAbsoluteWorksheetTranslationTest5()
        {
            this.worksheet2.Cells[0, 1].SetValue(20);

            RadExpression expression = ParseExpression("=Sheet2!A1", this.worksheet1, new CellIndex(1, 1));
            AssertTranslatedWorksheetCellReferenceRange(expression, this.worksheet2, new CellIndex(1, 2), 20, "Sheet2!B1", 0, 1, 0, 1);
        }

        [TestMethod]
        public void CellReferenceRangeWithAbsoluteWorksheetTranslationTest6()
        {
            this.worksheet2.Cells[0, 0].SetValue(10);

            RadExpression expression = ParseExpression("=Sheet2!$A$1", this.worksheet1, new CellIndex(1, 1));
            AssertTranslatedWorksheetCellReferenceRange(expression, this.worksheet2, new CellIndex(1, 2), 10, "Sheet2!$A$1", 0, 0, 0, 0);
        }

        [TestMethod]
        public void CellReferenceRangeWithInvalidWorksheetTranslationTest()
        {
            RadExpression expression = ParseExpression("=NoSheet!A1", this.worksheet1, new CellIndex(1, 1));
            AssertTranslatedWorksheetCellReferenceRangeError(expression, this.worksheet1, new CellIndex(2, 2), "NoSheet!B2");
        }

        [TestMethod]
        public void InvalidCellReferenceRangeWithAbsoluteWorksheetTranslationTest()
        {
            RadExpression expression = ParseExpression("=Sheet2!A1", this.worksheet1, new CellIndex(2, 2));
            AssertTranslatedWorksheetCellReferenceRangeError(expression, this.worksheet1, new CellIndex(0, 0), "Sheet2!#REF!");
        }

        [TestMethod]
        public void TranslateCellReferenceRangeWithSpecialWorksheetName1()
        {
            this.worksheet1.Cells[0, 0].SetValue(10);

            RadExpression expression = ParseExpression("='Sheet1'!A1", this.worksheet1, new CellIndex(1, 1));
            AssertTranslatedWorksheetCellReferenceRange(expression, this.worksheet2, new CellIndex(1, 1), 10, "'Sheet1'!A1", 0, 0, 0, 0);
        }

        [TestMethod]
        public void TranslateCellReferenceRangeWithSpecialWorksheetName2()
        {
            this.worksheet1.Cells[0, 0].SetValue(10);
            this.worksheet1.Name = "Sh''eet";

            RadExpression expression = ParseExpression("='Sh''''eet'!A1", this.worksheet1, new CellIndex(1, 1));
            AssertTranslatedWorksheetCellReferenceRange(expression, this.worksheet2, new CellIndex(1, 1), 10, "'Sh''''eet'!A1", 0, 0, 0, 0);
        }

        [TestMethod]
        public void TranslateCellReferenceRangeWithSpecialWorksheetName3()
        {
            this.worksheet1.Cells[0, 0].SetValue(10);
            this.worksheet1.Name = "S'h'e'e't";

            RadExpression expression = ParseExpression("='S''h''e''e''t'!A1", this.worksheet1, new CellIndex(1, 1));
            AssertTranslatedWorksheetCellReferenceRange(expression, this.worksheet2, new CellIndex(1, 1), 10, "'S''h''e''e''t'!A1", 0, 0, 0, 0);
        }

        [TestMethod]
        public void TranslateCellReferenceRangeWithSpecialWorksheetName4()
        {
            this.worksheet1.Cells[0, 0].SetValue(10);
            this.worksheet1.Name = "Sam's Worksheet";

            RadExpression expression = ParseExpression("='Sam''s Worksheet'!A1", this.worksheet1, new CellIndex(1, 1));
            AssertTranslatedWorksheetCellReferenceRange(expression, this.worksheet2, new CellIndex(1, 1), 10, "'Sam''s Worksheet'!A1", 0, 0, 0, 0);
        }

        [TestMethod]
        public void TranslateCellReferenceRangeWithSpecialWorksheetName5()
        {
            this.worksheet1.Cells[0, 0].SetValue(10);
            this.worksheet1.Name = "Sam's Worksheet #!";

            RadExpression expression = ParseExpression("='Sam''s Worksheet #!'!A1", this.worksheet1, new CellIndex(1, 1));
            AssertTranslatedWorksheetCellReferenceRange(expression, this.worksheet2, new CellIndex(1, 1), 10, "'Sam''s Worksheet #!'!A1", 0, 0, 0, 0);
        }

        [TestMethod]
        public void TranslateCellReferenceRangeWithSpecialWorksheetName6()
        {
            this.worksheet2.Name = " ";
            this.worksheet2.Cells[1, 1].SetValue(234);

            RadExpression expression = ParseExpression("=' '!A1", this.worksheet1, new CellIndex(0, 0));
            AssertTranslatedWorksheetCellReferenceRange(expression, this.worksheet1, new CellIndex(1, 1), 234, "' '!B2", 1, 1, 1, 1);
        }

        [TestMethod]
        public void TranslateCellReferenceRangeWithSpecialWorksheetName7()
        {
            this.worksheet2.Name = "_";
            this.worksheet2.Cells[2, 2].SetValue(48);

            RadExpression expression = ParseExpression("='_'!A1", this.worksheet1, new CellIndex(0, 0));
            AssertTranslatedWorksheetCellReferenceRange(expression, this.worksheet1, new CellIndex(2, 2), 48, "'_'!C3", 2, 2, 2, 2);
        }

        [TestMethod]
        public void TranslateCellReferenceUnionToSameWorksheet()
        {
            AssertTranslateExpressionStringCorrectness("=Sheet1!B1{ls}C1", this.worksheet1, new CellIndex(0, 0),
                "=Sheet1!B3{ls}C3", this.worksheet1, new CellIndex(2, 0));
        }

        [TestMethod]
        public void TranslateCellReferenceUnionInFunctionToSameWorksheet()
        {
            AssertTranslateExpressionStringCorrectness("=SUM((Sheet1!B1{ls}Sheet1!C1))", this.worksheet1, new CellIndex(0, 0),
               "=SUM((Sheet1!B3{ls}Sheet1!C3))", this.worksheet1, new CellIndex(2, 0));
        }

        [TestMethod]
        public void TranslateMultipleUnionInFunctionToSameWorksheet()
        {
            AssertTranslateExpressionStringCorrectness("=SUM(((Sheet1!B1{ls} F6){ls} Sheet1!C1))", this.worksheet1, new CellIndex(0, 0),
               "=SUM(((Sheet1!B3{ls} F8){ls} Sheet1!C3))", this.worksheet1, new CellIndex(2, 0));
        }

        [TestMethod]
        public void TranslateHalfAbsoluteUnionToOtherWorksheet()
        {
            AssertTranslateExpressionStringCorrectness("=Sheet1!B1{ls} C2", this.worksheet1, new CellIndex(0, 0),
               "=Sheet1!D3{ls} E4", this.worksheet2, new CellIndex(2, 2));
        }

        [TestMethod]
        public void TranslateRelativeUnionToOtherWorksheet()
        {
            AssertTranslateExpressionStringCorrectness("=B1{ls} C2", this.worksheet1, new CellIndex(0, 0),
               "=D3{ls} E4", this.worksheet2, new CellIndex(2, 2));
        }

        #endregion


        #region Cell Reference Range Cross Workbook Translation

        [TestMethod]
        public void TranslateCellReferenceRangeToAnotherWorkbook()
        {
            Worksheet targetWorksheet = TestHelper.CreateWorksheet();

            CellIndex cellIndexA1 = new CellIndex(0, 0);

            RadExpression expression = ParseExpression("=B1", this.worksheet1, cellIndexA1);
            CellReferenceRangeExpression translatedExpression = (CellReferenceRangeExpression)expression.CloneAndTranslate(
                targetWorksheet, cellIndexA1);

            Assert.AreEqual(targetWorksheet.Workbook, translatedExpression.Worksheet.Workbook);
        }

        #endregion
    }
}
