﻿using System;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal
{
    [TestClass]
    public class CellReferenceInsertRemoveCellsTests
    {
        #region Fields

        private Worksheet worksheet;

        private readonly CellIndex A1 = new CellIndex(0, 0);
        private readonly CellIndex A2 = new CellIndex(1, 0);
        private readonly CellIndex A4 = new CellIndex(3, 0);

        private readonly CellIndex B4 = new CellIndex(3, 1);
        private readonly CellIndex C4 = new CellIndex(3, 2);
        private readonly CellIndex D4 = new CellIndex(3, 3);
        private readonly CellIndex F4 = new CellIndex(3, 5);
        private readonly CellIndex A3 = new CellIndex(2, 0);

        private readonly CellRange A1B2 = new CellRange(0, 0, 1, 1);
        private readonly CellRange A4C4 = new CellRange(3, 0, 3, 2);
        private readonly CellRange A1A2 = new CellRange(0, 0, 1, 0);
        private readonly CellRange D1D2 = new CellRange(0, 3, 1, 3);
        private readonly CellRange A3B3 = new CellRange(2, 0, 2, 1);
        private readonly CellRange E1F2 = new CellRange(0, 4, 1, 5);
        private readonly CellRange C1D2 = new CellRange(0, 2, 1, 3);
        private readonly CellRange B1C4 = new CellRange(0, 1, 3, 2);

        #endregion


        #region Helper Methods

        private void SetValueInWorksheet(CellIndex index, string value)
        {
            this.worksheet.Cells[index].SetValue(value);
        }

        private void InsertCells(CellRange range, InsertShiftType insertShiftType)
        {
            this.worksheet.Cells[range].Insert(insertShiftType);
        }

        private void InsertCells(CellIndex index, InsertShiftType insertShiftType)
        {
            this.worksheet.Cells[index].Insert(insertShiftType);
        }

        private void RemoveCells(CellRange range, RemoveShiftType removeShiftType)
        {
            this.worksheet.Cells[range].Remove(removeShiftType);
        }

        private void RemoveCells(CellIndex index, RemoveShiftType removeShiftType)
        {
            this.worksheet.Cells[index].Remove(removeShiftType);
        }

        private void AssertExpressionCorrectness(CellIndex index, string expectedValue)
        {
            CellRange range;
            NameConverter.TryConvertCellNameToCellRange(expectedValue, out range);

            FormulaCellValue cellValue = (FormulaCellValue)this.worksheet.Cells[index].GetValue().Value;
            string valueAsString = SpreadsheetCultureHelper.ClearFormulaValue(cellValue.RawValue);
            Assert.AreEqual(expectedValue, valueAsString);

            CellReferenceRangeExpression cellReference = (CellReferenceRangeExpression)cellValue.Value;

            AssertExpressionPointsToRange(cellReference, range);
        }

        private static void AssertExpressionPointsToRange(CellReferenceRangeExpression expression, CellRange cellRange)
        {
            Assert.AreEqual(cellRange.FromIndex.RowIndex, expression.CellReferenceRange.FromCellReference.RowReference.ActualIndex);
            Assert.AreEqual(cellRange.FromIndex.ColumnIndex, expression.CellReferenceRange.FromCellReference.ColumnReference.ActualIndex);
            Assert.AreEqual(cellRange.ToIndex.RowIndex, expression.CellReferenceRange.ToCellReference.RowReference.ActualIndex);
            Assert.AreEqual(cellRange.ToIndex.ColumnIndex, expression.CellReferenceRange.ToCellReference.ColumnReference.ActualIndex);
        }

        #endregion


        #region Test Initialize and Cleanup

        [TestInitialize]
        public void TestInitialize()
        {
            this.worksheet = TestHelper.CreateWorksheet();
        }

        [TestCleanup]
        public void TestCleanup()
        {
            this.worksheet = null;
        }

        #endregion


        [TestMethod]
        public void InsertRightTranslateTargetPositiveOffset()
        {
            SetValueInWorksheet(B4, "=C1:D2");

            InsertCells(A1A2, InsertShiftType.Right);

            AssertExpressionCorrectness(B4, "D1:E2");
        }

        [TestMethod]
        public void RemoveLeftTranslateTargetPositiveOffset()
        {
            SetValueInWorksheet(B4, "=D1:E2");

            RemoveCells(A1A2, RemoveShiftType.Left);

            AssertExpressionCorrectness(B4, "C1:D2");
        }

        [TestMethod]
        public void InsertRightTranslateTargetNegativeOffset()
        {
            SetValueInWorksheet(C4, "=A1:B2");

            InsertCells(A1A2, InsertShiftType.Right);

            AssertExpressionCorrectness(C4, "B1:C2");
        }

        [TestMethod]
        public void RemoveLeftTranslateTargetNegativeOffset()
        {
            SetValueInWorksheet(C4, "=B1:C2");

            RemoveCells(A1A2, RemoveShiftType.Left);

            AssertExpressionCorrectness(C4, "A1:B2");
        }

        [TestMethod]
        public void InsertRightCellsBetweenSourceAndTargetPositiveOffset()
        {
            SetValueInWorksheet(B4, "=C1:D2");

            InsertCells(A3B3, InsertShiftType.Right);

            AssertExpressionCorrectness(B4, "C1:D2");
        }

        [TestMethod]
        public void RemoveLeftCellsBetweenSourceAndTargetPositiveOffset()
        {
            SetValueInWorksheet(B4, "=C1:D2");

            RemoveCells(A3B3, RemoveShiftType.Left);

            AssertExpressionCorrectness(B4, "C1:D2");
        }

        [TestMethod]
        public void InsertRightCellsBetweenSourceAndTargetNegativeOffset()
        {
            SetValueInWorksheet(C4, "=A1:B2");

            InsertCells(A3B3, InsertShiftType.Right);

            AssertExpressionCorrectness(C4, "A1:B2");
        }

        [TestMethod]
        public void RemoveLeftCellsBetweenSourceAndTargetNegativeOffset()
        {
            SetValueInWorksheet(C4, "=A1:B2");

            RemoveCells(A3B3, RemoveShiftType.Left);

            AssertExpressionCorrectness(C4, "A1:B2");

        }

        [TestMethod]
        public void InsertRightOneCellTranslateSourcePositiveOffset()
        {
            SetValueInWorksheet(B4, "=C1:D2");

            InsertCells(A4, InsertShiftType.Right);

            AssertExpressionCorrectness(C4, "C1:D2");
        }

        [TestMethod]
        public void RemoveLeftOneCellTranslateSourcePositiveOffset()
        {
            SetValueInWorksheet(C4, "=C1:D2");

            RemoveCells(A4, RemoveShiftType.Left);

            AssertExpressionCorrectness(B4, "C1:D2");
        }

        [TestMethod]
        public void InsertRightOneCellTranslateSourceNegativeOffset()
        {
            SetValueInWorksheet(C4, "=A1:B2");

            InsertCells(A4, InsertShiftType.Right);

            AssertExpressionCorrectness(D4, "A1:B2");
        }

        [TestMethod]
        public void RemoveLeftOneCellTranslateSourceNegativeOffset()
        {
            SetValueInWorksheet(D4, "=A1:B2");

            RemoveCells(A4, RemoveShiftType.Left);

            AssertExpressionCorrectness(C4, "A1:B2");
        }

        [TestMethod]
        public void InsertRightThreeCellsTranslateSourcePositiveOffset()
        {
            CellIndex E4 = new CellIndex(3, 4);
            SetValueInWorksheet(B4, "=C1:D2");

            InsertCells(A4C4, InsertShiftType.Right);

            AssertExpressionCorrectness(E4, "C1:D2");
        }

        [TestMethod]
        public void RemoveLeftThreeCellsTranslateSourcePositiveOffset()
        {
            CellIndex E4 = new CellIndex(3, 4);
            SetValueInWorksheet(E4, "=C1:D2");

            RemoveCells(A4C4, RemoveShiftType.Left);

            AssertExpressionCorrectness(B4, "C1:D2");

        }
        [TestMethod]
        public void InsertRightThreeCellsTranslateSourceNegativeOffset()
        {
            SetValueInWorksheet(C4, "=A1:B2");

            InsertCells(A4C4, InsertShiftType.Right);

            AssertExpressionCorrectness(F4, "A1:B2");
        }

        [TestMethod]
        public void RemoveLeftThreeCellsTranslateSourceNegativeOffset()
        {
            SetValueInWorksheet(F4, "=A1:B2");

            RemoveCells(A4C4, RemoveShiftType.Left);

            AssertExpressionCorrectness(C4, "A1:B2");
        }

        [TestMethod]
        public void InsertRightTranslateHalfTargetPositiveOffset()
        {
            SetValueInWorksheet(B4, "=C1:D2");

            InsertCells(A1, InsertShiftType.Right);

            AssertExpressionCorrectness(B4, "C1:D2");
        }

        [TestMethod]
        public void RemoveLeftTranslateHalfTargetPositiveOffset()
        {
            SetValueInWorksheet(B4, "=C1:D2");

            RemoveCells(A1, RemoveShiftType.Left);

            AssertExpressionCorrectness(B4, "C1:D2");
        }

        [TestMethod]
        public void InsertRightTranslateHalfTargetNegativeOffset()
        {
            SetValueInWorksheet(C4, "=A1:B2");

            InsertCells(A1, InsertShiftType.Right);

            AssertExpressionCorrectness(C4, "A1:B2");
        }

        [TestMethod]
        public void RemoveLeftTranslateHalfTargetNegativeOffset()
        {
            SetValueInWorksheet(C4, "=A1:B2");

            RemoveCells(A1, RemoveShiftType.Left);

            AssertExpressionCorrectness(C4, "A1:B2");
        }

        [TestMethod]
        public void InsertRightTranslateHalfTargetAndSourcePositiveOffset()
        {
            CellRange A2A4 = new CellRange(1, 0, 3, 0);
            SetValueInWorksheet(B4, "=C1:D2");

            InsertCells(A2A4, InsertShiftType.Right);

            AssertExpressionCorrectness(C4, "C1:D2");
        }

        [TestMethod]
        public void RemoveLeftTranslateHalfTargetAndSourcePositiveOffset()
        {
            CellRange A2A4 = new CellRange(1, 0, 3, 0);
            SetValueInWorksheet(C4, "=C1:D2");

            RemoveCells(A2A4, RemoveShiftType.Left);

            AssertExpressionCorrectness(B4, "C1:D2");
        }

        [TestMethod]
        public void InsertRightTranslateAfterSourcePositiveOffset()
        {
            SetValueInWorksheet(B4, "=C1:D2");

            InsertCells(C4, InsertShiftType.Right);

            AssertExpressionCorrectness(B4, "C1:D2");
        }

        [TestMethod]
        public void RemoveLeftTranslateAfterSourcePositiveOffset()
        {
            SetValueInWorksheet(B4, "=C1:D2");

            RemoveCells(C4, RemoveShiftType.Left);

            AssertExpressionCorrectness(B4, "C1:D2");
        }

        [TestMethod]
        public void InsertRightTranslateAfterSourceNegativeOffsety()
        {
            SetValueInWorksheet(C4, "=A1:B2");

            InsertCells(D4, InsertShiftType.Right);

            AssertExpressionCorrectness(C4, "A1:B2");
        }

        [TestMethod]
        public void RemoveLeftTranslateAfterSourceNegativeOffsety()
        {
            SetValueInWorksheet(C4, "=A1:B2");

            RemoveCells(D4, RemoveShiftType.Left);

            AssertExpressionCorrectness(C4, "A1:B2");
        }

        [TestMethod]
        public void InsertRightTranslateAfterTargetPositiveOffset()
        {
            SetValueInWorksheet(B4, "=C1:D2");

            InsertCells(E1F2, InsertShiftType.Right);

            AssertExpressionCorrectness(B4, "C1:D2");
        }

        [TestMethod]
        public void RemoveLeftTranslateAfterTargetPositiveOffset()
        {
            SetValueInWorksheet(B4, "=C1:D2");

            RemoveCells(E1F2, RemoveShiftType.Left);

            AssertExpressionCorrectness(B4, "C1:D2");
        }

        [TestMethod]
        public void InsertRightTranslateAfterTargetNegativeOffset()
        {
            SetValueInWorksheet(C4, "=A1:B2");

            InsertCells(C1D2, InsertShiftType.Right);

            AssertExpressionCorrectness(C4, "A1:B2");
        }

        [TestMethod]
        public void RemoveLeftTranslateAfterTargetNegativeOffset()
        {
            SetValueInWorksheet(C4, "=A1:B2");

            RemoveCells(C1D2, RemoveShiftType.Left);

            AssertExpressionCorrectness(C4, "A1:B2");
        }

        [TestMethod]
        public void InsertRightTranslateAfterSourceAndTargetPositiveOffset()
        {
            CellRange E1F6 = new CellRange(0, 4, 5, 5);
            SetValueInWorksheet(B4, "=C1:D2");

            InsertCells(E1F6, InsertShiftType.Right);

            AssertExpressionCorrectness(B4, "C1:D2");
        }

        [TestMethod]
        public void RemoveLeftTranslateAfterSourceAndTargetPositiveOffset()
        {
            CellRange E1F6 = new CellRange(0, 4, 5, 5);
            SetValueInWorksheet(B4, "=C1:D2");

            RemoveCells(E1F6, RemoveShiftType.Left);

            AssertExpressionCorrectness(B4, "C1:D2");
        }

        [TestMethod]
        public void InsertRightTranslateAfterSourceAndTargetNegativeOffset()
        {
            CellRange D1E6 = new CellRange(0, 3, 5, 4);
            SetValueInWorksheet(C4, "=A1:B2");

            InsertCells(D1E6, InsertShiftType.Right);

            AssertExpressionCorrectness(C4, "A1:B2");
        }

        [TestMethod]
        public void RemoveLeftTranslateAfterSourceAndTargetNegativeOffset()
        {
            CellRange D1E6 = new CellRange(0, 3, 5, 4);
            SetValueInWorksheet(C4, "=A1:B2");

            RemoveCells(D1E6, RemoveShiftType.Left);

            AssertExpressionCorrectness(C4, "A1:B2");
        }

        [TestMethod]
        public void InsertRightInTarget()
        {
            SetValueInWorksheet(B4, "=C1:D2");

            InsertCells(D1D2, InsertShiftType.Right);

            AssertExpressionCorrectness(B4, "C1:E2");
        }

        [TestMethod]
        public void InsertRightTwoCellsInTarget()
        {
            SetValueInWorksheet(B4, "=C1:E2");

            InsertCells(C1D2, InsertShiftType.Right);

            AssertExpressionCorrectness(B4, "E1:G2");
        }

        [TestMethod]
        public void InsertRightInTargetAndBeforeSource()
        {
            CellRange B1B4 = new CellRange(0, 1, 3, 1);
            SetValueInWorksheet(C4, "=A1:D2");

            InsertCells(B1B4, InsertShiftType.Right);

            AssertExpressionCorrectness(D4, "A1:E2");
        }

        [TestMethod]
        public void InsertRightThreeCellsInTargetAndBeforeSource()
        {
            CellIndex G2 = new CellIndex(3, 6);
            CellRange B1D4 = new CellRange(0, 1, 3, 3);
            SetValueInWorksheet(D4, "=A1:D2");

            InsertCells(B1D4, InsertShiftType.Right);

            AssertExpressionCorrectness(G2, "A1:G2");
        }

        [TestMethod]
        public void InsertRightInTargetAndAnfer()
        {
            CellRange D1F2 = new CellRange(0, 3, 1, 5);
            SetValueInWorksheet(B4, "=C1:D2");

            InsertCells(D1F2, InsertShiftType.Right);

            AssertExpressionCorrectness(B4, "C1:G2");
        }

        [TestMethod]
        public void InsertRightBeforeTargetAndAfter()
        {
            CellRange B1E2 = new CellRange(0, 1, 1, 4);
            SetValueInWorksheet(B4, "=C1:D2");

            InsertCells(B1E2, InsertShiftType.Right);

            AssertExpressionCorrectness(B4, "G1:H2");
        }

        [TestMethod]
        public void InsertRightBeforeAndInTarget()
        {
            CellRange B1D2 = new CellRange(0, 1, 1, 3);
            SetValueInWorksheet(B4, "=C1:D2");

            InsertCells(B1D2, InsertShiftType.Right);

            AssertExpressionCorrectness(B4, "F1:G2");
        }

        [TestMethod]
        public void InsertRightBeforeAndInTargetAndSource()
        {
            SetValueInWorksheet(D4, "=A1:B2");

            InsertCells(B1C4, InsertShiftType.Right);

            AssertExpressionCorrectness(F4, "A1:D2");
        }

        [TestMethod]
        public void RemoveLeftInTarget()
        {
            SetValueInWorksheet(B4, "=C1:E2");

            RemoveCells(D1D2, RemoveShiftType.Left);

            AssertExpressionCorrectness(B4, "C1:D2");
        }

        [TestMethod]
        public void RemoveLeftOneInAndOneAfterTarget()
        {
            SetValueInWorksheet(B4, "=C1:E2");

            RemoveCells(E1F2, RemoveShiftType.Left);

            AssertExpressionCorrectness(B4, "C1:D2");
        }

        [TestMethod]
        public void RemoveLeftTwoInAndOneAfterTarget()
        {
            CellRange D1F2 = new CellRange(0, 3, 1, 5);
            SetValueInWorksheet(B4, "=C1:E2");

            RemoveCells(D1F2, RemoveShiftType.Left);

            AssertExpressionCorrectness(B4, "C1:C2");
        }

        [TestMethod]
        public void RemoveLeftOneInAndTwoAfterTarget()
        {
            CellRange E1G2 = new CellRange(0, 4, 1, 6);
            SetValueInWorksheet(B4, "=C1:E2");

            RemoveCells(E1G2, RemoveShiftType.Left);

            AssertExpressionCorrectness(B4, "C1:D2");
        }

        [TestMethod]
        public void RemoveLeftWholeTarget()
        {
            CellRange B1E2 = new CellRange(0, 1, 1, 4);
            SetValueInWorksheet(B4, "=C1:E2");

            RemoveCells(B1E2, RemoveShiftType.Left);

            string value = ((FormulaCellValue)this.worksheet.Cells[B4].GetValue().Value).GetResultValueAsString(CellValueFormat.GeneralFormat);
            Assert.AreEqual(SpreadsheetCultureHelper.ReferenceErrorValue, value);
        }

        [TestMethod]
        public void RemoveLeftInTargetPositiveOffset()
        {
            SetValueInWorksheet(B4, "=C1:D2");

            RemoveCells(D1D2, RemoveShiftType.Left);

            AssertExpressionCorrectness(B4, "C1:C2");
        }

        [TestMethod]
        public void InsertRightTranslateInTargetNegativeOffset()
        {
            SetValueInWorksheet(C4, "=A1:B2");

            InsertCells(A1B2, InsertShiftType.Right);

            AssertExpressionCorrectness(C4, "C1:D2");
        }

        [TestMethod]
        public void RemoveLeftTranslateInTargetNegativeOffset()
        {
            SetValueInWorksheet(C4, "=C1:D2");

            RemoveCells(A1B2, RemoveShiftType.Left);

            AssertExpressionCorrectness(C4, "A1:B2");
        }

        [TestMethod]
        public void InsertDownTranslateTargetPositiveOffset()
        {
            SetValueInWorksheet(A1, "=C2:D3");

            InsertCells(C1D2, InsertShiftType.Down);

            AssertExpressionCorrectness(A1, "C4:D5");
        }

        [TestMethod]
        public void RemoveUpTranslateTargetPositiveOffset()
        {
            SetValueInWorksheet(A1, "=C4:D5");

            RemoveCells(C1D2, RemoveShiftType.Up);

            AssertExpressionCorrectness(A1, "C2:D3");
        }

        [TestMethod]
        public void InsertDownTranslateTargetNegativeOffset()
        {
            SetValueInWorksheet(D4, "=A1:B2");

            InsertCells(A1B2, InsertShiftType.Down);

            AssertExpressionCorrectness(D4, "A3:B4");
        }

        [TestMethod]
        public void RemoveUpTranslateTargetNegativeOffset()
        {
            SetValueInWorksheet(D4, "=A3:B4");

            RemoveCells(A1B2, RemoveShiftType.Up);

            AssertExpressionCorrectness(D4, "A1:B2");
        }

        [TestMethod]
        public void InsertDownTranslateCellsBetweenSourceAndTargetPositiveOffset()
        {
            CellRange B1B6 = new CellRange(0, 1, 5, 1);
            SetValueInWorksheet(A1, "=C2:D3");

            InsertCells(B1B6, InsertShiftType.Down);

            AssertExpressionCorrectness(A1, "C2:D3");
        }

        [TestMethod]
        public void RemoveUpTranslateCellsBetweenSourceAndTargetPositiveOffset()
        {
            CellRange B1B6 = new CellRange(0, 1, 5, 1);
            SetValueInWorksheet(A1, "=C2:D3");

            RemoveCells(B1B6, RemoveShiftType.Up);

            AssertExpressionCorrectness(A1, "C2:D3");
        }

        [TestMethod]
        public void InsertDownTranslateCellsBetweenSourceAndTargetNegativeOffset()
        {
            CellRange C1C6 = new CellRange(0, 2, 5, 2);
            SetValueInWorksheet(D4, "=A1:B2");

            InsertCells(C1C6, InsertShiftType.Down);

            AssertExpressionCorrectness(D4, "A1:B2");
        }

        [TestMethod]
        public void RemoveUpTranslateCellsBetweenSourceAndTargetNegativeOffset()
        {
            CellRange C1C6 = new CellRange(0, 2, 5, 2);
            SetValueInWorksheet(D4, "=A1:B2");

            RemoveCells(C1C6, RemoveShiftType.Up);

            AssertExpressionCorrectness(D4, "A1:B2");
        }

        [TestMethod]
        public void InsertDownOneCellTranslateSourcePositiveOffset()
        {
            SetValueInWorksheet(A1, "=C2:D3");

            InsertCells(A1, InsertShiftType.Down);

            AssertExpressionCorrectness(A2, "C2:D3");
        }

        [TestMethod]
        public void RemoveUpOneCellTranslateSourcePositiveOffset()
        {
            SetValueInWorksheet(A2, "=C2:D3");

            RemoveCells(A1, RemoveShiftType.Up);

            AssertExpressionCorrectness(A1, "C2:D3");
        }

        [TestMethod]
        public void InsertDownOneCellTranslateSourceNegativeOffset()
        {
            CellIndex D3 = new CellIndex(2, 3);
            CellIndex D5 = new CellIndex(4, 3);
            SetValueInWorksheet(D4, "=A1:B2");

            InsertCells(D3, InsertShiftType.Down);

            AssertExpressionCorrectness(D5, "A1:B2");
        }

        [TestMethod]
        public void RemoveUpOneCellTranslateSourceNegativeOffset()
        {
            CellIndex D3 = new CellIndex(2, 3);
            CellIndex D5 = new CellIndex(4, 3);
            SetValueInWorksheet(D4, "=A1:B2");

            RemoveCells(D3, RemoveShiftType.Up);

            AssertExpressionCorrectness(D3, "A1:B2");
        }

        [TestMethod]
        public void InsertDownThreeCellsTranslateSourcePositiveOffset()
        {
            CellRange A1A3 = new CellRange(0, 0, 2, 0);
            SetValueInWorksheet(A1, "=C2:D3");

            InsertCells(A1A3, InsertShiftType.Down);

            AssertExpressionCorrectness(A4, "C2:D3");
        }

        [TestMethod]
        public void RemoveUpThreeCellsTranslateSourcePositiveOffset()
        {
            CellRange A1A3 = new CellRange(0, 0, 2, 0);
            SetValueInWorksheet(A4, "=C2:D3");

            RemoveCells(A1A3, RemoveShiftType.Up);

            AssertExpressionCorrectness(A1, "C2:D3");
        }

        [TestMethod]
        public void InsertDownThreeCellsTranslateSourceNegativeOffset()
        {
            CellIndex D7 = new CellIndex(6, 3);
            CellRange D3D5 = new CellRange(2, 3, 4, 3);
            SetValueInWorksheet(D4, "=A1:B2");

            InsertCells(D3D5, InsertShiftType.Down);

            AssertExpressionCorrectness(D7, "A1:B2");
        }

        [TestMethod]
        public void RemoveUpThreeCellsTranslateSourceNegativeOffset()
        {
            CellIndex D7 = new CellIndex(6, 3);
            CellRange D3D5 = new CellRange(2, 3, 4, 3);
            SetValueInWorksheet(D7, "=A1:B2");

            RemoveCells(D3D5, RemoveShiftType.Up);

            AssertExpressionCorrectness(D4, "A1:B2");
        }

        [TestMethod]
        public void InsertDownTranslateHalfTargetPositiveOffset()
        {
            CellIndex C1 = new CellIndex(0, 2);
            SetValueInWorksheet(A1, "=C2:D3");

            InsertCells(C1, InsertShiftType.Down);

            AssertExpressionCorrectness(A1, "C2:D3");
        }

        [TestMethod]
        public void RemoveUpTranslateHalfTargetPositiveOffset()
        {
            CellIndex C1 = new CellIndex(0, 2);
            SetValueInWorksheet(A1, "=C2:D3");

            RemoveCells(C1, RemoveShiftType.Up);

            AssertExpressionCorrectness(A1, "C2:D3");
        }

        [TestMethod]
        public void InsertDownTranslateHalfTargetNegativeOffset()
        {
            CellIndex B1 = new CellIndex(0, 1);
            SetValueInWorksheet(D4, "=A1:B2");

            InsertCells(B1, InsertShiftType.Down);

            AssertExpressionCorrectness(D4, "A1:B2");
        }

        [TestMethod]
        public void RemoveUpTranslateHalfTargetNegativeOffset()
        {
            CellIndex B1 = new CellIndex(0, 1);
            SetValueInWorksheet(D4, "=A1:B2");

            RemoveCells(B1, RemoveShiftType.Up);

            AssertExpressionCorrectness(D4, "A1:B2");
        }

        [TestMethod]
        public void InsertDownTranslateAfterSourcePositiveOffset()
        {
            SetValueInWorksheet(A1, "=C2:D3");

            InsertCells(A2, InsertShiftType.Down);

            AssertExpressionCorrectness(A1, "C2:D3");
        }

        [TestMethod]
        public void RemoveUpTranslateAfterSourcePositiveOffset()
        {
            SetValueInWorksheet(A1, "=C2:D3");

            RemoveCells(A2, RemoveShiftType.Up);

            AssertExpressionCorrectness(A1, "C2:D3");
        }

        [TestMethod]
        public void InsertDownTranslateAfterTargetPositiveOffset()
        {
            CellRange C4D5 = new CellRange(3, 2, 4, 3);
            SetValueInWorksheet(A1, "=C2:D3");

            InsertCells(C4D5, InsertShiftType.Down);

            AssertExpressionCorrectness(A1, "C2:D3");
        }

        [TestMethod]
        public void RemoveUpTranslateAfterTargetPositiveOffset()
        {
            CellRange C4D5 = new CellRange(3, 2, 4, 3);
            SetValueInWorksheet(A1, "=C2:D3");

            RemoveCells(C4D5, RemoveShiftType.Up);

            AssertExpressionCorrectness(A1, "C2:D3");

        }

        [TestMethod]
        public void InsertDownTranslateAfterSourceAndTargetPositiveOffset()
        {
            CellRange A4E5 = new CellRange(3, 0, 4, 4);
            SetValueInWorksheet(A1, "=C2:D3");

            InsertCells(A4E5, InsertShiftType.Down);

            AssertExpressionCorrectness(A1, "C2:D3");
        }

        [TestMethod]
        public void RemoveUpTranslateAfterSourceAndTargetPositiveOffset()
        {
            CellRange A4E5 = new CellRange(3, 0, 4, 4);
            SetValueInWorksheet(A1, "=C2:D3");

            RemoveCells(A4E5, RemoveShiftType.Up);

            AssertExpressionCorrectness(A1, "C2:D3");
        }

        [TestMethod]
        public void InsertDownInTargetAndAnfer()
        {
            SetValueInWorksheet(A1, "=B2:C3");

            CellRange B3C4 = new CellRange(2, 1, 3, 2);
            InsertCells(B3C4, InsertShiftType.Down);

            AssertExpressionCorrectness(A1, "B2:C5");
        }

        [TestMethod]
        public void InsertDownBeforeTargetAndAfter()
        {
            SetValueInWorksheet(A1, "=B2:C3");

            InsertCells(B1C4, InsertShiftType.Down);

            AssertExpressionCorrectness(A1, "B6:C7");
        }

        [TestMethod]
        public void InsertDownBeforeAndInTarget()
        {
            SetValueInWorksheet(A1, "=B2:C3");

            CellRange B1C3 = new CellRange(0, 1, 2, 2);
            InsertCells(B1C3, InsertShiftType.Down);

            AssertExpressionCorrectness(A1, "B5:C6");
        }

        [TestMethod]
        public void InsertDownBeforeAndInTargetAndSource()
        {
            CellIndex A5 = new CellIndex(4, 0);
            CellRange A1C2 = new CellRange(0, 0, 1, 2);
            SetValueInWorksheet(A3, "=B2:C3");

            InsertCells(A1C2, InsertShiftType.Down);

            AssertExpressionCorrectness(A5, "B4:C5");
        }

        [TestMethod]
        public void RemoveUpInTarget()
        {
            SetValueInWorksheet(A3, "=B1:C3");

            CellRange B2C2 = new CellRange(1, 1, 1, 2);
            RemoveCells(B2C2, RemoveShiftType.Up);

            AssertExpressionCorrectness(A3, "B1:C2");
        }

        [TestMethod]
        public void RemoveUpOneInAndOneAfterTarget()
        {
            CellRange B3D4 = new CellRange(2, 1, 3, 3);
            SetValueInWorksheet(A3, "=B1:C3");

            RemoveCells(B3D4, RemoveShiftType.Up);

            AssertExpressionCorrectness(A3, "B1:C2");
        }

        [TestMethod]
        public void RemoveUpTwoInAndOneAfterTarget()
        {
            CellRange B2C4 = new CellRange(1, 1, 3, 2);
            SetValueInWorksheet(A3, "=B1:C3");

            RemoveCells(B2C4, RemoveShiftType.Up);

            AssertExpressionCorrectness(A3, "B1:C1");
        }

        [TestMethod]
        public void RemoveUpOneInAndTwoAfterTarget()
        {
            CellRange B3C5 = new CellRange(2, 1, 3, 2);
            SetValueInWorksheet(A3, "=B1:C3");

            RemoveCells(B3C5, RemoveShiftType.Up);

            AssertExpressionCorrectness(A3, "B1:C2");
        }

        [TestMethod]
        public void RemoveUpWholeTarget()
        {
            SetValueInWorksheet(A3, "=B1:C3");

            RemoveCells(B1C4, RemoveShiftType.Up);

            string value = ((FormulaCellValue)this.worksheet.Cells[A3].GetValue().Value).GetResultValueAsString(CellValueFormat.GeneralFormat);
            Assert.AreEqual(SpreadsheetCultureHelper.ReferenceErrorValue, value);
        }
    }
}
