﻿using System;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal
{
    internal static class CellReferenceTestHelper
    {
        public static void AssertTranslation(RadExpression translatedExpression, string expectedToString, int expectedFromIndexRow, int expectedFromIndexColumn, int expectedToIndexRow, int expectedToIndexColumn)
        {
            AssertToString(translatedExpression, expectedToString);
            Assert.AreEqual(((CellReferenceRangeExpression)translatedExpression).CellReferenceRange.FromCellReference.ActualRowIndex, expectedFromIndexRow);
            Assert.AreEqual(((CellReferenceRangeExpression)translatedExpression).CellReferenceRange.FromCellReference.ActualColumnIndex, expectedFromIndexColumn);
            Assert.AreEqual(((CellReferenceRangeExpression)translatedExpression).CellReferenceRange.ToCellReference.ActualRowIndex, expectedToIndexRow);
            Assert.AreEqual(((CellReferenceRangeExpression)translatedExpression).CellReferenceRange.ToCellReference.ActualColumnIndex, expectedToIndexColumn);
        }

        public static void AssertToString(RadExpression translatedExpression, string expectedToString)
        {
            Assert.IsTrue(expectedToString.Equals(translatedExpression.ToString(), StringComparison.CurrentCultureIgnoreCase));
        }

        public static void AssertNumberExpressionValue(RadExpression expression, double expectedValue)
        {
            NumberExpression resultExpression = expression.GetValueAsConstantExpression() as NumberExpression;
            Assert.IsNotNull(resultExpression);
            Assert.AreEqual(expectedValue, resultExpression.Value);
        }

        public static void CreateRangeAndAssertResult(CellAreaReference expectedRange, CellAreaReference range1, CellAreaReference range2, CellAreaReference range3 = null)
        {
            CellAreaReference mergedRange = CellAreaReference.CreateRange(range1, range2);

            if (range3 != null)
            {
                mergedRange = CellAreaReference.CreateRange(mergedRange, range3);
            }

            AssertRangeCorrectness(expectedRange, mergedRange);
        }

        public static void CreateIntersectionAndAssertResult(CellAreaReference expectedRange, CellAreaReference range1, CellAreaReference range2)
        {
            CellAreaReference intersectionRange = CellAreaReference.CreateIntersection(range1, range2);

            AssertRangeCorrectness(expectedRange, intersectionRange);
        }

        public static void CreateIntersectionAndAssertError(CellAreaReference range1, CellAreaReference range2)
        {
            CellAreaReference intersectionRange = CellAreaReference.CreateIntersection(range1, range2);

            Assert.IsNull(intersectionRange);
        }

        public static void AssertRangeCorrectness(CellAreaReference expectedRange, CellAreaReference actualRange)
        {
            Assert.AreEqual(actualRange.FromCellReference.ActualRowIndex, expectedRange.FromCellReference.ActualRowIndex);
            Assert.AreEqual(actualRange.FromCellReference.ActualColumnIndex, expectedRange.FromCellReference.ActualColumnIndex);
            Assert.AreEqual(actualRange.ToCellReference.ActualRowIndex, expectedRange.ToCellReference.ActualRowIndex);
            Assert.AreEqual(actualRange.ToCellReference.ActualColumnIndex, expectedRange.ToCellReference.ActualColumnIndex);
        }

        public static void AssertTranslatedFormula(Worksheet worksheet, FormulaCellValue cellValue, CellIndex cellIndex, double expectedResult)
        {
            worksheet.Cells[cellIndex].SetValue(cellValue.CloneAndTranslate(worksheet, cellIndex));
            FormulaCellValue translatedCellValue = worksheet.Cells[cellIndex].GetValue().Value as FormulaCellValue;

            Assert.IsNotNull(translatedCellValue);
            Assert.AreEqual(expectedResult, translatedCellValue.Value.GetValueAsConstantExpression().NumberValue());
        }

        public static void AssertTranslatedExpressionParenthesisCorrectness(string originalExpressionAsString, string expectedExpressionAsString, CellIndex fromCellIndex, CellIndex toCellIndex)
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            FormulaCellValue originalCellValue = TestHelper.CreateFormulaCellValue(originalExpressionAsString, worksheet, fromCellIndex);
            FormulaCellValue translatedCellValue = originalCellValue.CloneAndTranslate(worksheet, toCellIndex);

            Assert.AreEqual(expectedExpressionAsString, translatedCellValue.RawValue);
        }

        public static void AssertCellAreaReferenceValidityAndSize(CellAreaReference cellAreaReference, int columnsCount, int rowsCount)
        {
            Assert.IsTrue(cellAreaReference.IsValid);
            Assert.AreEqual(columnsCount, cellAreaReference.ColumnsCount);
            Assert.AreEqual(rowsCount, cellAreaReference.RowsCount);
        }

        public static void AssertCellReferenceCorrectness(CellReference cellReference, bool isColumnAbsolute, bool isRowAbsolute, int columnIndex, int rowIndex)
        {
            Assert.AreEqual(isColumnAbsolute, cellReference.ColumnReference.IsAbsolute);
            Assert.AreEqual(isRowAbsolute, cellReference.RowReference.IsAbsolute);
            Assert.AreEqual(columnIndex, cellReference.ActualColumnIndex);
            Assert.AreEqual(rowIndex, cellReference.ActualRowIndex);
        }

        public static RadExpression CloneAndTranslate(this RadExpression expression, Worksheet worksheet, CellIndex cellIndex)
        {
            ExpressionCloneAndTranslateContext context = new ExpressionCloneAndTranslateContext(worksheet, cellIndex.RowIndex, cellIndex.ColumnIndex, false);
            return expression.CloneAndTranslate(context);
        }
    }
}
