﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.DataValidation;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal.DataValidation
{
    [TestClass]
    public class WholeNumberDataValidationTests
    {
        #region Between

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationBetweenTwoNumbersValueOnTheLowerBound()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);
            worksheet.Cells[1, 0].SetValue(5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.Between;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "1");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationBetweenTwoNumbersValueOnTheUpperBound()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);
            worksheet.Cells[1, 0].SetValue(5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.Between;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "5");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationBetweenTwoNumbersValueInsideTheRange()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);
            worksheet.Cells[1, 0].SetValue(5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.Between;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "3");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationIgnoreBlank()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);
            worksheet.Cells[1, 0].SetValue(5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.IgnoreBlank = true;
            context.ComparisonOperator = ComparisonOperator.Between;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, string.Empty);

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationDoNotIgnoreBlank()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);
            worksheet.Cells[1, 0].SetValue(5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.IgnoreBlank = false;
            context.ComparisonOperator = ComparisonOperator.Between;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, string.Empty);

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationBetweenTwoNumbersValueBelowTheLowerBound()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);
            worksheet.Cells[1, 0].SetValue(5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.Between;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "0");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationBetweenTwoNumbersValueAboveTheUpperBound()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);
            worksheet.Cells[1, 0].SetValue(5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.Between;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "7");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationBetweenTwoNumbersValueOnTheLowerBoundDecimal()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);
            worksheet.Cells[1, 0].SetValue(5.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.Between;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "2");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationBetweenTwoNumbersValueOnTheUpperBoundDecimal()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);
            worksheet.Cells[1, 0].SetValue(5.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.Between;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "5");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationBetweenTwoNumbersValueInsideTheRangeDecimal()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);
            worksheet.Cells[1, 0].SetValue(5.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.Between;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "3");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationBetweenTwoNumbersValueBelowTheLowerBoundDecimal()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);
            worksheet.Cells[1, 0].SetValue(5.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.Between;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "0");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationBetweenTwoNumbersValueAboveTheUpperBoundDecimal()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);
            worksheet.Cells[1, 0].SetValue(5.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.Between;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "7");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationBetweenTwoNumbersDecimalValueOnTheLowerBoundDecimal()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);
            worksheet.Cells[1, 0].SetValue(5.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.Between;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "2.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationBetweenTwoNumbersDecimalValueOnTheUpperBoundDecimal()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);
            worksheet.Cells[1, 0].SetValue(5.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.Between;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "5.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationBetweenTwoNumbersDecimalValueInsideTheRangeDecimal()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);
            worksheet.Cells[1, 0].SetValue(5.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.Between;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "3.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationBetweenTwoNumbersDecimalValueBelowTheLowerBoundDecimal()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);
            worksheet.Cells[1, 0].SetValue(5.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.Between;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "0.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationBetweenTwoNumbersDecimalValueAboveTheUpperBoundDecimal()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);
            worksheet.Cells[1, 0].SetValue(5.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.Between;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "7.5");

            Assert.IsFalse(result);
        }

        #endregion


        #region Not Between

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotBetweenTwoNumbersValueOnTheLowerBound()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);
            worksheet.Cells[1, 0].SetValue(5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.NotBetween;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "1");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotBetweenTwoNumbersValueOnTheUpperBound()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);
            worksheet.Cells[1, 0].SetValue(5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.NotBetween;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotBetweenTwoNumbersValueInsideTheRange()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);
            worksheet.Cells[1, 0].SetValue(5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.NotBetween;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "3");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotBetweenTwoNumbersValueBelowTheLowerBound()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);
            worksheet.Cells[1, 0].SetValue(5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.NotBetween;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "0");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotBetweenTwoNumbersValueAboveTheUpperBound()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);
            worksheet.Cells[1, 0].SetValue(5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.NotBetween;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "7");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotBetweenTwoNumbersValueOnTheLowerBoundDecimal()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);
            worksheet.Cells[1, 0].SetValue(5.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.NotBetween;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "2");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotBetweenTwoNumbersValueOnTheUpperBoundDecimal()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);
            worksheet.Cells[1, 0].SetValue(5.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.NotBetween;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotBetweenTwoNumbersValueInsideTheRangeDecimal()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);
            worksheet.Cells[1, 0].SetValue(5.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.NotBetween;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "3");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotBetweenTwoNumbersValueBelowTheLowerBoundDecimal()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);
            worksheet.Cells[1, 0].SetValue(5.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.NotBetween;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "0");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotBetweenTwoNumbersValueAboveTheUpperBoundDecimal()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);
            worksheet.Cells[1, 0].SetValue(5.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.NotBetween;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "7");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotBetweenTwoNumbersDecimalValueOnTheLowerBound()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);
            worksheet.Cells[1, 0].SetValue(5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.NotBetween;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "1.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotBetweenTwoNumbersDecimalValueOnTheUpperBound()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);
            worksheet.Cells[1, 0].SetValue(5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.NotBetween;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "5.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotBetweenTwoNumbersDecimalValueInsideTheRange()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);
            worksheet.Cells[1, 0].SetValue(5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.NotBetween;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "3.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotBetweenTwoNumbersDecimalValueBelowTheLowerBound()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);
            worksheet.Cells[1, 0].SetValue(5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.NotBetween;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "0.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotBetweenTwoNumbersDecimalValueAboveTheUpperBound()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);
            worksheet.Cells[1, 0].SetValue(5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.Argument2 = "=A2";
            context.ComparisonOperator = ComparisonOperator.NotBetween;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "7.5");

            Assert.IsFalse(result);
        }

        #endregion


        #region Equal To

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationEqualToTheSameValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.EqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "1");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationEqualToTheDifferentValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.EqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "2");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationEqualToTheSameDecimalValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.EqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "1.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationEqualToTheDifferentDecimalValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.EqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "2.5");

            Assert.IsFalse(result);
        }

        #endregion


        #region Not Equal To

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotEqualToTheSameValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.NotEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "1");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotEqualToTheDifferentValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.NotEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "2");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotEqualToTheSameDecimalValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.NotEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, 2, 0, "1.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationNotEqualToTheDifferentDecimalValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.NotEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "2.5");

            Assert.IsFalse(result);
        }

        #endregion


        #region Greater Than

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationGreaterThanTheSameValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.GreaterThan;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "1");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationGreaterThanSmallerValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.GreaterThan;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "0");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationGreaterThanLargerValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.GreaterThan;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "5");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationGreaterThanTheSameDecimalValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.GreaterThan;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "1.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationGreaterThanSmallerDecimalValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.GreaterThan;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "0.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationGreaterThanLargerDecimalValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.GreaterThan;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "5.5");

            Assert.IsFalse(result);
        }

        #endregion


        #region Less Than

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationLessThanTheSameValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThan;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "1");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationLessThanSmallerValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThan;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "0");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationLessThanLargerValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThan;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationLessThanTheSameDecimalValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThan;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "1.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationLessThanSmallerDecimalValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThan;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "0.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationLessThanLargerDeciamlValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThan;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "5.5");

            Assert.IsFalse(result);
        }

        #endregion


        #region Greater or Equal

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationGreaterOrEqualTheSameValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.GreaterThanOrEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "1");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationGreaterOrEqualSmallerValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.GreaterThanOrEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "0");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationGreaterOrEqualLargerValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.GreaterThanOrEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "5");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationGreaterOrEqualTheSameDecimalValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.GreaterThanOrEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "1.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationGreaterOrEqualSmallerDecimalValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.GreaterThanOrEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "0.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationGreaterOrEqualLargerDecimalValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.GreaterThanOrEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "5.5");

            Assert.IsFalse(result);
        }

        #endregion


        #region Less or Equal

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationLessOrEqualTheSameValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThanOrEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "1");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationLessOrEqualSmallerValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThanOrEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "0");

            Assert.IsTrue(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationLessOrEqualLargerValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThanOrEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationLessOrEqualTheSameDecimalValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1.5);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThanOrEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "1.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationLessOrEqualSmallerDecimalValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThanOrEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "0.5");

            Assert.IsFalse(result);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestWholeNumberDataValidationLessOrEqualLargerDecimalValue()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThanOrEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            bool result = rule.Evaluate(worksheet, cellIndex, "5.5");

            Assert.IsFalse(result);
        }

        #endregion


        #region Insert Remove Cells Integration

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestDataValidationRuleContainingRelativeCellReferenceInsertCells()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThanOrEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            worksheet.Cells[2, 0].SetDataValidationRule(rule);

            worksheet.Cells[1, 0, 1, SpreadsheetDefaultValues.ColumnCount - 1].Insert(InsertShiftType.Down);

            rule = worksheet.Cells[2, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNull(rule);

            rule = worksheet.Cells[3, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNotNull(rule);

            string argument1 = rule.GetExpressionString(rule.Argument1, worksheet, 3, 0);
            Assert.AreEqual("=A1", argument1);

            workbook.History.Undo();

            rule = worksheet.Cells[3, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNull(rule);

            rule = worksheet.Cells[2, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNotNull(rule);

            argument1 = rule.GetExpressionString(rule.Argument1, worksheet, 2, 0);
            Assert.AreEqual("=A1", argument1);

            workbook.History.Redo();

            rule = worksheet.Cells[2, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNull(rule);

            rule = worksheet.Cells[3, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNotNull(rule);

            argument1 = rule.GetExpressionString(rule.Argument1, worksheet, 3, 0);
            Assert.AreEqual("=A1", argument1);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestDataValidationRuleContainingRelativeCellReferenceInsertRows()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThanOrEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            worksheet.Cells[2, 0].SetDataValidationRule(rule);

            worksheet.Rows[1].Insert();

            rule = worksheet.Cells[2, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNull(rule);

            rule = worksheet.Cells[3, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNotNull(rule);

            string argument1 = rule.GetExpressionString(rule.Argument1, worksheet, 3, 0);
            Assert.AreEqual("=A1", argument1);

            workbook.History.Undo();

            rule = worksheet.Cells[3, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNull(rule);

            rule = worksheet.Cells[2, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNotNull(rule);

            argument1 = rule.GetExpressionString(rule.Argument1, worksheet, 2, 0);
            Assert.AreEqual("=A1", argument1);

            workbook.History.Redo();

            rule = worksheet.Cells[2, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNull(rule);

            rule = worksheet.Cells[3, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNotNull(rule);

            argument1 = rule.GetExpressionString(rule.Argument1, worksheet, 3, 0);
            Assert.AreEqual("=A1", argument1);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestDataValidationRuleContainingRelativeCellReferenceDeleteCells()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThanOrEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            worksheet.Cells[2, 0].SetDataValidationRule(rule);

            worksheet.Cells[1, 0, 1, SpreadsheetDefaultValues.ColumnCount - 1].Remove(RemoveShiftType.Up);

            rule = worksheet.Cells[2, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNull(rule);

            rule = worksheet.Cells[1, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNotNull(rule);

            string argument1 = rule.GetExpressionString(rule.Argument1, worksheet, 1, 0);
            Assert.AreEqual("=A1", argument1);

            workbook.History.Undo();

            rule = worksheet.Cells[1, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNull(rule);

            rule = worksheet.Cells[2, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNotNull(rule);

            argument1 = rule.GetExpressionString(rule.Argument1, worksheet, 2, 0);
            Assert.AreEqual("=A1", argument1);

            workbook.History.Redo();

            rule = worksheet.Cells[2, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNull(rule);

            rule = worksheet.Cells[1, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNotNull(rule);

            argument1 = rule.GetExpressionString(rule.Argument1, worksheet, 1, 0);
            Assert.AreEqual("=A1", argument1);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestDataValidationRuleContainingRelativeCellReferenceDeleteRows()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThanOrEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            worksheet.Cells[2, 0].SetDataValidationRule(rule);

            worksheet.Rows[1].Remove();

            rule = worksheet.Cells[2, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNull(rule);

            rule = worksheet.Cells[1, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNotNull(rule);

            string argument1 = rule.GetExpressionString(rule.Argument1, worksheet, 1, 0);
            Assert.AreEqual("=A1", argument1);

            workbook.History.Undo();

            rule = worksheet.Cells[1, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNull(rule);

            rule = worksheet.Cells[2, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNotNull(rule);

            argument1 = rule.GetExpressionString(rule.Argument1, worksheet, 2, 0);
            Assert.AreEqual("=A1", argument1);

            workbook.History.Redo();

            rule = worksheet.Cells[2, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNull(rule);

            rule = worksheet.Cells[1, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNotNull(rule);

            argument1 = rule.GetExpressionString(rule.Argument1, worksheet, 1, 0);
            Assert.AreEqual("=A1", argument1);
        }

        #endregion


        #region Copy Paste Clear

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestDataValidationRuleContainingRelativeCellReferenceCopyPaste()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThanOrEqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            worksheet.Cells[2, 0].SetDataValidationRule(rule);

            WorksheetFragment fragment = worksheet.Cells[2, 0].Copy();
            worksheet.Cells[3, 0].Paste(fragment, PasteOptions.All);

            rule = worksheet.Cells[3, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            string argument1 = rule.GetExpressionString(rule.Argument1, worksheet, 3, 0);
            Assert.AreEqual("=A2", argument1);

            workbook.History.Undo();

            rule = worksheet.Cells[3, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNull(rule);

            workbook.History.Redo();

            rule = worksheet.Cells[3, 0].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            argument1 = rule.GetExpressionString(rule.Argument1, worksheet, 3, 0);
            Assert.AreEqual("=A2", argument1);
        }

        [TestMethod]
        [TestCategory("DataValidation")]
        public void TestDataValidationRuleContainingRelativeCellReferenceClear()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets[0];
            CellIndex cellIndex = new CellIndex(2, 0);

            worksheet.Cells[0, 0].SetValue(1);

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, cellIndex);
            context.Argument1 = "=A1";
            context.ComparisonOperator = ComparisonOperator.LessThanOrEqualsTo;

            CellRange cellRange = new CellRange(2, 0, 3, 0);

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            worksheet.Cells[cellRange].SetDataValidationRule(rule);

            worksheet.Cells[cellRange].Clear(ClearType.All);

            rule = worksheet.Cells[cellRange].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNull(rule);

            workbook.History.Undo();

            rule = worksheet.Cells[cellRange].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNotNull(rule);

            workbook.History.Redo();

            rule = worksheet.Cells[cellRange].GetDataValidationRule().Value as WholeNumberDataValidationRule;
            Assert.IsNull(rule);
        }

        #endregion
    }
}