﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Expressions;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal
{
    [TestClass]
    public class DefinedNameReferringTests : MultiCulturalTestClassBase
    {
        #region Fields

        private Workbook workbook;
        private Worksheet sheet1;
        private Worksheet sheet2;

        #endregion


        #region Properties

        public Workbook Workbook 
        { 
            get
            {
                return this.workbook; 
            }
        }

        public Worksheet Sheet1
        {
            get
            {
                return this.sheet1;
            }
        }

        public Worksheet Sheet2
        {
            get
            {
                return this.sheet2;
            }
        }

        #endregion


        #region Test Initialize & Clean up

        [TestInitialize]
        public void TestInitialize()
        {
            workbook = new Workbook();
            sheet1 = Workbook.Worksheets.Add();
            sheet2 = Workbook.Worksheets.Add();
        }

        [TestCleanup]
        public void TestCleanup()
        {
            sheet1 = null;
            sheet2 = null;
            workbook = null;
        }

        #endregion


        #region Helper Methods

        public ISpreadsheetName AddDefinedName(Worksheet sheet, string name, string value, CellIndex cellIndex)
        {
            return sheet.Names.Add(name, value, cellIndex);
        }
        
        public ISpreadsheetName AddDefinedName(Workbook workbook, string name, string value, CellIndex cellIndex)
        {
            return workbook.Names.Add(name, value, cellIndex);
        }

        public void RemoveDefinedName(Worksheet sheet, string name)
        {
            sheet.Names.Remove(name);
        }

        public void RemoveDefinedName(Workbook workbook, string name)
        {
            workbook.Names.Remove(name);
        }

        public void SetValue(Worksheet sheet, CellIndex cellIdex, string value)
        {
            sheet.Cells[cellIdex].SetValue(value);
        }

        public void AssertResult(Worksheet sheet, CellIndex cellIndex, string expectedResultAsString)
        {
            string resultValue = sheet.Cells[cellIndex].GetValue().Value.GetResultValueAsString(CellValueFormat.GeneralFormat);
            Assert.AreEqual(expectedResultAsString, resultValue);
        }

        public void UseDefinedNameAndAssertResult(Worksheet sheet, CellIndex cellIndex, string value, string expectedResultAsString)
        {
            this.SetValue(sheet, cellIndex, value);
            AssertResult(sheet, cellIndex, expectedResultAsString);
        }

        #endregion


        #region Referring global and local defined names tests

        [TestMethod]
        public void AddDefinedNameToWorkbookReferFromSheet1()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            AddDefinedName(Workbook, "name", "=5", cellIndex);
            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=name", "5");
        }

        [TestMethod]
        public void AddDefinedNameToSheet1ReferFromSheet1()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            AddDefinedName(Sheet1, "name", "=5", cellIndex);
            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=name", "5");
        }

        [TestMethod]
        public void AddDefinedNameToSheet1ReferFromSheet2()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            AddDefinedName(Sheet1, "name", "=5", cellIndex);
            UseDefinedNameAndAssertResult(Sheet2, cellIndex, "=name", ErrorExpressions.NameError.Value);
        }

        [TestMethod]
        public void AddDefinedNameToSheet1ReferFromSheet2WithQualifier()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            AddDefinedName(Sheet1, "name", "=5", cellIndex);
            UseDefinedNameAndAssertResult(Sheet2, cellIndex, "=Sheet1!name", "5");
        }

        [TestMethod]
        public void AddDefinedNameToWorkbookAndSheet1ReferFromSheet1()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            AddDefinedName(Workbook, "name", @"=""workbook""", cellIndex);
            AddDefinedName(Sheet1, "name", @"=""sheet1""", cellIndex);
            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=name", "sheet1");
        }

        [TestMethod]
        public void AddDefinedNameToWorkbookAndSheet1ReferFromSheet1WithSheetQualifier()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            AddDefinedName(Workbook, "name", @"=""workbook""", cellIndex);
            AddDefinedName(Sheet1, "name", @"=""sheet1""", cellIndex);
            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=Sheet1!name", "sheet1");
        }

        [TestMethod]
        public void AddDefinedNameToWorkbookAndSheet1ReferFromSheet1WithNonExistingQualifier()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            AddDefinedName(Workbook, "name", @"=""workbook""", cellIndex);
            AddDefinedName(Sheet1, "name", @"=""sheet1""", cellIndex);
            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=Tralala!name", ErrorExpressions.NameError.Value);
        }

        [TestMethod]
        public void AddDefinedNameToWorkbookAndSheet1ReferFromSheet1WithWorkbookQualifier()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            AddDefinedName(Workbook, "name", @"=""workbook""", cellIndex);
            AddDefinedName(Sheet1, "name", @"=""sheet1""", cellIndex);
            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=Book1!name", "workbook");
        }

        [TestMethod]
        public void AddDefinedNameToSheet1AndSheet2ReferFromSheet2()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            AddDefinedName(Sheet1, "name", @"=""sheet1""", cellIndex);
            AddDefinedName(Sheet2, "name", @"=""sheet2""", cellIndex);
            UseDefinedNameAndAssertResult(Sheet2, cellIndex, "=name", "sheet2");
        }

        [TestMethod]
        public void AddDefinedNameToSheet1AndSheet2ReferFromSheet2WithQualifier()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            AddDefinedName(Sheet1, "name", @"=""sheet1""", cellIndex);
            AddDefinedName(Sheet2, "name", @"=""sheet2""", cellIndex);
            UseDefinedNameAndAssertResult(Sheet2, cellIndex, "=Sheet1!name", "sheet1");
        }

        [TestMethod]
        public void ReferToNonExistingDefinedNameFromWorkbook()
        {
            CellIndex cellIndex = new CellIndex(0, 0);
            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=Book1!name", ErrorExpressions.NameError.Value);
        }

        [TestMethod]
        public void ReferToNonExistingDefinedNameFromSheet()
        {
            CellIndex cellIndex = new CellIndex(0, 0);
            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=Sheet1!name", ErrorExpressions.NameError.Value);
        }

        [TestMethod]
        public void ReferToDefinedNameFromNonExistingSheet()
        {
            CellIndex cellIndex = new CellIndex(0, 0);
            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=NoSuchSheet!name", ErrorExpressions.NameError.Value);
        }

        [TestMethod]
        public void ReferToDefinedNameFromNonExistingWorkbook()
        {
            CellIndex cellIndex = new CellIndex(0, 0);
            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=NoSuchBook.xlsx!name", ErrorExpressions.NameError.Value);
        }

        #endregion


        #region Defined names content tests

        [TestMethod]
        public void AddLocalDefinedNameContainingExpression()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            AddDefinedName(Sheet1, "name", "=1+2", cellIndex);
            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=name", "3");
        }

        [TestMethod]
        public void AddLocalDefinedNameContainingFunction()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            AddDefinedName(Sheet1, "name", "=SUM(1, 2, 3)", cellIndex);
            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=name", "6");
        }

        [TestMethod]
        public void AddLocalDefinedNameContainingAbsoluteCellReference()
        {
            Sheet1.Cells[0, 0].SetValue("spreadsheet");

            AddDefinedName(Sheet1, "name", "=Sheet1!$A$1", new CellIndex(2, 2));
            UseDefinedNameAndAssertResult(Sheet1, new CellIndex(1, 1), "=name", "spreadsheet");
        }

        [TestMethod]
        public void AddLocalDefinedNameContainingAbsoluteCellReferenceToOtherSheet()
        {
            Sheet1.Cells[0, 0].SetValue("spreadsheet");

            AddDefinedName(Sheet2, "name", "=Sheet1!$A$1", new CellIndex(2, 2));
            UseDefinedNameAndAssertResult(Sheet2, new CellIndex(1, 1), "=name", "spreadsheet");
        }

        [TestMethod]
        public void AddLocalDefinedNameContainingAbsoluteCellReferenceUpdatesCorrectly()
        {
            Sheet1.Cells[0, 0].SetValue("spreadsheet");

            AddDefinedName(Sheet1, "name", "=Sheet1!$A$1", new CellIndex(2, 2));
            UseDefinedNameAndAssertResult(Sheet1, new CellIndex(1, 1), "=name", "spreadsheet");

            Sheet1.Cells[0, 0].SetValue("15");
            string newResult = Sheet1.Cells[1, 1].GetValue().Value.GetResultValueAsString(CellValueFormat.GeneralFormat);
            Assert.AreEqual("15", newResult);
        }

        [TestMethod]
        public void AddLocalDefinedNameContainingRelativeCellReference()
        {
            Sheet1.Cells[1, 1].SetValue("spreadsheet");

            AddDefinedName(Sheet1, "name", "=Sheet1!A1", new CellIndex(1, 1));
            UseDefinedNameAndAssertResult(Sheet1, new CellIndex(2, 2), "=name", "spreadsheet");
        }
        
        [TestMethod]
        public void AddLocalDefinedNameContainingRelativeCellReferenceToOtherSheet()
        {
            Sheet1.Cells[1, 1].SetValue("spreadsheet");

            AddDefinedName(Sheet2, "name", "=Sheet1!A1", new CellIndex(1, 1));
            UseDefinedNameAndAssertResult(Sheet2, new CellIndex(2, 2), "=name", "spreadsheet");
        }

        [TestMethod]
        public void AddGlobalDefinedNameContainingRelativeCellReferenceWithoutQualifier()
        {
            Workbook.ActiveWorksheet = Sheet1;
            Sheet1.Cells[1, 1].SetValue("spreadsheet");

            AddDefinedName(Workbook, "name", "=A1", new CellIndex(1, 1));
            UseDefinedNameAndAssertResult(Sheet1, new CellIndex(2, 2), "=name", "spreadsheet");
        }

        [TestMethod]
        public void AddGlobalDefinedNameContainingRelativeCellReferenceWithoutQualifier2()
        {
            Workbook.ActiveWorksheet = Sheet2;
            Sheet2.Cells[1, 1].SetValue("spreadsheet");

            AddDefinedName(Workbook, "name", "=A1", new CellIndex(1, 1));
            UseDefinedNameAndAssertResult(Sheet2, new CellIndex(2, 2), "=name", "spreadsheet");
        }

        [TestMethod]
        public void AddDefinedNameContainingDefinedName()
        {
            AddDefinedName(Sheet1, "sheet1Name", "=10", new CellIndex(1, 1));
            AddDefinedName(Sheet2, "sheet2Name", "=Sheet1!sheet1Name", new CellIndex(1, 1));
            UseDefinedNameAndAssertResult(Sheet2, new CellIndex(2, 2), "=Sheet1!sheet1Name", "10");
        }

        [TestMethod]
        public void AddWorkbookDefinedNameCellReferenceWithoutQualifier()
        {
            Workbook.ActiveWorksheet = Sheet2;
            Sheet2.Cells[0, 0].SetValue("spreadsheet");

            AddDefinedName(Workbook, "aaa", "=A1", new CellIndex(1, 1));
            UseDefinedNameAndAssertResult(Sheet2, new CellIndex(1, 1), "=aaa", "spreadsheet");

            ISpreadsheetName name;
            Workbook.Names.TryGetSpreadsheetName("aaa", out name);
            Assert.AreEqual("=Sheet2!A1", name.RefersTo);
        }

        [TestMethod]
        public void AddWorkbookDefinedNameCellReferenceWithoutQualifier2()
        {
            Workbook.ActiveWorksheet = Sheet1;

            AddDefinedName(Workbook, "bbb", "=R15", new CellIndex(1, 1));
            UseDefinedNameAndAssertResult(Sheet2, new CellIndex(1, 1), "=bbb", "0");

            ISpreadsheetName name;
            Workbook.Names.TryGetSpreadsheetName("bbb", out name);
            Assert.AreEqual("=Sheet1!R15", name.RefersTo);
        }

        [TestMethod]
        public void AddWorkbookDefinedNameCellReferenceWithoutQualifier3()
        {
            Workbook.ActiveWorksheet = Sheet2;
            Sheet2.Name = "123";

            AddDefinedName(Workbook, "bbb", "=R15", new CellIndex(1, 1));
            UseDefinedNameAndAssertResult(Sheet2, new CellIndex(1, 1), "=bbb", "0");

            ISpreadsheetName name;
            Workbook.Names.TryGetSpreadsheetName("bbb", out name);
            Assert.AreEqual("='123'!R15", name.RefersTo);
        }

        [TestMethod]
        public void AddSheet1DefinedNameCellReferenceWithoutQualifier()
        {
            Sheet1.Cells[0, 0].SetValue("spreadsheet");

            AddDefinedName(Workbook, "aaa", "=A1", new CellIndex(1, 1));
            UseDefinedNameAndAssertResult(Sheet1, new CellIndex(1, 1), "=aaa", "spreadsheet");

            ISpreadsheetName name;
            Workbook.Names.TryGetSpreadsheetName("aaa", out name);
            Assert.AreEqual("=Sheet1!A1", name.RefersTo);
        }

        #endregion


        #region Refer, Add, Delete Scenarios

        [TestMethod]
        public void ReferAddLocalNameDeleteTest()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            SetValue(Sheet1, cellIndex, "=name/5");
            AssertResult(Sheet1, cellIndex, ErrorExpressions.NameError.Value);
            
            AddDefinedName(Sheet1, "name", "=5", cellIndex);
            AssertResult(Sheet1, cellIndex, "1");

            RemoveDefinedName(Sheet1, "name");
            AssertResult(Sheet1, cellIndex, ErrorExpressions.NameError.Value);
        }

        [TestMethod]
        public void ReferAddGlobalNameDeleteTest()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            SetValue(Sheet1, cellIndex, "=name/5");
            AssertResult(Sheet1, cellIndex, ErrorExpressions.NameError.Value);

            AddDefinedName(Workbook, "name", "=5", cellIndex);
            AssertResult(Sheet1, cellIndex, "1");

            RemoveDefinedName(Workbook, "name");
            AssertResult(Sheet1, cellIndex, ErrorExpressions.NameError.Value);
        }

        [TestMethod]
        public void ReferAddLocalNameWithQualifierDeleteTest()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            SetValue(Sheet1, cellIndex, "=Sheet1!name/5");
            AssertResult(Sheet1, cellIndex, ErrorExpressions.NameError.Value);

            AddDefinedName(Sheet1, "name", "=5", cellIndex);
            AssertResult(Sheet1, cellIndex, "1");

            RemoveDefinedName(Sheet1, "name");
            AssertResult(Sheet1, cellIndex, ErrorExpressions.NameError.Value);
        }

        [TestMethod]
        public void ReferAddGlobalNameWithQualifierDeleteTest()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            SetValue(Sheet1, cellIndex, "=name/5");
            AssertResult(Sheet1, cellIndex, ErrorExpressions.NameError.Value);

            AddDefinedName(Workbook, "name", "=5", cellIndex);
            AssertResult(Sheet1, cellIndex, "1");

            RemoveDefinedName(Workbook, "name");
            AssertResult(Sheet1, cellIndex, ErrorExpressions.NameError.Value);
        }

        #endregion


        #region Cyclic Defined Names

        [TestMethod]
        public void DefinedNameReferringItselfTest()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            AddDefinedName(Workbook, "name", "=name", cellIndex);
            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=name", ErrorExpressions.NameError.Value);
        }

        [TestMethod]
        public void TwoDefinedNameReferringEachOtherTest()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            AddDefinedName(Workbook, "a", "=b", cellIndex);
            AddDefinedName(Workbook, "b", "=a", cellIndex);
            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=a", ErrorExpressions.NameError.Value);
            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=b", ErrorExpressions.NameError.Value);
        }

        [TestMethod]
        public void ThreeDefinedNameReferringEachOtherTest()
        {
            CellIndex cellIndex = new CellIndex(0, 0);

            AddDefinedName(Workbook, "a", "=b", cellIndex);
            AddDefinedName(Workbook, "b", "=d", cellIndex);
            AddDefinedName(Workbook, "d", "=a", cellIndex);

            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=a", ErrorExpressions.NameError.Value);
            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=b", ErrorExpressions.NameError.Value);
            UseDefinedNameAndAssertResult(Sheet1, cellIndex, "=d", ErrorExpressions.NameError.Value);
        }

        #endregion


        #region Defined Names with Cell References update

        [TestMethod]
        public void DefinedNameWithAbsoluteReferenceUpdateTest()
        {
            CellIndex cellIndex = new CellIndex(0, 2);
            Workbook.ActiveWorksheet = Sheet1;

            AddDefinedName(Sheet1, "s", "=Sheet1!$A$1:$B$2", cellIndex);

            SetValue(Sheet1, cellIndex, "=SUM(s)");
            AssertResult(Sheet1, cellIndex, "0");

            SetValue(Sheet1, new CellIndex(0, 0), "1");
            AssertResult(Sheet1, cellIndex, "1");

            SetValue(Sheet1, new CellIndex(0, 1), "2");
            AssertResult(Sheet1, cellIndex, "3");

            SetValue(Sheet1, new CellIndex(1, 0), "3");
            AssertResult(Sheet1, cellIndex, "6");

            SetValue(Sheet1, new CellIndex(1, 1), "4");
            AssertResult(Sheet1, cellIndex, "10");
        }

        [TestMethod]
        public void DefinedNameWithRelativeReferenceUpdateTest()
        {
            CellIndex cellIndex = new CellIndex(5, 5);
            Workbook.ActiveWorksheet = Sheet1;

            AddDefinedName(Sheet1, "s", "=Sheet1!A1:B2", new CellIndex(2, 2));

            SetValue(Sheet1, cellIndex, "=SUM(s)");
            AssertResult(Sheet1, cellIndex, "0");

            SetValue(Sheet1, new CellIndex(0, 0), "1");
            AssertResult(Sheet1, cellIndex, "0");

            SetValue(Sheet1, new CellIndex(3, 3), "1");
            AssertResult(Sheet1, cellIndex, "1");

            SetValue(Sheet1, new CellIndex(3, 4), "2");
            AssertResult(Sheet1, cellIndex, "3");

            SetValue(Sheet1, new CellIndex(4, 3), "3");
            AssertResult(Sheet1, cellIndex, "6");

            SetValue(Sheet1, new CellIndex(4, 4), "4");
            AssertResult(Sheet1, cellIndex, "10");
        }

        [TestMethod]
        public void DefinedNameToCellReferenceUpdatedConsecutively()
        {
            CellIndex cellIndex = new CellIndex(1, 1);
            Workbook.ActiveWorksheet = Sheet1;

            AddDefinedName(Sheet1, "test", "=Sheet1!$A$1", new CellIndex(0, 0));

            SetValue(Sheet1, cellIndex, "=test");
            AssertResult(Sheet1, cellIndex, "0");

            SetValue(Sheet1, new CellIndex(0, 0), "1");
            AssertResult(Sheet1, cellIndex, "1");

            SetValue(Sheet1, new CellIndex(0, 0), "2");
            AssertResult(Sheet1, cellIndex, "2");

            SetValue(Sheet1, new CellIndex(0, 0), "5");
            AssertResult(Sheet1, cellIndex, "5");
        }

        #endregion


        #region Test Add, Delete, Update Methods of the Name Collection

        [TestMethod]
        [ExpectedException(typeof(SpreadsheetNameException))]
        public void AddDefinedNameIncorrectNameTest()
        {
            AddDefinedName(Workbook, "a a", "=5", new CellIndex(0, 0));
        }

        [TestMethod]
        [ExpectedException(typeof(SpreadsheetNameException))]
        public void AddDefinedNameWithNameThatAlreadyExists()
        {
            AddDefinedName(Workbook, "TestName", "=5", new CellIndex(0, 0));
            AddDefinedName(Workbook, "tESTName", "=5", new CellIndex(0, 0));
        }

        #endregion
    }
}
