﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Core.DataStructures;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Filtering;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;
using Telerik.Windows.Documents.Spreadsheet.TestsInternal.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal.Filtering
{
    [TestClass]
    public class AutoFilterTests
    {
        #region CellRange

        [TestMethod]
        public void AddRangeTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[FilterTestHelperInternal.DefaultCellRange].SetValue(1);

            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;

            CellRange expectedActualRange = new CellRange(FilterTestHelperInternal.DefaultCellRange.FromIndex.RowIndex + 1,
                FilterTestHelperInternal.DefaultCellRange.FromIndex.ColumnIndex,
                FilterTestHelperInternal.DefaultCellRange.ToIndex.RowIndex,
                FilterTestHelperInternal.DefaultCellRange.ToIndex.ColumnIndex);
            Assert.AreEqual(expectedActualRange, worksheet.Filter.ActualFilteredRange);
            FilterTestHelperInternal.AssertAllRowsAreUnhidden(worksheet, FilterTestHelperInternal.DefaultCellRange);
        }

        [TestMethod]
        public void AddRangeTestNoUsedCellRange()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            
            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;

            CellRange expectedActualRange = null;
            Assert.AreEqual(expectedActualRange, worksheet.Filter.ActualFilteredRange);
            FilterTestHelperInternal.AssertAllRowsAreUnhidden(worksheet, FilterTestHelperInternal.DefaultCellRange);
        }

        [TestMethod]
        public void AddRangeTestPartialUsedCellRange()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            CellRange usedRange = new CellRange(0, 0,
                FilterTestHelperInternal.DefaultCellRange.ToIndex.RowIndex - 2,
                FilterTestHelperInternal.DefaultCellRange.ToIndex.ColumnIndex - 2);
            worksheet.Cells[usedRange].SetValue(1);

            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;

            CellRange expectedActualRange = new CellRange(FilterTestHelperInternal.DefaultCellRange.FromIndex.RowIndex + 1,
                FilterTestHelperInternal.DefaultCellRange.FromIndex.ColumnIndex,
                FilterTestHelperInternal.DefaultCellRange.ToIndex.RowIndex - 2,
                FilterTestHelperInternal.DefaultCellRange.ToIndex.ColumnIndex);
            Assert.AreEqual(expectedActualRange, worksheet.Filter.ActualFilteredRange);
            FilterTestHelperInternal.AssertAllRowsAreUnhidden(worksheet, FilterTestHelperInternal.DefaultCellRange);
        }

        [TestMethod]
        public void AddRangeTestUndoRedo()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Cells[FilterTestHelperInternal.DefaultCellRange].SetValue(1);

            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            worksheet.Workbook.History.Undo();

            Assert.AreEqual(null, worksheet.Filter.ActualFilteredRange);
            Assert.AreEqual(null, worksheet.Filter.FilterRange);

            worksheet.Workbook.History.Redo();

            CellRange expectedActualRange = new CellRange(FilterTestHelperInternal.DefaultCellRange.FromIndex.RowIndex + 1,
                FilterTestHelperInternal.DefaultCellRange.FromIndex.ColumnIndex,
                FilterTestHelperInternal.DefaultCellRange.ToIndex.RowIndex,
                FilterTestHelperInternal.DefaultCellRange.ToIndex.ColumnIndex);
            Assert.AreEqual(expectedActualRange, worksheet.Filter.ActualFilteredRange);
            Assert.AreEqual(FilterTestHelperInternal.DefaultCellRange, worksheet.Filter.FilterRange);
            FilterTestHelperInternal.AssertAllRowsAreUnhidden(worksheet, FilterTestHelperInternal.DefaultCellRange);
        }

        [TestMethod]
        public void AddRangeFilterChangeValueSetNewRangeAndUndo()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FilterTestHelperInternal.SetValuesToWorksheetColumn(worksheet, 6, 5, new List<object>()
            {
                "1", "2", "3", "test"
            });

            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            IFilter filter = new ValuesCollectionFilter(0, FilterTestHelperInternal.StringValues, null, true);
            worksheet.Filter.SetFilter(filter);
            worksheet.Cells[6, 5, 6, 5].SetValue(3);

            worksheet.Filter.FilterRange = null;
            worksheet.Workbook.History.Undo();

            FilterTestHelperInternal.AssertCorrectRowsHidden(worksheet, new Dictionary<int, bool>()
            {
                {6, false},
                {7, false},
                {8, true},
                {9, false},
            });
        }

        [TestMethod]
        public void AddSingleRowRange()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Filter.FilterRange = new CellRange(0, 0, 0, 0);

            Assert.AreEqual(null, worksheet.Filter.ActualFilteredRange);
            FilterTestHelperInternal.AssertAllRowsAreUnhidden(worksheet, FilterTestHelperInternal.DefaultCellRange);
        }

        #endregion


        #region SetFilter

        [TestMethod]
        public void SetFilterTestColumnCount()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            
            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter);

            int filterColumnCount = worksheet.Filter.Filters.Count;
            Assert.AreEqual(1, filterColumnCount);
        }

        [TestMethod]
        public void SetFilterTestColumnCountUndoRedo()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            
            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter);

            worksheet.Workbook.History.Undo();

            FilterTestHelperInternal.AssertAllRowsAreUnhidden(worksheet, FilterTestHelperInternal.DefaultCellRange);
            int filterColumnCount = worksheet.Filter.Filters.Count;
            Assert.AreEqual(0, filterColumnCount);

            worksheet.Workbook.History.Redo();

            filterColumnCount = worksheet.Filter.Filters.Count;
            Assert.AreEqual(1, filterColumnCount);
        }

        [TestMethod]
        public void SetMultipleFiltersTestColumnCountUndoAndRedo()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            
            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            IFilter filter2 = new DynamicFilter(1, DynamicFilterType.December);
            worksheet.Filter.SetFilters(new IFilter[] {filter, filter2});

            worksheet.Workbook.History.Undo();

            FilterTestHelperInternal.AssertAllRowsAreUnhidden(worksheet, FilterTestHelperInternal.DefaultCellRange);
            int filterColumnCount = worksheet.Filter.Filters.Count;
            Assert.AreEqual(0, filterColumnCount);

            worksheet.Workbook.History.Redo();

            filterColumnCount = worksheet.Filter.Filters.Count;
            Assert.AreEqual(2, filterColumnCount);
        }

        [TestMethod]
        public void SetMultipleFiltersOneColumnTestColumnCount()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            
            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            IFilter filter2 = new DynamicFilter(0, DynamicFilterType.December);
            IFilter filter3 = new DynamicFilter(0, DynamicFilterType.November);
            worksheet.Filter.SetFilters(new IFilter[] {filter, filter2, filter3});

            int filterColumnCount = worksheet.Filter.Filters.Count;
            Assert.AreEqual(1, filterColumnCount);
            IFilter retrievedFilter = worksheet.Filter.GetFilter(0);
            Assert.AreEqual(filter3, retrievedFilter);
        }

        [TestMethod]
        public void SetFilterTestColumnIsCorrect()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            
            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter);

            IFilter retrievedFilter = worksheet.Filter.Filters.First();
            Assert.AreEqual(0, retrievedFilter.RelativeColumnIndex);
            Assert.AreEqual(filter, retrievedFilter);
        }

        [TestMethod]
        public void SetFilterOnTopOfAnother()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            List<object> values = new List<object>()
            {
                new DateTime(2013, 4, 2),
                new DateTime(2013, 12, 2),
            };
            FilterTestHelperInternal.SetValuesToWorksheetColumn(worksheet, 6, 5, values);

            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            IFilter filter = new DynamicFilter(0, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter);
            IFilter filter2 = new DynamicFilter(0, DynamicFilterType.December);
            worksheet.Filter.SetFilter(filter2);

            Dictionary<int, bool> expectedRows = new Dictionary<int, bool>()
            {
                {6, true},
                {7, false}
            };

            FilterTestHelperInternal.AssertCorrectRowsHidden(worksheet, expectedRows);
            IFilter retrievedFilter = worksheet.Filter.Filters.First();
            Assert.AreEqual(0, retrievedFilter.RelativeColumnIndex);
            Assert.AreEqual(filter2, retrievedFilter);
            Assert.AreEqual(1, worksheet.Filter.Filters.Count());
        }

        [TestMethod]
        public void SetFiltersOnTwoColumns()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            List<object> values = new List<object>() 
            { 
                new DateTime(1986, 4, 2),
                new DateTime(1986, 4, 3),
                1d,
                2d,
                new DateTime(2013, 1, 1),
                new DateTime(2013, 12, 5, 5, 5, 30),
                new DateTime(2014, 12, 5, 5, 5, 30)
            };
            FilterTestHelperInternal.SetValuesToWorksheetColumn(worksheet, 2, 5, values);

            List<object> values2 = new List<object>() 
            { 
                1d, 2d, 3d, 4d, 5d, 6d, 7d
            };
            FilterTestHelperInternal.SetValuesToWorksheetColumn(worksheet, 2, 6, values2);

            worksheet.Filter.FilterRange = new CellRange(1, 5, 8, 6);

            IFilter filter = new ValuesCollectionFilter(0, FilterTestHelperInternal.StringValues, FilterTestHelperInternal.DateValues, false);
            worksheet.Filter.SetFilter(filter);
            IFilter filter2 = new ValuesCollectionFilter(1, FilterTestHelperInternal.StringValues, FilterTestHelperInternal.DateValues, false);
            worksheet.Filter.SetFilter(filter2);

            Dictionary<int, bool> expectedRows = new Dictionary<int, bool>()
            {
               {1, false},
               {2, false},
               {3, true},
               {4, true},
               {5, true},
               {6, true},
               {7, true},
               {8, true},
            };
            FilterTestHelperInternal.AssertCorrectRowsHidden(worksheet, expectedRows);
        }

        [TestMethod]
        public void SetFilterChangeValueSetNewFilterAndUndo()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FilterTestHelperInternal.SetValuesToWorksheetColumn(worksheet, 6, 5, new List<object>()
            {
                "1", "2", "3", "test"
            });

            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            IFilter filter = new ValuesCollectionFilter(0, FilterTestHelperInternal.StringValues, null, true);
            worksheet.Filter.SetFilter(filter);
            worksheet.Cells[6, 5, 6, 5].SetValue(3);

            IFilter newFilter = new DynamicFilter(0, DynamicFilterType.April);
            worksheet.Filter.SetFilter(newFilter);
            worksheet.Workbook.History.Undo();

            FilterTestHelperInternal.AssertCorrectRowsHidden(worksheet, new Dictionary<int, bool>()
            {
                {6, false},
                {7, false},
                {8, true},
                {9, false},
            });
        }

        [TestMethod]
        public void SetFilterOnSingleRowRangeAssertNoException()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            worksheet.Filter.FilterRange = new CellRange(0, 0, 0, 0);
            IFilter filter = new ValuesCollectionFilter(0, FilterTestHelperInternal.StringValues, null, true);
            worksheet.Filter.SetFilter(filter);
        }

        #endregion


        #region RemoveFilter

        [TestMethod]
        public void RemoveFilterTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            IFilter filter = new DynamicFilter(1, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter);

            worksheet.Filter.RemoveFilter(1);

            int filterColumnCount = worksheet.Filter.Filters.Count;
            Assert.AreEqual(0, filterColumnCount);
        }

        [TestMethod]
        public void RemoveFilterTestUndoRedo()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            IFilter filter = new DynamicFilter(1, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter);

            worksheet.Filter.RemoveFilter(1);

            worksheet.Workbook.History.Undo();

            int filterCount = worksheet.Filter.Filters.Count;
            Assert.AreEqual(1, filterCount);
            IFilter retrievedFilter = worksheet.Filter.Filters.First();
            Assert.AreEqual(1, retrievedFilter.RelativeColumnIndex);
            Assert.AreEqual(filter, retrievedFilter);

            worksheet.Workbook.History.Redo();

            FilterTestHelperInternal.AssertAllRowsAreUnhidden(worksheet, FilterTestHelperInternal.DefaultCellRange);
            filterCount = worksheet.Filter.Filters.Count;
            Assert.AreEqual(0, filterCount);
        }

        [TestMethod]
        public void SetFilterChangeValueAndRemoveFilterUndo()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FilterTestHelperInternal.SetValuesToWorksheetColumn(worksheet, 6, 6, new List<object>()
            {
                new DateTime(1986, 4, 2),
                new DateTime(1986, 5, 2)
            });

            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            IFilter filter = new DynamicFilter(1, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter);
            worksheet.Cells[6, 6, 6, 6].SetValue(new DateTime(1986, 5, 2));

            worksheet.Filter.RemoveFilter(1);
            worksheet.Workbook.History.Undo();

            FilterTestHelperInternal.AssertCorrectRowsHidden(worksheet, new Dictionary<int, bool>()
            {
                {6, false},
                {7, true}
            });
        }

        [TestMethod]
        public void SetFiltersOnTwoColumnsAndRemoveSecond()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            List<object> values = new List<object>() 
            { 
                new DateTime(1986, 4, 2),
                new DateTime(1986, 4, 3),
                1d,
                2d,
                new DateTime(2013, 1, 1),
                new DateTime(2013, 12, 5, 5, 5, 30),
                new DateTime(2014, 12, 5, 5, 5, 30)
            };
            FilterTestHelperInternal.SetValuesToWorksheetColumn(worksheet, 2, 5, values);

            List<object> values2 = new List<object>() 
            { 
                1d, 2d, 3d, 4d, 5d, 6d, 7d
            };
            FilterTestHelperInternal.SetValuesToWorksheetColumn(worksheet, 2, 6, values2);

            worksheet.Filter.FilterRange = new CellRange(1, 5, 8, 6);

            IFilter filter = new ValuesCollectionFilter(0, FilterTestHelperInternal.StringValues, FilterTestHelperInternal.DateValues, false);
            worksheet.Filter.SetFilter(filter);
            IFilter filter2 = new ValuesCollectionFilter(1, FilterTestHelperInternal.StringValues, FilterTestHelperInternal.DateValues, false);
            worksheet.Filter.SetFilter(filter2);
            worksheet.Filter.RemoveFilter(1);

            Dictionary<int, bool> expectedRows = new Dictionary<int, bool>()
            {
               {1, false},
               {2, false},
               {3, true},
               {4, false},
               {5, false},
               {6, false},
               {7, false},
               {8, true},
            };
            FilterTestHelperInternal.AssertCorrectRowsHidden(worksheet, expectedRows);
        }

        [TestMethod]
        public void SetThreeFiltersUniquelyHiddenRowsTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            List<object> values0 = new List<object>() 
            { 
                new DateTime(1986, 4, 2),
                new DateTime(1986, 4, 3),
                1d,
                2d,
                new DateTime(2013, 1, 1),
                new DateTime(2013, 12, 5, 5, 5, 30),
                new DateTime(2014, 12, 5, 5, 5, 30)
            };
            FilterTestHelperInternal.SetValuesToWorksheetColumn(worksheet, 2, 5, values0);

            List<object> values1 = new List<object>() 
            { 
                1d, 2d, 3d, 4d, 5d, 6d, 7d
            };
            FilterTestHelperInternal.SetValuesToWorksheetColumn(worksheet, 2, 6, values1);

            worksheet.Filter.FilterRange = new CellRange(1, 5, 8, 6);
            IFilter filter0 = new ValuesCollectionFilter(0, null, FilterTestHelperInternal.DateValues, false);
            worksheet.Filter.SetFilter(filter0);
            IFilter filter1 = new ValuesCollectionFilter(1, FilterTestHelperInternal.StringValues, null, false);
            worksheet.Filter.SetFilter(filter1);

            Dictionary<int, bool> expectedRows = new Dictionary<int, bool>()
            {
               {1, false},
               {2, false},
               {3, true},
               {4, true},
               {5, true},
               {6, true},
               {7, true},
               {8, true},
            };
            FilterTestHelperInternal.AssertCorrectRowsHidden(worksheet, expectedRows);

            ICompressedList<bool> uniquelyHiddenRows0 = worksheet.Filter.GetRowsHiddenUniquelyByFilter(0);
            int hiddenRowCount0 = FilterTestHelperInternal.CountRowsFromRanges(uniquelyHiddenRows0);
            Assert.AreEqual(1, hiddenRowCount0);

            ICompressedList<bool> uniquelyHiddenRows1 = worksheet.Filter.GetRowsHiddenUniquelyByFilter(1);
            int hiddenRowCount1 = FilterTestHelperInternal.CountRowsFromRanges(uniquelyHiddenRows1);
            Assert.AreEqual(2, hiddenRowCount1);
        }

        #endregion


        #region Clear filter

        [TestMethod]
        public void ClearFilterTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            IFilter filter1 = new DynamicFilter(1, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter1);
            IFilter filter0 = new DynamicFilter(0, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter0);

            worksheet.Filter.ClearFilters();

            FilterTestHelperInternal.AssertAllRowsAreUnhidden(worksheet, FilterTestHelperInternal.DefaultCellRange);
            int filterColumnCount = worksheet.Filter.Filters.Count;
            Assert.AreEqual(0, filterColumnCount);
        }

        [TestMethod]
        public void SetFilterChangeValueClearFilterAndUndoTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FilterTestHelperInternal.SetValuesToWorksheetColumn(worksheet, 6, 5, new List<object>()
            {
                "1", "1.2", "2", "test"
            });
            FilterTestHelperInternal.SetValuesToWorksheetColumn(worksheet, 6, 6, new List<object>()
            {
                new DateTime(2013, 4, 4),
                new DateTime(2013, 4, 4),
                new DateTime(2013, 4, 4),
                new DateTime(2013, 4, 4),
            });
            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            IFilter filter = new ValuesCollectionFilter(0, FilterTestHelperInternal.StringValues, null, true);
            worksheet.Filter.SetFilter(filter);
            IFilter filter2 = new DynamicFilter(1, DynamicFilterType.April);
            worksheet.Filter.SetFilter(filter2);
            worksheet.Cells[6, 5, 6, 5].SetValue(100);

            worksheet.Filter.ClearFilters();

            FilterTestHelperInternal.AssertAllRowsAreUnhidden(worksheet, FilterTestHelperInternal.DefaultCellRange);
        }

        #endregion


        #region Reapply filter

        [TestMethod]
        public void ReapplyAfterChangingAValue()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            List<object> values = new List<object>() { 1d, 2d, 3d };
            FilterTestHelperInternal.SetValuesToWorksheetColumn(worksheet, 6, 5, values);

            worksheet.Filter.SetFilter(new ValuesCollectionFilter(0, FilterTestHelperInternal.StringValues, null, true));
            worksheet.Cells[6, 5, 6, 5].SetValue(3);
            worksheet.Filter.ReapplyFilter(0);

            Dictionary<int, bool> expectedRows = new Dictionary<int, bool>()
            {
                { 5, false },
                { 6, true },
                { 7, false },
                { 8, true },
            };
            FilterTestHelperInternal.AssertCorrectRowsHidden(worksheet, expectedRows);
        }

        [TestMethod]
        public void ReapplyAfterChangingAValueAndUndoAndRedo()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            List<object> values = new List<object>() { 1d, 2d, 3d };
            FilterTestHelperInternal.SetValuesToWorksheetColumn(worksheet, 6, 5, values);

            worksheet.Filter.SetFilter(new ValuesCollectionFilter(0, FilterTestHelperInternal.StringValues, null, true));
            worksheet.Cells[6, 5, 6, 5].SetValue(3);
            worksheet.Filter.ReapplyFilter(0);
            worksheet.Workbook.History.Undo();

            Dictionary<int, bool> expectedRows = new Dictionary<int, bool>()
            {
                { 5, false },
                { 6, false },
                { 7, false },
                { 8, true },
            };
            FilterTestHelperInternal.AssertCorrectRowsHidden(worksheet, expectedRows);

            worksheet.Workbook.History.Redo();

            expectedRows = new Dictionary<int, bool>()
            {
                { 5, false },
                { 6, true },
                { 7, false },
                { 8, true },
            };
            FilterTestHelperInternal.AssertCorrectRowsHidden(worksheet, expectedRows);
        }

        #endregion
    }
}
