﻿using System;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.TextBased.Csv;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal.FormatProviders
{
    [TestClass]
    public class ExportFormattedValuesToCsv : MultiCulturalTestClassBase
    {
        public void AssertCellValueExportedCorrectly(string cellValue, string formatString, string expectedExportValue)
        {
            RunTestInMultipleCultures(() => 
            {
                Workbook workbook = new Workbook();
                Worksheet worksheet = workbook.Worksheets.Add();

                worksheet.Cells[0, 0].SetValue(cellValue);
                if (!string.IsNullOrEmpty(formatString))
                {
                    worksheet.Cells[0, 0].SetFormat(new CellValueFormat(formatString));
                }

                string actualExportValue = new CsvFormatProvider().Export(workbook);
                actualExportValue = SanitizeNewLine(actualExportValue);
                Assert.AreEqual(expectedExportValue, actualExportValue);

            }, new string[] { "en-US" } );
        }

        private static string SanitizeNewLine(string value)
        {
            if (value.EndsWith("\r\n"))
            {
                return value.Substring(0, value.Length - 2);
            }

            return value;
        }

        [TestMethod]
        public void ExportGeneralFormatFomulaCellValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("=SUM(1, 2)", string.Empty, "3");
        }

        [TestMethod]
        public void ExportNumberFormatFomulaCellValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("=SUM(1, 2)", "0.00", "3.00");
        }
        
        [TestMethod]
        public void ExportCurrencyFormatFomulaCellValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("=SUM(1, 2)", "$#,##0.00", "$3.00");
        }
        
        [TestMethod]
        public void ExportDateFormatFomulaCellValueThatAutoSetsFormatToCsv()
        {
            this.AssertCellValueExportedCorrectly("=DATE(2020, 4, 10)", string.Empty, "4/10/2020");
        }    
    
        [TestMethod]
        public void ExportGeneralFormatNumberCellValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.56", string.Empty, "1234.56");
        }

        [TestMethod]
        public void ExportNumberFormatPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.56", "0.00", "1234.56");
        }

        [TestMethod]
        public void ExportNumberFormatColorPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.56", "0.00;[Red]0.00", "1234.56");
        }

        [TestMethod]
        public void ExportNumberFormatInvisibleSymbolsPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.56", "0.00_);(0.00)", "1234.56 ");
        }

        [TestMethod]
        public void ExportNumberFormatColorAndInvisibleSymbolsPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.56", "0.00_);[Red](0.00)", "1234.56 ");
        }

        [TestMethod]
        public void ExportNumberFormatNegativeValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.56", "0.00", "-1234.56");
        }

        [TestMethod]
        public void ExportNumberFormatColorNegativeValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.56", "0.00;[Red]0.00", "1234.56");
        }

        [TestMethod]
        public void ExportNumberFormatInvisibleSymbolsNegativeValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.56", "0.00_);(0.00)", "(1234.56)");
        }

        [TestMethod]
        public void ExportNumberFormatColorAndInvisibleSymbolsNegativeValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.56", "0.00_);[Red](0.00)", "(1234.56)");
        }

        [TestMethod]
        public void ExportCurrencyFormatPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "$#,##0.00", "\"$1,234.57\"");
        }

        [TestMethod]
        public void ExportCurrencyFormatColorPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "$#,##0.00;[Red]$#,##0.00", "\"$1,234.57\"");
        }

        [TestMethod]
        public void ExportCurrencyFormatInvisibleSymbolsPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "$#,##0.00_);($#,##0.00)", "\"$1,234.57 \"");
        }

        [TestMethod]
        public void ExportCurrencyFormatColorAndInvisibleSymbolsPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "$#,##0.00_);[Red]($#,##0.00)", "\"$1,234.57 \"");
        }

        [TestMethod]
        public void ExportCurrencyFormatNegativeValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.5678", "$#,##0.00", "\"-$1,234.57\"");
        }

        [TestMethod]
        public void ExportCurrencyFormatColorNegativeValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.5678", "$#,##0.00;[Red]$#,##0.00", "\"$1,234.57\"");
        }

        [TestMethod]
        public void ExportCurrencyFormatInvisibleSymbolsNegativeValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.5678", "$#,##0.00_);($#,##0.00)", "\"($1,234.57)\"");
        }

        [TestMethod]
        public void ExportCurrencyFormatColorAndInvisibleSymbolsNegativeValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.5678", "$#,##0.00_);[Red]($#,##0.00)", "\"($1,234.57)\"");
        }

        [TestMethod]
        public void ExportAccountingFormatPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "_($* #,##0.00_);_($* (#,##0.00);_($* \" - \"??_);_(@_)", "\" $1,234.57 \"");
        }

        [TestMethod]
        public void ExportAccountingFormatThreeDecPlacesPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "_($* #,##0.000_);_($* (#,##0.000);_($* \" - \"???_);_(@_)", "\" $1,234.568 \"");
        }

        [TestMethod]
        public void ExportAccountingFormatFourDecPlacesPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "_($* #,##0.0000_);_($* (#,##0.0000);_($* \" - \"???_);_(@_)", "\" $1,234.5678 \"");
        }

        [TestMethod]
        public void ExportAccountingFormatNoSymbolPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "_(* #,##0.00_);_(* (#,##0.00);_(* \" - \"??_);_(@_)", "\" 1,234.57 \"");
        }

        [TestMethod]
        public void ExportAccountingFormatUKPoundsSymbolPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "_-[$£-809]* #,##0.00_-;-[$£-809]* #,##0.00_-;_-[$£-809]* \" - \"??_-;_-@_-", "\" £1,234.57 \"");
        }

        [TestMethod]
        public void ExportAccountingFormatNegativeValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.5678", "_($* #,##0.00_);_($* (#,##0.00);_($* \" - \"??_);_(@_)", "\" $(1,234.57)\"");
        }

        [TestMethod]
        public void ExportAccountingFormatThreeDecPlacesNegativeValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.5678", "_($* #,##0.000_);_($* (#,##0.000);_($* \" - \"???_);_(@_)", "\" $(1,234.568)\"");
        }

        [TestMethod]
        public void ExportAccountingFormatFourDecPlacesNegativeValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.5678", "_($* #,##0.0000_);_($* (#,##0.0000);_($* \" - \"???_);_(@_)", "\" $(1,234.5678)\"");
        }

        [TestMethod]
        public void ExportAccountingFormatNoSymbolNegativeValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.5678", "_(* #,##0.00_);_(* (#,##0.00);_(* \" - \"??_);_(@_)", "\" (1,234.57)\"");
        }

        [TestMethod]
        public void ExportAccountingFormatUKPoundsSymbolNegativeValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.5678", "_-[$£-809]* #,##0.00_-;-[$£-809]* #,##0.00_-;_-[$£-809]* \" - \"??_-;_-@_-", "\"-£1,234.57 \"");
        }

        [TestMethod]
        public void ExportDateFormatMDYYYValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("4321.9876", "m/d/yyyy", "10/30/1911");
        }

        [TestMethod]
        public void ExportDateFormatMonthDDYYYValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("4321.9876", "[$-F800]dddd, mmmm dd, yyyy", "\"Monday, October 30, 1911\"");
        }

        [TestMethod]
        public void ExportDateFormatMDValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("4321.9876", "m/d;@", "10/30");
        }

        [TestMethod]
        public void ExportDateFormatMDYYValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("4321.9876", "m/d/yy;@", "10/30/11");
        }

        [TestMethod]
        public void ExportDateFormatMMDDYYValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("4321.9876", "mm/dd/yy;@", "10/30/11");
        }

        [TestMethod]
        public void ExportDateFormatDMMMValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("4321.9876", "[$-409]d-mmm;@", "30-Oct");
        }

        [TestMethod]
        public void ExportDateFormatDMMMYYValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("4321.9876", "[$-409]d-mmm-yy;@", "30-Oct-11");
        }

        [TestMethod]
        public void ExportDateFormatMMMYYValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("4321.9876", "[$-409]mmm-yy;@", "Oct-11");
        }

        [TestMethod]
        public void ExportDateFormatMMMMYYValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("4321.9876", "[$-409]mmmm-yy;@", "October-11");
        }

        [TestMethod]
        public void ExportTimeFormatHMMSSAMPMValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "[$-F400]h:mm:ss AM/PM", "1:37:38 PM");
        }

        [TestMethod]
        public void ExportTimeFormatHMMValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "h:mm;@", "13:37");
        }

        [TestMethod]
        public void ExportTimeFormatHMMAMPMValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "[$-409]h:mm AM/PM;@", "1:37 PM");
        }

        [TestMethod]
        public void ExportTimeFormatHMMSSValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "h:mm:ss;@", "13:37:38");
        }

        [TestMethod]
        public void ExportTimeFormatHMMSSAMPM2ValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "[$-409]h:mm:ss AM/PM;@", "1:37:38 PM");
        }

        [TestMethod]
        public void ExportTimeFormatMMSS0ValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "mm:ss.0;@", "37:37.9");
        }

        [TestMethod]
        public void ExportTimeFormatHinBracketsMMSSValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "[h]:mm:ss;@", "29629:37:38");
        }

        [TestMethod]
        public void ExportTimeFormatMDYYHMMAMPMValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "[$-409]m/d/yy h:mm AM/PM;@", "5/18/03 1:37 PM");
        }

        [TestMethod]
        public void ExportTimeFormatMDYYHMMValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "m/d/yy h:mm;@", "5/18/03 13:37");
        }

        [TestMethod]
        public void ExportPercentFormatTwoDecPlacesPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "0.00%", "123456.78%");
        }

        [TestMethod]
        public void ExportPercentFormatThreeDecPlacesPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "0.000%", "123456.780%");
        }

        [TestMethod]
        public void ExportPercentFormatFourDecPlacesPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "0.0000%", "123456.7800%");
        }

        [TestMethod]
        public void ExportPercentFormatZeroValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("0", "0.00%", "0.00%");
        }

        [TestMethod]
        public void ExportPercentFormatTwoDecPlacesNegativeValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.5678", "0.00%", "-123456.78%");
        }

        [TestMethod]
        public void ExportPercentFormatThreeDecPlacesNegativeValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.5678", "0.000%", "-123456.780%");
        }

        [TestMethod]
        public void ExportPercentFormatFourDecPlacesNegativeValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.5678", "0.0000%", "-123456.7800%");
        }

        [TestMethod]
        public void ExportFractionFormatOneDigitValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "# ?/?", "1234 4/7");
        }

        [TestMethod]
        public void ExportFractionFormatTwoDigitsValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "# ??/??", "1234 46/81");
        }

        [TestMethod]
        public void ExportFractionFormatThreeDigitsValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "# ???/???", "1234  67/118");
        }

        [TestMethod]
        public void ExportFractionFormatAsHalvesValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "# ?/2", "1234 1/2");
        }

        [TestMethod]
        public void ExportFractionFormatAsQuartersValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "# ?/4", "1234 2/4");
        }

        [TestMethod]
        public void ExportFractionFormatAsEighthsValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("9876.54321", "# ?/8", "9876 4/8");
        }

        [TestMethod]
        public void ExportFractionFormatAsSixteenthsValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "# ??/16", "1234  9/16");
        }

        [TestMethod]
        public void ExportFractionFormatAsTenthsValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "# ?/10", "1234 6/10");
        }

        [TestMethod]
        public void ExportFractionFormatAsHundredthsValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "# ??/100", "1234 57/100");
        }

        [TestMethod]
        public void ExportScientificFormatPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "0.00E+00", "1.23E+03");
        }

        [TestMethod]
        public void ExportScientificFormatThreeDecDigitsPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "0.000E+00", "1.235E+03");
        }

        [TestMethod]
        public void ExportScientificFormatFourDecDigitsPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "0.0000E+00", "1.2346E+03");
        }

        [TestMethod]
        public void ExportScientificFormatNoDecDigitsPositiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("1234.5678", "0E+00", "1E+03");
        }

        [TestMethod]
        public void ExportScientificFormatNagetiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.5678", "0.00E+00", "-1.23E+03");
        }

        [TestMethod]
        public void ExportScientificFormatThreeDecDigitsNagetiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.5678", "0.000E+00", "-1.235E+03");
        }

        [TestMethod]
        public void ExportScientificFormatFourDecDigitsNagetiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.5678", "0.0000E+00", "-1.2346E+03");
        }

        [TestMethod]
        public void ExportScientificFormatNoDecDigitsNagetiveValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("-1234.5678", "0E+00", "-1E+03");
        }

        [TestMethod]
        public void ExportScientificFormatZeroValueToCsv()
        {
            this.AssertCellValueExportedCorrectly("0", "0.00E+00", "0.00E+00");
        }
    }
}
