﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Expressions.Functions;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;
using Telerik.Windows.Documents.Spreadsheet.TestsInternal.Expressions;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal.Functions.FunctionHelperClassTests
{
    [TestClass]
    public class ReferenceArgumentConversion
    {
        [TestMethod]
        public void ConvertEmptyExpressionToReference()
        {
            FunctionsTestHelper.AssertConversionCorrectness(EmptyExpression.Empty, null, ArgumentType.Reference, ErrorExpressions.ValueError, 0, null);
        }

        [TestMethod]
        public void ConvertNumberExpressionToReference()
        {
            FunctionsTestHelper.AssertConversionCorrectness(NumberExpression.PI, null, ArgumentType.Reference, ErrorExpressions.ValueError, 0, null);
        }

        [TestMethod]
        public void ConvertArrayExpressionToReference()
        {
            ArrayExpression expression = new ArrayExpression(new RadExpression[,] { { new NumberExpression(5) } });
            FunctionsTestHelper.AssertConversionCorrectness(expression, null, ArgumentType.Reference, ErrorExpressions.ValueError, 0, null);
        }

        [TestMethod]
        public void ConvertCellReferenceExpressionToReference()
        {
            RadExpression expression = FunctionsTestHelper.GetIndirectArgument(string.Empty);
            FunctionsTestHelper.AssertConversionCorrectness(expression, null, ArgumentType.Reference, null, 1, expression);
        }

        [TestMethod]
        public void ConvertRangeExpressionToReference()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            CellAreaReference a1Range = CellAreaReference.CreateFromCellName("A1", new CellIndex(10, 10));
            CellAreaReference b2Range = CellAreaReference.CreateFromCellName("B2", new CellIndex(10, 10));
            CellReferenceRangeExpression a1CellReference = new CellReferenceRangeExpression(workbook, "Sheet1", false, a1Range);
            CellReferenceRangeExpression b2CellReference = new CellReferenceRangeExpression(workbook, "Sheet1", false, b2Range);

            RangeExpression expression = new RangeExpression(a1CellReference, b2CellReference);
            CellReferenceRangeExpression exprectedResult = new CellReferenceRangeExpression(workbook, "Sheet1", false, CellAreaReference.CreateRange(a1Range, b2Range));

            FunctionsTestHelper.AssertConversionCorrectness(expression, null, ArgumentType.Reference, null, 1, exprectedResult);
        }

        [TestMethod]
        public void ConvertIntersectionExpressionToReference()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            CellAreaReference a1Range = CellAreaReference.CreateFromCellName("A1", new CellIndex(10, 10));
            CellAreaReference b2Range = CellAreaReference.CreateFromCellName("B2", new CellIndex(10, 10));
            CellAreaReference c3Range = CellAreaReference.CreateFromCellName("C3", new CellIndex(10, 10));
            CellAreaReference d4Range = CellAreaReference.CreateFromCellName("D4", new CellIndex(10, 10));
            
            CellReferenceRangeExpression a1CellReference = new CellReferenceRangeExpression(workbook, "Sheet1", false, a1Range);
            CellReferenceRangeExpression b2CellReference = new CellReferenceRangeExpression(workbook, "Sheet1", false, b2Range);
            CellReferenceRangeExpression c3CellReference = new CellReferenceRangeExpression(workbook, "Sheet1", false, c3Range);
            CellReferenceRangeExpression d4CellReference = new CellReferenceRangeExpression(workbook, "Sheet1", false, d4Range);

            RangeExpression range1 = new RangeExpression(a1CellReference, c3CellReference);
            RangeExpression range2 = new RangeExpression(b2CellReference, d4CellReference);
            IntersectionExpression expression = new IntersectionExpression(range1, range2);
            CellReferenceRangeExpression exprectedResult = new CellReferenceRangeExpression(workbook, "Sheet1", false, CellAreaReference.CreateRange(b2Range, c3Range));

            FunctionsTestHelper.AssertConversionCorrectness(expression, null, ArgumentType.Reference, null, 1, exprectedResult);
        }

        [TestMethod]
        public void ConvertNameExpressionWithNumberToReference()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            workbook.Names.Add("test", "=123", new CellIndex(10, 10));

            SpreadsheetNameExpression expression = new SpreadsheetNameExpression(new ExpressionQualifierInfo(workbook, workbook.ActiveWorksheet, ""), new CellIndex(10, 10), "test");

            FunctionsTestHelper.AssertConversionCorrectness(expression, null, ArgumentType.Reference, ErrorExpressions.ValueError, 0, null);
        }

        [TestMethod]
        public void ConvertNameExpressionWithCellReferenceToReference()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            workbook.Names.Add("test", "=Sheet1!C3", new CellIndex(10, 10));

            CellAreaReference c3Range = CellAreaReference.CreateFromCellName("C3", new CellIndex(10, 10));
            CellReferenceRangeExpression c3CellReference = new CellReferenceRangeExpression(workbook, "Sheet1", true, c3Range);

            SpreadsheetNameExpression expression = new SpreadsheetNameExpression(new ExpressionQualifierInfo(workbook, workbook.ActiveWorksheet, ""), new CellIndex(10, 10), "test");

            FunctionsTestHelper.AssertConversionCorrectness(expression, null, ArgumentType.Reference, null, 1, c3CellReference);
        }

        [TestMethod]
        public void ConvertRangeExpressionOfReferenceAndNameToReference()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            workbook.Names.Add("test", "=Sheet1!B2", new CellIndex(10, 10));

            CellAreaReference a1Range = CellAreaReference.CreateFromCellName("A1", new CellIndex(10, 10));
            CellAreaReference b2Range = CellAreaReference.CreateFromCellName("B2", new CellIndex(10, 10));
            CellReferenceRangeExpression a1CellReference = new CellReferenceRangeExpression(workbook, "Sheet1", false, a1Range);
            SpreadsheetNameExpression nameExpression = new SpreadsheetNameExpression(new ExpressionQualifierInfo(workbook, workbook.ActiveWorksheet, ""), new CellIndex(10, 10), "test");

            RangeExpression expression = new RangeExpression(a1CellReference, nameExpression);
            CellReferenceRangeExpression exprectedResult = new CellReferenceRangeExpression(workbook, "Sheet1", false, CellAreaReference.CreateRange(a1Range, b2Range));

            FunctionsTestHelper.AssertConversionCorrectness(expression, null, ArgumentType.Reference, null, 1, exprectedResult);
        }

        [TestMethod]
        public void ConvertRangeExpressionOfTwoNamesToReference()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            workbook.Names.Add("testA", "=Sheet1!A1", new CellIndex(10, 10));
            workbook.Names.Add("testB", "=Sheet1!B2", new CellIndex(10, 10));

            CellAreaReference a1Range = CellAreaReference.CreateFromCellName("A1", new CellIndex(10, 10));
            CellAreaReference b2Range = CellAreaReference.CreateFromCellName("B2", new CellIndex(10, 10));
            
            SpreadsheetNameExpression aNameExpression = new SpreadsheetNameExpression(new ExpressionQualifierInfo(workbook, workbook.ActiveWorksheet, ""), new CellIndex(10, 10), "testA");
            SpreadsheetNameExpression bNameExpression = new SpreadsheetNameExpression(new ExpressionQualifierInfo(workbook, workbook.ActiveWorksheet, ""), new CellIndex(10, 10), "testB");

            RangeExpression expression = new RangeExpression(aNameExpression, bNameExpression);
            CellReferenceRangeExpression exprectedResult = new CellReferenceRangeExpression(workbook, "Sheet1", true, CellAreaReference.CreateRange(a1Range, b2Range));

            FunctionsTestHelper.AssertConversionCorrectness(expression, null, ArgumentType.Reference, null, 1, exprectedResult);
        }
    }
}

