﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Expressions;
using Telerik.Windows.Documents.Spreadsheet.Expressions.Functions;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;
using Telerik.Windows.Documents.Spreadsheet.TestsInternal.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal.Expressions
{
    public static class FunctionsTestHelper
    {
        #region Fields

        private readonly static Worksheet worksheet;

        #endregion


        #region Constructors

        static FunctionsTestHelper()
        {
            worksheet = TestHelper.CreateWorksheet();
            InitializeWorksheet();
        }

        #endregion


        #region Methods

        private static void InitializeWorksheet()
        {
            worksheet.Cells[0, 0].SetValue(10);
            worksheet.Cells[0, 1].SetValue("=CONCATENATE(\"20\")");
            worksheet.Cells[0, 2].SetValue("string");
            worksheet.Cells[0, 3].SetValue("=5/0");
            worksheet.Cells[0, 4].SetValue("=LN(0)");
            worksheet.Cells[0, 5].SetValue("true");
            worksheet.Cells[0, 6].SetValue("=MODE(1{ls}2{ls}3)".ReplaceListSeparator());
            
            worksheet.Cells[1, 0].SetValue("false");
            worksheet.Cells[1, 1].SetValue("true");
            worksheet.Cells[1, 2].SetValue(3);
            worksheet.Cells[1, 3].SetValue(4);
            worksheet.Cells[1, 4].SetValue("=CONCATENATE(\"5\")");
            worksheet.Cells[1, 5].SetValue("=CONCATENATE(\"6\")");
            worksheet.Cells[1, 6].SetValue("41320");

            worksheet.Cells[2, 0].SetValue((0.573599537037037).ToDefaultSpreadsheetCultureString());
            worksheet.Cells[2, 1].SetValue("=CONCATENATE(\"" + (0.573599537037037).ToDefaultSpreadsheetCultureString() + "\")");
            worksheet.Cells[10, 0].SetValue("=A12");
            worksheet.Cells[10, 1].SetValue("=A12:B12");
            worksheet.Cells[11, 0].SetValue(2);
            worksheet.Cells[11, 1].SetValue(3);
            worksheet.Cells[10, 2].SetValue("=C11");
        }

        public static void AssertFunctionResult(FunctionBase function, object expectedResult, params object[] arguments)
        {
            RadExpression[] expressionArguments = ConvertArgumentsToExpressions(arguments);

            RadExpression actualResultExpression = function.Evaluate(new FunctionEvaluationContext<RadExpression>(expressionArguments, null, 0, 0)).GetValue();
            RadExpression expectedResultExpression = ConvertArgumentToExpression(expectedResult);

            Assert.IsTrue(CompareExpressions(expectedResultExpression, actualResultExpression));
        }

        public static void AssertFunctionError(FunctionBase function, ErrorExpression errorExpression, params object[] arguments)
        {
            RadExpression[] expressionArguments = ConvertArgumentsToExpressions(arguments);
            RadExpression result = function.Evaluate(new FunctionEvaluationContext<RadExpression>(expressionArguments, null, 0, 0));

            Assert.AreEqual(errorExpression, result);
        }

        private static RadExpression[] ConvertArgumentsToExpressions(params object[] arguments)
        {
            List<RadExpression> expressionArguments = new List<RadExpression>();

            for (int i = 0; i < arguments.Length; i++)
            {
                RadExpression expression = ConvertArgumentToExpression(arguments[i]);
                if (expression != null)
                {
                    expressionArguments.Add(expression);
                }
            }

            return expressionArguments.ToArray();
        }

        private static RadExpression ConvertArgumentToExpression(object argument)
        {
            RadExpression argAsExpression = argument as RadExpression;
            if (argAsExpression != null)
            {
                return argAsExpression;
            }

            object[,] argAsArray = argument as object[,];
            if (argAsArray != null)
            {
                int rowCount = argAsArray.GetLength(0);
                int columnCount = argAsArray.GetLength(1);

                RadExpression[,] expressionArray = new RadExpression[rowCount, columnCount];

                for (int rowIndex = 0; rowIndex < rowCount; rowIndex++)
                {
                    for (int columnIndex = 0; columnIndex < columnCount; columnIndex++)
                    {
                        RadExpression result = ConvertArgumentToExpression(argAsArray[rowIndex, columnIndex]);
                        expressionArray[rowIndex, columnIndex] = result;
                    }
                }

                return new ArrayExpression(expressionArray);
            }

            string argAsString = argument as string;
            if (argAsString != null)
            {
                if (NameConverter.IsValidA1CellName(argAsString))
                {
                    return new CellReferenceRangeExpression(worksheet.Workbook, worksheet.Name, false, CellAreaReference.CreateFromCellName(argAsString, new CellIndex(20, 20)));
                }

                return new StringExpression(argAsString);
            }

            double argAsDouble;
            if (Double.TryParse(argument.ToString(), out argAsDouble))
            {
                return new NumberExpression(argAsDouble);
            }

            bool argAsBool;
            if (Boolean.TryParse(argument.ToString(), out argAsBool))
            {
                return argAsBool ? BooleanExpression.True : BooleanExpression.False;
            }

            return new StringExpression(argument.ToString());
        }

        private static bool CompareExpressions(RadExpression expected, RadExpression actual)
        {
            NumberExpression actualNumber = actual as NumberExpression;
            NumberExpression expectedNumber = expected as NumberExpression;

            if (actualNumber != null && expectedNumber != null)
            {
                return TestHelper.AreDoubleValuesEssentiallyEqual(actualNumber.Value, expectedNumber.Value);
            }

            StringExpression actualString = actual as StringExpression;
            StringExpression expectedString = expected as StringExpression;

            if (actualString != null && expectedString != null)
            {
                return String.Compare(actualString.Value, expectedString.Value) == 0;
            }

            BooleanExpression actualBoolean = actual as BooleanExpression;
            BooleanExpression expectedBoolean = expected as BooleanExpression;

            if (actualBoolean != null && expectedBoolean != null)
            {
                return actualBoolean.Value == expectedBoolean.Value;
            }

            ErrorExpression actualError = actual as ErrorExpression;
            ErrorExpression expectedError = expected as ErrorExpression;

            if (actualError != null && expectedError != null)
            {
                return actualError.Equals(expectedError);
            }

            return false;
        }

        public static void AssertConversionCorrectness(RadExpression expression, ArgumentConversionRules rules, ArgumentType type,
         ErrorExpression expectedError, int expectedAgrsCount, params object[] expectedArg)
        {
            List<object> convertedArguments = new List<object>();

            ErrorExpression errorExpression = FunctionHelper.TryConvertArgument(expression, rules, type, convertedArguments);

            Assert.AreEqual(expectedError, errorExpression);
            Assert.AreEqual(expectedAgrsCount, convertedArguments.Count);

            if (expectedAgrsCount > 0)
            {
                Assert.AreEqual(expectedArg.Length, convertedArguments.Count);

                for (int i = 0; i < expectedArg.Length; i++)
                {
                    Assert.AreEqual(expectedArg[i], convertedArguments[i]);    
                }
            }
        }

        public static RadExpression GetIndirectArgument(string cellValue)
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            workbook.Worksheets["Sheet1"].Cells[0, 0].SetValue(cellValue);
            return new CellReferenceRangeExpression(workbook, "Sheet1", false, CellAreaReference.CreateFromCellName("A1", new CellIndex(1, 1)));
        }

        #endregion
    }
}
