﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal
{
    [TestClass]
    public class HyperlinkTests
    {
        #region Fields

        private Worksheet worksheet;

        #endregion


        #region Initialize Methods

        [TestInitialize]
        public void TestInitializers()
        {
            this.worksheet = TestHelper.CreateWorksheet();
        }

        [TestCleanup]
        public void TestCleanup()
        {
            this.worksheet = null;
        }

        #endregion


        #region Helper Methods

        private SpreadsheetHyperlink CreateAndAddHyperlink(Worksheet worksheet)
        {
            CellRange hyperlinkRange = new CellRange(2, 2, 4, 4);
            return worksheet.Hyperlinks.Add(hyperlinkRange, HyperlinkInfo.CreateInDocumentHyperlink("U10:V15"));
        }

        private SpreadsheetHyperlink CreateAndAddHyperlink(Worksheet worksheet, CellRange range, string name)
        {
            return worksheet.Hyperlinks.Add(range, HyperlinkInfo.CreateHyperlink(name, name));
        }

        public bool TryFireHyperlinkCollectionChanging(Worksheet worksheet, Action[] actions, out int counter)
        {
            bool isEventFired = false;
            int innerCounter = 0;

            worksheet.Hyperlinks.Changing += (s, e) =>
            {
                isEventFired = !isEventFired;
                innerCounter++;
            };

            foreach (Action action in actions)
            {
                action();
            }

            counter = innerCounter;

            return isEventFired;
        }

        #endregion


        #region Test Methods

        [TestMethod]
        public void HyperlinkCollectionChangingEvent_WhenHyperlinkIsAdded_IsFired()
        {
            Action insertHyperlink = new Action(() => this.CreateAndAddHyperlink(this.worksheet));
            int actualEventFiredCount;

            bool isChangingEventFired = this.TryFireHyperlinkCollectionChanging(this.worksheet, new Action[] { insertHyperlink }, out actualEventFiredCount);

            int expectedEventFiredCount = 1;

            Assert.AreEqual(expectedEventFiredCount, actualEventFiredCount);
            Assert.IsTrue(isChangingEventFired);
        }

        [TestMethod]
        public void HyperlinkCollectionChangingEvent_WhenHyperlinkIsRemoved_IsFired()
        {
            SpreadsheetHyperlink hyperlink = this.CreateAndAddHyperlink(this.worksheet);

            Action removeHyperlink = new Action(() => this.worksheet.Hyperlinks.Remove(hyperlink));
            int actualEventFiredCount;

            bool isChangingEventFired = this.TryFireHyperlinkCollectionChanging(this.worksheet, new Action[] { removeHyperlink }, out actualEventFiredCount);

            int expectedEventFiredCount = 1;

            Assert.AreEqual(expectedEventFiredCount, actualEventFiredCount);
            Assert.IsTrue(isChangingEventFired);
        }

        [TestMethod]
        public void HyperlinkCollectionChangingEvent_WhenUndoIsExecuted_IsFired()
        {
            this.CreateAndAddHyperlink(this.worksheet);

            Action invokeUndo = new Action(() => this.worksheet.Workbook.History.Undo());
            int actualEventFiredCount;

            bool isChangingEventFired = this.TryFireHyperlinkCollectionChanging(this.worksheet, new Action[] { invokeUndo }, out actualEventFiredCount);

            int expectedEventFiredCount = 1;

            Assert.AreEqual(expectedEventFiredCount, actualEventFiredCount);
            Assert.IsTrue(isChangingEventFired);
        }

        [TestMethod]
        public void HyperlinkCollectionChangingEvent_WhenRedoIsExecuted_IsFired()
        {
            this.CreateAndAddHyperlink(this.worksheet);
            this.worksheet.Workbook.History.Undo();

            Action invokeRedo = new Action(() => this.worksheet.Workbook.History.Redo());
            int actualEventFiredCount;

            bool isChangingEventFired = this.TryFireHyperlinkCollectionChanging(this.worksheet, new Action[] { invokeRedo }, out actualEventFiredCount);

            int expectedEventFiredCount = 1;

            Assert.AreEqual(expectedEventFiredCount, actualEventFiredCount);
            Assert.IsTrue(isChangingEventFired);
        }

        [TestMethod]
        public void HyperlinkCollectionChangingEvent_WhenAddRemoveUndoRedoAreExecuted_IsFiredProperTimes()
        {
            SpreadsheetHyperlink hyperlink = this.CreateAndAddHyperlink(this.worksheet);

            Action[] actions = new Action[]
            {
                () => this.worksheet.Hyperlinks.Remove(hyperlink),
                () => this.worksheet.Workbook.History.Undo(),
                () => this.worksheet.Workbook.History.Redo()
            };
            int actualEventFiredCount;

            this.TryFireHyperlinkCollectionChanging(this.worksheet, actions, out actualEventFiredCount);

            int expectedEventFiredCount = 3;

            Assert.AreEqual(expectedEventFiredCount, actualEventFiredCount);
        }

        #endregion


        #region Sort Test Methods

        [TestMethod]
        public void SortHyperlinksTest()
        {
            int rowsCount = 4;
            int columnsCount = 3;

            List<List<CellRange>> hyperlinkRanges = new List<List<CellRange>>();

            for (int column = 0; column < columnsCount; column++)
            {
                hyperlinkRanges.Add(new List<CellRange>());
                for (int row = 0; row < rowsCount; row++)
                {
                    hyperlinkRanges[column].Add(new CellRange(row, column, row, column));
                    this.CreateAndAddHyperlink(this.worksheet, hyperlinkRanges[column][row], row.ToString());
                }
            }

            CellRange sortRange = new CellRange(0, 0, rowsCount - 1, columnsCount - 1);
            int[] sortedIndexes = new int[] { 3, 1, 2, 0 };
            var hyperlinks = this.worksheet.Hyperlinks;

            hyperlinks.Sort(sortRange, sortedIndexes);

            for (int column = 0; column < columnsCount; column++)
            {
                for (int row = 0; row < rowsCount; row++)
                {
                    SpreadsheetHyperlink hyperlink;
                    hyperlinks.TryGetHyperlink(hyperlinkRanges[column][row], out hyperlink);

                    Assert.AreEqual(sortedIndexes[row].ToString(), hyperlink.HyperlinkInfo.ScreenTip);
                }
            }
        }

        #endregion
    }
}