﻿using System;
using System.Globalization;
using System.IO;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.FormatProviders.OpenXml.Xlsx;
using Telerik.Windows.Documents.Spreadsheet.Formatting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.DataValidation;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal.ImportExport
{
    [TestClass]
    public class DataValidationMultiCulture : MultiCulturalImportExportTestClassBase
    {
        private string ReplaceListSeparators(string value)
        {
            return value.Replace("{ls}", FormatHelper.DefaultSpreadsheetCulture.ListSeparator);
        }

        public override Workbook ArrangeWorkbook()
        {
            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets["Sheet1"];
            Cells cells = worksheet.Cells;

            string argument = this.ReplaceListSeparators("=Sum(A2{ls} A3{ls} A4)");

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, 0, 0);
            context.Argument1 = argument;
            context.ComparisonOperator = ComparisonOperator.EqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            cells[0, 0].SetDataValidationRule(rule);

            return workbook;
        }

        [TestMethod]
        [TestCategory("ImportExport")]
        public void ExportAndImportMultiCultureDataValidationTest()
        {
            ArrangeWorkbookExportImportAssert();
        }

        public override void AssertWorkbookImportedCorrectly(Workbook workbook)
        {
            throw new NotImplementedException();
        }

        [TestMethod]
        [TestCategory("ImportExport")]
        public void ExportDataValidationRuleInCultureWithOneListSeparatorImportInCultureWithAnotherListSeparator()
        {
            CultureInfo exportCulture = new CultureInfo("es-VE");
            SpreadsheetCultureHelper exportCultureInfo = new SpreadsheetCultureHelper(exportCulture);

            FormatHelper.DefaultSpreadsheetCulture = exportCultureInfo;

            Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
            Worksheet worksheet = workbook.Worksheets["Sheet1"];
            Cells cells = worksheet.Cells;

            string argument = this.ReplaceListSeparators("=Sum(A2{ls} A3{ls} A4)");

            NumberDataValidationRuleContext context = new NumberDataValidationRuleContext(worksheet, 0, 0);
            context.Argument1 = argument;
            context.ComparisonOperator = ComparisonOperator.EqualsTo;

            WholeNumberDataValidationRule rule = new WholeNumberDataValidationRule(context);
            cells[0, 0].SetDataValidationRule(rule);

            using (Stream stream = new MemoryStream())
            {
                XlsxFormatProvider provider = new XlsxFormatProvider();
                provider.Export(workbook, stream);

                FormatHelper.DefaultSpreadsheetCulture = SpreadsheetCultureHelper.InvariantSpreadsheetCultureInfo;
                Workbook actualWorkbook = provider.Import(stream);

                TestHelper.AssertWorkbooksAreEqual(workbook, actualWorkbook);
            }
        }

        [TestMethod]
        [TestCategory("ImportExport")]
        public void ExportDataValidationListRuleInCultureWithOneListSeparatorImportInCultureWithAnotherListSeparator()
        {
            foreach (string culture in this.DefaultCultures)
            {
                CultureInfo exportCulture = new CultureInfo(culture);
                SpreadsheetCultureHelper exportCultureInfo = new SpreadsheetCultureHelper(exportCulture);

                FormatHelper.DefaultSpreadsheetCulture = exportCultureInfo;

                Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
                Worksheet worksheet = workbook.Worksheets["Sheet1"];
                Cells cells = worksheet.Cells;

                ListDataValidationRule rule = this.CreateListDataValidationRule(worksheet);
                cells[0, 0].SetDataValidationRule(rule);

                using (Stream stream = new MemoryStream())
                {
                    XlsxFormatProvider provider = new XlsxFormatProvider();
                    provider.Export(workbook, stream);

                    FormatHelper.DefaultSpreadsheetCulture = SpreadsheetCultureHelper.InvariantSpreadsheetCultureInfo;
                    Workbook actualWorkbook = provider.Import(stream);

                    ListDataValidationRule expectedRule = this.CreateListDataValidationRule(actualWorkbook.Worksheets[0]);
                    ListDataValidationRule actualRule = actualWorkbook.Worksheets[0].Cells[0, 0].GetDataValidationRule().Value as ListDataValidationRule;

                    Assert.AreEqual(expectedRule, actualRule);
                }
            }
        }

        [TestMethod]
        [TestCategory("ImportExport")]
        public void ExportDataValidationListRuleWithFormulaInCultureWithOneListSeparatorImportInCultureWithAnotherListSeparator()
        {
            foreach (string culture in this.DefaultCultures)
            {
                FormatHelper.DefaultSpreadsheetCulture = SpreadsheetCultureHelper.InvariantSpreadsheetCultureInfo;
                Workbook workbook = TestHelper.CreateNonEmptyWorkbook();
                Worksheet worksheet = workbook.Worksheets["Sheet1"];
                Cells cells = worksheet.Cells;

                string argument = "=OFFSET(P3:Q8,MATCH($A3,P3:P8,0)-1,1,2,1)";

                ListDataValidationRuleContext context = new ListDataValidationRuleContext(worksheet, 0, 0);
                context.Argument1 = argument;
                context.InCellDropdown = true;

                ListDataValidationRule rule = new ListDataValidationRule(context);
                cells[0, 0].SetDataValidationRule(rule);

                using (Stream stream = new MemoryStream())
                {
                    XlsxFormatProvider provider = new XlsxFormatProvider();
                    provider.Export(workbook, stream);

                    CultureInfo importCulture = new CultureInfo(culture);
                    SpreadsheetCultureHelper exportCultureInfo = new SpreadsheetCultureHelper(importCulture);
                    FormatHelper.DefaultSpreadsheetCulture = exportCultureInfo;

                    Workbook actualWorkbook = provider.Import(stream);
                    ListDataValidationRule actualRule = actualWorkbook.Worksheets[0].Cells[0, 0].GetDataValidationRule().Value as ListDataValidationRule;

                    Assert.AreEqual(CellValueType.Formula, actualRule.Argument1.ValueType);
                }
            }
        }

        private ListDataValidationRule CreateListDataValidationRule(Worksheet worksheet)
        {
            string argument = this.ReplaceListSeparators("1{ls}2{ls}3{ls}4{ls}5");

            ListDataValidationRuleContext context = new ListDataValidationRuleContext(worksheet, 0, 0);
            context.Argument1 = argument;
            context.InCellDropdown = true;

            ListDataValidationRule rule = new ListDataValidationRule(context);
            return rule;
        }
    }
}