﻿using System;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal.InsertRemoveCells
{
    [TestClass]
    public class PropertyChangedRaisedUponInsertRemoveCellsTests
    {
        #region Fields

        private Worksheet worksheet;
        private PropertyChangedEventAsserter propertyChangedAsserter;
        private CellsInsertRemoveEventAsserter cellsInsertRemoveAsserter;

        #endregion


        #region Test Initialize and Cleanup

        [TestInitialize]
        public void TestInitialize()
        {
            this.worksheet = TestHelper.CreateWorksheet();
        }

        [TestCleanup]
        public void TestCleanup()
        {
            this.worksheet = null;
            this.propertyChangedAsserter = null;
            this.cellsInsertRemoveAsserter = null;
        }

        public void InitializeAsserters()
        {
            this.propertyChangedAsserter = new PropertyChangedEventAsserter(this.worksheet.Cells);
            this.cellsInsertRemoveAsserter = new CellsInsertRemoveEventAsserter(this.worksheet.Cells);
        }

        #endregion


        #region Insert Cells Down

        [TestMethod]
        public void InsertCellsDownEmptyUsedCellRange()
        {
            CellRange range = new CellRange(0, 0, 0, 0);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Insert(InsertShiftType.Down);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        [TestMethod]
        public void InsertCellsDownWholeUsedCellRange()
        {
            CellRange range = new CellRange(0, 0, 0, 1);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Insert(InsertShiftType.Down);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 0, 1, 1));
        }

        [TestMethod]
        public void InsertCellsDownPartOfUsedCellRange()
        {
            CellRange range = new CellRange(0, 0, 0, 1);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[5, 5].SetValue(3);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Insert(InsertShiftType.Down);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 0, 6, 1));
        }

        [TestMethod]
        public void InsertCellsDownExpandUsedCellRange()
        {
            CellRange range = new CellRange(0, 0, 0, 1);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[5, 0].SetValue(4);
            this.worksheet.Cells[5, 5].SetValue(3);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Insert(InsertShiftType.Down);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 0, 6, 1));
        }

        [TestMethod]
        public void InsertCellsDownOutOfUsedCellRange()
        {
            CellRange range = new CellRange(6, 0, 8, 1);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[5, 0].SetValue(4);
            this.worksheet.Cells[5, 5].SetValue(3);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Insert(InsertShiftType.Down);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        #endregion


        #region Undo Insert Cells Down

        [TestMethod]
        public void UndoInsertCellsDownEmptyUsedCellRange()
        {
            CellRange range = new CellRange(0, 0, 0, 0);
            this.worksheet.Cells[range].Insert(InsertShiftType.Down);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        [TestMethod]
        public void UndoInsertCellsDownWholeUsedCellRange()
        {
            CellRange range = new CellRange(0, 0, 0, 1);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[range].Insert(InsertShiftType.Down);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 0, 1, 1));
        }

        [TestMethod]
        public void UndoInsertCellsDownPartOfUsedCellRange()
        {
            CellRange range = new CellRange(0, 0, 0, 1);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[5, 5].SetValue(3);
            this.worksheet.Cells[range].Insert(InsertShiftType.Down);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 0, 5, 1));
        }

        [TestMethod]
        public void UndoInsertCellsDownExpandUsedCellRange()
        {
            CellRange range = new CellRange(0, 0, 0, 1);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[5, 0].SetValue(4);
            this.worksheet.Cells[5, 5].SetValue(3);
            this.worksheet.Cells[range].Insert(InsertShiftType.Down);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 0, 6, 1));
        }


        [TestMethod]
        public void UndoInsertCellsDownOutOfUsedCellRange()
        {
            CellRange range = new CellRange(6, 0, 8, 1);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[5, 0].SetValue(4);
            this.worksheet.Cells[5, 5].SetValue(3);
            this.worksheet.Cells[range].Insert(InsertShiftType.Down);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        #endregion


        #region Insert Cells Right

        [TestMethod]
        public void InsertCellsRightEmptyUsedCellRange()
        {
            CellRange range = new CellRange(0, 0, 0, 0);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Insert(InsertShiftType.Right);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        [TestMethod]
        public void InsertCellsRightPartOfUsedCellRange()
        {
            CellRange range = new CellRange(0, 0, 1, 1);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 1].SetValue(2);
            this.worksheet.Cells[5, 5].SetValue(2);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Insert(InsertShiftType.Right);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 0, 1, 7));
        }

        [TestMethod]
        public void InsertCellsRightExpandUsedCellRange()
        {
            CellRange range = new CellRange(0, 4, 0, 5);
            this.worksheet.Cells[0, 4].SetValue(1);
            this.worksheet.Cells[0, 5].SetValue(2);
            this.worksheet.Cells[5, 5].SetValue(2);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Insert(InsertShiftType.Right);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 4, 0, 7));
        }

        [TestMethod]
        public void InsertCellsRightWholeUsedCellRange()
        {
            CellRange range = new CellRange(1, 1, 2, 2);
            this.worksheet.Cells[1, 1].SetValue(1);
            this.worksheet.Cells[2, 2].SetValue(2);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Insert(InsertShiftType.Right);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(1, 1, 2, 4));
        }

        [TestMethod]
        public void InsertCellsRightOutOfUsedCellRange()
        {
            CellRange range = new CellRange(0, 3, 1, 4);
            this.worksheet.Cells[1, 1].SetValue(1);
            this.worksheet.Cells[2, 2].SetValue(2);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Insert(InsertShiftType.Right);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        #endregion


        #region Undo Insert Cells Right

        [TestMethod]
        public void UndoInsertCellsRightEmptyUsedCellRange()
        {
            CellRange range = new CellRange(0, 0, 0, 1);
            this.worksheet.Cells[range].Insert(InsertShiftType.Right);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        [TestMethod]
        public void UndoInsertCellsRightWholeUsedCellRange()
        {
            CellRange range = new CellRange(1, 1, 1, 2);
            this.worksheet.Cells[1, 1].SetValue(1);
            this.worksheet.Cells[1, 2].SetValue(2);
            this.worksheet.Cells[range].Insert(InsertShiftType.Right);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(1, 1, 1, 4));
        }

        [TestMethod]
        public void UndoInsertCellsRightPartOfUsedCellRange()
        {
            CellRange range = new CellRange(0, 0, 0, 1);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[5, 5].SetValue(3);
            this.worksheet.Cells[range].Insert(InsertShiftType.Right);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 0, 0, 5));
        }

        [TestMethod]
        public void UndoInsertCellsRightExpandUsedCellRange()
        {
            CellRange range = new CellRange(0, 0, 0, 1);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[0, 5].SetValue(4);
            this.worksheet.Cells[5, 5].SetValue(3);
            this.worksheet.Cells[range].Insert(InsertShiftType.Right);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 0, 0, 7));
        }

        [TestMethod]
        public void UndoInsertCellsRightOutOfUsedCellRange()
        {
            CellRange range = new CellRange(0, 6, 1, 7);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[0, 5].SetValue(4);
            this.worksheet.Cells[5, 5].SetValue(3);
            this.worksheet.Cells[range].Insert(InsertShiftType.Right);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        #endregion


        #region Remove Cells Left

        [TestMethod]
        public void RemoveCellsLeftEmptyUsedCellRange()
        {
            CellRange range = new CellRange(0, 0, 0, 0);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Remove(RemoveShiftType.Left);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        [TestMethod]
        public void RemoveCellsLeftWholeUsedCellRange()
        {
            CellRange range = new CellRange(0, 1, 0, 2);
            this.worksheet.Cells[0, 1].SetValue(1);
            this.worksheet.Cells[0, 2].SetValue(2);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Remove(RemoveShiftType.Left);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, range);
        }

        [TestMethod]
        public void RemoveCellsLeftShrinkUsedCellRange()
        {
            CellRange range = new CellRange(1, 1, 2, 2);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 1].SetValue(2);
            this.worksheet.Cells[2, 2].SetValue(3);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Remove(RemoveShiftType.Left);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, range);
        }

        [TestMethod]
        public void RemoveCellsLeftOutOfUsedCellRange()
        {
            CellRange range = new CellRange(0, 3, 1, 4);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 1].SetValue(2);
            this.worksheet.Cells[2, 2].SetValue(3);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Remove(RemoveShiftType.Left);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        #endregion


        #region Undo Remove Cells Left

        [TestMethod]
        public void UndoRemoveCellsLeftEmptyUsedCellRange()
        {
            CellRange range = new CellRange(0, 0, 0, 0);
            this.worksheet.Cells[range].Remove(RemoveShiftType.Left);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        [TestMethod]
        public void UndoRemoveCellsLeftWholeUsedCellRange()
        {
            CellRange range = new CellRange(0, 1, 0, 2);
            this.worksheet.Cells[0, 1].SetValue(1);
            this.worksheet.Cells[0, 2].SetValue(2);
            this.worksheet.Cells[range].Remove(RemoveShiftType.Left);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 1, 0, 4));
        }

        [TestMethod]
        public void UndoRemoveCellsLeftShrinkUsedCellRange()
        {
            CellRange range = new CellRange(1, 1, 2, 2);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 1].SetValue(2);
            this.worksheet.Cells[2, 2].SetValue(3);
            this.worksheet.Cells[range].Remove(RemoveShiftType.Left);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(1, 1, 2, 4));
        }

        [TestMethod]
        public void UndoRemoveCellsLeftOutOfUsedCellRange()
        {
            CellRange range = new CellRange(0, 3, 1, 4);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 1].SetValue(2);
            this.worksheet.Cells[2, 2].SetValue(3);
            this.worksheet.Cells[range].Remove(RemoveShiftType.Left);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        #endregion


        #region Remove Cells Up

        [TestMethod]
        public void RemoveCellsUpEmptyUsedCellRange()
        {
            CellRange range = new CellRange(0, 0, 0, 0);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Remove(RemoveShiftType.Up);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        [TestMethod]
        public void RemoveCellsUpWholeUsedCellRange()
        {
            CellRange range = new CellRange(0, 1, 0, 2);
            this.worksheet.Cells[0, 1].SetValue(1);
            this.worksheet.Cells[0, 2].SetValue(2);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Remove(RemoveShiftType.Up);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, range);
        }

        [TestMethod]
        public void RemoveCellsUpShrinkUsedCellRange()
        {
            CellRange range = new CellRange(1, 1, 2, 2);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 1].SetValue(2);
            this.worksheet.Cells[2, 2].SetValue(3);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Remove(RemoveShiftType.Up);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, range);
        }

        [TestMethod]
        public void RemoveCellsUpOutOfUsedCellRange()
        {
            CellRange range = new CellRange(3, 0, 4, 1);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 1].SetValue(2);
            this.worksheet.Cells[2, 2].SetValue(3);
            this.InitializeAsserters();

            this.worksheet.Cells[range].Remove(RemoveShiftType.Up);

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, true);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        #endregion


        #region Undo Remove Cells Up

        [TestMethod]
        public void UndoRemoveCellsUpEmptyUsedCellRange()
        {
            CellRange range = new CellRange(0, 0, 0, 0);
            this.worksheet.Cells[range].Remove(RemoveShiftType.Up);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        [TestMethod]
        public void UndoRemoveCellsUpWholeUsedCellRange()
        {
            CellRange range = new CellRange(1, 0, 2, 0);
            this.worksheet.Cells[1, 0].SetValue(1);
            this.worksheet.Cells[2, 0].SetValue(2);
            this.worksheet.Cells[range].Remove(RemoveShiftType.Up);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(1, 0, 4, 0));
        }

        [TestMethod]
        public void UndoRemoveCellsUpShrinkUsedCellRange()
        {
            CellRange range = new CellRange(1, 1, 2, 2);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 1].SetValue(2);
            this.worksheet.Cells[2, 2].SetValue(3);
            this.worksheet.Cells[range].Remove(RemoveShiftType.Up);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(1, 1, 4, 2));
        }

        [TestMethod]
        public void UndoRemoveCellsUpOutOfUsedCellRange()
        {
            CellRange range = new CellRange(3, 0, 4, 1);
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 1].SetValue(2);
            this.worksheet.Cells[2, 2].SetValue(3);
            this.worksheet.Cells[range].Remove(RemoveShiftType.Up);
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, range, false);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        #endregion


        #region Insert Columns

        [TestMethod]
        public void InsertColumnEmptyUsedCellRange()
        {
            int index = 0;
            this.InitializeAsserters();

            this.worksheet.Columns[index].Insert();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(0, index, SpreadsheetDefaultValues.RowCount - 1, index), false);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        [TestMethod]
        public void InsertColumnPartOfUsedCellRange()
        {
            int index = 1;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.InitializeAsserters();

            this.worksheet.Columns[index].Insert();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(0, index, SpreadsheetDefaultValues.RowCount - 1, index), false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 1, 0, 2));
        }

        [TestMethod]
        public void InsertColumnWholeUsedCellRange()
        {
            int fromIndex = 0;
            int toIndex = 1;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[1, 1].SetValue(3);
            this.InitializeAsserters();

            this.worksheet.Columns[fromIndex, toIndex].Insert();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(0, fromIndex, SpreadsheetDefaultValues.RowCount - 1, toIndex), false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 0, 1, 3));
        }

        [TestMethod]
        public void InsertColumnOutOfUsedCellRange()
        {
            int fromIndex = 2;
            int toIndex = 3;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[1, 1].SetValue(3);
            this.InitializeAsserters();

            this.worksheet.Columns[fromIndex, toIndex].Insert();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(0, fromIndex, SpreadsheetDefaultValues.RowCount - 1, toIndex), false);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        #endregion


        #region Undo Insert Columns

        [TestMethod]
        public void UndoInsertColumnEmptyUsedCellRange()
        {
            int index = 0;
            this.worksheet.Columns[index].Insert();
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(0, index, SpreadsheetDefaultValues.RowCount - 1, index), true);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        [TestMethod]
        public void UndoInsertColumnPartOfUsedCellRange()
        {
            int index = 1;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Columns[index].Insert();
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(0, index, SpreadsheetDefaultValues.RowCount - 1, index), true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 1, 0, 2));
        }

        [TestMethod]
        public void UndoInsertColumnWholeUsedCellRange()
        {
            int fromIndex = 0;
            int toIndex = 1;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[1, 1].SetValue(3);
            this.worksheet.Columns[fromIndex, toIndex].Insert();
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(0, fromIndex, SpreadsheetDefaultValues.RowCount - 1, toIndex), true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 0, 1, 3));
        }

        [TestMethod]
        public void UndoInsertColumnOutOfUsedCellRange()
        {
            int fromIndex = 2;
            int toIndex = 3;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[1, 1].SetValue(3);
            this.worksheet.Columns[fromIndex, toIndex].Insert();
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(0, fromIndex, SpreadsheetDefaultValues.RowCount - 1, toIndex), true);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        #endregion


        #region Remove Columns

        [TestMethod]
        public void RemoveColumnEmptyUsedCellRange()
        {
            int index = 0;
            this.InitializeAsserters();

            this.worksheet.Columns[index].Remove();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(0, index, SpreadsheetDefaultValues.RowCount - 1, index), true);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        [TestMethod]
        public void RemoveColumnPartOfUsedCellRange()
        {
            int index = 1;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.InitializeAsserters();

            this.worksheet.Columns[index].Remove();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(0, index, SpreadsheetDefaultValues.RowCount - 1, index), true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 1, 0, 1));
        }

        [TestMethod]
        public void RemoveColumnWholeUsedCellRange()
        {
            int fromIndex = 0;
            int toIndex = 1;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[1, 1].SetValue(3);
            this.InitializeAsserters();

            this.worksheet.Columns[fromIndex, toIndex].Remove();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(0, fromIndex, SpreadsheetDefaultValues.RowCount - 1, toIndex), true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 0, 1, 1));
        }

        [TestMethod]
        public void RemoveColumnOutOfUsedCellRange()
        {
            int fromIndex = 2;
            int toIndex = 3;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[1, 1].SetValue(3);
            this.InitializeAsserters();

            this.worksheet.Columns[fromIndex, toIndex].Remove();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(0, fromIndex, SpreadsheetDefaultValues.RowCount - 1, toIndex), true);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        #endregion


        #region Undo Remove Columns

        [TestMethod]
        public void UndoRemoveColumnEmptyUsedCellRange()
        {
            int index = 0;
            this.worksheet.Columns[index].Remove();
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(0, index, SpreadsheetDefaultValues.RowCount - 1, index), false);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        [TestMethod]
        public void UndoRemoveColumnPartOfUsedCellRange()
        {
            int index = 1;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Columns[index].Remove();
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(0, index, SpreadsheetDefaultValues.RowCount - 1, index), false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 1, 0, 2));
        }

        [TestMethod]
        public void UndoRemoveColumnWholeUsedCellRange()
        {
            int fromIndex = 0;
            int toIndex = 1;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[1, 1].SetValue(3);
            this.worksheet.Columns[fromIndex, toIndex].Remove();
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(0, fromIndex, SpreadsheetDefaultValues.RowCount - 1, toIndex), false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 0, 1, 3));
        }

        [TestMethod]
        public void UndoRemoveColumnOutOfUsedCellRange()
        {
            int fromIndex = 2;
            int toIndex = 3;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[0, 1].SetValue(2);
            this.worksheet.Cells[1, 1].SetValue(3);
            this.worksheet.Columns[fromIndex, toIndex].Remove();
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(0, fromIndex, SpreadsheetDefaultValues.RowCount - 1, toIndex), false);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        #endregion


        #region Insert Rows

        [TestMethod]
        public void InsertRowEmptyUsedCellRange()
        {
            int index = 0;
            this.InitializeAsserters();

            this.worksheet.Rows[index].Insert();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(index, 0, index, SpreadsheetDefaultValues.ColumnCount - 1), false);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        [TestMethod]
        public void InsertRowPartOfUsedCellRange()
        {
            int index = 1;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 0].SetValue(2);
            this.InitializeAsserters();

            this.worksheet.Rows[index].Insert();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(index, 0, index, SpreadsheetDefaultValues.ColumnCount - 1), false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(1, 0, 2, 0));
        }

        [TestMethod]
        public void InsertRowWholeUsedCellRange()
        {
            int fromIndex = 0;
            int toIndex = 1;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 0].SetValue(2);
            this.worksheet.Cells[1, 1].SetValue(3);
            this.InitializeAsserters();

            this.worksheet.Rows[fromIndex, toIndex].Insert();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(fromIndex, 0, toIndex, SpreadsheetDefaultValues.ColumnCount - 1), false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 0, 3, 1));
        }

        [TestMethod]
        public void InsertRowOutOfUsedCellRange()
        {
            int fromIndex = 2;
            int toIndex = 3;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 0].SetValue(2);
            this.worksheet.Cells[1, 1].SetValue(3);
            this.InitializeAsserters();

            this.worksheet.Rows[fromIndex, toIndex].Insert();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(fromIndex, 0, toIndex, SpreadsheetDefaultValues.ColumnCount - 1), false);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        #endregion


        #region Undo Insert Rows

        [TestMethod]
        public void UndoInsertRowEmptyUsedCellRange()
        {
            int index = 0;
            this.worksheet.Rows[index].Insert();
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(index, 0, index, SpreadsheetDefaultValues.ColumnCount - 1), true);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        [TestMethod]
        public void UndoInsertRowPartOfUsedCellRange()
        {
            int index = 1;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 0].SetValue(2);
            this.worksheet.Rows[index].Insert();
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(index, 0, index, SpreadsheetDefaultValues.ColumnCount - 1), true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(1, 0, 2, 0));
        }

        [TestMethod]
        public void UndoInsertRowWholeUsedCellRange()
        {
            int fromIndex = 0;
            int toIndex = 1;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 0].SetValue(2);
            this.worksheet.Cells[1, 1].SetValue(3);
            this.worksheet.Rows[fromIndex, toIndex].Insert();
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(fromIndex, 0, toIndex, SpreadsheetDefaultValues.ColumnCount - 1), true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 0, 3, 1));
        }

        [TestMethod]
        public void UndoInsertRowOutOdUsedCellRange()
        {
            int fromIndex = 2;
            int toIndex = 3;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 0].SetValue(2);
            this.worksheet.Cells[1, 1].SetValue(3);
            this.worksheet.Rows[fromIndex, toIndex].Insert();
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(fromIndex, 0, toIndex, SpreadsheetDefaultValues.ColumnCount - 1), true);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        #endregion


        #region Remove Rows

        [TestMethod]
        public void RemoveRowEmptyUsedCellRange()
        {
            int index = 0;
            this.InitializeAsserters();

            this.worksheet.Rows[index].Remove();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(index, 0, index, SpreadsheetDefaultValues.ColumnCount - 1), true);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        [TestMethod]
        public void RemoveRowPartOfUsedCellRange()
        {
            int index = 1;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 0].SetValue(2);
            this.InitializeAsserters();

            this.worksheet.Rows[index].Remove();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(index, 0, index, SpreadsheetDefaultValues.ColumnCount - 1), true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(1, 0, 1, 0));
        }

        [TestMethod]
        public void RemoveRowWholeUsedCellRange()
        {
            int fromIndex = 0;
            int toIndex = 1;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 0].SetValue(2);
            this.worksheet.Cells[1, 1].SetValue(3);
            this.InitializeAsserters();

            this.worksheet.Rows[fromIndex, toIndex].Remove();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(fromIndex, 0, toIndex, SpreadsheetDefaultValues.ColumnCount - 1), true);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 0, 1, 1));
        }

        [TestMethod]
        public void RemoveRowOutOfUsedCellRange()
        {
            int fromIndex = 2;
            int toIndex = 3;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 0].SetValue(2);
            this.worksheet.Cells[1, 1].SetValue(3);
            this.InitializeAsserters();

            this.worksheet.Rows[fromIndex, toIndex].Remove();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(fromIndex, 0, toIndex, SpreadsheetDefaultValues.ColumnCount - 1), true);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        #endregion


        #region Undo Remove Rows

        [TestMethod]
        public void UndoRemoveRowEmptyUsedCellRange()
        {
            int index = 0;
            this.worksheet.Rows[index].Remove();
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(index, 0, index, SpreadsheetDefaultValues.ColumnCount - 1), false);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        [TestMethod]
        public void UndoRemoveRowPartOfUsedCellRange()
        {
            int index = 1;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 0].SetValue(2);
            this.worksheet.Rows[index].Remove();
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(index, 0, index, SpreadsheetDefaultValues.ColumnCount - 1), false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(1, 0, 2, 0));
        }

        [TestMethod]
        public void UndoRemoveRowWholeUsedCellRange()
        {
            int fromIndex = 0;
            int toIndex = 1;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 0].SetValue(2);
            this.worksheet.Cells[1, 1].SetValue(3);
            this.worksheet.Rows[fromIndex, toIndex].Remove();
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(fromIndex, 0, toIndex, SpreadsheetDefaultValues.ColumnCount - 1), false);
            this.propertyChangedAsserter.AssertCellPropertyChanged(CellPropertyDefinitions.ValueProperty, new CellRange(0, 0, 3, 1));
        }

        [TestMethod]
        public void UndoRemoveRowOutOfUsedCellRange()
        {
            int fromIndex = 2;
            int toIndex = 3;
            this.worksheet.Cells[0, 0].SetValue(1);
            this.worksheet.Cells[1, 0].SetValue(2);
            this.worksheet.Cells[1, 1].SetValue(3);
            this.worksheet.Rows[fromIndex, toIndex].Remove();
            this.InitializeAsserters();

            this.worksheet.Workbook.History.Undo();

            this.cellsInsertRemoveAsserter.AssertCellsInsertedRemoved(1, new CellRange(fromIndex, 0, toIndex, SpreadsheetDefaultValues.ColumnCount - 1), false);
            this.propertyChangedAsserter.AssertCellPropertyChangedNotFired();
        }

        #endregion
    }
}
