﻿using System;
using System.Windows;
using Telerik.Windows.Documents.Spreadsheet.Layout;
using Telerik.Windows.Documents.Spreadsheet.Layout.Layers;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.TestsInternal.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal.LayersTests
{
    public abstract class LayerTesterBase
    {
        protected static readonly double RowHeight = SpreadsheetDefaultValues.DefaultRowHeight;
        protected static readonly double ColumnWidth = SpreadsheetDefaultValues.DefaultColumnWidth;

        protected bool IsInitialized { get; private set; }
        protected double ScaleFactor { get; private set; }
        protected Point ScrollOffset { get; private set; }

        private Size ScaleSize
        {
            get
            {
                return new Size(this.ScaleFactor, this.ScaleFactor);
            }
        }

        protected void Initialize(Point scrollOffset, double scaleFactor)
        {
            this.ScrollOffset = scrollOffset;
            this.ScaleFactor = scaleFactor;

            this.IsInitialized = true;
        }

        protected void TestLayer(Worksheet worksheet, Size scaledViewportSize, Action<LayerWrapper> assertLayerUpdate, CellIndex frozenCellIndex = null)
        {
            this.OnBeforeTest();
            
            RadWorksheetLayout layout = worksheet.Workbook.GetWorksheetLayout(worksheet, true);

            SheetViewport sheetViewport = null;

            if (frozenCellIndex == null)
            {
                sheetViewport = SheetViewportFactory.CreateSheetViewport(layout, scaledViewportSize, this.ScrollOffset, this.ScaleSize);
            }
            else
            {
                Point frozenPoint = layout.GetTopLeftPointFromCellIndex(frozenCellIndex.RowIndex, frozenCellIndex.ColumnIndex, true);
                SizeI visibleSize = SpreadsheetDefaultValues.VisibleSize;
                Point fixedTopLeft = new Point(0, 0);
                Point scrollableTopLeft = frozenPoint.Translate(this.ScrollOffset);
                Point maxBottomRight = layout.GetMaxBottomRight(visibleSize);

                sheetViewport = new SheetViewport(layout, scaledViewportSize, this.ScaleSize, fixedTopLeft, scrollableTopLeft, frozenCellIndex, maxBottomRight);
            }                
                      
            WorksheetRenderUpdateContext updateContext = new WorksheetRenderUpdateContext(layout, sheetViewport, this.ScaleSize);

            LayerWrapper layer = this.CreateTestLayer();
            layer.UpdateRender(updateContext);

            assertLayerUpdate(layer);
        }

        protected abstract LayerWrapper CreateTestLayer();

        protected Rect TranslateAndScale(Rect rect, ViewportPaneType paneType)
        {
            double horizontalScroll = (paneType == ViewportPaneType.HorizontalScrollable || paneType == ViewportPaneType.Scrollable) ? (-this.ScrollOffset.X * this.ScaleFactor) : 0;
            double verticalScroll = (paneType == ViewportPaneType.VerticalScrollable || paneType == ViewportPaneType.Scrollable) ? (-this.ScrollOffset.Y * this.ScaleFactor) : 0;

            return rect.Scale(this.ScaleFactor).Translate(horizontalScroll, verticalScroll);
        }

        protected Point TranslateAndScale(Point point, ViewportPaneType paneType)
        {
            double horizontalScroll = -this.ScaleFactor * this.GetHorizontalScroll(paneType);
            double verticalScroll = -this.ScaleFactor * this.GetVerticalScroll(paneType);

            return point.Scale(this.ScaleFactor).Translate(horizontalScroll, verticalScroll);
        }

        protected double GetHorizontalScroll(ViewportPaneType paneType)
        {
            return (paneType == ViewportPaneType.HorizontalScrollable || paneType == ViewportPaneType.Scrollable) ? this.ScrollOffset.X : 0;
        }

        protected double GetVerticalScroll(ViewportPaneType paneType)
        {
            return (paneType == ViewportPaneType.VerticalScrollable || paneType == ViewportPaneType.Scrollable) ? this.ScrollOffset.Y : 0;
        }

        private void OnBeforeTest()
        {
            if (!this.IsInitialized)
            {
                throw new InvalidOperationException("Before testing you must call Initialize method!");
            }

            this.IsInitialized = false;
        }
    }
}
