﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Media;
using Telerik.Windows.Documents.Spreadsheet.Layout;
using Telerik.Windows.Documents.Spreadsheet.Layout.Layers;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;
using Telerik.Windows.Documents.Spreadsheet.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal.LayersTests
{
    internal static class RenderableComparer
    {
        private const double Epsilon = 1E-4;

        public static bool AreEqual(ShapeRenderable first, ShapeRenderable second)
        {
            Guard.ThrowExceptionIfNull(first, "first");
            Guard.ThrowExceptionIfNull(second, "second");

            bool result = first.GetType() == second.GetType() &&
                AreEqual(first.TopLeft, second.TopLeft) &&
                first.Width.Equals(second.Width) &&
                first.Height.Equals(second.Height);

            if (first is ImageRenderable)
            {
                ImageRenderable firstImage = first as ImageRenderable;
                ImageRenderable secondImage = second as ImageRenderable;

                result &= firstImage.Shape.IsHorizontallyFlipped.Equals(secondImage.Shape.IsHorizontallyFlipped) &&
                firstImage.Shape.IsVerticallyFlipped.Equals(secondImage.Shape.IsVerticallyFlipped) &&
                firstImage.Shape.RotationAngle.Equals(secondImage.Shape.RotationAngle);
            }

#if !SILVERLIGHT
            if(first is ChartRenderable)
            {
                ChartRenderable firstChart = first as ChartRenderable;
                ChartRenderable secondChart = second as ChartRenderable;

                result &= firstChart.Shape.Chart == secondChart.Shape.Chart;
            }
#endif
            return result;
        }

        public static bool AreEqual(LineRenderable first, LineRenderable second)
        {
            Guard.ThrowExceptionIfNull(first, "first");
            Guard.ThrowExceptionIfNull(second, "second");

            return AreEqual(first.X1, second.X1) &&
                AreEqual(first.X2, second.X2) &&
                AreEqual(first.Y1, second.Y1) &&
                AreEqual(first.Y2, second.Y2) &&
                first.Stroke.Equals(second.Stroke) &&
                first.StrokeThickness.Equals(second.StrokeThickness) &&
                first.CellBorderStyle.Equals(second.CellBorderStyle);
        }

        public static bool AreEqual<T>(RowColumnHeadingRenderable<T> first, RowColumnHeadingRenderable<T> second)
            where T : LayoutBox
        {
            Guard.ThrowExceptionIfNull(first, "first");
            Guard.ThrowExceptionIfNull(second, "second");

            return AreEqual(first.LayoutBox.BoundingRectangle, second.LayoutBox.BoundingRectangle) &&
                AreEqual(first.Rectangle, second.Rectangle) &&
                first.Text.Equals(second.Text) &&
                first.ScaleFactor.Equals(second.ScaleFactor);
        }

        public static bool AreEqual(GradientFillRenderable first, GradientFillRenderable second)
        {
            Guard.ThrowExceptionIfNull(first, "first");
            Guard.ThrowExceptionIfNull(second, "second");

            return AreEqual(first.BoundingRectangle, second.BoundingRectangle) &&
                first.GradientFill.Equals(second.GradientFill);
        }

        public static bool AreEqual(PatternFillRenderable first, PatternFillRenderable second)
        {
            Guard.ThrowExceptionIfNull(first, "first");
            Guard.ThrowExceptionIfNull(second, "second");

            return AreEqual(first.VisibleBoxes, second.VisibleBoxes) &&
                first.PatternFill.Equals(second.PatternFill);
        }

        public static bool AreEqual(TextBlockRenderable first, TextBlockRenderable second)
        {
            Guard.ThrowExceptionIfNull(first, "first");
            Guard.ThrowExceptionIfNull(second, "second");

            bool cellBoxEquality = AreEqual(first.CellBox, second.CellBox);
            bool textEquality = TelerikHelper.EqualsOfT(first.Text, second.Text);
            bool runsEquality = AreEqual(first.Runs, second.Runs);
            bool fontSizeEquality = AreEqual(first.FontSize, second.FontSize);
            bool fontFamilyEquality = first.FontFamily.Equals(second.FontFamily);
            bool fontWeightEquality = first.FontWeight.Equals(second.FontWeight);
            bool fontStyleEquality = first.FontStyle.Equals(second.FontStyle);
            bool textDecorationsEquality = TelerikHelper.EqualsOfT(first.TextDecorations, second.TextDecorations);
            bool foregroundEquality = first.Foreground.Equals(second.Foreground);
            bool textAlignmentEquality = first.TextAlignment.Equals(second.TextAlignment);
            bool wrappingEquality = first.TextWrapping.Equals(second.TextWrapping);
            bool scaleEquality = AreEqual(first.ScaleFactor, second.ScaleFactor);
            bool formatEquality = first.CellValueFormat.Equals(second.CellValueFormat);

            return cellBoxEquality &&
                textEquality &&
                runsEquality &&
                fontSizeEquality &&
                fontFamilyEquality &&
                fontWeightEquality &&
                fontStyleEquality &&
                textDecorationsEquality &&
                foregroundEquality &&
                textAlignmentEquality &&
                wrappingEquality &&
                scaleEquality &&
                formatEquality;
        }

        private static bool AreEqual(IEnumerable<RunRenderable> firstCollection, IEnumerable<RunRenderable> secondCollection)
        {
            RunRenderable[] firstArray = firstCollection.ToArray();
            RunRenderable[] secondArray = secondCollection.ToArray();

            if (firstArray.Length != secondArray.Length)
            {
                return false;
            }

            for (int i = 0; i < firstArray.Length; i++)
            {
                RunRenderable first = firstArray[i];
                RunRenderable second = secondArray[i];

                if (!AreEqual(first, second))
                {
                    return false;
                }
            }

            return true;
        }

        private static bool AreEqual(RunRenderable first, RunRenderable second)
        {
            Color? firstColor = first.Foreground != null ? (Color?)first.Foreground.Color : null;
            Color? secondColor = second.Foreground != null ? (Color?)second.Foreground.Color : null;

            bool textEquality = TelerikHelper.EqualsOfT(first.Text, second.Text);
            bool colorEquality = TelerikHelper.EqualsOfT(firstColor, secondColor);
            bool fontFamilyEquality = TelerikHelper.EqualsOfT(first.FontFamily, second.FontFamily);
            bool fontSizeEquality = AreEqual(first.FontSize, second.FontSize);

            return textEquality &&
                colorEquality &&
                fontFamilyEquality &&
                fontSizeEquality;
        }

        private static bool AreEqual(CellLayoutBox first, CellLayoutBox second)
        {
            return first.LongIndex.Equals(second.LongIndex) &&
                AreEqual(first.BoundingRectangle, second.BoundingRectangle);
        }

        private static bool AreEqual(double? x, double? y)
        {
            if (x.HasValue && y.HasValue)
            {
                return AreEqual(x.Value, y.Value);
            }
            else
            {
                return TelerikHelper.EqualsOfT(x, y);
            }
        }

        private static bool AreEqual(double x, double y)
        {
            return TestHelper.AreDoubleValuesEssentiallyEqual(x, y, Epsilon);
        }

        private static bool AreEqual(Point a, Point b)
        {
            return AreEqual(a.X, b.X) && AreEqual(a.Y, b.Y);
        }

        private static bool AreEqual(Size a, Size b)
        {
            return AreEqual(a.Width, b.Width) && AreEqual(a.Height, b.Height);
        }

        private static bool AreEqual(Rect a, Rect b)
        {
            return AreEqual(a.Left, b.Left) && AreEqual(a.Top, b.Top) && AreEqual(a.Right, b.Right) && AreEqual(a.Bottom, b.Bottom);
        }

        private static bool AreEqual(IEnumerable<Rect> a, IEnumerable<Rect> b)
        {
            Rect[] first = a.OrderBy(r => r.Left).ThenBy(r => r.Top).ThenBy(r => r.Width).ThenBy(r => r.Height).ToArray();
            Rect[] second = b.OrderBy(r => r.Left).ThenBy(r => r.Top).ThenBy(r => r.Width).ThenBy(r => r.Height).ToArray();

            if (first.Length != second.Length)
            {
                return false;
            }

            for (int i = 0; i < first.Length; i++)
            {
                if (!AreEqual(first[i], second[i]))
                {
                    return false;
                }
            }

            return true;
        }
    }
}
