﻿using System;
using System.Windows.Media;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Maths;

#if SILVERLIGHT
using Microsoft.Silverlight.Testing;
#endif

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal.Maths
{
    /// <summary>
    /// Unit tests of diverse algebraic operations.
    /// </summary>
    [TestClass]
    public class AlgebraTests
    {
        [TestMethod]
#if SILVERLIGHT 
        [Tag("Algebra")]
#else
        [TestCategory("Algebra")]
#endif
        public void LUDecompositionTest()
        {
            var m = new RMatrix(new[,] { { 1d, 2d }, { 3d, 4d } });
            var lum = AlgebraExtensions.LUDecompose(m);
            var expect = new RMatrix(new[,] { { 1d, 2d }, { 3d, -2d } });
            Assert.IsTrue(expect.Equals(lum));

            //    m = new RMatrix(new double[,] { { 2, 4, -6 }, { 6, -4, 2 }, { 4, 2, 6 } });
            //    lum = AlgebraExtensions.LUDecompose(m);
            //    expect = new RMatrix(
            //        new[,] {  
            //        { 2, 4, -6 },
            //        { 2, -6, 18 }, 
            //        { 3,2.66667, -28 } 
            //        }
            //);
            //    Assert.AreEqual(expect, lum);

            //m = new RMatrix(new double[,] { {2, 4 }, {6, -4 } });
            //lum = AlgebraExtensions.LUDecompose(m);
            //expect = new RMatrix(new double[,] { {2, 4 }, { 3, -16 } });
            //Assert.IsTrue(expect.Equals(lum));
        }

        private const double Accuracy = 1E-6;
        [TestMethod]
#if SILVERLIGHT 
        [Tag("Algebra")]
#else
        [TestCategory("Algebra")]
#endif
        public void LinearSolveTest()
        {
            var A = new RMatrix(3, 3);
            A[0, 0] = 2;
            A[0, 1] = 1;
            A[0, 2] = -1;
            A[1, 0] = -3;
            A[1, 1] = -1;
            A[1, 2] = 2;
            A[2, 0] = -2;
            A[2, 1] = 1;
            A[2, 2] = 2;
            var b = new RVector(3);
            b[0] = 8;
            b[1] = -11;
            b[2] = -3;

            var x = AlgebraExtensions.GaussJordan(A, b);
            Assert.AreEqual(2, x[0], Accuracy);
            Assert.AreEqual(3, x[1], Accuracy);
            Assert.AreEqual(-1, x[2], Accuracy);

            A = new RMatrix(new double[,] { { 2, 4, -6 }, { 6, -4, 2 }, { 4, 2, 6 } });
            b = new RVector(new double[] { 6, -2, 4 });
            x = AlgebraExtensions.GaussJordan(A, b);

            Assert.AreEqual(0.476190476190476, x[0], Accuracy);
            Assert.AreEqual(1.19047619047619, x[1], Accuracy);
            Assert.AreEqual(-0.0476190476190476, x[2], Accuracy);

            A = new RMatrix(new double[,] { { 4, 0, 1 }, { 0, 3, 2 }, { 1, 2, 4 } });
            b = new RVector(new double[] { 2, 1, 3 });
            x = AlgebraExtensions.GaussJacobi(A, b, 10, 1.0E-4);
            Assert.AreEqual(0.310397736820174, x[0], Accuracy);
            Assert.AreEqual(-0.17227270181287, x[1], Accuracy);
            Assert.AreEqual(0.751248669722443, x[2], Accuracy);

            // Gauss-Jordan is less precise
            x = AlgebraExtensions.GaussJordan(A, b);
            Assert.AreEqual(0.310397736820174, x[0], 0.1);
            Assert.AreEqual(-0.17227270181287, x[1], 0.1);
            Assert.AreEqual(0.751248669722443, x[2], 0.1);

            x = AlgebraExtensions.GaussSeidel(A, b);
            Assert.AreEqual(0.310397736820174, x[0], 0.1);
            Assert.AreEqual(-0.17227270181287, x[1], 0.1);
            Assert.AreEqual(0.751248669722443, x[2], 0.1);

        }

        [TestMethod]
        public void ScaleAtMatrixTest()
        {
            Matrix m = new Matrix(1, 2, 3, 4, 5, 6);

            Matrix actual = m.ScaleMatrixAt(10, 3, 5, 7);
            Matrix expected = new Matrix(10, 6, 30, 12, 5, 4);

            Assert.AreEqual(expected, actual);
            Assert.AreEqual(new Matrix(1, 2, 3, 4, 5, 6), m);
        }

        [TestMethod]
        public void ScaleMatrixTest()
        {
            Matrix m = new Matrix(1, 2, 3, 4, 5, 6);

            Matrix actual = m.ScaleMatrix(10, 3);
            Matrix expected = new Matrix(10, 6, 30, 12, 50, 18);

            Assert.AreEqual(expected, actual);
            Assert.AreEqual(new Matrix(1, 2, 3, 4, 5, 6), m);
        }

        [TestMethod]
        public void RotateAtMatrixTest()
        {
            Matrix m = new Matrix(1, 2, 3, 4, 5, 6);

            Matrix actual = m.RotateMatrixAt(32, 5, 7);
            Matrix expected = new Matrix(-0.211790432309984, 2.22601545654606, 0.424467231536458, 4.98195017732532, 5.52991926423321, 6.15195190384357);

            double tolerance = 1E-14;
            Assert.IsTrue(AreAlmostEqual(expected, actual, tolerance));
            Assert.AreEqual(new Matrix(1, 2, 3, 4, 5, 6), m);
        }

        [TestMethod]
        public void RotateMatrixTest()
        {
            Matrix m = new Matrix(1, 2, 3, 4, 5, 6);

            Matrix actual = m.RotateMatrix(32);
            Matrix expected = new Matrix(-0.211790432309984, 2.22601545654606, 0.424467231536458, 4.98195017732532, 1.0607248953829, 7.73788489810458);

            double tolerance = 1E-14;
            Assert.IsTrue(AreAlmostEqual(expected, actual, tolerance));
            Assert.AreEqual(new Matrix(1, 2, 3, 4, 5, 6), m);
        }

        [TestMethod]
        public void TranslateMatrixTest()
        {
            Matrix m = new Matrix(1, 2, 3, 4, 5, 6);

            Matrix actual = m.TranslateMatrix(4, 5);
            Matrix expected = new Matrix(1, 2, 3, 4, 9, 11);

            Assert.AreEqual(expected, actual);
            Assert.AreEqual(new Matrix(1, 2, 3, 4, 5, 6), m);
        }

        [TestMethod]
        public void MultiplyByMatrixTest()
        {
            Matrix m = new Matrix(1, 2, 3, 4, 5, 6);

            Matrix actual = m.MultiplyBy(m);
            Matrix expected = new Matrix(7, 10, 15, 22, 28, 40);

            Assert.AreEqual(expected, actual);
            Assert.AreEqual(new Matrix(1, 2, 3, 4, 5, 6), m);
        }

        [TestMethod]
        public void InverseMatrixTest()
        {
            Matrix m = new Matrix(1, 2, 3, 4, 5, 6);

            Matrix actual = m.InverseMatrix();
            Matrix expected = new Matrix(-2, 1, 1.5, -0.5, 1, -2);

            Assert.AreEqual(expected, actual);
            Assert.AreEqual(new Matrix(1, 2, 3, 4, 5, 6), m);
        }

        private static bool AreAlmostEqual(Matrix first, Matrix second, double epsilon)
        {
            return AreAlmostEqual(first.M11, second.M11, epsilon) &&
                AreAlmostEqual(first.M12, second.M12, epsilon) &&
                AreAlmostEqual(first.M21, second.M21, epsilon) &&
                AreAlmostEqual(first.M22, second.M22, epsilon) &&
                AreAlmostEqual(first.OffsetX, second.OffsetX, epsilon) &&
                AreAlmostEqual(first.OffsetY, second.OffsetY, epsilon);
        }

        private static bool AreAlmostEqual(double first, double second, double epsilon)
        {
            return System.Math.Abs(first - second) <= epsilon;
        }
    }
}