﻿using System;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Model;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal.Protection
{
    [TestClass]
    public class IsLockedPropertyTests
    {
        private static Workbook workbook;

        private static Workbook Workbook
        {
            get
            {
                return workbook;
            }
        }

        [TestInitialize]
        public void TestInitialize()
        {
            workbook = new Workbook();
            workbook.Worksheets.Add();

            workbook.Styles.Add("Protected", PropertySystem.CellStyleCategory.Custom, true);
            workbook.Styles["Protected"].IsLocked = true;
            workbook.Styles["Protected"].IncludeProtection = true;

            workbook.Styles.Add("Unprotected", PropertySystem.CellStyleCategory.Custom, true);
            workbook.Styles["Unprotected"].IsLocked = false;
            workbook.Styles["Unprotected"].IncludeProtection = true;
        }

        [TestCleanup]
        public void TestCleanup()
        {
            workbook = null;
        }

        [TestMethod]
        public void SetLocalValueToTrueProtectApplyStyleWithValueFalse()
        {
            // Arrange
            Worksheet worksheet = Workbook.ActiveWorksheet;
            CellSelection cellSelection = worksheet.Cells[0, 0];
            bool localValue = true;

            //Act
            cellSelection.SetIsLocked(localValue);
            worksheet.Protect(null, null);
            cellSelection.SetStyleName("Unprotected");

            //Assert
            bool isLocked = cellSelection.GetIsLocked().Value;
            Assert.AreEqual(localValue, isLocked);
        }

        [TestMethod]
        public void SetLocalValueToFalseProtectApplyStyleWithValueTrue()
        {
            // Arrange
            Worksheet worksheet = Workbook.ActiveWorksheet;
            CellSelection cellSelection = worksheet.Cells[0, 0];
            bool localValue = false;

            //Act
            cellSelection.SetIsLocked(localValue);
            worksheet.Protect(null, null);
            cellSelection.SetStyleName("Protected");

            //Assert
            bool isLocked = cellSelection.GetIsLocked().Value;
            Assert.AreEqual(localValue, isLocked);
        }

        [TestMethod]
        public void SetStyleWithValueTrueProtectApplyStyleWithValueFalse()
        {
            // Arrange
            Worksheet worksheet = Workbook.ActiveWorksheet;
            CellSelection cellSelection = worksheet.Cells[0, 0];

            //Act
            cellSelection.SetStyleName("Protected");
            worksheet.Protect(null, null);
            cellSelection.SetStyleName("Unprotected");

            //Assert
            bool isLocked = cellSelection.GetIsLocked().Value;
            Assert.AreEqual(true, isLocked);
        }

        [TestMethod]
        public void SetStyleWithValueFalseProtectApplyStyleWithValueTrue()
        {
            // Arrange
            Worksheet worksheet = Workbook.ActiveWorksheet;
            CellSelection cellSelection = worksheet.Cells[0, 0];

            //Act
            cellSelection.SetStyleName("Unprotected");
            worksheet.Protect(null, null);
            cellSelection.SetStyleName("Protected");

            //Assert
            bool isLocked = cellSelection.GetIsLocked().Value;
            Assert.AreEqual(false, isLocked);
        }

        [TestMethod]
        public void CopyValueStyleWithTrueAndPasteOnValueStyleWithFalse()
        {
            Worksheet worksheet = Workbook.ActiveWorksheet;
            CellSelection A1 = worksheet.Cells[0, 0];
            CellSelection B1 = worksheet.Cells[0, 1];
            A1.SetValue(1);
            A1.SetStyleName("Unprotected");
            B1.SetValue(2);
            B1.SetStyleName("Protected");

            WorksheetFragment fragment = B1.Copy();
            A1.Paste(fragment, PasteOptions.All);

            Assert.AreEqual(true, A1.GetIsLocked().Value);
        }

        [TestMethod]
        public void CopyValueStyleWithTrueProtectAndPasteOnValueStyleWithFalse()
        {
            Worksheet worksheet = Workbook.ActiveWorksheet;
            CellSelection A1 = worksheet.Cells[0, 0];
            CellSelection B1 = worksheet.Cells[0, 1];
            A1.SetValue(1);
            A1.SetStyleName("Unprotected");
            B1.SetValue(2);
            B1.SetStyleName("Protected");

            worksheet.Protect(null, null);
            WorksheetFragment fragment = B1.Copy();
            A1.Paste(fragment, PasteOptions.All);

            Assert.AreEqual(false, A1.GetIsLocked().Value);
        }

        [TestMethod]
        public void CopyValueStyleWithTrueProtectAndPasteOnValueStylesWithFalseAndTrue()
        {
            Worksheet worksheet = Workbook.ActiveWorksheet;
            CellSelection A1 = worksheet.Cells[0, 0];
            CellSelection B1 = worksheet.Cells[0, 1];
            CellSelection C1 = worksheet.Cells[0, 2];
            
            A1.SetStyleName("Unprotected");
            B1.SetStyleName("Protected");
            C1.SetStyleName("Protected");

            worksheet.Protect(null, null);
            WorksheetFragment fragment = B1.Copy();
            worksheet.Cells[0, 0, 0, 1].Paste(fragment, PasteOptions.All);

            Assert.AreEqual(false, A1.GetIsLocked().Value);
            Assert.AreEqual(true, B1.GetIsLocked().Value);
        }

        [TestMethod]
        public void CopyValueStyleWithFalseProtectAndPasteOnValueStylesWithFalseAndTrue()
        {
            Worksheet worksheet = Workbook.ActiveWorksheet;
            CellSelection A1 = worksheet.Cells[0, 0];
            CellSelection B1 = worksheet.Cells[0, 1];
            CellSelection C1 = worksheet.Cells[0, 2];

            A1.SetStyleName("Unprotected");
            B1.SetStyleName("Protected");
            C1.SetStyleName("Unprotected");

            worksheet.Protect(null, null);
            WorksheetFragment fragment = B1.Copy();
            worksheet.Cells[0, 0, 0, 1].Paste(fragment, PasteOptions.All);

            Assert.AreEqual(false, A1.GetIsLocked().Value);
            Assert.AreEqual(true, B1.GetIsLocked().Value);
        }
    }
}
