﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Media;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Shapes;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;
using Telerik.Windows.Documents.Spreadsheet.TestsInternal.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal.Shapes
{
    [TestClass]
    public class ShapeCollectionTests
    {
        private int collectionChangedCounter;

        private static readonly CellIndex standardCellIndex = new CellIndex(0, 0);
        private static readonly double width = 90;
        private static readonly double height = 90;

        [TestMethod]
        public void AddShapeToCollectionTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image = this.CreateImage(worksheet);

            worksheet.Shapes.Add(image);

            Assert.AreEqual(0, image.Id);
            Assert.AreEqual(1, worksheet.Shapes.Count);
            Assert.AreEqual("Picture 0", image.Name);
        }

        [TestMethod]
        public void AddShapeWithNameToCollectionTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image = this.CreateImage(worksheet);
            image.Name = "TestName";

            worksheet.Shapes.Add(image);

            Assert.AreEqual(0, image.Id);
            Assert.AreEqual(1, worksheet.Shapes.Count);
            Assert.AreEqual("TestName", image.Name);
        }

        [TestMethod]
        public void AddMultipleShapesTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);

            this.collectionChangedCounter = 0;
            worksheet.Shapes.Changed += Shapes_Changed;

            worksheet.Shapes.Add(new List<FloatingShapeBase> { image1, image2 });

            Assert.AreEqual(1, this.collectionChangedCounter);
            Assert.AreEqual(0, image1.Id);
            Assert.AreEqual(1, image2.Id);
            Assert.AreEqual(2, worksheet.Shapes.Count);
            Assert.AreEqual("Picture 0", image1.Name);
            Assert.AreEqual("Picture 1", image2.Name);
        }

        [TestMethod]
        [ExpectedException(typeof(InvalidOperationException))]
        public void AddShapeWithDifferentWorksheet()
        {
            Worksheet worksheet1 = TestHelper.CreateWorksheet();
            Worksheet worksheet2 = TestHelper.CreateWorksheet();
            FloatingShapeBase image = this.CreateImage(worksheet1);

            worksheet2.Shapes.Add(image);
        }

        [TestMethod]
        public void RemoveShapeFromCollectionTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(new List<FloatingShapeBase> { image1, image2 });

            worksheet.Shapes.Remove(image1);

            Assert.AreEqual(1, worksheet.Shapes.Count);
            Assert.AreEqual(image2.Id, worksheet.Shapes.First().Id);
        }

        [TestMethod]
        public void RemoveMultipleShapeFromCollectionTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(new List<FloatingShapeBase> { image1, image2 });

            this.collectionChangedCounter = 0;
            worksheet.Shapes.Changed += Shapes_Changed;

            worksheet.Shapes.Remove(new List<FloatingShapeBase> { image1, image2 });

            Assert.AreEqual(0, worksheet.Shapes.Count);
            Assert.AreEqual(1, this.collectionChangedCounter);
        }

        [TestMethod]
        public void RemoveNonexistingShapeTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);

            this.collectionChangedCounter = 0;
            worksheet.Shapes.Changed += Shapes_Changed;

            worksheet.Shapes.Remove(image2);

            Assert.AreEqual(1, worksheet.Shapes.Count);
            Assert.AreEqual(0, this.collectionChangedCounter);
        }

        [TestMethod]
        public void ClearShapeCollectionTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(new List<FloatingShapeBase> { image1, image2 });

            this.collectionChangedCounter = 0;
            worksheet.Shapes.Changed += Shapes_Changed;

            worksheet.Shapes.Clear();

            Assert.AreEqual(0, worksheet.Shapes.Count);
            Assert.AreEqual(1, this.collectionChangedCounter);
        }

        [TestMethod]
        public void IndexerGetTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(new List<FloatingShapeBase> { image1, image2 });

            FloatingShapeBase retrievedImage = worksheet.Shapes[1];

            Assert.AreEqual(image2, retrievedImage);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void IndexerGetTestOutOfRange()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(new List<FloatingShapeBase> { image1, image2 });

            FloatingShapeBase retrievedImage = worksheet.Shapes[2];
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void IndexerGetTestOutOfRange2()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();

            FloatingShapeBase retrievedImage = worksheet.Shapes[0];
        }

        [TestMethod]
        public void IndexerSetTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(new List<FloatingShapeBase> { image1, image2 });

            this.collectionChangedCounter = 0;
            worksheet.Shapes.Changed += Shapes_Changed;

            FloatingShapeBase image3 = this.CreateImage(worksheet);
            worksheet.Shapes[1] = image3;

            Assert.AreEqual(1, collectionChangedCounter);
            Assert.AreEqual(image1, worksheet.Shapes[0]);
            Assert.AreEqual(image3, worksheet.Shapes[1]);
            Assert.AreEqual(2, worksheet.Shapes.Count);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void IndexerSetTestNull()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(new List<FloatingShapeBase> { image1, image2 });

            worksheet.Shapes[1] = null;
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void IndexerSetTestOutOfRange()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(new List<FloatingShapeBase> { image1, image2 });

            FloatingShapeBase image3 = this.CreateImage(worksheet);
            worksheet.Shapes[2] = image3;
        }

        [TestMethod]
        public void BringForwardOneImageTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);

            worksheet.Shapes.BringForward(new List<FloatingShapeBase> { image1 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image2));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image3));
        }

        [TestMethod]
        public void BringForwardTopMostImageTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);

            worksheet.Shapes.BringForward(new List<FloatingShapeBase> { image3 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image2));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image3));
        }

        [TestMethod]
        public void BringForwardMultipleImagesTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);

            worksheet.Shapes.BringForward(new List<FloatingShapeBase> { image1, image2 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image3));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image2));
        }

        [TestMethod]
        public void BringForwardMultipleTopMostImagesTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            FloatingShapeBase image4 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);
            worksheet.Shapes.Add(image4);

            worksheet.Shapes.BringForward(new List<FloatingShapeBase> { image2, image3, image4 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image4));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image2));
            Assert.AreEqual(3, worksheet.Shapes.GetZIndex(image3));
        }

        [TestMethod]
        public void BringForwardMultipleNonConsequentialImagesTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            FloatingShapeBase image4 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);
            worksheet.Shapes.Add(image4);

            worksheet.Shapes.BringForward(new List<FloatingShapeBase> { image2, image4 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image3));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image2));
            Assert.AreEqual(3, worksheet.Shapes.GetZIndex(image4));
        }

        public void BringToFrontOneImageTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);

            worksheet.Shapes.BringToFront(new List<FloatingShapeBase> { image1 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image2));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image3));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image1));
        }

        [TestMethod]
        public void BringToFrontTopMostImageTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);

            worksheet.Shapes.BringToFront(new List<FloatingShapeBase> { image3 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image2));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image3));
        }

        [TestMethod]
        public void BringToFrontMultipleImagesTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            FloatingShapeBase image4 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);
            worksheet.Shapes.Add(image4);

            worksheet.Shapes.BringToFront(new List<FloatingShapeBase> { image1, image2 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image3));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image4));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(3, worksheet.Shapes.GetZIndex(image2));
        }

        [TestMethod]
        public void BringToFrontMultipleTopMostImagesTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            FloatingShapeBase image4 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);
            worksheet.Shapes.Add(image4);

            worksheet.Shapes.BringToFront(new List<FloatingShapeBase> { image2, image3, image4 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image2));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image3));
            Assert.AreEqual(3, worksheet.Shapes.GetZIndex(image4));
        }

        [TestMethod]
        public void BringToFrontMultipleNonConsequentialImagesTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            FloatingShapeBase image4 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);
            worksheet.Shapes.Add(image4);

            worksheet.Shapes.BringToFront(new List<FloatingShapeBase> { image1, image4 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image2));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image3));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(3, worksheet.Shapes.GetZIndex(image4));
        }

        [TestMethod]
        public void SendBackwardOneImageTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);

            worksheet.Shapes.SendBackward(new List<FloatingShapeBase> { image3 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image3));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image2));
        }

        [TestMethod]
        public void SendBackwardBottomMostImageTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);

            worksheet.Shapes.SendBackward(new List<FloatingShapeBase> { image1 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image2));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image3));
        }

        [TestMethod]
        public void SendBackwardMultipleImagesTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);

            worksheet.Shapes.SendBackward(new List<FloatingShapeBase> { image2, image3 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image2));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image3));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image1));
        }

        [TestMethod]
        public void SendBackwardMultipleBottomMostImagesTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            FloatingShapeBase image4 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);
            worksheet.Shapes.Add(image4);

            worksheet.Shapes.SendBackward(new List<FloatingShapeBase> { image1, image2, image3 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image2));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image3));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(3, worksheet.Shapes.GetZIndex(image4));
        }

        [TestMethod]
        public void SendBackwardMultipleNonConsequentialImagesTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            FloatingShapeBase image4 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);
            worksheet.Shapes.Add(image4);

            worksheet.Shapes.SendBackward(new List<FloatingShapeBase> { image1, image4 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image2));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image4));
            Assert.AreEqual(3, worksheet.Shapes.GetZIndex(image3));
        }

        [TestMethod]
        public void SendToBackOneImageTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);

            worksheet.Shapes.SendToBack(new List<FloatingShapeBase> { image3 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image3));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image2));
        }

        [TestMethod]
        public void SendToBackBottomMostImageTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);

            worksheet.Shapes.SendToBack(new List<FloatingShapeBase> { image1 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image2));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image3));
        }

        [TestMethod]
        public void SendToBackMultipleImagesTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            FloatingShapeBase image4 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);
            worksheet.Shapes.Add(image4);

            worksheet.Shapes.SendToBack(new List<FloatingShapeBase> { image3, image4 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image3));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image4));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(3, worksheet.Shapes.GetZIndex(image2));
        }

        [TestMethod]
        public void SendToBackMultipleBottomMostImagesTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            FloatingShapeBase image4 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);
            worksheet.Shapes.Add(image4);

            worksheet.Shapes.SendToBack(new List<FloatingShapeBase> { image1, image2, image3 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image2));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image3));
            Assert.AreEqual(3, worksheet.Shapes.GetZIndex(image4));
        }

        [TestMethod]
        public void SendToBackMultipleNonConsequentialImagesTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            FloatingShapeBase image2 = this.CreateImage(worksheet);
            FloatingShapeBase image3 = this.CreateImage(worksheet);
            FloatingShapeBase image4 = this.CreateImage(worksheet);
            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);
            worksheet.Shapes.Add(image3);
            worksheet.Shapes.Add(image4);

            worksheet.Shapes.SendToBack(new List<FloatingShapeBase> { image1, image3 });

            Assert.AreEqual(0, worksheet.Shapes.GetZIndex(image1));
            Assert.AreEqual(1, worksheet.Shapes.GetZIndex(image3));
            Assert.AreEqual(2, worksheet.Shapes.GetZIndex(image2));
            Assert.AreEqual(3, worksheet.Shapes.GetZIndex(image4));
        }

        [TestMethod]
        public void ShapeUsedCellRangeTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image = this.CreateImage(worksheet);

            worksheet.Shapes.Add(image);

            CellRange expectedRange = new CellRange(0, 0, 4, 1);
            Assert.AreEqual(expectedRange, worksheet.Shapes.UsedCellRange);
        }

        [TestMethod]
        public void ShapeUsedCellRangeRotatedImage()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image = this.CreateImage(worksheet);
            image.CellIndex = new CellIndex(3, 3);
            image.Width = 200;
            image.Height = 200;
            image.RotationAngle = 45;

            worksheet.Shapes.Add(image);

            CellRange expectedRange = new CellRange(0, 2, 15, 6);
            Assert.AreEqual(expectedRange, worksheet.Shapes.UsedCellRange);
        }

        [TestMethod]
        public void ShapeUsedCellRangeMultipleImages()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            image1.CellIndex = new CellIndex(3, 3);
            image1.Width = 200;
            image1.Height = 200;
            image1.RotationAngle = 45;

            FloatingShapeBase image2 = this.CreateImage(worksheet);
            image2.CellIndex = new CellIndex(15, 0);
            image2.Width = 200;
            image2.Height = 190;

            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);

            CellRange expectedRange = new CellRange(0, 0, 24, 6);
            Assert.AreEqual(expectedRange, worksheet.Shapes.UsedCellRange);
        }

        [TestMethod]
        public void ShapeUsedCellRangeResizedRowsAndColumnsTest()
        {
            Worksheet worksheet = TestHelper.CreateWorksheet();
            FloatingShapeBase image1 = this.CreateImage(worksheet);
            image1.CellIndex = new CellIndex(3, 3);
            image1.Width = 200;
            image1.Height = 200;
            image1.RotationAngle = 45;

            FloatingShapeBase image2 = this.CreateImage(worksheet);
            image2.CellIndex = new CellIndex(15, 0);
            image2.Width = 200;
            image2.Height = 190;

            worksheet.Shapes.Add(image1);
            worksheet.Shapes.Add(image2);

            worksheet.Rows[18].SetHeight(new RowHeight(100, true));
            worksheet.Columns[4].SetWidth(new ColumnWidth(150, true));

            CellRange expectedRange = new CellRange(0, 0, 20, 5);
            Assert.AreEqual(expectedRange, worksheet.Shapes.UsedCellRange);
        }

        private void Shapes_Changed(object sender, EventArgs e)
        {
            this.collectionChangedCounter++;
        }

        private FloatingImage CreateImage(Worksheet worksheet)
        {
            FloatingImage image = new FloatingImage(worksheet, standardCellIndex, 0, 0);
            Stream stream = TestHelperInternal.GetStreamFromFile("Penguins.jpg");
            using (stream)
            {
                image.ImageSource = new ImageSource(stream, "jpg");
            }
            image.Width = width;
            image.Height = height;

            return image;
        }
    }
}
