﻿using System;
using System.Linq;
using System.Windows;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Layout;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal.SheetViewportTests
{
    [TestClass]
    public class SheetViewport_FreezePanes_Should
    {
        #region Properties

        private SheetViewportTestsArgs Args { get; set; }

        private Size ViewportSize
        {
            get
            {
                return SheetViewportTestsHelper.ViewportSize;
            }
        }

        #endregion


        #region Methods

        private void AssertViewportRects(Rect expectedFixedRect, Rect expectedHorizontalScrollableRect, Rect expectedVerticalScrollableRect, Rect expectedScrollableRect)
        {
            Assert.AreEqual(expectedFixedRect, this.Args.SheetViewport[ViewportPaneType.Fixed].Rect);
            Assert.AreEqual(expectedHorizontalScrollableRect, this.Args.SheetViewport[ViewportPaneType.HorizontalScrollable].Rect);
            Assert.AreEqual(expectedVerticalScrollableRect, this.Args.SheetViewport[ViewportPaneType.VerticalScrollable].Rect);
            Assert.AreEqual(expectedScrollableRect, this.Args.SheetViewport[ViewportPaneType.Scrollable].Rect);
        }

        private void FreezePanes(CellIndex frozenCellIndex)
        {
            this.Args.FrozenCellIndex = frozenCellIndex;
            SheetViewportTestsHelper.FreezePanes(this.Args);
        }

        private void SetViewportTopLeftCellIndex(CellIndex topLeftCellIndex)
        {
            SheetViewportTestsHelper.SetViewportTopLeftCellIndex(this.Args, topLeftCellIndex);
        }

        [TestInitialize]
        public void TestInitialize()
        {
            RadWorksheetLayout radWorksheetLayout = new RadWorksheetLayout(new SizeI(SpreadsheetDefaultValues.ColumnCount, SpreadsheetDefaultValues.RowCount));
            radWorksheetLayout.SetWorksheet(TestHelper.CreateWorksheet(), false);
            radWorksheetLayout.Measure();

            this.Args = new SheetViewportTestsArgs(new SheetViewport(), radWorksheetLayout, new Size(1, 1), new CellIndex(0, 0));
            SheetViewportTestsHelper.InitializeViewport(this.Args);
        }

        #endregion


        #region Freeze without initial offset

        [TestMethod]
        public void FreezeAboveAndLeftFromSelectedCell_WhenThereIsNoHorizontalOrVerticalOffset()
        {
            CellIndex frozenCellIndex = new CellIndex(2, 2);
            this.FreezePanes(frozenCellIndex);

            double fixedX = 0;
            double fixedY = 0;
            double fixedWidth = 130;
            double fixedHeight = 40;

            double scrollableX = 130;
            double scrollableY = 40;
            double scrollableWidth = this.ViewportSize.Width - fixedWidth;
            double scrollableHeight = this.ViewportSize.Height - fixedHeight;

            Rect expectedFixedRect = new Rect(fixedX, fixedY, fixedWidth, fixedHeight);
            Rect expectedHorizontalScrollableRect = new Rect(scrollableX, fixedY, scrollableWidth, fixedHeight);
            Rect expectedVerticalScrollableRect = new Rect(fixedX, scrollableY, fixedWidth, scrollableHeight);
            Rect expectedScrollableRect = new Rect(scrollableX, scrollableY, scrollableWidth, scrollableHeight);

            this.AssertViewportRects(expectedFixedRect, expectedHorizontalScrollableRect, expectedVerticalScrollableRect, expectedScrollableRect);
        }

        [TestMethod]
        public void FreezeOnTheRightFromSelectedCell_WhenThereIsNoHorizontalOrVerticalOffset()
        {
            CellIndex frozenCellIndex = new CellIndex(0, 3);
            this.FreezePanes(frozenCellIndex);

            double fixedX = 0;
            double fixedY = 0;
            double fixedWidth = 195;
            double fixedHeight = 0;

            double scrollableX = 195;
            double scrollableY = 0;
            double scrollableWidth = this.ViewportSize.Width - fixedWidth;
            double scrollableHeight = this.ViewportSize.Height - fixedHeight;

            Rect expectedFixedRect = new Rect(fixedX, fixedY, fixedWidth, fixedHeight);
            Rect expectedHorizontalScrollableRect = new Rect(scrollableX, fixedY, scrollableWidth, fixedHeight);
            Rect expectedVerticalScrollableRect = new Rect(fixedX, scrollableY, fixedWidth, scrollableHeight);
            Rect expectedScrollableRect = new Rect(scrollableX, scrollableY, scrollableWidth, scrollableHeight);

            this.AssertViewportRects(expectedFixedRect, expectedHorizontalScrollableRect, expectedVerticalScrollableRect, expectedScrollableRect);
        }

        [TestMethod]
        public void FreezeAboveSelectedCell_WhenThereIsNoHorizontalOrVerticalOffset()
        {
            CellIndex frozenCellIndex = new CellIndex(3, 0);
            this.FreezePanes(frozenCellIndex);

            double fixedX = 0;
            double fixedY = 0;
            double fixedWidth = 0;
            double fixedHeight = 60;

            double scrollableX = 0;
            double scrollableY = 60;
            double scrollableWidth = this.ViewportSize.Width - fixedWidth;
            double scrollableHeight = this.ViewportSize.Height - fixedHeight;

            Rect expectedFixedRect = new Rect(fixedX, fixedY, fixedWidth, fixedHeight);
            Rect expectedHorizontalScrollableRect = new Rect(scrollableX, fixedY, scrollableWidth, fixedHeight);
            Rect expectedVerticalScrollableRect = new Rect(fixedX, scrollableY, fixedWidth, scrollableHeight);
            Rect expectedScrollableRect = new Rect(scrollableX, scrollableY, scrollableWidth, scrollableHeight);

            this.AssertViewportRects(expectedFixedRect, expectedHorizontalScrollableRect, expectedVerticalScrollableRect, expectedScrollableRect);
        }

        #endregion

        #region Freeze without initial offset hide columns

        [TestMethod]
        public void FreezeAboveAndLeftFromSelectedCellAndHideFirstColumn_WhenThereIsNoHorizontalOrVerticalOffset()
        {
            CellIndex frozenCellIndex = new CellIndex(1, 1);
            this.FreezePanes(frozenCellIndex);

            double fixedX = 0;
            double fixedY = 0;
            double fixedWidth = 0;
            double fixedHeight = 20;

            double scrollableX = 0;
            double scrollableY = 20;
            double scrollableWidth = this.ViewportSize.Width - fixedWidth;
            double scrollableHeight = this.ViewportSize.Height - fixedHeight;

            Rect expectedFixedRect = new Rect(fixedX, fixedY, fixedWidth, fixedHeight);
            Rect expectedHorizontalScrollableRect = new Rect(scrollableX, fixedY, scrollableWidth, fixedHeight);
            Rect expectedVerticalScrollableRect = new Rect(fixedX, scrollableY, fixedWidth, scrollableHeight);
            Rect expectedScrollableRect = new Rect(scrollableX, scrollableY, scrollableWidth, scrollableHeight);

            this.Args.Worksheet.Columns[0].SetHidden(true);
            this.SetViewportTopLeftCellIndex(new CellIndex(0, 0));

            this.AssertViewportRects(expectedFixedRect, expectedHorizontalScrollableRect, expectedVerticalScrollableRect, expectedScrollableRect);

            SheetViewport sheetViewport = this.Args.SheetViewport;

            CellRange expectedFixedRange = new CellRange(0, 0, 0, 0);
            CellRange expectedVerticalScrollbleRange = new CellRange(0, 0, 0, 0);
            CellRange expectedHorizontalScrollableRange = new CellRange(0, 1, 0, 19);
            CellRange expectedScrollableRange = new CellRange(1, 1, 35, 19);

            Assert.AreEqual(expectedFixedRange, sheetViewport[ViewportPaneType.Fixed].VisibleRange);
            Assert.AreEqual(expectedVerticalScrollbleRange, sheetViewport[ViewportPaneType.VerticalScrollable].VisibleRange);
            Assert.AreEqual(expectedHorizontalScrollableRange, sheetViewport[ViewportPaneType.HorizontalScrollable].VisibleRange);
            Assert.AreEqual(expectedScrollableRange, sheetViewport[ViewportPaneType.Scrollable].VisibleRange);
        }

        #endregion


        #region Freeze with initial offset

        [TestMethod]
        public void FreezeAboveAndLeftFromSelectedCell_WhenThereIsHorizontalAndVerticalOffset()
        {
            CellIndex topLeftCellIndex = new CellIndex(4, 4);
            this.SetViewportTopLeftCellIndex(topLeftCellIndex);

            CellIndex frozenCellIndex = new CellIndex(8, 8);
            this.FreezePanes(frozenCellIndex);

            double fixedX = 260;
            double fixedY = 80;
            double fixedWidth = 260;
            double fixedHeight = 80;

            double scrollableX = 520;
            double scrollableY = 160;
            double scrollableWidth = this.ViewportSize.Width - fixedWidth;
            double scrollableHeight = this.ViewportSize.Height - fixedHeight;

            Rect expectedFixedRect = new Rect(fixedX, fixedY, fixedWidth, fixedHeight);
            Rect expectedHorizontalScrollableRect = new Rect(scrollableX, fixedY, scrollableWidth, fixedHeight);
            Rect expectedVerticalScrollableRect = new Rect(fixedX, scrollableY, fixedWidth, scrollableHeight);
            Rect expectedScrollableRect = new Rect(scrollableX, scrollableY, scrollableWidth, scrollableHeight);

            this.AssertViewportRects(expectedFixedRect, expectedHorizontalScrollableRect, expectedVerticalScrollableRect, expectedScrollableRect);
        }

        [TestMethod]
        public void FreezeAboveAndLeftFromSelectedCell_WhenThereIsVerticalOffset()
        {
            CellIndex topLeftCellIndex = new CellIndex(4, 0);
            this.SetViewportTopLeftCellIndex(topLeftCellIndex);

            CellIndex frozenCellIndex = new CellIndex(8, 8);
            this.FreezePanes(frozenCellIndex);

            double fixedX = 0;
            double fixedY = 80;
            double fixedWidth = 520;
            double fixedHeight = 80;

            double scrollableX = 520;
            double scrollableY = 160;
            double scrollableWidth = this.ViewportSize.Width - fixedWidth;
            double scrollableHeight = this.ViewportSize.Height - fixedHeight;

            Rect expectedFixedRect = new Rect(fixedX, fixedY, fixedWidth, fixedHeight);
            Rect expectedHorizontalScrollableRect = new Rect(scrollableX, fixedY, scrollableWidth, fixedHeight);
            Rect expectedVerticalScrollableRect = new Rect(fixedX, scrollableY, fixedWidth, scrollableHeight);
            Rect expectedScrollableRect = new Rect(scrollableX, scrollableY, scrollableWidth, scrollableHeight);

            this.AssertViewportRects(expectedFixedRect, expectedHorizontalScrollableRect, expectedVerticalScrollableRect, expectedScrollableRect);
        }

        [TestMethod]
        public void FreezeAboveAndLeftFromSelectedCell_WhenThereIsHorizontalOffset()
        {
            CellIndex topLeftCellIndex = new CellIndex(0, 4);
            this.SetViewportTopLeftCellIndex(topLeftCellIndex);

            CellIndex frozenCellIndex = new CellIndex(8, 8);
            this.FreezePanes(frozenCellIndex);

            double fixedX = 260;
            double fixedY = 0;
            double fixedWidth = 260;
            double fixedHeight = 160;

            double scrollableX = 520;
            double scrollableY = 160;
            double scrollableWidth = this.ViewportSize.Width - fixedWidth;
            double scrollableHeight = this.ViewportSize.Height - fixedHeight;

            Rect expectedFixedRect = new Rect(fixedX, fixedY, fixedWidth, fixedHeight);
            Rect expectedHorizontalScrollableRect = new Rect(scrollableX, fixedY, scrollableWidth, fixedHeight);
            Rect expectedVerticalScrollableRect = new Rect(fixedX, scrollableY, fixedWidth, scrollableHeight);
            Rect expectedScrollableRect = new Rect(scrollableX, scrollableY, scrollableWidth, scrollableHeight);

            this.AssertViewportRects(expectedFixedRect, expectedHorizontalScrollableRect, expectedVerticalScrollableRect, expectedScrollableRect);
        }

        #endregion


        #region Freeze without initial offset and scroll after

        [TestMethod]
        public void FreezeAboveAndLeftFromSelectedCellAndScroll_WhenThereIsNoHorizontalOrVerticalOffset()
        {
            CellIndex frozenCellIndex = new CellIndex(2, 2);
            this.FreezePanes(frozenCellIndex);
            this.SetViewportTopLeftCellIndex(new CellIndex(5, 5));

            double fixedX = 0;
            double fixedY = 0;
            double fixedWidth = 130;
            double fixedHeight = 40;

            double scrollableX = 325;
            double scrollableY = 100;
            double scrollableWidth = this.ViewportSize.Width - fixedWidth;
            double scrollableHeight = this.ViewportSize.Height - fixedHeight;

            Rect expectedFixedRect = new Rect(fixedX, fixedY, fixedWidth, fixedHeight);
            Rect expectedHorizontalScrollableRect = new Rect(scrollableX, fixedY, scrollableWidth, fixedHeight);
            Rect expectedVerticalScrollableRect = new Rect(fixedX, scrollableY, fixedWidth, scrollableHeight);
            Rect expectedScrollableRect = new Rect(scrollableX, scrollableY, scrollableWidth, scrollableHeight);

            this.AssertViewportRects(expectedFixedRect, expectedHorizontalScrollableRect, expectedVerticalScrollableRect, expectedScrollableRect);
        }

        [TestMethod]
        public void FreezeOnTheRightFromSelectedCellAndScroll_WhenThereIsNoHorizontalOrVerticalOffset()
        {
            CellIndex frozenCellIndex = new CellIndex(0, 3);
            this.FreezePanes(frozenCellIndex);
            this.SetViewportTopLeftCellIndex(new CellIndex(5, 5));

            double fixedX = 0;
            double fixedY = 0;
            double fixedWidth = 195;
            double fixedHeight = 0;

            double scrollableX = 325;
            double scrollableY = 100;
            double scrollableWidth = this.ViewportSize.Width - fixedWidth;
            double scrollableHeight = this.ViewportSize.Height - fixedHeight;

            Rect expectedFixedRect = new Rect(fixedX, fixedY, fixedWidth, fixedHeight);
            Rect expectedHorizontalScrollableRect = new Rect(scrollableX, fixedY, scrollableWidth, fixedHeight);
            Rect expectedVerticalScrollableRect = new Rect(fixedX, scrollableY, fixedWidth, scrollableHeight);
            Rect expectedScrollableRect = new Rect(scrollableX, scrollableY, scrollableWidth, scrollableHeight);

            this.AssertViewportRects(expectedFixedRect, expectedHorizontalScrollableRect, expectedVerticalScrollableRect, expectedScrollableRect);
        }

        [TestMethod]
        public void FreezeAboveSelectedCellAndScroll_WhenThereIsNoHorizontalOrVerticalOffset()
        {
            CellIndex frozenCellIndex = new CellIndex(3, 0);
            this.FreezePanes(frozenCellIndex);
            this.SetViewportTopLeftCellIndex(new CellIndex(5, 5));

            double fixedX = 0;
            double fixedY = 0;
            double fixedWidth = 0;
            double fixedHeight = 60;

            double scrollableX = 325;
            double scrollableY = 100;
            double scrollableWidth = this.ViewportSize.Width - fixedWidth;
            double scrollableHeight = this.ViewportSize.Height - fixedHeight;

            Rect expectedFixedRect = new Rect(fixedX, fixedY, fixedWidth, fixedHeight);
            Rect expectedHorizontalScrollableRect = new Rect(scrollableX, fixedY, scrollableWidth, fixedHeight);
            Rect expectedVerticalScrollableRect = new Rect(fixedX, scrollableY, fixedWidth, scrollableHeight);
            Rect expectedScrollableRect = new Rect(scrollableX, scrollableY, scrollableWidth, scrollableHeight);

            this.AssertViewportRects(expectedFixedRect, expectedHorizontalScrollableRect, expectedVerticalScrollableRect, expectedScrollableRect);
        }

        #endregion


        #region Freeze with initial offset and scroll after

        [TestMethod]
        public void FreezeAboveAndLeftFromSelectedCellAndScroll_WhenThereIsHorizontalAndVerticalOffset()
        {
            this.SetViewportTopLeftCellIndex(new CellIndex(5, 5));

            CellIndex frozenCellIndex = new CellIndex(10, 10);
            this.FreezePanes(frozenCellIndex);

            this.SetViewportTopLeftCellIndex(new CellIndex(20, 20));

            double fixedX = 325;
            double fixedY = 100;
            double fixedWidth = 325;
            double fixedHeight = 100;

            double scrollableX = 1300;
            double scrollableY = 400;
            double scrollableWidth = this.ViewportSize.Width - fixedWidth;
            double scrollableHeight = this.ViewportSize.Height - fixedHeight;

            Rect expectedFixedRect = new Rect(fixedX, fixedY, fixedWidth, fixedHeight);
            Rect expectedHorizontalScrollableRect = new Rect(scrollableX, fixedY, scrollableWidth, fixedHeight);
            Rect expectedVerticalScrollableRect = new Rect(fixedX, scrollableY, fixedWidth, scrollableHeight);
            Rect expectedScrollableRect = new Rect(scrollableX, scrollableY, scrollableWidth, scrollableHeight);

            this.AssertViewportRects(expectedFixedRect, expectedHorizontalScrollableRect, expectedVerticalScrollableRect, expectedScrollableRect);
        }

        [TestMethod]
        public void FreezeAboveAndLeftFromSelectedCellAndScroll_WhenThereIsVerticalOffset()
        {
            this.SetViewportTopLeftCellIndex(new CellIndex(5, 0));

            CellIndex frozenCellIndex = new CellIndex(10, 10);
            this.FreezePanes(frozenCellIndex);

            this.SetViewportTopLeftCellIndex(new CellIndex(20, 20));

            double fixedX = 0;
            double fixedY = 100;
            double fixedWidth = 650;
            double fixedHeight = 100;

            double scrollableX = 1300;
            double scrollableY = 400;
            double scrollableWidth = this.ViewportSize.Width - fixedWidth;
            double scrollableHeight = this.ViewportSize.Height - fixedHeight;

            Rect expectedFixedRect = new Rect(fixedX, fixedY, fixedWidth, fixedHeight);
            Rect expectedHorizontalScrollableRect = new Rect(scrollableX, fixedY, scrollableWidth, fixedHeight);
            Rect expectedVerticalScrollableRect = new Rect(fixedX, scrollableY, fixedWidth, scrollableHeight);
            Rect expectedScrollableRect = new Rect(scrollableX, scrollableY, scrollableWidth, scrollableHeight);

            this.AssertViewportRects(expectedFixedRect, expectedHorizontalScrollableRect, expectedVerticalScrollableRect, expectedScrollableRect);
        }

        [TestMethod]
        public void FreezeAboveAndLeftFromSelectedCellAndScroll_WhenThereIsHorizontalOffset()
        {
            this.SetViewportTopLeftCellIndex(new CellIndex(0, 5));

            CellIndex frozenCellIndex = new CellIndex(10, 10);
            this.FreezePanes(frozenCellIndex);

            this.SetViewportTopLeftCellIndex(new CellIndex(20, 20));

            double fixedX = 325;
            double fixedY = 0;
            double fixedWidth = 325;
            double fixedHeight = 200;

            double scrollableX = 1300;
            double scrollableY = 400;
            double scrollableWidth = this.ViewportSize.Width - fixedWidth;
            double scrollableHeight = this.ViewportSize.Height - fixedHeight;

            Rect expectedFixedRect = new Rect(fixedX, fixedY, fixedWidth, fixedHeight);
            Rect expectedHorizontalScrollableRect = new Rect(scrollableX, fixedY, scrollableWidth, fixedHeight);
            Rect expectedVerticalScrollableRect = new Rect(fixedX, scrollableY, fixedWidth, scrollableHeight);
            Rect expectedScrollableRect = new Rect(scrollableX, scrollableY, scrollableWidth, scrollableHeight);

            this.AssertViewportRects(expectedFixedRect, expectedHorizontalScrollableRect, expectedVerticalScrollableRect, expectedScrollableRect);
        }

        [TestMethod]
        public void FreezeAboveFromSelectedCellAndScroll_WhenThereIsHorizontalOffset()
        {
            this.SetViewportTopLeftCellIndex(new CellIndex(5, 5));

            CellIndex frozenCellIndex = new CellIndex(10, 0);
            this.FreezePanes(frozenCellIndex);

            this.SetViewportTopLeftCellIndex(new CellIndex(20, 20));

            double fixedX = 325;
            double fixedY = 100;
            double fixedWidth = 0;
            double fixedHeight = 100;

            double scrollableX = 1300;
            double scrollableY = 400;
            double scrollableWidth = this.ViewportSize.Width - fixedWidth;
            double scrollableHeight = this.ViewportSize.Height - fixedHeight;

            Rect expectedFixedRect = new Rect(fixedX, fixedY, fixedWidth, fixedHeight);
            Rect expectedHorizontalScrollableRect = new Rect(scrollableX, fixedY, scrollableWidth, fixedHeight);
            Rect expectedVerticalScrollableRect = new Rect(fixedX, scrollableY, fixedWidth, scrollableHeight);
            Rect expectedScrollableRect = new Rect(scrollableX, scrollableY, scrollableWidth, scrollableHeight);

            this.AssertViewportRects(expectedFixedRect, expectedHorizontalScrollableRect, expectedVerticalScrollableRect, expectedScrollableRect);
        }

        [TestMethod]
        public void FreezeLeftFromSelectedCellAndScroll_WhenThereIsHorizontalOffset()
        {
            this.SetViewportTopLeftCellIndex(new CellIndex(5, 5));

            CellIndex frozenCellIndex = new CellIndex(0, 10);
            this.FreezePanes(frozenCellIndex);

            this.SetViewportTopLeftCellIndex(new CellIndex(20, 20));

            double fixedX = 325;
            double fixedY = 100;
            double fixedWidth = 325;
            double fixedHeight = 0;

            double scrollableX = 1300;
            double scrollableY = 400;
            double scrollableWidth = this.ViewportSize.Width - fixedWidth;
            double scrollableHeight = this.ViewportSize.Height - fixedHeight;

            Rect expectedFixedRect = new Rect(fixedX, fixedY, fixedWidth, fixedHeight);
            Rect expectedHorizontalScrollableRect = new Rect(scrollableX, fixedY, scrollableWidth, fixedHeight);
            Rect expectedVerticalScrollableRect = new Rect(fixedX, scrollableY, fixedWidth, scrollableHeight);
            Rect expectedScrollableRect = new Rect(scrollableX, scrollableY, scrollableWidth, scrollableHeight);

            this.AssertViewportRects(expectedFixedRect, expectedHorizontalScrollableRect, expectedVerticalScrollableRect, expectedScrollableRect);
        }

        #endregion
    }
}