﻿using System;
using System.IO;
using System.Linq;
using System.Windows;
using System.Windows.Media;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Telerik.Windows.Documents.Spreadsheet.Layout;
using Telerik.Windows.Documents.Spreadsheet.Model;
using Telerik.Windows.Documents.Spreadsheet.Model.Filtering;
using Telerik.Windows.Documents.Spreadsheet.Model.Printing;
using Telerik.Windows.Documents.Spreadsheet.Model.Protection;
using Telerik.Windows.Documents.Spreadsheet.Model.Shapes;
using Telerik.Windows.Documents.Spreadsheet.Model.Sorting;
using Telerik.Windows.Documents.Spreadsheet.PropertySystem;
using Telerik.Windows.Documents.Spreadsheet.Tests.Utilities;
using Telerik.Windows.Documents.Spreadsheet.TestsInternal.Utilities;

namespace Telerik.Windows.Documents.Spreadsheet.TestsInternal
{
    [TestClass]
    public class WorksheetTests
    {
        private static readonly IFill originalFill = new PatternFill(PatternType.Solid, System.Windows.Media.Colors.Red, System.Windows.Media.Colors.Red);
        private static readonly IFill modifiedFill = new PatternFill(PatternType.DiagonalStripe, System.Windows.Media.Colors.Blue, System.Windows.Media.Colors.Blue);

        [TestMethod]
        public void CopyFrom_WhenSourceWorksheetHasDefaultProperties_CopiedWorksheetIsSameAsTheOriginal()
        {
            // Arrange
            Worksheet worksheet = new Worksheet("Sheet1", new Workbook());

            // Act
            Worksheet copiedWorksheet = CreateCopiedWorksheet(worksheet);

            // Assert
            TestHelper.AssertWorksheetsPropertiesAreEqual(worksheet, copiedWorksheet);
        }

        [TestMethod]
        public void CopyFrom_WhenSourceWorksheetHasSetProperties_CopiedWorksheetIsSameAsTheOriginal()
        {
            // Arrange
            Worksheet worksheet = CreateWorksheetWithPropertiesTemplate("Sheet1");

            // Act
            Worksheet copiedWorksheet = CreateCopiedWorksheet(worksheet);

            // Assert
            TestHelper.AssertWorksheetsPropertiesAreEqual(worksheet, copiedWorksheet, false);
        }

        [TestMethod]
        public void CopyFrom_WhenSourceWorksheetHasDefaultPropertiesAndCopiedWorksheetHasModifiedThesePropertiesAfterCopyFrom_OriginalWorksheetIsUnchanged()
        {
            // Arrange
            Worksheet originalWorksheet = CreateWorksheetWithPropertiesTemplate("Sheet1");

            // Act
            Worksheet copiedWorksheet = CreateCopiedWorksheet(originalWorksheet);

            ModifyProperties(copiedWorksheet, false);

            // Assert
            TestHelper.AssertWorksheetsPropertiesAreEqual(CreateWorksheetWithPropertiesTemplate(originalWorksheet.Name), originalWorksheet, false);
        }

        [TestMethod]
        public void CopyFrom_WhenSourceWorksheetHasDefaultPropertiesAndSourceWorksheetHasModifiedThesePropertiesAfterCopyFrom_CopiedWorksheetIsUnchanged()
        {
            // Arrange
            Worksheet originalWorksheet = CreateWorksheetWithPropertiesTemplate("Sheet1");

            // Act
            Worksheet copiedWorksheet = CreateCopiedWorksheet(originalWorksheet);

            ModifyProperties(originalWorksheet);

            // Assert
            TestHelper.AssertWorksheetsPropertiesAreEqual(CreateWorksheetWithPropertiesTemplate(originalWorksheet.Name), copiedWorksheet, false);
        }

        [TestMethod]
        [ExpectedException(typeof(ProtectionException))]
        public void CopyFrom_WhenSourceWorksheetIsProtected_ProtectionIsPreserved()
        {
            // Arrange
            Worksheet sourceWorksheet = new Worksheet("source", new Workbook());
            string password = "pass";
            sourceWorksheet.Protect(password, sourceWorksheet.ProtectionOptions);

            // Act 
            Worksheet copiedWorksheet = CreateCopiedWorksheet(sourceWorksheet);

            // Assert
            copiedWorksheet.Protect(password, copiedWorksheet.ProtectionOptions);
        }

        [TestMethod]
        public void CopyFrom_WhenWorksheetIsProtected_CanBeUnprotected()
        {
            // Arrange
            Worksheet sourceWorksheet = new Worksheet("source", new Workbook());
            string password = "pass";
            sourceWorksheet.Protect(password, sourceWorksheet.ProtectionOptions);

            // Act 
            Worksheet copiedWorksheet = CreateCopiedWorksheet(sourceWorksheet);

            // Assert
            Assert.IsTrue(copiedWorksheet.Unprotect(password));
        }

        [TestMethod]
        public void CopyFrom_WhenSourceWorksheetHasDefinedNameWhichRefersToSheetName_DefinedNameInCopiedWorksheetPointsToCorrectWorksheet()
        {
            // Assert
            Worksheet sourceWorksheet = new Worksheet("SourceSheet", new Workbook());
            sourceWorksheet.Names.Add("test", "=SourceSheet!$A$1", new CellIndex(0, 0), "Comment");

            // Act
            Worksheet copiedWorksheet = CreateCopiedWorksheet(sourceWorksheet);

            // Assert
            ISpreadsheetName clonedName;
            copiedWorksheet.Names.TryGetSpreadsheetName("test", out clonedName);
            Assert.IsNotNull(clonedName);

            string expectedRefersTo = string.Format("={0}!$A$1", copiedWorksheet.Name);
            Assert.AreEqual(expectedRefersTo, clonedName.RefersTo);
        }

        [TestMethod]
        public void CopyFrom_WhenSourceWorksheetHasDefinedNameAndTheClonedNameRefersToCellValueIsModified_DefinedNameInSourceWorksheetIsUnchanged()
        {
            // Assert
            Workbook sourceWorkbook = new Workbook();
            Worksheet sourceWorksheet = sourceWorkbook.Worksheets.Add();
            sourceWorksheet.Name = "SourceSheet";
            sourceWorksheet.Cells[0, 0].SetValue(10);
            sourceWorksheet.Names.Add("test", "=SourceSheet!$A$1", new CellIndex(0, 0));
            sourceWorksheet.Cells[1, 1].SetValue("=test");

            // Act
            Worksheet copiedWorksheet = CreateCopiedWorksheet(sourceWorksheet);
            copiedWorksheet.Cells[0, 0].SetValue(15);

            // Assert
            ISpreadsheetName clonedName;
            copiedWorksheet.Names.TryGetSpreadsheetName("test", out clonedName);
            Assert.IsNotNull(clonedName);

            FormulaCellValue clonedFormulaCellValue = ((FormulaCellValue)copiedWorksheet.Cells[1, 1].GetValue().Value);
            string clonedValue = clonedFormulaCellValue.GetResultValueAsString(CellValueFormat.GeneralFormat);

            FormulaCellValue sourceFormulaCellValue = ((FormulaCellValue)sourceWorksheet.Cells[1, 1].GetValue().Value);
            string sourceValue = sourceFormulaCellValue.GetResultValueAsString(CellValueFormat.GeneralFormat);

            Assert.AreEqual("15", clonedValue);
            Assert.AreEqual("10", sourceValue);
        }

        [TestMethod]
        public void CopyFrom_WhenSourceWorksheetHasDefinedNameAndTheSourceNameRefersToCellValueIsModified_DefinedNameInCopiedWorksheetIsUnchanged()
        {
            // Assert
            Workbook sourceWorkbook = new Workbook();
            Worksheet sourceWorksheet = sourceWorkbook.Worksheets.Add();
            sourceWorksheet.Name = "SourceSheet";
            sourceWorksheet.Cells[0, 0].SetValue(10);
            sourceWorksheet.Names.Add("test", "=SourceSheet!$A$1", new CellIndex(0, 0));
            sourceWorksheet.Cells[1, 1].SetValue("=test");

            // Act
            Worksheet copiedWorksheet = CreateCopiedWorksheet(sourceWorksheet);
            sourceWorksheet.Cells[0, 0].SetValue(15);

            // Assert
            ISpreadsheetName clonedName;
            copiedWorksheet.Names.TryGetSpreadsheetName("test", out clonedName);
            Assert.IsNotNull(clonedName);

            FormulaCellValue clonedFormulaCellValue = ((FormulaCellValue)copiedWorksheet.Cells[1, 1].GetValue().Value);
            string clonedValue = clonedFormulaCellValue.GetResultValueAsString(CellValueFormat.GeneralFormat);

            FormulaCellValue sourceFormulaCellValue = ((FormulaCellValue)sourceWorksheet.Cells[1, 1].GetValue().Value);
            string sourceValue = sourceFormulaCellValue.GetResultValueAsString(CellValueFormat.GeneralFormat);

            Assert.AreEqual("10", clonedValue);
            Assert.AreEqual("15", sourceValue);
        }

        [TestMethod]
        public void CopyFrom_WhenSourceWorksheetHasDefinedNameAndTargetWorksheetIsFromOtherWorkbookAndTheSourceNameIsRenamed_DefinedNameInCopiedWorksheetIsUnchanged()
        {
            // Assert
            Workbook sourceWorkbook = new Workbook();
            Worksheet sourceWorksheet = sourceWorkbook.Worksheets.Add();
            sourceWorksheet.Name = "SourceSheet";
            sourceWorksheet.Cells[0, 0].SetValue(10);
            ISpreadsheetName spreadsheetName = sourceWorksheet.Names.Add("test", "=SourceSheet!$A$1", new CellIndex(0, 0));
            sourceWorksheet.Cells[1, 1].SetValue("=test");

            // Act
            Workbook targetWorkbook = new Workbook();
            Worksheet copiedWorksheet = targetWorkbook.Worksheets.Add();
            copiedWorksheet.CopyFrom(sourceWorksheet);

            sourceWorksheet.Cells[0, 0].SetValue(15);
            sourceWorksheet.Names.Update(spreadsheetName, "newName", spreadsheetName.RefersTo, new CellIndex(0, 0));
            sourceWorksheet.Cells[1, 1].SetValue("=newName");

            // Assert
            ISpreadsheetName clonedName;
            copiedWorksheet.Names.TryGetSpreadsheetName("test", out clonedName);
            Assert.IsNotNull(clonedName);

            FormulaCellValue clonedFormulaCellValue = ((FormulaCellValue)copiedWorksheet.Cells[1, 1].GetValue().Value);
            string clonedValue = clonedFormulaCellValue.GetResultValueAsString(CellValueFormat.GeneralFormat);

            FormulaCellValue sourceFormulaCellValue = ((FormulaCellValue)sourceWorksheet.Cells[1, 1].GetValue().Value);
            string sourceValue = sourceFormulaCellValue.GetResultValueAsString(CellValueFormat.GeneralFormat);

            Assert.AreEqual("10", clonedValue);
            Assert.AreEqual("15", sourceValue);
        }

        [TestMethod]
        public void CopyFrom_WhenSourceWorksheetHasDefinedNameWithScopeWorkbook_WorkbookOfCopiedWorksheetContainsDefinedName()
        {
            // Assert
            Workbook sourceWorkbook = new Workbook();
            Worksheet sourceWorksheet = sourceWorkbook.Worksheets.Add();
            sourceWorksheet.Name = "SourceSheet";
            sourceWorkbook.Names.Add("test", "=$A$1", new CellIndex(0, 0));
            sourceWorksheet.Cells[0, 0].SetValue(10);
            sourceWorksheet.Cells[1, 1].SetValue("=test");

            // Act
            Workbook workbook = new Workbook();
            Worksheet copiedWorksheet = workbook.Worksheets.Add();
            copiedWorksheet.CopyFrom(sourceWorksheet);

            // Assert
            ISpreadsheetName clonedName;
            copiedWorksheet.Workbook.Names.TryGetSpreadsheetName("test", out clonedName);
            Assert.IsNotNull(clonedName);

            FormulaCellValue clonedFormulaCellValue = ((FormulaCellValue)copiedWorksheet.Cells[1, 1].GetValue().Value);
            string clonedValue = clonedFormulaCellValue.GetResultValueAsString(CellValueFormat.GeneralFormat);

            FormulaCellValue sourceFormulaCellValue = ((FormulaCellValue)sourceWorksheet.Cells[1, 1].GetValue().Value);
            string sourceValue = sourceFormulaCellValue.GetResultValueAsString(CellValueFormat.GeneralFormat);

            Assert.AreEqual("10", clonedValue);
            Assert.AreEqual("10", sourceValue);
        }

        [TestMethod]
        public void CopyFrom_WhenSourceWorksheetHasDefinedNameWithScopeWorkbookAndTargetWorkbookHasDefinedNameWithSameName_ScopeOfTheClonedNameIsChangedToWorksheet()
        {
            // Assert
            string definedName = "test";

            Workbook sourceWorkbook = new Workbook();
            Worksheet sourceWorksheet = sourceWorkbook.Worksheets.Add();
            sourceWorksheet.Name = "SourceSheet";
            sourceWorkbook.Names.Add(definedName, "=SourceSheet!$A$1", new CellIndex(0, 0));
            sourceWorksheet.Cells[0, 0].SetValue(10);
            sourceWorksheet.Cells[1, 1].SetValue(string.Format("={0}", definedName));

            Workbook targetWorkbook = new Workbook();
            Worksheet targetWorksheet = targetWorkbook.Worksheets.Add();
            targetWorksheet.Workbook.Names.Add(definedName, "=$B$2", new CellIndex(1, 1));
            targetWorksheet.Cells[1, 1].SetValue(30);
            targetWorksheet.Cells[0, 0].SetValue(string.Format("={0}", definedName));

            // Act
            targetWorksheet.CopyFrom(sourceWorksheet);

            // Assert
            ISpreadsheetName clonedName;
            targetWorksheet.Names.TryGetSpreadsheetName(definedName, out clonedName);
            Assert.IsNotNull(clonedName);

            FormulaCellValue clonedFormulaCellValue = ((FormulaCellValue)targetWorksheet.Cells[1, 1].GetValue().Value);
            string clonedValue = clonedFormulaCellValue.GetResultValueAsString(CellValueFormat.GeneralFormat);

            FormulaCellValue sourceFormulaCellValue = ((FormulaCellValue)sourceWorksheet.Cells[1, 1].GetValue().Value);
            string sourceValue = sourceFormulaCellValue.GetResultValueAsString(CellValueFormat.GeneralFormat);

            Assert.AreEqual("10", clonedValue);
            Assert.AreEqual("10", sourceValue);
        }

        [TestMethod]
        public void CopyFrom_WhenSourceWorksheetContainsFormulaCellValueWhichUsesTheSheetName_FormulaCellValueInCopiedWorksheetRefersToCopiedWorksheet()
        {
            // Arrange
            Worksheet sourceWorksheet = new Worksheet("source", new Workbook());
            sourceWorksheet.Cells[0, 0].SetValue(10);
            CellIndex start = new CellIndex(1, 1);
            CellIndex end = new CellIndex(10, 10);
            sourceWorksheet.Cells[start, end].SetValue("=source!$A$1");

            // Act
            Worksheet copiedWorksheet = CreateCopiedWorksheet(sourceWorksheet);

            // Assert
            Assert.AreEqual("10", copiedWorksheet.Cells[0, 0].GetValue().Value.RawValue);

            for (int i = 1; i <= 10; i++)
            {
                for (int j = 1; j <= 10; j++)
                {
                    Assert.AreEqual(string.Format("={0}!$A$1", copiedWorksheet.Name), copiedWorksheet.Cells[i, j].GetValue().Value.RawValue);
                }
            }
        }

        [TestMethod]
        public void CopyFrom_WhenTargetAndSourceWorkbooksContainDifferentStylesWithEqualNamesAndCellRangeHasThisStyleInSourceWorksheet_StyleIsAppliedOnCorrectCellRangeInResultWorksheet()
        {
            // Arrange
            string styleName = "MyStyle";
            Workbook sourceWorkbook = new Workbook();
            CellStyle sourceStyle = sourceWorkbook.Styles.Add(styleName);
            sourceStyle.Fill = new PatternFill(PatternType.Solid, Colors.Red, Colors.Red);
            Worksheet sourceWorksheet = sourceWorkbook.Worksheets.Add();
            sourceWorksheet.Cells[1, 3, 5, 3].SetStyleName(styleName); // From B4 to B6

            Workbook targetWorkbook = new Workbook();
            targetWorkbook.Styles.Add(styleName);
            Worksheet actualWorksheet = targetWorkbook.Worksheets.Add();

            // Act
            actualWorksheet.CopyFrom(sourceWorksheet);
            string epxectedStyleName = styleName + " 1";

            // Assert
            Assert.AreEqual(epxectedStyleName, actualWorksheet.Cells[1, 3, 5, 3].GetStyleName().Value);
            Assert.AreEqual("Normal", actualWorksheet.Cells[0, 0, 0, 0].GetStyleName().Value);
        }

        [TestMethod]
        public void CopyFrom_DoesntPutActionInTheHistoryStack()
        {
            // Arrange
            Workbook workbook = new Workbook();
            Worksheet worksheet = CreateWorksheetWithPropertiesTemplate("Sheet1", workbook);
            worksheet.DefaultColumnWidth = new ColumnWidth(100, true);
            worksheet.DefaultRowHeight = new RowHeight(100, true);

            workbook.History.Clear();

            // Act
            workbook.Worksheets.Add().CopyFrom(worksheet);

            // Assert
            Assert.IsFalse(workbook.History.CanUndo);
            Assert.IsFalse(workbook.History.CanRedo);
        }

        [TestMethod]
        public void CopyFrom_WhenTargetWorksheetHasContent_OverridesTargetWorksheetContentWithSourceWorksheetContentExceptSheetName()
        {
            // Arrange
            string sheetName = "Target";
            Worksheet targetWorksheet = CreateWorksheetWithPropertiesTemplate(sheetName);

            // Act
            Worksheet sourceWorksheet = TestHelper.CreateWorksheet();
            sourceWorksheet.Name = "Source";

            targetWorksheet.CopyFrom(sourceWorksheet);

            // Assert
            Assert.AreEqual(sheetName, targetWorksheet.Name);
            TestHelper.AssertWorksheetsPropertiesAreEqual(sourceWorksheet, targetWorksheet, false);
        }

        // Memory leak test
        // [TestMethod]
        public void CopyFrom_WhenSourceWorksheetHasProperties_CopiedWorksheetDoesntHoldStrongReferenceWhichResultToMemoryLeak()
        {
            Worksheet copiedWorksheet = null;
            WeakReference reference = null;
            new Action(() =>
            {
                Worksheet sourceWorksheet = CreateWorksheetWithPropertiesTemplate("Sheet1");
                reference = new WeakReference(sourceWorksheet);
                copiedWorksheet = CreateCopiedWorksheet(sourceWorksheet);
                sourceWorksheet.Workbook.Worksheets.Remove(sourceWorksheet);
            })();

            GC.WaitForPendingFinalizers();
            GC.Collect();

            Assert.IsFalse(reference.IsAlive);
            Assert.IsNotNull(copiedWorksheet);
        }

        [TestMethod]
        public void CopyFrom_WhenSourceWorksheetContainsCellReferenceRangeExpressionWhichRefersToOtherSheetsFromAndTargetIsFromOtherWorkbook_CopiesWorksheetContentAndHasErrorExpressionForCellReferenceRangeExpression()
        {
            // Arrange
            Workbook sourceWorkbook = new Workbook();
            Worksheet source = sourceWorkbook.Worksheets.Add();

            Worksheet sourceRelatedWorksheet = sourceWorkbook.Worksheets.Add();
            sourceRelatedWorksheet.Cells[0, 0, 0, 0].SetValue(25); // SUMIF related.

            sourceRelatedWorksheet.Cells[0, 1, 2, 1].SetValue("test"); // COUNTIF related.

            sourceRelatedWorksheet.Cells[0, 2, 0, 2].SetValue(0.5); //LOOKUP related.
            sourceRelatedWorksheet.Cells[1, 2, 1, 2].SetValue(0.8);
            sourceRelatedWorksheet.Cells[2, 2, 2, 2].SetValue(1.2);
            sourceRelatedWorksheet.Cells[0, 3, 0, 3].SetValue("min");
            sourceRelatedWorksheet.Cells[1, 3, 1, 3].SetValue("middle");
            sourceRelatedWorksheet.Cells[2, 3, 2, 3].SetValue("max");

            string sourceName = sourceRelatedWorksheet.Name;

            string sumIfCellValue = "=SUMIF(" + sourceName + "!A1{ls}\">10\"{ls}B1)";
            sumIfCellValue = sumIfCellValue.ReplaceCultureDependentSeparators();
            source.Cells[0, 0, 0, 0].SetValue(sumIfCellValue);

            string countIfCellValue = "=COUNTIF(" + sourceName + "!B1:B3{ls}\"test\")";
            countIfCellValue = countIfCellValue.ReplaceCultureDependentSeparators();
            source.Cells[0, 1, 0, 1].SetValue(countIfCellValue);

            string lookupCellValue = "=LOOKUP(1{ls}" + sourceName + "!C1:C3{ls}" + sourceName + "!D1:D3)";
            lookupCellValue = lookupCellValue.ReplaceCultureDependentSeparators();
            source.Cells[0, 2, 0, 2].SetValue(lookupCellValue);

            // Act
            Worksheet target = TestHelper.CreateWorksheet();
            target.CopyFrom(source);

            // Assert
            AssertFormulaCellVaueIsReferenceError(new CellIndex(0, 0), new CellIndex(0, 0), target);
            AssertFormulaCellVaueIsReferenceError(new CellIndex(0, 1), new CellIndex(0, 1), target);
            AssertFormulaCellVaueIsReferenceError(new CellIndex(0, 2), new CellIndex(0, 2), target);
        }

        private static void AssertFormulaCellVaueIsReferenceError(CellIndex fromCellIndex, CellIndex toCellIndex, Worksheet worksheet)
        {
            FormulaCellValue value = worksheet.Cells[fromCellIndex, toCellIndex].GetValue().Value as FormulaCellValue;
            Assert.IsNotNull(value);
            Assert.AreEqual(CellValueType.Error, value.ResultValueType);
            Assert.AreEqual("#REF!", value.Value.GetValue().ToString());
        }

        private static Worksheet CreateCopiedWorksheet(Worksheet worksheet)
        {
            Worksheet copiedWorksheet = worksheet.Workbook.Worksheets.Add();
            copiedWorksheet.CopyFrom(worksheet);
            return copiedWorksheet;
        }

        private static Worksheet CreateWorksheetWithPropertiesTemplate(string name, Workbook workbook = null)
        {
            bool shouldSetName = false;
            if (workbook == null)
            {
                workbook = new Workbook();
                shouldSetName = true;
            }

            Worksheet worksheet = workbook.Worksheets.Add();
            if (shouldSetName)
            {
                worksheet.Name = name;
            }

            SetProperties(worksheet);

            return worksheet;
        }

        private static void ModifyProperties(Worksheet worksheet, bool shouldRemoveNames = true)
        {
            // Cells
            CellSelection cells = worksheet.Cells[0, 0, 0, 0];
            cells.SetValue("cloned test");
            cells.SetFontSize(30);
            cells.SetBorders(new CellBorders(new CellBorder(CellBorderStyle.Dashed, ThemableColor.FromArgb(255, 0, 0, 255))));

            // Merged cells
            worksheet.Cells[1, 1, 10, 10].Unmerge();

            // Rows and Columns
            worksheet.Rows[0].SetFill(modifiedFill);
            worksheet.Columns[0].SetFill(modifiedFill);

            // Protection
            worksheet.ProtectionOptions = new WorksheetProtectionOptions(false, false, false, false, false, false, false, false, false);

            // Page setup
            worksheet.WorksheetPageSetup.CenterHorizontally = false;
            worksheet.WorksheetPageSetup.CenterVertically = false;
            worksheet.WorksheetPageSetup.PrintOptions.IsDraftQuality = false;
            worksheet.WorksheetPageSetup.PrintArea.SetPrintArea(new CellRange(1, 1, 19, 19));
            worksheet.WorksheetPageSetup.PageBreaks.Clear();

            // ViewState
            worksheet.ViewState.ScaleFactor = new Size(0.5, 0.5);
            worksheet.ViewState.Pane = new Pane(new CellIndex(1, 1), 2, 2, ViewportPaneType.Scrollable);

            // Filter
            worksheet.Filter.RemoveFilter(0);
            IFilter fillFilter = new FillColorFilter(0, modifiedFill);
            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            worksheet.Filter.SetFilter(fillFilter);

            // Grouping
            worksheet.Rows[10, 12].Ungroup();

            // Sorting
            worksheet.SortState.Clear();

            // Shapes
            worksheet.Shapes[0].SetHeight(true, 100);

            // Names
            if (shouldRemoveNames)
            {
                worksheet.Names.Remove(worksheet.Name);
            }

            // Hyperlinks
            worksheet.Hyperlinks.Remove(worksheet.Hyperlinks.FirstOrDefault());

            // Direct properties
            worksheet.DefaultColumnWidth = new ColumnWidth(150, true);
            worksheet.DefaultRowHeight = new RowHeight(150, true);
        }

        private static void SetProperties(Worksheet worksheet)
        {
            // Cells
            CellSelection cells = worksheet.Cells[0, 0, 0, 0];
            cells.SetValue("test");
            cells.SetFontSize(20);
            cells.SetBorders(new CellBorders(new CellBorder(CellBorderStyle.DashDot, ThemableColor.FromArgb(255, 255, 0, 0))));

            // Merged cells
            worksheet.Cells[1, 1, 10, 10].Merge();

            // Rows and Columns
            worksheet.Rows[0].SetFill(originalFill);
            worksheet.Columns[0].SetFill(originalFill);

            // Protection
            worksheet.ProtectionOptions = new WorksheetProtectionOptions(true, true, true, true, true, true, true, true, true);

            // Page setup
            worksheet.WorksheetPageSetup.CenterHorizontally = true;
            worksheet.WorksheetPageSetup.CenterVertically = true;
            worksheet.WorksheetPageSetup.PrintOptions.IsDraftQuality = true;
            worksheet.WorksheetPageSetup.PrintArea.SetPrintArea(new CellRange(0, 0, 20, 20));
            worksheet.WorksheetPageSetup.PageBreaks.AddPageBreak(new PageBreak(PageBreakType.Horizontal, 0, 2, 3));
            worksheet.WorksheetPageSetup.HeaderFooterSettings.Header.LeftSection.Text = "left";
            worksheet.WorksheetPageSetup.HeaderFooterSettings.Header.CenterSection.Text = "center";
            worksheet.WorksheetPageSetup.HeaderFooterSettings.Header.RightSection.Text = "right";

            // ViewState
            worksheet.ViewState.ScaleFactor = new Size(0.8, 0.8);
            worksheet.ViewState.Pane = new Pane(new CellIndex(2, 2), 1, 1, ViewportPaneType.HorizontalScrollable);

            // Filter
            IFilter fillFilter = new FillColorFilter(0, originalFill);
            worksheet.Filter.FilterRange = FilterTestHelperInternal.DefaultCellRange;
            worksheet.Filter.SetFilter(fillFilter);

            // Grouping
            worksheet.Rows[10, 12].Group();

            // Sorting : Blue, Red, Green => Red, Blue, Green.
            worksheet.Cells[0, 15, 0, 15].SetFill(PatternFill.CreateSolidFill(ThemableColor.FromArgb(255, 0, 0, 255)));
            worksheet.Cells[1, 15, 1, 15].SetFill(PatternFill.CreateSolidFill(ThemableColor.FromArgb(255, 255, 0, 0)));
            worksheet.Cells[2, 15, 2, 15].SetFill(PatternFill.CreateSolidFill(ThemableColor.FromArgb(255, 0, 255, 0)));
            FillColorSortCondition sortCondition = new FillColorSortCondition(0, PatternFill.CreateSolidFill(ThemableColor.FromArgb(255, 255, 0, 0)), SortOrder.Ascending);
            worksheet.Cells[0, 15, 3, 15].Sort(sortCondition);

            // Shapes
            FloatingShapeBase image = CreateImage(worksheet);
            worksheet.Shapes.Add(image);

            // Names
            worksheet.Names.Add(worksheet.Name, "=10", new CellIndex(1, 1));

            // Hyperlinks
            worksheet.Hyperlinks.Add(new CellIndex(15, 15), HyperlinkInfo.CreateHyperlink("www.telerik.com"));

            // Direct properties
            worksheet.DefaultColumnWidth = new ColumnWidth(100, true);
            worksheet.DefaultRowHeight = new RowHeight(100, true);
            worksheet.Visibility = SheetVisibility.VeryHidden;
        }

        private static FloatingImage CreateImage(Worksheet worksheet)
        {
            FloatingImage image = new FloatingImage(worksheet, new CellIndex(0, 0), 0, 0);
            Stream stream = TestHelperInternal.GetStreamFromFile("Penguins.jpg");
            using (stream)
            {
                image.ImageSource = new Media.ImageSource(stream, "jpg");
            }
            image.Width = 90;
            image.Height = 90;

            return image;
        }
    }
}