﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Telerik.Documents.SpreadsheetStreaming.Core
{
    internal abstract class ChildEntityManagerBase<T>
        where T : IChildEntry
    {
        private readonly Dictionary<Type, Tuple<int, Type>> typeToRegisterIndex;
        private readonly Dictionary<Type, T> typeToInstance;
        private readonly Dictionary<Type, Func<T>> typeToFactory;
        private readonly Dictionary<int, T> indexToInstance;
        private int currentChildIndex;

        public ChildEntityManagerBase()
        {
            this.typeToRegisterIndex = new Dictionary<Type, Tuple<int, Type>>();
            this.typeToInstance = new Dictionary<Type, T>();
            this.typeToFactory = new Dictionary<Type, Func<T>>();
            this.indexToInstance = new Dictionary<int, T>();
            this.currentChildIndex = -1;
        }

        public void EnsureCanGetElement<TResult>()
            where TResult : T
        {
            Type type = typeof(TResult);

            Tuple<int, Type> childInfo = this.typeToRegisterIndex[type];
            int childIndex = childInfo.Item1;
            Type dependOnType = childInfo.Item2;

            this.ValidateExportOrder(dependOnType, childIndex);
            this.EnsureEndUsingPreviousChildInstance(childIndex);
            this.ValidatePreviousInstanceUsage(type);
        }

        public abstract TResult GetRegisteredChild<TResult>()
            where TResult : T;

        public bool TryGetLastChild<TResult>(out TResult child)
            where TResult : T
        {
            return this.TryGetChildAtIndex(this.currentChildIndex, out child);
        }

        public void RegisterChild<TResult>(Func<T> factory)
            where TResult : T
        {
            this.RegisterChild<TResult>(factory, null);
        }

        public void RegisterChild<TResult>(Func<T> factory, Type dependOnType)
            where TResult : T
        {
            int index = this.typeToRegisterIndex.Count;

            Tuple<int, Type> childInfo = new Tuple<int, Type>(index, dependOnType);
            this.typeToRegisterIndex.Add(typeof(TResult), childInfo);
            this.typeToFactory.Add(typeof(TResult), factory);
        }

        public int GetRegisteredChildIndex(Type type)
        {
            if (type == null ||
                !this.typeToRegisterIndex.ContainsKey(type))
            {
                return -1;
            }

            return this.typeToRegisterIndex[type].Item1;
        }

        protected TResult GetRegisteredChild<TResult>(bool reuseChild)
            where TResult : T
        {
            Type type = typeof(TResult);

            Tuple<int, Type> childInfo = this.typeToRegisterIndex[type];
            int childIndex = childInfo.Item1;
            Type dependOnType = childInfo.Item2;

            this.ValidateExportOrder(dependOnType, childIndex);
            this.EnsureEndUsingPreviousChildInstance(childIndex);

            this.currentChildIndex = childIndex;

            TResult instance;
            if (reuseChild && this.indexToInstance.ContainsKey(this.currentChildIndex))
            {
                instance = (TResult)this.indexToInstance[this.currentChildIndex];
            }
            else
            {
                instance = this.CreateChild<TResult>();
                this.indexToInstance[this.currentChildIndex] = instance;
            }

            return instance;
        }

        protected abstract void EnsureEndUsingChild(T previousInstance);

        private TResult CreateChild<TResult>()
            where TResult : T
        {
            Type type = typeof(TResult);

            this.ValidatePreviousInstanceUsage(type);

            TResult instance = (TResult)this.typeToFactory[type]();
            this.typeToInstance[type] = instance;

            return instance;
        }

        private void ValidatePreviousInstanceUsage(Type type)
        {
            if (this.typeToInstance.ContainsKey(type))
            {
                T oldInstance = this.typeToInstance[type];
                if (oldInstance.IsUsageBegan && !oldInstance.IsUsageCompleted)
                {
                    throw new InvalidOperationException("Can not start using new instance before stop using the previous.");
                }
            }
        }
 
        private void EnsureEndUsingPreviousChildInstance(int childIndex)
        {
            if (this.currentChildIndex >= 0 && this.currentChildIndex != childIndex)
            {
                T previousInstance = (T)this.indexToInstance[this.currentChildIndex];
                this.EnsureEndUsingChild(previousInstance);
            }
        }
 
        private void ValidateExportOrder(Type dependOnType, int childIndex)
        {
            bool dependencyIsNotExported = false;

            if (dependOnType != null && this.typeToRegisterIndex.ContainsKey(dependOnType))
            {
                Tuple<int, Type> dependency = this.typeToRegisterIndex[dependOnType];
                dependencyIsNotExported = !this.indexToInstance.ContainsKey(dependency.Item1);
            }

            if (childIndex < this.currentChildIndex || dependencyIsNotExported)
            {
                throw new InvalidOperationException("Invalid export order.");
            }
        }

        private bool TryGetChildAtIndex<TResult>(int index, out TResult child)
            where TResult : T
        {
            child = default(TResult);

            if (this.indexToInstance.ContainsKey(index))
            {
                child = (TResult)this.indexToInstance[index];
                return true;
            }

            return false;
        }
    }
}