﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Xml;
using Telerik.Documents.SpreadsheetStreaming.Core;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Core;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Model;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Model.Elements.Styles;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Model.Parts;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Model.Writers;
using Telerik.Documents.SpreadsheetStreaming.Model.Formatting;
using Telerik.Documents.SpreadsheetStreaming.Model.Relations;
#if !NETSTANDARD
using Telerik.Windows.Zip;
#else
using Telerik.Zip;
#endif

namespace Telerik.Documents.SpreadsheetStreaming.Exporters.Xlsx
{
    internal class XlsxAppendWorkbookExporter : XlsxWorkbookExporterBase
    {
        private readonly ZipArchive sourceArchive;
        private XmlDocument stylesDocument;
        private XmlDocument workbookDocument;
        private StylesRepository importedStyles;
        private List<Relationship> importedRelationships;

        public XlsxAppendWorkbookExporter(Stream source, Stream target)
            : base(target)
        {
            this.sourceArchive = new ZipArchive(source, ZipArchiveMode.Read, true, null);
            this.ReadContentTypesPart();
            this.ReadWorkbookRelationshipsPart();
            this.ReadStylesPart();
            this.ReadWorkbookPart();

            string[] partPathsToBeGenerated = new string[]
            {
                PartPaths.ContentTypesPath,
                PartPaths.StylesPartPath,
                PartPaths.WorkbookPartPath,
                PartPaths.WorkbookRelationshipsPartPath,
            };

            foreach (ZipArchiveEntry sourceEntry in this.sourceArchive.Entries)
            {
                if (partPathsToBeGenerated.Contains(sourceEntry.FullName))
                {
                    continue;
                }

                using (Stream sourceEntryStream = sourceEntry.Open())
                {
                    using (ZipArchiveEntry targetEntry = this.TargetArchive.CreateEntry(sourceEntry.FullName))
                    {
                        using (Stream targetEntryStream = targetEntry.Open())
                        {
                            sourceEntryStream.CopyTo(targetEntryStream);
                        }
                    }
                }
            }
        }

        public override bool IsAppender
        {
            get
            {
                return true;
            }
        }

        protected override StylesRepository InitializeStylesRepository()
        {
            this.importedStyles = new StylesRepository();
            return new StylesRepository();
        }

        protected override void WriteWorkbookPart()
        {
            using (Stream stream = new MemoryStream())
            {
                OpenXmlWriter writer = new OpenXmlWriter(stream);
                ElementContext context = new ElementContext(writer, null, this.Theme);

                IEnumerable<Relationship> worksheetRelationships = this.WorkbookRelationships.GetRelationships(XlsxRelationshipTypes.WorksheetRelationshipType);
                IEnumerable<Relationship> relationships = GetRelationshipsDifference(this.importedRelationships, worksheetRelationships);

                WorkbookElementWriter workbookElement = new WorkbookElementWriter();
                workbookElement.SetContext(context);
                workbookElement.Write(relationships.ToList());

                writer.Flush();

                stream.Seek(0, SeekOrigin.Begin);

                var targetDocument = this.workbookDocument;
                XmlDocument newDocument = new XmlDocument();
                newDocument.Load(stream);

                MergeDocuments(targetDocument, newDocument, false);
            }

            using (ZipArchiveEntry entry = this.TargetArchive.CreateEntry(PartPaths.WorkbookPartPath))
            {
                using (Stream stream = entry.Open())
                {
                    this.workbookDocument.Save(stream);
                }
            }
        }

        protected override void WriteStylesPart()
        {
            using (Stream stream = new MemoryStream())
            {
                OpenXmlWriter writer = new OpenXmlWriter(stream);
                ElementContext context = new ElementContext(writer, null, this.Theme);

                StylesRepository repository = GetStylesRepositoryDifference(this.importedStyles, this.StylesRepository);

                StyleSheetElement stylesElement = new StyleSheetElement();
                stylesElement.SetContext(context);
                stylesElement.Write(repository);

                writer.Flush();

                stream.Seek(0, SeekOrigin.Begin);

                XmlDocument newDocument = new XmlDocument();
                newDocument.Load(stream);

                var targetDocument = this.stylesDocument;
                MergeDocuments(targetDocument, newDocument, false);
            }

            using (var entry = this.TargetArchive.CreateEntry(PartPaths.StylesPartPath))
            {
                using (Stream stream = entry.Open())
                {
                    this.stylesDocument.Save(stream);
                }
            }
        }

        private static void MergeDocuments(XmlDocument targetDocument, XmlDocument sourceDocument, bool shouldUpdateChildrenCount)
        {
            foreach (object node in targetDocument.DocumentElement.ChildNodes)
            {
                XmlElement element = node as XmlElement;
                if (element != null)
                {
                    foreach (XmlElement sourceElement in sourceDocument.GetElementsByTagName(element.LocalName).Cast<XmlElement>())
                    {
                        foreach (XmlElement childNode in sourceElement.ChildNodes.Cast<XmlElement>())
                        {
                            XmlNode newChild = targetDocument.ImportNode(childNode, true);
                            newChild.Prefix = element.Prefix;
                            element.AppendChild(newChild);

                            if (shouldUpdateChildrenCount)
                            {
                                element.SetAttribute("count", element.ChildNodes.Count.ToString());
                            }
                        }
                    }
                }
            }
        }

        private static void CopyStyles(StylesRepository source, StylesRepository target)
        {
            foreach (var item in source.Borders)
            {
                target.AddBorders(item);
            }

            foreach (var item in source.CellFormats)
            {
                target.AddCellFormat(item);
            }

            foreach (var item in source.CellStyles)
            {
                target.AddCellStyle(item);
            }

            foreach (var item in source.CellStyleFormats)
            {
                target.AddCellStyleFormat(item);
            }

            foreach (var item in source.Fills)
            {
                target.AddFill(item);
            }

            foreach (var item in source.Fonts)
            {
                target.AddFont(item);
            }

            foreach (var item in source.NumberFormats)
            {
                target.AddNumberFormat(item);
            }
        }

        private static StylesRepository GetStylesRepositoryDifference(StylesRepository initial, StylesRepository updated)
        {
            StylesRepository result = new StylesRepository();

            foreach (var item in updated.Borders)
            {
                if (!initial.Borders.Contains(item))
                {
                    result.AddBorders(item);
                }
            }

            foreach (var item in updated.CellFormats)
            {
                if (!initial.CellFormats.Contains(item))
                {
                    result.AddCellFormat(item);
                }
            }

            foreach (var item in updated.CellStyles)
            {
                if (!initial.CellStyles.Contains(item))
                {
                    result.AddCellStyle(item);
                }
            }

            foreach (var item in updated.CellStyleFormats)
            {
                if (!initial.CellStyleFormats.Contains(item))
                {
                    result.AddCellStyleFormat(item);
                }
            }

            foreach (var item in updated.Fills)
            {
                if (!initial.Fills.Contains(item))
                {
                    result.AddFill(item);
                }
            }

            foreach (var item in updated.Fonts)
            {
                if (!initial.Fonts.Contains(item))
                {
                    result.AddFont(item);
                }
            }

            foreach (var item in updated.NumberFormats)
            {
                if (!initial.NumberFormats.Contains(item))
                {
                    result.AddNumberFormat(item);
                }
            }

            return result;
        }

        private static IEnumerable<Relationship> GetRelationshipsDifference(List<Relationship> initial, IEnumerable<Relationship> updated)
        {
            foreach (Relationship item in updated)
            {
                if (!initial.Where(p => p.RelationshipId == item.RelationshipId).Any())
                {
                    yield return item;
                }
            }
        }
        private void ReadContentTypesPart()
        {
            using (ContentTypesPart part = new ContentTypesPart(new PartContext(this.sourceArchive)))
            {
                part.Read(this.ContentTypes);
            }
        }

        private void ReadWorkbookRelationshipsPart()
        {
            using (WorkbookRelationshipsPart part = new WorkbookRelationshipsPart(new PartContext(this.sourceArchive)))
            {
                part.ReadRelationships(this.WorkbookRelationships);
            }
        }

        private void ReadStylesPart()
        {
            using (StylesPart part = new StylesPart(new PartContext(this.sourceArchive, this.Theme)))
            {
                part.Read(this.StylesRepository);
            }

            CopyStyles(this.StylesRepository, this.importedStyles);

            this.stylesDocument = new XmlDocument();
            using (ZipArchiveEntry entry = this.sourceArchive.Entries.Where(p => p.FullName == PartPaths.StylesPartPath).FirstOrDefault())
            {
                using (Stream stream = entry.Open())
                {
                    this.stylesDocument.Load(stream);
                }

                foreach (object node in this.stylesDocument.DocumentElement.ChildNodes)
                {
                    XmlElement element = node as XmlElement;
                    if (element != null)
                    {
                        ReadStylesPartElement(element);
                    }
                }
            }
        }

        private void ReadStylesPartElement(XmlElement element)
        {
            int count = element.ChildNodes.Count;
            switch (element.LocalName)
            {
                case XlsxElementNames.NumberFormatsElementName:
                    {
                        int maxNumberFormatId = count;
                        foreach (var child in element.ChildNodes.Cast<XmlElement>())
                        {
                            string value = child.GetAttribute("numFmtId");
                            if (!string.IsNullOrEmpty(value))
                            {
                                maxNumberFormatId = Math.Max(maxNumberFormatId, int.Parse(value) + 1);
                            }
                        }

                        this.StylesRepository.NumberFormatsStartIndex = maxNumberFormatId;
                    }
                    break;

                case XlsxElementNames.FontsElementName:
                    {
                        this.StylesRepository.FontsStartIndex = count;
                    }
                    break;

                case XlsxElementNames.FillsElementName:
                    {
                        this.StylesRepository.FillsStartIndex = count;
                    }
                    break;

                case XlsxElementNames.BordersElementName:
                    {
                        this.StylesRepository.BordersStartIndex = count;
                    }
                    break;

                case XlsxElementNames.CellStylesElementName:
                    {
                        this.StylesRepository.CellStylesStartIndex = count;
                    }
                    break;

                case XlsxElementNames.CellStyleFormatsElementName:
                    {
                        this.StylesRepository.CellStyleFormatsStartIndex = count;
                    }
                    break;

                case XlsxElementNames.CellFormatsElementName:
                    {
                        this.StylesRepository.CellFormatsStartIndex = count;
                    }
                    break;
            }
        }

        private void ReadWorkbookPart()
        {
            List<Relationship> relationships = new List<Relationship>();
            using (WorkbookPart part = new WorkbookPart(new PartContext(this.sourceArchive)))
            {
                part.Read(relationships);
            }

            foreach (Relationship relationship in relationships)
            {
                this.WorkbookRelationships.AddRelationship(relationship);
            }

            this.importedRelationships = relationships;

            this.workbookDocument = new XmlDocument();
            using (var entry = this.sourceArchive.Entries.Where(p => p.FullName == PartPaths.WorkbookPartPath).FirstOrDefault())
            {
                using (Stream stream = entry.Open())
                {
                    this.workbookDocument.Load(stream);
                }
            }
        }
    }
}
