﻿using System;
using System.Globalization;
using System.Linq;
using Telerik.Documents.SpreadsheetStreaming.Core;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Core.Types;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Model.Elements.Worksheet;
using Telerik.Documents.SpreadsheetStreaming.Model.Formatting;
using Telerik.Documents.SpreadsheetStreaming.Utilities;

namespace Telerik.Documents.SpreadsheetStreaming.Exporters.Xlsx
{
    internal sealed class XlsxCellExporter : EntityBase, ICellExporter
    {
        private readonly StylesRepository stylesRepository;
        private readonly int rowIndex;
        private readonly int columnIndex;
        private readonly CellElement cellElement;
        private string value;
        private CellValueType? valueType;
        private int? cellStyleId;
        private XlsxRowExporter row;

        internal XlsxCellExporter(XlsxRowExporter row, RowElement rowElement, StylesRepository stylesRepository, int rowIndex, int columnIndex)
        {
            Guard.ThrowExceptionIfGreaterThan(DefaultValues.ColumnCount - 1, columnIndex, "columnIndex");

            this.row = row;
            this.stylesRepository = stylesRepository;
            this.rowIndex = rowIndex;
            this.columnIndex = columnIndex;

            this.cellElement = rowElement.CreateCellElementWriter();
        }

        public void SetValue(string value)
        {
            Guard.ThrowExceptionIfNull(value, "value");

            this.value = value;

            // TODO: this have to be saved as CellValueType.Value using shared strings
            this.valueType = CellValueType.Text;
        }

        public void SetValue(double value)
        {
            Guard.ThrowExceptionIfNaN(value, "value");
            Guard.ThrowExceptionIfInfinity(value, "value");

            this.value = value.ToString(CultureInfo.InvariantCulture);
        }

        public void SetValue(bool value)
        {
            // TODO: Find the other ToString may be they need to use CultureInfo.InvariantCulture
            this.value = (value ? 1 : 0).ToString(CultureInfo.InvariantCulture);
            this.valueType = CellValueType.Boolean;
        }

        public void SetValue(DateTime value)
        {
            this.value = value.ToString("s", CultureInfo.InvariantCulture);
            this.valueType = CellValueType.Date;
        }

        public void SetFormula(string value)
        {
            Guard.ThrowExceptionIfNull(value, "value");

            this.value = value;
            this.valueType = CellValueType.Formula;
        }

        public void SetFormat(SpreadCellFormat cellFormat)
        {
            Guard.ThrowExceptionIfNull(cellFormat, "cellFormat");

            this.cellStyleId = this.stylesRepository.RegisterCellFormat(cellFormat);
        }

        // TODO: Fully implement this. CellValueType contains info where in the specification is this described.
        internal sealed override void CompleteWriteOverride()
        {
            if (this.value == null && !this.cellStyleId.HasValue)
            {
                return;
            }

            this.cellElement.Reference = new CellRef(this.rowIndex, this.columnIndex);

            if (this.cellStyleId.HasValue)
            {
                this.cellElement.StyleIndex = this.cellStyleId.Value;
            }

            if (this.valueType.HasValue && this.valueType.Value != CellValueType.Formula)
            {
                this.cellElement.CellDataType = CellTypesMapper.GetCellTypeName(this.valueType.Value);
            }

            this.cellElement.EnsureWritingStarted();

            if (this.valueType.HasValue)
            {
                switch (this.valueType)
                {
                    case CellValueType.Date:
                    case CellValueType.Boolean:
                    case CellValueType.Text:
                        this.cellElement.WriteValue(this.value);
                        break;

                    case CellValueType.Formula:
                        this.cellElement.WriteFormula(this.value);
                        break;

                    case CellValueType.Error:
                    case CellValueType.RichText:
                    case CellValueType.Number:
                    case CellValueType.SharedString:

                    default:
                        throw new NotImplementedException();
                }
            }
            else if (this.value != null)
            {
                this.cellElement.WriteValue(this.value);
            }

            this.cellElement.EnsureWritingEnded();
        }

        internal override void DisposeOverride()
        {
            this.row.NotifyCellDisposing();
            this.row = null;
            base.DisposeOverride();
        }
    }
}