﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Core;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Model.Parts;
using Telerik.Documents.SpreadsheetStreaming.Model.Formatting;
using Telerik.Documents.SpreadsheetStreaming.Model.Relations;
using Telerik.Documents.SpreadsheetStreaming.Model.Themes;
using Telerik.Documents.SpreadsheetStreaming.Utilities;

#if NETSTANDARD
using Telerik.Zip;
#else
using Telerik.Windows.Zip;
#endif

namespace Telerik.Documents.SpreadsheetStreaming.Exporters.Xlsx
{
    internal abstract class XlsxWorkbookExporterBase : WorkbookExporterBase
    {
        private static readonly List<char> forbiddenSheetNameSymbols = new List<char>(new char[] { '\\', '/', '?', '*', '[', ']', ':' });
        private static readonly int maxSheetNameLength = 31;
        private readonly Relationships workbookRelationships;
        private readonly ContentTypesRepository contentTypes;
        private readonly ZipArchive targetArchive;
        private readonly SpreadCellStyleCollection cellStyles;
        private readonly SpreadTheme theme;
        private StylesRepository stylesRepository;

        internal XlsxWorkbookExporterBase(Stream stream)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");

            this.targetArchive = new ZipArchive(stream, ZipArchiveMode.Create, true, null);
            this.workbookRelationships = new Relationships(this.IsAppender);
            this.contentTypes = new ContentTypesRepository();
            this.theme = SpreadPredefinedThemeSchemes.DefaultTheme;
            this.cellStyles = new SpreadCellStyleCollection();

            Telerik.Documents.SpreadsheetStreaming.Licensing.LicenseCheck.Validate(this);
        }

        public virtual bool IsAppender
        {
            get
            {
                return false;
            }
        }

        public override SpreadCellStyleCollection CellStyles
        {
            get
            {
                return this.cellStyles;
            }
        }

        protected ZipArchive TargetArchive
        {
            get
            {
                return this.targetArchive;
            }
        }

        protected Relationships WorkbookRelationships
        {
            get
            {
                return this.workbookRelationships;
            }
        }

        protected ContentTypesRepository ContentTypes
        {
            get
            {
                return this.contentTypes;
            }
        }

        protected StylesRepository StylesRepository
        {
            get
            {
                if (this.stylesRepository == null)
                {
                    this.stylesRepository = this.InitializeStylesRepository();
                }

                return this.stylesRepository;
            }
        }

        protected SpreadTheme Theme
        {
            get
            {
                return this.theme;
            }
        }

        public sealed override IWorksheetExporter CreateWorksheetExporterOverride(string name)
        {
            this.ValidateSheetName(name);

            int id = this.workbookRelationships.AddWorksheet(name);

            return new XlsxWorksheetExporter(this, id, this.StylesRepository, this.contentTypes, this.targetArchive);
        }

        public override IEnumerable<SheetInfo> GetSheetInfos()
        {
            return this.WorkbookRelationships.GetRelationships(XlsxRelationshipTypes.WorksheetRelationshipType)
                .OrderBy(relationship => relationship.Index)
                .Select(relationship => new SheetInfo(relationship.Name));
        }

        internal sealed override void DisposeOverride()
        {
            this.targetArchive.Dispose();

            base.DisposeOverride();
        }

        internal sealed override void CompleteWriteOverride()
        {
            bool hasWorksheets = this.WorkbookRelationships.GetRelationships(XlsxRelationshipTypes.WorksheetRelationshipType).Any();

            if (!hasWorksheets)
            {
                throw new InvalidOperationException("The workbook must have at least one worksheet.");
            }

            this.WriteWorkbookPart();
            this.WriteStylesPart();
            this.WriteWorkbookRelationshipsPart();
            ResourcesHelper.CopyResourcesToZipArchive(this.TargetArchive, this.ContentTypes);
            this.WriteContentTypesPart();
        }

        protected abstract StylesRepository InitializeStylesRepository();

        protected virtual void WriteWorkbookPart()
        {
            using (WorkbookPart part = new WorkbookPart(new PartContext(this.targetArchive)))
            {
                this.contentTypes.Register(part);
                IEnumerable<Relationship> relationships = this.workbookRelationships.GetRelationships(XlsxRelationshipTypes.WorksheetRelationshipType);
                var orderedRelationships = relationships.OrderBy(p => p.Id);
                List<Relationship> worksheetRelationships = orderedRelationships.ToList();
                part.WriteWorksheetRelationships(worksheetRelationships);
            }
        }

        protected virtual void WriteStylesPart()
        {
            using (StylesPart part = new StylesPart(new PartContext(this.targetArchive, this.theme)))
            {
                this.contentTypes.Register(part);
                part.Write(this.StylesRepository);
            }
        }

        protected void WriteWorkbookRelationshipsPart()
        {
            using (WorkbookRelationshipsPart part = new WorkbookRelationshipsPart(new PartContext(this.targetArchive)))
            {
                this.contentTypes.Register(part);
                part.WriteRelationships(this.workbookRelationships);
            }
        }

        protected void WriteContentTypesPart()
        {
            using (ContentTypesPart part = new ContentTypesPart(new PartContext(this.targetArchive)))
            {
                part.Write(contentTypes);
            }
        }

        private void ValidateSheetName(string sheetName)
        {
            if (this.workbookRelationships.ContainsWorksheetName(sheetName))
            {
                throw new ArgumentException(string.Format("Sheet with name \"{0}\" already exists.", sheetName));
            }

            if (sheetName.Length >= maxSheetNameLength)
            {
                throw new ArgumentException(string.Format("Sheet name should be limited to 31 characters."));
            }

            foreach (char symbol in sheetName)
            {
                if (forbiddenSheetNameSymbols.Contains(symbol))
                {
                    string forbiddenSymbolsExceptionMessage = string.Join(", ", forbiddenSheetNameSymbols);
                    throw new ArgumentException(string.Format("Sheet name contains at least one of the forbidden symbols: ({0}).", forbiddenSymbolsExceptionMessage));
                }
            }
        }
    }
}