﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Documents.SpreadsheetStreaming.Core;
using Telerik.Documents.SpreadsheetStreaming.Exporters.Xlsx;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Core;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Model.Elements.Worksheet;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Model.Parts;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Model.Writers.Worksheet;
using Telerik.Documents.SpreadsheetStreaming.Model;
using Telerik.Documents.SpreadsheetStreaming.Model.Formatting;
using Telerik.Documents.SpreadsheetStreaming.Utilities;

#if NETSTANDARD
using Telerik.Zip;
#else
using Telerik.Windows.Zip;
#endif

namespace Telerik.Documents.SpreadsheetStreaming
{
    internal sealed class XlsxWorksheetExporter : EntityBase, IWorksheetExporter
    {
        private readonly ExporterChildManager childManager;
        private readonly StylesRepository stylesRepository;
        private readonly WorksheetPartWriter part;
        private readonly List<SpreadCellRange> mergedCellsRanges;
        private XlsxWorkbookExporterBase workbook;
        private Queue<ColumnRange> columnRangesQueue;
        private bool hasSheetView;

        internal XlsxWorksheetExporter(XlsxWorkbookExporterBase workbook, int id, StylesRepository stylesRepository, ContentTypesRepository contentTypes, ZipArchive archive)
        {
            this.workbook = workbook;
            this.stylesRepository = stylesRepository;
            this.part = new WorksheetPartWriter(new PartContext(archive, id.ToString()));
            contentTypes.Register(this.part);
            this.mergedCellsRanges = new List<SpreadCellRange>();

            this.childManager = new ExporterChildManager();
            this.childManager.RegisterChild<XlsxWorksheetViewExporter>(this.CreateViewExporter);
            this.childManager.RegisterChild<XlsxColumnExporter>(this.CreateColumnExproter);
            this.childManager.RegisterChild<XlsxRowExporter>(this.CreateRowExproter);
            this.childManager.RegisterChild<XlsxMergedCellsExporter>(this.CreateMergedCellsExproter);
        }

        internal int RowIndex { get; set; }

        internal int ColumnIndex { get; set; }

        public IWorksheetViewExporter CreateWorksheetViewExporter()
        {
            if (this.hasSheetView)
            {
                throw new InvalidOperationException("Only one sheet view can be exported.");
            }

            XlsxWorksheetViewExporter sheetViewExporter = this.childManager.GetRegisteredChild<XlsxWorksheetViewExporter>();

            this.hasSheetView = true;

            return sheetViewExporter;
        }

        public IColumnExporter CreateColumnExporter()
        {
            this.childManager.EnsureCanGetElement<XlsxColumnExporter>();

            if (this.columnRangesQueue == null)
            {
                this.columnRangesQueue = new Queue<ColumnRange>();
            }

            XlsxColumnExporter column = this.childManager.GetRegisteredChild<XlsxColumnExporter>();
            this.ColumnIndex++;

            return column;
        }

        public void SkipColumns(int count)
        {
            Guard.ThrowExceptionIfLessThan(0, count, "count");

            this.childManager.EnsureCanGetElement<XlsxColumnExporter>();

            this.ColumnIndex += count;
        }

        public IRowExporter CreateRowExporter()
        {
            this.childManager.EnsureCanGetElement<XlsxRowExporter>();

            this.EnsureColumsAreExported();

            XlsxRowExporter row = this.childManager.GetRegisteredChild<XlsxRowExporter>();
            this.RowIndex++;

            return row;
        }

        public void SkipRows(int count)
        {
            Guard.ThrowExceptionIfLessThan(0, count, "count");

            this.childManager.EnsureCanGetElement<XlsxRowExporter>();

            this.RowIndex += count;
        }

        public void MergeCells(int fromRowIndex, int fromColumnIndex, int toRowIndex, int toColumnIndex)
        {
            using (var mergedCellsExporter = this.childManager.GetRegisteredChild<XlsxMergedCellsExporter>())
            {
                mergedCellsExporter.MergeCells(fromRowIndex, fromColumnIndex, toRowIndex, toColumnIndex);
            }
        }

        internal void NotifyColumnDisposing()
        {
            if (this.columnRangesQueue.Count > 1)
            {
                this.WriteColumn(this.columnRangesQueue.Dequeue());
            }
        }

        internal override void CompleteWriteOverride()
        {
            if (this.RowIndex == 0)
            {
                throw new InvalidOperationException("The worksheet must have at least one row.");
            }

            base.CompleteWriteOverride();
        }

        internal sealed override void DisposeOverride()
        {
            this.EnsureColumsAreExported();
            this.workbook.NotifyWorksheetDisposing();
            this.workbook = null;

            this.part.Dispose();
        }

        private XlsxWorksheetViewExporter CreateViewExporter()
        {
            SheetViewsElement sheetViews = this.part.RootElement.GetSheetViewsElementWriter();

            return new XlsxWorksheetViewExporter(sheetViews);
        }

        private XlsxColumnExporter CreateColumnExproter()
        {
            return new XlsxColumnExporter(this.columnRangesQueue, this.ColumnIndex, this.NotifyColumnDisposing);
        }

        private XlsxRowExporter CreateRowExproter()
        {
            SheetDataElement sheetData = this.part.RootElement.GetSheetDataElementWriter();
            sheetData.EnsureWritingStarted();

            return new XlsxRowExporter(sheetData, this.stylesRepository, this.RowIndex);
        }

        private XlsxMergedCellsExporter CreateMergedCellsExproter()
        {
            MergedCellsElementWriter mergedCellsCollectionWriter = this.part.RootElement.GetMergedCellsElementWriter();
            mergedCellsCollectionWriter.EnsureWritingStarted();

            return new XlsxMergedCellsExporter(mergedCellsCollectionWriter, this.mergedCellsRanges);
        }

        private void EnsureColumsAreExported()
        {
            if (this.columnRangesQueue != null)
            {
                while (this.columnRangesQueue.Count > 0)
                {
                    ColumnRange column = columnRangesQueue.Dequeue();
                    this.WriteColumn(column);
                }
            }
        }

        private void WriteColumn(ColumnRange columnRange)
        {
            ColumnsElement columns = this.part.RootElement.GetColumnsElementWriter();
            columns.EnsureWritingStarted();

            var columnElement = columns.CreateColumnElementWriter();
            columnElement.Write(columnRange);
        }
    }
}