﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using Telerik.Documents.SpreadsheetStreaming.Utilities;

namespace Telerik.Documents.SpreadsheetStreaming.ImportExport
{
    internal class ColorsHelper
    {
        private static readonly List<TintValues> tintValuesList = new List<TintValues>
        {
            new TintValues(0, 0, new double[] { 0.5, 0.35, 0.25, 0.15, 0.05 }),
            new TintValues(1, 19, new double[] { 0.9, 0.75, 0.5, 0.25, 0.1 }),
            new TintValues(20, 79, new double[] { 0.80, 0.60, 0.40, -0.25, -0.5 }),
            new TintValues(80, 99, new double[] { -0.10, -0.25, -0.50, -0.75, -0.9 }),
            new TintValues(100, 100, new double[] { -0.05, -0.15, -0.25, -0.35, -0.5 }),
        };

        public static SpreadColor HexStringToColor(string hexColor)
        {
            // Remove the # at the front.
            if (hexColor.StartsWith("#"))
            {
                hexColor = hexColor.Substring(1, hexColor.Length - 1);
            }

            byte a = 255;
            byte r = 255;
            byte g = 255;
            byte b = 255;

            int start = 0;

            // Handle ARGB strings (8 characters long).
            if (hexColor.Length == 8)
            {
                a = byte.Parse(hexColor.Substring(0, 2), NumberStyles.HexNumber);
                start = 2;
            }

            // Handle contracted RGB strings (3 characters long)
            if (hexColor.Length == 3)
            {
                hexColor = string.Format("{0}{0}{1}{1}{2}{2}", hexColor[0], hexColor[1], hexColor[2]);
            }

            // Convert RGB characters to bytes.
            r = byte.Parse(hexColor.Substring(start, 2), NumberStyles.HexNumber);
            g = byte.Parse(hexColor.Substring(start + 2, 2), NumberStyles.HexNumber);
            b = byte.Parse(hexColor.Substring(start + 4, 2), NumberStyles.HexNumber);

            return new SpreadColor(a, r, g, b);
        }

        public static SpreadColorShadeType? GetColorShadeTypeForColorAndTint(SpreadColor color, double tint)
        {
            HlsaColor hlsColor = HlsaColor.FromColor(color);
            var array = GetTintValuesForHlsaColor(hlsColor);

            double roundedTint = Math.Round(tint, 2);
            for (int i = 0; i < array.Length; i++)
            {
                double t = Math.Round(array[i], 2);
                if (ObjectExtensions.EqualsDouble(t, roundedTint))
                {
                    return (SpreadColorShadeType)i;
                }
            }

            return null;
        }

        public static double GetTintAndShadeForColorAndIndex(SpreadColor color, SpreadColorShadeType colorShadeType)
        {
            HlsaColor hlsColor = HlsaColor.FromColor(color);
            double[] array = GetTintValuesForHlsaColor(hlsColor);

            return array[(int)colorShadeType];
        }

        public static SpreadColor UpdateTint(SpreadColor color, double tint)
        {
            if (tint != 0)
            {
                HlsaColor hlsColor = HlsaColor.FromColor(color);
                int colorLuminosity = (int)(hlsColor.Luminance * 240);
                if (tint <= 0f)
                {
                    double luminosity = colorLuminosity * -tint;

                    return ColorFromHlsa(hlsColor.Hue, (colorLuminosity - luminosity) / 240, hlsColor.Saturation);
                }
                else
                {
                    int num = 240 - colorLuminosity;

                    return ColorFromHlsa(hlsColor.Hue, (colorLuminosity + (num * tint)) / 240, hlsColor.Saturation);
                }
            }
            else
            {
                return color;
            }
        }

        public static SpreadColor ColorFromHlsa(double hue, double luminosity, double saturation)
        {
            return HlsaColor.ToColor(new HlsaColor(hue, luminosity, saturation, 1));
        }

        private static double[] GetTintValuesForHlsaColor(HlsaColor color)
        {
            int luminosityInPercents = (int)(color.Luminance * 100);

            foreach (TintValues values in tintValuesList)
            {
                if (luminosityInPercents >= values.Start &&
                    luminosityInPercents <= values.End)
                {
                    return values.TintValuesArray;
                }
            }

            return null;
        }

        private class TintValues
        {
            private readonly int start;
            private readonly int end;
            private readonly double[] tintValues;

            public TintValues(int start, int end, double[] tintValues)
            {
                this.start = start;
                this.end = end;
                this.tintValues = tintValues;
            }

            public int Start
            {
                get
                {
                    return this.start;
                }
            }

            public int End
            {
                get
                {
                    return this.end;
                }
            }

            public double[] TintValuesArray
            {
                get
                {
                    return this.tintValues;
                }
            }
        }
    }
}