﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace Telerik.Documents.SpreadsheetStreaming.ImportExport.Core
{
    internal abstract class ConsecutiveElementBase : ElementBase
    {
        private ElementChildManager childElementsManager;

        public ConsecutiveElementBase()
        {
            this.childElementsManager = new ElementChildManager();
        }

        public void EnsureWritingStarted()
        {
            if (!this.IsWritingStarted)
            {
                this.BeginWriteElement();
            }
        }

        public void EnsureWritingEnded()
        {
            if (this.IsWritingStarted && !this.IsWritingEnded)
            {
                ConsecutiveElementBase instance = null;
                if (this.childElementsManager.TryGetLastChild(out instance))
                {
                    instance.EnsureWritingEnded();
                }

                this.EndWriteElement();
            }
        }

        public void BeginWriteElement()
        {
            if (this.IsWritingStarted)
            {
                throw new InvalidOperationException("Can not start already started element.");
            }

            this.BeginWrite();
        }

        public void EndWriteElement()
        {
            if (!this.IsWritingStarted)
            {
                throw new InvalidOperationException("Can not end an element before starting it.");
            }

            if (this.IsWritingEnded)
            {
                throw new InvalidOperationException("Can not end already ended element.");
            }

            this.EndWrite();
        }

        public void BeginReadElement()
        {
            if (this.IsReadingStarted)
            {
                throw new InvalidOperationException("Can not start already started element.");
            }

            this.BeginRead();
        }

        public void EndReadElement()
        {
            if (!this.IsReadingStarted)
            {
                throw new InvalidOperationException("Can not end an element before starting it.");
            }

            if (this.IsReadingEnded)
            {
                throw new InvalidOperationException("Can not end already ended element.");
            }

            this.EndRead();
        }

        protected bool ReadToElement<T>()
        {
            if (this.Reader.IsEndOfElement())
            {
                this.Reader.Read();
            }

            Type type = typeof(T);
            int index = this.GetRegisteredChildElementIndex(type);

            int readedElementIndex = this.GetCurrentElementIndex();

            while (readedElementIndex < index)
            {
                this.Reader.SkipElement();
                if (this.Reader.IsEndOfElement())
                {
                    this.Reader.Read();
                }

                readedElementIndex = this.GetCurrentElementIndex();
            }

            return readedElementIndex == index;
        }

        protected TResult GetRegisteredChildElement<TResult>()
            where TResult : ConsecutiveElementBase, new()
        {
            return this.childElementsManager.GetRegisteredChild<TResult>();
        }

        protected void RegisterChildElement<TResult>()
            where TResult : ElementBase, new()
        {
            this.RegisterChildElement<TResult>(null);
        }

        protected void RegisterChildElement<TResult>(Type dependOnType)
            where TResult : ElementBase, new()
        {
            this.childElementsManager.RegisterChild<TResult>(() =>
                {
                    ElementContext context = new ElementContext(this.Writer, this.Reader, this.Theme);
                    TResult instance = new TResult();
                    instance.SetContext(context);
                    return instance;
                },
                dependOnType);
        }

        protected TResult CreateChildElement<TResult>()
            where TResult : ElementBase, new()
        {
            ElementContext context = new ElementContext(this.Writer, this.Reader, this.Theme);
            TResult instance = new TResult();
            instance.SetContext(context);
            return instance;
        }

        private int GetCurrentElementIndex()
        {
            int readedElementIndex = int.MaxValue;
            string elementName = null;
            if (this.Reader.GetElementName(out elementName))
            {
                Type readedElementType = ElementsFactory.GetElementType(elementName);
                readedElementIndex = this.GetRegisteredChildElementIndex(readedElementType);
            }

            return readedElementIndex;
        }

        private int GetRegisteredChildElementIndex(Type type)
        {
            return this.childElementsManager.GetRegisteredChildIndex(type);
        }
    }
}