﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Xml;
using Telerik.Documents.SpreadsheetStreaming.Core;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Core.Attributes;
using Telerik.Documents.SpreadsheetStreaming.Model.Themes;

namespace Telerik.Documents.SpreadsheetStreaming.ImportExport.Core
{
    internal abstract class ElementBase : IChildEntry
    {
        private readonly Dictionary<string, OpenXmlAttributeBase> attributes;

        public ElementBase()
        {
            this.attributes = new Dictionary<string, OpenXmlAttributeBase>();
        }

        public bool IsEmptyElement { get; private set; }

        public virtual OpenXmlNamespace Namespace
        {
            get
            {
                return null;
            }
        }

        public virtual IEnumerable<OpenXmlNamespace> Namespaces
        {
            get
            {
                return Enumerable.Empty<OpenXmlNamespace>();
            }
        }

        public abstract string ElementName { get; }

        public bool IsWritingStarted { get; private set; }

        public bool IsWritingEnded { get; private set; }

        public bool IsReadingStarted { get; private set; }

        public bool IsReadingEnded { get; private set; }

        bool IChildEntry.IsUsageBegan
        {
            get
            {
                return this.IsWritingStarted || this.IsReadingStarted;
            }
        }

        bool IChildEntry.IsUsageCompleted
        {
            get 
            {
                return this.IsWritingEnded || this.IsReadingEnded;
            }
        }

        protected string InnerText
        {
            get;
            set;
        }

        protected virtual bool AlwaysExport
        {
            get
            {
                return true;
            }
        }

        protected SpreadTheme Theme { get; private set; }

        protected OpenXmlWriter Writer { get; private set; }

        protected OpenXmlReader Reader { get; private set; }

        public void SetContext(ElementContext context)
        {
            this.Writer = context.Writer;
            this.Reader = context.Reader;
            this.Theme = context.Theme;
        }

        protected void BeginWrite()
        {
            this.WriteElementStart();
            this.WriteAttributes();
            this.WriteInnerText();

            this.IsWritingStarted = true;
        }

        protected void EndWrite()
        {
            this.WriteElementEnd();

            this.IsWritingEnded = true;
        }

        protected void BeginRead()
        {
            this.ReadElementStart();
            this.IsEmptyElement = this.Reader.IsEmptyElement();
            this.ReadAttributes();
            this.Reader.Read();

            this.IsReadingStarted = true;
        }

        protected void EndRead()
        {
            this.IsReadingEnded = true;
        }

        protected bool ShouldExport()
        {
            if (this.AlwaysExport)
            {
                return true;
            }

            if (!string.IsNullOrEmpty(this.InnerText))
            {
                return true;
            }

            if (this.attributes.Values.Any(a => a.ShouldExport()))
            {
                return true;
            }

            return false;
        }

        protected OpenXmlCountAttribute RegisterCountAttribute()
        {
            OpenXmlCountAttribute countAttribute = new OpenXmlCountAttribute();
            this.RegisterAttribute(countAttribute);

            return countAttribute;
        }

        protected OpenXmlAttribute<TValue> RegisterAttribute<TValue>(string name, TValue defaultValue, bool isRequired = false)
        {
            return this.RegisterAttribute<TValue>(name, null, defaultValue, isRequired);
        }

        protected OpenXmlAttribute<TValue> RegisterAttribute<TValue>(string name, bool isRequired = false)
        {
            return this.RegisterAttribute<TValue>(name, null, default(TValue), isRequired);
        }

        protected OpenXmlAttribute<TValue> RegisterAttribute<TValue>(string name, OpenXmlNamespace ns, TValue defaultValue, bool isRequired = false)
        {
            OpenXmlAttribute<TValue> res = new OpenXmlAttribute<TValue>(name, ns, defaultValue, isRequired);
            this.RegisterAttribute(res);

            return res;
        }

        protected TValue RegisterAttribute<TValue>(TValue attribute)
            where TValue : OpenXmlAttributeBase
        {
            if (attribute.Namespace != null)
            {
                this.attributes[string.Format("{0}:{1}", attribute.Namespace.LocalName, attribute.Name)] = attribute;
            }
            else
            {
                this.attributes[attribute.Name] = attribute;
            }

            return attribute;
        }

        protected ElementBase CreateElement(string elementName)
        {
            ElementContext context = new ElementContext(this.Writer, this.Reader, this.Theme);
            return ElementsFactory.CreateElement(elementName, context);
        }

        protected void ReadInnerText()
        {
            string value;
            if (this.Reader.GetElementValue(out value))
            {
                this.InnerText = value;
                this.Reader.Read();
            }
            else
            {
                this.Reader.SkipElement();
            }
        }

        private void WriteElementStart()
        {
            this.Writer.WriteElementStart(this);
            foreach (OpenXmlNamespace ns in this.Namespaces)
            {
                this.Writer.WriteNamespace(ns);
            }
        }

        private void WriteAttributes()
        {
            foreach (KeyValuePair<string, OpenXmlAttributeBase> pair in this.attributes)
            {
                OpenXmlAttributeBase attribute = pair.Value;

                if (attribute.ShouldExport())
                {
                    this.Writer.WriteAttribute(attribute);
                }
                attribute.MarkAsWritten();
            }
        }

        private void WriteInnerText()
        {
            this.Writer.WriteValue(this.InnerText);
        }

        private void WriteElementEnd()
        {
            this.Writer.WriteElementEnd();
        }

        private void ReadElementStart()
        {
            while (this.Reader.GetNodeType() != XmlNodeType.Element)
            {
                if (!this.Reader.Read())
                {
                    throw new InvalidOperationException("Element not found");
                }
            }
        }

        private void ReadAttributes()
        {
            if (this.Reader.HasAttributes() && this.Reader.MoveToFirstAttribute())
            {
                OpenXmlAttributeBase openXmlAttribute;
                string attribute;
                string value;
                do
                {
                    if (!this.Reader.GetAttributeNameAndValue(out attribute, out value))
                    {
                        throw new InvalidOperationException("Attribute not found.");
                    }

                    if (this.attributes.TryGetValue(attribute, out openXmlAttribute))
                    {
                        openXmlAttribute.SetStringValue(value);
                    }
                }
                while (this.Reader.MoveToNextAttribute());
            }
        }
    }
}