﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Model.Elements;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Model.Elements.Styles;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Model.Elements.Worksheet;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Model.Writers;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Model.Writers.Worksheet.Styles;

namespace Telerik.Documents.SpreadsheetStreaming.ImportExport.Core
{
    internal static class ElementsFactory
    {
        private static readonly Dictionary<string, Type> elementNameToType;
        private static readonly Dictionary<Type, string> typeToElementName;
        private static readonly Dictionary<string, Func<ElementBase>> elementNameToCreateFunc;

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Maintainability", "CA1506:AvoidExcessiveClassCoupling")]
        static ElementsFactory()
        {
            elementNameToType = new Dictionary<string, Type>();
            typeToElementName = new Dictionary<Type, string>();
            elementNameToCreateFunc = new Dictionary<string, Func<ElementBase>>();

            RegisterFactoryMethod<DefaultElement>(XlsxElementNames.DefaultElementName);
            RegisterFactoryMethod<OverrideElement>(XlsxElementNames.OverrideElementName);
            RegisterFactoryMethod<RelationshipElement>(XlsxElementNames.RelationshipElementName);
            RegisterFactoryMethod<NumberFormatsElement>(XlsxElementNames.NumberFormatsElementName);
            RegisterFactoryMethod<NumberFormatElement>(XlsxElementNames.NumberFormatElementName);
            RegisterFactoryMethod<FontsElement>(XlsxElementNames.FontsElementName);
            RegisterFactoryMethod<FontElement>(XlsxElementNames.FontElementName);
            RegisterFactoryMethod<FontSizeElement>(XlsxElementNames.FontSizeElementName);
            RegisterFactoryMethod<ColorElement>(XlsxElementNames.ColorElementName);
            RegisterFactoryMethod<FontNameElement>(XlsxElementNames.FontNameElementName);
            RegisterFactoryMethod<FontFamilyElement>(XlsxElementNames.FontFamilyElementName);
            RegisterFactoryMethod<SchemeElement>(XlsxElementNames.SchemeElementName);
            RegisterFactoryMethod<BoldElement>(XlsxElementNames.BoldElementName);
            RegisterFactoryMethod<ItalicElement>(XlsxElementNames.ItalicElementName);
            RegisterFactoryMethod<UnderlineElement>(XlsxElementNames.UnderlineElementName);
            RegisterFactoryMethod<FillsElement>(XlsxElementNames.FillsElementName);
            RegisterFactoryMethod<FillElement>(XlsxElementNames.FillElementName);
            RegisterFactoryMethod<PatternFillElement>(XlsxElementNames.PatternFillElementName);
            RegisterFactoryMethod<BordersElement>(XlsxElementNames.BordersElementName);
            RegisterFactoryMethod<BorderElement>(XlsxElementNames.BorderElementName);
            RegisterFactoryMethod<LeftBorderElement>(XlsxElementNames.LeftBorderElementName);
            RegisterFactoryMethod<RightBorderElement>(XlsxElementNames.RightBorderElementName);
            RegisterFactoryMethod<TopBorderElement>(XlsxElementNames.TopBorderElementName);
            RegisterFactoryMethod<BottomBorderElement>(XlsxElementNames.BottomBorderElementName);
            RegisterFactoryMethod<DiagonalBorderElement>(XlsxElementNames.DiagonalBorderElementName);
            RegisterFactoryMethod<CellStyleFormatsElement>(XlsxElementNames.CellStyleFormatsElementName);
            RegisterFactoryMethod<FormatElement>(XlsxElementNames.FormatElementName);
            RegisterFactoryMethod<CellFormatsElement>(XlsxElementNames.CellFormatsElementName);
            RegisterFactoryMethod<CellStylesElement>(XlsxElementNames.CellStylesElementName);
            RegisterFactoryMethod<CellStyleElement>(XlsxElementNames.CellStyleElementName);
            RegisterFactoryMethod<DifferentialFormatsElement>(XlsxElementNames.DifferentialFormatsElementName);
            RegisterFactoryMethod<SheetsElement>(XlsxElementNames.SheetsElementName);
            RegisterFactoryMethod<SheetElement>(XlsxElementNames.SheetElementName);
            RegisterFactoryMethod<SheetDataElement>(XlsxElementNames.SheetDataElementName);
            RegisterFactoryMethod<ColumnsElement>(XlsxElementNames.ColumnsElementName);
            RegisterFactoryMethod<ColumnElement>(XlsxElementNames.ColumnElementName);
            RegisterFactoryMethod<RowElement>(XlsxElementNames.RowElementName);
            RegisterFactoryMethod<CellElement>(XlsxElementNames.CellElementName);
            RegisterFactoryMethod<CellValueElement>(XlsxElementNames.CellValueElementName);
            RegisterFactoryMethod<ForegroundColorElement>(XlsxElementNames.ForegroundColorElementName);
            RegisterFactoryMethod<BackgroundColorElement>(XlsxElementNames.BackgroundColorElementName);
            RegisterFactoryMethod<AlignmentElement>(XlsxElementNames.AlignmentElementName);
            RegisterFactoryMethod<GradientFillElement>(XlsxElementNames.GradientFillElementName);
            RegisterFactoryMethod<StopElement>(XlsxElementNames.StopElementName);
            RegisterFactoryMethod<VerticalAlignmentElement>(XlsxElementNames.VerticalAlignmentElementName);
            RegisterFactoryMethod<ProtectionElement>(XlsxElementNames.ProtectionElementName);
            RegisterFactoryMethod<StrikeElement>(XlsxElementNames.StrikeElementName);
        }

        public static ElementBase CreateElement(string elementName, ElementContext context)
        {
            ElementBase element = null;
            Func<ElementBase> createElementFunc = null;

            if (elementNameToCreateFunc.TryGetValue(elementName, out createElementFunc))
            {
                if (createElementFunc == null)
                {
                    return null;
                }

                element = createElementFunc();
                element.SetContext(context);
            }
            else
            {
                Debug.Assert(false, string.Format("{0} Element factory not found!", elementName));
            }

            return element;
        }

        public static Type GetElementType(string elementName)
        {
            return elementNameToType[elementName];
        }

        public static bool CanCreateElement(string elementName)
        {
            if (!elementNameToType.ContainsKey(elementName))
            {
                return false;
            }

            return elementNameToType[elementName] != null;
        }

        private static void RegisterFactoryMethod<T>(string elementName)
            where T : ElementBase, new()
        {
            Type type = typeof(T);
            elementNameToType.Add(elementName, type);
            elementNameToCreateFunc.Add(elementName, () => new T());
            if (type != null)
            {
                typeToElementName.Add(type, elementName);
            }
        }
    }
}