﻿using System;
using System.Globalization;
using System.Linq;

namespace Telerik.Documents.SpreadsheetStreaming.ImportExport.Core
{
    internal class OpenXmlCultureInfo
    {
        private readonly CultureInfo cultureInfo;

        public OpenXmlCultureInfo()
            : this(CultureInfo.CurrentCulture)
        {
        }

        public OpenXmlCultureInfo(CultureInfo cultureInfo)
        {
            this.cultureInfo = cultureInfo;
        }

        public CultureInfo CultureInfo
        {
            get
            {
                return this.cultureInfo;
            }
        }

        public string NumberGroupSeparator
        {
            get
            {
                return this.cultureInfo.NumberFormat.NumberGroupSeparator;
            }
        }

        public string NumberDecimalSeparator
        {
            get
            {
                return this.cultureInfo.NumberFormat.NumberDecimalSeparator;
            }
        }

        public static bool IsCharEqualTo(char ch, params string[] values)
        {
            string character = ch.ToString();
            for (int i = 0; i < values.Length; i++)
            {
                if (character == values[i])
                {
                    return true;
                }
            }

            return false;
        }

        public bool TryParseGeneral(string value, out double result)
        {
            value = value.Trim(' ');
            NumberStyles numberStyles = NumberStyles.AllowDecimalPoint | NumberStyles.AllowLeadingSign;
            return double.TryParse(value, numberStyles, this.cultureInfo, out result);
        }

        public bool TryParseNumber(string value, out double result)
        {
            result = 0;

            bool containsNumberGroupSeparator = value.Contains(this.NumberGroupSeparator);

            bool canParse = containsNumberGroupSeparator &&
                double.TryParse(value, NumberStyles.Number, this.cultureInfo, out result) &&
                this.ApplyAdditionalNumberConstraints(value);

            return canParse;
        }

        public bool TryParseDouble(string value, out double result)
        {
            return this.TryParseGeneral(value, out result) ||
                this.TryParseNumber(value, out result);
        }

        private static bool CheckGroupSeparatorsPositionedCorrectly(string value, string groupSeparator, string decimalSeparator, int[] groupSizes)
        {
            string plainString = GetValueWithoutFractionalPart(value, decimalSeparator);
            plainString = RemoveLeadingAndTrailingNonDigitSymbols(plainString, groupSeparator, decimalSeparator);

            string[] groups = plainString.Split(new string[] { groupSeparator }, StringSplitOptions.None);

            if (groups.Length == 1)
            {
                return true;
            }

            Func<int, int, bool> greaterThanComparer = new Func<int, int, bool>((i, j) => i >= j);
            Func<int, int, bool> equalsComparer = new Func<int, int, bool>((i, j) => i == j);

            if (!CompareValues(groupSizes, groups[0].Length, greaterThanComparer))
            {
                return false;
            }

            for (int i = 1; i < groups.Length; i++)
            {
                if (!CompareValues(groupSizes, groups[i].Length, equalsComparer))
                {
                    return false;
                }
            }

            return true;
        }

        private static string GetValueWithoutFractionalPart(string value, string decimalSeparator)
        {
            string result = value;
            int decimalSeparatorIndex = value.IndexOf(decimalSeparator);
            if (decimalSeparatorIndex != -1)
            {
                result = value.Substring(0, decimalSeparatorIndex);
            }

            return result;
        }

        private static string RemoveLeadingAndTrailingNonDigitSymbols(string value, string groupSeparator, string decimalSeparator)
        {
            string result = value;

            int firstDigitIndex = 0;
            for (int i = 0; i < value.Length; i++)
            {
                if (IsValidSymbol(value[i], groupSeparator, decimalSeparator))
                {
                    firstDigitIndex = i;
                    break;
                }
            }

            int lastDigitIndex = value.Length - 1;
            for (int i = lastDigitIndex; i >= 0; i--)
            {
                if (IsValidSymbol(value[i], groupSeparator, decimalSeparator))
                {
                    lastDigitIndex = i;
                    break;
                }
            }

            result = value.Substring(firstDigitIndex, lastDigitIndex - firstDigitIndex + 1);

            return result;
        }

        private static bool CompareValues(int[] availableValues, int requiredValue, Func<int, int, bool> comparer)
        {
            bool returnValue = false;
            foreach (int value in availableValues)
            {
                if (comparer(value, requiredValue))
                {
                    returnValue = true;
                    break;
                }
            }

            if (!returnValue)
            {
                return false;
            }

            return true;
        }

        private static bool IsValidSymbol(char c, string groupSeparator, string decimalSeparator)
        {
            return char.IsDigit(c) || IsCharEqualTo(c, groupSeparator, decimalSeparator);
        }

        private bool ApplyAdditionalNumberConstraints(string value)
        {
            if (value.IndexOf(this.NumberGroupSeparator) != -1)
            {
                return CheckGroupSeparatorsPositionedCorrectly(value, this.NumberGroupSeparator, this.NumberDecimalSeparator, this.CultureInfo.NumberFormat.NumberGroupSizes);
            }

            return false;
        }
    }
}