﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Documents.SpreadsheetStreaming.Model;
using Telerik.Documents.SpreadsheetStreaming.Utilities;

namespace Telerik.Documents.SpreadsheetStreaming.ImportExport.Core.Types
{
    internal class GradientFillInfo
    {
        private static readonly Dictionary<SpreadGradientType, Func<SpreadGradientFill, GradientFillInfo>> gradientTypeToInfoFactory;
        private List<GradientStop> stops;

        static GradientFillInfo()
        {
            gradientTypeToInfoFactory = new Dictionary<SpreadGradientType, Func<SpreadGradientFill, GradientFillInfo>>();
            InitGradientInfoFactories();
        }

        public GradientFillInfo()
        {
            this.Stops = new List<GradientStop>();
        }

        public GradientFillInfo(IEnumerable<GradientStop> stops,
            double? degree = null,
            GradientInfoType? infoType = null,
            double? left = null,
            double? top = null,
            double? right = null,
            double? bottom = null)
        {
            this.Stops = stops.ToList();
            this.Degree = degree;
            this.Left = left;
            this.Top = top;
            this.Right = right;
            this.Bottom = bottom;
            this.InfoType = infoType;
        }

        public List<GradientStop> Stops
        {
            get
            {
                return this.stops;
            }
            set
            {
                this.stops = value;
            }
        }

        public double? Degree { get; set; }

        public double? Left { get; set; }

        public double? Top { get; set; }

        public double? Right { get; set; }

        public double? Bottom { get; set; }

        public GradientInfoType? InfoType { get; set; }

        public static GradientFillInfo Create(SpreadGradientFill gradientFill)
        {
            return gradientTypeToInfoFactory[gradientFill.GradientType](gradientFill);
        }

        public SpreadGradientFill ToFill()
        {
            SpreadThemableColor color1 = null;
            SpreadThemableColor color2 = null;
            SpreadGradientType gradientType = SpreadGradientType.Horizontal;

            foreach (var gradeintStop in this.Stops)
            {
                if (color1 == null)
                {
                    color1 = gradeintStop.ThemableColor;
                }
                else if (color2 == null)
                {
                    color2 = gradeintStop.ThemableColor;
                }
            }

            Guard.ThrowExceptionIfNull(color1, "color1");
            Guard.ThrowExceptionIfNull(color2, "color2");

            SpreadGradientFill fakeGradientFill = new SpreadGradientFill(SpreadGradientType.Horizontal, color1, color2);
            foreach (var pair in gradientTypeToInfoFactory)
            {
                var fillInfoFactory = pair.Value;
                GradientFillInfo other = fillInfoFactory(fakeGradientFill);

                if (this.Stops.Count == other.Stops.Count &&
                    ObjectExtensions.EqualsOfT(this.Degree, other.Degree) &&
                    ObjectExtensions.EqualsOfT(this.Left, other.Left) &&
                    ObjectExtensions.EqualsOfT(this.Right, other.Right) &&
                    ObjectExtensions.EqualsOfT(this.Top, other.Top) &&
                    ObjectExtensions.EqualsOfT(this.Bottom, other.Bottom) &&
                    ObjectExtensions.EqualsOfT(this.InfoType, other.InfoType))
                {
                    bool areSame = true;
                    for (int i = 0; i < this.Stops.Count; i++)
                    {
                        if (this.stops[i].Position != other.stops[i].Position)
                        {
                            areSame = false;
                            break;
                        }
                    }

                    if (areSame)
                    {
                        gradientType = pair.Key;
                        break;
                    }
                }
            }

            return new SpreadGradientFill(gradientType, color1, color2);
        }

        private static void InitGradientInfoFactories()
        {
            RegisterGradientCreationMethod(SpreadGradientType.Horizontal,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops, 90);
                });

            RegisterGradientCreationMethod(SpreadGradientType.HorizontalReversed,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops, 270);
                });

            RegisterGradientCreationMethod(SpreadGradientType.HorizontalRepeated,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(0.5, gradientFill.Color2),
                        new GradientStop(1, gradientFill.Color1)
                    };

                    return new GradientFillInfo(gradientStops, 90);
                });

            RegisterGradientCreationMethod(SpreadGradientType.Vertical,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops);
                });

            RegisterGradientCreationMethod(SpreadGradientType.VerticalReversed,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops, 180);
                });

            RegisterGradientCreationMethod(SpreadGradientType.VerticalRepeated,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(0.5, gradientFill.Color2),
                        new GradientStop(1, gradientFill.Color1)
                    };

                    return new GradientFillInfo(gradientStops);
                });

            RegisterGradientCreationMethod(SpreadGradientType.DiagonalUp,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2),
                    };

                    return new GradientFillInfo(gradientStops, 45);
                });

            RegisterGradientCreationMethod(SpreadGradientType.DiagonalUpReversed,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2),
                    };

                    return new GradientFillInfo(gradientStops, 225);
                });

            RegisterGradientCreationMethod(SpreadGradientType.DiagonalUpRepeated,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(0.5, gradientFill.Color2),
                        new GradientStop(1, gradientFill.Color1),
                    };

                    return new GradientFillInfo(gradientStops, 45);
                });

            RegisterGradientCreationMethod(SpreadGradientType.DiagonalDown,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2),
                    };

                    return new GradientFillInfo(gradientStops, 135);
                });

            RegisterGradientCreationMethod(SpreadGradientType.DiagonalDownReversed,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2),
                    };

                    return new GradientFillInfo(gradientStops, 315);
                });

            RegisterGradientCreationMethod(SpreadGradientType.DiagonalDownRepeated,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(0.5, gradientFill.Color2),
                        new GradientStop(1, gradientFill.Color1),
                    };

                    return new GradientFillInfo(gradientStops, 135);
                });

            RegisterGradientCreationMethod(SpreadGradientType.FromTopLeftCorner,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops, infoType: GradientInfoType.Path);
                });

            RegisterGradientCreationMethod(SpreadGradientType.FromTopRightCorner,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops, infoType: GradientInfoType.Path, left: 1, right: 1);
                });

            RegisterGradientCreationMethod(SpreadGradientType.FromBottomLeftCorner,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops, infoType: GradientInfoType.Path, top: 1, bottom: 1);
                });

            RegisterGradientCreationMethod(SpreadGradientType.FromBottomRightCorner,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops, infoType: GradientInfoType.Path, left: 1, top: 1, right: 1, bottom: 1);
                });

            RegisterGradientCreationMethod(SpreadGradientType.FromCenter,
                (gradientFill) =>
                {
                    List<GradientStop> gradientStops = new List<GradientStop>
                    {
                        new GradientStop(0, gradientFill.Color1),
                        new GradientStop(1, gradientFill.Color2)
                    };

                    return new GradientFillInfo(gradientStops, infoType: GradientInfoType.Path, left: 0.5, top: 0.5, right: 0.5, bottom: 0.5);
                });
        }

        private static void RegisterGradientCreationMethod(SpreadGradientType gradientType, Func<SpreadGradientFill, GradientFillInfo> factory)
        {
            gradientTypeToInfoFactory.Add(gradientType, factory);
        }
    }
}