﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Core;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Core.Types;

namespace Telerik.Documents.SpreadsheetStreaming.ImportExport.Model.Elements.Styles
{
    internal class FontElement : DirectElementBase<FontProperties>
    {
        private readonly Dictionary<string, Action<ElementBase, FontProperties>> elementNameToReadMethod;

        public FontElement()
        {
            this.elementNameToReadMethod = new Dictionary<string, Action<ElementBase, FontProperties>>();
            this.elementNameToReadMethod.Add(XlsxElementNames.BoldElementName, this.ReadBoldElement);
            this.elementNameToReadMethod.Add(XlsxElementNames.ItalicElementName, this.ReadItalicElement);
            this.elementNameToReadMethod.Add(XlsxElementNames.ColorElementName, this.ReadColorElement);
            this.elementNameToReadMethod.Add(XlsxElementNames.FontSizeElementName, this.ReadFontSizeElement);
            this.elementNameToReadMethod.Add(XlsxElementNames.FontNameElementName, this.ReadFontNameElement);
            this.elementNameToReadMethod.Add(XlsxElementNames.UnderlineElementName, this.ReadUnderlineElement);
            this.elementNameToReadMethod.Add(XlsxElementNames.FontFamilyElementName, this.ReadFontFamilyElement);
            this.elementNameToReadMethod.Add(XlsxElementNames.SchemeElementName, this.ReadSchemeElement);
            this.elementNameToReadMethod.Add(XlsxElementNames.VerticalAlignmentElementName, this.ReadVerticalAlignmentElement);
            this.elementNameToReadMethod.Add(XlsxElementNames.StrikeElementName, this.ReadStrikeElement);
        }

        public override string ElementName
        {
            get
            {
                return XlsxElementNames.FontElementName;
            }
        }

        protected override void InitializeAttributesOverride(FontProperties value)
        {
        }

        protected override void WriteChildElementsOverride(FontProperties value)
        {
            if (value.IsBold.HasValue)
            {
                BoldElement bold = this.CreateChildElement<BoldElement>();
                bold.Write(value.IsBold.Value);
            }

            if (value.IsItalic.HasValue)
            {
                ItalicElement italic = this.CreateChildElement<ItalicElement>();
                italic.Write(value.IsItalic.Value);
            }

            if (value.ForeColor != null)
            {
                ColorElement color = this.CreateChildElement<ColorElement>();
                color.Write(value.ForeColor);
            }

            if (value.FontSize.HasValue)
            {
                FontSizeElement fontSize = this.CreateChildElement<FontSizeElement>();
                fontSize.Write(value.FontSize.Value);
            }

            if (value.FontFamily != null)
            {
                FontNameElement name = this.CreateChildElement<FontNameElement>();
                name.Write(value.FontFamily.GetActualValue(this.Theme));

                if (value.FontFamily.IsFromTheme)
                {
                    SchemeElement scheme = this.CreateChildElement<SchemeElement>();
                    scheme.Write(FontSchemesMapper.GetFontSchemeName(value.FontFamily.ThemeFontType));
                }
            }

            if (value.Underline.HasValue)
            {
                UnderlineElement underline = this.CreateChildElement<UnderlineElement>();
                underline.Write(UnderlineValuesMapper.GetUnderlineValueName(value.Underline.Value));
            }

            if (value.VerticalAlignment != null)
            {
                VerticalAlignmentElement verticalAlignment = this.CreateChildElement<VerticalAlignmentElement>();
                verticalAlignment.Write(value.VerticalAlignment);
            }

            if (value.Strike.HasValue)
            {
                StrikeElement strikeElement = this.CreateChildElement<StrikeElement>();
                strikeElement.Write(value.Strike.Value);
            }
        }

        protected override void CopyAttributesOverride(ref FontProperties value)
        {
        }

        protected override void ReadChildElementOverride(ElementBase element, ref FontProperties value)
        {
            this.elementNameToReadMethod[element.ElementName](element, value);
        }

        private void ReadBoldElement(ElementBase element, FontProperties value)
        {
            BoldElement boldElement = element as BoldElement;
            bool isBold = false;
            boldElement.Read(ref isBold);

            value.IsBold = isBold;
        }

        private void ReadItalicElement(ElementBase element, FontProperties value)
        {
            ItalicElement italicElelemt = element as ItalicElement;
            bool isItalic = false;
            italicElelemt.Read(ref isItalic);

            value.IsItalic = isItalic;
        }

        private void ReadColorElement(ElementBase element, FontProperties value)
        {
            ColorElement colorElement = element as ColorElement;
            SpreadThemableColor color = null;
            colorElement.Read(ref color);

            if (color != null)
            {
                value.ForeColor = color;
            }
        }

        private void ReadFontSizeElement(ElementBase element, FontProperties value)
        {
            FontSizeElement fontSizeElement = element as FontSizeElement;
            double fontSize = -1;
            fontSizeElement.Read(ref fontSize);

            if (fontSize > 0)
            {
                value.FontSize = fontSize;
            }
        }

        private void ReadFontNameElement(ElementBase element, FontProperties value)
        {
            // TODO: What if is from Theme? See the writing of this elelemt!
            FontNameElement fontNameElement = element as FontNameElement;
            string fontName = null;
            fontNameElement.Read(ref fontName);

            if (fontName != null)
            {
                value.FontFamily = new SpreadThemableFontFamily(fontName);
            }
        }

        private void ReadFontFamilyElement(ElementBase element, FontProperties value)
        {
            FontFamilyElement fontFamilyElement = element as FontFamilyElement;
            string fontFamily = null;
            fontFamilyElement.Read(ref fontFamily);

            if (fontFamily != null)
            {
                // TODO: This may be needed?
            }
        }

        private void ReadSchemeElement(ElementBase element, FontProperties value)
        {
            SchemeElement schemeElement = element as SchemeElement;
            string scheme = null;
            schemeElement.Read(ref scheme);

            if (scheme != null)
            {
                SpreadThemeFontType? themeFontType = FontSchemesMapper.GetFontScheme(scheme);
                if (themeFontType != null)
                {
                    value.FontFamily = new SpreadThemableFontFamily(themeFontType.Value);
                }
            }
        }

        private void ReadUnderlineElement(ElementBase element, FontProperties value)
        {
            UnderlineElement underlineElement = element as UnderlineElement;
            string underline = null;
            underlineElement.Read(ref underline);

            value.Underline = UnderlineValuesMapper.GetUnderlineValue(underline);
        }

        private void ReadVerticalAlignmentElement(ElementBase element, FontProperties value)
        {
            VerticalAlignmentElement verticalAlignmentElement = element as VerticalAlignmentElement;
            string verticalAlignment = null;
            verticalAlignmentElement.Read(ref verticalAlignment);

            value.VerticalAlignment = verticalAlignment;
        }

        private void ReadStrikeElement(ElementBase element, FontProperties value)
        {
            StrikeElement strikeElement = element as StrikeElement;
            bool strike = false;
            strikeElement.Read(ref strike);

            value.Strike = strike;
        }
    }
}