﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Core;
using Telerik.Documents.SpreadsheetStreaming.Model.Formatting;

namespace Telerik.Documents.SpreadsheetStreaming.ImportExport.Model.Elements.Styles
{
    internal class StyleSheetElement : DirectElementBase<StylesRepository>
    {
        private readonly Dictionary<string, Action<ElementBase, StylesRepository>> elementNameToReadAction;

        public StyleSheetElement()
        {
            this.elementNameToReadAction = new Dictionary<string, Action<ElementBase, StylesRepository>>();

            this.elementNameToReadAction.Add(XlsxElementNames.NumberFormatsElementName, this.ReadNumberFormats);
            this.elementNameToReadAction.Add(XlsxElementNames.FontsElementName, this.ReadFonts);
            this.elementNameToReadAction.Add(XlsxElementNames.FillsElementName, this.ReadFills);
            this.elementNameToReadAction.Add(XlsxElementNames.BordersElementName, this.ReadBorders);
            this.elementNameToReadAction.Add(XlsxElementNames.CellStyleFormatsElementName, this.ReadCellStyleFormats);
            this.elementNameToReadAction.Add(XlsxElementNames.CellFormatsElementName, this.ReadCellFormats);
            this.elementNameToReadAction.Add(XlsxElementNames.CellStylesElementName, this.ReadCellStyles);
            this.elementNameToReadAction.Add(XlsxElementNames.DifferentialFormatsElementName, this.ReadDxfs);
        }

        public override OpenXmlNamespace Namespace
        {
            get
            {
                return OpenXmlNamespaces.SpreadsheetMLNamespace;
            }
        }

        public override string ElementName
        {
            get
            {
                return XlsxElementNames.StyleSheetElementName;
            }
        }

        protected override void InitializeAttributesOverride(StylesRepository value)
        {
        }

        protected override void WriteChildElementsOverride(StylesRepository value)
        {
            this.WriteNumberFormat(value.NumberFormats.ToList());
            this.WriteFonts(value.Fonts.ToList());
            this.WriteFills(value.Fills.ToList());
            this.WriteBorders(value.Borders.ToList());
            this.WriteCellStyleFormats(value.CellStyleFormats.ToList());
            this.WriteCellFormats(value.CellFormats.ToList());
            this.WriteCellStyles(value.CellStyles.ToList());
            this.WriteDxfs();

            // TODO: There is another element called colors
        }

        protected override void CopyAttributesOverride(ref StylesRepository value)
        {
        }

        protected override void ReadChildElementOverride(ElementBase element, ref StylesRepository value)
        {
            this.elementNameToReadAction[element.ElementName](element, value);
        }

        private static int AddItemToStylesRepository<T>(T item, Action<T> addItemAction, int startIndex)
        {
            int oldIndex = startIndex;

            addItemAction(item);

            if (oldIndex == startIndex)
            {
                startIndex++;
            }

            return startIndex;
        }

        private void WriteNumberFormat(List<NumberFormat> numberFormats)
        {
            NumberFormatsElement numberFormatsElement = this.CreateChildElement<NumberFormatsElement>();
            numberFormatsElement.Write(numberFormats);
        }

        private void ReadNumberFormats(ElementBase element, StylesRepository styles)
        {
            NumberFormatsElement numberFormatsElement = element as NumberFormatsElement;
            List<NumberFormat> numberFormats = new List<NumberFormat>();

            numberFormatsElement.Read(ref numberFormats);

            foreach (NumberFormat item in numberFormats)
            {
                styles.NumberFormatsStartIndex = AddItemToStylesRepository(item, p => styles.AddNumberFormat(p), styles.NumberFormatsStartIndex);
            }
        }

        private void WriteFonts(List<FontProperties> fonts)
        {
            FontsElement fontsElement = this.CreateChildElement<FontsElement>();
            fontsElement.Write(fonts);
        }

        private void ReadFonts(ElementBase element, StylesRepository styles)
        {
            FontsElement fontsElement = element as FontsElement;
            List<FontProperties> fonts = new List<FontProperties>();

            fontsElement.Read(ref fonts);

            foreach (FontProperties item in fonts)
            {
                styles.FontsStartIndex = AddItemToStylesRepository(item, p => styles.AddFont(p), styles.FontsStartIndex);
            }
        }

        private void WriteFills(List<ISpreadFill> fills)
        {
            FillsElement fillsElement = this.CreateChildElement<FillsElement>();
            fillsElement.Write(fills);
        }

        private void ReadFills(ElementBase element, StylesRepository styles)
        {
            FillsElement fillsElement = element as FillsElement;
            List<ISpreadFill> fills = new List<ISpreadFill>();

            fillsElement.Read(ref fills);

            foreach (ISpreadFill item in fills)
            {
                styles.FillsStartIndex = AddItemToStylesRepository(item, p => styles.AddFill(p), styles.FillsStartIndex);
            }
        }

        private void WriteBorders(List<SpreadCellBorders> borders)
        {
            BordersElement bordersElement = this.CreateChildElement<BordersElement>();
            bordersElement.Write(borders);
        }

        private void ReadBorders(ElementBase element, StylesRepository styles)
        {
            BordersElement bordersElement = element as BordersElement;
            List<SpreadCellBorders> borders = new List<SpreadCellBorders>();

            bordersElement.Read(ref borders);

            foreach (SpreadCellBorders item in borders)
            {
                styles.BordersStartIndex = AddItemToStylesRepository(item, p => styles.AddBorders(p), styles.BordersStartIndex);
            }
        }

        private void WriteCellStyleFormats(List<DiferentialFormat> cellStyleFormats)
        {
            CellStyleFormatsElement cellStyleFormatsElement = this.CreateChildElement<CellStyleFormatsElement>();
            cellStyleFormatsElement.Write(cellStyleFormats);
        }

        private void ReadCellStyleFormats(ElementBase element, StylesRepository styles)
        {
            CellStyleFormatsElement cellStyleFormatsElement = element as CellStyleFormatsElement;
            List<DiferentialFormat> formats = new List<DiferentialFormat>();

            cellStyleFormatsElement.Read(ref formats);

            foreach (DiferentialFormat item in formats)
            {
                styles.CellStyleFormatsStartIndex = AddItemToStylesRepository(item, p => styles.AddCellStyleFormat(p), styles.CellStyleFormatsStartIndex);
            }
        }

        private void WriteCellFormats(List<DiferentialFormat> cellFormats)
        {
            CellFormatsElement cellFormatsElement = this.CreateChildElement<CellFormatsElement>();
            cellFormatsElement.Write(cellFormats);
        }

        private void ReadCellFormats(ElementBase element, StylesRepository styles)
        {
            CellFormatsElement cellFormatsElement = element as CellFormatsElement;
            List<DiferentialFormat> formats = new List<DiferentialFormat>();

            cellFormatsElement.Read(ref formats);

            foreach (DiferentialFormat item in formats)
            {
                styles.CellFormatsStartIndex = AddItemToStylesRepository(item, p => styles.AddCellFormat(p), styles.CellFormatsStartIndex);
            }
        }

        private void WriteCellStyles(List<CellStyleInfo> cellStyles)
        {
            CellStylesElement cellStylesElement = this.CreateChildElement<CellStylesElement>();
            cellStylesElement.Write(cellStyles);
        }

        private void ReadCellStyles(ElementBase element, StylesRepository styles)
        {
            CellStylesElement cellStylesElement = element as CellStylesElement;
            List<CellStyleInfo> formats = new List<CellStyleInfo>();

            cellStylesElement.Read(ref formats);

            foreach (CellStyleInfo item in formats)
            {
                styles.CellStylesStartIndex = AddItemToStylesRepository(item, p => styles.AddCellStyle(p), styles.CellStylesStartIndex);
            }
        }

        private void WriteDxfs()
        {
            // TODO: This is not completely implemented!
            DifferentialFormatsElement differentialFormatsElement = this.CreateChildElement<DifferentialFormatsElement>();
            differentialFormatsElement.Write(0);
        }

        private void ReadDxfs(ElementBase element, StylesRepository styles)
        {
            DifferentialFormatsElement differentialFormatsElement = element as DifferentialFormatsElement;
            int value = 0;
            differentialFormatsElement.Read(ref value);
        }
    }
}