﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Core;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Model.Parts;
using Telerik.Documents.SpreadsheetStreaming.Model.Formatting;
using Telerik.Documents.SpreadsheetStreaming.Model.Relations;
using Telerik.Documents.SpreadsheetStreaming.Model.Themes;
using Telerik.Documents.SpreadsheetStreaming.Utilities;

#if NETSTANDARD
using Telerik.Zip;
#else
using Telerik.Windows.Zip;
#endif

namespace Telerik.Documents.SpreadsheetStreaming.Importers
{
    internal class XlsxWorkbookImporter : IWorkbookImporter
    {
        private readonly ZipArchive archive;
        private readonly ContentTypesRepository contentTypes;
        private readonly Relationships workbookRelationships;
        private readonly StylesRepository stylesRepository;
        private readonly SpreadTheme theme;

        internal XlsxWorkbookImporter(Stream stream)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");

            this.theme = SpreadPredefinedThemeSchemes.DefaultTheme;
            this.archive = new ZipArchive(stream, ZipArchiveMode.Read, true, null);
            this.contentTypes = new ContentTypesRepository();
            this.workbookRelationships = new Relationships(import: true);
            this.stylesRepository = new StylesRepository();

            this.ReadContentTypesPart();
            this.ReadWorkbookRelationshipsPart();
            this.ReadStylesPart();
            this.ReadWorkbookPart();
        }

        ~XlsxWorkbookImporter()
        {
            this.Dispose(false);
        }

        public IEnumerable<IWorksheetImporter> WorksheetImporters
        {
            get
            {
                foreach (Relationship relationship in this.workbookRelationships.GetRelationships(XlsxRelationshipTypes.WorksheetRelationshipType))
                {
                    yield return new XlsxWorksheetImporter(this.archive, relationship.Name, relationship.Target);
                }
            }
        }

        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                this.archive.Dispose();
            }
        }

        private void ReadContentTypesPart()
        {
            using (ContentTypesPart part = new ContentTypesPart(new PartContext(this.archive)))
            {
                part.Read(this.contentTypes);
            }
        }

        private void ReadWorkbookRelationshipsPart()
        {
            using (WorkbookRelationshipsPart part = new WorkbookRelationshipsPart(new PartContext(this.archive)))
            {
                part.ReadRelationships(this.workbookRelationships);
            }
        }

        private void ReadStylesPart()
        {
            using (StylesPart part = new StylesPart(new PartContext(this.archive, this.theme)))
            {
                part.Read(this.stylesRepository);
            }
        }

        private void ReadWorkbookPart()
        {
            List<Relationship> relationships = new List<Relationship>();
            using (WorkbookPart part = new WorkbookPart(new PartContext(this.archive)))
            {
                part.Read(relationships);
            }

            foreach (Relationship relationship in relationships)
            {
                this.workbookRelationships.AddRelationship(relationship);
            }
        }
    }
}