﻿using System;
using System.Collections.Generic;
using System.Linq;
using Telerik.Documents.SpreadsheetStreaming.ImportExport.Model;
using Telerik.Documents.SpreadsheetStreaming.Utilities;

namespace Telerik.Documents.SpreadsheetStreaming.Model.Formatting
{
    internal class StylesRepository
    {
        private const int DefaultNumberFormatsStartIndex = 192;

        private readonly UniqueItemsStore<FontProperties> fonts;
        private readonly UniqueItemsStore<ISpreadFill> fills;
        private readonly UniqueItemsStore<SpreadCellBorders> borders;
        private readonly UniqueItemsStore<CellStyleInfo> cellStyles;
        private readonly UniqueItemsStore<DiferentialFormat> cellStyleFormats;
        private readonly UniqueItemsStore<DiferentialFormat> cellFormats;
        private readonly UniqueItemsStore<NumberFormat> numberFormats;
        private readonly SpreadCellStyle normalStyle;

        public StylesRepository()
        {
            this.fonts = new UniqueItemsStore<FontProperties>();
            this.fills = new UniqueItemsStore<ISpreadFill>();
            this.borders = new UniqueItemsStore<SpreadCellBorders>();
            this.cellStyles = new UniqueItemsStore<CellStyleInfo>();
            this.cellStyleFormats = new UniqueItemsStore<DiferentialFormat>();
            this.cellFormats = new UniqueItemsStore<DiferentialFormat>();
            this.numberFormats = new UniqueItemsStore<NumberFormat>(DefaultNumberFormatsStartIndex);
        }

        public StylesRepository(SpreadCellStyle normalStyle)
            : this()
        {
            this.normalStyle = normalStyle;
            this.InitializeBuiltInStyles();
        }

        public IEnumerable<FontProperties> Fonts
        {
            get
            {
                return this.fonts;
            }
        }

        public IEnumerable<ISpreadFill> Fills
        {
            get
            {
                return this.fills;
            }
        }

        public IEnumerable<SpreadCellBorders> Borders
        {
            get
            {
                return this.borders;
            }
        }

        public IEnumerable<CellStyleInfo> CellStyles
        {
            get
            {
                return this.cellStyles;
            }
        }

        public IEnumerable<DiferentialFormat> CellStyleFormats
        {
            get
            {
                return this.cellStyleFormats;
            }
        }

        public IEnumerable<DiferentialFormat> CellFormats
        {
            get
            {
                return this.cellFormats;
            }
        }

        public IEnumerable<NumberFormat> NumberFormats
        {
            get
            {
                return this.numberFormats;
            }
        }

        public int NumberFormatsStartIndex
        {
            get
            {
                return this.numberFormats.StartIndex;
            }
            set
            {
                int maxIndex = Math.Max(value, DefaultNumberFormatsStartIndex);
                this.numberFormats.StartIndex = maxIndex;
            }
        }

        public int FontsStartIndex
        {
            get
            {
                return this.fonts.StartIndex;
            }
            set
            {
                this.fonts.StartIndex = value;
            }
        }

        public int FillsStartIndex
        {
            get
            {
                return this.fills.StartIndex;
            }
            set
            {
                this.fills.StartIndex = value;
            }
        }

        public int BordersStartIndex
        {
            get
            {
                return this.borders.StartIndex;
            }
            set
            {
                this.borders.StartIndex = value;
            }
        }

        public int CellStylesStartIndex
        {
            get
            {
                return this.cellStyles.StartIndex;
            }
            set
            {
                this.cellStyles.StartIndex = value;
            }
        }

        public int CellStyleFormatsStartIndex
        {
            get
            {
                return this.cellStyleFormats.StartIndex;
            }
            set
            {
                this.cellStyleFormats.StartIndex = value;
            }
        }

        public int CellFormatsStartIndex
        {
            get
            {
                return this.cellFormats.StartIndex;
            }
            set
            {
                this.cellFormats.StartIndex = value;
            }
        }

        public int RegisterCellFormat(SpreadCellFormat cellFormat)
        {
            DiferentialFormat format = this.CellFormatToDiferentialFormat(cellFormat);

            int index = this.cellFormats.Add(format);

            return index;
        }

        public void AddNumberFormat(NumberFormat format)
        {
            if (format.NumberFormatId.HasValue)
            {
                this.numberFormats.Add(format.NumberFormatId.Value, format);
            }
            else
            {
                this.numberFormats.Add(format);
            }
        }

        public void AddFont(FontProperties font)
        {
            this.fonts.Add(font);
        }

        public void AddFill(ISpreadFill fill)
        {
            this.fills.Add(fill);
        }

        public void AddBorders(SpreadCellBorders borders)
        {
            this.borders.Add(borders);
        }

        public void AddCellStyleFormat(DiferentialFormat format)
        {
            this.cellStyleFormats.Add(format);
        }

        public void AddCellFormat(DiferentialFormat format)
        {
            this.cellFormats.Add(format);
        }

        public void AddCellStyle(CellStyleInfo style)
        {
            this.cellStyles.Add(style);
        }

        private static DiferentialFormat CopyAsCellDiferentialFormat(DiferentialFormat cellStyleFormat)
        {
            var copy = new DiferentialFormat();

            if (cellStyleFormat.ApplyBorder.HasValue && cellStyleFormat.ApplyBorder.Value)
            {
                copy.BorderId = cellStyleFormat.BorderId;
            }

            if (cellStyleFormat.ApplyFill.HasValue && cellStyleFormat.ApplyFill.Value)
            {
                copy.FillId = cellStyleFormat.FillId;
            }

            if (cellStyleFormat.ApplyFont.HasValue && cellStyleFormat.ApplyFont.Value)
            {
                copy.FontId = cellStyleFormat.FontId;
            }

            if (cellStyleFormat.ApplyNumberFormat.HasValue && cellStyleFormat.ApplyNumberFormat.Value)
            {
                copy.NumFmtId = cellStyleFormat.NumFmtId;
            }

            copy.FormatId = cellStyleFormat.FormatId;

            if (cellStyleFormat.ApplyAlignment.HasValue && cellStyleFormat.ApplyAlignment.Value)
            {
                copy.HorizontalAlignment = cellStyleFormat.HorizontalAlignment;
                copy.VerticalAlignment = cellStyleFormat.VerticalAlignment;
                copy.Indent = cellStyleFormat.Indent;
                copy.WrapText = cellStyleFormat.WrapText;
            }

            if (cellStyleFormat.ApplyProtection.HasValue && cellStyleFormat.ApplyProtection.Value)
            {
                copy.IsLocked = cellStyleFormat.IsLocked;
            }

            copy.ApplyAlignment = false;
            copy.ApplyBorder = false;
            copy.ApplyFill = false;
            copy.ApplyFont = false;
            copy.ApplyNumberFormat = false;
            copy.ApplyProtection = false;

            return copy;
        }

        private DiferentialFormat CellFormatToDiferentialFormat(SpreadCellFormat cellFormat)
        {
            DiferentialFormat format = this.AddCellStyle(cellFormat);

            this.CopyFormatProperties(cellFormat, format);

            return format;
        }

        private DiferentialFormat AddCellStyle(SpreadCellFormat cellFormat)
        {
            DiferentialFormat format = new DiferentialFormat();

            SpreadCellStyle cellStyle = cellFormat.CellStyle;
            if (cellStyle != null)
            {
                DiferentialFormat cellStyleFormat = this.CellStyleToDiferentialFormat(cellStyle);
                format = CopyAsCellDiferentialFormat(cellStyleFormat);

                int formattingRecordId;

                CellStyleInfo cellStyleModel = this.cellStyles.Where(p => p.Name == cellStyle.Name).FirstOrDefault();
                if (cellStyleModel == null)
                {
                    formattingRecordId = this.cellStyleFormats.Add(cellStyleFormat);

                    cellStyleModel = new CellStyleInfo(cellStyle.Name, formattingRecordId, cellStyle.BuiltinId);

                    this.cellStyles.Add(cellStyleModel);
                }
                else
                {
                    formattingRecordId = cellStyleModel.FormattingRecordId;
                    DiferentialFormat oldFormattingRecord = this.cellStyleFormats.GetValueByIndex(formattingRecordId);

                    if (!ObjectExtensions.EqualsOfT(oldFormattingRecord, cellStyleFormat))
                    {
                        this.cellStyleFormats.RemoveAtIndex(formattingRecordId);
                        this.cellStyleFormats.Add(formattingRecordId, cellStyleFormat);

                        this.UpdateOldCellsDiferentialFormats(formattingRecordId, oldFormattingRecord, cellStyleFormat);
                    }
                }

                format.FormatId = formattingRecordId;
            }

            return format;
        }

        private DiferentialFormat CellStyleToDiferentialFormat(SpreadCellStyle cellFormat)
        {
            DiferentialFormat format = new DiferentialFormat();

            this.CopyFormatProperties(cellFormat, format);

            return format;
        }

        private void CopyFormatProperties(SpreadCellFormatBase sourceFormat, DiferentialFormat destinationFormat)
        {
            if (sourceFormat.HasAnyBorderSet || (sourceFormat.ApplyBorderInternal.HasValue && sourceFormat.ApplyBorderInternal.Value))
            {
                SpreadCellBorders borders = new SpreadCellBorders();
                borders.Bottom = sourceFormat.BottomBorder;
                borders.DiagonalDown = sourceFormat.DiagonalDownBorder;
                borders.DiagonalUp = sourceFormat.DiagonalUpBorder;
                borders.Left = sourceFormat.LeftBorder;
                borders.Right = sourceFormat.RightBorder;
                borders.Top = sourceFormat.TopBorder;

                destinationFormat.BorderId = this.borders.Add(borders);
            }

            ISpreadFill fill = null;
            if (sourceFormat.ApplyFillInternal.HasValue && sourceFormat.ApplyFillInternal.Value)
            {
                fill = DefaultCellFormatPropertyValues.Fill;
                if (sourceFormat.Fill != null)
                {
                    fill = sourceFormat.Fill;
                }

                destinationFormat.FillId = this.fills.Add(fill);
            }

            if (sourceFormat.HasTextPropertiesSet || (sourceFormat.ApplyFontInternal.HasValue && sourceFormat.ApplyFontInternal.Value))
            {
                destinationFormat.FontId = this.fonts.Add(FontProperties.FromCellFormat(sourceFormat));
            }

            if (sourceFormat.NumberFormat != null && (sourceFormat.ApplyNumberFormatInternal.HasValue && sourceFormat.ApplyNumberFormatInternal.Value))
            {
                NumberFormat numberFormat = new NumberFormat();
                numberFormat.NumberFormatString = sourceFormat.NumberFormat;

                if (ObjectExtensions.EqualsOfT(sourceFormat.NumberFormat, DefaultCellFormatPropertyValues.NumberFormat))
                {
                    numberFormat.NumberFormatId = 0;
                }
                else
                {
                    numberFormat.NumberFormatId = this.numberFormats.GetNextIndex(numberFormat);
                }

                this.numberFormats.Add(numberFormat.NumberFormatId.Value, numberFormat);
                destinationFormat.NumFmtId = numberFormat.NumberFormatId.Value;
            }

            if (sourceFormat.HorizontalAlignment.HasValue)
            {
                destinationFormat.HorizontalAlignment = sourceFormat.HorizontalAlignment;
            }

            if (sourceFormat.VerticalAlignment.HasValue)
            {
                destinationFormat.VerticalAlignment = sourceFormat.VerticalAlignment;
            }

            if (sourceFormat.Indent.HasValue)
            {
                destinationFormat.Indent = sourceFormat.Indent;
            }

            if (sourceFormat.WrapText.HasValue)
            {
                destinationFormat.WrapText = sourceFormat.WrapText;
            }

            if (sourceFormat.ApplyBorderInternal.HasValue)
            {
                destinationFormat.ApplyBorder = sourceFormat.ApplyBorderInternal;
            }

            if (sourceFormat.ApplyFillInternal.HasValue)
            {
                destinationFormat.ApplyFill = sourceFormat.ApplyFillInternal;
            }

            if (sourceFormat.ApplyFontInternal.HasValue)
            {
                destinationFormat.ApplyFont = sourceFormat.ApplyFontInternal;
            }

            if (sourceFormat.ApplyNumberFormatInternal.HasValue)
            {
                destinationFormat.ApplyNumberFormat = sourceFormat.ApplyNumberFormatInternal;
            }

            if (sourceFormat.ApplyAlignmentInternal.HasValue)
            {
                destinationFormat.ApplyAlignment = sourceFormat.ApplyAlignmentInternal;
            }

            if (sourceFormat.ApplyProtectionInternal.HasValue)
            {
                destinationFormat.ApplyProtection = sourceFormat.ApplyProtectionInternal;
            }
        }

        private void UpdateOldCellsDiferentialFormats(int formattingRecordId, DiferentialFormat oldFormattingRecord, DiferentialFormat cellStyleFormat)
        {
            DiferentialFormat cellFormattingRecord = this.cellFormats.Where(p => p.FormatId == formattingRecordId).FirstOrDefault();
            if (cellFormattingRecord != null)
            {
                int id = this.cellFormats.GetIndexByValue(cellFormattingRecord);
                this.cellFormats.RemoveAtIndex(id);

                if (cellFormattingRecord.BorderId == oldFormattingRecord.BorderId &&
                    cellFormattingRecord.BorderId != cellStyleFormat.BorderId)
                {
                    cellFormattingRecord.BorderId = cellStyleFormat.BorderId;
                }

                if (cellFormattingRecord.FillId == oldFormattingRecord.FillId &&
                    cellFormattingRecord.FillId != cellStyleFormat.FillId)
                {
                    cellFormattingRecord.FillId = cellStyleFormat.FillId;
                }

                if (cellFormattingRecord.FontId == oldFormattingRecord.FontId &&
                    cellFormattingRecord.FontId != cellStyleFormat.FontId)
                {
                    cellFormattingRecord.FontId = cellStyleFormat.FontId;
                }

                if (cellFormattingRecord.FormatId == oldFormattingRecord.FormatId &&
                    cellFormattingRecord.FormatId != cellStyleFormat.FormatId)
                {
                    cellFormattingRecord.FormatId = cellStyleFormat.FormatId;
                }

                if (cellFormattingRecord.HorizontalAlignment == oldFormattingRecord.HorizontalAlignment &&
                    cellFormattingRecord.HorizontalAlignment != cellStyleFormat.HorizontalAlignment)
                {
                    cellFormattingRecord.HorizontalAlignment = cellStyleFormat.HorizontalAlignment;
                }

                if (cellFormattingRecord.Indent == oldFormattingRecord.Indent &&
                    cellFormattingRecord.Indent != cellStyleFormat.Indent)
                {
                    cellFormattingRecord.Indent = cellStyleFormat.Indent;
                }

                if (cellFormattingRecord.IsLocked == oldFormattingRecord.IsLocked &&
                    cellFormattingRecord.IsLocked != cellStyleFormat.IsLocked)
                {
                    cellFormattingRecord.IsLocked = cellStyleFormat.IsLocked;
                }

                if (cellFormattingRecord.NumFmtId == oldFormattingRecord.NumFmtId &&
                    cellFormattingRecord.NumFmtId != cellStyleFormat.NumFmtId)
                {
                    cellFormattingRecord.NumFmtId = cellStyleFormat.NumFmtId;
                }

                if (cellFormattingRecord.VerticalAlignment == oldFormattingRecord.VerticalAlignment &&
                    cellFormattingRecord.VerticalAlignment != cellStyleFormat.VerticalAlignment)
                {
                    cellFormattingRecord.VerticalAlignment = cellStyleFormat.VerticalAlignment;
                }

                if (cellFormattingRecord.WrapText == oldFormattingRecord.WrapText &&
                    cellFormattingRecord.WrapText != cellStyleFormat.WrapText)
                {
                    cellFormattingRecord.WrapText = cellStyleFormat.WrapText;
                }

                this.cellFormats.Add(id, cellFormattingRecord);
            }
        }

        private void InitializeBuiltInStyles()
        {
            SpreadCellFormat format = new SpreadCellFormat();
            format.CellStyle = this.normalStyle;
            this.RegisterCellFormat(format);

            this.fills.Add(new SpreadPatternFill(SpreadPatternType.Gray12Percent));
        }
    }
}