﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace Telerik.Documents.SpreadsheetStreaming.Model
{
    internal class SheetView
    {
        private int topLeftCellRowIndex;
        private int topLeftCellColumnIndex;
        private int activeSelectionCellRowIndex;
        private int activeSelectionCellColumnIndex;
        private List<SpreadCellRange> selectedRanges;
        private bool isActiveSelectionCellSet;

        public SheetView()
        {
            this.Pane = new Pane();
            this.TopLeftSelection = new Selection(SpreadPane.TopLeft);
            this.TopRightSelection = new Selection(SpreadPane.TopRight);
            this.BottomLeftSelection = new Selection(SpreadPane.BottomLeft);
            this.BottomRightSelection = new Selection(SpreadPane.BottomRight);

            this.ShouldShowGridLines = true;
            this.ShouldShowRowColumnHeaders = true;
        }

        public bool HasFreezePanes { get; private set; }

        public bool IsFirstRowFreeze { get; private set; }

        public bool IsFirstColumnFreeze { get; private set; }

        public int? ScaleFactor { get; private set; }

        public bool? ShouldShowGridLines { get; private set; }

        public bool? ShouldShowRowColumnHeaders { get; private set; }

        public Pane Pane { get; private set; }

        public Selection TopLeftSelection { get; private set; }

        public Selection TopRightSelection { get; private set; }

        public Selection BottomLeftSelection { get; private set; }

        public Selection BottomRightSelection { get; private set; }

        public void PrepareForExport()
        {
            Selection selection = this.TopLeftSelection;
            if (this.HasFreezePanes)
            {
                selection = this.BottomRightSelection;
            }

            if (!this.isActiveSelectionCellSet && this.selectedRanges != null)
            {
                var firstSelectionRange = this.selectedRanges.First();
                this.activeSelectionCellRowIndex = firstSelectionRange.FromRowIndex;
                this.activeSelectionCellColumnIndex = firstSelectionRange.FromColumnIndex;
            }

            if (this.isActiveSelectionCellSet && this.selectedRanges != null)
            {
                var rangesContainingActiveCell = this.selectedRanges.Where(p => p.ContainsCell(this.activeSelectionCellRowIndex, this.activeSelectionCellColumnIndex));
                if (!rangesContainingActiveCell.Any())
                {
                    throw new InvalidOperationException("The active cell should be part of the selection");
                }
            }

            selection.SetActiveCell(this.activeSelectionCellRowIndex, this.activeSelectionCellColumnIndex);
            selection.SetSelectedRanges(this.selectedRanges);
        }

        internal void SetTopLeftCell(int rowIndex, int columnIndex)
        {
            this.SetTopLeftCell(SpreadPane.TopLeft, rowIndex, columnIndex);
        }

        internal void SetTopLeftCell(SpreadPane pane, int rowIndex, int columnIndex)
        {
            switch (pane)
            {
                case SpreadPane.TopLeft:
                    {
                        this.topLeftCellRowIndex = rowIndex;
                        this.topLeftCellColumnIndex = columnIndex;
                        this.TopLeftSelection.SetActiveCell(rowIndex, columnIndex);
                        this.TopRightSelection.SetActiveCellRowIndex(rowIndex);
                        this.BottomLeftSelection.SetActiveCellColumnIndex(columnIndex);
                    }
                    break;
                case SpreadPane.TopRight:
                    {
                        this.topLeftCellRowIndex = rowIndex;
                        this.TopRightSelection.SetActiveCell(rowIndex, columnIndex);
                        this.TopLeftSelection.SetActiveCellRowIndex(rowIndex);
                        this.Pane.SetTopLeftColumnIndex(columnIndex);
                    }
                    break;
                case SpreadPane.BottomLeft:
                    {
                        this.BottomLeftSelection.SetActiveCell(rowIndex, columnIndex);
                        this.TopLeftSelection.SetActiveCellColumnIndex(columnIndex);
                        this.topLeftCellColumnIndex = columnIndex;
                        this.Pane.SetTopLeftRowIndex(rowIndex);
                    }
                    break;
                case SpreadPane.BottomRight:
                    {
                        this.TopRightSelection.SetActiveCellColumnIndex(columnIndex);
                        this.BottomLeftSelection.SetActiveCellRowIndex(rowIndex);
                        this.Pane.SetTopLeftCell(rowIndex, columnIndex);
                    }
                    break;
            }
        }

        internal void GetTopLeftCell(SpreadPane pane, out int rowIndex, out int columnIndex)
        {
            switch (pane)
            {
                case SpreadPane.TopLeft:
                    {
                        rowIndex = this.topLeftCellRowIndex;
                        columnIndex = this.topLeftCellColumnIndex;
                    }
                    break;
                case SpreadPane.TopRight:
                    {
                        this.TopRightSelection.GetActiveCell(out rowIndex, out columnIndex);
                    }
                    break;
                case SpreadPane.BottomLeft:
                    {
                        this.BottomLeftSelection.GetActiveCell(out rowIndex, out columnIndex);
                    }
                    break;
                case SpreadPane.BottomRight:
                    {
                        this.Pane.GetTopLeftCell(out rowIndex, out columnIndex);
                    }
                    break;

                default:
                    throw new ArgumentException("Not expected pane value.");
            }
        }

        internal void SetSplitOffset(int rowOffset, int columnOffset)
        {
            this.HasFreezePanes = rowOffset > 0 || columnOffset > 0;
            this.IsFirstColumnFreeze = rowOffset == 0 && columnOffset > 0;
            this.IsFirstRowFreeze = rowOffset > 0 && columnOffset == 0;

            this.Pane.SetSplitOffset(rowOffset, columnOffset);
            this.Pane.SetActivePane(SpreadPane.BottomRight);
            this.Pane.SetTopLeftCell(this.topLeftCellRowIndex + rowOffset, this.topLeftCellColumnIndex + columnOffset);

            this.TopRightSelection.SetActiveCell(this.topLeftCellRowIndex, this.topLeftCellColumnIndex + columnOffset);
            this.BottomLeftSelection.SetActiveCell(this.topLeftCellRowIndex + rowOffset, this.topLeftCellColumnIndex);
            this.BottomRightSelection.SetActiveCell(this.topLeftCellRowIndex + rowOffset, this.topLeftCellColumnIndex + columnOffset);
        }

        internal void SetScaleFactor(double percents)
        {
            this.ScaleFactor = (int)(percents * 100f);
        }

        internal void SetShouldShowGridLines(bool value)
        {
            this.ShouldShowGridLines = value;
        }

        internal void SetShouldShowRowColumnHeaders(bool value)
        {
            this.ShouldShowRowColumnHeaders = value;
        }

        internal void SetActiveSelectionCell(int rowIndex, int columnIndex)
        {
            this.isActiveSelectionCellSet = true;

            this.activeSelectionCellRowIndex = rowIndex;
            this.activeSelectionCellColumnIndex = columnIndex;
        }

        internal void AddSelectedRange(SpreadCellRange selectedRange)
        {
            if (this.selectedRanges == null)
            {
                this.selectedRanges = new List<SpreadCellRange>();
            }

            this.selectedRanges.Add(selectedRange);
        }
    }
}