﻿using System;
using System.Linq;

namespace Telerik.Documents.SpreadsheetStreaming
{
    /// <summary>
    /// Represent the base class for cell formats.
    /// </summary>
    public abstract class SpreadCellFormatBase
    {
        private SpreadBorder leftBorder;
        private SpreadBorder rightBorder;
        private SpreadBorder topBorder;
        private SpreadBorder bottomBorder;
        private SpreadBorder diagonalUpBorder;
        private SpreadBorder diagonalDownBorder;
        private ISpreadFill fill;
        private string numberFormat;
        private double? fontSize;
        private SpreadThemableColor foreColor;
        private bool? isBold;
        private bool? isItalic;
        private SpreadUnderlineType? underline;
        private SpreadThemableFontFamily fontFamily;
        private SpreadHorizontalAlignment? horizontalAlignment;
        private SpreadVerticalAlignment? verticalAlignment;
        private int? indent;
        private bool? wrapText;

        /// <summary>
        /// Gets or sets the left border.
        /// </summary>
        /// <value>The left border.</value>
        public SpreadBorder LeftBorder
        {
            get
            {
                return this.leftBorder;
            }
            set
            {
                if (this.leftBorder != value)
                {
                    this.leftBorder = value;
                    this.ApplyBorderInternal = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets the right border.
        /// </summary>
        /// <value>The right border.</value>
        public SpreadBorder RightBorder
        {
            get
            {
                return this.rightBorder;
            }
            set
            {
                if (this.rightBorder != value)
                {
                    this.rightBorder = value;
                    this.ApplyBorderInternal = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets the top border.
        /// </summary>
        /// <value>The top border.</value>
        public SpreadBorder TopBorder
        {
            get
            {
                return this.topBorder;
            }
            set
            {
                if (this.topBorder != value)
                {
                    this.topBorder = value;
                    this.ApplyBorderInternal = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets the bottom border.
        /// </summary>
        /// <value>The bottom border.</value>
        public SpreadBorder BottomBorder
        {
            get
            {
                return this.bottomBorder;
            }
            set
            {
                if (this.bottomBorder != value)
                {
                    this.bottomBorder = value;
                    this.ApplyBorderInternal = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets the diagonal up border.
        /// </summary>
        /// <value>The diagonal up border.</value>
        public SpreadBorder DiagonalUpBorder
        {
            get
            {
                return this.diagonalUpBorder;
            }
            set
            {
                if (this.diagonalUpBorder != value)
                {
                    this.diagonalUpBorder = value;
                    this.ApplyBorderInternal = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets the diagonal down border.
        /// </summary>
        /// <value>The diagonal down border.</value>
        public SpreadBorder DiagonalDownBorder
        {
            get
            {
                return this.diagonalDownBorder;
            }
            set
            {
                if (this.diagonalDownBorder != value)
                {
                    this.diagonalDownBorder = value;
                    this.ApplyBorderInternal = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets the fill.
        /// </summary>
        /// <value>The fill.</value>
        public ISpreadFill Fill
        {
            get
            {
                return this.fill;
            }
            set
            {
                if (this.fill != value)
                {
                    this.fill = value;
                    this.ApplyFillInternal = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets the number format.
        /// </summary>
        /// <value>The number format.</value>
        public string NumberFormat
        {
            get
            {
                return this.numberFormat;
            }
            set
            {
                if (this.numberFormat != value)
                {
                    this.numberFormat = value;
                    this.ApplyNumberFormatInternal = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets the size of the font.
        /// </summary>
        /// <value>The size of the font.</value>
        public double? FontSize
        {
            get
            {
                return this.fontSize;
            }
            set
            {
                if (this.fontSize != value)
                {
                    this.fontSize = value;
                    this.ApplyFontInternal = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets the fore color.
        /// </summary>
        /// <value>The fore color.</value>
        public SpreadThemableColor ForeColor
        {
            get
            {
                return this.foreColor;
            }
            set
            {
                if (this.foreColor != value)
                {
                    this.foreColor = value;
                    this.ApplyFontInternal = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the text is bold.
        /// </summary>
        /// <value>The value indicating whether the text bold.</value>
        public bool? IsBold
        {
            get
            {
                return this.isBold;
            }
            set
            {
                if (this.isBold != value)
                {
                    this.isBold = value;
                    this.ApplyFontInternal = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the text is italic.
        /// </summary>
        /// <value>The value indicating whether the text italic.</value>
        public bool? IsItalic
        {
            get
            {
                return this.isItalic;
            }
            set
            {
                if (this.isItalic != value)
                {
                    this.isItalic = value;
                    this.ApplyFontInternal = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets the underline type.
        /// </summary>
        /// <value>The underline type.</value>
        public SpreadUnderlineType? Underline
        {
            get
            {
                return this.underline;
            }
            set
            {
                if (this.underline != value)
                {
                    this.underline = value;
                    this.ApplyFontInternal = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets the font family.
        /// </summary>
        /// <value>The font family.</value>
        public SpreadThemableFontFamily FontFamily
        {
            get
            {
                return this.fontFamily;
            }
            set
            {
                if (this.fontFamily != value)
                {
                    this.fontFamily = value;
                    this.ApplyFontInternal = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets the horizontal alignment.
        /// </summary>
        /// <value>The horizontal alignment.</value>
        public SpreadHorizontalAlignment? HorizontalAlignment
        {
            get
            {
                return this.horizontalAlignment;
            }
            set
            {
                if (this.horizontalAlignment != value)
                {
                    this.horizontalAlignment = value;
                    this.ApplyAlignmentInternal = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets the vertical alignment.
        /// </summary>
        /// <value>The vertical alignment.</value>
        public SpreadVerticalAlignment? VerticalAlignment
        {
            get
            {
                return this.verticalAlignment;
            }
            set
            {
                if (this.verticalAlignment != value)
                {
                    this.verticalAlignment = value;
                    this.ApplyAlignmentInternal = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets the indent.
        /// </summary>
        /// <value>The indent.</value>
        public int? Indent
        {
            get
            {
                return this.indent;
            }
            set
            {
                if (this.indent != value)
                {
                    this.indent = value;
                    this.ApplyAlignmentInternal = true;
                }
            }
        }

        /// <summary>
        /// Gets or sets a value indicating if the text in a cell should be line-wrapped within the cell.
        /// </summary>
        /// <value>The value indicating if the text in a cell should be line-wrapped within the cell.</value>
        public bool? WrapText
        {
            get
            {
                return this.wrapText;
            }
            set
            {
                if (this.wrapText != value)
                {
                    this.wrapText = value;
                    this.ApplyAlignmentInternal = true;
                }
            }
        }

        internal bool? ApplyBorderInternal { get; set; }

        internal bool? ApplyFillInternal { get; set; }

        internal bool? ApplyFontInternal { get; set; }

        internal bool? ApplyNumberFormatInternal { get; set; }

        internal bool? ApplyAlignmentInternal { get; set; }

        internal bool? ApplyProtectionInternal { get; set; }

        internal bool HasTextPropertiesSet
        {
            get
            {
                return this.FontSize.HasValue ||
                    this.ForeColor != null ||
                    this.IsBold.HasValue ||
                    this.IsItalic.HasValue ||
                    this.Underline.HasValue ||
                    this.FontFamily != null;
            }
        }

        internal bool HasAnyBorderSet
        {
            get
            {
                return this.BottomBorder != null ||
                    this.DiagonalDownBorder != null ||
                    this.DiagonalUpBorder != null ||
                    this.LeftBorder != null ||
                    this.RightBorder != null ||
                    this.TopBorder != null;
            }
        }
    }
}