﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Telerik.Documents.SpreadsheetStreaming.Utilities;

namespace Telerik.Documents.SpreadsheetStreaming
{
    /// <summary>
    /// Represents a cell style collection.
    /// </summary>
    public class SpreadCellStyleCollection : IEnumerable<SpreadCellStyle>
    {
        internal static readonly string DefaultStyleName = "Normal";

        private static readonly Dictionary<string, int> styleToBuiltinId;
        private readonly Dictionary<string, SpreadCellStyle> nameToStyles;

        static SpreadCellStyleCollection()
        {
            styleToBuiltinId = new Dictionary<string, int>();
            RegisterStyleBuiltinId(DefaultStyleName, 0);
            RegisterStyleBuiltinId("Comma", 3);
            RegisterStyleBuiltinId("Currency", 4);
            RegisterStyleBuiltinId("Percent", 5);
            RegisterStyleBuiltinId("Comma [0]", 6);
            RegisterStyleBuiltinId("Currency [0]", 7);
            RegisterStyleBuiltinId("Hyperlink", 8);
            RegisterStyleBuiltinId("Followed Hyperlink", 9);
            RegisterStyleBuiltinId("Note", 10);
            RegisterStyleBuiltinId("Warning Text", 11);
            RegisterStyleBuiltinId("Title", 15);
            RegisterStyleBuiltinId("Heading 1", 16);
            RegisterStyleBuiltinId("Heading 2", 17);
            RegisterStyleBuiltinId("Heading 3", 18);
            RegisterStyleBuiltinId("Heading 4", 19);
            RegisterStyleBuiltinId("Input", 20);
            RegisterStyleBuiltinId("Output", 21);
            RegisterStyleBuiltinId("Calculation", 22);
            RegisterStyleBuiltinId("Check Cell", 23);
            RegisterStyleBuiltinId("Linked Cell", 24);
            RegisterStyleBuiltinId("Total", 25);
            RegisterStyleBuiltinId("Good", 26);
            RegisterStyleBuiltinId("Bad", 27);
            RegisterStyleBuiltinId("Neutral", 28);

            RegisterStyleBuiltinId("Accent1", 29);
            RegisterStyleBuiltinId("20% - Accent1", 30);
            RegisterStyleBuiltinId("40% - Accent1", 31);
            RegisterStyleBuiltinId("60% - Accent1", 32);

            RegisterStyleBuiltinId("Accent2", 33);
            RegisterStyleBuiltinId("20% - Accent2", 34);
            RegisterStyleBuiltinId("40% - Accent2", 35);
            RegisterStyleBuiltinId("60% - Accent2", 36);

            RegisterStyleBuiltinId("Accent3", 37);
            RegisterStyleBuiltinId("20% - Accent3", 38);
            RegisterStyleBuiltinId("40% - Accent3", 39);
            RegisterStyleBuiltinId("60% - Accent3", 40);

            RegisterStyleBuiltinId("Accent4", 41);
            RegisterStyleBuiltinId("20% - Accent4", 42);
            RegisterStyleBuiltinId("40% - Accent4", 43);
            RegisterStyleBuiltinId("60% - Accent4", 44);

            RegisterStyleBuiltinId("Accent5", 45);
            RegisterStyleBuiltinId("20% - Accent5", 46);
            RegisterStyleBuiltinId("40% - Accent5", 47);
            RegisterStyleBuiltinId("60% - Accent5", 48);

            RegisterStyleBuiltinId("Accent6", 49);
            RegisterStyleBuiltinId("20% - Accent6", 50);
            RegisterStyleBuiltinId("40% - Accent6", 51);
            RegisterStyleBuiltinId("60% - Accent6", 52);
            RegisterStyleBuiltinId("Explanatory Text", 53);
        }

        internal SpreadCellStyleCollection()
        {
            this.nameToStyles = new Dictionary<string, SpreadCellStyle>();

            this.InitDefaultStyles();
        }

        /// <summary>
        /// Gets the number of elements contained in this collection.
        /// </summary>
        /// <value>The number of elements contained in this collection.</value>
        public int Count
        {
            get
            {
                return this.nameToStyles.Keys.Count;
            }
        }

        /// <summary>
        /// Gets the <see cref="SpreadCellStyle" /> with the specified style name.
        /// </summary>
        /// <value>The spread cell style.</value>
        public SpreadCellStyle this[string styleName]
        {
            get
            {
                return this.GetByName(styleName);
            }
        }

        /// <summary>
        /// Determines whether the style with the specified name is in the <see cref="SpreadCellStyleCollection"/>.
        /// </summary>
        /// <param name="styleName">The name of the style.</param>
        /// <returns>True if it is present; otherwise, false.</returns>
        public bool Contains(string styleName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(styleName, "styleName");
            return this.nameToStyles.ContainsKey(styleName);
        }

        /// <summary>
        /// Determines whether the specified style is in the <see cref="SpreadCellStyleCollection"/>.
        /// </summary>
        /// <param name="style">The style.</param>
        /// <returns>True if it is present; otherwise, false.</returns>
        public bool Contains(SpreadCellStyle style)
        {
            Guard.ThrowExceptionIfNull(style, "style");
            return this.Contains(style.Name);
        }

        /// <summary>
        ///  Gets the <see cref="SpreadCellStyle" /> with the specified style name.
        /// </summary>
        /// <param name="styleName">The name of the style.</param>
        /// <returns>The style.</returns>
        public SpreadCellStyle GetByName(string styleName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(styleName, "styleName");

            SpreadCellStyle result = null;
            this.nameToStyles.TryGetValue(styleName, out result);

            return result;
        }

        /// <summary>
        /// Creates a new cell style with the specified name.
        /// </summary>
        /// <param name="styleName">The name of the new cell style.</param>
        /// <returns>The style.</returns>
        public SpreadCellStyle Add(string styleName)
        {
            Guard.ThrowExceptionIfNullOrEmpty(styleName, "styleName");

            if (this.Contains(styleName))
            {
                throw new InvalidOperationException(
                    string.Format("Style with name '{0}' already exists",
                    styleName));
            }

            SpreadCellStyle style = CreateCellStyle(styleName);

            this.AddInternal(style);

            return style;
        }

        /// <summary>
        /// Returns an enumerator that iterates through the the <see cref="SpreadCellStyleCollection"/>.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1" /> that can
        /// be used to iterate through the collection.
        /// </returns>
        public IEnumerator<SpreadCellStyle> GetEnumerator()
        {
            return this.nameToStyles.Values.GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that iterates through a the <see cref="SpreadCellStyleCollection"/>.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be
        /// used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return ((IEnumerable<SpreadCellStyle>)this).GetEnumerator();
        }

        internal void AddInternal(SpreadCellStyle style)
        {
            this.nameToStyles.Add(style.Name, style);
        }

        private static int? GetStyleBuiltinId(string styleName)
        {
            int id;
            if (!styleToBuiltinId.TryGetValue(styleName, out id))
            {
                return null;
            }

            return id;
        }

        private static void RegisterStyleBuiltinId(string styleName, int builtinId)
        {
            styleToBuiltinId.Add(styleName, builtinId);
        }

        private static SpreadCellStyle CreateCellStyle(string styleName)
        {
            int? builtinId = GetStyleBuiltinId(styleName);
            return new SpreadCellStyle(styleName, builtinId);
        }

        private void InitDefaultStyles()
        {
            this.AddGoodBadAndNeutralStyles();
            this.AddDataAndModelStyles();
            this.AddTitlesAndHeadingsStyles();
            this.AddThemedStyles();
            this.AddNumberFormatStyles();
        }

        private void AddGoodBadAndNeutralStyles()
        {
            SpreadCellStyle normalStyle = CreateCellStyle(DefaultStyleName);
            this.AddInternal(normalStyle);

            SpreadCellStyle badStyle = CreateCellStyle("Bad");
            badStyle.ForeColor = SpreadThemableColor.FromRgb(156, 0, 6);
            badStyle.Fill = SpreadPatternFill.CreateSolidFill(new SpreadColor(255, 199, 206));
            this.AddInternal(badStyle);

            SpreadCellStyle goodStyle = CreateCellStyle("Good");
            goodStyle.ForeColor = SpreadThemableColor.FromRgb(0, 97, 0);
            goodStyle.Fill = SpreadPatternFill.CreateSolidFill(new SpreadColor(198, 239, 206));
            this.AddInternal(goodStyle);

            SpreadCellStyle neutralStyle = CreateCellStyle("Neutral");
            neutralStyle.ForeColor = SpreadThemableColor.FromRgb(156, 101, 0);
            neutralStyle.Fill = SpreadPatternFill.CreateSolidFill(new SpreadColor(255, 235, 156));
            this.AddInternal(neutralStyle);
        }

        private void AddDataAndModelStyles()
        {
            SpreadCellStyle calculationStyle = CreateCellStyle("Calculation");
            calculationStyle.ForeColor = SpreadThemableColor.FromRgb(250, 125, 0);
            SpreadBorder cellBorder = new SpreadBorder(SpreadBorderStyle.Thin, SpreadThemableColor.FromRgb(127, 127, 127));
            calculationStyle.LeftBorder = cellBorder;
            calculationStyle.TopBorder = cellBorder;
            calculationStyle.RightBorder = cellBorder;
            calculationStyle.BottomBorder = cellBorder;
            calculationStyle.Fill = SpreadPatternFill.CreateSolidFill(new SpreadColor(242, 242, 242));
            this.AddInternal(calculationStyle);

            SpreadCellStyle checkCellStyle = CreateCellStyle("Check Cell");
            checkCellStyle.ForeColor = new SpreadThemableColor(SpreadThemeColorType.Light1);
            SpreadBorder checkCellBorder = new SpreadBorder(SpreadBorderStyle.Double, SpreadThemableColor.FromRgb(127, 127, 127));
            checkCellStyle.LeftBorder = checkCellBorder;
            checkCellStyle.TopBorder = checkCellBorder;
            checkCellStyle.RightBorder = checkCellBorder;
            checkCellStyle.BottomBorder = checkCellBorder;
            checkCellStyle.Fill = SpreadPatternFill.CreateSolidFill(new SpreadColor(165, 165, 165));

            this.AddInternal(checkCellStyle);

            SpreadCellStyle explanatoryStyle = CreateCellStyle("Explanatory Text");
            explanatoryStyle.ForeColor = SpreadThemableColor.FromRgb(127, 127, 127);
            explanatoryStyle.IsItalic = true;
            this.AddInternal(explanatoryStyle);

            SpreadCellStyle followedHyperlink = CreateCellStyle("Followed Hyperlink");
            followedHyperlink.FontFamily = new SpreadThemableFontFamily(SpreadThemeFontType.Minor);
            followedHyperlink.Underline = SpreadUnderlineType.Single;
            followedHyperlink.ForeColor = new SpreadThemableColor(SpreadThemeColorType.FollowedHyperlink);
            this.AddInternal(followedHyperlink);

            SpreadCellStyle hyperlink = CreateCellStyle("Hyperlink");
            hyperlink.FontFamily = new SpreadThemableFontFamily(SpreadThemeFontType.Minor);
            hyperlink.Underline = SpreadUnderlineType.Single;
            hyperlink.ForeColor = new SpreadThemableColor(SpreadThemeColorType.Hyperlink);
            this.AddInternal(hyperlink);

            SpreadCellStyle inputStyle = CreateCellStyle("Input");
            inputStyle.ForeColor = SpreadThemableColor.FromRgb(63, 63, 118);
            SpreadBorder inputCellBorder = new SpreadBorder(SpreadBorderStyle.Thin, new SpreadThemableColor(new SpreadColor(127, 127, 127)));
            inputStyle.LeftBorder = inputCellBorder;
            inputStyle.TopBorder = inputCellBorder;
            inputStyle.RightBorder = inputCellBorder;
            inputStyle.BottomBorder = inputCellBorder;
            inputStyle.Fill = SpreadPatternFill.CreateSolidFill(new SpreadColor(255, 204, 153));
            this.AddInternal(inputStyle);

            SpreadCellStyle linkedCellStyle = CreateCellStyle("Linked Cell");
            linkedCellStyle.ForeColor = SpreadThemableColor.FromRgb(250, 125, 0);
            SpreadBorder linkedCellBorder = new SpreadBorder(SpreadBorderStyle.Double, new SpreadThemableColor(new SpreadColor(255, 128, 1)));
            linkedCellStyle.BottomBorder = linkedCellBorder;
            this.AddInternal(linkedCellStyle);

            SpreadCellStyle noteStyle = CreateCellStyle("Note");
            SpreadBorder noteBorder = new SpreadBorder(SpreadBorderStyle.Thin, new SpreadThemableColor(new SpreadColor(178, 178, 178)));
            noteStyle.LeftBorder = noteBorder;
            noteStyle.TopBorder = noteBorder;
            noteStyle.RightBorder = noteBorder;
            noteStyle.BottomBorder = noteBorder;
            noteStyle.Fill = SpreadPatternFill.CreateSolidFill(new SpreadColor(255, 255, 204));
            this.AddInternal(noteStyle);

            SpreadCellStyle outputStyle = CreateCellStyle("Output");
            outputStyle.ForeColor = SpreadThemableColor.FromRgb(63, 63, 63);
            outputStyle.IsBold = true;
            SpreadBorder outputBorder = new SpreadBorder(SpreadBorderStyle.Thin, SpreadThemableColor.FromRgb(63, 63, 63));
            outputStyle.LeftBorder = outputBorder;
            outputStyle.TopBorder = outputBorder;
            outputStyle.RightBorder = outputBorder;
            outputStyle.BottomBorder = outputBorder;
            outputStyle.Fill = SpreadPatternFill.CreateSolidFill(new SpreadColor(242, 242, 242));
            this.AddInternal(outputStyle);

            SpreadCellStyle warningTextStyle = CreateCellStyle("Warning Text");
            warningTextStyle.ForeColor = SpreadThemableColor.FromRgb(255, 0, 0);
            this.AddInternal(warningTextStyle);
        }

        private void AddTitlesAndHeadingsStyles()
        {
            SpreadCellStyle heading1Style = CreateCellStyle("Heading 1");
            heading1Style.IsBold = true;
            heading1Style.FontSize = 15;
            heading1Style.ForeColor = new SpreadThemableColor(SpreadThemeColorType.Dark2);
            SpreadBorder heading1Border = new SpreadBorder(SpreadBorderStyle.Thick, new SpreadThemableColor(SpreadThemeColorType.Accent1));
            heading1Style.BottomBorder = heading1Border;
            this.AddInternal(heading1Style);

            SpreadCellStyle heading2Style = CreateCellStyle("Heading 2");
            heading2Style.IsBold = true;
            heading2Style.FontSize = 13;
            heading2Style.ForeColor = new SpreadThemableColor(SpreadThemeColorType.Dark2);
            SpreadBorder heading2Border = new SpreadBorder(SpreadBorderStyle.Thick, new SpreadThemableColor(SpreadThemeColorType.Accent1, SpreadColorShadeType.Shade2));
            heading2Style.BottomBorder = heading2Border;
            this.AddInternal(heading2Style);

            SpreadCellStyle heading3Style = CreateCellStyle("Heading 3");
            heading3Style.IsBold = true;
            heading3Style.ForeColor = new SpreadThemableColor(SpreadThemeColorType.Dark2);
            SpreadBorder heading3Border = new SpreadBorder(SpreadBorderStyle.Medium, new SpreadThemableColor(SpreadThemeColorType.Accent1, SpreadColorShadeType.Shade2));
            heading3Style.BottomBorder = heading3Border;
            this.AddInternal(heading3Style);

            SpreadCellStyle heading4Style = CreateCellStyle("Heading 4");
            heading4Style.IsBold = true;
            heading4Style.ForeColor = new SpreadThemableColor(SpreadThemeColorType.Dark2);
            this.AddInternal(heading4Style);

            SpreadCellStyle titleStyle = CreateCellStyle("Title");
            titleStyle.FontFamily = new SpreadThemableFontFamily(SpreadThemeFontType.Major);
            titleStyle.FontSize = 18;
            titleStyle.IsBold = true;
            titleStyle.ForeColor = new SpreadThemableColor(SpreadThemeColorType.Dark2);
            this.AddInternal(titleStyle);

            SpreadCellStyle totalStyle = CreateCellStyle("Total");
            totalStyle.IsBold = true;
            totalStyle.ForeColor = new SpreadThemableColor(SpreadThemeColorType.Dark1);
            SpreadBorder totalStyleTopBorder = new SpreadBorder(SpreadBorderStyle.Thin, new SpreadThemableColor(SpreadThemeColorType.Accent1, SpreadColorShadeType.Shade3));
            totalStyle.TopBorder = totalStyleTopBorder;
            SpreadBorder totalStyleBottomBorder = new SpreadBorder(SpreadBorderStyle.Double, new SpreadThemableColor(SpreadThemeColorType.Accent1, SpreadColorShadeType.Shade3));
            totalStyle.BottomBorder = totalStyleBottomBorder;
            this.AddInternal(totalStyle);
        }

        private void AddThemedStyles()
        {
            double tintAndShade = 0.8;
            for (int i = 0; i < 4; i++)
            {
                SpreadThemableColor foreColor = new SpreadThemableColor((i < 2) ? SpreadThemeColorType.Dark1 : SpreadThemeColorType.Light1);
                for (int j = (int)SpreadThemeColorType.Accent1; j <= (int)SpreadThemeColorType.Accent6; j++)
                {
                    SpreadThemeColorType themeColorType = (SpreadThemeColorType)j;
                    string styleName = themeColorType.ToString();
                    if (tintAndShade != 0)
                    {
                        styleName = string.Format("{0}% - ", (1 - tintAndShade) * 100) + styleName;
                    }

                    SpreadCellStyle style = CreateCellStyle(styleName);
                    style.ForeColor = foreColor;
                    if (tintAndShade == 0)
                    {
                        style.Fill = SpreadPatternFill.CreateSolidFill(new SpreadThemableColor(themeColorType, null));
                    }
                    else
                    {
                        style.Fill = SpreadPatternFill.CreateSolidFill(new SpreadThemableColor(themeColorType, tintAndShade));
                    }
                    this.AddInternal(style);
                }

                if (i < 2)
                {
                    tintAndShade -= 0.2;
                }
                else
                {
                    tintAndShade = 0;
                }
            }
        }

        private void AddNumberFormatStyles()
        {
            SpreadCellStyle commaStyle = CreateCellStyle("Comma");
            commaStyle.NumberFormat = "_(* #,##0.00_);_(* (#,##0.00);_(* \"-\"??_);_(@_)";
            this.AddInternal(commaStyle);

            SpreadCellStyle commaZeroStyle = CreateCellStyle("Comma [0]");
            commaZeroStyle.NumberFormat = "_(* #,##0_);_(* (#,##0);_(* \"-\"_);_(@_)";
            this.AddInternal(commaZeroStyle);

            SpreadCellStyle currencyStyle = CreateCellStyle("Currency");
            currencyStyle.NumberFormat = "_(\\$* #,##0.00_);_(\\$* (#,##0.00);_(\\$* \"-\"??_);_(@_)";
            this.AddInternal(currencyStyle);

            SpreadCellStyle currencyZeroStyle = CreateCellStyle("Currency [0]");
            currencyZeroStyle.NumberFormat = "_(\\$* #,##0_);_(\\$* (#,##0);_(\\$* \"-\"_);_(@_)";
            this.AddInternal(currencyZeroStyle);

            SpreadCellStyle percentStyle = CreateCellStyle("Percent");
            percentStyle.NumberFormat = "0%";
            this.AddInternal(percentStyle);
        }
    }
}