﻿using System;
using System.IO;
using Telerik.Documents.SpreadsheetStreaming.Exporters.Csv;
using Telerik.Documents.SpreadsheetStreaming.Exporters.Xlsx;
using Telerik.Documents.SpreadsheetStreaming.Utilities;

namespace Telerik.Documents.SpreadsheetStreaming
{
    /// <summary>
    /// Represents an entry point to export a spreadsheet.
    /// </summary>
    public static class SpreadExporter
    {
        /// <summary>
        /// Creates a workbook exporter.
        /// </summary>
        /// <param name="documentFormat">The document format.</param>
        /// <param name="stream">The output stream.</param>
        /// <returns>The workbook exporter.</returns>
        public static IWorkbookExporter CreateWorkbookExporter(SpreadDocumentFormat documentFormat, Stream stream)
        {
            return CreateWorkbookExporter(documentFormat, stream, SpreadExportMode.Create);
        }

        /// <summary>
        /// Creates the workbook exporter.
        /// </summary>
        /// <param name="documentFormat">The document format.</param>
        /// <param name="stream">The stream.</param>
        /// <param name="exportMode">The export mode.</param>
        /// <returns>The workbook exporter.</returns>
        public static IWorkbookExporter CreateWorkbookExporter(SpreadDocumentFormat documentFormat, Stream stream, SpreadExportMode exportMode)
        {
            Guard.ThrowExceptionIfNull(stream, "stream");

            stream.Seek(0, SeekOrigin.Begin);

            switch (documentFormat)
            {
                case SpreadDocumentFormat.Xlsx:
                    return CreateXlsxWorkbookExporter(stream, exportMode);
                case SpreadDocumentFormat.Csv:
                    return CreateCsvWorkbookExporter(stream, exportMode);
                default:
                    throw new NotSupportedException();
            }
        }

        private static IWorkbookExporter CreateXlsxWorkbookExporter(Stream stream, SpreadExportMode exportMode)
        {
            switch (exportMode)
            {
                case SpreadExportMode.Create:
                    {
                        return new XlsxWorkbookExporter(stream);
                    }
                case SpreadExportMode.Append:
                    {
                        if (stream.Length > 0)
                        {
                            using (MemoryStream sourceStream = new MemoryStream())
                            {
                                stream.CopyTo(sourceStream);
                                stream.Seek(0, SeekOrigin.Begin);
                                return new XlsxAppendWorkbookExporter(sourceStream, stream);
                            }
                        }
                        else
                        {
                            return new XlsxWorkbookExporter(stream);
                        }
                    }
                default:
                    throw new NotSupportedException();
            }
        }

        private static IWorkbookExporter CreateCsvWorkbookExporter(Stream stream, SpreadExportMode exportMode)
        {
            switch (exportMode)
            {
                case SpreadExportMode.Create:
                    {
                        return new CsvWorkbookExporter(stream, isAppending: false);
                    }
                case SpreadExportMode.Append:
                    {
                        if (stream.Length > 0)
                        {
                            return new CsvWorkbookExporter(stream, isAppending: true);
                        }
                        else
                        {
                            return new CsvWorkbookExporter(stream, isAppending: false);
                        }
                    }
                default:
                    throw new NotSupportedException();
            }
        }
    }
}