﻿using System;
using System.Linq;
using Telerik.Documents.SpreadsheetStreaming.Utilities;

namespace Telerik.Documents.SpreadsheetStreaming
{
    /// <summary>
    /// Describes a pattern type of fill.
    /// </summary>
    public class SpreadPatternFill : ISpreadFill
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="SpreadPatternFill" /> class.
        /// </summary>
        /// <param name="patternType">Type of the pattern.</param>
        public SpreadPatternFill(SpreadPatternType patternType) 
            : this(patternType, null, null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SpreadPatternFill" /> class.
        /// </summary>
        /// <param name="patternType">Type of the pattern.</param>
        /// <param name="patternColor">Color of the pattern.</param>
        /// <param name="backgroundColor">Color of the background.</param>
        public SpreadPatternFill(SpreadPatternType patternType, SpreadThemableColor patternColor, SpreadThemableColor backgroundColor)
        {
            this.PatternType = patternType;
            this.PatternColor = patternColor;
            this.BackgroundColor = backgroundColor;
        }

        /// <summary>
        /// Gets the type of the pattern.
        /// </summary>
        /// <value>The type of the pattern.</value>
        public SpreadPatternType PatternType { get; private set; }

        /// <summary>
        /// Gets the color of the pattern.
        /// </summary>
        /// <value>The color of the pattern.</value>
        public SpreadThemableColor PatternColor { get; private set; }

        /// <summary>
        /// Gets the color of the background.
        /// </summary>
        /// <value>The color of the background.</value>
        public SpreadThemableColor BackgroundColor { get; private set; }

        /// <summary>
        /// Creates a solid pattern fill.
        /// </summary>
        /// <param name="color">The color of the pattern.</param>
        /// <returns>Solid pattern fill.</returns>
        public static SpreadPatternFill CreateSolidFill(SpreadColor color)
        {
            Guard.ThrowExceptionIfNull(color, "color");

            return SpreadPatternFill.CreateSolidFill(new SpreadThemableColor(color));
        }

        /// <summary>
        /// Creates a solid pattern fill.
        /// </summary>
        /// <param name="color">The color of the pattern.</param>
        /// <returns>Solid pattern fill.</returns>
        public static SpreadPatternFill CreateSolidFill(SpreadThemableColor color)
        {
            Guard.ThrowExceptionIfNull(color, "color");

            return new SpreadPatternFill(SpreadPatternType.Solid, new SpreadThemableColor(new SpreadColor(0, 0, 0)), color);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// Returns true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            SpreadPatternFill other = obj as SpreadPatternFill;

            if (other == null)
            {
                return false;
            }

            return ObjectExtensions.EqualsOfT(this.PatternType, other.PatternType) &&
                ObjectExtensions.EqualsOfT(this.PatternColor, other.PatternColor) &&
                ObjectExtensions.EqualsOfT(this.BackgroundColor, other.BackgroundColor);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return ObjectExtensions.CombineHashCodes(
                this.PatternType.GetHashCodeOrZero(),
                this.PatternColor.GetHashCodeOrZero(),
                this.BackgroundColor.GetHashCodeOrZero());
        }
    }
}