﻿using System;
using System.Linq;
using Telerik.Documents.SpreadsheetStreaming.ImportExport;
using Telerik.Documents.SpreadsheetStreaming.Model.Themes;
using Telerik.Documents.SpreadsheetStreaming.Utilities;

namespace Telerik.Documents.SpreadsheetStreaming
{
    /// <summary>
    /// Represents a color which can be used in a theme.
    /// </summary>
    public class SpreadThemableColor
    {
        private readonly SpreadColor localValue;
        private readonly SpreadThemeColorType? themeColorType;
        private readonly SpreadColorShadeType? colorShadeType;
        private readonly double? tintAndShade;
        private readonly bool isAutomatic;

        /// <summary>
        /// Initializes a new instance of the <see cref="SpreadThemableColor" /> class.
        /// </summary>
        /// <param name="color">The color.</param>
        public SpreadThemableColor(SpreadColor color) : this(color, false)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SpreadThemableColor" /> class.
        /// </summary>
        /// <param name="color">The color.</param>
        /// <param name="isAutomatic">The is automatic.</param>
        public SpreadThemableColor(SpreadColor color, bool isAutomatic)
        {
            this.localValue = color;
            this.isAutomatic = isAutomatic;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SpreadThemableColor" /> class.
        /// </summary>
        /// <param name="themeColorType">Type of the theme color.</param>
        /// <param name="tintAndShade">The tint and shade.</param>
        public SpreadThemableColor(SpreadThemeColorType themeColorType, double tintAndShade) : this(themeColorType)
        {
            Guard.ThrowExceptionIfOutOfRange(-1, 1, tintAndShade, "tintAndShade");

            this.tintAndShade = tintAndShade;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SpreadThemableColor" /> class.
        /// </summary>
        /// <param name="themeColorType">Type of the theme color.</param>
        public SpreadThemableColor(SpreadThemeColorType themeColorType) : this(themeColorType, null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SpreadThemableColor" /> class.
        /// </summary>
        /// <param name="themeColorType">Type of the theme color.</param>
        /// <param name="colorShadeType">Type of the predefined tint and shade proportion for the themable color.</param>
        public SpreadThemableColor(SpreadThemeColorType themeColorType, SpreadColorShadeType? colorShadeType)
        {
            this.themeColorType = themeColorType;
            this.colorShadeType = colorShadeType;
        }

        /// <summary>
        /// Gets the local value of the color.
        /// </summary>
        /// <value>The local value of the color.</value>
        public SpreadColor LocalValue
        {
            get
            {
                return this.localValue;
            }
        }

        /// <summary>
        /// Gets the value indicating if the color is automatic. Automatic colors may be interpreted by a consumer as appropriate.
        /// </summary>
        /// <value>Value indicating if the color is automatic.</value>
        public bool IsAutomatic
        {
            get
            {
                return this.isAutomatic;
            }
        }

        /// <summary>
        /// Gets the type of the theme color.
        /// </summary>
        /// <value>The type of the theme color.</value>
        public SpreadThemeColorType ThemeColorType
        {
            get
            {
                return this.themeColorType.Value;
            }
        }

        /// <summary>
        /// Gets the type of the color shade.
        /// </summary>
        /// <value>The type of the color shade.</value>
        public SpreadColorShadeType? ColorShadeType
        {
            get
            {
                return this.colorShadeType;
            }
        }

        /// <summary>
        /// Gets the tint and shade proportion value. This value is applied over the theme color coming from the <see cref="SpreadThemeColorType" /> to determine the final color applied over the content.
        /// </summary>
        /// <value>The tint and shade value.</value>
        public double? TintAndShade
        {
            get
            {
                return this.tintAndShade;
            }
        }

        /// <summary>
        /// Gets a value indicating if the color comes from a theme.
        /// </summary>
        /// <value>The value indicating if the color comes from a theme.</value>
        public bool IsFromTheme
        {
            get
            {
                return this.themeColorType.HasValue;
            }
        }

        /// <summary>
        /// Creates a new instance of the <see cref="SpreadThemableColor" /> class.
        /// </summary>
        /// <param name="red">The red component.</param>
        /// <param name="green">The green component.</param>
        /// <param name="blue">The blue component.</param>
        /// <returns>Instance of the <see cref="SpreadThemableColor" /> class.</returns>
        public static SpreadThemableColor FromRgb(byte red, byte green, byte blue)
        {
            return new SpreadThemableColor(new SpreadColor(red, green, blue));
        }

        /// <summary>
        /// Determines whether the specified themable colors are equal.
        /// </summary>
        /// <returns>True if the themable colors are equal; otherwise, false.</returns>
        public static bool operator ==(SpreadThemableColor first, SpreadThemableColor second)
        {
            if ((object)first == null)
            {
                return (object)second == null;
            }

            return first.Equals(second);
        }

        /// <summary>
        /// Determines whether the specified themable colors are different.
        /// </summary>
        /// <returns>True if the themable colors are different; otherwise, false.</returns>
        public static bool operator !=(SpreadThemableColor first, SpreadThemableColor second)
        {
            return !(first == second);
        }

        /// <summary>
        /// Explicitly cast color to themable color.
        /// </summary>
        /// <param name="value">The color.</param>
        /// <returns>Themable color.</returns>
        public static explicit operator SpreadThemableColor(SpreadColor value)
        {
            return new SpreadThemableColor(value);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// Returns true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            SpreadThemableColor other = obj as SpreadThemableColor;

            if (other == null)
            {
                return false;
            }

            return ObjectExtensions.EqualsOfT(this.localValue, other.localValue) &&
                ObjectExtensions.EqualsOfT(this.isAutomatic, other.isAutomatic) &&
                ObjectExtensions.EqualsOfT(this.themeColorType, other.themeColorType) &&
                this.tintAndShade.EqualsDouble(other.tintAndShade) &&
                ObjectExtensions.EqualsOfT(this.colorShadeType, other.colorShadeType);
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return ObjectExtensions.CombineHashCodes(
                this.localValue.GetHashCodeOrZero(),
                this.themeColorType.GetHashCodeOrZero(),
                this.tintAndShade.GetHashCodeOrZero(),
                this.colorShadeType.GetHashCodeOrZero());
        }

        /// <summary>
        /// Returns a string that represents the themable color.
        /// </summary>
        /// <returns>A string that represents the themable color.</returns>
        public override string ToString()
        {
            if (this.IsFromTheme)
            {
                return this.ThemeColorType.ToString();
            }
            else
            {
                if (this.isAutomatic)
                {
                    return string.Format("Auto({0})", this.localValue.ToString());
                }
                else
                {
                    return this.localValue.ToString();
                }
            }
        }

        internal SpreadColor GetActualValue(SpreadTheme theme)
        {
            return this.GetActualValue(theme.ColorScheme);
        }

        internal SpreadColor GetActualValue(SpreadThemeColorScheme colorScheme)
        {
            double actualTintAndShade;
            SpreadColor color;
            if (this.IsFromTheme)
            {
                color = colorScheme[this.ThemeColorType].Color;

                if (this.tintAndShade.HasValue)
                {
                    actualTintAndShade = this.tintAndShade.Value;
                }
                else if (this.colorShadeType.HasValue)
                {
                    actualTintAndShade = colorScheme.GetTintAndShade(this.themeColorType.Value, this.colorShadeType.Value);
                }
                else
                {
                    actualTintAndShade = 0;
                }
            }
            else
            {
                color = this.localValue;
                actualTintAndShade = 0;
            }

            return ColorsHelper.UpdateTint(color, actualTintAndShade);
        }
    }
}