﻿using System;
using System.Linq;
using Telerik.Documents.SpreadsheetStreaming.Model.Themes;
using Telerik.Documents.SpreadsheetStreaming.Utilities;

namespace Telerik.Documents.SpreadsheetStreaming
{
    /// <summary>
    /// Represents a font family which can be used in a theme.
    /// </summary>
    public class SpreadThemableFontFamily
    {
        private readonly string localValue;
        private readonly SpreadThemeFontType? themeFontType;

        /// <summary>
        /// Initializes a new instance of the <see cref="SpreadThemableFontFamily" /> class.
        /// </summary>
        /// <param name="fontFamily">The font family name.</param>
        public SpreadThemableFontFamily(string fontFamily)
        {
            if (string.IsNullOrEmpty(fontFamily))
            {
                this.themeFontType = SpreadThemeFontType.Minor;
            }
            else
            {
                this.localValue = fontFamily;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SpreadThemableFontFamily" /> class.
        /// </summary>
        /// <param name="themeFontType">Type of the theme font.</param>
        public SpreadThemableFontFamily(SpreadThemeFontType themeFontType)
        {
            this.themeFontType = themeFontType;
        }

        /// <summary>
        /// Gets the local value of the font family.
        /// </summary>
        /// <value>The local value of the font family.</value>
        public string LocalValue
        {
            get
            {
                return this.localValue;
            }
        }

        /// <summary>
        /// Gets the type of the theme font.
        /// </summary>
        /// <value>The type of the theme font.</value>
        public SpreadThemeFontType ThemeFontType
        {
            get
            {
                return this.themeFontType.Value;
            }
        }

        /// <summary>
        /// Gets a value indicating if the font family comes from a theme.
        /// </summary>
        /// <value>The value indicating if the font family comes from a theme.</value>
        public bool IsFromTheme
        {
            get
            {
                return this.localValue == null;
            }
        }

        /// <summary>
        /// Determines whether the specified document themable font families are equal.
        /// </summary>
        /// <returns>True if the document themable font families are equal; otherwise, false.</returns>
        public static bool operator ==(SpreadThemableFontFamily first, SpreadThemableFontFamily second)
        {
            if ((object)first == null)
            {
                return (object)second == null;
            }

            return first.Equals(second);
        }

        /// <summary>
        /// Determines whether the specified document themable font families are different.
        /// </summary>
        /// <returns>True if the document themable font families are different; otherwise, false.</returns>
        public static bool operator !=(SpreadThemableFontFamily first, SpreadThemableFontFamily second)
        {
            return !(first == second);
        }

        /// <summary>
        /// Returns a string that represents the document themable font family.
        /// </summary>
        /// <returns>A string that represents the document themable font family.</returns>
        public override string ToString()
        {
            if (this.IsFromTheme)
            {
                return this.ThemeFontType.ToString();
            }
            else
            {
                return this.localValue.ToString();
            }
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object" /> is equal
        /// to the current <see cref="T:System.Object" />.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object" /> to compare with the current
        /// <see cref="T:System.Object" />.</param>
        /// <returns>
        /// Returns true if the specified <see cref="T:System.Object" /> is equal to the
        /// current <see cref="T:System.Object" />; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            SpreadThemableFontFamily other = obj as SpreadThemableFontFamily;

            if (other == null)
            {
                return false;
            }

            return ObjectExtensions.EqualsOfT(this.localValue, other.localValue) &&
                this.themeFontType == other.themeFontType;
        }

        /// <summary>
        /// Serves as a hash function for a particular type.
        /// </summary>
        /// <returns>A hash code for the current <see cref="T:System.Object" />.</returns>
        public override int GetHashCode()
        {
            return ObjectExtensions.CombineHashCodes(
                this.localValue.GetHashCodeOrZero(),
                this.themeFontType.GetHashCodeOrZero());
        }

        internal string GetActualValue(SpreadTheme theme)
        {
            return this.GetActualValue(theme.FontScheme);
        }

        private string GetActualValue(SpreadThemeFontScheme themeFontScheme, SpreadFontLanguageType fontLanguageType = SpreadFontLanguageType.Latin)
        {
            string fontFamily = this.IsFromTheme ? themeFontScheme[this.ThemeFontType][fontLanguageType].FontFamily : this.localValue;

            return fontFamily;
        }
    }
}
