﻿using System;
using System.Linq;
using System.Text;
using Telerik.Documents.SpreadsheetStreaming.Model;

namespace Telerik.Documents.SpreadsheetStreaming.Utilities
{
    internal static class NameConverter
    {
        private static readonly int LatinAlphabetLettersCount = 'Z' - 'A' + 1;

        /// <summary>
        /// Converts the row index to name.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <returns>The name.</returns>
        public static string ConvertRowIndexToName(int rowIndex)
        {
            rowIndex++;
            return rowIndex.ToString();
        }

        /// <summary>
        /// Converts the row name to index.
        /// </summary>
        /// <param name="rowName">Name of the row.</param>
        /// <returns>The index.</returns>
        public static int ConvertRowNameToIndex(string rowName)
        {
            int rowIndex;
            if (!int.TryParse(rowName, out rowIndex))
            {
                throw new InvalidOperationException(string.Format("'{0}' is invalid row name.", rowName));
            }

            rowIndex--;

            return rowIndex;
        }

        /// <summary>
        /// Converts the column index to name.
        /// </summary>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The name.</returns>
        public static string ConvertColumnIndexToName(int columnIndex)
        {
            string result = string.Empty;
            while (columnIndex >= 26)
            {
                result = (char)((columnIndex % LatinAlphabetLettersCount) + 'A') + result;
                columnIndex = (columnIndex / LatinAlphabetLettersCount) - 1;
            }

            result = (char)((columnIndex % LatinAlphabetLettersCount) + 'A') + result;

            return result;
        }

        /// <summary>
        /// Converts the cell index to name.
        /// </summary>
        /// <param name="rowIndex">Index of the row.</param>
        /// <param name="columnIndex">Index of the column.</param>
        /// <returns>The name.</returns>
        public static string ConvertCellIndexToName(int rowIndex, int columnIndex)
        {
            string rowName = ConvertRowIndexToName(rowIndex);
            string columnName = ConvertColumnIndexToName(columnIndex);

            return columnName + rowName;
        }

        public static void ConvertCellNameToIndex(string cellName, out int rowIndex, out int columnIndex)
        {
            int firstDigitIndex = GetFirstDigitIndex(cellName);

            rowIndex = ConvertRowNameToIndex(cellName.Substring(firstDigitIndex));
            columnIndex = ConvertColumnNameToIndex(cellName.Substring(0, firstDigitIndex));
        }

        public static void ConvertCellNameToIndex(string cellName, out bool isRowAbsolute, out int rowIndex, out bool isColumnAbsolute, out int columnIndex)
        {
            Guard.ThrowExceptionIfNullOrEmpty(cellName, "cellName");
            string sanitizedCellName = cellName.Replace("$", string.Empty);
            ConvertCellNameToIndex(sanitizedCellName, out rowIndex, out columnIndex);

            int dollarCount = cellName.Length - sanitizedCellName.Length;

            isColumnAbsolute = cellName.StartsWith("$");
            if (isColumnAbsolute)
            {
                dollarCount--;
            }

            isRowAbsolute = dollarCount > 0;
        }

        public static int ConvertColumnNameToIndex(string columnName)
        {
            int result;
            if (!TryConvertColumnNameToIndexInternal(columnName, out result))
            {
                throw new InvalidOperationException(string.Format("'{0}' is invalid column name.", columnName));
            }

            return result;
        }

        /// <summary>
        /// Converts the cell range to a name.
        /// </summary>
        /// <param name="fromRowIndex">Index of from row.</param>
        /// <param name="fromColumnIndex">Index of from column.</param>
        /// <param name="toRowIndex">Index of to row.</param>
        /// <param name="toColumnIndex">Index of to column.</param>
        /// <returns>The name.</returns>
        public static string ConvertCellRangeToName(int fromRowIndex, int fromColumnIndex, int toRowIndex, int toColumnIndex)
        {
            return ConvertCellIndexesToName(fromRowIndex, fromColumnIndex, toRowIndex, toColumnIndex);
        }

        /// <summary>
        /// Converts the cell indexes to a name.
        /// </summary>
        /// <param name="fromRowIndex">The start row index.</param>
        /// <param name="fromColumnIndex">The start column index.</param>
        /// <param name="toRowIndex">The end row index.</param>
        /// <param name="toColumnIndex">The end column index.</param>
        /// <returns>The name.</returns>
        public static string ConvertCellIndexesToName(int fromRowIndex, int fromColumnIndex, int toRowIndex, int toColumnIndex)
        {
            return string.Format("{0}:{1}", ConvertCellIndexToName(fromRowIndex, fromColumnIndex), ConvertCellIndexToName(toRowIndex, toColumnIndex));
        }

        internal static void ConvertCellNameToCellRange(string cellName, out SpreadCellRange cellRange)
        {
            Guard.ThrowExceptionIfNullOrEmpty(cellName, "cellName");

            string[] cellNames = cellName.Split(':');
            int row;
            int column;

            if (cellNames.Length == 1)
            {
                ConvertCellNameToIndex(cellName, out row, out column);
                cellRange = new SpreadCellRange(row, column, row, column);
            }
            else
            {
                int row1 = 0;
                int column1 = 0;
                bool isRowAbsolute;
                bool isColumnAbsolute;
                bool isRowAbsolute1;
                bool isColumnAbsolute1;

                ConvertCellNameToIndex(cellNames[0], out isRowAbsolute, out row, out isColumnAbsolute, out column);
                ConvertCellNameToIndex(cellNames[1], out isRowAbsolute1, out row1, out isColumnAbsolute1, out column1);

                cellRange = new SpreadCellRange(row, column, row1, column1);
            }
        }

        private static bool TryConvertColumnNameToIndexInternal(string columnName, out int index)
        {
            index = -1;

            if (!IsValidColumnName(columnName))
            {
                return false;
            }

            StringBuilder stringBuilder = new StringBuilder();
            for (int i = 0; i < columnName.Length; i++)
            {
                stringBuilder.Append(char.ToUpperInvariant(columnName[i]));
            }

            string columnNameToUpper = stringBuilder.ToString();

            int multiplier = LatinAlphabetLettersCount;
            index = columnNameToUpper[columnNameToUpper.Length - 1] - 'A';
            for (int i = columnNameToUpper.Length - 2; i >= 0; i--)
            {
                index += multiplier * (columnNameToUpper[i] - 'A' + 1);
                multiplier *= LatinAlphabetLettersCount;
            }

            return index <= DefaultValues.ColumnCount - 1;
        }

        private static bool IsValidColumnName(string columnName)
        {
            if (string.IsNullOrEmpty(columnName))
            {
                return false;
            }

            for (int i = 0; i < columnName.Length; i++)
            {
                if (!char.IsLetter(columnName[i]))
                {
                    return false;
                }
            }

            return true;
        }

        private static int GetFirstDigitIndex(string cellName)
        {
            for (int i = 0; i < cellName.Length; i++)
            {
                if (char.IsDigit(cellName[i]))
                {
                    return i;
                }
            }

            return 0;
        }
    }
}