﻿using System;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Base class for all block transformations.
    /// </summary>
    public abstract class BlockTransformBase : IBlockTransform
    {
        private TransformationHeader header = new TransformationHeader();

        /// <summary>Gets a value indicating whether the current transform can be reused.</summary>
        /// <returns>true if the current transform can be reused; otherwise, false.</returns>
        public abstract bool CanReuseTransform
        {
            get;
        }

        /// <summary>Gets a value indicating whether multiple blocks can be transformed.</summary>
        /// <returns>true if multiple blocks can be transformed; otherwise, false.</returns>
        public abstract bool CanTransformMultipleBlocks
        {
            get;
        }

        /// <summary>Gets transformation header (if required).</summary>
        public TransformationHeader Header
        {
            get
            {
                return this.header;
            }
        }

        /// <summary>Gets the input block size.</summary>
        /// <returns>The size of the input data blocks in bytes.</returns>
        public abstract int InputBlockSize
        {
            get;
        }

        /// <summary>Gets the output block size.</summary>
        /// <returns>The size of the output data blocks in bytes.</returns>
        public abstract int OutputBlockSize
        {
            get;
        }

        /// <summary>
        /// Gets or sets value which indicates whether the transformation uses
        /// input buffer of the fixed size.
        /// </summary>
        protected bool FixedInputBlockSize
        {
            get;
            set;
        }

        /// <summary>
        /// Releases the resources used by the current instance of the ZipArchive class.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Creates transformation header to be written into the output stream.
        /// </summary>
        public virtual void CreateHeader()
        {
        }

        /// <summary>
        /// Initialize reading of the transformation header.
        /// </summary>
        public virtual void InitHeaderReading()
        {
        }

        /// <summary>
        /// Process transformation header has been read.
        /// </summary>
        public virtual void ProcessHeader()
        {
            this.Header.BytesToRead = 0;
        }

        /// <summary>Transforms the specified region of the input byte array and copies the resulting transform to the specified region of the output byte array.</summary>
        /// <returns>The number of bytes written.</returns>
        /// <param name="inputBuffer">The input for which to compute the transform. </param>
        /// <param name="inputOffset">The offset into the input byte array from which to begin using data. </param>
        /// <param name="inputCount">The number of bytes in the input byte array to use as data. </param>
        /// <param name="outputBuffer">The output to which to write the transform. </param>
        /// <param name="outputOffset">The offset into the output byte array from which to begin writing data. </param>
        public abstract int TransformBlock(byte[] inputBuffer, int inputOffset, int inputCount, byte[] outputBuffer, int outputOffset);

        /// <summary>Transforms the specified region of the specified byte array.</summary>
        /// <returns>The computed transform.</returns>
        /// <param name="inputBuffer">The input for which to compute the transform. </param>
        /// <param name="inputOffset">The offset into the byte array from which to begin using data. </param>
        /// <param name="inputCount">The number of bytes in the byte array to use as data. </param>
        public abstract byte[] TransformFinalBlock(byte[] inputBuffer, int inputOffset, int inputCount);

        /// <summary>
        /// Releases the unmanaged resources used by the transform and optionally releases the managed resources. 
        /// </summary>
        /// <param name="disposing">Value which indicates whether both managed and unmanaged resources (true) on only unmanaged resources (false) should be released.</param>
        protected abstract void Dispose(bool disposing);

        /// <summary>
        /// Validates parameters of the input buffer.
        /// </summary>
        /// <param name="inputBuffer">The input for which to compute the transform. </param>
        /// <param name="inputOffset">The offset into the input byte array from which to begin using data. </param>
        /// <param name="inputCount">The number of bytes in the input byte array to use as data. </param>
        /// <param name="validateBlockSize">Indicates whether buffer block size should be validated. Should be true for the TransformBlock and false for the TransformFinalBlock.</param>
        /// <param name="allowZeroCount">Indicates whether count can be zero.</param>
        protected void ValidateInputBufferParameters(
            byte[] inputBuffer, 
            int inputOffset, 
            int inputCount, 
            bool validateBlockSize, 
            bool allowZeroCount)
        {
            OperationStream.ValidateBufferParameters(inputBuffer, inputOffset, inputCount, allowZeroCount);

            if (this.FixedInputBlockSize && validateBlockSize && inputCount % this.InputBlockSize != 0)
            {
                throw new ArgumentException("Invalid value.", "inputCount");
            }
        }

        /// <summary>
        /// Validates parameters of the transform operation.
        /// </summary>
        /// <param name="inputBuffer">The input for which to compute the transform. </param>
        /// <param name="inputOffset">The offset into the input byte array from which to begin using data. </param>
        /// <param name="inputCount">The number of bytes in the input byte array to use as data. </param>
        /// <param name="outputBuffer">The output to which to write the transform. </param>
        /// <param name="outputOffset">The offset into the output byte array from which to begin writing data. </param>
        /// <param name="allowZeroCount">Indicates whether input count can be zero.</param>
        protected void ValidateParameters(
            byte[] inputBuffer, 
            int inputOffset, 
            int inputCount, 
            byte[] outputBuffer,
            int outputOffset,
            bool allowZeroCount)
        {
            this.ValidateInputBufferParameters(inputBuffer, inputOffset, inputCount, true, allowZeroCount);
            OperationStream.ValidateBufferParameters(outputBuffer, outputOffset, 0, true);
        }
    }
}
