﻿#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Implements Adler-32 checksum algorithm.
    /// </summary>
    internal class Adler32 : IChecksumAlgorithm
    {
        /// <summary>
        /// Base for modulo arithmetic (largest prime smaller than 65536).
        /// </summary>
        private const uint Base = 65521;

        /// <summary>
        /// Number of iterations we can safely do before applying the modulo.
        /// </summary>
        private const int MaxIterations = 5552;
    
        /// <summary>
        /// Calculate checksum for the specified region of the input byte array.
        /// </summary>
        /// <param name="checksum">Checksum to update.</param>
        /// <param name="buffer">The input for which to compute the checksum.</param>
        /// <param name="offset">The offset into the input byte array from which to begin using data.</param>
        /// <param name="length">The number of bytes in the input byte array to use as data.</param>
        /// <returns>Updated checksum.</returns>
        public uint UpdateChecksum(uint checksum, byte[] buffer, int offset, int length)
        {
            if (checksum == 1 || buffer == null)
            {
                checksum = 1;
            }

            if (buffer != null)
            {
                OperationStream.ValidateBufferParameters(buffer, offset, length, true);

                uint lowerPart = checksum & 0xffff;
                uint upperPart = (checksum >> 16) & 0xffff;
                while (length > 0)
                {
                    int iterationCount = (length < MaxIterations) ? length : MaxIterations;
                    length -= iterationCount;
                    for (int iteration = 0; iteration < iterationCount; iteration++)
                    {
                        lowerPart += (uint)buffer[offset++] & 255;
                        upperPart += lowerPart;
                    }

                    lowerPart = lowerPart % Base;
                    upperPart = upperPart % Base;
                }

                checksum = (upperPart << 16) | lowerPart;
            }

            return checksum;
        }
    }
}
