﻿using System;
using System.IO;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Represents stream which allows read/write compressed information from/to given input stream.
    /// </summary>
    public class CompressedStream : OperationStream
    {
        private ICompressionAlgorithm algorithm;
        private long compressedSize = 0L;

        /// <summary>
        /// Initializes a new instance of the CompressedStream class.
        /// </summary>
        /// <param name="baseStream">The base input/output stream.</param>
        /// <param name="mode">Stream operational mode.</param>
        /// <param name="settings">Compression settings.</param>
        /// <exception cref="System.ArgumentOutOfRangeException">Specified mode is not allowed for the given stream.</exception>
        public CompressedStream(Stream baseStream, StreamOperationMode mode, CompressionSettings settings)
            : this(baseStream, mode, settings, true, null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the CompressedStream class.
        /// </summary>
        /// <param name="baseStream">The base input/output stream.</param>
        /// <param name="mode">Stream operational mode.</param>
        /// <param name="settings">Compression settings.</param>
        /// <param name="useCrc32">Indicates whether the CRC32 (true) or Adler32 (false) checksum algorithm will be used.</param>
        /// <param name="encryptionSettings">Encryption settings.</param>
        /// <exception cref="System.ArgumentOutOfRangeException">Specified mode is not allowed for the given stream.</exception>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "Crc")]
        public CompressedStream(
            Stream baseStream, 
            StreamOperationMode mode,
            CompressionSettings settings, 
            bool useCrc32, 
            EncryptionSettings encryptionSettings)
            : base(baseStream, mode)
        {
            baseStream = encryptionSettings != null ? 
                new CryptoStream(baseStream, mode, PlatformSettings.Manager.GetCryptoProvider(encryptionSettings)) : 
                baseStream;
            ICompressionAlgorithm compressionAlgorithm = ZipHelper.GetCompressionAlgorithm(settings);
            IChecksumAlgorithm checksumAlgorithm = useCrc32 ? 
                (IChecksumAlgorithm)(new Crc32()) : 
                (IChecksumAlgorithm)(new Adler32());
            this.Initialize(baseStream, compressionAlgorithm, checksumAlgorithm);
        }

        /// <summary>
        /// Initializes a new instance of the CompressedStream class.
        /// </summary>
        /// <param name="baseStream">The base input/output stream.</param>
        /// <param name="mode">Stream operational mode.</param>
        /// <param name="compressionAlgorithm">Compression algorithm.</param>
        /// <param name="checksumAlgorithm">Checksum algorithm.</param>
        /// <exception cref="System.ArgumentOutOfRangeException">Specified mode is not allowed for the given stream.</exception>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
        internal CompressedStream(Stream baseStream, StreamOperationMode mode, ICompressionAlgorithm compressionAlgorithm, IChecksumAlgorithm checksumAlgorithm)
            : base(baseStream, mode)
        {
            if (compressionAlgorithm == null)
            {
                throw new ArgumentNullException("compressionAlgorithm");
            }

            this.Initialize(baseStream, compressionAlgorithm, checksumAlgorithm);
        }

        /// <summary>
        /// Event occurs when calculation of the checksum for this stream is completed. 
        /// </summary>
        public event EventHandler ChecksumReady;

        /// <summary>
        /// Gets checksum calculated for this stream starting from 
        /// the first read/write operation and up to the Flush call.
        /// </summary>
        public long Checksum
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets the compressed size of the stream.
        /// </summary>
        public long CompressedSize
        {
            get
            {
                if (!this.IsDisposed)
                {
                    // 4.4.8, 4.4.9
                    // When a decryption header is present it 
                    // will be placed in front of the file data 
                    // and the value of the compressed file size 
                    // will include the bytes of the decryption header.
                    CryptoStream cryptoStream = this.BaseStream as CryptoStream;
                    if (cryptoStream != null)
                    {
                        return cryptoStream.TotalTransformedCount
                            + (cryptoStream.Transform != null && cryptoStream.Transform.Header.Buffer != null ?
                            (long)cryptoStream.Transform.Header.Buffer.Length :
                            0L);
                    }
                    else
                    {
                        return this.TotalTransformedCount
                            + (this.Transform != null 
                            && this.Transform.Header.Buffer != null
                            && this.Transform.Header.CountHeaderInCompressedSize ?
                            (long)this.Transform.Header.Buffer.Length :
                            0L);
                    }
                }
                else
                {
                    // Stream has been disposed.
                    // Return stored compressed size.
                    return this.compressedSize;
                }
            }
        }

        /// <summary>
        /// Gets or sets the checksum algorithm will be used during compression-decompression.
        /// </summary>
        internal IChecksumAlgorithm ChecksumAlgorithm
        {
            get;
            set;
        }
        
        /// <summary>
        /// Reads a sequence of bytes from the current stream and advances the position within the stream by the number of bytes read.
        /// </summary>
        /// <param name="buffer">An array of bytes. When this method returns, the buffer contains the specified byte array with the 
        /// values between offset and (offset + count - 1) replaced by the bytes read from the current source.</param>
        /// <param name="offset">The zero-based byte offset in buffer at which to begin storing the data read from the current stream.</param>
        /// <param name="count">The maximum number of bytes to be read from the current stream. </param>
        /// <returns>The total number of bytes read into the buffer. This can be less than the number of bytes requested if that many 
        /// bytes are not currently available, or zero (0) if the end of the stream has been reached.</returns>
        /// <exception cref="T:System.NotSupportedException">The <see cref="StreamOperationMode" /> associated with 
        /// current <see cref="OperationStream" /> object does not match the underlying stream.  
        /// For example, this exception is thrown when using <see cref="StreamOperationMode.Read" /> with an underlying stream that is write only.  </exception>
        /// <exception cref="T:System.ArgumentOutOfRangeException">The <paramref name="offset" /> parameter is less than zero.-or- The <paramref name="count" /> parameter is less than zero. </exception>
        /// <exception cref="T:System.ArgumentException">The sum of the <paramref name="count" /> and <paramref name="offset" /> parameters is larger than the length of the buffer. </exception>
        public override int Read(byte[] buffer, int offset, int count)
        {
            int bytesRead = base.Read(buffer, offset, count);

            if (bytesRead != 0)
            {
                if (this.ChecksumAlgorithm != null)
                {
                    this.Checksum = this.ChecksumAlgorithm.UpdateChecksum((uint)this.Checksum, buffer, offset, bytesRead);
                }
            }
            else
            {
                if (this.ChecksumReady != null)
                {
                    this.ChecksumReady(this, EventArgs.Empty);
                }
            }

            return bytesRead;
        }

        /// <summary>
        /// Sets the length of the current stream.
        /// </summary>
        /// <param name="value">The desired length of the current stream in bytes. </param>
        public override void SetLength(long value)
        {
            if (this.Mode == StreamOperationMode.Read)
            {
                CryptoStream cryptoStream = this.BaseStream as CryptoStream;
                if (cryptoStream != null)
                {
                    cryptoStream.SetLength(value);
                }
                base.SetLength(value);
            }
            else
            {
                throw new NotSupportedException();
            }
        }

        /// <summary>
        /// Writes a sequence of bytes to the current stream and advances the current position within this stream by the number of bytes written.
        /// </summary>
        /// <param name="buffer">An array of bytes. This method copies count bytes from buffer to the current stream. </param>
        /// <param name="offset">The zero-based byte offset in buffer at which to begin copying bytes to the current stream. </param>
        /// <param name="count">The number of bytes to be written to the current stream.</param>
        public override void Write(byte[] buffer, int offset, int count)
        {
            base.Write(buffer, offset, count);

            if (this.ChecksumAlgorithm != null)
            {
                this.Checksum = this.ChecksumAlgorithm.UpdateChecksum((uint)this.Checksum, buffer, offset, count);

                if (this.ChecksumReady != null)
                {
                    this.ChecksumReady(this, EventArgs.Empty);
                }
            }
        }

        /// <summary>
        /// Releases the unmanaged resources used by the Stream and optionally releases the managed resources. 
        /// </summary>
        /// <param name="disposing">Value which indicates whether both managed and unmanaged resources (true) on only unmanaged resources (false) should be released.</param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Usage", "CA2215:Dispose methods should call base class dispose")]
        protected override void Dispose(bool disposing)
        {
            if (!this.IsDisposed)
            {
                try
                {
                    if (!this.HasFlushedFinalBlock)
                    {
                        this.FlushFinalBlock();
                    }
                    this.compressedSize = this.CompressedSize;
                }
                finally
                {
                    this.algorithm = null;
                    base.Dispose(disposing);
                }
            }
        }

        /// <summary>
        /// Initialize compressed stream.
        /// </summary>
        /// <param name="baseStream">The base input/output stream.</param>
        /// <param name="compressionAlgorithm">Compression algorithm.</param>
        /// <param name="checksumAlgorithm">Checksum algorithm.</param>
        private void Initialize(Stream baseStream, ICompressionAlgorithm compressionAlgorithm, IChecksumAlgorithm checksumAlgorithm)
        {
            this.BaseStream = baseStream;
            this.algorithm = compressionAlgorithm;
            this.ChecksumAlgorithm = checksumAlgorithm;

            switch (this.Mode)
            {
                case StreamOperationMode.Read:
                    this.Transform = this.algorithm.CreateDecompressor();
                    break;

                case StreamOperationMode.Write:
                    this.Transform = this.algorithm.CreateCompressor();
                    break;
            }
        }
    }
}
