﻿using System;
using System.IO;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Represents base class for all compression and decompression functionality.
    /// </summary>
    internal abstract class CompressionTransformBase : BlockTransformBase
    {
        /// <summary>
        /// Initializes a new instance of the CompressionTransformBase class.
        /// </summary>
        public CompressionTransformBase()
        {
            this.FixedInputBlockSize = false;
        }

        /// <summary>Gets a value indicating whether the current transform can be reused.</summary>
        /// <returns>true if the current transform can be reused; otherwise, false.</returns>
        public override bool CanReuseTransform
        {
            get
            {
                return false;
            }
        }

        /// <summary>Gets a value indicating whether multiple blocks can be transformed.</summary>
        /// <returns>true if multiple blocks can be transformed; otherwise, false.</returns>
        public override bool CanTransformMultipleBlocks
        {
            get
            {
                return false;
            }
        }

        /// <summary>Gets the input block size.</summary>
        /// <returns>The size of the input data blocks in bytes.</returns>
        public override int InputBlockSize
        {
            get
            {
                return 8192;
            }
        }

        /// <summary>Gets the output block size.</summary>
        /// <returns>The size of the output data blocks in bytes.</returns>
        public override int OutputBlockSize
        {
            get
            {
                return 32768;
            }
        }

        protected int AvailableBytesIn
        {
            get;
            set;
        }

        protected byte[] InputBuffer
        {
            get;
            set;
        }

        protected int NextIn
        {
            get;
            set;
        }

        protected int TotalBytesIn
        {
            get;
            set;
        }

        protected int AvailableBytesOut
        {
            get;
            set;
        }

        protected byte[] OutputBuffer
        {
            get;
            set;
        }

        protected int NextOut
        {
            get;
            set;
        }

        protected int TotalBytesOut
        {
            get;
            set;
        }

        /// <summary>
        /// Transforms the specified region of the input byte array and copies
        /// the resulting transform to the specified region of the output byte array.
        /// </summary>
        /// <returns>The number of bytes written.</returns>
        /// <param name="inputBuffer">The input for which to compute the transform. </param>
        /// <param name="inputOffset">The offset into the input byte array from which to begin using data. </param>
        /// <param name="inputCount">The number of bytes in the input byte array to use as data. </param>
        /// <param name="outputBuffer">The output to which to write the transform. </param>
        /// <param name="outputOffset">The offset into the output byte array from which to begin writing data. </param>
        public override int TransformBlock(byte[] inputBuffer, int inputOffset, int inputCount, byte[] outputBuffer, int outputOffset)
        {
            this.ValidateParameters(inputBuffer, inputOffset, inputCount, outputBuffer, outputOffset, true);

            this.AvailableBytesIn = inputCount;
            this.InputBuffer = inputBuffer;
            this.AvailableBytesOut = outputBuffer.Length - outputOffset;
            this.OutputBuffer = outputBuffer;

            this.NextOut = outputOffset;
            this.NextIn = inputOffset;

            this.ProcessTransform(false);

            return this.NextOut - outputOffset;
        }

        /// <summary>Transforms the specified region of the specified byte array.</summary>
        /// <returns>The computed transform.</returns>
        /// <param name="inputBuffer">The input for which to compute the transform. </param>
        /// <param name="inputOffset">The offset into the byte array from which to begin using data. </param>
        /// <param name="inputCount">The number of bytes in the byte array to use as data. </param>
        public override byte[] TransformFinalBlock(byte[] inputBuffer, int inputOffset, int inputCount)
        {
            this.ValidateInputBufferParameters(inputBuffer, inputOffset, inputCount, false, true);

            this.AvailableBytesIn = inputCount;
            this.InputBuffer = inputBuffer;
            byte[] buffer;
            using (MemoryStream stream = new MemoryStream())
            {
                bool next;
                do
                {
                    byte[] outputBuffer = new byte[this.OutputBlockSize];

                    this.AvailableBytesOut = outputBuffer.Length;
                    this.OutputBuffer = outputBuffer;

                    this.NextOut = 0;
                    this.NextIn = 0;

                    next = this.ProcessTransform(true);
                    this.AvailableBytesIn = 0;

                    stream.Write(outputBuffer, 0, this.NextOut);
                }
                while (next);

                buffer = stream.ToArray();
            }

            return buffer;
        }

        /// <summary>
        /// Releases the unmanaged resources used by the transform and optionally releases the managed resources. 
        /// </summary>
        /// <param name="disposing">Value which indicates whether both managed and unmanaged resources (true) on only unmanaged resources (false) should be released.</param>
        protected override void Dispose(bool disposing)
        {
        }

        /// <summary>
        /// Transforms current input buffer.
        /// </summary>
        /// <param name="finalBlock">The final block flag.</param>
        /// <returns>True when output still available.</returns>
        protected abstract bool ProcessTransform(bool finalBlock);
    }
}
