﻿using System;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Compressor which implements Deflate compression.
    /// </summary>
    internal class DeflateCompressor : DeflateTransformBase
    {
        /// <summary>
        /// The default number of window bits for the Deflate algorithm.
        /// 15 is the maximum number of window bits for the Deflate algorithm (32K window).
        /// </summary>
        private const int WindowBitsDefault = 15;
        private const int DefaultMemoryLevel = 8;

        private const int StaticTreeBlock = 1;
        private const int DynamicTreeBlock = 2;

        private const int BufferSize = 8 * 2;
        private const int MinMatch = 3;
        private const int MaxMatch = 258;
        private const int MinLookAhead = MaxMatch + MinMatch + 1;
        private const int EndBlock = 256;

        private int goodLength;
        private int maxChainLength;
        private int maxLazy;
        private int niceLength;

        private CompressionMethod deflateFunction;

        // as the name implies
        private bool finishPending;

        // LZ77 window size (32K by default)
        private int windowSize;

        // log2(windowSize)  (8..16)
        private int windowBits;

        // windowSize - 1
        private int windowMask;

        // Sliding window. Input bytes are read into the second half of the window,
        // and move to the first half later to keep a dictionary of at least wSize
        // bytes. With this organization, matches are limited to a distance of
        // windowSize-MaxMatch bytes, but this ensures that IO is always
        // performed with a length multiple of the block size.
        private byte[] window;

        // Actual size of window: 2*windowSize, except when the user input buffer
        // is directly used as sliding window.
        private int actualWindowSize;

        // Link to older string with same hash index. To limit the size of this
        // array to 64K, this link is maintained only for the last 32K strings.
        // An index in this array is thus a window index modulo 32K.
        // Heads of the hash chains or Null.
        private short[] head;
        private short[] previous;

        // hash index of string to be inserted
        private int hashIndexOfString;

        // number of elements in hash table
        private int hashSize;

        // log2(hashSize)
        private int hashBits;

        // hashSize-1
        private int hashMask;

        // Number of bits by which hashIndexOfString must be shifted at each input
        // step. It must be such that after MinMatch steps, the oldest
        // byte no longer takes part in the hash key, that is:
        // hashShift * MinMatch >= hashBits
        private int hashShift;

        // Window position at the beginning of the current output block.
        // Gets negative when the window is moved backwards.
        private int blockStart;

        // length of best match
        private int matchLength;

        // previous match
        private int previousMatch;

        // set if previous match exists
        private int matchAvailable;

        // start of string to insert into
        private int startOfString;

        // start of matching string
        private int matchStart;

        // number of valid bytes ahead in window
        private int lookAhead;

        // Length of the best match at previous step. Matches not greater than this
        // are discarded. This is used in the lazy match evaluation.
        private int previousLength;

        // compression level (1..9)
        private int compressionLevel;

        // literal and length tree
        private short[] dynamicLiteralsTree;

        // distance tree
        private short[] dynamicDistanceTree;

        // Huffman tree for bit lengths
        private short[] dynamicBitLengthTree;

        // desc for literal tree
        private Tree treeLiterals = new Tree();

        // desc for distance tree
        private Tree treeDistances = new Tree();

        // desc for bit length tree
        private Tree treeBitLengths = new Tree();

        // index for literals or lengths
        private int lengthOffset;

        // Size of match buffer for literals/lengths
        private int literalsBufsize;

        // running index in literal buffer
        private int lastLiteral;

        // index into pending
        private int distanceOffset;

        // number of string matches in current block
        private int matches;

        // Output buffer.
        // Bits are inserted starting at the bottom
        // (least significant bits).
        private short bitsBuffer;

        // Number of valid bits in bits buffer.
        // All bits above the last valid bit are always zero.
        private int bitsValid;

        // output still pending - waiting to be compressed
        private byte[] pending;

        // index of next pending byte to output
        private int nextPending;

        // number of bytes in the pending buffer
        private int pendingCount;

        /// <summary>
        /// Initializes a new instance of the DeflateCompressor class.
        /// </summary>
        /// <param name="settings">Deflate settings.</param>
        public DeflateCompressor(DeflateSettings settings)
            : base(settings)
        {
            int heapSize = (2 * DeflateConstants.LiteralCodes) + 1;

            this.dynamicLiteralsTree = new short[heapSize * 2];

            // distance tree
            this.dynamicDistanceTree = new short[((2 * DeflateConstants.DynamicCodes) + 1) * 2];

            // Huffman tree for bit lengths
            this.dynamicBitLengthTree = new short[((2 * DeflateConstants.BitLengthCodes) + 1) * 2];

            this.BitLengthCount = new short[DeflateConstants.MaxBits + 1];

            this.Heap = new int[heapSize];
            this.Depth = new sbyte[heapSize];

            this.Initialize(settings);
        }

        private delegate DeflateBlockState CompressionMethod(bool flush);

        // number of codes at each bit length for an optimal tree
        public short[] BitLengthCount
        {
            get;
            private set;
        }

        // heap used to build the Huffman trees
        public int[] Heap
        {
            get;
            private set;
        }

        // number of elements in the heap
        public int HeapLength
        {
            get;
            set;
        }

        // element of largest frequency
        public int HeapMax
        {
            get;
            set;
        }

        // The sons of heap[n] are heap[2*n] and heap[2*n+1].
        // heap[0] is not used.
        // The same heap array is used to build all trees.
        // Depth of each subtree used as tie breaker for trees of equal frequency
        public sbyte[] Depth
        {
            get;
            private set;
        }

        // bit length of current block with optimal trees
        public int OptimalLength
        {
            get;
            set;
        }

        // bit length of current block with static trees
        public int StaticLength
        {
            get;
            set;
        }

        /// <summary>
        /// Creates RFC 1950 (ZLIB Compressed Data Format Specification version 3.3) header
        /// to be written into the output stream.
        /// </summary>
        public override void CreateHeader()
        {
            if (this.Settings.HeaderType == CompressedStreamHeader.ZLib)
            {
                byte[] byteHeader = new byte[2];

                int header = (DeflateConstants.HeaderDeflated + ((this.windowBits - 8) << 4)) << 8;
                int levelFlags = ((this.compressionLevel - 1) & 0xff) >> 1;

                if (levelFlags > 3)
                {
                    levelFlags = 3;
                }

                header |= levelFlags << 6;
                header += 31 - (header % 31);

                byteHeader[0] = (byte)(header >> 8);
                byteHeader[1] = (byte)header;

                this.Header.Buffer = byteHeader;
            }
        }

        /// <summary>
        /// Restore the heap property by moving down the tree starting at specified node,
        /// exchanging a node with the smallest of its two sons if necessary, stopping
        /// when the heap property is re-established (each father smaller than its two sons).
        /// </summary>
        /// <param name="tree">The tree.</param>
        /// <param name="nodeIndex">Index of node.</param>
        internal void DownHeap(short[] tree, int nodeIndex)
        {
            // Index of left son of node
            int index = nodeIndex << 1;
            int node = this.Heap[nodeIndex];

            while (index <= this.HeapLength)
            {
                // Set index to the smallest of the two sons:
                if (index < this.HeapLength && IsSmaller(tree, this.Heap[index + 1], this.Heap[index], this.Depth))
                {
                    index++;
                }

                // Exit if node is smaller than both sons
                if (IsSmaller(tree, node, this.Heap[index], this.Depth))
                {
                    break;
                }

                // Exchange node with the smallest son
                this.Heap[nodeIndex] = this.Heap[index];
                nodeIndex = index;

                // And continue down the tree, setting index to the left son of node
                index <<= 1;
            }

            this.Heap[nodeIndex] = node;
        }

        /// <summary>
        /// Transforms current input buffer.
        /// </summary>
        /// <param name="finalBlock">The final block flag.</param>
        /// <returns>True when still output available.</returns>
        protected override bool ProcessTransform(bool finalBlock)
        {
            this.Deflate(finalBlock);

            return this.pendingCount != 0;
        }

        private static bool IsSmaller(short[] tree, int nodeIndex1, int nodeIndex2, sbyte[] depth)
        {
            short node1 = tree[nodeIndex1 * 2];
            short node2 = tree[nodeIndex2 * 2];

            return node1 < node2 || (node1 == node2 && depth[nodeIndex1] <= depth[nodeIndex2]);
        }

        private void Deflate(bool flush)
        {
            if (this.AvailableBytesOut <= 1)
            {
                throw new InvalidOperationException("Output buffer is full");
            }

            // Flush as much pending output as possible
            if (this.pendingCount != 0)
            {
                this.FlushPending();
                if (this.AvailableBytesOut == 0)
                {
                    return;
                }
            }
            else if (this.AvailableBytesIn == 0 && !flush)
            {
                return;
            }

            // User must not provide more input after the first FINISH:
            if (this.finishPending && this.AvailableBytesIn != 0)
            {
                throw new InvalidOperationException("Status is finish, but input bytes count is not zero");
            }

            // Start a new block or continue the current one.
            if (this.AvailableBytesIn != 0 || this.lookAhead != 0 || (flush && !this.finishPending))
            {
                DeflateBlockState bstate = this.deflateFunction(flush);

                if (bstate == DeflateBlockState.FinishStarted || bstate == DeflateBlockState.FinishDone)
                {
                    this.finishPending = true;
                }

                if (bstate == DeflateBlockState.NeedMore || bstate == DeflateBlockState.FinishStarted)
                {
                    return;
                }

                if (bstate == DeflateBlockState.BlockDone)
                {
                    this.SendStoredBlock(0, 0, false);

                    this.FlushPending();
                    if (this.AvailableBytesOut == 0)
                    {
                        return;
                    }
                }
            }
        }

        /// <summary>
        /// Scan a literal or distance tree to determine the frequencies of the codes
        /// in the bit length tree.
        /// </summary>
        /// <param name="tree">The tree.</param>
        /// <param name="maxCode">Max code.</param>
        private void ScanTree(short[] tree, int maxCode)
        {
            // iterates over all tree elements
            int n;

            // last emitted length
            int lastLength = -1;

            // length of current code
            int currentLength;

            // length of next code
            int nextLength = (int)tree[1];

            // repeat count of the current code
            int count = 0;

            // max repeat count
            int maxCount = 7;

            // min repeat count
            int minCount = 4;

            if (nextLength == 0)
            {
                maxCount = 138;
                minCount = 3;
            }

            tree[((maxCode + 1) * 2) + 1] = (short)0x7fff;

            for (n = 0; n <= maxCode; n++)
            {
                currentLength = nextLength;
                nextLength = (int)tree[((n + 1) * 2) + 1];
                if (++count < maxCount && currentLength == nextLength)
                {
                    continue;
                }
                else if (count < minCount)
                {
                    this.dynamicBitLengthTree[currentLength * 2] = (short)(this.dynamicBitLengthTree[currentLength * 2] + count);
                }
                else if (currentLength != 0)
                {
                    if (currentLength != lastLength)
                    {
                        this.dynamicBitLengthTree[currentLength * 2]++;
                    }

                    this.dynamicBitLengthTree[DeflateConstants.Repeat3To6 * 2]++;
                }
                else if (count <= 10)
                {
                    this.dynamicBitLengthTree[DeflateConstants.RepeatZero3To10 * 2]++;
                }
                else
                {
                    this.dynamicBitLengthTree[DeflateConstants.RepeatZero11To138 * 2]++;
                }

                count = 0;
                lastLength = currentLength;
                if (nextLength == 0)
                {
                    maxCount = 138;
                    minCount = 3;
                }
                else if (currentLength == nextLength)
                {
                    maxCount = 6;
                    minCount = 3;
                }
                else
                {
                    maxCount = 7;
                    minCount = 4;
                }
            }
        }

        /// <summary>
        /// Construct the Huffman tree for the bit lengths.
        /// </summary>
        /// <returns>The index of the last bit length code to send.</returns>
        private int BuildBitLengthTree()
        {
            // index of last bit length code of non zero freq
            int maxIndex;

            // Determine the bit length frequencies for literal and distance trees
            this.ScanTree(this.dynamicLiteralsTree, this.treeLiterals.MaxCode);
            this.ScanTree(this.dynamicDistanceTree, this.treeDistances.MaxCode);

            // Build the bit length tree
            this.treeBitLengths.BuildTree(this);

            // this.OptimalLength now includes the length of the tree representations,
            // except the lengths of the bit lengths codes and the 5+5+4 bits for the counts.
            // Determine the number of bit length codes to send. The pkzip format
            // requires that at least 4 bit length codes be sent.
            // Specification says 3 but the actual value used is 4.
            for (maxIndex = DeflateConstants.BitLengthCodes - 1; maxIndex >= 3; maxIndex--)
            {
                if (this.dynamicBitLengthTree[(Tree.BitLengthOrder[maxIndex] * 2) + 1] != 0)
                {
                    break;
                }
            }

            // Update optimal length to include the bit length tree and counts
            this.OptimalLength += (3 * (maxIndex + 1)) + 5 + 5 + 4;

            return maxIndex;
        }

        /// <summary>
        /// Send the header for a block using dynamic Huffman trees: the counts,
        /// the lengths of the bit length codes, the literal tree and the distance tree.
        /// </summary>
        /// <param name="literalCodes">Length of literal codes.</param>
        /// <param name="distanceCodes">Length of distance codes.</param>
        /// <param name="bitLengthCodes">Length of bit length codes.</param>
        private void SendAllTrees(int literalCodes, int distanceCodes, int bitLengthCodes)
        {
            // index in bit length order
            int rank;

            this.SendBits(literalCodes - 257, 5);
            this.SendBits(distanceCodes - 1, 5);
            this.SendBits(bitLengthCodes - 4, 4);
            for (rank = 0; rank < bitLengthCodes; rank++)
            {
                this.SendBits(this.dynamicBitLengthTree[(Tree.BitLengthOrder[rank] * 2) + 1], 3);
            }

            // literal tree
            this.SendTree(this.dynamicLiteralsTree, literalCodes - 1);

            // distance tree
            this.SendTree(this.dynamicDistanceTree, distanceCodes - 1);
        }

        /// <summary>
        /// Send a literal or distance tree in compressed form,
        /// using the codes in bit length tree.
        /// </summary>
        /// <param name="tree">The tree.</param>
        /// <param name="maxCode">Max code.</param>
        private void SendTree(short[] tree, int maxCode)
        {
            // iterates over all tree elements
            int n;

            // last emitted length
            int lastLength = -1;

            // length of current code
            int currentLength;

            // length of next code
            int nextLength = tree[1];

            // repeat count of the current code
            int count = 0;

            // max repeat count
            int maxCount = 7;

            // min repeat count
            int minCount = 4;

            if (nextLength == 0)
            {
                maxCount = 138;
                minCount = 3;
            }

            for (n = 0; n <= maxCode; n++)
            {
                currentLength = nextLength;
                nextLength = tree[((n + 1) * 2) + 1];
                if (++count < maxCount && currentLength == nextLength)
                {
                    continue;
                }
                else if (count < minCount)
                {
                    do
                    {
                        this.SendCode(currentLength, this.dynamicBitLengthTree);
                    }
                    while (--count != 0);
                }
                else if (currentLength != 0)
                {
                    if (currentLength != lastLength)
                    {
                        this.SendCode(currentLength, this.dynamicBitLengthTree);
                        count--;
                    }

                    this.SendCode(DeflateConstants.Repeat3To6, this.dynamicBitLengthTree);
                    this.SendBits(count - 3, 2);
                }
                else if (count <= 10)
                {
                    this.SendCode(DeflateConstants.RepeatZero3To10, this.dynamicBitLengthTree);
                    this.SendBits(count - 3, 3);
                }
                else
                {
                    this.SendCode(DeflateConstants.RepeatZero11To138, this.dynamicBitLengthTree);
                    this.SendBits(count - 11, 7);
                }

                count = 0;
                lastLength = currentLength;
                if (nextLength == 0)
                {
                    maxCount = 138;
                    minCount = 3;
                }
                else if (currentLength == nextLength)
                {
                    maxCount = 6;
                    minCount = 3;
                }
                else
                {
                    maxCount = 7;
                    minCount = 4;
                }
            }
        }

        /// <summary>
        /// Output a block of bytes on the stream.
        /// </summary>
        /// <param name="buffer">Buffer.</param>
        /// <param name="start">Start index.</param>
        /// <param name="length">Length.</param>
        private void PutBytes(byte[] buffer, int start, int length)
        {
            Array.Copy(buffer, start, this.pending, this.pendingCount, length);
            this.pendingCount += length;
        }

        private void SendCode(int nodeIndex, short[] tree)
        {
            int index = nodeIndex * 2;
            this.SendBits(tree[index] & 0xffff, tree[index + 1] & 0xffff);
        }

        private void SendBits(int value, int length)
        {
            int len = length;
            unchecked
            {
                int bitwiseOperationResult = this.bitsBuffer | (short)((value << this.bitsValid) & 0xffff);
                this.bitsBuffer = (short)bitwiseOperationResult;

                if (this.bitsValid > (int)BufferSize - len)
                {
                    this.pending[this.pendingCount++] = (byte)this.bitsBuffer;
                    this.pending[this.pendingCount++] = (byte)(this.bitsBuffer >> 8);

                    this.bitsBuffer = (short)((uint)value >> (BufferSize - this.bitsValid));
                    this.bitsValid += len - BufferSize;
                }
                else
                {
                    this.bitsValid += len;
                }
            }
        }

        /// <summary>
        /// Save the match info and tally the frequency counts.
        /// </summary>
        /// <param name="distance">Distance.</param>
        /// <param name="lengthOrChar">Length or unmatched char.</param>
        /// <returns>Return true if the current block must be flushed.</returns>
        private bool TreeTally(int distance, int lengthOrChar)
        {
            this.pending[this.distanceOffset + (this.lastLiteral * 2)] = unchecked((byte)((uint)distance >> 8));
            this.pending[this.distanceOffset + (this.lastLiteral * 2) + 1] = unchecked((byte)distance);
            this.pending[this.lengthOffset + this.lastLiteral] = unchecked((byte)lengthOrChar);
            this.lastLiteral++;

            if (distance == 0)
            {
                // lengthOrChar is the unmatched char
                this.dynamicLiteralsTree[lengthOrChar * 2]++;
            }
            else
            {
                this.matches++;

                // Here, lengthOrChar is the match length - MIN_MATCH
                // distance = match distance - 1
                distance--;
                this.dynamicLiteralsTree[(Tree.LengthCode[lengthOrChar] + DeflateConstants.Literals + 1) * 2]++;
                this.dynamicDistanceTree[Tree.GetDistanceCode(distance) * 2]++;
            }

            if ((this.lastLiteral & 0x1fff) == 0 && this.compressionLevel > 2)
            {
                // Compute an upper bound for the compressed length
                int outputLength = this.lastLiteral << 3;
                int inputLength = this.startOfString - this.blockStart;
                int dcode;
                for (dcode = 0; dcode < DeflateConstants.DynamicCodes; dcode++)
                {
                    outputLength = (int)(outputLength
                        + ((int)this.dynamicDistanceTree[dcode * 2] * (5L + Tree.ExtraDistanceBits[dcode])));
                }

                outputLength >>= 3;
                if (this.matches < this.lastLiteral / 2 && outputLength < inputLength / 2)
                {
                    return true;
                }
            }

            return (this.lastLiteral == this.literalsBufsize - 1) || (this.lastLiteral == this.literalsBufsize);
        }

        /// <summary>
        /// Send the block data compressed using the given Huffman trees.
        /// </summary>
        /// <param name="literalTree">Literal tree.</param>
        /// <param name="distanceTree">Distance tree.</param>
        private void SendCompressedBlock(short[] literalTree, short[] distanceTree)
        {
            // distance of matched string
            int distance;

            // match length or unmatched char if distance is zero
            int lengthOrChar;

            // running index in buffer
            int index = 0;

            // the code to send
            int code;

            // number of extra bits to send
            int extra;

            if (this.lastLiteral != 0)
            {
                do
                {
                    int offset = this.distanceOffset + (index * 2);
                    distance = ((this.pending[offset] << 8) & 0xff00) |
                        (this.pending[offset + 1] & 0xff);

                    lengthOrChar = this.pending[this.lengthOffset + index] & 0xff;
                    index++;

                    if (distance == 0)
                    {
                        // send a literal byte
                        this.SendCode(lengthOrChar, literalTree);
                    }
                    else
                    {
                        // literal or match pair
                        // Here, lengthOrChar is the match length - MIN_MATCH
                        code = Tree.LengthCode[lengthOrChar];

                        // send the length code
                        this.SendCode(code + DeflateConstants.Literals + 1, literalTree);
                        extra = Tree.ExtraLengthBits[code];
                        if (extra != 0)
                        {
                            // send the extra length bits
                            lengthOrChar -= Tree.LengthBase[code];
                            this.SendBits(lengthOrChar, extra);
                        }

                        // distance is now the match distance - 1
                        distance--;
                        code = Tree.GetDistanceCode(distance);

                        // send the distance code
                        this.SendCode(code, distanceTree);

                        extra = Tree.ExtraDistanceBits[code];
                        if (extra != 0)
                        {
                            // send the extra distance bits
                            distance -= Tree.DistanceBase[code];
                            this.SendBits(distance, extra);
                        }
                    }
                }
                while (index < this.lastLiteral);
            }

            this.SendCode(EndBlock, literalTree);
        }

        /// <summary>
        /// Flush the bit buffer and align the output on a byte boundary.
        /// </summary>
        private void AlginOnByteBoundary()
        {
            if (this.bitsValid > 8)
            {
                this.pending[this.pendingCount++] = (byte)this.bitsBuffer;
                this.pending[this.pendingCount++] = (byte)(this.bitsBuffer >> 8);
            }
            else if (this.bitsValid > 0)
            {
                this.pending[this.pendingCount++] = (byte)this.bitsBuffer;
            }

            this.bitsBuffer = 0;
            this.bitsValid = 0;
        }

        /// <summary>
        /// Copy a stored block, storing first the length
        /// and its one's complement if requested.
        /// </summary>
        /// <param name="buffer">Buffer.</param>
        /// <param name="length">Length.</param>
        /// <param name="header">Should send the header.</param>
        private void CopyBlock(int buffer, int length, bool header)
        {
            // align on byte boundary
            this.AlginOnByteBoundary();

            if (header)
            {
                unchecked
                {
                    this.pending[this.pendingCount++] = (byte)length;
                    this.pending[this.pendingCount++] = (byte)(length >> 8);
                    this.pending[this.pendingCount++] = (byte)~length;
                    this.pending[this.pendingCount++] = (byte)(~length >> 8);
                }
            }

            this.PutBytes(this.window, buffer, length);
        }

        private void FlushBlockOnly(bool lastBlock)
        {
            this.TreeFlushBlock(this.blockStart >= 0 ? this.blockStart : -1, this.startOfString - this.blockStart, lastBlock);
            this.blockStart = this.startOfString;
            this.FlushPending();
        }

        /// <summary>
        /// Send a stored block.
        /// </summary>
        /// <param name="offset">Offset in window.</param>
        /// <param name="length">Length.</param>
        /// <param name="lastBlock">The flag of last block.</param>
        private void SendStoredBlock(int offset, int length, bool lastBlock)
        {
            // send a stored block type
            int last = lastBlock ? 1 : 0;
            this.SendBits(last, 3);

            // with header
            this.CopyBlock(offset, length, true);
        }

        /// <summary>
        /// Determine the best encoding for the current block: dynamic trees, static
        /// trees or store, and output the encoded block.
        /// </summary>
        /// <param name="offset">Offset in window.</param>
        /// <param name="length">Length.</param>
        /// <param name="lastBlock">The flag of last block.</param>
        private void TreeFlushBlock(int offset, int length, bool lastBlock)
        {
            // opt_len and static_len in bytes
            int optimalBytes;
            int staticBytes;

            // index of last bit length code of non zero freq
            int maxBitLengthIndex = 0;

            // Build the Huffman trees unless a stored block is forced
            if (this.compressionLevel > 0)
            {
                // Construct the literal and distance trees
                this.treeLiterals.BuildTree(this);

                this.treeDistances.BuildTree(this);

                // At this point, optimal length and static length are the total bit lengths
                // of the compressed block data, excluding the tree representations.
                // Build the bit length tree for the above two trees, and get the index
                // in bit length order of the last bit length code to send.
                maxBitLengthIndex = this.BuildBitLengthTree();

                // Determine the best encoding.
                // Compute first the block length in bytes.
                optimalBytes = (this.OptimalLength + 3 + 7) >> 3;
                staticBytes = (this.StaticLength + 3 + 7) >> 3;

                if (staticBytes <= optimalBytes)
                {
                    optimalBytes = staticBytes;
                }
            }
            else
            {
                // force a stored block
                optimalBytes = staticBytes = length + 5;
            }

            if (length + 4 <= optimalBytes && offset != -1)
            {
                // 4: two words for the lengths
                // The test buf != NULL is only necessary if LIT_BUFSIZE > WSIZE.
                // Otherwise we can't have processed more than WSIZE input bytes since
                // the last block flush, because compression would have been
                // successful. If LIT_BUFSIZE <= WSIZE, it is never too late to
                // transform a block into a stored block.
                this.SendStoredBlock(offset, length, lastBlock);
            }
            else
            {
                int last = lastBlock ? 1 : 0;
                if (staticBytes == optimalBytes)
                {
                    this.SendBits((StaticTreeBlock << 1) + last, 3);
                    this.SendCompressedBlock(StaticTree.LengthAndLiteralsTreeCodes, StaticTree.DistTreeCodes);
                }
                else
                {
                    this.SendBits((DynamicTreeBlock << 1) + last, 3);
                    this.SendAllTrees(this.treeLiterals.MaxCode + 1,
                        this.treeDistances.MaxCode + 1,
                        maxBitLengthIndex + 1);
                    this.SendCompressedBlock(this.dynamicLiteralsTree, this.dynamicDistanceTree);
                }
            }

            // The above check is made mod 2^32, for files larger than 512 MB
            // and uLong implemented on 32 bits.
            this.InitializeBlocks();

            if (lastBlock)
            {
                this.AlginOnByteBoundary();
            }
        }

        /// <summary>
        /// Fill the window if necessary.
        /// </summary>
        private void FillWindow()
        {
            // Amount of free space at the end of the window
            int more;
            int nextIndex;
            int previousIndex;

            do
            {
                more = this.actualWindowSize - this.lookAhead - this.startOfString;

                // Deal with 64K limit:
                if (more == 0 && this.startOfString == 0 && this.lookAhead == 0)
                {
                    more = this.windowSize;
                }
                else if (more == -1)
                {
                    // Very unlikely, but possible on 16 bit machine if startOfString == 0
                    // and this.lookAhead == 1 (input done one byte at time)
                    more--;
                }
                else if (this.startOfString >= this.windowSize + this.windowSize - MinLookAhead)
                {
                    // If the window is almost full and there is insufficient this.lookAhead,
                    // move the upper half to the lower one to make room in the upper half.
                    Array.Copy(this.window, this.windowSize, this.window, 0, this.windowSize);

                    this.matchStart -= this.windowSize;

                    // we now have startOfString >= MAX_DIST
                    this.startOfString -= this.windowSize;
                    this.blockStart -= this.windowSize;

                    // Slide the hash table (could be avoided with 32 bit values
                    // at the expense of memory usage). We slide even when level == 0
                    // to keep the hash table consistent if we switch back to level > 0
                    // later. (Using level 0 permanently is not an optimal usage of
                    // zlib, so we don't care about this pathological case.)
                    nextIndex = this.hashSize;
                    previousIndex = nextIndex;
                    do
                    {
                        int index = this.head[--previousIndex] & 0xffff;
                        this.head[previousIndex] = (short)(index >= this.windowSize ? index - this.windowSize : 0);
                    }
                    while (--nextIndex != 0);

                    nextIndex = this.windowSize;
                    previousIndex = nextIndex;
                    do
                    {
                        // If nextIndex is not on any hash chain, previous[nextIndex] is garbage but
                        // its value will never be used.
                        int index = this.previous[--previousIndex] & 0xffff;
                        this.previous[previousIndex] = (short)(index >= this.windowSize ? index - this.windowSize : 0);
                    }
                    while (--nextIndex != 0);

                    more += this.windowSize;
                }

                if (this.AvailableBytesIn == 0)
                {
                    return;
                }

                nextIndex = this.ReadBuffer(this.window, this.startOfString + this.lookAhead, more);
                this.lookAhead += nextIndex;

                // Initialize the hash value now that we have some input:
                if (this.lookAhead >= MinMatch)
                {
                    this.hashIndexOfString = this.window[this.startOfString] & 0xff;
                    this.hashIndexOfString = ((this.hashIndexOfString << this.hashShift)
                        ^ (this.window[this.startOfString + 1] & 0xff)) & this.hashMask;
                }
            }
            while (this.lookAhead < MinLookAhead && this.AvailableBytesIn != 0);
        }

        /// <summary>
        /// Compress as much as possible from the input stream, return the current
        /// block state.
        /// This function does not perform lazy evaluation of matches and inserts
        /// new strings in the dictionary only for unmatched strings or for short
        /// matches. It is used only for the fast compression options.
        /// </summary>
        /// <param name="flush">Flush flag.</param>
        /// <returns>Returns the current block state.</returns>
        private DeflateBlockState DeflateFast(bool flush)
        {
            // head of the hash chain
            int hashHead = 0;

            // set if current block must be flushed
            bool flushBlock;

            while (true)
            {
                // Make sure that we always have enough this.lookAhead, except
                // a case with last block in input buffer. We need MaxMatch bytes
                // for the next match, plus MinMatch bytes to insert the
                // string following the next match.
                if (this.lookAhead < MinLookAhead)
                {
                    this.FillWindow();
                    if (this.lookAhead < MinLookAhead && !flush)
                    {
                        return DeflateBlockState.NeedMore;
                    }

                    if (this.lookAhead == 0)
                    {
                        // flush the current block
                        break;
                    }
                }

                if (this.lookAhead >= MinMatch)
                {
                    // Insert the string window[startOfString .. startOfString+2] in the
                    // dictionary, and set hashHead to the head of the hash chain
                    hashHead = this.UpdateHead();
                }

                // Find the longest match, discarding those <= previous length.
                // At this point we have always this.matchLength < MinMatch
                if (hashHead != 0L && ((this.startOfString - hashHead) & 0xffff) <= this.windowSize - MinLookAhead)
                {
                    this.matchLength = this.LongestMatch(hashHead);
                }

                if (this.matchLength >= MinMatch)
                {
                    flushBlock = this.TreeTally(this.startOfString - this.matchStart, this.matchLength - MinMatch);

                    this.lookAhead -= this.matchLength;

                    // Insert new strings in the hash table only if the match length
                    // is not too large. This saves time but degrades compression.
                    if (this.matchLength <= this.maxLazy && this.lookAhead >= MinMatch)
                    {
                        // string at startOfString already in hash table
                        this.matchLength--;
                        do
                        {
                            this.startOfString++;

                            // startOfString never exceeds windowSize-MaxMatch,
                            // so there are always MinMatch bytes ahead.
                            hashHead = this.UpdateHead();
                        }
                        while (--this.matchLength != 0);

                        this.startOfString++;
                    }
                    else
                    {
                        // If this.lookAhead < MinMatch, this.startOfString is garbage,
                        // but it does not matter since it will be recomputed at next deflate call.
                        this.startOfString += this.matchLength;
                        this.matchLength = 0;
                        this.hashIndexOfString = this.window[this.startOfString] & 0xff;

                        this.hashIndexOfString = ((this.hashIndexOfString << this.hashShift)
                            ^ (this.window[this.startOfString + 1] & 0xff)) & this.hashMask;
                    }
                }
                else
                {
                    // No match, output a literal byte
                    flushBlock = this.TreeTally(0, this.window[this.startOfString] & 0xff);
                    this.lookAhead--;
                    this.startOfString++;
                }

                if (flushBlock)
                {
                    this.FlushBlockOnly(false);
                    if (this.AvailableBytesOut == 0)
                    {
                        return DeflateBlockState.NeedMore;
                    }
                }
            }

            return this.CompleteFlushBlock(flush);
        }

        private int UpdateHead()
        {
            this.hashIndexOfString = ((this.hashIndexOfString << this.hashShift)
                ^ (this.window[this.startOfString + MinMatch - 1] & 0xff)) & this.hashMask;

            int hashHead = this.head[this.hashIndexOfString] & 0xffff;
            this.previous[this.startOfString & this.windowMask] = this.head[this.hashIndexOfString];
            this.head[this.hashIndexOfString] = unchecked((short)this.startOfString);

            return hashHead;
        }

        private DeflateBlockState CompleteFlushBlock(bool flush)
        {
            this.FlushBlockOnly(flush);
            if (this.AvailableBytesOut == 0)
            {
                return flush ? DeflateBlockState.FinishStarted : DeflateBlockState.NeedMore;
            }

            return flush ? DeflateBlockState.FinishDone : DeflateBlockState.BlockDone;
        }

        /// <summary>
        /// Copy without compression as much as possible from the input buffer.
        /// </summary>
        /// <param name="flush">Flush flag.</param>
        /// <returns>Returns the current block state.</returns>
        private DeflateBlockState DeflateNone(bool flush)
        {
            // Stored blocks are limited to 0xffff bytes
            int maxBlockSize = 0xffff;
            int maxStart;
            if (maxBlockSize > this.pending.Length - 5)
            {
                // pending is limited to pending buffer size,
                // and each stored block has a 5 byte header
                maxBlockSize = this.pending.Length - 5;
            }

            // Copy as much as possible from input to output
            while (true)
            {
                // Fill the window as much as possible
                if (this.lookAhead <= 1)
                {
                    this.FillWindow();
                    if (this.lookAhead == 0 && !flush)
                    {
                        return DeflateBlockState.NeedMore;
                    }

                    if (this.lookAhead == 0)
                    {
                        // flush the current block
                        break;
                    }
                }

                this.startOfString += this.lookAhead;
                this.lookAhead = 0;

                // Emit a stored block if pending will be full
                maxStart = this.blockStart + maxBlockSize;
                if (this.startOfString == 0 || this.startOfString >= maxStart)
                {
                    this.lookAhead = (int)(this.startOfString - maxStart);
                    this.startOfString = (int)maxStart;

                    this.FlushBlockOnly(false);
                    if (this.AvailableBytesOut == 0)
                    {
                        return DeflateBlockState.NeedMore;
                    }
                }

                if (this.startOfString - this.blockStart >= this.windowSize - MinLookAhead)
                {
                    // Flush if we may have to slide, otherwise blockStart may become
                    // negative and the data will be gone
                    this.FlushBlockOnly(false);

                    if (this.AvailableBytesOut == 0)
                    {
                        return DeflateBlockState.NeedMore;
                    }
                }
            }

            return this.CompleteFlushBlock(flush);
        }

        /// <summary>
        /// Same as above, but achieves better compression. We use a lazy
        /// evaluation for matches: a match is finally adopted only if there is
        /// no better match at the next window position.
        /// </summary>
        /// <param name="flush">Flush flag.</param>
        /// <returns>Returns the current block state.</returns>
        private DeflateBlockState DeflateSlow(bool flush)
        {
            // head of hash chain
            int hashHead = 0;

            // set if current block must be flushed
            bool flushBlock;
            while (true)
            {
                // Make sure that we always have enough this.lookAhead, except
                // at the end of the input. We need MaxMatch bytes
                // for the next match, plus MinMatch bytes to insert the
                // string following the next match.
                if (this.lookAhead < MinLookAhead)
                {
                    this.FillWindow();
                    if (this.lookAhead < MinLookAhead && !flush)
                    {
                        return DeflateBlockState.NeedMore;
                    }

                    if (this.lookAhead == 0)
                    {
                        // flush the current block
                        break;
                    }
                }

                if (this.lookAhead >= MinMatch)
                {
                    // Insert the string window[startOfString .. startOfString+2] in the
                    // dictionary, and set hashHead to the head of the hash chain
                    hashHead = this.UpdateHead();
                }

                // Find the longest match, discarding those <= previousLength.
                this.previousLength = this.matchLength;
                this.previousMatch = this.matchStart;
                this.matchLength = MinMatch - 1;

                if (hashHead != 0 && this.previousLength < this.maxLazy &&
                    ((this.startOfString - hashHead) & 0xffff) <= this.windowSize - MinLookAhead)
                {
                    this.matchLength = this.LongestMatch(hashHead);

                    if (this.matchLength <= 5
                        && this.matchLength == MinMatch
                        && this.startOfString - this.matchStart > 4096)
                    {
                        // If previousMatch is also MinMatch, matchStart is garbage
                        // but we will ignore the current match anyway.
                        this.matchLength = MinMatch - 1;
                    }
                }

                // If there was a match at the previous step and the current
                // match is not better, output the previous match
                if (this.previousLength >= MinMatch && this.matchLength <= this.previousLength)
                {
                    int maxInsert = this.startOfString + this.lookAhead - MinMatch;
                    flushBlock = this.TreeTally(this.startOfString - 1 - this.previousMatch, this.previousLength - MinMatch);

                    // Insert in hash table all strings up to the end of the match.
                    // startOfString-1 and startOfString are already inserted. If there is not
                    // enough this.lookAhead, the last two strings are not inserted in
                    // the hash table.
                    this.lookAhead -= this.previousLength - 1;
                    this.previousLength -= 2;
                    do
                    {
                        if (++this.startOfString <= maxInsert)
                        {
                            hashHead = this.UpdateHead();
                        }
                    }
                    while (--this.previousLength != 0);

                    this.matchAvailable = 0;
                    this.matchLength = MinMatch - 1;
                    this.startOfString++;

                    if (flushBlock)
                    {
                        this.FlushBlockOnly(false);
                        if (this.AvailableBytesOut == 0)
                        {
                            return DeflateBlockState.NeedMore;
                        }
                    }
                }
                else if (this.matchAvailable != 0)
                {
                    // If there was no match at the previous position, output a
                    // single literal. If there was a match but the current match
                    // is longer, truncate the previous match to a single literal.
                    flushBlock = this.TreeTally(0, this.window[this.startOfString - 1] & 0xff);
                    if (flushBlock)
                    {
                        this.FlushBlockOnly(false);
                    }

                    this.startOfString++;
                    this.lookAhead--;
                    if (this.AvailableBytesOut == 0)
                    {
                        return DeflateBlockState.NeedMore;
                    }
                }
                else
                {
                    // There is no previous match to compare with,
                    // wait for the next step to decide.
                    this.matchAvailable = 1;
                    this.startOfString++;
                    this.lookAhead--;
                }
            }

            if (this.matchAvailable != 0)
            {
                flushBlock = this.TreeTally(0, this.window[this.startOfString - 1] & 0xff);
                this.matchAvailable = 0;
            }

            return this.CompleteFlushBlock(flush);
        }

        private int LongestMatch(int currentMatch)
        {
            // max hash chain length
            int chainLength = this.maxChainLength;

            // current string
            int scan = this.startOfString;

            // matched string
            int match;

            // length of current match
            int len;

            // best match length so far
            int bestLength = this.previousLength;
            int limit = this.startOfString > (this.windowSize - MinLookAhead)
                ? this.startOfString - (this.windowSize - MinLookAhead) : 0;
            int niceLength = this.niceLength;

            // Stop when current match becomes <= limit. To simplify the code,
            // we prevent matches with the string of window index 0.
            int mask = this.windowMask;

            int end = this.startOfString + MaxMatch;
            int scanEndIndex = scan + bestLength;
            byte scanEnd1 = this.window[scanEndIndex - 1];
            byte scanEnd = this.window[scanEndIndex];

            // The code is optimized for hashBits >= 8 and MaxMatch-2 multiple of 16.
            // It is easy to get rid of this optimization if necessary.
            // Do not waste too much time if we already have a good match
            if (this.previousLength >= this.goodLength)
            {
                chainLength >>= 2;
            }

            // Do not look for matches beyond the end of the input.
            // This is necessary to make deflate deterministic.
            if (niceLength > this.lookAhead)
            {
                niceLength = this.lookAhead;
            }

            do
            {
                match = currentMatch;

                if (this.window[match + bestLength] != scanEnd ||
                    this.window[match + bestLength - 1] != scanEnd1 ||
                    this.window[match] != this.window[scan] ||
                    this.window[++match] != this.window[scan + 1])
                {
                    // Skip to next match if the match length cannot increase
                    // or if the match length is less than 2
                    continue;
                }

                // The check at bestLength-1 can be removed because it will be made
                // again later. (This heuristic is not always a win.)
                // It is not necessary to compare scan[2] and match[2] since they
                // are always equal when the other bytes match, given that
                // the hash keys are equal and that hashBits >= 8.
                scan += 2;
                match++;

                // We check for insufficient this.lookAhead only every 8th comparison;
                // the 256th check will be made at startOfString+258.
                do
                {
                }
                while (this.window[++scan] == this.window[++match] &&
                       this.window[++scan] == this.window[++match] &&
                       this.window[++scan] == this.window[++match] &&
                       this.window[++scan] == this.window[++match] &&
                       this.window[++scan] == this.window[++match] &&
                       this.window[++scan] == this.window[++match] &&
                       this.window[++scan] == this.window[++match] &&
                       this.window[++scan] == this.window[++match] && scan < end);

                len = MaxMatch - (int)(end - scan);
                scan = end - MaxMatch;

                if (len > bestLength)
                {
                    this.matchStart = currentMatch;
                    bestLength = len;
                    if (len >= niceLength)
                    {
                        break;
                    }

                    scanEnd1 = this.window[scan + bestLength - 1];
                    scanEnd = this.window[scan + bestLength];
                }
            }
            while ((currentMatch = this.previous[currentMatch & mask] & 0xffff) > limit && --chainLength != 0);

            if (bestLength <= this.lookAhead)
            {
                return bestLength;
            }

            return this.lookAhead;
        }

        private void Initialize(DeflateSettings deflateSettings)
        {
            this.compressionLevel = (int)deflateSettings.CompressionLevel;
            this.windowBits = WindowBitsDefault;

            int memoryLevel = DefaultMemoryLevel;

            this.windowSize = 1 << this.windowBits;
            this.windowMask = this.windowSize - 1;

            this.hashBits = memoryLevel + 7;
            this.hashSize = 1 << this.hashBits;
            this.hashMask = this.hashSize - 1;
            this.hashShift = (this.hashBits + MinMatch - 1) / MinMatch;

            this.window = new byte[this.windowSize * 2];
            this.previous = new short[this.windowSize];
            this.head = new short[this.hashSize];

            // for memLevel==8, this will be 16384, 16k
            this.literalsBufsize = 1 << (memoryLevel + 6);

            // Use a single array as the buffer for data pending compression,
            // the output distance codes, and the output length codes (aka tree).
            // orig comment: This works just fine since the average
            // output size for (length, distance) codes is <= 24 bits.
            this.pending = new byte[this.literalsBufsize * 4];
            this.distanceOffset = this.literalsBufsize;
            this.lengthOffset = (1 + 2) * this.literalsBufsize;

            // So, for memLevel 8, the length of the pending buffer is 65536. 64k.
            // The first 16k are pending bytes.
            // The middle slice, of 32k, is used for distance codes.
            // The final 16k are length codes.
            this.TotalBytesIn = this.TotalBytesOut = 0;

            this.pendingCount = 0;
            this.nextPending = 0;

            this.InitializeTreeData();
            this.InitializeLazyMatch();
        }

        private void InitializeLazyMatch()
        {
            this.actualWindowSize = 2 * this.windowSize;

            // clear the hash
            Array.Clear(this.head, 0, this.hashSize);

            this.SetConfiguration(this.compressionLevel);

            this.startOfString = 0;
            this.blockStart = 0;
            this.lookAhead = 0;
            this.matchLength = this.previousLength = MinMatch - 1;
            this.matchAvailable = 0;
            this.hashIndexOfString = 0;
        }

        /// <summary>
        /// Initialize the tree data structures.
        /// </summary>
        private void InitializeTreeData()
        {
            this.treeLiterals.DynamicTree = this.dynamicLiteralsTree;
            this.treeLiterals.StaticTree = StaticTree.Literals;

            this.treeDistances.DynamicTree = this.dynamicDistanceTree;
            this.treeDistances.StaticTree = StaticTree.Distances;

            this.treeBitLengths.DynamicTree = this.dynamicBitLengthTree;
            this.treeBitLengths.StaticTree = StaticTree.BitLengths;

            this.bitsBuffer = 0;
            this.bitsValid = 0;

            // Initialize the first block
            this.InitializeBlocks();
        }

        private void InitializeBlocks()
        {
            // Initialize the trees.
            for (int i = 0; i < DeflateConstants.LiteralCodes; i++)
            {
                this.dynamicLiteralsTree[i * 2] = 0;
            }

            for (int i = 0; i < DeflateConstants.DynamicCodes; i++)
            {
                this.dynamicDistanceTree[i * 2] = 0;
            }

            for (int i = 0; i < DeflateConstants.BitLengthCodes; i++)
            {
                this.dynamicBitLengthTree[i * 2] = 0;
            }

            this.dynamicLiteralsTree[EndBlock * 2] = 1;
            this.OptimalLength = this.StaticLength = 0;
            this.lastLiteral = this.matches = 0;
        }

        /// <summary>
        /// Sets configuration parameters by the compression level.
        /// </summary>
        /// <param name="level">Compression level.</param>
        private void SetConfiguration(int level)
        {
            DeflateConfiguration configuration = DeflateConfiguration.Lookup(level);
            this.goodLength = configuration.GoodLength;
            this.maxChainLength = configuration.MaxChainLength;
            this.maxLazy = configuration.MaxLazy;
            this.niceLength = configuration.NiceLength;

            if (level > 0)
            {
                if (level > 3)
                {
                    this.deflateFunction = this.DeflateSlow;
                }
                else
                {
                    this.deflateFunction = this.DeflateFast;
                }
            }
            else
            {
                this.deflateFunction = this.DeflateNone;
            }
        }

        /// <summary>
        /// Flush as much pending output as possible.
        /// All deflate output goes through this function.
        /// </summary>
        private void FlushPending()
        {
            int length = this.pendingCount;

            if (length > this.AvailableBytesOut)
            {
                length = this.AvailableBytesOut;
            }

            if (length == 0)
            {
                return;
            }

            if (this.pending.Length <= this.nextPending ||
                this.OutputBuffer.Length <= this.NextOut ||
                this.pending.Length < (this.nextPending + length) ||
                this.OutputBuffer.Length < (this.NextOut + length))
            {
                throw new InvalidOperationException(string.Format("Invalid State. (pending.Length={0}, pendingCount={1})",
                    this.pending.Length,
                    this.pendingCount));
            }

            Array.Copy(this.pending, this.nextPending, this.OutputBuffer, this.NextOut, length);

            this.NextOut += length;
            this.nextPending += length;
            this.TotalBytesOut += length;
            this.AvailableBytesOut -= length;
            this.pendingCount -= length;
            if (this.pendingCount == 0)
            {
                this.nextPending = 0;
            }
        }

        /// <summary>
        /// Read a new buffer from the current input stream, update
        /// total number of bytes read.  All deflate input goes through
        /// this function.
        /// </summary>
        /// <param name="buffer">Buffer.</param>
        /// <param name="start">Start position in buffer.</param>
        /// <param name="size">Size.</param>
        /// <returns></returns>
        private int ReadBuffer(byte[] buffer, int start, int size)
        {
            int length = this.AvailableBytesIn;

            if (length > size)
            {
                length = size;
            }

            if (length == 0)
            {
                return 0;
            }

            this.AvailableBytesIn -= length;

            Array.Copy(this.InputBuffer, this.NextIn, buffer, start, length);
            this.NextIn += length;
            this.TotalBytesIn += length;

            return length;
        }
    }
}
