﻿using System;
using System.Collections.Generic;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Decompressor which implements Deflate compression.
    /// </summary>
    internal class DeflateDecompressor : DeflateTransformBase
    {
        // Length codes 257..285 base
        private static readonly int[] lengthBase = new int[]
        {
            3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27, 31,
            35, 43, 51, 59, 67, 83, 99, 115, 131, 163, 195, 227, 258
        };

        // Distance codes 0..29 base
        private static readonly int[] distanceBasePosition = new int[]
        {
            1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193, 257, 385, 513,
            769, 1025, 1537, 2049, 3073, 4097, 6145, 8193, 12289, 16385, 24577, 0, 0
        };

        private static readonly byte[] staticDistanceTreeTable = new byte[]
        {
            0, 16, 8, 24, 4, 20, 12, 28, 2, 18, 10, 26, 6, 22, 14, 30,
            1, 17, 9, 25, 5, 21, 13, 29, 3, 19, 11, 27, 7, 23, 15, 31
        };

        private DecompressorState state;
        private InputBitsBuffer input;
        private OutputWindow output;

        private int bfinal;
        private BlockType blockType;
        private byte[] blockLengthBuffer = new byte[4];
        private int blockLength;
        private int distanceCode;
        private int extraBits;
        private int loopCounter;
        private int literalLengthCodeCount;
        private int distanceCodeCount;
        private int codeLengthCodeCount;
        private int codeArraySize;
        private int lengthCode;

        private byte[] codeList;
        private byte[] codeLengthTreeCodeLength;

        private InflateTree literalLengthTree;
        private InflateTree distanceTree;
        private InflateTree codeLengthTree;

        private Queue<byte[]> pendingInput = new Queue<byte[]>();

        /// <summary>
        /// Initializes a new instance of the DeflateDecompressor class.
        /// </summary>
        /// <param name="settings">Deflate settings.</param>
        public DeflateDecompressor(DeflateSettings settings)
            : base(settings)
        {
            this.output = new OutputWindow();
            this.input = new InputBitsBuffer();
            this.codeList = new byte[DeflateConstants.MaxLiteralTreeElements + DeflateConstants.MaxDistanceTreeElements];
            this.codeLengthTreeCodeLength = new byte[DeflateConstants.BitLengthCodes];
            this.Reset();
        }

        /// <summary>
        /// Represents a state of decompressor process.
        /// </summary>
        private enum DecompressorState
        {
            ReadingBFinal = 2,
            ReadingBType = 3,
            ReadingNumLitCodes = 4,
            ReadingNumDistCodes = 5,
            ReadingNumCodeLengthCodes = 6,
            ReadingCodeLengthCodes = 7,
            ReadingTreeCodesBefore = 8,
            ReadingTreeCodesAfter = 9,
            DecodeTop = 10,
            HaveInitialLength = 11,
            HaveFullLength = 12,
            HaveDistCode = 13,
            UncompressedAligning = 15,
            UncompressedByte1 = 16,
            UncompressedByte2 = 17,
            UncompressedByte3 = 18,
            UncompressedByte4 = 19,
            DecodingUncompressed = 20,
            Done = 24
        }

        /// <summary>
        /// Represents a type of block in deflated data.
        /// </summary>
        private enum BlockType
        {
            Stored = 0,
            Static = 1,
            Dynamic = 2
        }

        /// <summary>Gets the output block size.</summary>
        /// <returns>The size of the output data blocks in bytes.</returns>
        public override int OutputBlockSize
        {
            get
            {
                return 262144;
            }
        }

        /// <summary>
        /// Initialize reading of the transformation header.
        /// </summary>
        public override void InitHeaderReading()
        {
            base.InitHeaderReading();

            if (this.Settings.HeaderType == CompressedStreamHeader.ZLib)
            {
                this.Header.BytesToRead = 2;
            }
        }

        /// <summary>
        /// Process transformation header has been read.
        /// </summary>
        public override void ProcessHeader()
        {
            base.ProcessHeader();

            if (this.Settings.HeaderType == CompressedStreamHeader.ZLib)
            {
                string message = string.Empty;
                if (this.Header.Buffer == null || this.Header.Buffer.Length != 2)
                {
                    message = "Invalid header length";
                }
                else
                {
                    int method = this.Header.Buffer[0];
                    int windowSize = (method >> 4) + 8;
                    int checkByte = this.Header.Buffer[1];
                    if ((method & 0x0f) != DeflateConstants.HeaderDeflated)
                    {
                        message = string.Format("Unknown compression method (0x{0:X2})", method);
                    }
                    else if (windowSize > DeflateConstants.MaxBits)
                    {
                        message = string.Format("Invalid window size ({0})", windowSize);
                    }
                    else if ((((method << 8) + checkByte) % 31) != 0)
                    {
                        message = "Invalid header";
                    }
                }

                if (message.Length != 0)
                {
                    throw new InvalidDataException(message);
                }
            }

            this.Header.BytesToRead = 0;
        }

        /// <summary>
        /// Transforms current input buffer.
        /// </summary>
        /// <param name="finalBlock">The final block flag.</param>
        /// <returns>True when still output available.</returns>
        protected override bool ProcessTransform(bool finalBlock)
        {
            this.SetInputBuffer();
            int length = this.Inflate(this.OutputBuffer, this.NextOut, this.AvailableBytesOut);
            this.NextOut += length;

            return this.output.AvailableBytes > 0
                || (this.state != DecompressorState.Done
                && (!this.input.InputRequired || this.pendingInput.Count > 0));
        }

        private static void ThrowInvalidData()
        {
            throw new InvalidDataException("Invalid data");
        }

        private static void ThrowInvalidDataGeneric()
        {
            throw new InvalidDataException();
        }

        private static void ThrowUnknownBlockType()
        {
            throw new InvalidDataException("Unknown block type");
        }

        private static void ThrowUnknownState()
        {
            throw new InvalidDataException("Unknown state");
        }

        private bool Decode()
        {
            if (this.state == DecompressorState.ReadingBFinal)
            {
                if (!this.input.CheckAvailable(1))
                {
                    return false;
                }

                this.bfinal = this.input.GetBits(1);
                this.state = DecompressorState.ReadingBType;
            }

            if (this.state == DecompressorState.ReadingBType)
            {
                if (!this.input.CheckAvailable(2))
                {
                    return false;
                }

                this.ProcessBlockType();
            }

            bool ready = this.CheckDecodeState();
            return ready;
        }

        private void ProcessBlockType()
        {
            this.blockType = (BlockType)this.input.GetBits(2);
            switch (this.blockType)
            {
                case BlockType.Dynamic:
                    this.state = DecompressorState.ReadingNumLitCodes;
                    break;

                case BlockType.Static:
                    this.literalLengthTree = InflateTree.StaticLiteralLengthTree;
                    this.distanceTree = InflateTree.StaticDistanceTree;
                    this.state = DecompressorState.DecodeTop;
                    break;

                case BlockType.Stored:
                    this.state = DecompressorState.UncompressedAligning;
                    break;

                default:
                    DeflateDecompressor.ThrowUnknownBlockType();
                    break;
            }
        }

        private bool CheckDecodeState()
        {
            bool ready = false;
            bool endOfBlock = false;
            if (this.blockType == BlockType.Dynamic)
            {
                if (this.state >= DecompressorState.DecodeTop)
                {
                    ready = this.DecodeBlock(out endOfBlock);
                }
                else
                {
                    ready = this.DecodeDynamicBlockHeader();
                }
            }
            else if (this.blockType != BlockType.Static)
            {
                if (this.blockType != BlockType.Stored)
                {
                    DeflateDecompressor.ThrowUnknownBlockType();
                }

                ready = this.DecodeUncompressedBlock(out endOfBlock);
            }
            else
            {
                ready = this.DecodeBlock(out endOfBlock);
            }

            if (endOfBlock && this.bfinal != 0)
            {
                this.state = DecompressorState.Done;
            }

            return ready;
        }

        private bool DecodeBlock(out bool endOfBlock)
        {
            int freeBytes = this.output.FreeBytes;
            endOfBlock = false;
            while (freeBytes > 258)
            {
                switch (this.state)
                {
                    case DecompressorState.DecodeTop:
                        {
                            bool? shouldReturn = this.DecodeTop(ref endOfBlock, ref freeBytes);
                            if (shouldReturn != null)
                            {
                                return (bool)shouldReturn;
                            }
                            continue;
                        }
                    case DecompressorState.HaveInitialLength:
                        {
                            bool? shouldReturn = this.DecodeInitialLength();
                            if (shouldReturn != null)
                            {
                                return (bool)shouldReturn;
                            }
                            continue;
                        }
                    case DecompressorState.HaveFullLength:
                        {
                            bool? shouldReturn = this.DecodeFullLength();
                            if (shouldReturn != null)
                            {
                                return (bool)shouldReturn;
                            }
                            continue;
                        }
                    case DecompressorState.HaveDistCode:
                        {
                            bool? shouldReturn = this.DecodeDistanceCode(ref freeBytes);
                            if (shouldReturn != null)
                            {
                                return (bool)shouldReturn;
                            }
                            continue;
                        }
                }

                DeflateDecompressor.ThrowUnknownState();
            }

            return true;
        }

        private bool? DecodeTop(ref bool endOfBlock, ref int freeBytes)
        {
            int nextSymbol = this.literalLengthTree.GetNextSymbol(this.input);
            if (nextSymbol < 0)
            {
                return false;
            }

            if (nextSymbol >= DeflateConstants.EndOfBlockCode)
            {
                if (nextSymbol == DeflateConstants.EndOfBlockCode)
                {
                    endOfBlock = true;
                    this.state = DecompressorState.ReadingBFinal;
                    return true;
                }
                else
                {
                    this.DecodeDistance(ref nextSymbol);
                    this.state = DecompressorState.HaveInitialLength;
                }
            }
            else
            {
                this.output.AddByte((byte)nextSymbol);
                freeBytes--;
            }

            return null;
        }

        private void DecodeDistance(ref int nextSymbol)
        {
            nextSymbol -= 257;
            if (nextSymbol < 8)
            {
                nextSymbol += 3;
                this.extraBits = 0;
            }
            else if (nextSymbol != 28)
            {
                if (nextSymbol < 0 || nextSymbol >= Tree.ExtraLengthBits.Length)
                {
                    DeflateDecompressor.ThrowInvalidData();
                }

                this.extraBits = Tree.ExtraLengthBits[nextSymbol];
            }
            else
            {
                nextSymbol = 258;
                this.extraBits = 0;
            }

            this.blockLength = nextSymbol;
        }

        private bool? DecodeInitialLength()
        {
            if (this.extraBits > 0)
            {
                int bits = this.input.GetBits(this.extraBits);
                if (bits < 0)
                {
                    return false;
                }

                if (this.blockLength < 0 || this.blockLength >= DeflateDecompressor.lengthBase.Length)
                {
                    DeflateDecompressor.ThrowInvalidData();
                }

                this.blockLength = DeflateDecompressor.lengthBase[this.blockLength] + bits;
            }

            this.state = DecompressorState.HaveFullLength;
            return null;
        }

        private bool? DecodeFullLength()
        {
            if (this.blockType != BlockType.Dynamic)
            {
                this.distanceCode = this.input.GetBits(5);
                if (this.distanceCode >= 0)
                {
                    this.distanceCode = DeflateDecompressor.staticDistanceTreeTable[this.distanceCode];
                }
            }
            else
            {
                this.distanceCode = this.distanceTree.GetNextSymbol(this.input);
            }

            if (this.distanceCode < 0)
            {
                return false;
            }

            this.state = DecompressorState.HaveDistCode;
            return null;
        }

        private bool? DecodeDistanceCode(ref int freeBytes)
        {
            int offset;
            if (this.distanceCode <= 3)
            {
                offset = this.distanceCode + 1;
            }
            else
            {
                this.extraBits = (this.distanceCode - 2) >> 1;
                int bits1 = this.input.GetBits(this.extraBits);
                if (bits1 < 0)
                {
                    return false;
                }

                offset = DeflateDecompressor.distanceBasePosition[this.distanceCode] + bits1;
            }

            this.output.Copy(this.blockLength, offset);
            freeBytes -= this.blockLength;
            this.state = DecompressorState.DecodeTop;

            return null;
        }

        private bool DecodeDynamicBlockHeader()
        {
            switch (this.state)
            {
                case DecompressorState.ReadingNumLitCodes:
                    {
                        this.literalLengthCodeCount = this.input.GetBits(5);
                        if (this.literalLengthCodeCount < 0)
                        {
                            return false;
                        }

                        this.literalLengthCodeCount += 257;
                        this.state = DecompressorState.ReadingNumDistCodes;
                        goto case DecompressorState.ReadingNumDistCodes;
                    }
                case DecompressorState.ReadingNumDistCodes:
                    {
                        this.distanceCodeCount = this.input.GetBits(5);
                        if (this.distanceCodeCount < 0)
                        {
                            return false;
                        }

                        this.distanceCodeCount++;
                        this.state = DecompressorState.ReadingNumCodeLengthCodes;
                        goto case DecompressorState.ReadingNumCodeLengthCodes;
                    }
                case DecompressorState.ReadingNumCodeLengthCodes:
                    {
                        this.codeLengthCodeCount = this.input.GetBits(4);
                        if (this.codeLengthCodeCount < 0)
                        {
                            return false;
                        }

                        this.codeLengthCodeCount += 4;
                        this.loopCounter = 0;
                        this.state = DecompressorState.ReadingCodeLengthCodes;
                        goto case DecompressorState.ReadingCodeLengthCodes;
                    }
                case DecompressorState.ReadingCodeLengthCodes:
                    {
                        if (!this.DecodeDynamicCodes())
                        {
                            return false;
                        }

                        this.state = DecompressorState.ReadingTreeCodesBefore;
                        goto case DecompressorState.ReadingTreeCodesBefore;
                    }
                case DecompressorState.ReadingTreeCodesBefore:
                case DecompressorState.ReadingTreeCodesAfter:
                    {
                        return this.ReadTreeCodes();
                    }
            }

            DeflateDecompressor.ThrowUnknownState();
            return true;
        }

        private bool DecodeDynamicCodes()
        {
            while (this.loopCounter < this.codeLengthCodeCount)
            {
                int bits = this.input.GetBits(3);
                if (bits < 0)
                {
                    return false;
                }

                this.codeLengthTreeCodeLength[Tree.BitLengthOrder[this.loopCounter]] = (byte)bits;
                this.loopCounter++;
            }

            for (int i = this.codeLengthCodeCount; i < Tree.BitLengthOrder.Length; i++)
            {
                this.codeLengthTreeCodeLength[Tree.BitLengthOrder[i]] = 0;
            }

            this.codeLengthTree = new InflateTree(this.codeLengthTreeCodeLength);
            this.codeArraySize = this.literalLengthCodeCount + this.distanceCodeCount;
            this.loopCounter = 0;

            return true;
        }

        private bool ReadTreeCodes()
        {
            while (this.loopCounter < this.codeArraySize)
            {
                if (this.state == DecompressorState.ReadingTreeCodesBefore)
                {
                    this.lengthCode = this.codeLengthTree.GetNextSymbol(this.input);
                    if (this.lengthCode < 0)
                    {
                        return false;
                    }
                }

                if (this.lengthCode > 15)
                {
                    if (!this.input.CheckAvailable(7))
                    {
                        this.state = DecompressorState.ReadingTreeCodesAfter;
                        return false;
                    }

                    this.SetPreviousCode();
                }
                else
                {
                    this.codeList[this.loopCounter++] = (byte)this.lengthCode;
                }

                this.state = DecompressorState.ReadingTreeCodesBefore;
            }

            this.SetTreeCodes();
            this.state = DecompressorState.DecodeTop;

            return true;
        }

        private void SetPreviousCode()
        {
            int counter;
            byte previousCode = 0;
            if (this.lengthCode == 16)
            {
                if (this.loopCounter == 0)
                {
                    DeflateDecompressor.ThrowInvalidDataGeneric();
                }

                previousCode = this.codeList[this.loopCounter - 1];
                counter = this.input.GetBits(2) + 3;
            }
            else if (this.lengthCode != 17)
            {
                counter = this.input.GetBits(7) + 11;
            }
            else
            {
                counter = this.input.GetBits(3) + 3;
            }

            if (this.loopCounter + counter > this.codeArraySize)
            {
                DeflateDecompressor.ThrowInvalidDataGeneric();
            }

            for (int i = 0; i < counter; i++)
            {
                this.codeList[this.loopCounter++] = previousCode;
            }
        }

        private void SetTreeCodes()
        {
            byte[] literalTreeCodeLength = new byte[DeflateConstants.MaxLiteralTreeElements];
            byte[] distanceTreeCodeLength = new byte[DeflateConstants.MaxDistanceTreeElements];

            Array.Copy(this.codeList, literalTreeCodeLength, this.literalLengthCodeCount);
            Array.Copy(this.codeList, this.literalLengthCodeCount, distanceTreeCodeLength, 0, this.distanceCodeCount);

            if (literalTreeCodeLength[DeflateConstants.EndOfBlockCode] == 0)
            {
                DeflateDecompressor.ThrowInvalidDataGeneric();
            }

            this.literalLengthTree = new InflateTree(literalTreeCodeLength);
            this.distanceTree = new InflateTree(distanceTreeCodeLength);
        }

        private bool DecodeUncompressedBlock(out bool endOfBlock)
        {
            endOfBlock = false;
            while (this.state != DecompressorState.DecodingUncompressed)
            {
                switch (this.state)
                {
                    case DecompressorState.UncompressedAligning:
                        {
                            this.input.SkipToByteBoundary();
                            this.state = DecompressorState.UncompressedByte1;
                            continue;
                        }

                    case DecompressorState.UncompressedByte1:
                    case DecompressorState.UncompressedByte2:
                    case DecompressorState.UncompressedByte3:
                    case DecompressorState.UncompressedByte4:
                        {
                            if (!this.DecodeBits())
                            {
                                return false;
                            }
                            continue;
                        }

                    default:
                        DeflateDecompressor.ThrowUnknownState();
                        break;
                }
            }

            int copied = this.output.ReadInput(this.input, this.blockLength);
            this.blockLength -= copied;
            if (this.blockLength == 0)
            {
                this.state = DecompressorState.ReadingBFinal;
                endOfBlock = true;
                return true;
            }

            return this.output.FreeBytes == 0;
        }

        private bool DecodeBits()
        {
            int bits = this.input.GetBits(8);
            if (bits < 0)
            {
                return false;
            }

            this.blockLengthBuffer[(int)this.state - (int)DecompressorState.UncompressedByte1] = (byte)bits;
            if (this.state == DecompressorState.UncompressedByte4)
            {
                this.blockLength = this.blockLengthBuffer[0] + ((int)this.blockLengthBuffer[1] * 256);
                int blockLengthInvertion = blockLengthBuffer[2] + ((int)blockLengthBuffer[3] * 256);

                if ((ushort)this.blockLength != (ushort)(~blockLengthInvertion))
                {
                    throw new InvalidDataException("Invalid block length");
                }
            }

            this.state = (DecompressorState)((int)this.state + 1);

            return true;
        }

        private int Inflate(byte[] bytes, int offset, int length)
        {
            int decompressed = 0;
            while (length > 0 && this.state != DecompressorState.Done)
            {
                int copied = this.output.Read(bytes, offset, length);

                if (copied > 0)
                {
                    offset += copied;
                    decompressed += copied;
                    length -= copied;
                }

                if (!this.Decode()
                && (this.state == DecompressorState.Done || !this.TrySetInputBuffer()))
                {
                    break;
                }
            }

            if (this.output.AvailableBytes > 0 && length > 0)
            {
                int copied = this.output.Read(bytes, offset, length);
                decompressed += copied;
            }

            return decompressed;
        }

        private void Reset()
        {
            this.state = DecompressorState.ReadingBFinal;
        }

        private void SetInput(byte[] inputBytes, int offset, int length)
        {
            this.input.SetBuffer(inputBytes, offset, length);
        }

        private void EnqueueInputBuffer()
        {
            if (this.AvailableBytesIn > 0)
            {
                byte[] buffer = new byte[this.AvailableBytesIn];
                Array.Copy(this.InputBuffer, 0, buffer, 0, buffer.Length);

                this.pendingInput.Enqueue(buffer);
            }
        }

        private void SetInputBuffer()
        {
            if (this.input.InputRequired)
            {
                byte[] buffer;

                if (this.pendingInput.Count > 0)
                {
                    this.EnqueueInputBuffer();

                    buffer = this.pendingInput.Dequeue();
                    this.AvailableBytesIn = buffer.Length;
                }
                else
                {
                    buffer = new byte[this.AvailableBytesIn];
                    Array.Copy(this.InputBuffer, 0, buffer, 0, buffer.Length);
                }

                this.SetInput(buffer, 0, this.AvailableBytesIn);
            }
            else
            {
                this.EnqueueInputBuffer();
            }
        }

        private bool TrySetInputBuffer()
        {
            if (this.input.InputRequired && this.pendingInput.Count > 0)
            {
                byte[] buffer = this.pendingInput.Dequeue();
                this.AvailableBytesIn = buffer.Length;
                this.SetInput(buffer, 0, this.AvailableBytesIn);

                return true;
            }

            return false;
        }
    }
}
