﻿using System;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Inflates data using a lookup table combined with a HuffmanTree.
    /// </summary>
    internal class InflateTree
    {
        private byte[] codeLengthArray;
        private short[] left;
        private short[] right;
        private short[] table;
        private int tableBits;
        private int tableMask;
        private int tableBitMask;

        /// <summary>
        /// Initializes static members of the InflateTree class.
        /// </summary>
        static InflateTree()
        {
            InflateTree.StaticLiteralLengthTree = new InflateTree(InflateTree.GetStaticLiteralTreeLength());
            InflateTree.StaticDistanceTree = new InflateTree(InflateTree.GetStaticDistanceTreeLength());
        }

        /// <summary>
        /// Initializes a new instance of the InflateTree class.
        /// </summary>
        public InflateTree(byte[] codeLengths)
        {
            this.codeLengthArray = codeLengths;
            if (this.codeLengthArray.Length == DeflateConstants.MaxLiteralTreeElements)
            {
                this.tableBits = 9;
            }
            else
            {
                this.tableBits = 7;
            }

            this.tableBitMask = 1 << (this.tableBits & 0x1f);
            this.tableMask = this.tableBitMask - 1;
            this.CreateTable();
        }

        public static InflateTree StaticDistanceTree
        {
            get;
            private set;
        }

        public static InflateTree StaticLiteralLengthTree
        {
            get;
            private set;
        }

        /// <summary>
        /// Tries to get enough bits from input and try to decode them.
        /// </summary>
        /// <param name="input">Input buffer.</param>
        /// <returns>Next symbol or -1 when there is no enough bits in input.</returns>
        public int GetNextSymbol(InputBitsBuffer input)
        {
            uint bits = input.Get16Bits();
            if (input.AvailableBits == 0)
            {
                return -1;
            }

            int nextSymbol = this.table[bits & this.tableMask];
            if (nextSymbol < 0)
            {
                uint bitsMask = (uint)this.tableBitMask;
                do
                {
                    nextSymbol = -nextSymbol;
                    nextSymbol = (bits & bitsMask) != 0 ? this.right[nextSymbol] : this.left[nextSymbol];
                    bitsMask <<= 1;
                }
                while (nextSymbol < 0);
            }

            int length = this.codeLengthArray[nextSymbol];
            if (length <= 0)
            {
                InflateTree.InvalidHuffmanData();
            }

            if (length > input.AvailableBits)
            {
                return -1;
            }

            input.SkipBits(length);

            return nextSymbol;
        }

        private static byte[] GetStaticDistanceTreeLength()
        {
            byte[] distanceTree = new byte[DeflateConstants.MaxDistanceTreeElements];
            for (int i = 0; i < DeflateConstants.MaxDistanceTreeElements; i++)
            {
                distanceTree[i] = 5;
            }

            return distanceTree;
        }

        private static byte[] GetStaticLiteralTreeLength()
        {
            byte[] literalTree = new byte[DeflateConstants.MaxLiteralTreeElements];
            for (int i = 0; i <= 143; i++)
            {
                literalTree[i] = 8;
            }
            for (int j = 144; j <= 255; j++)
            {
                literalTree[j] = 9;
            }
            for (int k = 256; k <= 279; k++)
            {
                literalTree[k] = 7;
            }
            for (int l = 280; l <= 287; l++)
            {
                literalTree[l] = 8;
            }

            return literalTree;
        }

        private static void InvalidHuffmanData()
        {
            throw new InvalidDataException("Invalid huffman data");
        }

        /// <summary>
        /// Calculate the huffman codes according to RFC 1951.
        /// </summary>
        /// <returns>Huffman codes.</returns>
        private uint[] CalculateHuffmanCode()
        {
            uint[] bitLengthCount = new uint[17];
            foreach (int codeLength in this.codeLengthArray)
            {
                bitLengthCount[codeLength]++;
            }

            bitLengthCount[0] = 0;

            uint code = 0;
            uint[] nextCode = new uint[17];
            for (int i = 1; i <= 16; i++)
            {
                code = (code + bitLengthCount[i - 1]) << 1;
                nextCode[i] = code;
            }

            uint[] codes = new uint[DeflateConstants.MaxLiteralTreeElements];
            for (int i = 0; i < this.codeLengthArray.Length; i++)
            {
                int length = this.codeLengthArray[i];
                if (length > 0)
                {
                    codes[i] = (uint)Tree.BitReverse((int)nextCode[length], length);
                    nextCode[length]++;
                }
            }

            return codes;
        }

        private void CreateTable()
        {
            short avail = (short)this.codeLengthArray.Length;
            uint[] codes = this.CalculateHuffmanCode();
            int size = this.codeLengthArray.Length * 2;

            this.table = new short[this.tableBitMask];

            this.left = new short[size];
            this.right = new short[size];

            for (int i = 0; i < this.codeLengthArray.Length; i++)
            {
                int length = this.codeLengthArray[i];
                if (length > 0)
                {
                    int start = (int)codes[i];
                    short code = (short)i;
                    if (length > this.tableBits)
                    {
                        this.BuildTree(code, start, length, ref avail);
                    }
                    else
                    {
                        this.DuplicateCode(code, start, length);
                    }
                }
            }
        }

        private void DuplicateCode(short code, int start, int length)
        {
            int offset = 1 << (length & 0x1f);
            if (start >= offset)
            {
                InflateTree.InvalidHuffmanData();
            }

            int repeats = 1 << (this.tableBits - length & 0x1f);
            for (int i = 0; i < repeats; i++)
            {
                this.table[start] = code;
                start += offset;
            }
        }

        private void BuildTree(short code, int start, int length, ref short avail)
        {
            int overflowBits = length - this.tableBits;
            int bitMask = this.tableBitMask;
            int index = start & this.tableMask;
            short[] usedTable = this.table;

            do
            {
                short value = usedTable[index];
                if (value == 0)
                {
                    value = (short)(-avail);
                    usedTable[index] = value;
                    avail = (short)(avail + 1);
                }

                if (value > 0)
                {
                    InflateTree.InvalidHuffmanData();
                }

                usedTable = (start & bitMask) == 0 ? this.left : this.right;

                bitMask <<= 1;
                index = -value;
                overflowBits--;
            }
            while (overflowBits != 0);

            usedTable[index] = code;
        }
    }
}
