﻿using System;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Represents input buffer for inflating data using Huffman coding.
    /// </summary>
    internal class InputBitsBuffer
    {
        private uint bitBuffer;
        private int availableBits;
        private byte[] buffer;
        private int end;
        private int start;

        /// <summary>
        /// Initializes a new instance of the InputBitsBuffer class.
        /// </summary>
        public InputBitsBuffer()
        {
        }

        /// <summary>
        /// Available bits in bit buffer.
        /// </summary>
        public int AvailableBits
        {
            get
            {
                return this.availableBits;
            }
        }

        /// <summary>
        /// Available bytes.
        /// </summary>
        public int AvailableBytes
        {
            get
            {
                return this.end - this.start + (this.availableBits / 8);
            }
        }

        /// <summary>
        /// Is input required.
        /// </summary>
        public bool InputRequired
        {
            get
            {
                return this.start == this.end;
            }
        }

        /// <summary>
        /// Checks available bits in the bit buffer.
        /// </summary>
        /// <param name="count">Count of bits.</param>
        /// <returns>True if available.</returns>
        public bool CheckAvailable(int count)
        {
            if (this.availableBits < count)
            {
                if (this.InputRequired)
                {
                    return false;
                }

                this.bitBuffer |= (uint)this.buffer[this.start++] << (this.availableBits & 0x1f);
                this.availableBits += 8; 

                if (this.availableBits < count)
                {
                    if (this.InputRequired)
                    {
                        return false;
                    }

                    this.bitBuffer |= (uint)this.buffer[this.start++] << (this.availableBits & 0x1f);
                    this.availableBits += 8;
                }
            }

            return true;
        }

        /// <summary>
        /// Gets available bits from buffer.
        /// </summary>
        /// <param name="count">Count of required bits.</param>
        /// <returns>Bits data.</returns>
        public int GetBits(int count)
        {
            if (!this.CheckAvailable(count))
            {
                return -1;
            }

            int bitMask = (int)(this.bitBuffer & GetBitMask(count));
            this.bitBuffer >>= count & 0x1f;
            this.availableBits -= count;

            return bitMask;
        }

        /// <summary>
        /// Read bytes to output buffer.
        /// </summary>
        /// <param name="output">Output buffer.</param>
        /// <param name="offset">Offset.</param>
        /// <param name="length">Length.</param>
        /// <returns>Count of bytes which are read.</returns>
        public int Read(byte[] output, int offset, int length)
        {
            int bytes = 0;
            while (this.availableBits > 0 && length > 0)
            {
                output[offset++] = (byte)this.bitBuffer;

                this.bitBuffer >>= 8;
                this.availableBits -= 8;
                length--;
                bytes++;
            }

            if (length == 0)
            {
                return bytes;
            }

            int rest = this.end - this.start;
            if (length > rest)
            {
                length = rest;
            }

            Array.Copy(this.buffer, this.start, output, offset, length);
            this.start += length;

            return bytes + length;
        }

        /// <summary>
        /// Set current working buffer.
        /// </summary>
        /// <param name="buffer">Bytes buffer.</param>
        /// <param name="offset">Offset.</param>
        /// <param name="length">Length.</param>
        public void SetBuffer(byte[] buffer, int offset, int length)
        {
            this.buffer = buffer;
            this.start = offset;
            this.end = offset + length;
        }

        /// <summary>
        /// Skips bits in bit buffer.
        /// </summary>
        /// <param name="count">Count of bits to skip.</param>
        public void SkipBits(int count)
        {
            this.bitBuffer >>= count & 0x1f;
            this.availableBits -= count;
        }

        /// <summary>
        /// Skips to the next byte boundary.
        /// </summary>
        public void SkipToByteBoundary()
        {
            this.bitBuffer >>= this.availableBits % 8 & 0x1f;
            this.availableBits -= this.availableBits % 8;
        }

        /// <summary>
        /// Gets 16 or more bits into bit buffer.
        /// </summary>
        /// <returns>Bit buffer.</returns>
        public uint Get16Bits()
        {
            if (this.availableBits < 8)
            {
                this.Get8Bits();
                this.Get8Bits();
            }
            else if (this.availableBits < 16)
            {
                this.Get8Bits();
            }

            return this.bitBuffer;
        }

        private static uint GetBitMask(int count)
        {
            return (uint)((1 << (count & 0x1f)) - 1);
        }

        private void Get8Bits()
        {
            if (this.start < this.end)
            {
                this.bitBuffer |= (uint)this.buffer[this.start++] << (this.availableBits & 0x1f);
                this.availableBits += 8;
            }
        }
    }
}
