﻿using System;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Represents output window for inflating data using Huffman coding.
    /// </summary>
    internal class OutputWindow
    {
        private const int WindowMask = 0x7fff;
        private const int WindowSize = 0x8000;

        private int end;
        private int availableBytes;
        private byte[] window = new byte[OutputWindow.WindowSize];

        /// <summary>
        /// Initializes a new instance of the OutputWindow class.
        /// </summary>
        public OutputWindow()
        {
        }

        /// <summary>
        /// Gets available bytes count.
        /// </summary>
        public int AvailableBytes
        {
            get
            {
                return this.availableBytes;
            }
        }

        /// <summary>
        /// Gets free bytes count.
        /// </summary>
        public int FreeBytes
        {
            get
            {
                return OutputWindow.WindowSize - this.availableBytes;
            }
        }

        /// <summary>
        /// Adds a byte to output window.
        /// </summary>
        /// <param name="value">Byte.</param>
        public void AddByte(byte value)
        {
            this.window[this.end++] = value;
            this.end &= OutputWindow.WindowMask;
            this.availableBytes++;
        }

        /// <summary>
        /// Copies bytes within output window.
        /// Moves backwards distance bytes and copy length bytes.
        /// </summary>
        /// <param name="length">Length.</param>
        /// <param name="distance">Distance.</param>
        public void Copy(int length, int distance)
        {
            this.availableBytes += length;
            int start = this.end - distance & OutputWindow.WindowMask;
            int border = OutputWindow.WindowSize - length;
            if (start > border || this.end >= border)
            {
                while (length-- > 0)
                {
                    this.window[this.end++] = this.window[start++];
                    this.end &= OutputWindow.WindowMask;
                    start &= OutputWindow.WindowMask;
                }
            }
            else
            {
                if (length > distance)
                {
                    while (length-- > 0)
                    {
                        this.window[this.end++] = this.window[start++];
                    }
                }
                else
                {
                    Array.Copy(this.window, start, this.window, this.end, length);
                    this.end += length;
                }
            }
        }

        /// <summary>
        /// Read bytes to output buffer.
        /// </summary>
        /// <param name="output">Output buffer.</param>
        /// <param name="offset">Offset.</param>
        /// <param name="length">Length.</param>
        /// <returns>Count of bytes which are read.</returns>
        public int Read(byte[] output, int offset, int length)
        {
            int endPosition;
            if (length <= this.availableBytes)
            {
                endPosition = (this.end - this.availableBytes + length) & OutputWindow.WindowMask;
            }
            else
            {
                endPosition = this.end;
                length = this.availableBytes;
            }

            int read = length;
            int rest = length - endPosition;
            if (rest > 0)
            {
                Array.Copy(this.window, OutputWindow.WindowSize - rest, output, offset, rest);
                offset += rest;
                length = endPosition;
            }

            Array.Copy(this.window, endPosition - length, output, offset, length);
            this.availableBytes -= read;

            return read;
        }

        /// <summary>
        /// Reads bytes from input.
        /// </summary>
        /// <param name="input">InputBitsBuffer.</param>
        /// <param name="length">Length.</param>
        /// <returns>Count of read bytes.</returns>
        public int ReadInput(InputBitsBuffer input, int length)
        {
            int read;
            int rest = OutputWindow.WindowSize - this.end;

            length = Math.Min(Math.Min(length, this.FreeBytes), input.AvailableBytes);
            if (length <= rest)
            {
                read = input.Read(this.window, this.end, length);
            }
            else
            {
                read = input.Read(this.window, this.end, rest);
                if (read == rest)
                {
                    read += input.Read(this.window, 0, length - rest);
                }
            }

            this.end = (this.end + read) & OutputWindow.WindowMask;
            this.availableBytes += read;

            return read;
        }
    }
}
