﻿using System;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    internal class LzmaBinaryTree
    {
        private const uint Hash2Size = 1 << 10;
        private const uint Hash3Size = 1 << 16;
        private const uint BT2HashSize = 1 << 16;
        private const uint StartMaxLen = 1;
        private const uint Hash3Offset = Hash2Size;
        private const uint EmptyHashValue = 0;
        private const uint NormalizeMaxValue = ((uint)1 << 31) - 1;

        private byte[] buffer = null;

        // offset from buffer of first byte when new block reading must be done
        private uint positionLimit;

        // if (true) then _streamPos shows real end of stream
        private bool inputComplete;

        private uint lastSafePosition;
        private uint bufferOffset;

        // current state of matches processing
        private uint[] currentDistances;
        private uint currentBufferIndex;
        private uint currentLengthLimit;
        private uint currentMatch;
        private uint currentMatchMaxLen;
        private uint currentMatchMinPos;
        private uint currentOffset;

        // Size of Allocated memory block
        private uint blockSize;

        // how many BYTEs must be kept in buffer before _pos
        private uint keepSizeBefore;

        // how many BYTEs must be kept buffer after _pos
        private uint keepSizeAfter;

        // offset (from _buffer) of curent byte
        private uint bufferIndex;

        // offset (from _buffer) of first not read byte from Stream
        private uint streamIndex;

        private byte[] inputBuffer;
        private int inputBufferLength;
        private int inputBufferIndex;
        private bool finalBlock;

        private uint cyclicBufferPos;
        private uint cyclicBufferSize = 0;
        private uint matchMaxLength;

        private uint[] son;
        private uint[] hash;

        private uint cutValue = 0xFF;
        private uint hashMask;
        private uint hashSizeSum = 0;

        private bool useHashArray = true;

        private uint hashDirectBytes = 0;
        private uint minMatchCheck = 4;
        private uint fixHashSize = Hash2Size + Hash3Size;
        private int startInputBufferIndex;

        public LzmaBinaryTree(int hashBytes, uint historySize, uint matchMaxLen, uint keepAddBufferAfter)
        {
            this.useHashArray = hashBytes > 2;
            if (this.useHashArray)
            {
                this.hashDirectBytes = 0;
                this.minMatchCheck = 4;
                this.fixHashSize = Hash2Size + Hash3Size;
            }
            else
            {
                this.hashDirectBytes = 2;
                this.minMatchCheck = 2 + 1;
                this.fixHashSize = 0;
            }

            uint keepAddBufferBefore = LzmaCompressor.OptimumsNumber;

            uint windowReservSize = 256
                + ((historySize + keepAddBufferBefore + matchMaxLen + keepAddBufferAfter) / 2);

            this.keepSizeBefore = historySize + keepAddBufferBefore;
            this.keepSizeAfter = matchMaxLen + keepAddBufferAfter;
            uint blockSize = this.keepSizeBefore + this.keepSizeAfter + windowReservSize;

            this.blockSize = blockSize;
            this.buffer = new byte[this.blockSize];

            this.lastSafePosition = this.blockSize - this.keepSizeAfter;

            this.matchMaxLength = matchMaxLen;
            this.cutValue = 16 + (this.matchMaxLength >> 1);

            uint size = historySize + 1;
            if (this.cyclicBufferSize != size)
            {
                this.son = new uint[(this.cyclicBufferSize = size) * 2];
            }

            this.CreateHash(historySize);
            this.Init();
        }

        public uint AvailableBytes
        {
            get
            {
                return this.streamIndex - this.bufferIndex;
            }
        }

        public bool HasInput
        {
            get
            {
                return this.inputBufferIndex - this.startInputBufferIndex < this.inputBufferLength;
            }
        }

        public bool Ready
        {
            get
            {
                if (!this.inputComplete)
                {
                    int size = (int)((0 - this.bufferOffset) + this.blockSize - this.streamIndex);
                    if (size != 0)
                    {
                        return false;
                    }
                }

                return true;
            }
        }

        public bool ReadyToMove
        {
            get
            {
                if (!this.inputComplete)
                {
                    uint position = this.bufferIndex + 1;
                    if (position > this.positionLimit)
                    {
                        uint pointerToPostion = this.bufferOffset + position;
                        if (pointerToPostion > this.lastSafePosition)
                        {
                            return true;
                        }
                    }
                }

                return false;
            }
        }

        public uint GetMatches(uint[] distances)
        {
            this.GetLengthLimit();
            if (this.currentLengthLimit == 0)
            {
                return this.currentLengthLimit;
            }

            this.SetInitialMatchValues(distances);
            this.CalculateHashValues(true);
            this.CheckDirectBytes();

            this.ProcessMatches();

            return this.currentOffset;
        }

        public void Skip(uint counter)
        {
            do
            {
                this.GetLengthLimit();
                if (this.currentLengthLimit == 0)
                {
                    continue;
                }

                this.SetInitialMatchValues();
                this.CalculateHashValues(false);

                this.ProcessMatches(false);
            }
            while (--counter != 0);
        }

        public void MoveBlock()
        {
            uint offset = (uint)this.bufferOffset + this.bufferIndex - this.keepSizeBefore;

            // we need one additional byte, since MovePos moves on 1 byte.
            if (offset > 0)
            {
                offset--;
            }

            int length = (int)((uint)this.bufferOffset + this.streamIndex - offset);
            Array.Copy(this.buffer, (int)offset, this.buffer, 0, length);
            this.bufferOffset -= offset;
        }

        public void SetInputBuffer(byte[] buffer, int offset, int length, bool finalBlock)
        {
            if (this.HasInput)
            {
                this.ReadCicleBlock();
            }

            if (this.HasInput)
            {
            }

            this.inputBuffer = buffer;
            this.inputBufferLength = length;
            this.startInputBufferIndex = this.inputBufferIndex = offset;
            this.finalBlock = finalBlock;

            this.ReadCicleBlock();
        }

        public byte GetIndexByte(int index)
        {
            return this.buffer[this.bufferOffset + this.bufferIndex + index];
        }

        // index + limit have not to exceed keepSizeAfter
        public uint GetMatchLen(int index, uint distance, uint limit)
        {
            if (this.inputComplete)
            {
                if ((this.bufferIndex + index) + limit > this.streamIndex)
                {
                    limit = this.streamIndex - (uint)(this.bufferIndex + index);
                }
            }

            distance++;

            uint pointer = this.bufferOffset + this.bufferIndex + (uint)index;

            uint i;
            for (i = 0; i < limit && this.buffer[pointer + i] == this.buffer[pointer + i - distance]; i++)
            {
            }

            return i;
        }

        private static void NormalizeLinks(uint[] items, uint itemsCounter, uint subValue)
        {
            for (uint i = 0; i < itemsCounter; i++)
            {
                uint value = items[i];
                if (value <= subValue)
                {
                    value = EmptyHashValue;
                }
                else
                {
                    value -= subValue;
                }

                items[i] = value;
            }
        }

        private void Init()
        {
            this.bufferOffset = 0;
            this.bufferIndex = 0;
            this.streamIndex = 0;
            this.inputComplete = false;

            for (uint i = 0; i < this.hashSizeSum; i++)
            {
                this.hash[i] = EmptyHashValue;
            }

            this.cyclicBufferPos = 0;

            this.ReduceOffsets(-1);
        }

        private void CreateHash(uint historySize)
        {
            uint hashSize = BT2HashSize;
            if (this.useHashArray)
            {
                hashSize = historySize - 1;
                hashSize |= hashSize >> 1;
                hashSize |= hashSize >> 2;
                hashSize |= hashSize >> 4;
                hashSize |= hashSize >> 8;
                hashSize >>= 1;
                hashSize |= 0xFFFF;

                if (hashSize > (1 << 24))
                {
                    hashSize >>= 1;
                }

                this.hashMask = hashSize;
                hashSize++;
                hashSize += this.fixHashSize;
            }

            if (hashSize != this.hashSizeSum)
            {
                this.hash = new uint[this.hashSizeSum = hashSize];
            }
        }

        private void Normalize()
        {
            uint subValue = this.bufferIndex - this.cyclicBufferSize;

            NormalizeLinks(this.son, this.cyclicBufferSize * 2, subValue);
            NormalizeLinks(this.hash, this.hashSizeSum, subValue);

            this.ReduceOffsets((int)subValue);
        }

        private void ReduceOffsets(int subValue)
        {
            this.bufferOffset += (uint)subValue;
            this.positionLimit -= (uint)subValue;
            this.bufferIndex -= (uint)subValue;
            this.streamIndex -= (uint)subValue;
        }

        private void ReadBlock()
        {
            if (!this.inputComplete)
            {
                while (true)
                {
                    int size = (int)((0 - this.bufferOffset) + this.blockSize - this.streamIndex);
                    if (size == 0)
                    {
                        this.SetRestInputBuffer();
                        return;
                    }

                    size = Math.Min(size, this.inputBufferLength + this.startInputBufferIndex - this.inputBufferIndex);
                    Array.Copy(this.inputBuffer, this.inputBufferIndex, this.buffer, (int)(this.bufferOffset + this.streamIndex), size);

                    this.inputBufferIndex += size;
                    this.streamIndex += (uint)size;
                    if (this.streamIndex >= this.bufferIndex + this.keepSizeAfter)
                    {
                        this.positionLimit = this.streamIndex - this.keepSizeAfter;
                    }

                    if (!this.HasInput)
                    {
                        if (this.finalBlock)
                        {
                            this.SetInputComplete();
                        }

                        return;
                    }
                }
            }
        }

        private void SetInputComplete()
        {
            this.positionLimit = this.streamIndex;
            uint pointerToPostion = this.bufferOffset + this.positionLimit;
            if (pointerToPostion > this.lastSafePosition)
            {
                this.positionLimit = (uint)(this.lastSafePosition - this.bufferOffset);
            }

            this.inputComplete = true;
        }

        private void SetRestInputBuffer()
        {
            int rest = this.inputBufferLength + this.startInputBufferIndex - this.inputBufferIndex;
            byte[] tempBuffer = new byte[rest];
            Array.Copy(this.inputBuffer, this.inputBufferIndex, tempBuffer, 0, rest);
            this.startInputBufferIndex = this.inputBufferIndex = 0;
            this.inputBufferLength = rest;
            this.inputBuffer = tempBuffer;
        }

        private void ReadCicleBlock()
        {
            this.TryMoveBlock();
            this.ReadBlock();
        }

        private bool TryMoveBlock()
        {
            if (this.bufferIndex > this.positionLimit)
            {
                uint pointerToPostion = this.bufferOffset + this.bufferIndex;
                if (pointerToPostion > this.lastSafePosition)
                {
                    this.MoveBlock();
                    return true;
                }
            }

            return false;
        }

        private void GetLengthLimit()
        {
            if (this.bufferIndex + this.matchMaxLength <= this.streamIndex)
            {
                this.currentLengthLimit = this.matchMaxLength;
            }
            else
            {
                this.currentLengthLimit = this.streamIndex - this.bufferIndex;
                if (this.currentLengthLimit < this.minMatchCheck)
                {
                    this.MovePos();
                    this.currentLengthLimit = 0;
                }
            }
        }

        private void MovePos()
        {
            if (++this.cyclicBufferPos >= this.cyclicBufferSize)
            {
                this.cyclicBufferPos = 0;
            }

            this.bufferIndex++;
            if (this.TryMoveBlock())
            {
                this.ReadBlock();
            }

            if (this.bufferIndex == NormalizeMaxValue)
            {
                this.Normalize();
            }
        }

        private void SetInitialMatchValues(uint[] distances = null)
        {
            this.currentDistances = distances;
            this.currentOffset = 0;
            this.currentMatchMinPos = this.bufferIndex > this.cyclicBufferSize ? this.bufferIndex - this.cyclicBufferSize : 0;
            this.currentBufferIndex = this.bufferOffset + this.bufferIndex;
            this.currentMatchMaxLen = StartMaxLen;
        }

        private void CalculateHashValues(bool matchProcess = true)
        {
            uint current = this.currentBufferIndex;
            uint hashValue, hash2Value = 0, hash3Value = 0;

            if (this.useHashArray)
            {
                uint temp = Crc32.Table[this.buffer[current]] ^ this.buffer[current + 1];
                hash2Value = temp & (Hash2Size - 1);
                temp ^= (uint)this.buffer[current + 2] << 8;

                hash3Value = temp & (Hash3Size - 1);
                hashValue = (temp ^ (Crc32.Table[this.buffer[current + 3]] << 5)) & this.hashMask;
            }
            else
            {
                hashValue = this.buffer[current] ^ ((uint)this.buffer[current + 1] << 8);
                this.currentMatch = this.hash[this.fixHashSize + hashValue];
            }

            this.currentMatch = this.hash[this.fixHashSize + hashValue];

            if (this.useHashArray)
            {
                uint curMatch2 = this.hash[hash2Value];
                this.hash[hash2Value] = this.bufferIndex;
                uint curMatch3 = this.hash[Hash3Offset + hash3Value];
                this.hash[Hash3Offset + hash3Value] = this.bufferIndex;

                if (matchProcess)
                {
                    this.UpdateMatches(curMatch2, curMatch3);
                }
            }
            else
            {
                hashValue = this.buffer[current] ^ ((uint)this.buffer[current + 1] << 8);
                this.currentMatch = this.hash[this.fixHashSize + hashValue];
            }

            this.hash[this.fixHashSize + hashValue] = this.bufferIndex;
        }

        private void UpdateMatches(uint curMatch2, uint curMatch3)
        {
            if (this.useHashArray)
            {
                uint[] distances = this.currentDistances;
                uint matchMinPos = this.currentMatchMinPos;
                uint offset = this.currentOffset;
                uint current = this.currentBufferIndex;

                if (curMatch2 > matchMinPos && this.buffer[this.bufferOffset + curMatch2] == this.buffer[current])
                {
                    distances[offset++] = this.currentMatchMaxLen = 2;
                    distances[offset++] = this.bufferIndex - curMatch2 - 1;
                }

                if (curMatch3 > matchMinPos && this.buffer[this.bufferOffset + curMatch3] == this.buffer[current])
                {
                    if (curMatch3 == curMatch2)
                    {
                        offset -= 2;
                    }

                    distances[offset++] = this.currentMatchMaxLen = 3;
                    distances[offset++] = this.bufferIndex - curMatch3 - 1;
                    curMatch2 = curMatch3;
                }

                if (offset != 0 && curMatch2 == this.currentMatch)
                {
                    offset -= 2;
                    this.currentMatchMaxLen = StartMaxLen;
                }

                this.currentOffset = offset;
            }
        }

        private void CheckDirectBytes()
        {
            if (this.hashDirectBytes != 0
                && this.currentMatch > this.currentMatchMinPos
                && this.buffer[this.bufferOffset + this.currentMatch + this.hashDirectBytes]
                    != this.buffer[this.currentBufferIndex + this.hashDirectBytes])
            {
                this.currentDistances[this.currentOffset++] = this.currentMatchMaxLen = this.hashDirectBytes;
                this.currentDistances[this.currentOffset++] = this.bufferIndex - this.currentMatch - 1;
            }
        }

        private void ProcessMatches(bool matchProcess = true)
        {
            uint pointer1 = this.cyclicBufferPos << 1;
            uint pointer0 = pointer1 + 1;
            uint current = this.currentBufferIndex;
            uint matchMinPos = this.currentMatchMinPos;
            uint count = this.cutValue;
            uint length0, length1;
            length0 = length1 = this.hashDirectBytes;

            while (true)
            {
                uint match = this.currentMatch;
                if (match <= matchMinPos || count-- == 0)
                {
                    this.son[pointer0] = this.son[pointer1] = EmptyHashValue;
                    break;
                }

                uint delta = this.bufferIndex - match;
                uint cyclicPosition = ((delta <= this.cyclicBufferPos) ?
                            (this.cyclicBufferPos - delta) :
                            (this.cyclicBufferPos - delta + this.cyclicBufferSize)) << 1;

                uint pointerByte1 = this.bufferOffset + match;
                uint length = Math.Min(length0, length1);

                if (this.buffer[pointerByte1 + length] == this.buffer[current + length])
                {
                    length = this.SkipToLimit(current, pointerByte1, length);
                    if (this.CheckMatchLength(matchProcess, cyclicPosition, length, pointer0, pointer1))
                    {
                        break;
                    }
                }

                if (this.buffer[pointerByte1 + length] < this.buffer[current + length])
                {
                    this.son[pointer1] = match;
                    pointer1 = cyclicPosition + 1;
                    this.currentMatch = this.son[pointer1];
                    length1 = length;
                }
                else
                {
                    this.son[pointer0] = match;
                    pointer0 = cyclicPosition;
                    this.currentMatch = this.son[pointer0];
                    length0 = length;
                }
            }

            this.MovePos();
        }

        private uint SkipToLimit(uint current, uint pointerByte1, uint length)
        {
            while (++length != this.currentLengthLimit)
            {
                if (this.buffer[pointerByte1 + length] != this.buffer[current + length])
                {
                    break;
                }
            }

            return length;
        }

        private bool CheckMatchLength(bool matchProcess, uint cyclicPosition, uint length, uint pointer0, uint pointer1)
        {
            bool updateLength = this.currentMatchMaxLen < length;
            if (matchProcess && updateLength)
            {
                uint[] distances = this.currentDistances;
                uint offset = this.currentOffset;
                uint delta = this.bufferIndex - this.currentMatch;

                distances[offset++] = this.currentMatchMaxLen = length;
                distances[offset++] = delta - 1;
                this.currentOffset = offset;
            }

            if ((!matchProcess || updateLength) && length == this.currentLengthLimit)
            {
                this.son[pointer1] = this.son[cyclicPosition];
                this.son[pointer0] = this.son[cyclicPosition + 1];
                return true;
            }

            return false;
        }
    }
}
