﻿using System;
using System.IO;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Represents the LZMA range encoder.
    /// </summary>
    internal class LzmaRangeEncoder : IDisposable
    {
        public const uint TopValue = 1 << 24;

        private byte cache;
        private uint cacheSize;
        private byte[] outputBuffer;
        private int nextOut;
        private MemoryStream pendingData;

        public LzmaRangeEncoder()
        {
            this.pendingData = new MemoryStream();
            this.Low = 0;
            this.Range = 0xFFFFFFFF;
            this.cacheSize = 1;
            this.cache = 0;
        }

        public ulong Low
        {
            get;
            set;
        }

        public uint Range
        {
            get;
            set;
        }

        public bool HasData
        {
            get
            {
                return this.pendingData.Length > 0;
            }
        }

        public int NextOut
        {
            get
            {
                return this.nextOut;
            }
        }

        private bool CanWrite
        {
            get
            {
                return this.NextOut < this.outputBuffer.Length;
            }
        }

        public void SetOutputBuffer(byte[] outputBuffer, int index)
        {
            this.outputBuffer = outputBuffer;
            this.nextOut = index;
            this.Flush();
        }

        public void FlushData()
        {
            for (int i = 0; i < 5; i++)
            {
                this.ShiftLow();
            }
        }

        public void Dispose()
        {
            if (this.pendingData != null)
            {
                this.pendingData.Dispose();
            }
        }

        public void ShiftLow()
        {
            if ((uint)this.Low < (uint)0xFF000000 || (uint)(this.Low >> 32) == 1)
            {
                byte temp = this.cache;
                do
                {
                    byte outputByte = (byte)(temp + (this.Low >> 32));
                    if (this.CanWrite)
                    {
                        this.outputBuffer[this.nextOut++] = outputByte;
                    }
                    else
                    {
                        this.pendingData.WriteByte(outputByte);
                    }

                    temp = 0xFF;
                }
                while (--this.cacheSize != 0);

                this.cache = (byte)(((uint)this.Low) >> 24);
            }

            this.cacheSize++;
            this.Low = ((uint)this.Low) << 8;
        }

        public void EncodeDirectBits(uint value, int totalBits)
        {
            for (int i = totalBits - 1; i >= 0; i--)
            {
                this.Range >>= 1;
                if (((value >> i) & 1) == 1)
                {
                    this.Low += this.Range;
                }

                if (this.Range < TopValue)
                {
                    this.Range <<= 8;
                    this.ShiftLow();
                }
            }
        }

        private void Flush()
        {
            if (this.pendingData.Length > 0)
            {
                if (this.pendingData.Position == this.pendingData.Length)
                {
                    this.pendingData.Position = 0;
                }

                int read = this.pendingData.Read(this.outputBuffer, this.NextOut, this.outputBuffer.Length - this.NextOut);
                this.nextOut += read;
                if (this.pendingData.Position == this.pendingData.Length)
                {
                    this.pendingData.SetLength(0);
                }
            }
        }
    }
}
