﻿using System;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Compressor which implements LZMA compression.
    /// </summary>
    internal class LzmaCompressor : LzmaTransformBase
    {
        public const uint OptimumsNumber = 1 << 12;

        private const int MinBlockSize = 8192;
        private const int MaxBlockSize = 262144;
        private const int DefaultDictionaryLogSize = 22;
        private const int DictionaryMaxCompressValue = 27;
        private const uint FastBytesDefaultNumber = 0x20;
        private const uint LengthSpecialSymbolsNumber =
            LzmaTransformBase.LowLengthSymbolsNumber + LzmaTransformBase.MiddleLengthSymbolsNumber;
        private const uint InfinityPrice = 0xFFFFFFF;
        private const byte FastSlotsNumber = 22;

        private static byte[] fastPositions = new byte[1 << 11];

        private LzmaState compressorState = new LzmaState();
        private bool finished;

        private byte currentPreviousByte;
        private uint[] repeaterDistances = new uint[LzmaTransformBase.RepeaterDistancesNumber];

        private LzmaOptimizationData[] optimizationData = new LzmaOptimizationData[OptimumsNumber];
        private LzmaBinaryTree matchFinder;
        private LzmaRangeEncoder rangeEncoder = new LzmaRangeEncoder();

        private LzmaRangeBitEncoder[] matchRangeBitEncoders =
            new LzmaRangeBitEncoder[LzmaTransformBase.StatesNumber << LzmaTransformBase.PosStatesBitsMax];
        private LzmaRangeBitEncoder[] repeaterRangeBitEncoders =
            new LzmaRangeBitEncoder[LzmaTransformBase.StatesNumber];
        private LzmaRangeBitEncoder[] repeaterG0RangeBitEncoders =
            new LzmaRangeBitEncoder[LzmaTransformBase.StatesNumber];
        private LzmaRangeBitEncoder[] repeaterG1RangeBitEncoders =
            new LzmaRangeBitEncoder[LzmaTransformBase.StatesNumber];
        private LzmaRangeBitEncoder[] repeaterG2RangeBitEncoders =
            new LzmaRangeBitEncoder[LzmaTransformBase.StatesNumber];
        private LzmaRangeBitEncoder[] repeaterLongRangeBitEncoders =
            new LzmaRangeBitEncoder[LzmaTransformBase.StatesNumber << LzmaTransformBase.PosStatesBitsMax];

        private LzmaBitTreeEncoder[] positionSlotEncoder = new LzmaBitTreeEncoder[LzmaTransformBase.LengthToPosStates];

        private LzmaRangeBitEncoder[] positionEncoders =
            new LzmaRangeBitEncoder[LzmaTransformBase.FullDistancesNumber - LzmaTransformBase.EndPosModelIndex];
        private LzmaBitTreeEncoder positionAlignEncoder = new LzmaBitTreeEncoder(LzmaTransformBase.AlignBitsNumber);

        private LzmaLengthTableEncoder lengthEncoder = new LzmaLengthTableEncoder();
        private LzmaLengthTableEncoder repeaterMatchLengthEncoder = new LzmaLengthTableEncoder();

        private LzmaLiteralEncoder literalEncoder;

        private uint[] matchDistances = new uint[(LzmaTransformBase.MatchMaxLen * 2) + 2];

        private uint configFastBytes = FastBytesDefaultNumber;
        private uint longestMatchLength;
        private uint distancePairsNumber;

        private uint additionalOffset;

        private uint optimumEndIndex;
        private uint optimumCurrentIndex;
        private bool longestMatchWasFound;

        private uint[] positionSlotPrices =
            new uint[1 << (LzmaTransformBase.PosSlotBitsNumber + LzmaTransformBase.LengthToPosStatesBits)];
        private uint[] distancesPrices =
            new uint[LzmaTransformBase.FullDistancesNumber << LzmaTransformBase.LengthToPosStatesBits];
        private uint[] alignPrices =
            new uint[LzmaTransformBase.AlignTableSize];
        private uint alignPriceCount;

        private uint configDistanceTableSize;
        private int configPositionStateBits;
        private uint configPositionStateMask;
        private int configLiteralPositionStateBits;
        private int configLiteralContextBits;
        private uint configDictionarySize;

        private long currentPosition;

        private LzmaMatchFinderType matchFinderType = LzmaMatchFinderType.BT4;

        private uint[] repeaters = new uint[LzmaTransformBase.RepeaterDistancesNumber];
        private uint[] repeaterLengths = new uint[LzmaTransformBase.RepeaterDistancesNumber];

        private uint[] tempPrices = new uint[LzmaTransformBase.FullDistancesNumber];

        private LzmaState currentState;
        private uint matchPriceCount;
        private uint optimumPosition;
        private uint optimumLength;
        private uint optimumLengthEnd;
        private uint optimumNewLength;
        private uint repeaterMaxIndex;
        private uint distancePairsCounter;
        private uint optimumCurrent;
        private uint optimumAvailableBytesFull;
        private uint optimumPositionState;

        static LzmaCompressor()
        {
            int counter = 2;
            fastPositions[0] = 0;
            fastPositions[1] = 1;
            for (byte slotFast = 2; slotFast < FastSlotsNumber; slotFast++)
            {
                uint limit = (uint)1 << ((slotFast >> 1) - 1);

                for (uint i = 0; i < limit; i++, counter++)
                {
                    fastPositions[counter] = slotFast;
                }
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="LzmaCompressor"/> class.
        /// </summary>
        /// <param name="settings">Settings.</param>
        public LzmaCompressor(LzmaSettings settings)
            : base(settings)
        {
            for (int i = 0; i < OptimumsNumber; i++)
            {
                this.optimizationData[i] = new LzmaOptimizationData();
            }

            for (int i = 0; i < LzmaTransformBase.LengthToPosStates; i++)
            {
                this.positionSlotEncoder[i] = new LzmaBitTreeEncoder(LzmaTransformBase.PosSlotBitsNumber);
            }

            this.Initialize();
        }

        /// <summary>Gets the output block size.</summary>
        /// <returns>The size of the output data blocks in bytes.</returns>
        public override int OutputBlockSize
        {
            get
            {
                int size = (int)this.configDictionarySize;
                if (size < MinBlockSize)
                {
                    return MinBlockSize;
                }
                else if (size > MaxBlockSize)
                {
                    return MaxBlockSize;
                }

                return size;
            }
        }

        /// <summary>
        /// Creates transformation header to be written into the output stream.
        /// </summary>
        public override void CreateHeader()
        {
            byte[] byteHeader;
            if (this.Settings.UseZipHeader)
            {
                byteHeader = new byte[]
                {
                    // version signature
                    0x09, 0x14, 0x05, 0x00,
                    0x00, 0x00, 0x00, 0x00, 0x00
                };

                this.CreatePropertiesHeader(byteHeader, 4);
                this.Header.CountHeaderInCompressedSize = true;
            }
            else
            {
                byteHeader = new byte[]
                {
                    0x00, 0x00, 0x00, 0x00, 0x00,

                    // size = -1
                    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
                };

                if (this.Settings.StreamLength > -1)
                {
                    byte[] size = BitConverter.GetBytes(this.Settings.StreamLength);
                    Array.Copy(size, 0, byteHeader, 5, size.Length);
                }

                this.CreatePropertiesHeader(byteHeader, 0);
            }

            this.Header.Buffer = byteHeader;
        }

        /// <summary>
        /// Transforms current input buffer.
        /// </summary>
        /// <param name="finalBlock">The final block flag.</param>
        /// <returns>True when output still available.</returns>
        protected override bool ProcessTransform(bool finalBlock)
        {
            this.rangeEncoder.SetOutputBuffer(this.OutputBuffer, this.NextOut);

            if (!this.finished)
            {
                this.matchFinder.SetInputBuffer(this.InputBuffer, this.NextIn, this.AvailableBytesIn, finalBlock);
                this.CodeOneBlock(finalBlock);
            }

            this.NextOut = this.rangeEncoder.NextOut;

            return this.rangeEncoder.HasData || this.matchFinder.HasInput;
        }

        /// <summary>
        /// Releases the unmanaged resources used by the transform and optionally releases the managed resources. 
        /// </summary>
        /// <param name="disposing">Value which indicates whether both managed and unmanaged resources
        /// (true) on only unmanaged resources (false) should be released.</param>
        protected override void Dispose(bool disposing)
        {
            this.rangeEncoder.Dispose();

            base.Dispose(disposing);
        }

        private static uint GetPosSlot(uint index)
        {
            if (index < (1 << 11))
            {
                return fastPositions[index];
            }

            if (index < (1 << 21))
            {
                return (uint)(fastPositions[index >> 10] + 20);
            }

            return (uint)(fastPositions[index >> 20] + 40);
        }

        private static uint GetPosSlot2(uint index)
        {
            if (index < (1 << 17))
            {
                return (uint)(fastPositions[index >> 6] + 12);
            }

            if (index < (1 << 27))
            {
                return (uint)(fastPositions[index >> 16] + 32);
            }

            return (uint)(fastPositions[index >> 26] + 52);
        }

        private static void CheckPrice(LzmaOptimizationData optimum, uint curAndLenPrice, uint current, uint backPrev)
        {
            if (curAndLenPrice < optimum.Price)
            {
                optimum.Price = curAndLenPrice;
                optimum.PosPrev = current;
                optimum.BackPrev = backPrev;
                optimum.Prev1IsChar = false;
            }
        }

        private void InitRepeaterDistances()
        {
            this.currentPreviousByte = 0;
            for (uint i = 0; i < LzmaTransformBase.RepeaterDistancesNumber; i++)
            {
                this.repeaterDistances[i] = 0;
            }
        }

        private void CreatePropertiesHeader(byte[] byteHeader, int propertiesIndex)
        {
            byteHeader[propertiesIndex] =
                (byte)((((this.configPositionStateBits * 5) + this.configLiteralPositionStateBits) * 9) + this.configLiteralContextBits);
            propertiesIndex++;
            for (int i = 0; i < 4; i++)
            {
                byteHeader[propertiesIndex + i] = (byte)((this.configDictionarySize >> (8 * i)) & 0xFF);
            }
        }

        private void InitEncoders()
        {
            this.InitRepeaterDistances();

            uint i;
            for (i = 0; i < LzmaTransformBase.StatesNumber; i++)
            {
                for (uint j = 0; j <= this.configPositionStateMask; j++)
                {
                    uint complexState = (i << LzmaTransformBase.PosStatesBitsMax) + j;
                    this.matchRangeBitEncoders[complexState].Init();
                    this.repeaterLongRangeBitEncoders[complexState].Init();
                }

                this.repeaterRangeBitEncoders[i].Init();
                this.repeaterG0RangeBitEncoders[i].Init();
                this.repeaterG1RangeBitEncoders[i].Init();
                this.repeaterG2RangeBitEncoders[i].Init();
            }

            for (i = 0; i < LzmaTransformBase.LengthToPosStates; i++)
            {
                this.positionSlotEncoder[i].Init();
            }

            for (i = 0; i < LzmaTransformBase.FullDistancesNumber - LzmaTransformBase.EndPosModelIndex; i++)
            {
                this.positionEncoders[i].Init();
            }

            this.lengthEncoder.Init((uint)1 << this.configPositionStateBits);
            this.repeaterMatchLengthEncoder.Init((uint)1 << this.configPositionStateBits);

            this.positionAlignEncoder.Init();

            this.longestMatchWasFound = false;
            this.optimumEndIndex = 0;
            this.optimumCurrentIndex = 0;
            this.additionalOffset = 0;
        }

        private void ReadMatchDistances(out uint length)
        {
            length = 0;
            this.distancePairsCounter = this.matchFinder.GetMatches(this.matchDistances);
            uint counter = this.distancePairsCounter;
            if (counter > 0)
            {
                length = this.matchDistances[counter - 2];
                if (length == this.configFastBytes)
                {
                    length += matchFinder.GetMatchLen((int)length - 1,
                        this.matchDistances[counter - 1],
                        LzmaTransformBase.MatchMaxLen - length);
                }
            }

            this.additionalOffset++;
        }

        private void MovePosition(uint counter)
        {
            if (counter > 0)
            {
                this.matchFinder.Skip(counter);
                this.additionalOffset += counter;
            }
        }

        private uint GetRepeaterLen1Price(LzmaState state, uint positionState)
        {
            uint position = (state.Index << LzmaTransformBase.PosStatesBitsMax)
                + positionState;

            return this.repeaterG0RangeBitEncoders[state.Index].GetPrice0()
                + this.repeaterLongRangeBitEncoders[position].GetPrice0();
        }

        private uint GetPureRepeaterPrice(uint index, LzmaState state, uint positionState)
        {
            uint price;
            if (index == 0)
            {
                price = this.repeaterG0RangeBitEncoders[state.Index].GetPrice0();
                uint position = (state.Index << LzmaTransformBase.PosStatesBitsMax)
                    + positionState;
                price += this.repeaterLongRangeBitEncoders[position].GetPrice1();
            }
            else
            {
                price = this.repeaterG0RangeBitEncoders[state.Index].GetPrice1();
                if (index == 1)
                {
                    price += this.repeaterG1RangeBitEncoders[state.Index].GetPrice0();
                }
                else
                {
                    price += this.repeaterG1RangeBitEncoders[state.Index].GetPrice1();
                    price += this.repeaterG2RangeBitEncoders[state.Index].GetPrice(index - 2);
                }
            }

            return price;
        }

        private uint GetRepeaterPrice(uint index, uint length, LzmaState state, uint positionState)
        {
            uint price = this.repeaterMatchLengthEncoder.GetPrice(length - LzmaTransformBase.MatchMinLength, positionState);

            return price + this.GetPureRepeaterPrice(index, state, positionState);
        }

        private uint GetPosLenPrice(uint position, uint length, uint positionState)
        {
            uint price;
            uint lenToPosState = LzmaState.GetLenToPosState(length);
            if (position < LzmaTransformBase.FullDistancesNumber)
            {
                price = this.distancesPrices[(lenToPosState * LzmaTransformBase.FullDistancesNumber) + position];
            }
            else
            {
                uint index = (lenToPosState << LzmaTransformBase.PosSlotBitsNumber) + GetPosSlot2(position);
                price = positionSlotPrices[index] + alignPrices[position & LzmaTransformBase.AlignMask];
            }

            return price + this.lengthEncoder.GetPrice(length - LzmaTransformBase.MatchMinLength, positionState);
        }

        private void Backward(uint current)
        {
            this.optimumEndIndex = current;
            uint posMem = this.optimizationData[current].PosPrev;
            uint backMem = this.optimizationData[current].BackPrev;
            do
            {
                if (this.optimizationData[current].Prev1IsChar)
                {
                    this.optimizationData[posMem].MakeAsChar();
                    this.optimizationData[posMem].PosPrev = posMem - 1;
                    if (this.optimizationData[current].Prev2)
                    {
                        this.optimizationData[posMem - 1].Prev1IsChar = false;
                        this.optimizationData[posMem - 1].PosPrev = this.optimizationData[current].PosPrev2;
                        this.optimizationData[posMem - 1].BackPrev = this.optimizationData[current].BackPrev2;
                    }
                }

                uint posPrev = posMem;
                uint backCur = backMem;

                backMem = this.optimizationData[posPrev].BackPrev;
                posMem = this.optimizationData[posPrev].PosPrev;

                this.optimizationData[posPrev].BackPrev = backCur;
                this.optimizationData[posPrev].PosPrev = current;
                current = posPrev;
            }
            while (current > 0);

            this.optimumPosition = this.optimizationData[0].BackPrev;
            this.optimumLength = this.optimumCurrentIndex = this.optimizationData[0].PosPrev;
        }

        private void GetOptimum(uint position)
        {
            this.optimumPosition = position;
            this.optimumPositionState = position & this.configPositionStateMask;

            bool shouldReturn = this.CheckOptimumIndex();
            if (shouldReturn)
            {
                return;
            }

            this.optimumCurrent = 0;
            while (true)
            {
                this.optimumCurrent++;
                this.optimumPosition++;
                this.optimumPositionState = this.optimumPosition & this.configPositionStateMask;

                shouldReturn = this.CheckFastBytes();
                if (shouldReturn)
                {
                    break;
                }

                this.GetCurrentState();
                this.ProcessOptimum();
            }
        }

        private void ProcessOptimum()
        {
            uint current = this.optimumCurrent;
            uint position = this.optimumPosition;
            uint curPrice = this.optimizationData[current].Price;
            byte currentByte = this.matchFinder.GetIndexByte(-1);
            byte matchByte = this.matchFinder.GetIndexByte((int)(-repeaters[0] - 2));
            uint positionState = this.optimumPositionState;
            uint index = (this.currentState.Index << LzmaTransformBase.PosStatesBitsMax) + positionState;
            var encoder = this.literalEncoder.GetSubCoder(position, this.matchFinder.GetIndexByte(0 - 2));
            LzmaOptimizationData nextOptimum = this.optimizationData[current + 1];
            uint curAnd1Price = curPrice +
                this.matchRangeBitEncoders[index].GetPrice0() +
                encoder.GetPrice(!this.currentState.IsCharState(), matchByte, currentByte);
            bool nextIsChar = false;
            if (curAnd1Price < nextOptimum.Price)
            {
                nextOptimum.Price = curAnd1Price;
                nextOptimum.PosPrev = current;
                nextOptimum.MakeAsChar();
                nextIsChar = true;
            }

            index = (this.currentState.Index << LzmaTransformBase.PosStatesBitsMax) + positionState;
            uint matchPrice = curPrice + this.matchRangeBitEncoders[index].GetPrice1();
            uint repMatchPrice = matchPrice + this.repeaterRangeBitEncoders[this.currentState.Index].GetPrice1();

            if (matchByte == currentByte &&
                !(nextOptimum.PosPrev < current && nextOptimum.BackPrev == 0))
            {
                uint shortRepPrice = repMatchPrice + this.GetRepeaterLen1Price(this.currentState, positionState);
                if (shortRepPrice <= nextOptimum.Price)
                {
                    nextOptimum.Price = shortRepPrice;
                    nextOptimum.PosPrev = current;
                    nextOptimum.MakeAsShortRep();
                    nextIsChar = true;
                }
            }

            uint availableBytes = this.GetAvailableBytes(current);
            if (availableBytes < 2)
            {
                return;
            }

            if (!nextIsChar && matchByte != currentByte)
            {
                this.TryLiteralRep0(curAnd1Price);
            }

            this.ProcessOptimumDistances(repMatchPrice, availableBytes, matchPrice);
        }

        private uint GetAvailableBytes(uint current)
        {
            this.optimumAvailableBytesFull = this.matchFinder.AvailableBytes + 1;
            this.optimumAvailableBytesFull = Math.Min(OptimumsNumber - 1 - current, this.optimumAvailableBytesFull);
            uint availableBytes = this.optimumAvailableBytesFull;
            if (availableBytes >= 2 && availableBytes > this.configFastBytes)
            {
                availableBytes = this.configFastBytes;
            }

            return availableBytes;
        }

        private void UpdatePositionLengthPrice(uint startLen, uint matchPrice)
        {
            uint current = this.optimumCurrent;
            uint normalMatchPrice = matchPrice + this.repeaterRangeBitEncoders[this.currentState.Index].GetPrice0();

            while (this.optimumLengthEnd < current + this.optimumNewLength)
            {
                this.optimizationData[++this.optimumLengthEnd].Price = InfinityPrice;
            }

            this.UpdatePositionLengthPrice(startLen, current, normalMatchPrice);
        }

        private void UpdatePositionLengthPrice(uint startLen, uint current, uint normalMatchPrice)
        {
            uint positionState = this.optimumPositionState;
            uint offset = 0;
            while (startLen > this.matchDistances[offset])
            {
                offset += 2;
            }

            for (uint lenTest = startLen;; lenTest++)
            {
                uint curBack = this.matchDistances[offset + 1];
                uint curAndLenPrice = normalMatchPrice + this.GetPosLenPrice(curBack, lenTest, positionState);
                CheckPrice(this.optimizationData[current + lenTest], curAndLenPrice, current, curBack + LzmaTransformBase.RepeaterDistancesNumber);
                if (lenTest == this.matchDistances[offset])
                {
                    this.CheckAvailableLength(lenTest, curBack, curAndLenPrice, curBack + LzmaTransformBase.RepeaterDistancesNumber, false);
                    offset += 2;
                    if (offset == this.distancePairsCounter)
                    {
                        break;
                    }
                }
            }
        }

        private void ProcessOptimumDistances(uint repMatchPrice, uint availableBytes, uint matchPrice)
        {
            // speed optimization
            uint startLen = 2;
            uint current = this.optimumCurrent;
            uint positionState = this.optimumPositionState;
            for (uint repIndex = 0; repIndex < LzmaTransformBase.RepeaterDistancesNumber; repIndex++)
            {
                uint lenTest = this.matchFinder.GetMatchLen(0 - 1, this.repeaters[repIndex], availableBytes);
                if (lenTest < 2)
                {
                    continue;
                }

                uint lenTestTemp = lenTest;
                do
                {
                    while (this.optimumLengthEnd < current + lenTest)
                    {
                        this.optimizationData[++this.optimumLengthEnd].Price = InfinityPrice;
                    }

                    uint curAndLenPrice = repMatchPrice
                        + this.GetRepeaterPrice(repIndex, lenTest, this.currentState, positionState);
                    CheckPrice(this.optimizationData[current + lenTest], curAndLenPrice, current, repIndex);
                }
                while (--lenTest >= 2);

                lenTest = lenTestTemp;

                if (repIndex == 0)
                {
                    startLen = lenTest + 1;
                }

                this.CheckAvailableLength(lenTest, repeaters[repIndex], repMatchPrice, repIndex);
            }

            if (this.optimumNewLength > availableBytes)
            {
                this.optimumNewLength = availableBytes;
                for (this.distancePairsCounter = 0;
                    this.optimumNewLength > this.matchDistances[this.distancePairsCounter];
                    this.distancePairsCounter += 2)
                {
                }

                this.matchDistances[this.distancePairsCounter] = this.optimumNewLength;
                this.distancePairsCounter += 2;
            }

            if (this.optimumNewLength >= startLen)
            {
                this.UpdatePositionLengthPrice(startLen, matchPrice);
            }
        }

        private void TryLiteralRep0(uint curAnd1Price)
        {
            uint current = this.optimumCurrent + 1;

            // try Literal + rep0
            uint limit = Math.Min(this.optimumAvailableBytesFull - 1, this.configFastBytes);
            uint lenTest2 = this.matchFinder.GetMatchLen(0, this.repeaters[0], limit);
            if (lenTest2 >= 2)
            {
                LzmaState state2 = this.currentState;
                state2.UpdateChar();
                uint posStateNext = (this.optimumPosition + 1) & this.configPositionStateMask;
                uint nextRepMatchPrice = curAnd1Price +
                    this.matchRangeBitEncoders[(state2.Index << LzmaTransformBase.PosStatesBitsMax) + posStateNext].GetPrice1() +
                    this.repeaterRangeBitEncoders[state2.Index].GetPrice1();
                uint offset = current + lenTest2;
                while (this.optimumLengthEnd < offset)
                {
                    this.optimizationData[++this.optimumLengthEnd].Price = InfinityPrice;
                }

                uint curAndLenPrice = nextRepMatchPrice
                    + this.GetRepeaterPrice(0, lenTest2, state2, posStateNext);
                LzmaOptimizationData optimum = this.optimizationData[offset];
                if (curAndLenPrice < optimum.Price)
                {
                    optimum.Price = curAndLenPrice;
                    optimum.PosPrev = current;
                    optimum.BackPrev = 0;
                    optimum.Prev1IsChar = true;
                    optimum.Prev2 = false;
                }
            }
        }

        private void CheckAvailableLength(uint lenTest, uint curBack, uint startPrice, uint repIndex, bool repeater = true)
        {
            if (lenTest < this.optimumAvailableBytesFull)
            {
                uint position = this.optimumPosition;
                uint current = this.optimumCurrent;
                uint positionState = this.optimumPositionState;
                uint limit = Math.Min(this.optimumAvailableBytesFull - 1 - lenTest, this.configFastBytes);
                uint lenTest2 = this.matchFinder.GetMatchLen((int)lenTest, curBack, limit);
                if (lenTest2 >= 2)
                {
                    uint curAndLenCharPrice = startPrice;
                    LzmaState state = this.currentState;
                    uint index;
                    uint posStateNext = (position + lenTest) & this.configPositionStateMask;
                    if (repeater)
                    {
                        state.UpdateRep();
                        index = (state.Index << LzmaTransformBase.PosStatesBitsMax) + posStateNext;
                        curAndLenCharPrice +=
                            this.GetRepeaterPrice(repIndex, lenTest, this.currentState, positionState) +
                            this.matchRangeBitEncoders[index].GetPrice0() +
                            this.literalEncoder.GetSubCoder(position + lenTest,
                            this.matchFinder.GetIndexByte((int)lenTest - 2)).GetPrice(true,
                            this.matchFinder.GetIndexByte((int)((int)lenTest - 1 - (int)(curBack + 1))),
                            this.matchFinder.GetIndexByte((int)lenTest - 1));
                    }
                    else
                    {
                        state.UpdateMatch();
                        index = (state.Index << LzmaTransformBase.PosStatesBitsMax) + posStateNext;
                        curAndLenCharPrice +=
                            this.matchRangeBitEncoders[index].GetPrice0() +
                            this.literalEncoder.GetSubCoder(position + lenTest,
                            this.matchFinder.GetIndexByte((int)lenTest - 2)).GetPrice(true,
                            this.matchFinder.GetIndexByte((int)lenTest - (int)(curBack + 1) - 1),
                            this.matchFinder.GetIndexByte((int)lenTest - 1));
                    }

                    state.UpdateChar();
                    posStateNext = (position + lenTest + 1) & this.configPositionStateMask;
                    index = (state.Index << LzmaTransformBase.PosStatesBitsMax) + posStateNext;
                    uint nextMatchPrice = curAndLenCharPrice + this.matchRangeBitEncoders[index].GetPrice1();
                    uint nextRepMatchPrice = nextMatchPrice + this.repeaterRangeBitEncoders[state.Index].GetPrice1();

                    uint offset = lenTest + 1 + lenTest2;
                    while (this.optimumLengthEnd < current + offset)
                    {
                        this.optimizationData[++this.optimumLengthEnd].Price = InfinityPrice;
                    }

                    uint curAndLenPrice = nextRepMatchPrice + this.GetRepeaterPrice(0, lenTest2, state, posStateNext);
                    LzmaOptimizationData optimum = this.optimizationData[current + offset];
                    if (curAndLenPrice < optimum.Price)
                    {
                        optimum.Price = curAndLenPrice;
                        optimum.PosPrev = current + lenTest + 1;
                        optimum.BackPrev = 0;
                        optimum.Prev1IsChar = true;
                        optimum.Prev2 = true;
                        optimum.PosPrev2 = current;
                        optimum.BackPrev2 = repIndex;
                    }
                }
            }
        }

        private LzmaState GetCurrentState()
        {
            uint current = this.optimumCurrent;
            uint posPrev = this.optimizationData[current].PosPrev;
            if (this.optimizationData[current].Prev1IsChar)
            {
                posPrev--;
                if (this.optimizationData[current].Prev2)
                {
                    this.currentState = this.optimizationData[this.optimizationData[current].PosPrev2].State;
                    if (this.optimizationData[current].BackPrev2 < LzmaTransformBase.RepeaterDistancesNumber)
                    {
                        this.currentState.UpdateRep();
                    }
                    else
                    {
                        this.currentState.UpdateMatch();
                    }
                }
                else
                {
                    this.currentState = optimizationData[posPrev].State;
                }

                this.currentState.UpdateChar();
            }
            else
            {
                this.currentState = optimizationData[posPrev].State;
            }

            this.UpdateCurrentState(posPrev);

            return currentState;
        }

        private void UpdateCurrentState(uint previousPosition)
        {
            uint current = this.optimumCurrent;
            if (previousPosition == current - 1)
            {
                if (this.optimizationData[current].IsShortRep())
                {
                    this.currentState.UpdateShortRep();
                }
                else
                {
                    this.currentState.UpdateChar();
                }
            }
            else
            {
                uint position;
                if (this.optimizationData[current].Prev1IsChar && this.optimizationData[current].Prev2)
                {
                    previousPosition = this.optimizationData[current].PosPrev2;
                    position = this.optimizationData[current].BackPrev2;
                    this.currentState.UpdateRep();
                }
                else
                {
                    position = this.optimizationData[current].BackPrev;
                    if (position < LzmaTransformBase.RepeaterDistancesNumber)
                    {
                        this.currentState.UpdateRep();
                    }
                    else
                    {
                        this.currentState.UpdateMatch();
                    }
                }

                this.UpdateRepeaters(position, previousPosition);
            }

            this.optimizationData[current].State = this.currentState;
            this.optimizationData[current].Backs0 = this.repeaters[0];
            this.optimizationData[current].Backs1 = this.repeaters[1];
            this.optimizationData[current].Backs2 = this.repeaters[2];
            this.optimizationData[current].Backs3 = this.repeaters[3];
        }

        private void UpdateRepeaters(uint position, uint previousPosition)
        {
            LzmaOptimizationData data = this.optimizationData[previousPosition];
            if (position < LzmaTransformBase.RepeaterDistancesNumber)
            {
                if (position == 0)
                {
                    this.repeaters[0] = data.Backs0;
                    this.repeaters[1] = data.Backs1;
                    this.repeaters[2] = data.Backs2;
                    this.repeaters[3] = data.Backs3;
                }
                else if (position == 1)
                {
                    this.repeaters[0] = data.Backs1;
                    this.repeaters[1] = data.Backs0;
                    this.repeaters[2] = data.Backs2;
                    this.repeaters[3] = data.Backs3;
                }
                else if (position == 2)
                {
                    this.repeaters[0] = data.Backs2;
                    this.repeaters[1] = data.Backs0;
                    this.repeaters[2] = data.Backs1;
                    this.repeaters[3] = data.Backs3;
                }
                else
                {
                    this.repeaters[0] = data.Backs3;
                    this.repeaters[1] = data.Backs0;
                    this.repeaters[2] = data.Backs1;
                    this.repeaters[3] = data.Backs2;
                }
            }
            else
            {
                this.repeaters[0] = position - LzmaTransformBase.RepeaterDistancesNumber;
                this.repeaters[1] = data.Backs0;
                this.repeaters[2] = data.Backs1;
                this.repeaters[3] = data.Backs2;
            }
        }

        private bool CheckFastBytes()
        {
            uint current = this.optimumCurrent;
            if (current == this.optimumLengthEnd)
            {
                this.Backward(current);
                return true;
            }

            uint newLength;
            this.ReadMatchDistances(out newLength);
            this.optimumNewLength = newLength;

            if (this.optimumNewLength >= this.configFastBytes)
            {
                this.distancePairsNumber = distancePairsCounter;
                this.longestMatchLength = this.optimumNewLength;
                this.longestMatchWasFound = true;
                this.Backward(current);
                return true;
            }

            return false;
        }

        private bool SetOptimizationData(uint lengthMain)
        {
            byte currentByte = this.matchFinder.GetIndexByte(0 - 1);
            byte matchByte = this.matchFinder.GetIndexByte((int)(0 - this.repeaterDistances[0] - 2));

            if (lengthMain < 2 && currentByte != matchByte && this.repeaterLengths[this.repeaterMaxIndex] < 2)
            {
                this.optimumPosition = (uint)0xFFFFFFFF;
                this.optimumLength = 1;
                return true;
            }

            uint position = (uint)this.currentPosition;
            uint positionState = this.optimumPositionState;

            this.optimizationData[0].State = this.compressorState;

            uint matchIndex = (compressorState.Index << LzmaTransformBase.PosStatesBitsMax) + positionState;
            bool charState = this.compressorState.IsCharState();
            var literalEncoder = this.literalEncoder.GetSubCoder(position, this.currentPreviousByte);

            this.optimizationData[1].Price = this.matchRangeBitEncoders[matchIndex].GetPrice0()
                + literalEncoder.GetPrice(!charState, matchByte, currentByte);
            this.optimizationData[1].MakeAsChar();

            uint matchPrice = this.matchRangeBitEncoders[(this.compressorState.Index << LzmaTransformBase.PosStatesBitsMax) + positionState].GetPrice1();
            uint repMatchPrice = matchPrice + repeaterRangeBitEncoders[this.compressorState.Index].GetPrice1();

            if (matchByte == currentByte)
            {
                uint shortRepPrice = repMatchPrice + GetRepeaterLen1Price(this.compressorState, positionState);
                if (shortRepPrice < this.optimizationData[1].Price)
                {
                    this.optimizationData[1].Price = shortRepPrice;
                    this.optimizationData[1].MakeAsShortRep();
                }
            }

            this.optimumLengthEnd = lengthMain >= this.repeaterLengths[this.repeaterMaxIndex]
                ? lengthMain
                : this.repeaterLengths[this.repeaterMaxIndex];

            if (this.optimumLengthEnd < 2)
            {
                this.optimumPosition = this.optimizationData[1].BackPrev;
                this.optimumLength = 1;
                return true;
            }

            this.SetOptimizationData(repMatchPrice, positionState);
            this.UpdateDistancePairs(matchPrice, lengthMain);

            return false;
        }

        private void UpdateDistancePairs(uint matchPrice, uint lengthMain)
        {
            uint normalMatchPrice = matchPrice + this.repeaterRangeBitEncoders[this.compressorState.Index].GetPrice0();
            uint length = (this.repeaterLengths[0] >= 2) ? this.repeaterLengths[0] + 1 : 2;
            if (length <= lengthMain)
            {
                uint positionState = this.optimumPositionState;
                uint offset = 0;
                while (length > this.matchDistances[offset])
                {
                    offset += 2;
                }

                for (;; length++)
                {
                    uint distance = this.matchDistances[offset + 1];
                    uint curAndLenPrice = normalMatchPrice + this.GetPosLenPrice(distance, length, positionState);
                    CheckPrice(this.optimizationData[length], curAndLenPrice, 0, distance + LzmaTransformBase.RepeaterDistancesNumber);
                    if (length == this.matchDistances[offset])
                    {
                        offset += 2;
                        if (offset == this.distancePairsCounter)
                        {
                            break;
                        }
                    }
                }
            }
        }

        private void SetOptimizationData(uint repMatchPrice, uint posState)
        {
            this.optimizationData[1].PosPrev = 0;

            this.optimizationData[0].Backs0 = this.repeaters[0];
            this.optimizationData[0].Backs1 = this.repeaters[1];
            this.optimizationData[0].Backs2 = this.repeaters[2];
            this.optimizationData[0].Backs3 = this.repeaters[3];

            uint length = this.optimumLengthEnd;
            do
            {
                this.optimizationData[length--].Price = InfinityPrice;
            }
            while (length >= 2);

            for (uint i = 0; i < LzmaTransformBase.RepeaterDistancesNumber; i++)
            {
                uint repLen = this.repeaterLengths[i];
                if (repLen < 2)
                {
                    continue;
                }

                uint price = repMatchPrice + this.GetPureRepeaterPrice(i, this.compressorState, posState);
                do
                {
                    uint curAndLenPrice = price + this.repeaterMatchLengthEncoder.GetPrice(repLen - 2, posState);
                    CheckPrice(this.optimizationData[repLen], curAndLenPrice, 0, i);
                }
                while (--repLen >= 2);
            }
        }

        private bool CalculateRepeaterMaxIndex(uint lengthMain)
        {
            this.repeaterMaxIndex = 0;
            for (uint i = 0; i < LzmaTransformBase.RepeaterDistancesNumber; i++)
            {
                this.repeaters[i] = this.repeaterDistances[i];
                this.repeaterLengths[i] = this.matchFinder.GetMatchLen(-1, this.repeaters[i], LzmaTransformBase.MatchMaxLen);
                if (this.repeaterLengths[i] > this.repeaterLengths[this.repeaterMaxIndex])
                {
                    this.repeaterMaxIndex = i;
                }
            }

            if (this.repeaterLengths[this.repeaterMaxIndex] >= this.configFastBytes)
            {
                this.optimumPosition = this.repeaterMaxIndex;
                this.optimumLength = this.repeaterLengths[this.repeaterMaxIndex];

                this.MovePosition(this.optimumLength - 1);
                return true;
            }

            if (lengthMain >= this.configFastBytes)
            {
                this.optimumPosition = this.matchDistances[this.distancePairsCounter - 1] + LzmaTransformBase.RepeaterDistancesNumber;
                this.MovePosition(lengthMain - 1);
                this.optimumLength = lengthMain;
                return true;
            }

            return false;
        }

        private bool CheckOptimumIndex()
        {
            if (this.optimumEndIndex != this.optimumCurrentIndex)
            {
                uint index = this.optimumCurrentIndex;
                this.optimumLength = this.optimizationData[index].PosPrev - index;
                this.optimumPosition = this.optimizationData[index].BackPrev;
                this.optimumCurrentIndex = this.optimizationData[index].PosPrev;
                return true;
            }

            this.optimumCurrentIndex = this.optimumEndIndex = 0;

            uint lengthMain;
            if (this.CalculateMatchDistances(out lengthMain))
            {
                this.optimumPosition = 0xFFFFFFFF;
                this.optimumLength = 1;
                return true;
            }

            if (this.CalculateRepeaterMaxIndex(lengthMain) || this.SetOptimizationData(lengthMain))
            {
                return true;
            }

            return false;
        }

        private bool CalculateMatchDistances(out uint lenMain)
        {
            if (!this.longestMatchWasFound)
            {
                this.ReadMatchDistances(out lenMain);
            }
            else
            {
                lenMain = this.longestMatchLength;
                this.distancePairsCounter = this.distancePairsNumber;
                this.longestMatchWasFound = false;
            }

            uint availableBytes = this.matchFinder.AvailableBytes + 1;
            if (availableBytes < 2)
            {
                return true;
            }

            return false;
        }

        private void WriteEndMarker(uint posState)
        {
            if (!this.Settings.UseZipHeader && this.Settings.StreamLength <= -1)
            {
                uint index = (this.compressorState.Index << LzmaTransformBase.PosStatesBitsMax) + posState;
                this.matchRangeBitEncoders[index].Encode(this.rangeEncoder, 1);
                this.repeaterRangeBitEncoders[this.compressorState.Index].Encode(this.rangeEncoder, 0);
                this.compressorState.UpdateMatch();

                uint length = LzmaTransformBase.MatchMinLength;
                this.lengthEncoder.Encode(this.rangeEncoder, length - LzmaTransformBase.MatchMinLength, posState);
                uint posSlot = (1 << LzmaTransformBase.PosSlotBitsNumber) - 1;
                uint lenToPosState = LzmaState.GetLenToPosState(length);
                this.positionSlotEncoder[lenToPosState].Encode(this.rangeEncoder, posSlot);
                int footerBits = 30;
                uint posReduced = (((uint)1) << footerBits) - 1;
                this.rangeEncoder.EncodeDirectBits(posReduced >> LzmaTransformBase.AlignBitsNumber,
                    footerBits - LzmaTransformBase.AlignBitsNumber);

                this.positionAlignEncoder.ReverseEncode(this.rangeEncoder, posReduced & LzmaTransformBase.AlignMask);
            }
        }

        private void Flush(bool finalBlock)
        {
            if (finalBlock)
            {
                this.finished = true;
                this.WriteEndMarker((uint)this.currentPosition & this.configPositionStateMask);
                this.rangeEncoder.FlushData();
            }
        }

        private void CodeOneBlock(bool finalBlock)
        {
            if (this.CodeFirstPart(finalBlock))
            {
                if (this.matchFinder.AvailableBytes == 0)
                {
                    this.Flush(finalBlock);
                }
                else
                {
                    bool ready = this.matchFinder.ReadyToMove;
                    while (!ready && (this.matchFinder.Ready || finalBlock))
                    {
                        this.GetOptimum((uint)this.currentPosition);

                        uint positionState = ((uint)this.currentPosition) & this.configPositionStateMask;
                        uint complexState = (this.compressorState.Index << LzmaTransformBase.PosStatesBitsMax) + positionState;
                        if (this.optimumLength == 1 && this.optimumPosition == 0xFFFFFFFF)
                        {
                            this.Code0(complexState);
                        }
                        else
                        {
                            this.Code1(complexState, positionState);
                        }

                        if (this.UpdateCodePosition(finalBlock))
                        {
                            return;
                        }
                    }

                    if (ready)
                    {
                        this.matchFinder.MoveBlock();
                    }
                }
            }
        }

        private bool UpdateCodePosition(bool finalBlock)
        {
            uint length = this.optimumLength;
            this.additionalOffset -= length;
            this.currentPosition += length;

            if (this.additionalOffset == 0)
            {
                if (this.matchPriceCount >= (1 << 7))
                {
                    this.FillDistancesPrices();
                }

                if (this.alignPriceCount >= LzmaTransformBase.AlignTableSize)
                {
                    this.FillAlignPrices();
                }

                if (this.matchFinder.AvailableBytes == 0)
                {
                    this.Flush(finalBlock);
                    return true;
                }
            }

            return false;
        }

        private void Code1(uint complexState, uint positionState)
        {
            uint position = this.optimumPosition;
            uint length = this.optimumLength;

            this.matchRangeBitEncoders[complexState].Encode(this.rangeEncoder, 1);
            if (position < LzmaTransformBase.RepeaterDistancesNumber)
            {
                this.Code1Encoding(complexState, positionState);
            }
            else
            {
                this.Code0Encoding(positionState);
            }

            this.currentPreviousByte = this.matchFinder.GetIndexByte((int)(length - 1 - this.additionalOffset));
        }

        private void Code0Encoding(uint positionState)
        {
            uint position = this.optimumPosition;
            uint length = this.optimumLength;

            this.repeaterRangeBitEncoders[this.compressorState.Index].Encode(this.rangeEncoder, 0);
            this.compressorState.UpdateMatch();
            this.lengthEncoder.Encode(this.rangeEncoder, length - LzmaTransformBase.MatchMinLength, positionState);
            position -= LzmaTransformBase.RepeaterDistancesNumber;
            uint posSlot = GetPosSlot(position);
            uint lenToPosState = LzmaState.GetLenToPosState(length);
            this.positionSlotEncoder[lenToPosState].Encode(this.rangeEncoder, posSlot);

            if (posSlot >= LzmaTransformBase.StartPosModelIndex)
            {
                int footerBits = (int)((posSlot >> 1) - 1);
                uint baseVal = (2 | (posSlot & 1)) << footerBits;
                uint posReduced = position - baseVal;

                if (posSlot < LzmaTransformBase.EndPosModelIndex)
                {
                    uint startIndex = baseVal - posSlot - 1;
                    LzmaBitTreeEncoder.ReverseEncode(this.positionEncoders, startIndex, this.rangeEncoder, footerBits, posReduced);
                }
                else
                {
                    this.rangeEncoder.EncodeDirectBits(posReduced >> LzmaTransformBase.AlignBitsNumber,
                        footerBits - LzmaTransformBase.AlignBitsNumber);
                    this.positionAlignEncoder.ReverseEncode(this.rangeEncoder,
                        posReduced & LzmaTransformBase.AlignMask);
                    this.alignPriceCount++;
                }
            }

            uint distance = position;
            for (uint i = LzmaTransformBase.RepeaterDistancesNumber - 1; i >= 1; i--)
            {
                this.repeaterDistances[i] = this.repeaterDistances[i - 1];
            }

            this.repeaterDistances[0] = distance;
            this.matchPriceCount++;

            this.optimumPosition = position;
        }

        private void Code1Encoding(uint complexState, uint positionState)
        {
            uint position = this.optimumPosition;
            uint length = this.optimumLength;

            this.repeaterRangeBitEncoders[this.compressorState.Index].Encode(this.rangeEncoder, 1);
            if (position == 0)
            {
                this.repeaterG0RangeBitEncoders[this.compressorState.Index].Encode(this.rangeEncoder, 0);
                if (length == 1)
                {
                    this.repeaterLongRangeBitEncoders[complexState].Encode(this.rangeEncoder, 0);
                }
                else
                {
                    this.repeaterLongRangeBitEncoders[complexState].Encode(this.rangeEncoder, 1);
                }
            }
            else
            {
                this.repeaterG0RangeBitEncoders[this.compressorState.Index].Encode(this.rangeEncoder, 1);
                if (position == 1)
                {
                    this.repeaterG1RangeBitEncoders[this.compressorState.Index].Encode(this.rangeEncoder, 0);
                }
                else
                {
                    this.repeaterG1RangeBitEncoders[this.compressorState.Index].Encode(this.rangeEncoder, 1);
                    this.repeaterG2RangeBitEncoders[this.compressorState.Index].Encode(this.rangeEncoder, position - 2);
                }
            }

            if (length == 1)
            {
                this.compressorState.UpdateShortRep();
            }
            else
            {
                this.repeaterMatchLengthEncoder.Encode(this.rangeEncoder,
                    length - LzmaTransformBase.MatchMinLength,
                    positionState);
                this.compressorState.UpdateRep();
            }

            uint distance = this.repeaterDistances[position];
            if (position != 0)
            {
                for (uint i = position; i >= 1; i--)
                {
                    this.repeaterDistances[i] = this.repeaterDistances[i - 1];
                }

                this.repeaterDistances[0] = distance;
            }
        }

        private void Code0(uint complexState)
        {
            this.matchRangeBitEncoders[complexState].Encode(this.rangeEncoder, 0);
            byte curentByte = this.matchFinder.GetIndexByte((int)(0 - this.additionalOffset));
            LzmaLiteralEncoder.Encoder subCoder =
                this.literalEncoder.GetSubCoder((uint)this.currentPosition, this.currentPreviousByte);

            if (!this.compressorState.IsCharState())
            {
                byte matchByte = this.matchFinder.GetIndexByte((int)(0 - this.repeaterDistances[0] - 1 - additionalOffset));
                subCoder.EncodeMatched(this.rangeEncoder, matchByte, curentByte);
            }
            else
            {
                subCoder.Encode(rangeEncoder, curentByte);
            }

            this.currentPreviousByte = curentByte;
            this.compressorState.UpdateChar();
        }

        private bool CodeFirstPart(bool finalBlock)
        {
            if (this.currentPosition == 0)
            {
                if (!this.matchFinder.Ready)
                {
                    return false;
                }

                if (this.matchFinder.AvailableBytes == 0)
                {
                    this.Flush(finalBlock);
                    return false;
                }

                uint length;
                this.ReadMatchDistances(out length);
                uint posState = (uint)this.currentPosition & this.configPositionStateMask;
                uint index = (this.compressorState.Index << LzmaTransformBase.PosStatesBitsMax) + posState;
                this.matchRangeBitEncoders[index].Encode(this.rangeEncoder, 0);
                this.compressorState.UpdateChar();

                byte curentByte = this.matchFinder.GetIndexByte((int)(0 - this.additionalOffset));
                this.literalEncoder.GetSubCoder((uint)this.currentPosition, this.currentPreviousByte).Encode(this.rangeEncoder, curentByte);
                this.currentPreviousByte = curentByte;
                this.additionalOffset--;
                this.currentPosition++;
            }

            return true;
        }

        private void Initialize()
        {
            this.ApplySettings();
            this.InitEncoders();

            this.FillDistancesPrices();
            this.FillAlignPrices();

            this.lengthEncoder.SetTableSize(configFastBytes + 1 - LzmaTransformBase.MatchMinLength);
            this.lengthEncoder.UpdateTables((uint)1 << this.configPositionStateBits);
            this.repeaterMatchLengthEncoder.SetTableSize(this.configFastBytes + 1 - LzmaTransformBase.MatchMinLength);
            this.repeaterMatchLengthEncoder.UpdateTables((uint)1 << this.configPositionStateBits);

            this.currentPosition = 0;
        }

        private void ApplySettings()
        {
            this.ApplyDictionarySize();
            this.ApplyPositionStateBits();
            this.ApplyLiteralContextBits();
            this.ApplyLiteralPositionBits();
            this.ApplyFastBytes();
            this.ApplyMatchFinderType();

            this.literalEncoder = new LzmaLiteralEncoder(this.configLiteralPositionStateBits, this.configLiteralContextBits);
        }

        private void ApplyDictionarySize()
        {
            int dictionary = this.Settings.DictionarySize;
            if (dictionary < LzmaTransformBase.DictionaryLogSizeMin)
            {
                dictionary = LzmaTransformBase.DictionaryLogSizeMin;
            }
            else if (dictionary > DictionaryMaxCompressValue)
            {
                dictionary = DictionaryMaxCompressValue;
            }

            int dictionarySize = 1 << dictionary;
            this.configDictionarySize = (uint)dictionarySize;
            int dictionaryLogSize;
            for (dictionaryLogSize = 0; dictionaryLogSize < (uint)DictionaryMaxCompressValue; dictionaryLogSize++)
            {
                if (dictionarySize <= ((uint)1 << dictionaryLogSize))
                {
                    break;
                }
            }

            this.configDistanceTableSize = (uint)dictionaryLogSize * 2;
        }

        private void ApplyPositionStateBits()
        {
            // Specifies number of position state bits for LZMA (0 <= x <= 4).
            int posStateBits = this.Settings.PositionStateBits;
            if (posStateBits < 0)
            {
                posStateBits = 0;
            }
            else if (posStateBits > LzmaTransformBase.PosStatesBitsEncodingMax)
            {
                posStateBits = LzmaTransformBase.PosStatesBitsEncodingMax;
            }

            this.configPositionStateBits = posStateBits;
            this.configPositionStateMask = (((uint)1) << this.configPositionStateBits) - 1;
        }

        private void ApplyLiteralContextBits()
        {
            // Specifies number of literal context bits for LZMA (0 <= x <= 8).
            int literalContextBits = this.Settings.LiteralContextBits;
            if (literalContextBits < 0)
            {
                literalContextBits = 0;
            }
            else if (literalContextBits > LzmaTransformBase.LiteralContextBitsMax)
            {
                literalContextBits = LzmaTransformBase.LiteralContextBitsMax;
            }

            this.configLiteralContextBits = literalContextBits;
        }

        private void ApplyLiteralPositionBits()
        {
            // Specifies number of literal position bits for LZMA (0 <= x <= 4).
            int literalPositionBits = this.Settings.LiteralPositionBits;
            if (literalPositionBits < 0)
            {
                literalPositionBits = 0;
            }
            else if (literalPositionBits > LzmaTransformBase.LiteralPosStatesBitsEncodingMax)
            {
                literalPositionBits = (int)LzmaTransformBase.LiteralPosStatesBitsEncodingMax;
            }

            this.configLiteralPositionStateBits = literalPositionBits;
        }

        private void ApplyFastBytes()
        {
            // Specifies number of fast bytes for LZ*.
            uint fastBytes = (uint)this.Settings.FastBytes;
            if (fastBytes < 5)
            {
                fastBytes = 5;
            }
            else if (fastBytes > LzmaTransformBase.MatchMaxLen)
            {
                fastBytes = LzmaTransformBase.MatchMaxLen;
            }

            this.configFastBytes = fastBytes;
        }

        private void ApplyMatchFinderType()
        {
            this.matchFinderType = this.Settings.MatchFinderType;

            int hashBytes = 4;
            if (this.matchFinderType == LzmaMatchFinderType.BT2)
            {
                hashBytes = 2;
            }

            this.matchFinder = new LzmaBinaryTree(hashBytes,
                this.configDictionarySize,
                this.configFastBytes,
                LzmaTransformBase.MatchMaxLen + 1);
        }

        private void FillDistancesPrices()
        {
            for (uint i = LzmaTransformBase.StartPosModelIndex; i < LzmaTransformBase.FullDistancesNumber; i++)
            { 
                uint posSlot = GetPosSlot(i);
                int footerBits = (int)((posSlot >> 1) - 1);
                uint baseVal = (2 | (posSlot & 1)) << footerBits;

                this.tempPrices[i] = LzmaBitTreeEncoder.ReverseGetPrice(this.positionEncoders, 
                    baseVal - posSlot - 1,
                    footerBits,
                    i - baseVal);
            }

            for (uint lenToPosState = 0; lenToPosState < LzmaTransformBase.LengthToPosStates; lenToPosState++)
            {
                uint posSlot;
                LzmaBitTreeEncoder encoder = this.positionSlotEncoder[lenToPosState];

                uint st = lenToPosState << LzmaTransformBase.PosSlotBitsNumber;
                for (posSlot = 0; posSlot < configDistanceTableSize; posSlot++)
                {
                    this.positionSlotPrices[st + posSlot] = encoder.GetPrice(posSlot);
                }

                for (posSlot = LzmaTransformBase.EndPosModelIndex; posSlot < configDistanceTableSize; posSlot++)
                {
                    uint value = (((posSlot >> 1) - 1) - LzmaTransformBase.AlignBitsNumber) << LzmaRangeBitEncoder.BitPriceShiftBits;
                    this.positionSlotPrices[st + posSlot] += value;
                }

                uint st2 = lenToPosState * LzmaTransformBase.FullDistancesNumber;
                uint i;
                for (i = 0; i < LzmaTransformBase.StartPosModelIndex; i++)
                {
                    this.distancesPrices[st2 + i] = this.positionSlotPrices[st + i];
                }

                for (; i < LzmaTransformBase.FullDistancesNumber; i++)
                {
                    this.distancesPrices[st2 + i] = this.positionSlotPrices[st + GetPosSlot(i)] + this.tempPrices[i];
                }
            }

            this.matchPriceCount = 0;
        }

        private void FillAlignPrices()
        {
            for (uint i = 0; i < LzmaTransformBase.AlignTableSize; i++)
            {
                this.alignPrices[i] = this.positionAlignEncoder.ReverseGetPrice(i);
            }

            this.alignPriceCount = 0;
        }
    }
}
