﻿using System;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Decompressor which implements LZMA decompression algorithm.
    /// </summary>
    internal class LzmaDecompressor : LzmaTransformBase
    {
        private LzmaState lzmaState;
        private uint dictionarySize;
        private uint dictionarySizeCheck;

        private LzmaLengthDecoder lengthDecoder;
        private LzmaLengthDecoder repeaterLengthDecoder;
        private LzmaLiteralDecoder literalDecoder;
        private LzmaBitTreeDecoder[] positionSlotDecoder = new LzmaBitTreeDecoder[LzmaTransformBase.LengthToPosStates];
        private LzmaOutputWindow outputWindow;

        private LzmaRangeDecoder rangeDecoder = new LzmaRangeDecoder();

        private LzmaRangeBitDecoder[] matchDecoders = new LzmaRangeBitDecoder[LzmaTransformBase.StatesNumber << LzmaTransformBase.PosStatesBitsMax];
        private LzmaRangeBitDecoder[] repeaterDecoders = new LzmaRangeBitDecoder[LzmaTransformBase.StatesNumber];
        private LzmaRangeBitDecoder[] repeaterG0Decoders = new LzmaRangeBitDecoder[LzmaTransformBase.StatesNumber];
        private LzmaRangeBitDecoder[] repeaterG1Decoders = new LzmaRangeBitDecoder[LzmaTransformBase.StatesNumber];
        private LzmaRangeBitDecoder[] repeaterG2Decoders = new LzmaRangeBitDecoder[LzmaTransformBase.StatesNumber];
        private LzmaRangeBitDecoder[] repeaterLongDecoders = new LzmaRangeBitDecoder[LzmaTransformBase.StatesNumber << LzmaTransformBase.PosStatesBitsMax];

        private LzmaRangeBitDecoder[] positionDecoders = new LzmaRangeBitDecoder[LzmaTransformBase.FullDistancesNumber - LzmaTransformBase.EndPosModelIndex];
        private LzmaBitTreeDecoder positionAlignDecoder = new LzmaBitTreeDecoder(LzmaTransformBase.AlignBitsNumber);
        private LzmaDecompressorState decompressState = LzmaDecompressorState.Start;

        private ulong currentPosition;
        private uint repeater = 0;
        private uint repeaterPosition1 = 0;
        private uint repeaterPosition2 = 0;
        private uint repeaterPosition3 = 0;
        private uint decodedLength;
        private uint positionStateMask;
        private uint decodedState;
        private uint positionState;
        private uint positionSlot;

        /// <summary>
        /// Initializes a new instance of the <see cref="LzmaDecompressor"/> class.
        /// </summary>
        /// <param name="settings">Settings.</param>
        public LzmaDecompressor(LzmaSettings settings)
            : base(settings)
        {
            this.dictionarySize = 0xFFFFFFFF;
            for (int i = 0; i < LzmaTransformBase.LengthToPosStates; i++)
            {
                this.positionSlotDecoder[i] = new LzmaBitTreeDecoder(LzmaTransformBase.PosSlotBitsNumber);
            }
        }

        /// <summary>
        /// Represents a state of decompressor process.
        /// </summary>
        private enum LzmaDecompressorState
        {
            Start,
            MatchDecoder,
            LiteralDecoder,
            RepeaterDecoder,
            RepeaterG0Decoder,
            Repeater0LongDecoder,
            RepeaterG1Decoder,
            RepeaterG2Decoder,
            RepeaterLengthDecoder,
            LengthDecoder,
            PosSlotDecoder,
            PosSlotProcess,
            ReverseDecoder,
            CopyBlock,
            Done,
        }

        /// <summary>Gets the output block size.</summary>
        /// <returns>The size of the output data blocks in bytes.</returns>
        public override int OutputBlockSize
        {
            get
            {
                return 262144;
            }
        }

        /// <summary>
        /// Initialize reading of the transformation header.
        /// </summary>
        public override void InitHeaderReading()
        {
            base.InitHeaderReading();

            if (this.Settings.UseZipHeader)
            {
                this.Header.BytesToRead = 14;
            }
            else
            {
                this.Header.BytesToRead = 18;
            }
        }

        /// <summary>
        /// Process transformation header has been read.
        /// </summary>
        public override void ProcessHeader()
        {
            base.ProcessHeader();

            this.SetProperties();

            this.Header.BytesToRead = 0;
        }

        protected override bool ProcessTransform(bool finalBlock)
        {
            this.rangeDecoder.FinalBlock = finalBlock;

            this.SetInputBuffer();
            this.outputWindow.SetOutputBuffer(this.OutputBuffer);
            this.outputWindow.Flush();

            while (!this.rangeDecoder.InputRequired
                && !this.outputWindow.Full
                && this.decompressState != LzmaDecompressorState.Done)
            {
                this.Decompress();
                this.outputWindow.Flush();
            }

            this.NextOut += this.outputWindow.Copied;

            return this.outputWindow.AvailableBytes > 0
                || (this.decompressState != LzmaDecompressorState.Done && !this.rangeDecoder.InputRequired);
        }

        private static void InvalidLzmaProperties()
        {
            throw new InvalidDataException("Invalid LZMA properties");
        }

        private void SetProperties()
        {
            string message = string.Empty;
            string invalidHeaderLength = "Invalid header length";
            if (this.Settings.UseZipHeader)
            {
                if (this.Header.Buffer == null || this.Header.Buffer.Length != 14)
                {
                    message = invalidHeaderLength;
                }
                else
                {
                    this.InitializeDecompressor(4);
                    this.Init(this.Header.Buffer, 9);
                }
            }
            else
            {
                if (this.Header.Buffer == null || this.Header.Buffer.Length != 18)
                {
                    message = invalidHeaderLength;
                }
                else
                {
                    this.InitializeDecompressor(0);
                    this.Init(this.Header.Buffer, 13);
                    this.Settings.InternalStreamLength = BitConverter.ToInt64(this.Header.Buffer, 5);
                }
            }

            if (message.Length != 0)
            {
                throw new InvalidDataException(message);
            }
        }

        private void InitializeDecompressor(int propertiesIndex)
        {
            byte[] properties = this.Header.Buffer;

            int lc = properties[propertiesIndex] % 9;
            int remainder = properties[propertiesIndex] / 9;
            int lp = remainder % 5;
            int pb = remainder / 5;
            if (pb > LzmaTransformBase.PosStatesBitsMax)
            {
                InvalidLzmaProperties();
            }

            uint dictionarySize = 0;
            propertiesIndex++;
            for (int i = 0; i < 4; i++)
            {
                dictionarySize += ((uint)properties[propertiesIndex + i]) << (i * 8);
            }

            this.SetDictionarySize(dictionarySize);
            SetLiteralProperties(lp, lc);
            SetPosBitsProperties(pb);
        }

        private void SetLiteralProperties(int lp, int lc)
        {
            if (lp > 8 || lc > 8)
            {
                InvalidLzmaProperties();
            }

            this.literalDecoder = new LzmaLiteralDecoder(lp, lc);
        }

        private void SetPosBitsProperties(int pb)
        {
            if (pb > LzmaTransformBase.PosStatesBitsMax)
            {
                InvalidLzmaProperties();
            }

            uint numPosStates = (uint)1 << pb;

            this.lengthDecoder = new LzmaLengthDecoder(numPosStates);
            this.repeaterLengthDecoder = new LzmaLengthDecoder(numPosStates);
            this.positionStateMask = numPosStates - 1;
        }

        private void SetDictionarySize(uint size)
        {
            this.dictionarySize = size;
            this.dictionarySizeCheck = Math.Max(this.dictionarySize, 1);
            uint blockSize = Math.Max(this.dictionarySizeCheck, 1 << 12);
            this.outputWindow = new LzmaOutputWindow(blockSize);
        }

        private void Init(byte[] inputBuffer, int startIndex)
        {
            this.lzmaState = new LzmaState();

            this.rangeDecoder.Init(inputBuffer, startIndex);

            uint i;
            for (i = 0; i < LzmaTransformBase.StatesNumber; i++)
            {
                for (uint j = 0; j <= this.positionStateMask; j++)
                {
                    uint index = (i << LzmaTransformBase.PosStatesBitsMax) + j;
                    this.matchDecoders[index].Init();
                    this.repeaterLongDecoders[index].Init();
                }

                this.repeaterDecoders[i].Init();
                this.repeaterG0Decoders[i].Init();
                this.repeaterG1Decoders[i].Init();
                this.repeaterG2Decoders[i].Init();
            }

            this.literalDecoder.Init();
            for (i = 0; i < LzmaTransformBase.LengthToPosStates; i++)
            {
                this.positionSlotDecoder[i].Init();
            }

            for (i = 0; i < LzmaTransformBase.FullDistancesNumber - LzmaTransformBase.EndPosModelIndex; i++)
            {
                this.positionDecoders[i].Init();
            }

            this.lengthDecoder.Init();
            this.repeaterLengthDecoder.Init();
            this.positionAlignDecoder.Init();
        }

        private void DecompressStart()
        {
            uint decoderIndex = this.lzmaState.Index << LzmaTransformBase.PosStatesBitsMax;
            this.decodedState = this.matchDecoders[decoderIndex].DecodeState(this.rangeDecoder);
            if (this.decodedState != 0)
            {
                if (this.currentPosition == 0
                    && (this.AvailableBytesIn == 0 || this.AvailableBytesIn == 5))
                {
                    this.decompressState = LzmaDecompressorState.Done;
                    return;
                }

                throw new InvalidDataException();
            }

            this.lzmaState.UpdateChar();

            byte decoded = this.literalDecoder.DecodeNormal(this.rangeDecoder, 0, 0);
            this.outputWindow.PutByte(decoded);
            this.currentPosition++;

            this.decompressState = LzmaDecompressorState.MatchDecoder;
        }

        private bool Decompress()
        {
            while (!this.rangeDecoder.InputRequired
                && !this.outputWindow.Full
                && this.decompressState != LzmaDecompressorState.Done)
            {
                switch (this.decompressState)
                {
                    case LzmaDecompressorState.Start:
                        this.DecompressStart();
                        continue;

                    case LzmaDecompressorState.MatchDecoder:
                        this.positionState = (uint)this.currentPosition & this.positionStateMask;
                        this.DecompressMatchDecoder();
                        continue;

                    case LzmaDecompressorState.LiteralDecoder:
                        this.DecompressLiteralDecoder();
                        continue;

                    case LzmaDecompressorState.RepeaterDecoder:
                        this.DecompressRepeaterDecoder();
                        continue;

                    case LzmaDecompressorState.RepeaterG0Decoder:
                        this.DecompressRepeaterG0Decoder();
                        continue;

                    case LzmaDecompressorState.Repeater0LongDecoder:
                        this.DecompressRepeater0LongDecoder();
                        continue;

                    case LzmaDecompressorState.RepeaterLengthDecoder:
                        this.DecompressRepeaterLengthDecoder();
                        continue;

                    case LzmaDecompressorState.RepeaterG1Decoder:
                        this.DecompressRepeaterG1Decoder();
                        continue;

                    case LzmaDecompressorState.RepeaterG2Decoder:
                        this.DecompressRepeaterG2Decoder();
                        continue;

                    case LzmaDecompressorState.LengthDecoder:
                        this.DecompressLengthDecoder();
                        continue;

                    case LzmaDecompressorState.PosSlotDecoder:
                        this.DecompressPosSlotDecoder();
                        continue;

                    case LzmaDecompressorState.PosSlotProcess:
                        this.DecompressPosSlotProcess();
                        continue;

                    case LzmaDecompressorState.ReverseDecoder:
                        this.DecompressReverseDecoder();
                        continue;

                    case LzmaDecompressorState.CopyBlock:
                        this.DecompressCopyBlock();
                        continue;
                }
            }

            return false;
        }

        private void DecompressMatchDecoder()
        {
            uint decoderIndex = (this.lzmaState.Index << LzmaTransformBase.PosStatesBitsMax) + this.positionState;
            uint state = this.matchDecoders[decoderIndex].DecodeState(this.rangeDecoder);
            bool inputRequired = this.rangeDecoder.InputRequired;
            bool ready = this.rangeDecoder.CheckInputRequired(1, true);
            if (!ready && !inputRequired && this.rangeDecoder.FinalBlock)
            {
                if (this.currentPosition < (ulong)this.Settings.InternalStreamLength)
                {
                    ready = this.rangeDecoder.CheckInputRequired(0, true);
                }
                else
                {
                    this.decompressState = LzmaDecompressorState.Done;
                }
            }

            if (ready)
            {
                this.decodedState = state;
                if (this.decodedState == 0)
                {
                    this.decompressState = LzmaDecompressorState.LiteralDecoder;
                }
                else
                {
                    this.decompressState = LzmaDecompressorState.RepeaterDecoder;
                }
            }
        }

        private void DecompressLiteralDecoder()
        {
            byte decoded;
            byte previousByte = this.outputWindow.GetByte(0);
            if (this.rangeDecoder.CheckInputRequired(0x100))
            {
                if (this.lzmaState.IsCharState())
                {
                    decoded = this.literalDecoder.DecodeNormal(this.rangeDecoder, (uint)this.currentPosition, previousByte);
                }
                else
                {
                    decoded = this.literalDecoder.DecodeWithMatchByte(this.rangeDecoder,
                        (uint)this.currentPosition,
                        previousByte,
                        this.outputWindow.GetByte(this.repeater));
                }

                this.outputWindow.PutByte(decoded);
                this.lzmaState.UpdateChar();
                this.currentPosition++;

                this.decompressState = LzmaDecompressorState.MatchDecoder;
            }
        }

        private void DecompressRepeaterDecoder()
        {
            uint state = this.repeaterDecoders[this.lzmaState.Index].DecodeState(this.rangeDecoder);
            if (!this.rangeDecoder.InputRequired)
            {
                this.decodedState = state;
                if (this.decodedState == 0)
                {
                    this.decompressState = LzmaDecompressorState.LengthDecoder;
                }
                else
                {
                    this.decompressState = LzmaDecompressorState.RepeaterG0Decoder;
                }
            }
        }

        private void DecompressRepeaterG0Decoder()
        {
            uint state = this.repeaterG0Decoders[this.lzmaState.Index].DecodeState(this.rangeDecoder);
            if (!this.rangeDecoder.InputRequired)
            {
                this.decodedState = state;
                if (this.decodedState == 0)
                {
                    this.decompressState = LzmaDecompressorState.Repeater0LongDecoder;
                }
                else
                {
                    this.decompressState = LzmaDecompressorState.RepeaterG1Decoder;
                }
            }
        }

        private void DecompressRepeater0LongDecoder()
        {
            uint index = (this.lzmaState.Index << LzmaTransformBase.PosStatesBitsMax) + positionState;
            uint state = this.repeaterLongDecoders[index].DecodeState(this.rangeDecoder);
            if (!this.rangeDecoder.InputRequired)
            {
                this.decodedState = state;
                if (this.decodedState == 0)
                {
                    this.lzmaState.UpdateShortRep();
                    this.outputWindow.PutByte(this.outputWindow.GetByte(this.repeater));
                    this.currentPosition++;
                    this.decompressState = LzmaDecompressorState.MatchDecoder;
                }
                else
                {
                    this.decompressState = LzmaDecompressorState.RepeaterLengthDecoder;
                }
            }
        }

        private void DecompressRepeaterLengthDecoder()
        {
            decodedLength = this.repeaterLengthDecoder.Decode(this.rangeDecoder, positionState) + LzmaTransformBase.MatchMinLength;
            if (!this.rangeDecoder.InputRequired)
            {
                this.lzmaState.UpdateRep();
                this.decompressState = LzmaDecompressorState.CopyBlock;
            }
        }

        private void DecompressRepeaterG1Decoder()
        {
            uint distance;
            uint state = this.repeaterG1Decoders[this.lzmaState.Index].DecodeState(this.rangeDecoder);
            if (!this.rangeDecoder.InputRequired)
            {
                this.decodedState = state;
                if (this.decodedState == 0)
                {
                    distance = this.repeaterPosition1;
                    this.repeaterPosition1 = this.repeater;
                    this.repeater = distance;
                    this.decompressState = LzmaDecompressorState.RepeaterLengthDecoder;
                }
                else
                {
                    this.decompressState = LzmaDecompressorState.RepeaterG2Decoder;
                }
            }
        }

        private void DecompressRepeaterG2Decoder()
        {
            uint distance;
            uint state = this.repeaterG2Decoders[this.lzmaState.Index].DecodeState(this.rangeDecoder);
            if (!this.rangeDecoder.InputRequired)
            {
                this.decodedState = state;
                if (decodedState == 0)
                {
                    distance = this.repeaterPosition2;
                }
                else
                {
                    distance = this.repeaterPosition3;
                    this.repeaterPosition3 = this.repeaterPosition2;
                }
                this.repeaterPosition2 = this.repeaterPosition1;
                this.repeaterPosition1 = this.repeater;
                this.repeater = distance;

                this.decompressState = LzmaDecompressorState.RepeaterLengthDecoder;
            }
        }

        private void DecompressCopyBlock()
        {
            if (this.repeater >= this.currentPosition || this.repeater >= this.dictionarySizeCheck)
            {
                if (this.repeater == 0xFFFFFFFF)
                {
                    this.decompressState = LzmaDecompressorState.Done;
                }
                else
                {
                    throw new InvalidDataException();
                }
            }
            else
            {
                this.outputWindow.CopyBlock(this.repeater, this.decodedLength);
                this.currentPosition += this.decodedLength;

                this.decompressState = LzmaDecompressorState.MatchDecoder;
            }
        }

        private void DecompressReverseDecoder()
        {
            uint add = this.positionAlignDecoder.ReverseDecode(this.rangeDecoder);
            if (!this.rangeDecoder.InputRequired)
            {
                this.repeater += add;
                this.decompressState = LzmaDecompressorState.CopyBlock;
            }
        }

        private void DecompressPosSlotProcess()
        {
            if (this.positionSlot >= LzmaTransformBase.StartPosModelIndex)
            {
                int numDirectBits = (int)((this.positionSlot >> 1) - 1);
                this.repeater = (2 | (this.positionSlot & 1)) << numDirectBits;

                if (this.positionSlot < LzmaTransformBase.EndPosModelIndex)
                {
                    uint add = LzmaBitTreeDecoder.ReverseDecode(this.positionDecoders,
                        this.repeater - this.positionSlot - 1,
                        this.rangeDecoder,
                        numDirectBits);

                    if (this.rangeDecoder.InputRequired)
                    {
                        return;
                    }

                    this.repeater += add;
                }
                else
                {
                    uint add = this.rangeDecoder.DecodeDirectBits(numDirectBits - LzmaTransformBase.AlignBitsNumber) << LzmaTransformBase.AlignBitsNumber;
                    if (this.rangeDecoder.InputRequired)
                    {
                        return;
                    }

                    this.repeater += add;
                    this.decompressState = LzmaDecompressorState.ReverseDecoder;

                    return;
                }
            }
            else
            {
                this.repeater = this.positionSlot;
            }

            this.decompressState = LzmaDecompressorState.CopyBlock;
        }

        private void DecompressPosSlotDecoder()
        {
            this.positionSlot = this.positionSlotDecoder[LzmaState.GetLenToPosState(this.decodedLength)].Decode(this.rangeDecoder);
            if (!this.rangeDecoder.InputRequired)
            {
                this.decompressState = LzmaDecompressorState.PosSlotProcess;
            }
        }

        private void DecompressLengthDecoder()
        {
            this.decodedLength = LzmaTransformBase.MatchMinLength
                + this.lengthDecoder.Decode(this.rangeDecoder, this.positionState);
            if (!this.rangeDecoder.InputRequired)
            {
                this.repeaterPosition3 = this.repeaterPosition2;
                this.repeaterPosition2 = this.repeaterPosition1;
                this.repeaterPosition1 = this.repeater;
                this.lzmaState.UpdateMatch();

                this.decompressState = LzmaDecompressorState.PosSlotDecoder;
            }
        }

        private void SetInput(byte[] inputBytes, int length)
        {
            this.rangeDecoder.SetBuffer(inputBytes, length);
        }

        private void SetInputBuffer()
        {
            if (this.AvailableBytesIn > 0)
            {
                byte[] buffer = new byte[this.AvailableBytesIn];
                Array.Copy(this.InputBuffer, 0, buffer, this.NextIn, buffer.Length);

                this.SetInput(buffer, this.AvailableBytesIn);
            }
        }
    }
}
