﻿#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Compression settings of the Deflate method.
    /// </summary>
    public class LzmaSettings : CompressionSettings
    {
        private int dictionarySize;
        private int positionStateBits;
        private int literalContextBits;
        private int literalPositionBits;
        private int fastBytes;
        private LzmaMatchFinderType matchFinderType;

        /// <summary>
        /// Initializes a new instance of the <see cref="LzmaSettings"/> class.
        /// </summary>
        public LzmaSettings()
        {
            this.Method = CompressionMethod.Lzma;
            this.UseZipHeader = false;
            this.dictionarySize = 23;
            this.positionStateBits = 2;
            this.literalContextBits = 3;
            this.literalPositionBits = 0;
            this.fastBytes = 32;
            this.matchFinderType = LzmaMatchFinderType.BT4;
            this.StreamLength = -1;
        }

        /// <summary>
        /// Gets or sets dictionary size [0 - 27].
        /// Default value is 23 (8MB).
        /// </summary>
        public int DictionarySize
        {
            get
            {
                return this.dictionarySize;
            }
            set
            {
                this.dictionarySize = value;
                this.OnPropertyChanged("DictionarySize");
            }
        }

        /// <summary>
        /// Gets or sets number of position state bits for LZMA [0 - 4].
        /// Default value is 2.
        /// </summary>
        public int PositionStateBits
        {
            get
            {
                return this.positionStateBits;
            }
            set
            {
                this.positionStateBits = value;
                this.OnPropertyChanged("PositionStateBits");
            }
        }

        /// <summary>
        /// Gets or sets number of literal context bits for LZMA [0 - 8].
        /// Default value is 3.
        /// </summary>
        public int LiteralContextBits
        {
            get
            {
                return this.literalContextBits;
            }
            set
            {
                this.literalContextBits = value;
                this.OnPropertyChanged("LiteralContextBits");
            }
        }

        /// <summary>
        /// Gets or sets number of literal position bits for LZMA [0 - 4].
        /// Default value is 3.
        /// </summary>
        public int LiteralPositionBits
        {
            get
            {
                return this.literalPositionBits;
            }
            set
            {
                this.literalPositionBits = value;
                this.OnPropertyChanged("LiteralPositionBits");
            }
        }

        /// <summary>
        /// Gets or sets number of fast bytes [5 - 273].
        /// Default value is 32.
        /// </summary>
        public int FastBytes
        {
            get
            {
                return this.fastBytes;
            }
            set
            {
                this.fastBytes = value;
                this.OnPropertyChanged("FastBytes");
            }
        }

        /// <summary>
        /// Gets or sets a type of the match finder.
        /// </summary>
        public LzmaMatchFinderType MatchFinderType
        {
            get
            {
                return this.matchFinderType;
            }
            set
            {
                this.matchFinderType = value;
                this.OnPropertyChanged("MatchFinderType");
            }
        }

        /// <summary>
        /// Gets or sets length of the stream for compressing.
        /// Used for single compressed streams only (not for ZIP archives).
        /// Allows to avoid using the end of stream marker for compressed stream.
        /// If it is set to -1, then the marker will be used.
        /// </summary>
        public long StreamLength
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets length of the stream for decompressing.
        /// </summary>
        internal long InternalStreamLength
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets a value which indicates whether
        /// the compression stream should use zip header type.
        /// </summary>
        internal bool UseZipHeader
        {
            get;
            set;
        }

        /// <summary>
        /// Copy settings from the given base settings.
        /// </summary>
        /// <param name="baseSettings">Base settings to copy from.</param>
        internal override void CopyFrom(CompressionSettings baseSettings)
        {
            LzmaSettings lzmaSettings = baseSettings as LzmaSettings;
            if (lzmaSettings != null)
            {
                this.UseZipHeader = lzmaSettings.UseZipHeader;
                this.DictionarySize = lzmaSettings.DictionarySize;
                this.PositionStateBits = lzmaSettings.PositionStateBits;
                this.LiteralContextBits = lzmaSettings.LiteralContextBits;
                this.LiteralPositionBits = lzmaSettings.LiteralPositionBits;
                this.FastBytes = lzmaSettings.FastBytes;
                this.MatchFinderType = lzmaSettings.MatchFinderType;
            }
        }

        /// <summary>
        /// Prepare settings for usage in zip archive entries.
        /// </summary>
        /// <param name="header">Central directory header.</param>
        internal override void PrepareForZip(CentralDirectoryHeader header = null)
        {
            base.PrepareForZip(header);

            this.UseZipHeader = true;

            if (header != null)
            {
                this.InternalStreamLength = header.UncompressedSize;
            }
        }
    }
}
