﻿using System;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Base class for the Store (no compression) transformation.
    /// </summary>
    internal class StoreTransformBase : BlockTransformBase
    {
        /// <summary>
        /// Initializes a new instance of the StoreTransformBase class.
        /// </summary>
        public StoreTransformBase()
        {
            this.FixedInputBlockSize = false;
        }

        /// <summary>Gets a value indicating whether the current transform can be reused.</summary>
        /// <returns>true if the current transform can be reused; otherwise, false.</returns>
        public override bool CanReuseTransform
        {
            get 
            { 
                return true; 
            }
        }

        /// <summary>Gets a value indicating whether multiple blocks can be transformed.</summary>
        /// <returns>true if multiple blocks can be transformed; otherwise, false.</returns>
        public override bool CanTransformMultipleBlocks
        {
            get
            {
                return true;
            }
        }

        /// <summary>Gets the input block size.</summary>
        /// <returns>The size of the input data blocks in bytes.</returns>
        public override int InputBlockSize
        {
            get
            {
                return 4096;
            }
        }

        /// <summary>Gets the output block size.</summary>
        /// <returns>The size of the output data blocks in bytes.</returns>
        public override int OutputBlockSize
        {
            get
            {
                return 4096;
            }
        }

        /// <summary>Transforms the specified region of the input byte array and copies the resulting transform to the specified region of the output byte array.</summary>
        /// <returns>The number of bytes written.</returns>
        /// <param name="inputBuffer">The input for which to compute the transform. </param>
        /// <param name="inputOffset">The offset into the input byte array from which to begin using data. </param>
        /// <param name="inputCount">The number of bytes in the input byte array to use as data. </param>
        /// <param name="outputBuffer">The output to which to write the transform. </param>
        /// <param name="outputOffset">The offset into the output byte array from which to begin writing data. </param>
        public override int TransformBlock(byte[] inputBuffer, int inputOffset, int inputCount, byte[] outputBuffer, int outputOffset)
        {
            this.ValidateParameters(inputBuffer, inputOffset, inputCount, outputBuffer, outputOffset, true);

            for (int i = 0; i < inputCount; i++)
            {
                outputBuffer[outputOffset + i] = inputBuffer[inputOffset + i];
            }

            return inputCount;
        }

        /// <summary>Transforms the specified region of the specified byte array.</summary>
        /// <returns>The computed transform.</returns>
        /// <param name="inputBuffer">The input for which to compute the transform. </param>
        /// <param name="inputOffset">The offset into the byte array from which to begin using data. </param>
        /// <param name="inputCount">The number of bytes in the byte array to use as data. </param>
        public override byte[] TransformFinalBlock(byte[] inputBuffer, int inputOffset, int inputCount)
        {
            this.ValidateInputBufferParameters(inputBuffer, inputOffset, inputCount, false, true);

            byte[] outputBuffer = new byte[inputCount];
            this.TransformBlock(inputBuffer, inputOffset, inputCount, outputBuffer, 0);
            return outputBuffer;
        }

        /// <summary>
        /// Releases the unmanaged resources used by the transform and optionally releases the managed resources. 
        /// </summary>
        /// <param name="disposing">Value which indicates whether both managed and unmanaged resources (true) on only unmanaged resources (false) should be released.</param>
        protected override void Dispose(bool disposing)
        {
        }
    }
}
