﻿using System;
using System.IO;
using System.Text;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Platform independent manager.
    /// </summary>
    internal class DefaultPlatformManager : IPlatformManager
    {
        /// <summary>
        /// Gets a platform-specific alternate character used to separate directory levels in a path string that reflects a hierarchical file system organization.
        /// </summary>
        public string AltDirectorySeparatorChar
        {
            get
            {
                return "/";
            }
        }

        /// <summary>
        /// Gets default encoding for this platform.
        /// </summary>
        public Encoding DefaultEncoding
        {
            get
            {
                return Encoding.UTF8;
            }
        }

        /// <summary>
        /// Gets a platform-specific character used to separate directory levels in a path string that reflects a hierarchical file system organization.
        /// </summary>
        public string DirectorySeparatorChar
        {
            get
            {
                return "\\";
            }
        }

        /// <summary>
        /// Creates temporary stream.
        /// </summary>
        /// <returns>Stream will be used for temporary operations.</returns>
        public Stream CreateTemporaryStream()
        {
            return new MemoryStream();
        }

        /// <summary>
        /// Deletes temporary stream.
        /// </summary>
        /// <param name="stream">Stream to delete.</param>
        public void DeleteTemporaryStream(Stream stream)
        {
        }

        /// <summary>
        /// Gets crypto provider initialized using given encryption settings.
        /// </summary>
        /// <param name="settings">Encryption settings.</param>
        /// <returns>Crypto provider.</returns>
        /// <exception cref="NotSupportedException">Specified crypto algorithm is not supported.</exception>
        public ICryptoProvider GetCryptoProvider(EncryptionSettings settings)
        {
            ICryptoProvider cryptoProvider = null;

            switch (settings.Algorithm.ToUpperInvariant())
            {
                case "DEFAULT":
                    cryptoProvider = new DefaultCryptoProvider();
                    break;

                default:
                    throw new NotSupportedException();
            }

            if (cryptoProvider != null)
            {
                cryptoProvider.Initialize(settings);
            }

            return cryptoProvider;
        }

        /// <summary>
        /// Indicates whether specified encoding is supported for this platform.
        /// </summary>
        /// <param name="encoding">Encoding.</param>
        /// <returns>true if encoding is allowed in the ZIP file.</returns>
        public bool IsEncodingSupported(Encoding encoding)
        {
            return (encoding == null)
                || !(encoding.Equals(Encoding.BigEndianUnicode)
                || encoding.Equals(Encoding.Unicode));
        }
    }
}