﻿using System;
using System.IO;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Cryptographic stream. Allows encrypt or decrypt information from the given input stream.
    /// </summary>
    internal class CryptoStream : OperationStream
    {
        private ICryptoProvider cryptoProvider;

        /// <summary>
        /// Initializes a new instance of the CryptoStream class.
        /// </summary>
        /// <param name="input">Input stream.</param>
        /// <param name="mode">Stream operational mode.</param>
        /// <param name="cryptoProvider">Crypto provider.</param>
        /// <exception cref="System.ArgumentOutOfRangeException">Specified mode is not allowed for the given stream.</exception>
        public CryptoStream(Stream input, StreamOperationMode mode, ICryptoProvider cryptoProvider)
            : base(input, mode)
        {
            if ((!input.CanRead && mode == StreamOperationMode.Read)
                || (!input.CanWrite && mode == StreamOperationMode.Write))
            {
                throw new ArgumentOutOfRangeException("mode");
            }

            this.cryptoProvider = cryptoProvider;

            switch (this.Mode)
            {
                case StreamOperationMode.Read:
                    this.Transform = this.cryptoProvider.CreateDecryptor();
                    break;

                case StreamOperationMode.Write:
                    this.Transform = this.cryptoProvider.CreateEncryptor();
                    break;
            }
        }

        /// <summary>
        /// Allows an object to try to free resources and perform other cleanup operations before it is reclaimed by garbage collection. 
        /// </summary>
        ~CryptoStream()
        {
            this.Dispose(false);
        }

        /// <summary>
        /// Releases the unmanaged resources used by the Stream and optionally releases the managed resources. 
        /// </summary>
        /// <param name="disposing">Value which indicates whether both managed and unmanaged resources (true) on only unmanaged resources (false) should be released.</param>
        protected override void Dispose(bool disposing)
        {
            this.cryptoProvider = null;
            base.Dispose(disposing);
        }
    }
}
