﻿using System;
using System.Text;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Base class for the transformations which implements traditional PKWARE encryption/decryption.
    /// </summary>
    internal abstract class DefaultCryptoTransformBase : BlockTransformBase
    {
        protected const int HeaderSize = 12;
        private uint[] encryptionKeys = { 0x12345678, 0x23456789, 0x34567890 };
        
        /// <summary>
        /// Initializes a new instance of the DefaultCryptoTransformBase class.
        /// </summary>
        public DefaultCryptoTransformBase()
        {
            this.FixedInputBlockSize = false;
            this.Header.Length = DefaultCryptoTransformBase.HeaderSize;
        }

        /// <summary>Gets a value indicating whether the current transform can be reused.</summary>
        /// <returns>true if the current transform can be reused; otherwise, false.</returns>
        public override bool CanReuseTransform
        {
            get 
            { 
                return true; 
            }
        }

        /// <summary>Gets a value indicating whether multiple blocks can be transformed.</summary>
        /// <returns>true if multiple blocks can be transformed; otherwise, false.</returns>
        public override bool CanTransformMultipleBlocks
        {
            get
            {
                return true;
            }
        }

        /// <summary>Gets the input block size.</summary>
        /// <returns>The size of the input data blocks in bytes.</returns>
        public override int InputBlockSize
        {
            get
            {
                return 12;
            }
        }

        /// <summary>Gets the output block size.</summary>
        /// <returns>The size of the output data blocks in bytes.</returns>
        public override int OutputBlockSize
        {
            get
            {
                return 12;
            }
        }

        /// <summary>
        /// Gets encoding byte.
        /// </summary>
        protected byte EncodingByte
        {
            get
            {
                uint t = (ushort)((uint)(this.encryptionKeys[2] & 0xFFFF) | 2);
                return (byte)((t * (t ^ 1)) >> 8);
            }
        }

        /// <summary>
        /// Creates transformation header to be written into the output stream.
        /// </summary>
        public override void CreateHeader()
        {
            Random random = new Random((int)DateTime.Now.Ticks);
            byte[] data = new byte[DefaultCryptoTransformBase.HeaderSize];
            random.NextBytes(data);
            if (this.Header.InitData != null)
            {
                data[DefaultCryptoTransformBase.HeaderSize - 1] = this.Header.InitData[0];
            }
            byte[] outputBuffer = new byte[DefaultCryptoTransformBase.HeaderSize];
            this.TransformBlock(data, 0, DefaultCryptoTransformBase.HeaderSize, outputBuffer, 0);
            this.Header.Buffer = outputBuffer;
        }

        /// <summary>
        /// Initialize reading of the transformation header.
        /// </summary>
        public override void InitHeaderReading()
        {
            this.Header.BytesToRead = DefaultCryptoTransformBase.HeaderSize;
        }

        /// <summary>
        /// Process transformation header has been read.
        /// </summary>
        public override void ProcessHeader()
        {
            byte[] outputBuffer = new byte[DefaultCryptoTransformBase.HeaderSize];
            this.TransformBlock(this.Header.Buffer, 0, this.Header.Buffer.Length, outputBuffer, 0);
            this.Header.BytesToRead = 0;
        }

        internal void InitKeys(string password)
        {
            byte[] buffer = Encoding.UTF8.GetBytes(password);
            foreach (byte byteValue in buffer)
            {
                this.UpdateKeys(byteValue);
            }
        }

        /// <summary>
        /// Releases the unmanaged resources used by the transform and optionally releases the managed resources. 
        /// </summary>
        /// <param name="disposing">Value which indicates whether both managed and unmanaged resources (true) on only unmanaged resources (false) should be released.</param>
        protected override void Dispose(bool disposing)
        {
        }

        /// <summary>
        /// Update encryption keys.
        /// </summary>
        /// <param name="byteValue">Byte.</param>
        protected void UpdateKeys(byte byteValue)
        {
            this.encryptionKeys[0] = Crc32.ComputeCrc32(encryptionKeys[0], byteValue);
            this.encryptionKeys[1] = encryptionKeys[1] + (byte)encryptionKeys[0];
            this.encryptionKeys[1] = (encryptionKeys[1] * 0x08088405) + 1;
            this.encryptionKeys[2] = Crc32.ComputeCrc32(encryptionKeys[2], (byte)(encryptionKeys[1] >> 24));
        }
    }
}
