﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Globalization;
using System.IO;

namespace Telerik.Windows.Zip
{
    /// <summary>
    /// Represents the ZipPackage class.
    /// </summary>
    [Obsolete("This class has been deprecated. Use ZipArchive instead.")]
    [EditorBrowsable(EditorBrowsableState.Never)]
    public class ZipPackage : ZipArchive
    {
        private const uint ArchiveAttribute = 32;

        private string zipFileName;
        private List<ZipPackageEntry> zipPackageEntries;

        /// <summary>
        /// Initializes a new instance of the ZipPackage class from the specified stream.
        /// </summary>
        /// <param name="stream">The stream that contains the archive to be read.</param>
        /// <param name="mode">Mode.</param>
        private ZipPackage(Stream stream, ZipArchiveMode mode)
            : base(stream, mode, false, null)
        {
        }

        /// <summary>
        /// Gets the file name for the ZipPackage.
        /// </summary>
        public string FileName
        {
            get
            {
                return this.zipFileName;
            }
        }

        /// <summary>
        /// Gets the collection of entries that are currently in the zip archive.
        /// </summary>
        /// <exception cref="NotSupportedException">The zip archive does not support reading.</exception>
        /// <exception cref="ObjectDisposedException">The zip archive has been disposed.</exception>
        /// <exception cref="InvalidDataException">The zip archive is corrupt, and its entries cannot be retrieved.</exception>
        public IList<ZipPackageEntry> ZipPackageEntries
        {
            get
            {
                List<ZipPackageEntry> list = new List<ZipPackageEntry>();
                foreach (ZipArchiveEntry entry in this.Entries)
                {
                    list.Add(new ZipPackageEntry(entry));
                }

                this.zipPackageEntries = list;
                return list;
            }
        }

        /// <summary>
        /// This method is used to create a ZipPackage from a stream.
        /// </summary>
        /// <param name="stream">Stream.</param>
        /// <returns>ZipPackage instance.</returns>
        public static ZipPackage Create(Stream stream)
        {
            ZipPackage zip = new ZipPackage(stream, ZipArchiveMode.Create);
            return zip;
        }

        /// <summary>
        /// This method is used to create a ZipPackage with the passed file name.
        /// </summary>
        /// <param name="fileName">File name.</param>
        /// <returns>ZipPackage instance.</returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Reliability", "CA2000:Dispose objects before losing scope")]
        public static ZipPackage CreateFile(string fileName)
        {
            Stream stream = new FileStream(fileName, FileMode.Create, FileAccess.ReadWrite, FileShare.None);
            ZipPackage package = ZipPackage.Create(stream);
            package.zipFileName = fileName;

            return package;
        }

        /// <summary>
        /// Checks whether the stream that represents a zip file is actually a zip file.
        /// </summary>
        /// <param name="stream">Stream.</param>
        /// <returns>True if the stream represents a zip file.</returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
        public static bool IsZipFile(Stream stream)
        {
            bool flag;
            if ((stream == null) || (stream.Length < 22))
            {
                return false;
            }

            long position = stream.Position;
            try
            {
                using (ZipPackage.Open(stream))
                {
                    flag = true;
                }
            }
            catch
            {
                flag = false;
            }
            finally
            {
                stream.Position = position;
            }

            return flag;
        }

        /// <summary>
        /// Checks whether the file with the passed file name is actually a zip file.
        /// </summary>
        /// <param name="fileName">File name.</param>
        /// <returns>True if the file represents a zip file.</returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
        public static bool IsZipFile(string fileName)
        {
            if (!File.Exists(fileName))
            {
                return false;
            }

            try
            {
                using (ZipPackage.OpenFile(fileName, FileAccess.Read))
                {
                    return true;
                }
            }
            catch
            {
                return false;
            }
        }

        /// <summary>
        /// Opens zip archive from the Stream.
        /// </summary>
        /// <param name="stream">Stream.</param>
        /// <returns>ZipPackage instance.</returns>
        public static ZipPackage Open(Stream stream)
        {
            ZipArchiveMode mode = stream.CanWrite ? ZipArchiveMode.Update : ZipArchiveMode.Read;
            ZipPackage zip = new ZipPackage(stream, mode);
            return zip;
        }

        /// <summary>
        /// This method is used to open a ZipPackage with the passed file name.
        /// </summary>
        /// <param name="fileName">File name.</param>
        /// <param name="access">File access.</param>
        /// <returns>ZipPackage instance.</returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Reliability", "CA2000:Dispose objects before losing scope")]
        public static ZipPackage OpenFile(string fileName, FileAccess access)
        {
            if (access != FileAccess.Read)
            {
                throw new InvalidOperationException("File should be opened with read access.");
            }

            Stream stream = new FileStream(fileName, FileMode.Open, access, FileShare.Read);
            ZipPackage package = ZipPackage.Open(stream);
            package.zipFileName = fileName;

            return package;
        }

        /// <summary>
        /// Adds a file with the passed file name in the ZipPackage.
        /// </summary>
        /// <param name="fileName">File name.</param>
        public void Add(string fileName)
        {
            string fileNameInZip = GetFileName(fileName);
            this.Add(fileName, fileNameInZip, CompressionType.Default);
        }

        /// <summary>
        /// Adds a file with the passed file name in the ZipPackage.
        /// </summary>
        public void Add(string fileName, CompressionType compressionType)
        {
            string fileNameInZip = GetFileName(fileName);
            this.Add(fileName, fileNameInZip, compressionType);
        }

        /// <summary>
        /// Adds the files from the passed IEnumerable of file names in the ZipPackage.
        /// </summary>
        public void Add(IEnumerable<string> fileNames)
        {
            foreach (string fileName in fileNames)
            {
                string fileNameInZip = GetFileName(fileName);
                this.Add(fileName, fileNameInZip, CompressionType.Default);
            }
        }

        /// <summary>
        /// Adds the files from the passed IEnumerable of file names in the ZipPackage.
        /// </summary>
        public void Add(IEnumerable<string> fileNames, CompressionType compressionType)
        {
            foreach (string fileName in fileNames)
            {
                string fileNameInZip = GetFileName(fileName);
                this.Add(fileName, fileNameInZip, compressionType);
            }
        }

        /// <summary>
        /// Adds a file with the passed file name in the ZipPackage and associates it with the passed file name in zip.
        /// </summary>
        public void Add(string fileName, string fileNameInZip)
        {
            this.Add(fileName, fileNameInZip, DateTime.Now, CompressionType.Default);
        }

        /// <summary>
        /// Adds a file with the passed file name in the ZipPackage and associates it with the passed file name in zip.
        /// </summary>
        public void Add(string fileName, string fileNameInZip, CompressionType compressionType)
        {
            this.Add(fileName, fileNameInZip, DateTime.Now, compressionType);
        }

        /// <summary>
        /// Adds a file with the passed file name in the ZipPackage, associates it with the passed file name in zip and sets a date time for the entry.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Globalization", "CA1309:UseOrdinalStringComparison", MessageId = "System.String.Compare(System.String,System.String,System.Globalization.CultureInfo,System.Globalization.CompareOptions)")]
        public void Add(string fileName, string fileNameInZip, DateTime dateTime)
        {
            if (string.Compare(fileName, this.FileName, CultureInfo.InvariantCulture, CompareOptions.IgnoreCase) == 0)
            {
                throw new ArgumentException("Can't add a zip file to itself.");
            }

            bool flag = File.Exists(fileName);
            bool flag2 = !flag && Directory.Exists(fileName);
            if (!flag && !flag2)
            {
                throw new FileNotFoundException(string.Format("File not found: '{0}'.", fileName));
            }

            this.AddEntry(fileName, fileNameInZip, dateTime, CompressionType.Default);
        }

        /// <summary>
        /// Adds a file with the passed file name in the ZipPackage, associates it with the passed file name in zip and sets a date time for the entry.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Globalization", "CA1309:UseOrdinalStringComparison", MessageId = "System.String.Compare(System.String,System.String,System.Globalization.CultureInfo,System.Globalization.CompareOptions)")]
        public void Add(string fileName, string fileNameInZip, DateTime dateTime, CompressionType compressionType)
        {
            if (string.Compare(fileName, this.FileName, CultureInfo.InvariantCulture, CompareOptions.IgnoreCase) == 0)
            {
                throw new ArgumentException("Can't add a zip file to itself.");
            }
            bool flag = File.Exists(fileName);
            bool flag2 = !flag && Directory.Exists(fileName);
            if (!flag && !flag2)
            {
                throw new FileNotFoundException(string.Format("File not found: '{0}'.", fileName));
            }

            this.AddEntry(fileName, fileNameInZip, dateTime, compressionType);
        }

        /// <summary>
        /// Adds a stream in the ZipPackage and associates it with the passed file name in zip.
        /// </summary>
        /// <param name="stream">Stream.</param>
        /// <param name="fileNameInZip">File name in zip archive.</param>
        public void AddStream(Stream stream, string fileNameInZip)
        {
            this.AddStream(stream, fileNameInZip, ZipCompression.Default, DateTime.Now, CompressionType.Default);
        }

        /// <summary>
        /// Adds a stream in the ZipPackage and associates it with the passed file name in zip.
        /// </summary>
        /// <param name="stream">Stream.</param>
        /// <param name="fileNameInZip">File name in zip archive.</param>
        /// <param name="compressionType">Compression type.</param>
        public void AddStream(Stream stream, string fileNameInZip, CompressionType compressionType)
        {
            this.AddStream(stream, fileNameInZip, ZipCompression.Default, DateTime.Now, compressionType);
        }

        /// <summary>
        /// Adds a stream in the ZipPackage, compresses it with the passed compress method,
        /// associates it with the passed file name in zip and sets a date time for the entry.
        /// </summary>
        /// <param name="stream">Stream.</param>
        /// <param name="fileNameInZip">File name in zip archive.</param>
        /// <param name="method">Compression level.</param>
        /// <param name="dateTime">Date and time of file.</param>
        public void AddStream(Stream stream, string fileNameInZip, ZipCompression method, DateTime dateTime)
        {
            this.AddStream(stream, fileNameInZip, method, dateTime, CompressionType.Default);
        }

        /// <summary>
        /// Adds a stream in the ZipPackage and associates it with the passed file name in zip.
        /// </summary>
        /// <param name="stream">Stream.</param>
        /// <param name="fileNameInZip">File name in zip archive.</param>
        /// <param name="method">Compression level.</param>
        /// <param name="dateTime">Date and time of file.</param>
        /// <param name="compressionType">Compression type.</param>
        public void AddStream(Stream stream, string fileNameInZip, ZipCompression method, DateTime dateTime, CompressionType compressionType)
        {
            if (compressionType == CompressionType.Lzma)
            {
                throw new NotSupportedException();
            }

            DeflateSettings settings = ZipOutputStream.CreateDeflateSettings(method);
            settings.HeaderType = CompressedStreamHeader.None;

            this.DeleteAvailableEntry(fileNameInZip);

            using (var entry = this.CreateEntry(fileNameInZip, settings))
            {
                entry.LastWriteTime = dateTime;
                entry.ExternalAttributes = (int)ZipPackage.ArchiveAttribute;
                Stream entryStream = entry.Open();
                stream.CopyTo(entryStream);
            }
        }

        /// <summary>
        /// Closes the ZipPackage.
        /// </summary>
        /// <param name="closeStream">If the parameter is set to true then closes the file.</param>
        public void Close(bool closeStream)
        {
            this.DisposeStreams(closeStream);
        }

        /// <summary>
        /// Gets the index of the entry in the list of entries of the ZipPackage.
        /// </summary>
        /// <param name="fileNameInZip">File name in zip archive.</param>
        /// <returns>Index of entry or -1 when the entry is not found.</returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Globalization", "CA1309:UseOrdinalStringComparison", MessageId = "System.String.Compare(System.String,System.String,System.Globalization.CultureInfo,System.Globalization.CompareOptions)")]
        public int IndexOf(string fileNameInZip)
        {
            int i = 0;
            var entries = this.ZipPackageEntries;
            foreach (ZipPackageEntry entry in entries)
            {
                if (string.Compare(entry.FileNameInZip, fileNameInZip, CultureInfo.InvariantCulture, CompareOptions.IgnoreCase) == 0)
                {
                    return i;
                }

                i++;
            }

            return -1;
        }

        /// <summary>
        /// Removes the passed entry from the ZipPackage.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1822:MarkMembersAsStatic")]
        public void RemoveEntry(ZipPackageEntry zipPackageEntry)
        {
            zipPackageEntry.Delete();
        }

        private static string GetFileName(string fileName)
        {
            fileName = Path.GetFileName(fileName);
            return fileName;
        }

        private void AddEntry(string fileName, string fileNameInZip, DateTime dateTime, CompressionType compressionType)
        {
            if (compressionType == CompressionType.Lzma)
            {
                throw new NotSupportedException();
            }

            if (File.Exists(fileName))
            {
                using (FileStream stream = new FileStream(fileName, FileMode.Open, FileAccess.Read, FileShare.ReadWrite))
                {
                    this.SaveEntry(stream, fileName, fileNameInZip, dateTime);
                    return;
                }
            }
            else if (Directory.Exists(fileName))
            {
                // save folder
                this.SaveEntry(null, fileName, fileNameInZip, dateTime);
            }
            else
            {
                throw new FileNotFoundException(string.Format("File not found: '{0}'.", fileName));
            }
        }

        private void DeleteAvailableEntry(string fileNameInZip)
        {
            int availableEntryIndex = this.IndexOf(fileNameInZip);
            if (availableEntryIndex != -1)
            {
                this.zipPackageEntries[availableEntryIndex].Delete();
            }
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Usage", "CA1801:ReviewUnusedParameters", MessageId = "dateTime")]
        private void SaveEntry(FileStream stream, string fileName, string fileNameInZip, DateTime dateTime)
        {
            FileInfo info = new FileInfo(fileName);
            int attributes = (int)info.Attributes;

            this.DeleteAvailableEntry(fileNameInZip);

            using (var entry = this.CreateEntry(fileNameInZip,
            new DeflateSettings()
            {
                CompressionLevel = CompressionLevel.Optimal,
                HeaderType = CompressedStreamHeader.None
            }))
            {
                Stream entryStream = entry.Open();
                if (stream != null)
                {
                    stream.CopyTo(entryStream);
                }

                entry.LastWriteTime = info.LastWriteTime;
                entry.ExternalAttributes = attributes;
            }
        }
    }
}
