﻿using System;
using System.IO;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Operational stream. Base class for cryptographic and compression streams.
    /// </summary>
    public abstract class OperationStream : Stream
    {
        private Stream baseStream;
        private StreamOperationMode streamMode;
        private IBlockTransform blockTransform = null;

        private byte[] inputBuffer;
        private int inputBufferIndex;

        private byte[] outputBuffer;
        private int outputBufferIndex;

        private bool finalBlockTransformed = false;

        private long totalPlainCount = 0;
        private long totalTransformedCount = 0;

        private long? length;
        private long startPosition = 0;

        /// <summary>
        /// Initializes a new instance of the OperationStream class.
        /// </summary>
        /// <param name="baseStream">The base input/output stream.</param>
        /// <param name="mode">Stream operational mode.</param>
        /// <exception cref="System.ArgumentOutOfRangeException">Specified mode is not allowed for the given stream.</exception>
        internal OperationStream(Stream baseStream, StreamOperationMode mode)
        {
            if ((!baseStream.CanRead && mode == StreamOperationMode.Read)
                || (!baseStream.CanWrite && mode == StreamOperationMode.Write))
            {
                throw new ArgumentOutOfRangeException("mode");
            }

            this.BaseStream = baseStream;
            this.streamMode = mode;
            this.IsDisposed = false;
        }

        /// <summary>
        /// Allows an object to try to free resources and perform other cleanup operations before it is reclaimed by garbage collection. 
        /// </summary>
        ~OperationStream()
        {
            this.Dispose(false);
        }

        /// <summary>
        /// Gets a value indicating whether the current stream supports reading.
        /// </summary>
        public override bool CanRead
        {
            get
            {
                return this.baseStream != null && this.streamMode == StreamOperationMode.Read;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the current stream supports seeking.
        /// </summary>
        public override bool CanSeek
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the current stream supports writing.
        /// </summary>
        public override bool CanWrite
        {
            get
            {
                return this.baseStream != null && this.streamMode == StreamOperationMode.Write;
            }
        }

        /// <summary>Gets a value which indicates whether the final buffer block has been written/read to/from the underlying stream. </summary>
        /// <returns>true if the final block has been flushed or end of underlying stream is reached; otherwise, false. </returns>
        public bool HasFlushedFinalBlock
        {
            get
            {
                return this.finalBlockTransformed;
            }
        }

        /// <summary>
        /// Gets the length in bytes of the stream.
        /// </summary>
        public override long Length
        {
            get
            {
                if (this.BaseStream != null)
                {
                    if (this.BaseStream.CanSeek)
                    {
                        if (this.Mode == StreamOperationMode.Read)
                        {
                            return this.length.Value;
                        }
                        else
                        {
                            return this.BaseStream.Length;
                        }
                    }
                    else
                    {
                        if (this.length.HasValue)
                        {
                            return this.length.Value;
                        }
                        else
                        {
                            return long.MaxValue;
                        }
                    }
                }
                else
                {
                    throw new NotSupportedException();
                }
            }
        }

        /// <summary>
        /// Gets or sets the position within the current stream.
        /// </summary>
        public override long Position
        {
            get
            {
                if (this.BaseStream != null)
                {
                    if (this.BaseStream.CanSeek)
                    {
                        if (this.Mode == StreamOperationMode.Read)
                        {
                            return this.BaseStream.Position - this.startPosition;
                        }
                        else
                        {
                            return this.BaseStream.Position;
                        }
                    }
                    else
                    {
                        return 0;
                    }
                }
                else
                {
                    throw new NotSupportedException();
                }
            }

            set
            {
                throw new NotSupportedException();
            }
        }

        /// <summary>
        /// Gets value which specify total plain bytes count (not-compressed and not-encrypted).
        /// </summary>
        public long TotalPlainCount
        {
            get
            {
                return this.totalPlainCount;
            }

            private set
            {
                this.totalPlainCount = value;
            }
        }

        /// <summary>
        /// Gets value which specify total transformed bytes count (compressed or encrypted).
        /// </summary>
        public long TotalTransformedCount
        {
            get
            {
                return this.totalTransformedCount;
            }

            private set
            {
                this.totalTransformedCount = value;
            }
        }

        /// <summary>
        /// Gets input stream.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
        internal Stream BaseStream
        {
            get
            {
                return this.baseStream;
            }

            set
            {
                this.baseStream = value;
                if (this.baseStream != null)
                {
                    if (this.Mode == StreamOperationMode.Read)
                    {
                        if (this.baseStream.CanSeek)
                        {
                            this.startPosition = this.baseStream.Position;
                            try
                            {
                                this.length = this.baseStream.Length;
                            }
                            catch
                            {
                                this.length = 0;
                            }
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Gets stream mode.
        /// </summary>
        internal StreamOperationMode Mode
        {
            get
            {
                return this.streamMode;
            }
        }

        /// <summary>
        /// Gets or sets value which indicates whether this stream is disposed already.
        /// </summary>
        internal bool IsDisposed
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets block transformation is used for read/write operations.
        /// </summary>
        internal IBlockTransform Transform
        {
            get
            {
                return this.blockTransform;
            }

            set
            {
                this.blockTransform = value;
                this.InitializeTransform();
                this.InitializeBuffers();
            }
        }

        /// <summary>
        /// Clears all buffers for this stream and causes any buffered data to be written to the underlying device.
        /// </summary>
        public override void Flush()
        {
            this.FlushFinalBlock();
        }

        /// <summary>
        /// Reads a sequence of bytes from the current stream and advances the position within the stream by the number of bytes read.
        /// </summary>
        /// <param name="buffer">An array of bytes. When this method returns, the buffer contains the specified byte array with the 
        /// values between offset and (offset + count - 1) replaced by the bytes read from the current source.</param>
        /// <param name="offset">The zero-based byte offset in buffer at which to begin storing the data read from the current stream.</param>
        /// <param name="count">The maximum number of bytes to be read from the current stream. </param>
        /// <returns>The total number of bytes read into the buffer. This can be less than the number of bytes requested if that many 
        /// bytes are not currently available, or zero (0) if the end of the stream has been reached.</returns>
        /// <exception cref="T:System.NotSupportedException">The <see cref="StreamOperationMode" /> associated with 
        /// current <see cref="OperationStream" /> object does not match the underlying stream.  
        /// For example, this exception is thrown when using <see cref="StreamOperationMode.Read" /> with an underlying stream that is write only.  </exception>
        /// <exception cref="T:System.ArgumentOutOfRangeException">The <paramref name="offset" /> parameter is less than zero.-or- The <paramref name="count" /> parameter is less than zero. </exception>
        /// <exception cref="T:System.ArgumentException">The sum of the <paramref name="count" /> and <paramref name="offset" /> parameters is larger than the length of the buffer. </exception>
        public override int Read(byte[] buffer, int offset, int count)
        {
            this.EnsureNotDisposed();
            OperationStream.ValidateBufferParameters(buffer, offset, count, true);
            if (!this.CanRead)
            {
                throw new NotSupportedException();
            }

            int bytesTransformed;
            int bytesLeft = count;
            int currentOffset = offset;

            if (this.outputBufferIndex != 0)
            {
                if (this.outputBufferIndex > count)
                {
                    // Output buffer contains required quantity of the transformed bytes. 
                    Buffer.BlockCopy(this.outputBuffer, 0, buffer, offset, count);
                    Buffer.BlockCopy(this.outputBuffer, count, this.outputBuffer, 0, this.outputBufferIndex - count);
                    this.outputBufferIndex = this.outputBufferIndex - count;
                    return count;
                }
                else
                {
                    // Output buffer contains just part of the required quantity of the transformed bytes.
                    Buffer.BlockCopy(this.outputBuffer, 0, buffer, offset, this.outputBufferIndex);
                    bytesLeft = bytesLeft - this.outputBufferIndex;
                    currentOffset = currentOffset + this.outputBufferIndex;
                    this.outputBufferIndex = 0;
                }
            }

            if (this.finalBlockTransformed)
            {
                return count - bytesLeft;
            }

            int bytesInBuffer = 0;
            if (bytesLeft > this.Transform.OutputBlockSize && this.Transform.CanTransformMultipleBlocks)
            {
                int readBufferSize = (bytesLeft / this.Transform.OutputBlockSize) * this.Transform.InputBlockSize;
                long bytesLeftInStream = this.Length - this.Position;
                readBufferSize = (int)Math.Min((long)readBufferSize, bytesLeftInStream);
                byte[] readBuffer = new byte[readBufferSize];

                Buffer.BlockCopy(this.inputBuffer, 0, readBuffer, 0, this.inputBufferIndex);

                bytesInBuffer = this.inputBufferIndex + this.baseStream.Read(readBuffer, this.inputBufferIndex, readBufferSize - this.inputBufferIndex);
                this.TotalTransformedCount += bytesInBuffer - this.inputBufferIndex;

                this.inputBufferIndex = 0;

                if (bytesInBuffer > this.Transform.InputBlockSize)
                {
                    int restPos = bytesInBuffer / this.Transform.InputBlockSize * this.Transform.InputBlockSize;
                    int restCount = bytesInBuffer - restPos;
                    if (restCount != 0)
                    {
                        this.inputBufferIndex = restCount;
                        Buffer.BlockCopy(readBuffer, restPos, this.inputBuffer, 0, restCount);
                    }

                    int transformBufferSize = (restPos / this.Transform.InputBlockSize) * this.Transform.OutputBlockSize;
                    byte[] transformBuffer = new byte[transformBufferSize];

                    bytesTransformed = this.Transform.TransformBlock(readBuffer, 0, restPos, transformBuffer, 0);
                    this.TotalPlainCount += bytesTransformed;

                    Buffer.BlockCopy(transformBuffer, 0, buffer, currentOffset, bytesTransformed);

                    Array.Clear(readBuffer, 0, (int)readBuffer.Length);
                    Array.Clear(transformBuffer, 0, (int)transformBuffer.Length);

                    bytesLeft = bytesLeft - bytesTransformed;
                    currentOffset = currentOffset + bytesTransformed;
                }
                else
                {
                    this.inputBuffer = readBuffer;
                    this.inputBufferIndex = bytesInBuffer;
                }
            }

            while (bytesLeft > 0)
            {
                while (this.inputBufferIndex < this.Transform.InputBlockSize)
                {
                    long bytesLeftInStream = this.Length - this.Position;
                    int bytesToRead = this.Transform.InputBlockSize - this.inputBufferIndex;
                    bytesToRead = (int)Math.Min((long)bytesToRead, bytesLeftInStream);
                    bytesInBuffer = this.baseStream.Read(this.inputBuffer, this.inputBufferIndex, bytesToRead);
                    this.TotalTransformedCount += bytesInBuffer;
                    if (bytesInBuffer == 0)
                    {
                        // We reach end of the base stream.
                        byte[] transformBuffer = this.Transform.TransformFinalBlock(this.inputBuffer, 0, this.inputBufferIndex);
                        this.TotalPlainCount += transformBuffer.Length;
                        this.outputBuffer = transformBuffer;
                        this.outputBufferIndex = transformBuffer.Length;
                        this.finalBlockTransformed = true;
                        if (bytesLeft >= this.outputBufferIndex)
                        {
                            Buffer.BlockCopy(this.outputBuffer, 0, buffer, currentOffset, this.outputBufferIndex);
                            bytesLeft = bytesLeft - this.outputBufferIndex;
                            this.outputBufferIndex = 0;
                            return count - bytesLeft;
                        }
                        else
                        {
                            Buffer.BlockCopy(this.outputBuffer, 0, buffer, currentOffset, bytesLeft);
                            this.outputBufferIndex = this.outputBufferIndex - bytesLeft;
                            Buffer.BlockCopy(this.outputBuffer, bytesLeft, this.outputBuffer, 0, this.outputBufferIndex);
                            return count;
                        }
                    }
                    this.inputBufferIndex = this.inputBufferIndex + bytesInBuffer;
                }

                bytesTransformed = this.Transform.TransformBlock(this.inputBuffer, 0, this.Transform.InputBlockSize, this.outputBuffer, 0);
                this.TotalPlainCount += bytesTransformed;
                this.inputBufferIndex = 0;
                if (bytesLeft < bytesTransformed)
                {
                    // We read transformed enough data to return.
                    Buffer.BlockCopy(this.outputBuffer, 0, buffer, currentOffset, bytesLeft);
                    this.outputBufferIndex = bytesTransformed - bytesLeft;
                    Buffer.BlockCopy(this.outputBuffer, bytesLeft, this.outputBuffer, 0, this.outputBufferIndex);
                    return count;
                }

                Buffer.BlockCopy(this.outputBuffer, 0, buffer, currentOffset, bytesTransformed);
                currentOffset = currentOffset + bytesTransformed;
                bytesLeft = bytesLeft - bytesTransformed;
            }

            return count;
        }

        /// <summary>
        /// Sets the position within the current stream. 
        /// </summary>
        /// <param name="offset">A byte offset relative to the origin parameter.</param>
        /// <param name="origin">A value of type SeekOrigin indicating the reference point used to obtain the new position.</param>
        /// <returns>The new position within the current stream.</returns>
        public override long Seek(long offset, SeekOrigin origin)
        {
            throw new NotSupportedException();
        }

        /// <summary>
        /// Sets the length of the current stream.
        /// </summary>
        /// <param name="value">The desired length of the current stream in bytes. </param>
        public override void SetLength(long value)
        {
            if (this.Mode == StreamOperationMode.Read)
            {
                this.length = value;
            }
            else
            {
                throw new NotSupportedException();
            }
        }

        /// <summary>
        /// Writes a sequence of bytes to the current stream and advances the current position within this stream by the number of bytes written.
        /// </summary>
        /// <param name="buffer">An array of bytes. This method copies count bytes from buffer to the current stream. </param>
        /// <param name="offset">The zero-based byte offset in buffer at which to begin copying bytes to the current stream. </param>
        /// <param name="count">The number of bytes to be written to the current stream.</param>
        public override void Write(byte[] buffer, int offset, int count)
        {
            this.EnsureNotDisposed();
            OperationStream.ValidateBufferParameters(buffer, offset, count, true);
            if (!this.CanWrite)
            {
                throw new NotSupportedException();
            }

            this.TotalPlainCount += count;

            int bytesTransformed;
            int bytesLeft = count;
            int currentOffset = offset;

            if (this.inputBufferIndex > 0)
            {
                if (count < this.Transform.InputBlockSize - this.inputBufferIndex)
                {
                    Buffer.BlockCopy(buffer, offset, this.inputBuffer, this.inputBufferIndex, count);
                    this.inputBufferIndex = this.inputBufferIndex + count;
                    return;
                }
                else
                {
                    Buffer.BlockCopy(buffer, offset, this.inputBuffer, this.inputBufferIndex, this.Transform.InputBlockSize - this.inputBufferIndex);
                    currentOffset = currentOffset + (this.Transform.InputBlockSize - this.inputBufferIndex);
                    bytesLeft = bytesLeft - (this.Transform.InputBlockSize - this.inputBufferIndex);
                    this.inputBufferIndex = this.Transform.InputBlockSize;
                }
            }

            if (this.outputBufferIndex > 0)
            {
                this.baseStream.Write(this.outputBuffer, 0, this.outputBufferIndex);
                this.outputBufferIndex = 0;
            }

            if (this.inputBufferIndex == this.Transform.InputBlockSize)
            {
                bytesTransformed = this.Transform.TransformBlock(this.inputBuffer, 0, this.Transform.InputBlockSize, this.outputBuffer, 0);
                this.TotalTransformedCount += bytesTransformed;
                this.baseStream.Write(this.outputBuffer, 0, bytesTransformed);
                this.inputBufferIndex = 0;
            }

            while (bytesLeft > 0)
            {
                if (bytesLeft < this.Transform.InputBlockSize)
                {
                    Buffer.BlockCopy(buffer, currentOffset, this.inputBuffer, 0, bytesLeft);
                    this.inputBufferIndex = this.inputBufferIndex + bytesLeft;
                    return;
                }

                if (!this.Transform.CanTransformMultipleBlocks)
                {
                    bytesTransformed = this.Transform.TransformBlock(buffer, currentOffset, this.Transform.InputBlockSize, this.outputBuffer, 0);
                    this.TotalTransformedCount += bytesTransformed;
                    this.baseStream.Write(this.outputBuffer, 0, bytesTransformed);
                    currentOffset = currentOffset + this.Transform.InputBlockSize;
                    bytesLeft = bytesLeft - this.Transform.InputBlockSize;
                }
                else
                {
                    int blocksCount = bytesLeft / this.Transform.InputBlockSize;
                    int bytesToTransform = blocksCount * this.Transform.InputBlockSize;
                    byte[] transformBuffer = new byte[blocksCount * this.Transform.OutputBlockSize];
                    bytesTransformed = this.Transform.TransformBlock(buffer, currentOffset, bytesToTransform, transformBuffer, 0);
                    this.TotalTransformedCount += bytesTransformed;
                    this.baseStream.Write(transformBuffer, 0, bytesTransformed);
                    currentOffset = currentOffset + bytesToTransform;
                    bytesLeft = bytesLeft - bytesToTransform;
                }
            }
        }

        /// <summary>
        /// Validate read/write operation parameters.
        /// </summary>
        /// <param name="buffer">Operation buffer.</param>
        /// <param name="offset">Offset.</param>
        /// <param name="count">Count.</param>
        /// <param name="allowZeroCount">Indicates whether count can be zero.</param>
        internal static void ValidateBufferParameters(byte[] buffer, int offset, int count, bool allowZeroCount)
        {
            if (buffer == null)
            {
                throw new ArgumentNullException("buffer");
            }
            if (offset < 0)
            {
                throw new ArgumentOutOfRangeException("offset");
            }
            if (count < 0 || (count == 0 && !allowZeroCount))
            {
                throw new ArgumentOutOfRangeException("count");
            }
            if ((int)buffer.Length - offset < count)
            {
                throw new ArgumentException("InvalidArgumentOffsetCount");
            }
        }

        /// <summary>Updates the underlying data source or repository with the current state of the buffer, then clears the buffer.</summary>
        /// <exception cref="NotSupportedException">The current stream is not writable.-or- The final block has already been transformed. </exception>
        internal virtual void FlushFinalBlock()
        {
            if (this.finalBlockTransformed)
            {
                throw new NotSupportedException("Can't flush final block twice");
            }

            byte[] transformedBlock = new byte[0];
            if (this.inputBuffer != null)
            {
                transformedBlock = this.Transform.TransformFinalBlock(this.inputBuffer, 0, this.inputBufferIndex);
            }

            this.finalBlockTransformed = true;
            this.TotalTransformedCount += transformedBlock.Length;

            if (this.baseStream != null && this.CanWrite)
            {
                if (this.outputBufferIndex > 0)
                {
                    this.baseStream.Write(this.outputBuffer, 0, this.outputBufferIndex);
                    this.outputBufferIndex = 0;
                }

                this.baseStream.Write(transformedBlock, 0, (int)transformedBlock.Length);
            }

            OperationStream operationStream = this.baseStream as OperationStream;
            if (operationStream == null)
            {
                try
                {
                    if (this.baseStream.CanRead)
                    {
                        this.baseStream.Flush();
                    }
                }
                catch (NotSupportedException)
                {
                }
            }
            else if (!operationStream.HasFlushedFinalBlock)
            {
                if (operationStream.CanWrite)
                {
                    operationStream.FlushFinalBlock();
                }
            }

            if (this.inputBuffer != null)
            {
                Array.Clear(this.inputBuffer, 0, (int)this.inputBuffer.Length);
            }

            if (this.outputBuffer != null)
            {
                Array.Clear(this.outputBuffer, 0, (int)this.outputBuffer.Length);
            }
        }

        /// <summary>
        /// Ensure that current stream is not disposed.
        /// </summary>
        protected void EnsureNotDisposed()
        {
            if (this.baseStream == null)
            {
                throw new ObjectDisposedException(null, "Stream closed");
            }
        }

        /// <summary>
        /// Releases the unmanaged resources used by the Stream and optionally releases the managed resources. 
        /// </summary>
        /// <param name="disposing">Value which indicates whether both managed and unmanaged resources (true) on only unmanaged resources (false) should be released.</param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Usage", "CA2215:Dispose methods should call base class dispose")]
        protected override void Dispose(bool disposing)
        {
            if (!this.IsDisposed)
            {
                try
                {
                    if (disposing)
                    {
                        if (!this.finalBlockTransformed && this.Transform != null)
                        {
                            this.FlushFinalBlock();
                        }

                        OperationStream operationStream = this.baseStream as OperationStream;
                        if (operationStream != null)
                        {
                            operationStream.Dispose();
                        }
                    }
                }
                finally
                {
                    try
                    {
                        this.finalBlockTransformed = true;

                        if (this.Transform != null)
                        {
                            this.Transform.Dispose();
                        }

                        if (this.inputBuffer != null)
                        {
                            Array.Clear(this.inputBuffer, 0, (int)this.inputBuffer.Length);
                        }

                        if (this.outputBuffer != null)
                        {
                            Array.Clear(this.outputBuffer, 0, (int)this.outputBuffer.Length);
                        }

                        this.inputBuffer = null;
                        this.outputBuffer = null;
                        this.blockTransform = null;
                        this.baseStream = null;
                    }
                    finally
                    {
                        this.IsDisposed = true;
                        base.Dispose(disposing);
                    }
                }
            }
        }

        /// <summary>
        /// Initialize internal buffers.
        /// </summary>
        private void InitializeBuffers()
        {
            if (this.blockTransform != null)
            {
                this.inputBuffer = new byte[this.blockTransform.InputBlockSize];
                this.outputBuffer = new byte[this.blockTransform.OutputBlockSize];
            }
        }

        /// <summary>
        /// Initialize transformation.
        /// </summary>
        private void InitializeTransform()
        {
            switch (this.Mode)
            {
                case StreamOperationMode.Read:
                    this.ReadTransformationHeader();
                    break;

                case StreamOperationMode.Write:
                    this.WriteTransformationHeader();
                    break;
            }
        }

        /// <summary>
        /// Read transformation header.
        /// </summary>
        private void ReadTransformationHeader()
        {
            this.Transform.InitHeaderReading();
            while (this.Transform.Header.BytesToRead > 0)
            {
                int newSize = this.Transform.Header.Buffer != null ?
                    this.Transform.Header.Buffer.Length + this.Transform.Header.BytesToRead :
                    this.Transform.Header.BytesToRead;

                byte[] buffer = new byte[newSize];
                int offset = 0;
                if (this.Transform.Header.Buffer != null)
                {
                    Array.Copy(this.Transform.Header.Buffer, buffer, this.Transform.Header.Buffer.Length);
                    offset = this.Transform.Header.Buffer.Length;
                }
                this.BaseStream.Read(buffer, offset, this.Transform.Header.BytesToRead);
                this.Transform.Header.Buffer = buffer;
                this.Transform.ProcessHeader();
            }
        }

        /// <summary>
        /// Write transformation header.
        /// </summary>
        private void WriteTransformationHeader()
        {
            this.Transform.CreateHeader();
            if (this.Transform.Header.Buffer != null)
            {
                this.BaseStream.Write(this.Transform.Header.Buffer, 0, this.Transform.Header.Buffer.Length);
            }
        }
    }
}
