﻿using System;
using System.Windows;
#if WPF
using System.Collections.Generic;
using System.Windows.Controls;
using System.Windows.Threading;
#endif

namespace Telerik.Windows.Zip.Tests
{
	public class AsyncUITest<T> : WindowedTest<T> where T : FrameworkElement, new()
	{
		private const int AsyncCallbackDelay = 100;

		public static Action Sleep100 = () => new Object();
		public static Action Sleep1000 = () => new Object();
#if WPF

		private const int DefaultTickInterval = 1;

		private Queue<AsyncWorkItem> itemQueue = new Queue<AsyncWorkItem>();

		public bool RemainingWork
		{
			get
			{
				return this.itemQueue.Count > 0;
			}
		}

		public DispatcherTimer Timer
		{
			get;
			set;
		}

		public Panel TestPanel
		{
			get;
			set;
		}

		public virtual void EnqueueAsyncWorkItem(AsyncWorkItem workItem)
		{
			this.itemQueue.Enqueue(workItem);
		}

		public virtual void EnqueueCallback(Action testCallbackDelegate)
		{
			this.EnqueueAsyncWorkItem(new AsyncCallbackWorkItem(testCallbackDelegate));
		}

		public virtual void EnqueueSleep(int sleepMillisecondsMinimum)
		{
			this.EnqueueAsyncWorkItem(new AsyncSleepWorkItem(sleepMillisecondsMinimum));
		}

		public virtual void EnqueueConditional(Func<bool> conditionalDelegate)
		{
			this.EnqueueAsyncWorkItem(new AsyncConditionalWorkItem(conditionalDelegate));
		}

		public virtual void EnqueueTestComplete()
		{
			this.Timer = new DispatcherTimer();
			this.Timer.Interval = new TimeSpan(0, 0, 0, 0, DefaultTickInterval);
			this.Timer.Tick += new EventHandler(this.OnTimerTick);
			this.Timer.Start();

			while (this.RemainingWork)
			{
				DoEvents();
			}
		}

		void OnTimerTick(object sender, EventArgs e)
		{
			this.Timer.Stop();

			AsyncWorkItem item = this.itemQueue.Peek();
			item.Invoke();

			if (item.IsComplete)
				this.itemQueue.Dequeue();

			if (this.RemainingWork)
				this.Timer.Start();
		}
#endif

		public virtual void EnqueueCallbacks(params Action[] testCallbackDelegates)
		{
			foreach (var item in testCallbackDelegates)
			{
				if (item == Sleep100)
				{
					this.EnqueueSleep(100);
				}
				else if (item == Sleep1000)
				{
					this.EnqueueSleep(1000);
				}
				else
				{
					this.EnqueueCallback(item);
				}
			}
		}

		public bool IsTestElementLoaded
		{
			get;
			set;
		}

		protected void LoadTestElement()
		{
			this.EnqueueCallback(() => this.TestElement.Loaded += delegate { this.IsTestElementLoaded = true; });
			this.EnqueueCallback(() => this.TestPanel.Children.Add(this.TestElement));
			this.EnqueueConditional(() => this.IsTestElementLoaded);
			this.EnqueueCallback(() => this.TestPanel.UpdateLayout());
		}

		protected void ClearTestElement()
		{
			this.EnqueueCallback(() => this.TestPanel.Children.Remove(this.TestElement));
			this.EnqueueCallback(() => this.IsTestElementLoaded = false);
		}

		protected virtual void PerformActions(Action[] actions)
		{
			foreach (Action action in actions)
			{
				Action capturedAction = action;
				this.PerformAction(capturedAction);
				this.PerformDoEvents();
			}
		}

		protected virtual void PerformPreLoadAction(Action action)
		{
			if (action == null)
				return;

			this.PerformAction(action);
		}

		protected virtual void PerformDoEvents()
		{
#if SILVERLIGHT
			this.EnqueueCallback(() => this.TestPanel.UpdateLayout());
			this.EnqueueSleep(AsyncCallbackDelay);
#endif

#if WPF
			this.EnqueueCallback(() => DoEvents());
#endif
		}

		protected virtual void PerformAction(Action action)
		{
			this.EnqueueCallback(() => action());
		}

		protected virtual void CreateAsyncTask(Action preLoadAction, params Action[] actions)
		{
			if (actions == null)
				throw new ArgumentNullException("actions", "Task cannot run without a specified set of actions");

			this.PerformPreLoadAction(preLoadAction);
			this.LoadTestElement();
			this.PerformActions(actions);
			this.ClearTestElement();
		}

		protected virtual void CreateAsyncTest(Action preLoadAction, params Action[] actions)
		{
#if WPF
			this.Window.Show();
#endif
			this.CreateAsyncTask(preLoadAction, actions);
			this.EnqueueTestComplete();
		}

		protected virtual void CreateTest(params Action[] actions)
		{
			this.CreateAsyncTest(null, actions);
		}

		public override void TestInitialize()
		{
			this.InitializeTestPanel();
			this.CreateTestElement();
#if (WPF)
			this.Window.Content = this.TestPanel;
#endif
		}

		public void InitializeTestPanel()
		{
#if (WPF)
			this.TestPanel = new Grid();
#endif
			this.TestPanel.Width = 800;
			this.TestPanel.Height = 600;
			this.TestPanel.HorizontalAlignment = HorizontalAlignment.Left;
			this.TestPanel.VerticalAlignment = VerticalAlignment.Top;
		}

		public void CreateTestElement()
		{
			this.TestElement = new T();
			this.IsTestElementLoaded = false;
		}

		public override void TestCleanup()
		{
			this.TestElement = default(T);
		}

		public override void Show()
		{
			this.TestPanel.Children.Add(this.TestElement);

			base.Show();
		}
	}
}
