﻿using System;
using System.IO;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Reflection;

#if NETCORE
namespace Telerik.Zip.Tests
#else
namespace Telerik.Windows.Zip.Tests
#endif
{
    [TestClass]
    public class ZipArchiveTest
    {
        [TestMethod]
        public void ZipCompressionTest()
        {
            const string fileName = "LoremIpsum.txt";
            const string filePath = "/Telerik.Windows.Zip.Tests;component/TestData/" + fileName;

            using (MemoryStream uncompressedStream = ZipTestsHepler.GetResourceStream(filePath))
            {
                MemoryStream compressedStream = new MemoryStream();
                using (ZipArchive zipArchive = new ZipArchive(compressedStream, ZipArchiveMode.Create, true, null))
                {
                    using (ZipArchiveEntry entry = zipArchive.CreateEntry(fileName))
                    {
                        Stream entryStream = entry.Open();
                        uncompressedStream.CopyTo(entryStream);
                    }
                }

                using (Stream inspectedStream = new MemoryStream())
                {
                    using (ZipArchive zipArchive = new ZipArchive(compressedStream))
                    {
                        ZipArchiveEntry entry = zipArchive.GetEntry(fileName);
                        Assert.IsNotNull(entry, "Entry is not found.");

                        Stream stream = entry.Open();
                        stream.CopyTo(inspectedStream);
                    }

                    Assert.IsTrue(ZipTestsHepler.CompareStreams(uncompressedStream, inspectedStream),
                        "Stream is broken after compress/decompress process.");
                }
            }
        }

        [TestMethod]
        public void ZipCompressedFileTest()
        {
            const string filePath = "/Telerik.Windows.Zip.Tests;component/TestData/TestZip.zip";

            MemoryStream zipStream = ZipTestsHepler.GetResourceStream(filePath);
            EntryData[] entriesData = ZipTestsHepler.GetCompressedFileData();

            using (ZipArchive zipArchive = new ZipArchive(zipStream))
            {
                var entries = zipArchive.Entries.ToArray();
                Assert.AreEqual<int>(entries.Length,
                    entriesData.Length,
                    "Count of created entries is incorrect.");

                for (int i = 0; i < entriesData.Length; i++)
                {
                    EntryData data = entriesData[i];
                    data.CompareToZipArchiveEntry(entries[i]);
                }
            }
        }

        [TestMethod]
        public void ZipEntriesTest()
        {
            const int fileCount = 10;
            const string fileNameFormat = "Hello World #{0}.txt";
            MemoryStream zipStream = new MemoryStream();
            using (ZipArchive zipArchive = new ZipArchive(zipStream, ZipArchiveMode.Create, true, null))
            {
                for (int i = 0; i < fileCount; i++)
                {
                    this.AddTestEntry(zipArchive, fileNameFormat, i);
                }
            }

            using (ZipArchive zipArchive = new ZipArchive(zipStream, ZipArchiveMode.Update, false, null))
            {
                var entries = zipArchive.Entries.ToArray();
                Assert.AreEqual<int>(entries.Length,
                    fileCount,
                    "Count of created entries is incorrect.");

                this.AddTestEntry(zipArchive, fileNameFormat, fileCount);

                entries = zipArchive.Entries.ToArray();
                Assert.AreEqual<int>(entries.Length,
                    fileCount + 1,
                    "Count of created entries is incorrect.");

                entries[0].Delete();
                entries[1].Delete();
                entries[2].Delete();

                entries = zipArchive.Entries.ToArray();
                Assert.AreEqual<int>(entries.Length,
                    fileCount - 2,
                    "Count of created entries is incorrect.");
            }
        }

        [TestMethod]
        public void ZipEntryDateTimeTest()
        {
            const string fileName = "Folder";
            DateTimeOffset testDateTime = new DateTime(2014, 2, 23, 15, 21, 32);

            MemoryStream zipStream = new MemoryStream();
            using (ZipArchive zipArchive = new ZipArchive(zipStream, ZipArchiveMode.Create, true, null))
            {
                using (ZipArchiveEntry entry = zipArchive.CreateEntry(fileName))
                {
                    entry.LastWriteTime = testDateTime;
                }
            }

            using (ZipArchive zipArchive = new ZipArchive(zipStream))
            {
                ZipArchiveEntry entry = zipArchive.GetEntry(fileName);
                Assert.IsNotNull(entry, "Entry is not found.");

                Assert.AreEqual<DateTimeOffset>(testDateTime, entry.LastWriteTime, "DateTime of entry is incorrect.");
            }
        }

        [TestMethod]
        public void ZipArchiveUpdateModeTest()
        {
            const string archivePath = "/Telerik.Windows.Zip.Tests;component/TestData/TestZip.zip";
            const string filePath = "/Telerik.Windows.Zip.Tests;component/TestData/test.xml";
            const string fileName = "Data/test.xml";

            MemoryStream zipStream = ZipTestsHepler.GetResourceStream(archivePath, false);
            using (ZipArchive zipArchive = new ZipArchive(zipStream, ZipArchiveMode.Update, true, null))
            {
                ZipArchiveEntry entry = zipArchive.GetEntry(fileName);
                Assert.IsNotNull(entry, "Entry is not found.");

                Stream entryStream = entry.Open();
                this.AddTestLine(entryStream);
            }

            using (MemoryStream stream = ZipTestsHepler.GetResourceStream(filePath, false))
            {
                this.AddTestLine(stream);

                using (Stream inspectedStream = new MemoryStream())
                {
                    using (ZipArchive zipArchive = new ZipArchive(zipStream))
                    {
                        ZipArchiveEntry entry = zipArchive.GetEntry(fileName);
                        Assert.IsNotNull(entry, "Entry is not found.");

                        Stream entryStream = entry.Open();
                        entryStream.CopyTo(inspectedStream);

                        Assert.IsTrue(ZipTestsHepler.CompareStreams(stream, inspectedStream),
                            "The entry is not updated.");
                    }
                }
            }
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException), "Stream must support writing and seeking.")]
        public void CreateZipArchiveWithNotSeekableStreamThrowsValidExceptionTest()
        {
            using (NotSeekAbleStream stream = new NotSeekAbleStream())
            {
                using (ZipArchive archive = new ZipArchive(stream, ZipArchiveMode.Create, true, null))
                {
                }
            }
        }

#if !SILVERLIGHT
        [TestMethod]
        public void ZipArchiveIsDisposedTest()
        {
            const string fileName = "LoremIpsum.txt";
            const string filePath = "/Telerik.Windows.Zip.Tests;component/TestData/" + fileName;

            MemoryStream stream = new MemoryStream();

            using (MemoryStream uncompressedStream = ZipTestsHepler.GetResourceStream(filePath))
            {
                MemoryStream compressedStream = new MemoryStream();

                ZipArchive zipArchive = new ZipArchive(compressedStream, ZipArchiveMode.Create, true, null);

                ZipArchiveEntry entry = zipArchive.CreateEntry(fileName);
                Stream entryStream = entry.Open();
                uncompressedStream.CopyTo(entryStream);

                ZipArchiveEntry firstEntry = zipArchive.Entries.First();
                CompressedStream compressedData = GetFieldValue<CompressedStream>(firstEntry, "compressedData");

                // assert before act
                bool isDisposed = GetPropertyValue<bool>(compressedData, "IsDisposed");
                Assert.IsFalse(isDisposed, "ZipArchiveEntry is already disposed.");

                // act
                zipArchive.Dispose();

                // assert
                isDisposed = GetPropertyValue<bool>(compressedData, "IsDisposed");
                Assert.IsTrue(isDisposed, "ZipArchiveEntry is not disposed properly.");
            }
        }

        private static T GetFieldValue<T>(object objectInstance, string fieldName)
        {
            FieldInfo[] fieldInfos = objectInstance.GetType().GetFields(BindingFlags.NonPublic | BindingFlags.Instance);

            FieldInfo fieldInfo = fieldInfos.Where(info => info.FieldType == typeof(T) && info.Name == fieldName).FirstOrDefault();

            T value = (T)fieldInfo.GetValue(objectInstance);

            return value;
        }

        private static T GetPropertyValue<T>(object objectInstance, string fieldName)
        {
            PropertyInfo[] propertyInfos = objectInstance.GetType().GetProperties(BindingFlags.NonPublic | BindingFlags.Instance);

            PropertyInfo propertyInfo = propertyInfos.Where(info => info.PropertyType == typeof(T) && info.Name == fieldName).FirstOrDefault();

            T value = (T)propertyInfo.GetValue(objectInstance, null);

            return value;
        }
#endif

        private void AddTestEntry(ZipArchive zipArchive, string fileNameFormat, int fileIndex)
        {
            MemoryStream testStream = new MemoryStream();
            using (StreamWriter writter = new StreamWriter(testStream))
            {
                string fileName = string.Format(fileNameFormat, fileIndex.ToString());
                writter.Write(fileName);
                testStream.Position = 0;
                using (ZipArchiveEntry entry = zipArchive.CreateEntry(fileName))
                {
                    Stream stream = entry.Open();
                    testStream.CopyTo(stream);
                }
            }
        }

        private void AddTestLine(Stream stream)
        {
            const string lineToAppend = "This line added to the text in update mode!";

            StreamReader reader = new StreamReader(stream);
            string line = reader.ReadToEnd();
            Assert.IsFalse(string.IsNullOrEmpty(line), "The stream of entry is not read.");

            StreamWriter writer = new StreamWriter(stream);
            writer.WriteLine(lineToAppend);
        }
    }
}
