﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Telerik.Windows.Zip.Extensions
{
    /// <summary>
    /// Platform manager which can be used with full version of the .NET Framework.
    /// </summary>
    public class DotNetPlatformManager : IPlatformManager
    {
        private static readonly string altDirectorySeparatorChar = new string(Path.AltDirectorySeparatorChar, 1);
        private static readonly string directorySeparatorChar = new string(Path.DirectorySeparatorChar, 1);

        private Dictionary<Stream, string> temporaryStreams = new Dictionary<Stream, string>();

        /// <summary>
        /// Initializes a new instance of the DotNetPlatformManager class.
        /// </summary>
        public DotNetPlatformManager()
        {
            this.TemporaryStreamType = TemporaryStreamType.Memory;
        }

        /// <summary>
        /// Gets a platform-specific alternate character used to separate directory levels in a path string that reflects a hierarchical file system organization.
        /// </summary>
        public string AltDirectorySeparatorChar
        {
            get
            {
                return DotNetPlatformManager.altDirectorySeparatorChar;
            }
        }

        /// <summary>
        /// Gets default encoding for this platform.
        /// </summary>
        public Encoding DefaultEncoding
        {
            get
            {
                return Encoding.UTF8;
            }
        }

        /// <summary>
        /// Gets a platform-specific character used to separate directory levels in a path string that reflects a hierarchical file system organization.
        /// </summary>
        public string DirectorySeparatorChar
        {
            get
            {
                return DotNetPlatformManager.directorySeparatorChar;
            }
        }

        /// <summary>
        /// Gets or sets type of the temporary stream. The default value is TemporaryStreamType.Memory.
        /// </summary>
        public TemporaryStreamType TemporaryStreamType
        {
            get;
            set;
        }

        /// <summary>
        /// Creates temporary stream.
        /// </summary>
        /// <returns>Stream will be used for temporary operations.</returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Reliability", "CA2000:Dispose objects before losing scope")]
        public Stream CreateTemporaryStream()
        {
            if (this.TemporaryStreamType == TemporaryStreamType.Memory)
            {
                return new MemoryStream();
            }
            else
            {
                string tempFile = System.IO.Path.GetTempFileName();
                FileStream tempStream = File.Open(tempFile, FileMode.Create, FileAccess.ReadWrite, FileShare.Read);
                this.temporaryStreams.Add(tempStream, tempFile);
                return tempStream;
            }
        }

        /// <summary>
        /// Deletes temporary stream.
        /// </summary>
        /// <param name="stream">Stream to delete.</param>
        public void DeleteTemporaryStream(Stream stream)
        {
            if (this.TemporaryStreamType == TemporaryStreamType.File)
            {
                string tempFile = null;
                if (this.temporaryStreams.TryGetValue(stream, out tempFile))
                {
                    System.IO.File.Delete(tempFile);
                    this.temporaryStreams.Remove(stream);
                }
            }
        }

        /// <summary>
        /// Gets crypto provider initialized using given encryption settings.
        /// </summary>
        /// <param name="settings">Encryption settings.</param>
        /// <returns>Crypto provider.</returns>
        /// <exception cref="NotSupportedException">Specified crypto algorithm is not supported.</exception>
        public ICryptoProvider GetCryptoProvider(EncryptionSettings settings)
        {
            ICryptoProvider cryptoProvider = null;

            switch (settings.Algorithm.ToUpperInvariant())
            {
                case "DEFAULT":
                    cryptoProvider = new DefaultCryptoProvider();
                    break;

                default:
                    throw new NotSupportedException();
            }

            if (cryptoProvider != null)
            {
                cryptoProvider.Initialize(settings);
            }

            return cryptoProvider;
        }

        /// <summary>
        /// Indicates whether specified encoding is supported for this platform.
        /// </summary>
        /// <param name="encoding">Encoding.</param>
        /// <returns>true if encoding is allowed in the ZIP file.</returns>
        public bool IsEncodingSupported(Encoding encoding)
        {
            return (encoding == null)
                || !(encoding.Equals(Encoding.BigEndianUnicode)
                || encoding.Equals(Encoding.Unicode));
        }
    }
}
