﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Telerik.Windows.Zip.Extensions
{
    /// <summary>
    /// Provides static methods for creating, extracting, and opening zip archives. 
    /// </summary>
    public static class ZipFile
    {
        /// <summary>
        /// Archives a file by compressing it and adding it to the zip archive.
        /// </summary>
        /// <param name="destination">The zip archive to add the file to.</param>
        /// <param name="sourceFileName">The path to the file to be archived. You can specify either a relative or an absolute path. A relative path is interpreted as relative to the current working directory.</param>
        /// <param name="entryName">The name of the entry to create in the zip archive.</param>
        /// <returns>New entry in archive.</returns>
        public static ZipArchiveEntry CreateEntryFromFile(
            this ZipArchive destination, 
            string sourceFileName, 
            string entryName)
        {
            return ZipFile.CreateEntryFromFile(destination, sourceFileName, entryName, null);
        }

        /// <summary>
        /// Archives a file by compressing it using the specified compression level and adding it to the zip archive.
        /// </summary>
        /// <param name="destination">The zip archive to add the file to.</param>
        /// <param name="sourceFileName">The path to the file to be archived. You can specify either a relative or an absolute path. A relative path is interpreted as relative to the current working directory.</param>
        /// <param name="entryName">The name of the entry to create in the zip archive.</param>
        /// <param name="compressionLevel">One of the enumeration values that indicates whether to emphasize speed or compression effectiveness when creating the entry.</param>
        /// <returns>New entry in archive.</returns>
        public static ZipArchiveEntry CreateEntryFromFile(
            this ZipArchive destination, 
            string sourceFileName, 
            string entryName, 
            CompressionLevel compressionLevel)
        {
            DeflateSettings deflateSettings = new DeflateSettings()
            {
                CompressionLevel = compressionLevel,
                HeaderType = CompressedStreamHeader.None
            };

            return ZipFile.CreateEntryFromFile(destination, sourceFileName, entryName, deflateSettings);
        }

        /// <summary>
        /// Archives a file by compressing it using the specified compression settings and adding it to the zip archive.
        /// </summary>
        /// <param name="destination">The zip archive to add the file to.</param>
        /// <param name="sourceFileName">The path to the file to be archived. You can specify either a relative or an absolute path. A relative path is interpreted as relative to the current working directory.</param>
        /// <param name="entryName">The name of the entry to create in the zip archive.</param>
        /// <param name="compressionSettings">Compression settings.</param>
        /// <returns>New entry in archive.</returns>
        public static ZipArchiveEntry CreateEntryFromFile(
            this ZipArchive destination,
            string sourceFileName,
            string entryName,
            CompressionSettings compressionSettings)
        {
            ZipArchiveEntry zipArchiveEntry;

            if (destination == null)
            {
                throw new ArgumentNullException("destination");
            }

            if (sourceFileName == null)
            {
                throw new ArgumentNullException("sourceFileName");
            }

            if (entryName == null)
            {
                throw new ArgumentNullException("entryName");
            }

            using (Stream fileStream = File.Open(sourceFileName, FileMode.Open, FileAccess.Read, FileShare.Read))
            {
                zipArchiveEntry = compressionSettings == null ?
                    destination.CreateEntry(entryName) :
                    destination.CreateEntry(entryName, compressionSettings);

                zipArchiveEntry.ExternalAttributes = (int)File.GetAttributes(sourceFileName);

                DateTime lastWriteTime = File.GetLastWriteTime(sourceFileName);
                if (lastWriteTime.Year < 1980 || lastWriteTime.Year > 2107)
                {
                    lastWriteTime = new DateTime(1980, 1, 1, 0, 0, 0);
                }

                zipArchiveEntry.LastWriteTime = lastWriteTime;

                Stream entryStream = zipArchiveEntry.Open();
                fileStream.CopyTo(entryStream);

                if (destination.Mode == ZipArchiveMode.Create)
                {
                    entryStream.Dispose();
                    zipArchiveEntry.Dispose();
                }
            }

            return zipArchiveEntry;
        }

        /// <summary>
        /// Creates a zip archive that contains the files and directories from the specified directory.
        /// </summary>
        /// <param name="sourceDirectoryName">The path to the directory to be archived, specified as a relative or absolute path. A relative path is interpreted as relative to the current working directory.</param>
        /// <param name="destinationArchiveFileName">The path of the archive to be created, specified as a relative or absolute path. A relative path is interpreted as relative to the current working directory.</param>
        public static void CreateFromDirectory(string sourceDirectoryName, string destinationArchiveFileName)
        {
            ZipFile.CreateFromDirectory(sourceDirectoryName, destinationArchiveFileName, null, false, null);
        }

        /// <summary>
        /// Creates a zip archive that contains the files and directories from the specified directory, uses the specified compression level, and optionally includes the base directory.
        /// </summary>
        /// <param name="sourceDirectoryName">The path to the directory to be archived, specified as a relative or absolute path. A relative path is interpreted as relative to the current working directory.</param>
        /// <param name="destinationArchiveFileName">The path of the archive to be created, specified as a relative or absolute path. A relative path is interpreted as relative to the current working directory.</param>
        /// <param name="compressionLevel">One of the enumeration values that indicates whether to emphasize speed or compression effectiveness when creating the entry.</param>
        /// <param name="includeBaseDirectory">True to include the directory name from sourceDirectoryName at the root of the archive; false to include only the contents of the directory.</param>
        public static void CreateFromDirectory(
            string sourceDirectoryName, 
            string destinationArchiveFileName, 
            CompressionLevel compressionLevel, 
            bool includeBaseDirectory)
        {
            DeflateSettings deflateSettings = new DeflateSettings()
            {
                CompressionLevel = compressionLevel,
                HeaderType = CompressedStreamHeader.None
            };

            ZipFile.CreateFromDirectory(sourceDirectoryName, destinationArchiveFileName, deflateSettings, includeBaseDirectory, null);
        }

        /// <summary>
        /// Creates a zip archive that contains the files and directories from the specified directory, uses the specified compression level, and optionally includes the base directory.
        /// </summary>
        /// <param name="sourceDirectoryName">The path to the directory to be archived, specified as a relative or absolute path. A relative path is interpreted as relative to the current working directory.</param>
        /// <param name="destinationArchiveFileName">The path of the archive to be created, specified as a relative or absolute path. A relative path is interpreted as relative to the current working directory.</param>
        /// <param name="compressionLevel">One of the enumeration values that indicates whether to emphasize speed or compression effectiveness when creating the entry.</param>
        /// <param name="includeBaseDirectory">True to include the directory name from sourceDirectoryName at the root of the archive; false to include only the contents of the directory.</param>
        /// <param name="entryNameEncoding">The encoding to use when reading or writing entry names in this archive. Specify a value for this parameter only when an encoding is required for interoperability with zip archive tools and libraries that do not support UTF-8 encoding for entry names.</param>
        public static void CreateFromDirectory(
            string sourceDirectoryName, 
            string destinationArchiveFileName, 
            CompressionLevel compressionLevel, 
            bool includeBaseDirectory, 
            Encoding entryNameEncoding)
        {
            DeflateSettings deflateSettings = new DeflateSettings()
            {
                CompressionLevel = compressionLevel,
                HeaderType = CompressedStreamHeader.None
            };

            ZipFile.CreateFromDirectory(sourceDirectoryName, destinationArchiveFileName, deflateSettings, includeBaseDirectory, entryNameEncoding);
        }

        /// <summary>
        /// Creates a zip archive that contains the files and directories from the specified directory, uses the specified compression settings, and optionally includes the base directory.
        /// </summary>
        /// <param name="sourceDirectoryName">The path to the directory to be archived, specified as a relative or absolute path. A relative path is interpreted as relative to the current working directory.</param>
        /// <param name="destinationArchiveFileName">The path of the archive to be created, specified as a relative or absolute path. A relative path is interpreted as relative to the current working directory.</param>
        /// <param name="compressionSettings">Compression settings.</param>
        /// <param name="includeBaseDirectory">True to include the directory name from sourceDirectoryName at the root of the archive; false to include only the contents of the directory.</param>
        /// <param name="entryNameEncoding">The encoding to use when reading or writing entry names in this archive. Specify a value for this parameter only when an encoding is required for interoperability with zip archive tools and libraries that do not support UTF-8 encoding for entry names.</param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Reliability", "CA2000:Dispose objects before losing scope")]
        public static void CreateFromDirectory(
            string sourceDirectoryName, 
            string destinationArchiveFileName, 
            CompressionSettings compressionSettings, 
            bool includeBaseDirectory, 
            Encoding entryNameEncoding)
        {
            char[] directorySeparatorChar = new char[] { Path.DirectorySeparatorChar, Path.AltDirectorySeparatorChar };

            sourceDirectoryName = Path.GetFullPath(sourceDirectoryName);
            destinationArchiveFileName = Path.GetFullPath(destinationArchiveFileName);

            using (ZipArchive zipArchive = ZipFile.Open(destinationArchiveFileName, ZipArchiveMode.Create, entryNameEncoding))
            {
                bool empty = true;

                DirectoryInfo directoryInfo = new DirectoryInfo(sourceDirectoryName);
                string fullName = directoryInfo.FullName;
                if (includeBaseDirectory && directoryInfo.Parent != null)
                {
                    fullName = directoryInfo.Parent.FullName;
                }

                foreach (FileSystemInfo fileSystemInfo in directoryInfo.EnumerateFileSystemInfos("*", SearchOption.AllDirectories))
                {
                    empty = false;
                    
                    int length = fileSystemInfo.FullName.Length - fullName.Length;
                    string entryName = fileSystemInfo.FullName.Substring(fullName.Length, length);
                    entryName = entryName.TrimStart(directorySeparatorChar);

                    DirectoryInfo subfolder = fileSystemInfo as DirectoryInfo;
                    if (subfolder != null)
                    {
                        if (ZipFile.IsDirectoryEmpty(subfolder))
                        {
                            zipArchive.CreateEntry(string.Concat(entryName, Path.DirectorySeparatorChar));
                        }
                    }
                    else
                    {
                        ZipFile.CreateEntryFromFile(zipArchive, fileSystemInfo.FullName, entryName, compressionSettings);
                    }
                }

                if (includeBaseDirectory && empty)
                {
                    zipArchive.CreateEntry(string.Concat(directoryInfo.Name, Path.DirectorySeparatorChar));
                }
            }
        }

        /// <summary>
        /// Extracts all the files in the specified zip archive to a directory on the file system.
        /// </summary>
        /// <param name="sourceArchiveFileName">The path to the archive that is to be extracted.</param>
        /// <param name="destinationDirectoryName">The path to the directory in which to place the extracted files, 
        /// specified as a relative or absolute path. A relative path is interpreted as relative to the current working directory.</param>
        public static void ExtractToDirectory(string sourceArchiveFileName, string destinationDirectoryName)
        {
            ZipFile.ExtractToDirectory(sourceArchiveFileName, destinationDirectoryName, null);
        }

        /// <summary>
        /// Extracts all the files in the specified zip archive to a directory on the file system and uses the specified character encoding for entry names.
        /// </summary>
        /// <param name="sourceArchiveFileName">The path to the archive that is to be extracted.</param>
        /// <param name="destinationDirectoryName">The path to the directory in which to place the extracted files, 
        /// specified as a relative or absolute path. A relative path is interpreted as relative to the current working directory.</param>
        /// <param name="entryNameEncoding">The encoding to use when reading or writing entry names in this archive. 
        /// Specify a value for this parameter only when an encoding is required for interoperability with zip archive 
        /// tools and libraries that do not support UTF-8 encoding for entry names.</param>
        public static void ExtractToDirectory(
            string sourceArchiveFileName, 
            string destinationDirectoryName, 
            Encoding entryNameEncoding)
        {
            if (sourceArchiveFileName == null)
            {
                throw new ArgumentNullException("sourceArchiveFileName");
            }

            using (ZipArchive zipArchive = ZipFile.Open(sourceArchiveFileName, ZipArchiveMode.Read, entryNameEncoding))
            {
                zipArchive.ExtractToDirectory(destinationDirectoryName);
            }
        }

        /// <summary>
        /// Extracts all the files in the zip archive to a directory on the file system.
        /// </summary>
        /// <param name="source">The zip archive to extract files from.</param>
        /// <param name="destinationDirectoryName">The path to the directory to place the extracted files in. You can specify either a relative or an absolute path. A relative path is interpreted as relative to the current working directory.</param>
        public static void ExtractToDirectory(this ZipArchive source, string destinationDirectoryName)
        {
            if (source == null)
            {
                throw new ArgumentNullException("source");
            }

            if (destinationDirectoryName == null)
            {
                throw new ArgumentNullException("destinationDirectoryName");
            }

            string fullName = Directory.CreateDirectory(destinationDirectoryName).FullName;

            foreach (ZipArchiveEntry entry in source.Entries)
            {
                string fullPath = Path.GetFullPath(Path.Combine(fullName, entry.FullName));
                if (!fullPath.StartsWith(fullName, StringComparison.OrdinalIgnoreCase))
                {
                    throw new IOException("Extracting results in outside");
                }

                if (Path.GetFileName(fullPath).Length != 0)
                {
                    Directory.CreateDirectory(Path.GetDirectoryName(fullPath));
                    entry.ExtractToFile(fullPath, false);
                }
                else
                {
                    if (entry.Length != (long)0)
                    {
                        throw new IOException("Directory name with data");
                    }
                    Directory.CreateDirectory(fullPath);
                }
            }
        }

        /// <summary>
        /// Extracts an entry in the zip archive to a file.
        /// </summary>
        /// <param name="source">The zip archive entry to extract a file from.</param>
        /// <param name="destinationFileName">The path of the file to create from the contents 
        /// of the entry. You can specify either a relative or an absolute path. A relative path 
        /// is interpreted as relative to the current working directory.</param>
        public static void ExtractToFile(this ZipArchiveEntry source, string destinationFileName)
        {
            source.ExtractToFile(destinationFileName, false);
        }

        /// <summary>
        /// Extracts an entry in the zip archive to a file, and optionally overwrites an existing file that has the same name.
        /// </summary>
        /// <param name="source">The zip archive entry to extract a file from.</param>
        /// <param name="destinationFileName">The path of the file to create from the contents 
        /// of the entry. You can specify either a relative or an absolute path. A relative path 
        /// is interpreted as relative to the current working directory.</param>
        /// <param name="overwrite">True to overwrite an existing file that has the same name as the destination file; otherwise, false.</param>
        public static void ExtractToFile(this ZipArchiveEntry source, string destinationFileName, bool overwrite)
        {
            if (source == null)
            {
                throw new ArgumentNullException("source");
            }

            if (destinationFileName == null)
            {
                throw new ArgumentNullException("destinationFileName");
            }

            using (Stream fileStream = File.Open(
                destinationFileName, 
                overwrite ? FileMode.Create : FileMode.CreateNew,
                FileAccess.Write, 
                FileShare.None))
            {
                using (Stream entryStream = source.Open())
                {
                    entryStream.CopyTo(fileStream);
                }
            }

            File.SetLastWriteTime(destinationFileName, source.LastWriteTime.DateTime);
        }

        /// <summary>
        /// Opens a zip archive at the specified path and in the specified mode.
        /// </summary>
        /// <param name="archiveFileName">The path to the archive to open, specified as a relative or absolute path. A relative path is interpreted as relative to the current working directory.</param>
        /// <param name="mode">One of the enumeration values that specifies the actions which are allowed on the entries in the opened archive.</param>
        /// <returns>The opened zip archive.</returns>
        public static ZipArchive Open(string archiveFileName, ZipArchiveMode mode)
        {
            return ZipFile.Open(archiveFileName, mode, null);
        }

        /// <summary>
        /// Opens a zip archive at the specified path and in the specified mode.
        /// </summary>
        /// <param name="archiveFileName">The path to the archive to open, specified as a relative or absolute path. A relative path is interpreted as relative to the current working directory.</param>
        /// <param name="mode">One of the enumeration values that specifies the actions which are allowed on the entries in the opened archive.</param>
        /// <param name="entryNameEncoding">The encoding to use when reading or writing entry names in this archive. Specify a value for this parameter only when an encoding is required for interoperability with zip archive tools and libraries that do not support UTF-8 encoding for entry names.</param>
        /// <returns>The opened zip archive.</returns>
        public static ZipArchive Open(string archiveFileName, ZipArchiveMode mode, Encoding entryNameEncoding)
        {
            FileMode fileMode;
            FileAccess fileAccess;
            FileShare fileShare;
            ZipArchive zipArchive;

            switch (mode)
            {
                case ZipArchiveMode.Read:
                    fileMode = FileMode.Open;
                    fileAccess = FileAccess.Read;
                    fileShare = FileShare.Read;
                    break;

                case ZipArchiveMode.Create:
                    fileMode = FileMode.CreateNew;
                    fileAccess = FileAccess.Write;
                    fileShare = FileShare.None;
                    break;

                case ZipArchiveMode.Update:
                    fileMode = FileMode.OpenOrCreate;
                    fileAccess = FileAccess.ReadWrite;
                    fileShare = FileShare.None;
                    break;

                default:
                    throw new ArgumentOutOfRangeException("mode");
            }

            FileStream fileStream = null;
            try
            {
                fileStream = File.Open(archiveFileName, fileMode, fileAccess, fileShare);
                zipArchive = new ZipArchive(fileStream, mode, false, entryNameEncoding);
            }
            catch
            {
                if (fileStream != null)
                {
                    fileStream.Dispose();
                }
                throw;
            }
            return zipArchive;
        }

        /// <summary>
        /// Opens a zip archive for reading at the specified path.
        /// </summary>
        /// <param name="archiveFileName">The path to the archive to open, specified as a relative or absolute path. 
        /// A relative path is interpreted as relative to the current working directory.</param>
        /// <returns>The opened zip archive.</returns>
        public static ZipArchive OpenRead(string archiveFileName)
        {
            return ZipFile.Open(archiveFileName, ZipArchiveMode.Read);
        }

        /// <summary>
        /// Indicates whether specified directory is empty or not.
        /// </summary>
        /// <param name="directoryInfo">Directory info.</param>
        /// <returns>True if directory is empty; otherwise - false.</returns>
        private static bool IsDirectoryEmpty(DirectoryInfo directoryInfo)
        {
            bool empty = true;

            using (IEnumerator<FileSystemInfo> enumerator = directoryInfo.EnumerateFileSystemInfos("*", SearchOption.AllDirectories).GetEnumerator())
            {
                if (enumerator.MoveNext())
                {
                    empty = false;
                }
            }

            return empty;
        }
    }
}
