﻿using System;
using System.IO;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Represents Zip64 end of central directory record described in the
    /// ZIP File Format Specification v6.3.3, #4.3.14.
    /// </summary>
    internal class Zip64EndOfCentralDirectoryRecord : ISpecData
    {
        /// <summary>
        /// Zip64 end of central directory record signature.
        /// </summary>
        public const uint Signature = 0x06064b50;

        /// <summary>
        /// Gets or sets size of zip64 end of central
        /// directory record.
        /// </summary>
        public ulong SizeOfZip64EndOfCentralDirectoryRecord
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets byte which indicates the ZIP specification version
        /// supported by the software used to encode the file.
        /// </summary>
        public byte VersionMadeBy
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets byte which indicates the compatibility
        /// of the file attribute information.
        /// </summary>
        public byte OsCompatibility
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets version needed to extract.
        /// </summary>
        public ushort VersionNeededToExtract
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets number of this disk.
        /// </summary>
        public uint NumberOfThisDisk
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets number of the disk with the start of the central directory.
        /// </summary>
        public uint NumberOfTheDiskWithTheStartOfTheCentralDirectory
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets total number of entries in the central directory on this disk.
        /// </summary>
        public ulong NumberOfEntriesInTheCentralDirectoryOnThisDisk
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets total number of entries in the central directory.
        /// </summary>
        public ulong NumberOfEntriesInTheCentralDirectory
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets size of the central directory.
        /// </summary>
        public ulong SizeOfCentralDirectory
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets offset of start of central directory with respect to the starting disk number.
        /// </summary>
        public ulong OffsetOfStartOfCentralDirectoryWithRespectToTheStartingDiskNumber
        {
            get;
            set;
        }

        /// <summary>
        /// Read data from the binary reader.
        /// </summary>
        /// <param name="reader">Binary reader to read data from.</param>
        /// <returns>True if success, otherwise false.</returns>
        public bool TryReadBlock(BinaryReader reader)
        {
            if (reader.ReadUInt32() != Zip64EndOfCentralDirectoryRecord.Signature)
            {
                return false;
            }

            this.SizeOfZip64EndOfCentralDirectoryRecord = reader.ReadUInt64();
            this.VersionMadeBy = reader.ReadByte();
            this.OsCompatibility = reader.ReadByte();
            this.VersionNeededToExtract = reader.ReadUInt16();
            this.NumberOfThisDisk = reader.ReadUInt32();
            this.NumberOfTheDiskWithTheStartOfTheCentralDirectory = reader.ReadUInt32();
            this.NumberOfEntriesInTheCentralDirectoryOnThisDisk = reader.ReadUInt64();
            this.NumberOfEntriesInTheCentralDirectory = reader.ReadUInt64();
            this.SizeOfCentralDirectory = reader.ReadUInt64();
            this.OffsetOfStartOfCentralDirectoryWithRespectToTheStartingDiskNumber = reader.ReadUInt64();

            return true;
        }

        /// <summary>
        /// Write data to the binary writer.
        /// </summary>
        /// <param name="writer">Binary writer to write data to.</param>
        public void WriteBlock(BinaryWriter writer)
        {
            byte[] fieldsData = this.GetBlock();

            this.SizeOfZip64EndOfCentralDirectoryRecord = (ulong)fieldsData.Length;

            writer.Write(Zip64EndOfCentralDirectoryRecord.Signature);
            writer.Write(this.SizeOfZip64EndOfCentralDirectoryRecord);
            writer.Write(fieldsData);
        }

        private byte[] GetBlock()
        {
            byte[] fieldsData;

            using (MemoryStream stream = new MemoryStream())
            {
                using (BinaryWriter writer = new BinaryWriter(stream))
                {
                    writer.Write(this.VersionMadeBy);
                    writer.Write(this.OsCompatibility);
                    writer.Write(this.VersionNeededToExtract);
                    writer.Write(this.NumberOfThisDisk);
                    writer.Write(this.NumberOfTheDiskWithTheStartOfTheCentralDirectory);
                    writer.Write(this.NumberOfEntriesInTheCentralDirectoryOnThisDisk);
                    writer.Write(this.NumberOfEntriesInTheCentralDirectory);
                    writer.Write(this.SizeOfCentralDirectory);
                    writer.Write(this.OffsetOfStartOfCentralDirectoryWithRespectToTheStartingDiskNumber);

                    fieldsData = stream.ToArray();
                }
            }

            return fieldsData;
        }
    }
}
