﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Text;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Represents a package of compressed files in the zip archive format. 
    /// </summary>
    public class ZipArchive : IDisposable, INotifyPropertyChanged
    {
        /// <summary>
        /// Value that describes the type of action the zip archive can perform on entries.
        /// </summary>
        private ZipArchiveMode archiveMode;

        /// <summary>
        /// Binary reader is used to read from working stream.
        /// </summary>
        private BinaryReader archiveReader;

        /// <summary>
        /// Binary writer is used to write to working stream.
        /// </summary>
        private BinaryWriter archiveWriter;

        /// <summary>
        /// Track whether Dispose has been called. 
        /// </summary>
        private bool disposed = false;

        /// <summary>
        /// Encoding of the entry name.
        /// </summary>
        private Encoding entryNameEncoding;

        /// <summary>
        /// Original archive stream. If this stream doesn't support seeking then 
        /// temporary working stream will be created.
        /// </summary>
        private Stream originalStream;

        /// <summary>
        /// Working archive stream. If original stream doesn't support seeking then 
        /// temporary working stream will be created.
        /// </summary>
        private Stream workingStream;

        /// <summary>
        /// True to leave the stream open after the ZipArchive object is disposed; otherwise, false.
        /// </summary>
        private bool leaveStreamOpen;

        /// <summary>
        /// Indicates whether the central directory have been read.
        /// </summary>
        private bool centralDirectoryRead;

        /// <summary>
        /// ZIP Archive End of Central Directory.
        /// </summary>
        private EndOfCentralDirectoryRecord endOfCentralDirectoryRecord;

        /// <summary>
        /// ZIP64 End of Central Directory Locator.
        /// </summary>
        private Zip64EndOfCentralDirectoryLocator zip64EndOfCentralDirectoryLocator;

        /// <summary>
        /// ZIP64 End of Central Directory Record.
        /// </summary>
        private Zip64EndOfCentralDirectoryRecord zip64EndOfCentralDirectoryRecord;

        /// <summary>
        /// ZIP entries.
        /// </summary>
        private Dictionary<string, ZipArchiveEntry> entries = new Dictionary<string, ZipArchiveEntry>();

        /// <summary>
        /// Initializes a new instance of the ZipArchive class from the specified stream.
        /// </summary>
        /// <param name="stream">The stream that contains the archive to be read.</param>
        public ZipArchive(Stream stream)
            : this(stream, ZipArchiveMode.Read, true, null, null, null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the ZipArchive class.
        /// </summary>
        /// <param name="stream">The stream that contains the archive to be read.</param>
        /// <param name="mode">One of the enumeration values that indicates whether the zip archive is used to read, create, or update entries.</param>
        /// <param name="leaveOpen">True to leave the stream open after the ZipArchive object is disposed; otherwise, false.</param>
        /// <param name="entryNameEncoding">The encoding to use when reading or writing entry names in this archive. Specify a value for this parameter 
        /// only when an encoding is required for interoperability with zip archive tools and libraries that do not support UTF-8 encoding for entry names.</param>
        public ZipArchive(Stream stream, ZipArchiveMode mode, bool leaveOpen, Encoding entryNameEncoding)
            : this(stream, mode, leaveOpen, entryNameEncoding, null, null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the ZipArchive class.
        /// </summary>
        /// <param name="stream">The stream that contains the archive to be read.</param>
        /// <param name="mode">One of the enumeration values that indicates whether the zip archive is used to read, create, or update entries.</param>
        /// <param name="leaveOpen">True to leave the stream open after the ZipArchive object is disposed; otherwise, false.</param>
        /// <param name="entryNameEncoding">The encoding to use when reading or writing entry names in this archive. Specify a value for this parameter 
        /// only when an encoding is required for interoperability with zip archive tools and libraries that do not support UTF-8 encoding for entry names.</param>
        /// <param name="compressionSettings">Compression settings.</param>
        /// <param name="encryptionSettings">Encryption settings.</param>
        public ZipArchive(Stream stream, ZipArchiveMode mode, bool leaveOpen, Encoding entryNameEncoding, CompressionSettings compressionSettings, EncryptionSettings encryptionSettings)
        {
            if (stream == null)
            {
                throw new ArgumentNullException("stream");
            }

            this.EntryNameEncoding = entryNameEncoding;

            if (compressionSettings == null)
            {
                compressionSettings = new DeflateSettings()
                {
                    HeaderType = CompressedStreamHeader.None
                };
            }

            this.CompressionSettings = compressionSettings;
            this.CompressionSettings.PrepareForZip();

            this.EncryptionSettings = encryptionSettings;

            this.Init(stream, mode, leaveOpen);
        }

        /// <summary>
        /// Allows an object to try to free resources and perform other cleanup operations before it is reclaimed by garbage collection. 
        /// </summary>
        ~ZipArchive()
        {
            this.Dispose(false);
        }

        /// <summary>
        /// Occurs when a property value changes.
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        /// <summary>
        /// Gets the collection of entries that are currently in the zip archive.
        /// </summary>
        /// <exception cref="NotSupportedException">The zip archive does not support reading.</exception>
        /// <exception cref="ObjectDisposedException">The zip archive has been disposed.</exception>
        /// <exception cref="InvalidDataException">The zip archive is corrupt, and its entries cannot be retrieved.</exception>
        public IEnumerable<ZipArchiveEntry> Entries
        {
            get
            {
                this.EnsureCentralDirectoryRead();
                foreach (ZipArchiveEntry entry in this.entries.Values)
                {
                    yield return entry;
                }
            }
        }

        /// <summary>
        /// Gets entry name encoding.
        /// </summary>
        public Encoding EntryNameEncoding
        {
            get
            {
                return this.entryNameEncoding;
            }

            private set
            {
                if (!PlatformSettings.Manager.IsEncodingSupported(value))
                {
                    throw new ArgumentException("Entry name encoding is not supported", "value");
                }
                this.entryNameEncoding = value;
            }
        }

        /// <summary>
        /// Gets a value that describes the type of action the zip archive can perform on entries.
        /// </summary>
        public ZipArchiveMode Mode
        {
            get
            {
                return this.archiveMode;
            }
        }

        /// <summary>
        /// Gets compression settings.
        /// </summary>
        internal CompressionSettings CompressionSettings
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets encryption settings.
        /// </summary>
        internal EncryptionSettings EncryptionSettings
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets number of the disk.
        /// </summary>
        internal uint NumberOfThisDisk
        {
            get
            {
                return this.zip64EndOfCentralDirectoryRecord == null ?
                    this.endOfCentralDirectoryRecord.NumberOfThisDisk :
                    this.zip64EndOfCentralDirectoryRecord.NumberOfThisDisk;
            }
        }

        /// <summary>
        /// Gets reader for the working stream.
        /// </summary>
        internal BinaryReader Reader
        {
            get
            {
                return this.archiveReader;
            }
        }

        /// <summary>
        /// Gets writer for the working stream.
        /// </summary>
        internal BinaryWriter Writer
        {
            get
            {
                return this.archiveWriter;
            }
        }

        /// <summary>
        /// Gets start of the central directory.
        /// </summary>
        private long CentralDirectoryStart
        {
            get
            {
                return this.zip64EndOfCentralDirectoryRecord == null ?
                    (long)this.endOfCentralDirectoryRecord.OffsetOfStartOfCentralDirectoryWithRespectToTheStartingDiskNumber :
                    (long)this.zip64EndOfCentralDirectoryRecord.OffsetOfStartOfCentralDirectoryWithRespectToTheStartingDiskNumber;
            }
        }

        private ulong NumberOfEntriesInTheCentralDirectory
        {
            get
            {
                return this.zip64EndOfCentralDirectoryRecord == null ?
                    (ulong)this.endOfCentralDirectoryRecord.NumberOfEntriesInTheCentralDirectory :
                    this.zip64EndOfCentralDirectoryRecord.NumberOfEntriesInTheCentralDirectory;
            }
        }

        /// <summary>
        /// Creates an empty entry that has the specified path and entry name in the zip archive.
        /// </summary>
        /// <param name="entryName">A path, relative to the root of the archive, that specifies the name of the entry to be created.</param>
        /// <returns>An empty entry in the zip archive.</returns>
        /// <exception cref="ArgumentException">The entry name is empty.</exception>
        /// <exception cref="ArgumentNullException">The entry name is null.</exception>
        /// <exception cref="NotSupportedException">The zip archive does not support writing.</exception>
        /// <exception cref="ObjectDisposedException">The zip archive has been disposed.</exception>
        public ZipArchiveEntry CreateEntry(string entryName)
        {
            ZipArchiveEntry entry = new ZipArchiveEntry(this, entryName, this.CompressionSettings);
            this.entries.Add(entry.FullName, entry);
            return entry;
        }

        /// <summary>
        /// Creates an empty entry that has the specified path and entry name in the zip archive.
        /// </summary>
        /// <param name="entryName">A path, relative to the root of the archive, that specifies the name of the entry to be created.</param>
        /// <param name="settings">Compression settings.</param>
        /// <returns>An empty entry in the zip archive.</returns>
        /// <exception cref="ArgumentException">The entry name is empty.</exception>
        /// <exception cref="ArgumentNullException">The entry name is null.</exception>
        /// <exception cref="NotSupportedException">The zip archive does not support writing.</exception>
        /// <exception cref="ObjectDisposedException">The zip archive has been disposed.</exception>
        public ZipArchiveEntry CreateEntry(string entryName, CompressionSettings settings)
        {
            ZipArchiveEntry entry = new ZipArchiveEntry(this, entryName, settings);
            this.entries.Add(entry.FullName, entry);
            return entry;
        }

        /// <summary>
        /// Releases the resources used by the current instance of the ZipArchive class.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Retrieves a wrapper for the specified entry in the zip archive.
        /// </summary>
        /// <param name="entryName">A path, relative to the root of the archive, that identifies the entry to retrieve.</param>
        /// <returns>A wrapper for the specified entry in the archive; null if the entry does not exist in the archive.</returns>
        /// <exception cref="ArgumentException">The entry name is empty.</exception>
        /// <exception cref="ArgumentNullException">The entry name is null.</exception>
        /// <exception cref="NotSupportedException">The zip archive does not support reading.</exception>
        /// <exception cref="ObjectDisposedException">The zip archive has been disposed.</exception>
        /// <exception cref="InvalidDataException">The zip archive is corrupt, and its entries cannot be retrieved.</exception>
        public ZipArchiveEntry GetEntry(string entryName)
        {
            ZipArchiveEntry zipArchiveEntry;
            if (entryName == null)
            {
                throw new ArgumentNullException("entryName");
            }
            if (this.Mode == ZipArchiveMode.Create)
            {
                throw new NotSupportedException("Can't get entry in the create mode.");
            }
            this.EnsureCentralDirectoryRead();
            this.entries.TryGetValue(entryName, out zipArchiveEntry);
            return zipArchiveEntry;
        }

        internal void ThrowIfDisposed()
        {
            if (this.disposed)
            {
                throw new ObjectDisposedException(this.GetType().Name);
            }
        }

        internal void RemoveEntry(ZipArchiveEntry entry)
        {
            this.entries.Remove(entry.FullName);
        }

        /// <summary>
        /// Release the unmanaged resources used by the current instance of the ZipArchive class.
        /// </summary>
        /// <param name="closeStream">True to leave the stream open after the ZipArchive object is disposed; otherwise, false.</param>
        internal void DisposeStreams(bool closeStream)
        {
            this.leaveStreamOpen = !closeStream;
            this.Dispose();
        }

        private void ClearEntries()
        {
            foreach (ZipArchiveEntry entry in this.entries.Values)
            {
                entry.Dispose();
                entry.Archive = null;
            }

            this.entries.Clear();
        }

        /// <summary>
        /// Called by the Dispose() and Finalize() methods to release the unmanaged 
        /// resources used by the current instance of the ZipArchive class, and optionally 
        /// finishes writing the archive and releases the managed resources.
        /// </summary>
        /// <param name="disposing">True to finish writing the archive and release unmanaged and managed resources; 
        /// false to release only unmanaged resources.</param>
        private void Dispose(bool disposing)
        {
            // Check to see if Dispose has already been called. 
            if (!this.disposed)
            {
                // If disposing equals true, dispose all managed  
                // and unmanaged resources. 
                if (disposing)
                {
                    // Dispose managed resources.
                    switch (this.Mode)
                    {
                        case ZipArchiveMode.Read:
                            this.DisposeStreams();
                            break;

                        default:
                            try
                            {
                                this.WriteArchive();
                                this.DisposeStreams();
                            }
                            catch (InvalidDataException)
                            {
                                this.DisposeStreams();
                                this.disposed = true;
                                throw;
                            }
                            break;
                    }
                }

                this.ClearEntries();

                // Call the appropriate methods to clean up  
                // unmanaged resources here. 
            }
            this.disposed = true;
        }

        /// <summary>
        /// Dispose streams.
        /// </summary>
        private void DisposeStreams()
        {
            if (!this.leaveStreamOpen)
            {
                this.workingStream.Dispose();
                if (this.originalStream != null)
                {
                    this.originalStream.Dispose();
                    PlatformSettings.Manager.DeleteTemporaryStream(this.workingStream);
                }

                if (this.archiveReader != null)
                {
#if NETSTANDARD
                    this.archiveReader.Dispose();
#else
                    this.archiveReader.Close();
#endif
                }

                if (this.archiveWriter != null)
                {
#if NETSTANDARD
                    this.archiveWriter.Dispose();
#else
                    this.archiveWriter.Close();
#endif
                }
            }
            else if (this.originalStream != null)
            {
                this.originalStream.Flush();
                this.workingStream.Dispose();
                PlatformSettings.Manager.DeleteTemporaryStream(this.workingStream);
            }
        }

        private void EnsureCentralDirectoryRead()
        {
            if (!this.centralDirectoryRead)
            {
                this.ReadCentralDirectory();
                this.centralDirectoryRead = true;
            }
        }

        private void Init(Stream stream, ZipArchiveMode mode, bool leaveOpen)
        {
            try
            {
                this.originalStream = null;
                switch (mode)
                {
                    case ZipArchiveMode.Read:
                        if (!stream.CanRead)
                        {
                            throw new ArgumentException("Stream must support reading.");
                        }

                        if (stream.CanSeek)
                        {
                            break;
                        }

                        this.originalStream = stream;
                        stream = PlatformSettings.Manager.CreateTemporaryStream();

#if (WPF || SILVERLIGHT || NETSTANDARD)
                        this.originalStream.CopyTo(stream);
#else
                        Telerik.WinControls.Zip.Extensions.ZipFile.CopyStreamTo(originalStream, stream);
#endif

                        stream.Seek(0L, SeekOrigin.Begin);
                        break;

                    case ZipArchiveMode.Create:
                        if (stream.CanWrite && stream.CanSeek)
                        {
                            stream.SetLength(0);
                            break;
                        }

                        throw new ArgumentException("Stream must support writing and seeking.");

                    case ZipArchiveMode.Update:
                        if (stream.CanRead && stream.CanSeek)
                        {
                            this.originalStream = stream;
                            stream = PlatformSettings.Manager.CreateTemporaryStream();
                            stream.Seek(0L, SeekOrigin.Begin);

                            if (stream.CanWrite)
                            {
                                break;
                            }
                        }

                        throw new ArgumentException("Stream must support reading, writing and seeking.");

                    default:
                        throw new ArgumentOutOfRangeException("mode");
                }

                this.archiveMode = mode;
                this.workingStream = stream;

                if (mode == ZipArchiveMode.Read)
                {
                    this.archiveWriter = null;
                    this.archiveReader = new BinaryReader(this.workingStream);
                }
                else if (mode == ZipArchiveMode.Update)
                {
                    this.archiveReader = new BinaryReader(this.originalStream);
                    this.archiveWriter = new BinaryWriter(this.workingStream);
                }
                else
                {
                    this.archiveReader = null;
                    this.archiveWriter = new BinaryWriter(this.workingStream);
                }

                this.centralDirectoryRead = false;
                this.leaveStreamOpen = leaveOpen;

                switch (mode)
                {
                    case ZipArchiveMode.Read:
                        this.ReadEndOfCentralDirectory();
                        break;

                    case ZipArchiveMode.Create:
                        this.centralDirectoryRead = true;
                        break;

                    case ZipArchiveMode.Update:
                        if (this.Reader.BaseStream.Length != 0L)
                        {
                            this.ReadEndOfCentralDirectory();
                            this.EnsureCentralDirectoryRead();

                            foreach (ZipArchiveEntry entry in this.Entries)
                            {
                                string message = null;
                                if (!entry.CheckIntegrity(out message))
                                {
                                    throw new InvalidDataException(message);
                                }
                            }
                        }
                        else
                        {
                            this.centralDirectoryRead = true;
                            break;
                        }
                        break;
                }
            }
            catch
            {
                if (this.originalStream != null)
                {
                    PlatformSettings.Manager.DeleteTemporaryStream(this.workingStream);
                }
                throw;
            }
        }

        private void ReadCentralDirectory()
        {
            this.Reader.BaseStream.Seek(this.CentralDirectoryStart, SeekOrigin.Begin);
            for (ulong i = 0; i < this.NumberOfEntriesInTheCentralDirectory; i++)
            {
                CentralDirectoryHeader zipCentralDirectoryHeader = new CentralDirectoryHeader();
                if (zipCentralDirectoryHeader.TryReadBlock(this.archiveReader))
                {
                    ZipArchiveEntry entry = new ZipArchiveEntry(this, zipCentralDirectoryHeader);
                    this.entries.Add(entry.FullName, entry);
                }
                else
                {
                    throw new InvalidDataException("Central directory header is broken.");
                }
            }
        }

        private void ReadEndOfCentralDirectory()
        {
            try
            {
                this.Reader.BaseStream.Seek((long)-EndOfCentralDirectoryRecord.StaticBlockLength, SeekOrigin.End);
                if (!ZipHelper.SeekBackwardsToSignature(this.Reader.BaseStream, EndOfCentralDirectoryRecord.Signature))
                {
                    throw new InvalidDataException("End of central directory not found.");
                }
                long position = this.Reader.BaseStream.Position;

                this.endOfCentralDirectoryRecord = new EndOfCentralDirectoryRecord();
                if (!this.endOfCentralDirectoryRecord.TryReadBlock(this.archiveReader))
                {
                    throw new InvalidDataException("End of central directory not found.");
                }

                if (this.endOfCentralDirectoryRecord.NumberOfThisDisk != this.endOfCentralDirectoryRecord.NumberOfTheDiskWithTheStartOfTheCentralDirectory)
                {
                    throw new InvalidDataException("Splitted archive is not supported.");
                }

                if (this.endOfCentralDirectoryRecord.NumberOfEntriesInTheCentralDirectory != this.endOfCentralDirectoryRecord.NumberOfEntriesInTheCentralDirectoryOnThisDisk)
                {
                    throw new InvalidDataException("Splitted archive is not supported.");
                }

                if (this.endOfCentralDirectoryRecord.NumberOfThisDisk == 0xFFFF
                    || this.endOfCentralDirectoryRecord.OffsetOfStartOfCentralDirectoryWithRespectToTheStartingDiskNumber == 0xFFFFFFFF
                    || this.endOfCentralDirectoryRecord.NumberOfEntriesInTheCentralDirectory == 0xFFFF)
                {
                    // ZIP64 specification should be used.
                    this.Reader.BaseStream.Seek(position - (long)Zip64EndOfCentralDirectoryLocator.StaticBlockLength, SeekOrigin.Begin);
                    if (ZipHelper.SeekBackwardsToSignature(this.Reader.BaseStream, Zip64EndOfCentralDirectoryLocator.Signature))
                    {
                        this.zip64EndOfCentralDirectoryLocator = new Zip64EndOfCentralDirectoryLocator();
                        if (!this.zip64EndOfCentralDirectoryLocator.TryReadBlock(this.archiveReader))
                        {
                            throw new InvalidDataException("ZIP64 End of central directory locator not found.");
                        }

                        if (this.zip64EndOfCentralDirectoryLocator.RelativeOffsetOfTheZip64EndOfCentralDirectoryRecord > 9223372036854775807L)
                        {
                            throw new InvalidDataException("Relative offset of the Zip64 End Of Central Directory Record is too big.");
                        }

                        long offsetOfZip64EOCD = (long)this.zip64EndOfCentralDirectoryLocator.RelativeOffsetOfTheZip64EndOfCentralDirectoryRecord;
                        this.Reader.BaseStream.Seek(offsetOfZip64EOCD, SeekOrigin.Begin);

                        this.zip64EndOfCentralDirectoryRecord = new Zip64EndOfCentralDirectoryRecord();
                        if (!this.zip64EndOfCentralDirectoryRecord.TryReadBlock(this.archiveReader))
                        {
                            throw new InvalidDataException("Zip64 End Of Central Directory Record not found.");
                        }

                        if (this.zip64EndOfCentralDirectoryRecord.NumberOfEntriesInTheCentralDirectory > 9223372036854775807L)
                        {
                            throw new InvalidDataException("Number of entries is too big.");
                        }

                        if (this.zip64EndOfCentralDirectoryRecord.OffsetOfStartOfCentralDirectoryWithRespectToTheStartingDiskNumber > 9223372036854775807L)
                        {
                            throw new InvalidDataException("Relative offset of the Central Directory Start is too big.");
                        }

                        if (this.zip64EndOfCentralDirectoryRecord.NumberOfEntriesInTheCentralDirectory != this.zip64EndOfCentralDirectoryRecord.NumberOfEntriesInTheCentralDirectoryOnThisDisk)
                        {
                            throw new InvalidDataException("Splitted archive is not supported.");
                        }
                    }
                }
                if (this.CentralDirectoryStart > this.Reader.BaseStream.Length)
                {
                    throw new InvalidDataException("Relative offset of the Central Directory Start is too big.");
                }
            }
            catch (EndOfStreamException endOfStreamException)
            {
                throw new InvalidDataException("Archive corrupted", endOfStreamException);
            }
            catch (IOException exc)
            {
                throw new InvalidDataException("Archive corrupted", exc);
            }
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
        private void OnPropertyChanged(string propertyName)
        {
            if (this.PropertyChanged != null)
            {
                this.PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
            }
        }

        /// <summary>
        /// Writes archive to the original stream.
        /// </summary>
        private void WriteArchive()
        {
            if (this.Mode == ZipArchiveMode.Update)
            {
                foreach (ZipArchiveEntry entry in this.entries.Values)
                {
                    entry.Dispose();
                }
            }

            long startOfCentralDirectory = this.Writer.BaseStream.Position;
            foreach (ZipArchiveEntry entry in this.entries.Values)
            {
                entry.WriteCentralDirectoryHeader();
            }

            long sizeOfTheCentralDirectory = this.Writer.BaseStream.Position - startOfCentralDirectory;

            bool useZip64 = false;
            if (startOfCentralDirectory >= uint.MaxValue
                || sizeOfTheCentralDirectory >= uint.MaxValue
                || this.entries.Count >= ushort.MaxValue)
            {
                useZip64 = true;
            }

            if (useZip64)
            {
                long position = this.Writer.BaseStream.Position;

                if (this.zip64EndOfCentralDirectoryRecord == null)
                {
                    this.zip64EndOfCentralDirectoryRecord = new Zip64EndOfCentralDirectoryRecord();
                }

                this.zip64EndOfCentralDirectoryRecord.NumberOfEntriesInTheCentralDirectory = (ulong)this.entries.Count;
                this.zip64EndOfCentralDirectoryRecord.NumberOfEntriesInTheCentralDirectoryOnThisDisk = (ulong)this.entries.Count;
                this.zip64EndOfCentralDirectoryRecord.OffsetOfStartOfCentralDirectoryWithRespectToTheStartingDiskNumber = (ulong)startOfCentralDirectory;
                this.zip64EndOfCentralDirectoryRecord.SizeOfCentralDirectory = (ulong)sizeOfTheCentralDirectory;
                this.zip64EndOfCentralDirectoryRecord.WriteBlock(this.Writer);

                if (this.zip64EndOfCentralDirectoryLocator == null)
                {
                    this.zip64EndOfCentralDirectoryLocator = new Zip64EndOfCentralDirectoryLocator();
                }

                this.zip64EndOfCentralDirectoryLocator.NumberOfTheDiskWithTheStartOfTheZip64EndOfCentralDirectory = 0;
                this.zip64EndOfCentralDirectoryLocator.NumberOfDisks = 1;
                this.zip64EndOfCentralDirectoryLocator.RelativeOffsetOfTheZip64EndOfCentralDirectoryRecord = (ulong)position;
                this.zip64EndOfCentralDirectoryLocator.WriteBlock(this.Writer);
            }

            if (this.endOfCentralDirectoryRecord == null)
            {
                this.endOfCentralDirectoryRecord = new EndOfCentralDirectoryRecord();
            }

            this.endOfCentralDirectoryRecord.NumberOfThisDisk = 0;
            this.endOfCentralDirectoryRecord.NumberOfTheDiskWithTheStartOfTheCentralDirectory = 0;
            this.endOfCentralDirectoryRecord.NumberOfEntriesInTheCentralDirectoryOnThisDisk = this.entries.Count < ushort.MaxValue ? (ushort)this.entries.Count : ushort.MaxValue;
            this.endOfCentralDirectoryRecord.NumberOfEntriesInTheCentralDirectory = this.entries.Count < ushort.MaxValue ? (ushort)this.entries.Count : ushort.MaxValue;
            this.endOfCentralDirectoryRecord.OffsetOfStartOfCentralDirectoryWithRespectToTheStartingDiskNumber = (uint)(startOfCentralDirectory > uint.MaxValue ? uint.MaxValue : startOfCentralDirectory);
            this.endOfCentralDirectoryRecord.SizeOfCentralDirectory = (uint)(sizeOfTheCentralDirectory > uint.MaxValue ? uint.MaxValue : sizeOfTheCentralDirectory);

            this.endOfCentralDirectoryRecord.WriteBlock(this.Writer);

            if (this.Mode == ZipArchiveMode.Update)
            {
                this.workingStream.Seek(0L, SeekOrigin.Begin);
                this.originalStream.Seek(0L, SeekOrigin.Begin);

#if (WPF || SILVERLIGHT || NETSTANDARD)
                this.workingStream.CopyTo(this.originalStream);
#else
                Telerik.WinControls.Zip.Extensions.ZipFile.CopyStreamTo(workingStream, originalStream);
#endif

                this.originalStream.SetLength(this.workingStream.Length);
            }
        }
    }
}
