﻿using System;
using System.ComponentModel;
using System.IO;
using System.Text;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Represents a compressed file within a zip archive.
    /// </summary>
    public class ZipArchiveEntry : IDisposable, INotifyPropertyChanged
    {
        private ZipArchive archiveRef = null;
        private CentralDirectoryHeader header = null;
        private CompressionSettings settings = null;
        private CompressedStream compressedData = null;
        private bool written = false;
        private bool existedInArchive = false;
        private LocalFileHeader localFileHeader = null;
        private long? offsetOfCompressedData = null;
        private Stream updatableData = null;
        private bool deleted = false;

        /// <summary>
        /// Track whether Dispose has been called. 
        /// </summary>
        private bool disposed = false;

        /// <summary>
        /// Initializes a new instance of the ZipArchiveEntry class.
        /// </summary>
        /// <param name="archive">Zip archive.</param>
        /// <param name="header">Central directory header correspondent to this entry.</param>
        internal ZipArchiveEntry(ZipArchive archive, CentralDirectoryHeader header)
        {
            this.Archive = archive;
            this.header = header;
            this.settings = ZipHelper.GetCompressionSettings((CompressionMethod)this.header.CompressionMethod, this.Archive.CompressionSettings);
            this.settings.PrepareForZip();

            this.existedInArchive = true;
        }

        /// <summary>
        /// Initializes a new instance of the ZipArchiveEntry class.
        /// </summary>
        /// <param name="archive">Zip archive.</param>
        /// <param name="entryName">Entry name.</param>
        internal ZipArchiveEntry(ZipArchive archive, string entryName)
        {
            this.Archive = archive;
            this.settings = this.Archive.CompressionSettings;
            this.header = new CentralDirectoryHeader();
            this.header.VersionNeededToExtract = (ushort)VersionNeededToExtract.Default;
            this.header.GeneralPurposeBitFlag = (ushort)GeneralPurposeBitFlag.ZeroLocalHeader;
            if (this.Archive.EncryptionSettings is DefaultEncryptionSettings)
            {
                this.header.GeneralPurposeBitFlag |= (ushort)GeneralPurposeBitFlag.FileIsEncrypted;
                this.ValidateVersionNeeded(VersionNeededToExtract.TraditionalEncryption);
            }
            this.FullName = entryName;
            this.LastWriteTime = DateTimeOffset.Now;
        }

        internal ZipArchiveEntry(ZipArchive archive, string entryName, CompressionSettings settings)
            : this(archive, entryName)
        {
            this.settings = settings;
            this.settings.PrepareForZip();
            this.CompressionMethod = settings.Method;
        }

        /// <summary>
        /// Occurs when a property value changes.
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        /// <summary>
        /// The zip archive that the entry belongs to, or null if the entry has been deleted.
        /// </summary>
        public ZipArchive Archive
        {
            get
            {
                return this.archiveRef;
            }

            set
            {
                this.archiveRef = value;
            }
        }

        /// <summary>
        /// Gets compressed size of the entry in the zip archive.
        /// </summary>
        public long CompressedLength
        {
            get
            {
                this.EnsureCentralDirectoryHeader();
                Zip64ExtraField zip64ExtraField = this.SelectZip64ExtraField();
                if (zip64ExtraField != null && zip64ExtraField.CompressedSize.HasValue)
                {
                    return (long)zip64ExtraField.CompressedSize.Value;
                }
                else
                {
                    return this.header.CompressedSize;
                }
            }

            private set
            {
                this.EnsureCentralDirectoryHeader();
                this.ValidateArchiveModeForWriting();

                if (value >= uint.MaxValue)
                {
                    Zip64ExtraField zip64ExtraField = this.EnsureZip64ExtraField();
                    zip64ExtraField.CompressedSize = (ulong)value;
                    this.header.CompressedSize = uint.MaxValue;
                }
                else
                {
                    this.header.CompressedSize = (uint)value;
                }
            }
        }

        /// <summary>
        /// Gets or sets external file attributes.
        /// </summary>
        public int ExternalAttributes
        {
            get
            {
                this.EnsureCentralDirectoryHeader();
                return (int)this.header.ExternalAttributes;
            }

            set
            {
                this.EnsureCentralDirectoryHeader();
                this.header.ExternalAttributes = (uint)value;
                this.OnPropertyChanged("ExternalAttributes");
            }
        }

        /// <summary>
        /// Gets the relative path of the entry in the zip archive.
        /// </summary>
        public string FullName
        {
            get
            {
                this.EnsureCentralDirectoryHeader();
                return this.DecodeEntryName(this.header.FileName);
            }

            private set
            {
                if (value == null)
                {
                    throw new ArgumentNullException("value");
                }
                this.EnsureCentralDirectoryHeader();
                string name = value;
                name = name.Replace(PlatformSettings.Manager.DirectorySeparatorChar, "/");
                name = name.Replace(PlatformSettings.Manager.AltDirectorySeparatorChar, "/");
                this.header.FileName = this.EncodeEntryName(name);
                if (ZipHelper.EndsWithDirChar(value))
                {
                    this.ValidateVersionNeeded(VersionNeededToExtract.ExplicitDirectory);
                }
                this.OnPropertyChanged("FullName");
            }
        }

        /// <summary>
        /// Gets or sets the last time the entry in the zip archive was changed.
        /// </summary>
        public DateTimeOffset LastWriteTime
        {
            get
            {
                this.EnsureCentralDirectoryHeader();
                DateTime dt = ZipHelper.PackedToDateTime(this.header.FileTime);
                DateTimeOffset writeTime = new DateTimeOffset(dt);
                return writeTime;
            }

            set
            {
                this.EnsureCentralDirectoryHeader();
                this.ValidateArchiveModeForWriting();
                if (value.DateTime.Year < 1980 || value.DateTime.Year > 2107)
                {
                    throw new ArgumentOutOfRangeException("value", "Date-Time out of range.");
                }
                DateTime dt = value.DateTime;
                uint fileTime = ZipHelper.DateTimeToPacked(dt);
                this.header.FileTime = fileTime;
            }
        }

        /// <summary>
        /// Gets the uncompressed size of the entry in the zip archive.
        /// </summary>
        public long Length
        {
            get
            {
                this.EnsureCentralDirectoryHeader();
                Zip64ExtraField zip64ExtraField = this.SelectZip64ExtraField();
                if (zip64ExtraField != null && zip64ExtraField.OriginalSize.HasValue)
                {
                    return (long)zip64ExtraField.OriginalSize.Value;
                }
                else
                {
                    return this.header.UncompressedSize;
                }
            }

            private set
            {
                this.EnsureCentralDirectoryHeader();
                this.ValidateArchiveModeForWriting();

                if (value >= uint.MaxValue)
                {
                    Zip64ExtraField zip64ExtraField = this.EnsureZip64ExtraField();
                    zip64ExtraField.OriginalSize = (ulong)value;
                    this.header.UncompressedSize = uint.MaxValue;
                }
                else
                {
                    this.header.UncompressedSize = (uint)value;
                }
            }
        }

        /// <summary>
        /// Gets the file name of the entry in the zip archive.
        /// </summary>
        public string Name
        {
            get
            {
                string[] segments = this.FullName.Split(
                    new string[] 
                    {
                        PlatformSettings.Manager.DirectorySeparatorChar,
                        PlatformSettings.Manager.AltDirectorySeparatorChar 
                    },
                    StringSplitOptions.None);
                return segments[segments.Length - 1];
            }
        }

        /// <summary>
        /// Gets or sets compression method.
        /// </summary>
        internal CompressionMethod CompressionMethod
        {
            get
            {
                this.EnsureCentralDirectoryHeader();
                return (CompressionMethod)this.header.CompressionMethod;
            }

            set
            {
                this.EnsureCentralDirectoryHeader();
                if (value == Zip.CompressionMethod.Deflate)
                {
                    this.ValidateVersionNeeded(VersionNeededToExtract.Deflate);
                }
                this.header.CompressionMethod = (ushort)value;
            }
        }

        /// <summary>
        /// Gets or sets offset of the compressed data.
        /// </summary>
        private long CompressedDataOffset
        {
            get
            {
                if (!this.offsetOfCompressedData.HasValue)
                {
                    this.Archive.Reader.BaseStream.Seek(this.LocalHeaderOffset, SeekOrigin.Begin);
                    this.localFileHeader = new LocalFileHeader();
                    if (!this.localFileHeader.TryReadBlock(this.Archive.Reader))
                    {
                        throw new InvalidDataException("Local file header is corrupted.");
                    }
                    this.offsetOfCompressedData = new long?(this.Archive.Reader.BaseStream.Position);
                }
                return this.offsetOfCompressedData.Value;
            }

            set
            {
                this.offsetOfCompressedData = value;
            }
        }

        /// <summary>
        /// Gets disk start number.
        /// </summary>
        private uint DiskStartNumber
        {
            get
            {
                this.EnsureCentralDirectoryHeader();
                Zip64ExtraField zip64ExtraField = this.SelectZip64ExtraField();
                if (zip64ExtraField != null && zip64ExtraField.DiskStartNumber.HasValue)
                {
                    return zip64ExtraField.DiskStartNumber.Value;
                }
                else
                {
                    return this.header.DiskNumberStart;
                }
            }
        }

        /// <summary>
        /// Gets or sets offset of the local header.
        /// </summary>
        private long LocalHeaderOffset
        {
            get
            {
                this.EnsureCentralDirectoryHeader();
                Zip64ExtraField zip64ExtraField = this.SelectZip64ExtraField();
                if (zip64ExtraField != null && zip64ExtraField.RelativeHeaderOffset.HasValue)
                {
                    return (long)zip64ExtraField.RelativeHeaderOffset.Value;
                }
                else
                {
                    return (long)this.header.LocalHeaderOffset;
                }
            }

            set
            {
                this.EnsureCentralDirectoryHeader();
                this.ValidateArchiveModeForWriting();

                if (value >= uint.MaxValue)
                {
                    Zip64ExtraField zip64ExtraField = this.EnsureZip64ExtraField();
                    zip64ExtraField.RelativeHeaderOffset = (ulong)value;
                    this.header.LocalHeaderOffset = uint.MaxValue;
                }
                else
                {
                    this.header.LocalHeaderOffset = (uint)value;
                }
            }
        }

        /// <summary>
        /// Gets temporary stream which contains uncompressed data for update.
        /// </summary>
        private Stream UpdatableData
        {
            get
            {
                if (this.updatableData == null)
                {
                    this.updatableData = PlatformSettings.Manager.CreateTemporaryStream();
                    if (this.existedInArchive)
                    {
                        using (Stream stream = this.OpenForReading())
                        {
                            try
                            {
#if (WPF || SILVERLIGHT || NETSTANDARD)
                stream.CopyTo(this.updatableData);
#else
                Telerik.WinControls.Zip.Extensions.ZipFile.CopyStreamTo(stream, updatableData);
#endif
                            }
                            catch (InvalidDataException)
                            {
                                this.updatableData.Dispose();
                                PlatformSettings.Manager.DeleteTemporaryStream(this.updatableData);
                                this.updatableData = null;
                                this.written = false;
                                throw;
                            }
                        }
                    }
                }
                return this.updatableData;
            }
        }

        /// <summary>
        /// Deletes the entry from the zip archive.
        /// </summary>
        /// <exception cref="IOException">The entry is already open for reading or writing.</exception>
        /// <exception cref="NotSupportedException">The zip archive for this entry was opened in a mode other than Update.</exception>
        /// <exception cref="ObjectDisposedException">The zip archive for this entry has been disposed.</exception>
        public void Delete()
        {
            if (this.Archive == null)
            {
                return;
            }

            if (this.Archive.Mode != ZipArchiveMode.Update)
            {
                throw new NotSupportedException("Entry can be deleted in Update mode only.");
            }

            this.Archive.ThrowIfDisposed();
            this.Archive.RemoveEntry(this);
            this.Archive = null;
            this.deleted = true;
            this.DisposeStreams();
        }

        /// <summary>
        /// Releases the resources used by the current instance of the ZipArchiveEntry class.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Opens the entry from the zip archive.
        /// </summary>
        /// <returns>The stream that represents the contents of the entry.</returns>
        /// <remarks>The resulting stream depends on the zip archive mode.
        /// If zip archive mode is <see cref="ZipArchiveMode.Create"/> then read-only stream without seeking support is returned (<see cref="CompressedStream"/>).
        /// If zip archive mode is <see cref="ZipArchiveMode.Read"/> then write-only stream without seeking support is returned (<see cref="CompressedStream"/>).
        /// If zip archive mode is <see cref="ZipArchiveMode.Update"/> then read/write stream which supports seeking is returned.
        /// </remarks>
        /// <exception cref="IOException">The entry is already currently open for writing.
        /// -or-
        /// The entry has been deleted from the archive.
        /// -or-
        /// The archive for this entry was opened with the Create mode, and this entry has already been written to.</exception>
        /// <exception cref="NotSupportedException">The entry is either missing from the archive or is corrupt and cannot be read. 
        /// -or-
        /// The entry has been compressed by using a compression method that is not supported.</exception>
        /// <exception cref="ObjectDisposedException">The zip archive for this entry has been disposed.</exception>
        public Stream Open()
        {
            switch (this.Archive.Mode)
            {
                case ZipArchiveMode.Read:
                    return this.OpenForReading();

                case ZipArchiveMode.Create:
                    return this.OpenForWriting();

                case ZipArchiveMode.Update:
                    return this.OpenForUpdate();
            }

            return null;
        }

        /// <summary>
        /// Checks entry integrity.
        /// </summary>
        /// <param name="message">Message will be thrown if entry don't pass integrity check.</param>
        /// <returns>True - if entry is OK; false - otherwise.</returns>
        internal bool CheckIntegrity(out string message)
        {
            message = null;
            if (this.existedInArchive)
            {
                if (!ZipHelper.IsCompressionMethodSupported(this.CompressionMethod))
                {
                    message = string.Format("Unsupported compression method: {0}.", this.CompressionMethod);
                    return false;
                }

                if (this.DiskStartNumber != this.Archive.NumberOfThisDisk)
                {
                    message = "Splitted archive is not supported.";
                    return false;
                }

                if (this.LocalHeaderOffset > this.Archive.Reader.BaseStream.Length)
                {
                    message = "Local file header is corrupted.";
                    return false;
                }

                this.Archive.Reader.BaseStream.Seek(this.LocalHeaderOffset, SeekOrigin.Begin);
                this.localFileHeader = new LocalFileHeader();

                if (!this.localFileHeader.TryReadBlock(this.Archive.Reader))
                {
                    message = "Local file header is corrupted.";
                    return false;
                }

                if (!this.ValidateLocalFileHeader())
                {
                    message = "Local file header is corrupted.";
                    return false;
                }

                if (this.CompressedDataOffset + this.CompressedLength > this.Archive.Reader.BaseStream.Length)
                {
                    message = "Local file header is corrupted.";
                    return false;
                }
            }
            return true;
        }

        /// <summary>
        /// Writes central directory header.
        /// </summary>
        internal void WriteCentralDirectoryHeader()
        {
            this.header.WriteBlock(this.Archive.Writer);
        }

        private void CompressedData_ChecksumReady(object sender, EventArgs e)
        {
            if (this.Archive.Mode == ZipArchiveMode.Create
                || this.Archive.Mode == ZipArchiveMode.Update)
            {
                this.header.Crc = (uint)this.compressedData.Checksum;
            }
            else
            {
                // Check CRC when necessary.
            }
        }

        private string DecodeEntryName(byte[] entryNameBytes)
        {
            Encoding encoding = this.GetEntryNameEncoding();
            return encoding.GetString(entryNameBytes, 0, entryNameBytes.Length);
        }

        /// <summary>
        /// Called by the Dispose() and Finalize() methods to release the unmanaged 
        /// resources used by the current instance of the ZipArchive class, and optionally 
        /// finishes writing the archive and releases the managed resources.
        /// </summary>
        /// <param name="disposing">True to finish writing the archive and release unmanaged and managed resources; 
        /// false to release only unmanaged resources.</param>
        private void Dispose(bool disposing)
        {
            if (!this.disposed)
            {
                if (disposing)
                {
                    this.FlushEntryData();
                    this.DisposeStreams();
                }

                this.disposed = true;
            }
        }

        private void DisposeStreams()
        {
            if (this.compressedData != null)
            {
                if (!this.deleted)
                {
                    this.compressedData.Dispose();
                }
                this.compressedData.ChecksumReady -= this.CompressedData_ChecksumReady;
                this.compressedData = null;
            }

            if (this.updatableData != null)
            {
                this.updatableData.Dispose();
                PlatformSettings.Manager.DeleteTemporaryStream(this.updatableData);
                this.updatableData = null;
            }
        }

        private byte[] EncodeEntryName(string entryName)
        {
            if (!string.IsNullOrEmpty(entryName))
            {
                Encoding entryNameEncoding = this.Archive == null ?
                    PlatformSettings.Manager.DefaultEncoding
                    : this.Archive.EntryNameEncoding ?? PlatformSettings.Manager.DefaultEncoding;
                bool utf8 = entryNameEncoding is UTF8Encoding ? entryNameEncoding.Equals(Encoding.UTF8) : false;
                if (utf8)
                {
                    this.header.GeneralPurposeBitFlag = (ushort)((GeneralPurposeBitFlag)this.header.GeneralPurposeBitFlag | GeneralPurposeBitFlag.EncodingUtf8);
                }
                else
                {
                    this.header.GeneralPurposeBitFlag = (ushort)((GeneralPurposeBitFlag)this.header.GeneralPurposeBitFlag & ~GeneralPurposeBitFlag.EncodingUtf8);
                }
                return entryNameEncoding.GetBytes(entryName);
            }
            else
            {
                return new byte[0];
            }
        }

        private void EnsureCentralDirectoryHeader()
        {
            if (this.header == null)
            {
                this.header = new CentralDirectoryHeader();
            }
        }

        private Zip64ExtraField EnsureZip64ExtraField()
        {
            Zip64ExtraField zip64ExtraField = null;
            
            foreach (ExtraFieldBase ef in this.header.ExtraFields)
            {
                zip64ExtraField = ef as Zip64ExtraField;
                if (zip64ExtraField != null)
                {
                    break;
                }
            }

            if (zip64ExtraField == null)
            {
                zip64ExtraField = new Zip64ExtraField(new CentralDirectoryHeaderInfo(this.header));
                this.header.ExtraFields.Add(zip64ExtraField);
            }

            return zip64ExtraField;
        }

        private void FlushEntryData()
        {
            if (this.Archive.Mode == ZipArchiveMode.Create)
            {
                if (this.compressedData != null)
                {
                    this.FlushFinalBlockAndDataDescriptor();
                }
            }
            else if (this.Archive.Mode == ZipArchiveMode.Update)
            {
                if (this.updatableData != null)
                {
                    this.OpenForWriting();
                    this.updatableData.Seek(0L, SeekOrigin.Begin);

#if (WPF || SILVERLIGHT || NETSTANDARD)
                    this.updatableData.CopyTo(this.compressedData);
#else
                    Telerik.WinControls.Zip.Extensions.ZipFile.CopyStreamTo(updatableData, compressedData);
#endif

                    this.FlushFinalBlockAndDataDescriptor();
                }
                else
                {
                    long dataOffset = this.CompressedDataOffset;
                    this.WriteLocalFileHeader();
                    this.Archive.Reader.BaseStream.Seek(dataOffset, SeekOrigin.Begin);
                    ZipHelper.CopyStream(this.Archive.Reader.BaseStream, this.Archive.Writer.BaseStream, this.CompressedLength);
                    this.WriteDataDescriptor();
                }
            }
        }

        private void FlushFinalBlockAndDataDescriptor()
        {
            if (!this.compressedData.HasFlushedFinalBlock)
            {
                this.compressedData.Flush();
            }
            this.CompressedLength = this.compressedData.CompressedSize;
            this.Length = this.compressedData.TotalPlainCount;
            this.WriteDataDescriptor();
        }

        private Encoding GetEntryNameEncoding()
        {
            Encoding entryNameEncoding;
            if (((GeneralPurposeBitFlag)this.header.GeneralPurposeBitFlag & GeneralPurposeBitFlag.EncodingUtf8) != 0)
            {
                entryNameEncoding = Encoding.UTF8;
            }
            else
            {
                entryNameEncoding = this.Archive == null ?
                    PlatformSettings.Manager.DefaultEncoding
                    : this.Archive.EntryNameEncoding ?? PlatformSettings.Manager.DefaultEncoding;
            }
            return entryNameEncoding;
        }

        private void OnPropertyChanged(string propertyName)
        {
            if (this.PropertyChanged != null)
            {
                this.PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
            }
        }

        private Stream OpenForReading()
        {
            string message = null;
            if (!this.CheckIntegrity(out message))
            {
                throw new InvalidDataException(message);
            }

            this.Archive.Reader.BaseStream.Position = this.CompressedDataOffset;

            this.compressedData = new CompressedStream(this.Archive.Reader.BaseStream, StreamOperationMode.Read, this.settings, true, this.Archive.EncryptionSettings);
            this.compressedData.ChecksumReady += this.CompressedData_ChecksumReady;
            this.compressedData.SetLength(this.CompressedLength);
            return this.compressedData;
        }

        private Stream OpenForUpdate()
        {
            if (this.written)
            {
                throw new IOException("Create mode writes entry once and one entry at a time");
            }

            this.UpdatableData.Seek(0L, SeekOrigin.Begin);
            return this.UpdatableData;
        }

        private Stream OpenForWriting()
        {
            if (this.written)
            {
                throw new IOException("Create mode writes entry once and one entry at a time");
            }

            this.WriteLocalFileHeader();

            if (this.Archive.EncryptionSettings != null)
            {
                DefaultEncryptionSettings encryptionSettings = this.Archive.EncryptionSettings as DefaultEncryptionSettings;
                if (encryptionSettings != null)
                {
                    encryptionSettings.FileTime = this.header.FileTime;
                }
            }

            this.compressedData = new CompressedStream(this.Archive.Writer.BaseStream, StreamOperationMode.Write, this.settings, true, this.Archive.EncryptionSettings);
            this.compressedData.ChecksumReady += this.CompressedData_ChecksumReady;
            return this.compressedData;
        }

        private Zip64ExtraField SelectZip64ExtraField()
        {
            Zip64ExtraField zip64ExtraField = null;

            if (this.header.ExtraFields.Count > 0)
            {
                foreach (ExtraFieldBase ef in this.header.ExtraFields)
                {
                    zip64ExtraField = ef as Zip64ExtraField;
                    if (zip64ExtraField != null)
                    {
                        break;
                    }
                }
            }

            return zip64ExtraField;
        }

        private bool ValidateLocalFileHeader()
        {
            return this.localFileHeader != null;
        }

        private void ValidateVersionNeeded(VersionNeededToExtract value)
        {
            if (this.header.VersionNeededToExtract < (ushort)value)
            {
                this.header.VersionNeededToExtract = (ushort)value;
            }
        }

        private void ValidateArchiveModeForWriting()
        {
            if (this.Archive.Mode == ZipArchiveMode.Read)
            {
                throw new NotSupportedException("Read only archive.");
            }
            if (this.Archive.Mode == ZipArchiveMode.Create && this.written)
            {
                throw new IOException("Entry is frozen after write.");
            }
        }

        private void WriteDataDescriptor()
        {
            if (this.CompressedLength >= uint.MaxValue
                || this.Length >= uint.MaxValue)
            {
                this.header.VersionNeededToExtract = (ushort)VersionNeededToExtract.Zip64;
            }
            DataDescriptor descr = DataDescriptor.FromFileHeader(this.header as FileHeaderBase);
            descr.WriteBlock(this.Archive.Writer);
        }

        private void WriteLocalFileHeader()
        {
            this.LocalHeaderOffset = this.Archive.Writer.BaseStream.Position;
            LocalFileHeader localHeader = new LocalFileHeader(this.header);
            localHeader.WriteBlock(this.Archive.Writer);
            this.CompressedDataOffset = this.Archive.Writer.BaseStream.Position;
        }
    }
}
