﻿using System;
using System.IO;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Provides common internal static methods.
    /// </summary>
    internal static class ZipHelper
    {
        internal static readonly DateTime InvalidDateTime = new DateTime(1980, 1, 1, 0, 0, 0, 0);

        /// <summary>
        /// Copy specified number of bytes from one stream to another.
        /// </summary>
        /// <param name="input">Input stream.</param>
        /// <param name="output">Output stream.</param>
        /// <param name="bytes">Number of bytes to copy.</param>
        internal static void CopyStream(Stream input, Stream output, long bytes)
        {
            int bufferSize = 32768;
            byte[] buffer = new byte[bufferSize];
            int read;
            while (bytes > 0 &&
                   (read = input.Read(buffer, 0, (int)Math.Min(bufferSize, bytes))) > 0)
            {
                output.Write(buffer, 0, read);
                bytes -= read;
            }
        }

        /// <summary>
        /// Converts .NET DateTime structure to the MS-DOS date-time.
        /// </summary>
        /// <param name="time">DateTime structure to convert.</param>
        /// <returns>Packed date-time.</returns>
        internal static uint DateTimeToPacked(DateTime time)
        {
            // see http://www.vsft.com/hal/dostime.htm for the format
            ushort packedDate = (ushort)((time.Day & 0x0000001F)
                | ((time.Month << 5) & 0x000001E0)
                | (((time.Year - 1980) << 9) & 0x0000FE00));
            ushort packedTime = (ushort)((time.Second / 2 & 0x0000001F)
                | ((time.Minute << 5) & 0x000007E0)
                | ((time.Hour << 11) & 0x0000F800));

            uint result = ((uint)(packedDate << 16)) | packedTime;
            return result;
        }

        /// <summary>
        /// Gets compression algorithm which corresponds to the given compression settings.
        /// </summary>
        /// <param name="settings">Compression settings to get algorithm for.</param>
        /// <returns>Compression algorithm.</returns>
        /// <exception cref="NotSupportedException">Compression method is not supported.</exception>
        internal static ICompressionAlgorithm GetCompressionAlgorithm(CompressionSettings settings)
        {
            ICompressionAlgorithm algorithm = null;

            switch (settings.Method)
            {
                case CompressionMethod.Stored:
                    algorithm = new StoreAlgorithm();
                    break;

                case CompressionMethod.Deflate:
                    algorithm = new DeflateAlgorithm();
                    break;

                case CompressionMethod.Lzma:
                    algorithm = new LzmaAlgorithm();
                    break;

                default:
                    throw new NotSupportedException(string.Format("Compression method {0} is not supported.", settings));
            }

            if (algorithm != null)
            {
                algorithm.Initialize(settings);
            }

            return algorithm;
        }

        /// <summary>
        /// Gets compression settings for the specified compression method.
        /// </summary>
        /// <param name="method">Compression method to get settings for.</param>
        /// <param name="baseSettings">Base settings to copy parameters from.</param>
        /// <returns>Compression settings.</returns>
        /// <exception cref="NotSupportedException">Compression method is not supported.</exception>
        internal static CompressionSettings GetCompressionSettings(CompressionMethod method, CompressionSettings baseSettings)
        {
            CompressionSettings settings = null;

            switch (method)
            {
                case CompressionMethod.Stored:
                    settings = new StoreSettings();
                    break;

                case CompressionMethod.Deflate:
                    settings = new DeflateSettings();
                    break;

                case CompressionMethod.Lzma:
                    settings = new LzmaSettings();
                    break;

                default:
                    throw new NotSupportedException(string.Format("Compression method {0} is not supported.", method));
            }

            if (baseSettings != null)
            {
                if (settings.GetType() == baseSettings.GetType())
                {
                    settings.CopyFrom(baseSettings);
                }
            }

            return settings;
        }

        /// <summary>
        /// Detect whether the given path string ends with directory separator char (i.e. given path represents directory).
        /// </summary>
        /// <param name="path">Path string.</param>
        /// <returns>True if path string ends with directory separator char; otherwise - false.</returns>
        internal static bool EndsWithDirChar(string path)
        {
            string trimmed = path.Trim();
            return trimmed.EndsWith(PlatformSettings.Manager.DirectorySeparatorChar) 
                || trimmed.EndsWith(PlatformSettings.Manager.AltDirectorySeparatorChar);
        }

        /// <summary>
        /// Gets value which indicates whether specified compression method is supported.
        /// </summary>
        /// <param name="method">Compression method to check.</param>
        /// <returns>True - if compression method is supported; false - otherwise.</returns>
        internal static bool IsCompressionMethodSupported(CompressionMethod method)
        {
            return method == CompressionMethod.Deflate
                || method == CompressionMethod.Stored
                || method == CompressionMethod.Lzma;
        }

        /// <summary>
        /// Converts MS-DOS date-time to the .NET DateTime structure.
        /// </summary>
        /// <param name="packedDateTime">Packed date-time to convert.</param>
        /// <returns>DataTime structure.</returns>
        internal static DateTime PackedToDateTime(uint packedDateTime)
        {
            if (packedDateTime == 0xFFFF || packedDateTime == 0)
            {
                return new System.DateTime(1995, 1, 1, 0, 0, 0, 0);  // return a fixed date when none is supplied.
            }

            ushort packedTime = unchecked((ushort)(packedDateTime & 0x0000ffff));
            ushort packedDate = unchecked((ushort)((packedDateTime & 0xffff0000) >> 16));

            int year = 1980 + ((packedDate & 0xFE00) >> 9);
            int month = (packedDate & 0x01E0) >> 5;
            int day = packedDate & 0x001F;

            int hour = (packedTime & 0xF800) >> 11;
            int minute = (packedTime & 0x07E0) >> 5;
            int second = (packedTime & 0x001F) * 2;

            // validation and error checking.
            // this is not foolproof but will catch most errors.
            if (second >= 60) 
            { 
                minute += second / 60;
                second = second % 60; 
            }

            if (minute >= 60) 
            { 
                hour += minute / 60; 
                minute = minute % 60; 
            }

            if (hour >= 24) 
            { 
                day += hour / 24; 
                hour = hour % 24; 
            }

            DateTime dateTime = System.DateTime.Now;
            try
            {
                dateTime = new System.DateTime(year, month, day, hour, minute, second, 0);
            }
            catch (System.ArgumentOutOfRangeException)
            {
                dateTime = ZipHelper.InvalidDateTime;
            }
            catch (System.ArgumentException)
            {
                dateTime = ZipHelper.InvalidDateTime;
            }

            return dateTime;
        }

        /// <summary>
        /// Read specified number of bytes from the given stream to the buffer.
        /// </summary>
        /// <param name="stream">Stream to read data from.</param>
        /// <param name="buffer">Buffer to write data to.</param>
        /// <param name="bytesToRead">Number of bytes to read.</param>
        internal static void ReadBytes(Stream stream, byte[] buffer, int bytesToRead)
        {
            int offset = 0;
            while (bytesToRead > 0)
            {
                int count = stream.Read(buffer, offset, bytesToRead);
                if (count == 0)
                {
                    throw new IOException("Unexpected End Of Stream");
                }
                offset = offset + count;
                bytesToRead = bytesToRead - count;
            }
        }

        /// <summary>
        /// Seek given stream backward to the data signature.
        /// </summary>
        /// <param name="stream">Stream to seek.</param>
        /// <param name="signatureToFind">Signature to find.</param>
        /// <returns>true if signature is found, otherwise false.</returns>
        internal static bool SeekBackwardsToSignature(Stream stream, uint signatureToFind)
        {
            int bufferPointer = 0;
            uint bufferValue = 0;
            byte[] numArray = new byte[32];
            bool beginOfStream = false;
            bool found = false;
            while (!found && !beginOfStream)
            {
                beginOfStream = ZipHelper.SeekBackwardsAndRead(stream, numArray, out bufferPointer);
                while (bufferPointer >= 0 && !found)
                {
                    bufferValue = bufferValue << 8 | numArray[bufferPointer];
                    if (bufferValue != signatureToFind)
                    {
                        bufferPointer--;
                    }
                    else
                    {
                        found = true;
                    }
                }
            }
            if (!found)
            {
                return false;
            }
            stream.Seek((long)bufferPointer, SeekOrigin.Current);
            return true;
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
        private static bool SeekBackwardsAndRead(Stream stream, byte[] buffer, out int bufferPointer)
        {
            if (stream.Position < (long)buffer.Length)
            {
                int position = (int)stream.Position;
                stream.Seek(0L, SeekOrigin.Begin);
                ZipHelper.ReadBytes(stream, buffer, position);
                stream.Seek(0L, SeekOrigin.Begin);
                bufferPointer = position - 1;
                return true;
            }
            stream.Seek((long)(-buffer.Length), SeekOrigin.Current);
            ZipHelper.ReadBytes(stream, buffer, buffer.Length);
            stream.Seek((long)(-buffer.Length), SeekOrigin.Current);
            bufferPointer = buffer.Length - 1;
            return false;
        }
    }
}
