﻿using System;
using System.IO;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Represents central directory header record described in the
    /// ZIP File Format Specification v6.3.3, #4.3.12.
    /// </summary>
    internal class CentralDirectoryHeader : FileHeaderBase, ISpecData
    {
        /// <summary>
        /// Central directory header signature.
        /// </summary>
        public const uint Signature = 0x02014b50;

        /// <summary>
        /// Size of the data block without signature and variable size fields.
        /// </summary>
        public const int StaticBlockLength = 42;

        /// <summary>
        /// Gets or sets byte which indicates the ZIP specification version
        /// supported by the software used to encode the file.
        /// </summary>
        public byte VersionMadeBy
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets byte which indicates the compatibility
        /// of the file attribute information.
        /// </summary>
        public byte OsCompatibility
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets disk number start.
        /// </summary>
        public ushort DiskNumberStart
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets internal file attributes.
        /// </summary>
        public ushort InternalAttributes
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets external file attributes.
        /// </summary>
        public uint ExternalAttributes
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets relative offset of local header.
        /// </summary>
        public uint LocalHeaderOffset
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets file comment.
        /// </summary>
        public byte[] FileComment
        {
            get;
            set;
        }

        /// <summary>
        /// Read data from the binary reader.
        /// </summary>
        /// <param name="reader">Binary reader to read data from.</param>
        /// <returns>True if success, otherwise false.</returns>
        public bool TryReadBlock(BinaryReader reader)
        {
            if (reader.ReadUInt32() != CentralDirectoryHeader.Signature)
            {
                return false;
            }

            this.VersionMadeBy = reader.ReadByte();
            this.OsCompatibility = reader.ReadByte();

            this.ReadFields(reader);

            int fileCommentLength = reader.ReadInt16();
            this.DiskNumberStart = reader.ReadUInt16();
            this.InternalAttributes = reader.ReadUInt16();
            this.ExternalAttributes = reader.ReadUInt32();
            this.LocalHeaderOffset = reader.ReadUInt32();

            this.ReadExtraData(reader);

            this.FileComment = reader.ReadBytes(fileCommentLength);

            return true;
        }

        /// <summary>
        /// Write data to the binary writer.
        /// </summary>
        /// <param name="writer">Binary writer to write data to.</param>
        public void WriteBlock(BinaryWriter writer)
        {
            writer.Write(CentralDirectoryHeader.Signature);

            writer.Write(this.VersionMadeBy);
            writer.Write(this.OsCompatibility);

            this.WriteFields(writer);

            writer.Write((ushort)(this.FileComment != null ? this.FileComment.Length : 0));
            writer.Write(this.DiskNumberStart);
            writer.Write(this.InternalAttributes);
            writer.Write(this.ExternalAttributes);
            writer.Write(this.LocalHeaderOffset);

            this.WriteExtraData(writer);

            if (this.FileComment != null)
            {
                writer.Write(this.FileComment);
            }
        }

        protected override FileHeaderInfoBase GetHeaderInfo()
        {
            CentralDirectoryHeaderInfo headerInfo = new CentralDirectoryHeaderInfo(this);

            return headerInfo;
        }
    }
}
