﻿using System;
using System.IO;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Represents data descriptor record described in the
    /// ZIP File Format Specification v6.3.3, #4.3.9.
    /// </summary>
    internal class DataDescriptor : DataDescriptorBase, ISpecData
    {
        /// <summary>
        /// Data descriptor header signature.
        /// </summary>
        public const uint Signature = 0x08074b50;

        /// <summary>
        /// Size of the data block without signature and variable size fields.
        /// </summary>
        public const int StaticBlockLength = 12;

        private bool useZip64 = false;

        /// <summary>
        /// Gets or sets compressed size.
        /// </summary>
        public ulong CompressedSizeZip64
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets uncompressed size.
        /// </summary>
        public ulong UncompressedSizeZip64
        {
            get;
            set;
        }

        /// <summary>
        /// Read data from the binary reader.
        /// </summary>
        /// <param name="reader">Binary reader to read data from.</param>
        /// <returns>True if success, otherwise false.</returns>
        public bool TryReadBlock(BinaryReader reader)
        {
            byte[] buffer = new byte[4];
            ZipHelper.ReadBytes(reader.BaseStream, buffer, buffer.Length);
            uint crcOrSignature = BitConverter.ToUInt32(buffer, 0);
            if (crcOrSignature == DataDescriptor.Signature)
            {
                this.ReadFields(reader);
            }
            else
            {
                this.Crc = crcOrSignature;
                this.ReadSize(reader);
            }

            return true;
        }

        /// <summary>
        /// Write data to the binary writer.
        /// </summary>
        /// <param name="writer">Binary writer to write data to.</param>
        public void WriteBlock(BinaryWriter writer)
        {
            writer.Write(DataDescriptor.Signature);

            if (!this.useZip64)
            {
                this.WriteFields(writer);
            }
            else
            {
                writer.Write(this.Crc);
                writer.Write(this.CompressedSizeZip64);
                writer.Write(this.UncompressedSizeZip64);
            }
        }

        /// <summary>
        /// Copy properties from the given file header to this object.
        /// </summary>
        /// <param name="fileHeader">File header to copy properties from.</param>
        internal static DataDescriptor FromFileHeader(FileHeaderBase fileHeader)
        {
            DataDescriptor descr = new DataDescriptor()
            {
                CompressedSize = fileHeader.CompressedSize,
                Crc = fileHeader.Crc,
                UncompressedSize = fileHeader.UncompressedSize
            };

            Zip64ExtraField zip64ExtraField = null;

            foreach (ExtraFieldBase ef in fileHeader.ExtraFields)
            {
                zip64ExtraField = ef as Zip64ExtraField;
                if (zip64ExtraField != null)
                {
                    break;
                }
            }

            if (zip64ExtraField != null)
            {
                descr.useZip64 = true;

                if (zip64ExtraField.CompressedSize.HasValue)
                {
                    descr.CompressedSizeZip64 = zip64ExtraField.CompressedSize.Value;
                }
                else
                {
                    descr.CompressedSizeZip64 = fileHeader.CompressedSize;
                }

                if (zip64ExtraField.OriginalSize.HasValue)
                {
                    descr.UncompressedSizeZip64 = zip64ExtraField.OriginalSize.Value;
                }
                else
                {
                    descr.UncompressedSizeZip64 = fileHeader.UncompressedSize;
                }
            }

            return descr;
        }
    }
}
