﻿using System;
using System.IO;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Represents end of central directory record described in the
    /// ZIP File Format Specification v6.3.3, #4.3.16.
    /// </summary>
    internal class EndOfCentralDirectoryRecord : ISpecData 
    {
        /// <summary>
        /// End of central directory signature.
        /// </summary>
        public const uint Signature = 0x06054b50;

        /// <summary>
        /// Size of the data block without signature and variable size fields.
        /// </summary>
        public const int StaticBlockLength = 18;

        /// <summary>
        /// Gets or sets number of this disk.
        /// </summary>
        public ushort NumberOfThisDisk
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets number of the disk with the start of the central directory.
        /// </summary>
        public ushort NumberOfTheDiskWithTheStartOfTheCentralDirectory
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets total number of entries in the central directory on this disk.
        /// </summary>
        public ushort NumberOfEntriesInTheCentralDirectoryOnThisDisk
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets total number of entries in the central directory.
        /// </summary>
        public ushort NumberOfEntriesInTheCentralDirectory
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets size of the central directory.
        /// </summary>
        public uint SizeOfCentralDirectory
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets offset of start of central directory with respect to the starting disk number.
        /// </summary>
        public uint OffsetOfStartOfCentralDirectoryWithRespectToTheStartingDiskNumber
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets .ZIP file comment.
        /// </summary>
        public byte[] ArchiveComment
        {
            get;
            set;
        }

        /// <summary>
        /// Read data from the binary reader.
        /// </summary>
        /// <param name="reader">Binary reader to read data from.</param>
        /// <returns>true if success, otherwise false.</returns>
        public bool TryReadBlock(BinaryReader reader)
        {
            if (reader.ReadUInt32() != EndOfCentralDirectoryRecord.Signature)
            {
                return false;
            } 

            this.NumberOfThisDisk = reader.ReadUInt16();
            this.NumberOfTheDiskWithTheStartOfTheCentralDirectory = reader.ReadUInt16();
            this.NumberOfEntriesInTheCentralDirectoryOnThisDisk = reader.ReadUInt16();
            this.NumberOfEntriesInTheCentralDirectory = reader.ReadUInt16();
            this.SizeOfCentralDirectory = reader.ReadUInt32();
            this.OffsetOfStartOfCentralDirectoryWithRespectToTheStartingDiskNumber = reader.ReadUInt32();
            this.ArchiveComment = reader.ReadBytes((int)reader.ReadUInt16());

            return true;
        }

        /// <summary>
        /// Write data to the binary writer.
        /// </summary>
        /// <param name="writer">Binary writer to write data to.</param>
        public void WriteBlock(BinaryWriter writer)
        {
            writer.Write(EndOfCentralDirectoryRecord.Signature);
            writer.Write(this.NumberOfThisDisk);
            writer.Write(this.NumberOfTheDiskWithTheStartOfTheCentralDirectory);
            writer.Write(this.NumberOfEntriesInTheCentralDirectoryOnThisDisk);
            writer.Write(this.NumberOfEntriesInTheCentralDirectory);
            writer.Write(this.SizeOfCentralDirectory);
            writer.Write(this.OffsetOfStartOfCentralDirectoryWithRespectToTheStartingDiskNumber);
            ushort commentLength = (this.ArchiveComment != null) ? 
                (ushort)this.ArchiveComment.Length : (ushort)0;
            writer.Write(commentLength);
            if (this.ArchiveComment != null)
            {
                writer.Write(this.ArchiveComment);
            }
        }
    }
}
