﻿using System;
using System.Collections.Generic;
using System.IO;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Represents base class for extra field described in the
    /// ZIP File Format Specification v6.3.3, #4.5.2.
    /// </summary>
    internal abstract class ExtraFieldBase
    {
        /// <summary>
        /// Gets known extra field type.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
        public ExtraFieldType ExtraFieldType
        {
            get
            {
                return GetExtraFieldType(this.HeaderId);
            }
        }

        /// <summary>
        /// Gets extra field type (Header ID).
        /// </summary>
        protected abstract ushort HeaderId
        {
            get;
        }

        /// <summary>
        /// Gets extra field collection.
        /// </summary>
        /// <param name="headerInfo">The header info.</param>
        /// <returns>IEnumerable of ExtraFieldBase instances.</returns>
        public static IEnumerable<ExtraFieldBase> GetExtraFields(FileHeaderInfoBase headerInfo)
        {
            MemoryStream stream = new MemoryStream(headerInfo.ExtraFieldsData);
            using (BinaryReader reader = new BinaryReader(stream))
            {
                while (stream.Position < stream.Length)
                {
                    ExtraFieldBase field = GetExtraField(headerInfo, reader);
                    yield return field;
                }
            }
        }

        /// <summary>
        /// Gets extra field data.
        /// </summary>
        /// <param name="fields">Extra field collection.</param>
        /// <returns>Extra field data.</returns>
        public static byte[] GetExtraFieldsData(IEnumerable<ExtraFieldBase> fields)
        {
            byte[] fieldsData;
            using (MemoryStream stream = new MemoryStream())
            {
                using (BinaryWriter writer = new BinaryWriter(stream))
                {
                    foreach (ExtraFieldBase field in fields)
                    {
                        writer.Write(field.HeaderId);
                        byte[] data = field.GetBlock();
                        writer.Write((ushort)data.Length);
                        writer.Write(data);
                    }

                    fieldsData = stream.ToArray();
                }
            }

            return fieldsData;
        }

        /// <summary>
        /// Should implement parsing of extra field data.
        /// </summary>
        /// <param name="extraData">Extra field data.</param>
        protected abstract void ParseBlock(byte[] extraData);

        /// <summary>
        /// Gets extra field data.
        /// </summary>
        /// <returns>Byte array of extra field data.</returns>
        protected abstract byte[] GetBlock();

        private static ExtraFieldBase GetExtraField(FileHeaderInfoBase headerInfo, BinaryReader reader)
        {
            ushort headerId = reader.ReadUInt16();
            ExtraFieldType extraFieldType = GetExtraFieldType(headerId);
            ushort length = reader.ReadUInt16();
            byte[] extraData = reader.ReadBytes(length);
            ExtraFieldBase field;

            switch (extraFieldType)
            {
                case ExtraFieldType.AesEncryption:
                    field = new AesEncryptionExtraField();
                    break;

                case ExtraFieldType.Zip64:
                    field = new Zip64ExtraField(headerInfo);
                    break;

                default:
                    field = new UnknownExtraField(headerId);
                    break;
            }

            field.ParseBlock(extraData);
            return field;
        }

        private static ExtraFieldType GetExtraFieldType(ushort headerID)
        {
            if (Enum.IsDefined(typeof(ExtraFieldType), (int)headerID))
            {
                return (ExtraFieldType)headerID;
            }

            return Zip.ExtraFieldType.Unknown;
        }
    }
}
