﻿using System;
using System.IO;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Represents Zip64 Extended Information Extra Field described in the
    /// ZIP File Format Specification v6.3.3, #4.5.3.
    /// </summary>
    internal class Zip64ExtraField : ExtraFieldBase
    {
        private readonly FileHeaderInfoBase headerInfo;

        internal Zip64ExtraField(FileHeaderInfoBase headerInfo)
        {
            this.headerInfo = headerInfo;
        }

        /// <summary>
        /// Gets or sets original uncompressed file size.
        /// </summary>
        public ulong? OriginalSize
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets size of compressed data.
        /// </summary>
        public ulong? CompressedSize
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets offset of local header record.
        /// </summary>
        public ulong? RelativeHeaderOffset
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets number of the disk on which this file starts.
        /// </summary>
        public uint? DiskStartNumber
        {
            get;
            set;
        }

        /// <summary>
        /// Gets extra field type (Header ID).
        /// </summary>
        protected override ushort HeaderId
        {
            get
            {
                return (ushort)ExtraFieldType.Zip64;
            }
        }

        /// <summary>
        /// Gets extra field data.
        /// </summary>
        /// <returns>Byte array of extra field data.</returns>
        protected override byte[] GetBlock()
        {
            byte[] fieldsData;
            using (MemoryStream stream = new MemoryStream())
            {
                using (BinaryWriter writer = new BinaryWriter(stream))
                {
                    if (this.OriginalSize != null)
                    {
                        writer.Write((ulong)this.OriginalSize);
                    }
                    if (this.CompressedSize != null)
                    {
                        writer.Write((ulong)this.CompressedSize);
                    }
                    if (this.RelativeHeaderOffset != null)
                    {
                        writer.Write((ulong)this.RelativeHeaderOffset);
                    }
                    if (this.DiskStartNumber != null)
                    {
                        writer.Write((uint)this.DiskStartNumber);
                    }

                    fieldsData = stream.ToArray();
                }
            }

            return fieldsData;
        }

        /// <summary>
        /// Implements parsing of extra field data.
        /// </summary>
        /// <param name="extraData">Extra field data.</param>
        protected override void ParseBlock(byte[] extraData)
        {
            int offset = 0;

            if (this.headerInfo.UncompressedSizeOverflow)
            {
                this.OriginalSize = BitConverter.ToUInt64(extraData, offset);
                offset += 8;
            }

            if (this.headerInfo.CompressedSizeOverflow)
            {
                this.CompressedSize = BitConverter.ToUInt64(extraData, offset);
                offset += 8;
            }

            CentralDirectoryHeaderInfo centralDirectoryHeaderInfo = this.headerInfo as CentralDirectoryHeaderInfo;
            if (centralDirectoryHeaderInfo != null && centralDirectoryHeaderInfo.LocalHeaderOffsetOverflow)
            {
                this.RelativeHeaderOffset = BitConverter.ToUInt64(extraData, offset);
                offset += 8;
            }

            if (extraData.Length >= offset + 24)
            {
                this.DiskStartNumber = BitConverter.ToUInt32(extraData, 24);
            }
        }
    }
}