﻿using System;
using System.Collections.Generic;
using System.IO;

#if NETSTANDARD
namespace Telerik.Zip
#else
namespace Telerik.Windows.Zip
#endif
{
    /// <summary>
    /// Represents file header base class for
    /// the local file header and central directory header
    /// which are described in the ZIP File Format Specification v6.3.3, #4.3.7 and #4.3.12.
    /// </summary>
    internal abstract class FileHeaderBase : DataDescriptorBase
    {
        private ushort extraFieldLength;
        private ushort fileNameLength;

        public FileHeaderBase()
        {
            this.ExtraFields = new List<ExtraFieldBase>();
        }

        /// <summary>
        /// Gets or sets version needed to extract.
        /// </summary>
        public ushort VersionNeededToExtract
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets general purpose bit flag.
        /// </summary>
        public ushort GeneralPurposeBitFlag
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets compression method.
        /// </summary>
        public ushort CompressionMethod
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets last modification file date and time.
        /// </summary>
        public uint FileTime
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets file name.
        /// </summary>
        public byte[] FileName
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets extra fields data.
        /// </summary>
        /// <value>The extra fields data.</value>
        public byte[] ExtraFieldsData
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets list of extra fields.
        /// </summary>
        public List<ExtraFieldBase> ExtraFields
        {
            get;
            set;
        }

        /// <summary>
        /// Copy properties from the given file header to this object.
        /// </summary>
        /// <param name="fileHeader">File header to copy properties from.</param>
        internal void FromFileHeader(FileHeaderBase fileHeader)
        {
            this.CompressedSize = fileHeader.CompressedSize;
            this.CompressionMethod = fileHeader.CompressionMethod;
            this.UncompressedSize = fileHeader.UncompressedSize;
            this.Crc = fileHeader.Crc;
            
            if (fileHeader.ExtraFieldsData != null)
            {
                this.ExtraFieldsData = new byte[fileHeader.ExtraFieldsData.Length];
                Array.Copy(fileHeader.ExtraFieldsData, this.ExtraFieldsData, fileHeader.ExtraFieldsData.Length);
                this.ExtraFields = new List<ExtraFieldBase>(ExtraFieldBase.GetExtraFields(this.GetHeaderInfo()));
            }

            if (fileHeader.FileName != null)
            {
                this.FileName = new byte[fileHeader.FileName.Length];
                Array.Copy(fileHeader.FileName, this.FileName, fileHeader.FileName.Length);
            }

            this.FileTime = fileHeader.FileTime;
            this.GeneralPurposeBitFlag = fileHeader.GeneralPurposeBitFlag;
            this.VersionNeededToExtract = fileHeader.VersionNeededToExtract;
        }

        protected abstract FileHeaderInfoBase GetHeaderInfo();

        /// <summary>
        /// Read data from the binary reader.
        /// </summary>
        /// <param name="reader">Binary reader to read data from.</param>
        protected override void ReadFields(BinaryReader reader)
        {
            this.VersionNeededToExtract = reader.ReadUInt16();
            this.GeneralPurposeBitFlag = reader.ReadUInt16();
            this.CompressionMethod = reader.ReadUInt16();
            this.FileTime = reader.ReadUInt32();

            base.ReadFields(reader);

            this.fileNameLength = reader.ReadUInt16();
            this.extraFieldLength = reader.ReadUInt16();
        }

        /// <summary>
        /// Read data from the binary reader.
        /// </summary>
        /// <param name="reader">Binary reader to read data from.</param>
        protected void ReadExtraData(BinaryReader reader)
        {
            this.FileName = reader.ReadBytes(this.fileNameLength);
            this.ExtraFieldsData = reader.ReadBytes(this.extraFieldLength);
            this.ExtraFields = new List<ExtraFieldBase>(ExtraFieldBase.GetExtraFields(this.GetHeaderInfo()));  
        }

        /// <summary>
        /// Write data to the binary writer.
        /// </summary>
        /// <param name="writer">Binary writer to write data to.</param>
        protected override void WriteFields(BinaryWriter writer)
        {
            writer.Write(this.VersionNeededToExtract);
            writer.Write(this.GeneralPurposeBitFlag);
            writer.Write(this.CompressionMethod);
            writer.Write(this.FileTime);

            base.WriteFields(writer);

            writer.Write((ushort)this.FileName.Length);

            if (this.ExtraFields != null)
            {
                this.ExtraFieldsData = ExtraFieldBase.GetExtraFieldsData(this.ExtraFields);
                writer.Write((ushort)this.ExtraFieldsData.Length);
            }
            else
            {
                writer.Write((ushort)0);
            }
        }

        /// <summary>
        /// Write data to the binary writer.
        /// </summary>
        /// <param name="writer">Binary writer to write data to.</param>
        protected void WriteExtraData(BinaryWriter writer)
        {
            writer.Write(this.FileName);
            if (this.ExtraFieldsData != null)
            {
                writer.Write(this.ExtraFieldsData);
            }
        }
    }
}
